// This file is @generated by crates/typos-dict/tests/codegen.rs
#![allow(clippy::unreadable_literal)]
#![allow(unreachable_pub)]

pub static WORD: dictgen::Map<dictgen::InsensitiveStr<'static>, &[&str]> = dictgen::Map {
    map: ::phf::Map {
        key: 16287231350648472473,
        disps: &[
            (0, 0),
            (0, 0),
            (0, 619),
            (0, 141),
            (0, 265),
            (0, 19),
            (0, 77),
            (0, 546),
            (0, 432),
            (0, 2),
            (0, 7),
            (0, 125),
            (0, 2358),
            (0, 582),
            (0, 162),
            (0, 15),
            (0, 693),
            (0, 142),
            (0, 0),
            (0, 27),
            (0, 27),
            (0, 2),
            (0, 29),
            (0, 174),
            (0, 6),
            (0, 0),
            (0, 191),
            (0, 1),
            (0, 243),
            (0, 131),
            (0, 129),
            (0, 6),
            (0, 1497),
            (0, 44),
            (0, 0),
            (0, 256),
            (0, 980),
            (0, 10),
            (0, 760),
            (0, 2),
            (0, 1255),
            (0, 1),
            (0, 25),
            (0, 25),
            (0, 1051),
            (0, 142),
            (0, 2),
            (0, 0),
            (0, 1422),
            (0, 108),
            (0, 1023),
            (0, 26),
            (0, 0),
            (0, 126),
            (0, 103),
            (0, 87),
            (0, 171),
            (0, 29),
            (0, 0),
            (0, 96),
            (0, 153),
            (0, 113),
            (0, 1),
            (0, 122),
            (0, 1194),
            (0, 6),
            (0, 17),
            (0, 814),
            (0, 299),
            (0, 2),
            (0, 149),
            (0, 182),
            (0, 233),
            (0, 282),
            (0, 5),
            (0, 223),
            (0, 7),
            (0, 297),
            (0, 203),
            (0, 16),
            (0, 2),
            (0, 0),
            (0, 6),
            (0, 7),
            (0, 4),
            (0, 139),
            (0, 59),
            (0, 13),
            (0, 57),
            (0, 21),
            (0, 0),
            (0, 52),
            (0, 425),
            (0, 0),
            (0, 2),
            (0, 8),
            (0, 3),
            (0, 5),
            (0, 5),
            (0, 77),
            (0, 101),
            (0, 1),
            (0, 9),
            (0, 2),
            (0, 0),
            (0, 98),
            (0, 1),
            (0, 0),
            (0, 12),
            (0, 3169),
            (0, 101),
            (0, 5),
            (0, 10),
            (0, 1),
            (0, 0),
            (0, 3),
            (0, 370),
            (0, 0),
            (0, 20),
            (0, 5),
            (0, 236),
            (0, 10),
            (0, 36),
            (0, 561),
            (0, 776),
            (0, 1),
            (0, 476),
            (0, 12),
            (0, 12),
            (0, 27),
            (0, 56),
            (0, 38),
            (0, 0),
            (0, 391),
            (0, 30),
            (0, 1321),
            (0, 354),
            (0, 34),
            (0, 2),
            (0, 0),
            (0, 15),
            (0, 581),
            (0, 1),
            (0, 232),
            (0, 623),
            (0, 446),
            (0, 4),
            (0, 22),
            (0, 336),
            (0, 320),
            (0, 116),
            (0, 26),
            (0, 91),
            (0, 2),
            (0, 3),
            (0, 0),
            (0, 2),
            (0, 13),
            (0, 619),
            (0, 0),
            (0, 289),
            (0, 417),
            (0, 0),
            (0, 0),
            (0, 5),
            (0, 149),
            (0, 0),
            (0, 2),
            (0, 734),
            (0, 202),
            (0, 15),
            (0, 602),
            (0, 936),
            (0, 106),
            (0, 877),
            (0, 132),
            (0, 12),
            (0, 5),
            (0, 5),
            (0, 0),
            (0, 90),
            (0, 31),
            (0, 0),
            (0, 0),
            (0, 14),
            (0, 742),
            (0, 83),
            (0, 1335),
            (0, 8),
            (0, 2),
            (0, 133),
            (0, 9),
            (0, 105),
            (0, 1116),
            (0, 518),
            (0, 1173),
            (0, 0),
            (0, 19),
            (0, 122),
            (0, 160),
            (0, 341),
            (0, 216),
            (0, 4),
            (0, 232),
            (0, 27),
            (0, 401),
            (0, 14),
            (0, 84),
            (0, 372),
            (0, 1569),
            (0, 272),
            (0, 4),
            (0, 7),
            (0, 32),
            (0, 0),
            (0, 1),
            (0, 0),
            (0, 229),
            (0, 17),
            (0, 13),
            (0, 5),
            (0, 58),
            (0, 183),
            (0, 458),
            (0, 91),
            (0, 27),
            (0, 1040),
            (0, 29),
            (0, 625),
            (0, 83),
            (0, 478),
            (0, 35),
            (0, 2),
            (0, 147),
            (0, 883),
            (0, 357),
            (0, 16),
            (0, 2302),
            (0, 0),
            (0, 8),
            (0, 550),
            (0, 656),
            (0, 11),
            (0, 305),
            (0, 619),
            (0, 1),
            (0, 175),
            (0, 552),
            (0, 19),
            (0, 15),
            (0, 194),
            (0, 928),
            (0, 298),
            (0, 172),
            (0, 123),
            (0, 130),
            (0, 1143),
            (0, 3),
            (0, 57),
            (0, 0),
            (0, 18),
            (0, 0),
            (0, 1203),
            (0, 102),
            (0, 0),
            (0, 376),
            (0, 8),
            (0, 1001),
            (0, 32),
            (0, 1199),
            (0, 39),
            (0, 0),
            (0, 83),
            (0, 40),
            (0, 0),
            (0, 18),
            (0, 1163),
            (0, 165),
            (0, 206),
            (0, 1481),
            (0, 5),
            (0, 17),
            (0, 23),
            (0, 4),
            (0, 3),
            (0, 15),
            (0, 104),
            (0, 1),
            (0, 31),
            (0, 19),
            (0, 678),
            (0, 9),
            (0, 34),
            (0, 0),
            (0, 19),
            (0, 716),
            (0, 382),
            (0, 169),
            (0, 269),
            (0, 8),
            (0, 210),
            (0, 9),
            (0, 517),
            (0, 5),
            (0, 1373),
            (0, 85),
            (0, 283),
            (0, 50),
            (0, 2708),
            (0, 80),
            (0, 1),
            (0, 349),
            (0, 992),
            (0, 444),
            (0, 6),
            (0, 301),
            (0, 826),
            (0, 1368),
            (0, 78),
            (0, 1),
            (0, 219),
            (0, 17),
            (0, 1),
            (0, 35),
            (0, 1057),
            (0, 1),
            (0, 797),
            (0, 163),
            (0, 171),
            (0, 77),
            (0, 54),
            (0, 8),
            (0, 130),
            (0, 357),
            (0, 1),
            (0, 72),
            (0, 184),
            (0, 1760),
            (0, 391),
            (0, 60),
            (0, 0),
            (0, 1510),
            (0, 814),
            (0, 132),
            (0, 86),
            (0, 92),
            (0, 17),
            (0, 1),
            (0, 3),
            (0, 7),
            (0, 8),
            (0, 66),
            (0, 5),
            (0, 423),
            (0, 5),
            (0, 52),
            (0, 7),
            (0, 778),
            (0, 14),
            (0, 42),
            (0, 0),
            (0, 46),
            (0, 233),
            (0, 0),
            (0, 12),
            (0, 503),
            (0, 19),
            (0, 69),
            (0, 23),
            (0, 25),
            (0, 0),
            (0, 9),
            (0, 2),
            (0, 1),
            (0, 13),
            (0, 1),
            (0, 449),
            (0, 200),
            (0, 76),
            (0, 135),
            (0, 0),
            (0, 22),
            (0, 0),
            (0, 428),
            (0, 103),
            (0, 29),
            (0, 3),
            (0, 0),
            (0, 416),
            (0, 16),
            (0, 14),
            (0, 4),
            (0, 2808),
            (0, 24),
            (0, 106),
            (0, 10),
            (0, 16),
            (0, 193),
            (0, 7),
            (0, 7),
            (0, 30),
            (0, 891),
            (0, 6),
            (0, 2),
            (0, 1397),
            (0, 25),
            (0, 2),
            (0, 6),
            (0, 60),
            (0, 8),
            (0, 0),
            (0, 506),
            (0, 82),
            (0, 16),
            (0, 1),
            (0, 1),
            (0, 474),
            (0, 2),
            (0, 1),
            (0, 13),
            (0, 17),
            (0, 900),
            (0, 37),
            (0, 0),
            (0, 324),
            (0, 3),
            (0, 1),
            (0, 149),
            (0, 572),
            (0, 99),
            (0, 19),
            (0, 3),
            (0, 14),
            (0, 3),
            (0, 552),
            (0, 26),
            (0, 159),
            (0, 227),
            (0, 103),
            (0, 176),
            (0, 0),
            (0, 1),
            (0, 574),
            (0, 392),
            (0, 45),
            (0, 5),
            (0, 562),
            (0, 444),
            (0, 10),
            (0, 133),
            (0, 784),
            (0, 10),
            (0, 12),
            (0, 1249),
            (0, 0),
            (0, 912),
            (0, 57),
            (0, 238),
            (0, 341),
            (0, 360),
            (0, 296),
            (0, 444),
            (0, 14),
            (0, 232),
            (0, 194),
            (0, 304),
            (0, 15),
            (0, 112),
            (0, 442),
            (0, 117),
            (0, 3),
            (0, 2),
            (0, 4),
            (0, 80),
            (0, 1),
            (0, 3),
            (0, 21),
            (0, 25),
            (0, 0),
            (0, 0),
            (0, 275),
            (0, 156),
            (0, 499),
            (0, 292),
            (0, 5),
            (0, 0),
            (0, 1521),
            (0, 1),
            (0, 4),
            (0, 221),
            (0, 79),
            (0, 302),
            (0, 175),
            (0, 130),
            (0, 1),
            (0, 3445),
            (0, 55),
            (0, 6),
            (0, 4),
            (0, 10),
            (0, 70),
            (0, 42),
            (0, 106),
            (0, 156),
            (0, 552),
            (0, 5),
            (0, 7),
            (0, 1021),
            (0, 187),
            (0, 199),
            (0, 507),
            (0, 273),
            (0, 5),
            (0, 0),
            (0, 236),
            (0, 9),
            (0, 2464),
            (0, 12),
            (0, 1011),
            (0, 57),
            (0, 10),
            (0, 1),
            (0, 247),
            (0, 2092),
            (0, 139),
            (0, 545),
            (0, 236),
            (0, 217),
            (0, 7),
            (0, 8),
            (0, 98),
            (0, 41),
            (0, 4),
            (0, 603),
            (0, 1476),
            (0, 493),
            (0, 55),
            (0, 3),
            (0, 29),
            (0, 3),
            (0, 904),
            (0, 670),
            (0, 6),
            (0, 122),
            (0, 2),
            (0, 0),
            (0, 93),
            (0, 9),
            (0, 313),
            (0, 154),
            (0, 1622),
            (0, 9),
            (0, 10),
            (0, 888),
            (0, 432),
            (0, 107),
            (0, 75),
            (0, 45),
            (0, 1164),
            (0, 0),
            (0, 89),
            (0, 1),
            (0, 0),
            (0, 2),
            (0, 130),
            (0, 0),
            (0, 0),
            (0, 87),
            (0, 1),
            (0, 59),
            (0, 21),
            (0, 162),
            (0, 22),
            (0, 2),
            (0, 13),
            (0, 35),
            (0, 5),
            (0, 372),
            (0, 69),
            (0, 30),
            (0, 37),
            (0, 780),
            (0, 0),
            (0, 85),
            (0, 99),
            (0, 0),
            (0, 7),
            (0, 7),
            (0, 32),
            (0, 4),
            (0, 197),
            (0, 278),
            (0, 34),
            (0, 393),
            (0, 23),
            (0, 146),
            (0, 9),
            (0, 4),
            (0, 3),
            (0, 8),
            (0, 5),
            (0, 1),
            (0, 9),
            (0, 1),
            (0, 124),
            (0, 245),
            (0, 224),
            (0, 10),
            (0, 1061),
            (0, 202),
            (0, 2),
            (0, 175),
            (0, 1),
            (0, 487),
            (0, 394),
            (0, 1536),
            (0, 21),
            (0, 76),
            (0, 6),
            (0, 60),
            (0, 19),
            (0, 12),
            (0, 117),
            (0, 1),
            (0, 2),
            (0, 3),
            (0, 18),
            (0, 66),
            (0, 27),
            (0, 507),
            (0, 173),
            (0, 180),
            (0, 63),
            (0, 821),
            (0, 165),
            (0, 3),
            (0, 2),
            (0, 79),
            (0, 0),
            (0, 77),
            (0, 1),
            (0, 5),
            (0, 469),
            (0, 227),
            (0, 18),
            (0, 0),
            (0, 74),
            (0, 853),
            (0, 60),
            (0, 0),
            (0, 19),
            (0, 79),
            (0, 98),
            (0, 0),
            (0, 388),
            (0, 85),
            (0, 197),
            (0, 4),
            (0, 33),
            (0, 1728),
            (0, 1),
            (0, 22),
            (0, 2),
            (0, 244),
            (0, 1094),
            (0, 45),
            (0, 27),
            (0, 62),
            (0, 136),
            (0, 419),
            (0, 39),
            (0, 1),
            (0, 762),
            (0, 5),
            (0, 0),
            (0, 398),
            (0, 10),
            (0, 213),
            (0, 25),
            (0, 17),
            (0, 0),
            (0, 32),
            (0, 42),
            (0, 19),
            (0, 456),
            (0, 7),
            (0, 0),
            (0, 166),
            (0, 1),
            (0, 23),
            (0, 3),
            (0, 436),
            (0, 251),
            (0, 38),
            (0, 19),
            (0, 102),
            (0, 5),
            (0, 69),
            (0, 34),
            (0, 104),
            (0, 1),
            (0, 259),
            (0, 69),
            (0, 0),
            (0, 485),
            (0, 16),
            (0, 3),
            (0, 2),
            (0, 82),
            (0, 1),
            (0, 1001),
            (0, 10),
            (0, 1),
            (0, 3),
            (0, 0),
            (0, 28),
            (0, 1),
            (0, 19),
            (0, 0),
            (0, 249),
            (0, 53),
            (0, 0),
            (0, 0),
            (0, 43),
            (0, 1),
            (0, 1808),
            (0, 32),
            (0, 45),
            (0, 2),
            (0, 1483),
            (0, 101),
            (0, 1),
            (0, 32),
            (0, 13),
            (0, 200),
            (0, 356),
            (0, 84),
            (0, 23),
            (0, 331),
            (0, 699),
            (0, 1),
            (0, 654),
            (0, 19),
            (0, 2),
            (0, 969),
            (0, 0),
            (0, 13),
            (0, 175),
            (0, 97),
            (0, 149),
            (0, 10),
            (0, 1308),
            (0, 21),
            (0, 126),
            (0, 34),
            (0, 267),
            (0, 8),
            (0, 1),
            (0, 604),
            (0, 0),
            (0, 2),
            (0, 29),
            (0, 25),
            (0, 5),
            (0, 66),
            (0, 110),
            (0, 1035),
            (0, 0),
            (0, 34),
            (0, 237),
            (0, 23),
            (0, 12),
            (0, 0),
            (0, 5),
            (0, 1),
            (0, 3),
            (0, 657),
            (0, 26),
            (0, 617),
            (0, 44),
            (0, 0),
            (0, 208),
            (0, 14),
            (0, 1627),
            (0, 1686),
            (0, 429),
            (0, 0),
            (0, 136),
            (0, 519),
            (0, 461),
            (0, 1176),
            (0, 23),
            (0, 2),
            (0, 4),
            (0, 231),
            (0, 305),
            (0, 5),
            (0, 808),
            (0, 2023),
            (0, 0),
            (0, 400),
            (0, 0),
            (0, 0),
            (0, 110),
            (0, 187),
            (0, 2),
            (0, 0),
            (0, 1337),
            (0, 37),
            (0, 103),
            (0, 6),
            (0, 1591),
            (0, 17),
            (0, 11),
            (0, 5),
            (0, 585),
            (0, 1),
            (0, 0),
            (0, 318),
            (0, 10),
            (0, 525),
            (0, 1572),
            (0, 254),
            (0, 8),
            (0, 1),
            (0, 72),
            (0, 93),
            (0, 127),
            (0, 2414),
            (0, 104),
            (0, 244),
            (0, 6),
            (0, 877),
            (0, 266),
            (0, 34),
            (0, 55),
            (0, 7),
            (0, 1),
            (0, 23),
            (0, 18),
            (0, 967),
            (0, 6),
            (0, 149),
            (0, 162),
            (0, 23),
            (0, 1),
            (0, 107),
            (0, 13),
            (0, 552),
            (0, 31),
            (0, 6),
            (0, 9),
            (0, 561),
            (0, 257),
            (0, 1),
            (0, 6),
            (0, 191),
            (0, 68),
            (0, 24),
            (0, 653),
            (0, 3),
            (0, 2049),
            (0, 8),
            (0, 613),
            (0, 195),
            (0, 76),
            (0, 5),
            (0, 0),
            (0, 0),
            (0, 387),
            (0, 23),
            (0, 9),
            (0, 29),
            (0, 0),
            (0, 481),
            (0, 1384),
            (0, 1),
            (0, 0),
            (0, 34),
            (0, 28),
            (0, 0),
            (0, 109),
            (0, 926),
            (0, 142),
            (0, 13),
            (0, 128),
            (0, 214),
            (0, 1),
            (0, 203),
            (0, 0),
            (0, 45),
            (0, 280),
            (0, 172),
            (0, 43),
            (0, 1),
            (0, 3),
            (0, 107),
            (0, 2446),
            (0, 731),
            (0, 134),
            (0, 7),
            (0, 8),
            (0, 39),
            (0, 48),
            (0, 156),
            (0, 23),
            (0, 49),
            (0, 0),
            (0, 0),
            (0, 41),
            (0, 661),
            (0, 2047),
            (0, 1),
            (0, 101),
            (0, 26),
            (0, 4),
            (0, 1),
            (0, 36),
            (0, 281),
            (0, 220),
            (0, 6),
            (0, 68),
            (0, 78),
            (0, 5),
            (0, 1),
            (0, 254),
            (0, 1),
            (0, 464),
            (0, 388),
            (0, 20),
            (0, 30),
            (0, 120),
            (0, 102),
            (0, 68),
            (0, 543),
            (0, 50),
            (0, 131),
            (0, 3),
            (0, 0),
            (0, 615),
            (0, 104),
            (0, 1327),
            (0, 361),
            (0, 351),
            (0, 510),
            (0, 462),
            (0, 96),
            (0, 520),
            (0, 2),
            (0, 76),
            (0, 3),
            (0, 102),
            (0, 6),
            (0, 118),
            (0, 10),
            (0, 8),
            (0, 65),
            (0, 125),
            (0, 22),
            (0, 1),
            (0, 5),
            (0, 113),
            (0, 217),
            (0, 234),
            (0, 92),
            (0, 0),
            (0, 1),
            (0, 14),
            (0, 1005),
            (0, 731),
            (0, 1),
            (0, 150),
            (0, 0),
            (0, 63),
            (0, 33),
            (0, 5),
            (0, 18),
            (0, 673),
            (0, 14),
            (0, 108),
            (0, 830),
            (0, 0),
            (0, 208),
            (0, 4),
            (0, 0),
            (0, 38),
            (0, 385),
            (0, 12),
            (0, 280),
            (0, 1179),
            (0, 25),
            (0, 4),
            (0, 295),
            (0, 76),
            (0, 12),
            (0, 3),
            (0, 33),
            (0, 213),
            (0, 11),
            (0, 1),
            (0, 61),
            (0, 204),
            (0, 12),
            (0, 810),
            (0, 3),
            (0, 0),
            (0, 0),
            (0, 313),
            (0, 13),
            (0, 757),
            (0, 5),
            (0, 0),
            (0, 570),
            (0, 392),
            (0, 343),
            (0, 6),
            (0, 9),
            (0, 772),
            (0, 255),
            (0, 595),
            (0, 130),
            (0, 4),
            (0, 100),
            (0, 1708),
            (0, 7),
            (0, 63),
            (0, 182),
            (0, 2),
            (0, 5),
            (0, 41),
            (0, 46),
            (0, 60),
            (0, 2),
            (0, 0),
            (0, 548),
            (0, 6),
            (0, 346),
            (0, 61),
            (0, 31),
            (0, 56),
            (0, 2719),
            (0, 1),
            (0, 0),
            (0, 2),
            (0, 59),
            (0, 384),
            (0, 66),
            (0, 15),
            (0, 105),
            (0, 961),
            (0, 2),
            (0, 2),
            (0, 106),
            (0, 14),
            (0, 75),
            (0, 1664),
            (0, 188),
            (0, 795),
            (0, 6),
            (0, 242),
            (0, 87),
            (0, 566),
            (0, 5),
            (0, 15),
            (0, 28),
            (0, 153),
            (0, 565),
            (0, 0),
            (0, 2),
            (0, 2),
            (0, 122),
            (0, 1434),
            (0, 3),
            (0, 1),
            (0, 351),
            (0, 124),
            (0, 20),
            (0, 40),
            (0, 586),
            (0, 20),
            (0, 189),
            (0, 1),
            (0, 4),
            (0, 8),
            (0, 77),
            (0, 0),
            (0, 267),
            (0, 35),
            (0, 316),
            (0, 39),
            (0, 0),
            (0, 7),
            (0, 0),
            (0, 21),
            (0, 9),
            (0, 3),
            (0, 4),
            (0, 353),
            (0, 2),
            (0, 10),
            (0, 15),
            (0, 18),
            (0, 629),
            (0, 173),
            (0, 36),
            (0, 899),
            (0, 2),
            (0, 0),
            (0, 0),
            (0, 19),
            (0, 9),
            (0, 0),
            (0, 309),
            (0, 0),
            (0, 93),
            (0, 66),
            (0, 83),
            (0, 538),
            (0, 264),
            (0, 0),
            (0, 1),
            (0, 2),
            (0, 653),
            (0, 0),
            (0, 8),
            (0, 21),
            (0, 105),
            (0, 10),
            (0, 31),
            (0, 82),
            (0, 133),
            (0, 34),
            (0, 1),
            (0, 524),
            (0, 531),
            (0, 23),
            (0, 97),
            (0, 106),
            (0, 114),
            (0, 14),
            (0, 69),
            (0, 0),
            (0, 2),
            (0, 17),
            (0, 12),
            (0, 1094),
            (0, 12),
            (0, 74),
            (0, 368),
            (0, 184),
            (0, 502),
            (0, 167),
            (0, 65),
            (0, 168),
            (0, 7),
            (0, 66),
            (0, 216),
            (0, 518),
            (0, 31),
            (0, 6),
            (0, 3),
            (0, 10),
            (0, 748),
            (0, 450),
            (0, 6),
            (0, 35),
            (0, 0),
            (0, 355),
            (0, 634),
            (0, 3745),
            (0, 0),
            (0, 19),
            (0, 514),
            (0, 6),
            (0, 1),
            (0, 0),
            (0, 2),
            (0, 5),
            (0, 112),
            (0, 633),
            (0, 18),
            (0, 134),
            (0, 0),
            (0, 126),
            (0, 79),
            (0, 97),
            (0, 62),
            (0, 41),
            (0, 23),
            (0, 1),
            (0, 107),
            (0, 3),
            (0, 43),
            (0, 83),
            (0, 90),
            (0, 506),
            (0, 60),
            (0, 6),
            (0, 20),
            (0, 217),
            (0, 1),
            (0, 121),
            (0, 328),
            (0, 5),
            (0, 0),
            (0, 1247),
            (0, 103),
            (0, 0),
            (0, 166),
            (0, 62),
            (0, 49),
            (0, 20),
            (0, 5),
            (0, 100),
            (0, 23),
            (0, 6),
            (0, 0),
            (0, 65),
            (0, 69),
            (0, 19),
            (0, 465),
            (0, 2),
            (0, 2),
            (0, 55),
            (0, 269),
            (0, 388),
            (0, 3),
            (0, 0),
            (0, 106),
            (0, 181),
            (0, 0),
            (0, 118),
            (0, 87),
            (0, 0),
            (0, 74),
            (0, 1),
            (0, 118),
            (0, 55),
            (0, 13),
            (0, 84),
            (0, 120),
            (0, 543),
            (0, 582),
            (0, 74),
            (0, 157),
            (0, 15),
            (0, 23),
            (0, 54),
            (0, 18),
            (0, 191),
            (0, 8),
            (0, 81),
            (0, 1044),
            (0, 32),
            (0, 370),
            (0, 579),
            (0, 9),
            (0, 47),
            (0, 183),
            (0, 6),
            (0, 6),
            (0, 16),
            (0, 35),
            (0, 25),
            (0, 0),
            (0, 91),
            (0, 21),
            (0, 58),
            (0, 522),
            (0, 0),
            (0, 0),
            (0, 25),
            (0, 528),
            (0, 0),
            (0, 34),
            (0, 0),
            (0, 2),
            (0, 8),
            (0, 142),
            (0, 515),
            (0, 1),
            (0, 0),
            (0, 14),
            (0, 8),
            (0, 31),
            (0, 411),
            (0, 504),
            (0, 1038),
            (0, 8),
            (0, 997),
            (0, 10),
            (0, 18),
            (0, 426),
            (0, 137),
            (0, 646),
            (0, 10),
            (0, 311),
            (0, 271),
            (0, 68),
            (0, 2036),
            (0, 13),
            (0, 1),
            (0, 28),
            (0, 11),
            (0, 6),
            (0, 2401),
            (0, 796),
            (0, 1),
            (0, 11),
            (0, 9),
            (0, 22),
            (0, 1650),
            (0, 2),
            (0, 52),
            (0, 246),
            (0, 24),
            (0, 50),
            (0, 105),
            (0, 71),
            (0, 1083),
            (0, 22),
            (0, 12),
            (0, 0),
            (0, 17),
            (0, 129),
            (0, 1),
            (0, 20),
            (0, 38),
            (0, 36),
            (0, 343),
            (0, 255),
            (0, 376),
            (0, 620),
            (0, 1),
            (0, 76),
            (0, 1),
            (0, 129),
            (0, 7),
            (0, 1),
            (0, 850),
            (0, 16),
            (0, 396),
            (0, 106),
            (0, 0),
            (0, 17),
            (0, 569),
            (0, 111),
            (0, 13),
            (0, 5),
            (0, 39),
            (0, 115),
            (0, 99),
            (0, 320),
            (0, 2),
            (0, 181),
            (0, 343),
            (0, 0),
            (0, 10),
            (0, 103),
            (0, 59),
            (0, 85),
            (0, 3),
            (0, 692),
            (0, 891),
            (0, 4),
            (0, 12),
            (0, 131),
            (0, 9),
            (0, 621),
            (0, 193),
            (0, 89),
            (0, 0),
            (0, 142),
            (0, 956),
            (0, 210),
            (0, 34),
            (0, 908),
            (0, 0),
            (0, 4),
            (0, 0),
            (0, 71),
            (0, 1),
            (0, 1),
            (0, 10),
            (0, 0),
            (0, 529),
            (0, 0),
            (0, 47),
            (0, 31),
            (0, 275),
            (0, 25),
            (0, 41),
            (0, 11),
            (0, 1),
            (0, 38),
            (0, 0),
            (0, 34),
            (0, 0),
            (0, 11),
            (0, 64),
            (0, 38),
            (0, 56),
            (0, 587),
            (0, 457),
            (0, 190),
            (0, 1),
            (0, 10),
            (0, 571),
            (0, 8),
            (0, 1790),
            (0, 7),
            (0, 78),
            (0, 1),
            (0, 1),
            (0, 479),
            (0, 71),
            (0, 281),
            (0, 73),
            (0, 145),
            (0, 5),
            (0, 13),
            (0, 1191),
            (0, 21),
            (0, 7),
            (0, 3),
            (0, 123),
            (0, 235),
            (0, 570),
            (0, 3),
            (0, 658),
            (0, 102),
            (0, 204),
            (0, 4),
            (0, 0),
            (0, 14),
            (0, 86),
            (0, 1),
            (0, 0),
            (0, 1),
            (0, 14),
            (0, 283),
            (0, 3),
            (0, 26),
            (0, 3),
            (0, 56),
            (0, 438),
            (0, 3),
            (0, 15),
            (0, 705),
            (0, 110),
            (0, 2),
            (0, 685),
            (0, 1),
            (0, 128),
            (0, 235),
            (0, 5),
            (0, 3),
            (0, 29),
            (0, 369),
            (0, 1),
            (0, 9),
            (0, 133),
            (0, 33),
            (0, 65),
            (0, 1),
            (0, 41),
            (0, 4),
            (0, 28),
            (0, 214),
            (0, 5),
            (0, 36),
            (0, 867),
            (0, 11),
            (0, 54),
            (0, 8),
            (0, 15),
            (0, 11),
            (0, 0),
            (0, 325),
            (0, 1),
            (0, 29),
            (0, 62),
            (0, 180),
            (0, 3),
            (0, 73),
            (0, 36),
            (0, 229),
            (0, 729),
            (0, 335),
            (0, 3),
            (0, 26),
            (0, 0),
            (0, 139),
            (0, 183),
            (0, 240),
            (0, 2),
            (0, 230),
            (0, 1007),
            (0, 132),
            (0, 10),
            (0, 365),
            (0, 476),
            (0, 266),
            (0, 94),
            (0, 473),
            (0, 515),
            (0, 4),
            (0, 1),
            (0, 367),
            (0, 867),
            (0, 0),
            (0, 108),
            (0, 156),
            (0, 155),
            (0, 18),
            (0, 16),
            (0, 1233),
            (0, 123),
            (0, 145),
            (0, 516),
            (0, 447),
            (0, 274),
            (0, 97),
            (0, 8),
            (0, 8),
            (0, 37),
            (0, 1),
            (0, 3),
            (0, 10),
            (0, 169),
            (0, 41),
            (0, 386),
            (0, 23),
            (0, 98),
            (0, 25),
            (0, 939),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 29),
            (0, 690),
            (0, 3),
            (0, 1),
            (0, 112),
            (0, 11),
            (0, 417),
            (0, 148),
            (0, 2021),
            (0, 0),
            (0, 160),
            (0, 223),
            (0, 603),
            (0, 20),
            (0, 23),
            (0, 76),
            (0, 67),
            (0, 2453),
            (0, 6),
            (0, 522),
            (0, 10),
            (0, 54),
            (0, 340),
            (0, 6),
            (0, 18),
            (0, 406),
            (0, 1),
            (0, 2),
            (0, 34),
            (0, 47),
            (0, 138),
            (0, 4),
            (0, 850),
            (0, 0),
            (0, 15),
            (0, 13),
            (0, 131),
            (0, 254),
            (0, 1780),
            (0, 1592),
            (0, 87),
            (0, 143),
            (0, 0),
            (0, 309),
            (0, 80),
            (0, 2),
            (0, 32),
            (0, 260),
            (0, 322),
            (0, 52),
            (0, 57),
            (0, 108),
            (0, 17),
            (0, 287),
            (0, 464),
            (0, 0),
            (0, 536),
            (0, 49),
            (0, 0),
            (0, 21),
            (0, 8),
            (0, 1),
            (0, 9),
            (0, 0),
            (0, 344),
            (0, 27),
            (0, 32),
            (0, 2),
            (0, 63),
            (0, 1),
            (0, 10),
            (0, 24),
            (0, 25),
            (0, 41),
            (0, 29),
            (0, 703),
            (0, 335),
            (0, 787),
            (0, 4),
            (0, 372),
            (0, 118),
            (0, 1),
            (0, 0),
            (0, 1013),
            (0, 0),
            (0, 223),
            (0, 0),
            (0, 11),
            (0, 26),
            (0, 88),
            (0, 635),
            (0, 1),
            (0, 2447),
            (0, 16),
            (0, 894),
            (0, 3),
            (0, 2762),
            (0, 16),
            (0, 396),
            (0, 989),
            (0, 129),
            (0, 123),
            (0, 8),
            (0, 42),
            (0, 8),
            (0, 80),
            (0, 11),
            (0, 46),
            (0, 0),
            (0, 78),
            (0, 25),
            (0, 15),
            (0, 1),
            (0, 1807),
            (0, 3233),
            (0, 10),
            (0, 22),
            (0, 1783),
            (0, 4),
            (0, 634),
            (0, 258),
            (0, 846),
            (0, 4),
            (0, 2),
            (0, 636),
            (0, 149),
            (0, 121),
            (0, 16),
            (0, 1084),
            (0, 11),
            (0, 114),
            (0, 5),
            (0, 12),
            (0, 150),
            (0, 53),
            (0, 12),
            (0, 2),
            (0, 11),
            (0, 10),
            (0, 989),
            (0, 1424),
            (0, 12),
            (0, 2),
            (0, 1),
            (0, 0),
            (0, 0),
            (0, 36),
            (0, 591),
            (0, 3830),
            (0, 1),
            (0, 799),
            (0, 21),
            (0, 69),
            (0, 427),
            (0, 24),
            (0, 28),
            (0, 480),
            (0, 272),
            (0, 81),
            (0, 0),
            (0, 301),
            (0, 30),
            (0, 0),
            (0, 18),
            (0, 2),
            (0, 199),
            (0, 25),
            (0, 57),
            (0, 5),
            (0, 373),
            (0, 2),
            (0, 104),
            (0, 119),
            (0, 64),
            (0, 0),
            (0, 269),
            (0, 25),
            (0, 14),
            (0, 19),
            (0, 13),
            (0, 81),
            (0, 8),
            (0, 223),
            (0, 114),
            (0, 101),
            (0, 28),
            (0, 52),
            (0, 284),
            (0, 0),
            (0, 12),
            (0, 1),
            (0, 0),
            (0, 2),
            (0, 65),
            (0, 966),
            (0, 4556),
            (0, 106),
            (0, 8),
            (0, 0),
            (0, 61),
            (0, 1382),
            (0, 9),
            (0, 129),
            (0, 1),
            (0, 8),
            (0, 273),
            (0, 0),
            (0, 485),
            (0, 1),
            (0, 19),
            (0, 0),
            (0, 208),
            (0, 0),
            (0, 24),
            (0, 105),
            (0, 611),
            (0, 133),
            (0, 49),
            (0, 66),
            (0, 0),
            (0, 33),
            (0, 41),
            (0, 149),
            (0, 0),
            (0, 721),
            (0, 26),
            (0, 47),
            (0, 9),
            (0, 7),
            (0, 3),
            (0, 56),
            (0, 87),
            (0, 21),
            (0, 2),
            (0, 146),
            (0, 7),
            (0, 5),
            (0, 487),
            (0, 64),
            (0, 12),
            (0, 58),
            (0, 217),
            (0, 12),
            (0, 129),
            (0, 3014),
            (0, 34),
            (0, 24),
            (0, 268),
            (0, 89),
            (0, 543),
            (0, 26),
            (0, 75),
            (0, 339),
            (0, 55),
            (0, 39),
            (0, 56),
            (0, 1),
            (0, 24),
            (0, 203),
            (0, 54),
            (0, 0),
            (0, 94),
            (0, 99),
            (0, 193),
            (0, 10),
            (0, 77),
            (0, 736),
            (0, 0),
            (0, 1125),
            (0, 492),
            (0, 16),
            (0, 421),
            (0, 647),
            (0, 478),
            (0, 14),
            (0, 2589),
            (0, 297),
            (0, 236),
            (0, 53),
            (0, 4),
            (0, 24),
            (0, 1375),
            (0, 707),
            (0, 51),
            (0, 4),
            (0, 265),
            (0, 329),
            (0, 41),
            (0, 123),
            (0, 2),
            (0, 1933),
            (0, 1322),
            (0, 144),
            (0, 1574),
            (0, 3),
            (0, 401),
            (0, 412),
            (0, 21),
            (0, 914),
            (0, 27),
            (0, 10),
            (0, 202),
            (0, 17),
            (0, 4),
            (0, 51),
            (0, 164),
            (0, 235),
            (0, 6),
            (0, 0),
            (0, 552),
            (0, 153),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 2),
            (0, 46),
            (0, 181),
            (0, 617),
            (0, 90),
            (0, 14),
            (0, 2),
            (0, 32),
            (0, 4),
            (0, 43),
            (0, 171),
            (0, 0),
            (0, 1),
            (0, 19),
            (0, 35),
            (0, 419),
            (0, 4),
            (0, 1),
            (0, 77),
            (0, 36),
            (0, 31),
            (0, 146),
            (0, 1),
            (0, 1),
            (0, 1945),
            (0, 585),
            (0, 228),
            (0, 0),
            (0, 16),
            (0, 1),
            (0, 2),
            (0, 8),
            (0, 1),
            (0, 13),
            (0, 2789),
            (0, 356),
            (0, 180),
            (0, 1465),
            (0, 7),
            (0, 556),
            (0, 15),
            (0, 3),
            (0, 782),
            (0, 1),
            (0, 1054),
            (0, 27),
            (0, 0),
            (0, 0),
            (0, 296),
            (0, 2),
            (0, 0),
            (0, 1),
            (0, 47),
            (0, 662),
            (0, 21),
            (0, 3),
            (0, 403),
            (0, 16),
            (0, 2),
            (0, 33),
            (0, 55),
            (0, 21),
            (0, 7),
            (0, 712),
            (0, 48),
            (0, 0),
            (0, 2086),
            (0, 11),
            (0, 2),
            (0, 402),
            (0, 0),
            (0, 58),
            (0, 1),
            (0, 15),
            (0, 0),
            (0, 386),
            (0, 0),
            (0, 85),
            (0, 962),
            (0, 179),
            (0, 192),
            (0, 57),
            (0, 4),
            (0, 50),
            (0, 6),
            (0, 4),
            (0, 44),
            (0, 0),
            (0, 0),
            (0, 81),
            (0, 546),
            (0, 2),
            (0, 184),
            (0, 14),
            (0, 0),
            (0, 387),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 117),
            (0, 74),
            (0, 0),
            (0, 148),
            (0, 2),
            (0, 2),
            (0, 3),
            (0, 0),
            (0, 32),
            (0, 140),
            (0, 547),
            (0, 176),
            (0, 137),
            (0, 1),
            (0, 9),
            (0, 7),
            (0, 1325),
            (0, 370),
            (0, 6),
            (0, 6),
            (0, 442),
            (0, 42),
            (0, 3),
            (0, 283),
            (0, 47),
            (0, 59),
            (0, 32),
            (0, 0),
            (0, 1),
            (0, 3),
            (0, 1747),
            (0, 36),
            (0, 208),
            (0, 77),
            (0, 4),
            (0, 26),
            (0, 23),
            (0, 66),
            (0, 274),
            (0, 7),
            (0, 91),
            (0, 4),
            (0, 21),
            (0, 280),
            (0, 1287),
            (0, 758),
            (0, 206),
            (0, 21),
            (0, 4),
            (0, 0),
            (0, 75),
            (0, 143),
            (0, 201),
            (0, 38),
            (0, 460),
            (0, 1032),
            (0, 678),
            (0, 36),
            (0, 120),
            (0, 14),
            (0, 7),
            (0, 348),
            (0, 0),
            (0, 3),
            (0, 373),
            (0, 24),
            (0, 38),
            (0, 256),
            (0, 3),
            (0, 0),
            (0, 10),
            (0, 3),
            (0, 20),
            (0, 3),
            (0, 1692),
            (0, 55),
            (0, 36),
            (0, 323),
            (0, 988),
            (0, 5),
            (0, 364),
            (0, 1),
            (0, 3),
            (0, 31),
            (0, 53),
            (0, 1),
            (0, 39),
            (0, 3),
            (0, 11),
            (0, 0),
            (0, 319),
            (0, 13),
            (0, 648),
            (0, 888),
            (0, 570),
            (0, 2),
            (0, 184),
            (0, 27),
            (0, 40),
            (0, 2),
            (0, 28),
            (0, 199),
            (0, 45),
            (0, 2648),
            (0, 5),
            (0, 292),
            (0, 3),
            (0, 868),
            (0, 741),
            (0, 14),
            (0, 7),
            (0, 25),
            (0, 1),
            (0, 411),
            (0, 298),
            (0, 270),
            (0, 614),
            (0, 529),
            (0, 0),
            (0, 784),
            (0, 1111),
            (0, 376),
            (0, 1881),
            (0, 0),
            (0, 25),
            (0, 8),
            (0, 195),
            (0, 4016),
            (0, 22),
            (0, 18),
            (0, 188),
            (0, 26),
            (0, 717),
            (0, 448),
            (0, 2),
            (0, 43),
            (0, 398),
            (0, 356),
            (0, 966),
            (0, 1),
            (0, 92),
            (0, 296),
            (0, 17),
            (0, 1),
            (0, 0),
            (0, 284),
            (0, 0),
            (0, 8),
            (0, 0),
            (0, 0),
            (0, 308),
            (0, 611),
            (0, 366),
            (0, 601),
            (0, 1263),
            (0, 40),
            (0, 58),
            (0, 14),
            (0, 2692),
            (0, 16),
            (0, 103),
            (0, 2537),
            (0, 14),
            (0, 0),
            (0, 1),
            (0, 122),
            (0, 246),
            (0, 1984),
            (0, 369),
            (0, 655),
            (0, 15),
            (0, 55),
            (0, 16),
            (0, 494),
            (0, 2),
            (0, 42),
            (0, 13),
            (0, 31),
            (0, 669),
            (0, 1),
            (0, 50),
            (0, 9),
            (0, 26),
            (0, 42),
            (0, 0),
            (0, 0),
            (0, 565),
            (0, 0),
            (0, 450),
            (0, 124),
            (0, 3),
            (0, 46),
            (0, 16),
            (0, 1),
            (0, 405),
            (0, 274),
            (0, 132),
            (0, 935),
            (0, 232),
            (0, 14),
            (0, 1),
            (0, 426),
            (0, 310),
            (0, 0),
            (0, 44),
            (0, 3186),
            (0, 122),
            (0, 60),
            (0, 150),
            (0, 13),
            (0, 84),
            (0, 23),
            (0, 110),
            (0, 12),
            (0, 1341),
            (0, 44),
            (0, 204),
            (0, 34),
            (0, 171),
            (0, 0),
            (0, 2470),
            (0, 773),
            (0, 5),
            (0, 4),
            (0, 4),
            (0, 5),
            (0, 0),
            (0, 373),
            (0, 355),
            (0, 6),
            (0, 1666),
            (0, 342),
            (0, 3),
            (0, 39),
            (0, 40),
            (0, 4),
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 18),
            (0, 0),
            (0, 51),
            (0, 1),
            (0, 1829),
            (0, 1532),
            (0, 1069),
            (0, 6),
            (0, 164),
            (0, 132),
            (0, 2),
            (0, 59),
            (0, 53),
            (0, 3),
            (0, 304),
            (0, 0),
            (0, 234),
            (0, 307),
            (0, 6),
            (0, 66),
            (0, 5),
            (0, 803),
            (0, 23),
            (0, 278),
            (0, 3),
            (0, 260),
            (0, 36),
            (0, 192),
            (0, 45),
            (0, 199),
            (0, 12),
            (0, 517),
            (0, 115),
            (0, 69),
            (0, 21),
            (0, 5),
            (0, 26),
            (0, 104),
            (0, 474),
            (0, 273),
            (0, 1),
            (0, 0),
            (0, 171),
            (0, 7),
            (0, 91),
            (0, 211),
            (0, 5),
            (0, 1147),
            (0, 293),
            (0, 222),
            (0, 11),
            (0, 194),
            (0, 51),
            (0, 61),
            (0, 40),
            (0, 7),
            (0, 601),
            (0, 264),
            (0, 46),
            (0, 321),
            (0, 10),
            (0, 0),
            (0, 1),
            (0, 602),
            (0, 120),
            (0, 298),
            (0, 731),
            (0, 40),
            (0, 264),
            (0, 61),
            (0, 82),
            (0, 0),
            (0, 1),
            (0, 26),
            (0, 5),
            (0, 61),
            (0, 0),
            (0, 271),
            (0, 419),
            (0, 25),
            (0, 1),
            (0, 0),
            (0, 219),
            (0, 35),
            (0, 20),
            (0, 0),
            (0, 2),
            (0, 699),
            (0, 17),
            (0, 44),
            (0, 16),
            (0, 25),
            (0, 358),
            (0, 1322),
            (0, 442),
            (0, 3),
            (0, 32),
            (0, 2),
            (0, 116),
            (0, 5),
            (0, 92),
            (0, 387),
            (0, 0),
            (0, 113),
            (0, 17),
            (0, 9),
            (0, 6),
            (0, 34),
            (0, 11),
            (0, 26),
            (0, 1233),
            (0, 0),
            (0, 0),
            (0, 94),
            (0, 193),
            (0, 1),
            (0, 897),
            (0, 8),
            (0, 2),
            (0, 117),
            (0, 200),
            (0, 305),
            (0, 214),
            (0, 1030),
            (0, 295),
            (0, 40),
            (0, 14),
            (0, 537),
            (0, 738),
            (0, 0),
            (0, 17),
            (0, 448),
            (0, 0),
            (0, 1044),
            (0, 13),
            (0, 0),
            (0, 480),
            (0, 33),
            (0, 9),
            (0, 25),
            (0, 550),
            (0, 5),
            (0, 1441),
            (0, 315),
            (0, 3),
            (0, 85),
            (0, 3),
            (0, 1412),
            (0, 0),
            (0, 122),
            (0, 0),
            (0, 15),
            (0, 1786),
            (0, 63),
            (0, 28),
            (0, 0),
            (0, 201),
            (0, 59),
            (0, 6),
            (0, 3),
            (0, 1),
            (0, 0),
            (0, 1079),
            (0, 200),
            (0, 299),
            (0, 15),
            (0, 143),
            (0, 2),
            (0, 6),
            (0, 116),
            (0, 19),
            (0, 1),
            (0, 865),
            (0, 85),
            (0, 0),
            (0, 473),
            (0, 2874),
            (0, 11),
            (0, 50),
            (0, 0),
            (0, 1207),
            (0, 38),
            (0, 127),
            (0, 0),
            (0, 0),
            (0, 2009),
            (0, 4),
            (0, 2598),
            (0, 5),
            (0, 1),
            (0, 74),
            (0, 68),
            (0, 461),
            (0, 45),
            (0, 32),
            (0, 68),
            (0, 0),
            (0, 7),
            (0, 6),
            (0, 378),
            (0, 128),
            (0, 302),
            (0, 338),
            (0, 122),
            (0, 23),
            (0, 14),
            (0, 227),
            (0, 101),
            (0, 50),
            (0, 430),
            (0, 93),
            (0, 0),
            (0, 37),
            (0, 212),
            (0, 19),
            (0, 0),
            (0, 111),
            (0, 31),
            (0, 253),
            (0, 158),
            (0, 0),
            (0, 49),
            (0, 0),
            (0, 25),
            (0, 16),
            (0, 0),
            (0, 56),
            (0, 180),
            (0, 1331),
            (0, 145),
            (0, 145),
            (0, 14),
            (0, 5),
            (0, 1),
            (0, 1385),
            (0, 0),
            (0, 440),
            (0, 55),
            (0, 200),
            (0, 151),
            (0, 1),
            (0, 9),
            (0, 1),
            (0, 1),
            (0, 70),
            (0, 16),
            (0, 14),
            (0, 2),
            (0, 239),
            (0, 38),
            (0, 53),
            (0, 77),
            (0, 0),
            (0, 206),
            (0, 698),
            (0, 0),
            (0, 12),
            (0, 395),
            (0, 1),
            (0, 75),
            (0, 6),
            (0, 5),
            (0, 7),
            (0, 71),
            (0, 8),
            (0, 64),
            (0, 503),
            (0, 2),
            (0, 1020),
            (0, 20),
            (0, 28),
            (0, 265),
            (0, 104),
            (0, 11),
            (0, 12),
            (0, 29),
            (0, 13),
            (0, 0),
            (0, 0),
            (0, 5),
            (0, 21),
            (0, 10),
            (0, 58),
            (0, 6),
            (0, 25),
            (0, 11),
            (0, 17),
            (0, 85),
            (0, 11),
            (0, 4),
            (0, 192),
            (0, 2),
            (0, 11),
            (0, 8),
            (0, 2),
            (0, 16),
            (0, 277),
            (0, 23),
            (0, 73),
            (0, 33),
            (0, 1),
            (0, 213),
            (0, 41),
            (0, 52),
            (0, 24),
            (0, 154),
            (0, 0),
            (0, 644),
            (0, 52),
            (0, 2622),
            (0, 709),
            (0, 661),
            (0, 444),
            (0, 364),
            (0, 327),
            (0, 169),
            (0, 165),
            (0, 3),
            (0, 159),
            (0, 32),
            (0, 1344),
            (0, 0),
            (0, 10),
            (0, 1),
            (0, 7),
            (0, 26),
            (0, 4098),
            (0, 441),
            (0, 287),
            (0, 2049),
            (0, 345),
            (0, 540),
            (0, 1909),
            (0, 146),
            (0, 402),
            (0, 117),
            (0, 114),
            (0, 318),
            (0, 8),
            (0, 116),
            (0, 17),
            (0, 12),
            (0, 56),
            (0, 274),
            (0, 0),
            (0, 2453),
            (0, 0),
            (0, 117),
            (0, 194),
            (0, 293),
            (0, 1),
            (0, 619),
            (0, 77),
            (0, 1056),
            (0, 138),
            (0, 340),
            (0, 68),
            (0, 46),
            (0, 202),
            (0, 34),
            (0, 91),
            (0, 37),
            (0, 54),
            (0, 2),
            (0, 0),
            (0, 265),
            (0, 209),
            (0, 29),
            (0, 2938),
            (0, 0),
            (0, 2),
            (0, 171),
            (0, 1),
            (0, 18),
            (0, 4),
            (0, 983),
            (0, 692),
            (0, 597),
            (0, 211),
            (0, 15),
            (0, 462),
            (0, 226),
            (0, 0),
            (0, 2),
            (0, 768),
            (0, 2),
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 59),
            (0, 929),
            (0, 1),
            (0, 1315),
            (0, 326),
            (0, 1528),
            (0, 84),
            (0, 3),
            (0, 2),
            (0, 5),
            (0, 5),
            (0, 268),
            (0, 793),
            (0, 29),
            (0, 1),
            (0, 20),
            (0, 309),
            (0, 168),
            (0, 0),
            (0, 1196),
            (0, 8),
            (0, 1),
            (0, 109),
            (0, 447),
            (0, 96),
            (0, 5),
            (0, 99),
            (0, 263),
            (0, 18),
            (0, 53),
            (0, 0),
            (0, 122),
            (0, 275),
            (0, 1),
            (0, 71),
            (0, 0),
            (0, 13),
            (0, 1120),
            (0, 86),
            (0, 517),
            (0, 42),
            (0, 86),
            (0, 17),
            (0, 6),
            (0, 77),
            (0, 28),
            (0, 445),
            (0, 143),
            (0, 2),
            (0, 0),
            (0, 598),
            (0, 793),
            (0, 291),
            (0, 240),
            (0, 21),
            (0, 359),
            (0, 1664),
            (0, 1),
            (0, 350),
            (0, 198),
            (0, 565),
            (0, 873),
            (0, 1270),
            (0, 8),
            (0, 912),
            (0, 87),
            (0, 15),
            (0, 16),
            (0, 0),
            (0, 0),
            (0, 480),
            (0, 44),
            (0, 1615),
            (0, 16),
            (0, 21),
            (0, 0),
            (0, 11),
            (0, 486),
            (0, 394),
            (0, 0),
            (0, 1726),
            (0, 11),
            (0, 29),
            (0, 135),
            (0, 234),
            (0, 0),
            (0, 75),
            (0, 359),
            (0, 298),
            (0, 172),
            (0, 0),
            (0, 68),
            (0, 1586),
            (0, 0),
            (0, 470),
            (0, 464),
            (0, 98),
            (0, 32),
            (0, 599),
            (0, 1343),
            (0, 13),
            (0, 858),
            (0, 0),
            (0, 955),
            (0, 199),
            (0, 1),
            (0, 0),
            (0, 1),
            (0, 158),
            (0, 3),
            (0, 433),
            (0, 0),
            (0, 56),
            (0, 54),
            (0, 4415),
            (0, 31),
            (0, 0),
            (0, 0),
            (0, 89),
            (0, 235),
            (0, 1),
            (0, 6),
            (0, 3),
            (0, 2),
            (0, 160),
            (0, 55),
            (0, 2065),
            (0, 540),
            (0, 6),
            (0, 20),
            (0, 13),
            (0, 0),
            (0, 206),
            (0, 418),
            (0, 160),
            (0, 0),
            (0, 17),
            (0, 0),
            (0, 61),
            (0, 29),
            (0, 100),
            (0, 0),
            (0, 0),
            (0, 64),
            (0, 121),
            (0, 813),
            (0, 0),
            (0, 492),
            (0, 260),
            (0, 67),
            (0, 0),
            (0, 456),
            (0, 1481),
            (0, 0),
            (0, 5),
            (0, 408),
            (0, 42),
            (0, 9),
            (0, 110),
            (0, 183),
            (0, 692),
            (0, 431),
            (0, 9),
            (0, 136),
            (0, 88),
            (0, 228),
            (0, 44),
            (0, 78),
            (0, 6),
            (0, 18),
            (0, 140),
            (0, 243),
            (0, 146),
            (0, 241),
            (0, 5),
            (0, 89),
            (0, 0),
            (0, 6),
            (0, 10),
            (0, 12),
            (0, 290),
            (0, 554),
            (0, 95),
            (0, 428),
            (0, 1),
            (0, 19),
            (0, 151),
            (0, 79),
            (0, 66),
            (0, 9),
            (0, 7),
            (0, 1087),
            (0, 34),
            (0, 79),
            (0, 3),
            (0, 14),
            (0, 2),
            (0, 9),
            (0, 47),
            (0, 8),
            (0, 16),
            (0, 0),
            (0, 28),
            (0, 1),
            (0, 61),
            (0, 10),
            (0, 5),
            (0, 23),
            (0, 1069),
            (0, 0),
            (0, 46),
            (0, 0),
            (0, 124),
            (0, 695),
            (0, 0),
            (0, 1844),
            (0, 0),
            (0, 360),
            (0, 21),
            (0, 84),
            (0, 3),
            (0, 111),
            (0, 103),
            (0, 1),
            (0, 0),
            (0, 1),
            (0, 577),
            (0, 422),
            (0, 0),
            (0, 73),
            (0, 0),
            (0, 0),
            (0, 268),
            (0, 0),
            (0, 446),
            (0, 40),
            (0, 459),
            (0, 1331),
            (0, 49),
            (0, 5),
            (0, 82),
            (0, 402),
            (0, 1326),
            (0, 46),
            (0, 330),
            (0, 0),
            (0, 80),
            (0, 16),
            (0, 0),
            (0, 144),
            (0, 1563),
            (0, 1061),
            (0, 8),
            (0, 2251),
            (0, 9),
            (0, 2129),
            (0, 9),
            (0, 5),
            (0, 715),
            (0, 3985),
            (0, 212),
            (0, 1),
            (0, 0),
            (0, 1),
            (0, 7),
            (0, 99),
            (0, 4),
            (0, 34),
            (0, 21),
            (0, 0),
            (0, 1740),
            (0, 808),
            (0, 0),
            (0, 28),
            (0, 31),
            (0, 3),
            (0, 31),
            (0, 3321),
            (0, 12),
            (0, 5),
            (0, 1),
            (0, 1372),
            (0, 350),
            (0, 10),
            (0, 0),
            (0, 0),
            (0, 2),
            (0, 100),
            (0, 277),
            (0, 136),
            (0, 168),
            (0, 108),
            (0, 71),
            (0, 1),
            (0, 38),
            (0, 170),
            (0, 2300),
            (0, 57),
            (0, 278),
            (0, 445),
            (0, 1),
            (0, 1),
            (0, 51),
            (0, 0),
            (0, 31),
            (0, 101),
            (0, 6),
            (0, 27),
            (0, 583),
            (0, 3),
            (0, 55),
            (0, 342),
            (0, 13),
            (0, 2334),
            (0, 13),
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 287),
            (0, 83),
            (0, 188),
            (0, 596),
            (0, 0),
            (0, 53),
            (0, 11),
            (0, 429),
            (0, 7),
            (0, 71),
            (0, 600),
            (0, 7),
            (0, 537),
            (0, 25),
            (0, 0),
            (0, 775),
            (0, 1),
            (0, 52),
            (0, 9),
            (0, 49),
            (0, 922),
            (0, 597),
            (0, 492),
            (0, 0),
            (0, 442),
            (0, 0),
            (0, 19),
            (0, 178),
            (0, 17),
            (0, 0),
            (0, 18),
            (0, 30),
            (0, 7),
            (0, 123),
            (0, 2),
            (0, 119),
            (0, 1),
            (0, 1),
            (0, 553),
            (0, 183),
            (0, 6),
            (0, 1061),
            (0, 252),
            (0, 1),
            (0, 13),
            (0, 10),
            (0, 2017),
            (0, 1),
            (0, 9),
            (0, 176),
            (0, 0),
            (0, 2121),
            (0, 0),
            (0, 20),
            (0, 222),
            (0, 4),
            (0, 34),
            (0, 23),
            (0, 92),
            (0, 12),
            (0, 29),
            (0, 91),
            (0, 79),
            (0, 28),
            (0, 1424),
            (0, 298),
            (0, 1048),
            (0, 156),
            (0, 0),
            (0, 186),
            (0, 860),
            (0, 126),
            (0, 51),
            (0, 12),
            (0, 261),
            (0, 63),
            (0, 35),
            (0, 166),
            (0, 0),
            (0, 2),
            (0, 4),
            (0, 1117),
            (0, 1850),
            (0, 3),
            (0, 157),
            (0, 767),
            (0, 189),
            (0, 1944),
            (0, 0),
            (0, 127),
            (0, 6),
            (0, 4),
            (0, 30),
            (0, 10),
            (0, 192),
            (0, 302),
            (0, 18),
            (0, 3),
            (0, 0),
            (0, 2151),
            (0, 1688),
            (0, 17),
            (0, 1234),
            (0, 10),
            (0, 115),
            (0, 1403),
            (0, 818),
            (0, 14),
            (0, 23),
            (0, 60),
            (0, 1),
            (0, 33),
            (0, 3),
            (0, 1),
            (0, 1202),
            (0, 1553),
            (0, 53),
            (0, 74),
            (0, 12),
            (0, 105),
            (0, 3),
            (0, 0),
            (0, 47),
            (0, 37),
            (0, 159),
            (0, 496),
            (0, 2),
            (0, 24),
            (0, 0),
            (0, 109),
            (0, 2124),
            (0, 2469),
            (0, 7),
            (0, 1),
            (0, 858),
            (0, 940),
            (0, 19),
            (0, 4),
            (0, 0),
            (0, 5),
            (0, 2822),
            (0, 102),
            (0, 26),
            (0, 103),
            (0, 1895),
            (0, 120),
            (0, 90),
            (0, 37),
            (0, 73),
            (0, 11),
            (0, 514),
            (0, 0),
            (0, 4),
            (0, 119),
            (0, 798),
            (0, 417),
            (0, 3),
            (0, 41),
            (0, 0),
            (0, 82),
            (0, 298),
            (0, 32),
            (0, 2),
            (0, 43),
            (0, 110),
            (0, 0),
            (0, 3),
            (0, 124),
            (0, 1614),
            (0, 945),
            (0, 288),
            (0, 0),
            (0, 44),
            (0, 3),
            (0, 118),
            (0, 82),
            (0, 18),
            (0, 373),
            (0, 1),
            (0, 0),
            (0, 252),
            (0, 255),
            (0, 930),
            (0, 892),
            (0, 381),
            (0, 590),
            (0, 8),
            (0, 27),
            (0, 297),
            (0, 860),
            (0, 30),
            (0, 4),
            (0, 8),
            (0, 2),
            (0, 520),
            (0, 40),
            (0, 179),
            (0, 0),
            (0, 226),
            (0, 0),
            (0, 1598),
            (0, 2),
            (0, 2179),
            (0, 1),
            (0, 140),
            (0, 57),
            (0, 58),
            (0, 4),
            (0, 441),
            (0, 0),
            (0, 257),
            (0, 17),
            (0, 0),
            (0, 8),
            (0, 0),
            (0, 2380),
            (0, 582),
            (0, 1),
            (0, 264),
            (0, 15),
            (0, 0),
            (0, 0),
            (0, 758),
            (0, 87),
            (0, 1),
            (0, 277),
            (0, 2),
            (0, 10),
            (0, 256),
            (0, 40),
            (0, 386),
            (0, 3),
            (0, 349),
            (0, 15),
            (0, 881),
            (0, 1156),
            (0, 36),
            (0, 953),
            (0, 598),
            (0, 119),
            (0, 7),
            (0, 43),
            (0, 80),
            (0, 500),
            (0, 66),
            (0, 102),
            (0, 443),
            (0, 55),
            (0, 43),
            (0, 183),
            (0, 237),
            (0, 6),
            (0, 4),
            (0, 110),
            (0, 164),
            (0, 22),
            (0, 0),
            (0, 147),
            (0, 211),
            (0, 6),
            (0, 33),
            (0, 138),
            (0, 0),
            (0, 2),
            (0, 6),
            (0, 20),
            (0, 215),
            (0, 82),
            (0, 7),
            (0, 76),
            (0, 378),
            (0, 3),
            (0, 10),
            (0, 936),
            (0, 1),
            (0, 23),
            (0, 61),
            (0, 35),
            (0, 2165),
            (0, 1429),
            (0, 1),
            (0, 41),
            (0, 37),
            (0, 1),
            (0, 0),
            (0, 398),
            (0, 442),
            (0, 1),
            (0, 0),
            (0, 14),
            (0, 0),
            (0, 905),
            (0, 0),
            (0, 43),
            (0, 196),
            (0, 0),
            (0, 93),
            (0, 56),
            (0, 258),
            (0, 9),
            (0, 183),
            (0, 4),
            (0, 1),
            (0, 1),
            (0, 3),
            (0, 181),
            (0, 22),
            (0, 847),
            (0, 370),
            (0, 136),
            (0, 475),
            (0, 109),
            (0, 133),
            (0, 24),
            (0, 189),
            (0, 78),
            (0, 120),
            (0, 273),
            (0, 10),
            (0, 218),
            (0, 38),
            (0, 50),
            (0, 0),
            (0, 569),
            (0, 773),
            (0, 6),
            (0, 71),
            (0, 12),
            (0, 7),
            (0, 473),
            (0, 721),
            (0, 455),
            (0, 1218),
            (0, 1),
            (0, 280),
            (0, 1144),
            (0, 270),
            (0, 593),
            (0, 7),
            (0, 13),
            (0, 215),
            (0, 1),
            (0, 0),
            (0, 622),
            (0, 68),
            (0, 137),
            (0, 19),
            (0, 72),
            (0, 12),
            (0, 1726),
            (0, 0),
            (0, 15),
            (0, 735),
            (0, 298),
            (0, 401),
            (0, 107),
            (0, 9),
            (0, 0),
            (0, 4),
            (0, 2650),
            (0, 0),
            (0, 116),
            (0, 62),
            (0, 483),
            (0, 8),
            (0, 21),
            (0, 195),
            (0, 0),
            (0, 11),
            (0, 113),
            (0, 49),
            (0, 3),
            (0, 102),
            (0, 303),
            (0, 0),
            (0, 44),
            (0, 21),
            (0, 16),
            (0, 238),
            (0, 135),
            (0, 86),
            (0, 604),
            (0, 3),
            (0, 4),
            (0, 1780),
            (0, 290),
            (0, 261),
            (0, 11),
            (0, 399),
            (0, 10),
            (0, 0),
            (0, 2882),
            (0, 0),
            (0, 0),
            (0, 7),
            (0, 8),
            (0, 0),
            (0, 96),
            (0, 0),
            (0, 0),
            (0, 444),
            (0, 81),
            (0, 1832),
            (0, 18),
            (0, 198),
            (0, 796),
            (0, 271),
            (0, 74),
            (0, 1372),
            (0, 0),
            (0, 7),
            (0, 0),
            (0, 20),
            (0, 9),
            (0, 18),
            (0, 3905),
            (0, 5),
            (0, 338),
            (0, 801),
            (0, 1624),
            (0, 577),
            (0, 0),
            (0, 320),
            (0, 333),
            (0, 9),
            (0, 89),
            (0, 1),
            (0, 529),
            (0, 127),
            (0, 217),
            (0, 0),
            (0, 37),
            (0, 18),
            (0, 45),
            (0, 3),
            (0, 231),
            (0, 0),
            (0, 435),
            (0, 13),
            (0, 5),
            (0, 0),
            (0, 0),
            (0, 2),
            (0, 263),
            (0, 303),
            (0, 412),
            (0, 116),
            (0, 105),
            (0, 946),
            (0, 1201),
            (0, 34),
            (0, 36),
            (0, 227),
            (0, 265),
            (0, 339),
            (0, 11),
            (0, 6),
            (0, 167),
            (0, 856),
            (0, 34),
            (0, 2334),
            (0, 339),
            (0, 25),
            (0, 58),
            (0, 3),
            (0, 239),
            (0, 9),
            (0, 2808),
            (0, 25),
            (0, 945),
            (0, 101),
            (0, 75),
            (0, 370),
            (0, 7),
            (0, 40),
            (0, 344),
            (0, 80),
            (0, 291),
            (0, 153),
            (0, 0),
            (0, 967),
            (0, 669),
            (0, 12),
            (0, 753),
            (0, 0),
            (0, 771),
            (0, 0),
            (0, 6),
            (0, 0),
            (0, 1),
            (0, 24),
            (0, 2),
            (0, 36),
            (0, 601),
            (0, 36),
            (0, 133),
            (0, 34),
            (0, 0),
            (0, 41),
            (0, 303),
            (0, 670),
            (0, 10),
            (0, 26),
            (0, 418),
            (0, 0),
            (0, 8),
            (0, 28),
            (0, 93),
            (0, 108),
            (0, 11),
            (0, 12),
            (0, 22),
            (0, 254),
            (0, 263),
            (0, 126),
            (0, 391),
            (0, 61),
            (0, 0),
            (0, 7),
            (0, 239),
            (0, 101),
            (0, 225),
            (0, 175),
            (0, 5),
            (0, 177),
            (0, 44),
            (0, 24),
            (0, 3),
            (0, 68),
            (0, 15),
            (0, 64),
            (0, 719),
            (0, 31),
            (0, 10),
            (0, 17),
            (0, 492),
            (0, 54),
            (0, 3),
            (0, 2152),
            (0, 0),
            (0, 132),
            (0, 0),
            (0, 678),
            (0, 131),
            (0, 3158),
            (0, 7),
            (0, 0),
            (0, 1160),
            (0, 22),
            (0, 8),
            (0, 1476),
            (0, 62),
            (0, 28),
            (0, 31),
            (0, 145),
            (0, 2),
            (0, 0),
            (0, 1130),
            (0, 6),
            (0, 20),
            (0, 11),
            (0, 10),
            (0, 38),
            (0, 113),
            (0, 40),
            (0, 832),
            (0, 200),
            (0, 9),
            (0, 43),
            (0, 32),
            (0, 64),
            (0, 2381),
            (0, 299),
            (0, 8),
            (0, 223),
            (0, 41),
            (0, 10),
            (0, 272),
            (0, 161),
            (0, 489),
            (0, 7),
            (0, 127),
            (0, 1084),
            (0, 18),
            (0, 226),
            (0, 9),
            (0, 1623),
            (0, 1251),
            (0, 2),
            (0, 229),
            (0, 0),
            (0, 505),
            (0, 6),
            (0, 7),
            (0, 105),
            (0, 17),
            (0, 628),
            (0, 130),
            (0, 9),
            (0, 138),
            (0, 0),
            (0, 132),
            (0, 84),
            (0, 4),
            (0, 14),
            (0, 23),
            (0, 12),
            (0, 34),
            (0, 5),
            (0, 7),
            (0, 72),
            (0, 21),
            (0, 47),
            (0, 224),
            (0, 2),
            (0, 66),
            (0, 0),
            (0, 405),
            (0, 46),
            (0, 184),
            (0, 9),
            (0, 666),
            (0, 18),
            (0, 4),
            (0, 42),
            (0, 3),
            (0, 30),
            (0, 57),
            (0, 0),
            (0, 84),
            (0, 442),
            (0, 5),
            (0, 63),
            (0, 445),
            (0, 48),
            (0, 135),
            (0, 964),
            (0, 15),
            (0, 86),
            (0, 5),
            (0, 2),
            (0, 19),
            (0, 291),
            (0, 37),
            (0, 2),
            (0, 0),
            (0, 53),
            (0, 152),
            (0, 395),
            (0, 6),
            (0, 31),
            (0, 44),
            (0, 17),
            (0, 2766),
            (0, 174),
            (0, 45),
            (0, 114),
            (0, 22),
            (0, 12),
            (0, 5),
            (0, 1514),
            (0, 287),
            (0, 1),
            (0, 543),
            (0, 18),
            (0, 771),
            (0, 941),
            (0, 12),
            (0, 66),
            (0, 218),
            (0, 143),
            (0, 11),
            (0, 19),
            (0, 475),
            (0, 2108),
            (0, 4),
            (0, 185),
            (0, 84),
            (0, 4),
            (0, 1473),
            (0, 0),
            (0, 7),
            (0, 407),
            (0, 1),
            (0, 154),
            (0, 2),
            (0, 84),
            (0, 5),
            (0, 95),
            (0, 70),
            (0, 24),
            (0, 26),
            (0, 5),
            (0, 5),
            (0, 180),
            (0, 345),
            (0, 322),
            (0, 45),
            (0, 891),
            (0, 5),
            (0, 655),
            (0, 42),
            (0, 265),
            (0, 682),
            (0, 212),
            (0, 0),
            (0, 1),
            (0, 48),
            (0, 514),
            (0, 2),
            (0, 0),
            (0, 15),
            (0, 2784),
            (0, 40),
            (0, 9),
            (0, 486),
            (0, 3),
            (0, 281),
            (0, 8),
            (0, 1),
            (0, 7),
            (0, 304),
            (0, 13),
            (0, 586),
            (0, 8),
            (0, 2),
            (0, 82),
            (0, 281),
            (0, 3),
            (0, 33),
            (0, 2),
            (0, 1),
            (0, 194),
            (0, 93),
            (0, 280),
            (0, 376),
            (0, 466),
            (0, 0),
            (0, 450),
            (0, 442),
            (0, 11),
            (0, 4),
            (0, 790),
            (0, 16),
            (0, 85),
            (0, 492),
            (0, 272),
            (0, 118),
            (0, 1),
            (0, 101),
            (0, 8),
            (0, 406),
            (0, 0),
            (0, 28),
            (0, 28),
            (0, 33),
            (0, 137),
            (0, 177),
            (0, 995),
            (0, 4),
            (0, 673),
            (0, 39),
            (0, 32),
            (0, 5),
            (0, 15),
            (0, 70),
            (0, 619),
            (0, 1075),
            (0, 419),
            (0, 228),
            (0, 610),
            (0, 10),
            (0, 371),
            (0, 243),
            (0, 68),
            (0, 5),
            (0, 276),
            (0, 0),
            (0, 193),
            (0, 0),
            (0, 230),
            (0, 367),
            (0, 0),
            (0, 2690),
            (0, 21),
            (0, 238),
            (0, 904),
            (0, 17),
            (0, 105),
            (0, 38),
            (0, 14),
            (0, 9),
            (0, 1),
            (0, 595),
            (0, 851),
            (0, 104),
            (0, 0),
            (0, 0),
            (0, 30),
            (0, 49),
            (0, 157),
            (0, 2),
            (0, 10),
            (0, 14),
            (0, 101),
            (0, 845),
            (0, 538),
            (0, 5),
            (0, 1),
            (0, 386),
            (0, 1770),
            (0, 122),
            (0, 4),
            (0, 1892),
            (0, 1),
            (0, 14),
            (0, 101),
            (0, 620),
            (0, 2),
            (0, 203),
            (0, 16),
            (0, 1),
            (0, 67),
            (0, 63),
            (0, 187),
            (0, 695),
            (0, 428),
            (0, 1106),
            (0, 651),
            (0, 1),
            (0, 45),
            (0, 137),
            (0, 293),
            (0, 1),
            (0, 3),
            (0, 49),
            (0, 1),
            (0, 210),
            (0, 2),
            (0, 148),
            (0, 6),
            (0, 451),
            (0, 1852),
            (0, 74),
            (0, 79),
            (0, 351),
            (0, 94),
            (0, 682),
            (0, 1),
            (0, 29),
            (0, 0),
            (0, 366),
            (0, 41),
            (0, 0),
            (0, 1113),
            (0, 6),
            (0, 0),
            (0, 121),
            (0, 619),
            (0, 2),
            (0, 5),
            (0, 217),
            (0, 1013),
            (0, 912),
            (0, 16),
            (0, 276),
            (0, 1636),
            (0, 63),
            (0, 2),
            (0, 1),
            (0, 16),
            (0, 5),
            (0, 8),
            (0, 128),
            (0, 14),
            (0, 124),
            (0, 853),
            (0, 1713),
            (0, 2),
            (0, 6),
            (0, 193),
            (0, 127),
            (0, 1294),
            (0, 20),
            (0, 48),
            (0, 7),
            (0, 527),
            (0, 7),
            (0, 1),
            (0, 0),
            (0, 0),
            (0, 27),
            (0, 1238),
            (0, 5),
            (0, 1753),
            (0, 121),
            (0, 17),
            (0, 133),
            (0, 11),
            (0, 456),
            (0, 399),
            (0, 11),
            (0, 67),
            (0, 208),
            (0, 104),
            (0, 148),
            (0, 555),
            (0, 452),
            (0, 10),
            (0, 1),
            (0, 221),
            (0, 176),
            (0, 3),
            (0, 23),
            (0, 0),
            (0, 54),
            (0, 29),
            (0, 10),
            (0, 0),
            (0, 93),
            (0, 48),
            (0, 2419),
            (0, 40),
            (0, 4),
            (0, 0),
            (0, 1048),
            (0, 418),
            (0, 607),
            (0, 1601),
            (0, 3439),
            (0, 5),
            (0, 193),
            (0, 398),
            (0, 405),
            (0, 323),
            (0, 16),
            (0, 2),
            (0, 14),
            (0, 52),
            (0, 4),
            (0, 9),
            (0, 12),
            (0, 40),
            (0, 2),
            (0, 0),
            (0, 573),
            (0, 15),
            (0, 4),
            (0, 177),
            (0, 173),
            (0, 5),
            (0, 914),
            (0, 267),
            (0, 800),
            (0, 120),
            (0, 0),
            (0, 2794),
            (0, 81),
            (0, 37),
            (0, 146),
            (0, 127),
            (0, 90),
            (0, 4499),
            (0, 3),
            (0, 214),
            (0, 122),
            (0, 699),
            (0, 972),
            (0, 0),
            (0, 3),
            (0, 37),
            (0, 0),
            (0, 0),
            (0, 249),
            (0, 1321),
            (0, 0),
            (0, 234),
            (0, 583),
            (0, 152),
            (0, 18),
            (0, 34),
            (0, 916),
            (0, 36),
            (0, 254),
            (0, 38),
            (0, 2),
            (0, 0),
            (0, 44),
            (0, 230),
            (0, 665),
            (0, 218),
            (0, 5),
            (0, 1116),
            (0, 1059),
            (0, 642),
            (0, 72),
            (0, 83),
            (0, 1),
            (0, 826),
            (0, 1),
            (0, 439),
            (0, 24),
            (0, 5),
            (0, 99),
            (0, 254),
            (0, 19),
            (0, 1),
            (0, 0),
            (0, 355),
            (0, 1152),
            (0, 72),
            (0, 487),
            (0, 18),
            (0, 232),
            (0, 4),
            (0, 459),
            (0, 0),
            (0, 1),
            (0, 38),
            (0, 148),
            (0, 21),
            (0, 107),
            (0, 105),
            (0, 1),
            (0, 6),
            (0, 281),
            (0, 0),
            (0, 38),
            (0, 13),
            (0, 9),
            (0, 1114),
            (0, 4),
            (0, 1097),
            (0, 7),
            (0, 1002),
            (0, 52),
            (0, 232),
            (0, 339),
            (0, 68),
            (0, 4),
            (0, 794),
            (0, 6),
            (0, 16),
            (0, 4),
            (0, 14),
            (0, 4),
            (0, 1),
            (0, 3),
            (0, 239),
            (0, 15),
            (0, 6),
            (0, 2115),
            (0, 956),
            (0, 28),
            (0, 15),
            (0, 169),
            (0, 0),
            (0, 186),
            (0, 98),
            (0, 13),
            (0, 0),
            (0, 4),
            (0, 2331),
            (0, 25),
            (0, 28),
            (0, 44),
            (0, 25),
            (0, 1073),
            (0, 1),
            (0, 76),
            (0, 4),
            (0, 4),
            (0, 0),
            (0, 13),
            (0, 72),
            (0, 150),
            (0, 367),
            (0, 5),
            (0, 22),
            (0, 7),
            (0, 3),
            (0, 1),
            (0, 160),
            (0, 0),
            (0, 94),
            (0, 35),
            (0, 3),
            (0, 2256),
            (0, 212),
            (0, 273),
            (0, 0),
            (0, 0),
            (0, 5),
            (0, 21),
            (0, 57),
            (0, 11),
            (0, 1873),
            (0, 47),
            (0, 21),
            (0, 8),
            (0, 0),
            (0, 2),
            (0, 6),
            (0, 199),
            (0, 100),
            (0, 198),
            (0, 1),
            (0, 1136),
            (0, 2),
            (0, 26),
            (0, 240),
            (0, 76),
            (0, 3),
            (0, 39),
            (0, 675),
            (0, 0),
            (0, 758),
            (0, 5),
            (0, 3),
            (0, 0),
            (0, 173),
            (0, 350),
            (0, 316),
            (0, 5),
            (0, 8),
            (0, 676),
            (0, 692),
            (0, 13),
            (0, 125),
            (0, 10),
            (0, 61),
            (0, 1),
            (0, 93),
            (0, 1),
            (0, 6),
            (0, 114),
            (0, 84),
            (0, 120),
            (0, 0),
            (0, 3191),
            (0, 370),
            (0, 7),
            (0, 12),
            (0, 8),
            (0, 7),
            (0, 1),
            (0, 11),
            (0, 612),
            (0, 6),
            (0, 47),
            (0, 806),
            (0, 219),
            (0, 30),
            (0, 1328),
            (0, 35),
            (0, 124),
            (0, 18),
            (0, 671),
            (0, 29),
            (0, 0),
            (0, 22),
            (0, 1),
            (0, 194),
            (0, 204),
            (0, 45),
            (0, 662),
            (0, 30),
            (0, 315),
            (0, 49),
            (0, 33),
            (0, 7),
            (0, 373),
            (0, 56),
            (0, 0),
            (0, 50),
            (0, 111),
            (0, 16),
            (0, 929),
            (0, 2),
            (0, 464),
            (0, 4),
            (0, 707),
            (0, 8),
            (0, 72),
            (0, 1647),
            (0, 4),
            (0, 78),
            (0, 31),
            (0, 44),
            (0, 8),
            (0, 14),
            (0, 11),
            (0, 4),
            (0, 1),
            (0, 55),
            (0, 0),
            (0, 229),
            (0, 85),
            (0, 182),
            (0, 784),
            (0, 774),
            (0, 1489),
            (0, 256),
            (0, 51),
            (0, 12),
            (0, 192),
            (0, 382),
            (0, 51),
            (0, 0),
            (0, 938),
            (0, 2),
            (0, 5),
            (0, 4),
            (0, 620),
            (0, 99),
            (0, 1),
            (0, 41),
            (0, 74),
            (0, 6),
            (0, 0),
            (0, 980),
            (0, 141),
            (0, 123),
            (0, 3),
            (0, 0),
            (0, 0),
            (0, 951),
            (0, 1),
            (0, 6),
            (0, 1470),
            (0, 632),
            (0, 7),
            (0, 0),
            (0, 389),
            (0, 179),
            (0, 6),
            (0, 20),
            (0, 0),
            (0, 80),
            (0, 239),
            (0, 107),
            (0, 56),
            (0, 28),
            (0, 2332),
            (0, 683),
            (0, 5),
            (0, 742),
            (0, 172),
            (0, 2),
            (0, 4),
            (0, 1276),
            (0, 114),
            (0, 15),
            (0, 1),
            (0, 389),
            (0, 6),
            (0, 11),
            (0, 4),
            (0, 0),
            (0, 0),
            (0, 1),
            (0, 633),
            (0, 123),
            (0, 4),
            (0, 214),
            (0, 0),
            (0, 39),
            (0, 546),
            (0, 19),
            (0, 228),
            (0, 44),
            (0, 573),
            (0, 9),
            (0, 7),
            (0, 2),
            (0, 33),
            (0, 4),
            (0, 18),
            (0, 910),
            (0, 43),
            (0, 390),
            (0, 54),
            (0, 295),
            (0, 43),
            (0, 30),
            (0, 419),
            (0, 17),
            (0, 13),
            (0, 38),
            (0, 134),
            (0, 9),
            (0, 0),
            (0, 620),
            (0, 107),
            (0, 22),
            (0, 48),
            (0, 1304),
            (0, 1888),
            (0, 238),
            (0, 63),
            (0, 564),
            (0, 152),
            (0, 561),
            (0, 11),
            (0, 6),
            (0, 89),
            (0, 513),
            (0, 3),
            (0, 2257),
            (0, 0),
            (0, 512),
            (0, 297),
            (0, 6),
            (0, 838),
            (0, 38),
            (0, 2),
            (0, 126),
            (0, 3),
            (0, 5),
            (0, 41),
            (0, 19),
            (0, 3),
            (0, 426),
            (0, 0),
            (0, 0),
            (0, 278),
            (0, 25),
            (0, 72),
            (0, 10),
            (0, 93),
            (0, 674),
            (0, 400),
            (0, 149),
            (0, 11),
            (0, 23),
            (0, 6),
            (0, 1029),
            (0, 5),
            (0, 118),
            (0, 24),
            (0, 846),
            (0, 6),
            (0, 1307),
            (0, 648),
            (0, 292),
            (0, 146),
            (0, 50),
            (0, 561),
            (0, 2301),
            (0, 493),
            (0, 96),
            (0, 866),
            (0, 1239),
            (0, 128),
            (0, 261),
            (0, 0),
            (0, 2),
            (0, 274),
            (0, 950),
            (0, 3),
            (0, 365),
            (0, 28),
            (0, 2440),
            (0, 1807),
            (0, 1138),
            (0, 87),
            (0, 1305),
            (0, 173),
            (0, 275),
            (0, 510),
            (0, 28),
            (0, 4),
            (0, 39),
            (0, 49),
            (0, 0),
            (0, 37),
            (0, 0),
            (0, 82),
            (0, 650),
            (0, 19),
            (0, 0),
            (0, 7),
            (0, 727),
            (0, 9),
            (0, 253),
            (0, 397),
            (0, 11),
            (0, 139),
            (0, 498),
            (0, 1),
            (0, 539),
            (0, 1281),
            (0, 332),
            (0, 48),
            (0, 9),
            (0, 16),
            (0, 7),
            (0, 40),
            (0, 7),
            (0, 2),
            (0, 578),
            (0, 1),
            (0, 605),
            (0, 317),
            (0, 0),
            (0, 6),
            (0, 1397),
            (0, 1918),
            (0, 579),
            (0, 1467),
            (0, 236),
            (0, 6),
            (0, 398),
            (0, 0),
            (0, 464),
            (0, 139),
            (0, 78),
            (0, 67),
            (0, 309),
            (0, 129),
            (0, 67),
            (0, 95),
            (0, 17),
            (0, 720),
            (0, 16),
            (0, 143),
            (0, 411),
            (0, 8),
            (0, 53),
            (0, 506),
            (0, 543),
            (0, 35),
            (0, 97),
            (0, 82),
            (0, 40),
            (0, 67),
            (0, 2155),
            (0, 2),
            (0, 5),
            (0, 367),
            (0, 120),
            (0, 39),
            (0, 870),
            (0, 1),
            (0, 14),
            (0, 184),
            (0, 306),
            (0, 2),
            (0, 488),
            (0, 160),
            (0, 5),
            (0, 1479),
            (0, 44),
            (0, 114),
            (0, 46),
            (0, 1),
            (0, 30),
            (0, 255),
            (0, 1163),
            (0, 62),
            (0, 1691),
            (0, 24),
            (0, 10),
            (0, 1962),
            (0, 8),
            (0, 12),
            (0, 540),
            (0, 69),
            (0, 171),
            (0, 60),
            (0, 83),
            (0, 322),
            (0, 67),
            (0, 16),
            (0, 33),
            (0, 54),
            (0, 354),
            (0, 6),
            (0, 14),
            (0, 43),
            (0, 66),
            (0, 84),
            (0, 371),
            (0, 4),
            (0, 221),
            (0, 649),
            (0, 0),
            (0, 7),
            (0, 0),
            (0, 74),
            (0, 151),
            (0, 525),
            (0, 106),
            (0, 539),
            (0, 8),
            (0, 109),
            (0, 108),
            (0, 8),
            (0, 0),
            (0, 56),
            (0, 46),
            (0, 4),
            (0, 205),
            (0, 51),
            (0, 10),
            (0, 36),
            (0, 489),
            (0, 779),
            (0, 128),
            (0, 1),
            (0, 23),
            (0, 2),
            (0, 590),
            (0, 974),
            (0, 830),
            (0, 665),
            (0, 228),
            (0, 262),
            (0, 26),
            (0, 123),
            (0, 333),
            (0, 1872),
            (0, 1196),
            (0, 225),
            (0, 25),
            (0, 1),
            (0, 1),
            (0, 509),
            (0, 67),
            (0, 0),
            (0, 760),
            (0, 137),
            (0, 10),
            (0, 813),
            (0, 5),
            (0, 177),
            (0, 190),
            (0, 125),
            (0, 74),
            (0, 1207),
            (0, 27),
            (0, 1),
            (0, 602),
            (0, 130),
            (0, 1383),
            (0, 147),
            (0, 326),
            (0, 9),
            (0, 57),
            (0, 241),
            (0, 149),
            (0, 190),
            (0, 192),
            (0, 8),
            (0, 8),
            (0, 225),
            (0, 291),
            (0, 125),
            (0, 5),
            (0, 66),
            (0, 13),
            (0, 5),
            (0, 6),
            (0, 228),
            (0, 604),
            (0, 1),
            (0, 1),
            (0, 144),
            (0, 40),
            (0, 2),
            (0, 20),
            (0, 296),
            (0, 1059),
            (0, 20),
            (0, 579),
            (0, 4),
            (0, 693),
            (0, 92),
            (0, 0),
            (0, 1984),
            (0, 3),
            (0, 8),
            (0, 2),
            (0, 1218),
            (0, 1),
            (0, 606),
            (0, 14),
            (0, 295),
            (0, 10),
            (0, 55),
            (0, 119),
            (0, 232),
            (0, 186),
            (0, 261),
            (0, 6),
            (0, 0),
            (0, 38),
            (0, 4),
            (0, 20),
            (0, 0),
            (0, 25),
            (0, 32),
            (0, 303),
            (0, 141),
            (0, 16),
            (0, 131),
            (0, 181),
            (0, 454),
            (0, 224),
            (0, 9),
            (0, 148),
            (0, 33),
            (0, 277),
            (0, 32),
            (0, 668),
            (0, 92),
            (0, 0),
            (0, 194),
            (0, 87),
            (0, 1391),
            (0, 2),
            (0, 11),
            (0, 17),
            (0, 33),
            (0, 799),
            (0, 0),
            (0, 929),
            (0, 349),
            (0, 121),
            (0, 850),
            (0, 16),
            (0, 13),
            (0, 246),
            (0, 1),
            (0, 1489),
            (0, 1),
            (0, 439),
            (0, 882),
            (0, 1),
            (0, 21),
            (0, 225),
            (0, 2),
            (0, 5),
            (0, 91),
            (0, 884),
            (0, 0),
            (0, 0),
            (0, 2058),
            (0, 49),
            (0, 69),
            (0, 942),
            (0, 10),
            (0, 103),
            (0, 1073),
            (0, 2),
            (0, 555),
            (0, 1),
            (0, 178),
            (0, 723),
            (0, 0),
            (0, 8),
            (0, 1414),
            (0, 297),
            (0, 0),
            (0, 6),
            (0, 811),
            (0, 43),
            (0, 244),
            (0, 128),
            (0, 57),
            (0, 1246),
            (0, 92),
            (0, 33),
            (0, 1),
            (0, 4),
            (0, 3),
            (0, 68),
            (0, 7),
            (0, 738),
            (0, 24),
            (0, 11),
            (0, 121),
            (0, 41),
            (0, 717),
            (0, 2),
            (0, 25),
            (0, 9),
            (0, 1338),
            (0, 0),
            (0, 0),
            (0, 49),
            (0, 594),
            (0, 21),
            (0, 22),
            (0, 28),
            (0, 73),
            (0, 4),
            (0, 194),
            (0, 289),
            (0, 206),
            (0, 588),
            (0, 888),
            (0, 512),
            (0, 34),
            (0, 12),
            (0, 642),
            (0, 92),
            (0, 2),
            (0, 301),
            (0, 29),
            (0, 822),
            (0, 1824),
            (0, 98),
            (0, 288),
            (0, 183),
            (0, 0),
            (0, 2),
            (0, 14),
            (0, 835),
            (0, 132),
            (0, 622),
            (0, 5),
            (0, 0),
            (0, 676),
            (0, 116),
            (0, 22),
            (0, 4),
            (0, 128),
            (0, 45),
            (0, 1),
            (0, 4),
            (0, 47),
            (0, 61),
            (0, 959),
            (0, 25),
            (0, 276),
            (0, 138),
            (0, 27),
            (0, 1156),
            (0, 510),
            (0, 380),
            (0, 3),
            (0, 2),
            (0, 7),
            (0, 43),
            (0, 181),
            (0, 64),
            (0, 0),
            (0, 3),
            (0, 118),
            (0, 1),
            (0, 0),
            (0, 15),
            (0, 212),
            (0, 1245),
            (0, 19),
            (0, 3),
            (0, 9),
            (0, 1075),
            (0, 47),
            (0, 37),
            (0, 772),
            (0, 8),
            (0, 111),
            (0, 0),
            (0, 1),
            (0, 33),
            (0, 988),
            (0, 4),
            (0, 127),
            (0, 502),
            (0, 3),
            (0, 1154),
            (0, 239),
            (0, 893),
            (0, 161),
            (0, 85),
            (0, 774),
            (0, 8),
            (0, 376),
            (0, 260),
            (0, 115),
            (0, 7),
            (0, 17),
            (0, 1),
            (0, 14),
            (0, 115),
            (0, 649),
            (0, 255),
            (0, 56),
            (0, 2),
            (0, 1),
            (0, 121),
            (0, 11),
            (0, 21),
            (0, 7),
            (0, 64),
            (0, 75),
            (0, 13),
            (0, 322),
            (0, 0),
            (0, 20),
            (0, 3451),
            (0, 1),
            (0, 51),
            (0, 3),
            (0, 806),
            (0, 1375),
            (0, 363),
            (0, 1),
            (0, 14),
            (0, 93),
            (0, 154),
            (0, 851),
            (0, 34),
            (0, 39),
            (0, 10),
            (0, 588),
            (0, 11),
            (0, 27),
            (0, 8),
            (0, 5),
            (0, 22),
            (0, 516),
            (0, 8),
            (0, 8),
            (0, 42),
            (0, 30),
            (0, 38),
            (0, 681),
            (0, 687),
            (0, 130),
            (0, 56),
            (0, 507),
            (0, 0),
            (0, 15),
            (0, 174),
            (0, 5),
            (0, 21),
            (0, 1189),
            (0, 343),
            (0, 2),
            (0, 4),
            (0, 9),
            (0, 4408),
            (0, 65),
            (0, 7),
            (0, 641),
            (0, 3),
            (0, 2138),
            (0, 273),
            (0, 446),
            (0, 118),
            (0, 1580),
            (0, 1),
            (0, 1480),
            (0, 441),
            (0, 1171),
            (0, 169),
            (0, 15),
            (0, 0),
            (0, 175),
            (0, 503),
            (0, 171),
            (0, 3),
            (0, 3),
            (0, 989),
            (0, 200),
            (0, 742),
            (0, 4),
            (0, 3),
            (0, 8),
            (0, 130),
            (0, 1003),
            (0, 418),
            (0, 128),
            (0, 860),
            (0, 3),
            (0, 1391),
            (0, 155),
            (0, 117),
            (0, 3766),
            (0, 121),
            (0, 78),
            (0, 0),
            (0, 768),
            (0, 35),
            (0, 13),
            (0, 0),
            (0, 2473),
            (0, 722),
            (0, 0),
            (0, 0),
            (0, 137),
            (0, 41),
            (0, 588),
            (0, 583),
            (0, 0),
            (0, 0),
            (0, 6),
            (0, 2441),
            (0, 1728),
            (0, 40),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 4),
            (0, 182),
            (0, 546),
            (0, 18),
            (0, 219),
            (0, 2182),
            (0, 2),
            (0, 100),
            (0, 436),
            (0, 5),
            (0, 4337),
            (0, 0),
            (0, 225),
            (0, 3),
            (0, 2030),
            (0, 433),
            (0, 7),
            (0, 1155),
            (0, 16),
            (0, 519),
            (0, 1),
            (0, 37),
            (0, 9),
            (0, 7),
            (0, 634),
            (0, 32),
            (0, 167),
            (0, 797),
            (0, 714),
            (0, 31),
            (0, 8),
            (0, 0),
            (0, 90),
            (0, 267),
            (0, 1612),
            (0, 0),
            (0, 1),
            (0, 51),
            (0, 52),
            (0, 1953),
            (0, 1381),
            (0, 824),
            (0, 1),
            (0, 0),
            (0, 1418),
            (0, 1288),
            (0, 2),
            (0, 66),
            (0, 49),
            (0, 0),
            (0, 2),
            (0, 41),
            (0, 13),
            (0, 0),
            (0, 0),
            (0, 693),
            (0, 2),
            (0, 29),
            (0, 0),
            (0, 167),
            (0, 26),
            (0, 490),
            (0, 23),
            (0, 4),
            (0, 14),
            (0, 581),
            (0, 945),
            (0, 53),
            (0, 4),
            (0, 788),
            (0, 16),
            (0, 0),
            (0, 160),
            (0, 1464),
            (0, 159),
            (0, 5),
            (0, 888),
            (0, 181),
            (0, 2676),
            (0, 2),
            (0, 0),
            (0, 14),
            (0, 901),
            (0, 1223),
            (0, 2),
            (0, 568),
            (0, 71),
            (0, 229),
            (0, 0),
            (0, 497),
            (0, 767),
            (0, 32),
            (0, 37),
            (0, 1339),
            (0, 961),
            (0, 35),
            (0, 63),
            (0, 7),
            (0, 495),
            (0, 311),
            (0, 0),
            (0, 82),
            (0, 381),
            (0, 56),
            (0, 125),
            (0, 0),
            (0, 427),
            (0, 0),
            (0, 2),
            (0, 3414),
            (0, 3332),
            (0, 6),
            (0, 19),
            (0, 1),
            (0, 882),
            (0, 181),
            (0, 71),
            (0, 18),
            (0, 34),
            (0, 31),
            (0, 334),
            (0, 2990),
            (0, 589),
            (0, 95),
            (0, 0),
            (0, 216),
            (0, 8),
            (0, 0),
            (0, 4144),
            (0, 184),
            (0, 222),
            (0, 1573),
            (0, 125),
            (0, 98),
            (0, 3),
            (0, 178),
            (0, 12),
            (0, 320),
            (0, 131),
            (0, 278),
            (0, 149),
            (0, 14),
            (0, 86),
            (0, 20),
            (0, 72),
            (0, 11),
            (0, 1016),
            (0, 27),
            (0, 167),
            (0, 763),
            (0, 35),
            (0, 115),
            (0, 6),
            (0, 43),
            (0, 17),
            (0, 59),
            (0, 5759),
            (0, 18),
            (0, 306),
            (0, 1),
            (0, 354),
            (0, 635),
            (0, 2112),
            (0, 4),
            (0, 114),
            (0, 317),
            (0, 15),
            (0, 125),
            (0, 1117),
            (0, 239),
            (0, 122),
            (0, 13),
            (0, 58),
            (0, 13),
            (0, 1354),
            (0, 4),
            (0, 5),
            (0, 94),
            (0, 310),
            (0, 34),
            (0, 3),
            (0, 740),
            (0, 9),
            (0, 19),
            (0, 230),
            (0, 857),
            (0, 10),
            (0, 1073),
            (0, 80),
            (0, 7),
            (0, 12),
            (0, 56),
            (0, 6),
            (0, 190),
            (0, 705),
            (0, 45),
            (0, 48),
            (0, 23),
            (0, 3),
            (0, 11),
            (0, 2),
            (0, 17),
            (0, 198),
            (0, 0),
            (0, 13),
            (0, 704),
            (0, 1),
            (0, 2),
            (0, 12),
            (0, 1933),
            (0, 0),
            (0, 562),
            (0, 678),
            (0, 1),
            (0, 528),
            (0, 4),
            (0, 338),
            (0, 0),
            (0, 62),
            (0, 0),
            (0, 35),
            (0, 10),
            (0, 91),
            (0, 54),
            (0, 34),
            (0, 603),
            (0, 4),
            (0, 0),
            (0, 136),
            (0, 778),
            (0, 9),
            (0, 252),
            (0, 1),
            (0, 3),
            (0, 3),
            (0, 2),
            (0, 5),
            (0, 1545),
            (0, 7),
            (0, 1),
            (0, 0),
            (0, 8),
            (0, 368),
            (0, 74),
            (0, 74),
            (0, 157),
            (0, 35),
            (0, 259),
            (0, 513),
            (0, 587),
            (0, 550),
            (0, 501),
            (0, 164),
            (0, 522),
            (0, 287),
            (0, 11),
            (0, 88),
            (0, 220),
            (0, 1126),
            (0, 99),
            (0, 678),
            (0, 639),
            (0, 19),
            (0, 1229),
            (0, 50),
            (0, 2849),
            (0, 387),
            (0, 166),
            (0, 16),
            (0, 1525),
            (0, 210),
            (0, 185),
            (0, 2),
            (0, 2),
            (0, 105),
            (0, 1077),
            (0, 214),
            (0, 3),
            (0, 163),
            (0, 241),
            (0, 194),
            (0, 8),
            (0, 1),
            (0, 4),
            (0, 185),
            (0, 7),
            (0, 25),
            (0, 4),
            (0, 1920),
            (0, 647),
            (0, 6),
            (0, 83),
            (0, 1891),
            (0, 682),
            (0, 192),
            (0, 12),
            (0, 941),
            (0, 714),
            (0, 0),
            (0, 0),
            (0, 144),
            (0, 144),
            (0, 28),
            (0, 11),
            (0, 14),
            (0, 59),
            (0, 1264),
            (0, 15),
            (0, 2),
            (0, 588),
            (0, 2),
            (0, 2515),
            (0, 0),
            (0, 0),
            (0, 33),
            (0, 1339),
            (0, 869),
            (0, 21),
            (0, 550),
            (0, 203),
            (0, 8),
            (0, 11),
            (0, 1552),
            (0, 107),
            (0, 9),
            (0, 1242),
            (0, 880),
            (0, 2),
            (0, 9),
            (0, 496),
            (0, 497),
            (0, 2901),
            (0, 221),
            (0, 0),
            (0, 0),
            (0, 85),
            (0, 1),
            (0, 2),
            (0, 18),
            (0, 126),
            (0, 159),
            (0, 107),
            (0, 271),
            (0, 327),
            (0, 3352),
            (0, 834),
            (0, 18),
            (0, 14),
            (0, 422),
            (0, 82),
            (0, 7),
            (0, 1145),
            (0, 1556),
            (0, 1408),
            (0, 52),
            (0, 538),
            (0, 916),
            (0, 135),
            (0, 72),
            (0, 30),
            (0, 34),
            (0, 1144),
            (0, 40),
            (0, 614),
            (0, 9),
            (0, 365),
            (0, 515),
            (0, 51),
            (0, 105),
            (0, 580),
            (0, 1140),
            (0, 4),
            (0, 3),
            (0, 114),
            (0, 199),
            (0, 164),
            (0, 435),
            (0, 25),
            (0, 1585),
            (0, 0),
            (0, 5),
            (0, 1),
            (0, 48),
            (0, 455),
            (0, 156),
            (0, 39),
            (0, 5),
            (0, 0),
            (0, 208),
            (0, 85),
            (0, 20),
            (0, 8),
            (0, 5),
            (0, 1865),
            (0, 24),
            (0, 1291),
            (0, 15),
            (0, 18),
            (0, 104),
            (0, 115),
            (0, 47),
            (0, 67),
            (0, 28),
            (0, 70),
            (0, 23),
            (0, 25),
            (0, 123),
            (0, 1),
            (0, 5),
            (0, 0),
            (0, 59),
            (0, 207),
            (0, 50),
            (0, 293),
            (0, 1),
            (0, 0),
            (0, 2687),
            (0, 0),
            (0, 22),
            (0, 0),
            (0, 15),
            (0, 81),
            (0, 243),
            (0, 10),
            (0, 11),
            (0, 26),
            (0, 0),
            (0, 591),
            (0, 22),
            (0, 1),
            (0, 5),
            (0, 150),
            (0, 2),
            (0, 10),
            (0, 1122),
            (0, 613),
            (0, 128),
            (0, 7),
            (0, 57),
            (0, 1066),
            (0, 438),
            (0, 38),
            (0, 362),
            (0, 1538),
            (0, 4),
            (0, 2),
            (0, 79),
            (0, 21),
            (0, 18),
            (0, 419),
            (0, 1077),
            (0, 14),
            (0, 1633),
            (0, 432),
            (0, 21),
            (0, 346),
            (0, 662),
            (0, 109),
            (0, 1),
            (0, 511),
            (0, 4),
            (0, 4),
            (0, 290),
            (0, 1106),
            (0, 0),
            (0, 108),
            (0, 2),
            (0, 810),
            (0, 29),
            (0, 2015),
            (0, 0),
            (0, 2),
            (0, 170),
            (0, 15),
            (0, 46),
            (0, 367),
            (0, 61),
            (0, 5),
            (0, 140),
            (0, 10),
            (0, 3),
            (0, 265),
            (0, 69),
            (0, 101),
            (0, 242),
            (0, 0),
            (0, 37),
            (0, 72),
            (0, 125),
            (0, 109),
            (0, 103),
            (0, 278),
            (0, 974),
            (0, 45),
            (0, 4),
            (0, 562),
            (0, 42),
            (0, 95),
            (0, 33),
            (0, 254),
            (0, 20),
            (0, 335),
            (0, 0),
            (0, 0),
            (0, 572),
            (0, 14),
            (0, 3),
            (0, 384),
            (0, 1),
            (0, 64),
            (0, 1065),
            (0, 93),
            (0, 105),
            (0, 445),
            (0, 70),
            (0, 178),
            (0, 6),
            (0, 25),
            (0, 18),
            (0, 1157),
            (0, 64),
            (0, 7),
            (0, 64),
            (0, 108),
            (0, 174),
            (0, 0),
            (0, 2),
            (0, 117),
            (0, 707),
            (0, 49),
            (0, 13),
            (0, 421),
            (0, 60),
            (0, 10),
            (0, 194),
            (0, 178),
            (0, 17),
            (0, 896),
            (0, 105),
            (0, 96),
            (0, 54),
            (0, 154),
            (0, 0),
            (0, 2),
            (0, 973),
            (0, 4),
            (0, 97),
            (0, 1),
            (0, 125),
            (0, 2),
            (0, 1006),
            (0, 0),
            (0, 334),
            (0, 14),
            (0, 2156),
            (0, 0),
            (0, 460),
            (0, 377),
            (0, 2),
            (0, 134),
            (0, 1376),
            (0, 182),
            (0, 544),
            (0, 23),
            (0, 1423),
            (0, 10),
            (0, 4),
            (0, 14),
            (0, 12),
            (0, 152),
            (0, 194),
            (0, 90),
            (0, 0),
            (0, 6),
            (0, 2396),
            (0, 2),
            (0, 1463),
            (0, 1008),
            (0, 12),
            (0, 0),
            (0, 46),
            (0, 3083),
            (0, 1583),
            (0, 916),
            (0, 1535),
            (0, 153),
            (0, 1795),
            (0, 2),
            (0, 164),
            (0, 213),
            (0, 29),
            (0, 53),
            (0, 0),
            (0, 638),
            (0, 87),
            (0, 1),
            (0, 431),
            (0, 312),
            (0, 9),
            (0, 49),
            (0, 1494),
            (0, 41),
            (0, 18),
            (0, 46),
            (0, 1586),
            (0, 8),
            (0, 91),
            (0, 2),
            (0, 72),
            (0, 516),
            (0, 71),
            (0, 0),
            (0, 283),
            (0, 1661),
            (0, 20),
            (0, 4),
            (0, 4),
            (0, 61),
            (0, 988),
            (0, 195),
            (0, 6),
            (0, 1690),
            (0, 159),
            (0, 72),
            (0, 253),
            (0, 2052),
            (0, 668),
            (0, 2),
            (0, 2001),
            (0, 2),
            (0, 21),
            (0, 0),
            (0, 757),
            (0, 13),
            (0, 21),
            (0, 104),
            (0, 276),
            (0, 28),
            (0, 2045),
            (0, 58),
            (0, 44),
            (0, 403),
            (0, 0),
            (0, 1135),
            (0, 11),
            (0, 121),
            (0, 15),
            (0, 116),
            (0, 15),
            (0, 171),
            (0, 797),
            (0, 514),
            (0, 0),
            (0, 307),
            (0, 207),
            (0, 462),
            (0, 124),
            (0, 1),
            (0, 137),
            (0, 54),
            (0, 283),
            (0, 4),
            (0, 16),
            (0, 7),
            (0, 287),
            (0, 7),
            (0, 746),
            (0, 2),
            (0, 1),
            (0, 170),
            (0, 2996),
            (0, 128),
            (0, 915),
            (0, 113),
            (0, 638),
            (0, 210),
            (0, 8),
            (0, 108),
            (0, 4),
            (0, 0),
            (0, 5),
            (0, 9),
            (0, 788),
            (0, 48),
            (0, 16),
            (0, 165),
            (0, 718),
            (0, 0),
            (0, 1),
            (0, 260),
            (0, 0),
            (0, 419),
            (0, 266),
            (0, 8),
            (0, 83),
            (0, 45),
            (0, 177),
            (0, 5),
            (0, 1825),
            (0, 5),
            (0, 4288),
            (0, 17),
            (0, 0),
            (0, 77),
            (0, 25),
            (0, 6),
            (0, 297),
            (0, 5),
            (0, 99),
            (0, 392),
            (0, 26),
            (0, 2716),
            (0, 3),
            (0, 0),
            (0, 3),
            (0, 554),
            (0, 52),
            (0, 653),
            (0, 3),
            (0, 1366),
            (0, 19),
            (0, 11),
            (0, 0),
            (0, 1117),
            (0, 1193),
            (0, 0),
            (0, 481),
            (0, 64),
            (0, 293),
            (0, 131),
            (0, 85),
            (0, 128),
            (0, 1),
            (0, 4),
            (0, 54),
            (0, 364),
            (0, 26),
            (0, 10),
            (0, 6),
            (0, 0),
            (0, 1),
            (0, 20),
            (0, 146),
            (0, 4),
            (0, 0),
            (0, 469),
            (0, 24),
            (0, 451),
            (0, 121),
            (0, 8),
            (0, 75),
            (0, 20),
            (0, 2),
            (0, 1),
            (0, 68),
            (0, 16),
            (0, 416),
            (0, 75),
            (0, 455),
            (0, 765),
            (0, 6),
            (0, 9),
            (0, 12),
            (0, 228),
            (0, 245),
            (0, 838),
            (0, 27),
            (0, 137),
            (0, 303),
            (0, 9),
            (0, 17),
            (0, 12),
            (0, 15),
            (0, 0),
            (0, 3),
            (0, 6),
            (0, 365),
            (0, 228),
            (0, 4),
            (0, 38),
            (0, 0),
            (0, 171),
            (0, 196),
            (0, 25),
            (0, 522),
            (0, 343),
            (0, 0),
            (0, 1),
            (0, 205),
            (0, 19),
            (0, 144),
            (0, 47),
            (0, 38),
            (0, 35),
            (0, 545),
            (0, 48),
            (0, 31),
            (0, 0),
            (0, 1096),
            (0, 350),
            (0, 3026),
            (0, 91),
            (0, 1086),
            (0, 111),
            (0, 76),
            (0, 500),
            (0, 3),
            (0, 4),
            (0, 1400),
            (0, 312),
            (0, 705),
            (0, 871),
            (0, 932),
            (0, 256),
            (0, 29),
            (0, 2),
            (0, 10),
            (0, 1381),
            (0, 0),
            (0, 16),
            (0, 5),
            (0, 15),
            (0, 135),
            (0, 28),
            (0, 249),
            (0, 537),
            (0, 3180),
            (0, 6),
            (0, 2448),
            (0, 45),
            (0, 12),
            (0, 11),
            (0, 231),
            (0, 4),
            (0, 5),
            (0, 207),
            (0, 42),
            (0, 483),
            (0, 5157),
            (0, 12),
            (0, 42),
            (0, 88),
            (0, 202),
            (0, 49),
            (0, 3),
            (0, 1367),
            (0, 1),
            (0, 426),
            (0, 0),
            (0, 24),
            (0, 70),
            (0, 3143),
            (0, 190),
            (0, 97),
            (0, 40),
            (0, 452),
            (0, 2109),
            (0, 17),
            (0, 563),
            (0, 1),
            (0, 125),
            (0, 0),
            (0, 671),
            (0, 946),
            (0, 46),
            (0, 4),
            (0, 3392),
            (0, 471),
            (0, 542),
            (0, 472),
            (0, 1),
            (0, 14),
            (0, 1),
            (0, 38),
            (0, 5346),
            (0, 113),
            (0, 134),
            (0, 27),
            (0, 1404),
            (0, 69),
            (0, 4367),
            (0, 26),
            (0, 759),
            (0, 667),
            (0, 52),
            (0, 68),
            (0, 248),
            (0, 133),
            (0, 52),
            (0, 36),
            (0, 482),
            (0, 1),
            (0, 17),
            (0, 149),
            (0, 268),
            (0, 3),
            (0, 1664),
            (0, 13),
            (0, 1),
            (0, 1),
            (0, 9),
            (0, 94),
            (0, 1886),
            (0, 21),
            (0, 279),
            (0, 16),
            (0, 151),
            (0, 333),
            (0, 229),
            (0, 924),
            (0, 5),
            (0, 67),
            (0, 1),
            (0, 76),
            (0, 203),
            (0, 56),
            (0, 79),
            (0, 1071),
            (0, 0),
            (0, 1676),
            (0, 0),
            (0, 173),
            (0, 14),
            (0, 10),
            (0, 3),
            (0, 45),
            (0, 1),
            (0, 3),
            (0, 50),
            (0, 72),
            (0, 73),
            (0, 27),
            (0, 0),
            (0, 88),
            (0, 237),
            (0, 72),
            (0, 576),
            (0, 18),
            (0, 1040),
            (0, 2205),
            (0, 814),
            (0, 16),
            (0, 3),
            (0, 40),
            (0, 97),
            (0, 47),
            (0, 2959),
            (0, 6),
            (0, 188),
            (0, 0),
            (0, 5),
            (0, 16),
            (0, 32),
            (0, 61),
            (0, 516),
            (0, 31),
            (0, 3141),
            (0, 50),
            (0, 104),
            (0, 0),
            (0, 159),
            (0, 20),
            (0, 2151),
            (0, 115),
            (0, 31),
            (0, 18),
            (0, 122),
            (0, 784),
            (0, 1231),
            (0, 5),
            (0, 600),
            (0, 89),
            (0, 10),
            (0, 0),
            (0, 38),
            (0, 430),
            (0, 10),
            (0, 55),
            (0, 145),
            (0, 163),
            (0, 1),
            (0, 241),
            (0, 151),
            (0, 170),
            (0, 5),
            (0, 973),
            (0, 20),
            (0, 157),
            (0, 1062),
            (0, 346),
            (0, 10),
            (0, 585),
            (0, 7),
            (0, 0),
            (0, 1123),
            (0, 2),
            (0, 4),
            (0, 105),
            (0, 0),
            (0, 12),
            (0, 42),
            (0, 1152),
            (0, 513),
            (0, 4347),
            (0, 0),
            (0, 3),
            (0, 273),
            (0, 4),
            (0, 160),
            (0, 6),
            (0, 1578),
            (0, 195),
            (0, 11),
            (0, 0),
            (0, 0),
            (0, 1618),
            (0, 32),
            (0, 1),
            (0, 100),
            (0, 0),
            (0, 857),
            (0, 542),
            (0, 3),
            (0, 4),
            (0, 862),
            (0, 17),
            (0, 225),
            (0, 44),
            (0, 344),
            (0, 26),
            (0, 16),
            (0, 251),
            (0, 468),
            (0, 2281),
            (0, 1),
            (0, 29),
            (0, 752),
            (0, 2),
            (0, 371),
            (0, 29),
            (0, 9),
            (0, 0),
            (0, 6),
            (0, 1068),
            (0, 48),
            (0, 258),
            (0, 25),
            (0, 1987),
            (0, 2120),
            (0, 0),
            (0, 60),
            (0, 421),
            (0, 1),
            (0, 231),
            (0, 478),
            (0, 6),
            (0, 39),
            (0, 0),
            (0, 0),
            (0, 360),
            (0, 1224),
            (0, 7),
            (0, 16),
            (0, 244),
            (0, 250),
            (0, 38),
            (0, 708),
            (0, 109),
            (0, 38),
            (0, 163),
            (0, 109),
            (0, 1963),
            (0, 743),
            (0, 53),
            (0, 194),
            (0, 32),
            (0, 557),
            (0, 1),
            (0, 994),
            (0, 108),
            (0, 16),
            (0, 503),
            (0, 309),
            (0, 615),
            (0, 10),
            (0, 50),
            (0, 26),
            (0, 9),
            (0, 35),
            (0, 3475),
            (0, 135),
            (0, 22),
            (0, 576),
            (0, 7),
            (0, 157),
            (0, 2),
            (0, 19),
            (0, 379),
            (0, 192),
            (0, 2569),
            (0, 1),
            (0, 714),
            (0, 4),
            (0, 253),
            (0, 649),
            (0, 77),
            (0, 1082),
            (0, 297),
            (0, 32),
            (0, 0),
            (0, 6),
            (0, 8),
            (0, 1),
            (0, 5),
            (0, 39),
            (0, 2889),
            (0, 1504),
            (0, 0),
            (0, 52),
            (0, 4902),
            (0, 844),
            (0, 867),
            (0, 0),
            (0, 18),
            (0, 187),
            (0, 327),
            (0, 92),
            (0, 1),
            (0, 6),
            (0, 812),
            (0, 256),
            (0, 86),
            (0, 2),
            (0, 162),
            (0, 9),
            (0, 0),
            (0, 265),
            (0, 1050),
            (0, 333),
            (0, 7),
            (0, 11),
            (0, 71),
            (0, 3),
            (0, 17),
            (0, 272),
            (0, 2),
            (0, 1271),
            (0, 211),
            (0, 2272),
            (0, 49),
            (0, 253),
            (0, 5242),
            (0, 176),
            (0, 762),
            (0, 1),
            (0, 39),
            (0, 126),
            (0, 6),
            (0, 13),
            (0, 0),
            (0, 13),
            (0, 5),
            (0, 20),
            (0, 135),
            (0, 354),
            (0, 74),
            (0, 291),
            (0, 57),
            (0, 2939),
            (0, 31),
            (0, 10),
            (0, 299),
            (0, 1739),
            (0, 231),
            (0, 113),
            (0, 12),
            (0, 44),
            (0, 64),
            (0, 4163),
            (0, 1205),
            (0, 1097),
            (0, 496),
            (0, 2),
            (0, 111),
            (0, 549),
            (0, 33),
            (0, 33),
            (0, 924),
            (0, 727),
            (0, 7),
            (0, 53),
            (0, 267),
            (0, 12),
            (0, 3),
            (0, 2),
            (0, 2082),
            (0, 8),
            (0, 5),
            (0, 1663),
            (0, 2434),
            (0, 0),
            (0, 1322),
            (0, 256),
            (0, 5),
            (0, 4),
            (0, 3),
            (0, 50),
            (0, 11),
            (0, 1),
            (0, 895),
            (0, 457),
            (0, 23),
            (0, 747),
            (0, 406),
            (0, 4),
            (0, 2),
            (0, 12),
            (0, 2),
            (0, 47),
            (0, 240),
            (0, 29),
            (0, 345),
            (0, 1354),
            (0, 291),
            (0, 2325),
            (0, 1778),
            (0, 203),
            (0, 2),
            (0, 0),
            (0, 31),
            (0, 15),
            (0, 565),
            (0, 1313),
            (0, 1),
            (0, 3),
            (0, 1459),
            (0, 865),
            (0, 662),
            (0, 103),
            (0, 29),
            (0, 5),
            (0, 122),
            (0, 1308),
            (0, 47),
            (0, 4),
            (0, 771),
            (0, 430),
            (0, 243),
            (0, 94),
            (0, 651),
            (0, 2),
            (0, 87),
            (0, 910),
            (0, 31),
            (0, 2470),
            (0, 1),
            (0, 1366),
            (0, 1),
            (0, 6),
            (0, 71),
            (0, 0),
            (0, 23),
            (0, 209),
            (0, 1565),
            (0, 618),
            (0, 0),
            (0, 8),
            (0, 393),
            (0, 587),
            (0, 5),
            (0, 0),
            (0, 917),
            (0, 25),
            (0, 58),
            (0, 4346),
            (0, 370),
            (0, 1),
            (0, 5),
            (0, 10),
            (0, 1),
            (0, 0),
            (0, 219),
            (0, 47),
            (0, 2),
            (0, 238),
            (0, 22),
            (0, 61),
            (0, 0),
            (0, 365),
            (0, 338),
            (0, 2406),
            (0, 76),
            (0, 9),
            (0, 20),
            (0, 1063),
            (0, 2),
            (0, 49),
            (0, 94),
            (0, 125),
            (0, 75),
            (0, 9),
            (0, 1447),
            (0, 114),
            (0, 287),
            (0, 49),
            (0, 865),
            (0, 186),
            (0, 36),
            (0, 6922),
            (0, 21),
            (0, 1),
            (0, 12),
            (0, 106),
            (0, 501),
            (0, 140),
            (0, 273),
            (0, 16),
            (0, 19),
            (0, 39),
            (0, 7),
            (0, 415),
            (0, 6),
            (0, 1014),
            (0, 0),
            (0, 32),
            (0, 145),
            (0, 744),
            (0, 818),
            (0, 4),
            (0, 0),
            (0, 242),
            (0, 6),
            (0, 1),
            (0, 1261),
            (0, 2068),
            (0, 0),
            (0, 0),
            (0, 2),
            (0, 76),
            (0, 899),
            (0, 19),
            (0, 1163),
            (0, 5288),
            (0, 155),
            (0, 848),
            (0, 0),
            (0, 8),
            (0, 148),
            (0, 30),
            (0, 1408),
            (0, 1),
            (0, 0),
            (0, 45),
            (0, 22),
            (0, 214),
            (0, 211),
            (0, 305),
            (0, 354),
            (0, 13),
            (0, 333),
            (0, 43),
            (0, 242),
            (0, 83),
            (0, 369),
            (0, 2007),
            (0, 59),
            (0, 790),
            (0, 2),
            (0, 20),
            (0, 1990),
            (0, 0),
            (0, 8),
            (0, 845),
            (0, 1943),
            (0, 826),
            (0, 0),
            (0, 4),
            (0, 808),
            (0, 362),
            (0, 345),
            (0, 4434),
            (0, 68),
            (0, 8),
            (0, 3),
            (0, 1),
            (0, 631),
            (0, 25),
            (0, 40),
            (0, 1078),
            (0, 5),
            (0, 0),
            (0, 75),
            (0, 166),
            (0, 15),
            (0, 3216),
            (0, 159),
            (0, 197),
            (0, 0),
            (0, 82),
            (0, 53),
            (0, 474),
            (0, 8),
            (0, 63),
            (0, 278),
            (0, 0),
            (0, 23),
            (0, 11),
            (0, 51),
            (0, 8),
            (0, 2),
            (0, 524),
            (0, 6),
            (0, 3206),
            (0, 39),
            (0, 37),
            (0, 1670),
            (0, 85),
            (0, 361),
            (0, 9),
            (0, 5),
            (0, 191),
            (0, 39),
            (0, 330),
            (0, 124),
            (0, 58),
            (0, 182),
            (0, 1089),
            (0, 45),
            (0, 49),
            (0, 621),
            (0, 1),
            (0, 5),
            (0, 129),
            (0, 53),
            (0, 1582),
            (0, 223),
            (0, 2977),
            (0, 392),
            (0, 73),
            (0, 38),
            (0, 4),
            (0, 229),
            (0, 1120),
            (0, 345),
            (0, 405),
            (0, 5040),
            (0, 9),
            (0, 112),
            (0, 5),
            (0, 507),
            (0, 3),
            (0, 1755),
            (0, 24),
            (0, 9),
            (0, 733),
            (0, 2),
            (0, 40),
            (0, 808),
            (0, 191),
            (0, 3),
            (0, 56),
            (0, 98),
            (0, 172),
            (0, 1),
            (0, 73),
            (0, 1456),
            (0, 3),
            (0, 340),
            (0, 220),
            (0, 996),
            (0, 2),
            (0, 0),
            (0, 9),
            (0, 7),
            (0, 10),
            (0, 1880),
            (0, 86),
            (0, 105),
            (0, 304),
            (0, 89),
            (0, 145),
            (0, 245),
            (0, 389),
            (0, 15),
            (0, 156),
            (0, 185),
            (0, 3270),
            (0, 105),
            (0, 42),
            (0, 1),
            (0, 745),
            (0, 1393),
            (0, 2),
            (0, 42),
            (0, 111),
            (0, 69),
            (0, 7),
            (0, 90),
            (0, 469),
            (0, 301),
            (0, 11),
            (0, 92),
            (0, 1),
            (0, 4),
            (0, 622),
            (0, 7),
            (0, 0),
            (0, 225),
            (0, 335),
            (0, 6),
            (0, 0),
            (0, 2327),
            (0, 0),
            (0, 749),
            (0, 47),
            (0, 854),
            (0, 57),
            (0, 49),
            (0, 3),
            (0, 44),
            (0, 26),
            (0, 488),
            (0, 960),
            (0, 16),
            (0, 57),
            (0, 632),
            (0, 3),
            (0, 12),
            (0, 317),
            (0, 783),
            (0, 79),
            (0, 4078),
            (0, 2),
            (0, 574),
            (0, 3),
            (0, 4115),
            (0, 598),
            (0, 5),
            (0, 14),
            (0, 15),
            (0, 42),
            (0, 153),
            (0, 2901),
            (0, 0),
            (0, 141),
            (0, 0),
            (0, 1),
            (0, 961),
            (0, 0),
            (0, 1),
            (0, 0),
            (0, 38),
            (0, 36),
            (0, 16),
            (0, 2),
            (0, 612),
            (0, 124),
            (0, 0),
            (0, 195),
            (0, 790),
            (0, 2421),
            (0, 101),
            (0, 16),
            (0, 76),
            (0, 1275),
            (0, 324),
            (0, 1683),
            (0, 2229),
            (0, 6),
            (0, 106),
            (0, 57),
            (0, 2135),
            (0, 19),
            (0, 0),
            (0, 2021),
            (0, 387),
            (0, 605),
            (0, 54),
            (0, 14),
            (0, 2),
            (0, 10),
            (0, 8),
            (0, 18),
            (0, 825),
            (0, 47),
            (0, 44),
            (0, 548),
            (0, 44),
            (0, 447),
            (0, 9),
            (0, 0),
            (0, 0),
            (0, 1546),
            (0, 29),
            (0, 73),
            (0, 208),
            (0, 83),
            (0, 0),
            (0, 210),
            (0, 40),
            (0, 1),
            (0, 255),
            (0, 45),
            (0, 75),
            (0, 2646),
            (0, 15),
            (0, 107),
            (0, 538),
            (0, 844),
            (0, 214),
            (0, 95),
            (0, 831),
            (0, 2144),
            (0, 809),
            (0, 0),
            (0, 144),
            (0, 0),
            (0, 1),
            (0, 16),
            (0, 2),
            (0, 142),
            (0, 4128),
            (0, 0),
            (0, 0),
            (0, 858),
            (0, 286),
            (0, 32),
            (0, 0),
            (0, 0),
            (0, 1832),
            (0, 1),
            (0, 782),
            (0, 0),
            (0, 3),
            (0, 12),
            (0, 1032),
            (0, 1382),
            (0, 168),
            (0, 395),
            (0, 1251),
            (0, 808),
            (0, 13),
            (0, 1165),
            (0, 84),
            (0, 6),
            (0, 548),
            (0, 489),
            (0, 673),
            (0, 1187),
            (0, 1310),
            (0, 2159),
            (0, 149),
            (0, 17),
            (0, 1),
            (0, 100),
            (0, 146),
            (0, 266),
            (0, 76),
            (0, 92),
            (0, 0),
            (0, 105),
            (0, 1433),
            (0, 24),
            (0, 2),
            (0, 28),
            (0, 0),
            (0, 2),
            (0, 412),
            (0, 330),
            (0, 95),
            (0, 188),
            (0, 269),
            (0, 200),
            (0, 1),
            (0, 72),
            (0, 0),
            (0, 254),
            (0, 123),
            (0, 1406),
            (0, 361),
            (0, 15),
            (0, 0),
            (0, 1174),
            (0, 786),
            (0, 111),
            (0, 494),
            (0, 1350),
            (0, 25),
            (0, 151),
            (0, 25),
            (0, 20),
            (0, 17),
            (0, 3580),
            (0, 99),
            (0, 1640),
            (0, 165),
            (0, 135),
            (0, 1664),
            (0, 199),
            (0, 2),
            (0, 5),
            (0, 878),
            (0, 0),
            (0, 42),
            (0, 0),
            (0, 24),
            (0, 265),
            (0, 88),
            (0, 49),
            (0, 14),
            (0, 1055),
            (0, 529),
            (0, 2244),
            (0, 6),
            (0, 61),
            (0, 1),
            (0, 2),
            (0, 37),
            (0, 172),
            (0, 25),
            (0, 7),
            (0, 0),
            (0, 23),
            (0, 174),
            (0, 28),
            (0, 170),
            (0, 331),
            (0, 0),
            (0, 11),
            (0, 2116),
            (0, 4),
            (0, 26),
            (0, 104),
            (0, 2),
            (0, 11),
            (0, 5),
            (0, 0),
            (0, 120),
            (0, 791),
            (0, 117),
            (0, 27),
            (0, 123),
            (0, 38),
            (0, 19),
            (0, 720),
            (0, 106),
            (0, 290),
            (0, 17),
            (0, 733),
            (0, 85),
            (0, 2),
            (0, 1),
            (0, 397),
            (0, 1),
            (0, 1276),
            (0, 2),
            (0, 390),
            (0, 2282),
            (0, 9),
            (0, 1335),
            (0, 3083),
            (0, 153),
            (0, 8),
            (0, 0),
            (0, 24),
            (0, 263),
            (0, 1779),
            (0, 340),
            (0, 358),
            (0, 299),
            (0, 58),
            (0, 65),
            (0, 1340),
            (0, 664),
            (0, 3129),
            (0, 326),
            (0, 1),
            (0, 379),
            (0, 12),
            (0, 67),
            (0, 29),
            (0, 174),
            (0, 3157),
            (0, 11),
            (0, 35),
            (0, 0),
            (0, 291),
            (0, 58),
            (0, 16),
            (0, 337),
            (0, 7),
            (0, 2443),
            (0, 822),
            (0, 235),
            (0, 98),
            (0, 4),
            (0, 1219),
            (0, 86),
            (0, 61),
            (0, 21),
            (0, 21),
            (0, 10),
            (0, 55),
            (0, 8),
            (0, 17),
            (0, 30),
            (0, 5),
            (0, 175),
            (0, 14),
            (0, 115),
            (0, 82),
            (0, 7),
            (0, 971),
            (0, 20),
            (0, 62),
            (0, 58),
            (0, 8),
            (0, 596),
            (0, 2351),
            (0, 78),
            (0, 313),
            (0, 252),
            (0, 1748),
            (0, 0),
            (0, 32),
            (0, 16),
            (0, 92),
            (0, 53),
            (0, 10),
            (0, 1),
            (0, 438),
            (0, 7874),
            (0, 5),
            (0, 0),
            (0, 24),
            (0, 4062),
            (0, 0),
            (0, 297),
            (0, 45),
            (0, 21),
            (0, 879),
            (0, 1003),
            (0, 87),
            (0, 2),
            (0, 1),
            (0, 939),
            (0, 893),
            (0, 0),
            (0, 4),
            (0, 17),
            (0, 16),
            (0, 31),
            (0, 77),
            (0, 37),
            (0, 24),
            (0, 24),
            (0, 48),
            (0, 43),
            (0, 414),
            (0, 4),
            (0, 7),
            (0, 20),
            (0, 26),
            (0, 660),
            (0, 1622),
            (0, 586),
            (0, 5),
            (0, 11),
            (0, 77),
            (0, 917),
            (0, 353),
            (0, 160),
            (0, 353),
            (0, 87),
            (0, 18),
            (0, 1677),
            (0, 31),
            (0, 78),
            (0, 336),
            (0, 7),
            (0, 2),
            (0, 3),
            (0, 131),
            (0, 41),
            (0, 4823),
            (0, 680),
            (0, 22),
            (0, 540),
            (0, 5),
            (0, 284),
            (0, 50),
            (0, 2251),
            (0, 12),
            (0, 192),
            (0, 44),
            (0, 68),
            (0, 1),
            (0, 418),
            (0, 1114),
            (0, 377),
            (0, 68),
            (0, 2961),
            (0, 362),
            (0, 5),
            (0, 20),
            (0, 538),
            (0, 349),
            (0, 48),
            (0, 124),
            (0, 8),
            (0, 338),
            (0, 1),
            (0, 19),
            (0, 11),
            (0, 1057),
            (0, 609),
            (0, 365),
            (0, 589),
            (0, 167),
            (0, 290),
            (0, 318),
            (0, 96),
            (0, 4),
            (0, 371),
            (0, 9),
            (0, 72),
            (0, 57),
            (0, 980),
            (0, 160),
            (0, 991),
            (0, 11),
            (0, 0),
            (0, 3042),
            (0, 221),
            (0, 5),
            (0, 2),
            (0, 27),
            (0, 575),
            (0, 2),
            (0, 274),
            (0, 5),
            (0, 436),
            (0, 21),
            (0, 3677),
            (0, 28),
            (0, 15),
            (0, 22),
            (0, 6),
            (0, 1),
            (0, 84),
            (0, 0),
            (0, 7),
            (0, 15),
            (0, 94),
            (0, 54),
            (0, 0),
            (0, 7),
            (0, 582),
            (0, 185),
            (0, 7),
            (0, 583),
            (0, 26),
            (0, 277),
            (0, 730),
            (0, 225),
            (0, 354),
            (0, 3257),
            (0, 147),
            (0, 10),
            (0, 5),
            (0, 689),
            (0, 219),
            (0, 43),
            (0, 7),
            (0, 243),
            (0, 943),
            (0, 58),
            (0, 11),
            (0, 8),
            (0, 31),
            (0, 1),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 1),
            (0, 80),
            (0, 18),
            (0, 3662),
            (0, 149),
            (0, 2),
            (0, 1944),
            (0, 228),
            (0, 524),
            (0, 340),
            (0, 6),
            (0, 101),
            (0, 0),
            (0, 407),
            (0, 3896),
            (0, 559),
            (0, 1068),
            (0, 716),
            (0, 833),
            (0, 265),
            (0, 157),
            (0, 1287),
            (0, 0),
            (0, 25),
            (0, 701),
            (0, 527),
            (0, 61),
            (0, 650),
            (0, 1833),
            (0, 2942),
            (0, 2023),
            (0, 126),
            (0, 1962),
            (0, 71),
            (0, 439),
            (0, 135),
            (0, 38),
            (0, 162),
            (0, 291),
            (0, 1114),
            (0, 21),
            (0, 1075),
            (0, 160),
            (0, 4726),
            (0, 85),
            (0, 403),
            (0, 288),
            (0, 104),
            (0, 0),
            (0, 24),
            (0, 305),
            (0, 346),
            (0, 643),
            (0, 581),
            (0, 0),
            (0, 581),
            (0, 0),
            (0, 2),
            (0, 10),
            (0, 29),
            (0, 73),
            (0, 69),
            (0, 236),
            (0, 2),
            (0, 11),
            (0, 317),
            (0, 54),
            (0, 330),
            (0, 346),
            (0, 7),
            (0, 371),
            (0, 449),
            (0, 382),
            (0, 390),
            (0, 32),
            (0, 6228),
            (0, 5),
            (0, 4186),
            (0, 2658),
            (0, 2),
            (0, 12),
            (0, 54),
            (0, 250),
            (0, 1621),
            (0, 629),
            (0, 1),
            (0, 11),
            (0, 54),
            (0, 0),
            (0, 22),
            (0, 4),
            (0, 15),
            (0, 43),
            (0, 2764),
            (0, 258),
            (0, 4143),
            (0, 16),
            (0, 55),
            (0, 996),
            (0, 266),
            (0, 516),
            (0, 216),
            (0, 2541),
            (0, 1875),
            (0, 1408),
            (0, 59),
            (0, 22),
            (0, 334),
            (0, 49),
            (0, 693),
            (0, 24),
            (0, 19),
            (0, 2),
            (0, 7),
            (0, 1453),
            (0, 20),
            (0, 5),
            (0, 906),
            (0, 152),
            (0, 60),
            (0, 23),
            (0, 7),
            (0, 2518),
            (0, 1),
            (0, 353),
            (0, 1),
            (0, 417),
            (0, 8),
            (0, 5100),
            (0, 779),
            (0, 163),
            (0, 2),
            (0, 697),
            (0, 238),
            (0, 381),
            (0, 50),
            (0, 131),
            (0, 9),
            (0, 584),
            (0, 203),
            (0, 45),
            (0, 49),
            (0, 458),
            (0, 963),
            (0, 22),
            (0, 1737),
            (0, 7),
            (0, 164),
            (0, 8),
            (0, 63),
            (0, 0),
            (0, 2788),
            (0, 1),
            (0, 2),
            (0, 19),
            (0, 15),
            (0, 2126),
            (0, 79),
            (0, 19),
            (0, 981),
            (0, 0),
            (0, 258),
            (0, 1530),
            (0, 651),
            (0, 204),
            (0, 8),
            (0, 1744),
            (0, 297),
            (0, 3),
            (0, 926),
            (0, 8),
            (0, 277),
            (0, 0),
            (0, 27),
            (0, 49),
            (0, 271),
            (0, 175),
            (0, 9),
            (0, 1),
            (0, 45),
            (0, 5),
            (0, 54),
            (0, 2844),
            (0, 247),
            (0, 40),
            (0, 0),
            (0, 29),
            (0, 212),
            (0, 0),
            (0, 347),
            (0, 19),
            (0, 5),
            (0, 32),
            (0, 3),
            (0, 224),
            (0, 6),
            (0, 209),
            (0, 74),
            (0, 451),
            (0, 1428),
            (0, 31),
            (0, 3685),
            (0, 954),
            (0, 298),
            (0, 19),
            (0, 40),
            (0, 427),
            (0, 1770),
            (0, 44),
            (0, 583),
            (0, 0),
            (0, 11),
            (0, 1082),
            (0, 151),
            (0, 179),
            (0, 50),
            (0, 54),
            (0, 296),
            (0, 66),
            (0, 160),
            (0, 1),
            (0, 118),
            (0, 681),
            (0, 252),
            (0, 322),
            (0, 1287),
            (0, 328),
            (0, 11),
            (0, 5454),
            (0, 750),
            (0, 80),
            (0, 14),
            (0, 1129),
            (0, 0),
            (0, 29),
            (0, 373),
            (0, 571),
            (0, 19),
            (0, 118),
            (0, 364),
            (0, 56),
            (0, 43),
            (0, 469),
            (0, 1528),
            (0, 5),
            (0, 270),
            (0, 463),
            (0, 64),
            (0, 2),
            (0, 18),
            (0, 49),
            (0, 5),
            (0, 879),
            (0, 263),
            (0, 6),
            (0, 31),
            (0, 631),
            (0, 19),
            (0, 2),
            (0, 988),
            (0, 247),
            (0, 46),
            (0, 105),
            (0, 1521),
            (0, 27),
            (0, 39),
            (0, 143),
            (0, 65),
            (0, 5),
            (0, 13),
            (0, 1976),
            (0, 0),
            (0, 76),
            (0, 1),
            (0, 726),
            (0, 435),
            (0, 194),
            (0, 3112),
            (0, 537),
            (0, 5),
            (0, 2368),
            (0, 52),
            (0, 1036),
            (0, 6),
            (0, 539),
            (0, 227),
            (0, 49),
            (0, 13),
            (0, 1),
            (0, 421),
            (0, 1383),
            (0, 7),
            (0, 0),
            (0, 464),
            (0, 2175),
            (0, 20),
            (0, 882),
            (0, 0),
            (0, 1),
            (0, 1),
            (0, 709),
            (0, 854),
            (0, 60),
            (0, 39),
            (0, 332),
            (0, 126),
            (0, 7),
            (0, 7836),
            (0, 417),
            (0, 157),
            (0, 282),
            (0, 4),
            (0, 6),
            (0, 24),
            (0, 206),
            (0, 26),
            (0, 29),
            (0, 236),
            (0, 207),
            (0, 0),
            (0, 82),
            (0, 85),
            (0, 367),
            (0, 29),
            (0, 1142),
            (0, 686),
            (0, 8),
            (0, 58),
            (0, 1102),
            (0, 0),
            (0, 19),
            (0, 982),
            (0, 5072),
            (0, 23),
            (0, 33),
            (0, 242),
            (0, 77),
            (0, 212),
            (0, 612),
            (0, 151),
            (0, 15),
            (0, 24),
            (0, 2),
            (0, 50),
            (0, 111),
            (0, 1747),
            (0, 1002),
            (0, 9),
            (0, 4),
            (0, 1),
            (0, 2827),
            (0, 12),
            (0, 2),
            (0, 39),
            (0, 154),
            (0, 436),
            (0, 1566),
            (0, 62),
            (0, 11),
            (0, 0),
            (0, 1),
            (0, 1235),
            (0, 32),
            (0, 3380),
            (0, 375),
            (0, 87),
            (0, 30),
            (0, 979),
            (0, 535),
            (0, 1),
            (0, 2647),
            (0, 10),
            (0, 8),
            (0, 1),
            (0, 487),
            (0, 99),
            (0, 12),
            (0, 8),
            (0, 91),
            (0, 603),
            (0, 5),
            (0, 385),
            (0, 5),
            (0, 131),
            (0, 2360),
            (0, 43),
            (0, 301),
            (0, 85),
            (0, 42),
            (0, 2678),
            (0, 3599),
            (0, 72),
            (0, 1),
            (0, 4279),
            (0, 1077),
            (0, 9),
            (0, 738),
            (0, 191),
            (0, 2638),
            (0, 1),
            (0, 136),
            (0, 784),
            (0, 1),
            (0, 262),
            (0, 154),
            (0, 91),
            (0, 147),
            (0, 0),
            (0, 663),
            (0, 187),
            (0, 56),
            (0, 466),
            (0, 366),
            (0, 21),
            (0, 356),
            (0, 917),
            (0, 17),
            (0, 50),
            (0, 709),
            (0, 1),
            (0, 28),
            (0, 4442),
            (0, 9),
            (0, 17),
            (0, 421),
            (0, 11502),
            (0, 15),
            (0, 180),
            (0, 479),
            (0, 48),
            (0, 186),
            (0, 1),
            (0, 0),
            (0, 4236),
            (0, 3352),
            (0, 28),
            (0, 3),
            (0, 0),
            (0, 5),
            (0, 42),
            (0, 7),
            (0, 46),
            (0, 1548),
            (0, 1418),
            (0, 127),
            (0, 14),
            (0, 193),
            (0, 266),
            (0, 56),
            (0, 299),
            (0, 2354),
            (0, 1756),
            (0, 3),
            (0, 133),
            (0, 8),
            (0, 1427),
            (0, 388),
            (0, 11),
            (0, 306),
            (0, 44),
            (0, 2765),
            (0, 5),
            (0, 25),
            (0, 0),
            (0, 3),
            (0, 71),
            (0, 1227),
            (0, 242),
            (0, 547),
            (0, 703),
            (0, 505),
            (0, 36),
            (0, 221),
            (0, 156),
            (0, 20),
            (0, 11),
            (0, 14),
            (0, 9),
            (0, 37),
            (0, 5306),
            (0, 61),
            (0, 738),
            (0, 3),
            (0, 101),
            (0, 6),
            (0, 13),
            (0, 426),
            (0, 41),
            (0, 0),
            (0, 1),
            (0, 2),
            (0, 6),
            (0, 1095),
            (0, 63),
            (0, 2),
            (0, 11),
            (0, 1347),
            (0, 7),
            (0, 856),
            (0, 163),
            (0, 103),
            (0, 40),
            (0, 92),
            (0, 15),
            (0, 659),
            (0, 132),
            (0, 57),
            (0, 1),
            (0, 106),
            (0, 3),
            (0, 192),
            (0, 9),
            (0, 10),
            (0, 7),
            (0, 21),
            (0, 20),
            (0, 116),
            (0, 8400),
            (0, 0),
            (0, 0),
            (0, 91),
            (0, 1208),
            (0, 708),
            (0, 59),
            (0, 189),
            (0, 129),
            (0, 10),
            (0, 287),
            (0, 11),
            (0, 0),
            (0, 75),
            (0, 413),
            (0, 351),
            (0, 1),
            (0, 329),
            (0, 21),
            (0, 2),
            (0, 371),
            (0, 108),
            (0, 636),
            (0, 49),
            (0, 87),
            (0, 334),
            (0, 3),
            (0, 1539),
            (0, 84),
            (0, 2223),
            (0, 283),
            (0, 49),
            (0, 10),
            (0, 347),
            (0, 116),
            (0, 111),
            (0, 345),
            (0, 10),
            (0, 95),
            (0, 8),
            (0, 8881),
            (0, 382),
            (0, 83),
            (0, 7),
            (0, 976),
            (0, 3199),
            (0, 10),
            (0, 56),
            (0, 62),
            (0, 3),
            (0, 97),
            (0, 353),
            (0, 0),
            (0, 246),
            (0, 32),
            (0, 39),
            (0, 468),
            (0, 33),
            (0, 1267),
            (0, 11),
            (0, 424),
            (0, 18),
            (0, 577),
            (0, 1),
            (0, 484),
            (0, 741),
            (0, 1),
            (0, 101),
            (0, 307),
            (0, 0),
            (0, 3498),
            (0, 735),
            (0, 252),
            (0, 230),
            (0, 173),
            (0, 3631),
            (0, 3349),
            (0, 156),
            (0, 23),
            (0, 0),
            (0, 19),
            (0, 8),
            (0, 30),
            (0, 31),
            (0, 255),
            (0, 76),
            (0, 285),
            (0, 0),
            (0, 96),
            (0, 155),
            (0, 855),
            (0, 0),
            (0, 373),
            (0, 101),
            (0, 1196),
            (0, 280),
            (0, 110),
            (0, 963),
            (0, 5),
            (0, 2),
            (0, 155),
            (0, 245),
            (0, 764),
            (0, 7),
            (0, 15),
            (0, 1473),
            (0, 9),
            (0, 215),
            (0, 283),
            (0, 2),
            (0, 5),
            (0, 704),
            (0, 36),
            (0, 2264),
            (0, 3),
            (0, 9),
            (0, 1057),
            (0, 29),
            (0, 153),
            (0, 348),
            (0, 93),
            (0, 11),
            (0, 89),
            (0, 1365),
            (0, 149),
            (0, 542),
            (0, 1),
            (0, 6),
            (0, 185),
            (0, 58),
            (0, 3369),
            (0, 0),
            (0, 1),
            (0, 25),
            (0, 79),
            (0, 3),
            (0, 117),
            (0, 0),
            (0, 87),
            (0, 300),
            (0, 101),
            (0, 354),
            (0, 3982),
            (0, 383),
            (0, 1351),
            (0, 2037),
            (0, 22),
            (0, 87),
            (0, 148),
            (0, 0),
            (0, 0),
            (0, 225),
            (0, 106),
            (0, 9),
            (0, 26),
            (0, 0),
            (0, 146),
            (0, 175),
            (0, 375),
            (0, 0),
            (0, 4912),
            (0, 0),
            (0, 5),
            (0, 56),
            (0, 0),
            (0, 175),
            (0, 1887),
            (0, 4),
            (0, 135),
            (0, 0),
            (0, 0),
            (0, 24),
            (0, 337),
            (0, 2405),
            (0, 198),
            (0, 9),
            (0, 0),
            (0, 125),
            (0, 7),
            (0, 0),
            (0, 186),
            (0, 53),
            (0, 1107),
            (0, 112),
            (0, 251),
            (0, 2),
            (0, 29),
            (0, 164),
            (0, 543),
            (0, 1216),
            (0, 6),
            (0, 754),
            (0, 127),
            (0, 28),
            (0, 2153),
            (0, 6),
            (0, 318),
            (0, 6),
            (0, 29),
            (0, 13),
            (0, 4275),
            (0, 376),
            (0, 189),
            (0, 1395),
            (0, 0),
            (0, 38),
            (0, 299),
            (0, 160),
            (0, 3120),
            (0, 16),
            (0, 142),
            (0, 398),
            (0, 64),
            (0, 830),
            (0, 58),
            (0, 582),
            (0, 7),
            (0, 258),
            (0, 1609),
            (0, 18),
            (0, 10),
            (0, 4),
            (0, 3),
            (0, 3796),
            (0, 272),
            (0, 8),
            (0, 2641),
            (0, 2),
            (0, 1076),
            (0, 84),
            (0, 8),
            (0, 0),
            (0, 12),
            (0, 3),
            (0, 17),
            (0, 1956),
            (0, 0),
            (0, 1),
            (0, 368),
            (0, 93),
            (0, 18),
            (0, 6),
            (0, 6),
            (0, 200),
            (0, 618),
            (0, 6158),
            (0, 15),
            (0, 1),
            (0, 1),
            (0, 265),
            (0, 1),
            (0, 70),
            (0, 257),
            (0, 748),
            (0, 2463),
            (0, 19),
            (0, 83),
            (0, 6),
            (0, 178),
            (0, 9),
            (0, 28),
            (0, 65),
            (0, 22),
            (0, 18),
            (0, 517),
            (0, 779),
            (0, 28),
            (0, 132),
            (0, 2622),
            (0, 60),
            (0, 22),
            (0, 6),
            (0, 237),
            (0, 235),
            (0, 1009),
            (0, 381),
            (0, 13),
            (0, 181),
            (0, 1404),
            (0, 18),
            (0, 3),
            (0, 106),
            (0, 422),
            (0, 116),
            (0, 3),
            (0, 0),
            (0, 412),
            (0, 88),
            (0, 153),
            (0, 2287),
            (0, 246),
            (0, 3),
            (0, 0),
            (0, 504),
            (0, 8),
            (0, 2),
            (0, 70),
            (0, 371),
            (0, 205),
            (0, 4),
            (0, 1374),
            (0, 3),
            (0, 0),
            (0, 34),
            (0, 15),
            (0, 136),
            (0, 16),
            (0, 931),
            (0, 8),
            (0, 18),
            (0, 485),
            (0, 4),
            (0, 189),
            (0, 1831),
            (0, 3),
            (0, 4),
            (0, 532),
            (0, 2),
            (0, 0),
            (0, 84),
            (0, 0),
            (0, 1261),
            (0, 57),
            (0, 339),
            (0, 0),
            (0, 743),
            (0, 261),
            (0, 271),
            (0, 18),
            (0, 199),
            (0, 20),
            (0, 95),
            (0, 669),
            (0, 825),
            (0, 2177),
            (0, 0),
            (0, 1),
            (0, 31),
            (0, 1),
            (0, 688),
            (0, 0),
            (0, 3),
            (0, 12),
            (0, 19),
            (0, 380),
            (0, 625),
            (0, 1964),
            (0, 379),
            (0, 166),
            (0, 192),
            (0, 166),
            (0, 783),
            (0, 226),
            (0, 219),
            (0, 2724),
            (0, 59),
            (0, 268),
            (0, 56),
            (0, 85),
            (0, 1969),
            (0, 29),
            (0, 1),
            (0, 623),
            (0, 459),
            (0, 810),
            (0, 202),
            (0, 87),
            (0, 0),
            (0, 174),
            (0, 45),
            (0, 2),
            (0, 2),
            (0, 0),
            (0, 1130),
            (0, 9),
            (0, 579),
            (0, 0),
            (0, 102),
            (0, 1),
            (0, 158),
            (0, 6),
            (0, 0),
            (0, 7),
            (0, 993),
            (0, 7),
            (0, 547),
            (0, 30),
            (0, 384),
            (0, 257),
            (0, 882),
            (0, 3),
            (0, 56),
            (0, 2643),
            (0, 117),
            (0, 61),
            (0, 0),
            (0, 8),
            (0, 24),
            (0, 0),
            (0, 953),
            (0, 11),
            (0, 301),
            (0, 857),
            (0, 3),
            (0, 1016),
            (0, 256),
            (0, 286),
            (0, 31),
            (0, 25),
            (0, 6),
            (0, 351),
            (0, 3710),
            (0, 320),
            (0, 29),
            (0, 4427),
            (0, 97),
            (0, 26),
            (0, 1260),
            (0, 2463),
            (0, 7),
            (0, 125),
            (0, 180),
            (0, 2),
            (0, 3),
            (0, 1),
            (0, 13),
            (0, 129),
            (0, 169),
            (0, 18),
            (0, 139),
            (0, 3612),
            (0, 37),
            (0, 7),
            (0, 340),
            (0, 35),
            (0, 9),
            (0, 18),
            (0, 458),
            (0, 5),
            (0, 58),
            (0, 0),
            (0, 1335),
            (0, 17),
            (0, 61),
            (0, 5),
            (0, 1),
            (0, 74),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 145),
            (0, 447),
            (0, 1),
            (0, 350),
            (0, 8),
            (0, 310),
            (0, 22),
            (0, 133),
            (0, 497),
            (0, 4),
            (0, 3365),
            (0, 28),
            (0, 82),
            (0, 61),
            (0, 22),
            (0, 14),
            (0, 1366),
            (0, 11),
            (0, 115),
            (0, 45),
            (0, 11),
            (0, 18),
            (0, 64),
            (0, 260),
            (0, 3),
            (0, 43),
            (0, 12),
            (0, 81),
            (0, 4),
            (0, 138),
            (0, 1381),
            (0, 64),
            (0, 1915),
            (0, 3),
            (0, 13),
            (0, 508),
            (0, 16),
            (0, 157),
            (0, 219),
            (0, 1),
            (0, 125),
            (0, 9),
            (0, 84),
            (0, 1),
            (0, 4547),
            (0, 0),
            (0, 741),
            (0, 7297),
            (0, 7),
            (0, 0),
            (0, 34),
            (0, 26),
            (0, 588),
            (0, 4),
            (0, 2361),
            (0, 2829),
            (0, 5430),
            (0, 327),
            (0, 21),
            (0, 1771),
            (0, 4),
            (0, 268),
            (0, 0),
            (0, 10),
            (0, 1176),
            (0, 129),
            (0, 4),
            (0, 330),
            (0, 1),
            (0, 188),
            (0, 7),
            (0, 19),
            (0, 2),
            (0, 1338),
            (0, 4607),
            (0, 155),
            (0, 120),
            (0, 40),
            (0, 2),
            (0, 79),
            (0, 0),
            (0, 477),
            (0, 33),
            (0, 54),
            (0, 0),
            (0, 126),
            (0, 100),
            (0, 101),
            (0, 639),
            (0, 226),
            (0, 581),
            (0, 4),
            (0, 50),
            (0, 344),
            (0, 0),
            (0, 0),
            (0, 8),
            (0, 0),
            (0, 25),
            (0, 0),
            (0, 56),
            (0, 40),
            (0, 49),
            (0, 70),
            (0, 15),
            (0, 11),
            (0, 10),
            (0, 778),
            (0, 2),
            (0, 3),
            (0, 4),
            (0, 179),
            (0, 119),
            (0, 175),
            (0, 1330),
            (0, 1934),
            (0, 21),
            (0, 166),
            (0, 17),
            (0, 207),
            (0, 0),
            (0, 4),
            (0, 4),
            (0, 7),
            (0, 102),
            (0, 2),
            (0, 337),
            (0, 25),
            (0, 0),
            (0, 533),
            (0, 132),
            (0, 1),
            (0, 698),
            (0, 7),
            (0, 837),
            (0, 497),
            (0, 38),
            (0, 56),
            (0, 258),
            (0, 1),
            (0, 172),
            (0, 165),
            (0, 74),
            (0, 993),
            (0, 4),
            (0, 3),
            (0, 55),
            (0, 67),
            (0, 1123),
            (0, 145),
            (0, 95),
            (0, 214),
            (0, 55),
            (0, 27),
            (0, 681),
            (0, 3327),
            (0, 0),
            (0, 242),
            (0, 779),
            (0, 415),
            (0, 0),
            (0, 225),
            (0, 2760),
            (0, 215),
            (0, 619),
            (0, 36),
            (0, 107),
            (0, 47),
            (0, 9),
            (0, 3375),
            (0, 11),
            (0, 952),
            (0, 8),
            (0, 324),
            (0, 21),
            (0, 0),
            (0, 261),
            (0, 9),
            (0, 389),
            (0, 892),
            (0, 170),
            (0, 261),
            (0, 152),
            (0, 1143),
            (0, 0),
            (0, 0),
            (0, 688),
            (0, 491),
            (0, 2487),
            (0, 20),
            (0, 51),
            (0, 768),
            (0, 1802),
            (0, 209),
            (0, 36),
            (0, 664),
            (0, 0),
            (0, 403),
            (0, 11),
            (0, 0),
            (0, 770),
            (0, 19),
            (0, 389),
            (0, 646),
            (0, 407),
            (0, 14),
            (0, 324),
            (0, 65),
            (0, 194),
            (0, 0),
            (0, 255),
            (0, 9),
            (0, 4254),
            (0, 480),
            (0, 315),
            (0, 0),
            (0, 1119),
            (0, 30),
            (0, 762),
            (0, 333),
            (0, 11),
            (0, 53),
            (0, 4494),
            (0, 34),
            (0, 4),
            (0, 607),
            (0, 1023),
            (0, 65),
            (0, 3),
            (0, 1377),
            (0, 3049),
            (0, 23),
            (0, 229),
            (0, 27),
            (0, 96),
            (0, 218),
            (0, 157),
            (0, 399),
            (0, 462),
            (0, 6),
            (0, 596),
            (0, 0),
            (0, 3),
            (0, 848),
            (0, 0),
            (0, 4006),
            (0, 1866),
            (0, 1949),
            (0, 274),
            (0, 65),
            (0, 46),
            (0, 108),
            (0, 112),
            (0, 33),
            (0, 12),
            (0, 369),
            (0, 116),
            (0, 51),
            (0, 6),
            (0, 2924),
            (0, 134),
            (0, 1),
            (0, 210),
            (0, 9),
            (0, 34),
            (0, 2618),
            (0, 13),
            (0, 59),
            (0, 0),
            (0, 718),
            (0, 2517),
            (0, 0),
            (0, 81),
            (0, 254),
            (0, 106),
            (0, 0),
            (0, 81),
            (0, 4154),
            (0, 719),
            (0, 0),
            (0, 0),
            (0, 39),
            (0, 1),
            (0, 853),
            (0, 2),
            (0, 3585),
            (0, 1351),
            (0, 1513),
            (0, 1075),
            (0, 0),
            (0, 450),
            (0, 297),
            (0, 52),
            (0, 0),
            (0, 579),
            (0, 391),
            (0, 0),
            (0, 2),
            (0, 135),
            (0, 865),
            (0, 789),
            (0, 848),
            (0, 0),
            (0, 140),
            (0, 1334),
            (0, 1485),
            (0, 547),
            (0, 9),
            (0, 69),
            (0, 115),
            (0, 51),
            (0, 13),
            (0, 458),
            (0, 60),
            (0, 11),
            (0, 537),
            (0, 2468),
            (0, 6),
            (0, 21),
            (0, 10405),
            (0, 2),
            (0, 170),
            (0, 0),
            (0, 777),
            (0, 2),
            (0, 97),
            (0, 79),
            (0, 46),
            (0, 2451),
            (0, 399),
            (0, 486),
            (0, 224),
            (0, 3),
            (0, 718),
            (0, 110),
            (0, 5671),
            (0, 15),
            (0, 709),
            (0, 22),
            (0, 3),
            (0, 1298),
            (0, 968),
            (0, 333),
            (0, 10),
            (0, 0),
            (0, 1739),
            (0, 187),
            (0, 107),
            (0, 1710),
            (0, 52),
            (0, 69),
            (0, 0),
            (0, 4),
            (0, 754),
            (0, 129),
            (0, 5111),
            (0, 204),
            (0, 501),
            (0, 21),
            (0, 1),
            (0, 416),
            (0, 1),
            (0, 22),
            (0, 13),
            (0, 738),
            (0, 157),
            (0, 21),
            (0, 14),
            (0, 1081),
            (0, 12),
            (0, 11),
            (0, 1),
            (0, 0),
            (0, 4424),
            (0, 306),
            (0, 6177),
            (0, 374),
            (0, 3),
            (0, 3417),
            (0, 0),
            (0, 21),
            (0, 637),
            (0, 22),
            (0, 447),
            (0, 18),
            (0, 7),
            (0, 47),
            (0, 585),
            (0, 3),
            (0, 1),
            (0, 2476),
            (0, 19),
            (0, 236),
            (0, 83),
            (0, 1),
            (0, 451),
            (0, 26),
            (0, 732),
            (0, 4),
            (0, 963),
            (0, 3216),
            (0, 52),
            (0, 12),
            (0, 302),
            (0, 22),
            (0, 321),
            (0, 37),
            (0, 22),
            (0, 804),
            (0, 289),
            (0, 0),
            (0, 14),
            (0, 81),
            (0, 2187),
            (0, 48),
            (0, 31),
            (0, 824),
            (0, 12),
            (0, 0),
            (0, 1),
            (0, 50),
            (0, 9278),
            (0, 1869),
            (0, 1162),
            (0, 0),
            (0, 0),
            (0, 987),
            (0, 34),
            (0, 46),
            (0, 62),
            (0, 1436),
            (0, 33),
            (0, 739),
            (0, 2525),
            (0, 362),
            (0, 1374),
            (0, 211),
            (0, 0),
            (0, 25),
            (0, 261),
            (0, 1678),
            (0, 0),
            (0, 3),
            (0, 3266),
            (0, 465),
            (0, 356),
            (0, 27),
            (0, 15),
            (0, 161),
            (0, 4582),
            (0, 156),
            (0, 0),
            (0, 1338),
            (0, 1788),
            (0, 58),
            (0, 44),
            (0, 390),
            (0, 1155),
            (0, 9),
            (0, 825),
            (0, 64),
            (0, 198),
            (0, 1),
            (0, 432),
            (0, 2075),
            (0, 314),
            (0, 3),
            (0, 22),
            (0, 0),
            (0, 4),
            (0, 705),
            (0, 6),
            (0, 9),
            (0, 3652),
            (0, 139),
            (0, 3),
            (0, 4),
            (0, 209),
            (0, 5),
            (0, 3144),
            (0, 84),
            (0, 12),
            (0, 0),
            (0, 170),
            (0, 2942),
            (0, 7),
            (0, 0),
            (0, 1789),
            (0, 37),
            (0, 436),
            (0, 606),
            (0, 2),
            (0, 84),
            (0, 11),
            (0, 130),
            (0, 350),
            (0, 4),
            (0, 47),
            (0, 0),
            (0, 0),
            (0, 362),
            (0, 0),
            (0, 82),
            (0, 244),
            (0, 718),
            (0, 14),
            (0, 234),
            (0, 1664),
            (0, 29),
            (0, 1),
            (0, 172),
            (0, 6),
            (0, 353),
            (0, 304),
            (0, 6),
            (0, 31),
            (0, 5),
            (0, 1122),
            (0, 370),
            (0, 1338),
            (0, 320),
            (0, 5),
            (0, 869),
            (0, 1211),
            (0, 72),
            (0, 0),
            (0, 155),
            (0, 224),
            (0, 824),
            (0, 674),
            (0, 304),
            (0, 329),
            (0, 257),
            (0, 7),
            (0, 122),
            (0, 1046),
            (0, 1937),
            (0, 238),
            (0, 3),
            (0, 746),
            (0, 941),
            (0, 38),
            (0, 2176),
            (0, 1),
            (0, 58),
            (0, 1957),
            (0, 9),
            (0, 37),
            (0, 79),
            (0, 94),
            (0, 120),
            (0, 1142),
            (0, 1841),
            (0, 1),
            (0, 84),
            (0, 145),
            (0, 3523),
            (0, 25),
            (0, 363),
            (0, 1504),
            (0, 1),
            (0, 0),
            (0, 118),
            (0, 232),
            (0, 0),
            (0, 95),
            (0, 2149),
            (0, 2771),
            (0, 702),
            (0, 4),
            (0, 1),
            (0, 170),
            (0, 147),
            (0, 42),
            (0, 1710),
            (0, 730),
            (0, 115),
            (0, 1407),
            (0, 26),
            (0, 4143),
            (0, 12),
            (0, 104),
            (0, 97),
            (0, 1215),
            (0, 40),
            (0, 296),
            (0, 5),
            (0, 47),
            (0, 5),
            (0, 91),
            (0, 316),
            (0, 558),
            (0, 3),
            (0, 43),
            (0, 330),
            (0, 700),
            (0, 5),
            (0, 238),
            (0, 8),
            (0, 121),
            (0, 20),
            (0, 5),
            (0, 19),
            (0, 1206),
            (0, 7),
            (0, 186),
            (0, 759),
            (0, 10132),
            (0, 65),
            (0, 6),
            (0, 288),
            (0, 2402),
            (0, 56),
            (0, 217),
            (0, 153),
            (0, 135),
            (0, 22),
            (0, 5154),
            (0, 1),
            (0, 289),
            (0, 1),
            (0, 5),
            (0, 5),
            (0, 43),
            (0, 94),
            (0, 3),
            (0, 0),
            (0, 0),
            (0, 182),
            (0, 606),
            (0, 2),
            (0, 1280),
            (0, 43),
            (0, 3092),
            (0, 4),
            (0, 12),
            (0, 582),
            (0, 1495),
            (0, 11),
            (0, 259),
            (0, 405),
            (0, 4),
            (0, 0),
            (0, 154),
            (0, 0),
            (0, 55),
            (0, 1),
            (0, 147),
            (0, 8),
            (0, 1485),
            (0, 0),
            (0, 4611),
            (0, 4332),
            (0, 569),
            (0, 167),
            (0, 247),
            (0, 431),
            (0, 143),
            (0, 162),
            (0, 412),
            (0, 10),
            (0, 87),
            (0, 1381),
            (0, 482),
            (0, 1378),
            (0, 110),
            (0, 193),
            (0, 50),
            (0, 27),
            (0, 153),
            (0, 380),
            (0, 420),
            (0, 66),
            (0, 98),
            (0, 11),
            (0, 1046),
            (0, 66),
            (0, 4024),
            (0, 58),
            (0, 5225),
            (0, 667),
            (0, 684),
            (0, 15),
            (0, 1),
            (0, 5374),
            (0, 0),
            (0, 1717),
            (0, 11),
            (0, 2),
            (0, 4176),
            (0, 817),
            (0, 0),
            (0, 4),
            (0, 14),
            (0, 17),
            (0, 0),
            (0, 170),
            (0, 457),
            (0, 15),
            (0, 854),
            (0, 450),
            (0, 1),
            (0, 52),
            (0, 615),
            (0, 30),
            (0, 523),
            (0, 7),
            (0, 10),
            (0, 61),
            (0, 671),
            (0, 0),
            (0, 104),
            (0, 182),
            (0, 308),
            (0, 432),
            (0, 0),
            (0, 38),
            (0, 2),
            (0, 0),
            (0, 649),
            (0, 0),
            (0, 1487),
            (0, 3),
            (0, 11),
            (0, 234),
            (0, 496),
            (0, 160),
            (0, 27),
            (0, 4),
            (0, 2),
            (0, 31),
            (0, 312),
            (0, 20),
            (0, 13954),
            (0, 63),
            (0, 26),
            (0, 88),
            (0, 333),
            (0, 3),
            (0, 992),
            (0, 586),
            (0, 24),
            (0, 232),
            (0, 89),
            (0, 0),
            (0, 715),
            (0, 65),
            (0, 3),
            (0, 41),
            (0, 554),
            (0, 28),
            (0, 1),
            (0, 2),
            (0, 1),
            (0, 58),
            (0, 282),
            (0, 1058),
            (0, 46),
            (0, 19),
            (0, 1257),
            (0, 0),
            (0, 0),
            (0, 321),
            (0, 6),
            (0, 4),
            (0, 4),
            (0, 622),
            (0, 48),
            (0, 317),
            (0, 8),
            (0, 747),
            (0, 350),
            (0, 27),
            (0, 333),
            (0, 2),
            (0, 497),
            (0, 7),
            (0, 179),
            (0, 27),
            (0, 618),
            (0, 218),
            (0, 0),
            (0, 101),
            (0, 346),
            (0, 43),
            (0, 1069),
            (0, 400),
            (0, 3319),
            (0, 218),
            (0, 414),
            (0, 2),
            (0, 550),
            (0, 2),
            (0, 0),
            (0, 2593),
            (0, 4466),
            (0, 442),
            (0, 51),
            (0, 0),
            (0, 393),
            (0, 563),
            (0, 81),
            (0, 675),
            (0, 113),
            (0, 185),
            (0, 2),
            (0, 0),
            (0, 1798),
            (0, 0),
            (0, 1268),
            (0, 2843),
            (0, 64),
            (0, 1),
            (0, 155),
            (0, 7),
            (0, 2586),
            (0, 0),
            (0, 56),
            (0, 102),
            (0, 235),
            (0, 0),
            (0, 42),
            (0, 203),
            (0, 2),
            (0, 5),
            (0, 65),
            (0, 46),
            (0, 15),
            (0, 925),
            (0, 417),
            (0, 0),
            (0, 5),
            (0, 2668),
            (0, 184),
            (0, 0),
            (0, 1),
            (0, 0),
            (0, 1),
            (0, 4),
            (0, 50),
            (0, 1068),
            (0, 216),
            (0, 287),
            (0, 49),
            (0, 964),
            (0, 3276),
            (0, 2830),
            (0, 0),
            (0, 113),
            (0, 11),
            (0, 35),
            (0, 8),
            (0, 31),
            (0, 4),
            (0, 3328),
            (0, 0),
            (0, 705),
            (0, 52),
            (0, 896),
            (0, 69),
            (0, 13),
            (0, 32),
            (0, 178),
            (0, 7),
            (0, 278),
            (0, 531),
            (0, 1),
            (0, 1),
            (0, 150),
            (0, 375),
            (0, 73),
            (0, 3),
            (0, 1751),
            (0, 409),
            (0, 632),
            (0, 45),
            (0, 0),
            (0, 2),
            (0, 5),
            (0, 61),
            (0, 73),
            (0, 710),
            (0, 58),
            (0, 216),
            (0, 11),
            (0, 6),
            (0, 111),
            (0, 91),
            (0, 82),
            (0, 146),
            (0, 46),
            (0, 177),
            (0, 45),
            (0, 437),
            (0, 16),
            (0, 24),
            (0, 2739),
            (0, 11),
            (0, 0),
            (0, 5056),
            (0, 163),
            (0, 80),
            (0, 13),
            (0, 7),
            (0, 2),
            (0, 4),
            (0, 286),
            (0, 260),
            (0, 11),
            (0, 6),
            (0, 172),
            (0, 43),
            (0, 480),
            (0, 91),
            (0, 764),
            (0, 9),
            (0, 712),
            (0, 1662),
            (0, 4),
            (0, 11),
            (0, 15),
            (0, 71),
            (0, 119),
            (0, 6),
            (0, 50),
            (0, 622),
            (0, 13),
            (0, 43),
            (0, 3),
            (0, 489),
            (0, 552),
            (0, 1),
            (0, 290),
            (0, 5314),
            (0, 373),
            (0, 196),
            (0, 162),
            (0, 445),
            (0, 0),
            (0, 18),
            (0, 36),
            (0, 254),
            (0, 0),
            (0, 32),
            (0, 88),
            (0, 30),
            (0, 649),
            (0, 0),
            (0, 275),
            (0, 304),
            (0, 15),
            (0, 127),
            (0, 6269),
            (0, 31),
            (0, 23),
            (0, 122),
            (0, 1617),
            (0, 139),
            (0, 1309),
            (0, 3),
            (0, 7158),
            (0, 0),
            (0, 244),
            (0, 24),
            (0, 306),
            (0, 106),
            (0, 148),
            (0, 407),
            (0, 50),
            (0, 1744),
            (0, 3095),
            (0, 3368),
            (0, 2103),
            (0, 707),
            (0, 3),
            (0, 59),
            (0, 346),
            (0, 0),
            (0, 74),
            (0, 39),
            (0, 7),
            (0, 1448),
            (0, 230),
            (0, 99),
            (0, 0),
            (0, 0),
            (0, 1087),
            (0, 19),
            (0, 4),
            (0, 616),
            (0, 4),
            (0, 290),
            (0, 64),
            (0, 37),
            (0, 39),
            (0, 0),
            (0, 4049),
            (0, 674),
            (0, 65),
            (0, 663),
            (0, 120),
            (0, 416),
            (0, 109),
            (0, 0),
            (0, 388),
            (0, 472),
            (0, 25),
            (0, 279),
            (0, 570),
            (0, 0),
            (0, 0),
            (0, 261),
            (0, 2972),
            (0, 5),
            (0, 1),
            (0, 0),
            (0, 1),
            (0, 3),
            (0, 323),
            (0, 785),
            (0, 111),
            (0, 2522),
            (0, 355),
            (0, 1098),
            (0, 1),
            (0, 222),
            (0, 1),
            (0, 1),
            (0, 10),
            (0, 24),
            (0, 13),
            (0, 682),
            (0, 827),
            (0, 2),
            (0, 493),
            (0, 14),
            (0, 6),
            (0, 558),
            (0, 1),
            (0, 1148),
            (0, 80),
            (0, 190),
            (0, 72),
            (0, 204),
            (0, 3),
            (0, 588),
            (0, 275),
            (0, 578),
            (0, 5),
            (0, 15),
            (0, 1683),
            (0, 513),
            (0, 240),
            (0, 1794),
            (0, 1507),
            (0, 150),
            (0, 10),
            (0, 10),
            (0, 3),
            (0, 327),
            (0, 299),
            (0, 1),
            (0, 13),
            (0, 65),
            (0, 2267),
            (0, 110),
            (0, 348),
            (0, 894),
            (0, 63),
            (0, 994),
            (0, 72),
            (0, 362),
            (0, 192),
            (0, 2),
            (0, 3382),
            (0, 165),
            (0, 810),
            (0, 139),
            (0, 80),
            (0, 3700),
            (0, 6),
            (0, 726),
            (0, 30),
            (0, 53),
            (0, 308),
            (0, 31),
            (0, 747),
            (0, 32),
            (0, 3509),
            (0, 3),
            (0, 3),
            (0, 924),
            (0, 63),
            (0, 399),
            (0, 3),
            (0, 10),
            (0, 543),
            (0, 121),
            (0, 474),
            (0, 20),
            (0, 116),
            (0, 1),
            (0, 4),
            (0, 15),
            (0, 2116),
            (0, 564),
            (0, 460),
            (0, 17),
            (0, 1526),
            (0, 36),
            (0, 15),
            (0, 1526),
            (0, 76),
            (0, 13456),
            (0, 414),
            (0, 132),
            (0, 15),
            (0, 86),
            (0, 91),
            (0, 436),
            (0, 234),
            (0, 0),
            (0, 13),
            (0, 60),
            (0, 1451),
            (0, 181),
            (0, 33),
            (0, 37),
            (0, 26),
            (0, 59),
            (0, 983),
            (0, 1354),
            (0, 2),
            (0, 14),
            (0, 23),
            (0, 181),
            (0, 679),
            (0, 228),
            (0, 1),
            (0, 3),
            (0, 3880),
            (0, 8),
            (0, 292),
            (0, 7),
            (0, 0),
            (0, 26),
            (0, 0),
            (0, 5200),
            (0, 67),
            (0, 53),
            (0, 0),
            (0, 14),
            (0, 0),
            (0, 567),
            (0, 224),
            (0, 2),
            (0, 811),
            (0, 10),
            (0, 254),
            (0, 181),
            (0, 1),
            (0, 511),
            (0, 131),
            (0, 182),
            (0, 1121),
            (0, 2526),
            (0, 0),
            (0, 1588),
            (0, 16),
            (0, 581),
            (0, 22),
            (0, 393),
            (0, 1871),
            (0, 212),
            (0, 23),
            (0, 363),
            (0, 36),
            (0, 88),
            (0, 1155),
            (0, 43),
            (0, 213),
            (0, 941),
            (0, 1107),
            (0, 10),
            (0, 1404),
            (0, 37),
            (0, 5575),
            (0, 60),
            (0, 1591),
            (0, 10),
            (0, 192),
            (0, 574),
            (0, 48),
            (0, 67),
            (0, 5809),
            (0, 530),
            (0, 363),
            (0, 467),
            (0, 5),
            (0, 122),
            (0, 8),
            (0, 1248),
            (0, 17),
            (0, 3582),
            (0, 113),
            (0, 145),
            (0, 1326),
            (0, 1),
            (0, 4),
            (0, 323),
            (0, 311),
            (0, 45),
            (0, 108),
            (0, 1419),
            (0, 268),
            (0, 6),
            (0, 522),
            (0, 51),
            (0, 43),
            (0, 1),
            (0, 5),
            (0, 59),
            (0, 311),
            (0, 484),
            (0, 488),
            (0, 59),
            (0, 4395),
            (0, 1),
            (0, 14),
            (0, 2210),
            (0, 236),
            (0, 154),
            (0, 269),
            (0, 98),
            (0, 328),
            (0, 200),
            (0, 6726),
            (0, 66),
            (0, 167),
            (0, 334),
            (0, 41),
            (0, 678),
            (0, 2136),
            (0, 942),
            (0, 911),
            (0, 1008),
            (0, 612),
            (0, 12),
            (0, 984),
            (0, 511),
            (0, 22),
            (0, 574),
            (0, 194),
            (0, 1558),
            (0, 11),
            (0, 295),
            (0, 3077),
            (0, 0),
            (0, 10),
            (0, 4),
            (0, 19),
            (0, 673),
            (0, 404),
            (0, 3920),
            (0, 10),
            (0, 190),
            (0, 1126),
            (0, 4),
            (0, 204),
            (0, 479),
            (0, 698),
            (0, 22),
            (0, 0),
            (0, 41),
            (0, 1424),
            (0, 149),
            (0, 3),
            (0, 3),
            (0, 5),
            (0, 15),
            (0, 213),
            (0, 0),
            (0, 1721),
            (0, 1172),
            (0, 104),
            (0, 0),
            (0, 1),
            (0, 1322),
            (0, 12),
            (0, 192),
            (0, 360),
            (0, 327),
            (0, 290),
            (0, 1),
            (0, 674),
            (0, 117),
            (0, 0),
            (0, 4),
            (0, 1),
            (0, 52),
            (0, 409),
            (0, 22),
            (0, 990),
            (0, 452),
            (0, 1643),
            (0, 91),
            (0, 728),
            (0, 642),
            (0, 0),
            (0, 57),
            (0, 131),
            (0, 296),
            (0, 374),
            (0, 250),
            (0, 306),
            (0, 4),
            (0, 220),
            (0, 602),
            (0, 112),
            (0, 6692),
            (0, 477),
            (0, 199),
            (0, 31),
            (0, 2025),
            (0, 37),
            (0, 1184),
            (0, 394),
            (0, 0),
            (0, 86),
            (0, 83),
            (0, 650),
            (0, 507),
            (0, 1),
            (0, 87),
            (0, 322),
            (0, 4),
            (0, 151),
            (0, 5),
            (0, 47),
            (0, 4124),
            (0, 79),
            (0, 250),
            (0, 659),
            (0, 68),
            (0, 578),
            (0, 578),
            (0, 0),
            (0, 137),
            (0, 5),
            (0, 0),
            (0, 1143),
            (0, 28),
            (0, 417),
            (0, 1716),
            (0, 10238),
            (0, 51),
            (0, 40),
            (0, 6296),
            (0, 0),
            (0, 2),
            (0, 608),
            (0, 132),
            (0, 8548),
            (0, 29),
            (0, 1910),
            (0, 2659),
            (0, 4320),
            (0, 823),
            (0, 9),
            (0, 983),
            (0, 1483),
            (0, 54),
            (0, 11),
            (0, 87),
            (0, 636),
            (0, 23),
            (0, 0),
            (0, 161),
            (0, 313),
            (0, 1722),
            (0, 6244),
            (0, 0),
            (0, 26),
            (0, 180),
            (0, 290),
            (0, 3),
            (0, 3683),
            (0, 12),
            (0, 0),
            (0, 22),
            (0, 305),
            (0, 4),
            (0, 383),
            (0, 26),
            (0, 7672),
            (0, 0),
            (0, 46),
            (0, 567),
            (0, 8),
            (0, 1034),
            (0, 132),
            (0, 1100),
            (0, 13),
            (0, 0),
            (0, 655),
            (0, 2),
            (0, 10),
            (0, 2853),
            (0, 1301),
            (0, 218),
            (0, 782),
            (0, 411),
            (0, 142),
            (0, 1187),
            (0, 0),
            (0, 1),
            (0, 138),
            (0, 13),
            (0, 0),
            (0, 411),
            (0, 1),
            (0, 495),
            (0, 252),
            (0, 471),
            (0, 0),
            (0, 13),
            (0, 88),
            (0, 30),
            (0, 13),
            (0, 3),
            (0, 0),
            (0, 444),
            (0, 13),
            (0, 517),
            (0, 179),
            (0, 138),
            (0, 2519),
            (0, 617),
            (0, 24),
            (0, 254),
            (0, 258),
            (0, 1045),
            (0, 1465),
            (0, 65),
            (0, 0),
            (0, 487),
            (0, 12),
            (0, 795),
            (0, 419),
            (0, 1880),
            (0, 1726),
            (0, 1985),
            (0, 1),
            (0, 9),
            (0, 200),
            (0, 1),
            (0, 2769),
            (0, 135),
            (0, 3849),
            (0, 55),
            (0, 134),
            (0, 48),
            (0, 1187),
            (0, 0),
            (0, 18),
            (0, 19),
            (0, 184),
            (0, 1433),
            (0, 5),
            (0, 159),
            (0, 1126),
            (0, 848),
            (0, 374),
            (0, 201),
            (0, 757),
            (0, 95),
            (0, 0),
            (0, 147),
            (0, 1),
            (0, 1247),
            (0, 49),
            (0, 3009),
            (0, 47),
            (0, 2),
            (0, 1796),
            (0, 0),
            (0, 2),
            (0, 202),
            (0, 61),
            (0, 6),
            (0, 3),
            (0, 497),
            (0, 64),
            (0, 665),
            (0, 19),
            (0, 298),
            (0, 124),
            (0, 7),
            (0, 227),
            (0, 28),
            (0, 2319),
            (0, 193),
            (0, 0),
            (0, 410),
            (0, 21),
            (0, 1),
            (0, 1772),
            (0, 2),
            (0, 5),
            (0, 1622),
            (0, 1377),
            (0, 204),
            (0, 132),
            (0, 3),
            (0, 755),
            (0, 2511),
            (0, 12),
            (0, 17),
            (0, 2459),
            (0, 471),
            (0, 235),
            (0, 16),
            (0, 218),
            (0, 6),
            (0, 1139),
            (0, 83),
            (0, 3900),
            (0, 12),
            (0, 12),
            (0, 0),
            (0, 3),
            (0, 2442),
            (0, 117),
            (0, 427),
            (0, 5),
            (0, 24),
            (0, 4),
            (0, 182),
            (0, 0),
            (0, 17),
            (0, 2731),
            (0, 419),
            (0, 6),
            (0, 868),
            (0, 27),
            (0, 0),
            (0, 0),
            (0, 1097),
            (0, 41),
            (0, 0),
            (0, 1839),
            (0, 8),
            (0, 1752),
            (0, 51),
            (0, 59),
            (0, 217),
            (0, 580),
            (0, 135),
            (0, 9),
            (0, 2),
            (0, 150),
            (0, 1322),
            (0, 30),
            (0, 21),
            (0, 1),
            (0, 262),
            (0, 11),
            (0, 182),
            (0, 100),
            (0, 119),
            (0, 41),
            (0, 2),
            (0, 78),
            (0, 0),
            (0, 1),
            (0, 1060),
            (0, 7067),
            (0, 0),
            (0, 30),
            (0, 787),
            (0, 214),
            (0, 35),
            (0, 668),
            (0, 9670),
            (0, 1217),
            (0, 1058),
            (0, 0),
            (0, 213),
            (0, 517),
            (0, 471),
            (0, 346),
            (0, 42),
            (0, 599),
            (0, 30),
            (0, 0),
            (0, 14),
            (0, 47),
            (0, 3),
            (0, 27),
            (0, 362),
            (0, 215),
            (0, 996),
            (0, 2527),
            (0, 888),
            (0, 12),
            (0, 107),
            (0, 812),
            (0, 21),
            (0, 518),
            (0, 0),
            (0, 0),
            (0, 2),
            (0, 4146),
            (0, 3),
            (0, 121),
            (0, 3295),
            (0, 291),
            (0, 3781),
            (0, 11),
            (0, 0),
            (0, 336),
            (0, 270),
            (0, 69),
            (0, 0),
            (0, 69),
            (0, 27),
            (0, 329),
            (0, 23),
            (0, 2),
            (0, 32),
            (0, 2652),
            (0, 3832),
            (0, 32),
            (0, 12),
            (0, 0),
            (0, 6),
            (0, 43),
            (0, 18),
            (0, 108),
            (0, 449),
            (0, 55),
            (0, 6),
            (0, 146),
            (0, 339),
            (0, 18),
            (0, 29),
            (0, 33),
            (0, 448),
            (0, 3),
            (0, 1915),
            (1, 22),
            (0, 10208),
            (0, 1),
            (0, 396),
            (0, 308),
            (0, 4896),
            (0, 303),
            (0, 2),
            (0, 0),
            (0, 7),
            (0, 2477),
            (0, 12),
            (0, 10),
            (0, 230),
            (0, 143),
            (0, 5),
            (0, 1401),
            (0, 2),
            (0, 2),
            (0, 401),
            (0, 109),
            (0, 361),
            (0, 192),
            (0, 362),
            (0, 11),
            (0, 27),
            (0, 1431),
            (0, 1),
            (0, 9),
            (0, 25),
            (0, 1844),
            (0, 5),
            (0, 134),
            (0, 2),
            (0, 415),
            (0, 268),
            (0, 554),
            (0, 0),
            (0, 162),
            (0, 2776),
            (0, 6),
            (0, 38),
            (0, 0),
            (0, 2048),
            (0, 150),
            (0, 549),
            (0, 23),
            (0, 337),
            (0, 1456),
            (0, 20),
            (0, 2057),
            (0, 126),
            (0, 3760),
            (0, 3509),
            (0, 713),
            (0, 450),
            (0, 854),
            (0, 2175),
            (0, 2),
            (0, 5),
            (0, 3),
            (0, 1584),
            (0, 246),
            (0, 167),
            (0, 1151),
            (0, 3),
            (0, 577),
            (0, 9000),
            (0, 3108),
            (0, 1),
            (0, 41),
            (0, 189),
            (0, 1),
            (0, 1241),
            (0, 551),
            (0, 525),
            (0, 1328),
            (0, 672),
            (0, 9),
            (0, 87),
            (0, 0),
            (0, 33),
            (0, 5356),
            (0, 234),
            (0, 1),
            (0, 127),
            (0, 179),
            (0, 2),
            (0, 174),
            (0, 22),
            (0, 13),
            (0, 57),
            (0, 222),
            (0, 263),
            (0, 3488),
            (0, 7379),
            (0, 9),
            (0, 561),
            (0, 5),
            (0, 12),
            (0, 643),
            (0, 371),
            (0, 1),
            (0, 57),
            (0, 2108),
            (0, 413),
            (0, 33),
            (0, 11),
            (0, 1491),
            (0, 32),
            (0, 4),
            (0, 8),
            (0, 5),
            (0, 68),
            (0, 1143),
            (0, 1378),
            (0, 242),
            (0, 1570),
            (0, 156),
            (0, 6),
            (0, 21),
            (0, 434),
            (0, 0),
            (0, 91),
            (0, 417),
            (0, 2),
            (0, 30),
            (0, 618),
            (0, 7),
            (0, 2),
            (0, 69),
            (0, 167),
            (0, 0),
            (0, 24),
            (0, 924),
            (0, 909),
            (0, 1005),
            (0, 188),
            (0, 13),
            (0, 46),
            (0, 1157),
            (0, 4),
            (0, 141),
            (0, 45),
            (0, 27),
            (0, 0),
            (0, 3121),
            (0, 1329),
            (0, 1064),
            (0, 274),
            (0, 675),
            (0, 10),
            (0, 1953),
            (0, 3719),
            (0, 1177),
            (0, 0),
            (0, 616),
            (0, 4),
            (0, 204),
            (0, 1404),
            (0, 2241),
            (0, 1),
            (0, 597),
            (0, 10625),
            (0, 125),
            (0, 0),
            (0, 147),
            (0, 46),
            (0, 0),
            (0, 286),
            (0, 9),
            (0, 9),
            (0, 27),
            (0, 968),
            (0, 1050),
            (0, 247),
            (0, 64),
            (0, 233),
            (0, 89),
            (0, 2),
            (0, 142),
            (0, 2),
            (0, 41),
            (0, 1),
            (0, 4),
            (0, 407),
            (0, 17),
            (0, 3),
            (0, 6),
            (0, 732),
            (0, 1352),
            (0, 269),
            (0, 40),
            (0, 2231),
            (0, 58),
            (0, 47),
            (0, 2),
            (0, 919),
            (0, 9),
            (0, 2156),
            (0, 38),
            (0, 852),
            (0, 128),
            (0, 5),
            (0, 1952),
            (0, 190),
            (0, 298),
            (0, 389),
            (0, 1),
            (0, 14),
            (0, 1457),
            (0, 1082),
            (0, 1169),
            (0, 28),
            (0, 4818),
            (0, 0),
            (0, 1029),
            (0, 0),
            (0, 380),
            (0, 258),
            (0, 54),
            (0, 258),
            (0, 37),
            (0, 2),
            (0, 109),
            (0, 13),
            (0, 9),
            (0, 347),
            (0, 45),
            (0, 1199),
            (0, 13),
            (0, 3707),
            (0, 183),
            (0, 631),
            (0, 20),
            (0, 178),
            (0, 10),
            (0, 785),
            (0, 3),
            (0, 4358),
            (0, 63),
            (0, 294),
            (0, 2039),
            (0, 826),
            (0, 1832),
            (0, 5222),
            (0, 397),
            (0, 7),
            (0, 365),
            (0, 2),
            (0, 1033),
            (0, 43),
            (0, 113),
            (0, 1219),
            (0, 238),
            (0, 1934),
            (0, 3),
            (0, 8),
            (0, 7150),
            (0, 55),
            (0, 811),
            (0, 91),
            (0, 1223),
            (0, 0),
            (0, 163),
            (0, 170),
            (0, 474),
            (0, 72),
            (0, 20),
            (0, 8),
            (0, 829),
            (0, 7),
            (0, 245),
            (0, 70),
            (0, 370),
            (0, 147),
            (0, 3),
            (0, 91),
            (0, 137),
            (0, 3),
            (0, 24),
            (0, 13),
            (0, 4),
            (0, 6242),
            (0, 6),
            (0, 493),
            (0, 879),
            (0, 46),
            (0, 3),
            (0, 233),
            (0, 96),
            (0, 21),
            (0, 11),
            (0, 1650),
            (0, 6),
            (0, 24),
            (0, 1274),
            (0, 9),
            (0, 0),
            (0, 1231),
            (0, 1002),
            (0, 2),
            (0, 5092),
            (0, 444),
            (0, 620),
            (0, 7),
            (0, 4),
            (0, 142),
            (0, 62),
            (0, 9),
            (0, 66),
            (0, 771),
            (0, 0),
            (0, 2),
            (0, 1157),
            (0, 0),
            (0, 11),
            (0, 0),
            (0, 5),
            (0, 551),
            (0, 6),
            (0, 126),
            (0, 1),
            (0, 58),
            (0, 8),
            (0, 1332),
            (0, 0),
            (0, 261),
            (0, 77),
            (0, 1925),
            (0, 686),
            (0, 504),
            (0, 3),
            (0, 85),
            (0, 418),
            (0, 4),
            (0, 35),
            (0, 400),
            (0, 176),
            (0, 3),
            (0, 38),
            (0, 108),
            (0, 322),
            (0, 1761),
            (0, 10868),
            (0, 57),
            (0, 204),
            (0, 7),
            (0, 331),
            (0, 1296),
            (0, 1),
            (0, 2039),
            (0, 33),
            (0, 36),
            (0, 70),
            (0, 237),
            (0, 185),
            (0, 18),
            (0, 170),
            (0, 1335),
            (0, 0),
            (0, 3),
            (0, 4),
            (0, 1950),
            (0, 580),
            (0, 141),
            (0, 8010),
            (0, 5),
            (0, 1),
            (0, 1251),
            (0, 1573),
            (0, 162),
            (0, 2),
            (0, 390),
            (0, 131),
            (0, 36),
            (0, 5),
            (0, 14374),
            (0, 698),
            (0, 35),
            (0, 0),
            (0, 52),
            (0, 1827),
            (0, 0),
            (0, 326),
            (0, 17),
            (0, 1),
            (0, 11),
            (0, 896),
            (0, 0),
            (0, 117),
            (0, 2727),
            (0, 160),
            (0, 24),
            (0, 1263),
            (0, 1598),
            (0, 0),
            (0, 539),
            (0, 75),
            (0, 390),
            (0, 4777),
            (0, 12),
            (0, 38),
            (0, 35),
            (0, 52),
            (0, 2362),
            (0, 0),
            (0, 2119),
            (0, 37),
            (0, 26),
            (0, 6094),
            (0, 4202),
            (0, 1307),
            (0, 2513),
            (0, 865),
            (0, 441),
            (0, 1),
            (0, 181),
            (0, 968),
            (0, 1),
            (0, 0),
            (0, 5848),
            (0, 261),
            (0, 0),
            (0, 1731),
            (0, 234),
            (0, 3004),
            (0, 42),
            (0, 63),
            (0, 170),
            (0, 269),
            (0, 8),
            (0, 278),
            (0, 296),
            (0, 475),
            (0, 37),
            (0, 16),
            (0, 11),
            (0, 6),
            (0, 1),
            (0, 7),
            (0, 0),
            (0, 6),
            (0, 146),
            (0, 91),
            (0, 4078),
            (0, 69),
            (0, 14),
            (0, 7663),
            (0, 1),
            (0, 24),
            (0, 714),
            (0, 0),
            (0, 512),
            (0, 14),
            (0, 18),
            (0, 2099),
            (0, 195),
            (0, 430),
            (0, 102),
            (0, 1278),
            (0, 4054),
            (0, 1130),
            (0, 0),
            (0, 2340),
            (0, 1050),
            (0, 31),
            (0, 10),
            (0, 926),
            (0, 2403),
            (0, 129),
            (0, 3117),
            (0, 1),
            (0, 15955),
            (0, 3985),
            (0, 0),
            (0, 3658),
            (0, 1),
            (0, 31),
            (0, 1157),
            (0, 1),
            (0, 1393),
            (0, 446),
            (0, 2136),
            (0, 66),
            (0, 2593),
            (0, 4),
            (0, 893),
            (0, 109),
            (0, 1174),
            (0, 1787),
            (0, 8),
            (0, 8),
            (0, 193),
            (0, 2),
            (0, 2),
            (0, 0),
            (0, 35),
            (0, 5),
            (0, 0),
            (0, 885),
            (0, 670),
            (0, 128),
            (0, 1659),
            (0, 2444),
            (0, 54),
            (0, 15),
            (0, 4202),
            (0, 2),
            (0, 2129),
            (0, 91),
            (0, 6268),
            (0, 128),
            (0, 1),
            (0, 24),
            (0, 1640),
            (0, 6),
            (0, 4288),
            (0, 59),
            (0, 46),
            (0, 0),
            (0, 884),
            (0, 0),
            (0, 89),
            (0, 987),
            (0, 1197),
            (0, 48),
            (0, 18),
            (0, 1070),
            (0, 8540),
            (0, 2490),
            (0, 810),
            (0, 64),
            (0, 481),
            (0, 72),
            (0, 525),
            (0, 200),
            (0, 0),
            (0, 292),
            (0, 757),
            (0, 3886),
            (0, 297),
            (0, 1),
            (0, 2581),
            (0, 53),
            (0, 187),
            (0, 5361),
            (0, 13),
            (0, 1694),
            (0, 740),
            (0, 3),
            (0, 13),
            (0, 357),
            (0, 1864),
            (0, 135),
            (0, 1),
            (0, 661),
            (0, 43),
            (0, 2328),
            (0, 2188),
            (0, 3443),
            (0, 1050),
            (0, 2742),
            (0, 1641),
            (0, 15),
            (0, 2),
            (0, 434),
            (0, 1300),
            (0, 4215),
            (0, 78),
            (0, 706),
            (0, 2),
            (0, 11086),
            (0, 0),
            (0, 319),
            (0, 532),
            (0, 379),
            (0, 46),
            (0, 4),
            (0, 55),
            (0, 346),
            (0, 1603),
            (0, 1458),
            (0, 383),
            (0, 2811),
            (0, 1),
            (0, 1210),
            (0, 0),
            (0, 29),
            (0, 1321),
            (0, 187),
            (0, 316),
            (0, 1207),
            (0, 27),
            (0, 339),
            (0, 1),
            (0, 0),
            (0, 112),
            (0, 13),
            (0, 1000),
            (0, 152),
            (0, 113),
            (0, 71),
            (0, 3066),
            (0, 2248),
            (0, 53),
            (0, 1048),
            (0, 584),
            (0, 143),
            (0, 209),
            (0, 19),
            (0, 5553),
            (0, 1541),
            (0, 3),
            (0, 374),
            (0, 0),
            (0, 965),
            (0, 179),
            (0, 197),
            (0, 87),
            (0, 3),
            (0, 234),
            (0, 539),
            (0, 639),
            (0, 1432),
            (0, 36),
            (0, 2),
            (0, 8684),
            (0, 191),
            (0, 382),
            (0, 6),
            (0, 541),
            (0, 197),
            (0, 6444),
            (0, 29),
            (0, 5),
            (0, 73),
            (0, 1302),
            (0, 6),
            (0, 879),
            (0, 214),
            (0, 26),
            (0, 1),
            (0, 1530),
            (0, 31),
            (0, 7047),
            (0, 12291),
            (0, 940),
            (0, 6),
            (0, 3),
            (0, 34),
            (0, 914),
            (0, 15),
            (0, 37),
            (0, 29),
            (0, 333),
            (0, 2251),
            (0, 12),
            (0, 158),
            (0, 2067),
            (0, 0),
            (0, 1),
            (0, 1715),
            (0, 82),
            (0, 9),
            (0, 11),
            (0, 708),
            (0, 343),
            (0, 2311),
            (0, 4),
            (0, 16),
            (0, 126),
            (0, 37),
            (0, 33),
            (0, 1418),
            (0, 844),
            (0, 0),
            (0, 4),
            (0, 0),
            (0, 19),
            (0, 13),
            (0, 220),
            (0, 633),
            (0, 197),
            (0, 131),
            (0, 1814),
            (0, 1118),
            (0, 109),
            (0, 1),
            (0, 248),
            (0, 18),
            (0, 914),
            (0, 95),
            (0, 245),
            (0, 17),
            (0, 5),
            (0, 36),
            (0, 1057),
            (0, 2441),
            (0, 1085),
            (0, 0),
            (0, 845),
            (0, 31),
            (0, 2158),
            (0, 2),
            (0, 687),
            (0, 264),
            (0, 2561),
            (0, 100),
            (0, 188),
            (0, 711),
            (0, 40),
            (0, 163),
            (0, 543),
            (0, 4159),
            (0, 35),
            (0, 2),
            (0, 116),
            (0, 40),
            (0, 616),
            (0, 1444),
            (0, 10),
            (0, 0),
            (0, 890),
            (0, 1818),
            (0, 7),
            (0, 1973),
            (0, 0),
            (0, 2065),
            (0, 10),
            (0, 0),
            (0, 85),
            (0, 2),
            (0, 1144),
            (0, 1),
            (0, 8),
            (0, 145),
            (0, 18),
            (0, 191),
            (0, 323),
            (0, 484),
            (0, 76),
            (0, 3),
            (0, 117),
            (0, 16186),
            (0, 0),
            (0, 154),
            (0, 2087),
            (0, 287),
            (0, 3527),
            (0, 7443),
            (0, 2288),
            (0, 317),
            (0, 1533),
            (0, 35),
            (0, 2280),
            (0, 6),
            (0, 166),
            (0, 440),
            (0, 1806),
            (0, 1475),
            (0, 9316),
            (0, 9),
            (0, 35),
            (0, 86),
            (0, 4),
            (0, 473),
            (0, 1243),
            (0, 20),
            (0, 3),
            (0, 184),
            (0, 325),
            (0, 93),
            (0, 135),
            (0, 27),
            (0, 1),
            (0, 15),
            (0, 1091),
            (0, 3458),
            (0, 4),
            (0, 140),
            (0, 22),
            (0, 93),
            (0, 25),
            (0, 108),
            (0, 7),
            (0, 584),
            (0, 2),
            (0, 8),
            (0, 1),
            (0, 0),
            (0, 329),
            (0, 52),
            (0, 1915),
            (0, 982),
            (0, 365),
            (0, 251),
            (0, 0),
            (0, 180),
            (0, 346),
            (0, 363),
            (0, 702),
            (0, 288),
            (0, 14),
            (0, 270),
            (0, 6),
            (0, 12),
            (0, 1646),
            (0, 5),
            (0, 16),
            (0, 2496),
            (0, 384),
            (0, 20),
            (0, 93),
            (0, 103),
            (0, 5837),
            (0, 0),
            (0, 5),
            (0, 502),
            (0, 82),
            (0, 26),
            (0, 669),
            (0, 764),
            (0, 2151),
            (0, 106),
            (0, 22),
            (0, 1135),
            (0, 3),
            (0, 284),
            (0, 3173),
            (0, 170),
            (0, 1),
            (0, 12),
            (0, 21),
            (0, 168),
            (0, 789),
            (0, 741),
            (0, 629),
            (0, 4),
            (0, 30),
            (0, 5),
            (0, 847),
            (0, 134),
            (0, 35),
            (0, 244),
            (0, 442),
            (0, 0),
            (0, 2),
            (0, 306),
            (0, 12),
            (0, 43),
            (0, 525),
            (0, 216),
            (0, 33),
            (0, 5),
            (0, 11),
            (0, 197),
            (0, 46),
            (0, 25),
            (0, 7),
            (0, 93),
            (0, 5),
            (0, 18),
            (0, 4),
            (0, 15),
            (0, 196),
            (0, 1158),
            (0, 10),
            (0, 1),
            (0, 725),
            (0, 3),
            (0, 953),
            (0, 568),
            (0, 411),
            (0, 40),
            (0, 3779),
            (0, 2935),
            (0, 4452),
            (0, 151),
            (0, 642),
            (0, 1),
            (0, 622),
            (0, 173),
            (0, 5),
            (0, 2542),
            (0, 564),
            (0, 1682),
            (0, 5),
            (0, 6),
            (0, 470),
            (0, 1),
            (0, 349),
            (0, 13),
            (0, 1597),
            (0, 326),
            (0, 63),
            (0, 0),
            (0, 189),
            (0, 422),
            (0, 1294),
            (0, 0),
            (0, 14),
            (0, 7),
            (0, 5),
            (0, 9),
            (0, 633),
            (0, 0),
            (0, 4593),
            (0, 106),
            (0, 783),
            (0, 557),
            (0, 0),
            (0, 113),
            (0, 1086),
            (0, 25),
            (0, 814),
            (0, 1262),
            (0, 2404),
            (0, 6),
            (0, 5),
            (0, 11),
            (0, 11),
            (0, 31),
            (0, 960),
            (0, 484),
            (0, 18),
            (0, 2188),
            (0, 114),
            (0, 59),
            (0, 611),
            (0, 3810),
            (0, 56),
            (0, 4),
            (0, 215),
            (0, 100),
            (0, 2777),
            (0, 1952),
            (0, 3),
            (0, 9),
            (0, 81),
            (0, 993),
            (0, 4797),
            (0, 3),
            (0, 10923),
            (0, 321),
            (0, 36),
            (0, 632),
            (0, 41),
            (0, 40),
            (0, 1062),
            (0, 1575),
            (0, 23),
            (0, 134),
            (0, 62),
            (0, 1284),
            (0, 46),
            (0, 24),
            (0, 101),
            (0, 9),
            (0, 370),
            (0, 546),
            (0, 8674),
            (0, 0),
            (0, 143),
            (0, 2687),
            (0, 257),
            (0, 277),
            (0, 1093),
            (0, 8667),
            (0, 87),
            (0, 8347),
            (0, 370),
            (0, 3855),
            (0, 9),
            (0, 81),
            (0, 33),
            (0, 5633),
            (0, 490),
            (0, 7),
            (0, 66),
            (0, 43),
            (0, 0),
            (0, 346),
            (0, 3480),
            (0, 1),
            (0, 17),
            (0, 33),
            (0, 1474),
            (0, 239),
            (0, 769),
            (0, 22),
            (0, 4250),
            (0, 706),
            (0, 446),
            (0, 10),
            (0, 1),
            (0, 156),
            (0, 52),
            (0, 357),
            (0, 734),
            (0, 410),
            (0, 68),
            (0, 1),
            (0, 318),
            (0, 19896),
            (0, 9),
            (0, 44),
            (0, 1879),
            (0, 372),
            (0, 13),
            (0, 1417),
            (0, 0),
            (0, 28),
            (0, 188),
            (0, 88),
            (0, 781),
            (0, 259),
            (0, 12),
            (0, 13),
            (0, 2139),
            (0, 758),
            (0, 11),
            (0, 42),
            (0, 12),
            (0, 1),
            (0, 4),
            (0, 1),
            (0, 1353),
            (0, 125),
            (0, 124),
            (0, 85),
            (0, 5013),
            (0, 498),
            (0, 3),
            (0, 549),
            (0, 101),
            (0, 2050),
            (0, 4),
            (0, 1611),
            (0, 0),
            (0, 233),
            (0, 19),
            (0, 9),
            (0, 811),
            (0, 201),
            (0, 17),
            (0, 288),
            (0, 2654),
            (0, 207),
            (0, 982),
            (0, 83),
            (0, 17),
            (0, 150),
            (0, 5350),
            (0, 884),
            (0, 111),
            (0, 0),
            (0, 269),
            (0, 43),
            (0, 31),
            (0, 0),
            (0, 211),
            (0, 0),
            (0, 635),
            (0, 89),
            (0, 167),
            (0, 1184),
            (0, 108),
            (0, 38),
            (0, 1446),
            (0, 71),
            (0, 1),
            (0, 0),
            (0, 260),
            (0, 0),
            (0, 58),
            (0, 25),
            (0, 0),
            (0, 697),
            (0, 3267),
            (0, 90),
            (0, 11),
            (0, 2500),
            (0, 173),
            (0, 1),
            (0, 4284),
            (0, 4),
            (0, 2),
            (0, 23),
            (0, 20),
            (0, 55),
            (0, 2418),
            (0, 28),
            (0, 458),
            (0, 2485),
            (0, 2),
            (0, 6),
            (0, 728),
            (0, 415),
            (0, 1138),
            (0, 6),
            (0, 441),
            (0, 6209),
            (0, 2396),
            (0, 5143),
            (0, 167),
            (0, 318),
            (0, 11),
            (0, 205),
            (0, 6),
            (0, 73),
            (0, 63),
            (0, 0),
            (0, 3871),
            (0, 347),
            (0, 380),
            (0, 0),
            (0, 725),
            (0, 90),
            (0, 3),
            (0, 0),
            (0, 3),
            (0, 4),
            (0, 172),
            (0, 51),
            (0, 1793),
            (0, 41),
            (0, 22),
            (0, 2),
            (0, 18),
            (0, 220),
            (0, 6),
            (0, 4),
            (0, 4911),
            (0, 25),
            (0, 13),
            (0, 2290),
            (0, 705),
            (0, 376),
            (0, 0),
            (0, 29),
            (0, 653),
            (0, 82),
            (0, 51),
            (0, 1648),
            (0, 2324),
            (0, 2290),
            (0, 64),
            (0, 2189),
            (0, 32),
            (0, 25),
            (0, 62),
            (0, 61),
            (0, 4932),
            (0, 70),
            (0, 20),
            (0, 2822),
            (0, 1),
            (0, 1424),
            (0, 16),
            (0, 896),
            (0, 1107),
            (0, 1393),
            (0, 2027),
            (0, 727),
            (0, 123),
            (0, 12),
            (0, 4212),
            (0, 8),
            (0, 2563),
            (0, 208),
            (0, 2102),
            (0, 3103),
            (0, 17),
            (0, 2),
            (0, 2),
            (0, 160),
            (0, 30),
            (0, 137),
            (0, 1192),
            (0, 3),
            (0, 393),
            (0, 0),
            (0, 543),
            (0, 360),
            (0, 3791),
            (0, 2),
            (0, 2247),
            (0, 306),
            (0, 2349),
            (0, 58),
            (0, 1758),
            (0, 891),
            (0, 1957),
            (0, 287),
            (0, 85),
            (0, 0),
            (0, 1158),
            (0, 14),
            (0, 1),
            (0, 12),
            (0, 697),
            (0, 231),
            (0, 9),
            (0, 19),
            (0, 89),
            (0, 0),
            (0, 7),
            (0, 1408),
            (0, 1),
            (0, 3418),
            (0, 6),
            (0, 33),
            (0, 1921),
            (0, 2416),
            (0, 92),
            (0, 7),
            (0, 151),
            (0, 5141),
            (0, 7),
            (0, 218),
            (0, 161),
            (0, 1552),
            (0, 309),
            (0, 136),
            (0, 282),
            (0, 1320),
            (0, 40),
            (0, 1065),
            (0, 0),
            (0, 77),
            (0, 123),
            (0, 0),
            (0, 2930),
            (0, 212),
            (0, 77),
            (0, 134),
            (0, 1567),
            (0, 2133),
            (0, 126),
            (0, 17),
            (0, 219),
            (0, 51),
            (0, 0),
            (0, 989),
            (0, 2473),
            (0, 59),
            (0, 13),
            (0, 650),
            (0, 0),
            (0, 1),
            (0, 23),
            (0, 0),
            (0, 4),
            (0, 52),
            (0, 119),
            (0, 0),
            (0, 176),
            (0, 14),
            (0, 3),
            (0, 6),
            (0, 2618),
            (0, 671),
            (0, 77),
            (0, 39),
            (0, 564),
            (0, 280),
            (0, 16),
            (0, 3),
            (0, 2),
            (0, 1352),
            (0, 1492),
            (0, 91),
            (0, 780),
            (0, 2),
            (0, 2706),
            (0, 17),
            (0, 12),
            (0, 1),
            (0, 402),
            (0, 80),
            (0, 0),
            (0, 765),
            (0, 316),
            (0, 13),
            (0, 13),
            (0, 10),
            (0, 154),
            (0, 739),
            (0, 7091),
            (0, 653),
            (0, 267),
            (0, 65),
            (0, 42),
            (0, 4549),
            (0, 76),
            (0, 335),
            (0, 317),
            (0, 33),
            (0, 2115),
            (0, 10),
            (0, 4168),
            (0, 1585),
            (0, 3),
            (0, 77),
            (0, 2540),
            (0, 345),
            (0, 37),
            (0, 24),
            (0, 379),
            (0, 0),
            (0, 2),
            (0, 963),
            (0, 455),
            (0, 85),
            (0, 6024),
            (0, 1274),
            (0, 1477),
            (0, 9),
            (0, 799),
            (0, 830),
            (0, 17),
            (0, 3558),
            (0, 721),
            (0, 483),
            (0, 195),
            (0, 33),
            (0, 610),
            (0, 9006),
            (0, 4),
            (0, 359),
            (0, 79),
            (0, 1723),
            (0, 15),
            (0, 522),
            (0, 6),
            (0, 130),
            (0, 2290),
            (0, 145),
            (0, 60),
            (0, 411),
            (0, 1),
            (0, 0),
            (0, 343),
            (0, 752),
            (0, 575),
            (0, 69),
            (0, 950),
            (0, 25),
            (0, 18),
            (0, 10),
            (0, 2595),
            (0, 368),
            (0, 2355),
            (0, 67),
            (0, 817),
            (0, 2515),
            (0, 77),
            (0, 77),
            (0, 0),
            (0, 172),
            (0, 170),
            (0, 17),
            (0, 1633),
            (0, 78),
            (0, 60),
            (0, 1656),
            (0, 27),
            (0, 5859),
            (0, 44),
            (0, 38),
            (0, 89),
            (0, 1375),
            (0, 1323),
            (0, 692),
            (0, 99),
            (0, 175),
            (0, 534),
            (0, 3421),
            (0, 597),
            (0, 83),
            (0, 0),
            (0, 363),
            (0, 158),
            (0, 2755),
            (0, 213),
            (0, 397),
            (0, 498),
            (0, 389),
            (0, 1629),
            (0, 103),
            (0, 3),
            (0, 388),
            (0, 67),
            (0, 0),
            (0, 0),
            (0, 1385),
            (0, 203),
            (0, 0),
            (0, 953),
            (0, 1308),
            (0, 6087),
            (0, 2),
            (0, 480),
            (0, 0),
            (0, 178),
            (0, 45),
            (0, 32),
            (0, 1711),
            (0, 7333),
            (0, 0),
            (0, 460),
            (0, 2),
            (0, 1016),
            (0, 177),
            (0, 2),
            (0, 202),
            (0, 5),
            (0, 2269),
            (0, 42),
            (0, 509),
            (0, 33),
            (0, 5),
            (0, 394),
            (0, 108),
            (0, 132),
            (0, 3560),
            (0, 169),
            (0, 27),
            (0, 922),
            (0, 334),
            (0, 2596),
            (0, 1346),
            (0, 125),
            (0, 8),
            (0, 2517),
            (0, 3154),
            (0, 2186),
            (0, 9),
            (0, 77),
            (0, 0),
            (0, 327),
            (0, 177),
            (0, 13480),
            (0, 156),
            (0, 65),
            (0, 1019),
            (0, 231),
            (0, 756),
            (0, 31),
            (0, 12288),
            (0, 59),
            (0, 214),
            (0, 80),
            (0, 424),
            (0, 21),
            (0, 378),
            (0, 2566),
            (0, 3),
            (0, 205),
            (0, 104),
            (0, 2275),
            (0, 1),
            (0, 20),
            (0, 1060),
            (0, 26),
            (0, 39),
            (0, 1616),
            (0, 25),
            (0, 3),
            (0, 7),
            (0, 1),
            (0, 3194),
            (0, 1),
            (0, 198),
            (0, 1),
            (0, 44),
            (0, 270),
            (0, 278),
            (0, 25),
            (0, 7),
            (0, 10),
            (0, 7520),
            (0, 1),
            (0, 21),
            (0, 0),
            (0, 5893),
            (0, 195),
            (0, 121),
            (0, 2356),
            (0, 44),
            (0, 27),
            (0, 143),
            (0, 0),
            (0, 1178),
            (0, 3307),
            (0, 791),
            (0, 2),
            (0, 5),
            (0, 215),
            (0, 9),
            (0, 104),
            (0, 20),
            (0, 1),
            (0, 2514),
            (0, 74),
            (0, 2354),
            (0, 282),
            (0, 124),
            (0, 81),
            (0, 58),
            (0, 36),
            (0, 106),
            (0, 317),
            (0, 710),
            (0, 6),
            (0, 96),
            (0, 0),
            (0, 1135),
            (0, 4),
            (0, 347),
            (0, 1116),
            (0, 1586),
            (0, 5262),
            (0, 358),
            (0, 0),
            (0, 5),
            (0, 478),
            (0, 400),
            (0, 407),
            (0, 4402),
            (0, 6841),
            (0, 1),
            (0, 2),
            (0, 115),
            (0, 40),
            (0, 949),
            (0, 32),
            (0, 1),
            (0, 104),
            (0, 5610),
            (0, 5),
            (0, 827),
            (0, 9353),
            (0, 1803),
            (0, 4),
            (0, 1440),
            (0, 1711),
            (0, 1911),
            (0, 1336),
            (0, 900),
            (0, 137),
            (0, 32),
            (0, 717),
            (0, 3),
            (0, 1),
            (0, 127),
            (0, 148),
            (0, 155),
            (0, 22),
            (0, 408),
            (0, 19),
            (0, 1),
            (0, 155),
            (0, 805),
            (0, 349),
            (0, 288),
            (0, 1046),
            (0, 2),
            (0, 5),
            (0, 7),
            (0, 50),
            (0, 173),
            (0, 178),
            (0, 166),
            (0, 5),
            (0, 134),
            (0, 103),
            (0, 2999),
            (0, 1277),
            (0, 2),
            (0, 35),
            (0, 1266),
            (0, 1606),
            (0, 41),
            (0, 38),
            (0, 115),
            (0, 1),
            (0, 498),
            (0, 458),
            (0, 142),
            (0, 33),
            (0, 6),
            (0, 1),
            (0, 1390),
            (0, 13),
            (0, 19993),
            (0, 5),
            (0, 1487),
            (0, 8),
            (0, 4071),
            (0, 4),
            (0, 912),
            (0, 4),
            (0, 1450),
            (0, 441),
            (0, 34),
            (0, 4),
            (0, 0),
            (0, 1554),
            (0, 164),
            (0, 0),
            (0, 392),
            (0, 258),
            (0, 5),
            (0, 122),
            (0, 1148),
            (0, 4),
            (0, 3156),
            (0, 8664),
            (0, 19),
            (0, 1950),
            (0, 31),
            (0, 879),
            (0, 90),
            (0, 144),
            (0, 2002),
            (0, 4604),
            (0, 447),
            (0, 55),
            (0, 13),
            (0, 12),
            (0, 3),
            (0, 1260),
            (0, 3),
            (0, 1122),
            (0, 1509),
            (0, 6),
            (0, 53),
            (0, 116),
            (0, 67),
            (0, 32),
            (0, 1326),
            (0, 210),
            (0, 32),
            (0, 123),
            (0, 138),
            (0, 154),
            (0, 49),
            (0, 1),
            (0, 187),
            (0, 5354),
            (0, 2212),
            (0, 61),
            (0, 61),
            (0, 74),
            (0, 3205),
            (0, 76),
            (0, 62),
            (0, 2897),
            (0, 1468),
            (0, 4777),
            (0, 21),
            (0, 9),
            (0, 138),
            (0, 2930),
            (0, 1189),
            (0, 579),
            (0, 7),
            (0, 2767),
            (0, 2107),
            (0, 1),
            (0, 0),
            (0, 238),
            (0, 920),
            (0, 6893),
            (0, 49),
            (0, 2174),
            (0, 31),
            (0, 0),
            (0, 630),
            (0, 1043),
            (0, 506),
            (0, 623),
            (0, 0),
            (0, 147),
            (0, 2),
            (0, 2406),
            (0, 15),
            (0, 230),
            (0, 398),
            (0, 17),
            (0, 163),
            (0, 697),
            (0, 6),
            (0, 647),
            (0, 3220),
            (0, 193),
            (0, 231),
            (0, 11),
            (0, 565),
            (0, 177),
            (0, 144),
            (0, 128),
            (0, 4),
            (0, 402),
            (0, 952),
            (0, 84),
            (0, 93),
            (0, 5),
            (0, 24),
            (0, 694),
            (0, 44),
            (0, 149),
            (0, 2),
            (0, 67),
            (0, 281),
            (0, 1017),
            (0, 0),
            (0, 289),
            (0, 15),
            (0, 18),
            (0, 694),
            (0, 698),
            (0, 102),
            (0, 19),
            (0, 1561),
            (0, 719),
            (0, 537),
            (0, 1045),
            (0, 326),
            (0, 3),
            (0, 339),
            (0, 57),
            (0, 2193),
            (0, 54),
            (0, 2842),
            (0, 882),
            (0, 1874),
            (0, 690),
            (0, 3294),
            (0, 269),
            (0, 607),
            (0, 630),
            (0, 4),
            (0, 3),
            (0, 29),
            (0, 4323),
            (0, 2341),
            (0, 2325),
            (0, 639),
            (0, 1),
            (0, 16),
            (0, 352),
            (0, 644),
            (0, 2),
            (0, 183),
            (0, 82),
            (0, 149),
            (0, 31),
            (0, 5),
            (0, 2729),
            (0, 102),
            (0, 1690),
            (0, 157),
            (0, 1034),
            (0, 848),
            (0, 1753),
            (0, 261),
            (0, 52),
            (0, 62),
            (0, 22),
            (0, 3741),
            (0, 21),
            (0, 9192),
            (0, 73),
            (0, 82),
            (0, 195),
            (0, 489),
            (0, 12333),
            (0, 4428),
            (0, 13),
            (0, 491),
            (0, 11),
            (0, 2780),
            (0, 33),
            (0, 31),
            (0, 703),
            (0, 0),
            (0, 1),
            (0, 58),
            (0, 1),
            (0, 23104),
            (0, 1450),
            (0, 98),
            (0, 15),
            (0, 151),
            (0, 41),
            (0, 2),
            (0, 206),
            (0, 2354),
            (0, 8),
            (0, 3906),
            (0, 562),
            (0, 16),
            (0, 34),
            (0, 844),
            (0, 157),
            (0, 4728),
            (0, 1836),
            (0, 0),
            (0, 38),
            (0, 1045),
            (0, 13),
            (0, 129),
            (0, 0),
            (0, 554),
            (0, 51),
            (0, 7763),
            (0, 13),
            (0, 27),
            (0, 1211),
            (0, 212),
            (0, 24),
            (0, 9),
            (0, 1106),
            (0, 3),
            (0, 34),
            (0, 1759),
            (0, 7),
            (0, 6),
            (0, 979),
            (0, 161),
            (0, 11),
            (0, 2249),
            (0, 261),
            (0, 4050),
            (0, 266),
            (0, 41),
            (0, 140),
            (0, 27),
            (0, 5),
            (0, 0),
            (0, 322),
            (0, 47),
            (0, 117),
            (0, 573),
            (0, 5962),
            (0, 30),
            (0, 4),
            (0, 194),
            (0, 9),
            (0, 10),
            (0, 43),
            (0, 3),
            (0, 80),
            (0, 0),
            (0, 1094),
            (0, 297),
            (0, 62),
            (0, 97),
            (0, 3170),
            (0, 45),
            (0, 413),
            (0, 2133),
            (0, 0),
            (0, 57),
            (0, 24),
            (0, 14634),
            (0, 7),
            (0, 423),
            (0, 21),
            (0, 3393),
            (0, 492),
            (0, 9),
            (0, 8564),
            (0, 979),
            (0, 1903),
            (0, 3),
            (0, 505),
            (0, 1390),
            (0, 1285),
            (0, 10399),
            (0, 168),
            (0, 730),
            (0, 3),
            (0, 436),
            (0, 0),
            (0, 9),
            (0, 7),
            (0, 1640),
            (0, 767),
            (0, 624),
            (0, 408),
            (0, 305),
            (0, 2348),
            (0, 140),
            (0, 2262),
            (0, 39),
            (0, 2107),
            (0, 10),
            (0, 119),
            (0, 275),
            (0, 1),
            (0, 8),
            (0, 994),
            (0, 517),
            (0, 10),
            (0, 6),
            (0, 371),
            (0, 1549),
            (0, 1280),
            (0, 284),
            (0, 467),
            (0, 11),
            (0, 3510),
            (0, 4523),
            (0, 149),
            (0, 977),
            (0, 23),
            (0, 1911),
            (0, 4),
            (0, 238),
            (0, 1714),
            (0, 103),
            (0, 2445),
            (0, 823),
            (0, 3846),
            (0, 9),
            (0, 44),
            (0, 1086),
            (0, 48),
            (0, 216),
            (0, 4),
            (0, 0),
            (0, 5232),
            (0, 15),
            (0, 0),
            (0, 0),
            (0, 404),
            (0, 1128),
            (0, 7),
            (0, 12),
            (0, 1258),
            (0, 2190),
            (0, 1284),
            (0, 2571),
            (0, 627),
            (0, 44),
            (0, 360),
            (0, 13),
            (0, 791),
            (0, 10),
            (0, 12),
            (0, 26),
            (0, 12),
            (0, 1643),
            (0, 281),
            (0, 42),
            (0, 42),
            (0, 199),
            (0, 268),
            (0, 17),
            (0, 8),
            (0, 527),
            (0, 39),
            (0, 1083),
            (0, 42),
            (0, 73),
            (0, 3805),
            (0, 4),
            (0, 1113),
            (0, 1213),
            (0, 34),
            (0, 18),
            (0, 0),
            (0, 100),
            (0, 8),
            (0, 455),
            (0, 212),
            (0, 42),
            (0, 43),
            (0, 651),
            (0, 12),
            (0, 134),
            (0, 921),
            (0, 2859),
            (0, 510),
            (0, 16),
            (0, 518),
            (0, 90),
            (0, 6),
            (0, 1947),
            (0, 9332),
            (0, 23871),
            (0, 17),
            (0, 0),
            (0, 152),
            (0, 405),
            (0, 1461),
            (0, 0),
            (0, 38),
            (0, 104),
            (0, 2),
            (0, 3302),
            (0, 1),
            (0, 1897),
            (0, 2135),
            (0, 1476),
            (0, 270),
            (0, 844),
            (0, 10720),
            (0, 51),
            (0, 1719),
            (0, 81),
            (0, 4),
            (0, 1935),
            (0, 183),
            (0, 300),
            (0, 39),
            (0, 3725),
            (0, 3348),
            (0, 30),
            (0, 29),
            (0, 5),
            (0, 8),
            (0, 1777),
            (0, 1087),
            (0, 11),
            (0, 19),
            (0, 17),
            (0, 2189),
            (0, 393),
            (0, 1947),
            (0, 353),
            (0, 332),
            (0, 9),
            (0, 7),
            (0, 5142),
            (0, 0),
            (0, 13),
            (0, 690),
            (0, 4780),
            (0, 191),
            (0, 150),
            (0, 2113),
            (0, 1433),
            (0, 3234),
            (0, 397),
            (0, 73),
            (0, 577),
            (0, 47),
            (0, 99),
            (0, 73),
            (0, 8845),
            (0, 195),
            (0, 1349),
            (0, 3),
            (0, 11),
            (0, 44),
            (0, 2655),
            (0, 13),
            (0, 9),
            (0, 212),
            (0, 54),
            (0, 129),
            (0, 304),
            (0, 54),
            (0, 2),
            (0, 2),
            (0, 0),
            (0, 11),
            (0, 1090),
            (0, 5),
            (0, 670),
            (0, 112),
            (0, 86),
            (0, 571),
            (0, 6296),
            (0, 0),
            (0, 59),
            (0, 1),
            (0, 3087),
            (0, 505),
            (0, 91),
            (0, 26),
            (0, 349),
            (0, 16),
            (0, 830),
            (0, 24),
            (0, 14969),
            (0, 525),
            (0, 1511),
            (0, 2700),
            (0, 27),
            (0, 1473),
            (0, 14),
            (0, 2804),
            (0, 115),
            (0, 369),
            (0, 69),
            (0, 25),
            (0, 777),
            (0, 11),
            (0, 31),
            (0, 177),
            (0, 370),
            (0, 6335),
            (0, 434),
            (0, 5707),
            (0, 3734),
            (0, 1107),
            (0, 5),
            (0, 8468),
            (0, 11),
            (0, 2944),
            (0, 105),
            (0, 306),
            (0, 60),
            (0, 620),
            (0, 1297),
            (0, 2855),
            (0, 44),
            (0, 1697),
            (0, 10410),
            (0, 1903),
            (0, 6),
            (0, 1875),
            (0, 756),
            (0, 3330),
            (0, 9),
            (0, 309),
            (0, 10409),
            (0, 0),
            (0, 148),
            (0, 642),
            (0, 41),
            (0, 3568),
            (0, 8690),
            (0, 4),
            (0, 1683),
            (0, 8),
            (0, 1405),
            (0, 211),
            (0, 586),
            (0, 38),
            (0, 126),
            (0, 921),
            (0, 9),
            (0, 43),
            (0, 71),
            (0, 78),
            (0, 127),
            (0, 84),
            (0, 325),
            (0, 21),
            (0, 11),
            (0, 26),
            (0, 8763),
            (0, 653),
            (0, 333),
            (0, 46),
            (0, 20),
            (0, 374),
            (0, 18),
            (0, 560),
            (0, 24),
            (0, 0),
            (0, 1343),
            (0, 54),
            (0, 1430),
            (0, 0),
            (0, 132),
            (0, 50),
            (0, 170),
            (0, 99),
            (0, 2810),
            (0, 585),
            (0, 391),
            (0, 29),
            (0, 3),
            (0, 9),
            (0, 5),
            (0, 0),
            (0, 277),
            (0, 5701),
            (0, 1319),
            (0, 439),
            (0, 359),
            (0, 39),
            (0, 12),
            (0, 87),
            (0, 1484),
            (0, 1),
            (0, 36),
            (0, 17),
            (0, 15),
            (0, 5348),
            (0, 797),
            (0, 1),
            (0, 1285),
            (0, 900),
            (0, 27),
            (0, 999),
            (0, 4007),
            (0, 172),
            (0, 7),
            (0, 976),
            (0, 16),
            (0, 35),
            (0, 207),
            (0, 192),
            (0, 0),
            (0, 195),
            (0, 1),
            (0, 248),
            (0, 11),
            (0, 8390),
            (0, 672),
            (0, 538),
            (0, 1),
            (0, 21),
            (0, 1753),
            (0, 2419),
            (0, 53),
            (0, 0),
            (0, 8),
            (0, 4),
            (0, 14),
            (0, 33),
            (0, 9),
            (0, 75),
            (0, 49),
            (0, 1),
            (0, 4),
            (0, 108),
            (0, 1338),
            (0, 2),
            (0, 122),
            (0, 5),
            (0, 264),
            (0, 80),
            (0, 2389),
            (0, 35),
            (0, 3),
            (0, 324),
            (0, 2),
            (0, 59),
            (0, 71),
            (0, 1),
            (0, 4),
            (0, 4),
            (0, 4993),
            (0, 988),
            (0, 1706),
            (0, 40),
            (0, 5787),
            (0, 100),
            (0, 50),
            (0, 7983),
            (0, 2963),
            (0, 343),
            (0, 4413),
            (0, 230),
            (0, 3196),
            (0, 249),
            (0, 3),
            (0, 639),
            (0, 1816),
            (0, 564),
            (0, 4884),
            (0, 59),
            (0, 29),
            (0, 1112),
            (0, 820),
            (0, 1908),
            (0, 15),
            (0, 2),
            (0, 0),
            (0, 200),
            (0, 259),
            (0, 22),
            (0, 2),
            (0, 694),
            (0, 120),
            (0, 1869),
            (0, 25),
            (0, 133),
            (0, 41),
            (0, 3),
            (0, 5),
            (0, 5),
            (0, 1),
            (0, 8049),
            (0, 739),
            (0, 394),
            (0, 2622),
            (0, 1153),
            (0, 3785),
            (0, 86),
            (0, 5),
            (0, 1),
            (0, 14),
            (0, 10),
            (0, 451),
            (0, 59),
            (0, 0),
            (0, 89),
            (0, 34),
            (0, 47),
            (0, 33),
            (0, 108),
            (0, 42),
            (0, 381),
            (0, 1612),
            (0, 20),
            (0, 1279),
            (0, 2451),
            (0, 27),
            (0, 2),
            (0, 1),
            (0, 28),
            (0, 19),
            (0, 0),
            (0, 763),
            (0, 476),
            (0, 1),
            (0, 2246),
            (0, 129),
            (0, 10070),
            (0, 564),
            (0, 1950),
            (0, 341),
            (0, 5),
            (0, 0),
            (0, 291),
            (0, 1436),
            (0, 5887),
            (0, 0),
            (0, 1182),
            (0, 35),
            (0, 1496),
            (0, 117),
            (0, 0),
            (0, 151),
            (0, 7),
            (0, 5328),
            (0, 58),
            (0, 1),
            (0, 266),
            (0, 405),
            (0, 1020),
            (0, 547),
            (0, 2),
            (0, 793),
            (0, 192),
            (0, 19),
            (0, 1335),
            (0, 1311),
            (0, 3),
            (0, 168),
            (0, 2),
            (0, 1),
            (0, 312),
            (0, 2417),
            (0, 863),
            (0, 1053),
            (0, 2405),
            (0, 558),
            (0, 21),
            (0, 176),
            (0, 258),
            (0, 157),
            (0, 894),
            (0, 2186),
            (0, 41),
            (0, 2917),
            (0, 616),
            (0, 9),
            (0, 98),
            (0, 55),
            (0, 0),
            (0, 130),
            (0, 528),
            (0, 4),
            (0, 403),
            (0, 366),
            (0, 1375),
            (0, 7),
            (0, 1503),
            (0, 1380),
            (0, 53),
            (0, 837),
            (0, 4),
            (0, 26),
            (0, 67),
            (0, 118),
            (0, 1),
            (0, 215),
            (0, 563),
            (0, 179),
            (0, 39),
            (0, 9),
            (0, 1406),
            (0, 110),
            (0, 0),
            (0, 220),
            (0, 180),
            (0, 2752),
            (0, 5207),
            (0, 2855),
            (0, 0),
            (0, 155),
            (0, 570),
            (0, 7),
            (0, 876),
            (0, 9),
            (0, 385),
            (0, 1492),
            (0, 372),
            (0, 697),
            (0, 2124),
            (0, 0),
            (0, 46),
            (0, 992),
            (0, 12),
            (0, 18),
            (0, 16),
            (0, 204),
            (0, 17),
            (0, 312),
            (0, 774),
            (0, 2154),
            (0, 441),
            (0, 2231),
            (0, 2959),
            (0, 1),
            (0, 0),
            (0, 0),
            (0, 596),
            (0, 6),
            (0, 1),
            (0, 13334),
            (0, 355),
            (0, 38),
            (0, 24),
            (0, 138),
            (0, 354),
            (0, 16),
            (0, 1580),
            (0, 12),
            (0, 837),
            (0, 189),
            (0, 7),
            (0, 72),
            (0, 1479),
            (0, 12169),
            (0, 5),
            (0, 602),
            (0, 111),
            (0, 6478),
            (0, 152),
            (0, 155),
            (0, 1813),
            (0, 7),
            (0, 18),
            (0, 69),
            (0, 1),
            (0, 20),
            (0, 0),
            (0, 1124),
            (0, 31),
            (0, 17),
            (0, 432),
            (0, 61),
            (0, 1),
            (0, 10),
            (0, 67),
            (0, 138),
            (0, 4943),
            (0, 110),
            (0, 69),
            (0, 17),
            (0, 448),
            (0, 607),
            (0, 37),
            (0, 0),
            (0, 396),
            (0, 0),
            (0, 526),
            (0, 557),
            (0, 3822),
            (0, 0),
            (0, 3),
            (0, 113),
            (0, 3761),
            (0, 5120),
            (0, 136),
            (0, 400),
            (0, 15381),
            (0, 2092),
            (0, 139),
            (0, 556),
            (0, 127),
            (0, 743),
            (0, 5682),
            (0, 120),
            (0, 14),
            (0, 612),
            (0, 404),
            (0, 16),
            (0, 1732),
            (0, 40),
            (0, 22),
            (0, 1144),
            (0, 344),
            (0, 975),
            (0, 1755),
            (0, 1406),
            (0, 0),
            (0, 6),
            (0, 410),
            (0, 414),
            (0, 4003),
            (0, 60),
            (0, 1526),
            (0, 13),
            (0, 4748),
            (0, 2783),
            (0, 32),
            (0, 7631),
            (0, 1570),
            (0, 624),
            (0, 11),
            (0, 2),
            (0, 26),
            (0, 84),
            (0, 846),
            (0, 0),
            (0, 1043),
            (0, 61),
            (0, 56),
            (0, 5456),
            (0, 123),
            (0, 29),
            (0, 1671),
            (0, 26),
            (0, 2),
            (0, 119),
            (0, 20),
            (0, 58),
            (0, 568),
            (0, 8),
            (0, 51),
            (0, 2733),
            (0, 34),
            (0, 3910),
            (0, 564),
            (0, 34),
            (0, 0),
            (0, 10),
            (0, 1096),
            (0, 224),
            (0, 190),
            (0, 0),
            (0, 95),
            (0, 339),
            (0, 4519),
            (0, 1),
            (0, 247),
            (0, 16),
            (0, 19),
            (0, 4),
            (0, 476),
            (0, 409),
            (0, 184),
            (0, 24),
            (0, 27),
            (0, 1),
            (0, 20),
            (0, 297),
            (0, 251),
            (0, 18),
            (0, 482),
            (0, 74),
            (0, 138),
            (0, 1258),
            (0, 2),
            (0, 39),
            (0, 22),
            (0, 0),
            (0, 199),
            (0, 96),
            (0, 1340),
            (0, 78),
            (0, 517),
            (0, 462),
            (0, 4944),
            (0, 275),
            (0, 112),
            (0, 3054),
            (0, 572),
            (0, 0),
            (0, 11),
            (0, 6327),
            (0, 4743),
            (0, 12),
            (0, 60),
            (0, 66),
            (0, 296),
            (0, 272),
            (0, 155),
            (0, 798),
            (0, 61),
            (0, 56),
            (0, 3466),
            (0, 1961),
            (0, 9),
            (0, 63),
            (0, 5),
            (0, 14645),
            (0, 1),
            (0, 3),
            (0, 69),
            (0, 432),
            (0, 0),
            (0, 74),
            (0, 4577),
            (0, 0),
            (0, 201),
            (0, 0),
            (0, 162),
            (0, 276),
            (0, 24),
            (0, 316),
            (0, 78),
            (0, 42),
            (0, 5240),
            (0, 1154),
            (0, 92),
            (0, 140),
            (0, 408),
            (0, 1365),
            (0, 1417),
            (0, 6358),
            (0, 7095),
            (0, 2794),
            (0, 316),
            (0, 28),
            (0, 147),
            (0, 2),
            (0, 400),
            (0, 1523),
            (0, 85),
            (0, 43),
            (0, 1),
            (0, 1227),
            (0, 17),
            (0, 38),
            (0, 2498),
            (0, 17),
            (0, 278),
            (0, 8),
            (0, 160),
            (0, 898),
            (0, 2559),
            (0, 994),
            (0, 29),
            (0, 22755),
            (0, 928),
            (0, 0),
            (0, 11),
            (0, 1357),
            (0, 225),
            (0, 3158),
            (0, 3091),
            (0, 498),
            (0, 58),
            (0, 6966),
            (0, 17),
            (0, 795),
            (0, 238),
            (0, 204),
            (0, 9),
            (0, 4218),
            (0, 0),
            (0, 3030),
            (0, 1),
            (0, 1),
            (0, 262),
            (0, 616),
            (0, 317),
            (0, 51),
            (0, 3579),
            (0, 29),
            (0, 1787),
            (0, 4),
            (0, 8),
            (0, 6356),
            (0, 399),
            (0, 0),
            (0, 168),
            (0, 2819),
            (0, 2978),
            (0, 498),
            (0, 9),
            (0, 64),
            (0, 493),
            (0, 419),
            (0, 21),
            (0, 565),
            (0, 429),
            (0, 69),
            (0, 1809),
            (0, 2821),
            (0, 18),
            (0, 311),
            (0, 5),
            (0, 163),
            (0, 37),
            (0, 21),
            (0, 36),
            (0, 675),
            (0, 6247),
            (0, 251),
            (0, 733),
            (0, 1128),
            (0, 6),
            (0, 4402),
            (0, 6490),
            (0, 11),
            (0, 223),
            (0, 102),
            (0, 25),
            (0, 47),
            (0, 402),
            (0, 0),
            (0, 2703),
            (0, 0),
            (0, 234),
            (0, 2264),
            (0, 4379),
            (0, 0),
            (0, 6807),
            (0, 3842),
            (0, 39),
            (0, 16),
            (0, 7641),
            (0, 79),
            (0, 323),
            (0, 1482),
            (0, 66),
            (0, 142),
            (0, 17),
            (0, 19),
            (0, 4906),
            (0, 698),
            (0, 18),
            (0, 656),
            (0, 5),
            (0, 3610),
            (0, 4202),
            (0, 10817),
            (0, 241),
            (0, 8205),
            (0, 1749),
            (0, 1122),
            (0, 239),
            (0, 145),
            (0, 301),
            (0, 175),
            (0, 1389),
            (0, 99),
            (0, 0),
            (0, 2026),
            (0, 3),
            (0, 15261),
            (0, 0),
            (0, 51),
            (0, 25),
            (0, 349),
            (0, 0),
            (0, 1298),
            (0, 0),
            (0, 84),
            (0, 6),
            (0, 744),
            (0, 173),
            (0, 103),
            (0, 170),
            (0, 256),
            (0, 7),
            (0, 225),
            (0, 1011),
            (0, 51),
            (0, 6),
            (0, 3561),
            (0, 66),
            (0, 33),
            (0, 4530),
            (0, 15),
            (0, 46),
            (0, 1604),
            (0, 1213),
            (0, 342),
            (0, 10),
            (0, 563),
            (0, 952),
            (0, 264),
            (0, 3840),
            (0, 22),
            (0, 18),
            (0, 953),
            (0, 3),
            (0, 303),
            (0, 1138),
            (0, 24),
            (0, 4520),
            (0, 496),
            (0, 1154),
            (0, 2268),
            (0, 1714),
            (0, 461),
            (0, 1470),
            (0, 0),
            (0, 3),
            (0, 1399),
            (0, 2807),
            (0, 19),
            (0, 2),
            (0, 89),
            (0, 343),
            (0, 534),
            (0, 1874),
            (0, 57),
            (0, 125),
            (0, 10),
            (0, 8168),
            (0, 8),
            (0, 2772),
            (0, 11496),
            (0, 4136),
            (0, 3379),
            (0, 27),
            (0, 32),
            (0, 21),
            (0, 18),
            (0, 210),
            (0, 30),
            (0, 0),
            (0, 221),
            (0, 5380),
            (0, 33),
            (0, 2726),
            (0, 136),
            (0, 161),
            (0, 9),
            (0, 264),
            (0, 219),
            (0, 28),
            (0, 41),
            (0, 244),
            (0, 1401),
            (0, 1),
            (0, 3384),
            (0, 9555),
            (0, 97),
            (0, 67),
            (0, 95),
            (0, 3040),
            (0, 3137),
            (0, 85),
            (0, 1),
            (0, 1),
            (0, 252),
            (0, 111),
            (0, 181),
            (0, 22974),
            (0, 146),
            (0, 15206),
            (0, 704),
            (0, 9181),
            (0, 561),
            (0, 3984),
            (0, 9799),
            (0, 2726),
            (0, 86),
            (0, 2),
            (0, 839),
            (0, 0),
            (0, 1),
            (0, 79),
            (0, 58),
            (0, 219),
            (0, 5970),
            (0, 0),
            (0, 714),
            (0, 126),
            (0, 4098),
            (0, 1),
            (0, 1227),
            (0, 1),
            (0, 12),
            (0, 9121),
            (0, 0),
            (0, 5364),
            (0, 1434),
            (0, 214),
            (0, 1496),
            (0, 513),
            (0, 261),
            (0, 0),
            (0, 3),
            (0, 0),
            (0, 151),
            (0, 1800),
            (0, 75),
            (0, 1),
            (0, 43),
            (0, 13),
            (0, 11),
            (0, 2224),
            (0, 26),
            (0, 18),
            (0, 5125),
            (0, 64),
            (0, 2),
            (0, 479),
            (0, 2),
            (0, 1026),
            (0, 62),
            (0, 21543),
            (0, 280),
            (0, 632),
            (0, 31),
            (0, 634),
            (0, 3701),
            (0, 17),
            (0, 2265),
            (0, 95),
            (0, 188),
            (0, 4120),
            (0, 644),
            (0, 155),
            (0, 520),
            (0, 138),
            (0, 444),
            (0, 751),
            (0, 11763),
            (0, 1707),
            (0, 1),
            (0, 266),
            (0, 26),
            (0, 5617),
            (0, 644),
            (0, 1),
            (0, 33),
            (0, 2070),
            (0, 25),
            (0, 2054),
            (0, 805),
            (0, 440),
            (0, 13123),
            (0, 88),
            (0, 9),
            (0, 29),
            (0, 2),
            (0, 15),
            (0, 4868),
            (0, 767),
            (0, 990),
            (0, 13),
            (0, 12),
            (0, 27),
            (0, 7008),
            (0, 7),
            (0, 1423),
            (0, 10984),
            (0, 3306),
            (0, 912),
            (0, 413),
            (0, 19),
            (0, 8),
            (0, 3878),
            (0, 8),
            (0, 285),
            (0, 221),
            (0, 0),
            (0, 2835),
            (0, 310),
            (0, 34),
            (0, 349),
            (0, 30),
            (0, 3474),
            (0, 2),
            (0, 400),
            (0, 1),
            (0, 8),
            (0, 2737),
            (0, 4),
            (0, 150),
            (0, 349),
            (0, 4608),
            (0, 535),
            (0, 688),
            (0, 0),
            (0, 0),
            (0, 26),
            (0, 56),
            (0, 37),
            (0, 16),
            (0, 100),
            (0, 621),
            (0, 9),
            (0, 1532),
            (0, 5),
            (0, 829),
            (0, 228),
            (0, 42),
            (0, 0),
            (0, 0),
            (0, 46),
            (0, 16440),
            (0, 3456),
            (0, 64),
            (0, 39),
            (0, 933),
            (0, 1),
            (0, 1164),
            (0, 114),
            (0, 88),
            (0, 1),
            (0, 0),
            (0, 79),
            (0, 65),
            (0, 157),
            (0, 75),
            (0, 638),
            (0, 48),
            (0, 206),
            (0, 1104),
            (0, 256),
            (0, 0),
            (0, 0),
            (0, 651),
            (0, 2660),
            (0, 5),
            (0, 23),
            (0, 782),
            (0, 41),
            (0, 1),
            (0, 8),
            (0, 20303),
            (0, 16),
            (0, 828),
            (0, 346),
            (0, 1873),
            (0, 353),
            (0, 4116),
            (0, 226),
            (0, 44),
            (0, 24),
            (0, 2329),
            (0, 5609),
            (0, 0),
            (0, 488),
            (0, 141),
            (0, 3),
            (0, 1405),
            (0, 1518),
            (0, 986),
            (0, 5),
            (0, 1),
            (0, 1110),
            (0, 827),
            (0, 1),
            (0, 11559),
            (0, 0),
            (0, 37),
            (0, 4598),
            (0, 32),
            (0, 152),
            (0, 0),
            (0, 4303),
            (0, 276),
            (0, 48),
            (0, 0),
            (0, 6887),
            (0, 5),
            (0, 0),
            (0, 8547),
            (0, 401),
            (0, 4207),
            (0, 599),
            (0, 2587),
            (0, 0),
            (0, 489),
            (0, 917),
            (0, 3),
            (0, 225),
            (0, 920),
            (0, 215),
            (0, 4839),
            (0, 40),
            (0, 1281),
            (0, 456),
            (0, 2),
            (0, 2420),
            (0, 0),
            (0, 220),
            (0, 755),
            (0, 0),
            (0, 388),
            (0, 8),
            (0, 517),
            (0, 5),
            (0, 27),
            (0, 67),
            (0, 20644),
            (0, 142),
            (0, 8),
            (0, 202),
            (0, 4209),
            (0, 12772),
            (0, 1221),
            (0, 10054),
            (0, 750),
            (0, 127),
            (0, 4),
            (0, 263),
            (0, 116),
            (0, 340),
            (0, 2),
            (0, 3916),
            (0, 3530),
            (0, 50),
            (0, 1),
            (0, 511),
            (0, 5140),
            (0, 113),
            (0, 8),
            (0, 3),
            (0, 1142),
            (0, 29),
            (0, 11209),
            (0, 15),
            (0, 158),
            (0, 1920),
            (0, 3808),
            (0, 299),
            (0, 5218),
            (0, 51),
            (0, 12764),
            (0, 883),
            (0, 39),
            (0, 44),
            (0, 813),
            (0, 17),
            (0, 2174),
            (0, 395),
            (0, 8721),
            (0, 1397),
            (0, 2291),
            (0, 2744),
            (0, 98),
            (0, 2546),
            (0, 3032),
            (0, 0),
            (0, 3004),
            (0, 110),
            (0, 568),
            (0, 5),
            (0, 18479),
            (0, 836),
            (0, 1657),
            (0, 113),
            (0, 36),
            (0, 169),
            (0, 0),
            (0, 2),
            (0, 160),
            (0, 1216),
            (0, 3060),
            (0, 302),
            (0, 140),
            (0, 158),
            (0, 4807),
            (0, 1730),
            (0, 2),
            (0, 236),
            (0, 195),
            (0, 5256),
            (0, 15),
            (0, 104),
            (0, 2492),
            (0, 1),
            (0, 1717),
            (0, 8),
            (0, 98),
            (0, 31),
            (0, 14),
            (0, 952),
            (0, 4),
            (0, 3),
            (0, 6481),
            (0, 561),
            (0, 2890),
            (0, 1058),
            (0, 8),
            (0, 5),
            (0, 25),
            (0, 14),
            (0, 3),
            (0, 5998),
            (0, 4848),
            (0, 6015),
            (0, 53),
            (0, 33),
            (0, 16228),
            (0, 2069),
            (0, 206),
            (0, 1355),
            (0, 1),
            (0, 167),
            (0, 623),
            (0, 642),
            (0, 8),
            (0, 506),
            (0, 3826),
            (0, 120),
            (0, 134),
            (0, 71),
            (0, 484),
            (0, 144),
            (0, 417),
            (0, 308),
            (0, 12),
            (0, 0),
            (0, 15),
            (0, 1),
            (0, 5),
            (0, 3247),
            (0, 782),
            (0, 65),
            (0, 1),
            (0, 1460),
            (0, 56),
            (0, 3166),
            (0, 726),
            (0, 15),
            (0, 3280),
            (0, 44),
            (0, 94),
            (0, 489),
            (0, 2751),
            (0, 77),
            (0, 5184),
            (0, 19),
            (0, 167),
            (0, 201),
            (0, 1),
            (0, 60),
            (0, 199),
            (0, 4),
            (0, 44),
            (0, 334),
            (0, 13805),
            (0, 651),
            (0, 58),
            (0, 476),
            (0, 243),
            (0, 4276),
            (0, 0),
            (0, 433),
            (0, 173),
            (0, 1859),
            (0, 87),
            (0, 3728),
            (0, 12141),
            (0, 62),
            (0, 4),
            (0, 541),
            (0, 421),
            (0, 25),
            (0, 278),
            (0, 238),
            (0, 193),
            (0, 247),
            (0, 913),
            (0, 0),
            (0, 1),
            (0, 1416),
            (0, 5375),
            (0, 1903),
            (0, 28),
            (0, 1),
            (0, 3289),
            (0, 611),
            (0, 47),
            (0, 98),
            (0, 2922),
            (0, 270),
            (0, 10),
            (0, 286),
            (0, 119),
            (0, 65),
            (0, 1579),
            (0, 3738),
            (0, 0),
            (0, 839),
            (0, 103),
            (0, 67),
            (0, 105),
            (0, 6785),
            (0, 157),
            (0, 936),
            (0, 2336),
            (0, 1114),
            (0, 1493),
            (0, 2185),
            (0, 0),
            (0, 250),
            (0, 4242),
            (0, 97),
            (0, 1563),
            (0, 4985),
            (0, 6332),
            (0, 451),
            (0, 653),
            (0, 368),
            (0, 36),
            (0, 389),
            (0, 18),
            (0, 5),
            (0, 309),
            (0, 324),
            (0, 1384),
            (0, 193),
            (0, 998),
            (0, 292),
            (0, 329),
            (0, 22),
            (0, 102),
            (0, 407),
            (0, 245),
            (0, 7969),
            (0, 11),
            (0, 44),
            (0, 219),
            (0, 4453),
            (0, 1072),
            (0, 580),
            (0, 110),
            (0, 899),
            (0, 1),
            (0, 217),
            (0, 7),
            (0, 2072),
            (0, 2231),
            (0, 13),
            (0, 9),
            (0, 6),
            (0, 397),
            (0, 8287),
            (0, 0),
            (0, 310),
            (0, 4595),
            (0, 10271),
            (0, 8649),
            (0, 2185),
            (0, 2406),
            (0, 1807),
            (0, 142),
            (0, 12308),
            (0, 6329),
            (0, 2203),
            (0, 1194),
            (0, 1500),
            (0, 2752),
            (0, 194),
            (0, 0),
            (0, 65),
            (0, 63),
            (0, 7933),
            (0, 1855),
            (0, 316),
            (0, 1),
            (0, 1304),
            (0, 87),
            (0, 8911),
            (0, 4),
            (0, 92),
            (0, 4),
            (0, 41),
            (0, 25),
            (0, 1314),
            (0, 396),
            (0, 95),
            (0, 1),
            (0, 4801),
            (0, 146),
            (0, 6),
            (0, 2917),
            (0, 3),
            (0, 859),
            (0, 3155),
            (0, 3384),
            (0, 65),
            (0, 1077),
            (0, 22),
            (0, 1070),
            (0, 335),
            (0, 23),
            (0, 820),
            (0, 108),
            (0, 985),
            (0, 13),
            (0, 10),
            (0, 61),
            (0, 134),
            (0, 277),
            (0, 22),
            (0, 140),
            (0, 2207),
            (0, 1619),
            (0, 28),
            (0, 786),
            (0, 1768),
            (0, 1),
            (0, 2311),
            (0, 657),
            (0, 8836),
            (0, 168),
            (0, 351),
            (0, 0),
            (0, 178),
            (0, 354),
            (0, 15092),
            (0, 4),
            (0, 12273),
            (0, 283),
            (0, 7),
            (0, 11377),
            (0, 140),
            (0, 10770),
            (0, 1535),
            (0, 939),
            (0, 40),
            (0, 1),
            (0, 8),
            (0, 830),
            (0, 256),
            (0, 407),
            (0, 61),
            (0, 452),
            (0, 4343),
            (0, 5378),
            (0, 2666),
            (0, 81),
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 682),
            (0, 9),
            (0, 14),
            (0, 1066),
            (0, 98),
            (0, 827),
            (0, 14),
            (0, 2),
            (0, 6),
            (0, 2261),
            (0, 300),
            (0, 171),
            (0, 12),
            (0, 892),
            (0, 148),
            (0, 22),
            (0, 4593),
            (0, 24),
            (0, 294),
            (0, 3836),
            (0, 2678),
            (0, 2253),
            (0, 9),
            (0, 26),
            (0, 59),
            (0, 1),
            (0, 0),
            (0, 0),
            (0, 10861),
            (0, 3074),
            (0, 111),
            (0, 20),
            (0, 10135),
            (0, 418),
            (0, 1162),
            (0, 30),
            (0, 116),
            (0, 960),
            (0, 2554),
            (0, 406),
            (0, 5),
            (0, 10),
            (0, 7265),
            (0, 623),
            (0, 1860),
            (0, 2317),
            (0, 323),
            (0, 693),
            (0, 23825),
            (0, 4),
            (0, 6130),
            (0, 1),
            (0, 7),
            (0, 707),
            (0, 8),
            (0, 923),
            (0, 4),
            (0, 11463),
            (0, 6),
            (0, 1029),
            (0, 2046),
            (0, 1865),
            (0, 119),
            (0, 26),
            (0, 96),
            (0, 7),
            (0, 18),
            (0, 119),
            (0, 104),
            (0, 76),
            (0, 14028),
            (0, 12),
            (0, 27),
            (0, 43),
            (0, 1321),
            (0, 438),
            (0, 1621),
            (0, 6),
            (0, 18),
            (0, 1318),
            (0, 43),
            (0, 364),
            (0, 7),
            (0, 1),
            (0, 4816),
            (0, 112),
            (0, 1080),
            (0, 4097),
            (0, 0),
            (0, 388),
            (0, 1),
            (0, 30),
            (0, 87),
            (0, 205),
            (0, 6512),
            (0, 40),
            (0, 2826),
            (0, 8067),
            (0, 2945),
            (0, 41),
            (0, 881),
            (0, 99),
            (0, 6808),
            (0, 15),
            (0, 42),
            (0, 0),
            (0, 121),
            (0, 1),
            (0, 998),
            (0, 1248),
            (0, 3),
            (0, 32),
            (0, 0),
            (0, 2),
            (0, 2635),
            (0, 110),
            (0, 84),
            (0, 1032),
            (0, 226),
            (0, 9847),
            (0, 8534),
            (0, 2182),
            (0, 4),
            (0, 2),
            (0, 2982),
            (0, 200),
            (0, 1448),
            (0, 196),
            (0, 2959),
            (0, 322),
            (0, 63),
            (0, 654),
            (0, 1140),
            (0, 0),
            (0, 0),
            (0, 2977),
            (0, 1954),
            (0, 3221),
            (0, 1),
            (0, 522),
            (0, 253),
            (0, 2039),
            (0, 10),
            (0, 5270),
            (0, 1038),
            (0, 1953),
            (0, 255),
            (0, 170),
            (0, 11),
            (0, 13190),
            (0, 1),
            (0, 0),
            (0, 2556),
            (0, 123),
            (0, 2954),
            (0, 6),
            (0, 7830),
            (0, 69),
            (0, 6),
            (0, 0),
            (0, 27),
            (0, 862),
            (0, 1652),
            (0, 420),
            (0, 761),
            (0, 2),
            (0, 11835),
            (0, 311),
            (0, 1081),
            (0, 0),
            (0, 12),
            (0, 51),
            (0, 3),
            (0, 3192),
            (0, 267),
            (0, 7),
            (0, 550),
            (0, 1702),
            (0, 2262),
            (0, 3957),
            (0, 3675),
            (0, 81),
            (0, 244),
            (0, 78),
            (0, 70),
            (0, 67),
            (0, 7),
            (0, 799),
            (0, 9069),
            (0, 243),
            (0, 11112),
            (0, 620),
            (0, 44),
            (0, 40),
            (0, 984),
            (0, 1097),
            (0, 531),
            (0, 11),
            (0, 40),
            (0, 2),
            (0, 0),
            (0, 59),
            (0, 119),
            (0, 103),
            (0, 6211),
            (0, 61),
            (0, 39),
            (0, 1503),
            (0, 120),
            (0, 24),
            (0, 19),
            (0, 34),
            (0, 379),
            (0, 4384),
            (0, 1097),
            (0, 134),
            (0, 7715),
            (0, 1),
            (0, 40),
            (0, 4839),
            (0, 27),
            (0, 1152),
            (0, 69),
            (0, 16179),
            (0, 109),
            (0, 513),
            (0, 1),
            (0, 10),
            (0, 4),
            (0, 10),
            (0, 8169),
            (0, 389),
            (0, 270),
            (0, 8415),
            (0, 508),
            (0, 77),
            (0, 308),
            (0, 32),
            (0, 68),
            (0, 20868),
            (0, 186),
            (0, 258),
            (0, 87),
            (0, 15710),
            (0, 116),
            (0, 1788),
            (0, 159),
            (0, 887),
            (0, 1307),
            (0, 627),
            (0, 170),
            (0, 0),
            (0, 52),
            (0, 1),
            (0, 109),
            (0, 408),
            (0, 4562),
            (0, 468),
            (0, 3141),
            (0, 797),
            (0, 80),
            (0, 7470),
            (0, 7),
            (0, 232),
            (0, 3143),
            (0, 17510),
            (0, 1711),
            (0, 12),
            (0, 611),
            (0, 12),
            (0, 228),
            (0, 760),
            (0, 50),
            (0, 2024),
            (0, 374),
            (0, 2),
            (0, 1087),
            (0, 4646),
            (0, 0),
            (0, 3569),
            (0, 0),
            (0, 14288),
            (0, 31),
            (0, 13850),
            (0, 269),
            (0, 32),
            (0, 0),
            (0, 60),
            (0, 0),
            (0, 2730),
            (0, 26),
            (0, 3792),
            (0, 40),
            (0, 22),
            (0, 414),
            (0, 4),
            (0, 21),
            (0, 413),
            (0, 981),
            (0, 526),
            (0, 176),
            (0, 0),
            (0, 27),
            (0, 72),
            (0, 833),
            (0, 36),
            (0, 6),
            (0, 799),
            (0, 5988),
            (0, 704),
            (0, 2),
            (0, 8358),
            (0, 522),
            (0, 0),
            (0, 4),
            (0, 1),
            (0, 34),
            (0, 2),
            (0, 13),
            (0, 20),
            (0, 181),
            (0, 1),
            (0, 10554),
            (0, 32),
            (0, 715),
            (0, 23),
            (0, 46),
            (0, 587),
            (0, 56),
            (0, 212),
            (0, 478),
            (0, 540),
            (0, 0),
            (0, 80),
            (0, 163),
            (0, 12867),
            (0, 927),
            (0, 316),
            (0, 1105),
            (0, 1),
            (0, 828),
            (0, 420),
            (1, 0),
            (0, 21),
            (0, 1),
            (0, 283),
            (0, 3),
            (0, 8315),
            (0, 34),
            (0, 3439),
            (0, 6),
            (0, 4988),
            (0, 3217),
            (0, 5584),
            (0, 58),
            (0, 291),
            (0, 241),
            (0, 1364),
            (0, 32),
            (0, 211),
            (0, 9),
            (0, 22),
            (0, 162),
            (0, 139),
            (0, 17),
            (0, 3),
            (0, 14010),
            (0, 519),
            (0, 2298),
            (0, 182),
            (0, 9676),
            (0, 2),
            (0, 3943),
            (0, 499),
            (0, 2597),
            (0, 7787),
            (0, 23),
            (0, 3),
            (0, 107),
            (0, 3),
            (0, 1280),
            (0, 20),
            (0, 853),
            (0, 1704),
            (0, 0),
            (0, 3835),
            (0, 0),
            (0, 729),
            (0, 59),
            (0, 855),
            (0, 1531),
            (0, 523),
            (0, 2),
            (0, 2),
            (0, 25886),
            (0, 237),
            (0, 3),
            (0, 1114),
            (0, 13348),
            (0, 2),
            (0, 2),
            (0, 19),
            (0, 8465),
            (0, 13796),
            (0, 888),
            (0, 756),
            (0, 11704),
            (0, 9),
            (0, 84),
            (0, 3),
            (0, 74),
            (0, 23),
            (0, 625),
            (0, 0),
            (0, 981),
            (0, 2),
            (0, 80),
            (0, 257),
            (0, 950),
            (0, 3328),
            (0, 41),
            (0, 13003),
            (0, 4),
            (0, 29),
            (0, 41),
            (0, 402),
            (0, 0),
            (0, 6564),
            (0, 11),
            (0, 6637),
            (0, 1515),
            (0, 5320),
            (0, 218),
            (0, 0),
            (0, 783),
            (0, 3923),
            (0, 180),
            (0, 25),
            (0, 60),
            (0, 374),
            (0, 186),
            (0, 56),
            (0, 15),
            (0, 1832),
            (0, 86),
            (0, 11770),
            (0, 823),
            (0, 39),
            (0, 0),
            (0, 8025),
            (0, 181),
            (0, 0),
            (0, 12479),
            (0, 3),
            (0, 0),
            (0, 282),
            (0, 242),
            (0, 7438),
            (0, 19),
            (0, 17),
            (0, 204),
            (0, 2024),
            (0, 7872),
            (0, 431),
            (0, 2),
            (0, 124),
            (0, 2),
            (0, 2200),
            (0, 225),
            (0, 64),
            (0, 11),
            (0, 201),
            (0, 3),
            (0, 306),
            (0, 9),
            (0, 363),
            (0, 55),
            (0, 170),
            (0, 3342),
            (0, 23),
            (0, 1189),
            (0, 2495),
            (0, 11),
            (0, 509),
            (0, 172),
            (0, 741),
            (0, 347),
            (0, 27),
            (0, 125),
            (0, 3),
            (0, 7568),
            (0, 19),
            (0, 290),
            (0, 7990),
            (0, 985),
            (0, 930),
            (0, 53),
            (0, 26),
            (0, 2340),
            (0, 42),
            (0, 14),
            (0, 1075),
            (0, 2556),
            (0, 4699),
            (0, 10),
            (0, 422),
            (0, 15495),
            (0, 286),
            (0, 93),
            (0, 6909),
            (0, 38),
            (0, 2237),
            (0, 2),
            (0, 105),
            (0, 241),
            (0, 721),
            (0, 31),
            (0, 2235),
            (0, 1928),
            (0, 20),
            (0, 468),
            (0, 158),
            (0, 241),
            (0, 2145),
            (0, 1896),
            (0, 24),
            (0, 14989),
            (0, 2),
            (0, 467),
            (0, 190),
            (0, 208),
            (0, 1),
            (0, 878),
            (0, 2654),
            (0, 45),
            (0, 69),
            (0, 101),
            (0, 126),
            (0, 1018),
            (0, 2),
            (0, 19),
            (0, 45),
            (0, 1558),
            (0, 382),
            (0, 556),
            (0, 892),
            (0, 728),
            (0, 0),
            (0, 4),
            (0, 0),
            (0, 2519),
            (0, 50),
            (0, 3627),
            (0, 263),
            (0, 1197),
            (0, 61),
            (0, 630),
            (0, 351),
            (0, 67),
            (0, 625),
            (0, 192),
            (0, 212),
            (0, 177),
            (0, 50),
            (0, 627),
            (0, 63),
            (0, 21),
            (0, 8832),
            (0, 344),
            (0, 821),
            (0, 6339),
            (0, 7),
            (0, 713),
            (0, 5952),
            (0, 2604),
            (0, 0),
            (0, 781),
            (0, 636),
            (0, 190),
            (0, 255),
            (0, 3),
            (0, 785),
            (0, 1),
            (0, 17),
            (0, 7301),
            (0, 44),
            (0, 10006),
            (0, 129),
            (0, 0),
            (0, 187),
            (0, 150),
            (0, 678),
            (0, 15),
            (0, 172),
            (0, 2949),
            (0, 12),
            (0, 303),
            (0, 3),
            (0, 1815),
            (0, 60),
            (0, 2297),
            (0, 16),
            (0, 5),
            (0, 57),
            (0, 2),
            (0, 256),
            (0, 169),
            (0, 365),
            (0, 432),
            (0, 3046),
            (0, 0),
            (0, 0),
            (0, 525),
            (0, 5925),
            (0, 8),
            (0, 1),
            (0, 19),
            (0, 16),
            (0, 2),
            (0, 0),
            (0, 201),
            (0, 5524),
            (0, 811),
            (0, 9385),
            (0, 1916),
            (0, 17),
            (0, 7596),
            (0, 8),
            (0, 19),
            (0, 12),
            (0, 41),
            (0, 55),
            (0, 184),
            (0, 295),
            (0, 150),
            (0, 12),
            (0, 4),
            (0, 4499),
            (0, 16),
            (0, 412),
            (0, 3727),
            (0, 5),
            (0, 107),
            (0, 799),
            (0, 3914),
            (0, 21),
            (0, 747),
            (0, 18),
            (0, 621),
            (0, 655),
            (0, 55),
            (0, 0),
            (0, 1363),
            (0, 378),
            (0, 2),
            (0, 0),
            (0, 2),
            (0, 786),
            (0, 324),
            (0, 5233),
            (0, 331),
            (0, 6120),
            (0, 363),
            (0, 3014),
            (0, 17),
            (0, 5165),
            (0, 7791),
            (0, 1922),
            (0, 3),
            (0, 4),
            (0, 1078),
            (0, 1546),
            (0, 3644),
            (0, 219),
            (0, 0),
            (0, 31060),
            (0, 75),
            (0, 8333),
            (0, 319),
            (0, 17),
            (0, 346),
            (0, 12),
            (0, 37),
            (0, 53),
            (0, 1),
            (0, 1603),
            (0, 536),
            (0, 16382),
            (0, 292),
            (0, 127),
            (0, 4348),
            (0, 388),
            (0, 5),
            (0, 9364),
            (0, 13908),
            (0, 6),
            (0, 29),
            (0, 2747),
            (0, 73),
            (0, 602),
            (0, 25636),
            (0, 1248),
            (0, 122),
            (0, 131),
            (0, 0),
            (0, 14),
            (0, 9),
            (0, 829),
            (0, 110),
            (0, 15),
            (0, 134),
            (0, 1771),
            (0, 176),
            (0, 3526),
            (0, 371),
            (0, 115),
            (0, 5),
            (0, 2421),
            (0, 45),
            (0, 4148),
            (0, 4571),
            (0, 10),
            (0, 1042),
            (0, 349),
            (0, 330),
            (0, 4477),
            (0, 1930),
            (0, 234),
            (0, 210),
            (0, 27),
            (0, 903),
            (0, 0),
            (0, 219),
            (0, 94),
            (0, 128),
            (0, 64),
            (0, 3),
            (0, 980),
            (0, 3457),
            (0, 0),
            (0, 0),
            (0, 7137),
            (0, 8),
            (0, 121),
            (0, 8),
            (0, 149),
            (0, 0),
            (0, 917),
            (0, 3292),
            (0, 36),
            (0, 1164),
            (0, 1),
            (0, 10),
            (0, 6689),
            (0, 445),
            (0, 9),
            (0, 36),
            (0, 50),
            (0, 9396),
            (0, 692),
            (0, 5770),
            (0, 0),
            (0, 0),
            (0, 0),
            (0, 281),
            (0, 4199),
            (0, 13227),
            (0, 491),
            (0, 617),
            (0, 1667),
            (0, 311),
            (0, 344),
            (0, 3933),
            (0, 157),
            (0, 1752),
            (0, 9),
            (0, 248),
            (0, 7882),
            (0, 134),
            (0, 1091),
            (0, 1),
            (0, 29),
            (0, 8),
            (0, 3),
            (0, 213),
            (0, 45),
            (0, 142),
            (0, 2049),
            (0, 1549),
            (0, 19463),
            (0, 3505),
            (0, 3),
            (0, 1033),
            (0, 71),
            (0, 1764),
            (0, 2),
            (0, 952),
            (0, 6),
            (0, 4388),
            (0, 602),
            (0, 1158),
            (0, 256),
            (0, 3114),
            (0, 4),
            (0, 28),
            (0, 1),
            (0, 157),
            (0, 0),
            (0, 11),
            (0, 1297),
            (0, 295),
            (0, 177),
            (0, 2),
            (0, 0),
            (0, 6507),
            (0, 12980),
            (0, 1),
            (0, 3),
            (0, 2732),
            (0, 41),
            (0, 425),
            (0, 204),
            (0, 19),
            (0, 413),
            (0, 3878),
            (0, 1520),
            (0, 323),
            (0, 2445),
            (0, 109),
            (0, 186),
            (0, 999),
            (0, 273),
            (0, 1),
            (0, 9447),
            (0, 1887),
            (0, 5258),
            (0, 11203),
            (0, 2),
            (0, 1038),
            (0, 3),
            (0, 8830),
            (0, 19),
            (0, 3170),
            (0, 4),
            (0, 2714),
            (0, 160),
            (0, 4),
            (0, 48),
            (0, 1761),
            (0, 56),
            (0, 710),
            (0, 1996),
            (0, 136),
            (0, 1718),
            (0, 1),
            (0, 54),
            (0, 28),
            (0, 324),
            (0, 670),
            (0, 12),
            (0, 33),
            (0, 336),
            (0, 171),
            (0, 20),
            (0, 1),
            (0, 0),
            (0, 101),
            (0, 178),
            (0, 132),
            (0, 0),
            (0, 0),
            (0, 1),
            (0, 615),
            (0, 3253),
            (0, 0),
            (0, 95),
            (0, 133),
            (0, 55),
            (0, 602),
            (0, 1),
            (0, 245),
            (0, 974),
            (0, 1258),
            (0, 2062),
            (0, 33),
            (0, 3471),
            (0, 413),
            (0, 11881),
            (0, 223),
            (0, 403),
            (0, 265),
            (0, 176),
            (0, 2626),
            (0, 3956),
            (0, 23),
            (0, 165),
            (0, 201),
            (0, 1074),
            (0, 17),
            (0, 486),
            (0, 994),
            (0, 220),
            (0, 26),
            (0, 185),
            (0, 12616),
            (0, 24),
            (0, 0),
            (0, 3),
            (0, 324),
            (0, 525),
            (0, 6),
            (0, 3),
            (0, 4708),
            (0, 27),
            (0, 8),
            (0, 0),
            (0, 140),
            (0, 1085),
            (0, 12499),
            (0, 2),
            (0, 166),
            (0, 295),
            (0, 34),
            (0, 14),
            (0, 171),
            (0, 1256),
            (0, 127),
            (0, 364),
            (0, 2),
            (0, 41),
            (0, 892),
            (0, 343),
            (0, 4),
            (0, 1147),
            (0, 15095),
            (0, 6150),
            (0, 2),
            (0, 20),
            (0, 43),
            (0, 27),
            (0, 11453),
            (0, 1196),
            (0, 20579),
            (0, 11),
            (0, 2789),
            (0, 0),
            (0, 6),
            (0, 3218),
            (0, 462),
            (0, 90),
            (0, 2),
            (0, 449),
            (0, 0),
            (0, 28),
            (0, 1967),
            (0, 37),
            (0, 221),
            (0, 0),
            (0, 7353),
            (0, 8304),
            (0, 8124),
            (0, 120),
            (0, 5220),
            (0, 235),
            (0, 236),
            (0, 30),
            (0, 2067),
            (0, 92),
            (0, 9),
            (0, 354),
            (0, 214),
            (0, 5),
            (0, 1),
            (0, 9),
            (0, 138),
            (0, 101),
            (0, 1991),
            (0, 0),
            (0, 100),
            (0, 7558),
            (0, 50),
            (0, 439),
            (0, 110),
            (0, 2),
            (0, 212),
            (0, 2423),
            (0, 13362),
            (0, 137),
            (0, 1913),
            (0, 28),
            (0, 99),
            (0, 4823),
            (0, 5745),
            (0, 66),
            (0, 411),
            (0, 226),
            (0, 8179),
            (0, 2004),
            (0, 57),
            (0, 23),
            (0, 2124),
            (0, 185),
            (0, 5714),
            (0, 95),
            (0, 111),
            (0, 80),
            (0, 9392),
            (0, 175),
            (0, 10),
            (0, 1546),
            (0, 5),
            (0, 1191),
            (0, 16),
            (0, 56),
            (0, 699),
            (0, 378),
            (0, 2444),
            (0, 18),
            (0, 15),
            (0, 1657),
            (0, 190),
            (0, 14),
            (0, 1),
            (0, 1035),
            (0, 10308),
            (0, 97),
            (0, 7923),
            (0, 1415),
            (0, 6),
            (0, 6615),
            (0, 157),
            (0, 4),
            (0, 7063),
            (0, 1),
            (0, 60),
            (0, 1850),
            (0, 39),
            (0, 221),
            (0, 14601),
            (0, 6835),
            (0, 86),
            (0, 18690),
            (0, 299),
            (0, 1),
            (0, 17),
            (0, 69),
            (0, 2293),
            (0, 339),
            (0, 403),
            (0, 119),
            (0, 19),
            (0, 10073),
            (0, 366),
            (0, 1),
            (0, 624),
            (0, 97),
            (0, 7),
            (0, 16255),
            (0, 1557),
            (0, 1),
            (0, 0),
            (0, 3192),
            (0, 1363),
            (0, 55),
            (0, 269),
            (0, 843),
            (0, 0),
            (0, 22),
            (0, 2688),
            (0, 13),
            (0, 1),
            (0, 78),
            (0, 1),
            (0, 0),
            (0, 893),
            (0, 10445),
            (0, 681),
            (0, 466),
            (0, 3825),
            (0, 8225),
            (0, 5521),
            (0, 11903),
            (0, 0),
            (0, 882),
            (0, 603),
            (0, 343),
            (0, 93),
            (0, 6),
            (0, 389),
            (0, 40),
            (0, 1980),
            (0, 1304),
            (0, 19),
            (0, 3391),
            (0, 1),
            (0, 233),
            (0, 1903),
            (0, 10509),
            (0, 135),
            (0, 403),
            (0, 8748),
            (0, 81),
            (0, 8),
            (0, 1),
            (0, 1527),
            (0, 0),
            (0, 19),
            (0, 54),
            (0, 321),
            (0, 689),
            (0, 130),
            (0, 386),
            (0, 4),
            (0, 2615),
            (0, 468),
            (0, 5598),
            (0, 7),
            (0, 149),
            (0, 8254),
            (0, 3),
            (0, 14),
            (0, 32),
            (0, 426),
            (0, 10450),
            (0, 4),
            (0, 3),
            (0, 711),
            (0, 53),
            (0, 1529),
            (0, 10),
            (0, 3291),
            (0, 516),
            (0, 3),
            (0, 184),
            (0, 11),
            (0, 35),
            (0, 141),
            (0, 429),
            (0, 6),
            (0, 1),
            (0, 0),
            (0, 10995),
            (0, 290),
            (0, 838),
            (0, 28),
            (0, 36),
            (0, 25),
            (0, 1781),
            (0, 283),
            (0, 127),
            (0, 7),
            (0, 25),
            (0, 37),
            (0, 282),
            (0, 895),
            (0, 1633),
            (0, 0),
            (0, 41),
            (0, 2687),
            (0, 12557),
            (0, 777),
            (0, 964),
            (0, 3205),
            (0, 1),
            (0, 206),
            (0, 3),
            (0, 1012),
            (0, 9),
            (0, 2535),
            (0, 23),
            (0, 153),
            (0, 5),
            (0, 6669),
            (0, 12),
            (0, 168),
            (0, 27),
            (0, 60),
            (0, 1117),
            (0, 484),
            (0, 2),
            (0, 194),
            (0, 12),
            (0, 65),
            (0, 359),
            (0, 2106),
            (0, 29),
            (0, 894),
            (0, 381),
            (0, 837),
            (0, 7489),
            (0, 614),
            (0, 298),
            (0, 124),
            (0, 1415),
            (0, 3),
            (0, 577),
            (0, 1657),
            (0, 22),
            (0, 67),
            (0, 156),
            (0, 16),
            (0, 21),
            (0, 251),
            (0, 149),
            (0, 290),
            (0, 423),
            (0, 117),
            (0, 439),
            (0, 40),
            (0, 771),
            (0, 15872),
            (0, 4005),
            (0, 101),
            (0, 996),
            (0, 1313),
            (0, 5),
            (0, 28),
            (0, 1416),
            (0, 42),
            (0, 82),
            (0, 24),
            (0, 124),
            (0, 1020),
            (0, 14670),
            (0, 605),
            (0, 172),
            (0, 1530),
            (0, 0),
            (0, 1025),
            (0, 11),
            (0, 1331),
            (0, 0),
            (0, 164),
            (0, 277),
            (0, 3567),
            (0, 41),
            (0, 335),
            (0, 12),
            (0, 353),
            (0, 15),
            (0, 2335),
            (0, 13),
            (0, 16440),
            (0, 21),
            (0, 1054),
            (0, 0),
            (0, 56),
            (0, 2338),
            (0, 15291),
            (0, 65),
            (0, 5030),
            (0, 736),
            (0, 2023),
            (0, 1185),
            (0, 1540),
            (0, 4855),
            (0, 770),
            (0, 71),
            (0, 95),
            (0, 13),
            (0, 3110),
            (0, 1838),
            (0, 99),
            (0, 26),
            (0, 65),
            (0, 896),
            (0, 5),
            (0, 7),
            (0, 1424),
            (0, 7485),
            (0, 1066),
            (0, 0),
            (0, 78),
            (0, 400),
            (0, 385),
            (0, 61),
            (0, 363),
            (0, 2168),
            (0, 489),
            (0, 235),
            (0, 3),
            (0, 30),
            (0, 337),
            (0, 127),
            (0, 437),
            (0, 140),
            (0, 75),
            (0, 18),
            (0, 1914),
            (0, 40),
            (0, 125),
            (0, 641),
            (0, 496),
            (0, 0),
            (0, 105),
            (0, 10367),
            (0, 491),
            (0, 7),
            (0, 937),
            (0, 514),
            (0, 2),
            (0, 72),
            (0, 3115),
            (0, 17084),
            (0, 1392),
            (0, 69),
            (0, 2016),
            (0, 541),
            (0, 2673),
            (0, 0),
            (0, 804),
            (0, 13923),
            (0, 14),
            (0, 22179),
            (0, 0),
            (0, 3312),
            (0, 3),
            (0, 11),
            (0, 559),
            (0, 2),
            (0, 1538),
            (0, 24),
            (0, 13),
            (0, 3233),
            (0, 562),
            (0, 34),
            (0, 3188),
            (0, 428),
            (0, 6),
            (0, 26),
            (0, 23),
            (0, 1185),
            (0, 1438),
            (0, 7),
            (0, 19),
            (0, 3853),
            (0, 39),
            (0, 0),
            (0, 628),
            (0, 48),
            (0, 1296),
            (0, 769),
            (0, 8496),
            (0, 271),
            (0, 200),
            (0, 52571),
            (0, 32),
            (0, 2),
            (0, 10973),
            (0, 1496),
            (0, 666),
            (0, 2310),
            (0, 2),
            (0, 2157),
            (0, 277),
            (0, 12041),
            (0, 4807),
            (0, 6),
            (0, 2516),
            (0, 157),
            (0, 2551),
            (0, 554),
            (0, 64),
            (0, 6),
            (0, 3979),
            (0, 174),
            (0, 807),
            (0, 4065),
            (0, 762),
            (0, 0),
            (0, 15),
            (0, 64),
            (0, 125),
            (0, 13771),
            (0, 2672),
            (0, 6282),
            (0, 151),
            (0, 23),
            (0, 27),
            (0, 0),
            (0, 5),
            (0, 4042),
            (0, 113),
            (0, 109),
            (0, 73),
            (0, 0),
            (0, 7964),
            (0, 2110),
            (0, 13),
            (0, 482),
            (0, 1253),
            (0, 13),
            (0, 8),
            (0, 71),
            (0, 172),
            (0, 2775),
            (0, 1171),
            (0, 130),
            (0, 202),
            (0, 311),
            (0, 0),
            (0, 760),
            (0, 65),
            (0, 2448),
            (0, 3938),
            (0, 3081),
            (0, 130),
            (0, 2741),
            (0, 11),
            (0, 59),
            (0, 0),
            (0, 39),
            (0, 168),
            (0, 556),
            (0, 120),
            (0, 1654),
            (0, 3134),
            (0, 24),
            (0, 2),
            (0, 32),
            (0, 9),
            (0, 7351),
            (0, 15),
            (0, 584),
            (0, 3620),
            (0, 21546),
            (0, 1069),
            (0, 83),
            (0, 0),
            (0, 63),
            (0, 0),
            (0, 316),
            (0, 1170),
            (0, 128),
            (0, 565),
            (0, 476),
            (0, 69),
            (0, 23),
            (0, 122),
            (0, 1053),
            (0, 2147),
            (0, 35),
            (0, 8383),
            (0, 36),
            (0, 3797),
            (0, 6293),
            (0, 23),
            (0, 70),
            (0, 10313),
            (0, 6),
            (0, 116),
            (0, 1),
            (0, 3970),
            (0, 161),
            (0, 1935),
            (0, 86),
            (0, 631),
            (0, 1729),
            (0, 213),
            (0, 22),
            (0, 715),
            (0, 311),
            (0, 2760),
            (0, 57),
            (0, 4363),
            (0, 1),
            (0, 368),
            (0, 6),
            (0, 7),
            (0, 47),
            (0, 2258),
            (0, 0),
            (0, 7),
            (0, 1786),
            (0, 10),
            (0, 0),
            (0, 334),
            (0, 1),
            (0, 1),
            (0, 2),
            (0, 96),
            (0, 226),
            (0, 0),
            (0, 79),
            (0, 28),
            (0, 2162),
            (0, 2),
            (0, 43),
            (0, 8773),
            (0, 31),
            (0, 9),
            (0, 583),
            (0, 18),
            (0, 50),
            (0, 124),
            (0, 1031),
            (0, 8),
            (0, 1236),
            (0, 1),
            (0, 56),
            (0, 4476),
            (0, 419),
            (0, 3270),
            (0, 21271),
            (0, 2326),
            (0, 8778),
            (0, 3330),
            (0, 1667),
            (0, 1103),
            (0, 128),
            (0, 1),
            (0, 47),
            (0, 128),
            (0, 10129),
            (0, 5855),
            (0, 1893),
            (0, 3305),
            (0, 335),
            (0, 205),
            (0, 34),
            (0, 10004),
            (0, 3619),
            (0, 6263),
            (0, 22335),
            (0, 1411),
            (0, 1),
            (0, 89),
            (0, 46),
            (0, 499),
            (0, 389),
            (0, 241),
            (0, 77),
            (0, 369),
            (0, 2480),
            (0, 252),
            (0, 611),
            (0, 2583),
            (0, 769),
            (0, 2364),
            (0, 451),
            (0, 172),
            (0, 4414),
            (0, 738),
            (0, 1499),
            (0, 25),
            (0, 2033),
            (0, 17619),
            (0, 9),
            (0, 12),
            (0, 326),
            (0, 0),
            (0, 70),
            (0, 53),
            (0, 5023),
            (0, 6184),
            (0, 3),
            (0, 1613),
            (0, 8112),
            (0, 4972),
            (0, 4254),
            (0, 16398),
            (0, 19),
            (0, 22126),
            (0, 684),
            (0, 256),
            (0, 6955),
            (0, 72),
            (0, 16),
            (0, 1964),
            (0, 387),
            (0, 33),
            (0, 246),
            (0, 5490),
            (0, 0),
            (0, 1494),
            (0, 611),
            (0, 613),
            (0, 150),
            (0, 0),
            (0, 1450),
            (0, 3700),
            (0, 4),
            (0, 1440),
            (0, 1149),
            (0, 498),
            (0, 1421),
            (0, 242),
            (0, 484),
            (0, 1),
            (0, 55),
            (0, 623),
            (0, 578),
            (0, 17),
            (0, 18),
            (0, 477),
            (0, 141),
            (0, 40),
            (0, 753),
            (0, 186),
            (0, 402),
            (0, 18),
            (0, 4456),
            (0, 12896),
            (0, 27),
            (0, 248),
            (0, 475),
            (0, 0),
            (0, 6753),
            (0, 5),
            (0, 497),
            (0, 4),
            (0, 5),
            (0, 1),
            (0, 2304),
            (0, 33),
            (0, 3),
            (0, 1741),
            (0, 1708),
            (0, 278),
            (0, 33),
            (0, 2119),
            (0, 12597),
            (0, 9),
            (0, 0),
            (0, 79),
            (0, 1050),
            (0, 1698),
            (0, 32),
            (0, 0),
            (0, 14),
            (0, 456),
            (0, 2253),
            (0, 907),
            (0, 0),
            (0, 8620),
            (0, 77),
            (0, 1449),
            (0, 5342),
            (0, 0),
            (0, 2139),
            (0, 3),
            (0, 0),
            (0, 109),
            (0, 3527),
            (0, 217),
            (0, 11),
            (0, 1265),
            (0, 8),
            (0, 85),
            (0, 33341),
            (0, 18415),
            (0, 18609),
            (0, 3),
            (0, 141),
            (0, 1340),
            (0, 4),
            (0, 7142),
            (0, 326),
            (0, 2),
            (0, 0),
            (0, 47),
            (0, 1408),
            (0, 7132),
            (0, 1367),
            (0, 1),
            (0, 4),
            (0, 961),
            (0, 201),
            (0, 4203),
            (0, 6245),
            (0, 1),
            (0, 760),
            (0, 1530),
            (0, 43),
            (0, 2711),
            (0, 33),
            (0, 409),
            (0, 3),
            (0, 136),
            (0, 2589),
            (0, 4050),
            (0, 159),
            (0, 16400),
            (0, 1261),
            (0, 651),
            (0, 547),
            (0, 1390),
            (0, 15005),
            (0, 1033),
            (0, 132),
            (0, 43),
            (0, 677),
            (0, 3325),
            (0, 3),
            (0, 9534),
            (0, 0),
            (0, 400),
            (0, 1),
            (0, 1383),
            (0, 1780),
            (0, 965),
            (0, 2544),
            (0, 939),
            (0, 2),
            (0, 153),
            (0, 85),
            (0, 145),
            (0, 43),
            (0, 1021),
            (0, 1523),
            (0, 26),
            (0, 12137),
            (0, 418),
            (0, 1172),
            (0, 319),
            (0, 16),
            (0, 8),
            (0, 0),
            (0, 48),
            (0, 230),
            (0, 272),
            (0, 167),
            (0, 25919),
            (0, 16),
            (0, 2399),
            (0, 7),
            (0, 14819),
            (0, 118),
            (0, 4869),
            (0, 14),
            (0, 0),
            (0, 880),
            (0, 37),
            (0, 1439),
            (0, 1350),
            (0, 5105),
            (0, 1072),
            (0, 1361),
            (0, 2),
            (0, 9392),
            (0, 9794),
            (0, 3),
            (0, 1137),
            (0, 804),
            (0, 5714),
            (0, 1),
            (0, 846),
            (0, 105),
            (0, 454),
            (0, 9088),
            (0, 17),
            (0, 0),
            (0, 536),
            (0, 14007),
            (0, 77),
            (0, 0),
            (0, 59),
            (0, 4951),
            (0, 564),
            (0, 2845),
            (0, 6),
            (0, 1018),
            (0, 26),
            (0, 164),
            (0, 8194),
            (0, 48),
            (0, 3),
            (0, 3),
            (0, 476),
            (0, 50),
            (0, 17),
            (0, 84),
            (0, 4590),
            (0, 4),
            (0, 363),
            (0, 0),
            (0, 111),
            (0, 43),
            (0, 521),
            (0, 0),
            (0, 0),
            (0, 2267),
            (0, 7),
            (0, 8),
            (0, 1),
            (0, 17),
            (0, 1),
            (0, 1),
            (0, 5),
            (0, 2518),
            (0, 29),
            (0, 16),
            (0, 0),
            (0, 344),
            (0, 295),
            (0, 120),
            (0, 252),
            (0, 4255),
            (0, 2627),
            (0, 9249),
            (0, 191),
            (0, 2693),
            (0, 5024),
            (0, 30),
            (0, 0),
            (0, 20366),
            (0, 1),
            (0, 182),
            (0, 1),
            (0, 9509),
            (0, 391),
            (0, 2539),
            (0, 3),
            (0, 31),
            (0, 53),
            (0, 54),
            (0, 0),
            (0, 1),
            (0, 405),
            (0, 24815),
            (0, 10193),
            (0, 198),
            (0, 2),
            (0, 165),
            (0, 968),
            (0, 39),
            (0, 2049),
            (0, 20194),
            (0, 1),
            (0, 22),
            (0, 6653),
            (0, 189),
            (0, 139),
            (0, 2437),
            (0, 14697),
            (0, 1852),
            (0, 0),
            (0, 0),
            (0, 2531),
            (0, 58),
            (0, 2319),
            (0, 3),
            (0, 6),
            (0, 1037),
            (0, 20794),
            (0, 284),
            (0, 25),
            (0, 149),
            (0, 9),
            (0, 680),
            (0, 1693),
            (0, 24),
            (0, 1),
            (0, 629),
            (0, 3049),
            (0, 2552),
            (0, 33),
            (0, 988),
            (0, 11),
            (0, 705),
            (0, 109),
            (0, 5),
            (0, 2),
            (0, 2283),
            (0, 3),
            (0, 352),
            (0, 7),
            (0, 998),
            (0, 52),
            (0, 7185),
            (0, 12),
            (0, 1),
            (0, 4136),
            (0, 2934),
            (0, 10118),
            (0, 114),
            (0, 6),
            (0, 1812),
            (0, 1660),
            (0, 2734),
            (0, 1183),
            (0, 10537),
            (0, 7460),
            (0, 23003),
            (0, 3),
            (0, 194),
            (0, 1883),
            (0, 493),
            (0, 66),
            (0, 3230),
            (0, 1134),
            (0, 1792),
            (0, 559),
            (0, 21623),
            (0, 2825),
            (0, 1384),
            (0, 964),
            (0, 2),
            (0, 134),
            (0, 3011),
            (0, 5319),
            (0, 349),
            (0, 1),
            (0, 0),
            (0, 13),
            (0, 167),
            (0, 4695),
            (0, 25),
            (0, 134),
            (0, 6023),
            (0, 220),
            (0, 2),
            (0, 4),
            (0, 11298),
            (0, 149),
            (0, 11),
            (0, 287),
            (0, 13),
            (0, 0),
            (0, 193),
            (0, 35),
            (0, 273),
            (0, 3),
            (0, 689),
            (0, 6307),
            (0, 19852),
            (0, 287),
            (0, 394),
            (0, 0),
            (0, 1610),
            (0, 1),
            (0, 70),
            (0, 1567),
            (0, 1),
            (0, 1),
            (0, 54),
            (0, 620),
            (0, 3809),
            (0, 1481),
            (0, 0),
            (0, 0),
            (0, 16),
            (0, 728),
            (0, 91),
            (0, 12),
            (0, 867),
            (0, 31),
            (0, 12126),
            (0, 0),
            (0, 13394),
            (0, 30582),
            (0, 53),
            (0, 0),
            (0, 8272),
            (0, 34),
            (0, 191),
            (0, 487),
            (0, 126),
            (0, 3393),
            (0, 1),
            (0, 160),
            (0, 4615),
            (0, 655),
            (0, 120),
            (0, 544),
            (0, 29),
            (0, 21),
            (0, 0),
            (0, 6216),
            (0, 4),
            (0, 78),
            (0, 8001),
            (0, 140),
            (0, 58),
            (0, 5638),
            (0, 32),
            (0, 261),
            (0, 19),
            (0, 7784),
            (0, 10),
            (0, 476),
            (0, 5),
            (0, 298),
            (0, 2216),
            (0, 115),
            (0, 897),
            (0, 43),
            (0, 1),
            (0, 0),
            (0, 972),
            (0, 3893),
            (0, 53),
            (0, 0),
            (0, 59),
            (0, 143),
            (0, 32336),
            (0, 2584),
            (0, 202),
            (0, 45),
            (0, 5935),
            (0, 109),
            (0, 1826),
            (0, 0),
            (0, 14017),
            (0, 5),
            (0, 2352),
            (0, 592),
            (0, 5647),
            (0, 252),
            (0, 2327),
            (0, 21),
            (0, 33841),
            (0, 232),
            (0, 658),
            (0, 223),
            (0, 47),
            (0, 4158),
            (0, 33),
            (0, 6),
            (0, 1163),
            (0, 22331),
            (0, 167),
            (0, 986),
            (0, 1031),
            (0, 0),
            (0, 70),
            (0, 37),
            (0, 47),
            (0, 130),
            (0, 139),
            (0, 164),
            (0, 21239),
            (0, 438),
            (0, 12186),
            (0, 3178),
            (0, 603),
            (0, 0),
            (0, 3539),
            (0, 12),
            (0, 636),
            (0, 8),
            (0, 2058),
            (0, 0),
            (0, 341),
            (0, 2000),
            (0, 7355),
            (0, 4038),
            (0, 44),
            (0, 2597),
            (0, 0),
            (0, 5),
            (0, 21731),
            (0, 5),
            (0, 5052),
            (0, 570),
            (0, 6),
            (0, 4),
            (0, 173),
            (0, 92),
            (0, 1),
            (0, 5),
            (0, 78),
            (0, 11),
            (0, 82),
            (0, 131),
            (0, 3619),
            (0, 392),
            (0, 249),
            (0, 85),
            (0, 1384),
            (0, 42),
            (0, 2988),
            (0, 0),
            (0, 1),
            (0, 17055),
            (0, 7),
            (0, 0),
            (0, 175),
            (0, 11),
            (0, 4799),
            (0, 13076),
            (0, 14),
            (0, 6897),
            (0, 8216),
            (0, 12),
            (0, 26),
            (0, 60),
            (0, 375),
            (0, 169),
            (0, 51),
            (0, 1969),
            (0, 1995),
            (0, 3),
            (0, 351),
            (0, 312),
            (0, 237),
            (0, 745),
            (0, 668),
            (0, 47),
            (0, 473),
            (0, 3027),
            (0, 0),
            (0, 300),
            (0, 995),
            (0, 1),
            (0, 25774),
            (0, 85),
            (0, 767),
            (0, 18799),
            (0, 2),
            (0, 18),
            (0, 4156),
            (0, 22),
            (0, 1537),
            (0, 345),
            (0, 313),
            (0, 373),
            (0, 4),
            (0, 90),
            (0, 1192),
            (0, 5),
            (0, 549),
            (0, 16),
            (0, 130),
            (0, 30),
            (0, 14824),
            (0, 616),
            (0, 72),
            (0, 5840),
            (0, 0),
            (0, 74),
            (0, 3),
            (0, 21),
            (0, 0),
            (0, 828),
            (0, 0),
            (0, 4),
            (0, 65),
            (0, 1763),
            (0, 9),
            (0, 173),
            (0, 55),
            (0, 6630),
            (0, 94),
            (0, 6),
            (0, 2),
            (0, 6),
            (0, 511),
            (0, 5819),
            (0, 2607),
            (0, 54),
            (0, 545),
            (0, 677),
            (0, 2291),
            (0, 83),
            (0, 1),
            (0, 38),
            (0, 330),
            (0, 1949),
            (0, 83),
            (0, 561),
            (0, 396),
            (0, 135),
            (0, 21312),
            (0, 2),
            (0, 0),
            (0, 1471),
            (0, 0),
            (0, 125),
            (0, 89),
            (0, 2019),
            (0, 18986),
            (0, 13494),
            (0, 12353),
            (0, 14864),
            (0, 6),
            (0, 2),
            (0, 5454),
            (0, 2968),
            (0, 313),
            (0, 13975),
            (0, 1),
            (0, 12),
            (0, 6614),
            (0, 87),
            (0, 891),
            (0, 234),
            (0, 65),
            (0, 34),
            (0, 31),
            (0, 9171),
            (0, 12),
            (0, 577),
            (0, 1408),
            (0, 3466),
            (0, 3589),
            (0, 225),
            (0, 22),
            (0, 5),
            (0, 3944),
            (0, 129),
            (0, 170),
            (0, 0),
            (0, 224),
            (0, 2018),
            (0, 0),
            (0, 866),
            (0, 116),
            (0, 7661),
            (0, 25227),
            (0, 2610),
            (0, 4),
            (0, 3063),
            (0, 15836),
            (0, 51),
            (0, 479),
            (0, 29),
            (0, 702),
            (0, 19),
            (0, 6552),
            (0, 3),
            (0, 241),
            (0, 354),
            (0, 22),
            (0, 610),
            (0, 4),
            (0, 367),
            (0, 4360),
            (0, 1478),
            (0, 5),
            (0, 63),
            (0, 8226),
            (0, 3192),
            (0, 0),
            (0, 5060),
            (0, 546),
            (0, 22249),
            (0, 6),
            (0, 2261),
            (0, 340),
            (0, 51),
            (0, 5234),
            (0, 4546),
            (0, 1),
            (0, 508),
            (0, 1804),
            (0, 83),
            (0, 61),
            (0, 13366),
            (0, 21309),
            (0, 181),
            (0, 226),
            (0, 7485),
            (0, 1103),
            (0, 486),
            (0, 1329),
            (0, 35),
            (0, 14155),
            (0, 10012),
            (0, 10),
            (0, 1477),
            (0, 1784),
            (0, 5824),
            (0, 21283),
            (0, 196),
            (0, 1),
            (0, 266),
            (0, 60),
            (0, 4572),
            (0, 7845),
            (0, 65),
            (0, 667),
            (0, 375),
            (0, 16),
            (0, 985),
            (0, 51),
            (0, 35),
            (0, 2012),
            (0, 188),
            (0, 571),
            (0, 16876),
            (0, 2594),
            (0, 2374),
            (0, 460),
            (0, 1343),
            (0, 4),
            (0, 55),
            (0, 293),
            (0, 2163),
            (0, 1292),
            (0, 3508),
            (0, 178),
            (0, 65),
            (0, 21),
            (0, 4888),
            (0, 84),
            (0, 909),
            (0, 2),
            (0, 26),
            (0, 14410),
            (0, 1221),
            (0, 12),
            (0, 8),
            (0, 146),
            (0, 12),
            (0, 26174),
            (0, 1764),
            (0, 7610),
            (0, 3),
            (0, 460),
            (0, 0),
            (0, 1886),
            (0, 3),
            (0, 8761),
            (0, 6137),
            (0, 4433),
            (0, 834),
            (0, 25),
            (0, 577),
            (0, 16167),
            (0, 21269),
            (0, 294),
            (0, 33),
            (0, 1389),
            (0, 39),
            (0, 475),
            (0, 4948),
            (0, 15849),
            (0, 128),
            (0, 4629),
            (0, 6),
            (0, 273),
            (0, 12),
            (0, 12434),
            (0, 107),
            (0, 137),
            (0, 5336),
            (0, 144),
            (0, 4054),
            (0, 80),
            (0, 36),
            (0, 0),
            (0, 512),
            (0, 298),
            (0, 2),
            (0, 131),
            (0, 16840),
            (0, 211),
            (0, 687),
            (0, 7),
            (0, 47),
            (0, 422),
            (0, 8096),
            (0, 1020),
            (0, 2314),
            (0, 381),
            (0, 943),
            (0, 22),
            (0, 22),
            (0, 1410),
            (0, 43),
            (0, 4413),
            (0, 6043),
            (0, 716),
            (0, 1697),
            (0, 23956),
            (0, 0),
            (0, 248),
            (0, 354),
            (0, 100),
            (0, 34),
            (0, 7159),
            (0, 692),
            (0, 883),
            (0, 1900),
            (0, 423),
            (0, 1),
            (0, 877),
            (0, 6954),
            (0, 36),
            (0, 28),
            (0, 2838),
            (0, 63),
            (0, 2),
            (0, 137),
            (0, 3563),
            (0, 111),
            (0, 5140),
            (0, 1640),
            (0, 1079),
            (0, 1434),
            (0, 37),
            (0, 464),
            (0, 7),
            (0, 4553),
            (0, 45),
            (0, 1969),
            (0, 818),
            (0, 159),
            (0, 11),
            (0, 0),
            (0, 1),
            (0, 163),
            (0, 6692),
            (0, 22),
            (0, 3433),
            (0, 611),
            (0, 183),
            (0, 1314),
            (0, 33348),
            (0, 120),
            (0, 1),
            (0, 8482),
            (0, 5),
            (0, 3981),
            (0, 6432),
            (0, 29),
            (0, 7011),
            (0, 503),
            (0, 0),
            (0, 0),
            (0, 48),
            (0, 4420),
            (0, 10),
            (0, 231),
            (0, 5817),
            (0, 10402),
            (0, 16),
            (0, 111),
            (0, 187),
            (0, 1320),
            (0, 1),
            (0, 27),
            (0, 438),
            (0, 51),
            (0, 74),
            (0, 86),
            (0, 665),
            (0, 3631),
            (0, 4843),
            (0, 10),
            (0, 3667),
            (0, 1),
            (0, 1467),
            (0, 1555),
            (0, 8188),
            (0, 20),
            (0, 1),
            (0, 6332),
            (0, 2375),
            (0, 9304),
            (0, 8),
            (0, 14644),
            (0, 161),
            (0, 11),
            (0, 252),
            (0, 5902),
            (0, 8),
            (0, 19617),
            (0, 2),
            (0, 1199),
            (0, 122),
            (0, 2),
            (0, 107),
            (0, 2),
            (0, 22),
            (0, 5644),
            (0, 5),
            (0, 859),
            (0, 2719),
            (0, 2696),
            (0, 5494),
            (0, 400),
            (0, 2),
            (0, 130),
            (0, 10),
            (0, 266),
            (0, 43),
            (0, 32),
            (0, 6431),
            (0, 716),
            (0, 2577),
            (0, 70),
            (0, 5288),
            (0, 985),
            (0, 150),
            (0, 63),
            (0, 1469),
            (0, 74),
            (0, 255),
            (0, 3),
            (0, 419),
            (0, 19),
            (0, 458),
            (0, 194),
            (0, 41),
            (0, 15556),
            (0, 219),
            (0, 2),
            (0, 378),
            (0, 2734),
            (0, 10),
            (0, 12551),
            (0, 981),
            (0, 56),
            (0, 1),
            (0, 0),
            (0, 4),
            (0, 625),
            (0, 1),
            (0, 1),
            (0, 5902),
            (0, 1419),
            (0, 7),
            (0, 1504),
            (0, 6482),
            (0, 2145),
            (0, 3639),
            (0, 9),
            (0, 1495),
            (0, 1),
            (0, 7),
            (0, 32),
            (0, 16),
            (0, 1458),
            (0, 688),
            (0, 203),
            (0, 0),
            (0, 89),
            (0, 8),
            (0, 6698),
            (0, 2869),
            (0, 186),
            (0, 301),
            (0, 323),
            (0, 231),
            (0, 55),
            (0, 0),
            (0, 4),
            (0, 5),
            (0, 0),
            (0, 1253),
            (0, 548),
            (0, 6),
            (0, 5),
            (0, 3300),
            (0, 59),
            (0, 118),
            (0, 45),
            (0, 82),
            (0, 8127),
            (0, 67),
            (0, 4351),
            (0, 28),
            (0, 943),
            (0, 61),
            (0, 287),
            (0, 4790),
            (0, 1),
            (0, 4694),
            (0, 22),
            (0, 2003),
            (0, 3),
            (0, 558),
            (0, 3),
            (0, 1465),
            (0, 0),
            (0, 11804),
            (0, 256),
            (0, 13),
            (0, 0),
            (0, 3645),
            (0, 2574),
            (0, 95),
            (0, 3122),
            (0, 60),
            (0, 178),
            (0, 1557),
            (0, 3225),
            (0, 761),
            (0, 14),
            (0, 1),
            (0, 6),
            (0, 2109),
            (0, 1938),
            (0, 773),
            (0, 660),
            (0, 11346),
            (0, 14),
            (0, 61),
            (0, 710),
            (0, 3083),
            (0, 829),
            (0, 161),
            (0, 411),
            (0, 2767),
            (0, 108),
            (0, 33381),
            (0, 687),
            (0, 7587),
            (0, 541),
            (0, 59),
            (0, 196),
            (0, 187),
            (0, 17997),
            (0, 13),
            (0, 0),
            (0, 6491),
            (0, 7155),
            (0, 21),
            (0, 6976),
            (0, 12),
            (0, 202),
            (0, 207),
            (0, 75),
            (0, 274),
            (0, 4),
            (0, 90),
            (0, 7505),
            (0, 0),
            (0, 249),
            (0, 21794),
            (0, 116),
            (0, 16463),
            (0, 8442),
            (0, 3),
            (0, 51),
            (0, 7000),
            (0, 353),
            (0, 83),
            (0, 311),
            (0, 13582),
            (0, 7129),
            (0, 951),
            (0, 0),
            (0, 241),
            (0, 1256),
            (0, 579),
            (0, 3412),
            (0, 230),
            (0, 1571),
            (0, 8065),
            (0, 9727),
            (0, 0),
            (0, 1870),
            (0, 12),
            (0, 2541),
            (0, 4),
            (0, 20991),
            (0, 3929),
            (0, 1),
            (0, 2208),
            (0, 9448),
            (0, 19424),
            (0, 205),
            (0, 15822),
            (0, 2),
            (0, 36),
            (0, 0),
            (0, 2),
            (0, 821),
            (0, 4),
            (0, 3767),
            (0, 18109),
            (0, 4130),
            (0, 732),
            (0, 4360),
            (0, 17335),
            (0, 314),
            (0, 25),
            (0, 371),
            (0, 26544),
            (0, 13),
            (0, 1418),
            (0, 1953),
            (0, 0),
            (0, 5553),
            (0, 0),
            (0, 1),
            (0, 362),
            (0, 84),
            (0, 14),
            (0, 8),
            (0, 51),
            (0, 215),
            (0, 9717),
            (0, 889),
            (0, 9),
            (0, 48),
            (0, 1257),
            (0, 452),
            (0, 9139),
            (0, 121),
            (0, 8),
            (0, 1708),
            (0, 6578),
            (0, 6384),
            (0, 20),
            (0, 270),
            (0, 21606),
            (0, 1243),
            (0, 686),
            (0, 2107),
            (0, 196),
            (0, 0),
            (0, 418),
            (0, 2047),
            (0, 11416),
            (0, 76),
            (0, 10),
            (0, 3876),
            (0, 950),
            (0, 10441),
            (0, 107),
            (0, 65),
            (0, 2),
            (0, 8),
            (0, 16616),
            (0, 2),
            (0, 1),
            (0, 1997),
            (0, 1605),
            (0, 659),
            (0, 1),
            (0, 3),
            (0, 3986),
            (0, 11),
            (0, 32755),
            (0, 4513),
            (0, 31),
            (0, 2),
            (0, 9),
            (0, 17993),
            (0, 6932),
            (0, 8924),
            (0, 1044),
            (0, 0),
            (0, 14638),
            (0, 0),
            (0, 2242),
            (0, 35718),
            (0, 1),
            (0, 25),
            (0, 1651),
            (0, 8),
            (0, 16),
            (0, 79),
            (0, 13787),
            (0, 214),
            (0, 7254),
            (0, 1296),
            (0, 4616),
            (0, 295),
            (0, 134),
            (0, 0),
            (0, 16),
            (0, 1960),
            (0, 8840),
            (0, 15),
            (0, 2308),
            (0, 16789),
            (0, 21613),
            (0, 257),
            (0, 581),
            (0, 250),
            (0, 18),
            (0, 8),
            (0, 414),
            (0, 32),
            (0, 4846),
            (0, 10),
            (0, 13762),
            (0, 19318),
            (0, 275),
            (0, 1738),
            (0, 461),
            (0, 155),
            (0, 166),
            (0, 10440),
            (0, 153),
            (0, 1003),
            (0, 602),
            (0, 31349),
            (0, 1877),
            (0, 1181),
            (0, 9090),
            (0, 22260),
            (0, 82),
            (0, 2578),
            (0, 76),
            (0, 7),
            (0, 81),
            (0, 10),
            (0, 2019),
            (0, 104),
            (0, 280),
            (0, 133),
            (0, 1715),
            (0, 1527),
            (0, 1027),
            (0, 236),
            (0, 3),
            (0, 47),
            (0, 365),
            (0, 4716),
            (0, 2011),
            (0, 204),
            (0, 1008),
            (0, 6652),
            (0, 17387),
            (0, 21448),
            (0, 3),
            (0, 2469),
            (0, 347),
            (0, 2259),
            (0, 4310),
            (0, 876),
            (0, 1550),
            (0, 4),
            (0, 29668),
            (0, 10756),
            (0, 21),
            (0, 125),
            (0, 1),
            (0, 437),
            (0, 7000),
            (0, 3),
            (0, 10941),
            (0, 3022),
            (0, 16639),
            (0, 2034),
            (0, 0),
            (0, 162),
            (0, 420),
            (0, 8324),
            (0, 77),
            (0, 22),
            (0, 64),
            (0, 1099),
            (0, 0),
            (0, 0),
            (0, 3580),
            (0, 1319),
            (0, 1399),
            (0, 706),
            (0, 1),
            (0, 34),
            (0, 7),
            (0, 15),
            (0, 5711),
            (0, 374),
            (0, 1146),
            (0, 22919),
            (0, 1093),
            (0, 4),
            (0, 0),
            (0, 640),
            (0, 29015),
            (0, 450),
            (0, 13961),
            (0, 491),
            (0, 11030),
            (0, 799),
            (0, 279),
            (0, 252),
            (0, 678),
            (0, 3621),
            (0, 1846),
            (0, 3479),
            (0, 5625),
            (0, 2478),
            (0, 217),
            (0, 107),
            (0, 7),
            (0, 9681),
            (0, 8976),
            (0, 0),
            (0, 52),
            (0, 1191),
            (0, 69),
            (0, 43),
            (0, 37),
            (0, 44964),
            (0, 15),
            (0, 39),
            (0, 82),
            (0, 3),
            (0, 7),
            (0, 13),
            (0, 543),
            (0, 1562),
            (0, 345),
            (0, 93),
            (0, 175),
            (0, 78),
            (0, 1250),
            (0, 106),
            (0, 281),
            (0, 0),
            (0, 9910),
            (0, 379),
            (0, 19208),
            (0, 764),
            (0, 3),
            (0, 9),
            (0, 0),
            (0, 7515),
            (0, 19),
            (0, 4991),
            (0, 107),
            (0, 20),
            (0, 0),
            (0, 10795),
            (0, 6185),
            (0, 1395),
            (0, 0),
            (0, 4223),
            (0, 13731),
            (0, 408),
            (0, 9198),
            (0, 2),
            (0, 6485),
            (0, 4937),
            (0, 2136),
            (0, 1367),
            (0, 5612),
            (0, 463),
            (0, 16964),
            (0, 80),
            (0, 42),
            (0, 58),
            (0, 9),
            (0, 73),
            (0, 4892),
            (0, 5),
            (0, 102),
            (0, 6),
            (0, 2711),
            (0, 1629),
            (0, 3466),
            (0, 3262),
            (0, 66),
            (0, 850),
            (0, 35),
            (0, 34),
            (0, 13),
            (0, 74),
            (0, 968),
            (0, 16),
            (0, 981),
            (0, 231),
            (0, 23),
            (0, 3353),
            (0, 379),
            (0, 7351),
            (0, 361),
            (0, 4667),
            (0, 918),
            (0, 6),
            (0, 6167),
            (0, 0),
            (0, 2474),
            (0, 134),
            (0, 9615),
            (0, 941),
            (0, 1654),
            (0, 35),
            (0, 4),
            (0, 1267),
            (0, 11099),
            (0, 8806),
            (0, 56),
            (0, 7),
            (0, 4135),
            (0, 0),
            (0, 7),
            (0, 3130),
            (0, 489),
            (0, 20),
            (0, 573),
            (0, 89),
            (0, 48),
            (0, 3),
            (0, 226),
            (0, 24684),
            (0, 1864),
            (0, 308),
            (0, 0),
            (0, 862),
            (0, 156),
            (0, 6603),
            (0, 419),
            (0, 26485),
            (0, 4475),
            (0, 49519),
            (0, 1706),
            (0, 599),
            (0, 832),
            (0, 5),
            (0, 9742),
            (0, 1789),
            (0, 2271),
            (0, 857),
            (0, 6176),
            (0, 830),
            (0, 81),
            (0, 10),
            (0, 12506),
            (0, 2667),
            (0, 1),
            (0, 5),
            (0, 2),
            (0, 4117),
            (0, 2),
            (0, 5064),
            (0, 0),
            (0, 0),
            (0, 326),
            (0, 10460),
            (0, 3053),
            (0, 159),
            (0, 1),
            (0, 2118),
            (0, 1639),
            (0, 210),
            (0, 379),
            (0, 5),
            (0, 350),
            (0, 172),
            (0, 144),
            (0, 343),
            (0, 756),
            (0, 55),
            (0, 771),
            (0, 29),
            (0, 513),
            (0, 183),
            (0, 274),
            (0, 15),
            (0, 3519),
            (0, 2),
            (0, 46),
            (0, 583),
            (0, 0),
            (0, 4911),
            (0, 3451),
            (0, 99),
            (0, 86),
            (0, 87),
            (0, 1627),
            (0, 12),
            (0, 22),
            (0, 17),
            (0, 3654),
            (0, 6004),
            (0, 442),
            (0, 193),
            (0, 2),
            (0, 5313),
            (0, 700),
            (0, 74),
            (0, 22),
            (0, 4596),
            (0, 4365),
            (0, 67),
            (0, 356),
            (0, 2211),
            (0, 210),
            (0, 259),
            (0, 2213),
            (0, 8490),
            (0, 109),
            (0, 115),
            (0, 9840),
            (0, 432),
            (0, 97),
            (0, 201),
            (0, 0),
            (0, 0),
            (0, 1774),
            (0, 41),
            (0, 11),
            (0, 2),
            (0, 44),
            (0, 0),
            (0, 22684),
            (0, 118),
            (0, 2074),
            (0, 336),
            (0, 8241),
            (0, 189),
            (0, 139),
            (0, 0),
            (0, 0),
            (0, 22),
            (0, 2),
            (0, 843),
            (0, 0),
            (0, 3296),
            (0, 3495),
            (0, 6),
            (0, 2209),
            (0, 8629),
            (0, 375),
            (0, 39),
            (0, 12),
            (0, 5275),
            (0, 105),
            (0, 1119),
            (0, 4558),
            (0, 1005),
            (0, 64),
            (0, 66),
            (0, 103),
            (0, 37),
            (0, 9279),
            (0, 17),
            (0, 6419),
            (0, 2192),
            (0, 14),
            (0, 4280),
            (0, 0),
            (0, 2530),
            (0, 841),
            (0, 4665),
            (0, 17),
            (0, 3091),
            (0, 33505),
            (0, 39),
            (0, 8085),
            (0, 4333),
            (0, 108),
            (0, 27),
            (0, 1293),
            (0, 8155),
            (0, 0),
            (0, 239),
            (0, 222),
            (0, 1331),
            (0, 9983),
            (0, 5523),
            (0, 122),
            (0, 1974),
            (0, 8040),
            (0, 2),
            (0, 39),
            (0, 712),
            (0, 38),
            (0, 2019),
            (0, 4454),
            (0, 2049),
            (0, 4),
            (0, 5352),
            (0, 0),
            (0, 784),
            (0, 15080),
            (0, 11863),
            (0, 8678),
            (0, 273),
            (0, 4880),
            (0, 27),
            (0, 8),
            (0, 0),
            (0, 4629),
            (0, 28),
            (0, 4279),
            (0, 9503),
            (0, 236),
            (0, 1317),
            (0, 1278),
            (0, 1746),
            (0, 21429),
            (0, 2170),
            (0, 1),
            (0, 675),
            (0, 221),
            (0, 1),
            (0, 3086),
            (0, 78),
            (0, 22218),
            (0, 184),
            (0, 5),
            (0, 13541),
            (0, 10),
            (0, 88),
            (0, 1245),
            (0, 1026),
            (0, 42),
            (0, 1105),
            (0, 0),
            (0, 103),
            (0, 97),
            (0, 16533),
            (0, 11959),
            (0, 562),
            (0, 3),
            (0, 884),
            (0, 2520),
            (0, 63),
            (0, 23),
            (0, 404),
            (0, 1929),
            (0, 25),
            (0, 1446),
            (0, 601),
            (0, 400),
            (0, 6823),
            (0, 10014),
            (0, 1468),
            (0, 46),
            (0, 10531),
            (0, 4068),
            (0, 229),
            (0, 68),
            (0, 63),
            (0, 664),
            (0, 8),
            (0, 0),
            (0, 1701),
            (0, 4395),
            (0, 7442),
            (0, 4),
            (0, 307),
            (0, 4),
            (0, 65),
            (0, 329),
            (0, 0),
            (0, 2),
            (0, 1057),
            (0, 4467),
            (0, 45),
            (0, 503),
            (0, 25829),
            (0, 1115),
            (0, 8686),
            (0, 111),
            (0, 2593),
            (0, 1),
            (0, 5397),
            (0, 2690),
            (0, 1458),
            (0, 13184),
            (0, 10666),
            (0, 252),
            (0, 421),
            (0, 2170),
            (0, 166),
            (0, 21),
            (0, 131),
            (0, 7),
            (0, 6492),
            (0, 1253),
            (0, 14),
            (0, 6173),
            (0, 17),
            (0, 8),
            (0, 983),
            (0, 1902),
            (0, 216),
            (0, 4312),
            (0, 80),
            (0, 816),
            (0, 9708),
            (0, 150),
            (0, 3478),
            (0, 4972),
            (0, 6856),
            (0, 279),
            (0, 17),
            (0, 10),
            (0, 66),
            (0, 115),
            (0, 5),
            (0, 7),
            (0, 489),
            (0, 5569),
            (0, 0),
            (0, 11),
            (0, 5160),
            (0, 6964),
            (0, 2976),
            (0, 28879),
            (0, 2),
            (0, 82),
            (0, 9391),
            (0, 5),
            (0, 0),
            (0, 0),
            (0, 2283),
            (0, 998),
            (0, 12),
            (0, 2),
            (0, 0),
            (0, 150),
            (0, 764),
            (0, 7925),
            (0, 16),
            (0, 560),
            (0, 297),
            (0, 1847),
            (0, 10522),
            (0, 183),
            (0, 0),
            (0, 331),
            (0, 51006),
            (0, 4427),
            (0, 2063),
            (0, 24073),
            (0, 0),
            (0, 4553),
            (0, 1397),
            (0, 463),
            (0, 221),
            (0, 0),
            (0, 40),
            (0, 88),
            (0, 5),
            (0, 12604),
            (0, 4756),
            (0, 1),
            (0, 675),
            (0, 145),
            (0, 168),
            (0, 18),
            (0, 1568),
            (0, 12),
            (0, 2208),
            (0, 920),
            (0, 28565),
            (0, 2420),
            (0, 1357),
            (0, 2148),
            (0, 60),
            (0, 2),
            (0, 4082),
            (0, 689),
            (0, 432),
            (0, 2),
            (0, 63),
            (0, 6),
            (0, 36),
            (0, 106),
            (0, 1421),
            (0, 7051),
            (0, 7661),
            (0, 4262),
            (0, 365),
            (0, 93),
            (0, 1),
            (0, 4239),
            (0, 33),
            (0, 51),
            (0, 3),
            (0, 15918),
            (0, 43),
            (0, 7744),
            (0, 8),
            (0, 1857),
            (0, 41),
            (0, 2900),
            (0, 4007),
            (0, 12),
            (0, 1),
            (0, 28),
            (0, 11508),
            (0, 119),
            (0, 7602),
            (0, 16565),
            (0, 6982),
            (0, 0),
            (0, 9813),
            (0, 36),
            (0, 5),
            (0, 21419),
            (0, 20),
            (0, 23414),
            (0, 2768),
            (0, 1944),
            (0, 1629),
            (0, 5064),
            (0, 326),
            (0, 13981),
            (0, 13499),
            (0, 11792),
            (0, 6424),
            (0, 811),
            (0, 3),
            (0, 199),
            (0, 834),
            (0, 28456),
            (0, 2),
            (0, 4),
            (0, 1552),
            (0, 7052),
            (0, 296),
            (0, 1276),
            (0, 265),
            (0, 10),
            (0, 3494),
            (0, 3),
            (0, 0),
            (0, 165),
            (0, 8),
            (0, 168),
            (0, 8191),
            (0, 60),
            (0, 8955),
            (0, 224),
            (0, 5106),
            (0, 0),
            (0, 3),
            (0, 19),
            (0, 2455),
            (0, 932),
            (0, 217),
            (0, 21),
            (0, 2),
            (0, 2),
            (0, 32),
            (0, 99),
            (0, 20),
            (0, 15),
            (0, 90),
            (0, 1226),
            (0, 443),
            (0, 2639),
            (0, 974),
            (0, 53),
            (0, 23943),
            (0, 34),
            (0, 186),
            (0, 1),
            (0, 80),
            (0, 127),
            (0, 24639),
            (0, 2353),
            (0, 284),
            (0, 5119),
            (0, 1767),
            (0, 0),
            (0, 6050),
            (0, 1564),
            (0, 1827),
            (0, 489),
            (0, 254),
            (0, 36663),
            (0, 38501),
            (0, 3),
            (0, 331),
            (0, 959),
            (0, 9),
            (0, 2181),
            (0, 0),
            (0, 3365),
            (0, 744),
            (0, 7929),
            (0, 6),
            (0, 1874),
            (0, 4),
            (0, 410),
            (0, 874),
            (0, 2949),
            (0, 15371),
            (0, 5),
            (0, 6655),
            (0, 15),
            (0, 8),
            (0, 5),
            (0, 48),
            (0, 3833),
            (0, 11551),
            (0, 1392),
            (0, 3370),
            (0, 323),
            (0, 1),
            (0, 732),
            (0, 16694),
            (0, 3532),
            (0, 125),
            (0, 63),
            (0, 7168),
            (0, 1173),
            (0, 13176),
            (0, 7491),
            (0, 90),
            (0, 890),
            (0, 40),
            (0, 4608),
            (0, 8848),
            (0, 22781),
            (0, 35),
            (0, 0),
            (0, 10),
            (0, 908),
            (0, 16),
            (0, 0),
            (0, 126),
            (0, 80),
            (0, 9162),
            (0, 35),
            (0, 41),
            (0, 0),
            (0, 2939),
            (0, 11409),
            (0, 6760),
            (0, 143),
            (0, 21443),
            (0, 2491),
            (0, 206),
            (0, 1619),
            (0, 15722),
            (0, 2278),
            (0, 2039),
            (0, 38),
            (0, 663),
            (0, 1123),
            (0, 1889),
            (0, 12),
            (0, 4),
            (0, 22),
            (0, 4285),
            (0, 224),
            (0, 11839),
            (0, 1),
            (0, 2),
            (0, 39),
            (0, 2),
            (0, 168),
            (0, 219),
            (0, 2),
            (0, 411),
            (0, 639),
            (0, 1687),
            (0, 14),
            (0, 17),
            (0, 25),
            (0, 22),
            (0, 5659),
            (0, 9),
            (0, 8),
            (0, 0),
            (0, 19),
            (0, 410),
            (0, 5046),
            (0, 524),
            (0, 2037),
            (0, 7956),
            (0, 58),
            (0, 185),
            (0, 1685),
            (0, 1480),
            (0, 116),
            (0, 84),
            (0, 2412),
            (0, 168),
            (0, 8202),
            (0, 11084),
            (0, 1791),
            (0, 4922),
            (0, 75),
            (0, 960),
            (0, 166),
            (0, 1),
            (0, 0),
            (0, 32),
            (0, 55),
            (0, 20456),
            (0, 130),
            (0, 1498),
            (0, 221),
            (0, 30749),
            (0, 62606),
            (0, 1472),
            (0, 1213),
            (0, 2377),
            (0, 16),
            (0, 1876),
            (0, 8312),
            (0, 316),
            (0, 41),
            (1, 1049),
            (0, 179),
            (0, 7744),
            (0, 3548),
            (0, 20910),
            (0, 45),
            (0, 519),
            (0, 2632),
            (0, 41),
            (0, 6114),
            (0, 0),
            (0, 2038),
            (0, 19),
            (0, 6024),
            (0, 2411),
            (0, 486),
            (0, 10532),
            (0, 148),
            (0, 2889),
            (0, 122),
            (0, 6921),
            (0, 2484),
            (0, 193),
            (0, 31412),
            (0, 5),
            (0, 15),
            (0, 7),
            (0, 0),
            (0, 2335),
            (0, 174),
            (0, 514),
            (0, 41),
            (0, 310),
            (0, 15843),
            (0, 33),
            (0, 3),
            (0, 10),
            (0, 5122),
            (0, 680),
            (0, 705),
            (0, 68611),
            (0, 0),
            (0, 10),
            (0, 3773),
            (0, 7904),
            (0, 3),
            (0, 18),
            (0, 5088),
            (0, 9792),
            (0, 374),
            (0, 763),
            (0, 1),
            (0, 19),
            (0, 5),
            (0, 232),
            (0, 1),
            (0, 1202),
            (0, 32962),
            (0, 274),
            (0, 119),
            (0, 3969),
            (0, 1),
            (0, 402),
            (0, 23),
            (0, 264),
            (0, 827),
            (0, 44969),
            (0, 29816),
            (0, 2096),
            (0, 11),
            (0, 296),
            (0, 2),
            (0, 0),
            (0, 79),
            (0, 32843),
            (0, 7),
            (0, 8),
            (0, 8500),
            (0, 4786),
            (0, 344),
            (0, 14),
            (0, 69),
            (0, 123),
            (0, 4868),
            (0, 101),
            (0, 813),
            (0, 0),
            (0, 16),
            (0, 1102),
            (0, 343),
            (0, 3355),
            (0, 8),
            (0, 3519),
            (0, 534),
            (0, 3539),
            (0, 7638),
            (0, 24031),
            (0, 784),
            (0, 0),
            (0, 10905),
            (0, 99),
            (0, 3),
            (0, 547),
            (0, 20),
            (0, 49),
            (0, 778),
            (0, 9413),
            (0, 103),
            (0, 621),
            (0, 0),
            (0, 14392),
            (0, 3),
            (0, 1),
            (0, 50),
            (0, 1098),
            (0, 354),
            (0, 5),
            (0, 5),
            (0, 13355),
            (0, 89),
            (0, 2),
            (0, 388),
            (0, 7766),
            (0, 90),
            (0, 175),
            (0, 1583),
            (0, 736),
            (0, 7561),
            (0, 17859),
            (0, 32436),
            (0, 10),
            (0, 457),
            (0, 413),
            (0, 111),
            (0, 110),
            (0, 43),
            (0, 105),
            (0, 88),
            (0, 3641),
            (0, 2724),
            (0, 20749),
            (0, 2124),
            (0, 263),
            (0, 1),
            (0, 3),
            (0, 12),
            (0, 11326),
            (0, 41),
            (0, 0),
            (0, 219),
            (0, 174),
            (0, 3),
            (0, 1907),
            (0, 7578),
            (0, 0),
            (0, 3100),
            (0, 35),
            (0, 314),
            (0, 8107),
            (0, 583),
            (0, 2492),
            (0, 906),
            (0, 18),
            (0, 39316),
            (0, 225),
            (0, 280),
            (0, 0),
            (0, 36447),
            (0, 26559),
            (0, 10),
            (0, 76),
            (0, 176),
            (0, 3),
            (0, 15384),
            (0, 18043),
            (0, 24),
            (0, 8),
            (0, 919),
            (0, 0),
            (0, 7755),
            (0, 329),
            (0, 94),
            (0, 2575),
            (0, 73),
            (0, 0),
            (0, 308),
            (0, 682),
            (0, 4000),
            (0, 272),
            (0, 2768),
            (0, 848),
            (0, 41),
            (0, 12748),
            (0, 3191),
            (0, 12),
            (0, 77),
            (0, 12201),
            (0, 9),
            (0, 30),
            (0, 1),
            (0, 326),
            (0, 0),
            (0, 208),
            (0, 15),
            (0, 167),
            (0, 7),
            (0, 73),
            (0, 955),
            (0, 1596),
            (0, 639),
            (0, 0),
            (0, 163),
            (0, 4149),
            (0, 20011),
            (0, 16),
            (0, 57484),
            (0, 23409),
            (0, 97),
            (0, 5872),
            (0, 9),
            (0, 26),
            (0, 6),
            (0, 1562),
            (0, 9836),
            (0, 242),
            (0, 270),
            (0, 4),
            (0, 8942),
            (0, 1087),
            (0, 1),
            (0, 39109),
            (0, 803),
            (0, 95),
            (0, 2415),
            (0, 66),
            (0, 59),
            (0, 1322),
            (0, 4118),
            (0, 1),
            (0, 2),
            (0, 0),
            (0, 0),
            (0, 147),
            (0, 35110),
            (0, 63),
            (0, 193),
            (0, 18),
            (0, 18),
            (0, 6),
            (0, 938),
            (0, 49),
            (0, 37),
            (0, 2813),
            (0, 9350),
            (0, 1),
            (0, 1024),
            (0, 11542),
            (0, 7826),
            (0, 325),
            (0, 9810),
            (0, 394),
            (0, 438),
            (0, 0),
            (0, 0),
            (0, 8),
            (0, 21),
            (0, 11),
            (0, 3067),
            (0, 717),
            (0, 8718),
            (0, 3),
            (0, 73),
            (0, 2231),
            (0, 483),
            (0, 2266),
            (0, 2080),
            (0, 13192),
            (0, 50),
            (0, 41267),
            (0, 5934),
            (0, 11404),
            (0, 2925),
            (0, 34),
            (0, 16),
            (0, 7),
            (0, 191),
            (0, 318),
            (0, 155),
            (0, 3277),
            (0, 67),
            (0, 270),
            (0, 2227),
            (0, 5269),
            (0, 24617),
            (0, 0),
            (0, 0),
            (0, 4116),
            (0, 33122),
            (0, 261),
            (0, 50),
            (0, 764),
            (0, 3005),
            (0, 0),
            (0, 234),
            (0, 84),
            (0, 122),
            (0, 7),
            (0, 31),
            (0, 31),
            (0, 7510),
            (0, 4663),
            (0, 1545),
            (0, 22486),
            (0, 3751),
            (0, 531),
            (0, 5493),
            (0, 4),
            (0, 7099),
            (0, 4747),
            (0, 7),
            (0, 388),
            (0, 1119),
            (0, 70),
            (0, 31),
            (0, 83),
            (0, 60),
            (0, 489),
            (0, 26018),
            (0, 12),
            (0, 5635),
            (0, 1408),
            (0, 1),
            (0, 5),
            (0, 1640),
            (0, 60),
            (0, 1252),
            (0, 3),
            (0, 904),
            (0, 1519),
            (0, 176),
            (0, 141),
            (0, 2801),
            (0, 4062),
            (0, 256),
            (0, 23),
            (0, 25835),
            (0, 21861),
            (0, 23),
            (0, 1844),
            (0, 1841),
            (0, 87),
            (0, 986),
            (0, 17676),
            (0, 28),
            (0, 24550),
            (0, 10540),
            (0, 0),
            (0, 232),
            (0, 849),
            (0, 26837),
            (0, 4796),
            (0, 1530),
            (0, 17567),
            (0, 62171),
            (0, 167),
            (0, 5490),
            (0, 22932),
            (0, 6661),
            (0, 40),
            (0, 919),
            (0, 2),
            (0, 146),
            (0, 10),
            (0, 3621),
            (0, 2),
            (0, 26),
            (0, 592),
            (0, 1844),
            (0, 1),
            (0, 93),
            (0, 11),
            (0, 44),
            (0, 1196),
            (0, 60),
            (0, 232),
            (0, 319),
            (0, 16477),
            (0, 765),
            (0, 52),
            (0, 5370),
            (0, 5847),
            (0, 22692),
            (0, 399),
            (0, 108),
            (0, 22),
            (0, 899),
            (0, 429),
            (0, 72),
            (0, 17),
            (0, 91),
            (0, 124),
            (0, 0),
            (0, 8313),
            (0, 2),
            (0, 9537),
            (0, 0),
            (0, 145),
            (0, 1775),
            (0, 14790),
            (0, 46),
            (0, 5357),
            (0, 70),
            (0, 75094),
            (0, 10),
            (0, 5),
            (0, 48),
            (0, 85),
            (0, 7),
            (0, 15230),
            (0, 249),
            (0, 2),
            (0, 283),
            (0, 2435),
            (0, 9596),
            (0, 15),
            (0, 20),
            (0, 7684),
            (0, 11304),
            (0, 597),
            (0, 153),
            (0, 5198),
            (0, 13),
            (0, 1620),
            (0, 7320),
            (0, 78),
            (0, 28724),
            (0, 4907),
            (0, 2424),
            (0, 58),
            (0, 3372),
            (0, 0),
            (0, 4611),
            (0, 1),
            (0, 2351),
            (0, 15493),
            (0, 5113),
            (0, 7473),
            (0, 91),
            (0, 221),
            (0, 8),
            (0, 0),
            (0, 17),
            (0, 39),
            (1, 616),
            (0, 28),
            (0, 4652),
            (0, 31),
            (0, 4238),
            (0, 718),
            (0, 31727),
            (0, 89),
            (0, 10),
            (0, 44),
            (0, 6070),
            (0, 47),
            (0, 0),
            (0, 365),
            (0, 4303),
            (0, 7689),
            (0, 275),
            (0, 3994),
            (0, 3929),
            (0, 14523),
            (0, 11),
            (0, 6950),
            (0, 4),
            (0, 16),
            (0, 14956),
            (0, 72),
            (0, 1189),
            (0, 16),
            (0, 2789),
            (0, 8683),
            (0, 2612),
            (0, 164),
            (0, 32),
            (0, 2134),
            (0, 4887),
            (0, 180),
            (0, 0),
            (0, 90),
            (0, 34),
            (0, 12344),
            (0, 5287),
            (0, 9),
            (0, 5330),
            (0, 1),
            (0, 9),
            (0, 512),
            (0, 5991),
            (0, 0),
            (0, 2738),
            (0, 7464),
            (0, 30171),
            (0, 447),
            (0, 394),
            (0, 74),
            (0, 1420),
            (0, 43),
            (0, 16740),
            (0, 2200),
            (0, 530),
            (0, 4),
            (0, 8),
            (0, 1),
            (0, 3268),
            (0, 80625),
            (0, 17),
            (0, 22),
            (0, 3006),
            (0, 44),
            (0, 705),
            (0, 3412),
            (0, 106),
            (0, 1646),
            (0, 1),
            (0, 4),
            (0, 0),
            (0, 29694),
            (0, 15219),
            (0, 0),
            (0, 46575),
            (0, 31576),
            (0, 29),
            (0, 1626),
            (0, 6),
            (0, 20280),
            (0, 62014),
            (0, 46),
            (0, 14),
            (0, 15488),
            (0, 1417),
            (0, 0),
            (0, 2260),
            (0, 1271),
            (0, 52),
            (0, 73),
            (0, 1381),
            (0, 356),
            (0, 794),
            (0, 164),
            (0, 703),
            (0, 1984),
            (0, 60),
            (0, 18575),
            (0, 1),
            (0, 553),
            (0, 1749),
            (0, 774),
            (0, 1),
            (0, 9),
            (0, 68),
            (0, 399),
            (0, 318),
            (0, 12),
            (0, 291),
            (0, 298),
            (0, 11),
            (0, 7),
            (0, 35626),
            (0, 0),
            (0, 122),
            (0, 2587),
            (0, 11),
            (0, 4007),
            (0, 1564),
            (0, 9070),
            (0, 10),
            (0, 4203),
            (0, 1),
            (0, 18074),
            (0, 5),
            (0, 3666),
            (0, 10),
            (0, 0),
            (0, 1143),
            (0, 387),
            (0, 46),
            (0, 2346),
            (0, 4896),
            (0, 290),
            (0, 372),
            (0, 2691),
            (0, 147),
            (0, 466),
            (0, 2),
            (0, 202),
            (0, 11385),
            (0, 608),
            (0, 159),
            (0, 4200),
            (0, 28),
            (0, 7530),
            (0, 41831),
            (0, 6919),
            (0, 3895),
            (0, 2),
            (0, 782),
            (0, 7081),
            (0, 3792),
            (0, 8750),
            (0, 2508),
            (0, 2),
            (0, 10751),
            (0, 10202),
            (0, 96),
            (0, 23),
            (0, 67),
            (0, 6279),
            (0, 0),
            (0, 1),
            (0, 8417),
            (0, 110),
            (1, 7077),
            (0, 0),
            (0, 22),
            (0, 15472),
            (0, 1296),
            (0, 160),
            (0, 1),
            (0, 279),
            (0, 4894),
            (0, 1385),
            (0, 200),
            (0, 1359),
            (0, 5895),
            (0, 35),
            (0, 13287),
            (0, 268),
            (0, 616),
            (0, 2),
            (0, 46),
            (0, 37185),
            (0, 17),
            (0, 646),
            (0, 2492),
            (0, 687),
            (0, 46104),
            (0, 725),
            (0, 4508),
            (0, 523),
            (0, 20),
            (0, 2),
            (0, 3104),
            (0, 313),
            (0, 13450),
            (0, 845),
            (0, 16862),
            (0, 30),
            (0, 2078),
            (0, 3861),
            (0, 1),
            (0, 19229),
            (0, 20433),
            (0, 2719),
            (0, 62713),
            (0, 2064),
            (0, 25),
            (0, 482),
            (0, 0),
            (0, 144),
            (0, 4846),
            (0, 2644),
            (0, 2),
            (0, 1972),
            (0, 15718),
            (0, 1193),
            (0, 457),
            (0, 0),
            (0, 6665),
            (0, 390),
            (0, 12),
            (0, 10977),
            (0, 17),
            (0, 56156),
            (0, 154),
            (0, 154),
            (0, 12755),
            (0, 36140),
            (0, 10997),
            (0, 13873),
            (0, 3),
            (0, 1322),
            (0, 147),
            (0, 22),
            (0, 73912),
            (0, 1155),
            (0, 371),
            (0, 7663),
            (0, 3),
            (0, 230),
            (0, 2556),
            (0, 3893),
            (0, 18),
            (0, 51),
            (0, 44160),
            (0, 9),
            (0, 2729),
            (0, 4),
            (0, 33335),
            (0, 0),
            (0, 30),
            (0, 4647),
            (0, 55),
            (0, 587),
            (0, 51),
            (0, 803),
            (0, 155),
            (0, 4574),
            (0, 188),
            (0, 3),
            (0, 2974),
            (0, 18442),
            (0, 2463),
            (0, 0),
            (0, 1566),
            (0, 6322),
            (0, 250),
            (0, 25675),
            (0, 9),
            (0, 701),
            (0, 349),
            (0, 21912),
            (0, 33949),
            (0, 3),
            (0, 20448),
            (0, 61),
            (0, 115),
            (0, 15),
            (0, 17287),
            (0, 48),
            (0, 1203),
            (0, 30),
            (0, 9),
            (0, 110),
            (0, 36),
            (0, 98),
            (0, 2444),
            (0, 15254),
            (0, 4763),
            (0, 0),
            (0, 1565),
            (0, 361),
            (0, 6),
            (0, 8145),
            (0, 12),
            (0, 846),
            (0, 3027),
            (0, 25),
            (0, 22),
            (0, 3874),
            (0, 9179),
            (0, 51766),
            (0, 447),
            (0, 3210),
            (0, 4),
            (0, 31311),
            (0, 1),
            (0, 13),
            (0, 4),
            (0, 7526),
            (0, 24),
            (0, 9871),
            (0, 0),
            (0, 94),
            (0, 428),
            (0, 63),
            (0, 2),
            (0, 566),
            (0, 10811),
            (0, 282),
            (0, 173),
            (0, 10301),
            (0, 4648),
            (0, 41),
            (0, 999),
            (0, 7),
            (0, 69),
            (0, 107),
            (0, 798),
            (0, 34),
            (0, 1),
            (0, 88),
            (0, 5429),
            (0, 10),
            (0, 2),
            (0, 1231),
            (0, 0),
            (0, 3137),
            (0, 2),
            (0, 0),
            (0, 16193),
            (0, 879),
            (0, 17),
            (0, 10),
            (0, 6394),
            (0, 106),
            (0, 1091),
            (0, 1362),
            (0, 156),
            (0, 16),
            (0, 52187),
            (0, 47),
            (0, 188),
            (0, 1),
            (0, 205),
            (0, 482),
            (0, 1580),
            (0, 11),
            (0, 2),
            (0, 127),
            (0, 58),
            (0, 22824),
            (0, 5),
            (0, 243),
            (0, 10),
            (0, 33),
            (0, 0),
            (0, 50181),
            (0, 1008),
            (0, 57),
            (0, 9),
            (0, 4),
            (0, 2334),
            (0, 3),
            (0, 21),
            (0, 1872),
            (0, 12730),
            (0, 2),
            (0, 5),
            (0, 26180),
            (0, 7),
            (0, 1553),
            (0, 314),
            (0, 1089),
            (0, 14),
            (0, 489),
            (0, 6317),
            (0, 0),
            (0, 105),
            (0, 42),
            (0, 37),
            (0, 283),
            (0, 4),
            (0, 6199),
            (0, 660),
            (0, 5),
            (0, 77),
            (0, 4503),
            (0, 18533),
            (0, 71),
            (0, 75),
            (0, 34),
            (0, 160),
            (0, 2387),
            (0, 31),
            (0, 15544),
            (0, 527),
            (0, 5888),
            (0, 1133),
            (0, 0),
            (0, 2),
            (0, 53),
            (0, 13441),
            (0, 157),
            (0, 18893),
            (0, 230),
            (0, 35),
            (0, 4),
            (0, 325),
            (0, 864),
            (0, 297),
            (0, 924),
            (0, 25),
            (0, 14569),
            (0, 377),
            (0, 16394),
            (0, 18290),
            (0, 15),
            (0, 198),
            (0, 2043),
            (0, 12),
            (0, 3646),
            (0, 5),
            (0, 24),
            (0, 652),
            (0, 0),
            (0, 64),
            (0, 153),
            (0, 96),
            (0, 24276),
            (0, 227),
            (0, 0),
            (0, 16209),
            (0, 3359),
            (0, 8208),
            (0, 7276),
            (0, 473),
            (0, 11863),
            (0, 5819),
            (0, 2065),
            (0, 3),
            (0, 38594),
            (0, 0),
            (0, 23954),
            (0, 5194),
            (0, 2178),
            (0, 0),
            (0, 70),
            (0, 7693),
            (0, 3413),
            (0, 0),
            (0, 83),
            (0, 96),
            (0, 17640),
            (0, 725),
            (0, 204),
            (0, 1423),
            (0, 8884),
            (0, 2325),
            (0, 172),
            (0, 0),
            (0, 4087),
            (0, 96),
            (0, 35711),
            (0, 5),
            (0, 5023),
            (0, 57),
            (0, 22),
            (0, 5310),
            (0, 4),
            (0, 9164),
            (0, 9),
            (0, 657),
            (0, 1777),
            (0, 1),
            (0, 61),
            (0, 7186),
            (0, 14515),
            (0, 2),
            (0, 1213),
            (0, 4709),
            (0, 18),
            (0, 6642),
            (0, 6),
            (0, 7),
            (0, 15244),
            (0, 11048),
            (0, 153),
            (0, 25),
            (0, 11247),
            (0, 4),
            (0, 34),
            (0, 0),
            (0, 1262),
            (0, 10),
            (0, 9),
            (0, 2165),
            (0, 421),
            (0, 12052),
            (0, 204),
            (0, 4234),
            (0, 0),
            (0, 29662),
            (0, 87),
            (0, 61),
            (0, 5618),
            (0, 389),
            (0, 0),
            (0, 6906),
            (0, 247),
            (0, 84),
            (0, 1923),
            (0, 21),
            (0, 9218),
            (0, 801),
            (0, 1),
            (0, 3045),
            (0, 39),
            (0, 17),
            (0, 42747),
            (0, 166),
            (0, 35),
            (0, 16270),
            (0, 261),
            (0, 1181),
            (0, 3),
            (0, 177),
            (0, 37691),
            (0, 2),
            (0, 0),
            (0, 11),
            (0, 14),
            (0, 826),
            (0, 9030),
            (0, 963),
            (0, 8220),
            (0, 2289),
            (0, 11801),
            (0, 55),
            (0, 1859),
            (0, 79),
            (0, 130),
            (0, 5),
            (0, 23366),
            (0, 1),
            (0, 20),
            (0, 3),
            (0, 9070),
            (0, 57695),
            (0, 616),
            (0, 5017),
            (0, 9576),
            (0, 221),
            (0, 25),
            (0, 93),
            (0, 303),
            (0, 1838),
            (0, 1),
            (0, 2535),
            (0, 144),
            (0, 17),
            (0, 23902),
            (0, 56),
            (0, 10776),
            (0, 0),
            (0, 38),
            (0, 22),
            (0, 0),
            (0, 7),
            (0, 0),
            (0, 12069),
            (0, 4540),
            (0, 0),
            (0, 7),
            (0, 19),
            (0, 0),
            (0, 2102),
            (0, 242),
            (0, 32),
            (0, 67),
            (0, 12726),
            (0, 54992),
            (0, 114),
            (0, 10),
            (0, 162),
            (0, 8050),
            (0, 3696),
            (0, 27564),
            (0, 1),
            (0, 8),
            (0, 4),
            (0, 37),
            (0, 4608),
            (0, 283),
            (0, 1257),
            (0, 0),
            (0, 6785),
            (0, 69),
            (0, 1245),
            (0, 209),
            (0, 54),
            (0, 711),
            (0, 24),
            (0, 16422),
            (0, 232),
            (0, 20),
            (0, 443),
            (0, 2689),
            (0, 858),
            (0, 607),
            (0, 7126),
            (0, 168),
            (0, 5496),
            (0, 1502),
            (0, 6848),
            (0, 8552),
            (0, 4570),
            (0, 2639),
            (0, 157),
            (0, 10331),
            (0, 21312),
            (0, 1),
            (0, 43340),
            (0, 2),
            (0, 446),
            (0, 373),
            (0, 12),
            (0, 2),
            (0, 5),
            (0, 41973),
            (0, 1),
            (0, 124),
            (0, 412),
            (0, 425),
            (0, 4),
            (0, 6),
            (0, 249),
            (0, 431),
            (0, 3819),
            (0, 42),
            (0, 10),
            (0, 489),
            (0, 2169),
            (0, 1232),
            (0, 184),
            (0, 5886),
            (0, 37547),
            (0, 1),
            (0, 8977),
            (0, 3),
            (0, 115),
            (0, 7),
            (0, 2291),
            (0, 3559),
            (0, 440),
            (0, 9803),
            (0, 19),
            (0, 2202),
            (0, 1),
            (0, 803),
            (0, 21),
            (0, 54),
            (0, 191),
            (0, 7885),
            (0, 5),
            (0, 4601),
            (0, 1051),
            (0, 571),
            (0, 687),
            (0, 18897),
            (0, 41582),
            (0, 327),
            (0, 1806),
            (0, 98),
            (0, 385),
            (0, 1013),
            (0, 18),
            (0, 274),
            (0, 7918),
            (0, 0),
            (0, 1997),
            (0, 11688),
            (0, 3534),
            (0, 499),
            (0, 2),
            (0, 85),
            (0, 663),
            (0, 46),
            (0, 73),
            (0, 131),
            (0, 759),
            (0, 0),
            (0, 5),
            (0, 8),
            (0, 35),
            (0, 377),
            (0, 1502),
            (0, 11414),
            (0, 14),
            (0, 199),
            (0, 152),
            (0, 0),
            (0, 59),
            (0, 296),
            (0, 2),
            (0, 1067),
            (0, 3),
            (0, 13210),
            (0, 128),
            (0, 0),
            (0, 3826),
            (0, 1806),
            (0, 333),
            (0, 10374),
            (0, 672),
            (0, 1941),
            (0, 43594),
            (0, 18217),
            (0, 103),
            (0, 25657),
            (0, 9661),
            (0, 13),
            (0, 28),
            (0, 12),
            (0, 20),
            (0, 6965),
            (0, 451),
            (0, 17),
            (0, 1765),
            (0, 40),
            (0, 23),
            (0, 363),
            (0, 202),
            (0, 11402),
            (0, 49),
            (0, 6037),
            (0, 55),
            (0, 443),
            (0, 26),
            (0, 142),
            (0, 15845),
            (0, 464),
            (0, 44),
            (0, 9),
            (0, 2822),
            (0, 95),
            (0, 7),
            (0, 58030),
            (0, 3),
            (0, 1139),
            (0, 5601),
            (0, 508),
            (0, 8),
            (0, 52363),
            (0, 142),
            (0, 6822),
            (0, 5563),
            (0, 5),
            (0, 4958),
            (0, 103),
            (0, 531),
            (0, 101),
            (0, 2840),
            (0, 20705),
            (0, 6797),
            (0, 3032),
            (0, 2502),
            (0, 988),
            (0, 436),
            (0, 25293),
            (0, 5504),
            (0, 2510),
            (0, 29),
            (0, 3),
            (0, 5096),
            (0, 11266),
            (0, 11954),
            (0, 11),
            (0, 87),
            (0, 12145),
            (0, 1925),
            (0, 2434),
            (0, 2309),
            (0, 88),
            (0, 51),
            (0, 341),
            (0, 49),
            (0, 1499),
            (0, 1),
            (0, 2),
            (0, 263),
            (0, 85096),
            (0, 181),
            (0, 86198),
            (0, 32049),
            (0, 25),
            (0, 661),
            (0, 11112),
            (0, 0),
            (0, 84040),
            (0, 20051),
            (0, 6344),
            (0, 20),
            (0, 34160),
            (0, 41091),
            (0, 58),
            (0, 3984),
            (0, 303),
            (0, 2856),
            (0, 1428),
            (0, 42),
            (0, 27),
            (0, 7780),
            (0, 3),
            (0, 104),
            (0, 269),
            (0, 11),
            (0, 49),
            (0, 11318),
            (0, 1),
            (0, 7),
            (0, 92),
            (0, 1269),
            (0, 57),
            (0, 1114),
            (0, 2),
            (0, 30964),
            (0, 171),
            (0, 97),
            (0, 508),
            (0, 2),
            (0, 51),
            (0, 3047),
            (0, 215),
            (0, 867),
            (0, 0),
            (0, 2947),
            (0, 201),
            (0, 3702),
            (0, 2543),
            (0, 34950),
            (0, 708),
            (0, 230),
            (0, 4),
            (0, 8675),
            (0, 2099),
            (0, 166),
            (0, 27),
            (0, 9),
            (0, 169),
            (0, 3330),
            (0, 3),
            (0, 22882),
            (0, 37),
            (0, 176),
            (0, 3037),
            (0, 6127),
            (0, 4621),
            (0, 17),
            (0, 78348),
            (0, 25325),
            (0, 4317),
            (0, 7914),
            (0, 9562),
            (0, 7761),
            (0, 9),
            (0, 41),
            (0, 2982),
            (0, 2),
            (0, 318),
            (0, 210),
            (0, 3998),
            (0, 10886),
            (0, 71),
            (0, 246),
            (0, 95),
            (0, 212),
            (0, 3),
            (0, 7),
            (0, 25),
            (0, 72189),
            (0, 18320),
            (0, 139),
            (0, 0),
            (0, 12),
            (0, 7583),
            (0, 9114),
            (0, 4),
            (0, 433),
            (0, 54),
            (0, 363),
            (0, 1185),
            (0, 2539),
            (0, 790),
            (0, 273),
            (0, 56),
            (0, 3369),
            (0, 1),
            (0, 7),
            (0, 14055),
            (0, 27979),
            (0, 1),
            (0, 19),
            (0, 5),
            (0, 6496),
            (0, 5824),
            (0, 4),
            (0, 45),
            (0, 26),
            (0, 0),
            (0, 22635),
            (0, 6108),
            (0, 37705),
            (0, 10072),
            (0, 0),
            (0, 5459),
            (0, 52058),
            (0, 65),
            (0, 26),
            (0, 120),
            (0, 146),
            (0, 3),
            (0, 220),
            (0, 10613),
            (0, 2),
            (0, 25),
            (0, 992),
            (0, 2819),
            (0, 184),
            (0, 11962),
            (0, 32),
            (0, 42155),
            (0, 41),
            (0, 2879),
            (0, 10),
            (0, 521),
            (0, 0),
            (0, 30),
            (0, 2925),
            (0, 1339),
            (0, 1828),
            (0, 0),
            (0, 2),
            (0, 0),
            (0, 488),
            (0, 1311),
            (0, 872),
            (0, 37),
            (0, 4),
            (0, 1),
            (0, 2533),
            (0, 0),
            (0, 9),
            (0, 134),
            (0, 135),
            (0, 2358),
            (0, 549),
            (0, 3323),
            (0, 60),
            (0, 15903),
            (0, 20471),
            (0, 561),
            (0, 101),
            (0, 361),
            (0, 27),
            (0, 9502),
            (0, 2),
            (0, 52),
            (0, 3807),
            (0, 15),
            (0, 536),
            (0, 3267),
            (0, 7),
            (0, 67748),
            (0, 13876),
            (0, 29457),
            (0, 3628),
            (0, 4),
            (0, 1043),
            (0, 101),
            (0, 1),
            (0, 8756),
            (0, 169),
            (0, 73),
            (0, 5958),
            (0, 60),
            (0, 2072),
            (0, 27484),
            (0, 93),
            (0, 2197),
            (0, 18369),
            (0, 0),
            (0, 116),
            (0, 24),
            (0, 1449),
            (0, 468),
            (0, 1),
            (0, 1),
            (0, 987),
            (0, 66),
            (0, 12088),
            (0, 12842),
            (0, 976),
            (0, 2054),
            (0, 19503),
            (0, 69),
            (0, 11089),
            (0, 35405),
            (0, 138),
            (0, 145),
            (0, 1),
            (0, 38758),
            (0, 16939),
            (0, 1554),
            (0, 3479),
            (0, 9950),
            (0, 27230),
            (0, 18752),
            (0, 20),
            (0, 1),
            (0, 309),
            (0, 405),
            (0, 317),
            (0, 24706),
            (0, 33),
            (0, 7),
            (0, 31848),
            (0, 2784),
            (0, 730),
            (0, 773),
            (0, 5),
            (0, 258),
            (0, 2029),
            (0, 0),
            (0, 3),
            (0, 3),
            (0, 0),
            (0, 921),
            (0, 1043),
            (0, 48),
            (0, 79020),
            (0, 1282),
            (0, 202),
            (0, 22030),
            (0, 12850),
            (0, 1945),
            (0, 3636),
            (0, 260),
            (0, 126),
            (0, 8),
            (0, 138),
            (0, 28),
            (0, 251),
            (0, 199),
            (0, 615),
            (0, 46),
            (0, 1479),
            (0, 9809),
            (0, 23),
            (0, 3),
            (0, 2709),
            (0, 270),
            (0, 1),
            (1, 3756),
            (0, 29),
            (0, 10),
            (0, 237),
            (0, 516),
            (0, 27),
            (0, 73756),
            (0, 419),
            (0, 425),
            (0, 54),
            (0, 57065),
            (0, 9751),
            (0, 583),
            (0, 7298),
            (0, 5376),
            (0, 60),
            (0, 9021),
            (0, 3),
            (0, 2937),
            (0, 72920),
            (0, 1246),
            (0, 8),
            (0, 7),
            (0, 4950),
            (0, 9),
            (0, 13641),
            (0, 136),
            (0, 4471),
            (0, 1486),
            (0, 7388),
            (0, 21),
            (0, 231),
            (0, 136),
            (0, 53),
            (0, 44),
            (0, 2252),
            (0, 138),
            (0, 64),
            (0, 96),
            (0, 5329),
            (0, 427),
            (0, 658),
            (0, 123),
            (0, 16),
            (0, 906),
            (0, 13),
            (0, 114),
            (0, 20987),
            (0, 3753),
            (0, 174),
            (0, 44953),
            (0, 25),
            (0, 356),
            (0, 26),
            (0, 20),
            (0, 1047),
            (0, 1536),
            (0, 65),
            (0, 56),
            (0, 1),
            (0, 689),
            (0, 265),
            (0, 23),
            (0, 9045),
            (0, 289),
            (0, 15919),
            (0, 2252),
            (0, 6161),
            (0, 39587),
            (0, 189),
            (0, 499),
            (0, 5),
            (0, 29907),
            (0, 3),
            (0, 1),
            (0, 387),
            (0, 873),
            (0, 11373),
            (0, 6),
            (0, 8545),
            (0, 22595),
            (0, 0),
            (0, 4),
            (0, 12260),
            (0, 160),
            (0, 26),
            (0, 41815),
            (0, 11772),
            (0, 41199),
            (0, 3),
            (0, 1161),
            (0, 84499),
            (0, 681),
            (0, 3083),
            (0, 5),
            (0, 44124),
            (0, 31),
            (0, 91),
            (0, 9229),
            (0, 2),
            (0, 88),
            (0, 8151),
            (0, 2695),
            (0, 2),
            (0, 0),
            (0, 4715),
            (0, 0),
            (0, 38784),
            (0, 16073),
            (0, 1),
            (0, 2145),
            (0, 250),
            (0, 78656),
        ],
        entries: &[
            (dictgen::InsensitiveStr::Ascii("vassalls"), &["vassals"]),
            (dictgen::InsensitiveStr::Ascii("eveolving"), &["evolving"]),
            (
                dictgen::InsensitiveStr::Ascii("remmebering"),
                &["remembering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabalistically"),
                &["probabilistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serviceble"),
                &["serviceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("betwee"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("henious"), &["heinous"]),
            (
                dictgen::InsensitiveStr::Ascii("screenshit"),
                &["screenshot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmaceutial"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibilies"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("schaedule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("fingertrips"),
                &["fingertips"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantly"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("realation"), &["relation"]),
            (
                dictgen::InsensitiveStr::Ascii("desriptive"),
                &["descriptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuppancy"), &["occupancy"]),
            (dictgen::InsensitiveStr::Ascii("regualr"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("sratches"),
                &["scratches", "snatches", "starches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nummerators"),
                &["numerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("liquidas"), &["liquids"]),
            (
                dictgen::InsensitiveStr::Ascii("instituion"),
                &["institution"],
            ),
            (dictgen::InsensitiveStr::Ascii("unline"), &["unlike"]),
            (
                dictgen::InsensitiveStr::Ascii("asscociations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recquires"),
                &["requires", "reacquires"],
            ),
            (dictgen::InsensitiveStr::Ascii("endevour"), &["endeavour"]),
            (
                dictgen::InsensitiveStr::Ascii("availabality"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configutres"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("posption"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("emulaton"), &["emulation"]),
            (dictgen::InsensitiveStr::Ascii("somethin"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsecuential"),
                &["inconsequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("allwed"), &["allowed"]),
            (
                dictgen::InsensitiveStr::Ascii("provenence"),
                &["provenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquaruims"), &["aquariums"]),
            (
                dictgen::InsensitiveStr::Ascii("instlaling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("contanined"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("ponited"), &["pointed"]),
            (dictgen::InsensitiveStr::Ascii("christin"), &["christian"]),
            (
                dictgen::InsensitiveStr::Ascii("independnt"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaiter"), &["animator"]),
            (
                dictgen::InsensitiveStr::Ascii("accomplise"),
                &["accomplishes", "accomplish", "accomplice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sikscreened"),
                &["silkscreened"],
            ),
            (dictgen::InsensitiveStr::Ascii("reccording"), &["recording"]),
            (dictgen::InsensitiveStr::Ascii("assiociate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("researvation"),
                &["reservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("associeate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("misimformed"),
                &["misinformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("messangers"),
                &["messengers"],
            ),
            (dictgen::InsensitiveStr::Ascii("failuring"), &["failing"]),
            (dictgen::InsensitiveStr::Ascii("developper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("bootrap"), &["bootstrap"]),
            (
                dictgen::InsensitiveStr::Ascii("legitamite"),
                &["legitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterpreneurs"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegociations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initatives"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("ultimae"), &["ultimate"]),
            (dictgen::InsensitiveStr::Ascii("deletetes"), &["deletes"]),
            (
                dictgen::InsensitiveStr::Ascii("elagance"),
                &["elegant", "elegance"],
            ),
            (dictgen::InsensitiveStr::Ascii("provicing"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("dellocated"),
                &["deallocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("memonics"), &["mnemonics"]),
            (dictgen::InsensitiveStr::Ascii("smeesters"), &["semesters"]),
            (dictgen::InsensitiveStr::Ascii("forgeiner"), &["foreigner"]),
            (dictgen::InsensitiveStr::Ascii("functiions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("iniitials"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("execeeds"), &["exceeds"]),
            (
                dictgen::InsensitiveStr::Ascii("milisecond"),
                &["millisecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("geenrate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("capabiltity"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("lauguages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("commuication"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trangulations"),
                &["triangulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overestemating"),
                &["overestimating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sasuage"), &["sausage"]),
            (dictgen::InsensitiveStr::Ascii("aspriin"), &["aspirin"]),
            (dictgen::InsensitiveStr::Ascii("despides"), &["despised"]),
            (
                dictgen::InsensitiveStr::Ascii("alpabetical"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexplicibly"),
                &["inexplicably"],
            ),
            (dictgen::InsensitiveStr::Ascii("suficated"), &["suffocated"]),
            (dictgen::InsensitiveStr::Ascii("presearved"), &["preserved"]),
            (
                dictgen::InsensitiveStr::Ascii("toothbruch"),
                &["toothbrush"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mintoring"),
                &["mentoring", "monitoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzied"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("hotpsot"), &["hotspot"]),
            (
                dictgen::InsensitiveStr::Ascii("coppied"),
                &["copied", "coppiced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understants"),
                &["understands", "understates"],
            ),
            (dictgen::InsensitiveStr::Ascii("stupitidy"), &["stupidity"]),
            (dictgen::InsensitiveStr::Ascii("interferre"), &["interfere"]),
            (
                dictgen::InsensitiveStr::Ascii("checkstumming"),
                &["checksumming"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgasmes"), &["orgasms"]),
            (
                dictgen::InsensitiveStr::Ascii("tructs"),
                &["structs", "trusts", "trucks", "tracts", "truces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("safepooints"),
                &["safepoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("acconting"), &["accounting"]),
            (
                dictgen::InsensitiveStr::Ascii("inititialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteactively"),
                &["interactively"],
            ),
            (dictgen::InsensitiveStr::Ascii("beauquets"), &["bouquets"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicaor"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierchical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inheirtance"),
                &["inheritance"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrytpts"), &["encrypts"]),
            (
                dictgen::InsensitiveStr::Ascii("commmunication"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutlithreading"),
                &["multithreading"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootstra"), &["bootstrap"]),
            (
                dictgen::InsensitiveStr::Ascii("libertairan"),
                &["libertarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("naturely"), &["naturally"]),
            (dictgen::InsensitiveStr::Ascii("diffence"), &["difference"]),
            (dictgen::InsensitiveStr::Ascii("infininty"), &["infinity"]),
            (dictgen::InsensitiveStr::Ascii("suppier"), &["supplier"]),
            (dictgen::InsensitiveStr::Ascii("enbiggen"), &["embiggen"]),
            (dictgen::InsensitiveStr::Ascii("proteceted"), &["protected"]),
            (
                dictgen::InsensitiveStr::Ascii("convertirle"),
                &["convertible"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinals"), &["signals"]),
            (dictgen::InsensitiveStr::Ascii("metting"), &["meeting"]),
            (dictgen::InsensitiveStr::Ascii("repblic"), &["republic"]),
            (dictgen::InsensitiveStr::Ascii("snydrome"), &["syndrome"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiazations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("resultung"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("closts"),
                &["closest", "closets"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrenges"), &["arranges"]),
            (
                dictgen::InsensitiveStr::Ascii("coutermeasuere"),
                &["countermeasure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cellabrated"),
                &["celebrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impementated"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("aftrerward"), &["afterward"]),
            (
                dictgen::InsensitiveStr::Ascii("ridiculosly"),
                &["ridiculously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuprizingly"),
                &["unsurprisingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("commedians"), &["comedians"]),
            (
                dictgen::InsensitiveStr::Ascii("thermodinamics"),
                &["thermodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consomation"),
                &["consolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inititial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("perturbates"), &["perturbs"]),
            (dictgen::InsensitiveStr::Ascii("fualt"), &["fault"]),
            (dictgen::InsensitiveStr::Ascii("cleen"), &["clean"]),
            (dictgen::InsensitiveStr::Ascii("decond"), &["decode"]),
            (
                dictgen::InsensitiveStr::Ascii("convinence"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("misspel"), &["misspell"]),
            (
                dictgen::InsensitiveStr::Ascii("conditinal"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cherchs"), &["churches"]),
            (dictgen::InsensitiveStr::Ascii("disclosue"), &["disclosure"]),
            (dictgen::InsensitiveStr::Ascii("enumates"), &["enumerates"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabule"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("binominal"), &["binomial"]),
            (dictgen::InsensitiveStr::Ascii("defendeers"), &["defender"]),
            (
                dictgen::InsensitiveStr::Ascii("susecptible"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("wtiches"), &["witches"]),
            (dictgen::InsensitiveStr::Ascii("croation"), &["croatian"]),
            (
                dictgen::InsensitiveStr::Ascii("terroristes"),
                &["terrorists"],
            ),
            (dictgen::InsensitiveStr::Ascii("becuasse"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("entir"),
                &["entire", "enter"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpect"), &["inspect"]),
            (
                dictgen::InsensitiveStr::Ascii("exagerates"),
                &["exaggerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("flecther"), &["fletcher"]),
            (dictgen::InsensitiveStr::Ascii("synonims"), &["synonyms"]),
            (dictgen::InsensitiveStr::Ascii("freqent"), &["frequent"]),
            (dictgen::InsensitiveStr::Ascii("styleing"), &["styling"]),
            (dictgen::InsensitiveStr::Ascii("oterwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("idices"), &["indices"]),
            (
                dictgen::InsensitiveStr::Ascii("actuionable"),
                &["actionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("acqures"), &["acquires"]),
            (
                dictgen::InsensitiveStr::Ascii("dispapointed"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("inproves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("recofig"), &["reconfig"]),
            (dictgen::InsensitiveStr::Ascii("govorment"), &["government"]),
            (
                dictgen::InsensitiveStr::Ascii("origininating"),
                &["originating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gunslingner"),
                &["gunslinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apretiating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutiples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("dependd"), &["depended"]),
            (
                dictgen::InsensitiveStr::Ascii("inialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggegator"), &["aggregator"]),
            (dictgen::InsensitiveStr::Ascii("briten"), &["brighten"]),
            (
                dictgen::InsensitiveStr::Ascii("charaterizes"),
                &["characterizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropollis"),
                &["metropolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconcistencies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accellerated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychodelic"),
                &["psychedelic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philospohical"),
                &["philosophical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccessble"),
                &["inaccessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("confog"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("simplificaiton"),
                &["simplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("travesing"), &["traversing"]),
            (
                dictgen::InsensitiveStr::Ascii("agressviely"),
                &["aggressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("apears"), &["appears"]),
            (dictgen::InsensitiveStr::Ascii("pssibly"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("importanty"),
                &["importantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consituent"),
                &["constituent"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgarded"), &["upgraded"]),
            (
                dictgen::InsensitiveStr::Ascii("ineresting"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brainwased"),
                &["brainwashed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("couroption"),
                &["corruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agigressive"),
                &["aggressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multpilication"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionares"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deoendencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("thiis"), &["this"]),
            (
                dictgen::InsensitiveStr::Ascii("trasncoding"),
                &["transcoding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backrefence"),
                &["backreference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abigiously"),
                &["ambiguously"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahhels"), &["channels"]),
            (dictgen::InsensitiveStr::Ascii("recpies"), &["recipes"]),
            (
                dictgen::InsensitiveStr::Ascii("proejctions"),
                &["projections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogenenous"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("paased"), &["passed"]),
            (dictgen::InsensitiveStr::Ascii("intiutive"), &["intuitive"]),
            (dictgen::InsensitiveStr::Ascii("antartic"), &["antarctic"]),
            (dictgen::InsensitiveStr::Ascii("diagnosin"), &["diagnosing"]),
            (dictgen::InsensitiveStr::Ascii("psydonyms"), &["pseudonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("docuemtnation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fielesystem"),
                &["filesystem"],
            ),
            (dictgen::InsensitiveStr::Ascii("savgroup"), &["savegroup"]),
            (
                dictgen::InsensitiveStr::Ascii("excecptional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaation"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("recurssion"), &["recursion"]),
            (dictgen::InsensitiveStr::Ascii("copmiling"), &["compiling"]),
            (
                dictgen::InsensitiveStr::Ascii("interspereses"),
                &["intersperses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prematurley"),
                &["prematurely"],
            ),
            (dictgen::InsensitiveStr::Ascii("hybirds"), &["hybrids"]),
            (
                dictgen::InsensitiveStr::Ascii("somthingelse"),
                &["somethingelse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technitions"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porjection"),
                &["projection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchendise"),
                &["merchandise"],
            ),
            (dictgen::InsensitiveStr::Ascii("archval"), &["archival"]),
            (
                dictgen::InsensitiveStr::Ascii("percission"),
                &["percussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("succesor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("approximatelly"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("deubuging"), &["debugging"]),
            (
                dictgen::InsensitiveStr::Ascii("inconlusive"),
                &["inconclusive"],
            ),
            (dictgen::InsensitiveStr::Ascii("haulted"), &["halted"]),
            (
                dictgen::InsensitiveStr::Ascii("enrolement"),
                &["enrollment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pessimisitic"),
                &["pessimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("stupdily"), &["stupidly"]),
            (dictgen::InsensitiveStr::Ascii("adorbale"), &["adorable"]),
            (dictgen::InsensitiveStr::Ascii("inbedded"), &["imbedded"]),
            (
                dictgen::InsensitiveStr::Ascii("categoriei"),
                &["categorize"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouer"), &["outer"]),
            (dictgen::InsensitiveStr::Ascii("appendex"), &["appendix"]),
            (
                dictgen::InsensitiveStr::Ascii("capabilies"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("validion"), &["validation"]),
            (dictgen::InsensitiveStr::Ascii("athleticos"), &["athletics"]),
            (dictgen::InsensitiveStr::Ascii("blancked"), &["blanked"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronized"),
                &["synchronized"],
            ),
            (dictgen::InsensitiveStr::Ascii("begain"), &["began"]),
            (dictgen::InsensitiveStr::Ascii("underfiend"), &["undefined"]),
            (
                dictgen::InsensitiveStr::Ascii("cellebrating"),
                &["celebrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("turain"), &["terrain"]),
            (dictgen::InsensitiveStr::Ascii("functiuons"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("ponter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("enumurating"),
                &["enumerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("proejct"), &["project"]),
            (dictgen::InsensitiveStr::Ascii("sercive"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("mayybe"), &["maybe"]),
            (
                dictgen::InsensitiveStr::Ascii("acccordingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("poosible"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("highst"), &["highest"]),
            (
                dictgen::InsensitiveStr::Ascii("disperportionate"),
                &["disproportionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practictitioner"),
                &["practitioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imporvement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhanes"), &["enhances"]),
            (
                dictgen::InsensitiveStr::Ascii("accoustically"),
                &["acoustically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noe"),
                &["not", "no", "node", "know", "now", "note", "none"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tarnsparent"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedamic"), &["academic"]),
            (
                dictgen::InsensitiveStr::Ascii("conneticut"),
                &["connecticut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepparation"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarbage"), &["garbage"]),
            (dictgen::InsensitiveStr::Ascii("ethnicitiy"), &["ethnicity"]),
            (
                dictgen::InsensitiveStr::Ascii("mansalughter"),
                &["manslaughter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportinate"),
                &["proportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("readnig"), &["reading"]),
            (
                dictgen::InsensitiveStr::Ascii("responcible"),
                &["responsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("auhtorize"), &["authorize"]),
            (dictgen::InsensitiveStr::Ascii("fillled"), &["filled"]),
            (
                dictgen::InsensitiveStr::Ascii("enumerables"),
                &["enumerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("classicals"), &["classics"]),
            (dictgen::InsensitiveStr::Ascii("normlized"), &["normalized"]),
            (dictgen::InsensitiveStr::Ascii("opitonal"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("proffesor"), &["professor"]),
            (dictgen::InsensitiveStr::Ascii("foze"), &["froze", "doze"]),
            (dictgen::InsensitiveStr::Ascii("trrigered"), &["triggered"]),
            (dictgen::InsensitiveStr::Ascii("changnig"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("costumiser"),
                &["customiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("disbelif"), &["disbelief"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintenace"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriarcal"),
                &["patriarchal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carnavores"),
                &["carnivores"],
            ),
            (dictgen::InsensitiveStr::Ascii("coutries"), &["countries"]),
            (dictgen::InsensitiveStr::Ascii("olt"), &["old"]),
            (dictgen::InsensitiveStr::Ascii("temaplated"), &["templated"]),
            (
                dictgen::InsensitiveStr::Ascii("rsizing"),
                &["resizing", "sizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detectsions"),
                &["detections"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuction"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("offten"), &["often"]),
            (
                dictgen::InsensitiveStr::Ascii("individules"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlytical"), &["analytical"]),
            (
                dictgen::InsensitiveStr::Ascii("perfermances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecrting"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoductions"),
                &["reproductions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyticall"),
                &["analytical", "analytically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("altruistisch"),
                &["altruistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("classifes"), &["classifies"]),
            (
                dictgen::InsensitiveStr::Ascii("bueraucrats"),
                &["bureaucrats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prebiously"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subfoilders"),
                &["subdolders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diminsioned"),
                &["dimensioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referenencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartain"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("reaeched"), &["reached"]),
            (dictgen::InsensitiveStr::Ascii("datadsir"), &["datadir"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmectrically"),
                &["asymmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guareenteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("exponets"), &["exponents"]),
            (dictgen::InsensitiveStr::Ascii("operarion"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("creteria"), &["criteria"]),
            (
                dictgen::InsensitiveStr::Ascii("absoltuely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("homeonwer"), &["homeowner"]),
            (dictgen::InsensitiveStr::Ascii("negotate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("relaised"), &["realised"]),
            (
                dictgen::InsensitiveStr::Ascii("concequently"),
                &["consequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("detault"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("memebrs"), &["members"]),
            (dictgen::InsensitiveStr::Ascii("interpets"), &["interprets"]),
            (
                dictgen::InsensitiveStr::Ascii("coneptually"),
                &["conceptually"],
            ),
            (dictgen::InsensitiveStr::Ascii("milages"), &["mileages"]),
            (dictgen::InsensitiveStr::Ascii("residant"), &["resident"]),
            (dictgen::InsensitiveStr::Ascii("colects"), &["collects"]),
            (
                dictgen::InsensitiveStr::Ascii("emphesizes"),
                &["emphasizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("chrsitian"), &["christian"]),
            (dictgen::InsensitiveStr::Ascii("partiot"), &["patriot"]),
            (
                dictgen::InsensitiveStr::Ascii("statistacally"),
                &["statistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("hevy"), &["heavy"]),
            (dictgen::InsensitiveStr::Ascii("terrorits"), &["terrorist"]),
            (
                dictgen::InsensitiveStr::Ascii("accedentally"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noticeabley"),
                &["noticeably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deadlifters"),
                &["deadlifts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedohpilia"),
                &["pedophilia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alocatable"),
                &["allocatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motation"),
                &["notation", "rotation", "motivation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virualized"),
                &["virtualized", "visualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploaeded"), &["uploaded"]),
            (dictgen::InsensitiveStr::Ascii("iteratiors"), &["iterators"]),
            (dictgen::InsensitiveStr::Ascii("feminsits"), &["feminists"]),
            (
                dictgen::InsensitiveStr::Ascii("requenst"),
                &["request", "requests"],
            ),
            (dictgen::InsensitiveStr::Ascii("fideltiy"), &["fidelity"]),
            (
                dictgen::InsensitiveStr::Ascii("reconnctions"),
                &["reconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intriniscally"),
                &["intrinsically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colonializm"),
                &["colonialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("characeter"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("vertially"), &["vertically"]),
            (
                dictgen::InsensitiveStr::Ascii("manupulation"),
                &["manipulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("normlised"), &["normalised"]),
            (
                dictgen::InsensitiveStr::Ascii("accrediated"),
                &["accredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cartrigdes"),
                &["cartridges"],
            ),
            (dictgen::InsensitiveStr::Ascii("electrones"), &["electrons"]),
            (dictgen::InsensitiveStr::Ascii("renewl"), &["renewal"]),
            (
                dictgen::InsensitiveStr::Ascii("copyrughts"),
                &["copyrights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("istallation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trempoline"),
                &["trampoline"],
            ),
            (dictgen::InsensitiveStr::Ascii("tigth"), &["tight"]),
            (dictgen::InsensitiveStr::Ascii("exhistence"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("reaktivated"),
                &["reactivated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilizaton"),
                &["civilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infektious"),
                &["infectious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beliefable"),
                &["believable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wonderously"),
                &["wondrously"],
            ),
            (dictgen::InsensitiveStr::Ascii("amonsgt"), &["amongst"]),
            (dictgen::InsensitiveStr::Ascii("insered"), &["inserted"]),
            (
                dictgen::InsensitiveStr::Ascii("translattable"),
                &["translatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbouthood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investiages"),
                &["investigates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multuplayer"),
                &["multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimazes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("exploitions"),
                &["explosions", "exploitations", "exploits"],
            ),
            (dictgen::InsensitiveStr::Ascii("aixs"), &["axis"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanetation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conflictin"),
                &["conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("mircales"), &["miracles"]),
            (dictgen::InsensitiveStr::Ascii("execeeded"), &["exceeded"]),
            (dictgen::InsensitiveStr::Ascii("curiosly"), &["curiously"]),
            (
                dictgen::InsensitiveStr::Ascii("sleectable"),
                &["selectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinined"), &["combined"]),
            (dictgen::InsensitiveStr::Ascii("excuted"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("apends"), &["appends"]),
            (
                dictgen::InsensitiveStr::Ascii("instituto"),
                &["institution"],
            ),
            (dictgen::InsensitiveStr::Ascii("hothsot"), &["hotshot"]),
            (dictgen::InsensitiveStr::Ascii("resonabelt"), &["resonate"]),
            (dictgen::InsensitiveStr::Ascii("checkkbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratuions"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quatertion"),
                &["quaternion"],
            ),
            (dictgen::InsensitiveStr::Ascii("inables"), &["enables"]),
            (dictgen::InsensitiveStr::Ascii("eurpoean"), &["european"]),
            (dictgen::InsensitiveStr::Ascii("broadcas"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("decathalon"), &["decathlon"]),
            (
                dictgen::InsensitiveStr::Ascii("mischievious"),
                &["mischievous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privelaged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaches"), &["caches"]),
            (
                dictgen::InsensitiveStr::Ascii("recrutiting"),
                &["recruiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("scollbars"), &["scrollbars"]),
            (
                dictgen::InsensitiveStr::Ascii("transterring"),
                &["transferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("eearly"), &["early"]),
            (dictgen::InsensitiveStr::Ascii("mecanics"), &["mechanics"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreagation"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("peerowet"), &["pirouette"]),
            (
                dictgen::InsensitiveStr::Ascii("commnetators"),
                &["commentators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasferred"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserializazion"),
                &["deserialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sypmathize"),
                &["sympathize"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragrap"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("exeuctions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("windshiled"),
                &["windshield"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("traveerse"), &["traverse"]),
            (
                dictgen::InsensitiveStr::Ascii("spirtiually"),
                &["spiritually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasforming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redemtions"),
                &["redemptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("icelings"), &["ceilings"]),
            (dictgen::InsensitiveStr::Ascii("recyled"), &["recycled"]),
            (
                dictgen::InsensitiveStr::Ascii("ideosyncracies"),
                &["ideosyncrasies", "idiosyncrasies"],
            ),
            (dictgen::InsensitiveStr::Ascii("procedger"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("extraxted"), &["extracted"]),
            (
                dictgen::InsensitiveStr::Ascii("intilaizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilpicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("conclue"), &["conclude"]),
            (
                dictgen::InsensitiveStr::Ascii("rewriable"),
                &["rewritable", "reliable"],
            ),
            (dictgen::InsensitiveStr::Ascii("limite"), &["limit"]),
            (dictgen::InsensitiveStr::Ascii("credate"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("compaines"), &["companies"]),
            (dictgen::InsensitiveStr::Ascii("versios"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("dabilitating"),
                &["debilitating"],
            ),
            (dictgen::InsensitiveStr::Ascii("credntial"), &["credential"]),
            (
                dictgen::InsensitiveStr::Ascii("connatations"),
                &["connotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacame"), &["became"]),
            (
                dictgen::InsensitiveStr::Ascii("attribuito"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpact"), &["compact"]),
            (dictgen::InsensitiveStr::Ascii("librairies"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("cricitals"), &["criticals"]),
            (
                dictgen::InsensitiveStr::Ascii("hypvisors"),
                &["hypervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitionns"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychologial"),
                &["psychological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beautyfied"),
                &["beautified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominaron"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("roels"), &["roles"]),
            (
                dictgen::InsensitiveStr::Ascii("defintians"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simlations"),
                &["simulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemetns"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulataeously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptot"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("binraries"), &["binaries"]),
            (dictgen::InsensitiveStr::Ascii("declarates"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("complite"),
                &["complete", "compile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composistion"),
                &["compositions", "composition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsequental"),
                &["inconsequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("ojective"), &["objective"]),
            (dictgen::InsensitiveStr::Ascii("gerilla"), &["guerilla"]),
            (
                dictgen::InsensitiveStr::Ascii("whitrespace"),
                &["whitespace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alrorythms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("immedidate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("pompting"), &["prompting"]),
            (
                dictgen::InsensitiveStr::Ascii("quanifies"),
                &["quantifies", "qualifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumemnt"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("insatll"), &["install"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrarely"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("preproces"), &["preprocess"]),
            (
                dictgen::InsensitiveStr::Ascii("unauthroized"),
                &["unauthorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyring"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("accountatns"),
                &["accountants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("westmisnter"),
                &["westminster"],
            ),
            (dictgen::InsensitiveStr::Ascii("spsacing"), &["spacing"]),
            (dictgen::InsensitiveStr::Ascii("sythesis"), &["synthesis"]),
            (dictgen::InsensitiveStr::Ascii("condesend"), &["condensed"]),
            (
                dictgen::InsensitiveStr::Ascii("apprixamation"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apppriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("refelct"), &["reflect"]),
            (
                dictgen::InsensitiveStr::Ascii("beding"),
                &["bedding", "begging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stati"),
                &["statuses", "state", "static"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amendmends"),
                &["amendments"],
            ),
            (dictgen::InsensitiveStr::Ascii("inderect"), &["indirect"]),
            (dictgen::InsensitiveStr::Ascii("sandwitch"), &["sandwich"]),
            (
                dictgen::InsensitiveStr::Ascii("grandising"),
                &["aggrandising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("govermnent"),
                &["governments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwhlemed"),
                &["overwhelmed"],
            ),
            (dictgen::InsensitiveStr::Ascii("stabl"), &["stable"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsultion"),
                &["encapsulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("codesbases"), &["codebases"]),
            (
                dictgen::InsensitiveStr::Ascii("bounhdaries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("worksace"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("brobabilistic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassemble"),
                &["disassemble"],
            ),
            (dictgen::InsensitiveStr::Ascii("cvignore"), &["cvsignore"]),
            (dictgen::InsensitiveStr::Ascii("posession"), &["possession"]),
            (
                dictgen::InsensitiveStr::Ascii("assoiations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specualtions"),
                &["speculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepenently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("lannguage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("pront"),
                &["print", "pronto", "prone", "front", "prong"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvoteas"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("momemtary"), &["momentary"]),
            (dictgen::InsensitiveStr::Ascii("existince"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("craete"),
                &["create", "crate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tortillla"), &["tortilla"]),
            (
                dictgen::InsensitiveStr::Ascii("powerfisting"),
                &["powerlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplivity"),
                &["simplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("orangerd"), &["orangered"]),
            (
                dictgen::InsensitiveStr::Ascii("authenricator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confussed"),
                &["confused", "confessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbbacks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("annoyningly"),
                &["annoyingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglegible"),
                &["negligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommnding"),
                &["recommending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstruncted"),
                &["reconstructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitency"),
                &["consistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invitating"),
                &["invitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arithmetc"), &["arithmetic"]),
            (dictgen::InsensitiveStr::Ascii("sellect"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("omains"), &["domains"]),
            (
                dictgen::InsensitiveStr::Ascii("fraudelant"),
                &["fraudulent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exectuioner"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("expeption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("stablility"), &["stability"]),
            (dictgen::InsensitiveStr::Ascii("samilarly"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("proficit"), &["proficient"]),
            (
                dictgen::InsensitiveStr::Ascii("desserialising"),
                &["deserialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complicite"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterizes"),
                &["parameterizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enemployment"),
                &["unemployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("givving"), &["giving"]),
            (
                dictgen::InsensitiveStr::Ascii("destinaitons"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaiance"),
                &["variance", "valiance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("withrdawal"),
                &["withdrawals"],
            ),
            (dictgen::InsensitiveStr::Ascii("warinings"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifcally"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("seacrches"), &["searches"]),
            (dictgen::InsensitiveStr::Ascii("plethorian"), &["plethora"]),
            (
                dictgen::InsensitiveStr::Ascii("approoximations"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("diverisfy"), &["diversify"]),
            (dictgen::InsensitiveStr::Ascii("inactivly"), &["inactively"]),
            (
                dictgen::InsensitiveStr::Ascii("limitiaion"),
                &["limitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("librari"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("chekout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("renosance"),
                &["renaissance", "resonance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parternships"),
                &["partnership"],
            ),
            (dictgen::InsensitiveStr::Ascii("sesison"), &["session"]),
            (
                dictgen::InsensitiveStr::Ascii("accummulative"),
                &["accumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jailbrocken"),
                &["jailbroken"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nasturshem"),
                &["nasturtium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elemnetary"),
                &["elementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifing"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("centrilise"),
                &["centralise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuretion"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significato"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("separtion"), &["separation"]),
            (dictgen::InsensitiveStr::Ascii("foootball"), &["football"]),
            (
                dictgen::InsensitiveStr::Ascii("transilvania"),
                &["transylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappoint"),
                &["disappoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("streamos"), &["streams"]),
            (
                dictgen::InsensitiveStr::Ascii("perforaming"),
                &["performing"],
            ),
            (dictgen::InsensitiveStr::Ascii("excaped"), &["escaped"]),
            (
                dictgen::InsensitiveStr::Ascii("instinctivly"),
                &["instinctively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilegier"),
                &["privilege"],
            ),
            (dictgen::InsensitiveStr::Ascii("funrel"), &["funeral"]),
            (dictgen::InsensitiveStr::Ascii("properely"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("preferables"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disscovered"),
                &["discovered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progapations"),
                &["propagations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cthuhlu"), &["cthulhu"]),
            (dictgen::InsensitiveStr::Ascii("sycles"), &["cycles"]),
            (
                dictgen::InsensitiveStr::Ascii("combied"),
                &["combined", "combed"],
            ),
            (dictgen::InsensitiveStr::Ascii("hygine"), &["hygiene"]),
            (
                dictgen::InsensitiveStr::Ascii("infinitly"),
                &["infinity", "infinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculalated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insurgancy"),
                &["insurgency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decrased"),
                &["decreased", "deceased"],
            ),
            (dictgen::InsensitiveStr::Ascii("eachtoher"), &["eachother"]),
            (
                dictgen::InsensitiveStr::Ascii("ctificate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdifiers"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("compansates"),
                &["compensates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcribers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("wordlwide"), &["worldwide"]),
            (
                dictgen::InsensitiveStr::Ascii("transperency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succeessful"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasmitting"),
                &["transmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deterioriating"),
                &["deteriorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("everone"), &["everyone"]),
            (dictgen::InsensitiveStr::Ascii("dedsired"), &["desired"]),
            (dictgen::InsensitiveStr::Ascii("exterbal"), &["external"]),
            (dictgen::InsensitiveStr::Ascii("capitans"), &["captains"]),
            (dictgen::InsensitiveStr::Ascii("loding"), &["loading"]),
            (
                dictgen::InsensitiveStr::Ascii("disenchanged"),
                &["disenchanted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactetistic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("normanday"), &["normandy"]),
            (dictgen::InsensitiveStr::Ascii("installates"), &["installs"]),
            (
                dictgen::InsensitiveStr::Ascii("implicitely"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("yiu"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("dockefile"), &["dockerfile"]),
            (
                dictgen::InsensitiveStr::Ascii("contributuon"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("requriing"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("peculair"), &["peculiar"]),
            (dictgen::InsensitiveStr::Ascii("atttempt"), &["attempt"]),
            (
                dictgen::InsensitiveStr::Ascii("degenrates"),
                &["degenerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cricital"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("flawleslly"),
                &["flawlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronts"),
                &["prints", "prongs", "pronto"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expecially"),
                &["especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randonmess"),
                &["randomness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepearting"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("normolize"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("cannnot"), &["cannot"]),
            (
                dictgen::InsensitiveStr::Ascii("allocatiom"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wraper"), &["wrapper"]),
            (
                dictgen::InsensitiveStr::Ascii("initilialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilializers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fexibility"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("republicons"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("soovinears"), &["souvenirs"]),
            (
                dictgen::InsensitiveStr::Ascii("essentualy"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinquishable"),
                &["distinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaning"), &["containing"]),
            (dictgen::InsensitiveStr::Ascii("associatd"), &["associated"]),
            (dictgen::InsensitiveStr::Ascii("voif"), &["void"]),
            (
                dictgen::InsensitiveStr::Ascii("comfrontation"),
                &["confrontation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upsrteamer"),
                &["upstreamer"],
            ),
            (dictgen::InsensitiveStr::Ascii("appearant"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("benjing"), &["binging"]),
            (
                dictgen::InsensitiveStr::Ascii("inialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ulteration"),
                &["alteration"],
            ),
            (dictgen::InsensitiveStr::Ascii("remontly"), &["remotely"]),
            (
                dictgen::InsensitiveStr::Ascii("propetiies"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("grabage"), &["garbage"]),
            (dictgen::InsensitiveStr::Ascii("slewthed"), &["sleuthed"]),
            (
                dictgen::InsensitiveStr::Ascii("assosiations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influancing"),
                &["influencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perentage"),
                &["percentage", "parentage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condig"),
                &["config", "coding", "condign"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pollutin"),
                &["polluting", "pollution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchoronous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastrcutures"),
                &["infrastructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootrapping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutexs"), &["mutexes"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("adrresses"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("corretors"), &["correctors"]),
            (dictgen::InsensitiveStr::Ascii("tightare"), &["tighter"]),
            (
                dictgen::InsensitiveStr::Ascii("uplodad"),
                &["upload", "uploaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bangaldeshi"),
                &["bangladesh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychedellics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("finanical"), &["financial"]),
            (
                dictgen::InsensitiveStr::Ascii("impelemented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tremondously"),
                &["tremendously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviornemnt"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descroption"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forwared"),
                &["forwarded", "forward"],
            ),
            (dictgen::InsensitiveStr::Ascii("percieves"), &["perceives"]),
            (dictgen::InsensitiveStr::Ascii("hopefullt"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("contiainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("execitioner"),
                &["executioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responibilities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architechts"),
                &["architect", "architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combininatorics"),
                &["combinatorics"],
            ),
            (dictgen::InsensitiveStr::Ascii("excersizes"), &["exercises"]),
            (
                dictgen::InsensitiveStr::Ascii("initliasation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialiazes"),
                &["specializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("creenshot"), &["screenshot"]),
            (dictgen::InsensitiveStr::Ascii("exerpts"), &["excerpts"]),
            (dictgen::InsensitiveStr::Ascii("thershold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("unssuported"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theyre"),
                &["their", "there"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperaly"), &["separately"]),
            (
                dictgen::InsensitiveStr::Ascii("whent"),
                &["went", "when", "wheat", "whet"],
            ),
            (dictgen::InsensitiveStr::Ascii("calculs"), &["calculus"]),
            (
                dictgen::InsensitiveStr::Ascii("ambassader"),
                &["ambassador"],
            ),
            (dictgen::InsensitiveStr::Ascii("acepter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnfered"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("possable"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("vresions"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("costumary"), &["customary"]),
            (
                dictgen::InsensitiveStr::Ascii("narritives"),
                &["narratives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepparatelly"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("dabree"), &["debris"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhoof"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customziations"),
                &["customizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theresholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diskretization"),
                &["discretization"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitimg"), &["limiting"]),
            (dictgen::InsensitiveStr::Ascii("traversare"), &["traverse"]),
            (
                dictgen::InsensitiveStr::Ascii("appriciation"),
                &["appreciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("montoring"),
                &["monitoring", "mentoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleeteness"),
                &["completeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enchentments"),
                &["enchantments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technicall"),
                &["technically", "technical"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconc"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessing"),
                &["successive", "succeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfromances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("whindow"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("apretiative"),
                &["appreciative"],
            ),
            (dictgen::InsensitiveStr::Ascii("neigbhor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("reconection"),
                &["reconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("vecotr"), &["vector"]),
            (
                dictgen::InsensitiveStr::Ascii("relisations"),
                &["realisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconects"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcably"), &["forcibly"]),
            (
                dictgen::InsensitiveStr::Ascii("conicidentally"),
                &["coincidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aprreciates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recusive"),
                &["recursive", "reclusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redepmtion"),
                &["redemption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerise"),
                &["parameterise"],
            ),
            (dictgen::InsensitiveStr::Ascii("addmins"), &["admins"]),
            (dictgen::InsensitiveStr::Ascii("peapel"), &["people"]),
            (dictgen::InsensitiveStr::Ascii("coverign"), &["covering"]),
            (dictgen::InsensitiveStr::Ascii("seinors"), &["seniors"]),
            (dictgen::InsensitiveStr::Ascii("adbuct"), &["abduct"]),
            (dictgen::InsensitiveStr::Ascii("moribdly"), &["morbidly"]),
            (
                dictgen::InsensitiveStr::Ascii("articuling"),
                &["articulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematicaly"),
                &["mathematically"],
            ),
            (dictgen::InsensitiveStr::Ascii("obnject"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("declerates"),
                &["decelerates", "declares"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restaurantes"),
                &["restaurants"],
            ),
            (dictgen::InsensitiveStr::Ascii("cauess"), &["causes"]),
            (dictgen::InsensitiveStr::Ascii("calaber"), &["caliber"]),
            (
                dictgen::InsensitiveStr::Ascii("chrisitanity"),
                &["christianity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attraktion"),
                &["attraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sementation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ovelapping"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dentities"),
                &["identities", "densities", "entities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functioality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangemenets"),
                &["arrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strin"),
                &["string", "strine", "stein"],
            ),
            (dictgen::InsensitiveStr::Ascii("quaternin"), &["quaternion"]),
            (
                dictgen::InsensitiveStr::Ascii("planeswlakers"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmbedding"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("interrpution"),
                &["interruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("assending"), &["ascending"]),
            (
                dictgen::InsensitiveStr::Ascii("wrokspaces"),
                &["workspaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizelled"), &["chiselled"]),
            (
                dictgen::InsensitiveStr::Ascii("dewrapping"),
                &["unwrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusotmers"), &["customers"]),
            (
                dictgen::InsensitiveStr::Ascii("annoyn"),
                &["annoy", "annoying"],
            ),
            (dictgen::InsensitiveStr::Ascii("propegate"), &["propagate"]),
            (
                dictgen::InsensitiveStr::Ascii("guarneteing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("palestein"), &["palestine"]),
            (
                dictgen::InsensitiveStr::Ascii("measuer"),
                &["measure", "measurer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditial"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("searh"), &["search"]),
            (dictgen::InsensitiveStr::Ascii("oredered"), &["ordered"]),
            (dictgen::InsensitiveStr::Ascii("currentpos"), &["currents"]),
            (dictgen::InsensitiveStr::Ascii("detatches"), &["detaches"]),
            (
                dictgen::InsensitiveStr::Ascii("provie"),
                &["provide", "prove"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilcked"), &["clicked"]),
            (dictgen::InsensitiveStr::Ascii("harwdare"), &["hardware"]),
            (dictgen::InsensitiveStr::Ascii("backpsace"), &["backspace"]),
            (
                dictgen::InsensitiveStr::Ascii("parliementary"),
                &["parliamentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("trippin"), &["tripping"]),
            (
                dictgen::InsensitiveStr::Ascii("quantitaive"),
                &["quantitative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partecipant"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("amgiguous"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("uncommited"),
                &["uncommitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("assurred"), &["assured"]),
            (dictgen::InsensitiveStr::Ascii("trasnport"), &["transport"]),
            (dictgen::InsensitiveStr::Ascii("godess"), &["goddess"]),
            (
                dictgen::InsensitiveStr::Ascii("supposingly"),
                &["supposedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("pluggins"), &["plugins"]),
            (dictgen::InsensitiveStr::Ascii("sourround"), &["surround"]),
            (dictgen::InsensitiveStr::Ascii("aprooved"), &["approved"]),
            (dictgen::InsensitiveStr::Ascii("principial"), &["principal"]),
            (dictgen::InsensitiveStr::Ascii("exturding"), &["extruding"]),
            (dictgen::InsensitiveStr::Ascii("opeatrors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("autogrpah"), &["autograph"]),
            (
                dictgen::InsensitiveStr::Ascii("constratints"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compadibility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accointants"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("appeded"), &["appended"]),
            (dictgen::InsensitiveStr::Ascii("serveces"), &["services"]),
            (
                dictgen::InsensitiveStr::Ascii("classificaitons"),
                &["classifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("statmenet"), &["statement"]),
            (
                dictgen::InsensitiveStr::Ascii("functining"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("intircate"), &["intricate"]),
            (dictgen::InsensitiveStr::Ascii("brugundy"), &["burgundy"]),
            (dictgen::InsensitiveStr::Ascii("replys"), &["replies"]),
            (
                dictgen::InsensitiveStr::Ascii("deslecting"),
                &["deselecting", "deflecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("snashots"), &["snapshots"]),
            (dictgen::InsensitiveStr::Ascii("beginng"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("preocessing"),
                &["processing", "preprocessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirting"), &["writing"]),
            (dictgen::InsensitiveStr::Ascii("eviserate"), &["eviscerate"]),
            (dictgen::InsensitiveStr::Ascii("denseley"), &["densely"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsitently"),
                &["inconsistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srceenshot"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("vesioning"), &["versioning"]),
            (dictgen::InsensitiveStr::Ascii("caspules"), &["capsules"]),
            (
                dictgen::InsensitiveStr::Ascii("hemmorhages"),
                &["hemorrhages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modificaction"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenattions"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aposthrophe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsessfull"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("acsending"), &["ascending"]),
            (dictgen::InsensitiveStr::Ascii("addressses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("comforable"),
                &["comfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proctecting"),
                &["protecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("additon"), &["addition"]),
            (
                dictgen::InsensitiveStr::Ascii("misognyists"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("reservered"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("exapand"), &["expand"]),
            (dictgen::InsensitiveStr::Ascii("sercret"), &["secret"]),
            (
                dictgen::InsensitiveStr::Ascii("newspapaers"),
                &["newspapers"],
            ),
            (dictgen::InsensitiveStr::Ascii("tutoriel"), &["tutorial"]),
            (
                dictgen::InsensitiveStr::Ascii("scandinavien"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhetorisch"), &["rhetoric"]),
            (dictgen::InsensitiveStr::Ascii("azmith"), &["azimuth"]),
            (
                dictgen::InsensitiveStr::Ascii("quiest"),
                &["quest", "quiet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulbearbility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definately"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecte"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("unreplacalbe"),
                &["unreplaceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntector"), &["connector"]),
            (dictgen::InsensitiveStr::Ascii("asbtract"), &["abstract"]),
            (dictgen::InsensitiveStr::Ascii("favorit"), &["favorite"]),
            (
                dictgen::InsensitiveStr::Ascii("freqencies"),
                &["frequencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soundtracs"),
                &["soundtracks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermient"),
                &["experiments", "experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("scantuary"), &["sanctuary"]),
            (
                dictgen::InsensitiveStr::Ascii("deperecates"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("snipet"), &["snippet"]),
            (dictgen::InsensitiveStr::Ascii("covenat"), &["covenant"]),
            (
                dictgen::InsensitiveStr::Ascii("proceedures"),
                &["procedures"],
            ),
            (dictgen::InsensitiveStr::Ascii("summmaries"), &["summaries"]),
            (dictgen::InsensitiveStr::Ascii("noramlly"), &["normally"]),
            (dictgen::InsensitiveStr::Ascii("sqaures"), &["squares"]),
            (dictgen::InsensitiveStr::Ascii("aggreator"), &["aggregator"]),
            (
                dictgen::InsensitiveStr::Ascii("applyicable"),
                &["applicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archieve"),
                &["achieve", "archive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificationn"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceedd"), &["conceded"]),
            (
                dictgen::InsensitiveStr::Ascii("fortunatly"),
                &["fortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pshycologically"),
                &["psychologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionaltiies"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rembembered"),
                &["remembered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tocuhscreen"),
                &["touchscreen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beaurocratical"),
                &["bureaucratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resembelance"),
                &["resemblance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formualted"),
                &["formulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminstratively"),
                &["administratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("examnple"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("safeing"), &["saving"]),
            (
                dictgen::InsensitiveStr::Ascii("specifi"),
                &["specific", "specify"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percession"),
                &["percussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("restrore"), &["restore"]),
            (
                dictgen::InsensitiveStr::Ascii("insuffcient"),
                &["insufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sexulaized"),
                &["sexualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cateogorical"),
                &["categorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsquent"), &["subsequent"]),
            (
                dictgen::InsensitiveStr::Ascii("shareholds"),
                &["shareholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("targted"), &["targeted"]),
            (dictgen::InsensitiveStr::Ascii("tacticaly"), &["tactically"]),
            (dictgen::InsensitiveStr::Ascii("ambiguos"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("reasonablly"),
                &["reasonably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ineffciently"),
                &["inefficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("varigated"), &["variegated"]),
            (
                dictgen::InsensitiveStr::Ascii("detertmines"),
                &["determines"],
            ),
            (dictgen::InsensitiveStr::Ascii("phoentic"), &["phonetic"]),
            (dictgen::InsensitiveStr::Ascii("claeners"), &["cleaners"]),
            (
                dictgen::InsensitiveStr::Ascii("recommanding"),
                &["recommending"],
            ),
            (dictgen::InsensitiveStr::Ascii("failrues"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("clyamore"), &["claymore"]),
            (
                dictgen::InsensitiveStr::Ascii("combintaions"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performence"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("socieites"), &["societies"]),
            (dictgen::InsensitiveStr::Ascii("extrated"), &["extracted"]),
            (dictgen::InsensitiveStr::Ascii("interewbs"), &["interwebs"]),
            (
                dictgen::InsensitiveStr::Ascii("repectable"),
                &["repeatable", "respectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("proocols"), &["protocols"]),
            (
                dictgen::InsensitiveStr::Ascii("silouwetes"),
                &["silhouettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcments"),
                &["announcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentato"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("oarticle"), &["particle"]),
            (
                dictgen::InsensitiveStr::Ascii("equivivalent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premission"),
                &["permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("snowflaek"), &["snowflake"]),
            (dictgen::InsensitiveStr::Ascii("requesed"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("aggraveted"),
                &["aggravated"],
            ),
            (dictgen::InsensitiveStr::Ascii("squashgin"), &["squashing"]),
            (dictgen::InsensitiveStr::Ascii("backsapce"), &["backspace"]),
            (dictgen::InsensitiveStr::Ascii("aack"), &["ack"]),
            (dictgen::InsensitiveStr::Ascii("campains"), &["campaigns"]),
            (dictgen::InsensitiveStr::Ascii("exapanding"), &["expanding"]),
            (dictgen::InsensitiveStr::Ascii("inabling"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("calibartion"),
                &["calibration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomadating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("horphan"), &["orphan"]),
            (
                dictgen::InsensitiveStr::Ascii("cheecks"),
                &["checks", "cheeks"],
            ),
            (dictgen::InsensitiveStr::Ascii("beligan"), &["belgian"]),
            (dictgen::InsensitiveStr::Ascii("relevent"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("directlly"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("individiually"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("touristey"), &["touristy"]),
            (dictgen::InsensitiveStr::Ascii("preferis"), &["prefers"]),
            (dictgen::InsensitiveStr::Ascii("tunnell"), &["tunnel"]),
            (dictgen::InsensitiveStr::Ascii("otherwuise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("alcohalics"),
                &["alcoholics"],
            ),
            (dictgen::InsensitiveStr::Ascii("shebeng"), &["shebang"]),
            (
                dictgen::InsensitiveStr::Ascii("reprsentations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("drwawing"), &["drawing"]),
            (dictgen::InsensitiveStr::Ascii("disirable"), &["desirable"]),
            (dictgen::InsensitiveStr::Ascii("obserer"), &["observer"]),
            (
                dictgen::InsensitiveStr::Ascii("discconencts"),
                &["disconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captializes"),
                &["capitalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessesarry"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("neckbreads"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternavtes"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legislatiors"),
                &["legislators"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrollbae"), &["scrollbar"]),
            (dictgen::InsensitiveStr::Ascii("minimium"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("opearations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("fitured"),
                &["figured", "featured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpindicular"),
                &["perpendicular"],
            ),
            (dictgen::InsensitiveStr::Ascii("saveing"), &["saving"]),
            (
                dictgen::InsensitiveStr::Ascii("contracter"),
                &["contractor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilavent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialitze"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("explanes"), &["explains"]),
            (dictgen::InsensitiveStr::Ascii("activy"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("ovelaps"), &["overlaps"]),
            (
                dictgen::InsensitiveStr::Ascii("teminations"),
                &["terminations"],
            ),
            (dictgen::InsensitiveStr::Ascii("covver"), &["cover"]),
            (
                dictgen::InsensitiveStr::Ascii("provisioing"),
                &["provisioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqiures"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("regularlises"),
                &["regularises"],
            ),
            (dictgen::InsensitiveStr::Ascii("montanta"), &["montana"]),
            (dictgen::InsensitiveStr::Ascii("huminoid"), &["humanoid"]),
            (dictgen::InsensitiveStr::Ascii("analitics"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("boundaryi"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("substitues"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whitspaces"),
                &["whitespace", "whitespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixeid"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("repliac"), &["replica"]),
            (
                dictgen::InsensitiveStr::Ascii("platfotmer"),
                &["platformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("ivocation"), &["invocation"]),
            (
                dictgen::InsensitiveStr::Ascii("adoloscent"),
                &["adolescent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassines"),
                &["assassins", "assassinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immortaliy"),
                &["immortality"],
            ),
            (dictgen::InsensitiveStr::Ascii("mige"), &["midge"]),
            (dictgen::InsensitiveStr::Ascii("argumnet"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("simiarity"), &["similarity"]),
            (dictgen::InsensitiveStr::Ascii("retinew"), &["retinue"]),
            (
                dictgen::InsensitiveStr::Ascii("akkumulative"),
                &["accumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("reslving"), &["resolving"]),
            (dictgen::InsensitiveStr::Ascii("otimally"), &["optimally"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptoin"),
                &["descriptions", "description"],
            ),
            (dictgen::InsensitiveStr::Ascii("britened"), &["brightened"]),
            (dictgen::InsensitiveStr::Ascii("minimsing"), &["minimising"]),
            (
                dictgen::InsensitiveStr::Ascii("aribtrarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("commiters"), &["committers"]),
            (dictgen::InsensitiveStr::Ascii("topoplogy"), &["topology"]),
            (dictgen::InsensitiveStr::Ascii("assfalts"), &["asphalts"]),
            (dictgen::InsensitiveStr::Ascii("melineum"), &["millennium"]),
            (
                dictgen::InsensitiveStr::Ascii("enchamtment"),
                &["enchantment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluclater"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("minerial"), &["mineral"]),
            (
                dictgen::InsensitiveStr::Ascii("suuporting"),
                &["supporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realtivity"),
                &["relativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("exepcted"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("additivies"),
                &["additive", "additives"],
            ),
            (dictgen::InsensitiveStr::Ascii("eihter"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("contestat"),
                &["contestants"],
            ),
            (dictgen::InsensitiveStr::Ascii("mellineums"), &["millennia"]),
            (
                dictgen::InsensitiveStr::Ascii("experement"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("convrets"), &["converts"]),
            (dictgen::InsensitiveStr::Ascii("cnfig"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("genralises"),
                &["generalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idealistisch"),
                &["idealistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlicitly"), &["explicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("transaprency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empahsizing"),
                &["emphasizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptionms"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("iraninas"), &["iranians"]),
            (dictgen::InsensitiveStr::Ascii("ressources"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("veriftying"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("poylmer"), &["polymer"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdentedly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("temprarary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("inflatie"), &["inflate"]),
            (dictgen::InsensitiveStr::Ascii("vicotrian"), &["victorian"]),
            (
                dictgen::InsensitiveStr::Ascii("multifuntion"),
                &["multifunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrinsincs"),
                &["intrinsics"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabably"), &["capably"]),
            (
                dictgen::InsensitiveStr::Ascii("identiviert"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructiin"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("exproting"), &["exporting"]),
            (
                dictgen::InsensitiveStr::Ascii("buildpackge"),
                &["buildpackage"],
            ),
            (dictgen::InsensitiveStr::Ascii("genearlly"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("consestently"),
                &["consistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("impicit"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("ussual"), &["usual"]),
            (
                dictgen::InsensitiveStr::Ascii("configirations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfortablity"),
                &["comfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("currrent"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("semseter"), &["semester"]),
            (dictgen::InsensitiveStr::Ascii("debouces"), &["debounces"]),
            (dictgen::InsensitiveStr::Ascii("longevitiy"), &["longevity"]),
            (
                dictgen::InsensitiveStr::Ascii("coflicting"),
                &["conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("charictor"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("missen"), &["mizzen"]),
            (
                dictgen::InsensitiveStr::Ascii("fanservise"),
                &["fanservice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coniderations"),
                &["considerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("regultor"), &["regulator"]),
            (dictgen::InsensitiveStr::Ascii("anothe"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("hirachy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("transparant"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legislacion"),
                &["legislation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obatinable"),
                &["obtainable"],
            ),
            (dictgen::InsensitiveStr::Ascii("causeed"), &["caused"]),
            (dictgen::InsensitiveStr::Ascii("enhanse"), &["enhances"]),
            (dictgen::InsensitiveStr::Ascii("podule"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("succcessful"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconetions"),
                &["reconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocorrrect"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("torwards"), &["towards"]),
            (dictgen::InsensitiveStr::Ascii("anouncer"), &["announcer"]),
            (
                dictgen::InsensitiveStr::Ascii("subconscous"),
                &["subconscious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabil"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromossome"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("meassures"), &["measures"]),
            (
                dictgen::InsensitiveStr::Ascii("sterilizin"),
                &["sterilizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("manfest"), &["manifest"]),
            (
                dictgen::InsensitiveStr::Ascii("subexpresssion"),
                &["subexpression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orienatation"),
                &["orientation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("halluciantion"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("welfair"), &["welfare"]),
            (dictgen::InsensitiveStr::Ascii("lightnig"), &["lightning"]),
            (
                dictgen::InsensitiveStr::Ascii("cannotes"),
                &["cannot", "connotes"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepect"), &["accept"]),
            (dictgen::InsensitiveStr::Ascii("udpater"), &["updater"]),
            (
                dictgen::InsensitiveStr::Ascii("uncommtited"),
                &["uncommitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulaion"), &["regulation"]),
            (dictgen::InsensitiveStr::Ascii("dreawn"), &["drawn"]),
            (dictgen::InsensitiveStr::Ascii("vriety"), &["variety"]),
            (
                dictgen::InsensitiveStr::Ascii("gernerating"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stringfied"),
                &["stringified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instabilty"),
                &["instability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("productioin"),
                &["productions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physicions"),
                &["physicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("termporal"), &["temporal"]),
            (dictgen::InsensitiveStr::Ascii("sholuld"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakground"), &["background"]),
            (dictgen::InsensitiveStr::Ascii("abudcting"), &["abducting"]),
            (dictgen::InsensitiveStr::Ascii("usernname"), &["username"]),
            (dictgen::InsensitiveStr::Ascii("redefinin"), &["redefining"]),
            (dictgen::InsensitiveStr::Ascii("warninfs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("reconenctions"),
                &["reconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("langualge"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("braista"), &["barista"]),
            (dictgen::InsensitiveStr::Ascii("prioritze"), &["prioritize"]),
            (dictgen::InsensitiveStr::Ascii("identtify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("attriute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("hardwdare"), &["hardware"]),
            (
                dictgen::InsensitiveStr::Ascii("tranalations"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaetive"), &["creative"]),
            (dictgen::InsensitiveStr::Ascii("conitinue"), &["continue"]),
            (dictgen::InsensitiveStr::Ascii("opem"), &["open"]),
            (
                dictgen::InsensitiveStr::Unicode("сontainors"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicaion"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("formulayic"), &["formulaic"]),
            (
                dictgen::InsensitiveStr::Ascii("handshkng"),
                &["handshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("peacefuly"), &["peacefully"]),
            (dictgen::InsensitiveStr::Ascii("recenty"), &["recently"]),
            (dictgen::InsensitiveStr::Ascii("ues"), &["use", "yes"]),
            (
                dictgen::InsensitiveStr::Ascii("conveinence"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialiezd"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bofere"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("concatene"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("globlaly"), &["globally"]),
            (
                dictgen::InsensitiveStr::Ascii("unsychronized"),
                &["unsynchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underlfowing"),
                &["underflowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intressted"),
                &["interested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrtibuted"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fraudalent"),
                &["fraudulent"],
            ),
            (dictgen::InsensitiveStr::Ascii("jalibreak"), &["jailbreak"]),
            (
                dictgen::InsensitiveStr::Ascii("reuplaodad"),
                &["reupload", "reuploaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterprets"), &["interprets"]),
            (dictgen::InsensitiveStr::Ascii("pretene"), &["pretense"]),
            (
                dictgen::InsensitiveStr::Ascii("struttural"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpahbetical"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifty"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("defailed"),
                &["detailed", "derailed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrainting"),
                &["constraining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compitance"),
                &["compliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catholocism"),
                &["catholicism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreggation"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opproximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("celing"), &["ceiling"]),
            (
                dictgen::InsensitiveStr::Ascii("reportresouces"),
                &["reportresources"],
            ),
            (dictgen::InsensitiveStr::Ascii("easilty"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("interator"), &["iterator"]),
            (
                dictgen::InsensitiveStr::Ascii("unitilizing"),
                &["uninitializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cauhgt"), &["caught"]),
            (dictgen::InsensitiveStr::Ascii("ccahe"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("silhouete"), &["silhouette"]),
            (
                dictgen::InsensitiveStr::Ascii("configurtoin"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capcaitance"),
                &["capacitance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrantied"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lannistars"),
                &["lannisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocilating"),
                &["oscillating"],
            ),
            (dictgen::InsensitiveStr::Ascii("billboars"), &["billboards"]),
            (dictgen::InsensitiveStr::Ascii("secue"), &["secure"]),
            (
                dictgen::InsensitiveStr::Ascii("assmptions"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tamplating"),
                &["templating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presenation"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("navigtion"), &["navigation"]),
            (
                dictgen::InsensitiveStr::Ascii("wallpappers"),
                &["wallpapers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aknowledging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiability"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcouments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("suprizingly"),
                &["surprisingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensionned"),
                &["dimensioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("explitict"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("relocaition"),
                &["relocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confriming"),
                &["confirming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demesticated"),
                &["domesticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("berekley"), &["berkeley"]),
            (
                dictgen::InsensitiveStr::Ascii("reorginized"),
                &["reorganized"],
            ),
            (dictgen::InsensitiveStr::Ascii("langhts"), &["lengths"]),
            (dictgen::InsensitiveStr::Ascii("hanbdle"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("createors"), &["creators"]),
            (dictgen::InsensitiveStr::Ascii("capitas"), &["capitals"]),
            (
                dictgen::InsensitiveStr::Ascii("retransfered"),
                &["retransferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("conpared"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("indepenendently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastropically"),
                &["catastrophically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precendencies"),
                &["precedences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmentv"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("ditionary"), &["dictionary"]),
            (dictgen::InsensitiveStr::Ascii("imparied"), &["impaired"]),
            (
                dictgen::InsensitiveStr::Ascii("cerrebrally"),
                &["cerebrally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sevrity"), &["severity"]),
            (
                dictgen::InsensitiveStr::Ascii("paramterizations"),
                &["parametrizations", "parameterizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrater"),
                &[
                    "demonstrator",
                    "demonstrated",
                    "demonstrates",
                    "demonstrate",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("persoanl"), &["personal"]),
            (dictgen::InsensitiveStr::Ascii("itnerpret"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("preservaton"),
                &["preservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("respponse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("acitvate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("retroactivly"),
                &["retroactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimmentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("execueted"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("chuncksize"), &["chunksize"]),
            (dictgen::InsensitiveStr::Ascii("notorized"), &["notarized"]),
            (dictgen::InsensitiveStr::Ascii("lanuch"), &["launch"]),
            (
                dictgen::InsensitiveStr::Ascii("sements"),
                &["cements", "segments"],
            ),
            (dictgen::InsensitiveStr::Ascii("acitively"), &["actively"]),
            (
                dictgen::InsensitiveStr::Ascii("economiclly"),
                &["economically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exspecially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("conainted"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("reffer"),
                &["refer", "reefer"],
            ),
            (dictgen::InsensitiveStr::Ascii("questioed"), &["questioned"]),
            (
                dictgen::InsensitiveStr::Ascii("incremenetd"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referencin"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminante"),
                &["discriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kwnoledgeable"),
                &["knowledgeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoization"),
                &["authorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("swapiness"), &["swappiness"]),
            (dictgen::InsensitiveStr::Ascii("receeding"), &["receding"]),
            (dictgen::InsensitiveStr::Ascii("expers"), &["express"]),
            (
                dictgen::InsensitiveStr::Ascii("intialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("lincence"), &["licence"]),
            (dictgen::InsensitiveStr::Ascii("migh"), &["might"]),
            (dictgen::InsensitiveStr::Ascii("meoldies"), &["melodies"]),
            (
                dictgen::InsensitiveStr::Ascii("americams"),
                &["americas", "americans"],
            ),
            (dictgen::InsensitiveStr::Ascii("parcer"), &["parser"]),
            (dictgen::InsensitiveStr::Ascii("structurs"), &["structures"]),
            (dictgen::InsensitiveStr::Ascii("availlable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("intitals"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("apointment"),
                &["appointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheepherd"), &["shepherd"]),
            (dictgen::InsensitiveStr::Ascii("maxosx"), &["macosx"]),
            (dictgen::InsensitiveStr::Ascii("garantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("needes"),
                &["needs", "needed", "needles"],
            ),
            (dictgen::InsensitiveStr::Ascii("symobolic"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("doublelit"), &["doublelift"]),
            (dictgen::InsensitiveStr::Ascii("expetions"), &["exceptions"]),
            (dictgen::InsensitiveStr::Ascii("minimial"), &["minimal"]),
            (
                dictgen::InsensitiveStr::Ascii("formidible"),
                &["formidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facilitait"),
                &["facilitate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consultent"),
                &["consultant"],
            ),
            (dictgen::InsensitiveStr::Ascii("coexinsts"), &["coexists"]),
            (dictgen::InsensitiveStr::Ascii("punshier"), &["punisher"]),
            (dictgen::InsensitiveStr::Ascii("nottaions"), &["notations"]),
            (
                dictgen::InsensitiveStr::Ascii("distorsions"),
                &["distortions"],
            ),
            (dictgen::InsensitiveStr::Ascii("basebame"), &["basename"]),
            (dictgen::InsensitiveStr::Ascii("pensinula"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("inproovment"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corpration"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autherises"),
                &["authorises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downtokers"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placeholer"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellegance"), &["elegance"]),
            (
                dictgen::InsensitiveStr::Ascii("expedetion"),
                &["expedition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stanardised"),
                &["standardised"],
            ),
            (dictgen::InsensitiveStr::Ascii("programmme"), &["programme"]),
            (
                dictgen::InsensitiveStr::Ascii("pretecting"),
                &["protecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("callcack"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("antarcitc"), &["antarctic"]),
            (
                dictgen::InsensitiveStr::Ascii("unnaceptably"),
                &["unacceptably"],
            ),
            (dictgen::InsensitiveStr::Ascii("presistion"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecessairy"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juse"),
                &["just", "juice", "jude", "june"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrived"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("seperatley"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("micosoft"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("augmnetation"),
                &["augmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecntration"),
                &["concentrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expreesion"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutipl"),
                &["multiple", "multiply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleratie"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("querstionnaires"),
                &["questionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("memebrof"), &["memberof"]),
            (
                dictgen::InsensitiveStr::Ascii("encarceration"),
                &["incarceration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facilitiates"),
                &["facilitates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onomonopea"),
                &["onomatopoeia"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpapted"), &["adapted"]),
            (
                dictgen::InsensitiveStr::Ascii("particulalry"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("spagnum"), &["sphagnum"]),
            (
                dictgen::InsensitiveStr::Ascii("narcissisist"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjectivety"),
                &["subjectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("blopgpot"), &["blogpost"]),
            (dictgen::InsensitiveStr::Ascii("notbeooks"), &["notebooks"]),
            (dictgen::InsensitiveStr::Ascii("deteceted"), &["detected"]),
            (
                dictgen::InsensitiveStr::Ascii("congresssman"),
                &["congressman"],
            ),
            (dictgen::InsensitiveStr::Ascii("speherical"), &["spherical"]),
            (dictgen::InsensitiveStr::Ascii("quesets"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("registoring"),
                &["registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaptate"), &["adapt"]),
            (dictgen::InsensitiveStr::Ascii("boofcamp"), &["bootcamp"]),
            (
                dictgen::InsensitiveStr::Ascii("photograhy"),
                &["photography"],
            ),
            (dictgen::InsensitiveStr::Ascii("caceling"), &["canceling"]),
            (dictgen::InsensitiveStr::Ascii("unbounaded"), &["unbounded"]),
            (
                dictgen::InsensitiveStr::Ascii("configuation"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrijske"),
                &["industries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predessecors"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeeficient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("searchs"), &["searches"]),
            (
                dictgen::InsensitiveStr::Ascii("authentation"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuitng"),
                &["distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("segault"), &["segfault"]),
            (
                dictgen::InsensitiveStr::Ascii("multilangual"),
                &["multilingual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfeccion"),
                &["perfection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionaly"),
                &["professionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecentages"),
                &["percentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("zomebie"), &["zombie"]),
            (dictgen::InsensitiveStr::Ascii("miltiples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("organische"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("elemt"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("philosophios"),
                &["philosophies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconstency"),
                &["inconsistency", "inconstancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatted"), &["operated"]),
            (dictgen::InsensitiveStr::Ascii("loacte"), &["locate"]),
            (dictgen::InsensitiveStr::Ascii("diagrama"), &["diagram"]),
            (dictgen::InsensitiveStr::Ascii("reuqires"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactureds"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atachments"),
                &["attachments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bottlnecks"),
                &["bottlenecks"],
            ),
            (dictgen::InsensitiveStr::Ascii("numbbering"), &["numbering"]),
            (dictgen::InsensitiveStr::Ascii("chidlish"), &["childish"]),
            (dictgen::InsensitiveStr::Ascii("candiadate"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("vegatarian"),
                &["vegetarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("creatation"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("analyseras"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasuticals"),
                &["pharmaceuticals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeatadly"),
                &["repeatedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ironamn"), &["ironman"]),
            (dictgen::InsensitiveStr::Ascii("inlcludes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("representitive"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sapphirre"), &["sapphire"]),
            (dictgen::InsensitiveStr::Ascii("possesess"), &["possesses"]),
            (
                dictgen::InsensitiveStr::Ascii("invisibiltiy"),
                &["invisibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvient"),
                &["inconvenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indentifing"),
                &["identifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afterwords"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecify"),
                &["specify", "pacify"],
            ),
            (dictgen::InsensitiveStr::Ascii("chunc"), &["chunk"]),
            (
                dictgen::InsensitiveStr::Ascii("sympathethic"),
                &["sympathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgruonds"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecessar"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("proabably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("overpooling"),
                &["overlooking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifikation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanatsies"), &["fantasies"]),
            (dictgen::InsensitiveStr::Ascii("automatied"), &["automate"]),
            (
                dictgen::InsensitiveStr::Ascii("exects"),
                &["expects", "ejects", "exerts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algotighms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccommendation"),
                &["recommendation"],
            ),
            (dictgen::InsensitiveStr::Ascii("regsiters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("combininator"),
                &["combinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reolve"),
                &["resolve", "revolve"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptoinal"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stegnographic"),
                &["steganographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("splel"), &["spell", "spiel"]),
            (dictgen::InsensitiveStr::Ascii("spefiific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferencce"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("regxp"), &["regexp"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplacation"),
                &["deduplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itermediate"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("restatin"), &["restating"]),
            (dictgen::InsensitiveStr::Ascii("fasicst"), &["fascist"]),
            (dictgen::InsensitiveStr::Ascii("blockchan"), &["blockchain"]),
            (
                dictgen::InsensitiveStr::Ascii("remarkablly"),
                &["remarkably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customizible"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorrd"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("templarios"), &["templars"]),
            (dictgen::InsensitiveStr::Ascii("curent"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("valuble"), &["valuable"]),
            (
                dictgen::InsensitiveStr::Ascii("corresspondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devleopment"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagnosied"), &["diagnose"]),
            (
                dictgen::InsensitiveStr::Ascii("paranornal"),
                &["paranormal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creationkit"),
                &["creationist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatanates"),
                &["concatenates"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneected"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("expriment"), &["experiment"]),
            (dictgen::InsensitiveStr::Ascii("spectrolab"), &["spectral"]),
            (dictgen::InsensitiveStr::Ascii("coallesced"), &["coalesced"]),
            (
                dictgen::InsensitiveStr::Ascii("facilititate"),
                &["facilitate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("micorphones"),
                &["microphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("endpdoints"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("respectivly"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("euqally"), &["equally"]),
            (dictgen::InsensitiveStr::Ascii("encryted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("formidabelt"),
                &["formidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benelovent"),
                &["benevolent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declaracion"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficiics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("capitilise"),
                &["capitalise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservativeky"),
                &["conservatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("ntegrate"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("alues"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("sensivity"),
                &["sensitivity", "sensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpexcted"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("featre"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("formattters"),
                &["formatters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("codde"),
                &["code", "coded", "coddle"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecause"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("wastefullness"),
                &["wastefulness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontaniously"),
                &["spontaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoirthms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("thant"), &["than", "that"]),
            (dictgen::InsensitiveStr::Ascii("endfi"), &["endif"]),
            (dictgen::InsensitiveStr::Ascii("baptims"), &["baptism"]),
            (dictgen::InsensitiveStr::Ascii("lasgana"), &["lasagna"]),
            (dictgen::InsensitiveStr::Ascii("resulves"), &["resolves"]),
            (dictgen::InsensitiveStr::Ascii("exampele"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("indenpendent"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("screeen"), &["screen"]),
            (dictgen::InsensitiveStr::Ascii("sylibol"), &["syllable"]),
            (
                dictgen::InsensitiveStr::Ascii("tramsmission"),
                &["transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("enries"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("bloock"), &["block"]),
            (dictgen::InsensitiveStr::Ascii("freqency"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("conpress"), &["compress"]),
            (dictgen::InsensitiveStr::Ascii("imlicit"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("appearnce"), &["appearance"]),
            (dictgen::InsensitiveStr::Ascii("dissrupted"), &["disrupted"]),
            (dictgen::InsensitiveStr::Ascii("complile"), &["compile"]),
            (dictgen::InsensitiveStr::Ascii("expception"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("representativas"),
                &["representatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contine"),
                &["continue", "contain"],
            ),
            (dictgen::InsensitiveStr::Ascii("attibute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("repayemnt"), &["repayment"]),
            (dictgen::InsensitiveStr::Ascii("reprsents"), &["represents"]),
            (dictgen::InsensitiveStr::Ascii("tomor"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("symoblically"),
                &["symbolically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighorhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficeid"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("congfigure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("spcifying"), &["specifying"]),
            (
                dictgen::InsensitiveStr::Ascii("inflatabale"),
                &["inflatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benefecial"),
                &["beneficial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanicsms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatief"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delporable"),
                &["deplorable"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppored"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("unintialised"),
                &["uninitialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimliarly"),
                &["dissimilarly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolagised"),
                &["apologised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequneces"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimziation"),
                &["optimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("paranoida"), &["paranoia"]),
            (dictgen::InsensitiveStr::Ascii("obsereved"), &["observed"]),
            (dictgen::InsensitiveStr::Ascii("newcastel"), &["newcastle"]),
            (
                dictgen::InsensitiveStr::Ascii("hotizontal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmiate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortihms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoatiated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacemnt"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirucular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("instatances"),
                &["instances"],
            ),
            (dictgen::InsensitiveStr::Ascii("releiver"), &["reliever"]),
            (dictgen::InsensitiveStr::Ascii("annnotator"), &["annotator"]),
            (
                dictgen::InsensitiveStr::Ascii("correspendence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerised"),
                &["parameterised"],
            ),
            (dictgen::InsensitiveStr::Ascii("protocoles"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("deveolp"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("cahotic"), &["chaotic"]),
            (dictgen::InsensitiveStr::Ascii("titile"), &["title"]),
            (
                dictgen::InsensitiveStr::Ascii("offets"),
                &["offsets", "offers"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenticles"), &["tentacles"]),
            (dictgen::InsensitiveStr::Ascii("perfecctly"), &["perfectly"]),
            (
                dictgen::InsensitiveStr::Ascii("unbraikable"),
                &["unbreakable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profesion"),
                &["profusion", "profession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialiazing"),
                &["serializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associacted"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("remotly"), &["remotely"]),
            (dictgen::InsensitiveStr::Ascii("statisitic"), &["statistic"]),
            (dictgen::InsensitiveStr::Ascii("meida"), &["media"]),
            (dictgen::InsensitiveStr::Ascii("adrerssed"), &["addressed"]),
            (
                dictgen::InsensitiveStr::Ascii("configursations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limtations"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("evicition"), &["eviction"]),
            (dictgen::InsensitiveStr::Ascii("bounraies"), &["boundaries"]),
            (
                dictgen::InsensitiveStr::Ascii("qeustioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("devolop"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("canaceled"), &["canceled"]),
            (dictgen::InsensitiveStr::Ascii("quatnize"), &["quantize"]),
            (
                dictgen::InsensitiveStr::Ascii("standarisation"),
                &["standardisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambadexterouseness"),
                &["ambidextrousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmnetal"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persistant"),
                &["persistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teoretical"),
                &["theoretical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuares"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fetures"),
                &["features", "futures"],
            ),
            (dictgen::InsensitiveStr::Ascii("techer"), &["teacher"]),
            (dictgen::InsensitiveStr::Ascii("requrests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("complexe"), &["complex"]),
            (
                dictgen::InsensitiveStr::Ascii("incease"),
                &["increase", "incense"],
            ),
            (dictgen::InsensitiveStr::Ascii("dignostic"), &["diagnostic"]),
            (dictgen::InsensitiveStr::Ascii("bethseda"), &["bethesda"]),
            (dictgen::InsensitiveStr::Ascii("dictinary"), &["dictionary"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitplies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invariabley"),
                &["invariably"],
            ),
            (dictgen::InsensitiveStr::Ascii("surfaec"), &["surface"]),
            (dictgen::InsensitiveStr::Ascii("minmize"), &["minimize"]),
            (dictgen::InsensitiveStr::Ascii("respource"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("vertecies"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("protocolos"), &["protocols"]),
            (
                dictgen::InsensitiveStr::Ascii("perfact"),
                &["prefect", "perfect"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundares"), &["boundaries"]),
            (
                dictgen::InsensitiveStr::Ascii("multlication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("comiplers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("asynschronly"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reigstries"),
                &["registries"],
            ),
            (dictgen::InsensitiveStr::Ascii("thurrow"), &["thorough"]),
            (
                dictgen::InsensitiveStr::Ascii("perferential"),
                &["preferential"],
            ),
            (dictgen::InsensitiveStr::Ascii("taiwanesse"), &["taiwanese"]),
            (
                dictgen::InsensitiveStr::Ascii("factorizaiton"),
                &["factorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("misformed"), &["malformed"]),
            (
                dictgen::InsensitiveStr::Ascii("practitionner"),
                &["practitioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscirber"),
                &["subscribers", "subscriber"],
            ),
            (dictgen::InsensitiveStr::Ascii("activstes"), &["activates"]),
            (dictgen::InsensitiveStr::Ascii("aovoided"), &["avoided"]),
            (
                dictgen::InsensitiveStr::Ascii("indepenence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("analitycs"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("uptdate"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("paranoicas"), &["paranoia"]),
            (dictgen::InsensitiveStr::Ascii("funktions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("interplotated"),
                &["interpolated"],
            ),
            (dictgen::InsensitiveStr::Ascii("himselv"), &["himself"]),
            (dictgen::InsensitiveStr::Ascii("matchies"), &["matches"]),
            (dictgen::InsensitiveStr::Ascii("cannisters"), &["canisters"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibilitiy"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("curreent"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("compabilities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devconainer"),
                &["devcontainer"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaied"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("addiditive"), &["additive"]),
            (
                dictgen::InsensitiveStr::Ascii("ansyert"),
                &["answer", "answered"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenitcal"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("behvior"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("adustment"), &["adjustment"]),
            (
                dictgen::InsensitiveStr::Ascii("compositore"),
                &["composite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasactions"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offsers"),
                &["offers", "offsets"],
            ),
            (dictgen::InsensitiveStr::Ascii("delevoped"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("maintaine"),
                &["maintain", "maintained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reseletion"),
                &["reselection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenciating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("succecor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("represensts"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalses"),
                &["capitalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("hardward"), &["hardware"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnxostic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconf"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("portait"), &["portrait"]),
            (
                dictgen::InsensitiveStr::Ascii("parachutage"),
                &["parachute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregorative"),
                &["prerogative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categorisin"),
                &["categorising"],
            ),
            (dictgen::InsensitiveStr::Ascii("matresses"), &["mattresses"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoemanses"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("slach"), &["slash"]),
            (
                dictgen::InsensitiveStr::Ascii("iteracted"),
                &["interacted", "iterated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferente"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furhermore"),
                &["furthermore"],
            ),
            (dictgen::InsensitiveStr::Ascii("precisley"), &["precisely"]),
            (
                dictgen::InsensitiveStr::Ascii("ochestrating"),
                &["orchestrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresspondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphatized"),
                &["emphasized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capcaitors"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enclycopedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("capsumel"), &["capsule"]),
            (
                dictgen::InsensitiveStr::Ascii("checkeerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("dismisals"), &["dismissal"]),
            (dictgen::InsensitiveStr::Ascii("sesnor"), &["sensor"]),
            (dictgen::InsensitiveStr::Ascii("overheared"), &["overhead"]),
            (
                dictgen::InsensitiveStr::Ascii("astranomical"),
                &["astronomical"],
            ),
            (dictgen::InsensitiveStr::Ascii("buda"), &["buddha"]),
            (dictgen::InsensitiveStr::Ascii("ofuscated"), &["obfuscated"]),
            (dictgen::InsensitiveStr::Ascii("tyuphos"), &["typhus"]),
            (dictgen::InsensitiveStr::Ascii("proberly"), &["properly"]),
            (dictgen::InsensitiveStr::Ascii("belyed"), &["belied"]),
            (dictgen::InsensitiveStr::Ascii("questons"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("garanty"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("sketpicism"),
                &["skepticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("infared"), &["infrared"]),
            (
                dictgen::InsensitiveStr::Ascii("accelleration"),
                &["acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallizable"),
                &["parallelizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("analzyer"), &["analyzer"]),
            (
                dictgen::InsensitiveStr::Ascii("infrormation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("fadind"), &["fading"]),
            (
                dictgen::InsensitiveStr::Ascii("folllows"),
                &["follows", "followings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucesssful"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("bargaing"), &["bargaining"]),
            (
                dictgen::InsensitiveStr::Ascii("addressings"),
                &["addressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("requird"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("defenesman"),
                &["defenseman"],
            ),
            (dictgen::InsensitiveStr::Ascii("satisify"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("backgroud"), &["background"]),
            (dictgen::InsensitiveStr::Ascii("anyhting"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("infreqeuntcy"),
                &["infrequency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreggators"),
                &["aggregators"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusotmer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("unilateraly"),
                &["unilaterally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucessful"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("valeus"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("combinaisons"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ioclt"), &["ioctl"]),
            (dictgen::InsensitiveStr::Ascii("frostig"), &["frosting"]),
            (dictgen::InsensitiveStr::Ascii("resevoir"), &["reservoir"]),
            (
                dictgen::InsensitiveStr::Ascii("mailiciously"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constributor"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhapsodomy"), &["rhapsody"]),
            (
                dictgen::InsensitiveStr::Ascii("intermittment"),
                &["intermittent"],
            ),
            (dictgen::InsensitiveStr::Ascii("susinctly"), &["succinctly"]),
            (
                dictgen::InsensitiveStr::Ascii("flacouring"),
                &["flavouring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dditionally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastructuur"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchamarks"),
                &["benchmarks"],
            ),
            (dictgen::InsensitiveStr::Ascii("notitifers"), &["notifiers"]),
            (dictgen::InsensitiveStr::Ascii("compainion"), &["companion"]),
            (
                dictgen::InsensitiveStr::Ascii("akumulates"),
                &["accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbaled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("communcating"),
                &["communicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialzied"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cllosely"), &["closely"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("nickanme"), &["nickname"]),
            (dictgen::InsensitiveStr::Ascii("traked"), &["tracked"]),
            (dictgen::InsensitiveStr::Ascii("braverly"), &["bravery"]),
            (dictgen::InsensitiveStr::Ascii("legths"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("chanpionship"),
                &["championship"],
            ),
            (dictgen::InsensitiveStr::Ascii("timeserie"), &["timeseries"]),
            (dictgen::InsensitiveStr::Ascii("humantiy"), &["humanity"]),
            (dictgen::InsensitiveStr::Ascii("numberal"), &["numeral"]),
            (
                dictgen::InsensitiveStr::Ascii("pususading"),
                &["persuading"],
            ),
            (dictgen::InsensitiveStr::Ascii("notifty"), &["notify"]),
            (
                dictgen::InsensitiveStr::Ascii("negoatiable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("contigues"), &["contiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("accepetably"),
                &["acceptably"],
            ),
            (dictgen::InsensitiveStr::Ascii("centipeed"), &["centipede"]),
            (
                dictgen::InsensitiveStr::Ascii("introdiced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condidates"),
                &["candidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resgistrations"),
                &["registrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("accidant"), &["accident"]),
            (
                dictgen::InsensitiveStr::Ascii("scandanivian"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("europian"), &["european"]),
            (
                dictgen::InsensitiveStr::Ascii("ramificaitons"),
                &["ramifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerimented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("itegral"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("monolouge"), &["monologue"]),
            (dictgen::InsensitiveStr::Ascii("oher"), &["her", "other"]),
            (
                dictgen::InsensitiveStr::Ascii("defenderlos"),
                &["defenders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwwin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("reorer"), &["reorder"]),
            (dictgen::InsensitiveStr::Ascii("analogeous"), &["analogous"]),
            (dictgen::InsensitiveStr::Ascii("mimimizes"), &["minimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("atructural"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improbements"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("vioalte"), &["violate"]),
            (
                dictgen::InsensitiveStr::Ascii("fontains"),
                &["fountains", "contains"],
            ),
            (dictgen::InsensitiveStr::Ascii("messgaing"), &["messaging"]),
            (dictgen::InsensitiveStr::Ascii("eyerone"), &["everyone"]),
            (dictgen::InsensitiveStr::Ascii("erlier"), &["earlier"]),
            (dictgen::InsensitiveStr::Ascii("booee"), &["buoy"]),
            (dictgen::InsensitiveStr::Ascii("indrect"), &["indirect"]),
            (
                dictgen::InsensitiveStr::Ascii("implememts"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viusalized"),
                &["visualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sagitarius"),
                &["sagittarius"],
            ),
            (dictgen::InsensitiveStr::Ascii("floridia"), &["florida"]),
            (dictgen::InsensitiveStr::Ascii("cehcksums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("dictonary"), &["dictionary"]),
            (
                dictgen::InsensitiveStr::Ascii("heavywieghts"),
                &["heavyweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desychronized"),
                &["desynchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dermatoligist"),
                &["dermatologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoaiations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regsitration"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterize"),
                &["parameterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("trotzkism"), &["trotskyism"]),
            (
                dictgen::InsensitiveStr::Ascii("inclutions"),
                &["inclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("boomarks"), &["bookmarks"]),
            (dictgen::InsensitiveStr::Ascii("imorting"), &["importing"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnted"), &["commented"]),
            (
                dictgen::InsensitiveStr::Ascii("distuingished"),
                &["distinguished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappline"),
                &["discipline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("especting"),
                &["expecting", "respecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("norification"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearators"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("ordner"), &["order"]),
            (dictgen::InsensitiveStr::Ascii("camperos"), &["campers"]),
            (
                dictgen::InsensitiveStr::Ascii("charactertistics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registartion"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasparent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approopriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompotent"),
                &["incompetent"],
            ),
            (dictgen::InsensitiveStr::Ascii("planation"), &["plantation"]),
            (
                dictgen::InsensitiveStr::Ascii("informatiei"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articualted"),
                &["articulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("disregrad"), &["disregard"]),
            (dictgen::InsensitiveStr::Ascii("nuisans"), &["nuisance"]),
            (dictgen::InsensitiveStr::Ascii("ivokes"), &["invokes"]),
            (
                dictgen::InsensitiveStr::Ascii("chaning"),
                &["chaining", "changing"],
            ),
            (dictgen::InsensitiveStr::Ascii("myhtical"), &["mythical"]),
            (dictgen::InsensitiveStr::Ascii("everage"), &["average"]),
            (
                dictgen::InsensitiveStr::Ascii("essencially"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paravirutalisation"),
                &["paravirtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceration"), &["creation"]),
            (
                dictgen::InsensitiveStr::Ascii("phillosophically"),
                &["philosophically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preformance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("desscribe"), &["describe"]),
            (
                dictgen::InsensitiveStr::Ascii("liberoffice"),
                &["libreoffice"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfified"), &["verified"]),
            (dictgen::InsensitiveStr::Ascii("levle"), &["level"]),
            (
                dictgen::InsensitiveStr::Ascii("microtransacions"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabniet"), &["cabinet"]),
            (dictgen::InsensitiveStr::Ascii("starigth"), &["straight"]),
            (dictgen::InsensitiveStr::Ascii("staurates"), &["saturates"]),
            (dictgen::InsensitiveStr::Ascii("asertive"), &["assertive"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiifeirs"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("peristed"), &["persisted"]),
            (dictgen::InsensitiveStr::Ascii("sistemed"), &["systemed"]),
            (dictgen::InsensitiveStr::Ascii("contextes"), &["contexts"]),
            (dictgen::InsensitiveStr::Ascii("aplyed"), &["applied"]),
            (
                dictgen::InsensitiveStr::Ascii("assyemetric"),
                &["asymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("qust"), &["quest"]),
            (dictgen::InsensitiveStr::Ascii("demorcracy"), &["democracy"]),
            (dictgen::InsensitiveStr::Ascii("vampiers"), &["vampires"]),
            (
                dictgen::InsensitiveStr::Ascii("respectible"),
                &["respectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiteis"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("altruistric"),
                &["altruistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenedents"),
                &["dependents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collender"),
                &["cullender", "colander"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulataniously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("extration"), &["extraction"]),
            (
                dictgen::InsensitiveStr::Ascii("grammaticly"),
                &["grammatical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retransfering"),
                &["retransferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("dumbbels"), &["dumbbells"]),
            (dictgen::InsensitiveStr::Ascii("initilals"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("etsbalishment"),
                &["establishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchangin"), &["exchanging"]),
            (dictgen::InsensitiveStr::Ascii("flexibile"), &["flexible"]),
            (
                dictgen::InsensitiveStr::Ascii("deteted"),
                &["detected", "deleted"],
            ),
            (dictgen::InsensitiveStr::Ascii("brievely"), &["briefly"]),
            (dictgen::InsensitiveStr::Ascii("amethsyt"), &["amethyst"]),
            (dictgen::InsensitiveStr::Ascii("artifactos"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("serialisin"),
                &["serialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("helemts"), &["helmets"]),
            (dictgen::InsensitiveStr::Ascii("wheareas"), &["whereas"]),
            (
                dictgen::InsensitiveStr::Ascii("detech"),
                &["detach", "detect"],
            ),
            (dictgen::InsensitiveStr::Ascii("delcaring"), &["declaring"]),
            (
                dictgen::InsensitiveStr::Ascii("dispensories"),
                &["dispensaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transators"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("affaris"), &["affairs"]),
            (
                dictgen::InsensitiveStr::Ascii("comman"),
                &["command", "common", "comma", "commas"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envionment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("byt"), &["byte"]),
            (dictgen::InsensitiveStr::Ascii("youngents"), &["youngest"]),
            (dictgen::InsensitiveStr::Ascii("thhank"), &["thank"]),
            (
                dictgen::InsensitiveStr::Ascii("implementasi"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comutation"),
                &["commutation", "computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deposin"), &["deposing"]),
            (
                dictgen::InsensitiveStr::Ascii("declators"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparissions"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertional"),
                &["proportional"],
            ),
            (dictgen::InsensitiveStr::Ascii("manatory"), &["mandatory"]),
            (dictgen::InsensitiveStr::Ascii("optiion"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("substargets"),
                &["subtargets"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoltue"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpet"),
                &["misinterpret"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmimate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ostensibily"),
                &["ostensibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimiately"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranzporter"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("orginize"), &["organize"]),
            (dictgen::InsensitiveStr::Ascii("shwoing"), &["showing"]),
            (dictgen::InsensitiveStr::Ascii("oxigen"), &["oxygen"]),
            (dictgen::InsensitiveStr::Ascii("oeprating"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("controllore"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downroaded"),
                &["downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("pericings"), &["piercings"]),
            (
                dictgen::InsensitiveStr::Ascii("higlighting"),
                &["highlighting"],
            ),
            (dictgen::InsensitiveStr::Ascii("malins"), &["maligns"]),
            (dictgen::InsensitiveStr::Ascii("stabilzer"), &["stabilizer"]),
            (dictgen::InsensitiveStr::Ascii("securrly"), &["securely"]),
            (dictgen::InsensitiveStr::Ascii("ealrier"), &["earlier"]),
            (
                dictgen::InsensitiveStr::Ascii("miltiplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloted"), &["bloated"]),
            (dictgen::InsensitiveStr::Ascii("constrint"), &["constraint"]),
            (dictgen::InsensitiveStr::Ascii("emmitting"), &["emitting"]),
            (
                dictgen::InsensitiveStr::Ascii("aleviating"),
                &["alleviating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generat"),
                &["generate", "general"],
            ),
            (dictgen::InsensitiveStr::Ascii("transsend"), &["transcend"]),
            (dictgen::InsensitiveStr::Ascii("povide"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("soket"), &["socket"]),
            (
                dictgen::InsensitiveStr::Ascii("repeteadly"),
                &["repeatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homeopahty"),
                &["homeopathy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("toubleshoot"),
                &["troubleshoot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculaion"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("elementry"), &["elementary"]),
            (
                dictgen::InsensitiveStr::Ascii("distributons"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palestinain"),
                &["palestinians"],
            ),
            (dictgen::InsensitiveStr::Ascii("optening"), &["opening"]),
            (dictgen::InsensitiveStr::Ascii("molestare"), &["molester"]),
            (dictgen::InsensitiveStr::Ascii("oerations"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("discourarged"),
                &["discouraged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremetric"),
                &["parametric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termporarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("malaira"), &["malaria"]),
            (
                dictgen::InsensitiveStr::Ascii("declerating"),
                &["decelerating", "declaring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genereation"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissamblies"),
                &["disassemblies"],
            ),
            (dictgen::InsensitiveStr::Ascii("everywher"), &["everywhere"]),
            (
                dictgen::InsensitiveStr::Ascii("tranformations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuing"), &["confusing"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubstansiated"),
                &["unsubstantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addministrative"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vocabualries"),
                &["vocabularies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mircowaves"),
                &["microwaves"],
            ),
            (dictgen::InsensitiveStr::Ascii("erroneus"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("cannonically"),
                &["canonically"],
            ),
            (dictgen::InsensitiveStr::Ascii("guareente"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("distiguished"),
                &["distinguished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjuncion"),
                &["conjunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("immobilien"), &["immobile"]),
            (
                dictgen::InsensitiveStr::Ascii("professers"),
                &["professors"],
            ),
            (dictgen::InsensitiveStr::Ascii("featurin"), &["featuring"]),
            (
                dictgen::InsensitiveStr::Ascii("cholestorol"),
                &["cholesterol"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inintializer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwiase"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("investiagte"),
                &["investigate"],
            ),
            (dictgen::InsensitiveStr::Ascii("undertand"), &["understand"]),
            (
                dictgen::InsensitiveStr::Ascii("braodcasts"),
                &["broadcasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caclulation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("contniued"), &["continued"]),
            (dictgen::InsensitiveStr::Ascii("rosponse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("downgades"), &["downgrades"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("bankgok"), &["bangkok"]),
            (
                dictgen::InsensitiveStr::Ascii("driectories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinded"), &["combined"]),
            (
                dictgen::InsensitiveStr::Ascii("functitonal"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("maimum"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("adpated"), &["adapted"]),
            (
                dictgen::InsensitiveStr::Ascii("bloks"),
                &["blocks", "blokes"],
            ),
            (dictgen::InsensitiveStr::Ascii("strikely"), &["strikingly"]),
            (dictgen::InsensitiveStr::Ascii("enoght"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("connctor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("configirable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("senarreos"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("keyensian"), &["keynesian"]),
            (dictgen::InsensitiveStr::Ascii("coelesce"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("workbanchs"),
                &["workbenches"],
            ),
            (dictgen::InsensitiveStr::Ascii("calilng"), &["calling"]),
            (
                dictgen::InsensitiveStr::Ascii("resolucion"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intorducing"),
                &["introducing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiomeout"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("omnisicent"),
                &["omniscient"],
            ),
            (dictgen::InsensitiveStr::Ascii("thne"), &["then"]),
            (dictgen::InsensitiveStr::Ascii("idenify"), &["identify"]),
            (
                dictgen::InsensitiveStr::Ascii("sociopolical"),
                &["sociological"],
            ),
            (dictgen::InsensitiveStr::Ascii("distater"), &["disaster"]),
            (
                dictgen::InsensitiveStr::Ascii("competetion"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("usefulu"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("commplied"), &["complied"]),
            (dictgen::InsensitiveStr::Ascii("revelent"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("inmense"), &["immense"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastrutures"),
                &["infrastructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("carryng"), &["carrying"]),
            (dictgen::InsensitiveStr::Ascii("subroutie"), &["subroutine"]),
            (
                dictgen::InsensitiveStr::Ascii("iteratoins"),
                &["iterations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("longet"),
                &["longest", "longer"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciclical"), &["cyclical"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregetions"),
                &["aggregations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definiatively"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("collecton"), &["collection"]),
            (dictgen::InsensitiveStr::Ascii("excpet"), &["except"]),
            (dictgen::InsensitiveStr::Ascii("iconifie"), &["iconify"]),
            (
                dictgen::InsensitiveStr::Ascii("conparators"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancls"), &["cancels"]),
            (
                dictgen::InsensitiveStr::Ascii("mathemetical"),
                &["mathematical"],
            ),
            (dictgen::InsensitiveStr::Ascii("elgant"), &["elegant"]),
            (dictgen::InsensitiveStr::Ascii("deperecate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("compleet"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("appneding"), &["appending"]),
            (dictgen::InsensitiveStr::Ascii("cymptum"), &["symptom"]),
            (
                dictgen::InsensitiveStr::Ascii("battlechip"),
                &["battleship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("driectives"),
                &["directives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("educting"),
                &["educating", "deducting"],
            ),
            (dictgen::InsensitiveStr::Ascii("subjec"), &["subject"]),
            (
                dictgen::InsensitiveStr::Ascii("assocications"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("visualy"), &["visually"]),
            (
                dictgen::InsensitiveStr::Ascii("componenets"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intializaze"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("braodband"), &["broadband"]),
            (
                dictgen::InsensitiveStr::Ascii("apocalypitic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("metatata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("excitment"), &["excitement"]),
            (dictgen::InsensitiveStr::Ascii("verrical"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("techonlogical"),
                &["technological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proyect"),
                &["project", "protect"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimissals"), &["dismissals"]),
            (dictgen::InsensitiveStr::Ascii("heaby"), &["heavy"]),
            (dictgen::InsensitiveStr::Ascii("putposes"), &["purposes"]),
            (dictgen::InsensitiveStr::Ascii("recyle"), &["recycle"]),
            (
                dictgen::InsensitiveStr::Ascii("widnow"),
                &["widow", "window"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fucntioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("sevrally"), &["severally"]),
            (
                dictgen::InsensitiveStr::Ascii("initialyses"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("marineras"), &["mariners"]),
            (dictgen::InsensitiveStr::Ascii("secodary"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("sunbinterpreter"),
                &["subinterpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priods"),
                &["periods", "priors", "prods"],
            ),
            (dictgen::InsensitiveStr::Ascii("renqueued"), &["reenqueued"]),
            (dictgen::InsensitiveStr::Ascii("duellin"), &["duelling"]),
            (dictgen::InsensitiveStr::Ascii("regerts"), &["regrets"]),
            (dictgen::InsensitiveStr::Ascii("consants"), &["constants"]),
            (
                dictgen::InsensitiveStr::Ascii("acknolwedges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidebtly"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolutistes"),
                &["absolutists", "absolutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("radation"), &["radiation"]),
            (
                dictgen::InsensitiveStr::Ascii("repositoriy"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("supprort"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("imposible"), &["impossible"]),
            (dictgen::InsensitiveStr::Ascii("chahges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("interferencies"),
                &["interferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizontaal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("smalles"), &["smallest"]),
            (
                dictgen::InsensitiveStr::Ascii("legistation"),
                &["legislation"],
            ),
            (dictgen::InsensitiveStr::Ascii("spiritis"), &["spirits"]),
            (dictgen::InsensitiveStr::Ascii("fountian"), &["fountain"]),
            (dictgen::InsensitiveStr::Ascii("ouput"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("pennisular"),
                &["peninsular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extracter"),
                &["extractor", "extracted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scuccessful"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opertaions"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinguishe"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internatioanlly"),
                &["internationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nortification"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipelins"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("dealerhsip"),
                &["dealerships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configureations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precessors"),
                &["predecessors", "processors"],
            ),
            (dictgen::InsensitiveStr::Ascii("failuer"), &["failure"]),
            (
                dictgen::InsensitiveStr::Ascii("earhtquakes"),
                &["earthquakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charakters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("birght"), &["bright"]),
            (dictgen::InsensitiveStr::Ascii("mutuall"), &["mutually"]),
            (dictgen::InsensitiveStr::Ascii("chnanging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("pseudopoential"),
                &["pseudopotential"],
            ),
            (dictgen::InsensitiveStr::Ascii("globel"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("progerssive"),
                &["progressives"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylidner"), &["cylinder"]),
            (
                dictgen::InsensitiveStr::Ascii("represantative"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sponsers"), &["sponsors"]),
            (
                dictgen::InsensitiveStr::Ascii("comlications"),
                &["complications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adversiting"),
                &["advertising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrucction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("memebers"), &["members"]),
            (
                dictgen::InsensitiveStr::Ascii("anlytically"),
                &["analytically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otate"),
                &["notate", "rotate"],
            ),
            (dictgen::InsensitiveStr::Ascii("placment"), &["placement"]),
            (
                dictgen::InsensitiveStr::Ascii("characetrs"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("takedn"), &["taken"]),
            (dictgen::InsensitiveStr::Ascii("fullly"), &["fully"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbathe"),
                &["masturbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pormise"), &["promise"]),
            (
                dictgen::InsensitiveStr::Ascii("spreasheets"),
                &["spreadsheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neratives"),
                &["narratives", "negatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exporte"),
                &["export", "exported"],
            ),
            (dictgen::InsensitiveStr::Ascii("croatioa"), &["croatia"]),
            (
                dictgen::InsensitiveStr::Ascii("accomondation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximatley"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("partiall"), &["partially"]),
            (
                dictgen::InsensitiveStr::Ascii("decomposeion"),
                &["decomposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlooming"),
                &["overlooking"],
            ),
            (dictgen::InsensitiveStr::Ascii("intergity"), &["integrity"]),
            (
                dictgen::InsensitiveStr::Ascii("vierification"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("serach"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("installemnt"),
                &["installment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paravirutalise"),
                &["paravirtualise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("toether"),
                &["together", "tether"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validadion"),
                &["validation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currespondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sistematizing"),
                &["systematizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("probally"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("otimises"), &["optimises"]),
            (
                dictgen::InsensitiveStr::Ascii("partifular"),
                &["particular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deminsions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortimic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singlepalyer"),
                &["singleplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("chiense"), &["chinese"]),
            (dictgen::InsensitiveStr::Ascii("possessess"), &["possesses"]),
            (dictgen::InsensitiveStr::Ascii("auhtorise"), &["authorise"]),
            (dictgen::InsensitiveStr::Ascii("arrising"), &["arising"]),
            (dictgen::InsensitiveStr::Ascii("magentic"), &["magnetic"]),
            (
                dictgen::InsensitiveStr::Ascii("spezialiced"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("syles"), &["styles"]),
            (
                dictgen::InsensitiveStr::Ascii("racionalization"),
                &["rationalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinemental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuing"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordonates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimial"), &["optimal"]),
            (
                dictgen::InsensitiveStr::Ascii("unstall"),
                &["install", "uninstall"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfifying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("hopful"), &["hopeful"]),
            (dictgen::InsensitiveStr::Ascii("abvailable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("catgorizations"),
                &["categorizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("somoenes"), &["someones"]),
            (
                dictgen::InsensitiveStr::Ascii("inapproprietly"),
                &["inappropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ojbectives"),
                &["objectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("revists"), &["revisits"]),
            (dictgen::InsensitiveStr::Ascii("algorythim"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("positve"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("specifyig"), &["specifying"]),
            (dictgen::InsensitiveStr::Ascii("provison"), &["provision"]),
            (dictgen::InsensitiveStr::Ascii("resetted"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("advanatges"),
                &["advantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrigeraters"),
                &["refrigerators", "refrigerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteristcs"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownershp"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("debuggger"), &["debugger"]),
            (dictgen::InsensitiveStr::Ascii("sythetic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("vaccums"), &["vacuums"]),
            (dictgen::InsensitiveStr::Ascii("indictor"), &["indicator"]),
            (
                dictgen::InsensitiveStr::Ascii("crrespondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("portabel"), &["portable"]),
            (
                dictgen::InsensitiveStr::Ascii("naviagation"),
                &["navigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnanged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("worstation"),
                &["workstation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nativey"), &["natively"]),
            (dictgen::InsensitiveStr::Ascii("uncompres"), &["uncompress"]),
            (
                dictgen::InsensitiveStr::Ascii("gratituous"),
                &["gratuitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insignifiant"),
                &["insignificant"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuckt"), &["stuck"]),
            (dictgen::InsensitiveStr::Ascii("inpeach"), &["impeach"]),
            (dictgen::InsensitiveStr::Ascii("clonnes"), &["clones"]),
            (dictgen::InsensitiveStr::Ascii("origional"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("yoursleves"),
                &["yourselves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("actication"),
                &["activation"],
            ),
            (dictgen::InsensitiveStr::Ascii("apostels"), &["apostles"]),
            (
                dictgen::InsensitiveStr::Ascii("contencious"),
                &["contentious"],
            ),
            (dictgen::InsensitiveStr::Ascii("morgatges"), &["mortgages"]),
            (dictgen::InsensitiveStr::Ascii("tiggered"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("crackes"),
                &["crackles", "cracks", "crackers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfiying"),
                &["satisfying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritzations"),
                &["prioritizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerablility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transphopic"),
                &["transphobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("farse"), &["farce"]),
            (
                dictgen::InsensitiveStr::Ascii("acknolwedge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mimaching"),
                &["mismatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("vectror"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("fautured"), &["featured"]),
            (
                dictgen::InsensitiveStr::Ascii("cnfiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretets"),
                &["interpreters", "interprets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspdonent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlehips"),
                &["battleship"],
            ),
            (dictgen::InsensitiveStr::Ascii("concides"), &["concise"]),
            (
                dictgen::InsensitiveStr::Ascii("persecucion"),
                &["persecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recuriters"),
                &["recruiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("metada"), &["metadata"]),
            (
                dictgen::InsensitiveStr::Ascii("infridgement"),
                &["infringement"],
            ),
            (dictgen::InsensitiveStr::Ascii("downlaod"), &["download"]),
            (dictgen::InsensitiveStr::Ascii("tempoaray"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinater"),
                &["coordinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("rithm"), &["rhythm"]),
            (
                dictgen::InsensitiveStr::Ascii("inisialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("knowlingly"), &["knowingly"]),
            (dictgen::InsensitiveStr::Ascii("copmlete"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("strukture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("addiotionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("presidin"), &["presiding"]),
            (dictgen::InsensitiveStr::Ascii("curretly"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("iplement"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("conplimentary"),
                &["complimentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("chipet"), &["chipset"]),
            (
                dictgen::InsensitiveStr::Ascii("reconections"),
                &["reconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illsutration"),
                &["illustrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resotred"),
                &["restored", "resorted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cattleship"),
                &["battleship"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwers"), &["answers"]),
            (dictgen::InsensitiveStr::Ascii("syntihetic"), &["synthetic"]),
            (
                dictgen::InsensitiveStr::Ascii("mchanicals"),
                &["mechanicals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evelutionary"),
                &["evolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expoentially"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indipendently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlyzer"), &["analyzer"]),
            (dictgen::InsensitiveStr::Ascii("colum"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("aganst"), &["against"]),
            (dictgen::InsensitiveStr::Ascii("assinging"), &["assigning"]),
            (
                dictgen::InsensitiveStr::Ascii("singelarly"),
                &["singularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("histry"), &["history"]),
            (dictgen::InsensitiveStr::Ascii("hispanicus"), &["hispanics"]),
            (dictgen::InsensitiveStr::Ascii("excetuted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("parallellisation"),
                &["parallelisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omnishints"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("willingless"),
                &["willingness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessllation"),
                &["tessellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beauracratical"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonimus"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("nobady"), &["nobody"]),
            (
                dictgen::InsensitiveStr::Ascii("obsolesence"),
                &["obsolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("francsico"), &["francisco"]),
            (dictgen::InsensitiveStr::Ascii("feeedback"), &["feedback"]),
            (
                dictgen::InsensitiveStr::Ascii("convesions"),
                &["conversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwasys"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("contanins"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("freames"), &["frames"]),
            (
                dictgen::InsensitiveStr::Ascii("interneters"),
                &["internets"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepatates"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("interpertation"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("destop"), &["desktop"]),
            (dictgen::InsensitiveStr::Ascii("plugable"), &["pluggable"]),
            (dictgen::InsensitiveStr::Ascii("destort"), &["distort"]),
            (
                dictgen::InsensitiveStr::Ascii("postprocesing"),
                &["postprocessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survivabiity"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("follwoeiwng"),
                &["following"],
            ),
            (dictgen::InsensitiveStr::Ascii("rainbowers"), &["rainbows"]),
            (dictgen::InsensitiveStr::Ascii("desribing"), &["describing"]),
            (dictgen::InsensitiveStr::Ascii("cobined"), &["combined"]),
            (
                dictgen::InsensitiveStr::Ascii("sociophatic"),
                &["sociopathic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrantying"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("herculase"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("testiment"), &["testament"]),
            (dictgen::InsensitiveStr::Ascii("callint"), &["calling"]),
            (
                dictgen::InsensitiveStr::Ascii("efficitently"),
                &["efficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooprative"),
                &["cooperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maneouvred"),
                &["manoeuvred"],
            ),
            (dictgen::InsensitiveStr::Ascii("waisline"), &["waistline"]),
            (
                dictgen::InsensitiveStr::Ascii("performane"),
                &["performances", "performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("certianly"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("conffig"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("unaliged"), &["unaligned"]),
            (dictgen::InsensitiveStr::Ascii("ignortd"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("sirvayls"), &["surveils"]),
            (
                dictgen::InsensitiveStr::Ascii("aaproximates"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconeecting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgarading"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("bebongs"), &["belongs"]),
            (
                dictgen::InsensitiveStr::Ascii("organisert"),
                &["organisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowedging"),
                &["acknowledging"],
            ),
            (dictgen::InsensitiveStr::Ascii("vritual"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("creationsts"),
                &["creationists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oragnisation"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofcousre"), &["ofcourse"]),
            (
                dictgen::InsensitiveStr::Ascii("encyclepedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("dunegons"), &["dungeons"]),
            (dictgen::InsensitiveStr::Ascii("amzaing"), &["amazing"]),
            (dictgen::InsensitiveStr::Ascii("uffers"), &["buffers"]),
            (
                dictgen::InsensitiveStr::Ascii("succssfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribition"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoilday"), &["holiday"]),
            (
                dictgen::InsensitiveStr::Ascii("agressively"),
                &["aggressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("centerd"), &["centered"]),
            (dictgen::InsensitiveStr::Ascii("sherlok"), &["sherlock"]),
            (
                dictgen::InsensitiveStr::Ascii("expemental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrspondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introduciton"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("contetion"), &["contention"]),
            (
                dictgen::InsensitiveStr::Ascii("vunlerability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("unwiedly"), &["unwieldy"]),
            (
                dictgen::InsensitiveStr::Ascii("formidabble"),
                &["formidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtenticated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("quarante"), &["quarantine"]),
            (dictgen::InsensitiveStr::Ascii("subproces"), &["subprocess"]),
            (dictgen::InsensitiveStr::Ascii("assemly"), &["assembly"]),
            (dictgen::InsensitiveStr::Ascii("shystem"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("suplies"), &["supplies"]),
            (dictgen::InsensitiveStr::Ascii("htpt"), &["http"]),
            (dictgen::InsensitiveStr::Ascii("conumer"), &["consumer"]),
            (dictgen::InsensitiveStr::Ascii("parameteer"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("redefintions"),
                &["redefinitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("excutes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("accumulater"),
                &["accumulator", "accumulated", "accumulates", "accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexaul"),
                &["homosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derugulation"),
                &["deregulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("measues"), &["measures"]),
            (dictgen::InsensitiveStr::Ascii("ignorme"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("permuates"),
                &["permutates", "permutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("jumpt"), &["jumped", "jump"]),
            (dictgen::InsensitiveStr::Ascii("analized"), &["analyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("supsiciously"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inkonsistent"),
                &["inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("nightlty"), &["nightly"]),
            (
                dictgen::InsensitiveStr::Ascii("preblematic"),
                &["problematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("buthcered"), &["butchered"]),
            (dictgen::InsensitiveStr::Ascii("cloess"), &["closes"]),
            (
                dictgen::InsensitiveStr::Ascii("motorcicle"),
                &["motorcycle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliencing"), &["silencing"]),
            (
                dictgen::InsensitiveStr::Ascii("cimptomaticly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("acctually"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("corrispondances"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumulated"), &["simulated"]),
            (dictgen::InsensitiveStr::Ascii("siad"), &["said"]),
            (
                dictgen::InsensitiveStr::Ascii("transfromations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("piglrim"), &["pilgrim"]),
            (
                dictgen::InsensitiveStr::Ascii("owerflowing"),
                &["overflowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ununsupported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("aimd"), &["aimed"]),
            (
                dictgen::InsensitiveStr::Ascii("featuers"),
                &["features", "feathers"],
            ),
            (dictgen::InsensitiveStr::Ascii("atll"), &["all"]),
            (dictgen::InsensitiveStr::Ascii("chaacters"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("subseqeunt"),
                &["subsequent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expendeble"),
                &["expendable"],
            ),
            (dictgen::InsensitiveStr::Ascii("aktivated"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("idication"), &["indication"]),
            (
                dictgen::InsensitiveStr::Ascii("ocilaters"),
                &["oscillators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distibutions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("haeder"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("misogynsitic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisions"),
                &["commissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("baoynet"), &["bayonet"]),
            (dictgen::InsensitiveStr::Ascii("naviagte"), &["navigate"]),
            (
                dictgen::InsensitiveStr::Ascii("assoications"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteritisc"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpter"), &["accepter"]),
            (dictgen::InsensitiveStr::Ascii("fonction"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("initiailising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("nastyness"), &["nastiness"]),
            (dictgen::InsensitiveStr::Ascii("verically"), &["vertically"]),
            (
                dictgen::InsensitiveStr::Ascii("temporaily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignmnt"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("incorperation"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("multliple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("economicaly"),
                &["economically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strengthining"),
                &["strengthening"],
            ),
            (dictgen::InsensitiveStr::Ascii("defailt"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("sussinct"), &["succinct"]),
            (dictgen::InsensitiveStr::Ascii("litquid"), &["liquid"]),
            (dictgen::InsensitiveStr::Ascii("conincides"), &["coincides"]),
            (
                dictgen::InsensitiveStr::Ascii("convinient"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scatch"),
                &["catch", "scratch", "sketch"],
            ),
            (dictgen::InsensitiveStr::Ascii("reast"), &["rest"]),
            (dictgen::InsensitiveStr::Ascii("churchers"), &["churches"]),
            (
                dictgen::InsensitiveStr::Ascii("certificiate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instlled"),
                &["installed", "instilled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differientiates"),
                &["differentiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("activelly"), &["actively"]),
            (dictgen::InsensitiveStr::Ascii("backdor"), &["backdoor"]),
            (
                dictgen::InsensitiveStr::Ascii("undocummented"),
                &["undocumented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominationals"),
                &["denominations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installtion"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influanced"),
                &["influenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("autmotive"), &["automotive"]),
            (
                dictgen::InsensitiveStr::Ascii("cocuments"),
                &["document", "documents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowlegeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("erntry"), &["entry"]),
            (
                dictgen::InsensitiveStr::Ascii("pitchforkes"),
                &["pitchforks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimizaton"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnicated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("fumctions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("hesistant"), &["hesitant"]),
            (dictgen::InsensitiveStr::Ascii("trukish"), &["turkish"]),
            (
                dictgen::InsensitiveStr::Ascii("icrementing"),
                &["incrementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibiltiies"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anoymously"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternatvies"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cholestoral"),
                &["cholesterol"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfearing"),
                &["interfering"],
            ),
            (dictgen::InsensitiveStr::Ascii("transfomm"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("cirlces"), &["circles"]),
            (
                dictgen::InsensitiveStr::Ascii("completelly"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("depsoited"), &["deposited"]),
            (
                dictgen::InsensitiveStr::Ascii("extenstions"),
                &["extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanaticos"), &["fanatics"]),
            (
                dictgen::InsensitiveStr::Ascii("instrucitonal"),
                &["instructional"],
            ),
            (dictgen::InsensitiveStr::Ascii("criteras"), &["criteria"]),
            (dictgen::InsensitiveStr::Ascii("natrually"), &["naturally"]),
            (dictgen::InsensitiveStr::Ascii("cruetly"), &["cruelty"]),
            (
                dictgen::InsensitiveStr::Ascii("cehckboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eleimination"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucceed"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("thems"), &["them", "themes"]),
            (
                dictgen::InsensitiveStr::Ascii("vegitarion"),
                &["vegetarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("degrassse"), &["degrasse"]),
            (
                dictgen::InsensitiveStr::Ascii("synonymnous"),
                &["synonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("provine"), &["province"]),
            (dictgen::InsensitiveStr::Ascii("mornign"), &["morning"]),
            (
                dictgen::InsensitiveStr::Ascii("strage"),
                &["storage", "strange"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innecessarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcct"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("specificallly"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wariors"), &["warriors"]),
            (dictgen::InsensitiveStr::Ascii("intiially"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhodd"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("mountble"), &["mountable"]),
            (dictgen::InsensitiveStr::Ascii("midpints"), &["midpoints"]),
            (dictgen::InsensitiveStr::Ascii("addersses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastrinating"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("currenlty"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("quarentine"),
                &["quarantine"],
            ),
            (dictgen::InsensitiveStr::Ascii("fauture"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("sourt"),
                &["sort", "south", "sour"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailsed"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("gool"), &["ghoul"]),
            (
                dictgen::InsensitiveStr::Ascii("authorititive"),
                &["authoritative"],
            ),
            (dictgen::InsensitiveStr::Ascii("obstable"), &["obstacle"]),
            (dictgen::InsensitiveStr::Ascii("halpoints"), &["halfpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("explot"),
                &["exploit", "explore"],
            ),
            (dictgen::InsensitiveStr::Ascii("entchant"), &["enchant"]),
            (dictgen::InsensitiveStr::Ascii("traiging"), &["triaging"]),
            (dictgen::InsensitiveStr::Ascii("permature"), &["premature"]),
            (
                dictgen::InsensitiveStr::Ascii("unitilising"),
                &["uninitialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("ypes"), &["types"]),
            (dictgen::InsensitiveStr::Ascii("consumend"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("frecuented"),
                &["frequented"],
            ),
            (dictgen::InsensitiveStr::Ascii("sciripts"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("expansie"), &["expansive"]),
            (
                dictgen::InsensitiveStr::Ascii("prixy"),
                &["proxy", "privy", "pricy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepretator"),
                &["interpretor", "interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommpresed"),
                &["uncompressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("linar"), &["linear"]),
            (dictgen::InsensitiveStr::Ascii("regualtor"), &["regulator"]),
            (
                dictgen::InsensitiveStr::Ascii("particpating"),
                &["participating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interlaving"),
                &["interlacing", "interleaving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opperations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sessio"), &["session"]),
            (dictgen::InsensitiveStr::Ascii("desicions"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("ciguretes"), &["cigarettes"]),
            (dictgen::InsensitiveStr::Ascii("objejct"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("comonent"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("redable"), &["readable"]),
            (
                dictgen::InsensitiveStr::Ascii("possiblility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chesckerboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("cushin"), &["cushion"]),
            (
                dictgen::InsensitiveStr::Ascii("guarentying"),
                &["guarantee", "guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepenent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsisencies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscarraige"),
                &["miscarriage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("historiaan"),
                &["historians"],
            ),
            (dictgen::InsensitiveStr::Ascii("leagal"), &["legal"]),
            (
                dictgen::InsensitiveStr::Ascii("ingrediants"),
                &["ingredients"],
            ),
            (dictgen::InsensitiveStr::Ascii("missence"), &["missense"]),
            (
                dictgen::InsensitiveStr::Ascii("strucrured"),
                &["structured"],
            ),
            (dictgen::InsensitiveStr::Ascii("strengts"), &["strengths"]),
            (dictgen::InsensitiveStr::Ascii("pinockle"), &["pinochle"]),
            (
                dictgen::InsensitiveStr::Ascii("unavaialbel"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservativsm"),
                &["conservatism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curiousities"),
                &["curiosities"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjectief"),
                &["subjective"],
            ),
            (dictgen::InsensitiveStr::Ascii("qest"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("cusomizations"),
                &["customizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsequentual"),
                &["inconsequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimiiters"),
                &["delimiters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accessiores"),
                &["accessories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shound"),
                &["should", "sound"],
            ),
            (dictgen::InsensitiveStr::Ascii("anuled"), &["annulled"]),
            (
                dictgen::InsensitiveStr::Ascii("envoker"),
                &["evoker", "invoker"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagge"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditialy"),
                &["unconditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("combaines"), &["combines"]),
            (dictgen::InsensitiveStr::Ascii("deplyos"), &["deploys"]),
            (dictgen::InsensitiveStr::Ascii("iteraes"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("absoultly"), &["absolutely"]),
            (
                dictgen::InsensitiveStr::Ascii("legenderies"),
                &["legendaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("barriors"), &["barriers"]),
            (
                dictgen::InsensitiveStr::Ascii("analyitical"),
                &["analytical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("providor"),
                &["provider", "providore"],
            ),
            (dictgen::InsensitiveStr::Ascii("contnt"), &["content"]),
            (dictgen::InsensitiveStr::Ascii("cashieres"), &["cashiers"]),
            (dictgen::InsensitiveStr::Ascii("burguny"), &["burgundy"]),
            (dictgen::InsensitiveStr::Ascii("bacomes"), &["becomes"]),
            (dictgen::InsensitiveStr::Ascii("suseedes"), &["secedes"]),
            (
                dictgen::InsensitiveStr::Ascii("technichian"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categogies"),
                &["categories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumualtes"),
                &["accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("malless"), &["malice"]),
            (dictgen::InsensitiveStr::Ascii("evaulates"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("calcultated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbelievablly"),
                &["unbelievably"],
            ),
            (dictgen::InsensitiveStr::Ascii("filterin"), &["filtering"]),
            (
                dictgen::InsensitiveStr::Ascii("expectaions"),
                &["expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submissiona"),
                &["submission", "submissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modle"),
                &["model", "module"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("washignton"),
                &["washington"],
            ),
            (dictgen::InsensitiveStr::Ascii("genitics"), &["genetics"]),
            (dictgen::InsensitiveStr::Ascii("mabe"), &["maybe"]),
            (
                dictgen::InsensitiveStr::Ascii("appretince"),
                &["apprentice"],
            ),
            (dictgen::InsensitiveStr::Ascii("smooghing"), &["smoothing"]),
            (dictgen::InsensitiveStr::Ascii("seriesi"), &["series"]),
            (
                dictgen::InsensitiveStr::Ascii("clonned"),
                &["cloned", "clowned", "conned"],
            ),
            (dictgen::InsensitiveStr::Ascii("settngs"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("straightenend"),
                &["straightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("arramging"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("practicallity"),
                &["practically", "practicality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traslating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("helment"), &["helmet"]),
            (
                dictgen::InsensitiveStr::Ascii("supportet"),
                &["supporter", "supported"],
            ),
            (dictgen::InsensitiveStr::Ascii("whtielist"), &["whitelist"]),
            (
                dictgen::InsensitiveStr::Ascii("incremnets"),
                &["increments"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecrtors"), &["selectors"]),
            (dictgen::InsensitiveStr::Ascii("butifuly"), &["beautifully"]),
            (
                dictgen::InsensitiveStr::Ascii("repostiory"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supllementing"),
                &["supplementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrouding"),
                &["surrounding"],
            ),
            (dictgen::InsensitiveStr::Ascii("fequently"), &["frequently"]),
            (dictgen::InsensitiveStr::Ascii("tottehnam"), &["tottenham"]),
            (dictgen::InsensitiveStr::Ascii("sispect"), &["suspect"]),
            (
                dictgen::InsensitiveStr::Ascii("assoziative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("tottenahm"), &["tottenham"]),
            (
                dictgen::InsensitiveStr::Ascii("prdpagating"),
                &["propagating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negociating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favilitate"),
                &["facilitate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envirnmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaible"), &["reliable"]),
            (dictgen::InsensitiveStr::Ascii("specfies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("calcuations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("classsroom"), &["classroom"]),
            (
                dictgen::InsensitiveStr::Ascii("kubermetes"),
                &["kubernetes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worning"),
                &["warning", "morning"],
            ),
            (dictgen::InsensitiveStr::Ascii("morgage"), &["mortgage"]),
            (dictgen::InsensitiveStr::Ascii("oigins"), &["origins"]),
            (
                dictgen::InsensitiveStr::Ascii("exeggerating"),
                &["exaggerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcionally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripitor"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbaks"), &["callbacks"]),
            (dictgen::InsensitiveStr::Ascii("folowinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("modiifier"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("unversity"), &["university"]),
            (
                dictgen::InsensitiveStr::Ascii("providors"),
                &["providers", "providores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intterupting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrelaible"),
                &["unreliable"],
            ),
            (dictgen::InsensitiveStr::Ascii("weigthed"), &["weighted"]),
            (dictgen::InsensitiveStr::Ascii("internaly"), &["internally"]),
            (dictgen::InsensitiveStr::Ascii("pxoxied"), &["proxied"]),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedorens"),
                &["tetrahedrons"],
            ),
            (dictgen::InsensitiveStr::Ascii("exterlal"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("conspriacies"),
                &["conspiracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workaraound"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("struggel"), &["struggle"]),
            (dictgen::InsensitiveStr::Ascii("prviate"), &["private"]),
            (dictgen::InsensitiveStr::Ascii("indicatior"), &["indicator"]),
            (dictgen::InsensitiveStr::Ascii("combindes"), &["combines"]),
            (dictgen::InsensitiveStr::Ascii("gnawwed"), &["gnawed"]),
            (
                dictgen::InsensitiveStr::Ascii("requresting"),
                &["requesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("floaing"),
                &["floating", "flowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholarhips"),
                &["scholarship", "scholarships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelerar"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("proccesed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("instrutor"), &["instructor"]),
            (dictgen::InsensitiveStr::Ascii("devangari"), &["devanagari"]),
            (
                dictgen::InsensitiveStr::Ascii("fonctionnalies"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pascheurising"),
                &["pasteurising"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeators"), &["operators"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferencable"),
                &["dereferenceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("mamentos"), &["mementos"]),
            (
                dictgen::InsensitiveStr::Ascii("chekerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depretiates"),
                &["depreciates", "deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("trubbles"), &["troubles"]),
            (
                dictgen::InsensitiveStr::Ascii("peaces"),
                &["pieces", "peace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apropriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagerrated"),
                &["exaggerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavailablility"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defnitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genrations"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("febewary"), &["february"]),
            (dictgen::InsensitiveStr::Ascii("aquiessed"), &["acquiesced"]),
            (dictgen::InsensitiveStr::Ascii("numbe"), &["number", "numb"]),
            (dictgen::InsensitiveStr::Ascii("poitns"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("vluable"), &["valuable"]),
            (
                dictgen::InsensitiveStr::Ascii("streichung"),
                &["stretching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insenstively"),
                &["insensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benechmarks"),
                &["benchmarks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consdidered"),
                &["considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegdes"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarcaration"),
                &["incarceration"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabke"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("neightbouring"),
                &["neighbouring"],
            ),
            (dictgen::InsensitiveStr::Ascii("augmnet"), &["augment"]),
            (dictgen::InsensitiveStr::Ascii("singuarl"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("porposes"),
                &["proposes", "purposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluculated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("rejcted"), &["rejected"]),
            (dictgen::InsensitiveStr::Ascii("industy"), &["industry"]),
            (
                dictgen::InsensitiveStr::Ascii("noramlizing"),
                &["normalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represneting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceppt"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("univeristy"),
                &["university"],
            ),
            (dictgen::InsensitiveStr::Ascii("reviewr"), &["reviewer"]),
            (
                dictgen::InsensitiveStr::Ascii("accentuatin"),
                &["accentuating", "accentuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("artice"), &["article"]),
            (dictgen::InsensitiveStr::Ascii("cleary"), &["clearly"]),
            (
                dictgen::InsensitiveStr::Ascii("placd"),
                &["placed", "place"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomonal"),
                &["phenomenal"],
            ),
            (dictgen::InsensitiveStr::Ascii("wapper"), &["wrapper"]),
            (dictgen::InsensitiveStr::Ascii("becacdd"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("republis"), &["republics"]),
            (dictgen::InsensitiveStr::Ascii("efffective"), &["effective"]),
            (dictgen::InsensitiveStr::Ascii("workbnech"), &["workbench"]),
            (dictgen::InsensitiveStr::Ascii("quartery"), &["quarterly"]),
            (dictgen::InsensitiveStr::Ascii("vertixs"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("proceeeding"),
                &["proceeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("scritped"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("vetrex"), &["vertex"]),
            (dictgen::InsensitiveStr::Ascii("executding"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("organsier"), &["organiser"]),
            (dictgen::InsensitiveStr::Ascii("informarle"), &["informal"]),
            (dictgen::InsensitiveStr::Ascii("cotain"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("skeletton"), &["skeleton"]),
            (dictgen::InsensitiveStr::Ascii("extrememe"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("nubmered"), &["numbered"]),
            (dictgen::InsensitiveStr::Ascii("sonewhat"), &["somewhat"]),
            (dictgen::InsensitiveStr::Ascii("numers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("ccertificated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("enlose"), &["enclose"]),
            (dictgen::InsensitiveStr::Ascii("unrwitten"), &["unwritten"]),
            (
                dictgen::InsensitiveStr::Ascii("averging"),
                &["averaging", "avenging"],
            ),
            (dictgen::InsensitiveStr::Ascii("evluation"), &["evaluation"]),
            (
                dictgen::InsensitiveStr::Ascii("arithemtic"),
                &["arithmetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allcoating"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachterize"),
                &["characterize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mchanically"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reurning"),
                &["returning", "rerunning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggegations"),
                &["aggregations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("routies"),
                &["routines", "routes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distructive"),
                &["destructive"],
            ),
            (dictgen::InsensitiveStr::Ascii("godlfish"), &["goldfish"]),
            (dictgen::InsensitiveStr::Ascii("englih"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("disscovers"), &["discovers"]),
            (
                dictgen::InsensitiveStr::Ascii("standrdize"),
                &["standardize"],
            ),
            (dictgen::InsensitiveStr::Ascii("disktop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("underlflows"),
                &["underflows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nexting"),
                &["nesting", "texting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experse"),
                &["expense", "express"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuccessfullly"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("excepton"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("delibirate"),
                &["deliberate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tangeantial"),
                &["tangential"],
            ),
            (dictgen::InsensitiveStr::Ascii("releaing"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("tetsters"), &["testers"]),
            (dictgen::InsensitiveStr::Ascii("descrite"), &["discrete"]),
            (
                dictgen::InsensitiveStr::Ascii("intestinas"),
                &["intestines"],
            ),
            (dictgen::InsensitiveStr::Ascii("rackit"), &["racket"]),
            (
                dictgen::InsensitiveStr::Ascii("incompetentence"),
                &["incompetence"],
            ),
            (dictgen::InsensitiveStr::Ascii("polcies"), &["policies"]),
            (
                dictgen::InsensitiveStr::Ascii("caluclating"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fashist"), &["fascist"]),
            (
                dictgen::InsensitiveStr::Ascii("disclousre"),
                &["disclosure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equiavlent"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggestief"),
                &["suggestive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ptiched"), &["pitched"]),
            (dictgen::InsensitiveStr::Ascii("extactly"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("etensions"), &["extensions"]),
            (
                dictgen::InsensitiveStr::Ascii("cerativity"),
                &["creativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("lauches"), &["launches"]),
            (
                dictgen::InsensitiveStr::Ascii("daclaration"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("perferm"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("liberalest"), &["liberate"]),
            (dictgen::InsensitiveStr::Ascii("numher"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("hypocrates"),
                &["hypocrites"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrady"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("dubley"), &["doubly"]),
            (dictgen::InsensitiveStr::Ascii("goldamn"), &["goldman"]),
            (
                dictgen::InsensitiveStr::Ascii("confinguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("certaionty"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("priortising"),
                &["prioritising"],
            ),
            (dictgen::InsensitiveStr::Ascii("retunrned"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("dismabiguation"),
                &["disambiguation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("believeable"),
                &["believable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofusing"), &["confusing"]),
            (
                dictgen::InsensitiveStr::Ascii("esponding"),
                &["desponding", "responding"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomates"), &["automates"]),
            (dictgen::InsensitiveStr::Ascii("beghavior"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("sequencs"),
                &["sequences", "sequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("fewwer"), &["fewer"]),
            (
                dictgen::InsensitiveStr::Ascii("critisizes"),
                &["criticizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommented"),
                &["recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orienatated"),
                &["orientated"],
            ),
            (dictgen::InsensitiveStr::Ascii("decalares"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("appropropriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribytes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subterranian"),
                &["subterranean"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gometrically"),
                &["geometrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("immitated"), &["imitated"]),
            (dictgen::InsensitiveStr::Ascii("isntantly"), &["instantly"]),
            (
                dictgen::InsensitiveStr::Ascii("downagrades"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("anhoter"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("spefify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("catedgories"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("inclution"), &["inclusion"]),
            (
                dictgen::InsensitiveStr::Ascii("confugured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("despict"), &["depict"]),
            (dictgen::InsensitiveStr::Ascii("siffixing"), &["suffixing"]),
            (
                dictgen::InsensitiveStr::Ascii("ubsubstantiated"),
                &["unsubstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("includding"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("updters"), &["updaters"]),
            (dictgen::InsensitiveStr::Ascii("opitmizer"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("chacking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("comppilers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("degenarating"),
                &["degenerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("garrions"), &["garrison"]),
            (
                dictgen::InsensitiveStr::Ascii("distincion"),
                &["distinction"],
            ),
            (dictgen::InsensitiveStr::Ascii("bibilcal"), &["biblical"]),
            (dictgen::InsensitiveStr::Ascii("modellling"), &["modelling"]),
            (
                dictgen::InsensitiveStr::Ascii("inisializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("adavance"), &["advance"]),
            (dictgen::InsensitiveStr::Ascii("parametar"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("arguement"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitalize"),
                &["uninitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("preimums"), &["premiums"]),
            (dictgen::InsensitiveStr::Ascii("granolla"), &["granola"]),
            (
                dictgen::InsensitiveStr::Ascii("expedicion"),
                &["expedition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stardardized"),
                &["standardized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collission"),
                &["collisions", "collision", "collusion"],
            ),
            (dictgen::InsensitiveStr::Ascii("rapells"), &["rappells"]),
            (dictgen::InsensitiveStr::Ascii("hanler"), &["handler"]),
            (
                dictgen::InsensitiveStr::Ascii("automically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("kubernets"), &["kubernetes"]),
            (
                dictgen::InsensitiveStr::Ascii("perventing"),
                &["preventing", "perverting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quadruplin"),
                &["quadrupling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admittadly"),
                &["admittedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comrpomising"),
                &["compromising"],
            ),
            (dictgen::InsensitiveStr::Ascii("notfy"), &["notify"]),
            (
                dictgen::InsensitiveStr::Ascii("instruccion"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misdimeanor"),
                &["misdemeanor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("principels"),
                &["principals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanicly"),
                &["mechanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrcting"),
                &["constructing", "constricting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extravagent"),
                &["extravagant"],
            ),
            (dictgen::InsensitiveStr::Ascii("analouges"), &["analogues"]),
            (
                dictgen::InsensitiveStr::Ascii("calculautions"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("naybhorly"), &["neighborly"]),
            (dictgen::InsensitiveStr::Ascii("obsessin"), &["obsession"]),
            (dictgen::InsensitiveStr::Ascii("vitrual"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("miracalous"),
                &["miraculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("vlarge"), &["large"]),
            (
                dictgen::InsensitiveStr::Ascii("vunlerabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("patince"), &["patience"]),
            (dictgen::InsensitiveStr::Ascii("interioare"), &["interior"]),
            (
                dictgen::InsensitiveStr::Ascii("updatr"),
                &["updater", "update"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acommodated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sycning"), &["syncing"]),
            (dictgen::InsensitiveStr::Ascii("outwiegh"), &["outweigh"]),
            (
                dictgen::InsensitiveStr::Ascii("sigularities"),
                &["singularities"],
            ),
            (dictgen::InsensitiveStr::Ascii("clearning"), &["clearing"]),
            (
                dictgen::InsensitiveStr::Ascii("supernatrual"),
                &["supernatural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypocrities"),
                &["hypocrite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disillisioned"),
                &["disillusioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("usued"), &["used"]),
            (
                dictgen::InsensitiveStr::Ascii("intertwinned"),
                &["intertwined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemishpere"),
                &["hemisphere"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesections"),
                &["intersections"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefifics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("repsectable"),
                &["respectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("escation"), &["escalation"]),
            (dictgen::InsensitiveStr::Ascii("candidats"), &["candidates"]),
            (dictgen::InsensitiveStr::Ascii("attriubute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("measuements"),
                &["measurements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apppreciating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("worspaces"), &["workspaces"]),
            (dictgen::InsensitiveStr::Ascii("chizzil"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("eliminster"),
                &["eliminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("publishd"), &["published"]),
            (
                dictgen::InsensitiveStr::Ascii("intereactions"),
                &["intersections", "interactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("rouines"), &["routines"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsulting"),
                &["encapsulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deductiable"),
                &["deductible"],
            ),
            (dictgen::InsensitiveStr::Ascii("adrea"), &["area"]),
            (dictgen::InsensitiveStr::Ascii("decrytped"), &["decrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("chocoaltes"),
                &["chocolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("facttory"), &["factory"]),
            (dictgen::InsensitiveStr::Ascii("exersises"), &["exercises"]),
            (dictgen::InsensitiveStr::Ascii("succee"), &["succeed"]),
            (
                dictgen::InsensitiveStr::Ascii("conenction"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valentimes"),
                &["valentines"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownes"), &["owns", "ones"]),
            (dictgen::InsensitiveStr::Ascii("enforceing"), &["enforcing"]),
            (dictgen::InsensitiveStr::Ascii("foldin"), &["folding"]),
            (
                dictgen::InsensitiveStr::Ascii("reseraches"),
                &["researches"],
            ),
            (dictgen::InsensitiveStr::Ascii("istallers"), &["installers"]),
            (dictgen::InsensitiveStr::Ascii("albiet"), &["albeit"]),
            (dictgen::InsensitiveStr::Ascii("maximial"), &["maximal"]),
            (dictgen::InsensitiveStr::Ascii("oarcles"), &["oracles"]),
            (
                dictgen::InsensitiveStr::Ascii("ommishiences"),
                &["omnisciences"],
            ),
            (dictgen::InsensitiveStr::Ascii("chisles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("somewaht"), &["somewhat"]),
            (
                dictgen::InsensitiveStr::Ascii("ocasionaly"),
                &["occasionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seemlessly"),
                &["seamlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("chescksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("paramteters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajustment"), &["adjustment"]),
            (
                dictgen::InsensitiveStr::Ascii("playthtoughs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("resierfs"), &["reiserfs"]),
            (
                dictgen::InsensitiveStr::Ascii("gradiating"),
                &["graduating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proletaricat"),
                &["proletariat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("telaportation"),
                &["teleportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approimation"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramameters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytle"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("resignition"),
                &["resignation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disclamier"),
                &["disclaimer"],
            ),
            (dictgen::InsensitiveStr::Ascii("beloging"), &["belonging"]),
            (
                dictgen::InsensitiveStr::Ascii("antripanewer"),
                &["entrepreneur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intializer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("emplyers"), &["employers"]),
            (
                dictgen::InsensitiveStr::Ascii("directorires"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congresmen"),
                &["congressmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryfier"), &["verifier"]),
            (
                dictgen::InsensitiveStr::Ascii("descritprs"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrooll"), &["scroll"]),
            (dictgen::InsensitiveStr::Ascii("realsied"), &["realised"]),
            (dictgen::InsensitiveStr::Ascii("slogen"), &["slogan"]),
            (dictgen::InsensitiveStr::Ascii("athesim"), &["atheism"]),
            (
                dictgen::InsensitiveStr::Ascii("instuctors"),
                &["instructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccessing"), &["accessing"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncicalize"),
                &["canonicalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computaton"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("miralces"), &["miracles"]),
            (
                dictgen::InsensitiveStr::Ascii("charactears"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaiates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctiption"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("normlizes"), &["normalizes"]),
            (dictgen::InsensitiveStr::Ascii("statemnt"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("bouund"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("propgated"), &["propagated"]),
            (dictgen::InsensitiveStr::Ascii("severless"), &["serverless"]),
            (dictgen::InsensitiveStr::Ascii("pojrected"), &["projected"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentle"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapaiting"),
                &["dissipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("followinig"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("gauage"), &["gauge"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("gnerative"), &["generative"]),
            (
                dictgen::InsensitiveStr::Ascii("resposiblities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsers"),
                &["users", "unsets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qiestioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("dploy"), &["deploy"]),
            (dictgen::InsensitiveStr::Ascii("datatore"), &["datastore"]),
            (dictgen::InsensitiveStr::Ascii("bahaviors"), &["behaviors"]),
            (dictgen::InsensitiveStr::Ascii("ancedotes"), &["anecdotes"]),
            (
                dictgen::InsensitiveStr::Ascii("involuntarely"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("tavelling"), &["travelling"]),
            (
                dictgen::InsensitiveStr::Ascii("originallity"),
                &["originality"],
            ),
            (dictgen::InsensitiveStr::Ascii("iddle"), &["idle"]),
            (dictgen::InsensitiveStr::Ascii("diserable"), &["desirable"]),
            (dictgen::InsensitiveStr::Ascii("releaasing"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("piclking"), &["pickling"]),
            (dictgen::InsensitiveStr::Ascii("featues"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticte"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstraccions"),
                &["abstractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriping"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("cources"),
                &["courses", "coerces", "sources"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rembering"),
                &["remembering", "renumbering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commissionar"),
                &["commissioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("berserkr"), &["berserker"]),
            (dictgen::InsensitiveStr::Ascii("cehcks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("culmulating"),
                &["cumulating", "culminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("opration"), &["operation"]),
            (
                dictgen::InsensitiveStr::Ascii("cheecklists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chloropleth"),
                &["choropleth"],
            ),
            (dictgen::InsensitiveStr::Ascii("nstalling"), &["installing"]),
            (dictgen::InsensitiveStr::Ascii("oyur"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("fils"),
                &["fills", "files", "file"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimmentelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("believeing"), &["believing"]),
            (dictgen::InsensitiveStr::Ascii("saxaphone"), &["saxophone"]),
            (
                dictgen::InsensitiveStr::Ascii("advantegeous"),
                &["advantageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disastreous"),
                &["disastrous"],
            ),
            (dictgen::InsensitiveStr::Ascii("spammeur"), &["spammer"]),
            (
                dictgen::InsensitiveStr::Ascii("analgoues"),
                &["analogues", "analogous"],
            ),
            (dictgen::InsensitiveStr::Ascii("pulic"), &["public"]),
            (dictgen::InsensitiveStr::Ascii("dipolma"), &["diploma"]),
            (dictgen::InsensitiveStr::Ascii("stle"), &["style"]),
            (dictgen::InsensitiveStr::Ascii("pseduo"), &["pseudo"]),
            (dictgen::InsensitiveStr::Ascii("clhorine"), &["chlorine"]),
            (
                dictgen::InsensitiveStr::Ascii("internatioanlists"),
                &["internationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("ratpure"), &["rapture"]),
            (
                dictgen::InsensitiveStr::Ascii("migate"),
                &["migrate", "mitigate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accordinly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsteram"), &["upstream"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacuters"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("preoxide"), &["peroxide"]),
            (dictgen::InsensitiveStr::Ascii("rebuiling"), &["rebuilding"]),
            (
                dictgen::InsensitiveStr::Ascii("compasssion"),
                &["compassion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differances"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychiatist"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("generages"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("apropriately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("nurtient"), &["nutrient"]),
            (
                dictgen::InsensitiveStr::Ascii("infograpgic"),
                &["infographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ist"),
                &["is", "it", "its", "sit", "list"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfearnce"),
                &["interference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accepptance"),
                &["acceptance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inapropriate"),
                &["inappropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminitering"),
                &["administering"],
            ),
            (dictgen::InsensitiveStr::Ascii("compositie"), &["composite"]),
            (
                dictgen::InsensitiveStr::Ascii("aggegation"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("taret"),
                &["target", "tarot"],
            ),
            (dictgen::InsensitiveStr::Ascii("otpimiser"), &["optimiser"]),
            (
                dictgen::InsensitiveStr::Ascii("abstrctness"),
                &["abstractness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprentatives"),
                &["representatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("flavourus"), &["flavours"]),
            (dictgen::InsensitiveStr::Ascii("cocktrice"), &["cockatrice"]),
            (
                dictgen::InsensitiveStr::Ascii("misdameanor"),
                &["misdemeanor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atatchment"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("distantce"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("revrse"), &["reverse"]),
            (
                dictgen::InsensitiveStr::Ascii("istallations"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorrig"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("muiltipliers"),
                &["multipliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncicalization"),
                &["canonicalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administrer"),
                &["administer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organsiations"),
                &["organisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("toppingest"), &["toppings"]),
            (
                dictgen::InsensitiveStr::Ascii("assocaitive"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sapmle"), &["sample"]),
            (
                dictgen::InsensitiveStr::Ascii("gropus"),
                &["groups", "gropes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punishmet"),
                &["punishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("adusted"), &["adjusted"]),
            (
                dictgen::InsensitiveStr::Ascii("downlowding"),
                &["downloading"],
            ),
            (dictgen::InsensitiveStr::Ascii("interrrupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("arquived"), &["archived"]),
            (dictgen::InsensitiveStr::Ascii("arugably"), &["arguably"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticaes"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("joineable"), &["joinable"]),
            (
                dictgen::InsensitiveStr::Ascii("standardisin"),
                &["standardising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registerd"),
                &["registered", "registers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gargage"),
                &["garbage", "garage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficieirs"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("graphis"), &["graphics"]),
            (dictgen::InsensitiveStr::Ascii("funstion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("nomral"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("poreclain"), &["porcelain"]),
            (dictgen::InsensitiveStr::Ascii("libraray"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("sence"), &["sense", "since"]),
            (
                dictgen::InsensitiveStr::Ascii("amealearator"),
                &["ameliorator"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatyepes"), &["datatypes"]),
            (
                dictgen::InsensitiveStr::Ascii("indetification"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decsriptor"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("comnplete"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("compoenent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("progammatically"),
                &["programmatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclcoking"),
                &["overclocking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicaple"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("impaitent"), &["impatient"]),
            (dictgen::InsensitiveStr::Ascii("increae"), &["increase"]),
            (dictgen::InsensitiveStr::Ascii("basially"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("industriella"),
                &["industrial"],
            ),
            (dictgen::InsensitiveStr::Ascii("moderats"), &["moderates"]),
            (
                dictgen::InsensitiveStr::Ascii("hystorically"),
                &["hysterically"],
            ),
            (dictgen::InsensitiveStr::Ascii("garanties"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("condtionals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("guitards"), &["guitars"]),
            (
                dictgen::InsensitiveStr::Ascii("anmespaces"),
                &["namespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("expercting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("frustraited"),
                &["frustrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("squirrelies"),
                &["squirrels"],
            ),
            (dictgen::InsensitiveStr::Ascii("expains"), &["explains"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubuted"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcaultes"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("boofays"), &["buffets"]),
            (dictgen::InsensitiveStr::Ascii("protlets"), &["portlets"]),
            (dictgen::InsensitiveStr::Ascii("digsted"), &["digested"]),
            (
                dictgen::InsensitiveStr::Ascii("comaparator"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("artficial"), &["artificial"]),
            (dictgen::InsensitiveStr::Ascii("behabior"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("emoi"), &["emoji"]),
            (dictgen::InsensitiveStr::Ascii("permise"), &["premise"]),
            (dictgen::InsensitiveStr::Ascii("feetur"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("procoessing"),
                &["processing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiazers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nealy"),
                &["nearly", "newly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compre"),
                &["compare", "compère"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preisdents"),
                &["presidents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aprehensive"),
                &["apprehensive"],
            ),
            (dictgen::InsensitiveStr::Ascii("precense"), &["presence"]),
            (dictgen::InsensitiveStr::Ascii("desginer"), &["designer"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapoimted"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("actuallly"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("requency"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("qualfy"), &["qualify"]),
            (
                dictgen::InsensitiveStr::Ascii("equaivalent"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("countains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("intermedate"),
                &["intermediate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("truncaction"),
                &["truncation"],
            ),
            (dictgen::InsensitiveStr::Ascii("centruy"), &["century"]),
            (
                dictgen::InsensitiveStr::Ascii("proletariet"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("privided"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("excplict"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("cassm"), &["chasm"]),
            (
                dictgen::InsensitiveStr::Ascii("lighthning"),
                &["lightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseuticals"),
                &["pharmaceuticals"],
            ),
            (dictgen::InsensitiveStr::Ascii("hygeine"), &["hygiene"]),
            (
                dictgen::InsensitiveStr::Ascii("overrideen"),
                &["overridden"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execultable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("broadcat"),
                &["broadcasts", "broadcast"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomizer"),
                &["randomizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("neightbor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("pairokeelly"),
                &["parochially"],
            ),
            (dictgen::InsensitiveStr::Ascii("halarious"), &["hilarious"]),
            (dictgen::InsensitiveStr::Ascii("platimun"), &["platinum"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitialzed"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("nawshusly"), &["nauseously"]),
            (
                dictgen::InsensitiveStr::Ascii("insufficant"),
                &["insufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enginereed"),
                &["engineered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introveted"),
                &["introverted"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialate"), &["dilate"]),
            (
                dictgen::InsensitiveStr::Ascii("calcuators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimzers"), &["optimizers"]),
            (dictgen::InsensitiveStr::Ascii("icleandic"), &["icelandic"]),
            (
                dictgen::InsensitiveStr::Ascii("triagularity"),
                &["triangularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occaisonally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("bullevard"), &["boulevard"]),
            (dictgen::InsensitiveStr::Ascii("progammed"), &["programmed"]),
            (
                dictgen::InsensitiveStr::Ascii("contracition"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranmission"),
                &["transmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deathamtch"),
                &["deathmatch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imprisonent"),
                &["imprisonment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepulchrlly"),
                &["sepulchrally"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrisinc"), &["intrinsic"]),
            (
                dictgen::InsensitiveStr::Ascii("intoleranse"),
                &["intolerance"],
            ),
            (dictgen::InsensitiveStr::Ascii("expersses"), &["expresses"]),
            (
                dictgen::InsensitiveStr::Ascii("crenditals"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitutivo"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differenctly"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplistisch"),
                &["simplistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("readible"), &["readable"]),
            (dictgen::InsensitiveStr::Ascii("inuition"), &["intuition"]),
            (dictgen::InsensitiveStr::Ascii("lcoally"), &["locally"]),
            (
                dictgen::InsensitiveStr::Ascii("mutiplicity"),
                &["multiplicity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompasitating"),
                &["incapacitating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antagonisitic"),
                &["antagonistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sherlcok"), &["sherlock"]),
            (
                dictgen::InsensitiveStr::Ascii("assignmetns"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepreate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("nsted"), &["nested"]),
            (
                dictgen::InsensitiveStr::Ascii("downgardes"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("intergrity"), &["integrity"]),
            (dictgen::InsensitiveStr::Ascii("supossing"), &["supposing"]),
            (dictgen::InsensitiveStr::Ascii("supose"), &["suppose"]),
            (
                dictgen::InsensitiveStr::Ascii("conflcited"),
                &["conflicted"],
            ),
            (dictgen::InsensitiveStr::Ascii("merajes"), &["mirages"]),
            (dictgen::InsensitiveStr::Ascii("listenters"), &["listeners"]),
            (dictgen::InsensitiveStr::Ascii("trew"), &["threw", "true"]),
            (dictgen::InsensitiveStr::Ascii("commment"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("metioned"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("cirular"), &["circular"]),
            (dictgen::InsensitiveStr::Ascii("accortind"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("recongnizes"),
                &["recognizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("makign"), &["making"]),
            (dictgen::InsensitiveStr::Ascii("completen"), &["complement"]),
            (dictgen::InsensitiveStr::Ascii("arrangerad"), &["arranged"]),
            (
                dictgen::InsensitiveStr::Ascii("checkpoings"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenenticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("bagded"), &["badged"]),
            (dictgen::InsensitiveStr::Ascii("attachtes"), &["attaches"]),
            (dictgen::InsensitiveStr::Ascii("wepbage"), &["webpage"]),
            (
                dictgen::InsensitiveStr::Ascii("discourges"),
                &["discourages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initator"),
                &["imitator", "initiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontaneuosly"),
                &["spontaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("candidite"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdenently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepretors"),
                &["interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubeshooting"),
                &["troubleshooting"],
            ),
            (dictgen::InsensitiveStr::Ascii("comenting"), &["commenting"]),
            (dictgen::InsensitiveStr::Ascii("densitiy"), &["density"]),
            (dictgen::InsensitiveStr::Ascii("cordinate"), &["coordinate"]),
            (dictgen::InsensitiveStr::Ascii("selectd"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("acceeptors"), &["acceptors"]),
            (dictgen::InsensitiveStr::Ascii("weahter"), &["weather"]),
            (dictgen::InsensitiveStr::Ascii("sinoid"), &["sinusoid"]),
            (dictgen::InsensitiveStr::Ascii("handshke"), &["handshake"]),
            (
                dictgen::InsensitiveStr::Ascii("conpetitions"),
                &["competitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("identicial"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("stupiditiy"), &["stupidity"]),
            (dictgen::InsensitiveStr::Ascii("cylces"), &["cycles"]),
            (dictgen::InsensitiveStr::Ascii("immedate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("conecept"), &["concept"]),
            (
                dictgen::InsensitiveStr::Ascii("deassering"),
                &["deasserting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("translpant"),
                &["transplants"],
            ),
            (dictgen::InsensitiveStr::Ascii("ebcuase"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("contuining"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaquirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plaguarism"),
                &["plagiarism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asssertively"),
                &["assertively"],
            ),
            (dictgen::InsensitiveStr::Ascii("objtains"), &["obtains"]),
            (dictgen::InsensitiveStr::Ascii("syntesis"), &["synthesis"]),
            (dictgen::InsensitiveStr::Ascii("tabualtor"), &["tabulator"]),
            (
                dictgen::InsensitiveStr::Ascii("concantenate"),
                &["concatenate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarasingly"),
                &["embarrassingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberivated"),
                &["abbreviated"],
            ),
            (dictgen::InsensitiveStr::Ascii("desctop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("diaciritcs"),
                &["diacritics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("staggerwing"),
                &["staggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("loclae"), &["locale"]),
            (
                dictgen::InsensitiveStr::Ascii("composistes"),
                &["composites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anntations"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("stratgey"), &["strategy"]),
            (
                dictgen::InsensitiveStr::Ascii("javescript"),
                &["javascript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stroyboard"),
                &["storyboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relfections"),
                &["reflections"],
            ),
            (dictgen::InsensitiveStr::Ascii("proedural"), &["procedural"]),
            (
                dictgen::InsensitiveStr::Ascii("subtrafuge"),
                &["subterfuge"],
            ),
            (dictgen::InsensitiveStr::Ascii("ephimeral"), &["ephemeral"]),
            (dictgen::InsensitiveStr::Ascii("methpd"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("belyes"), &["belies"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangaelbe"),
                &["arrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaluates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("dicision"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("vsibility"), &["visibility"]),
            (
                dictgen::InsensitiveStr::Ascii("abriviation"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("kown"), &["known"]),
            (
                dictgen::InsensitiveStr::Ascii("catagorizing"),
                &["categorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("callapsed"), &["collapsed"]),
            (dictgen::InsensitiveStr::Ascii("underfined"), &["undefined"]),
            (dictgen::InsensitiveStr::Ascii("bassics"), &["basics"]),
            (dictgen::InsensitiveStr::Ascii("excptions"), &["exceptions"]),
            (
                dictgen::InsensitiveStr::Ascii("abbreveations"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certficates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("loepard"), &["leopard"]),
            (
                dictgen::InsensitiveStr::Ascii("insteallation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rendayvoued"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthoganal"),
                &["orthogonal"],
            ),
            (dictgen::InsensitiveStr::Ascii("apppear"), &["appear"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptionnal"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("calll"), &["call"]),
            (
                dictgen::InsensitiveStr::Ascii("intuitivley"),
                &["intuitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("serinomy"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("sciencers"), &["sciences"]),
            (
                dictgen::InsensitiveStr::Ascii("identiies"),
                &["identities", "identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desecripter"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaitive"),
                &["initiative"],
            ),
            (dictgen::InsensitiveStr::Ascii("countinual"), &["continual"]),
            (
                dictgen::InsensitiveStr::Ascii("impossbily"),
                &["impossibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wapping"),
                &["wrapping", "swapping", "warping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalsing"),
                &["capitalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("upddated"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("triator"), &["traitor"]),
            (
                dictgen::InsensitiveStr::Ascii("strategems"),
                &["strategies"],
            ),
            (dictgen::InsensitiveStr::Ascii("varants"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("defragmenation"),
                &["defragmentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nauseos"), &["nauseous"]),
            (
                dictgen::InsensitiveStr::Ascii("fianlization"),
                &["finalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("relesed"), &["released"]),
            (dictgen::InsensitiveStr::Ascii("coruching"), &["crouching"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplents"),
                &["reimplements", "reimplants"],
            ),
            (dictgen::InsensitiveStr::Ascii("expressy"), &["expressly"]),
            (
                dictgen::InsensitiveStr::Ascii("indoctronated"),
                &["indoctrinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("paralels"), &["parallels"]),
            (
                dictgen::InsensitiveStr::Ascii("classication"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("submition"), &["submission"]),
            (
                dictgen::InsensitiveStr::Ascii("comparativley"),
                &["comparatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactionn"),
                &["interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("devive"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("advertiers"),
                &["advertisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tiemstamped"),
                &["timestamped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downsteramers"),
                &["downstreamers"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanceling"), &["canceling"]),
            (dictgen::InsensitiveStr::Ascii("speciefies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("straange"), &["strange"]),
            (
                dictgen::InsensitiveStr::Ascii("optimizatons"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenedently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispell"), &["dispel"]),
            (dictgen::InsensitiveStr::Ascii("calclator"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("deletin"),
                &["deleting", "deletion"],
            ),
            (dictgen::InsensitiveStr::Ascii("errect"), &["erect"]),
            (dictgen::InsensitiveStr::Ascii("timestaps"), &["timestamps"]),
            (dictgen::InsensitiveStr::Ascii("conteined"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("shenangians"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefented"),
                &["prevented", "presented"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomtical"), &["automatic"]),
            (
                dictgen::InsensitiveStr::Ascii("transpaernt"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispoosables"),
                &["disposables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unversionned"),
                &["unversioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("tridnet"), &["trident"]),
            (dictgen::InsensitiveStr::Ascii("pscyhotic"), &["psychotic"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicae"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("potentiel"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("restricion"),
                &["restriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("coummunity"), &["community"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuccessfull"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcesses"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("messaages"), &["messages"]),
            (
                dictgen::InsensitiveStr::Ascii("respectabil"),
                &["respectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuents"), &["students"]),
            (
                dictgen::InsensitiveStr::Ascii("agresssive"),
                &["aggressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("malariya"), &["malaria"]),
            (
                dictgen::InsensitiveStr::Ascii("hypotheticaly"),
                &["hypothetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mamages"), &["manages"]),
            (dictgen::InsensitiveStr::Ascii("releoad"), &["reload"]),
            (
                dictgen::InsensitiveStr::Ascii("accodringly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("imapired"), &["impaired"]),
            (dictgen::InsensitiveStr::Ascii("pojector"), &["projector"]),
            (dictgen::InsensitiveStr::Ascii("unfoulded"), &["unfolded"]),
            (
                dictgen::InsensitiveStr::Ascii("remaines"),
                &["remains", "remained"],
            ),
            (dictgen::InsensitiveStr::Ascii("nofifies"), &["notifies"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequsite"),
                &["prerequisite"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("newine"), &["newline"]),
            (
                dictgen::InsensitiveStr::Ascii("nanoparicles"),
                &["nanoparticles"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixifed"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("loaclized"), &["localized"]),
            (dictgen::InsensitiveStr::Ascii("everytiem"), &["everytime"]),
            (
                dictgen::InsensitiveStr::Ascii("canocicalizing"),
                &["canonicalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("intenral"), &["internal"]),
            (dictgen::InsensitiveStr::Ascii("backwrad"), &["backward"]),
            (dictgen::InsensitiveStr::Ascii("solfs"), &["solves"]),
            (dictgen::InsensitiveStr::Ascii("paramter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("imrpoved"), &["improved"]),
            (
                dictgen::InsensitiveStr::Ascii("trotskyite"),
                &["trotskyist"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghtly"), &["lengthy"]),
            (
                dictgen::InsensitiveStr::Ascii("retroactivelly"),
                &["retroactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constained"),
                &["constrained", "contained"],
            ),
            (dictgen::InsensitiveStr::Ascii("clisheys"), &["clichés"]),
            (dictgen::InsensitiveStr::Ascii("freedum"), &["freedom"]),
            (
                dictgen::InsensitiveStr::Ascii("windsheilds"),
                &["windshield"],
            ),
            (dictgen::InsensitiveStr::Ascii("chlild"), &["child"]),
            (
                dictgen::InsensitiveStr::Ascii("corresondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfiion"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("daugher"), &["daughter"]),
            (
                dictgen::InsensitiveStr::Ascii("midifelder"),
                &["midfielder"],
            ),
            (dictgen::InsensitiveStr::Ascii("prophacy"), &["prophecy"]),
            (dictgen::InsensitiveStr::Ascii("goldburger"), &["goldberg"]),
            (
                dictgen::InsensitiveStr::Ascii("mastubration"),
                &["masturbation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succeeed"),
                &["succeed", "succeeded"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissable"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("prejeduced"),
                &["prejudiced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bookmakred"),
                &["bookmarked"],
            ),
            (dictgen::InsensitiveStr::Ascii("contruct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("suply"), &["supply"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuer"),
                &["contribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frations"),
                &["fractions", "rations"],
            ),
            (dictgen::InsensitiveStr::Ascii("appned"), &["append"]),
            (dictgen::InsensitiveStr::Ascii("necessaery"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("trikery"), &["trickery"]),
            (dictgen::InsensitiveStr::Ascii("preests"), &["presets"]),
            (
                dictgen::InsensitiveStr::Ascii("restorarion"),
                &["restoration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulernabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtitel"), &["subtitle"]),
            (dictgen::InsensitiveStr::Ascii("mashal"), &["marshal"]),
            (dictgen::InsensitiveStr::Ascii("seperate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("cermonially"),
                &["ceremonially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemmorhaged"),
                &["hemorrhaged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufaturers"),
                &["manufacturers"],
            ),
            (dictgen::InsensitiveStr::Ascii("fransisco"), &["francisco"]),
            (
                dictgen::InsensitiveStr::Ascii("invovlement"),
                &["involvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirtyness"), &["dirtiness"]),
            (
                dictgen::InsensitiveStr::Ascii("substansial"),
                &["substantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associetes"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaod"), &["reload"]),
            (dictgen::InsensitiveStr::Ascii("impoves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("pipepline"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("extisting"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("thinckness"), &["thickness"]),
            (dictgen::InsensitiveStr::Ascii("akss"), &["asks", "ass"]),
            (
                dictgen::InsensitiveStr::Ascii("maketplaces"),
                &["marketplaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntain"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("assember"), &["assembler"]),
            (
                dictgen::InsensitiveStr::Ascii("inconveience"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("affilliate"), &["affiliate"]),
            (
                dictgen::InsensitiveStr::Ascii("availaliable"),
                &["available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnosticsm"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("newcaslte"), &["newcastle"]),
            (
                dictgen::InsensitiveStr::Ascii("adjustsments"),
                &["adjustments"],
            ),
            (dictgen::InsensitiveStr::Ascii("filelds"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaited"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nastertiems"),
                &["nasturtiums"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collatoral"),
                &["collateral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canabelyses"),
                &["cannibalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("triumpth"), &["triumph"]),
            (
                dictgen::InsensitiveStr::Ascii("neckboards"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorcement"),
                &["endorsement", "enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("stpiped"), &["stripped"]),
            (dictgen::InsensitiveStr::Ascii("copryight"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("intrval"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("apeared"), &["appeared"]),
            (
                dictgen::InsensitiveStr::Ascii("methodoly"),
                &["methodology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disambiguaiton"),
                &["disambiguation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chcks"), &["checks"]),
            (dictgen::InsensitiveStr::Ascii("deterine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiing"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politickin"),
                &["politician", "politicking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chronoligical"),
                &["chronological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contemporany"),
                &["contemporary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fascistisk"), &["fascists"]),
            (dictgen::InsensitiveStr::Ascii("guyzers"), &["geysers"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplementions"),
                &["reimplementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("recepits"), &["receipts"]),
            (
                dictgen::InsensitiveStr::Ascii("signifcantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("boudary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("sloothing"), &["sleuthing"]),
            (dictgen::InsensitiveStr::Ascii("generaties"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("onlsaught"), &["onslaught"]),
            (dictgen::InsensitiveStr::Ascii("weaponary"), &["weaponry"]),
            (dictgen::InsensitiveStr::Ascii("sirection"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("provideres"), &["providers"]),
            (dictgen::InsensitiveStr::Ascii("thpe"), &["type"]),
            (dictgen::InsensitiveStr::Ascii("commite"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("skurging"), &["scourging"]),
            (
                dictgen::InsensitiveStr::Ascii("occuppation"),
                &["occupation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("toplogical"),
                &["topological"],
            ),
            (dictgen::InsensitiveStr::Ascii("dilay"), &["delay"]),
            (
                dictgen::InsensitiveStr::Ascii("pairokeally"),
                &["parochially"],
            ),
            (dictgen::InsensitiveStr::Ascii("directroy"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("throttes"),
                &["throttles", "trots"],
            ),
            (dictgen::InsensitiveStr::Ascii("exptra"), &["extra"]),
            (
                dictgen::InsensitiveStr::Ascii("battelfield"),
                &["battlefield"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradditions"),
                &["traditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("defyning"), &["defining"]),
            (dictgen::InsensitiveStr::Ascii("doiing"), &["doing"]),
            (dictgen::InsensitiveStr::Ascii("cameleons"), &["chameleons"]),
            (
                dictgen::InsensitiveStr::Ascii("requirment"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structring"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probllematic"),
                &["problematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("obvisious"), &["obvious"]),
            (
                dictgen::InsensitiveStr::Ascii("claravoyantes"),
                &["clairvoyants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginazation"),
                &["organization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expriencing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electrnoics"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dueing"),
                &["doing", "during", "dueling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("artificielly"),
                &["artificially"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifyer"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("substituts"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blitzkreig"),
                &["blitzkrieg"],
            ),
            (dictgen::InsensitiveStr::Ascii("aroud"), &["around"]),
            (
                dictgen::InsensitiveStr::Ascii("extraterrestial"),
                &["extraterrestrial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocroprocessors"),
                &["microprocessors"],
            ),
            (dictgen::InsensitiveStr::Ascii("signaures"), &["signatures"]),
            (dictgen::InsensitiveStr::Ascii("corrrect"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("wrkstation"),
                &["workstation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirects"), &["directs"]),
            (
                dictgen::InsensitiveStr::Ascii("suplimented"),
                &["supplemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punishemnt"),
                &["punishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("parragraph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("chancging"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("princinpal"), &["principal"]),
            (
                dictgen::InsensitiveStr::Ascii("narcissisism"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaguelly"), &["vaguely"]),
            (dictgen::InsensitiveStr::Ascii("ocurr"), &["occur"]),
            (dictgen::InsensitiveStr::Ascii("prirority"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("depdent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("bokmarks"), &["bookmarks"]),
            (
                dictgen::InsensitiveStr::Ascii("promordials"),
                &["primordials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("candidatas"),
                &["candidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacialises"),
                &["specialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirearcically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("conecpt"), &["concept"]),
            (
                dictgen::InsensitiveStr::Ascii("priortizations"),
                &["prioritizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thressholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retramsmits"),
                &["retransmits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calucalting"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignoerd"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("empited"), &["emptied"]),
            (dictgen::InsensitiveStr::Ascii("follewing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("zoomming"), &["zooming"]),
            (
                dictgen::InsensitiveStr::Ascii("aggragated"),
                &["aggregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependencied"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostnae"), &["hostname"]),
            (dictgen::InsensitiveStr::Ascii("mispells"), &["misspells"]),
            (dictgen::InsensitiveStr::Ascii("cleint"), &["client"]),
            (dictgen::InsensitiveStr::Ascii("proptery"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("explusions"),
                &["explosions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aletrs"),
                &["alerts", "alters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordines"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containting"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliving"),
                &["delivering", "delving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsencored"),
                &["uncensored"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipleine"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("commano"), &["commando"]),
            (dictgen::InsensitiveStr::Ascii("roations"), &["rotations"]),
            (
                dictgen::InsensitiveStr::Ascii("workagound"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovewrite"), &["overwrite"]),
            (dictgen::InsensitiveStr::Ascii("denstiy"), &["density"]),
            (
                dictgen::InsensitiveStr::Ascii("suspisions"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("discraced"), &["disgraced"]),
            (dictgen::InsensitiveStr::Ascii("seconly"), &["secondly"]),
            (dictgen::InsensitiveStr::Ascii("winn"), &["win"]),
            (dictgen::InsensitiveStr::Ascii("protcted"), &["protected"]),
            (dictgen::InsensitiveStr::Ascii("anntenas"), &["antennas"]),
            (dictgen::InsensitiveStr::Ascii("maximini"), &["maximizing"]),
            (
                dictgen::InsensitiveStr::Ascii("contamporaries"),
                &["contemporaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supllement"),
                &["supplement"],
            ),
            (dictgen::InsensitiveStr::Ascii("insersted"), &["inserted"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitializing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chcekerboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("brunettte"), &["brunette"]),
            (dictgen::InsensitiveStr::Ascii("slefies"), &["selfies"]),
            (dictgen::InsensitiveStr::Ascii("nubers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("ostencibly"),
                &["ostensibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("denpendent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("generaing"), &["generating"]),
            (dictgen::InsensitiveStr::Ascii("varaiances"), &["variances"]),
            (dictgen::InsensitiveStr::Ascii("fustrated"), &["frustrated"]),
            (dictgen::InsensitiveStr::Ascii("deveoper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("maintioned"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("critrion"), &["criterion"]),
            (
                dictgen::InsensitiveStr::Ascii("normailzation"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("redliens"), &["redlines"]),
            (dictgen::InsensitiveStr::Ascii("udate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("interection"),
                &["interaction", "intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("poitnd"), &["pointed"]),
            (
                dictgen::InsensitiveStr::Ascii("ideosyncratic"),
                &["idiosyncratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("tollerance"), &["tolerance"]),
            (dictgen::InsensitiveStr::Ascii("ignortable"), &["ignorable"]),
            (dictgen::InsensitiveStr::Ascii("allocaton"), &["allocation"]),
            (
                dictgen::InsensitiveStr::Ascii("subbet"),
                &["subset", "subnet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invairably"),
                &["invariably"],
            ),
            (dictgen::InsensitiveStr::Ascii("kronicle"), &["chronicle"]),
            (
                dictgen::InsensitiveStr::Ascii("psychologits"),
                &["psychologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorthyn"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("froced"), &["forced"]),
            (dictgen::InsensitiveStr::Ascii("recangle"), &["rectangle"]),
            (dictgen::InsensitiveStr::Ascii("eveluator"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("attriuted"), &["attributed"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaocracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("indiciate"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("updtes"), &["updates"]),
            (
                dictgen::InsensitiveStr::Ascii("arhitectures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thoruoghly"),
                &["thoroughly"],
            ),
            (dictgen::InsensitiveStr::Ascii("boardband"), &["broadband"]),
            (
                dictgen::InsensitiveStr::Ascii("silowetes"),
                &["silhouettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instituional"),
                &["institutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("bording"), &["boarding"]),
            (
                dictgen::InsensitiveStr::Ascii("simualting"),
                &["simulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("anguage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("controversa"),
                &["controversial"],
            ),
            (dictgen::InsensitiveStr::Ascii("conmands"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("suplly"),
                &["supply", "sully"],
            ),
            (dictgen::InsensitiveStr::Ascii("critria"), &["criteria"]),
            (dictgen::InsensitiveStr::Ascii("faymus"), &["famous"]),
            (dictgen::InsensitiveStr::Ascii("timzezones"), &["timezones"]),
            (dictgen::InsensitiveStr::Ascii("characture"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("unrecognizeable"),
                &["unrecognizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adevnturing"),
                &["adventuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("improrting"), &["importing"]),
            (
                dictgen::InsensitiveStr::Ascii("represanting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("triangls"), &["triangles"]),
            (dictgen::InsensitiveStr::Ascii("accepecter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("ortolaryngologist"),
                &["otolaryngologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("endcoding"), &["encoding"]),
            (
                dictgen::InsensitiveStr::Ascii("attrbiutions"),
                &["attributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("methpds"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("decoartions"),
                &["decoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("impleiment"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("cahsier"), &["cashier"]),
            (
                dictgen::InsensitiveStr::Ascii("paremeterless"),
                &["parameterless"],
            ),
            (dictgen::InsensitiveStr::Ascii("incompete"), &["incomplete"]),
            (dictgen::InsensitiveStr::Ascii("currentusa"), &["currents"]),
            (
                dictgen::InsensitiveStr::Ascii("executatables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("comares"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("paticularly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("simulato"), &["simulation"]),
            (
                dictgen::InsensitiveStr::Ascii("incarseration"),
                &["incarceration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacmenet"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquaint"), &["acquaint"]),
            (
                dictgen::InsensitiveStr::Ascii("handsharing"),
                &["handshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("sterotype"), &["stereotype"]),
            (
                dictgen::InsensitiveStr::Ascii("exhuasting"),
                &["exhausting"],
            ),
            (dictgen::InsensitiveStr::Ascii("conidered"), &["considered"]),
            (dictgen::InsensitiveStr::Ascii("stremed"), &["streamed"]),
            (
                dictgen::InsensitiveStr::Ascii("concurrant"),
                &["concurrent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("headquater"),
                &["headquarter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restricitons"),
                &["restrictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffentiate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reagrdless"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiby"),
                &["compatibly", "compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesssed"), &["accessed"]),
            (dictgen::InsensitiveStr::Ascii("pragam"), &["pragma"]),
            (dictgen::InsensitiveStr::Ascii("cutomised"), &["customised"]),
            (
                dictgen::InsensitiveStr::Ascii("informatino"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreconized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoit"), &["exploit"]),
            (dictgen::InsensitiveStr::Ascii("interwebbs"), &["interwebs"]),
            (dictgen::InsensitiveStr::Ascii("similat"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("coverge"),
                &["coverage", "converge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocroprocessor"),
                &["microprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survivalibity"),
                &["survivability"],
            ),
            (dictgen::InsensitiveStr::Ascii("retirned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("prosperos"), &["prosperous"]),
            (dictgen::InsensitiveStr::Ascii("spcecified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("whatevr"), &["whatever"]),
            (dictgen::InsensitiveStr::Ascii("reuest"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("niether"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("ofset"), &["offset"]),
            (
                dictgen::InsensitiveStr::Ascii("corretion"),
                &["correction", "correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasposing"),
                &["transposing"],
            ),
            (dictgen::InsensitiveStr::Ascii("assiciete"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("expetctedly"),
                &["expectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("gnerally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("desinged"), &["designed"]),
            (
                dictgen::InsensitiveStr::Ascii("intercouse"),
                &["intercourse"],
            ),
            (dictgen::InsensitiveStr::Ascii("appeares"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("eventaully"),
                &["eventually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvenice"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coreespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resurections"),
                &["resurrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("coerscive"), &["coercive"]),
            (dictgen::InsensitiveStr::Ascii("maintians"), &["maintains"]),
            (dictgen::InsensitiveStr::Ascii("sperate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("diagonsed"), &["diagnosed"]),
            (dictgen::InsensitiveStr::Ascii("lithuiana"), &["lithuania"]),
            (dictgen::InsensitiveStr::Ascii("shapening"), &["sharpening"]),
            (
                dictgen::InsensitiveStr::Ascii("compansations"),
                &["compensations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tesselates"),
                &["tessellates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticiing"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidentely"),
                &["confidently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thumbnal"),
                &["thumbnails", "thumbnail"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("becames"),
                &["becomes", "became"],
            ),
            (dictgen::InsensitiveStr::Ascii("droppin"), &["dropping"]),
            (dictgen::InsensitiveStr::Ascii("palyers"), &["players"]),
            (
                dictgen::InsensitiveStr::Ascii("administraively"),
                &["administratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthymic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("volatges"), &["voltages"]),
            (
                dictgen::InsensitiveStr::Ascii("susequences"),
                &["subsequences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeditor"),
                &["competitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenanegins"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("curvture"), &["curvature"]),
            (dictgen::InsensitiveStr::Ascii("protocalls"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("grpahs"), &["graphs"]),
            (dictgen::InsensitiveStr::Ascii("claculate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("defnitive"), &["definitive"]),
            (dictgen::InsensitiveStr::Ascii("arrises"), &["arises"]),
            (dictgen::InsensitiveStr::Ascii("endpont"), &["endpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("activition"),
                &["activation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onmisience"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucrates"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanshake"), &["handshake"]),
            (
                dictgen::InsensitiveStr::Ascii("intimadated"),
                &["intimidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contignuous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denegrating"),
                &["denigrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("kiyacker"), &["kayaker"]),
            (dictgen::InsensitiveStr::Ascii("utimate"), &["ultimate"]),
            (dictgen::InsensitiveStr::Ascii("overlow"), &["overflow"]),
            (dictgen::InsensitiveStr::Ascii("werid"), &["weird"]),
            (dictgen::InsensitiveStr::Ascii("actualky"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("spectatores"),
                &["spectators"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetxure"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("corrcted"), &["corrected"]),
            (
                dictgen::InsensitiveStr::Ascii("correspending"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferentiators"),
                &["differentiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overcompansations"),
                &["overcompensations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interplotating"),
                &["interpolating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conttonwood"),
                &["cottonwood"],
            ),
            (dictgen::InsensitiveStr::Ascii("nammed"), &["named"]),
            (dictgen::InsensitiveStr::Ascii("normolise"), &["normalise"]),
            (
                dictgen::InsensitiveStr::Ascii("missisipi"),
                &["mississippi"],
            ),
            (dictgen::InsensitiveStr::Ascii("byttes"), &["bytes"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailabe"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dynamlically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciruclar"), &["circular"]),
            (dictgen::InsensitiveStr::Ascii("termined"), &["terminated"]),
            (
                dictgen::InsensitiveStr::Ascii("preservating"),
                &["preservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lighhtning"),
                &["lightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograhed"),
                &["photographed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unihabited"),
                &["uninhabited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sementing"),
                &["cementing", "segmenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("authoises"), &["authorises"]),
            (dictgen::InsensitiveStr::Ascii("dopped"), &["dropped"]),
            (dictgen::InsensitiveStr::Ascii("asistant"), &["assistant"]),
            (
                dictgen::InsensitiveStr::Ascii("legitimicy"),
                &["legitimacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("echosystem"), &["ecosystem"]),
            (
                dictgen::InsensitiveStr::Ascii("suggetsing"),
                &["suggesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deisgnated"),
                &["designated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differientiating"),
                &["differentiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("lizens"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("regulae"), &["regulate"]),
            (
                dictgen::InsensitiveStr::Ascii("anialating"),
                &["annihilating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dysentry"), &["dysentery"]),
            (
                dictgen::InsensitiveStr::Ascii("systematiclly"),
                &["systematically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("campainging"),
                &["campaigning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathimatic"),
                &["mathematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detramental"),
                &["detrimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("overloded"), &["overloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("optomisers"),
                &["optimisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetimentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countriside"),
                &["countryside"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tragicallly"),
                &["tragically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutomizer"), &["customizer"]),
            (
                dictgen::InsensitiveStr::Ascii("definaltey"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negogiated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("deplyoed"), &["deployed"]),
            (dictgen::InsensitiveStr::Ascii("carreful"), &["careful"]),
            (dictgen::InsensitiveStr::Ascii("reguarding"), &["regarding"]),
            (dictgen::InsensitiveStr::Ascii("execudes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenaded"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliaze"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("maritan"), &["martian"]),
            (
                dictgen::InsensitiveStr::Ascii("approxime"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investingations"),
                &["investigations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anntoations"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("warrnings"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("jusitfication"),
                &["justifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("athesits"), &["atheists"]),
            (dictgen::InsensitiveStr::Ascii("hydropile"), &["hydrophile"]),
            (dictgen::InsensitiveStr::Ascii("mismaches"), &["mismatches"]),
            (dictgen::InsensitiveStr::Ascii("packacge"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("homineum"), &["hominem"]),
            (
                dictgen::InsensitiveStr::Ascii("participativo"),
                &["participation"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("rarified"), &["rarefied"]),
            (dictgen::InsensitiveStr::Ascii("capbale"), &["capable"]),
            (dictgen::InsensitiveStr::Ascii("entents"), &["intents"]),
            (dictgen::InsensitiveStr::Ascii("ressize"), &["resize"]),
            (
                dictgen::InsensitiveStr::Ascii("measurmenet"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountabillity"),
                &["accountability"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacels"), &["cancels"]),
            (dictgen::InsensitiveStr::Ascii("workaound"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("teaspon"), &["teaspoon"]),
            (dictgen::InsensitiveStr::Ascii("diganosis"), &["diagnosis"]),
            (dictgen::InsensitiveStr::Ascii("brianer"), &["brainer"]),
            (
                dictgen::InsensitiveStr::Ascii("reciprocrate"),
                &["reciprocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cencter"), &["center"]),
            (dictgen::InsensitiveStr::Ascii("toughts"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("untuitively"),
                &["unintuitively", "intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnreability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associaiton"),
                &["associations", "association"],
            ),
            (dictgen::InsensitiveStr::Ascii("refreshs"), &["refreshes"]),
            (dictgen::InsensitiveStr::Ascii("hydrolics"), &["hydraulics"]),
            (
                dictgen::InsensitiveStr::Ascii("componentes"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exculsivly"),
                &["exclusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotication"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("americanus"), &["americans"]),
            (dictgen::InsensitiveStr::Ascii("erronenous"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("primtiive"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("cincinnatti"),
                &["cincinnati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapporving"),
                &["disapproving"],
            ),
            (dictgen::InsensitiveStr::Ascii("scedules"), &["schedules"]),
            (
                dictgen::InsensitiveStr::Ascii("conffigures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("targests"), &["targets"]),
            (dictgen::InsensitiveStr::Ascii("rocessor"), &["processor"]),
            (
                dictgen::InsensitiveStr::Ascii("administratiors"),
                &["administrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imputting"),
                &["inputting", "imputing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overhwlemingly"),
                &["overwhelmingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemementing"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyprocites"),
                &["hypocrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongnising"),
                &["recognising"],
            ),
            (dictgen::InsensitiveStr::Ascii("timdelta"), &["timedelta"]),
            (dictgen::InsensitiveStr::Ascii("fschk"), &["fsck"]),
            (dictgen::InsensitiveStr::Ascii("execuated"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("sqashing"), &["squashing"]),
            (dictgen::InsensitiveStr::Ascii("querst"), &["quest"]),
            (dictgen::InsensitiveStr::Ascii("porviding"), &["providing"]),
            (dictgen::InsensitiveStr::Ascii("sequecence"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("omplementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("recitfy"), &["rectify"]),
            (
                dictgen::InsensitiveStr::Ascii("specificatons"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inperfections"),
                &["imperfections"],
            ),
            (dictgen::InsensitiveStr::Ascii("standrd"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("immedeate"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("inisializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("confiures"), &["configures"]),
            (
                dictgen::InsensitiveStr::Ascii("bouns"),
                &["bounds", "bonus", "bound"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("economistes"),
                &["economists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derogetory"),
                &["derogatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envionmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("affinitiy"), &["affinity"]),
            (
                dictgen::InsensitiveStr::Ascii("comparsion"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("affinily"), &["affinity"]),
            (dictgen::InsensitiveStr::Ascii("decendent"), &["descendent"]),
            (
                dictgen::InsensitiveStr::Ascii("dependecies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinical"), &["cynical"]),
            (
                dictgen::InsensitiveStr::Ascii("aborigional"),
                &["aboriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prophechies"),
                &["prophecies"],
            ),
            (dictgen::InsensitiveStr::Ascii("disticnt"), &["distinct"]),
            (dictgen::InsensitiveStr::Ascii("alignemnt"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("numper"),
                &["number", "bumper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("bewteen"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("ines"), &["lines"]),
            (
                dictgen::InsensitiveStr::Ascii("interfase"),
                &["interfaces", "interface"],
            ),
            (dictgen::InsensitiveStr::Ascii("esimate"), &["estimate"]),
            (
                dictgen::InsensitiveStr::Ascii("custamizing"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authroisation"),
                &["authorisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("calss"), &["calls", "class"]),
            (
                dictgen::InsensitiveStr::Ascii("untranslateable"),
                &["untranslatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("declarate"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("attriutes"), &["attributes"]),
            (
                dictgen::InsensitiveStr::Ascii("caues"),
                &["causes", "cause"],
            ),
            (dictgen::InsensitiveStr::Ascii("becasuse"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("cleasne"), &["cleanse"]),
            (dictgen::InsensitiveStr::Ascii("widthes"), &["widths"]),
            (
                dictgen::InsensitiveStr::Ascii("discriminitory"),
                &["discriminatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accesiblity"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singeles"),
                &["singles", "signals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitalize"),
                &["reinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("findn"), &["find"]),
            (
                dictgen::InsensitiveStr::Ascii("convigures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacthing"), &["patching"]),
            (
                dictgen::InsensitiveStr::Ascii("scientifcaly"),
                &["scientifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitmains"), &["vitamins"]),
            (dictgen::InsensitiveStr::Ascii("miniscuel"), &["miniscule"]),
            (
                dictgen::InsensitiveStr::Ascii("geogcountry"),
                &["geocountry"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrything"), &["writhing"]),
            (dictgen::InsensitiveStr::Ascii("makrsman"), &["marksman"]),
            (
                dictgen::InsensitiveStr::Ascii("theorhetically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("occording"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("purcahed"), &["purchased"]),
            (dictgen::InsensitiveStr::Ascii("researved"), &["reserved"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratioms"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("canew"), &["canoe"]),
            (dictgen::InsensitiveStr::Ascii("acatemic"), &["academic"]),
            (dictgen::InsensitiveStr::Ascii("coexhists"), &["coexists"]),
            (
                dictgen::InsensitiveStr::Ascii("subcatogory"),
                &["subcategory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisssioning"),
                &["commissioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("stremaing"), &["streaming"]),
            (dictgen::InsensitiveStr::Ascii("dscribing"), &["describing"]),
            (dictgen::InsensitiveStr::Ascii("galcier"), &["glacier"]),
            (dictgen::InsensitiveStr::Ascii("sattisfied"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("reinfoce"), &["reinforce"]),
            (dictgen::InsensitiveStr::Ascii("ontainor"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("opinoins"), &["opinions"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuiliti"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qualificaitons"),
                &["qualifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifiating"),
                &["certificating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("feserialized"),
                &["deserialized", "reserialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("objectivs"), &["objectives"]),
            (
                dictgen::InsensitiveStr::Ascii("transmision"),
                &["transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("mectrics"), &["metrics"]),
            (dictgen::InsensitiveStr::Ascii("bload"), &["bloat"]),
            (dictgen::InsensitiveStr::Ascii("sinular"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("repitive"), &["repetitive"]),
            (dictgen::InsensitiveStr::Ascii("situatuion"), &["situation"]),
            (
                dictgen::InsensitiveStr::Ascii("capatilization"),
                &["capitalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("innvoate"), &["innovate"]),
            (
                dictgen::InsensitiveStr::Ascii("peridiocal"),
                &["periodical"],
            ),
            (dictgen::InsensitiveStr::Ascii("seccess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("embeder"), &["embedder"]),
            (dictgen::InsensitiveStr::Ascii("favritt"), &["favorite"]),
            (dictgen::InsensitiveStr::Ascii("regardnig"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrated"),
                &["aggregated", "aggravated"],
            ),
            (dictgen::InsensitiveStr::Ascii("capabably"), &["capably"]),
            (dictgen::InsensitiveStr::Ascii("specemin"), &["specimen"]),
            (dictgen::InsensitiveStr::Ascii("ummark"), &["unmark"]),
            (
                dictgen::InsensitiveStr::Ascii("wirtable"),
                &["writable", "writeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custoising"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("kidnappade"), &["kidnapped"]),
            (
                dictgen::InsensitiveStr::Ascii("cleande"),
                &["cleaned", "cleanse"],
            ),
            (dictgen::InsensitiveStr::Ascii("amercian"), &["american"]),
            (
                dictgen::InsensitiveStr::Ascii("palestiniens"),
                &["palestinians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archaoelogy"),
                &["archaeology", "archeology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhilates"),
                &["annihilates"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrerupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("bahaviour"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("blockes"),
                &["blockers", "blocked", "blocks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alighed"),
                &["aligned", "alighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reevalulate"),
                &["reevaluate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cacellation"),
                &["cancellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommandation"),
                &["recommendation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bargani"), &["bargain"]),
            (dictgen::InsensitiveStr::Ascii("esging"), &["edging"]),
            (dictgen::InsensitiveStr::Ascii("applogy"), &["apology"]),
            (dictgen::InsensitiveStr::Ascii("wirk"), &["work"]),
            (
                dictgen::InsensitiveStr::Ascii("utilsation"),
                &["utilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("comittees"), &["committees"]),
            (dictgen::InsensitiveStr::Ascii("cteator"), &["creator"]),
            (
                dictgen::InsensitiveStr::Ascii("condolances"),
                &["condolences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assiocated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frambuffer"),
                &["framebuffer"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsessie"), &["obsessive"]),
            (dictgen::InsensitiveStr::Ascii("diffulcty"), &["difficulty"]),
            (dictgen::InsensitiveStr::Ascii("miminal"), &["minimal"]),
            (
                dictgen::InsensitiveStr::Ascii("contributons"),
                &["contributors", "contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("stutdown"), &["shutdown"]),
            (
                dictgen::InsensitiveStr::Ascii("overcloked"),
                &["overclocked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desciptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proffessing"),
                &["professing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedrans"),
                &["tetrahedrons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrincs"),
                &["intrinsics", "instincts"],
            ),
            (dictgen::InsensitiveStr::Ascii("distractia"), &["district"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapolatin"),
                &["extrapolating", "extrapolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("clitoridis"), &["clitoris"]),
            (dictgen::InsensitiveStr::Ascii("entiy"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("startng"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("liberry"),
                &["library", "liberty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeuing"),
                &["neighbouring"],
            ),
            (dictgen::InsensitiveStr::Ascii("modualr"), &["modular"]),
            (
                dictgen::InsensitiveStr::Ascii("midfeilders"),
                &["midfielders"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifes"), &["lives"]),
            (dictgen::InsensitiveStr::Ascii("frequant"), &["frequent"]),
            (dictgen::InsensitiveStr::Ascii("dislpays"), &["displays"]),
            (
                dictgen::InsensitiveStr::Ascii("supporte"),
                &["supported", "supporter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interripts"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corretness"),
                &["correctness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probility"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classsic"),
                &["classics", "classic"],
            ),
            (dictgen::InsensitiveStr::Ascii("derivativ"), &["derivative"]),
            (dictgen::InsensitiveStr::Ascii("keyesnian"), &["keynesian"]),
            (dictgen::InsensitiveStr::Ascii("clent"), &["client"]),
            (dictgen::InsensitiveStr::Ascii("systen"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("unavaiable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrutors"),
                &["instructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadaquate"),
                &["inadequate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("errornously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validaitons"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soruces"),
                &["sources", "spruces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepertators"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntact"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("fautures"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("unrechable"),
                &["unreachable"],
            ),
            (dictgen::InsensitiveStr::Ascii("elelmental"), &["elemental"]),
            (
                dictgen::InsensitiveStr::Ascii("stratup"),
                &["startup", "stratum", "stratus"],
            ),
            (dictgen::InsensitiveStr::Ascii("becoms"), &["becomes"]),
            (dictgen::InsensitiveStr::Ascii("relizing"), &["realizing"]),
            (
                dictgen::InsensitiveStr::Ascii("understadnable"),
                &["understandable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ligher"),
                &["lighter", "liar", "liger"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramesr"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("aboslute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncalize"),
                &["canonicalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("benging"), &["binging"]),
            (
                dictgen::InsensitiveStr::Ascii("replicaitons"),
                &["replications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weaponaries"),
                &["weaponry", "weaponries"],
            ),
            (dictgen::InsensitiveStr::Ascii("famliy"), &["family"]),
            (dictgen::InsensitiveStr::Ascii("copyrigt"), &["copyright"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dyanamically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("comaparing"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("devleop"), &["develop"]),
            (
                dictgen::InsensitiveStr::Ascii("assocciates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomits"), &["commits"]),
            (
                dictgen::InsensitiveStr::Ascii("parentesis"),
                &["parentheses", "parenthesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cantainers"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downrgades"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("critcial"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("fianlised"), &["finalised"]),
            (dictgen::InsensitiveStr::Ascii("sqiushy"), &["squishy"]),
            (
                dictgen::InsensitiveStr::Ascii("intilisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpsonding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranlates"), &["translates"]),
            (dictgen::InsensitiveStr::Ascii("otpimizer"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("illegales"), &["illegals"]),
            (dictgen::InsensitiveStr::Ascii("italias"), &["italians"]),
            (
                dictgen::InsensitiveStr::Ascii("colletively"),
                &["collectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bangladash"),
                &["bangladesh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coommitting"),
                &["committing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminoloy"),
                &["terminology"],
            ),
            (dictgen::InsensitiveStr::Ascii("demension"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("genreator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("procesd"), &["proceeds"]),
            (dictgen::InsensitiveStr::Ascii("tortillia"), &["tortilla"]),
            (dictgen::InsensitiveStr::Ascii("valdation"), &["validation"]),
            (
                dictgen::InsensitiveStr::Ascii("compramised"),
                &["compromised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpaly"),
                &["counterplay"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfortablly"),
                &["comfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diectional"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaccuuming"), &["vacuuming"]),
            (dictgen::InsensitiveStr::Ascii("consifer"), &["consider"]),
            (dictgen::InsensitiveStr::Ascii("simbolic"), &["symbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("conveninent"),
                &["convenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperators"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("compatablity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewiev"), &["review"]),
            (
                dictgen::InsensitiveStr::Ascii("jusitification"),
                &["justification"],
            ),
            (dictgen::InsensitiveStr::Ascii("anc"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("skool"), &["school"]),
            (dictgen::InsensitiveStr::Ascii("govermetn"), &["government"]),
            (dictgen::InsensitiveStr::Ascii("metamage"), &["metagame"]),
            (dictgen::InsensitiveStr::Ascii("wrapps"), &["wraps"]),
            (dictgen::InsensitiveStr::Ascii("lolipop"), &["lollipop"]),
            (dictgen::InsensitiveStr::Ascii("definend"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("northerend"), &["northern"]),
            (dictgen::InsensitiveStr::Ascii("argumentos"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("cipheruite"),
                &["ciphersuite"],
            ),
            (dictgen::InsensitiveStr::Ascii("stategy"), &["strategy"]),
            (dictgen::InsensitiveStr::Ascii("reguard"), &["regard"]),
            (
                dictgen::InsensitiveStr::Ascii("collaborato"),
                &["collaboration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calssifying"),
                &["classifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantazising"),
                &["fantasising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authrorities"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oportunity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minneapols"),
                &["minneapolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("clearity"), &["clarity"]),
            (
                dictgen::InsensitiveStr::Ascii("continuse"),
                &["continues", "continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minization"),
                &["minimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contagioso"),
                &["contagious"],
            ),
            (dictgen::InsensitiveStr::Ascii("germanos"), &["germans"]),
            (
                dictgen::InsensitiveStr::Ascii("provinicial"),
                &["provincial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assotiating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("prolem"), &["problem"]),
            (
                dictgen::InsensitiveStr::Ascii("increadible"),
                &["incredible"],
            ),
            (dictgen::InsensitiveStr::Ascii("etablish"), &["establish"]),
            (
                dictgen::InsensitiveStr::Ascii("househoulds"),
                &["households"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripitive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redners"),
                &["renders", "redness"],
            ),
            (dictgen::InsensitiveStr::Ascii("beautful"), &["beautiful"]),
            (
                dictgen::InsensitiveStr::Ascii("identifiter"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("oredred"), &["ordered"]),
            (
                dictgen::InsensitiveStr::Ascii("comapnions"),
                &["companions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exludes"),
                &["excludes", "exudes"],
            ),
            (dictgen::InsensitiveStr::Ascii("calulator"), &["calculator"]),
            (dictgen::InsensitiveStr::Ascii("zealouts"), &["zealous"]),
            (dictgen::InsensitiveStr::Ascii("camelions"), &["chameleons"]),
            (
                dictgen::InsensitiveStr::Ascii("farhenheit"),
                &["fahrenheit"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfied"), &["verified"]),
            (
                dictgen::InsensitiveStr::Ascii("interperted"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranceivers"),
                &["transceivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("tradtions"), &["traditions"]),
            (dictgen::InsensitiveStr::Ascii("messsaged"), &["messaged"]),
            (dictgen::InsensitiveStr::Ascii("befried"), &["befriend"]),
            (dictgen::InsensitiveStr::Ascii("obssesive"), &["obsessive"]),
            (dictgen::InsensitiveStr::Ascii("guaranted"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("ankshiosly"), &["anxiously"]),
            (
                dictgen::InsensitiveStr::Ascii("simulater"),
                &["simulators", "simulated", "simulates", "simulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apostrobphe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imlementations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("flasely"), &["falsely"]),
            (dictgen::InsensitiveStr::Ascii("applyed"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("chechksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("aboves"), &["above"]),
            (dictgen::InsensitiveStr::Ascii("chercks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("promprted"), &["prompted"]),
            (dictgen::InsensitiveStr::Ascii("bechmarks"), &["benchmarks"]),
            (
                dictgen::InsensitiveStr::Ascii("pregressively"),
                &["progressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curruptible"),
                &["corruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("menioned"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("satisifed"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("actuion"), &["action"]),
            (dictgen::InsensitiveStr::Ascii("coalases"), &["coalesces"]),
            (
                dictgen::InsensitiveStr::Ascii("quarintine"),
                &["quarantine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwiting"),
                &["overwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undebiably"),
                &["undeniably"],
            ),
            (dictgen::InsensitiveStr::Ascii("repaires"), &["repairs"]),
            (dictgen::InsensitiveStr::Ascii("privde"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("protoco"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("quanity"), &["quantity"]),
            (dictgen::InsensitiveStr::Ascii("citeria"), &["criteria"]),
            (dictgen::InsensitiveStr::Ascii("analist"), &["analyst"]),
            (
                dictgen::InsensitiveStr::Ascii("counterfits"),
                &["counterfeits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoindents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("erroreous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("contemple"),
                &["contemplate"],
            ),
            (dictgen::InsensitiveStr::Ascii("achilleous"), &["achilles"]),
            (
                dictgen::InsensitiveStr::Ascii("siganlling"),
                &["signalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlockes"), &["unlocks"]),
            (dictgen::InsensitiveStr::Ascii("chercklist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("seriouly"), &["seriously"]),
            (
                dictgen::InsensitiveStr::Ascii("overwrittten"),
                &["overwritten"],
            ),
            (dictgen::InsensitiveStr::Ascii("lentiles"), &["lentils"]),
            (dictgen::InsensitiveStr::Ascii("preety"), &["pretty"]),
            (dictgen::InsensitiveStr::Ascii("comlicate"), &["complicate"]),
            (dictgen::InsensitiveStr::Ascii("spefcifics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuiton"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("litterfinger"),
                &["littlefinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accodomating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjascent"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("spetial"),
                &["spatial", "special"],
            ),
            (dictgen::InsensitiveStr::Ascii("amatur"), &["amateur"]),
            (
                dictgen::InsensitiveStr::Ascii("virualisation"),
                &["virtualisation", "visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("metod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("undoubtedbly"),
                &["undoubtedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("buiilding"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("statistircs"),
                &["statistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionalies"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("devitaion"), &["deviation"]),
            (
                dictgen::InsensitiveStr::Ascii("ure"),
                &["sure", "ire", "are", "urea", "rue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlighed"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("appluads"), &["applauds"]),
            (dictgen::InsensitiveStr::Ascii("startnig"), &["starting"]),
            (dictgen::InsensitiveStr::Ascii("referernce"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("hgolding"), &["holding"]),
            (
                dictgen::InsensitiveStr::Ascii("discredid"),
                &["discredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuztomizer"),
                &["customizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspont"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("synonoms"), &["synonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("cooordinate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("timneouts"), &["timeouts"]),
            (
                dictgen::InsensitiveStr::Ascii("gerenation"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integerates"),
                &["integrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("itegrated"), &["integrated"]),
            (
                dictgen::InsensitiveStr::Ascii("reconcilation"),
                &["reconciliation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pilup"), &["pileup"]),
            (dictgen::InsensitiveStr::Ascii("entbook"), &["netbook"]),
            (
                dictgen::InsensitiveStr::Ascii("prelimiary"),
                &["preliminary"],
            ),
            (dictgen::InsensitiveStr::Ascii("invole"), &["involve"]),
            (dictgen::InsensitiveStr::Ascii("anotation"), &["annotation"]),
            (
                dictgen::InsensitiveStr::Ascii("orcestrates"),
                &["orchestrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instuctions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("levetate"), &["levitate"]),
            (
                dictgen::InsensitiveStr::Ascii("caluclated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("delalyed"), &["delayed"]),
            (dictgen::InsensitiveStr::Ascii("eanabling"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("congugation"),
                &["conjugation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyporcite"), &["hypocrite"]),
            (
                dictgen::InsensitiveStr::Ascii("massachussetts"),
                &["massachusetts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represendatives"),
                &["representatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("expained"), &["explained"]),
            (
                dictgen::InsensitiveStr::Ascii("comparission"),
                &["comparisons", "comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrucutions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("agrv"), &["argv"]),
            (
                dictgen::InsensitiveStr::Ascii("imlimentation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatching"), &["detaching"]),
            (
                dictgen::InsensitiveStr::Ascii("operationnal"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seleting"),
                &["selecting", "deleting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miinimizing"),
                &["minimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("refleting"), &["reflecting"]),
            (
                dictgen::InsensitiveStr::Ascii("cockateels"),
                &["cockatiels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governemntal"),
                &["governmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("thugh"), &["though"]),
            (
                dictgen::InsensitiveStr::Ascii("averges"),
                &["averages", "avenges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriminate"),
                &["discriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciaed"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("soovineer"), &["souvenir"]),
            (dictgen::InsensitiveStr::Ascii("mutbale"), &["mutable"]),
            (
                dictgen::InsensitiveStr::Ascii("distintly"),
                &["distinctly", "distantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("violoating"), &["violating"]),
            (dictgen::InsensitiveStr::Ascii("dispeners"), &["dispenser"]),
            (dictgen::InsensitiveStr::Ascii("cacuasian"), &["caucasian"]),
            (dictgen::InsensitiveStr::Ascii("asterois"), &["asteroids"]),
            (
                dictgen::InsensitiveStr::Ascii("handlebards"),
                &["handlebars"],
            ),
            (dictgen::InsensitiveStr::Ascii("inidcator"), &["indicator"]),
            (dictgen::InsensitiveStr::Ascii("torpeados"), &["torpedoes"]),
            (dictgen::InsensitiveStr::Ascii("exclsuive"), &["exclusive"]),
            (dictgen::InsensitiveStr::Ascii("studnet"), &["student"]),
            (
                dictgen::InsensitiveStr::Ascii("costomization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("generage"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoweldge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwealming"),
                &["overwhelming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introduses"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondindly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyizes"), &["analyzes"]),
            (
                dictgen::InsensitiveStr::Ascii("dynamicaly"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("diposable"), &["disposable"]),
            (dictgen::InsensitiveStr::Ascii("editot"), &["editor"]),
            (
                dictgen::InsensitiveStr::Ascii("unoticeable"),
                &["unnoticeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photagrapher"),
                &["photographer"],
            ),
            (dictgen::InsensitiveStr::Ascii("medioce"), &["mediocre"]),
            (
                dictgen::InsensitiveStr::Ascii("immadiatly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanpsnots"), &["snapshots"]),
            (dictgen::InsensitiveStr::Ascii("rebild"), &["rebuild"]),
            (
                dictgen::InsensitiveStr::Ascii("combability"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convergance"),
                &["convergence"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimizing"), &["optimizing"]),
            (dictgen::InsensitiveStr::Ascii("tockens"), &["tokens"]),
            (dictgen::InsensitiveStr::Ascii("framented"), &["fragmented"]),
            (dictgen::InsensitiveStr::Ascii("quesrs"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("exicutable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("subscritpt"), &["subscript"]),
            (dictgen::InsensitiveStr::Ascii("nderline"), &["underline"]),
            (dictgen::InsensitiveStr::Ascii("itegers"), &["integers"]),
            (dictgen::InsensitiveStr::Ascii("averhaul"), &["overhaul"]),
            (
                dictgen::InsensitiveStr::Ascii("narls"),
                &[
                    "earls", "farls", "gnarls", "marls", "nails", "narks", "snarls",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("backware"), &["backward"]),
            (dictgen::InsensitiveStr::Ascii("neigbouds"), &["neighbours"]),
            (
                dictgen::InsensitiveStr::Ascii("transparancy"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakcaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("survivabillity"),
                &["survivability"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuld"), &["rebuild"]),
            (
                dictgen::InsensitiveStr::Ascii("constents"),
                &["constants", "consents", "contents"],
            ),
            (dictgen::InsensitiveStr::Ascii("lasonya"), &["lasagna"]),
            (dictgen::InsensitiveStr::Ascii("firsbee"), &["frisbee"]),
            (
                dictgen::InsensitiveStr::Ascii("sucesssfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("califronia"),
                &["california"],
            ),
            (dictgen::InsensitiveStr::Ascii("reevalute"), &["reevaluate"]),
            (dictgen::InsensitiveStr::Ascii("gioen"), &["given"]),
            (dictgen::InsensitiveStr::Ascii("narctoics"), &["narcotics"]),
            (
                dictgen::InsensitiveStr::Ascii("calcultation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nueronal"), &["neuronal"]),
            (dictgen::InsensitiveStr::Ascii("descrie"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("artibrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("righmost"), &["rightmost"]),
            (dictgen::InsensitiveStr::Ascii("inomplete"), &["incomplete"]),
            (dictgen::InsensitiveStr::Ascii("analsying"), &["analysing"]),
            (dictgen::InsensitiveStr::Ascii("secratary"), &["secretary"]),
            (dictgen::InsensitiveStr::Ascii("evaluting"), &["evaluating"]),
            (
                dictgen::InsensitiveStr::Ascii("multliplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancllation"),
                &["cancellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropration"),
                &["appropriation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wraapps"), &["wraps"]),
            (dictgen::InsensitiveStr::Ascii("nofity"), &["notify"]),
            (dictgen::InsensitiveStr::Ascii("prinicpal"), &["principal"]),
            (dictgen::InsensitiveStr::Ascii("comon"), &["common"]),
            (dictgen::InsensitiveStr::Ascii("blieving"), &["believing"]),
            (dictgen::InsensitiveStr::Ascii("aginst"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumnet"),
                &["instrument"],
            ),
            (dictgen::InsensitiveStr::Ascii("playlst"), &["playlist"]),
            (
                dictgen::InsensitiveStr::Ascii("predicitons"),
                &["predictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("optios"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("technnical"), &["technical"]),
            (
                dictgen::InsensitiveStr::Ascii("transmiters"),
                &["transmitters"],
            ),
            (dictgen::InsensitiveStr::Ascii("heroicus"), &["heroics"]),
            (dictgen::InsensitiveStr::Ascii("shpaes"), &["shapes"]),
            (
                dictgen::InsensitiveStr::Ascii("sepeartion"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("uites"), &["suites"]),
            (
                dictgen::InsensitiveStr::Ascii("acquifer"),
                &["acquire", "aquifer"],
            ),
            (dictgen::InsensitiveStr::Ascii("plebicite"), &["plebiscite"]),
            (dictgen::InsensitiveStr::Ascii("annnounced"), &["announced"]),
            (dictgen::InsensitiveStr::Ascii("magnatude"), &["magnitude"]),
            (
                dictgen::InsensitiveStr::Ascii("funtionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpectives"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessealted"),
                &["tessellated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cought"),
                &["caught", "cough", "fought"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbeus"), &["neighbours"]),
            (
                dictgen::InsensitiveStr::Ascii("dirrectories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("extepects"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("legac"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("ireelevant"),
                &["irrelevant"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurding"), &["recurring"]),
            (dictgen::InsensitiveStr::Ascii("fontonfig"), &["fontconfig"]),
            (dictgen::InsensitiveStr::Ascii("quesiton"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("specal"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("asborbs"), &["absorbs"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendly"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("sizebale"), &["sizeable"]),
            (
                dictgen::InsensitiveStr::Ascii("alternador"),
                &["alternator"],
            ),
            (dictgen::InsensitiveStr::Ascii("matirx"), &["matrix"]),
            (
                dictgen::InsensitiveStr::Ascii("trememdous"),
                &["tremendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("proejcted"), &["projected"]),
            (
                dictgen::InsensitiveStr::Ascii("dynamincally"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhaution"), &["exhaustion"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawalls"),
                &["withdrawals"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuggs"), &["debugs"]),
            (dictgen::InsensitiveStr::Ascii("clusetr"), &["cluster"]),
            (dictgen::InsensitiveStr::Ascii("eqipped"), &["equipped"]),
            (dictgen::InsensitiveStr::Ascii("wbsites"), &["websites"]),
            (
                dictgen::InsensitiveStr::Ascii("oganization"),
                &["organization"],
            ),
            (dictgen::InsensitiveStr::Ascii("releated"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("photographied"),
                &["photographed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retcieved"),
                &["received", "retrieved"],
            ),
            (dictgen::InsensitiveStr::Ascii("economicos"), &["economics"]),
            (dictgen::InsensitiveStr::Ascii("hydregon"), &["hydrogen"]),
            (
                dictgen::InsensitiveStr::Ascii("organisaiton"),
                &["organisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconviniance"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("begiin"), &["begin"]),
            (dictgen::InsensitiveStr::Ascii("sawtaying"), &["sautéing"]),
            (
                dictgen::InsensitiveStr::Ascii("activiation"),
                &["activation"],
            ),
            (dictgen::InsensitiveStr::Ascii("instancen"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("farction"),
                &["fraction", "faction"],
            ),
            (dictgen::InsensitiveStr::Ascii("respoduce"), &["reproduce"]),
            (dictgen::InsensitiveStr::Ascii("tortiose"), &["tortoise"]),
            (dictgen::InsensitiveStr::Ascii("picknicker"), &["picnicker"]),
            (dictgen::InsensitiveStr::Ascii("sidelinjen"), &["sideline"]),
            (dictgen::InsensitiveStr::Ascii("purtained"), &["pertained"]),
            (
                dictgen::InsensitiveStr::Ascii("sensationel"),
                &["sensational"],
            ),
            (dictgen::InsensitiveStr::Ascii("soldger"), &["soldier"]),
            (dictgen::InsensitiveStr::Ascii("lepoard"), &["leopard"]),
            (dictgen::InsensitiveStr::Ascii("bactrack"), &["backtrack"]),
            (
                dictgen::InsensitiveStr::Ascii("consideratin"),
                &["considerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorthm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("allegiences"),
                &["allegiances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleeting"),
                &["completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("deciede"), &["decide"]),
            (
                dictgen::InsensitiveStr::Ascii("asignmends"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("extimates"), &["estimates"]),
            (dictgen::InsensitiveStr::Ascii("farthere"), &["farther"]),
            (dictgen::InsensitiveStr::Ascii("dingees"), &["dinghies"]),
            (dictgen::InsensitiveStr::Ascii("tranclate"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("proyects"),
                &["projects", "protects"],
            ),
            (dictgen::InsensitiveStr::Ascii("presentase"), &["presents"]),
            (dictgen::InsensitiveStr::Ascii("exsists"), &["exists"]),
            (
                dictgen::InsensitiveStr::Ascii("onomanopea"),
                &["onomatopoeia"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacoon"), &["cocoon"]),
            (
                dictgen::InsensitiveStr::Ascii("pnuemonic"),
                &["mnemonic", "pneumonic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrube"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speadsheets"),
                &["spreadsheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusstom"), &["custom"]),
            (
                dictgen::InsensitiveStr::Ascii("charactiristically"),
                &["characteristically"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemtent"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("hurrikanes"),
                &["hurricanes"],
            ),
            (dictgen::InsensitiveStr::Ascii("carmony"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("presciuosly"),
                &["preciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamendalist"),
                &["fundamentalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("beieve"), &["believe"]),
            (
                dictgen::InsensitiveStr::Ascii("simultanous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapoined"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartiuons"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flter"),
                &["filter", "falter", "flutter", "flatter", "floater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cativating"),
                &["activating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedections"),
                &["detections"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartmaan"), &["cartman"]),
            (dictgen::InsensitiveStr::Ascii("poisitions"), &["positions"]),
            (dictgen::InsensitiveStr::Ascii("pevious"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("acender"), &["ascender"]),
            (dictgen::InsensitiveStr::Ascii("mamaged"), &["managed"]),
            (dictgen::InsensitiveStr::Ascii("efect"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("paramemeters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimatione"),
                &["estimation"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfom"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("compatibe"), &["compatible"]),
            (dictgen::InsensitiveStr::Ascii("ofrom"), &["from"]),
            (
                dictgen::InsensitiveStr::Ascii("usnupported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("solvin"), &["solving"]),
            (dictgen::InsensitiveStr::Ascii("diganosed"), &["diagnosed"]),
            (dictgen::InsensitiveStr::Ascii("veature"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("actresss"),
                &["actresses", "actress"],
            ),
            (dictgen::InsensitiveStr::Ascii("nohypen"), &["nohyphen"]),
            (
                dictgen::InsensitiveStr::Ascii("bostom"),
                &["boston", "bottom"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrupted"), &["encrypted"]),
            (dictgen::InsensitiveStr::Ascii("winndows"), &["windows"]),
            (dictgen::InsensitiveStr::Ascii("exceuctive"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("powerhorse"),
                &["powerhouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanciation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lightnting"),
                &["lightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellipis"), &["ellipsis"]),
            (
                dictgen::InsensitiveStr::Ascii("appliations"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("notasion"), &["notation"]),
            (
                dictgen::InsensitiveStr::Ascii("reevaulating"),
                &["reevaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimesnsions"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("delimitis"), &["delimits"]),
            (dictgen::InsensitiveStr::Ascii("solate"), &["isolate"]),
            (dictgen::InsensitiveStr::Ascii("typscript"), &["typescript"]),
            (dictgen::InsensitiveStr::Ascii("asssets"), &["assets"]),
            (dictgen::InsensitiveStr::Ascii("datatpe"), &["datatype"]),
            (
                dictgen::InsensitiveStr::Ascii("defnitively"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("datastrcuture"),
                &["datastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("exerternal"), &["external"]),
            (dictgen::InsensitiveStr::Ascii("neum"), &["enum"]),
            (dictgen::InsensitiveStr::Ascii("paresr"), &["parser"]),
            (
                dictgen::InsensitiveStr::Ascii("altenately"),
                &["alternately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traingulate"),
                &["triangulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wieghed"), &["weighed"]),
            (
                dictgen::InsensitiveStr::Ascii("resonsibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("sarimony"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("muc"), &["much"]),
            (dictgen::InsensitiveStr::Ascii("locaizes"), &["localizes"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheviing"), &["achieving"]),
            (dictgen::InsensitiveStr::Ascii("pyhsical"), &["physical"]),
            (dictgen::InsensitiveStr::Ascii("maxinums"), &["maximums"]),
            (
                dictgen::InsensitiveStr::Ascii("assits"),
                &["assist", "assists"],
            ),
            (dictgen::InsensitiveStr::Ascii("decieved"), &["deceived"]),
            (
                dictgen::InsensitiveStr::Ascii("alterantives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("beggins"), &["begins"]),
            (
                dictgen::InsensitiveStr::Ascii("yourselfe"),
                &["yourself", "yourselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentinte"), &["sentient"]),
            (dictgen::InsensitiveStr::Ascii("shifs"), &["shifts"]),
            (dictgen::InsensitiveStr::Ascii("windwos"), &["windows"]),
            (dictgen::InsensitiveStr::Ascii("strech"), &["stretch"]),
            (dictgen::InsensitiveStr::Ascii("froce"), &["force"]),
            (
                dictgen::InsensitiveStr::Ascii("propectuses"),
                &["prospectuses"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpter"), &["adapter"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("grizzley"), &["grizzly"]),
            (dictgen::InsensitiveStr::Ascii("replayes"), &["replays"]),
            (
                dictgen::InsensitiveStr::Ascii("corralation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffuculty"),
                &["difficulty"],
            ),
            (dictgen::InsensitiveStr::Ascii("inraged"), &["enraged"]),
            (dictgen::InsensitiveStr::Ascii("journies"), &["journeys"]),
            (
                dictgen::InsensitiveStr::Ascii("comression"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminise"), &["reminisce"]),
            (dictgen::InsensitiveStr::Ascii("inbraces"), &["embraces"]),
            (
                dictgen::InsensitiveStr::Ascii("isplayed"),
                &["displayed", "misplayed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unpridictable"),
                &["unpredictable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfotunately"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("behaiour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("tcppcheck"), &["cppcheck"]),
            (
                dictgen::InsensitiveStr::Ascii("recommmended"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("legionis"), &["legions"]),
            (
                dictgen::InsensitiveStr::Ascii("evisceratin"),
                &["eviscerating", "evisceration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissertaion"),
                &["dissertation"],
            ),
            (dictgen::InsensitiveStr::Ascii("confir"), &["confirm"]),
            (dictgen::InsensitiveStr::Ascii("experance"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("philosophycal"),
                &["philosophically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privaleges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("falltrough"),
                &["fallthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("repbulic"), &["republic"]),
            (dictgen::InsensitiveStr::Ascii("loobpack"), &["loopback"]),
            (dictgen::InsensitiveStr::Ascii("comnnected"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("visuallisation"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("precisson"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("forwrad"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("converation"),
                &["conversation"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectpro"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("hipopotamus"),
                &["hippopotamus"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsteamed"), &["upstreamed"]),
            (
                dictgen::InsensitiveStr::Ascii("recommentation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participaste"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertaion"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("operatins"), &["operations"]),
            (dictgen::InsensitiveStr::Ascii("modifified"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("ecouragements"),
                &["encouragements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailzation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("tentacuel"), &["tentacle"]),
            (
                dictgen::InsensitiveStr::Ascii("disfunctional"),
                &["dysfunctional"],
            ),
            (dictgen::InsensitiveStr::Ascii("constansts"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("goddman"), &["goddamn"]),
            (
                dictgen::InsensitiveStr::Ascii("patriarhal"),
                &["patriarchal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inequilvalent"),
                &["inequivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pitchfords"),
                &["pitchforks"],
            ),
            (dictgen::InsensitiveStr::Ascii("methon"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("attatch"), &["attach"]),
            (
                dictgen::InsensitiveStr::Ascii("interpolatation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deconding"), &["decoding"]),
            (
                dictgen::InsensitiveStr::Ascii("chocalates"),
                &["chocolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("servises"), &["services"]),
            (dictgen::InsensitiveStr::Ascii("annonomus"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("unessecarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validatin"),
                &["validation", "validating"],
            ),
            (dictgen::InsensitiveStr::Ascii("inifinite"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("diappearing"),
                &["disappearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("adverties"), &["advertises"]),
            (dictgen::InsensitiveStr::Ascii("intenion"), &["intention"]),
            (dictgen::InsensitiveStr::Ascii("chizling"), &["chiseling"]),
            (
                dictgen::InsensitiveStr::Ascii("incrementers"),
                &["increments", "incrementors"],
            ),
            (dictgen::InsensitiveStr::Ascii("propechy"), &["prophecy"]),
            (dictgen::InsensitiveStr::Ascii("depndents"), &["dependents"]),
            (dictgen::InsensitiveStr::Ascii("exectution"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("positves"), &["positives"]),
            (
                dictgen::InsensitiveStr::Ascii("inadecuate"),
                &["inadequate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvineances"),
                &["inconveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("kentuckey"), &["kentucky"]),
            (dictgen::InsensitiveStr::Ascii("generatng"), &["generating"]),
            (
                dictgen::InsensitiveStr::Ascii("hierachycal"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disicplines"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("peanuchle"), &["pinochle"]),
            (dictgen::InsensitiveStr::Ascii("perverst"), &["pervert"]),
            (dictgen::InsensitiveStr::Ascii("bffer"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("stroylines"),
                &["storylines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifuable"),
                &["modifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("osffsetting"),
                &["offsetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatin"),
                &["calculations", "calculating", "calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("collectng"), &["collecting"]),
            (
                dictgen::InsensitiveStr::Ascii("pronte"),
                &["prone", "pronate", "pronto", "ponte"],
            ),
            (dictgen::InsensitiveStr::Ascii("improoves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("insterrupt"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("progagating"),
                &["propagating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratiutous"),
                &["gratuitous"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrrupts"), &["interrupts"]),
            (
                dictgen::InsensitiveStr::Ascii("componenet"),
                &["components", "component"],
            ),
            (dictgen::InsensitiveStr::Ascii("invarant"), &["invariant"]),
            (dictgen::InsensitiveStr::Ascii("equiment"), &["equipment"]),
            (
                dictgen::InsensitiveStr::Ascii("incoherantly"),
                &["incoherently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scatcher"),
                &["catcher", "scratcher", "sketcher"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependened"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("delgating"), &["delegating"]),
            (
                dictgen::InsensitiveStr::Ascii("strawbarry"),
                &["strawberry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohpecies"),
                &["prophecies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("threasholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalantly"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("footprinst"),
                &["footprints"],
            ),
            (dictgen::InsensitiveStr::Ascii("excectors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("inkompatible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elemente"),
                &["element", "elements"],
            ),
            (dictgen::InsensitiveStr::Ascii("enumurate"), &["enumerate"]),
            (
                dictgen::InsensitiveStr::Ascii("associdates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("viewpoit"), &["viewpoints"]),
            (dictgen::InsensitiveStr::Ascii("dictaters"), &["dictates"]),
            (dictgen::InsensitiveStr::Ascii("schedualed"), &["scheduled"]),
            (
                dictgen::InsensitiveStr::Ascii("refurbised"),
                &["refurbished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propostions"),
                &["proportions"],
            ),
            (dictgen::InsensitiveStr::Ascii("styilistic"), &["stylistic"]),
            (dictgen::InsensitiveStr::Ascii("cyrrilic"), &["cyrillic"]),
            (dictgen::InsensitiveStr::Ascii("algotighm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("unnofficial"),
                &["unofficial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corespondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("austrailia"), &["australia"]),
            (
                dictgen::InsensitiveStr::Ascii("exaggurated"),
                &["exaggerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticiaption"),
                &["anticipation"],
            ),
            (dictgen::InsensitiveStr::Ascii("parametter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("custoizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("certainy"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("informaion"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("becuase"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("gorwing"), &["growing"]),
            (dictgen::InsensitiveStr::Ascii("colectors"), &["collectors"]),
            (
                dictgen::InsensitiveStr::Ascii("converston"),
                &["conversation", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("spoaces"), &["spaces"]),
            (
                dictgen::InsensitiveStr::Ascii("underscrore"),
                &["underscore"],
            ),
            (dictgen::InsensitiveStr::Ascii("severiy"), &["severity"]),
            (
                dictgen::InsensitiveStr::Ascii("caracteristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serailzied"),
                &["serialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealearators"),
                &["ameliorators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kimerrically"),
                &["chimerically"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyelnol"), &["tylenol"]),
            (dictgen::InsensitiveStr::Ascii("expeced"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("dispse"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("tramsmit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("inexperineced"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("adequat"), &["adequate"]),
            (
                dictgen::InsensitiveStr::Ascii("thasnk"),
                &["thanks", "thank"],
            ),
            (dictgen::InsensitiveStr::Ascii("acclerate"), &["accelerate"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotatible"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbacsks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("indentifer"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("esclude"), &["exclude"]),
            (dictgen::InsensitiveStr::Ascii("carefullly"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("screenhots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("baesd"), &["based"]),
            (
                dictgen::InsensitiveStr::Ascii("oficionado"),
                &["aficionado"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilaizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hamburgeres"),
                &["hamburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("squres"),
                &["squares", "squires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benfit"),
                &["benefit", "befit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outragious"),
                &["outrageous"],
            ),
            (dictgen::InsensitiveStr::Ascii("trian"), &["train", "trial"]),
            (
                dictgen::InsensitiveStr::Ascii("repotes"),
                &["reports", "remotes", "reputes"],
            ),
            (dictgen::InsensitiveStr::Ascii("incative"), &["inactive"]),
            (
                dictgen::InsensitiveStr::Ascii("innitializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hunagry"), &["hungary"]),
            (
                dictgen::InsensitiveStr::Ascii("adapation"),
                &["adaptation", "adoption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recriational"),
                &["recreational"],
            ),
            (dictgen::InsensitiveStr::Ascii("visibley"), &["visibly"]),
            (dictgen::InsensitiveStr::Ascii("methond"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("caffeinne"), &["caffeine"]),
            (dictgen::InsensitiveStr::Ascii("exceuction"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("juipter"), &["jupiter"]),
            (dictgen::InsensitiveStr::Ascii("analaysis"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("definicon"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("approppriately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("beseiging"), &["besieging"]),
            (
                dictgen::InsensitiveStr::Ascii("disproportianate"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("salavge"), &["salvage"]),
            (
                dictgen::InsensitiveStr::Ascii("informatiomal"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthinically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemetns"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("docuemnting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satellitte"),
                &["satellites", "satellite"],
            ),
            (dictgen::InsensitiveStr::Ascii("hundreths"), &["hundredths"]),
            (
                dictgen::InsensitiveStr::Ascii("inefficent"),
                &["inefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difine"),
                &["define", "divine"],
            ),
            (dictgen::InsensitiveStr::Ascii("forteen"), &["fourteen"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatabilities"),
                &["incompatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("depsite"), &["despite"]),
            (dictgen::InsensitiveStr::Ascii("firware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("mplement"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenetions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommndation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactiristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontaneuos"),
                &["spontaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronized"),
                &["synchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffferences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaparators"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconcistently"),
                &["inconsistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trajatories"),
                &["trajectories"],
            ),
            (dictgen::InsensitiveStr::Ascii("deffinite"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("generaotr"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("embdedding"), &["embedding"]),
            (dictgen::InsensitiveStr::Ascii("uniuqe"), &["unique"]),
            (dictgen::InsensitiveStr::Ascii("taryvon"), &["trayvon"]),
            (
                dictgen::InsensitiveStr::Ascii("calculuators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("enew"), &["new"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawles"),
                &["withdrawals"],
            ),
            (dictgen::InsensitiveStr::Ascii("licemses"), &["licenses"]),
            (dictgen::InsensitiveStr::Ascii("dirctors"), &["directors"]),
            (
                dictgen::InsensitiveStr::Ascii("weaknessses"),
                &["weaknesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("posiitives"), &["positives"]),
            (dictgen::InsensitiveStr::Ascii("animetor"), &["animator"]),
            (dictgen::InsensitiveStr::Ascii("releavant"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("conecting"), &["connecting"]),
            (
                dictgen::InsensitiveStr::Ascii("repeadetly"),
                &["repeatedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("opiniones"), &["opinions"]),
            (
                dictgen::InsensitiveStr::Ascii("themselfe"),
                &["themselves", "themself"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afforts"),
                &["affords", "efforts"],
            ),
            (dictgen::InsensitiveStr::Ascii("homniem"), &["hominem"]),
            (dictgen::InsensitiveStr::Ascii("araound"), &["around"]),
            (
                dictgen::InsensitiveStr::Ascii("devrices"),
                &["devices", "dervices"],
            ),
            (dictgen::InsensitiveStr::Ascii("ststions"), &["stations"]),
            (
                dictgen::InsensitiveStr::Ascii("overpowred"),
                &["overpowered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeition"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosed"), &["closed"]),
            (
                dictgen::InsensitiveStr::Ascii("personel"),
                &["personnel", "personal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rollercoaser"),
                &["rollercoaster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("halluzination"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuilts"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("encyclopeadia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialsed"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imobilisation"),
                &["immobilisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projetcting"),
                &["projecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("depressin"), &["depression"]),
            (dictgen::InsensitiveStr::Ascii("nutirent"), &["nutrient"]),
            (
                dictgen::InsensitiveStr::Ascii("identidies"),
                &["identifies", "identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomated"), &["automated"]),
            (dictgen::InsensitiveStr::Ascii("consise"), &["concise"]),
            (dictgen::InsensitiveStr::Ascii("critism"), &["criticism"]),
            (
                dictgen::InsensitiveStr::Ascii("substansive"),
                &["substantive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vitualises"),
                &["virtualises", "visualises"],
            ),
            (dictgen::InsensitiveStr::Ascii("detorit"), &["detroit"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponed"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("itegrity"), &["integrity"]),
            (dictgen::InsensitiveStr::Ascii("oour"), &["our"]),
            (
                dictgen::InsensitiveStr::Ascii("systhemwindow"),
                &["systemwindow"],
            ),
            (dictgen::InsensitiveStr::Ascii("destryong"), &["destroying"]),
            (dictgen::InsensitiveStr::Ascii("relaesing"), &["releasing"]),
            (
                dictgen::InsensitiveStr::Ascii("documenations"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("protedcted"), &["protected"]),
            (dictgen::InsensitiveStr::Ascii("statemets"), &["statements"]),
            (
                dictgen::InsensitiveStr::Ascii("adjustible"),
                &["adjustable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backslases"),
                &["backslashes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commnadline"),
                &["commandline"],
            ),
            (dictgen::InsensitiveStr::Ascii("elimiates"), &["eliminates"]),
            (
                dictgen::InsensitiveStr::Ascii("teleportaton"),
                &["teleportation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gziniflate"), &["gzinflate"]),
            (dictgen::InsensitiveStr::Ascii("certiain"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("tiome"), &["time", "tome"]),
            (dictgen::InsensitiveStr::Ascii("assisstant"), &["assistant"]),
            (dictgen::InsensitiveStr::Ascii("pople"), &["people"]),
            (dictgen::InsensitiveStr::Ascii("unwrappin"), &["unwrapping"]),
            (dictgen::InsensitiveStr::Ascii("achievd"), &["achieved"]),
            (
                dictgen::InsensitiveStr::Ascii("accumuated"),
                &["accumulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destiantions"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("havee"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("collaobrative"),
                &["collaborative"],
            ),
            (dictgen::InsensitiveStr::Ascii("nane"), &["name"]),
            (dictgen::InsensitiveStr::Ascii("fedorahs"), &["fedoras"]),
            (
                dictgen::InsensitiveStr::Ascii("truamatized"),
                &["traumatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contignous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deminstrations"),
                &["demonstrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("settele"), &["settle"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilitys"),
                &["vulnerabilities", "vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("vanadlism"), &["vandalism"]),
            (
                dictgen::InsensitiveStr::Ascii("checherboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccord"), &["accord"]),
            (
                dictgen::InsensitiveStr::Ascii("documentatons"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apreciative"),
                &["appreciative"],
            ),
            (dictgen::InsensitiveStr::Ascii("bastino"), &["bastion"]),
            (dictgen::InsensitiveStr::Ascii("dokc"), &["dock"]),
            (dictgen::InsensitiveStr::Ascii("availalbel"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("spontanious"),
                &["spontaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sytactical"),
                &["syntactical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequensing"),
                &["sequencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("excistence"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("constitues"),
                &["constitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("umcomfortably"),
                &["uncomfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("reanmed"), &["renamed"]),
            (
                dictgen::InsensitiveStr::Ascii("contened"),
                &["contained", "contend"],
            ),
            (dictgen::InsensitiveStr::Ascii("barnches"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("whislist"), &["wishlist"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitoned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("redircts"), &["redirects"]),
            (dictgen::InsensitiveStr::Ascii("boundart"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("extravigant"),
                &["extravagant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cateogorically"),
                &["categorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preporation"),
                &["preparation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstrution"),
                &["reconstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relevation"),
                &["revelation"],
            ),
            (dictgen::InsensitiveStr::Ascii("communit"), &["communist"]),
            (
                dictgen::InsensitiveStr::Ascii("trogliditic"),
                &["troglodytic"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgarades"), &["upgrades"]),
            (
                dictgen::InsensitiveStr::Ascii("paralelized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulatiors"),
                &["regulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribtuion"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poossibility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referended"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("redneckers"), &["rednecks"]),
            (dictgen::InsensitiveStr::Ascii("ownner"), &["owner"]),
            (dictgen::InsensitiveStr::Ascii("publising"), &["publishing"]),
            (
                dictgen::InsensitiveStr::Ascii("distinugish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesernames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("somethinbg"),
                &["somethings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequentually"),
                &["sequentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitification"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teamfigths"),
                &["teamfights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egistering"),
                &["registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("consisten"), &["consistent"]),
            (dictgen::InsensitiveStr::Ascii("muiltiply"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("esitmates"), &["estimates"]),
            (
                dictgen::InsensitiveStr::Ascii("parenthese"),
                &["parentheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidenty"),
                &["presidency"],
            ),
            (dictgen::InsensitiveStr::Ascii("vulerable"), &["vulnerable"]),
            (
                dictgen::InsensitiveStr::Ascii("completd"),
                &["complete", "completed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentrant"),
                &["concentrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algotighmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("mointors"), &["monitors"]),
            (
                dictgen::InsensitiveStr::Ascii("approximatly"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subitle"),
                &["subtitle", "subtle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasferring"),
                &["transferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("passagens"), &["passages"]),
            (
                dictgen::InsensitiveStr::Ascii("populatoin"),
                &["populations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gunanine"), &["guanine"]),
            (dictgen::InsensitiveStr::Ascii("proeprty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("sturucture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("authenentication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configursing"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("bolow"), &["below"]),
            (
                dictgen::InsensitiveStr::Ascii("guareentees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("entoties"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("rearranement"),
                &["rearrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referncing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleanliess"),
                &["cleanliness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwritren"),
                &["overwritten"],
            ),
            (dictgen::InsensitiveStr::Ascii("sring"), &["string"]),
            (
                dictgen::InsensitiveStr::Ascii("physcologist"),
                &["psychologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondign"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deinitalizes"),
                &["deinitializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improovment"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("parsial"), &["partial"]),
            (dictgen::InsensitiveStr::Ascii("releations"), &["relations"]),
            (
                dictgen::InsensitiveStr::Ascii("captialises"),
                &["capitalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("veyr"), &["very"]),
            (dictgen::InsensitiveStr::Ascii("arguements"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("sisteming"),
                &["systemic", "stemming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgrated"),
                &["downgrade", "downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("passivs"), &["passives"]),
            (
                dictgen::InsensitiveStr::Ascii("unoriginial"),
                &["unoriginal"],
            ),
            (dictgen::InsensitiveStr::Ascii("snese"), &["sneeze"]),
            (
                dictgen::InsensitiveStr::Ascii("cancelleable"),
                &["cancellable"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisonis"), &["poisons"]),
            (dictgen::InsensitiveStr::Ascii("channe"), &["channel"]),
            (
                dictgen::InsensitiveStr::Ascii("apppendages"),
                &["appendages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternativets"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("enumate"), &["enumerate"]),
            (
                dictgen::InsensitiveStr::Ascii("depandantly"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metabolics"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("embrago"), &["embargo"]),
            (dictgen::InsensitiveStr::Ascii("mismactch"), &["mismatch"]),
            (
                dictgen::InsensitiveStr::Ascii("retribvtion"),
                &["retribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("setyp"), &["setup"]),
            (
                dictgen::InsensitiveStr::Ascii("comressing"),
                &["compressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("curently"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("libraire"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("pyrimads"), &["pyramids"]),
            (dictgen::InsensitiveStr::Ascii("powefull"), &["powerful"]),
            (
                dictgen::InsensitiveStr::Ascii("declartion"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("fianlize"), &["finalize"]),
            (
                dictgen::InsensitiveStr::Ascii("intstrumentation"),
                &["instrumentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcultates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geneic"),
                &["generic", "genetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutiplayer"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbsting"),
                &["masturbating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pairokeelity"),
                &["parochiality"],
            ),
            (dictgen::InsensitiveStr::Ascii("iwth"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("extrimities"),
                &["extremities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphrial"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("themperatures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("mircowave"), &["microwave"]),
            (
                dictgen::InsensitiveStr::Ascii("miscallenous"),
                &["miscellaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rseolution"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("desending"), &["descending"]),
            (
                dictgen::InsensitiveStr::Ascii("initilaizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoligical"),
                &["theological"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdules"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("alrelady"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("identtified"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("apdats"), &["adapts"]),
            (dictgen::InsensitiveStr::Ascii("skipeable"), &["skippable"]),
            (dictgen::InsensitiveStr::Ascii("refenence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("calagry"), &["calgary"]),
            (dictgen::InsensitiveStr::Ascii("criticable"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("individualty"),
                &["individuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenendecy"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implentation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("numercial"), &["numerical"]),
            (dictgen::InsensitiveStr::Ascii("arcitect"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("avialble"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("contraciction"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtaulization"),
                &["virtualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misisonaries"),
                &["missionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("normnal"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("hamburgr"), &["hamburgers"]),
            (dictgen::InsensitiveStr::Ascii("exprecting"), &["expecting"]),
            (dictgen::InsensitiveStr::Ascii("recommand"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("diagramms"), &["diagrams"]),
            (dictgen::InsensitiveStr::Ascii("filese"), &["files"]),
            (dictgen::InsensitiveStr::Ascii("abslote"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("sucesss"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("implemantation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("famlilies"), &["families"]),
            (dictgen::InsensitiveStr::Ascii("keyboaard"), &["keyboard"]),
            (
                dictgen::InsensitiveStr::Ascii("definiting"),
                &["defining", "definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resotrations"),
                &["restorations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponsdents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("msytical"), &["mystical"]),
            (
                dictgen::InsensitiveStr::Ascii("dispaching"),
                &["dispatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("scaepgoat"), &["scapegoat"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapoloate"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuiting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("begininer"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("exspense"), &["expense"]),
            (
                dictgen::InsensitiveStr::Ascii("loosy"),
                &["lossy", "lousy", "loose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructer"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("intitiate"), &["initiate"]),
            (dictgen::InsensitiveStr::Ascii("enabeld"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("manufaturer"),
                &["manufacturer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backeneds"),
                &["backends", "blackens"],
            ),
            (dictgen::InsensitiveStr::Ascii("distriute"), &["distribute"]),
            (
                dictgen::InsensitiveStr::Ascii("successesful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("sertainly"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("absailing"), &["abseiling"]),
            (dictgen::InsensitiveStr::Ascii("bacground"), &["background"]),
            (
                dictgen::InsensitiveStr::Ascii("reguarldess"),
                &["regardless"],
            ),
            (dictgen::InsensitiveStr::Ascii("aspiraton"), &["aspiration"]),
            (
                dictgen::InsensitiveStr::Ascii("referrencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotribute"), &["contribute"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacin"), &["pacing", "pain"]),
            (
                dictgen::InsensitiveStr::Ascii("insturmentation"),
                &["instrumentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decompreses"),
                &["decompresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("processees"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("repsository"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("clsoely"), &["closely"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrevated"),
                &["abbreviated"],
            ),
            (dictgen::InsensitiveStr::Ascii("comfused"), &["confused"]),
            (dictgen::InsensitiveStr::Ascii("avergaes"), &["averages"]),
            (
                dictgen::InsensitiveStr::Ascii("origining"),
                &["originating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynmaics"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("reviw"), &["review"]),
            (
                dictgen::InsensitiveStr::Ascii("slects"),
                &["selects", "elects", "sects"],
            ),
            (dictgen::InsensitiveStr::Ascii("accet"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("incrmenting"),
                &["incrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("overriden"), &["overridden"]),
            (dictgen::InsensitiveStr::Ascii("arraay"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("firendly"), &["friendly"]),
            (
                dictgen::InsensitiveStr::Ascii("pleasently"),
                &["pleasantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("propely"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("upsrteaming"),
                &["upstreaming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("folllowinwg"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genererater"),
                &["generator"],
            ),
            (dictgen::InsensitiveStr::Ascii("availale"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("splittin"), &["splitting"]),
            (dictgen::InsensitiveStr::Ascii("escalative"), &["escalate"]),
            (
                dictgen::InsensitiveStr::Ascii("alternativ"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("propagte"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("returen"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("hunman"), &["human"]),
            (dictgen::InsensitiveStr::Ascii("instruced"), &["instructed"]),
            (dictgen::InsensitiveStr::Ascii("studenst"), &["students"]),
            (
                dictgen::InsensitiveStr::Ascii("inteligability"),
                &["intelligibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabe"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimiliary"),
                &["dissimilarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parenteses"),
                &["parentheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("diective"), &["directive"]),
            (dictgen::InsensitiveStr::Ascii("doctirne"), &["doctrine"]),
            (
                dictgen::InsensitiveStr::Ascii("sluaghtered"),
                &["slaughtered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallunication"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("busienss"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("surroud"), &["surround"]),
            (
                dictgen::InsensitiveStr::Ascii("objectivify"),
                &["objectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ostricized"),
                &["ostracized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defintitively"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consctruction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifyer"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttacked"), &["attacked"]),
            (
                dictgen::InsensitiveStr::Ascii("uninterneted"),
                &["uninterested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anphatheaters"),
                &["amphitheaters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autononous"),
                &["autonomous"],
            ),
            (dictgen::InsensitiveStr::Ascii("astroid"), &["asteroid"]),
            (dictgen::InsensitiveStr::Ascii("herculeans"), &["hercules"]),
            (
                dictgen::InsensitiveStr::Ascii("factionnal"),
                &["fractional", "factional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resestatus"),
                &["resetstatus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interacive"),
                &["interactive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marhsmallows"),
                &["marshmallows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investingation"),
                &["investigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deteministically"),
                &["deterministically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construce"),
                &["construct", "construe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenificate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntructs"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("amernians"), &["armenians"]),
            (dictgen::InsensitiveStr::Ascii("waitinig"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("witdths"), &["widths"]),
            (
                dictgen::InsensitiveStr::Ascii("shortenning"),
                &["shortening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guantanomo"),
                &["guantanamo"],
            ),
            (dictgen::InsensitiveStr::Ascii("causioned"), &["cautioned"]),
            (dictgen::InsensitiveStr::Ascii("feasable"), &["feasible"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetimines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("senario"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("catgorized"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplets"), &["simplest"]),
            (
                dictgen::InsensitiveStr::Ascii("acceeptance"),
                &["acceptance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorespondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("unkonw"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("annoncing"), &["announcing"]),
            (
                dictgen::InsensitiveStr::Ascii("paralleized"),
                &["parallelized"],
            ),
            (dictgen::InsensitiveStr::Ascii("patientens"), &["patients"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualtion"),
                &["calculations", "calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fause"), &["false"]),
            (dictgen::InsensitiveStr::Ascii("addrass"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("collaboritavely"),
                &["collaboratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("amaizingly"), &["amazingly"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sympathyze"),
                &["sympathize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arranagement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanatsize"), &["fantasize"]),
            (dictgen::InsensitiveStr::Ascii("finsishes"), &["finishes"]),
            (dictgen::InsensitiveStr::Ascii("idetify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("caputure"), &["capture"]),
            (
                dictgen::InsensitiveStr::Ascii("imporovements"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("porbably"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("cartladge"), &["cartilage"]),
            (dictgen::InsensitiveStr::Ascii("scritpt"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("flzgs"), &["flags"]),
            (dictgen::InsensitiveStr::Ascii("reposonse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("unboun"), &["unbound"]),
            (
                dictgen::InsensitiveStr::Ascii("authentificate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahlpas"), &["alphas"]),
            (dictgen::InsensitiveStr::Ascii("mallicious"), &["malicious"]),
            (
                dictgen::InsensitiveStr::Ascii("temorarily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitaions"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reducin"), &["reducing"]),
            (dictgen::InsensitiveStr::Ascii("vetween"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("conservacion"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pracitcally"),
                &["practically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respitatory"),
                &["respiratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("faptastically"),
                &["fantastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("leyers"), &["layers"]),
            (dictgen::InsensitiveStr::Ascii("cofigs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("braodcasting"),
                &["broadcasting"],
            ),
            (dictgen::InsensitiveStr::Ascii("daquiris"), &["daiquiris"]),
            (dictgen::InsensitiveStr::Ascii("cerates"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("autherisation"),
                &["authorisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finctionally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("compleetd"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("tesitmony"), &["testimony"]),
            (
                dictgen::InsensitiveStr::Ascii("implementataion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensionaility"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnearbuility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("drver"), &["driver"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesable"), &["accessible"]),
            (
                dictgen::InsensitiveStr::Ascii("asserively"),
                &["assertively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrispond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("patrios"), &["patriots"]),
            (dictgen::InsensitiveStr::Ascii("gud"), &["good"]),
            (dictgen::InsensitiveStr::Ascii("hanbdles"), &["handles"]),
            (dictgen::InsensitiveStr::Ascii("consultat"), &["consultant"]),
            (dictgen::InsensitiveStr::Ascii("exposito"), &["exposition"]),
            (
                dictgen::InsensitiveStr::Ascii("condulences"),
                &["condolences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcotries"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("directirx"), &["directrix"]),
            (
                dictgen::InsensitiveStr::Ascii("controoler"),
                &["controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("shouldbe"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("comando"), &["commando"]),
            (dictgen::InsensitiveStr::Ascii("provicdes"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("permable"), &["permeable"]),
            (dictgen::InsensitiveStr::Ascii("beginin"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("referecing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("behabviour"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("deactivatin"),
                &["deactivating", "deactivation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docuemtned"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("experienc"), &["experience"]),
            (dictgen::InsensitiveStr::Ascii("opatque"), &["opaque"]),
            (
                dictgen::InsensitiveStr::Ascii("integratie"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hadnlers"), &["handlers"]),
            (
                dictgen::InsensitiveStr::Ascii("swtizerland"),
                &["switzerland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("granchildren"),
                &["grandchildren"],
            ),
            (dictgen::InsensitiveStr::Ascii("expeceting"), &["expecting"]),
            (dictgen::InsensitiveStr::Ascii("uptate"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("otherways"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("legislatie"),
                &["legislative"],
            ),
            (dictgen::InsensitiveStr::Ascii("covinced"), &["convinced"]),
            (
                dictgen::InsensitiveStr::Ascii("portuguise"),
                &["portuguese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virutalisation"),
                &["virtualisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertainy"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("conencted"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("furmula"), &["formula"]),
            (dictgen::InsensitiveStr::Ascii("trhusters"), &["thrusters"]),
            (
                dictgen::InsensitiveStr::Ascii("remenicent"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanatsy"), &["fantasy"]),
            (
                dictgen::InsensitiveStr::Ascii("cointerpoint"),
                &["counterpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomponied"),
                &["accompanied"],
            ),
            (dictgen::InsensitiveStr::Ascii("expolding"), &["exploding"]),
            (
                dictgen::InsensitiveStr::Ascii("prestegious"),
                &["prestigious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transprted"),
                &["transported"],
            ),
            (dictgen::InsensitiveStr::Ascii("copany"), &["company"]),
            (
                dictgen::InsensitiveStr::Ascii("transfromate"),
                &["transform", "transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propserity"),
                &["prosperity"],
            ),
            (dictgen::InsensitiveStr::Ascii("reverseed"), &["reversed"]),
            (
                dictgen::InsensitiveStr::Ascii("rememebered"),
                &["remembered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approproximated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltrater"),
                &["infiltrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("xepects"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("excect"),
                &["except", "expect"],
            ),
            (dictgen::InsensitiveStr::Ascii("respod"), &["respond"]),
            (dictgen::InsensitiveStr::Ascii("ediited"), &["edited"]),
            (
                dictgen::InsensitiveStr::Ascii("assocciations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postscritp"),
                &["postscript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outlowed"),
                &["outflowed", "outlawed"],
            ),
            (dictgen::InsensitiveStr::Ascii("piority"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("selecetions"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("artistas"), &["artists"]),
            (dictgen::InsensitiveStr::Ascii("exernal"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("disconetion"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expieriencing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsisency"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("sizez"), &["sizes", "sized"]),
            (
                dictgen::InsensitiveStr::Ascii("observasion"),
                &["observations"],
            ),
            (dictgen::InsensitiveStr::Ascii("generaml"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("eclipes"), &["eclipse"]),
            (dictgen::InsensitiveStr::Ascii("physiqe"), &["physique"]),
            (
                dictgen::InsensitiveStr::Ascii("cobinations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tansformation"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissovle"), &["dissolve"]),
            (dictgen::InsensitiveStr::Ascii("informarlo"), &["informal"]),
            (dictgen::InsensitiveStr::Ascii("feeze"), &["freeze"]),
            (
                dictgen::InsensitiveStr::Ascii("interesseted"),
                &["interested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassembles"),
                &["disassembles"],
            ),
            (dictgen::InsensitiveStr::Ascii("anthing"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("befor"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("distribtutors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southamton"),
                &["southampton"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrive"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("reflecto"), &["reflection"]),
            (dictgen::InsensitiveStr::Ascii("seqeuncer"), &["sequencer"]),
            (dictgen::InsensitiveStr::Ascii("effciency"), &["efficiency"]),
            (dictgen::InsensitiveStr::Ascii("souvineer"), &["souvenir"]),
            (dictgen::InsensitiveStr::Ascii("internest"), &["internets"]),
            (
                dictgen::InsensitiveStr::Ascii("replicaion"),
                &["replication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acumalating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cherckboxes"),
                &["checkboxes"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizziles"), &["chisels"]),
            (dictgen::InsensitiveStr::Ascii("futherize"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("harrassment"),
                &["harassment"],
            ),
            (dictgen::InsensitiveStr::Ascii("insetad"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("imaghe"), &["image"]),
            (dictgen::InsensitiveStr::Ascii("travaller"), &["traveller"]),
            (dictgen::InsensitiveStr::Ascii("sonetimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("expirie"), &["expire"]),
            (dictgen::InsensitiveStr::Ascii("mchu"), &["much"]),
            (
                dictgen::InsensitiveStr::Ascii("strucutures"),
                &["structures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexpectedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlighterd"),
                &["highlighted", "highlighter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturdating"),
                &["masturbating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execulatbles"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transaccion"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearates"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("uninterruptable"),
                &["uninterruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactaristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("desisions"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("colide"), &["collide"]),
            (dictgen::InsensitiveStr::Ascii("hamburgare"), &["hamburger"]),
            (
                dictgen::InsensitiveStr::Ascii("assumuptions"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electic"),
                &["electric", "eclectic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excist"),
                &["exist", "excise"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("generting"), &["generating"]),
            (
                dictgen::InsensitiveStr::Ascii("thiank"),
                &["thank", "think"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intergated"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("zoombies"), &["zombies"]),
            (dictgen::InsensitiveStr::Ascii("exeuctors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("waitting"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("expanion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("sucessed"), &["succeeded"]),
            (
                dictgen::InsensitiveStr::Ascii("assingments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("enalbes"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsultions"),
                &["encapsulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("barrells"), &["barrels"]),
            (
                dictgen::InsensitiveStr::Ascii("confgurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolte"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("segmentaion"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegitation"),
                &["vegetation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undertacker"),
                &["undertaker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitativo"),
                &["limitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suprassing"),
                &["surpassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("thannking"), &["thanking"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimizers"),
                &["optimizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("heigh"), &["height", "high"]),
            (dictgen::InsensitiveStr::Ascii("glaobal"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("baising"),
                &["raising", "biasing", "braising", "bailing", "baiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("mazilla"), &["mozilla"]),
            (
                dictgen::InsensitiveStr::Ascii("negothiates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("concerete"), &["concrete"]),
            (dictgen::InsensitiveStr::Ascii("mintues"), &["minutes"]),
            (dictgen::InsensitiveStr::Ascii("carbus"), &["cardbus"]),
            (
                dictgen::InsensitiveStr::Ascii("penninsulas"),
                &["peninsulas"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unncessary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amerliorated"),
                &["ameliorated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bumber"),
                &["bomber", "bummer", "bumper", "number"],
            ),
            (dictgen::InsensitiveStr::Ascii("testicel"), &["testicle"]),
            (dictgen::InsensitiveStr::Ascii("algorhythm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("prfix"), &["prefix"]),
            (
                dictgen::InsensitiveStr::Ascii("desoriented"),
                &["disoriented"],
            ),
            (dictgen::InsensitiveStr::Ascii("eception"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("neratively"),
                &["narratively", "negatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliased"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("uniqeness"), &["uniqueness"]),
            (
                dictgen::InsensitiveStr::Ascii("thunderboldt"),
                &["thunderbolt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professinally"),
                &["professionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ackowledging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("slections"),
                &["selections", "sections", "elections"],
            ),
            (dictgen::InsensitiveStr::Ascii("restanti"), &["restarting"]),
            (dictgen::InsensitiveStr::Ascii("visheators"), &["vitiators"]),
            (dictgen::InsensitiveStr::Ascii("ediits"), &["edits"]),
            (
                dictgen::InsensitiveStr::Ascii("exlanation"),
                &["explanation"],
            ),
            (dictgen::InsensitiveStr::Ascii("clinicas"), &["clinics"]),
            (
                dictgen::InsensitiveStr::Ascii("provincija"),
                &["provincial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emumerator"),
                &["enumerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("billionairre"),
                &["billionaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("visiblity"), &["visibility"]),
            (dictgen::InsensitiveStr::Ascii("tution"), &["tuition"]),
            (dictgen::InsensitiveStr::Ascii("handeldy"), &["handedly"]),
            (
                dictgen::InsensitiveStr::Ascii("funchtioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("agggregate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("opposit"), &["opposite"]),
            (
                dictgen::InsensitiveStr::Ascii("readabillity"),
                &["readability"],
            ),
            (dictgen::InsensitiveStr::Ascii("categeory"), &["category"]),
            (
                dictgen::InsensitiveStr::Ascii("partions"),
                &["partitions", "portions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incovenient"),
                &["inconvenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("itme"), &["item", "time"]),
            (dictgen::InsensitiveStr::Ascii("knowledg"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("constainer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("obscuirty"), &["obscurity"]),
            (
                dictgen::InsensitiveStr::Ascii("phisosophy"),
                &["philosophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclocing"),
                &["overclocking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandizes"),
                &["aggrandizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenificators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("oterwize"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("conbination"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectionas"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorne"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("arhives"), &["archives"]),
            (dictgen::InsensitiveStr::Ascii("goint"), &["going"]),
            (dictgen::InsensitiveStr::Ascii("faulure"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("leciester"), &["leicester"]),
            (
                dictgen::InsensitiveStr::Ascii("certiticated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("numnbered"), &["numbered"]),
            (dictgen::InsensitiveStr::Ascii("apperent"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("qulified"), &["qualified"]),
            (dictgen::InsensitiveStr::Ascii("wokrers"), &["workers"]),
            (dictgen::InsensitiveStr::Ascii("becomin"), &["becoming"]),
            (
                dictgen::InsensitiveStr::Ascii("accceptance"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("matchig"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("depdencentes"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssessing"), &["assessing"]),
            (dictgen::InsensitiveStr::Ascii("positioin"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("godpseed"), &["godspeed"]),
            (
                dictgen::InsensitiveStr::Ascii("donwloadable"),
                &["downloadable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individial"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("sueprset"), &["superset"]),
            (
                dictgen::InsensitiveStr::Ascii("subexperession"),
                &["subexpression"],
            ),
            (dictgen::InsensitiveStr::Ascii("spsace"), &["space"]),
            (dictgen::InsensitiveStr::Ascii("fandation"), &["foundation"]),
            (
                dictgen::InsensitiveStr::Ascii("understadably"),
                &["understandably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cocatenated"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactue"),
                &["interacted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrective"),
                &["restrictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("funiture"), &["furniture"]),
            (
                dictgen::InsensitiveStr::Ascii("professorin"),
                &["profession"],
            ),
            (dictgen::InsensitiveStr::Ascii("seriosly"), &["seriously"]),
            (dictgen::InsensitiveStr::Ascii("executute"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("equilibruim"),
                &["equilibrium"],
            ),
            (dictgen::InsensitiveStr::Ascii("collased"), &["collapsed"]),
            (
                dictgen::InsensitiveStr::Ascii("asertively"),
                &["assertively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimate"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcondes"),
                &["transcodes"],
            ),
            (dictgen::InsensitiveStr::Ascii("plethroa"), &["plethora"]),
            (
                dictgen::InsensitiveStr::Ascii("lifestlyes"),
                &["lifestyles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anononymous"),
                &["anonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anthropoly"),
                &["anthropology"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilmbers"), &["climbers"]),
            (dictgen::InsensitiveStr::Ascii("markedet"), &["marketed"]),
            (
                dictgen::InsensitiveStr::Ascii("parallellizes"),
                &["parallelizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("inividual"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("srcolled"), &["scrolled"]),
            (dictgen::InsensitiveStr::Ascii("expectus"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("organsie"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("linewdith"), &["linewidth"]),
            (dictgen::InsensitiveStr::Ascii("sistance"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("unanymous"), &["unanimous"]),
            (dictgen::InsensitiveStr::Ascii("definined"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("premanent"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("lighlty"), &["lightly"]),
            (
                dictgen::InsensitiveStr::Ascii("admininstrator"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("copiling"), &["compiling"]),
            (dictgen::InsensitiveStr::Ascii("imorts"), &["imports"]),
            (
                dictgen::InsensitiveStr::Ascii("coditionally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victorinos"),
                &["victorious"],
            ),
            (dictgen::InsensitiveStr::Ascii("reudction"), &["reduction"]),
            (dictgen::InsensitiveStr::Ascii("dungeoness"), &["dungeons"]),
            (dictgen::InsensitiveStr::Ascii("inhertis"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("fromerly"), &["formerly"]),
            (dictgen::InsensitiveStr::Ascii("presearves"), &["preserves"]),
            (
                dictgen::InsensitiveStr::Ascii("guarrented"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("chlidrens"), &["childrens"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomnace"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modernination"),
                &["modernization"],
            ),
            (dictgen::InsensitiveStr::Ascii("unplesent"), &["unpleasant"]),
            (dictgen::InsensitiveStr::Ascii("finanace"), &["finance"]),
            (dictgen::InsensitiveStr::Ascii("wheren"), &["where"]),
            (dictgen::InsensitiveStr::Ascii("remoived"), &["removed"]),
            (
                dictgen::InsensitiveStr::Ascii("distinguise"),
                &["distinguished", "distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassinare"),
                &["assassinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frams"),
                &["frames", "grams", "drams", "crams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irresponsibile"),
                &["irresponsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapprouving"),
                &["disapproving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juggurnaut"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purchaseing"),
                &["purchasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("onlt"), &["only"]),
            (dictgen::InsensitiveStr::Ascii("markter"), &["marker"]),
            (dictgen::InsensitiveStr::Ascii("mutilpath"), &["multipath"]),
            (dictgen::InsensitiveStr::Ascii("cafetiera"), &["cafeteria"]),
            (dictgen::InsensitiveStr::Ascii("formuals"), &["formulas"]),
            (
                dictgen::InsensitiveStr::Ascii("suffiecient"),
                &["sufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("laucnhed"), &["launched"]),
            (
                dictgen::InsensitiveStr::Ascii("instroduces"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installationa"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambuguity"), &["ambiguity"]),
            (dictgen::InsensitiveStr::Ascii("nessesary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("customisin"),
                &["customising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentat"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caricaturile"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaizers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("connecs"), &["connects"]),
            (
                dictgen::InsensitiveStr::Ascii("periwinkel"),
                &["periwinkle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvished"),
                &["improvised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countires"),
                &["countries", "counties"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggragate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("recjects"), &["rejects"]),
            (dictgen::InsensitiveStr::Ascii("withhout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("consonent"), &["consonant"]),
            (
                dictgen::InsensitiveStr::Ascii("protability"),
                &["portability", "probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscirptions"),
                &["unsubscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("creatred"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("mutbal"), &["mutable"]),
            (
                dictgen::InsensitiveStr::Ascii("gauranteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("trollade"), &["trolled"]),
            (
                dictgen::InsensitiveStr::Ascii("reducter"),
                &["reducer", "redacter"],
            ),
            (dictgen::InsensitiveStr::Ascii("receiv"), &["receive"]),
            (dictgen::InsensitiveStr::Ascii("inbeetwen"), &["inbetween"]),
            (
                dictgen::InsensitiveStr::Ascii("subtsitutable"),
                &["substitutable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginizer"),
                &["recognizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("regisetr"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("autocmpletes"),
                &["autocompletes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfored"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("confiugre"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("initializaers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enrollemnt"),
                &["enrollment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("menberships"),
                &["memberships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmenation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialzed"),
                &["deserialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferrable"),
                &["preferable"],
            ),
            (dictgen::InsensitiveStr::Ascii("univiersal"), &["universal"]),
            (
                dictgen::InsensitiveStr::Ascii("inpredictable"),
                &["unpredictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("incremet"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("transmitt"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("inforamtion"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("mitaken"), &["mistaken"]),
            (dictgen::InsensitiveStr::Ascii("assmbler"), &["assembler"]),
            (dictgen::InsensitiveStr::Ascii("eorrs"), &["errors", "errs"]),
            (dictgen::InsensitiveStr::Ascii("activty"), &["activity"]),
            (
                dictgen::InsensitiveStr::Ascii("catastronphic"),
                &["catastrophic"],
            ),
            (dictgen::InsensitiveStr::Ascii("crticised"), &["criticised"]),
            (dictgen::InsensitiveStr::Ascii("acocunts"), &["accounts"]),
            (dictgen::InsensitiveStr::Ascii("udateed"), &["updated"]),
            (
                dictgen::InsensitiveStr::Ascii("collapted"),
                &["coapted", "collapse", "collapsed", "collated"],
            ),
            (dictgen::InsensitiveStr::Ascii("deisgner"), &["designer"]),
            (dictgen::InsensitiveStr::Ascii("forcaster"), &["forecaster"]),
            (dictgen::InsensitiveStr::Ascii("progrmae"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("searcahble"),
                &["searchable"],
            ),
            (dictgen::InsensitiveStr::Ascii("impelemnt"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("paragaraphs"),
                &["paragraphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idendifying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("attahced"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("unknowningly"),
                &["unknowingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compositin"),
                &["compositing", "composition"],
            ),
            (dictgen::InsensitiveStr::Ascii("cretating"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("regessed"),
                &["regressed", "recessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("mudule"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplaicate"),
                &["deduplicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lonly"), &["lonely", "only"]),
            (
                dictgen::InsensitiveStr::Ascii("acumulative"),
                &["accumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("naybourly"),
                &["neighbourly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ceckpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplyfications"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("macthups"), &["matchups"]),
            (dictgen::InsensitiveStr::Ascii("nax"), &["max", "nad"]),
            (
                dictgen::InsensitiveStr::Ascii("cronologically"),
                &["chronologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doen"),
                &["done", "don", "doesn"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persistantly"),
                &["persistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misforture"),
                &["misfortune"],
            ),
            (dictgen::InsensitiveStr::Ascii("optonally"), &["optionally"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsrapped"),
                &["bootstrapped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumised"),
                &["customised"],
            ),
            (dictgen::InsensitiveStr::Ascii("templaces"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("escluding"), &["excluding"]),
            (dictgen::InsensitiveStr::Ascii("confids"), &["confides"]),
            (dictgen::InsensitiveStr::Ascii("journalits"), &["journals"]),
            (
                dictgen::InsensitiveStr::Ascii("veitnamese"),
                &["vietnamese"],
            ),
            (dictgen::InsensitiveStr::Ascii("soemtimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("variatoin"), &["variation"]),
            (dictgen::InsensitiveStr::Ascii("anlysed"), &["analysed"]),
            (dictgen::InsensitiveStr::Ascii("commencera"), &["commerce"]),
            (dictgen::InsensitiveStr::Ascii("enthicity"), &["ethnicity"]),
            (dictgen::InsensitiveStr::Ascii("aysnc"), &["async"]),
            (
                dictgen::InsensitiveStr::Ascii("sarting"),
                &["starting", "sorting"],
            ),
            (dictgen::InsensitiveStr::Ascii("artillerly"), &["artillery"]),
            (dictgen::InsensitiveStr::Ascii("orpan"), &["orphan"]),
            (dictgen::InsensitiveStr::Ascii("clampping"), &["clamping"]),
            (dictgen::InsensitiveStr::Ascii("ovrlap"), &["overlap"]),
            (
                dictgen::InsensitiveStr::Ascii("openbrower"),
                &["openbrowser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitimisations"),
                &["optimisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tourmanent"),
                &["tournaments"],
            ),
            (dictgen::InsensitiveStr::Ascii("instabce"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("multlipliers"),
                &["multipliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefedined"),
                &["predefined"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwwers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("decordes"), &["decodes"]),
            (dictgen::InsensitiveStr::Ascii("beginns"), &["begins"]),
            (
                dictgen::InsensitiveStr::Ascii("medicore"),
                &["mediocre", "medicare"],
            ),
            (dictgen::InsensitiveStr::Ascii("roration"), &["rotation"]),
            (
                dictgen::InsensitiveStr::Ascii("phlematous"),
                &["phlegmatous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experamented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompleteness"),
                &["incompleteness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generalyse"),
                &["generalise"],
            ),
            (dictgen::InsensitiveStr::Ascii("definito"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("tranzistor"),
                &["transistor"],
            ),
            (dictgen::InsensitiveStr::Ascii("reniging"), &["reneging"]),
            (dictgen::InsensitiveStr::Ascii("delagates"), &["delegates"]),
            (
                dictgen::InsensitiveStr::Ascii("pasturising"),
                &["pasteurising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reigstration"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speicfication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rienforcement"),
                &["reinforcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("resistered"), &["resisted"]),
            (
                dictgen::InsensitiveStr::Ascii("exercizers"),
                &["exercisers", "exorcists"],
            ),
            (dictgen::InsensitiveStr::Ascii("stach"), &["stack"]),
            (dictgen::InsensitiveStr::Ascii("poistion"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("combanation"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("councidentally"),
                &["coincidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeltes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("revoluitonary"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implausibe"),
                &["implausible"],
            ),
            (dictgen::InsensitiveStr::Ascii("constatn"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("turtorials"), &["tutorials"]),
            (
                dictgen::InsensitiveStr::Ascii("defaoulting"),
                &["defaulting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afterthougth"),
                &["afterthought"],
            ),
            (dictgen::InsensitiveStr::Ascii("apointing"), &["appointing"]),
            (
                dictgen::InsensitiveStr::Ascii("terminatior"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exspectedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persentiles"),
                &["percentiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrucured"),
                &["restructured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funvtionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unliaterally"),
                &["unilaterally"],
            ),
            (dictgen::InsensitiveStr::Ascii("expepects"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("solitudine"), &["solitude"]),
            (dictgen::InsensitiveStr::Ascii("cahcs"), &["caches"]),
            (
                dictgen::InsensitiveStr::Ascii("trewthfully"),
                &["truthfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("invokves"), &["invokes"]),
            (
                dictgen::InsensitiveStr::Ascii("intorduces"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreamedd"),
                &["upstreamed"],
            ),
            (dictgen::InsensitiveStr::Ascii("owerwrite"), &["overwrite"]),
            (
                dictgen::InsensitiveStr::Ascii("configuartions"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advatanges"),
                &["advantages"],
            ),
            (dictgen::InsensitiveStr::Ascii("assymptote"), &["asymptote"]),
            (
                dictgen::InsensitiveStr::Ascii("accomadated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("directrive"), &["directive"]),
            (
                dictgen::InsensitiveStr::Ascii("anniversery"),
                &["anniversary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalubale"),
                &["invaluable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vertifications"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoudt"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("scandinavica"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("inquisitr"), &["inquisitor"]),
            (dictgen::InsensitiveStr::Ascii("appplies"), &["applies"]),
            (
                dictgen::InsensitiveStr::Ascii("possiability"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("writig"), &["writing"]),
            (dictgen::InsensitiveStr::Ascii("undergard"), &["undergrad"]),
            (dictgen::InsensitiveStr::Ascii("eemove"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("succesful"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("aribiter"), &["arbiter"]),
            (dictgen::InsensitiveStr::Ascii("geeric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("dicusses"), &["discusses"]),
            (dictgen::InsensitiveStr::Ascii("therafter"), &["thereafter"]),
            (dictgen::InsensitiveStr::Ascii("whisperes"), &["whispers"]),
            (
                dictgen::InsensitiveStr::Ascii("definitve"),
                &["definite", "definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("defering"), &["deferring"]),
            (
                dictgen::InsensitiveStr::Ascii("antibiotcs"),
                &["antibiotics"],
            ),
            (dictgen::InsensitiveStr::Ascii("meterial"), &["material"]),
            (dictgen::InsensitiveStr::Ascii("critqued"), &["critiqued"]),
            (dictgen::InsensitiveStr::Ascii("simetrie"), &["symmetry"]),
            (dictgen::InsensitiveStr::Ascii("gnerate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("spesifical"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("serebral"), &["cerebral"]),
            (dictgen::InsensitiveStr::Ascii("togled"), &["toggled"]),
            (dictgen::InsensitiveStr::Ascii("persented"), &["presented"]),
            (dictgen::InsensitiveStr::Ascii("booleamn"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("fantazise"), &["fantasise"]),
            (
                dictgen::InsensitiveStr::Ascii("quiestioned"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("compex"), &["complex"]),
            (dictgen::InsensitiveStr::Ascii("fialogues"), &["dialogues"]),
            (
                dictgen::InsensitiveStr::Ascii("authethenticating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifaced"),
                &["sacrificed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challeners"),
                &["challengers"],
            ),
            (dictgen::InsensitiveStr::Ascii("regualator"), &["regulator"]),
            (dictgen::InsensitiveStr::Ascii("cteates"), &["creates"]),
            (dictgen::InsensitiveStr::Ascii("clas"), &["class"]),
            (dictgen::InsensitiveStr::Ascii("exansion"), &["expansion"]),
            (
                dictgen::InsensitiveStr::Ascii("desintegration"),
                &["disintegration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chesckboxes"),
                &["checkboxes"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeuired"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("degarded"), &["degraded"]),
            (dictgen::InsensitiveStr::Ascii("councelor"), &["counselor"]),
            (
                dictgen::InsensitiveStr::Ascii("underwares"),
                &["underwater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotiote"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cataclsym"), &["cataclysm"]),
            (
                dictgen::InsensitiveStr::Ascii("congragation"),
                &["congregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chameleooon"),
                &["chameleon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistant"),
                &["consistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversimplificaton"),
                &["oversimplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbalnaced"),
                &["unbalanced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adverserial"),
                &["adversarial"],
            ),
            (dictgen::InsensitiveStr::Ascii("selet"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivites"),
                &["sensitivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsildes"),
                &["reconciles"],
            ),
            (dictgen::InsensitiveStr::Ascii("epxanding"), &["expanding"]),
            (dictgen::InsensitiveStr::Ascii("momementum"), &["momentum"]),
            (
                dictgen::InsensitiveStr::Ascii("unspecifiec"),
                &["unspecific"],
            ),
            (dictgen::InsensitiveStr::Ascii("tupe"), &["tuple", "type"]),
            (dictgen::InsensitiveStr::Ascii("proider"), &["provider"]),
            (
                dictgen::InsensitiveStr::Ascii("agina"),
                &["again", "angina"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonised"),
                &["synchronised"],
            ),
            (dictgen::InsensitiveStr::Ascii("delegatie"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("thursdey"), &["thursday"]),
            (
                dictgen::InsensitiveStr::Ascii("duplicting"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dogdammit"), &["goddammit"]),
            (dictgen::InsensitiveStr::Ascii("folwoing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("enumarating"),
                &["enumerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mantions"),
                &["mentions", "mansions"],
            ),
            (dictgen::InsensitiveStr::Ascii("finnishes"), &["finishes"]),
            (
                dictgen::InsensitiveStr::Ascii("compativilities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructer"),
                &["instructor", "instructed", "instruct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experation"),
                &["expiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philadelpia"),
                &["philadelphia"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshale"), &["handshake"]),
            (dictgen::InsensitiveStr::Ascii("uages"), &["usages"]),
            (dictgen::InsensitiveStr::Ascii("proffesed"), &["professed"]),
            (dictgen::InsensitiveStr::Ascii("jurnaler"), &["journaler"]),
            (dictgen::InsensitiveStr::Ascii("cancles"), &["cancels"]),
            (dictgen::InsensitiveStr::Ascii("categoize"), &["categorize"]),
            (
                dictgen::InsensitiveStr::Ascii("dependes"),
                &["dependees", "depends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insturction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("firgure"), &["figure"]),
            (dictgen::InsensitiveStr::Ascii("assembley"), &["assembly"]),
            (dictgen::InsensitiveStr::Ascii("valkyire"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("relativety"),
                &["relativity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visibililty"),
                &["visibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilites"),
                &["stabilizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simeltaneously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcioned"), &["functioned"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotate"),
                &["negotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitrues"), &["virtues"]),
            (dictgen::InsensitiveStr::Ascii("mailformed"), &["malformed"]),
            (
                dictgen::InsensitiveStr::Ascii("falled"),
                &["fell", "failed", "fallen", "felled"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciept"), &["receipt"]),
            (
                dictgen::InsensitiveStr::Ascii("rollercoater"),
                &["rollercoaster"],
            ),
            (dictgen::InsensitiveStr::Ascii("shifteer"), &["shifter"]),
            (dictgen::InsensitiveStr::Ascii("conflcit"), &["conflict"]),
            (
                dictgen::InsensitiveStr::Ascii("consctructor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("deplyed"), &["deployed"]),
            (dictgen::InsensitiveStr::Ascii("upperace"), &["uppercase"]),
            (dictgen::InsensitiveStr::Ascii("indexig"), &["indexing"]),
            (dictgen::InsensitiveStr::Ascii("emphatize"), &["empathize"]),
            (dictgen::InsensitiveStr::Ascii("bracets"), &["brackets"]),
            (dictgen::InsensitiveStr::Ascii("cimmanded"), &["commanded"]),
            (
                dictgen::InsensitiveStr::Ascii("refectored"),
                &["refactored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sympathatic"),
                &["sympathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("epectations"),
                &["expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jounralistic"),
                &["journalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chiper"),
                &["cipher", "chipper", "chimer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unneccecarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interription"),
                &["interruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workarrounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiniatorial"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("langauge"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("ggogle"),
                &["goggle", "google"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitabiliy"),
                &["profitability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regarging"),
                &["regarding", "recharging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propietries"),
                &["proprietaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("asscoaied"), &["associated"]),
            (
                dictgen::InsensitiveStr::Ascii("reportidly"),
                &["reportedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambigiuous"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("sectionning"),
                &["sectioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conicidental"),
                &["coincidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependncies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("comannds"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("porgressed"),
                &["progressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassociated"),
                &["disassociated", "dissociated"],
            ),
            (dictgen::InsensitiveStr::Ascii("satarted"), &["started"]),
            (
                dictgen::InsensitiveStr::Ascii("copilates"),
                &["compiles", "copulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wheater"),
                &["weather", "whether"],
            ),
            (dictgen::InsensitiveStr::Ascii("defitnely"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("questoinnaires"),
                &["questionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("timestans"), &["timespans"]),
            (dictgen::InsensitiveStr::Ascii("caculated"), &["calculated"]),
            (
                dictgen::InsensitiveStr::Ascii("conversacional"),
                &["conversational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compitation"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("execucte"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("subscibed"), &["subscribed"]),
            (
                dictgen::InsensitiveStr::Ascii("promisculous"),
                &["promiscuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("konservatism"),
                &["conservatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("technik"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("corellations"),
                &["correlations"],
            ),
            (dictgen::InsensitiveStr::Ascii("compliers"), &["compilers"]),
            (dictgen::InsensitiveStr::Ascii("ostridge"), &["ostrich"]),
            (dictgen::InsensitiveStr::Ascii("streoid"), &["steroid"]),
            (dictgen::InsensitiveStr::Ascii("spezific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("sophisitcated"),
                &["sophisticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microcentres"),
                &["microcenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("buad"), &["baud"]),
            (
                dictgen::InsensitiveStr::Ascii("intelligantly"),
                &["intelligently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unstability"),
                &["instability"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectopm"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("opertae"), &["operate"]),
            (
                dictgen::InsensitiveStr::Ascii("specialied"),
                &["specialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philadelpha"),
                &["philadelphia"],
            ),
            (dictgen::InsensitiveStr::Ascii("existy"), &["exist"]),
            (dictgen::InsensitiveStr::Ascii("folloing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("hader"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("agitatin"),
                &["agitating", "agitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rivarlies"), &["rivalries"]),
            (dictgen::InsensitiveStr::Ascii("colission"), &["collision"]),
            (
                dictgen::InsensitiveStr::Ascii("baised"),
                &["raised", "biased", "braised", "bailed", "baited"],
            ),
            (dictgen::InsensitiveStr::Ascii("teminator"), &["terminator"]),
            (
                dictgen::InsensitiveStr::Ascii("aggegrated"),
                &["aggregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("banglidesh"),
                &["bangladesh"],
            ),
            (dictgen::InsensitiveStr::Ascii("disalbes"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoziated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facillitate"),
                &["facilitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sicinct"), &["succinct"]),
            (
                dictgen::InsensitiveStr::Ascii("incarantion"),
                &["incarnation"],
            ),
            (dictgen::InsensitiveStr::Ascii("switcheasy"), &["switches"]),
            (
                dictgen::InsensitiveStr::Ascii("succesfull"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("adventage"), &["advantage"]),
            (dictgen::InsensitiveStr::Ascii("stranglove"), &["strangle"]),
            (dictgen::InsensitiveStr::Ascii("remvoe"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("schoo"), &["school"]),
            (
                dictgen::InsensitiveStr::Ascii("paragrpahs"),
                &["paragraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("champane"), &["champagne"]),
            (dictgen::InsensitiveStr::Ascii("alinment"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("histgrams"), &["histograms"]),
            (
                dictgen::InsensitiveStr::Ascii("decommissionn"),
                &["decommission"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomate"), &["automate"]),
            (dictgen::InsensitiveStr::Ascii("unsetset"), &["unset"]),
            (dictgen::InsensitiveStr::Ascii("occuption"), &["occupation"]),
            (
                dictgen::InsensitiveStr::Ascii("drastical"),
                &["drastically", "drastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acses"),
                &["cases", "access"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissisitic"),
                &["narcissistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheeving"), &["achieving"]),
            (
                dictgen::InsensitiveStr::Ascii("corospondances"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiantence"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirector"), &["director"]),
            (dictgen::InsensitiveStr::Ascii("encruptor"), &["encryptor"]),
            (dictgen::InsensitiveStr::Ascii("nawshea"), &["nausea"]),
            (dictgen::InsensitiveStr::Ascii("traner"), &["trainer"]),
            (
                dictgen::InsensitiveStr::Ascii("authentificator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alterarion"),
                &["alteration"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatorss"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("usuer"), &["user"]),
            (dictgen::InsensitiveStr::Ascii("srollbars"), &["scrollbars"]),
            (dictgen::InsensitiveStr::Ascii("anythig"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("movemment"), &["movement"]),
            (dictgen::InsensitiveStr::Ascii("virtical"), &["vertical"]),
            (dictgen::InsensitiveStr::Ascii("glichy"), &["glitchy"]),
            (dictgen::InsensitiveStr::Ascii("textuer"), &["texture"]),
            (
                dictgen::InsensitiveStr::Ascii("behavoiural"),
                &["behavioural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capcaities"),
                &["capacities"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrupto"), &["corruption"]),
            (
                dictgen::InsensitiveStr::Ascii("contatenate"),
                &["concatenate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renforcement"),
                &["reinforcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("varilues"),
                &["values", "variables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allcoateing"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hiddin"),
                &["hidden", "hiding"],
            ),
            (dictgen::InsensitiveStr::Ascii("probablybe"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("clapsing"), &["collapsing"]),
            (
                dictgen::InsensitiveStr::Ascii("essentaily"),
                &["essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("multibye"), &["multibyte"]),
            (dictgen::InsensitiveStr::Ascii("esacped"), &["escaped"]),
            (
                dictgen::InsensitiveStr::Ascii("computtaion"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("prarameter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("buckes"),
                &["buckets", "bucks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specailize"),
                &["specializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("postiion"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("suspiciouly"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoporated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reseved"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("seee"), &["see"]),
            (
                dictgen::InsensitiveStr::Ascii("accrediation"),
                &["accreditation"],
            ),
            (dictgen::InsensitiveStr::Ascii("progidy"), &["prodigy"]),
            (
                dictgen::InsensitiveStr::Ascii("structering"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mamber"),
                &["member", "amber"],
            ),
            (dictgen::InsensitiveStr::Ascii("registrart"), &["registrar"]),
            (
                dictgen::InsensitiveStr::Ascii("summarizin"),
                &["summarizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresonding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associdated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("secundary"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("transfomation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interleavin"),
                &["interleaving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utiliatrian"),
                &["utilitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expotition"),
                &["exposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fraternety"),
                &["fraternity"],
            ),
            (dictgen::InsensitiveStr::Ascii("triguered"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("astronautlis"),
                &["astronauts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalistisk"),
                &["journalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("invaid"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("enginineer"), &["engineer"]),
            (
                dictgen::InsensitiveStr::Ascii("expereience"),
                &["experience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinquished"),
                &["distinguished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentciating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("commti"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("creaator"), &["creator"]),
            (dictgen::InsensitiveStr::Ascii("inluence"), &["influence"]),
            (dictgen::InsensitiveStr::Ascii("statestic"), &["statistic"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonatedly"),
                &["opinionatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("healtiest"),
                &["healthiest", "heartiest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprixamating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispathed"), &["dispatched"]),
            (
                dictgen::InsensitiveStr::Ascii("temprament"),
                &["temperament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extravegant"),
                &["extravagant"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulls"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("echanisms"), &["mechanisms"]),
            (
                dictgen::InsensitiveStr::Ascii("resgistering"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manfacturer"),
                &["manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("itsems"), &["items"]),
            (
                dictgen::InsensitiveStr::Ascii("conculsion"),
                &["conclusion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestions"),
                &["manifestations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paravirutalize"),
                &["paravirtualize"],
            ),
            (dictgen::InsensitiveStr::Ascii("inhereted"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaising"),
                &["generalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqcuires"), &["acquires"]),
            (dictgen::InsensitiveStr::Ascii("herclues"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("offcials"), &["officials"]),
            (dictgen::InsensitiveStr::Ascii("visheation"), &["vitiation"]),
            (dictgen::InsensitiveStr::Ascii("usally"), &["usually"]),
            (dictgen::InsensitiveStr::Ascii("timneout"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("distnguish"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoritiers"),
                &["authorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("breweres"), &["brewers"]),
            (
                dictgen::InsensitiveStr::Ascii("regrad"),
                &["regard", "regrade"],
            ),
            (dictgen::InsensitiveStr::Ascii("clener"), &["cleaner"]),
            (dictgen::InsensitiveStr::Ascii("maximuum"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("suppost"),
                &["supports", "support", "suppose", "supposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("arguemet"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("legitematly"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devastatin"),
                &["devastating", "devastation"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsidiy"), &["subsidy"]),
            (dictgen::InsensitiveStr::Ascii("normol"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("warnig"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoom"),
                &["perfume", "perform"],
            ),
            (dictgen::InsensitiveStr::Ascii("coulour"), &["colour"]),
            (
                dictgen::InsensitiveStr::Ascii("cannote"),
                &["cannot", "connote"],
            ),
            (dictgen::InsensitiveStr::Ascii("declred"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("ackowledges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiors"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninteristing"),
                &["uninteresting"],
            ),
            (dictgen::InsensitiveStr::Ascii("provincies"), &["province"]),
            (dictgen::InsensitiveStr::Ascii("excange"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("bethesa"), &["bethesda"]),
            (dictgen::InsensitiveStr::Ascii("suffic"), &["suffix"]),
            (
                dictgen::InsensitiveStr::Ascii("intemediate"),
                &["intermediate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("garranteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("genration"), &["generation"]),
            (dictgen::InsensitiveStr::Ascii("fiedlity"), &["fidelity"]),
            (dictgen::InsensitiveStr::Ascii("hybridus"), &["hybrids"]),
            (dictgen::InsensitiveStr::Ascii("dissmisses"), &["dismisses"]),
            (dictgen::InsensitiveStr::Ascii("tresures"), &["treasures"]),
            (dictgen::InsensitiveStr::Ascii("guarunty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("neighors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("hourgalss"), &["hourglass"]),
            (dictgen::InsensitiveStr::Ascii("clanups"), &["cleanups"]),
            (dictgen::InsensitiveStr::Ascii("checkkers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("bridaging"), &["brigading"]),
            (dictgen::InsensitiveStr::Ascii("insterters"), &["inserters"]),
            (dictgen::InsensitiveStr::Ascii("siutable"), &["suitable"]),
            (
                dictgen::InsensitiveStr::Ascii("experimed"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("featurd"), &["featured"]),
            (
                dictgen::InsensitiveStr::Ascii("intialistion"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occuptions"),
                &["occupations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequenzes"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("trilogoy"), &["trilogy"]),
            (dictgen::InsensitiveStr::Ascii("proocedure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("gentailia"), &["genitalia"]),
            (
                dictgen::InsensitiveStr::Ascii("miliciously"),
                &["maliciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("gemometric"), &["geometric"]),
            (dictgen::InsensitiveStr::Ascii("cofidence"), &["confidence"]),
            (
                dictgen::InsensitiveStr::Ascii("omnishients"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indisputibly"),
                &["indisputably"],
            ),
            (dictgen::InsensitiveStr::Ascii("calandar"), &["calendar"]),
            (
                dictgen::InsensitiveStr::Ascii("differntly"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressentations"),
                &["presentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorindates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("privelege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("unballance"), &["unbalance"]),
            (dictgen::InsensitiveStr::Ascii("checout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("automaticlly"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sociologia"),
                &["sociological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accessibilty"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermienting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("evesdrops"), &["eavesdrops"]),
            (
                dictgen::InsensitiveStr::Ascii("barycentic"),
                &["barycentric"],
            ),
            (dictgen::InsensitiveStr::Ascii("spartians"), &["spartans"]),
            (
                dictgen::InsensitiveStr::Ascii("indonasian"),
                &["indonesian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficiied"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ukrainiens"),
                &["ukrainians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destionation"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("anconda"), &["anaconda"]),
            (
                dictgen::InsensitiveStr::Ascii("transfroms"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("strenth"), &["strength"]),
            (dictgen::InsensitiveStr::Ascii("eqaully"), &["equally"]),
            (dictgen::InsensitiveStr::Ascii("fowlling"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("flie"), &["file"]),
            (
                dictgen::InsensitiveStr::Ascii("fpr"),
                &["for", "far", "fps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precedencs"),
                &["precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomrating"), &["formatting"]),
            (dictgen::InsensitiveStr::Ascii("pipsline"), &["pipeline"]),
            (
                dictgen::InsensitiveStr::Ascii("assisstants"),
                &["assistants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrections"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("rasons"), &["reasons"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialze"),
                &["deserialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("previvous"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("cotaining"), &["containing"]),
            (dictgen::InsensitiveStr::Ascii("auxililary"), &["auxiliary"]),
            (
                dictgen::InsensitiveStr::Ascii("interliveing"),
                &["interleaving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("follfowiwng"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contensious"),
                &["contentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satruation"),
                &["saturation"],
            ),
            (dictgen::InsensitiveStr::Ascii("headhsot"), &["headshot"]),
            (dictgen::InsensitiveStr::Ascii("enumrate"), &["enumerate"]),
            (
                dictgen::InsensitiveStr::Ascii("vaildating"),
                &["validating"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparaste"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("locte"), &["locate"]),
            (
                dictgen::InsensitiveStr::Ascii("accidantely"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertexts"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("emproves"), &["improves"]),
            (
                dictgen::InsensitiveStr::Ascii("selectin"),
                &["selection", "selecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("flacor"), &["flavor"]),
            (dictgen::InsensitiveStr::Ascii("qeuests"), &["quests"]),
            (dictgen::InsensitiveStr::Ascii("checboxes"), &["checkboxes"]),
            (
                dictgen::InsensitiveStr::Ascii("automotice"),
                &["automotive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ejaculatte"), &["ejaculate"]),
            (dictgen::InsensitiveStr::Ascii("conotrols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("huricane"), &["hurricane"]),
            (
                dictgen::InsensitiveStr::Ascii("perforances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugnal"), &["signal"]),
            (
                dictgen::InsensitiveStr::Ascii("aaproximated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cholocate"), &["chocolate"]),
            (
                dictgen::InsensitiveStr::Ascii("underlowed"),
                &["underflowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playthroughers"),
                &["playthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprieciate"),
                &["appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermientation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nummbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("microtransations"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customizeable"),
                &["customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instnsiated"),
                &["instantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("unquottes"), &["unquotes"]),
            (dictgen::InsensitiveStr::Ascii("hardwoord"), &["hardwood"]),
            (dictgen::InsensitiveStr::Ascii("antia"), &["anti"]),
            (
                dictgen::InsensitiveStr::Ascii("imformational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("pavin"), &["paving", "pain"]),
            (dictgen::InsensitiveStr::Ascii("enty"), &["entry", "entity"]),
            (dictgen::InsensitiveStr::Ascii("laborerers"), &["laborers"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantasiaing"),
                &["fantasizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underlowing"),
                &["underflowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("campain"), &["campaign"]),
            (dictgen::InsensitiveStr::Ascii("emough"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("scraching"), &["scratching"]),
            (dictgen::InsensitiveStr::Ascii("byciclist"), &["bicyclist"]),
            (
                dictgen::InsensitiveStr::Ascii("sanetisers"),
                &["sanitisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("topoligy"), &["topology"]),
            (
                dictgen::InsensitiveStr::Ascii("immoratlity"),
                &["immortality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maliciusally"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proftiable"),
                &["profitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("beld"), &["held"]),
            (dictgen::InsensitiveStr::Ascii("probblems"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("projetctor"), &["projector"]),
            (dictgen::InsensitiveStr::Ascii("restat"), &["restart"]),
            (dictgen::InsensitiveStr::Ascii("cookeies"), &["cookies"]),
            (
                dictgen::InsensitiveStr::Ascii("monumentul"),
                &["monumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("creatie"), &["creatine"]),
            (dictgen::InsensitiveStr::Ascii("exactely"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("comptuers"), &["computers"]),
            (
                dictgen::InsensitiveStr::Ascii("tyrantical"),
                &["tyrannical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docummentation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decices"),
                &["decides", "devices", "deciles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consctructive"),
                &["constructive"],
            ),
            (dictgen::InsensitiveStr::Ascii("engeneers"), &["engineers"]),
            (
                dictgen::InsensitiveStr::Ascii("contenants"),
                &["continents"],
            ),
            (dictgen::InsensitiveStr::Ascii("midpionts"), &["midpoints"]),
            (dictgen::InsensitiveStr::Ascii("teritary"), &["tertiary"]),
            (dictgen::InsensitiveStr::Ascii("optionss"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("promitives"),
                &["primitives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complilations"),
                &["complications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistankely"),
                &["mistakenly"],
            ),
            (dictgen::InsensitiveStr::Ascii("excutive"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("disappearence"),
                &["disappearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summerized"),
                &["summarized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropreations"),
                &["appropriations"],
            ),
            (dictgen::InsensitiveStr::Ascii("matainer"), &["maintainer"]),
            (dictgen::InsensitiveStr::Ascii("mereged"), &["merged"]),
            (
                dictgen::InsensitiveStr::Ascii("transmorfers"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyitic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("structuing"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("enmum"), &["enum"]),
            (
                dictgen::InsensitiveStr::Ascii("provencial"),
                &["provincial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correxpondance"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condiitonals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulgara"), &["bulgaria"]),
            (dictgen::InsensitiveStr::Ascii("repaitnt"), &["repaint"]),
            (
                dictgen::InsensitiveStr::Ascii("tempearure"),
                &["temperature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiye"),
                &["initialize", "initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownward"), &["onward"]),
            (
                dictgen::InsensitiveStr::Ascii("redumndancy"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("brekpoint"), &["breakpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("polymoprhs"),
                &["polymorphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("keychan"), &["keychain"]),
            (
                dictgen::InsensitiveStr::Ascii("cilincer"),
                &["cylinder", "silencer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compesitions"),
                &["compositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlescar"),
                &["battlestar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensinoality"),
                &["dimensionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("typpes"), &["types"]),
            (
                dictgen::InsensitiveStr::Ascii("dominanting"),
                &["dominating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequental"),
                &["consequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desitnation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrtibute"),
                &["distribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("noramal"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("projetct"), &["project"]),
            (
                dictgen::InsensitiveStr::Ascii("subroutins"),
                &["subroutines"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcaed"), &["spaced"]),
            (dictgen::InsensitiveStr::Ascii("scafold"), &["scaffold"]),
            (dictgen::InsensitiveStr::Ascii("frameword"), &["framework"]),
            (
                dictgen::InsensitiveStr::Ascii("suppressin"),
                &["suppression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilty"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discernable"),
                &["discernible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wraping"),
                &["wrapping", "warping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeliveably"),
                &["unbelievably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antialising"),
                &["antialiasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfactly"), &["perfectly"]),
            (
                dictgen::InsensitiveStr::Ascii("versatillity"),
                &["versatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("immserive"), &["immersive"]),
            (dictgen::InsensitiveStr::Ascii("poins"), &["points"]),
            (
                dictgen::InsensitiveStr::Ascii("aplicability"),
                &["applicability"],
            ),
            (dictgen::InsensitiveStr::Ascii("neruons"), &["neurons"]),
            (dictgen::InsensitiveStr::Ascii("shortcurt"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("tiitle"), &["title"]),
            (
                dictgen::InsensitiveStr::Ascii("apdatations"),
                &["adaptations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cehcklists"),
                &["checklists"],
            ),
            (dictgen::InsensitiveStr::Ascii("fabulos"), &["fabulous"]),
            (dictgen::InsensitiveStr::Ascii("weev"), &["weave"]),
            (dictgen::InsensitiveStr::Ascii("portugues"), &["portuguese"]),
            (
                dictgen::InsensitiveStr::Ascii("compensentate"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("officionados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("keywards"), &["keywords"]),
            (dictgen::InsensitiveStr::Ascii("consumate"), &["consummate"]),
            (dictgen::InsensitiveStr::Ascii("reduzing"), &["reducing"]),
            (dictgen::InsensitiveStr::Ascii("wotks"), &["works"]),
            (
                dictgen::InsensitiveStr::Ascii("atorecovery"),
                &["autorecovery"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defaulz"),
                &["default", "defaults"],
            ),
            (dictgen::InsensitiveStr::Ascii("symmetic"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("enthaply"), &["enthalpy"]),
            (
                dictgen::InsensitiveStr::Ascii("uninstlaler"),
                &["uninstaller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("godafther"), &["godfather"]),
            (dictgen::InsensitiveStr::Ascii("argemnt"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("overloooked"),
                &["overlooked"],
            ),
            (dictgen::InsensitiveStr::Ascii("flexble"), &["flexible"]),
            (dictgen::InsensitiveStr::Ascii("atrittion"), &["attrition"]),
            (dictgen::InsensitiveStr::Ascii("vaidates"), &["validates"]),
            (
                dictgen::InsensitiveStr::Ascii("psychyatrists"),
                &["psychiatrists"],
            ),
            (dictgen::InsensitiveStr::Ascii("pubished"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("vertrices"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("contructs"), &["constructs"]),
            (dictgen::InsensitiveStr::Ascii("embbedders"), &["embedders"]),
            (
                dictgen::InsensitiveStr::Ascii("implimenation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrubures"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autheticating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("oriign"), &["origin"]),
            (dictgen::InsensitiveStr::Ascii("syringae"), &["syringe"]),
            (dictgen::InsensitiveStr::Ascii("ivoked"), &["invoked"]),
            (
                dictgen::InsensitiveStr::Ascii("conventiently"),
                &["conveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("tanseint"), &["transient"]),
            (dictgen::InsensitiveStr::Ascii("posifions"), &["positions"]),
            (
                dictgen::InsensitiveStr::Ascii("conventioally"),
                &["conventionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("averageifs"), &["averages"]),
            (dictgen::InsensitiveStr::Ascii("prepairing"), &["preparing"]),
            (dictgen::InsensitiveStr::Ascii("octects"), &["octets"]),
            (
                dictgen::InsensitiveStr::Ascii("placeholdes"),
                &["placeholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("harmonisch"), &["harmonic"]),
            (dictgen::InsensitiveStr::Ascii("indicatess"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("reconeects"),
                &["reconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processsing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("exitation"), &["excitation"]),
            (
                dictgen::InsensitiveStr::Ascii("prefent"),
                &["prevent", "present", "prefect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psycological"),
                &["psychological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repressenting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bruning"), &["burning"]),
            (
                dictgen::InsensitiveStr::Ascii("typograhic"),
                &["typographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("autofilt"), &["autofilter"]),
            (dictgen::InsensitiveStr::Ascii("sturtures"), &["structures"]),
            (dictgen::InsensitiveStr::Ascii("frozee"), &["frozen"]),
            (dictgen::InsensitiveStr::Ascii("supreses"), &["suppresses"]),
            (dictgen::InsensitiveStr::Ascii("diamon"), &["diamond"]),
            (dictgen::InsensitiveStr::Ascii("anecodtes"), &["anecdotes"]),
            (dictgen::InsensitiveStr::Ascii("emojiis"), &["emojis"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoziator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("incalid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("definitavely"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theresholding"),
                &["thresholding"],
            ),
            (dictgen::InsensitiveStr::Ascii("exampple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("trangulate"),
                &["triangulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("theroy"), &["theory"]),
            (
                dictgen::InsensitiveStr::Ascii("appropropiately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("andtoid"), &["android"]),
            (
                dictgen::InsensitiveStr::Ascii("defanitely"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("thorasic"), &["thoracic"]),
            (dictgen::InsensitiveStr::Ascii("requeim"), &["requiem"]),
            (dictgen::InsensitiveStr::Ascii("atomical"), &["atomic"]),
            (dictgen::InsensitiveStr::Ascii("adealide"), &["adelaide"]),
            (
                dictgen::InsensitiveStr::Ascii("recogniton"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legionairs"),
                &["legionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("hiearchy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("sopunded"), &["sounded"]),
            (dictgen::InsensitiveStr::Ascii("kend"), &["kind"]),
            (
                dictgen::InsensitiveStr::Ascii("mutimodule"),
                &["multimodule"],
            ),
            (dictgen::InsensitiveStr::Ascii("servantes"), &["servants"]),
            (
                dictgen::InsensitiveStr::Ascii("streamining"),
                &["streaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("interrut"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilitiy"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("bund"), &["bind", "bound"]),
            (dictgen::InsensitiveStr::Ascii("tekn"), &["taken", "tekken"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficiallifed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deterministc"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decalratively"),
                &["declaratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("blosum"), &["blossom"]),
            (
                dictgen::InsensitiveStr::Ascii("labarotory"),
                &["laboratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("portagonists"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("galsgow"), &["glasgow"]),
            (
                dictgen::InsensitiveStr::Ascii("disappearred"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("penerator"), &["penetrator"]),
            (
                dictgen::InsensitiveStr::Ascii("placholder"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("alforithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("reoders"), &["reorders"]),
            (dictgen::InsensitiveStr::Ascii("represets"), &["represents"]),
            (dictgen::InsensitiveStr::Ascii("seraches"), &["searches"]),
            (
                dictgen::InsensitiveStr::Ascii("supllemented"),
                &["supplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("colective"), &["collective"]),
            (dictgen::InsensitiveStr::Ascii("expport"), &["export"]),
            (dictgen::InsensitiveStr::Ascii("alpahbets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("negoticator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenrication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("eletrodes"), &["electrodes"]),
            (
                dictgen::InsensitiveStr::Ascii("acdept"),
                &["accept", "adept"],
            ),
            (dictgen::InsensitiveStr::Ascii("worksapce"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("headqaurters"),
                &["headquarters"],
            ),
            (dictgen::InsensitiveStr::Ascii("rountinely"), &["routinely"]),
            (dictgen::InsensitiveStr::Ascii("recordarme"), &["recorder"]),
            (dictgen::InsensitiveStr::Ascii("suggetsed"), &["suggested"]),
            (dictgen::InsensitiveStr::Ascii("hotsopt"), &["hotspot"]),
            (
                dictgen::InsensitiveStr::Ascii("copiler"),
                &["compiler", "copier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoriziation"),
                &["authorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appoximating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("constarin"), &["constrain"]),
            (dictgen::InsensitiveStr::Ascii("automatted"), &["automate"]),
            (dictgen::InsensitiveStr::Ascii("fideling"), &["fielding"]),
            (
                dictgen::InsensitiveStr::Ascii("variabled"),
                &["variables", "variable"],
            ),
            (dictgen::InsensitiveStr::Ascii("certan"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("depracation"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credetials"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representaions"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interseted"),
                &["interested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nightmeres"),
                &["nightmares"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimdnters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("orcale"), &["oracle"]),
            (
                dictgen::InsensitiveStr::Ascii("popuplating"),
                &["populating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinkarnation"),
                &["reincarnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectaulars"),
                &["spectaculars"],
            ),
            (dictgen::InsensitiveStr::Ascii("synonmys"), &["synonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("paritioning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("susequent"), &["subsequent"]),
            (
                dictgen::InsensitiveStr::Ascii("invarients"),
                &["invariants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cercomstantial"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reparamterized"),
                &["reparameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("parrallell"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("stancers"), &["stances"]),
            (dictgen::InsensitiveStr::Ascii("thast"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("crosair"), &["corsair"]),
            (dictgen::InsensitiveStr::Ascii("wiil"), &["will"]),
            (dictgen::InsensitiveStr::Ascii("follwoin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("locaitions"), &["locations"]),
            (dictgen::InsensitiveStr::Ascii("problemas"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("coproration"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weghts"),
                &["weights", "wights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kommas"),
                &["comas", "commas"],
            ),
            (dictgen::InsensitiveStr::Ascii("mysogyny"), &["misogyny"]),
            (
                dictgen::InsensitiveStr::Ascii("correltions"),
                &["correlations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermoneter"),
                &["thermometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchronusly"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("porject"), &["project"]),
            (
                dictgen::InsensitiveStr::Ascii("belligerant"),
                &["belligerent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("easthetically"),
                &["esthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingestigator"),
                &["investigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibative"),
                &["prohibitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("undrstand"), &["understand"]),
            (dictgen::InsensitiveStr::Ascii("modyfing"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("anticipatin"),
                &["anticipation", "anticipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaege"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("judegmental"),
                &["judgmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throught"),
                &["thought", "through", "throughout"],
            ),
            (dictgen::InsensitiveStr::Ascii("hitory"), &["history"]),
            (
                dictgen::InsensitiveStr::Ascii("seqeuental"),
                &["sequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("eleminate"), &["eliminate"]),
            (
                dictgen::InsensitiveStr::Ascii("indonesean"),
                &["indonesian"],
            ),
            (dictgen::InsensitiveStr::Ascii("flexiliy"), &["flexibly"]),
            (dictgen::InsensitiveStr::Ascii("alongsize"), &["alongside"]),
            (dictgen::InsensitiveStr::Ascii("tranform"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("progres"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("worflow"), &["workflow"]),
            (
                dictgen::InsensitiveStr::Ascii("fingernal"),
                &["fingernails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suffisticated"),
                &["sophisticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aditionally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definetlely"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("siffix"), &["suffix"]),
            (
                dictgen::InsensitiveStr::Ascii("topoplogies"),
                &["topologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("anomally"), &["anomaly"]),
            (
                dictgen::InsensitiveStr::Ascii("pennsylvnia"),
                &["pennsylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construrcts"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnticators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capaticitor"),
                &["capacitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("resouce"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("coflicts"), &["conflicts"]),
            (dictgen::InsensitiveStr::Ascii("uhandled"), &["unhandled"]),
            (
                dictgen::InsensitiveStr::Ascii("elimintates"),
                &["eliminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("embeddin"), &["embedding"]),
            (dictgen::InsensitiveStr::Ascii("assoziate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("posthomous"),
                &["posthumous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciste"),
                &["appreciates", "appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("squirrles"), &["squirrels"]),
            (
                dictgen::InsensitiveStr::Ascii("cinncinati"),
                &["cincinnati"],
            ),
            (dictgen::InsensitiveStr::Ascii("mony"), &["money"]),
            (
                dictgen::InsensitiveStr::Ascii("respberries"),
                &["raspberries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dinamically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetsting"), &["testing"]),
            (dictgen::InsensitiveStr::Ascii("sabatosh"), &["sabotage"]),
            (dictgen::InsensitiveStr::Ascii("delimitter"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("aaccession"), &["accession"]),
            (dictgen::InsensitiveStr::Ascii("messeges"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("assime"), &["assume"]),
            (
                dictgen::InsensitiveStr::Ascii("ehtnicities"),
                &["ethnicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inplements"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("veriable"),
                &["variable", "veritable", "verifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competiveness"),
                &["competitiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refelctive"),
                &["reflective"],
            ),
            (dictgen::InsensitiveStr::Ascii("ridiculled"), &["ridicule"]),
            (dictgen::InsensitiveStr::Ascii("ahev"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("raeds"), &["reads"]),
            (dictgen::InsensitiveStr::Ascii("hesistated"), &["hesitated"]),
            (dictgen::InsensitiveStr::Ascii("securit"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("wehreas"), &["whereas"]),
            (
                dictgen::InsensitiveStr::Ascii("correspindence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integrade"),
                &["integrated", "integrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("uqest"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("regulatons"),
                &["regulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sionists"), &["zionists"]),
            (
                dictgen::InsensitiveStr::Ascii("consenquently"),
                &["consequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("upladed"), &["uploaded"]),
            (dictgen::InsensitiveStr::Ascii("disbling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("ordinarly"), &["ordinary"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapounted"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("teminal"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("proteciton"),
                &["protections"],
            ),
            (dictgen::InsensitiveStr::Ascii("normlises"), &["normalises"]),
            (dictgen::InsensitiveStr::Ascii("shwashank"), &["shawshank"]),
            (
                dictgen::InsensitiveStr::Ascii("implemet"),
                &["implements", "implement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comftorably"),
                &["comfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleifs"), &["beliefs"]),
            (
                dictgen::InsensitiveStr::Ascii("furstrated"),
                &["frustrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propositivo"),
                &["proposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governmeant"),
                &["governmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloows"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainters"),
                &["maintainers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hexigonal"), &["hexagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("regrads"),
                &["regards", "regrades"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanlification"),
                &["qualification", "quantification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaoidable"),
                &["unavoidable"],
            ),
            (dictgen::InsensitiveStr::Ascii("trotzki"), &["trotsky"]),
            (dictgen::InsensitiveStr::Ascii("ethcially"), &["ethically"]),
            (dictgen::InsensitiveStr::Ascii("simleys"), &["smileys"]),
            (
                dictgen::InsensitiveStr::Ascii("protagonits"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("alhabet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("reinforcemnt"),
                &["reinforcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progessbar"),
                &["progressbar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexperssion"),
                &["subexpression"],
            ),
            (dictgen::InsensitiveStr::Ascii("amaterus"), &["amateurs"]),
            (dictgen::InsensitiveStr::Ascii("damenor"), &["demeanor"]),
            (
                dictgen::InsensitiveStr::Ascii("architectual"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ventalation"),
                &["ventilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exitsing"),
                &["existing", "exiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooefficient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("jeircho"), &["jericho"]),
            (
                dictgen::InsensitiveStr::Ascii("informfation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("arleady"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditianally"),
                &["unconditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("verticla"), &["vertical"]),
            (dictgen::InsensitiveStr::Ascii("techniucal"), &["technical"]),
            (
                dictgen::InsensitiveStr::Ascii("atferthoughts"),
                &["afterthoughts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconsistencies"),
                &["inconsistencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("followign"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("introdusing"),
                &["introducing"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiblke"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("superfulous"),
                &["superfluous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otimisation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("manifestus"), &["manifests"]),
            (dictgen::InsensitiveStr::Ascii("detroy"), &["destroy"]),
            (
                dictgen::InsensitiveStr::Ascii("superios"),
                &["superiors", "superior"],
            ),
            (dictgen::InsensitiveStr::Ascii("leraned"), &["learned"]),
            (dictgen::InsensitiveStr::Ascii("impeed"), &["impede"]),
            (
                dictgen::InsensitiveStr::Ascii("experimmentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incrementaly"),
                &["incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understant"),
                &["understand", "understate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhasution"),
                &["exhaustion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accesories"),
                &["accessories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusivas"),
                &["exclusives"],
            ),
            (dictgen::InsensitiveStr::Ascii("abondened"), &["abandoned"]),
            (
                dictgen::InsensitiveStr::Ascii("autoincrememnt"),
                &["autoincrement"],
            ),
            (dictgen::InsensitiveStr::Ascii("sufficent"), &["sufficient"]),
            (
                dictgen::InsensitiveStr::Ascii("quesetionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ungrapeful"),
                &["ungrateful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanitifications"),
                &["quantifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaing"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("specimine"), &["specimen"]),
            (
                dictgen::InsensitiveStr::Ascii("aritficially"),
                &["artificially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affirmitve"),
                &["affirmative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumulatin"),
                &["accumulation", "accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("thhinking"), &["thinking"]),
            (
                dictgen::InsensitiveStr::Ascii("esitmations"),
                &["estimations"],
            ),
            (dictgen::InsensitiveStr::Ascii("urainum"), &["uranium"]),
            (
                dictgen::InsensitiveStr::Ascii("indroducing"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcurations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("posistions"), &["positions"]),
            (dictgen::InsensitiveStr::Ascii("espace"), &["escape"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinitively"),
                &["indefinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("assitant"), &["assistant"]),
            (dictgen::InsensitiveStr::Ascii("precisly"), &["precisely"]),
            (dictgen::InsensitiveStr::Ascii("addresssed"), &["addressed"]),
            (
                dictgen::InsensitiveStr::Ascii("forgivens"),
                &["forgiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("custumary"), &["customary"]),
            (dictgen::InsensitiveStr::Ascii("mexicants"), &["mexicans"]),
            (dictgen::InsensitiveStr::Ascii("redifine"), &["redefine"]),
            (
                dictgen::InsensitiveStr::Ascii("generlised"),
                &["generalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceptabel"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelization"),
                &["parallelization"],
            ),
            (dictgen::InsensitiveStr::Ascii("colapsed"), &["collapsed"]),
            (dictgen::InsensitiveStr::Ascii("outpun"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("algorthim"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("ensconsed"), &["ensconced"]),
            (
                dictgen::InsensitiveStr::Ascii("referecence"),
                &["reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("travercal"), &["traversal"]),
            (dictgen::InsensitiveStr::Ascii("strenghts"), &["strengths"]),
            (dictgen::InsensitiveStr::Ascii("utillised"), &["utilised"]),
            (dictgen::InsensitiveStr::Ascii("lubircant"), &["lubricant"]),
            (
                dictgen::InsensitiveStr::Ascii("soveregnity"),
                &["sovereignty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optionall"),
                &["optional", "optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveniece"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("deviatie"), &["deviate"]),
            (dictgen::InsensitiveStr::Ascii("contacs"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("conditianaly"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substituable"),
                &["substitutable"],
            ),
            (dictgen::InsensitiveStr::Ascii("deflectin"), &["deflection"]),
            (dictgen::InsensitiveStr::Ascii("trhreads"), &["threads"]),
            (dictgen::InsensitiveStr::Ascii("doub"), &["doubt", "daub"]),
            (dictgen::InsensitiveStr::Ascii("entraced"), &["entranced"]),
            (dictgen::InsensitiveStr::Ascii("unlcear"), &["unclear"]),
            (dictgen::InsensitiveStr::Ascii("reacords"), &["records"]),
            (
                dictgen::InsensitiveStr::Ascii("executalble"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("vanguad"), &["vanguard"]),
            (
                dictgen::InsensitiveStr::Ascii("analitically"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cenario"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("encompusing"),
                &["encompassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("heacy"), &["heavy"]),
            (dictgen::InsensitiveStr::Ascii("routeros"), &["routers"]),
            (
                dictgen::InsensitiveStr::Ascii("demograhic"),
                &["demographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporve"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("fiedls"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("diection"),
                &["direction", "dissection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requiement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("prolicks"), &["prolix"]),
            (
                dictgen::InsensitiveStr::Ascii("assocciating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("plcement"), &["placement"]),
            (
                dictgen::InsensitiveStr::Ascii("compatabilities"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("thq"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("cahanges"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("triancles"), &["triangles"]),
            (
                dictgen::InsensitiveStr::Ascii("authetnicate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficilleir"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapporval"),
                &["disapproval"],
            ),
            (dictgen::InsensitiveStr::Ascii("stanp"), &["stamp"]),
            (dictgen::InsensitiveStr::Ascii("unplease"), &["displease"]),
            (
                dictgen::InsensitiveStr::Ascii("nasturshum"),
                &["nasturtium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rotationd"),
                &["rotation", "rotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chagning"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("egotisitcal"),
                &["egotistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sycronously"),
                &["synchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("timestmp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("tounge"), &["tongue"]),
            (dictgen::InsensitiveStr::Ascii("crace"), &["grace", "crate"]),
            (
                dictgen::InsensitiveStr::Ascii("complate"),
                &["complacent", "complete"],
            ),
            (dictgen::InsensitiveStr::Ascii("boounds"), &["bounds"]),
            (
                dictgen::InsensitiveStr::Ascii("convinceing"),
                &["convincing"],
            ),
            (dictgen::InsensitiveStr::Ascii("repackged"), &["repackaged"]),
            (
                dictgen::InsensitiveStr::Ascii("financialy"),
                &["financially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exilerating"),
                &["exhilarating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compartmet"),
                &["compartment"],
            ),
            (dictgen::InsensitiveStr::Ascii("reallllly"), &["really"]),
            (dictgen::InsensitiveStr::Ascii("comrads"), &["comrades"]),
            (dictgen::InsensitiveStr::Ascii("remainung"), &["remaining"]),
            (dictgen::InsensitiveStr::Ascii("haa"), &["has"]),
            (dictgen::InsensitiveStr::Ascii("digitzed"), &["digitized"]),
            (dictgen::InsensitiveStr::Ascii("rceating"), &["creating"]),
            (dictgen::InsensitiveStr::Ascii("peristent"), &["persistent"]),
            (dictgen::InsensitiveStr::Ascii("possably"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("alginment"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("affinitze"), &["affinitize"]),
            (
                dictgen::InsensitiveStr::Ascii("inisialise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chaeckpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("encripted"), &["encrypted"]),
            (dictgen::InsensitiveStr::Ascii("aisian"), &["asian"]),
            (dictgen::InsensitiveStr::Ascii("scpeter"), &["scepter"]),
            (
                dictgen::InsensitiveStr::Ascii("anonimised"),
                &["anonymised"],
            ),
            (dictgen::InsensitiveStr::Ascii("saddenes"), &["saddens"]),
            (
                dictgen::InsensitiveStr::Ascii("dimineshes"),
                &["diminishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inforational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("variaty"), &["variety"]),
            (
                dictgen::InsensitiveStr::Ascii("hesistation"),
                &["hesitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemeneter"),
                &["implementer"],
            ),
            (dictgen::InsensitiveStr::Ascii("opayk"), &["opaque"]),
            (dictgen::InsensitiveStr::Ascii("multiplis"), &["multiplies"]),
            (dictgen::InsensitiveStr::Ascii("tedeous"), &["tedious"]),
            (
                dictgen::InsensitiveStr::Ascii("intriquing"),
                &["intriguing"],
            ),
            (dictgen::InsensitiveStr::Ascii("servocing"), &["servicing"]),
            (
                dictgen::InsensitiveStr::Ascii("cirumvention"),
                &["circumvention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periwankle"),
                &["periwinkle"],
            ),
            (dictgen::InsensitiveStr::Ascii("patrones"), &["patrons"]),
            (dictgen::InsensitiveStr::Ascii("ewll"), &["well"]),
            (
                dictgen::InsensitiveStr::Ascii("rasbperries"),
                &["raspberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsribed"), &["subscribed"]),
            (
                dictgen::InsensitiveStr::Ascii("recretional"),
                &["recreational"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanaguge"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("servival"), &["survival"]),
            (dictgen::InsensitiveStr::Ascii("canabels"), &["cannibals"]),
            (dictgen::InsensitiveStr::Ascii("planatery"), &["planetary"]),
            (
                dictgen::InsensitiveStr::Ascii("silckscreened"),
                &["silkscreened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarcelated"),
                &["incarcerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninspried"),
                &["uninspired"],
            ),
            (dictgen::InsensitiveStr::Ascii("berkley"), &["berkeley"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoteur"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archictecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theree"),
                &["there", "three"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electronicas"),
                &["electronics"],
            ),
            (dictgen::InsensitiveStr::Ascii("brazilias"), &["brazilians"]),
            (
                dictgen::InsensitiveStr::Ascii("acheivable"),
                &["achievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("wealthly"), &["wealthy"]),
            (dictgen::InsensitiveStr::Ascii("inhert"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("breakes"), &["breaks"]),
            (dictgen::InsensitiveStr::Ascii("applayed"), &["applied"]),
            (
                dictgen::InsensitiveStr::Ascii("toothbursh"),
                &["toothbrush"],
            ),
            (dictgen::InsensitiveStr::Ascii("rmoving"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("diected"),
                &["directed", "dissected"],
            ),
            (dictgen::InsensitiveStr::Ascii("peforming"), &["performing"]),
            (dictgen::InsensitiveStr::Ascii("reloade"), &["reload"]),
            (dictgen::InsensitiveStr::Ascii("corecct"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("distructors"),
                &["destructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("alsways"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("regeistered"),
                &["registered"],
            ),
            (dictgen::InsensitiveStr::Ascii("transpart"), &["transport"]),
            (dictgen::InsensitiveStr::Ascii("algorithmo"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("performas"), &["performs"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("passagers"), &["passages"]),
            (
                dictgen::InsensitiveStr::Ascii("maesurements"),
                &["measurements"],
            ),
            (dictgen::InsensitiveStr::Ascii("pice"), &["piece"]),
            (
                dictgen::InsensitiveStr::Ascii("imbalenced"),
                &["imbalanced"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaibly"), &["reliably"]),
            (dictgen::InsensitiveStr::Ascii("renuion"), &["reunion"]),
            (
                dictgen::InsensitiveStr::Ascii("revealtion"),
                &["revelations"],
            ),
            (dictgen::InsensitiveStr::Ascii("declareted"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("repsonsive"),
                &["responsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("beneith"), &["beneath"]),
            (dictgen::InsensitiveStr::Ascii("vertextes"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("accumlated"),
                &["accumulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accommadates"),
                &["accommodates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncomented"),
                &["uncommented"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrtiping"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("checkked"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("remainig"), &["remaining"]),
            (
                dictgen::InsensitiveStr::Ascii("disadvandages"),
                &["disadvantaged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idividuals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immigrents"),
                &["immigrants"],
            ),
            (dictgen::InsensitiveStr::Ascii("supportts"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("involuntarly"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("impluses"), &["impulses"]),
            (
                dictgen::InsensitiveStr::Ascii("acquinated"),
                &["acquainted"],
            ),
            (dictgen::InsensitiveStr::Ascii("delte"), &["delete"]),
            (
                dictgen::InsensitiveStr::Ascii("metabloism"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("penatgon"), &["pentagon"]),
            (
                dictgen::InsensitiveStr::Ascii("notificactions"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicrectly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("mozillla"), &["mozilla"]),
            (dictgen::InsensitiveStr::Ascii("abosrbs"), &["absorbs"]),
            (
                dictgen::InsensitiveStr::Ascii("ntification"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranactions"),
                &["transactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("workdlows"), &["workflows"]),
            (
                dictgen::InsensitiveStr::Ascii("handshares"),
                &["handshakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosciating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificy"),
                &["specify", "specificity", "specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frightining"),
                &["frightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("eanble"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("mysteriouly"),
                &["mysteriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("schemetic"), &["schematic"]),
            (dictgen::InsensitiveStr::Ascii("varous"), &["various"]),
            (
                dictgen::InsensitiveStr::Ascii("optomizers"),
                &["optimizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempalting"),
                &["templating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attentuation"),
                &["attenuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aktively"), &["actively"]),
            (dictgen::InsensitiveStr::Ascii("squirrelus"), &["squirrels"]),
            (dictgen::InsensitiveStr::Ascii("retunrns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("filenmae"), &["filename"]),
            (dictgen::InsensitiveStr::Ascii("unkown"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("adatper"), &["adapter"]),
            (
                dictgen::InsensitiveStr::Ascii("quesetioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabele"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("incremneted"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantistically"),
                &["fantastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("programms"), &["programs"]),
            (dictgen::InsensitiveStr::Ascii("princliple"), &["principle"]),
            (
                dictgen::InsensitiveStr::Ascii("allocaions"),
                &["allocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapated"), &["dissipated"]),
            (
                dictgen::InsensitiveStr::Ascii("abstracto"),
                &["abstraction", "abstract"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanatsic"), &["fanatics"]),
            (
                dictgen::InsensitiveStr::Ascii("exhautiveness"),
                &["exhaustiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangaing"), &["managing"]),
            (dictgen::InsensitiveStr::Ascii("valkryie"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("whick"),
                &["which", "whisk", "wick", "whack"],
            ),
            (dictgen::InsensitiveStr::Ascii("intverted"), &["inverted"]),
            (
                dictgen::InsensitiveStr::Ascii("ridicoulusly"),
                &["ridiculously"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocuppied"), &["occupied"]),
            (dictgen::InsensitiveStr::Ascii("truley"), &["truly"]),
            (dictgen::InsensitiveStr::Ascii("libary"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("eto"),
                &["to", "ego", "veto", "wto"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predocessor"),
                &["predecessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versionaddded"),
                &["versionadded"],
            ),
            (dictgen::InsensitiveStr::Ascii("suefull"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("assymmetric"),
                &["asymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressents"), &["presents"]),
            (dictgen::InsensitiveStr::Ascii("meanting"), &["meaning"]),
            (
                dictgen::InsensitiveStr::Ascii("homosexuais"),
                &["homosexuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("deamon"), &["daemon"]),
            (
                dictgen::InsensitiveStr::Ascii("sturctured"),
                &["structured"],
            ),
            (dictgen::InsensitiveStr::Ascii("zipers"), &["zippers"]),
            (
                dictgen::InsensitiveStr::Ascii("enginerring"),
                &["engineering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentatl"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("univesities"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("immediae"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("attribued"), &["attributed"]),
            (dictgen::InsensitiveStr::Ascii("thouch"), &["touch"]),
            (dictgen::InsensitiveStr::Ascii("modifled"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("errrors"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronunciato"),
                &["pronunciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaquire"),
                &["reacquire", "require"],
            ),
            (dictgen::InsensitiveStr::Ascii("preshis"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("compresor"), &["compressor"]),
            (
                dictgen::InsensitiveStr::Ascii("preferencfe"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promot"),
                &["prompt", "promote"],
            ),
            (dictgen::InsensitiveStr::Ascii("expalined"), &["explained"]),
            (dictgen::InsensitiveStr::Ascii("chatous"), &["châteaux"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuprised"),
                &["unsurprised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interviwed"),
                &["interviewed"],
            ),
            (dictgen::InsensitiveStr::Ascii("vell"), &["well"]),
            (dictgen::InsensitiveStr::Ascii("creaet"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("luancher"), &["launcher"]),
            (
                dictgen::InsensitiveStr::Ascii("recconetion"),
                &["reconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregnanies"),
                &["pregnancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("ubernetes"), &["kubernetes"]),
            (
                dictgen::InsensitiveStr::Ascii("sythesising"),
                &["synthesising"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootraps"), &["bootstraps"]),
            (
                dictgen::InsensitiveStr::Ascii("comventional"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("larvay"), &["larvae"]),
            (
                dictgen::InsensitiveStr::Ascii("independnet"),
                &["independents", "independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("sawtays"), &["sautés"]),
            (dictgen::InsensitiveStr::Ascii("abriters"), &["arbiters"]),
            (dictgen::InsensitiveStr::Ascii("bangkock"), &["bangkok"]),
            (dictgen::InsensitiveStr::Ascii("pionere"), &["pioneer"]),
            (
                dictgen::InsensitiveStr::Ascii("unsignificant"),
                &["insignificant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rleases"),
                &["releases", "pleases", "leases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipualting"),
                &["manipulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retransmited"),
                &["retransmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrideable"),
                &["overridable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encryptrion"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("barrackus"), &["barracks"]),
            (dictgen::InsensitiveStr::Ascii("sesitive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("captuing"), &["capturing"]),
            (
                dictgen::InsensitiveStr::Ascii("contrubute"),
                &["contribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("negotatie"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("paln"),
                &["plan", "pain", "palm"],
            ),
            (dictgen::InsensitiveStr::Ascii("pevented"), &["prevented"]),
            (dictgen::InsensitiveStr::Ascii("aruging"), &["arguing"]),
            (dictgen::InsensitiveStr::Ascii("fullill"), &["fulfill"]),
            (dictgen::InsensitiveStr::Ascii("udpaet"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("polinomial"),
                &["polynomial"],
            ),
            (dictgen::InsensitiveStr::Ascii("triked"), &["tricked"]),
            (
                dictgen::InsensitiveStr::Ascii("suppreses"),
                &["suppress", "suppresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("metaphores"), &["metaphors"]),
            (dictgen::InsensitiveStr::Ascii("primiarily"), &["primarily"]),
            (dictgen::InsensitiveStr::Ascii("ody"), &["body"]),
            (
                dictgen::InsensitiveStr::Ascii("reformated"),
                &["reformatted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organsiers"),
                &["organisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendayvoo"), &["rendezvous"]),
            (dictgen::InsensitiveStr::Ascii("functiuon"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("properity"),
                &["property", "proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("calidates"), &["validates"]),
            (dictgen::InsensitiveStr::Ascii("ponts"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("archieval"), &["archival"]),
            (
                dictgen::InsensitiveStr::Ascii("asscciation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comverting"),
                &["converting"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounrady"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("exexute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("princeple"), &["principle"]),
            (
                dictgen::InsensitiveStr::Ascii("prgressively"),
                &["progressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practicaly"),
                &["practically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberivation"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("getttext"), &["gettext"]),
            (dictgen::InsensitiveStr::Ascii("indentit"), &["identity"]),
            (dictgen::InsensitiveStr::Ascii("deactive"), &["deactivate"]),
            (dictgen::InsensitiveStr::Ascii("mandarian"), &["mandarin"]),
            (dictgen::InsensitiveStr::Ascii("maanger"), &["manager"]),
            (dictgen::InsensitiveStr::Ascii("admendment"), &["amendment"]),
            (
                dictgen::InsensitiveStr::Ascii("monolothic"),
                &["monolithic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparision"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquantaince"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("determin"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificas"),
                &["sacrifices"],
            ),
            (dictgen::InsensitiveStr::Ascii("insaller"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("bumpted"), &["bumped"]),
            (
                dictgen::InsensitiveStr::Ascii("iterrupting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninsipred"),
                &["uninspired"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("transmittor"),
                &["transmitter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autenticated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compliants"),
                &["complaints"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbles"), &["enables"]),
            (dictgen::InsensitiveStr::Ascii("ipmroved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("scheduals"), &["schedules"]),
            (dictgen::InsensitiveStr::Ascii("mdoe"), &["doe", "mode"]),
            (
                dictgen::InsensitiveStr::Ascii("italicizin"),
                &["italicizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("europians"), &["europeans"]),
            (
                dictgen::InsensitiveStr::Ascii("participatns"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionaities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("reappered"), &["reappeared"]),
            (dictgen::InsensitiveStr::Ascii("addaptive"), &["adaptive"]),
            (
                dictgen::InsensitiveStr::Ascii("insturcting"),
                &["instructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registring"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperislist"),
                &["imperialist"],
            ),
            (dictgen::InsensitiveStr::Ascii("orangism"), &["organism"]),
            (dictgen::InsensitiveStr::Ascii("cuestion"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("stattus"), &["status"]),
            (dictgen::InsensitiveStr::Ascii("tesellate"), &["tessellate"]),
            (dictgen::InsensitiveStr::Ascii("controlas"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("indivudal"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("abanonded"), &["abandoned"]),
            (
                dictgen::InsensitiveStr::Ascii("communistes"),
                &["communists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conincident"),
                &["coincident"],
            ),
            (dictgen::InsensitiveStr::Ascii("oblisk"), &["obelisk"]),
            (dictgen::InsensitiveStr::Ascii("fractalers"), &["fractals"]),
            (dictgen::InsensitiveStr::Ascii("exerbate"), &["exacerbate"]),
            (dictgen::InsensitiveStr::Ascii("pulishes"), &["publishes"]),
            (
                dictgen::InsensitiveStr::Ascii("accumlator"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("farmework"), &["framework"]),
            (dictgen::InsensitiveStr::Ascii("histroies"), &["histories"]),
            (dictgen::InsensitiveStr::Ascii("opression"), &["oppression"]),
            (dictgen::InsensitiveStr::Ascii("goverance"), &["governance"]),
            (
                dictgen::InsensitiveStr::Ascii("thereaputic"),
                &["therapeutic"],
            ),
            (dictgen::InsensitiveStr::Ascii("acrued"), &["accrued"]),
            (dictgen::InsensitiveStr::Ascii("restire"), &["restore"]),
            (dictgen::InsensitiveStr::Ascii("sheping"), &["shaping"]),
            (dictgen::InsensitiveStr::Ascii("whitepace"), &["whitespace"]),
            (
                dictgen::InsensitiveStr::Ascii("retroactivily"),
                &["retroactively"],
            ),
            (dictgen::InsensitiveStr::Ascii("manuallly"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("britening"),
                &["brightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("varification"),
                &["verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quesrionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradictie"),
                &["contradicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controverial"),
                &["controversial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificaitons"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("mocules"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("capsulers"), &["capsules"]),
            (
                dictgen::InsensitiveStr::Ascii("correlatas"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("backwardes"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("unespaced"), &["unescaped"]),
            (
                dictgen::InsensitiveStr::Ascii("baksetball"),
                &["basketball"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derivativos"),
                &["derivatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modefications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrehouses"), &["warehouses"]),
            (dictgen::InsensitiveStr::Ascii("betweed"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("qualifed"), &["qualified"]),
            (dictgen::InsensitiveStr::Ascii("operaor"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("bounrary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("prooerty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("avalable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("cahces"), &["caches"]),
            (dictgen::InsensitiveStr::Ascii("mulitpart"), &["multipart"]),
            (
                dictgen::InsensitiveStr::Ascii("accummulator"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("temprorary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("gess"), &["guess"]),
            (dictgen::InsensitiveStr::Ascii("cartrige"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("massachsuetts"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("opionally"), &["optionally"]),
            (
                dictgen::InsensitiveStr::Ascii("paticipations"),
                &["participations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopmepage"), &["homepage"]),
            (dictgen::InsensitiveStr::Ascii("serciving"), &["servicing"]),
            (dictgen::InsensitiveStr::Ascii("chanck"), &["chunk"]),
            (
                dictgen::InsensitiveStr::Ascii("contermporaneous"),
                &["contemporaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscrubed"),
                &["subscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("velolcity"), &["velocity"]),
            (dictgen::InsensitiveStr::Ascii("gruseome"), &["gruesome"]),
            (dictgen::InsensitiveStr::Ascii("alumi"), &["alumni"]),
            (dictgen::InsensitiveStr::Ascii("avove"), &["above"]),
            (
                dictgen::InsensitiveStr::Ascii("discourgae"),
                &["discourages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anononymously"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchrnonization"),
                &["synchronization"],
            ),
            (dictgen::InsensitiveStr::Ascii("decosing"), &["decoding"]),
            (dictgen::InsensitiveStr::Ascii("constaint"), &["constraint"]),
            (
                dictgen::InsensitiveStr::Ascii("matcht"),
                &["match", "matched"],
            ),
            (dictgen::InsensitiveStr::Ascii("carreid"), &["carried"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpecedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificaton"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desencitized"),
                &["desensitized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pershusly"), &["preciously"]),
            (
                dictgen::InsensitiveStr::Ascii("otifications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("procecure"), &["procedure"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistant"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bazare"),
                &["bazaar", "bizarre"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigthstone"),
                &["sightstone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perticular"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("endge"), &["edge", "end"]),
            (
                dictgen::InsensitiveStr::Ascii("excercising"),
                &["exercising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pubilcations"),
                &["publications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whipsering"),
                &["whispering"],
            ),
            (dictgen::InsensitiveStr::Ascii("eductes"), &["educates"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alegorical"),
                &["allegorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostris"), &["nostrils"]),
            (dictgen::InsensitiveStr::Ascii("recogiser"), &["recogniser"]),
            (
                dictgen::InsensitiveStr::Ascii("abstruct"),
                &["abstract", "obstruct"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrytpt"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("caledars"), &["calendars"]),
            (dictgen::InsensitiveStr::Ascii("warantee"), &["warranty"]),
            (
                dictgen::InsensitiveStr::Ascii("schows"),
                &["shows", "chows"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiailly"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("capabicities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("keykode"), &["keycode"]),
            (dictgen::InsensitiveStr::Ascii("schmeas"), &["schemas"]),
            (dictgen::InsensitiveStr::Ascii("hypotesis"), &["hypothesis"]),
            (
                dictgen::InsensitiveStr::Ascii("insepctions"),
                &["inspections"],
            ),
            (dictgen::InsensitiveStr::Ascii("nmae"), &["name"]),
            (
                dictgen::InsensitiveStr::Ascii("interchangable"),
                &["interchangeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unessecarrily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snowboaring"),
                &["snowboarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("onborad"), &["onboard"]),
            (
                dictgen::InsensitiveStr::Ascii("aassociated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advantagious"),
                &["advantageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decloration"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagnal"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("acrosss"), &["across"]),
            (
                dictgen::InsensitiveStr::Ascii("celebrationists"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("begininers"), &["beginners"]),
            (dictgen::InsensitiveStr::Ascii("elemenets"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("considerant"),
                &["considerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("reruning"), &["rerunning"]),
            (dictgen::InsensitiveStr::Ascii("sistematy"), &["systematy"]),
            (dictgen::InsensitiveStr::Ascii("confrims"), &["confirms"]),
            (
                dictgen::InsensitiveStr::Ascii("participat"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicitable"),
                &["predictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("shfts"), &["shifts"]),
            (dictgen::InsensitiveStr::Ascii("nastalgea"), &["nostalgia"]),
            (
                dictgen::InsensitiveStr::Ascii("cmpression"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransacitons"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("definie"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("senoirs"), &["seniors"]),
            (dictgen::InsensitiveStr::Ascii("scipping"), &["skipping"]),
            (dictgen::InsensitiveStr::Ascii("accussed"), &["accused"]),
            (dictgen::InsensitiveStr::Ascii("quesions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("iteratered"), &["iterated"]),
            (dictgen::InsensitiveStr::Ascii("overwtach"), &["overwatch"]),
            (
                dictgen::InsensitiveStr::Ascii("combinacion"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriximately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("siganls"), &["signals"]),
            (dictgen::InsensitiveStr::Ascii("sorround"), &["surround"]),
            (
                dictgen::InsensitiveStr::Ascii("huminatarian"),
                &["humanitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("nonsene"), &["nonsense"]),
            (
                dictgen::InsensitiveStr::Ascii("exorts"),
                &["exports", "extorts", "exerts", "exhorts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconencted"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplying"), &["applying"]),
            (dictgen::InsensitiveStr::Ascii("botstrap"), &["bootstrap"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcuate"),
                &["recalculate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competetiveness"),
                &["competitiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfmroming"),
                &["transforming"],
            ),
            (dictgen::InsensitiveStr::Ascii("procude"), &["produce"]),
            (
                dictgen::InsensitiveStr::Ascii("yhanks"),
                &["thanks", "yanks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recusively"),
                &["recursively", "reclusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatons"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("athiesm"), &["atheism"]),
            (
                dictgen::InsensitiveStr::Ascii("desensitzed"),
                &["desensitized"],
            ),
            (dictgen::InsensitiveStr::Ascii("namesapce"), &["namespace"]),
            (
                dictgen::InsensitiveStr::Ascii("pornograpy"),
                &["pornography"],
            ),
            (dictgen::InsensitiveStr::Ascii("spicific"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("reknown"), &["renown"]),
            (dictgen::InsensitiveStr::Ascii("misreably"), &["miserably"]),
            (
                dictgen::InsensitiveStr::Ascii("inteligent"),
                &["intelligent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagerizes"),
                &["plagiarizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reqirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identation"),
                &["indentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intgration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("burritio"), &["burrito"]),
            (dictgen::InsensitiveStr::Ascii("eifnach"), &["einfach"]),
            (
                dictgen::InsensitiveStr::Ascii("correctivos"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("comfirmed"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("hanel"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("primarlily"), &["primarily"]),
            (
                dictgen::InsensitiveStr::Ascii("servents"),
                &["servants", "serpents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inddividual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("instanty"), &["instantly"]),
            (
                dictgen::InsensitiveStr::Ascii("undertstands"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumet"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("randayvoo"), &["rendezvous"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiasation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propigation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeatation"),
                &["repetition"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeator"), &["creator"]),
            (dictgen::InsensitiveStr::Ascii("updraging"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("buttosn"), &["buttons"]),
            (dictgen::InsensitiveStr::Ascii("vengaence"), &["vengeance"]),
            (dictgen::InsensitiveStr::Ascii("dezine"), &["design"]),
            (
                dictgen::InsensitiveStr::Ascii("deficiancy"),
                &["deficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("exdcutive"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("summurises"),
                &["summarises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amelearated"),
                &["ameliorated"],
            ),
            (dictgen::InsensitiveStr::Ascii("workdlow"), &["workflow"]),
            (dictgen::InsensitiveStr::Ascii("comlainer"), &["complainer"]),
            (dictgen::InsensitiveStr::Ascii("numbet"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("visualsation"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fillling"), &["filling"]),
            (dictgen::InsensitiveStr::Ascii("reacently"), &["recently"]),
            (
                dictgen::InsensitiveStr::Ascii("tennant"),
                &["tenant", "pennant"],
            ),
            (dictgen::InsensitiveStr::Ascii("victums"), &["victims"]),
            (
                dictgen::InsensitiveStr::Ascii("teleporing"),
                &["teleporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("coud"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("exhcuast"), &["exhaust"]),
            (
                dictgen::InsensitiveStr::Ascii("celebrasion"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("orcestras"), &["orchestras"]),
            (
                dictgen::InsensitiveStr::Ascii("getimezone"),
                &["gettimezone"],
            ),
            (dictgen::InsensitiveStr::Ascii("deffaults"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("scandivania"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjascencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("recject"), &["reject"]),
            (
                dictgen::InsensitiveStr::Ascii("handshackes"),
                &["handshakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolitated"),
                &["consolidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invulnerble"),
                &["invulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimazed"), &["minimized"]),
            (dictgen::InsensitiveStr::Ascii("invokin"), &["invoking"]),
            (
                dictgen::InsensitiveStr::Ascii("renderrs"),
                &["renderers", "renders"],
            ),
            (dictgen::InsensitiveStr::Ascii("catptured"), &["captured"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentais"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inearisation"),
                &["linearisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congifured"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportaed"),
                &["incorporated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialyzed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("serailise"), &["serialise"]),
            (dictgen::InsensitiveStr::Ascii("strengthn"), &["strengthen"]),
            (dictgen::InsensitiveStr::Ascii("receied"), &["received"]),
            (
                dictgen::InsensitiveStr::Ascii("publisgers"),
                &["publishers"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoited"), &["exploited"]),
            (dictgen::InsensitiveStr::Ascii("propogate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("desingn"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("rememors"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("conchance"), &["conscience"]),
            (dictgen::InsensitiveStr::Ascii("templeos"), &["temples"]),
            (
                dictgen::InsensitiveStr::Ascii("comparotor"),
                &["comparator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overheading"),
                &["overheating"],
            ),
            (dictgen::InsensitiveStr::Ascii("standrad"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("bollcoks"), &["bollocks"]),
            (dictgen::InsensitiveStr::Ascii("indvidual"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("priviate"), &["private"]),
            (dictgen::InsensitiveStr::Ascii("enhaced"), &["enhanced"]),
            (
                dictgen::InsensitiveStr::Ascii("unattracive"),
                &["unattractive"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdate"), &["anecdote"]),
            (dictgen::InsensitiveStr::Ascii("propehts"), &["prophets"]),
            (dictgen::InsensitiveStr::Ascii("suprress"), &["suppress"]),
            (dictgen::InsensitiveStr::Ascii("summmers"), &["summers"]),
            (dictgen::InsensitiveStr::Ascii("cheks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("distructions"),
                &["distractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrdinating"),
                &["coordinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("movespeeed"), &["movespeed"]),
            (
                dictgen::InsensitiveStr::Ascii("daming"),
                &["damning", "damping", "doming", "gaming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repsonsibly"),
                &["responsibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("bomberos"), &["bombers"]),
            (dictgen::InsensitiveStr::Ascii("disuptes"), &["disputes"]),
            (
                dictgen::InsensitiveStr::Ascii("erformance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("abbrviate"), &["abbreviate"]),
            (dictgen::InsensitiveStr::Ascii("operaiton"), &["operation"]),
            (
                dictgen::InsensitiveStr::Ascii("intercorse"),
                &["intercourse"],
            ),
            (dictgen::InsensitiveStr::Ascii("recoginize"), &["recognize"]),
            (dictgen::InsensitiveStr::Ascii("deconded"), &["decoded"]),
            (
                dictgen::InsensitiveStr::Ascii("occurd"),
                &["occurred", "occur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculuations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiint"), &["point"]),
            (dictgen::InsensitiveStr::Ascii("relatated"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("perpose"),
                &["purpose", "propose"],
            ),
            (dictgen::InsensitiveStr::Ascii("goverining"), &["governing"]),
            (dictgen::InsensitiveStr::Ascii("landscaps"), &["landscapes"]),
            (
                dictgen::InsensitiveStr::Ascii("negotioble"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bodyheight"),
                &["bodyweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("palne"), &["plane"]),
            (dictgen::InsensitiveStr::Ascii("disscovery"), &["discovery"]),
            (dictgen::InsensitiveStr::Ascii("montanha"), &["montana"]),
            (dictgen::InsensitiveStr::Ascii("austrija"), &["austria"]),
            (
                dictgen::InsensitiveStr::Ascii("obejctions"),
                &["objections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convertirea"),
                &["converter"],
            ),
            (dictgen::InsensitiveStr::Ascii("suporting"), &["supporting"]),
            (dictgen::InsensitiveStr::Ascii("adultrey"), &["adultery"]),
            (dictgen::InsensitiveStr::Ascii("broardcast"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("prjection"), &["projection"]),
            (dictgen::InsensitiveStr::Ascii("webistes"), &["websites"]),
            (dictgen::InsensitiveStr::Ascii("iteraton"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("governmet"),
                &["governments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematicas"),
                &["mathematics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persistented"),
                &["persisted"],
            ),
            (dictgen::InsensitiveStr::Ascii("veiwers"), &["viewers"]),
            (dictgen::InsensitiveStr::Ascii("cousings"), &["cousins"]),
            (dictgen::InsensitiveStr::Ascii("aboslving"), &["absolving"]),
            (
                dictgen::InsensitiveStr::Ascii("gunsiinger"),
                &["gunslinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("algerbaic"), &["algebraic"]),
            (dictgen::InsensitiveStr::Ascii("adaptated"), &["adapted"]),
            (
                dictgen::InsensitiveStr::Ascii("caharacters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicergence"),
                &["divergence"],
            ),
            (dictgen::InsensitiveStr::Ascii("projectie"), &["projectile"]),
            (dictgen::InsensitiveStr::Ascii("veiwer"), &["viewer"]),
            (dictgen::InsensitiveStr::Ascii("nechanism"), &["mechanism"]),
            (dictgen::InsensitiveStr::Ascii("pruchased"), &["purchased"]),
            (dictgen::InsensitiveStr::Ascii("dependandt"), &["dependant"]),
            (dictgen::InsensitiveStr::Ascii("occurrin"), &["occurring"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifeir"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("observin"), &["observing"]),
            (
                dictgen::InsensitiveStr::Ascii("acheivment"),
                &["achievement"],
            ),
            (dictgen::InsensitiveStr::Ascii("parrallels"), &["parallels"]),
            (dictgen::InsensitiveStr::Ascii("ovveride"), &["override"]),
            (dictgen::InsensitiveStr::Ascii("poistivly"), &["positively"]),
            (dictgen::InsensitiveStr::Ascii("languauge"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("ninjs"),
                &["ninja", "ninjas"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unvavailable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repuglicans"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("bradcast"), &["broadcast"]),
            (
                dictgen::InsensitiveStr::Ascii("misogynyst"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facors"),
                &["favors", "fakers"],
            ),
            (dictgen::InsensitiveStr::Ascii("smoebody"), &["somebody"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiatons"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("errouneously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chaarcters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimitade"),
                &["intimidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tortilini"), &["tortellini"]),
            (dictgen::InsensitiveStr::Ascii("beatels"), &["beatles"]),
            (dictgen::InsensitiveStr::Ascii("beatiful"), &["beautiful"]),
            (
                dictgen::InsensitiveStr::Ascii("covnersation"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("accidentes"), &["accidents"]),
            (
                dictgen::InsensitiveStr::Ascii("hipotetically"),
                &["hypothetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prologomena"),
                &["prolegomena"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depcrecates"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ediitable"), &["editable"]),
            (
                dictgen::InsensitiveStr::Ascii("accepatble"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potientials"),
                &["potentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alterative"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoymous"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("otherwrite"), &["overwrite"]),
            (
                dictgen::InsensitiveStr::Ascii("microhpone"),
                &["microphone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscirptions"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passtrough"),
                &["passthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnostocism"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulrt"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("identifieras"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undestructible"),
                &["indestructible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("solidariety"),
                &["solidarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("clishay"), &["cliché"]),
            (
                dictgen::InsensitiveStr::Ascii("instuments"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analye"),
                &["analyse", "analyze"],
            ),
            (dictgen::InsensitiveStr::Ascii("moddeled"), &["modeled"]),
            (dictgen::InsensitiveStr::Ascii("bried"), &["brief", "bride"]),
            (
                dictgen::InsensitiveStr::Ascii("asociation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomsphere"),
                &["atmosphere"],
            ),
            (dictgen::InsensitiveStr::Ascii("invoce"), &["invoice"]),
            (
                dictgen::InsensitiveStr::Ascii("sympothetic"),
                &["sympathetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("varsiables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("inititlized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cammond"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("wrod"), &["word"]),
            (
                dictgen::InsensitiveStr::Ascii("cofficient"),
                &["coefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occassional"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avaialability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("positvely"), &["positively"]),
            (dictgen::InsensitiveStr::Ascii("atttribute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("exatcly"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("incementally"),
                &["incrementally"],
            ),
            (dictgen::InsensitiveStr::Ascii("implememnt"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("dpendently"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nominatino"),
                &["nominations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cassete"), &["cassette"]),
            (
                dictgen::InsensitiveStr::Ascii("electrinic"),
                &["electrician"],
            ),
            (dictgen::InsensitiveStr::Ascii("parametre"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("datatime"), &["datetime"]),
            (
                dictgen::InsensitiveStr::Ascii("destinarion"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecessairly"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformate"),
                &["transform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassocations"),
                &["disassociations"],
            ),
            (dictgen::InsensitiveStr::Ascii("signture"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("descreasing"),
                &["decreasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bookmakrs"), &["bookmarks"]),
            (
                dictgen::InsensitiveStr::Ascii("summury"),
                &["summary", "summery"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intruduction"),
                &["introduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particapate"),
                &["participate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitivly"),
                &["competitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("archvier"), &["archiver"]),
            (
                dictgen::InsensitiveStr::Ascii("universaly"),
                &["universally"],
            ),
            (dictgen::InsensitiveStr::Ascii("consious"), &["conscious"]),
            (
                dictgen::InsensitiveStr::Ascii("possitioned"),
                &["positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expontential"),
                &["exponential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underheight"),
                &["underweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calciulating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anayltically"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("depcits"), &["depicts"]),
            (
                dictgen::InsensitiveStr::Ascii("unconsistent"),
                &["inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistical"), &["mystical"]),
            (dictgen::InsensitiveStr::Ascii("redeable"), &["readable"]),
            (
                dictgen::InsensitiveStr::Ascii("descirptor"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representating"),
                &["representation", "representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persepctive"),
                &["perspective"],
            ),
            (dictgen::InsensitiveStr::Ascii("acients"), &["ancients"]),
            (dictgen::InsensitiveStr::Ascii("furition"), &["fruition"]),
            (
                dictgen::InsensitiveStr::Ascii("transprent"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("veamantly"), &["vehemently"]),
            (
                dictgen::InsensitiveStr::Ascii("electroyltes"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteraors"), &["iterators"]),
            (dictgen::InsensitiveStr::Ascii("spainsh"), &["spanish"]),
            (dictgen::InsensitiveStr::Ascii("neveloped"), &["enveloped"]),
            (
                dictgen::InsensitiveStr::Ascii("discconections"),
                &["disconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgarted"),
                &["downgraded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceeptable"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("invalis"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("convuluted"),
                &["convoluted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplaicating"),
                &["duplicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destruktive"),
                &["destructive"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntents"), &["contents"]),
            (
                dictgen::InsensitiveStr::Ascii("correxpond"),
                &["correspond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreviatins"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatons"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termimally"),
                &["terminally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoorithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("impletment"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("spleled"),
                &["spelled", "spieled"],
            ),
            (dictgen::InsensitiveStr::Ascii("swicth"), &["switch"]),
            (dictgen::InsensitiveStr::Ascii("visblie"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("debbuger"), &["debugger"]),
            (dictgen::InsensitiveStr::Ascii("specfying"), &["specifying"]),
            (dictgen::InsensitiveStr::Ascii("embebbed"), &["embedded"]),
            (
                dictgen::InsensitiveStr::Ascii("subsituent"),
                &["substituent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("principlas"),
                &["principals"],
            ),
            (dictgen::InsensitiveStr::Ascii("follos"), &["follows"]),
            (dictgen::InsensitiveStr::Ascii("wigdets"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("bitiwse"), &["bitwise"]),
            (dictgen::InsensitiveStr::Ascii("redefinine"), &["redefine"]),
            (
                dictgen::InsensitiveStr::Ascii("theoreticl"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("restired"), &["restored"]),
            (
                dictgen::InsensitiveStr::Ascii("administratio"),
                &["administrator", "administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("langages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("depdendant"), &["dependant"]),
            (
                dictgen::InsensitiveStr::Ascii("determinse"),
                &["determines"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinopses"), &["synopses"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoteds"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanification"),
                &["qualification", "quantification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("girlfrinds"),
                &["girlfriends"],
            ),
            (dictgen::InsensitiveStr::Ascii("dagners"), &["dangers"]),
            (
                dictgen::InsensitiveStr::Ascii("reminiscense"),
                &["reminiscence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compromize"),
                &["compromise"],
            ),
            (dictgen::InsensitiveStr::Ascii("debugginf"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("haved"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("derefrence"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignemt"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("progarammability"),
                &["programmability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitifes"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perosnalities"),
                &["personalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("adherance"), &["adherence"]),
            (dictgen::InsensitiveStr::Ascii("multpily"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("proptly"), &["promptly"]),
            (
                dictgen::InsensitiveStr::Ascii("neglibible"),
                &["negligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("insipre"), &["inspire"]),
            (
                dictgen::InsensitiveStr::Ascii("intsalling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("coallesence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thougt"),
                &["thought", "though"],
            ),
            (dictgen::InsensitiveStr::Ascii("tacticts"), &["tactics"]),
            (dictgen::InsensitiveStr::Ascii("circlin"), &["circling"]),
            (
                dictgen::InsensitiveStr::Ascii("corrrespond"),
                &["correspond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpoder"),
                &["transported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disnabilities"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thorwn"),
                &["thrown", "thorn"],
            ),
            (dictgen::InsensitiveStr::Ascii("escapse"), &["escapes"]),
            (
                dictgen::InsensitiveStr::Ascii("nasturshems"),
                &["nasturtiums"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagnosi"), &["diagnostic"]),
            (
                dictgen::InsensitiveStr::Ascii("instrucional"),
                &["instructional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canibalizing"),
                &["cannibalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("secutity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("sumbits"), &["submits"]),
            (dictgen::InsensitiveStr::Ascii("vriament"), &["vraiment"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnearabule"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("assighors"), &["assignors"]),
            (
                dictgen::InsensitiveStr::Ascii("charicters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifing"),
                &["simplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statu"),
                &["status", "statue"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceeeding"), &["exceeding"]),
            (
                dictgen::InsensitiveStr::Ascii("furtermore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patial"),
                &["spatial", "partial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ivocations"),
                &["invocations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specififers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("instales"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("tpyically"), &["typically"]),
            (
                dictgen::InsensitiveStr::Ascii("unknowinly"),
                &["unknowingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adborbing"),
                &["adsorbing", "absorbing"],
            ),
            (dictgen::InsensitiveStr::Ascii("garantees"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("rationalizaiton"),
                &["rationalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entertaing"),
                &["entertaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteger"), &["integer"]),
            (dictgen::InsensitiveStr::Ascii("ortogonal"), &["orthogonal"]),
            (dictgen::InsensitiveStr::Ascii("globbally"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("cleaneup"), &["cleanup"]),
            (
                dictgen::InsensitiveStr::Ascii("unprobably"),
                &["improbably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politicans"),
                &["politicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("roaylties"), &["royalties"]),
            (dictgen::InsensitiveStr::Ascii("usinging"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("dinosuar"), &["dinosaur"]),
            (dictgen::InsensitiveStr::Ascii("montitor"), &["monitor"]),
            (dictgen::InsensitiveStr::Ascii("vaccuumed"), &["vacuumed"]),
            (
                dictgen::InsensitiveStr::Ascii("expermentations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("oyu"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("thows"),
                &["those", "throws", "tows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncorrectly"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsloete"), &["obsolete"]),
            (
                dictgen::InsensitiveStr::Ascii("speicfications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ancapsulated"),
                &["encapsulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pershus"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("outsidr"), &["outsider"]),
            (dictgen::InsensitiveStr::Ascii("amalgums"), &["amalgams"]),
            (dictgen::InsensitiveStr::Ascii("onlu"), &["only"]),
            (dictgen::InsensitiveStr::Ascii("alltough"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("staggaring"),
                &["staggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("tipicals"), &["typical"]),
            (dictgen::InsensitiveStr::Ascii("alinments"), &["alignments"]),
            (
                dictgen::InsensitiveStr::Ascii("temperarily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("erraneously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomralizes"),
                &["normalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfenses"), &["offenses"]),
            (dictgen::InsensitiveStr::Ascii("tetching"), &["fetching"]),
            (
                dictgen::InsensitiveStr::Ascii("authenenticators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitue"),
                &["constitutes", "constitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphasizin"),
                &["emphasizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throroughly"),
                &["thoroughly"],
            ),
            (dictgen::InsensitiveStr::Ascii("branckes"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("adjactend"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbothoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("adresing"), &["addressing"]),
            (
                dictgen::InsensitiveStr::Ascii("gentlemanne"),
                &["gentlemen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civelisation"),
                &["civilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("colses"), &["closes"]),
            (dictgen::InsensitiveStr::Ascii("survivied"), &["survived"]),
            (
                dictgen::InsensitiveStr::Ascii("americanis"),
                &["americas", "americans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posioned"),
                &["poisoned", "positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuratuion"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("miminize"), &["minimize"]),
            (
                dictgen::InsensitiveStr::Ascii("rewatchimg"),
                &["rewatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understnads"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aboroginal"),
                &["aboriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("errorneusly"),
                &["erroneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("libraby"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("alogirhtm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("interplotates"),
                &["interpolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarunteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowledagble"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("imliment"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("unforetunately"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvemnets"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawed"), &["drew", "drawn"]),
            (dictgen::InsensitiveStr::Ascii("formattd"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("auhtor"), &["author"]),
            (
                dictgen::InsensitiveStr::Ascii("decsending"),
                &["descending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequesites"),
                &["prerequisites"],
            ),
            (dictgen::InsensitiveStr::Ascii("comammand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("chache"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("thigns"), &["things"]),
            (
                dictgen::InsensitiveStr::Ascii("permantly"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclicked"),
                &["overclocked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automatin"),
                &["automating", "automation", "automatic", "automaton"],
            ),
            (dictgen::InsensitiveStr::Ascii("sttetings"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanetelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("ventillate"), &["ventilate"]),
            (
                dictgen::InsensitiveStr::Ascii("volatillity"),
                &["volatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("aksing"), &["asking"]),
            (dictgen::InsensitiveStr::Ascii("arlighty"), &["alrighty"]),
            (dictgen::InsensitiveStr::Ascii("trabajao"), &["trabajo"]),
            (dictgen::InsensitiveStr::Ascii("defecit"), &["deficit"]),
            (
                dictgen::InsensitiveStr::Ascii("probablility"),
                &["probability"],
            ),
            (dictgen::InsensitiveStr::Ascii("agression"), &["aggression"]),
            (dictgen::InsensitiveStr::Ascii("michagan"), &["michigan"]),
            (dictgen::InsensitiveStr::Ascii("retrevier"), &["retriever"]),
            (
                dictgen::InsensitiveStr::Ascii("aymptotically"),
                &["asymptotically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusstoms"), &["customs"]),
            (dictgen::InsensitiveStr::Ascii("unoffical"), &["unofficial"]),
            (dictgen::InsensitiveStr::Ascii("windoe"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("authenificates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("kighbosh"), &["kibosh"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstanings"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("anulling"), &["annulling"]),
            (dictgen::InsensitiveStr::Ascii("campaignes"), &["campaigns"]),
            (dictgen::InsensitiveStr::Ascii("chaneging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("relativitly"),
                &["relativity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psudonymous"),
                &["pseudonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("updats"), &["updates"]),
            (
                dictgen::InsensitiveStr::Ascii("recongnition"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctiosn"),
                &["correction", "corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("assciate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("represenative"),
                &["representative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversatiosn"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("epty"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("vritualize"),
                &["virtualize"],
            ),
            (dictgen::InsensitiveStr::Ascii("preverved"), &["preserved"]),
            (dictgen::InsensitiveStr::Ascii("itertation"), &["iteration"]),
            (dictgen::InsensitiveStr::Ascii("goveremnt"), &["government"]),
            (
                dictgen::InsensitiveStr::Ascii("maintaners"),
                &["maintainers"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpsects"), &["inspects"]),
            (
                dictgen::InsensitiveStr::Ascii("foreseaable"),
                &["foreseeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("comliant"), &["compliant"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitalizations"),
                &["reinitializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celars"),
                &["cellars", "clears"],
            ),
            (dictgen::InsensitiveStr::Ascii("unversal"), &["universal"]),
            (dictgen::InsensitiveStr::Ascii("thesholds"), &["thresholds"]),
            (dictgen::InsensitiveStr::Ascii("syllabel"), &["syllable"]),
            (
                dictgen::InsensitiveStr::Ascii("performanes"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redundnacy"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("follewed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoweldged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimise"), &["minimise"]),
            (dictgen::InsensitiveStr::Ascii("progaramms"), &["programs"]),
            (dictgen::InsensitiveStr::Ascii("comrpess"), &["compress"]),
            (
                dictgen::InsensitiveStr::Ascii("thorougnly"),
                &["thoroughly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obhectification"),
                &["objectification"],
            ),
            (dictgen::InsensitiveStr::Ascii("typcast"), &["typecast"]),
            (
                dictgen::InsensitiveStr::Ascii("dicussions"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perposed"),
                &["proposed", "purposed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurrances"),
                &["recurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelligente"),
                &["intelligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unifed"),
                &["unified", "united", "unfed"],
            ),
            (dictgen::InsensitiveStr::Ascii("simbol"), &["symbol"]),
            (dictgen::InsensitiveStr::Ascii("thikness"), &["thickness"]),
            (
                dictgen::InsensitiveStr::Ascii("supppressed"),
                &["suppressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependding"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("theire"), &["their"]),
            (dictgen::InsensitiveStr::Ascii("explicity"), &["explicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepreneuer"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("separartor"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("granchild"), &["grandchild"]),
            (
                dictgen::InsensitiveStr::Ascii("retartation"),
                &["retardation"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteraterate"), &["iterate"]),
            (
                dictgen::InsensitiveStr::Ascii("percenatge"),
                &["percentages", "percentage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoroporates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indimidation"),
                &["intimidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alises"), &["aliases"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmectries"),
                &["asymmetries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otpimising"),
                &["optimising"],
            ),
            (dictgen::InsensitiveStr::Ascii("polulated"), &["populated"]),
            (dictgen::InsensitiveStr::Ascii("anye"), &["any", "ante"]),
            (dictgen::InsensitiveStr::Ascii("publiaher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("sythesise"), &["synthesise"]),
            (
                dictgen::InsensitiveStr::Ascii("registerring"),
                &["registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("capcaitor"), &["capacitor"]),
            (dictgen::InsensitiveStr::Ascii("poiters"), &["pointers"]),
            (dictgen::InsensitiveStr::Ascii("interbal"), &["internal"]),
            (dictgen::InsensitiveStr::Ascii("spport"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("athist"),
                &["atheist", "athirst"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functinality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architetcurally"),
                &["architecturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kernul"),
                &["colonel", "kernel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethenticator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ceaser"),
                &["cesar", "caesar"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgiused"), &["disguised"]),
            (
                dictgen::InsensitiveStr::Ascii("mercenaires"),
                &["mercenaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwwise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("compulosry"),
                &["compulsory"],
            ),
            (dictgen::InsensitiveStr::Ascii("scenarions"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("atendee"), &["attendee"]),
            (dictgen::InsensitiveStr::Ascii("untoched"), &["untouched"]),
            (
                dictgen::InsensitiveStr::Ascii("unvailability"),
                &["unavailability"],
            ),
            (dictgen::InsensitiveStr::Ascii("handsake"), &["handshake"]),
            (dictgen::InsensitiveStr::Ascii("decribes"), &["describes"]),
            (dictgen::InsensitiveStr::Ascii("cassims"), &["chasms"]),
            (
                dictgen::InsensitiveStr::Ascii("fascinatin"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("retireds"), &["retires"]),
            (dictgen::InsensitiveStr::Ascii("possble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("cardiovasuclar"),
                &["cardiovascular"],
            ),
            (dictgen::InsensitiveStr::Ascii("doubious"), &["dubious"]),
            (
                dictgen::InsensitiveStr::Ascii("schedualers"),
                &["schedulers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidentically"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifcation"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schizoprhenic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("mostlky"), &["mostly"]),
            (
                dictgen::InsensitiveStr::Ascii("redused"),
                &["reduced", "refused", "reused"],
            ),
            (dictgen::InsensitiveStr::Ascii("clapsed"), &["collapsed"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinitly"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramaterisation"),
                &["parameterisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sighths"),
                &["scythes", "sights"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribure"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("sufficated"),
                &["suffocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("herocis"), &["heroics"]),
            (dictgen::InsensitiveStr::Ascii("usind"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("prject"), &["project"]),
            (
                dictgen::InsensitiveStr::Ascii("circunvention"),
                &["circumvention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compoonents"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("illumini"), &["illuminati"]),
            (dictgen::InsensitiveStr::Ascii("perfom"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("giv"), &["give"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscribin"),
                &["unsubscribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("chiselles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("strenghening"),
                &["strengthening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appliacations"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("meddos"), &["meadows"]),
            (
                dictgen::InsensitiveStr::Ascii("tramas"),
                &["trams", "traumas"],
            ),
            (dictgen::InsensitiveStr::Ascii("insanelly"), &["insanely"]),
            (dictgen::InsensitiveStr::Ascii("sepecify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangteable"),
                &["arrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("arry"), &["array", "carry"]),
            (dictgen::InsensitiveStr::Ascii("scema"), &["schema"]),
            (dictgen::InsensitiveStr::Ascii("bitfilelds"), &["bitfields"]),
            (dictgen::InsensitiveStr::Ascii("negarive"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("impcat"), &["impact"]),
            (
                dictgen::InsensitiveStr::Ascii("annonating"),
                &["annotating"],
            ),
            (dictgen::InsensitiveStr::Ascii("avods"), &["avoids"]),
            (
                dictgen::InsensitiveStr::Ascii("expirimental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symtomatic"),
                &["symptomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("smealting"), &["smelting"]),
            (dictgen::InsensitiveStr::Ascii("serentiy"), &["serenity"]),
            (dictgen::InsensitiveStr::Ascii("versatel"), &["versatile"]),
            (dictgen::InsensitiveStr::Ascii("mamory"), &["memory"]),
            (
                dictgen::InsensitiveStr::Ascii("accomidated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("isoloated"), &["isolated"]),
            (
                dictgen::InsensitiveStr::Ascii("extraodrinarily"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("satuaday"), &["saturday"]),
            (
                dictgen::InsensitiveStr::Ascii("territoris"),
                &["territories"],
            ),
            (dictgen::InsensitiveStr::Ascii("exctacted"), &["extracted"]),
            (
                dictgen::InsensitiveStr::Ascii("transtitions"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rehtorical"),
                &["rhetorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("resovled"), &["resolved"]),
            (dictgen::InsensitiveStr::Ascii("separed"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("pasitioning"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("ultimative"), &["ultimate"]),
            (
                dictgen::InsensitiveStr::Ascii("maybellline"),
                &["maybelline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemememntation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("culumative"),
                &["cumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchaging"), &["exchanging"]),
            (
                dictgen::InsensitiveStr::Ascii("intialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbacion"),
                &["masturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("workboos"), &["workbooks"]),
            (
                dictgen::InsensitiveStr::Ascii("partitially"),
                &["partially"],
            ),
            (dictgen::InsensitiveStr::Ascii("aserting"), &["asserting"]),
            (
                dictgen::InsensitiveStr::Ascii("controversey"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleaving"), &["believing"]),
            (dictgen::InsensitiveStr::Ascii("moeny"), &["money"]),
            (dictgen::InsensitiveStr::Ascii("assmumes"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("agregators"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restruction"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulenrability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreceate"),
                &["deprecate", "depreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("manging"), &["managing"]),
            (
                dictgen::InsensitiveStr::Ascii("retunring"),
                &["returning", "retuning"],
            ),
            (dictgen::InsensitiveStr::Ascii("roatation"), &["rotation"]),
            (
                dictgen::InsensitiveStr::Ascii("painlkillers"),
                &["painkillers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rollercaoster"),
                &["rollercoaster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relfection"),
                &["reflection"],
            ),
            (dictgen::InsensitiveStr::Ascii("avergae"), &["average"]),
            (
                dictgen::InsensitiveStr::Ascii("mantaining"),
                &["maintaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("tabls"), &["tables"]),
            (
                dictgen::InsensitiveStr::Ascii("requriements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("lexial"), &["lexical"]),
            (dictgen::InsensitiveStr::Ascii("requime"), &["requiem"]),
            (
                dictgen::InsensitiveStr::Ascii("malicousally"),
                &["maliciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("cubburds"), &["cupboards"]),
            (dictgen::InsensitiveStr::Ascii("keyowrds"), &["keywords"]),
            (
                dictgen::InsensitiveStr::Ascii("undescored"),
                &["underscored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volonteers"),
                &["volunteers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accompagning"),
                &["accompanying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliase"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaiblale"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("valdate"), &["validate"]),
            (dictgen::InsensitiveStr::Ascii("uggested"), &["suggested"]),
            (
                dictgen::InsensitiveStr::Ascii("encyclopeida"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("maintaint"), &["maintain"]),
            (
                dictgen::InsensitiveStr::Ascii("paravirutalised"),
                &["paravirtualised"],
            ),
            (dictgen::InsensitiveStr::Ascii("wakup"), &["wakeup"]),
            (dictgen::InsensitiveStr::Ascii("higeine"), &["hygiene"]),
            (dictgen::InsensitiveStr::Ascii("iniitally"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("greatful"), &["grateful"]),
            (dictgen::InsensitiveStr::Ascii("acccept"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("ridicilousness"),
                &["ridiculousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reorganisin"),
                &["reorganising"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethikng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("contibutions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("epiphanny"), &["epiphany"]),
            (dictgen::InsensitiveStr::Ascii("excerise"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("reinstal"), &["reinstall"]),
            (dictgen::InsensitiveStr::Ascii("somehere"), &["somewhere"]),
            (dictgen::InsensitiveStr::Ascii("hcange"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucesssfuly"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("succsses"), &["successes"]),
            (
                dictgen::InsensitiveStr::Ascii("prioritzation"),
                &["prioritization"],
            ),
            (dictgen::InsensitiveStr::Ascii("assiming"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("psudonymity"),
                &["pseudonymity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referances"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ethincally"),
                &["ethnically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descentences"),
                &["descendents"],
            ),
            (dictgen::InsensitiveStr::Ascii("contstrain"), &["constrain"]),
            (dictgen::InsensitiveStr::Ascii("aproached"), &["approached"]),
            (dictgen::InsensitiveStr::Ascii("aqcuired"), &["acquired"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecesarrily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("acuracy"), &["accuracy"]),
            (
                dictgen::InsensitiveStr::Ascii("impementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vessles"), &["vessels"]),
            (dictgen::InsensitiveStr::Ascii("privilegie"), &["privilege"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchonous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commerorative"),
                &["commemorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("handbok"), &["handbook"]),
            (dictgen::InsensitiveStr::Ascii("mananges"), &["manages"]),
            (
                dictgen::InsensitiveStr::Ascii("interfeers"),
                &["interferes"],
            ),
            (dictgen::InsensitiveStr::Ascii("technolgy"), &["technology"]),
            (dictgen::InsensitiveStr::Ascii("producto"), &["production"]),
            (dictgen::InsensitiveStr::Ascii("debnia"), &["debian"]),
            (dictgen::InsensitiveStr::Ascii("bouind"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("verical"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("comaptibility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializatin"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("goldfisch"), &["goldfish"]),
            (dictgen::InsensitiveStr::Ascii("annotaors"), &["annotators"]),
            (dictgen::InsensitiveStr::Ascii("donwvote"), &["downvote"]),
            (dictgen::InsensitiveStr::Ascii("threrefore"), &["therefore"]),
            (
                dictgen::InsensitiveStr::Ascii("accesiiblity"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emmediately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("curret"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("cahgne"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("dispoasbles"),
                &["disposables"],
            ),
            (dictgen::InsensitiveStr::Ascii("matchin"), &["matching"]),
            (dictgen::InsensitiveStr::Ascii("seqense"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("exellent"), &["excellent"]),
            (
                dictgen::InsensitiveStr::Ascii("compatability"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("acitive"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("expiremented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alighn"),
                &["align", "alight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("booes"),
                &["boos", "booze", "buoys"],
            ),
            (dictgen::InsensitiveStr::Ascii("currest"), &["currents"]),
            (
                dictgen::InsensitiveStr::Ascii("preferaable"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstructers"),
                &["reconstructors", "reconstructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("torrest"), &["torrents"]),
            (dictgen::InsensitiveStr::Ascii("fater"), &["faster"]),
            (dictgen::InsensitiveStr::Ascii("seperares"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("recoginise"), &["recognise"]),
            (
                dictgen::InsensitiveStr::Ascii("credintials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("examlpe"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdenent"),
                &["independents", "independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlaysis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("incroporated"),
                &["incorporated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laf"),
                &["kaf", "kaph", "lac", "lad", "lag", "laugh", "leaf", "loaf"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centralisin"),
                &["centralising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colser"),
                &["closer", "colder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deifinitions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("judgemet"), &["judgements"]),
            (dictgen::InsensitiveStr::Ascii("guarnanty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("incured"), &["incurred"]),
            (
                dictgen::InsensitiveStr::Ascii("comprimised"),
                &["compromised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommded"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("decyrpt"), &["decrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("attributrion"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sperations"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("optio"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("aglorithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("indoneisan"),
                &["indonesian"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrice"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("lithuana"), &["lithuania"]),
            (dictgen::InsensitiveStr::Ascii("roiginals"), &["originals"]),
            (dictgen::InsensitiveStr::Ascii("aggrgate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("excepion"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("conditionns"),
                &["conditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("haethen"), &["heathen"]),
            (
                dictgen::InsensitiveStr::Ascii("parrallelise"),
                &["parallelise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charicterized"),
                &["characterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerain"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("mocrocodes"),
                &["microcodes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoatiations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vehciles"), &["vehicles"]),
            (dictgen::InsensitiveStr::Ascii("methodes"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("confugirations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("konwledgeable"),
                &["knowledgeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conider"),
                &["consider", "conifer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annyoingly"),
                &["annoyingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("quaterion"), &["quaternion"]),
            (dictgen::InsensitiveStr::Ascii("oultlines"), &["outlines"]),
            (
                dictgen::InsensitiveStr::Ascii("reconsturctors"),
                &["reconstructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("acounts"), &["accounts"]),
            (dictgen::InsensitiveStr::Ascii("magnitue"), &["magnitude"]),
            (dictgen::InsensitiveStr::Ascii("imprtant"), &["important"]),
            (dictgen::InsensitiveStr::Ascii("enforrce"), &["enforce"]),
            (dictgen::InsensitiveStr::Ascii("cjoices"), &["choices"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrviation"),
                &["abbreviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esitmation"),
                &["estimation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rleased"),
                &["released", "pleased", "leased"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udnerstanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("rectengle"), &["rectangle"]),
            (
                dictgen::InsensitiveStr::Ascii("oficionados"),
                &["aficionados"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("californican"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previvously"),
                &["previously"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyschotic"), &["psychotic"]),
            (
                dictgen::InsensitiveStr::Ascii("sublassing"),
                &["subclassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniator"), &["initiator"]),
            (
                dictgen::InsensitiveStr::Ascii("cannibalsm"),
                &["cannibalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefififed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("accomodation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condictionally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cateogrizations"),
                &["categorizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scaramento"),
                &["sacramento"],
            ),
            (dictgen::InsensitiveStr::Ascii("trunctes"), &["truncates"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivties"),
                &["sensitivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reassinging"),
                &["reassigning"],
            ),
            (dictgen::InsensitiveStr::Ascii("personaes"), &["personas"]),
            (dictgen::InsensitiveStr::Ascii("entension"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("urfaced"), &["surfaced"]),
            (dictgen::InsensitiveStr::Ascii("dcok"), &["dock"]),
            (dictgen::InsensitiveStr::Ascii("owener"), &["owner"]),
            (dictgen::InsensitiveStr::Ascii("faciltate"), &["facilitate"]),
            (
                dictgen::InsensitiveStr::Ascii("contradice"),
                &["contradicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsrcibed"),
                &["subscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontainous"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifieras"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("animaors"), &["animators"]),
            (
                dictgen::InsensitiveStr::Ascii("directoves"),
                &["directives"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfitiers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("soiurce"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("aplhabetical"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("itterative"), &["iterative"]),
            (dictgen::InsensitiveStr::Ascii("guarentys"), &["guarantees"]),
            (dictgen::InsensitiveStr::Ascii("scaleing"), &["scaling"]),
            (
                dictgen::InsensitiveStr::Ascii("freameworks"),
                &["frameworks"],
            ),
            (dictgen::InsensitiveStr::Ascii("acecss"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("unti"), &["until"]),
            (
                dictgen::InsensitiveStr::Ascii("threscholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satistaction"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("redumndant"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("interveneing"),
                &["intervening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversubcribed"),
                &["oversubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("constrans"), &["constrains"]),
            (
                dictgen::InsensitiveStr::Ascii("deprication"),
                &["deprivation", "deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("requireing"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("permenent"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("inculded"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("entchants"), &["enchants"]),
            (
                dictgen::InsensitiveStr::Ascii("summerising"),
                &["summarising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simaltaneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outpus"),
                &["output", "outputs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commenstatus"),
                &["commentstatus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governmential"),
                &["governmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteization"),
                &["characterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("crashign"), &["crashing"]),
            (dictgen::InsensitiveStr::Ascii("consntants"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("hirarcy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("reigsry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("attched"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("principas"), &["principals"]),
            (dictgen::InsensitiveStr::Ascii("namspaces"), &["namespaces"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitializer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("pltform"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("opertaes"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("multiverese"),
                &["multiverse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susceptibile"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloesely"), &["closely"]),
            (dictgen::InsensitiveStr::Ascii("chizziled"), &["chiseled"]),
            (dictgen::InsensitiveStr::Ascii("nuknown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("ampitheaters"),
                &["amphitheaters"],
            ),
            (dictgen::InsensitiveStr::Ascii("thah"), &["than"]),
            (
                dictgen::InsensitiveStr::Ascii("satisfactoraly"),
                &["satisfactory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("romanitcally"),
                &["romantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exprression"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectioms"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capacators"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transsmision"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialistions"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("structre"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("otimistic"), &["optimistic"]),
            (
                dictgen::InsensitiveStr::Ascii("complicaed"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartize"),
                &["standardize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimmetrical"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasformer"),
                &["transformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatble"),
                &["calculatable", "calculable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nastertiem"),
                &["nasturtium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acitivations"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tripels"), &["triples"]),
            (
                dictgen::InsensitiveStr::Ascii("shceduling"),
                &["scheduling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifizes"),
                &["sacrifices"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguired"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("experesed"), &["expressed"]),
            (dictgen::InsensitiveStr::Ascii("thialand"), &["thailand"]),
            (
                dictgen::InsensitiveStr::Ascii("restructed"),
                &["restricted", "restructured"],
            ),
            (dictgen::InsensitiveStr::Ascii("buidlers"), &["builders"]),
            (
                dictgen::InsensitiveStr::Ascii("theologicial"),
                &["theological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchronlous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceessor"), &["processor"]),
            (dictgen::InsensitiveStr::Ascii("exeucting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("braodcast"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("arrages"), &["arranges"]),
            (
                dictgen::InsensitiveStr::Ascii("improssible"),
                &["impossible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mamberships"),
                &["memberships"],
            ),
            (dictgen::InsensitiveStr::Ascii("retruns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("reulted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("constituye"),
                &["constitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excluslvely"),
                &["exclusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountats"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("boradband"), &["broadband"]),
            (dictgen::InsensitiveStr::Ascii("allococate"), &["allocate"]),
            (dictgen::InsensitiveStr::Ascii("applyd"), &["applied"]),
            (
                dictgen::InsensitiveStr::Ascii("intializtions"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("nutritents"), &["nutrients"]),
            (
                dictgen::InsensitiveStr::Ascii("replacated"),
                &["replicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrribution"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpotting"),
                &["transporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("hese"), &["these"]),
            (dictgen::InsensitiveStr::Ascii("valkriye"), &["valkyrie"]),
            (dictgen::InsensitiveStr::Ascii("logitudes"), &["longitudes"]),
            (dictgen::InsensitiveStr::Ascii("upgades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("ponints"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("allocted"), &["allocated"]),
            (
                dictgen::InsensitiveStr::Ascii("destructo"),
                &["destruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurises"),
                &["pasteurises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understnadable"),
                &["understandable"],
            ),
            (dictgen::InsensitiveStr::Ascii("accapting"), &["accepting"]),
            (dictgen::InsensitiveStr::Ascii("proprely"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("accompanyed"),
                &["accompanied"],
            ),
            (dictgen::InsensitiveStr::Ascii("meessaging"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("accorddingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("shouldnot"), &["shouldnt"]),
            (dictgen::InsensitiveStr::Ascii("asteorid"), &["asteroid"]),
            (dictgen::InsensitiveStr::Ascii("diplomancy"), &["diplomacy"]),
            (
                dictgen::InsensitiveStr::Ascii("interrrupts"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amiguities"),
                &["ambiguities"],
            ),
            (dictgen::InsensitiveStr::Ascii("obtaiend"), &["obtained"]),
            (dictgen::InsensitiveStr::Ascii("homewolrd"), &["homeworld"]),
            (dictgen::InsensitiveStr::Ascii("eclispe"), &["eclipse"]),
            (dictgen::InsensitiveStr::Ascii("dependicy"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("repelases"), &["releases"]),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cercumstance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("asignee"), &["assignee"]),
            (
                dictgen::InsensitiveStr::Ascii("authoratative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noninitalized"),
                &["noninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("uglyness"), &["ugliness"]),
            (
                dictgen::InsensitiveStr::Ascii("mediciney"),
                &["mediciny", "medicine", "medicinal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custamizes"),
                &["customizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("opacitiy"), &["opacity"]),
            (dictgen::InsensitiveStr::Ascii("optionaly"), &["optionally"]),
            (dictgen::InsensitiveStr::Ascii("temperal"), &["temporal"]),
            (
                dictgen::InsensitiveStr::Ascii("expropiation"),
                &["expropriation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deliviered"), &["delivered"]),
            (dictgen::InsensitiveStr::Ascii("arounf"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("travaler"), &["traveler"]),
            (dictgen::InsensitiveStr::Ascii("forgein"), &["foreign"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmistic"),
                &["optimistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximeting"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommmenting"),
                &["uncommenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethimg"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("assimptions"),
                &["assumptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("threhold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("paricularly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("scopin"), &["scoping"]),
            (dictgen::InsensitiveStr::Ascii("notidied"), &["notified"]),
            (
                dictgen::InsensitiveStr::Ascii("directionary"),
                &["dictionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituee"),
                &["prostitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("egistries"), &["registries"]),
            (
                dictgen::InsensitiveStr::Ascii("prediceted"),
                &["predicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("notaion"), &["notation"]),
            (
                dictgen::InsensitiveStr::Ascii("umpredictable"),
                &["unpredictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("quanitze"), &["quantize"]),
            (
                dictgen::InsensitiveStr::Ascii("dependencis"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unchangable"),
                &["unchangeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summerization"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("lunguage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("assigneble"),
                &["assignable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependong"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("sustitution"),
                &["substitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("igonorando"), &["ignorando"]),
            (
                dictgen::InsensitiveStr::Ascii("exlcusives"),
                &["exclusives"],
            ),
            (dictgen::InsensitiveStr::Ascii("refaction"), &["refraction"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigioso"),
                &["prestigious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhilerates"),
                &["exhilarates"],
            ),
            (dictgen::InsensitiveStr::Ascii("comand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("necessarally"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("famiy"), &["family"]),
            (dictgen::InsensitiveStr::Ascii("numberring"), &["numbering"]),
            (dictgen::InsensitiveStr::Ascii("appendig"), &["appending"]),
            (
                dictgen::InsensitiveStr::Ascii("religiousy"),
                &["religiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substantialy"),
                &["substantially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilibium"),
                &["equilibrium"],
            ),
            (dictgen::InsensitiveStr::Ascii("assitsant"), &["assistant"]),
            (
                dictgen::InsensitiveStr::Ascii("dependice"),
                &["dependency", "dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("methids"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("predicement"),
                &["predicament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercontinential"),
                &["intercontinental"],
            ),
            (dictgen::InsensitiveStr::Ascii("valtages"), &["voltages"]),
            (dictgen::InsensitiveStr::Ascii("fitlered"), &["filtered"]),
            (dictgen::InsensitiveStr::Ascii("apppender"), &["appender"]),
            (
                dictgen::InsensitiveStr::Ascii("sourcs"),
                &["sources", "source"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("mutiplied"), &["multiplied"]),
            (
                dictgen::InsensitiveStr::Ascii("posibility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("islamisists"),
                &["islamists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("octohedron"),
                &["octahedron"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestado"),
                &["manifesto"],
            ),
            (dictgen::InsensitiveStr::Ascii("karakoe"), &["karaoke"]),
            (
                dictgen::InsensitiveStr::Ascii("determinas"),
                &["determines"],
            ),
            (dictgen::InsensitiveStr::Ascii("whitsle"), &["whistle"]),
            (dictgen::InsensitiveStr::Ascii("cafetera"), &["cafeteria"]),
            (dictgen::InsensitiveStr::Ascii("exexutes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("downoloaded"),
                &["downloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("skateing"), &["skating"]),
            (
                dictgen::InsensitiveStr::Ascii("magneficent"),
                &["magnificent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrabution"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissaponited"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("minized"), &["minimized"]),
            (dictgen::InsensitiveStr::Ascii("connetced"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("initialliser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiuguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("treasue"), &["treasure"]),
            (dictgen::InsensitiveStr::Ascii("adivsory"), &["advisory"]),
            (dictgen::InsensitiveStr::Ascii("indigenos"), &["indigenous"]),
            (dictgen::InsensitiveStr::Ascii("wirte"), &["write"]),
            (
                dictgen::InsensitiveStr::Ascii("monopolice"),
                &["monopolies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exluding"),
                &["excluding", "exuding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dircetional"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("diong"), &["doing"]),
            (dictgen::InsensitiveStr::Ascii("flippin"), &["flipping"]),
            (dictgen::InsensitiveStr::Ascii("ciphrs"), &["ciphers"]),
            (dictgen::InsensitiveStr::Ascii("editied"), &["edited"]),
            (dictgen::InsensitiveStr::Ascii("abolute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("predicitive"),
                &["predictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatiion"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistatch"), &["mismatch"]),
            (dictgen::InsensitiveStr::Ascii("regurlarly"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("celisus"), &["celsius"]),
            (
                dictgen::InsensitiveStr::Ascii("suggestieve"),
                &["suggestive"],
            ),
            (dictgen::InsensitiveStr::Ascii("accessile"), &["accessible"]),
            (dictgen::InsensitiveStr::Ascii("origines"), &["origins"]),
            (dictgen::InsensitiveStr::Ascii("offeres"), &["offers"]),
            (dictgen::InsensitiveStr::Ascii("techician"), &["technician"]),
            (dictgen::InsensitiveStr::Ascii("frequncy"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("seeems"), &["seems"]),
            (dictgen::InsensitiveStr::Ascii("sitill"), &["still"]),
            (
                dictgen::InsensitiveStr::Ascii("interactons"),
                &["interacts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philedelphia"),
                &["philadelphia"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomrrow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("seperatated"),
                &["separated"],
            ),
            (dictgen::InsensitiveStr::Ascii("athiests"), &["atheists"]),
            (
                dictgen::InsensitiveStr::Ascii("apppreciates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("collectes"), &["collects"]),
            (dictgen::InsensitiveStr::Ascii("scriptng"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("archaelogy"),
                &["archaeology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatin"),
                &["combinations", "combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfecct"), &["perfect"]),
            (dictgen::InsensitiveStr::Ascii("incrmenet"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("empiers"), &["empires"]),
            (dictgen::InsensitiveStr::Ascii("suppsed"), &["supposed"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialize"),
                &["initialize", "initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("seezes"), &["seizes"]),
            (dictgen::InsensitiveStr::Ascii("maesuring"), &["measuring"]),
            (dictgen::InsensitiveStr::Ascii("falesly"), &["falsely"]),
            (
                dictgen::InsensitiveStr::Ascii("unavilable"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("whcih"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("sufficate"), &["suffocate"]),
            (
                dictgen::InsensitiveStr::Ascii("knowlegable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("spinlcok"), &["spinlock"]),
            (dictgen::InsensitiveStr::Ascii("firmwwre"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("salveof"), &["slaveof"]),
            (
                dictgen::InsensitiveStr::Ascii("relationshp"),
                &["relationships"],
            ),
            (dictgen::InsensitiveStr::Ascii("maline"), &["malign"]),
            (
                dictgen::InsensitiveStr::Ascii("prostituye"),
                &["prostitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("klunky"), &["clunky"]),
            (dictgen::InsensitiveStr::Ascii("argumeny"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("overlfows"), &["overflows"]),
            (dictgen::InsensitiveStr::Ascii("activatd"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("subseqent"), &["subsequent"]),
            (dictgen::InsensitiveStr::Ascii("decisison"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("dumbell"), &["dumbbell"]),
            (
                dictgen::InsensitiveStr::Ascii("destruccion"),
                &["destruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endien"),
                &["endian", "indian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substractions"),
                &["subtractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("adeqaute"), &["adequate"]),
            (
                dictgen::InsensitiveStr::Ascii("cerified"),
                &["certified", "verified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comforming"),
                &["comforting", "conforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coexhisting"),
                &["coexisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("aniway"), &["anyway"]),
            (
                dictgen::InsensitiveStr::Ascii("dissspointed"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppeners"), &["openers"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliverse"),
                &["multiverse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compaining"),
                &["complaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeal"), &["appeal"]),
            (
                dictgen::InsensitiveStr::Ascii("presrciption"),
                &["prescriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("physican"), &["physician"]),
            (dictgen::InsensitiveStr::Ascii("horrable"), &["horrible"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("excempt"), &["exempt"]),
            (
                dictgen::InsensitiveStr::Ascii("authetnicates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("separtors"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contengency"),
                &["contingency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivelants"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architectur"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnning"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("aequidistant"),
                &["equidistant"],
            ),
            (dictgen::InsensitiveStr::Ascii("disard"), &["discard"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminiti"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("softwrae"), &["software"]),
            (dictgen::InsensitiveStr::Ascii("animaiting"), &["animating"]),
            (
                dictgen::InsensitiveStr::Ascii("physisions"),
                &["physicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stimulaton"),
                &["stimulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missalignment"),
                &["misalignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repored"),
                &["reported", "reposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("refusin"), &["refusing"]),
            (
                dictgen::InsensitiveStr::Ascii("identifyes"),
                &["identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("shortend"), &["shortened"]),
            (dictgen::InsensitiveStr::Ascii("seperare"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("manufature"),
                &["manufacture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progapated"),
                &["propagated"],
            ),
            (dictgen::InsensitiveStr::Ascii("honeslty"), &["honestly"]),
            (
                dictgen::InsensitiveStr::Ascii("subconsicous"),
                &["subconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("stadnards"), &["standards"]),
            (
                dictgen::InsensitiveStr::Ascii("illustrato"),
                &["illustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("intiate"), &["initiate"]),
            (dictgen::InsensitiveStr::Ascii("follwer"), &["follower"]),
            (
                dictgen::InsensitiveStr::Ascii("susbtantial"),
                &["substantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("techeis"), &["techies"]),
            (
                dictgen::InsensitiveStr::Ascii("evauluations"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesssory"), &["accessory"]),
            (dictgen::InsensitiveStr::Ascii("vassales"), &["vassals"]),
            (
                dictgen::InsensitiveStr::Ascii("proocessing"),
                &["processing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enccryption"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyhon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("conigure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("emai"), &["email"]),
            (dictgen::InsensitiveStr::Ascii("kow"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("situationnal"),
                &["situational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fille"),
                &["file", "fill", "filled"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeendix"), &["appendix"]),
            (
                dictgen::InsensitiveStr::Ascii("definitivno"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resubstituion"),
                &["resubstitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkare"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("freqeuntly"),
                &["frequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("seseeds"), &["secedes"]),
            (dictgen::InsensitiveStr::Ascii("extendes"), &["extends"]),
            (dictgen::InsensitiveStr::Ascii("heaviliy"), &["heavily"]),
            (dictgen::InsensitiveStr::Ascii("correte"), &["correlate"]),
            (dictgen::InsensitiveStr::Ascii("follwiwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("zombes"), &["zombies"]),
            (dictgen::InsensitiveStr::Ascii("listenes"), &["listens"]),
            (dictgen::InsensitiveStr::Ascii("chuncked"), &["chunked"]),
            (dictgen::InsensitiveStr::Ascii("arctifact"), &["artifact"]),
            (dictgen::InsensitiveStr::Ascii("cluases"), &["clauses"]),
            (
                dictgen::InsensitiveStr::Ascii("modivational"),
                &["motivational"],
            ),
            (dictgen::InsensitiveStr::Ascii("streammed"), &["streamed"]),
            (
                dictgen::InsensitiveStr::Ascii("fuctionality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("forworded"), &["forwarded"]),
            (dictgen::InsensitiveStr::Ascii("repblics"), &["republics"]),
            (
                dictgen::InsensitiveStr::Ascii("comuting"),
                &["commuting", "computing"],
            ),
            (dictgen::InsensitiveStr::Ascii("asessing"), &["assessing"]),
            (dictgen::InsensitiveStr::Ascii("lesbains"), &["lesbians"]),
            (
                dictgen::InsensitiveStr::Ascii("comediantes"),
                &["comedians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combintations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonize"),
                &["synchronize"],
            ),
            (dictgen::InsensitiveStr::Ascii("limti"), &["limit"]),
            (dictgen::InsensitiveStr::Ascii("expeience"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("abstruction"),
                &["abstraction", "obstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucesful"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("edcdic"), &["ebcdic"]),
            (
                dictgen::InsensitiveStr::Ascii("parametters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corporatoins"),
                &["corporations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrupttion"),
                &["interruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaugely"), &["vaguely"]),
            (
                dictgen::InsensitiveStr::Ascii("methodoligies"),
                &["methodologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("straigth"), &["straight"]),
            (
                dictgen::InsensitiveStr::Ascii("definitin"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("xmdoel"), &["xmodel"]),
            (dictgen::InsensitiveStr::Ascii("posifion"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("enthisiasts"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("selcet"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("losd"),
                &["lost", "loss", "lose", "load"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanndle"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("satursday"), &["saturday"]),
            (
                dictgen::InsensitiveStr::Ascii("composutes"),
                &["composites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buddhistisk"),
                &["buddhists"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginnigs"), &["beginnings"]),
            (
                dictgen::InsensitiveStr::Ascii("defiantely"),
                &["definitely", "defiantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("metods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("hijkacks"), &["hijacks"]),
            (
                dictgen::InsensitiveStr::Ascii("manufcaturers"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("delared"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("simulatanious"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("complette"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("reelation"), &["relation"]),
            (
                dictgen::InsensitiveStr::Ascii("decalarator"),
                &["declarator"],
            ),
            (dictgen::InsensitiveStr::Ascii("effekt"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("incmopatible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("dismante"), &["dismantle"]),
            (dictgen::InsensitiveStr::Ascii("carimony"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("ugraded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("convertire"), &["converter"]),
            (dictgen::InsensitiveStr::Ascii("autmation"), &["automation"]),
            (
                dictgen::InsensitiveStr::Ascii("expeirimenter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("newmonitis"),
                &["pneumonitis"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssert"), &["assert"]),
            (dictgen::InsensitiveStr::Ascii("thaughts"), &["thoughts"]),
            (dictgen::InsensitiveStr::Ascii("lockscren"), &["lockscreen"]),
            (dictgen::InsensitiveStr::Ascii("sunifre"), &["sunfire"]),
            (dictgen::InsensitiveStr::Ascii("mutiply"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("scollbar"), &["scrollbar"]),
            (
                dictgen::InsensitiveStr::Ascii("aggresssion"),
                &["aggression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perdictable"),
                &["predictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sluaghter"), &["slaughter"]),
            (dictgen::InsensitiveStr::Ascii("decerning"), &["discerning"]),
            (
                dictgen::InsensitiveStr::Ascii("expriation"),
                &["expiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("enoungh"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("paramterts"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("lvoe"), &["love"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrations"),
                &["aggregations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finall"),
                &["final", "finally"],
            ),
            (dictgen::InsensitiveStr::Ascii("melbounre"), &["melbourne"]),
            (dictgen::InsensitiveStr::Ascii("trians"), &["trains"]),
            (dictgen::InsensitiveStr::Ascii("compiliant"), &["compliant"]),
            (
                dictgen::InsensitiveStr::Ascii("ganerating"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginses"),
                &["recognises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idications"),
                &["indications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defleciton"),
                &["deflection"],
            ),
            (dictgen::InsensitiveStr::Ascii("wqs"), &["was"]),
            (dictgen::InsensitiveStr::Ascii("voluntarly"), &["voluntary"]),
            (dictgen::InsensitiveStr::Ascii("reposity"), &["repository"]),
            (
                dictgen::InsensitiveStr::Ascii("skeptisicm"),
                &["skepticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimaly"),
                &["legitimacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("dramtic"), &["dramatic"]),
            (
                dictgen::InsensitiveStr::Ascii("effciently"),
                &["efficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperitive"),
                &["imperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogirhtmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("santized"), &["sanitized"]),
            (dictgen::InsensitiveStr::Ascii("visheate"), &["vitiate"]),
            (dictgen::InsensitiveStr::Ascii("guidence"), &["guidance"]),
            (
                dictgen::InsensitiveStr::Ascii("dectections"),
                &["detections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecjtures"),
                &["conjectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("shaprness"), &["sharpness"]),
            (
                dictgen::InsensitiveStr::Ascii("unresonable"),
                &["unreasonable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjecutre"),
                &["conjecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("enouncter"), &["encounter"]),
            (
                dictgen::InsensitiveStr::Ascii("consideren"),
                &["considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postgressql"),
                &["postgresql"],
            ),
            (dictgen::InsensitiveStr::Ascii("reversably"), &["reversal"]),
            (dictgen::InsensitiveStr::Ascii("collum"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("enlsave"), &["enslave"]),
            (
                dictgen::InsensitiveStr::Ascii("highligjted"),
                &["highlighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessefully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("clickbat"), &["clickbait"]),
            (dictgen::InsensitiveStr::Ascii("cenerated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("kommissioner"),
                &["commissioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constracted"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("anout"), &["about"]),
            (
                dictgen::InsensitiveStr::Ascii("guarnetees"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessitites"),
                &["necessities"],
            ),
            (dictgen::InsensitiveStr::Ascii("scannig"), &["scanning"]),
            (dictgen::InsensitiveStr::Ascii("dataste"), &["dataset"]),
            (
                dictgen::InsensitiveStr::Ascii("stadnardizing"),
                &["standardizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceates"), &["creates"]),
            (dictgen::InsensitiveStr::Ascii("calibler"), &["caliber"]),
            (
                dictgen::InsensitiveStr::Ascii("mulfunctioning"),
                &["malfunctioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isconnection"),
                &["isconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoemr"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("counsellling"),
                &["counselling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypocrises"),
                &["hypocrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfotable"),
                &["comfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enchances"),
                &["enhances", "enchants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muliplexer"),
                &["multiplexer"],
            ),
            (dictgen::InsensitiveStr::Ascii("techology"), &["technology"]),
            (dictgen::InsensitiveStr::Ascii("membrance"), &["membrane"]),
            (dictgen::InsensitiveStr::Ascii("cataylst"), &["catalyst"]),
            (dictgen::InsensitiveStr::Ascii("omages"), &["homages"]),
            (
                dictgen::InsensitiveStr::Ascii("vaid"),
                &["valid", "void", "vapid"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hesistations"),
                &["hesitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sweatshrit"),
                &["sweatshirt"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabeled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perticularly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("staticaly"), &["statically"]),
            (
                dictgen::InsensitiveStr::Ascii("ueslessness"),
                &["uselessness"],
            ),
            (dictgen::InsensitiveStr::Ascii("rectifiy"), &["rectify"]),
            (dictgen::InsensitiveStr::Ascii("knowlege"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("asscoitaing"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("assignted"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("botton"),
                &["bottom", "button"],
            ),
            (dictgen::InsensitiveStr::Ascii("primevil"), &["primeval"]),
            (dictgen::InsensitiveStr::Ascii("worspace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("starlted"), &["startled"]),
            (dictgen::InsensitiveStr::Ascii("widgtes"), &["widgets"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminato"),
                &["contamination"],
            ),
            (dictgen::InsensitiveStr::Ascii("slowely"), &["slowly"]),
            (
                dictgen::InsensitiveStr::Ascii("interative"),
                &["interactive", "iterative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forementionned"),
                &["aforementioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("dictateurs"), &["dictates"]),
            (dictgen::InsensitiveStr::Ascii("labellin"), &["labelling"]),
            (
                dictgen::InsensitiveStr::Ascii("satety"),
                &["safety", "satiety"],
            ),
            (dictgen::InsensitiveStr::Ascii("descions"), &["decisions"]),
            (
                dictgen::InsensitiveStr::Ascii("allowences"),
                &["allowances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzelosly"),
                &["overzealously"],
            ),
            (dictgen::InsensitiveStr::Ascii("flexibel"), &["flexible"]),
            (dictgen::InsensitiveStr::Ascii("diffucult"), &["difficult"]),
            (
                dictgen::InsensitiveStr::Ascii("constallations"),
                &["constellations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sttutering"),
                &["stuttering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bordlerands"),
                &["borderlands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefereably"),
                &["preferably"],
            ),
            (dictgen::InsensitiveStr::Ascii("releaf"), &["relief"]),
            (
                dictgen::InsensitiveStr::Ascii("speces"),
                &["spaces", "species", "spices"],
            ),
            (dictgen::InsensitiveStr::Ascii("abiove"), &["above"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstanting"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("studis"),
                &["studies", "studios"],
            ),
            (dictgen::InsensitiveStr::Ascii("extoics"), &["exotics"]),
            (
                dictgen::InsensitiveStr::Ascii("decrambled"),
                &["descrambled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivqlent"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("stiples"), &["stipples"]),
            (
                dictgen::InsensitiveStr::Ascii("exepectations"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mobiel"), &["mobile"]),
            (dictgen::InsensitiveStr::Ascii("attributen"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("creationistas"),
                &["creationists"],
            ),
            (dictgen::InsensitiveStr::Ascii("emtpiness"), &["emptiness"]),
            (dictgen::InsensitiveStr::Ascii("buldning"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("perferrances"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuppier"), &["occupier"]),
            (
                dictgen::InsensitiveStr::Ascii("socailizing"),
                &["socializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("annointing"), &["anointing"]),
            (dictgen::InsensitiveStr::Ascii("sensibile"), &["sensible"]),
            (
                dictgen::InsensitiveStr::Ascii("subsconcious"),
                &["subconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("wireframw"), &["wireframe"]),
            (
                dictgen::InsensitiveStr::Ascii("misscounted"),
                &["miscounted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annihilited"),
                &["annihilated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralellizes"),
                &["parallelizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("saniziter"), &["sanitizer"]),
            (dictgen::InsensitiveStr::Ascii("accsepted"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("atempts"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("distict"),
                &["distinct", "district"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreators"), &["operators"]),
            (
                dictgen::InsensitiveStr::Ascii("reinforcemt"),
                &["reinforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("isses"), &["issues"]),
            (
                dictgen::InsensitiveStr::Ascii("undertandable"),
                &["understandable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associtive"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fileststems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("tsuanmi"), &["tsunami"]),
            (dictgen::InsensitiveStr::Ascii("embasssy"), &["embassy"]),
            (dictgen::InsensitiveStr::Ascii("sorrounds"), &["surrounds"]),
            (dictgen::InsensitiveStr::Ascii("belowe"), &["below"]),
            (
                dictgen::InsensitiveStr::Ascii("guerrillera"),
                &["guerrilla"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearched"),
                &["searched", "researched", "reached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traditiona"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("prexifing"), &["prefixing"]),
            (dictgen::InsensitiveStr::Ascii("senetence"), &["sentence"]),
            (dictgen::InsensitiveStr::Ascii("optmizers"), &["optimizers"]),
            (dictgen::InsensitiveStr::Ascii("improter"), &["importer"]),
            (dictgen::InsensitiveStr::Ascii("atremis"), &["artemis"]),
            (
                dictgen::InsensitiveStr::Ascii("diffentiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("houskeeping"),
                &["housekeeping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simuated"),
                &["simulated", "situated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenstionality"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomodata"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpahabetically"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("forece"), &["force"]),
            (
                dictgen::InsensitiveStr::Ascii("condradictions"),
                &["contradictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialiazation"),
                &["serialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangmenet"),
                &["rearrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("garantors"), &["guarantors"]),
            (
                dictgen::InsensitiveStr::Ascii("implecations"),
                &["implications"],
            ),
            (dictgen::InsensitiveStr::Ascii("hovewer"), &["however"]),
            (dictgen::InsensitiveStr::Ascii("relpase"), &["relapse"]),
            (dictgen::InsensitiveStr::Ascii("rememered"), &["remembered"]),
            (
                dictgen::InsensitiveStr::Ascii("processsors"),
                &["processors"],
            ),
            (dictgen::InsensitiveStr::Ascii("comanded"), &["commanded"]),
            (
                dictgen::InsensitiveStr::Ascii("forgivness"),
                &["forgiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integerating"),
                &["integrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("destkop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("blockhains"),
                &["blockchains"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuctor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("consitionally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirtualisaion"),
                &["devirtualisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmign"),
                &["programming"],
            ),
            (dictgen::InsensitiveStr::Ascii("debbuging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("shoould"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("unimplemneted"),
                &["unimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("urugauy"), &["uruguay"]),
            (
                dictgen::InsensitiveStr::Ascii("depenendency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationilist"),
                &["sensationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heapdhones"),
                &["headphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("faceboook"), &["facebook"]),
            (
                dictgen::InsensitiveStr::Ascii("validataion"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("immortalis"), &["immortals"]),
            (
                dictgen::InsensitiveStr::Ascii("simmetriclly"),
                &["symmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compuslory"),
                &["compulsory"],
            ),
            (dictgen::InsensitiveStr::Ascii("idirectly"), &["indirectly"]),
            (dictgen::InsensitiveStr::Ascii("wounderful"), &["wonderful"]),
            (dictgen::InsensitiveStr::Ascii("treshhold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("panckaes"), &["pancakes"]),
            (dictgen::InsensitiveStr::Ascii("hode"), &["node"]),
            (dictgen::InsensitiveStr::Ascii("execures"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("deffinitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techmological"),
                &["technological"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetrainty"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("advace"), &["advance"]),
            (dictgen::InsensitiveStr::Ascii("faciliate"), &["facilitate"]),
            (
                dictgen::InsensitiveStr::Ascii("unilaterlly"),
                &["unilaterally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("makretplace"),
                &["marketplace"],
            ),
            (dictgen::InsensitiveStr::Ascii("saddends"), &["saddens"]),
            (
                dictgen::InsensitiveStr::Ascii("instnsiations"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generalice"),
                &["generalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpature"), &["capture"]),
            (
                dictgen::InsensitiveStr::Ascii("inofrmational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("paload"), &["payload"]),
            (dictgen::InsensitiveStr::Ascii("draging"), &["dragging"]),
            (
                dictgen::InsensitiveStr::Ascii("recrutiing"),
                &["recruiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintiainer"),
                &["maintainer"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrewebs"), &["interwebs"]),
            (
                dictgen::InsensitiveStr::Ascii("inspeciton"),
                &["inspections", "inspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depressief"),
                &["depressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monagomous"),
                &["monogamous"],
            ),
            (dictgen::InsensitiveStr::Ascii("pruchase"), &["purchase"]),
            (dictgen::InsensitiveStr::Ascii("prametrs"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("promts"), &["prompts"]),
            (dictgen::InsensitiveStr::Ascii("unworty"), &["unworthy"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrifises"),
                &["sacrifices"],
            ),
            (dictgen::InsensitiveStr::Ascii("documants"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("somwbody"), &["somebody"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrias"),
                &["industries"],
            ),
            (dictgen::InsensitiveStr::Ascii("beuty"), &["beauty"]),
            (dictgen::InsensitiveStr::Ascii("attackes"), &["attackers"]),
            (dictgen::InsensitiveStr::Ascii("requed"), &["requeued"]),
            (dictgen::InsensitiveStr::Ascii("trubines"), &["turbines"]),
            (
                dictgen::InsensitiveStr::Ascii("redistributivo"),
                &["redistribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("democrasies"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscommuniation"),
                &["miscommunication"],
            ),
            (dictgen::InsensitiveStr::Ascii("avacodos"), &["avocados"]),
            (dictgen::InsensitiveStr::Ascii("aggreeing"), &["agreeing"]),
            (dictgen::InsensitiveStr::Ascii("pulished"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("uperclass"), &["upperclass"]),
            (dictgen::InsensitiveStr::Ascii("newtwork"), &["network"]),
            (dictgen::InsensitiveStr::Ascii("feromone"), &["pheromone"]),
            (
                dictgen::InsensitiveStr::Ascii("nieghborhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articluating"),
                &["articulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("listerner"), &["listener"]),
            (
                dictgen::InsensitiveStr::Ascii("sexualizied"),
                &["sexualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("utlizes"), &["utilizes"]),
            (
                dictgen::InsensitiveStr::Ascii("advertiements"),
                &["advertisements"],
            ),
            (dictgen::InsensitiveStr::Ascii("suplant"), &["supplant"]),
            (
                dictgen::InsensitiveStr::Ascii("explicite"),
                &["explicit", "explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartive"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavioal"), &["behavioral"]),
            (dictgen::InsensitiveStr::Ascii("acceps"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("requirig"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("recoreds"), &["records"]),
            (dictgen::InsensitiveStr::Ascii("telegrapgh"), &["telegraph"]),
            (dictgen::InsensitiveStr::Ascii("handcuffes"), &["handcuffs"]),
            (
                dictgen::InsensitiveStr::Ascii("geneticlly"),
                &["genetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("serailize"), &["serialize"]),
            (dictgen::InsensitiveStr::Ascii("overyhped"), &["overhyped"]),
            (dictgen::InsensitiveStr::Ascii("redandant"), &["redundant"]),
            (dictgen::InsensitiveStr::Ascii("overdrev"), &["overdrive"]),
            (dictgen::InsensitiveStr::Ascii("orderd"), &["ordered"]),
            (dictgen::InsensitiveStr::Ascii("hostame"), &["hostname"]),
            (
                dictgen::InsensitiveStr::Ascii("administratiei"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("medicince"), &["medicine"]),
            (dictgen::InsensitiveStr::Ascii("diagramm"), &["diagram"]),
            (
                dictgen::InsensitiveStr::Ascii("prematurly"),
                &["prematurely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entrepreneus"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificatios"),
                &["notification", "notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitvely"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsrteam"), &["upstream"]),
            (dictgen::InsensitiveStr::Ascii("dosuments"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("soflty"), &["softly"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaiing"),
                &["reallocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrepencies"),
                &["discrepancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("deocded"), &["decoded"]),
            (
                dictgen::InsensitiveStr::Ascii("reconeected"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("popularily"),
                &["popularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institucionalized"),
                &["institutionalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("poket"), &["pocket"]),
            (
                dictgen::InsensitiveStr::Ascii("functinally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkspace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("enabe"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("devcie"), &["device"]),
            (dictgen::InsensitiveStr::Ascii("examlpes"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("comapny"), &["company"]),
            (dictgen::InsensitiveStr::Ascii("moderatey"), &["moderately"]),
            (
                dictgen::InsensitiveStr::Ascii("comperative"),
                &["comparative"],
            ),
            (dictgen::InsensitiveStr::Ascii("priases"), &["praises"]),
            (
                dictgen::InsensitiveStr::Ascii("normallising"),
                &["normalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("privalege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("omajes"), &["homages"]),
            (
                dictgen::InsensitiveStr::Ascii("mississipppi"),
                &["mississippi"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carier"),
                &["carrier", "career"],
            ),
            (dictgen::InsensitiveStr::Ascii("meditatie"), &["meditate"]),
            (dictgen::InsensitiveStr::Ascii("searhed"), &["searched"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("assmuming"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("pacakages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("uninteresed"),
                &["uninterested"],
            ),
            (dictgen::InsensitiveStr::Ascii("touchda"), &["touchpad"]),
            (dictgen::InsensitiveStr::Ascii("additiions"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("lanucher"), &["launcher"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessefully"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("specfiic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("defulting"), &["defaulting"]),
            (dictgen::InsensitiveStr::Ascii("mistread"), &["mistreated"]),
            (dictgen::InsensitiveStr::Ascii("commandpod"), &["commando"]),
            (dictgen::InsensitiveStr::Ascii("glyphes"), &["glyphs"]),
            (
                dictgen::InsensitiveStr::Ascii("magnifient"),
                &["magnificent"],
            ),
            (dictgen::InsensitiveStr::Ascii("requseted"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("airzona"), &["arizona"]),
            (dictgen::InsensitiveStr::Ascii("modfides"), &["modifies"]),
            (dictgen::InsensitiveStr::Ascii("mindcarck"), &["mindcrack"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappears"),
                &["disappears"],
            ),
            (dictgen::InsensitiveStr::Ascii("dessigners"), &["designers"]),
            (dictgen::InsensitiveStr::Ascii("dividened"), &["dividend"]),
            (
                dictgen::InsensitiveStr::Ascii("enumeratable"),
                &["enumerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("xonstruction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoincremet"),
                &["autoincrement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("methematician"),
                &["mathematician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumptous"),
                &["presumptuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("releife"), &["relief"]),
            (dictgen::InsensitiveStr::Ascii("hiygeine"), &["hygiene"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabillity"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbirary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("reaquests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("efficeint"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("predding"), &["pressing"]),
            (
                dictgen::InsensitiveStr::Ascii("targettting"),
                &["targeting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implimentaion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissadvantage"),
                &["disadvantage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curvatrues"),
                &["curvatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconeect"),
                &["disconnect"],
            ),
            (dictgen::InsensitiveStr::Ascii("inifitely"), &["infinitely"]),
            (
                dictgen::InsensitiveStr::Ascii("buros"),
                &["bureaus", "burros"],
            ),
            (dictgen::InsensitiveStr::Ascii("wayferer"), &["wayfarer"]),
            (
                dictgen::InsensitiveStr::Ascii("dictonaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("singuar"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("listenres"), &["listeners"]),
            (
                dictgen::InsensitiveStr::Ascii("flacourings"),
                &["flavourings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origninated"),
                &["originated"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjacancy"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("ythe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("extrordinarily"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("evryone"), &["everyone"]),
            (dictgen::InsensitiveStr::Ascii("imporot"), &["import"]),
            (dictgen::InsensitiveStr::Ascii("shledon"), &["sheldon"]),
            (dictgen::InsensitiveStr::Ascii("multcast"), &["multicast"]),
            (
                dictgen::InsensitiveStr::Ascii("glorifindel"),
                &["glorified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troups"),
                &["troupes", "troops"],
            ),
            (dictgen::InsensitiveStr::Ascii("datesets"), &["datasets"]),
            (dictgen::InsensitiveStr::Ascii("withrawal"), &["withdrawal"]),
            (dictgen::InsensitiveStr::Ascii("existung"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("indepentendly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("misaken"), &["mistaken"]),
            (
                dictgen::InsensitiveStr::Ascii("grievences"),
                &["grievances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundementally"),
                &["fundamentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("nornal"), &["normal"]),
            (
                dictgen::InsensitiveStr::Ascii("suffiently"),
                &["sufficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dael"),
                &["deal", "dial", "dahl"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cincinnasti"),
                &["cincinnati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priortizing"),
                &["prioritizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegetarin"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delibaretely"),
                &["deliberately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioting"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresser"),
                &["compressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conflifting"),
                &["conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("linceses"), &["licenses"]),
            (
                dictgen::InsensitiveStr::Ascii("spefifiable"),
                &["specifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("excersized"), &["exercised"]),
            (dictgen::InsensitiveStr::Ascii("lifeycle"), &["lifecycle"]),
            (dictgen::InsensitiveStr::Ascii("quailty"), &["quality"]),
            (dictgen::InsensitiveStr::Ascii("chtulhu"), &["cthulhu"]),
            (
                dictgen::InsensitiveStr::Ascii("inituition"),
                &["initiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("liases"), &["liaises"]),
            (dictgen::InsensitiveStr::Ascii("addjusts"), &["adjusts"]),
            (
                dictgen::InsensitiveStr::Ascii("contrbutors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknwoing"), &["unknowing"]),
            (dictgen::InsensitiveStr::Ascii("adaptes"), &["adapters"]),
            (
                dictgen::InsensitiveStr::Ascii("recalulation"),
                &["recalculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expendatures"),
                &["expenditures"],
            ),
            (dictgen::InsensitiveStr::Ascii("miltiple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("mnethods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("logially"), &["logically"]),
            (
                dictgen::InsensitiveStr::Ascii("educationnal"),
                &["educational"],
            ),
            (dictgen::InsensitiveStr::Ascii("addedd"), &["added"]),
            (dictgen::InsensitiveStr::Ascii("veryifies"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("regenaration"),
                &["regeneration"],
            ),
            (dictgen::InsensitiveStr::Ascii("stlakers"), &["stalkers"]),
            (dictgen::InsensitiveStr::Ascii("driagrams"), &["diagrams"]),
            (
                dictgen::InsensitiveStr::Ascii("delagating"),
                &["delegating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desginated"),
                &["designated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reniforcements"),
                &["reinforcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hankerchifs"),
                &["handkerchiefs"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraged"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("beleaver"), &["believer"]),
            (
                dictgen::InsensitiveStr::Ascii("midfieldes"),
                &["midfielders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pshycologist"),
                &["psychologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanitory"),
                &["explanatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("pensle"), &["pencil"]),
            (dictgen::InsensitiveStr::Ascii("betweel"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("markede"), &["marketed"]),
            (dictgen::InsensitiveStr::Ascii("fabrices"), &["fabrics"]),
            (dictgen::InsensitiveStr::Ascii("evrithing"), &["everything"]),
            (
                dictgen::InsensitiveStr::Ascii("exapnsions"),
                &["expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degragation"),
                &["degradation"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsi"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("forunner"), &["forerunner"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendents"),
                &["dependents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pshycopath"),
                &["psychopath"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archicturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("notning"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("avnegers"), &["avengers"]),
            (
                dictgen::InsensitiveStr::Ascii("noticication"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unkowningly"),
                &["unknowingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourgood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("durabiltiy"),
                &["durability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mezmorized"),
                &["mesmerized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infinately"),
                &["infinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmets"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("callig"), &["calling"]),
            (dictgen::InsensitiveStr::Ascii("agrresive"), &["aggressive"]),
            (
                dictgen::InsensitiveStr::Ascii("evangelia"),
                &["evangelical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootraping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("formulat"), &["formula"]),
            (
                dictgen::InsensitiveStr::Ascii("intitiators"),
                &["initiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("cursosr"), &["cursors"]),
            (
                dictgen::InsensitiveStr::Ascii("limitiations"),
                &["limitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponentiella"),
                &["exponential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ostensably"),
                &["ostensibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("isue"), &["issue"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstanta"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("californai"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cellebration"),
                &["celebration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutliplier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("firmwar"), &["firmware"]),
            (
                dictgen::InsensitiveStr::Ascii("conflicing"),
                &["conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("protess"), &["protests"]),
            (
                dictgen::InsensitiveStr::Ascii("coreespond"),
                &["correspond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delibarate"),
                &["deliberate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfitiers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasengers"), &["passengers"]),
            (dictgen::InsensitiveStr::Ascii("mailling"), &["mailing"]),
            (
                dictgen::InsensitiveStr::Ascii("dinamicaly"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("unassings"), &["unassigns"]),
            (dictgen::InsensitiveStr::Ascii("octobor"), &["october"]),
            (
                dictgen::InsensitiveStr::Ascii("autheritatively"),
                &["authoritatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficillies"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrptions"),
                &["encryptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tirangles"), &["triangles"]),
            (
                dictgen::InsensitiveStr::Ascii("replaceemnt"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportae"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accerelate"),
                &["accelerated", "accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generalizng"),
                &["generalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("streem"), &["stream"]),
            (dictgen::InsensitiveStr::Ascii("sabatage"), &["sabotage"]),
            (
                dictgen::InsensitiveStr::Ascii("irrationallity"),
                &["irrationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryavon"), &["trayvon"]),
            (dictgen::InsensitiveStr::Ascii("aleady"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("rthe"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("throaths"), &["throats"]),
            (
                dictgen::InsensitiveStr::Ascii("cdecompress"),
                &["decompress"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minoroties"),
                &["minorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("readiable"), &["readable"]),
            (dictgen::InsensitiveStr::Ascii("registory"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("assgins"), &["assigns"]),
            (
                dictgen::InsensitiveStr::Ascii("terminaisons"),
                &["terminations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoted"),
                &["reported", "reposted", "reputed", "repeated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconceved"),
                &["preconceived"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterpeace"),
                &["masterpiece"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exeriences"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("priror"), &["prior"]),
            (dictgen::InsensitiveStr::Ascii("augmneter"), &["augmenter"]),
            (
                dictgen::InsensitiveStr::Ascii("unappealling"),
                &["unappealing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serializatioin"),
                &["serialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patrent"),
                &["parent", "patent", "patron"],
            ),
            (dictgen::InsensitiveStr::Ascii("packtes"), &["packets"]),
            (
                dictgen::InsensitiveStr::Ascii("experementers"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("conneyct"), &["connect"]),
            (dictgen::InsensitiveStr::Ascii("sutisfied"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("relising"), &["realising"]),
            (
                dictgen::InsensitiveStr::Ascii("vaidity"),
                &["validity", "vapidity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calncellation"),
                &["cancellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("euthanesia"),
                &["euthanasia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulithreaded"),
                &["multithreaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decendend"),
                &["descendent", "descended"],
            ),
            (dictgen::InsensitiveStr::Ascii("strugle"), &["struggle"]),
            (
                dictgen::InsensitiveStr::Ascii("ourselve"),
                &["ourself", "ourselves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarunteing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("whaat"), &["what"]),
            (
                dictgen::InsensitiveStr::Ascii("asychronously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assymmetrical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinstatiation"),
                &["reinstantiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("siute"), &["suite"]),
            (
                dictgen::InsensitiveStr::Ascii("consumsion"),
                &["consumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colapsible"),
                &["collapsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ganerators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("eachotehr"), &["eachother"]),
            (dictgen::InsensitiveStr::Ascii("develpoed"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("definations"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetative"),
                &["repetitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("inports"), &["imports"]),
            (
                dictgen::InsensitiveStr::Ascii("riminicing"),
                &["reminiscing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investegators"),
                &["investigators"],
            ),
            (dictgen::InsensitiveStr::Ascii("intriuged"), &["intrigued"]),
            (dictgen::InsensitiveStr::Ascii("helpder"), &["helper"]),
            (dictgen::InsensitiveStr::Ascii("usuable"), &["usable"]),
            (
                dictgen::InsensitiveStr::Ascii("generaions"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("kocalized"), &["localized"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprihensible"),
                &["incomprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("indecated"), &["indicated"]),
            (dictgen::InsensitiveStr::Ascii("detahced"), &["detached"]),
            (dictgen::InsensitiveStr::Ascii("sript"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("nighlty"), &["nightly"]),
            (dictgen::InsensitiveStr::Ascii("dpeloy"), &["deploy"]),
            (
                dictgen::InsensitiveStr::Ascii("manetained"),
                &["maintained"],
            ),
            (dictgen::InsensitiveStr::Ascii("kepet"), &["kept"]),
            (dictgen::InsensitiveStr::Ascii("originl"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("timmestamps"),
                &["timestamps"],
            ),
            (dictgen::InsensitiveStr::Ascii("unich"), &["unix"]),
            (dictgen::InsensitiveStr::Ascii("extremem"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("cockatils"), &["cocktails"]),
            (dictgen::InsensitiveStr::Ascii("argumenr"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("combinate"),
                &["combine", "combination", "combined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("migrateable"),
                &["migratable"],
            ),
            (dictgen::InsensitiveStr::Ascii("spehres"), &["spheres"]),
            (
                dictgen::InsensitiveStr::Ascii("perperties"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("templarius"), &["templars"]),
            (
                dictgen::InsensitiveStr::Ascii("fignernails"),
                &["fingernails"],
            ),
            (dictgen::InsensitiveStr::Ascii("offereings"), &["offerings"]),
            (dictgen::InsensitiveStr::Ascii("publick"), &["public"]),
            (
                dictgen::InsensitiveStr::Ascii("pressiously"),
                &["preciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("freqeuncies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("amublance"), &["ambulance"]),
            (
                dictgen::InsensitiveStr::Ascii("avary"),
                &["every", "aviary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synoynmous"),
                &["synonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlusive"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("accomdation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraemters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("malpractce"),
                &["malpractice"],
            ),
            (dictgen::InsensitiveStr::Ascii("rythem"), &["rhythm"]),
            (dictgen::InsensitiveStr::Ascii("cooldows"), &["cooldowns"]),
            (dictgen::InsensitiveStr::Ascii("snippests"), &["snippets"]),
            (dictgen::InsensitiveStr::Ascii("polotic"), &["politic"]),
            (
                dictgen::InsensitiveStr::Ascii("flawlessley"),
                &["flawlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("errecting"), &["erecting"]),
            (dictgen::InsensitiveStr::Ascii("eveing"), &["evening"]),
            (
                dictgen::InsensitiveStr::Ascii("outragously"),
                &["outrageously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaginery"),
                &["imaginary", "imagery"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjusments"),
                &["adjustments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isotrophically"),
                &["isotropically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identidying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("badnwidth"), &["bandwidth"]),
            (dictgen::InsensitiveStr::Ascii("updationg"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("converstaion"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comminucating"),
                &["communicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwwarded"), &["forwarded"]),
            (
                dictgen::InsensitiveStr::Ascii("owervrites"),
                &["overwrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photogrpahers"),
                &["photographers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missionairy"),
                &["missionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("mariyln"), &["marilyn"]),
            (
                dictgen::InsensitiveStr::Ascii("initiavite"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumising"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("tennesee"), &["tennessee"]),
            (dictgen::InsensitiveStr::Ascii("memmory"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("warnng"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("monteral"), &["montreal"]),
            (
                dictgen::InsensitiveStr::Ascii("intializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("compilters"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapointing"),
                &["disappointing"],
            ),
            (dictgen::InsensitiveStr::Ascii("decission"), &["decision"]),
            (
                dictgen::InsensitiveStr::Ascii("negotionator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procederal"),
                &["procedural"],
            ),
            (dictgen::InsensitiveStr::Ascii("retourned"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("consuptions"),
                &["consumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberbilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("lesiban"), &["lesbian"]),
            (dictgen::InsensitiveStr::Ascii("shdaer"), &["shader"]),
            (dictgen::InsensitiveStr::Ascii("sytem"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("sigaled"), &["signaled"]),
            (
                dictgen::InsensitiveStr::Ascii("tranmitter"),
                &["transmitter"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentense"), &["sentence"]),
            (dictgen::InsensitiveStr::Ascii("suceeds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("degeneret"), &["degenerate"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhoof"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centenal"),
                &["centennial", "sentinel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condfigures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shuit"),
                &["shut", "suit", "shit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compulsery"),
                &["compulsory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("personalitie"),
                &["personality"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbot"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("checkear"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("projectin"), &["projection"]),
            (dictgen::InsensitiveStr::Ascii("manigests"), &["manifests"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacturas"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tupparware"),
                &["tupperware"],
            ),
            (dictgen::InsensitiveStr::Ascii("vareity"), &["variety"]),
            (dictgen::InsensitiveStr::Ascii("ignorat"), &["ignorant"]),
            (dictgen::InsensitiveStr::Ascii("conctats"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("initilizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremantal"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mofification"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecttors"),
                &["connectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("garabage"), &["garbage"]),
            (dictgen::InsensitiveStr::Ascii("rgister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnositic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingnores"), &["ignores"]),
            (
                dictgen::InsensitiveStr::Ascii("scren"),
                &["screen", "screw"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momementary"),
                &["momentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("happned"), &["happened"]),
            (
                dictgen::InsensitiveStr::Ascii("includd"),
                &["included", "include"],
            ),
            (dictgen::InsensitiveStr::Ascii("indisious"), &["insidious"]),
            (
                dictgen::InsensitiveStr::Ascii("proporition"),
                &["proposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstante"), &["bastante"]),
            (
                dictgen::InsensitiveStr::Ascii("indianopolis"),
                &["indianapolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("fewd"), &["few", "feud"]),
            (
                dictgen::InsensitiveStr::Ascii("invisbility"),
                &["invisibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("recuiting"), &["recruiting"]),
            (
                dictgen::InsensitiveStr::Ascii("opportuity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectivas"),
                &["objectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("youforic"), &["euphoric"]),
            (
                dictgen::InsensitiveStr::Ascii("authoritatian"),
                &["authoritarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implamented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("minmal"), &["minimal"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminaton"),
                &["contamination"],
            ),
            (dictgen::InsensitiveStr::Ascii("liscenses"), &["licenses"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubting"),
                &["distributing", "disrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("excentric"), &["eccentric"]),
            (dictgen::InsensitiveStr::Ascii("incorect"), &["incorrect"]),
            (
                dictgen::InsensitiveStr::Ascii("parallized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beneficiul"),
                &["beneficial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonization"),
                &["synchronization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schenanigans"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("levleling"), &["leveling"]),
            (dictgen::InsensitiveStr::Ascii("mial"), &["mail"]),
            (
                dictgen::InsensitiveStr::Ascii("transcocder"),
                &["transcoder"],
            ),
            (dictgen::InsensitiveStr::Ascii("narcissm"), &["narcissism"]),
            (dictgen::InsensitiveStr::Ascii("offine"), &["offline"]),
            (
                dictgen::InsensitiveStr::Ascii("chacacters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("progated"), &["propagated"]),
            (dictgen::InsensitiveStr::Ascii("returrn"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("resturants"),
                &["restaurants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportes"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("importatnt"), &["important"]),
            (dictgen::InsensitiveStr::Ascii("densly"), &["densely"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timestaped"),
                &["timestamped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepndence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ackumulates"),
                &["accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("margings"), &["margins"]),
            (dictgen::InsensitiveStr::Ascii("includea"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("comflicting"),
                &["conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighor"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("agragates"), &["aggregates"]),
            (
                dictgen::InsensitiveStr::Ascii("measurmenets"),
                &["measurements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliberatley"),
                &["deliberately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anectodally"),
                &["anecdotally"],
            ),
            (dictgen::InsensitiveStr::Ascii("quoshant"), &["quotient"]),
            (dictgen::InsensitiveStr::Ascii("acocunt"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("progessed"), &["progressed"]),
            (
                dictgen::InsensitiveStr::Ascii("miraculaous"),
                &["miraculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahchedb"), &["cachedb"]),
            (
                dictgen::InsensitiveStr::Ascii("illegitemate"),
                &["illegitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tresholds"), &["thresholds"]),
            (
                dictgen::InsensitiveStr::Ascii("millioniare"),
                &["millionaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectivly"),
                &["objectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasmited"),
                &["transmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("companians"),
                &["companions"],
            ),
            (dictgen::InsensitiveStr::Ascii("instealled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("coutn"), &["count"]),
            (
                dictgen::InsensitiveStr::Ascii("verificiation"),
                &["verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unversality"),
                &["universality"],
            ),
            (dictgen::InsensitiveStr::Ascii("shleters"), &["shelters"]),
            (
                dictgen::InsensitiveStr::Ascii("unsunscribe"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogenist"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ussers"),
                &["users", "ushers"],
            ),
            (dictgen::InsensitiveStr::Ascii("puplished"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("mantained"), &["maintained"]),
            (dictgen::InsensitiveStr::Ascii("darkenss"), &["darkness"]),
            (dictgen::InsensitiveStr::Ascii("prceeded"), &["preceded"]),
            (
                dictgen::InsensitiveStr::Ascii("concidering"),
                &["considering"],
            ),
            (dictgen::InsensitiveStr::Ascii("expemplary"), &["exemplary"]),
            (dictgen::InsensitiveStr::Ascii("keychian"), &["keychain"]),
            (
                dictgen::InsensitiveStr::Ascii("emperially"),
                &["imperially"],
            ),
            (dictgen::InsensitiveStr::Ascii("braodly"), &["broadly"]),
            (dictgen::InsensitiveStr::Ascii("impulsemos"), &["impulses"]),
            (dictgen::InsensitiveStr::Ascii("expoent"), &["exponent"]),
            (dictgen::InsensitiveStr::Ascii("extrating"), &["extracting"]),
            (
                dictgen::InsensitiveStr::Ascii("realisticlly"),
                &["realistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wathc"), &["watch"]),
            (
                dictgen::InsensitiveStr::Ascii("occasionaly"),
                &["occasionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catestrophic"),
                &["catastrophic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derviatives"),
                &["derivatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporories"),
                &["temporaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("requierd"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("compeare"), &["compare"]),
            (dictgen::InsensitiveStr::Ascii("accoutant"), &["accountant"]),
            (dictgen::InsensitiveStr::Ascii("itegrate"), &["integrate"]),
            (
                dictgen::InsensitiveStr::Ascii("dependencys"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultanaeously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("involes"), &["involves"]),
            (
                dictgen::InsensitiveStr::Ascii("algorthyms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquanitance"),
                &["acquaintance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumputous"),
                &["presumptuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("jsut"), &["just"]),
            (
                dictgen::InsensitiveStr::Ascii("expermimenters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidencey"),
                &["presidency"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmend"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("atract"), &["attract"]),
            (dictgen::InsensitiveStr::Ascii("procesor"), &["processor"]),
            (dictgen::InsensitiveStr::Ascii("kncokback"), &["knockback"]),
            (dictgen::InsensitiveStr::Ascii("functioin"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("whn"), &["when"]),
            (
                dictgen::InsensitiveStr::Ascii("finctionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delpoyment"),
                &["deployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qick"),
                &["quick", "pick", "tick", "kick", "wick", "sick"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demphasize"),
                &["deemphasize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coexhistence"),
                &["coexistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("doese"), &["does"]),
            (dictgen::InsensitiveStr::Ascii("downaload"), &["download"]),
            (dictgen::InsensitiveStr::Ascii("optins"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("socumentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conmand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("distingusihing"),
                &["distinguishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrutible"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependencias"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decomtamination"),
                &["decontamination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("departmens"),
                &["departments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scence"),
                &["scene", "science", "sense"],
            ),
            (dictgen::InsensitiveStr::Ascii("pattrns"), &["patterns"]),
            (dictgen::InsensitiveStr::Ascii("dimond"), &["diamond"]),
            (dictgen::InsensitiveStr::Ascii("assumin"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("unexecptedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("debloking"), &["deblocking"]),
            (
                dictgen::InsensitiveStr::Ascii("subscribirse"),
                &["subscriber"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structored"),
                &["structured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocorrent"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("erorneus"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("insitutional"),
                &["institutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immadiately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconencting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("constatnts"), &["constants"]),
            (
                dictgen::InsensitiveStr::Ascii("subconsciosly"),
                &["subconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("withou"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("refedendum"),
                &["referendum"],
            ),
            (dictgen::InsensitiveStr::Ascii("inot"), &["into"]),
            (dictgen::InsensitiveStr::Ascii("cardboad"), &["cardboard"]),
            (dictgen::InsensitiveStr::Ascii("compareing"), &["comparing"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcuated"),
                &["recalculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("akkumulating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparispon"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refirgerator"),
                &["refrigerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("callenges"), &["challenges"]),
            (
                dictgen::InsensitiveStr::Ascii("unexceptionnal"),
                &["unexceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complecations"),
                &["complications"],
            ),
            (dictgen::InsensitiveStr::Ascii("aadd"), &["add"]),
            (
                dictgen::InsensitiveStr::Ascii("steroetyping"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readmition"),
                &["readmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doocumentary"),
                &["documentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activtes"),
                &["activates", "activities"],
            ),
            (dictgen::InsensitiveStr::Ascii("satrudays"), &["saturdays"]),
            (
                dictgen::InsensitiveStr::Ascii("generalizaton"),
                &["generalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanbdled"), &["handled"]),
            (dictgen::InsensitiveStr::Ascii("improt"), &["import"]),
            (
                dictgen::InsensitiveStr::Ascii("repititive"),
                &["repetitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distroyers"),
                &["destroyers"],
            ),
            (dictgen::InsensitiveStr::Ascii("canye"), &["canaille"]),
            (
                dictgen::InsensitiveStr::Ascii("additionallly"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("doucmentes"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("accomidate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactherise"),
                &["characterise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presedential"),
                &["presidential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionaary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("feautres"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("campatibly"),
                &["compatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheet"),
                &["cheat", "sheet", "chest", "cheer", "cheek"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrappin"), &["wrapping"]),
            (dictgen::InsensitiveStr::Ascii("henderence"), &["hindrance"]),
            (dictgen::InsensitiveStr::Ascii("timestemp"), &["timestamp"]),
            (
                dictgen::InsensitiveStr::Ascii("accidenlty"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversimplyfication"),
                &["oversimplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("braisin"), &["braising"]),
            (
                dictgen::InsensitiveStr::Ascii("braket"),
                &["bracket", "brake"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inhernetly"),
                &["inherently"],
            ),
            (dictgen::InsensitiveStr::Ascii("agrrement"), &["agreement"]),
            (dictgen::InsensitiveStr::Ascii("alwo"), &["also"]),
            (
                dictgen::InsensitiveStr::Ascii("experimneted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("openend"), &["opened"]),
            (
                dictgen::InsensitiveStr::Ascii("anaytically"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("massivley"), &["massively"]),
            (dictgen::InsensitiveStr::Ascii("reliased"), &["realised"]),
            (dictgen::InsensitiveStr::Ascii("begginning"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("syraucse"), &["syracuse"]),
            (
                dictgen::InsensitiveStr::Ascii("enrollement"),
                &["enrollment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlastion"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("macth"),
                &["match", "math", "mach"],
            ),
            (dictgen::InsensitiveStr::Ascii("ailgn"), &["align"]),
            (
                dictgen::InsensitiveStr::Ascii("ontain"),
                &["contain", "obtain"],
            ),
            (dictgen::InsensitiveStr::Ascii("followig"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("evalueating"),
                &["evaluating"],
            ),
            (dictgen::InsensitiveStr::Ascii("efficcient"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("backwardss"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("abasin"), &["abasing"]),
            (
                dictgen::InsensitiveStr::Ascii("intrumenting"),
                &["instrumenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfroamtion"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylynder"), &["cylinder"]),
            (
                dictgen::InsensitiveStr::Ascii("triscadecafobia"),
                &["triskaidekaphobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bittersweat"),
                &["bittersweet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempoparily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incatations"),
                &["incantations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilzation"),
                &["stabilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("thirlling"), &["thrilling"]),
            (
                dictgen::InsensitiveStr::Ascii("combatible"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiially"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("internatioanlism"),
                &["internationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("undefinded"), &["undefined"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arametrical"),
                &["parametrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exparation"),
                &["expiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremmentally"),
                &["incrementally"],
            ),
            (dictgen::InsensitiveStr::Ascii("secpter"), &["scepter"]),
            (dictgen::InsensitiveStr::Ascii("airbrone"), &["airborne"]),
            (dictgen::InsensitiveStr::Ascii("fedorans"), &["fedoras"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapating"),
                &["dissipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("colgone"), &["cologne"]),
            (
                dictgen::InsensitiveStr::Ascii("increades"),
                &["increased", "increases"],
            ),
            (dictgen::InsensitiveStr::Ascii("pojecting"), &["projecting"]),
            (
                dictgen::InsensitiveStr::Ascii("mircotransactions"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("arugment"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("intantiate"),
                &["instantiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apoligising"),
                &["apologising"],
            ),
            (dictgen::InsensitiveStr::Ascii("makro"), &["macro"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegogtiations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicatbly"),
                &["predictably"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomatos"), &["tomatoes"]),
            (
                dictgen::InsensitiveStr::Ascii("pedestrains"),
                &["pedestrians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configrued"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("assemblys"), &["assemblies"]),
            (
                dictgen::InsensitiveStr::Ascii("ellipitcal"),
                &["elliptical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoistory"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("negoator"), &["negotiator"]),
            (dictgen::InsensitiveStr::Ascii("loder"), &["loader"]),
            (
                dictgen::InsensitiveStr::Ascii("normalizaiton"),
                &["normalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refreshener"),
                &["refresher"],
            ),
            (dictgen::InsensitiveStr::Ascii("locae"), &["locate"]),
            (
                dictgen::InsensitiveStr::Ascii("successfull"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifetiem"), &["lifetime"]),
            (
                dictgen::InsensitiveStr::Ascii("preemptable"),
                &["preemptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("atelast"), &["atleast"]),
            (
                dictgen::InsensitiveStr::Ascii("surporters"),
                &["supporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wildreness"),
                &["wilderness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("packging"),
                &["packaging", "packing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sometiem"),
                &["sometime", "sometimes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armenina"),
                &["armenian", "armenia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathalogists"),
                &["pathologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("contrsts"), &["contrasts"]),
            (dictgen::InsensitiveStr::Ascii("penguines"), &["penguins"]),
            (dictgen::InsensitiveStr::Ascii("unvisisted"), &["unvisited"]),
            (dictgen::InsensitiveStr::Ascii("warninf"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("authenciation"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("sescedes"), &["secedes"]),
            (
                dictgen::InsensitiveStr::Ascii("assymptotic"),
                &["asymptotic"],
            ),
            (dictgen::InsensitiveStr::Ascii("absodefly"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("disconeection"),
                &["disconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("swtiched"), &["switched"]),
            (
                dictgen::InsensitiveStr::Ascii("architechure"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconcious"),
                &["unconscious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decriptions"),
                &["descriptions", "decryptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("elephans"), &["elephants"]),
            (dictgen::InsensitiveStr::Ascii("gettitems"), &["getitems"]),
            (dictgen::InsensitiveStr::Ascii("insatlled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("subtelty"), &["subtlety"]),
            (dictgen::InsensitiveStr::Ascii("inbalance"), &["imbalance"]),
            (
                dictgen::InsensitiveStr::Ascii("initaitions"),
                &["initiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("finshing"), &["finishing"]),
            (
                dictgen::InsensitiveStr::Ascii("unoperational"),
                &["nonoperational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charaterizing"),
                &["characterizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("medevil"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("reponder"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("furher"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("rendereds"),
                &["rendered", "renders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominater"),
                &["denominator", "denominated", "denominates", "denominate"],
            ),
            (dictgen::InsensitiveStr::Ascii("roganism"), &["organism"]),
            (
                dictgen::InsensitiveStr::Ascii("calculattions"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desperating"),
                &["desperation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphapetical"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("appenines"), &["apennines"]),
            (dictgen::InsensitiveStr::Ascii("wnating"), &["wanting"]),
            (dictgen::InsensitiveStr::Ascii("arguemnet"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("interracting"),
                &["interacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("menthod"),
                &["method", "menthol"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conparisons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntacticly"),
                &["syntactically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ivalidate"), &["invalidate"]),
            (dictgen::InsensitiveStr::Ascii("specfiics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("subfeatutures"),
                &["subfeatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undicovered"),
                &["undiscovered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearraning"),
                &["rearranging"],
            ),
            (dictgen::InsensitiveStr::Ascii("letmost"), &["leftmost"]),
            (dictgen::InsensitiveStr::Ascii("retalaite"), &["retaliate"]),
            (dictgen::InsensitiveStr::Ascii("follwoeing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("sprintas"), &["sprints"]),
            (dictgen::InsensitiveStr::Ascii("endianity"), &["endianness"]),
            (dictgen::InsensitiveStr::Ascii("eithe"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("authentificates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributers"),
                &["distributors", "distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("riendeers"), &["reindeers"]),
            (
                dictgen::InsensitiveStr::Ascii("therem"),
                &["there", "theorem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spoffing"),
                &["spoofing", "spaffing", "scoffing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experssion"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periodicy"),
                &["periodicity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsesfull"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("pickyune"), &["picayune"]),
            (
                dictgen::InsensitiveStr::Ascii("confernece"),
                &["conferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microwavees"),
                &["microwaves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("widhtpoints"),
                &["widthpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aformentioned"),
                &["aforementioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissmantles"),
                &["dismantles"],
            ),
            (dictgen::InsensitiveStr::Ascii("orientned"), &["oriented"]),
            (dictgen::InsensitiveStr::Ascii("kronicler"), &["chronicler"]),
            (dictgen::InsensitiveStr::Ascii("unusre"), &["unsure"]),
            (dictgen::InsensitiveStr::Ascii("regardign"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("variationnal"),
                &["variational"],
            ),
            (dictgen::InsensitiveStr::Ascii("hipocrit"), &["hypocrite"]),
            (
                dictgen::InsensitiveStr::Ascii("compositoin"),
                &["compositions"],
            ),
            (dictgen::InsensitiveStr::Ascii("vartical"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("termporaries"),
                &["temporaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquainted"), &["acquainted"]),
            (
                dictgen::InsensitiveStr::Ascii("idependence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("calclated"), &["calculated"]),
            (dictgen::InsensitiveStr::Ascii("hardwiered"), &["hardwired"]),
            (
                dictgen::InsensitiveStr::Ascii("deamiguates"),
                &["disambiguates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pascheurizes"),
                &["pasteurizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("palyer"), &["player"]),
            (
                dictgen::InsensitiveStr::Ascii("populatiry"),
                &["popularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("caendar"), &["calendar"]),
            (
                dictgen::InsensitiveStr::Ascii("californien"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alse"),
                &["else", "also", "false"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectificaton"),
                &["objectification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diferentiation"),
                &["differentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("remvoes"), &["removes"]),
            (
                dictgen::InsensitiveStr::Ascii("absoulutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consueling"),
                &["consulting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cognatious"),
                &["contagious"],
            ),
            (dictgen::InsensitiveStr::Ascii("scanerios"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("arcylic"), &["acrylic"]),
            (
                dictgen::InsensitiveStr::Ascii("backgounds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("yur"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalty"),
                &["nationality"],
            ),
            (dictgen::InsensitiveStr::Ascii("crowkay"), &["croquet"]),
            (dictgen::InsensitiveStr::Ascii("idividual"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("lcoals"), &["locals"]),
            (dictgen::InsensitiveStr::Ascii("pumpknis"), &["pumpkins"]),
            (
                dictgen::InsensitiveStr::Ascii("perposefully"),
                &["purposefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("icreasing"), &["increasing"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiamtions"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("versoions"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("contraveining"),
                &["contravening"],
            ),
            (dictgen::InsensitiveStr::Ascii("supspected"), &["suspected"]),
            (
                dictgen::InsensitiveStr::Ascii("beautifyl"),
                &["beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("culmulated"),
                &["cumulated", "culminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("headlesss"), &["headless"]),
            (dictgen::InsensitiveStr::Ascii("accpt"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("ramdomisation"),
                &["randomisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("redemeed"), &["redeemed"]),
            (dictgen::InsensitiveStr::Ascii("roachers"), &["roaches"]),
            (
                dictgen::InsensitiveStr::Ascii("appicability"),
                &["applicability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersecing"),
                &["intersecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("quizzs"), &["quizzes"]),
            (dictgen::InsensitiveStr::Ascii("classfier"), &["classifier"]),
            (dictgen::InsensitiveStr::Ascii("nameing"), &["naming"]),
            (
                dictgen::InsensitiveStr::Ascii("empiricaly"),
                &["empirically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parragraphs"),
                &["paragraphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grammartical"),
                &["grammatical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferefernce"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("appathetic"), &["apathetic"]),
            (
                dictgen::InsensitiveStr::Ascii("subsidizies"),
                &["subsidize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genererations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjacantly"),
                &["adjacently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subfolfers"),
                &["subfolders"],
            ),
            (dictgen::InsensitiveStr::Ascii("targt"), &["target"]),
            (dictgen::InsensitiveStr::Ascii("speficifcs"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmerical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antidepresssant"),
                &["antidepressant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cyrstallisation"),
                &["crystallisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gorillla"), &["gorilla"]),
            (dictgen::InsensitiveStr::Ascii("unflaged"), &["unflagged"]),
            (
                dictgen::InsensitiveStr::Ascii("retquirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommenation"),
                &["recommendation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempdate"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("feedbakc"), &["feedback"]),
            (
                dictgen::InsensitiveStr::Ascii("oppositition"),
                &["opposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("certin"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("cetified"), &["certified"]),
            (dictgen::InsensitiveStr::Ascii("ponit"), &["point"]),
            (dictgen::InsensitiveStr::Ascii("ellignton"), &["ellington"]),
            (dictgen::InsensitiveStr::Ascii("calalog"), &["catalog"]),
            (
                dictgen::InsensitiveStr::Ascii("regarldess"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innotation"),
                &["innovation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coinitailize"),
                &["coinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryying"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimlarlity"),
                &["dissimilarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("illegalis"), &["illegals"]),
            (
                dictgen::InsensitiveStr::Ascii("oging"),
                &["going", "ogling"],
            ),
            (dictgen::InsensitiveStr::Ascii("poitnes"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("unselct"), &["unselect"]),
            (dictgen::InsensitiveStr::Ascii("positionl"), &["positional"]),
            (dictgen::InsensitiveStr::Ascii("asshatts"), &["asshats"]),
            (dictgen::InsensitiveStr::Ascii("depent"), &["depend"]),
            (dictgen::InsensitiveStr::Ascii("tabels"), &["tables"]),
            (
                dictgen::InsensitiveStr::Ascii("alllocating"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applciations"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomrals"), &["normals"]),
            (
                dictgen::InsensitiveStr::Ascii("geogrpahic"),
                &["geographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("autoonf"), &["autoconf"]),
            (dictgen::InsensitiveStr::Ascii("suppply"), &["supply"]),
            (dictgen::InsensitiveStr::Ascii("riddlin"), &["riddling"]),
            (
                dictgen::InsensitiveStr::Ascii("paricipating"),
                &["participating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dowloading"),
                &["downloading"],
            ),
            (dictgen::InsensitiveStr::Ascii("buildes"), &["builders"]),
            (dictgen::InsensitiveStr::Ascii("witho"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("qutoes"), &["quotes"]),
            (dictgen::InsensitiveStr::Ascii("connecotr"), &["connector"]),
            (dictgen::InsensitiveStr::Ascii("eceonomy"), &["economy"]),
            (
                dictgen::InsensitiveStr::Ascii("deterministicly"),
                &["deterministically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developent"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheltics"), &["athletics"]),
            (dictgen::InsensitiveStr::Ascii("proclami"), &["proclaim"]),
            (dictgen::InsensitiveStr::Ascii("hrlps"), &["helps"]),
            (dictgen::InsensitiveStr::Ascii("podfie"), &["podfile"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmmetrical"),
                &["asymmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("horriblly"), &["horribly"]),
            (dictgen::InsensitiveStr::Ascii("catacylsm"), &["cataclysm"]),
            (dictgen::InsensitiveStr::Ascii("ignorming"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("scrpted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("hallunications"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tieing"), &["tying"]),
            (dictgen::InsensitiveStr::Ascii("methots"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("reprots"), &["reports"]),
            (dictgen::InsensitiveStr::Ascii("innovaties"), &["innovate"]),
            (dictgen::InsensitiveStr::Ascii("arsenaals"), &["arsenals"]),
            (dictgen::InsensitiveStr::Ascii("ontairo"), &["ontario"]),
            (dictgen::InsensitiveStr::Ascii("disancing"), &["distancing"]),
            (
                dictgen::InsensitiveStr::Ascii("paramterical"),
                &["parametric", "parametrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silhouweted"),
                &["silhouetted"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofcoure"), &["ofcourse"]),
            (
                dictgen::InsensitiveStr::Ascii("tournamets"),
                &["tournaments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimilarily"),
                &["dissimilarly"],
            ),
            (dictgen::InsensitiveStr::Ascii("llinear"), &["linear"]),
            (
                dictgen::InsensitiveStr::Ascii("connatation"),
                &["connotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconceieved"),
                &["preconceived"],
            ),
            (dictgen::InsensitiveStr::Ascii("lookes"), &["looks"]),
            (dictgen::InsensitiveStr::Ascii("faill"), &["fail"]),
            (
                dictgen::InsensitiveStr::Ascii("activateing"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirvayl"), &["surveil"]),
            (
                dictgen::InsensitiveStr::Ascii("teathers"),
                &[
                    "tethers", "teachers", "theaters", "feathers", "weathers", "leathers",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derectories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instututionalized"),
                &["institutionalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("develompental"),
                &["developmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destribution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("plasticosa"), &["plastics"]),
            (
                dictgen::InsensitiveStr::Ascii("resonsible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bitterwseet"),
                &["bittersweet"],
            ),
            (dictgen::InsensitiveStr::Ascii("commiest"), &["commits"]),
            (dictgen::InsensitiveStr::Ascii("impcating"), &["impacting"]),
            (dictgen::InsensitiveStr::Ascii("sperately"), &["separately"]),
            (dictgen::InsensitiveStr::Ascii("chckbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("ambadexterousness"),
                &["ambidextrousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("chipp"), &["chip"]),
            (
                dictgen::InsensitiveStr::Ascii("expirementing"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumernt"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnets"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signularity"),
                &["singularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offensivly"),
                &["offensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechnism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("slected"),
                &["selected", "elected"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirle"), &["circle"]),
            (
                dictgen::InsensitiveStr::Ascii("finanacially"),
                &["financially"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("negothiations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("creepgin"), &["creeping"]),
            (
                dictgen::InsensitiveStr::Ascii("directivos"),
                &["directions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnearbuilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sypmtomatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("porposed"), &["proposed"]),
            (
                dictgen::InsensitiveStr::Ascii("sentively"),
                &["sensitively", "sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lowerst"),
                &["lowest", "lowers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extarnally"),
                &["externally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speparation"),
                &["separation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caridovascular"),
                &["cardiovascular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upredictable"),
                &["unpredictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("thannked"), &["thanked"]),
            (dictgen::InsensitiveStr::Ascii("economisch"), &["economic"]),
            (dictgen::InsensitiveStr::Ascii("maitain"), &["maintain"]),
            (dictgen::InsensitiveStr::Ascii("carnomial"), &["ceremonial"]),
            (dictgen::InsensitiveStr::Ascii("synopsys"), &["synopsis"]),
            (
                dictgen::InsensitiveStr::Ascii("transseuxal"),
                &["transsexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irradating"),
                &["irradiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("refereced"), &["referenced"]),
            (dictgen::InsensitiveStr::Ascii("celestail"), &["celestial"]),
            (dictgen::InsensitiveStr::Ascii("graident"), &["gradient"]),
            (
                dictgen::InsensitiveStr::Ascii("improovments"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounray"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("releates"), &["relates"]),
            (dictgen::InsensitiveStr::Ascii("rathar"), &["rather"]),
            (dictgen::InsensitiveStr::Ascii("nneeded"), &["needed"]),
            (dictgen::InsensitiveStr::Ascii("errorneous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("antagonstic"),
                &["antagonist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ridicilous"),
                &["ridiculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("angshiuos"), &["anxious"]),
            (dictgen::InsensitiveStr::Ascii("claymoe"), &["claymore"]),
            (dictgen::InsensitiveStr::Ascii("committy"), &["committee"]),
            (
                dictgen::InsensitiveStr::Ascii("mindlessley"),
                &["mindlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncializes"),
                &["canonicalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("recusrsive"), &["recursive"]),
            (
                dictgen::InsensitiveStr::Ascii("exnternality"),
                &["externality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappointing"),
                &["disappointing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcendentational"),
                &["transcendental"],
            ),
            (dictgen::InsensitiveStr::Ascii("prividing"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("restauraunt"),
                &["restaurant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anniversairy"),
                &["anniversary"],
            ),
            (dictgen::InsensitiveStr::Ascii("issuses"), &["issues"]),
            (
                dictgen::InsensitiveStr::Ascii("contraversial"),
                &["controversial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corersponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phisically"),
                &["physically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propteries"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invesitgators"),
                &["investigators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acumulating"),
                &["accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("onbtain"), &["obtain"]),
            (dictgen::InsensitiveStr::Ascii("underlow"), &["underflow"]),
            (dictgen::InsensitiveStr::Ascii("opps"), &["oops"]),
            (
                dictgen::InsensitiveStr::Ascii("consequentually"),
                &["consequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowedges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prodced"),
                &["produced", "proceed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispossed"),
                &["disposed", "dispossessed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("framethrower"),
                &["flamethrower"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprion"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("microscoptic"),
                &["microscopic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracterise"),
                &["characterise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superviser"),
                &["supervisor"],
            ),
            (dictgen::InsensitiveStr::Ascii("manyally"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("extermal"),
                &["external", "extremal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumwent"),
                &["circumvent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubcription"),
                &["unsubscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bellweather"),
                &["bellwether"],
            ),
            (dictgen::InsensitiveStr::Ascii("prioirty"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("xepecting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("diferrence"),
                &["difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("higer"), &["higher"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowedgement"),
                &["acknowledgement"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarchit"), &["anarchist"]),
            (dictgen::InsensitiveStr::Ascii("initmacy"), &["intimacy"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditial"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("anounces"), &["announces"]),
            (
                dictgen::InsensitiveStr::Ascii("discconecct"),
                &["disconnect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaquirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribuors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("comomand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("chceklists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggegating"),
                &["aggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microprocesspr"),
                &["microprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorythemic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibiilities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("selectoin"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplementaion"),
                &["reimplementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dbeian"), &["debian"]),
            (dictgen::InsensitiveStr::Ascii("itereates"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("referrred"), &["referred"]),
            (dictgen::InsensitiveStr::Ascii("cllouds"), &["clouds"]),
            (
                dictgen::InsensitiveStr::Ascii("specificied"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstraties"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiators"),
                &["combinators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respectivley"),
                &["respectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sturturing"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("backwarks"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("abstrakts"), &["abstracts"]),
            (dictgen::InsensitiveStr::Ascii("insertin"), &["inserting"]),
            (dictgen::InsensitiveStr::Ascii("scannner"), &["scanner"]),
            (
                dictgen::InsensitiveStr::Ascii("attivating"),
                &["activating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememnted"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microvaxes"),
                &["microwaves"],
            ),
            (dictgen::InsensitiveStr::Ascii("yougest"), &["youngest"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanment"),
                &["enchantment"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecrts"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("identitifier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("illess"), &["illness"]),
            (dictgen::InsensitiveStr::Ascii("necesssity"), &["necessity"]),
            (
                dictgen::InsensitiveStr::Ascii("componenent"),
                &["component"],
            ),
            (dictgen::InsensitiveStr::Ascii("charactetr"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("provilege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("separatly"), &["separately"]),
            (
                dictgen::InsensitiveStr::Ascii("unoffocial"),
                &["unofficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("teamplated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("isssues"), &["issues"]),
            (
                dictgen::InsensitiveStr::Ascii("conributors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("condemmed"), &["condemned"]),
            (dictgen::InsensitiveStr::Ascii("majoroty"), &["majority"]),
            (dictgen::InsensitiveStr::Ascii("argumment"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("depents"),
                &["dependents", "depends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professinal"),
                &["professional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("hsve"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("expepected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("understads"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annonymouse"),
                &["anonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reoslutions"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chronicals"),
                &["chronicles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contridict"),
                &["contradict"],
            ),
            (dictgen::InsensitiveStr::Ascii("fotograph"), &["photograph"]),
            (dictgen::InsensitiveStr::Ascii("cacelled"), &["cancelled"]),
            (
                dictgen::InsensitiveStr::Ascii("specififying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professioal"),
                &["professional"],
            ),
            (dictgen::InsensitiveStr::Ascii("doucheus"), &["douches"]),
            (dictgen::InsensitiveStr::Ascii("shirely"), &["shirley"]),
            (
                dictgen::InsensitiveStr::Ascii("obtianable"),
                &["obtainable"],
            ),
            (dictgen::InsensitiveStr::Ascii("analize"), &["analyze"]),
            (dictgen::InsensitiveStr::Ascii("hsotname"), &["hostname"]),
            (
                dictgen::InsensitiveStr::Ascii("ellimination"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceutable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitverse"),
                &["multiverse"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaloable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("axixsymmetric"),
                &["axisymmetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decompositons"),
                &["decompositions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ojects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("reparamterizes"),
                &["reparameterizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleratio"),
                &["accelerator", "acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("primarially"),
                &["primarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("castleos"), &["castles"]),
            (dictgen::InsensitiveStr::Ascii("tehn"), &["then"]),
            (
                dictgen::InsensitiveStr::Ascii("operativos"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contagiosum"),
                &["contagious"],
            ),
            (dictgen::InsensitiveStr::Ascii("mccarhty"), &["mccarthy"]),
            (dictgen::InsensitiveStr::Ascii("intiial"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("evolutionarilly"),
                &["evolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("exntry"), &["entry"]),
            (dictgen::InsensitiveStr::Ascii("modfiders"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("coalaesced"), &["coalesced"]),
            (
                dictgen::InsensitiveStr::Ascii("shenenigans"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("estimage"), &["estimate"]),
            (
                dictgen::InsensitiveStr::Ascii("allegiancie"),
                &["allegiance"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicided"), &["decided"]),
            (
                dictgen::InsensitiveStr::Ascii("charascters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("propse"), &["propose"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnosics"),
                &["diagnostics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underestiamte"),
                &["underestimated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australianess"),
                &["australians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perordered"),
                &["preordered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reposiotories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centrilize"),
                &["centralize"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumne"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("wishpers"), &["whispers"]),
            (
                dictgen::InsensitiveStr::Ascii("interruting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimansions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosohpical"),
                &["philosophical"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdency"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("precicion"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("surpresses"),
                &["suppresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("disablede"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("expoert"),
                &["export", "expert"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paliamentarian"),
                &["parliamentarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("avearges"), &["averages"]),
            (
                dictgen::InsensitiveStr::Ascii("heartbrake"),
                &["heartbreak"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordinarilly"),
                &["extraordinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plase"),
                &["place", "please", "phase", "plaice"],
            ),
            (dictgen::InsensitiveStr::Ascii("captais"), &["captains"]),
            (dictgen::InsensitiveStr::Ascii("tengant"), &["tangent"]),
            (dictgen::InsensitiveStr::Ascii("realyl"), &["really"]),
            (
                dictgen::InsensitiveStr::Ascii("suspisious"),
                &["suspicious"],
            ),
            (dictgen::InsensitiveStr::Ascii("copyrigth"), &["copyright"]),
            (
                dictgen::InsensitiveStr::Ascii("utilizating"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsciption"),
                &["subscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminised"), &["reminisced"]),
            (dictgen::InsensitiveStr::Ascii("funcition"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("overwhlem"), &["overwhelm"]),
            (dictgen::InsensitiveStr::Ascii("arrivies"), &["arrives"]),
            (
                dictgen::InsensitiveStr::Ascii("healty"),
                &["healthy", "hearty"],
            ),
            (dictgen::InsensitiveStr::Ascii("kimerra"), &["chimera"]),
            (
                dictgen::InsensitiveStr::Ascii("sarcasitcally"),
                &["sarcastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wnt"), &["want"]),
            (
                dictgen::InsensitiveStr::Ascii("narcassism"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("writtable"), &["writable"]),
            (
                dictgen::InsensitiveStr::Ascii("reparing"),
                &["repairing", "reparsing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disribution"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deailing"),
                &["dealing", "detailing", "derailing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicslose"), &["disclose"]),
            (
                dictgen::InsensitiveStr::Ascii("cellabrates"),
                &["celebrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operaitonal"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annuciators"),
                &["annunciators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufaucturing"),
                &["manufacturing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devision"),
                &["division", "decision", "revision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attaindre"),
                &["attainder", "attained"],
            ),
            (dictgen::InsensitiveStr::Ascii("happaning"), &["happening"]),
            (dictgen::InsensitiveStr::Ascii("divinin"), &["divining"]),
            (dictgen::InsensitiveStr::Ascii("ouptuting"), &["outputting"]),
            (
                dictgen::InsensitiveStr::Ascii("recconeection"),
                &["reconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("betales"), &["beatles"]),
            (dictgen::InsensitiveStr::Ascii("complatly"), &["completely"]),
            (
                dictgen::InsensitiveStr::Ascii("agreemnets"),
                &["agreements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectutors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("illustraded"),
                &["illustrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("doublely"), &["doubly"]),
            (dictgen::InsensitiveStr::Ascii("forwading"), &["forwarding"]),
            (dictgen::InsensitiveStr::Ascii("extimate"), &["estimate"]),
            (
                dictgen::InsensitiveStr::Ascii("unvailable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custoniser"),
                &["customiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("reccurrent"), &["recurrent"]),
            (dictgen::InsensitiveStr::Ascii("raptores"), &["raptors"]),
            (
                dictgen::InsensitiveStr::Ascii("initliazations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associatio"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("trucated"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("leagelise"), &["legalise"]),
            (dictgen::InsensitiveStr::Ascii("spefeis"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("offens"),
                &["offend", "offends", "offense", "offers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psudoinverse"),
                &["pseudoinverse"],
            ),
            (dictgen::InsensitiveStr::Ascii("fotograf"), &["photograph"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbering"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("instint"), &["instinct"]),
            (dictgen::InsensitiveStr::Ascii("techniqe"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaitors"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("extractet"), &["extracted"]),
            (
                dictgen::InsensitiveStr::Ascii("execulusive"),
                &["exclusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimenary"),
                &["complimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pararmeters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribtuion"),
                &["distributions", "distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poistively"),
                &["positively"],
            ),
            (dictgen::InsensitiveStr::Ascii("pointseta"), &["poinsettia"]),
            (
                dictgen::InsensitiveStr::Ascii("reponsiveness"),
                &["responsiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commonhealth"),
                &["commonwealth"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplying"),
                &["simplifying", "implying", "simply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schedualing"),
                &["scheduling"],
            ),
            (dictgen::InsensitiveStr::Ascii("raplace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("illuninati"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("domianting"),
                &["dominating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wellingtion"),
                &["wellington"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("folling"),
                &["following", "falling", "rolling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconnets"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("listeing"), &["listening"]),
            (dictgen::InsensitiveStr::Ascii("paraniac"), &["paranoiac"]),
            (
                dictgen::InsensitiveStr::Ascii("correlatos"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("detialed"), &["detailed"]),
            (
                dictgen::InsensitiveStr::Ascii("simptumatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("esists"), &["exists"]),
            (dictgen::InsensitiveStr::Ascii("removeting"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("diappeared"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("teapsoon"), &["teaspoon"]),
            (
                dictgen::InsensitiveStr::Ascii("delimititations"),
                &["delimitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresssed"),
                &["compressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("charater"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("challeged"), &["challenged"]),
            (dictgen::InsensitiveStr::Ascii("famework"), &["framework"]),
            (dictgen::InsensitiveStr::Ascii("specivied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("objctions"), &["objections"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanenting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addtionaly"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("impropely"), &["improperly"]),
            (dictgen::InsensitiveStr::Ascii("tageted"), &["targeted"]),
            (
                dictgen::InsensitiveStr::Ascii("thinkg"),
                &["think", "thing", "things"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transormer"),
                &["transformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incohorent"),
                &["incoherent"],
            ),
            (dictgen::InsensitiveStr::Ascii("alsmost"), &["almost"]),
            (dictgen::InsensitiveStr::Ascii("sequense"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("apetite"), &["appetite"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomfortabel"),
                &["uncomfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatelyly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparititive"),
                &["comparative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heding"),
                &["heading", "heeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relocateble"),
                &["relocatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortan"), &["fortran"]),
            (dictgen::InsensitiveStr::Ascii("ontacts"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("canoicalizes"),
                &["canonicalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyjacker"), &["hijacker"]),
            (dictgen::InsensitiveStr::Ascii("benge"), &["binge"]),
            (
                dictgen::InsensitiveStr::Ascii("probabe"),
                &["probable", "probate"],
            ),
            (dictgen::InsensitiveStr::Ascii("usetnet"), &["usenet"]),
            (
                dictgen::InsensitiveStr::Ascii("cooedinated"),
                &["coordinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionability"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprectation"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("meldoic"), &["melodic"]),
            (dictgen::InsensitiveStr::Ascii("szie"), &["size"]),
            (
                dictgen::InsensitiveStr::Ascii("insallation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinte"),
                &["determine", "determined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substences"),
                &["substances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congifuration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guantamamo"),
                &["guantanamo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ony"),
                &["only", "on", "one"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deompression"),
                &["decompression"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomeer"), &["performer"]),
            (dictgen::InsensitiveStr::Ascii("sundey"), &["sunday"]),
            (dictgen::InsensitiveStr::Ascii("defaulh"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("accepte"), &["accept"]),
            (dictgen::InsensitiveStr::Ascii("teacers"), &["teachers"]),
            (dictgen::InsensitiveStr::Ascii("concepta"), &["conceptual"]),
            (
                dictgen::InsensitiveStr::Ascii("speficable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viatnamese"),
                &["vietnamese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intraverts"),
                &["introverts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchroously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentations"),
                &["concatenations"],
            ),
            (dictgen::InsensitiveStr::Ascii("beacaon"), &["beacon"]),
            (dictgen::InsensitiveStr::Ascii("steadilly"), &["steadily"]),
            (
                dictgen::InsensitiveStr::Ascii("perdicament"),
                &["predicament"],
            ),
            (dictgen::InsensitiveStr::Ascii("thankyoou"), &["thankyou"]),
            (dictgen::InsensitiveStr::Ascii("executuing"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("functoning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excectables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minimsation"),
                &["minimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessesery"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("grahpically"),
                &["graphically"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtarger"), &["subtarget"]),
            (
                dictgen::InsensitiveStr::Ascii("intstuction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coressponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprting"), &["exporting"]),
            (dictgen::InsensitiveStr::Ascii("wekly"), &["weekly"]),
            (dictgen::InsensitiveStr::Ascii("lanugage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("accends"),
                &["ascends", "accents"],
            ),
            (dictgen::InsensitiveStr::Ascii("whoes"), &["whose"]),
            (dictgen::InsensitiveStr::Ascii("requisit"), &["requisite"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchroneously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("rougly"), &["roughly"]),
            (dictgen::InsensitiveStr::Ascii("terminarla"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("elevateion"), &["elevation"]),
            (
                dictgen::InsensitiveStr::Ascii("introversa"),
                &["introverts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aletrnating"),
                &["alternating"],
            ),
            (dictgen::InsensitiveStr::Ascii("correctlly"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("passionetly"),
                &["passionately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millionairre"),
                &["millionaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reporisaties"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolorance"), &["tolerance"]),
            (
                dictgen::InsensitiveStr::Ascii("perputually"),
                &["perpetually"],
            ),
            (dictgen::InsensitiveStr::Ascii("madatory"), &["mandatory"]),
            (
                dictgen::InsensitiveStr::Ascii("pshyciatric"),
                &["psychiatric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irraditating"),
                &["irradiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectivos"),
                &["collections"],
            ),
            (dictgen::InsensitiveStr::Ascii("addiing"), &["adding"]),
            (dictgen::InsensitiveStr::Ascii("ccorrect"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("numnbering"), &["numbering"]),
            (dictgen::InsensitiveStr::Ascii("accepor"), &["acceptor"]),
            (
                dictgen::InsensitiveStr::Ascii("alterantively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("yhe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("tehnicians"),
                &["technicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("ardiuno"), &["arduino"]),
            (
                dictgen::InsensitiveStr::Ascii("constatnly"),
                &["constantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conspriacy"),
                &["conspiracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("rednering"), &["rendering"]),
            (dictgen::InsensitiveStr::Ascii("examplify"), &["exemplify"]),
            (
                dictgen::InsensitiveStr::Ascii("instromented"),
                &["instrumented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenetations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discusssion"),
                &["discussions", "discussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obseravtions"),
                &["observations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coefficiens"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("memoty"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("encourge"), &["encourage"]),
            (dictgen::InsensitiveStr::Ascii("persewing"), &["pursuing"]),
            (dictgen::InsensitiveStr::Ascii("occrring"), &["occurring"]),
            (dictgen::InsensitiveStr::Ascii("barnchers"), &["branchers"]),
            (
                dictgen::InsensitiveStr::Ascii("malfuncion"),
                &["malfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("tringket"), &["trinket"]),
            (dictgen::InsensitiveStr::Ascii("cnfigs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("timetamps"), &["timestamps"]),
            (
                dictgen::InsensitiveStr::Ascii("desciminate"),
                &["discriminate", "disseminate", "decimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("giltchy"), &["glitchy"]),
            (dictgen::InsensitiveStr::Ascii("secertly"), &["secretly"]),
            (dictgen::InsensitiveStr::Ascii("modiffer"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("possibe"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("appropiration"),
                &["appropriation"],
            ),
            (dictgen::InsensitiveStr::Ascii("obeserves"), &["observes"]),
            (dictgen::InsensitiveStr::Ascii("conigured"), &["configured"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomeances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("bount"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("addresing"), &["addressing"]),
            (dictgen::InsensitiveStr::Ascii("strugglin"), &["struggling"]),
            (dictgen::InsensitiveStr::Ascii("digesty"), &["digest"]),
            (dictgen::InsensitiveStr::Ascii("breanches"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("inheritend"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("moutn"), &["mount"]),
            (
                dictgen::InsensitiveStr::Ascii("algoritmos"),
                &["algorithms", "algorithm"],
            ),
            (dictgen::InsensitiveStr::Ascii("accdients"), &["accidents"]),
            (
                dictgen::InsensitiveStr::Ascii("constrainst"),
                &["constraint", "constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertaed"), &["operated"]),
            (dictgen::InsensitiveStr::Ascii("visuabs"), &["visuals"]),
            (dictgen::InsensitiveStr::Ascii("bitcion"), &["bitcoin"]),
            (
                dictgen::InsensitiveStr::Ascii("selfeshness"),
                &["selfishness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directries"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignenment"),
                &["assignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfie"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrait"),
                &["constraints", "constraint"],
            ),
            (dictgen::InsensitiveStr::Ascii("veritcal"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("uninteresing"),
                &["uninteresting"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitural"), &["virtual"]),
            (dictgen::InsensitiveStr::Ascii("troublin"), &["troubling"]),
            (dictgen::InsensitiveStr::Ascii("srver"), &["server"]),
            (dictgen::InsensitiveStr::Ascii("foontnotes"), &["footnotes"]),
            (dictgen::InsensitiveStr::Ascii("threates"), &["threatens"]),
            (
                dictgen::InsensitiveStr::Ascii("detcetions"),
                &["detections"],
            ),
            (dictgen::InsensitiveStr::Ascii("semicoli"), &["semicolons"]),
            (
                dictgen::InsensitiveStr::Ascii("subsribing"),
                &["subscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertsing"),
                &["advertising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurare"),
                &["configure"],
            ),
            (dictgen::InsensitiveStr::Ascii("psedeo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoonded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intergration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("memicked"), &["mimicked"]),
            (
                dictgen::InsensitiveStr::Ascii("recongizable"),
                &["recognizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accociating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsetlling"),
                &["unsettling"],
            ),
            (dictgen::InsensitiveStr::Ascii("sapeena"), &["subpoena"]),
            (
                dictgen::InsensitiveStr::Ascii("comporable"),
                &["comparable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intristically"),
                &["intrinsically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assotiations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("scolls"), &["scrolls"]),
            (dictgen::InsensitiveStr::Ascii("fmaily"), &["family"]),
            (dictgen::InsensitiveStr::Ascii("senerio"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactuing"),
                &["manufacturing"],
            ),
            (dictgen::InsensitiveStr::Ascii("frmat"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("glight"), &["flight"]),
            (dictgen::InsensitiveStr::Ascii("buillt"), &["built"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronise"),
                &["synchronise"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifcycles"), &["lifecycles"]),
            (
                dictgen::InsensitiveStr::Ascii("scandonavia"),
                &["scandinavia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illumintati"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("hunrgy"), &["hungry"]),
            (dictgen::InsensitiveStr::Ascii("disussion"), &["discussion"]),
            (dictgen::InsensitiveStr::Ascii("committ"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("inisghts"), &["insights"]),
            (dictgen::InsensitiveStr::Ascii("actuell"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("messenged"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("deativates"),
                &["deactivates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elecation"),
                &["elevation", "election"],
            ),
            (dictgen::InsensitiveStr::Ascii("hmtl"), &["html"]),
            (dictgen::InsensitiveStr::Ascii("comlaints"), &["complaints"]),
            (dictgen::InsensitiveStr::Ascii("orignial"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("bulletprooof"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denomenations"),
                &["denominations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pririty"),
                &["priority", "privity"],
            ),
            (dictgen::InsensitiveStr::Ascii("uplodaded"), &["uploaded"]),
            (
                dictgen::InsensitiveStr::Ascii("transmiter"),
                &["transmitter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituion"),
                &["prostitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownersip"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("samilar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("microtranscation"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofuse"), &["confuse"]),
            (dictgen::InsensitiveStr::Ascii("depositas"), &["deposits"]),
            (
                dictgen::InsensitiveStr::Ascii("participats"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occure"),
                &["occur", "occurred"],
            ),
            (dictgen::InsensitiveStr::Ascii("profesors"), &["professors"]),
            (dictgen::InsensitiveStr::Ascii("ipubrofen"), &["ibuprofen"]),
            (
                dictgen::InsensitiveStr::Ascii("initialitzing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interperters"),
                &["interpreter", "interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepearators"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("repect"), &["respect"]),
            (dictgen::InsensitiveStr::Ascii("driveris"), &["drivers"]),
            (dictgen::InsensitiveStr::Ascii("wispering"), &["whispering"]),
            (
                dictgen::InsensitiveStr::Ascii("optimasation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("desigs"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtuions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortunte"), &["fortune"]),
            (dictgen::InsensitiveStr::Ascii("associete"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciatin"),
                &["appreciation", "appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequencin"), &["sequencing"]),
            (dictgen::InsensitiveStr::Ascii("confusting"), &["confusing"]),
            (dictgen::InsensitiveStr::Ascii("magintude"), &["magnitude"]),
            (dictgen::InsensitiveStr::Ascii("plagerism"), &["plagiarism"]),
            (
                dictgen::InsensitiveStr::Ascii("repressented"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persitence"),
                &["persistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vitualizes"),
                &["virtualizes", "visualizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiazing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("priniciple"), &["principle"]),
            (dictgen::InsensitiveStr::Ascii("simplistes"), &["simplest"]),
            (dictgen::InsensitiveStr::Ascii("execustion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelisation"),
                &["parallelisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chocolote"), &["chocolate"]),
            (dictgen::InsensitiveStr::Ascii("recoves"), &["recovers"]),
            (
                dictgen::InsensitiveStr::Ascii("gamifications"),
                &["ramifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("assertios"), &["assertions"]),
            (
                dictgen::InsensitiveStr::Ascii("appplications"),
                &["applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extremety"),
                &["extremity", "extremely"],
            ),
            (dictgen::InsensitiveStr::Ascii("headses"), &["headsets"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisactory"),
                &["satisfactory"],
            ),
            (dictgen::InsensitiveStr::Ascii("implament"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("pathc"), &["patch"]),
            (
                dictgen::InsensitiveStr::Ascii("docuemtning"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("paralell"), &["parallel"]),
            (
                dictgen::InsensitiveStr::Ascii("erested"),
                &["arrested", "erected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memorozation"),
                &["memorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("depracate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("businees"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("updaer"), &["updater"]),
            (dictgen::InsensitiveStr::Ascii("adequite"), &["adequate"]),
            (dictgen::InsensitiveStr::Ascii("situdios"), &["studios"]),
            (dictgen::InsensitiveStr::Ascii("intedned"), &["intended"]),
            (dictgen::InsensitiveStr::Ascii("obejct"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("watchog"), &["watchdog"]),
            (dictgen::InsensitiveStr::Ascii("pompt"), &["prompt"]),
            (
                dictgen::InsensitiveStr::Ascii("coneecting"),
                &["connecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equiverlant"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeses"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("persistense"),
                &["persistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("blcoks"), &["blocks"]),
            (
                dictgen::InsensitiveStr::Ascii("lighers"),
                &["lighters", "liars", "ligers"],
            ),
            (dictgen::InsensitiveStr::Ascii("manuley"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapoliation"),
                &["extrapolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("configres"), &["configures"]),
            (
                dictgen::InsensitiveStr::Ascii("collorscheme"),
                &["colorscheme"],
            ),
            (dictgen::InsensitiveStr::Ascii("emabling"), &["enabling"]),
            (dictgen::InsensitiveStr::Ascii("sanetize"), &["sanitize"]),
            (
                dictgen::InsensitiveStr::Ascii("authetnicators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainwrek"), &["trainwreck"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplater"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congfigurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcluator"),
                &["calculators", "calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("prfiler"), &["profiler"]),
            (
                dictgen::InsensitiveStr::Ascii("philospohies"),
                &["philosophies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privledged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacoons"), &["cocoons"]),
            (dictgen::InsensitiveStr::Ascii("muscially"), &["musically"]),
            (dictgen::InsensitiveStr::Ascii("advantag"), &["advantage"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriatness"),
                &["appropriateness"],
            ),
            (dictgen::InsensitiveStr::Ascii("excluing"), &["excluding"]),
            (dictgen::InsensitiveStr::Ascii("despenser"), &["dispenser"]),
            (
                dictgen::InsensitiveStr::Ascii("consciense"),
                &["conscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("packags"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("concentrare"),
                &["concentrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desentitized"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gerenating"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaaluation"), &["valuation"]),
            (dictgen::InsensitiveStr::Ascii("transvorm"), &["transform"]),
            (
                dictgen::InsensitiveStr::Ascii("difficutlies"),
                &["difficulties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisaions"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("autors"), &["authors"]),
            (dictgen::InsensitiveStr::Ascii("neighbout"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("comarison"), &["comparison"]),
            (
                dictgen::InsensitiveStr::Ascii("constructtors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compitition"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("resessed"), &["recessed"]),
            (dictgen::InsensitiveStr::Ascii("contrrols"), &["controls"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpeectedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("episonage"), &["espionage"]),
            (dictgen::InsensitiveStr::Ascii("uncehck"), &["uncheck"]),
            (dictgen::InsensitiveStr::Ascii("gosples"), &["gospels"]),
            (
                dictgen::InsensitiveStr::Ascii("supremicist"),
                &["supremacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegitarians"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("caht"), &["chat"]),
            (
                dictgen::InsensitiveStr::Ascii("technologial"),
                &["technological"],
            ),
            (dictgen::InsensitiveStr::Ascii("prasies"), &["praises"]),
            (dictgen::InsensitiveStr::Ascii("vave"), &["have", "valve"]),
            (
                dictgen::InsensitiveStr::Ascii("atttributions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("contanier"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("spefiifies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("subtitutions"),
                &["substitutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutnerpart"),
                &["counterpart"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monoploies"),
                &["monopolies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimentry"),
                &["complimentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangse"), &["arranges"]),
            (dictgen::InsensitiveStr::Ascii("boyfriens"), &["boyfriends"]),
            (dictgen::InsensitiveStr::Ascii("filenname"), &["filename"]),
            (dictgen::InsensitiveStr::Ascii("journalims"), &["journals"]),
            (
                dictgen::InsensitiveStr::Ascii("openes"),
                &["opens", "openness"],
            ),
            (dictgen::InsensitiveStr::Ascii("monotir"), &["monitor"]),
            (dictgen::InsensitiveStr::Ascii("netwrork"), &["network"]),
            (
                dictgen::InsensitiveStr::Ascii("compleatness"),
                &["completeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("intervalls"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("distoriton"),
                &["distortion"],
            ),
            (dictgen::InsensitiveStr::Ascii("furether"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumnets"),
                &["instruments"],
            ),
            (dictgen::InsensitiveStr::Ascii("asborb"), &["absorb"]),
            (dictgen::InsensitiveStr::Ascii("factualy"), &["factually"]),
            (dictgen::InsensitiveStr::Ascii("summerize"), &["summarize"]),
            (
                dictgen::InsensitiveStr::Ascii("instatiate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("gague"), &["gage"]),
            (dictgen::InsensitiveStr::Ascii("messgaes"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("depciting"), &["depicting"]),
            (dictgen::InsensitiveStr::Ascii("theis"), &["this", "thesis"]),
            (dictgen::InsensitiveStr::Ascii("algorithem"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("definatley"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ampatheaters"),
                &["amphitheaters"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanlder"), &["sandler"]),
            (
                dictgen::InsensitiveStr::Ascii("cyptography"),
                &["cryptography"],
            ),
            (dictgen::InsensitiveStr::Ascii("musculaire"), &["muscular"]),
            (
                dictgen::InsensitiveStr::Ascii("fasinating"),
                &["fascinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissrupt"), &["disrupt"]),
            (dictgen::InsensitiveStr::Ascii("porposals"), &["proposals"]),
            (
                dictgen::InsensitiveStr::Ascii("subcatagories"),
                &["subcategories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronize"),
                &["synchronize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadeguate"),
                &["inadequate"],
            ),
            (dictgen::InsensitiveStr::Ascii("phsyique"), &["physique"]),
            (
                dictgen::InsensitiveStr::Ascii("redemtpion"),
                &["redemption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assginment"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaibale"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("jodpers"), &["jodhpurs"]),
            (
                dictgen::InsensitiveStr::Ascii("complicatin"),
                &["complicit", "complicating", "complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("modeul"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("appretiates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performans"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminicient"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mushrom"), &["mushroom"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoledgments"),
                &["acknowledgments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstadings"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentenceing"),
                &["sentencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("singl"), &["single"]),
            (dictgen::InsensitiveStr::Ascii("obvoius"), &["obvious"]),
            (
                dictgen::InsensitiveStr::Ascii("custoisable"),
                &["customisable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaccurring"),
                &["recurring"],
            ),
            (dictgen::InsensitiveStr::Ascii("accual"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("exmaines"), &["examines"]),
            (dictgen::InsensitiveStr::Ascii("folliwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("thare"), &["there"]),
            (dictgen::InsensitiveStr::Ascii("legngth"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("wrokspace"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("chcekpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkcing"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("erorneusly"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prototypin"),
                &["prototyping"],
            ),
            (dictgen::InsensitiveStr::Ascii("liverpoool"), &["liverpool"]),
            (dictgen::InsensitiveStr::Ascii("hatchig"), &["hatching"]),
            (dictgen::InsensitiveStr::Ascii("pathnme"), &["pathname"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronises"),
                &["synchronises"],
            ),
            (dictgen::InsensitiveStr::Ascii("acronymn"), &["acronym"]),
            (dictgen::InsensitiveStr::Ascii("santcuary"), &["sanctuary"]),
            (dictgen::InsensitiveStr::Ascii("mistery"), &["mystery"]),
            (dictgen::InsensitiveStr::Ascii("serie"), &["series"]),
            (dictgen::InsensitiveStr::Ascii("acual"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalisme"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mangaer"),
                &["manager", "manger"],
            ),
            (dictgen::InsensitiveStr::Ascii("exagonal"), &["hexagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("humiliaton"),
                &["humiliation"],
            ),
            (dictgen::InsensitiveStr::Ascii("florecen"), &["florence"]),
            (
                dictgen::InsensitiveStr::Ascii("disttributed"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disillusionned"),
                &["disillusioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninitailized"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("finxing"), &["fixing"]),
            (dictgen::InsensitiveStr::Ascii("uptional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("placeholde"),
                &["placeholder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialialization"),
                &["serialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conparisions"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denpendences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("expells"), &["expels"]),
            (dictgen::InsensitiveStr::Ascii("aggregete"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("insdie"), &["inside"]),
            (dictgen::InsensitiveStr::Ascii("puch"), &["push"]),
            (dictgen::InsensitiveStr::Ascii("subsysytem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("atomticlly"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("divintiy"), &["divinity"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunatel"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrictings"),
                &["restricting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorised"),
                &["prioritised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransacting"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("degarde"), &["degrade"]),
            (
                dictgen::InsensitiveStr::Ascii("quesrioned"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("numberator"), &["numerator"]),
            (dictgen::InsensitiveStr::Ascii("neeeding"), &["needing"]),
            (
                dictgen::InsensitiveStr::Ascii("authrorized"),
                &["authorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("prorgam"), &["program"]),
            (dictgen::InsensitiveStr::Ascii("upacked"), &["unpacked"]),
            (dictgen::InsensitiveStr::Ascii("docuemnt"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("langurage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("identifyed"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("equipments"), &["equipment"]),
            (dictgen::InsensitiveStr::Ascii("bafore"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("blance"),
                &["balance", "glance", "lance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sypmtom"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("behvioral"), &["behavioral"]),
            (
                dictgen::InsensitiveStr::Ascii("desparation"),
                &["desperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("carnigie"), &["carnegie"]),
            (
                dictgen::InsensitiveStr::Ascii("coincidense"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sevice"),
                &["service", "device"],
            ),
            (dictgen::InsensitiveStr::Ascii("sysyems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posibilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paranthesis"),
                &["parenthesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coprighted"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combainatorial"),
                &["combinatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probelmatic"),
                &["problematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("assinment"), &["assignment"]),
            (dictgen::InsensitiveStr::Ascii("proosed"), &["proposed"]),
            (
                dictgen::InsensitiveStr::Ascii("implemnetation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilzer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("fourty"), &["forty"]),
            (
                dictgen::InsensitiveStr::Ascii("acqusitions"),
                &["acquisitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suroundings"),
                &["surroundings"],
            ),
            (dictgen::InsensitiveStr::Ascii("tention"), &["tension"]),
            (
                dictgen::InsensitiveStr::Ascii("unabled"),
                &["unable", "enabled", "disabled"],
            ),
            (dictgen::InsensitiveStr::Ascii("septemer"), &["september"]),
            (
                dictgen::InsensitiveStr::Ascii("dicriminates"),
                &["discriminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("scripot"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("supoosedly"),
                &["supposedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enmvironment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompartibly"),
                &["incompatibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ubiquitos"), &["ubiquitous"]),
            (
                dictgen::InsensitiveStr::Ascii("hirarcies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contorller"),
                &["controllers", "controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("raotat"), &["rotate"]),
            (
                dictgen::InsensitiveStr::Ascii("subresoures"),
                &["subresources"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attenutaion"),
                &["attenuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australin"),
                &["australians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mondey"),
                &["monday", "money", "monkey"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplyfying"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("monsday"), &["monday"]),
            (dictgen::InsensitiveStr::Ascii("inspite"), &["inspire"]),
            (dictgen::InsensitiveStr::Ascii("follind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("tiemstamp"), &["timestamp"]),
            (
                dictgen::InsensitiveStr::Ascii("instrction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asssessments"),
                &["assessments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apsiration"),
                &["aspiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direccional"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("damanged"), &["damaged"]),
            (dictgen::InsensitiveStr::Ascii("marignal"), &["marginal"]),
            (dictgen::InsensitiveStr::Ascii("utilze"), &["utilize"]),
            (dictgen::InsensitiveStr::Ascii("konstants"), &["constants"]),
            (
                dictgen::InsensitiveStr::Ascii("exixts"),
                &["exists", "exits"],
            ),
            (dictgen::InsensitiveStr::Ascii("commutated"), &["commuted"]),
            (
                dictgen::InsensitiveStr::Ascii("marignally"),
                &["marginally"],
            ),
            (dictgen::InsensitiveStr::Ascii("progarm"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("corretions"),
                &["corrections", "correlations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interacties"),
                &["interacts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tahnkfully"),
                &["thankfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("kowledge"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("waitign"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("whihc"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("shuold"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("gerenator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeard"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipsampling"),
                &["multisampling"],
            ),
            (dictgen::InsensitiveStr::Ascii("coumpounds"), &["compounds"]),
            (dictgen::InsensitiveStr::Ascii("navigatest"), &["navigates"]),
            (
                dictgen::InsensitiveStr::Ascii("originially"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("intertia"), &["inertia"]),
            (dictgen::InsensitiveStr::Ascii("tigthened"), &["tightened"]),
            (
                dictgen::InsensitiveStr::Ascii("strawbeary"),
                &["strawberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("teamfighs"), &["teamfights"]),
            (
                dictgen::InsensitiveStr::Ascii("rehabilitatin"),
                &["rehabilitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undeinably"),
                &["undeniably"],
            ),
            (dictgen::InsensitiveStr::Ascii("nromandy"), &["normandy"]),
            (dictgen::InsensitiveStr::Ascii("termimals"), &["terminals"]),
            (dictgen::InsensitiveStr::Ascii("boundimg"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("espacing"), &["escaping"]),
            (
                dictgen::InsensitiveStr::Ascii("renegociation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("heuristc"), &["heuristic"]),
            (
                dictgen::InsensitiveStr::Ascii("reminisent"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("reponsive"), &["responsive"]),
            (
                dictgen::InsensitiveStr::Ascii("adreniline"),
                &["adrenaline"],
            ),
            (dictgen::InsensitiveStr::Ascii("handelr"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("redundand"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("armegaddon"),
                &["armageddon"],
            ),
            (dictgen::InsensitiveStr::Ascii("accronyms"), &["acronyms"]),
            (dictgen::InsensitiveStr::Ascii("adheasives"), &["adhesives"]),
            (
                dictgen::InsensitiveStr::Ascii("muliplicative"),
                &["multiplicative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discourged"),
                &["discouraged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meanigfull"),
                &["meaningful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desiging"),
                &["designing", "desiring"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcers"), &["producers"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenet"),
                &["implements", "implement"],
            ),
            (dictgen::InsensitiveStr::Ascii("electroal"), &["electoral"]),
            (dictgen::InsensitiveStr::Ascii("refreshses"), &["refreshes"]),
            (
                dictgen::InsensitiveStr::Ascii("intepolating"),
                &["interpolating"],
            ),
            (dictgen::InsensitiveStr::Ascii("smapling"), &["sampling"]),
            (dictgen::InsensitiveStr::Ascii("contro"), &["control"]),
            (dictgen::InsensitiveStr::Ascii("impliment"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("excecises"), &["exercises"]),
            (
                dictgen::InsensitiveStr::Ascii("substraction"),
                &["subtraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequnce"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarantees"),
                &["guarantees", "quarantines"],
            ),
            (dictgen::InsensitiveStr::Ascii("specieal"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("overtunred"),
                &["overturned"],
            ),
            (dictgen::InsensitiveStr::Ascii("spriritual"), &["spiritual"]),
            (dictgen::InsensitiveStr::Ascii("sebastain"), &["sebastian"]),
            (dictgen::InsensitiveStr::Ascii("catridge"), &["cartridge"]),
            (dictgen::InsensitiveStr::Ascii("signauture"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("contect"),
                &["contact", "context", "connect"],
            ),
            (dictgen::InsensitiveStr::Ascii("byteoder"), &["byteorder"]),
            (dictgen::InsensitiveStr::Ascii("transalte"), &["translate"]),
            (dictgen::InsensitiveStr::Ascii("folls"), &["follows"]),
            (dictgen::InsensitiveStr::Ascii("crulety"), &["cruelty"]),
            (dictgen::InsensitiveStr::Ascii("protestes"), &["protests"]),
            (dictgen::InsensitiveStr::Ascii("proceses"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("opulates"),
                &["populates", "opiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("prfile"), &["profile"]),
            (dictgen::InsensitiveStr::Ascii("detectes"), &["detects"]),
            (dictgen::InsensitiveStr::Ascii("endever"), &["endeavor"]),
            (dictgen::InsensitiveStr::Ascii("progamms"), &["programs"]),
            (
                dictgen::InsensitiveStr::Ascii("distinctin"),
                &["distinctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("separetor"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("ondexing"), &["indexing"]),
            (dictgen::InsensitiveStr::Ascii("depolyed"), &["deployed"]),
            (
                dictgen::InsensitiveStr::Ascii("raplacement"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thoose"),
                &["those", "choose"],
            ),
            (dictgen::InsensitiveStr::Ascii("origininal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("incliude"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("accumultate"),
                &["accumulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("descision"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("vauling"), &["valuing"]),
            (dictgen::InsensitiveStr::Ascii("nuremburg"), &["nuremberg"]),
            (dictgen::InsensitiveStr::Ascii("promots"), &["promotes"]),
            (dictgen::InsensitiveStr::Ascii("choicing"), &["choosing"]),
            (
                dictgen::InsensitiveStr::Ascii("embededding"),
                &["embedding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humanitites"),
                &["humanities"],
            ),
            (dictgen::InsensitiveStr::Ascii("buhddism"), &["buddhism"]),
            (
                dictgen::InsensitiveStr::Ascii("ralatively"),
                &["relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificastions"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhencements"),
                &["enhancements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deccrements"),
                &["decrements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfacin"),
                &["interfacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pratictioners"),
                &["practitioners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutaters"),
                &["mutators", "mutates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparaison"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciffication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("collsion"), &["collision"]),
            (dictgen::InsensitiveStr::Ascii("releived"), &["relieved"]),
            (
                dictgen::InsensitiveStr::Ascii("cutomizing"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecret"),
                &["secret", "erect"],
            ),
            (dictgen::InsensitiveStr::Ascii("deadlifs"), &["deadlifts"]),
            (
                dictgen::InsensitiveStr::Ascii("hadshake"),
                &["handshake", "headshake"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkmeat"), &["checkmate"]),
            (dictgen::InsensitiveStr::Ascii("seziure"), &["seizure"]),
            (dictgen::InsensitiveStr::Ascii("suficate"), &["suffocate"]),
            (dictgen::InsensitiveStr::Ascii("realisme"), &["realise"]),
            (dictgen::InsensitiveStr::Ascii("avodacos"), &["avocados"]),
            (
                dictgen::InsensitiveStr::Ascii("tempraily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complemetary"),
                &["complementary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chihuahuita"),
                &["chihuahua"],
            ),
            (dictgen::InsensitiveStr::Ascii("resposne"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("rearrangs"), &["rearranges"]),
            (
                dictgen::InsensitiveStr::Ascii("readed"),
                &["read", "readd", "readded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theorician"),
                &["theoretician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immsersively"),
                &["immersively"],
            ),
            (dictgen::InsensitiveStr::Ascii("shuttind"), &["shutting"]),
            (dictgen::InsensitiveStr::Ascii("kindoms"), &["kingdoms"]),
            (dictgen::InsensitiveStr::Ascii("strutures"), &["structures"]),
            (dictgen::InsensitiveStr::Ascii("reqiore"), &["require"]),
            (
                dictgen::InsensitiveStr::Ascii("dcumentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vactors"), &["vectors"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("checcklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrivates"),
                &["aggravates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapplines"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("intruige"), &["intrigue"]),
            (dictgen::InsensitiveStr::Ascii("asender"), &["ascender"]),
            (
                dictgen::InsensitiveStr::Ascii("sentancing"),
                &["sentencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missclassified"),
                &["misclassified"],
            ),
            (dictgen::InsensitiveStr::Ascii("atempt"), &["attempt"]),
            (
                dictgen::InsensitiveStr::Ascii("scavangers"),
                &["scavengers"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolue"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("assumued"), &["assumed"]),
            (dictgen::InsensitiveStr::Ascii("cleanies"), &["cleanse"]),
            (dictgen::InsensitiveStr::Ascii("rhapsoy"), &["rhapsody"]),
            (
                dictgen::InsensitiveStr::Ascii("associcating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checckerboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("hardend"), &["hardened"]),
            (dictgen::InsensitiveStr::Ascii("godd"), &["good", "god"]),
            (dictgen::InsensitiveStr::Ascii("adovcated"), &["advocated"]),
            (
                dictgen::InsensitiveStr::Ascii("essentiels"),
                &["essentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("distatse"), &["distaste"]),
            (
                dictgen::InsensitiveStr::Ascii("approximes"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitraion"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("desriptor"), &["descriptor"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialisaitons"),
                &["specialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumualtive"),
                &["accumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("strciter"), &["stricter"]),
            (
                dictgen::InsensitiveStr::Ascii("coalacence"),
                &["coalescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("curremt"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("auidobooks"),
                &["audiobooks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deaktivate"),
                &["deactivate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propability"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("turthfully"),
                &["truthfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("ewxception"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphysisch"),
                &["metaphysics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructino"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("obejcts"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("ideosincratic"),
                &["idiosyncratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("condidate"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("paravirutalization"),
                &["paravirtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("timoeuts"), &["timeouts"]),
            (
                dictgen::InsensitiveStr::Ascii("cooordinates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("itereative"), &["iterative"]),
            (dictgen::InsensitiveStr::Ascii("sepcify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("essentiatlly"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compromsie"),
                &["compromises"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusotodian"), &["custodian"]),
            (dictgen::InsensitiveStr::Ascii("libreries"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("imrovements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("logarithmetic"),
                &["logarithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("conicide"), &["coincide"]),
            (
                dictgen::InsensitiveStr::Ascii("finctionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempertaures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotaition"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronisation"),
                &["synchronisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parlamentary"),
                &["parliamentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("pumpinks"), &["pumpkins"]),
            (
                dictgen::InsensitiveStr::Ascii("activationg"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatined"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("delapidated"),
                &["dilapidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tentacls"), &["tentacles"]),
            (
                dictgen::InsensitiveStr::Ascii("chronciles"),
                &["chronicles"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssasin"), &["assassin"]),
            (
                dictgen::InsensitiveStr::Ascii("absolutley"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("reseearch"), &["research"]),
            (
                dictgen::InsensitiveStr::Ascii("idendifiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiallised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("pavillion"), &["pavilion"]),
            (dictgen::InsensitiveStr::Ascii("garillas"), &["guerillas"]),
            (dictgen::InsensitiveStr::Ascii("unbouund"), &["unbound"]),
            (dictgen::InsensitiveStr::Ascii("lemosine"), &["limousine"]),
            (dictgen::InsensitiveStr::Ascii("arthirtis"), &["arthritis"]),
            (dictgen::InsensitiveStr::Ascii("filebame"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("battlesaur"),
                &["battlestar"],
            ),
            (dictgen::InsensitiveStr::Ascii("hellow"), &["hello"]),
            (dictgen::InsensitiveStr::Ascii("instnces"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("allocarion"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bengzhai"), &["benghazi"]),
            (
                dictgen::InsensitiveStr::Ascii("metophorical"),
                &["metaphorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerablities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("informable"), &["informal"]),
            (
                dictgen::InsensitiveStr::Ascii("associeates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefifed"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("publshing"), &["publishing"]),
            (dictgen::InsensitiveStr::Ascii("analisys"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("collecte"), &["collective"]),
            (
                dictgen::InsensitiveStr::Ascii("communicatin"),
                &["communications", "communication", "communicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comeptitions"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufficently"),
                &["sufficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepecifies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("overreactiong"),
                &["overreacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tabluating"),
                &["tabulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallely"),
                &["parallelly"],
            ),
            (dictgen::InsensitiveStr::Ascii("dowrd"), &["dword"]),
            (
                dictgen::InsensitiveStr::Ascii("unrepsonsive"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("vactor"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("miniture"), &["miniature"]),
            (dictgen::InsensitiveStr::Ascii("wondeirng"), &["wondering"]),
            (
                dictgen::InsensitiveStr::Ascii("espond"),
                &["respond", "despond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transperencies"),
                &["transparencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeficients"),
                &["coefficients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exting"),
                &["existing", "exiting", "texting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undergradute"),
                &["undergraduate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standarized"),
                &["standardized"],
            ),
            (dictgen::InsensitiveStr::Ascii("faillover"), &["failover"]),
            (
                dictgen::InsensitiveStr::Ascii("distributore"),
                &["distribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("healther"), &["healthier"]),
            (dictgen::InsensitiveStr::Ascii("invlisible"), &["invisible"]),
            (dictgen::InsensitiveStr::Ascii("destroyedd"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalizacion"),
                &["capitalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timming"),
                &["timing", "trimming"],
            ),
            (dictgen::InsensitiveStr::Ascii("oudated"), &["outdated"]),
            (dictgen::InsensitiveStr::Ascii("furtherst"), &["furthest"]),
            (dictgen::InsensitiveStr::Ascii("argeuments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("effcient"), &["efficient"]),
            (
                dictgen::InsensitiveStr::Ascii("circluarly"),
                &["circularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("unroated"), &["unrotated"]),
            (
                dictgen::InsensitiveStr::Ascii("parkay"),
                &["parkway", "parlay", "parquet"],
            ),
            (dictgen::InsensitiveStr::Ascii("picknicer"), &["picnicker"]),
            (dictgen::InsensitiveStr::Ascii("penisse"), &["penises"]),
            (
                dictgen::InsensitiveStr::Ascii("aross"),
                &["across", "arose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assisnate"),
                &["assassinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alteratively"),
                &["alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reallocaiting"),
                &["reallocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("releave"), &["relieve"]),
            (
                dictgen::InsensitiveStr::Ascii("deffer"),
                &["differ", "defer"],
            ),
            (dictgen::InsensitiveStr::Ascii("sibtitle"), &["subtitle"]),
            (
                dictgen::InsensitiveStr::Ascii("calbirating"),
                &["calibrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("routiner"), &["routine"]),
            (
                dictgen::InsensitiveStr::Ascii("aletrnately"),
                &["alternately"],
            ),
            (dictgen::InsensitiveStr::Ascii("vektor"), &["vector"]),
            (
                dictgen::InsensitiveStr::Ascii("concentating"),
                &["concentrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("opposiste"), &["opposites"]),
            (
                dictgen::InsensitiveStr::Ascii("utilizacion"),
                &["utilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("acryllic"), &["acrylic"]),
            (dictgen::InsensitiveStr::Ascii("targested"), &["targeted"]),
            (dictgen::InsensitiveStr::Ascii("wriets"), &["writes"]),
            (dictgen::InsensitiveStr::Ascii("downvotr"), &["downvoters"]),
            (dictgen::InsensitiveStr::Ascii("dyansty"), &["dynasty"]),
            (dictgen::InsensitiveStr::Ascii("accepects"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("utilizies"),
                &["utilities", "utilizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("termials"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("damon"),
                &["daemon", "demon"],
            ),
            (dictgen::InsensitiveStr::Ascii("waiitng"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("htacces"), &["htaccess"]),
            (dictgen::InsensitiveStr::Ascii("eroneous"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("coutnered"), &["countered"]),
            (dictgen::InsensitiveStr::Ascii("trollys"), &["trolleys"]),
            (dictgen::InsensitiveStr::Ascii("theroists"), &["theorists"]),
            (dictgen::InsensitiveStr::Ascii("purused"), &["pursued"]),
            (dictgen::InsensitiveStr::Ascii("advacned"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("inhertances"),
                &["inheritances"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifing"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("januaray"), &["january"]),
            (dictgen::InsensitiveStr::Ascii("wouldnot"), &["wouldnt"]),
            (dictgen::InsensitiveStr::Ascii("improoving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("returnned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("buitins"), &["builtins"]),
            (
                dictgen::InsensitiveStr::Ascii("indestructble"),
                &["indestructible"],
            ),
            (dictgen::InsensitiveStr::Ascii("dahsboard"), &["dashboard"]),
            (
                dictgen::InsensitiveStr::Ascii("reccurence"),
                &["recurrence"],
            ),
            (dictgen::InsensitiveStr::Ascii("bernouilli"), &["bernoulli"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitiviy"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("trageted"), &["targeted"]),
            (
                dictgen::InsensitiveStr::Ascii("introverso"),
                &["introverts"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecets"), &["selects"]),
            (dictgen::InsensitiveStr::Ascii("sanwich"), &["sandwich"]),
            (dictgen::InsensitiveStr::Ascii("infectin"), &["infections"]),
            (dictgen::InsensitiveStr::Ascii("produser"), &["producer"]),
            (
                dictgen::InsensitiveStr::Ascii("recommndations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("receving"), &["receiving"]),
            (dictgen::InsensitiveStr::Ascii("regultors"), &["regulators"]),
            (dictgen::InsensitiveStr::Ascii("juvenilles"), &["juvenile"]),
            (
                dictgen::InsensitiveStr::Ascii("virtualiziation"),
                &["virtualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extentions"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swtichback"),
                &["switchback"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sattellites"),
                &["satellites"],
            ),
            (dictgen::InsensitiveStr::Ascii("preapres"), &["prepares"]),
            (dictgen::InsensitiveStr::Ascii("ganerates"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("deadpol"), &["deadpool"]),
            (dictgen::InsensitiveStr::Ascii("categorry"), &["category"]),
            (
                dictgen::InsensitiveStr::Ascii("depercating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrrive"), &["arrive"]),
            (
                dictgen::InsensitiveStr::Ascii("dyamically"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ackumulate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperataly"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challanges"),
                &["challenges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predujiced"),
                &["prejudiced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandizement"),
                &["aggrandizement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoritmes"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acclerometers"),
                &["accelerometers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hte"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("encocdes"), &["encodes"]),
            (
                dictgen::InsensitiveStr::Ascii("ideologins"),
                &["ideologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socialicing"),
                &["socializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comvolution"),
                &["convolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("instlaler"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("mehodical"), &["methodical"]),
            (
                dictgen::InsensitiveStr::Ascii("asssociations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplicte"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("reynols"), &["reynolds"]),
            (dictgen::InsensitiveStr::Ascii("actiual"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("cloack"), &["cloak"]),
            (
                dictgen::InsensitiveStr::Ascii("extraordinay"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("temporar"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("mastrubate"),
                &["masturbate"],
            ),
            (dictgen::InsensitiveStr::Ascii("syrains"), &["syrians"]),
            (
                dictgen::InsensitiveStr::Ascii("provid"),
                &["provide", "prove", "proved", "proves"],
            ),
            (dictgen::InsensitiveStr::Ascii("utitlty"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("premable"), &["preamble"]),
            (
                dictgen::InsensitiveStr::Ascii("hights"),
                &["height", "heights"],
            ),
            (dictgen::InsensitiveStr::Ascii("overhpyed"), &["overhyped"]),
            (dictgen::InsensitiveStr::Ascii("sceanrio"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("neverthless"),
                &["nevertheless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fased"),
                &["fazed", "phased"],
            ),
            (dictgen::InsensitiveStr::Ascii("editible"), &["editable"]),
            (dictgen::InsensitiveStr::Ascii("definatly"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("interepolated"),
                &["interpolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsbiilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("berform"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("remeote"), &["remote"]),
            (dictgen::InsensitiveStr::Ascii("desgin"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("presere"), &["preserve"]),
            (
                dictgen::InsensitiveStr::Ascii("experimtenel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignoret"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("reuplaoders"),
                &["reuploaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firefigthers"),
                &["firefighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("prseent"), &["present"]),
            (dictgen::InsensitiveStr::Ascii("idetifier"), &["identifier"]),
            (dictgen::InsensitiveStr::Ascii("fasodd"), &["facade"]),
            (dictgen::InsensitiveStr::Ascii("corssfire"), &["crossfire"]),
            (
                dictgen::InsensitiveStr::Ascii("particiapnts"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accoundants"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurence"), &["recurrence"]),
            (
                dictgen::InsensitiveStr::Ascii("metropoliton"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("shawshak"), &["shawshank"]),
            (dictgen::InsensitiveStr::Ascii("laregly"), &["largely"]),
            (dictgen::InsensitiveStr::Ascii("immediet"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("posponed"), &["postponed"]),
            (
                dictgen::InsensitiveStr::Ascii("commmunicates"),
                &["communicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoruthmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arithmitic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("drvier"), &["driver"]),
            (
                dictgen::InsensitiveStr::Ascii("highlithing"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certicate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ohter"), &["other"]),
            (
                dictgen::InsensitiveStr::Ascii("idead"),
                &["idea", "ideas", "ideal", "dead"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangtes"), &["arranges"]),
            (dictgen::InsensitiveStr::Ascii("preven"), &["prevent"]),
            (
                dictgen::InsensitiveStr::Ascii("sharin"),
                &["sharing", "sharia", "sarin"],
            ),
            (dictgen::InsensitiveStr::Ascii("lentghs"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("customizatons"),
                &["customizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehere"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("sepulchrly"),
                &["sepulchrally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interatellar"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coallescing"),
                &["coalescing"],
            ),
            (dictgen::InsensitiveStr::Ascii("scirpts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("warppring"),
                &["warping", "wrapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conter"),
                &["conteur", "counter"],
            ),
            (dictgen::InsensitiveStr::Ascii("plaese"), &["please"]),
            (dictgen::InsensitiveStr::Ascii("convigure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("trustwortyness"),
                &["trustworthiness"],
            ),
            (dictgen::InsensitiveStr::Ascii("deleteing"), &["deleting"]),
            (dictgen::InsensitiveStr::Ascii("nummerals"), &["numerals"]),
            (
                dictgen::InsensitiveStr::Ascii("spacify"),
                &["specify", "pacify"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciprocoals"),
                &["reciprocals"],
            ),
            (dictgen::InsensitiveStr::Ascii("spririt"), &["spirit"]),
            (dictgen::InsensitiveStr::Ascii("exhibitons"), &["exhibits"]),
            (dictgen::InsensitiveStr::Ascii("replcias"), &["replicas"]),
            (
                dictgen::InsensitiveStr::Ascii("restes"),
                &["reset", "rests"],
            ),
            (dictgen::InsensitiveStr::Ascii("eminate"), &["emanate"]),
            (
                dictgen::InsensitiveStr::Ascii("regardeless"),
                &["regardless"],
            ),
            (dictgen::InsensitiveStr::Ascii("polygoon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("selecetor"), &["selector"]),
            (dictgen::InsensitiveStr::Ascii("triolgy"), &["trilogy"]),
            (
                dictgen::InsensitiveStr::Ascii("decalarations"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traingulating"),
                &["triangulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defailing"),
                &["detailing", "derailing"],
            ),
            (dictgen::InsensitiveStr::Ascii("devicdes"), &["devices"]),
            (dictgen::InsensitiveStr::Ascii("millenna"), &["millennia"]),
            (dictgen::InsensitiveStr::Ascii("ckecking"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("fuctionoid"),
                &["functionoid"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challening"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuildt"), &["rebuilt"]),
            (
                dictgen::InsensitiveStr::Ascii("comisioning"),
                &["commissioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmitted"), &["committed"]),
            (
                dictgen::InsensitiveStr::Ascii("rewrire"),
                &["rewrite", "rewire"],
            ),
            (dictgen::InsensitiveStr::Ascii("cigarrette"), &["cigarette"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialsie"),
                &["deserialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagaes"), &["messages"]),
            (
                dictgen::InsensitiveStr::Ascii("procedurial"),
                &["procedural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deivatives"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("figgures"), &["figures"]),
            (dictgen::InsensitiveStr::Ascii("hvae"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("djanog"), &["django"]),
            (dictgen::InsensitiveStr::Ascii("quessants"), &["croissants"]),
            (dictgen::InsensitiveStr::Ascii("availaable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("innitials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentater"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("resedue"), &["residue"]),
            (
                dictgen::InsensitiveStr::Ascii("authenciator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aaproximate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("respods"), &["responds"]),
            (
                dictgen::InsensitiveStr::Ascii("businessen"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("eygptian"), &["egyptian"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnearable"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confederancy"),
                &["confederacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disicpline"),
                &["discipline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocrosecond"),
                &["microsecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlayzed"), &["analyzed"]),
            (dictgen::InsensitiveStr::Ascii("mototola"), &["motorola"]),
            (
                dictgen::InsensitiveStr::Ascii("circuncision"),
                &["circumcision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encompuses"),
                &["encompasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("unued"), &["unused"]),
            (dictgen::InsensitiveStr::Ascii("genereated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("annoucement"),
                &["announcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("schudule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("traslucency"),
                &["translucency"],
            ),
            (dictgen::InsensitiveStr::Ascii("basicley"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("evning"), &["evening"]),
            (
                dictgen::InsensitiveStr::Ascii("architectrual"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communucation"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("methaporically"),
                &["metaphorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceutioner"),
                &["executioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partioned"),
                &["partitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reacher"),
                &["richer", "reader"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhiliated"),
                &["annihilated"],
            ),
            (dictgen::InsensitiveStr::Ascii("aligments"), &["alignments"]),
            (
                dictgen::InsensitiveStr::Ascii("attemptting"),
                &["attempting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nimphomaniac"),
                &["nymphomaniac"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("namesspaces"),
                &["namespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("accptor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("constrcuct"), &["construct"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastion"),
                &["procreation"],
            ),
            (dictgen::InsensitiveStr::Ascii("eeger"), &["eager"]),
            (
                dictgen::InsensitiveStr::Ascii("conrrespondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissbles"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("envirnments"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jouranlist"),
                &["journalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceppting"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("vitenamese"),
                &["vietnamese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecticunts"),
                &["connecticut"],
            ),
            (dictgen::InsensitiveStr::Ascii("reflesh"), &["refresh"]),
            (
                dictgen::InsensitiveStr::Ascii("distrtibuting"),
                &["distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("yaerly"), &["yearly"]),
            (dictgen::InsensitiveStr::Ascii("removeing"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("mommentarily"),
                &["momentarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfoemed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("termo"), &["thermo"]),
            (dictgen::InsensitiveStr::Ascii("varinats"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("knowleagable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("toronta"), &["toronto"]),
            (dictgen::InsensitiveStr::Ascii("witout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("millimitres"),
                &["millimetres"],
            ),
            (dictgen::InsensitiveStr::Ascii("upating"), &["updating"]),
            (dictgen::InsensitiveStr::Ascii("disrm"), &["disarm"]),
            (dictgen::InsensitiveStr::Ascii("idealistc"), &["idealistic"]),
            (dictgen::InsensitiveStr::Ascii("versoins"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("unforttunately"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationalizm"),
                &["sensationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justifible"),
                &["justifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("apppearing"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("hae"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("envioments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlysers"), &["analysers"]),
            (dictgen::InsensitiveStr::Ascii("gymnists"), &["gymnasts"]),
            (dictgen::InsensitiveStr::Ascii("spahgetti"), &["spaghetti"]),
            (dictgen::InsensitiveStr::Ascii("debufs"), &["debugfs"]),
            (dictgen::InsensitiveStr::Ascii("paragrah"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecessiarlly"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegoation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civelisations"),
                &["civilisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertinant"), &["pertinent"]),
            (
                dictgen::InsensitiveStr::Ascii("prescriped"),
                &["prescribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("energis"), &["energies"]),
            (
                dictgen::InsensitiveStr::Ascii("allegainces"),
                &["allegiances"],
            ),
            (dictgen::InsensitiveStr::Ascii("opaykely"), &["opaquely"]),
            (dictgen::InsensitiveStr::Ascii("traget"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("uptionally"),
                &["optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemipsheres"),
                &["hemisphere", "hemispheres"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlato"),
                &["correlation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetser"), &["tester"]),
            (
                dictgen::InsensitiveStr::Ascii("purpetuating"),
                &["perpetuating"],
            ),
            (dictgen::InsensitiveStr::Ascii("buglaria"), &["bulgaria"]),
            (dictgen::InsensitiveStr::Ascii("fastests"), &["fastest"]),
            (dictgen::InsensitiveStr::Ascii("retrievied"), &["retrieved"]),
            (dictgen::InsensitiveStr::Ascii("trougth"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("imprefect"), &["imperfect"]),
            (dictgen::InsensitiveStr::Ascii("foloowong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("hasing"), &["hashing"]),
            (dictgen::InsensitiveStr::Ascii("timeots"), &["timeouts"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatablity"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signatue"),
                &["signatures", "signature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ghostcript"),
                &["ghostscript"],
            ),
            (dictgen::InsensitiveStr::Ascii("spawend"), &["spawned"]),
            (dictgen::InsensitiveStr::Ascii("moleclues"), &["molecules"]),
            (dictgen::InsensitiveStr::Ascii("alignrigh"), &["alignright"]),
            (dictgen::InsensitiveStr::Ascii("qustions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("circulater"),
                &["circulator", "circulated", "circulates", "circulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("contactes"), &["contacts"]),
            (dictgen::InsensitiveStr::Ascii("reqiores"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("eloquintly"),
                &["eloquently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sophicated"),
                &["sophisticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stying"),
                &["staying", "styling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiniators"),
                &["combinators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archeaologist"),
                &["archeologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desallocated"),
                &["deallocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("fewsha"), &["fuchsia"]),
            (dictgen::InsensitiveStr::Ascii("tookit"), &["toolkit"]),
            (dictgen::InsensitiveStr::Ascii("frequancy"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("contacentaion"),
                &["concatenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("udpdated"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("adjsuts"), &["adjusts"]),
            (
                dictgen::InsensitiveStr::Ascii("suppurters"),
                &["supporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increaments"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proceessing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("theive"), &["thief"]),
            (dictgen::InsensitiveStr::Ascii("confunsed"), &["confused"]),
            (dictgen::InsensitiveStr::Ascii("requres"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("pascheurization"),
                &["pasteurization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalistas"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("unwieldly"), &["unwieldy"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoof"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("clsoing"), &["closing"]),
            (
                dictgen::InsensitiveStr::Ascii("controversal"),
                &["controversial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outlowing"),
                &["outflowing", "outlawing"],
            ),
            (dictgen::InsensitiveStr::Ascii("activison"), &["activision"]),
            (dictgen::InsensitiveStr::Ascii("abundaces"), &["abundances"]),
            (dictgen::InsensitiveStr::Ascii("sumarized"), &["summarized"]),
            (
                dictgen::InsensitiveStr::Ascii("motorcyce"),
                &["motorcycles"],
            ),
            (dictgen::InsensitiveStr::Ascii("rotaion"), &["rotation"]),
            (
                dictgen::InsensitiveStr::Ascii("parameteras"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notwhithstanding"),
                &["notwithstanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooporative"),
                &["cooperative"],
            ),
            (dictgen::InsensitiveStr::Ascii("caluculate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("excempted"), &["exempted"]),
            (dictgen::InsensitiveStr::Ascii("effecive"), &["effective"]),
            (dictgen::InsensitiveStr::Ascii("eczecutes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("inputed"), &["inputted"]),
            (dictgen::InsensitiveStr::Ascii("addopt"), &["adopt"]),
            (dictgen::InsensitiveStr::Ascii("peices"), &["pieces"]),
            (dictgen::InsensitiveStr::Ascii("gallaxies"), &["galaxies"]),
            (dictgen::InsensitiveStr::Ascii("refertence"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("substatial"),
                &["substantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("stlyish"), &["stylish"]),
            (
                dictgen::InsensitiveStr::Ascii("apdating"),
                &["updating", "adapting"],
            ),
            (dictgen::InsensitiveStr::Ascii("coexinst"), &["coexist"]),
            (
                dictgen::InsensitiveStr::Ascii("initilialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("schcema"), &["schema"]),
            (dictgen::InsensitiveStr::Ascii("immatably"), &["immutably"]),
            (
                dictgen::InsensitiveStr::Ascii("quesitoning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("enclosin"), &["enclosing"]),
            (
                dictgen::InsensitiveStr::Ascii("multpilying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("shieldd"), &["shielded"]),
            (dictgen::InsensitiveStr::Ascii("fineshes"), &["finesse"]),
            (dictgen::InsensitiveStr::Ascii("confirmd"), &["confirmed"]),
            (
                dictgen::InsensitiveStr::Ascii("refelection"),
                &["reflection"],
            ),
            (dictgen::InsensitiveStr::Ascii("symbl"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("domonstrate"),
                &["demonstrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("giltched"), &["glitched"]),
            (dictgen::InsensitiveStr::Ascii("criticisme"), &["criticise"]),
            (dictgen::InsensitiveStr::Ascii("maube"), &["maybe", "mauve"]),
            (
                dictgen::InsensitiveStr::Ascii("restricte"),
                &["restrictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mezmorising"),
                &["mesmerising"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcitions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("constratining"),
                &["constraining"],
            ),
            (dictgen::InsensitiveStr::Ascii("iranias"), &["iranians"]),
            (dictgen::InsensitiveStr::Ascii("numbred"), &["numbered"]),
            (
                dictgen::InsensitiveStr::Ascii("srcollbars"),
                &["scrollbars"],
            ),
            (dictgen::InsensitiveStr::Ascii("offener"), &["oftener"]),
            (
                dictgen::InsensitiveStr::Ascii("commpatibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimisim"), &["optimism"]),
            (
                dictgen::InsensitiveStr::Ascii("brainwahsing"),
                &["brainwashing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reorganizin"),
                &["reorganizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calucation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theroetical"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("draculla"), &["dracula"]),
            (
                dictgen::InsensitiveStr::Ascii("deactived"),
                &["deactivated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajust"), &["adjust"]),
            (
                dictgen::InsensitiveStr::Ascii("environmnet"),
                &["environments", "environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("lepracy"), &["leprosy"]),
            (
                dictgen::InsensitiveStr::Ascii("probabilite"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("pleacing"), &["placing"]),
            (dictgen::InsensitiveStr::Ascii("duobles"), &["doubles"]),
            (dictgen::InsensitiveStr::Ascii("divder"), &["divider"]),
            (
                dictgen::InsensitiveStr::Ascii("documantations"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("optinal"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("smal"), &["small"]),
            (dictgen::InsensitiveStr::Ascii("anonomous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("formtating"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("requireed"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("torrentas"), &["torrents"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessflly"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciritcality"),
                &["criticality"],
            ),
            (dictgen::InsensitiveStr::Ascii("clyinder"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("brusselers"), &["brussels"]),
            (dictgen::InsensitiveStr::Ascii("sceond"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("exlcuded"), &["excluded"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretting"),
                &["interpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("whan"), &["want", "when"]),
            (
                dictgen::InsensitiveStr::Ascii("fransiscans"),
                &["franciscans"],
            ),
            (dictgen::InsensitiveStr::Ascii("darnkess"), &["darkness"]),
            (
                dictgen::InsensitiveStr::Ascii("manafacturers"),
                &["manufacturers"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("milimetres"),
                &["millimetres"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regconisers"),
                &["recognisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenanagans"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("lcoalizes"), &["localizes"]),
            (
                dictgen::InsensitiveStr::Ascii("disingenous"),
                &["disingenuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefected"), &["perfected"]),
            (
                dictgen::InsensitiveStr::Ascii("councelled"),
                &["counselled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibitifs"),
                &["prohibits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referensing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spesifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("scrached"), &["scratched"]),
            (dictgen::InsensitiveStr::Ascii("oppinion"), &["opinion"]),
            (dictgen::InsensitiveStr::Ascii("suscribes"), &["subscribes"]),
            (dictgen::InsensitiveStr::Ascii("variery"), &["variety"]),
            (dictgen::InsensitiveStr::Ascii("convertis"), &["converts"]),
            (
                dictgen::InsensitiveStr::Ascii("accestor"),
                &["accessor", "ancestor"],
            ),
            (dictgen::InsensitiveStr::Ascii("modulues"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("seacrcher"), &["searcher"]),
            (
                dictgen::InsensitiveStr::Ascii("synonmyous"),
                &["synonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alers"),
                &["alerts", "alters"],
            ),
            (dictgen::InsensitiveStr::Ascii("repositry"), &["repository"]),
            (dictgen::InsensitiveStr::Ascii("folde"), &["folder", "fold"]),
            (dictgen::InsensitiveStr::Ascii("innoucous"), &["innocuous"]),
            (
                dictgen::InsensitiveStr::Ascii("loclaizing"),
                &["localizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("flage"), &["flags", "flag"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrarilly"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maginficent"),
                &["magnificent"],
            ),
            (dictgen::InsensitiveStr::Ascii("inintially"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("availibilities"),
                &["availabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("competely"), &["completely"]),
            (dictgen::InsensitiveStr::Ascii("porfile"), &["profile"]),
            (
                dictgen::InsensitiveStr::Ascii("fromidable"),
                &["formidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aborded"),
                &["aborted", "boarded"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartledge"), &["cartilage"]),
            (
                dictgen::InsensitiveStr::Ascii("fallthough"),
                &["fallthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingushing"),
                &["distinguishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgartes"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("adpat"), &["adapt"]),
            (
                dictgen::InsensitiveStr::Ascii("javascritp"),
                &["javascript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socialistos"),
                &["socialists"],
            ),
            (dictgen::InsensitiveStr::Ascii("defininng"), &["defining"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlators"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("lcuase"), &["clause"]),
            (dictgen::InsensitiveStr::Ascii("mechanim"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("decalaratively"),
                &["declaratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("befreind"), &["befriend"]),
            (
                dictgen::InsensitiveStr::Ascii("retalitation"),
                &["retaliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustrasion"),
                &["illustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucessors"), &["successors"]),
            (
                dictgen::InsensitiveStr::Ascii("assumpetions"),
                &["assumptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("strenghen"), &["strengthen"]),
            (dictgen::InsensitiveStr::Ascii("liteicon"), &["litecoin"]),
            (dictgen::InsensitiveStr::Ascii("locaised"), &["localised"]),
            (
                dictgen::InsensitiveStr::Ascii("cheets"),
                &["cheats", "sheets", "chests", "cheers", "cheeks"],
            ),
            (dictgen::InsensitiveStr::Ascii("coypright"), &["copyright"]),
            (
                dictgen::InsensitiveStr::Ascii("physciatrists"),
                &["psychiatrists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectivley"),
                &["collectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depndences"),
                &["dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensitity"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("gadgest"), &["gadgets"]),
            (dictgen::InsensitiveStr::Ascii("iniection"), &["injection"]),
            (dictgen::InsensitiveStr::Ascii("calciulate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("guass"), &["gauss"]),
            (dictgen::InsensitiveStr::Ascii("neighburs"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("weaknesess"),
                &["weaknesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recipiants"),
                &["recipients"],
            ),
            (dictgen::InsensitiveStr::Ascii("crdit"), &["credit"]),
            (dictgen::InsensitiveStr::Ascii("algorthym"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("assumuing"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("flaghsip"), &["flagship"]),
            (dictgen::InsensitiveStr::Ascii("metropos"), &["metropolis"]),
            (
                dictgen::InsensitiveStr::Ascii("critisisms"),
                &["criticisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefitler"), &["prefilter"]),
            (
                dictgen::InsensitiveStr::Ascii("insantiate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("inposible"), &["impossible"]),
            (dictgen::InsensitiveStr::Ascii("reease"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("trubadours"),
                &["troubadours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanciated"),
                &["instantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authecation"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dominantion"),
                &["domination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assistanse"),
                &["assistants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintiaining"),
                &["maintaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communikays"),
                &["communiqués"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negosiations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexcepted"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifieced"),
                &["sacrificed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailaze"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("neibors"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("direcctorys"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encorporation"),
                &["incorporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monstrostiy"),
                &["monstrosity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intutively"),
                &["intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mininising"),
                &["minimising"],
            ),
            (dictgen::InsensitiveStr::Ascii("utiliza"), &["utilize"]),
            (dictgen::InsensitiveStr::Ascii("skwyard"), &["skyward"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalisitc"),
                &["nationalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comtamination"),
                &["contamination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mangnitudes"),
                &["magnitudes"],
            ),
            (dictgen::InsensitiveStr::Ascii("jepoardy"), &["jeopardy"]),
            (dictgen::InsensitiveStr::Ascii("asterorid"), &["asteroid"]),
            (
                dictgen::InsensitiveStr::Ascii("commect"),
                &["connect", "comment"],
            ),
            (dictgen::InsensitiveStr::Ascii("aeriels"), &["aerials"]),
            (
                dictgen::InsensitiveStr::Ascii("eveluating"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entrepreuner"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("finsishing"), &["finishing"]),
            (
                dictgen::InsensitiveStr::Ascii("discconencting"),
                &["disconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("remmeber"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("authros"), &["authors"]),
            (
                dictgen::InsensitiveStr::Ascii("priortized"),
                &["prioritized"],
            ),
            (dictgen::InsensitiveStr::Ascii("autolaod"), &["autoload"]),
            (dictgen::InsensitiveStr::Ascii("garfeild"), &["garfield"]),
            (
                dictgen::InsensitiveStr::Ascii("aaproximations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atttachments"),
                &["attachments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trotzkyists"),
                &["trotskyists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critising"),
                &["criticising"],
            ),
            (dictgen::InsensitiveStr::Ascii("cace"), &["cache"]),
            (
                dictgen::InsensitiveStr::Ascii("calidating"),
                &["validating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissconnected"),
                &["disconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentitaction"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("anotator"), &["annotator"]),
            (dictgen::InsensitiveStr::Ascii("reosne"), &["reason"]),
            (dictgen::InsensitiveStr::Ascii("reuqested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("itialize"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("metabolisim"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("hedaers"), &["headers"]),
            (
                dictgen::InsensitiveStr::Ascii("transporder"),
                &["transporter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neightborhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgrouund"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indciating"),
                &["indicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("challeges"), &["challenges"]),
            (
                dictgen::InsensitiveStr::Ascii("rudimentory"),
                &["rudimentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("initialy"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("invokee"),
                &["invoked", "invoke"],
            ),
            (dictgen::InsensitiveStr::Ascii("expetced"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("extits"),
                &["exits", "excites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terrerists"),
                &["terrorists"],
            ),
            (dictgen::InsensitiveStr::Ascii("chowsing"), &["choosing"]),
            (dictgen::InsensitiveStr::Ascii("chanegs"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("excludling"), &["excluding"]),
            (
                dictgen::InsensitiveStr::Ascii("fomulating"),
                &["formulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("expertis"), &["experts"]),
            (dictgen::InsensitiveStr::Ascii("compeltly"), &["completely"]),
            (dictgen::InsensitiveStr::Ascii("debbugged"), &["debugged"]),
            (
                dictgen::InsensitiveStr::Ascii("conrrespondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolagising"),
                &["apologising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurently"),
                &["recurrently"],
            ),
            (dictgen::InsensitiveStr::Ascii("interupt"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("regestering"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropolos"),
                &["metropolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependdent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("confiugurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ankshus"), &["anxious"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomring"),
                &["performing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahgners"), &["changers"]),
            (dictgen::InsensitiveStr::Ascii("valkyre"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapeard"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("biginner"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("urethrea"), &["urethra"]),
            (dictgen::InsensitiveStr::Ascii("craceful"), &["graceful"]),
            (
                dictgen::InsensitiveStr::Ascii("posiblity"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("preccedent"), &["precedent"]),
            (
                dictgen::InsensitiveStr::Ascii("aggragates"),
                &["aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrensically"),
                &["intrinsically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrdinate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelly"),
                &["parallelly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mecahnically"),
                &["mechanically"],
            ),
            (dictgen::InsensitiveStr::Ascii("lmiter"), &["limiter"]),
            (
                dictgen::InsensitiveStr::Ascii("opportunaity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveluate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("preceived"), &["perceived"]),
            (
                dictgen::InsensitiveStr::Ascii("appropritaely"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirucumstances"),
                &["circumstances"],
            ),
            (dictgen::InsensitiveStr::Ascii("inventers"), &["inventors"]),
            (dictgen::InsensitiveStr::Ascii("esecute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("districs"), &["districts"]),
            (dictgen::InsensitiveStr::Ascii("acordinng"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("definintive"),
                &["definitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crediential"),
                &["credential"],
            ),
            (dictgen::InsensitiveStr::Ascii("definining"), &["defining"]),
            (
                dictgen::InsensitiveStr::Ascii("temperatues"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("standardss"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("defnition"), &["definition"]),
            (dictgen::InsensitiveStr::Ascii("couuld"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("unmodfied"), &["unmodified"]),
            (
                dictgen::InsensitiveStr::Ascii("restrition"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitutes"),
                &["constitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsrcibers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("claering"), &["clearing"]),
            (dictgen::InsensitiveStr::Ascii("tahnk"), &["thank"]),
            (
                dictgen::InsensitiveStr::Ascii("neccsessarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtoritative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispbibute"),
                &["distribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("chescklist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("usiing"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("mobule"),
                &["module", "mobile"],
            ),
            (dictgen::InsensitiveStr::Ascii("aovoid"), &["avoid"]),
            (dictgen::InsensitiveStr::Ascii("declarted"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("januar"), &["january"]),
            (dictgen::InsensitiveStr::Ascii("liftimes"), &["lifetimes"]),
            (dictgen::InsensitiveStr::Ascii("knowlage"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("seraching"), &["searching"]),
            (
                dictgen::InsensitiveStr::Ascii("justfications"),
                &["justifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("emultor"), &["emulator"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchnous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surreptious"),
                &["surreptitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("metadta"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("quadrupple"), &["quadruple"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchronuous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techniciens"),
                &["technicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("soemhow"), &["somehow"]),
            (
                dictgen::InsensitiveStr::Ascii("demodualtor"),
                &["demodulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("unhealthly"), &["unhealthy"]),
            (dictgen::InsensitiveStr::Ascii("extention"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("pheonixes"), &["phoenixes"]),
            (dictgen::InsensitiveStr::Ascii("uestions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("commnads"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("membersip"), &["membership"]),
            (dictgen::InsensitiveStr::Ascii("platfomrs"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("captability"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaious"), &["various"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbathon"),
                &["masturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dulicated"), &["duplicated"]),
            (dictgen::InsensitiveStr::Ascii("raotations"), &["rotations"]),
            (dictgen::InsensitiveStr::Ascii("accrording"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("developpment"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addes"),
                &["adds", "added", "adders", "address"],
            ),
            (dictgen::InsensitiveStr::Ascii("nortmally"), &["normally"]),
            (dictgen::InsensitiveStr::Ascii("conosle"), &["console"]),
            (dictgen::InsensitiveStr::Ascii("zefers"), &["zephyrs"]),
            (
                dictgen::InsensitiveStr::Ascii("investingated"),
                &["investigated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hietus"), &["hiatus"]),
            (
                dictgen::InsensitiveStr::Ascii("disturbante"),
                &["disturbance"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaaged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("stategised"),
                &["strategised"],
            ),
            (dictgen::InsensitiveStr::Ascii("anphibian"), &["amphibian"]),
            (dictgen::InsensitiveStr::Ascii("repostig"), &["reposting"]),
            (dictgen::InsensitiveStr::Ascii("kidknapped"), &["kidnapped"]),
            (
                dictgen::InsensitiveStr::Ascii("consistuents"),
                &["constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("glyping"), &["glyphing"]),
            (
                dictgen::InsensitiveStr::Ascii("orchestraed"),
                &["orchestrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("transferes"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("eighten"), &["eighteen"]),
            (
                dictgen::InsensitiveStr::Ascii("expermenetel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemped"), &["attempted"]),
            (dictgen::InsensitiveStr::Ascii("impossble"), &["impossible"]),
            (
                dictgen::InsensitiveStr::Ascii("chancellour"),
                &["chancellor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("departement"),
                &["department"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ifnormation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("idetities"), &["identities"]),
            (
                dictgen::InsensitiveStr::Ascii("clairvoiantes"),
                &["clairvoyants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibilitity"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("wille"), &["will"]),
            (dictgen::InsensitiveStr::Ascii("arrenge"), &["arrange"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilties"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirutalise"),
                &["devirtualise"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("udo"),
                &["ado", "judo", "sudo", "udon", "ufo", "undo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenater"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignnore"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("anynomous"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("chcekout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("emenet"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("samwich"), &["sandwich"]),
            (
                dictgen::InsensitiveStr::Ascii("ridiculious"),
                &["ridiculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("commtited"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("exclive"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("desiged"),
                &["designed", "desired"],
            ),
            (dictgen::InsensitiveStr::Ascii("maibox"), &["mailbox"]),
            (
                dictgen::InsensitiveStr::Ascii("independnently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistkake"), &["mistake"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmmetrically"),
                &["asymmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistensy"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("aspahlt"), &["asphalt"]),
            (
                dictgen::InsensitiveStr::Ascii("fallthruogh"),
                &["fallthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("planeswaller"),
                &["planeswalker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyclopidia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spesialisation"),
                &["specialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fursthermore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenatations"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reesrved"), &["reserved"]),
            (
                dictgen::InsensitiveStr::Ascii("trustworhy"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequantly"),
                &["frequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represnetations"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatios"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conetxts"),
                &["contexts", "connects"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoughful"), &["thoughtful"]),
            (dictgen::InsensitiveStr::Ascii("subcript"), &["subscript"]),
            (
                dictgen::InsensitiveStr::Ascii("specailized"),
                &["specialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucccessul"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("konwn"), &["known"]),
            (dictgen::InsensitiveStr::Ascii("reloccated"), &["relocated"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhtmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separatiste"),
                &["separates"],
            ),
            (dictgen::InsensitiveStr::Ascii("guranteed"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("methodd"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("sopunds"), &["sounds"]),
            (
                dictgen::InsensitiveStr::Ascii("templeting"),
                &["templating"],
            ),
            (dictgen::InsensitiveStr::Ascii("undetware"), &["underwear"]),
            (dictgen::InsensitiveStr::Ascii("tabe"), &["table", "tab"]),
            (
                dictgen::InsensitiveStr::Ascii("abstracton"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("formate"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("aracnid"), &["arachnid"]),
            (
                dictgen::InsensitiveStr::Ascii("symphatetic"),
                &["sympathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknwoledged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depentences"),
                &["dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuuity"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redefiniton"),
                &["redefinition"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwaring"), &["forwarding"]),
            (dictgen::InsensitiveStr::Ascii("stroner"), &["stronger"]),
            (dictgen::InsensitiveStr::Ascii("stilus"), &["stylus"]),
            (dictgen::InsensitiveStr::Ascii("chnging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("infomartion"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("viabiliy"), &["viability"]),
            (
                dictgen::InsensitiveStr::Ascii("geographacilly"),
                &["geographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origninating"),
                &["originating"],
            ),
            (dictgen::InsensitiveStr::Ascii("extracing"), &["extracting"]),
            (
                dictgen::InsensitiveStr::Ascii("filessystems"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracterizes"),
                &["characterizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrisincally"),
                &["intrinsically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canniablism"),
                &["cannibalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncahnged"), &["unchanged"]),
            (dictgen::InsensitiveStr::Ascii("conflic"), &["conflict"]),
            (dictgen::InsensitiveStr::Ascii("expantion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("parmsean"), &["parmesan"]),
            (dictgen::InsensitiveStr::Ascii("datatpyes"), &["datatypes"]),
            (dictgen::InsensitiveStr::Ascii("serveral"), &["several"]),
            (
                dictgen::InsensitiveStr::Ascii("comprihend"),
                &["comprehend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititializers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceopt"),
                &["exempt", "except"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostalgica"), &["nostalgia"]),
            (dictgen::InsensitiveStr::Ascii("waviers"), &["waivers"]),
            (dictgen::InsensitiveStr::Ascii("contagios"), &["contagious"]),
            (dictgen::InsensitiveStr::Ascii("dsplaying"), &["displaying"]),
            (dictgen::InsensitiveStr::Ascii("dynaminc"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("stablizie"), &["stabilize"]),
            (
                dictgen::InsensitiveStr::Ascii("annihliated"),
                &["annihilated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degradating"),
                &["degradation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practioner"),
                &["practitioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("acumulate"), &["accumulate"]),
            (dictgen::InsensitiveStr::Ascii("remmve"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("proffes"), &["profess"]),
            (dictgen::InsensitiveStr::Ascii("tehre"), &["there"]),
            (dictgen::InsensitiveStr::Ascii("surveill"), &["surveil"]),
            (
                dictgen::InsensitiveStr::Ascii("asdignments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("zaelots"), &["zealots"]),
            (
                dictgen::InsensitiveStr::Ascii("corintheans"),
                &["corinthians"],
            ),
            (dictgen::InsensitiveStr::Ascii("discconet"), &["disconnect"]),
            (dictgen::InsensitiveStr::Ascii("avengence"), &["vengeance"]),
            (dictgen::InsensitiveStr::Ascii("argments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("formaths"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("diffference"),
                &["difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("lithuanina"), &["lithuania"]),
            (
                dictgen::InsensitiveStr::Ascii("desperetly"),
                &["desperately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acculumate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reporitory"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("liuke"), &["like"]),
            (
                dictgen::InsensitiveStr::Ascii("descritptor"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condolencies"),
                &["condolences"],
            ),
            (dictgen::InsensitiveStr::Ascii("attacehd"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("acitivty"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("symbsols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("tje"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinete"),
                &["indefinite"],
            ),
            (dictgen::InsensitiveStr::Ascii("calssify"), &["classify"]),
            (dictgen::InsensitiveStr::Ascii("prcoessed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("conflits"), &["conflicts"]),
            (
                dictgen::InsensitiveStr::Ascii("rewinded"),
                &["rewound", "reminded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concurment"),
                &["concurrent"],
            ),
            (dictgen::InsensitiveStr::Ascii("disposin"), &["disposing"]),
            (dictgen::InsensitiveStr::Ascii("ithe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("transfere"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("visti"), &["visit"]),
            (dictgen::InsensitiveStr::Ascii("excludle"), &["exclude"]),
            (dictgen::InsensitiveStr::Ascii("identty"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiifers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("departmet"),
                &["departments"],
            ),
            (dictgen::InsensitiveStr::Ascii("extaction"), &["extraction"]),
            (
                dictgen::InsensitiveStr::Ascii("conditionnally"),
                &["conditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("erminated"), &["terminated"]),
            (dictgen::InsensitiveStr::Ascii("wrtie"), &["write"]),
            (dictgen::InsensitiveStr::Ascii("paragraps"), &["paragraphs"]),
            (dictgen::InsensitiveStr::Ascii("cahgnes"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("wishlit"), &["wishlist"]),
            (
                dictgen::InsensitiveStr::Ascii("envrionmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("prelayed"), &["replayed"]),
            (
                dictgen::InsensitiveStr::Ascii("accientally"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("locatins"), &["locations"]),
            (
                dictgen::InsensitiveStr::Ascii("explisitly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("reservin"), &["reserving"]),
            (
                dictgen::InsensitiveStr::Ascii("instansiate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("blankes"), &["blankets"]),
            (dictgen::InsensitiveStr::Ascii("buddist"), &["buddhist"]),
            (dictgen::InsensitiveStr::Ascii("continers"), &["containers"]),
            (
                dictgen::InsensitiveStr::Ascii("verfification"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("frgament"), &["fragment"]),
            (
                dictgen::InsensitiveStr::Ascii("infecteous"),
                &["infectious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counsilers"),
                &["counselors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreaction"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("departue"), &["departure"]),
            (
                dictgen::InsensitiveStr::Ascii("cutomizations"),
                &["customizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("escalatin"),
                &["escalation", "escalating"],
            ),
            (dictgen::InsensitiveStr::Ascii("popoen"), &["popen"]),
            (
                dictgen::InsensitiveStr::Ascii("alloated"),
                &["allocated", "allotted"],
            ),
            (dictgen::InsensitiveStr::Ascii("effforts"), &["efforts"]),
            (dictgen::InsensitiveStr::Ascii("hurricans"), &["hurricanes"]),
            (dictgen::InsensitiveStr::Ascii("handerler"), &["handler"]),
            (
                dictgen::InsensitiveStr::Ascii("dependdently"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritities"),
                &["priorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partitinioning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("classroms"), &["classrooms"]),
            (
                dictgen::InsensitiveStr::Ascii("proporties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woarkaround"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reposonding"),
                &["responding"],
            ),
            (dictgen::InsensitiveStr::Ascii("protecion"), &["protection"]),
            (
                dictgen::InsensitiveStr::Ascii("subsitututes"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costums"),
                &["costumes", "customs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaminatie"),
                &["contaminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporachable"),
                &["approachable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentatation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("initailly"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("vaues"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("exhanging"), &["exchanging"]),
            (
                dictgen::InsensitiveStr::Ascii("dedupliate"),
                &["deduplicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mershall"), &["marshall"]),
            (dictgen::InsensitiveStr::Ascii("refacted"), &["refracted"]),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedora"),
                &["tetrahedra"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribude"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("tempermeant"),
                &["temperament"],
            ),
            (dictgen::InsensitiveStr::Ascii("redesings"), &["redesigns"]),
            (
                dictgen::InsensitiveStr::Ascii("athletisicm"),
                &["athleticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assigne"),
                &["assign", "assigned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannibalizin"),
                &["cannibalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegoitations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatening"),
                &["concatenating"],
            ),
            (dictgen::InsensitiveStr::Ascii("eatting"), &["eating"]),
            (
                dictgen::InsensitiveStr::Ascii("perisisting"),
                &["persisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuptable"), &["reputable"]),
            (
                dictgen::InsensitiveStr::Ascii("horozontally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresspond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("incluedd"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("initaite"), &["initiate"]),
            (
                dictgen::InsensitiveStr::Ascii("funktioning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("folllowiwng"),
                &["following"],
            ),
            (dictgen::InsensitiveStr::Ascii("owerread"), &["overread"]),
            (
                dictgen::InsensitiveStr::Ascii("adjascently"),
                &["adjacently"],
            ),
            (dictgen::InsensitiveStr::Ascii("traker"), &["tracker"]),
            (dictgen::InsensitiveStr::Ascii("conern"), &["concern"]),
            (dictgen::InsensitiveStr::Ascii("chainge"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("restoding"), &["restoring"]),
            (dictgen::InsensitiveStr::Ascii("pressented"), &["presented"]),
            (
                dictgen::InsensitiveStr::Ascii("maintentance"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("retarted"), &["restarted"]),
            (dictgen::InsensitiveStr::Ascii("regularis"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("comecting"), &["connecting"]),
            (
                dictgen::InsensitiveStr::Ascii("micorcenter"),
                &["microcenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplate"),
                &["deduplicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisic"),
                &["characteristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiting"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("golitah"), &["goliath"]),
            (dictgen::InsensitiveStr::Ascii("behavriour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("encouters"), &["encounters"]),
            (
                dictgen::InsensitiveStr::Ascii("miscommunicatie"),
                &["miscommunication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resurected"),
                &["resurrected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmiations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitalise"),
                &["reinitialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissctor"), &["dissector"]),
            (dictgen::InsensitiveStr::Ascii("somtimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("eliminase"), &["eliminates"]),
            (
                dictgen::InsensitiveStr::Ascii("availabillity"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("corretes"), &["correlates"]),
            (dictgen::InsensitiveStr::Ascii("queris"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("intertvined"),
                &["intertwined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighrourhoods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("chassy"), &["chassis"]),
            (
                dictgen::InsensitiveStr::Ascii("mathetician"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("satsify"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("mohammedans"), &["muslims"]),
            (
                dictgen::InsensitiveStr::Ascii("narrativas"),
                &["narratives"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceepter"), &["accepter"]),
            (dictgen::InsensitiveStr::Ascii("knowlegde"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangent"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("catloags"), &["catalogs"]),
            (dictgen::InsensitiveStr::Ascii("commannds"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("trakcers"), &["trackers"]),
            (dictgen::InsensitiveStr::Ascii("materias"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("proportionallity"),
                &["proportionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("simluated"), &["simulated"]),
            (
                dictgen::InsensitiveStr::Ascii("wordarounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasverses"),
                &["traverses", "transverses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relatibely"),
                &["relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapprovin"),
                &["disapproving"],
            ),
            (dictgen::InsensitiveStr::Ascii("copenhagn"), &["copenhagen"]),
            (dictgen::InsensitiveStr::Ascii("playstlye"), &["playstyle"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchonously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("igore"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("erasuer"),
                &["erasure", "eraser"],
            ),
            (dictgen::InsensitiveStr::Ascii("glitchs"), &["glitches"]),
            (dictgen::InsensitiveStr::Ascii("compling"), &["compiling"]),
            (dictgen::InsensitiveStr::Ascii("enchantig"), &["enchanting"]),
            (dictgen::InsensitiveStr::Ascii("aiffer"), &["differ"]),
            (
                dictgen::InsensitiveStr::Ascii("deferentiator"),
                &["differentiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("besic"), &["basic"]),
            (dictgen::InsensitiveStr::Ascii("exchenged"), &["exchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("approoximately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conferene"),
                &["conferences", "conference"],
            ),
            (dictgen::InsensitiveStr::Ascii("frowraded"), &["forwarded"]),
            (
                dictgen::InsensitiveStr::Ascii("custamized"),
                &["customized"],
            ),
            (dictgen::InsensitiveStr::Ascii("unusal"), &["unusual"]),
            (
                dictgen::InsensitiveStr::Ascii("masturabtion"),
                &["masturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("preperint"), &["preprint"]),
            (
                dictgen::InsensitiveStr::Ascii("cirticising"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intallation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("signiture"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("syntesizing"),
                &["synthesizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("signall"), &["signal"]),
            (dictgen::InsensitiveStr::Ascii("availavle"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("nam"), &["name"]),
            (dictgen::InsensitiveStr::Ascii("emtpy"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("turstworthy"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulbearabule"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("necesarry"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("spcifications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("vamipres"), &["vampires"]),
            (
                dictgen::InsensitiveStr::Ascii("prerelased"),
                &["prereleased"],
            ),
            (dictgen::InsensitiveStr::Ascii("kyrillic"), &["cyrillic"]),
            (dictgen::InsensitiveStr::Ascii("disutils"), &["distutils"]),
            (
                dictgen::InsensitiveStr::Ascii("prematurelly"),
                &["prematurely"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequenzer"), &["sequencer"]),
            (dictgen::InsensitiveStr::Ascii("caputred"), &["captured"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretatons"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weightlifitng"),
                &["weightlifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("enlighment"), &["enlighten"]),
            (
                dictgen::InsensitiveStr::Ascii("exracting"),
                &["extracting", "exacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("bussines"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("distincitve"),
                &["distinctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temperement"),
                &["temperament"],
            ),
            (dictgen::InsensitiveStr::Ascii("facilitiy"), &["facility"]),
            (dictgen::InsensitiveStr::Ascii("outweight"), &["outweigh"]),
            (dictgen::InsensitiveStr::Ascii("generaor"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("charcteristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scientiests"),
                &["scientists"],
            ),
            (dictgen::InsensitiveStr::Ascii("oject"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("strkethrough"),
                &["strikethrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatablities"),
                &["incompatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigfault"), &["segfault"]),
            (dictgen::InsensitiveStr::Ascii("confingure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("saleries"), &["salaries"]),
            (
                dictgen::InsensitiveStr::Ascii("namesapces"),
                &["namespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("tawainese"), &["taiwanese"]),
            (
                dictgen::InsensitiveStr::Ascii("analyitcaly"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("whos"), &["whose"]),
            (
                dictgen::InsensitiveStr::Ascii("exepectation"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("execellent"), &["excellent"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomenace"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("blacklit"), &["blacklist"]),
            (
                dictgen::InsensitiveStr::Ascii("singnificantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("poeoples"), &["peoples"]),
            (
                dictgen::InsensitiveStr::Ascii("adivce"),
                &["advice", "advise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bortherhood"),
                &["brotherhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countepart"),
                &["counteract"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporiate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encapsualtion"),
                &["encapsulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rickoshaying"),
                &["ricocheting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attirbutes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disucussion"),
                &["discussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveneince"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("stabalize"), &["stabilize"]),
            (dictgen::InsensitiveStr::Ascii("initiale"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("umbrealla"), &["umbrella"]),
            (
                dictgen::InsensitiveStr::Ascii("ostrasized"),
                &["ostracized"],
            ),
            (dictgen::InsensitiveStr::Ascii("gotee"), &["goatee"]),
            (dictgen::InsensitiveStr::Ascii("variabels"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("kubenertes"),
                &["kubernetes"],
            ),
            (dictgen::InsensitiveStr::Ascii("comminism"), &["communism"]),
            (
                dictgen::InsensitiveStr::Ascii("appearnace"),
                &["appearances", "appearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("disovered"), &["discovered"]),
            (
                dictgen::InsensitiveStr::Ascii("unintuitve"),
                &["unintuitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("euqations"), &["equations"]),
            (
                dictgen::InsensitiveStr::Ascii("rememembers"),
                &["remembers"],
            ),
            (dictgen::InsensitiveStr::Ascii("nombering"), &["numbering"]),
            (
                dictgen::InsensitiveStr::Ascii("corospondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("faukty"), &["faulty"]),
            (dictgen::InsensitiveStr::Ascii("scaleable"), &["scalable"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetemine"),
                &["amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangmeent"),
                &["arrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexcected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialsiation"),
                &["deserialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntraining"),
                &["constraining"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwwong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("antennnas"), &["antennas"]),
            (
                dictgen::InsensitiveStr::Ascii("constrints"),
                &["constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsit"), &["exist", "exit"]),
            (
                dictgen::InsensitiveStr::Ascii("deativation"),
                &["deactivation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("veritcally"),
                &["vertically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarentining"),
                &["quarantining"],
            ),
            (dictgen::InsensitiveStr::Ascii("nonesense"), &["nonsense"]),
            (dictgen::InsensitiveStr::Ascii("earnt"), &["earned"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangmeents"),
                &["arrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avantageous"),
                &["advantageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanetizing"),
                &["sanitizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contravercies"),
                &["controversies"],
            ),
            (dictgen::InsensitiveStr::Ascii("egenral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("theraphy"), &["therapy"]),
            (
                dictgen::InsensitiveStr::Ascii("appericated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneartor"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("termianl"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("switchter"), &["switcher"]),
            (
                dictgen::InsensitiveStr::Ascii("prosperety"),
                &["prosperity"],
            ),
            (dictgen::InsensitiveStr::Ascii("automtic"), &["automatic"]),
            (dictgen::InsensitiveStr::Ascii("realate"), &["relate"]),
            (
                dictgen::InsensitiveStr::Ascii("propertion"),
                &["proportion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subfromats"),
                &["subformats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertatianism"),
                &["libertarianism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shakespare"),
                &["shakespeare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typdef"),
                &["typed", "typedef"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avisoriy"),
                &["advisory", "advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("trinekts"), &["trinkets"]),
            (dictgen::InsensitiveStr::Ascii("sketchey"), &["sketchy"]),
            (dictgen::InsensitiveStr::Ascii("reponses"), &["responses"]),
            (dictgen::InsensitiveStr::Ascii("condensend"), &["condensed"]),
            (
                dictgen::InsensitiveStr::Ascii("individualls"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questonable"),
                &["questionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("paymnet"), &["payment"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatly"),
                &["alternately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differenciated"),
                &["differentiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovewrote"), &["overwrote"]),
            (
                dictgen::InsensitiveStr::Ascii("monumentais"),
                &["monuments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchroniously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nighborhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprocible"),
                &["reproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disssociated"),
                &["dissociated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclockign"),
                &["overclocking"],
            ),
            (dictgen::InsensitiveStr::Ascii("asserive"), &["assertive"]),
            (
                dictgen::InsensitiveStr::Ascii("paschurised"),
                &["pasteurised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("livescream"),
                &["livestream"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspinding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fraudulant"),
                &["fraudulent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificatin"),
                &["certificating", "certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("endianess"), &["endianness"]),
            (dictgen::InsensitiveStr::Ascii("vallue"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("quyality"), &["quality"]),
            (
                dictgen::InsensitiveStr::Ascii("philosophicaly"),
                &["philosophically"],
            ),
            (dictgen::InsensitiveStr::Ascii("inheritted"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("diolog"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("instabillity"),
                &["instability"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkloads"), &["workloads"]),
            (dictgen::InsensitiveStr::Ascii("squadroon"), &["squadron"]),
            (dictgen::InsensitiveStr::Ascii("explitit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("contstrained"),
                &["constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succedding"),
                &["succeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitabillity"),
                &["profitability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilazation"),
                &["capitalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("simuators"), &["simulators"]),
            (
                dictgen::InsensitiveStr::Ascii("prodceding"),
                &["proceeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("evindence"), &["evidence"]),
            (dictgen::InsensitiveStr::Ascii("bugdets"), &["budgets"]),
            (dictgen::InsensitiveStr::Ascii("ejacualte"), &["ejaculate"]),
            (dictgen::InsensitiveStr::Ascii("knwo"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("oldes"), &["oldest"]),
            (dictgen::InsensitiveStr::Ascii("oppenents"), &["opponents"]),
            (dictgen::InsensitiveStr::Ascii("signins"), &["signings"]),
            (dictgen::InsensitiveStr::Ascii("preparin"), &["preparing"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditionnaly"),
                &["unconditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfix"), &["prefix"]),
            (dictgen::InsensitiveStr::Ascii("vhild"), &["child"]),
            (dictgen::InsensitiveStr::Ascii("certaing"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("consentration"),
                &["concentration"],
            ),
            (dictgen::InsensitiveStr::Ascii("expolit"), &["exploit"]),
            (dictgen::InsensitiveStr::Ascii("indiania"), &["indiana"]),
            (dictgen::InsensitiveStr::Ascii("specias"), &["specials"]),
            (dictgen::InsensitiveStr::Ascii("coinsided"), &["coincided"]),
            (dictgen::InsensitiveStr::Ascii("fugures"), &["figures"]),
            (
                dictgen::InsensitiveStr::Ascii("separatring"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannabilyzing"),
                &["cannibalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprecaited"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticdote"),
                &["anecdote", "antidote"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectatular"),
                &["spectacular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirimentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transtions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("crossfie"), &["crossfire"]),
            (
                dictgen::InsensitiveStr::Ascii("catapillars"),
                &["caterpillars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubelshooting"),
                &["troubleshooting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphreals"),
                &["peripherals"],
            ),
            (dictgen::InsensitiveStr::Ascii("detenction"), &["detection"]),
            (
                dictgen::InsensitiveStr::Ascii("thatnkfully"),
                &["thankfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profides"),
                &["provides", "profiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gunatanamo"),
                &["guantanamo"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatiling"), &["detailing"]),
            (
                dictgen::InsensitiveStr::Ascii("exlpicitly"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("struggeling"),
                &["struggling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congratualting"),
                &["congratulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstructing"),
                &["abstracting", "obstructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("faillure"), &["failure"]),
            (
                dictgen::InsensitiveStr::Ascii("cocnurency"),
                &["concurrency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuais"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fabricatie"),
                &["fabricated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creadentialed"),
                &["credentialed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sycronizes"),
                &["synchronizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customsiers"),
                &["customisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particualrly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miliciousally"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiened"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("fintie"), &["finite"]),
            (dictgen::InsensitiveStr::Ascii("regester"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("afeter"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("professioanl"),
                &["professionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("franchsies"),
                &["franchises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unaccesible"),
                &["inaccessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("cachline"), &["cacheline"]),
            (
                dictgen::InsensitiveStr::Ascii("vists"),
                &["fists", "lists", "visits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruptablity"),
                &["interruptibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undecideable"),
                &["undecidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conributions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("toiletts"), &["toilets"]),
            (dictgen::InsensitiveStr::Ascii("hourlgass"), &["hourglass"]),
            (
                dictgen::InsensitiveStr::Ascii("tablepsoons"),
                &["tablespoons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerbly"),
                &["considerably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descirptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instll"),
                &["install", "instill"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumamry"), &["summary"]),
            (dictgen::InsensitiveStr::Ascii("assitants"), &["assistants"]),
            (
                dictgen::InsensitiveStr::Ascii("liminations"),
                &["limitations", "laminations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loged"),
                &["logged", "lodged", "longed"],
            ),
            (dictgen::InsensitiveStr::Ascii("gimmicy"), &["gimmicky"]),
            (
                dictgen::InsensitiveStr::Ascii("cirumstances"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certitication"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misconecption"),
                &["misconceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenative"), &["tentative"]),
            (dictgen::InsensitiveStr::Ascii("cylcic"), &["cyclic"]),
            (
                dictgen::InsensitiveStr::Ascii("diminising"),
                &["diminishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("focuss"), &["focus"]),
            (dictgen::InsensitiveStr::Ascii("avacado"), &["avocado"]),
            (
                dictgen::InsensitiveStr::Ascii("relm"),
                &["elm", "helm", "realm", "ream", "rem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplinera"),
                &["disciplinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("appenging"), &["appending"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonaire"),
                &["opinionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexadeciaml"),
                &["hexadecimal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increamentally"),
                &["incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("feminitity"),
                &["femininity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiasum"),
                &["enthusiasm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundemantals"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misundertsanding"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("milimeters"),
                &["millimeters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertiies"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("resert"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstale"),
                &["reinstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("needlesslly"),
                &["needlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootstap"), &["bootstrap"]),
            (dictgen::InsensitiveStr::Ascii("turorials"), &["tutorials"]),
            (
                dictgen::InsensitiveStr::Ascii("convinsing"),
                &["convincing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predoiminates"),
                &["predominates"],
            ),
            (dictgen::InsensitiveStr::Ascii("excesvly"), &["excessively"]),
            (dictgen::InsensitiveStr::Ascii("firrst"), &["first"]),
            (
                dictgen::InsensitiveStr::Ascii("trucnations"),
                &["truncations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufication"),
                &["suffocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbiitrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("reistence"), &["resistance"]),
            (
                dictgen::InsensitiveStr::Ascii("clocksourc"),
                &["clocksource"],
            ),
            (dictgen::InsensitiveStr::Ascii("rappellin"), &["rappelling"]),
            (dictgen::InsensitiveStr::Ascii("faktored"), &["factored"]),
            (dictgen::InsensitiveStr::Ascii("redircet"), &["redirect"]),
            (
                dictgen::InsensitiveStr::Ascii("perpatuating"),
                &["perpetuating"],
            ),
            (dictgen::InsensitiveStr::Ascii("changge"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("decroation"),
                &["decoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("mecahnic"), &["mechanic"]),
            (
                dictgen::InsensitiveStr::Ascii("domesticted"),
                &["domesticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbiatraily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configaration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investegating"),
                &["investigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cramugin"), &["curmudgeon"]),
            (dictgen::InsensitiveStr::Ascii("combacks"), &["comebacks"]),
            (dictgen::InsensitiveStr::Ascii("penindg"), &["pending"]),
            (dictgen::InsensitiveStr::Ascii("regiser"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("becuae"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("disscused"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("uplfiting"), &["uplifting"]),
            (dictgen::InsensitiveStr::Ascii("vacumms"), &["vacuums"]),
            (dictgen::InsensitiveStr::Ascii("compelx"), &["complex"]),
            (dictgen::InsensitiveStr::Ascii("theri"), &["their", "there"]),
            (dictgen::InsensitiveStr::Ascii("ignornable"), &["ignorable"]),
            (dictgen::InsensitiveStr::Ascii("archiects"), &["architects"]),
            (
                dictgen::InsensitiveStr::Ascii("disssociates"),
                &["dissociates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimaneted"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissmissing"),
                &["dismissing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tontiguous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconfortable"),
                &["uncomfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thernodynamics"),
                &["thermodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decsribing"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mannually"), &["manually"]),
            (dictgen::InsensitiveStr::Ascii("ulteriour"), &["ulterior"]),
            (
                dictgen::InsensitiveStr::Ascii("teathered"),
                &["tethered", "feathered", "weathered"],
            ),
            (dictgen::InsensitiveStr::Ascii("toxiticy"), &["toxicity"]),
            (dictgen::InsensitiveStr::Ascii("instrinsic"), &["intrinsic"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinators"),
                &["terminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("allaince"), &["alliance"]),
            (dictgen::InsensitiveStr::Ascii("disagres"), &["disagrees"]),
            (
                dictgen::InsensitiveStr::Ascii("syptomatic"),
                &["symptomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("airrcraft"), &["aircraft"]),
            (dictgen::InsensitiveStr::Ascii("indever"), &["endeavor"]),
            (
                dictgen::InsensitiveStr::Ascii("adaquately"),
                &["adequately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overmapping"),
                &["overlapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("delcining"), &["declining"]),
            (
                dictgen::InsensitiveStr::Ascii("positiviely"),
                &["positively"],
            ),
            (dictgen::InsensitiveStr::Ascii("pateince"), &["patience"]),
            (dictgen::InsensitiveStr::Ascii("paernt"), &["parent"]),
            (dictgen::InsensitiveStr::Ascii("overiding"), &["overriding"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstitue"),
                &["reconstitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("dipsosing"), &["disposing"]),
            (dictgen::InsensitiveStr::Ascii("preapares"), &["prepares"]),
            (
                dictgen::InsensitiveStr::Ascii("comparabil"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("autherize"), &["authorize"]),
            (dictgen::InsensitiveStr::Ascii("protestas"), &["protests"]),
            (dictgen::InsensitiveStr::Ascii("coexsit"), &["coexist"]),
            (dictgen::InsensitiveStr::Ascii("potentital"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("evaliators"),
                &["evaluators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insistince"),
                &["insistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatanating"),
                &["concatenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acommodating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vectrors"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("norwiegan"), &["norwegian"]),
            (dictgen::InsensitiveStr::Ascii("mechanisim"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("negociator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("swticher"), &["switcher"]),
            (dictgen::InsensitiveStr::Ascii("complated"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("extrememly"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("materilize"),
                &["materialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nead"),
                &["need", "head", "knead"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forcefullly"),
                &["forcefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("thimngs"), &["things"]),
            (
                dictgen::InsensitiveStr::Ascii("abstructed"),
                &["abstracted", "obstructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("pont"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoteros"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimute"), &["minute"]),
            (dictgen::InsensitiveStr::Ascii("scrits"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("transters"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("defectus"), &["defects"]),
            (dictgen::InsensitiveStr::Ascii("concreet"), &["concrete"]),
            (
                dictgen::InsensitiveStr::Ascii("resemblense"),
                &["resembles"],
            ),
            (dictgen::InsensitiveStr::Ascii("feminsim"), &["feminism"]),
            (
                dictgen::InsensitiveStr::Ascii("deamonized"),
                &["daemonized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unecesarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disssembled"),
                &["disassembled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fomulation"),
                &["formulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colonialsm"),
                &["colonialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprasal"), &["appraisal"]),
            (
                dictgen::InsensitiveStr::Ascii("autonagotiation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplifer"), &["simplifier"]),
            (
                dictgen::InsensitiveStr::Ascii("availablility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kenels"),
                &["kernels", "kennels"],
            ),
            (dictgen::InsensitiveStr::Ascii("analityc"), &["analytic"]),
            (dictgen::InsensitiveStr::Ascii("atalog"), &["catalog"]),
            (dictgen::InsensitiveStr::Ascii("sirectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("unexcectedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("durationn"), &["duration"]),
            (dictgen::InsensitiveStr::Ascii("introdues"), &["introduces"]),
            (dictgen::InsensitiveStr::Ascii("opiton"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("unsupproted"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("templaced"), &["templated"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("poitner"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("cyrstalline"),
                &["crystalline"],
            ),
            (dictgen::InsensitiveStr::Ascii("justfying"), &["justifying"]),
            (
                dictgen::InsensitiveStr::Ascii("comporators"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("pattterns"), &["patterns"]),
            (dictgen::InsensitiveStr::Ascii("panmedic"), &["pandemic"]),
            (
                dictgen::InsensitiveStr::Ascii("effortlessy"),
                &["effortlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invulnerbale"),
                &["invulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sligt"), &["slight"]),
            (dictgen::InsensitiveStr::Ascii("seprators"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("depreceated"),
                &["deprecated", "depreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avaliability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("amazaing"), &["amazing"]),
            (dictgen::InsensitiveStr::Ascii("resoreces"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("paltette"), &["palette"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratrions"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("colmun"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("executng"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("camparative"),
                &["comparative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepenendent"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("depnd"), &["depend"]),
            (
                dictgen::InsensitiveStr::Ascii("delaership"),
                &["dealership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosohpies"),
                &["philosophies"],
            ),
            (dictgen::InsensitiveStr::Ascii("opperate"), &["operate"]),
            (
                dictgen::InsensitiveStr::Ascii("consitututing"),
                &["constituting"],
            ),
            (dictgen::InsensitiveStr::Ascii("totatl"), &["total"]),
            (
                dictgen::InsensitiveStr::Ascii("intimidacion"),
                &["intimidation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparizons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgrating"),
                &["downgrading"],
            ),
            (dictgen::InsensitiveStr::Ascii("scandalosi"), &["scandals"]),
            (dictgen::InsensitiveStr::Ascii("mentoned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("callibrates"),
                &["calibrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncicalizes"),
                &["canonicalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstuct"),
                &["reconstruct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("princepals"),
                &["principals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalulations"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bascktrack"), &["backtrack"]),
            (dictgen::InsensitiveStr::Ascii("unnown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("compromized"),
                &["compromised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegoatiation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("enlsaved"), &["enslaved"]),
            (
                dictgen::InsensitiveStr::Ascii("accomplisses"),
                &["accomplishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("switiching"), &["switching"]),
            (dictgen::InsensitiveStr::Ascii("apporach"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("absoltely"), &["absolutely"]),
            (dictgen::InsensitiveStr::Ascii("assignmet"), &["assignment"]),
            (
                dictgen::InsensitiveStr::Ascii("cloisonays"),
                &["cloisonnés"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprot"), &["export"]),
            (dictgen::InsensitiveStr::Ascii("deselct"), &["deselect"]),
            (dictgen::InsensitiveStr::Ascii("recommds"), &["recommends"]),
            (
                dictgen::InsensitiveStr::Ascii("delpoyments"),
                &["deployments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mensions"),
                &["mansions", "mentions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controveries"),
                &["controversies"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrete"), &["concrete"]),
            (dictgen::InsensitiveStr::Ascii("opactity"), &["opacity"]),
            (dictgen::InsensitiveStr::Ascii("carcus"), &["carcass"]),
            (dictgen::InsensitiveStr::Ascii("anuling"), &["annulling"]),
            (
                dictgen::InsensitiveStr::Ascii("horizontallly"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assicietion"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socialistes"),
                &["socialists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreegators"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simptomaticlly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cupbaord"), &["cupboard"]),
            (
                dictgen::InsensitiveStr::Ascii("deathmacth"),
                &["deathmatch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internt"),
                &["internet", "intent", "intern"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecctions"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trumendously"),
                &["tremendously"],
            ),
            (dictgen::InsensitiveStr::Ascii("modyfier"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("mentionnned"),
                &["mentioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absodeflyly"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("educacional"),
                &["educational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preciding"),
                &["preceding", "presiding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("landacapes"),
                &["landscapes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whsipering"),
                &["whispering"],
            ),
            (dictgen::InsensitiveStr::Ascii("decalared"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("recive"), &["receive"]),
            (
                dictgen::InsensitiveStr::Ascii("priorising"),
                &["prioritising"],
            ),
            (dictgen::InsensitiveStr::Ascii("herarchy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("operatings"),
                &["operations", "operating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigativo"),
                &["investigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sertificated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("yello"), &["yellow"]),
            (dictgen::InsensitiveStr::Ascii("bengeing"), &["bingeing"]),
            (dictgen::InsensitiveStr::Ascii("striaghts"), &["straights"]),
            (dictgen::InsensitiveStr::Ascii("foreces"), &["forces"]),
            (
                dictgen::InsensitiveStr::Ascii("arametrically"),
                &["parametrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoulndt"), &["shouldnt"]),
            (dictgen::InsensitiveStr::Ascii("enterred"), &["entered"]),
            (
                dictgen::InsensitiveStr::Ascii("underlfows"),
                &["underflows"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoutely"), &["absolutely"]),
            (
                dictgen::InsensitiveStr::Ascii("wokrstation"),
                &["workstation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representas"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaginativo"),
                &["imagination"],
            ),
            (dictgen::InsensitiveStr::Ascii("chiziling"), &["chiseling"]),
            (
                dictgen::InsensitiveStr::Ascii("manafestation"),
                &["manifestation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bandwdith"), &["bandwidth"]),
            (
                dictgen::InsensitiveStr::Ascii("interpeter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthenticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("shapin"), &["shaping"]),
            (
                dictgen::InsensitiveStr::Ascii("despectively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("stubbs"), &["stubs"]),
            (
                dictgen::InsensitiveStr::Ascii("unprecendented"),
                &["unprecedented"],
            ),
            (dictgen::InsensitiveStr::Ascii("formt"), &["format", "form"]),
            (
                dictgen::InsensitiveStr::Ascii("altruisitic"),
                &["altruistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicate"), &["dictate"]),
            (dictgen::InsensitiveStr::Ascii("lond"), &["long"]),
            (
                dictgen::InsensitiveStr::Ascii("numearation"),
                &["numeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("uestioned"), &["questioned"]),
            (dictgen::InsensitiveStr::Ascii("agruements"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("compilicated"),
                &["complicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininistrator"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolvte"), &["absolve"]),
            (
                dictgen::InsensitiveStr::Ascii("destributes"),
                &["distributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chanc"),
                &["chance", "chunk"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monitoreado"),
                &["monitored"],
            ),
            (dictgen::InsensitiveStr::Ascii("fallbacl"), &["fallback"]),
            (
                dictgen::InsensitiveStr::Ascii("questionne"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("yosemitie"), &["yosemite"]),
            (dictgen::InsensitiveStr::Ascii("emodifying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("namess"), &["names"]),
            (
                dictgen::InsensitiveStr::Ascii("transofrming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertaryanism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("anddroid"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("dilligent"), &["diligent"]),
            (dictgen::InsensitiveStr::Ascii("anwsered"), &["answered"]),
            (
                dictgen::InsensitiveStr::Ascii("inequallity"),
                &["inequality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabibility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkcpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("synonymis"), &["synonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("interacte"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkis"), &["checks"]),
            (dictgen::InsensitiveStr::Ascii("submittin"), &["submitting"]),
            (
                dictgen::InsensitiveStr::Ascii("negotators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("soley"), &["solely"]),
            (dictgen::InsensitiveStr::Ascii("armstrog"), &["armstrong"]),
            (
                dictgen::InsensitiveStr::Ascii("pefect"),
                &["perfect", "prefect", "defect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commections"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalueted"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("ligts"),
                &["lights", "lifts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbriviate"),
                &["abbreviate"],
            ),
            (dictgen::InsensitiveStr::Ascii("handedlly"), &["handedly"]),
            (dictgen::InsensitiveStr::Ascii("frok"), &["from"]),
            (dictgen::InsensitiveStr::Ascii("snythesis"), &["synthesis"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvineance"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("shif"), &["shift"]),
            (dictgen::InsensitiveStr::Ascii("someting"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("discriptor"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("midifiers"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhaagen"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("gir"), &["git"]),
            (dictgen::InsensitiveStr::Ascii("rythems"), &["rhythms"]),
            (dictgen::InsensitiveStr::Ascii("forefit"), &["forfeit"]),
            (
                dictgen::InsensitiveStr::Ascii("suscessfull"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sunconsciously"),
                &["subconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desgustingly"),
                &["disgustingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constuctions"),
                &["constructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassembled"),
                &["disassembled"],
            ),
            (dictgen::InsensitiveStr::Ascii("hardocde"), &["hardcode"]),
            (dictgen::InsensitiveStr::Ascii("epihpany"), &["epiphany"]),
            (dictgen::InsensitiveStr::Ascii("anyother"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("interviening"),
                &["interviewing"],
            ),
            (dictgen::InsensitiveStr::Ascii("preocessor"), &["processor"]),
            (
                dictgen::InsensitiveStr::Ascii("achivements"),
                &["achievements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowledgebly"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validade"),
                &["validated", "validate"],
            ),
            (dictgen::InsensitiveStr::Ascii("leats"), &["least"]),
            (
                dictgen::InsensitiveStr::Ascii("kidnappning"),
                &["kidnapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlcudes"), &["excludes"]),
            (
                dictgen::InsensitiveStr::Ascii("unforutunately"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("glitchd"), &["glitched"]),
            (
                dictgen::InsensitiveStr::Ascii("theoratical"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("bunlder"), &["bundler"]),
            (
                dictgen::InsensitiveStr::Ascii("narl"),
                &[
                    "earl", "farl", "gnarl", "marl", "nail", "nark", "nary", "snarl",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("wheres"), &["whereas"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wrokstation"),
                &["workstation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignord"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("sugggested"), &["suggested"]),
            (
                dictgen::InsensitiveStr::Ascii("indonisian"),
                &["indonesian"],
            ),
            (dictgen::InsensitiveStr::Ascii("asterick"), &["asterisk"]),
            (dictgen::InsensitiveStr::Ascii("woud"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("absense"), &["absence"]),
            (dictgen::InsensitiveStr::Ascii("centuties"), &["centuries"]),
            (dictgen::InsensitiveStr::Ascii("camperas"), &["campers"]),
            (dictgen::InsensitiveStr::Ascii("unsupress"), &["unsuppress"]),
            (
                dictgen::InsensitiveStr::Ascii("implementacion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("editers"),
                &["editors", "edits"],
            ),
            (dictgen::InsensitiveStr::Ascii("paladines"), &["paladins"]),
            (dictgen::InsensitiveStr::Ascii("franlkin"), &["franklin"]),
            (
                dictgen::InsensitiveStr::Ascii("transisitor"),
                &["transistor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manisfestations"),
                &["manifestations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlightenend"),
                &["enlightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("amplitud"), &["amplitude"]),
            (
                dictgen::InsensitiveStr::Ascii("letgitimate"),
                &["legitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versioming"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("goruotine"), &["goroutine"]),
            (
                dictgen::InsensitiveStr::Ascii("troubleshooot"),
                &["troubleshoot"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispite"), &["despite"]),
            (
                dictgen::InsensitiveStr::Ascii("volenteers"),
                &["volunteers"],
            ),
            (dictgen::InsensitiveStr::Ascii("zinoists"), &["zionists"]),
            (dictgen::InsensitiveStr::Ascii("enconnding"), &["encoding"]),
            (
                dictgen::InsensitiveStr::Ascii("puplish"),
                &["publish", "purplish"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacakage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("serialialised"),
                &["serialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abrreviation"),
                &["abbreviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiatiors"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commuinications"),
                &["communications"],
            ),
            (dictgen::InsensitiveStr::Ascii("theshold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("admitedly"), &["admittedly"]),
            (dictgen::InsensitiveStr::Ascii("illustre"), &["illustrate"]),
            (
                dictgen::InsensitiveStr::Ascii("rennaisance"),
                &["renaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predictibility"),
                &["predictability"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutlicast"), &["multicast"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarcically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("breef"), &["brief", "beef"]),
            (dictgen::InsensitiveStr::Ascii("abudction"), &["abduction"]),
            (dictgen::InsensitiveStr::Ascii("teest"), &["test"]),
            (
                dictgen::InsensitiveStr::Ascii("policie"),
                &["policies", "policy", "police"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuasse"),
                &["continues"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conceptuel"),
                &["conceptual"],
            ),
            (dictgen::InsensitiveStr::Ascii("itialise"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("capsuels"), &["capsules"]),
            (dictgen::InsensitiveStr::Ascii("stupire"), &["stupider"]),
            (dictgen::InsensitiveStr::Ascii("filess"), &["files"]),
            (
                dictgen::InsensitiveStr::Ascii("depanding"),
                &["depending", "demanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("preambel"), &["preamble"]),
            (
                dictgen::InsensitiveStr::Ascii("unresgisters"),
                &["unregisters"],
            ),
            (dictgen::InsensitiveStr::Ascii("encoutter"), &["encounter"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaition"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("definig"), &["defining"]),
            (dictgen::InsensitiveStr::Ascii("nesponse"), &["response"]),
            (
                dictgen::InsensitiveStr::Ascii("minues"),
                &["menus", "minus", "minuses", "minutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconition"),
                &["recognition"],
            ),
            (dictgen::InsensitiveStr::Ascii("fethced"), &["fetched"]),
            (dictgen::InsensitiveStr::Ascii("ariborne"), &["airborne"]),
            (dictgen::InsensitiveStr::Ascii("charsima"), &["charisma"]),
            (
                dictgen::InsensitiveStr::Ascii("hamburguers"),
                &["hamburgers"],
            ),
            (dictgen::InsensitiveStr::Ascii("backwward"), &["backward"]),
            (dictgen::InsensitiveStr::Ascii("hydogen"), &["hydrogen"]),
            (dictgen::InsensitiveStr::Ascii("anomynous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("desentisized"),
                &["desensitized"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("comnplex"), &["complex"]),
            (dictgen::InsensitiveStr::Ascii("punicode"), &["punycode"]),
            (dictgen::InsensitiveStr::Ascii("debufffs"), &["debuffs"]),
            (dictgen::InsensitiveStr::Ascii("meaing"), &["meaning"]),
            (
                dictgen::InsensitiveStr::Ascii("liberterians"),
                &["libertarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyward"), &["keyword"]),
            (
                dictgen::InsensitiveStr::Ascii("steryotyping"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privisions"),
                &["provisions"],
            ),
            (dictgen::InsensitiveStr::Ascii("downalod"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("normaizing"),
                &["normalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("furncae"), &["furnace"]),
            (dictgen::InsensitiveStr::Ascii("servivors"), &["survivors"]),
            (dictgen::InsensitiveStr::Ascii("regarting"), &["regarding"]),
            (dictgen::InsensitiveStr::Ascii("oldests"), &["oldest"]),
            (dictgen::InsensitiveStr::Ascii("carolinia"), &["carolina"]),
            (dictgen::InsensitiveStr::Ascii("prescibe"), &["prescribe"]),
            (
                dictgen::InsensitiveStr::Ascii("billegerent"),
                &["belligerent"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriipting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("betrayd"), &["betrayed"]),
            (dictgen::InsensitiveStr::Ascii("fomr"), &["from", "form"]),
            (dictgen::InsensitiveStr::Ascii("quivalent"), &["equivalent"]),
            (dictgen::InsensitiveStr::Ascii("anoter"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("complicatii"),
                &["complicit"],
            ),
            (dictgen::InsensitiveStr::Ascii("fifht"), &["fifth", "fight"]),
            (
                dictgen::InsensitiveStr::Ascii("necassrily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("presssing"), &["pressing"]),
            (dictgen::InsensitiveStr::Ascii("revese"), &["reverse"]),
            (
                dictgen::InsensitiveStr::Ascii("devastanti"),
                &["devastating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unequality"),
                &["inequality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buyed"),
                &["bought", "buyer"],
            ),
            (dictgen::InsensitiveStr::Ascii("maibe"), &["maybe"]),
            (dictgen::InsensitiveStr::Ascii("dasdot"), &["dashdot"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretate"),
                &["interpret"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandizing"),
                &["aggrandizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("assembeld"), &["assembled"]),
            (dictgen::InsensitiveStr::Ascii("emualtion"), &["emulation"]),
            (
                dictgen::InsensitiveStr::Ascii("unstalling"),
                &["installing", "uninstalling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interbread"),
                &["interbreed", "interbred"],
            ),
            (dictgen::InsensitiveStr::Ascii("deahtly"), &["deathly"]),
            (
                dictgen::InsensitiveStr::Ascii("democratisch"),
                &["democratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("lancin"), &["lancing"]),
            (dictgen::InsensitiveStr::Ascii("etropy"), &["entropy"]),
            (
                dictgen::InsensitiveStr::Ascii("disctinguish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("algnment"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("resetoring"), &["restoring"]),
            (
                dictgen::InsensitiveStr::Ascii("disgarded"),
                &["discarded", "discarted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dection"),
                &[
                    "detection",
                    "section",
                    "direction",
                    "deduction",
                    "diction",
                    "depiction",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("sunderlad"), &["sunderland"]),
            (
                dictgen::InsensitiveStr::Ascii("methaporical"),
                &["metaphorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astonising"),
                &["astonishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("organisten"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("penetraton"),
                &["penetration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condict"),
                &["conduct", "convict"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oakerously"),
                &["ocherously"],
            ),
            (dictgen::InsensitiveStr::Ascii("soodonim"), &["pseudonym"]),
            (dictgen::InsensitiveStr::Ascii("permites"), &["permits"]),
            (dictgen::InsensitiveStr::Ascii("parallely"), &["parallelly"]),
            (
                dictgen::InsensitiveStr::Ascii("neighroring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beautilful"),
                &["beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagiarius"),
                &["plagiarism"],
            ),
            (dictgen::InsensitiveStr::Ascii("ublishes"), &["publishes"]),
            (
                dictgen::InsensitiveStr::Ascii("prelimanary"),
                &["preliminary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacializations"),
                &["specializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranasction"),
                &["transaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disssembly"),
                &["disassembly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precessor"),
                &["predecessor", "processor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rectanglar"),
                &["rectangular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("midfiedlers"),
                &["midfielders"],
            ),
            (dictgen::InsensitiveStr::Ascii("mortage"), &["mortgage"]),
            (
                dictgen::InsensitiveStr::Ascii("trasverse"),
                &["traverse", "transverse"],
            ),
            (dictgen::InsensitiveStr::Ascii("projetcts"), &["projects"]),
            (dictgen::InsensitiveStr::Ascii("videogams"), &["videogames"]),
            (dictgen::InsensitiveStr::Ascii("capabable"), &["capable"]),
            (dictgen::InsensitiveStr::Ascii("refect"), &["reflect"]),
            (
                dictgen::InsensitiveStr::Ascii("depenedencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("decomiler"), &["decompiler"]),
            (
                dictgen::InsensitiveStr::Ascii("repressents"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availabiliy"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("odasees"), &["odysseys"]),
            (
                dictgen::InsensitiveStr::Ascii("qustionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrealeased"),
                &["unreleased"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrucures"),
                &["restructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penetrarion"),
                &["penetration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miinimization"),
                &["minimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("miniatute"), &["miniature"]),
            (dictgen::InsensitiveStr::Ascii("woiuld"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("speciel"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("checkcerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstatiated"),
                &["uninstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("canidate"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("exprimentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeasing"), &["releasing"]),
            (
                dictgen::InsensitiveStr::Ascii("reproductivo"),
                &["reproduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranmsission"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarases"),
                &["embarrasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currecting"),
                &["correcting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downoading"),
                &["downloading"],
            ),
            (dictgen::InsensitiveStr::Ascii("pticher"), &["pitcher"]),
            (
                dictgen::InsensitiveStr::Ascii("ciruclator"),
                &["circulator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("levensthein"),
                &["levenshtein"],
            ),
            (dictgen::InsensitiveStr::Ascii("buitton"), &["button"]),
            (dictgen::InsensitiveStr::Ascii("conpilers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("subdivisio"),
                &["subdivision"],
            ),
            (dictgen::InsensitiveStr::Ascii("varifying"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("implmentation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("faiulres"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("depandent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("accorind"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("acusation"),
                &["causation", "accusation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chisiles"), &["chisels"]),
            (dictgen::InsensitiveStr::Ascii("diggesting"), &["digesting"]),
            (
                dictgen::InsensitiveStr::Ascii("practictitioners"),
                &["practitioners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grapefriut"),
                &["grapefruit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("circut"), &["circuit"]),
            (
                dictgen::InsensitiveStr::Ascii("architetcures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfifies"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unapretiated"),
                &["unappreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalesc"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("inheritences"),
                &["inheritances"],
            ),
            (dictgen::InsensitiveStr::Ascii("digial"), &["digital"]),
            (
                dictgen::InsensitiveStr::Ascii("applycation"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("synonymes"), &["synonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("algotighmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installaitons"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("detaild"), &["detailed"]),
            (dictgen::InsensitiveStr::Ascii("bitnaps"), &["bitmaps"]),
            (
                dictgen::InsensitiveStr::Ascii("serverites"),
                &["severities"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaites"), &["locates"]),
            (dictgen::InsensitiveStr::Ascii("ppyint"), &["pyint"]),
            (dictgen::InsensitiveStr::Ascii("desingner"), &["designer"]),
            (
                dictgen::InsensitiveStr::Ascii("conflictd"),
                &["conflicts", "conflicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnaslator"),
                &["translator"],
            ),
            (dictgen::InsensitiveStr::Ascii("freqeunt"), &["frequent"]),
            (
                dictgen::InsensitiveStr::Ascii("uncousciously"),
                &["unconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commenteers"),
                &["commenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steryotypes"),
                &["stereotypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconecction"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curce"),
                &["course", "curse", "curve"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartator"),
                &["declarator"],
            ),
            (dictgen::InsensitiveStr::Ascii("heaveny"), &["heavenly"]),
            (dictgen::InsensitiveStr::Ascii("heped"), &["helped"]),
            (dictgen::InsensitiveStr::Ascii("hierarcy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("comparistons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sumbissions"),
                &["submissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatilibities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compule"),
                &["compute", "compile"],
            ),
            (dictgen::InsensitiveStr::Ascii("agorithms"), &["algorithms"]),
            (
                dictgen::InsensitiveStr::Ascii("sympatheic"),
                &["sympathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelisation"),
                &["parallelisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("spliting"), &["splitting"]),
            (
                dictgen::InsensitiveStr::Ascii("endocer"),
                &["endorser", "enforcer", "encoder"],
            ),
            (dictgen::InsensitiveStr::Ascii("middte"), &["middle"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamers"),
                &["performers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strenghtened"),
                &["strengthened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortismic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decprecated"),
                &["deprecated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutlitasking"),
                &["multitasking"],
            ),
            (dictgen::InsensitiveStr::Ascii("nife"), &["knife"]),
            (
                dictgen::InsensitiveStr::Ascii("mocrocontrollers"),
                &["microcontrollers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complainte"),
                &["complained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commponents"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("fireballls"), &["fireballs"]),
            (dictgen::InsensitiveStr::Ascii("clsoe"), &["close"]),
            (dictgen::InsensitiveStr::Ascii("coexsits"), &["coexists"]),
            (dictgen::InsensitiveStr::Ascii("userr"), &["user"]),
            (
                dictgen::InsensitiveStr::Ascii("neighourhoods"),
                &["neighbourhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beanched"),
                &["branched", "benched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("obatain"), &["obtain"]),
            (dictgen::InsensitiveStr::Ascii("actice"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("netocde"), &["netcode"]),
            (
                dictgen::InsensitiveStr::Ascii("coommitters"),
                &["committers"],
            ),
            (dictgen::InsensitiveStr::Ascii("miminised"), &["minimised"]),
            (dictgen::InsensitiveStr::Ascii("delievred"), &["delivered"]),
            (dictgen::InsensitiveStr::Ascii("defeault"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptoins"),
                &["exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battelfront"),
                &["battlefront"],
            ),
            (dictgen::InsensitiveStr::Ascii("struggeld"), &["struggled"]),
            (dictgen::InsensitiveStr::Ascii("renamin"), &["renaming"]),
            (dictgen::InsensitiveStr::Ascii("tracsoded"), &["transcoded"]),
            (dictgen::InsensitiveStr::Ascii("scripots"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("depsise"), &["despise"]),
            (dictgen::InsensitiveStr::Ascii("pusnishes"), &["punishes"]),
            (dictgen::InsensitiveStr::Ascii("epsorts"), &["esports"]),
            (
                dictgen::InsensitiveStr::Ascii("annonomusally"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inplementations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("warriros"), &["warriors"]),
            (
                dictgen::InsensitiveStr::Ascii("appealling"),
                &["appealing", "appalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("nautils"), &["nautilus"]),
            (
                dictgen::InsensitiveStr::Ascii("varaiations"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("buttonns"), &["buttons"]),
            (
                dictgen::InsensitiveStr::Ascii("definatelty"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experineced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insternally"),
                &["internally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utilizaiton"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docmuneting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("harrased"), &["harassed"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanent"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("witht"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("forvatter"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("confgiure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("mastrubation"),
                &["masturbation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpectend"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boootstraps"),
                &["bootstraps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promotin"),
                &["promoting", "promotion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autmoation"),
                &["automation"],
            ),
            (dictgen::InsensitiveStr::Ascii("degrads"), &["degrades"]),
            (dictgen::InsensitiveStr::Ascii("thickenss"), &["thickness"]),
            (
                dictgen::InsensitiveStr::Ascii("supplamental"),
                &["supplemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetntication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("infering"), &["inferring"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotitions"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submerines"),
                &["submarines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessleated"),
                &["tessellated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparoched"),
                &["approached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defint"),
                &["definite", "define"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginizing"),
                &["recognizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpsition"),
                &["transposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inscrutible"),
                &["inscrutable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrispondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("roperty"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("indroduced"),
                &["introduced"],
            ),
            (dictgen::InsensitiveStr::Ascii("jitterring"), &["jittering"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplexer"),
                &["multiplexer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanichal"),
                &["mechanical"],
            ),
            (dictgen::InsensitiveStr::Ascii("underfow"), &["underflow"]),
            (dictgen::InsensitiveStr::Ascii("obdisian"), &["obsidian"]),
            (
                dictgen::InsensitiveStr::Ascii("corse"),
                &[
                    "course", "coarse", "core", "corpse", "cors", "corset", "curse", "horse",
                    "norse", "torse", "worse",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("lications"), &["locations"]),
            (dictgen::InsensitiveStr::Ascii("declarasen"), &["declares"]),
            (dictgen::InsensitiveStr::Ascii("loggger"), &["logger"]),
            (dictgen::InsensitiveStr::Ascii("facinate"), &["fascinate"]),
            (
                dictgen::InsensitiveStr::Ascii("checkstuming"),
                &["checksumming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiatves"),
                &["initiatives", "initiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("menacin"), &["menacing"]),
            (dictgen::InsensitiveStr::Ascii("anomynity"), &["anonymity"]),
            (dictgen::InsensitiveStr::Ascii("dargging"), &["dragging"]),
            (dictgen::InsensitiveStr::Ascii("indicents"), &["incidents"]),
            (
                dictgen::InsensitiveStr::Ascii("unsccessful"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("sharraids"), &["charades"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoticator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("inculde"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("expirenced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equavalents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandity"), &["sanity"]),
            (
                dictgen::InsensitiveStr::Ascii("ejaculatin"),
                &["ejaculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paychopathic"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyar"), &["year", "eyas"]),
            (dictgen::InsensitiveStr::Ascii("dingee"), &["dinghy"]),
            (
                dictgen::InsensitiveStr::Ascii("technichan"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrriding"),
                &["overriding"],
            ),
            (dictgen::InsensitiveStr::Ascii("eahc"), &["each"]),
            (
                dictgen::InsensitiveStr::Ascii("moisturizor"),
                &["moisturizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("upadated"), &["updated"]),
            (
                dictgen::InsensitiveStr::Ascii("suspecions"),
                &["suspicions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temparatures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititative"),
                &["initiative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectus"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrirbuted"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicatiosn"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderrers"), &["renderers"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotaiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bullevards"),
                &["boulevards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficilly"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("orthagnal"), &["orthogonal"]),
            (
                dictgen::InsensitiveStr::Ascii("porbability"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("islamisters"),
                &["islamists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apologisin"),
                &["apologising"],
            ),
            (dictgen::InsensitiveStr::Ascii("secund"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitudes"),
                &["prostitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistancies"),
                &["resistances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arthimetic"),
                &["arithmetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verication"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrutinity"), &["scrutiny"]),
            (dictgen::InsensitiveStr::Ascii("generete"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("acively"), &["actively"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpextedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("throtlled"), &["throttled"]),
            (dictgen::InsensitiveStr::Ascii("sysmte"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("clonet"), &["clone"]),
            (
                dictgen::InsensitiveStr::Ascii("possiility"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainigns"), &["trainings"]),
            (dictgen::InsensitiveStr::Ascii("rekursion"), &["recursion"]),
            (
                dictgen::InsensitiveStr::Ascii("munipulative"),
                &["manipulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srcollable"),
                &["scrollable"],
            ),
            (dictgen::InsensitiveStr::Ascii("callss"), &["calls"]),
            (dictgen::InsensitiveStr::Ascii("fasen"), &["fasten"]),
            (dictgen::InsensitiveStr::Ascii("chillead"), &["chilled"]),
            (dictgen::InsensitiveStr::Ascii("trudnle"), &["trundle"]),
            (
                dictgen::InsensitiveStr::Ascii("portabillity"),
                &["portability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarentined"),
                &["quarantined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structiring"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("parallell"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("darma"), &["dharma"]),
            (
                dictgen::InsensitiveStr::Ascii("recivership"),
                &["receivership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supervisiors"),
                &["supervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sizemology"),
                &["seismology"],
            ),
            (dictgen::InsensitiveStr::Ascii("faggotus"), &["faggots"]),
            (
                dictgen::InsensitiveStr::Ascii("credientals"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimentionnal"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("alcemy"), &["alchemy"]),
            (dictgen::InsensitiveStr::Ascii("incldes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("substition"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("switserland"),
                &["switzerland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remarkabley"),
                &["remarkably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accestors"),
                &["accessors", "ancestors"],
            ),
            (dictgen::InsensitiveStr::Ascii("constacted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("aoache"), &["apache"]),
            (dictgen::InsensitiveStr::Ascii("pretects"), &["protects"]),
            (
                dictgen::InsensitiveStr::Ascii("transciever"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("stiky"), &["sticky"]),
            (
                dictgen::InsensitiveStr::Ascii("furthermor"),
                &["furthermore"],
            ),
            (dictgen::InsensitiveStr::Ascii("incluse"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("descriotor"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalutations"),
                &["evaluations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfriently"),
                &["unfriendly"],
            ),
            (dictgen::InsensitiveStr::Ascii("melborune"), &["melbourne"]),
            (
                dictgen::InsensitiveStr::Ascii("indepenedence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("reallt"), &["really"]),
            (dictgen::InsensitiveStr::Ascii("executuion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("compres"),
                &["compress", "compares"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrabil"),
                &["demonstrably"],
            ),
            (dictgen::InsensitiveStr::Ascii("bisunes"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("homogenoues"),
                &["homogeneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defintively"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("allright"), &["alright"]),
            (dictgen::InsensitiveStr::Ascii("projetile"), &["projectile"]),
            (
                dictgen::InsensitiveStr::Ascii("indepenedent"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("protoyped"), &["prototyped"]),
            (
                dictgen::InsensitiveStr::Ascii("carefull"),
                &["careful", "carefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significent"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("loadin"), &["loading"]),
            (dictgen::InsensitiveStr::Ascii("allegence"), &["allegiance"]),
            (dictgen::InsensitiveStr::Ascii("depencent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainter"),
                &["maintainer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentul"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeectedly"),
                &["expectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("guiseppe"), &["giuseppe"]),
            (
                dictgen::InsensitiveStr::Ascii("heaviweight"),
                &["heavyweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementaion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inbankment"),
                &["embankment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eveolutionary"),
                &["evolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("mena"), &["mean"]),
            (dictgen::InsensitiveStr::Ascii("fisist"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("heigths"), &["heights"]),
            (
                dictgen::InsensitiveStr::Ascii("protectice"),
                &["protective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("especailly"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("immedates"), &["immediates"]),
            (dictgen::InsensitiveStr::Ascii("comlaint"), &["complaint"]),
            (
                dictgen::InsensitiveStr::Ascii("priortizes"),
                &["prioritizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attributo"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpand"),
                &["command", "compand"],
            ),
            (dictgen::InsensitiveStr::Ascii("mattreses"), &["mattress"]),
            (
                dictgen::InsensitiveStr::Ascii("genenerated"),
                &["generated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepcial"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("clenase"), &["cleanse"]),
            (
                dictgen::InsensitiveStr::Ascii("indevering"),
                &["endeavoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doucmenter"),
                &["documenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("destoy"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("changelin"), &["changeling"]),
            (
                dictgen::InsensitiveStr::Ascii("mathemetics"),
                &["mathematics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synthesed"),
                &["synthesized"],
            ),
            (dictgen::InsensitiveStr::Ascii("srroy"), &["sorry"]),
            (dictgen::InsensitiveStr::Ascii("descover"), &["discover"]),
            (
                dictgen::InsensitiveStr::Ascii("experimdnt"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("humanitary"), &["humanity"]),
            (dictgen::InsensitiveStr::Ascii("algorrithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("depeneds"), &["depends"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulises"),
                &["virtualises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expendature"),
                &["expenditure"],
            ),
            (dictgen::InsensitiveStr::Ascii("utiliz"), &["utilize"]),
            (dictgen::InsensitiveStr::Ascii("inershial"), &["inertial"]),
            (dictgen::InsensitiveStr::Ascii("cllosest"), &["closest"]),
            (
                dictgen::InsensitiveStr::Ascii("psychadelic"),
                &["psychedelic"],
            ),
            (dictgen::InsensitiveStr::Ascii("commenetd"), &["commented"]),
            (dictgen::InsensitiveStr::Ascii("magents"), &["magnets"]),
            (dictgen::InsensitiveStr::Ascii("serilize"), &["serialize"]),
            (
                dictgen::InsensitiveStr::Ascii("environmenets"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("emaill"), &["email"]),
            (
                dictgen::InsensitiveStr::Ascii("endevering"),
                &["endeavoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intereference"),
                &["interference"],
            ),
            (dictgen::InsensitiveStr::Ascii("treausres"), &["treasures"]),
            (
                dictgen::InsensitiveStr::Ascii("experinementally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("infintiy"), &["infinity"]),
            (dictgen::InsensitiveStr::Ascii("cosnsrain"), &["constrain"]),
            (
                dictgen::InsensitiveStr::Ascii("haemorrhagin"),
                &["haemorrhaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("frecuent"), &["frequent"]),
            (
                dictgen::InsensitiveStr::Ascii("notififation"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkoslovakia"),
                &["czechoslovakia"],
            ),
            (dictgen::InsensitiveStr::Ascii("dichtomy"), &["dichotomy"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisists"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("regurally"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("limitter"), &["limiter"]),
            (dictgen::InsensitiveStr::Ascii("scetion"), &["section"]),
            (
                dictgen::InsensitiveStr::Ascii("editabiliity"),
                &["editability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accleometer"),
                &["accelerometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meda"),
                &["medal", "media", "meta"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmetned"), &["segmented"]),
            (dictgen::InsensitiveStr::Ascii("bitis"), &["bits"]),
            (dictgen::InsensitiveStr::Ascii("numebers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("compromisng"),
                &["compromising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uploaed"),
                &["upload", "uploaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("califorina"),
                &["california"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demolation"),
                &["demolition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemoragic"),
                &["haemorrhagic", "hemorrhagic"],
            ),
            (dictgen::InsensitiveStr::Ascii("newsanses"), &["nuisances"]),
            (
                dictgen::InsensitiveStr::Ascii("filghts"),
                &["flights", "fights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbaing"),
                &["masturbating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("xenophibia"),
                &["xenophobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("lukid"), &["likud", "lucid"]),
            (dictgen::InsensitiveStr::Ascii("benine"), &["benign"]),
            (
                dictgen::InsensitiveStr::Ascii("assosiating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previsouly"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physcedelic"),
                &["psychedelic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("subisdy"), &["subsidy"]),
            (dictgen::InsensitiveStr::Ascii("thiss"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("eazy"), &["easy"]),
            (dictgen::InsensitiveStr::Ascii("absout"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("cimptomaticaly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcordings"),
                &["transcodings"],
            ),
            (dictgen::InsensitiveStr::Ascii("swaering"), &["swearing"]),
            (dictgen::InsensitiveStr::Ascii("morhpine"), &["morphine"]),
            (dictgen::InsensitiveStr::Ascii("unicrons"), &["unicorns"]),
            (dictgen::InsensitiveStr::Ascii("inspecor"), &["inspector"]),
            (dictgen::InsensitiveStr::Ascii("refferring"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("seagments"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("sufferage"), &["suffrage"]),
            (dictgen::InsensitiveStr::Ascii("oficial"), &["official"]),
            (
                dictgen::InsensitiveStr::Ascii("supervisers"),
                &["supervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuety"),
                &["continuity"],
            ),
            (dictgen::InsensitiveStr::Ascii("confilct"), &["conflict"]),
            (dictgen::InsensitiveStr::Ascii("strentgh"), &["strength"]),
            (
                dictgen::InsensitiveStr::Ascii("ligthening"),
                &["lightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("earyly"), &["eerily"]),
            (
                dictgen::InsensitiveStr::Ascii("comuted"),
                &["commuted", "computed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visulization"),
                &["visualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signficiance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumentent"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("duirng"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("comunists"), &["communists"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialzing"),
                &["deserializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mairabd"), &["mariadb"]),
            (
                dictgen::InsensitiveStr::Ascii("individials"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bueraucracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("festivle"), &["festive"]),
            (dictgen::InsensitiveStr::Ascii("variavle"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("fincally"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("resouced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("perticulars"),
                &["particulars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peferences"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcrtly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("valuabe"), &["valuable"]),
            (dictgen::InsensitiveStr::Ascii("emptyness"), &["emptiness"]),
            (
                dictgen::InsensitiveStr::Ascii("tranpshobic"),
                &["transphobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomoorw"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("proficientcy"),
                &["proficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("aliging"), &["aligning"]),
            (dictgen::InsensitiveStr::Ascii("pbject"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("repetatively"),
                &["repetitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("polution"), &["pollution"]),
            (
                dictgen::InsensitiveStr::Ascii("renegociators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realitively"),
                &["relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponsdingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("pontential"), &["potential"]),
            (dictgen::InsensitiveStr::Ascii("checksm"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("psuedocolor"),
                &["pseudocolor"],
            ),
            (dictgen::InsensitiveStr::Ascii("orienatate"), &["orientate"]),
            (dictgen::InsensitiveStr::Ascii("propehcy"), &["prophecy"]),
            (
                dictgen::InsensitiveStr::Ascii("mechnanisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzelus"),
                &["overzealous"],
            ),
            (dictgen::InsensitiveStr::Ascii("awfuly"), &["awfully"]),
            (
                dictgen::InsensitiveStr::Ascii("dissobedient"),
                &["disobedient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workststions"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("curiousity"), &["curiosity"]),
            (
                dictgen::InsensitiveStr::Ascii("randayvoued"),
                &["rendezvoused"],
            ),
            (dictgen::InsensitiveStr::Ascii("explciti"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("sumbit"), &["submit"]),
            (dictgen::InsensitiveStr::Ascii("beastley"), &["beastly"]),
            (dictgen::InsensitiveStr::Ascii("trimed"), &["trimmed"]),
            (dictgen::InsensitiveStr::Ascii("idons"), &["icons"]),
            (
                dictgen::InsensitiveStr::Ascii("totalitaran"),
                &["totalitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("worklfow"), &["workflow"]),
            (
                dictgen::InsensitiveStr::Ascii("descandants"),
                &["descendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aquaintance"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguler"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("chlidish"), &["childish"]),
            (
                dictgen::InsensitiveStr::Ascii("transpancies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("connctors"), &["connectors"]),
            (
                dictgen::InsensitiveStr::Ascii("partcularly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("bundeled"), &["bundled"]),
            (dictgen::InsensitiveStr::Ascii("laucnhes"), &["launches"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotatiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unefficiently"),
                &["inefficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("slienced"), &["silenced"]),
            (dictgen::InsensitiveStr::Ascii("pytyon"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("devloppers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupplication"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsed"),
                &["used", "unused", "unset"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyclopeadic"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otpimisations"),
                &["optimisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incentivos"),
                &["incentives"],
            ),
            (dictgen::InsensitiveStr::Ascii("aparthide"), &["apartheid"]),
            (dictgen::InsensitiveStr::Ascii("acitivated"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("availble"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("imporoved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("sucessor"), &["successor"]),
            (dictgen::InsensitiveStr::Ascii("publush"), &["publish"]),
            (
                dictgen::InsensitiveStr::Ascii("applicabiltiy"),
                &["applicability"],
            ),
            (dictgen::InsensitiveStr::Ascii("paitently"), &["patiently"]),
            (dictgen::InsensitiveStr::Ascii("recommdend"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("assimialted"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("convereted"), &["converted"]),
            (dictgen::InsensitiveStr::Ascii("someoneis"), &["someones"]),
            (
                dictgen::InsensitiveStr::Ascii("opitions"),
                &["options", "opinions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pshycosis"), &["psychosis"]),
            (dictgen::InsensitiveStr::Ascii("metatadata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("acivates"), &["activates"]),
            (dictgen::InsensitiveStr::Ascii("assigee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("beanchmark"), &["benchmark"]),
            (dictgen::InsensitiveStr::Ascii("confims"), &["confirms"]),
            (dictgen::InsensitiveStr::Ascii("autoattak"), &["autoattack"]),
            (
                dictgen::InsensitiveStr::Ascii("hyperthropy"),
                &["hypertrophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tepmorarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("obilvion"), &["oblivion"]),
            (dictgen::InsensitiveStr::Ascii("immobilie"), &["immobile"]),
            (dictgen::InsensitiveStr::Ascii("cnannels"), &["channels"]),
            (
                dictgen::InsensitiveStr::Ascii("additinally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("suficient"), &["sufficient"]),
            (
                dictgen::InsensitiveStr::Ascii("immortailty"),
                &["immortality"],
            ),
            (dictgen::InsensitiveStr::Ascii("succuss"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streigt"),
                &["straight", "strait"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competenet"),
                &["competence"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranports"), &["transports"]),
            (dictgen::InsensitiveStr::Ascii("importnce"), &["importance"]),
            (
                dictgen::InsensitiveStr::Ascii("bulnerability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapprove"),
                &["disapprove"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certfication"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("intagible"), &["intangible"]),
            (dictgen::InsensitiveStr::Ascii("kidnly"), &["kindly"]),
            (
                dictgen::InsensitiveStr::Ascii("conrruption"),
                &["corruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrivute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("somwhere"), &["somewhere"]),
            (dictgen::InsensitiveStr::Ascii("flatteing"), &["flattening"]),
            (
                dictgen::InsensitiveStr::Ascii("concatentates"),
                &["concatenates"],
            ),
            (dictgen::InsensitiveStr::Ascii("empheral"), &["ephemeral"]),
            (
                dictgen::InsensitiveStr::Ascii("mechinisms"),
                &["mechanisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("neight"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("inventario"), &["inventor"]),
            (dictgen::InsensitiveStr::Ascii("ooutputs"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("checkeer"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("egenralizes"),
                &["generalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("rentres"), &["renters"]),
            (
                dictgen::InsensitiveStr::Ascii("surrenderred"),
                &["surrendered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transplanet"),
                &["transplant"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedemies"), &["academies"]),
            (dictgen::InsensitiveStr::Ascii("mesaging"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("simmilarity"),
                &["similarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditianally"),
                &["conditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("assighers"), &["assigners"]),
            (
                dictgen::InsensitiveStr::Ascii("accompinied"),
                &["accompanied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fascisation"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("mportant"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("specifities"),
                &["specifics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitalises"),
                &["reinitialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("lients"), &["clients"]),
            (
                dictgen::InsensitiveStr::Ascii("opthalmology"),
                &["ophthalmology"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapats"), &["dissipates"]),
            (
                dictgen::InsensitiveStr::Ascii("desccription"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("emperial"), &["imperial"]),
            (dictgen::InsensitiveStr::Ascii("etherel"), &["ethereal"]),
            (dictgen::InsensitiveStr::Ascii("operartor"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("threatend"), &["threatened"]),
            (
                dictgen::InsensitiveStr::Ascii("appriceate"),
                &["appreciates", "appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("priotiry"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("lewows"), &["luaus"]),
            (
                dictgen::InsensitiveStr::Ascii("documentating"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aphabetic"), &["alphabetic"]),
            (
                dictgen::InsensitiveStr::Ascii("finisch"),
                &["finish", "finnish"],
            ),
            (dictgen::InsensitiveStr::Ascii("convertr"), &["converter"]),
            (dictgen::InsensitiveStr::Ascii("maual"), &["manual", "maul"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrivating"),
                &["aggravating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swiched"),
                &["switched", "swished"],
            ),
            (dictgen::InsensitiveStr::Ascii("writtened"), &["written"]),
            (
                dictgen::InsensitiveStr::Ascii("recipercality"),
                &["reciprocality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recnt"),
                &["recent", "recant", "rent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difficulte"),
                &["difficulties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiuration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oders"),
                &["orders", "coders", "odors", "doers"],
            ),
            (dictgen::InsensitiveStr::Ascii("donejuns"), &["dungeons"]),
            (
                dictgen::InsensitiveStr::Ascii("capitlaist"),
                &["capitalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysin"), &["analysing"]),
            (
                dictgen::InsensitiveStr::Ascii("cehckerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delaration"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("numberred"), &["numbered"]),
            (
                dictgen::InsensitiveStr::Ascii("figurestyle"),
                &["figurestyles"],
            ),
            (dictgen::InsensitiveStr::Ascii("degrate"), &["degrade"]),
            (dictgen::InsensitiveStr::Ascii("voyour"), &["voyeur"]),
            (
                dictgen::InsensitiveStr::Ascii("funcionalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("intiution"), &["intuition"]),
            (dictgen::InsensitiveStr::Ascii("spcified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("datatyepe"), &["datatype"]),
            (
                dictgen::InsensitiveStr::Ascii("checkkerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("thck"), &["thick"]),
            (
                dictgen::InsensitiveStr::Ascii("tabulatin"),
                &["tabulating", "tabulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mozzilla"), &["mozilla"]),
            (dictgen::InsensitiveStr::Ascii("implemets"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegogtiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oprionally"),
                &["optionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("foresnic"), &["forensic"]),
            (
                dictgen::InsensitiveStr::Ascii("undertoker"),
                &["undertaker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabile"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("produsers"), &["producers"]),
            (dictgen::InsensitiveStr::Ascii("numberss"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("contenplate"),
                &["contemplate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispicable"),
                &["despicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("efinite"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("pregancy"), &["pregnancy"]),
            (
                dictgen::InsensitiveStr::Ascii("paragraghs"),
                &["paragraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("appneded"), &["appended"]),
            (dictgen::InsensitiveStr::Ascii("augmneted"), &["augmented"]),
            (
                dictgen::InsensitiveStr::Ascii("representaional"),
                &["representational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occassionaly"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdedded"), &["embedded"]),
            (
                dictgen::InsensitiveStr::Ascii("unvelievably"),
                &["unbelievably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomidations"),
                &["accommodations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silhouetist"),
                &["silhouettist"],
            ),
            (dictgen::InsensitiveStr::Ascii("concieted"), &["conceited"]),
            (dictgen::InsensitiveStr::Ascii("werider"), &["weirder"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronisers"),
                &["synchronisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descovered"),
                &["discovered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recognice"),
                &["recognize", "recognise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descipable"),
                &["despicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressults"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("servce"),
                &["service", "serve"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefix"), &["prefix"]),
            (dictgen::InsensitiveStr::Ascii("ueried"), &["queried"]),
            (
                dictgen::InsensitiveStr::Ascii("clasically"),
                &["classically"],
            ),
            (dictgen::InsensitiveStr::Ascii("sementic"), &["semantic"]),
            (dictgen::InsensitiveStr::Ascii("createn"), &["creatine"]),
            (dictgen::InsensitiveStr::Ascii("militais"), &["militias"]),
            (dictgen::InsensitiveStr::Ascii("baloons"), &["balloons"]),
            (
                dictgen::InsensitiveStr::Ascii("descreptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoveniences"),
                &["inconveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaibility"),
                &["reliability"],
            ),
            (dictgen::InsensitiveStr::Ascii("abilty"), &["ability"]),
            (
                dictgen::InsensitiveStr::Ascii("tooo"),
                &["todo", "too", "took", "tool"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("housekeepin"),
                &["housekeeping"],
            ),
            (dictgen::InsensitiveStr::Ascii("enclsoure"), &["enclosure"]),
            (
                dictgen::InsensitiveStr::Ascii("dialoges"),
                &["dialogs", "dialogues"],
            ),
            (dictgen::InsensitiveStr::Ascii("thirsday"), &["thursday"]),
            (
                dictgen::InsensitiveStr::Ascii("generationers"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("adivsors"), &["advisors"]),
            (dictgen::InsensitiveStr::Ascii("bloekes"), &["blokes"]),
            (
                dictgen::InsensitiveStr::Ascii("apperciate"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sorkflow"), &["workflow"]),
            (
                dictgen::InsensitiveStr::Ascii("intresection"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canabelyzes"),
                &["cannibalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containging"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("opton"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("dominace"),
                &["dominance", "dominate"],
            ),
            (dictgen::InsensitiveStr::Ascii("islamit"), &["islamist"]),
            (
                dictgen::InsensitiveStr::Ascii("fraturnity"),
                &["fraternity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubadors"),
                &["troubadours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skatin"),
                &["skating", "statin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derepcated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("vrersion"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("precondtions"),
                &["preconditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecoding"),
                &["encoding", "decoding"],
            ),
            (dictgen::InsensitiveStr::Ascii("mecahnism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("temperatur"),
                &["temperature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyticals"),
                &["analytics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understandig"),
                &["understanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horiztonal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("caculator"), &["calculator"]),
            (dictgen::InsensitiveStr::Ascii("updat"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("alailable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("caler"), &["caller"]),
            (
                dictgen::InsensitiveStr::Ascii("permissoin"),
                &["permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("modiifying"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("sturggling"),
                &["struggling"],
            ),
            (dictgen::InsensitiveStr::Ascii("privaledge"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("expres"), &["express"]),
            (dictgen::InsensitiveStr::Ascii("saddnes"), &["saddens"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmiated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defautting"),
                &["defaulting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensistively"),
                &["sensitively", "sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisitics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancinets"), &["ancients"]),
            (dictgen::InsensitiveStr::Ascii("acqueos"), &["aqueous"]),
            (dictgen::InsensitiveStr::Ascii("endeveres"), &["endeavors"]),
            (
                dictgen::InsensitiveStr::Ascii("mathemagically"),
                &["mathematically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallelpiped"),
                &["parallelepiped"],
            ),
            (dictgen::InsensitiveStr::Ascii("modelinng"), &["modeling"]),
            (dictgen::InsensitiveStr::Ascii("exectured"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("extreanous"),
                &["extraneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermients"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sheparding"),
                &["shepherding"],
            ),
            (dictgen::InsensitiveStr::Ascii("acepters"), &["accepters"]),
            (
                dictgen::InsensitiveStr::Ascii("corrspondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("prayrys"), &["prairies"]),
            (dictgen::InsensitiveStr::Ascii("inlies"), &["inlines"]),
            (dictgen::InsensitiveStr::Ascii("expressie"), &["expressive"]),
            (dictgen::InsensitiveStr::Ascii("outfidel"), &["outfield"]),
            (
                dictgen::InsensitiveStr::Ascii("ccordinates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("apporved"), &["approved"]),
            (dictgen::InsensitiveStr::Ascii("menutitems"), &["menuitems"]),
            (
                dictgen::InsensitiveStr::Ascii("siliences"),
                &["silences", "saliences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingegrated"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("flahsed"), &["flashed"]),
            (dictgen::InsensitiveStr::Ascii("appeasr"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("booshelves"),
                &["bookshelves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sudnerland"),
                &["sunderland"],
            ),
            (dictgen::InsensitiveStr::Ascii("counseil"), &["counsel"]),
            (dictgen::InsensitiveStr::Ascii("keupoints"), &["keypoints"]),
            (
                dictgen::InsensitiveStr::Ascii("uncorectly"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("conaines"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("unintentionnal"),
                &["unintentional"],
            ),
            (dictgen::InsensitiveStr::Ascii("exaplain"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("aserted"), &["asserted"]),
            (dictgen::InsensitiveStr::Ascii("syteny"), &["synteny"]),
            (dictgen::InsensitiveStr::Ascii("sivible"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("wordpres"), &["wordpress"]),
            (
                dictgen::InsensitiveStr::Ascii("expereince"),
                &["experience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intreeguing"),
                &["intriguing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ironcially"),
                &["ironically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyticly"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("brocken"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("trapeziodal"),
                &["trapezoidal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcorded"),
                &["transcoded"],
            ),
            (dictgen::InsensitiveStr::Ascii("puroses"), &["purposes"]),
            (
                dictgen::InsensitiveStr::Ascii("placehoders"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volunterily"),
                &["voluntarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucrural"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physcopaths"),
                &["psychopaths"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregression"),
                &["progression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sceenshots"),
                &["screenshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transportarse"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("quanitify"), &["quantify"]),
            (dictgen::InsensitiveStr::Ascii("suceeding"), &["succeeding"]),
            (
                dictgen::InsensitiveStr::Ascii("discconeccts"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveyrones"), &["everyones"]),
            (dictgen::InsensitiveStr::Ascii("peremiter"), &["perimeter"]),
            (dictgen::InsensitiveStr::Ascii("levl"), &["level"]),
            (dictgen::InsensitiveStr::Ascii("stoream"), &["stream"]),
            (
                dictgen::InsensitiveStr::Ascii("civiliaztion"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("proeceudre"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("necessaray"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("trriger"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("blievers"), &["believers"]),
            (
                dictgen::InsensitiveStr::Ascii("confideny"),
                &["confidently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coprorates"),
                &["corporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("minise"), &["minimise"]),
            (
                dictgen::InsensitiveStr::Ascii("combinatation"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("rephras"), &["rephrase"]),
            (
                dictgen::InsensitiveStr::Ascii("virualise"),
                &["virtualise", "visualise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathmatically"),
                &["mathematically"],
            ),
            (dictgen::InsensitiveStr::Ascii("suceessor"), &["successor"]),
            (dictgen::InsensitiveStr::Ascii("terimnate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("ccpcheck"), &["cppcheck"]),
            (
                dictgen::InsensitiveStr::Ascii("pshycology"),
                &["psychology"],
            ),
            (dictgen::InsensitiveStr::Ascii("innocuos"), &["innocuous"]),
            (dictgen::InsensitiveStr::Ascii("recreted"), &["recreated"]),
            (
                dictgen::InsensitiveStr::Ascii("reasonnably"),
                &["reasonably"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcusing"), &["focusing"]),
            (
                dictgen::InsensitiveStr::Ascii("neveretheless"),
                &["nevertheless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncializing"),
                &["canonicalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("problably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("figuretively"),
                &["figuratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separeters"),
                &["separators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workstaition"),
                &["workstation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaruntes"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("psychedleic"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("randomrs"), &["randoms"]),
            (
                dictgen::InsensitiveStr::Ascii("scartching"),
                &["scratching"],
            ),
            (dictgen::InsensitiveStr::Ascii("appetities"), &["appetite"]),
            (dictgen::InsensitiveStr::Ascii("handelrs"), &["handlers"]),
            (
                dictgen::InsensitiveStr::Ascii("interactve"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoding"), &["holding"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhologists"),
                &["psychologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("suseed"), &["secede"]),
            (dictgen::InsensitiveStr::Ascii("literative"), &["literate"]),
            (dictgen::InsensitiveStr::Ascii("tungues"), &["tongues"]),
            (
                dictgen::InsensitiveStr::Ascii("interefaced"),
                &["interfaced"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensibel"), &["sensible"]),
            (dictgen::InsensitiveStr::Ascii("visilbity"), &["visibility"]),
            (dictgen::InsensitiveStr::Ascii("birdges"), &["bridges"]),
            (dictgen::InsensitiveStr::Ascii("heveanly"), &["heavenly"]),
            (
                dictgen::InsensitiveStr::Ascii("inadequete"),
                &["inadequate"],
            ),
            (dictgen::InsensitiveStr::Ascii("answhared"), &["answered"]),
            (
                dictgen::InsensitiveStr::Ascii("brillianty"),
                &["brilliantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ininity"),
                &["infinity", "inanity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritzing"),
                &["prioritizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obsolutism"),
                &["absolutism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polularity"),
                &["popularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("injest"), &["ingest"]),
            (dictgen::InsensitiveStr::Ascii("absolvin"), &["absolving"]),
            (
                dictgen::InsensitiveStr::Ascii("collaboratie"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("valif"), &["valid"]),
            (
                dictgen::InsensitiveStr::Ascii("infrustructures"),
                &["infrastructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kilomiters"),
                &["kilometers"],
            ),
            (dictgen::InsensitiveStr::Ascii("comaring"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("explins"), &["explains"]),
            (dictgen::InsensitiveStr::Ascii("dynmaic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("fritional"), &["frictional"]),
            (
                dictgen::InsensitiveStr::Ascii("differenciate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inefficiant"),
                &["inefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("receviers"), &["receivers"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequistes"),
                &["prerequisites"],
            ),
            (dictgen::InsensitiveStr::Ascii("reprodice"), &["reproduce"]),
            (dictgen::InsensitiveStr::Ascii("middel"), &["middle"]),
            (
                dictgen::InsensitiveStr::Ascii("difuse"),
                &["diffuse", "defuse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normlisation"),
                &["normalisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disatvantage"),
                &["disadvantage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("saveguarding"),
                &["safeguarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("looback"), &["loopback"]),
            (
                dictgen::InsensitiveStr::Ascii("tnankfully"),
                &["thankfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expection"),
                &["exception", "expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bleading"), &["bleeding"]),
            (
                dictgen::InsensitiveStr::Ascii("greenhosue"),
                &["greenhouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renaissence"),
                &["renaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrainting"),
                &["restraining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activiates"),
                &["activist", "activates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buisnesses"),
                &["businesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fingerpints"),
                &["fingerprints", "fingertips"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanse"),
                &["instances", "instance"],
            ),
            (dictgen::InsensitiveStr::Ascii("thenes"), &["themes"]),
            (
                dictgen::InsensitiveStr::Ascii("differnece"),
                &["difference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verificaions"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("faiures"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("signicant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("whther"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("pregancies"),
                &["pregnancies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stange"),
                &["strange", "stage", "stance"],
            ),
            (dictgen::InsensitiveStr::Ascii("manuels"), &["manuals"]),
            (
                dictgen::InsensitiveStr::Ascii("consisntently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throtlling"),
                &["throttling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sillibuses"),
                &["syllabuses"],
            ),
            (dictgen::InsensitiveStr::Ascii("mommentary"), &["momentary"]),
            (
                dictgen::InsensitiveStr::Ascii("constuction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresposive"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("directpry"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("fremented"), &["fermented"]),
            (
                dictgen::InsensitiveStr::Ascii("uninfroms"),
                &["uninforms", "uniforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("chlidren"), &["children"]),
            (
                dictgen::InsensitiveStr::Ascii("businessemen"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("omre"), &["more"]),
            (dictgen::InsensitiveStr::Ascii("softwares"), &["software"]),
            (dictgen::InsensitiveStr::Ascii("acccords"), &["accords"]),
            (
                dictgen::InsensitiveStr::Ascii("deregistartion"),
                &["deregistration"],
            ),
            (dictgen::InsensitiveStr::Ascii("decodded"), &["decoded"]),
            (dictgen::InsensitiveStr::Ascii("accending"), &["ascending"]),
            (dictgen::InsensitiveStr::Ascii("regsisters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("alterating"),
                &["alteration", "altering", "alternating"],
            ),
            (dictgen::InsensitiveStr::Ascii("namespce"), &["namespace"]),
            (
                dictgen::InsensitiveStr::Ascii("imablanced"),
                &["imbalanced"],
            ),
            (dictgen::InsensitiveStr::Ascii("diectly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("mergge"), &["merge"]),
            (dictgen::InsensitiveStr::Ascii("cermaic"), &["ceramic"]),
            (
                dictgen::InsensitiveStr::Ascii("relentlessely"),
                &["relentlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compantability"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custonisers"),
                &["customisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("apporoved"), &["approved"]),
            (dictgen::InsensitiveStr::Ascii("filtersing"), &["filtering"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamnetalists"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("appared"), &["appeared"]),
            (dictgen::InsensitiveStr::Ascii("improbing"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("diferrs"), &["differs"]),
            (dictgen::InsensitiveStr::Ascii("accorted"), &["accorded"]),
            (dictgen::InsensitiveStr::Ascii("contniues"), &["continues"]),
            (
                dictgen::InsensitiveStr::Ascii("accessoirez"),
                &["accessorize", "accessories"],
            ),
            (dictgen::InsensitiveStr::Ascii("refernces"), &["references"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occasinaly"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("februar"), &["february"]),
            (
                dictgen::InsensitiveStr::Ascii("discusions"),
                &["discussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chronice"), &["chronicles"]),
            (
                dictgen::InsensitiveStr::Ascii("equuivalence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchangees"), &["exchanges"]),
            (
                dictgen::InsensitiveStr::Ascii("riased"),
                &["biased", "raised"],
            ),
            (dictgen::InsensitiveStr::Ascii("mixure"), &["mixture"]),
            (dictgen::InsensitiveStr::Ascii("algoorithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("funtionally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("upport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("apapt"), &["adapt"]),
            (
                dictgen::InsensitiveStr::Ascii("exagerrating"),
                &["exaggerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowwed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("prosesses"),
                &["processes", "professes", "possesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignenmentes"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("intercat"), &["interact"]),
            (
                dictgen::InsensitiveStr::Ascii("legnedaries"),
                &["legendaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("victorina"), &["victorian"]),
            (dictgen::InsensitiveStr::Ascii("drowt"), &["drouth"]),
            (
                dictgen::InsensitiveStr::Ascii("chckerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copenhaguen"),
                &["copenhagen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genereators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylzers"), &["analyzers"]),
            (dictgen::InsensitiveStr::Ascii("kroeans"), &["koreans"]),
            (
                dictgen::InsensitiveStr::Ascii("spirituallly"),
                &["spiritually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patially"),
                &["partially", "spatially"],
            ),
            (dictgen::InsensitiveStr::Ascii("swith"), &["switch"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferantial"),
                &["preferential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("slite"),
                &["elite", "site", "sleight", "slide"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optizimations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ane"), &["and", "one"]),
            (
                dictgen::InsensitiveStr::Ascii("unneccecary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colliquial"),
                &["colloquial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furthremore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architurally"),
                &["architecturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatenade"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("airlfow"), &["airflow"]),
            (dictgen::InsensitiveStr::Ascii("sentance"), &["sentence"]),
            (dictgen::InsensitiveStr::Ascii("accep"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("hemoraged"),
                &["hemorrhaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciveing"), &["receiving"]),
            (
                dictgen::InsensitiveStr::Ascii("improbible"),
                &["improbable"],
            ),
            (dictgen::InsensitiveStr::Ascii("terroist"), &["terrorist"]),
            (dictgen::InsensitiveStr::Ascii("vermuth"), &["vermouth"]),
            (
                dictgen::InsensitiveStr::Ascii("disconecting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("beond"), &["beyond"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuition"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("wavelnes"), &["wavelines"]),
            (dictgen::InsensitiveStr::Ascii("srcipts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("arethmetic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("lingth"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalizatoin"),
                &["capitalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitionals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("decod"), &["decode"]),
            (dictgen::InsensitiveStr::Ascii("databse"), &["database"]),
            (
                dictgen::InsensitiveStr::Ascii("tangenitally"),
                &["tangentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("honeymooon"), &["honeymoon"]),
            (dictgen::InsensitiveStr::Ascii("limititer"), &["limiter"]),
            (dictgen::InsensitiveStr::Ascii("reverved"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("meny"), &["menu", "many"]),
            (
                dictgen::InsensitiveStr::Ascii("automacially"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrances"), &["arranges"]),
            (dictgen::InsensitiveStr::Ascii("cmmented"), &["commented"]),
            (dictgen::InsensitiveStr::Ascii("datebase"), &["database"]),
            (dictgen::InsensitiveStr::Ascii("paletable"), &["palatable"]),
            (
                dictgen::InsensitiveStr::Ascii("desagreements"),
                &["disagreements"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("scottisch"), &["scottish"]),
            (dictgen::InsensitiveStr::Ascii("swithc"), &["switch"]),
            (dictgen::InsensitiveStr::Ascii("obvisou"), &["obvious"]),
            (dictgen::InsensitiveStr::Ascii("videogaem"), &["videogame"]),
            (
                dictgen::InsensitiveStr::Ascii("conpensation"),
                &["compensation"],
            ),
            (dictgen::InsensitiveStr::Ascii("asympyote"), &["asymptote"]),
            (
                dictgen::InsensitiveStr::Ascii("jurisdiciton"),
                &["jurisdiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communin"),
                &["communing", "communion", "communis"],
            ),
            (dictgen::InsensitiveStr::Ascii("applyied"), &["applied"]),
            (
                dictgen::InsensitiveStr::Ascii("socializare"),
                &["socialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("porcess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("votage"),
                &["voltage", "voyage"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccorded"), &["accorded"]),
            (
                dictgen::InsensitiveStr::Ascii("ingerdients"),
                &["ingredients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsolicitied"),
                &["unsolicited"],
            ),
            (dictgen::InsensitiveStr::Ascii("plageries"), &["plagiaries"]),
            (dictgen::InsensitiveStr::Ascii("simplyfy"), &["simplify"]),
            (dictgen::InsensitiveStr::Ascii("excided"), &["exceeded"]),
            (dictgen::InsensitiveStr::Ascii("dashboars"), &["dashboards"]),
            (dictgen::InsensitiveStr::Ascii("earlyer"), &["earlier"]),
            (
                dictgen::InsensitiveStr::Ascii("enclude"),
                &["include", "exclude"],
            ),
            (dictgen::InsensitiveStr::Ascii("defintion"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("recognizible"),
                &["recognizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwee"), &["ennui"]),
            (
                dictgen::InsensitiveStr::Ascii("responciveness"),
                &["responsiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionel"),
                &["professional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eatch"),
                &["death", "each", "teach", "watch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senzationalist"),
                &["sensationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interperter"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("strcutre"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("dredentials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("cigarrett"), &["cigarette"]),
            (dictgen::InsensitiveStr::Ascii("undedined"), &["undefined"]),
            (dictgen::InsensitiveStr::Ascii("diasbled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("unesacpe"), &["unescape"]),
            (
                dictgen::InsensitiveStr::Ascii("micortransactions"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("linesw"), &["lines"]),
            (dictgen::InsensitiveStr::Ascii("carrotts"), &["carrots"]),
            (dictgen::InsensitiveStr::Ascii("mutliples"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("prospettive"),
                &["prospective"],
            ),
            (dictgen::InsensitiveStr::Ascii("recongise"), &["recognise"]),
            (dictgen::InsensitiveStr::Ascii("prayries"), &["prairies"]),
            (
                dictgen::InsensitiveStr::Ascii("hostorical"),
                &["historical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contestion"),
                &["contention"],
            ),
            (dictgen::InsensitiveStr::Ascii("coinside"), &["coincide"]),
            (
                dictgen::InsensitiveStr::Ascii("diminsional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intuitevely"),
                &["intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corinthias"),
                &["corinthians"],
            ),
            (dictgen::InsensitiveStr::Ascii("detrmined"), &["determined"]),
            (dictgen::InsensitiveStr::Ascii("avaluated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("syling"), &["styling"]),
            (dictgen::InsensitiveStr::Ascii("zelaots"), &["zealots"]),
            (
                dictgen::InsensitiveStr::Ascii("autocomple"),
                &["autocomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("characted"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("mamager"), &["manager"]),
            (dictgen::InsensitiveStr::Ascii("ehtnicity"), &["ethnicity"]),
            (
                dictgen::InsensitiveStr::Ascii("scavanging"),
                &["scavenging"],
            ),
            (dictgen::InsensitiveStr::Ascii("emualtor"), &["emulator"]),
            (
                dictgen::InsensitiveStr::Ascii("reassureing"),
                &["reassuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcotories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contingous"),
                &["contiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("adbomen"), &["abdomen"]),
            (
                dictgen::InsensitiveStr::Ascii("unauthenicated"),
                &["unauthenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenat"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("relise"), &["realise"]),
            (
                dictgen::InsensitiveStr::Ascii("confilcting"),
                &["conflicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitilised"),
                &["uninitialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("manges"), &["manages"]),
            (dictgen::InsensitiveStr::Ascii("perfomers"), &["performers"]),
            (dictgen::InsensitiveStr::Ascii("admiraals"), &["admirals"]),
            (dictgen::InsensitiveStr::Ascii("depenent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("gaurentee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("intresect"), &["intersect"]),
            (dictgen::InsensitiveStr::Ascii("differenct"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("noexistent"),
                &["nonexistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("crowbahr"), &["crowbar"]),
            (dictgen::InsensitiveStr::Ascii("buildding"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("blanced"),
                &["balanced", "glanced", "lanced"],
            ),
            (dictgen::InsensitiveStr::Ascii("enteprise"), &["enterprise"]),
            (
                dictgen::InsensitiveStr::Ascii("notificications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscriptin"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("squsihy"), &["squishy"]),
            (dictgen::InsensitiveStr::Ascii("masteris"), &["masteries"]),
            (dictgen::InsensitiveStr::Ascii("remanining"), &["remaining"]),
            (
                dictgen::InsensitiveStr::Ascii("resintallation"),
                &["reinstallation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("observating"),
                &["observation"],
            ),
            (dictgen::InsensitiveStr::Ascii("confrm"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("montions"),
                &["mentions", "motions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relocaiting"),
                &["relocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("androinds"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("bastracter"),
                &["abstracter"],
            ),
            (dictgen::InsensitiveStr::Ascii("recrutier"), &["recruiter"]),
            (
                dictgen::InsensitiveStr::Ascii("submarinas"),
                &["submarines"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereal"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("forwward"), &["forward"]),
            (dictgen::InsensitiveStr::Ascii("qeueing"), &["queueing"]),
            (dictgen::InsensitiveStr::Ascii("resluting"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("mangementt"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("thorughly"), &["thoroughly"]),
            (dictgen::InsensitiveStr::Ascii("nestolgic"), &["nostalgic"]),
            (
                dictgen::InsensitiveStr::Ascii("intervalles"),
                &["intervals"],
            ),
            (dictgen::InsensitiveStr::Ascii("dashboad"), &["dashboard"]),
            (dictgen::InsensitiveStr::Ascii("backfeild"), &["backfield"]),
            (
                dictgen::InsensitiveStr::Ascii("religioners"),
                &["religions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("powerprint"),
                &["powerpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pitchferks"),
                &["pitchforks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exeuctables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisond"), &["poisoned"]),
            (
                dictgen::InsensitiveStr::Ascii("mantioning"),
                &["mentioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missionares"),
                &["missionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transverers"),
                &["traversers", "transverters", "transferers"],
            ),
            (dictgen::InsensitiveStr::Ascii("javscript"), &["javascript"]),
            (dictgen::InsensitiveStr::Ascii("internface"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregeted"),
                &["aggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("rith"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unattendend"),
                &["unattended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinare"),
                &["determine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscociates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relentlessley"),
                &["relentlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("kubirck"), &["kubrick"]),
            (
                dictgen::InsensitiveStr::Ascii("consideratoin"),
                &["considerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("probablle"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("relcation"), &["relocation"]),
            (
                dictgen::InsensitiveStr::Ascii("costumizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("eratic"), &["erratic"]),
            (
                dictgen::InsensitiveStr::Ascii("neigbouring"),
                &["neighbouring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enchancment"),
                &["enchantment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubelshoot"),
                &["troubleshoot"],
            ),
            (dictgen::InsensitiveStr::Ascii("appplying"), &["applying"]),
            (dictgen::InsensitiveStr::Ascii("dirceted"), &["directed"]),
            (dictgen::InsensitiveStr::Ascii("nevelop"), &["envelop"]),
            (
                dictgen::InsensitiveStr::Ascii("encrcypting"),
                &["encrypting"],
            ),
            (dictgen::InsensitiveStr::Ascii("resourece"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilysing"),
                &["cannibalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("contracr"), &["contractor"]),
            (
                dictgen::InsensitiveStr::Ascii("aletrnatively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("calculat"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("driveing"), &["driving"]),
            (
                dictgen::InsensitiveStr::Ascii("serviciable"),
                &["serviceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satistical"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derviative"),
                &["derivative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualion"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("afer"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("internvals"), &["intervals"]),
            (dictgen::InsensitiveStr::Ascii("shanghia"), &["shanghai"]),
            (
                dictgen::InsensitiveStr::Ascii("publicizin"),
                &["publicizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undersood"),
                &["understood", "undersold"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missconfigures"),
                &["misconfigures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramedicks"),
                &["paramedics"],
            ),
            (dictgen::InsensitiveStr::Ascii("negoziate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("differene"),
                &["differences", "difference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mositurizing"),
                &["moisturizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avent"),
                &["advent", "event"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuded"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("foloowiwng"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourgs"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("epxands"), &["expands"]),
            (dictgen::InsensitiveStr::Ascii("audtion"), &["audition"]),
            (
                dictgen::InsensitiveStr::Ascii("sponsorhips"),
                &["sponsorship", "sponsorships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterrupted"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("accademy"), &["academy"]),
            (dictgen::InsensitiveStr::Ascii("compnents"), &["components"]),
            (dictgen::InsensitiveStr::Ascii("throough"), &["thorough"]),
            (dictgen::InsensitiveStr::Ascii("regarg"), &["regard"]),
            (
                dictgen::InsensitiveStr::Ascii("develompent"),
                &["developments", "development"],
            ),
            (dictgen::InsensitiveStr::Ascii("gam"), &["game"]),
            (
                dictgen::InsensitiveStr::Ascii("ligthweight"),
                &["lightweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("recommed"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("questoinnaire"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekced"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("anicents"), &["ancients"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtentication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaters"),
                &["initiators", "initiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("activisim"), &["activism"]),
            (dictgen::InsensitiveStr::Ascii("acccording"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("connetors"), &["connectors"]),
            (dictgen::InsensitiveStr::Ascii("ceck"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("divsiors"), &["divisors"]),
            (dictgen::InsensitiveStr::Ascii("porcelina"), &["porcelain"]),
            (
                dictgen::InsensitiveStr::Ascii("calssrooms"),
                &["classrooms"],
            ),
            (dictgen::InsensitiveStr::Ascii("phramacy"), &["pharmacy"]),
            (dictgen::InsensitiveStr::Ascii("easiy"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("warrn"), &["warn"]),
            (dictgen::InsensitiveStr::Ascii("clloses"), &["closes"]),
            (dictgen::InsensitiveStr::Ascii("comany"), &["company"]),
            (
                dictgen::InsensitiveStr::Ascii("orthonormalizatin"),
                &["orthonormalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("antiviurs"), &["antivirus"]),
            (dictgen::InsensitiveStr::Ascii("mainfesto"), &["manifesto"]),
            (dictgen::InsensitiveStr::Ascii("updter"), &["updater"]),
            (
                dictgen::InsensitiveStr::Ascii("propietry"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("descripes"), &["describes"]),
            (dictgen::InsensitiveStr::Ascii("boold"), &["bold", "bool"]),
            (dictgen::InsensitiveStr::Ascii("locatoin"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("legnth"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("micoscopic"),
                &["microscopic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architetural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("aactual"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("firtsly"), &["firstly"]),
            (
                dictgen::InsensitiveStr::Ascii("formativos"),
                &["formations"],
            ),
            (dictgen::InsensitiveStr::Ascii("derypt"), &["decrypt"]),
            (dictgen::InsensitiveStr::Ascii("misogynt"), &["misogynist"]),
            (
                dictgen::InsensitiveStr::Ascii("jorunaling"),
                &["journaling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improofement"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecesasrily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conted"),
                &["counted", "counter", "countered"],
            ),
            (dictgen::InsensitiveStr::Ascii("tamplated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("redudent"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("suplementary"),
                &["supplementary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlefied"),
                &["battlefield"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understandible"),
                &["understandable"],
            ),
            (dictgen::InsensitiveStr::Ascii("similiair"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("intangibile"),
                &["intangible"],
            ),
            (dictgen::InsensitiveStr::Ascii("microship"), &["microchip"]),
            (dictgen::InsensitiveStr::Ascii("partitial"), &["partial"]),
            (
                dictgen::InsensitiveStr::Ascii("assocaitions"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negativelly"),
                &["negatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactuer"),
                &["manufacture", "manufacturer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufficating"),
                &["suffocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlyaing"),
                &["overlaying"],
            ),
            (dictgen::InsensitiveStr::Ascii("scientifc"), &["scientific"]),
            (
                dictgen::InsensitiveStr::Ascii("authroritative"),
                &["authoritative"],
            ),
            (dictgen::InsensitiveStr::Ascii("portioon"), &["portion"]),
            (dictgen::InsensitiveStr::Ascii("inalid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("prometheas"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("nilihism"), &["nihilism"]),
            (dictgen::InsensitiveStr::Ascii("excactly"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("inititializer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("situaions"), &["situations"]),
            (dictgen::InsensitiveStr::Ascii("updaed"), &["updated"]),
            (
                dictgen::InsensitiveStr::Ascii("themselve"),
                &["themselves", "themself"],
            ),
            (dictgen::InsensitiveStr::Ascii("vasectomey"), &["vasectomy"]),
            (
                dictgen::InsensitiveStr::Ascii("oglin"),
                &["ogling", "goblin"],
            ),
            (dictgen::InsensitiveStr::Ascii("randum"), &["random"]),
            (
                dictgen::InsensitiveStr::Ascii("cappability"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("conncting"), &["connecting"]),
            (
                dictgen::InsensitiveStr::Ascii("selction"),
                &["selection", "section"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscarraiges"),
                &["miscarriage"],
            ),
            (dictgen::InsensitiveStr::Ascii("identier"), &["identifier"]),
            (
                dictgen::InsensitiveStr::Ascii("consecutevily"),
                &["consecutively"],
            ),
            (dictgen::InsensitiveStr::Ascii("pinoneered"), &["pioneered"]),
            (
                dictgen::InsensitiveStr::Ascii("environmontal"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refartor"),
                &["refactor", "refractor"],
            ),
            (dictgen::InsensitiveStr::Ascii("fluctaute"), &["fluctuate"]),
            (dictgen::InsensitiveStr::Ascii("racaus"), &["raucous"]),
            (
                dictgen::InsensitiveStr::Ascii("periperhal"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permormance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahdidate"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("clevery"), &["cleverly"]),
            (dictgen::InsensitiveStr::Ascii("shreshold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("univerities"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("confure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunalty"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("offseting"), &["offsetting"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginzers"),
                &["recognizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("testomony"), &["testimony"]),
            (
                dictgen::InsensitiveStr::Ascii("ajurnment"),
                &["adjournment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randayvoos"),
                &["rendezvous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amelearator"),
                &["ameliorator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uneccessarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("transferts"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("ungoldy"), &["ungodly"]),
            (
                dictgen::InsensitiveStr::Ascii("proceeedures"),
                &["procedures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("examinining"),
                &["examining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inappropriatelly"),
                &["inappropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compiliance"),
                &["compliance"],
            ),
            (dictgen::InsensitiveStr::Ascii("plarize"), &["polarize"]),
            (dictgen::InsensitiveStr::Ascii("wayoint"), &["waypoint"]),
            (
                dictgen::InsensitiveStr::Ascii("restaraunts"),
                &["restaurants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discretiztions"),
                &["discretizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("donw"), &["down", "done"]),
            (dictgen::InsensitiveStr::Ascii("awkard"), &["awkward"]),
            (dictgen::InsensitiveStr::Ascii("compresed"), &["compressed"]),
            (
                dictgen::InsensitiveStr::Ascii("persentations"),
                &["presentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("spaws"), &["spawns"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsitancy"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("veneuzela"), &["venezuela"]),
            (
                dictgen::InsensitiveStr::Ascii("apologistics"),
                &["apologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("curefully"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("subclasseses"),
                &["subclasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beautifull"),
                &["beautiful", "beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstratin"),
                &["demonstration", "demonstrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("comipled"), &["compiled"]),
            (dictgen::InsensitiveStr::Ascii("aquarim"), &["aquarium"]),
            (
                dictgen::InsensitiveStr::Ascii("acciednetally"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dalmations"),
                &["dalmatians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomponying"),
                &["accompanying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("engeneering"),
                &["engineering"],
            ),
            (dictgen::InsensitiveStr::Ascii("comerant"), &["cormorant"]),
            (
                dictgen::InsensitiveStr::Ascii("reattachement"),
                &["reattachment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oints"),
                &["points", "pints", "joints", "lints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulamentation"),
                &["regulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comversion"),
                &["conversion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heretosexual"),
                &["heterosexual"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifcic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("dallocate"), &["deallocate"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthrouhgh"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("dividin"), &["dividing"]),
            (dictgen::InsensitiveStr::Ascii("spefy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("artificailly"),
                &["artificially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imrpovement"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceutives"),
                &["executives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repentence"),
                &["repentance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusaists"),
                &["enthusiasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scrit"),
                &["scrip", "script"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automattically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("experiece"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("atmosphir"),
                &["atmospheric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilties"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("numearate"), &["numerate"]),
            (
                dictgen::InsensitiveStr::Ascii("booloaders"),
                &["bootloaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polynomical"),
                &["polynomial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aotomatical"),
                &["automatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sourrounding"),
                &["surrounding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validatons"),
                &["validations", "validators"],
            ),
            (dictgen::InsensitiveStr::Ascii("traspose"), &["transpose"]),
            (
                dictgen::InsensitiveStr::Ascii("philippides"),
                &["philippines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intorduction"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("sublcuase"), &["subclause"]),
            (dictgen::InsensitiveStr::Ascii("obolete"), &["obsolete"]),
            (dictgen::InsensitiveStr::Ascii("seguoys"), &["segues"]),
            (
                dictgen::InsensitiveStr::Ascii("rasterizor"),
                &["rasterizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitiaions"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("becaming"), &["becoming"]),
            (dictgen::InsensitiveStr::Ascii("paragaphs"), &["paragraphs"]),
            (dictgen::InsensitiveStr::Ascii("modfities"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("diffculties"),
                &["difficulties"],
            ),
            (dictgen::InsensitiveStr::Ascii("debugggs"), &["debugs"]),
            (
                dictgen::InsensitiveStr::Ascii("guarenteers"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("chckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("gylph"), &["glyph"]),
            (
                dictgen::InsensitiveStr::Ascii("apocalypes"),
                &["apocalypse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dperecates"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("som"), &["some"]),
            (dictgen::InsensitiveStr::Ascii("longtitude"), &["longitude"]),
            (
                dictgen::InsensitiveStr::Ascii("negothiator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutlithreaded"),
                &["multithreaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("intiation"), &["initiation"]),
            (
                dictgen::InsensitiveStr::Ascii("broadcastors"),
                &["broadcasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("phsically"), &["physically"]),
            (
                dictgen::InsensitiveStr::Ascii("initalizable"),
                &["initializable"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfoms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("crahed"), &["crashed"]),
            (
                dictgen::InsensitiveStr::Ascii("indien"),
                &["indian", "endian"],
            ),
            (dictgen::InsensitiveStr::Ascii("exibited"), &["exhibited"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterized"),
                &["characterized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoritate"),
                &["authoritative"],
            ),
            (dictgen::InsensitiveStr::Ascii("transofrm"), &["transform"]),
            (
                dictgen::InsensitiveStr::Ascii("enumearation"),
                &["enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtractly"),
                &["abstractly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boootstrapper"),
                &["bootstrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("inster"), &["insert"]),
            (
                dictgen::InsensitiveStr::Ascii("conceviably"),
                &["conceivably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indecators"),
                &["indicators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facilitatin"),
                &["facilitating", "facilitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("otpions"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaing"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destiation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discared"),
                &["discarded", "discard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promethese"),
                &["prometheus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explorerers"),
                &["explorers"],
            ),
            (dictgen::InsensitiveStr::Ascii("assagne"), &["assange"]),
            (
                dictgen::InsensitiveStr::Ascii("thumbnials"),
                &["thumbnails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrisincs"),
                &["intrinsics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexepected"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocarnia"), &["ocarina"]),
            (dictgen::InsensitiveStr::Ascii("swithable"), &["switchable"]),
            (
                dictgen::InsensitiveStr::Ascii("hightlighting"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabillity"),
                &["probability"],
            ),
            (dictgen::InsensitiveStr::Ascii("enebaled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("creens"),
                &["screens", "greens", "creeps", "careens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("youforically"),
                &["euphorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefifeir"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("caugt"), &["caught"]),
            (dictgen::InsensitiveStr::Ascii("thow"), &["throw", "tow"]),
            (
                dictgen::InsensitiveStr::Ascii("attrbiuting"),
                &["attributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indentifers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rationallity"),
                &["rationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoicative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierarachies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("garabge"), &["garbage"]),
            (dictgen::InsensitiveStr::Ascii("milions"), &["millions"]),
            (
                dictgen::InsensitiveStr::Ascii("shortining"),
                &["shortening"],
            ),
            (dictgen::InsensitiveStr::Ascii("transter"), &["transfer"]),
            (dictgen::InsensitiveStr::Ascii("inbetweeen"), &["inbetween"]),
            (dictgen::InsensitiveStr::Ascii("anly"), &["only", "any"]),
            (dictgen::InsensitiveStr::Ascii("opimize"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("abstratness"),
                &["abstractness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destibation"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("digitis"), &["digits"]),
            (
                dictgen::InsensitiveStr::Ascii("antrhopology"),
                &["anthropology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itterations"),
                &["iterations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agrandized"),
                &["aggrandized"],
            ),
            (dictgen::InsensitiveStr::Ascii("taregt"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("sepparations"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("difussion"), &["diffusion"]),
            (dictgen::InsensitiveStr::Ascii("astrixes"), &["asterisks"]),
            (dictgen::InsensitiveStr::Ascii("functiom"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("propatagion"),
                &["propagation"],
            ),
            (dictgen::InsensitiveStr::Ascii("progaramm"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiecation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("authers"), &["authors"]),
            (
                dictgen::InsensitiveStr::Ascii("behaivoural"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("opperated"), &["operated"]),
            (dictgen::InsensitiveStr::Ascii("resoource"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("swithed"), &["switched"]),
            (dictgen::InsensitiveStr::Ascii("misson"), &["mission"]),
            (dictgen::InsensitiveStr::Ascii("correc"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("develpoments"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("horozon"), &["horizon"]),
            (
                dictgen::InsensitiveStr::Ascii("indecating"),
                &["indicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoud"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("nucular"), &["nuclear"]),
            (dictgen::InsensitiveStr::Ascii("imformed"), &["informed"]),
            (
                dictgen::InsensitiveStr::Ascii("resgistered"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combnation"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribusion"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delaratively"),
                &["declaratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("pupulate"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("combiniations"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("greenhoe"), &["greenhouse"]),
            (
                dictgen::InsensitiveStr::Ascii("unforunately"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repesentation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propogation"),
                &["propagation"],
            ),
            (dictgen::InsensitiveStr::Ascii("baristia"), &["barista"]),
            (dictgen::InsensitiveStr::Ascii("prorgress"), &["progress"]),
            (
                dictgen::InsensitiveStr::Ascii("disrespekt"),
                &["disrespect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreviato"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("enablen"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("libertariaism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("souces"), &["sources"]),
            (
                dictgen::InsensitiveStr::Ascii("recommondation"),
                &["recommendation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nacrotics"), &["narcotics"]),
            (
                dictgen::InsensitiveStr::Ascii("groudnbreaking"),
                &["groundbreaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("discoverd"), &["discovered"]),
            (dictgen::InsensitiveStr::Ascii("downoaded"), &["downloaded"]),
            (dictgen::InsensitiveStr::Ascii("sleects"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("autorealease"),
                &["autorelease"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicabile"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizzelle"), &["chisel"]),
            (dictgen::InsensitiveStr::Ascii("defianlty"), &["defiantly"]),
            (
                dictgen::InsensitiveStr::Ascii("phychologists"),
                &["psychologists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estiamative"),
                &["estimative"],
            ),
            (dictgen::InsensitiveStr::Ascii("consummes"), &["consumes"]),
            (dictgen::InsensitiveStr::Ascii("hallowean"), &["halloween"]),
            (dictgen::InsensitiveStr::Ascii("rearrage"), &["rearrange"]),
            (dictgen::InsensitiveStr::Ascii("infiltre"), &["infiltrate"]),
            (
                dictgen::InsensitiveStr::Ascii("unaviodable"),
                &["unavoidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surelly"),
                &["surely", "surreally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rinosaruss"),
                &["rhinoceroses"],
            ),
            (dictgen::InsensitiveStr::Ascii("grop"), &["group", "drop"]),
            (dictgen::InsensitiveStr::Ascii("finilizes"), &["finalizes"]),
            (
                dictgen::InsensitiveStr::Ascii("duplacation"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inflatible"),
                &["inflatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commonweatlh"),
                &["commonwealth"],
            ),
            (dictgen::InsensitiveStr::Ascii("vlued"), &["valued"]),
            (
                dictgen::InsensitiveStr::Ascii("enumarations"),
                &["enumerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguarls"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("japaneese"), &["japanese"]),
            (
                dictgen::InsensitiveStr::Ascii("thron"),
                &["thrown", "throne"],
            ),
            (dictgen::InsensitiveStr::Ascii("goups"), &["groups"]),
            (dictgen::InsensitiveStr::Ascii("interioara"), &["interior"]),
            (
                dictgen::InsensitiveStr::Ascii("scarch"),
                &["search", "scorch", "scratch", "starch", "scarce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congradulations"),
                &["congratulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("frquent"), &["frequent"]),
            (dictgen::InsensitiveStr::Ascii("addrersss"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("reliabily"),
                &["reliability", "reliably"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyhsicaly"), &["physically"]),
            (
                dictgen::InsensitiveStr::Ascii("inferfacing"),
                &["interfacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interseccts"),
                &["intersects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfitication"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registerered"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallifes"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("constats"), &["constants"]),
            (
                dictgen::InsensitiveStr::Ascii("cataclismic"),
                &["cataclysmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegothiating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("assiged"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("antaganists"),
                &["antagonists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexpense"),
                &["inexpensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partisioning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("suger"), &["sugar"]),
            (dictgen::InsensitiveStr::Ascii("sctosman"), &["scotsman"]),
            (
                dictgen::InsensitiveStr::Ascii("corrosponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrespondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("nusiance"), &["nuisance"]),
            (dictgen::InsensitiveStr::Ascii("moprh"), &["morph"]),
            (dictgen::InsensitiveStr::Ascii("recipiant"), &["recipient"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchorous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functinoally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("purcahsed"), &["purchased"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncally"),
                &["canonically"],
            ),
            (dictgen::InsensitiveStr::Ascii("trriggers"), &["triggers"]),
            (dictgen::InsensitiveStr::Ascii("selecetion"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("machne"), &["machine"]),
            (dictgen::InsensitiveStr::Ascii("shileding"), &["shielding"]),
            (
                dictgen::InsensitiveStr::Ascii("consentrating"),
                &["concentrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monothilic"),
                &["monolithic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convient"),
                &["convenient", "convent"],
            ),
            (dictgen::InsensitiveStr::Ascii("istening"), &["listening"]),
            (dictgen::InsensitiveStr::Ascii("obseleted"), &["obsoleted"]),
            (dictgen::InsensitiveStr::Ascii("senaireo"), &["scenario"]),
            (dictgen::InsensitiveStr::Ascii("misstypes"), &["mistypes"]),
            (
                dictgen::InsensitiveStr::Ascii("authroising"),
                &["authorising"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnnel"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("maraduer"), &["marauder"]),
            (
                dictgen::InsensitiveStr::Ascii("dipolmatic"),
                &["diplomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("daimond"), &["diamond"]),
            (dictgen::InsensitiveStr::Ascii("brin"), &["bring"]),
            (
                dictgen::InsensitiveStr::Ascii("disastisfied"),
                &["dissatisfied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cnversation"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelereation"),
                &["acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempermental"),
                &["temperamental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralelised"),
                &["parallelised"],
            ),
            (dictgen::InsensitiveStr::Ascii("spepc"), &["spec"]),
            (
                dictgen::InsensitiveStr::Ascii("chromosoms"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("usefull"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("picnickin"), &["picnicking"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabiltiy"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promsicuous"),
                &["promiscuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adivse"),
                &["advise", "advice"],
            ),
            (dictgen::InsensitiveStr::Ascii("proctects"), &["protects"]),
            (
                dictgen::InsensitiveStr::Ascii("indocternated"),
                &["indoctrinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxited"), &["excited"]),
            (
                dictgen::InsensitiveStr::Ascii("recommpiled"),
                &["recompiled"],
            ),
            (dictgen::InsensitiveStr::Ascii("smaple"), &["sample"]),
            (
                dictgen::InsensitiveStr::Ascii("dimemsions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimziers"),
                &["optimizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apeendages"),
                &["appendages"],
            ),
            (dictgen::InsensitiveStr::Ascii("filller"), &["filler"]),
            (
                dictgen::InsensitiveStr::Ascii("boxe"),
                &["boxes", "box", "boxer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postitively"),
                &["positively"],
            ),
            (dictgen::InsensitiveStr::Ascii("colera"), &["cholera"]),
            (dictgen::InsensitiveStr::Ascii("autistisch"), &["autistic"]),
            (dictgen::InsensitiveStr::Ascii("availablee"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasmission"),
                &["transmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mischieveously"),
                &["mischievously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restransmitted"),
                &["retransmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declrations"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("fatigure"), &["fatigue"]),
            (
                dictgen::InsensitiveStr::Ascii("denomimators"),
                &["denominators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparible"),
                &["compatible", "comparable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superseedd"),
                &["superseded"],
            ),
            (dictgen::InsensitiveStr::Ascii("composits"), &["composites"]),
            (dictgen::InsensitiveStr::Ascii("dispoves"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("clonin"), &["cloning"]),
            (dictgen::InsensitiveStr::Ascii("conesencus"), &["consensus"]),
            (dictgen::InsensitiveStr::Ascii("physiscian"), &["physician"]),
            (dictgen::InsensitiveStr::Ascii("haording"), &["hoarding"]),
            (
                dictgen::InsensitiveStr::Ascii("abandonning"),
                &["abandoning"],
            ),
            (dictgen::InsensitiveStr::Ascii("variabeles"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("invesitgate"),
                &["investigate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inplemented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adquisition"),
                &["acquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("ihaca"), &["ithaca"]),
            (dictgen::InsensitiveStr::Ascii("tichness"), &["thickness"]),
            (
                dictgen::InsensitiveStr::Ascii("aspriations"),
                &["aspirations"],
            ),
            (dictgen::InsensitiveStr::Ascii("involtues"), &["involutes"]),
            (dictgen::InsensitiveStr::Ascii("hime"), &["him"]),
            (dictgen::InsensitiveStr::Ascii("meauring"), &["measuring"]),
            (
                dictgen::InsensitiveStr::Ascii("discussiong"),
                &["discussing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siloueting"),
                &["silhouetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("istantiated"),
                &["instantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instroduction"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolerence"), &["tolerance"]),
            (
                dictgen::InsensitiveStr::Ascii("buissinesses"),
                &["businesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekcsums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("temaplating"),
                &["templating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreliablity"),
                &["unreliability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counceling"),
                &["counseling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creationisim"),
                &["creationism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorsments"),
                &["endorsements"],
            ),
            (dictgen::InsensitiveStr::Ascii("awming"), &["awning"]),
            (
                dictgen::InsensitiveStr::Ascii("overlowed"),
                &["overflowed", "overloaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitlaism"),
                &["capitalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("explan"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("attepmpt"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("ediable"), &["editable"]),
            (dictgen::InsensitiveStr::Ascii("degradead"), &["degraded"]),
            (dictgen::InsensitiveStr::Ascii("conservare"), &["conserve"]),
            (dictgen::InsensitiveStr::Ascii("anaylzer"), &["analyzer"]),
            (dictgen::InsensitiveStr::Ascii("origniate"), &["originate"]),
            (
                dictgen::InsensitiveStr::Ascii("metropolois"),
                &["metropolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converdation"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("volumetic"), &["volumetric"]),
            (dictgen::InsensitiveStr::Ascii("allmost"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("noify"),
                &["notify", "noisy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parlimentary"),
                &["parliamentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("credis"), &["credits"]),
            (dictgen::InsensitiveStr::Ascii("conetor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("cripts"),
                &["crypts", "scripts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underpowred"),
                &["underpowered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("varriations"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("birghten"), &["brighten"]),
            (dictgen::InsensitiveStr::Ascii("thraeds"), &["threads"]),
            (dictgen::InsensitiveStr::Ascii("necesasary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("hexigons"), &["hexagons"]),
            (dictgen::InsensitiveStr::Ascii("centipedle"), &["centipede"]),
            (
                dictgen::InsensitiveStr::Ascii("overlappin"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterizes"),
                &["parameterizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scourcer"),
                &["scourer", "sorcerer", "scouser"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleived"), &["believed"]),
            (dictgen::InsensitiveStr::Ascii("actaly"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("expiriation"),
                &["expiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertix"), &["vertex"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifiaction"),
                &["qualification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannabilyzed"),
                &["cannibalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("vauge"), &["vague"]),
            (
                dictgen::InsensitiveStr::Ascii("misoginysts"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("enquing"), &["enqueuing"]),
            (dictgen::InsensitiveStr::Ascii("anohter"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("insignificat"),
                &["insignificant"],
            ),
            (dictgen::InsensitiveStr::Ascii("projetcted"), &["projected"]),
            (
                dictgen::InsensitiveStr::Ascii("altenating"),
                &["alternating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refactrion"),
                &["refraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valentiens"),
                &["valentines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performancewise"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitations"), &["situations"]),
            (dictgen::InsensitiveStr::Ascii("cannont"), &["cannot"]),
            (
                dictgen::InsensitiveStr::Ascii("probematic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplication"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwritin"),
                &["overwriting"],
            ),
            (dictgen::InsensitiveStr::Ascii("unforseen"), &["unforeseen"]),
            (dictgen::InsensitiveStr::Ascii("raosting"), &["roasting"]),
            (dictgen::InsensitiveStr::Ascii("retyring"), &["retrying"]),
            (
                dictgen::InsensitiveStr::Ascii("sinal"),
                &["signal", "spinal"],
            ),
            (dictgen::InsensitiveStr::Ascii("varifies"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("centrilised"),
                &["centralised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactersitic"),
                &["characteristics", "characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("disires"), &["desires"]),
            (dictgen::InsensitiveStr::Ascii("directsion"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("bakcground"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assymtotically"),
                &["asymptotically"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentienl"), &["sentinel"]),
            (dictgen::InsensitiveStr::Ascii("submite"), &["submit"]),
            (
                dictgen::InsensitiveStr::Ascii("projecitons"),
                &["projections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utilitisation"),
                &["utilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeleted"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("preventation"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("excerised"), &["exercised"]),
            (
                dictgen::InsensitiveStr::Ascii("reprocuces"),
                &["reproduces", "reprocures"],
            ),
            (dictgen::InsensitiveStr::Ascii("shouuld"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("spritns"), &["sprints"]),
            (
                dictgen::InsensitiveStr::Ascii("corralated"),
                &["correlated"],
            ),
            (dictgen::InsensitiveStr::Ascii("alreadt"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("conquerring"),
                &["conquering"],
            ),
            (dictgen::InsensitiveStr::Ascii("appoves"), &["approves"]),
            (dictgen::InsensitiveStr::Ascii("bethedsa"), &["bethesda"]),
            (dictgen::InsensitiveStr::Ascii("solerly"), &["solely"]),
            (
                dictgen::InsensitiveStr::Ascii("accesssibility"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("symplify"), &["simplify"]),
            (dictgen::InsensitiveStr::Ascii("prooblems"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("sometihn"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("auotmating"),
                &["automating"],
            ),
            (dictgen::InsensitiveStr::Ascii("selcets"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("guarunteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auotattack"),
                &["autoattack"],
            ),
            (dictgen::InsensitiveStr::Ascii("depond"), &["depend"]),
            (
                dictgen::InsensitiveStr::Ascii("embaraasing"),
                &["embarrassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexul"),
                &["homosexuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("tieth"), &["tithe"]),
            (dictgen::InsensitiveStr::Ascii("abstrctly"), &["abstractly"]),
            (
                dictgen::InsensitiveStr::Ascii("comparisson"),
                &["comparisons", "comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcationing"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extermporaneous"),
                &["extemporaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("answerting"), &["answering"]),
            (
                dictgen::InsensitiveStr::Ascii("legendariske"),
                &["legendaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomadation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undersatnds"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("deisgn"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("refesh"), &["refresh"]),
            (dictgen::InsensitiveStr::Ascii("multilies"), &["multiplies"]),
            (
                dictgen::InsensitiveStr::Ascii("fondamental"),
                &["fundamental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("samaphores"),
                &["semaphores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eficiently"),
                &["efficiently", "deficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("fronteir"), &["frontier"]),
            (dictgen::InsensitiveStr::Ascii("animatted"), &["animated"]),
            (dictgen::InsensitiveStr::Ascii("annotaged"), &["annotated"]),
            (
                dictgen::InsensitiveStr::Ascii("etamologies"),
                &["etymologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contatenations"),
                &["concatenations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percusssion"),
                &["percussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubstanciated"),
                &["unsubstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("soemwhat"), &["somewhat"]),
            (dictgen::InsensitiveStr::Ascii("stiches"), &["stitches"]),
            (dictgen::InsensitiveStr::Ascii("syntakts"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("neighobrhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("devlove"), &["devolve"]),
            (dictgen::InsensitiveStr::Ascii("allws"), &["allows"]),
            (dictgen::InsensitiveStr::Ascii("maxiize"), &["maximize"]),
            (
                dictgen::InsensitiveStr::Ascii("referneced"),
                &["referenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unacknownledged"),
                &["unacknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("contxt"), &["context"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrafices"),
                &["sacrifices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuaring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automotove"),
                &["automotive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restauration"),
                &["restoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("attackeras"), &["attackers"]),
            (dictgen::InsensitiveStr::Ascii("debuged"), &["debugged"]),
            (dictgen::InsensitiveStr::Ascii("intrinsict"), &["intrinsic"]),
            (
                dictgen::InsensitiveStr::Ascii("regenertive"),
                &["regenerative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("youthinasia"),
                &["euthanasia"],
            ),
            (dictgen::InsensitiveStr::Ascii("ruleboook"), &["rulebook"]),
            (dictgen::InsensitiveStr::Ascii("domiantes"), &["dominates"]),
            (dictgen::InsensitiveStr::Ascii("nationas"), &["nationals"]),
            (dictgen::InsensitiveStr::Ascii("finil"), &["final"]),
            (
                dictgen::InsensitiveStr::Ascii("homoepathy"),
                &["homoeopathy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarenteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicetely"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("summming"), &["summing"]),
            (dictgen::InsensitiveStr::Ascii("cleverleys"), &["cleverly"]),
            (
                dictgen::InsensitiveStr::Ascii("requied"),
                &["required", "requited"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterface"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("inacuraccies"),
                &["inaccuracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulider"), &["builder"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistencias"),
                &["inconsistencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("discuess"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("darws"), &["draws"]),
            (dictgen::InsensitiveStr::Ascii("stupidfree"), &["stupider"]),
            (
                dictgen::InsensitiveStr::Ascii("enforsers"),
                &["endorsers", "enforcers"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceoting"), &["accepting"]),
            (dictgen::InsensitiveStr::Ascii("elcipse"), &["eclipse"]),
            (dictgen::InsensitiveStr::Ascii("strores"), &["stores"]),
            (
                dictgen::InsensitiveStr::Ascii("tranceiver"),
                &["transceiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronizer"),
                &["synchronizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("rabits"), &["rabbits"]),
            (dictgen::InsensitiveStr::Ascii("brigthest"), &["brightest"]),
            (
                dictgen::InsensitiveStr::Ascii("ralationships"),
                &["relationships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innitialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occourence"),
                &["occurrence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cllouded"), &["clouded"]),
            (
                dictgen::InsensitiveStr::Ascii("sucesssful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("continu"), &["continue"]),
            (dictgen::InsensitiveStr::Ascii("alledge"), &["allege"]),
            (dictgen::InsensitiveStr::Ascii("mininum"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("asymetri"),
                &["asymmetric", "asymmetry"],
            ),
            (dictgen::InsensitiveStr::Ascii("allegric"), &["allergic"]),
            (dictgen::InsensitiveStr::Ascii("derageable"), &["dirigible"]),
            (
                dictgen::InsensitiveStr::Ascii("convincente"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vallays"),
                &["valets", "valleys"],
            ),
            (dictgen::InsensitiveStr::Ascii("counrty"), &["country"]),
            (
                dictgen::InsensitiveStr::Ascii("neeeed"),
                &["need", "needed"],
            ),
            (dictgen::InsensitiveStr::Ascii("hrlper"), &["helper"]),
            (dictgen::InsensitiveStr::Ascii("snaphsot"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("conneted"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("disticts"), &["districts"]),
            (dictgen::InsensitiveStr::Ascii("cetifies"), &["certifies"]),
            (dictgen::InsensitiveStr::Ascii("vaalue"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("attacthed"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("beinning"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("valule"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("inidicates"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("illegetimate"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conlcuding"),
                &["concluding"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorthn"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("physcology"),
                &["psychology"],
            ),
            (dictgen::InsensitiveStr::Ascii("volxel"), &["voxel"]),
            (dictgen::InsensitiveStr::Ascii("genarate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("datatbase"), &["database"]),
            (
                dictgen::InsensitiveStr::Ascii("constucted"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("templaras"), &["templars"]),
            (dictgen::InsensitiveStr::Ascii("eczecute"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("collissions"),
                &["collisions"],
            ),
            (dictgen::InsensitiveStr::Ascii("restiring"), &["restoring"]),
            (dictgen::InsensitiveStr::Ascii("fashoined"), &["fashioned"]),
            (dictgen::InsensitiveStr::Ascii("liased"), &["liaised"]),
            (
                dictgen::InsensitiveStr::Ascii("spefcifieir"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("labaled"), &["labeled"]),
            (dictgen::InsensitiveStr::Ascii("evalutor"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("asssasins"), &["assassins"]),
            (dictgen::InsensitiveStr::Ascii("oly"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("identicially"),
                &["identically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacluates"), &["calculates"]),
            (
                dictgen::InsensitiveStr::Ascii("pertubation"),
                &["perturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("crystalisk"), &["crystals"]),
            (
                dictgen::InsensitiveStr::Ascii("disconneced"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoter"), &["supporter"]),
            (
                dictgen::InsensitiveStr::Ascii("delimeters"),
                &["delimiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("plateu"), &["plateau"]),
            (
                dictgen::InsensitiveStr::Ascii("visualsations"),
                &["visualisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("swastikka"), &["swastika"]),
            (
                dictgen::InsensitiveStr::Ascii("sumulation"),
                &["simulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("disccused"), &["discussed"]),
            (
                dictgen::InsensitiveStr::Ascii("muliplayer"),
                &["multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("autisitc"), &["autistic"]),
            (dictgen::InsensitiveStr::Ascii("strengt"), &["strength"]),
            (
                dictgen::InsensitiveStr::Ascii("produktive"),
                &["productive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instaciate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("abriter"), &["arbiter"]),
            (dictgen::InsensitiveStr::Ascii("moratlity"), &["mortality"]),
            (dictgen::InsensitiveStr::Ascii("asterices"), &["asterisks"]),
            (dictgen::InsensitiveStr::Ascii("deadpoool"), &["deadpool"]),
            (dictgen::InsensitiveStr::Ascii("canews"), &["canoes"]),
            (dictgen::InsensitiveStr::Ascii("hurricance"), &["hurricane"]),
            (dictgen::InsensitiveStr::Ascii("cecksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("thyriod"), &["thyroid"]),
            (
                dictgen::InsensitiveStr::Ascii("catigorically"),
                &["categorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demultiplexor"),
                &["demultiplexer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unserstands"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreprenures"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proector"),
                &["protector", "projector", "proctor"],
            ),
            (dictgen::InsensitiveStr::Ascii("newthon"), &["newton"]),
            (dictgen::InsensitiveStr::Ascii("acivation"), &["activation"]),
            (
                dictgen::InsensitiveStr::Ascii("progagators"),
                &["propagators"],
            ),
            (dictgen::InsensitiveStr::Ascii("direrctor"), &["director"]),
            (dictgen::InsensitiveStr::Ascii("keept"), &["kept"]),
            (
                dictgen::InsensitiveStr::Ascii("optimitations"),
                &["optimizations", "optimisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completito"),
                &["completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("synthesys"), &["synthesis"]),
            (
                dictgen::InsensitiveStr::Ascii("partecipated"),
                &["participated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetoires"),
                &["repertoires"],
            ),
            (dictgen::InsensitiveStr::Ascii("assyemetry"), &["asymmetry"]),
            (dictgen::InsensitiveStr::Ascii("advandcing"), &["advancing"]),
            (dictgen::InsensitiveStr::Ascii("extremly"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("incomfortable"),
                &["uncomfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("momemts"), &["moments"]),
            (dictgen::InsensitiveStr::Ascii("silicoon"), &["silicon"]),
            (dictgen::InsensitiveStr::Ascii("sucide"), &["suicide"]),
            (dictgen::InsensitiveStr::Ascii("forcefuly"), &["forcefully"]),
            (
                dictgen::InsensitiveStr::Ascii("missionaris"),
                &["missionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("runnigng"), &["running"]),
            (
                dictgen::InsensitiveStr::Ascii("underwealming"),
                &["underwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("conanical"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("interveiw"), &["interview"]),
            (
                dictgen::InsensitiveStr::Ascii("unevaluted"),
                &["unevaluated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacaged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("cathegorizes"),
                &["categorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("applyies"), &["applies"]),
            (dictgen::InsensitiveStr::Ascii("operatate"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("divet"), &["divot"]),
            (
                dictgen::InsensitiveStr::Ascii("interesected"),
                &["intersected"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomrer"), &["performer"]),
            (
                dictgen::InsensitiveStr::Ascii("overlauded"),
                &["overloaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encycolpedias"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("scketches"), &["sketches"]),
            (dictgen::InsensitiveStr::Ascii("foget"), &["forget"]),
            (
                dictgen::InsensitiveStr::Ascii("egalitatian"),
                &["egalitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("beccause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("bloodbore"), &["bloodborne"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstatiated"),
                &["reinstantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacific"),
                &["specific", "pacific"],
            ),
            (dictgen::InsensitiveStr::Ascii("agnositc"), &["agnostic"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplicativo"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("tensiones"), &["tensions"]),
            (dictgen::InsensitiveStr::Ascii("stongly"), &["strongly"]),
            (
                dictgen::InsensitiveStr::Ascii("peripeheral"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strigification"),
                &["stringification"],
            ),
            (dictgen::InsensitiveStr::Ascii("sibing"), &["sibling"]),
            (
                dictgen::InsensitiveStr::Ascii("beliv"),
                &["believe", "belief"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doamin"),
                &["domain", "dopamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wearning"),
                &[
                    "warning", "wearing", "earning", "learning", "yearning", "wearying",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("opertaing"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitaion"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nethod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("contiguious"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coonnection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depndances"),
                &["dependances"],
            ),
            (dictgen::InsensitiveStr::Ascii("nullfiy"), &["nullify"]),
            (dictgen::InsensitiveStr::Ascii("personalis"), &["personas"]),
            (dictgen::InsensitiveStr::Ascii("upsrteams"), &["upstreams"]),
            (
                dictgen::InsensitiveStr::Ascii("spontanes"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanlders"), &["handlers"]),
            (dictgen::InsensitiveStr::Ascii("triplin"), &["tripling"]),
            (
                dictgen::InsensitiveStr::Ascii("absorpsion"),
                &["absorption"],
            ),
            (dictgen::InsensitiveStr::Ascii("mobiliy"), &["mobility"]),
            (dictgen::InsensitiveStr::Ascii("creeative"), &["creative"]),
            (
                dictgen::InsensitiveStr::Ascii("trasformative"),
                &["transformative"],
            ),
            (dictgen::InsensitiveStr::Ascii("insdead"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("unregsitering"),
                &["unregistering"],
            ),
            (dictgen::InsensitiveStr::Ascii("revelance"), &["relevance"]),
            (dictgen::InsensitiveStr::Ascii("romoted"), &["remoted"]),
            (dictgen::InsensitiveStr::Ascii("chocolots"), &["chocolates"]),
            (dictgen::InsensitiveStr::Ascii("hellvua"), &["helluva"]),
            (
                dictgen::InsensitiveStr::Ascii("preposters"),
                &["preposterous"],
            ),
            (dictgen::InsensitiveStr::Ascii("seve"), &["save", "sieve"]),
            (dictgen::InsensitiveStr::Ascii("endpoing"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("ingorant"), &["ignorant"]),
            (
                dictgen::InsensitiveStr::Ascii("overritten"),
                &["overwritten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onmiscient"),
                &["omniscient"],
            ),
            (dictgen::InsensitiveStr::Ascii("scalines"), &["scanlines"]),
            (
                dictgen::InsensitiveStr::Ascii("anonymousny"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entiry"),
                &["entity", "entry", "entirely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphetimin"),
                &["amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beind"),
                &["behind", "being"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appeciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("fability"), &["facility"]),
            (dictgen::InsensitiveStr::Ascii("bicylces"), &["bicycles"]),
            (dictgen::InsensitiveStr::Ascii("currentely"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("beyound"), &["beyond"]),
            (
                dictgen::InsensitiveStr::Ascii("cpmpression"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("dosens"), &["dozens"]),
            (dictgen::InsensitiveStr::Ascii("tenacles"), &["tentacles"]),
            (
                dictgen::InsensitiveStr::Ascii("anisotrophically"),
                &["anisotropically"],
            ),
            (dictgen::InsensitiveStr::Ascii("preveiwer"), &["previewer"]),
            (
                dictgen::InsensitiveStr::Ascii("sensationalim"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("armenains"), &["armenians"]),
            (
                dictgen::InsensitiveStr::Ascii("genreating"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("caeses"), &["ceases"]),
            (
                dictgen::InsensitiveStr::Ascii("increade"),
                &["increase", "increased"],
            ),
            (dictgen::InsensitiveStr::Ascii("nevelopes"), &["envelopes"]),
            (dictgen::InsensitiveStr::Ascii("shwon"), &["shown"]),
            (
                dictgen::InsensitiveStr::Ascii("analtyical"),
                &["analytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("releaseing"), &["releasing"]),
            (
                dictgen::InsensitiveStr::Ascii("caclulator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("outrside"), &["outside"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuyes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workfaround"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("mplements"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("instantiatoins"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paricipated"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("alredy"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("diaplays"), &["displays"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaing"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestents"),
                &["protestants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conparative"),
                &["comparative"],
            ),
            (dictgen::InsensitiveStr::Ascii("gigaybte"), &["gigabyte"]),
            (dictgen::InsensitiveStr::Ascii("apparen"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("listenns"), &["listens"]),
            (dictgen::InsensitiveStr::Ascii("pillards"), &["pillars"]),
            (dictgen::InsensitiveStr::Ascii("breapoint"), &["breakpoint"]),
            (dictgen::InsensitiveStr::Ascii("esccape"), &["escape"]),
            (dictgen::InsensitiveStr::Ascii("modifiying"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("uncompatible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructted"),
                &["constructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frivilously"),
                &["frivolously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadverdently"),
                &["inadvertently"],
            ),
            (dictgen::InsensitiveStr::Ascii("recjecting"), &["rejecting"]),
            (dictgen::InsensitiveStr::Ascii("animore"), &["anymore"]),
            (dictgen::InsensitiveStr::Ascii("eanbles"), &["enables"]),
            (dictgen::InsensitiveStr::Ascii("critisice"), &["criticise"]),
            (
                dictgen::InsensitiveStr::Ascii("acceotance"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("reupladed"), &["reuploaded"]),
            (
                dictgen::InsensitiveStr::Ascii("restrcuture"),
                &["restructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spearation"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("afteroon"), &["afternoon"]),
            (dictgen::InsensitiveStr::Ascii("yongest"), &["youngest"]),
            (
                dictgen::InsensitiveStr::Ascii("unreacahable"),
                &["unreachable"],
            ),
            (dictgen::InsensitiveStr::Ascii("basicly"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("satistic"),
                &["statistic", "sadistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("hinderance"), &["hindrance"]),
            (
                dictgen::InsensitiveStr::Ascii("accommidation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assimulate"),
                &["assimilate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comdemnation"),
                &["condemnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abandonnent"),
                &["abandonment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidentail"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("pronounes"), &["pronouns"]),
            (dictgen::InsensitiveStr::Ascii("exaustive"), &["exhaustive"]),
            (
                dictgen::InsensitiveStr::Ascii("funktioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("devels"), &["delves"]),
            (dictgen::InsensitiveStr::Ascii("metrci"), &["metric"]),
            (
                dictgen::InsensitiveStr::Ascii("unreaponsive"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("missils"), &["missiles"]),
            (dictgen::InsensitiveStr::Ascii("concelaer"), &["concealer"]),
            (dictgen::InsensitiveStr::Ascii("nueron"), &["neuron"]),
            (
                dictgen::InsensitiveStr::Ascii("reosurcing"),
                &["resourcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("storeage"), &["storage"]),
            (dictgen::InsensitiveStr::Ascii("ceartype"), &["cleartype"]),
            (dictgen::InsensitiveStr::Ascii("greather"), &["greater"]),
            (
                dictgen::InsensitiveStr::Ascii("asycnhronous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("generelly"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("sharpining"),
                &["sharpening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegments"),
                &["acknowledgments"],
            ),
            (dictgen::InsensitiveStr::Ascii("thwe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("goosegumps"),
                &["goosebumps"],
            ),
            (dictgen::InsensitiveStr::Ascii("aestethic"), &["aesthetic"]),
            (dictgen::InsensitiveStr::Ascii("muhc"), &["much"]),
            (dictgen::InsensitiveStr::Ascii("journied"), &["journeyed"]),
            (dictgen::InsensitiveStr::Ascii("cyllinders"), &["cylinders"]),
            (
                dictgen::InsensitiveStr::Ascii("succsfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("zerod"), &["zeroed"]),
            (
                dictgen::InsensitiveStr::Ascii("comparrator"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("versio"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("compemdium"),
                &["compendium"],
            ),
            (dictgen::InsensitiveStr::Ascii("updatae"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("exeeds"), &["exceeds"]),
            (dictgen::InsensitiveStr::Ascii("shfits"), &["shifts"]),
            (dictgen::InsensitiveStr::Ascii("alignd"), &["aligned"]),
            (dictgen::InsensitiveStr::Ascii("filetered"), &["filtered"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropriatley"),
                &["inappropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaiable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suffiency"),
                &["sufficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("ethicaly"), &["ethically"]),
            (dictgen::InsensitiveStr::Ascii("timesamps"), &["timestamps"]),
            (
                dictgen::InsensitiveStr::Ascii("positivitely"),
                &["positivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("thess"), &["this", "these"]),
            (dictgen::InsensitiveStr::Ascii("recort"), &["record"]),
            (dictgen::InsensitiveStr::Ascii("develp"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("statsit"), &["statist"]),
            (dictgen::InsensitiveStr::Ascii("diables"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("genereting"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentraded"),
                &["concentrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apdater"),
                &["updater", "adapter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colleciton"),
                &["collection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrastically"),
                &["sarcastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("feture"),
                &["feature", "future"],
            ),
            (dictgen::InsensitiveStr::Ascii("toloerant"), &["tolerant"]),
            (dictgen::InsensitiveStr::Ascii("persuing"), &["pursuing"]),
            (
                dictgen::InsensitiveStr::Ascii("temporafily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("sawnson"), &["swanson"]),
            (dictgen::InsensitiveStr::Ascii("immidiate"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("succseeding"),
                &["succeeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("surkarta"), &["surakarta"]),
            (
                dictgen::InsensitiveStr::Ascii("certificaat"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signficantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("orginates"), &["originates"]),
            (dictgen::InsensitiveStr::Ascii("yuo"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("rememebering"),
                &["remembering"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroolls"), &["scrolls"]),
            (
                dictgen::InsensitiveStr::Ascii("encosings"),
                &["enclosings", "encodings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surpporters"),
                &["supporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vartically"),
                &["vertically"],
            ),
            (dictgen::InsensitiveStr::Ascii("distates"), &["distaste"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaites"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cicular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonrous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("codespel"), &["codespell"]),
            (
                dictgen::InsensitiveStr::Ascii("endocing"),
                &["endorsing", "enforcing", "encoding"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchaned"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("elasping"), &["elapsing"]),
            (
                dictgen::InsensitiveStr::Ascii("hertically"),
                &["hectically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferanced"),
                &["dereferenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eqivalents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("striclty"), &["strictly"]),
            (dictgen::InsensitiveStr::Ascii("bevorhand"), &["beforehand"]),
            (dictgen::InsensitiveStr::Ascii("postdam"), &["potsdam"]),
            (dictgen::InsensitiveStr::Ascii("costantly"), &["constantly"]),
            (
                dictgen::InsensitiveStr::Ascii("comfortabel"),
                &["comfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otpimistic"),
                &["optimistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparanetly"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrms"), &["alarms"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrovated"),
                &["aggravated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participent"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("apsirin"), &["aspirin"]),
            (dictgen::InsensitiveStr::Ascii("iselfe"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("apocalype"), &["apocalypse"]),
            (
                dictgen::InsensitiveStr::Ascii("actvations"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("messenging"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("appropaches"),
                &["approaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("railrod"), &["railroad"]),
            (dictgen::InsensitiveStr::Ascii("fasend"), &["fastened"]),
            (dictgen::InsensitiveStr::Ascii("sensative"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("sereously"), &["seriously"]),
            (dictgen::InsensitiveStr::Ascii("xwindows"), &["x"]),
            (
                dictgen::InsensitiveStr::Ascii("capabillity"),
                &["capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinstallng"),
                &["reinstalling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comutations"),
                &["computations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentratie"),
                &["concentrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("thouhgt"), &["thought"]),
            (dictgen::InsensitiveStr::Ascii("vertbrae"), &["vertebrae"]),
            (dictgen::InsensitiveStr::Ascii("externels"), &["externals"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiteid"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("interro"), &["interior"]),
            (dictgen::InsensitiveStr::Ascii("valitdity"), &["validity"]),
            (
                dictgen::InsensitiveStr::Ascii("environmentaly"),
                &["environmentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenciates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("toekn"), &["token"]),
            (dictgen::InsensitiveStr::Ascii("satified"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("asycnhronously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("wht"), &["what"]),
            (
                dictgen::InsensitiveStr::Ascii("congnitively"),
                &["cognitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilzing"),
                &["stabilizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marginilized"),
                &["marginalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("runntimes"), &["runtimes"]),
            (dictgen::InsensitiveStr::Ascii("literture"), &["literature"]),
            (dictgen::InsensitiveStr::Ascii("opimistic"), &["optimistic"]),
            (
                dictgen::InsensitiveStr::Ascii("restorani"),
                &["restoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("grahic"), &["graphic"]),
            (
                dictgen::InsensitiveStr::Ascii("preditermined"),
                &["predetermined"],
            ),
            (dictgen::InsensitiveStr::Ascii("squritle"), &["squirtle"]),
            (dictgen::InsensitiveStr::Ascii("infine"), &["inline"]),
            (dictgen::InsensitiveStr::Ascii("crypro"), &["crypto"]),
            (dictgen::InsensitiveStr::Ascii("temparal"), &["temporal"]),
            (dictgen::InsensitiveStr::Ascii("mport"), &["import"]),
            (dictgen::InsensitiveStr::Ascii("vyre"), &["very"]),
            (dictgen::InsensitiveStr::Ascii("pacjaging"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("reseached"), &["researched"]),
            (
                dictgen::InsensitiveStr::Ascii("programmeurs"),
                &["programmer"],
            ),
            (dictgen::InsensitiveStr::Ascii("turrain"), &["terrain"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worhsipping"),
                &["worshipping"],
            ),
            (dictgen::InsensitiveStr::Ascii("cobvers"), &["covers"]),
            (dictgen::InsensitiveStr::Ascii("valubale"), &["valuable"]),
            (
                dictgen::InsensitiveStr::Ascii("misfortunte"),
                &["misfortune"],
            ),
            (dictgen::InsensitiveStr::Ascii("devaults"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("aquaduct"), &["aqueduct"]),
            (dictgen::InsensitiveStr::Ascii("expliciet"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("obeserver"), &["observer"]),
            (dictgen::InsensitiveStr::Ascii("sycled"), &["cycled"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ansester"), &["ancestor"]),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexplaned"),
                &["unexplained"],
            ),
            (dictgen::InsensitiveStr::Ascii("autonomity"), &["autonomy"]),
            (
                dictgen::InsensitiveStr::Ascii("opertional"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("weas"), &["was"]),
            (dictgen::InsensitiveStr::Ascii("concerend"), &["concerned"]),
            (dictgen::InsensitiveStr::Ascii("giveing"), &["giving"]),
            (
                dictgen::InsensitiveStr::Ascii("acknodledgment"),
                &["acknowledgment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empricically"),
                &["empirically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approproate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("scandales"), &["scandals"]),
            (
                dictgen::InsensitiveStr::Ascii("specialies"),
                &["specializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprensentations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("faught"), &["fought"]),
            (
                dictgen::InsensitiveStr::Ascii("dependencie"),
                &["dependence", "dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercated"),
                &["interacted"],
            ),
            (dictgen::InsensitiveStr::Ascii("quee"), &["queue"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainablity"),
                &["maintainability"],
            ),
            (dictgen::InsensitiveStr::Ascii("examplle"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("authecating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("phyiscal"), &["physical"]),
            (
                dictgen::InsensitiveStr::Ascii("heterosexuella"),
                &["heterosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envolved"),
                &["involved", "evolved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructior"),
                &["instructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encoutered"),
                &["encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("sastified"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("complmenet"),
                &["complement"],
            ),
            (dictgen::InsensitiveStr::Ascii("freezs"), &["freezes"]),
            (dictgen::InsensitiveStr::Ascii("outlinin"), &["outlining"]),
            (
                dictgen::InsensitiveStr::Ascii("extact"),
                &["extract", "exact"],
            ),
            (dictgen::InsensitiveStr::Ascii("oriantate"), &["orientate"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliactions"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custome"),
                &["custom", "customs", "costume", "customer"],
            ),
            (dictgen::InsensitiveStr::Ascii("allk"), &["all", "ally"]),
            (
                dictgen::InsensitiveStr::Ascii("unfirom"),
                &["uniform", "unfirm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imrovement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("justfied"), &["justified"]),
            (dictgen::InsensitiveStr::Ascii("ibraries"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("kickstarer"),
                &["kickstarter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismantlin"),
                &["dismantling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corordination"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accerelating"),
                &["accelerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecssarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("menssaged"), &["messaged"]),
            (dictgen::InsensitiveStr::Ascii("tey"), &["they"]),
            (
                dictgen::InsensitiveStr::Ascii("fingertops"),
                &["fingertips"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitaley"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("meskeeto"), &["mosquito"]),
            (dictgen::InsensitiveStr::Ascii("simpilfy"), &["simplify"]),
            (dictgen::InsensitiveStr::Ascii("ajasence"), &["adjacence"]),
            (
                dictgen::InsensitiveStr::Ascii("starwberries"),
                &["strawberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("namned"), &["named"]),
            (
                dictgen::InsensitiveStr::Ascii("archieving"),
                &["achieving", "archiving"],
            ),
            (dictgen::InsensitiveStr::Ascii("capabilty"), &["capability"]),
            (dictgen::InsensitiveStr::Ascii("prposer"), &["proposer"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilitiy"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeditive"),
                &["competitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orgin"),
                &["origin", "organ"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conclsuions"),
                &["concussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwiong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("ignorre"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("comarisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("lits"), &["list"]),
            (dictgen::InsensitiveStr::Ascii("specifcier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("encoedings"), &["encodings"]),
            (dictgen::InsensitiveStr::Ascii("whicth"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("quanify"),
                &["quantify", "qualify"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internatioanlist"),
                &["internationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("couters"),
                &["counters", "routers", "scouters"],
            ),
            (dictgen::InsensitiveStr::Ascii("psot"), &["post"]),
            (dictgen::InsensitiveStr::Ascii("loacate"), &["locate"]),
            (
                dictgen::InsensitiveStr::Ascii("constructos"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("oilgarchy"), &["oligarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("caracterizing"),
                &["characterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyperbrophy"),
                &["hypertrophy"],
            ),
            (dictgen::InsensitiveStr::Ascii("reomving"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("thanskgiving"),
                &["thanksgiving"],
            ),
            (dictgen::InsensitiveStr::Ascii("setence"), &["sentence"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidatin"),
                &["invalidating", "invalidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("immenantly"), &["eminently"]),
            (dictgen::InsensitiveStr::Ascii("viwers"), &["viewers"]),
            (dictgen::InsensitiveStr::Ascii("staoshi"), &["satoshi"]),
            (dictgen::InsensitiveStr::Ascii("chanage"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("iland"), &["island"]),
            (dictgen::InsensitiveStr::Ascii("surgest"), &["suggest"]),
            (dictgen::InsensitiveStr::Ascii("trength"), &["strength"]),
            (dictgen::InsensitiveStr::Ascii("particales"), &["particles"]),
            (
                dictgen::InsensitiveStr::Ascii("asscoitaes"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("leutenant"), &["lieutenant"]),
            (
                dictgen::InsensitiveStr::Ascii("installion"),
                &["installing", "installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("resutling"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("postpocessing"),
                &["postprocessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ostentibly"),
                &["ostensibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporpriated"),
                &["appropriated"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("diffirences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicte"),
                &["explicate", "explicit"],
            ),
            (dictgen::InsensitiveStr::Ascii("globas"), &["globals"]),
            (
                dictgen::InsensitiveStr::Ascii("lockscreeen"),
                &["lockscreen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassemblies"),
                &["disassemblies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronunciaton"),
                &["pronunciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normailzing"),
                &["normalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("unspected"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("reprodicibly"),
                &["reproducibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corralates"),
                &["correlates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeleivably"),
                &["unbelievably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predefiined"),
                &["predefined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collenders"),
                &["cullenders", "colanders"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntent"), &["content"]),
            (dictgen::InsensitiveStr::Ascii("passowrd"), &["password"]),
            (
                dictgen::InsensitiveStr::Ascii("inhabitats"),
                &["inhabitants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inspectons"),
                &["inspections"],
            ),
            (dictgen::InsensitiveStr::Ascii("inititive"), &["initiative"]),
            (
                dictgen::InsensitiveStr::Ascii("spiritualty"),
                &["spirituality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irreplacable"),
                &["irreplaceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schemetics"),
                &["schematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("girafffe"), &["giraffe"]),
            (dictgen::InsensitiveStr::Ascii("behaviros"), &["behaviors"]),
            (dictgen::InsensitiveStr::Ascii("fromal"), &["formal"]),
            (dictgen::InsensitiveStr::Ascii("ater"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("reintarnation"),
                &["reincarnation"],
            ),
            (dictgen::InsensitiveStr::Ascii("genitalis"), &["genitals"]),
            (
                dictgen::InsensitiveStr::Ascii("beliavable"),
                &["believable"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoise"), &["customise"]),
            (
                dictgen::InsensitiveStr::Ascii("preleminary"),
                &["preliminary"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuasble"), &["reusable"]),
            (
                dictgen::InsensitiveStr::Ascii("initiailizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionray"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutsceen"), &["cutscene"]),
            (dictgen::InsensitiveStr::Ascii("violenty"), &["violently"]),
            (dictgen::InsensitiveStr::Ascii("offcial"), &["official"]),
            (
                dictgen::InsensitiveStr::Ascii("marshmalllows"),
                &["marshmallows"],
            ),
            (dictgen::InsensitiveStr::Ascii("istener"), &["listener"]),
            (
                dictgen::InsensitiveStr::Ascii("individiuals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congugating"),
                &["conjugating"],
            ),
            (dictgen::InsensitiveStr::Ascii("condmened"), &["condemned"]),
            (dictgen::InsensitiveStr::Ascii("exceutive"), &["executive"]),
            (dictgen::InsensitiveStr::Ascii("supsect"), &["suspect"]),
            (
                dictgen::InsensitiveStr::Ascii("suscpetible"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("believr"), &["believer"]),
            (
                dictgen::InsensitiveStr::Ascii("extensiability"),
                &["extensibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("usiability"), &["usability"]),
            (dictgen::InsensitiveStr::Ascii("implmenet"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("americares"), &["americas"]),
            (dictgen::InsensitiveStr::Ascii("liklihood"), &["likelihood"]),
            (dictgen::InsensitiveStr::Ascii("neighber"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("abstacted"), &["abstracted"]),
            (
                dictgen::InsensitiveStr::Ascii("conotation"),
                &["connotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organisaton"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("slashs"), &["slashes"]),
            (
                dictgen::InsensitiveStr::Ascii("occuppations"),
                &["occupations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chached"), &["cached"]),
            (dictgen::InsensitiveStr::Ascii("symtom"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("retriger"), &["retrigger"]),
            (dictgen::InsensitiveStr::Ascii("peacify"), &["pacify"]),
            (dictgen::InsensitiveStr::Ascii("conect"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("quesionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionaty"),
                &["dictionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instnatiations"),
                &["instantiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("heve"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("assimalate"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("blbo"), &["blob"]),
            (
                dictgen::InsensitiveStr::Ascii("specialazation"),
                &["specialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanving"), &["advancing"]),
            (dictgen::InsensitiveStr::Ascii("attemting"), &["attempting"]),
            (dictgen::InsensitiveStr::Ascii("probablly"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("disgraseful"),
                &["disgraceful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registraration"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicatino"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("bbooleans"), &["booleans"]),
            (dictgen::InsensitiveStr::Ascii("parametrs"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("qestioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaetd"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("annonceur"), &["announcer"]),
            (
                dictgen::InsensitiveStr::Ascii("carrer"),
                &["career", "carrier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consecuentes"),
                &["consequences"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgardes"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("siplifies"), &["simplifies"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatabily"),
                &["incompatibility", "incompatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilevants"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckecksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("animatied"), &["animated"]),
            (
                dictgen::InsensitiveStr::Ascii("comporesser"),
                &["compressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("aloow"), &["allow"]),
            (dictgen::InsensitiveStr::Ascii("reclami"), &["reclaim"]),
            (dictgen::InsensitiveStr::Ascii("moudles"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("edior"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("immortas"), &["immortals"]),
            (
                dictgen::InsensitiveStr::Ascii("reactionairy"),
                &["reactionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("proponenet"), &["proponent"]),
            (dictgen::InsensitiveStr::Ascii("bellonging"), &["belonging"]),
            (dictgen::InsensitiveStr::Ascii("skepticals"), &["skeptics"]),
            (
                dictgen::InsensitiveStr::Ascii("protruberance"),
                &["protuberance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasturisation"),
                &["pasteurisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctures"),
                &["correctors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("magnifacent"),
                &["magnificent"],
            ),
            (dictgen::InsensitiveStr::Ascii("queston"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("faclons"), &["falcons"]),
            (
                dictgen::InsensitiveStr::Ascii("commects"),
                &["connects", "comments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forst"),
                &["first", "forced"],
            ),
            (dictgen::InsensitiveStr::Ascii("oridnary"), &["ordinary"]),
            (dictgen::InsensitiveStr::Ascii("morphen"), &["morphine"]),
            (
                dictgen::InsensitiveStr::Ascii("interchangebale"),
                &["interchangeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termianted"),
                &["terminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inplicitly"),
                &["implicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attachemnts"),
                &["attachments"],
            ),
            (dictgen::InsensitiveStr::Ascii("portoflio"), &["portfolio"]),
            (
                dictgen::InsensitiveStr::Ascii("anonimusally"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanetel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesfull"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("destinatons"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("obviosly"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasist"),
                &["pharmacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("actal"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("yera"), &["year"]),
            (
                dictgen::InsensitiveStr::Ascii("canocicalize"),
                &["canonicalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emply"),
                &["employ", "empty", "imply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationslism"),
                &["sensationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributave"),
                &["distributive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envonment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("latets"), &["latest"]),
            (dictgen::InsensitiveStr::Ascii("robutness"), &["robustness"]),
            (dictgen::InsensitiveStr::Ascii("loacals"), &["locals"]),
            (dictgen::InsensitiveStr::Ascii("recusin"), &["recusing"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastrinates"),
                &["procrastinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("permiums"), &["premiums"]),
            (
                dictgen::InsensitiveStr::Ascii("supplimentary"),
                &["supplementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("resotring"), &["restoring"]),
            (
                dictgen::InsensitiveStr::Ascii("usign"),
                &["using", "unsign"],
            ),
            (dictgen::InsensitiveStr::Ascii("authoize"), &["authorize"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbohoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallizing"),
                &["parallelizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dosclosures"),
                &["disclosures"],
            ),
            (dictgen::InsensitiveStr::Ascii("stradegy"), &["strategy"]),
            (
                dictgen::InsensitiveStr::Ascii("archeaology"),
                &["archeology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribbution"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("tombsones"), &["tombstones"]),
            (
                dictgen::InsensitiveStr::Ascii("conviently"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combintaion"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoeers"), &["followers"]),
            (
                dictgen::InsensitiveStr::Ascii("coorperation"),
                &["cooperation", "corporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("glitchly"), &["glitchy"]),
            (dictgen::InsensitiveStr::Ascii("youu"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotiotions"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("zeor"), &["zero"]),
            (
                dictgen::InsensitiveStr::Ascii("stornfront"),
                &["stormfront"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniuplating"),
                &["manipulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconditon"),
                &["precondition"],
            ),
            (dictgen::InsensitiveStr::Ascii("tagert"), &["target"]),
            (dictgen::InsensitiveStr::Ascii("brisben"), &["brisbane"]),
            (
                dictgen::InsensitiveStr::Ascii("incomapatible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undertsands"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polymoprhism"),
                &["polymorphism"],
            ),
            (dictgen::InsensitiveStr::Ascii("priciples"), &["principles"]),
            (dictgen::InsensitiveStr::Ascii("marineros"), &["mariners"]),
            (dictgen::InsensitiveStr::Ascii("cuestions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("unatteneded"),
                &["unattended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copehnagen"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("serialsed"), &["serialised"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusists"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckoud"), &["cloud"]),
            (dictgen::InsensitiveStr::Ascii("workins"), &["workings"]),
            (dictgen::InsensitiveStr::Ascii("cliper"), &["clipper"]),
            (dictgen::InsensitiveStr::Ascii("selectons"), &["selections"]),
            (dictgen::InsensitiveStr::Ascii("opereate"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("independ"), &["independent"]),
            (dictgen::InsensitiveStr::Ascii("manaager"), &["manager"]),
            (
                dictgen::InsensitiveStr::Ascii("surpresss"),
                &["suppress", "suppresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("inavlid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("itialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("medows"), &["meadows"]),
            (dictgen::InsensitiveStr::Ascii("laboreres"), &["laborers"]),
            (
                dictgen::InsensitiveStr::Ascii("neigbhorhood"),
                &["neighborhoods", "neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vawdvillean"),
                &["vaudevillian"],
            ),
            (dictgen::InsensitiveStr::Ascii("refelect"), &["reflect"]),
            (dictgen::InsensitiveStr::Ascii("appetitite"), &["appetite"]),
            (dictgen::InsensitiveStr::Ascii("difusive"), &["diffusive"]),
            (
                dictgen::InsensitiveStr::Ascii("deployement"),
                &["deployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("containerr"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("samller"), &["smaller"]),
            (
                dictgen::InsensitiveStr::Ascii("stadnardization"),
                &["standardization"],
            ),
            (dictgen::InsensitiveStr::Ascii("naybor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("identtification"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("incluides"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("relyes"),
                &["relies", "realize", "realise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playgerise"),
                &["plagiarise"],
            ),
            (dictgen::InsensitiveStr::Ascii("godds"), &["goods", "gods"]),
            (
                dictgen::InsensitiveStr::Ascii("authenification"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hesitstion"),
                &["hesitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("repond"), &["respond"]),
            (dictgen::InsensitiveStr::Ascii("agant"), &["agent"]),
            (
                dictgen::InsensitiveStr::Ascii("angshusness"),
                &["anxiousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("lisetner"), &["listener"]),
            (dictgen::InsensitiveStr::Ascii("arbutrary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("reproduciblity"),
                &["reproducibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediaately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aproximations"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("importamnt"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("reoccurrin"),
                &["reoccurring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorises"),
                &["prioritises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefcifieid"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("exteded"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("stablizied"),
                &["stabilized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bracnhed"), &["branched"]),
            (dictgen::InsensitiveStr::Ascii("laziliy"), &["lazily"]),
            (dictgen::InsensitiveStr::Ascii("outpsoken"), &["outspoken"]),
            (
                dictgen::InsensitiveStr::Ascii("pariticipations"),
                &["participations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferrences"),
                &["preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncluding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("overidable"),
                &["overridable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cassawory"), &["cassowary"]),
            (dictgen::InsensitiveStr::Ascii("palnning"), &["planning"]),
            (dictgen::InsensitiveStr::Ascii("academci"), &["academic"]),
            (dictgen::InsensitiveStr::Ascii("widgest"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("changinge"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("tipically"), &["typically"]),
            (
                dictgen::InsensitiveStr::Ascii("vulneabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachterised"),
                &["characterised"],
            ),
            (dictgen::InsensitiveStr::Ascii("funniliy"), &["funnily"]),
            (dictgen::InsensitiveStr::Ascii("waitng"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("foramtted"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("protectings"),
                &["protections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("incides"), &["indices"]),
            (dictgen::InsensitiveStr::Ascii("cyclyc"), &["cyclic"]),
            (
                dictgen::InsensitiveStr::Ascii("resursively"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("configutre"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("infomativeness"),
                &["informativeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deseralises"),
                &["deserialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("patriotas"), &["patriots"]),
            (
                dictgen::InsensitiveStr::Ascii("stornghold"),
                &["stronghold"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestent"),
                &["protestant"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebulds"), &["rebuilds"]),
            (
                dictgen::InsensitiveStr::Ascii("environtment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("senteneces"), &["sentences"]),
            (
                dictgen::InsensitiveStr::Ascii("noral"),
                &["normal", "moral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversimplifaction"),
                &["oversimplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("shadder"), &["shader"]),
            (
                dictgen::InsensitiveStr::Ascii("counterside"),
                &["countryside"],
            ),
            (dictgen::InsensitiveStr::Ascii("lauched"), &["launched"]),
            (
                dictgen::InsensitiveStr::Ascii("necessarrily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("contestors"), &["contests"]),
            (dictgen::InsensitiveStr::Ascii("percisely"), &["precisely"]),
            (
                dictgen::InsensitiveStr::Ascii("conrtibuting"),
                &["contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bachleors"), &["bachelors"]),
            (dictgen::InsensitiveStr::Ascii("semaphors"), &["semaphores"]),
            (dictgen::InsensitiveStr::Ascii("sprecial"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("ressurect"), &["resurrect"]),
            (dictgen::InsensitiveStr::Ascii("frined"), &["friend"]),
            (dictgen::InsensitiveStr::Ascii("midified"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("againn"), &["again"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridinating"),
                &["coordinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fancyness"), &["fanciness"]),
            (dictgen::InsensitiveStr::Ascii("scrabing"), &["scraping"]),
            (dictgen::InsensitiveStr::Ascii("releaases"), &["releases"]),
            (dictgen::InsensitiveStr::Ascii("opertaor"), &["operator"]),
            (
                dictgen::InsensitiveStr::Ascii("hallucinaitons"),
                &["hallucination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accompagniment"),
                &["accompaniment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("warrenties"),
                &["warranties"],
            ),
            (dictgen::InsensitiveStr::Ascii("ffor"), &["for"]),
            (dictgen::InsensitiveStr::Ascii("stimulai"), &["stimuli"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminted"),
                &["contaminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commemmorate"),
                &["commemorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recrutiers"),
                &["recruiters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizarem"),
                &["organizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("pubishing"), &["publishing"]),
            (dictgen::InsensitiveStr::Ascii("victoires"), &["victories"]),
            (dictgen::InsensitiveStr::Ascii("asend"), &["ascend"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("woftware"), &["software"]),
            (
                dictgen::InsensitiveStr::Ascii("cosntructors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("verstaile"), &["versatile"]),
            (dictgen::InsensitiveStr::Ascii("leght"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("interaktivly"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleanders"),
                &["cleaners", "calendars", "cleansers"],
            ),
            (dictgen::InsensitiveStr::Ascii("issusing"), &["issuing"]),
            (
                dictgen::InsensitiveStr::Ascii("verifiactions"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("birthdsy"), &["birthdays"]),
            (
                dictgen::InsensitiveStr::Ascii("disruptivo"),
                &["disruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("versiomed"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("dowgrades"), &["downgrades"]),
            (
                dictgen::InsensitiveStr::Ascii("psycholoog"),
                &["psychology"],
            ),
            (dictgen::InsensitiveStr::Ascii("specift"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("uncetainties"),
                &["uncertainties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defintiive"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("remorted"), &["reported"]),
            (
                dictgen::InsensitiveStr::Ascii("propertiess"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("colomns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("sentires"), &["sentries"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmetetric"),
                &["asymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("thtread"), &["thread"]),
            (
                dictgen::InsensitiveStr::Ascii("dissobediant"),
                &["disobedient"],
            ),
            (dictgen::InsensitiveStr::Ascii("slaptoon"), &["splatoon"]),
            (
                dictgen::InsensitiveStr::Ascii("sensationaism"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("coersces"), &["coerces"]),
            (dictgen::InsensitiveStr::Ascii("africanas"), &["africans"]),
            (
                dictgen::InsensitiveStr::Ascii("boradcasting"),
                &["broadcasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangaeble"),
                &["rearrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dafault"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("kimerric"), &["chimeric"]),
            (dictgen::InsensitiveStr::Ascii("poninter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("visualizating"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("bziped"), &["bzipped"]),
            (
                dictgen::InsensitiveStr::Ascii("initiializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccesibles"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conspiraces"),
                &["conspiracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("inititals"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("execpt"), &["except"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluatin"),
                &["evaluating", "evaluation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbhoring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dffer"),
                &["differ", "defer", "offer"],
            ),
            (dictgen::InsensitiveStr::Ascii("movepseed"), &["movespeed"]),
            (dictgen::InsensitiveStr::Ascii("alignmets"), &["alignments"]),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbcribing"),
                &["subscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adming"),
                &["admin", "arming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modiification"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("noices"), &["noises"]),
            (dictgen::InsensitiveStr::Ascii("passwrods"), &["passwords"]),
            (
                dictgen::InsensitiveStr::Ascii("belligirent"),
                &["belligerent"],
            ),
            (dictgen::InsensitiveStr::Ascii("precurors"), &["precursor"]),
            (dictgen::InsensitiveStr::Ascii("caribles"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("assined"), &["assigned"]),
            (dictgen::InsensitiveStr::Ascii("esportes"), &["esports"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("qoutes"), &["quotes"]),
            (
                dictgen::InsensitiveStr::Ascii("comparioss"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("coloum"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("perticipated"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("extedners"), &["extenders"]),
            (
                dictgen::InsensitiveStr::Ascii("backwars"),
                &["backward", "backwards"],
            ),
            (dictgen::InsensitiveStr::Ascii("trigering"), &["triggering"]),
            (dictgen::InsensitiveStr::Ascii("behavrior"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("intallationpath"),
                &["installationpath"],
            ),
            (dictgen::InsensitiveStr::Ascii("gneral"), &["general"]),
            (
                dictgen::InsensitiveStr::Ascii("converaging"),
                &["converging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subredddits"),
                &["subreddits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offet"),
                &["offset", "offer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soverignity"),
                &["sovereignty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appoitment"),
                &["appointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("dwarvens"), &["dwarves"]),
            (
                dictgen::InsensitiveStr::Ascii("initiasing"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("orpaned"), &["orphaned"]),
            (dictgen::InsensitiveStr::Ascii("sycn"), &["sync"]),
            (dictgen::InsensitiveStr::Ascii("noraml"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("cipboard"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("incrediably"),
                &["incredibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuqesting"),
                &["requesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocatin"),
                &["allocating", "allocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremmentd"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("determins"), &["determines"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchron"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simpliifcations"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("dektops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("stoopidity"), &["stupidity"]),
            (dictgen::InsensitiveStr::Ascii("suppses"), &["supposes"]),
            (
                dictgen::InsensitiveStr::Ascii("intergates"),
                &["integrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differentl"),
                &["differential", "differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traumatisch"),
                &["traumatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retquiressl"),
                &["requiressl"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurisve"), &["recursive"]),
            (
                dictgen::InsensitiveStr::Ascii("indecations"),
                &["indications"],
            ),
            (dictgen::InsensitiveStr::Ascii("jorney"), &["journey"]),
            (dictgen::InsensitiveStr::Ascii("oakery"), &["ochery"]),
            (dictgen::InsensitiveStr::Ascii("fike"), &["file"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsdent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfacting"),
                &["satisfaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genenerally"),
                &["generally"],
            ),
            (dictgen::InsensitiveStr::Ascii("deflaut"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("unregistred"),
                &["unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consquences"),
                &["consequences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proble"),
                &["probe", "probably", "problem", "probable", "provable"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrapers"), &["wrappers"]),
            (dictgen::InsensitiveStr::Ascii("smoot"), &["smooth"]),
            (
                dictgen::InsensitiveStr::Ascii("troubeshooted"),
                &["troubleshooted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanetory"),
                &["explanatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthongally"),
                &["orthogonally"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefic"), &["prefix"]),
            (
                dictgen::InsensitiveStr::Ascii("negosiators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inialiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidentcy"),
                &["presidency"],
            ),
            (dictgen::InsensitiveStr::Ascii("uniforn"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("derefenreced"),
                &["dereferenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visibiltiy"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrapp"), &["wrap"]),
            (dictgen::InsensitiveStr::Ascii("damanges"), &["damages"]),
            (
                dictgen::InsensitiveStr::Ascii("persistente"),
                &["persistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("silouet"), &["silhouette"]),
            (dictgen::InsensitiveStr::Ascii("underuns"), &["underruns"]),
            (dictgen::InsensitiveStr::Ascii("accptable"), &["acceptable"]),
            (
                dictgen::InsensitiveStr::Ascii("libertarianisme"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("telporting"),
                &["teleporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deacitvated"),
                &["deactivated"],
            ),
            (dictgen::InsensitiveStr::Ascii("meachinism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("phylosophers"),
                &["philosophers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintainance"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("subequent"), &["subsequent"]),
            (
                dictgen::InsensitiveStr::Ascii("countertraps"),
                &["counterparts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordingate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("merrors"), &["mirrors"]),
            (
                dictgen::InsensitiveStr::Ascii("disuss"),
                &["discuss", "disuse"],
            ),
            (dictgen::InsensitiveStr::Ascii("generell"), &["general"]),
            (
                dictgen::InsensitiveStr::Ascii("similtaneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrbutes"),
                &["distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("unrnaked"), &["unranked"]),
            (
                dictgen::InsensitiveStr::Ascii("definishion"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminisentky"),
                &["reminiscently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concission"),
                &["concession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udnerlying"),
                &["underlying"],
            ),
            (dictgen::InsensitiveStr::Ascii("viee"), &["view"]),
            (dictgen::InsensitiveStr::Ascii("cematery"), &["cemetery"]),
            (dictgen::InsensitiveStr::Ascii("assignts"), &["assigns"]),
            (dictgen::InsensitiveStr::Ascii("grammers"), &["grammars"]),
            (dictgen::InsensitiveStr::Ascii("sentinals"), &["sentinels"]),
            (dictgen::InsensitiveStr::Ascii("comtpon"), &["compton"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenatation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nstalled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("cacl"), &["calc"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmaceudical"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("recommande"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrarion"),
                &["arbitration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decremeanting"),
                &["decrementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innitialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registrato"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represened"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("produkt"), &["product"]),
            (dictgen::InsensitiveStr::Ascii("mltibyte"), &["multibyte"]),
            (
                dictgen::InsensitiveStr::Ascii("rationalizaton"),
                &["rationalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechnanical"),
                &["mechanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interetsed"),
                &["interested"],
            ),
            (dictgen::InsensitiveStr::Ascii("minites"), &["minutes"]),
            (
                dictgen::InsensitiveStr::Ascii("experimintations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ncludes"),
                &["includes", "concludes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tigthening"),
                &["tightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multimational"),
                &["multinational"],
            ),
            (dictgen::InsensitiveStr::Ascii("resevering"), &["reserving"]),
            (
                dictgen::InsensitiveStr::Ascii("atomtic"),
                &["atomic", "automatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("esperate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("abstact"), &["abstract"]),
            (dictgen::InsensitiveStr::Ascii("sequared"), &["squared"]),
            (dictgen::InsensitiveStr::Ascii("spiritd"), &["spirited"]),
            (dictgen::InsensitiveStr::Ascii("chennel"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("conincide"), &["coincide"]),
            (dictgen::InsensitiveStr::Ascii("resease"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("initializationg"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("potpurri"), &["potpourri"]),
            (dictgen::InsensitiveStr::Ascii("mehod"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("anlayses"), &["analyses"]),
            (dictgen::InsensitiveStr::Ascii("priorit"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("benifitting"),
                &["benefitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("deoes"), &["does"]),
            (
                dictgen::InsensitiveStr::Ascii("embarressing"),
                &["embarrassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declareation"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommonding"),
                &["recommending"],
            ),
            (dictgen::InsensitiveStr::Ascii("decideable"), &["decidable"]),
            (dictgen::InsensitiveStr::Ascii("scientits"), &["scientist"]),
            (
                dictgen::InsensitiveStr::Ascii("supermakter"),
                &["supermarkets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatins"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existend"),
                &["existed", "existent"],
            ),
            (dictgen::InsensitiveStr::Ascii("assasins"), &["assassins"]),
            (dictgen::InsensitiveStr::Ascii("fitered"), &["filtered"]),
            (dictgen::InsensitiveStr::Ascii("accounant"), &["accountant"]),
            (dictgen::InsensitiveStr::Ascii("reudces"), &["reduces"]),
            (
                dictgen::InsensitiveStr::Ascii("microvaves"),
                &["microwaves"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoprted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("refreing"), &["referring"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasociated"),
                &["disassociated"],
            ),
            (dictgen::InsensitiveStr::Ascii("miligram"), &["milligram"]),
            (
                dictgen::InsensitiveStr::Ascii("heighest"),
                &["heights", "highest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassmble"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scientisst"),
                &["scientists", "scientist"],
            ),
            (dictgen::InsensitiveStr::Ascii("implicity"), &["implicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("expresison"),
                &["expressions", "expression"],
            ),
            (dictgen::InsensitiveStr::Ascii("treatsie"), &["treaties"]),
            (dictgen::InsensitiveStr::Ascii("confuzze"), &["confuse"]),
            (
                dictgen::InsensitiveStr::Ascii("canonalization"),
                &["canonicalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wastefull"),
                &["wasteful", "wastefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("detectaron"), &["detector"]),
            (dictgen::InsensitiveStr::Ascii("icense"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("summmed"), &["summed"]),
            (dictgen::InsensitiveStr::Ascii("protestos"), &["protests"]),
            (
                dictgen::InsensitiveStr::Ascii("reserached"),
                &["researched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rhinosaruss"),
                &["rhinoceroses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggergated"),
                &["aggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pointeur"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("documemting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("intercats"), &["interacts"]),
            (
                dictgen::InsensitiveStr::Ascii("classications"),
                &["classifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mearly"),
                &["merely", "nearly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementet"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("rrase"), &["erase"]),
            (dictgen::InsensitiveStr::Ascii("yourr"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("vritualenv"),
                &["virtualenv"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entent"),
                &["intent", "extent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainfests"), &["manifests"]),
            (dictgen::InsensitiveStr::Ascii("brusk"), &["brusque"]),
            (dictgen::InsensitiveStr::Ascii("kinfs"), &["kinds"]),
            (dictgen::InsensitiveStr::Ascii("porvided"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("fullfilled"), &["fulfilled"]),
            (
                dictgen::InsensitiveStr::Ascii("runnging"),
                &["running", "rummaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragmants"), &["fragments"]),
            (
                dictgen::InsensitiveStr::Ascii("aprroximations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insesitiveness"),
                &["insensitiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("astethics"), &["aesthetics"]),
            (
                dictgen::InsensitiveStr::Ascii("transluent"),
                &["translucent"],
            ),
            (dictgen::InsensitiveStr::Ascii("tipe"), &["type", "tip"]),
            (dictgen::InsensitiveStr::Ascii("abviously"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("combatibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("accmulate"), &["accumulate"]),
            (dictgen::InsensitiveStr::Ascii("middlewar"), &["middleware"]),
            (dictgen::InsensitiveStr::Ascii("schduler"), &["scheduler"]),
            (
                dictgen::InsensitiveStr::Ascii("bittersweeet"),
                &["bittersweet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditiond"),
                &["conditioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("offloded"), &["offloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("sabatoshing"),
                &["sabotaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("classees"), &["classes"]),
            (dictgen::InsensitiveStr::Ascii("rendring"), &["rendering"]),
            (dictgen::InsensitiveStr::Ascii("butttons"), &["buttons"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiationg"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("skools"), &["schools"]),
            (dictgen::InsensitiveStr::Ascii("conerter"), &["converter"]),
            (dictgen::InsensitiveStr::Ascii("mindcrak"), &["mindcrack"]),
            (
                dictgen::InsensitiveStr::Ascii("conjugaged"),
                &["conjugated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronised"),
                &["synchronised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kidnergarten"),
                &["kindergarten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnarability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaninings"), &["meanings"]),
            (dictgen::InsensitiveStr::Ascii("chekcmate"), &["checkmate"]),
            (dictgen::InsensitiveStr::Ascii("termniate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("otherwhile"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("sketpic"), &["skeptic"]),
            (
                dictgen::InsensitiveStr::Ascii("calclators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhod"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("approuval"), &["approval"]),
            (dictgen::InsensitiveStr::Ascii("wiyh"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("porfiler"), &["profiler"]),
            (dictgen::InsensitiveStr::Ascii("listent"), &["listen"]),
            (dictgen::InsensitiveStr::Ascii("distanca"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("expirencing"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorns"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("fiferox"), &["firefox"]),
            (dictgen::InsensitiveStr::Ascii("assimes"), &["assumes"]),
            (dictgen::InsensitiveStr::Ascii("simluate"), &["simulate"]),
            (
                dictgen::InsensitiveStr::Ascii("intance"),
                &["instance", "intense"],
            ),
            (dictgen::InsensitiveStr::Ascii("kresh"), &["crèche"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitilaized"),
                &["reinitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("purtains"), &["pertains"]),
            (
                dictgen::InsensitiveStr::Ascii("formatiing"),
                &["formatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiancial"),
                &["experiential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("villin"),
                &["villain", "villi", "villein"],
            ),
            (dictgen::InsensitiveStr::Ascii("instalded"), &["installed"]),
            (
                dictgen::InsensitiveStr::Ascii("indonistan"),
                &["indonesian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psichological"),
                &["psychological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucsessfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("expalin"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("wierdness"), &["weirdness"]),
            (
                dictgen::InsensitiveStr::Ascii("wordaround"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvenince"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definetively"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currespondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("promptes"), &["prompts"]),
            (
                dictgen::InsensitiveStr::Ascii("midfiedler"),
                &["midfielder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominaters"),
                &["denominators", "denominates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrigerar"),
                &["refrigerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("outfiled"), &["outfield"]),
            (
                dictgen::InsensitiveStr::Ascii("sist"),
                &["cyst", "sift", "sits"],
            ),
            (dictgen::InsensitiveStr::Ascii("dilema"), &["dilemma"]),
            (dictgen::InsensitiveStr::Ascii("excute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("lenghthen"), &["lengthen"]),
            (dictgen::InsensitiveStr::Ascii("auromated"), &["automated"]),
            (dictgen::InsensitiveStr::Ascii("eaither"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoitator"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centrafuge"),
                &["centrifuge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumulted"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("diretion"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("desition"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("variences"), &["variances"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenions"),
                &["concatenations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transportor"),
                &["transporter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("courcing"),
                &["coercing", "sourcing", "coursing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performamce"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchronusly"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrution"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("succecors"), &["successors"]),
            (
                dictgen::InsensitiveStr::Ascii("unintialization"),
                &["uninitialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cafays"), &["cafes"]),
            (dictgen::InsensitiveStr::Ascii("qouting"), &["quoting"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetrifies"), &["certifies"]),
            (dictgen::InsensitiveStr::Ascii("latitute"), &["latitude"]),
            (dictgen::InsensitiveStr::Ascii("symobilic"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("ppelines"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("inititlised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("stonger"), &["stronger"]),
            (
                dictgen::InsensitiveStr::Ascii("depdenency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breadtfeeding"),
                &["breastfeeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("resmebles"), &["resembles"]),
            (
                dictgen::InsensitiveStr::Ascii("degenratet"),
                &["degenerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluater"),
                &["evaluator", "evaluated", "evaluates", "evaluate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproduciton"),
                &["reproduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelizing"),
                &["parallelizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terimnally"),
                &["terminally"],
            ),
            (dictgen::InsensitiveStr::Ascii("activisit"), &["activist"]),
            (dictgen::InsensitiveStr::Ascii("turtels"), &["turtles"]),
            (
                dictgen::InsensitiveStr::Ascii("scandimania"),
                &["scandinavia"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepe"), &["shape"]),
            (
                dictgen::InsensitiveStr::Ascii("combanations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decorelate"),
                &["decorrelate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kubenretes"),
                &["kubernetes"],
            ),
            (dictgen::InsensitiveStr::Ascii("restoiring"), &["restoring"]),
            (dictgen::InsensitiveStr::Ascii("suppports"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("plaestine"), &["palestine"]),
            (
                dictgen::InsensitiveStr::Ascii("consevible"),
                &["conceivable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemetened"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investagator"),
                &["investigator"],
            ),
            (dictgen::InsensitiveStr::Ascii("colecting"), &["collecting"]),
            (dictgen::InsensitiveStr::Ascii("tempels"), &["temples"]),
            (dictgen::InsensitiveStr::Ascii("variablen"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("forthcomin"),
                &["forthcoming"],
            ),
            (dictgen::InsensitiveStr::Ascii("qiest"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("cringeworhty"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("naybors"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("intiialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbiddin"), &["forbidding"]),
            (dictgen::InsensitiveStr::Ascii("liberalis"), &["liberals"]),
            (dictgen::InsensitiveStr::Ascii("whythout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("reluctanty"),
                &["reluctantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overshaddowed"),
                &["overshadowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forewarded"),
                &["forwarded", "forewarned"],
            ),
            (dictgen::InsensitiveStr::Ascii("msis"), &["miss"]),
            (
                dictgen::InsensitiveStr::Ascii("distructor"),
                &["destructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("standrds"), &["standards"]),
            (
                dictgen::InsensitiveStr::Ascii("obsessivley"),
                &["obsessive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionailies"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("copytight"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("altrusim"), &["altruism"]),
            (
                dictgen::InsensitiveStr::Ascii("aquisition"),
                &["acquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("languagees"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilyzes"),
                &["cannibalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outnumbed"),
                &["outnumbered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("freez"),
                &["frees", "freeze"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconstititional"),
                &["unconstitutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("airial"),
                &["aerial", "arial"],
            ),
            (dictgen::InsensitiveStr::Ascii("deverloped"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretad"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("librabrie"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("schedulier"), &["scheduler"]),
            (dictgen::InsensitiveStr::Ascii("faills"), &["fails"]),
            (dictgen::InsensitiveStr::Ascii("gladiatr"), &["gladiator"]),
            (
                dictgen::InsensitiveStr::Ascii("examplee"),
                &["example", "examples"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanetising"),
                &["sanitising"],
            ),
            (dictgen::InsensitiveStr::Ascii("meeds"), &["needs"]),
            (dictgen::InsensitiveStr::Ascii("dicrectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplates"),
                &["deduplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("obtainabe"), &["obtainable"]),
            (
                dictgen::InsensitiveStr::Ascii("sanctionned"),
                &["sanctioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("fleed"), &["fled", "freed"]),
            (
                dictgen::InsensitiveStr::Ascii("progresive"),
                &["progressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsmitted"),
                &["transmitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("enignes"), &["engines"]),
            (dictgen::InsensitiveStr::Ascii("opiods"), &["opioids"]),
            (
                dictgen::InsensitiveStr::Ascii("performanve"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("redution"), &["reduction"]),
            (
                dictgen::InsensitiveStr::Ascii("transitors"),
                &["transistor", "transistors"],
            ),
            (dictgen::InsensitiveStr::Ascii("filanames"), &["filenames"]),
            (
                dictgen::InsensitiveStr::Ascii("aaccessibility"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloese"), &["close"]),
            (dictgen::InsensitiveStr::Ascii("construde"), &["construed"]),
            (
                dictgen::InsensitiveStr::Ascii("enhacement"),
                &["enhancement"],
            ),
            (dictgen::InsensitiveStr::Ascii("compltes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("integating"),
                &["integrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concerntrating"),
                &["concentrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperetive"),
                &["imperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succecfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcular"), &["calculator"]),
            (dictgen::InsensitiveStr::Ascii("ahmond"), &["almond"]),
            (
                dictgen::InsensitiveStr::Ascii("tenatiously"),
                &["tenaciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("scaricity"), &["scarcity"]),
            (dictgen::InsensitiveStr::Ascii("cotroll"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("architectually"),
                &["architecturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represssion"),
                &["repression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supscription"),
                &["subscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("receet"), &["receipt"]),
            (
                dictgen::InsensitiveStr::Ascii("deffinition"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapionted"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("fushed"), &["flushed"]),
            (dictgen::InsensitiveStr::Ascii("appers"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiatin"),
                &["renegotiating", "renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dumbells"), &["dumbbells"]),
            (dictgen::InsensitiveStr::Ascii("modyfying"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("unrepetent"),
                &["unrepentant"],
            ),
            (dictgen::InsensitiveStr::Ascii("suprising"), &["surprising"]),
            (dictgen::InsensitiveStr::Ascii("devovle"), &["devolve"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificng"),
                &["sacrificing"],
            ),
            (dictgen::InsensitiveStr::Ascii("propgates"), &["propagates"]),
            (dictgen::InsensitiveStr::Ascii("direftory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("tenacitiy"), &["tenacity"]),
            (
                dictgen::InsensitiveStr::Ascii("simplifiying"),
                &["simplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuirement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("tradion"), &["tradition"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangtements"),
                &["rearrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotionation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsitution"),
                &["substitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreves"), &["retrieves"]),
            (dictgen::InsensitiveStr::Ascii("eement"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("compativle"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("invertes"), &["inverse"]),
            (
                dictgen::InsensitiveStr::Ascii("documentationn"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transtator"),
                &["translator"],
            ),
            (dictgen::InsensitiveStr::Ascii("congigure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("confilcted"),
                &["conflicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspongd"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("visibleable"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("vitrole"), &["vitriol"]),
            (
                dictgen::InsensitiveStr::Ascii("distingusih"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("inluding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("develepmont"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flanethrower"),
                &["flamethrower"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contstruction"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("regisrers"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("vaoided"),
                &["avoided", "voided"],
            ),
            (dictgen::InsensitiveStr::Ascii("allopone"), &["allophone"]),
            (dictgen::InsensitiveStr::Ascii("toggels"), &["toggles"]),
            (
                dictgen::InsensitiveStr::Ascii("freindship"),
                &["friendship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthimic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dirct"),
                &["direct", "dict", "dirt"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroolled"), &["scrolled"]),
            (dictgen::InsensitiveStr::Ascii("soddering"), &["soldering"]),
            (
                dictgen::InsensitiveStr::Ascii("tessleations"),
                &["tessellations"],
            ),
            (dictgen::InsensitiveStr::Ascii("combaine"), &["combine"]),
            (
                dictgen::InsensitiveStr::Ascii("operater"),
                &["operator", "operated", "operates", "operate"],
            ),
            (dictgen::InsensitiveStr::Ascii("possibble"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("guitarit"), &["guitarist"]),
            (
                dictgen::InsensitiveStr::Ascii("schizophernic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("siginifies"), &["signifies"]),
            (dictgen::InsensitiveStr::Ascii("whitspace"), &["whitespace"]),
            (dictgen::InsensitiveStr::Ascii("somethiing"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("manageed"), &["managed"]),
            (
                dictgen::InsensitiveStr::Ascii("retireving"),
                &["retrieving"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimal"), &["minimal"]),
            (dictgen::InsensitiveStr::Ascii("clossed"), &["closed"]),
            (
                dictgen::InsensitiveStr::Ascii("obligitory"),
                &["obligatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("dowgrade"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("tournemant"),
                &["tournament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vietnamesea"),
                &["vietnamese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archicture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refectoring"),
                &["refactoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sarcasticlly"),
                &["sarcastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("bebore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("cche"), &["cache"]),
            (
                dictgen::InsensitiveStr::Ascii("troglodite"),
                &["troglodyte"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistansen"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("excetion"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("persciuos"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("endiness"), &["endianness"]),
            (dictgen::InsensitiveStr::Ascii("threatment"), &["treatment"]),
            (
                dictgen::InsensitiveStr::Ascii("exceedinly"),
                &["exceedingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indoktrination"),
                &["indoctrination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signularly"),
                &["singularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transacitonal"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("sortig"), &["sorting"]),
            (dictgen::InsensitiveStr::Ascii("partians"), &["partisan"]),
            (
                dictgen::InsensitiveStr::Ascii("readablity"),
                &["readability"],
            ),
            (dictgen::InsensitiveStr::Ascii("airscape"), &["airspace"]),
            (dictgen::InsensitiveStr::Ascii("subtrings"), &["substrings"]),
            (
                dictgen::InsensitiveStr::Ascii("manslaugher"),
                &["manslaughter"],
            ),
            (dictgen::InsensitiveStr::Ascii("ballanced"), &["balanced"]),
            (
                dictgen::InsensitiveStr::Ascii("correllation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grammarical"),
                &["grammatical"],
            ),
            (dictgen::InsensitiveStr::Ascii("coumpound"), &["compound"]),
            (dictgen::InsensitiveStr::Ascii("ciph"), &["cipher", "chip"]),
            (dictgen::InsensitiveStr::Ascii("temrinate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("indxing"), &["indexing"]),
            (
                dictgen::InsensitiveStr::Ascii("renegosiation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alsot"), &["also"]),
            (dictgen::InsensitiveStr::Ascii("enabked"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowladgements"),
                &["acknowledgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbhorhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutlinationals"),
                &["multinational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("docuemnts"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("disagreemet"),
                &["disagreements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mircophone"),
                &["microphone"],
            ),
            (dictgen::InsensitiveStr::Ascii("expeense"), &["expense"]),
            (dictgen::InsensitiveStr::Ascii("forin"), &["foreign"]),
            (
                dictgen::InsensitiveStr::Ascii("ubiquituously"),
                &["ubiquitously"],
            ),
            (dictgen::InsensitiveStr::Ascii("propetrys"), &["properties"]),
            (
                dictgen::InsensitiveStr::Ascii("preferrence"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfoemant"),
                &["performant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("involintary"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("degraderad"), &["degraded"]),
            (
                dictgen::InsensitiveStr::Ascii("correted"),
                &["corrected", "correlated"],
            ),
            (dictgen::InsensitiveStr::Ascii("simpliest"), &["simplest"]),
            (
                dictgen::InsensitiveStr::Ascii("comunicated"),
                &["communicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasforms"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogenously"),
                &["homogeneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highighting"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inflitrator"),
                &["infiltrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostspot"), &["hotspot"]),
            (dictgen::InsensitiveStr::Ascii("sanple"), &["sample"]),
            (dictgen::InsensitiveStr::Ascii("streatched"), &["stretched"]),
            (dictgen::InsensitiveStr::Ascii("suscribe"), &["subscribe"]),
            (dictgen::InsensitiveStr::Ascii("drotmund"), &["dortmund"]),
            (dictgen::InsensitiveStr::Ascii("lucifear"), &["lucifer"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcuation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechines"), &["machines"]),
            (
                dictgen::InsensitiveStr::Ascii("repedability"),
                &["repeatability"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartiladge"), &["cartilage"]),
            (dictgen::InsensitiveStr::Ascii("mailny"), &["mainly"]),
            (
                dictgen::InsensitiveStr::Ascii("competators"),
                &["competitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexepcted"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("ascneding"), &["ascending"]),
            (dictgen::InsensitiveStr::Ascii("rebouns"), &["rebounds"]),
            (dictgen::InsensitiveStr::Ascii("enehanced"), &["enhanced"]),
            (dictgen::InsensitiveStr::Ascii("decideds"), &["decides"]),
            (dictgen::InsensitiveStr::Ascii("halfiax"), &["halifax"]),
            (
                dictgen::InsensitiveStr::Ascii("rispective"),
                &["respective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depercated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqire"), &["require"]),
            (
                dictgen::InsensitiveStr::Ascii("initiatiating"),
                &["initiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("depretiate"),
                &["depreciate", "deprecate"],
            ),
            (dictgen::InsensitiveStr::Ascii("brightn"), &["brighten"]),
            (
                dictgen::InsensitiveStr::Ascii("prominately"),
                &["prominently", "predominately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strat"),
                &["start", "strata"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("futhroc"),
                &["futhark", "futhorc"],
            ),
            (dictgen::InsensitiveStr::Ascii("argsuments"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("contianment"),
                &["containment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coproccessors"),
                &["coprocessors"],
            ),
            (dictgen::InsensitiveStr::Ascii("borgwasy"), &["bourgeoisie"]),
            (
                dictgen::InsensitiveStr::Ascii("unsupport"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrbiutes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("gigabye"), &["gigabyte"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondente"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciriteria"), &["criteria"]),
            (dictgen::InsensitiveStr::Ascii("suggeting"), &["suggesting"]),
            (dictgen::InsensitiveStr::Ascii("scaner"), &["scanner"]),
            (
                dictgen::InsensitiveStr::Ascii("accessbility"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusotm"), &["custom"]),
            (
                dictgen::InsensitiveStr::Ascii("rquest"),
                &["request", "quest"],
            ),
            (dictgen::InsensitiveStr::Ascii("ertoneous"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("defaulds"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("bounaries"), &["boundaries"]),
            (
                dictgen::InsensitiveStr::Ascii("rotatons"),
                &["rotations", "rotators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetification"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("romoting"), &["remoting"]),
            (
                dictgen::InsensitiveStr::Ascii("readdrsses"),
                &["readdresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrgument"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("hirsohima"), &["hiroshima"]),
            (
                dictgen::InsensitiveStr::Ascii("implemeneting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychopaat"),
                &["psychopath"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eigenevector"),
                &["eigenvector"],
            ),
            (dictgen::InsensitiveStr::Ascii("arcticles"), &["articles"]),
            (
                dictgen::InsensitiveStr::Ascii("instatiating"),
                &["instantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("instrcuts"), &["instructs"]),
            (dictgen::InsensitiveStr::Ascii("etamology"), &["etymology"]),
            (
                dictgen::InsensitiveStr::Ascii("definayely"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucesive"), &["successive"]),
            (
                dictgen::InsensitiveStr::Ascii("youseff"),
                &["yousef", "yourself"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contexta"),
                &["contextual", "context"],
            ),
            (dictgen::InsensitiveStr::Ascii("dangerus"), &["dangers"]),
            (dictgen::InsensitiveStr::Ascii("mnenomic"), &["mnemonic"]),
            (dictgen::InsensitiveStr::Ascii("noticabe"), &["noticeable"]),
            (
                dictgen::InsensitiveStr::Ascii("functiioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("frament"), &["fragment"]),
            (dictgen::InsensitiveStr::Ascii("nomber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("gouvener"), &["governor"]),
            (dictgen::InsensitiveStr::Ascii("copayble"), &["copyable"]),
            (
                dictgen::InsensitiveStr::Ascii("accuastion"),
                &["accusation"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrrypted"), &["encrypted"]),
            (dictgen::InsensitiveStr::Ascii("annoted"), &["annotated"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamtive"),
                &["performative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sutatus"), &["status"]),
            (dictgen::InsensitiveStr::Ascii("ancedote"), &["anecdote"]),
            (dictgen::InsensitiveStr::Ascii("lecteurs"), &["lectures"]),
            (dictgen::InsensitiveStr::Ascii("appplet"), &["applet"]),
            (dictgen::InsensitiveStr::Ascii("nightime"), &["nighttime"]),
            (dictgen::InsensitiveStr::Ascii("colapse"), &["collapse"]),
            (dictgen::InsensitiveStr::Ascii("apporoves"), &["approves"]),
            (
                dictgen::InsensitiveStr::Ascii("thermodynaics"),
                &["thermodynamics"],
            ),
            (dictgen::InsensitiveStr::Ascii("miscrosoft"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("possiblilty"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equavalence"),
                &["equivalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("erally"),
                &["orally", "really"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procudes"),
                &["procures", "produces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fases"),
                &["fazes", "phases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synomymous"),
                &["synonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifiy"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("bitcase"), &["bitcast"]),
            (dictgen::InsensitiveStr::Ascii("asii"), &["ascii"]),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiate"),
                &["negotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irraditates"),
                &["irradiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("monglos"), &["mongols"]),
            (
                dictgen::InsensitiveStr::Ascii("theorically"),
                &["theoretically"],
            ),
            (dictgen::InsensitiveStr::Ascii("enocdes"), &["encodes"]),
            (
                dictgen::InsensitiveStr::Ascii("dpendencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("loacates"), &["locates"]),
            (dictgen::InsensitiveStr::Ascii("froday"), &["friday"]),
            (dictgen::InsensitiveStr::Ascii("prayry"), &["prairie"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellising"),
                &["parallelising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtracter"),
                &["abstracter"],
            ),
            (dictgen::InsensitiveStr::Ascii("cernomies"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("appriceates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("duing"), &["during"]),
            (
                dictgen::InsensitiveStr::Ascii("indicatee"),
                &["indicates", "indicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropriatly"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("retieve"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("warnned"), &["warned"]),
            (dictgen::InsensitiveStr::Ascii("aboute"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("egards"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("abliity"), &["ability"]),
            (dictgen::InsensitiveStr::Ascii("plarizes"), &["polarizes"]),
            (
                dictgen::InsensitiveStr::Ascii("configurrations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciall"),
                &["special", "specially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infedility"),
                &["infidelity"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploaeds"), &["uploads"]),
            (
                dictgen::InsensitiveStr::Ascii("disaproval"),
                &["disapproval"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performinc"),
                &["performing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstallimg"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("allcoated"), &["allocated"]),
            (dictgen::InsensitiveStr::Ascii("plannig"), &["planning"]),
            (
                dictgen::InsensitiveStr::Ascii("parahprase"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singaled"),
                &["signaled", "singled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singularily"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("secod"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("ahed"), &["ahead"]),
            (dictgen::InsensitiveStr::Ascii("neccessary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("pesudo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("captialise"),
                &["capitalise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("produceds"),
                &["produces", "produced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplimenting"),
                &["reimplementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("begiinning"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("somce"), &["some"]),
            (dictgen::InsensitiveStr::Ascii("emmission"), &["emission"]),
            (
                dictgen::InsensitiveStr::Ascii("intelligient"),
                &["intelligent"],
            ),
            (dictgen::InsensitiveStr::Ascii("impulisve"), &["impulsive"]),
            (
                dictgen::InsensitiveStr::Ascii("sensititivy"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("scnearios"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhopath"),
                &["psychopath"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparntly"),
                &["transparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterpreneur"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("mawsoleum"), &["mausoleum"]),
            (dictgen::InsensitiveStr::Ascii("negotable"), &["negotiable"]),
            (
                dictgen::InsensitiveStr::Ascii("metabolizm"),
                &["metabolism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sinagog"),
                &["synagog", "synagogue"],
            ),
            (dictgen::InsensitiveStr::Ascii("defishant"), &["deficient"]),
            (
                dictgen::InsensitiveStr::Ascii("calculuation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("streamear"), &["streamer"]),
            (
                dictgen::InsensitiveStr::Ascii("outperfome"),
                &["outperform"],
            ),
            (dictgen::InsensitiveStr::Ascii("moddelled"), &["modelled"]),
            (dictgen::InsensitiveStr::Ascii("ypu"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegoziation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("forse"), &["force"]),
            (
                dictgen::InsensitiveStr::Ascii("contitnent"),
                &["contingent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reassurring"),
                &["reassuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectins"),
                &["connects", "connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("establises"),
                &["establishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("resposnes"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("interspersin"),
                &["interspersing", "interspersion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmisssion"),
                &["transmissions", "transmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradicitng"),
                &["contradicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersteller"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinaters"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fursest"),
                &["first", "furthest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccurently"),
                &["recurrently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palcements"),
                &["placements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effeciency"),
                &["efficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("presssed"), &["pressed"]),
            (
                dictgen::InsensitiveStr::Ascii("distribtion"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aproximately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatyeps"), &["datatypes"]),
            (
                dictgen::InsensitiveStr::Ascii("parititions"),
                &["partitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astonashing"),
                &["astonishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mezmorizing"),
                &["mesmerizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corss"),
                &["cross", "course"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciry"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("excludde"), &["exclude"]),
            (dictgen::InsensitiveStr::Ascii("shoftware"), &["software"]),
            (
                dictgen::InsensitiveStr::Ascii("recconecting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("theorits"), &["theorist"]),
            (
                dictgen::InsensitiveStr::Ascii("confidentiel"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchainged"), &["unchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("complicacion"),
                &["complication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fertilzier"),
                &["fertilizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("intervall"), &["interval"]),
            (
                dictgen::InsensitiveStr::Ascii("defamating"),
                &["defamation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessairily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("emproving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("depoly"), &["deploy"]),
            (
                dictgen::InsensitiveStr::Ascii("miscommuniction"),
                &["miscommunication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significently"),
                &["significantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furutistic"),
                &["futuristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sbaolutely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("regnerate"), &["regenerate"]),
            (
                dictgen::InsensitiveStr::Ascii("ampehtamine"),
                &["amphetamine"],
            ),
            (dictgen::InsensitiveStr::Ascii("exponants"), &["exponents"]),
            (dictgen::InsensitiveStr::Ascii("nessicarry"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("siloweting"),
                &["silhouetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconsicously"),
                &["unconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrobution"),
                &["retribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jackonsville"),
                &["jacksonville"],
            ),
            (dictgen::InsensitiveStr::Ascii("donain"), &["domain"]),
            (dictgen::InsensitiveStr::Ascii("misandrony"), &["misandry"]),
            (dictgen::InsensitiveStr::Ascii("charatcer"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("interfacces"),
                &["interfaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitaor"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acitvision"),
                &["activision"],
            ),
            (dictgen::InsensitiveStr::Ascii("stlye"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("startlisteneing"),
                &["startlistening"],
            ),
            (dictgen::InsensitiveStr::Ascii("writters"), &["writers"]),
            (dictgen::InsensitiveStr::Ascii("onsluaght"), &["onslaught"]),
            (dictgen::InsensitiveStr::Ascii("unwraping"), &["unwrapping"]),
            (dictgen::InsensitiveStr::Ascii("glichted"), &["glitched"]),
            (dictgen::InsensitiveStr::Ascii("subsytems"), &["subsystems"]),
            (
                dictgen::InsensitiveStr::Ascii("offensivelly"),
                &["offensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandobx"), &["sandbox"]),
            (
                dictgen::InsensitiveStr::Ascii("configurtions"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresovled"),
                &["unresolved"],
            ),
            (dictgen::InsensitiveStr::Ascii("monsteres"), &["monsters"]),
            (dictgen::InsensitiveStr::Ascii("sourrounds"), &["surrounds"]),
            (
                dictgen::InsensitiveStr::Ascii("consant"),
                &["constant", "consent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasmitters"),
                &["transmitters"],
            ),
            (dictgen::InsensitiveStr::Ascii("compabile"), &["compatible"]),
            (dictgen::InsensitiveStr::Ascii("polpulate"), &["populate"]),
            (dictgen::InsensitiveStr::Ascii("upercased"), &["uppercased"]),
            (dictgen::InsensitiveStr::Ascii("zucheeni"), &["zucchini"]),
            (dictgen::InsensitiveStr::Ascii("plguin"), &["plugin"]),
            (dictgen::InsensitiveStr::Ascii("leagelity"), &["legality"]),
            (
                dictgen::InsensitiveStr::Ascii("constrcutor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgaded"), &["upgraded"]),
            (
                dictgen::InsensitiveStr::Ascii("deliverying"),
                &["delivering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteres"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("incredibe"), &["incredible"]),
            (
                dictgen::InsensitiveStr::Ascii("modificiation"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prejudgudice"),
                &["prejudice"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomorroe"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("beglium"), &["belgium"]),
            (
                dictgen::InsensitiveStr::Ascii("dismantleing"),
                &["dismantling"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccriping"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("authenication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("akumulating"),
                &["accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("funvtion"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("specificateions"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dermitologist"),
                &["dermatologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperialims"),
                &["imperialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("trnsfer"), &["transfer"]),
            (
                dictgen::InsensitiveStr::Ascii("psychopatic"),
                &["psychopathic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("campaiging"),
                &["campaigning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequeny"),
                &["frequency", "frequently", "frequent"],
            ),
            (dictgen::InsensitiveStr::Ascii("threshol"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("acendencey"),
                &["ascendency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interepolation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("progamers"), &["programers"]),
            (
                dictgen::InsensitiveStr::Ascii("reorded"),
                &["reorder", "recorded", "reordered", "reworded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrainted"),
                &["restrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completelty"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("expersive"), &["expensive"]),
            (dictgen::InsensitiveStr::Ascii("sciprts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("enviromnents"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagiariasm"),
                &["plagiarism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viewpionts"),
                &["viewpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sists"),
                &["cysts", "sifts", "sits"],
            ),
            (dictgen::InsensitiveStr::Ascii("southren"), &["southern"]),
            (
                dictgen::InsensitiveStr::Ascii("signabl"),
                &["signable", "signal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterizin"),
                &["characterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalulation"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("miagical"), &["magical"]),
            (dictgen::InsensitiveStr::Ascii("atquired"), &["acquired"]),
            (dictgen::InsensitiveStr::Ascii("forhead"), &["forehead"]),
            (
                dictgen::InsensitiveStr::Ascii("characteers"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iteartions"),
                &["iterations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intializating"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasres"),
                &["parses", "pastes", "passes", "pares"],
            ),
            (dictgen::InsensitiveStr::Ascii("desplayed"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("confirmaed"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("chioces"), &["choices"]),
            (dictgen::InsensitiveStr::Ascii("comile"), &["compile"]),
            (
                dictgen::InsensitiveStr::Ascii("specialices"),
                &["specializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transivity"),
                &["transitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("toword"), &["toward"]),
            (dictgen::InsensitiveStr::Ascii("lighness"), &["lightness"]),
            (
                dictgen::InsensitiveStr::Ascii("disccussions"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patrens"),
                &["patron", "patterns", "patrons"],
            ),
            (dictgen::InsensitiveStr::Ascii("nothig"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("promissing"), &["promising"]),
            (
                dictgen::InsensitiveStr::Ascii("reslutions"),
                &["resolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("numberr"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("decreas"), &["decrease"]),
            (
                dictgen::InsensitiveStr::Ascii("paramterising"),
                &["parameterising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intterupted"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomaptibelity"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acadimically"),
                &["academically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mardown"), &["markdown"]),
            (dictgen::InsensitiveStr::Ascii("hge"), &["he"]),
            (dictgen::InsensitiveStr::Ascii("oprate"), &["operate"]),
            (
                dictgen::InsensitiveStr::Ascii("irresponsble"),
                &["irresponsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("replcaing"), &["replacing"]),
            (dictgen::InsensitiveStr::Ascii("lisens"), &["listens"]),
            (
                dictgen::InsensitiveStr::Ascii("televesion"),
                &["television"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacth"), &["patch"]),
            (dictgen::InsensitiveStr::Ascii("scrutiney"), &["scrutiny"]),
            (
                dictgen::InsensitiveStr::Ascii("transmitsions"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuned"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("insterrupting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipulted"),
                &["manipulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("whome"), &["whom"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterized"),
                &["parameterized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfuncitonal"),
                &["dysfunctional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cricitally"),
                &["critically"],
            ),
            (dictgen::InsensitiveStr::Ascii("caculate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("correltor"), &["correlator"]),
            (
                dictgen::InsensitiveStr::Ascii("corosponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("retring"), &["retrying"]),
            (dictgen::InsensitiveStr::Ascii("nuetral"), &["neutral"]),
            (
                dictgen::InsensitiveStr::Ascii("advandce"),
                &["advance", "advanced"],
            ),
            (dictgen::InsensitiveStr::Ascii("montaeg"), &["montage"]),
            (
                dictgen::InsensitiveStr::Ascii("xeonphobia"),
                &["xenophobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualizaitons"),
                &["visualizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiatin"),
                &["initiation", "initiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustratie"),
                &["illustrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("aborigene"), &["aborigine"]),
            (dictgen::InsensitiveStr::Ascii("kenrels"), &["kernels"]),
            (dictgen::InsensitiveStr::Ascii("herat"), &["heart"]),
            (dictgen::InsensitiveStr::Ascii("symphonity"), &["symphony"]),
            (
                dictgen::InsensitiveStr::Ascii("inheritablility"),
                &["inheritability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automatico"),
                &["automation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placehoulder"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdedder"), &["embedder"]),
            (
                dictgen::InsensitiveStr::Ascii("unristricted"),
                &["unrestricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("leyering"), &["layering"]),
            (
                dictgen::InsensitiveStr::Ascii("proprogates"),
                &["propagates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laveled"),
                &["leveled", "labeled"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepairs"), &["prepares"]),
            (dictgen::InsensitiveStr::Ascii("noen"), &["none", "neon"]),
            (
                dictgen::InsensitiveStr::Ascii("recompuuting"),
                &["recomputing"],
            ),
            (dictgen::InsensitiveStr::Ascii("motorollas"), &["motorola"]),
            (dictgen::InsensitiveStr::Ascii("archivel"), &["archival"]),
            (dictgen::InsensitiveStr::Ascii("conbtrols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("cloudfare"), &["cloudflare"]),
            (
                dictgen::InsensitiveStr::Ascii("expetimentelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unimplmented"),
                &["unimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("postphoned"), &["postponed"]),
            (dictgen::InsensitiveStr::Ascii("excludin"), &["excluding"]),
            (dictgen::InsensitiveStr::Ascii("lauchers"), &["launchers"]),
            (dictgen::InsensitiveStr::Ascii("vasall"), &["vassal"]),
            (dictgen::InsensitiveStr::Ascii("jamaina"), &["jamaican"]),
            (dictgen::InsensitiveStr::Ascii("upddates"), &["updates"]),
            (
                dictgen::InsensitiveStr::Ascii("esteblished"),
                &["established"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indistinquishable"),
                &["indistinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahtletes"), &["athletes"]),
            (
                dictgen::InsensitiveStr::Ascii("parafanalia"),
                &["paraphernalia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preints"),
                &["prints", "preprints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschiatrist"),
                &["psychiatrist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decorativo"),
                &["decoration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forgotting"),
                &["forgetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberarion"),
                &["liberation"],
            ),
            (dictgen::InsensitiveStr::Ascii("respoders"), &["responders"]),
            (dictgen::InsensitiveStr::Ascii("illigel"), &["illegal"]),
            (dictgen::InsensitiveStr::Ascii("agonstic"), &["agnostic"]),
            (dictgen::InsensitiveStr::Ascii("xepect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("proposin"), &["proposing"]),
            (dictgen::InsensitiveStr::Ascii("normale"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("stimulans"), &["stimulants"]),
            (dictgen::InsensitiveStr::Ascii("sur"), &["sure"]),
            (dictgen::InsensitiveStr::Ascii("alwaty"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("caucasin"), &["caucasian"]),
            (dictgen::InsensitiveStr::Ascii("maight"), &["might"]),
            (
                dictgen::InsensitiveStr::Ascii("interplotate"),
                &["interpolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("starups"), &["startups"]),
            (
                dictgen::InsensitiveStr::Ascii("incredable"),
                &["incredible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjustement"),
                &["adjustment"],
            ),
            (dictgen::InsensitiveStr::Ascii("compiator"), &["compilator"]),
            (
                dictgen::InsensitiveStr::Ascii("uncesessarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispairty"), &["disparity"]),
            (dictgen::InsensitiveStr::Ascii("exspenses"), &["expenses"]),
            (dictgen::InsensitiveStr::Ascii("chechked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("orientaion"),
                &["orientation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condiotions"),
                &["conditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("attahed"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("illinos"), &["illinois"]),
            (
                dictgen::InsensitiveStr::Ascii("vewed"),
                &["viewed", "vowed", "vexed"],
            ),
            (dictgen::InsensitiveStr::Ascii("deterant"), &["deterrent"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaiations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomicly"), &["atomically"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessful"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("roataion"), &["rotation"]),
            (dictgen::InsensitiveStr::Ascii("difficuly"), &["difficulty"]),
            (dictgen::InsensitiveStr::Ascii("invloving"), &["involving"]),
            (
                dictgen::InsensitiveStr::Ascii("courthourse"),
                &["courthouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momomentos"),
                &["mementos", "moments"],
            ),
            (dictgen::InsensitiveStr::Ascii("maintream"), &["mainstream"]),
            (dictgen::InsensitiveStr::Ascii("airbourne"), &["airborne"]),
            (
                dictgen::InsensitiveStr::Ascii("intimadate"),
                &["intimidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoinds"), &["avoids"]),
            (dictgen::InsensitiveStr::Ascii("clusterm"), &["cluster"]),
            (
                dictgen::InsensitiveStr::Ascii("ramdomness"),
                &["randomness"],
            ),
            (dictgen::InsensitiveStr::Ascii("timtout"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("regresssive"),
                &["regressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("globbal"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowledget"),
                &["acknowledgment"],
            ),
            (dictgen::InsensitiveStr::Ascii("ranage"), &["range"]),
            (dictgen::InsensitiveStr::Ascii("anddroids"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("generatins"),
                &["generations", "generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("championshiop"),
                &["championships"],
            ),
            (dictgen::InsensitiveStr::Ascii("detele"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("asociated"), &["associated"]),
            (dictgen::InsensitiveStr::Ascii("arbitual"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("updting"), &["updating"]),
            (dictgen::InsensitiveStr::Ascii("stipling"), &["stippling"]),
            (dictgen::InsensitiveStr::Ascii("extedn"), &["extend"]),
            (dictgen::InsensitiveStr::Ascii("cointain"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("publshers"), &["publishers"]),
            (
                dictgen::InsensitiveStr::Ascii("dramaticlly"),
                &["dramatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarantene"),
                &["quarantine"],
            ),
            (dictgen::InsensitiveStr::Ascii("reinfoces"), &["reinforces"]),
            (
                dictgen::InsensitiveStr::Ascii("vitualized"),
                &["virtualized", "visualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reflektion"),
                &["reflection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconecctions"),
                &["disconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecassary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrievate"),
                &["abbreviate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uppstreaming"),
                &["upstreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("compposite"), &["composite"]),
            (dictgen::InsensitiveStr::Ascii("virtiol"), &["vitriol"]),
            (dictgen::InsensitiveStr::Ascii("cukoo"), &["cuckoo"]),
            (dictgen::InsensitiveStr::Ascii("strenght"), &["strength"]),
            (dictgen::InsensitiveStr::Ascii("tequests"), &["requests"]),
            (
                dictgen::InsensitiveStr::Ascii("otations"),
                &["notations", "rotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bussy"), &["busy"]),
            (dictgen::InsensitiveStr::Ascii("retreved"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("equialents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcy"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("temlates"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("lattest"), &["latest"]),
            (dictgen::InsensitiveStr::Ascii("fliper"), &["flipper"]),
            (
                dictgen::InsensitiveStr::Ascii("commadnline"),
                &["commandline"],
            ),
            (dictgen::InsensitiveStr::Ascii("calbirate"), &["calibrate"]),
            (dictgen::InsensitiveStr::Ascii("pevents"), &["prevents"]),
            (dictgen::InsensitiveStr::Ascii("matadata"), &["metadata"]),
            (
                dictgen::InsensitiveStr::Ascii("validatior"),
                &["validator", "validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("postitives"), &["positives"]),
            (
                dictgen::InsensitiveStr::Ascii("prerogitive"),
                &["prerogative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identtation"),
                &["indentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("lieved"), &["lived"]),
            (
                dictgen::InsensitiveStr::Ascii("camoflaguing"),
                &["camouflaging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tanslating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("unexpetct"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("dependente"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("regreses"), &["regress"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterise"),
                &["characterise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("licese"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("langueage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("confiming"), &["confirming"]),
            (
                dictgen::InsensitiveStr::Ascii("necessesarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recgonizes"),
                &["recognizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilesation"),
                &["civilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("improof"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("sintakts"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("proponants"),
                &["proponents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificaton"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("privelige"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("winnpieg"), &["winnipeg"]),
            (
                dictgen::InsensitiveStr::Ascii("linguisticas"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("presener"), &["presenter"]),
            (
                dictgen::InsensitiveStr::Ascii("discritizations"),
                &["discretizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("newmonic"), &["pneumonic"]),
            (dictgen::InsensitiveStr::Ascii("goldike"), &["godlike"]),
            (dictgen::InsensitiveStr::Ascii("documnets"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("apocalipse"),
                &["apocalypse"],
            ),
            (dictgen::InsensitiveStr::Ascii("ashpalt"), &["asphalt"]),
            (
                dictgen::InsensitiveStr::Ascii("archievers"),
                &["achievers", "archivers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtenticator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compredded"),
                &["compressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("practic"), &["practice"]),
            (dictgen::InsensitiveStr::Ascii("espisodes"), &["episodes"]),
            (dictgen::InsensitiveStr::Ascii("contigent"), &["contingent"]),
            (
                dictgen::InsensitiveStr::Ascii("individuales"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("otion"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("usful"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("additioanal"),
                &["additional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contracing"),
                &["contracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgrateful"),
                &["disgraceful"],
            ),
            (dictgen::InsensitiveStr::Ascii("repilcas"), &["replicas"]),
            (
                dictgen::InsensitiveStr::Ascii("procrascinating"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("unheathly"), &["unhealthy"]),
            (dictgen::InsensitiveStr::Ascii("espcially"), &["especially"]),
            (
                dictgen::InsensitiveStr::Ascii("photographics"),
                &["photographs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiantances"),
                &["acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("betwern"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("priviously"),
                &["previously"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasnmit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("uncorelated"),
                &["uncorrelated"],
            ),
            (dictgen::InsensitiveStr::Ascii("progarmed"), &["programmed"]),
            (dictgen::InsensitiveStr::Ascii("oboleted"), &["obsoleted"]),
            (
                dictgen::InsensitiveStr::Ascii("repititively"),
                &["repetitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carthagian"),
                &["carthaginian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procastrination"),
                &["procrastination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanatin"),
                &["explanations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addintional"),
                &["additional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuses"),
                &["excuses", "executes"],
            ),
            (dictgen::InsensitiveStr::Ascii("decalare"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("chizzel"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("autocoomit"),
                &["autocommit"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirorring"), &["mirroring"]),
            (dictgen::InsensitiveStr::Ascii("otyher"), &["other"]),
            (dictgen::InsensitiveStr::Ascii("adpater"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("arouind"), &["around"]),
            (
                dictgen::InsensitiveStr::Ascii("descriotors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("italianas"), &["italians"]),
            (
                dictgen::InsensitiveStr::Ascii("stadnardisation"),
                &["standardisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("proocesses"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("bloodboner"),
                &["bloodborne"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiator"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rebuplicans"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("elelment"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("convigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiailized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("revsions"), &["revisions"]),
            (dictgen::InsensitiveStr::Ascii("modiifiers"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("adjancencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("categroy"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("protals"), &["portals"]),
            (
                dictgen::InsensitiveStr::Ascii("unreigstering"),
                &["unregistering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcdorys"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("wednesdsy"), &["wednesdays"]),
            (dictgen::InsensitiveStr::Ascii("testsing"), &["testing"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccuraccy"),
                &["inaccuracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interecting"),
                &["interacting", "intersecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("spolied"), &["spoiled"]),
            (dictgen::InsensitiveStr::Ascii("specidic"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreggations"),
                &["aggregations"],
            ),
            (dictgen::InsensitiveStr::Ascii("carnavor"), &["carnivore"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprehensibe"),
                &["incomprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("copyies"), &["copies"]),
            (
                dictgen::InsensitiveStr::Ascii("oversubcribing"),
                &["oversubscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequentally"),
                &["sequentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disctinctions"),
                &["distinctions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possesssion"),
                &["possessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("surrouds"), &["surrounds"]),
            (dictgen::InsensitiveStr::Ascii("acessory"), &["accessory"]),
            (dictgen::InsensitiveStr::Ascii("ducment"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("premption"), &["preemption"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpetedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("overlflows"), &["overflows"]),
            (dictgen::InsensitiveStr::Ascii("satifying"), &["satisfying"]),
            (
                dictgen::InsensitiveStr::Ascii("acknownledgements"),
                &["acknowledgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mitigaiton"),
                &["mitigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimach"), &["mismatch"]),
            (
                dictgen::InsensitiveStr::Ascii("apprecited"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("logisticas"), &["logistics"]),
            (
                dictgen::InsensitiveStr::Ascii("utillisation"),
                &["utilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("regsters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("applainces"),
                &["appliances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characater"),
                &["caricature", "character"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simuate"),
                &["simulate", "situate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neede"),
                &["needed", "need", "needle"],
            ),
            (dictgen::InsensitiveStr::Ascii("assingnee"), &["assignee"]),
            (
                dictgen::InsensitiveStr::Ascii("annoucements"),
                &["announcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geenrators"),
                &["generators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positivy"),
                &["positivity", "positive", "positively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segretation"),
                &["segregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneting"), &["connecting"]),
            (dictgen::InsensitiveStr::Ascii("abudcted"), &["abducted"]),
            (
                dictgen::InsensitiveStr::Ascii("versiond"),
                &["versions", "versioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("univeristies"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("pauloads"), &["payloads"]),
            (dictgen::InsensitiveStr::Ascii("overands"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("functoned"), &["functioned"]),
            (dictgen::InsensitiveStr::Ascii("origiginal"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("unpriviledged"),
                &["unprivileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoliute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("bucker"), &["bucket"]),
            (
                dictgen::InsensitiveStr::Ascii("affilitations"),
                &["affiliations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indigenious"),
                &["indigenous"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppen"), &["open"]),
            (
                dictgen::InsensitiveStr::Ascii("communitites"),
                &["communities"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirriring"), &["mirroring"]),
            (dictgen::InsensitiveStr::Ascii("peole"), &["people"]),
            (dictgen::InsensitiveStr::Ascii("saturdsy"), &["saturdays"]),
            (
                dictgen::InsensitiveStr::Ascii("unterstands"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("javasciript"),
                &["javascript"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconised"), &["recognised"]),
            (
                dictgen::InsensitiveStr::Ascii("obersvation"),
                &["observation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forseeable"),
                &["foreseeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnosticm"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("organzies"), &["organizes"]),
            (dictgen::InsensitiveStr::Ascii("attirtion"), &["attrition"]),
            (
                dictgen::InsensitiveStr::Ascii("depreceates"),
                &["deprecates", "depreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extractnow"),
                &["extraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfetch"), &["prefetch"]),
            (
                dictgen::InsensitiveStr::Ascii("probabalities"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("atrtribute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("dfference"), &["difference"]),
            (
                dictgen::InsensitiveStr::Ascii("configueres"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("recordss"), &["records"]),
            (
                dictgen::InsensitiveStr::Ascii("heidelburg"),
                &["heidelberg"],
            ),
            (dictgen::InsensitiveStr::Ascii("samwiched"), &["sandwiched"]),
            (dictgen::InsensitiveStr::Ascii("labael"), &["label"]),
            (
                dictgen::InsensitiveStr::Ascii("omishients"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("referecne"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("revolvs"), &["revolves"]),
            (dictgen::InsensitiveStr::Ascii("conseat"), &["conceit"]),
            (dictgen::InsensitiveStr::Ascii("finailze"), &["finalize"]),
            (
                dictgen::InsensitiveStr::Ascii("documentaion"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unappretiative"),
                &["unappreciative"],
            ),
            (dictgen::InsensitiveStr::Ascii("wathever"), &["whatever"]),
            (
                dictgen::InsensitiveStr::Ascii("ansalization"),
                &["nasalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doublerift"),
                &["doublelift"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdot"), &["anecdote"]),
            (dictgen::InsensitiveStr::Ascii("integates"), &["integrates"]),
            (dictgen::InsensitiveStr::Ascii("ths"), &["the", "this"]),
            (
                dictgen::InsensitiveStr::Ascii("implementattions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcitons"),
                &["constructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curriculem"),
                &["curriculum"],
            ),
            (dictgen::InsensitiveStr::Ascii("shpapes"), &["shapes"]),
            (dictgen::InsensitiveStr::Ascii("toliets"), &["toilets"]),
            (
                dictgen::InsensitiveStr::Ascii("beaurocracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("thirldy"), &["thirdly"]),
            (dictgen::InsensitiveStr::Ascii("mutch"), &["much"]),
            (
                dictgen::InsensitiveStr::Ascii("conversationg"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fromation"), &["formation"]),
            (dictgen::InsensitiveStr::Ascii("chararcter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("effectivenss"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratutiously"),
                &["gratuitously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controllling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeakable"),
                &["unbeatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("alst"), &["last"]),
            (dictgen::InsensitiveStr::Ascii("ssome"), &["some"]),
            (
                dictgen::InsensitiveStr::Ascii("coeefficients"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("texturers"), &["textures"]),
            (
                dictgen::InsensitiveStr::Ascii("significante"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("extendet"), &["extended"]),
            (dictgen::InsensitiveStr::Ascii("withun"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("setions"), &["sections"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilitie"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("attension"), &["attention"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseutically"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnessessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomat"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("peripathetic"),
                &["peripatetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("nasting"), &["nesting"]),
            (
                dictgen::InsensitiveStr::Ascii("reregisteration"),
                &["reregistration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sematnic"), &["semantic"]),
            (dictgen::InsensitiveStr::Ascii("critcally"), &["critically"]),
            (
                dictgen::InsensitiveStr::Ascii("autorization"),
                &["authorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutilcast"), &["multicast"]),
            (dictgen::InsensitiveStr::Ascii("diffeent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("exchagnes"), &["exchanges"]),
            (
                dictgen::InsensitiveStr::Ascii("defination"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("esacpes"), &["escapes"]),
            (dictgen::InsensitiveStr::Ascii("appnedix"), &["appendix"]),
            (
                dictgen::InsensitiveStr::Ascii("conversino"),
                &["conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explotiative"),
                &["exploitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("louisvile"), &["louisville"]),
            (
                dictgen::InsensitiveStr::Ascii("constisting"),
                &["consisting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conciveable"),
                &["conceivable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrbuting"),
                &["attributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbutor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independenly"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unannimously"),
                &["unanimously"],
            ),
            (dictgen::InsensitiveStr::Ascii("quanties"), &["quantities"]),
            (dictgen::InsensitiveStr::Ascii("apostroph"), &["apostrophe"]),
            (dictgen::InsensitiveStr::Ascii("acotr"), &["actor"]),
            (
                dictgen::InsensitiveStr::Ascii("consistantly"),
                &["consistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("blocack"), &["blockack"]),
            (dictgen::InsensitiveStr::Ascii("improbe"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("declaremos"), &["declares"]),
            (dictgen::InsensitiveStr::Ascii("qeueueing"), &["queueing"]),
            (dictgen::InsensitiveStr::Ascii("pengiun"), &["penguin"]),
            (
                dictgen::InsensitiveStr::Ascii("rotats"),
                &["rotates", "rotate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentcating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuers"), &["occurs"]),
            (
                dictgen::InsensitiveStr::Ascii("assoiative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("anomolies"), &["anomalies"]),
            (dictgen::InsensitiveStr::Ascii("jornals"), &["journals"]),
            (
                dictgen::InsensitiveStr::Ascii("ploretariat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("disaled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("optimizating"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("preshus"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("definitly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("reverve"), &["reserve"]),
            (dictgen::InsensitiveStr::Ascii("documnts"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("dysphroia"), &["dysphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("ressemble"),
                &["reassemble", "resemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessfull"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("courtsey"), &["courtesy"]),
            (
                dictgen::InsensitiveStr::Ascii("accordeons"),
                &["accordions"],
            ),
            (dictgen::InsensitiveStr::Ascii("directes"), &["directs"]),
            (dictgen::InsensitiveStr::Ascii("alogirthm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("cemetary"), &["cemetery"]),
            (dictgen::InsensitiveStr::Ascii("trasitive"), &["transitive"]),
            (dictgen::InsensitiveStr::Ascii("propsect"), &["prospect"]),
            (dictgen::InsensitiveStr::Ascii("inplement"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("calulators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconeects"),
                &["reconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classificatin"),
                &["classification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrealisitc"),
                &["unrealistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vibrater"),
                &["vibrator", "vibrated", "vibrates", "vibrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("skelatel"), &["skeletal"]),
            (
                dictgen::InsensitiveStr::Ascii("annaverseries"),
                &["anniversaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicidated"),
                &["indicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("phonecian"), &["phoenecian"]),
            (
                dictgen::InsensitiveStr::Ascii("runtine"),
                &["runtime", "routine"],
            ),
            (dictgen::InsensitiveStr::Ascii("warrning"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("acuired"), &["acquired"]),
            (dictgen::InsensitiveStr::Ascii("chassims"), &["chasms"]),
            (dictgen::InsensitiveStr::Ascii("nutritios"), &["nutritious"]),
            (dictgen::InsensitiveStr::Ascii("urfaces"), &["surfaces"]),
            (
                dictgen::InsensitiveStr::Ascii("duplation"),
                &["duplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("spammare"), &["spammer"]),
            (
                dictgen::InsensitiveStr::Ascii("algoirthmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physcopath"),
                &["psychopath"],
            ),
            (dictgen::InsensitiveStr::Ascii("exnternal"), &["external"]),
            (dictgen::InsensitiveStr::Ascii("hacthing"), &["hatching"]),
            (dictgen::InsensitiveStr::Ascii("enbabled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("exprectations"),
                &["expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behavrioral"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("logrithmical"),
                &["logarithmical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overcoding"),
                &["overcoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoitates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberries"),
                &["libraries", "liberties"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshage"), &["handshake"]),
            (
                dictgen::InsensitiveStr::Ascii("yourselfes"),
                &["yourselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("alerady"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("alignation"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("recrooter"), &["recruiter"]),
            (dictgen::InsensitiveStr::Ascii("officeal"), &["official"]),
            (dictgen::InsensitiveStr::Ascii("countriy"), &["country"]),
            (
                dictgen::InsensitiveStr::Ascii("deutch"),
                &["deutsch", "dutch"],
            ),
            (dictgen::InsensitiveStr::Ascii("cockateel"), &["cockatiel"]),
            (
                dictgen::InsensitiveStr::Ascii("dielectirc"),
                &["dielectric"],
            ),
            (dictgen::InsensitiveStr::Ascii("strcuture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("corener"),
                &["corner", "coroner"],
            ),
            (dictgen::InsensitiveStr::Ascii("leasure"), &["leisure"]),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonise"),
                &["synchronise"],
            ),
            (dictgen::InsensitiveStr::Ascii("beerer"), &["bearer"]),
            (
                dictgen::InsensitiveStr::Ascii("restrainig"),
                &["restraining"],
            ),
            (dictgen::InsensitiveStr::Ascii("regisered"), &["registered"]),
            (dictgen::InsensitiveStr::Ascii("chemcial"), &["chemical"]),
            (
                dictgen::InsensitiveStr::Ascii("wishpering"),
                &["whispering"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruzes"), &["cruises"]),
            (
                dictgen::InsensitiveStr::Ascii("satsifying"),
                &["satisfying"],
            ),
            (dictgen::InsensitiveStr::Ascii("malayisa"), &["malaysia"]),
            (
                dictgen::InsensitiveStr::Ascii("neglicible"),
                &["negligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranclucent"),
                &["translucent"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrise"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("advanatages"),
                &["advantages"],
            ),
            (dictgen::InsensitiveStr::Ascii("absestos"), &["asbestos"]),
            (dictgen::InsensitiveStr::Ascii("apoligize"), &["apologize"]),
            (dictgen::InsensitiveStr::Ascii("supressor"), &["suppressor"]),
            (dictgen::InsensitiveStr::Ascii("termminal"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("divorse"), &["divorce"]),
            (dictgen::InsensitiveStr::Ascii("seledts"), &["selects"]),
            (dictgen::InsensitiveStr::Ascii("preferrred"), &["preferred"]),
            (
                dictgen::InsensitiveStr::Ascii("recommennd"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("restroing"), &["restoring"]),
            (
                dictgen::InsensitiveStr::Ascii("overwride"),
                &["overwrite", "override"],
            ),
            (dictgen::InsensitiveStr::Ascii("modidfiers"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialisaing"),
                &["deserialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("arhitects"), &["architects"]),
            (dictgen::InsensitiveStr::Ascii("machanism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("thermodymamics"),
                &["thermodynamics"],
            ),
            (dictgen::InsensitiveStr::Ascii("subcribed"), &["subscribed"]),
            (
                dictgen::InsensitiveStr::Ascii("instrcutino"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenicates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresopndingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("warninsg"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("locailty"), &["locality"]),
            (dictgen::InsensitiveStr::Ascii("ancesteres"), &["ancestors"]),
            (dictgen::InsensitiveStr::Ascii("badnwagon"), &["bandwagon"]),
            (dictgen::InsensitiveStr::Ascii("timemout"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("copyrigted"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("differnce"), &["difference"]),
            (dictgen::InsensitiveStr::Ascii("miinimises"), &["minimises"]),
            (dictgen::InsensitiveStr::Ascii("recurent"), &["recurrent"]),
            (dictgen::InsensitiveStr::Ascii("usuage"), &["usage"]),
            (dictgen::InsensitiveStr::Ascii("tyo"), &["to"]),
            (
                dictgen::InsensitiveStr::Ascii("kidnappping"),
                &["kidnapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("colorao"), &["colorado"]),
            (dictgen::InsensitiveStr::Ascii("projctors"), &["projectors"]),
            (dictgen::InsensitiveStr::Ascii("negotated"), &["negotiated"]),
            (dictgen::InsensitiveStr::Ascii("imroved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("synoynms"), &["synonyms"]),
            (dictgen::InsensitiveStr::Ascii("avation"), &["aviation"]),
            (dictgen::InsensitiveStr::Ascii("lisenses"), &["licenses"]),
            (dictgen::InsensitiveStr::Ascii("formuladas"), &["formulas"]),
            (dictgen::InsensitiveStr::Ascii("dracual"), &["dracula"]),
            (dictgen::InsensitiveStr::Ascii("multplier"), &["multiplier"]),
            (dictgen::InsensitiveStr::Ascii("assumus"), &["assumes"]),
            (dictgen::InsensitiveStr::Ascii("revrieves"), &["retrieves"]),
            (dictgen::InsensitiveStr::Ascii("analysiing"), &["analysing"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availaability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("glodberg"), &["goldberg"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbibtary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("inactieve"), &["inactive"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomenting"),
                &["uncommenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartization"),
                &["standardization"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneccts"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("docunment"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("automoterator"),
                &["automoderator"],
            ),
            (dictgen::InsensitiveStr::Ascii("immitating"), &["imitating"]),
            (
                dictgen::InsensitiveStr::Ascii("approximed"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitemately"),
                &["legitimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("setion"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("aizmuth"), &["azimuth"]),
            (dictgen::InsensitiveStr::Ascii("teamspeack"), &["teamspeak"]),
            (
                dictgen::InsensitiveStr::Ascii("contribued"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorizin"),
                &["prioritizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("alighment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("necesarrily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descendands"),
                &["descendants"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpatures"), &["captures"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonize"),
                &["synchronize"],
            ),
            (dictgen::InsensitiveStr::Ascii("onl"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterization"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chesckpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("literae"), &["literate"]),
            (dictgen::InsensitiveStr::Ascii("ahtlete"), &["athlete"]),
            (
                dictgen::InsensitiveStr::Ascii("calbirator"),
                &["calibrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dynmically"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armageedon"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogynisic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omniverous"),
                &["omnivorous"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("precicely"), &["precisely"]),
            (
                dictgen::InsensitiveStr::Ascii("woked"),
                &["worked", "woke", "woken", "waked"],
            ),
            (dictgen::InsensitiveStr::Ascii("umbrellla"), &["umbrella"]),
            (dictgen::InsensitiveStr::Ascii("butotn"), &["button"]),
            (dictgen::InsensitiveStr::Ascii("boyfirend"), &["boyfriend"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffient"),
                &["inefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("incrments"), &["increments"]),
            (dictgen::InsensitiveStr::Ascii("caputre"), &["capture"]),
            (dictgen::InsensitiveStr::Ascii("retriev"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("heursitic"), &["heuristic"]),
            (dictgen::InsensitiveStr::Ascii("dentical"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("prset"), &["preset"]),
            (dictgen::InsensitiveStr::Ascii("asssertion"), &["assertion"]),
            (
                dictgen::InsensitiveStr::Ascii("straightned"),
                &["straightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluetors"),
                &["evaluators"],
            ),
            (dictgen::InsensitiveStr::Ascii("remainds"), &["remains"]),
            (
                dictgen::InsensitiveStr::Ascii("missmatches"),
                &["mismatches"],
            ),
            (dictgen::InsensitiveStr::Ascii("adresses"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("prodeed"), &["proceed"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("halfed"), &["halved"]),
            (dictgen::InsensitiveStr::Ascii("protoytpe"), &["prototype"]),
            (
                dictgen::InsensitiveStr::Ascii("convenions"),
                &["conventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unformated"),
                &["unformatted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remane"),
                &["remain", "remake", "rename"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vocabualry"),
                &["vocabulary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reveive"),
                &["receive", "revive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purposefuly"),
                &["purposefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agricolture"),
                &["agriculture"],
            ),
            (dictgen::InsensitiveStr::Ascii("inacurate"), &["inaccurate"]),
            (dictgen::InsensitiveStr::Ascii("assigment"), &["assignment"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitifer"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspontences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelerater"),
                &["accelerator", "accelerated", "accelerates", "accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peresent"),
                &["present", "presents", "presence", "percent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspnded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferible"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdencencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symapthize"),
                &["sympathize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasition"),
                &["transition", "tradition"],
            ),
            (dictgen::InsensitiveStr::Ascii("charecter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("comaptibelities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifiication"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounadary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("culumns"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("detrmining"),
                &["determining"],
            ),
            (dictgen::InsensitiveStr::Ascii("involing"), &["involving"]),
            (
                dictgen::InsensitiveStr::Ascii("cannibilism"),
                &["cannibalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interratial"),
                &["interracial"],
            ),
            (dictgen::InsensitiveStr::Ascii("caeseing"), &["ceasing"]),
            (
                dictgen::InsensitiveStr::Ascii("anonimized"),
                &["anonymized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlsators"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("peninusla"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("specialisaiton"),
                &["specialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponentation"),
                &["exponentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("queiried"), &["queried"]),
            (dictgen::InsensitiveStr::Ascii("resulvers"), &["resolvers"]),
            (
                dictgen::InsensitiveStr::Ascii("consciouslly"),
                &["consciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifciated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("warrenty"), &["warranty"]),
            (
                dictgen::InsensitiveStr::Ascii("precedance"),
                &["precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorted"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("examinin"), &["examining"]),
            (dictgen::InsensitiveStr::Ascii("configuare"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("cicruit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("puplic"), &["public"]),
            (
                dictgen::InsensitiveStr::Ascii("resest"),
                &["reset", "recessed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kindgergarten"),
                &["kindergarten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcultators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("overfows"), &["overflows"]),
            (
                dictgen::InsensitiveStr::Ascii("civelesations"),
                &["civilisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triditional"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmiting"), &["emitting"]),
            (dictgen::InsensitiveStr::Ascii("physicans"), &["physicians"]),
            (
                dictgen::InsensitiveStr::Ascii("assyemetrical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constainers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("layringes"), &["larynges"]),
            (
                dictgen::InsensitiveStr::Ascii("incorperate"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiramtion"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantacizing"),
                &["fantasizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incognition"),
                &["incognito"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virutalise"),
                &["virtualise"],
            ),
            (dictgen::InsensitiveStr::Ascii("satement"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("invertendo"), &["inverted"]),
            (
                dictgen::InsensitiveStr::Ascii("guaduloupe"),
                &["guadalupe", "guadeloupe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlaizes"),
                &["generalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("codesream"), &["codestream"]),
            (dictgen::InsensitiveStr::Ascii("congfig"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("apeending"), &["appending"]),
            (dictgen::InsensitiveStr::Ascii("noramalise"), &["normalise"]),
            (
                dictgen::InsensitiveStr::Ascii("archictures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("texure"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("contary"), &["contrary"]),
            (dictgen::InsensitiveStr::Ascii("apparrent"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("giser"), &["geyser"]),
            (
                dictgen::InsensitiveStr::Ascii("emdeddings"),
                &["embeddings"],
            ),
            (dictgen::InsensitiveStr::Ascii("combained"), &["combined"]),
            (
                dictgen::InsensitiveStr::Ascii("decomplier"),
                &["decompiler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derprecating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("exculded"), &["excluded"]),
            (dictgen::InsensitiveStr::Ascii("attemmpt"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("aquits"), &["acquits"]),
            (
                dictgen::InsensitiveStr::Ascii("unconfiged"),
                &["unconfigured"],
            ),
            (dictgen::InsensitiveStr::Ascii("dateset"), &["dataset"]),
            (dictgen::InsensitiveStr::Ascii("logial"), &["logical"]),
            (dictgen::InsensitiveStr::Ascii("claring"), &["clearing"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccesible"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("booltloaders"),
                &["bootloaders"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinsoaurs"), &["dinosaurs"]),
            (
                dictgen::InsensitiveStr::Ascii("experimntelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaive"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("probablistic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternadors"),
                &["alternators"],
            ),
            (dictgen::InsensitiveStr::Ascii("persitent"), &["persistent"]),
            (dictgen::InsensitiveStr::Ascii("interntal"), &["internal"]),
            (dictgen::InsensitiveStr::Ascii("eleased"), &["released"]),
            (
                dictgen::InsensitiveStr::Ascii("apollstree"),
                &["upholstery"],
            ),
            (dictgen::InsensitiveStr::Ascii("assgin"), &["assign"]),
            (dictgen::InsensitiveStr::Ascii("laundrey"), &["laundry"]),
            (dictgen::InsensitiveStr::Ascii("subystem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("transpareny"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtituted"),
                &["substituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("partices"), &["particles"]),
            (dictgen::InsensitiveStr::Ascii("appove"), &["approve"]),
            (
                dictgen::InsensitiveStr::Ascii("crytopgraphic"),
                &["cryptographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negativaty"),
                &["negativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("begginging"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("transcrpit"),
                &["transcript"],
            ),
            (dictgen::InsensitiveStr::Ascii("powerpot"), &["powerpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("correspodents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenisonality"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphisizes"),
                &["emphasizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("optino"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("qustionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambassaders"),
                &["ambassadors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customziing"),
                &["customizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitmaks"), &["bitmask"]),
            (dictgen::InsensitiveStr::Ascii("vendeta"), &["vendetta"]),
            (dictgen::InsensitiveStr::Ascii("sytntax"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("objectificiation"),
                &["objectification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captialisation"),
                &["capitalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("recors"), &["records"]),
            (
                dictgen::InsensitiveStr::Ascii("requiremenets"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strengtened"),
                &["strengthened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singools"),
                &["singles", "signals"],
            ),
            (dictgen::InsensitiveStr::Ascii("transiten"), &["transient"]),
            (dictgen::InsensitiveStr::Ascii("shutodwn"), &["shutdown"]),
            (
                dictgen::InsensitiveStr::Ascii("launh"),
                &["launch", "laugh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wheights"),
                &["weights", "heights"],
            ),
            (dictgen::InsensitiveStr::Ascii("kuberetes"), &["kubernetes"]),
            (dictgen::InsensitiveStr::Ascii("vehicule"), &["vehicle"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanetally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("culrpit"), &["culprit"]),
            (dictgen::InsensitiveStr::Ascii("metioning"), &["mentioning"]),
            (dictgen::InsensitiveStr::Ascii("trown"), &["thrown"]),
            (
                dictgen::InsensitiveStr::Ascii("psychedelicious"),
                &["psychedelics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constriant"),
                &["constraint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checlerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favilitating"),
                &["facilitating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaklyrie"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("reseration"),
                &["reservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfify"), &["verify"]),
            (
                dictgen::InsensitiveStr::Ascii("noramlized"),
                &["normalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomata"), &["automata"]),
            (
                dictgen::InsensitiveStr::Ascii("decompositon"),
                &["decomposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("structere"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("adderessing"),
                &["addressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("othre"), &["other", "ochre"]),
            (dictgen::InsensitiveStr::Ascii("patroit"), &["patriot"]),
            (
                dictgen::InsensitiveStr::Ascii("villification"),
                &["vilification"],
            ),
            (dictgen::InsensitiveStr::Ascii("dervied"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("backpakcs"), &["backpacks"]),
            (
                dictgen::InsensitiveStr::Ascii("stardardizes"),
                &["standardizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("addapts"), &["adapts"]),
            (dictgen::InsensitiveStr::Ascii("wipin"), &["wiping"]),
            (dictgen::InsensitiveStr::Ascii("livley"), &["lively"]),
            (dictgen::InsensitiveStr::Ascii("uploadds"), &["uploads"]),
            (dictgen::InsensitiveStr::Ascii("evoluton"), &["evolution"]),
            (dictgen::InsensitiveStr::Ascii("actuaal"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("ploblems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("invariablly"),
                &["invariably"],
            ),
            (dictgen::InsensitiveStr::Ascii("preivew"), &["preview"]),
            (dictgen::InsensitiveStr::Ascii("minizes"), &["minimizes"]),
            (dictgen::InsensitiveStr::Ascii("tranducer"), &["transducer"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilty"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currenices"),
                &["currencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("speling"), &["spelling"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinnates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touchsceen"),
                &["touchscreen"],
            ),
            (dictgen::InsensitiveStr::Ascii("diminsion"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("anaytical"), &["analytical"]),
            (dictgen::InsensitiveStr::Ascii("pedophilie"), &["pedophile"]),
            (dictgen::InsensitiveStr::Ascii("reecord"), &["record"]),
            (dictgen::InsensitiveStr::Ascii("favorits"), &["favorites"]),
            (
                dictgen::InsensitiveStr::Ascii("ambassidors"),
                &["ambassadors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponetiation"),
                &["exponentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("necessay"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("cubburd"), &["cupboard"]),
            (
                dictgen::InsensitiveStr::Ascii("emphesized"),
                &["emphasized"],
            ),
            (dictgen::InsensitiveStr::Ascii("existiong"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("domesitcated"),
                &["domesticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbets"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("squeakey"), &["squeaky"]),
            (
                dictgen::InsensitiveStr::Ascii("stripting"),
                &["scripting", "stripping"],
            ),
            (dictgen::InsensitiveStr::Ascii("constat"), &["constant"]),
            (
                dictgen::InsensitiveStr::Ascii("horizontale"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depricates"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("igzorting"), &["exhorting"]),
            (
                dictgen::InsensitiveStr::Ascii("originatin"),
                &["originating", "origination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassembly"),
                &["disassembly"],
            ),
            (dictgen::InsensitiveStr::Ascii("stormade"), &["stormed"]),
            (dictgen::InsensitiveStr::Ascii("synatx"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("propotions"),
                &["proportions", "promotions"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabile"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("publucation"),
                &["publication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condifurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("casued"), &["caused"]),
            (dictgen::InsensitiveStr::Ascii("patroitic"), &["patriotic"]),
            (
                dictgen::InsensitiveStr::Ascii("interactice"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("desribe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("lmiters"), &["limiters"]),
            (dictgen::InsensitiveStr::Ascii("wonderlad"), &["wonderland"]),
            (dictgen::InsensitiveStr::Ascii("procdural"), &["procedural"]),
            (
                dictgen::InsensitiveStr::Ascii("revoluions"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminaison"),
                &["termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("iternator"), &["iterator"]),
            (
                dictgen::InsensitiveStr::Ascii("consequenty"),
                &["consequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("harmoniacs"), &["harmonic"]),
            (dictgen::InsensitiveStr::Ascii("actullay"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("supected"), &["suspected"]),
            (
                dictgen::InsensitiveStr::Ascii("futer"),
                &["further", "future"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsessfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimzie"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("complaing"),
                &["complaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experementation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginser"),
                &["recogniser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percusions"),
                &["percussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crsytalline"),
                &["crystalline"],
            ),
            (dictgen::InsensitiveStr::Ascii("soory"), &["sorry"]),
            (dictgen::InsensitiveStr::Ascii("ovrride"), &["override"]),
            (
                dictgen::InsensitiveStr::Ascii("optomising"),
                &["optimising"],
            ),
            (dictgen::InsensitiveStr::Ascii("rationaly"), &["rationally"]),
            (dictgen::InsensitiveStr::Ascii("defiend"), &["defined"]),
            (
                dictgen::InsensitiveStr::Ascii("normallises"),
                &["normalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("announcemets"),
                &["announcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("correced"), &["corrected"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinatin"),
                &["coordinating", "coordination"],
            ),
            (dictgen::InsensitiveStr::Ascii("declearing"), &["declaring"]),
            (dictgen::InsensitiveStr::Ascii("terurns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("refrerenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoting"), &["supporting"]),
            (
                dictgen::InsensitiveStr::Ascii("therapautic"),
                &["therapeutic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cintainer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("diagonsis"), &["diagnosis"]),
            (
                dictgen::InsensitiveStr::Ascii("commmitment"),
                &["commitment"],
            ),
            (dictgen::InsensitiveStr::Ascii("coralina"), &["carolina"]),
            (dictgen::InsensitiveStr::Ascii("veryfing"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("corrputed"), &["corrupted"]),
            (dictgen::InsensitiveStr::Ascii("comarable"), &["comparable"]),
            (
                dictgen::InsensitiveStr::Ascii("namesd"),
                &["named", "names"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microscoop"),
                &["microscope"],
            ),
            (dictgen::InsensitiveStr::Ascii("qunatum"), &["quantum"]),
            (dictgen::InsensitiveStr::Ascii("mesages"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("retreeve"), &["retrieve"]),
            (
                dictgen::InsensitiveStr::Ascii("resepective"),
                &["respective"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanages"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("enumurator"),
                &["enumerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nedding"),
                &["needing", "nodding"],
            ),
            (dictgen::InsensitiveStr::Ascii("succedes"), &["succeeds"]),
            (
                dictgen::InsensitiveStr::Ascii("interruptis"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carbohdyrates"),
                &["carbohydrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("holdins"), &["holdings"]),
            (dictgen::InsensitiveStr::Ascii("obersvant"), &["observant"]),
            (
                dictgen::InsensitiveStr::Ascii("confliting"),
                &["conflating", "conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("compielers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("frequenies"),
                &["frequencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skeptisism"),
                &["skepticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("strinsg"), &["strings"]),
            (
                dictgen::InsensitiveStr::Ascii("euthanaisa"),
                &["euthanasia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("portraiing"),
                &["portraying"],
            ),
            (dictgen::InsensitiveStr::Ascii("sillybus"), &["syllabus"]),
            (
                dictgen::InsensitiveStr::Ascii("pariticipant"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politicanti"),
                &["politician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diagnositics"),
                &["diagnostics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misognyistic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("daquiri"), &["daiquiri"]),
            (
                dictgen::InsensitiveStr::Ascii("triscaidecafobia"),
                &["triskaidekaphobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misspellled"),
                &["misspelled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intolernace"),
                &["intolerance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experminent"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("aircrafts"), &["aircraft"]),
            (
                dictgen::InsensitiveStr::Ascii("manualls"),
                &["manuals", "manually"],
            ),
            (dictgen::InsensitiveStr::Ascii("tucan"), &["toucan"]),
            (
                dictgen::InsensitiveStr::Ascii("arameterization"),
                &["parameterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sowe"), &["sow"]),
            (dictgen::InsensitiveStr::Ascii("moveement"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("multplicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("venezuella"), &["venezuela"]),
            (
                dictgen::InsensitiveStr::Ascii("opprotunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkcsums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("diferrentiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calncelling"),
                &["cancelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("robotis"), &["robotics"]),
            (dictgen::InsensitiveStr::Ascii("chenging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatiev"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorrrectly"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("varialbe"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("dueration"), &["duration"]),
            (dictgen::InsensitiveStr::Ascii("bundeler"), &["bundler"]),
            (
                dictgen::InsensitiveStr::Ascii("commitmet"),
                &["commitments"],
            ),
            (dictgen::InsensitiveStr::Ascii("comor"), &["color"]),
            (
                dictgen::InsensitiveStr::Ascii("aprreciating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathimatician"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("binarry"), &["binary"]),
            (dictgen::InsensitiveStr::Ascii("leceister"), &["leicester"]),
            (dictgen::InsensitiveStr::Ascii("chochkas"), &["tchotchkes"]),
            (
                dictgen::InsensitiveStr::Ascii("hetrogeneous"),
                &["heterogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("apect"), &["aspect"]),
            (
                dictgen::InsensitiveStr::Ascii("erraticly"),
                &["erratically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostethics"),
                &["prosthetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("suplle"), &["supple"]),
            (
                dictgen::InsensitiveStr::Ascii("descirptive"),
                &["descriptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("stronlgy"), &["strongly"]),
            (dictgen::InsensitiveStr::Ascii("vecvtors"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("eyt"), &["yet"]),
            (
                dictgen::InsensitiveStr::Ascii("corressponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("detailled"), &["detailed"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretato"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("congif"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("decomissioning"),
                &["decommissioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("supoprt"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("shoult"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("featurng"), &["featuring"]),
            (
                dictgen::InsensitiveStr::Ascii("replacemends"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscripton"),
                &["subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interected"),
                &["interacted", "intersected"],
            ),
            (dictgen::InsensitiveStr::Ascii("solutiin"), &["solution"]),
            (
                dictgen::InsensitiveStr::Ascii("sirectionless"),
                &["directionless"],
            ),
            (dictgen::InsensitiveStr::Ascii("undescore"), &["underscore"]),
            (
                dictgen::InsensitiveStr::Ascii("slector"),
                &["selector", "sector", "elector"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outragesouly"),
                &["outrageously"],
            ),
            (dictgen::InsensitiveStr::Ascii("norml"), &["normal", "norm"]),
            (
                dictgen::InsensitiveStr::Ascii("subsitutuent"),
                &["substituent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infomational"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilitiy"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("avare"), &["aware"]),
            (dictgen::InsensitiveStr::Ascii("vaules"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("mucnhies"), &["munchies"]),
            (dictgen::InsensitiveStr::Ascii("milivolts"), &["millivolts"]),
            (dictgen::InsensitiveStr::Ascii("injustics"), &["injustices"]),
            (
                dictgen::InsensitiveStr::Ascii("thse"),
                &["the", "these", "this", "those"],
            ),
            (dictgen::InsensitiveStr::Ascii("finishs"), &["finishes"]),
            (
                dictgen::InsensitiveStr::Ascii("capibilities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("subbtle"), &["subtle"]),
            (
                dictgen::InsensitiveStr::Ascii("addiionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("srings"), &["strings"]),
            (
                dictgen::InsensitiveStr::Ascii("singols"),
                &["signals", "singles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somethinhg"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanetizer"), &["sanitizer"]),
            (dictgen::InsensitiveStr::Ascii("meoldic"), &["melodic"]),
            (
                dictgen::InsensitiveStr::Ascii("govormental"),
                &["governmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retributioon"),
                &["retribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pusblishing"),
                &["publishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbhor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvieniently"),
                &["inconveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilsations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alhabeticaly"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceptioin"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("invulnarable"),
                &["invulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("direvctory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("surgury"), &["surgery"]),
            (dictgen::InsensitiveStr::Ascii("defaulfs"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("abudances"), &["abundances"]),
            (
                dictgen::InsensitiveStr::Ascii("contigency"),
                &["contingency"],
            ),
            (dictgen::InsensitiveStr::Ascii("obeservers"), &["observers"]),
            (dictgen::InsensitiveStr::Ascii("limiation"), &["limitation"]),
            (dictgen::InsensitiveStr::Ascii("innosense"), &["innocence"]),
            (
                dictgen::InsensitiveStr::Ascii("crresponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("veryfication"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("corus"), &["chorus"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapulate"),
                &["extrapolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coniders"),
                &["considers", "conifers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expexnasions"),
                &["expansions"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfromed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("harldine"), &["hardline"]),
            (
                dictgen::InsensitiveStr::Ascii("justifcation"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presure"),
                &["pressure", "presume"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synthesasia"),
                &["synthesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciving"), &["receiving"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastibating"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("medevial"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("macor"), &["macro"]),
            (
                dictgen::InsensitiveStr::Ascii("chechkboxes"),
                &["checkboxes"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrption"), &["encryption"]),
            (
                dictgen::InsensitiveStr::Ascii("professionallism"),
                &["professionalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defautly"),
                &["defaultly", "default"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challanged"),
                &["challenged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previleged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoer"), &["follower"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtat"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("kepping"), &["keeping"]),
            (
                dictgen::InsensitiveStr::Ascii("begininings"),
                &["beginnings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriving"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("deprectate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("indpendence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commant"),
                &["command", "comment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palythrough"),
                &["playthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("vairances"), &["variances"]),
            (
                dictgen::InsensitiveStr::Ascii("schozophrenia"),
                &["schizophrenia"],
            ),
            (dictgen::InsensitiveStr::Ascii("identic"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("desciribed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstrats"),
                &["demonstrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arcehtypes"),
                &["archetypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("typechek"), &["typecheck"]),
            (
                dictgen::InsensitiveStr::Ascii("assmebling"),
                &["assembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("embededded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("regiems"), &["regimes"]),
            (dictgen::InsensitiveStr::Ascii("cleannups"), &["cleanups"]),
            (
                dictgen::InsensitiveStr::Ascii("ephemereally"),
                &["ephemerally"],
            ),
            (dictgen::InsensitiveStr::Ascii("upploads"), &["uploads"]),
            (
                dictgen::InsensitiveStr::Ascii("ellitpical"),
                &["elliptical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motherbard"),
                &["motherboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scarificing"),
                &["sacrificing"],
            ),
            (dictgen::InsensitiveStr::Ascii("beseige"), &["besiege"]),
            (dictgen::InsensitiveStr::Ascii("comapre"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("acchieveable"),
                &["achievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicitelly"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("landshapes"),
                &["landscapes"],
            ),
            (dictgen::InsensitiveStr::Ascii("utitlties"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("sepecifier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("exorbitent"),
                &["exorbitant"],
            ),
            (dictgen::InsensitiveStr::Ascii("reponding"), &["responding"]),
            (
                dictgen::InsensitiveStr::Ascii("identifieed"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("preriod"), &["period"]),
            (dictgen::InsensitiveStr::Ascii("measue"), &["measure"]),
            (
                dictgen::InsensitiveStr::Ascii("comptabile"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuethest"), &["furthest"]),
            (dictgen::InsensitiveStr::Ascii("dprecates"), &["deprecates"]),
            (dictgen::InsensitiveStr::Ascii("hardwarte"), &["hardware"]),
            (dictgen::InsensitiveStr::Ascii("euklidean"), &["euclidean"]),
            (
                dictgen::InsensitiveStr::Ascii("disatisfaction"),
                &["dissatisfaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workbenchs"),
                &["workbenches"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfieir"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("enhancenment"),
                &["enhancement"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbr"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("dasy"),
                &["daisy", "dash", "days", "easy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasversed"),
                &["traversed", "transversed"],
            ),
            (dictgen::InsensitiveStr::Ascii("menues"), &["menus"]),
            (dictgen::InsensitiveStr::Ascii("appllied"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("operting"), &["operating"]),
            (dictgen::InsensitiveStr::Ascii("euqality"), &["equality"]),
            (
                dictgen::InsensitiveStr::Ascii("progagated"),
                &["propagated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ehthernet"), &["ethernet"]),
            (
                dictgen::InsensitiveStr::Ascii("componding"),
                &["compounding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("methos"),
                &["methods", "method"],
            ),
            (dictgen::InsensitiveStr::Ascii("restet"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("isntrument"),
                &["instrument"],
            ),
            (dictgen::InsensitiveStr::Ascii("elimnates"), &["eliminates"]),
            (
                dictgen::InsensitiveStr::Ascii("impossibel"),
                &["impossibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqiored"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("executabe"), &["executable"]),
            (dictgen::InsensitiveStr::Ascii("guarnete"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("vunrability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("iin"), &["in"]),
            (
                dictgen::InsensitiveStr::Ascii("oakereously"),
                &["ocherously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definintion"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfunctionnal"),
                &["dysfunctional"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlysing"), &["analysing"]),
            (
                dictgen::InsensitiveStr::Ascii("allcateing"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componsate"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distincive"),
                &["distinctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custonised"),
                &["customised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diagronalizations"),
                &["diagonalizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determiens"),
                &["determines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transplantes"),
                &["transplants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficity"),
                &["specificity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxiates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("junglign"), &["jungling"]),
            (
                dictgen::InsensitiveStr::Ascii("paravirtualiation"),
                &["paravirtualisation", "paravirtualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essenatially"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exlusionary"),
                &["exclusionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apostrophei"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permuated"),
                &["permutated", "permuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hipothetically"),
                &["hypothetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("testting"), &["testing"]),
            (dictgen::InsensitiveStr::Ascii("followint"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinally"),
                &["terminally"],
            ),
            (dictgen::InsensitiveStr::Ascii("narcissim"), &["narcissism"]),
            (
                dictgen::InsensitiveStr::Ascii("converying"),
                &["converting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pretentieus"),
                &["pretentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pertitence"),
                &["pertinence", "persistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuido"), &["studio"]),
            (dictgen::InsensitiveStr::Ascii("wresters"), &["wrestlers"]),
            (dictgen::InsensitiveStr::Ascii("costomer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("intential"),
                &["intentional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterz"),
                &["characterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("shaddow"), &["shadow"]),
            (dictgen::InsensitiveStr::Ascii("negotaite"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("definateky"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapoled"),
                &["extrapolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginised"),
                &["recognised"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxiumums"), &["maximums"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimising"),
                &["optimising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogrithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("mispelt"), &["misspelt"]),
            (dictgen::InsensitiveStr::Ascii("delagate"), &["delegate"]),
            (
                dictgen::InsensitiveStr::Ascii("interescted"),
                &["intersected"],
            ),
            (dictgen::InsensitiveStr::Ascii("ukrainina"), &["ukrainian"]),
            (dictgen::InsensitiveStr::Ascii("archtypes"), &["archetypes"]),
            (dictgen::InsensitiveStr::Ascii("countert"), &["counteract"]),
            (
                dictgen::InsensitiveStr::Ascii("libguistic"),
                &["linguistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("unpplied"), &["unapplied"]),
            (dictgen::InsensitiveStr::Ascii("uniqe"), &["unique"]),
            (
                dictgen::InsensitiveStr::Ascii("emprisonment"),
                &["imprisonment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychodelics"),
                &["psychedelics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whiped"),
                &["whipped", "wiped"],
            ),
            (dictgen::InsensitiveStr::Ascii("quiestions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("neckbeardos"),
                &["neckbeards"],
            ),
            (dictgen::InsensitiveStr::Ascii("bufffering"), &["buffering"]),
            (
                dictgen::InsensitiveStr::Ascii("vegetairan"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocat"), &["allocate"]),
            (dictgen::InsensitiveStr::Ascii("nasts"), &["masts", "nests"]),
            (dictgen::InsensitiveStr::Ascii("compresss"), &["compress"]),
            (
                dictgen::InsensitiveStr::Ascii("mechancial"),
                &["mechanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiblities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progagations"),
                &["propagations"],
            ),
            (dictgen::InsensitiveStr::Ascii("porgramme"), &["programme"]),
            (
                dictgen::InsensitiveStr::Ascii("emissed"),
                &["amassed", "amiss"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assmumption"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("hrlpers"), &["helpers"]),
            (dictgen::InsensitiveStr::Ascii("publishr"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("configureable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("feals"), &["feels"]),
            (dictgen::InsensitiveStr::Ascii("bounhdary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("propmts"), &["prompts"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("monoclinc"), &["monoclinic"]),
            (dictgen::InsensitiveStr::Ascii("usied"), &["busied", "used"]),
            (dictgen::InsensitiveStr::Ascii("sirected"), &["directed"]),
            (dictgen::InsensitiveStr::Ascii("articafts"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("inteaction"),
                &["interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("adveristy"), &["adversity"]),
            (
                dictgen::InsensitiveStr::Ascii("thunderbot"),
                &["thunderbolt"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapear"), &["disappear"]),
            (
                dictgen::InsensitiveStr::Ascii("attriution"),
                &["attribution", "attrition"],
            ),
            (dictgen::InsensitiveStr::Ascii("edtions"), &["editions"]),
            (
                dictgen::InsensitiveStr::Ascii("percentille"),
                &["percentile"],
            ),
            (dictgen::InsensitiveStr::Ascii("hsitory"), &["history"]),
            (
                dictgen::InsensitiveStr::Ascii("recognizion"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformacion"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginers"), &["beginners"]),
            (
                dictgen::InsensitiveStr::Ascii("limters"),
                &["limiters", "linters"],
            ),
            (dictgen::InsensitiveStr::Ascii("conservies"), &["conserve"]),
            (dictgen::InsensitiveStr::Ascii("decrment"), &["decrement"]),
            (dictgen::InsensitiveStr::Ascii("peodphile"), &["pedophile"]),
            (dictgen::InsensitiveStr::Ascii("jeoprady"), &["jeopardy"]),
            (
                dictgen::InsensitiveStr::Ascii("convoultions"),
                &["convolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developme"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deinitailse"),
                &["deinitialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribues"), &["attributes"]),
            (
                dictgen::InsensitiveStr::Ascii("inconistencies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quots"),
                &["quotes", "quits"],
            ),
            (dictgen::InsensitiveStr::Ascii("peventing"), &["preventing"]),
            (dictgen::InsensitiveStr::Ascii("viariables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("longiude"), &["longitude"]),
            (
                dictgen::InsensitiveStr::Ascii("similiarty"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutlitude"), &["multitude"]),
            (
                dictgen::InsensitiveStr::Ascii("libertariansim"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("indonesa"), &["indonesia"]),
            (dictgen::InsensitiveStr::Ascii("instanze"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("laughablly"), &["laughably"]),
            (dictgen::InsensitiveStr::Ascii("initailse"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("pragamtic"), &["pragmatic"]),
            (dictgen::InsensitiveStr::Ascii("accended"), &["ascended"]),
            (dictgen::InsensitiveStr::Ascii("madule"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("prohibirte"), &["prohibit"]),
            (
                dictgen::InsensitiveStr::Ascii("charitible"),
                &["charitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversatoin"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarnacion"),
                &["incarnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpsondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presitgious"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("roate"), &["rotate"]),
            (
                dictgen::InsensitiveStr::Ascii("shepperding"),
                &["shepherding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depentency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otimization"),
                &["optimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciricular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("adjustament"),
                &["adjustment"],
            ),
            (dictgen::InsensitiveStr::Ascii("weilded"), &["wielded"]),
            (dictgen::InsensitiveStr::Ascii("modlues"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("installl"), &["install"]),
            (dictgen::InsensitiveStr::Ascii("juxtified"), &["justified"]),
            (dictgen::InsensitiveStr::Ascii("interprut"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("eary"),
                &["eery", "early", "easy", "eerie"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterration"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("guaruanteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insustainable"),
                &["unsustainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrilated"),
                &["correlated"],
            ),
            (dictgen::InsensitiveStr::Ascii("messenge"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("specialzes"),
                &["specializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("varriants"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("initailses"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hopsitalized"),
                &["hospitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificationkits"),
                &["certifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("tobbaco"), &["tobacco"]),
            (
                dictgen::InsensitiveStr::Ascii("responsabile"),
                &["responsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("recived"), &["received"]),
            (
                dictgen::InsensitiveStr::Ascii("specififier"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("numerious"), &["numerous"]),
            (dictgen::InsensitiveStr::Ascii("mathcing"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("prostiture"),
                &["prostitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repesenting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("frontlinie"), &["frontline"]),
            (dictgen::InsensitiveStr::Ascii("ommits"), &["omits"]),
            (
                dictgen::InsensitiveStr::Ascii("egostitical"),
                &["egotistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misoygnist"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("impoved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("israleis"), &["israelis"]),
            (dictgen::InsensitiveStr::Ascii("ceckboxes"), &["checkboxes"]),
            (dictgen::InsensitiveStr::Ascii("utiltities"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("knoledge"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("destoried"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("boyfrients"),
                &["boyfriends"],
            ),
            (dictgen::InsensitiveStr::Ascii("knownledge"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("toursity"), &["touristy"]),
            (
                dictgen::InsensitiveStr::Ascii("suggestins"),
                &["suggestions"],
            ),
            (dictgen::InsensitiveStr::Ascii("represnts"), &["represents"]),
            (dictgen::InsensitiveStr::Ascii("brigaged"), &["brigade"]),
            (
                dictgen::InsensitiveStr::Ascii("characthers"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("hmdi"), &["hdmi"]),
            (
                dictgen::InsensitiveStr::Ascii("preferenes"),
                &["preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condictioning"),
                &["conditioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("margianlly"),
                &["marginally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferance"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossiblble"),
                &["impossible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functonality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destryoing"),
                &["destroying"],
            ),
            (dictgen::InsensitiveStr::Ascii("outupt"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("pricision"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("frimware"), &["firmware"]),
            (
                dictgen::InsensitiveStr::Ascii("addrersssed"),
                &["addressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unimpressd"),
                &["unimpressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inelligible"),
                &["ineligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteritic"),
                &["characteristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimmenters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathes"),
                &["paths", "patches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperitave"),
                &["imperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripters"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("corproate"), &["corporate"]),
            (dictgen::InsensitiveStr::Ascii("useddd"), &["used"]),
            (dictgen::InsensitiveStr::Ascii("collectin"), &["collection"]),
            (dictgen::InsensitiveStr::Ascii("forulas"), &["formulas"]),
            (dictgen::InsensitiveStr::Ascii("prolbems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("speacial"),
                &["special", "spacial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maninpulations"),
                &["manipulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactets"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("porblem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("anxiuosly"), &["anxiously"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomited"),
                &["uncommitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definatlely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expectionally"),
                &["exceptionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("engery"), &["energy"]),
            (dictgen::InsensitiveStr::Ascii("arbitray"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("spectauclar"),
                &["spectacular"],
            ),
            (dictgen::InsensitiveStr::Ascii("hepled"), &["helped"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenstional"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumentes"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("recongnizer"),
                &["recognizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculcated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassemblying"),
                &["disassembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("auther"), &["author"]),
            (
                dictgen::InsensitiveStr::Ascii("transimssion"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("argubly"), &["arguably"]),
            (
                dictgen::InsensitiveStr::Ascii("measurd"),
                &["measured", "measure"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrtten"), &["written"]),
            (
                dictgen::InsensitiveStr::Ascii("compiltaion"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetsers"), &["testers"]),
            (dictgen::InsensitiveStr::Ascii("passsing"), &["passing"]),
            (dictgen::InsensitiveStr::Ascii("staement"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("scupture"), &["sculpture"]),
            (dictgen::InsensitiveStr::Ascii("residule"), &["residue"]),
            (
                dictgen::InsensitiveStr::Ascii("recomendation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatatbility"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("somewher"), &["somewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("blueprintcss"),
                &["blueprints"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceptted"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("discreto"), &["discretion"]),
            (
                dictgen::InsensitiveStr::Ascii("drescription"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("elmenets"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("anlytics"), &["analytics"]),
            (
                dictgen::InsensitiveStr::Ascii("availibility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("grabed"), &["grabbed"]),
            (dictgen::InsensitiveStr::Ascii("embrios"), &["embryos"]),
            (dictgen::InsensitiveStr::Ascii("indcates"), &["indicates"]),
            (dictgen::InsensitiveStr::Ascii("ignitin"), &["ignition"]),
            (
                dictgen::InsensitiveStr::Ascii("translatron"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealiarating"),
                &["ameliorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("crossfiter"), &["crossfire"]),
            (dictgen::InsensitiveStr::Ascii("hiddden"), &["hidden"]),
            (
                dictgen::InsensitiveStr::Ascii("enhancemants"),
                &["enhancements"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellide"), &["elide"]),
            (dictgen::InsensitiveStr::Ascii("averageed"), &["averaged"]),
            (dictgen::InsensitiveStr::Ascii("cashires"), &["cashiers"]),
            (dictgen::InsensitiveStr::Ascii("coutners"), &["counters"]),
            (dictgen::InsensitiveStr::Ascii("discotek"), &["discotheque"]),
            (dictgen::InsensitiveStr::Ascii("passivelly"), &["passively"]),
            (dictgen::InsensitiveStr::Ascii("spefififer"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("overlordess"),
                &["overlords"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procentual"),
                &["percentual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protruberances"),
                &["protuberances"],
            ),
            (dictgen::InsensitiveStr::Ascii("updatin"), &["updating"]),
            (dictgen::InsensitiveStr::Ascii("acssume"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("contancted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("mape"), &["map"]),
            (
                dictgen::InsensitiveStr::Ascii("adjsutments"),
                &["adjustments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstated"),
                &["demonstrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envirionments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("groouped"), &["grouped"]),
            (dictgen::InsensitiveStr::Ascii("anmesia"), &["amnesia"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastinatin"),
                &["procrastination", "procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("opposits"), &["opposites"]),
            (
                dictgen::InsensitiveStr::Ascii("suborutine"),
                &["subroutine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitabilty"),
                &["profitability"],
            ),
            (dictgen::InsensitiveStr::Ascii("spammear"), &["spammer"]),
            (dictgen::InsensitiveStr::Ascii("scetioned"), &["sectioned"]),
            (
                dictgen::InsensitiveStr::Ascii("assoicating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subequently"),
                &["subsequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconeections"),
                &["reconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("pundents"), &["pundits"]),
            (
                dictgen::InsensitiveStr::Ascii("newmatically"),
                &["pneumatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("devoloped"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhodd"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stimilants"),
                &["stimulants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructtor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveinent"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steriotyping"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deployements"),
                &["deployments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassociative"),
                &["dissociative"],
            ),
            (dictgen::InsensitiveStr::Ascii("maschinery"), &["machinery"]),
            (dictgen::InsensitiveStr::Ascii("altriusm"), &["altruism"]),
            (
                dictgen::InsensitiveStr::Ascii("butterfiles"),
                &["butterflies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technicaly"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiyhin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("discrimante"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneating"), &["generating"]),
            (
                dictgen::InsensitiveStr::Ascii("stipped"),
                &["stripped", "stopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anager"),
                &["manager", "anger"],
            ),
            (dictgen::InsensitiveStr::Ascii("systhem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("delutional"),
                &["delusional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publicisin"),
                &["publicising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anthropoloy"),
                &["anthropology"],
            ),
            (dictgen::InsensitiveStr::Ascii("benchmarck"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("realignements"),
                &["realignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("drnik"), &["drink"]),
            (dictgen::InsensitiveStr::Ascii("indiquates"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("automagicall"),
                &["automagically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgournds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("potatoe"), &["potato"]),
            (
                dictgen::InsensitiveStr::Ascii("tendensies"),
                &["tendencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloacate"), &["allocate"]),
            (dictgen::InsensitiveStr::Ascii("multliply"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("initailsiation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("furtest"), &["furthest"]),
            (dictgen::InsensitiveStr::Ascii("behavour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("sprinf"), &["sprintf"]),
            (dictgen::InsensitiveStr::Ascii("intresst"), &["interest"]),
            (
                dictgen::InsensitiveStr::Ascii("disadvanteged"),
                &["disadvantaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("anynomity"), &["anonymity"]),
            (
                dictgen::InsensitiveStr::Ascii("delcarative"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("bytets"), &["bytes"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilism"),
                &["cannibalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("insepected"), &["inspected"]),
            (dictgen::InsensitiveStr::Ascii("agressvie"), &["aggressive"]),
            (
                dictgen::InsensitiveStr::Ascii("custumizable"),
                &["customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acculumation"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacusly"), &["vacuously"]),
            (
                dictgen::InsensitiveStr::Ascii("separetions"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermimenter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unphyiscal"),
                &["unphysical"],
            ),
            (dictgen::InsensitiveStr::Ascii("hosrtname"), &["hostname"]),
            (dictgen::InsensitiveStr::Ascii("morrisound"), &["morrison"]),
            (
                dictgen::InsensitiveStr::Ascii("calcultors"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortuante"), &["fortunate"]),
            (dictgen::InsensitiveStr::Ascii("brigthen"), &["brighten"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinator"),
                &["terminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("paresrs"), &["parsers"]),
            (dictgen::InsensitiveStr::Ascii("selecet"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("pallette"), &["palette"]),
            (dictgen::InsensitiveStr::Ascii("minimini"), &["minimizing"]),
            (dictgen::InsensitiveStr::Ascii("qurying"), &["querying"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnetally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoluitons"),
                &["resolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("senitnel"), &["sentinel"]),
            (
                dictgen::InsensitiveStr::Ascii("comittment"),
                &["commitment"],
            ),
            (dictgen::InsensitiveStr::Ascii("ginee"), &["guinea"]),
            (dictgen::InsensitiveStr::Ascii("incurrs"), &["incurs"]),
            (dictgen::InsensitiveStr::Ascii("directy"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("ramdomising"),
                &["randomising"],
            ),
            (dictgen::InsensitiveStr::Ascii("caridac"), &["cardiac"]),
            (dictgen::InsensitiveStr::Ascii("nazereth"), &["nazareth"]),
            (
                dictgen::InsensitiveStr::Ascii("instatiations"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captacitors"),
                &["capacitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("obtion"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("constructie"),
                &["constructive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consturcted"),
                &["constructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canabelyze"),
                &["cannibalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operatational"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectute"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("privelleges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transferrd"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apreteating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rigtheousness"),
                &["righteousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("nce"), &["nice", "once"]),
            (
                dictgen::InsensitiveStr::Ascii("hurriganes"),
                &["hurricanes"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesitcles"), &["testicles"]),
            (
                dictgen::InsensitiveStr::Ascii("leavin"),
                &["leaving", "leaven"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initation"),
                &["initiation", "imitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersperesing"),
                &["interspersing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mambership"),
                &["membership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renaiscance"),
                &["renaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("thruout"), &["throughout"]),
            (dictgen::InsensitiveStr::Ascii("acquried"), &["acquired"]),
            (
                dictgen::InsensitiveStr::Ascii("congrigation"),
                &["congregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("startlxde"), &["startled"]),
            (dictgen::InsensitiveStr::Ascii("achors"), &["anchors"]),
            (dictgen::InsensitiveStr::Ascii("gentialia"), &["genitalia"]),
            (
                dictgen::InsensitiveStr::Ascii("ubsubscribed"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("platnuim"), &["platinum"]),
            (dictgen::InsensitiveStr::Ascii("worknig"), &["working"]),
            (dictgen::InsensitiveStr::Ascii("dijsktra"), &["dijkstra"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterprating"),
                &["misinterpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adequatley"),
                &["adequately"],
            ),
            (dictgen::InsensitiveStr::Ascii("itereate"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("annonce"), &["announce"]),
            (
                dictgen::InsensitiveStr::Ascii("dublications"),
                &["publications", "duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("histroy"), &["history"]),
            (
                dictgen::InsensitiveStr::Ascii("assassinats"),
                &["assassins", "assassinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("respondas"), &["responds"]),
            (
                dictgen::InsensitiveStr::Ascii("fallhrough"),
                &["fallthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("bactracks"), &["backtracks"]),
            (dictgen::InsensitiveStr::Ascii("bengahzi"), &["benghazi"]),
            (dictgen::InsensitiveStr::Ascii("syringue"), &["syringe"]),
            (
                dictgen::InsensitiveStr::Ascii("subscritpions"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggragation"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("responeds"), &["responds"]),
            (
                dictgen::InsensitiveStr::Ascii("stroes"),
                &["stores", "strokes", "strobes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeurhing"),
                &["neighbouring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deomcratic"),
                &["democratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("potental"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("chekpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demographical"),
                &["demographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techiniques"),
                &["techniques"],
            ),
            (dictgen::InsensitiveStr::Ascii("canceles"), &["cancels"]),
            (
                dictgen::InsensitiveStr::Ascii("feasability"),
                &["feasibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasforms"), &["transforms"]),
            (
                dictgen::InsensitiveStr::Ascii("dismbiguate"),
                &["disambiguate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substutite"),
                &["substitutes", "substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("recomput"), &["recompute"]),
            (dictgen::InsensitiveStr::Ascii("iteates"), &["iterates"]),
            (
                dictgen::InsensitiveStr::Ascii("docuemnted"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("vailidty"), &["validity"]),
            (dictgen::InsensitiveStr::Ascii("absovling"), &["absolving"]),
            (dictgen::InsensitiveStr::Ascii("itearted"), &["iterated"]),
            (dictgen::InsensitiveStr::Ascii("uruguary"), &["uruguay"]),
            (
                dictgen::InsensitiveStr::Ascii("succesully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonexistant"),
                &["nonexistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immortaly"),
                &["immortality"],
            ),
            (dictgen::InsensitiveStr::Ascii("assosiate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("dissambling"),
                &["disassembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("equipmentd"), &["equipment"]),
            (dictgen::InsensitiveStr::Ascii("dignosing"), &["diagnosing"]),
            (dictgen::InsensitiveStr::Ascii("expectd"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("lunaticos"), &["lunatics"]),
            (
                dictgen::InsensitiveStr::Ascii("miscelaneous"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakends"), &["backends"]),
            (dictgen::InsensitiveStr::Ascii("comect"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("cylyndrical"),
                &["cylindrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("receivd"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("distnce"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("commedian"), &["comedian"]),
            (
                dictgen::InsensitiveStr::Ascii("ditributed"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("semaphor"), &["semaphore"]),
            (dictgen::InsensitiveStr::Ascii("thsnk"), &["thank"]),
            (dictgen::InsensitiveStr::Ascii("againe"), &["again"]),
            (
                dictgen::InsensitiveStr::Ascii("indeginous"),
                &["indigenous"],
            ),
            (dictgen::InsensitiveStr::Ascii("compnay"), &["company"]),
            (dictgen::InsensitiveStr::Ascii("handhake"), &["handshake"]),
            (dictgen::InsensitiveStr::Ascii("manfiesto"), &["manifesto"]),
            (dictgen::InsensitiveStr::Ascii("aggreived"), &["aggrieved"]),
            (dictgen::InsensitiveStr::Ascii("moibles"), &["mobiles"]),
            (
                dictgen::InsensitiveStr::Ascii("practicioner"),
                &["practitioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("progapate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("suceses"), &["successes"]),
            (
                dictgen::InsensitiveStr::Ascii("fingerpoint"),
                &["fingerprint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acftually"),
                &["factually", "actually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evauluators"),
                &["evaluators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consctructors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("anitvirus"), &["antivirus"]),
            (
                dictgen::InsensitiveStr::Ascii("assosciation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatie"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("achevie"), &["achieve"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinitive"),
                &["indefinite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upsreaming"),
                &["upstreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("fitler"), &["filter"]),
            (
                dictgen::InsensitiveStr::Ascii("selectivley"),
                &["selectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downstiars"),
                &["downstairs"],
            ),
            (dictgen::InsensitiveStr::Ascii("conpares"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("conpleteness"),
                &["completeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenitier"), &["identifier"]),
            (dictgen::InsensitiveStr::Ascii("armchar"), &["armchair"]),
            (
                dictgen::InsensitiveStr::Ascii("distribitor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nunnecssary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("rountine"), &["routine"]),
            (dictgen::InsensitiveStr::Ascii("movemnets"), &["movements"]),
            (
                dictgen::InsensitiveStr::Ascii("particiapte"),
                &["participate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhouds"),
                &["neighbourhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aceleration"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancerns"), &["cancers"]),
            (dictgen::InsensitiveStr::Ascii("guaranteey"), &["guaranty"]),
            (
                dictgen::InsensitiveStr::Ascii("discusssions"),
                &["discussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("marginaal"), &["marginal"]),
            (dictgen::InsensitiveStr::Ascii("entitites"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("attampting"),
                &["attempting"],
            ),
            (dictgen::InsensitiveStr::Ascii("assertin"), &["assertions"]),
            (dictgen::InsensitiveStr::Ascii("alhapets"), &["alphabets"]),
            (dictgen::InsensitiveStr::Ascii("exolicit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("promethesus"),
                &["prometheus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amatures"),
                &["amateurs", "armatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("catesion"), &["cartesian"]),
            (
                dictgen::InsensitiveStr::Ascii("ciruclation"),
                &["circulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cients"), &["clients"]),
            (dictgen::InsensitiveStr::Ascii("shevles"), &["shelves"]),
            (
                dictgen::InsensitiveStr::Ascii("expeimentelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("usser"), &["user", "usher"]),
            (
                dictgen::InsensitiveStr::Ascii("splet"),
                &["spelt", "split", "splat"],
            ),
            (dictgen::InsensitiveStr::Ascii("priorites"), &["priorities"]),
            (
                dictgen::InsensitiveStr::Ascii("imigrated"),
                &["emigrated", "immigrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componenents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agggregated"),
                &["aggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("poluting"), &["polluting"]),
            (
                dictgen::InsensitiveStr::Ascii("superceding"),
                &["superseding"],
            ),
            (dictgen::InsensitiveStr::Ascii("elminate"), &["eliminate"]),
            (dictgen::InsensitiveStr::Ascii("servising"), &["servicing"]),
            (
                dictgen::InsensitiveStr::Ascii("eviserates"),
                &["eviscerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boilderplate"),
                &["boilerplate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exprimented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("upater"), &["updater"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitplier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("modift"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapoling"),
                &["extrapolating"],
            ),
            (dictgen::InsensitiveStr::Ascii("daita"), &["data"]),
            (
                dictgen::InsensitiveStr::Ascii("spacialising"),
                &["specialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revew"),
                &["review", "renew", "revel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frankensite"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sportsmansship"),
                &["sportsmanship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrespekting"),
                &["disrespecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("higth"), &["height"]),
            (
                dictgen::InsensitiveStr::Ascii("approproximately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("dateteims"), &["datetimes"]),
            (dictgen::InsensitiveStr::Ascii("yeld"), &["yield"]),
            (
                dictgen::InsensitiveStr::Ascii("accicently"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongizing"),
                &["recognizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceelings"), &["ceilings"]),
            (
                dictgen::InsensitiveStr::Ascii("hyphotesis"),
                &["hypothesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoprt"), &["report"]),
            (
                dictgen::InsensitiveStr::Ascii("justiifcation"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alienatin"),
                &["alienating", "alienation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomorro"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("speciffies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("progagation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substibute"),
                &["substitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiable"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("minial"), &["minimal"]),
            (
                dictgen::InsensitiveStr::Ascii("deterimnes"),
                &["determines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationalsim"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouevre"), &["oeuvre"]),
            (
                dictgen::InsensitiveStr::Ascii("socioecomonic"),
                &["socioeconomic"],
            ),
            (dictgen::InsensitiveStr::Ascii("novisheate"), &["novitiate"]),
            (dictgen::InsensitiveStr::Ascii("fillls"), &["fills"]),
            (
                dictgen::InsensitiveStr::Ascii("commites"),
                &["commits", "committed"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaned"), &["meant"]),
            (
                dictgen::InsensitiveStr::Ascii("nowns"),
                &["knowns", "nouns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgruntaled"),
                &["disgruntled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubeshooters"),
                &["troubleshooters"],
            ),
            (dictgen::InsensitiveStr::Ascii("recuisive"), &["recursive"]),
            (dictgen::InsensitiveStr::Ascii("clevely"), &["cleverly"]),
            (
                dictgen::InsensitiveStr::Ascii("spefififers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefally"),
                &["specially", "specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singnificant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("cywgin"), &["cygwin"]),
            (
                dictgen::InsensitiveStr::Ascii("trustworthly"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferances"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("colordao"), &["colorado"]),
            (dictgen::InsensitiveStr::Ascii("lettters"), &["letters"]),
            (
                dictgen::InsensitiveStr::Ascii("strucutural"),
                &["structural"],
            ),
            (dictgen::InsensitiveStr::Ascii("taveled"), &["traveled"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciatve"),
                &["appreciative"],
            ),
            (dictgen::InsensitiveStr::Ascii("howover"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlations"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("becnhmark"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("manigested"),
                &["manifested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclimation"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("leran"), &["learn"]),
            (
                dictgen::InsensitiveStr::Ascii("archaology"),
                &["archeology", "archaeology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularlising"),
                &["regularising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porgrammeer"),
                &["programmer"],
            ),
            (dictgen::InsensitiveStr::Ascii("sofwtare"), &["software"]),
            (dictgen::InsensitiveStr::Ascii("patenterad"), &["patented"]),
            (dictgen::InsensitiveStr::Ascii("resolviste"), &["resolves"]),
            (dictgen::InsensitiveStr::Ascii("sapces"), &["spaces"]),
            (
                dictgen::InsensitiveStr::Ascii("regaurdless"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unacompanied"),
                &["unaccompanied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatabable"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfurrd"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("prommpt"), &["prompt"]),
            (
                dictgen::InsensitiveStr::Ascii("separetors"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("forat"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("conparison"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consttructers"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("excutors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("intersectuion"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("workspces"), &["workspaces"]),
            (dictgen::InsensitiveStr::Ascii("tio"), &["to"]),
            (dictgen::InsensitiveStr::Ascii("outpouts"), &["outputs"]),
            (
                dictgen::InsensitiveStr::Ascii("precesion"),
                &["precession", "precision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fanfaction"),
                &["fanfiction"],
            ),
            (dictgen::InsensitiveStr::Ascii("implets"), &["implements"]),
            (dictgen::InsensitiveStr::Ascii("exhuasive"), &["exhaustive"]),
            (dictgen::InsensitiveStr::Ascii("reshedule"), &["reschedule"]),
            (dictgen::InsensitiveStr::Ascii("containns"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("deifnes"), &["defines"]),
            (dictgen::InsensitiveStr::Ascii("celllars"), &["cellars"]),
            (dictgen::InsensitiveStr::Ascii("inderts"), &["inserts"]),
            (dictgen::InsensitiveStr::Ascii("disccuss"), &["discuss"]),
            (
                dictgen::InsensitiveStr::Ascii("exerimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decompresor"),
                &["decompressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meger"),
                &["merger", "merge", "meager"],
            ),
            (dictgen::InsensitiveStr::Ascii("soldires"), &["soldiers"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalististic"),
                &["nationalistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciales"), &["specials"]),
            (
                dictgen::InsensitiveStr::Ascii("persitently"),
                &["persistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("troubing"), &["troubling"]),
            (
                dictgen::InsensitiveStr::Ascii("acheivments"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("entryes"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("agruably"), &["arguably"]),
            (
                dictgen::InsensitiveStr::Ascii("confrirmation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gloriuos"), &["glorious"]),
            (dictgen::InsensitiveStr::Ascii("paeth"), &["path"]),
            (
                dictgen::InsensitiveStr::Ascii("autheniticators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiztions"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("startin"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("intilized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacjkward"), &["backward"]),
            (
                dictgen::InsensitiveStr::Ascii("creater"),
                &[
                    "creature", "creator", "crater", "created", "creates", "create",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("reformerad"), &["reformed"]),
            (
                dictgen::InsensitiveStr::Ascii("kimerrical"),
                &["chimerical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ccessible"), &["accessible"]),
            (dictgen::InsensitiveStr::Ascii("deovtion"), &["devotion"]),
            (
                dictgen::InsensitiveStr::Ascii("perceptiuon"),
                &["perception"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoerist"), &["theorist"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsequentional"),
                &["inconsequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("sover"), &["solver"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowldegements"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("binares"), &["binaries"]),
            (dictgen::InsensitiveStr::Ascii("systhems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("authentciators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doocumentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoeong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("computationalaly"),
                &["computationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("readble"), &["readable"]),
            (
                dictgen::InsensitiveStr::Ascii("avalibaility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("startegies"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuerable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("nasueous"), &["nauseous"]),
            (
                dictgen::InsensitiveStr::Ascii("acadamically"),
                &["academically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserailizes"),
                &["deserializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulbearable"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("simualte"), &["simulate"]),
            (dictgen::InsensitiveStr::Ascii("dota"), &["data"]),
            (dictgen::InsensitiveStr::Ascii("buuld"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("prepartions"),
                &["preparations"],
            ),
            (dictgen::InsensitiveStr::Ascii("amendement"), &["amendment"]),
            (dictgen::InsensitiveStr::Ascii("microsft"), &["microsoft"]),
            (dictgen::InsensitiveStr::Ascii("actiive"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("elelments"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("photogrpah"),
                &["photographs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corellation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instiatiate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("subculter"), &["subculture"]),
            (dictgen::InsensitiveStr::Ascii("labirynth"), &["labyrinth"]),
            (dictgen::InsensitiveStr::Ascii("tanslates"), &["translates"]),
            (
                dictgen::InsensitiveStr::Ascii("beutification"),
                &["beautification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implicitley"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("coform"), &["conform"]),
            (dictgen::InsensitiveStr::Ascii("minsicule"), &["miniscule"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequiste"),
                &["prerequisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("turains"), &["terrains"]),
            (dictgen::InsensitiveStr::Ascii("freckels"), &["freckles"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenisons"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("paricles"), &["particles"]),
            (dictgen::InsensitiveStr::Ascii("victem"), &["victim"]),
            (
                dictgen::InsensitiveStr::Ascii("combatents"),
                &["combatants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("couterpart"),
                &["counterpart"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimimal"), &["optimal"]),
            (
                dictgen::InsensitiveStr::Ascii("favilitator"),
                &["facilitator"],
            ),
            (dictgen::InsensitiveStr::Ascii("knoweldge"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("preliminarly"),
                &["preliminary"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangted"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("diagnally"), &["diagonally"]),
            (dictgen::InsensitiveStr::Ascii("marilyin"), &["marilyn"]),
            (
                dictgen::InsensitiveStr::Ascii("councellors"),
                &["councillors", "counsellors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imprementation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substiution"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relocateable"),
                &["relocatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("accross"), &["across"]),
            (dictgen::InsensitiveStr::Ascii("lenths"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("garuantees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("handlign"), &["handling"]),
            (dictgen::InsensitiveStr::Ascii("seperared"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("custcene"), &["cutscene"]),
            (dictgen::InsensitiveStr::Ascii("tolerenace"), &["tolerance"]),
            (
                dictgen::InsensitiveStr::Ascii("comparions"),
                &["comparison", "comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crednetials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("assuption"), &["assumption"]),
            (dictgen::InsensitiveStr::Ascii("imapct"), &["impact"]),
            (dictgen::InsensitiveStr::Ascii("runnining"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("fior"), &["for"]),
            (dictgen::InsensitiveStr::Ascii("visuales"), &["visuals"]),
            (dictgen::InsensitiveStr::Ascii("overloade"), &["overload"]),
            (dictgen::InsensitiveStr::Ascii("dudo"), &["sudo"]),
            (
                dictgen::InsensitiveStr::Ascii("coallessing"),
                &["coalescing"],
            ),
            (dictgen::InsensitiveStr::Ascii("pretendend"), &["pretended"]),
            (dictgen::InsensitiveStr::Ascii("marbels"), &["marbles"]),
            (dictgen::InsensitiveStr::Ascii("associats"), &["associates"]),
            (
                dictgen::InsensitiveStr::Ascii("quivalence"),
                &["equivalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estiamatively"),
                &["estimatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("progrewss"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("analyzis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("annihalition"),
                &["annihilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meoization"),
                &["memoization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utilitization"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summurized"),
                &["summarized"],
            ),
            (dictgen::InsensitiveStr::Ascii("startegic"), &["strategic"]),
            (dictgen::InsensitiveStr::Ascii("anlyzing"), &["analyzing"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstncial"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("necssary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("bolean"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("burnign"), &["burning"]),
            (dictgen::InsensitiveStr::Ascii("archtect"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("emphatised"),
                &["emphasised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associato"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("folwoiong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("destroyeer"),
                &["destroyers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laest"),
                &["least", "latest"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirynge"), &["syringe"]),
            (
                dictgen::InsensitiveStr::Ascii("estabished"),
                &["established"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lexigraphic"),
                &["lexicographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnteed"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("einstances"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("assasssin"),
                &["assassins", "assassin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoticating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("biginning"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("imperealist"),
                &["imperialist"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterrating"), &["iterating"]),
            (dictgen::InsensitiveStr::Ascii("annyos"), &["annoys"]),
            (dictgen::InsensitiveStr::Ascii("keybroad"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("finsihes"), &["finishes"]),
            (dictgen::InsensitiveStr::Ascii("specifiec"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("relievin"), &["relieving"]),
            (
                dictgen::InsensitiveStr::Ascii("audodetect"),
                &["autodetect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituiton"),
                &["prostitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakup"), &["backup"]),
            (
                dictgen::InsensitiveStr::Ascii("copyeighted"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pretedermined"),
                &["predetermined"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrekcing"), &["wrecking"]),
            (
                dictgen::InsensitiveStr::Ascii("containint"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("copys"), &["copies"]),
            (
                dictgen::InsensitiveStr::Ascii("miticateing"),
                &["mitigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("triagular"), &["triangular"]),
            (dictgen::InsensitiveStr::Ascii("genearally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("iomaped"), &["iomapped"]),
            (dictgen::InsensitiveStr::Ascii("amrchairs"), &["armchairs"]),
            (
                dictgen::InsensitiveStr::Ascii("creatin"),
                &["creation", "creating", "creatine", "cretin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symmectries"),
                &["symmetries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recipientes"),
                &["recipients"],
            ),
            (dictgen::InsensitiveStr::Ascii("salaraies"), &["salaries"]),
            (
                dictgen::InsensitiveStr::Ascii("simptumaticaly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doocumentations"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitts"), &["limits"]),
            (
                dictgen::InsensitiveStr::Ascii("congratule"),
                &["congratulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviornment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conferances"),
                &["conferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("fromating"), &["formatting"]),
            (
                dictgen::InsensitiveStr::Ascii("activite"),
                &["activities", "activity", "activate"],
            ),
            (dictgen::InsensitiveStr::Ascii("neames"), &["names"]),
            (dictgen::InsensitiveStr::Ascii("syntak"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("contributiont"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("certainity"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("atomiticity"),
                &["atomicity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstucts"),
                &["reconstructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("provive"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("unopend"), &["unopened"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabitlies"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("baubes"), &["baubles"]),
            (dictgen::InsensitiveStr::Ascii("betehsda"), &["bethesda"]),
            (dictgen::InsensitiveStr::Ascii("mirgates"), &["migrates"]),
            (dictgen::InsensitiveStr::Ascii("geryscale"), &["greyscale"]),
            (
                dictgen::InsensitiveStr::Ascii("unperturbe"),
                &["unperturbed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanatized"), &["sanitized"]),
            (dictgen::InsensitiveStr::Ascii("scroates"), &["socrates"]),
            (dictgen::InsensitiveStr::Ascii("togheter"), &["together"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallics"),
                &["specifics"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostspots"), &["hotspots"]),
            (dictgen::InsensitiveStr::Ascii("blief"), &["belief"]),
            (
                dictgen::InsensitiveStr::Ascii("infulences"),
                &["influences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introsepection"),
                &["introspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinstaled"),
                &["reinstalled"],
            ),
            (dictgen::InsensitiveStr::Ascii("refeshed"), &["refreshed"]),
            (
                dictgen::InsensitiveStr::Ascii("prevente"),
                &["prevented", "prevent"],
            ),
            (dictgen::InsensitiveStr::Ascii("releaased"), &["released"]),
            (
                dictgen::InsensitiveStr::Ascii("catterpillars"),
                &["caterpillars"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhauting"), &["exhausting"]),
            (dictgen::InsensitiveStr::Ascii("chizziling"), &["chiseling"]),
            (dictgen::InsensitiveStr::Ascii("pipeliend"), &["pipelined"]),
            (dictgen::InsensitiveStr::Ascii("rmove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("concatinates"),
                &["concatenates"],
            ),
            (dictgen::InsensitiveStr::Ascii("usuallly"), &["usually"]),
            (
                dictgen::InsensitiveStr::Ascii("destruktors"),
                &["destructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("fiercly"), &["fiercely"]),
            (
                dictgen::InsensitiveStr::Ascii("visability"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("interivew"), &["interview"]),
            (dictgen::InsensitiveStr::Ascii("assersive"), &["assertive"]),
            (dictgen::InsensitiveStr::Ascii("upgading"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("traumitized"),
                &["traumatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferencing"),
                &["dereferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypethetically"),
                &["hypothetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connnectors"),
                &["connectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("grouepd"), &["grouped"]),
            (
                dictgen::InsensitiveStr::Ascii("partitiones"),
                &["partitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutionilized"),
                &["institutionalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("tropcial"), &["tropical"]),
            (dictgen::InsensitiveStr::Ascii("distince"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("skillshits"),
                &["skillshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoting"), &["annotating"]),
            (
                dictgen::InsensitiveStr::Ascii("utiliztion"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscirbed"),
                &["subscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedestrain"),
                &["pedestrian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innitialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibley"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinite"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viusalizes"),
                &["visualizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahgning"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("assignend"), &["assigned"]),
            (dictgen::InsensitiveStr::Ascii("somethng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("enivornments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("propsal"), &["proposal"]),
            (
                dictgen::InsensitiveStr::Ascii("distrobuting"),
                &["distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dfined"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("addcits"), &["addicts"]),
            (dictgen::InsensitiveStr::Ascii("amensia"), &["amnesia"]),
            (dictgen::InsensitiveStr::Ascii("instamt"), &["instant"]),
            (
                dictgen::InsensitiveStr::Ascii("intstructing"),
                &["instructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("testice"), &["testicle"]),
            (dictgen::InsensitiveStr::Ascii("adrersses"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("armistace"), &["armistice"]),
            (
                dictgen::InsensitiveStr::Ascii("dpeendences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceptd"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("orginiated"),
                &["originated"],
            ),
            (dictgen::InsensitiveStr::Ascii("crytpic"), &["cryptic"]),
            (
                dictgen::InsensitiveStr::Ascii("scatches"),
                &["catches", "scratches", "sketches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpaerntly"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("diabler"), &["disabler"]),
            (dictgen::InsensitiveStr::Ascii("wednesdey"), &["wednesday"]),
            (
                dictgen::InsensitiveStr::Ascii("argumements"),
                &["arguments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marger"),
                &["merger", "marker"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomtries"), &["geometries"]),
            (dictgen::InsensitiveStr::Ascii("squrtile"), &["squirtle"]),
            (dictgen::InsensitiveStr::Ascii("brwosing"), &["browsing"]),
            (dictgen::InsensitiveStr::Ascii("menetion"), &["mention"]),
            (dictgen::InsensitiveStr::Ascii("hunderds"), &["hundreds"]),
            (
                dictgen::InsensitiveStr::Ascii("respecitve"),
                &["respective"],
            ),
            (dictgen::InsensitiveStr::Ascii("aything"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("dialectos"), &["dialects"]),
            (dictgen::InsensitiveStr::Ascii("feltcher"), &["fletcher"]),
            (dictgen::InsensitiveStr::Ascii("clinet"), &["client"]),
            (dictgen::InsensitiveStr::Ascii("emlation"), &["emulation"]),
            (dictgen::InsensitiveStr::Ascii("evertyime"), &["everytime"]),
            (dictgen::InsensitiveStr::Ascii("srcolls"), &["scrolls"]),
            (
                dictgen::InsensitiveStr::Ascii("commutablility"),
                &["commutability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peprocessor"),
                &["preprocessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("sofomores"), &["sophomores"]),
            (dictgen::InsensitiveStr::Ascii("geters"), &["getters"]),
            (dictgen::InsensitiveStr::Ascii("colletive"), &["collective"]),
            (dictgen::InsensitiveStr::Ascii("cylcist"), &["cyclist"]),
            (
                dictgen::InsensitiveStr::Ascii("restriccion"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sybolically"),
                &["symbolically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheerleadra"),
                &["cheerleader"],
            ),
            (dictgen::InsensitiveStr::Ascii("airrports"), &["airports"]),
            (
                dictgen::InsensitiveStr::Ascii("tangetially"),
                &["tangentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("abtracted"), &["abstracted"]),
            (dictgen::InsensitiveStr::Ascii("intereview"), &["interview"]),
            (
                dictgen::InsensitiveStr::Ascii("contiunous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrinution"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("defeine"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("quadraped"), &["quadruped"]),
            (
                dictgen::InsensitiveStr::Ascii("recustions"),
                &["recursions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qualifikation"),
                &["qualification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occationly"),
                &["occasionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thouroghly"),
                &["thoroughly"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiill"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("consideracion"),
                &["consideration"],
            ),
            (dictgen::InsensitiveStr::Ascii("carnomies"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("micorwaves"),
                &["microwaves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximited"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goreshadowing"),
                &["foreshadowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantaneu"),
                &["instantaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("cambodai"), &["cambodia"]),
            (dictgen::InsensitiveStr::Ascii("vlone"), &["clone"]),
            (
                dictgen::InsensitiveStr::Ascii("poesession"),
                &["possession"],
            ),
            (dictgen::InsensitiveStr::Ascii("anytihng"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("centerns"), &["centers"]),
            (dictgen::InsensitiveStr::Ascii("artillary"), &["artillery"]),
            (dictgen::InsensitiveStr::Ascii("comamnds"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("discretation"),
                &["discretization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repreciate"),
                &["depreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("myslef"), &["myself"]),
            (dictgen::InsensitiveStr::Ascii("roght"), &["right"]),
            (dictgen::InsensitiveStr::Ascii("boradly"), &["broadly"]),
            (dictgen::InsensitiveStr::Ascii("ellitot"), &["elliott"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegothiation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hiptser"), &["hipster"]),
            (dictgen::InsensitiveStr::Ascii("attchment"), &["attachment"]),
            (
                dictgen::InsensitiveStr::Ascii("horizantally"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sevral"), &["several"]),
            (
                dictgen::InsensitiveStr::Ascii("sarts"),
                &["stars", "starts"],
            ),
            (dictgen::InsensitiveStr::Ascii("assissting"), &["assisting"]),
            (
                dictgen::InsensitiveStr::Ascii("deffinitively"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuids"), &["rebuilds"]),
            (
                dictgen::InsensitiveStr::Ascii("asnychronous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("miliraty"), &["military"]),
            (dictgen::InsensitiveStr::Ascii("checkcsum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("tablespooons"),
                &["tablespoon"],
            ),
            (dictgen::InsensitiveStr::Ascii("computre"), &["compute"]),
            (dictgen::InsensitiveStr::Ascii("torelance"), &["tolerance"]),
            (dictgen::InsensitiveStr::Ascii("andoird"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("awakend"), &["awakened"]),
            (dictgen::InsensitiveStr::Ascii("occourring"), &["occurring"]),
            (dictgen::InsensitiveStr::Ascii("tattooos"), &["tattoos"]),
            (dictgen::InsensitiveStr::Ascii("excersise"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("syracue"), &["syracuse"]),
            (dictgen::InsensitiveStr::Ascii("simlicity"), &["simplicity"]),
            (
                dictgen::InsensitiveStr::Ascii("drivern"),
                &["driven", "driver"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffussion"), &["diffusion"]),
            (dictgen::InsensitiveStr::Ascii("abliities"), &["abilities"]),
            (
                dictgen::InsensitiveStr::Ascii("auotmatically"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilcating"),
                &["complicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merget"),
                &["merge", "merged", "merger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifible"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("arquive"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("alocating"), &["allocating"]),
            (
                dictgen::InsensitiveStr::Ascii("discontigious"),
                &["discontiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilckers"), &["clickers"]),
            (
                dictgen::InsensitiveStr::Ascii("vegetarion"),
                &["vegetarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubtions"),
                &["distributions", "disruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("burtsing"), &["bursting"]),
            (
                dictgen::InsensitiveStr::Ascii("misrepresenation"),
                &["misrepresentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counstructor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotiotations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defninitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condidional"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descendend"),
                &["descended", "descendent"],
            ),
            (dictgen::InsensitiveStr::Ascii("receiption"), &["reception"]),
            (dictgen::InsensitiveStr::Ascii("condiiton"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("unsigne"), &["unsigned"]),
            (
                dictgen::InsensitiveStr::Ascii("configiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astethically"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transacrtional"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirertion"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("shdow"), &["shadow", "show"]),
            (dictgen::InsensitiveStr::Ascii("firmeware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("anyaway"), &["anyway"]),
            (dictgen::InsensitiveStr::Ascii("epilepsey"), &["epilepsy"]),
            (
                dictgen::InsensitiveStr::Ascii("proocedures"),
                &["procedures"],
            ),
            (dictgen::InsensitiveStr::Ascii("churchs"), &["churches"]),
            (dictgen::InsensitiveStr::Ascii("unahppy"), &["unhappy"]),
            (
                dictgen::InsensitiveStr::Ascii("separatedly"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepatated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifiies"),
                &["qualifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("versatille"), &["versatile"]),
            (
                dictgen::InsensitiveStr::Ascii("theoeretics"),
                &["theoretics"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrnage"), &["arrange"]),
            (dictgen::InsensitiveStr::Ascii("ehenever"), &["whenever"]),
            (dictgen::InsensitiveStr::Ascii("reoslve"), &["resolve"]),
            (dictgen::InsensitiveStr::Ascii("evovler"), &["evolver"]),
            (dictgen::InsensitiveStr::Ascii("lanaguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("ganes"), &["games"]),
            (dictgen::InsensitiveStr::Ascii("camaera"), &["camera"]),
            (dictgen::InsensitiveStr::Ascii("assingn"), &["assign"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("deposite"), &["deposit"]),
            (dictgen::InsensitiveStr::Ascii("cantainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("attribiutes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("holf"), &["hold"]),
            (dictgen::InsensitiveStr::Ascii("secretley"), &["secretly"]),
            (dictgen::InsensitiveStr::Ascii("disgustoso"), &["disgusts"]),
            (
                dictgen::InsensitiveStr::Ascii("argicultural"),
                &["agricultural"],
            ),
            (dictgen::InsensitiveStr::Ascii("concevied"), &["conceived"]),
            (
                dictgen::InsensitiveStr::Ascii("perpective"),
                &["perspective", "perceptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratuituous"),
                &["gratuitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectivily"),
                &["collectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifcation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutsence"), &["cutscene"]),
            (dictgen::InsensitiveStr::Ascii("mirorr"), &["mirror"]),
            (dictgen::InsensitiveStr::Ascii("seziures"), &["seizures"]),
            (dictgen::InsensitiveStr::Ascii("scenaireo"), &["scenario"]),
            (dictgen::InsensitiveStr::Ascii("woring"), &["working"]),
            (
                dictgen::InsensitiveStr::Ascii("opimization"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experession"),
                &["expression"],
            ),
            (dictgen::InsensitiveStr::Ascii("driectx"), &["directx"]),
            (dictgen::InsensitiveStr::Ascii("negoatiate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("inidicating"),
                &["indicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unformattet"),
                &["unformatted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somethingest"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("stunami"), &["tsunami"]),
            (dictgen::InsensitiveStr::Ascii("assining"), &["assigning"]),
            (
                dictgen::InsensitiveStr::Ascii("priortises"),
                &["prioritises"],
            ),
            (dictgen::InsensitiveStr::Ascii("failied"), &["failed"]),
            (dictgen::InsensitiveStr::Ascii("gruped"), &["grouped"]),
            (
                dictgen::InsensitiveStr::Ascii("desitinations"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ortolaryngology"),
                &["otolaryngology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precidence"),
                &["precedence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transtioning"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoratically"),
                &["theoretically"],
            ),
            (dictgen::InsensitiveStr::Ascii("atached"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("gramatical"),
                &["grammatical"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosmeticos"), &["cosmetics"]),
            (dictgen::InsensitiveStr::Ascii("ulimate"), &["ultimate"]),
            (
                dictgen::InsensitiveStr::Ascii("confirguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("maneouvre"), &["manoeuvre"]),
            (
                dictgen::InsensitiveStr::Ascii("intruiging"),
                &["intriguing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wyas"), &["ways"]),
            (
                dictgen::InsensitiveStr::Ascii("paramterizing"),
                &["parameterizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arachive"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("comands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("preapred"), &["prepared"]),
            (
                dictgen::InsensitiveStr::Ascii("consntantly"),
                &["constantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decleratively"),
                &["declaratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholarstic"),
                &["scholastic", "scholarly"],
            ),
            (dictgen::InsensitiveStr::Ascii("grapgical"), &["graphical"]),
            (dictgen::InsensitiveStr::Ascii("patronas"), &["patrons"]),
            (
                dictgen::InsensitiveStr::Ascii("envolving"),
                &["involving", "evolving"],
            ),
            (dictgen::InsensitiveStr::Ascii("objectss"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("anonymou"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("tak"), &["take"]),
            (
                dictgen::InsensitiveStr::Ascii("geneations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highligths"),
                &["highlights"],
            ),
            (dictgen::InsensitiveStr::Ascii("imortance"), &["importance"]),
            (
                dictgen::InsensitiveStr::Ascii("asymetrical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massachustts"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("addidition"), &["addition"]),
            (
                dictgen::InsensitiveStr::Ascii("opjectifications"),
                &["objectifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("spawnned"), &["spawned"]),
            (
                dictgen::InsensitiveStr::Ascii("condtioned"),
                &["conditioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("stragegy"), &["strategy"]),
            (dictgen::InsensitiveStr::Ascii("wizrd"), &["wizard"]),
            (dictgen::InsensitiveStr::Ascii("guarente"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("nothern"), &["northern"]),
            (
                dictgen::InsensitiveStr::Ascii("pascheurise"),
                &["pasteurise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sprin"),
                &[
                    "spring", "sprint", "spin", "spain", "sarin", "sprig", "sprit",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("overhread"), &["overhead"]),
            (
                dictgen::InsensitiveStr::Ascii("aprroximately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paragaraphas"),
                &["paragraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepeters"), &["accepters"]),
            (
                dictgen::InsensitiveStr::Ascii("archictural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counries"),
                &["countries", "counties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumeratin"),
                &["enumerating", "enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribucion"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("assummes"), &["assumes"]),
            (dictgen::InsensitiveStr::Ascii("knowngly"), &["knowingly"]),
            (
                dictgen::InsensitiveStr::Ascii("expemtions"),
                &["exemptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destroyd"),
                &["destroyed", "destroys"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrupteds"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifciate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriburtion"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("mandarijn"), &["mandarin"]),
            (dictgen::InsensitiveStr::Ascii("sitirring"), &["stirring"]),
            (
                dictgen::InsensitiveStr::Ascii("sweetshart"),
                &["sweetheart"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheilds"), &["shields"]),
            (
                dictgen::InsensitiveStr::Ascii("ancapsulating"),
                &["encapsulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggresively"),
                &["aggressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introducin"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architectue"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("advnace"), &["advance"]),
            (
                dictgen::InsensitiveStr::Ascii("notidication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("strngs"), &["strings"]),
            (dictgen::InsensitiveStr::Ascii("cartilidge"), &["cartilage"]),
            (dictgen::InsensitiveStr::Ascii("separatae"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("patholgoical"),
                &["pathological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coditionals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("sourec"), &["source"]),
            (dictgen::InsensitiveStr::Ascii("dektop"), &["desktop"]),
            (dictgen::InsensitiveStr::Ascii("accessort"), &["accessor"]),
            (
                dictgen::InsensitiveStr::Ascii("acqauinted"),
                &["acquainted"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysus"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("containg"), &["containing"]),
            (
                dictgen::InsensitiveStr::Ascii("enviromentally"),
                &["environmentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hurricanger"),
                &["hurricane"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqensed"), &["sequenced"]),
            (dictgen::InsensitiveStr::Ascii("ethernal"), &["eternal"]),
            (dictgen::InsensitiveStr::Ascii("uncanney"), &["uncanny"]),
            (dictgen::InsensitiveStr::Ascii("managet"), &["manager"]),
            (dictgen::InsensitiveStr::Ascii("fnishing"), &["finishing"]),
            (dictgen::InsensitiveStr::Ascii("heavnely"), &["heavenly"]),
            (dictgen::InsensitiveStr::Ascii("evluate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("preservare"), &["preserve"]),
            (
                dictgen::InsensitiveStr::Ascii("envoked"),
                &["evoked", "invoked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceuctions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("amde"), &["made"]),
            (dictgen::InsensitiveStr::Ascii("fusipn"), &["fusion"]),
            (dictgen::InsensitiveStr::Ascii("undfined"), &["undefined"]),
            (
                dictgen::InsensitiveStr::Ascii("fluctuaties"),
                &["fluctuate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolists"),
                &["capitalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovserver"), &["observer"]),
            (dictgen::InsensitiveStr::Ascii("seccond"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("gigibit"), &["gigabit"]),
            (
                dictgen::InsensitiveStr::Ascii("expessions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffenet"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("linguisitc"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("acctivates"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("declaraters"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strems"),
                &["streams", "stems", "strums"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reportd"),
                &["reported", "reports"],
            ),
            (dictgen::InsensitiveStr::Ascii("paristan"), &["partisan"]),
            (dictgen::InsensitiveStr::Ascii("costumize"), &["customize"]),
            (
                dictgen::InsensitiveStr::Ascii("fullscrean"),
                &["fullscreen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggergates"),
                &["aggregates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahin"), &["chain", "cabin"]),
            (
                dictgen::InsensitiveStr::Ascii("resursions"),
                &["recursions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficitifed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appareance"),
                &["appearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiaists"),
                &["enthusiasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluetion"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("avalible"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("noticabley"),
                &["noticeably"],
            ),
            (dictgen::InsensitiveStr::Ascii("soparnos"), &["sopranos"]),
            (dictgen::InsensitiveStr::Ascii("lonliness"), &["loneliness"]),
            (
                dictgen::InsensitiveStr::Ascii("acchievements"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("glodfish"), &["goldfish"]),
            (
                dictgen::InsensitiveStr::Ascii("classificaion"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("anialate"), &["annihilate"]),
            (
                dictgen::InsensitiveStr::Ascii("compromissen"),
                &["compromise"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapened"), &["happened"]),
            (
                dictgen::InsensitiveStr::Ascii("proecessing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("paremter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("collapese"), &["collapse"]),
            (dictgen::InsensitiveStr::Ascii("sufaces"), &["surfaces"]),
            (
                dictgen::InsensitiveStr::Ascii("survice"),
                &["service", "survive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrophize"),
                &["catastrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facination"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("posisble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("thouroughly"),
                &["thoroughly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomticly"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cashe"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("utput"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("verificaitons"),
                &["verifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminsitrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("nombers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("integraitons"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("lincenses"), &["licenses"]),
            (dictgen::InsensitiveStr::Ascii("detecions"), &["detections"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboor"),
                &["neighbor", "neighbour"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspndences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrespondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaite"), &["locate"]),
            (dictgen::InsensitiveStr::Ascii("intiails"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("regualarly"), &["regularly"]),
            (
                dictgen::InsensitiveStr::Ascii("informatoins"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("passowords"), &["passwords"]),
            (dictgen::InsensitiveStr::Ascii("musuem"), &["museum"]),
            (
                dictgen::InsensitiveStr::Ascii("conspiricies"),
                &["conspiracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invitacion"),
                &["invitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summersalt"),
                &["somersault"],
            ),
            (dictgen::InsensitiveStr::Ascii("fourties"), &["forties"]),
            (
                dictgen::InsensitiveStr::Ascii("sychroniser"),
                &["synchroniser"],
            ),
            (dictgen::InsensitiveStr::Ascii("bufferred"), &["buffered"]),
            (
                dictgen::InsensitiveStr::Ascii("sucesfuly"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unorginial"),
                &["unoriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acticating"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("probleem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("backtic"), &["backtick"]),
            (dictgen::InsensitiveStr::Ascii("nothinh"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("denpends"), &["depends"]),
            (dictgen::InsensitiveStr::Ascii("engouh"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("propulstion"),
                &["propulsion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executibles"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporaching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invulverable"),
                &["invulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("constanct"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("deifning"), &["defining"]),
            (dictgen::InsensitiveStr::Ascii("mainfest"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("datatytpes"), &["datatypes"]),
            (dictgen::InsensitiveStr::Ascii("subsysthem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("conifigured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("costomize"), &["customize"]),
            (
                dictgen::InsensitiveStr::Ascii("customable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwareds"), &["forwards"]),
            (
                dictgen::InsensitiveStr::Ascii("sensationails"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("prerelase"), &["prerelease"]),
            (
                dictgen::InsensitiveStr::Ascii("styed"),
                &["stayed", "styled"],
            ),
            (dictgen::InsensitiveStr::Ascii("vor"), &["for"]),
            (
                dictgen::InsensitiveStr::Ascii("transcorders"),
                &["transcoders"],
            ),
            (dictgen::InsensitiveStr::Ascii("fallbcks"), &["fallbacks"]),
            (dictgen::InsensitiveStr::Ascii("returnes"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("contributo"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lockacreen"),
                &["lockscreen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorothms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("snytheses"), &["syntheses"]),
            (dictgen::InsensitiveStr::Ascii("expectes"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("distniguished"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuppy"), &["occupy"]),
            (
                dictgen::InsensitiveStr::Ascii("lightenting"),
                &["lightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("incomming"), &["incoming"]),
            (
                dictgen::InsensitiveStr::Ascii("descroptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("quries"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("docuemntation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manupilates"),
                &["manipulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiance"),
                &["experience"],
            ),
            (dictgen::InsensitiveStr::Ascii("chemestry"), &["chemistry"]),
            (
                dictgen::InsensitiveStr::Ascii("assyemetrically"),
                &["asymmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("geerics"), &["generics"]),
            (dictgen::InsensitiveStr::Ascii("vallues"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("fonetic"), &["phonetic"]),
            (dictgen::InsensitiveStr::Ascii("cetain"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("cabnet"), &["cabinet"]),
            (dictgen::InsensitiveStr::Ascii("blulets"), &["bullets"]),
            (dictgen::InsensitiveStr::Ascii("eurpoeans"), &["europeans"]),
            (
                dictgen::InsensitiveStr::Ascii("moleculaire"),
                &["molecular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caterpiller"),
                &["caterpillar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recontructions"),
                &["reconstructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("quailified"), &["qualified"]),
            (dictgen::InsensitiveStr::Ascii("apaul"), &["appall"]),
            (dictgen::InsensitiveStr::Ascii("endponits"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("informtional"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("passords"), &["passwords"]),
            (dictgen::InsensitiveStr::Ascii("speciifed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("interrputions"),
                &["interruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganeric"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("emprovement"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulfunction"),
                &["malfunction", "multifunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("abodmen"), &["abdomen"]),
            (
                dictgen::InsensitiveStr::Ascii("arbritations"),
                &["arbitrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("connin"), &["conning"]),
            (dictgen::InsensitiveStr::Ascii("fliping"), &["flipping"]),
            (dictgen::InsensitiveStr::Ascii("cylce"), &["cycle"]),
            (dictgen::InsensitiveStr::Ascii("galllery"), &["gallery"]),
            (
                dictgen::InsensitiveStr::Ascii("circunvents"),
                &["circumvents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circunventing"),
                &["circumventing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ampitude"), &["amplitude"]),
            (dictgen::InsensitiveStr::Ascii("pinncale"), &["pinnacle"]),
            (dictgen::InsensitiveStr::Ascii("instaler"), &["installer"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowledgin"),
                &["acknowledging"],
            ),
            (dictgen::InsensitiveStr::Ascii("coyotees"), &["coyotes"]),
            (
                dictgen::InsensitiveStr::Ascii("referenses"),
                &["referees", "references"],
            ),
            (dictgen::InsensitiveStr::Ascii("plannign"), &["planning"]),
            (dictgen::InsensitiveStr::Ascii("tkaes"), &["takes"]),
            (
                dictgen::InsensitiveStr::Ascii("regeressing"),
                &["regressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynamcly"), &["dynamically"]),
            (dictgen::InsensitiveStr::Ascii("androiders"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("prefferably"),
                &["preferably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoitates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underlyign"),
                &["underlying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sematically"),
                &["semantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspecifiecd"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporximately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("pharmsci"), &["pharmacist"]),
            (
                dictgen::InsensitiveStr::Ascii("threated"),
                &["threaded", "threatened", "treated"],
            ),
            (dictgen::InsensitiveStr::Ascii("messuring"), &["measuring"]),
            (
                dictgen::InsensitiveStr::Ascii("constrating"),
                &["constraint", "contrasting", "constraining"],
            ),
            (dictgen::InsensitiveStr::Ascii("thnakyou"), &["thankyou"]),
            (dictgen::InsensitiveStr::Ascii("classifed"), &["classified"]),
            (
                dictgen::InsensitiveStr::Ascii("reattache"),
                &["reattach", "reattached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypertrophey"),
                &["hypertrophy"],
            ),
            (dictgen::InsensitiveStr::Ascii("redered"), &["rendered"]),
            (
                dictgen::InsensitiveStr::Ascii("redistirbute"),
                &["redistribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacwards"), &["backwards"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnsporter"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavours"), &["behaviours"]),
            (dictgen::InsensitiveStr::Ascii("oeprate"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("myselfes"), &["myself"]),
            (dictgen::InsensitiveStr::Ascii("accumalte"), &["accumulate"]),
            (
                dictgen::InsensitiveStr::Ascii("strengtening"),
                &["strengthening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissonence"),
                &["dissonance"],
            ),
            (dictgen::InsensitiveStr::Ascii("incremeant"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("compilacion"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("playble"), &["playable"]),
            (
                dictgen::InsensitiveStr::Ascii("succussful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("excetutes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("autheticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("decsions"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("squirrl"), &["squirrel"]),
            (dictgen::InsensitiveStr::Ascii("prefex"), &["prefix"]),
            (
                dictgen::InsensitiveStr::Ascii("presenations"),
                &["presentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("detatils"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("fildered"), &["filtered"]),
            (dictgen::InsensitiveStr::Ascii("upploaders"), &["uploaders"]),
            (
                dictgen::InsensitiveStr::Ascii("octives"),
                &["actives", "octaves"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreeable"), &["agreeable"]),
            (
                dictgen::InsensitiveStr::Ascii("expextedly"),
                &["expectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("oeprated"), &["operated"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithimical"),
                &["algorithmic", "algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("funides"), &["fundies"]),
            (
                dictgen::InsensitiveStr::Ascii("enumertion"),
                &["enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condolensces"),
                &["condolences"],
            ),
            (dictgen::InsensitiveStr::Ascii("vairable"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("fertalizer"),
                &["fertilizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("evey"), &["every"]),
            (dictgen::InsensitiveStr::Ascii("renmants"), &["remnants"]),
            (
                dictgen::InsensitiveStr::Ascii("formidabile"),
                &["formidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mozzorella"),
                &["mozzarella"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplpied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("hepers"), &["helpers"]),
            (
                dictgen::InsensitiveStr::Ascii("superseedes"),
                &["supersedes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sute"),
                &["site", "suite", "suit"],
            ),
            (dictgen::InsensitiveStr::Ascii("becauase"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("lamdas"), &["lambdas"]),
            (dictgen::InsensitiveStr::Ascii("practicess"), &["practice"]),
            (
                dictgen::InsensitiveStr::Ascii("provids"),
                &["provides", "proves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("butterflyes"),
                &["butterflies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconetions"),
                &["disconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("suvh"), &["such"]),
            (
                dictgen::InsensitiveStr::Ascii("impleneting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instutitional"),
                &["institutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("businesss"), &["businesses"]),
            (dictgen::InsensitiveStr::Ascii("refresing"), &["refreshing"]),
            (dictgen::InsensitiveStr::Ascii("rotat"), &["rotate"]),
            (
                dictgen::InsensitiveStr::Ascii("boundardies"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("comminist"), &["communist"]),
            (
                dictgen::InsensitiveStr::Ascii("crytographic"),
                &["cryptographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("functoins"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("uncoment"), &["uncomment"]),
            (dictgen::InsensitiveStr::Ascii("envolope"), &["envelope"]),
            (dictgen::InsensitiveStr::Ascii("pyrhon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("yomen"), &["yeomen"]),
            (dictgen::InsensitiveStr::Ascii("crashaes"), &["crashes"]),
            (dictgen::InsensitiveStr::Ascii("ceation"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("comparisen"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("rehabilitaiton"),
                &["rehabilitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiteir"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("paitence"), &["patience"]),
            (
                dictgen::InsensitiveStr::Ascii("orginiates"),
                &["originates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ststion"), &["station"]),
            (dictgen::InsensitiveStr::Ascii("obvioulsy"), &["obviously"]),
            (dictgen::InsensitiveStr::Ascii("angziety"), &["anxiety"]),
            (
                dictgen::InsensitiveStr::Ascii("proteccion"),
                &["protection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typecheking"),
                &["typechecking"],
            ),
            (dictgen::InsensitiveStr::Ascii("surgey"), &["surgery"]),
            (
                dictgen::InsensitiveStr::Ascii("encrpytion"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("encyrpt"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("purchasin"), &["purchasing"]),
            (
                dictgen::InsensitiveStr::Ascii("assosiated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("announcin"), &["announcing"]),
            (
                dictgen::InsensitiveStr::Ascii("callengers"),
                &["challengers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleanded"),
                &["cleaned", "cleansed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contancting"),
                &["contracting", "contacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discontined"),
                &["discontinued"],
            ),
            (dictgen::InsensitiveStr::Ascii("coulored"), &["colored"]),
            (dictgen::InsensitiveStr::Ascii("rountines"), &["routines"]),
            (
                dictgen::InsensitiveStr::Ascii("renegothiated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("underwaer"), &["underwear"]),
            (
                dictgen::InsensitiveStr::Ascii("cantaining"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("membeor"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("hemisperes"),
                &["hemispheres"],
            ),
            (dictgen::InsensitiveStr::Ascii("deffaulted"), &["defaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("standrdisation"),
                &["standardisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("edfault"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("herione"), &["heroine"]),
            (dictgen::InsensitiveStr::Ascii("tanlged"), &["tangled"]),
            (
                dictgen::InsensitiveStr::Ascii("substitutos"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expewriments"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("poducible"), &["producible"]),
            (dictgen::InsensitiveStr::Ascii("warining"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("commont"),
                &["common", "comment"],
            ),
            (dictgen::InsensitiveStr::Ascii("timetstamp"), &["timestamp"]),
            (
                dictgen::InsensitiveStr::Ascii("verifie"),
                &["verify", "verified"],
            ),
            (dictgen::InsensitiveStr::Ascii("estmated"), &["estimated"]),
            (
                dictgen::InsensitiveStr::Ascii("principias"),
                &["principals"],
            ),
            (dictgen::InsensitiveStr::Ascii("refleted"), &["reflected"]),
            (
                dictgen::InsensitiveStr::Ascii("uesfulness"),
                &["usefulness"],
            ),
            (dictgen::InsensitiveStr::Ascii("blaimed"), &["blamed"]),
            (
                dictgen::InsensitiveStr::Ascii("amendmenter"),
                &["amendment"],
            ),
            (dictgen::InsensitiveStr::Ascii("entroy"), &["entropy"]),
            (
                dictgen::InsensitiveStr::Ascii("referening"),
                &["referencing", "refereeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeust"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("inspriational"),
                &["inspirational"],
            ),
            (dictgen::InsensitiveStr::Ascii("misspeled"), &["misspelled"]),
            (dictgen::InsensitiveStr::Ascii("fornats"), &["formats"]),
            (dictgen::InsensitiveStr::Ascii("acceeptor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("menbers"), &["members"]),
            (
                dictgen::InsensitiveStr::Ascii("supermacist"),
                &["supremacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remoes"),
                &["removes", "remotes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertania"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categogical"),
                &["categorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origianlly"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("appart"), &["apart"]),
            (dictgen::InsensitiveStr::Ascii("relizes"), &["realizes"]),
            (dictgen::InsensitiveStr::Ascii("markkup"), &["markup"]),
            (dictgen::InsensitiveStr::Ascii("virutal"), &["virtual"]),
            (dictgen::InsensitiveStr::Ascii("obsolte"), &["obsolete"]),
            (
                dictgen::InsensitiveStr::Ascii("settlemetns"),
                &["settlements"],
            ),
            (dictgen::InsensitiveStr::Ascii("debgger"), &["debugger"]),
            (
                dictgen::InsensitiveStr::Ascii("desingation"),
                &["designation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formattings"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("studdy"), &["study"]),
            (dictgen::InsensitiveStr::Ascii("templeted"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("meixcans"), &["mexicans"]),
            (
                dictgen::InsensitiveStr::Ascii("reconfingure"),
                &["reconfigure"],
            ),
            (dictgen::InsensitiveStr::Ascii("assotiate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("bilzzard"), &["blizzard"]),
            (dictgen::InsensitiveStr::Ascii("insomina"), &["insomnia"]),
            (
                dictgen::InsensitiveStr::Ascii("identtifiable"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("rrrored"), &["errored"]),
            (
                dictgen::InsensitiveStr::Ascii("apostrophied"),
                &["apostrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("happpening"), &["happening"]),
            (dictgen::InsensitiveStr::Ascii("declarded"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("suplants"), &["supplants"]),
            (dictgen::InsensitiveStr::Ascii("genereate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("academis"),
                &["academics", "academies"],
            ),
            (dictgen::InsensitiveStr::Ascii("assitsed"), &["assisted"]),
            (
                dictgen::InsensitiveStr::Ascii("declerators"),
                &["declarators", "decelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("exclusie"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("umemployment"),
                &["unemployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("hipsterest"), &["hipsters"]),
            (dictgen::InsensitiveStr::Ascii("fewg"), &["few", "fugue"]),
            (
                dictgen::InsensitiveStr::Ascii("screebshot"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("edige"), &["edge"]),
            (dictgen::InsensitiveStr::Ascii("instnce"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("rinosarus"), &["rhinoceros"]),
            (dictgen::InsensitiveStr::Ascii("detctors"), &["detectors"]),
            (
                dictgen::InsensitiveStr::Ascii("segrigation"),
                &["segregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucceeded"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compositied"),
                &["composite"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectts"), &["connects"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosynchrasy"),
                &["idiosyncrasy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depretiation"),
                &["depreciation", "deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("saskatchewinian"),
                &["saskatchewan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conficts"),
                &["conflicts", "convicts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("divisbility"),
                &["divisibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("prgram"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("afficianados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("retigger"), &["retrigger"]),
            (dictgen::InsensitiveStr::Ascii("amost"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("transparcency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ontent"),
                &["content", "intent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrtups"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palestininans"),
                &["palestinians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconvencional"),
                &["unconventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("chracters"), &["characters"]),
            (dictgen::InsensitiveStr::Ascii("grandize"), &["aggrandize"]),
            (
                dictgen::InsensitiveStr::Ascii("tranferring"),
                &["transferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("assoicate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("satandard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("languaje"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("electroate"),
                &["electorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificiations"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("searchg"),
                &["search", "searching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("establishement"),
                &["establishment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sugguestion"),
                &["suggestion"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependet"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("inclueds"), &["includes"]),
            (dictgen::InsensitiveStr::Ascii("schmes"), &["schemes"]),
            (dictgen::InsensitiveStr::Ascii("pijeons"), &["pigeons"]),
            (dictgen::InsensitiveStr::Ascii("labryinth"), &["labyrinth"]),
            (
                dictgen::InsensitiveStr::Ascii("presidenital"),
                &["presidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("recevier"), &["receiver"]),
            (dictgen::InsensitiveStr::Ascii("peirod"), &["period"]),
            (
                dictgen::InsensitiveStr::Ascii("dupplicated"),
                &["duplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstarte"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("headphoens"),
                &["headphones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenitcate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wisnton"), &["winston"]),
            (
                dictgen::InsensitiveStr::Ascii("inializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("addreses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("sovereighty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanetized"), &["sanitized"]),
            (dictgen::InsensitiveStr::Ascii("vriables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("impemented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("perimetr"), &["perimeter"]),
            (dictgen::InsensitiveStr::Ascii("acount"), &["account"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("capicator"), &["capacitor"]),
            (
                dictgen::InsensitiveStr::Ascii("tablespons"),
                &["tablespoons"],
            ),
            (dictgen::InsensitiveStr::Ascii("usuing"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("exlcusively"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("corparate"), &["corporate"]),
            (
                dictgen::InsensitiveStr::Ascii("dscriptors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significatly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("carange"), &["carnage"]),
            (dictgen::InsensitiveStr::Ascii("endrose"), &["endorse"]),
            (
                dictgen::InsensitiveStr::Ascii("visuabized"),
                &["visualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("genearate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("dungeoun"), &["dungeon"]),
            (
                dictgen::InsensitiveStr::Ascii("aggressivley"),
                &["aggressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("dubios"), &["dubious"]),
            (
                dictgen::InsensitiveStr::Ascii("destroyeds"),
                &["destroyers"],
            ),
            (dictgen::InsensitiveStr::Ascii("couting"), &["counting"]),
            (dictgen::InsensitiveStr::Ascii("listned"), &["listened"]),
            (
                dictgen::InsensitiveStr::Ascii("proffessional"),
                &["professional"],
            ),
            (dictgen::InsensitiveStr::Ascii("peolpe"), &["people"]),
            (
                dictgen::InsensitiveStr::Ascii("refererences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("creats"), &["creates"]),
            (dictgen::InsensitiveStr::Ascii("accepptors"), &["acceptors"]),
            (dictgen::InsensitiveStr::Ascii("synposis"), &["synopsis"]),
            (dictgen::InsensitiveStr::Ascii("howeever"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("eleminated"),
                &["eliminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongnizing"),
                &["recognizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cafine"), &["caffeine"]),
            (dictgen::InsensitiveStr::Ascii("hadled"), &["handled"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhegan"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsnking"), &["thanking"]),
            (dictgen::InsensitiveStr::Ascii("decreace"), &["decrease"]),
            (
                dictgen::InsensitiveStr::Ascii("reconcilin"),
                &["reconciling"],
            ),
            (dictgen::InsensitiveStr::Ascii("unannomus"), &["unanimous"]),
            (dictgen::InsensitiveStr::Ascii("juveline"), &["juvenile"]),
            (
                dictgen::InsensitiveStr::Ascii("cosolation"),
                &["consolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arudino"), &["arduino"]),
            (
                dictgen::InsensitiveStr::Ascii("uncondtional"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("dictionry"), &["dictionary"]),
            (dictgen::InsensitiveStr::Ascii("archvers"), &["archivers"]),
            (dictgen::InsensitiveStr::Ascii("bunbler"), &["bundler"]),
            (
                dictgen::InsensitiveStr::Ascii("disconnectes"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmplary"), &["exemplary"]),
            (
                dictgen::InsensitiveStr::Ascii("framlayout"),
                &["framelayout"],
            ),
            (dictgen::InsensitiveStr::Ascii("uniformes"), &["uniforms"]),
            (dictgen::InsensitiveStr::Ascii("atttempted"), &["attempted"]),
            (
                dictgen::InsensitiveStr::Ascii("cantalopes"),
                &["cantaloupes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confoguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inventivos"),
                &["inventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convering"),
                &["converting", "covering", "converging", "conveying"],
            ),
            (dictgen::InsensitiveStr::Ascii("militiants"), &["militias"]),
            (dictgen::InsensitiveStr::Ascii("spearate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("synoyms"), &["synonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("absintence"),
                &["abstinence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nucleous"),
                &["nucleus", "nucleolus"],
            ),
            (dictgen::InsensitiveStr::Ascii("processos"), &["processors"]),
            (
                dictgen::InsensitiveStr::Ascii("asymettric"),
                &["asymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("esssential"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("computtaions"),
                &["computations"],
            ),
            (dictgen::InsensitiveStr::Ascii("quicly"), &["quickly"]),
            (dictgen::InsensitiveStr::Ascii("winnnig"), &["winning"]),
            (dictgen::InsensitiveStr::Ascii("condemend"), &["condemned"]),
            (dictgen::InsensitiveStr::Ascii("trigged"), &["triggered"]),
            (dictgen::InsensitiveStr::Ascii("allso"), &["also"]),
            (
                dictgen::InsensitiveStr::Ascii("environemntal"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripition"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("calssical"), &["classical"]),
            (dictgen::InsensitiveStr::Ascii("untill"), &["until"]),
            (dictgen::InsensitiveStr::Ascii("exectued"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("destiantion"),
                &["destinations", "destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("endianes"), &["endianness"]),
            (dictgen::InsensitiveStr::Ascii("reatil"), &["retail"]),
            (dictgen::InsensitiveStr::Ascii("collpse"), &["collapse"]),
            (
                dictgen::InsensitiveStr::Ascii("distriution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("detcts"), &["detects"]),
            (
                dictgen::InsensitiveStr::Ascii("informatike"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publsihers"),
                &["publishers"],
            ),
            (dictgen::InsensitiveStr::Ascii("recievd"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("sequencies"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("commponent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("monumentaal"),
                &["monumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("taged"), &["tagged"]),
            (dictgen::InsensitiveStr::Ascii("posistion"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("fethces"), &["fetches"]),
            (dictgen::InsensitiveStr::Ascii("dispalyes"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("soultions"), &["solutions"]),
            (dictgen::InsensitiveStr::Ascii("proocol"), &["protocol"]),
            (
                dictgen::InsensitiveStr::Ascii("machanically"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straightward"),
                &["straightwards"],
            ),
            (dictgen::InsensitiveStr::Ascii("absenses"), &["absences"]),
            (dictgen::InsensitiveStr::Ascii("aucitons"), &["auctions"]),
            (
                dictgen::InsensitiveStr::Ascii("infestaion"),
                &["infestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("resolvs"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("lifesystem"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtitecture"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("polcy"), &["policy"]),
            (
                dictgen::InsensitiveStr::Ascii("accend"),
                &["ascend", "accent"],
            ),
            (dictgen::InsensitiveStr::Ascii("damange"), &["damage"]),
            (dictgen::InsensitiveStr::Ascii("allocaing"), &["allocating"]),
            (
                dictgen::InsensitiveStr::Ascii("visted"),
                &["visited", "listed", "vested"],
            ),
            (dictgen::InsensitiveStr::Ascii("trubles"), &["troubles"]),
            (dictgen::InsensitiveStr::Ascii("registerss"), &["registers"]),
            (dictgen::InsensitiveStr::Ascii("crusdaers"), &["crusaders"]),
            (
                dictgen::InsensitiveStr::Ascii("spefification"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("gaus"), &["gauss", "gauze"]),
            (
                dictgen::InsensitiveStr::Ascii("redner"),
                &["render", "redder"],
            ),
            (dictgen::InsensitiveStr::Ascii("faciltiy"), &["facility"]),
            (dictgen::InsensitiveStr::Ascii("refreshd"), &["refreshed"]),
            (
                dictgen::InsensitiveStr::Ascii("nastershums"),
                &["nasturtiums"],
            ),
            (dictgen::InsensitiveStr::Ascii("duleing"), &["dueling"]),
            (dictgen::InsensitiveStr::Ascii("labbeled"), &["labeled"]),
            (
                dictgen::InsensitiveStr::Ascii("ineritance"),
                &["inheritance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjuste"),
                &["adjust", "adjusted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implimcitly"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("movemement"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("reminiscin"),
                &["reminiscing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimziing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("operares"), &["operates"]),
            (dictgen::InsensitiveStr::Ascii("platnium"), &["platinum"]),
            (
                dictgen::InsensitiveStr::Ascii("commatas"),
                &["commas", "commata"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostitutie"),
                &["prostitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pascheurizing"),
                &["pasteurizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mercenares"),
                &["mercenaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("anouncing"), &["announcing"]),
            (
                dictgen::InsensitiveStr::Ascii("visiblility"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("katemine"), &["ketamine"]),
            (
                dictgen::InsensitiveStr::Ascii("deode"),
                &["decode", "diode"],
            ),
            (dictgen::InsensitiveStr::Ascii("routins"), &["routines"]),
            (dictgen::InsensitiveStr::Ascii("behaivor"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("previes"), &["previews"]),
            (
                dictgen::InsensitiveStr::Ascii("termianlly"),
                &["terminally"],
            ),
            (dictgen::InsensitiveStr::Ascii("geration"), &["generation"]),
            (dictgen::InsensitiveStr::Ascii("reltive"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("cimmetricaly"),
                &["symmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbode"), &["verbose"]),
            (
                dictgen::InsensitiveStr::Ascii("thermostasts"),
                &["thermostats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espects"),
                &["expects", "aspects", "respects"],
            ),
            (dictgen::InsensitiveStr::Ascii("thirteeen"), &["thirteen"]),
            (dictgen::InsensitiveStr::Ascii("praefix"), &["prefix"]),
            (dictgen::InsensitiveStr::Ascii("frequetly"), &["frequently"]),
            (
                dictgen::InsensitiveStr::Ascii("problamatic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequenty"),
                &["frequently", "frequent"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfier"), &["verifier"]),
            (
                dictgen::InsensitiveStr::Ascii("photochopped"),
                &["photoshopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prestiguous"),
                &["prestigious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermiment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condradicted"),
                &["contradicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calidation"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("canoncial"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("algorhithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("soundtrakcs"),
                &["soundtracks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifiested"),
                &["manifested"],
            ),
            (dictgen::InsensitiveStr::Ascii("autor"), &["author"]),
            (dictgen::InsensitiveStr::Ascii("hansome"), &["handsome"]),
            (
                dictgen::InsensitiveStr::Ascii("hyopthetical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcor"), &["director"]),
            (
                dictgen::InsensitiveStr::Ascii("emotionaly"),
                &["emotionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streaching"),
                &["stretching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dramaticaly"),
                &["dramatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("koncentration"),
                &["concentration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveinently"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particiaption"),
                &["participation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardbaord"), &["cardboard"]),
            (dictgen::InsensitiveStr::Ascii("roducts"), &["products"]),
            (dictgen::InsensitiveStr::Ascii("downlond"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("progessbars"),
                &["progressbars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disastrose"),
                &["disastrous"],
            ),
            (dictgen::InsensitiveStr::Ascii("etherent"), &["ethernet"]),
            (
                dictgen::InsensitiveStr::Ascii("signging"),
                &["signing", "singing"],
            ),
            (dictgen::InsensitiveStr::Ascii("phatnom"), &["phantom"]),
            (dictgen::InsensitiveStr::Ascii("enfornced"), &["enforced"]),
            (dictgen::InsensitiveStr::Ascii("aatribute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotited"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("revoce"), &["revoke"]),
            (
                dictgen::InsensitiveStr::Ascii("desintation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprocuced"),
                &["reproduced", "reprocured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availaliability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("westmonster"),
                &["westminster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlightin"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrctions"),
                &["constructions", "constrictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("diplomma"), &["diploma"]),
            (
                dictgen::InsensitiveStr::Ascii("transgenres"),
                &["transgender"],
            ),
            (dictgen::InsensitiveStr::Ascii("momentus"), &["moments"]),
            (dictgen::InsensitiveStr::Ascii("theramin"), &["theremin"]),
            (dictgen::InsensitiveStr::Ascii("loacated"), &["located"]),
            (dictgen::InsensitiveStr::Ascii("portrary"), &["portray"]),
            (
                dictgen::InsensitiveStr::Ascii("designacion"),
                &["designation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootsrapper"),
                &["bootstrapper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convienance"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individially"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependices"),
                &["dependencies", "dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trancoders"),
                &["transcoders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconeting"),
                &["disconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreivation"),
                &["abbreviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("xenophibic"),
                &["xenophobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contructors"),
                &["contractors", "constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scarching"),
                &["searching", "scorching", "scratching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooefficients"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("unworthly"), &["unworthy"]),
            (dictgen::InsensitiveStr::Ascii("accpeter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("abolustely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("negotaiate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("aftterwards"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparasions"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("intenally"), &["internally"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowldging"),
                &["acknowledging"],
            ),
            (dictgen::InsensitiveStr::Ascii("estimateor"), &["estimator"]),
            (
                dictgen::InsensitiveStr::Ascii("familairity"),
                &["familiarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("chemsitry"), &["chemistry"]),
            (
                dictgen::InsensitiveStr::Ascii("judgemenal"),
                &["judgemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorizeed"),
                &["authorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("keysenian"), &["keynesian"]),
            (
                dictgen::InsensitiveStr::Ascii("paralelising"),
                &["parallelising"],
            ),
            (dictgen::InsensitiveStr::Ascii("johhny"), &["johnny"]),
            (
                dictgen::InsensitiveStr::Ascii("leaglizing"),
                &["legalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("workench"), &["workbench"]),
            (dictgen::InsensitiveStr::Ascii("chiledren"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("personis"), &["persons"]),
            (dictgen::InsensitiveStr::Ascii("mercurey"), &["mercury"]),
            (
                dictgen::InsensitiveStr::Ascii("instroducing"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicationb"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permormances"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volenteered"),
                &["volunteered"],
            ),
            (dictgen::InsensitiveStr::Ascii("aafter"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("votlage"), &["voltage"]),
            (dictgen::InsensitiveStr::Ascii("inititials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("imprefection"),
                &["imperfections"],
            ),
            (dictgen::InsensitiveStr::Ascii("addressess"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("advicing"), &["advising"]),
            (dictgen::InsensitiveStr::Ascii("framw"), &["frame"]),
            (dictgen::InsensitiveStr::Ascii("oredering"), &["ordering"]),
            (dictgen::InsensitiveStr::Ascii("poissble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("disapperaing"),
                &["disappearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differentes"),
                &["differences", "difference", "different"],
            ),
            (dictgen::InsensitiveStr::Ascii("consuminng"), &["consuming"]),
            (dictgen::InsensitiveStr::Ascii("resolvendo"), &["resolved"]),
            (dictgen::InsensitiveStr::Ascii("referrin"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("trasports"), &["transports"]),
            (
                dictgen::InsensitiveStr::Ascii("reocmpression"),
                &["recompression"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("possobility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extenisons"),
                &["extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("thereaded"), &["threaded"]),
            (
                dictgen::InsensitiveStr::Ascii("tokenziation"),
                &["tokenization"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanaguages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("effectiviness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervation"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adequitely"),
                &["adequately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cofigurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("concludin"), &["concluding"]),
            (dictgen::InsensitiveStr::Ascii("feek"), &["feel"]),
            (dictgen::InsensitiveStr::Ascii("acccesses"), &["accesses"]),
            (dictgen::InsensitiveStr::Ascii("beavior"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("intermidiate"),
                &["intermediate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("veryfications"),
                &["verifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfonctional"),
                &["dysfunctional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fingertits"),
                &["fingertips"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motorcicles"),
                &["motorcycles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firendlies"),
                &["friendlies"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumbmitted"), &["submitted"]),
            (
                dictgen::InsensitiveStr::Ascii("efficiantly"),
                &["efficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacables"),
                &["replaceables"],
            ),
            (dictgen::InsensitiveStr::Ascii("clicnial"), &["clinical"]),
            (
                dictgen::InsensitiveStr::Ascii("addminister"),
                &["administer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entoxication"),
                &["intoxication"],
            ),
            (dictgen::InsensitiveStr::Ascii("succseeds"), &["succeeds"]),
            (
                dictgen::InsensitiveStr::Ascii("pressureing"),
                &["pressuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistance"),
                &["consistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("unusaully"), &["unusually"]),
            (dictgen::InsensitiveStr::Ascii("libraary"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("assumpted"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("defacation"),
                &["defamation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fileystems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpsected"), &["inspected"]),
            (dictgen::InsensitiveStr::Ascii("redneckest"), &["rednecks"]),
            (dictgen::InsensitiveStr::Ascii("vibratr"), &["vibrator"]),
            (
                dictgen::InsensitiveStr::Ascii("expanisons"),
                &["expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concecutively"),
                &["consecutively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detrimential"),
                &["detrimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitavely"),
                &["competitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvemenet"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("sended"), &["sent"]),
            (
                dictgen::InsensitiveStr::Ascii("achived"),
                &["achieved", "archived"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wollow"),
                &["follow", "wallow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("censibility"),
                &["sensibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comformation"),
                &["conformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polically"),
                &["politically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particularily"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("glichtes"), &["glitches"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditinally"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intensionaly"),
                &["intentionally", "intensionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constains"),
                &["constrains", "contains"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaisn"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("referemdum"),
                &["referendum"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinamic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("behavoirs"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("thight"),
                &["tight", "thigh", "fight"],
            ),
            (dictgen::InsensitiveStr::Ascii("novemer"), &["november"]),
            (dictgen::InsensitiveStr::Ascii("dependded"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("tookits"), &["toolkits"]),
            (dictgen::InsensitiveStr::Ascii("espaces"), &["escapes"]),
            (
                dictgen::InsensitiveStr::Ascii("definjtely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbsted"),
                &["masturbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converterd"),
                &["converted", "converter"],
            ),
            (dictgen::InsensitiveStr::Ascii("testifiy"), &["testify"]),
            (
                dictgen::InsensitiveStr::Ascii("dispositon"),
                &["disposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("spetsific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("detecte"),
                &["detected", "detect", "detects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("menion"),
                &["mention", "minion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technitian"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomendatoins"),
                &["recommendations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesction"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("covereing"), &["covering"]),
            (
                dictgen::InsensitiveStr::Ascii("descripton"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neiborhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restoreble"),
                &["restorable"],
            ),
            (dictgen::InsensitiveStr::Ascii("becauses"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("outperfom"), &["outperform"]),
            (dictgen::InsensitiveStr::Ascii("lincolon"), &["lincoln"]),
            (
                dictgen::InsensitiveStr::Ascii("assembladge"),
                &["assemblage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefcifiying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shcizophrenic"),
                &["schizophrenic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trotzkyites"),
                &["trotskyists"],
            ),
            (dictgen::InsensitiveStr::Ascii("resovles"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegogtiation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("serieses"), &["series"]),
            (dictgen::InsensitiveStr::Ascii("hypocrasy"), &["hypocrisy"]),
            (dictgen::InsensitiveStr::Ascii("gauntles"), &["gauntlets"]),
            (dictgen::InsensitiveStr::Ascii("calsses"), &["classes"]),
            (
                dictgen::InsensitiveStr::Ascii("thermastat"),
                &["thermostat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiugurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("persuaso"), &["persuasion"]),
            (dictgen::InsensitiveStr::Ascii("apretiate"), &["appreciate"]),
            (dictgen::InsensitiveStr::Ascii("perssed"), &["pressed"]),
            (
                dictgen::InsensitiveStr::Ascii("origianally"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("filesstem"), &["filesystem"]),
            (
                dictgen::InsensitiveStr::Ascii("containts"),
                &["contains", "constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hsitorical"),
                &["historical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasported"),
                &["transported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedplication"),
                &["deduplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predeclarnig"),
                &["predeclaring"],
            ),
            (dictgen::InsensitiveStr::Ascii("overral"), &["overall"]),
            (dictgen::InsensitiveStr::Ascii("associdate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("celebrite"),
                &["celebrities"],
            ),
            (dictgen::InsensitiveStr::Ascii("ednpoint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("datatye"), &["datatype"]),
            (dictgen::InsensitiveStr::Ascii("adjcentcy"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("bloek"), &["bloke"]),
            (dictgen::InsensitiveStr::Ascii("aluminimum"), &["aluminium"]),
            (dictgen::InsensitiveStr::Ascii("bloeks"), &["blokes"]),
            (
                dictgen::InsensitiveStr::Ascii("desserialises"),
                &["deserialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("montiors"), &["monitors"]),
            (dictgen::InsensitiveStr::Ascii("versioon"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("transperancy"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purchacing"),
                &["purchasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("outisder"), &["outsider"]),
            (dictgen::InsensitiveStr::Ascii("mosnter"), &["monster"]),
            (
                dictgen::InsensitiveStr::Ascii("flyes"),
                &["flies", "flyers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterprenuers"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antecedant"),
                &["antecedent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mircoservices"),
                &["microservices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibiilities"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammortizes"), &["amortizes"]),
            (
                dictgen::InsensitiveStr::Ascii("acctivation"),
                &["activation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcripting"),
                &["transcribing", "transcription"],
            ),
            (dictgen::InsensitiveStr::Ascii("proeprly"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocatin"),
                &["reallocating", "reallocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deregistrated"),
                &["deregistered"],
            ),
            (dictgen::InsensitiveStr::Ascii("columnts"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("multiprocesing"),
                &["multiprocessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dealilng"), &["dealing"]),
            (
                dictgen::InsensitiveStr::Ascii("perpatrator"),
                &["perpetrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("expemplar"), &["exemplar"]),
            (dictgen::InsensitiveStr::Ascii("subract"), &["subtract"]),
            (dictgen::InsensitiveStr::Ascii("excption"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("technicallly"),
                &["technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requets"),
                &["request", "requests"],
            ),
            (dictgen::InsensitiveStr::Ascii("goverment"), &["government"]),
            (dictgen::InsensitiveStr::Ascii("packadge"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("minimun"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("likewis"), &["likewise"]),
            (
                dictgen::InsensitiveStr::Ascii("parallises"),
                &["parallelises"],
            ),
            (dictgen::InsensitiveStr::Ascii("padedd"), &["padded"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuit"),
                &["contribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommandes"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("socialsim"), &["socialism"]),
            (dictgen::InsensitiveStr::Ascii("topolgy"), &["topology"]),
            (
                dictgen::InsensitiveStr::Ascii("srceenshots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambedded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("errot"), &["error"]),
            (
                dictgen::InsensitiveStr::Ascii("verications"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("opperates"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("calcalations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interects"),
                &["interacts", "intersects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scaleability"),
                &["scalability"],
            ),
            (dictgen::InsensitiveStr::Ascii("clikcbait"), &["clickbait"]),
            (dictgen::InsensitiveStr::Ascii("hrlp"), &["help"]),
            (
                dictgen::InsensitiveStr::Ascii("knowldgeable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sponses"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("reclutantly"),
                &["reluctantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jurisdiktion"),
                &["jurisdiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastrukture"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instalaltion"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("interere"), &["interfere"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratin"),
                &[
                    "configurations",
                    "configuration",
                    "configurating",
                    "configuring",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificato"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redability"),
                &["readability"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressets"), &["resets"]),
            (
                dictgen::InsensitiveStr::Ascii("absolutisme"),
                &["absolutes", "absolutism"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyled"), &["cycled"]),
            (
                dictgen::InsensitiveStr::Ascii("confingures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaruantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("expidition"),
                &["expedition"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasform"), &["transform"]),
            (
                dictgen::InsensitiveStr::Ascii("disssembling"),
                &["disassembling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("futuristisch"),
                &["futuristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("secenes"),
                &["scenes", "secedes"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreiver"), &["retriever"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commodoties"),
                &["commodities"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplacted"), &["duplicated"]),
            (dictgen::InsensitiveStr::Ascii("cipers"), &["ciphers"]),
            (
                dictgen::InsensitiveStr::Ascii("qiestioned"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("arramge"), &["arrange"]),
            (
                dictgen::InsensitiveStr::Ascii("containees"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceps"), &["excepts"]),
            (dictgen::InsensitiveStr::Ascii("markown"), &["markdown"]),
            (
                dictgen::InsensitiveStr::Ascii("prepostorous"),
                &["preposterous"],
            ),
            (dictgen::InsensitiveStr::Ascii("procalim"), &["proclaim"]),
            (
                dictgen::InsensitiveStr::Ascii("incombatibly"),
                &["incompatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indentifed"),
                &["identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistencias"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("woundering"), &["wondering"]),
            (
                dictgen::InsensitiveStr::Ascii("subconsciouly"),
                &["subconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("porperty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("annotion"), &["annotation"]),
            (
                dictgen::InsensitiveStr::Ascii("confict"),
                &["conflict", "convict"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmits"), &["emits"]),
            (
                dictgen::InsensitiveStr::Ascii("republians"),
                &["republicans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuccesfull"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analitycal"),
                &["analytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("thaliand"), &["thailand"]),
            (
                dictgen::InsensitiveStr::Ascii("panding"),
                &["landing", "panting"],
            ),
            (dictgen::InsensitiveStr::Ascii("engagment"), &["engagement"]),
            (
                dictgen::InsensitiveStr::Ascii("competation"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("langague"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("assumung"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasudical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confisgated"),
                &["confiscated"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalation"), &["evaluation"]),
            (
                dictgen::InsensitiveStr::Ascii("deflektion"),
                &["deflection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplinas"),
                &["disciplines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convertable"),
                &["convertible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properitary"),
                &["proprietary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consoltation"),
                &["consultation"],
            ),
            (dictgen::InsensitiveStr::Ascii("peotry"), &["poetry"]),
            (dictgen::InsensitiveStr::Ascii("trancoded"), &["transcoded"]),
            (
                dictgen::InsensitiveStr::Ascii("democracis"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("escapemant"),
                &["escapement"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwin"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("prolitariat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpting"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("supportors"),
                &["supporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathifnder"),
                &["pathfinder"],
            ),
            (dictgen::InsensitiveStr::Ascii("backpacs"), &["backpacks"]),
            (dictgen::InsensitiveStr::Ascii("illegalls"), &["illegals"]),
            (dictgen::InsensitiveStr::Ascii("acepting"), &["accepting"]),
            (dictgen::InsensitiveStr::Ascii("algoithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("treasurin"), &["treasuring"]),
            (
                dictgen::InsensitiveStr::Ascii("sacle"),
                &["scale", "sale", "sable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefinely"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fonctionnality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjection"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persicuted"),
                &["persecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantiating"),
                &["differentiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasmissions"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("domecrat"), &["democrat"]),
            (
                dictgen::InsensitiveStr::Ascii("termintating"),
                &["terminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polysaccaride"),
                &["polysaccharide"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyprto"), &["crypto"]),
            (dictgen::InsensitiveStr::Ascii("gettign"), &["getting"]),
            (
                dictgen::InsensitiveStr::Ascii("valumes"),
                &["volumes", "values"],
            ),
            (dictgen::InsensitiveStr::Ascii("previos"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("edn"), &["end"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithsm"),
                &["algorithms", "algorithm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insignificunt"),
                &["insignificant"],
            ),
            (dictgen::InsensitiveStr::Ascii("uppstream"), &["upstream"]),
            (dictgen::InsensitiveStr::Ascii("retrurns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("sparator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("rovided"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("sharpley"), &["sharply"]),
            (
                dictgen::InsensitiveStr::Ascii("influental"),
                &["influential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interlaves"),
                &["interlaces", "interleaves"],
            ),
            (dictgen::InsensitiveStr::Ascii("referenace"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("trancivers"),
                &["transceivers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoltuions"),
                &["revolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhcange"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("interpertive"),
                &["interpretive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecesarry"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommmitted"),
                &["uncommitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("abilitiy"), &["ability"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotionating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demograpics"),
                &["demographics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fanfcition"),
                &["fanfiction"],
            ),
            (dictgen::InsensitiveStr::Ascii("stummac"), &["stomach"]),
            (dictgen::InsensitiveStr::Ascii("medadata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("develping"), &["developing"]),
            (dictgen::InsensitiveStr::Ascii("rebease"), &["rebase"]),
            (
                dictgen::InsensitiveStr::Ascii("pitmap"),
                &["pixmap", "bitmap"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqeeze"), &["squeeze"]),
            (dictgen::InsensitiveStr::Ascii("devuce"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("prouncements"),
                &["pronouncements"],
            ),
            (dictgen::InsensitiveStr::Ascii("througth"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("cahce"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("whethter"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("alreadu"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("whehter"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("discources"), &["discourse"]),
            (
                dictgen::InsensitiveStr::Ascii("unresgister"),
                &["unregister"],
            ),
            (dictgen::InsensitiveStr::Ascii("parem"), &["param"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiazes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjasancies"),
                &["adjacencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abosrption"),
                &["absorption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("actitivies"),
                &["activities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ovservable"),
                &["observable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installaion"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overkapping"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("navagitation"),
                &["navigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimensin"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("legistative"),
                &["legislative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendances"),
                &["dependances"],
            ),
            (dictgen::InsensitiveStr::Ascii("excusin"), &["excusing"]),
            (
                dictgen::InsensitiveStr::Ascii("diretional"),
                &["directional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsription"),
                &["subscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("anympore"), &["anymore"]),
            (
                dictgen::InsensitiveStr::Ascii("approimations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreggating"),
                &["aggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependendences"),
                &["dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handeled"),
                &["handled", "handheld"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefcifiable"),
                &["specifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("enouhg"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("familiair"), &["familiar"]),
            (dictgen::InsensitiveStr::Ascii("gettng"), &["getting"]),
            (dictgen::InsensitiveStr::Ascii("allcoater"), &["allocator"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhelemed"),
                &["overwhelmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructr"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessecary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("etablished"),
                &["established"],
            ),
            (dictgen::InsensitiveStr::Ascii("dezember"), &["december"]),
            (
                dictgen::InsensitiveStr::Ascii("sumbarines"),
                &["submarines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconditionnally"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerimentally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("scyhter"), &["scyther"]),
            (
                dictgen::InsensitiveStr::Ascii("kindergarden"),
                &["kindergarten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropropriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantasticaly"),
                &["fantastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intidimation"),
                &["intimidation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inacccessible"),
                &["inaccessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("agaonst"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("aloowances"),
                &["allowances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condescensing"),
                &["condescension"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confirmity"),
                &["conformity"],
            ),
            (dictgen::InsensitiveStr::Ascii("rathern"), &["rather"]),
            (
                dictgen::InsensitiveStr::Ascii("culutrally"),
                &["culturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("advertied"), &["advertised"]),
            (dictgen::InsensitiveStr::Ascii("knowling"), &["knowing"]),
            (dictgen::InsensitiveStr::Ascii("puasing"), &["pausing"]),
            (dictgen::InsensitiveStr::Ascii("chamption"), &["champion"]),
            (dictgen::InsensitiveStr::Ascii("genrate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("spoffs"),
                &["spoofs", "spaffs", "scoffs"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitant"), &["militant"]),
            (dictgen::InsensitiveStr::Ascii("aggresive"), &["aggressive"]),
            (dictgen::InsensitiveStr::Ascii("exatly"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("additioanally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoperated"),
                &["incorporated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chocolotes"),
                &["chocolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("threshhold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("analyseas"), &["analyses"]),
            (dictgen::InsensitiveStr::Ascii("durig"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("nightlie"), &["nightlife"]),
            (dictgen::InsensitiveStr::Ascii("accesible"), &["accessible"]),
            (
                dictgen::InsensitiveStr::Ascii("accerelates"),
                &["accelerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("necassary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("catalsyt"), &["catalyst"]),
            (
                dictgen::InsensitiveStr::Ascii("interchangeabley"),
                &["interchangeably"],
            ),
            (dictgen::InsensitiveStr::Ascii("declaire"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("depsawn"), &["despawn"]),
            (
                dictgen::InsensitiveStr::Ascii("addiditional"),
                &["additional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complicare"),
                &["complicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutli"), &["multi"]),
            (
                dictgen::InsensitiveStr::Ascii("schizophrena"),
                &["schizophrenia"],
            ),
            (dictgen::InsensitiveStr::Ascii("halp"), &["help"]),
            (
                dictgen::InsensitiveStr::Ascii("formalizin"),
                &["formalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autogenration"),
                &["autogeneration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontanous"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("translath"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("unscubscribed"),
                &["subscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civalazations"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statestically"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantition"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschopaths"),
                &["psychopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("laguagues"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("warcarft"), &["warcraft"]),
            (
                dictgen::InsensitiveStr::Ascii("dissadvantages"),
                &["disadvantages"],
            ),
            (dictgen::InsensitiveStr::Ascii("weired"), &["weird"]),
            (
                dictgen::InsensitiveStr::Ascii("certifcate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("globabl"), &["global"]),
            (dictgen::InsensitiveStr::Ascii("clsoest"), &["closest"]),
            (
                dictgen::InsensitiveStr::Ascii("imaginatie"),
                &["imaginative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advices"),
                &["advice", "advises"],
            ),
            (dictgen::InsensitiveStr::Ascii("compreser"), &["compressor"]),
            (dictgen::InsensitiveStr::Ascii("cyclinder"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("politcal"), &["political"]),
            (dictgen::InsensitiveStr::Ascii("cyrstal"), &["crystal"]),
            (dictgen::InsensitiveStr::Ascii("tempory"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("prefereed"), &["preferred"]),
            (
                dictgen::InsensitiveStr::Ascii("commene"),
                &["comment", "commend", "commence", "commune"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumtion"),
                &["consumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("asemblies"), &["assemblies"]),
            (dictgen::InsensitiveStr::Ascii("actural"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("unspefixeid"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpetectly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("toursim"), &["tourism"]),
            (
                dictgen::InsensitiveStr::Ascii("certficated"),
                &["certificated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outsouring"),
                &["outsourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathelogical"),
                &["pathological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplinery"),
                &["disciplinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthenticating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("curser"), &["cursor"]),
            (dictgen::InsensitiveStr::Ascii("keyobard"), &["keyboard"]),
            (
                dictgen::InsensitiveStr::Ascii("photogrophers"),
                &["photographers"],
            ),
            (dictgen::InsensitiveStr::Ascii("advatage"), &["advantage"]),
            (dictgen::InsensitiveStr::Ascii("vizual"), &["visual"]),
            (
                dictgen::InsensitiveStr::Ascii("monstorsity"),
                &["monstrosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirlcing"), &["circling"]),
            (dictgen::InsensitiveStr::Ascii("sircular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpt"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("authoized"), &["authorized"]),
            (dictgen::InsensitiveStr::Ascii("blongs"), &["belongs"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitors"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helicpoters"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("particual"), &["particular"]),
            (dictgen::InsensitiveStr::Ascii("argement"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("confiugration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompresors"),
                &["uncompressors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viusalises"),
                &["visualises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosquite"), &["mosquitoes"]),
            (
                dictgen::InsensitiveStr::Ascii("relization"),
                &["realization"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissplay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("touchapd"), &["touchpad"]),
            (
                dictgen::InsensitiveStr::Ascii("representatino"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("keywprds"), &["keywords"]),
            (
                dictgen::InsensitiveStr::Ascii("explorating"),
                &["exploration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amature"),
                &["armature", "amateur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persciuosly"),
                &["preciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasformable"),
                &["transformable"],
            ),
            (dictgen::InsensitiveStr::Ascii("seeen"), &["seen"]),
            (
                dictgen::InsensitiveStr::Ascii("compaitbility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("dentify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("callbacls"), &["callbacks"]),
            (dictgen::InsensitiveStr::Ascii("doscloses"), &["discloses"]),
            (
                dictgen::InsensitiveStr::Ascii("assoziation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("ptyhon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("reportes"), &["reporters"]),
            (dictgen::InsensitiveStr::Ascii("conets"), &["connects"]),
            (
                dictgen::InsensitiveStr::Ascii("cliking"),
                &["clicking", "clinking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pilgrimmage"),
                &["pilgrimage"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandales"), &["sandals"]),
            (
                dictgen::InsensitiveStr::Ascii("compuation"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("entitity"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("esge"), &["edge"]),
            (
                dictgen::InsensitiveStr::Ascii("convencen"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("aproaches"), &["approaches"]),
            (dictgen::InsensitiveStr::Ascii("dogders"), &["dodgers"]),
            (
                dictgen::InsensitiveStr::Ascii("usrers"),
                &["users", "ushers", "usurers"],
            ),
            (dictgen::InsensitiveStr::Ascii("destance"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("directsong"),
                &["directions"],
            ),
            (dictgen::InsensitiveStr::Ascii("stragers"), &["strangers"]),
            (dictgen::InsensitiveStr::Ascii("interrutp"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("perpatuated"),
                &["perpetuated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sergaent"), &["sergeant"]),
            (dictgen::InsensitiveStr::Ascii("queu"), &["queue"]),
            (
                dictgen::InsensitiveStr::Ascii("structiral"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmens"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("withdral"), &["withdrawal"]),
            (
                dictgen::InsensitiveStr::Ascii("distractons"),
                &["distracts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generatign"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("objeect"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillally"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuseable"), &["reusable"]),
            (dictgen::InsensitiveStr::Ascii("verivy"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("calllee"), &["callee"]),
            (dictgen::InsensitiveStr::Ascii("adderssed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("thougts"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("silouetes"),
                &["silhouettes"],
            ),
            (dictgen::InsensitiveStr::Ascii("colleting"), &["collecting"]),
            (dictgen::InsensitiveStr::Ascii("sugest"), &["suggest"]),
            (
                dictgen::InsensitiveStr::Ascii("commectors"),
                &["connectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nitification"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermeneting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("taht"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("sehll"), &["sell", "shell"]),
            (
                dictgen::InsensitiveStr::Ascii("comfirmations"),
                &["confirmations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerelases"),
                &["prereleases"],
            ),
            (dictgen::InsensitiveStr::Ascii("acutal"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("franes"),
                &["frames", "cranes", "francs"],
            ),
            (dictgen::InsensitiveStr::Ascii("infp"), &["info"]),
            (
                dictgen::InsensitiveStr::Ascii("unitilises"),
                &["uninitialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prevailling"),
                &["prevailing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configiured"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleratoring"),
                &["accelerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakwards"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("idented"), &["indented"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulatin"),
                &["manipulation", "manipulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilited"),
                &["stabilized"],
            ),
            (dictgen::InsensitiveStr::Ascii("swither"), &["switcher"]),
            (
                dictgen::InsensitiveStr::Ascii("understatin"),
                &["understating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcdig"),
                &["transcoding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibiliies"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("compaired"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("termintion"),
                &["termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("devce"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("expericene"),
                &["experience"],
            ),
            (dictgen::InsensitiveStr::Ascii("studetns"), &["students"]),
            (dictgen::InsensitiveStr::Ascii("sanpshot"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("vertiece"), &["vertex"]),
            (dictgen::InsensitiveStr::Ascii("apperture"), &["aperture"]),
            (dictgen::InsensitiveStr::Ascii("symbos"), &["symbols"]),
            (
                dictgen::InsensitiveStr::Ascii("algorihtmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustrums"),
                &["frustums", "frusta"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnind"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("penciles"), &["pencils"]),
            (
                dictgen::InsensitiveStr::Ascii("copyritghts"),
                &["copyrights"],
            ),
            (dictgen::InsensitiveStr::Ascii("valentien"), &["valentine"]),
            (
                dictgen::InsensitiveStr::Ascii("connetable"),
                &["connectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("evreytime"), &["everytime"]),
            (dictgen::InsensitiveStr::Ascii("cuplrit"), &["culprit"]),
            (dictgen::InsensitiveStr::Ascii("occurreed"), &["occurred"]),
            (dictgen::InsensitiveStr::Ascii("filaneme"), &["filename"]),
            (dictgen::InsensitiveStr::Ascii("optomizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("emulaters"),
                &["emulators", "emulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustraion"),
                &["illustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("spilts"), &["splits"]),
            (
                dictgen::InsensitiveStr::Ascii("anectotally"),
                &["anecdotally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continouesly"),
                &["continuously"],
            ),
            (dictgen::InsensitiveStr::Ascii("neady"), &["needy"]),
            (
                dictgen::InsensitiveStr::Ascii("aniversary"),
                &["anniversary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confogurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adnimistrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("grahpical"), &["graphical"]),
            (
                dictgen::InsensitiveStr::Ascii("innecesary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("ublished"), &["published"]),
            (
                dictgen::InsensitiveStr::Ascii("steroetypes"),
                &["stereotypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprentation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thruoghput"),
                &["throughput"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritpors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("latice"), &["lattice"]),
            (
                dictgen::InsensitiveStr::Ascii("scroollbars"),
                &["scrollbars"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesulting"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("dubling"),
                &["doubling", "dublin"],
            ),
            (dictgen::InsensitiveStr::Ascii("becomming"), &["becoming"]),
            (
                dictgen::InsensitiveStr::Ascii("sorces"),
                &["sources", "forces", "sores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintaing"),
                &["maintaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnances"), &["chances"]),
            (dictgen::InsensitiveStr::Ascii("prodution"), &["production"]),
            (dictgen::InsensitiveStr::Ascii("viewpiont"), &["viewpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("remenissence"),
                &["reminiscence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restauraunts"),
                &["restaurants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unqualifed"),
                &["unqualified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stattistics"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("buget"), &["budget"]),
            (dictgen::InsensitiveStr::Ascii("avaiblable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("resintalling"),
                &["reinstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("extermes"), &["extremes"]),
            (
                dictgen::InsensitiveStr::Ascii("construrtors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("stornegst"), &["strongest"]),
            (dictgen::InsensitiveStr::Ascii("visulize"), &["visualize"]),
            (
                dictgen::InsensitiveStr::Ascii("compeeting"),
                &["completing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dregee"), &["degree"]),
            (
                dictgen::InsensitiveStr::Ascii("visuabization"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("herad"), &["heard", "hera"]),
            (
                dictgen::InsensitiveStr::Ascii("stcokbrush"),
                &["stockbrush"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tunraround"),
                &["turnaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfieid"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("ytou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("progesss"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("hsitoric"), &["historic"]),
            (
                dictgen::InsensitiveStr::Ascii("milicousally"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechananism"),
                &["mechanism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functitoned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("curcuit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("seperetor"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("subsriber"), &["subscriber"]),
            (dictgen::InsensitiveStr::Ascii("replicaes"), &["replicates"]),
            (
                dictgen::InsensitiveStr::Ascii("counsellng"),
                &["counselling"],
            ),
            (dictgen::InsensitiveStr::Ascii("flattern"), &["flatten"]),
            (dictgen::InsensitiveStr::Ascii("mergging"), &["merging"]),
            (
                dictgen::InsensitiveStr::Ascii("avalibility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifgured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("volumne"), &["volume"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatieve"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytler"), &["styler"]),
            (
                dictgen::InsensitiveStr::Ascii("connonations"),
                &["connotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("midponts"), &["midpoints"]),
            (dictgen::InsensitiveStr::Ascii("gauntelts"), &["gauntlets"]),
            (
                dictgen::InsensitiveStr::Ascii("resusing"),
                &["reusing", "refusing", "resuming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reformattd"),
                &["reformatted"],
            ),
            (dictgen::InsensitiveStr::Ascii("indizies"), &["indices"]),
            (
                dictgen::InsensitiveStr::Ascii("boootstrappers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srew"),
                &["screw", "shrew", "sew"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comporession"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificatins"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("infintely"), &["infinitely"]),
            (
                dictgen::InsensitiveStr::Ascii("dicriminate"),
                &["discriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("turkisch"), &["turkish"]),
            (
                dictgen::InsensitiveStr::Ascii("qaulifiers"),
                &["qualifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apologiste"),
                &["apologise", "apologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomer"), &["performer"]),
            (
                dictgen::InsensitiveStr::Ascii("belives"),
                &["believes", "beliefs"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrats"), &["abstracts"]),
            (dictgen::InsensitiveStr::Ascii("throwgh"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("posiitivity"),
                &["positivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("varations"), &["variations"]),
            (dictgen::InsensitiveStr::Ascii("booleen"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("acrynoms"), &["acronyms"]),
            (
                dictgen::InsensitiveStr::Ascii("microhpones"),
                &["microphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("managably"), &["manageably"]),
            (dictgen::InsensitiveStr::Ascii("alignmnet"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("tabkes"), &["tables"]),
            (
                dictgen::InsensitiveStr::Ascii("accomplissements"),
                &["accomplishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interferring"),
                &["interfering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconnet"), &["reconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("proffesors"),
                &["professors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluacion"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("executting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("conceed"), &["concede"]),
            (dictgen::InsensitiveStr::Ascii("ntegrated"), &["integrated"]),
            (
                dictgen::InsensitiveStr::Ascii("checkcerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("physisist"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("ananlyzed"), &["analyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("primatives"),
                &["primitives"],
            ),
            (dictgen::InsensitiveStr::Ascii("evluator"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("overwites"), &["overwrites"]),
            (dictgen::InsensitiveStr::Ascii("hexgaonal"), &["hexagonal"]),
            (dictgen::InsensitiveStr::Ascii("regually"), &["regularly"]),
            (
                dictgen::InsensitiveStr::Ascii("unarchving"),
                &["unarchiving"],
            ),
            (dictgen::InsensitiveStr::Ascii("capicity"), &["capacity"]),
            (
                dictgen::InsensitiveStr::Ascii("pivate"),
                &["private", "pirate"],
            ),
            (dictgen::InsensitiveStr::Ascii("setis"), &["settees"]),
            (dictgen::InsensitiveStr::Ascii("eddit"), &["edit"]),
            (dictgen::InsensitiveStr::Ascii("parmameter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("intoxicted"),
                &["intoxicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtaulize"),
                &["virtualize"],
            ),
            (dictgen::InsensitiveStr::Ascii("planteary"), &["planetary"]),
            (
                dictgen::InsensitiveStr::Ascii("stimualted"),
                &["stimulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invaldiates"),
                &["invalidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturture"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("benfited"), &["benefited"]),
            (dictgen::InsensitiveStr::Ascii("faggotts"), &["faggots"]),
            (
                dictgen::InsensitiveStr::Ascii("activiations"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("concurer"), &["conqueror"]),
            (
                dictgen::InsensitiveStr::Ascii("itegration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("mehods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("remasterd"), &["remastered"]),
            (dictgen::InsensitiveStr::Ascii("fialogue"), &["dialogue"]),
            (dictgen::InsensitiveStr::Ascii("swearengin"), &["swearing"]),
            (
                dictgen::InsensitiveStr::Ascii("relentlesly"),
                &["relentlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("originnally"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("equps"), &["equips"]),
            (dictgen::InsensitiveStr::Ascii("publsihes"), &["publishes"]),
            (dictgen::InsensitiveStr::Ascii("cannbial"), &["cannibal"]),
            (dictgen::InsensitiveStr::Ascii("existencd"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("particals"), &["particles"]),
            (dictgen::InsensitiveStr::Ascii("paritiy"), &["parity"]),
            (
                dictgen::InsensitiveStr::Ascii("skillsofts"),
                &["skillshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availalability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("amercias"), &["americas"]),
            (dictgen::InsensitiveStr::Ascii("tidyness"), &["tidiness"]),
            (dictgen::InsensitiveStr::Ascii("contaned"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("projecties"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percetages"),
                &["percentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouptut"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("memery"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("reslve"), &["resolve"]),
            (
                dictgen::InsensitiveStr::Ascii("substanses"),
                &["substances"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconecct"), &["reconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("transmorgs"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantzation"),
                &["quantization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aboriginial"),
                &["aboriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("includeds"),
                &["includes", "included"],
            ),
            (dictgen::InsensitiveStr::Ascii("biologiset"), &["biologist"]),
            (dictgen::InsensitiveStr::Ascii("orcestra"), &["orchestra"]),
            (dictgen::InsensitiveStr::Ascii("spectifies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("overwirtten"),
                &["overwritten"],
            ),
            (dictgen::InsensitiveStr::Ascii("changging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("accessoires"),
                &["accessories", "accessorise"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifiees"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("perfermence"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trottling"),
                &["throttling", "trotting"],
            ),
            (dictgen::InsensitiveStr::Ascii("esger"), &["edger"]),
            (dictgen::InsensitiveStr::Ascii("nottaion"), &["notation"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsiquential"),
                &["inconsequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighobrhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("famlily"), &["family"]),
            (dictgen::InsensitiveStr::Ascii("discoverys"), &["discovers"]),
            (
                dictgen::InsensitiveStr::Ascii("abundacies"),
                &["abundances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idempontent"),
                &["idempotent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depecating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mkaing"), &["making"]),
            (dictgen::InsensitiveStr::Ascii("attampt"), &["attempt"]),
            (
                dictgen::InsensitiveStr::Ascii("explanantions"),
                &["explanations"],
            ),
            (dictgen::InsensitiveStr::Ascii("repblican"), &["republican"]),
            (
                dictgen::InsensitiveStr::Ascii("conflciting"),
                &["conflicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("ebcidc"), &["ebcdic"]),
            (dictgen::InsensitiveStr::Ascii("forcast"), &["forecast"]),
            (dictgen::InsensitiveStr::Ascii("polariy"), &["polarity"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuons"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decrecendo"),
                &["decrescendo"],
            ),
            (dictgen::InsensitiveStr::Ascii("casue"), &["cause"]),
            (dictgen::InsensitiveStr::Ascii("duplicats"), &["duplicates"]),
            (dictgen::InsensitiveStr::Ascii("mdify"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("symphatizes"),
                &["sympathize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnforms"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoits"), &["exploits"]),
            (
                dictgen::InsensitiveStr::Ascii("compativly"),
                &["compatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustraton"),
                &["illustration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocatation"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasient"), &["transient"]),
            (
                dictgen::InsensitiveStr::Ascii("deciptions"),
                &["depictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimeting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("iligal"), &["illegal"]),
            (
                dictgen::InsensitiveStr::Ascii("prostethic"),
                &["prosthetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypens"), &["hyphens"]),
            (dictgen::InsensitiveStr::Ascii("mysekf"), &["myself"]),
            (
                dictgen::InsensitiveStr::Ascii("consittent"),
                &["consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("bodiese"), &["bodies"]),
            (dictgen::InsensitiveStr::Ascii("brightoner"), &["brighten"]),
            (
                dictgen::InsensitiveStr::Ascii("triange"),
                &["triangle", "triage"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutomises"), &["customises"]),
            (dictgen::InsensitiveStr::Ascii("acronymes"), &["acronyms"]),
            (dictgen::InsensitiveStr::Ascii("paralle"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("keybaord"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("natches"), &["matches"]),
            (
                dictgen::InsensitiveStr::Ascii("levenhstein"),
                &["levenshtein"],
            ),
            (dictgen::InsensitiveStr::Ascii("patinet"), &["patient"]),
            (dictgen::InsensitiveStr::Ascii("organzier"), &["organizer"]),
            (
                dictgen::InsensitiveStr::Ascii("freedeom"),
                &["freedom", "freedoms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathfindir"),
                &["pathfinder"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemements"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("permanant"), &["permanent"]),
            (
                dictgen::InsensitiveStr::Ascii("exporer"),
                &["explorer", "exporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellement"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("probblez"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("operted"), &["operated"]),
            (
                dictgen::InsensitiveStr::Ascii("attriubtes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("percrent"), &["percent"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisicion"),
                &["inquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraye"), &["array"]),
            (
                dictgen::InsensitiveStr::Ascii("pecified"),
                &["specified", "pacified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superstisious"),
                &["superstitious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviornemntal"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegitimt"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transferrred"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("exictes"), &["excites"]),
            (dictgen::InsensitiveStr::Ascii("workapace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("prodces"), &["produces"]),
            (
                dictgen::InsensitiveStr::Ascii("indentaction"),
                &["indentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappearaing"),
                &["disappearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("accelerte"), &["accelerate"]),
            (
                dictgen::InsensitiveStr::Ascii("compensacion"),
                &["compensation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinosaus"), &["dinosaurs"]),
            (dictgen::InsensitiveStr::Ascii("poposal"), &["proposal"]),
            (
                dictgen::InsensitiveStr::Ascii("predacessors"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlucent"),
                &["translucent"],
            ),
            (dictgen::InsensitiveStr::Ascii("distrbute"), &["distribute"]),
            (dictgen::InsensitiveStr::Ascii("shaneal"), &["chenille"]),
            (dictgen::InsensitiveStr::Ascii("defaoulted"), &["defaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("diplomacia"),
                &["diplomatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prequesites"),
                &["prerequisites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissicm"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("untqueue"), &["unqueue"]),
            (dictgen::InsensitiveStr::Ascii("camoufle"), &["camouflage"]),
            (dictgen::InsensitiveStr::Ascii("wizrds"), &["wizards"]),
            (dictgen::InsensitiveStr::Ascii("quicket"), &["quickest"]),
            (
                dictgen::InsensitiveStr::Ascii("infograhic"),
                &["infographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collapseing"),
                &["collapsing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentaire"),
                &["documentaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inhertance"),
                &["inheritance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ressembling"),
                &["resembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("geoio"), &["geoip"]),
            (dictgen::InsensitiveStr::Ascii("intergate"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("sryians"), &["syrians"]),
            (dictgen::InsensitiveStr::Ascii("ignorms"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("infoms"), &["informs"]),
            (dictgen::InsensitiveStr::Ascii("cnosole"), &["console"]),
            (
                dictgen::InsensitiveStr::Ascii("explcitily"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breaktrhough"),
                &["breakthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighboors"),
                &["neighbors", "neighbours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comiplations"),
                &["compilations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("untils"),
                &["until", "utils"],
            ),
            (dictgen::InsensitiveStr::Ascii("trunactes"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("copletely"), &["completely"]),
            (dictgen::InsensitiveStr::Ascii("centrid"), &["centroid"]),
            (
                dictgen::InsensitiveStr::Ascii("hieroglphs"),
                &["hieroglyphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapoliates"),
                &["extrapolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomodations"),
                &["accommodations"],
            ),
            (dictgen::InsensitiveStr::Ascii("failture"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("sclupture"), &["sculpture"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequesite"),
                &["prerequisite"],
            ),
            (dictgen::InsensitiveStr::Ascii("beauti"), &["beauty"]),
            (dictgen::InsensitiveStr::Ascii("glitchty"), &["glitchy"]),
            (dictgen::InsensitiveStr::Ascii("advertie"), &["advertise"]),
            (
                dictgen::InsensitiveStr::Ascii("absoluthly"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alloate"),
                &["allocate", "allotted", "allot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpahabetical"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("amplitde"), &["amplitude"]),
            (dictgen::InsensitiveStr::Ascii("elephantos"), &["elephants"]),
            (
                dictgen::InsensitiveStr::Ascii("apperarances"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriotion"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("decieve"), &["deceive"]),
            (
                dictgen::InsensitiveStr::Ascii("apporoximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepereated"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("cheesse"), &["cheeses"]),
            (
                dictgen::InsensitiveStr::Ascii("goalkeeprs"),
                &["goalkeeper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transphoic"),
                &["transphobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("toothbruth"),
                &["toothbrush"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docmentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("intreface"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("concernt"), &["concert"]),
            (dictgen::InsensitiveStr::Ascii("rewachted"), &["rewatched"]),
            (
                dictgen::InsensitiveStr::Ascii("analyste"),
                &["analyse", "analyst"],
            ),
            (dictgen::InsensitiveStr::Ascii("incldues"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("conversons"),
                &["conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submariens"),
                &["submarines"],
            ),
            (dictgen::InsensitiveStr::Ascii("requestsed"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("exchenges"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("argmument"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("specifid"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("certficiates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("streatch"), &["stretch"]),
            (dictgen::InsensitiveStr::Ascii("miticates"), &["mitigates"]),
            (dictgen::InsensitiveStr::Ascii("imliments"), &["implements"]),
            (dictgen::InsensitiveStr::Ascii("luanches"), &["launches"]),
            (dictgen::InsensitiveStr::Ascii("cristal"), &["crystal"]),
            (dictgen::InsensitiveStr::Ascii("beutiful"), &["beautiful"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimilare"),
                &["dissimilar", "dissimilarly"],
            ),
            (dictgen::InsensitiveStr::Ascii("uneared"), &["unearned"]),
            (
                dictgen::InsensitiveStr::Ascii("calclulating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteactions"),
                &["interactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shanging"),
                &["changing", "shanking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("piches"),
                &["pinches", "pitches"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrubte"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("siluette"), &["silhouette"]),
            (
                dictgen::InsensitiveStr::Ascii("beaurocratic"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("wieghts"), &["weights"]),
            (dictgen::InsensitiveStr::Ascii("hypocriscy"), &["hypocrisy"]),
            (
                dictgen::InsensitiveStr::Ascii("chornological"),
                &["chronological"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparism"), &["comparison"]),
            (dictgen::InsensitiveStr::Ascii("comerants"), &["cormorants"]),
            (dictgen::InsensitiveStr::Ascii("integate"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("possibley"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("composision"),
                &["compositions"],
            ),
            (dictgen::InsensitiveStr::Ascii("purposley"), &["purposely"]),
            (dictgen::InsensitiveStr::Ascii("contatin"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("experimemtal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senstivity"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("sysdamin"), &["sysadmin"]),
            (dictgen::InsensitiveStr::Ascii("umless"), &["unless"]),
            (
                dictgen::InsensitiveStr::Ascii("inforamation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("imbaress"), &["embarrass"]),
            (dictgen::InsensitiveStr::Ascii("symtpoms"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("circluar"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("deacitivation"),
                &["deactivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("insterts"), &["inserts"]),
            (
                dictgen::InsensitiveStr::Ascii("scientifcally"),
                &["scientifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proselytisin"),
                &["proselytising"],
            ),
            (dictgen::InsensitiveStr::Ascii("quoteed"), &["quoted"]),
            (
                dictgen::InsensitiveStr::Ascii("initailizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sherif"), &["sheriff"]),
            (dictgen::InsensitiveStr::Ascii("finialized"), &["finalized"]),
            (
                dictgen::InsensitiveStr::Ascii("ligitamately"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialzied"),
                &["deserialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromosone"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("seethin"), &["seething"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdendently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preservato"),
                &["preservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("temorary"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("thicking"),
                &["thinking", "thickening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininistrative"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("magent"),
                &["magenta", "magnet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinacion"),
                &["determination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctnes"),
                &["correctness"],
            ),
            (dictgen::InsensitiveStr::Ascii("accssible"), &["accessible"]),
            (dictgen::InsensitiveStr::Ascii("heper"), &["helper"]),
            (dictgen::InsensitiveStr::Ascii("nurishes"), &["nourishes"]),
            (
                dictgen::InsensitiveStr::Ascii("intiutively"),
                &["intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approrpriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterpolating"),
                &["interpolating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cooment"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("genralise"), &["generalise"]),
            (dictgen::InsensitiveStr::Ascii("vbsrcript"), &["vbscript"]),
            (
                dictgen::InsensitiveStr::Ascii("signifficantly"),
                &["significantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ananlytical"),
                &["analytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("oepn"), &["open"]),
            (
                dictgen::InsensitiveStr::Ascii("lieutentant"),
                &["lieutenant"],
            ),
            (dictgen::InsensitiveStr::Ascii("hashtabke"), &["hashtable"]),
            (dictgen::InsensitiveStr::Ascii("distrubte"), &["distribute"]),
            (dictgen::InsensitiveStr::Ascii("inquiriy"), &["inquiry"]),
            (
                dictgen::InsensitiveStr::Ascii("unepectedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compensatie"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprieatry"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("suceess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("espression"), &["espresso"]),
            (dictgen::InsensitiveStr::Ascii("stealty"), &["stealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("deffering"),
                &["differing", "deferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procedurile"),
                &["procedure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigaiton"),
                &["investigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("variantion"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("aribtration"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("multipy"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("herely"),
                &["hereby", "heresy", "merely", "here"],
            ),
            (dictgen::InsensitiveStr::Ascii("pestore"), &["petstore"]),
            (dictgen::InsensitiveStr::Ascii("perfoming"), &["performing"]),
            (
                dictgen::InsensitiveStr::Ascii("renegosiators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("polutes"), &["pollutes"]),
            (dictgen::InsensitiveStr::Ascii("flexibele"), &["flexible"]),
            (dictgen::InsensitiveStr::Ascii("persisits"), &["persists"]),
            (
                dictgen::InsensitiveStr::Ascii("functionalilties"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repraesentational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("setiing"), &["setting"]),
            (
                dictgen::InsensitiveStr::Ascii("allocaiton"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("applikays"), &["appliqués"]),
            (dictgen::InsensitiveStr::Ascii("visibilit"), &["visibility"]),
            (
                dictgen::InsensitiveStr::Ascii("positivisty"),
                &["positivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrolled"), &["controlled"]),
            (
                dictgen::InsensitiveStr::Ascii("devirutalization"),
                &["devirtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("reprort"), &["report"]),
            (
                dictgen::InsensitiveStr::Ascii("methology"),
                &["methodology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumlative"),
                &["accumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("remingotn"), &["remington"]),
            (
                dictgen::InsensitiveStr::Ascii("finaly"),
                &["finally", "finale"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relocaiton"),
                &["relocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("integated"), &["integrated"]),
            (
                dictgen::InsensitiveStr::Ascii("libertatian"),
                &["libertarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("popularest"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("downlaoding"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustartor"),
                &["illustrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catergories"),
                &["categories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mercenaire"),
                &["mercenaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("veresion"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("wavelenghts"),
                &["wavelength", "wavelengths"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invididual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("templaetd"), &["templated"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcualte"),
                &["recalculate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suborutines"),
                &["subroutines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressd"),
                &["pressed", "press"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acumulation"),
                &["accumulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expementally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrija"),
                &["industrial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("medicaitons"),
                &["medications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthynic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutomisers"),
                &["customisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controvercy"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("beween"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("micrsoft"), &["microsoft"]),
            (dictgen::InsensitiveStr::Ascii("evluated"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("philospoher"),
                &["philosopher"],
            ),
            (dictgen::InsensitiveStr::Ascii("skpetics"), &["skeptics"]),
            (
                dictgen::InsensitiveStr::Ascii("geographicial"),
                &["geographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intsallation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("assistat"), &["assistants"]),
            (dictgen::InsensitiveStr::Ascii("certaily"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("pomise"), &["promise"]),
            (dictgen::InsensitiveStr::Ascii("bzte"), &["byte"]),
            (dictgen::InsensitiveStr::Ascii("concurrin"), &["concurring"]),
            (
                dictgen::InsensitiveStr::Ascii("perenially"),
                &["perennially"],
            ),
            (dictgen::InsensitiveStr::Ascii("notidy"), &["notify"]),
            (dictgen::InsensitiveStr::Ascii("resulver"), &["resolver"]),
            (
                dictgen::InsensitiveStr::Ascii("skpeticism"),
                &["skepticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statictical"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("faild"),
                &["failed", "fail", "fails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequentry"),
                &["frequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("striekr"), &["striker"]),
            (dictgen::InsensitiveStr::Ascii("intergral"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("resutled"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("cutoms"),
                &["custom", "customs"],
            ),
            (dictgen::InsensitiveStr::Ascii("initialial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("valu"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("covriate"), &["covariate"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstallling"),
                &["reinstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("manualyl"), &["manually"]),
            (dictgen::InsensitiveStr::Ascii("registor"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("barrriers"), &["barriers"]),
            (
                dictgen::InsensitiveStr::Ascii("vunerabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lewchemia"),
                &["leukaemia", "leukemia"],
            ),
            (dictgen::InsensitiveStr::Ascii("deigits"), &["digits"]),
            (
                dictgen::InsensitiveStr::Ascii("interation"),
                &["iteration", "interaction", "integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("humiditiy"), &["humidity"]),
            (dictgen::InsensitiveStr::Ascii("patname"), &["pathname"]),
            (dictgen::InsensitiveStr::Ascii("clevelry"), &["cleverly"]),
            (
                dictgen::InsensitiveStr::Ascii("eptrapolate"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("claening"), &["cleaning"]),
            (dictgen::InsensitiveStr::Ascii("expeects"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("rfeturned"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("combintation"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialication"),
                &["specialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("respoting"), &["reposting"]),
            (dictgen::InsensitiveStr::Ascii("vrsatile"), &["versatile"]),
            (
                dictgen::InsensitiveStr::Ascii("reaising"),
                &["realising", "raising"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapative"), &["adaptive"]),
            (
                dictgen::InsensitiveStr::Ascii("implementase"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("proivder"), &["provider"]),
            (
                dictgen::InsensitiveStr::Ascii("chatched"),
                &["caught", "chatted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percetnage"),
                &["percentages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invisblity"),
                &["invisibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("hierarhcy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("trasfer"), &["transfer"]),
            (dictgen::InsensitiveStr::Ascii("warniong"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("separetly"), &["separately"]),
            (
                dictgen::InsensitiveStr::Ascii("contradicitons"),
                &["contradictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalaescence"),
                &["coalescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("pannels"), &["panels"]),
            (dictgen::InsensitiveStr::Ascii("transvers"), &["transverse"]),
            (
                dictgen::InsensitiveStr::Ascii("trubadors"),
                &["troubadours"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepes"), &["shapes"]),
            (dictgen::InsensitiveStr::Ascii("dignose"), &["diagnose"]),
            (dictgen::InsensitiveStr::Ascii("precisily"), &["precisely"]),
            (dictgen::InsensitiveStr::Ascii("rendeirng"), &["rendering"]),
            (
                dictgen::InsensitiveStr::Ascii("resistans"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifac"), &["artifact"]),
            (dictgen::InsensitiveStr::Ascii("resuorces"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("hanppened"), &["happened"]),
            (
                dictgen::InsensitiveStr::Ascii("unterstand"),
                &["understand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("disbaling"), &["disabling"]),
            (
                dictgen::InsensitiveStr::Ascii("activaters"),
                &["activates", "activators", "actuators"],
            ),
            (dictgen::InsensitiveStr::Ascii("prvate"), &["private"]),
            (dictgen::InsensitiveStr::Ascii("sandalls"), &["sandals"]),
            (
                dictgen::InsensitiveStr::Ascii("assosciative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("arange"), &["arrange"]),
            (dictgen::InsensitiveStr::Ascii("actiavted"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("corupted"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("christinaity"),
                &["christianity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfrorming"),
                &["performing"],
            ),
            (dictgen::InsensitiveStr::Ascii("deiners"), &["deniers"]),
            (
                dictgen::InsensitiveStr::Ascii("stationair"),
                &["stationary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docomentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tentalce"), &["tentacle"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nast"),
                &["mast", "nasty", "nest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cumulitive"),
                &["cumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyckes"), &["cycles"]),
            (
                dictgen::InsensitiveStr::Ascii("capabilitiesu"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("immobilzer"), &["immobile"]),
            (
                dictgen::InsensitiveStr::Ascii("privisoned"),
                &["provisioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("itterates"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("tnan"), &["than"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstratie"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expet"),
                &["expect", "expat", "expert"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquarims"), &["aquariums"]),
            (
                dictgen::InsensitiveStr::Ascii("autoconpletion"),
                &["autocompletion"],
            ),
            (dictgen::InsensitiveStr::Ascii("sligtly"), &["slightly"]),
            (
                dictgen::InsensitiveStr::Ascii("maanagement"),
                &["management"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapating"),
                &["dissipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoruces"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("unbareable"),
                &["unbearable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellecual"),
                &["intellectual"],
            ),
            (dictgen::InsensitiveStr::Ascii("metris"), &["metrics"]),
            (dictgen::InsensitiveStr::Ascii("scenaireos"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("mathemeticians"),
                &["mathematicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referencable"),
                &["referenceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializedd"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("activness"), &["activeness"]),
            (dictgen::InsensitiveStr::Ascii("anoyed"), &["annoyed"]),
            (dictgen::InsensitiveStr::Ascii("exhisting"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("congfiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("canidates"), &["candidates"]),
            (
                dictgen::InsensitiveStr::Ascii("brackground"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resgination"),
                &["resignation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternitively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefics"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("aethists"), &["atheists"]),
            (dictgen::InsensitiveStr::Ascii("tortillias"), &["tortilla"]),
            (dictgen::InsensitiveStr::Ascii("sometim"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("arrauy"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("agsinst"), &["against"]),
            (dictgen::InsensitiveStr::Ascii("cantalope"), &["cantaloupe"]),
            (
                dictgen::InsensitiveStr::Ascii("indendent"),
                &["indent", "indented", "independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrelavant"),
                &["irrelevant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportionel"),
                &["proportional"],
            ),
            (dictgen::InsensitiveStr::Ascii("porprties"), &["properties"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsitratively"),
                &["administratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conclsuion"),
                &["concussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("asscociate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrevates"),
                &["aggravates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditoins"),
                &["conditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpatrent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achive"),
                &["achieve", "archive"],
            ),
            (dictgen::InsensitiveStr::Ascii("annotae"), &["annotate"]),
            (dictgen::InsensitiveStr::Ascii("nurishing"), &["nourishing"]),
            (
                dictgen::InsensitiveStr::Ascii("assicieted"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mayu"), &["may"]),
            (dictgen::InsensitiveStr::Ascii("rgisters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("cononicalizing"),
                &["canonicalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsyncronized"),
                &["unsynchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasturized"),
                &["pasteurized"],
            ),
            (dictgen::InsensitiveStr::Ascii("reluctanct"), &["reluctant"]),
            (
                dictgen::InsensitiveStr::Ascii("amalgomated"),
                &["amalgamated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reectangular"),
                &["rectangular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extendible"),
                &["extendable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarenty"),
                &["guaranty", "guarantee"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliam"), &["claim"]),
            (
                dictgen::InsensitiveStr::Ascii("congfigured"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplciated"),
                &["duplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developmenet"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("ileagle"), &["illegal"]),
            (dictgen::InsensitiveStr::Ascii("stattistic"), &["statistic"]),
            (
                dictgen::InsensitiveStr::Ascii("illustartion"),
                &["illustration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjectivos"),
                &["adjectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("cappacitor"), &["capacitor"]),
            (dictgen::InsensitiveStr::Ascii("istambul"), &["istanbul"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readabilty"),
                &["readability"],
            ),
            (dictgen::InsensitiveStr::Ascii("leghts"), &["lengths"]),
            (dictgen::InsensitiveStr::Ascii("defaulr"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("decleres"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("reintepreted"),
                &["reinterpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrivial"), &["arrival"]),
            (dictgen::InsensitiveStr::Ascii("locatoins"), &["locations"]),
            (dictgen::InsensitiveStr::Ascii("unlcuky"), &["unlucky"]),
            (
                dictgen::InsensitiveStr::Ascii("retranmissions"),
                &["retransmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documemtations"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("recconet"), &["reconnect"]),
            (dictgen::InsensitiveStr::Ascii("veryfieng"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitiers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("forfiet"), &["forfeit"]),
            (dictgen::InsensitiveStr::Ascii("inlude"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("lightwieght"),
                &["lightweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("suffex"), &["suffix"]),
            (
                dictgen::InsensitiveStr::Ascii("determinaton"),
                &["determination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pahts"),
                &["paths", "pats", "parts"],
            ),
            (dictgen::InsensitiveStr::Ascii("lengthh"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("homophopia"),
                &["homophobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirrect"), &["direct"]),
            (
                dictgen::InsensitiveStr::Ascii("converved"),
                &["conserved", "conversed", "converted", "converged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photogropher"),
                &["photographer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compenstated"),
                &["compensated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deamonisation"),
                &["daemonisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementeing"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalutation"),
                &["evaluation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("granulatity"),
                &["granularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("finially"), &["finally"]),
            (
                dictgen::InsensitiveStr::Ascii("delimititing"),
                &["delimiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuati"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straigthen"),
                &["straighten"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyton"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("keybooard"), &["keyboard"]),
            (
                dictgen::InsensitiveStr::Ascii("bootstrapers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filld"),
                &["filled", "filed", "fill"],
            ),
            (dictgen::InsensitiveStr::Ascii("integerate"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("borad"), &["board"]),
            (dictgen::InsensitiveStr::Ascii("satsified"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("approximat"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caclulates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("stylisch"), &["stylish"]),
            (
                dictgen::InsensitiveStr::Ascii("mision"),
                &["mission", "minion", "vision"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpusts"), &["inputs"]),
            (dictgen::InsensitiveStr::Ascii("materlas"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("conntaining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innacuracy"),
                &["inaccuracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersectin"),
                &["intersection", "intersecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilvalents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("kubenetes"), &["kubernetes"]),
            (
                dictgen::InsensitiveStr::Ascii("appriximated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfigures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrouage"),
                &["discourages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitor"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("alognment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("constituant"),
                &["constituent"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendirers"), &["renderers"]),
            (dictgen::InsensitiveStr::Ascii("sanetise"), &["sanitise"]),
            (dictgen::InsensitiveStr::Ascii("chalenged"), &["challenged"]),
            (
                dictgen::InsensitiveStr::Ascii("assingment"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("excrept"), &["excerpt"]),
            (dictgen::InsensitiveStr::Ascii("traverssal"), &["traversal"]),
            (dictgen::InsensitiveStr::Ascii("coutner"), &["counter"]),
            (
                dictgen::InsensitiveStr::Ascii("consenusal"),
                &["consensual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represnting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ristrict"), &["restrict"]),
            (
                dictgen::InsensitiveStr::Ascii("enchantmants"),
                &["enchantments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kongregation"),
                &["congregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palce"),
                &["place", "palace"],
            ),
            (dictgen::InsensitiveStr::Ascii("resaerch"), &["research"]),
            (dictgen::InsensitiveStr::Ascii("cheduled"), &["scheduled"]),
            (
                dictgen::InsensitiveStr::Ascii("satifaction"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("replentish"), &["replenish"]),
            (
                dictgen::InsensitiveStr::Ascii("inpoverished"),
                &["impoverished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undertsood"),
                &["understood"],
            ),
            (dictgen::InsensitiveStr::Ascii("trubble"), &["trouble"]),
            (dictgen::InsensitiveStr::Ascii("untimatly"), &["ultimately"]),
            (dictgen::InsensitiveStr::Ascii("dangerouns"), &["dangerous"]),
            (
                dictgen::InsensitiveStr::Ascii("respondendo"),
                &["responded"],
            ),
            (dictgen::InsensitiveStr::Ascii("cailbre"), &["calibre"]),
            (dictgen::InsensitiveStr::Ascii("salery"), &["salary"]),
            (
                dictgen::InsensitiveStr::Ascii("esponsibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("pennisula"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("repeticion"),
                &["repetition"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartlege"), &["cartilage"]),
            (
                dictgen::InsensitiveStr::Ascii("boardcasting"),
                &["broadcasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seqeuncers"),
                &["sequencers"],
            ),
            (dictgen::InsensitiveStr::Ascii("anythng"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("downlowded"),
                &["downloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("approuve"), &["approve"]),
            (dictgen::InsensitiveStr::Ascii("decocdings"), &["decodings"]),
            (
                dictgen::InsensitiveStr::Ascii("comminucate"),
                &["communicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaates"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("tentaively"),
                &["tentatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherats"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("expcting"), &["expecting"]),
            (dictgen::InsensitiveStr::Ascii("htings"), &["things"]),
            (
                dictgen::InsensitiveStr::Ascii("accessability"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("satarting"), &["starting"]),
            (dictgen::InsensitiveStr::Ascii("unsinged"), &["unsigned"]),
            (
                dictgen::InsensitiveStr::Ascii("coersitivity"),
                &["coercitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticiszed"),
                &["criticise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("formuale"), &["formulae"]),
            (
                dictgen::InsensitiveStr::Ascii("definine"),
                &["define", "definite"],
            ),
            (dictgen::InsensitiveStr::Ascii("assemblin"), &["assembling"]),
            (dictgen::InsensitiveStr::Ascii("nubering"), &["numbering"]),
            (
                dictgen::InsensitiveStr::Ascii("goegraphic"),
                &["geographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoures"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("curnels"), &["colonels"]),
            (dictgen::InsensitiveStr::Ascii("casltes"), &["castles"]),
            (
                dictgen::InsensitiveStr::Ascii("mathimatics"),
                &["mathematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("launher"), &["launcher"]),
            (dictgen::InsensitiveStr::Ascii("lastes"), &["latest"]),
            (dictgen::InsensitiveStr::Ascii("collpsing"), &["collapsing"]),
            (dictgen::InsensitiveStr::Ascii("vview"), &["view"]),
            (dictgen::InsensitiveStr::Ascii("eauality"), &["equality"]),
            (dictgen::InsensitiveStr::Ascii("declears"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("manifacturers"),
                &["manufacturers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repositary"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("upadates"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("otherwhise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappointment"),
                &["disappointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("asserion"), &["assertion"]),
            (
                dictgen::InsensitiveStr::Ascii("deomnstration"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experessions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("reudce"), &["reduce"]),
            (
                dictgen::InsensitiveStr::Ascii("inforgivable"),
                &["unforgivable"],
            ),
            (dictgen::InsensitiveStr::Ascii("mananged"), &["managed"]),
            (
                dictgen::InsensitiveStr::Ascii("compiliation"),
                &["compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("netural"),
                &["neutral", "natural"],
            ),
            (dictgen::InsensitiveStr::Ascii("precsions"), &["precisions"]),
            (
                dictgen::InsensitiveStr::Ascii("trucations"),
                &["truncations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accoundant"),
                &["accountant"],
            ),
            (dictgen::InsensitiveStr::Ascii("lcoated"), &["located"]),
            (dictgen::InsensitiveStr::Ascii("exhauted"), &["exhausted"]),
            (
                dictgen::InsensitiveStr::Ascii("callef"),
                &["called", "caller"],
            ),
            (dictgen::InsensitiveStr::Ascii("sparyed"), &["sprayed"]),
            (
                dictgen::InsensitiveStr::Ascii("heursitics"),
                &["heuristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mastquerade"),
                &["masquerade"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relcutantly"),
                &["reluctantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archietecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignnored"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("metapackges"),
                &["metapackages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythesized"),
                &["synthesized"],
            ),
            (dictgen::InsensitiveStr::Ascii("creepig"), &["creeping"]),
            (dictgen::InsensitiveStr::Ascii("upstreems"), &["upstreams"]),
            (dictgen::InsensitiveStr::Ascii("ocassion"), &["occasion"]),
            (dictgen::InsensitiveStr::Ascii("computin"), &["computing"]),
            (dictgen::InsensitiveStr::Ascii("lcoalhost"), &["localhost"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("usrename"), &["username"]),
            (
                dictgen::InsensitiveStr::Ascii("impermable"),
                &["impermeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("intalles"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("opeartor"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("vrey"), &["very"]),
            (dictgen::InsensitiveStr::Ascii("arealdy"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("relativy"), &["relativity"]),
            (
                dictgen::InsensitiveStr::Ascii("passionant"),
                &["passionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("symptomps"), &["symptoms"]),
            (
                dictgen::InsensitiveStr::Ascii("catholacism"),
                &["catholicism"],
            ),
            (dictgen::InsensitiveStr::Ascii("quiessent"), &["quiescent"]),
            (
                dictgen::InsensitiveStr::Ascii("calclation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("resumin"), &["resuming"]),
            (dictgen::InsensitiveStr::Ascii("euphorical"), &["euphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropiate"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("containned"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("gullbile"), &["gullible"]),
            (dictgen::InsensitiveStr::Ascii("strpiped"), &["stripped"]),
            (dictgen::InsensitiveStr::Ascii("variatnts"), &["variants"]),
            (dictgen::InsensitiveStr::Ascii("natched"), &["matched"]),
            (
                dictgen::InsensitiveStr::Ascii("boyfreinds"),
                &["boyfriends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custonized"),
                &["customized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistendtly"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallellization"),
                &["parallelization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibity"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("shcolars"), &["scholars"]),
            (
                dictgen::InsensitiveStr::Ascii("whitholding"),
                &["withholding"],
            ),
            (dictgen::InsensitiveStr::Ascii("weeek"), &["week"]),
            (
                dictgen::InsensitiveStr::Ascii("soundtraks"),
                &["soundtracks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itereatively"),
                &["iteratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("verigated"), &["variegated"]),
            (dictgen::InsensitiveStr::Ascii("elctrons"), &["electrons"]),
            (
                dictgen::InsensitiveStr::Ascii("exceoted"),
                &["excepted", "exempted"],
            ),
            (dictgen::InsensitiveStr::Ascii("textur"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("variey"), &["variety"]),
            (dictgen::InsensitiveStr::Ascii("paches"), &["patches"]),
            (dictgen::InsensitiveStr::Ascii("iteation"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instalation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("itslev"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("compeittion"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unstalles"),
                &["installs", "uninstalls"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steriotypical"),
                &["stereotypical"],
            ),
            (dictgen::InsensitiveStr::Ascii("viwed"), &["viewed"]),
            (
                dictgen::InsensitiveStr::Ascii("garuantied"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("unvierse"), &["universe"]),
            (
                dictgen::InsensitiveStr::Ascii("displacemnt"),
                &["displacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("sissering"), &["scissoring"]),
            (
                dictgen::InsensitiveStr::Ascii("itentional"),
                &["intentional"],
            ),
            (dictgen::InsensitiveStr::Ascii("comtain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("termiantions"),
                &["terminations"],
            ),
            (dictgen::InsensitiveStr::Ascii("fielneame"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("disamgiguation"),
                &["disambiguation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbraser"), &["embracer"]),
            (dictgen::InsensitiveStr::Ascii("monitord"), &["monitored"]),
            (
                dictgen::InsensitiveStr::Ascii("algotirhmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immigrato"),
                &["immigration"],
            ),
            (dictgen::InsensitiveStr::Ascii("formated"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("translaction"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchronlously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucsessor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("irrevelant"),
                &["irrelevant", "irreverent"],
            ),
            (dictgen::InsensitiveStr::Ascii("reccomend"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("literarly"), &["literary"]),
            (
                dictgen::InsensitiveStr::Ascii("inisialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtaulise"),
                &["virtualise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretated"),
                &["misinterpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("chiselle"), &["chisel"]),
            (dictgen::InsensitiveStr::Ascii("literarlly"), &["literary"]),
            (dictgen::InsensitiveStr::Ascii("sautayes"), &["sautés"]),
            (
                dictgen::InsensitiveStr::Ascii("executabled"),
                &["executable", "executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoicated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("loiusiana"), &["louisiana"]),
            (dictgen::InsensitiveStr::Ascii("pariticles"), &["particles"]),
            (dictgen::InsensitiveStr::Ascii("grwo"), &["grow"]),
            (dictgen::InsensitiveStr::Ascii("balse"), &["false"]),
            (dictgen::InsensitiveStr::Ascii("abtract"), &["abstract"]),
            (
                dictgen::InsensitiveStr::Ascii("eeverywhere"),
                &["everywhere"],
            ),
            (dictgen::InsensitiveStr::Ascii("littel"), &["little"]),
            (dictgen::InsensitiveStr::Ascii("polymoprh"), &["polymorph"]),
            (dictgen::InsensitiveStr::Ascii("xeonblade"), &["xenoblade"]),
            (
                dictgen::InsensitiveStr::Ascii("schem"),
                &["scheme", "schema"],
            ),
            (dictgen::InsensitiveStr::Ascii("compliler"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("inititatives"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alligeance"),
                &["allegiance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenatry"),
                &["documentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("articat"), &["artifact"]),
            (dictgen::InsensitiveStr::Ascii("lackering"), &["lacquering"]),
            (
                dictgen::InsensitiveStr::Ascii("submittion"),
                &["submitting", "submission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecentiles"),
                &["percentiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particualr"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("deterministinc"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("detructed"), &["destructed"]),
            (
                dictgen::InsensitiveStr::Ascii("certficate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("hadling"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("confugires"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initated"),
                &["initiated", "imitated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinlges"), &["singles"]),
            (dictgen::InsensitiveStr::Ascii("schedulin"), &["scheduling"]),
            (
                dictgen::InsensitiveStr::Ascii("strangly"),
                &["strangely", "strange", "strangle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitation"),
                &["arbitration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("bannet"), &["bayonet"]),
            (
                dictgen::InsensitiveStr::Ascii("capacitores"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholarhip"),
                &["scholarship"],
            ),
            (dictgen::InsensitiveStr::Ascii("currrency"), &["currency"]),
            (dictgen::InsensitiveStr::Ascii("hurricaine"), &["hurricane"]),
            (
                dictgen::InsensitiveStr::Ascii("renaiisance"),
                &["renaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimizaer"), &["optimizer"]),
            (
                dictgen::InsensitiveStr::Ascii("christines"),
                &["christians"],
            ),
            (dictgen::InsensitiveStr::Ascii("feeded"), &["fed"]),
            (dictgen::InsensitiveStr::Ascii("spiritus"), &["spirits"]),
            (
                dictgen::InsensitiveStr::Ascii("instrinics"),
                &["intrinsics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dpend"),
                &["depend", "spend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbaring"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("jefferry"), &["jeffery"]),
            (
                dictgen::InsensitiveStr::Ascii("histori"),
                &["history", "historic"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgards"), &["discards"]),
            (dictgen::InsensitiveStr::Ascii("abouta"), &["about"]),
            (
                dictgen::InsensitiveStr::Ascii("sevices"),
                &["services", "devices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceleration"),
                &["acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protelariat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("hungarin"), &["hungarian"]),
            (dictgen::InsensitiveStr::Ascii("toopology"), &["topology"]),
            (dictgen::InsensitiveStr::Ascii("modlue"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("cretated"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("handlbars"), &["handlebars"]),
            (dictgen::InsensitiveStr::Ascii("synagouge"), &["synagogue"]),
            (dictgen::InsensitiveStr::Ascii("bundeles"), &["bundles"]),
            (dictgen::InsensitiveStr::Ascii("roestta"), &["rosetta"]),
            (dictgen::InsensitiveStr::Ascii("intructs"), &["instructs"]),
            (
                dictgen::InsensitiveStr::Ascii("multipliciaton"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("iritating"), &["irritating"]),
            (
                dictgen::InsensitiveStr::Ascii("rennovation"),
                &["renovation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceassers"), &["caesars"]),
            (
                dictgen::InsensitiveStr::Ascii("unresovlable"),
                &["unresolvable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("athleticsm"),
                &["athleticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justificativos"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrobutors"),
                &["distributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("poltical"), &["political"]),
            (
                dictgen::InsensitiveStr::Ascii("perputuating"),
                &["perpetuating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatted"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availabability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refenences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuplic"), &["republic"]),
            (
                dictgen::InsensitiveStr::Ascii("concentrait"),
                &["concentrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuining"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arameterizes"),
                &["parameterizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("foller"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("geomertic"), &["geometric"]),
            (dictgen::InsensitiveStr::Ascii("populaties"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("direktional"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("serveless"), &["serverless"]),
            (dictgen::InsensitiveStr::Ascii("clannand"), &["clannad"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicaing"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authecated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asymmetri"),
                &["asymmetric", "asymmetry"],
            ),
            (dictgen::InsensitiveStr::Ascii("comsumed"), &["consumed"]),
            (dictgen::InsensitiveStr::Ascii("abuot"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("optimistc"), &["optimistic"]),
            (dictgen::InsensitiveStr::Ascii("oprator"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("speficiier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("establishemnt"),
                &["establishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrimiate"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmained"), &["examined"]),
            (dictgen::InsensitiveStr::Ascii("componnet"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("exeception"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("wehenever"), &["whenever"]),
            (
                dictgen::InsensitiveStr::Ascii("venezuelian"),
                &["venezuela"],
            ),
            (dictgen::InsensitiveStr::Ascii("slippes"), &["slippers"]),
            (dictgen::InsensitiveStr::Ascii("liscenced"), &["licenced"]),
            (dictgen::InsensitiveStr::Ascii("colsed"), &["closed"]),
            (dictgen::InsensitiveStr::Ascii("manulally"), &["manually"]),
            (dictgen::InsensitiveStr::Ascii("deccimal"), &["decimal"]),
            (
                dictgen::InsensitiveStr::Ascii("nessecarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influented"),
                &["influenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definishions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outrageoulsy"),
                &["outrageously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laceholder"),
                &["placeholder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represending"),
                &["representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simbolically"),
                &["symbolically"],
            ),
            (dictgen::InsensitiveStr::Ascii("frigign"), &["friggin"]),
            (
                dictgen::InsensitiveStr::Ascii("epxerimental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("cupboad"), &["cupboard"]),
            (
                dictgen::InsensitiveStr::Ascii("compliations"),
                &["compilation", "compilations", "complications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etrors"),
                &["errors", "terrors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptoys"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seskatchewan"),
                &["saskatchewan"],
            ),
            (dictgen::InsensitiveStr::Ascii("buidings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("instatiation"),
                &["instantiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("treausre"), &["treasure"]),
            (
                dictgen::InsensitiveStr::Ascii("contiuguous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jurisdiccion"),
                &["jurisdiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("northwestener"),
                &["northwestern"],
            ),
            (dictgen::InsensitiveStr::Ascii("alllocated"), &["allocated"]),
            (dictgen::InsensitiveStr::Ascii("excerciser"), &["exerciser"]),
            (dictgen::InsensitiveStr::Ascii("irelevent"), &["irrelevant"]),
            (
                dictgen::InsensitiveStr::Ascii("implimentors"),
                &["implementers"],
            ),
            (dictgen::InsensitiveStr::Ascii("beliveble"), &["believable"]),
            (
                dictgen::InsensitiveStr::Ascii("deporarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("amrchair"), &["armchair"]),
            (dictgen::InsensitiveStr::Ascii("agrgument"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("skelton"), &["skeleton"]),
            (
                dictgen::InsensitiveStr::Ascii("intimiated"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experieced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asymetically"),
                &["asymmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("whoms"), &["whom", "whose"]),
            (
                dictgen::InsensitiveStr::Ascii("algorrithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramecias"),
                &["paramedics"],
            ),
            (dictgen::InsensitiveStr::Ascii("definatey"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("irraditated"),
                &["irradiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychologycal"),
                &["psychologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiatied"), &["initiated"]),
            (dictgen::InsensitiveStr::Ascii("ake"), &["ache"]),
            (dictgen::InsensitiveStr::Ascii("abstrctor"), &["abstractor"]),
            (dictgen::InsensitiveStr::Ascii("becaouse"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulacion"),
                &["manipulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionay"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscoitaions"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("becauseq"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("representstive"),
                &["representatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gereations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("olbiterated"),
                &["obliterated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescrition"),
                &["prescription"],
            ),
            (dictgen::InsensitiveStr::Ascii("texual"), &["textual"]),
            (dictgen::InsensitiveStr::Ascii("chernboyl"), &["chernobyl"]),
            (dictgen::InsensitiveStr::Ascii("querrying"), &["querying"]),
            (dictgen::InsensitiveStr::Ascii("evovles"), &["evolves"]),
            (dictgen::InsensitiveStr::Ascii("pojection"), &["projection"]),
            (dictgen::InsensitiveStr::Ascii("pipeines"), &["pipelines"]),
            (dictgen::InsensitiveStr::Ascii("sudmobule"), &["submodule"]),
            (
                dictgen::InsensitiveStr::Ascii("experimten"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trignametric"),
                &["trigonometric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repacement"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("bottome"), &["bottom"]),
            (
                dictgen::InsensitiveStr::Ascii("compartent"),
                &["compartment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vitualize"),
                &["virtualize", "visualize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("anotators"), &["annotators"]),
            (
                dictgen::InsensitiveStr::Ascii("swiching"),
                &["switching", "swishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangts"),
                &["rearrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("phazing"), &["phasing"]),
            (dictgen::InsensitiveStr::Ascii("tutoriales"), &["tutorials"]),
            (dictgen::InsensitiveStr::Ascii("letivicus"), &["leviticus"]),
            (
                dictgen::InsensitiveStr::Ascii("handrwiting"),
                &["handwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suscedes"),
                &["secedes", "succeeds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psycholigists"),
                &["psychologists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indespensable"),
                &["indispensable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformates"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuite"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("interveen"), &["intervene"]),
            (dictgen::InsensitiveStr::Ascii("swaer"), &["swear"]),
            (
                dictgen::InsensitiveStr::Ascii("attributred"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotpyical"),
                &["stereotypical"],
            ),
            (dictgen::InsensitiveStr::Ascii("fowarded"), &["forwarded"]),
            (
                dictgen::InsensitiveStr::Ascii("psychiatrict"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("surronds"), &["surrounds"]),
            (
                dictgen::InsensitiveStr::Ascii("beanches"),
                &["benches", "branches", "beaches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accuratley"),
                &["accurately"],
            ),
            (dictgen::InsensitiveStr::Ascii("unqoutes"), &["unquotes"]),
            (
                dictgen::InsensitiveStr::Ascii("dissatisifed"),
                &["dissatisfied"],
            ),
            (dictgen::InsensitiveStr::Ascii("sickamore"), &["sycamore"]),
            (
                dictgen::InsensitiveStr::Ascii("corordinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profissional"),
                &["professional"],
            ),
            (dictgen::InsensitiveStr::Ascii("inheirt"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("regisers"), &["registers"]),
            (dictgen::InsensitiveStr::Ascii("choosin"), &["choosing"]),
            (
                dictgen::InsensitiveStr::Ascii("facilitare"),
                &["facilitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lithuaina"), &["lithuania"]),
            (dictgen::InsensitiveStr::Ascii("peicemeal"), &["piecemeal"]),
            (
                dictgen::InsensitiveStr::Ascii("frustrastion"),
                &["frustrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destuctors"),
                &["destructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("allowoed"), &["allowed"]),
            (dictgen::InsensitiveStr::Ascii("nomalised"), &["normalised"]),
            (dictgen::InsensitiveStr::Ascii("rhythim"), &["rhythm"]),
            (dictgen::InsensitiveStr::Ascii("hypothess"), &["hypotheses"]),
            (
                dictgen::InsensitiveStr::Ascii("assasined"),
                &["assassinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatenading"),
                &["concatenating"],
            ),
            (dictgen::InsensitiveStr::Ascii("neworks"), &["networks"]),
            (
                dictgen::InsensitiveStr::Ascii("organizacion"),
                &["organization"],
            ),
            (dictgen::InsensitiveStr::Ascii("curcuits"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("triscadecaphobia"),
                &["triskaidekaphobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percential"),
                &["percentile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mirrorowing"),
                &["mirroring"],
            ),
            (dictgen::InsensitiveStr::Ascii("pinpoit"), &["pinpoint"]),
            (dictgen::InsensitiveStr::Ascii("indirecty"), &["indirectly"]),
            (dictgen::InsensitiveStr::Ascii("uyou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("virtaul"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("quartercask"),
                &["quarterbacks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaetion"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("envrions"), &["environs"]),
            (
                dictgen::InsensitiveStr::Ascii("burueacrats"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("satus"), &["status"]),
            (
                dictgen::InsensitiveStr::Ascii("controdicting"),
                &["contradicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintaince"),
                &["maintained", "maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microcender"),
                &["microcenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("chipps"), &["chips"]),
            (
                dictgen::InsensitiveStr::Ascii("chrash"),
                &["crash", "thrash", "trash"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminl"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("archaelogists"),
                &["archaeologists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("netruality"),
                &["neutrality"],
            ),
            (dictgen::InsensitiveStr::Ascii("appnedage"), &["appendage"]),
            (dictgen::InsensitiveStr::Ascii("doalog"), &["dialog"]),
            (dictgen::InsensitiveStr::Ascii("arguabley"), &["arguably"]),
            (dictgen::InsensitiveStr::Ascii("sugestion"), &["suggestion"]),
            (
                dictgen::InsensitiveStr::Ascii("definaitly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arranment"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("dingity"), &["dignity"]),
            (dictgen::InsensitiveStr::Ascii("affaires"), &["affairs"]),
            (
                dictgen::InsensitiveStr::Ascii("structurin"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatendated"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispensarios"),
                &["dispensaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coomitters"),
                &["committers"],
            ),
            (dictgen::InsensitiveStr::Ascii("numebr"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("cahsiers"), &["cashiers"]),
            (
                dictgen::InsensitiveStr::Ascii("thunderblot"),
                &["thunderbolt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represnets"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lisenced"),
                &["silenced", "licenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sophistacated"),
                &["sophisticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("beacuoup"), &["beaucoup"]),
            (dictgen::InsensitiveStr::Ascii("distancin"), &["distancing"]),
            (
                dictgen::InsensitiveStr::Ascii("misisonary"),
                &["missionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("monumentus"), &["monuments"]),
            (
                dictgen::InsensitiveStr::Ascii("discrimnator"),
                &["discriminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ptting"),
                &["putting", "pitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acconplishment"),
                &["accomplishment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predominantley"),
                &["predominantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("attornies"), &["attorneys"]),
            (dictgen::InsensitiveStr::Ascii("vallys"), &["valleys"]),
            (
                dictgen::InsensitiveStr::Ascii("subdivised"),
                &["subdivided"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternaive"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disparingly"),
                &["disparagingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("baroke"), &["baroque"]),
            (dictgen::InsensitiveStr::Ascii("dispaly"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("genereating"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgronds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceededs"), &["proceeds"]),
            (dictgen::InsensitiveStr::Ascii("cletics"), &["celtics"]),
            (
                dictgen::InsensitiveStr::Ascii("translatio"),
                &["translator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illumanati"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thorughout"),
                &["throughout"],
            ),
            (dictgen::InsensitiveStr::Ascii("binairies"), &["binaries"]),
            (dictgen::InsensitiveStr::Ascii("nned"), &["need"]),
            (dictgen::InsensitiveStr::Ascii("porivdes"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("renegociable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcurator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("calicum"), &["calcium"]),
            (dictgen::InsensitiveStr::Ascii("probaby"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("withought"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("recognizeable"),
                &["recognizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifity"),
                &["specificity"],
            ),
            (dictgen::InsensitiveStr::Ascii("descriced"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("semaintics"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("syschronize"),
                &["synchronize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destructro"),
                &["destructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnostacism"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("nieveatay"), &["naivete"]),
            (dictgen::InsensitiveStr::Ascii("dominoin"), &["dominion"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprenurs"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functtional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shystemerror"),
                &["systemerror"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosectuors"),
                &["prosecutors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configrations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("slewths"), &["sleuths"]),
            (dictgen::InsensitiveStr::Ascii("meesages"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("charizma"), &["charisma"]),
            (
                dictgen::InsensitiveStr::Ascii("relocaiing"),
                &["relocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("groth"), &["growth"]),
            (
                dictgen::InsensitiveStr::Ascii("triming"),
                &["trimming", "timing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rectangulaire"),
                &["rectangular"],
            ),
            (dictgen::InsensitiveStr::Ascii("breakoint"), &["breakpoint"]),
            (dictgen::InsensitiveStr::Ascii("palced"), &["placed"]),
            (dictgen::InsensitiveStr::Ascii("perfoms"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("economicas"), &["economics"]),
            (dictgen::InsensitiveStr::Ascii("indiaan"), &["indiana"]),
            (
                dictgen::InsensitiveStr::Ascii("laborotory"),
                &["laboratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplissment"),
                &["accomplishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("shbang"), &["shebang"]),
            (
                dictgen::InsensitiveStr::Ascii("accidetly"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("withholdng"),
                &["withholding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifastation"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("traveleres"), &["travelers"]),
            (dictgen::InsensitiveStr::Ascii("vessells"), &["vessels"]),
            (dictgen::InsensitiveStr::Ascii("religeous"), &["religious"]),
            (dictgen::InsensitiveStr::Ascii("ownerd"), &["owned"]),
            (
                dictgen::InsensitiveStr::Ascii("littelfinger"),
                &["littlefinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dumplications"),
                &["duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("publisehr"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("contrtoller"),
                &["controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("configred"), &["configured"]),
            (
                dictgen::InsensitiveStr::Ascii("unappeasing"),
                &["unappealing"],
            ),
            (dictgen::InsensitiveStr::Ascii("erease"), &["erase"]),
            (
                dictgen::InsensitiveStr::Ascii("catepillars"),
                &["caterpillars"],
            ),
            (dictgen::InsensitiveStr::Ascii("cauncks"), &["canucks"]),
            (
                dictgen::InsensitiveStr::Ascii("correspontends"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaxating"),
                &["relaxation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meachinisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("newmonectomy"),
                &["pneumonectomy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterprices"),
                &["enterprises"],
            ),
            (dictgen::InsensitiveStr::Ascii("boeleans"), &["booleans"]),
            (
                dictgen::InsensitiveStr::Ascii("depentencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("varation"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("implementtaion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interaxtion"),
                &["interaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compied"),
                &["compiled", "copied", "complied"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifify"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("collectos"), &["collectors"]),
            (
                dictgen::InsensitiveStr::Ascii("scenegraaphs"),
                &["scenegraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("testin"), &["testing"]),
            (
                dictgen::InsensitiveStr::Ascii("misdemenaor"),
                &["misdemeanor"],
            ),
            (dictgen::InsensitiveStr::Ascii("browine"), &["brownie"]),
            (
                dictgen::InsensitiveStr::Ascii("univerally"),
                &["universally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreigster"),
                &["unregister"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipluated"),
                &["manipulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("togetheer"), &["together"]),
            (dictgen::InsensitiveStr::Ascii("burried"), &["buried"]),
            (dictgen::InsensitiveStr::Ascii("actualiy"), &["actuality"]),
            (dictgen::InsensitiveStr::Ascii("spartsn"), &["spartans"]),
            (dictgen::InsensitiveStr::Ascii("causeing"), &["causing"]),
            (dictgen::InsensitiveStr::Ascii("ponint"), &["point"]),
            (dictgen::InsensitiveStr::Ascii("audcaity"), &["audacity"]),
            (dictgen::InsensitiveStr::Ascii("tryng"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("reguarlize"),
                &["regularize"],
            ),
            (dictgen::InsensitiveStr::Ascii("extreamly"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("anouncements"),
                &["announcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfuncion"),
                &["dysfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("quitly"), &["quietly"]),
            (dictgen::InsensitiveStr::Ascii("tagrets"), &["targets"]),
            (
                dictgen::InsensitiveStr::Ascii("unreprociblbe"),
                &["unreproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaction"),
                &["reaction", "relation", "redaction", "relocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compativility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("salaris"), &["salaries"]),
            (
                dictgen::InsensitiveStr::Ascii("calcuating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretting"),
                &["misinterpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brigthened"),
                &["brightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("detal"), &["detail"]),
            (
                dictgen::InsensitiveStr::Ascii("consequenses"),
                &["consequences"],
            ),
            (dictgen::InsensitiveStr::Ascii("almightly"), &["almighty"]),
            (
                dictgen::InsensitiveStr::Ascii("professionnal"),
                &["professional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administored"),
                &["administer", "administered"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipeliens"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("inclinato"),
                &["inclination"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinosar"), &["dinosaur"]),
            (dictgen::InsensitiveStr::Ascii("nawshus"), &["nauseous"]),
            (dictgen::InsensitiveStr::Ascii("blanketts"), &["blankets"]),
            (
                dictgen::InsensitiveStr::Ascii("callibration"),
                &["calibration"],
            ),
            (dictgen::InsensitiveStr::Ascii("userrnames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("wholeheartdly"),
                &["wholeheartedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("assemmbly"), &["assembly"]),
            (dictgen::InsensitiveStr::Ascii("indicte"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("follewer"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("depdencent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("alternarive"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("gloablly"), &["globally"]),
            (
                dictgen::InsensitiveStr::Ascii("functioalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("dleays"), &["delays"]),
            (dictgen::InsensitiveStr::Ascii("lanuching"), &["launching"]),
            (dictgen::InsensitiveStr::Ascii("broswing"), &["browsing"]),
            (dictgen::InsensitiveStr::Ascii("reutnrs"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("genertions"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("thingking"), &["thinking"]),
            (dictgen::InsensitiveStr::Ascii("accuarcy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("enrgy"), &["energy"]),
            (
                dictgen::InsensitiveStr::Ascii("incosistencies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubsrcibed"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("feezes"), &["freezes"]),
            (dictgen::InsensitiveStr::Ascii("apreture"), &["aperture"]),
            (
                dictgen::InsensitiveStr::Ascii("achieveble"),
                &["achievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("popualted"), &["populated"]),
            (dictgen::InsensitiveStr::Ascii("caractere"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("proessing"), &["processing"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscribade"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("cmmanding"), &["commanding"]),
            (
                dictgen::InsensitiveStr::Ascii("propulaios"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("confernce"), &["conference"]),
            (dictgen::InsensitiveStr::Ascii("buggest"), &["biggest"]),
            (
                dictgen::InsensitiveStr::Ascii("functtionalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("motiation"), &["motivation"]),
            (
                dictgen::InsensitiveStr::Ascii("hopefull"),
                &["hopeful", "hopefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("baase"), &["abase", "base"]),
            (dictgen::InsensitiveStr::Ascii("foloowin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("carnavors"), &["carnivores"]),
            (
                dictgen::InsensitiveStr::Ascii("sythesises"),
                &["synthesises"],
            ),
            (dictgen::InsensitiveStr::Ascii("terrform"), &["terraform"]),
            (dictgen::InsensitiveStr::Ascii("reservouir"), &["reservoir"]),
            (dictgen::InsensitiveStr::Ascii("lastr"), &["last"]),
            (
                dictgen::InsensitiveStr::Ascii("religiosly"),
                &["religiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formadible"),
                &["formidable"],
            ),
            (dictgen::InsensitiveStr::Ascii("wieh"), &["view"]),
            (dictgen::InsensitiveStr::Ascii("samll"), &["small"]),
            (
                dictgen::InsensitiveStr::Ascii("repsectfully"),
                &["respectfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("androind"), &["android"]),
            (
                dictgen::InsensitiveStr::Ascii("singelplayer"),
                &["singleplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("decalred"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("interrum"), &["interim"]),
            (
                dictgen::InsensitiveStr::Ascii("underwarter"),
                &["underwater"],
            ),
            (dictgen::InsensitiveStr::Ascii("avarageing"), &["averaging"]),
            (
                dictgen::InsensitiveStr::Ascii("gauntlettes"),
                &["gauntlets"],
            ),
            (dictgen::InsensitiveStr::Ascii("abandones"), &["abandons"]),
            (dictgen::InsensitiveStr::Ascii("wrappr"), &["wrapper"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminoti"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubing"),
                &["disturbing"],
            ),
            (dictgen::InsensitiveStr::Ascii("searchd"), &["searched"]),
            (dictgen::InsensitiveStr::Ascii("becayse"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("aeropsace"), &["aerospace"]),
            (
                dictgen::InsensitiveStr::Ascii("verificato"),
                &["verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermarkten"),
                &["supermarket"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("servent"),
                &["servant", "serpent", "fervent"],
            ),
            (dictgen::InsensitiveStr::Ascii("fornat"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("identifers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("picutre"), &["picture"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiase"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifcated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("consuemr"), &["consumer"]),
            (dictgen::InsensitiveStr::Ascii("coccupied"), &["occupied"]),
            (dictgen::InsensitiveStr::Ascii("nextwork"), &["network"]),
            (dictgen::InsensitiveStr::Ascii("entoty"), &["entity"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidante"),
                &["invalidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desposition"),
                &["disposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emition"),
                &["emission", "emotion"],
            ),
            (dictgen::InsensitiveStr::Ascii("comapison"), &["comparison"]),
            (
                dictgen::InsensitiveStr::Ascii("replase"),
                &["replaces", "replace", "relapse", "rephase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncesnored"),
                &["uncensored"],
            ),
            (dictgen::InsensitiveStr::Ascii("discribed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("hypothethical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciritical"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("thoughput"), &["throughput"]),
            (
                dictgen::InsensitiveStr::Ascii("intepretor"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("potetinal"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("uniformy"),
                &["uniformly", "uniform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepondance"),
                &["preponderance"],
            ),
            (dictgen::InsensitiveStr::Ascii("speeedy"), &["speedy"]),
            (dictgen::InsensitiveStr::Ascii("satic"), &["static"]),
            (dictgen::InsensitiveStr::Ascii("effient"), &["efficient"]),
            (
                dictgen::InsensitiveStr::Ascii("spradsheet"),
                &["spreadsheet"],
            ),
            (dictgen::InsensitiveStr::Ascii("fineses"), &["finesse"]),
            (
                dictgen::InsensitiveStr::Ascii("sophosticated"),
                &["sophisticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sillabuses"),
                &["syllabuses"],
            ),
            (dictgen::InsensitiveStr::Ascii("restult"), &["result"]),
            (dictgen::InsensitiveStr::Ascii("ussername"), &["username"]),
            (dictgen::InsensitiveStr::Ascii("indcators"), &["indicators"]),
            (
                dictgen::InsensitiveStr::Ascii("hypotheticly"),
                &["hypothetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vertification"),
                &["certification", "verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciricularize"),
                &["circularize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deseralized"),
                &["deserialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bothe"), &["both"]),
            (
                dictgen::InsensitiveStr::Ascii("definutely"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("perseeded"), &["preceded"]),
            (dictgen::InsensitiveStr::Ascii("entitiys"), &["entities"]),
            (dictgen::InsensitiveStr::Ascii("srcoll"), &["scroll"]),
            (
                dictgen::InsensitiveStr::Ascii("wearnings"),
                &["warnings", "earnings", "yearnings"],
            ),
            (dictgen::InsensitiveStr::Ascii("preprare"), &["prepare"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("hinduisim"), &["hinduism"]),
            (dictgen::InsensitiveStr::Ascii("stutus"), &["status"]),
            (dictgen::InsensitiveStr::Ascii("instealler"), &["installer"]),
            (
                dictgen::InsensitiveStr::Ascii("procucing"),
                &["procuring", "producing"],
            ),
            (dictgen::InsensitiveStr::Ascii("learnd"), &["learned"]),
            (
                dictgen::InsensitiveStr::Ascii("authentictes"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiecne"),
                &["experience"],
            ),
            (dictgen::InsensitiveStr::Ascii("yeterday"), &["yesterday"]),
            (
                dictgen::InsensitiveStr::Ascii("coordanate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("iplements"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("definiations"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("irrelvant"), &["irrelevant"]),
            (dictgen::InsensitiveStr::Ascii("pinetrest"), &["pinterest"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("commongly"), &["commonly"]),
            (dictgen::InsensitiveStr::Ascii("vengance"), &["vengeance"]),
            (dictgen::InsensitiveStr::Ascii("sbaolute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("applciation"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dpeendency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuetable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuesly"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accellerate"),
                &["accelerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("snowlfake"), &["snowflake"]),
            (dictgen::InsensitiveStr::Ascii("monnths"), &["months"]),
            (
                dictgen::InsensitiveStr::Ascii("ternimally"),
                &["terminally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beggin"),
                &["begin", "begging"],
            ),
            (dictgen::InsensitiveStr::Ascii("fushing"), &["flushing"]),
            (dictgen::InsensitiveStr::Ascii("fassade"), &["facade"]),
            (
                dictgen::InsensitiveStr::Ascii("aplitude"),
                &["amplitude", "aptitude"],
            ),
            (dictgen::InsensitiveStr::Ascii("carosels"), &["carousels"]),
            (
                dictgen::InsensitiveStr::Ascii("nieghbourhoods"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("vindicte"), &["vindictive"]),
            (dictgen::InsensitiveStr::Ascii("prooper"), &["proper"]),
            (
                dictgen::InsensitiveStr::Ascii("interogation"),
                &["interrogation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scroolbars"),
                &["scrollbars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cononicalizes"),
                &["canonicalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("prmitive"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("depseration"),
                &["desperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fornesic"), &["forensic"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucccessfully"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indispensibility"),
                &["indispensability"],
            ),
            (dictgen::InsensitiveStr::Ascii("esporst"), &["esports"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliations"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensability"),
                &["sensibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("muti"), &["multi"]),
            (dictgen::InsensitiveStr::Ascii("nmbering"), &["numbering"]),
            (dictgen::InsensitiveStr::Ascii("guardianis"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("copmile"), &["compile"]),
            (dictgen::InsensitiveStr::Ascii("audactiy"), &["audacity"]),
            (dictgen::InsensitiveStr::Ascii("mch"), &["much"]),
            (dictgen::InsensitiveStr::Ascii("guarntee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("metatada"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("userrspace"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("cacluation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("campiagn"), &["campaign"]),
            (dictgen::InsensitiveStr::Ascii("soliliquy"), &["soliloquy"]),
            (
                dictgen::InsensitiveStr::Ascii("cateogrical"),
                &["categorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("singalled"), &["signalled"]),
            (
                dictgen::InsensitiveStr::Ascii("complimentje"),
                &["complimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phramacist"),
                &["pharmacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("createion"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("straigths"), &["straights"]),
            (
                dictgen::InsensitiveStr::Ascii("organizatinal"),
                &["organizational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adevntures"),
                &["adventures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phisiological"),
                &["physiological"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifyng"), &["specifying"]),
            (dictgen::InsensitiveStr::Ascii("leage"), &["league"]),
            (
                dictgen::InsensitiveStr::Ascii("responsds"),
                &["responds", "responses"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistaek"), &["mistake"]),
            (
                dictgen::InsensitiveStr::Ascii("convsation"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properry"),
                &["property", "properly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physicallity"),
                &["physically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esponse"),
                &["response", "espouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approbiately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibillity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("captin"), &["captain"]),
            (
                dictgen::InsensitiveStr::Ascii("confogures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("establishs"),
                &["establishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("navitation"),
                &["navigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstaled"),
                &["uninstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successsful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahnging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("controllng"),
                &["controlling"],
            ),
            (dictgen::InsensitiveStr::Ascii("origianls"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("helictoper"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleean"), &["clean"]),
            (dictgen::InsensitiveStr::Ascii("camboda"), &["cambodia"]),
            (
                dictgen::InsensitiveStr::Ascii("schma"),
                &["schema", "schwa"],
            ),
            (dictgen::InsensitiveStr::Ascii("untraind"), &["untrained"]),
            (dictgen::InsensitiveStr::Ascii("omitt"), &["omit"]),
            (dictgen::InsensitiveStr::Ascii("reveiwing"), &["reviewing"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("stormde"), &["stormed"]),
            (
                dictgen::InsensitiveStr::Ascii("kolonization"),
                &["colonization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unforutunate"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("inrested"), &["inserted"]),
            (
                dictgen::InsensitiveStr::Ascii("apsirations"),
                &["aspirations"],
            ),
            (dictgen::InsensitiveStr::Ascii("othr"), &["other"]),
            (
                dictgen::InsensitiveStr::Ascii("insuffiency"),
                &["insufficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadvertendly"),
                &["inadvertently"],
            ),
            (dictgen::InsensitiveStr::Ascii("decoraded"), &["decorated"]),
            (
                dictgen::InsensitiveStr::Ascii("coditional"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderdtanding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("probbly"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("notifiying"), &["notifying"]),
            (dictgen::InsensitiveStr::Ascii("mdoel"), &["model"]),
            (dictgen::InsensitiveStr::Ascii("bourgeoius"), &["bourgeois"]),
            (dictgen::InsensitiveStr::Ascii("assihnee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("asbestoast"), &["asbestos"]),
            (
                dictgen::InsensitiveStr::Ascii("formalisin"),
                &["formalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("iunior"), &["junior"]),
            (
                dictgen::InsensitiveStr::Ascii("separatley"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissallow"), &["disallow"]),
            (dictgen::InsensitiveStr::Ascii("trakers"), &["trackers"]),
            (dictgen::InsensitiveStr::Ascii("qtuie"), &["quite", "quiet"]),
            (dictgen::InsensitiveStr::Ascii("ryou"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonizing"),
                &["synchronizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("univercity"),
                &["university"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaiants"), &["variants"]),
            (dictgen::InsensitiveStr::Ascii("grea"), &["great"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcualted"),
                &["recalculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classsically"),
                &["classically"],
            ),
            (dictgen::InsensitiveStr::Ascii("jeapardy"), &["jeopardy"]),
            (dictgen::InsensitiveStr::Ascii("questio"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("appointmet"),
                &["appointments", "appointment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresers"),
                &["compressors"],
            ),
            (dictgen::InsensitiveStr::Ascii("reseponse"), &["response"]),
            (
                dictgen::InsensitiveStr::Ascii("originiality"),
                &["originality"],
            ),
            (dictgen::InsensitiveStr::Ascii("heronie"), &["heroine"]),
            (
                dictgen::InsensitiveStr::Ascii("chronologicly"),
                &["chronological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approproximation"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directtories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("arborting"), &["aborting"]),
            (dictgen::InsensitiveStr::Ascii("fesiable"), &["feasible"]),
            (dictgen::InsensitiveStr::Ascii("rescuse"), &["rescues"]),
            (
                dictgen::InsensitiveStr::Ascii("priortisation"),
                &["prioritisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaitng"), &["animating"]),
            (
                dictgen::InsensitiveStr::Ascii("deprecatin"),
                &["deprecating", "deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("followng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("distrct"), &["district"]),
            (dictgen::InsensitiveStr::Ascii("modfies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("gunsligner"),
                &["gunslinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simmetrical"),
                &["symmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("satistics"), &["statistics"]),
            (
                dictgen::InsensitiveStr::Ascii("ressurrects"),
                &["resurrects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conferencin"),
                &["conferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoducing"),
                &["reproducing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reanming"), &["renaming"]),
            (dictgen::InsensitiveStr::Ascii("interrput"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("fimily"), &["family"]),
            (dictgen::InsensitiveStr::Ascii("japanes"), &["japanese"]),
            (dictgen::InsensitiveStr::Ascii("premeire"), &["premiere"]),
            (
                dictgen::InsensitiveStr::Ascii("tranitioning"),
                &["transitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocos"), &["allocs"]),
            (dictgen::InsensitiveStr::Ascii("accepable"), &["acceptable"]),
            (dictgen::InsensitiveStr::Ascii("spoonfulls"), &["spoonfuls"]),
            (
                dictgen::InsensitiveStr::Ascii("approxamating"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relatioship"),
                &["relationship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubitors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("songles"),
                &["singles", "dongles"],
            ),
            (dictgen::InsensitiveStr::Ascii("equl"), &["equal"]),
            (
                dictgen::InsensitiveStr::Ascii("verifiaction"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteartes"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("deceber"), &["december"]),
            (
                dictgen::InsensitiveStr::Ascii("experieces"),
                &["experiences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detetes"),
                &["deletes", "detects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depoloyment"),
                &["deployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("ourcing"), &["sourcing"]),
            (
                dictgen::InsensitiveStr::Ascii("epxansions"),
                &["expansions"],
            ),
            (dictgen::InsensitiveStr::Ascii("paniced"), &["panicked"]),
            (dictgen::InsensitiveStr::Ascii("ignorgg"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("finilisation"),
                &["finalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("failling"), &["failing"]),
            (dictgen::InsensitiveStr::Ascii("contaier"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("accorging"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("arrestos"), &["arrests"]),
            (dictgen::InsensitiveStr::Ascii("reavealed"), &["revealed"]),
            (dictgen::InsensitiveStr::Ascii("condered"), &["considered"]),
            (dictgen::InsensitiveStr::Ascii("potraits"), &["portraits"]),
            (
                dictgen::InsensitiveStr::Ascii("pipeling"),
                &["pipelining", "pipeline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalueation"),
                &["evaluation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hilighters"),
                &["highlighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("symble"), &["symbol"]),
            (dictgen::InsensitiveStr::Ascii("boredoom"), &["boredom"]),
            (
                dictgen::InsensitiveStr::Ascii("portugeuse"),
                &["portuguese"],
            ),
            (dictgen::InsensitiveStr::Ascii("morevoer"), &["moreover"]),
            (dictgen::InsensitiveStr::Ascii("tracsodes"), &["transcodes"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactuters"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappeared"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("owerflow"), &["overflow"]),
            (dictgen::InsensitiveStr::Ascii("semgents"), &["segments"]),
            (
                dictgen::InsensitiveStr::Ascii("paermissions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("violoations"),
                &["violations"],
            ),
            (dictgen::InsensitiveStr::Ascii("shortkuts"), &["shortcuts"]),
            (
                dictgen::InsensitiveStr::Ascii("incompetance"),
                &["incompetence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("angshisness"),
                &["anxiousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("unanomus"), &["unanimous"]),
            (
                dictgen::InsensitiveStr::Ascii("individuelly"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functiional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agressiveness"),
                &["aggressiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("tahnkful"), &["thankful"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitilized"),
                &["reinitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omisience"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deductable"),
                &["deductible"],
            ),
            (dictgen::InsensitiveStr::Ascii("deelte"), &["delete"]),
            (
                dictgen::InsensitiveStr::Ascii("alled"),
                &["called", "allied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contiously"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheerlearder"),
                &["cheerleader"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depricating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("scruitny"), &["scrutiny"]),
            (dictgen::InsensitiveStr::Ascii("debuging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("alphapet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("statulines"),
                &["statuslines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condradictory"),
                &["contradictory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conetivity"),
                &["connectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrementing"),
                &["decrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("follewong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("backbacking"),
                &["backpacking"],
            ),
            (dictgen::InsensitiveStr::Ascii("poinyent"), &["poignant"]),
            (
                dictgen::InsensitiveStr::Ascii("acitvating"),
                &["activating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalistc"),
                &["nationalistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("emcompass"), &["encompass"]),
            (dictgen::InsensitiveStr::Ascii("appy"), &["apply"]),
            (
                dictgen::InsensitiveStr::Ascii("appriceating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("intensitiy"), &["intensity"]),
            (
                dictgen::InsensitiveStr::Ascii("intertwinded"),
                &["intertwined"],
            ),
            (dictgen::InsensitiveStr::Ascii("adultary"), &["adultery"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondd"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("remot"), &["remote"]),
            (dictgen::InsensitiveStr::Ascii("poligons"), &["polygons"]),
            (dictgen::InsensitiveStr::Ascii("cadidates"), &["candidates"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangmeents"),
                &["rearrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("allwoing"), &["allowing"]),
            (
                dictgen::InsensitiveStr::Ascii("unequivalent"),
                &["nonequivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkeed"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("respektive"),
                &["respective"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusomizer"), &["customizer"]),
            (
                dictgen::InsensitiveStr::Ascii("dismension"),
                &["dimension", "dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("dignoses"), &["diagnoses"]),
            (dictgen::InsensitiveStr::Ascii("azmiths"), &["azimuths"]),
            (dictgen::InsensitiveStr::Ascii("chekcers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("favilitation"),
                &["facilitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("phtread"), &["pthread"]),
            (
                dictgen::InsensitiveStr::Ascii("powerhourse"),
                &["powerhouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoourcing"),
                &["resourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannabolism"),
                &["cannibalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("encoeders"), &["encoders"]),
            (dictgen::InsensitiveStr::Ascii("evauluate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("pratcise"), &["practise"]),
            (
                dictgen::InsensitiveStr::Ascii("blochchain"),
                &["blockchain"],
            ),
            (dictgen::InsensitiveStr::Ascii("persews"), &["pursues"]),
            (dictgen::InsensitiveStr::Ascii("generaged"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("unquotin"), &["unquoting"]),
            (
                dictgen::InsensitiveStr::Ascii("resuced"),
                &["rescued", "reduced"],
            ),
            (dictgen::InsensitiveStr::Ascii("installls"), &["installs"]),
            (
                dictgen::InsensitiveStr::Ascii("exixting"),
                &["existing", "exiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmenting"),
                &["commenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifiier"),
                &["classifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configursed"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarmatic"),
                &["programmatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptve"),
                &["descriptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("preorderes"), &["preorders"]),
            (dictgen::InsensitiveStr::Ascii("opearate"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("warrante"), &["warranties"]),
            (dictgen::InsensitiveStr::Ascii("settup"), &["setup"]),
            (dictgen::InsensitiveStr::Ascii("laso"), &["also", "lasso"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticaed"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("secconds"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("hypocrits"), &["hypocrites"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisiations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimesions"), &["dimensions"]),
            (
                dictgen::InsensitiveStr::Ascii("analyitcal"),
                &["analytical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininistration"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("enocders"), &["encoders"]),
            (
                dictgen::InsensitiveStr::Ascii("reestructures"),
                &["restructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inplications"),
                &["implications"],
            ),
            (dictgen::InsensitiveStr::Ascii("prohibt"), &["prohibit"]),
            (
                dictgen::InsensitiveStr::Ascii("alognments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("netropolitan"),
                &["metropolitan"],
            ),
            (dictgen::InsensitiveStr::Ascii("patrolls"), &["patrols"]),
            (dictgen::InsensitiveStr::Ascii("exlcusive"), &["exclusive"]),
            (dictgen::InsensitiveStr::Ascii("querries"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("comvention"),
                &["convention"],
            ),
            (dictgen::InsensitiveStr::Ascii("rembember"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("capsuls"), &["capsules"]),
            (dictgen::InsensitiveStr::Ascii("ethnicaly"), &["ethnically"]),
            (dictgen::InsensitiveStr::Ascii("samori"), &["samurai"]),
            (dictgen::InsensitiveStr::Ascii("furutre"), &["future"]),
            (dictgen::InsensitiveStr::Ascii("brownis"), &["brownies"]),
            (
                dictgen::InsensitiveStr::Ascii("downtoking"),
                &["downvoting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questioing"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("diplay"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("configuarion"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("tratior"), &["traitor"]),
            (dictgen::InsensitiveStr::Ascii("surregat"), &["surrogate"]),
            (
                dictgen::InsensitiveStr::Ascii("compainons"),
                &["companions"],
            ),
            (dictgen::InsensitiveStr::Ascii("initital"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("reapeating"), &["repeating"]),
            (dictgen::InsensitiveStr::Ascii("skepitcal"), &["skeptical"]),
            (dictgen::InsensitiveStr::Ascii("multuple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("retransmissons"),
                &["retransmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhaposdy"), &["rhapsody"]),
            (dictgen::InsensitiveStr::Ascii("functiion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("vibratin"), &["vibration"]),
            (dictgen::InsensitiveStr::Ascii("locaized"), &["localized"]),
            (dictgen::InsensitiveStr::Ascii("racistes"), &["racists"]),
            (
                dictgen::InsensitiveStr::Ascii("belittlling"),
                &["belittling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miserablely"),
                &["miserably"],
            ),
            (dictgen::InsensitiveStr::Ascii("incluided"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("donwvoting"),
                &["downvoting"],
            ),
            (dictgen::InsensitiveStr::Ascii("deinite"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("deisgning"), &["designing"]),
            (dictgen::InsensitiveStr::Ascii("wuth"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("reutnred"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("orhtogonality"),
                &["orthogonality"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfences"), &["offences"]),
            (dictgen::InsensitiveStr::Ascii("iself"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("disposible"),
                &["disposable"],
            ),
            (dictgen::InsensitiveStr::Ascii("recangles"), &["rectangles"]),
            (dictgen::InsensitiveStr::Ascii("expasion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("plarised"), &["polarised"]),
            (dictgen::InsensitiveStr::Ascii("whench"), &["whence"]),
            (dictgen::InsensitiveStr::Ascii("infroms"), &["informs"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeuthood"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfied"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("souvinneers"),
                &["souvenirs"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininize"), &["minimize"]),
            (dictgen::InsensitiveStr::Ascii("guardianes"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("coreolis"), &["coriolis"]),
            (
                dictgen::InsensitiveStr::Ascii("carinomially"),
                &["ceremonially"],
            ),
            (dictgen::InsensitiveStr::Ascii("enitties"), &["entities"]),
            (dictgen::InsensitiveStr::Ascii("assumbler"), &["assembler"]),
            (
                dictgen::InsensitiveStr::Ascii("throtting"),
                &["throttling", "trotting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deslects"),
                &["deselects", "deflects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integraded"),
                &["integrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geoemtries"),
                &["geometries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affact"),
                &["affect", "effect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransacciones"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("hidded"), &["hidden"]),
            (dictgen::InsensitiveStr::Ascii("uniqness"), &["uniqueness"]),
            (dictgen::InsensitiveStr::Ascii("premius"), &["premiums"]),
            (dictgen::InsensitiveStr::Ascii("inertion"), &["insertion"]),
            (dictgen::InsensitiveStr::Ascii("honory"), &["honorary"]),
            (dictgen::InsensitiveStr::Ascii("numvbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("questionned"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("reaaly"), &["really"]),
            (dictgen::InsensitiveStr::Ascii("conplete"), &["complete"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrifacing"),
                &["sacrificing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscibing"),
                &["unsubscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminscently"),
                &["reminiscently"],
            ),
            (dictgen::InsensitiveStr::Ascii("messge"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("survivin"), &["surviving"]),
            (dictgen::InsensitiveStr::Ascii("vallies"), &["valleys"]),
            (
                dictgen::InsensitiveStr::Ascii("coordintate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruncing"), &["crunching"]),
            (
                dictgen::InsensitiveStr::Ascii("dermatalogist"),
                &["dermatologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepndents"),
                &["independents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparantly"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierarchichally"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resemebling"),
                &["resembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiail"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("modifdy"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("subproceess"),
                &["subprocess"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminatti"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("rivision"), &["revision"]),
            (
                dictgen::InsensitiveStr::Ascii("headrom"),
                &["headroom", "bedroom"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visibiliy"),
                &["visibility", "visibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curriculim"),
                &["curriculum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsadized"),
                &["subsidized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appripriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("facsists"), &["fascists"]),
            (
                dictgen::InsensitiveStr::Ascii("missunderstanding"),
                &["misunderstanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("extented"), &["extended"]),
            (dictgen::InsensitiveStr::Ascii("callser"), &["caller"]),
            (dictgen::InsensitiveStr::Ascii("aways"), &["always", "away"]),
            (dictgen::InsensitiveStr::Ascii("udpated"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("clipoing"), &["clipping"]),
            (dictgen::InsensitiveStr::Ascii("magnetis"), &["magnets"]),
            (dictgen::InsensitiveStr::Ascii("arrane"), &["arrange"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliazers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentris"), &["sentries"]),
            (dictgen::InsensitiveStr::Ascii("intevals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucccessful"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincidantal"),
                &["coincidental"],
            ),
            (dictgen::InsensitiveStr::Ascii("desorder"), &["disorder"]),
            (dictgen::InsensitiveStr::Ascii("pershisly"), &["preciously"]),
            (
                dictgen::InsensitiveStr::Ascii("mutlipliers"),
                &["multipliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrespectul"),
                &["disrespectful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correesponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("proxyed"), &["proxied"]),
            (
                dictgen::InsensitiveStr::Ascii("experements"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("clokc"), &["clock"]),
            (
                dictgen::InsensitiveStr::Ascii("subcontraint"),
                &["subconstraint"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyjacking"), &["hijacking"]),
            (
                dictgen::InsensitiveStr::Ascii("renegerate"),
                &["regenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("uername"), &["username"]),
            (dictgen::InsensitiveStr::Ascii("uplods"), &["uploads"]),
            (dictgen::InsensitiveStr::Ascii("saphires"), &["sapphires"]),
            (dictgen::InsensitiveStr::Ascii("preferered"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("thumbnals"), &["thumbnails"]),
            (dictgen::InsensitiveStr::Ascii("leting"), &["letting"]),
            (dictgen::InsensitiveStr::Ascii("initalese"), &["initialese"]),
            (dictgen::InsensitiveStr::Ascii("rentime"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("appicative"),
                &["applicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("formatin"), &["formation"]),
            (dictgen::InsensitiveStr::Ascii("certian"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("raspbery"), &["raspberry"]),
            (dictgen::InsensitiveStr::Ascii("declarated"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("plcaeholder"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("directix"), &["directx"]),
            (dictgen::InsensitiveStr::Ascii("coommittee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("mountian"), &["mountain"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaer"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgrunted"),
                &["disgruntled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyerarchical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentaing"),
                &["concatenating"],
            ),
            (dictgen::InsensitiveStr::Ascii("aircaft"), &["aircraft"]),
            (dictgen::InsensitiveStr::Ascii("nedd"), &["need"]),
            (
                dictgen::InsensitiveStr::Ascii("potentiallly"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outperfoem"),
                &["outperform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("africaners"),
                &["africans", "afrikaners", "africanders", "afrikanders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creativley"),
                &["creatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitalizing"),
                &["reinitializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenets"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("viscreal"), &["visceral"]),
            (dictgen::InsensitiveStr::Ascii("previus"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("shystemwindow"),
                &["systemwindow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limititations"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("corerctly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("stlaker"), &["stalker"]),
            (
                dictgen::InsensitiveStr::Ascii("brazilains"),
                &["brazilians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniuplative"),
                &["manipulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("xyophone"), &["xylophone"]),
            (dictgen::InsensitiveStr::Ascii("spcies"), &["species"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbehood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("autmaton"), &["automaton"]),
            (
                dictgen::InsensitiveStr::Ascii("partiularly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detinations"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("asymetriy"), &["asymmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("introvered"),
                &["introverted"],
            ),
            (dictgen::InsensitiveStr::Ascii("moustahce"), &["moustache"]),
            (dictgen::InsensitiveStr::Ascii("calllbacks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("preseleting"),
                &["preselecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representaiton"),
                &["representations", "representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("shutdownm"), &["shutdown"]),
            (
                dictgen::InsensitiveStr::Ascii("knowledgabel"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deintialize"),
                &["deinitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assimilant"),
                &["assimilate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continus"),
                &["continues", "continue", "continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cononicalization"),
                &["canonicalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inapporpriate"),
                &["inappropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constain"),
                &["constrain", "contain"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirectors"), &["directors"]),
            (dictgen::InsensitiveStr::Ascii("paramaeter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("reinfocement"),
                &["reinforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("prevend"), &["prevent"]),
            (dictgen::InsensitiveStr::Ascii("scandalos"), &["scandals"]),
            (dictgen::InsensitiveStr::Ascii("worthles"), &["worthless"]),
            (dictgen::InsensitiveStr::Ascii("diagonse"), &["diagnose"]),
            (
                dictgen::InsensitiveStr::Ascii("downloadbel"),
                &["downloadable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narled"),
                &["gnarled", "nailed", "narked", "snarled"],
            ),
            (dictgen::InsensitiveStr::Ascii("collsions"), &["collisions"]),
            (dictgen::InsensitiveStr::Ascii("cammander"), &["commander"]),
            (
                dictgen::InsensitiveStr::Ascii("expentitures"),
                &["expenditures"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrythes"), &["writhes"]),
            (dictgen::InsensitiveStr::Ascii("singulaire"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("paceholder"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("thiniks"), &["thinks"]),
            (dictgen::InsensitiveStr::Ascii("worhtless"), &["worthless"]),
            (
                dictgen::InsensitiveStr::Ascii("contracters"),
                &["contractors"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignmen"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("acticate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("selectivly"),
                &["selectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("curnel"), &["colonel"]),
            (dictgen::InsensitiveStr::Ascii("scheduld"), &["scheduled"]),
            (
                dictgen::InsensitiveStr::Ascii("descritpros"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbility"), &["inability"]),
            (dictgen::InsensitiveStr::Ascii("bacends"), &["backends"]),
            (
                dictgen::InsensitiveStr::Ascii("apretiated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("shouws"), &["shows"]),
            (dictgen::InsensitiveStr::Ascii("unreliabe"), &["unreliable"]),
            (dictgen::InsensitiveStr::Ascii("gocde"), &["gcode"]),
            (
                dictgen::InsensitiveStr::Ascii("indepentent"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulitated"), &["mutilated"]),
            (dictgen::InsensitiveStr::Ascii("insistant"), &["insistent"]),
            (
                dictgen::InsensitiveStr::Ascii("meaurement"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatably"),
                &["incompatibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("deinition"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("decompresed"),
                &["decompressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expierienced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negligance"),
                &["negligence"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntehses"), &["syntheses"]),
            (dictgen::InsensitiveStr::Ascii("anulled"), &["annulled"]),
            (
                dictgen::InsensitiveStr::Ascii("synchornization"),
                &["synchronization"],
            ),
            (dictgen::InsensitiveStr::Ascii("phanthom"), &["phantom"]),
            (dictgen::InsensitiveStr::Ascii("speficiic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("greeneer"), &["greener"]),
            (dictgen::InsensitiveStr::Ascii("meditato"), &["meditation"]),
            (
                dictgen::InsensitiveStr::Ascii("informatie"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opperational"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("devaite"), &["deviate"]),
            (dictgen::InsensitiveStr::Ascii("dusbtep"), &["dubstep"]),
            (
                dictgen::InsensitiveStr::Ascii("procuce"),
                &["procure", "produce"],
            ),
            (dictgen::InsensitiveStr::Ascii("alreeady"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("spectularly"),
                &["spectacularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("speelling"), &["spelling"]),
            (dictgen::InsensitiveStr::Ascii("superioare"), &["superior"]),
            (dictgen::InsensitiveStr::Ascii("anaytic"), &["analytic"]),
            (dictgen::InsensitiveStr::Ascii("mointored"), &["monitored"]),
            (
                dictgen::InsensitiveStr::Ascii("idealisitic"),
                &["idealistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("retore"), &["restore"]),
            (dictgen::InsensitiveStr::Ascii("shfted"), &["shifted"]),
            (
                dictgen::InsensitiveStr::Ascii("renaissaince"),
                &["renaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("effctive"), &["effective"]),
            (
                dictgen::InsensitiveStr::Ascii("opportunistly"),
                &["opportunistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminats"), &["terminates"]),
            (dictgen::InsensitiveStr::Ascii("foreing"), &["foreign"]),
            (
                dictgen::InsensitiveStr::Ascii("miscallenaous"),
                &["miscellaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dnamically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhenced"), &["enhanced"]),
            (
                dictgen::InsensitiveStr::Ascii("consistendly"),
                &["consistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("furthrest"), &["furthest"]),
            (dictgen::InsensitiveStr::Ascii("apparence"), &["appearance"]),
            (
                dictgen::InsensitiveStr::Ascii("desychronize"),
                &["desynchronize"],
            ),
            (dictgen::InsensitiveStr::Ascii("immeadiate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("mapings"), &["mappings"]),
            (
                dictgen::InsensitiveStr::Ascii("initialistion"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghthy"), &["lengthy"]),
            (dictgen::InsensitiveStr::Ascii("decrpyt"), &["decrypt"]),
            (dictgen::InsensitiveStr::Ascii("visiably"), &["visibly"]),
            (dictgen::InsensitiveStr::Ascii("emiitter"), &["emitter"]),
            (dictgen::InsensitiveStr::Ascii("andelabra"), &["candelabra"]),
            (
                dictgen::InsensitiveStr::Ascii("accociation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiessing"),
                &["acquiescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sweathsirt"),
                &["sweatshirt"],
            ),
            (dictgen::InsensitiveStr::Ascii("dolphis"), &["dolphins"]),
            (
                dictgen::InsensitiveStr::Ascii("mcroscopies"),
                &["microscopies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterprested"),
                &["misinterpret"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanentally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("matierals"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("preoccuption"),
                &["preoccupation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reviere"), &["reviewer"]),
            (dictgen::InsensitiveStr::Ascii("ntegrates"), &["integrates"]),
            (
                dictgen::InsensitiveStr::Ascii("extraciton"),
                &["extraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rachet"),
                &["ratchet", "sachet", "cachet", "racket"],
            ),
            (dictgen::InsensitiveStr::Ascii("eligable"), &["eligible"]),
            (dictgen::InsensitiveStr::Ascii("sanbox"), &["sandbox"]),
            (
                dictgen::InsensitiveStr::Ascii("vegetarianas"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("idaeidae"), &["idea"]),
            (dictgen::InsensitiveStr::Ascii("benghzai"), &["benghazi"]),
            (
                dictgen::InsensitiveStr::Ascii("deducatble"),
                &["deductible"],
            ),
            (dictgen::InsensitiveStr::Ascii("activitis"), &["activities"]),
            (
                dictgen::InsensitiveStr::Ascii("organisaions"),
                &["organisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("awfull"),
                &["awful", "awfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("detemines"), &["determines"]),
            (dictgen::InsensitiveStr::Ascii("normlise"), &["normalise"]),
            (
                dictgen::InsensitiveStr::Ascii("expectating"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressent"), &["present"]),
            (dictgen::InsensitiveStr::Ascii("evluates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("roviding"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("cirumvented"),
                &["circumvented"],
            ),
            (dictgen::InsensitiveStr::Ascii("insertsect"), &["intersect"]),
            (dictgen::InsensitiveStr::Ascii("thje"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("contentos"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("requestied"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("agregating"),
                &["aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("multilpe"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("statememts"),
                &["statements"],
            ),
            (dictgen::InsensitiveStr::Ascii("athenean"), &["athenian"]),
            (
                dictgen::InsensitiveStr::Ascii("experiementation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rsicv"), &["riscv"]),
            (dictgen::InsensitiveStr::Ascii("catloaged"), &["cataloged"]),
            (
                dictgen::InsensitiveStr::Ascii("assymetrically"),
                &["asymmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("unwraped"), &["unwrapped"]),
            (dictgen::InsensitiveStr::Ascii("strarted"), &["started"]),
            (
                dictgen::InsensitiveStr::Ascii("humillation"),
                &["humiliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confing"),
                &["config", "confine"],
            ),
            (dictgen::InsensitiveStr::Ascii("omaj"), &["homage", "oman"]),
            (
                dictgen::InsensitiveStr::Ascii("triangulataion"),
                &["triangulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwrd"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("posisitioning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frankenstines"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysees"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("interplation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("contener"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("unabling"),
                &["unable", "disabling", "enabling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parameterts"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusors"), &["cursors"]),
            (dictgen::InsensitiveStr::Ascii("swithces"), &["switches"]),
            (
                dictgen::InsensitiveStr::Ascii("radeuses"),
                &["radii", "radiuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cicled"),
                &["cycled", "circled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copmilation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("constand"), &["constant"]),
            (
                dictgen::InsensitiveStr::Ascii("communsits"),
                &["communists"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatabable"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("origined"), &["originated"]),
            (dictgen::InsensitiveStr::Ascii("wendsday"), &["wednesday"]),
            (dictgen::InsensitiveStr::Ascii("generats"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("coordenation"),
                &["coordination"],
            ),
            (dictgen::InsensitiveStr::Ascii("accademies"), &["academies"]),
            (
                dictgen::InsensitiveStr::Ascii("custumisable"),
                &["customisable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocnfiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcemnets"),
                &["announcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("quitted"), &["quit"]),
            (dictgen::InsensitiveStr::Ascii("televison"), &["television"]),
            (dictgen::InsensitiveStr::Ascii("custormer"), &["customer"]),
            (dictgen::InsensitiveStr::Ascii("structire"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("southanpton"),
                &["southampton"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crytpographically"),
                &["cryptographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consecuently"),
                &["consequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dishcarged"),
                &["discharged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foraigners"),
                &["foreigners"],
            ),
            (dictgen::InsensitiveStr::Ascii("aceptance"), &["acceptance"]),
            (
                dictgen::InsensitiveStr::Ascii("acceotable"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("paritial"), &["partial"]),
            (dictgen::InsensitiveStr::Ascii("prophesie"), &["prophecies"]),
            (dictgen::InsensitiveStr::Ascii("atomaton"), &["automaton"]),
            (dictgen::InsensitiveStr::Ascii("scurtiny"), &["scrutiny"]),
            (dictgen::InsensitiveStr::Ascii("algorythem"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("supspiciously"),
                &["suspiciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("anserer"), &["answerer"]),
            (dictgen::InsensitiveStr::Ascii("resuction"), &["reduction"]),
            (dictgen::InsensitiveStr::Ascii("charimsa"), &["charisma"]),
            (dictgen::InsensitiveStr::Ascii("occurrred"), &["occurred"]),
            (
                dictgen::InsensitiveStr::Ascii("satistifies"),
                &["satisfies"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgusied"), &["disguised"]),
            (dictgen::InsensitiveStr::Ascii("wavelengh"), &["wavelength"]),
            (
                dictgen::InsensitiveStr::Ascii("surrorates"),
                &["surrogates"],
            ),
            (dictgen::InsensitiveStr::Ascii("bointer"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("participte"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sobreity"), &["sobriety"]),
            (
                dictgen::InsensitiveStr::Ascii("corparation"),
                &["corporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("princila"), &["principals"]),
            (dictgen::InsensitiveStr::Ascii("templete"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("aristotel"), &["aristotle"]),
            (dictgen::InsensitiveStr::Ascii("noramals"), &["normals"]),
            (dictgen::InsensitiveStr::Ascii("tempraray"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("sentamental"),
                &["sentimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("phonemena"), &["phenomena"]),
            (dictgen::InsensitiveStr::Ascii("writye"), &["write"]),
            (
                dictgen::InsensitiveStr::Ascii("cannibalisim"),
                &["cannibalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exections"),
                &["executions", "ejections", "elections", "exertions"],
            ),
            (dictgen::InsensitiveStr::Ascii("involding"), &["involving"]),
            (
                dictgen::InsensitiveStr::Ascii("hierachically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("addrees"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("tefine"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("statup"), &["startup"]),
            (dictgen::InsensitiveStr::Ascii("gmaertag"), &["gamertag"]),
            (dictgen::InsensitiveStr::Ascii("librrary"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("reserching"),
                &["researching"],
            ),
            (dictgen::InsensitiveStr::Ascii("comadres"), &["comrades"]),
            (
                dictgen::InsensitiveStr::Ascii("colonialsim"),
                &["colonialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("comletion"), &["completion"]),
            (dictgen::InsensitiveStr::Ascii("protectos"), &["protectors"]),
            (dictgen::InsensitiveStr::Ascii("qwessants"), &["croissants"]),
            (dictgen::InsensitiveStr::Ascii("handshacke"), &["handshake"]),
            (dictgen::InsensitiveStr::Ascii("stringet"), &["stringent"]),
            (dictgen::InsensitiveStr::Ascii("monarkey"), &["monarchy"]),
            (dictgen::InsensitiveStr::Ascii("ahppen"), &["happen"]),
            (dictgen::InsensitiveStr::Ascii("pervail"), &["prevail"]),
            (
                dictgen::InsensitiveStr::Ascii("administratieve"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigularity"),
                &["singularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificare"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("territority"),
                &["territory"],
            ),
            (dictgen::InsensitiveStr::Ascii("stampade"), &["stamped"]),
            (
                dictgen::InsensitiveStr::Ascii("represend"),
                &["represented", "represent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizativos"),
                &["organizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envioronmental"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supramacist"),
                &["supremacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("loade"), &["load"]),
            (
                dictgen::InsensitiveStr::Ascii("envirnoments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirctives"), &["directives"]),
            (dictgen::InsensitiveStr::Ascii("wipoing"), &["wiping"]),
            (dictgen::InsensitiveStr::Ascii("endcoders"), &["encoders"]),
            (dictgen::InsensitiveStr::Ascii("winndow"), &["window"]),
            (dictgen::InsensitiveStr::Ascii("executued"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("nethods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("showede"), &["showered"]),
            (dictgen::InsensitiveStr::Ascii("toolsbox"), &["toolbox"]),
            (dictgen::InsensitiveStr::Ascii("napolen"), &["napoleon"]),
            (
                dictgen::InsensitiveStr::Ascii("resistanes"),
                &["resistances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unofficail"),
                &["unofficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("reslover"), &["resolver"]),
            (
                dictgen::InsensitiveStr::Ascii("miscellanious"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("devovled"), &["devolved"]),
            (
                dictgen::InsensitiveStr::Ascii("neighoring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associuated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("endiannes"), &["endianness"]),
            (dictgen::InsensitiveStr::Ascii("republian"), &["republican"]),
            (
                dictgen::InsensitiveStr::Ascii("successfule"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("fitnes"), &["fitness"]),
            (
                dictgen::InsensitiveStr::Ascii("convorsation"),
                &["conversation"],
            ),
            (dictgen::InsensitiveStr::Ascii("beahvior"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("morrocco"), &["morocco"]),
            (
                dictgen::InsensitiveStr::Ascii("institucion"),
                &["institution"],
            ),
            (dictgen::InsensitiveStr::Ascii("teminates"), &["terminates"]),
            (dictgen::InsensitiveStr::Ascii("prepar"), &["prepare"]),
            (
                dictgen::InsensitiveStr::Ascii("justications"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjectivley"),
                &["subjectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("nihilsim"), &["nihilism"]),
            (
                dictgen::InsensitiveStr::Ascii("trasaction"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("lackered"), &["lacquered"]),
            (dictgen::InsensitiveStr::Ascii("proecess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("outlows"),
                &["outflows", "outlaws"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegallity"),
                &["illegally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolutelly"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attatchments"),
                &["attachments"],
            ),
            (dictgen::InsensitiveStr::Ascii("outputes"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("rewitten"), &["rewritten"]),
            (
                dictgen::InsensitiveStr::Ascii("marshmalow"),
                &["marshmallow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disingenously"),
                &["disingenuously"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumning"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("paly"),
                &["play", "poly", "pally", "palsy", "palmy"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterest"), &["interest"]),
            (dictgen::InsensitiveStr::Ascii("estoeric"), &["esoteric"]),
            (dictgen::InsensitiveStr::Ascii("inporter"), &["importer"]),
            (
                dictgen::InsensitiveStr::Ascii("implemementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwayse"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("coreesponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("semnatics"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("contemperary"),
                &["contemporary"],
            ),
            (dictgen::InsensitiveStr::Ascii("pinnalce"), &["pinnacle"]),
            (dictgen::InsensitiveStr::Ascii("expencive"), &["expensive"]),
            (
                dictgen::InsensitiveStr::Ascii("compoments"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stlyesheet"),
                &["stylesheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiteirs"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("faiulre"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("cieling"), &["ceiling"]),
            (
                dictgen::InsensitiveStr::Ascii("contirbutors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplha"), &["alpha"]),
            (dictgen::InsensitiveStr::Ascii("seing"), &["seeing"]),
            (
                dictgen::InsensitiveStr::Ascii("saskatcehwan"),
                &["saskatchewan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hunred"),
                &["hundred", "hunted"],
            ),
            (dictgen::InsensitiveStr::Ascii("priemere"), &["premiere"]),
            (dictgen::InsensitiveStr::Ascii("queriable"), &["queryable"]),
            (
                dictgen::InsensitiveStr::Ascii("notifictations"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("dictactes"), &["dictates"]),
            (dictgen::InsensitiveStr::Ascii("obsolite"), &["obsolete"]),
            (dictgen::InsensitiveStr::Ascii("unsigend"), &["unsigned"]),
            (
                dictgen::InsensitiveStr::Ascii("semiconducter"),
                &["semiconductor"],
            ),
            (dictgen::InsensitiveStr::Ascii("benghai"), &["benghazi"]),
            (dictgen::InsensitiveStr::Ascii("chedules"), &["schedules"]),
            (dictgen::InsensitiveStr::Ascii("obsidion"), &["obsidian"]),
            (dictgen::InsensitiveStr::Ascii("pretendas"), &["pretends"]),
            (
                dictgen::InsensitiveStr::Ascii("transcocded"),
                &["transcoded"],
            ),
            (dictgen::InsensitiveStr::Ascii("tombsone"), &["tombstone"]),
            (dictgen::InsensitiveStr::Ascii("sugesting"), &["suggesting"]),
            (dictgen::InsensitiveStr::Ascii("bood"), &["boot"]),
            (dictgen::InsensitiveStr::Ascii("moderatore"), &["moderate"]),
            (dictgen::InsensitiveStr::Ascii("lifspan"), &["lifespan"]),
            (
                dictgen::InsensitiveStr::Ascii("predefineds"),
                &["predefined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ukrainains"),
                &["ukrainians"],
            ),
            (dictgen::InsensitiveStr::Ascii("ablout"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("liberae"), &["liberate"]),
            (
                dictgen::InsensitiveStr::Ascii("acessories"),
                &["accessories"],
            ),
            (dictgen::InsensitiveStr::Ascii("exacutor"), &["executor"]),
            (dictgen::InsensitiveStr::Ascii("atributes"), &["attributes"]),
            (
                dictgen::InsensitiveStr::Ascii("intimidatin"),
                &["intimidation", "intimidating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptiors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggeted"), &["suggested"]),
            (dictgen::InsensitiveStr::Ascii("separeter"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("compelte"), &["complete"]),
            (
                dictgen::InsensitiveStr::Ascii("religously"),
                &["religiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhilerating"),
                &["exhilarating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validaters"),
                &["validators", "validates"],
            ),
            (dictgen::InsensitiveStr::Ascii("imovable"), &["immovable"]),
            (dictgen::InsensitiveStr::Ascii("contolled"), &["controlled"]),
            (
                dictgen::InsensitiveStr::Ascii("subsysthems"),
                &["subsystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complitely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplicatess"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("viwe"), &["view"]),
            (
                dictgen::InsensitiveStr::Ascii("refelctions"),
                &["reflections"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplfy"), &["simplify"]),
            (dictgen::InsensitiveStr::Ascii("structetr"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("rescedule"), &["reschedule"]),
            (dictgen::InsensitiveStr::Ascii("resurcing"), &["resourcing"]),
            (
                dictgen::InsensitiveStr::Ascii("hypertrohpy"),
                &["hypertrophy"],
            ),
            (dictgen::InsensitiveStr::Ascii("exclusuive"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendent"),
                &["dependent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggets"),
                &["suggest", "suggests"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchrouns"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("unexected"), &["unexpected"]),
            (dictgen::InsensitiveStr::Ascii("compaible"), &["compatible"]),
            (dictgen::InsensitiveStr::Ascii("wehre"), &["where"]),
            (dictgen::InsensitiveStr::Ascii("buldings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("teraforming"),
                &["terraforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prejudgudiced"),
                &["prejudiced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commoditites"),
                &["commodities"],
            ),
            (dictgen::InsensitiveStr::Ascii("girlling"), &["grilling"]),
            (dictgen::InsensitiveStr::Ascii("ensuers"), &["ensures"]),
            (
                dictgen::InsensitiveStr::Ascii("stratagically"),
                &["strategically"],
            ),
            (dictgen::InsensitiveStr::Ascii("articel"), &["article"]),
            (dictgen::InsensitiveStr::Ascii("detaill"), &["detail"]),
            (dictgen::InsensitiveStr::Ascii("analyzator"), &["analyzer"]),
            (dictgen::InsensitiveStr::Ascii("wissling"), &["whistling"]),
            (dictgen::InsensitiveStr::Ascii("qualites"), &["qualities"]),
            (dictgen::InsensitiveStr::Ascii("spirtied"), &["spirited"]),
            (dictgen::InsensitiveStr::Ascii("ineriting"), &["inheriting"]),
            (dictgen::InsensitiveStr::Ascii("gorry"), &["gory"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoteing"),
                &["downvoting"],
            ),
            (dictgen::InsensitiveStr::Ascii("attendace"), &["attendance"]),
            (dictgen::InsensitiveStr::Ascii("towords"), &["towards"]),
            (
                dictgen::InsensitiveStr::Ascii("contagiosa"),
                &["contagious"],
            ),
            (dictgen::InsensitiveStr::Ascii("devicde"), &["device"]),
            (dictgen::InsensitiveStr::Ascii("symetries"), &["symmetries"]),
            (
                dictgen::InsensitiveStr::Ascii("disassmeble"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("menssages"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("accrate"), &["accurate"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilished"),
                &["established", "stabilised"],
            ),
            (dictgen::InsensitiveStr::Ascii("smoetimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("obection"), &["objection"]),
            (
                dictgen::InsensitiveStr::Ascii("udnerpowered"),
                &["underpowered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loggin"),
                &["login", "logging"],
            ),
            (dictgen::InsensitiveStr::Ascii("roundign"), &["rounding"]),
            (dictgen::InsensitiveStr::Ascii("uprades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("seventen"), &["seventeen"]),
            (
                dictgen::InsensitiveStr::Ascii("methodoligy"),
                &["methodology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcris"),
                &["transcripts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilzed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("everythng"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("intergrate"), &["integrate"]),
            (
                dictgen::InsensitiveStr::Ascii("unmisakable"),
                &["unmistakable"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutliply"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("communicatoin"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowldges"),
                &["acknowledges"],
            ),
            (dictgen::InsensitiveStr::Ascii("requre"), &["require"]),
            (
                dictgen::InsensitiveStr::Ascii("supproters"),
                &["supporters"],
            ),
            (dictgen::InsensitiveStr::Ascii("deletetion"), &["deletion"]),
            (dictgen::InsensitiveStr::Ascii("polygond"), &["polygons"]),
            (
                dictgen::InsensitiveStr::Ascii("initaitors"),
                &["initiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chirstians"),
                &["christians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appereances"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unneedingly"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammoung"), &["among"]),
            (dictgen::InsensitiveStr::Ascii("otherws"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("condomns"), &["condoms"]),
            (dictgen::InsensitiveStr::Ascii("cotroller"), &["controller"]),
            (dictgen::InsensitiveStr::Ascii("miht"), &["might"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("effedts"), &["effects"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropriet"),
                &["inappropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostalga"), &["nostalgia"]),
            (
                dictgen::InsensitiveStr::Ascii("presuasive"),
                &["persuasive"],
            ),
            (dictgen::InsensitiveStr::Ascii("enumarate"), &["enumerate"]),
            (dictgen::InsensitiveStr::Ascii("skurges"), &["scourges"]),
            (
                dictgen::InsensitiveStr::Ascii("tranfering"),
                &["transferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterperet"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("agreing"), &["agreeing"]),
            (dictgen::InsensitiveStr::Ascii("anarchisim"), &["anarchism"]),
            (
                dictgen::InsensitiveStr::Ascii("commomplace"),
                &["commonplace"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgest"), &["digest"]),
            (dictgen::InsensitiveStr::Ascii("roundinf"), &["rounding"]),
            (
                dictgen::InsensitiveStr::Ascii("propegator"),
                &["propagator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcationalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpoluation"),
                &["interpolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asendencey"),
                &["ascendency"],
            ),
            (dictgen::InsensitiveStr::Ascii("repulican"), &["republican"]),
            (
                dictgen::InsensitiveStr::Ascii("convienent"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utilitatian"),
                &["utilitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("goruping"), &["grouping"]),
            (
                dictgen::InsensitiveStr::Ascii("deprivition"),
                &["deprivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicovers"), &["discovers"]),
            (
                dictgen::InsensitiveStr::Ascii("stormfrount"),
                &["stormfront"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessecity"), &["necessity"]),
            (dictgen::InsensitiveStr::Ascii("childbird"), &["childbirth"]),
            (
                dictgen::InsensitiveStr::Ascii("chage"),
                &["change", "charge"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfeclty"), &["perfectly"]),
            (dictgen::InsensitiveStr::Ascii("overwath"), &["overwatch"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainces"),
                &["maintenances"],
            ),
            (dictgen::InsensitiveStr::Ascii("deviders"), &["dividers"]),
            (
                dictgen::InsensitiveStr::Ascii("antibiodics"),
                &["antibiotics"],
            ),
            (dictgen::InsensitiveStr::Ascii("refeence"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("unbalenced"),
                &["unbalanced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectacularely"),
                &["spectacularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependending"),
                &["depending"],
            ),
            (dictgen::InsensitiveStr::Ascii("visercal"), &["visceral"]),
            (
                dictgen::InsensitiveStr::Ascii("derevatives"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("celullar"), &["cellular"]),
            (
                dictgen::InsensitiveStr::Ascii("virutalenv"),
                &["virtualenv"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitaion"), &["limitation"]),
            (dictgen::InsensitiveStr::Ascii("abosrbed"), &["absorbed"]),
            (
                dictgen::InsensitiveStr::Ascii("optimizaions"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("decomposit"), &["decompose"]),
            (dictgen::InsensitiveStr::Ascii("unrpoven"), &["unproven"]),
            (
                dictgen::InsensitiveStr::Ascii("contributiongs"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descencing"),
                &["descending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infintesimal"),
                &["infinitesimal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generelizing"),
                &["generalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("naviagtors"),
                &["navigators"],
            ),
            (dictgen::InsensitiveStr::Ascii("publissher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("buider"), &["builder"]),
            (
                dictgen::InsensitiveStr::Ascii("gravitacional"),
                &["gravitational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acccumulate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motifications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convirting"),
                &["converting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicriminated"),
                &["discriminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("recoginse"), &["recognise"]),
            (
                dictgen::InsensitiveStr::Ascii("marshmellows"),
                &["marshmallows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentiation"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("upagrade"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("speciefically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuncurrent"),
                &["concurrent"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlcusion"), &["inclusion"]),
            (dictgen::InsensitiveStr::Ascii("losely"), &["loosely"]),
            (dictgen::InsensitiveStr::Ascii("dobles"), &["doubles"]),
            (
                dictgen::InsensitiveStr::Ascii("semgmentation"),
                &["segmentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("crytsal"), &["crystal"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidare"),
                &["invalidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("attractes"), &["attracts"]),
            (dictgen::InsensitiveStr::Ascii("humitidy"), &["humidity"]),
            (dictgen::InsensitiveStr::Ascii("ultamite"), &["ultimate"]),
            (
                dictgen::InsensitiveStr::Ascii("persvasive"),
                &["persuasive"],
            ),
            (dictgen::InsensitiveStr::Ascii("qestion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortuantly"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooduct"), &["product"]),
            (dictgen::InsensitiveStr::Ascii("repsonses"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("scrrenshot"),
                &["screenshot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construst"),
                &["constructs", "construct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisn"),
                &["comparisons", "comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("trcer"), &["tracer"]),
            (
                dictgen::InsensitiveStr::Ascii("complitation"),
                &["complication", "compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contructed"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("degreees"), &["degrees"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessfually"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("enties"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("contition"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("superseeding"),
                &["superseding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissatsified"),
                &["dissatisfied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statmenmts"),
                &["statements"],
            ),
            (dictgen::InsensitiveStr::Ascii("secretas"), &["secrets"]),
            (
                dictgen::InsensitiveStr::Ascii("aproaching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixally"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("examinating"),
                &["examining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expendeture"),
                &["expenditure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inmplementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gitars"), &["guitars"]),
            (
                dictgen::InsensitiveStr::Ascii("simetricaly"),
                &["symmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structered"),
                &["structured"],
            ),
            (dictgen::InsensitiveStr::Ascii("srevice"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("rregister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("overlorded"),
                &["overloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("decemer"), &["december"]),
            (
                dictgen::InsensitiveStr::Ascii("topolgical"),
                &["topological"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangaed"), &["managed"]),
            (dictgen::InsensitiveStr::Ascii("coversity"), &["coverity"]),
            (
                dictgen::InsensitiveStr::Ascii("virtualisaion"),
                &["virtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("marijauna"), &["marijuana"]),
            (
                dictgen::InsensitiveStr::Ascii("logarithimic"),
                &["logarithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphasiszed"),
                &["emphasized"],
            ),
            (dictgen::InsensitiveStr::Ascii("spetember"), &["september"]),
            (dictgen::InsensitiveStr::Ascii("acivate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("permisison"),
                &["permissions", "permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("occour"), &["occur"]),
            (dictgen::InsensitiveStr::Ascii("settign"), &["setting"]),
            (
                dictgen::InsensitiveStr::Ascii("workarould"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illsutrator"),
                &["illustrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("positionns"), &["positions"]),
            (
                dictgen::InsensitiveStr::Ascii("implemetned"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reptuation"),
                &["reputation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unanimuous"), &["unanimous"]),
            (
                dictgen::InsensitiveStr::Ascii("initialice"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("witha"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("chandaliers"),
                &["chandeliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("filtred"), &["filtered"]),
            (
                dictgen::InsensitiveStr::Ascii("workstaton"),
                &["workstation"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconary"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("mililiter"), &["milliliter"]),
            (dictgen::InsensitiveStr::Ascii("detecters"), &["detectors"]),
            (dictgen::InsensitiveStr::Ascii("htts"), &["https"]),
            (dictgen::InsensitiveStr::Ascii("descriibed"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("regart"), &["regard"]),
            (dictgen::InsensitiveStr::Ascii("ampliflier"), &["amplifier"]),
            (
                dictgen::InsensitiveStr::Ascii("consistes"),
                &["consists", "consisted"],
            ),
            (dictgen::InsensitiveStr::Ascii("chatolic"), &["catholic"]),
            (dictgen::InsensitiveStr::Ascii("descrybe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("oposition"), &["opposition"]),
            (
                dictgen::InsensitiveStr::Ascii("concerte"),
                &["concrete", "concerts", "concert", "concerto", "concerted"],
            ),
            (dictgen::InsensitiveStr::Ascii("sesion"), &["session"]),
            (
                dictgen::InsensitiveStr::Ascii("absoluutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactherising"),
                &["characterising"],
            ),
            (dictgen::InsensitiveStr::Ascii("beetween"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("commonwelth"),
                &["commonwealth"],
            ),
            (dictgen::InsensitiveStr::Ascii("narcisst"), &["narcissist"]),
            (dictgen::InsensitiveStr::Ascii("coginto"), &["cognito"]),
            (dictgen::InsensitiveStr::Ascii("impune"), &["impugn"]),
            (dictgen::InsensitiveStr::Ascii("kindapped"), &["kidnapped"]),
            (
                dictgen::InsensitiveStr::Ascii("beuracracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("thenn"), &["then"]),
            (
                dictgen::InsensitiveStr::Ascii("craetor"),
                &["crater", "creator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncencored"),
                &["uncensored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indcations"),
                &["indications"],
            ),
            (dictgen::InsensitiveStr::Ascii("archiv"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("acadimy"), &["academy"]),
            (dictgen::InsensitiveStr::Ascii("capablity"), &["capability"]),
            (
                dictgen::InsensitiveStr::Ascii("detecter"),
                &["detector", "detected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invidually"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procedes"),
                &["proceeds", "precedes"],
            ),
            (dictgen::InsensitiveStr::Ascii("ussefully"), &["usefully"]),
            (
                dictgen::InsensitiveStr::Ascii("improovements"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("enableing"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("therapuetic"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresible"),
                &["compressible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redudantly"),
                &["redundantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarrasment"),
                &["embarrassment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("artifiically"),
                &["artificially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciriticality"),
                &["criticality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precendances"),
                &["precedences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interplotation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("buissiness"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("migriane"), &["migraine"]),
            (
                dictgen::InsensitiveStr::Ascii("sepatations"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("coulmn"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("intangiable"),
                &["intangible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutablility"),
                &["mutability"],
            ),
            (dictgen::InsensitiveStr::Ascii("wishlsit"), &["wishlist"]),
            (
                dictgen::InsensitiveStr::Ascii("rondazyvoued"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("econimists"),
                &["economists"],
            ),
            (dictgen::InsensitiveStr::Ascii("timeuot"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("conversaion"),
                &["conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("servioce"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("malyasia"), &["malaysia"]),
            (
                dictgen::InsensitiveStr::Ascii("ambassodors"),
                &["ambassadors"],
            ),
            (dictgen::InsensitiveStr::Ascii("gainined"), &["gained"]),
            (
                dictgen::InsensitiveStr::Ascii("intolerent"),
                &["intolerant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expresions"),
                &["expression", "expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("primaris"), &["primaries"]),
            (dictgen::InsensitiveStr::Ascii("configued"), &["configured"]),
            (
                dictgen::InsensitiveStr::Ascii("equivilant"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlelated"),
                &["correlated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performnace"),
                &["performances", "performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("blokcing"), &["blocking"]),
            (dictgen::InsensitiveStr::Ascii("abstactly"), &["abstractly"]),
            (dictgen::InsensitiveStr::Ascii("liberatore"), &["liberate"]),
            (dictgen::InsensitiveStr::Ascii("ditsnace"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("prioritsation"),
                &["prioritisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniuplations"),
                &["manipulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("neccecary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("whitsling"), &["whistling"]),
            (
                dictgen::InsensitiveStr::Ascii("consensuel"),
                &["consensual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstace"),
                &["circumstance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgorund"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correllated"),
                &["correlated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mchanics"), &["mechanics"]),
            (dictgen::InsensitiveStr::Ascii("calees"), &["callees"]),
            (
                dictgen::InsensitiveStr::Ascii("investigstion"),
                &["investigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagnosted"), &["diagnose"]),
            (dictgen::InsensitiveStr::Ascii("administr"), &["administer"]),
            (dictgen::InsensitiveStr::Ascii("hyjacked"), &["hijacked"]),
            (dictgen::InsensitiveStr::Ascii("langeuages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("perfexcion"),
                &["perfection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prespective"),
                &["perspective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overpowerd"),
                &["overpowered"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifu"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("suspectiable"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("protential"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("extesnions"),
                &["extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("memicking"), &["mimicking"]),
            (dictgen::InsensitiveStr::Ascii("jewllery"), &["jewellery"]),
            (
                dictgen::InsensitiveStr::Ascii("commening"),
                &["commenting", "commending", "commencing", "communing"],
            ),
            (dictgen::InsensitiveStr::Ascii("depoisted"), &["deposited"]),
            (dictgen::InsensitiveStr::Ascii("ancestory"), &["ancestry"]),
            (dictgen::InsensitiveStr::Ascii("beeings"), &["beings"]),
            (
                dictgen::InsensitiveStr::Ascii("houndred"),
                &["hundred", "hounded"],
            ),
            (dictgen::InsensitiveStr::Ascii("simpify"), &["simplify"]),
            (
                dictgen::InsensitiveStr::Ascii("livestreem"),
                &["livestream"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenisonalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportians"),
                &["proportions"],
            ),
            (dictgen::InsensitiveStr::Ascii("feminen"), &["feminine"]),
            (
                dictgen::InsensitiveStr::Ascii("unfourtunetly"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("piched"),
                &["picked", "pinched", "pitched"],
            ),
            (dictgen::InsensitiveStr::Ascii("refromist"), &["reformist"]),
            (dictgen::InsensitiveStr::Ascii("observare"), &["observer"]),
            (dictgen::InsensitiveStr::Ascii("exceprt"), &["excerpt"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaucats"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("thisle"), &["thistle"]),
            (dictgen::InsensitiveStr::Ascii("reopended"), &["reopened"]),
            (
                dictgen::InsensitiveStr::Ascii("gamms"),
                &["gammas", "games"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremeanted"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("dteect"), &["detect"]),
            (
                dictgen::InsensitiveStr::Ascii("assicietes"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("sommething"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("institutie"), &["institute"]),
            (dictgen::InsensitiveStr::Ascii("expertas"), &["experts"]),
            (
                dictgen::InsensitiveStr::Ascii("brigdes"),
                &["bridges", "brigades"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prediciton"),
                &["prediction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obtaint"),
                &["obtain", "obtained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrants"),
                &["constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("typess"), &["types"]),
            (dictgen::InsensitiveStr::Ascii("bitcon"), &["bitcoin"]),
            (dictgen::InsensitiveStr::Ascii("defintily"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("collectivelly"),
                &["collectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understandble"),
                &["understandable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experminentally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitirs"), &["stirs"]),
            (dictgen::InsensitiveStr::Ascii("semamphore"), &["semaphore"]),
            (
                dictgen::InsensitiveStr::Ascii("distinghishable"),
                &["distinguishable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placeholers"),
                &["placeholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("finanaces"), &["finances"]),
            (dictgen::InsensitiveStr::Ascii("diarea"), &["diarrhea"]),
            (dictgen::InsensitiveStr::Ascii("desktopos"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("bobmers"), &["bombers"]),
            (
                dictgen::InsensitiveStr::Ascii("caclucations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aproved"), &["approved"]),
            (dictgen::InsensitiveStr::Ascii("maxumum"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("reltaionship"),
                &["relationship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consultaion"),
                &["consultation"],
            ),
            (dictgen::InsensitiveStr::Ascii("blcoking"), &["blocking"]),
            (dictgen::InsensitiveStr::Ascii("anarkists"), &["anarchists"]),
            (
                dictgen::InsensitiveStr::Ascii("parliments"),
                &["parliaments"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimphos"), &["nymphos"]),
            (
                dictgen::InsensitiveStr::Ascii("provinical"),
                &["provincial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antidepresssants"),
                &["antidepressants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficialliers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bariner"), &["brainer"]),
            (
                dictgen::InsensitiveStr::Ascii("copmonents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decalration"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experincing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclockig"),
                &["overclocking"],
            ),
            (dictgen::InsensitiveStr::Ascii("grafics"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("environental"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescripcion"),
                &["prescription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amelearative"),
                &["ameliorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("hemispere"), &["hemisphere"]),
            (dictgen::InsensitiveStr::Ascii("descrease"), &["decrease"]),
            (
                dictgen::InsensitiveStr::Ascii("placehold"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("digitaly"), &["digitally"]),
            (
                dictgen::InsensitiveStr::Ascii("chrsitians"),
                &["christians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabiities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("encryped"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("provinciaal"),
                &["provincial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aford"),
                &["afford", "afore"],
            ),
            (dictgen::InsensitiveStr::Ascii("virtaully"), &["virtually"]),
            (dictgen::InsensitiveStr::Ascii("editiable"), &["editable"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprehesible"),
                &["incomprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifikation"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorging"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("diasbling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("linkinag"), &["linking"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowlede"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("voleyball"), &["volleyball"]),
            (
                dictgen::InsensitiveStr::Ascii("directoriei"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrmities"),
                &["extremities"],
            ),
            (dictgen::InsensitiveStr::Ascii("oiginal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("courtrom"), &["courtroom"]),
            (dictgen::InsensitiveStr::Ascii("procesure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("bitocins"), &["bitcoins"]),
            (
                dictgen::InsensitiveStr::Ascii("corrently"),
                &["correctly", "currently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representatin"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigatio"),
                &["investigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separeting"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreating"),
                &["aggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individus"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocurrance"), &["occurrence"]),
            (
                dictgen::InsensitiveStr::Ascii("uderstandably"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlysing"), &["analysing"]),
            (dictgen::InsensitiveStr::Ascii("pytohn"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("subscrpition"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asphyxation"),
                &["asphyxiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompartible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasformation"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("secified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("propper"), &["proper"]),
            (
                dictgen::InsensitiveStr::Ascii("intantiates"),
                &["instantiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("obivius"), &["obvious"]),
            (
                dictgen::InsensitiveStr::Ascii("poperty"),
                &["property", "properly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straiht"),
                &["straight", "strait"],
            ),
            (dictgen::InsensitiveStr::Ascii("polute"), &["pollute"]),
            (
                dictgen::InsensitiveStr::Ascii("unrealisitic"),
                &["unrealistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cholesteral"),
                &["cholesterol"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modernizationz"),
                &["modernizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("languajes"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("jaunary"), &["january"]),
            (dictgen::InsensitiveStr::Ascii("proteinas"), &["proteins"]),
            (
                dictgen::InsensitiveStr::Ascii("selecrtable"),
                &["selectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senitivity"),
                &["sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuousiy"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attiributes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("obediance"), &["obedience"]),
            (
                dictgen::InsensitiveStr::Ascii("provokative"),
                &["provocative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weakend"),
                &["weakened", "weekend"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosumed"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("simpathizers"),
                &["sympathizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanved"), &["advanced"]),
            (dictgen::InsensitiveStr::Ascii("intamicy"), &["intimacy"]),
            (dictgen::InsensitiveStr::Ascii("beginging"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondnts"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integarted"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mvoes"), &["moves"]),
            (
                dictgen::InsensitiveStr::Ascii("provisionging"),
                &["provisioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("difussive"), &["diffusive"]),
            (dictgen::InsensitiveStr::Ascii("intquiry"), &["inquiry"]),
            (dictgen::InsensitiveStr::Ascii("poitn"), &["point"]),
            (dictgen::InsensitiveStr::Ascii("dregree"), &["degree"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotiotation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experemented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constraits"),
                &["constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("expeenses"), &["expenses"]),
            (
                dictgen::InsensitiveStr::Ascii("intentionaly"),
                &["intentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("monteize"), &["monetize"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleread"),
                &["accelerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("thonic"), &["chthonic"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibe"),
                &["responsive", "responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reparamterize"),
                &["reparameterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("suystem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("scrambleing"),
                &["scrambling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("slaughterd"),
                &["slaughtered"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuacasian"), &["caucasian"]),
            (dictgen::InsensitiveStr::Ascii("codo"), &["code"]),
            (dictgen::InsensitiveStr::Ascii("veiwed"), &["viewed"]),
            (
                dictgen::InsensitiveStr::Ascii("orthoganalize"),
                &["orthogonalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradititional"),
                &["traditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedestirans"),
                &["pedestrians"],
            ),
            (dictgen::InsensitiveStr::Ascii("poupular"), &["popular"]),
            (dictgen::InsensitiveStr::Ascii("alpahs"), &["alphas"]),
            (dictgen::InsensitiveStr::Ascii("apendices"), &["appendices"]),
            (dictgen::InsensitiveStr::Ascii("commnting"), &["commenting"]),
            (dictgen::InsensitiveStr::Ascii("wonderous"), &["wondrous"]),
            (dictgen::InsensitiveStr::Ascii("complted"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("deciaml"), &["decimal"]),
            (
                dictgen::InsensitiveStr::Ascii("shanenigans"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revews"),
                &["reviews", "renews", "revels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undefinetively"),
                &["indefinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("denpend"), &["depend"]),
            (dictgen::InsensitiveStr::Ascii("cehckbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("autoincrementive"),
                &["autoincrement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intruductory"),
                &["introductory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photagraphers"),
                &["photographers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referendom"),
                &["referendum"],
            ),
            (dictgen::InsensitiveStr::Ascii("existet"), &["existent"]),
            (dictgen::InsensitiveStr::Ascii("uitls"), &["utils"]),
            (dictgen::InsensitiveStr::Ascii("realte"), &["relate"]),
            (dictgen::InsensitiveStr::Ascii("debudgs"), &["debugs"]),
            (dictgen::InsensitiveStr::Ascii("suscriber"), &["subscriber"]),
            (dictgen::InsensitiveStr::Ascii("persual"), &["perusal"]),
            (dictgen::InsensitiveStr::Ascii("trimuph"), &["triumph"]),
            (
                dictgen::InsensitiveStr::Ascii("indicativo"),
                &["indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparising"),
                &["comparing"],
            ),
            (dictgen::InsensitiveStr::Ascii("obays"), &["obeys"]),
            (dictgen::InsensitiveStr::Ascii("tranfer"), &["transfer"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptin"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("pollind"), &["polling"]),
            (
                dictgen::InsensitiveStr::Ascii("resloution"),
                &["resolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("gettetx"), &["gettext"]),
            (
                dictgen::InsensitiveStr::Ascii("innovativo"),
                &["innovation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temperarures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoticable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesring"),
                &["interfering"],
            ),
            (dictgen::InsensitiveStr::Ascii("plas"), &["plays"]),
            (
                dictgen::InsensitiveStr::Ascii("sensationality"),
                &["sensationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("athough"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsisten"),
                &["inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("techique"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("rsourced"),
                &["resourced", "sourced"],
            ),
            (dictgen::InsensitiveStr::Ascii("exclusiev"), &["exclusive"]),
            (dictgen::InsensitiveStr::Ascii("inviation"), &["invitation"]),
            (dictgen::InsensitiveStr::Ascii("hitogram"), &["histogram"]),
            (dictgen::InsensitiveStr::Ascii("amoungst"), &["amongst"]),
            (
                dictgen::InsensitiveStr::Ascii("arquitects"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plauthroughs"),
                &["playthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrible"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("potencial"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("astronaught"),
                &["astronaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreegates"),
                &["aggregates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ankshusly"), &["anxiously"]),
            (dictgen::InsensitiveStr::Ascii("ppcheck"), &["cppcheck"]),
            (
                dictgen::InsensitiveStr::Ascii("interepretations"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resuses"),
                &["reuses", "refuses", "resumes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progammers"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathod"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("verbatum"), &["verbatim"]),
            (dictgen::InsensitiveStr::Ascii("multpile"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("psychopathes"),
                &["psychopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("carosel"), &["carousel"]),
            (
                dictgen::InsensitiveStr::Ascii("dscriptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infestaton"),
                &["infestation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sorce"),
                &["source", "force", "sore"],
            ),
            (dictgen::InsensitiveStr::Ascii("verisons"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("performe"),
                &["perform", "performed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenonemal"),
                &["phenomenal"],
            ),
            (dictgen::InsensitiveStr::Ascii("costant"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("equil"), &["equal"]),
            (dictgen::InsensitiveStr::Ascii("supprter"), &["supporter"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalesl"),
                &["nationals"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambuigity"), &["ambiguity"]),
            (dictgen::InsensitiveStr::Ascii("supsend"), &["suspend"]),
            (dictgen::InsensitiveStr::Ascii("aadding"), &["adding"]),
            (dictgen::InsensitiveStr::Ascii("driveable"), &["drivable"]),
            (dictgen::InsensitiveStr::Ascii("expoter"), &["exporter"]),
            (dictgen::InsensitiveStr::Ascii("soildly"), &["solidly"]),
            (
                dictgen::InsensitiveStr::Ascii("partitoned"),
                &["partitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("logictech"), &["logitech"]),
            (
                dictgen::InsensitiveStr::Ascii("workspsaces"),
                &["workspaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("desigers"), &["designers"]),
            (dictgen::InsensitiveStr::Ascii("extrinsict"), &["extrinsic"]),
            (dictgen::InsensitiveStr::Ascii("eitquette"), &["etiquette"]),
            (
                dictgen::InsensitiveStr::Ascii("heirarchal"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superceeding"),
                &["superseding"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpoy"), &["coy", "copy"]),
            (dictgen::InsensitiveStr::Ascii("succecion"), &["succession"]),
            (dictgen::InsensitiveStr::Ascii("polocy"), &["policy"]),
            (dictgen::InsensitiveStr::Ascii("functionss"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("constitutues"),
                &["constitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generatons"),
                &["generators", "generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("michellle"), &["michelle"]),
            (
                dictgen::InsensitiveStr::Ascii("topoligical"),
                &["topological"],
            ),
            (dictgen::InsensitiveStr::Ascii("surfce"), &["surface"]),
            (
                dictgen::InsensitiveStr::Ascii("respnsiveness"),
                &["responsiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("referr"), &["refer"]),
            (dictgen::InsensitiveStr::Ascii("paficist"), &["pacifist"]),
            (dictgen::InsensitiveStr::Ascii("igonore"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("mutally"), &["mutually"]),
            (
                dictgen::InsensitiveStr::Ascii("responsebilities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scubscriber"),
                &["subscriber"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbort"),
                &["abort", "abbot"],
            ),
            (dictgen::InsensitiveStr::Ascii("parralely"), &["parallelly"]),
            (dictgen::InsensitiveStr::Ascii("rendiring"), &["rendering"]),
            (
                dictgen::InsensitiveStr::Ascii("commentater"),
                &["commentator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explictily"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("depictin"), &["depictions"]),
            (dictgen::InsensitiveStr::Ascii("crititical"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("intagration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("desing"), &["design"]),
            (
                dictgen::InsensitiveStr::Ascii("discconeccting"),
                &["disconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotionations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simlarlity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammenities"), &["amenities"]),
            (dictgen::InsensitiveStr::Ascii("documenet"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("theisitc"), &["theistic"]),
            (
                dictgen::InsensitiveStr::Ascii("propagtion"),
                &["propagation"],
            ),
            (dictgen::InsensitiveStr::Ascii("particuar"), &["particular"]),
            (
                dictgen::InsensitiveStr::Ascii("intilises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("branchsi"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("stlyes"), &["styles"]),
            (
                dictgen::InsensitiveStr::Ascii("replicatin"),
                &["replicating", "replication"],
            ),
            (dictgen::InsensitiveStr::Ascii("knw"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("gnerates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("inisialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulit"), &["built"]),
            (dictgen::InsensitiveStr::Ascii("squardon"), &["squadron"]),
            (
                dictgen::InsensitiveStr::Ascii("alternitivly"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("yorkshie"), &["yorkshire"]),
            (dictgen::InsensitiveStr::Ascii("collecing"), &["collecting"]),
            (dictgen::InsensitiveStr::Ascii("wrapeprs"), &["wrappers"]),
            (
                dictgen::InsensitiveStr::Ascii("adventue"),
                &["adventures", "adventure"],
            ),
            (dictgen::InsensitiveStr::Ascii("vicintiy"), &["vicinity"]),
            (
                dictgen::InsensitiveStr::Ascii("interrruptions"),
                &["interruptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marshmalows"),
                &["marshmallows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectoion"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contridicts"),
                &["contradicts"],
            ),
            (dictgen::InsensitiveStr::Ascii("enumated"), &["enumerated"]),
            (
                dictgen::InsensitiveStr::Ascii("relpacements"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("messasged"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("quantaties"),
                &["quantities"],
            ),
            (dictgen::InsensitiveStr::Ascii("chopipng"), &["chopping"]),
            (dictgen::InsensitiveStr::Ascii("misbehaive"), &["misbehave"]),
            (dictgen::InsensitiveStr::Ascii("apdate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbehoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("argumentents"),
                &["arguments"],
            ),
            (dictgen::InsensitiveStr::Ascii("havn"), &["haven"]),
            (
                dictgen::InsensitiveStr::Ascii("formualtions"),
                &["formulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exacutions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("nullabour"), &["nullarbor"]),
            (dictgen::InsensitiveStr::Ascii("sourcen"), &["source"]),
            (dictgen::InsensitiveStr::Ascii("perisists"), &["persists"]),
            (
                dictgen::InsensitiveStr::Ascii("incorperating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sykward"), &["skyward"]),
            (dictgen::InsensitiveStr::Ascii("tabes"), &["tables", "tabs"]),
            (dictgen::InsensitiveStr::Ascii("aligining"), &["aligning"]),
            (
                dictgen::InsensitiveStr::Ascii("perminantly"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represneted"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("utillising"), &["utilising"]),
            (dictgen::InsensitiveStr::Ascii("whih"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("misundersatnding"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accreditied"),
                &["accredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attributin"),
                &["attributing", "attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("larer"),
                &["larger", "later", "layer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("importantce"),
                &["importance"],
            ),
            (dictgen::InsensitiveStr::Ascii("interfear"), &["interfere"]),
            (
                dictgen::InsensitiveStr::Ascii("antagnoists"),
                &["antagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperical"), &["spherical"]),
            (
                dictgen::InsensitiveStr::Ascii("compinsated"),
                &["compensated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bliefs"), &["beliefs"]),
            (dictgen::InsensitiveStr::Ascii("kinght"), &["knight"]),
            (
                dictgen::InsensitiveStr::Ascii("fomulas"),
                &["formula", "formulas"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exeecutable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentraions"),
                &["concentrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reveret"), &["revert"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialsing"),
                &["deserialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metholodogy"),
                &["methodology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplacates"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("acchiev"), &["achieve"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisiert"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcues"), &["produces"]),
            (
                dictgen::InsensitiveStr::Ascii("reproduciablity"),
                &["reproducibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("veritical"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("autmatically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("retieving"), &["retrieving"]),
            (
                dictgen::InsensitiveStr::Ascii("bargianing"),
                &["bargaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("housholds"), &["households"]),
            (dictgen::InsensitiveStr::Ascii("excercise"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("vacummes"), &["vacuums"]),
            (dictgen::InsensitiveStr::Ascii("nonsence"), &["nonsense"]),
            (dictgen::InsensitiveStr::Ascii("iritate"), &["irritate"]),
            (
                dictgen::InsensitiveStr::Ascii("simplisity"),
                &["simplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("loclaly"), &["locally"]),
            (
                dictgen::InsensitiveStr::Ascii("obtaineble"),
                &["obtainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constituintes"),
                &["constituents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("democratas"), &["democrats"]),
            (dictgen::InsensitiveStr::Ascii("surpases"), &["surpass"]),
            (
                dictgen::InsensitiveStr::Ascii("proprotional"),
                &["proportional"],
            ),
            (dictgen::InsensitiveStr::Ascii("encosers"), &["encoders"]),
            (dictgen::InsensitiveStr::Ascii("yyou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("wresseler"), &["wrestler"]),
            (
                dictgen::InsensitiveStr::Ascii("experessing"),
                &["expressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sependent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("raedy"), &["ready"]),
            (
                dictgen::InsensitiveStr::Ascii("statistcally"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperratly"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusivley"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("confudion"), &["confusion"]),
            (
                dictgen::InsensitiveStr::Ascii("picoseonds"),
                &["picoseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("learnig"), &["learning"]),
            (
                dictgen::InsensitiveStr::Ascii("colonizaton"),
                &["colonization"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambibuity"), &["ambiguity"]),
            (
                dictgen::InsensitiveStr::Ascii("comparisment"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schduleing"),
                &["scheduling"],
            ),
            (dictgen::InsensitiveStr::Ascii("eligble"), &["eligible"]),
            (
                dictgen::InsensitiveStr::Ascii("tansferred"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("indx"), &["index"]),
            (dictgen::InsensitiveStr::Ascii("personsa"), &["personas"]),
            (dictgen::InsensitiveStr::Ascii("disclamer"), &["disclaimer"]),
            (
                dictgen::InsensitiveStr::Ascii("independenet"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckeck"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("succesul"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("cotaves"), &["octaves"]),
            (dictgen::InsensitiveStr::Ascii("peopl"), &["people"]),
            (
                dictgen::InsensitiveStr::Ascii("autcomplete"),
                &["autocomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("scracthed"), &["scratched"]),
            (
                dictgen::InsensitiveStr::Ascii("interpeted"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("missspells"), &["misspells"]),
            (
                dictgen::InsensitiveStr::Ascii("guarantess"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesystes"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordenates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("defautts"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("dasboard"), &["dashboard"]),
            (
                dictgen::InsensitiveStr::Ascii("initializs"),
                &["initialize", "initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arguin"),
                &["arguing", "airgun"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indecation"),
                &["indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registerations"),
                &["registrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foled"),
                &["fold", "fooled", "foiled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhancments"),
                &["enhancements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somehtings"),
                &["somethings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explenatory"),
                &["explanatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimillar"),
                &["dissimilar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documeneting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("overfl"), &["overflow"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiastics"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("namne"), &["name"]),
            (
                dictgen::InsensitiveStr::Ascii("delimitiers"),
                &["delimiters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envinronment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operatos"),
                &["operators", "operates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wholeheartidly"),
                &["wholeheartedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocaiing"),
                &["allocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("alreaady"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("thoughout"), &["throughout"]),
            (
                dictgen::InsensitiveStr::Ascii("dispachers"),
                &["dispatchers"],
            ),
            (dictgen::InsensitiveStr::Ascii("emotinal"), &["emotional"]),
            (
                dictgen::InsensitiveStr::Ascii("inconviniencys"),
                &["inconveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scuccessully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("knockous"), &["noxious"]),
            (
                dictgen::InsensitiveStr::Ascii("consectively"),
                &["consecutively"],
            ),
            (dictgen::InsensitiveStr::Ascii("sybsystem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("collabration"),
                &["collaboration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repete"),
                &["repeat", "repute", "replete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suffciency"),
                &["sufficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("existig"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("projecter"),
                &["projector", "projected", "project"],
            ),
            (dictgen::InsensitiveStr::Ascii("vakued"), &["valued"]),
            (
                dictgen::InsensitiveStr::Ascii("palestina"),
                &["palestinian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funxtional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinsm"),
                &["determinism"],
            ),
            (dictgen::InsensitiveStr::Ascii("delate"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("trottle"), &["throttle"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtorising"),
                &["authorising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribbutions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("itereated"), &["iterated"]),
            (
                dictgen::InsensitiveStr::Ascii("correspontend"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patameters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improofment"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmit"), &["emit"]),
            (dictgen::InsensitiveStr::Ascii("unorotated"), &["unrotated"]),
            (dictgen::InsensitiveStr::Ascii("roganisms"), &["organisms"]),
            (dictgen::InsensitiveStr::Ascii("operattor"), &["operator"]),
            (
                dictgen::InsensitiveStr::Ascii("quadrapedal"),
                &["quadrupedal"],
            ),
            (dictgen::InsensitiveStr::Ascii("apartide"), &["apartheid"]),
            (
                dictgen::InsensitiveStr::Ascii("prosecutie"),
                &["prosecuted"],
            ),
            (dictgen::InsensitiveStr::Ascii("suspenso"), &["suspension"]),
            (dictgen::InsensitiveStr::Ascii("milimeter"), &["millimeter"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boycutting"),
                &["boycotting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaases"),
                &["releases", "relapses"],
            ),
            (dictgen::InsensitiveStr::Ascii("cntains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forgettting"),
                &["forgetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("championshp"),
                &["championship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrbution"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ponctuation"),
                &["punctuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respesctive"),
                &["respective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionos"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("drived"),
                &["derived", "drove", "driven"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeating"),
                &["completing", "competing"],
            ),
            (dictgen::InsensitiveStr::Ascii("segwayed"), &["segued"]),
            (
                dictgen::InsensitiveStr::Ascii("architectues"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("screebs"), &["screens"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bombardeada"),
                &["bombarded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprezentative"),
                &["representative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essesitally"),
                &["essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("existsing"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("portalis"), &["portals"]),
            (dictgen::InsensitiveStr::Ascii("speficed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("opetionally"),
                &["optionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("suceesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("targetd"), &["targeted"]),
            (dictgen::InsensitiveStr::Ascii("infromal"), &["informal"]),
            (dictgen::InsensitiveStr::Ascii("dungeos"), &["dungeons"]),
            (dictgen::InsensitiveStr::Ascii("dedicato"), &["dedication"]),
            (dictgen::InsensitiveStr::Ascii("transeat"), &["translates"]),
            (dictgen::InsensitiveStr::Ascii("premptive"), &["preemptive"]),
            (dictgen::InsensitiveStr::Ascii("eneters"), &["enters"]),
            (dictgen::InsensitiveStr::Ascii("engeries"), &["energies"]),
            (
                dictgen::InsensitiveStr::Ascii("intimitading"),
                &["intimidating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restorating"),
                &["restoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("precess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("constratined"),
                &["constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptiom"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("talbe"), &["table"]),
            (dictgen::InsensitiveStr::Ascii("accouns"), &["accounts"]),
            (dictgen::InsensitiveStr::Ascii("devaint"), &["deviant"]),
            (
                dictgen::InsensitiveStr::Ascii("transfered"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("sypmtoms"), &["symptoms"]),
            (
                dictgen::InsensitiveStr::Ascii("detectiona"),
                &["detection", "detections"],
            ),
            (dictgen::InsensitiveStr::Ascii("updaers"), &["updaters"]),
            (dictgen::InsensitiveStr::Ascii("haricut"), &["haircut"]),
            (dictgen::InsensitiveStr::Ascii("swithch"), &["switch"]),
            (
                dictgen::InsensitiveStr::Ascii("concered"),
                &["concede", "concerned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expierences"),
                &["experiences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concellation"),
                &["cancellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalisitic"),
                &["nationalistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("knarled"), &["gnarled"]),
            (dictgen::InsensitiveStr::Ascii("previouls"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("strawbeery"),
                &["strawberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhausing"), &["exhausting"]),
            (dictgen::InsensitiveStr::Ascii("finilise"), &["finalise"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenion"),
                &["concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indigneous"),
                &["indigenous"],
            ),
            (dictgen::InsensitiveStr::Ascii("execive"), &["excessive"]),
            (
                dictgen::InsensitiveStr::Ascii("finctional"),
                &["functional", "fictional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchrously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostehtic"),
                &["prosthetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("evesdrop"), &["eavesdrop"]),
            (dictgen::InsensitiveStr::Ascii("dismisse"), &["dismissive"]),
            (
                dictgen::InsensitiveStr::Ascii("stlyesheets"),
                &["stylesheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imanent"),
                &["eminent", "imminent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adolence"),
                &["adolescent", "adolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("workarond"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("materla"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallizations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presed"),
                &["preset", "pressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opinyonable"),
                &["opinionable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millesecond"),
                &["millisecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("complet"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("resgisters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("intstrument"),
                &["instrument"],
            ),
            (dictgen::InsensitiveStr::Ascii("editiing"), &["editing"]),
            (dictgen::InsensitiveStr::Ascii("condion"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("peristently"),
                &["persistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideologicaly"),
                &["ideologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limination"),
                &["limitation", "lamination"],
            ),
            (dictgen::InsensitiveStr::Ascii("moduile"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("alghoritmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentait"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("custumise"), &["customise"]),
            (
                dictgen::InsensitiveStr::Ascii("operarional"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("entried"), &["entries"]),
            (
                dictgen::InsensitiveStr::Ascii("narcississm"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("intestins"), &["intestines"]),
            (dictgen::InsensitiveStr::Ascii("bre"), &["be", "brie"]),
            (dictgen::InsensitiveStr::Ascii("backgrond"), &["background"]),
            (
                dictgen::InsensitiveStr::Ascii("applifiers"),
                &["amplifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("seantor"), &["senator"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatoin"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolens"), &["tokens"]),
            (dictgen::InsensitiveStr::Ascii("stalekrs"), &["stalkers"]),
            (dictgen::InsensitiveStr::Ascii("kubenates"), &["kubernetes"]),
            (dictgen::InsensitiveStr::Ascii("alue"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("trackign"), &["tracking"]),
            (
                dictgen::InsensitiveStr::Ascii("independece"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disssemble"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithsms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interprise"),
                &["enterprise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desciription"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approprpiate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("redirrect"), &["redirect"]),
            (dictgen::InsensitiveStr::Ascii("elignment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("xepectedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integradion"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calle"),
                &["called", "caller", "calls"],
            ),
            (dictgen::InsensitiveStr::Ascii("provieders"), &["providers"]),
            (dictgen::InsensitiveStr::Ascii("comipler"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("patern"), &["pattern"]),
            (
                dictgen::InsensitiveStr::Ascii("subdiretories"),
                &["subdirectories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achives"),
                &["achieves", "archives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoagressive"),
                &["autoaggressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendance"),
                &["dependance"],
            ),
            (dictgen::InsensitiveStr::Ascii("stilletto"), &["stiletto"]),
            (dictgen::InsensitiveStr::Ascii("diganose"), &["diagnose"]),
            (dictgen::InsensitiveStr::Ascii("rouding"), &["rounding"]),
            (
                dictgen::InsensitiveStr::Ascii("negociates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocaed"), &["allocated"]),
            (
                dictgen::InsensitiveStr::Ascii("abstanence"),
                &["abstinence"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeurying"), &["querying"]),
            (dictgen::InsensitiveStr::Ascii("innput"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("prohibted"), &["prohibited"]),
            (
                dictgen::InsensitiveStr::Ascii("santiation"),
                &["sanitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ointer"),
                &["pointer", "linter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelises"),
                &["parallelises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanantion"),
                &["explanation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conteners"), &["containers"]),
            (
                dictgen::InsensitiveStr::Ascii("notfications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmened"), &["segmented"]),
            (dictgen::InsensitiveStr::Ascii("requieres"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("deserializae"),
                &["deserialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiifally"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repraesentations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("animatng"), &["animating"]),
            (
                dictgen::InsensitiveStr::Ascii("inheriable"),
                &["inheritable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conntections"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("recreatie"), &["recreate"]),
            (dictgen::InsensitiveStr::Ascii("autmobile"), &["automobile"]),
            (dictgen::InsensitiveStr::Ascii("refernced"), &["referenced"]),
            (
                dictgen::InsensitiveStr::Ascii("characterstics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procedrual"),
                &["procedural"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiblec"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("dissct"), &["dissect"]),
            (dictgen::InsensitiveStr::Ascii("noteworty"), &["noteworthy"]),
            (dictgen::InsensitiveStr::Ascii("reamde"), &["readme"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvinent"),
                &["inconvenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiemzone"), &["timezone"]),
            (dictgen::InsensitiveStr::Ascii("childen"), &["children"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumentaiton"),
                &["instrumentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("descion"), &["decision"]),
            (
                dictgen::InsensitiveStr::Ascii("especialyl"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignes"), &["aligns"]),
            (
                dictgen::InsensitiveStr::Ascii("initilialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impelementations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misalinged"),
                &["misaligned"],
            ),
            (dictgen::InsensitiveStr::Ascii("givem"), &["given"]),
            (dictgen::InsensitiveStr::Ascii("gargoils"), &["gargoyles"]),
            (
                dictgen::InsensitiveStr::Ascii("simultaenously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrosponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheher"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("archiectures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incrementos"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tained"),
                &["stained", "tainted"],
            ),
            (dictgen::InsensitiveStr::Ascii("bimtap"), &["bitmap"]),
            (
                dictgen::InsensitiveStr::Ascii("dependancy"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tinterrupts"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisin"),
                &["characterising"],
            ),
            (dictgen::InsensitiveStr::Ascii("sice"), &["since"]),
            (dictgen::InsensitiveStr::Ascii("exprots"), &["exports"]),
            (dictgen::InsensitiveStr::Ascii("hodlings"), &["holdings"]),
            (dictgen::InsensitiveStr::Ascii("circuitery"), &["circuitry"]),
            (dictgen::InsensitiveStr::Ascii("scootes"), &["scooters"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoenstly"), &["honestly"]),
            (dictgen::InsensitiveStr::Ascii("dupicated"), &["duplicated"]),
            (
                dictgen::InsensitiveStr::Ascii("secons"),
                &["second", "seconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convigured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("mnemoncis"), &["mnemonics"]),
            (dictgen::InsensitiveStr::Ascii("haranguin"), &["haranguing"]),
            (
                dictgen::InsensitiveStr::Ascii("unconsiderate"),
                &["inconsiderate"],
            ),
            (dictgen::InsensitiveStr::Ascii("seige"), &["siege"]),
            (
                dictgen::InsensitiveStr::Ascii("delimitions"),
                &["delimitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemtentations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continutation"),
                &["continuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirimenting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("farest"),
                &["fairest", "farthest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buearucratic"),
                &["bureaucratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferend"),
                &["preferred", "preference"],
            ),
            (dictgen::InsensitiveStr::Ascii("extreems"), &["extremes"]),
            (dictgen::InsensitiveStr::Ascii("restircts"), &["restricts"]),
            (dictgen::InsensitiveStr::Ascii("porved"), &["proved"]),
            (
                dictgen::InsensitiveStr::Ascii("califnoria"),
                &["californian"],
            ),
            (dictgen::InsensitiveStr::Ascii("organisim"), &["organism"]),
            (dictgen::InsensitiveStr::Ascii("tarhet"), &["target"]),
            (dictgen::InsensitiveStr::Ascii("cusomer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrivated"),
                &["aggravated"],
            ),
            (dictgen::InsensitiveStr::Ascii("introdued"), &["introduced"]),
            (dictgen::InsensitiveStr::Ascii("compliles"), &["compiles"]),
            (dictgen::InsensitiveStr::Ascii("clpboard"), &["clipboard"]),
            (dictgen::InsensitiveStr::Ascii("osffset"), &["offset"]),
            (dictgen::InsensitiveStr::Ascii("rememers"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("nnot"), &["not"]),
            (
                dictgen::InsensitiveStr::Ascii("initializeing"),
                &["initializing", "intitializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootstraped"),
                &["bootstrapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("pompay"), &["pompeii"]),
            (dictgen::InsensitiveStr::Ascii("algnments"), &["alignments"]),
            (dictgen::InsensitiveStr::Ascii("synamic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("simetric"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("errro"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("fiedl"), &["field"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplayer"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sinlgeplayer"),
                &["singleplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("precceded"), &["preceded"]),
            (
                dictgen::InsensitiveStr::Ascii("monitering"),
                &["monitoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soovinneers"),
                &["souvenirs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differeces"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("comppiler"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("serialialize"),
                &["serialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unathorized"),
                &["unauthorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ajustments"),
                &["adjustments"],
            ),
            (dictgen::InsensitiveStr::Ascii("multiles"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("destinaion"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retreaving"),
                &["retrieving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obstruktion"),
                &["obstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("malfucntions"),
                &["malfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("marliyn"), &["marilyn"]),
            (
                dictgen::InsensitiveStr::Ascii("emobdiment"),
                &["embodiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscciating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegdement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("broswable"), &["browsable"]),
            (dictgen::InsensitiveStr::Ascii("idenfies"), &["identifies"]),
            (dictgen::InsensitiveStr::Ascii("carrrier"), &["carrier"]),
            (
                dictgen::InsensitiveStr::Ascii("disrecpect"),
                &["disrespect"],
            ),
            (dictgen::InsensitiveStr::Ascii("intreeg"), &["intrigue"]),
            (
                dictgen::InsensitiveStr::Ascii("abstrcting"),
                &["abstracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseudicaly"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deploymenet"),
                &["deployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refubrished"),
                &["refurbished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addeds"),
                &["added", "adders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contanining"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardnials"), &["cardinals"]),
            (dictgen::InsensitiveStr::Ascii("laguages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("biscut"), &["biscuit"]),
            (dictgen::InsensitiveStr::Ascii("warumup"), &["warmup"]),
            (dictgen::InsensitiveStr::Ascii("weavin"), &["weaving"]),
            (dictgen::InsensitiveStr::Ascii("prioritiy"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("controdiction"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcing"),
                &["constructing", "construing"],
            ),
            (dictgen::InsensitiveStr::Ascii("variuous"), &["various"]),
            (
                dictgen::InsensitiveStr::Ascii("potentails"),
                &["potentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("subcriber"), &["subscriber"]),
            (
                dictgen::InsensitiveStr::Ascii("proective"),
                &["protective", "projective", "proactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("unbounnd"), &["unbound"]),
            (dictgen::InsensitiveStr::Ascii("smae"), &["same"]),
            (dictgen::InsensitiveStr::Ascii("lnguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("ancester"), &["ancestor"]),
            (dictgen::InsensitiveStr::Ascii("otpimize"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("behavoral"), &["behavioral"]),
            (
                dictgen::InsensitiveStr::Ascii("itnerpreted"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodiving"), &["providing"]),
            (dictgen::InsensitiveStr::Ascii("qulity"), &["quality"]),
            (dictgen::InsensitiveStr::Ascii("indicatet"), &["indicated"]),
            (
                dictgen::InsensitiveStr::Ascii("buring"),
                &["burying", "burning", "burin", "during"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciprociate"),
                &["reciprocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("restires"), &["restores"]),
            (dictgen::InsensitiveStr::Ascii("grappnel"), &["grapple"]),
            (
                dictgen::InsensitiveStr::Ascii("necesssities"),
                &["necessities"],
            ),
            (dictgen::InsensitiveStr::Ascii("onwee"), &["ennui"]),
            (dictgen::InsensitiveStr::Ascii("mirro"), &["mirror"]),
            (
                dictgen::InsensitiveStr::Ascii("absoultely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("formating"), &["formatting"]),
            (dictgen::InsensitiveStr::Ascii("disuade"), &["dissuade"]),
            (dictgen::InsensitiveStr::Ascii("silentely"), &["silently"]),
            (dictgen::InsensitiveStr::Ascii("posisson"), &["poisson"]),
            (dictgen::InsensitiveStr::Ascii("propert"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("sawteed"), &["sautéd"]),
            (dictgen::InsensitiveStr::Ascii("birhtday"), &["birthday"]),
            (
                dictgen::InsensitiveStr::Ascii("capatilist"),
                &["capitalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disambigation"),
                &["disambiguation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("councel"),
                &["council", "counsel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedophilies"),
                &["pedophile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imlimenting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticaitng"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("escpe"), &["escape"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvienience"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("banch"),
                &["bank", "branch", "bench"],
            ),
            (dictgen::InsensitiveStr::Ascii("algortims"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("doument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("immdediately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("vizualise"), &["visualise"]),
            (
                dictgen::InsensitiveStr::Ascii("apreciation"),
                &["appreciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rotatin"),
                &["rotating", "rotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatates"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("boundatries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achieveds"),
                &["achieves", "achieved"],
            ),
            (dictgen::InsensitiveStr::Ascii("disscuesed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("comsummed"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("resolutinos"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconeted"),
                &["disconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hesistating"),
                &["hesitating"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcultes"), &["calculates"]),
            (dictgen::InsensitiveStr::Ascii("beliving"), &["believing"]),
            (
                dictgen::InsensitiveStr::Ascii("determinitic"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermodyanmics"),
                &["thermodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expansiones"),
                &["expansions"],
            ),
            (dictgen::InsensitiveStr::Ascii("wdth"), &["width"]),
            (dictgen::InsensitiveStr::Ascii("sycther"), &["scyther"]),
            (dictgen::InsensitiveStr::Ascii("excedded"), &["exceeded"]),
            (dictgen::InsensitiveStr::Ascii("exexuted"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("revelaing"), &["revealing"]),
            (dictgen::InsensitiveStr::Ascii("umark"), &["unmark"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("asetic"), &["ascetic"]),
            (dictgen::InsensitiveStr::Ascii("warwcik"), &["warwick"]),
            (dictgen::InsensitiveStr::Ascii("extedned"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("quanitities"),
                &["quantities"],
            ),
            (dictgen::InsensitiveStr::Ascii("deprectae"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("animetes"), &["animates"]),
            (dictgen::InsensitiveStr::Ascii("willfullly"), &["willfully"]),
            (dictgen::InsensitiveStr::Ascii("dafaut"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("reactionair"),
                &["reactionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliuent"), &["client"]),
            (
                dictgen::InsensitiveStr::Ascii("repeatible"),
                &["repeatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helipcopter"),
                &["helicopter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singulatiry"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("struggele"), &["struggle"]),
            (dictgen::InsensitiveStr::Ascii("fiddley"), &["fiddly"]),
            (dictgen::InsensitiveStr::Ascii("deeep"), &["deep"]),
            (dictgen::InsensitiveStr::Ascii("kurdisch"), &["kurdish"]),
            (dictgen::InsensitiveStr::Ascii("inititates"), &["initiates"]),
            (
                dictgen::InsensitiveStr::Ascii("consire"),
                &["conspire", "consider"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctructors"),
                &["destructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("multipes"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("certaintty"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("preferabbly"),
                &["preferably"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuetes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("unconvensional"),
                &["unconventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("failde"), &["failed"]),
            (
                dictgen::InsensitiveStr::Ascii("acomplishment"),
                &["accomplishment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remifications"),
                &["ramifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("delare"), &["declare"]),
            (
                dictgen::InsensitiveStr::Ascii("outragiously"),
                &["outrageously"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitween"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("syfs"), &["sysfs"]),
            (
                dictgen::InsensitiveStr::Ascii("nominacion"),
                &["nomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("languagues"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("thirtyth"), &["thirtieth"]),
            (
                dictgen::InsensitiveStr::Ascii("manifestanti"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("presssures"), &["pressures"]),
            (
                dictgen::InsensitiveStr::Ascii("allocaters"),
                &["allocators", "allocates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evauluation"),
                &["evaluation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decraese"),
                &["decrease", "decease"],
            ),
            (dictgen::InsensitiveStr::Ascii("buidling"), &["building"]),
            (dictgen::InsensitiveStr::Ascii("promt"), &["prompt"]),
            (
                dictgen::InsensitiveStr::Ascii("chcekboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dominatin"),
                &["domination", "dominating"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqensing"), &["sequencing"]),
            (
                dictgen::InsensitiveStr::Ascii("apporached"),
                &["approached"],
            ),
            (dictgen::InsensitiveStr::Ascii("mecanisms"), &["mechanisms"]),
            (dictgen::InsensitiveStr::Ascii("glipmse"), &["glimpse"]),
            (
                dictgen::InsensitiveStr::Ascii("generationals"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("curerncy"), &["currency"]),
            (dictgen::InsensitiveStr::Ascii("expell"), &["expel"]),
            (dictgen::InsensitiveStr::Ascii("separted"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("campere"),
                &["ampere", "compare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disenginuous"),
                &["disingenuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("processer"), &["processor"]),
            (
                dictgen::InsensitiveStr::Ascii("misdeamenors"),
                &["misdemeanor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incrementarlo"),
                &["incremental"],
            ),
            (dictgen::InsensitiveStr::Ascii("separe"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("discoveryd"),
                &["discovered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entrapeneur"),
                &["entrepreneur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exorbirant"),
                &["exorbitant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controleld"),
                &["controlled"],
            ),
            (dictgen::InsensitiveStr::Ascii("unafected"), &["unaffected"]),
            (dictgen::InsensitiveStr::Ascii("boundries"), &["boundaries"]),
            (dictgen::InsensitiveStr::Ascii("dischard"), &["discharged"]),
            (
                dictgen::InsensitiveStr::Ascii("dosctrings"),
                &["docstrings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrobutor"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodived"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("prpeparations"),
                &["preparations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficialliable"),
                &["specifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("humanitis"), &["humanities"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplent"),
                &["reimplement", "reimplant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backhacking"),
                &["backpacking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neigbhourhood"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("debudg"), &["debug"]),
            (
                dictgen::InsensitiveStr::Ascii("constact"),
                &["contact", "constant"],
            ),
            (dictgen::InsensitiveStr::Ascii("hvaing"), &["having"]),
            (dictgen::InsensitiveStr::Ascii("destory"), &["destroy"]),
            (
                dictgen::InsensitiveStr::Ascii("occasioanlly"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatcted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("comits"), &["commits"]),
            (
                dictgen::InsensitiveStr::Ascii("avaiibility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("stlyed"), &["styled"]),
            (
                dictgen::InsensitiveStr::Ascii("atmosphereic"),
                &["atmospheric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encarcerated"),
                &["incarcerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tripplets"), &["triplets"]),
            (dictgen::InsensitiveStr::Ascii("organim"), &["organism"]),
            (dictgen::InsensitiveStr::Ascii("surivial"), &["survival"]),
            (dictgen::InsensitiveStr::Ascii("ganerate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("aproched"), &["approached"]),
            (dictgen::InsensitiveStr::Ascii("inovative"), &["innovative"]),
            (dictgen::InsensitiveStr::Ascii("decieving"), &["deceiving"]),
            (dictgen::InsensitiveStr::Ascii("metions"), &["mentions"]),
            (dictgen::InsensitiveStr::Ascii("builing"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("multyplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximetly"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("perssiuos"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("fordwarded"), &["forwarded"]),
            (dictgen::InsensitiveStr::Ascii("pristen"), &["pristine"]),
            (dictgen::InsensitiveStr::Ascii("allerts"), &["alerts"]),
            (dictgen::InsensitiveStr::Ascii("columnn"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("agnosticim"),
                &["agnosticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destiations"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accessibiltiy"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perticipate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("chateaos"), &["châteaux"]),
            (
                dictgen::InsensitiveStr::Ascii("intersperese"),
                &["intersperse"],
            ),
            (dictgen::InsensitiveStr::Ascii("hindrence"), &["hindrance"]),
            (dictgen::InsensitiveStr::Ascii("miticator"), &["mitigator"]),
            (
                dictgen::InsensitiveStr::Ascii("specialites"),
                &["specializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reseraching"),
                &["researching"],
            ),
            (dictgen::InsensitiveStr::Ascii("probabably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("multinationella"),
                &["multinational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertaintly"),
                &["uncertainty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calclulation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categroized"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("eminated"), &["emanated"]),
            (dictgen::InsensitiveStr::Ascii("maintiain"), &["maintain"]),
            (dictgen::InsensitiveStr::Ascii("acccounted"), &["accounted"]),
            (dictgen::InsensitiveStr::Ascii("functionns"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("temperamant"),
                &["temperament"],
            ),
            (dictgen::InsensitiveStr::Ascii("texxt"), &["text"]),
            (dictgen::InsensitiveStr::Ascii("removerd"), &["removed"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("safeguared"),
                &["safeguard", "safeguarded"],
            ),
            (dictgen::InsensitiveStr::Ascii("mappble"), &["mappable"]),
            (dictgen::InsensitiveStr::Ascii("instlal"), &["instal"]),
            (dictgen::InsensitiveStr::Ascii("contrller"), &["controller"]),
            (
                dictgen::InsensitiveStr::Ascii("renegociates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("merchans"), &["merchants"]),
            (dictgen::InsensitiveStr::Ascii("adeilade"), &["adelaide"]),
            (
                dictgen::InsensitiveStr::Ascii("guantanemo"),
                &["guantanamo"],
            ),
            (dictgen::InsensitiveStr::Ascii("amtrix"), &["matrix"]),
            (dictgen::InsensitiveStr::Ascii("updload"), &["upload"]),
            (dictgen::InsensitiveStr::Ascii("porshan"), &["portion"]),
            (
                dictgen::InsensitiveStr::Ascii("algoprithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summurizing"),
                &["summarizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostalgija"), &["nostalgia"]),
            (
                dictgen::InsensitiveStr::Ascii("ceasin"),
                &["ceasing", "casein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protoyping"),
                &["prototyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instromentations"),
                &["instrumentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irresponcible"),
                &["irresponsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ethnocentricm"),
                &["ethnocentrism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intorverts"),
                &["introverts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreeming"),
                &["upstreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("paretns"), &["parents"]),
            (
                dictgen::InsensitiveStr::Ascii("vocabularily"),
                &["vocabulary"],
            ),
            (dictgen::InsensitiveStr::Ascii("itereators"), &["iterators"]),
            (
                dictgen::InsensitiveStr::Ascii("dynamcally"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("promptus"), &["prompts"]),
            (dictgen::InsensitiveStr::Ascii("turnning"), &["turning"]),
            (dictgen::InsensitiveStr::Ascii("encoeder"), &["encoder"]),
            (
                dictgen::InsensitiveStr::Ascii("agressivity"),
                &["aggressivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("popuplar"), &["popular"]),
            (dictgen::InsensitiveStr::Ascii("palletes"), &["palettes"]),
            (dictgen::InsensitiveStr::Ascii("sinagpore"), &["singapore"]),
            (dictgen::InsensitiveStr::Ascii("manyals"), &["manuals"]),
            (dictgen::InsensitiveStr::Ascii("therien"), &["therein"]),
            (dictgen::InsensitiveStr::Ascii("shippment"), &["shipment"]),
            (dictgen::InsensitiveStr::Ascii("fedility"), &["fidelity"]),
            (dictgen::InsensitiveStr::Ascii("injustie"), &["injustices"]),
            (
                dictgen::InsensitiveStr::Ascii("contibutes"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("venyets"), &["vignettes"]),
            (dictgen::InsensitiveStr::Ascii("ttest"), &["test", "attest"]),
            (
                dictgen::InsensitiveStr::Ascii("palistinians"),
                &["palestinians"],
            ),
            (dictgen::InsensitiveStr::Ascii("reptils"), &["reptiles"]),
            (
                dictgen::InsensitiveStr::Ascii("peformances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("werever"), &["wherever"]),
            (dictgen::InsensitiveStr::Ascii("provinciae"), &["province"]),
            (dictgen::InsensitiveStr::Ascii("astrounat"), &["astronaut"]),
            (dictgen::InsensitiveStr::Ascii("remembee"), &["remembered"]),
            (dictgen::InsensitiveStr::Ascii("skeletl"), &["skeletal"]),
            (dictgen::InsensitiveStr::Ascii("complting"), &["completing"]),
            (
                dictgen::InsensitiveStr::Ascii("excected"),
                &["executed", "expected"],
            ),
            (dictgen::InsensitiveStr::Ascii("obviosu"), &["obvious"]),
            (dictgen::InsensitiveStr::Ascii("snashot"), &["snapshot"]),
            (
                dictgen::InsensitiveStr::Ascii("defiintely"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesnors"), &["tensors"]),
            (
                dictgen::InsensitiveStr::Ascii("proficieny"),
                &["proficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consending"),
                &["consenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("interepret"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("communitcates"),
                &["communicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensititivity"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("millenial"), &["millennial"]),
            (dictgen::InsensitiveStr::Ascii("enouraged"), &["encouraged"]),
            (dictgen::InsensitiveStr::Ascii("breakpont"), &["breakpoint"]),
            (dictgen::InsensitiveStr::Ascii("ussernames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("surpressions"),
                &["suppressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aknowledge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("poppup"), &["popup"]),
            (dictgen::InsensitiveStr::Ascii("signales"), &["signals"]),
            (
                dictgen::InsensitiveStr::Ascii("implementantions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combatabts"),
                &["combatants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinaiton"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotiotion"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obliterared"),
                &["obliterated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntactally"),
                &["syntactically"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtilte"), &["subtitle"]),
            (dictgen::InsensitiveStr::Ascii("speicified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("probalby"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("unrepetant"),
                &["unrepentant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogeneize"),
                &["homogenize"],
            ),
            (dictgen::InsensitiveStr::Ascii("opionion"), &["opinion"]),
            (
                dictgen::InsensitiveStr::Ascii("syntesises"),
                &["synthesises"],
            ),
            (dictgen::InsensitiveStr::Ascii("browswer"), &["browser"]),
            (
                dictgen::InsensitiveStr::Ascii("stimulents"),
                &["stimulants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimatin"),
                &["estimation", "estimating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculcators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inclued"),
                &["included", "include"],
            ),
            (dictgen::InsensitiveStr::Ascii("juxtifies"), &["justifies"]),
            (dictgen::InsensitiveStr::Ascii("cardnial"), &["cardinal"]),
            (dictgen::InsensitiveStr::Ascii("cmmands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("expext"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsstrappers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facilties"),
                &["facilities", "faculties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilpicities"),
                &["multiplicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arameterized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("sourcelss"), &["sourceless"]),
            (dictgen::InsensitiveStr::Ascii("zefer"), &["zephyr"]),
            (
                dictgen::InsensitiveStr::Ascii("viralently"),
                &["virulently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplications"),
                &["multiplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("hace"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("satistfies"), &["satisfies"]),
            (dictgen::InsensitiveStr::Ascii("mencaing"), &["menacing"]),
            (dictgen::InsensitiveStr::Ascii("angostic"), &["agnostic"]),
            (dictgen::InsensitiveStr::Ascii("valiate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("independentness"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectt"), &["connect"]),
            (dictgen::InsensitiveStr::Ascii("adbucting"), &["abducting"]),
            (dictgen::InsensitiveStr::Ascii("chanager"), &["changer"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("viedo"), &["video"]),
            (
                dictgen::InsensitiveStr::Ascii("transmisisons"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrat"),
                &["constraint", "contrast"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registreed"),
                &["registered"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrrupt"), &["corrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("charactristic"),
                &["characteristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapears"),
                &["disappears"],
            ),
            (dictgen::InsensitiveStr::Ascii("clloud"), &["cloud"]),
            (
                dictgen::InsensitiveStr::Ascii("juggernaunts"),
                &["juggernaut"],
            ),
            (dictgen::InsensitiveStr::Ascii("compagnion"), &["companion"]),
            (dictgen::InsensitiveStr::Ascii("tempororay"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("tabl"), &["table"]),
            (dictgen::InsensitiveStr::Ascii("algotrithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("cusines"), &["cuisines"]),
            (
                dictgen::InsensitiveStr::Ascii("sugggestion"),
                &["suggestion"],
            ),
            (dictgen::InsensitiveStr::Ascii("purcahse"), &["purchase"]),
            (dictgen::InsensitiveStr::Ascii("henc"), &["hence"]),
            (
                dictgen::InsensitiveStr::Ascii("intergrations"),
                &["integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uplaodad"),
                &["upload", "uploaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("apporving"), &["approving"]),
            (dictgen::InsensitiveStr::Ascii("accecpt"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("organizativo"),
                &["organization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homophopic"),
                &["homophobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuallly"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurively"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("wnet"), &["went"]),
            (dictgen::InsensitiveStr::Ascii("reserverd"), &["reserved"]),
            (
                dictgen::InsensitiveStr::Ascii("metapackge"),
                &["metapackage"],
            ),
            (dictgen::InsensitiveStr::Ascii("deleleted"), &["deleted"]),
            (
                dictgen::InsensitiveStr::Ascii("confidantial"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("somewere"), &["somewhere"]),
            (dictgen::InsensitiveStr::Ascii("pathane"), &["pathname"]),
            (
                dictgen::InsensitiveStr::Ascii("arquitectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("organsied"), &["organised"]),
            (
                dictgen::InsensitiveStr::Ascii("protaginst"),
                &["protagonist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondg"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpretator"),
                &["perpetrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auospacing"),
                &["autospacing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conetnt"), &["content"]),
            (dictgen::InsensitiveStr::Ascii("attmpted"), &["attempted"]),
            (dictgen::InsensitiveStr::Ascii("cutomers"), &["customers"]),
            (dictgen::InsensitiveStr::Ascii("eqauality"), &["equality"]),
            (dictgen::InsensitiveStr::Ascii("miraculu"), &["miraculous"]),
            (
                dictgen::InsensitiveStr::Ascii("popuplation"),
                &["population"],
            ),
            (dictgen::InsensitiveStr::Ascii("weeknd"), &["weekend"]),
            (dictgen::InsensitiveStr::Ascii("holdiays"), &["holidays"]),
            (
                dictgen::InsensitiveStr::Ascii("chaeckboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchromously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contriubtions"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortimically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("gernerally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("cutom"), &["custom"]),
            (dictgen::InsensitiveStr::Ascii("appology"), &["apology"]),
            (dictgen::InsensitiveStr::Ascii("necessery"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("edning"),
                &["edging", "ending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("futurisitc"),
                &["futuristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completeing"),
                &["completing"],
            ),
            (dictgen::InsensitiveStr::Ascii("includs"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("gernerations"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("inncrement"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("themseleves"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("commitus"), &["commits"]),
            (dictgen::InsensitiveStr::Ascii("assit"), &["assist"]),
            (dictgen::InsensitiveStr::Ascii("samped"), &["sampled"]),
            (dictgen::InsensitiveStr::Ascii("bindins"), &["bindings"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinnate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstunding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("discoveres"), &["discovers"]),
            (dictgen::InsensitiveStr::Ascii("probllem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("pepperonni"), &["pepperoni"]),
            (
                dictgen::InsensitiveStr::Ascii("charactristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anitbiotics"),
                &["antibiotics"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("saftey"), &["safety"]),
            (
                dictgen::InsensitiveStr::Ascii("intiializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntacts"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("declartated"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("permision"), &["permission"]),
            (
                dictgen::InsensitiveStr::Ascii("overfapping"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difficaulty"),
                &["difficulty"],
            ),
            (dictgen::InsensitiveStr::Ascii("affintize"), &["affinitize"]),
            (dictgen::InsensitiveStr::Ascii("skywalkr"), &["skywalker"]),
            (
                dictgen::InsensitiveStr::Ascii("transperant"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mesoneen"), &["mezzanine"]),
            (dictgen::InsensitiveStr::Ascii("modfifies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("amendmants"),
                &["amendments"],
            ),
            (dictgen::InsensitiveStr::Ascii("jusitfy"), &["justify"]),
            (
                dictgen::InsensitiveStr::Ascii("contigiously"),
                &["contiguously"],
            ),
            (dictgen::InsensitiveStr::Ascii("fullscren"), &["fullscreen"]),
            (
                dictgen::InsensitiveStr::Ascii("encouragin"),
                &["encouraging"],
            ),
            (dictgen::InsensitiveStr::Ascii("publised"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("paralells"), &["parallels"]),
            (
                dictgen::InsensitiveStr::Ascii("constitutin"),
                &["constituting", "constitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepecification"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volunteeer"),
                &["volunteers"],
            ),
            (dictgen::InsensitiveStr::Ascii("initalize"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("copyeights"),
                &["copyrights"],
            ),
            (dictgen::InsensitiveStr::Ascii("alocholic"), &["alcoholic"]),
            (
                dictgen::InsensitiveStr::Ascii("minneaoplis"),
                &["minneapolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("desnse"), &["dense"]),
            (dictgen::InsensitiveStr::Ascii("verstion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("sorrogate"), &["surrogate"]),
            (
                dictgen::InsensitiveStr::Ascii("destributed"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyperobola"), &["hyperbola"]),
            (dictgen::InsensitiveStr::Ascii("extermism"), &["extremism"]),
            (dictgen::InsensitiveStr::Ascii("spcae"), &["space"]),
            (dictgen::InsensitiveStr::Ascii("orignally"), &["originally"]),
            (dictgen::InsensitiveStr::Ascii("floppys"), &["floppies"]),
            (dictgen::InsensitiveStr::Ascii("faincee"), &["fiancee"]),
            (dictgen::InsensitiveStr::Ascii("opjection"), &["objection"]),
            (dictgen::InsensitiveStr::Ascii("explicilt"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("shandeleer"),
                &["chandelier"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrytion"), &["encryption"]),
            (
                dictgen::InsensitiveStr::Ascii("invoves"),
                &["involves", "invokes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availabable"),
                &["available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distriuting"),
                &["distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ugprades"), &["upgrades"]),
            (
                dictgen::InsensitiveStr::Ascii("opportuny"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrokload"), &["workload"]),
            (
                dictgen::InsensitiveStr::Ascii("notifactions"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncomfertable"),
                &["uncomfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("presense"), &["presence"]),
            (dictgen::InsensitiveStr::Ascii("applience"), &["appliance"]),
            (
                dictgen::InsensitiveStr::Ascii("blokcss"),
                &["blocks", "blokes"],
            ),
            (dictgen::InsensitiveStr::Ascii("soilders"), &["soldiers"]),
            (dictgen::InsensitiveStr::Ascii("acidental"), &["accidental"]),
            (
                dictgen::InsensitiveStr::Ascii("replacte"),
                &["replace", "replicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("conlcudes"), &["concludes"]),
            (
                dictgen::InsensitiveStr::Ascii("enmvironmental"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoatiating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathalogist"),
                &["pathologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifiying"),
                &["classifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disasembler"),
                &["disassembler"],
            ),
            (dictgen::InsensitiveStr::Ascii("zionistes"), &["zionists"]),
            (dictgen::InsensitiveStr::Ascii("complelx"), &["complex"]),
            (
                dictgen::InsensitiveStr::Ascii("automicatilly"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coprrighted"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("toruneys"), &["tourneys"]),
            (dictgen::InsensitiveStr::Ascii("bayliwick"), &["bailiwick"]),
            (
                dictgen::InsensitiveStr::Ascii("sement"),
                &["cement", "segment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceopted"),
                &["exempted", "excepted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proporpotion"),
                &["proportion"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpeloys"), &["deploys"]),
            (dictgen::InsensitiveStr::Ascii("peageant"), &["pageant"]),
            (dictgen::InsensitiveStr::Ascii("acovados"), &["avocados"]),
            (dictgen::InsensitiveStr::Ascii("offest"), &["offset"]),
            (
                dictgen::InsensitiveStr::Ascii("undenaibly"),
                &["undeniably"],
            ),
            (dictgen::InsensitiveStr::Ascii("applogise"), &["apologise"]),
            (dictgen::InsensitiveStr::Ascii("ownder"), &["owner"]),
            (dictgen::InsensitiveStr::Ascii("circuts"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("nives"),
                &[
                    "dives", "fives", "hives", "knives", "nieves", "nines", "wives",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("martix"), &["matrix"]),
            (
                dictgen::InsensitiveStr::Ascii("phychological"),
                &["psychological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undermineing"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simuates"),
                &["simulates", "situates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ludcrious"), &["ludicrous"]),
            (dictgen::InsensitiveStr::Ascii("compation"), &["compaction"]),
            (dictgen::InsensitiveStr::Ascii("calcable"), &["calculable"]),
            (dictgen::InsensitiveStr::Ascii("encripts"), &["encrypts"]),
            (dictgen::InsensitiveStr::Ascii("desgined"), &["designed"]),
            (dictgen::InsensitiveStr::Ascii("opearting"), &["operating"]),
            (dictgen::InsensitiveStr::Ascii("defitnaly"), &["defiantly"]),
            (
                dictgen::InsensitiveStr::Ascii("approximetely"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("climbes"), &["climbers"]),
            (
                dictgen::InsensitiveStr::Ascii("desregarding"),
                &["disregarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infringeing"),
                &["infringing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ampilfying"),
                &["amplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administre"),
                &["administer"],
            ),
            (dictgen::InsensitiveStr::Ascii("ablities"), &["abilities"]),
            (dictgen::InsensitiveStr::Ascii("sylabus"), &["syllabus"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleartion"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("plagery"), &["plagiary"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminauti"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generaters"),
                &["generates", "generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("climer"), &["climber"]),
            (
                dictgen::InsensitiveStr::Ascii("claibscale"),
                &["calibscale"],
            ),
            (dictgen::InsensitiveStr::Ascii("ternimate"), &["terminate"]),
            (
                dictgen::InsensitiveStr::Ascii("exculsively"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("noramlise"), &["normalise"]),
            (
                dictgen::InsensitiveStr::Ascii("subscirption"),
                &["subscriptions", "subscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("enfornce"), &["enforce"]),
            (
                dictgen::InsensitiveStr::Ascii("censorhsip"),
                &["censorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("gliphs"), &["glyphs"]),
            (
                dictgen::InsensitiveStr::Ascii("malfunciton"),
                &["malfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreieve"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("fetishers"), &["fetishes"]),
            (dictgen::InsensitiveStr::Ascii("chizzelles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("authencticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lushis"), &["luscious"]),
            (dictgen::InsensitiveStr::Ascii("overlapp"), &["overlap"]),
            (
                dictgen::InsensitiveStr::Ascii("experiminters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benckmarks"),
                &["benchmarks"],
            ),
            (dictgen::InsensitiveStr::Ascii("retruning"), &["returning"]),
            (dictgen::InsensitiveStr::Ascii("wirless"), &["wireless"]),
            (dictgen::InsensitiveStr::Ascii("unaminous"), &["unanimous"]),
            (
                dictgen::InsensitiveStr::Ascii("cellabrating"),
                &["celebrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("riminices"), &["reminisces"]),
            (dictgen::InsensitiveStr::Ascii("assemlies"), &["assemblies"]),
            (dictgen::InsensitiveStr::Ascii("itselv"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("dimaond"), &["diamond"]),
            (
                dictgen::InsensitiveStr::Ascii("inizializations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditonals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("canadias"), &["canadians"]),
            (dictgen::InsensitiveStr::Ascii("macaronni"), &["macaroni"]),
            (dictgen::InsensitiveStr::Ascii("indoro"), &["indoor"]),
            (dictgen::InsensitiveStr::Ascii("chaecked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("allegeance"),
                &["allegiance"],
            ),
            (dictgen::InsensitiveStr::Ascii("legoins"), &["legions"]),
            (
                dictgen::InsensitiveStr::Ascii("collaberative"),
                &["collaborative"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuqest"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("pojects"), &["projects"]),
            (
                dictgen::InsensitiveStr::Ascii("precussion"),
                &["percussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inherantly"),
                &["inherently"],
            ),
            (dictgen::InsensitiveStr::Ascii("cherckers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("accesssible"),
                &["accessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprixamations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technicain"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigater"),
                &["investigator"],
            ),
            (dictgen::InsensitiveStr::Ascii("adressed"), &["addressed"]),
            (
                dictgen::InsensitiveStr::Ascii("misogonist"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesed"), &["used"]),
            (
                dictgen::InsensitiveStr::Ascii("atttributing"),
                &["attributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siezing"),
                &["seizing", "sizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("insterter"), &["inserter"]),
            (dictgen::InsensitiveStr::Ascii("expicitly"), &["explicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("annaversery"),
                &["anniversary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contemplare"),
                &["contemplate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reste"),
                &["reset", "rest", "rests"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quess"),
                &["guess", "quests"],
            ),
            (dictgen::InsensitiveStr::Ascii("nodel"), &["model", "nodal"]),
            (dictgen::InsensitiveStr::Ascii("claendars"), &["calendars"]),
            (
                dictgen::InsensitiveStr::Ascii("contatiners"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hannibl"), &["hannibal"]),
            (dictgen::InsensitiveStr::Ascii("indevers"), &["endeavors"]),
            (
                dictgen::InsensitiveStr::Ascii("negoatiator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segemntation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progresssion"),
                &["progressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("defitinly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("lintain"), &["lintian"]),
            (
                dictgen::InsensitiveStr::Ascii("unreigstered"),
                &["unregistered"],
            ),
            (dictgen::InsensitiveStr::Ascii("realllly"), &["really"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiifations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggestied"),
                &["suggestive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excersizing"),
                &["exercising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotaitions"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chescks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("motivacional"),
                &["motivational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emprisoned"),
                &["imprisoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("obtasisned"), &["obtained"]),
            (
                dictgen::InsensitiveStr::Ascii("paranetrized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaulted"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("dimnention"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("specialiced"),
                &["specialized", "specialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inacurracies"),
                &["inaccuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchroneous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("aprox"), &["approx"]),
            (dictgen::InsensitiveStr::Ascii("naything"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("ventialtion"),
                &["ventilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("respondus"), &["responds"]),
            (dictgen::InsensitiveStr::Ascii("pandorra"), &["pandora"]),
            (dictgen::InsensitiveStr::Ascii("cancuks"), &["canucks"]),
            (dictgen::InsensitiveStr::Ascii("retrviever"), &["retriever"]),
            (dictgen::InsensitiveStr::Ascii("ypou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("survays"), &["surveys"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutoin"),
                &["revolutions", "revolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("lating"), &["latin"]),
            (
                dictgen::InsensitiveStr::Ascii("apppropriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostles"), &["hostels"]),
            (dictgen::InsensitiveStr::Ascii("crutialy"), &["crucially"]),
            (
                dictgen::InsensitiveStr::Ascii("reparamterised"),
                &["reparameterised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistuency"),
                &["constituency", "consistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critizising"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitilizes"),
                &["uninitializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("superheo"), &["superhero"]),
            (dictgen::InsensitiveStr::Ascii("redesing"), &["redesign"]),
            (
                dictgen::InsensitiveStr::Ascii("appeciative"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premultiplcation"),
                &["premultiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("pripheral"), &["peripheral"]),
            (
                dictgen::InsensitiveStr::Ascii("sumbodules"),
                &["submodules"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionning"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvemnt"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarterbackers"),
                &["quarterbacks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opereations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotangens"), &["cotangents"]),
            (
                dictgen::InsensitiveStr::Ascii("unreigsters"),
                &["unregisters"],
            ),
            (dictgen::InsensitiveStr::Ascii("positionn"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("concurreny"),
                &["concurrency"],
            ),
            (dictgen::InsensitiveStr::Ascii("recend"), &["rescind"]),
            (dictgen::InsensitiveStr::Ascii("commputing"), &["computing"]),
            (dictgen::InsensitiveStr::Ascii("exorcizer"), &["exorcist"]),
            (
                dictgen::InsensitiveStr::Ascii("specificically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropritate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphedamines"),
                &["amphetamines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influneced"),
                &["influenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concequences"),
                &["consequences"],
            ),
            (dictgen::InsensitiveStr::Ascii("botloader"), &["bootloader"]),
            (dictgen::InsensitiveStr::Ascii("legnths"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("instersections"),
                &["intersections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executeble"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parametres"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("mtuually"), &["mutually"]),
            (dictgen::InsensitiveStr::Ascii("susseeded"), &["seceded"]),
            (
                dictgen::InsensitiveStr::Ascii("acquistion"),
                &["acquisition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industructible"),
                &["indestructible"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiaton"), &["initiation"]),
            (
                dictgen::InsensitiveStr::Ascii("distribtution"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("insteead"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("fasle"), &["false"]),
            (dictgen::InsensitiveStr::Ascii("minimimum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("verticel"), &["vertical"]),
            (dictgen::InsensitiveStr::Ascii("seperatly"), &["separately"]),
            (
                dictgen::InsensitiveStr::Ascii("appoitments"),
                &["appointments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temprararly"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchronus"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptuve"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagerated"),
                &["exaggerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("arsnic"), &["arsenic"]),
            (dictgen::InsensitiveStr::Ascii("shwo"), &["show"]),
            (dictgen::InsensitiveStr::Ascii("anroids"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("choronological"),
                &["chronological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("currate"), &["curate"]),
            (
                dictgen::InsensitiveStr::Ascii("accumalated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("capmbell"), &["campbell"]),
            (
                dictgen::InsensitiveStr::Ascii("documenetations"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("switerzland"),
                &["switzerland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intalleing"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("internetts"), &["internets"]),
            (dictgen::InsensitiveStr::Ascii("sectoning"), &["sectioning"]),
            (
                dictgen::InsensitiveStr::Ascii("checkings"),
                &["checkins", "checks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiable"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("thinigs"), &["things"]),
            (dictgen::InsensitiveStr::Ascii("carrige"), &["carriage"]),
            (
                dictgen::InsensitiveStr::Ascii("cances"),
                &["cancel", "cancels", "chances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimtenations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ccontrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("pennsylvannia"),
                &["pennsylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaneouly"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("conflics"), &["conflicts"]),
            (
                dictgen::InsensitiveStr::Ascii("symobolically"),
                &["symbolically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idemopotent"),
                &["idempotent"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetrified"), &["certified"]),
            (dictgen::InsensitiveStr::Ascii("qwessant"), &["croissant"]),
            (
                dictgen::InsensitiveStr::Ascii("synthesesia"),
                &["synthesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("principall"), &["principal"]),
            (
                dictgen::InsensitiveStr::Ascii("recocnised"),
                &["recognised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfactorilly"),
                &["satisfactory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrction"),
                &["restriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdending"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("multipe"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("despatcher"),
                &["dispatcher"],
            ),
            (dictgen::InsensitiveStr::Ascii("ktichen"), &["kitchen"]),
            (dictgen::InsensitiveStr::Ascii("ehance"), &["enhance"]),
            (dictgen::InsensitiveStr::Ascii("cilyndrs"), &["cylinders"]),
            (
                dictgen::InsensitiveStr::Ascii("unregualted"),
                &["unregulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminaed"), &["terminated"]),
            (dictgen::InsensitiveStr::Ascii("samling"), &["sampling"]),
            (dictgen::InsensitiveStr::Ascii("evem"), &["even", "ever"]),
            (
                dictgen::InsensitiveStr::Ascii("botivational"),
                &["motivational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribuo"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("engeneer"), &["engineer"]),
            (dictgen::InsensitiveStr::Ascii("garentee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("congnitive"), &["cognitive"]),
            (dictgen::InsensitiveStr::Ascii("marxisits"), &["marxists"]),
            (
                dictgen::InsensitiveStr::Ascii("instumental"),
                &["instrumental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impelementation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destanations"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("contrieve"), &["contrive"]),
            (
                dictgen::InsensitiveStr::Ascii("associcates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("aboved"), &["above"]),
            (dictgen::InsensitiveStr::Ascii("strutcure"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("appointement"),
                &["appointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("windowz"), &["windows"]),
            (
                dictgen::InsensitiveStr::Ascii("prosterity"),
                &["prosperity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insteall"),
                &["install", "instead"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colloqiual"),
                &["colloquial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("welathiest"),
                &["wealthiest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cigarretts"),
                &["cigarettes"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitalizes"),
                &["uninitializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("neigborgh"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("totalitarion"),
                &["totalitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("genearated"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("unwirtten"), &["unwritten"]),
            (dictgen::InsensitiveStr::Ascii("confesos"), &["confess"]),
            (dictgen::InsensitiveStr::Ascii("importnt"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("combiatorics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propotional"),
                &["proportional", "promotional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifiate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("algotihm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("exprensive"),
                &["expressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ligthen"), &["lighten"]),
            (
                dictgen::InsensitiveStr::Ascii("canonivally"),
                &["canonically"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaiing"), &["locating"]),
            (dictgen::InsensitiveStr::Ascii("contatins"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("interseccion"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("accss"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("oje"), &["one"]),
            (dictgen::InsensitiveStr::Ascii("desctruct"), &["destruct"]),
            (
                dictgen::InsensitiveStr::Ascii("categoizes"),
                &["categorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("effecks"), &["effects"]),
            (dictgen::InsensitiveStr::Ascii("analysises"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("overpowed"),
                &["overpowered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactherised"),
                &["characterised"],
            ),
            (dictgen::InsensitiveStr::Ascii("anytghing"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("convulated"),
                &["convoluted"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaitons"), &["locations"]),
            (dictgen::InsensitiveStr::Ascii("jniw"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("procotols"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("seduciton"), &["seduction"]),
            (
                dictgen::InsensitiveStr::Ascii("explainatory"),
                &["explanatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("cralwed"), &["crawled"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawning"),
                &["withdrawing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirtes"), &["writes"]),
            (dictgen::InsensitiveStr::Ascii("subelemet"), &["subelement"]),
            (dictgen::InsensitiveStr::Ascii("mertic"), &["metric"]),
            (
                dictgen::InsensitiveStr::Ascii("preseleted"),
                &["preselected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternavtely"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("activies"), &["activities"]),
            (dictgen::InsensitiveStr::Ascii("truncatd"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("secuences"), &["sequences"]),
            (
                dictgen::InsensitiveStr::Ascii("ceratively"),
                &["creatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("patcket"), &["packet"]),
            (
                dictgen::InsensitiveStr::Ascii("retargetting"),
                &["retargeting"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekd"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("creaeting"), &["creating"]),
            (dictgen::InsensitiveStr::Ascii("wrapepd"), &["wrapped"]),
            (
                dictgen::InsensitiveStr::Ascii("asthetical"),
                &["aesthetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assimilare"),
                &["assimilate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certi"),
                &["certificate", "certify"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revove"),
                &["revolve", "remove"],
            ),
            (dictgen::InsensitiveStr::Ascii("mappin"), &["mapping"]),
            (dictgen::InsensitiveStr::Ascii("originas"), &["origins"]),
            (
                dictgen::InsensitiveStr::Ascii("conraction"),
                &["contraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pertubations"),
                &["perturbations"],
            ),
            (dictgen::InsensitiveStr::Ascii("handedy"), &["handedly"]),
            (dictgen::InsensitiveStr::Ascii("archviers"), &["archivers"]),
            (dictgen::InsensitiveStr::Ascii("recrusion"), &["recursion"]),
            (dictgen::InsensitiveStr::Ascii("mebmers"), &["members"]),
            (dictgen::InsensitiveStr::Ascii("komma"), &["coma", "comma"]),
            (dictgen::InsensitiveStr::Ascii("angshus"), &["anxious"]),
            (dictgen::InsensitiveStr::Ascii("downgradei"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("initiializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("impusle"), &["impulse"]),
            (dictgen::InsensitiveStr::Ascii("utitities"), &["utilities"]),
            (
                dictgen::InsensitiveStr::Ascii("tangencially"),
                &["tangentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("styrofaom"), &["styrofoam"]),
            (dictgen::InsensitiveStr::Ascii("limititing"), &["limiting"]),
            (dictgen::InsensitiveStr::Ascii("compilr"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("optimzation"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doucmenters"),
                &["documenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("importd"), &["imported"]),
            (dictgen::InsensitiveStr::Ascii("tempral"), &["temporal"]),
            (dictgen::InsensitiveStr::Ascii("scehdule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("unsage"),
                &["unsafe", "usage", "unstage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provicative"),
                &["provocative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexperessions"),
                &["subexpressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("referered"), &["referred"]),
            (dictgen::InsensitiveStr::Ascii("rhapson"), &["raphson"]),
            (dictgen::InsensitiveStr::Ascii("evalating"), &["evaluating"]),
            (dictgen::InsensitiveStr::Ascii("varidic"), &["variadic"]),
            (
                dictgen::InsensitiveStr::Ascii("supllies"),
                &["supplies", "sullies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralelism"),
                &["parallelism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicriminator"),
                &["discriminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("opderands"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("aroara"), &["aurora"]),
            (dictgen::InsensitiveStr::Ascii("hapenns"), &["happens"]),
            (dictgen::InsensitiveStr::Ascii("ecomonic"), &["economic"]),
            (
                dictgen::InsensitiveStr::Ascii("alliegance"),
                &["allegiance"],
            ),
            (dictgen::InsensitiveStr::Ascii("expersion"), &["expression"]),
            (
                dictgen::InsensitiveStr::Ascii("enlightended"),
                &["enlightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("notifed"), &["notified"]),
            (dictgen::InsensitiveStr::Ascii("binidng"), &["binding"]),
            (dictgen::InsensitiveStr::Ascii("spesify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("subpecies"), &["subspecies"]),
            (
                dictgen::InsensitiveStr::Ascii("recipercity"),
                &["reciprocity"],
            ),
            (dictgen::InsensitiveStr::Ascii("inifinity"), &["infinity"]),
            (
                dictgen::InsensitiveStr::Ascii("implenting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("medifors"), &["metaphors"]),
            (
                dictgen::InsensitiveStr::Ascii("terminially"),
                &["terminally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundametals"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invers"),
                &["inverse", "invert"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceratures"), &["creatures"]),
            (
                dictgen::InsensitiveStr::Ascii("descvriptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychemedics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggessive"), &["aggressive"]),
            (
                dictgen::InsensitiveStr::Ascii("generlization"),
                &["generalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fanatsizing"),
                &["fantasizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("subquue"), &["subqueue"]),
            (dictgen::InsensitiveStr::Ascii("agreeded"), &["agreed"]),
            (dictgen::InsensitiveStr::Ascii("calescing"), &["coalescing"]),
            (
                dictgen::InsensitiveStr::Ascii("alcoholisim"),
                &["alcoholism"],
            ),
            (dictgen::InsensitiveStr::Ascii("costants"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("expecte"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("undesireable"),
                &["undesirable"],
            ),
            (dictgen::InsensitiveStr::Ascii("stelathy"), &["stealthy"]),
            (dictgen::InsensitiveStr::Ascii("wrokflow"), &["workflow"]),
            (
                dictgen::InsensitiveStr::Ascii("centrifigual"),
                &["centrifugal"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreand"), &["operand"]),
            (
                dictgen::InsensitiveStr::Ascii("statese"),
                &["statuses", "states"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("graviometry"),
                &["gravimetry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegatarians"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("minstries"), &["ministries"]),
            (
                dictgen::InsensitiveStr::Ascii("cotnaining"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("valuin"), &["valuing"]),
            (
                dictgen::InsensitiveStr::Ascii("procuding"),
                &["procuring", "producing"],
            ),
            (dictgen::InsensitiveStr::Ascii("picthes"), &["pitches"]),
            (dictgen::InsensitiveStr::Ascii("switchters"), &["switchers"]),
            (
                dictgen::InsensitiveStr::Ascii("annihilatin"),
                &["annihilating", "annihilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhods"),
                &["neighbourhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("messging"),
                &["messaging", "messing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speperateing"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("disscus"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("ublicly"), &["publicly"]),
            (dictgen::InsensitiveStr::Ascii("fromatter"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetnel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("dellocate"), &["deallocate"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappered"),
                &["disappeared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("withdrawls"),
                &["withdrawals", "withdraws"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stablilization"),
                &["stabilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequentail"),
                &["sequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtted"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("revewrse"), &["reverse"]),
            (
                dictgen::InsensitiveStr::Ascii("mainfestation"),
                &["manifestation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantiation"),
                &["quantization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overries"),
                &["overrides", "ovaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realy"),
                &["really", "relay", "real"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporalily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consiciousness"),
                &["consciousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowldge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("hepling"), &["helping"]),
            (dictgen::InsensitiveStr::Ascii("upagrades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("dprecate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("negativley"),
                &["negatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verifikation"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("misreable"), &["miserable"]),
            (
                dictgen::InsensitiveStr::Ascii("susceptiable"),
                &["susceptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcationed"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("breackets"), &["brackets"]),
            (dictgen::InsensitiveStr::Ascii("depositis"), &["deposits"]),
            (
                dictgen::InsensitiveStr::Ascii("inlightening"),
                &["enlightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succeesive"),
                &["successive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consifered"),
                &["considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indifferant"),
                &["indifferent"],
            ),
            (dictgen::InsensitiveStr::Ascii("indcate"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("differenent"),
                &["different"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practive"),
                &["practice", "proactive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatibil"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("nibmle"), &["nimble"]),
            (
                dictgen::InsensitiveStr::Ascii("datatstructure"),
                &["datastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("amplifiy"), &["amplify"]),
            (
                dictgen::InsensitiveStr::Ascii("incooperated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("supprot"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabillities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performace"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("ralations"), &["relations"]),
            (
                dictgen::InsensitiveStr::Ascii("controlllers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("signautre"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("psrameter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("audioboook"), &["audiobook"]),
            (
                dictgen::InsensitiveStr::Ascii("apretiates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("piar"),
                &["pair", "pier", "pliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("identifcal"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("renaissanse"),
                &["renaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perosnality"),
                &["personality"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoerted"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("concurrectly"),
                &["concurrently"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbabling"), &["enabling"]),
            (dictgen::InsensitiveStr::Ascii("bynasty"), &["dynasty"]),
            (dictgen::InsensitiveStr::Ascii("whsiper"), &["whisper"]),
            (dictgen::InsensitiveStr::Ascii("suprass"), &["surpass"]),
            (dictgen::InsensitiveStr::Ascii("responeded"), &["responded"]),
            (
                dictgen::InsensitiveStr::Ascii("verifcations"),
                &["verifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminitions"),
                &["admonitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufficiennt"),
                &["sufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparatusess"),
                &["apparatuses"],
            ),
            (dictgen::InsensitiveStr::Ascii("animeting"), &["animating"]),
            (dictgen::InsensitiveStr::Ascii("taxonmy"), &["taxonomy"]),
            (dictgen::InsensitiveStr::Ascii("seracher"), &["searcher"]),
            (
                dictgen::InsensitiveStr::Ascii("corispondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unimplemened"),
                &["unimplemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpary"),
                &["counterplay"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtitectural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintersting"),
                &["uninteresting"],
            ),
            (dictgen::InsensitiveStr::Ascii("thereshold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("getoject"), &["getobject"]),
            (
                dictgen::InsensitiveStr::Ascii("trasmitted"),
                &["transmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("broadacasting"),
                &["broadcasting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanisng"), &["cleansing"]),
            (dictgen::InsensitiveStr::Ascii("brigther"), &["brighter"]),
            (
                dictgen::InsensitiveStr::Ascii("colllection"),
                &["collection"],
            ),
            (dictgen::InsensitiveStr::Ascii("redefinied"), &["redefine"]),
            (
                dictgen::InsensitiveStr::Ascii("sepearating"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("playright"), &["playwright"]),
            (dictgen::InsensitiveStr::Ascii("throuhput"), &["throughput"]),
            (dictgen::InsensitiveStr::Ascii("widespred"), &["widespread"]),
            (
                dictgen::InsensitiveStr::Ascii("adrerssing"),
                &["addressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suprimacist"),
                &["supremacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacestomy"), &["vasectomy"]),
            (dictgen::InsensitiveStr::Ascii("legnedary"), &["legendary"]),
            (dictgen::InsensitiveStr::Ascii("constuct"), &["construct"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregaters"),
                &["aggregators", "aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disptacher"),
                &["dispatcher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnanteing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificsted"),
                &["certificated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aftrerwards"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemends"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insignifigant"),
                &["insignificant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspendents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("proclam"), &["proclaim"]),
            (
                dictgen::InsensitiveStr::Ascii("automatially"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("transcocde"), &["transcode"]),
            (dictgen::InsensitiveStr::Ascii("loacalhost"), &["localhost"]),
            (dictgen::InsensitiveStr::Ascii("acording"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("courtesty"), &["courtesy"]),
            (dictgen::InsensitiveStr::Ascii("enabel"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("decrations"),
                &["decorations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asymtomatically"),
                &["asymptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("provoder"), &["provider"]),
            (dictgen::InsensitiveStr::Ascii("hysterica"), &["hysteria"]),
            (
                dictgen::InsensitiveStr::Ascii("conslutant"),
                &["consultant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaccinato"),
                &["vaccination"],
            ),
            (dictgen::InsensitiveStr::Ascii("reseted"), &["reset"]),
            (dictgen::InsensitiveStr::Ascii("prescius"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("disptach"), &["dispatch"]),
            (dictgen::InsensitiveStr::Ascii("descuss"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("simlator"), &["simulator"]),
            (dictgen::InsensitiveStr::Ascii("flacors"), &["flavors"]),
            (dictgen::InsensitiveStr::Ascii("pertinate"), &["pertinent"]),
            (dictgen::InsensitiveStr::Ascii("fliter"), &["filter"]),
            (
                dictgen::InsensitiveStr::Ascii("couses"),
                &["courses", "causes"],
            ),
            (dictgen::InsensitiveStr::Ascii("eclise"), &["eclipse"]),
            (dictgen::InsensitiveStr::Ascii("tringkets"), &["trinkets"]),
            (dictgen::InsensitiveStr::Ascii("websties"), &["websites"]),
            (
                dictgen::InsensitiveStr::Ascii("repeatition"),
                &["repetition"],
            ),
            (dictgen::InsensitiveStr::Ascii("aristote"), &["aristotle"]),
            (dictgen::InsensitiveStr::Ascii("hanlded"), &["handled"]),
            (
                dictgen::InsensitiveStr::Ascii("californina"),
                &["californian"],
            ),
            (dictgen::InsensitiveStr::Ascii("blcok"), &["block"]),
            (
                dictgen::InsensitiveStr::Ascii("servive"),
                &["service", "survive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demolicion"),
                &["demolition"],
            ),
            (dictgen::InsensitiveStr::Ascii("behvaiors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtuting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("earliear"), &["earlier"]),
            (
                dictgen::InsensitiveStr::Ascii("reccurrences"),
                &["recurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoledging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compairson"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrubure"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("ciritcs"), &["critics"]),
            (
                dictgen::InsensitiveStr::Ascii("tring"),
                &["trying", "string", "ring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validadions"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cliffhager"),
                &["cliffhanger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplorabel"),
                &["deplorable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalueble"),
                &["invaluable"],
            ),
            (dictgen::InsensitiveStr::Ascii("propmpt"), &["prompt"]),
            (dictgen::InsensitiveStr::Ascii("poisonus"), &["poisons"]),
            (dictgen::InsensitiveStr::Ascii("intefrace"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("correspndence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfiing"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("daugthers"), &["daughters"]),
            (
                dictgen::InsensitiveStr::Ascii("descendenta"),
                &["descendents"],
            ),
            (dictgen::InsensitiveStr::Ascii("naxima"), &["maxima"]),
            (dictgen::InsensitiveStr::Ascii("destrois"), &["destroys"]),
            (dictgen::InsensitiveStr::Ascii("exempel"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("gerate"),
                &["generate", "gyrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccope"), &["scope"]),
            (
                dictgen::InsensitiveStr::Ascii("delibitating"),
                &["debilitating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviorment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetically"), &["vertically"]),
            (dictgen::InsensitiveStr::Ascii("territorry"), &["territory"]),
            (
                dictgen::InsensitiveStr::Ascii("satuation"),
                &["saturation", "situation"],
            ),
            (dictgen::InsensitiveStr::Ascii("lightining"), &["lighting"]),
            (dictgen::InsensitiveStr::Ascii("differnent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("undestands"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succesdsful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulits"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("resistence"),
                &["resistance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positiong"),
                &["positioning", "position"],
            ),
            (dictgen::InsensitiveStr::Ascii("aceept"), &["accept"]),
            (dictgen::InsensitiveStr::Ascii("keybiding"), &["keybinding"]),
            (dictgen::InsensitiveStr::Ascii("esudo"), &["sudo"]),
            (
                dictgen::InsensitiveStr::Ascii("unmistakeably"),
                &["unmistakably"],
            ),
            (dictgen::InsensitiveStr::Ascii("brusies"), &["bruises"]),
            (
                dictgen::InsensitiveStr::Ascii("caluculation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememberes"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("evaulator"), &["evaluator"]),
            (
                dictgen::InsensitiveStr::Ascii("pictureskely"),
                &["picturesquely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assembalge"),
                &["assemblage"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanlder"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("dahboard"), &["dashboard"]),
            (
                dictgen::InsensitiveStr::Ascii("adminstrator"),
                &["administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adborbed"),
                &["adsorbed", "absorbed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictonnary"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("premiee"), &["premiere"]),
            (
                dictgen::InsensitiveStr::Ascii("insruction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowledgeble"),
                &["knowledgeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identtities"),
                &["identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("intutive"), &["intuitive"]),
            (
                dictgen::InsensitiveStr::Ascii("administartor"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("archine"), &["archive"]),
            (
                dictgen::InsensitiveStr::Ascii("replaca"),
                &["replica", "replace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphacized"),
                &["emphasized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justificatin"),
                &["justifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("catlayst"), &["catalyst"]),
            (dictgen::InsensitiveStr::Ascii("ghandi"), &["gandhi"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretaion"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defintiions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("disposel"), &["disposal"]),
            (dictgen::InsensitiveStr::Ascii("derprecate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("mingiame"), &["minigame"]),
            (dictgen::InsensitiveStr::Ascii("oprations"), &["operations"]),
            (dictgen::InsensitiveStr::Ascii("intagrate"), &["integrate"]),
            (
                dictgen::InsensitiveStr::Ascii("electrisity"),
                &["electricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("restict"), &["restrict"]),
            (dictgen::InsensitiveStr::Ascii("numberals"), &["numerals"]),
            (
                dictgen::InsensitiveStr::Ascii("preinted"),
                &["printed", "preprinted"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconects"), &["reconnects"]),
            (dictgen::InsensitiveStr::Ascii("mayonase"), &["mayonnaise"]),
            (dictgen::InsensitiveStr::Ascii("creaete"), &["create"]),
            (
                dictgen::InsensitiveStr::Ascii("acitvations"),
                &["activations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kalidescopes"),
                &["kaleidoscopes"],
            ),
            (dictgen::InsensitiveStr::Ascii("transeint"), &["transient"]),
            (
                dictgen::InsensitiveStr::Ascii("abbreviatons"),
                &["abbreviations"],
            ),
            (dictgen::InsensitiveStr::Ascii("appplied"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("cureful"), &["careful"]),
            (
                dictgen::InsensitiveStr::Ascii("adge"),
                &["edge", "badge", "adage"],
            ),
            (dictgen::InsensitiveStr::Ascii("geenrated"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("realoding"), &["reloading"]),
            (dictgen::InsensitiveStr::Ascii("reaserch"), &["research"]),
            (dictgen::InsensitiveStr::Ascii("guarnatee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("trgger"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("carnberry"), &["cranberry"]),
            (dictgen::InsensitiveStr::Ascii("milicous"), &["malicious"]),
            (
                dictgen::InsensitiveStr::Ascii("understantably"),
                &["understandably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincedences"),
                &["coincidences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticater"),
                &[
                    "authenticator",
                    "authenticated",
                    "authenticates",
                    "authenticate",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pipelinining"),
                &["pipelining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astronouts"),
                &["astronauts"],
            ),
            (dictgen::InsensitiveStr::Ascii("apolgize"), &["apologize"]),
            (dictgen::InsensitiveStr::Ascii("beforing"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("wndow"), &["window"]),
            (dictgen::InsensitiveStr::Ascii("himslef"), &["himself"]),
            (
                dictgen::InsensitiveStr::Ascii("truct"),
                &["struct", "trust", "truck", "tract", "truce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fals"),
                &["false", "fails", "falls"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotionators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("timespanps"), &["timespans"]),
            (dictgen::InsensitiveStr::Ascii("importerad"), &["imported"]),
            (
                dictgen::InsensitiveStr::Ascii("managebale"),
                &["manageable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxites"), &["excites"]),
            (dictgen::InsensitiveStr::Ascii("bacause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("transmitts"), &["transmits"]),
            (dictgen::InsensitiveStr::Ascii("recievs"), &["receives"]),
            (dictgen::InsensitiveStr::Ascii("pallettes"), &["palettes"]),
            (dictgen::InsensitiveStr::Ascii("unitss"), &["units"]),
            (
                dictgen::InsensitiveStr::Ascii("demostration"),
                &["demonstration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnanges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("hirearchical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intevening"),
                &["intervening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electromagnatic"),
                &["electromagnetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("containig"), &["containing"]),
            (
                dictgen::InsensitiveStr::Ascii("singool"),
                &["single", "signal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatibilty"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapperared"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeg"), &["open"]),
            (dictgen::InsensitiveStr::Ascii("yelded"), &["yielded"]),
            (dictgen::InsensitiveStr::Ascii("moent"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("motivationnal"),
                &["motivational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comtaminated"),
                &["contaminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountas"),
                &["accountants", "accounts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparity"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("vunrable"), &["vulnerable"]),
            (
                dictgen::InsensitiveStr::Ascii("implicatons"),
                &["implications"],
            ),
            (dictgen::InsensitiveStr::Ascii("metarials"), &["materials"]),
            (dictgen::InsensitiveStr::Ascii("builter"), &["builder"]),
            (dictgen::InsensitiveStr::Ascii("entirerly"), &["entirely"]),
            (dictgen::InsensitiveStr::Ascii("disapperar"), &["disappear"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoind"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("weree"), &["were"]),
            (dictgen::InsensitiveStr::Ascii("pompts"), &["prompts"]),
            (dictgen::InsensitiveStr::Ascii("botom"), &["bottom"]),
            (
                dictgen::InsensitiveStr::Ascii("curcumstantial"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvment"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("outragous"), &["outrageous"]),
            (
                dictgen::InsensitiveStr::Ascii("appearantly"),
                &["apparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("extraced"), &["extracted"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronizes"),
                &["synchronizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("fobid"), &["forbid"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterizes"),
                &["characterizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sarcams"), &["sarcasm"]),
            (
                dictgen::InsensitiveStr::Ascii("psychedellic"),
                &["psychedelic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reappering"),
                &["reappearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("narhwal"), &["narwhal"]),
            (
                dictgen::InsensitiveStr::Ascii("copengagen"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("campany"), &["company"]),
            (
                dictgen::InsensitiveStr::Ascii("policys"),
                &["police", "policies"],
            ),
            (dictgen::InsensitiveStr::Ascii("mocrocode"), &["microcode"]),
            (dictgen::InsensitiveStr::Ascii("iterrate"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("recogised"), &["recognised"]),
            (
                dictgen::InsensitiveStr::Ascii("purpse"),
                &["purpose", "purse", "purple"],
            ),
            (dictgen::InsensitiveStr::Ascii("cretators"), &["creators"]),
            (
                dictgen::InsensitiveStr::Ascii("jurisidctions"),
                &["jurisdictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("jumo"), &["jump"]),
            (
                dictgen::InsensitiveStr::Ascii("relashionships"),
                &["relationships"],
            ),
            (dictgen::InsensitiveStr::Ascii("unwrritten"), &["unwritten"]),
            (dictgen::InsensitiveStr::Ascii("algorithym"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiifes"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("accquiring"), &["acquiring"]),
            (
                dictgen::InsensitiveStr::Ascii("abbriviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermodynamcis"),
                &["thermodynamics"],
            ),
            (dictgen::InsensitiveStr::Ascii("improvess"), &["improves"]),
            (
                dictgen::InsensitiveStr::Ascii("instantitation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitalising"),
                &["reinitialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pontentials"),
                &["potentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("zukeenees"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("regresas"), &["regress"]),
            (
                dictgen::InsensitiveStr::Ascii("coveniences"),
                &["conveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucsess"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("diminsihing"),
                &["diminishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("erturned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("headahces"), &["headaches"]),
            (
                dictgen::InsensitiveStr::Ascii("cleander"),
                &["cleaner", "calendar", "cleanser"],
            ),
            (dictgen::InsensitiveStr::Ascii("encodnings"), &["encodings"]),
            (dictgen::InsensitiveStr::Ascii("cyclinders"), &["cylinders"]),
            (
                dictgen::InsensitiveStr::Ascii("interaxction"),
                &["interaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbiratily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("laready"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("annay"), &["annoy", "any"]),
            (dictgen::InsensitiveStr::Ascii("brewerey"), &["brewery"]),
            (
                dictgen::InsensitiveStr::Ascii("dublication"),
                &["duplication", "publication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiest"),
                &["enthusiast"],
            ),
            (dictgen::InsensitiveStr::Ascii("mecanism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("architecturs"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disengenuous"),
                &["disingenuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutermeasures"),
                &["countermeasures"],
            ),
            (dictgen::InsensitiveStr::Ascii("apppend"), &["append"]),
            (
                dictgen::InsensitiveStr::Ascii("represenations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cluase"), &["clause"]),
            (
                dictgen::InsensitiveStr::Ascii("provies"),
                &["provides", "proves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtiects"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calibraitons"),
                &["calibrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permutaions"),
                &["permutations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ghostscritp"),
                &["ghostscript"],
            ),
            (dictgen::InsensitiveStr::Ascii("mindlessy"), &["mindlessly"]),
            (
                dictgen::InsensitiveStr::Ascii("discconeection"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typicall"),
                &["typically", "typical"],
            ),
            (dictgen::InsensitiveStr::Ascii("organims"), &["organism"]),
            (
                dictgen::InsensitiveStr::Ascii("federacion"),
                &["federation"],
            ),
            (dictgen::InsensitiveStr::Ascii("startuip"), &["startup"]),
            (
                dictgen::InsensitiveStr::Ascii("positivisme"),
                &["positives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regitration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cariied"), &["carried"]),
            (dictgen::InsensitiveStr::Ascii("skepticons"), &["skeptics"]),
            (
                dictgen::InsensitiveStr::Ascii("vegeterians"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acctions"),
                &["actions", "auctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("reductins"), &["reductions"]),
            (
                dictgen::InsensitiveStr::Ascii("superifical"),
                &["superficial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplacations"),
                &["duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("varible"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("physiologial"),
                &["physiological"],
            ),
            (dictgen::InsensitiveStr::Ascii("invlaid"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("failicies"), &["facilities"]),
            (dictgen::InsensitiveStr::Ascii("coallesces"), &["coalesces"]),
            (dictgen::InsensitiveStr::Ascii("pachages"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("simpley"), &["simply"]),
            (dictgen::InsensitiveStr::Ascii("pilar"), &["pillar"]),
            (dictgen::InsensitiveStr::Ascii("reacahable"), &["reachable"]),
            (
                dictgen::InsensitiveStr::Ascii("charactheristically"),
                &["characteristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmnetations"),
                &["segmentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("coommitter"), &["committer"]),
            (dictgen::InsensitiveStr::Ascii("parsially"), &["partially"]),
            (
                dictgen::InsensitiveStr::Ascii("tempuratures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shopkeeepers"),
                &["shopkeepers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralellism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("retriving"), &["retrieving"]),
            (dictgen::InsensitiveStr::Ascii("paramss"), &["params"]),
            (dictgen::InsensitiveStr::Ascii("logicaly"), &["logically"]),
            (
                dictgen::InsensitiveStr::Ascii("maybellibe"),
                &["maybelline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desserialisation"),
                &["deserialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dysphora"), &["dysphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("documening"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranforming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refleciton"),
                &["reflections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitutuent"),
                &["constituent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpencies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("amrstrong"), &["armstrong"]),
            (dictgen::InsensitiveStr::Ascii("accces"), &["access"]),
            (
                dictgen::InsensitiveStr::Ascii("brotherood"),
                &["brotherhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaialbe"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("althorithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("neame"), &["name"]),
            (
                dictgen::InsensitiveStr::Ascii("substracting"),
                &["subtracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("awrnings"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("triangel"), &["triangle"]),
            (
                dictgen::InsensitiveStr::Ascii("bounadries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribuition"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amerliorating"),
                &["ameliorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("comfuses"), &["confuses"]),
            (dictgen::InsensitiveStr::Ascii("cehcklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("ampethamine"),
                &["amphetamine"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaptare"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("srcolling"), &["scrolling"]),
            (dictgen::InsensitiveStr::Ascii("acending"), &["ascending"]),
            (dictgen::InsensitiveStr::Ascii("disbles"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("technqiues"),
                &["techniques"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccessors"), &["accessors"]),
            (dictgen::InsensitiveStr::Ascii("visibile"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("impletation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistmatch"), &["mismatch"]),
            (
                dictgen::InsensitiveStr::Ascii("qualfiiers"),
                &["qualifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegetarianos"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("speficilliers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcationality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vernaculaire"),
                &["vernacular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propereties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentatlly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anialated"),
                &["annihilated"],
            ),
            (dictgen::InsensitiveStr::Ascii("botstraps"), &["bootstraps"]),
            (dictgen::InsensitiveStr::Ascii("exand"), &["expand"]),
            (
                dictgen::InsensitiveStr::Ascii("colelctive"),
                &["collective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheerleards"),
                &["cheerleaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknownledgments"),
                &["acknowledgments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electrocity"),
                &["electricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("retirase"), &["retires"]),
            (
                dictgen::InsensitiveStr::Ascii("inhertig"),
                &["inheriting", "inherited"],
            ),
            (dictgen::InsensitiveStr::Ascii("opion"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("unserstanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinsiter"), &["sinister"]),
            (dictgen::InsensitiveStr::Ascii("undeneath"), &["underneath"]),
            (dictgen::InsensitiveStr::Ascii("sohuld"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("punctation"),
                &["punctuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antisipating"),
                &["anticipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("containted"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("hightlights"),
                &["highlights"],
            ),
            (dictgen::InsensitiveStr::Ascii("dyanmite"), &["dynamite"]),
            (
                dictgen::InsensitiveStr::Ascii("itnerprelation"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inted"),
                &["inetd", "intend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlisation"),
                &["generalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("createur"), &["creature"]),
            (dictgen::InsensitiveStr::Ascii("suggessted"), &["suggested"]),
            (dictgen::InsensitiveStr::Ascii("craced"), &["graced"]),
            (dictgen::InsensitiveStr::Ascii("divion"), &["division"]),
            (
                dictgen::InsensitiveStr::Ascii("southernors"),
                &["southerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ervices"),
                &["services", "cervices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definetily"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("suffaces"), &["surfaces"]),
            (dictgen::InsensitiveStr::Ascii("bottm"), &["bottom"]),
            (
                dictgen::InsensitiveStr::Ascii("dections"),
                &[
                    "detections",
                    "sections",
                    "directions",
                    "deductions",
                    "depictions",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specfiiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previosuly"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pestisides"),
                &["pesticides"],
            ),
            (dictgen::InsensitiveStr::Ascii("tsamina"), &["stamina"]),
            (dictgen::InsensitiveStr::Ascii("buissness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("rqeuests"),
                &["requests", "quests"],
            ),
            (dictgen::InsensitiveStr::Ascii("basod"), &["based"]),
            (dictgen::InsensitiveStr::Ascii("berfore"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("cymptumaticly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("decoposes"), &["decomposes"]),
            (dictgen::InsensitiveStr::Ascii("obious"), &["obvious"]),
            (
                dictgen::InsensitiveStr::Ascii("experinemented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreemer"),
                &["upstreamer"],
            ),
            (dictgen::InsensitiveStr::Ascii("vrify"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("validtor"), &["validator"]),
            (
                dictgen::InsensitiveStr::Ascii("convserion"),
                &["conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("strategis"), &["strategies"]),
            (dictgen::InsensitiveStr::Ascii("throen"), &["thrown"]),
            (dictgen::InsensitiveStr::Ascii("mdifier"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("bottlenck"), &["bottleneck"]),
            (dictgen::InsensitiveStr::Ascii("cynisicm"), &["cynicism"]),
            (
                dictgen::InsensitiveStr::Ascii("tacticallity"),
                &["tactically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubstantiative"),
                &["unsubstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("optionnal"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("coneninet"), &["convenient"]),
            (
                dictgen::InsensitiveStr::Ascii("normall"),
                &["normal", "normally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonizations"),
                &["synchronizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifiactions"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vallidation"),
                &["validation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implenmenting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("blieve"), &["believe"]),
            (dictgen::InsensitiveStr::Ascii("speecified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("medicacion"),
                &["medication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chekclists"),
                &["checklists"],
            ),
            (dictgen::InsensitiveStr::Ascii("godspede"), &["godspeed"]),
            (
                dictgen::InsensitiveStr::Ascii("downloadmanger"),
                &["downloadmanager"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introuduces"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anpatheaters"),
                &["amphitheaters"],
            ),
            (dictgen::InsensitiveStr::Ascii("recources"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("scheduales"), &["schedules"]),
            (
                dictgen::InsensitiveStr::Ascii("reproduktion"),
                &["reproduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aqquaintances"),
                &["acquaintances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misoginist"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfomations"),
                &["transformations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prediously"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowleding"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("businesse"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoed"), &["exposed"]),
            (dictgen::InsensitiveStr::Ascii("unlikey"), &["unlikely"]),
            (
                dictgen::InsensitiveStr::Ascii("aborigenal"),
                &["aboriginal"],
            ),
            (dictgen::InsensitiveStr::Ascii("audo"), &["audio"]),
            (
                dictgen::InsensitiveStr::Ascii("migratin"),
                &["migrating", "migration"],
            ),
            (dictgen::InsensitiveStr::Ascii("maunally"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("incrimental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalising"),
                &["orthogonalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("omlet"), &["omelet"]),
            (
                dictgen::InsensitiveStr::Ascii("negitiators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("upppercase"), &["uppercase"]),
            (dictgen::InsensitiveStr::Ascii("browesr"), &["browser"]),
            (dictgen::InsensitiveStr::Ascii("itnernal"), &["internal"]),
            (
                dictgen::InsensitiveStr::Ascii("rolepalying"),
                &["roleplaying"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginnners"), &["beginners"]),
            (
                dictgen::InsensitiveStr::Ascii("sunglassses"),
                &["sunglasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("thniks"), &["thinks"]),
            (dictgen::InsensitiveStr::Ascii("cahange"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("subracted"), &["subtracted"]),
            (dictgen::InsensitiveStr::Ascii("curency"), &["currency"]),
            (
                dictgen::InsensitiveStr::Ascii("affectionnate"),
                &["affectionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("patials"), &["partials"]),
            (dictgen::InsensitiveStr::Ascii("stalkear"), &["stalker"]),
            (
                dictgen::InsensitiveStr::Ascii("coordintates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("insurence"), &["insurance"]),
            (dictgen::InsensitiveStr::Ascii("trriggered"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("maintnance"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessisary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("topolgoy"), &["topology"]),
            (
                dictgen::InsensitiveStr::Ascii("obtainabie"),
                &["obtainable"],
            ),
            (dictgen::InsensitiveStr::Ascii("indcated"), &["indicated"]),
            (
                dictgen::InsensitiveStr::Ascii("complimentoni"),
                &["complimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrupsed"), &["corrupted"]),
            (dictgen::InsensitiveStr::Ascii("databaase"), &["database"]),
            (
                dictgen::InsensitiveStr::Ascii("apporximate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahhel"), &["channel"]),
            (
                dictgen::InsensitiveStr::Ascii("accidentky"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeatios"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("impossibily"),
                &["impossibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singnatures"),
                &["signatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbrases"), &["embraces"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inumerable"),
                &["enumerable", "innumerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercontinetal"),
                &["intercontinental"],
            ),
            (dictgen::InsensitiveStr::Ascii("shceduled"), &["scheduled"]),
            (dictgen::InsensitiveStr::Ascii("selete"), &["delete"]),
            (
                dictgen::InsensitiveStr::Ascii("collequial"),
                &["colloquial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standarizes"),
                &["standardizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repersentation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muliplicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("threee"), &["three"]),
            (dictgen::InsensitiveStr::Ascii("eplicitly"), &["explicitly"]),
            (dictgen::InsensitiveStr::Ascii("glpyhs"), &["glyphs"]),
            (
                dictgen::InsensitiveStr::Ascii("elliminated"),
                &["eliminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremintal"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnformation"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exctracts"), &["extracts"]),
            (dictgen::InsensitiveStr::Ascii("catpure"), &["capture"]),
            (
                dictgen::InsensitiveStr::Ascii("adjustes"),
                &["adjusted", "adjusts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subredditors"),
                &["subreddits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("picure"),
                &["picture", "epicure"],
            ),
            (dictgen::InsensitiveStr::Ascii("diarrheoa"), &["diarrhea"]),
            (dictgen::InsensitiveStr::Ascii("dicovered"), &["discovered"]),
            (dictgen::InsensitiveStr::Ascii("excliudes"), &["excludes"]),
            (dictgen::InsensitiveStr::Ascii("archiect"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("responsbility"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifificate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermittient"),
                &["intermittent"],
            ),
            (dictgen::InsensitiveStr::Ascii("placmenet"), &["placement"]),
            (dictgen::InsensitiveStr::Ascii("enebled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("notorization"),
                &["notarization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dangerouly"),
                &["dangerously"],
            ),
            (dictgen::InsensitiveStr::Ascii("tmorrow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("painkills"),
                &["painkillers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jounralists"),
                &["journalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlog"), &["analog"]),
            (
                dictgen::InsensitiveStr::Ascii("beed"),
                &["bead", "been", "beer", "bees", "beet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitlety"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("randomely"), &["randomly"]),
            (
                dictgen::InsensitiveStr::Ascii("autheniticating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idetification"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtitile"), &["subtitle"]),
            (dictgen::InsensitiveStr::Ascii("strating"), &["starting"]),
            (dictgen::InsensitiveStr::Ascii("documentes"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallate"), &["collate"]),
            (
                dictgen::InsensitiveStr::Ascii("perliferates"),
                &["proliferates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sectopning"),
                &["sectioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("trafficed"), &["trafficked"]),
            (dictgen::InsensitiveStr::Ascii("sacaled"), &["scaled"]),
            (dictgen::InsensitiveStr::Ascii("returnd"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("remasterred"),
                &["remastered"],
            ),
            (dictgen::InsensitiveStr::Ascii("caputured"), &["captured"]),
            (dictgen::InsensitiveStr::Ascii("simbols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("milimetre"), &["millimetre"]),
            (
                dictgen::InsensitiveStr::Ascii("charactersm"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltartor"),
                &["infiltrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("pormised"), &["promised"]),
            (
                dictgen::InsensitiveStr::Ascii("equivalenet"),
                &["equivalents", "equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinisitic"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethime"), &["sometime"]),
            (
                dictgen::InsensitiveStr::Ascii("unregognised"),
                &["unrecognised"],
            ),
            (dictgen::InsensitiveStr::Ascii("strigs"), &["strings"]),
            (
                dictgen::InsensitiveStr::Ascii("multuplications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introspectin"),
                &["introspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsisent"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discusss"),
                &["discuss", "discusses"],
            ),
            (dictgen::InsensitiveStr::Ascii("clearely"), &["clearly"]),
            (dictgen::InsensitiveStr::Ascii("catergory"), &["category"]),
            (
                dictgen::InsensitiveStr::Ascii("expereincing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resturaunt"),
                &["restaurant"],
            ),
            (dictgen::InsensitiveStr::Ascii("tapitude"), &["aptitude"]),
            (dictgen::InsensitiveStr::Ascii("batistia"), &["batista"]),
            (
                dictgen::InsensitiveStr::Ascii("mathmatical"),
                &["mathematical"],
            ),
            (dictgen::InsensitiveStr::Ascii("appache"), &["apache"]),
            (
                dictgen::InsensitiveStr::Ascii("trinagulating"),
                &["triangulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("winodws"), &["windows"]),
            (dictgen::InsensitiveStr::Ascii("jouney"), &["journey"]),
            (
                dictgen::InsensitiveStr::Ascii("speficialleirs"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliverately"),
                &["deliberately"],
            ),
            (dictgen::InsensitiveStr::Ascii("balcanes"), &["balances"]),
            (
                dictgen::InsensitiveStr::Ascii("substential"),
                &["substantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("updatig"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("intregrating"),
                &["integrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibiblity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horicontally"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("inclduing"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("sighth"),
                &["scythe", "sight"],
            ),
            (dictgen::InsensitiveStr::Ascii("webapge"), &["webpage"]),
            (
                dictgen::InsensitiveStr::Ascii("playthourghs"),
                &["playthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anthropolgy"),
                &["anthropology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efore"),
                &["afore", "before"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strengthin"),
                &["strengthen"],
            ),
            (dictgen::InsensitiveStr::Ascii("embeeded"), &["embedded"]),
            (
                dictgen::InsensitiveStr::Ascii("considerstion"),
                &["considerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orangatang"),
                &["orangutang", "orangutan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assciations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diappearances"),
                &["disappearances"],
            ),
            (dictgen::InsensitiveStr::Ascii("invervals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("perscription"),
                &["prescription"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathches"), &["matches"]),
            (
                dictgen::InsensitiveStr::Ascii("differnence"),
                &["difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("resutl"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiamtely"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("modidfying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("inifite"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("simultanaeous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("abbouts"), &["abbots"]),
            (
                dictgen::InsensitiveStr::Ascii("similiarity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("aninator"), &["animator"]),
            (dictgen::InsensitiveStr::Ascii("marraige"), &["marriage"]),
            (dictgen::InsensitiveStr::Ascii("origials"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("reruirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforcemant"),
                &["enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("enalbe"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("monolistic"),
                &["monolithic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consuelling"),
                &["counselling"],
            ),
            (dictgen::InsensitiveStr::Ascii("breacking"), &["breaking"]),
            (
                dictgen::InsensitiveStr::Ascii("microtransactios"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breastfeeing"),
                &["breastfeeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("liftime"), &["lifetime"]),
            (
                dictgen::InsensitiveStr::Ascii("counrtyside"),
                &["countryside"],
            ),
            (dictgen::InsensitiveStr::Ascii("pattrn"), &["pattern"]),
            (dictgen::InsensitiveStr::Ascii("intrrupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("archiac"), &["archaic"]),
            (dictgen::InsensitiveStr::Ascii("nombered"), &["numbered"]),
            (dictgen::InsensitiveStr::Ascii("intereface"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("sophmore"), &["sophomore"]),
            (
                dictgen::InsensitiveStr::Ascii("dictionaires"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("euqivalently"),
                &["equivalently"],
            ),
            (dictgen::InsensitiveStr::Ascii("appraoch"), &["approach"]),
            (
                dictgen::InsensitiveStr::Ascii("differentiatiations"),
                &["differentiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delivatives"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("realtd"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitoning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circunstance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("aparent"), &["apparent"]),
            (
                dictgen::InsensitiveStr::Ascii("unavaialability"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arranable"),
                &["arrangeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensasional"),
                &["sensational"],
            ),
            (dictgen::InsensitiveStr::Ascii("enpoints"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("encryptiion"),
                &["encryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosistency"),
                &["consistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrct"), &["abstract"]),
            (dictgen::InsensitiveStr::Ascii("validasted"), &["validated"]),
            (
                dictgen::InsensitiveStr::Ascii("strigified"),
                &["stringified"],
            ),
            (dictgen::InsensitiveStr::Ascii("orhpan"), &["orphan"]),
            (
                dictgen::InsensitiveStr::Ascii("photographyi"),
                &["photographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegoptionsotiations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectaculaire"),
                &["spectacular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insensive"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compinsating"),
                &["compensating"],
            ),
            (dictgen::InsensitiveStr::Ascii("escuses"), &["excuses"]),
            (
                dictgen::InsensitiveStr::Ascii("mysteriosly"),
                &["mysteriously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrat"),
                &["demonstrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammends"), &["amends"]),
            (dictgen::InsensitiveStr::Ascii("alcohal"), &["alcohol"]),
            (
                dictgen::InsensitiveStr::Ascii("speciifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("shll"), &["shell"]),
            (dictgen::InsensitiveStr::Ascii("othersie"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("counpounds"), &["compounds"]),
            (dictgen::InsensitiveStr::Ascii("everaged"), &["averaged"]),
            (
                dictgen::InsensitiveStr::Ascii("coalascece"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conventionss"),
                &["conventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("restrores"), &["restores"]),
            (dictgen::InsensitiveStr::Ascii("regurlars"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("devfines"), &["defines"]),
            (
                dictgen::InsensitiveStr::Ascii("pricinpals"),
                &["principals"],
            ),
            (dictgen::InsensitiveStr::Ascii("similiary"), &["similarity"]),
            (dictgen::InsensitiveStr::Ascii("modulair"), &["modular"]),
            (dictgen::InsensitiveStr::Ascii("zobriust"), &["zobrist"]),
            (
                dictgen::InsensitiveStr::Ascii("qualificaton"),
                &["qualification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveniantly"),
                &["conveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("extehd"), &["extend"]),
            (
                dictgen::InsensitiveStr::Ascii("anihilates"),
                &["annihilates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingished"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("appar"), &["appear"]),
            (
                dictgen::InsensitiveStr::Ascii("continant"),
                &["continental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exect"),
                &["exact", "expect", "eject", "exert"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("conerting"), &["converting"]),
            (
                dictgen::InsensitiveStr::Ascii("ckeckerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missin"),
                &["mission", "missing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pennsyvlania"),
                &["pennsylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabeles"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("logictical"),
                &["logistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramaters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("metatdata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("verticall"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("distinghish"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnofficially"),
                &["unofficially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialsied"),
                &["deserialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underesitmate"),
                &["underestimated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segementing"),
                &["segmenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passhtrough"),
                &["passthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("banannas"), &["bananas"]),
            (
                dictgen::InsensitiveStr::Ascii("permssions"),
                &["permissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloowing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("disadvantged"),
                &["disadvantaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("quiery"), &["query"]),
            (
                dictgen::InsensitiveStr::Ascii("particaulr"),
                &["particular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatios"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alturistic"),
                &["altruistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("statless"), &["stateless"]),
            (
                dictgen::InsensitiveStr::Ascii("lighthearded"),
                &["lighthearted"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicly"), &["explicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("corresopnding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ossociations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creddential"),
                &["credential"],
            ),
            (dictgen::InsensitiveStr::Ascii("deature"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("availaibility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disssociate"),
                &["dissociate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electroncs"),
                &["electrons", "electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reveiwes"),
                &["reviews", "reviewers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemetnation"),
                &["implementations", "implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rewatchibg"),
                &["rewatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bastracting"),
                &["abstracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("pxoxies"), &["proxies"]),
            (dictgen::InsensitiveStr::Ascii("execure"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("servies"), &["services"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferncing"),
                &["dereferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrevates"),
                &["abbreviates"],
            ),
            (dictgen::InsensitiveStr::Ascii("skeletones"), &["skeletons"]),
            (
                dictgen::InsensitiveStr::Ascii("rittled"),
                &["rattled", "riddled"],
            ),
            (dictgen::InsensitiveStr::Ascii("docuented"), &["documented"]),
            (dictgen::InsensitiveStr::Ascii("annoucner"), &["announcer"]),
            (
                dictgen::InsensitiveStr::Ascii("socreboard"),
                &["scoreboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("incedent"), &["incident"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeorhods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("goig"), &["going"]),
            (
                dictgen::InsensitiveStr::Ascii("metroplois"),
                &["metropolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dermatolagist"),
                &["dermatologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("frameowrk"), &["framework"]),
            (dictgen::InsensitiveStr::Ascii("dpendent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("encompused"),
                &["encompassed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializaing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirtualizaiton"),
                &["devirtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sawtees"), &["sautés"]),
            (
                dictgen::InsensitiveStr::Ascii("intereacts"),
                &["interfaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjunciton"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arameterizing"),
                &["parameterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("struturing"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("billioniare"),
                &["billionaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavialble"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sematnically"),
                &["semantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechansims"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("builers"),
                &["builders", "boilers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bombarment"),
                &["bombardment"],
            ),
            (dictgen::InsensitiveStr::Ascii("psuh"), &["push"]),
            (
                dictgen::InsensitiveStr::Ascii("progammatic"),
                &["programmatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("eqautions"), &["equations"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunantly"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("wbsite"), &["website"]),
            (dictgen::InsensitiveStr::Ascii("adequit"), &["adequate"]),
            (
                dictgen::InsensitiveStr::Ascii("predicessor"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("octohedra"), &["octahedra"]),
            (
                dictgen::InsensitiveStr::Ascii("organizors"),
                &["organizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepate"), &["prepare"]),
            (dictgen::InsensitiveStr::Ascii("puplishes"), &["publishes"]),
            (
                dictgen::InsensitiveStr::Ascii("silkcreened"),
                &["silkscreened"],
            ),
            (dictgen::InsensitiveStr::Ascii("electical"), &["electrical"]),
            (
                dictgen::InsensitiveStr::Ascii("comander"),
                &["commander", "commandeer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adressable"),
                &["addressable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precidents"),
                &["precedents", "presidents"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoved"), &["resolved"]),
            (dictgen::InsensitiveStr::Ascii("garphical"), &["graphical"]),
            (dictgen::InsensitiveStr::Ascii("covarage"), &["coverage"]),
            (
                dictgen::InsensitiveStr::Ascii("mircroprocessor"),
                &["microprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffrerently"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formtatters"),
                &["formatters"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprear"), &["appear"]),
            (
                dictgen::InsensitiveStr::Ascii("spefifiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swtichover"),
                &["switchover"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beatifully"),
                &["beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representatitive"),
                &["representative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deseralizing"),
                &["deserializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fignerprint"),
                &["fingerprint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superfluious"),
                &["superfluous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isntallers"),
                &["installers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excptionally"),
                &["exceptionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigaters"),
                &["investigators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precausions"),
                &["precautions"],
            ),
            (dictgen::InsensitiveStr::Ascii("rulle"), &["rule"]),
            (
                dictgen::InsensitiveStr::Ascii("vulerability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("cachable"), &["cacheable"]),
            (
                dictgen::InsensitiveStr::Ascii("amealiarated"),
                &["ameliorated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuiting"),
                &["distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conecpts"), &["concepts"]),
            (dictgen::InsensitiveStr::Ascii("protocool"), &["protocol"]),
            (
                dictgen::InsensitiveStr::Ascii("sufficiancy"),
                &["sufficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invisivility"),
                &["invisibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("reseves"), &["reserves"]),
            (dictgen::InsensitiveStr::Ascii("malining"), &["maligning"]),
            (dictgen::InsensitiveStr::Ascii("sacrifies"), &["sacrifices"]),
            (dictgen::InsensitiveStr::Ascii("ulitmatum"), &["ultimatum"]),
            (dictgen::InsensitiveStr::Ascii("mirrorr"), &["mirror"]),
            (dictgen::InsensitiveStr::Ascii("workspsace"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("permisssion"),
                &["permissions", "permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghths"), &["lengths"]),
            (dictgen::InsensitiveStr::Ascii("flpppy"), &["floppy"]),
            (
                dictgen::InsensitiveStr::Ascii("simulatons"),
                &["simulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("filesysem"), &["filesystem"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailiable"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("rekursed"), &["recursed"]),
            (
                dictgen::InsensitiveStr::Ascii("iresistibly"),
                &["irresistibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destinaton"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("gemerate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("theologial"),
                &["theological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("samwiching"),
                &["sandwiching"],
            ),
            (dictgen::InsensitiveStr::Ascii("detctives"), &["detectives"]),
            (
                dictgen::InsensitiveStr::Ascii("serialuzing"),
                &["serializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("browers"), &["browsers"]),
            (dictgen::InsensitiveStr::Ascii("probobly"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("multples"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("mulithreading"),
                &["multithreading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowedged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stacionary"),
                &["stationary"],
            ),
            (dictgen::InsensitiveStr::Ascii("sneding"), &["sending"]),
            (
                dictgen::InsensitiveStr::Ascii("rienforced"),
                &["reenforced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograhs"),
                &["photographs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilvalent"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacuses"), &["caucuses"]),
            (
                dictgen::InsensitiveStr::Ascii("rudimentry"),
                &["rudimentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("absrob"), &["absorb"]),
            (
                dictgen::InsensitiveStr::Ascii("parrallelised"),
                &["parallelised"],
            ),
            (dictgen::InsensitiveStr::Ascii("applaying"), &["applying"]),
            (dictgen::InsensitiveStr::Ascii("impleted"), &["implemented"]),
            (
                dictgen::InsensitiveStr::Ascii("perliferating"),
                &["proliferating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplishent"),
                &["accomplishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("patriarh"), &["patriarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("repsented"),
                &["repented", "represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wighting"),
                &["weighting", "weighing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepndent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precondtionners"),
                &["preconditioners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highligting"),
                &["highlighting"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulsims"), &["muslims"]),
            (dictgen::InsensitiveStr::Ascii("arrnaged"), &["arranged"]),
            (
                dictgen::InsensitiveStr::Ascii("guaruanteing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("thresolds"), &["thresholds"]),
            (dictgen::InsensitiveStr::Ascii("strinq"), &["string"]),
            (
                dictgen::InsensitiveStr::Ascii("administraron"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("daed"), &["dead"]),
            (
                dictgen::InsensitiveStr::Ascii("explanition"),
                &["explanation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiailisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoch"), &["epoch"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperatelly"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("arameter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("effeckt"), &["effect"]),
            (dictgen::InsensitiveStr::Ascii("alpabet"), &["alphabet"]),
            (dictgen::InsensitiveStr::Ascii("dependig"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("guarenteedeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaitonship"),
                &["relationships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphbetical"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appricating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("naturually"), &["naturally"]),
            (
                dictgen::InsensitiveStr::Ascii("apostraphe"),
                &["apostrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("frogot"), &["forgot"]),
            (
                dictgen::InsensitiveStr::Ascii("xenophonic"),
                &["xenophobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constency"),
                &["consistency", "constancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("genrals"), &["generals"]),
            (
                dictgen::InsensitiveStr::Ascii("substitutents"),
                &["substituents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backaloriettes"),
                &["baccalaureates"],
            ),
            (dictgen::InsensitiveStr::Ascii("keword"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("siblind"), &["sibling"]),
            (
                dictgen::InsensitiveStr::Ascii("characteristsics"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("compleeted"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("carbohidrates"),
                &["carbohydrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subfuntion"),
                &["subfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemens"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("decose"), &["decode"]),
            (
                dictgen::InsensitiveStr::Ascii("depdenently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("rder"), &["order"]),
            (dictgen::InsensitiveStr::Ascii("playe"), &["played", "play"]),
            (
                dictgen::InsensitiveStr::Ascii("potentiomenters"),
                &["potentiometers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profesional"),
                &["professional"],
            ),
            (dictgen::InsensitiveStr::Ascii("forgoten"), &["forgotten"]),
            (dictgen::InsensitiveStr::Ascii("spefixifer"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("dubsetp"), &["dubstep"]),
            (dictgen::InsensitiveStr::Ascii("queueu"), &["queue"]),
            (dictgen::InsensitiveStr::Ascii("climing"), &["climbing"]),
            (dictgen::InsensitiveStr::Ascii("overlflow"), &["overflow"]),
            (
                dictgen::InsensitiveStr::Ascii("acrtivations"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcoked"), &["docked"]),
            (
                dictgen::InsensitiveStr::Ascii("hundret"),
                &["hundred", "hundreds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assihnments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("prgresses"), &["progresses"]),
            (
                dictgen::InsensitiveStr::Ascii("assiociative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provocotive"),
                &["provocative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepedence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportated"),
                &["incorporated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("housingkeeping"),
                &["housekeeping"],
            ),
            (dictgen::InsensitiveStr::Ascii("facourite"), &["favourite"]),
            (dictgen::InsensitiveStr::Ascii("oprions"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("fermenterad"),
                &["fermented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fallicious"),
                &["fallacious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infromatoin"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphiral"),
                &["peripheral"],
            ),
            (dictgen::InsensitiveStr::Ascii("whitin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("digset"), &["digest"]),
            (dictgen::InsensitiveStr::Ascii("byets"), &["bytes"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptoinally"),
                &["exceptionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independendly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("favourtie"), &["favourite"]),
            (dictgen::InsensitiveStr::Ascii("expressley"), &["expressly"]),
            (
                dictgen::InsensitiveStr::Ascii("interrutpion"),
                &["interruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriptions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbrase"), &["embrace"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptino"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("posssibly"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("retranser"), &["retransfer"]),
            (
                dictgen::InsensitiveStr::Ascii("unrandimized"),
                &["unrandomized"],
            ),
            (dictgen::InsensitiveStr::Ascii("urrlib"), &["urllib"]),
            (
                dictgen::InsensitiveStr::Ascii("comparispons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("appoligize"), &["apologize"]),
            (dictgen::InsensitiveStr::Ascii("uffering"), &["buffering"]),
            (dictgen::InsensitiveStr::Ascii("docuemtns"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("ingrediets"),
                &["ingredient"],
            ),
            (dictgen::InsensitiveStr::Ascii("narwharl"), &["narwhal"]),
            (dictgen::InsensitiveStr::Ascii("toursit"), &["tourist"]),
            (dictgen::InsensitiveStr::Ascii("temproal"), &["temporal"]),
            (dictgen::InsensitiveStr::Ascii("regulax"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("encprytion"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("epties"), &["empties"]),
            (
                dictgen::InsensitiveStr::Ascii("catastraphe"),
                &["catastrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagnotic"), &["diagnostic"]),
            (dictgen::InsensitiveStr::Ascii("escused"), &["excused"]),
            (dictgen::InsensitiveStr::Ascii("mirrorred"), &["mirrored"]),
            (
                dictgen::InsensitiveStr::Ascii("administres"),
                &["administer", "administers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understannd"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("excectes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("infinie"), &["infinite"]),
            (dictgen::InsensitiveStr::Ascii("breks"), &["breaks"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnplant"),
                &["transplant"],
            ),
            (dictgen::InsensitiveStr::Ascii("gratful"), &["grateful"]),
            (
                dictgen::InsensitiveStr::Ascii("gynmastics"),
                &["gymnastics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controvery"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("confiugs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("discription"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommpression"),
                &["uncompression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithemic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("invokving"), &["invoking"]),
            (
                dictgen::InsensitiveStr::Ascii("reaponsibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("pigun"), &["pigeon"]),
            (
                dictgen::InsensitiveStr::Ascii("desitnations"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propectively"),
                &["projectively", "prospectively", "protectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcocdings"),
                &["transcodings"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimyze"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("retrict"),
                &["restrict", "retract"],
            ),
            (dictgen::InsensitiveStr::Ascii("necesasry"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("recontructs"),
                &["reconstructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identtically"),
                &["identically"],
            ),
            (dictgen::InsensitiveStr::Ascii("backrace"), &["backtrace"]),
            (
                dictgen::InsensitiveStr::Ascii("brancing"),
                &["branching", "bracing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificatinos"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("kibutz"), &["kibbutz"]),
            (
                dictgen::InsensitiveStr::Ascii("appened"),
                &["append", "appended", "happened"],
            ),
            (dictgen::InsensitiveStr::Ascii("alreadly"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("declrator"), &["declarator"]),
            (dictgen::InsensitiveStr::Ascii("synonymn"), &["synonym"]),
            (dictgen::InsensitiveStr::Ascii("modesl"), &["models"]),
            (dictgen::InsensitiveStr::Ascii("messgaed"), &["messaged"]),
            (dictgen::InsensitiveStr::Ascii("heigest"), &["highest"]),
            (
                dictgen::InsensitiveStr::Ascii("disposeables"),
                &["disposables"],
            ),
            (dictgen::InsensitiveStr::Ascii("facirity"), &["facility"]),
            (
                dictgen::InsensitiveStr::Ascii("convertibile"),
                &["convertible"],
            ),
            (dictgen::InsensitiveStr::Ascii("workign"), &["working"]),
            (
                dictgen::InsensitiveStr::Ascii("rige"),
                &["rice", "ride", "ridge", "rigs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monstrocity"),
                &["monstrosity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ininteresting"),
                &["uninteresting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpant"),
                &["transplant", "transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("anceint"), &["ancient"]),
            (dictgen::InsensitiveStr::Ascii("ctuhlhu"), &["cthulhu"]),
            (
                dictgen::InsensitiveStr::Ascii("materilism"),
                &["materialism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worstations"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("revison"), &["revision"]),
            (
                dictgen::InsensitiveStr::Ascii("subsribers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("fractalis"), &["fractals"]),
            (dictgen::InsensitiveStr::Ascii("resepect"), &["respect"]),
            (dictgen::InsensitiveStr::Ascii("dettaches"), &["detaches"]),
            (
                dictgen::InsensitiveStr::Ascii("microseond"),
                &["microsecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfied"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasmission"),
                &["transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("loyality"), &["loyalty"]),
            (
                dictgen::InsensitiveStr::Ascii("innaccuracies"),
                &["inaccuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("obeserve"), &["observe"]),
            (
                dictgen::InsensitiveStr::Ascii("extremistas"),
                &["extremists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsiquently"),
                &["subsequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scorebaord"),
                &["scoreboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("whheel"), &["wheel"]),
            (dictgen::InsensitiveStr::Ascii("critera"), &["criteria"]),
            (
                dictgen::InsensitiveStr::Ascii("validataions"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenaniganns"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrvieve"), &["retrieve"]),
            (
                dictgen::InsensitiveStr::Ascii("firend"),
                &["friend", "fired", "fiend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negativeity"),
                &["negativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("presnted"), &["presented"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellization"),
                &["parallelization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("councidental"),
                &["coincidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprocucing"),
                &["reproducing", "reprocuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excisted"),
                &["existed", "excised"],
            ),
            (dictgen::InsensitiveStr::Ascii("attachged"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("apapts"), &["adapts"]),
            (dictgen::InsensitiveStr::Ascii("pakistian"), &["pakistani"]),
            (
                dictgen::InsensitiveStr::Ascii("algorrithms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("phsical"), &["physical"]),
            (dictgen::InsensitiveStr::Ascii("assembeler"), &["assembler"]),
            (dictgen::InsensitiveStr::Ascii("assmebly"), &["assembly"]),
            (dictgen::InsensitiveStr::Ascii("unexpectd"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("custoizing"),
                &["customizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("porfolio"), &["portfolio"]),
            (
                dictgen::InsensitiveStr::Ascii("concantenations"),
                &["concatenations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continentul"),
                &["continental"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeutions"), &["executions"]),
            (
                dictgen::InsensitiveStr::Ascii("triagulator"),
                &["triangulator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expepectedly"),
                &["expectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("spannig"), &["spanning"]),
            (dictgen::InsensitiveStr::Ascii("distacnes"), &["distances"]),
            (dictgen::InsensitiveStr::Ascii("embaras"), &["embarrass"]),
            (dictgen::InsensitiveStr::Ascii("sucesion"), &["succession"]),
            (dictgen::InsensitiveStr::Ascii("commontly"), &["commonly"]),
            (
                dictgen::InsensitiveStr::Ascii("compession"),
                &["compassion", "compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("desgining"), &["designing"]),
            (
                dictgen::InsensitiveStr::Ascii("checkkboxes"),
                &["checkboxes"],
            ),
            (dictgen::InsensitiveStr::Ascii("initilal"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("obervs"), &["observes"]),
            (
                dictgen::InsensitiveStr::Ascii("destryiong"),
                &["destroying"],
            ),
            (dictgen::InsensitiveStr::Ascii("acroos"), &["across"]),
            (
                dictgen::InsensitiveStr::Ascii("contradictons"),
                &["contradicts"],
            ),
            (dictgen::InsensitiveStr::Ascii("necesseary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misgoynist"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("chahge"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("preformer"), &["performer"]),
            (dictgen::InsensitiveStr::Ascii("itenerary"), &["itinerary"]),
            (dictgen::InsensitiveStr::Unicode("сontain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("incremetal"),
                &["incremental"],
            ),
            (dictgen::InsensitiveStr::Ascii("normany"), &["normandy"]),
            (dictgen::InsensitiveStr::Ascii("configuer"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("sction"),
                &["section", "suction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regiserted"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactureras"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("etcp"), &["etc"]),
            (
                dictgen::InsensitiveStr::Ascii("valentinos"),
                &["valentines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polynominal"),
                &["polynomial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rturns"),
                &["returns", "turns"],
            ),
            (dictgen::InsensitiveStr::Ascii("heroicos"), &["heroics"]),
            (
                dictgen::InsensitiveStr::Ascii("interuppted"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("penwar"), &["peignoir"]),
            (
                dictgen::InsensitiveStr::Ascii("australina"),
                &["australians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trajactory"),
                &["trajectory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenently"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privetized"),
                &["privatized"],
            ),
            (dictgen::InsensitiveStr::Ascii("robustnes"), &["robustness"]),
            (dictgen::InsensitiveStr::Ascii("appraent"), &["apparent"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhague"),
                &["copenhagen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotionable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desriptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("enameld"), &["enameled"]),
            (dictgen::InsensitiveStr::Ascii("generaed"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("autentificated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("attmept"), &["attempt"]),
            (
                dictgen::InsensitiveStr::Ascii("accension"),
                &["accession", "ascension"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefified"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proporting"),
                &["proportion"],
            ),
            (dictgen::InsensitiveStr::Ascii("characer"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("methodds"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("camoflaged"),
                &["camouflaged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termiation"),
                &["termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("abiut"), &["about"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotors"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoucnes"), &["announces"]),
            (dictgen::InsensitiveStr::Ascii("synanym"), &["synonym"]),
            (dictgen::InsensitiveStr::Ascii("theiv"), &["thief"]),
            (dictgen::InsensitiveStr::Ascii("godounov"), &["godunov"]),
            (
                dictgen::InsensitiveStr::Ascii("sesson"),
                &["session", "season", "lesson"],
            ),
            (dictgen::InsensitiveStr::Ascii("animatte"), &["animate"]),
            (dictgen::InsensitiveStr::Ascii("kepps"), &["keeps"]),
            (dictgen::InsensitiveStr::Ascii("acceepters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("emphaize"), &["emphasize"]),
            (dictgen::InsensitiveStr::Ascii("endcodes"), &["encodes"]),
            (
                dictgen::InsensitiveStr::Ascii("operatinal"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chalenging"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particalur"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("observarse"), &["observers"]),
            (dictgen::InsensitiveStr::Ascii("cerbreus"), &["cerberus"]),
            (
                dictgen::InsensitiveStr::Ascii("predeceding"),
                &["preceding"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrachs"), &["scratches"]),
            (
                dictgen::InsensitiveStr::Ascii("paticipated"),
                &["participated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overengeneering"),
                &["overengineering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antaganist"),
                &["antagonist"],
            ),
            (dictgen::InsensitiveStr::Ascii("gloiath"), &["goliath"]),
            (dictgen::InsensitiveStr::Ascii("aspected"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("dacquiri"), &["daiquiri"]),
            (
                dictgen::InsensitiveStr::Ascii("guadulupe"),
                &["guadalupe", "guadeloupe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extacted"),
                &["extracted", "exacted"],
            ),
            (dictgen::InsensitiveStr::Ascii("edpoint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("reacked"), &["reached"]),
            (
                dictgen::InsensitiveStr::Ascii("normarlized"),
                &["normalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasnform"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("guarnateee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("journalizm"),
                &["journalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("befre"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("instad"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("activitiy"), &["activity"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualter"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irresepective"),
                &["irrespective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touchscreeen"),
                &["touchscreen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremnetal"),
                &["incremental"],
            ),
            (dictgen::InsensitiveStr::Ascii("oracels"), &["oracles"]),
            (dictgen::InsensitiveStr::Ascii("reqests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("updted"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("hilights"), &["highlights"]),
            (dictgen::InsensitiveStr::Ascii("colsing"), &["closing"]),
            (dictgen::InsensitiveStr::Ascii("docuement"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("ligthning"), &["lightning"]),
            (
                dictgen::InsensitiveStr::Ascii("tranverse"),
                &["transverse", "traverse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momentarly"),
                &["momentarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoindingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philippinas"),
                &["philippines"],
            ),
            (dictgen::InsensitiveStr::Ascii("negtative"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("prvilege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("indivdual"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("detcted"), &["detected"]),
            (dictgen::InsensitiveStr::Ascii("suplement"), &["supplement"]),
            (dictgen::InsensitiveStr::Ascii("inherints"), &["inherits"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuppoted"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicial"),
                &["initial", "indicial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independt"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("assosciate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("intermedie"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmetns"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("abstacter"), &["abstracter"]),
            (dictgen::InsensitiveStr::Ascii("warnign"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("absolaute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("pachage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("aleged"), &["alleged"]),
            (dictgen::InsensitiveStr::Ascii("navgiator"), &["navigator"]),
            (
                dictgen::InsensitiveStr::Ascii("cconsidered"),
                &["considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sinagogs"),
                &["synagogs", "synagogues"],
            ),
            (dictgen::InsensitiveStr::Ascii("wthin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("vasalls"), &["vassals"]),
            (dictgen::InsensitiveStr::Ascii("flavord"), &["flavored"]),
            (dictgen::InsensitiveStr::Ascii("technics"), &["techniques"]),
            (
                dictgen::InsensitiveStr::Ascii("prescribtion"),
                &["prescription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creaing"),
                &["creating", "creasing", "creaking", "creaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqenses"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("pallete"), &["palette"]),
            (
                dictgen::InsensitiveStr::Ascii("completetion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represendation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("strengh"), &["strength"]),
            (
                dictgen::InsensitiveStr::Ascii("oridinal"),
                &["ordinal", "original"],
            ),
            (dictgen::InsensitiveStr::Ascii("timemouts"), &["timeouts"]),
            (
                dictgen::InsensitiveStr::Ascii("authentcated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("seralized"), &["serialized"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubtor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expandibility"),
                &["expandability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intentines"),
                &["intestines"],
            ),
            (dictgen::InsensitiveStr::Ascii("macthers"), &["matchers"]),
            (dictgen::InsensitiveStr::Ascii("wekk"), &["week"]),
            (dictgen::InsensitiveStr::Ascii("suppoorter"), &["supporter"]),
            (dictgen::InsensitiveStr::Ascii("illiegal"), &["illegal"]),
            (dictgen::InsensitiveStr::Ascii("elsewere"), &["elsewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("initiatve"),
                &["initiate", "initiative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantaneus"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscried"),
                &["unsubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ointing"),
                &["pointing", "jointing", "linting"],
            ),
            (dictgen::InsensitiveStr::Ascii("palster"), &["plaster"]),
            (dictgen::InsensitiveStr::Ascii("assertino"), &["assertion"]),
            (dictgen::InsensitiveStr::Ascii("preventes"), &["prevents"]),
            (dictgen::InsensitiveStr::Ascii("complexy"), &["complexity"]),
            (dictgen::InsensitiveStr::Ascii("ecidious"), &["deciduous"]),
            (dictgen::InsensitiveStr::Ascii("destruyed"), &["destroyed"]),
            (dictgen::InsensitiveStr::Ascii("jorunaled"), &["journaled"]),
            (dictgen::InsensitiveStr::Ascii("whne"), &["when"]),
            (
                dictgen::InsensitiveStr::Ascii("questiong"),
                &["questioning", "question"],
            ),
            (dictgen::InsensitiveStr::Ascii("expceting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("overwitten"),
                &["overwritten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desinations"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanitions"),
                &["explanations"],
            ),
            (dictgen::InsensitiveStr::Ascii("impiled"), &["implied"]),
            (dictgen::InsensitiveStr::Ascii("siplifier"), &["simplifier"]),
            (dictgen::InsensitiveStr::Ascii("boffer"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("ususal"), &["usual"]),
            (dictgen::InsensitiveStr::Ascii("cusomary"), &["customary"]),
            (dictgen::InsensitiveStr::Ascii("buhddists"), &["buddhists"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibilitys"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitialized"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("wnat"), &["want", "what"]),
            (
                dictgen::InsensitiveStr::Ascii("currects"),
                &["corrects", "currents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreasonalby"),
                &["unreasonably"],
            ),
            (dictgen::InsensitiveStr::Ascii("coppyright"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("imbrase"), &["embrace"]),
            (
                dictgen::InsensitiveStr::Ascii("resuorcing"),
                &["resourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effiently"),
                &["efficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiugred"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("referance"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("aesthatics"),
                &["aesthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("diappear"), &["disappear"]),
            (
                dictgen::InsensitiveStr::Ascii("mediterrannean"),
                &["mediterranean"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nueroinformatics"),
                &["neuroinformatics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organisere"),
                &["organisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("evyrthing"), &["everything"]),
            (
                dictgen::InsensitiveStr::Ascii("sturctures"),
                &["structures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structruing"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathmatics"),
                &["mathematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("trpoical"), &["tropical"]),
            (
                dictgen::InsensitiveStr::Ascii("aesthitically"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absoletely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("sykwalker"), &["skywalker"]),
            (
                dictgen::InsensitiveStr::Ascii("deined"),
                &["denied", "defined"],
            ),
            (dictgen::InsensitiveStr::Ascii("teminals"), &["terminals"]),
            (dictgen::InsensitiveStr::Ascii("canclling"), &["cancelling"]),
            (dictgen::InsensitiveStr::Ascii("comitee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("claendar"), &["calendar"]),
            (
                dictgen::InsensitiveStr::Ascii("acomplishes"),
                &["accomplishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possessin"),
                &["possessions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimmensional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cristallisation"),
                &["crystallisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crationists"),
                &["creationists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonization"),
                &["synchronization"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeases"), &["releases"]),
            (
                dictgen::InsensitiveStr::Ascii("confimations"),
                &["confirmations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configiurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("beyoncye"), &["beyonce"]),
            (
                dictgen::InsensitiveStr::Ascii("philosohpically"),
                &["philosophically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggragators"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirumvents"),
                &["circumvents"],
            ),
            (dictgen::InsensitiveStr::Ascii("styhe"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("compatablie"),
                &["compatible", "compatibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("necessaru"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("symptumaticly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appearently"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("listensing"),
                &["listening", "licensing"],
            ),
            (dictgen::InsensitiveStr::Ascii("otehrwize"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("writabe"), &["writable"]),
            (dictgen::InsensitiveStr::Ascii("gottleib"), &["gottlieb"]),
            (
                dictgen::InsensitiveStr::Ascii("resumbmitting"),
                &["resubmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronounceing"),
                &["pronouncing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagarisim"),
                &["plagiarism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neuroligical"),
                &["neurological"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeares"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("recuriter"), &["recruiter"]),
            (
                dictgen::InsensitiveStr::Ascii("milktoasts"),
                &["milquetoasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proplusion"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("whislting"), &["whistling"]),
            (dictgen::InsensitiveStr::Ascii("symhpony"), &["symphony"]),
            (dictgen::InsensitiveStr::Ascii("apporval"), &["approval"]),
            (dictgen::InsensitiveStr::Ascii("oppsite"), &["opposite"]),
            (
                dictgen::InsensitiveStr::Ascii("knowledgebale"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("strugling"), &["struggling"]),
            (dictgen::InsensitiveStr::Ascii("asutria"), &["austria"]),
            (
                dictgen::InsensitiveStr::Ascii("backrounds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("concedendo"), &["conceded"]),
            (
                dictgen::InsensitiveStr::Ascii("sived"),
                &["saved", "sieved"],
            ),
            (dictgen::InsensitiveStr::Ascii("takoeff"), &["takeoff"]),
            (
                dictgen::InsensitiveStr::Ascii("immatable"),
                &["immutable", "imitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("vlones"), &["clones"]),
            (dictgen::InsensitiveStr::Ascii("reminsce"), &["reminisce"]),
            (dictgen::InsensitiveStr::Ascii("exlicit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("exceopting"),
                &["exempting", "excepting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralelly"),
                &["parallelly"],
            ),
            (dictgen::InsensitiveStr::Ascii("genuinly"), &["genuinely"]),
            (
                dictgen::InsensitiveStr::Ascii("juristiction"),
                &["jurisdiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facisnation"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("termintor"), &["terminator"]),
            (
                dictgen::InsensitiveStr::Ascii("comtaining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inforamtional"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("systens"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("challeger"), &["challenger"]),
            (dictgen::InsensitiveStr::Ascii("placeemnt"), &["placement"]),
            (dictgen::InsensitiveStr::Ascii("dervived"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("ukrainie"), &["ukraine"]),
            (dictgen::InsensitiveStr::Ascii("othere"), &["other"]),
            (
                dictgen::InsensitiveStr::Ascii("appreacites"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrutor"), &["destructor"]),
            (
                dictgen::InsensitiveStr::Ascii("specificated"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("unter"), &["under"]),
            (dictgen::InsensitiveStr::Ascii("witrhout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("exsistence"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("genererating"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("negotioate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("straigthforward"),
                &["straightforward"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphetamines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("changlog"), &["changelog"]),
            (dictgen::InsensitiveStr::Ascii("mdoelling"), &["modelling"]),
            (dictgen::InsensitiveStr::Ascii("performaed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("canides"), &["candies"]),
            (
                dictgen::InsensitiveStr::Ascii("surplanted"),
                &["supplanted"],
            ),
            (dictgen::InsensitiveStr::Ascii("motorolla"), &["motorola"]),
            (
                dictgen::InsensitiveStr::Ascii("asymptoptical"),
                &["asymptotical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumires"),
                &["consumerism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assersively"),
                &["assertively"],
            ),
            (dictgen::InsensitiveStr::Ascii("upadater"), &["updater"]),
            (dictgen::InsensitiveStr::Ascii("orginised"), &["organised"]),
            (
                dictgen::InsensitiveStr::Ascii("commmunicating"),
                &["communicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("relcaim"), &["reclaim"]),
            (dictgen::InsensitiveStr::Ascii("vieport"), &["viewport"]),
            (dictgen::InsensitiveStr::Ascii("illegallly"), &["illegally"]),
            (
                dictgen::InsensitiveStr::Ascii("disassebled"),
                &["disassembled"],
            ),
            (dictgen::InsensitiveStr::Ascii("scannin"), &["scanning"]),
            (dictgen::InsensitiveStr::Ascii("stll"), &["still"]),
            (
                dictgen::InsensitiveStr::Ascii("disertation"),
                &["dissertation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hashnaps"), &["hashmaps"]),
            (
                dictgen::InsensitiveStr::Ascii("motivatinal"),
                &["motivational"],
            ),
            (dictgen::InsensitiveStr::Ascii("excesss"), &["excess"]),
            (dictgen::InsensitiveStr::Ascii("trasport"), &["transport"]),
            (
                dictgen::InsensitiveStr::Ascii("dimentionnalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registing"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organistaions"),
                &["organisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precenence"),
                &["preference", "precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapat"), &["dissipate"]),
            (dictgen::InsensitiveStr::Ascii("cirillic"), &["cyrillic"]),
            (dictgen::InsensitiveStr::Ascii("uegister"), &["unregister"]),
            (dictgen::InsensitiveStr::Ascii("relapes"), &["relapse"]),
            (
                dictgen::InsensitiveStr::Ascii("appllications"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("finanize"), &["finalize"]),
            (dictgen::InsensitiveStr::Ascii("embarras"), &["embarrass"]),
            (
                dictgen::InsensitiveStr::Ascii("correnspondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encouranged"),
                &["encouraged"],
            ),
            (dictgen::InsensitiveStr::Ascii("ojbects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("incredably"),
                &["incredibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundations"),
                &["foundations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reincarantion"),
                &["reincarnation"],
            ),
            (dictgen::InsensitiveStr::Ascii("oiginally"), &["originally"]),
            (dictgen::InsensitiveStr::Ascii("justifed"), &["justified"]),
            (
                dictgen::InsensitiveStr::Ascii("linkes"),
                &["lines", "links", "linked", "likes"],
            ),
            (dictgen::InsensitiveStr::Ascii("examind"), &["examined"]),
            (
                dictgen::InsensitiveStr::Ascii("modificaiton"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("escaltes"), &["escalates"]),
            (
                dictgen::InsensitiveStr::Ascii("searche"),
                &["search", "searched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propagatex"),
                &["propagates", "propagated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("matterns"),
                &["matters", "patterns"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloow"), &["follow"]),
            (
                dictgen::InsensitiveStr::Ascii("covergence"),
                &["convergence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcpondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiemstep"), &["timestep"]),
            (dictgen::InsensitiveStr::Ascii("integrety"), &["integrity"]),
            (
                dictgen::InsensitiveStr::Ascii("abords"),
                &["aborts", "boards"],
            ),
            (dictgen::InsensitiveStr::Ascii("trnasfers"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("silouete"), &["silhouette"]),
            (
                dictgen::InsensitiveStr::Ascii("depretiated"),
                &["depreciated", "deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hellfie"), &["hellfire"]),
            (dictgen::InsensitiveStr::Ascii("obersvers"), &["observers"]),
            (
                dictgen::InsensitiveStr::Ascii("technniques"),
                &["techniques"],
            ),
            (dictgen::InsensitiveStr::Ascii("handels"), &["handles"]),
            (
                dictgen::InsensitiveStr::Ascii("particaluar"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("vidoes"), &["videos"]),
            (dictgen::InsensitiveStr::Ascii("bacehlor"), &["bachelor"]),
            (
                dictgen::InsensitiveStr::Ascii("intervenin"),
                &["intervening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("torchilght"),
                &["torchlight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlefiend"),
                &["battlefield"],
            ),
            (dictgen::InsensitiveStr::Ascii("algrithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("recreatin"),
                &["recreating", "recreation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distibuted"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("divinitory"), &["divinity"]),
            (
                dictgen::InsensitiveStr::Ascii("increadibly"),
                &["incredibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentions"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosumptions"),
                &["consumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automtically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cehcker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("unstructions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanatge"), &["advantage"]),
            (dictgen::InsensitiveStr::Ascii("brobably"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("clitorius"), &["clitoris"]),
            (
                dictgen::InsensitiveStr::Ascii("moisutrizing"),
                &["moisturizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinfigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthagonalize"),
                &["orthogonalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("excerising"),
                &["exercising"],
            ),
            (dictgen::InsensitiveStr::Ascii("tannheill"), &["tannehill"]),
            (dictgen::InsensitiveStr::Ascii("driftig"), &["drifting"]),
            (dictgen::InsensitiveStr::Ascii("rerference"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("missionera"),
                &["missionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("truthfullly"),
                &["truthfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependnts"), &["dependents"]),
            (
                dictgen::InsensitiveStr::Ascii("decomissioned"),
                &["decommissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bueraucratic"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("equipmnet"), &["equipment"]),
            (
                dictgen::InsensitiveStr::Ascii("inpenetrable"),
                &["impenetrable"],
            ),
            (dictgen::InsensitiveStr::Ascii("treaeted"), &["treated"]),
            (
                dictgen::InsensitiveStr::Ascii("executuble"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoepfully"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("atttacks"), &["attacks"]),
            (dictgen::InsensitiveStr::Ascii("authrored"), &["authored"]),
            (dictgen::InsensitiveStr::Ascii("csaes"), &["cases"]),
            (dictgen::InsensitiveStr::Ascii("captians"), &["captains"]),
            (dictgen::InsensitiveStr::Ascii("cartrdige"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("strategiclly"),
                &["strategically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competetors"),
                &["competitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advantageus"),
                &["advantageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reqeusting"),
                &["requesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("arays"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("kewords"), &["keywords"]),
            (
                dictgen::InsensitiveStr::Ascii("devolopement"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("tere"), &["there", "here"]),
            (
                dictgen::InsensitiveStr::Ascii("broderlands"),
                &["borderlands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrtib"), &["contrib"]),
            (dictgen::InsensitiveStr::Ascii("sponzored"), &["sponsored"]),
            (dictgen::InsensitiveStr::Ascii("actauly"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("avaialbel"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("cootdinating"),
                &["coordinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("irritatie"), &["irritate"]),
            (dictgen::InsensitiveStr::Ascii("othervise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("instruccional"),
                &["instructional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("eache"), &["each", "ache"]),
            (dictgen::InsensitiveStr::Ascii("pleasent"), &["pleasant"]),
            (dictgen::InsensitiveStr::Ascii("capitalze"), &["capitalize"]),
            (dictgen::InsensitiveStr::Ascii("atrifacts"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("classfications"),
                &["classifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antarctida"),
                &["antarctica"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcifies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("transmition"),
                &["transmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("popolating"),
                &["populating"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloccated"), &["allocated"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchrounsly"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obamination"),
                &["abomination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reposoitory"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("cafs"), &["calves"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuited"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("comercial"), &["commercial"]),
            (
                dictgen::InsensitiveStr::Ascii("deterministische"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processd"),
                &["process", "processed"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntacs"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("depenencys"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("odf"), &["of"]),
            (
                dictgen::InsensitiveStr::Ascii("interefered"),
                &["interfered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribbuting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tegister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("experct"),
                &["expect", "excerpt"],
            ),
            (dictgen::InsensitiveStr::Ascii("similuate"), &["simulate"]),
            (
                dictgen::InsensitiveStr::Ascii("envirionmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("guardain"), &["guardian"]),
            (dictgen::InsensitiveStr::Ascii("usibility"), &["usability"]),
            (
                dictgen::InsensitiveStr::Ascii("depandances"),
                &["dependances"],
            ),
            (dictgen::InsensitiveStr::Ascii("equestion"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("copmletes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("compelting"),
                &["completing"],
            ),
            (dictgen::InsensitiveStr::Ascii("asborbes"), &["absorbs"]),
            (dictgen::InsensitiveStr::Ascii("dermines"), &["determines"]),
            (dictgen::InsensitiveStr::Ascii("replubic"), &["republic"]),
            (
                dictgen::InsensitiveStr::Ascii("challeneger"),
                &["challenger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cardiocascular"),
                &["cardiovascular"],
            ),
            (dictgen::InsensitiveStr::Ascii("artemius"), &["artemis"]),
            (
                dictgen::InsensitiveStr::Ascii("illitirate"),
                &["illiterate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mapable"), &["mappable"]),
            (
                dictgen::InsensitiveStr::Ascii("backjacking"),
                &["backpacking"],
            ),
            (dictgen::InsensitiveStr::Ascii("brower"), &["browser"]),
            (dictgen::InsensitiveStr::Ascii("animaton"), &["animation"]),
            (dictgen::InsensitiveStr::Ascii("lokaly"), &["locally"]),
            (
                dictgen::InsensitiveStr::Ascii("unneccesarry"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nore"),
                &["nor", "more", "node", "note"],
            ),
            (dictgen::InsensitiveStr::Ascii("medicad"), &["medicaid"]),
            (
                dictgen::InsensitiveStr::Ascii("origina"),
                &["origin", "original"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignmant"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("taliored"), &["tailored"]),
            (dictgen::InsensitiveStr::Ascii("expepts"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("normalyl"), &["normally"]),
            (
                dictgen::InsensitiveStr::Ascii("combininatorial"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("counciler"), &["councilor"]),
            (dictgen::InsensitiveStr::Ascii("painfuly"), &["painfully"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriatedly"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("gentelmen"), &["gentlemen"]),
            (
                dictgen::InsensitiveStr::Ascii("discriptors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("girlfrients"),
                &["girlfriends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diferrentiation"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pluggin"),
                &["plugin", "plugging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intillectually"),
                &["intellectually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuprizing"),
                &["unsurprising"],
            ),
            (dictgen::InsensitiveStr::Ascii("excata"), &["exacta"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegothiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitaliser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("rertieves"), &["retrieves"]),
            (dictgen::InsensitiveStr::Ascii("spawing"), &["spawning"]),
            (dictgen::InsensitiveStr::Ascii("yesturday"), &["yesterday"]),
            (dictgen::InsensitiveStr::Ascii("coontext"), &["context"]),
            (
                dictgen::InsensitiveStr::Ascii("shaerd"),
                &["shared", "shard"],
            ),
            (dictgen::InsensitiveStr::Ascii("montioned"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("speciifc"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("nimphomania"),
                &["nymphomania"],
            ),
            (dictgen::InsensitiveStr::Ascii("sighrynges"), &["syringes"]),
            (dictgen::InsensitiveStr::Ascii("besically"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("excesv"), &["excessive"]),
            (dictgen::InsensitiveStr::Ascii("strign"), &["string"]),
            (dictgen::InsensitiveStr::Ascii("bayoent"), &["bayonet"]),
            (dictgen::InsensitiveStr::Ascii("ganerated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("prosectuor"),
                &["prosecutor"],
            ),
            (dictgen::InsensitiveStr::Ascii("franches"), &["franchise"]),
            (dictgen::InsensitiveStr::Ascii("islamsits"), &["islamists"]),
            (
                dictgen::InsensitiveStr::Ascii("configuered"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("wedensday"), &["wednesday"]),
            (dictgen::InsensitiveStr::Ascii("widghets"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("sybmols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("refudn"), &["refund"]),
            (
                dictgen::InsensitiveStr::Ascii("cringewrothy"),
                &["cringeworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconect"),
                &["disconnect"],
            ),
            (dictgen::InsensitiveStr::Ascii("coloums"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("diconnects"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambitous"), &["ambitious"]),
            (dictgen::InsensitiveStr::Ascii("dupliate"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporpriations"),
                &["appropriations"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcoesses"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("knockbak"), &["knockback"]),
            (dictgen::InsensitiveStr::Ascii("critisize"), &["criticize"]),
            (dictgen::InsensitiveStr::Ascii("usuability"), &["usability"]),
            (
                dictgen::InsensitiveStr::Ascii("chechkpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overidding"),
                &["overriding"],
            ),
            (dictgen::InsensitiveStr::Ascii("boooleans"), &["booleans"]),
            (dictgen::InsensitiveStr::Ascii("sucessot"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("identifiaction"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("configues"), &["configures"]),
            (dictgen::InsensitiveStr::Ascii("mimaches"), &["mismatches"]),
            (dictgen::InsensitiveStr::Ascii("verticed"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("quesetionnaires"),
                &["questionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("rederer"), &["renderer"]),
            (dictgen::InsensitiveStr::Ascii("examles"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("intilizers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("speparate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("benifited"), &["benefited"]),
            (
                dictgen::InsensitiveStr::Ascii("getttimeofday"),
                &["gettimeofday"],
            ),
            (dictgen::InsensitiveStr::Ascii("multiplaye"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("anywere"), &["anywhere"]),
            (dictgen::InsensitiveStr::Ascii("jist"), &["gist"]),
            (dictgen::InsensitiveStr::Ascii("inseet"), &["insert"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninhabbited"),
                &["uninhabited"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakcwards"), &["backwards"]),
            (
                dictgen::InsensitiveStr::Ascii("empressing"),
                &["impressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrdinates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("mantioned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("awknowledgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuuing"),
                &["continuing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryahrd"), &["tryhard"]),
            (dictgen::InsensitiveStr::Ascii("nver"), &["never"]),
            (dictgen::InsensitiveStr::Ascii("legaxy"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("outragerous"),
                &["outrageous"],
            ),
            (dictgen::InsensitiveStr::Ascii("unirom"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("accomplisments"),
                &["accomplishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creationsm"),
                &["creationism"],
            ),
            (dictgen::InsensitiveStr::Ascii("smilar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("electritian"),
                &["electrician"],
            ),
            (dictgen::InsensitiveStr::Ascii("acctivated"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("caculates"), &["calculates"]),
            (dictgen::InsensitiveStr::Ascii("raotated"), &["rotated"]),
            (dictgen::InsensitiveStr::Ascii("cpatured"), &["captured"]),
            (dictgen::InsensitiveStr::Ascii("allready"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchroniously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("tructures"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("availabilty"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("incgonito"), &["incognito"]),
            (
                dictgen::InsensitiveStr::Ascii("eleiminated"),
                &["eliminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reresh"), &["refresh"]),
            (
                dictgen::InsensitiveStr::Ascii("contrainers"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nightmates"),
                &["nightmares"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualisin"),
                &["visualising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissodance"),
                &["dissonance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimititer"),
                &["delimiter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgrouunds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("classifiy"), &["classify"]),
            (dictgen::InsensitiveStr::Ascii("haynus"), &["heinous"]),
            (dictgen::InsensitiveStr::Ascii("addjust"), &["adjust"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscritpiton"),
                &["unsubscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("devicesr"), &["devices"]),
            (
                dictgen::InsensitiveStr::Ascii("acknownledgment"),
                &["acknowledgment"],
            ),
            (dictgen::InsensitiveStr::Ascii("copletes"), &["completes"]),
            (dictgen::InsensitiveStr::Ascii("retieves"), &["retrieves"]),
            (dictgen::InsensitiveStr::Ascii("camerea"), &["camera"]),
            (
                dictgen::InsensitiveStr::Ascii("fram"),
                &["frame", "gram", "dram", "cram"],
            ),
            (dictgen::InsensitiveStr::Ascii("pharoh"), &["pharaoh"]),
            (dictgen::InsensitiveStr::Ascii("worrry"), &["worry"]),
            (dictgen::InsensitiveStr::Ascii("validd"), &["valid"]),
            (
                dictgen::InsensitiveStr::Ascii("productivos"),
                &["productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("corectors"), &["correctors"]),
            (dictgen::InsensitiveStr::Ascii("traveresed"), &["traversed"]),
            (
                dictgen::InsensitiveStr::Ascii("connetcions"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("encorders"), &["encoders"]),
            (
                dictgen::InsensitiveStr::Ascii("cannel"),
                &["channel", "cancel", "canal"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrokbook"), &["workbook"]),
            (
                dictgen::InsensitiveStr::Ascii("nastertiums"),
                &["nasturtiums"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhodds"),
                &["neighbourhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterises"),
                &["parameterises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmacuetical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hurricain"), &["hurricane"]),
            (
                dictgen::InsensitiveStr::Ascii("restuctures"),
                &["restructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requesites"),
                &["requisites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capicitance"),
                &["capacitance"],
            ),
            (dictgen::InsensitiveStr::Ascii("abdomnial"), &["abdominal"]),
            (
                dictgen::InsensitiveStr::Ascii("missionnary"),
                &["missionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("descktops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("templares"), &["templars"]),
            (
                dictgen::InsensitiveStr::Ascii("computacional"),
                &["computational"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendezous"), &["rendezvous"]),
            (
                dictgen::InsensitiveStr::Ascii("deficately"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceratinty"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("expetimentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("gemetry"), &["geometry"]),
            (
                dictgen::InsensitiveStr::Ascii("correspdoningly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("flakyness"), &["flakiness"]),
            (dictgen::InsensitiveStr::Ascii("dedidates"), &["dedicates"]),
            (dictgen::InsensitiveStr::Ascii("franciso"), &["francisco"]),
            (
                dictgen::InsensitiveStr::Ascii("relativated"),
                &["relative", "relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masogynistic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("typpically"), &["typically"]),
            (dictgen::InsensitiveStr::Ascii("miminally"), &["minimally"]),
            (
                dictgen::InsensitiveStr::Ascii("chaeckpoint"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elimenates"),
                &["eliminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("relativly"), &["relatively"]),
            (dictgen::InsensitiveStr::Ascii("inventaron"), &["inventor"]),
            (dictgen::InsensitiveStr::Ascii("coveres"), &["covers"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitions"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mccarthyst"),
                &["mccarthyist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemements"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unneccessarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersectionals"),
                &["intersections"],
            ),
            (dictgen::InsensitiveStr::Ascii("attand"), &["attend"]),
            (dictgen::InsensitiveStr::Ascii("anticipe"), &["anticipate"]),
            (
                dictgen::InsensitiveStr::Ascii("orthorombic"),
                &["orthorhombic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlytic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendets"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("returm"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("intimitate"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexuels"),
                &["homosexuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyle"), &["cycle"]),
            (
                dictgen::InsensitiveStr::Ascii("directores"),
                &["directors", "directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("spearator"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("dealth"),
                &["dealt", "death"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connektors"),
                &["connectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("refelects"), &["reflects"]),
            (
                dictgen::InsensitiveStr::Ascii("obhectifying"),
                &["objectifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inclinaison"),
                &["inclination"],
            ),
            (dictgen::InsensitiveStr::Ascii("orphanded"), &["orphaned"]),
            (
                dictgen::InsensitiveStr::Ascii("agument"),
                &["argument", "augment"],
            ),
            (dictgen::InsensitiveStr::Ascii("orginate"), &["originate"]),
            (
                dictgen::InsensitiveStr::Ascii("appraoching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatanated"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogirhtmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summurization"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("misquitos"), &["mosquitos"]),
            (dictgen::InsensitiveStr::Ascii("aribitrary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibitting"),
                &["prohibiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interlave"),
                &["interlace", "interleave"],
            ),
            (dictgen::InsensitiveStr::Ascii("extimated"), &["estimated"]),
            (dictgen::InsensitiveStr::Ascii("oen"), &["one"]),
            (
                dictgen::InsensitiveStr::Ascii("psysiological"),
                &["physiological"],
            ),
            (dictgen::InsensitiveStr::Ascii("sructure"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("hardenend"), &["hardened"]),
            (
                dictgen::InsensitiveStr::Ascii("attivation"),
                &["activation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrypiton"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("insigts"), &["insights"]),
            (
                dictgen::InsensitiveStr::Ascii("rebuilts"),
                &["rebuilds", "rebuilt", "rebuild"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahtiest"), &["atheist"]),
            (
                dictgen::InsensitiveStr::Ascii("disciplies"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("postponin"), &["postponing"]),
            (
                dictgen::InsensitiveStr::Ascii("descryption"),
                &["description", "decryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountents"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("carfull"), &["careful"]),
            (dictgen::InsensitiveStr::Ascii("falsk"), &["flask", "false"]),
            (
                dictgen::InsensitiveStr::Ascii("overrridden"),
                &["overridden"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomended"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("charocal"), &["charcoal"]),
            (dictgen::InsensitiveStr::Ascii("perilious"), &["perilous"]),
            (
                dictgen::InsensitiveStr::Ascii("equivlents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("evrything"), &["everything"]),
            (
                dictgen::InsensitiveStr::Ascii("contradtion"),
                &["contraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("rapsadys"), &["rhapsodies"]),
            (
                dictgen::InsensitiveStr::Ascii("seceeded"),
                &["succeeded", "seceded"],
            ),
            (dictgen::InsensitiveStr::Ascii("dropable"), &["droppable"]),
            (
                dictgen::InsensitiveStr::Ascii("comtemplating"),
                &["contemplating"],
            ),
            (dictgen::InsensitiveStr::Ascii("charascter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("seralization"),
                &["serialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("algoriths"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("folfers"), &["folders"]),
            (dictgen::InsensitiveStr::Ascii("editior"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("arros"), &["arrows"]),
            (dictgen::InsensitiveStr::Ascii("langual"), &["lingual"]),
            (
                dictgen::InsensitiveStr::Ascii("copyrughted"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("paticle"), &["particle"]),
            (dictgen::InsensitiveStr::Ascii("atribute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("initaites"), &["initiates"]),
            (dictgen::InsensitiveStr::Ascii("chimeny"), &["chimney"]),
            (
                dictgen::InsensitiveStr::Ascii("intorverted"),
                &["introverted"],
            ),
            (dictgen::InsensitiveStr::Ascii("aftzerward"), &["afterward"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitfication"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discapline"),
                &["discipline"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtring"), &["substring"]),
            (dictgen::InsensitiveStr::Ascii("tabluate"), &["tabulate"]),
            (
                dictgen::InsensitiveStr::Ascii("fallatious"),
                &["fallacious"],
            ),
            (dictgen::InsensitiveStr::Ascii("underwar"), &["underwater"]),
            (
                dictgen::InsensitiveStr::Ascii("certificion"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("ternimal"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("notifing"), &["notifying"]),
            (dictgen::InsensitiveStr::Ascii("deverlop"), &["develop"]),
            (
                dictgen::InsensitiveStr::Ascii("inferfaces"),
                &["interfaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("conet"), &["connect"]),
            (dictgen::InsensitiveStr::Ascii("mulipart"), &["multipart"]),
            (
                dictgen::InsensitiveStr::Ascii("professoinal"),
                &["professionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("adventurs"), &["adventures"]),
            (dictgen::InsensitiveStr::Ascii("executue"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("playersare"),
                &["playerbase"],
            ),
            (dictgen::InsensitiveStr::Ascii("toggeling"), &["toggling"]),
            (dictgen::InsensitiveStr::Ascii("explecit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("plabeswalker"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("prevet"), &["prevent"]),
            (
                dictgen::InsensitiveStr::Ascii("paralelisation"),
                &["parallelisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("neigbhour"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("finilly"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("traslated"), &["translated"]),
            (
                dictgen::InsensitiveStr::Ascii("estabilishing"),
                &["establishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("althetics"), &["athletics"]),
            (dictgen::InsensitiveStr::Ascii("appares"), &["appears"]),
            (dictgen::InsensitiveStr::Ascii("imploys"), &["employs"]),
            (dictgen::InsensitiveStr::Ascii("exchanced"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("symmetral"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("authorithy"), &["authority"]),
            (dictgen::InsensitiveStr::Ascii("expanions"), &["expansions"]),
            (
                dictgen::InsensitiveStr::Ascii("competitivley"),
                &["competitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalistens"),
                &["journalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramteric"),
                &["parametric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliasers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("analaog"), &["analog"]),
            (
                dictgen::InsensitiveStr::Ascii("accusition"),
                &["accusation"],
            ),
            (dictgen::InsensitiveStr::Ascii("silhuette"), &["silhouette"]),
            (
                dictgen::InsensitiveStr::Ascii("independntly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("oylmpics"), &["olympics"]),
            (dictgen::InsensitiveStr::Ascii("inestart"), &["linestart"]),
            (dictgen::InsensitiveStr::Ascii("bugous"), &["bogus"]),
            (
                dictgen::InsensitiveStr::Ascii("soler"),
                &["solver", "solar", "solely"],
            ),
            (dictgen::InsensitiveStr::Ascii("donwgrade"), &["downgrade"]),
            (dictgen::InsensitiveStr::Ascii("expalins"), &["explains"]),
            (dictgen::InsensitiveStr::Ascii("deocode"), &["decode"]),
            (dictgen::InsensitiveStr::Ascii("scopeing"), &["scoping"]),
            (
                dictgen::InsensitiveStr::Ascii("autogenratet"),
                &["autogenerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("larvy"), &["larvae"]),
            (dictgen::InsensitiveStr::Ascii("refinin"), &["refining"]),
            (
                dictgen::InsensitiveStr::Ascii("conncetions"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chcekpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certficiation"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecutural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("detemined"), &["determined"]),
            (dictgen::InsensitiveStr::Ascii("satisfiied"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("acitvity"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("parmaters"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibitly"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antibioitcs"),
                &["antibiotic", "antibiotics"],
            ),
            (dictgen::InsensitiveStr::Ascii("attmpting"), &["attempting"]),
            (
                dictgen::InsensitiveStr::Ascii("dumplicating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("staurate"), &["saturate"]),
            (
                dictgen::InsensitiveStr::Ascii("cellabration"),
                &["celebration"],
            ),
            (dictgen::InsensitiveStr::Ascii("brittish"), &["british"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corospondingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("piguns"), &["pigeons"]),
            (
                dictgen::InsensitiveStr::Ascii("overcloaking"),
                &["overclocking"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpporter"), &["supporter"]),
            (
                dictgen::InsensitiveStr::Ascii("missmatching"),
                &["mismatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nedles"),
                &["needles", "needless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitilise"),
                &["uninitialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoziation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrolation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compenent"),
                &["component", "competent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elastisize"),
                &["elasticize"],
            ),
            (dictgen::InsensitiveStr::Ascii("whetehr"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("ascneds"), &["ascends"]),
            (dictgen::InsensitiveStr::Ascii("comlains"), &["complains"]),
            (dictgen::InsensitiveStr::Ascii("lokal"), &["local"]),
            (dictgen::InsensitiveStr::Ascii("diableing"), &["disabling"]),
            (
                dictgen::InsensitiveStr::Ascii("incremeants"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directroies"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("prgressed"), &["progressed"]),
            (dictgen::InsensitiveStr::Ascii("guarntes"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("bookeeping"),
                &["bookkeeping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encapsualte"),
                &["encapsulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harcode"),
                &["hardcode", "charcode"],
            ),
            (dictgen::InsensitiveStr::Ascii("butthoel"), &["butthole"]),
            (
                dictgen::InsensitiveStr::Ascii("execises"),
                &["excises", "exercises"],
            ),
            (dictgen::InsensitiveStr::Ascii("somwtimes"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("interacing"),
                &["interacting", "interfacing", "interlacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfmormance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("promisin"), &["promising"]),
            (dictgen::InsensitiveStr::Ascii("somehwo"), &["somehow"]),
            (dictgen::InsensitiveStr::Ascii("marniers"), &["mariners"]),
            (dictgen::InsensitiveStr::Ascii("conists"), &["consists"]),
            (dictgen::InsensitiveStr::Ascii("contstant"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("builddings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("laceholders"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsituents"),
                &["substituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitato"), &["limitation"]),
            (
                dictgen::InsensitiveStr::Ascii("prexisting"),
                &["preexisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("austrila"), &["austria"]),
            (dictgen::InsensitiveStr::Ascii("retrn"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("preample"), &["preamble"]),
            (
                dictgen::InsensitiveStr::Ascii("deomgraphics"),
                &["demographics"],
            ),
            (dictgen::InsensitiveStr::Ascii("litecion"), &["litecoin"]),
            (dictgen::InsensitiveStr::Ascii("aledged"), &["alleged"]),
            (dictgen::InsensitiveStr::Ascii("inspried"), &["inspired"]),
            (dictgen::InsensitiveStr::Ascii("neglible"), &["negligible"]),
            (dictgen::InsensitiveStr::Ascii("coordate"), &["coordinate"]),
            (dictgen::InsensitiveStr::Ascii("certfied"), &["certified"]),
            (
                dictgen::InsensitiveStr::Ascii("comformations"),
                &["conformations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addrerssser"),
                &["addresser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reasearcher"),
                &["researcher"],
            ),
            (dictgen::InsensitiveStr::Ascii("umbrela"), &["umbrella"]),
            (dictgen::InsensitiveStr::Ascii("diql"), &["dial"]),
            (
                dictgen::InsensitiveStr::Ascii("triangultaion"),
                &["triangulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introcuced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electricion"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comtemporary"),
                &["contemporary"],
            ),
            (dictgen::InsensitiveStr::Ascii("bimask"), &["bitmask"]),
            (dictgen::InsensitiveStr::Ascii("wordlview"), &["worldview"]),
            (
                dictgen::InsensitiveStr::Ascii("conviniency"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("proeperty"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("substitutin"),
                &["substituting", "substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("squirel"),
                &["squirtle", "squirrel"],
            ),
            (dictgen::InsensitiveStr::Ascii("avobe"), &["above"]),
            (dictgen::InsensitiveStr::Ascii("wya"), &["way"]),
            (dictgen::InsensitiveStr::Ascii("bumby"), &["bumpy"]),
            (
                dictgen::InsensitiveStr::Ascii("argumentatie"),
                &["argumentative"],
            ),
            (dictgen::InsensitiveStr::Ascii("secifies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("inpsector"), &["inspector"]),
            (
                dictgen::InsensitiveStr::Ascii("comparatie"),
                &["comparative"],
            ),
            (dictgen::InsensitiveStr::Ascii("mormonsim"), &["mormonism"]),
            (dictgen::InsensitiveStr::Ascii("rulebok"), &["rulebook"]),
            (
                dictgen::InsensitiveStr::Ascii("declaratons"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("veicles"), &["vehicles"]),
            (dictgen::InsensitiveStr::Ascii("enbling"), &["enabling"]),
            (dictgen::InsensitiveStr::Ascii("sart"), &["star", "start"]),
            (dictgen::InsensitiveStr::Ascii("armistis"), &["armistice"]),
            (dictgen::InsensitiveStr::Ascii("relaimed"), &["reclaimed"]),
            (dictgen::InsensitiveStr::Ascii("qauntum"), &["quantum"]),
            (dictgen::InsensitiveStr::Ascii("calloed"), &["called"]),
            (dictgen::InsensitiveStr::Ascii("palatte"), &["palette"]),
            (dictgen::InsensitiveStr::Ascii("adresed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("staionary"), &["stationary"]),
            (dictgen::InsensitiveStr::Ascii("indiands"), &["indians"]),
            (dictgen::InsensitiveStr::Ascii("simlpy"), &["simply"]),
            (
                dictgen::InsensitiveStr::Ascii("carciature"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transofrmer"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("expepect"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("independet"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenificating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitifier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("novembeard"), &["november"]),
            (dictgen::InsensitiveStr::Ascii("educaiton"), &["education"]),
            (
                dictgen::InsensitiveStr::Ascii("contradics"),
                &["contradicts"],
            ),
            (dictgen::InsensitiveStr::Ascii("jelousy"), &["jealousy"]),
            (
                dictgen::InsensitiveStr::Ascii("conspiracize"),
                &["conspiracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparitors"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumation"),
                &["enumeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("influeced"), &["influenced"]),
            (dictgen::InsensitiveStr::Ascii("portalus"), &["portals"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetraitor"),
                &["perpetrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("unamed"), &["unnamed"]),
            (dictgen::InsensitiveStr::Ascii("perferred"), &["preferred"]),
            (
                dictgen::InsensitiveStr::Ascii("philospohy"),
                &["philosophy"],
            ),
            (dictgen::InsensitiveStr::Ascii("santizer"), &["sanitizer"]),
            (dictgen::InsensitiveStr::Ascii("nucelar"), &["nuclear"]),
            (
                dictgen::InsensitiveStr::Ascii("dinosaurios"),
                &["dinosaurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("multiplys"), &["multiplies"]),
            (dictgen::InsensitiveStr::Ascii("asiprin"), &["aspirin"]),
            (dictgen::InsensitiveStr::Ascii("strectch"), &["stretch"]),
            (dictgen::InsensitiveStr::Ascii("tweleve"), &["twelve"]),
            (
                dictgen::InsensitiveStr::Ascii("scientificlly"),
                &["scientifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntacts"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("disastrosa"),
                &["disastrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifiation"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmitions"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrantieing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vidoegames"),
                &["videogames"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overhtinking"),
                &["overthinking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintinaing"),
                &["maintaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("onwwards"), &["onwards"]),
            (
                dictgen::InsensitiveStr::Ascii("availabity"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fltering"),
                &["filtering", "faltering", "fluttering", "flattering"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomtric"), &["geometric"]),
            (dictgen::InsensitiveStr::Ascii("sentreis"), &["sentries"]),
            (dictgen::InsensitiveStr::Ascii("ahppy"), &["happy"]),
            (dictgen::InsensitiveStr::Ascii("depolying"), &["deploying"]),
            (dictgen::InsensitiveStr::Ascii("worlld"), &["world"]),
            (dictgen::InsensitiveStr::Ascii("alyer"), &["layer"]),
            (dictgen::InsensitiveStr::Ascii("redunancy"), &["redundancy"]),
            (dictgen::InsensitiveStr::Ascii("beliefed"), &["believed"]),
            (dictgen::InsensitiveStr::Ascii("pipelien"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("testng"), &["testing"]),
            (dictgen::InsensitiveStr::Ascii("assum"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("nineth"), &["ninth"]),
            (
                dictgen::InsensitiveStr::Ascii("initailized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("paranoya"), &["paranoia"]),
            (
                dictgen::InsensitiveStr::Ascii("ortogonality"),
                &["orthogonality"],
            ),
            (dictgen::InsensitiveStr::Ascii("migitate"), &["mitigate"]),
            (dictgen::InsensitiveStr::Ascii("depdended"), &["depended"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotitations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evengalical"),
                &["evangelical"],
            ),
            (dictgen::InsensitiveStr::Ascii("implusive"), &["impulsive"]),
            (dictgen::InsensitiveStr::Ascii("orcales"), &["oracles"]),
            (
                dictgen::InsensitiveStr::Ascii("distriutes"),
                &["distributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antibioitic"),
                &["antibiotic"],
            ),
            (dictgen::InsensitiveStr::Ascii("makros"), &["macros"]),
            (dictgen::InsensitiveStr::Ascii("propbably"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("oeprates"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("capatilized"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transisitions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("youlogy"), &["eulogy"]),
            (
                dictgen::InsensitiveStr::Ascii("attatchment"),
                &["attachment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undenyably"),
                &["undeniably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsaveis"),
                &["responsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucsessive"),
                &["successive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("decocder"), &["decoder"]),
            (
                dictgen::InsensitiveStr::Ascii("instantaneos"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegitime"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repreesnted"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleetly"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloowed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("actived"),
                &["activated", "active"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convuluting"),
                &["convoluting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranversal"),
                &["transversal", "traversal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referincing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discipliniary"),
                &["disciplinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantiaties"),
                &["instantiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeriment"), &["experiment"]),
            (dictgen::InsensitiveStr::Ascii("curteus"), &["courteous"]),
            (
                dictgen::InsensitiveStr::Ascii("imprisonned"),
                &["imprisoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("evreyones"), &["everyones"]),
            (dictgen::InsensitiveStr::Ascii("armpitt"), &["armpit"]),
            (dictgen::InsensitiveStr::Ascii("mssing"), &["missing"]),
            (
                dictgen::InsensitiveStr::Ascii("expentancy"),
                &["expectancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetsing"), &["testing"]),
            (dictgen::InsensitiveStr::Ascii("ealier"), &["earlier"]),
            (dictgen::InsensitiveStr::Ascii("sighn"), &["sign"]),
            (dictgen::InsensitiveStr::Ascii("driecting"), &["directing"]),
            (
                dictgen::InsensitiveStr::Ascii("gerenators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("decutable"), &["deductible"]),
            (
                dictgen::InsensitiveStr::Ascii("assasinated"),
                &["assassinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticisn"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virualization"),
                &["visualization", "virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("volounteer"), &["volunteer"]),
            (
                dictgen::InsensitiveStr::Ascii("nstallation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separtators"),
                &["separators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opportuntiy"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("regurlar"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("blonging"), &["belonging"]),
            (
                dictgen::InsensitiveStr::Ascii("noramlising"),
                &["normalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downlaoded"),
                &["downloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("defineed"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("claened"), &["cleaned"]),
            (dictgen::InsensitiveStr::Ascii("progreess"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("floppyies"), &["floppies"]),
            (
                dictgen::InsensitiveStr::Ascii("championchips"),
                &["championship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("homeworks"), &["homework"]),
            (
                dictgen::InsensitiveStr::Ascii("transfmromations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("torelable"), &["tolerable"]),
            (
                dictgen::InsensitiveStr::Ascii("diferred"),
                &["differed", "deferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("chasims"), &["chasms"]),
            (dictgen::InsensitiveStr::Ascii("literarely"), &["literary"]),
            (
                dictgen::InsensitiveStr::Ascii("documentery"),
                &["documentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("asymmery"), &["asymmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("recconection"),
                &["reconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitied"), &["limited"]),
            (dictgen::InsensitiveStr::Ascii("pratice"), &["practice"]),
            (
                dictgen::InsensitiveStr::Ascii("banchmarked"),
                &["benchmarked"],
            ),
            (dictgen::InsensitiveStr::Ascii("clien"), &["client"]),
            (dictgen::InsensitiveStr::Ascii("loclaizes"), &["localizes"]),
            (
                dictgen::InsensitiveStr::Ascii("missclassifies"),
                &["misclassifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("freindlies"),
                &["friendlies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissict"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolised"),
                &["capitalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delibirately"),
                &["deliberately"],
            ),
            (dictgen::InsensitiveStr::Ascii("gerenics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("politicing"),
                &["politician", "politicking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sovereigny"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("inershia"), &["inertia"]),
            (
                dictgen::InsensitiveStr::Ascii("checkcboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandinava"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("requiered"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("ideologe"), &["ideologies"]),
            (dictgen::InsensitiveStr::Ascii("warehoses"), &["warehouses"]),
            (dictgen::InsensitiveStr::Ascii("deboucing"), &["debouncing"]),
            (
                dictgen::InsensitiveStr::Ascii("cusstomized"),
                &["customized"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrument"), &["instrument"]),
            (
                dictgen::InsensitiveStr::Ascii("immediantely"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitamitely"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("busniessmen"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispather"), &["dispatcher"]),
            (dictgen::InsensitiveStr::Ascii("remvoing"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("segementations"),
                &["segmentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transprency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heteresexual"),
                &["heterosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabulility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("layrinks"), &["larynx"]),
            (dictgen::InsensitiveStr::Ascii("servise"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("sewrvice"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("acyrlic"), &["acrylic"]),
            (dictgen::InsensitiveStr::Ascii("refrers"), &["refers"]),
            (dictgen::InsensitiveStr::Ascii("offenest"), &["oftenest"]),
            (dictgen::InsensitiveStr::Ascii("naieve"), &["naive"]),
            (
                dictgen::InsensitiveStr::Ascii("reprsented"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congigurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inspecitons"),
                &["inspections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotatible"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("evoluated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("varioables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("ymbol"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("boganveelias"),
                &["bougainvilleas"],
            ),
            (dictgen::InsensitiveStr::Ascii("nihilim"), &["nihilism"]),
            (dictgen::InsensitiveStr::Ascii("thopmson"), &["thompson"]),
            (
                dictgen::InsensitiveStr::Ascii("pressiuosly"),
                &["preciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsolated"), &["obsoleted"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotitations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constraing"),
                &["constraining", "constraint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beliveably"),
                &["believably"],
            ),
            (dictgen::InsensitiveStr::Ascii("jus"), &["just"]),
            (dictgen::InsensitiveStr::Ascii("connetet"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("meeing"), &["meeting"]),
            (dictgen::InsensitiveStr::Ascii("harrassed"), &["harassed"]),
            (dictgen::InsensitiveStr::Ascii("vectores"), &["vectors"]),
            (
                dictgen::InsensitiveStr::Ascii("accurracies"),
                &["accuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusivly"),
                &["exclusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitionals"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restriktive"),
                &["restrictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoyences"),
                &["annoyances"],
            ),
            (dictgen::InsensitiveStr::Ascii("specailty"), &["specialty"]),
            (dictgen::InsensitiveStr::Ascii("oregeno"), &["oregano"]),
            (dictgen::InsensitiveStr::Ascii("exhanges"), &["exchanges"]),
            (
                dictgen::InsensitiveStr::Ascii("insturmentals"),
                &["instrumental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contritutions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("peaples"), &["peoples"]),
            (dictgen::InsensitiveStr::Ascii("ansered"), &["answered"]),
            (dictgen::InsensitiveStr::Ascii("langiages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("conectors"), &["connectors"]),
            (
                dictgen::InsensitiveStr::Ascii("calculalate"),
                &["calculate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarator"),
                &["ameliorator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goalkepeer"),
                &["goalkeeper"],
            ),
            (dictgen::InsensitiveStr::Ascii("desogner"), &["designer"]),
            (
                dictgen::InsensitiveStr::Ascii("conideration"),
                &["consideration"],
            ),
            (dictgen::InsensitiveStr::Ascii("manfesto"), &["manifesto"]),
            (
                dictgen::InsensitiveStr::Ascii("adventrues"),
                &["adventures"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypocriet"), &["hypocrite"]),
            (
                dictgen::InsensitiveStr::Ascii("anecdotale"),
                &["anecdote", "anecdotal", "anecdotally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attachmants"),
                &["attachments"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomant"), &["performant"]),
            (dictgen::InsensitiveStr::Ascii("assihnees"), &["assignees"]),
            (dictgen::InsensitiveStr::Ascii("finxed"), &["fixed"]),
            (
                dictgen::InsensitiveStr::Ascii("sarimonies"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternatevly"),
                &["alternately"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplhabet"), &["alphabet"]),
            (dictgen::InsensitiveStr::Ascii("becuz"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("contradiccion"),
                &["contradiction"],
            ),
            (dictgen::InsensitiveStr::Ascii("shgould"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("perosn"), &["person"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiecations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("requrie"), &["require"]),
            (dictgen::InsensitiveStr::Ascii("subroutin"), &["subroutine"]),
            (
                dictgen::InsensitiveStr::Ascii("shdows"),
                &["shadows", "shows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architechs"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("margers"),
                &["mergers", "markers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticiaptes"),
                &["anticipates"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrytpted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("virualize"),
                &["virtualize", "visualize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastrinated"),
                &["procrastinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidencial"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("exclusiv"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("desentisize"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempurtures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconnction"),
                &["reconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dynamicily"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touranment"),
                &["tournaments"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknouwn"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("despide"), &["despised"]),
            (
                dictgen::InsensitiveStr::Ascii("spacialized"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("duobled"), &["doubled"]),
            (dictgen::InsensitiveStr::Ascii("reslovers"), &["resolvers"]),
            (
                dictgen::InsensitiveStr::Ascii("insesitively"),
                &["insensitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("tarceback"), &["traceback"]),
            (
                dictgen::InsensitiveStr::Ascii("authethicates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("htink"), &["think"]),
            (
                dictgen::InsensitiveStr::Ascii("completitions"),
                &["completions", "competitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("diaplayed"), &["displayed"]),
            (
                dictgen::InsensitiveStr::Ascii("multplayer"),
                &["multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("unhelathy"), &["unhealthy"]),
            (dictgen::InsensitiveStr::Ascii("afor"), &["for"]),
            (dictgen::InsensitiveStr::Ascii("beginings"), &["beginnings"]),
            (
                dictgen::InsensitiveStr::Ascii("enoucnters"),
                &["encounters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etends"),
                &["extends", "attends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multuplier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("anual"), &["annual"]),
            (
                dictgen::InsensitiveStr::Ascii("heders"),
                &["headers", "herders"],
            ),
            (dictgen::InsensitiveStr::Ascii("temprol"), &["temporal"]),
            (dictgen::InsensitiveStr::Ascii("swedisn"), &["swedish"]),
            (
                dictgen::InsensitiveStr::Ascii("unannomusly"),
                &["unanimously"],
            ),
            (dictgen::InsensitiveStr::Ascii("referened"), &["referenced"]),
            (
                dictgen::InsensitiveStr::Ascii("asymptoptic"),
                &["asymptotic"],
            ),
            (dictgen::InsensitiveStr::Ascii("ueful"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("invisiable"), &["invisible"]),
            (
                dictgen::InsensitiveStr::Ascii("fluoroscent"),
                &["fluorescent"],
            ),
            (dictgen::InsensitiveStr::Ascii("spindrel"), &["spindle"]),
            (dictgen::InsensitiveStr::Ascii("decid"), &["decide"]),
            (
                dictgen::InsensitiveStr::Ascii("interuptability"),
                &["interruptibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempation"), &["temptation"]),
            (
                dictgen::InsensitiveStr::Ascii("interruptting"),
                &["interrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("queset"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("coalscence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intentialy"),
                &["intentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppse"), &["suppose"]),
            (
                dictgen::InsensitiveStr::Ascii("singularty"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("usptarts"), &["upstarts"]),
            (dictgen::InsensitiveStr::Ascii("undeflows"), &["underflows"]),
            (
                dictgen::InsensitiveStr::Ascii("initliazation"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressivsm"),
                &["progressives"],
            ),
            (dictgen::InsensitiveStr::Ascii("undfine"), &["undefine"]),
            (
                dictgen::InsensitiveStr::Ascii("authoratatively"),
                &["authoritatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penatenturies"),
                &["penitentiaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multliplied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("warninings"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gramatically"),
                &["grammatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("confunder"), &["confounder"]),
            (dictgen::InsensitiveStr::Ascii("stumbeld"), &["stumbled"]),
            (
                dictgen::InsensitiveStr::Ascii("normlising"),
                &["normalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monogameous"),
                &["monogamous"],
            ),
            (dictgen::InsensitiveStr::Ascii("speep"), &["sleep", "speed"]),
            (dictgen::InsensitiveStr::Ascii("disgned"), &["designed"]),
            (dictgen::InsensitiveStr::Ascii("quotting"), &["quoting"]),
            (dictgen::InsensitiveStr::Ascii("dictioary"), &["dictionary"]),
            (
                dictgen::InsensitiveStr::Ascii("neutrailty"),
                &["neutrality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automonomous"),
                &["autonomous"],
            ),
            (dictgen::InsensitiveStr::Ascii("watchdong"), &["watchdog"]),
            (
                dictgen::InsensitiveStr::Ascii("syncronises"),
                &["synchronises"],
            ),
            (dictgen::InsensitiveStr::Ascii("transcrit"), &["transcript"]),
            (
                dictgen::InsensitiveStr::Ascii("utiltiarian"),
                &["utilitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("buitlin"), &["builtin"]),
            (
                dictgen::InsensitiveStr::Ascii("archetecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("decribed"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("landscae"), &["landscapes"]),
            (
                dictgen::InsensitiveStr::Ascii("discontinuty"),
                &["discontinuity"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbous"), &["verbose"]),
            (
                dictgen::InsensitiveStr::Ascii("apliances"),
                &["appliances", "alliances"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitier"), &["limiter"]),
            (
                dictgen::InsensitiveStr::Ascii("fuctionally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersecton"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastonating"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("extist"), &["exist"]),
            (
                dictgen::InsensitiveStr::Ascii("interraction"),
                &["interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceved"), &["concede"]),
            (
                dictgen::InsensitiveStr::Ascii("errorneously"),
                &["erroneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("venezeula"), &["venezuela"]),
            (dictgen::InsensitiveStr::Ascii("charachter"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("icelandinc"), &["icelandic"]),
            (dictgen::InsensitiveStr::Ascii("skirmesh"), &["skirmish"]),
            (
                dictgen::InsensitiveStr::Ascii("dpeendents"),
                &["dependents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartisation"),
                &["standardisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuted"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisitcally"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenificated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sirectories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchagnge"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("generelization"),
                &["generalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrpt"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("lookp"),
                &["look", "looks", "lookup"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opearitons"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexpect"),
                &["inexpected", "unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("priodic"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("archtiectural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowleged"),
                &["acknowledge", "acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("tourisy"), &["touristy"]),
            (dictgen::InsensitiveStr::Ascii("opeartes"), &["operates"]),
            (dictgen::InsensitiveStr::Ascii("exeuted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("opthalmologist"),
                &["ophthalmologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertisiment"),
                &["advertisement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundrasing"),
                &["fundraising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributior"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("stemmin"), &["stemming"]),
            (dictgen::InsensitiveStr::Ascii("morges"), &["morgues"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulenv"),
                &["virtualenv"],
            ),
            (dictgen::InsensitiveStr::Ascii("conveting"), &["converting"]),
            (dictgen::InsensitiveStr::Ascii("imforms"), &["informs"]),
            (dictgen::InsensitiveStr::Ascii("medoicre"), &["mediocre"]),
            (
                dictgen::InsensitiveStr::Ascii("derogatery"),
                &["derogatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionalites"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("mystrows"), &["maestros"]),
            (dictgen::InsensitiveStr::Ascii("acception"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("ameelarated"),
                &["ameliorated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priveliges"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("accodring"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("sodders"), &["solders"]),
            (dictgen::InsensitiveStr::Ascii("recored"), &["recorded"]),
            (dictgen::InsensitiveStr::Ascii("detctor"), &["detector"]),
            (
                dictgen::InsensitiveStr::Ascii("overweigth"),
                &["overweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloreen"), &["chlorine"]),
            (
                dictgen::InsensitiveStr::Ascii("hystericly"),
                &["hysterical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrollers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("equivlent"), &["equivalent"]),
            (dictgen::InsensitiveStr::Ascii("intrigured"), &["intrigue"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpectdly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsivle"),
                &["responsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumulotor"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardaic"), &["cardiac"]),
            (
                dictgen::InsensitiveStr::Ascii("infestions"),
                &["infections"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtters"), &["formatters"]),
            (dictgen::InsensitiveStr::Ascii("futture"), &["future"]),
            (dictgen::InsensitiveStr::Ascii("dinamics"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("transverer"),
                &["traverser", "transverter", "transferer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintenence"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photografy"),
                &["photography"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanell"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("menetions"), &["mentions"]),
            (
                dictgen::InsensitiveStr::Ascii("heistation"),
                &["hesitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimaze"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("extenison"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("insititution"),
                &["institution"],
            ),
            (dictgen::InsensitiveStr::Ascii("editiojn"), &["edition"]),
            (dictgen::InsensitiveStr::Ascii("exrension"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("aproximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conidition"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("dipslay"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("retreivable"),
                &["retrievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elamentries"),
                &["elementaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agggregation"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicating"), &["dictating"]),
            (
                dictgen::InsensitiveStr::Ascii("configureing"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceutes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("contribtor"),
                &["contributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciatie"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momomentary"),
                &["momentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("happending"), &["happening"]),
            (dictgen::InsensitiveStr::Ascii("isteners"), &["listeners"]),
            (
                dictgen::InsensitiveStr::Ascii("acceptence"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("guerrilas"), &["guerrillas"]),
            (dictgen::InsensitiveStr::Ascii("failuers"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("progressivas"),
                &["progressives"],
            ),
            (dictgen::InsensitiveStr::Ascii("failsave"), &["failsafe"]),
            (
                dictgen::InsensitiveStr::Ascii("alignmments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composuting"),
                &["compositing"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeations"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("strenghtening"),
                &["strengthening"],
            ),
            (dictgen::InsensitiveStr::Ascii("acorting"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("elment"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("harranged"),
                &["arranged", "harangued"],
            ),
            (dictgen::InsensitiveStr::Ascii("divisable"), &["divisible"]),
            (
                dictgen::InsensitiveStr::Ascii("intitialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechansim"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("sepecifiers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("renoun"), &["renown"]),
            (
                dictgen::InsensitiveStr::Ascii("armagedons"),
                &["armageddons"],
            ),
            (dictgen::InsensitiveStr::Ascii("definifing"), &["defining"]),
            (dictgen::InsensitiveStr::Ascii("acoount"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("stategise"), &["strategise"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifiing"),
                &["qualifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("encordings"), &["encodings"]),
            (dictgen::InsensitiveStr::Ascii("hhttps"), &["https"]),
            (
                dictgen::InsensitiveStr::Ascii("volentiered"),
                &["volunteered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playtrhough"),
                &["playthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("overloards"), &["overload"]),
            (
                dictgen::InsensitiveStr::Ascii("documentory"),
                &["documentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differrence"),
                &["difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("mehses"), &["meshes"]),
            (
                dictgen::InsensitiveStr::Ascii("outsourceing"),
                &["outsourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregration"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boootstrapped"),
                &["bootstrapped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surroudings"),
                &["surroundings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progamatically"),
                &["programmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("genetive"), &["genitive"]),
            (dictgen::InsensitiveStr::Ascii("takeway"), &["takeaway"]),
            (
                dictgen::InsensitiveStr::Ascii("nutritinal"),
                &["nutritional"],
            ),
            (dictgen::InsensitiveStr::Ascii("comopose"), &["compose"]),
            (dictgen::InsensitiveStr::Ascii("comunist"), &["communist"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinations"),
                &["terminations"],
            ),
            (dictgen::InsensitiveStr::Ascii("valide"), &["valid"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotion"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportionaly"),
                &["proportionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("accunt"), &["account"]),
            (
                dictgen::InsensitiveStr::Ascii("administraions"),
                &["administrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuations"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofer"), &["offer"]),
            (dictgen::InsensitiveStr::Ascii("waitong"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("siganlled"), &["signalled"]),
            (
                dictgen::InsensitiveStr::Ascii("freindzoned"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("spectracal"), &["spectral"]),
            (
                dictgen::InsensitiveStr::Ascii("suggesttions"),
                &["suggestions"],
            ),
            (dictgen::InsensitiveStr::Ascii("recusrion"), &["recursion"]),
            (
                dictgen::InsensitiveStr::Ascii("valudating"),
                &["validating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decreate"),
                &["decrease", "desecrate", "destroy", "discrete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degnerates"),
                &["degenerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("varified"), &["verified"]),
            (
                dictgen::InsensitiveStr::Ascii("difusing"),
                &["diffusing", "defusing"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragmnet"), &["fragment"]),
            (dictgen::InsensitiveStr::Ascii("streamm"), &["stream"]),
            (dictgen::InsensitiveStr::Ascii("pretens"), &["pretends"]),
            (
                dictgen::InsensitiveStr::Ascii("patriatism"),
                &["patriotism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descussing"),
                &["discussing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcemnts"),
                &["announcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("awesomelly"), &["awesomely"]),
            (dictgen::InsensitiveStr::Ascii("resuorced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("antagnoist"),
                &["antagonist"],
            ),
            (dictgen::InsensitiveStr::Ascii("failured"), &["failed"]),
            (dictgen::InsensitiveStr::Ascii("forwad"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("apparances"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insantiates"),
                &["instantiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("birghtening"),
                &["brightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pretendendo"),
                &["pretended"],
            ),
            (dictgen::InsensitiveStr::Ascii("shpae"), &["shape"]),
            (
                dictgen::InsensitiveStr::Ascii("scripturae"),
                &["scriptures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternativey"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("incements"), &["increments"]),
            (dictgen::InsensitiveStr::Ascii("funczion"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("precentile"),
                &["percentile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncondisionaly"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peristence"),
                &["persistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("opinoniated"),
                &["opinionated"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceepted"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("branck"), &["branch"]),
            (dictgen::InsensitiveStr::Ascii("abput"), &["about"]),
            (
                dictgen::InsensitiveStr::Ascii("simlifying"),
                &["simplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derprecation"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bascily"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawels"),
                &["withdrawals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterprenuer"),
                &["entrepreneur"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloacks"), &["cloaks"]),
            (dictgen::InsensitiveStr::Ascii("plattoe"), &["plateau"]),
            (
                dictgen::InsensitiveStr::Ascii("overcloaked"),
                &["overclocked"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressapee"), &["recipe"]),
            (dictgen::InsensitiveStr::Ascii("hihglight"), &["highlight"]),
            (
                dictgen::InsensitiveStr::Ascii("miscummunication"),
                &["miscommunication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrope"),
                &["catastrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("penatlies"), &["penalties"]),
            (dictgen::InsensitiveStr::Ascii("girlfried"), &["girlfriend"]),
            (
                dictgen::InsensitiveStr::Ascii("fwankenstein"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotaition"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnesessarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("displatched"),
                &["dispatched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rickoshayed"),
                &["ricocheted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teetotlers"),
                &["teetotalers"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutes"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("occulation"),
                &["occupation"],
            ),
            (dictgen::InsensitiveStr::Ascii("currentlys"), &["currents"]),
            (dictgen::InsensitiveStr::Ascii("iteators"), &["iterators"]),
            (dictgen::InsensitiveStr::Ascii("marrige"), &["marriage"]),
            (
                dictgen::InsensitiveStr::Ascii("internval"),
                &["interval", "internal"],
            ),
            (dictgen::InsensitiveStr::Ascii("davantage"), &["advantage"]),
            (
                dictgen::InsensitiveStr::Ascii("summarisin"),
                &["summarising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coyprights"),
                &["copyrights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferncencer"),
                &["dereferencer"],
            ),
            (dictgen::InsensitiveStr::Ascii("pllatforms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("postiton"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("cauched"), &["caught"]),
            (dictgen::InsensitiveStr::Ascii("fantasic"), &["fantastic"]),
            (dictgen::InsensitiveStr::Ascii("potenial"), &["potential"]),
            (dictgen::InsensitiveStr::Ascii("capible"), &["capable"]),
            (dictgen::InsensitiveStr::Ascii("syracusae"), &["syracuse"]),
            (dictgen::InsensitiveStr::Ascii("comandos"), &["commandos"]),
            (dictgen::InsensitiveStr::Ascii("extnend"), &["extend"]),
            (dictgen::InsensitiveStr::Ascii("louisianna"), &["louisiana"]),
            (dictgen::InsensitiveStr::Ascii("execues"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("unsoliciated"),
                &["unsolicited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomondating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mediphors"), &["metaphors"]),
            (
                dictgen::InsensitiveStr::Ascii("cuztomizations"),
                &["customizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("strechted"), &["stretched"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificare"),
                &["sacrifice"],
            ),
            (dictgen::InsensitiveStr::Ascii("socripting"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("warnnings"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("neumeric"), &["numeric"]),
            (
                dictgen::InsensitiveStr::Ascii("farehnheit"),
                &["fahrenheit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sourrounded"),
                &["surrounded"],
            ),
            (dictgen::InsensitiveStr::Ascii("rigoreous"), &["rigorous"]),
            (dictgen::InsensitiveStr::Ascii("listeneres"), &["listeners"]),
            (
                dictgen::InsensitiveStr::Ascii("awkwardess"),
                &["awkwardness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continualy"),
                &["continually"],
            ),
            (dictgen::InsensitiveStr::Ascii("polietly"), &["politely"]),
            (dictgen::InsensitiveStr::Ascii("beteen"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("outtermost"), &["outermost"]),
            (
                dictgen::InsensitiveStr::Ascii("documentarsi"),
                &["documentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignent"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("authendicates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avaluating"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ontrollers"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inappropriat"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdence"), &["dependence"]),
            (dictgen::InsensitiveStr::Ascii("garbagge"), &["garbage"]),
            (
                dictgen::InsensitiveStr::Ascii("infeccious"),
                &["infectious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capablities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("exexution"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("mechnanism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("infiltratior"),
                &["infiltrator"],
            ),
            (dictgen::InsensitiveStr::Unicode("paínt"), &["paint"]),
            (dictgen::InsensitiveStr::Ascii("aftzer"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("iniates"), &["initiates"]),
            (dictgen::InsensitiveStr::Ascii("eerpom"), &["eeprom"]),
            (dictgen::InsensitiveStr::Ascii("populative"), &["populate"]),
            (dictgen::InsensitiveStr::Ascii("ahve"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("divice"),
                &["device", "divide"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoltuions"),
                &["resolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("protines"), &["proteins"]),
            (dictgen::InsensitiveStr::Ascii("jscipt"), &["jscript"]),
            (
                dictgen::InsensitiveStr::Ascii("trignometric"),
                &["trigonometric"],
            ),
            (dictgen::InsensitiveStr::Ascii("modue"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbatie"),
                &["masturbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unititialized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microsecsond"),
                &["microsecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmamanged"), &["unmanaged"]),
            (dictgen::InsensitiveStr::Ascii("namtables"), &["nametables"]),
            (
                dictgen::InsensitiveStr::Ascii("compitable"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("autmatic"), &["automatic"]),
            (dictgen::InsensitiveStr::Ascii("finallizes"), &["finalizes"]),
            (dictgen::InsensitiveStr::Ascii("revlover"), &["revolver"]),
            (dictgen::InsensitiveStr::Ascii("tagnets"), &["tangents"]),
            (
                dictgen::InsensitiveStr::Ascii("execuitable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("poing"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("defininitions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("licenes"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("theese"), &["these"]),
            (
                dictgen::InsensitiveStr::Ascii("enviornments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("impot"), &["import"]),
            (dictgen::InsensitiveStr::Ascii("hnadsfree"), &["handsfree"]),
            (dictgen::InsensitiveStr::Ascii("snaped"), &["snapped"]),
            (
                dictgen::InsensitiveStr::Ascii("sepina"),
                &["subpena", "subpoena"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continenet"),
                &["continents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frane"),
                &["frame", "france", "crane", "franc"],
            ),
            (dictgen::InsensitiveStr::Ascii("lookin"), &["looking"]),
            (dictgen::InsensitiveStr::Ascii("cheksums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("bilangual"), &["bilingual"]),
            (dictgen::InsensitiveStr::Ascii("wariwck"), &["warwick"]),
            (
                dictgen::InsensitiveStr::Ascii("chromosones"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("spripting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("entend"),
                &["intend", "extend"],
            ),
            (dictgen::InsensitiveStr::Ascii("sopund"), &["sound"]),
            (
                dictgen::InsensitiveStr::Ascii("decomposin"),
                &["decomposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarantied"),
                &["quarantine"],
            ),
            (dictgen::InsensitiveStr::Ascii("curvtures"), &["curvatures"]),
            (dictgen::InsensitiveStr::Ascii("socratees"), &["socrates"]),
            (dictgen::InsensitiveStr::Ascii("keyboar"), &["keyboard"]),
            (
                dictgen::InsensitiveStr::Ascii("intelligance"),
                &["intelligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vailidating"),
                &["validating"],
            ),
            (dictgen::InsensitiveStr::Ascii("coaleses"), &["coalesces"]),
            (
                dictgen::InsensitiveStr::Ascii("authencation"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoziates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temptatation"),
                &["temptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("embracin"), &["embracing"]),
            (
                dictgen::InsensitiveStr::Ascii("conflited"),
                &["conflicted", "conflated"],
            ),
            (dictgen::InsensitiveStr::Ascii("abilites"), &["abilities"]),
            (
                dictgen::InsensitiveStr::Ascii("tiebreakker"),
                &["tiebreaker"],
            ),
            (dictgen::InsensitiveStr::Ascii("breathis"), &["breaths"]),
            (dictgen::InsensitiveStr::Ascii("priots"), &["priors"]),
            (
                dictgen::InsensitiveStr::Ascii("introducton"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("seriuosly"), &["seriously"]),
            (
                dictgen::InsensitiveStr::Ascii("reconciliates"),
                &["reconciles"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciguret"), &["cigaret"]),
            (
                dictgen::InsensitiveStr::Ascii("sandwhiching"),
                &["sandwiching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internations"),
                &["interactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyed"),
                &["analysed", "analyzed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overtunned"),
                &["overturned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postiively"),
                &["positively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impremented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("totalitarni"),
                &["totalitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("tarvis"), &["travis"]),
            (dictgen::InsensitiveStr::Ascii("backlghts"), &["backlights"]),
            (
                dictgen::InsensitiveStr::Ascii("contole"),
                &["control", "console", "condole"],
            ),
            (dictgen::InsensitiveStr::Ascii("perusaded"), &["persuaded"]),
            (
                dictgen::InsensitiveStr::Ascii("commuincation"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postcondtions"),
                &["postconditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprectaing"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("snowballes"), &["snowballs"]),
            (dictgen::InsensitiveStr::Ascii("ablums"), &["albums"]),
            (dictgen::InsensitiveStr::Ascii("refromed"), &["reformed"]),
            (
                dictgen::InsensitiveStr::Ascii("judgementals"),
                &["judgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relatiopnship"),
                &["relationship"],
            ),
            (dictgen::InsensitiveStr::Ascii("itention"), &["intention"]),
            (dictgen::InsensitiveStr::Ascii("ecryption"), &["encryption"]),
            (
                dictgen::InsensitiveStr::Ascii("writechetque"),
                &["writecheque"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("initializa"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("analgous"), &["analogous"]),
            (dictgen::InsensitiveStr::Ascii("selecrt"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaing"), &["animating"]),
            (dictgen::InsensitiveStr::Ascii("unprecise"), &["imprecise"]),
            (
                dictgen::InsensitiveStr::Ascii("facilty"),
                &["facility", "faculty"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardridge"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("directionnal"),
                &["directional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperaters"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("redirec"), &["redirect"]),
            (
                dictgen::InsensitiveStr::Ascii("presecuted"),
                &["persecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dossapointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misfourtunes"),
                &["misfortunes"],
            ),
            (dictgen::InsensitiveStr::Ascii("isntaller"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("teminated"), &["terminated"]),
            (
                dictgen::InsensitiveStr::Ascii("perferrmence"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("bachlers"), &["bachelors"]),
            (dictgen::InsensitiveStr::Ascii("stnad"), &["stand"]),
            (dictgen::InsensitiveStr::Ascii("someboby"), &["somebody"]),
            (dictgen::InsensitiveStr::Ascii("singleon"), &["singleton"]),
            (dictgen::InsensitiveStr::Ascii("opperators"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("listernes"), &["listeners"]),
            (
                dictgen::InsensitiveStr::Ascii("sissers"),
                &["scissors", "sisters"],
            ),
            (dictgen::InsensitiveStr::Ascii("conumers"), &["consumers"]),
            (
                dictgen::InsensitiveStr::Ascii("incrmental"),
                &["incremental"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixutre"), &["fixture"]),
            (dictgen::InsensitiveStr::Ascii("lookig"), &["looking"]),
            (
                dictgen::InsensitiveStr::Ascii("impelentation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenatations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("subscirpt"), &["subscript"]),
            (
                dictgen::InsensitiveStr::Ascii("hypocriticial"),
                &["hypocritical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonimous"),
                &["autonomous"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrokflows"), &["workflows"]),
            (
                dictgen::InsensitiveStr::Ascii("manetainer"),
                &["maintainer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arious"),
                &["various", "areous"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbes"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("grap"),
                &["grep", "grape", "graph", "grasp"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heirachically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordonate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occassionally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("blokcer"), &["blocker"]),
            (
                dictgen::InsensitiveStr::Ascii("avalability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("warningss"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("custoizers"),
                &["customizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("squre"),
                &["square", "sure", "squire"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoulte"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("funvtioning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promters"),
                &["promoters", "prompters"],
            ),
            (dictgen::InsensitiveStr::Ascii("treshold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("themselfs"), &["themselves"]),
            (dictgen::InsensitiveStr::Ascii("anoxeria"), &["anorexia"]),
            (
                dictgen::InsensitiveStr::Ascii("exectables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("apended"), &["appended"]),
            (
                dictgen::InsensitiveStr::Ascii("failting"),
                &["failing", "faulting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perimissions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaginaton"),
                &["imagination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("councelling"),
                &["counselling"],
            ),
            (dictgen::InsensitiveStr::Ascii("invisibile"), &["invisible"]),
            (
                dictgen::InsensitiveStr::Ascii("reparamterization"),
                &["reparameterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impplemented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acditionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("trunciated"), &["truncated"]),
            (
                dictgen::InsensitiveStr::Ascii("stomrfront"),
                &["stormfront"],
            ),
            (dictgen::InsensitiveStr::Ascii("assighor"), &["assignor"]),
            (dictgen::InsensitiveStr::Ascii("valailable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("optmize"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("surviver"), &["survivor"]),
            (
                dictgen::InsensitiveStr::Ascii("distration"),
                &["distortion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncensered"),
                &["uncensored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreggates"),
                &["aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chipertexts"),
                &["ciphertexts"],
            ),
            (dictgen::InsensitiveStr::Ascii("rutines"), &["routines"]),
            (
                dictgen::InsensitiveStr::Ascii("properries"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alleigances"),
                &["allegiance", "allegiances"],
            ),
            (dictgen::InsensitiveStr::Ascii("avearge"), &["average"]),
            (
                dictgen::InsensitiveStr::Ascii("atmopsheric"),
                &["atmospheric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deambiguation"),
                &["disambiguation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calssifier"),
                &["classifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("recconect"), &["reconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("koordination"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restorting"),
                &["restarting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dorment"), &["dormant"]),
            (dictgen::InsensitiveStr::Ascii("notmally"), &["normally"]),
            (dictgen::InsensitiveStr::Ascii("idenfier"), &["identifier"]),
            (
                dictgen::InsensitiveStr::Ascii("achievemnt"),
                &["achievement"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoiwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("finctions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("huanted"), &["haunted"]),
            (dictgen::InsensitiveStr::Ascii("tabeview"), &["tabview"]),
            (dictgen::InsensitiveStr::Ascii("depoist"), &["deposit"]),
            (dictgen::InsensitiveStr::Ascii("upperaced"), &["uppercased"]),
            (
                dictgen::InsensitiveStr::Ascii("introvertie"),
                &["introverted"],
            ),
            (dictgen::InsensitiveStr::Ascii("shperes"), &["spheres"]),
            (dictgen::InsensitiveStr::Ascii("preperes"), &["prepares"]),
            (
                dictgen::InsensitiveStr::Ascii("retrictive"),
                &["restrictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilcatedly"),
                &["complicatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aribitrarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shtoping"),
                &["stopping", "shopping"],
            ),
            (dictgen::InsensitiveStr::Ascii("befoer"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("defendis"), &["defends"]),
            (dictgen::InsensitiveStr::Ascii("ancestore"), &["ancestor"]),
            (dictgen::InsensitiveStr::Ascii("entirley"), &["entirely"]),
            (dictgen::InsensitiveStr::Ascii("revrieved"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("transending"),
                &["transcending"],
            ),
            (dictgen::InsensitiveStr::Ascii("acticates"), &["activates"]),
            (dictgen::InsensitiveStr::Ascii("ndoe"), &["node"]),
            (dictgen::InsensitiveStr::Ascii("stuctured"), &["structured"]),
            (dictgen::InsensitiveStr::Ascii("simplfier"), &["simplifier"]),
            (
                dictgen::InsensitiveStr::Ascii("scientficly"),
                &["scientifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("behvaior"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("mainstrem"), &["mainstream"]),
            (
                dictgen::InsensitiveStr::Ascii("sponsership"),
                &["sponsorship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighboorhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("knarls"), &["gnarls"]),
            (
                dictgen::InsensitiveStr::Ascii("thining"),
                &["thinking", "thinning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tangientally"),
                &["tangentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offlaoding"),
                &["offloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contianers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("articule"), &["articulate"]),
            (dictgen::InsensitiveStr::Ascii("southen"), &["southern"]),
            (
                dictgen::InsensitiveStr::Ascii("minneapolites"),
                &["minneapolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hydropilic"),
                &["hydrophilic"],
            ),
            (dictgen::InsensitiveStr::Ascii("perseeding"), &["preceding"]),
            (
                dictgen::InsensitiveStr::Ascii("dissallowed"),
                &["disallowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("rferences"), &["references"]),
            (dictgen::InsensitiveStr::Ascii("entitities"), &["entities"]),
            (dictgen::InsensitiveStr::Ascii("numbbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("contigious"),
                &["contagious", "contiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("rackits"), &["rackets"]),
            (
                dictgen::InsensitiveStr::Ascii("ivalidation"),
                &["invalidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cavarly"), &["cavalry"]),
            (dictgen::InsensitiveStr::Ascii("peanochle"), &["pinochle"]),
            (dictgen::InsensitiveStr::Ascii("pulgged"), &["plugged"]),
            (dictgen::InsensitiveStr::Ascii("movemments"), &["movements"]),
            (dictgen::InsensitiveStr::Ascii("obleek"), &["oblique"]),
            (
                dictgen::InsensitiveStr::Ascii("combainator"),
                &["combinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mediocirty"),
                &["mediocrity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physicials"),
                &["physicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("comlex"), &["complex"]),
            (dictgen::InsensitiveStr::Ascii("everthing"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("builded"), &["built"]),
            (dictgen::InsensitiveStr::Ascii("redenderer"), &["renderer"]),
            (dictgen::InsensitiveStr::Ascii("pression"), &["pressing"]),
            (
                dictgen::InsensitiveStr::Ascii("ureferenced"),
                &["unreferenced", "referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("shurely"), &["surely"]),
            (dictgen::InsensitiveStr::Ascii("amalgum"), &["amalgam"]),
            (dictgen::InsensitiveStr::Ascii("fedreally"), &["federally"]),
            (dictgen::InsensitiveStr::Ascii("artisitc"), &["artistic"]),
            (dictgen::InsensitiveStr::Ascii("marlformed"), &["malformed"]),
            (dictgen::InsensitiveStr::Ascii("orignate"), &["originate"]),
            (dictgen::InsensitiveStr::Ascii("thesits"), &["theists"]),
            (
                dictgen::InsensitiveStr::Ascii("referundum"),
                &["referendum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertisors"),
                &["advertisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepaire"), &["prepare"]),
            (
                dictgen::InsensitiveStr::Ascii("classsifies"),
                &["classifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transferd"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioted"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recursivly"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonymosly"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("essiential"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combindations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnences"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("extrats"), &["extracts"]),
            (dictgen::InsensitiveStr::Ascii("nmbered"), &["numbered"]),
            (dictgen::InsensitiveStr::Ascii("speicfies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("borke"), &["broke"]),
            (dictgen::InsensitiveStr::Ascii("fujistu"), &["fujitsu"]),
            (dictgen::InsensitiveStr::Ascii("wiscles"), &["whistles"]),
            (dictgen::InsensitiveStr::Ascii("wholy"), &["wholly", "holy"]),
            (dictgen::InsensitiveStr::Ascii("negible"), &["negligible"]),
            (
                dictgen::InsensitiveStr::Ascii("ntegrations"),
                &["integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adresees"),
                &["addresses", "addressees"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranposed"), &["transposed"]),
            (dictgen::InsensitiveStr::Ascii("anounced"), &["announced"]),
            (dictgen::InsensitiveStr::Ascii("ralation"), &["relation"]),
            (dictgen::InsensitiveStr::Ascii("childern"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("udid"), &["uuid"]),
            (
                dictgen::InsensitiveStr::Ascii("guarnateees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("proteced"), &["protected"]),
            (dictgen::InsensitiveStr::Ascii("acitons"), &["actions"]),
            (dictgen::InsensitiveStr::Ascii("welocme"), &["welcome"]),
            (
                dictgen::InsensitiveStr::Ascii("suppressingd"),
                &["suppressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("releation"), &["relation"]),
            (
                dictgen::InsensitiveStr::Ascii("mutithreading"),
                &["multithreading"],
            ),
            (dictgen::InsensitiveStr::Ascii("balace"), &["balance"]),
            (
                dictgen::InsensitiveStr::Ascii("heavyweigt"),
                &["heavyweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("factorys"),
                &["factors", "factories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("settlemets"),
                &["settlements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disributor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilse"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independens"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("reliablely"), &["reliably"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotitaions"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egotestical"),
                &["egotistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socailized"),
                &["socialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("levelign"), &["leveling"]),
            (dictgen::InsensitiveStr::Ascii("vacumming"), &["vacuuming"]),
            (dictgen::InsensitiveStr::Ascii("ocasions"), &["occasions"]),
            (dictgen::InsensitiveStr::Ascii("rekenton"), &["renekton"]),
            (dictgen::InsensitiveStr::Ascii("funvtions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("centain"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("analyszing"),
                &["analyzing", "analysing"],
            ),
            (dictgen::InsensitiveStr::Ascii("diselect"), &["deselect"]),
            (dictgen::InsensitiveStr::Ascii("emited"), &["emitted"]),
            (
                dictgen::InsensitiveStr::Ascii("authorizaton"),
                &["authorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistanses"),
                &["resistances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bacgrounds"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelereated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celebartion"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanelog"), &["changelog"]),
            (dictgen::InsensitiveStr::Ascii("stressende"), &["stressed"]),
            (dictgen::InsensitiveStr::Ascii("lighbulbs"), &["lightbulbs"]),
            (dictgen::InsensitiveStr::Ascii("insstances"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("transfom"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("cataclym"), &["cataclysm"]),
            (
                dictgen::InsensitiveStr::Ascii("alorithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("referere"), &["referee"]),
            (
                dictgen::InsensitiveStr::Ascii("remaning"),
                &["remaining", "renaming", "remaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("malpractise"),
                &["malpractice"],
            ),
            (dictgen::InsensitiveStr::Ascii("umit"), &["unit"]),
            (
                dictgen::InsensitiveStr::Ascii("placemet"),
                &["placements", "placement", "placemat"],
            ),
            (dictgen::InsensitiveStr::Ascii("algohmic"), &["algorithmic"]),
            (dictgen::InsensitiveStr::Ascii("handlare"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("otpimise"), &["optimise"]),
            (dictgen::InsensitiveStr::Ascii("problable"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("disolved"), &["dissolved"]),
            (dictgen::InsensitiveStr::Ascii("partiiton"), &["partition"]),
            (
                dictgen::InsensitiveStr::Ascii("acceptaable"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("carvinal"), &["carnival"]),
            (dictgen::InsensitiveStr::Ascii("balconey"), &["balcony"]),
            (dictgen::InsensitiveStr::Ascii("airrflow"), &["airflow"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffektive"),
                &["ineffective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("butterfies"),
                &["butterflies"],
            ),
            (dictgen::InsensitiveStr::Ascii("immediat"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("quotaion"), &["quotation"]),
            (
                dictgen::InsensitiveStr::Ascii("relocaitons"),
                &["relocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuase"), &["cause"]),
            (
                dictgen::InsensitiveStr::Ascii("botstrapper"),
                &["bootstrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuillt"), &["rebuilt"]),
            (
                dictgen::InsensitiveStr::Ascii("positionly"),
                &["positional"],
            ),
            (dictgen::InsensitiveStr::Ascii("domait"), &["domain"]),
            (
                dictgen::InsensitiveStr::Ascii("applicatios"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("globlally"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("incorreect"), &["incorrect"]),
            (
                dictgen::InsensitiveStr::Ascii("unaccessible"),
                &["inaccessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("operaring"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("senitments"),
                &["sentiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("expeceted"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisssism"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("invisivble"), &["invisible"]),
            (
                dictgen::InsensitiveStr::Ascii("insignficant"),
                &["insignificant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patnerships"),
                &["partnerships"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacarstic"), &["sarcastic"]),
            (
                dictgen::InsensitiveStr::Ascii("corrdinators"),
                &["coordinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("asociates"), &["associates"]),
            (
                dictgen::InsensitiveStr::Ascii("insteallations"),
                &["installations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colloquail"),
                &["colloquial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roated"),
                &["rotated", "roasted"],
            ),
            (dictgen::InsensitiveStr::Ascii("encahnt"), &["enchant"]),
            (
                dictgen::InsensitiveStr::Ascii("parlaments"),
                &["parliaments"],
            ),
            (dictgen::InsensitiveStr::Ascii("hysteriska"), &["hysteria"]),
            (dictgen::InsensitiveStr::Ascii("complians"), &["complains"]),
            (dictgen::InsensitiveStr::Ascii("reosnable"), &["reasonable"]),
            (
                dictgen::InsensitiveStr::Ascii("preceed"),
                &["precede", "proceed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serrebrally"),
                &["cerebrally"],
            ),
            (dictgen::InsensitiveStr::Ascii("realibly"), &["reliably"]),
            (
                dictgen::InsensitiveStr::Ascii("recommmending"),
                &["recommending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demographis"),
                &["demographics"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntaxg"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("deprectaes"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("contrieved"), &["contrived"]),
            (dictgen::InsensitiveStr::Ascii("bonsues"), &["bonuses"]),
            (dictgen::InsensitiveStr::Ascii("positoins"), &["positions"]),
            (dictgen::InsensitiveStr::Ascii("addresed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("depelted"), &["depleted"]),
            (dictgen::InsensitiveStr::Ascii("freeezes"), &["freezes"]),
            (dictgen::InsensitiveStr::Ascii("forwarrd"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("everyhting"),
                &["everything"],
            ),
            (dictgen::InsensitiveStr::Ascii("appoligies"), &["apologies"]),
            (
                dictgen::InsensitiveStr::Ascii("nomally"),
                &["normally", "nominally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceintist"), &["scientist"]),
            (
                dictgen::InsensitiveStr::Ascii("fustration"),
                &["frustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpted"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("reigstrations"),
                &["registrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("garantied"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("vallid"), &["valid"]),
            (dictgen::InsensitiveStr::Ascii("opions"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("renaissace"),
                &["renaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sleeped"), &["slept"]),
            (
                dictgen::InsensitiveStr::Ascii("imformation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avisoriyes"),
                &["advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("wroldwide"), &["worldwide"]),
            (dictgen::InsensitiveStr::Ascii("becaseu"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("skateborad"),
                &["skateboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("resset"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoticating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("micrcontroller"),
                &["microcontroller"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatated"), &["operated"]),
            (
                dictgen::InsensitiveStr::Ascii("specificlly"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("flaged"), &["flagged"]),
            (dictgen::InsensitiveStr::Ascii("togetehr"), &["together"]),
            (dictgen::InsensitiveStr::Ascii("fitering"), &["filtering"]),
            (
                dictgen::InsensitiveStr::Ascii("illegitamte"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specializate"),
                &["specialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autenticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("suspendes"), &["suspense"]),
            (dictgen::InsensitiveStr::Ascii("availanle"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("admininistering"),
                &["administering"],
            ),
            (dictgen::InsensitiveStr::Ascii("epandable"), &["expandable"]),
            (
                dictgen::InsensitiveStr::Ascii("anticuated"),
                &["antiquated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tahn"), &["than"]),
            (dictgen::InsensitiveStr::Ascii("corsspost"), &["crosspost"]),
            (
                dictgen::InsensitiveStr::Ascii("proceshandler"),
                &["processhandler"],
            ),
            (dictgen::InsensitiveStr::Ascii("instsances"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("suppied"), &["supplied"]),
            (dictgen::InsensitiveStr::Ascii("withelist"), &["whitelist"]),
            (dictgen::InsensitiveStr::Ascii("liason"), &["liaison"]),
            (
                dictgen::InsensitiveStr::Ascii("eviserated"),
                &["eviscerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independetn"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("ironicly"), &["ironically"]),
            (dictgen::InsensitiveStr::Ascii("dynammic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("ntegrals"), &["integrals"]),
            (dictgen::InsensitiveStr::Ascii("resutls"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("superflouous"),
                &["superfluous"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceenshot"), &["screenshot"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithmms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("longitme"), &["longtime"]),
            (
                dictgen::InsensitiveStr::Ascii("missinterpreting"),
                &["misinterpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolule"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("definitios"),
                &["definition", "definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mastutbation"),
                &["masturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("recieveing"), &["receiving"]),
            (
                dictgen::InsensitiveStr::Ascii("cound"),
                &["could", "count", "found"],
            ),
            (dictgen::InsensitiveStr::Ascii("assassian"), &["assassin"]),
            (
                dictgen::InsensitiveStr::Ascii("hospotality"),
                &["hospitality"],
            ),
            (dictgen::InsensitiveStr::Ascii("yourselvs"), &["yourselves"]),
            (
                dictgen::InsensitiveStr::Ascii("qualitying"),
                &["qualifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("characgter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("demontration"),
                &["demonstration"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlporer"), &["explorer"]),
            (
                dictgen::InsensitiveStr::Ascii("quesitonable"),
                &["questionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cllosing"), &["closing"]),
            (dictgen::InsensitiveStr::Ascii("indicees"), &["indices"]),
            (
                dictgen::InsensitiveStr::Ascii("phychopathic"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuils"), &["rebuilds"]),
            (
                dictgen::InsensitiveStr::Ascii("confimation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertaing"), &["pertaining"]),
            (
                dictgen::InsensitiveStr::Ascii("clearnances"),
                &["clearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditinals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("juducial"), &["judicial"]),
            (dictgen::InsensitiveStr::Ascii("photograh"), &["photograph"]),
            (dictgen::InsensitiveStr::Ascii("japanses"), &["japanese"]),
            (dictgen::InsensitiveStr::Ascii("overtun"), &["overturn"]),
            (
                dictgen::InsensitiveStr::Ascii("imcomplete"),
                &["incomplete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponders"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuplads"), &["reuploads"]),
            (dictgen::InsensitiveStr::Ascii("contzains"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("mulitiply"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithmos"),
                &["algorithm", "algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postioning"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("familair"), &["familiar"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccessibile"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downagraded"),
                &["downgraded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stuructured"),
                &["structured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poistioning"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("bachler"), &["bachelor"]),
            (dictgen::InsensitiveStr::Ascii("upates"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("keneysian"), &["keynesian"]),
            (dictgen::InsensitiveStr::Ascii("pretensje"), &["pretense"]),
            (
                dictgen::InsensitiveStr::Ascii("inspirating"),
                &["inspiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructo"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaailable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidantially"),
                &["confidentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("attatched"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("administed"),
                &["administered"],
            ),
            (dictgen::InsensitiveStr::Ascii("undeflow"), &["underflow"]),
            (dictgen::InsensitiveStr::Ascii("squirle"), &["squirrel"]),
            (
                dictgen::InsensitiveStr::Ascii("belove"),
                &["below", "beloved"],
            ),
            (dictgen::InsensitiveStr::Ascii("expidited"), &["expedited"]),
            (
                dictgen::InsensitiveStr::Ascii("stereotipical"),
                &["stereotypical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciliast"),
                &["specialists"],
            ),
            (dictgen::InsensitiveStr::Ascii("abadoned"), &["abandoned"]),
            (
                dictgen::InsensitiveStr::Ascii("pakistanezi"),
                &["pakistani"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuires"),
                &["requires", "rewires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirearchically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("rutimes"), &["runtimes"]),
            (
                dictgen::InsensitiveStr::Ascii("overengeneer"),
                &["overengineer"],
            ),
            (dictgen::InsensitiveStr::Ascii("fromt"), &["from", "front"]),
            (dictgen::InsensitiveStr::Ascii("reagarding"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("particaular"),
                &["particular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappinted"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumisers"),
                &["customisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("healtheast"),
                &["healthiest"],
            ),
            (dictgen::InsensitiveStr::Ascii("strengten"), &["strengthen"]),
            (dictgen::InsensitiveStr::Ascii("ontraio"), &["ontario"]),
            (dictgen::InsensitiveStr::Ascii("cxan"), &["cyan"]),
            (
                dictgen::InsensitiveStr::Ascii("messager"),
                &["massager", "messenger", "message"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infailable"),
                &["infallible"],
            ),
            (dictgen::InsensitiveStr::Ascii("selcting"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("acknownledge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("accaptor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("inovice"), &["invoice"]),
            (dictgen::InsensitiveStr::Ascii("callcain"), &["callchain"]),
            (
                dictgen::InsensitiveStr::Ascii("succussfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspodent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("yuforically"),
                &["euphorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("yoiur"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("miserablly"), &["miserably"]),
            (dictgen::InsensitiveStr::Ascii("refocuss"), &["refocus"]),
            (dictgen::InsensitiveStr::Ascii("finilize"), &["finalize"]),
            (
                dictgen::InsensitiveStr::Ascii("disonnected"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("streames"), &["streams"]),
            (dictgen::InsensitiveStr::Ascii("traveersed"), &["traversed"]),
            (dictgen::InsensitiveStr::Ascii("destroing"), &["destroying"]),
            (
                dictgen::InsensitiveStr::Ascii("restictive"),
                &["restrictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("megathred"), &["megathread"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleartors"),
                &["accelerators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretarea"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("alegbraic"), &["algebraic"]),
            (dictgen::InsensitiveStr::Ascii("reslove"), &["resolve"]),
            (dictgen::InsensitiveStr::Ascii("runtimr"), &["runtime"]),
            (dictgen::InsensitiveStr::Ascii("pracical"), &["practical"]),
            (
                dictgen::InsensitiveStr::Ascii("blackshits"),
                &["blacksmith"],
            ),
            (dictgen::InsensitiveStr::Ascii("archietect"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("qeustionnaire"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("struggeled"), &["struggled"]),
            (dictgen::InsensitiveStr::Ascii("stiring"), &["stirring"]),
            (
                dictgen::InsensitiveStr::Ascii("insatllations"),
                &["installations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differentiatin"),
                &["differentiating", "differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acclerometer"),
                &["accelerometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolagetic"),
                &["apologetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constributors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("thurdsays"), &["thursdays"]),
            (
                dictgen::InsensitiveStr::Ascii("inteligantly"),
                &["intelligently"],
            ),
            (dictgen::InsensitiveStr::Ascii("askign"), &["asking"]),
            (
                dictgen::InsensitiveStr::Ascii("iniations"),
                &["initiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peritinkle"),
                &["periwinkle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encapsualted"),
                &["encapsulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("qoute"), &["quote"]),
            (
                dictgen::InsensitiveStr::Ascii("artifically"),
                &["artificially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaccinaties"),
                &["vaccinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elipticity"),
                &["ellipticity"],
            ),
            (dictgen::InsensitiveStr::Ascii("deductie"), &["deductible"]),
            (
                dictgen::InsensitiveStr::Ascii("posssibilities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("refernece"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("differenciation"),
                &["differentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("domapine"), &["dopamine"]),
            (dictgen::InsensitiveStr::Ascii("standartd"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("commection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaction"),
                &["contacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementataions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("protrait"), &["portrait"]),
            (dictgen::InsensitiveStr::Ascii("dopmaine"), &["dopamine"]),
            (dictgen::InsensitiveStr::Ascii("relaized"), &["realized"]),
            (dictgen::InsensitiveStr::Ascii("desireably"), &["desirably"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplicacion"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cusotodians"),
                &["custodians"],
            ),
            (dictgen::InsensitiveStr::Ascii("sustem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("victorieuse"),
                &["victories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("platofrmer"),
                &["platformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulids"), &["builds"]),
            (
                dictgen::InsensitiveStr::Ascii("represensible"),
                &["reprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millimiters"),
                &["millimeters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coversions"),
                &["conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vientamese"),
                &["vietnamese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interersting"),
                &["interpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catepillar"),
                &["caterpillar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenpenendtly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("magnesuim"), &["magnesium"]),
            (dictgen::InsensitiveStr::Ascii("condem"), &["condemn"]),
            (
                dictgen::InsensitiveStr::Ascii("coencidental"),
                &["coincidental"],
            ),
            (dictgen::InsensitiveStr::Ascii("billon"), &["billion"]),
            (
                dictgen::InsensitiveStr::Ascii("nawseously"),
                &["nauseously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parshal"),
                &["marshal", "partial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succeesful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("enterance"), &["entrance"]),
            (
                dictgen::InsensitiveStr::Ascii("inspiritional"),
                &["inspirational"],
            ),
            (dictgen::InsensitiveStr::Ascii("cimptums"), &["symptoms"]),
            (
                dictgen::InsensitiveStr::Ascii("unauthenticed"),
                &["unauthenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternatievs"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conponents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentcation"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("reivew"), &["review"]),
            (
                dictgen::InsensitiveStr::Ascii("configuers"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("brakley"), &["barkley"]),
            (dictgen::InsensitiveStr::Ascii("survivour"), &["survivor"]),
            (dictgen::InsensitiveStr::Ascii("elemets"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulaters"),
                &["manipulators", "manipulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("antibiotc"), &["antibiotic"]),
            (dictgen::InsensitiveStr::Ascii("witheld"), &["withheld"]),
            (
                dictgen::InsensitiveStr::Ascii("observerable"),
                &["observable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumbitter"), &["submitter"]),
            (
                dictgen::InsensitiveStr::Ascii("sapeenaing"),
                &["subpoenaing"],
            ),
            (dictgen::InsensitiveStr::Ascii("stdanard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("alllocate"), &["allocate"]),
            (dictgen::InsensitiveStr::Ascii("packaaging"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("valules"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("optimiality"),
                &["optimality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utilzation"),
                &["utilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("dosclosed"), &["disclosed"]),
            (dictgen::InsensitiveStr::Ascii("cartride"), &["cartridge"]),
            (dictgen::InsensitiveStr::Ascii("oveload"), &["overload"]),
            (
                dictgen::InsensitiveStr::Ascii("retreevable"),
                &["retrievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expersions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("heorics"), &["heroics"]),
            (dictgen::InsensitiveStr::Ascii("severites"), &["severities"]),
            (
                dictgen::InsensitiveStr::Ascii("comparotors"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("alegience"), &["allegiance"]),
            (
                dictgen::InsensitiveStr::Ascii("vertabraes"),
                &["vertebraes"],
            ),
            (dictgen::InsensitiveStr::Ascii("secuence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("schdule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("asnychronously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("geerated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("conplimented"),
                &["complimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("partiular"), &["particular"]),
            (dictgen::InsensitiveStr::Ascii("worfklow"), &["workflow"]),
            (dictgen::InsensitiveStr::Ascii("bandwidthm"), &["bandwidth"]),
            (dictgen::InsensitiveStr::Ascii("srink"), &["shrink"]),
            (dictgen::InsensitiveStr::Ascii("expiriy"), &["expiry"]),
            (dictgen::InsensitiveStr::Ascii("faetures"), &["features"]),
            (dictgen::InsensitiveStr::Ascii("tweeked"), &["tweaked"]),
            (
                dictgen::InsensitiveStr::Ascii("alis"),
                &["alas", "alias", "alms", "axis"],
            ),
            (dictgen::InsensitiveStr::Ascii("pumkin"), &["pumpkin"]),
            (dictgen::InsensitiveStr::Ascii("esumption"), &["resumption"]),
            (dictgen::InsensitiveStr::Ascii("subdoamin"), &["subdomain"]),
            (
                dictgen::InsensitiveStr::Ascii("immigrantes"),
                &["immigrants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prejects"),
                &["prefects", "projects"],
            ),
            (dictgen::InsensitiveStr::Ascii("coleegs"), &["colleagues"]),
            (
                dictgen::InsensitiveStr::Ascii("legalimate"),
                &["legitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrys"), &["retries"]),
            (dictgen::InsensitiveStr::Ascii("soneone"), &["someone"]),
            (
                dictgen::InsensitiveStr::Ascii("validatiors"),
                &["validators", "validations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggree"), &["agree"]),
            (
                dictgen::InsensitiveStr::Ascii("inpersonating"),
                &["impersonating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playwrites"),
                &["playwrights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quites"),
                &["quiets", "quotes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotypying"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appplication"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eumeration"),
                &["enumeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("symmectry"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonative"),
                &["opinionative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coninstallable"),
                &["coinstallable"],
            ),
            (dictgen::InsensitiveStr::Ascii("checed"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("notifyable"),
                &["notifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("horizotal"), &["horizontal"]),
            (dictgen::InsensitiveStr::Ascii("perferrs"), &["prefers"]),
            (dictgen::InsensitiveStr::Ascii("sanytise"), &["sanitise"]),
            (
                dictgen::InsensitiveStr::Ascii("initailzations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("amateus"), &["amateurs"]),
            (
                dictgen::InsensitiveStr::Ascii("preallocationg"),
                &["preallocating", "preallocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sortrage"),
                &["storage", "shortage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplaications"),
                &["duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("mothing"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("absoluate"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("postitioned"),
                &["positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitified"),
                &["identifies", "identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompetive"),
                &["uncompetitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrirbute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("eneables"), &["enables"]),
            (dictgen::InsensitiveStr::Ascii("phenonema"), &["phenomena"]),
            (
                dictgen::InsensitiveStr::Ascii("consistute"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequncer"), &["sequencer"]),
            (dictgen::InsensitiveStr::Ascii("physicis"), &["physicians"]),
            (
                dictgen::InsensitiveStr::Ascii("alphabeticly"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repedeatly"),
                &["repeatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandianvians"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silouwetist"),
                &["silhouettist"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheecksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("lighthness"), &["lightness"]),
            (dictgen::InsensitiveStr::Ascii("oputput"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("jumoed"), &["jumped"]),
            (
                dictgen::InsensitiveStr::Ascii("handshaging"),
                &["handshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("culitvate"), &["cultivate"]),
            (
                dictgen::InsensitiveStr::Ascii("collegaues"),
                &["colleagues"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transistions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("massarce"), &["massacre"]),
            (dictgen::InsensitiveStr::Ascii("occasinal"), &["occasional"]),
            (dictgen::InsensitiveStr::Ascii("kernerls"), &["kernels"]),
            (
                dictgen::InsensitiveStr::Ascii("renegothiators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specilaized"),
                &["specialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("exactlly"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("temportal"), &["temporal"]),
            (
                dictgen::InsensitiveStr::Ascii("interregation"),
                &["interrogation"],
            ),
            (dictgen::InsensitiveStr::Ascii("confugure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("servives"),
                &["services", "survives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriminating"),
                &["discriminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("abolsute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("annotaiotn"),
                &["annotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicionary"), &["dictionary"]),
            (dictgen::InsensitiveStr::Ascii("goood"), &["good"]),
            (dictgen::InsensitiveStr::Ascii("spefifeid"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("regession"),
                &["regression", "recession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consultanti"),
                &["consultation"],
            ),
            (dictgen::InsensitiveStr::Ascii("htaching"), &["hatching"]),
            (
                dictgen::InsensitiveStr::Ascii("assignement"),
                &["assignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journolists"),
                &["journalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacremento"),
                &["sacramento"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inchoerent"),
                &["incoherent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multuplying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("deendency"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("mutiplicities"),
                &["multiplicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comftorable"),
                &["comfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accseptably"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserailising"),
                &["deserialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninitailised"),
                &["uninitialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("vicitms"), &["victims"]),
            (dictgen::InsensitiveStr::Ascii("iternation"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("configruation"),
                &["configurations", "configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regneration"),
                &["regeneration"],
            ),
            (dictgen::InsensitiveStr::Ascii("comnpletes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("spectatularly"),
                &["spectacularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("constnat"), &["constant"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstratbly"),
                &["demonstrably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unambigious"),
                &["unambiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delibertly"),
                &["deliberately"],
            ),
            (dictgen::InsensitiveStr::Ascii("existantes"), &["existent"]),
            (
                dictgen::InsensitiveStr::Ascii("associeting"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coventions"),
                &["conventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("diatance"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("selecrtion"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("amkefiles"), &["makefiles"]),
            (dictgen::InsensitiveStr::Ascii("proced"), &["proceed"]),
            (dictgen::InsensitiveStr::Ascii("lmitation"), &["limitation"]),
            (dictgen::InsensitiveStr::Ascii("arhitect"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("vaticaan"), &["vatican"]),
            (dictgen::InsensitiveStr::Ascii("virgintiy"), &["virginity"]),
            (
                dictgen::InsensitiveStr::Ascii("sedentarity"),
                &["sedentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("origion"), &["origin"]),
            (dictgen::InsensitiveStr::Ascii("neighbers"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("downaloading"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectoins"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unoptimice"),
                &["unoptimise", "unoptimize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fruadulent"),
                &["fraudulent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrativno"),
                &["demonstrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("underlyng"), &["underlying"]),
            (dictgen::InsensitiveStr::Ascii("collegue"), &["colleague"]),
            (dictgen::InsensitiveStr::Ascii("commpiling"), &["compiling"]),
            (
                dictgen::InsensitiveStr::Ascii("overzealisly"),
                &["overzealously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumizes"),
                &["customizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hiearchical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("genuises"), &["geniuses"]),
            (
                dictgen::InsensitiveStr::Ascii("disconeected"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("pilgramig"), &["pilgrimage"]),
            (dictgen::InsensitiveStr::Ascii("vresion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("ultimatly"), &["ultimately"]),
            (
                dictgen::InsensitiveStr::Ascii("specificatin"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissitic"),
                &["narcissistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extremeties"),
                &["extremities"],
            ),
            (dictgen::InsensitiveStr::Ascii("compliling"), &["compiling"]),
            (
                dictgen::InsensitiveStr::Ascii("disposses"),
                &["disposes", "dispossess"],
            ),
            (dictgen::InsensitiveStr::Ascii("anxiuos"), &["anxious"]),
            (
                dictgen::InsensitiveStr::Ascii("popluations"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auxilleries"),
                &["auxiliaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("childlren"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("blodk"), &["block"]),
            (
                dictgen::InsensitiveStr::Ascii("refectivity"),
                &["reflectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wavelenghs"),
                &["wavelength", "wavelengths"],
            ),
            (dictgen::InsensitiveStr::Ascii("elimintate"), &["eliminate"]),
            (
                dictgen::InsensitiveStr::Ascii("accomdates"),
                &["accommodates"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragrahs"), &["paragraphs"]),
            (dictgen::InsensitiveStr::Ascii("sanatize"), &["sanitize"]),
            (dictgen::InsensitiveStr::Ascii("tarcing"), &["tracing"]),
            (
                dictgen::InsensitiveStr::Ascii("implemted"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outskirters"),
                &["outskirts"],
            ),
            (dictgen::InsensitiveStr::Ascii("grpahic"), &["graphic"]),
            (
                dictgen::InsensitiveStr::Ascii("geenrations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highligher"),
                &["highlighter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapparances"),
                &["disappearances"],
            ),
            (dictgen::InsensitiveStr::Ascii("socila"), &["social"]),
            (dictgen::InsensitiveStr::Ascii("firggin"), &["friggin"]),
            (
                dictgen::InsensitiveStr::Ascii("eeded"),
                &["needed", "heeded", "seeded", "weeded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endoced"),
                &["endorsed", "enforced", "encoded"],
            ),
            (dictgen::InsensitiveStr::Ascii("conversly"), &["conversely"]),
            (
                dictgen::InsensitiveStr::Ascii("intiialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("monarkies"), &["monarchies"]),
            (dictgen::InsensitiveStr::Ascii("chromiun"), &["chromium"]),
            (
                dictgen::InsensitiveStr::Ascii("checkpoing"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boudnaries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("neibor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("unrestrcited"),
                &["unrestricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("yestruday"), &["yesterday"]),
            (dictgen::InsensitiveStr::Ascii("jurnaled"), &["journaled"]),
            (dictgen::InsensitiveStr::Ascii("acovado"), &["avocado"]),
            (dictgen::InsensitiveStr::Ascii("disgustus"), &["disgusts"]),
            (
                dictgen::InsensitiveStr::Ascii("intellecutally"),
                &["intellectually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprogramms"),
                &["reprograms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wonderwand"),
                &["wonderland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delaractions"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("multilple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("lesssons"), &["lessons"]),
            (dictgen::InsensitiveStr::Ascii("branckets"), &["brackets"]),
            (dictgen::InsensitiveStr::Ascii("regiter"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("scartch"), &["scratch"]),
            (dictgen::InsensitiveStr::Ascii("historcal"), &["historical"]),
            (dictgen::InsensitiveStr::Ascii("stationy"), &["stationary"]),
            (dictgen::InsensitiveStr::Ascii("componnent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("subcriptions"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eliminacion"),
                &["elimination"],
            ),
            (dictgen::InsensitiveStr::Ascii("itertions"), &["iterations"]),
            (
                dictgen::InsensitiveStr::Ascii("penetraion"),
                &["penetration"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaate"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("fopr"), &["for"]),
            (dictgen::InsensitiveStr::Ascii("aleniates"), &["alienates"]),
            (
                dictgen::InsensitiveStr::Ascii("consolacion"),
                &["consolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rocessed"),
                &["processed", "recessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("definet"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("invokations"),
                &["invocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cigurets"), &["cigarets"]),
            (dictgen::InsensitiveStr::Ascii("glyh"), &["glyph"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterizations"),
                &["characterizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("squrriel"), &["squirrel"]),
            (dictgen::InsensitiveStr::Ascii("genearic"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("momentos"),
                &["moments", "mementos"],
            ),
            (dictgen::InsensitiveStr::Ascii("follewiwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("retsart"), &["restart"]),
            (
                dictgen::InsensitiveStr::Ascii("consipracy"),
                &["conspiracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corellating"),
                &["correlating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depercation"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interestinly"),
                &["interestingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensitiv"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("essesital"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("herathstone"),
                &["hearthstone"],
            ),
            (dictgen::InsensitiveStr::Ascii("oterwice"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("endpints"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("recocnisable"),
                &["recognisable", "reconcilable"],
            ),
            (dictgen::InsensitiveStr::Ascii("messeged"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("caclucators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("fougth"), &["fought"]),
            (dictgen::InsensitiveStr::Ascii("produceres"), &["produces"]),
            (dictgen::InsensitiveStr::Ascii("reccommend"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarhcically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dragonus"), &["dragons"]),
            (dictgen::InsensitiveStr::Ascii("cotnained"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("foloower"), &["follower"]),
            (
                dictgen::InsensitiveStr::Ascii("elamentaries"),
                &["elementaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("rescourced"), &["resourced"]),
            (dictgen::InsensitiveStr::Ascii("workfows"), &["workflows"]),
            (
                dictgen::InsensitiveStr::Ascii("imperiarist"),
                &["imperialist"],
            ),
            (dictgen::InsensitiveStr::Ascii("erronoeus"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("mentiones"), &["mentions"]),
            (dictgen::InsensitiveStr::Ascii("commannd"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("profressions"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiatie"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reassocition"),
                &["reassociation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidance"),
                &["confidence"],
            ),
            (dictgen::InsensitiveStr::Ascii("extremley"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("symptumatic"),
                &["symptomatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmissione"),
                &["transmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respectuful"),
                &["respectful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminalogy"),
                &["terminology"],
            ),
            (dictgen::InsensitiveStr::Ascii("ususally"), &["usually"]),
            (
                dictgen::InsensitiveStr::Ascii("proviede"),
                &["provide", "provided"],
            ),
            (dictgen::InsensitiveStr::Ascii("standbuy"), &["standby"]),
            (dictgen::InsensitiveStr::Ascii("contatcted"), &["contacted"]),
            (
                dictgen::InsensitiveStr::Ascii("contoles"),
                &["controls", "consoles", "condoles"],
            ),
            (dictgen::InsensitiveStr::Ascii("tymecode"), &["timecode"]),
            (
                dictgen::InsensitiveStr::Ascii("concatented"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("nervious"), &["nervous"]),
            (
                dictgen::InsensitiveStr::Ascii("thereatfer"),
                &["thereafter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conficting"),
                &["conflicting", "convicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentaton"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("errros"), &["errors"]),
            (dictgen::InsensitiveStr::Ascii("dowlink"), &["downlink"]),
            (
                dictgen::InsensitiveStr::Ascii("destrucion"),
                &["destruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("caryons"), &["crayons"]),
            (dictgen::InsensitiveStr::Ascii("chaecks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("authendicators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("mordibly"), &["morbidly"]),
            (dictgen::InsensitiveStr::Ascii("yestarday"), &["yesterday"]),
            (dictgen::InsensitiveStr::Ascii("seprarate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("democray"), &["democracy"]),
            (
                dictgen::InsensitiveStr::Ascii("coexhistance"),
                &["coexistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimnetelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretet"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presentato"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inclue"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("abberivations"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabitilities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("packaed"),
                &["packaged", "packed"],
            ),
            (dictgen::InsensitiveStr::Ascii("momery"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("reprase"), &["reparse"]),
            (dictgen::InsensitiveStr::Ascii("componant"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("camparison"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deploiment"),
                &["deployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempramental"),
                &["temperamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("comitees"), &["committees"]),
            (dictgen::InsensitiveStr::Ascii("asssesses"), &["assesses"]),
            (dictgen::InsensitiveStr::Ascii("bouyed"), &["buoyed"]),
            (
                dictgen::InsensitiveStr::Ascii("assues"),
                &["assumes", "assures"],
            ),
            (dictgen::InsensitiveStr::Ascii("optioins"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("fluorecence"),
                &["fluorescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condicts"),
                &["conducts", "convicts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selecter"),
                &["selector", "selected", "select"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemtary"), &["elementary"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondng"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveniece"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("versiom"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("sociapaths"),
                &["sociopaths"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asignments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("trilineal"), &["trilinear"]),
            (dictgen::InsensitiveStr::Ascii("peninsual"), &["peninsula"]),
            (dictgen::InsensitiveStr::Ascii("cellst"), &["cells"]),
            (
                dictgen::InsensitiveStr::Ascii("clea"),
                &["clean", "clear", "cleat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonomously"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("applays"), &["applies"]),
            (
                dictgen::InsensitiveStr::Ascii("resistanse"),
                &["resistances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catterpillar"),
                &["caterpillar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triathalons"),
                &["triathlons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preemtipble"),
                &["preemptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analtyically"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ampty"), &["empty"]),
            (dictgen::InsensitiveStr::Ascii("unkmown"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("adjascency"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("sporles"), &["sproles"]),
            (dictgen::InsensitiveStr::Ascii("caregory"), &["category"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferentiating"),
                &["differentiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("etrailer"), &["retailer"]),
            (
                dictgen::InsensitiveStr::Ascii("unserstand"),
                &["understand"],
            ),
            (dictgen::InsensitiveStr::Ascii("goverend"), &["governed"]),
            (
                dictgen::InsensitiveStr::Ascii("impelements"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsolited"), &["obsoleted"]),
            (
                dictgen::InsensitiveStr::Ascii("acendence"),
                &["ascendence", "ascendance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximetes"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("laods"), &["loads"]),
            (
                dictgen::InsensitiveStr::Ascii("unintentionnally"),
                &["unintentionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggesstions"),
                &["suggestions"],
            ),
            (dictgen::InsensitiveStr::Ascii("optizmied"), &["optimized"]),
            (dictgen::InsensitiveStr::Ascii("jouneys"), &["journeys"]),
            (
                dictgen::InsensitiveStr::Ascii("bootstapped"),
                &["bootstrapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("clasrooms"), &["classrooms"]),
            (dictgen::InsensitiveStr::Ascii("bestialiy"), &["bestiality"]),
            (dictgen::InsensitiveStr::Ascii("coplete"), &["complete"]),
            (
                dictgen::InsensitiveStr::Ascii("expermenetal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbatium"), &["verbatim"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronizations"),
                &["synchronizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("facist"), &["fascist"]),
            (
                dictgen::InsensitiveStr::Ascii("necesarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("singnature"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("unsetted"),
                &["unset", "unseated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collabaration"),
                &["collaboration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanotrum"), &["santorum"]),
            (
                dictgen::InsensitiveStr::Ascii("ubitquitous"),
                &["ubiquitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directinla"),
                &["directional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admittadely"),
                &["admittedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("offficial"), &["official"]),
            (dictgen::InsensitiveStr::Ascii("wuold"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("druing"), &["during"]),
            (
                dictgen::InsensitiveStr::Ascii("brilliante"),
                &["brilliance"],
            ),
            (dictgen::InsensitiveStr::Ascii("librairy"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("tabacco"), &["tobacco"]),
            (dictgen::InsensitiveStr::Ascii("execunes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("quesr"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("placemenets"),
                &["placements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preprocesser"),
                &["preprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanatiation"),
                &["instantiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrbiute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("warppred"),
                &["warped", "wrapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("prameter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("sourct"), &["source"]),
            (dictgen::InsensitiveStr::Ascii("mustash"), &["mustache"]),
            (dictgen::InsensitiveStr::Ascii("assserted"), &["asserted"]),
            (
                dictgen::InsensitiveStr::Ascii("positiveity"),
                &["positivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relect"),
                &["reelect", "reflect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oultinenodes"),
                &["outlinenodes"],
            ),
            (dictgen::InsensitiveStr::Ascii("horizonts"), &["horizons"]),
            (dictgen::InsensitiveStr::Ascii("stubmled"), &["stumbled"]),
            (dictgen::InsensitiveStr::Ascii("cleantup"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("inherted"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("catholisicm"),
                &["catholicism"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeueue"), &["queue"]),
            (dictgen::InsensitiveStr::Ascii("wehther"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("spsaced"), &["spaced"]),
            (dictgen::InsensitiveStr::Ascii("reoslved"), &["resolved"]),
            (
                dictgen::InsensitiveStr::Ascii("assciation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availablity"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differentiater"),
                &[
                    "differentiator",
                    "differentiated",
                    "differentiates",
                    "differentiate",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("detecion"), &["detection"]),
            (dictgen::InsensitiveStr::Ascii("sutract"), &["subtract"]),
            (dictgen::InsensitiveStr::Ascii("arraived"), &["arrived"]),
            (dictgen::InsensitiveStr::Ascii("apparrel"), &["apparel"]),
            (dictgen::InsensitiveStr::Ascii("presuade"), &["persuade"]),
            (
                dictgen::InsensitiveStr::Ascii("handshacking"),
                &["handshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("visitng"), &["visiting"]),
            (dictgen::InsensitiveStr::Ascii("arthrose"), &["arthrosis"]),
            (dictgen::InsensitiveStr::Ascii("priave"), &["private"]),
            (
                dictgen::InsensitiveStr::Ascii("impersonationg"),
                &["impersonating"],
            ),
            (dictgen::InsensitiveStr::Ascii("countys"), &["counties"]),
            (
                dictgen::InsensitiveStr::Ascii("illistrate"),
                &["illiterate", "illustrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bordelrands"),
                &["borderlands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catergorized"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("regconise"), &["recognise"]),
            (
                dictgen::InsensitiveStr::Ascii("implamentations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("samruai"), &["samurai"]),
            (
                dictgen::InsensitiveStr::Ascii("incombatible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("annonator"), &["annotator"]),
            (dictgen::InsensitiveStr::Ascii("hapen"), &["happen"]),
            (
                dictgen::InsensitiveStr::Ascii("singals"),
                &["signals", "singles"],
            ),
            (dictgen::InsensitiveStr::Ascii("inolving"), &["involving"]),
            (
                dictgen::InsensitiveStr::Ascii("serveritie"),
                &["severity", "severities"],
            ),
            (dictgen::InsensitiveStr::Ascii("popullate"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomeamces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("attirubte"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("atuhenticators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewritting"), &["rewriting"]),
            (
                dictgen::InsensitiveStr::Ascii("concuring"),
                &["concurring", "conquering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restricing"),
                &["restricting"],
            ),
            (dictgen::InsensitiveStr::Ascii("aganist"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("allcoators"),
                &["allocators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheeckpoint"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activater"),
                &[
                    "activator",
                    "actuator",
                    "activated",
                    "activates",
                    "activate",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("reigining"), &["reigning"]),
            (dictgen::InsensitiveStr::Ascii("belgain"), &["belgian"]),
            (
                dictgen::InsensitiveStr::Ascii("develpoing"),
                &["developing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrtuctive"),
                &["destructive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquaintaces"),
                &["acquaintances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nofications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("inhibated"), &["inhibited"]),
            (
                dictgen::InsensitiveStr::Ascii("technicly"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("srceens"), &["screens"]),
            (
                dictgen::InsensitiveStr::Ascii("affirmate"),
                &["affirmative", "affirm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supplimented"),
                &["supplemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documantaries"),
                &["documentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardianl"), &["cardinal"]),
            (
                dictgen::InsensitiveStr::Ascii("terrforming"),
                &["terraforming"],
            ),
            (dictgen::InsensitiveStr::Ascii("tabel"), &["table"]),
            (dictgen::InsensitiveStr::Ascii("declated"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("vritualise"),
                &["virtualise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexpessions"),
                &["subexpressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resposnibility"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incentivare"),
                &["incentive"],
            ),
            (dictgen::InsensitiveStr::Ascii("vas"), &["was"]),
            (dictgen::InsensitiveStr::Ascii("writre"), &["writer"]),
            (
                dictgen::InsensitiveStr::Ascii("forgieners"),
                &["foreigners"],
            ),
            (dictgen::InsensitiveStr::Ascii("denomal"), &["denormal"]),
            (dictgen::InsensitiveStr::Ascii("opose"), &["oppose"]),
            (dictgen::InsensitiveStr::Ascii("existting"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("aotomatically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("curvatrue"), &["curvature"]),
            (
                dictgen::InsensitiveStr::Ascii("elligible"),
                &["eligible", "illegible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("numberically"),
                &["numerically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructicon"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("dran"), &["drawn"]),
            (dictgen::InsensitiveStr::Ascii("milleneum"), &["millennium"]),
            (dictgen::InsensitiveStr::Ascii("procelain"), &["porcelain"]),
            (dictgen::InsensitiveStr::Ascii("intersts"), &["interests"]),
            (dictgen::InsensitiveStr::Ascii("chagne"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("convincin"), &["convincing"]),
            (
                dictgen::InsensitiveStr::Ascii("sctioned"),
                &["sectioned", "suctioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommedations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("probleme"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("challanage"), &["challenge"]),
            (dictgen::InsensitiveStr::Ascii("perfored"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("circulaire"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnosctic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("vibratie"), &["vibrate"]),
            (
                dictgen::InsensitiveStr::Ascii("hypertropy"),
                &["hypertrophy"],
            ),
            (dictgen::InsensitiveStr::Ascii("facor"), &["favor", "faker"]),
            (
                dictgen::InsensitiveStr::Ascii("throttoling"),
                &["throttling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executuons"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenicaed"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("unning"), &["running"]),
            (
                dictgen::InsensitiveStr::Ascii("implenetations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("befoerhand"),
                &["beforehand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptionns"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("siganling"), &["signaling"]),
            (
                dictgen::InsensitiveStr::Ascii("surroundig"),
                &["surroundings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trancating"),
                &["truncating"],
            ),
            (dictgen::InsensitiveStr::Ascii("existying"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("granade"), &["grenade"]),
            (dictgen::InsensitiveStr::Ascii("easthetics"), &["esthetics"]),
            (dictgen::InsensitiveStr::Ascii("burtality"), &["brutality"]),
            (dictgen::InsensitiveStr::Ascii("depresso"), &["depression"]),
            (dictgen::InsensitiveStr::Ascii("omlets"), &["omelets"]),
            (
                dictgen::InsensitiveStr::Ascii("epression"),
                &["expression", "depression", "repression"],
            ),
            (dictgen::InsensitiveStr::Ascii("purhcase"), &["purchase"]),
            (dictgen::InsensitiveStr::Ascii("artices"), &["articles"]),
            (
                dictgen::InsensitiveStr::Ascii("mechamisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insentivie"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lotharingen"),
                &["lothringen"],
            ),
            (dictgen::InsensitiveStr::Ascii("ggogled"), &["googled"]),
            (
                dictgen::InsensitiveStr::Ascii("corectness"),
                &["correctness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practitions"),
                &["practitioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("ptorions"), &["portions"]),
            (
                dictgen::InsensitiveStr::Ascii("costum"),
                &["custom", "costume"],
            ),
            (dictgen::InsensitiveStr::Ascii("xenophoby"), &["xenophobia"]),
            (
                dictgen::InsensitiveStr::Ascii("transparrent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralellising"),
                &["parallelising"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqired"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("destribute"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thraded"),
                &["threaded", "traded"],
            ),
            (dictgen::InsensitiveStr::Ascii("wtihin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("attachmant"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("paylood"), &["payload"]),
            (dictgen::InsensitiveStr::Ascii("algorithme"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("torrentors"), &["torrents"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentella"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("syndrom"), &["syndrome"]),
            (
                dictgen::InsensitiveStr::Ascii("reitrement"),
                &["retirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferrers"), &["referrers"]),
            (
                dictgen::InsensitiveStr::Ascii("competitivos"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmissin"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disemination"),
                &["dissemination"],
            ),
            (dictgen::InsensitiveStr::Ascii("reipient"), &["recipient"]),
            (
                dictgen::InsensitiveStr::Ascii("familiarze"),
                &["familiarize"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislaying"), &["displaying"]),
            (dictgen::InsensitiveStr::Ascii("poduct"), &["product"]),
            (
                dictgen::InsensitiveStr::Ascii("considerstions"),
                &["considerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("continaer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("surpemacist"),
                &["supremacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomodating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("interakt"), &["interact"]),
            (dictgen::InsensitiveStr::Ascii("mandaste"), &["mandates"]),
            (dictgen::InsensitiveStr::Ascii("thniking"), &["thinking"]),
            (dictgen::InsensitiveStr::Ascii("measuing"), &["measuring"]),
            (
                dictgen::InsensitiveStr::Ascii("benevolant"),
                &["benevolent"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbales"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("resteraunts"),
                &["restaurants"],
            ),
            (dictgen::InsensitiveStr::Ascii("espsilon"), &["epsilon"]),
            (
                dictgen::InsensitiveStr::Ascii("dissambly"),
                &["disassembly"],
            ),
            (dictgen::InsensitiveStr::Ascii("buidls"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("pacjage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("handshare"), &["handshake"]),
            (dictgen::InsensitiveStr::Ascii("compilter"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("cirles"), &["circles"]),
            (
                dictgen::InsensitiveStr::Ascii("athletisim"),
                &["athleticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("esimators"), &["estimators"]),
            (
                dictgen::InsensitiveStr::Ascii("presumpteous"),
                &["presumptuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documanted"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("primitves"), &["primitives"]),
            (dictgen::InsensitiveStr::Ascii("embeddded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("checcked"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("soltions"), &["solutions"]),
            (dictgen::InsensitiveStr::Ascii("bufffers"), &["buffers"]),
            (dictgen::InsensitiveStr::Ascii("wisconisn"), &["wisconsin"]),
            (dictgen::InsensitiveStr::Ascii("uncorrect"), &["incorrect"]),
            (dictgen::InsensitiveStr::Ascii("negotible"), &["negotiable"]),
            (
                dictgen::InsensitiveStr::Ascii("calcurated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("critierion"), &["criterion"]),
            (dictgen::InsensitiveStr::Ascii("versin"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacteres"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excesively"),
                &["excessively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememted"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goegraphically"),
                &["geographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convertables"),
                &["convertibles"],
            ),
            (dictgen::InsensitiveStr::Ascii("boudns"), &["bounds"]),
            (dictgen::InsensitiveStr::Ascii("calculted"), &["calculated"]),
            (dictgen::InsensitiveStr::Ascii("analyseles"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("manupulates"),
                &["manipulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumizers"),
                &["customizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reponsibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("anguluar"), &["angular"]),
            (dictgen::InsensitiveStr::Ascii("docuhey"), &["douchey"]),
            (
                dictgen::InsensitiveStr::Ascii("executeables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecure"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proirities"),
                &["priorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presentated"),
                &["presented"],
            ),
            (dictgen::InsensitiveStr::Ascii("inported"), &["imported"]),
            (
                dictgen::InsensitiveStr::Ascii("suspicisons"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("talsiman"), &["talisman"]),
            (
                dictgen::InsensitiveStr::Ascii("carachteristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneation"), &["generation"]),
            (
                dictgen::InsensitiveStr::Ascii("superouman"),
                &["superhuman"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomalizing"),
                &["normalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("additonally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("originales"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("diassociations"),
                &["disassociations", "dissociations"],
            ),
            (dictgen::InsensitiveStr::Ascii("improove"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("unnnecessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomoetry"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("atatch"), &["attach"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiseing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prejiduces"),
                &["prejudices"],
            ),
            (dictgen::InsensitiveStr::Ascii("nesesary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("subscriped"),
                &["subscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termnially"),
                &["terminally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reate"),
                &["create", "relate"],
            ),
            (dictgen::InsensitiveStr::Ascii("stivk"), &["stick"]),
            (dictgen::InsensitiveStr::Ascii("iternates"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("seamlessy"), &["seamlessly"]),
            (dictgen::InsensitiveStr::Ascii("alrightey"), &["alrighty"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregrators"),
                &["aggregators"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqiurtle"), &["squirtle"]),
            (dictgen::InsensitiveStr::Ascii("ingredent"), &["ingredient"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinitelly"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gerates"),
                &["generates", "gyrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conficted"),
                &["conflicted", "convicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chronologial"),
                &["chronological"],
            ),
            (dictgen::InsensitiveStr::Ascii("quartlery"), &["quarterly"]),
            (dictgen::InsensitiveStr::Ascii("acumalate"), &["accumulate"]),
            (dictgen::InsensitiveStr::Ascii("progesses"), &["progresses"]),
            (
                dictgen::InsensitiveStr::Ascii("popoulation"),
                &["population"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automoblies"),
                &["automobile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insufficaint"),
                &["insufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("beaucop"), &["beaucoup"]),
            (dictgen::InsensitiveStr::Ascii("submarie"), &["submarines"]),
            (dictgen::InsensitiveStr::Ascii("resaon"), &["reason"]),
            (dictgen::InsensitiveStr::Ascii("ocurring"), &["occurring"]),
            (
                dictgen::InsensitiveStr::Ascii("suffciently"),
                &["sufficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconveniently"),
                &["inconveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("veiwership"),
                &["viewership"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssertive"), &["assertive"]),
            (
                dictgen::InsensitiveStr::Ascii("calclulates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlaysing"), &["analysing"]),
            (
                dictgen::InsensitiveStr::Ascii("derrivative"),
                &["derivative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professoras"),
                &["professors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structoral"),
                &["structural"],
            ),
            (dictgen::InsensitiveStr::Ascii("litteraly"), &["literally"]),
            (dictgen::InsensitiveStr::Ascii("hardare"), &["hardware"]),
            (dictgen::InsensitiveStr::Ascii("analzyes"), &["analyzes"]),
            (dictgen::InsensitiveStr::Ascii("execuet"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("wokrloads"), &["workloads"]),
            (dictgen::InsensitiveStr::Ascii("requrires"), &["requires"]),
            (dictgen::InsensitiveStr::Ascii("privilidge"), &["privilege"]),
            (
                dictgen::InsensitiveStr::Ascii("adminitrative"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continentes"),
                &["continents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefecth"),
                &["prefetch", "prefect"],
            ),
            (dictgen::InsensitiveStr::Ascii("givveing"), &["giving"]),
            (
                dictgen::InsensitiveStr::Ascii("creaentials"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crystallizin"),
                &["crystallizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compariosn"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplcity"), &["simplicity"]),
            (
                dictgen::InsensitiveStr::Ascii("enterrnal"),
                &["external", "internal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controleurs"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscibed"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("dumspter"), &["dumpster"]),
            (dictgen::InsensitiveStr::Ascii("extreamely"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("execluded"), &["excluded"]),
            (dictgen::InsensitiveStr::Ascii("segmentes"), &["segments"]),
            (
                dictgen::InsensitiveStr::Ascii("encourageing"),
                &["encouraging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efficience"),
                &["efficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("whhen"), &["when"]),
            (dictgen::InsensitiveStr::Ascii("resorece"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("aggessively"),
                &["aggressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequenec"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("axises"), &["axes"]),
            (
                dictgen::InsensitiveStr::Ascii("accpetably"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parametics"),
                &["paramedics"],
            ),
            (dictgen::InsensitiveStr::Ascii("impements"), &["implements"]),
            (dictgen::InsensitiveStr::Ascii("copuling"), &["coupling"]),
            (
                dictgen::InsensitiveStr::Ascii("litterature"),
                &["literature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocassionally"),
                &["occasionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scintillatqt"),
                &["scintillaqt"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimze"), &["minimize"]),
            (dictgen::InsensitiveStr::Ascii("requsite"), &["requisite"]),
            (
                dictgen::InsensitiveStr::Ascii("frigthening"),
                &["frightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("porition"), &["portion"]),
            (dictgen::InsensitiveStr::Ascii("inialize"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("traansforming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restircted"),
                &["restricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("cupborad"), &["cupboard"]),
            (
                dictgen::InsensitiveStr::Ascii("presbateriens"),
                &["presbyterians"],
            ),
            (dictgen::InsensitiveStr::Ascii("clitories"), &["clitoris"]),
            (dictgen::InsensitiveStr::Ascii("shoudlers"), &["shoulders"]),
            (dictgen::InsensitiveStr::Ascii("hax"), &["hex"]),
            (
                dictgen::InsensitiveStr::Ascii("conrruptions"),
                &["corruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mye"), &["may", "my"]),
            (
                dictgen::InsensitiveStr::Ascii("ganeration"),
                &["generation"],
            ),
            (dictgen::InsensitiveStr::Ascii("addnos"), &["addons"]),
            (
                dictgen::InsensitiveStr::Ascii("correspdonence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("advetise"), &["advertise"]),
            (
                dictgen::InsensitiveStr::Ascii("teleportion"),
                &["teleporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("supsicion"), &["suspicion"]),
            (
                dictgen::InsensitiveStr::Ascii("succeessive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("congugate"), &["conjugate"]),
            (dictgen::InsensitiveStr::Ascii("geerate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("coallesceing"),
                &["coalescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configiures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("specualr"), &["specular"]),
            (
                dictgen::InsensitiveStr::Ascii("systamatic"),
                &["systematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infastructures"),
                &["infrastructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comphrensive"),
                &["comprehensive"],
            ),
            (dictgen::InsensitiveStr::Ascii("largst"), &["largest"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilishing"),
                &["establishing", "stabilising"],
            ),
            (dictgen::InsensitiveStr::Ascii("onece"), &["once"]),
            (
                dictgen::InsensitiveStr::Ascii("restirctions"),
                &["restrictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reallocaing"),
                &["reallocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("atfer"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("sytling"), &["styling"]),
            (
                dictgen::InsensitiveStr::Ascii("implausble"),
                &["implausible"],
            ),
            (dictgen::InsensitiveStr::Ascii("acelerate"), &["accelerate"]),
            (dictgen::InsensitiveStr::Ascii("soemthing"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("presbaterians"),
                &["presbyterians"],
            ),
            (dictgen::InsensitiveStr::Ascii("preceede"), &["precede"]),
            (
                dictgen::InsensitiveStr::Ascii("annotaiton"),
                &["annotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("layrinx"), &["larynx"]),
            (
                dictgen::InsensitiveStr::Ascii("mediteranean"),
                &["mediterranean"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretarte"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypcorite"), &["hypocrite"]),
            (
                dictgen::InsensitiveStr::Ascii("democraphic"),
                &["demographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequence"),
                &["frequencies", "frequency"],
            ),
            (dictgen::InsensitiveStr::Ascii("tirggered"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("restaraunt"),
                &["restaurant"],
            ),
            (dictgen::InsensitiveStr::Ascii("coherancy"), &["coherency"]),
            (dictgen::InsensitiveStr::Ascii("afriad"), &["afraid"]),
            (
                dictgen::InsensitiveStr::Ascii("compuatation"),
                &["computation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postgress"),
                &["postgresql", "postgres"],
            ),
            (dictgen::InsensitiveStr::Ascii("laod"), &["load"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminen"),
                &["contaminant", "containment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anniverary"),
                &["anniversary"],
            ),
            (dictgen::InsensitiveStr::Ascii("previusly"), &["previously"]),
            (dictgen::InsensitiveStr::Ascii("alloctors"), &["allocators"]),
            (dictgen::InsensitiveStr::Ascii("specifer"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("captable"), &["capable"]),
            (
                dictgen::InsensitiveStr::Ascii("brilliany"),
                &["brilliantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instititional"),
                &["institutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiasations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("disupted"), &["disputed"]),
            (dictgen::InsensitiveStr::Ascii("collapsers"), &["collapses"]),
            (dictgen::InsensitiveStr::Ascii("littele"), &["little"]),
            (dictgen::InsensitiveStr::Ascii("competin"), &["competing"]),
            (
                dictgen::InsensitiveStr::Ascii("packged"),
                &["packaged", "packed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrispondant"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("cteateing"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("electricrain"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unperturb"),
                &["unperturbed"],
            ),
            (dictgen::InsensitiveStr::Ascii("simiar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("convenince"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatitable"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemtented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("relatative"), &["relative"]),
            (dictgen::InsensitiveStr::Ascii("actiavtes"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("experimmentations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrities"), &["thirties"]),
            (
                dictgen::InsensitiveStr::Ascii("respectabile"),
                &["respectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("instertion"), &["insertion"]),
            (
                dictgen::InsensitiveStr::Ascii("gratuidous"),
                &["gratuitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obliteraded"),
                &["obliterated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaction"),
                &["collection"],
            ),
            (dictgen::InsensitiveStr::Ascii("ariport"), &["airport"]),
            (
                dictgen::InsensitiveStr::Ascii("cilents"),
                &["clients", "silents", "silence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanotory"),
                &["explanatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sorroundings"),
                &["surroundings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confoguring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beleifes"),
                &["beliefs", "believes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monomophized"),
                &["monomorphized"],
            ),
            (dictgen::InsensitiveStr::Ascii("thouhgts"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("convice"),
                &["convince", "convict"],
            ),
            (dictgen::InsensitiveStr::Ascii("commt"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("finsihing"), &["finishing"]),
            (
                dictgen::InsensitiveStr::Ascii("reconecction"),
                &["reconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("propblem"), &["problem"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsdences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availaiability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpoloations"),
                &["interpolations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("konstellation"),
                &["constellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("declinig"), &["declining"]),
            (dictgen::InsensitiveStr::Ascii("methot"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("fontrier"), &["frontier"]),
            (
                dictgen::InsensitiveStr::Ascii("inferfaced"),
                &["interfaced"],
            ),
            (dictgen::InsensitiveStr::Ascii("obervant"), &["observant"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubiotion"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("sintacts"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("cilnatro"), &["cilantro"]),
            (
                dictgen::InsensitiveStr::Ascii("dictioaries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("texchnically"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("restord"), &["restored"]),
            (dictgen::InsensitiveStr::Ascii("libaries"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("deprivating"),
                &["deprivation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintialzied"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("resorcing"), &["resourcing"]),
            (
                dictgen::InsensitiveStr::Ascii("brillaince"),
                &["brilliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunatly"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crache"),
                &["cache", "crèche"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrepsect"),
                &["disrespect"],
            ),
            (dictgen::InsensitiveStr::Ascii("somehtin"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("valkues"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("soufces"), &["sources"]),
            (
                dictgen::InsensitiveStr::Ascii("independend"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("knoe"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("consitutional"),
                &["constitutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("ordert"), &["ordered"]),
            (dictgen::InsensitiveStr::Ascii("accrose"), &["across"]),
            (dictgen::InsensitiveStr::Ascii("cirumflex"), &["circumflex"]),
            (
                dictgen::InsensitiveStr::Ascii("comparation"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhancment"),
                &["enhancement"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleints"), &["clients"]),
            (
                dictgen::InsensitiveStr::Ascii("apreciates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("milestons"), &["milestones"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequisties"),
                &["prerequisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precedeces"),
                &["precedes", "precedents"],
            ),
            (dictgen::InsensitiveStr::Ascii("biulding"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("nastersium"),
                &["nasturtium"],
            ),
            (dictgen::InsensitiveStr::Ascii("htis"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("depandant"), &["dependant"]),
            (
                dictgen::InsensitiveStr::Ascii("continutiy"),
                &["continuity"],
            ),
            (dictgen::InsensitiveStr::Ascii("differect"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controvertial"),
                &["controversial"],
            ),
            (dictgen::InsensitiveStr::Ascii("techniks"), &["techniques"]),
            (
                dictgen::InsensitiveStr::Ascii("trings"),
                &["strings", "rings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conplications"),
                &["complications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archietects"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successfulln"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcievers"),
                &["transceivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounbdary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("probilistic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beanchmarks"),
                &["benchmarks"],
            ),
            (dictgen::InsensitiveStr::Ascii("neares"), &["nearest"]),
            (
                dictgen::InsensitiveStr::Ascii("integraton"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("artificiel"),
                &["artificial"],
            ),
            (dictgen::InsensitiveStr::Ascii("summore"), &["summoner"]),
            (dictgen::InsensitiveStr::Ascii("perosnal"), &["personal"]),
            (dictgen::InsensitiveStr::Ascii("reachged"), &["reached"]),
            (
                dictgen::InsensitiveStr::Ascii("expeirimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exchagnging"),
                &["exchanging"],
            ),
            (dictgen::InsensitiveStr::Ascii("exipre"), &["expire"]),
            (
                dictgen::InsensitiveStr::Ascii("automodertor"),
                &["automoderator"],
            ),
            (dictgen::InsensitiveStr::Ascii("californa"), &["california"]),
            (dictgen::InsensitiveStr::Ascii("dirrecting"), &["directing"]),
            (dictgen::InsensitiveStr::Ascii("orlenas"), &["orleans"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiaze"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availaility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("hurdels"), &["hurdles"]),
            (dictgen::InsensitiveStr::Ascii("alinged"), &["alined"]),
            (
                dictgen::InsensitiveStr::Ascii("attrubutes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("addrersses"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("potition"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("materalists"),
                &["materialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breakthrouhg"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("respwans"), &["respawns"]),
            (dictgen::InsensitiveStr::Ascii("aincent"), &["ancient"]),
            (
                dictgen::InsensitiveStr::Ascii("rimanissed"),
                &["reminisced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitfied"),
                &["identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooridnate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimam"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("pefore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("therre"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("compelation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("disble"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("distriminator"),
                &["discriminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("poduce"), &["produce"]),
            (dictgen::InsensitiveStr::Ascii("methid"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("carimonial"),
                &["ceremonial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersectons"),
                &["intersections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplishemnt"),
                &["accomplishment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimiliarity"),
                &["dissimilarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("mananging"), &["managing"]),
            (dictgen::InsensitiveStr::Ascii("stutied"), &["studied"]),
            (
                dictgen::InsensitiveStr::Ascii("differntials"),
                &["differentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("bassis"), &["basis"]),
            (dictgen::InsensitiveStr::Ascii("invovle"), &["involve"]),
            (dictgen::InsensitiveStr::Ascii("genitala"), &["genitalia"]),
            (dictgen::InsensitiveStr::Ascii("afte"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("aminosity"), &["animosity"]),
            (
                dictgen::InsensitiveStr::Ascii("dalled"),
                &["called", "dolled", "dallied"],
            ),
            (dictgen::InsensitiveStr::Ascii("additioal"), &["additional"]),
            (
                dictgen::InsensitiveStr::Ascii("effeciently"),
                &["efficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("annnouncer"), &["announcer"]),
            (
                dictgen::InsensitiveStr::Ascii("previousely"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("losting"),
                &["losing", "listing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanetiser"), &["sanitiser"]),
            (
                dictgen::InsensitiveStr::Ascii("oversubscribin"),
                &["oversubscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependdencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("realtes"), &["relates"]),
            (dictgen::InsensitiveStr::Ascii("corruput"), &["corrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("horitzonal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("ehanced"), &["enhanced"]),
            (dictgen::InsensitiveStr::Ascii("hermertic"), &["hermetic"]),
            (
                dictgen::InsensitiveStr::Ascii("resroucing"),
                &["resourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restaurents"),
                &["restaurants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decorataions"),
                &["decorations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsless"),
                &["unless", "useless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crystallisin"),
                &["crystallising"],
            ),
            (dictgen::InsensitiveStr::Ascii("usiage"), &["usage"]),
            (
                dictgen::InsensitiveStr::Ascii("algorwwithm"),
                &["algorithm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properyy"),
                &["property", "properly"],
            ),
            (dictgen::InsensitiveStr::Ascii("highjack"), &["hijack"]),
            (dictgen::InsensitiveStr::Ascii("paramer"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("counterfiet"),
                &["counterfeit"],
            ),
            (dictgen::InsensitiveStr::Ascii("definete"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("ecounters"), &["encounters"]),
            (dictgen::InsensitiveStr::Ascii("reasnable"), &["reasonable"]),
            (
                dictgen::InsensitiveStr::Ascii("indentifes"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genaration"),
                &["generation"],
            ),
            (dictgen::InsensitiveStr::Ascii("depoying"), &["deploying"]),
            (
                dictgen::InsensitiveStr::Ascii("electrolye"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("diretives"), &["directives"]),
            (
                dictgen::InsensitiveStr::Ascii("lightweigh"),
                &["lightweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("conerning"), &["concerning"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeorhod"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("looopback"), &["loopback"]),
            (dictgen::InsensitiveStr::Ascii("methoden"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("socities"), &["societies"]),
            (
                dictgen::InsensitiveStr::Ascii("adolencence"),
                &["adolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("registors"), &["registers"]),
            (dictgen::InsensitiveStr::Ascii("socres"), &["scores"]),
            (
                dictgen::InsensitiveStr::Ascii("allocatioin"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("glithcy"), &["glitchy"]),
            (
                dictgen::InsensitiveStr::Ascii("probbaility"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etend"),
                &["extend", "attend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilplied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("petterns"), &["patterns"]),
            (dictgen::InsensitiveStr::Ascii("userpsace"), &["userspace"]),
            (dictgen::InsensitiveStr::Ascii("misue"), &["misuse"]),
            (dictgen::InsensitiveStr::Ascii("excecuted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("austensibly"),
                &["ostensibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("winge"),
                &[
                    "wing", "wings", "winger", "winged", "whine", "twinge", "singe", "wince",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("imobile"), &["immobile"]),
            (
                dictgen::InsensitiveStr::Ascii("enterprsie"),
                &["enterprises"],
            ),
            (dictgen::InsensitiveStr::Ascii("costruct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("wrokbooks"), &["workbooks"]),
            (
                dictgen::InsensitiveStr::Ascii("republician"),
                &["republican"],
            ),
            (dictgen::InsensitiveStr::Ascii("oberflow"), &["overflow"]),
            (
                dictgen::InsensitiveStr::Ascii("mophologies"),
                &["morphologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("caracter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("programemer"),
                &["programmer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangning"),
                &["rearranging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreacating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("problme"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("perpares"), &["prepares"]),
            (dictgen::InsensitiveStr::Ascii("thouse"), &["those"]),
            (dictgen::InsensitiveStr::Ascii("improted"), &["imported"]),
            (dictgen::InsensitiveStr::Ascii("cygin"), &["cygwin"]),
            (dictgen::InsensitiveStr::Ascii("problaems"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("fomratter"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("insturctions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlabled"), &["unlabeled"]),
            (
                dictgen::InsensitiveStr::Ascii("likelyhood"),
                &["likelihood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ejacualtions"),
                &["ejaculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derogotary"),
                &["derogatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentratons"),
                &["concentrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("laout"), &["layout"]),
            (dictgen::InsensitiveStr::Ascii("lanauage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("sepereting"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("infinitey"), &["infinity"]),
            (dictgen::InsensitiveStr::Ascii("commandoes"), &["commandos"]),
            (dictgen::InsensitiveStr::Ascii("deamanding"), &["demanding"]),
            (dictgen::InsensitiveStr::Ascii("adressing"), &["addressing"]),
            (dictgen::InsensitiveStr::Ascii("possesss"), &["possesses"]),
            (
                dictgen::InsensitiveStr::Ascii("suplemental"),
                &["supplemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticipato"),
                &["anticipation", "anticipated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volumn"),
                &["volume", "column"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeender"), &["appender"]),
            (
                dictgen::InsensitiveStr::Ascii("subraction"),
                &["subtraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibitted"),
                &["prohibited"],
            ),
            (dictgen::InsensitiveStr::Ascii("varities"), &["varieties"]),
            (dictgen::InsensitiveStr::Ascii("architet"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("looop"), &["loop"]),
            (
                dictgen::InsensitiveStr::Ascii("plases"),
                &["places", "pleases", "phases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("airplance"),
                &["airplane", "airplanes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rquirements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("finisches"), &["finishes"]),
            (
                dictgen::InsensitiveStr::Ascii("paramormal"),
                &["paranormal"],
            ),
            (dictgen::InsensitiveStr::Ascii("sholder"), &["shoulder"]),
            (dictgen::InsensitiveStr::Ascii("optimzier"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("tallents"), &["tallest"]),
            (dictgen::InsensitiveStr::Ascii("shiiped"), &["shipped"]),
            (
                dictgen::InsensitiveStr::Ascii("artificiella"),
                &["artificial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpitrators"),
                &["perpetrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscritpions"),
                &["unsubscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parilamentary"),
                &["parliamentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conpensating"),
                &["compensating"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimise"), &["optimise"]),
            (dictgen::InsensitiveStr::Ascii("tillted"), &["tilted"]),
            (
                dictgen::InsensitiveStr::Ascii("unitialise"),
                &["uninitialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fabriacted"),
                &["fabricated"],
            ),
            (dictgen::InsensitiveStr::Ascii("seetings"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("delimitiaions"),
                &["delimitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("benetif"), &["benefit"]),
            (dictgen::InsensitiveStr::Ascii("charcater"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("colubmia"), &["columbia"]),
            (dictgen::InsensitiveStr::Ascii("corrcting"), &["correcting"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("distnctes"), &["distances"]),
            (dictgen::InsensitiveStr::Ascii("toos"), &["tools"]),
            (
                dictgen::InsensitiveStr::Ascii("attrubites"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptonally"),
                &["exceptionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("natioanlistic"),
                &["nationalistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("fullsceen"), &["fullscreen"]),
            (dictgen::InsensitiveStr::Ascii("visuaises"), &["visualises"]),
            (dictgen::InsensitiveStr::Ascii("returnng"), &["returning"]),
            (
                dictgen::InsensitiveStr::Ascii("deregulatin"),
                &["deregulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlimeted"), &["unlimited"]),
            (dictgen::InsensitiveStr::Ascii("explitic"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("funnilly"), &["funnily"]),
            (dictgen::InsensitiveStr::Ascii("metata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("defered"), &["deferred"]),
            (dictgen::InsensitiveStr::Ascii("pomised"), &["promised"]),
            (dictgen::InsensitiveStr::Ascii("briuser"), &["bruiser"]),
            (dictgen::InsensitiveStr::Ascii("hallowen"), &["halloween"]),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheat"),
                &["spreadsheet"],
            ),
            (dictgen::InsensitiveStr::Ascii("randmom"), &["random"]),
            (dictgen::InsensitiveStr::Ascii("rememored"), &["remembered"]),
            (dictgen::InsensitiveStr::Ascii("assighees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("inofrmation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scriptores"),
                &["scriptures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrouned"),
                &["surround", "surrounded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deallocaed"),
                &["deallocated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificies"),
                &["specifics", "specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnangelog"), &["changelog"]),
            (
                dictgen::InsensitiveStr::Ascii("beleifable"),
                &["believable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analitycally"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsfucate"), &["obfuscate"]),
            (dictgen::InsensitiveStr::Ascii("imolicit"), &["implicit"]),
            (
                dictgen::InsensitiveStr::Ascii("distirbutes"),
                &["distributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proleteriat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("recations"), &["creations"]),
            (
                dictgen::InsensitiveStr::Ascii("excitiment"),
                &["excitement"],
            ),
            (dictgen::InsensitiveStr::Ascii("coordiate"), &["coordinate"]),
            (dictgen::InsensitiveStr::Ascii("redisign"), &["redesign"]),
            (dictgen::InsensitiveStr::Ascii("browisng"), &["browsing"]),
            (dictgen::InsensitiveStr::Ascii("actial"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("osterage"), &["ostrich"]),
            (
                dictgen::InsensitiveStr::Ascii("repesentational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("musuems"), &["museums"]),
            (dictgen::InsensitiveStr::Ascii("exisit"), &["exist"]),
            (dictgen::InsensitiveStr::Ascii("jjust"), &["just"]),
            (
                dictgen::InsensitiveStr::Ascii("clairfying"),
                &["clarifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("ankshiuos"), &["anxious"]),
            (
                dictgen::InsensitiveStr::Ascii("disobediance"),
                &["disobedience"],
            ),
            (dictgen::InsensitiveStr::Ascii("sansitizer"), &["sanitizer"]),
            (dictgen::InsensitiveStr::Ascii("unweildy"), &["unwieldy"]),
            (
                dictgen::InsensitiveStr::Ascii("alternavte"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("climateers"), &["climates"]),
            (dictgen::InsensitiveStr::Ascii("extraxts"), &["extracts"]),
            (dictgen::InsensitiveStr::Ascii("mageta"), &["magenta"]),
            (dictgen::InsensitiveStr::Ascii("regisrtry"), &["registry"]),
            (
                dictgen::InsensitiveStr::Ascii("sourcde"),
                &["sourced", "source"],
            ),
            (dictgen::InsensitiveStr::Ascii("terrable"), &["terrible"]),
            (
                dictgen::InsensitiveStr::Ascii("gaurentees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulitples"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("desintaiton"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("frecuency"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("devellopment"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("planeswaler"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploygon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("parralels"), &["parallels"]),
            (dictgen::InsensitiveStr::Ascii("straightn"), &["straighten"]),
            (dictgen::InsensitiveStr::Ascii("deciemal"), &["decimal"]),
            (dictgen::InsensitiveStr::Ascii("agreeed"), &["agreed"]),
            (
                dictgen::InsensitiveStr::Ascii("miscelleneous"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("sonetime"), &["sometime"]),
            (
                dictgen::InsensitiveStr::Ascii("excessivley"),
                &["excessively"],
            ),
            (dictgen::InsensitiveStr::Ascii("capapcitor"), &["capacitor"]),
            (
                dictgen::InsensitiveStr::Ascii("kickstartr"),
                &["kickstarter"],
            ),
            (dictgen::InsensitiveStr::Ascii("settlment"), &["settlement"]),
            (
                dictgen::InsensitiveStr::Ascii("transportion"),
                &["transporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsiziding"),
                &["subsidizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammounted"), &["amounted"]),
            (
                dictgen::InsensitiveStr::Ascii("iteratin"),
                &["iterating", "iteration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obversation"),
                &["observation"],
            ),
            (dictgen::InsensitiveStr::Ascii("generaion"), &["generation"]),
            (dictgen::InsensitiveStr::Ascii("htiboxes"), &["hitboxes"]),
            (
                dictgen::InsensitiveStr::Ascii("threatning"),
                &["threatening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacter"),
                &["manufacture", "manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouldore"), &["boulder"]),
            (dictgen::InsensitiveStr::Ascii("thoes"), &["those"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhaven"),
                &["copenhagen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadventently"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proccesing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("gimmickly"), &["gimmicky"]),
            (
                dictgen::InsensitiveStr::Ascii("efetivity"),
                &["effectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolodated"),
                &["consolidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("thgat"), &["that"]),
            (
                dictgen::InsensitiveStr::Ascii("counterprodutive"),
                &["counterproductive"],
            ),
            (dictgen::InsensitiveStr::Ascii("finisched"), &["finished"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegoatiations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("idential"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("breadcumbs"),
                &["breadcrumbs"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("otpimised"), &["optimised"]),
            (
                dictgen::InsensitiveStr::Ascii("acocuntant"),
                &["accountant"],
            ),
            (dictgen::InsensitiveStr::Ascii("burtst"), &["burst"]),
            (dictgen::InsensitiveStr::Ascii("tarpolin"), &["tarpaulin"]),
            (dictgen::InsensitiveStr::Ascii("oversved"), &["observed"]),
            (dictgen::InsensitiveStr::Ascii("gamemdoe"), &["gamemode"]),
            (dictgen::InsensitiveStr::Ascii("ultimely"), &["ultimately"]),
            (dictgen::InsensitiveStr::Ascii("naviagted"), &["navigated"]),
            (dictgen::InsensitiveStr::Ascii("todya"), &["today"]),
            (dictgen::InsensitiveStr::Ascii("reuseing"), &["reusing"]),
            (dictgen::InsensitiveStr::Ascii("stortage"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("ttesting"),
                &["test", "attesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("neslave"), &["enslave"]),
            (
                dictgen::InsensitiveStr::Ascii("consentrated"),
                &["concentrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pypass"), &["bypass"]),
            (dictgen::InsensitiveStr::Ascii("assingning"), &["assigning"]),
            (
                dictgen::InsensitiveStr::Ascii("quarterblack"),
                &["quarterback"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlege"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rimanissing"),
                &["reminiscing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sysytem"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("heatbeat"), &["heartbeat"]),
            (
                dictgen::InsensitiveStr::Ascii("subractions"),
                &["subtractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("virament"), &["vraiment"]),
            (dictgen::InsensitiveStr::Ascii("explorare"), &["explorer"]),
            (
                dictgen::InsensitiveStr::Ascii("abondoning"),
                &["abandoning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encryptation"),
                &["encryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recovation"),
                &["revocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("topoligies"),
                &["topologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improbments"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("pararaphs"), &["paragraphs"]),
            (dictgen::InsensitiveStr::Ascii("migrans"), &["migrants"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimiliarly"),
                &["dissimilarly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapoonted"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attmepting"),
                &["attempting"],
            ),
            (dictgen::InsensitiveStr::Ascii("poject"), &["project"]),
            (dictgen::InsensitiveStr::Ascii("tremendos"), &["tremendous"]),
            (dictgen::InsensitiveStr::Ascii("ahd"), &["had", "and"]),
            (dictgen::InsensitiveStr::Ascii("oprimize"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatieves"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("szenario"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("connecotrs"),
                &["connectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("landins"), &["landings"]),
            (
                dictgen::InsensitiveStr::Ascii("comperators"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("shs"), &["nhs", "ssh"]),
            (
                dictgen::InsensitiveStr::Ascii("temparature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("repsonds"), &["responds"]),
            (
                dictgen::InsensitiveStr::Ascii("persuasian"),
                &["persuasion"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprience"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("destuction"),
                &["destruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertended"), &["pretended"]),
            (
                dictgen::InsensitiveStr::Ascii("promocional"),
                &["promotional"],
            ),
            (dictgen::InsensitiveStr::Ascii("gitar"), &["guitar"]),
            (
                dictgen::InsensitiveStr::Ascii("sympathtic"),
                &["sympathetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("acoutsic"), &["acoustic"]),
            (
                dictgen::InsensitiveStr::Ascii("experiements"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("remvoves"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("selecgt"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("trhe"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("bayge"), &["beige"]),
            (dictgen::InsensitiveStr::Ascii("chihuahuha"), &["chihuahua"]),
            (dictgen::InsensitiveStr::Ascii("forword"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("refartoring"),
                &["refactoring", "refractoring"],
            ),
            (dictgen::InsensitiveStr::Ascii("iimmune"), &["immune"]),
            (dictgen::InsensitiveStr::Ascii("accsepters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("rendereer"), &["renderer"]),
            (
                dictgen::InsensitiveStr::Ascii("extruciating"),
                &["excruciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("inavalid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlsated"),
                &["translated"],
            ),
            (dictgen::InsensitiveStr::Ascii("oustider"), &["outsider"]),
            (
                dictgen::InsensitiveStr::Ascii("prgressive"),
                &["progressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("herarchical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaint"), &["defiant"]),
            (
                dictgen::InsensitiveStr::Ascii("thirs"),
                &["third", "thirst"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reisntalled"),
                &["reinstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succesfullly"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("feezers"), &["freezers"]),
            (
                dictgen::InsensitiveStr::Ascii("importence"),
                &["importance", "impotence"],
            ),
            (dictgen::InsensitiveStr::Ascii("splited"), &["split"]),
            (
                dictgen::InsensitiveStr::Ascii("beastiaries"),
                &["bestiaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("allwoed"), &["allowed"]),
            (
                dictgen::InsensitiveStr::Ascii("conecutive"),
                &["consecutive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversetimating"),
                &["overestimating"],
            ),
            (dictgen::InsensitiveStr::Ascii("generattes"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("finnsih"), &["finnish"]),
            (
                dictgen::InsensitiveStr::Ascii("irresponsibe"),
                &["irresponsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("crusiing"), &["cruising"]),
            (dictgen::InsensitiveStr::Ascii("seprately"), &["separately"]),
            (
                dictgen::InsensitiveStr::Ascii("artihmetic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoxeric"), &["anorexic"]),
            (dictgen::InsensitiveStr::Ascii("weppons"), &["weapons"]),
            (dictgen::InsensitiveStr::Ascii("bookmars"), &["bookmarks"]),
            (dictgen::InsensitiveStr::Ascii("regads"), &["regards"]),
            (
                dictgen::InsensitiveStr::Ascii("unintesting"),
                &["uninteresting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unforgivible"),
                &["unforgivable"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmet"), &["comment"]),
            (
                dictgen::InsensitiveStr::Ascii("announceing"),
                &["announcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reqistration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("puprose"), &["purpose"]),
            (dictgen::InsensitiveStr::Ascii("archtiect"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("recntly"), &["recently"]),
            (
                dictgen::InsensitiveStr::Ascii("paceholders"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberatrian"),
                &["libertarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheudled"), &["scheduled"]),
            (dictgen::InsensitiveStr::Ascii("melineums"), &["millennia"]),
            (dictgen::InsensitiveStr::Ascii("inlcusive"), &["inclusive"]),
            (dictgen::InsensitiveStr::Ascii("disappar"), &["disappear"]),
            (
                dictgen::InsensitiveStr::Ascii("getegories"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("siguret"), &["cigaret"]),
            (dictgen::InsensitiveStr::Ascii("agruement"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("charasteristic"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misanderstood"),
                &["misunderstood"],
            ),
            (dictgen::InsensitiveStr::Ascii("treatis"), &["treaties"]),
            (dictgen::InsensitiveStr::Ascii("refreshr"), &["refresher"]),
            (
                dictgen::InsensitiveStr::Ascii("surroundign"),
                &["surroundings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unpredicatable"),
                &["unpredictable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecttion"),
                &["connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("desolve"), &["dissolve"]),
            (dictgen::InsensitiveStr::Ascii("routien"), &["routine"]),
            (
                dictgen::InsensitiveStr::Ascii("utiliterian"),
                &["utilitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("arround"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("enoughts"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("cipersuites"),
                &["ciphersuites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdenence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuileoga"), &["cuileog"]),
            (dictgen::InsensitiveStr::Ascii("libararies"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("wreslte"), &["wrestle"]),
            (
                dictgen::InsensitiveStr::Ascii("rererence"),
                &["reference", "reverence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoizing"),
                &["authorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragpraph"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("efines"),
                &["defines", "refines"],
            ),
            (dictgen::InsensitiveStr::Ascii("slavage"), &["salvage"]),
            (dictgen::InsensitiveStr::Ascii("nunbers"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("inpacts"), &["impacts"]),
            (
                dictgen::InsensitiveStr::Ascii("confiurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempatures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("fliters"), &["filters"]),
            (
                dictgen::InsensitiveStr::Ascii("buisinesses"),
                &["businesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victoriosa"),
                &["victorious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disabmiguate"),
                &["disambiguate"],
            ),
            (dictgen::InsensitiveStr::Ascii("devolops"), &["develops"]),
            (dictgen::InsensitiveStr::Ascii("attachen"), &["attach"]),
            (dictgen::InsensitiveStr::Ascii("anthings"), &["anythings"]),
            (
                dictgen::InsensitiveStr::Ascii("successfui"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondgly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proccessors"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helpes"),
                &["helps", "helpers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregrated"),
                &["aggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cursro"), &["cursor"]),
            (dictgen::InsensitiveStr::Ascii("porsalin"), &["porcelain"]),
            (
                dictgen::InsensitiveStr::Ascii("carmonially"),
                &["ceremonially"],
            ),
            (dictgen::InsensitiveStr::Ascii("comptuer"), &["computer"]),
            (dictgen::InsensitiveStr::Ascii("pris"), &["prise", "prism"]),
            (
                dictgen::InsensitiveStr::Ascii("deteriate"),
                &["deteriorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmations"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("accssor"), &["accessor"]),
            (dictgen::InsensitiveStr::Ascii("varients"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("fraudolent"),
                &["fraudulent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abnoramlly"),
                &["abnormally"],
            ),
            (dictgen::InsensitiveStr::Ascii("assuma"), &["assume"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstential"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenatelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtarcts"), &["subtracts"]),
            (
                dictgen::InsensitiveStr::Ascii("proyecting"),
                &["projecting", "protecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("apealed"), &["appealed"]),
            (dictgen::InsensitiveStr::Ascii("passswords"), &["passwords"]),
            (
                dictgen::InsensitiveStr::Ascii("rightiousness"),
                &["righteousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("resovling"), &["resolving"]),
            (dictgen::InsensitiveStr::Ascii("qulaities"), &["qualities"]),
            (dictgen::InsensitiveStr::Ascii("cheker"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("calncels"), &["cancels"]),
            (
                dictgen::InsensitiveStr::Ascii("autokorrect"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("futurers"), &["futures"]),
            (dictgen::InsensitiveStr::Ascii("grenads"), &["grenades"]),
            (
                dictgen::InsensitiveStr::Ascii("hallcuination"),
                &["hallucination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desparate"),
                &["desperate", "disparate"],
            ),
            (dictgen::InsensitiveStr::Ascii("camelion"), &["chameleon"]),
            (dictgen::InsensitiveStr::Ascii("pickniced"), &["picnicked"]),
            (
                dictgen::InsensitiveStr::Ascii("powerlfiting"),
                &["powerlifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("zise"), &["size"]),
            (dictgen::InsensitiveStr::Ascii("malclom"), &["malcolm"]),
            (dictgen::InsensitiveStr::Ascii("maintan"), &["maintain"]),
            (
                dictgen::InsensitiveStr::Ascii("formualtion"),
                &["formulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("promtion"), &["promotion"]),
            (dictgen::InsensitiveStr::Ascii("separors"), &["separators"]),
            (dictgen::InsensitiveStr::Ascii("destroi"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("thighten"), &["tighten"]),
            (dictgen::InsensitiveStr::Ascii("inforce"), &["enforce"]),
            (dictgen::InsensitiveStr::Ascii("contrete"), &["concrete"]),
            (dictgen::InsensitiveStr::Ascii("exata"), &["exacta"]),
            (
                dictgen::InsensitiveStr::Ascii("stuations"),
                &["situations", "stations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptionnalism"),
                &["exceptionalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singuarity"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("perturbate"), &["perturb"]),
            (dictgen::InsensitiveStr::Ascii("notmutch"), &["notmuch"]),
            (dictgen::InsensitiveStr::Ascii("indicatie"), &["indicative"]),
            (
                dictgen::InsensitiveStr::Ascii("follwowiong"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomeing"),
                &["performing"],
            ),
            (dictgen::InsensitiveStr::Ascii("compleats"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("contignent"),
                &["contingent"],
            ),
            (dictgen::InsensitiveStr::Ascii("compleete"), &["complete"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretion"),
                &["misinterpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("orginized"), &["organized"]),
            (
                dictgen::InsensitiveStr::Ascii("transparaently"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuped"), &["occupied"]),
            (dictgen::InsensitiveStr::Ascii("hesiate"), &["hesitate"]),
            (dictgen::InsensitiveStr::Ascii("thourogh"), &["thorough"]),
            (dictgen::InsensitiveStr::Ascii("scrambeld"), &["scrambled"]),
            (dictgen::InsensitiveStr::Ascii("spashed"), &["splashed"]),
            (dictgen::InsensitiveStr::Ascii("moneky"), &["monkey"]),
            (
                dictgen::InsensitiveStr::Ascii("horizintal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("timestmps"), &["timestamps"]),
            (dictgen::InsensitiveStr::Ascii("exposting"), &["exposing"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tensionors"), &["tensions"]),
            (dictgen::InsensitiveStr::Ascii("oridinates"), &["ordinates"]),
            (
                dictgen::InsensitiveStr::Ascii("picturesk"),
                &["picturesque"],
            ),
            (dictgen::InsensitiveStr::Ascii("buiilder"), &["builder"]),
            (
                dictgen::InsensitiveStr::Ascii("recomputin"),
                &["recomputing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mintutes"), &["minutes"]),
            (dictgen::InsensitiveStr::Ascii("peopel"), &["people"]),
            (dictgen::InsensitiveStr::Ascii("sympathie"), &["sympathize"]),
            (dictgen::InsensitiveStr::Ascii("proposels"), &["proposes"]),
            (dictgen::InsensitiveStr::Ascii("uptadeable"), &["updatable"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundarys"), &["boundaries"]),
            (dictgen::InsensitiveStr::Ascii("bouning"), &["bounding"]),
            (
                dictgen::InsensitiveStr::Ascii("individiual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("relavence"), &["relevance"]),
            (dictgen::InsensitiveStr::Ascii("seledt"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("tornadoe"), &["tornado"]),
            (dictgen::InsensitiveStr::Ascii("merchat"), &["merchant"]),
            (
                dictgen::InsensitiveStr::Ascii("vitualise"),
                &["virtualise", "visualise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wincheseter"),
                &["winchester"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrupible"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commandore"),
                &["commanders"],
            ),
            (dictgen::InsensitiveStr::Ascii("reveales"), &["reveals"]),
            (
                dictgen::InsensitiveStr::Ascii("plagierism"),
                &["plagiarism"],
            ),
            (dictgen::InsensitiveStr::Ascii("enxt"), &["next"]),
            (
                dictgen::InsensitiveStr::Ascii("incraments"),
                &["increments"],
            ),
            (dictgen::InsensitiveStr::Ascii("asistance"), &["assistance"]),
            (dictgen::InsensitiveStr::Ascii("uderstand"), &["understand"]),
            (dictgen::InsensitiveStr::Ascii("copute"), &["compute"]),
            (
                dictgen::InsensitiveStr::Ascii("insentive"),
                &["incentive", "insensitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("garphed"), &["graphed"]),
            (dictgen::InsensitiveStr::Ascii("asai"), &["asia"]),
            (
                dictgen::InsensitiveStr::Ascii("breakpoing"),
                &["breakpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("hardwod"), &["hardwood"]),
            (
                dictgen::InsensitiveStr::Ascii("blackberrry"),
                &["blackberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("seriuos"), &["serious"]),
            (
                dictgen::InsensitiveStr::Ascii("implementtaions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalnce"),
                &["equivalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distibution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("poining"), &["pointing"]),
            (dictgen::InsensitiveStr::Ascii("dyanmics"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("illustrater"),
                &["illustrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docuementation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interationally"),
                &["internationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterpolate"),
                &["interpolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verisions"),
                &["revisions", "versions"],
            ),
            (dictgen::InsensitiveStr::Ascii("obession"), &["obsession"]),
            (dictgen::InsensitiveStr::Ascii("armeniam"), &["armenian"]),
            (dictgen::InsensitiveStr::Ascii("backdooor"), &["backdoor"]),
            (dictgen::InsensitiveStr::Ascii("suppurter"), &["supporter"]),
            (dictgen::InsensitiveStr::Ascii("satisfing"), &["satisfying"]),
            (dictgen::InsensitiveStr::Ascii("pantomine"), &["pantomime"]),
            (dictgen::InsensitiveStr::Ascii("unmanged"), &["unmanaged"]),
            (dictgen::InsensitiveStr::Ascii("coulndt"), &["couldnt"]),
            (dictgen::InsensitiveStr::Ascii("closse"), &["close"]),
            (
                dictgen::InsensitiveStr::Ascii("estabilishments"),
                &["establishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accordindly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretationen"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lightheared"),
                &["lighthearted"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutatbly"), &["mutably"]),
            (
                dictgen::InsensitiveStr::Ascii("defferent"),
                &["different", "deferent"],
            ),
            (dictgen::InsensitiveStr::Ascii("awared"), &["awarded"]),
            (
                dictgen::InsensitiveStr::Ascii("neigbourhood"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("soundard"), &["soundcard"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpert"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperhical"), &["spherical"]),
            (
                dictgen::InsensitiveStr::Ascii("acctivations"),
                &["activations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissconnects"),
                &["disconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifites"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarassimg"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("needels"), &["needles"]),
            (dictgen::InsensitiveStr::Ascii("udates"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("finde"), &["find"]),
            (dictgen::InsensitiveStr::Ascii("insitute"), &["institute"]),
            (dictgen::InsensitiveStr::Ascii("alegance"), &["allegiance"]),
            (
                dictgen::InsensitiveStr::Ascii("cammanders"),
                &["commanders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("areodynamics"),
                &["aerodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recrutiment"),
                &["recruitment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contitionally"),
                &["conditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("ect"), &["etc"]),
            (
                dictgen::InsensitiveStr::Ascii("comparitave"),
                &["comparative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenderlas"),
                &["defenders"],
            ),
            (dictgen::InsensitiveStr::Ascii("tortialls"), &["tortillas"]),
            (dictgen::InsensitiveStr::Ascii("attruibute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("publically"), &["publicly"]),
            (dictgen::InsensitiveStr::Ascii("mahcine"), &["machine"]),
            (dictgen::InsensitiveStr::Ascii("igzorts"), &["exhorts"]),
            (dictgen::InsensitiveStr::Ascii("averloads"), &["overloads"]),
            (dictgen::InsensitiveStr::Ascii("samping"), &["sampling"]),
            (
                dictgen::InsensitiveStr::Ascii("excisting"),
                &["existing", "excising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininister"),
                &["administer"],
            ),
            (dictgen::InsensitiveStr::Ascii("zionisim"), &["zionism"]),
            (dictgen::InsensitiveStr::Ascii("disucuss"), &["discuss"]),
            (
                dictgen::InsensitiveStr::Ascii("terrirorial"),
                &["territorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefifying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("vecvtor"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("airrport"), &["airport"]),
            (
                dictgen::InsensitiveStr::Ascii("semapthores"),
                &["semaphores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifeir"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("alvorithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsequantial"),
                &["inconsequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("reasonal"), &["reasonably"]),
            (
                dictgen::InsensitiveStr::Ascii("electornic"),
                &["electronic"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifes"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("signficiantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("projext"), &["project"]),
            (
                dictgen::InsensitiveStr::Ascii("prototpyes"),
                &["prototypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("parellel"), &["parallel"]),
            (
                dictgen::InsensitiveStr::Ascii("spefifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("calbire"), &["calibre"]),
            (
                dictgen::InsensitiveStr::Ascii("sharpenning"),
                &["sharpening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particpant"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("inifinte"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("navigaters"),
                &["navigators", "navigates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retroactivley"),
                &["retroactively"],
            ),
            (dictgen::InsensitiveStr::Ascii("oragnizer"), &["organizer"]),
            (
                dictgen::InsensitiveStr::Ascii("fingeprint"),
                &["fingerprint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internacional"),
                &["international"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminolgy"),
                &["terminology"],
            ),
            (dictgen::InsensitiveStr::Ascii("practicval"), &["practical"]),
            (
                dictgen::InsensitiveStr::Ascii("intialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("consoluted"), &["consulted"]),
            (
                dictgen::InsensitiveStr::Ascii("pseudorinverse"),
                &["pseudoinverse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappearnace"),
                &["disappearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinatior"),
                &["coordinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("annother"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("sweatshort"),
                &["sweatshirt"],
            ),
            (dictgen::InsensitiveStr::Ascii("liscence"), &["licence"]),
            (dictgen::InsensitiveStr::Ascii("descirbe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("meausures"), &["measures"]),
            (dictgen::InsensitiveStr::Ascii("unifom"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("pasturizing"),
                &["pasteurizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("waterprooof"),
                &["waterproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cononicalize"),
                &["canonicalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("claculation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("teachnig"), &["teaching"]),
            (dictgen::InsensitiveStr::Ascii("informerad"), &["informed"]),
            (dictgen::InsensitiveStr::Ascii("slectable"), &["selectable"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanents"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifies"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstandind"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilaized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstadning"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("restorated"), &["restored"]),
            (dictgen::InsensitiveStr::Ascii("pircuit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("comacting"), &["compacting"]),
            (
                dictgen::InsensitiveStr::Ascii("chrashing"),
                &["crashing", "thrashing", "trashing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleaers"),
                &["clears", "cleaners", "cleavers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimissively"),
                &["dismissively"],
            ),
            (dictgen::InsensitiveStr::Ascii("overwelm"), &["overwhelm"]),
            (
                dictgen::InsensitiveStr::Ascii("expermientel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("merang"), &["meringue"]),
            (dictgen::InsensitiveStr::Ascii("incrase"), &["increase"]),
            (dictgen::InsensitiveStr::Ascii("acqueus"), &["aqueous"]),
            (
                dictgen::InsensitiveStr::Ascii("compunation"),
                &["computation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplorible"),
                &["deplorable"],
            ),
            (dictgen::InsensitiveStr::Ascii("meanining"), &["meaning"]),
            (dictgen::InsensitiveStr::Ascii("enfores"), &["enforces"]),
            (dictgen::InsensitiveStr::Ascii("rearranes"), &["rearranges"]),
            (dictgen::InsensitiveStr::Ascii("childeren"), &["children"]),
            (
                dictgen::InsensitiveStr::Ascii("contatenates"),
                &["concatenates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rehabilitaion"),
                &["rehabilitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallellised"),
                &["parallelised"],
            ),
            (dictgen::InsensitiveStr::Ascii("garphing"), &["graphing"]),
            (
                dictgen::InsensitiveStr::Ascii("explictely"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("mmoderator"), &["moderator"]),
            (dictgen::InsensitiveStr::Ascii("aplhabets"), &["alphabets"]),
            (dictgen::InsensitiveStr::Ascii("stressade"), &["stressed"]),
            (dictgen::InsensitiveStr::Ascii("aricraft"), &["aircraft"]),
            (
                dictgen::InsensitiveStr::Ascii("garanteeed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminales"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("neverhteless"),
                &["nevertheless"],
            ),
            (dictgen::InsensitiveStr::Ascii("elliminate"), &["eliminate"]),
            (dictgen::InsensitiveStr::Ascii("detceting"), &["detecting"]),
            (
                dictgen::InsensitiveStr::Ascii("contination"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloumn"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("mirgated"), &["migrated"]),
            (
                dictgen::InsensitiveStr::Ascii("experiming"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("chivaly"), &["chivalry"]),
            (
                dictgen::InsensitiveStr::Ascii("misceancellous"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("withdrawin"), &["withdrawn"]),
            (dictgen::InsensitiveStr::Ascii("detectionn"), &["detection"]),
            (dictgen::InsensitiveStr::Ascii("pershis"), &["precious"]),
            (
                dictgen::InsensitiveStr::Ascii("decroative"),
                &["decorative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighboorhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("concering"), &["concerning"]),
            (
                dictgen::InsensitiveStr::Ascii("fragmantation"),
                &["fragmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independenant"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("apparrels"), &["apparels"]),
            (
                dictgen::InsensitiveStr::Ascii("emphaising"),
                &["emphasising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranparencies"),
                &["transparencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tiples"),
                &["triples", "tuples"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellements"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("skillsto"), &["skillshots"]),
            (
                dictgen::InsensitiveStr::Ascii("massachusettes"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("sparates"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("compresors"),
                &["compressor", "compressors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defrentiates"),
                &["differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinating"),
                &["determining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definatetly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("infermon"), &["inferno"]),
            (dictgen::InsensitiveStr::Ascii("coccinele"), &["coccinelle"]),
            (dictgen::InsensitiveStr::Ascii("loaclize"), &["localize"]),
            (
                dictgen::InsensitiveStr::Ascii("refferencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("charctors"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("jusifications"),
                &["justifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramedis"), &["paramedics"]),
            (
                dictgen::InsensitiveStr::Ascii("regnerating"),
                &["regenerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("folllowin"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("discreditied"),
                &["discredited"],
            ),
            (dictgen::InsensitiveStr::Ascii("insipring"), &["inspiring"]),
            (
                dictgen::InsensitiveStr::Ascii("versiobing"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("dabase"), &["database"]),
            (dictgen::InsensitiveStr::Ascii("impliciet"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("embeedder"), &["embedder"]),
            (dictgen::InsensitiveStr::Ascii("wiredest"), &["weirdest"]),
            (
                dictgen::InsensitiveStr::Ascii("proconceived"),
                &["preconceived"],
            ),
            (dictgen::InsensitiveStr::Ascii("curretnly"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("absoutly"), &["absolutely"]),
            (
                dictgen::InsensitiveStr::Ascii("recuning"),
                &["reckoning", "recusing", "retuning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diasbility"),
                &["disability"],
            ),
            (dictgen::InsensitiveStr::Ascii("apperead"), &["appeared"]),
            (dictgen::InsensitiveStr::Ascii("uesrs"), &["users"]),
            (dictgen::InsensitiveStr::Ascii("substanse"), &["substances"]),
            (
                dictgen::InsensitiveStr::Ascii("interripted"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("pubishers"), &["publishers"]),
            (dictgen::InsensitiveStr::Ascii("applicble"), &["applicable"]),
            (
                dictgen::InsensitiveStr::Ascii("boootloaders"),
                &["bootloaders"],
            ),
            (dictgen::InsensitiveStr::Ascii("refure"), &["refuse"]),
            (
                dictgen::InsensitiveStr::Ascii("hierchal"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("agreeeable"), &["agreeable"]),
            (
                dictgen::InsensitiveStr::Ascii("customizeble"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ntegers"), &["integers"]),
            (dictgen::InsensitiveStr::Ascii("vigliant"), &["vigilant"]),
            (
                dictgen::InsensitiveStr::Ascii("circunvented"),
                &["circumvented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arithmatic"),
                &["arithmetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formallises"),
                &["formalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("navgiate"), &["navigate"]),
            (dictgen::InsensitiveStr::Ascii("diagnosies"), &["diagnosis"]),
            (
                dictgen::InsensitiveStr::Ascii("reinfocements"),
                &["reinforcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("intinite"), &["infinite"]),
            (dictgen::InsensitiveStr::Ascii("hsould"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("cataalogue"), &["catalogue"]),
            (dictgen::InsensitiveStr::Ascii("declartor"), &["declarator"]),
            (
                dictgen::InsensitiveStr::Ascii("antoganists"),
                &["antagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("abble"), &["able"]),
            (
                dictgen::InsensitiveStr::Ascii("allegience"),
                &["allegiance"],
            ),
            (dictgen::InsensitiveStr::Ascii("signular"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("metatatble"), &["metatable"]),
            (dictgen::InsensitiveStr::Ascii("aperature"), &["aperture"]),
            (
                dictgen::InsensitiveStr::Ascii("charecteristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("headder"), &["header"]),
            (dictgen::InsensitiveStr::Ascii("benifical"), &["beneficial"]),
            (dictgen::InsensitiveStr::Ascii("fomula"), &["formula"]),
            (dictgen::InsensitiveStr::Ascii("enought"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("scestion"), &["section"]),
            (
                dictgen::InsensitiveStr::Ascii("characteter"),
                &["character"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("despicaple"),
                &["despicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcoess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpierence"),
                &["inexperienced", "inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimations"), &["animations"]),
            (dictgen::InsensitiveStr::Ascii("circlebs"), &["circles"]),
            (
                dictgen::InsensitiveStr::Ascii("procedureal"),
                &["procedural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refertenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrguments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("exceding"), &["exceeding"]),
            (
                dictgen::InsensitiveStr::Ascii("uncerainty"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("columne"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("remaing"), &["remaining"]),
            (dictgen::InsensitiveStr::Ascii("wilcards"), &["wildcards"]),
            (dictgen::InsensitiveStr::Ascii("beggingin"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("developrs"), &["develops"]),
            (dictgen::InsensitiveStr::Ascii("corrrupted"), &["corrupted"]),
            (dictgen::InsensitiveStr::Ascii("dito"), &["ditto"]),
            (
                dictgen::InsensitiveStr::Ascii("invinceble"),
                &["invincible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protostants"),
                &["protestants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intgrations"),
                &["integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termperature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("intefaced"), &["interfaced"]),
            (dictgen::InsensitiveStr::Ascii("methdos"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("coducter"), &["conductor"]),
            (
                dictgen::InsensitiveStr::Ascii("deifinition"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("precition"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("verions"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("carrien"), &["carrier"]),
            (
                dictgen::InsensitiveStr::Ascii("converves"),
                &["conserves", "converses", "converges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cofrimation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofocurse"), &["ofcourse"]),
            (dictgen::InsensitiveStr::Ascii("sortings"), &["sorting"]),
            (
                dictgen::InsensitiveStr::Ascii("accumulaters"),
                &["accumulators", "accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("interaly"), &["internally"]),
            (dictgen::InsensitiveStr::Ascii("expecet"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("cringeworhy"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("temporaere"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimiliarty"),
                &["dissimilarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximites"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrigthy"), &["alrighty"]),
            (dictgen::InsensitiveStr::Ascii("whilsting"), &["whistling"]),
            (
                dictgen::InsensitiveStr::Ascii("indicatin"),
                &["indicating", "indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complition"),
                &["completion", "compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plarformer"),
                &["platformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitude"),
                &["substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("improofs"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("eveluated"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("undenialbly"),
                &["undeniably"],
            ),
            (dictgen::InsensitiveStr::Ascii("maitainer"), &["maintainer"]),
            (dictgen::InsensitiveStr::Ascii("touristly"), &["touristy"]),
            (
                dictgen::InsensitiveStr::Ascii("linerisation"),
                &["linearisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homeowneris"),
                &["homeowners"],
            ),
            (dictgen::InsensitiveStr::Ascii("worrysome"), &["worrisome"]),
            (dictgen::InsensitiveStr::Ascii("motiion"), &["motion"]),
            (
                dictgen::InsensitiveStr::Ascii("repsecting"),
                &["respecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("senarios"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("initliaser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedamy"), &["academy"]),
            (
                dictgen::InsensitiveStr::Ascii("resolutins"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construting"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("icrease"), &["increase"]),
            (dictgen::InsensitiveStr::Ascii("bufferered"), &["buffered"]),
            (dictgen::InsensitiveStr::Ascii("datatyoe"), &["datatype"]),
            (dictgen::InsensitiveStr::Ascii("epsiode"), &["episode"]),
            (
                dictgen::InsensitiveStr::Ascii("diffently"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("priortise"), &["prioritise"]),
            (
                dictgen::InsensitiveStr::Ascii("audomoderator"),
                &["automoderator"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallesse"), &["coalesce"]),
            (dictgen::InsensitiveStr::Ascii("reguire"), &["require"]),
            (dictgen::InsensitiveStr::Ascii("comiple"), &["compile"]),
            (
                dictgen::InsensitiveStr::Ascii("relinqushment"),
                &["relinquishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("collectems"), &["collects"]),
            (
                dictgen::InsensitiveStr::Ascii("guantanoma"),
                &["guantanamo"],
            ),
            (dictgen::InsensitiveStr::Ascii("brithdays"), &["birthdays"]),
            (
                dictgen::InsensitiveStr::Ascii("excectives"),
                &["executives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generetions"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commens"),
                &["comments", "commons", "commends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perameter"),
                &["parameter", "perimeter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompresor"),
                &["uncompressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretion"),
                &["interpreting", "interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("failled"), &["failed"]),
            (
                dictgen::InsensitiveStr::Ascii("condtional"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("prototoype"), &["prototype"]),
            (dictgen::InsensitiveStr::Ascii("permition"), &["permission"]),
            (dictgen::InsensitiveStr::Ascii("redundent"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("cofficients"),
                &["coefficients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convesations"),
                &["conversations"],
            ),
            (dictgen::InsensitiveStr::Ascii("negoticate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("arbritation"),
                &["arbitration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unpreparred"),
                &["unprepared"],
            ),
            (dictgen::InsensitiveStr::Ascii("chirstian"), &["christian"]),
            (
                dictgen::InsensitiveStr::Ascii("reconising"),
                &["recognising"],
            ),
            (dictgen::InsensitiveStr::Ascii("converte"), &["converts"]),
            (
                dictgen::InsensitiveStr::Ascii("foramatting"),
                &["formatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppurtunities"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skateborading"),
                &["skateboarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheild"), &["shield"]),
            (
                dictgen::InsensitiveStr::Ascii("celebratin"),
                &["celebration", "celebrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("havign"), &["having"]),
            (
                dictgen::InsensitiveStr::Ascii("variebility"),
                &["variability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requiremnts"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("rquesting"), &["requesting"]),
            (dictgen::InsensitiveStr::Ascii("preverve"), &["preserve"]),
            (dictgen::InsensitiveStr::Ascii("lage"), &["large"]),
            (dictgen::InsensitiveStr::Ascii("inadequet"), &["inadequate"]),
            (dictgen::InsensitiveStr::Ascii("execuates"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("politiicans"),
                &["politician"],
            ),
            (dictgen::InsensitiveStr::Ascii("happines"), &["happiness"]),
            (
                dictgen::InsensitiveStr::Ascii("psychiatrisch"),
                &["psychiatric"],
            ),
            (dictgen::InsensitiveStr::Ascii("futrue"), &["future"]),
            (
                dictgen::InsensitiveStr::Ascii("direcetories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignormd"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("hosteles"), &["hostels"]),
            (
                dictgen::InsensitiveStr::Ascii("distributibe"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("appeard"), &["appeared"]),
            (dictgen::InsensitiveStr::Ascii("endores"), &["endorse"]),
            (dictgen::InsensitiveStr::Ascii("subtittles"), &["subtitles"]),
            (dictgen::InsensitiveStr::Ascii("checls"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("absorbsion"),
                &["absorption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funchtionns"),
                &["functions"],
            ),
            (dictgen::InsensitiveStr::Ascii("identfies"), &["identifies"]),
            (dictgen::InsensitiveStr::Ascii("truied"), &["tried"]),
            (dictgen::InsensitiveStr::Ascii("heorine"), &["heroine"]),
            (dictgen::InsensitiveStr::Ascii("demonizin"), &["demonizing"]),
            (dictgen::InsensitiveStr::Ascii("sufferring"), &["suffering"]),
            (dictgen::InsensitiveStr::Ascii("notmalize"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("cloudes"), &["clouds"]),
            (dictgen::InsensitiveStr::Ascii("cahacter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("accessiblity"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplations"),
                &["duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("extedns"), &["extends"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusivos"),
                &["exclusives"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeenders"), &["appenders"]),
            (dictgen::InsensitiveStr::Ascii("utiltiy"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("allwys"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("muliplied"), &["multiplied"]),
            (dictgen::InsensitiveStr::Ascii("unbanend"), &["unbanned"]),
            (dictgen::InsensitiveStr::Ascii("immboile"), &["immobile"]),
            (
                dictgen::InsensitiveStr::Ascii("transimtted"),
                &["transmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetrification"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrests"), &["interests"]),
            (
                dictgen::InsensitiveStr::Ascii("fasing"),
                &["fazing", "phasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scubscribes"),
                &["subscribes"],
            ),
            (dictgen::InsensitiveStr::Ascii("famillies"), &["families"]),
            (
                dictgen::InsensitiveStr::Ascii("distubring"),
                &["disturbing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adivsoriy"),
                &["advisory", "advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlaytic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("luietenant"),
                &["lieutenant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intuitevly"),
                &["intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("closests"),
                &["closest", "closets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irregulierties"),
                &["irregularities"],
            ),
            (dictgen::InsensitiveStr::Ascii("etmyology"), &["etymology"]),
            (
                dictgen::InsensitiveStr::Ascii("proected"),
                &["protected", "projected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spoff"),
                &["spoof", "spiff", "spaff", "scoff"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrrupting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrigarator"),
                &["refrigerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifig"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("aeriel"), &["aerial"]),
            (
                dictgen::InsensitiveStr::Ascii("confugration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("devired"), &["derived"]),
            (
                dictgen::InsensitiveStr::Ascii("conquoring"),
                &["conquering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associeated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetificates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comporessors"),
                &["compressors"],
            ),
            (dictgen::InsensitiveStr::Ascii("yout"), &["your", "you"]),
            (dictgen::InsensitiveStr::Ascii("resorces"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("generalisin"),
                &["generalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("renember"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("visitign"), &["visiting"]),
            (dictgen::InsensitiveStr::Ascii("obejction"), &["objection"]),
            (
                dictgen::InsensitiveStr::Ascii("geometites"),
                &["geometries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convineance"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwowin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("poniters"), &["pointers"]),
            (dictgen::InsensitiveStr::Ascii("mundande"), &["mundane"]),
            (dictgen::InsensitiveStr::Ascii("guradian"), &["guardian"]),
            (
                dictgen::InsensitiveStr::Ascii("conceivibly"),
                &["conceivably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entrys"),
                &["entries", "entry"],
            ),
            (dictgen::InsensitiveStr::Ascii("viusalise"), &["visualise"]),
            (dictgen::InsensitiveStr::Ascii("tracsoder"), &["transcoder"]),
            (
                dictgen::InsensitiveStr::Ascii("americains"),
                &["americas", "americans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invote"),
                &["invoke", "invite"],
            ),
            (dictgen::InsensitiveStr::Ascii("netboook"), &["netbook"]),
            (dictgen::InsensitiveStr::Ascii("viscious"), &["vicious"]),
            (dictgen::InsensitiveStr::Ascii("volumnes"), &["volumes"]),
            (dictgen::InsensitiveStr::Ascii("naybhors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("thhere"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("hidding"),
                &["hiding", "hidden"],
            ),
            (dictgen::InsensitiveStr::Ascii("missiony"), &["missionary"]),
            (dictgen::InsensitiveStr::Ascii("threefor"), &["therefor"]),
            (
                dictgen::InsensitiveStr::Ascii("dissconnect"),
                &["disconnect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scalarizaiton"),
                &["scalarization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotitaion"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("parachutte"), &["parachute"]),
            (
                dictgen::InsensitiveStr::Ascii("absuolatism"),
                &["absolutism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workaroudn"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("resolvinf"), &["resolving"]),
            (
                dictgen::InsensitiveStr::Ascii("refeinements"),
                &["refinements"],
            ),
            (dictgen::InsensitiveStr::Ascii("usersapce"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("authoities"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summarizaion"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("effecked"), &["effected"]),
            (dictgen::InsensitiveStr::Ascii("modefied"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("psychedlic"),
                &["psychedelic"],
            ),
            (dictgen::InsensitiveStr::Ascii("prophey"), &["prophecy"]),
            (
                dictgen::InsensitiveStr::Ascii("mirconesia"),
                &["micronesia"],
            ),
            (dictgen::InsensitiveStr::Ascii("attanded"), &["attended"]),
            (dictgen::InsensitiveStr::Ascii("preferly"), &["preferably"]),
            (dictgen::InsensitiveStr::Ascii("uderlying"), &["underlying"]),
            (
                dictgen::InsensitiveStr::Ascii("tightenting"),
                &["tightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("painfull"),
                &["painful", "painfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("preculded"), &["precluded"]),
            (dictgen::InsensitiveStr::Ascii("execturing"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("apostrafes"),
                &["apostrophes"],
            ),
            (dictgen::InsensitiveStr::Ascii("compunds"), &["compounds"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentors"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("convised"), &["convinced"]),
            (dictgen::InsensitiveStr::Ascii("nnumber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("sponosor"), &["sponsor"]),
            (
                dictgen::InsensitiveStr::Ascii("expetimented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacinate"), &["vaccinate"]),
            (dictgen::InsensitiveStr::Ascii("sclaing"), &["scaling"]),
            (
                dictgen::InsensitiveStr::Ascii("directorie"),
                &["directory", "directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("walkthough"),
                &["walkthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associeative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrerencial"),
                &["referential"],
            ),
            (dictgen::InsensitiveStr::Ascii("spwaning"), &["spawning"]),
            (dictgen::InsensitiveStr::Ascii("whyth"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("crationist"),
                &["creationist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninfromed"),
                &["uninformed", "uniformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buisnessman"),
                &["businessman"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reolving"),
                &["resolving", "revolving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupplicates"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("vehcile"), &["vehicle"]),
            (
                dictgen::InsensitiveStr::Ascii("rationalizin"),
                &["rationalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reutnring"), &["returning"]),
            (
                dictgen::InsensitiveStr::Ascii("accdiently"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sighrynge"), &["syringe"]),
            (
                dictgen::InsensitiveStr::Ascii("supressable"),
                &["suppressible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitons"),
                &["competitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instlling"),
                &["installing", "instilling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proceessors"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sizemologogical"),
                &["seismological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissisions"),
                &["permissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruicial"), &["crucial"]),
            (
                dictgen::InsensitiveStr::Ascii("windowd"),
                &["windows", "windowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entertainted"),
                &["entertained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarantena"),
                &["quarantine"],
            ),
            (dictgen::InsensitiveStr::Ascii("istalling"), &["installing"]),
            (
                dictgen::InsensitiveStr::Ascii("definatily"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("resolt"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("doucmentated"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguards"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("fasicsm"), &["fascism"]),
            (dictgen::InsensitiveStr::Ascii("veticle"), &["vehicle"]),
            (dictgen::InsensitiveStr::Ascii("elctronic"), &["electronic"]),
            (
                dictgen::InsensitiveStr::Ascii("rearch"),
                &["search", "research", "reach"],
            ),
            (dictgen::InsensitiveStr::Ascii("manement"), &["management"]),
            (dictgen::InsensitiveStr::Ascii("speficieid"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("anjanew"), &["ingenue"]),
            (
                dictgen::InsensitiveStr::Ascii("condolonces"),
                &["condolences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appnedages"),
                &["appendages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimentionality"),
                &["dimensionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("properrys"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("libraarie"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("expolitative"),
                &["exploitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("whereever"), &["wherever"]),
            (dictgen::InsensitiveStr::Ascii("meatbals"), &["meatballs"]),
            (
                dictgen::InsensitiveStr::Ascii("irreversibile"),
                &["irreversible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paricipations"),
                &["participations"],
            ),
            (dictgen::InsensitiveStr::Ascii("maches"), &["matches"]),
            (dictgen::InsensitiveStr::Ascii("prority"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("constributions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("conly"), &["only"]),
            (dictgen::InsensitiveStr::Ascii("brigthly"), &["brightly"]),
            (dictgen::InsensitiveStr::Ascii("dashoboard"), &["dashboard"]),
            (
                dictgen::InsensitiveStr::Ascii("proerty"),
                &["property", "poetry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitucion"),
                &["constitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("repharse"), &["rephrase"]),
            (dictgen::InsensitiveStr::Ascii("tollerable"), &["tolerable"]),
            (dictgen::InsensitiveStr::Ascii("infornt"), &["infront"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomeamce"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("repitle"), &["reptile"]),
            (dictgen::InsensitiveStr::Ascii("reqisters"), &["registers"]),
            (dictgen::InsensitiveStr::Ascii("allwance"), &["allowance"]),
            (dictgen::InsensitiveStr::Ascii("wiredness"), &["weirdness"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptuvely"),
                &["descriptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("accapt"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("reproducble"),
                &["reproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintentially"),
                &["unintentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("labbels"), &["labels"]),
            (dictgen::InsensitiveStr::Ascii("payemnt"), &["payment"]),
            (dictgen::InsensitiveStr::Ascii("adobted"), &["adopted"]),
            (dictgen::InsensitiveStr::Ascii("husban"), &["husband"]),
            (dictgen::InsensitiveStr::Ascii("nstalls"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("apostrafy"), &["apostrophe"]),
            (dictgen::InsensitiveStr::Ascii("autosae"), &["autosave"]),
            (
                dictgen::InsensitiveStr::Ascii("ovious"),
                &["obvious", "odious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forgiviness"),
                &["forgiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthenticates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporoximating"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconistent"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("falltroughs"),
                &["fallthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("untargetted"),
                &["untargeted"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperato"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("grpahical"), &["graphical"]),
            (
                dictgen::InsensitiveStr::Ascii("visualizatons"),
                &["visualizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spirutuality"),
                &["spirituality"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecable"), &["selectable"]),
            (dictgen::InsensitiveStr::Ascii("symboles"), &["symbols"]),
            (
                dictgen::InsensitiveStr::Ascii("icremented"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("besitality"),
                &["bestiality"],
            ),
            (dictgen::InsensitiveStr::Ascii("conenctor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("resubmittin"),
                &["resubmitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("prvileges"), &["privileges"]),
            (
                dictgen::InsensitiveStr::Ascii("desciprion"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("positiviy"), &["positivity"]),
            (dictgen::InsensitiveStr::Ascii("imples"), &["implies"]),
            (dictgen::InsensitiveStr::Ascii("fobidding"), &["forbidding"]),
            (
                dictgen::InsensitiveStr::Ascii("appreaciates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfoemers"),
                &["performers"],
            ),
            (dictgen::InsensitiveStr::Ascii("producted"), &["produced"]),
            (dictgen::InsensitiveStr::Ascii("tocksen"), &["toxin"]),
            (
                dictgen::InsensitiveStr::Ascii("attribures"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethenticates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispoables"),
                &["disposables"],
            ),
            (dictgen::InsensitiveStr::Ascii("delievere"), &["delivered"]),
            (dictgen::InsensitiveStr::Ascii("dymanics"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("invincable"),
                &["invincible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bioligically"),
                &["biologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("igoring"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("normaly"), &["normally"]),
            (
                dictgen::InsensitiveStr::Ascii("sorrounded"),
                &["surrounded"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyllow"), &["yellow"]),
            (dictgen::InsensitiveStr::Ascii("anualized"), &["annualized"]),
            (dictgen::InsensitiveStr::Ascii("disgns"), &["designs"]),
            (dictgen::InsensitiveStr::Ascii("timesptamp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("opitons"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("attemppt"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("escalte"), &["escalate"]),
            (dictgen::InsensitiveStr::Ascii("cyclistes"), &["cyclists"]),
            (dictgen::InsensitiveStr::Ascii("deligate"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("checkum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("stogare"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("enlish"),
                &["english", "enlist"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("apprecite"), &["appreciate"]),
            (
                dictgen::InsensitiveStr::Ascii("inititialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitpliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("miinimum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("caffinee"), &["caffeine"]),
            (
                dictgen::InsensitiveStr::Ascii("companinion"),
                &["companion"],
            ),
            (dictgen::InsensitiveStr::Ascii("adviseur"), &["adviser"]),
            (
                dictgen::InsensitiveStr::Ascii("retransmisson"),
                &["retransmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("instument"), &["instrument"]),
            (
                dictgen::InsensitiveStr::Ascii("abreviation"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("succesive"), &["successive"]),
            (dictgen::InsensitiveStr::Ascii("sufix"), &["suffix"]),
            (
                dictgen::InsensitiveStr::Ascii("incremeanting"),
                &["incrementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcdories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("weahters"), &["weathers"]),
            (
                dictgen::InsensitiveStr::Ascii("battelships"),
                &["battleship"],
            ),
            (dictgen::InsensitiveStr::Ascii("oricles"), &["oracles"]),
            (dictgen::InsensitiveStr::Ascii("presicion"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("emblamatic"),
                &["emblematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonized"),
                &["synchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambluances"),
                &["ambulances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradditional"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("overheards"), &["overhead"]),
            (
                dictgen::InsensitiveStr::Ascii("rechargable"),
                &["rechargeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assigmnents"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convination"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("dowrds"), &["dwords"]),
            (dictgen::InsensitiveStr::Ascii("conservite"), &["conserve"]),
            (
                dictgen::InsensitiveStr::Ascii("inflamation"),
                &["inflammation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procuder"),
                &["procurer", "producer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidenly"),
                &["confidently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convaxness"),
                &["convexness"],
            ),
            (dictgen::InsensitiveStr::Ascii("defice"), &["device"]),
            (dictgen::InsensitiveStr::Ascii("shdowed"), &["shadowed"]),
            (dictgen::InsensitiveStr::Ascii("ellapsed"), &["elapsed"]),
            (dictgen::InsensitiveStr::Ascii("convinse"), &["convinces"]),
            (
                dictgen::InsensitiveStr::Ascii("transocded"),
                &["transcoded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expressable"),
                &["expressible"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumets"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("durring"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("instered"), &["inserted"]),
            (
                dictgen::InsensitiveStr::Ascii("encomapsses"),
                &["encompasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("interpet"), &["interpret"]),
            (dictgen::InsensitiveStr::Ascii("stategize"), &["strategize"]),
            (
                dictgen::InsensitiveStr::Ascii("imprementations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nutrutional"),
                &["nutritional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpreation"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("momento"), &["memento"]),
            (dictgen::InsensitiveStr::Ascii("sautayed"), &["sautéd"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifier"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("offspirng"), &["offspring"]),
            (dictgen::InsensitiveStr::Ascii("thge"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("kindgergarden"),
                &["kindergarten"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciriculum"), &["curriculum"]),
            (dictgen::InsensitiveStr::Ascii("exorcizers"), &["exorcists"]),
            (dictgen::InsensitiveStr::Ascii("decpetive"), &["deceptive"]),
            (dictgen::InsensitiveStr::Ascii("godlberg"), &["goldberg"]),
            (
                dictgen::InsensitiveStr::Ascii("deplicates"),
                &["duplicates", "replicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalations"),
                &["evaluations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequencies"),
                &["consequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("unneded"), &["unneeded"]),
            (
                dictgen::InsensitiveStr::Ascii("spontaneus"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("asuming"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("aggragating"),
                &["aggravating", "aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("accorddion"), &["accordion"]),
            (dictgen::InsensitiveStr::Ascii("pythjon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("assumang"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("unbelievablely"),
                &["unbelievably"],
            ),
            (dictgen::InsensitiveStr::Ascii("conained"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("exitss"),
                &["exists", "exits"],
            ),
            (dictgen::InsensitiveStr::Ascii("marmelade"), &["marmalade"]),
            (dictgen::InsensitiveStr::Ascii("reapeated"), &["repeated"]),
            (dictgen::InsensitiveStr::Ascii("stocastic"), &["stochastic"]),
            (
                dictgen::InsensitiveStr::Ascii("currecnies"),
                &["currencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("preferrs"), &["prefers"]),
            (dictgen::InsensitiveStr::Ascii("biejing"), &["beijing"]),
            (dictgen::InsensitiveStr::Ascii("decissions"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("recepient"), &["recipient"]),
            (dictgen::InsensitiveStr::Ascii("ukarine"), &["ukraine"]),
            (
                dictgen::InsensitiveStr::Ascii("deregualtion"),
                &["deregulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caclulated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mamagers"), &["managers"]),
            (
                dictgen::InsensitiveStr::Ascii("inialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("solications"),
                &["solicitations", "solidifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjectivs"), &["adjectives"]),
            (
                dictgen::InsensitiveStr::Ascii("reaise"),
                &["realise", "raise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prospertiy"),
                &["prosperity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visibily"),
                &["visibility", "visibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("containner"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("expeect"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("exnternalities"),
                &["externalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("legeacy"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("contibution"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geometrires"),
                &["geometries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtaulizes"),
                &["virtualizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addrerssses"),
                &["addresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplacating"),
                &["duplicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiniation"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surveilence"),
                &["surveillance"],
            ),
            (dictgen::InsensitiveStr::Ascii("govemrent"), &["government"]),
            (dictgen::InsensitiveStr::Ascii("uprade"), &["upgrade"]),
            (dictgen::InsensitiveStr::Ascii("aranging"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("milisecons"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("inferrence"), &["inference"]),
            (dictgen::InsensitiveStr::Ascii("appriate"), &["appropriate"]),
            (dictgen::InsensitiveStr::Ascii("seprating"), &["separating"]),
            (dictgen::InsensitiveStr::Ascii("abcense"), &["absence"]),
            (dictgen::InsensitiveStr::Ascii("chnanges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("challangers"),
                &["challengers"],
            ),
            (dictgen::InsensitiveStr::Ascii("peroids"), &["periods"]),
            (dictgen::InsensitiveStr::Ascii("unesacped"), &["unescaped"]),
            (dictgen::InsensitiveStr::Ascii("algorthms"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("anaytics"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("protcools"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("consummers"), &["consumers"]),
            (dictgen::InsensitiveStr::Ascii("moddels"), &["models"]),
            (dictgen::InsensitiveStr::Ascii("creatien"), &["creatine"]),
            (
                dictgen::InsensitiveStr::Ascii("excectable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoiciation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("declre"), &["declare"]),
            (
                dictgen::InsensitiveStr::Ascii("chekcpoint"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utlisation"),
                &["utilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("skippin"), &["skipping"]),
            (
                dictgen::InsensitiveStr::Ascii("unselecgtes"),
                &["unselects"],
            ),
            (dictgen::InsensitiveStr::Ascii("curriculm"), &["curriculum"]),
            (dictgen::InsensitiveStr::Ascii("defensie"), &["defensive"]),
            (
                dictgen::InsensitiveStr::Ascii("noramalized"),
                &["normalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimitation"),
                &["optimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximizin"), &["maximizing"]),
            (dictgen::InsensitiveStr::Ascii("culd"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("oublisher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("ednpoints"), &["endpoints"]),
            (dictgen::InsensitiveStr::Ascii("guarrante"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("technicallity"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("steped"), &["stepped"]),
            (dictgen::InsensitiveStr::Ascii("texline"), &["textline"]),
            (
                dictgen::InsensitiveStr::Ascii("tyin"),
                &["tying", "thin", "twin", "tin"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeation"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("segemnet"), &["segment"]),
            (
                dictgen::InsensitiveStr::Ascii("reguirements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("usused"), &["unused"]),
            (
                dictgen::InsensitiveStr::Ascii("decleration"),
                &["declaration", "deceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisement"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exportet"),
                &["exported", "exporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("interracts"), &["interacts"]),
            (
                dictgen::InsensitiveStr::Ascii("collecters"),
                &["collectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumise"),
                &["costumes", "customise"],
            ),
            (dictgen::InsensitiveStr::Ascii("surrrogate"), &["surrogate"]),
            (
                dictgen::InsensitiveStr::Ascii("legitimetly"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legimitately"),
                &["legitimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("dysoptian"), &["dystopian"]),
            (dictgen::InsensitiveStr::Ascii("mutipath"), &["multipath"]),
            (dictgen::InsensitiveStr::Ascii("celll"), &["cell"]),
            (dictgen::InsensitiveStr::Ascii("chaeckbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("nautlius"), &["nautilus"]),
            (
                dictgen::InsensitiveStr::Ascii("mouthpeices"),
                &["mouthpiece"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speeping"),
                &["sleeping", "speeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempating"),
                &["templating", "tempting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnessesarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("satify"), &["satisfy"]),
            (
                dictgen::InsensitiveStr::Ascii("transparanceies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("midotwn"), &["midtown"]),
            (
                dictgen::InsensitiveStr::Ascii("incompartibilities"),
                &["incompatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregistation"),
                &["unregistration"],
            ),
            (dictgen::InsensitiveStr::Ascii("priroity"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("lunaticus"), &["lunatics"]),
            (dictgen::InsensitiveStr::Ascii("anoymous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("permissoins"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mdifications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloocks"), &["blocks"]),
            (dictgen::InsensitiveStr::Ascii("stetement"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("outputing"), &["outputting"]),
            (
                dictgen::InsensitiveStr::Ascii("characteters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("meethods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("plcae"), &["place"]),
            (dictgen::InsensitiveStr::Ascii("questoned"), &["questioned"]),
            (
                dictgen::InsensitiveStr::Ascii("intecepting"),
                &["intercepting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condionally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incrememtal"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcment"),
                &["announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agravating"),
                &["aggravating"],
            ),
            (dictgen::InsensitiveStr::Ascii("unnmatched"), &["unmatched"]),
            (dictgen::InsensitiveStr::Ascii("succssors"), &["successors"]),
            (
                dictgen::InsensitiveStr::Ascii("preventivo"),
                &["prevention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accommodatin"),
                &["accommodating", "accommodation"],
            ),
            (dictgen::InsensitiveStr::Ascii("recogninse"), &["recognise"]),
            (dictgen::InsensitiveStr::Ascii("onoly"), &["only"]),
            (dictgen::InsensitiveStr::Ascii("packaes"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("availaible"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("ajustement"),
                &["adjustment"],
            ),
            (dictgen::InsensitiveStr::Ascii("europenas"), &["europeans"]),
            (dictgen::InsensitiveStr::Ascii("exitting"), &["exiting"]),
            (dictgen::InsensitiveStr::Ascii("vaidator"), &["validator"]),
            (dictgen::InsensitiveStr::Ascii("flawlessy"), &["flawlessly"]),
            (dictgen::InsensitiveStr::Ascii("dowloaded"), &["downloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbothood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("obecting"), &["objecting"]),
            (
                dictgen::InsensitiveStr::Ascii("additonaly"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("zeoring"), &["zeroing"]),
            (
                dictgen::InsensitiveStr::Ascii("ointed"),
                &["pointed", "jointed", "linted"],
            ),
            (dictgen::InsensitiveStr::Ascii("dircetory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("spripts"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("lanuguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("suplied"), &["supplied"]),
            (
                dictgen::InsensitiveStr::Ascii("effords"),
                &["efforts", "affords"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inforamational"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internationnal"),
                &["international"],
            ),
            (dictgen::InsensitiveStr::Ascii("fielname"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalis"),
                &["fundamentals"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorith"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("intermeidate"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("serailze"), &["serialize"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsitering"),
                &["administering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resilently"),
                &["resiliently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpepted"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theorectically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowlegdable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrid"), &["third"]),
            (
                dictgen::InsensitiveStr::Ascii("inpractically"),
                &["impractically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstrating"),
                &["abstracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geographicaly"),
                &["geographically"],
            ),
            (dictgen::InsensitiveStr::Ascii("maneer"), &["manner"]),
            (
                dictgen::InsensitiveStr::Ascii("allowes"),
                &["allowed", "allows"],
            ),
            (dictgen::InsensitiveStr::Ascii("decited"), &["decided"]),
            (dictgen::InsensitiveStr::Ascii("initialyl"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("interstection"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("enerator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("abrubtly"), &["abruptly"]),
            (
                dictgen::InsensitiveStr::Ascii("appens"),
                &["appends", "happens", "aspens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("windsheild"),
                &["windshield"],
            ),
            (dictgen::InsensitiveStr::Ascii("accademia"), &["academia"]),
            (
                dictgen::InsensitiveStr::Ascii("terfformed"),
                &["terraformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condeferacy"),
                &["confederacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("signuature"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("indictement"),
                &["indictment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unducumented"),
                &["undocumented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecifies"),
                &["specifies", "pacifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenating"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("missonary"), &["missionary"]),
            (
                dictgen::InsensitiveStr::Ascii("autogenrated"),
                &["autogenerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirrectly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("addresfs"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("japenese"), &["japanese"]),
            (dictgen::InsensitiveStr::Ascii("calrify"), &["clarify"]),
            (
                dictgen::InsensitiveStr::Ascii("tessleating"),
                &["tessellating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounaded"), &["bounded"]),
            (dictgen::InsensitiveStr::Ascii("meltodwn"), &["meltdown"]),
            (dictgen::InsensitiveStr::Ascii("aldready"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("certificat"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("amernian"), &["armenian"]),
            (dictgen::InsensitiveStr::Ascii("arcival"), &["archival"]),
            (dictgen::InsensitiveStr::Ascii("dissole"), &["dissolve"]),
            (dictgen::InsensitiveStr::Ascii("horible"), &["horrible"]),
            (
                dictgen::InsensitiveStr::Ascii("ampliflying"),
                &["amplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("allwo"), &["allow"]),
            (dictgen::InsensitiveStr::Ascii("dows"), &["does", "down"]),
            (dictgen::InsensitiveStr::Ascii("requiested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("adderessed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("exanding"), &["expanding"]),
            (dictgen::InsensitiveStr::Ascii("jave"), &["java", "have"]),
            (
                dictgen::InsensitiveStr::Ascii("republcians"),
                &["republicans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachterisation"),
                &["characterisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("failded"), &["failed"]),
            (dictgen::InsensitiveStr::Ascii("origigin"), &["origin"]),
            (
                dictgen::InsensitiveStr::Ascii("rationalizating"),
                &["rationalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("postition"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("remplace"), &["replace"]),
            (dictgen::InsensitiveStr::Ascii("becommes"), &["becomes"]),
            (
                dictgen::InsensitiveStr::Ascii("muscel"),
                &["muscle", "mussel"],
            ),
            (dictgen::InsensitiveStr::Ascii("symptomes"), &["symptoms"]),
            (
                dictgen::InsensitiveStr::Ascii("historicly"),
                &["historical"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssistant"), &["assistant"]),
            (dictgen::InsensitiveStr::Ascii("gyspies"), &["gypsies"]),
            (
                dictgen::InsensitiveStr::Ascii("identificaiton"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagionals"), &["diagonals"]),
            (dictgen::InsensitiveStr::Ascii("adjative"), &["adjective"]),
            (
                dictgen::InsensitiveStr::Ascii("sequenstial"),
                &["sequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingnore"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("exposted"), &["exposed"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("crtical"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("respsectively"),
                &["respectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extenions"),
                &["extension", "extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifcats"), &["artifacts"]),
            (dictgen::InsensitiveStr::Ascii("usiable"), &["usable"]),
            (dictgen::InsensitiveStr::Ascii("adresse"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlated"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("searachable"),
                &["searchable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacturors"),
                &["manufacturers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocomplte"),
                &["autocomplete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ternimated"),
                &["terminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presenning"),
                &["presenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("thnaks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("fromated"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("outbuts"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("retireves"), &["retrieves"]),
            (
                dictgen::InsensitiveStr::Ascii("identitical"),
                &["identical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formatable"),
                &["formattable"],
            ),
            (dictgen::InsensitiveStr::Ascii("accont"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("apparing"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("digtial"), &["digital"]),
            (dictgen::InsensitiveStr::Ascii("unknonwn"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("identicly"),
                &["identically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corinthiens"),
                &["corinthians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prestigieus"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("regeister"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("goddanm"), &["goddam"]),
            (dictgen::InsensitiveStr::Ascii("enabels"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("esentially"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rationnal"),
                &["rational", "rationale"],
            ),
            (dictgen::InsensitiveStr::Ascii("manange"), &["manage"]),
            (
                dictgen::InsensitiveStr::Ascii("executionees"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("otside"), &["outside"]),
            (
                dictgen::InsensitiveStr::Ascii("enterpries"),
                &["enterprises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistakingly"),
                &["mistakenly"],
            ),
            (dictgen::InsensitiveStr::Ascii("opjected"), &["objected"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnostiocs"),
                &["diagnostics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appnedices"),
                &["appendices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sontrollers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("validat"), &["validate"]),
            (dictgen::InsensitiveStr::Ascii("suject"), &["subject"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelising"),
                &["parallelising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reupladad"),
                &["reupload", "reuploaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degenrating"),
                &["degenerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremenets"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remenissent"),
                &["reminiscent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterpirses"),
                &["enterprises"],
            ),
            (dictgen::InsensitiveStr::Ascii("injcetion"), &["injection"]),
            (
                dictgen::InsensitiveStr::Ascii("exceot"),
                &["except", "exempt"],
            ),
            (dictgen::InsensitiveStr::Ascii("haviest"), &["heaviest"]),
            (
                dictgen::InsensitiveStr::Ascii("inidcating"),
                &["indicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mmbers"), &["members"]),
            (
                dictgen::InsensitiveStr::Ascii("introsepectable"),
                &["introspectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ipmrovement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("efficieint"), &["efficient"]),
            (
                dictgen::InsensitiveStr::Ascii("plaestinians"),
                &["palestinians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpats"),
                &["counterparts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delivermode"),
                &["deliverymode"],
            ),
            (dictgen::InsensitiveStr::Ascii("competeted"), &["competed"]),
            (dictgen::InsensitiveStr::Ascii("elloitt"), &["elliott"]),
            (dictgen::InsensitiveStr::Ascii("crtically"), &["critically"]),
            (
                dictgen::InsensitiveStr::Ascii("alllocation"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("divrese"), &["diverse"]),
            (dictgen::InsensitiveStr::Ascii("alimunum"), &["aluminum"]),
            (dictgen::InsensitiveStr::Ascii("sopohmore"), &["sophomore"]),
            (
                dictgen::InsensitiveStr::Ascii("foreignese"),
                &["foreigners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comferting"),
                &["comforting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressumable"),
                &["presumable"],
            ),
            (dictgen::InsensitiveStr::Ascii("validitity"), &["validity"]),
            (
                dictgen::InsensitiveStr::Ascii("exhasuting"),
                &["exhausting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oblitarated"),
                &["obliterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsetp"), &["offset"]),
            (dictgen::InsensitiveStr::Ascii("extrac"), &["extract"]),
            (dictgen::InsensitiveStr::Ascii("scriopts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("caluculating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aglorithms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("weither"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("convexe"),
                &["convex", "convexes"],
            ),
            (dictgen::InsensitiveStr::Ascii("momement"), &["moment"]),
            (dictgen::InsensitiveStr::Ascii("stricty"), &["strictly"]),
            (
                dictgen::InsensitiveStr::Ascii("supsenseful"),
                &["suspenseful"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgarding"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("apacolypse"),
                &["apocalypse"],
            ),
            (dictgen::InsensitiveStr::Ascii("splaton"), &["splatoon"]),
            (
                dictgen::InsensitiveStr::Ascii("incorporte"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalnet"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("lieing"), &["lying"]),
            (
                dictgen::InsensitiveStr::Ascii("missisippi"),
                &["mississippi"],
            ),
            (dictgen::InsensitiveStr::Ascii("ininitely"), &["infinitely"]),
            (dictgen::InsensitiveStr::Ascii("octogons"), &["octagons"]),
            (dictgen::InsensitiveStr::Ascii("copmiler"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Unicode("сontaining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appericates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("exampples"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("belivable"), &["believable"]),
            (dictgen::InsensitiveStr::Ascii("aligmnet"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("superfical"),
                &["superficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("conpletes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("interpolaiton"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dragable"), &["draggable"]),
            (
                dictgen::InsensitiveStr::Ascii("oscillater"),
                &["oscillator", "oscillated", "oscillates", "oscillate"],
            ),
            (dictgen::InsensitiveStr::Ascii("witnesess"), &["witnesses"]),
            (
                dictgen::InsensitiveStr::Ascii("achievemints"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("liekable"), &["likable"]),
            (
                dictgen::InsensitiveStr::Ascii("writte"),
                &["write", "written"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elipsises"),
                &["ellipses", "ellipsis"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutsom"), &["custom"]),
            (
                dictgen::InsensitiveStr::Ascii("objectificaiton"),
                &["objectification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onmishience"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatbases"), &["databases"]),
            (dictgen::InsensitiveStr::Ascii("orignates"), &["originates"]),
            (
                dictgen::InsensitiveStr::Ascii("exaggarating"),
                &["exaggerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("proocessed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("retyrable"), &["retryable"]),
            (
                dictgen::InsensitiveStr::Ascii("necessaraly"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("incmrement"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("lmitations"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("strart"), &["start"]),
            (dictgen::InsensitiveStr::Ascii("legenday"), &["legendary"]),
            (
                dictgen::InsensitiveStr::Ascii("calclulators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("coupel"), &["couple"]),
            (
                dictgen::InsensitiveStr::Ascii("oppresssing"),
                &["oppressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tese"),
                &["these", "tease", "terse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competiters"),
                &["competitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("volatage"), &["voltage"]),
            (
                dictgen::InsensitiveStr::Ascii("harvasting"),
                &["harvesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("instealls"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("doorjam"), &["doorjamb"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomftorably"),
                &["uncomfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("assocaite"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("separing"), &["separating"]),
            (
                dictgen::InsensitiveStr::Ascii("sutisfying"),
                &["satisfying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insinuationg"),
                &["insinuating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associetions"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("challege"), &["challenge"]),
            (
                dictgen::InsensitiveStr::Ascii("counterfeight"),
                &["counterfeit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifiiers"),
                &["classifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("involantary"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("becuse"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("intitial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("shfiting"), &["shifting"]),
            (dictgen::InsensitiveStr::Ascii("internetes"), &["internets"]),
            (
                dictgen::InsensitiveStr::Ascii("admininstrative"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enligthening"),
                &["enlightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interelaved"),
                &["interleaved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discraceful"),
                &["disgraceful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consluting"),
                &["consulting"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribte"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("annoucners"),
                &["announcers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effctiveness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depedences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("targting"), &["targeting"]),
            (dictgen::InsensitiveStr::Ascii("reevaled"), &["revealed"]),
            (dictgen::InsensitiveStr::Ascii("couresty"), &["courtesy"]),
            (dictgen::InsensitiveStr::Ascii("labratory"), &["laboratory"]),
            (dictgen::InsensitiveStr::Ascii("procdures"), &["procedures"]),
            (
                dictgen::InsensitiveStr::Ascii("universitets"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("smoehow"), &["somehow"]),
            (dictgen::InsensitiveStr::Ascii("visbility"), &["visibility"]),
            (
                dictgen::InsensitiveStr::Ascii("bimillenial"),
                &["bimillennial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimptomatic"),
                &["symptomatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexcpected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omishiences"),
                &["omnisciences"],
            ),
            (dictgen::InsensitiveStr::Ascii("plaform"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("emabaroged"), &["embargoed"]),
            (dictgen::InsensitiveStr::Ascii("colunm"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("turncation"),
                &["truncation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinsitically"),
                &["deterministically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocasion"), &["occasion"]),
            (dictgen::InsensitiveStr::Ascii("opereating"), &["operating"]),
            (dictgen::InsensitiveStr::Ascii("cateogory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("colection"), &["collection"]),
            (dictgen::InsensitiveStr::Ascii("loobacks"), &["loopbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistendt"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfotunate"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvottes"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envrironmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyour"), &["your", "tour"]),
            (dictgen::InsensitiveStr::Ascii("indains"), &["indians"]),
            (
                dictgen::InsensitiveStr::Ascii("despatchers"),
                &["dispatchers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmunicated"),
                &["communicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimalt"), &["minimalist"]),
            (dictgen::InsensitiveStr::Ascii("delimtier"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("philophsy"), &["philosophy"]),
            (dictgen::InsensitiveStr::Ascii("timetamp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("repsect"), &["respect"]),
            (dictgen::InsensitiveStr::Ascii("sampe"), &["sample"]),
            (dictgen::InsensitiveStr::Ascii("columnas"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("serialsied"),
                &["serialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undetstand"),
                &["understand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touranments"),
                &["tournaments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonsignificant"),
                &["insignificant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colume"),
                &["column", "volume"],
            ),
            (dictgen::InsensitiveStr::Ascii("africain"), &["african"]),
            (dictgen::InsensitiveStr::Ascii("abtracts"), &["abstracts"]),
            (
                dictgen::InsensitiveStr::Ascii("frankenstiner"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protoytpes"),
                &["prototypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("eqution"), &["equation"]),
            (
                dictgen::InsensitiveStr::Ascii("assimiliate"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("trafic"), &["traffic"]),
            (dictgen::InsensitiveStr::Ascii("runnner"), &["runner"]),
            (
                dictgen::InsensitiveStr::Ascii("authroitative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governmetn"),
                &["governments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bounaaries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contined"),
                &["continued", "contained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraodrinary"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyswords"), &["keywords"]),
            (dictgen::InsensitiveStr::Ascii("separartes"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("simular"),
                &["similar", "simulator", "singular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attendence"),
                &["attendance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charismastic"),
                &["charismatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expereinces"),
                &["experiences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breastfeading"),
                &["breastfeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blackhwaks"),
                &["blackhawks"],
            ),
            (dictgen::InsensitiveStr::Ascii("diveristy"), &["diversity"]),
            (
                dictgen::InsensitiveStr::Ascii("deteched"),
                &["detached", "detected"],
            ),
            (dictgen::InsensitiveStr::Ascii("behaiviuor"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("miinimizes"), &["minimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("experienshial"),
                &["experiential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penetracion"),
                &["penetration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neightobrhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interacion"),
                &["interaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplities"),
                &["simplicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("muder"), &["murder"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegded"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlsations"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("divinitiy"), &["divinity"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisscism"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("depnds"), &["depends"]),
            (dictgen::InsensitiveStr::Ascii("fulfiling"), &["fulfilling"]),
            (dictgen::InsensitiveStr::Ascii("cobines"), &["combines"]),
            (
                dictgen::InsensitiveStr::Ascii("kidknappees"),
                &["kidnappees"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciffier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("functionnal"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highligjtes"),
                &["highlights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarmable"),
                &["programmable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsturcting"),
                &["reconstructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("emphaised"), &["emphasised"]),
            (dictgen::InsensitiveStr::Ascii("zionistisk"), &["zionists"]),
            (dictgen::InsensitiveStr::Ascii("verboase"), &["verbose"]),
            (
                dictgen::InsensitiveStr::Ascii("initialzers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("relpace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("intilizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivlanets"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprixamates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("thinik"), &["think"]),
            (dictgen::InsensitiveStr::Ascii("pulishers"), &["publishers"]),
            (dictgen::InsensitiveStr::Ascii("locaiting"), &["locating"]),
            (dictgen::InsensitiveStr::Ascii("generaste"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("psaswd"), &["passwd"]),
            (
                dictgen::InsensitiveStr::Ascii("detmatologist"),
                &["dermatologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterm"),
                &["term", "item", "intern"],
            ),
            (dictgen::InsensitiveStr::Ascii("coninc"), &["conic"]),
            (dictgen::InsensitiveStr::Ascii("forciblly"), &["forcibly"]),
            (dictgen::InsensitiveStr::Ascii("depreate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("concentrade"),
                &["concentrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bottlenect"),
                &["bottleneck"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interaccion"),
                &["interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoudln"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("pefered"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("detcting"), &["detecting"]),
            (
                dictgen::InsensitiveStr::Ascii("parctically"),
                &["practically"],
            ),
            (dictgen::InsensitiveStr::Ascii("enourmous"), &["enormous"]),
            (dictgen::InsensitiveStr::Ascii("streamade"), &["streamed"]),
            (
                dictgen::InsensitiveStr::Ascii("occurances"),
                &["occurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pattersn"),
                &["patterson", "patterns"],
            ),
            (dictgen::InsensitiveStr::Ascii("shopuld"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("defin"), &["define"]),
            (
                dictgen::InsensitiveStr::Ascii("pecifier"),
                &["specifier", "pacifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("configl"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("diretions"), &["directions"]),
            (dictgen::InsensitiveStr::Ascii("croch"), &["crotch"]),
            (dictgen::InsensitiveStr::Ascii("creaets"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("potetinals"),
                &["potentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("stduents"), &["students"]),
            (
                dictgen::InsensitiveStr::Ascii("experimteners"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("sparkel"), &["sparkle"]),
            (
                dictgen::InsensitiveStr::Ascii("asscociating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("producerats"),
                &["producers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealiarates"),
                &["ameliorates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhhoods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("succsess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("dirving"), &["driving"]),
            (dictgen::InsensitiveStr::Ascii("flexable"), &["flexible"]),
            (
                dictgen::InsensitiveStr::Ascii("refenrenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("otheriwse"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("insersting"), &["inserting"]),
            (
                dictgen::InsensitiveStr::Ascii("expeiriments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepolate"),
                &["interpolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("artistc"), &["artistic"]),
            (dictgen::InsensitiveStr::Ascii("somhow"), &["somehow"]),
            (
                dictgen::InsensitiveStr::Ascii("tansactional"),
                &["transactional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasitional"),
                &["transitional", "traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("anyhwere"), &["anywhere"]),
            (
                dictgen::InsensitiveStr::Ascii("manetainers"),
                &["maintainers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongizes"),
                &["recognizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("validaing"), &["validating"]),
            (dictgen::InsensitiveStr::Ascii("hounour"), &["honour"]),
            (
                dictgen::InsensitiveStr::Ascii("compatator"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("turaround"), &["turnaround"]),
            (
                dictgen::InsensitiveStr::Ascii("incrmeneted"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("observaiton"),
                &["observant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberbuility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("montanya"), &["montana"]),
            (dictgen::InsensitiveStr::Ascii("fontier"), &["frontier"]),
            (
                dictgen::InsensitiveStr::Ascii("reselction"),
                &["reselection"],
            ),
            (dictgen::InsensitiveStr::Ascii("intialze"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("frequenct"),
                &["frequency", "frequent"],
            ),
            (dictgen::InsensitiveStr::Ascii("smackdwon"), &["smackdown"]),
            (
                dictgen::InsensitiveStr::Ascii("tansitioned"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("chinmey"), &["chimney"]),
            (dictgen::InsensitiveStr::Ascii("narural"), &["natural"]),
            (
                dictgen::InsensitiveStr::Ascii("impovershied"),
                &["impoverished"],
            ),
            (dictgen::InsensitiveStr::Ascii("steorid"), &["steroid"]),
            (dictgen::InsensitiveStr::Ascii("synopsies"), &["synopsis"]),
            (
                dictgen::InsensitiveStr::Ascii("serinomies"),
                &["ceremonies"],
            ),
            (dictgen::InsensitiveStr::Ascii("enginges"), &["engines"]),
            (
                dictgen::InsensitiveStr::Ascii("procriation"),
                &["procreation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announched"),
                &["announce", "announced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronisation"),
                &["synchronisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilzing"), &["utilizing"]),
            (
                dictgen::InsensitiveStr::Ascii("instersectional"),
                &["intersectional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deteremind"),
                &["determined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhilated"),
                &["annihilated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interescting"),
                &["intersecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimptumaticly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceuctor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("definetaly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("cros"), &["cross"]),
            (
                dictgen::InsensitiveStr::Ascii("philadalphia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imtimidating"),
                &["intimidating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotaiations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulithread"),
                &["multithread"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compitetion"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropaching"),
                &["approaching"],
            ),
            (dictgen::InsensitiveStr::Ascii("uknowns"), &["unknowns"]),
            (dictgen::InsensitiveStr::Ascii("rennovated"), &["renovated"]),
            (dictgen::InsensitiveStr::Ascii("ordr"), &["order"]),
            (dictgen::InsensitiveStr::Ascii("occrrence"), &["occurrence"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitures"),
                &["prostitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sintacs"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("arrranged"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("psitoined"), &["positioned"]),
            (
                dictgen::InsensitiveStr::Ascii("verisoners"),
                &["versioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehnology"), &["technology"]),
            (dictgen::InsensitiveStr::Ascii("checcking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("officeals"), &["officials"]),
            (dictgen::InsensitiveStr::Ascii("ccomment"), &["comment"]),
            (
                dictgen::InsensitiveStr::Ascii("juxtifying"),
                &["justifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("thei"), &["their", "they"]),
            (
                dictgen::InsensitiveStr::Ascii("trriggering"),
                &["triggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("sgadow"), &["shadow"]),
            (dictgen::InsensitiveStr::Ascii("reaceive"), &["receive"]),
            (
                dictgen::InsensitiveStr::Ascii("equvalence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("enttry"), &["entry"]),
            (
                dictgen::InsensitiveStr::Ascii("intermedidate"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarenteedy"), &["guaranty"]),
            (
                dictgen::InsensitiveStr::Ascii("moderatorin"),
                &["moderation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventourus"),
                &["adventurous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strenghened"),
                &["strengthened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuent"),
                &["contingent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outlow"),
                &["outflow", "outlaw"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predective"),
                &["predictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjustificat"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulatenously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninterruped"),
                &["uninterrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("mericful"), &["merciful"]),
            (dictgen::InsensitiveStr::Ascii("requities"), &["requisites"]),
            (dictgen::InsensitiveStr::Ascii("rationel"), &["rationale"]),
            (dictgen::InsensitiveStr::Ascii("collpase"), &["collapse"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherizing"),
                &["characterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oficianados"),
                &["aficionados"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replasing"),
                &["replacing", "relapsing", "rephasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecxeptionally"),
                &["exceptionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recund"),
                &["fecund", "reckoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("squirrle"), &["squirrel"]),
            (dictgen::InsensitiveStr::Ascii("necessaary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("rute"),
                &["brute", "route", "rule"],
            ),
            (dictgen::InsensitiveStr::Ascii("murr"), &["myrrh"]),
            (dictgen::InsensitiveStr::Ascii("cofirming"), &["confirming"]),
            (
                dictgen::InsensitiveStr::Ascii("kyrptonite"),
                &["kryptonite"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraive"), &["arrive"]),
            (dictgen::InsensitiveStr::Ascii("tillts"), &["tilts"]),
            (
                dictgen::InsensitiveStr::Ascii("arithmetisch"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("meetign"), &["meeting"]),
            (
                dictgen::InsensitiveStr::Ascii("purpotedly"),
                &["purportedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomidating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("enouugh"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("rearrangd"), &["rearranged"]),
            (
                dictgen::InsensitiveStr::Ascii("invertibrates"),
                &["invertebrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("almighy"), &["almighty"]),
            (
                dictgen::InsensitiveStr::Ascii("appropiate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrepetitiously"),
                &["surreptitiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributon"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noification"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creationnist"),
                &["creationist"],
            ),
            (dictgen::InsensitiveStr::Ascii("instlaled"), &["installed"]),
            (
                dictgen::InsensitiveStr::Ascii("inpsecting"),
                &["inspecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomatters"), &["formatters"]),
            (dictgen::InsensitiveStr::Ascii("memers"), &["members"]),
            (dictgen::InsensitiveStr::Ascii("sructured"), &["structured"]),
            (dictgen::InsensitiveStr::Ascii("occasin"), &["occasion"]),
            (dictgen::InsensitiveStr::Ascii("radicalis"), &["radicals"]),
            (
                dictgen::InsensitiveStr::Ascii("atatchable"),
                &["attachable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrisponsible"),
                &["irresponsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("brighly"), &["brightly"]),
            (
                dictgen::InsensitiveStr::Ascii("libertariansm"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("inhibt"), &["inhibit"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborghs"),
                &["neighbors"],
            ),
            (dictgen::InsensitiveStr::Ascii("comsuming"), &["consuming"]),
            (dictgen::InsensitiveStr::Ascii("soovinear"), &["souvenir"]),
            (dictgen::InsensitiveStr::Ascii("succeess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("turkoise"), &["turquoise"]),
            (
                dictgen::InsensitiveStr::Ascii("destinatino"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("accouted"), &["accounted"]),
            (dictgen::InsensitiveStr::Ascii("mysterous"), &["mysterious"]),
            (dictgen::InsensitiveStr::Ascii("evalutae"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("pathcing"), &["patching"]),
            (
                dictgen::InsensitiveStr::Ascii("wighed"),
                &["weighed", "wicked"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequenzed"), &["sequenced"]),
            (
                dictgen::InsensitiveStr::Ascii("illitarate"),
                &["illiterate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termniations"),
                &["terminations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preccedences"),
                &["precedences"],
            ),
            (dictgen::InsensitiveStr::Ascii("pactches"), &["patches"]),
            (
                dictgen::InsensitiveStr::Ascii("refererencing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailsers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeging"), &["opening"]),
            (dictgen::InsensitiveStr::Ascii("checsums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("techicians"),
                &["technicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("columnns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("imporoves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("implenet"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("sisnce"), &["since"]),
            (dictgen::InsensitiveStr::Ascii("patienty"), &["patiently"]),
            (dictgen::InsensitiveStr::Ascii("diffculty"), &["difficulty"]),
            (dictgen::InsensitiveStr::Ascii("esimated"), &["estimated"]),
            (
                dictgen::InsensitiveStr::Ascii("efffectively"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steriotypes"),
                &["stereotypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnostisim"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("simuator"), &["simulator"]),
            (
                dictgen::InsensitiveStr::Ascii("mischeivous"),
                &["mischievous"],
            ),
            (dictgen::InsensitiveStr::Ascii("treasuers"), &["treasures"]),
            (
                dictgen::InsensitiveStr::Ascii("checkoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frowarding"),
                &["forwarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassinted"),
                &["assassinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajusting"), &["adjusting"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstandingly"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("disucssed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("convet"), &["convert"]),
            (dictgen::InsensitiveStr::Ascii("resoucing"), &["resourcing"]),
            (
                dictgen::InsensitiveStr::Ascii("controlelr"),
                &["controllers", "controller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumations"),
                &["enumerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerator"), &["creator"]),
            (dictgen::InsensitiveStr::Ascii("appeciate"), &["appreciate"]),
            (
                dictgen::InsensitiveStr::Ascii("euthanazia"),
                &["euthanasia"],
            ),
            (dictgen::InsensitiveStr::Ascii("senteces"), &["sentences"]),
            (
                dictgen::InsensitiveStr::Ascii("reevalutate"),
                &["reevaluate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributiong"),
                &["contribution", "contributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legendaryes"),
                &["legendaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychiatrics"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("astronat"), &["astronaut"]),
            (
                dictgen::InsensitiveStr::Ascii("affirmitave"),
                &["affirmative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifcation"),
                &["simplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("improts"), &["imports"]),
            (dictgen::InsensitiveStr::Ascii("refrehser"), &["refresher"]),
            (
                dictgen::InsensitiveStr::Ascii("avdisories"),
                &["advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("lamdba"), &["lambda"]),
            (dictgen::InsensitiveStr::Ascii("simpifier"), &["simplifier"]),
            (
                dictgen::InsensitiveStr::Ascii("simplfications"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("equvilent"), &["equivalent"]),
            (
                dictgen::InsensitiveStr::Ascii("concersation"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("covnertors"),
                &["convertors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialisazion"),
                &["deserialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifited"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaged"), &["gaged"]),
            (dictgen::InsensitiveStr::Ascii("checkcers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("adolecent"), &["adolescent"]),
            (dictgen::InsensitiveStr::Ascii("remanined"), &["remained"]),
            (dictgen::InsensitiveStr::Ascii("dpuble"), &["double"]),
            (dictgen::InsensitiveStr::Ascii("obstruced"), &["obstructed"]),
            (
                dictgen::InsensitiveStr::Ascii("calcluations"),
                &["calculation", "calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartrigde"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronised"),
                &["synchronised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architectureal"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismentled"),
                &["dismantled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretarem"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("peoms"), &["poems"]),
            (
                dictgen::InsensitiveStr::Ascii("cinematograhpy"),
                &["cinematography"],
            ),
            (dictgen::InsensitiveStr::Ascii("mcroscope"), &["microscope"]),
            (
                dictgen::InsensitiveStr::Ascii("acumulates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffferently"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("divisble"), &["divisible"]),
            (dictgen::InsensitiveStr::Ascii("diectives"), &["directives"]),
            (dictgen::InsensitiveStr::Ascii("stealthify"), &["stealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("relintquish"),
                &["relinquish"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiythout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("mucuous"), &["mucous"]),
            (dictgen::InsensitiveStr::Ascii("rehersal"), &["rehearsal"]),
            (
                dictgen::InsensitiveStr::Ascii("suscede"),
                &["secede", "succeed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conitional"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconnecters"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("temolo"), &["tremolo"]),
            (
                dictgen::InsensitiveStr::Ascii("understandebly"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshaks"), &["handshakes"]),
            (
                dictgen::InsensitiveStr::Ascii("recogniced"),
                &["recognised", "recognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("momments"), &["moments"]),
            (
                dictgen::InsensitiveStr::Ascii("weilding"),
                &["wielding", "welding"],
            ),
            (dictgen::InsensitiveStr::Ascii("alo"), &["also"]),
            (
                dictgen::InsensitiveStr::Ascii("peferentially"),
                &["preferentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrerat"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copulatin"),
                &["copulating", "copulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grammaticallity"),
                &["grammatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("retangles"), &["rectangles"]),
            (dictgen::InsensitiveStr::Ascii("deomcrat"), &["democrat"]),
            (
                dictgen::InsensitiveStr::Ascii("vulbearbilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("coninues"), &["continues"]),
            (dictgen::InsensitiveStr::Ascii("plataeu"), &["plateau"]),
            (
                dictgen::InsensitiveStr::Ascii("insipration"),
                &["inspiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("turtleh"), &["turtle"]),
            (
                dictgen::InsensitiveStr::Ascii("thunderbots"),
                &["thunderbolt"],
            ),
            (dictgen::InsensitiveStr::Ascii("initally"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("kewwords"), &["keywords"]),
            (dictgen::InsensitiveStr::Ascii("compier"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("figuratevely"),
                &["figuratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grammaticaly"),
                &["grammatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("facinated"), &["fascinated"]),
            (dictgen::InsensitiveStr::Ascii("featres"), &["features"]),
            (dictgen::InsensitiveStr::Ascii("revereted"), &["reverted"]),
            (dictgen::InsensitiveStr::Ascii("snytax"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("thatnked"), &["thanked"]),
            (dictgen::InsensitiveStr::Ascii("htlm"), &["html"]),
            (dictgen::InsensitiveStr::Ascii("setteing"), &["setting"]),
            (
                dictgen::InsensitiveStr::Ascii("reminscing"),
                &["reminiscing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stregnthening"),
                &["strengthening"],
            ),
            (dictgen::InsensitiveStr::Ascii("volatge"), &["voltage"]),
            (
                dictgen::InsensitiveStr::Ascii("rudimentally"),
                &["rudimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaginacion"),
                &["imagination"],
            ),
            (dictgen::InsensitiveStr::Ascii("alllocates"), &["allocates"]),
            (dictgen::InsensitiveStr::Ascii("botttom"), &["bottom"]),
            (
                dictgen::InsensitiveStr::Ascii("consumpion"),
                &["consumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critisicing"),
                &["criticizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("alligning"), &["aligning"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotiotations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("feaure"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("complisance"),
                &["complaisance", "compliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("californica"),
                &["california"],
            ),
            (dictgen::InsensitiveStr::Ascii("neet"), &["need", "neat"]),
            (dictgen::InsensitiveStr::Ascii("curevs"), &["curves"]),
            (
                dictgen::InsensitiveStr::Ascii("orginisations"),
                &["organisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("interacs"), &["interacts"]),
            (
                dictgen::InsensitiveStr::Ascii("resourcde"),
                &["resourced", "resource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arhitectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("girefing"), &["griefing"]),
            (dictgen::InsensitiveStr::Ascii("highligth"), &["highlight"]),
            (
                dictgen::InsensitiveStr::Ascii("productoin"),
                &["productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("perserves"), &["preserves"]),
            (dictgen::InsensitiveStr::Ascii("notidying"), &["notifying"]),
            (
                dictgen::InsensitiveStr::Ascii("stanardisation"),
                &["standardisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("printes"), &["printers"]),
            (
                dictgen::InsensitiveStr::Ascii("contraticting"),
                &["contradicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parellelise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("vanillia"), &["vanilla"]),
            (dictgen::InsensitiveStr::Ascii("enoung"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("disrispectful"),
                &["disrespectful"],
            ),
            (dictgen::InsensitiveStr::Ascii("copiing"), &["copying"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecassarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("personalitity"),
                &["personality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vinigarettes"),
                &["vinaigrettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("systemwindiow"),
                &["systemwindow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trogladite"),
                &["troglodyte"],
            ),
            (dictgen::InsensitiveStr::Ascii("similarily"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("mercusry"), &["mercury"]),
            (dictgen::InsensitiveStr::Ascii("genitalias"), &["genitals"]),
            (dictgen::InsensitiveStr::Ascii("restults"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("disassmbly"),
                &["disassembly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistencey"),
                &["consistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributers"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("deviiates"), &["deviates"]),
            (
                dictgen::InsensitiveStr::Ascii("cauilflower"),
                &["cauliflower"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genered"),
                &["generated", "gendered", "gender"],
            ),
            (dictgen::InsensitiveStr::Ascii("progamer"), &["programer"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpred"),
                &["misinterpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singatures"),
                &["signatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("secenarios"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("variabele"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("openion"), &["opinion"]),
            (
                dictgen::InsensitiveStr::Ascii("interanlly"),
                &["internally"],
            ),
            (dictgen::InsensitiveStr::Ascii("replaint"), &["repaint"]),
            (
                dictgen::InsensitiveStr::Ascii("terminting"),
                &["terminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("visibles"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("promotted"), &["promoted"]),
            (dictgen::InsensitiveStr::Ascii("keystokes"), &["keystrokes"]),
            (dictgen::InsensitiveStr::Ascii("idenifies"), &["identifies"]),
            (dictgen::InsensitiveStr::Ascii("lesda"), &["leads"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnparent"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("declair"), &["declare"]),
            (
                dictgen::InsensitiveStr::Ascii("cimptumatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("htmp"), &["html"]),
            (dictgen::InsensitiveStr::Ascii("neigher"), &["neither"]),
            (
                dictgen::InsensitiveStr::Ascii("flashflame"),
                &["flashframe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscelleaneous"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("orderering"), &["ordering"]),
            (dictgen::InsensitiveStr::Ascii("algoprithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("tendencias"),
                &["tendencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondnces"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("salying"), &["slaying"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelizes"),
                &["parallelizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("princesss"), &["princesses"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplimentation"),
                &["reimplementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("acturally"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("descides"), &["decides"]),
            (dictgen::InsensitiveStr::Ascii("interesed"), &["interested"]),
            (
                dictgen::InsensitiveStr::Ascii("stragger"),
                &["straggler", "stagger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reganerative"),
                &["regenerative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ukraininas"),
                &["ukrainians"],
            ),
            (dictgen::InsensitiveStr::Ascii("kepoints"), &["keypoints"]),
            (
                dictgen::InsensitiveStr::Ascii("palestenian"),
                &["palestinian"],
            ),
            (dictgen::InsensitiveStr::Ascii("appaers"), &["appears"]),
            (dictgen::InsensitiveStr::Ascii("unsiged"), &["unsigned"]),
            (dictgen::InsensitiveStr::Ascii("quadrdatic"), &["quadratic"]),
            (dictgen::InsensitiveStr::Ascii("engieneers"), &["engineers"]),
            (dictgen::InsensitiveStr::Ascii("managets"), &["managers"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpetctly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("updloads"), &["uploads"]),
            (dictgen::InsensitiveStr::Ascii("unpluggin"), &["unplugging"]),
            (
                dictgen::InsensitiveStr::Ascii("authencticated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("wikileakes"), &["wikileaks"]),
            (
                dictgen::InsensitiveStr::Ascii("customisatons"),
                &["customisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mircosoft"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("collaberation"),
                &["collaboration"],
            ),
            (dictgen::InsensitiveStr::Ascii("singify"), &["signify"]),
            (dictgen::InsensitiveStr::Ascii("advocay"), &["advocacy"]),
            (dictgen::InsensitiveStr::Ascii("evalueate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("languaces"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("develpments"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphetamins"),
                &["amphetamines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefention"),
                &["prevention"],
            ),
            (dictgen::InsensitiveStr::Ascii("angrilly"), &["angrily"]),
            (
                dictgen::InsensitiveStr::Ascii("integratin"),
                &["integrating", "integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbuted"),
                &["distribute", "distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experences"),
                &["experiences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hanidcapped"),
                &["handicapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("resart"), &["restart"]),
            (
                dictgen::InsensitiveStr::Ascii("pasring"),
                &["parsing", "pasting", "passing", "paring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monstrisity"),
                &["monstrosity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agreableness"),
                &["agreeableness"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpye"), &["type"]),
            (dictgen::InsensitiveStr::Ascii("profesor"), &["professor"]),
            (
                dictgen::InsensitiveStr::Ascii("liberalnim"),
                &["liberalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sesitively"),
                &["sensitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("ssociate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("corresopndence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("enginer"), &["engineer"]),
            (
                dictgen::InsensitiveStr::Ascii("vizualising"),
                &["visualising"],
            ),
            (dictgen::InsensitiveStr::Ascii("alphabeast"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("teleporation"),
                &["teleportation"],
            ),
            (dictgen::InsensitiveStr::Ascii("thruogh"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("prespectives"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nummerically"),
                &["numerically"],
            ),
            (dictgen::InsensitiveStr::Ascii("packag"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("genaric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("lokale"), &["locale"]),
            (dictgen::InsensitiveStr::Ascii("anlytic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("blokcs"),
                &["blocks", "blokes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("queres"),
                &["queries", "queers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parenthsis"),
                &["parenthesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("foundaries"), &["foundries"]),
            (dictgen::InsensitiveStr::Ascii("miltary"), &["military"]),
            (
                dictgen::InsensitiveStr::Ascii("religeously"),
                &["religiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntructor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insantiation"),
                &["instantiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragrphs"), &["paragraphs"]),
            (dictgen::InsensitiveStr::Ascii("impied"), &["implied"]),
            (dictgen::InsensitiveStr::Ascii("devied"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliacated"),
                &["duplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("telegrah"), &["telegraph"]),
            (dictgen::InsensitiveStr::Ascii("exchage"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("errornous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("daemonified"),
                &["daemonised", "daemonized"],
            ),
            (dictgen::InsensitiveStr::Ascii("motoral"), &["motorola"]),
            (
                dictgen::InsensitiveStr::Ascii("calculatess"),
                &["calculated", "calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("joystic"), &["joystick"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinated"),
                &["terminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilitiy"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("initalizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("yatchs"), &["yachts"]),
            (dictgen::InsensitiveStr::Ascii("pefer"), &["prefer"]),
            (dictgen::InsensitiveStr::Ascii("cafeteira"), &["cafeteria"]),
            (
                dictgen::InsensitiveStr::Ascii("collpsible"),
                &["collapsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presbaterian"),
                &["presbyterian"],
            ),
            (dictgen::InsensitiveStr::Ascii("divde"), &["divide"]),
            (
                dictgen::InsensitiveStr::Ascii("permmissions"),
                &["permissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenfitify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("arbiraty"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("finacial"), &["financial"]),
            (dictgen::InsensitiveStr::Ascii("protocal"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("sylablle"), &["syllable"]),
            (
                dictgen::InsensitiveStr::Ascii("methodoloy"),
                &["methodology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooresponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trunacting"),
                &["truncating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undergradate"),
                &["undergraduate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinateion"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("fufill"), &["fulfill"]),
            (dictgen::InsensitiveStr::Ascii("lockingf"), &["locking"]),
            (
                dictgen::InsensitiveStr::Ascii("solification"),
                &["solidification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architctural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("aleternate"), &["alternate"]),
            (dictgen::InsensitiveStr::Ascii("chaneges"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("pyscic"), &["psychic"]),
            (
                dictgen::InsensitiveStr::Ascii("exeptional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("disparite"), &["disparate"]),
            (
                dictgen::InsensitiveStr::Ascii("confrimations"),
                &["confirmations"],
            ),
            (dictgen::InsensitiveStr::Ascii("offxet"), &["offset"]),
            (
                dictgen::InsensitiveStr::Ascii("repoducibility"),
                &["reproducibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electronis"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferrably"),
                &["preferably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosseses"),
                &["processes", "possesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("natinals"), &["nationals"]),
            (
                dictgen::InsensitiveStr::Ascii("anphitheaters"),
                &["amphitheaters"],
            ),
            (dictgen::InsensitiveStr::Ascii("animatting"), &["animating"]),
            (dictgen::InsensitiveStr::Ascii("mabey"), &["maybe"]),
            (dictgen::InsensitiveStr::Ascii("dignosis"), &["diagnosis"]),
            (dictgen::InsensitiveStr::Ascii("usinng"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("wehn"), &["when"]),
            (dictgen::InsensitiveStr::Ascii("luch"), &["lunch"]),
            (dictgen::InsensitiveStr::Ascii("kidknap"), &["kidnap"]),
            (dictgen::InsensitiveStr::Ascii("anouncers"), &["announcers"]),
            (dictgen::InsensitiveStr::Ascii("tge"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("storse"), &["stores"]),
            (
                dictgen::InsensitiveStr::Ascii("reopsitory"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcpondingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrgue"), &["argue"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomdents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outperfomr"),
                &["outperform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indistuingishable"),
                &["indistinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("skwalk"), &["squawk"]),
            (
                dictgen::InsensitiveStr::Ascii("algorightms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thre"),
                &["three", "there", "their", "the"],
            ),
            (dictgen::InsensitiveStr::Ascii("fillaments"), &["filaments"]),
            (
                dictgen::InsensitiveStr::Ascii("sympathsize"),
                &["sympathize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("principalis"),
                &["principals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitited"),
                &["delimited"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprears"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("reparametrize"),
                &["reparameterize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documetation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumirem"),
                &["consumerism"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeueues"), &["queues"]),
            (dictgen::InsensitiveStr::Ascii("bigrading"), &["brigading"]),
            (
                dictgen::InsensitiveStr::Ascii("initiall"),
                &["initial", "initially"],
            ),
            (dictgen::InsensitiveStr::Ascii("kilometes"), &["kilometers"]),
            (
                dictgen::InsensitiveStr::Ascii("innoavtion"),
                &["innovations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versoining"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("featute"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("amphatheaters"),
                &["amphitheaters"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoiser"), &["customiser"]),
            (dictgen::InsensitiveStr::Ascii("longers"), &["longer"]),
            (dictgen::InsensitiveStr::Ascii("donwsides"), &["downsides"]),
            (
                dictgen::InsensitiveStr::Ascii("ambiguuousness"),
                &["ambiguousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("soldgered"), &["soldiered"]),
            (dictgen::InsensitiveStr::Ascii("runinng"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("usible"), &["usable"]),
            (
                dictgen::InsensitiveStr::Ascii("disastrious"),
                &["disastrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decompresion"),
                &["decompression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copilated"),
                &["compiled", "copulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheduler"), &["scheduler"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibililties"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalleing"), &["coalescing"]),
            (dictgen::InsensitiveStr::Ascii("datection"), &["detection"]),
            (dictgen::InsensitiveStr::Ascii("depostied"), &["deposited"]),
            (dictgen::InsensitiveStr::Ascii("eveolved"), &["evolved"]),
            (dictgen::InsensitiveStr::Ascii("regularlly"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("recuring"), &["recurring"]),
            (dictgen::InsensitiveStr::Ascii("dymanite"), &["dynamite"]),
            (dictgen::InsensitiveStr::Ascii("sabatoshed"), &["sabotaged"]),
            (dictgen::InsensitiveStr::Ascii("mascarra"), &["mascara"]),
            (dictgen::InsensitiveStr::Ascii("horizonal"), &["horizontal"]),
            (dictgen::InsensitiveStr::Ascii("priting"), &["printing"]),
            (dictgen::InsensitiveStr::Ascii("statubar"), &["statusbar"]),
            (dictgen::InsensitiveStr::Ascii("eastwoood"), &["eastwood"]),
            (dictgen::InsensitiveStr::Ascii("chari"), &["chair"]),
            (dictgen::InsensitiveStr::Ascii("attept"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("flavorade"), &["flavored"]),
            (
                dictgen::InsensitiveStr::Ascii("registored"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciaton"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vecotor"), &["vector"]),
            (
                dictgen::InsensitiveStr::Ascii("legalizacion"),
                &["legalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("ininite"), &["infinite"]),
            (dictgen::InsensitiveStr::Ascii("poitless"), &["pointless"]),
            (dictgen::InsensitiveStr::Ascii("numnbers"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("withh"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("slac"), &["slack"]),
            (
                dictgen::InsensitiveStr::Ascii("highighters"),
                &["highlighters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnparencies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("discarge"), &["discharge"]),
            (dictgen::InsensitiveStr::Ascii("netwroks"), &["networks"]),
            (dictgen::InsensitiveStr::Ascii("vetexes"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("intuitivly"),
                &["intuitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutsences"), &["cutscenes"]),
            (
                dictgen::InsensitiveStr::Ascii("attribting"),
                &["attributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proejction"),
                &["projection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triagulations"),
                &["triangulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("possabily"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("assignenmet"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("jailbrake"), &["jailbreak"]),
            (dictgen::InsensitiveStr::Ascii("strigify"), &["stringify"]),
            (dictgen::InsensitiveStr::Ascii("raotates"), &["rotates"]),
            (dictgen::InsensitiveStr::Ascii("clesius"), &["celsius"]),
            (dictgen::InsensitiveStr::Ascii("sommehow"), &["somehow"]),
            (dictgen::InsensitiveStr::Ascii("cusomizes"), &["customizes"]),
            (dictgen::InsensitiveStr::Ascii("beautyfull"), &["beautiful"]),
            (
                dictgen::InsensitiveStr::Ascii("coutnerpoint"),
                &["counterpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thiner"),
                &["thinner", "thinker"],
            ),
            (dictgen::InsensitiveStr::Ascii("entris"), &["entries"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentts"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("otain"), &["obtain"]),
            (
                dictgen::InsensitiveStr::Ascii("voxes"),
                &["voxels", "voxel"],
            ),
            (dictgen::InsensitiveStr::Ascii("devloping"), &["developing"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeurgh"),
                &["neighbour"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deattaching"),
                &["detaching"],
            ),
            (dictgen::InsensitiveStr::Ascii("overfow"), &["overflow"]),
            (
                dictgen::InsensitiveStr::Ascii("prefereable"),
                &["preferable"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenical"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("prevelance"),
                &["prevalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("unexpcted"), &["unexpected"]),
            (dictgen::InsensitiveStr::Ascii("riendeer"), &["reindeer"]),
            (
                dictgen::InsensitiveStr::Ascii("explaination"),
                &["explanation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alllready"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("carmonial"), &["ceremonial"]),
            (
                dictgen::InsensitiveStr::Ascii("poposing"),
                &["proposing", "opposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rescursively"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armagaddon"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trigggering"),
                &["triggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("collosal"), &["colossal"]),
            (dictgen::InsensitiveStr::Ascii("contaire"), &["containers"]),
            (
                dictgen::InsensitiveStr::Ascii("corsshairs"),
                &["crosshairs"],
            ),
            (dictgen::InsensitiveStr::Ascii("apologism"), &["apology"]),
            (
                dictgen::InsensitiveStr::Ascii("spellshecking"),
                &["spellchecking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifciates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congretation"),
                &["congregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrophie"),
                &["catastrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("anotating"), &["annotating"]),
            (dictgen::InsensitiveStr::Ascii("raidance"), &["radiance"]),
            (dictgen::InsensitiveStr::Ascii("medow"), &["meadow"]),
            (dictgen::InsensitiveStr::Ascii("mutlipes"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("gemetrical"),
                &["geometrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceted"), &["concede"]),
            (dictgen::InsensitiveStr::Ascii("caendars"), &["calendars"]),
            (dictgen::InsensitiveStr::Ascii("siblin"), &["sibling"]),
            (
                dictgen::InsensitiveStr::Ascii("definitivley"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminstered"),
                &["administered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defendernos"),
                &["defenders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("californinan"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconets"),
                &["disconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparoches"),
                &["approaches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homesexuality"),
                &["homosexuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterization"),
                &["parametrization", "parameterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("assesing"), &["assessing"]),
            (dictgen::InsensitiveStr::Ascii("pakistanti"), &["pakistani"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreations"),
                &["aggregations"],
            ),
            (dictgen::InsensitiveStr::Ascii("presnter"), &["presenter"]),
            (dictgen::InsensitiveStr::Ascii("opeartors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("anitques"), &["antiques"]),
            (dictgen::InsensitiveStr::Ascii("popolate"), &["populate"]),
            (dictgen::InsensitiveStr::Ascii("resrouces"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("turncations"),
                &["truncations"],
            ),
            (dictgen::InsensitiveStr::Ascii("alawys"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("socttish"), &["scottish"]),
            (
                dictgen::InsensitiveStr::Ascii("perminently"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thundebird"),
                &["thunderbird"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disenntangled"),
                &["disentangled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explnation"),
                &["explanation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcluates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cumbserome"),
                &["cumbersome"],
            ),
            (dictgen::InsensitiveStr::Ascii("embbeder"), &["embedder"]),
            (
                dictgen::InsensitiveStr::Ascii("midification"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempertures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadventage"),
                &["disadvantage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemmorhaging"),
                &["hemorrhaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanbdling"), &["handling"]),
            (dictgen::InsensitiveStr::Ascii("sydnicate"), &["syndicate"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuye"),
                &["contribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expierence"),
                &["experience"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomating"), &["automating"]),
            (
                dictgen::InsensitiveStr::Ascii("substaintially"),
                &["substantially"],
            ),
            (dictgen::InsensitiveStr::Ascii("downlowd"), &["download"]),
            (dictgen::InsensitiveStr::Ascii("commoditiy"), &["commodity"]),
            (dictgen::InsensitiveStr::Ascii("opaykes"), &["opaques"]),
            (dictgen::InsensitiveStr::Ascii("weirods"), &["weirdos"]),
            (dictgen::InsensitiveStr::Ascii("dictacting"), &["dictating"]),
            (
                dictgen::InsensitiveStr::Ascii("commndline"),
                &["commandline"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoning"), &["reasoning"]),
            (dictgen::InsensitiveStr::Ascii("dialgs"), &["dialogs"]),
            (
                dictgen::InsensitiveStr::Ascii("modifcication"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cartdriges"),
                &["cartridges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interscetion"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("telemetery"), &["telemetry"]),
            (dictgen::InsensitiveStr::Ascii("eample"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("friendy"), &["friendly"]),
            (
                dictgen::InsensitiveStr::Ascii("civalezation"),
                &["civilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("testestorone"),
                &["testosterone"],
            ),
            (dictgen::InsensitiveStr::Ascii("comamnded"), &["commanded"]),
            (dictgen::InsensitiveStr::Ascii("unealthy"), &["unhealthy"]),
            (dictgen::InsensitiveStr::Ascii("solderded"), &["soldered"]),
            (dictgen::InsensitiveStr::Ascii("succcesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("mulitipart"), &["multipart"]),
            (
                dictgen::InsensitiveStr::Ascii("motherboad"),
                &["motherboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("techinque"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("syntesised"),
                &["synthesised"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqires"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("diaplaying"),
                &["displaying"],
            ),
            (dictgen::InsensitiveStr::Ascii("aobut"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("secords"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("satsifies"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("overestimateing"),
                &["overestimating"],
            ),
            (dictgen::InsensitiveStr::Ascii("spannin"), &["spanning"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulisation"),
                &["virtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("luanching"), &["launching"]),
            (dictgen::InsensitiveStr::Ascii("capturd"), &["captured"]),
            (dictgen::InsensitiveStr::Ascii("amplifer"), &["amplifier"]),
            (
                dictgen::InsensitiveStr::Ascii("subsidiced"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutaes"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("submachne"), &["submachine"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecssary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("stdoud"), &["stdout"]),
            (dictgen::InsensitiveStr::Ascii("otpimally"), &["optimally"]),
            (dictgen::InsensitiveStr::Ascii("momoent"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("diassociates"),
                &["disassociates", "dissociates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalutaeing"),
                &["evaluating"],
            ),
            (dictgen::InsensitiveStr::Ascii("develoing"), &["developing"]),
            (
                dictgen::InsensitiveStr::Ascii("figuritively"),
                &["figuratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overheathing"),
                &["overheating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburgs"),
                &["cheeseburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singificance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sociaty"), &["society"]),
            (
                dictgen::InsensitiveStr::Ascii("envireonments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("exammple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("admininistratively"),
                &["administratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("realizin"), &["realizing"]),
            (dictgen::InsensitiveStr::Ascii("brodcast"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("inhertied"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("delgation"), &["delegation"]),
            (dictgen::InsensitiveStr::Ascii("alpabets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("similare"),
                &["similar", "similarly"],
            ),
            (dictgen::InsensitiveStr::Ascii("hydropobe"), &["hydrophobe"]),
            (dictgen::InsensitiveStr::Ascii("santitizer"), &["sanitizer"]),
            (
                dictgen::InsensitiveStr::Ascii("previuosly"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultanious"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("slcies"), &["slices"]),
            (
                dictgen::InsensitiveStr::Ascii("alternarively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("emergerd"), &["emerged"]),
            (dictgen::InsensitiveStr::Ascii("otho"), &["otoh"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhythmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("toughtly"), &["tightly"]),
            (dictgen::InsensitiveStr::Ascii("greneer"), &["greener"]),
            (dictgen::InsensitiveStr::Ascii("relyably"), &["reliably"]),
            (
                dictgen::InsensitiveStr::Ascii("vigoruosly"),
                &["vigorously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tralling"),
                &[
                    "thralling",
                    "trailing",
                    "travelling",
                    "trialling",
                    "trilling",
                    "trolling",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("inpsire"), &["inspire"]),
            (
                dictgen::InsensitiveStr::Ascii("scrings"),
                &["strings", "springs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southerers"),
                &["southerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paragraphes"),
                &["paragraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("avtivity"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("garphic"), &["graphic"]),
            (dictgen::InsensitiveStr::Ascii("apprentie"), &["apprentice"]),
            (dictgen::InsensitiveStr::Ascii("inspireras"), &["inspires"]),
            (dictgen::InsensitiveStr::Ascii("girafes"), &["giraffes"]),
            (
                dictgen::InsensitiveStr::Ascii("renovatin"),
                &["renovating", "renovation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diagonstics"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("crpytic"), &["cryptic"]),
            (
                dictgen::InsensitiveStr::Ascii("superhereos"),
                &["superhero"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imigration"),
                &["emigration", "immigration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bounduaries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleetness"),
                &["completeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("normalizd"), &["normalized"]),
            (dictgen::InsensitiveStr::Ascii("initting"), &["initing"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbarte"),
                &["masturbate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("screenaver"),
                &["screensaver"],
            ),
            (dictgen::InsensitiveStr::Ascii("styiles"), &["styles"]),
            (
                dictgen::InsensitiveStr::Ascii("troglodites"),
                &["troglodytes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fltered"),
                &["filtered", "faltered", "fluttered", "flattered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opportunies"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerabely"),
                &["considerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("vavlue"), &["value"]),
            (
                dictgen::InsensitiveStr::Ascii("conciderate"),
                &["considerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("impule"), &["impulse"]),
            (dictgen::InsensitiveStr::Ascii("flitered"), &["filtered"]),
            (
                dictgen::InsensitiveStr::Ascii("suceessive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("neightobrs"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("untilities"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("bouregois"), &["bourgeois"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangaelbe"),
                &["rearrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("vill"), &["will"]),
            (dictgen::InsensitiveStr::Ascii("temove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("caricatore"),
                &["caricature"],
            ),
            (dictgen::InsensitiveStr::Ascii("morg"), &["morgue"]),
            (
                dictgen::InsensitiveStr::Ascii("implicictly"),
                &["implicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfucntion"),
                &["dysfunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dredential"),
                &["credential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teory"),
                &["theory", "tory", "teary"],
            ),
            (dictgen::InsensitiveStr::Ascii("shorctut"), &["shortcut"]),
            (
                dictgen::InsensitiveStr::Ascii("civalesations"),
                &["civilisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susceptable"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("freiting"), &["freighting"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecesary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signleplayer"),
                &["singleplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("commandi"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("neiter"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("progamed"), &["programed"]),
            (dictgen::InsensitiveStr::Ascii("emasc"), &["emacs"]),
            (dictgen::InsensitiveStr::Ascii("spefififes"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanentations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicsloses"), &["discloses"]),
            (dictgen::InsensitiveStr::Ascii("hikin"), &["hiking"]),
            (
                dictgen::InsensitiveStr::Ascii("definitlry"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previuously"),
                &["previously"],
            ),
            (dictgen::InsensitiveStr::Ascii("utputs"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("trialer"), &["trailer"]),
            (dictgen::InsensitiveStr::Ascii("avioded"), &["avoided"]),
            (dictgen::InsensitiveStr::Ascii("firmnware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("branchces"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("johanine"), &["johannine"]),
            (
                dictgen::InsensitiveStr::Ascii("accurences"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("desaster"), &["disaster"]),
            (dictgen::InsensitiveStr::Ascii("disastros"), &["disastrous"]),
            (dictgen::InsensitiveStr::Ascii("theather"), &["theater"]),
            (
                dictgen::InsensitiveStr::Ascii("sugguestions"),
                &["suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comprension"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("onging"), &["ongoing"]),
            (dictgen::InsensitiveStr::Ascii("amendmets"), &["amendments"]),
            (dictgen::InsensitiveStr::Ascii("loclaize"), &["localize"]),
            (dictgen::InsensitiveStr::Ascii("maylasia"), &["malaysia"]),
            (dictgen::InsensitiveStr::Ascii("identital"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("franchices"),
                &["franchises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foreruners"),
                &["forerunners"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwardsm"), &["forwards"]),
            (
                dictgen::InsensitiveStr::Ascii("orchestrial"),
                &["orchestra"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specilization"),
                &["specialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissino"),
                &["permission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transaciton"),
                &["transactions", "transaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingreidents"),
                &["ingredient"],
            ),
            (dictgen::InsensitiveStr::Ascii("instrall"), &["install"]),
            (
                dictgen::InsensitiveStr::Ascii("unefficiency"),
                &["inefficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("cimptum"), &["symptom"]),
            (
                dictgen::InsensitiveStr::Ascii("connvertions"),
                &["conversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("carmtan"), &["cartman"]),
            (dictgen::InsensitiveStr::Ascii("prohpecy"), &["prophecy"]),
            (dictgen::InsensitiveStr::Ascii("atteched"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("notario"), &["ontario"]),
            (dictgen::InsensitiveStr::Ascii("widghet"), &["widget"]),
            (
                dictgen::InsensitiveStr::Ascii("annonceurs"),
                &["announcers"],
            ),
            (dictgen::InsensitiveStr::Ascii("actualoy"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("annotatin"),
                &["annotating", "annotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwinwg"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("healtchare"),
                &["healthcare"],
            ),
            (dictgen::InsensitiveStr::Ascii("swiflty"), &["swiftly"]),
            (dictgen::InsensitiveStr::Ascii("retalier"), &["retailer"]),
            (
                dictgen::InsensitiveStr::Ascii("westmnister"),
                &["westminster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegetarain"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("alllows"), &["allows"]),
            (dictgen::InsensitiveStr::Ascii("precice"), &["precise"]),
            (
                dictgen::InsensitiveStr::Ascii("fabircation"),
                &["fabrication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitifying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("oscilator"), &["oscillator"]),
            (
                dictgen::InsensitiveStr::Ascii("painkilllers"),
                &["painkillers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconeections"),
                &["reconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mecanically"),
                &["mechanically"],
            ),
            (dictgen::InsensitiveStr::Ascii("organismen"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomfertably"),
                &["uncomfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkbooks"), &["workbooks"]),
            (dictgen::InsensitiveStr::Ascii("arraies"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("onstant"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("udnercut"), &["undercut"]),
            (dictgen::InsensitiveStr::Ascii("deatils"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("surgested"), &["suggested"]),
            (
                dictgen::InsensitiveStr::Ascii("executation"),
                &["execution"],
            ),
            (dictgen::InsensitiveStr::Ascii("deocdings"), &["decodings"]),
            (dictgen::InsensitiveStr::Ascii("noy"), &["not"]),
            (dictgen::InsensitiveStr::Ascii("literla"), &["literal"]),
            (dictgen::InsensitiveStr::Ascii("termindate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("wonderes"), &["wonders"]),
            (dictgen::InsensitiveStr::Ascii("strawbery"), &["strawberry"]),
            (
                dictgen::InsensitiveStr::Ascii("propertary"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("providee"), &["providence"]),
            (dictgen::InsensitiveStr::Ascii("tigger"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("miniumum"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("wonderfuly"),
                &["wonderfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("addiotions"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("wohle"), &["whole"]),
            (
                dictgen::InsensitiveStr::Ascii("incriments"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporarilly"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strack"),
                &["stack", "track"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sythensize"),
                &["synthesize"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialetcs"), &["dialects"]),
            (
                dictgen::InsensitiveStr::Ascii("ellpitical"),
                &["elliptical"],
            ),
            (dictgen::InsensitiveStr::Ascii("picnick"), &["picnic"]),
            (
                dictgen::InsensitiveStr::Ascii("embarrasement"),
                &["embarrassment"],
            ),
            (dictgen::InsensitiveStr::Ascii("chisling"), &["chiseling"]),
            (
                dictgen::InsensitiveStr::Ascii("differenciates"),
                &["differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhooods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("strerrror"), &["strerror"]),
            (
                dictgen::InsensitiveStr::Ascii("asscociated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("samarai"), &["samurai"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitalised"),
                &["uninitialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("annointed"), &["anointed"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerbilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conceviable"),
                &["conceivable"],
            ),
            (dictgen::InsensitiveStr::Ascii("misued"), &["misused"]),
            (
                dictgen::InsensitiveStr::Ascii("fituring"),
                &["figuring", "featuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analystic"),
                &["analytic", "analyst"],
            ),
            (dictgen::InsensitiveStr::Ascii("knarl"), &["gnarl"]),
            (
                dictgen::InsensitiveStr::Ascii("persenting"),
                &["presenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tablesppons"),
                &["tablespoons"],
            ),
            (dictgen::InsensitiveStr::Ascii("roelplay"), &["roleplay"]),
            (dictgen::InsensitiveStr::Ascii("plagarism"), &["plagiarism"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitae"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounnds"), &["bounds"]),
            (dictgen::InsensitiveStr::Ascii("enhabce"), &["enhance"]),
            (
                dictgen::InsensitiveStr::Ascii("disscussing"),
                &["discussing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucesfully"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("decimls"), &["decimals"]),
            (
                dictgen::InsensitiveStr::Ascii("mehcanically"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizaiton"),
                &["organization"],
            ),
            (dictgen::InsensitiveStr::Ascii("generalbs"), &["generals"]),
            (dictgen::InsensitiveStr::Ascii("debiab"), &["debian"]),
            (
                dictgen::InsensitiveStr::Ascii("passthough"),
                &["passthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("implenment"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("optiopn"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("segmente"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("insonmia"), &["insomnia"]),
            (dictgen::InsensitiveStr::Ascii("rectanlge"), &["rectangle"]),
            (dictgen::InsensitiveStr::Ascii("slipperies"), &["slippers"]),
            (dictgen::InsensitiveStr::Ascii("exipred"), &["expired"]),
            (dictgen::InsensitiveStr::Ascii("chainges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("tempororarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("declairs"), &["declares"]),
            (dictgen::InsensitiveStr::Ascii("sligntly"), &["slightly"]),
            (
                dictgen::InsensitiveStr::Ascii("finctioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoud"), &["avoid"]),
            (dictgen::InsensitiveStr::Ascii("targered"), &["targeted"]),
            (dictgen::InsensitiveStr::Ascii("martain"), &["martian"]),
            (dictgen::InsensitiveStr::Ascii("oppisite"), &["opposite"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisatin"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipualted"),
                &["manipulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("naximal"), &["maximal"]),
            (
                dictgen::InsensitiveStr::Ascii("satisfacory"),
                &["satisfactory"],
            ),
            (dictgen::InsensitiveStr::Ascii("spirites"), &["spirits"]),
            (dictgen::InsensitiveStr::Ascii("respnd"), &["respond"]),
            (
                dictgen::InsensitiveStr::Ascii("javsscript"),
                &["javascript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginized"),
                &["recognized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selecctions"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmples"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("afghanastan"),
                &["afghanistan"],
            ),
            (dictgen::InsensitiveStr::Ascii("resistes"), &["resists"]),
            (
                dictgen::InsensitiveStr::Ascii("saclar"),
                &["scalar", "sacral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congratualtes"),
                &["congratulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("zookeinei"), &["zucchinis"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finialization"),
                &["finalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authendicate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("creedence"), &["credence"]),
            (
                dictgen::InsensitiveStr::Ascii("combintion"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("lowecase"), &["lowercase"]),
            (
                dictgen::InsensitiveStr::Ascii("repeted"),
                &["repeated", "reputed", "repented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifiantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("analists"), &["analysts"]),
            (
                dictgen::InsensitiveStr::Ascii("campatible"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complusion"),
                &["compulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomred"), &["performed"]),
            (
                dictgen::InsensitiveStr::Ascii("tabulater"),
                &["tabulator", "tabulated", "tabulates", "tabulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("drats"), &["drafts"]),
            (dictgen::InsensitiveStr::Ascii("setitngs"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("supplie"),
                &["supplied", "supplier", "supply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regresssion"),
                &["regression"],
            ),
            (dictgen::InsensitiveStr::Ascii("timzone"), &["timezone"]),
            (dictgen::InsensitiveStr::Ascii("eneded"), &["ended"]),
            (
                dictgen::InsensitiveStr::Ascii("accumualted"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("substem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("unerstanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("manualy"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("funciotnalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpatuate"),
                &["perpetuate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("newletters"),
                &["newsletters"],
            ),
            (dictgen::InsensitiveStr::Ascii("beng"), &["being"]),
            (dictgen::InsensitiveStr::Ascii("parasitter"), &["parasite"]),
            (dictgen::InsensitiveStr::Ascii("lpatform"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("dangereous"), &["dangerous"]),
            (
                dictgen::InsensitiveStr::Ascii("ostrazised"),
                &["ostracised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarenteing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("complats"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("circumferencial"),
                &["circumferential"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispartiy"), &["disparity"]),
            (
                dictgen::InsensitiveStr::Ascii("inefficients"),
                &["inefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innefficient"),
                &["inefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anphetamines"),
                &["amphetamines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massachussets"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("genereral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("outoign"), &["outgoing"]),
            (
                dictgen::InsensitiveStr::Ascii("transcevier"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("waclott"), &["walcott"]),
            (
                dictgen::InsensitiveStr::Ascii("extremistisk"),
                &["extremists"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgn"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("comapnion"), &["companion"]),
            (dictgen::InsensitiveStr::Ascii("guarrentee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("internetest"),
                &["internets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilzers"),
                &["stabilizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorespondend"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responeding"),
                &["responding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analystically"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgrate"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("relaased"),
                &["released", "relapsed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outlcassed"),
                &["outclassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("entquired"), &["enquired"]),
            (
                dictgen::InsensitiveStr::Ascii("seperatism"),
                &["separatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("camoflage"), &["camouflage"]),
            (dictgen::InsensitiveStr::Ascii("animatiing"), &["animating"]),
            (
                dictgen::InsensitiveStr::Ascii("extenisbility"),
                &["extensibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("realiest"), &["earliest"]),
            (
                dictgen::InsensitiveStr::Ascii("interferece"),
                &["interference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anitbiotic"),
                &["antibiotic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlightment"),
                &["enlightenment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bliever"), &["believer"]),
            (dictgen::InsensitiveStr::Ascii("embeders"), &["embedders"]),
            (
                dictgen::InsensitiveStr::Ascii("expementelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ethincities"),
                &["ethnicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("overvize"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("olbigatory"),
                &["obligatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("dishonord"), &["dishonored"]),
            (dictgen::InsensitiveStr::Ascii("requiering"), &["requiring"]),
            (
                dictgen::InsensitiveStr::Ascii("eloctrolytes"),
                &["electrolytes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialisatons"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("velcoity"), &["velocity"]),
            (dictgen::InsensitiveStr::Ascii("isalnders"), &["islanders"]),
            (dictgen::InsensitiveStr::Ascii("defintely"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotionates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarrenty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("ebst"), &["best"]),
            (
                dictgen::InsensitiveStr::Ascii("connectiongs"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administors"),
                &["administrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disillation"),
                &["distillation"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifyed"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("somethinng"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherided"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("repeatitions"),
                &["repetitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parliamone"),
                &["parliament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("depeds"), &["depends"]),
            (dictgen::InsensitiveStr::Ascii("canaceling"), &["canceling"]),
            (
                dictgen::InsensitiveStr::Ascii("saturacion"),
                &["saturation"],
            ),
            (dictgen::InsensitiveStr::Ascii("retourn"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("whta"), &["what"]),
            (dictgen::InsensitiveStr::Ascii("caledar"), &["calendar"]),
            (
                dictgen::InsensitiveStr::Ascii("mispellings"),
                &["misspellings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferenser"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("peforated"), &["perforated"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhodds"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("barracs"), &["barracks"]),
            (
                dictgen::InsensitiveStr::Ascii("renderering"),
                &["rendering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartizator"),
                &["standardizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhasutive"),
                &["exhaustive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationalisim"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("chacacter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("indivuduals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciato"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("maturiry"), &["maturity"]),
            (dictgen::InsensitiveStr::Ascii("attachs"), &["attaches"]),
            (
                dictgen::InsensitiveStr::Ascii("sentimentul"),
                &["sentimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforcees"), &["enforces"]),
            (
                dictgen::InsensitiveStr::Ascii("liberacion"),
                &["liberation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicaters"),
                &["indicates", "indicators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regisrations"),
                &["registrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("squrared"), &["squared"]),
            (dictgen::InsensitiveStr::Ascii("optoin"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("enbrace"), &["embrace"]),
            (
                dictgen::InsensitiveStr::Ascii("recconeccting"),
                &["reconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighrourhood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remplacement"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugery"), &["surgery"]),
            (dictgen::InsensitiveStr::Ascii("tomoroow"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("enfocer"), &["enforcer"]),
            (
                dictgen::InsensitiveStr::Ascii("represenable"),
                &["representable"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomates"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstate"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allignmenterror"),
                &["alignmenterror"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumirse"),
                &["consumerism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convenietly"),
                &["conveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("epidsodes"), &["episodes"]),
            (dictgen::InsensitiveStr::Ascii("defensen"), &["defenseman"]),
            (dictgen::InsensitiveStr::Ascii("rysnc"), &["rsync"]),
            (
                dictgen::InsensitiveStr::Ascii("acocuntants"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("sripts"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("comprable"), &["comparable"]),
            (dictgen::InsensitiveStr::Ascii("xgetttext"), &["xgettext"]),
            (
                dictgen::InsensitiveStr::Ascii("hirarchies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spirituella"),
                &["spiritually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrcting"),
                &["restarting"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiaters"), &["waiters"]),
            (
                dictgen::InsensitiveStr::Ascii("operattions"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("wednesay"), &["wednesday"]),
            (dictgen::InsensitiveStr::Ascii("lloking"), &["looking"]),
            (dictgen::InsensitiveStr::Ascii("contrbute"), &["contribute"]),
            (
                dictgen::InsensitiveStr::Ascii("probablilistic"),
                &["probabilistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("estoniya"), &["estonia"]),
            (
                dictgen::InsensitiveStr::Ascii("certifificated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("assingned"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("spectifications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electromagntic"),
                &["electromagnetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("subrouine"), &["subroutine"]),
            (
                dictgen::InsensitiveStr::Ascii("virutalization"),
                &["virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("proporty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("decrments"), &["decrements"]),
            (dictgen::InsensitiveStr::Ascii("bumpded"), &["bumped"]),
            (
                dictgen::InsensitiveStr::Ascii("beautifullly"),
                &["beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoritorian"),
                &["authoritarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("ummatched"), &["unmatched"]),
            (dictgen::InsensitiveStr::Ascii("birsbane"), &["brisbane"]),
            (
                dictgen::InsensitiveStr::Ascii("capbailities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("consructs"), &["constructs"]),
            (dictgen::InsensitiveStr::Ascii("owrk"), &["work"]),
            (dictgen::InsensitiveStr::Ascii("idicates"), &["indicates"]),
            (dictgen::InsensitiveStr::Ascii("regsiter"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("interleaed"),
                &["interleaved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaliability"),
                &["unavailability"],
            ),
            (dictgen::InsensitiveStr::Ascii("sessoions"), &["sessions"]),
            (
                dictgen::InsensitiveStr::Ascii("homoegenous"),
                &["homogeneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intertaining"),
                &["entertaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readbility"),
                &["readability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canocicalized"),
                &["canonicalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somethingis"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoidence"), &["avoidance"]),
            (dictgen::InsensitiveStr::Ascii("standar"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("rectangluar"),
                &["rectangular"],
            ),
            (dictgen::InsensitiveStr::Ascii("magnited"), &["magnitude"]),
            (dictgen::InsensitiveStr::Ascii("presidet"), &["presidents"]),
            (dictgen::InsensitiveStr::Ascii("readig"), &["reading"]),
            (
                dictgen::InsensitiveStr::Ascii("requiriements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("placebro"), &["placebo"]),
            (dictgen::InsensitiveStr::Ascii("exercse"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("opimal"), &["optimal"]),
            (dictgen::InsensitiveStr::Ascii("develped"), &["developed"]),
            (dictgen::InsensitiveStr::Ascii("confgi"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("folloed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("wieth"), &["width"]),
            (dictgen::InsensitiveStr::Ascii("algined"), &["alined"]),
            (dictgen::InsensitiveStr::Ascii("cheeckers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("allert"), &["alert"]),
            (dictgen::InsensitiveStr::Ascii("anonimity"), &["anonymity"]),
            (
                dictgen::InsensitiveStr::Ascii("seperatable"),
                &["separable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("babysittting"),
                &["babysitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("depressie"), &["depressive"]),
            (
                dictgen::InsensitiveStr::Ascii("familiaries"),
                &["familiarize"],
            ),
            (dictgen::InsensitiveStr::Ascii("fufilling"), &["fulfilling"]),
            (dictgen::InsensitiveStr::Ascii("importnat"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("backprogagation"),
                &["backpropagation"],
            ),
            (dictgen::InsensitiveStr::Ascii("depndence"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("hotest"),
                &["hottest", "honest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conteneurs"),
                &["contenders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalution"),
                &["evaluation", "evolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entchantable"),
                &["enchantable"],
            ),
            (dictgen::InsensitiveStr::Ascii("constract"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("aktions"), &["actions"]),
            (
                dictgen::InsensitiveStr::Ascii("psychologicaly"),
                &["psychologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procesures"),
                &["procedures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amublances"),
                &["ambulances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udnerestimate"),
                &["underestimated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specififies"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("worket"), &["worker"]),
            (dictgen::InsensitiveStr::Ascii("contemplat"), &["contempt"]),
            (
                dictgen::InsensitiveStr::Ascii("coomenting"),
                &["commenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approperiate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concencrate"),
                &["concentrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncetain"), &["uncertain"]),
            (dictgen::InsensitiveStr::Ascii("arranent"), &["arrangement"]),
            (
                dictgen::InsensitiveStr::Ascii("progapator"),
                &["propagator"],
            ),
            (dictgen::InsensitiveStr::Ascii("transmmit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("paramerisation"),
                &["parameterisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("liscensing"), &["licensing"]),
            (dictgen::InsensitiveStr::Ascii("socailist"), &["socialist"]),
            (dictgen::InsensitiveStr::Ascii("languges"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("sargant"), &["sergeant"]),
            (dictgen::InsensitiveStr::Ascii("aligntment"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("shuoldnt"), &["shouldnt"]),
            (dictgen::InsensitiveStr::Ascii("narly"), &["gnarly"]),
            (dictgen::InsensitiveStr::Ascii("copoying"), &["copying"]),
            (
                dictgen::InsensitiveStr::Ascii("normls"),
                &["normals", "norms"],
            ),
            (dictgen::InsensitiveStr::Ascii("fiield"), &["field"]),
            (dictgen::InsensitiveStr::Ascii("marijuanna"), &["marijuana"]),
            (dictgen::InsensitiveStr::Ascii("recpetive"), &["receptive"]),
            (dictgen::InsensitiveStr::Ascii("concluse"), &["conclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("concantenation"),
                &["concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determenistic"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ampethamines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("chawk"), &["chalk"]),
            (dictgen::InsensitiveStr::Ascii("sematics"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("charatcers"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("belwo"), &["below"]),
            (
                dictgen::InsensitiveStr::Ascii("respectivelly"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("rectrangle"), &["rectangle"]),
            (
                dictgen::InsensitiveStr::Ascii("chansellor"),
                &["chancellor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rastersizer"),
                &["rasterizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("smaall"), &["small"]),
            (dictgen::InsensitiveStr::Ascii("natigate"), &["navigate"]),
            (dictgen::InsensitiveStr::Ascii("suppplied"), &["supplied"]),
            (dictgen::InsensitiveStr::Ascii("beoynce"), &["beyonce"]),
            (
                dictgen::InsensitiveStr::Ascii("regresssions"),
                &["regressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavivour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("privously"), &["previously"]),
            (dictgen::InsensitiveStr::Ascii("extrenaly"), &["externally"]),
            (dictgen::InsensitiveStr::Ascii("heartbeast"), &["heartbeat"]),
            (dictgen::InsensitiveStr::Ascii("dicates"), &["dictates"]),
            (dictgen::InsensitiveStr::Ascii("reoport"), &["report"]),
            (dictgen::InsensitiveStr::Ascii("serbain"), &["serbian"]),
            (
                dictgen::InsensitiveStr::Ascii("committters"),
                &["committers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contibutor"),
                &["contributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosume"),
                &["consume", "costume"],
            ),
            (dictgen::InsensitiveStr::Ascii("formater"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("sacrifise"), &["sacrifices"]),
            (
                dictgen::InsensitiveStr::Ascii("newslatter"),
                &["newsletter"],
            ),
            (dictgen::InsensitiveStr::Ascii("pthreds"), &["pthreads"]),
            (
                dictgen::InsensitiveStr::Ascii("definisions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("imlement"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("commmented"), &["commented"]),
            (dictgen::InsensitiveStr::Ascii("valdates"), &["validates"]),
            (dictgen::InsensitiveStr::Ascii("internetus"), &["internets"]),
            (
                dictgen::InsensitiveStr::Ascii("pairocheally"),
                &["parochially"],
            ),
            (dictgen::InsensitiveStr::Ascii("candiens"), &["candies"]),
            (dictgen::InsensitiveStr::Ascii("elsef"), &["elseif"]),
            (
                dictgen::InsensitiveStr::Ascii("reurn"),
                &["return", "rerun"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhing"),
                &["neighbouring"],
            ),
            (dictgen::InsensitiveStr::Ascii("eitehr"), &["either"]),
            (dictgen::InsensitiveStr::Ascii("numbbered"), &["numbered"]),
            (
                dictgen::InsensitiveStr::Ascii("athleticm"),
                &["athleticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("prescibes"), &["prescribes"]),
            (
                dictgen::InsensitiveStr::Ascii("consecuences"),
                &["consequences"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbbor"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("freeezers"), &["freezers"]),
            (
                dictgen::InsensitiveStr::Ascii("coniguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subrotuines"),
                &["subroutines"],
            ),
            (dictgen::InsensitiveStr::Ascii("canging"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("sandstrom"), &["sandstorm"]),
            (dictgen::InsensitiveStr::Ascii("abusrd"), &["absurd"]),
            (dictgen::InsensitiveStr::Ascii("structual"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("configuracion"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imsensitive"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimintal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merly"),
                &["merely", "formerly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buttom"),
                &["button", "bottom"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obliberated"),
                &["obliterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("higway"), &["highway"]),
            (
                dictgen::InsensitiveStr::Ascii("flaskbacks"),
                &["flashbacks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workaounds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("resisitor"), &["resistor"]),
            (dictgen::InsensitiveStr::Ascii("bumpt"), &["bump"]),
            (dictgen::InsensitiveStr::Ascii("loctation"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("equpped"), &["equipped"]),
            (dictgen::InsensitiveStr::Ascii("revist"), &["revisit"]),
            (
                dictgen::InsensitiveStr::Ascii("comapatible"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confgiurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acommodates"),
                &["accommodates"],
            ),
            (dictgen::InsensitiveStr::Ascii("memorizin"), &["memorizing"]),
            (dictgen::InsensitiveStr::Ascii("adjatates"), &["agitates"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginzing"),
                &["recognizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streight"),
                &["straight", "strait"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dilligently"),
                &["diligently"],
            ),
            (dictgen::InsensitiveStr::Ascii("futrure"), &["future"]),
            (
                dictgen::InsensitiveStr::Ascii("experimantel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("viusally"), &["visually"]),
            (dictgen::InsensitiveStr::Ascii("sryacuse"), &["syracuse"]),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalise"),
                &["orthogonalise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiallises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marging"),
                &["margin", "merging"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownders"), &["wonders"]),
            (dictgen::InsensitiveStr::Ascii("unocode"), &["unicode"]),
            (
                dictgen::InsensitiveStr::Ascii("independense"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("blizzcoin"), &["blizzcon"]),
            (
                dictgen::InsensitiveStr::Ascii("justifiaction"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentaes"),
                &["concatenates"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectes"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("uptreams"), &["upstreams"]),
            (dictgen::InsensitiveStr::Ascii("stuggles"), &["struggles"]),
            (
                dictgen::InsensitiveStr::Ascii("precaucion"),
                &["precaution"],
            ),
            (dictgen::InsensitiveStr::Ascii("addonts"), &["addons"]),
            (dictgen::InsensitiveStr::Ascii("ingored"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("concequent"),
                &["consequent"],
            ),
            (dictgen::InsensitiveStr::Ascii("attented"), &["attended"]),
            (
                dictgen::InsensitiveStr::Ascii("exclamating"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("radus"), &["radius"]),
            (
                dictgen::InsensitiveStr::Ascii("depcrecated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("corellate"), &["correlate"]),
            (
                dictgen::InsensitiveStr::Ascii("reluctanly"),
                &["reluctantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bastardizin"),
                &["bastardizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("crytpo"), &["crypto"]),
            (dictgen::InsensitiveStr::Ascii("reccursion"), &["recursion"]),
            (
                dictgen::InsensitiveStr::Ascii("philosopies"),
                &["philosophies"],
            ),
            (dictgen::InsensitiveStr::Ascii("executtion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("horyzontally"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("loca"), &["local"]),
            (
                dictgen::InsensitiveStr::Ascii("inhabitents"),
                &["inhabitants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponted"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwhere"), &["elsewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("termintors"),
                &["terminators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncessarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progresions"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiases"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predections"),
                &["predictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arameterize"),
                &["parameterize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serwer"),
                &["server", "sewer"],
            ),
            (dictgen::InsensitiveStr::Ascii("coduct"), &["conduct"]),
            (
                dictgen::InsensitiveStr::Ascii("intitaliser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("pararms"), &["params"]),
            (
                dictgen::InsensitiveStr::Ascii("headquarer"),
                &["headquarter"],
            ),
            (dictgen::InsensitiveStr::Ascii("homless"), &["homeless"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotionations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternitiv"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sircles"), &["circles"]),
            (dictgen::InsensitiveStr::Ascii("dentifies"), &["identifies"]),
            (dictgen::InsensitiveStr::Ascii("lightes"), &["lighters"]),
            (
                dictgen::InsensitiveStr::Ascii("negoziation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instsallations"),
                &["installations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialitsing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contirbutes"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("bountry"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("dashboads"), &["dashboards"]),
            (dictgen::InsensitiveStr::Ascii("smoewhat"), &["somewhat"]),
            (dictgen::InsensitiveStr::Ascii("endig"), &["ending"]),
            (dictgen::InsensitiveStr::Ascii("concretly"), &["concretely"]),
            (dictgen::InsensitiveStr::Ascii("nutriet"), &["nutrient"]),
            (dictgen::InsensitiveStr::Ascii("varously"), &["variously"]),
            (
                dictgen::InsensitiveStr::Ascii("benevalent"),
                &["benevolent"],
            ),
            (dictgen::InsensitiveStr::Ascii("transferr"), &["transfer"]),
            (dictgen::InsensitiveStr::Ascii("succseed"), &["succeed"]),
            (
                dictgen::InsensitiveStr::Ascii("archiectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("nonte"), &["note"]),
            (
                dictgen::InsensitiveStr::Ascii("workarouds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoreixa"), &["anorexia"]),
            (dictgen::InsensitiveStr::Ascii("updgrading"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("imperfactions"),
                &["imperfections"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilty"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("conifgs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("apperciating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opportunitiy"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloesst"), &["closest"]),
            (
                dictgen::InsensitiveStr::Ascii("exauhstion"),
                &["exhaustion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumisable"),
                &["customisable"],
            ),
            (dictgen::InsensitiveStr::Ascii("opimiser"), &["optimiser"]),
            (dictgen::InsensitiveStr::Ascii("litigato"), &["litigation"]),
            (dictgen::InsensitiveStr::Ascii("deriair"), &["derriere"]),
            (dictgen::InsensitiveStr::Ascii("customsie"), &["customise"]),
            (
                dictgen::InsensitiveStr::Ascii("highschoool"),
                &["highschool"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techanically"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desnity"),
                &["density", "destiny"],
            ),
            (dictgen::InsensitiveStr::Ascii("protcool"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("monestic"), &["monastic"]),
            (dictgen::InsensitiveStr::Ascii("seseeded"), &["seceded"]),
            (
                dictgen::InsensitiveStr::Ascii("mezmorizer"),
                &["mesmerizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justifes"),
                &["justifies", "justices"],
            ),
            (dictgen::InsensitiveStr::Ascii("sionist"), &["zionist"]),
            (dictgen::InsensitiveStr::Ascii("edtiable"), &["editable"]),
            (dictgen::InsensitiveStr::Ascii("contina"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("tema"), &["team"]),
            (dictgen::InsensitiveStr::Ascii("downgaded"), &["downgraded"]),
            (dictgen::InsensitiveStr::Ascii("amater"), &["amateur"]),
            (dictgen::InsensitiveStr::Ascii("generatted"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("supppored"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("overcomming"),
                &["overcoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misygonistic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("delagated"), &["delegated"]),
            (dictgen::InsensitiveStr::Ascii("oyurs"), &["yours"]),
            (
                dictgen::InsensitiveStr::Ascii("aborginial"),
                &["aboriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("planeswlaker"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("jusifying"), &["justifying"]),
            (
                dictgen::InsensitiveStr::Ascii("intagrations"),
                &["integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breakpoitns"),
                &["breakpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("destops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("distnct"), &["distinct"]),
            (
                dictgen::InsensitiveStr::Ascii("personalitys"),
                &["personalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("curisers"), &["cruisers"]),
            (dictgen::InsensitiveStr::Ascii("asemble"), &["assemble"]),
            (dictgen::InsensitiveStr::Ascii("earller"), &["earlier"]),
            (dictgen::InsensitiveStr::Ascii("serailzed"), &["serialized"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapointted"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("troosso"), &["trousseau"]),
            (dictgen::InsensitiveStr::Ascii("cellural"), &["cellular"]),
            (dictgen::InsensitiveStr::Ascii("offfer"), &["offer"]),
            (dictgen::InsensitiveStr::Ascii("quotent"), &["quotient"]),
            (
                dictgen::InsensitiveStr::Ascii("fucntioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebiult"), &["rebuilt"]),
            (
                dictgen::InsensitiveStr::Ascii("distributiors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplicated"),
                &["deprecated", "duplicated", "replicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("apendix"), &["appendix"]),
            (
                dictgen::InsensitiveStr::Ascii("envvironment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("armchars"), &["armchairs"]),
            (
                dictgen::InsensitiveStr::Ascii("impossiblely"),
                &["impossibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dardenelles"),
                &["dardanelles"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcts"), &["products"]),
            (
                dictgen::InsensitiveStr::Ascii("pariticipated"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("flewant"), &["fluent"]),
            (dictgen::InsensitiveStr::Ascii("reverce"), &["reverse"]),
            (
                dictgen::InsensitiveStr::Ascii("aftzerwards"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancllations"),
                &["cancellations"],
            ),
            (dictgen::InsensitiveStr::Ascii("occurrs"), &["occurs"]),
            (
                dictgen::InsensitiveStr::Ascii("electroncis"),
                &["electrons"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloesly"), &["closely"]),
            (dictgen::InsensitiveStr::Ascii("tomarow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitics"),
                &["specifics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("campatibilities"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("champagen"), &["champagne"]),
            (
                dictgen::InsensitiveStr::Ascii("simulacion"),
                &["simulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("offstets"), &["offsets"]),
            (dictgen::InsensitiveStr::Ascii("ignora"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("fartherest"), &["furthest"]),
            (
                dictgen::InsensitiveStr::Ascii("friendzonned"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("tottneham"), &["tottenham"]),
            (dictgen::InsensitiveStr::Ascii("auhtors"), &["authors"]),
            (
                dictgen::InsensitiveStr::Ascii("correspendent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periodicly"),
                &["periodically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchronus"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admittetly"),
                &["admittedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplicitas"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("experince"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("frigthened"),
                &["frightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("blapshemy"), &["blasphemy"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualions"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wokers"),
                &["workers", "wakers"],
            ),
            (dictgen::InsensitiveStr::Ascii("militians"), &["militias"]),
            (
                dictgen::InsensitiveStr::Ascii("volumns"),
                &["volumes", "columns"],
            ),
            (dictgen::InsensitiveStr::Ascii("exansive"), &["expansive"]),
            (
                dictgen::InsensitiveStr::Ascii("maliciousally"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mircophones"),
                &["microphones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesysthem"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationella"),
                &["sensational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obsolutist"),
                &["absolutist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conolization"),
                &["colonization"],
            ),
            (dictgen::InsensitiveStr::Ascii("copiese"), &["copies"]),
            (
                dictgen::InsensitiveStr::Ascii("insuffciently"),
                &["insufficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("clientelle"), &["clientele"]),
            (
                dictgen::InsensitiveStr::Ascii("cringeworthly"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("aktivates"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("tempary"),
                &["temporary", "temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("listeneing"), &["listening"]),
            (
                dictgen::InsensitiveStr::Ascii("reconnetions"),
                &["reconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tesselatad"),
                &["tessellated"],
            ),
            (dictgen::InsensitiveStr::Ascii("protobol"), &["protocol"]),
            (
                dictgen::InsensitiveStr::Ascii("replacemenets"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhancemenmt"),
                &["enhancement"],
            ),
            (dictgen::InsensitiveStr::Ascii("requrested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("feild"), &["field"]),
            (
                dictgen::InsensitiveStr::Ascii("commutating"),
                &["commuting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finilising"),
                &["finalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("downsiders"), &["downsides"]),
            (dictgen::InsensitiveStr::Ascii("denining"), &["defining"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhithms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("puffered"), &["buffered"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondends"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("includig"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("reponded"), &["responded"]),
            (
                dictgen::InsensitiveStr::Ascii("initialializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("provicded"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("naibor"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("attrubite"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("lebanesse"), &["lebanese"]),
            (dictgen::InsensitiveStr::Ascii("aburpt"), &["abrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("deminsional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insesitivity"),
                &["insensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginasations"),
                &["organisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encaspulating"),
                &["encapsulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcutlates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegatiotiation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alighning"),
                &["aligning", "alighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apretiation"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sancturay"), &["sanctuary"]),
            (
                dictgen::InsensitiveStr::Ascii("instutions"),
                &["intuitions", "institutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("plateua"), &["plateau"]),
            (
                dictgen::InsensitiveStr::Ascii("hahve"),
                &["have", "halve", "half"],
            ),
            (dictgen::InsensitiveStr::Ascii("incrieased"), &["increased"]),
            (dictgen::InsensitiveStr::Ascii("banchmark"), &["benchmark"]),
            (dictgen::InsensitiveStr::Ascii("accoding"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("debgu"), &["debug"]),
            (
                dictgen::InsensitiveStr::Ascii("enounctering"),
                &["encountering"],
            ),
            (dictgen::InsensitiveStr::Ascii("truncting"), &["truncating"]),
            (
                dictgen::InsensitiveStr::Ascii("itialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinforcemnts"),
                &["reinforcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poperly"),
                &["properly", "property"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nestalgicly"),
                &["nostalgically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multpilications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vunrabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imediatly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("familiary"),
                &["familiarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("certaion"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("interitance"),
                &["inheritance"],
            ),
            (dictgen::InsensitiveStr::Ascii("karam"), &["karma"]),
            (
                dictgen::InsensitiveStr::Ascii("frequenzies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosnters"), &["monsters"]),
            (
                dictgen::InsensitiveStr::Ascii("politiikan"),
                &["politician"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainifests"), &["manifests"]),
            (
                dictgen::InsensitiveStr::Ascii("coneectors"),
                &["connectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstuctions"),
                &["reconstructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("immitation"), &["imitation"]),
            (
                dictgen::InsensitiveStr::Ascii("cimptomatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("secion"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("constructt"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("reolace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("sotrage"),
                &["storage", "shortage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submissies"),
                &["submissive"],
            ),
            (dictgen::InsensitiveStr::Ascii("begginners"), &["beginners"]),
            (dictgen::InsensitiveStr::Ascii("mis"), &["miss", "mist"]),
            (dictgen::InsensitiveStr::Ascii("buiild"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("especifically"),
                &["specifically", "especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("cursader"), &["crusader"]),
            (dictgen::InsensitiveStr::Ascii("preseve"), &["preserve"]),
            (
                dictgen::InsensitiveStr::Ascii("complciations"),
                &["complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("greande"), &["grenade"]),
            (dictgen::InsensitiveStr::Ascii("primarilly"), &["primarily"]),
            (
                dictgen::InsensitiveStr::Ascii("dinamicly"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprieciating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("harasment"), &["harassment"]),
            (dictgen::InsensitiveStr::Ascii("unlcoks"), &["unlocks"]),
            (dictgen::InsensitiveStr::Ascii("archving"), &["archiving"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetuties"),
                &["perpetuates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propegates"),
                &["propagates"],
            ),
            (dictgen::InsensitiveStr::Ascii("flexilbe"), &["flexible"]),
            (
                dictgen::InsensitiveStr::Ascii("interwining"),
                &["intertwining"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecise"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("excutions"), &["executions"]),
            (
                dictgen::InsensitiveStr::Ascii("squred"),
                &["squared", "squired"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterpeices"),
                &["masterpiece"],
            ),
            (dictgen::InsensitiveStr::Ascii("seranomy"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("dillimport"), &["dllimport"]),
            (dictgen::InsensitiveStr::Ascii("timeot"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuion"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percenatges"),
                &["percentages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutlithread"),
                &["multithread"],
            ),
            (dictgen::InsensitiveStr::Ascii("worbench"), &["workbench"]),
            (
                dictgen::InsensitiveStr::Ascii("evaulations"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("insteed"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("languanage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("declaritively"),
                &["declaratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelligant"),
                &["intelligent"],
            ),
            (dictgen::InsensitiveStr::Ascii("plastre"), &["plaster"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosyncracy"),
                &["idiosyncrasy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specializaton"),
                &["specialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("gltichy"), &["glitchy"]),
            (
                dictgen::InsensitiveStr::Ascii("antagonostic"),
                &["antagonist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opportunty"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("altough"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("slaugthered"),
                &["slaughtered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivelent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulatious"),
                &["simulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumizations"),
                &["customizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bactracking"),
                &["backtracking"],
            ),
            (dictgen::InsensitiveStr::Ascii("expeensive"), &["expensive"]),
            (
                dictgen::InsensitiveStr::Ascii("perdictions"),
                &["predictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("kindgoms"), &["kingdoms"]),
            (
                dictgen::InsensitiveStr::Ascii("disatisfied"),
                &["dissatisfied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comuters"),
                &["computers", "commuters"],
            ),
            (dictgen::InsensitiveStr::Ascii("strokin"), &["stroking"]),
            (
                dictgen::InsensitiveStr::Ascii("convinences"),
                &["conveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlightnement"),
                &["enlightenment"],
            ),
            (dictgen::InsensitiveStr::Ascii("unrankend"), &["unranked"]),
            (dictgen::InsensitiveStr::Ascii("aso"), &["also"]),
            (dictgen::InsensitiveStr::Ascii("ontacted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("corrupeted"), &["corrupted"]),
            (dictgen::InsensitiveStr::Ascii("seleceting"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("redirction"),
                &["redirection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectin"),
                &["connecting", "connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("peicewise"), &["piecewise"]),
            (dictgen::InsensitiveStr::Ascii("resuts"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("suscribers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("mebrains"), &["membranes"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colonozation"),
                &["colonization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administraion"),
                &["administration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interstingly"),
                &["interestingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismanlting"),
                &["dismantling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interwebers"),
                &["interwebs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workstaitions"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("syas"), &["says"]),
            (dictgen::InsensitiveStr::Ascii("decortive"), &["decorative"]),
            (dictgen::InsensitiveStr::Ascii("enabeling"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("adoptors"),
                &["adopters", "adaptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libreoffie"),
                &["libreoffice"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighour"), &["neighbour"]),
            (
                dictgen::InsensitiveStr::Ascii("consructor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finsh"),
                &["finish", "finch"],
            ),
            (dictgen::InsensitiveStr::Ascii("exansions"), &["expansions"]),
            (
                dictgen::InsensitiveStr::Ascii("fullsceren"),
                &["fullscreen"],
            ),
            (dictgen::InsensitiveStr::Ascii("negativ"), &["negative"]),
            (
                dictgen::InsensitiveStr::Ascii("differents"),
                &["different", "difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("firmwware"), &["firmware"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotioation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referers"),
                &["referrer", "referrers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caharcters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("veganisim"), &["veganism"]),
            (dictgen::InsensitiveStr::Ascii("numer"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("instruemnt"),
                &["instrument"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturuct"), &["struct"]),
            (
                dictgen::InsensitiveStr::Ascii("consecuence"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suceession"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinbe"), &["combined"]),
            (
                dictgen::InsensitiveStr::Ascii("peased"),
                &["leased", "pieced", "pleased"],
            ),
            (dictgen::InsensitiveStr::Ascii("leninent"), &["lenient"]),
            (
                dictgen::InsensitiveStr::Ascii("crystalization"),
                &["crystallization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resolutons"),
                &["resolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignaed"), &["aligned"]),
            (dictgen::InsensitiveStr::Ascii("spaeced"), &["spaced"]),
            (dictgen::InsensitiveStr::Ascii("headests"), &["headsets"]),
            (
                dictgen::InsensitiveStr::Ascii("altenatively"),
                &["alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penetratin"),
                &["penetration"],
            ),
            (dictgen::InsensitiveStr::Ascii("deteced"), &["detected"]),
            (
                dictgen::InsensitiveStr::Ascii("immaturaty"),
                &["immaturity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comine"),
                &["combine", "coming"],
            ),
            (dictgen::InsensitiveStr::Ascii("brnaches"), &["branches"]),
            (
                dictgen::InsensitiveStr::Ascii("autocorrext"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("rotataions"), &["rotations"]),
            (
                dictgen::InsensitiveStr::Ascii("incorretly"),
                &["incorrectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionnal"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("suddeny"), &["suddenly"]),
            (dictgen::InsensitiveStr::Ascii("ramains"), &["remains"]),
            (
                dictgen::InsensitiveStr::Ascii("atuhenticated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("presciuos"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("norhern"), &["northern"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilish"),
                &["establish", "stabilise"],
            ),
            (dictgen::InsensitiveStr::Ascii("exclud"), &["exclude"]),
            (
                dictgen::InsensitiveStr::Ascii("benfits"),
                &["benefits", "befits"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemement"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("matc"), &["match"]),
            (
                dictgen::InsensitiveStr::Ascii("probablities"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("quanitzed"), &["quantized"]),
            (
                dictgen::InsensitiveStr::Ascii("implicite"),
                &["implicit", "implicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flacorings"),
                &["flavorings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persausion"),
                &["persuasion"],
            ),
            (dictgen::InsensitiveStr::Ascii("oberv"), &["observe"]),
            (dictgen::InsensitiveStr::Ascii("coyp"), &["copy"]),
            (
                dictgen::InsensitiveStr::Ascii("disappered"),
                &["disappeared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("actuall"),
                &["actually", "actual"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordind"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("cahr"), &["char"]),
            (
                dictgen::InsensitiveStr::Ascii("instllation"),
                &["installation", "instillation"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheam"), &["schema"]),
            (
                dictgen::InsensitiveStr::Ascii("easer"),
                &["easier", "eraser"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomander"), &["commander"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratioons"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("adeliade"), &["adelaide"]),
            (
                dictgen::InsensitiveStr::Ascii("progapates"),
                &["propagates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cathegorical"),
                &["categorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coexsiting"),
                &["coexisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressulting"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("configruable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mismaching"),
                &["mismatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("primimitives"),
                &["primitives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparenlty"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acomplishing"),
                &["accomplishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sopporting"),
                &["supporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconfure"),
                &["reconfigure"],
            ),
            (dictgen::InsensitiveStr::Ascii("pobably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("constrcution"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miinimisation"),
                &["minimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaites"), &["animates"]),
            (dictgen::InsensitiveStr::Ascii("loclaes"), &["locales"]),
            (
                dictgen::InsensitiveStr::Ascii("herarchies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfriendy"), &["unfriendly"]),
            (
                dictgen::InsensitiveStr::Ascii("responesive"),
                &["responsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforecer"), &["enforcer"]),
            (dictgen::InsensitiveStr::Ascii("insertino"), &["inserting"]),
            (
                dictgen::InsensitiveStr::Ascii("continenal"),
                &["continental"],
            ),
            (dictgen::InsensitiveStr::Ascii("dervie"), &["derive"]),
            (dictgen::InsensitiveStr::Ascii("caputres"), &["captures"]),
            (
                dictgen::InsensitiveStr::Ascii("inititializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("againt"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("referencd"),
                &["referenced", "reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("depeding"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("adventrous"),
                &["adventurous"],
            ),
            (dictgen::InsensitiveStr::Ascii("choseen"), &["chosen"]),
            (
                dictgen::InsensitiveStr::Ascii("guarrantyd"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminacion"),
                &["discrimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("naviagate"), &["navigate"]),
            (
                dictgen::InsensitiveStr::Ascii("minimazing"),
                &["minimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("capacior"), &["capacitor"]),
            (dictgen::InsensitiveStr::Ascii("impuner"), &["impugner"]),
            (dictgen::InsensitiveStr::Ascii("operatator"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("jetbrain"), &["jetbrains"]),
            (
                dictgen::InsensitiveStr::Ascii("collobarate"),
                &["collaborate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decceleration"),
                &["deceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neruoscience"),
                &["neuroscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("inspectin"), &["inspection"]),
            (dictgen::InsensitiveStr::Ascii("failres"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("directrories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmars"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounsd"), &["bounds"]),
            (dictgen::InsensitiveStr::Ascii("tgat"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("cosndier"), &["consider"]),
            (
                dictgen::InsensitiveStr::Ascii("retaurants"),
                &["restaurants"],
            ),
            (dictgen::InsensitiveStr::Ascii("memmber"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepeneur"),
                &["entrepreneur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabitiliby"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("synteses"), &["syntheses"]),
            (
                dictgen::InsensitiveStr::Ascii("abstrction"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("corresond"), &["correspond"]),
            (dictgen::InsensitiveStr::Ascii("extrotion"), &["extortion"]),
            (dictgen::InsensitiveStr::Ascii("addessed"), &["addressed"]),
            (
                dictgen::InsensitiveStr::Ascii("syntatically"),
                &["syntactically"],
            ),
            (dictgen::InsensitiveStr::Ascii("grafitti"), &["graffiti"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibting"),
                &["prohibiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreplacable"),
                &["unreplaceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertx"), &["vertex"]),
            (dictgen::InsensitiveStr::Ascii("audbile"), &["audible"]),
            (
                dictgen::InsensitiveStr::Ascii("dynamicallly"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("religionens"),
                &["religions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proffesionals"),
                &["professionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("librabary"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("presevering"),
                &["preserving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priveldges"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("jamine"), &["jasmine"]),
            (
                dictgen::InsensitiveStr::Ascii("expements"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discuusion"),
                &["discussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apllications"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("isntance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("paraghraph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("reboto"), &["reboot"]),
            (
                dictgen::InsensitiveStr::Ascii("dokcerd"),
                &["dockerd", "docked", "docker"],
            ),
            (dictgen::InsensitiveStr::Ascii("embassay"), &["embassy"]),
            (dictgen::InsensitiveStr::Ascii("dinosaures"), &["dinosaurs"]),
            (dictgen::InsensitiveStr::Ascii("evalutator"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("samler"), &["sampler"]),
            (
                dictgen::InsensitiveStr::Ascii("environmentl"),
                &["environmentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("decalre"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("alliases"), &["aliases"]),
            (
                dictgen::InsensitiveStr::Ascii("inexblicably"),
                &["inexplicably"],
            ),
            (dictgen::InsensitiveStr::Ascii("memoey"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("gurading"), &["guarding"]),
            (dictgen::InsensitiveStr::Ascii("propable"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("scucess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("stipler"), &["stippler"]),
            (
                dictgen::InsensitiveStr::Ascii("peassignment"),
                &["reassignment", "preassignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("rethoric"), &["rhetoric"]),
            (
                dictgen::InsensitiveStr::Ascii("stattistically"),
                &["statistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("musculair"), &["muscular"]),
            (dictgen::InsensitiveStr::Ascii("nulllable"), &["nullable"]),
            (
                dictgen::InsensitiveStr::Ascii("extraoridnary"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("meningful"), &["meaningful"]),
            (dictgen::InsensitiveStr::Ascii("assunbler"), &["assembler"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifiying"),
                &["qualifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifieer"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("promtp"), &["prompt"]),
            (dictgen::InsensitiveStr::Ascii("budled"), &["bundled"]),
            (
                dictgen::InsensitiveStr::Ascii("theorectical"),
                &["theoretical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("censurship"),
                &["censorship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knoledgeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("archvied"), &["archived"]),
            (dictgen::InsensitiveStr::Ascii("attachting"), &["attaching"]),
            (dictgen::InsensitiveStr::Ascii("smaples"), &["samples"]),
            (
                dictgen::InsensitiveStr::Ascii("consectuive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("desingns"), &["designs"]),
            (dictgen::InsensitiveStr::Ascii("foregrond"), &["foreground"]),
            (
                dictgen::InsensitiveStr::Ascii("definatively"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("dulaity"), &["duality"]),
            (
                dictgen::InsensitiveStr::Ascii("indempotent"),
                &["idempotent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronisations"),
                &["synchronisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("weekedn"), &["weekend"]),
            (
                dictgen::InsensitiveStr::Ascii("comparissons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flexability"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misintepret"),
                &["misinterpret"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibilties"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obsolutely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("reodered"), &["reordered"]),
            (dictgen::InsensitiveStr::Ascii("pendatic"), &["pedantic"]),
            (
                dictgen::InsensitiveStr::Ascii("inmigrants"),
                &["immigrants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusuives"),
                &["exclusives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemeted"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gauarenteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosiphy"),
                &["philosophy"],
            ),
            (dictgen::InsensitiveStr::Ascii("knowning"), &["knowing"]),
            (
                dictgen::InsensitiveStr::Ascii("dulications"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("sebrian"), &["serbian"]),
            (dictgen::InsensitiveStr::Ascii("poety"), &["poetry"]),
            (
                dictgen::InsensitiveStr::Ascii("referecned"),
                &["referenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuting"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chararacters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomobiles"),
                &["automobiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("terrorisim"), &["terrorism"]),
            (
                dictgen::InsensitiveStr::Ascii("efine"),
                &["define", "refine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotianing"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("simptums"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("guaratees"), &["guarantees"]),
            (dictgen::InsensitiveStr::Ascii("serapate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("reconncted"),
                &["reconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("nugfix"), &["bugfix"]),
            (dictgen::InsensitiveStr::Ascii("coexhisted"), &["coexisted"]),
            (dictgen::InsensitiveStr::Ascii("patterened"), &["patterned"]),
            (dictgen::InsensitiveStr::Ascii("asscoitae"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("exceuctors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("filemane"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("charecteristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inefficeint"),
                &["inefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("percise"), &["precise"]),
            (
                dictgen::InsensitiveStr::Ascii("contreception"),
                &["contraception"],
            ),
            (dictgen::InsensitiveStr::Ascii("comflict"), &["conflict"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("pixles"), &["pixels"]),
            (dictgen::InsensitiveStr::Ascii("pagagraph"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("underhwelming"),
                &["underwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("semphore"), &["semaphore"]),
            (dictgen::InsensitiveStr::Ascii("versoioned"), &["versioned"]),
            (
                dictgen::InsensitiveStr::Ascii("simplyfing"),
                &["simplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docuimented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiantorics"),
                &["combinatorics"],
            ),
            (dictgen::InsensitiveStr::Ascii("proprly"), &["properly"]),
            (dictgen::InsensitiveStr::Ascii("chesck"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegogtiations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aswered"), &["answered"]),
            (dictgen::InsensitiveStr::Ascii("serialsise"), &["serialise"]),
            (dictgen::InsensitiveStr::Ascii("rostaing"), &["roasting"]),
            (dictgen::InsensitiveStr::Ascii("notifes"), &["notifies"]),
            (dictgen::InsensitiveStr::Ascii("repositor"), &["repository"]),
            (
                dictgen::InsensitiveStr::Ascii("instroductions"),
                &["introductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("heatbeats"), &["heartbeats"]),
            (
                dictgen::InsensitiveStr::Ascii("accellerators"),
                &["accelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("decidin"), &["deciding"]),
            (
                dictgen::InsensitiveStr::Ascii("compatilibity"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("advence"), &["advance"]),
            (dictgen::InsensitiveStr::Ascii("eaxctly"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("depandencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annonymously"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("penguinos"), &["penguins"]),
            (
                dictgen::InsensitiveStr::Ascii("repeadedly"),
                &["repeatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annihliation"),
                &["annihilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discontiguities"),
                &["discontinuities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rictatorship"),
                &["dictatorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("toatul"), &["total"]),
            (dictgen::InsensitiveStr::Ascii("annoate"), &["annotate"]),
            (dictgen::InsensitiveStr::Ascii("repace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("compensatin"),
                &["compensation", "compensating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accommidated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("docments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("transocders"),
                &["transcoders"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanauge"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("activete"), &["activate"]),
            (dictgen::InsensitiveStr::Ascii("acccepting"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("enchantmant"),
                &["enchantment"],
            ),
            (dictgen::InsensitiveStr::Ascii("preamle"), &["preamble"]),
            (dictgen::InsensitiveStr::Ascii("periodes"), &["periods"]),
            (
                dictgen::InsensitiveStr::Ascii("compabillity"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("luckliy"), &["luckily"]),
            (
                dictgen::InsensitiveStr::Ascii("donwloading"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singeltons"),
                &["singletons"],
            ),
            (dictgen::InsensitiveStr::Ascii("fintuned"), &["finetuned"]),
            (dictgen::InsensitiveStr::Ascii("sepator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("discovrs"), &["discovers"]),
            (dictgen::InsensitiveStr::Ascii("immersie"), &["immerse"]),
            (
                dictgen::InsensitiveStr::Ascii("confingurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrinsinc"), &["intrinsic"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotioation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recquiring"),
                &["requiring", "reacquiring"],
            ),
            (dictgen::InsensitiveStr::Ascii("offaet"), &["offset"]),
            (dictgen::InsensitiveStr::Ascii("laringes"), &["larynxes"]),
            (
                dictgen::InsensitiveStr::Ascii("integrats"),
                &["integrates", "integrals"],
            ),
            (dictgen::InsensitiveStr::Ascii("dystpoian"), &["dystopian"]),
            (
                dictgen::InsensitiveStr::Ascii("administrez"),
                &["administer"],
            ),
            (dictgen::InsensitiveStr::Ascii("somwtime"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("exixt"), &["exist", "exit"]),
            (dictgen::InsensitiveStr::Ascii("imporsts"), &["imports"]),
            (dictgen::InsensitiveStr::Ascii("programd"), &["programed"]),
            (dictgen::InsensitiveStr::Ascii("agreesive"), &["aggressive"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhod"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agircultural"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shanged"),
                &["changed", "shanked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleratin"),
                &["accelerating", "acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converastions"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("flahs"), &["flash", "flags"]),
            (dictgen::InsensitiveStr::Ascii("annayed"), &["annoyed"]),
            (
                dictgen::InsensitiveStr::Ascii("disgrunteld"),
                &["disgruntled"],
            ),
            (dictgen::InsensitiveStr::Ascii("worload"), &["workload"]),
            (dictgen::InsensitiveStr::Ascii("disretion"), &["discretion"]),
            (dictgen::InsensitiveStr::Ascii("lenghty"), &["lengthy"]),
            (
                dictgen::InsensitiveStr::Ascii("programmming"),
                &["programming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copmletions"),
                &["completions"],
            ),
            (dictgen::InsensitiveStr::Ascii("populary"), &["popularity"]),
            (
                dictgen::InsensitiveStr::Ascii("unemployeed"),
                &["unemployed"],
            ),
            (dictgen::InsensitiveStr::Ascii("earliet"), &["earliest"]),
            (dictgen::InsensitiveStr::Ascii("kmow"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("butterly"), &["butterfly"]),
            (dictgen::InsensitiveStr::Ascii("prefering"), &["preferring"]),
            (
                dictgen::InsensitiveStr::Ascii("certificte"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parokeally"),
                &["parochially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaptible"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circuvention"),
                &["circumvention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capter"),
                &["captor", "chapter"],
            ),
            (dictgen::InsensitiveStr::Ascii("thikn"), &["think"]),
            (
                dictgen::InsensitiveStr::Ascii("extraordinarly"),
                &["extraordinary", "extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("aswer"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("forsaw"), &["foresaw"]),
            (dictgen::InsensitiveStr::Ascii("directely"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("tranaction"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("orthodx"), &["orthodox"]),
            (
                dictgen::InsensitiveStr::Ascii("undiserable"),
                &["undesirable"],
            ),
            (dictgen::InsensitiveStr::Ascii("metalic"), &["metallic"]),
            (dictgen::InsensitiveStr::Ascii("specefic"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("reparamterisation"),
                &["reparameterisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("abandonde"), &["abandoned"]),
            (dictgen::InsensitiveStr::Ascii("callack"), &["callback"]),
            (
                dictgen::InsensitiveStr::Ascii("benovelent"),
                &["benevolent"],
            ),
            (dictgen::InsensitiveStr::Ascii("jenkin"), &["jenkins"]),
            (
                dictgen::InsensitiveStr::Ascii("resticting"),
                &["restricting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nescesarrily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("symobol"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("sutracting"),
                &["subtracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("brabarian"), &["barbarian"]),
            (dictgen::InsensitiveStr::Ascii("aparatus"), &["apparatus"]),
            (dictgen::InsensitiveStr::Ascii("garnison"), &["garrison"]),
            (
                dictgen::InsensitiveStr::Ascii("miracluous"),
                &["miraculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("overturing"), &["overturn"]),
            (dictgen::InsensitiveStr::Ascii("intefered"), &["interfered"]),
            (
                dictgen::InsensitiveStr::Ascii("metropolian"),
                &["metropolitan"],
            ),
            (dictgen::InsensitiveStr::Ascii("quantuum"), &["quantum"]),
            (dictgen::InsensitiveStr::Ascii("sinceer"), &["sincere"]),
            (dictgen::InsensitiveStr::Ascii("confunse"), &["confuse"]),
            (dictgen::InsensitiveStr::Ascii("resverving"), &["reserving"]),
            (dictgen::InsensitiveStr::Ascii("preservin"), &["preserving"]),
            (dictgen::InsensitiveStr::Ascii("stabillity"), &["stability"]),
            (
                dictgen::InsensitiveStr::Ascii("blamethrower"),
                &["flamethrower"],
            ),
            (dictgen::InsensitiveStr::Ascii("coctail"), &["cocktail"]),
            (
                dictgen::InsensitiveStr::Ascii("congragulations"),
                &["congratulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mices"), &["mice"]),
            (
                dictgen::InsensitiveStr::Ascii("concatation"),
                &["concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anitdepressants"),
                &["antidepressants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immidately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("billboad"), &["billboard"]),
            (dictgen::InsensitiveStr::Ascii("manevuer"), &["maneuver"]),
            (
                dictgen::InsensitiveStr::Ascii("foreingers"),
                &["foreigners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defaulst"),
                &["defaults", "default"],
            ),
            (dictgen::InsensitiveStr::Ascii("patirots"), &["patriots"]),
            (dictgen::InsensitiveStr::Ascii("scuptures"), &["sculptures"]),
            (
                dictgen::InsensitiveStr::Ascii("acceptabelt"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attennuation"),
                &["attenuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insturctional"),
                &["instructional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrupttions"),
                &["interruptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscritpitons"),
                &["unsubscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("teatotler"), &["teetotaler"]),
            (dictgen::InsensitiveStr::Ascii("adjacentsy"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("storise"), &["stories"]),
            (dictgen::InsensitiveStr::Ascii("neighbet"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("incrompehensible"),
                &["incomprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechlist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("satelitte"), &["satellite"]),
            (dictgen::InsensitiveStr::Ascii("evauluated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("delegative"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("chheck"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("specificying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertices"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwritting"),
                &["overwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discribing"),
                &["describing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplyfies"),
                &["simplifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ehtnically"),
                &["ethnically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciprocant"),
                &["reciprocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucratische"),
                &["bureaucratic"],
            ),
            (
                dictgen::InsensitiveStr::Unicode("сontainers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacluator"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("authencicate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("numercially"),
                &["numerically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriximate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tabele"), &["table"]),
            (
                dictgen::InsensitiveStr::Ascii("unrealistc"),
                &["unrealistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitror"),
                &["arbiter", "arbitrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmissable"),
                &["transmissible"],
            ),
            (dictgen::InsensitiveStr::Ascii("weigth"), &["weight"]),
            (dictgen::InsensitiveStr::Ascii("charators"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misoganistic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("aneels"), &["anneals"]),
            (dictgen::InsensitiveStr::Ascii("whery"), &["where"]),
            (
                dictgen::InsensitiveStr::Ascii("organisationnels"),
                &["organisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("geniunely"), &["genuinely"]),
            (
                dictgen::InsensitiveStr::Ascii("patriarca"),
                &["patriarchal"],
            ),
            (dictgen::InsensitiveStr::Ascii("anotates"), &["annotates"]),
            (dictgen::InsensitiveStr::Ascii("processig"), &["processing"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumentul"),
                &["instrumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("equippied"), &["equipped"]),
            (dictgen::InsensitiveStr::Ascii("killingest"), &["killings"]),
            (dictgen::InsensitiveStr::Ascii("saty"), &["stay"]),
            (dictgen::InsensitiveStr::Ascii("pakvage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("vacums"), &["vacuums"]),
            (dictgen::InsensitiveStr::Ascii("suprised"), &["surprised"]),
            (
                dictgen::InsensitiveStr::Ascii("encaspulates"),
                &["encapsulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("stautses"), &["statuses"]),
            (
                dictgen::InsensitiveStr::Ascii("derevative"),
                &["derivative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoretocally"),
                &["theoretically"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixeed"), &["fixed"]),
            (dictgen::InsensitiveStr::Ascii("speciafied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("perscribe"), &["prescribe"]),
            (dictgen::InsensitiveStr::Ascii("aggreagate"), &["aggregate"]),
            (
                dictgen::InsensitiveStr::Ascii("disrepsecting"),
                &["disrespecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specyfying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangaging"), &["managing"]),
            (dictgen::InsensitiveStr::Ascii("allinaces"), &["alliances"]),
            (
                dictgen::InsensitiveStr::Ascii("disctinctive"),
                &["distinctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disiplined"),
                &["disciplined"],
            ),
            (dictgen::InsensitiveStr::Ascii("theer"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("prequisites"),
                &["prerequisites"],
            ),
            (dictgen::InsensitiveStr::Ascii("threshould"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("redundacy"), &["redundancy"]),
            (dictgen::InsensitiveStr::Ascii("derfined"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("meaninngs"), &["meanings"]),
            (dictgen::InsensitiveStr::Ascii("alchol"), &["alcohol"]),
            (
                dictgen::InsensitiveStr::Ascii("exprimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostprocessing"),
                &["postprocessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("navtive"), &["native"]),
            (
                dictgen::InsensitiveStr::Ascii("interferire"),
                &["interfere"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopfully"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("varables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("bangledash"),
                &["bangladesh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultenious"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("someon"), &["someone"]),
            (dictgen::InsensitiveStr::Ascii("usng"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("austitic"), &["autistic"]),
            (
                dictgen::InsensitiveStr::Ascii("resquesting"),
                &["requesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("entpoint"), &["endpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("corrdination"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vollyeball"),
                &["volleyball"],
            ),
            (dictgen::InsensitiveStr::Ascii("inticrate"), &["intricate"]),
            (dictgen::InsensitiveStr::Ascii("ressolves"), &["resolves"]),
            (dictgen::InsensitiveStr::Ascii("simialrly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteamers"),
                &["downstreamers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilivians"), &["civilians"]),
            (
                dictgen::InsensitiveStr::Ascii("hexademical"),
                &["hexadecimal"],
            ),
            (dictgen::InsensitiveStr::Ascii("webages"), &["webpages"]),
            (dictgen::InsensitiveStr::Ascii("kackies"), &["khakis"]),
            (
                dictgen::InsensitiveStr::Ascii("maniplulate"),
                &["manipulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("brwosed"), &["browsed"]),
            (dictgen::InsensitiveStr::Ascii("matreial"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("personnally"),
                &["personally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknownledged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("instancce"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("rythm"), &["rhythm"]),
            (dictgen::InsensitiveStr::Ascii("pinuchle"), &["pinochle"]),
            (dictgen::InsensitiveStr::Ascii("cabinate"), &["cabinet"]),
            (
                dictgen::InsensitiveStr::Ascii("necessesities"),
                &["necessities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasred"),
                &["parsed", "pasted", "passed", "pared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannibalisin"),
                &["cannibalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("napolean"), &["napoleon"]),
            (dictgen::InsensitiveStr::Ascii("eletronic"), &["electronic"]),
            (
                dictgen::InsensitiveStr::Ascii("surviced"),
                &["serviced", "survived"],
            ),
            (dictgen::InsensitiveStr::Ascii("interiour"), &["interior"]),
            (
                dictgen::InsensitiveStr::Ascii("idenitties"),
                &["identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("dividendos"), &["dividends"]),
            (dictgen::InsensitiveStr::Ascii("modiifies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("perfurring"),
                &["preferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indroductory"),
                &["introductory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograpphed"),
                &["photographed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenctication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("innosent"), &["innocent"]),
            (
                dictgen::InsensitiveStr::Ascii("decotations"),
                &["decorations"],
            ),
            (dictgen::InsensitiveStr::Ascii("snodwen"), &["snowden"]),
            (dictgen::InsensitiveStr::Ascii("intiated"), &["initiated"]),
            (dictgen::InsensitiveStr::Ascii("metifors"), &["metaphors"]),
            (dictgen::InsensitiveStr::Ascii("featrued"), &["featured"]),
            (dictgen::InsensitiveStr::Ascii("semicolor"), &["semicolon"]),
            (
                dictgen::InsensitiveStr::Ascii("socailists"),
                &["socialists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifizing"),
                &["sacrificing"],
            ),
            (dictgen::InsensitiveStr::Ascii("comple"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("wallpapes"), &["wallpapers"]),
            (dictgen::InsensitiveStr::Ascii("somethhing"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("correspendences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("datas"), &["data"]),
            (
                dictgen::InsensitiveStr::Ascii("eleminating"),
                &["eliminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("othewice"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("amelearate"),
                &["ameliorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfomfortable"),
                &["uncomfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("misstype"), &["mistype"]),
            (dictgen::InsensitiveStr::Ascii("throuh"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastibation"),
                &["procrastination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outrageouly"),
                &["outrageously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outperfrom"),
                &["outperform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelarete"),
                &["accelerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("grandise"), &["aggrandise"]),
            (dictgen::InsensitiveStr::Ascii("enduces"), &["induces"]),
            (
                dictgen::InsensitiveStr::Ascii("secceeding"),
                &["succeeding", "seceding"],
            ),
            (dictgen::InsensitiveStr::Ascii("claener"), &["cleaner"]),
            (dictgen::InsensitiveStr::Ascii("teches"), &["teaches"]),
            (
                dictgen::InsensitiveStr::Ascii("sparations"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desperatedly"),
                &["desperately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commemoratin"),
                &["commemorating", "commemoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("epoches"), &["epochs"]),
            (dictgen::InsensitiveStr::Ascii("conflift"), &["conflict"]),
            (
                dictgen::InsensitiveStr::Ascii("intelligble"),
                &["intelligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanuages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("summariaztion"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomentry"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("decodeable"), &["decodable"]),
            (dictgen::InsensitiveStr::Ascii("instancd"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("exceuctes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("optionn"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("criticiszing"),
                &["criticising"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarunte"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("masster"), &["master"]),
            (
                dictgen::InsensitiveStr::Ascii("configurrable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortihmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangments"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("revserse"), &["reverse"]),
            (dictgen::InsensitiveStr::Ascii("litihum"), &["lithium"]),
            (dictgen::InsensitiveStr::Ascii("wnats"), &["wants"]),
            (
                dictgen::InsensitiveStr::Ascii("egotistcal"),
                &["egotistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("facce"), &["face"]),
            (dictgen::InsensitiveStr::Ascii("helath"), &["health"]),
            (dictgen::InsensitiveStr::Ascii("enjoing"), &["enjoying"]),
            (dictgen::InsensitiveStr::Ascii("rahter"), &["rather"]),
            (dictgen::InsensitiveStr::Ascii("marhsal"), &["marshal"]),
            (
                dictgen::InsensitiveStr::Ascii("stegnography"),
                &["steganography"],
            ),
            (dictgen::InsensitiveStr::Ascii("princeses"), &["princess"]),
            (
                dictgen::InsensitiveStr::Ascii("unspefififed"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("greenalnd"), &["greenland"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitite"),
                &["prostitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contentended"),
                &["contended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handwritin"),
                &["handwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patrairchy"),
                &["patriarchy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articifial"),
                &["artificial"],
            ),
            (dictgen::InsensitiveStr::Ascii("delevopp"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("inserst"), &["insert"]),
            (
                dictgen::InsensitiveStr::Ascii("masculanity"),
                &["masculinity"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortunat"), &["fortunate"]),
            (dictgen::InsensitiveStr::Ascii("adquiring"), &["acquiring"]),
            (dictgen::InsensitiveStr::Ascii("intterrupt"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("premilinary"),
                &["preliminary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigificant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("suvenear"), &["souvenir"]),
            (dictgen::InsensitiveStr::Ascii("circuis"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("specfiically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidade"),
                &["intimidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caraciture"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrolating"),
                &["correlating"],
            ),
            (dictgen::InsensitiveStr::Ascii("polishees"), &["polishes"]),
            (
                dictgen::InsensitiveStr::Ascii("interresting"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credentail"),
                &["credential"],
            ),
            (dictgen::InsensitiveStr::Ascii("inresting"), &["inserting"]),
            (dictgen::InsensitiveStr::Ascii("organizuje"), &["organize"]),
            (dictgen::InsensitiveStr::Ascii("commenet"), &["comment"]),
            (
                dictgen::InsensitiveStr::Ascii("anphetamine"),
                &["amphetamine"],
            ),
            (dictgen::InsensitiveStr::Ascii("seiving"), &["sieving"]),
            (
                dictgen::InsensitiveStr::Ascii("surronding"),
                &["surrounding"],
            ),
            (dictgen::InsensitiveStr::Ascii("poitives"), &["positives"]),
            (
                dictgen::InsensitiveStr::Ascii("tha"),
                &["than", "that", "the"],
            ),
            (dictgen::InsensitiveStr::Ascii("amrenian"), &["armenian"]),
            (dictgen::InsensitiveStr::Ascii("hieghts"), &["heights"]),
            (dictgen::InsensitiveStr::Ascii("densicy"), &["density"]),
            (
                dictgen::InsensitiveStr::Ascii("equillibriation"),
                &["equilibriation"],
            ),
            (dictgen::InsensitiveStr::Ascii("callis"), &["callus"]),
            (
                dictgen::InsensitiveStr::Ascii("disirability"),
                &["desirability"],
            ),
            (dictgen::InsensitiveStr::Ascii("brusses"), &["brussels"]),
            (
                dictgen::InsensitiveStr::Ascii("conspirancy"),
                &["conspiracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilizied"),
                &["stabilize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrption"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("makefle"), &["makefile"]),
            (dictgen::InsensitiveStr::Ascii("maxium"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("perservation"),
                &["preservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resurreccion"),
                &["resurrection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmenetation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniulating"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("abd"), &["and", "bad"]),
            (
                dictgen::InsensitiveStr::Ascii("reccommendations"),
                &["recommendations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ommisions"),
                &["omissions", "commissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tructuring"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("westdern"), &["western"]),
            (
                dictgen::InsensitiveStr::Ascii("misdemenor"),
                &["misdemeanor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partcipates"),
                &["participates"],
            ),
            (dictgen::InsensitiveStr::Ascii("velcity"), &["velocity"]),
            (
                dictgen::InsensitiveStr::Ascii("uderstanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficifc"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("stereotpyes"),
                &["stereotypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinater"),
                &["combinator", "combiner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furhtermore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approimately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixeir"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("cracefulness"),
                &["gracefulness"],
            ),
            (dictgen::InsensitiveStr::Ascii("scartched"), &["scratched"]),
            (dictgen::InsensitiveStr::Ascii("installes"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("silentyly"), &["silently"]),
            (dictgen::InsensitiveStr::Ascii("pomises"), &["promises"]),
            (dictgen::InsensitiveStr::Ascii("depictes"), &["depicts"]),
            (
                dictgen::InsensitiveStr::Ascii("muliplication"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplyoment"),
                &["deployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("subfolfer"), &["subfolder"]),
            (
                dictgen::InsensitiveStr::Ascii("obilgatory"),
                &["obligatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglicting"),
                &["neglecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qustioning"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estiamations"),
                &["estimations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sunderlund"),
                &["sunderland"],
            ),
            (dictgen::InsensitiveStr::Ascii("semgented"), &["segmented"]),
            (dictgen::InsensitiveStr::Ascii("buttin"), &["button"]),
            (
                dictgen::InsensitiveStr::Ascii("practially"),
                &["practically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exacution"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("bult"), &["built"]),
            (dictgen::InsensitiveStr::Ascii("wieght"), &["weight"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsdence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("becahse"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("facinates"), &["fascinates"]),
            (dictgen::InsensitiveStr::Ascii("contorler"), &["controller"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribiute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("rlation"), &["relation"]),
            (dictgen::InsensitiveStr::Ascii("sory"), &["sorry", "sort"]),
            (dictgen::InsensitiveStr::Ascii("verisoner"), &["versioner"]),
            (
                dictgen::InsensitiveStr::Ascii("aggravanti"),
                &["aggravating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimater"),
                &["estimator", "estimated", "estimates", "estimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessaryly"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authendication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("containzer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("satistfying"),
                &["satisfying"],
            ),
            (dictgen::InsensitiveStr::Ascii("toglled"), &["toggled"]),
            (dictgen::InsensitiveStr::Ascii("exptect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("varaint"), &["variant"]),
            (dictgen::InsensitiveStr::Ascii("perfroms"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("otimism"), &["optimism"]),
            (
                dictgen::InsensitiveStr::Ascii("distroying"),
                &["destroying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsittute"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemetend"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regconised"),
                &["recognised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rranslation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inserster"), &["inserter"]),
            (
                dictgen::InsensitiveStr::Ascii("sizemologists"),
                &["seismologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("datetiem"), &["datetime"]),
            (
                dictgen::InsensitiveStr::Ascii("appening"),
                &["appending", "happening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executin"),
                &["execution", "executing"],
            ),
            (dictgen::InsensitiveStr::Ascii("avearged"), &["averaged"]),
            (dictgen::InsensitiveStr::Ascii("deuling"), &["dueling"]),
            (
                dictgen::InsensitiveStr::Ascii("astethicism"),
                &["aestheticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimdnter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwide"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("segmen"), &["segment"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualtor"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaialbility"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simptumaticlly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("advertisy"), &["adversity"]),
            (dictgen::InsensitiveStr::Ascii("rainbos"), &["rainbows"]),
            (
                dictgen::InsensitiveStr::Ascii("docstatistik"),
                &["docstatistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifficance"),
                &["significance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesystyem"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulaitons"),
                &["simulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subproccesed"),
                &["subprocessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("togggle"), &["toggle"]),
            (
                dictgen::InsensitiveStr::Ascii("identifaction"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("deconders"), &["decoders"]),
            (
                dictgen::InsensitiveStr::Ascii("renegothiates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowladged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceeds"), &["concedes"]),
            (dictgen::InsensitiveStr::Ascii("chipest"), &["chipset"]),
            (dictgen::InsensitiveStr::Ascii("qouted"), &["quoted"]),
            (dictgen::InsensitiveStr::Ascii("midtwon"), &["midtown"]),
            (dictgen::InsensitiveStr::Ascii("monstros"), &["monstrous"]),
            (dictgen::InsensitiveStr::Ascii("ramdomly"), &["randomly"]),
            (
                dictgen::InsensitiveStr::Ascii("canibalize"),
                &["cannibalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generattions"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("assertio"), &["assertion"]),
            (
                dictgen::InsensitiveStr::Ascii("containners"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerize"),
                &["parameterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("acrynom"), &["acronym"]),
            (dictgen::InsensitiveStr::Ascii("coomposed"), &["composed"]),
            (
                dictgen::InsensitiveStr::Ascii("combaination"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangt"), &["arrangement"]),
            (
                dictgen::InsensitiveStr::Ascii("accelration"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebellis"), &["rebellious"]),
            (dictgen::InsensitiveStr::Ascii("patternn"), &["pattern"]),
            (dictgen::InsensitiveStr::Ascii("articals"), &["articles"]),
            (dictgen::InsensitiveStr::Ascii("functionon"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("referering"),
                &["referring", "refereeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutualy"), &["mutually"]),
            (
                dictgen::InsensitiveStr::Ascii("alternavtives"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requiing"),
                &["requiring", "requiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("losted"),
                &["listed", "lost", "lasted"],
            ),
            (dictgen::InsensitiveStr::Ascii("teaher"), &["teacher"]),
            (dictgen::InsensitiveStr::Ascii("objcted"), &["objected"]),
            (dictgen::InsensitiveStr::Ascii("specficies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowiedged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstrations"),
                &["abstractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsidiezed"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulsr"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("languahe"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("technolog"), &["technology"]),
            (dictgen::InsensitiveStr::Ascii("repaces"), &["replaces"]),
            (dictgen::InsensitiveStr::Ascii("scrols"), &["scrolls"]),
            (
                dictgen::InsensitiveStr::Ascii("indentleveal"),
                &["indentlevel"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanaced"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("anpitheater"),
                &["amphitheater"],
            ),
            (dictgen::InsensitiveStr::Ascii("assisant"), &["assistant"]),
            (dictgen::InsensitiveStr::Ascii("buildls"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("weeving"), &["weaving"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitalises"),
                &["uninitialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenesless"),
                &["defenseless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compulsary"),
                &["compulsory"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyrpto"), &["crypto"]),
            (dictgen::InsensitiveStr::Ascii("confugire"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("idenity"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("mezmorizes"),
                &["mesmerizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sutisfy"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("preferas"), &["prefers"]),
            (
                dictgen::InsensitiveStr::Ascii("notofocations"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalueting"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deactivetes"),
                &["deactivates"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaecksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("templaets"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("assocated"), &["associated"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherisation"),
                &["characterisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicitions"),
                &["predictions", "predications"],
            ),
            (dictgen::InsensitiveStr::Ascii("engineeer"), &["engineer"]),
            (
                dictgen::InsensitiveStr::Ascii("problimatic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lood"),
                &["blood", "flood", "lewd", "look", "loom", "mood"],
            ),
            (dictgen::InsensitiveStr::Ascii("tucans"), &["toucans"]),
            (
                dictgen::InsensitiveStr::Ascii("initialzations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precedessor"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedamics"), &["academics"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetuas"),
                &["perpetuates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitializer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("chuks"), &["chunks"]),
            (dictgen::InsensitiveStr::Ascii("powerded"), &["powdered"]),
            (dictgen::InsensitiveStr::Ascii("rether"), &["rather"]),
            (
                dictgen::InsensitiveStr::Ascii("persents"),
                &["presents", "percents"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlyse"), &["analyse"]),
            (dictgen::InsensitiveStr::Ascii("extreams"), &["extremes"]),
            (dictgen::InsensitiveStr::Ascii("genereted"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("hach"),
                &["hatch", "hack", "hash"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribtion"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("telemery"), &["telemetry"]),
            (dictgen::InsensitiveStr::Ascii("precessed"), &["processed"]),
            (
                dictgen::InsensitiveStr::Ascii("exaggerrate"),
                &["exaggerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corospondance"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderered"), &["rendered"]),
            (dictgen::InsensitiveStr::Ascii("realeased"), &["released"]),
            (dictgen::InsensitiveStr::Ascii("draconain"), &["draconian"]),
            (dictgen::InsensitiveStr::Ascii("indentity"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("libertaryan"),
                &["libertarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desciribing"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("outbut"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("aguments"),
                &["arguments", "augments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangmenets"),
                &["rearrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquanitances"),
                &["acquaintance", "acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("updaing"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("ckeckpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeletion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("methed"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("inspiracion"),
                &["inspiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("recogizes"), &["recognizes"]),
            (dictgen::InsensitiveStr::Ascii("downgarte"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("missmatchd"),
                &["mismatched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addresse"),
                &["addresses", "address"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeperest"), &["creepers"]),
            (
                dictgen::InsensitiveStr::Ascii("rademptions"),
                &["redemptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rleasing"),
                &["releasing", "pleasing", "leasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuiding"), &["rebuilding"]),
            (
                dictgen::InsensitiveStr::Ascii("indiscrimnate"),
                &["indiscriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introdecks"),
                &["introduces"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaualts"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("confurmed"),
                &["confirmed", "conformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("devlops"), &["develops"]),
            (
                dictgen::InsensitiveStr::Ascii("preinting"),
                &["printing", "preprinting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetations"),
                &["repetitions", "reputations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hould"), &["hold", "should"]),
            (dictgen::InsensitiveStr::Ascii("inferrable"), &["inferable"]),
            (
                dictgen::InsensitiveStr::Ascii("carimonies"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emmisarries"),
                &["emissaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralelizes"),
                &["parallelizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitimization"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retribucion"),
                &["retribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermodynamcs"),
                &["thermodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invinicble"),
                &["invincible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verifyable"),
                &["verifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unanomusly"),
                &["unanimously"],
            ),
            (dictgen::InsensitiveStr::Ascii("baptisim"), &["baptism"]),
            (
                dictgen::InsensitiveStr::Ascii("cyphertext"),
                &["ciphertext"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasnfer"), &["transfer"]),
            (dictgen::InsensitiveStr::Ascii("satsohi"), &["satoshi"]),
            (
                dictgen::InsensitiveStr::Ascii("interseccted"),
                &["intersected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremement"),
                &["increment"],
            ),
            (dictgen::InsensitiveStr::Ascii("chrsitmas"), &["christmas"]),
            (dictgen::InsensitiveStr::Ascii("efficency"), &["efficiency"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcelated"),
                &["recalculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("durabilites"),
                &["durabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("compnaies"), &["companies"]),
            (dictgen::InsensitiveStr::Ascii("widesread"), &["widespread"]),
            (dictgen::InsensitiveStr::Ascii("triptish"), &["triptych"]),
            (dictgen::InsensitiveStr::Ascii("detatiled"), &["detailed"]),
            (
                dictgen::InsensitiveStr::Ascii("posion"),
                &["poison", "psion", "position"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corporacion"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfroming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuigring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("healthire"), &["healthier"]),
            (dictgen::InsensitiveStr::Ascii("functonal"), &["functional"]),
            (
                dictgen::InsensitiveStr::Ascii("unrepentent"),
                &["unrepentant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibil"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("equalizin"), &["equalizing"]),
            (
                dictgen::InsensitiveStr::Ascii("sbsampling"),
                &["subsampling"],
            ),
            (dictgen::InsensitiveStr::Ascii("peciluar"), &["peculiar"]),
            (dictgen::InsensitiveStr::Ascii("syatems"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("distacned"), &["distanced"]),
            (
                dictgen::InsensitiveStr::Ascii("legimitate"),
                &["legitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroe"), &["score", "scrod"]),
            (dictgen::InsensitiveStr::Ascii("bourgoeis"), &["bourgeois"]),
            (
                dictgen::InsensitiveStr::Ascii("restaraunteur"),
                &["restaurateur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demudulator"),
                &["demodulator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corosbonding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("gradaully"), &["gradually"]),
            (dictgen::InsensitiveStr::Ascii("volentier"), &["volunteer"]),
            (dictgen::InsensitiveStr::Ascii("primitivs"), &["primitives"]),
            (
                dictgen::InsensitiveStr::Ascii("coallescence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characeristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremements"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniputing"),
                &["manipulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realtionships"),
                &["relationships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompaitiblity"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograhper"),
                &["photographer"],
            ),
            (dictgen::InsensitiveStr::Ascii("taits"), &["taints"]),
            (dictgen::InsensitiveStr::Ascii("baasically"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("dentering"), &["centering"]),
            (
                dictgen::InsensitiveStr::Ascii("developoment"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("openened"), &["opened"]),
            (dictgen::InsensitiveStr::Ascii("especialy"), &["especially"]),
            (
                dictgen::InsensitiveStr::Ascii("develoment"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("assistsnt"), &["assistants"]),
            (
                dictgen::InsensitiveStr::Ascii("copenaghen"),
                &["copenhagen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegosiations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("efects"), &["effects"]),
            (
                dictgen::InsensitiveStr::Ascii("uneccesary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("queryies"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("iresistable"),
                &["irresistible"],
            ),
            (dictgen::InsensitiveStr::Ascii("sartups"), &["startups"]),
            (
                dictgen::InsensitiveStr::Ascii("accumalate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brigthness"),
                &["brightness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("championsihp"),
                &["championships"],
            ),
            (dictgen::InsensitiveStr::Ascii("desiger"), &["designer"]),
            (
                dictgen::InsensitiveStr::Ascii("underminding"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incooperates"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkt"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("eill"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("popultaion"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confussing"),
                &["confusing", "confessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbibtrarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("signfies"), &["signifies"]),
            (dictgen::InsensitiveStr::Ascii("fragements"), &["fragments"]),
            (dictgen::InsensitiveStr::Ascii("filesyste"), &["filesystem"]),
            (
                dictgen::InsensitiveStr::Ascii("maniuplate"),
                &["manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("packe"),
                &["packed", "packet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignenments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("differnt"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("difficulites"),
                &["difficulties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underastimate"),
                &["underestimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorced"),
                &["endorsed", "enforced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beavioural"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("meagtron"), &["megatron"]),
            (dictgen::InsensitiveStr::Ascii("properlly"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("congradulates"),
                &["congratulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("changning"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("copyritght"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("volitale"), &["volatile"]),
            (
                dictgen::InsensitiveStr::Ascii("accounants"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("vanila"), &["vanilla"]),
            (
                dictgen::InsensitiveStr::Ascii("workaroumd"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("volontary"), &["voluntary"]),
            (
                dictgen::InsensitiveStr::Ascii("broacasting"),
                &["broadcasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competions"),
                &["completions", "competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valentones"),
                &["valentines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undersetimated"),
                &["underestimated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correposding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transgenger"),
                &["transgender"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizell"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("expressivos"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("batsita"), &["batista"]),
            (
                dictgen::InsensitiveStr::Ascii("recconenction"),
                &["reconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecodte"), &["anecdote"]),
            (
                dictgen::InsensitiveStr::Ascii("incorrecty"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("enahnces"), &["enhances"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubtors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversubscibe"),
                &["oversubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("modee"), &["mode"]),
            (
                dictgen::InsensitiveStr::Ascii("propective"),
                &["projective", "prospective", "protective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoritical"),
                &["theoretical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthenticators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contibuted"),
                &["contributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribyes"), &["attributes"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiatians"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("exculding"), &["excluding"]),
            (
                dictgen::InsensitiveStr::Ascii("marshamllow"),
                &["marshmallow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrrided"),
                &["overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomment"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("allocae"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("hospitalzed"),
                &["hospitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particluar"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("finialise"), &["finalise"]),
            (dictgen::InsensitiveStr::Ascii("envlope"), &["envelope"]),
            (dictgen::InsensitiveStr::Ascii("bitamps"), &["bitmaps"]),
            (
                dictgen::InsensitiveStr::Ascii("inhomogenous"),
                &["inhomogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimmicked"), &["mimicked"]),
            (
                dictgen::InsensitiveStr::Ascii("dehyrdation"),
                &["dehydration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoticates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("toxicitity"), &["toxicity"]),
            (dictgen::InsensitiveStr::Ascii("soovinneer"), &["souvenir"]),
            (dictgen::InsensitiveStr::Ascii("echanism"), &["mechanism"]),
            (dictgen::InsensitiveStr::Ascii("chedule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("reimpplement"),
                &["reimplement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registrering"),
                &["registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininos"), &["minions"]),
            (dictgen::InsensitiveStr::Ascii("referiang"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("lolerant"), &["tolerant"]),
            (dictgen::InsensitiveStr::Ascii("systax"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("posin"),
                &["posing", "posit", "rosin"],
            ),
            (dictgen::InsensitiveStr::Ascii("poentials"), &["potentials"]),
            (dictgen::InsensitiveStr::Ascii("swtiching"), &["switching"]),
            (dictgen::InsensitiveStr::Ascii("abritrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("ethincity"), &["ethnicity"]),
            (
                dictgen::InsensitiveStr::Ascii("complicaitons"),
                &["complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakcends"), &["backends"]),
            (
                dictgen::InsensitiveStr::Ascii("separartions"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coerse"),
                &["coerce", "coarse", "course"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpiled"), &["compiled"]),
            (dictgen::InsensitiveStr::Ascii("socal"), &["social"]),
            (
                dictgen::InsensitiveStr::Ascii("transporation"),
                &["transportation"],
            ),
            (dictgen::InsensitiveStr::Ascii("guiness"), &["guinness"]),
            (
                dictgen::InsensitiveStr::Ascii("topoligically"),
                &["topologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriubuting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("protfolio"), &["portfolio"]),
            (
                dictgen::InsensitiveStr::Ascii("emplyments"),
                &["employments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lexographic"),
                &["lexicographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contemporaneus"),
                &["contemporaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derivitive"),
                &["derivative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intergal"),
                &["integral", "interval", "internal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utilizization"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventurious"),
                &["adventurous"],
            ),
            (dictgen::InsensitiveStr::Ascii("penguings"), &["penguins"]),
            (
                dictgen::InsensitiveStr::Ascii("impovement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucessfl"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("equesting"), &["requesting"]),
            (
                dictgen::InsensitiveStr::Ascii("overlowing"),
                &["overflowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("observabe"), &["observable"]),
            (
                dictgen::InsensitiveStr::Ascii("additionnally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("trascend"), &["transcend"]),
            (dictgen::InsensitiveStr::Ascii("quations"), &["equations"]),
            (
                dictgen::InsensitiveStr::Ascii("plcaements"),
                &["placements"],
            ),
            (dictgen::InsensitiveStr::Ascii("facorites"), &["favorites"]),
            (dictgen::InsensitiveStr::Ascii("dilineate"), &["delineate"]),
            (dictgen::InsensitiveStr::Ascii("stadnard"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoindences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cyclindrically"),
                &["cylindrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proceding"),
                &["proceeding", "preceding"],
            ),
            (dictgen::InsensitiveStr::Ascii("trunctated"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("sadning"), &["sanding"]),
            (
                dictgen::InsensitiveStr::Ascii("improvmenets"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("provder"), &["provider"]),
            (
                dictgen::InsensitiveStr::Ascii("characterisitcs"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vetoring"),
                &["vectoring", "vetoing"],
            ),
            (dictgen::InsensitiveStr::Ascii("summenors"), &["summoners"]),
            (dictgen::InsensitiveStr::Ascii("utiilty"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("accruacy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("sincerly"), &["sincerely"]),
            (dictgen::InsensitiveStr::Ascii("iternating"), &["iterating"]),
            (
                dictgen::InsensitiveStr::Ascii("philosohpy"),
                &["philosophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrationaly"),
                &["irrationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("brite"), &["bright"]),
            (dictgen::InsensitiveStr::Ascii("aspec"), &["aspect"]),
            (dictgen::InsensitiveStr::Ascii("senaireos"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("resonsibility"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unprecendeted"),
                &["unprecedented"],
            ),
            (dictgen::InsensitiveStr::Ascii("othopedic"), &["orthopedic"]),
            (dictgen::InsensitiveStr::Ascii("opaqe"), &["opaque"]),
            (dictgen::InsensitiveStr::Ascii("syled"), &["styled"]),
            (
                dictgen::InsensitiveStr::Ascii("represeting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("suprisd"), &["surprised"]),
            (
                dictgen::InsensitiveStr::Ascii("prioritisin"),
                &["prioritising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasitions"),
                &["transitions", "traditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fullfilment"),
                &["fulfilment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rebiulding"),
                &["rebuilding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inneffectual"),
                &["ineffectual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sociopatic"),
                &["sociopathic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurizes"),
                &["pasteurizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("feodras"), &["fedoras"]),
            (
                dictgen::InsensitiveStr::Ascii("dispalcements"),
                &["displacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("makretplaces"),
                &["marketplaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimisers"), &["optimisers"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiifiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visuabised"),
                &["visualised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("puncutation"),
                &["punctuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scheduleing"),
                &["scheduling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empahsized"),
                &["emphasized"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefernec"), &["preference"]),
            (
                dictgen::InsensitiveStr::Ascii("radiaoctive"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("diverstiy"), &["diversity"]),
            (
                dictgen::InsensitiveStr::Ascii("trinagulated"),
                &["triangulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sizeble"),
                &["sizable", "sizeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistakently"),
                &["mistakenly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootraped"),
                &["bootstrapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("squarey"), &["squarely"]),
            (dictgen::InsensitiveStr::Ascii("daimonds"), &["diamonds"]),
            (
                dictgen::InsensitiveStr::Ascii("verficiation"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("convesion"), &["conversion"]),
            (
                dictgen::InsensitiveStr::Ascii("pedictions"),
                &["depictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonogotiation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("travesy"), &["travesty"]),
            (dictgen::InsensitiveStr::Ascii("cricketts"), &["crickets"]),
            (dictgen::InsensitiveStr::Ascii("experence"), &["experience"]),
            (dictgen::InsensitiveStr::Ascii("asphlat"), &["asphalt"]),
            (dictgen::InsensitiveStr::Ascii("obivous"), &["obvious"]),
            (dictgen::InsensitiveStr::Ascii("positoin"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("bondary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("conidtion"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("insignificent"),
                &["insignificant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decomressed"),
                &["decompressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriebutes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parantheses"),
                &["parentheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("demenor"), &["demeanor"]),
            (
                dictgen::InsensitiveStr::Ascii("traditionnal"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("garantor"), &["guarantor"]),
            (dictgen::InsensitiveStr::Ascii("debugg"), &["debug"]),
            (dictgen::InsensitiveStr::Ascii("jornal"), &["journal"]),
            (
                dictgen::InsensitiveStr::Ascii("constratins"),
                &["constraints", "constrains"],
            ),
            (dictgen::InsensitiveStr::Ascii("defiened"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("memroy"), &["memory"]),
            (
                dictgen::InsensitiveStr::Ascii("distribtutor"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("peculure"), &["peculiar"]),
            (
                dictgen::InsensitiveStr::Ascii("dimesionalities"),
                &["dimensionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("betrayeado"), &["betrayed"]),
            (dictgen::InsensitiveStr::Ascii("sequenece"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("definitlely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metephorically"),
                &["metaphorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("followoing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("replacted"),
                &["replaced", "replicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnates"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinar"),
                &["coordinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizzles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("parrallelized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persits"),
                &["persist", "persists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaponsibility"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppoerters"),
                &["supporters"],
            ),
            (dictgen::InsensitiveStr::Ascii("differetn"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("slaughted"),
                &["slaughtered"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoolute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("neighours"), &["neighbours"]),
            (dictgen::InsensitiveStr::Ascii("evalutors"), &["evaluators"]),
            (
                dictgen::InsensitiveStr::Ascii("initiazing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("variaties"), &["varieties"]),
            (
                dictgen::InsensitiveStr::Ascii("architetcure"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("ublic"), &["public"]),
            (dictgen::InsensitiveStr::Ascii("acchieve"), &["achieve"]),
            (
                dictgen::InsensitiveStr::Ascii("emplying"),
                &["employing", "implying", "emptying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermeneters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palestinan"),
                &["palestinian"],
            ),
            (dictgen::InsensitiveStr::Ascii("hierarachy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("exectutor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("anticipacion"),
                &["anticipation"],
            ),
            (dictgen::InsensitiveStr::Ascii("degreee"), &["degree"]),
            (dictgen::InsensitiveStr::Ascii("lavendr"), &["lavender"]),
            (dictgen::InsensitiveStr::Ascii("emporer"), &["emperor"]),
            (dictgen::InsensitiveStr::Ascii("modeifier"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("predicessors"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rebuildling"),
                &["rebuilding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valuse"),
                &["values", "value", "valise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemorrhagin"),
                &["hemorrhaging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrsepondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprsentation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cambrige"), &["cambridge"]),
            (
                dictgen::InsensitiveStr::Ascii("initialzie"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("offst"), &["offset"]),
            (dictgen::InsensitiveStr::Ascii("verrsion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("imddle"), &["middle"]),
            (dictgen::InsensitiveStr::Ascii("poniter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("handshages"),
                &["handshakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("simposn"), &["simpson"]),
            (
                dictgen::InsensitiveStr::Ascii("dumplicated"),
                &["duplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferr"), &["refer"]),
            (dictgen::InsensitiveStr::Ascii("jstu"), &["just"]),
            (
                dictgen::InsensitiveStr::Ascii("cabability"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("skyp"), &["skip", "skype"]),
            (dictgen::InsensitiveStr::Ascii("recognie"), &["recognizes"]),
            (dictgen::InsensitiveStr::Ascii("sizos"), &["sizes"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonated"),
                &["opinionated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mebrans"), &["membranes"]),
            (dictgen::InsensitiveStr::Ascii("centrase"), &["centres"]),
            (
                dictgen::InsensitiveStr::Ascii("rediculous"),
                &["ridiculous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geneological"),
                &["genealogical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowladge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handly"),
                &["handy", "handily"],
            ),
            (dictgen::InsensitiveStr::Ascii("attirbute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("virutals"), &["virtuals"]),
            (
                dictgen::InsensitiveStr::Ascii("viusalised"),
                &["visualised"],
            ),
            (dictgen::InsensitiveStr::Ascii("harge"), &["charge"]),
            (dictgen::InsensitiveStr::Ascii("porgram"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("transvormation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classsrooms"),
                &["classrooms"],
            ),
            (dictgen::InsensitiveStr::Ascii("parttition"), &["partition"]),
            (dictgen::InsensitiveStr::Ascii("assginees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("compatiable"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("fequent"), &["frequent"]),
            (dictgen::InsensitiveStr::Ascii("retreeved"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("elimiating"),
                &["eliminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fonctioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("eeproom"), &["eeprom"]),
            (
                dictgen::InsensitiveStr::Ascii("benifitted"),
                &["benefitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seramonial"),
                &["ceremonial"],
            ),
            (dictgen::InsensitiveStr::Ascii("penisula"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("birmignham"),
                &["birmingham"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nuetrality"),
                &["neutrality"],
            ),
            (dictgen::InsensitiveStr::Ascii("ohterwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("geneates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("comparrisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("decocde"), &["decode"]),
            (dictgen::InsensitiveStr::Ascii("alignmeent"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("chiziled"), &["chiseled"]),
            (dictgen::InsensitiveStr::Ascii("resotrers"), &["restorers"]),
            (dictgen::InsensitiveStr::Ascii("liberaries"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("calander"),
                &["calendar", "colander"],
            ),
            (dictgen::InsensitiveStr::Ascii("princple"), &["principle"]),
            (dictgen::InsensitiveStr::Ascii("motified"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("throast"), &["throats"]),
            (
                dictgen::InsensitiveStr::Ascii("cristallize"),
                &["crystallize"],
            ),
            (dictgen::InsensitiveStr::Ascii("anserers"), &["answerers"]),
            (dictgen::InsensitiveStr::Ascii("concensors"), &["consensus"]),
            (
                dictgen::InsensitiveStr::Ascii("helicopers"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("statitic"), &["statistic"]),
            (dictgen::InsensitiveStr::Ascii("mintue"), &["minute"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtorization"),
                &["authorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revelatons"),
                &["revelations"],
            ),
            (dictgen::InsensitiveStr::Ascii("addded"), &["added"]),
            (dictgen::InsensitiveStr::Ascii("claimes"), &["claims"]),
            (dictgen::InsensitiveStr::Ascii("longtiem"), &["longtime"]),
            (dictgen::InsensitiveStr::Ascii("charisa"), &["charisma"]),
            (
                dictgen::InsensitiveStr::Ascii("connstraint"),
                &["constraint"],
            ),
            (dictgen::InsensitiveStr::Ascii("leigons"), &["legions"]),
            (dictgen::InsensitiveStr::Ascii("temptative"), &["tentative"]),
            (dictgen::InsensitiveStr::Ascii("torrentes"), &["torrents"]),
            (
                dictgen::InsensitiveStr::Ascii("currespond"),
                &["correspond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swizterland"),
                &["switzerland"],
            ),
            (dictgen::InsensitiveStr::Ascii("spligs"), &["splits"]),
            (dictgen::InsensitiveStr::Ascii("guradians"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("paddding"), &["padding"]),
            (
                dictgen::InsensitiveStr::Ascii("blanck"),
                &["blank", "black"],
            ),
            (dictgen::InsensitiveStr::Ascii("liquidos"), &["liquids"]),
            (
                dictgen::InsensitiveStr::Ascii("overblocking"),
                &["overclocking"],
            ),
            (
                dictgen::InsensitiveStr::Unicode("clockwíse"),
                &["clockwise"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifacs"), &["artifacts"]),
            (dictgen::InsensitiveStr::Ascii("graet"), &["great"]),
            (
                dictgen::InsensitiveStr::Ascii("contrubite"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("buildnings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoteurs"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternatievly"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("assigneees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("eescription"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resumitted"),
                &["resubmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satsifactory"),
                &["satisfactory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disqualifed"),
                &["disqualified"],
            ),
            (dictgen::InsensitiveStr::Ascii("cambrdige"), &["cambridge"]),
            (dictgen::InsensitiveStr::Ascii("preculde"), &["preclude"]),
            (dictgen::InsensitiveStr::Ascii("extermest"), &["extremest"]),
            (
                dictgen::InsensitiveStr::Ascii("accomdating"),
                &["accommodating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagairism"),
                &["plagiarism"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatyes"), &["datatypes"]),
            (dictgen::InsensitiveStr::Ascii("illution"), &["illusion"]),
            (dictgen::InsensitiveStr::Ascii("halfitme"), &["halftime"]),
            (dictgen::InsensitiveStr::Ascii("jekins"), &["jenkins"]),
            (dictgen::InsensitiveStr::Ascii("befure"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("abritration"),
                &["arbitration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depands"),
                &["depends", "demands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropolians"),
                &["metropolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigacion"),
                &["investigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efficienty"),
                &["efficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimickin"), &["mimicking"]),
            (dictgen::InsensitiveStr::Ascii("normallise"), &["normalise"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaties"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("despiration"),
                &["desperation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentories"),
                &["documentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("conitnue"), &["continue"]),
            (dictgen::InsensitiveStr::Ascii("serimonys"), &["ceremonies"]),
            (dictgen::InsensitiveStr::Ascii("mastermid"), &["mastermind"]),
            (dictgen::InsensitiveStr::Ascii("microtax"), &["microatx"]),
            (dictgen::InsensitiveStr::Ascii("partiets"), &["parties"]),
            (
                dictgen::InsensitiveStr::Ascii("aribtrations"),
                &["arbitrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("quikly"), &["quickly"]),
            (dictgen::InsensitiveStr::Ascii("jont"), &["joint"]),
            (
                dictgen::InsensitiveStr::Ascii("resilence"),
                &["resilience", "residence"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchaged"), &["unchanged"]),
            (dictgen::InsensitiveStr::Ascii("immideate"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("perfroming"),
                &["performing"],
            ),
            (dictgen::InsensitiveStr::Ascii("autherity"), &["authority"]),
            (
                dictgen::InsensitiveStr::Ascii("pertrubing"),
                &["perturbing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implenetation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alvorithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destintions"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatlye"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replentishes"),
                &["replenishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("condemnig"), &["condemning"]),
            (
                dictgen::InsensitiveStr::Ascii("acquiessed"),
                &["acquiesced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depicitons"),
                &["depictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spesifications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("scripoted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("progressiste"),
                &["progresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("managar"), &["manager"]),
            (dictgen::InsensitiveStr::Ascii("emtpied"), &["emptied"]),
            (
                dictgen::InsensitiveStr::Ascii("parntership"),
                &["partnership"],
            ),
            (dictgen::InsensitiveStr::Ascii("israelies"), &["israelis"]),
            (
                dictgen::InsensitiveStr::Ascii("foramtting"),
                &["formatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civelesation"),
                &["civilisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checckpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("muyst"), &["must"]),
            (
                dictgen::InsensitiveStr::Ascii("millisencond"),
                &["millisecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahrges"), &["charges"]),
            (dictgen::InsensitiveStr::Ascii("spritual"), &["spiritual"]),
            (dictgen::InsensitiveStr::Ascii("perhas"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("ovride"), &["override"]),
            (dictgen::InsensitiveStr::Ascii("unmacthed"), &["unmatched"]),
            (
                dictgen::InsensitiveStr::Ascii("additiional"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("poointed"), &["pointed"]),
            (
                dictgen::InsensitiveStr::Ascii("stanardizing"),
                &["standardizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoroporate"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("writtin"),
                &["written", "writing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scroollable"),
                &["scrollable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallifer"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepdent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininstratively"),
                &["administratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("gernade"), &["grenade"]),
            (dictgen::InsensitiveStr::Ascii("expolits"), &["exploits"]),
            (dictgen::InsensitiveStr::Ascii("thik"), &["thick", "think"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhagun"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmits"), &["commits"]),
            (
                dictgen::InsensitiveStr::Ascii("punctution"),
                &["punctuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("trancodes"), &["transcodes"]),
            (dictgen::InsensitiveStr::Ascii("quesioned"), &["questioned"]),
            (
                dictgen::InsensitiveStr::Ascii("dispertion"),
                &["dispersion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deletetions"),
                &["deletions"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifing"), &["unifying"]),
            (
                dictgen::InsensitiveStr::Ascii("implaments"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distintions"),
                &["distinctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("anitime"), &["anytime"]),
            (dictgen::InsensitiveStr::Ascii("speperator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("voilation"), &["violation"]),
            (dictgen::InsensitiveStr::Ascii("globlal"), &["global"]),
            (dictgen::InsensitiveStr::Ascii("mimatch"), &["mismatch"]),
            (dictgen::InsensitiveStr::Ascii("objetions"), &["objections"]),
            (
                dictgen::InsensitiveStr::Ascii("hallucinaton"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("reprsent"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("nymber"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("destinaions"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coerscitive"),
                &["coercitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appearanes"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arcaheology"),
                &["archeology"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticies"), &["criticise"]),
            (
                dictgen::InsensitiveStr::Ascii("experineces"),
                &["experiences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("radioacive"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("rearrang"), &["rearrange"]),
            (
                dictgen::InsensitiveStr::Ascii("distinghishing"),
                &["distinguishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("singsog"), &["singsong"]),
            (
                dictgen::InsensitiveStr::Ascii("multpliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("formatt"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("dedikation"),
                &["dedication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facsinated"),
                &["fascinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("waterlemon"),
                &["watermelon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefied"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mantenance"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insctuctional"),
                &["instructional"],
            ),
            (dictgen::InsensitiveStr::Ascii("comfirms"), &["confirms"]),
            (dictgen::InsensitiveStr::Ascii("rumorus"), &["rumors"]),
            (
                dictgen::InsensitiveStr::Ascii("incorparates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scientificaly"),
                &["scientifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neader"),
                &["header", "kneader"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmisison"),
                &["transmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimenters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecor"), &["selector"]),
            (dictgen::InsensitiveStr::Ascii("burlgar"), &["burglar"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("grups"), &["groups"]),
            (dictgen::InsensitiveStr::Ascii("temproy"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("caputring"), &["capturing"]),
            (dictgen::InsensitiveStr::Ascii("interuupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("percetage"), &["percentage"]),
            (dictgen::InsensitiveStr::Ascii("porcessor"), &["processor"]),
            (
                dictgen::InsensitiveStr::Ascii("hoested"),
                &["hosted", "hoisted"],
            ),
            (dictgen::InsensitiveStr::Ascii("bachleor"), &["bachelor"]),
            (
                dictgen::InsensitiveStr::Ascii("approxamation"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disatissfied"),
                &["dissatisfied"],
            ),
            (dictgen::InsensitiveStr::Ascii("aviod"), &["avoid"]),
            (
                dictgen::InsensitiveStr::Ascii("firefigter"),
                &["firefighter"],
            ),
            (dictgen::InsensitiveStr::Ascii("chorline"), &["chlorine"]),
            (dictgen::InsensitiveStr::Ascii("sellection"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("prefarable"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colections"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimmental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustartions"),
                &["illustrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("calrity"), &["clarity"]),
            (dictgen::InsensitiveStr::Ascii("therwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("recalim"), &["reclaim"]),
            (
                dictgen::InsensitiveStr::Ascii("spped"),
                &["speed", "sped", "sipped", "sapped", "supped", "sopped"],
            ),
            (dictgen::InsensitiveStr::Ascii("surevy"), &["survey"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotite"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectualy"),
                &["intellectually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmergable"),
                &["unmergeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("debending"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("exploracion"),
                &["exploration"],
            ),
            (dictgen::InsensitiveStr::Ascii("imgrants"), &["migrants"]),
            (dictgen::InsensitiveStr::Ascii("akumulate"), &["accumulate"]),
            (dictgen::InsensitiveStr::Ascii("intriduce"), &["introduce"]),
            (
                dictgen::InsensitiveStr::Ascii("perticipating"),
                &["participating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investingates"),
                &["investigates"],
            ),
            (dictgen::InsensitiveStr::Ascii("consturct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("otherise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("dissasters"), &["disasters"]),
            (
                dictgen::InsensitiveStr::Ascii("plagerised"),
                &["plagiarised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appearence"),
                &["appearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissimiar"), &["dissimilar"]),
            (dictgen::InsensitiveStr::Ascii("resembels"), &["resembles"]),
            (dictgen::InsensitiveStr::Ascii("cotave"), &["octave"]),
            (dictgen::InsensitiveStr::Ascii("beseiged"), &["besieged"]),
            (dictgen::InsensitiveStr::Ascii("functon"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("identfy"), &["identify"]),
            (
                dictgen::InsensitiveStr::Ascii("respnsible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deamonified"),
                &["daemonised", "daemonized"],
            ),
            (dictgen::InsensitiveStr::Ascii("loclas"), &["locals"]),
            (dictgen::InsensitiveStr::Ascii("convulute"), &["convolute"]),
            (dictgen::InsensitiveStr::Ascii("agressie"), &["aggressive"]),
            (dictgen::InsensitiveStr::Ascii("asside"), &["aside"]),
            (dictgen::InsensitiveStr::Ascii("astethic"), &["aesthetic"]),
            (dictgen::InsensitiveStr::Ascii("relocaing"), &["relocating"]),
            (
                dictgen::InsensitiveStr::Ascii("increading"),
                &["increasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubitor"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("timere"), &["timer"]),
            (dictgen::InsensitiveStr::Ascii("analouge"), &["analogue"]),
            (dictgen::InsensitiveStr::Ascii("acadmia"), &["academia"]),
            (dictgen::InsensitiveStr::Ascii("watermask"), &["watermark"]),
            (dictgen::InsensitiveStr::Ascii("geocentic"), &["geocentric"]),
            (dictgen::InsensitiveStr::Ascii("royalites"), &["royalties"]),
            (
                dictgen::InsensitiveStr::Ascii("activisits"),
                &["activist", "activists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpletely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inituialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("amaters"), &["amateurs"]),
            (
                dictgen::InsensitiveStr::Ascii("estbalishment"),
                &["establishments", "establishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesed"), &["accessed"]),
            (dictgen::InsensitiveStr::Ascii("animattes"), &["animates"]),
            (
                dictgen::InsensitiveStr::Ascii("idetifying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("absovled"), &["absolved"]),
            (dictgen::InsensitiveStr::Ascii("execuctes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("downlads"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("adderss"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("avialible"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("acutality"), &["actuality"]),
            (dictgen::InsensitiveStr::Ascii("accepetors"), &["acceptors"]),
            (
                dictgen::InsensitiveStr::Ascii("perpandicularly"),
                &["perpendicularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("puinsher"), &["punisher"]),
            (
                dictgen::InsensitiveStr::Ascii("refrubished"),
                &["refurbished"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressolver"), &["resolver"]),
            (
                dictgen::InsensitiveStr::Ascii("rountriping"),
                &["roundtripping"],
            ),
            (dictgen::InsensitiveStr::Ascii("deifne"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("analtyic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("propagationn"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auxilliaries"),
                &["auxiliaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("parameds"), &["paramedics"]),
            (dictgen::InsensitiveStr::Ascii("reffering"), &["referring"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensinal"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("implemend"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("supresses"), &["suppresses"]),
            (dictgen::InsensitiveStr::Ascii("blatanty"), &["blatantly"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotioations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preficiency"),
                &["proficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charaterization"),
                &["characterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("componens"), &["components"]),
            (dictgen::InsensitiveStr::Ascii("ukranian"), &["ukrainian"]),
            (
                dictgen::InsensitiveStr::Ascii("unqiueness"),
                &["uniqueness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpetrater"),
                &["perpetrator", "perpetrated", "perpetrates", "perpetrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("callabcks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("parralellization"),
                &["parallelization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cornders"), &["corners"]),
            (dictgen::InsensitiveStr::Ascii("prposiing"), &["proposing"]),
            (
                dictgen::InsensitiveStr::Ascii("quarted"),
                &["quarter", "quartered"],
            ),
            (dictgen::InsensitiveStr::Ascii("substitue"), &["substitute"]),
            (dictgen::InsensitiveStr::Ascii("goergia"), &["georgia"]),
            (
                dictgen::InsensitiveStr::Ascii("convertions"),
                &["conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indictating"),
                &["indicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrresponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("tonange"), &["tonnage"]),
            (
                dictgen::InsensitiveStr::Ascii("compenents"),
                &["components", "competence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emergenies"),
                &["emergencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operacional"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaeses"), &["releases"]),
            (dictgen::InsensitiveStr::Ascii("pressre"), &["pressure"]),
            (
                dictgen::InsensitiveStr::Ascii("trancoding"),
                &["transcoding"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuurrents"), &["currents"]),
            (
                dictgen::InsensitiveStr::Ascii("containings"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoluts"), &["absolutes"]),
            (
                dictgen::InsensitiveStr::Ascii("convicitons"),
                &["convictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheduald"), &["scheduled"]),
            (dictgen::InsensitiveStr::Ascii("forr"), &["for"]),
            (
                dictgen::InsensitiveStr::Ascii("comrpessed"),
                &["compressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyklopedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("retaurant"), &["restaurant"]),
            (dictgen::InsensitiveStr::Ascii("coleasing"), &["coalescing"]),
            (dictgen::InsensitiveStr::Ascii("acronmys"), &["acronyms"]),
            (
                dictgen::InsensitiveStr::Ascii("chromosoom"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("jurnaling"), &["journaling"]),
            (
                dictgen::InsensitiveStr::Ascii("parapharsing"),
                &["paraphrasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("optet"), &["opted"]),
            (
                dictgen::InsensitiveStr::Ascii("creationnism"),
                &["creationism"],
            ),
            (dictgen::InsensitiveStr::Ascii("courching"), &["crouching"]),
            (
                dictgen::InsensitiveStr::Ascii("inapporpriately"),
                &["inappropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicfics"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("compilles"), &["compiles"]),
            (dictgen::InsensitiveStr::Ascii("perforamed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("doccument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("justificatons"),
                &["justifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("coodinate"), &["coordinate"]),
            (
                dictgen::InsensitiveStr::Ascii("concentartion"),
                &["concentrations", "concentration"],
            ),
            (dictgen::InsensitiveStr::Ascii("listensers"), &["listeners"]),
            (dictgen::InsensitiveStr::Ascii("lesbianese"), &["lesbians"]),
            (dictgen::InsensitiveStr::Ascii("definetyl"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("eddited"), &["edited"]),
            (
                dictgen::InsensitiveStr::Ascii("appareances"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corporativo"),
                &["corporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("analig"), &["analog"]),
            (
                dictgen::InsensitiveStr::Ascii("intrerupted"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovrridden"), &["overridden"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapeared"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("eacf"), &["each"]),
            (
                dictgen::InsensitiveStr::Ascii("accepectably"),
                &["acceptably"],
            ),
            (dictgen::InsensitiveStr::Ascii("ptichfork"), &["pitchfork"]),
            (
                dictgen::InsensitiveStr::Ascii("traslalation"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forminable"),
                &["formidable"],
            ),
            (dictgen::InsensitiveStr::Ascii("toritlla"), &["tortilla"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpetctedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("aboslve"), &["absolve"]),
            (dictgen::InsensitiveStr::Ascii("randomez"), &["randomized"]),
            (
                dictgen::InsensitiveStr::Ascii("algotihmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("satuadays"), &["saturdays"]),
            (
                dictgen::InsensitiveStr::Ascii("livecycles"),
                &["lifecycles"],
            ),
            (dictgen::InsensitiveStr::Ascii("globla"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("comstraint"),
                &["constraint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("farmasudic"),
                &["pharmaceutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captialized"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimiate"),
                &["legitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sencondary"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("complainging"),
                &["complaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("crystalys"), &["crystals"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomforyable"),
                &["uncomfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviencece"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introdutions"),
                &["introductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("gernerated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("verificaiton"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("brakpoint"), &["breakpoint"]),
            (dictgen::InsensitiveStr::Ascii("advsiors"), &["advisors"]),
            (dictgen::InsensitiveStr::Ascii("folllow"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("warninigs"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("academica"), &["academia"]),
            (dictgen::InsensitiveStr::Ascii("soldies"), &["soldiers"]),
            (
                dictgen::InsensitiveStr::Ascii("maintenenace"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrtizations"),
                &["discretizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("downladed"), &["downloaded"]),
            (dictgen::InsensitiveStr::Ascii("preversed"), &["preserved"]),
            (
                dictgen::InsensitiveStr::Ascii("continient"),
                &["contingent"],
            ),
            (dictgen::InsensitiveStr::Ascii("pleasse"), &["please"]),
            (dictgen::InsensitiveStr::Ascii("stratagy"), &["strategy"]),
            (dictgen::InsensitiveStr::Ascii("qualfied"), &["qualified"]),
            (
                dictgen::InsensitiveStr::Ascii("multipiler"),
                &["multiplier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggesst"),
                &["suggests", "suggest"],
            ),
            (dictgen::InsensitiveStr::Ascii("metabloic"), &["metabolic"]),
            (dictgen::InsensitiveStr::Ascii("adjasence"), &["adjacence"]),
            (
                dictgen::InsensitiveStr::Ascii("idealogies"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("develpo"), &["develop"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificie"),
                &["sacrificed"],
            ),
            (dictgen::InsensitiveStr::Ascii("aroung"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("creditted"), &["credited"]),
            (
                dictgen::InsensitiveStr::Ascii("especcialy"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("wolrdly"), &["worldly"]),
            (dictgen::InsensitiveStr::Ascii("confgure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("marjiuana"), &["marijuana"]),
            (dictgen::InsensitiveStr::Ascii("spaltoon"), &["splatoon"]),
            (
                dictgen::InsensitiveStr::Ascii("coeficient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("tansact"), &["transact"]),
            (dictgen::InsensitiveStr::Ascii("tranport"), &["transport"]),
            (
                dictgen::InsensitiveStr::Ascii("retalitated"),
                &["retaliated"],
            ),
            (dictgen::InsensitiveStr::Ascii("avatards"), &["avatars"]),
            (dictgen::InsensitiveStr::Ascii("follinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("belifs"), &["beliefs"]),
            (dictgen::InsensitiveStr::Ascii("alignmnent"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("mispell"), &["misspell"]),
            (
                dictgen::InsensitiveStr::Ascii("suggetsions"),
                &["suggestions"],
            ),
            (dictgen::InsensitiveStr::Ascii("menstural"), &["menstrual"]),
            (dictgen::InsensitiveStr::Ascii("mathewes"), &["mathews"]),
            (
                dictgen::InsensitiveStr::Ascii("experiement"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("condions"), &["conditions"]),
            (dictgen::InsensitiveStr::Ascii("overboad"), &["overboard"]),
            (
                dictgen::InsensitiveStr::Ascii("experienceing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dynamnically"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pesticedes"),
                &["pesticides"],
            ),
            (dictgen::InsensitiveStr::Ascii("benedickt"), &["benedict"]),
            (
                dictgen::InsensitiveStr::Ascii("replacaments"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("counsil"), &["counsel"]),
            (
                dictgen::InsensitiveStr::Ascii("coruptions"),
                &["corruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekc"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("youa"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("unale"), &["unable"]),
            (dictgen::InsensitiveStr::Ascii("metropols"), &["metropolis"]),
            (dictgen::InsensitiveStr::Ascii("admited"), &["admitted"]),
            (dictgen::InsensitiveStr::Ascii("fracter"), &["fracture"]),
            (dictgen::InsensitiveStr::Ascii("chateo"), &["château"]),
            (dictgen::InsensitiveStr::Ascii("spellign"), &["spelling"]),
            (
                dictgen::InsensitiveStr::Ascii("customing"),
                &["customizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conenct"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("contanting"),
                &["contacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("reques"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("democracts"), &["democrat"]),
            (
                dictgen::InsensitiveStr::Ascii("highlightning"),
                &["highlighting"],
            ),
            (dictgen::InsensitiveStr::Ascii("withhin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("liesure"), &["leisure"]),
            (
                dictgen::InsensitiveStr::Ascii("excalmation"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chamnge"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("moderateurs"),
                &["moderates"],
            ),
            (dictgen::InsensitiveStr::Ascii("pheonix"), &["phoenix"]),
            (dictgen::InsensitiveStr::Ascii("exlain"), &["explain"]),
            (
                dictgen::InsensitiveStr::Ascii("legitamate"),
                &["legitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabilty"),
                &["probability"],
            ),
            (dictgen::InsensitiveStr::Ascii("detialing"), &["detailing"]),
            (dictgen::InsensitiveStr::Ascii("topologie"), &["topology"]),
            (dictgen::InsensitiveStr::Ascii("dipatch"), &["dispatch"]),
            (dictgen::InsensitiveStr::Ascii("focuse"), &["focus"]),
            (
                dictgen::InsensitiveStr::Ascii("genreations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheeckpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("deviices"), &["devices"]),
            (dictgen::InsensitiveStr::Ascii("folers"), &["folders"]),
            (
                dictgen::InsensitiveStr::Ascii("extrimists"),
                &["extremists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desserializes"),
                &["deserializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("redaundant"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("disambiguaing"),
                &["disambiguating"],
            ),
            (dictgen::InsensitiveStr::Ascii("initialls"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("experments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mansluaghter"),
                &["manslaughter"],
            ),
            (dictgen::InsensitiveStr::Ascii("handele"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("emapthy"), &["empathy"]),
            (dictgen::InsensitiveStr::Ascii("galdiator"), &["gladiator"]),
            (dictgen::InsensitiveStr::Ascii("draculea"), &["dracula"]),
            (dictgen::InsensitiveStr::Ascii("combins"), &["combines"]),
            (dictgen::InsensitiveStr::Ascii("informe"), &["inform"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptios"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mising"), &["missing"]),
            (dictgen::InsensitiveStr::Ascii("chossen"), &["chosen"]),
            (dictgen::InsensitiveStr::Ascii("succedeed"), &["succeeded"]),
            (dictgen::InsensitiveStr::Ascii("shortuts"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("qery"), &["query"]),
            (dictgen::InsensitiveStr::Ascii("indicaite"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("dicitonary"),
                &["dictionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("birmimgham"),
                &["birmingham"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parenthises"),
                &["parenthesis", "parentheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("aray"), &["array"]),
            (
                dictgen::InsensitiveStr::Ascii("nexted"),
                &["nested", "texted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dsyfunction"),
                &["dysfunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarranted"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commiti"),
                &["committee", "committing", "commit"],
            ),
            (dictgen::InsensitiveStr::Ascii("acepted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("dicslosure"),
                &["disclosure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redesinging"),
                &["redesigning"],
            ),
            (dictgen::InsensitiveStr::Ascii("songlar"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("popularaty"),
                &["popularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqential"), &["sequential"]),
            (dictgen::InsensitiveStr::Ascii("availbel"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("commonds"),
                &["commands", "commons", "commends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overcompansated"),
                &["overcompensated"],
            ),
            (dictgen::InsensitiveStr::Ascii("decion"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("genarator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("naugty"), &["naughty"]),
            (
                dictgen::InsensitiveStr::Ascii("essentually"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ridicoulus"),
                &["ridiculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("falsh"), &["flash", "false"]),
            (
                dictgen::InsensitiveStr::Ascii("apppearance"),
                &["appearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaukt"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("trunctate"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("monagomy"), &["monogamy"]),
            (dictgen::InsensitiveStr::Ascii("resoltion"), &["resolution"]),
            (
                dictgen::InsensitiveStr::Ascii("achiavable"),
                &["achievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("earthqaukes"),
                &["earthquakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("magicin"), &["magician"]),
            (dictgen::InsensitiveStr::Ascii("trainig"), &["training"]),
            (
                dictgen::InsensitiveStr::Ascii("imortantly"),
                &["importantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciprocite"),
                &["reciprocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trancation"),
                &["truncation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disaggretates"),
                &["disaggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminatorn"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skateboaring"),
                &["skateboarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmemorated"),
                &["commemorated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("originiated"),
                &["originated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fractoinal"),
                &["fractional"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoy"), &["annoy"]),
            (dictgen::InsensitiveStr::Ascii("heapdhone"), &["headphone"]),
            (
                dictgen::InsensitiveStr::Ascii("enternal"),
                &["internal", "external", "eternal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asssociation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brigde"),
                &["bridge", "brigade"],
            ),
            (dictgen::InsensitiveStr::Ascii("eiter"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("behabiours"),
                &["behaviours"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocal"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubitions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("resitor"), &["resistor"]),
            (dictgen::InsensitiveStr::Ascii("qestioned"), &["questioned"]),
            (dictgen::InsensitiveStr::Ascii("ahered"), &["adhered"]),
            (
                dictgen::InsensitiveStr::Ascii("sestatusmsg"),
                &["setstatusmsg"],
            ),
            (dictgen::InsensitiveStr::Ascii("ophans"), &["orphans"]),
            (
                dictgen::InsensitiveStr::Ascii("handkerchifs"),
                &["handkerchiefs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gracelfuly"),
                &["gracefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("corretor"), &["corrector"]),
            (
                dictgen::InsensitiveStr::Ascii("devritualization"),
                &["devirtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulds"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("scrirpt"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("noth"), &["north"]),
            (
                dictgen::InsensitiveStr::Ascii("chrashs"),
                &["crashes", "thrashes", "trashes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepedantly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("interanl"), &["internal"]),
            (dictgen::InsensitiveStr::Ascii("spansih"), &["spanish"]),
            (
                dictgen::InsensitiveStr::Ascii("abadonding"),
                &["abandoning"],
            ),
            (dictgen::InsensitiveStr::Ascii("covinces"), &["convinces"]),
            (
                dictgen::InsensitiveStr::Ascii("concatened"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misrepresentated"),
                &["misrepresented"],
            ),
            (dictgen::InsensitiveStr::Ascii("singualr"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawl"),
                &["withdrawal", "withdraw"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculador"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("udpatable"), &["updatable"]),
            (dictgen::InsensitiveStr::Ascii("driect"), &["direct"]),
            (dictgen::InsensitiveStr::Ascii("gobal"), &["global"]),
            (dictgen::InsensitiveStr::Ascii("forceably"), &["forcibly"]),
            (
                dictgen::InsensitiveStr::Ascii("highe"),
                &["high", "higher", "highs"],
            ),
            (dictgen::InsensitiveStr::Ascii("identiied"), &["identified"]),
            (dictgen::InsensitiveStr::Ascii("cylcops"), &["cyclops"]),
            (dictgen::InsensitiveStr::Ascii("rheotric"), &["rhetoric"]),
            (dictgen::InsensitiveStr::Ascii("compressd"), &["compressed"]),
            (
                dictgen::InsensitiveStr::Ascii("dusfunctional"),
                &["dysfunctional"],
            ),
            (dictgen::InsensitiveStr::Ascii("capsulse"), &["capsules"]),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonisation"),
                &["synchronisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantenous"),
                &["instantaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("etiquete"), &["etiquette"]),
            (
                dictgen::InsensitiveStr::Ascii("atmoshpere"),
                &["atmosphere"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfieirs"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("trascription"),
                &["transcription"],
            ),
            (dictgen::InsensitiveStr::Ascii("scavanged"), &["scavenged"]),
            (dictgen::InsensitiveStr::Ascii("cuases"), &["causes"]),
            (dictgen::InsensitiveStr::Ascii("konwledge"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("vulneable"), &["vulnerable"]),
            (dictgen::InsensitiveStr::Ascii("symboll"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("declartators"),
                &["declarators"],
            ),
            (dictgen::InsensitiveStr::Ascii("tunrtable"), &["turntable"]),
            (
                dictgen::InsensitiveStr::Ascii("subconcsiously"),
                &["subconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bangladesch"),
                &["bangladesh"],
            ),
            (dictgen::InsensitiveStr::Ascii("unx"), &["unix"]),
            (dictgen::InsensitiveStr::Ascii("imakes"), &["makes"]),
            (
                dictgen::InsensitiveStr::Ascii("stratchpad"),
                &["scratchpad"],
            ),
            (dictgen::InsensitiveStr::Ascii("approch"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("relfected"), &["reflected"]),
            (
                dictgen::InsensitiveStr::Ascii("confifures"),
                &["configures", "confitures"],
            ),
            (dictgen::InsensitiveStr::Ascii("acessing"), &["accessing"]),
            (
                dictgen::InsensitiveStr::Ascii("agrgressively"),
                &["aggressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predesessor"),
                &["predecessors"],
            ),
            (dictgen::InsensitiveStr::Ascii("illegalest"), &["illegals"]),
            (
                dictgen::InsensitiveStr::Ascii("placehodler"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("verious"), &["various"]),
            (dictgen::InsensitiveStr::Ascii("lisencing"), &["licencing"]),
            (
                dictgen::InsensitiveStr::Ascii("resintallations"),
                &["reinstallations"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplated"), &["duplicated"]),
            (
                dictgen::InsensitiveStr::Ascii("correpsondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerais"),
                &["considers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustarted"),
                &["illustrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("epressions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("qudratic"), &["quadratic"]),
            (
                dictgen::InsensitiveStr::Ascii("agricutlure"),
                &["agriculture"],
            ),
            (dictgen::InsensitiveStr::Ascii("reseve"), &["reserve"]),
            (dictgen::InsensitiveStr::Ascii("facebok"), &["facebook"]),
            (dictgen::InsensitiveStr::Ascii("aborbent"), &["absorbent"]),
            (
                dictgen::InsensitiveStr::Ascii("officionado"),
                &["aficionado"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stratigically"),
                &["strategically"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrance"), &["arrange"]),
            (dictgen::InsensitiveStr::Ascii("exhists"), &["exists"]),
            (dictgen::InsensitiveStr::Ascii("consitent"), &["consistent"]),
            (
                dictgen::InsensitiveStr::Ascii("missconfiguring"),
                &["misconfiguring"],
            ),
            (dictgen::InsensitiveStr::Ascii("beaslty"), &["beastly"]),
            (dictgen::InsensitiveStr::Ascii("genuinley"), &["genuinely"]),
            (dictgen::InsensitiveStr::Ascii("accecess"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("apartmens"), &["apartments"]),
            (
                dictgen::InsensitiveStr::Ascii("disguisting"),
                &["disgusting"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfifies"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("tuhmbnails"),
                &["thumbnails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insuspecting"),
                &["unsuspecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexpencive"),
                &["inexpensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggetsion"),
                &["suggestion"],
            ),
            (dictgen::InsensitiveStr::Ascii("actresess"), &["actresses"]),
            (dictgen::InsensitiveStr::Ascii("comfuse"), &["confuse"]),
            (
                dictgen::InsensitiveStr::Ascii("propoerties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porcessors"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structurels"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("gernerate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("licsenses"), &["licenses"]),
            (
                dictgen::InsensitiveStr::Ascii("consctructed"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("foorter"), &["footer"]),
            (dictgen::InsensitiveStr::Ascii("markeras"), &["markers"]),
            (
                dictgen::InsensitiveStr::Ascii("investigatiors"),
                &["investigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislpayed"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("framei"), &["frame"]),
            (
                dictgen::InsensitiveStr::Ascii("antibioticos"),
                &["antibiotics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialied"),
                &["initialized", "initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("heristic"), &["heuristic"]),
            (
                dictgen::InsensitiveStr::Ascii("staisfaction"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("govoner"), &["governor"]),
            (dictgen::InsensitiveStr::Ascii("scetching"), &["sketching"]),
            (dictgen::InsensitiveStr::Ascii("relesases"), &["releases"]),
            (
                dictgen::InsensitiveStr::Ascii("lifestiles"),
                &["lifestyles"],
            ),
            (dictgen::InsensitiveStr::Ascii("recruitcs"), &["recruits"]),
            (dictgen::InsensitiveStr::Ascii("statisic"), &["statistic"]),
            (
                dictgen::InsensitiveStr::Ascii("acquition"),
                &["acquisition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychidelic"),
                &["psychedelic"],
            ),
            (dictgen::InsensitiveStr::Ascii("valuled"), &["valued"]),
            (
                dictgen::InsensitiveStr::Ascii("diconnecting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("chemisty"), &["chemistry"]),
            (dictgen::InsensitiveStr::Ascii("relesaed"), &["released"]),
            (dictgen::InsensitiveStr::Ascii("happended"), &["happened"]),
            (dictgen::InsensitiveStr::Ascii("employeed"), &["employed"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamendalists"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("schown"), &["shown"]),
            (
                dictgen::InsensitiveStr::Ascii("intepretes"),
                &["interprets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techologies"),
                &["technologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsevers"), &["observers"]),
            (dictgen::InsensitiveStr::Ascii("statstics"), &["statistics"]),
            (dictgen::InsensitiveStr::Ascii("orede"), &["order"]),
            (dictgen::InsensitiveStr::Ascii("thubmnail"), &["thumbnail"]),
            (dictgen::InsensitiveStr::Ascii("leagcy"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("inforation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("grabing"), &["grabbing"]),
            (dictgen::InsensitiveStr::Ascii("genratet"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("ugglyness"), &["ugliness"]),
            (
                dictgen::InsensitiveStr::Ascii("accepectable"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgraddes"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("garfiled"), &["garfield"]),
            (
                dictgen::InsensitiveStr::Ascii("worshopping"),
                &["worshipping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definision"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("jewelrey"), &["jewelry"]),
            (dictgen::InsensitiveStr::Ascii("clairty"), &["clarity"]),
            (
                dictgen::InsensitiveStr::Ascii("apporpriating"),
                &["appropriating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocmpleted"),
                &["autocompleted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("haemorrage"),
                &["haemorrhage"],
            ),
            (dictgen::InsensitiveStr::Ascii("dum"), &["dumb"]),
            (
                dictgen::InsensitiveStr::Ascii("pyschologist"),
                &["psychologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("exisitence"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("alledges"), &["alleges"]),
            (dictgen::InsensitiveStr::Ascii("absoloute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("evalueations"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefices"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("moounted"), &["mounted"]),
            (dictgen::InsensitiveStr::Ascii("bettween"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("accummulate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coonstructed"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("submision"), &["submission"]),
            (
                dictgen::InsensitiveStr::Ascii("frightenend"),
                &["frightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("optaionl"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("partirion"), &["partition"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenae"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("suscript"), &["subscript"]),
            (dictgen::InsensitiveStr::Ascii("varialble"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("fobidden"), &["forbidden"]),
            (dictgen::InsensitiveStr::Ascii("stilll"), &["still"]),
            (dictgen::InsensitiveStr::Ascii("apsects"), &["aspects"]),
            (dictgen::InsensitiveStr::Ascii("exaplained"), &["explained"]),
            (
                dictgen::InsensitiveStr::Ascii("referce"),
                &["reference", "refers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutomising"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("crahs"), &["crash", "crass"]),
            (
                dictgen::InsensitiveStr::Ascii("entretained"),
                &["entertained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validtions"),
                &["validations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("constraction"),
                &["construction", "constriction", "contraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("authur"), &["author"]),
            (dictgen::InsensitiveStr::Ascii("curiostiy"), &["curiosity"]),
            (dictgen::InsensitiveStr::Ascii("portriat"), &["portrait"]),
            (dictgen::InsensitiveStr::Ascii("condigure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("custumers"), &["customers"]),
            (
                dictgen::InsensitiveStr::Ascii("millisoconds"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("sovietes"), &["soviets"]),
            (dictgen::InsensitiveStr::Ascii("fasterners"), &["fasteners"]),
            (
                dictgen::InsensitiveStr::Ascii("encounting"),
                &["encountering"],
            ),
            (dictgen::InsensitiveStr::Ascii("confrimed"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("chopy"), &["choppy", "chop"]),
            (
                dictgen::InsensitiveStr::Ascii("exchnaging"),
                &["exchanging"],
            ),
            (dictgen::InsensitiveStr::Ascii("dokcer"), &["docker"]),
            (
                dictgen::InsensitiveStr::Ascii("conescutive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimises"), &["optimises"]),
            (
                dictgen::InsensitiveStr::Ascii("degradacion"),
                &["degradation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caliofrnia"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabalized"),
                &["stabilized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruptability"),
                &["interruptibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("betwween"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("saphire"), &["sapphire"]),
            (dictgen::InsensitiveStr::Ascii("diasbles"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("issuse"),
                &["issues", "issue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coinsidence"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plian"),
                &["plain", "pliant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finilization"),
                &["finalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equailateral"),
                &["equilateral"],
            ),
            (dictgen::InsensitiveStr::Ascii("exonorate"), &["exonerate"]),
            (dictgen::InsensitiveStr::Ascii("napommes"), &["napalms"]),
            (dictgen::InsensitiveStr::Ascii("coalacing"), &["coalescing"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnarabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momementarily"),
                &["momentarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoiurced"), &["resourced"]),
            (dictgen::InsensitiveStr::Ascii("buinsess"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("thicknes"),
                &["thickness", "thickens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calucaltes"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("eanbled"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("beaviors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("exceopts"),
                &["exempts", "excepts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uestioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgiginals"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("addditionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("anedoctal"), &["anecdotal"]),
            (dictgen::InsensitiveStr::Ascii("contiued"), &["continued"]),
            (dictgen::InsensitiveStr::Ascii("quarentee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("supprots"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("panthoen"), &["pantheon"]),
            (dictgen::InsensitiveStr::Ascii("italains"), &["italians"]),
            (
                dictgen::InsensitiveStr::Ascii("medeterranean"),
                &["mediterranean"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyebros"), &["eyebrows"]),
            (dictgen::InsensitiveStr::Ascii("reniger"), &["reneger"]),
            (dictgen::InsensitiveStr::Ascii("contatined"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("dirrections"),
                &["directions"],
            ),
            (dictgen::InsensitiveStr::Ascii("agai"), &["again"]),
            (
                dictgen::InsensitiveStr::Ascii("prefences"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("janurary"), &["january"]),
            (dictgen::InsensitiveStr::Ascii("rizes"), &["rises"]),
            (dictgen::InsensitiveStr::Ascii("recomment"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("alisases"), &["aliases"]),
            (
                dictgen::InsensitiveStr::Ascii("incidentul"),
                &["incidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cigarattes"),
                &["cigarettes"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuqiring"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("statuss"), &["status"]),
            (
                dictgen::InsensitiveStr::Ascii("rearranable"),
                &["rearrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("maily"), &["mainly"]),
            (dictgen::InsensitiveStr::Ascii("pupose"), &["purpose"]),
            (dictgen::InsensitiveStr::Ascii("asess"), &["assess"]),
            (
                dictgen::InsensitiveStr::Ascii("constitutinal"),
                &["constitutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("zoombie"), &["zombie"]),
            (
                dictgen::InsensitiveStr::Ascii("authenricators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("freqently"), &["frequently"]),
            (
                dictgen::InsensitiveStr::Ascii("eligbility"),
                &["eligibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installated"),
                &["installed"],
            ),
            (dictgen::InsensitiveStr::Ascii("polyons"), &["polygons"]),
            (dictgen::InsensitiveStr::Ascii("swiftley"), &["swiftly"]),
            (dictgen::InsensitiveStr::Ascii("mor"), &["more"]),
            (
                dictgen::InsensitiveStr::Ascii("imperical"),
                &["empirical", "imperial"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporved"), &["improved"]),
            (
                dictgen::InsensitiveStr::Ascii("chipersuites"),
                &["ciphersuites"],
            ),
            (dictgen::InsensitiveStr::Ascii("strecth"), &["stretch"]),
            (
                dictgen::InsensitiveStr::Ascii("phylosophy"),
                &["philosophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constacts"),
                &["contacts", "constants"],
            ),
            (dictgen::InsensitiveStr::Ascii("pobular"), &["popular"]),
            (dictgen::InsensitiveStr::Ascii("duplicas"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("comformers"),
                &["conformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrdinated"),
                &["coordinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intruductions"),
                &["introductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("puls"), &["pulse", "plus"]),
            (
                dictgen::InsensitiveStr::Ascii("prabability"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favoutrable"),
                &["favourable"],
            ),
            (dictgen::InsensitiveStr::Ascii("metaphoras"), &["metaphors"]),
            (
                dictgen::InsensitiveStr::Ascii("asserttions"),
                &["assertions"],
            ),
            (dictgen::InsensitiveStr::Ascii("webste"), &["website"]),
            (
                dictgen::InsensitiveStr::Ascii("ejacualtion"),
                &["ejaculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("smartre"), &["smarter"]),
            (
                dictgen::InsensitiveStr::Ascii("battlefont"),
                &["battlefront"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacteurs"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuct"), &["struct"]),
            (
                dictgen::InsensitiveStr::Ascii("humanitarien"),
                &["humanitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoprorates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operaional"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongnised"),
                &["recognised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moderacion"),
                &["moderation"],
            ),
            (dictgen::InsensitiveStr::Ascii("increse"), &["increase"]),
            (
                dictgen::InsensitiveStr::Ascii("ovewritten"),
                &["overwritten"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundery"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfifier"),
                &["identifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inefficienct"),
                &["inefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compliactions"),
                &["complication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recquire"),
                &["require", "reacquire"],
            ),
            (dictgen::InsensitiveStr::Ascii("requeting"), &["requesting"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtenticate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excetutions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("disciplers"), &["disciples"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrer"),
                &["arbiter", "arbitrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperated"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("boolens"), &["booleans"]),
            (dictgen::InsensitiveStr::Ascii("syntesise"), &["synthesise"]),
            (
                dictgen::InsensitiveStr::Ascii("thumbnailers"),
                &["thumbnails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proviers"),
                &["providers", "provers"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramegter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("macarino"), &["macaroni"]),
            (
                dictgen::InsensitiveStr::Ascii("expermenetelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("labaeled"), &["labeled"]),
            (
                dictgen::InsensitiveStr::Ascii("unessecarry"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fligth"), &["flight"]),
            (dictgen::InsensitiveStr::Ascii("langauages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("contorlled"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserializaes"),
                &["deserializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbaling"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("backgrount"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychopathische"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("wehere"), &["where"]),
            (
                dictgen::InsensitiveStr::Ascii("accesssiblity"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialisationg"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("purturb"), &["perturb"]),
            (
                dictgen::InsensitiveStr::Ascii("negligeble"),
                &["negligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("intendos"), &["intends"]),
            (
                dictgen::InsensitiveStr::Ascii("decompositing"),
                &["decomposing"],
            ),
            (dictgen::InsensitiveStr::Ascii("comapator"), &["comparator"]),
            (
                dictgen::InsensitiveStr::Ascii("singoolarity"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("clossest"), &["closest"]),
            (dictgen::InsensitiveStr::Ascii("langths"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("ertoneously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoconpletes"),
                &["autocompletes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functonalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetation"),
                &["repetition", "reputation"],
            ),
            (dictgen::InsensitiveStr::Ascii("errer"), &["error"]),
            (
                dictgen::InsensitiveStr::Ascii("accumuating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliacates"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("algohms"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("needlessy"), &["needlessly"]),
            (dictgen::InsensitiveStr::Ascii("extnesion"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("elseof"), &["elseif"]),
            (
                dictgen::InsensitiveStr::Ascii("neccesarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexexpected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environemnts"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecessery"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagerizing"),
                &["plagiarizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wearned"),
                &["warned", "earned", "learned", "yearned"],
            ),
            (dictgen::InsensitiveStr::Ascii("hospots"), &["hotspots"]),
            (dictgen::InsensitiveStr::Ascii("dirver"), &["driver"]),
            (
                dictgen::InsensitiveStr::Ascii("initliasers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pidgeon"),
                &["pigeon", "pidgin"],
            ),
            (dictgen::InsensitiveStr::Ascii("shorcut"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("bbefore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("culturaly"), &["culturally"]),
            (dictgen::InsensitiveStr::Ascii("alway"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("tyhpos"), &["typos"]),
            (dictgen::InsensitiveStr::Ascii("setttings"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("claime"), &["claim"]),
            (dictgen::InsensitiveStr::Ascii("possesion"), &["possession"]),
            (dictgen::InsensitiveStr::Ascii("hoook"), &["hook"]),
            (
                dictgen::InsensitiveStr::Ascii("mouspointer"),
                &["mousepointer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproducable"),
                &["reproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("bastracts"), &["abstracts"]),
            (dictgen::InsensitiveStr::Ascii("sesions"), &["sessions"]),
            (
                dictgen::InsensitiveStr::Ascii("cinematagraphy"),
                &["cinematography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaneuos"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiniator"),
                &["combinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprieciates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("wothin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("silicoln"), &["silicon"]),
            (
                dictgen::InsensitiveStr::Ascii("temparement"),
                &["temperament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceedigly"),
                &["exceedingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authecators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orienetaion"),
                &["orientation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sovereingty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsribe"), &["subscribe"]),
            (dictgen::InsensitiveStr::Ascii("cryto"), &["crypto"]),
            (dictgen::InsensitiveStr::Ascii("imporatant"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("democraticlly"),
                &["democratically"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaods"), &["reloads"]),
            (dictgen::InsensitiveStr::Ascii("instert"), &["insert"]),
            (
                dictgen::InsensitiveStr::Ascii("dessication"),
                &["desiccation"],
            ),
            (dictgen::InsensitiveStr::Ascii("desine"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("releaves"), &["relieves"]),
            (dictgen::InsensitiveStr::Ascii("tabualtes"), &["tabulates"]),
            (dictgen::InsensitiveStr::Ascii("finanaced"), &["financed"]),
            (dictgen::InsensitiveStr::Ascii("soembody"), &["somebody"]),
            (
                dictgen::InsensitiveStr::Ascii("iteractive"),
                &["interactive", "iterative"],
            ),
            (dictgen::InsensitiveStr::Ascii("jamsine"), &["jasmine"]),
            (
                dictgen::InsensitiveStr::Ascii("andled"),
                &["handled", "dandled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thankfullly"),
                &["thankfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illigetimate"),
                &["illegitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("instincs"), &["instincts"]),
            (
                dictgen::InsensitiveStr::Ascii("conventinal"),
                &["conventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumptuious"),
                &["presumptuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cofigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agriculteurs"),
                &["agriculture"],
            ),
            (dictgen::InsensitiveStr::Ascii("scahr"), &["schar"]),
            (
                dictgen::InsensitiveStr::Ascii("troublshooting"),
                &["troubleshooting"],
            ),
            (dictgen::InsensitiveStr::Ascii("parocheal"), &["parochial"]),
            (dictgen::InsensitiveStr::Ascii("anaylzing"), &["analyzing"]),
            (
                dictgen::InsensitiveStr::Ascii("chrashed"),
                &["crashed", "thrashed", "trashed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significan"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("missourri"), &["missouri"]),
            (dictgen::InsensitiveStr::Ascii("segmetn"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("accesssor"), &["accessor"]),
            (dictgen::InsensitiveStr::Ascii("histroian"), &["historian"]),
            (
                dictgen::InsensitiveStr::Ascii("acknownledgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthousiastic"),
                &["enthusiastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lcoalization"),
                &["localization"],
            ),
            (dictgen::InsensitiveStr::Ascii("advaned"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("clifthanger"),
                &["cliffhanger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esponses"),
                &["responses", "espouses"],
            ),
            (dictgen::InsensitiveStr::Ascii("destroied"), &["destroyed"]),
            (dictgen::InsensitiveStr::Ascii("headshoots"), &["headshot"]),
            (
                dictgen::InsensitiveStr::Ascii("wholeheartely"),
                &["wholeheartedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("recongnise"), &["recognise"]),
            (dictgen::InsensitiveStr::Ascii("displayd"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("activsted"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("adventuruous"),
                &["adventurous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componennts"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("inveral"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("perheaps"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("satisfiies"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("contenental"),
                &["continental"],
            ),
            (dictgen::InsensitiveStr::Ascii("desingning"), &["designing"]),
            (
                dictgen::InsensitiveStr::Ascii("metophorically"),
                &["metaphorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwhemed"),
                &["overwhelmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orcestrate"),
                &["orchestrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplictions"),
                &["duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("aroudn"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("prodction"), &["production"]),
            (
                dictgen::InsensitiveStr::Ascii("typographyc"),
                &["typographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fascization"),
                &["fascination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coefficeint"),
                &["coefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("championsips"),
                &["championships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("knarly"), &["gnarly"]),
            (
                dictgen::InsensitiveStr::Ascii("spreadshet"),
                &["spreadsheet"],
            ),
            (dictgen::InsensitiveStr::Ascii("prommpts"), &["prompts"]),
            (
                dictgen::InsensitiveStr::Ascii("pojrection"),
                &["projection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsiduary"),
                &["subsidiary"],
            ),
            (dictgen::InsensitiveStr::Ascii("lagacy"), &["legacy"]),
            (dictgen::InsensitiveStr::Ascii("channging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("oppurtunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("siffixed"), &["suffixed"]),
            (
                dictgen::InsensitiveStr::Ascii("concequence"),
                &["consequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("priot"), &["prior"]),
            (dictgen::InsensitiveStr::Ascii("inclusice"), &["inclusive"]),
            (dictgen::InsensitiveStr::Ascii("solutide"), &["solitude"]),
            (
                dictgen::InsensitiveStr::Ascii("ambassedor"),
                &["ambassador"],
            ),
            (dictgen::InsensitiveStr::Ascii("pancakers"), &["pancakes"]),
            (
                dictgen::InsensitiveStr::Ascii("atteributes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("chatch"), &["catch"]),
            (
                dictgen::InsensitiveStr::Ascii("attribiuted"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("varifiers"), &["verifiers"]),
            (dictgen::InsensitiveStr::Ascii("falbacks"), &["fallbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("dulication"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrepancy"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("raspberrry"), &["raspberry"]),
            (dictgen::InsensitiveStr::Ascii("proivded"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("incorpration"),
                &["incorporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inmediatly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("penalizin"), &["penalizing"]),
            (dictgen::InsensitiveStr::Ascii("abl"), &["able"]),
            (
                dictgen::InsensitiveStr::Ascii("philippenis"),
                &["philippines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructiob"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperince"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("interesat"), &["interest"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumenal"),
                &["instrumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("squirl"), &["squirrel"]),
            (dictgen::InsensitiveStr::Ascii("commadn"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaatlly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("tdoo"), &["todo"]),
            (dictgen::InsensitiveStr::Ascii("paradiggum"), &["paradigm"]),
            (dictgen::InsensitiveStr::Ascii("repitles"), &["reptiles"]),
            (
                dictgen::InsensitiveStr::Ascii("symmmetrical"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afterhtoughts"),
                &["afterthoughts"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrater"), &["abstracter"]),
            (dictgen::InsensitiveStr::Ascii("genenerate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("kiyacking"), &["kayaking"]),
            (dictgen::InsensitiveStr::Ascii("funciton"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("profied"), &["profiled"]),
            (
                dictgen::InsensitiveStr::Ascii("unconnectes"),
                &["unconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costruction"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("afternooon"), &["afternoon"]),
            (dictgen::InsensitiveStr::Ascii("hypvisor"), &["hypervisor"]),
            (dictgen::InsensitiveStr::Ascii("permade"), &["premade"]),
            (
                dictgen::InsensitiveStr::Ascii("aletrnatives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabed"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("consideret"),
                &["considerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cordinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contectual"),
                &["contextual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifaces"),
                &["sacrifices"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheltes"), &["athletes"]),
            (dictgen::InsensitiveStr::Ascii("confgs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("derferencing"),
                &["dereferencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforers"), &["enforcers"]),
            (
                dictgen::InsensitiveStr::Ascii("subtiles"),
                &["subfiles", "subtitles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegociate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaptibele"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critizizing"),
                &["criticizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("catptures"), &["captures"]),
            (
                dictgen::InsensitiveStr::Ascii("informatations"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("entierly"), &["entirely"]),
            (
                dictgen::InsensitiveStr::Ascii("straightowards"),
                &["straightwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surbert"),
                &["sherbert", "sherbet"],
            ),
            (dictgen::InsensitiveStr::Ascii("importas"), &["imports"]),
            (dictgen::InsensitiveStr::Ascii("delaring"), &["declaring"]),
            (
                dictgen::InsensitiveStr::Ascii("itegrations"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("foucsed"), &["focused"]),
            (dictgen::InsensitiveStr::Ascii("naferious"), &["nefarious"]),
            (
                dictgen::InsensitiveStr::Ascii("implememtation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradional"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("betwwen"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("certifacte"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilibrim"),
                &["equilibrium"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprarent"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("wokrer"), &["worker"]),
            (dictgen::InsensitiveStr::Ascii("emulaion"), &["emulation"]),
            (dictgen::InsensitiveStr::Ascii("inorrect"), &["incorrect"]),
            (dictgen::InsensitiveStr::Ascii("qeustions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("electricman"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cathegorizing"),
                &["categorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conclued"), &["conclude"]),
            (dictgen::InsensitiveStr::Ascii("empolyees"), &["employees"]),
            (dictgen::InsensitiveStr::Ascii("eurpean"), &["european"]),
            (
                dictgen::InsensitiveStr::Ascii("predetirmined"),
                &["predetermined"],
            ),
            (dictgen::InsensitiveStr::Ascii("defetead"), &["defeated"]),
            (
                dictgen::InsensitiveStr::Ascii("sungalsses"),
                &["sunglasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertarianists"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpiration"),
                &["inspiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aqcuainted"),
                &["acquainted"],
            ),
            (dictgen::InsensitiveStr::Ascii("convery"), &["convert"]),
            (
                dictgen::InsensitiveStr::Ascii("supplemets"),
                &["supplements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compenidum"),
                &["compendium"],
            ),
            (dictgen::InsensitiveStr::Ascii("countig"), &["counting"]),
            (
                dictgen::InsensitiveStr::Ascii("functtioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("activtion"), &["activation"]),
            (dictgen::InsensitiveStr::Ascii("unknon"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("tramsmissions"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptioms"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("regsitre"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("geogria"), &["georgia"]),
            (
                dictgen::InsensitiveStr::Ascii("convnetions"),
                &["conventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepeter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("antiquited"),
                &["antiquated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consisntency"),
                &["consistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncicalizing"),
                &["canonicalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("carptener"), &["carpenter"]),
            (
                dictgen::InsensitiveStr::Ascii("unxepectedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distibutor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maybilline"),
                &["maybelline"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirived"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("considier"), &["consider"]),
            (
                dictgen::InsensitiveStr::Ascii("interseption"),
                &["interception"],
            ),
            (dictgen::InsensitiveStr::Ascii("excedeed"), &["exceeded"]),
            (
                dictgen::InsensitiveStr::Ascii("paraemeters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemementations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checlboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preprocesing"),
                &["preprocessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("pesticids"), &["pesticides"]),
            (dictgen::InsensitiveStr::Ascii("vindows"), &["windows"]),
            (dictgen::InsensitiveStr::Ascii("disnabling"), &["disabling"]),
            (
                dictgen::InsensitiveStr::Ascii("combinaters"),
                &["combinators", "combiners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provisoined"),
                &["provisioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("batchleurs"), &["bachelors"]),
            (dictgen::InsensitiveStr::Ascii("elemants"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("versfion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("privider"), &["provider"]),
            (dictgen::InsensitiveStr::Ascii("cacluated"), &["calculated"]),
            (dictgen::InsensitiveStr::Ascii("accapted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilitiesy"),
                &["vulnerability", "vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrypte"),
                &["encrypted", "encrypt"],
            ),
            (dictgen::InsensitiveStr::Ascii("naximum"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("functrion"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("assimilerat"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("belivebly"), &["believably"]),
            (dictgen::InsensitiveStr::Ascii("malasyian"), &["malaysian"]),
            (
                dictgen::InsensitiveStr::Ascii("depenendecies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("ueser"), &["user"]),
            (
                dictgen::InsensitiveStr::Ascii("alteranted"),
                &["alternated"],
            ),
            (dictgen::InsensitiveStr::Ascii("gosepls"), &["gospels"]),
            (dictgen::InsensitiveStr::Ascii("depreacte"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("privilegien"),
                &["privilege"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommneded"),
                &["recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeerable"),
                &["unbearable"],
            ),
            (dictgen::InsensitiveStr::Ascii("minisclue"), &["miniscule"]),
            (
                dictgen::InsensitiveStr::Ascii("annhiliation"),
                &["annihilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("artmeis"), &["artemis"]),
            (dictgen::InsensitiveStr::Ascii("journyed"), &["journeyed"]),
            (dictgen::InsensitiveStr::Ascii("lightrooom"), &["lightroom"]),
            (dictgen::InsensitiveStr::Ascii("seege"), &["siege"]),
            (dictgen::InsensitiveStr::Ascii("possobily"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("aggreates"), &["aggregates"]),
            (dictgen::InsensitiveStr::Ascii("speperate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("comminity"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("invesrse"), &["inverse"]),
            (dictgen::InsensitiveStr::Ascii("intesnely"), &["intensely"]),
            (
                dictgen::InsensitiveStr::Ascii("possiblilties"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("befroe"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("drivr"), &["driver"]),
            (dictgen::InsensitiveStr::Ascii("sincereley"), &["sincerely"]),
            (dictgen::InsensitiveStr::Ascii("rangeo"), &["range"]),
            (dictgen::InsensitiveStr::Ascii("wokrspace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("moderaters"), &["moderates"]),
            (dictgen::InsensitiveStr::Ascii("coem"), &["come"]),
            (dictgen::InsensitiveStr::Ascii("otherwyse"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("filesstems"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermidiery"),
                &["intermediary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predecesor"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimzied"), &["optimized"]),
            (dictgen::InsensitiveStr::Ascii("feonsay"), &["fiancée"]),
            (
                dictgen::InsensitiveStr::Ascii("peripheals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disingenuious"),
                &["disingenuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignory"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("perferrmances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("srarted"), &["started"]),
            (dictgen::InsensitiveStr::Ascii("astronuat"), &["astronaut"]),
            (
                dictgen::InsensitiveStr::Ascii("subresoure"),
                &["subresource"],
            ),
            (dictgen::InsensitiveStr::Ascii("intsalls"), &["installs"]),
            (
                dictgen::InsensitiveStr::Ascii("consective"),
                &["consecutive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exlusively"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("markters"), &["markers"]),
            (
                dictgen::InsensitiveStr::Ascii("camparator"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreated"), &["aggregated"]),
            (
                dictgen::InsensitiveStr::Ascii("speres"),
                &["spheres", "spares", "spires", "spores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventagous"),
                &["advantageous"],
            ),
            (dictgen::InsensitiveStr::Ascii("soldgers"), &["soldiers"]),
            (dictgen::InsensitiveStr::Ascii("namepace"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("univesity"), &["university"]),
            (
                dictgen::InsensitiveStr::Ascii("encouragment"),
                &["encouragement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstenance"),
                &["abstinence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understandablely"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("regisrty"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("geniune"), &["genuine"]),
            (dictgen::InsensitiveStr::Ascii("retuerned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("instanes"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("waitung"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("struggeles"), &["struggles"]),
            (
                dictgen::InsensitiveStr::Ascii("naybourhood"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("atack"), &["attack"]),
            (dictgen::InsensitiveStr::Ascii("decosings"), &["decodings"]),
            (dictgen::InsensitiveStr::Ascii("zionsits"), &["zionists"]),
            (dictgen::InsensitiveStr::Ascii("kookoo"), &["cuckoo"]),
            (
                dictgen::InsensitiveStr::Ascii("deveolpment"),
                &["developments", "development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilezations"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seprations"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("deafualts"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("stadius"), &["stadiums"]),
            (
                dictgen::InsensitiveStr::Ascii("dependdence"),
                &["dependence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synctactically"),
                &["syntactically"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinally"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("staminia"), &["stamina"]),
            (dictgen::InsensitiveStr::Ascii("amuch"), &["much"]),
            (dictgen::InsensitiveStr::Ascii("aledge"), &["allege"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreegating"),
                &["aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("garuantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("correcpondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixels"), &["pixels"]),
            (dictgen::InsensitiveStr::Ascii("retirns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("linarly"), &["linearly"]),
            (dictgen::InsensitiveStr::Ascii("nirvanna"), &["nirvana"]),
            (
                dictgen::InsensitiveStr::Ascii("checckboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customzied"),
                &["customized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempreture"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("recompuute"), &["recompute"]),
            (
                dictgen::InsensitiveStr::Ascii("addiditionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("atempted"), &["attempted"]),
            (dictgen::InsensitiveStr::Ascii("poiunter"), &["pointer"]),
            (dictgen::InsensitiveStr::Ascii("miniums"), &["minimums"]),
            (
                dictgen::InsensitiveStr::Ascii("banchmarks"),
                &["benchmarks"],
            ),
            (dictgen::InsensitiveStr::Ascii("steeles"), &["steelers"]),
            (dictgen::InsensitiveStr::Ascii("elememtal"), &["elemental"]),
            (dictgen::InsensitiveStr::Ascii("limtiing"), &["limiting"]),
            (
                dictgen::InsensitiveStr::Ascii("attachtments"),
                &["attachments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successfuily"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrilation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shamelessy"),
                &["shamelessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combainations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertainity"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("sability"), &["stability"]),
            (
                dictgen::InsensitiveStr::Ascii("descryptions"),
                &["descriptions", "decryptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("halfs"), &["halves"]),
            (
                dictgen::InsensitiveStr::Ascii("unforms"),
                &["uniforms", "informs"],
            ),
            (dictgen::InsensitiveStr::Ascii("projctor"), &["projector"]),
            (
                dictgen::InsensitiveStr::Ascii("initializaion"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovrides"), &["overrides"]),
            (dictgen::InsensitiveStr::Ascii("frustm"), &["frustum"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitevely"),
                &["competitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("thikns"), &["thinks"]),
            (dictgen::InsensitiveStr::Ascii("behaivour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("triffered"), &["triggered"]),
            (dictgen::InsensitiveStr::Ascii("uplaodaded"), &["uploaded"]),
            (dictgen::InsensitiveStr::Ascii("somestime"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("torndao"), &["tornado"]),
            (
                dictgen::InsensitiveStr::Ascii("multidimentionnal"),
                &["multidimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("craetors"),
                &["craters", "creators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bookkeeiping"),
                &["bookkeeping"],
            ),
            (dictgen::InsensitiveStr::Ascii("unreadble"), &["unreadable"]),
            (dictgen::InsensitiveStr::Ascii("witthout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("subfroms"), &["subforms"]),
            (dictgen::InsensitiveStr::Ascii("deltes"), &["deletes"]),
            (dictgen::InsensitiveStr::Ascii("proceesor"), &["processor"]),
            (
                dictgen::InsensitiveStr::Ascii("chloesterol"),
                &["cholesterol"],
            ),
            (dictgen::InsensitiveStr::Ascii("beflow"), &["below"]),
            (dictgen::InsensitiveStr::Ascii("considder"), &["consider"]),
            (dictgen::InsensitiveStr::Ascii("magensium"), &["magnesium"]),
            (dictgen::InsensitiveStr::Ascii("secifier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("internationnalists"),
                &["internationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifieing"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("costumizing"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fiel"),
                &["feel", "field", "file", "phial"],
            ),
            (dictgen::InsensitiveStr::Ascii("lepard"), &["leopard"]),
            (
                dictgen::InsensitiveStr::Ascii("outsoursed"),
                &["outsourced"],
            ),
            (dictgen::InsensitiveStr::Ascii("mecahinsm"), &["mechanism"]),
            (dictgen::InsensitiveStr::Ascii("analiser"), &["analyser"]),
            (
                dictgen::InsensitiveStr::Ascii("widhtpoint"),
                &["widthpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("settinga"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("hijkacking"), &["hijacking"]),
            (dictgen::InsensitiveStr::Ascii("banruptcy"), &["bankruptcy"]),
            (
                dictgen::InsensitiveStr::Ascii("lighthosue"),
                &["lighthouse"],
            ),
            (dictgen::InsensitiveStr::Ascii("shephed"), &["shepherd"]),
            (
                dictgen::InsensitiveStr::Ascii("introductie"),
                &["introduce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projectils"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hadshakes"),
                &["handshakes", "headshakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virualised"),
                &["virtualised", "visualised"],
            ),
            (dictgen::InsensitiveStr::Ascii("wehnever"), &["whenever"]),
            (dictgen::InsensitiveStr::Ascii("dedault"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("recendes"), &["rescinds"]),
            (dictgen::InsensitiveStr::Ascii("montlhy"), &["monthly"]),
            (dictgen::InsensitiveStr::Ascii("seacrchers"), &["searchers"]),
            (
                dictgen::InsensitiveStr::Ascii("administraton"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("warror"), &["warrior"]),
            (
                dictgen::InsensitiveStr::Ascii("intermediat"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("corelated"), &["correlated"]),
            (
                dictgen::InsensitiveStr::Ascii("asociations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sutffed"), &["stuffed"]),
            (dictgen::InsensitiveStr::Ascii("modfided"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("denominacion"),
                &["denomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("bearword"), &["bareword"]),
            (
                dictgen::InsensitiveStr::Ascii("immediatley"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("obcject"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("aninated"), &["animated"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoemamces"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsistainable"),
                &["unsustainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("balony"),
                &["baloney", "bologna"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfacted"), &["perfected"]),
            (dictgen::InsensitiveStr::Ascii("ligths"), &["lights"]),
            (dictgen::InsensitiveStr::Ascii("hyrdaulic"), &["hydraulic"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrotion"),
                &["distortion"],
            ),
            (dictgen::InsensitiveStr::Ascii("historgram"), &["histogram"]),
            (dictgen::InsensitiveStr::Ascii("cambride"), &["cambridge"]),
            (
                dictgen::InsensitiveStr::Ascii("incorparated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepeted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("annotaitons"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("happyness"), &["happiness"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("enumrates"), &["enumerates"]),
            (
                dictgen::InsensitiveStr::Ascii("algorgithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatior"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("notorios"), &["notorious"]),
            (dictgen::InsensitiveStr::Ascii("paralisys"), &["paralysis"]),
            (dictgen::InsensitiveStr::Ascii("sesceded"), &["seceded"]),
            (
                dictgen::InsensitiveStr::Ascii("inteact"),
                &["interact", "intact"],
            ),
            (dictgen::InsensitiveStr::Ascii("analised"), &["analysed"]),
            (dictgen::InsensitiveStr::Ascii("fatures"), &["features"]),
            (dictgen::InsensitiveStr::Ascii("sudu"), &["sudo"]),
            (
                dictgen::InsensitiveStr::Ascii("mircotransaction"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("napoelon"), &["napoleon"]),
            (dictgen::InsensitiveStr::Ascii("excedding"), &["exceeding"]),
            (
                dictgen::InsensitiveStr::Ascii("cleaing"),
                &["cleaning", "clearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcription"),
                &["subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterizing"),
                &["parameterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplpying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("playerss"), &["players"]),
            (
                dictgen::InsensitiveStr::Ascii("modififying"),
                &["modifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntatic"), &["syntactic"]),
            (dictgen::InsensitiveStr::Ascii("exces"), &["excess"]),
            (
                dictgen::InsensitiveStr::Ascii("annnotators"),
                &["annotators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertius"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("subree"), &["subtree"]),
            (
                dictgen::InsensitiveStr::Ascii("assignements"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inddividually"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wapped"),
                &["wrapped", "swapped", "warped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unseted"),
                &["unset", "unseated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocilator"), &["oscillator"]),
            (
                dictgen::InsensitiveStr::Ascii("highlited"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("inludung"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("tihs"), &["this"]),
            (
                dictgen::InsensitiveStr::Ascii("unavilability"),
                &["unavailability"],
            ),
            (dictgen::InsensitiveStr::Ascii("konws"), &["knows"]),
            (dictgen::InsensitiveStr::Ascii("vieports"), &["viewports"]),
            (dictgen::InsensitiveStr::Ascii("coommit"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("ediges"), &["edges"]),
            (
                dictgen::InsensitiveStr::Ascii("sepeartors"),
                &["separators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undermineras"),
                &["undermines"],
            ),
            (dictgen::InsensitiveStr::Ascii("miticated"), &["mitigated"]),
            (
                dictgen::InsensitiveStr::Ascii("horzontally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regresstion"),
                &["regression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliberatey"),
                &["deliberately"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolut"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("cirrently"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("wissle"), &["whistle"]),
            (dictgen::InsensitiveStr::Ascii("cahrge"), &["charge"]),
            (
                dictgen::InsensitiveStr::Ascii("desintegrated"),
                &["disintegrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("errornoeusly"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphatizing"),
                &["emphasizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("memonic"), &["mnemonic"]),
            (
                dictgen::InsensitiveStr::Ascii("inititaive"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchromous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("westernerns"),
                &["westerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriipted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("comaprators"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objetively"),
                &["objectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("dleayed"), &["delayed"]),
            (dictgen::InsensitiveStr::Ascii("sepcified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("restransmit"),
                &["retransmit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atferwards"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competetively"),
                &["competitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificastion"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracterize"),
                &["characterize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhiliations"),
                &["annihilations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pavillions"), &["pavilions"]),
            (dictgen::InsensitiveStr::Ascii("blonged"), &["belonged"]),
            (dictgen::InsensitiveStr::Ascii("biinary"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("deprovation"),
                &["deprivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("realiable"), &["reliable"]),
            (dictgen::InsensitiveStr::Ascii("dieabled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("aceptor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("connonical"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("nestwork"), &["network"]),
            (dictgen::InsensitiveStr::Ascii("deatch"), &["detach"]),
            (dictgen::InsensitiveStr::Ascii("jenkkins"), &["jenkins"]),
            (dictgen::InsensitiveStr::Ascii("suppopose"), &["suppose"]),
            (dictgen::InsensitiveStr::Ascii("managment"), &["management"]),
            (dictgen::InsensitiveStr::Ascii("runting"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("corresspondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("resulved"), &["resolved"]),
            (
                dictgen::InsensitiveStr::Ascii("indenation"),
                &["indentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronting"),
                &["printing", "pronating", "proning"],
            ),
            (dictgen::InsensitiveStr::Ascii("protosed"), &["proposed"]),
            (dictgen::InsensitiveStr::Ascii("manifiesto"), &["manifesto"]),
            (dictgen::InsensitiveStr::Ascii("ehough"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("progamable"),
                &["programmable"],
            ),
            (dictgen::InsensitiveStr::Ascii("develoers"), &["developers"]),
            (dictgen::InsensitiveStr::Ascii("fle"), &["file"]),
            (
                dictgen::InsensitiveStr::Ascii("corinthianos"),
                &["corinthians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimitating"),
                &["intimidating"],
            ),
            (dictgen::InsensitiveStr::Ascii("withough"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("comform"),
                &["conform", "comfort"],
            ),
            (dictgen::InsensitiveStr::Ascii("genrics"), &["generics"]),
            (dictgen::InsensitiveStr::Ascii("valdator"), &["validator"]),
            (dictgen::InsensitiveStr::Ascii("tyepof"), &["typeof"]),
            (dictgen::InsensitiveStr::Ascii("somwho"), &["somehow"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainting"),
                &["maintaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porbabilities"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancelability"),
                &["cancellability"],
            ),
            (dictgen::InsensitiveStr::Ascii("vriginity"), &["virginity"]),
            (dictgen::InsensitiveStr::Ascii("sketchs"), &["sketches"]),
            (
                dictgen::InsensitiveStr::Ascii("resposibilities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installatation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("variabl"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("messasge"),
                &["message", "messages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initaitive"),
                &["initiative"],
            ),
            (dictgen::InsensitiveStr::Ascii("oepning"), &["opening"]),
            (dictgen::InsensitiveStr::Ascii("folllowers"), &["followers"]),
            (
                dictgen::InsensitiveStr::Ascii("transkription"),
                &["transcription"],
            ),
            (dictgen::InsensitiveStr::Ascii("ridiculose"), &["ridicule"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactuerd"),
                &["manufactured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communistas"),
                &["communists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturabte"),
                &["masturbate"],
            ),
            (dictgen::InsensitiveStr::Ascii("zookeinee"), &["zucchini"]),
            (dictgen::InsensitiveStr::Ascii("ofr"), &["for"]),
            (
                dictgen::InsensitiveStr::Ascii("facinating"),
                &["fascinating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intereseting"),
                &["interesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("quitt"), &["quit"]),
            (
                dictgen::InsensitiveStr::Ascii("responsiblity"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simpplifications"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("srtaddle"), &["straddle"]),
            (
                dictgen::InsensitiveStr::Ascii("agknowledge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vehemantly"),
                &["vehemently"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallating"), &["collating"]),
            (dictgen::InsensitiveStr::Ascii("mangetic"), &["magnetic"]),
            (
                dictgen::InsensitiveStr::Ascii("cylindically"),
                &["cylindrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("waranties"), &["warranties"]),
            (
                dictgen::InsensitiveStr::Ascii("exersising"),
                &["exercising"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefers"), &["prefers"]),
            (dictgen::InsensitiveStr::Ascii("hoppefully"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("ocraina"), &["ocarina"]),
            (
                dictgen::InsensitiveStr::Ascii("unanwsered"),
                &["unanswered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fanslaughter"),
                &["manslaughter"],
            ),
            (dictgen::InsensitiveStr::Ascii("privilges"), &["privileges"]),
            (
                dictgen::InsensitiveStr::Ascii("cosumes"),
                &["consumes", "costumes"],
            ),
            (dictgen::InsensitiveStr::Ascii("isssue"), &["issue"]),
            (
                dictgen::InsensitiveStr::Ascii("resurecting"),
                &["resurrecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("propaged"), &["propagated"]),
            (
                dictgen::InsensitiveStr::Ascii("greysacles"),
                &["greyscales"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionalitis"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("addiotion"), &["addition"]),
            (dictgen::InsensitiveStr::Ascii("folllowind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("involuntarity"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccuracy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("aotomate"), &["automate"]),
            (dictgen::InsensitiveStr::Ascii("operandes"), &["operands"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferncer"),
                &["dereferencer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extendsions"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opthamologist"),
                &["ophthalmologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("clagary"), &["calgary"]),
            (dictgen::InsensitiveStr::Ascii("jupyther"), &["jupyter"]),
            (dictgen::InsensitiveStr::Ascii("linueps"), &["lineups"]),
            (
                dictgen::InsensitiveStr::Ascii("potencially"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("telementry"), &["telemetry"]),
            (
                dictgen::InsensitiveStr::Ascii("interseting"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mircocenter"),
                &["microcenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("abscence"), &["absence"]),
            (dictgen::InsensitiveStr::Ascii("apppended"), &["appended"]),
            (dictgen::InsensitiveStr::Ascii("abck"), &["back", "aback"]),
            (dictgen::InsensitiveStr::Ascii("diggests"), &["digests"]),
            (dictgen::InsensitiveStr::Ascii("follwoeind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("disaspointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionar"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decruption"),
                &["decryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cocurrency"),
                &["concurrency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commonwelath"),
                &["commonwealth"],
            ),
            (dictgen::InsensitiveStr::Ascii("invovles"), &["involves"]),
            (
                dictgen::InsensitiveStr::Ascii("annecessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("geoemtry"), &["geometry"]),
            (
                dictgen::InsensitiveStr::Ascii("capitlizes"),
                &["capitalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("warinigs"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("overaall"), &["overall"]),
            (dictgen::InsensitiveStr::Ascii("diarreah"), &["diarrhea"]),
            (
                dictgen::InsensitiveStr::Ascii("mantainable"),
                &["maintainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strigifying"),
                &["stringifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("arthorsis"), &["arthrosis"]),
            (dictgen::InsensitiveStr::Ascii("redict"), &["redirect"]),
            (
                dictgen::InsensitiveStr::Ascii("reurned"),
                &["returned", "reran"],
            ),
            (dictgen::InsensitiveStr::Ascii("critereon"), &["criterion"]),
            (
                dictgen::InsensitiveStr::Ascii("surrended"),
                &["surrounded", "surrendered"],
            ),
            (dictgen::InsensitiveStr::Ascii("decribe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("venemous"), &["venomous"]),
            (
                dictgen::InsensitiveStr::Ascii("seperating"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("musils"),
                &["muscles", "mussels"],
            ),
            (dictgen::InsensitiveStr::Ascii("delemiter"), &["delimiter"]),
            (
                dictgen::InsensitiveStr::Ascii("appriximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhaustin"), &["exhaustion"]),
            (dictgen::InsensitiveStr::Ascii("steadliy"), &["steadily"]),
            (dictgen::InsensitiveStr::Ascii("affiars"), &["affairs"]),
            (
                dictgen::InsensitiveStr::Ascii("simultaneoulsy"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("clinicaly"), &["clinically"]),
            (
                dictgen::InsensitiveStr::Ascii("prepartion"),
                &["preparation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bangledesh"),
                &["bangladesh"],
            ),
            (dictgen::InsensitiveStr::Ascii("dermine"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("elgible"), &["eligible"]),
            (dictgen::InsensitiveStr::Ascii("sucessufl"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("reverese"), &["reverse"]),
            (
                dictgen::InsensitiveStr::Ascii("eldistribution"),
                &["redistribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independed"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combindator"),
                &["combinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcedure"), &["procedure"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomdingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("duting"), &["during"]),
            (
                dictgen::InsensitiveStr::Ascii("rollertoaster"),
                &["rollercoaster"],
            ),
            (dictgen::InsensitiveStr::Ascii("collone"), &["cologne"]),
            (dictgen::InsensitiveStr::Ascii("filsystem"), &["filesystem"]),
            (dictgen::InsensitiveStr::Ascii("cotaniner"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalistisen"),
                &["nationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("permitts"), &["permits"]),
            (dictgen::InsensitiveStr::Ascii("asumes"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("conecntrates"),
                &["concentrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("broblems"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("eneabled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("datin"),
                &["dating", "satin", "latin"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepetate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("slippey"), &["slippery"]),
            (dictgen::InsensitiveStr::Ascii("liekwise"), &["likewise"]),
            (dictgen::InsensitiveStr::Ascii("invocie"), &["invoice"]),
            (dictgen::InsensitiveStr::Ascii("heros"), &["heroes"]),
            (
                dictgen::InsensitiveStr::Ascii("interruptable"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centrigrade"),
                &["centigrade"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermimentelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleanding"),
                &["cleaning", "cleansing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("referenc"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("annonomusly"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistend"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underminining"),
                &["undermining"],
            ),
            (dictgen::InsensitiveStr::Ascii("resolvin"), &["resolving"]),
            (dictgen::InsensitiveStr::Ascii("cutomise"), &["customise"]),
            (dictgen::InsensitiveStr::Ascii("blbos"), &["blobs"]),
            (
                dictgen::InsensitiveStr::Ascii("disriminator"),
                &["discriminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("becaused"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("granilarity"),
                &["granularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("epxand"), &["expand"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("attatching"), &["attaching"]),
            (
                dictgen::InsensitiveStr::Ascii("compitability"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irreplacalbe"),
                &["irreplaceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sorrry"), &["sorry"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferance"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("gererally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("preccede"), &["precede"]),
            (dictgen::InsensitiveStr::Ascii("silencin"), &["silencing"]),
            (dictgen::InsensitiveStr::Ascii("perfurd"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("religiones"), &["religions"]),
            (
                dictgen::InsensitiveStr::Ascii("tranistional"),
                &["transitional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assymptotes"),
                &["asymptotes"],
            ),
            (dictgen::InsensitiveStr::Ascii("highligts"), &["highlights"]),
            (
                dictgen::InsensitiveStr::Ascii("abandining"),
                &["abandoning"],
            ),
            (dictgen::InsensitiveStr::Ascii("highjacked"), &["hijacked"]),
            (dictgen::InsensitiveStr::Ascii("frrom"), &["from"]),
            (dictgen::InsensitiveStr::Ascii("escpaing"), &["escaping"]),
            (
                dictgen::InsensitiveStr::Ascii("extimators"),
                &["estimators"],
            ),
            (dictgen::InsensitiveStr::Ascii("intials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("liberatrians"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compulsing"),
                &["compulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("priveledge"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("forsane"), &["forsaken"]),
            (
                dictgen::InsensitiveStr::Ascii("laboritory"),
                &["laboratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("sillabus"), &["syllabus"]),
            (dictgen::InsensitiveStr::Ascii("priestood"), &["priesthood"]),
            (
                dictgen::InsensitiveStr::Ascii("hilighter"),
                &["highlighter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misrepresentatie"),
                &["misrepresentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("lauch"), &["launch"]),
            (
                dictgen::InsensitiveStr::Ascii("compatility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troughout"),
                &["throughout", "throughput"],
            ),
            (dictgen::InsensitiveStr::Ascii("lso"), &["also"]),
            (
                dictgen::InsensitiveStr::Ascii("hestiation"),
                &["hesitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("accodions"), &["accordions"]),
            (
                dictgen::InsensitiveStr::Ascii("highloader"),
                &["highlander"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paychologist"),
                &["psychologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quating"),
                &["quoting", "squatting", "equating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimentery"),
                &["complimentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("comsetics"), &["cosmetics"]),
            (
                dictgen::InsensitiveStr::Ascii("spefixications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("messsengers"),
                &["messengers"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganeral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("embbeders"), &["embedders"]),
            (
                dictgen::InsensitiveStr::Ascii("doubletquote"),
                &["doublequote"],
            ),
            (dictgen::InsensitiveStr::Ascii("betweem"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("lincenced"), &["licenced"]),
            (
                dictgen::InsensitiveStr::Ascii("preseverence"),
                &["perseverance"],
            ),
            (dictgen::InsensitiveStr::Ascii("atmospere"), &["atmosphere"]),
            (
                dictgen::InsensitiveStr::Ascii("fulfullment"),
                &["fulfillment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parliamant"),
                &["parliament"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceuting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("rebllions"), &["rebellions"]),
            (dictgen::InsensitiveStr::Ascii("aninates"), &["animates"]),
            (dictgen::InsensitiveStr::Ascii("chochka"), &["tchotchke"]),
            (dictgen::InsensitiveStr::Ascii("waypoing"), &["waypoint"]),
            (dictgen::InsensitiveStr::Ascii("sexualizd"), &["sexualized"]),
            (
                dictgen::InsensitiveStr::Ascii("identifierad"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("localed"), &["located"]),
            (dictgen::InsensitiveStr::Ascii("inclded"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("rountriped"),
                &["roundtripped"],
            ),
            (dictgen::InsensitiveStr::Ascii("replcaced"), &["replaced"]),
            (
                dictgen::InsensitiveStr::Ascii("trascational"),
                &["transactional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomizes"),
                &["randomizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kaleidoscopal"),
                &["kaleidoscopic"],
            ),
            (dictgen::InsensitiveStr::Ascii("actvate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("silouweted"),
                &["silhouetted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plaigarism"),
                &["plagiarism"],
            ),
            (dictgen::InsensitiveStr::Ascii("terorrist"), &["terrorist"]),
            (dictgen::InsensitiveStr::Ascii("sporious"), &["spurious"]),
            (
                dictgen::InsensitiveStr::Ascii("implmenetaion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("priviledge"), &["privilege"]),
            (
                dictgen::InsensitiveStr::Ascii("improvmenet"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associting"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("catagory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("checher"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("metacharaters"),
                &["metacharacters"],
            ),
            (dictgen::InsensitiveStr::Ascii("apoligise"), &["apologise"]),
            (dictgen::InsensitiveStr::Ascii("detectivs"), &["detectives"]),
            (dictgen::InsensitiveStr::Ascii("definance"), &["defiance"]),
            (
                dictgen::InsensitiveStr::Ascii("circumsition"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("funvion"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpereinced"),
                &["inexperienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authroized"),
                &["authorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realisitcally"),
                &["realistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("requesteed"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("clicklabe"), &["clickable"]),
            (dictgen::InsensitiveStr::Ascii("sedn"), &["send"]),
            (
                dictgen::InsensitiveStr::Ascii("ackowledgement"),
                &["acknowledgement"],
            ),
            (dictgen::InsensitiveStr::Ascii("attempst"), &["attempts"]),
            (dictgen::InsensitiveStr::Ascii("alos"), &["also"]),
            (dictgen::InsensitiveStr::Ascii("eddge"), &["edge"]),
            (
                dictgen::InsensitiveStr::Ascii("inconviniantly"),
                &["inconveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripriptors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depentently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilties"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("diffencne"), &["difference"]),
            (dictgen::InsensitiveStr::Ascii("groupping"), &["grouping"]),
            (
                dictgen::InsensitiveStr::Ascii("midifications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("manally"), &["manually"]),
            (dictgen::InsensitiveStr::Ascii("recogonize"), &["recognize"]),
            (
                dictgen::InsensitiveStr::Ascii("contration"),
                &["contraction", "contraption"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetainty"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("excetutor"), &["executor"]),
            (dictgen::InsensitiveStr::Ascii("trucnates"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("bjects"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("matricy"), &["matrices"]),
            (
                dictgen::InsensitiveStr::Ascii("introductury"),
                &["introductory"],
            ),
            (dictgen::InsensitiveStr::Ascii("multilpy"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("independnetly"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatabillity"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("howeveer"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("encose"),
                &["enclose", "encode"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrolling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcatagory"),
                &["subcategory"],
            ),
            (dictgen::InsensitiveStr::Ascii("vageuly"), &["vaguely"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaizations"),
                &["generalizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("preppend"), &["prepend"]),
            (
                dictgen::InsensitiveStr::Ascii("photographier"),
                &["photographer"],
            ),
            (dictgen::InsensitiveStr::Ascii("effcts"), &["effects"]),
            (
                dictgen::InsensitiveStr::Ascii("aprreciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumad"), &["assumed"]),
            (dictgen::InsensitiveStr::Ascii("beneeth"), &["beneath"]),
            (
                dictgen::InsensitiveStr::Ascii("unviersities"),
                &["universities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perogrative"),
                &["prerogative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emprovements"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbirtary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("varstiy"), &["varsity"]),
            (dictgen::InsensitiveStr::Ascii("togle"), &["toggle"]),
            (
                dictgen::InsensitiveStr::Ascii("abosulutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscarrige"),
                &["miscarriage"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisonos"), &["poisons"]),
            (
                dictgen::InsensitiveStr::Ascii("shamelesly"),
                &["shamelessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("melinneum"), &["millennium"]),
            (
                dictgen::InsensitiveStr::Ascii("denomonator"),
                &["denominator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstanidng"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("platoform"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("corses"),
                &[
                    "courses", "cores", "curses", "horses", "norses", "corpses", "torses",
                    "corsets",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominador"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("asteorids"), &["asteroids"]),
            (dictgen::InsensitiveStr::Ascii("perforamer"), &["performer"]),
            (dictgen::InsensitiveStr::Ascii("brwose"), &["browse"]),
            (dictgen::InsensitiveStr::Ascii("lisensed"), &["licensed"]),
            (
                dictgen::InsensitiveStr::Ascii("compulsivley"),
                &["compulsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aestheticaly"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propect"),
                &["project", "prospect", "protect"],
            ),
            (dictgen::InsensitiveStr::Ascii("umnarshal"), &["unmarshal"]),
            (dictgen::InsensitiveStr::Ascii("trainng"), &["training"]),
            (dictgen::InsensitiveStr::Ascii("prooxy"), &["proxy"]),
            (dictgen::InsensitiveStr::Ascii("pnatheon"), &["pantheon"]),
            (
                dictgen::InsensitiveStr::Ascii("resolition"),
                &["resolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("returend"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("illegitimite"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deintializing"),
                &["deinitializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("righteos"), &["righteous"]),
            (
                dictgen::InsensitiveStr::Ascii("independance"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("satelites"), &["satellites"]),
            (
                dictgen::InsensitiveStr::Ascii("techniquest"),
                &["techniques"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitfies"),
                &["identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("reviwing"), &["reviewing"]),
            (dictgen::InsensitiveStr::Ascii("fashiond"), &["fashioned"]),
            (dictgen::InsensitiveStr::Ascii("zlot"), &["slot"]),
            (
                dictgen::InsensitiveStr::Ascii("manupilate"),
                &["manipulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrog"), &["wrong"]),
            (dictgen::InsensitiveStr::Ascii("comsetic"), &["cosmetic"]),
            (
                dictgen::InsensitiveStr::Ascii("implimentaions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaition"), &["animation"]),
            (dictgen::InsensitiveStr::Ascii("connic"), &["conic"]),
            (dictgen::InsensitiveStr::Ascii("vitories"), &["victories"]),
            (
                dictgen::InsensitiveStr::Ascii("estabishes"),
                &["establishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonexisted"),
                &["nonexistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("tect"), &["test", "text"]),
            (
                dictgen::InsensitiveStr::Ascii("linguitics"),
                &["linguistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("singlely"), &["singly"]),
            (
                dictgen::InsensitiveStr::Ascii("testasterone"),
                &["testosterone"],
            ),
            (dictgen::InsensitiveStr::Ascii("cecking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("intervine"), &["intervene"]),
            (
                dictgen::InsensitiveStr::Ascii("meaurements"),
                &["measurements"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifrom"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacreficed"),
                &["sacrificed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computanti"),
                &["computation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ressurected"),
                &["resurrected"],
            ),
            (dictgen::InsensitiveStr::Ascii("returnig"), &["returning"]),
            (dictgen::InsensitiveStr::Ascii("weirldy"), &["weirdly"]),
            (
                dictgen::InsensitiveStr::Ascii("attributted"),
                &["attribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("idealsim"), &["idealism"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreggated"),
                &["aggregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obeservation"),
                &["observation"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadmics"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("provode"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("functino"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuccee"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("massoose"), &["masseuse"]),
            (dictgen::InsensitiveStr::Ascii("previouse"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("parrallel"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("sotre"), &["store"]),
            (dictgen::InsensitiveStr::Ascii("friedly"), &["friendly"]),
            (
                dictgen::InsensitiveStr::Ascii("westerness"),
                &["westerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valiidation"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("signatues"), &["signatures"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapoited"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressulted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("professonial"),
                &["professionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkklist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("transfos"), &["transforms"]),
            (dictgen::InsensitiveStr::Ascii("condenm"), &["condemn"]),
            (
                dictgen::InsensitiveStr::Ascii("despciable"),
                &["despicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frontlinjen"),
                &["frontline"],
            ),
            (dictgen::InsensitiveStr::Ascii("configurse"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("accidentaly"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosemtics"), &["cosmetics"]),
            (
                dictgen::InsensitiveStr::Ascii("europeaness"),
                &["europeans"],
            ),
            (dictgen::InsensitiveStr::Ascii("flashs"), &["flashes"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimentel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccurs"),
                &["recurs", "reoccurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanaguges"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("renewabe"), &["renewables"]),
            (dictgen::InsensitiveStr::Ascii("anistropy"), &["anisotropy"]),
            (dictgen::InsensitiveStr::Ascii("deviding"), &["dividing"]),
            (
                dictgen::InsensitiveStr::Ascii("acelerating"),
                &["accelerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proecting"),
                &["protecting", "projecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dsktop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinates"),
                &["terminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resturctured"),
                &["restructured"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfor"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("whis"), &["this", "whisk"]),
            (dictgen::InsensitiveStr::Ascii("variablas"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("enpoint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("difficault"), &["difficult"]),
            (
                dictgen::InsensitiveStr::Ascii("agrandizes"),
                &["aggrandizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("oveloads"), &["overloads"]),
            (
                dictgen::InsensitiveStr::Ascii("sratching"),
                &["scratching", "snatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unawnsered"),
                &["unanswered"],
            ),
            (dictgen::InsensitiveStr::Ascii("prjects"), &["projects"]),
            (dictgen::InsensitiveStr::Ascii("deplyment"), &["deployment"]),
            (
                dictgen::InsensitiveStr::Ascii("rember"),
                &["remember", "renumber", "member"],
            ),
            (dictgen::InsensitiveStr::Ascii("determinee"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("hyjack"), &["hijack"]),
            (
                dictgen::InsensitiveStr::Ascii("retardatron"),
                &["retardation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaiantly"), &["valiantly"]),
            (
                dictgen::InsensitiveStr::Ascii("fiew"),
                &["few", "flew", "view"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearions"), &["operations"]),
            (dictgen::InsensitiveStr::Ascii("upgrads"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("indecator"), &["indicator"]),
            (dictgen::InsensitiveStr::Ascii("trivials"), &["trivial"]),
            (
                dictgen::InsensitiveStr::Ascii("neiborhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("thorws"), &["throws"]),
            (
                dictgen::InsensitiveStr::Ascii("bookkeepin"),
                &["bookkeeping"],
            ),
            (dictgen::InsensitiveStr::Ascii("whith"), &["with", "which"]),
            (
                dictgen::InsensitiveStr::Ascii("disproportionaly"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("sketpical"), &["skeptical"]),
            (dictgen::InsensitiveStr::Ascii("authorded"), &["authored"]),
            (dictgen::InsensitiveStr::Ascii("decriptor"), &["descriptor"]),
            (dictgen::InsensitiveStr::Ascii("foever"), &["forever"]),
            (dictgen::InsensitiveStr::Ascii("sirvayled"), &["surveiled"]),
            (dictgen::InsensitiveStr::Ascii("zoocheenei"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("indexin"), &["indexing"]),
            (
                dictgen::InsensitiveStr::Ascii("drasitcally"),
                &["drastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("fligths"), &["flights"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficents"),
                &["coefficients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constituite"),
                &["constitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidental"),
                &["presidential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltrar"),
                &["infiltrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservatisim"),
                &["conservatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpters"), &["adapters"]),
            (
                dictgen::InsensitiveStr::Ascii("monopilies"),
                &["monopolies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assupmption"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciritic"), &["critic"]),
            (
                dictgen::InsensitiveStr::Ascii("complanied"),
                &["complained"],
            ),
            (dictgen::InsensitiveStr::Ascii("doess"), &["does"]),
            (dictgen::InsensitiveStr::Ascii("remeotes"), &["remotes"]),
            (
                dictgen::InsensitiveStr::Ascii("cooresponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triuangulate"),
                &["triangulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("courtrooom"), &["courtroom"]),
            (dictgen::InsensitiveStr::Ascii("exchnging"), &["exchanging"]),
            (dictgen::InsensitiveStr::Ascii("prctiles"), &["percentiles"]),
            (
                dictgen::InsensitiveStr::Ascii("condifurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimilia"),
                &["dissimilar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethicating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restaruant"),
                &["restaurant"],
            ),
            (dictgen::InsensitiveStr::Ascii("deciple"), &["disciple"]),
            (dictgen::InsensitiveStr::Ascii("yoyr"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("acess"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("hangig"), &["hanging"]),
            (
                dictgen::InsensitiveStr::Ascii("precuations"),
                &["precautions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsessors"),
                &["successors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitipliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("qucik"), &["quick"]),
            (dictgen::InsensitiveStr::Ascii("squshed"), &["squashed"]),
            (
                dictgen::InsensitiveStr::Ascii("sucsessful"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standarize"),
                &["standardize"],
            ),
            (dictgen::InsensitiveStr::Ascii("loadous"), &["loadouts"]),
            (
                dictgen::InsensitiveStr::Ascii("discreminates"),
                &["discriminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immigraiton"),
                &["immigration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handshking"),
                &["handshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("chrisitan"), &["christian"]),
            (
                dictgen::InsensitiveStr::Ascii("secene"),
                &["scene", "serene", "secede"],
            ),
            (dictgen::InsensitiveStr::Ascii("rechargin"), &["recharging"]),
            (
                dictgen::InsensitiveStr::Ascii("ceritificate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("threfore"), &["therefore"]),
            (
                dictgen::InsensitiveStr::Ascii("personalie"),
                &["personalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concesions"),
                &["concession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brakedowns"),
                &["breakdowns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tendincies"),
                &["tendencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychriatic"),
                &["psychiatric"],
            ),
            (dictgen::InsensitiveStr::Ascii("exterioara"), &["exterior"]),
            (
                dictgen::InsensitiveStr::Ascii("introduccion"),
                &["introduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirtualizaion"),
                &["devirtualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otated"),
                &["notated", "rotated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniuplator"),
                &["manipulator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obstructin"),
                &["obstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("tottaly"), &["totally"]),
            (dictgen::InsensitiveStr::Ascii("resizin"), &["resizing"]),
            (dictgen::InsensitiveStr::Ascii("donwload"), &["download"]),
            (dictgen::InsensitiveStr::Ascii("resusable"), &["reusable"]),
            (dictgen::InsensitiveStr::Ascii("methd"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("gauarantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabililties"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jailbroaken"),
                &["jailbroken"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyeshdaow"), &["eyeshadow"]),
            (dictgen::InsensitiveStr::Ascii("silybuses"), &["syllabuses"]),
            (
                dictgen::InsensitiveStr::Ascii("authendicator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handshales"),
                &["handshakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expiremental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarrssing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("aleniated"), &["alienated"]),
            (dictgen::InsensitiveStr::Ascii("obhects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticaing"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberablility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indentifiers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("tendancy"), &["tendency"]),
            (dictgen::InsensitiveStr::Ascii("adatpted"), &["adapted"]),
            (dictgen::InsensitiveStr::Ascii("stautes"), &["statues"]),
            (dictgen::InsensitiveStr::Ascii("enrypt"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("upstrema"), &["upstream"]),
            (dictgen::InsensitiveStr::Ascii("recevies"), &["receives"]),
            (dictgen::InsensitiveStr::Ascii("lighbulb"), &["lightbulb"]),
            (dictgen::InsensitiveStr::Ascii("icreased"), &["increased"]),
            (
                dictgen::InsensitiveStr::Ascii("chararacter"),
                &["character"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatio"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficiteid"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuploder"), &["reuploader"]),
            (
                dictgen::InsensitiveStr::Ascii("orientiation"),
                &["orientation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependntly"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexualtiy"),
                &["homosexuality"],
            ),
            (dictgen::InsensitiveStr::Ascii("carfully"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("excentricity"),
                &["eccentricity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("christmans"),
                &["christians"],
            ),
            (dictgen::InsensitiveStr::Ascii("enouph"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("mangeled"), &["mangled"]),
            (dictgen::InsensitiveStr::Ascii("niearest"), &["nearest"]),
            (dictgen::InsensitiveStr::Ascii("compaint"), &["complaint"]),
            (dictgen::InsensitiveStr::Ascii("somethn"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("inspriation"),
                &["inspiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("dettached"), &["detached"]),
            (dictgen::InsensitiveStr::Ascii("suppoort"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("adavantage"), &["advantage"]),
            (
                dictgen::InsensitiveStr::Ascii("provervial"),
                &["proverbial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphapetic"),
                &["alphabetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlcine"), &["incline"]),
            (
                dictgen::InsensitiveStr::Ascii("innapropriate"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("documnt"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("considerato"),
                &["consideration"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwer"), &["answer"]),
            (
                dictgen::InsensitiveStr::Ascii("precumputed"),
                &["precomputed"],
            ),
            (dictgen::InsensitiveStr::Ascii("addiitons"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("gereated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("characterstic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("shotcut"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("gamepaly"), &["gameplay"]),
            (dictgen::InsensitiveStr::Ascii("statmenmt"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("playbody"), &["playboy"]),
            (
                dictgen::InsensitiveStr::Ascii("discconetion"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocciative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefetchs"), &["prefetches"]),
            (
                dictgen::InsensitiveStr::Ascii("furstrations"),
                &["frustrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influincing"),
                &["influencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degration"),
                &["degradation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adolsecent"),
                &["adolescent"],
            ),
            (dictgen::InsensitiveStr::Ascii("massagens"), &["massages"]),
            (
                dictgen::InsensitiveStr::Ascii("concentartions"),
                &["concentrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sybsystems"),
                &["subsystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placematt"),
                &["placemat", "placement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectivety"),
                &["objectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coornates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("assualted"), &["assaulted"]),
            (dictgen::InsensitiveStr::Ascii("dicarding"), &["discarding"]),
            (dictgen::InsensitiveStr::Ascii("sopport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("pronouce"), &["pronounce"]),
            (dictgen::InsensitiveStr::Ascii("receiveing"), &["receiving"]),
            (dictgen::InsensitiveStr::Ascii("equippment"), &["equipment"]),
            (
                dictgen::InsensitiveStr::Ascii("intriuging"),
                &["intriguing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conctructors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predictious"),
                &["predictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contextuel"),
                &["contextual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegitarions"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intereferences"),
                &["interferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinateions"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chandeleer"),
                &["chandelier"],
            ),
            (dictgen::InsensitiveStr::Ascii("opereates"), &["operates"]),
            (dictgen::InsensitiveStr::Ascii("exessive"), &["excessive"]),
            (dictgen::InsensitiveStr::Ascii("wekaer"), &["weaker"]),
            (
                dictgen::InsensitiveStr::Ascii("skitsofrinics"),
                &["schizophrenics"],
            ),
            (dictgen::InsensitiveStr::Ascii("occurance"), &["occurrence"]),
            (
                dictgen::InsensitiveStr::Ascii("subdirectorys"),
                &["subdirectories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constractor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("disccussed"), &["discussed"]),
            (
                dictgen::InsensitiveStr::Ascii("persumptuous"),
                &["presumptuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insistance"),
                &["insistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workarround"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enfringement"),
                &["infringement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desscribing"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("consol"), &["console"]),
            (dictgen::InsensitiveStr::Ascii("rebision"), &["revision"]),
            (dictgen::InsensitiveStr::Ascii("campaings"), &["campaigns"]),
            (
                dictgen::InsensitiveStr::Ascii("exhilaratin"),
                &["exhilarating", "exhilaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("conecctor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("crytocurrency"),
                &["cryptocurrency"],
            ),
            (dictgen::InsensitiveStr::Ascii("medievel"), &["medieval"]),
            (
                dictgen::InsensitiveStr::Ascii("overclicking"),
                &["overclocking"],
            ),
            (dictgen::InsensitiveStr::Ascii("centruies"), &["centuries"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanentel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreasonablly"),
                &["unreasonably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomalising"),
                &["normalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nurtitional"),
                &["nutritional"],
            ),
            (dictgen::InsensitiveStr::Ascii("registrs"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("instrance"),
                &["instance", "insurance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degenerare"),
                &["degenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("pendig"), &["pending"]),
            (dictgen::InsensitiveStr::Ascii("coallessed"), &["coalesced"]),
            (
                dictgen::InsensitiveStr::Ascii("negotioated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bakcgrounds"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteligance"),
                &["intelligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dipose"),
                &["dispose", "depose"],
            ),
            (dictgen::InsensitiveStr::Ascii("werstler"), &["wrestler"]),
            (dictgen::InsensitiveStr::Ascii("vanishd"), &["vanished"]),
            (dictgen::InsensitiveStr::Ascii("asendancy"), &["ascendancy"]),
            (dictgen::InsensitiveStr::Ascii("virture"), &["virtue"]),
            (
                dictgen::InsensitiveStr::Ascii("atttachment"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("shuttdown"), &["shutdown"]),
            (dictgen::InsensitiveStr::Ascii("delfect"), &["deflect"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetnation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("zimpap"), &["zipmap"]),
            (
                dictgen::InsensitiveStr::Ascii("approxametely"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backreferrence"),
                &["backreference"],
            ),
            (dictgen::InsensitiveStr::Ascii("inersial"), &["inertial"]),
            (dictgen::InsensitiveStr::Ascii("atended"), &["attended"]),
            (
                dictgen::InsensitiveStr::Ascii("exhautively"),
                &["exhaustively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterization"),
                &["parameterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("gimmics"), &["gimmicks"]),
            (dictgen::InsensitiveStr::Ascii("seatch"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("unbeateble"),
                &["unbeatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("bebefore"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("positoning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straitforward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("toi"), &["to", "toy"]),
            (dictgen::InsensitiveStr::Ascii("opeates"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("conciveably"),
                &["conceivably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoperation"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("galllerys"), &["galleries"]),
            (dictgen::InsensitiveStr::Ascii("cabinettas"), &["cabinets"]),
            (dictgen::InsensitiveStr::Ascii("pinpiont"), &["pinpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("statuts"),
                &["statutes", "status", "statues"],
            ),
            (dictgen::InsensitiveStr::Ascii("lengthes"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("critizicing"),
                &["criticizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iplementing"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponteneous"),
                &["spontaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplishements"),
                &["accomplishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggrgates"), &["aggregates"]),
            (
                dictgen::InsensitiveStr::Ascii("shtoppes"),
                &["stops", "shops"],
            ),
            (dictgen::InsensitiveStr::Ascii("brtually"), &["brutally"]),
            (dictgen::InsensitiveStr::Ascii("audiobok"), &["audiobook"]),
            (dictgen::InsensitiveStr::Ascii("chargind"), &["charging"]),
            (
                dictgen::InsensitiveStr::Ascii("deceptivley"),
                &["deceptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annilihate"),
                &["annihilated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaiters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticapated"),
                &["anticipated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commercialy"),
                &["commercially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admittidly"),
                &["admittedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("chipslect"), &["chipselect"]),
            (
                dictgen::InsensitiveStr::Ascii("equilbirium"),
                &["equilibrium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finalle"),
                &["finale", "finally"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangagers"), &["managers"]),
            (dictgen::InsensitiveStr::Ascii("illionis"), &["illinois"]),
            (dictgen::InsensitiveStr::Ascii("opitmises"), &["optimises"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulater"),
                &["manipulator", "manipulated", "manipulates", "manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumabely"),
                &["presumably"],
            ),
            (dictgen::InsensitiveStr::Ascii("trrigers"), &["triggers"]),
            (dictgen::InsensitiveStr::Ascii("foreamrs"), &["forearms"]),
            (
                dictgen::InsensitiveStr::Ascii("integreations"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ublish"), &["publish"]),
            (
                dictgen::InsensitiveStr::Ascii("perameters"),
                &["parameters", "perimeters"],
            ),
            (dictgen::InsensitiveStr::Ascii("ened"), &["need"]),
            (
                dictgen::InsensitiveStr::Ascii("coprorations"),
                &["corporations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composities"),
                &["composite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalzation"),
                &["capitalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rsources"),
                &["resources", "sources"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iteracting"),
                &["interacting", "iterating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favouritsm"),
                &["favourites"],
            ),
            (dictgen::InsensitiveStr::Ascii("correlatd"), &["correlated"]),
            (dictgen::InsensitiveStr::Ascii("careflly"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("fermintation"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("oposite"), &["opposite"]),
            (dictgen::InsensitiveStr::Ascii("admiraal"), &["admiral"]),
            (
                dictgen::InsensitiveStr::Ascii("stram"),
                &["steam", "stream", "tram"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interragation"),
                &["interrogation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structually"),
                &["structurally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitralrily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("tournamet"), &["tournament"]),
            (
                dictgen::InsensitiveStr::Ascii("reprented"),
                &["repented", "represented", "reprinted"],
            ),
            (dictgen::InsensitiveStr::Ascii("liberas"), &["liberals"]),
            (dictgen::InsensitiveStr::Ascii("gattling"), &["gatling"]),
            (dictgen::InsensitiveStr::Ascii("refillls"), &["refills"]),
            (dictgen::InsensitiveStr::Ascii("generaors"), &["generators"]),
            (
                dictgen::InsensitiveStr::Ascii("relatievly"),
                &["relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("radioactice"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhibtion"), &["exhibition"]),
            (dictgen::InsensitiveStr::Ascii("speficeis"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("replicaition"),
                &["replication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retransferrd"),
                &["retransferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternatvely"),
                &["alternately", "alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("trggering"), &["triggering"]),
            (
                dictgen::InsensitiveStr::Ascii("uncompletely"),
                &["incompletely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("erroneouly"),
                &["erroneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("furter"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("depature"), &["departure"]),
            (
                dictgen::InsensitiveStr::Ascii("synthesing"),
                &["synthesizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectiom"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atructures"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("mysterieus"), &["mysteries"]),
            (dictgen::InsensitiveStr::Ascii("assignt"), &["assign"]),
            (dictgen::InsensitiveStr::Ascii("incedents"), &["incidents"]),
            (
                dictgen::InsensitiveStr::Ascii("reconecctions"),
                &["reconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handwritng"),
                &["handwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stalagtites"),
                &["stalagmites", "stalactites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achievemts"),
                &["achieves", "achievements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremential"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rondazyvou"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("datbases"), &["databases"]),
            (
                dictgen::InsensitiveStr::Ascii("instuctional"),
                &["instructional"],
            ),
            (dictgen::InsensitiveStr::Ascii("disruptin"), &["disruption"]),
            (
                dictgen::InsensitiveStr::Ascii("banched"),
                &["branched", "benched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioratize"),
                &["prioritize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derference"),
                &["deference", "dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("declering"), &["declaring"]),
            (
                dictgen::InsensitiveStr::Ascii("relocaitions"),
                &["relocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("decsended"), &["descended"]),
            (
                dictgen::InsensitiveStr::Ascii("chandeleers"),
                &["chandeliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("fairoh"), &["pharaoh"]),
            (dictgen::InsensitiveStr::Ascii("frocing"), &["forcing"]),
            (dictgen::InsensitiveStr::Ascii("baslines"), &["baselines"]),
            (
                dictgen::InsensitiveStr::Ascii("duplacated"),
                &["duplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funvtioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragrpah"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("relize"), &["realize"]),
            (dictgen::InsensitiveStr::Ascii("fetishs"), &["fetishes"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvienent"),
                &["inconvenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("universiy"), &["university"]),
            (dictgen::InsensitiveStr::Ascii("chruches"), &["churches"]),
            (dictgen::InsensitiveStr::Ascii("ckeckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("severirirty"), &["severity"]),
            (dictgen::InsensitiveStr::Ascii("verticles"), &["verticals"]),
            (
                dictgen::InsensitiveStr::Ascii("noticies"),
                &["notices", "notifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompatibility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictadorship"),
                &["dictatorship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deselcting"),
                &["deselecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("deamands"), &["demands"]),
            (dictgen::InsensitiveStr::Ascii("gurrantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("correlacion"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incentivice"),
                &["incentive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ridiculue"), &["ridicule"]),
            (dictgen::InsensitiveStr::Ascii("ternament"), &["tournament"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmisation"),
                &["optimisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefifed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whatosever"),
                &["whatsoever"],
            ),
            (dictgen::InsensitiveStr::Ascii("contextl"), &["contextual"]),
            (dictgen::InsensitiveStr::Ascii("prefernce"), &["preference"]),
            (
                dictgen::InsensitiveStr::Ascii("interfernce"),
                &["interference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appereance"),
                &["appearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosophae"),
                &["philosopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configares"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("avance"), &["advance"]),
            (
                dictgen::InsensitiveStr::Ascii("suppurtive"),
                &["supportive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interprered"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("buiilt"), &["built"]),
            (dictgen::InsensitiveStr::Ascii("diety"), &["deity"]),
            (dictgen::InsensitiveStr::Ascii("intilaize"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("jurisfiction"),
                &["jurisdiction"],
            ),
            (dictgen::InsensitiveStr::Ascii("clinets"), &["clients"]),
            (dictgen::InsensitiveStr::Ascii("falied"), &["failed"]),
            (
                dictgen::InsensitiveStr::Ascii("condescenion"),
                &["condescension"],
            ),
            (dictgen::InsensitiveStr::Ascii("anull"), &["annul"]),
            (dictgen::InsensitiveStr::Ascii("dicsuss"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("consdiers"), &["considers"]),
            (
                dictgen::InsensitiveStr::Ascii("bootstraping"),
                &["bootstrapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplised"),
                &["accomplished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precausion"),
                &["precaution"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabients"), &["cabinets"]),
            (
                dictgen::InsensitiveStr::Ascii("excptional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conenctions"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonther"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eleiminating"),
                &["eliminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("rummble"), &["rumble"]),
            (dictgen::InsensitiveStr::Ascii("bruses"), &["bruises"]),
            (dictgen::InsensitiveStr::Ascii("bandwagoon"), &["bandwagon"]),
            (
                dictgen::InsensitiveStr::Ascii("ditorconfig"),
                &["editorconfig"],
            ),
            (dictgen::InsensitiveStr::Ascii("catherdal"), &["cathedral"]),
            (dictgen::InsensitiveStr::Ascii("outskrits"), &["outskirts"]),
            (dictgen::InsensitiveStr::Ascii("formmats"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("accelerade"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forgivenness"),
                &["forgiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("reviws"), &["reviews"]),
            (dictgen::InsensitiveStr::Ascii("infaltes"), &["inflates"]),
            (dictgen::InsensitiveStr::Ascii("influnce"), &["influence"]),
            (
                dictgen::InsensitiveStr::Ascii("compolsory"),
                &["compulsory"],
            ),
            (dictgen::InsensitiveStr::Ascii("adeladie"), &["adelaide"]),
            (
                dictgen::InsensitiveStr::Ascii("alltogehter"),
                &["altogether"],
            ),
            (dictgen::InsensitiveStr::Ascii("rightoues"), &["righteous"]),
            (
                dictgen::InsensitiveStr::Ascii("confgiured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("analoge"), &["analogue"]),
            (dictgen::InsensitiveStr::Ascii("varaiables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("diffrently"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desable"),
                &["disable", "desirable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperation"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reposiry"), &["repository"]),
            (dictgen::InsensitiveStr::Ascii("intervalos"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("inceased"),
                &["increased", "incensed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oce"),
                &["once", "one", "ore"],
            ),
            (dictgen::InsensitiveStr::Ascii("decreypted"), &["decrypted"]),
            (dictgen::InsensitiveStr::Ascii("silenty"), &["silently"]),
            (dictgen::InsensitiveStr::Ascii("chanceled"), &["canceled"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomd"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("chocoalte"), &["chocolate"]),
            (
                dictgen::InsensitiveStr::Ascii("circumcisied"),
                &["circumcised"],
            ),
            (dictgen::InsensitiveStr::Ascii("durning"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("opearor"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("offficer"), &["officer"]),
            (dictgen::InsensitiveStr::Ascii("segfualt"), &["segfault"]),
            (dictgen::InsensitiveStr::Ascii("dthe"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("statistcs"), &["statistics"]),
            (
                dictgen::InsensitiveStr::Ascii("interrruptible"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicitave"),
                &["indicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabke"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("transcording"),
                &["transcoding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undeniablely"),
                &["undeniably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapperars"),
                &["disappears"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypotetical"),
                &["hypothetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concsiously"),
                &["consciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("individul"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("recusrive"), &["recursive"]),
            (dictgen::InsensitiveStr::Ascii("caesarian"), &["caesarean"]),
            (dictgen::InsensitiveStr::Ascii("allignemnt"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstatiates"),
                &["reinstantiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaulate"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("retursn"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("garantie"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("extrem"),
                &["extremum", "extreme"],
            ),
            (dictgen::InsensitiveStr::Ascii("olmypic"), &["olympic"]),
            (
                dictgen::InsensitiveStr::Ascii("directoies"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("slightlty"), &["slightly"]),
            (dictgen::InsensitiveStr::Ascii("srunk"), &["shrunk"]),
            (
                dictgen::InsensitiveStr::Ascii("approproately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("spotfiy"), &["spotify"]),
            (
                dictgen::InsensitiveStr::Ascii("representativo"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("activistas"), &["activists"]),
            (
                dictgen::InsensitiveStr::Ascii("deffinitive"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("saem"), &["same"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactering"),
                &["manufacturing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubance"),
                &["disturbance"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryifying"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("projeting"), &["projecting"]),
            (dictgen::InsensitiveStr::Ascii("suuport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("follwoein"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("expecets"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("teliportation"),
                &["teleportation"],
            ),
            (dictgen::InsensitiveStr::Ascii("callibrate"), &["calibrate"]),
            (dictgen::InsensitiveStr::Ascii("collpases"), &["collapses"]),
            (
                dictgen::InsensitiveStr::Ascii("communicatons"),
                &["communications"],
            ),
            (dictgen::InsensitiveStr::Ascii("becausee"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("disucssion"),
                &["discussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wighted"),
                &["weighted", "weighed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrctors"),
                &["constructors", "constrictors"],
            ),
            (dictgen::InsensitiveStr::Ascii("resourecs"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("exppressed"), &["expressed"]),
            (
                dictgen::InsensitiveStr::Ascii("virtualizaion"),
                &["virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("observeres"), &["observers"]),
            (
                dictgen::InsensitiveStr::Ascii("comperhension"),
                &["comprehension"],
            ),
            (dictgen::InsensitiveStr::Ascii("userrs"), &["users"]),
            (
                dictgen::InsensitiveStr::Ascii("emplied"),
                &["implied", "emptied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcitonalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("athletiscm"),
                &["athleticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmarized"),
                &["summarized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intolorant"),
                &["intolerant"],
            ),
            (dictgen::InsensitiveStr::Ascii("decrption"), &["decryption"]),
            (
                dictgen::InsensitiveStr::Ascii("variantions"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("eigtheen"), &["eighteen"]),
            (
                dictgen::InsensitiveStr::Ascii("missingassignement"),
                &["missingassignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("usres"), &["users", "uses"]),
            (dictgen::InsensitiveStr::Ascii("liquiidity"), &["liquidity"]),
            (dictgen::InsensitiveStr::Ascii("equvalent"), &["equivalent"]),
            (
                dictgen::InsensitiveStr::Ascii("coincedental"),
                &["coincidental"],
            ),
            (dictgen::InsensitiveStr::Ascii("hungs"), &["hangs", "hung"]),
            (dictgen::InsensitiveStr::Ascii("searialize"), &["serialize"]),
            (dictgen::InsensitiveStr::Ascii("awkawrdly"), &["awkwardly"]),
            (
                dictgen::InsensitiveStr::Ascii("meterosexual"),
                &["heterosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallellises"),
                &["parallelises"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabking"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("imcompatible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaccuums"), &["vacuums"]),
            (dictgen::InsensitiveStr::Ascii("upddate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("dictaionary"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpises"), &["surprises"]),
            (dictgen::InsensitiveStr::Ascii("worngs"), &["wrongs"]),
            (dictgen::InsensitiveStr::Ascii("langauges"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("lincesed"), &["licensed"]),
            (dictgen::InsensitiveStr::Ascii("modifieres"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("serveiced"), &["serviced"]),
            (
                dictgen::InsensitiveStr::Ascii("begind"),
                &["behind", "begin"],
            ),
            (dictgen::InsensitiveStr::Ascii("upadting"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("disrecpecting"),
                &["disrespecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provsionally"),
                &["provisionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservativs"),
                &["conservatism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oprimizations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("romote"), &["remote"]),
            (dictgen::InsensitiveStr::Ascii("roesetta"), &["rosetta"]),
            (dictgen::InsensitiveStr::Ascii("catpture"), &["capture"]),
            (dictgen::InsensitiveStr::Ascii("aand"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("subsripts"), &["subscripts"]),
            (dictgen::InsensitiveStr::Ascii("changin"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("whitesspaces"),
                &["whitespaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramedicos"),
                &["paramedics"],
            ),
            (dictgen::InsensitiveStr::Ascii("usrerspace"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("onmishients"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("languaqe"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uggestions"),
                &["suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randazyvoo"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("romanmania"), &["romanian"]),
            (
                dictgen::InsensitiveStr::Ascii("additioanl"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloocates"), &["allocates"]),
            (
                dictgen::InsensitiveStr::Ascii("recconencted"),
                &["reconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("covarege"), &["coverage"]),
            (dictgen::InsensitiveStr::Ascii("jasmien"), &["jasmine"]),
            (
                dictgen::InsensitiveStr::Ascii("dprecation"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifcations"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("substact"), &["subtract"]),
            (dictgen::InsensitiveStr::Ascii("politicas"), &["politics"]),
            (dictgen::InsensitiveStr::Ascii("confuze"), &["confuse"]),
            (
                dictgen::InsensitiveStr::Ascii("optionnaly"),
                &["optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iternations"),
                &["iterations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lable"),
                &["label", "ladle", "labile", "able"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arcitecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dentification"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("fnctions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("adherin"),
                &["adhering", "cadherin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intsrumental"),
                &["instrumental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("observatrions"),
                &["observations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persmissions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afficionado"),
                &["aficionado"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanaging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("interefering"),
                &["interfering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelorates"),
                &["accelerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyborads"), &["keyboards"]),
            (
                dictgen::InsensitiveStr::Ascii("microntrollers"),
                &["microcontrollers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviced"),
                &["convinced", "convicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repuglican"),
                &["republican"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accquainted"),
                &["acquainted"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatins"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("flexibiltiy"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connctions"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptoy"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahears"), &["adheres"]),
            (
                dictgen::InsensitiveStr::Ascii("naybourhoods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("higly"), &["highly"]),
            (dictgen::InsensitiveStr::Ascii("installeer"), &["installer"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangable"),
                &["arrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("parralel"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("proteines"), &["proteins"]),
            (
                dictgen::InsensitiveStr::Ascii("configuretions"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progapation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mitsubushi"),
                &["mitsubishi"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlyzing"), &["analyzing"]),
            (dictgen::InsensitiveStr::Ascii("attacht"), &["attach"]),
            (
                dictgen::InsensitiveStr::Ascii("epigramic"),
                &["epigrammatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("uite"), &["suite", "unite"]),
            (
                dictgen::InsensitiveStr::Ascii("accidentley"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("survay"), &["survey"]),
            (
                dictgen::InsensitiveStr::Ascii("locatio"),
                &["locations", "location"],
            ),
            (dictgen::InsensitiveStr::Ascii("baught"), &["bought"]),
            (dictgen::InsensitiveStr::Ascii("mroe"), &["more"]),
            (dictgen::InsensitiveStr::Ascii("oftern"), &["often"]),
            (dictgen::InsensitiveStr::Ascii("nummerous"), &["numerous"]),
            (
                dictgen::InsensitiveStr::Ascii("worthing"),
                &["worth", "meriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("waths"),
                &["whats", "watches"],
            ),
            (dictgen::InsensitiveStr::Ascii("testicels"), &["testicles"]),
            (dictgen::InsensitiveStr::Ascii("limmits"), &["limits"]),
            (
                dictgen::InsensitiveStr::Ascii("professionalisme"),
                &["professionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("talbes"), &["tables"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoledgment"),
                &["acknowledgment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonrously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmneting"),
                &["segmenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("excract"), &["extract"]),
            (dictgen::InsensitiveStr::Ascii("ceate"), &["create"]),
            (
                dictgen::InsensitiveStr::Ascii("accaptable"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degnerated"),
                &["degenerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimdntal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestantes"),
                &["protestants"],
            ),
            (dictgen::InsensitiveStr::Ascii("keywoed"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("unio"), &["union"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastinarting"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("shcool"), &["school"]),
            (dictgen::InsensitiveStr::Ascii("veiws"), &["views"]),
            (
                dictgen::InsensitiveStr::Ascii("significaint"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficer"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("uless"), &["unless"]),
            (
                dictgen::InsensitiveStr::Ascii("littlepinger"),
                &["littlefinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancelaltion"),
                &["cancelation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cneter"), &["center"]),
            (
                dictgen::InsensitiveStr::Ascii("regess"),
                &["regress", "recess"],
            ),
            (dictgen::InsensitiveStr::Ascii("factores"), &["factors"]),
            (dictgen::InsensitiveStr::Ascii("comfusing"), &["confusing"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("feches"), &["fetches"]),
            (dictgen::InsensitiveStr::Ascii("tentacel"), &["tentacle"]),
            (
                dictgen::InsensitiveStr::Ascii("occasionals"),
                &["occasions"],
            ),
            (dictgen::InsensitiveStr::Ascii("acrages"), &["acreages"]),
            (dictgen::InsensitiveStr::Ascii("remeotely"), &["remotely"]),
            (dictgen::InsensitiveStr::Ascii("samlers"), &["samplers"]),
            (dictgen::InsensitiveStr::Ascii("phyisque"), &["physique"]),
            (
                dictgen::InsensitiveStr::Ascii("unstructred"),
                &["unstructured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phillippines"),
                &["philippines"],
            ),
            (dictgen::InsensitiveStr::Ascii("validata"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("aqcuisitions"),
                &["acquisitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("blessd"), &["blessed"]),
            (
                dictgen::InsensitiveStr::Ascii("volunterring"),
                &["volunteering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintentionaly"),
                &["unintentionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representitve"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifeycles"), &["lifecycles"]),
            (dictgen::InsensitiveStr::Ascii("pericing"), &["piercing"]),
            (dictgen::InsensitiveStr::Ascii("proceedure"), &["procedure"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmimately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requiriement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("woudlnt"), &["wouldnt"]),
            (dictgen::InsensitiveStr::Ascii("basilcy"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("orginy"), &["originy"]),
            (dictgen::InsensitiveStr::Ascii("rendererd"), &["rendered"]),
            (dictgen::InsensitiveStr::Ascii("harcoding"), &["hardcoding"]),
            (dictgen::InsensitiveStr::Ascii("soultion"), &["solution"]),
            (dictgen::InsensitiveStr::Ascii("featue"), &["feature"]),
            (dictgen::InsensitiveStr::Ascii("costexpr"), &["constexpr"]),
            (dictgen::InsensitiveStr::Ascii("hopsital"), &["hospital"]),
            (dictgen::InsensitiveStr::Ascii("ispatches"), &["dispatches"]),
            (
                dictgen::InsensitiveStr::Ascii("experaments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negiotating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devlopments"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("comamnd"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("officealy"), &["officially"]),
            (
                dictgen::InsensitiveStr::Ascii("bumbing"),
                &["bombing", "bumping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogeneus"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("sysmbolic"), &["symbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("innacessible"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifiguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathched"), &["matched"]),
            (dictgen::InsensitiveStr::Ascii("conatin"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("antoganist"),
                &["antagonist"],
            ),
            (dictgen::InsensitiveStr::Ascii("copmonent"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("pigins"), &["pigeons"]),
            (
                dictgen::InsensitiveStr::Ascii("prejeduces"),
                &["prejudices"],
            ),
            (dictgen::InsensitiveStr::Ascii("assingns"), &["assigns"]),
            (
                dictgen::InsensitiveStr::Ascii("certifices"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("trigggered"), &["triggered"]),
            (dictgen::InsensitiveStr::Ascii("varibaless"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("cocurrently"),
                &["concurrently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeficents"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("contians"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("indentify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("receent"), &["recent"]),
            (dictgen::InsensitiveStr::Ascii("isntalls"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("ramdom"), &["random"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnetel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbritrarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoaiating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmnets"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("erly"), &["early"]),
            (
                dictgen::InsensitiveStr::Ascii("contstraint"),
                &["constraint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenionalities"),
                &["dimensionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("originsl"), &["originals"]),
            (dictgen::InsensitiveStr::Ascii("goregous"), &["gorgeous"]),
            (
                dictgen::InsensitiveStr::Ascii("agricultue"),
                &["agriculture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precomuted"),
                &["precomputed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfitifed"),
                &["identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innersection"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("posiiton"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("refartored"),
                &["refactored", "refracted"],
            ),
            (dictgen::InsensitiveStr::Ascii("selcts"), &["selects"]),
            (dictgen::InsensitiveStr::Ascii("abundent"), &["abundant"]),
            (
                dictgen::InsensitiveStr::Ascii("percievabley"),
                &["perceivably"],
            ),
            (dictgen::InsensitiveStr::Ascii("liebrals"), &["liberals"]),
            (
                dictgen::InsensitiveStr::Ascii("releaes"),
                &["release", "releases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confedaracy"),
                &["confederacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutlipart"), &["multipart"]),
            (dictgen::InsensitiveStr::Ascii("modfiying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("quatratic"), &["quadratic"]),
            (dictgen::InsensitiveStr::Ascii("ussually"), &["usually"]),
            (dictgen::InsensitiveStr::Ascii("dandlin"), &["dandling"]),
            (dictgen::InsensitiveStr::Ascii("decorders"), &["decoders"]),
            (
                dictgen::InsensitiveStr::Ascii("speciffications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("debbuged"), &["debugged"]),
            (dictgen::InsensitiveStr::Ascii("electhor"), &["electro"]),
            (
                dictgen::InsensitiveStr::Ascii("polynomicals"),
                &["polynomials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("messurment"),
                &["measurement"],
            ),
            (dictgen::InsensitiveStr::Ascii("thisy"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("finilised"), &["finalised"]),
            (
                dictgen::InsensitiveStr::Ascii("confederecy"),
                &["confederacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspdons"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("commom"), &["common"]),
            (
                dictgen::InsensitiveStr::Ascii("adolescant"),
                &["adolescent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terimnations"),
                &["terminations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annnouncements"),
                &["announcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("impelement"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("voight"), &["voigt"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentt"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimice"),
                &["optimise", "optimize"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanve"), &["advance"]),
            (
                dictgen::InsensitiveStr::Ascii("implenmented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("uery"), &["query", "very"]),
            (
                dictgen::InsensitiveStr::Ascii("paretheses"),
                &["parentheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkstums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("protaginists"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("platorm"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("historicans"),
                &["historians"],
            ),
            (dictgen::InsensitiveStr::Ascii("comletely"), &["completely"]),
            (
                dictgen::InsensitiveStr::Ascii("captialization"),
                &["capitalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thinkgs"),
                &["thinks", "things"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temafights"),
                &["teamfights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opinyonate"),
                &["opinionated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reanames"), &["renames"]),
            (
                dictgen::InsensitiveStr::Ascii("axissymmetric"),
                &["axisymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("negotitate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("specefication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("tdod"), &["todo"]),
            (
                dictgen::InsensitiveStr::Ascii("pathalogies"),
                &["pathologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("hemorage"), &["hemorrhage"]),
            (
                dictgen::InsensitiveStr::Ascii("everithing"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parokeelity"),
                &["parochiality"],
            ),
            (dictgen::InsensitiveStr::Ascii("balanceada"), &["balanced"]),
            (dictgen::InsensitiveStr::Ascii("nomes"), &["gnomes"]),
            (dictgen::InsensitiveStr::Ascii("secord"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("oneyway"), &["oneway"]),
            (
                dictgen::InsensitiveStr::Ascii("continouos"),
                &["continuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("jurnyes"), &["journeys"]),
            (
                dictgen::InsensitiveStr::Ascii("nueroscience"),
                &["neuroscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrelyable"),
                &["unreliable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unecessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("testiclees"), &["testicles"]),
            (dictgen::InsensitiveStr::Ascii("warinig"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("pieline"), &["pipeline"]),
            (
                dictgen::InsensitiveStr::Ascii("presentacion"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciepent"), &["recipient"]),
            (dictgen::InsensitiveStr::Ascii("chechksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("represenst"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remenisses"),
                &["reminisces"],
            ),
            (dictgen::InsensitiveStr::Ascii("grafic"), &["graphic"]),
            (dictgen::InsensitiveStr::Ascii("caffeen"), &["caffeine"]),
            (dictgen::InsensitiveStr::Ascii("lnguages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("extened"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("spirtiuality"),
                &["spirituality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerets"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intensitive"),
                &["insensitive", "intensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransactioms"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mastermined"),
                &["mastermind"],
            ),
            (dictgen::InsensitiveStr::Ascii("arguemtn"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("uninialize"),
                &["uninitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("movememt"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("insuffient"),
                &["insufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("peircings"), &["piercings"]),
            (dictgen::InsensitiveStr::Ascii("theshhold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("experiantial"),
                &["experiential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("piching"),
                &["picking", "pinching", "pitching"],
            ),
            (dictgen::InsensitiveStr::Ascii("acesses"), &["accesses"]),
            (
                dictgen::InsensitiveStr::Ascii("borderlanders"),
                &["borderlands"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordian"), &["accordion"]),
            (dictgen::InsensitiveStr::Ascii("relavent"), &["relevant"]),
            (
                dictgen::InsensitiveStr::Ascii("unexplaind"),
                &["unexplained"],
            ),
            (dictgen::InsensitiveStr::Ascii("sleectors"), &["selectors"]),
            (
                dictgen::InsensitiveStr::Ascii("sistematies"),
                &["systematies"],
            ),
            (dictgen::InsensitiveStr::Ascii("recomend"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("burritto"), &["burrito"]),
            (
                dictgen::InsensitiveStr::Ascii("similer"),
                &["similar", "simile", "similes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chamiponship"),
                &["championships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fingernals"),
                &["fingernails"],
            ),
            (dictgen::InsensitiveStr::Ascii("follew"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("porblems"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("facsism"), &["fascism"]),
            (
                dictgen::InsensitiveStr::Ascii("cristallised"),
                &["crystallised"],
            ),
            (dictgen::InsensitiveStr::Ascii("intesity"), &["intensity"]),
            (
                dictgen::InsensitiveStr::Ascii("humaniterian"),
                &["humanitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanuted"), &["haunted"]),
            (dictgen::InsensitiveStr::Ascii("attrribute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("acessed"), &["accessed"]),
            (
                dictgen::InsensitiveStr::Ascii("reconfugire"),
                &["reconfigure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("populatins"),
                &["populations"],
            ),
            (dictgen::InsensitiveStr::Ascii("neccisary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigiosi"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("noit"), &["not"]),
            (dictgen::InsensitiveStr::Ascii("evalulator"), &["evaluator"]),
            (
                dictgen::InsensitiveStr::Ascii("supernarkets"),
                &["supermarkets"],
            ),
            (dictgen::InsensitiveStr::Ascii("sattelits"), &["satellites"]),
            (dictgen::InsensitiveStr::Ascii("casims"), &["chasms"]),
            (
                dictgen::InsensitiveStr::Ascii("overfowing"),
                &["overflowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorightmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("assset"), &["asset"]),
            (
                dictgen::InsensitiveStr::Ascii("incoperates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verticallity"),
                &["vertically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preexisiting"),
                &["preexisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("publsihed"), &["published"]),
            (
                dictgen::InsensitiveStr::Ascii("assingnments"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreivated"),
                &["abbreviated"],
            ),
            (dictgen::InsensitiveStr::Ascii("impreve"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("cheked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("repsositories"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("soveregin"), &["sovereign"]),
            (
                dictgen::InsensitiveStr::Ascii("intregrates"),
                &["integrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("subscibe"), &["subscribe"]),
            (
                dictgen::InsensitiveStr::Ascii("industriels"),
                &["industries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acommodate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suplemented"),
                &["supplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("chuckels"), &["chuckles"]),
            (dictgen::InsensitiveStr::Ascii("hatian"), &["haitian"]),
            (dictgen::InsensitiveStr::Ascii("marbleds"), &["marbles"]),
            (
                dictgen::InsensitiveStr::Ascii("congratualtion"),
                &["congratulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubeshoots"),
                &["troubleshoots"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceet"), &["conceit"]),
            (dictgen::InsensitiveStr::Ascii("esimating"), &["estimating"]),
            (dictgen::InsensitiveStr::Ascii("usebility"), &["usability"]),
            (
                dictgen::InsensitiveStr::Ascii("consistuting"),
                &["constituting", "consisting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entirey"),
                &["entirety", "entirely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conttroller"),
                &["controller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automoderador"),
                &["automoderator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticially"),
                &["critically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conersation"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtorisation"),
                &["authorisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perceptoin"),
                &["perceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrohpe"),
                &["catastrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranmpoline"),
                &["trampoline"],
            ),
            (dictgen::InsensitiveStr::Ascii("currupted"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("collectoin"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphrials"),
                &["peripherals"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmneted"), &["segmented"]),
            (
                dictgen::InsensitiveStr::Ascii("poritioning"),
                &["portioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("pengwen"), &["penguin"]),
            (dictgen::InsensitiveStr::Ascii("facilites"), &["facilities"]),
            (
                dictgen::InsensitiveStr::Ascii("compexities"),
                &["complexities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsturction"),
                &["reconstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normazation"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoudn"), &["shouldn"]),
            (
                dictgen::InsensitiveStr::Ascii("supllements"),
                &["supplements"],
            ),
            (dictgen::InsensitiveStr::Ascii("aregument"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("volcaron"), &["volcano"]),
            (dictgen::InsensitiveStr::Ascii("searcg"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("interdependant"),
                &["interdependent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trogladitical"),
                &["troglodytical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integreates"),
                &["integrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prececessor"),
                &["predecessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("withdrawling"),
                &["withdrawing"],
            ),
            (dictgen::InsensitiveStr::Ascii("compilies"), &["compiles"]),
            (dictgen::InsensitiveStr::Ascii("rduce"), &["reduce"]),
            (dictgen::InsensitiveStr::Ascii("finnesse"), &["finesse"]),
            (dictgen::InsensitiveStr::Ascii("encrpt"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("indecates"), &["indicates"]),
            (dictgen::InsensitiveStr::Ascii("vairous"), &["various"]),
            (dictgen::InsensitiveStr::Ascii("whenevery"), &["whenever"]),
            (dictgen::InsensitiveStr::Ascii("forntpage"), &["frontpage"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("livliehood"),
                &["livelihood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coverting"),
                &["converting", "covering", "coveting"],
            ),
            (dictgen::InsensitiveStr::Ascii("generagor"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("correcor"), &["corrector"]),
            (dictgen::InsensitiveStr::Ascii("inerested"), &["interested"]),
            (dictgen::InsensitiveStr::Ascii("indugle"), &["indulge"]),
            (dictgen::InsensitiveStr::Ascii("ocilates"), &["oscillates"]),
            (dictgen::InsensitiveStr::Ascii("pligrim"), &["pilgrim"]),
            (dictgen::InsensitiveStr::Ascii("subpackge"), &["subpackage"]),
            (
                dictgen::InsensitiveStr::Ascii("resurse"),
                &["recurse", "resource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benifitial"),
                &["beneficial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availabaility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("varnings"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("dsyfunctional"),
                &["dysfunctional"],
            ),
            (dictgen::InsensitiveStr::Ascii("funchtions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("perfecty"), &["perfectly"]),
            (dictgen::InsensitiveStr::Ascii("godspeeed"), &["godspeed"]),
            (dictgen::InsensitiveStr::Ascii("refectiv"), &["reflective"]),
            (
                dictgen::InsensitiveStr::Ascii("coporations"),
                &["corporations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eextraction"),
                &["extraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consisently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreagates"),
                &["aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rezurrection"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistrows"), &["maestros"]),
            (
                dictgen::InsensitiveStr::Ascii("sepetarors"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("contenht"), &["content"]),
            (dictgen::InsensitiveStr::Ascii("weild"), &["wield", "wild"]),
            (dictgen::InsensitiveStr::Ascii("intquire"), &["inquire"]),
            (
                dictgen::InsensitiveStr::Ascii("uniterruptable"),
                &["uninterruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwhelimg"),
                &["underwhelming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difficutly"),
                &["difficulty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publikation"),
                &["publication"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomotive"), &["automotive"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdented"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegetatian"),
                &["vegetarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seqenstial"),
                &["sequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controceptives"),
                &["contraceptives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accociated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represet"),
                &["represents", "represent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("songle"),
                &["single", "dongle"],
            ),
            (dictgen::InsensitiveStr::Ascii("mointor"), &["monitor"]),
            (
                dictgen::InsensitiveStr::Ascii("underlfowed"),
                &["underflowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("veregated"), &["variegated"]),
            (
                dictgen::InsensitiveStr::Ascii("representate"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("consctruct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("refacor"), &["refactor"]),
            (dictgen::InsensitiveStr::Ascii("jamaicaman"), &["jamaican"]),
            (dictgen::InsensitiveStr::Ascii("furthur"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("aparatuses"),
                &["apparatuses"],
            ),
            (dictgen::InsensitiveStr::Ascii("instancs"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("identitier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("breshed"), &["brushed"]),
            (dictgen::InsensitiveStr::Ascii("seqences"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("symbolls"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("algortihm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("powerlifitng"),
                &["powerlifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("greately"), &["greatly"]),
            (
                dictgen::InsensitiveStr::Ascii("captialists"),
                &["capitalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrbute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("atfernoons"),
                &["afternoons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("officeally"),
                &["officially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valiables"),
                &["variables", "valuables"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcued"), &["produced"]),
            (dictgen::InsensitiveStr::Ascii("escapeable"), &["escapable"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpereince"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownwership"), &["ownership"]),
            (
                dictgen::InsensitiveStr::Ascii("confidenciall"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("aswers"), &["answers"]),
            (
                dictgen::InsensitiveStr::Ascii("leprachans"),
                &["leprechauns"],
            ),
            (dictgen::InsensitiveStr::Ascii("commercie"), &["commerce"]),
            (
                dictgen::InsensitiveStr::Ascii("respawining"),
                &["respawning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrispondance"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multinatinal"),
                &["multinational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interate"),
                &["iterate", "integrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constarnation"),
                &["consternation"],
            ),
            (dictgen::InsensitiveStr::Ascii("throughts"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("exchausting"),
                &["exhausting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dargons"), &["dragons"]),
            (dictgen::InsensitiveStr::Ascii("tabualte"), &["tabulate"]),
            (
                dictgen::InsensitiveStr::Ascii("substitures"),
                &["substitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlayzing"), &["analyzing"]),
            (
                dictgen::InsensitiveStr::Ascii("undertanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("parasits"), &["parasites"]),
            (
                dictgen::InsensitiveStr::Ascii("paramatrically"),
                &["parametrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("akkumulates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolutlye"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("remembe"), &["remember"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastructions"),
                &["infrastructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandivanian"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traslators"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("psudocode"), &["pseudocode"]),
            (dictgen::InsensitiveStr::Ascii("chckboxes"), &["checkboxes"]),
            (dictgen::InsensitiveStr::Ascii("heplful"), &["helpful"]),
            (dictgen::InsensitiveStr::Ascii("vrituoso"), &["virtuoso"]),
            (dictgen::InsensitiveStr::Ascii("passoword"), &["password"]),
            (dictgen::InsensitiveStr::Ascii("miticate"), &["mitigate"]),
            (
                dictgen::InsensitiveStr::Ascii("requirmentes"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("peedmonts"), &["piedmonts"]),
            (dictgen::InsensitiveStr::Ascii("severily"), &["severely"]),
            (dictgen::InsensitiveStr::Ascii("matrces"), &["matrices"]),
            (dictgen::InsensitiveStr::Ascii("sasy"), &["says", "sassy"]),
            (
                dictgen::InsensitiveStr::Ascii("combininers"),
                &["combiners"],
            ),
            (dictgen::InsensitiveStr::Ascii("enouth"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("speaces"),
                &["spaces", "species"],
            ),
            (dictgen::InsensitiveStr::Ascii("egisters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("nieghbours"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapte"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("opensll"), &["openssl"]),
            (
                dictgen::InsensitiveStr::Ascii("unconcistent"),
                &["inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("butterlfy"), &["butterfly"]),
            (
                dictgen::InsensitiveStr::Ascii("prepetuating"),
                &["perpetuating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevitibly"),
                &["inevitably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrisponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backet"),
                &["bracket", "basket", "bucket", "backed", "backer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beleved"),
                &["believed", "beloved"],
            ),
            (dictgen::InsensitiveStr::Ascii("ptotocol"), &["protocol"]),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplicities"),
                &["multiplicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationaliy"),
                &["nationality"],
            ),
            (dictgen::InsensitiveStr::Ascii("decypher"), &["decipher"]),
            (
                dictgen::InsensitiveStr::Ascii("authrorising"),
                &["authorising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitting"),
                &["delimiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitabile"),
                &["profitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promimently"),
                &["prominently"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturcts"), &["structs"]),
            (
                dictgen::InsensitiveStr::Ascii("accepectance"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("efective"), &["effective"]),
            (dictgen::InsensitiveStr::Ascii("oustiders"), &["outsiders"]),
            (
                dictgen::InsensitiveStr::Ascii("experimelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribuges"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excpetions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("connetec"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("enforncement"),
                &["enforcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibily"),
                &["possibility", "possibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disposicion"),
                &["disposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thundervolt"),
                &["thunderbolt"],
            ),
            (dictgen::InsensitiveStr::Ascii("submergerd"), &["submerged"]),
            (
                dictgen::InsensitiveStr::Ascii("integracion"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntained"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("genertive"), &["generative"]),
            (
                dictgen::InsensitiveStr::Ascii("invisibiity"),
                &["invisibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispraportionate"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlcluding"), &["excluding"]),
            (dictgen::InsensitiveStr::Ascii("unshfited"), &["unshifted"]),
            (
                dictgen::InsensitiveStr::Ascii("transportarme"),
                &["transporter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temperaturas"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("autosence"), &["autosense"]),
            (dictgen::InsensitiveStr::Ascii("achiev"), &["achieve"]),
            (dictgen::InsensitiveStr::Ascii("eighteeen"), &["eighteen"]),
            (
                dictgen::InsensitiveStr::Ascii("exeedingly"),
                &["exceedingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefitlering"),
                &["prefiltering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrispondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arragement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("broadbad"), &["broadband"]),
            (
                dictgen::InsensitiveStr::Ascii("unoptimzed"),
                &["unoptimized"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimizers"), &["optimizers"]),
            (dictgen::InsensitiveStr::Ascii("streched"), &["stretched"]),
            (
                dictgen::InsensitiveStr::Ascii("symbonname"),
                &["symbolname"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveniant"),
                &["inconvenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosphies"),
                &["philosophies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyplopedia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempararily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("allainces"), &["alliances"]),
            (dictgen::InsensitiveStr::Ascii("refeshes"), &["refreshes"]),
            (dictgen::InsensitiveStr::Ascii("occurded"), &["occurred"]),
            (dictgen::InsensitiveStr::Ascii("inheriet"), &["inherit"]),
            (
                dictgen::InsensitiveStr::Ascii("environnements"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("receivin"), &["receiving"]),
            (
                dictgen::InsensitiveStr::Ascii("previleges"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertinet"), &["pertinent"]),
            (dictgen::InsensitiveStr::Ascii("tringale"), &["triangle"]),
            (
                dictgen::InsensitiveStr::Ascii("pseudononymous"),
                &["pseudonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("groosumly"),
                &["gruesome", "gruesomely"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtiltes"), &["subtitles"]),
            (
                dictgen::InsensitiveStr::Ascii("reuqirements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("resarts"), &["restarts"]),
            (
                dictgen::InsensitiveStr::Ascii("incontrollably"),
                &["uncontrollably"],
            ),
            (dictgen::InsensitiveStr::Ascii("expreince"), &["experience"]),
            (dictgen::InsensitiveStr::Ascii("follewiong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("lippizaner"),
                &["lipizzaner"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapive"), &["adaptive"]),
            (dictgen::InsensitiveStr::Ascii("optomiser"), &["optimiser"]),
            (dictgen::InsensitiveStr::Ascii("lefted"), &["left"]),
            (dictgen::InsensitiveStr::Ascii("organzied"), &["organized"]),
            (dictgen::InsensitiveStr::Ascii("scketched"), &["sketched"]),
            (
                dictgen::InsensitiveStr::Ascii("costructions"),
                &["constructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryies"), &["tries"]),
            (dictgen::InsensitiveStr::Ascii("interwine"), &["intertwine"]),
            (
                dictgen::InsensitiveStr::Ascii("uncommmented"),
                &["uncommented"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtatter"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("habaeus"), &["habeas"]),
            (dictgen::InsensitiveStr::Ascii("inactiv"), &["inactive"]),
            (
                dictgen::InsensitiveStr::Ascii("ackowledged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpectadly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("kubernete"), &["kubernetes"]),
            (dictgen::InsensitiveStr::Ascii("opeinging"), &["opening"]),
            (dictgen::InsensitiveStr::Ascii("becausae"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("threadng"), &["threading"]),
            (
                dictgen::InsensitiveStr::Ascii("sportsmamship"),
                &["sportsmanship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suficiently"),
                &["sufficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atuhentication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("grup"), &["group"]),
            (dictgen::InsensitiveStr::Ascii("functtion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("befoe"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("ecourages"), &["encourages"]),
            (
                dictgen::InsensitiveStr::Ascii("repositiroes"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkflows"), &["workflows"]),
            (
                dictgen::InsensitiveStr::Ascii("noncombatents"),
                &["noncombatants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcultations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priodicals"),
                &["periodicals"],
            ),
            (dictgen::InsensitiveStr::Ascii("enqueing"), &["enqueuing"]),
            (
                dictgen::InsensitiveStr::Ascii("fisical"),
                &["physical", "fiscal"],
            ),
            (dictgen::InsensitiveStr::Ascii("shft"), &["shift"]),
            (
                dictgen::InsensitiveStr::Ascii("collabsible"),
                &["collapsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rittler"),
                &["rattler", "riddler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adolescene"),
                &["adolescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streightness"),
                &["straightness"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparring"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("charctor"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("custume"), &["custom"]),
            (dictgen::InsensitiveStr::Ascii("studetn"), &["student"]),
            (dictgen::InsensitiveStr::Ascii("colouer"), &["colour"]),
            (dictgen::InsensitiveStr::Ascii("similarely"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("spectacullar"),
                &["spectacularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("insteading"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("siliencing"), &["silencing"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhhods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempremental"),
                &["temperamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbraced"), &["embraced"]),
            (dictgen::InsensitiveStr::Ascii("pubblic"), &["public"]),
            (
                dictgen::InsensitiveStr::Ascii("flexibilty"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catgorical"),
                &["categorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illumianti"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstactness"),
                &["abstractness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasformations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("respresent"), &["represent"]),
            (
                dictgen::InsensitiveStr::Ascii("dividor"),
                &["divider", "divisor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behaivours"),
                &["behaviours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preficing"),
                &["prefixing", "prefacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ditionaries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailziations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instalement"),
                &["installment"],
            ),
            (dictgen::InsensitiveStr::Ascii("purhcased"), &["purchased"]),
            (
                dictgen::InsensitiveStr::Ascii("decriptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalleses"), &["coalesces"]),
            (dictgen::InsensitiveStr::Ascii("bethdesa"), &["bethesda"]),
            (dictgen::InsensitiveStr::Ascii("baasis"), &["basis"]),
            (dictgen::InsensitiveStr::Ascii("veru"), &["very"]),
            (
                dictgen::InsensitiveStr::Ascii("introductary"),
                &["introductory"],
            ),
            (dictgen::InsensitiveStr::Ascii("notive"), &["notice"]),
            (dictgen::InsensitiveStr::Ascii("peferably"), &["preferably"]),
            (
                dictgen::InsensitiveStr::Ascii("installataions"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutilsite"), &["multisite"]),
            (dictgen::InsensitiveStr::Ascii("yorksher"), &["yorkshire"]),
            (dictgen::InsensitiveStr::Ascii("exteders"), &["extenders"]),
            (dictgen::InsensitiveStr::Ascii("tyrrany"), &["tyranny"]),
            (
                dictgen::InsensitiveStr::Ascii("establishd"),
                &["established"],
            ),
            (dictgen::InsensitiveStr::Ascii("spesified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("formallizes"),
                &["formalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heartborken"),
                &["heartbroken"],
            ),
            (dictgen::InsensitiveStr::Ascii("follfowed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("sawstika"), &["swastika"]),
            (
                dictgen::InsensitiveStr::Ascii("mediocricy"),
                &["mediocrity"],
            ),
            (dictgen::InsensitiveStr::Ascii("neptun"), &["neptune"]),
            (
                dictgen::InsensitiveStr::Ascii("downrgading"),
                &["downgrading"],
            ),
            (dictgen::InsensitiveStr::Ascii("dasboards"), &["dashboards"]),
            (
                dictgen::InsensitiveStr::Ascii("govermnment"),
                &["government"],
            ),
            (dictgen::InsensitiveStr::Ascii("avisory"), &["advisory"]),
            (dictgen::InsensitiveStr::Ascii("glagsow"), &["glasgow"]),
            (dictgen::InsensitiveStr::Ascii("resmeble"), &["resemble"]),
            (
                dictgen::InsensitiveStr::Ascii("processinf"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("intiator"), &["initiator"]),
            (dictgen::InsensitiveStr::Ascii("metics"), &["metrics"]),
            (
                dictgen::InsensitiveStr::Ascii("achivement"),
                &["achievement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atrocitites"),
                &["atrocities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilitu"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializexd"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("overrrun"), &["overrun"]),
            (
                dictgen::InsensitiveStr::Ascii("tabelspoons"),
                &["tablespoons"],
            ),
            (dictgen::InsensitiveStr::Ascii("rachives"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("translastion"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpeters"),
                &["interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("folderd"),
                &["folders", "folded", "folder"],
            ),
            (dictgen::InsensitiveStr::Ascii("variatnt"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("invunlerable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpolaion"),
                &["interpolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intructors"),
                &["instructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("servicable"),
                &["serviceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("culprints"), &["culprits"]),
            (
                dictgen::InsensitiveStr::Ascii("distingushes"),
                &["distinguishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("populationes"),
                &["populations"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparse"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("workspae"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmiately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exchenging"),
                &["exchanging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perdiction"),
                &["prediction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferenes"),
                &["dereferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("secundly"), &["secondly"]),
            (dictgen::InsensitiveStr::Ascii("exapnds"), &["expands"]),
            (
                dictgen::InsensitiveStr::Ascii("containeres"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrucution"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("harrass"), &["harass"]),
            (dictgen::InsensitiveStr::Ascii("covanent"), &["covenant"]),
            (
                dictgen::InsensitiveStr::Ascii("capababilities"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberalsim"),
                &["liberalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("ideallly"), &["ideally"]),
            (dictgen::InsensitiveStr::Ascii("obstacal"), &["obstacle"]),
            (
                dictgen::InsensitiveStr::Ascii("misrepresantation"),
                &["misrepresentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purchashing"),
                &["purchasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tripel"), &["triple"]),
            (dictgen::InsensitiveStr::Ascii("limition"), &["limitation"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomeance"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wrkstations"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("quried"), &["queried"]),
            (dictgen::InsensitiveStr::Ascii("errorred"), &["errored"]),
            (
                dictgen::InsensitiveStr::Ascii("nutritionnal"),
                &["nutritional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restricteds"),
                &["restricts"],
            ),
            (dictgen::InsensitiveStr::Ascii("updateing"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("deffinetly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commtiment"),
                &["commitments"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggestons"), &["suggests"]),
            (
                dictgen::InsensitiveStr::Ascii("instantiatin"),
                &["instantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("pediod"), &["period"]),
            (dictgen::InsensitiveStr::Ascii("whaever"), &["whatever"]),
            (dictgen::InsensitiveStr::Ascii("zobrish"), &["zobrist"]),
            (
                dictgen::InsensitiveStr::Ascii("percievably"),
                &["perceivably"],
            ),
            (dictgen::InsensitiveStr::Ascii("amaturs"), &["amateurs"]),
            (
                dictgen::InsensitiveStr::Ascii("releationships"),
                &["relationships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semgentation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiator"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excelerates"),
                &["accelerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotes"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commentart"),
                &["commentator"],
            ),
            (dictgen::InsensitiveStr::Ascii("remenant"), &["remnant"]),
            (dictgen::InsensitiveStr::Ascii("intregrals"), &["integrals"]),
            (dictgen::InsensitiveStr::Ascii("regulaer"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("thoug"), &["though"]),
            (dictgen::InsensitiveStr::Ascii("jamiaca"), &["jamaica"]),
            (dictgen::InsensitiveStr::Ascii("veicle"), &["vehicle"]),
            (
                dictgen::InsensitiveStr::Ascii("uncounciously"),
                &["unconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("saveguards"),
                &["safeguards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approproximations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doccumenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condicting"),
                &["conducting", "convicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturggled"), &["struggled"]),
            (dictgen::InsensitiveStr::Ascii("temafight"), &["teamfight"]),
            (
                dictgen::InsensitiveStr::Ascii("encorde"),
                &["encode", "encore"],
            ),
            (dictgen::InsensitiveStr::Ascii("ayways"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("succedd"), &["succeed"]),
            (
                dictgen::InsensitiveStr::Ascii("afghansitan"),
                &["afghanistan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitplicities"),
                &["multiplicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("entquires"), &["enquires"]),
            (dictgen::InsensitiveStr::Ascii("solfed"), &["solved"]),
            (dictgen::InsensitiveStr::Ascii("distanse"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("depandance"),
                &["dependance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequeces"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("suasage"), &["sausage"]),
            (dictgen::InsensitiveStr::Ascii("superbock"), &["superblock"]),
            (
                dictgen::InsensitiveStr::Ascii("placedholder"),
                &["placeholder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilty"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pennsylvanica"),
                &["pennsylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aesthestically"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclockd"),
                &["overclocked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilizing"),
                &["capitalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpecting"), &["inspecting"]),
            (dictgen::InsensitiveStr::Ascii("nulll"), &["null"]),
            (dictgen::InsensitiveStr::Ascii("underying"), &["underlying"]),
            (dictgen::InsensitiveStr::Ascii("delievered"), &["delivered"]),
            (dictgen::InsensitiveStr::Ascii("nostolgic"), &["nostalgic"]),
            (dictgen::InsensitiveStr::Ascii("minimzing"), &["minimizing"]),
            (dictgen::InsensitiveStr::Ascii("resuilted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeckerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expanssion"),
                &["expansions", "expansion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controdictions"),
                &["contradictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("vavle"), &["valve"]),
            (dictgen::InsensitiveStr::Ascii("lasagnea"), &["lasagna"]),
            (dictgen::InsensitiveStr::Ascii("inintial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("abdonimal"), &["abdominal"]),
            (dictgen::InsensitiveStr::Ascii("langugage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("somehing"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("identidy"),
                &["identify", "identity"],
            ),
            (dictgen::InsensitiveStr::Ascii("spindel"), &["spindle"]),
            (dictgen::InsensitiveStr::Ascii("bounardy"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("andriod"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("garenteed"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("uninterpretted"),
                &["uninterpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("switcn"), &["switch"]),
            (dictgen::InsensitiveStr::Ascii("ofcrouse"), &["ofcourse"]),
            (
                dictgen::InsensitiveStr::Ascii("imperialisim"),
                &["imperialism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missplacing"),
                &["misplacing"],
            ),
            (dictgen::InsensitiveStr::Ascii("denilas"), &["denials"]),
            (dictgen::InsensitiveStr::Ascii("tnanked"), &["thanked"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeurhoors"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("addopted"), &["adopted"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpreted"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzealos"),
                &["overzealous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritied"),
                &["prioritize"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsued"), &["unused"]),
            (
                dictgen::InsensitiveStr::Ascii("configuraions"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosnsrained"),
                &["constrained"],
            ),
            (dictgen::InsensitiveStr::Ascii("donejun"), &["dungeon"]),
            (dictgen::InsensitiveStr::Ascii("nigthmare"), &["nightmare"]),
            (dictgen::InsensitiveStr::Ascii("chruch"), &["church"]),
            (
                dictgen::InsensitiveStr::Ascii("constribute"),
                &["contribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaited"), &["animated"]),
            (
                dictgen::InsensitiveStr::Ascii("intervines"),
                &["intervenes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprecating"),
                &["deprecating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurtations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("connction"), &["connection"]),
            (dictgen::InsensitiveStr::Ascii("anihilate"), &["annihilate"]),
            (dictgen::InsensitiveStr::Ascii("chernobly"), &["chernobyl"]),
            (dictgen::InsensitiveStr::Ascii("collisson"), &["collisions"]),
            (
                dictgen::InsensitiveStr::Ascii("correpondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("jonatahn"), &["jonathan"]),
            (dictgen::InsensitiveStr::Ascii("probabely"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("shotdown"), &["shutdown"]),
            (dictgen::InsensitiveStr::Ascii("ultiamte"), &["ultimate"]),
            (dictgen::InsensitiveStr::Ascii("austeer"), &["austere"]),
            (dictgen::InsensitiveStr::Ascii("pulisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("discontued"),
                &["discounted"],
            ),
            (dictgen::InsensitiveStr::Ascii("comverted"), &["converted"]),
            (
                dictgen::InsensitiveStr::Ascii("territoires"),
                &["territories"],
            ),
            (dictgen::InsensitiveStr::Ascii("udpates"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("versiobs"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("aggreecate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("execubale"), &["executable"]),
            (dictgen::InsensitiveStr::Ascii("defore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("asymetric"), &["asymmetric"]),
            (
                dictgen::InsensitiveStr::Ascii("overshadowered"),
                &["overshadowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("apaptive"), &["adaptive"]),
            (
                dictgen::InsensitiveStr::Ascii("implicitily"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("detatching"), &["detaching"]),
            (dictgen::InsensitiveStr::Ascii("anteenas"), &["antennas"]),
            (
                dictgen::InsensitiveStr::Ascii("constracts"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("designad"), &["designated"]),
            (dictgen::InsensitiveStr::Ascii("besidies"), &["besides"]),
            (dictgen::InsensitiveStr::Ascii("committted"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("tressle"), &["trestle"]),
            (
                dictgen::InsensitiveStr::Ascii("completeted"),
                &["completed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("themsleves"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("ascpect"), &["aspect"]),
            (dictgen::InsensitiveStr::Ascii("milicious"), &["malicious"]),
            (
                dictgen::InsensitiveStr::Ascii("knowledgble"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replentishing"),
                &["replenishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("webage"), &["webpage"]),
            (dictgen::InsensitiveStr::Ascii("inidicate"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("successfulyl"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtractness"),
                &["abstractness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devaulting"),
                &["defaulting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doubleiift"),
                &["doublelift"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configureble"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("expidite"), &["expedite"]),
            (
                dictgen::InsensitiveStr::Ascii("androgenous"),
                &["androgynous"],
            ),
            (dictgen::InsensitiveStr::Ascii("virignity"), &["virginity"]),
            (
                dictgen::InsensitiveStr::Ascii("traditition"),
                &["tradition"],
            ),
            (dictgen::InsensitiveStr::Ascii("slibling"), &["sibling"]),
            (
                dictgen::InsensitiveStr::Ascii("speculatie"),
                &["speculative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scnadinavian"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("beweeen"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("underminig"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versatality"),
                &["versatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("innoculate"), &["inoculate"]),
            (
                dictgen::InsensitiveStr::Ascii("effectivley"),
                &["effectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("handskake"), &["handshake"]),
            (
                dictgen::InsensitiveStr::Ascii("subrutines"),
                &["subroutines"],
            ),
            (dictgen::InsensitiveStr::Ascii("regionaal"), &["regional"]),
            (
                dictgen::InsensitiveStr::Ascii("busineses"),
                &["business", "businesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("distncte"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("alternetives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("playge"), &["plague"]),
            (
                dictgen::InsensitiveStr::Ascii("duplciates"),
                &["duplicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilaziaton"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregetors"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("octohedral"),
                &["octahedral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repressent"),
                &["represents", "represent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senzationalism"),
                &["sensationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inhibating"),
                &["inhibiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("sizre"), &["size"]),
            (
                dictgen::InsensitiveStr::Ascii("experimment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinats"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("corsiar"), &["corsair"]),
            (
                dictgen::InsensitiveStr::Ascii("mountpiont"),
                &["mountpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("resistendo"), &["resisted"]),
            (
                dictgen::InsensitiveStr::Ascii("manafactures"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fution"),
                &["function", "futon"],
            ),
            (dictgen::InsensitiveStr::Ascii("visable"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("constrant"), &["constraint"]),
            (
                dictgen::InsensitiveStr::Ascii("docomented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coloered"),
                &["colored", "coloured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depandences"),
                &["dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circuvented"),
                &["circumvented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiansial"),
                &["experiential"],
            ),
            (dictgen::InsensitiveStr::Ascii("allianses"), &["alliances"]),
            (
                dictgen::InsensitiveStr::Ascii("accress"),
                &["access", "address"],
            ),
            (dictgen::InsensitiveStr::Ascii("depndancy"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("abnd"), &["and", "band"]),
            (dictgen::InsensitiveStr::Ascii("deteremine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("checcklists"),
                &["checklists"],
            ),
            (dictgen::InsensitiveStr::Ascii("problim"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("stoage"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("studi"),
                &["study", "studio"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flawlessely"),
                &["flawlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("inquriy"), &["inquiry"]),
            (dictgen::InsensitiveStr::Ascii("larrry"), &["larry"]),
            (dictgen::InsensitiveStr::Ascii("sepeicfy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("mainained"), &["maintained"]),
            (
                dictgen::InsensitiveStr::Ascii("equivallent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seeting"),
                &["seating", "setting", "seething"],
            ),
            (dictgen::InsensitiveStr::Ascii("comressed"), &["compressed"]),
            (
                dictgen::InsensitiveStr::Ascii("spefixations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("secton"), &["section"]),
            (
                dictgen::InsensitiveStr::Ascii("specatotors"),
                &["spectators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identfication"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("stablity"), &["stability"]),
            (
                dictgen::InsensitiveStr::Ascii("appreicates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siffixation"),
                &["suffixation", "suffocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("motivatie"), &["motivate"]),
            (dictgen::InsensitiveStr::Ascii("allocatedp"), &["allocated"]),
            (dictgen::InsensitiveStr::Ascii("compund"), &["compound"]),
            (
                dictgen::InsensitiveStr::Ascii("facilitant"),
                &["facilitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrappper"), &["wrapper"]),
            (
                dictgen::InsensitiveStr::Ascii("fluctuatin"),
                &["fluctuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("barbarions"),
                &["barbarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdenncies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("redidual"), &["residual"]),
            (dictgen::InsensitiveStr::Ascii("alogning"), &["aligning"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstraby"),
                &["demonstrably"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchngng"), &["exchanging"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificin"),
                &["sacrificing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teridactyl"),
                &["pterodactyl"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrpyting"),
                &["encrypting"],
            ),
            (dictgen::InsensitiveStr::Ascii("expading"), &["expanding"]),
            (
                dictgen::InsensitiveStr::Ascii("increamenting"),
                &["incrementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontaenously"),
                &["spontaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("comfig"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenatations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgerade"), &["upgrade"]),
            (dictgen::InsensitiveStr::Ascii("areguments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("expctedly"), &["expectedly"]),
            (dictgen::InsensitiveStr::Ascii("graudates"), &["graduates"]),
            (dictgen::InsensitiveStr::Ascii("decalaring"), &["declaring"]),
            (
                dictgen::InsensitiveStr::Ascii("multidimension"),
                &["multidimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("conseedd"), &["conceded"]),
            (dictgen::InsensitiveStr::Ascii("conerters"), &["converters"]),
            (dictgen::InsensitiveStr::Ascii("coalased"), &["coalesced"]),
            (dictgen::InsensitiveStr::Ascii("leesures"), &["leisures"]),
            (dictgen::InsensitiveStr::Ascii("advocat"), &["advocate"]),
            (dictgen::InsensitiveStr::Ascii("montajes"), &["montages"]),
            (dictgen::InsensitiveStr::Ascii("ryhthm"), &["rhythm"]),
            (dictgen::InsensitiveStr::Ascii("assicates"), &["associates"]),
            (dictgen::InsensitiveStr::Ascii("agaist"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("paranthes"),
                &["parentheses", "parenthesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("assoiciate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("bais"),
                &["bias", "basis", "bails", "baits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("familiarty"),
                &["familiarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("linguisticos"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("opacy"), &["opacity"]),
            (
                dictgen::InsensitiveStr::Ascii("pilgramige"),
                &["pilgrimage"],
            ),
            (dictgen::InsensitiveStr::Ascii("precesses"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("bureuacratic"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("alorithms"), &["algorithms"]),
            (
                dictgen::InsensitiveStr::Ascii("devide"),
                &["divide", "device"],
            ),
            (dictgen::InsensitiveStr::Ascii("applaude"), &["applaud"]),
            (dictgen::InsensitiveStr::Ascii("flacours"), &["flavours"]),
            (
                dictgen::InsensitiveStr::Ascii("continuosly"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participanting"),
                &["participating"],
            ),
            (dictgen::InsensitiveStr::Ascii("reutrn"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("nner"), &["inner"]),
            (
                dictgen::InsensitiveStr::Ascii("colerscheme"),
                &["colorscheme"],
            ),
            (dictgen::InsensitiveStr::Ascii("flacons"), &["falcons"]),
            (
                dictgen::InsensitiveStr::Ascii("mutimarked"),
                &["multimarked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dumbfouded"),
                &["dumbfounded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalizer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constractions"),
                &["constrictions", "constructions", "contractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("econimically"),
                &["economically"],
            ),
            (dictgen::InsensitiveStr::Ascii("composant"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpacted"), &["impacted"]),
            (dictgen::InsensitiveStr::Ascii("acquited"), &["acquitted"]),
            (
                dictgen::InsensitiveStr::Ascii("probabiliste"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistendtly"),
                &["inconsistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerization"),
                &["parameterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improtantly"),
                &["importantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("reaonably"), &["reasonably"]),
            (
                dictgen::InsensitiveStr::Ascii("programmmed"),
                &["programme"],
            ),
            (dictgen::InsensitiveStr::Ascii("porduct"), &["product"]),
            (dictgen::InsensitiveStr::Ascii("rorating"), &["rotating"]),
            (
                dictgen::InsensitiveStr::Ascii("microstansactions"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpeected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconcequential"),
                &["inconsequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exploitaiton"),
                &["exploitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unabel"), &["unable"]),
            (
                dictgen::InsensitiveStr::Ascii("unpriveleged"),
                &["unprivileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("appllies"), &["applies"]),
            (dictgen::InsensitiveStr::Ascii("dimisses"), &["dismisses"]),
            (
                dictgen::InsensitiveStr::Ascii("phlem"),
                &["phlegm", "phloem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anniversry"),
                &["anniversary"],
            ),
            (dictgen::InsensitiveStr::Ascii("mehtods"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guantanameow"),
                &["guantanamo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapprovel"),
                &["disapproval"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validading"),
                &["validating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mandrain"), &["mandarin"]),
            (
                dictgen::InsensitiveStr::Ascii("unncessarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("detecttive"), &["detective"]),
            (dictgen::InsensitiveStr::Ascii("righetous"), &["righteous"]),
            (
                dictgen::InsensitiveStr::Ascii("unpredictible"),
                &["unpredictable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intidimate"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepresents"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("encocde"), &["encode"]),
            (dictgen::InsensitiveStr::Ascii("commetn"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("takeing"), &["taking"]),
            (
                dictgen::InsensitiveStr::Ascii("encompesses"),
                &["encompasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declatations"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocompete"),
                &["autocomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("bundlin"), &["bundling"]),
            (
                dictgen::InsensitiveStr::Ascii("trasparency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onmipotent"),
                &["omnipotent"],
            ),
            (dictgen::InsensitiveStr::Ascii("keywod"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("serches"), &["searches"]),
            (
                dictgen::InsensitiveStr::Ascii("representatie"),
                &["representatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplei"), &["simply"]),
            (
                dictgen::InsensitiveStr::Ascii("atttributes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitututed"),
                &["constituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("scultpure"), &["sculpture"]),
            (dictgen::InsensitiveStr::Ascii("absovle"), &["absolve"]),
            (dictgen::InsensitiveStr::Ascii("sitl"), &["still"]),
            (
                dictgen::InsensitiveStr::Ascii("requireds"),
                &["required", "requires"],
            ),
            (dictgen::InsensitiveStr::Ascii("forutunate"), &["fortunate"]),
            (
                dictgen::InsensitiveStr::Ascii("ovrridables"),
                &["overridables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assfalting"),
                &["asphalting"],
            ),
            (dictgen::InsensitiveStr::Ascii("klick"), &["click"]),
            (
                dictgen::InsensitiveStr::Ascii("straighted"),
                &["straightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("decdoing"), &["decoding"]),
            (
                dictgen::InsensitiveStr::Ascii("ridicously"),
                &["ridiculously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referenceas"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("dropabel"), &["droppable"]),
            (
                dictgen::InsensitiveStr::Ascii("kubernates"),
                &["kubernetes"],
            ),
            (dictgen::InsensitiveStr::Ascii("discused"), &["discussed"]),
            (
                dictgen::InsensitiveStr::Ascii("extrenally"),
                &["externally"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfset"), &["offset"]),
            (dictgen::InsensitiveStr::Ascii("upadters"), &["updaters"]),
            (
                dictgen::InsensitiveStr::Ascii("cnfiguring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitition"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estiamating"),
                &["estimating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instalations"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("brosers"), &["browsers"]),
            (dictgen::InsensitiveStr::Ascii("patameter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("deferenced"),
                &["dereference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcitons"),
                &["directions"],
            ),
            (dictgen::InsensitiveStr::Ascii("modeuls"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("cannt"), &["cannot"]),
            (dictgen::InsensitiveStr::Ascii("adapers"), &["adapters"]),
            (
                dictgen::InsensitiveStr::Ascii("changee"),
                &["change", "changed", "changes"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppsing"), &["supposing"]),
            (
                dictgen::InsensitiveStr::Ascii("paralelises"),
                &["parallelises"],
            ),
            (dictgen::InsensitiveStr::Ascii("variabbles"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("scetched"), &["sketched"]),
            (dictgen::InsensitiveStr::Ascii("gradute"), &["graduate"]),
            (dictgen::InsensitiveStr::Ascii("launchered"), &["launched"]),
            (
                dictgen::InsensitiveStr::Ascii("charaterize"),
                &["characterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("occourred"), &["occurred"]),
            (
                dictgen::InsensitiveStr::Ascii("invocaition"),
                &["invocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discouranged"),
                &["discouraged"],
            ),
            (dictgen::InsensitiveStr::Ascii("possessers"), &["possesses"]),
            (dictgen::InsensitiveStr::Ascii("upsteams"), &["upstreams"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("datebases"), &["databases"]),
            (
                dictgen::InsensitiveStr::Ascii("impeccible"),
                &["impeccable"],
            ),
            (dictgen::InsensitiveStr::Ascii("contestas"), &["contests"]),
            (dictgen::InsensitiveStr::Ascii("renderesd"), &["rendered"]),
            (
                dictgen::InsensitiveStr::Ascii("installaters"),
                &["installers"],
            ),
            (dictgen::InsensitiveStr::Ascii("firmeare"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("clicable"), &["clickable"]),
            (
                dictgen::InsensitiveStr::Ascii("fo"),
                &["of", "for", "do", "go", "to"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incosistent"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excetional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("deciedes"), &["decides"]),
            (
                dictgen::InsensitiveStr::Ascii("anniversy"),
                &["anniversary"],
            ),
            (dictgen::InsensitiveStr::Ascii("singature"), &["signature"]),
            (dictgen::InsensitiveStr::Ascii("deficeint"), &["deficient"]),
            (dictgen::InsensitiveStr::Ascii("yhou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("encyrpted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("continunity"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lacklaster"),
                &["lackluster"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsusually"), &["unusually"]),
            (
                dictgen::InsensitiveStr::Ascii("alernatively"),
                &["alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimlarly"),
                &["dissimilarly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcultuur"),
                &["subculture"],
            ),
            (dictgen::InsensitiveStr::Ascii("otpimism"), &["optimism"]),
            (
                dictgen::InsensitiveStr::Ascii("physioligical"),
                &["physiological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elegible"),
                &["eligible", "illegible", "legible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ducumenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competion"),
                &["competition", "completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("convergin"), &["converging"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessfually"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("extepect"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("individauls"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("contentn"), &["content"]),
            (
                dictgen::InsensitiveStr::Ascii("freqiently"),
                &["frequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("permisson"), &["permission"]),
            (dictgen::InsensitiveStr::Ascii("sumodules"), &["submodules"]),
            (dictgen::InsensitiveStr::Ascii("clutsers"), &["clusters"]),
            (
                dictgen::InsensitiveStr::Ascii("spesifiers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("volunatry"), &["voluntary"]),
            (dictgen::InsensitiveStr::Ascii("boads"), &["boards"]),
            (dictgen::InsensitiveStr::Ascii("absuers"), &["abusers"]),
            (
                dictgen::InsensitiveStr::Ascii("foreigness"),
                &["foreigners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colleced"),
                &["collected", "coalesced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consecurive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("thier"), &["their"]),
            (dictgen::InsensitiveStr::Ascii("enhace"), &["enhance"]),
            (
                dictgen::InsensitiveStr::Ascii("deutschalnd"),
                &["deutschland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increassing"),
                &["increasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreperure"),
                &["entrepreneur"],
            ),
            (dictgen::InsensitiveStr::Ascii("hda"), &["had"]),
            (dictgen::InsensitiveStr::Ascii("schoold"), &["schooled"]),
            (dictgen::InsensitiveStr::Ascii("mimmicks"), &["mimics"]),
            (dictgen::InsensitiveStr::Ascii("dijstra"), &["dijkstra"]),
            (dictgen::InsensitiveStr::Ascii("relcutant"), &["reluctant"]),
            (
                dictgen::InsensitiveStr::Ascii("disributors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrolations"),
                &["correlations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comptetion"),
                &["completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("symmetri"), &["symmetry"]),
            (dictgen::InsensitiveStr::Ascii("primitiv"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("toloerances"),
                &["tolerances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effortlesly"),
                &["effortlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("silabuses"), &["syllabuses"]),
            (
                dictgen::InsensitiveStr::Ascii("linguistcs"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("payed"), &["paid"]),
            (dictgen::InsensitiveStr::Ascii("conaining"), &["containing"]),
            (
                dictgen::InsensitiveStr::Ascii("graphiclly"),
                &["graphically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("histroical"),
                &["historical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transatctions"),
                &["transactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("aledges"), &["alleges"]),
            (
                dictgen::InsensitiveStr::Ascii("correctings"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("corecting"), &["correcting"]),
            (
                dictgen::InsensitiveStr::Ascii("confererate"),
                &["confederate"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefixing"), &["prefixing"]),
            (
                dictgen::InsensitiveStr::Ascii("installiing"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanpu"), &["cleanup"]),
            (
                dictgen::InsensitiveStr::Ascii("incurruptable"),
                &["incorruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucessuflly"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtting"), &["formatting"]),
            (dictgen::InsensitiveStr::Ascii("protecs"), &["protects"]),
            (
                dictgen::InsensitiveStr::Ascii("derivitave"),
                &["derivative"],
            ),
            (dictgen::InsensitiveStr::Ascii("unil"), &["until"]),
            (
                dictgen::InsensitiveStr::Ascii("apparrelling"),
                &["apparelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("stopry"), &["story"]),
            (
                dictgen::InsensitiveStr::Ascii("disproprotionate"),
                &["disproportionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnantees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispayes"), &["displays"]),
            (
                dictgen::InsensitiveStr::Ascii("pennensular"),
                &["peninsular"],
            ),
            (dictgen::InsensitiveStr::Ascii("tigthen"), &["tighten"]),
            (
                dictgen::InsensitiveStr::Ascii("cosntraint"),
                &["constraint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstantual"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissipatin"),
                &["dissipating", "dissipation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dungeong"), &["dungeon"]),
            (dictgen::InsensitiveStr::Ascii("colletors"), &["collectors"]),
            (dictgen::InsensitiveStr::Ascii("neesting"), &["nesting"]),
            (
                dictgen::InsensitiveStr::Ascii("priod"),
                &["period", "prior", "prod", "pried"],
            ),
            (dictgen::InsensitiveStr::Ascii("numbring"), &["numbering"]),
            (
                dictgen::InsensitiveStr::Ascii("subscibers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("aremnians"), &["armenians"]),
            (
                dictgen::InsensitiveStr::Ascii("decsription"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("interrup"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("continens"), &["continents"]),
            (
                dictgen::InsensitiveStr::Ascii("nominativo"),
                &["nomination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manupulate"),
                &["manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminuti"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("preciselly"), &["precisely"]),
            (
                dictgen::InsensitiveStr::Ascii("stabiliting"),
                &["stabilizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaiables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("anothers"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("impossibul"),
                &["impossibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denumenator"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("adultey"), &["adultery"]),
            (
                dictgen::InsensitiveStr::Ascii("accumualtion"),
                &["accumulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upladad"),
                &["upload", "uploaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("libraies"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabily"),
                &["vulnerably", "vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfomed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("groupt"), &["grouped"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowledgemnt"),
                &["acknowledgement"],
            ),
            (dictgen::InsensitiveStr::Ascii("configute"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("gravitatiei"),
                &["gravitate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declarater"),
                &["declarator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechaniclly"),
                &["mechanically"],
            ),
            (dictgen::InsensitiveStr::Ascii("aldutery"), &["adultery"]),
            (
                dictgen::InsensitiveStr::Ascii("corospondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autorizations"),
                &["authorizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("explit"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("consicous"), &["conscious"]),
            (dictgen::InsensitiveStr::Ascii("eather"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("continueously"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallization"),
                &["parallelization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpatrently"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("resgister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("dprecating"),
                &["deprecating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangments"),
                &["rearrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("eraised"), &["erased"]),
            (dictgen::InsensitiveStr::Ascii("reqister"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("sepcific"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("pennal"), &["panel"]),
            (
                dictgen::InsensitiveStr::Ascii("cuves"),
                &["curves", "cubes", "caves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woker"),
                &["worker", "waker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componet"),
                &["components", "component"],
            ),
            (dictgen::InsensitiveStr::Ascii("apapter"), &["adapter"]),
            (
                dictgen::InsensitiveStr::Ascii("nontheless"),
                &["nonetheless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacting"),
                &["replacing", "replicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalites"),
                &["nationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loiusville"),
                &["louisville"],
            ),
            (dictgen::InsensitiveStr::Ascii("clude"), &["clued", "clue"]),
            (
                dictgen::InsensitiveStr::Ascii("tramuatized"),
                &["traumatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("variales"),
                &["variables", "variates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conteiners"),
                &["contenders", "containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("primive"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("unreqested"),
                &["unrequested"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconncts"), &["reconnects"]),
            (dictgen::InsensitiveStr::Ascii("suporters"), &["supporters"]),
            (
                dictgen::InsensitiveStr::Ascii("disinteresed"),
                &["disinterested"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapaits"), &["dissipates"]),
            (dictgen::InsensitiveStr::Ascii("withoud"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("choclate"), &["chocolate"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowladgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramlization"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("currens"), &["currents"]),
            (dictgen::InsensitiveStr::Ascii("instsalls"), &["installs"]),
            (
                dictgen::InsensitiveStr::Ascii("spcification"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatenes"),
                &["concatenates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticallty"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accrodingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("apsotle"), &["apostle"]),
            (
                dictgen::InsensitiveStr::Ascii("complextions"),
                &["complexions"],
            ),
            (dictgen::InsensitiveStr::Ascii("architecht"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("fter"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuccessfuly"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnts"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("infantis"), &["infants"]),
            (dictgen::InsensitiveStr::Ascii("brussles"), &["brussels"]),
            (dictgen::InsensitiveStr::Ascii("crystsl"), &["crystals"]),
            (dictgen::InsensitiveStr::Ascii("jaguares"), &["jaguars"]),
            (dictgen::InsensitiveStr::Ascii("ploygonal"), &["polygonal"]),
            (
                dictgen::InsensitiveStr::Ascii("atributing"),
                &["attributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remanin"),
                &["remain", "remaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilplying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("polishuset"), &["polishes"]),
            (
                dictgen::InsensitiveStr::Ascii("negogtiator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("arount"), &["around"]),
            (
                dictgen::InsensitiveStr::Ascii("reviewsectio"),
                &["reviewsection"],
            ),
            (dictgen::InsensitiveStr::Ascii("alyers"), &["layers"]),
            (
                dictgen::InsensitiveStr::Ascii("trilinar"),
                &["trilinear", "trillionaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlains"), &["explains"]),
            (
                dictgen::InsensitiveStr::Ascii("ilumination"),
                &["illumination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prinicpals"),
                &["principals"],
            ),
            (dictgen::InsensitiveStr::Ascii("personnels"), &["personnel"]),
            (
                dictgen::InsensitiveStr::Ascii("reuqirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singoolarly"),
                &["singularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbaited"),
                &["masturbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("instatance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("cylcone"), &["cyclone"]),
            (
                dictgen::InsensitiveStr::Ascii("intervenion"),
                &["intervening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tacitcally"),
                &["tactically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitions"),
                &["limitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypotethical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("qwirky"), &["quirky"]),
            (
                dictgen::InsensitiveStr::Ascii("reolved"),
                &["resolved", "revolved"],
            ),
            (dictgen::InsensitiveStr::Ascii("trottin"), &["trotting"]),
            (dictgen::InsensitiveStr::Ascii("seleciton"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("stabalization"),
                &["stabilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decendants"),
                &["descendants"],
            ),
            (dictgen::InsensitiveStr::Ascii("underdong"), &["underdog"]),
            (
                dictgen::InsensitiveStr::Ascii("explotacion"),
                &["exploration"],
            ),
            (dictgen::InsensitiveStr::Ascii("updadates"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("rror"), &["error"]),
            (
                dictgen::InsensitiveStr::Ascii("legistators"),
                &["legislators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fertizilers"),
                &["fertilizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknkown"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("equiptment"), &["equipment"]),
            (
                dictgen::InsensitiveStr::Ascii("tesellated"),
                &["tessellated"],
            ),
            (dictgen::InsensitiveStr::Ascii("willaims"), &["williams"]),
            (
                dictgen::InsensitiveStr::Ascii("minipulative"),
                &["manipulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cariciture"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistribuito"),
                &["redistribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcasts"), &["forecasts"]),
            (
                dictgen::InsensitiveStr::Ascii("dependancey"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("usrer"),
                &["user", "surer", "usher", "usurer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("furtehr"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("almoast"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("unnesesarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("eaisest"), &["easiest"]),
            (
                dictgen::InsensitiveStr::Ascii("proceddings"),
                &["proceedings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interships"),
                &["internships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcontructors"),
                &["subconstructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vistors"),
                &["visitors", "visors", "victors"],
            ),
            (dictgen::InsensitiveStr::Ascii("pucini"), &["puccini"]),
            (
                dictgen::InsensitiveStr::Ascii("specfiications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("fobids"), &["forbids"]),
            (dictgen::InsensitiveStr::Ascii("aggergate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("ballancers"), &["balancers"]),
            (
                dictgen::InsensitiveStr::Ascii("paralleising"),
                &["parallelising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxiated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("driagramming"),
                &["diagramming"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemenal"), &["elemental"]),
            (dictgen::InsensitiveStr::Ascii("familar"), &["familiar"]),
            (dictgen::InsensitiveStr::Ascii("previal"), &["prevail"]),
            (dictgen::InsensitiveStr::Ascii("cusstomize"), &["customize"]),
            (
                dictgen::InsensitiveStr::Ascii("configation"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("slotable"), &["slottable"]),
            (dictgen::InsensitiveStr::Ascii("dynamitage"), &["dynamite"]),
            (
                dictgen::InsensitiveStr::Ascii("completily"),
                &["complexity"],
            ),
            (dictgen::InsensitiveStr::Ascii("inventios"), &["inventions"]),
            (
                dictgen::InsensitiveStr::Ascii("placeemnts"),
                &["placements"],
            ),
            (dictgen::InsensitiveStr::Ascii("folower"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("popluar"), &["popular"]),
            (
                dictgen::InsensitiveStr::Ascii("humanitarna"),
                &["humanitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouptput"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("definin"), &["defining"]),
            (dictgen::InsensitiveStr::Ascii("relesasing"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("singnalled"), &["signalled"]),
            (dictgen::InsensitiveStr::Ascii("paramert"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("arhivers"), &["archivers"]),
            (dictgen::InsensitiveStr::Ascii("obtions"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("clsoer"), &["closer"]),
            (dictgen::InsensitiveStr::Ascii("squattin"), &["squatting"]),
            (
                dictgen::InsensitiveStr::Ascii("acquaintinces"),
                &["acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("objcting"), &["objecting"]),
            (dictgen::InsensitiveStr::Ascii("deattaches"), &["detaches"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubition"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embendings"),
                &["embeddings"],
            ),
            (dictgen::InsensitiveStr::Ascii("chercksum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("misundarstanding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawin"), &["drawing"]),
            (dictgen::InsensitiveStr::Ascii("mustator"), &["mutator"]),
            (
                dictgen::InsensitiveStr::Ascii("incrfemental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicatbility"),
                &["predictability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mangae"),
                &["manage", "manga"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchamrks"),
                &["benchmarks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justificaiton"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcationally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventerous"),
                &["adventurous"],
            ),
            (dictgen::InsensitiveStr::Ascii("eleate"), &["relate"]),
            (dictgen::InsensitiveStr::Ascii("meerly"), &["merely"]),
            (
                dictgen::InsensitiveStr::Ascii("definative"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("amny"), &["many"]),
            (dictgen::InsensitiveStr::Ascii("jpin"), &["join"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapates"),
                &["dissipates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straightden"),
                &["straightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("naviagting"),
                &["navigating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expidentures"),
                &["expenditures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neadless"),
                &["needless", "needles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplificaitons"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfroms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("wioth"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("clitiros"), &["clitoris"]),
            (
                dictgen::InsensitiveStr::Ascii("retangular"),
                &["rectangular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demostrating"),
                &["demonstrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutsomer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("reconneted"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consentiously"),
                &["conscientiously"],
            ),
            (dictgen::InsensitiveStr::Ascii("oakerous"), &["ocherous"]),
            (dictgen::InsensitiveStr::Ascii("orginated"), &["originated"]),
            (dictgen::InsensitiveStr::Ascii("intimaty"), &["intimately"]),
            (
                dictgen::InsensitiveStr::Ascii("pronography"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depandency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrestination"),
                &["procrastination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pojrectors"),
                &["projectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarate"),
                &["ameliorate"],
            ),
            (dictgen::InsensitiveStr::Ascii("convecx"), &["convex"]),
            (dictgen::InsensitiveStr::Ascii("reguator"), &["regulator"]),
            (
                dictgen::InsensitiveStr::Ascii("cercumstancial"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctroyers"),
                &["destroyers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acclerators"),
                &["accelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("deadlfits"), &["deadlifts"]),
            (dictgen::InsensitiveStr::Ascii("qurey"), &["query"]),
            (dictgen::InsensitiveStr::Ascii("promisses"), &["promises"]),
            (dictgen::InsensitiveStr::Ascii("pehaps"), &["perhaps"]),
            (
                dictgen::InsensitiveStr::Ascii("peventions"),
                &["preventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("engish"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("liveprool"), &["liverpool"]),
            (
                dictgen::InsensitiveStr::Ascii("exepriment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceots"),
                &["excepts", "exempts"],
            ),
            (dictgen::InsensitiveStr::Ascii("speeeding"), &["speeding"]),
            (
                dictgen::InsensitiveStr::Ascii("sumbissive"),
                &["submissive"],
            ),
            (dictgen::InsensitiveStr::Ascii("releaved"), &["relieved"]),
            (
                dictgen::InsensitiveStr::Ascii("barbariens"),
                &["barbarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("diverison"), &["diversion"]),
            (dictgen::InsensitiveStr::Ascii("exploites"), &["exploits"]),
            (dictgen::InsensitiveStr::Ascii("naemspace"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("iniside"), &["inside"]),
            (
                dictgen::InsensitiveStr::Ascii("expetiment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitlty"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("anotate"), &["annotate"]),
            (dictgen::InsensitiveStr::Ascii("uncoments"), &["uncomments"]),
            (dictgen::InsensitiveStr::Ascii("strecthes"), &["stretches"]),
            (
                dictgen::InsensitiveStr::Ascii("apocolapse"),
                &["apocalypse"],
            ),
            (dictgen::InsensitiveStr::Ascii("invarders"), &["invaders"]),
            (dictgen::InsensitiveStr::Ascii("repport"), &["report"]),
            (
                dictgen::InsensitiveStr::Ascii("charactristically"),
                &["characteristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compartors"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deambigation"),
                &["disambiguation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhilations"),
                &["annihilations"],
            ),
            (dictgen::InsensitiveStr::Ascii("microsot"), &["microsoft"]),
            (dictgen::InsensitiveStr::Ascii("forgetten"), &["forgotten"]),
            (
                dictgen::InsensitiveStr::Ascii("explotiation"),
                &["exploitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectivy"),
                &["connectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partitition"),
                &["partition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inidication"),
                &["indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interaktions"),
                &["interactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumulaton"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("depndant"), &["dependant"]),
            (dictgen::InsensitiveStr::Ascii("neckbears"), &["neckbeards"]),
            (
                dictgen::InsensitiveStr::Ascii("recursevily"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatter"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("whicih"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("motiviation"),
                &["motivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreval"), &["retrieval"]),
            (dictgen::InsensitiveStr::Ascii("analsyes"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("athelticism"),
                &["athleticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promotinal"),
                &["promotional"],
            ),
            (dictgen::InsensitiveStr::Ascii("techiques"), &["techniques"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalistiskt"),
                &["nationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("horriblely"), &["horribly"]),
            (dictgen::InsensitiveStr::Ascii("sirectly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("seranomys"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("absoluately"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionaltiy"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("bufer"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("fequency"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("namespae"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("rertiever"), &["retriever"]),
            (
                dictgen::InsensitiveStr::Ascii("insistente"),
                &["insistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aboriganal"),
                &["aboriginal"],
            ),
            (dictgen::InsensitiveStr::Ascii("anteena"), &["antenna"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensonal"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenght"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("addiotional"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("owudl"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("batery"), &["battery"]),
            (
                dictgen::InsensitiveStr::Ascii("abselutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversaton"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randomeness"),
                &["randomness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberatrianism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("steriel"), &["sterile"]),
            (
                dictgen::InsensitiveStr::Ascii("insentives"),
                &["incentives"],
            ),
            (dictgen::InsensitiveStr::Ascii("servans"), &["servants"]),
            (dictgen::InsensitiveStr::Ascii("yoyu"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("vriable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("reliefing"), &["relieving"]),
            (dictgen::InsensitiveStr::Ascii("ibrary"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("iniation"), &["initiation"]),
            (dictgen::InsensitiveStr::Ascii("melodis"), &["melodies"]),
            (
                dictgen::InsensitiveStr::Ascii("unpredicatble"),
                &["unpredictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("soldily"), &["solidly"]),
            (dictgen::InsensitiveStr::Ascii("tcpdumpp"), &["tcpdump"]),
            (
                dictgen::InsensitiveStr::Ascii("reserrection"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("hdinsight"), &["hindsight"]),
            (dictgen::InsensitiveStr::Ascii("speshel"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("rounting"),
                &["routing", "rounding"],
            ),
            (dictgen::InsensitiveStr::Ascii("enpdoints"), &["endpoints"]),
            (dictgen::InsensitiveStr::Ascii("inproved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("coneects"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("ccale"), &["scale"]),
            (
                dictgen::InsensitiveStr::Ascii("handycapped"),
                &["handicapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("exctractor"), &["extractor"]),
            (
                dictgen::InsensitiveStr::Ascii("introdution"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("periode"), &["period"]),
            (dictgen::InsensitiveStr::Ascii("waring"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("preparetions"),
                &["preparations"],
            ),
            (dictgen::InsensitiveStr::Ascii("neumonia"), &["pneumonia"]),
            (dictgen::InsensitiveStr::Ascii("conflictos"), &["conflicts"]),
            (
                dictgen::InsensitiveStr::Ascii("enlightenmnt"),
                &["enlightenment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initaiting"),
                &["initiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("awrning"),
                &["awning", "warning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nitifications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("commansd"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("happeneing"), &["happening"]),
            (
                dictgen::InsensitiveStr::Ascii("wavelegnth"),
                &["wavelength"],
            ),
            (dictgen::InsensitiveStr::Ascii("jeruslaem"), &["jerusalem"]),
            (dictgen::InsensitiveStr::Ascii("organered"), &["orangered"]),
            (dictgen::InsensitiveStr::Ascii("ingores"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("cotinue"), &["continue"]),
            (dictgen::InsensitiveStr::Ascii("periodisch"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("missunderstood"),
                &["misunderstood"],
            ),
            (dictgen::InsensitiveStr::Ascii("ternimals"), &["terminals"]),
            (dictgen::InsensitiveStr::Ascii("golaith"), &["goliath"]),
            (
                dictgen::InsensitiveStr::Ascii("workaronds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suscessful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatoutg"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("dissipatore"),
                &["dissipate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kroniclers"),
                &["chroniclers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouying"), &["buoying"]),
            (
                dictgen::InsensitiveStr::Ascii("powerhsell"),
                &["powershell"],
            ),
            (dictgen::InsensitiveStr::Ascii("prividers"), &["providers"]),
            (dictgen::InsensitiveStr::Ascii("hanppens"), &["happens"]),
            (
                dictgen::InsensitiveStr::Ascii("funchtioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("signapore"), &["singapore"]),
            (
                dictgen::InsensitiveStr::Ascii("neighrorhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencicated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalisters"),
                &["journalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtorizing"),
                &["authorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("assimed"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("captacitance"),
                &["capacitance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surpressing"),
                &["suppressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decomposiion"),
                &["decomposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transforums"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technoligical"),
                &["technological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unables"),
                &["enables", "unable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythesize"),
                &["synthesize"],
            ),
            (dictgen::InsensitiveStr::Ascii("suspsend"), &["suspend"]),
            (
                dictgen::InsensitiveStr::Ascii("manifactures"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("wroker"), &["worker"]),
            (
                dictgen::InsensitiveStr::Ascii("canonivalizes"),
                &["canonicalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("meeeting"), &["meeting"]),
            (dictgen::InsensitiveStr::Ascii("pseude"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("perogative"),
                &["prerogative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capatibility"),
                &["compatibility", "capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("querry"),
                &["quarry", "query"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumisations"),
                &["customisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("shitstom"), &["shitstorm"]),
            (
                dictgen::InsensitiveStr::Ascii("descriibing"),
                &["describing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnearble"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("genarally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("gonig"), &["going"]),
            (
                dictgen::InsensitiveStr::Ascii("adivsories"),
                &["advisories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documantation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("channged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("grizzlay"), &["grizzly"]),
            (
                dictgen::InsensitiveStr::Ascii("generostiy"),
                &["generosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("includ"), &["include"]),
            (dictgen::InsensitiveStr::Ascii("fasion"), &["fashion"]),
            (dictgen::InsensitiveStr::Ascii("wroet"), &["wrote"]),
            (
                dictgen::InsensitiveStr::Ascii("authoratitatively"),
                &["authoritatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("honoric"), &["honorific"]),
            (dictgen::InsensitiveStr::Ascii("bachelores"), &["bachelors"]),
            (
                dictgen::InsensitiveStr::Ascii("oversubcribe"),
                &["oversubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretating"),
                &["misinterpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("preffered"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("windosr"), &["windsor"]),
            (dictgen::InsensitiveStr::Ascii("polular"), &["popular"]),
            (
                dictgen::InsensitiveStr::Ascii("miselaneously"),
                &["miscellaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("scetions"), &["sections"]),
            (
                dictgen::InsensitiveStr::Ascii("exclaimations"),
                &["exclamations"],
            ),
            (dictgen::InsensitiveStr::Ascii("afircan"), &["african"]),
            (dictgen::InsensitiveStr::Ascii("langagues"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("provideds"),
                &["provides", "provided"],
            ),
            (dictgen::InsensitiveStr::Ascii("feasble"), &["feasible"]),
            (dictgen::InsensitiveStr::Ascii("expanble"), &["expandable"]),
            (
                dictgen::InsensitiveStr::Ascii("absuolately"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressetting"), &["resetting"]),
            (dictgen::InsensitiveStr::Ascii("chescked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreegation"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpossibility"),
                &["impossibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("displa"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("meanin"), &["meaning"]),
            (dictgen::InsensitiveStr::Ascii("tothiba"), &["toshiba"]),
            (dictgen::InsensitiveStr::Ascii("touble"), &["trouble"]),
            (dictgen::InsensitiveStr::Ascii("decriment"), &["decrement"]),
            (
                dictgen::InsensitiveStr::Ascii("aglorithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("generaly"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("jusifies"), &["justifies"]),
            (dictgen::InsensitiveStr::Ascii("arquives"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("disqustingly"),
                &["disgustingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulgs"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("folllowiong"),
                &["following"],
            ),
            (dictgen::InsensitiveStr::Ascii("chimmeny"), &["chimney"]),
            (
                dictgen::InsensitiveStr::Ascii("specifigations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("ealry"), &["early"]),
            (dictgen::InsensitiveStr::Ascii("isssued"), &["issued"]),
            (dictgen::InsensitiveStr::Ascii("millitant"), &["militant"]),
            (
                dictgen::InsensitiveStr::Ascii("reductible"),
                &["deductible", "reducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correllating"),
                &["correlating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucrures"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostanme"), &["hostname"]),
            (
                dictgen::InsensitiveStr::Ascii("accepetance"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("grapic"), &["graphic"]),
            (dictgen::InsensitiveStr::Ascii("throughly"), &["thoroughly"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptiont"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("chcklist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("retangle"), &["rectangle"]),
            (
                dictgen::InsensitiveStr::Ascii("astonoshing"),
                &["astonishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semiconducters"),
                &["semiconductors"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavin"), &["behaving"]),
            (dictgen::InsensitiveStr::Ascii("eighter"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("reimpliment"),
                &["reimplement"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeing"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("professionnally"),
                &["professionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("siringes"), &["syringes"]),
            (
                dictgen::InsensitiveStr::Ascii("objectivelly"),
                &["objectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("stratups"), &["startups"]),
            (dictgen::InsensitiveStr::Ascii("poriod"), &["period"]),
            (dictgen::InsensitiveStr::Ascii("delpeted"), &["depleted"]),
            (dictgen::InsensitiveStr::Ascii("colateral"), &["collateral"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosyncracies"),
                &["idiosyncrasies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presistency"),
                &["persistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("exat"), &["exact"]),
            (dictgen::InsensitiveStr::Ascii("hiddent"), &["hidden"]),
            (dictgen::InsensitiveStr::Ascii("gatway"), &["gateway"]),
            (dictgen::InsensitiveStr::Ascii("poit"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("porportioning"),
                &["proportioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandchildern"),
                &["grandchildren"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoints"), &["accounts"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetutate"),
                &["perpetuate"],
            ),
            (dictgen::InsensitiveStr::Ascii("simulatie"), &["simulate"]),
            (
                dictgen::InsensitiveStr::Ascii("governmanet"),
                &["governmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("oeration"), &["operation"]),
            (
                dictgen::InsensitiveStr::Ascii("consuption"),
                &["consumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("transiet"), &["transient"]),
            (dictgen::InsensitiveStr::Ascii("inlcluding"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("fliped"), &["flipped"]),
            (
                dictgen::InsensitiveStr::Ascii("califorian"),
                &["californian"],
            ),
            (dictgen::InsensitiveStr::Ascii("refrom"), &["reform"]),
            (dictgen::InsensitiveStr::Ascii("cheeper"), &["cheaper"]),
            (dictgen::InsensitiveStr::Ascii("unvalid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("enourmously"),
                &["enormously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentictor"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalizations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fortelling"),
                &["foretelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("quafeur"), &["coiffure"]),
            (
                dictgen::InsensitiveStr::Ascii("executabels"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obscolescence"),
                &["obsolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("uestion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("lient"),
                &["client", "clients"],
            ),
            (dictgen::InsensitiveStr::Ascii("essense"), &["essence"]),
            (dictgen::InsensitiveStr::Ascii("inctroduce"), &["introduce"]),
            (dictgen::InsensitiveStr::Ascii("aesily"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("balencers"), &["balancers"]),
            (
                dictgen::InsensitiveStr::Ascii("algortithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("remooving"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("rollercoaters"),
                &["rollercoaster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoporate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootsraps"), &["bootstraps"]),
            (dictgen::InsensitiveStr::Ascii("vaccumed"), &["vacuumed"]),
            (dictgen::InsensitiveStr::Ascii("sequncers"), &["sequencers"]),
            (dictgen::InsensitiveStr::Ascii("bsuiness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("triggereing"),
                &["triggering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complections"),
                &["completions"],
            ),
            (dictgen::InsensitiveStr::Ascii("impeared"), &["impaired"]),
            (dictgen::InsensitiveStr::Ascii("hieroglph"), &["hieroglyph"]),
            (dictgen::InsensitiveStr::Ascii("exceededd"), &["exceeded"]),
            (dictgen::InsensitiveStr::Ascii("objcts"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("ugprade"), &["upgrade"]),
            (dictgen::InsensitiveStr::Ascii("pachaged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("disply"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("environents"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sllocation"),
                &["allocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neading"),
                &["heading", "kneading", "needing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbbreviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contanimated"),
                &["contaminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrcutors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("filltering"), &["filtering"]),
            (dictgen::InsensitiveStr::Ascii("occsional"), &["occasional"]),
            (dictgen::InsensitiveStr::Ascii("cosumers"), &["consumers"]),
            (
                dictgen::InsensitiveStr::Ascii("consumbale"),
                &["consumables"],
            ),
            (dictgen::InsensitiveStr::Ascii("bizare"), &["bizarre"]),
            (
                dictgen::InsensitiveStr::Ascii("palestinains"),
                &["palestinians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concret"),
                &["concert", "concrete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intorductions"),
                &["introductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("capcities"), &["capacities"]),
            (
                dictgen::InsensitiveStr::Ascii("reccurrently"),
                &["recurrently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soundtrakk"),
                &["soundtrack"],
            ),
            (dictgen::InsensitiveStr::Ascii("accompagny"), &["accompany"]),
            (
                dictgen::InsensitiveStr::Ascii("lavelled"),
                &["labelled", "levelled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copenhangen"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepair"), &["prepare"]),
            (dictgen::InsensitiveStr::Ascii("platofrms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("refridgerators"),
                &["refrigerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("founf"), &["found"]),
            (dictgen::InsensitiveStr::Ascii("concelear"), &["concealer"]),
            (
                dictgen::InsensitiveStr::Ascii("disucussing"),
                &["discussing"],
            ),
            (dictgen::InsensitiveStr::Ascii("netwroked"), &["networked"]),
            (dictgen::InsensitiveStr::Ascii("htikn"), &["think"]),
            (dictgen::InsensitiveStr::Ascii("opthalmic"), &["ophthalmic"]),
            (dictgen::InsensitiveStr::Ascii("intgers"), &["integers"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhoud"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("wtyle"), &["style"]),
            (dictgen::InsensitiveStr::Ascii("integrarla"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("conntect"), &["connect"]),
            (dictgen::InsensitiveStr::Ascii("produsing"), &["producing"]),
            (dictgen::InsensitiveStr::Ascii("nieghbors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("framws"), &["frames"]),
            (
                dictgen::InsensitiveStr::Ascii("outcalssed"),
                &["outclassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("harrasing"), &["harassing"]),
            (
                dictgen::InsensitiveStr::Ascii("distributer"),
                &["distributor", "distributed", "distributes", "distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifitions"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitfileld"), &["bitfield"]),
            (dictgen::InsensitiveStr::Ascii("cuasality"), &["causality"]),
            (
                dictgen::InsensitiveStr::Ascii("delevoping"),
                &["developing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etroneously"),
                &["erroneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("speaches"), &["speeches"]),
            (
                dictgen::InsensitiveStr::Ascii("calbirated"),
                &["calibrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("attepted"), &["attempted"]),
            (dictgen::InsensitiveStr::Ascii("accapts"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("cource"),
                &["course", "coerce", "source"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stroy"),
                &["story", "destroy"],
            ),
            (dictgen::InsensitiveStr::Ascii("jelous"), &["jealous"]),
            (
                dictgen::InsensitiveStr::Ascii("predacessor"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("pursiut"), &["pursuit"]),
            (dictgen::InsensitiveStr::Ascii("unkonwn"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("interep"), &["interop"]),
            (dictgen::InsensitiveStr::Ascii("bnecause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("callinth"), &["calling"]),
            (dictgen::InsensitiveStr::Ascii("retunrning"), &["returning"]),
            (
                dictgen::InsensitiveStr::Ascii("gurentees"),
                &["guarantees", "guarantee"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoaiates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomendatoin"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archeytpes"),
                &["archetypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("syndiacte"), &["syndicate"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpenisve"),
                &["inexpensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessite"),
                &["necessities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlising"),
                &["generalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("meantioned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("conversationable"),
                &["conversational"],
            ),
            (dictgen::InsensitiveStr::Ascii("truged"), &["trudged"]),
            (
                dictgen::InsensitiveStr::Ascii("ciphersuits"),
                &["ciphersuites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressumed"),
                &["presumed", "pressured"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesitcle"), &["testicle"]),
            (
                dictgen::InsensitiveStr::Ascii("therepists"),
                &["therapists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condemming"),
                &["condemning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuigrations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumetns"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("subrountines"),
                &["subroutines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcisissm"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("statisfies"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlates"),
                &["translates"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoruced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("evalutaors"),
                &["evaluators"],
            ),
            (dictgen::InsensitiveStr::Ascii("shcedule"), &["schedule"]),
            (dictgen::InsensitiveStr::Ascii("calcucate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("repalys"), &["replays"]),
            (dictgen::InsensitiveStr::Ascii("privilegs"), &["privileges"]),
            (dictgen::InsensitiveStr::Ascii("destrucor"), &["destructor"]),
            (
                dictgen::InsensitiveStr::Ascii("descripions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliseshow"), &["slideshow"]),
            (dictgen::InsensitiveStr::Ascii("buid"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("ducumented"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingition"), &["ignition"]),
            (
                dictgen::InsensitiveStr::Ascii("deconstracted"),
                &["deconstructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaildators"),
                &["validators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialiased"),
                &["specialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("benjer"), &["binger"]),
            (dictgen::InsensitiveStr::Ascii("morotola"), &["motorola"]),
            (dictgen::InsensitiveStr::Ascii("blackhaws"), &["blackhawks"]),
            (dictgen::InsensitiveStr::Ascii("constrast"), &["contrast"]),
            (
                dictgen::InsensitiveStr::Ascii("particpated"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("thourough"), &["thorough"]),
            (dictgen::InsensitiveStr::Ascii("perclude"), &["preclude"]),
            (dictgen::InsensitiveStr::Ascii("setyps"), &["setups"]),
            (dictgen::InsensitiveStr::Ascii("qeuestions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("dirations"), &["durations"]),
            (dictgen::InsensitiveStr::Ascii("supprters"), &["supporters"]),
            (dictgen::InsensitiveStr::Ascii("extracion"), &["extraction"]),
            (
                dictgen::InsensitiveStr::Ascii("instruting"),
                &["instructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genearating"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmane"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("chceks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("caluclations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forwareded"),
                &["forward", "forwarded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parcticularly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metodology"),
                &["methodology"],
            ),
            (dictgen::InsensitiveStr::Ascii("animets"), &["animates"]),
            (
                dictgen::InsensitiveStr::Ascii("destinato"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cumultative"),
                &["cumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleaed"),
                &["cleaned", "cleared"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticists"), &["critics"]),
            (
                dictgen::InsensitiveStr::Ascii("delagations"),
                &["delegations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formualtes"),
                &["formulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temportarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("legitimt"), &["legitimate"]),
            (dictgen::InsensitiveStr::Ascii("nodels"), &["models"]),
            (dictgen::InsensitiveStr::Ascii("surounds"), &["surrounds"]),
            (dictgen::InsensitiveStr::Ascii("determen"), &["determines"]),
            (
                dictgen::InsensitiveStr::Ascii("uniterrupted"),
                &["uninterrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("borrwable"), &["borrowable"]),
            (dictgen::InsensitiveStr::Ascii("ocurrs"), &["occurs"]),
            (
                dictgen::InsensitiveStr::Ascii("chercklists"),
                &["checklists"],
            ),
            (dictgen::InsensitiveStr::Ascii("comprassem"), &["compress"]),
            (
                dictgen::InsensitiveStr::Ascii("disappeard"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("suspeneded"), &["suspended"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifierais"),
                &["qualifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("historinhas"),
                &["historians"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetical"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("initating"),
                &["imitating", "initiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoorts"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("repostd"), &["reposted"]),
            (
                dictgen::InsensitiveStr::Ascii("configaring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("forfit"), &["forfeit"]),
            (dictgen::InsensitiveStr::Ascii("obvisous"), &["obvious"]),
            (dictgen::InsensitiveStr::Ascii("ded"), &["dead"]),
            (
                dictgen::InsensitiveStr::Ascii("napoleonian"),
                &["napoleonic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularlizes"),
                &["regularizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annonation"),
                &["annotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("platoe"), &["plateau"]),
            (
                dictgen::InsensitiveStr::Ascii("neumonitis"),
                &["pneumonitis"],
            ),
            (dictgen::InsensitiveStr::Ascii("weaknesss"), &["weaknesses"]),
            (dictgen::InsensitiveStr::Ascii("chceking"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("multplications"),
                &["multiplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("scramblies"), &["scramble"]),
            (dictgen::InsensitiveStr::Ascii("forearmes"), &["forearms"]),
            (dictgen::InsensitiveStr::Ascii("dsplay"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("identificacion"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimumum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("analyitcs"), &["analytics"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomfrotably"),
                &["uncomfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("soical"), &["social"]),
            (dictgen::InsensitiveStr::Ascii("certern"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("irrispective"),
                &["irrespective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatator"),
                &["combinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maneouvres"),
                &["manoeuvres"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresion"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("regon"), &["region"]),
            (
                dictgen::InsensitiveStr::Ascii("creae"),
                &["create", "crease"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetimentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valume"),
                &["volume", "value"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("medatite"), &["meditate"]),
            (dictgen::InsensitiveStr::Ascii("subtitls"), &["subtitles"]),
            (
                dictgen::InsensitiveStr::Ascii("caling"),
                &["calling", "scaling", "culling"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearted"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("sucessess"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("assuptions"),
                &["assumptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("treates"), &["treats"]),
            (dictgen::InsensitiveStr::Ascii("offcers"), &["officers"]),
            (
                dictgen::InsensitiveStr::Ascii("respositories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accruacies"),
                &["accuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrbutions"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registrates"),
                &["registers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreacite"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("absurdy"), &["absurdly"]),
            (
                dictgen::InsensitiveStr::Ascii("homogenity"),
                &["homogeneity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geerations"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("zelaous"), &["zealous"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriatley"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("discraces"), &["disgraces"]),
            (dictgen::InsensitiveStr::Ascii("clickear"), &["clicker"]),
            (
                dictgen::InsensitiveStr::Ascii("capabiltiies"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("condifure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("othewise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("homosexualiy"),
                &["homosexuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcurates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituees"),
                &["prostitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("playists"), &["playlists"]),
            (
                dictgen::InsensitiveStr::Ascii("effecively"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desciption"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("whil"), &["while", "will"]),
            (dictgen::InsensitiveStr::Ascii("installled"), &["installed"]),
            (
                dictgen::InsensitiveStr::Ascii("unscubscribe"),
                &["subscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determening"),
                &["determining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finacially"),
                &["financially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructin"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("follewers"), &["followers"]),
            (
                dictgen::InsensitiveStr::Ascii("abbriviations"),
                &["abbreviations"],
            ),
            (dictgen::InsensitiveStr::Ascii("postgrsql"), &["postgresql"]),
            (dictgen::InsensitiveStr::Ascii("congifure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("unbreakeble"),
                &["unbreakable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abanondment"),
                &["abandonment"],
            ),
            (dictgen::InsensitiveStr::Ascii("luandry"), &["laundry"]),
            (dictgen::InsensitiveStr::Ascii("rotaitons"), &["rotations"]),
            (dictgen::InsensitiveStr::Ascii("inprooves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("muliply"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratins"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrads"), &["threads"]),
            (dictgen::InsensitiveStr::Ascii("vyer"), &["very"]),
            (dictgen::InsensitiveStr::Ascii("insterad"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("fantasazing"),
                &["fantasizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devlivering"),
                &["delivering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unstaller"),
                &["installer", "uninstaller"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffcult"), &["difficult"]),
            (dictgen::InsensitiveStr::Ascii("breefly"), &["briefly"]),
            (
                dictgen::InsensitiveStr::Ascii("investigationes"),
                &["investigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("romoteing"), &["remoting"]),
            (
                dictgen::InsensitiveStr::Ascii("disusses"),
                &["discusses", "disuses"],
            ),
            (dictgen::InsensitiveStr::Ascii("deault"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("moisturier"),
                &["moisturizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("versital"), &["versatile"]),
            (dictgen::InsensitiveStr::Ascii("farsical"), &["farcical"]),
            (
                dictgen::InsensitiveStr::Ascii("currenct"),
                &["current", "currency"],
            ),
            (dictgen::InsensitiveStr::Ascii("rigourous"), &["rigorous"]),
            (dictgen::InsensitiveStr::Ascii("forground"), &["foreground"]),
            (dictgen::InsensitiveStr::Ascii("vacines"), &["vaccines"]),
            (dictgen::InsensitiveStr::Ascii("potentail"), &["potential"]),
            (dictgen::InsensitiveStr::Ascii("asshates"), &["asshats"]),
            (dictgen::InsensitiveStr::Ascii("researves"), &["reserves"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliviring"),
                &["delivering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("biddin"),
                &["bidding", "bidden"],
            ),
            (dictgen::InsensitiveStr::Ascii("sintax"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("unanamious"), &["unanimous"]),
            (dictgen::InsensitiveStr::Ascii("fecthes"), &["fetches"]),
            (dictgen::InsensitiveStr::Ascii("actulay"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("understoon"),
                &["understood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepererates"),
                &["separates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectivo"),
                &["collection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpatuates"),
                &["perpetuates"],
            ),
            (dictgen::InsensitiveStr::Ascii("reacing"), &["reaching"]),
            (
                dictgen::InsensitiveStr::Ascii("contributon"),
                &["contribution", "contributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitutn"),
                &["constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("libgng"), &["libpng"]),
            (dictgen::InsensitiveStr::Ascii("chceker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("manifestion"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("declase"), &["declares"]),
            (dictgen::InsensitiveStr::Ascii("intrinisc"), &["intrinsic"]),
            (
                dictgen::InsensitiveStr::Ascii("propoganda"),
                &["propaganda"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosiacitions"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwolrd"),
                &["underworld"],
            ),
            (dictgen::InsensitiveStr::Ascii("athetlic"), &["athletic"]),
            (
                dictgen::InsensitiveStr::Ascii("interogators"),
                &["interrogators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusisast"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("skateboad"), &["skateboard"]),
            (dictgen::InsensitiveStr::Ascii("cuonter"), &["counter"]),
            (
                dictgen::InsensitiveStr::Ascii("circomstances"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excaustion"),
                &["exhaustion"],
            ),
            (dictgen::InsensitiveStr::Ascii("lazer"), &["laser"]),
            (
                dictgen::InsensitiveStr::Ascii("reaoned"),
                &["reasoned", "rezoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("oultline"), &["outline"]),
            (
                dictgen::InsensitiveStr::Ascii("accessibily"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiuring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyliders"), &["cylinders"]),
            (dictgen::InsensitiveStr::Ascii("soild"), &["solid"]),
            (
                dictgen::InsensitiveStr::Ascii("aknowledgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cristallizes"),
                &["crystallizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repositorys"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("excuting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("foppys"), &["floppies"]),
            (
                dictgen::InsensitiveStr::Ascii("crsytallisation"),
                &["crystallisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("offcie"), &["office"]),
            (dictgen::InsensitiveStr::Ascii("adatpers"), &["adapters"]),
            (dictgen::InsensitiveStr::Ascii("libralie"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("benficial"), &["beneficial"]),
            (dictgen::InsensitiveStr::Ascii("poistions"), &["positions"]),
            (
                dictgen::InsensitiveStr::Ascii("remembrence"),
                &["remembrance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strecthing"),
                &["stretching"],
            ),
            (dictgen::InsensitiveStr::Ascii("mappeds"), &["mapped"]),
            (dictgen::InsensitiveStr::Ascii("iranain"), &["iranian"]),
            (dictgen::InsensitiveStr::Ascii("respondis"), &["responds"]),
            (dictgen::InsensitiveStr::Ascii("symmmetry"), &["symmetry"]),
            (dictgen::InsensitiveStr::Ascii("pounts"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("overlaod"), &["overload"]),
            (
                dictgen::InsensitiveStr::Ascii("synchronizin"),
                &["synchronizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbetwen"), &["inbetween"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilitation"),
                &["stabilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciazlize"),
                &["specialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intentionly"),
                &["intentional", "intentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("foraign"), &["foreign"]),
            (dictgen::InsensitiveStr::Ascii("ntegral"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("stabilzed"), &["stabilized"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncialize"),
                &["canonicalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("melatonian"), &["melatonin"]),
            (dictgen::InsensitiveStr::Ascii("returnning"), &["returning"]),
            (dictgen::InsensitiveStr::Ascii("briges"), &["bridges"]),
            (dictgen::InsensitiveStr::Ascii("lenggth"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("coneptual"), &["conceptual"]),
            (
                dictgen::InsensitiveStr::Ascii("calcutations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foreward"),
                &["foreword", "forward", "forewarn"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remenissing"),
                &["reminiscing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developpments"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequeseces"),
                &["consequences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synonynous"),
                &["synonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demodulaters"),
                &["demodulators", "demodulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rucuperate"),
                &["recuperate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuatable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("aceptable"), &["acceptable"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhemling"),
                &["overwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("formualate"), &["formulate"]),
            (dictgen::InsensitiveStr::Ascii("creaators"), &["creators"]),
            (dictgen::InsensitiveStr::Ascii("swicher"), &["switcher"]),
            (
                dictgen::InsensitiveStr::Ascii("recource"),
                &["resource", "recourse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciniplary"),
                &["disciplinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("matreials"), &["materials"]),
            (dictgen::InsensitiveStr::Ascii("jurnal"), &["journal"]),
            (dictgen::InsensitiveStr::Ascii("circuling"), &["circling"]),
            (dictgen::InsensitiveStr::Ascii("checlist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("duaghter"), &["daughter"]),
            (dictgen::InsensitiveStr::Ascii("abuseres"), &["abusers"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusivitiy"),
                &["exclusivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("extening"), &["extending"]),
            (
                dictgen::InsensitiveStr::Ascii("continuning"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transvorms"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornogrophy"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowedge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("overvride"), &["override"]),
            (
                dictgen::InsensitiveStr::Ascii("represetation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("forumlas"), &["formulas"]),
            (
                dictgen::InsensitiveStr::Ascii("atructuring"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("plagerize"), &["plagiarize"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptyion"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstructer"),
                &["reconstructor", "reconstruct"],
            ),
            (dictgen::InsensitiveStr::Ascii("withput"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("progates"), &["propagates"]),
            (
                dictgen::InsensitiveStr::Ascii("predictons"),
                &["predictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypethetical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("carpetner"), &["carpenter"]),
            (
                dictgen::InsensitiveStr::Ascii("exporession"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("headquaters"),
                &["headquarters"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimised"), &["minimised"]),
            (
                dictgen::InsensitiveStr::Ascii("soruce"),
                &["source", "spruce"],
            ),
            (dictgen::InsensitiveStr::Ascii("collpses"), &["collapses"]),
            (dictgen::InsensitiveStr::Ascii("allternate"), &["alternate"]),
            (
                dictgen::InsensitiveStr::Ascii("irreproducable"),
                &["irreproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("occulsion"), &["occlusion"]),
            (
                dictgen::InsensitiveStr::Ascii("decremenetd"),
                &["decremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimeter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("skippping"), &["skipping"]),
            (dictgen::InsensitiveStr::Ascii("reched"), &["reached"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutinary"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrengements"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("lookuup"), &["lookup"]),
            (dictgen::InsensitiveStr::Ascii("whell"), &["wheel", "well"]),
            (dictgen::InsensitiveStr::Ascii("affinty"), &["affinity"]),
            (
                dictgen::InsensitiveStr::Ascii("presidenta"),
                &["presidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("intstantly"), &["instantly"]),
            (dictgen::InsensitiveStr::Ascii("allacrity"), &["alacrity"]),
            (dictgen::InsensitiveStr::Ascii("appraches"), &["approaches"]),
            (
                dictgen::InsensitiveStr::Ascii("sirections"),
                &["directions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropriatons"),
                &["appropriations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proficiancy"),
                &["proficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transsexuella"),
                &["transsexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerablly"),
                &["considerably"],
            ),
            (dictgen::InsensitiveStr::Ascii("hambruger"), &["hamburger"]),
            (dictgen::InsensitiveStr::Ascii("nothng"), &["nothing"]),
            (
                dictgen::InsensitiveStr::Ascii("desrciption"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("polygammy"), &["polygamy"]),
            (dictgen::InsensitiveStr::Ascii("animaiters"), &["animators"]),
            (
                dictgen::InsensitiveStr::Ascii("descrptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("systimatic"),
                &["systematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("dertails"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("spefiied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("steryotype"),
                &["stereotype"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straigtforward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("friendle"), &["friendlies"]),
            (dictgen::InsensitiveStr::Ascii("micorsoft"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("enchantent"),
                &["enchantment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatablitiy"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("monlayer"), &["monolayer"]),
            (dictgen::InsensitiveStr::Ascii("sctipted"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("librabie"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("presentatuion"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("refinment"), &["refinement"]),
            (
                dictgen::InsensitiveStr::Ascii("shenadigans"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verly"),
                &["very", "verily", "overly"],
            ),
            (dictgen::InsensitiveStr::Ascii("changges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("itializers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealearate"),
                &["ameliorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutomisable"),
                &["customisable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrytpting"),
                &["encrypting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nutritian"),
                &["nutritional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immigranti"),
                &["immigration"],
            ),
            (dictgen::InsensitiveStr::Ascii("nominato"), &["nomination"]),
            (
                dictgen::InsensitiveStr::Ascii("spectacularily"),
                &["spectacularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("depndency"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("modifes"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("conclusiones"),
                &["conclusions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participatin"),
                &["participation", "participating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esseintially"),
                &["essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("headerr"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetuare"),
                &["perpetuate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspodences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("secify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("resett"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("grandchilderen"),
                &["grandchildren"],
            ),
            (dictgen::InsensitiveStr::Ascii("consolin"), &["consoling"]),
            (dictgen::InsensitiveStr::Ascii("defineas"), &["defines"]),
            (dictgen::InsensitiveStr::Ascii("clasic"), &["classic"]),
            (
                dictgen::InsensitiveStr::Ascii("contirbuted"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inforrmation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deifnately"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("releifed"), &["relieved"]),
            (
                dictgen::InsensitiveStr::Ascii("obfuscafted"),
                &["obfuscated"],
            ),
            (dictgen::InsensitiveStr::Ascii("soundtrak"), &["soundtrack"]),
            (dictgen::InsensitiveStr::Ascii("evalutive"), &["evaluative"]),
            (
                dictgen::InsensitiveStr::Ascii("happend"),
                &["happened", "happens", "happen"],
            ),
            (dictgen::InsensitiveStr::Ascii("catagori"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("backneds"), &["backends"]),
            (dictgen::InsensitiveStr::Ascii("occours"), &["occurs"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitioning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unifomtity"),
                &["uniformity"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetifying"), &["certifying"]),
            (
                dictgen::InsensitiveStr::Ascii("decraeses"),
                &["decreases", "deceases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforcable"),
                &["enforceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpecetd"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expieriences"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("readdrss"), &["readdress"]),
            (
                dictgen::InsensitiveStr::Ascii("embarassing"),
                &["embarrassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archietectural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandinaiva"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resorucing"),
                &["resourcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("neeedle"), &["needle"]),
            (dictgen::InsensitiveStr::Ascii("sysems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("tranistioned"),
                &["transitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("whatsover"), &["whatsoever"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatrs"),
                &["alternators"],
            ),
            (dictgen::InsensitiveStr::Ascii("igzorter"), &["exhorter"]),
            (dictgen::InsensitiveStr::Ascii("ndefined"), &["undefined"]),
            (
                dictgen::InsensitiveStr::Ascii("nesseccarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorperates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermiedate"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcaes"), &["spaces"]),
            (dictgen::InsensitiveStr::Ascii("provdie"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("pluged"), &["plugged"]),
            (
                dictgen::InsensitiveStr::Ascii("demostrated"),
                &["demonstrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reagards"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("suppies"), &["supplies"]),
            (dictgen::InsensitiveStr::Ascii("designd"), &["designed"]),
            (dictgen::InsensitiveStr::Ascii("cabnets"), &["cabinets"]),
            (dictgen::InsensitiveStr::Ascii("detial"), &["detail"]),
            (dictgen::InsensitiveStr::Ascii("formerlly"), &["formerly"]),
            (dictgen::InsensitiveStr::Ascii("maintaines"), &["maintains"]),
            (
                dictgen::InsensitiveStr::Ascii("gratest"),
                &["greatest", "grates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisioning"),
                &["commissioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("senerity"), &["serenity"]),
            (dictgen::InsensitiveStr::Ascii("svelt"), &["svelte"]),
            (dictgen::InsensitiveStr::Ascii("magnatiude"), &["magnitude"]),
            (
                dictgen::InsensitiveStr::Ascii("volounteering"),
                &["volunteering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porportions"),
                &["proportions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitle"),
                &["subtitle", "substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometimed"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("pronous"), &["pronouns"]),
            (
                dictgen::InsensitiveStr::Ascii("achievments"),
                &["achievements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destructer"),
                &["destructor", "destruct"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaridans"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("appareil"), &["apparel"]),
            (dictgen::InsensitiveStr::Ascii("heriocs"), &["heroics"]),
            (
                dictgen::InsensitiveStr::Ascii("constructopr"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("kiyacked"), &["kayaked"]),
            (dictgen::InsensitiveStr::Ascii("idealisim"), &["idealism"]),
            (
                dictgen::InsensitiveStr::Ascii("recepticals"),
                &["receptacles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallicunation"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("caclate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("suuported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("grahics"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("oppportunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("expemt"), &["exempt"]),
            (
                dictgen::InsensitiveStr::Ascii("southerness"),
                &["southerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerising"),
                &["parameterising"],
            ),
            (dictgen::InsensitiveStr::Ascii("randomes"), &["randoms"]),
            (dictgen::InsensitiveStr::Ascii("abandonded"), &["abandoned"]),
            (
                dictgen::InsensitiveStr::Ascii("acheviements"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("transplat"), &["transplant"]),
            (dictgen::InsensitiveStr::Ascii("invincibe"), &["invincible"]),
            (dictgen::InsensitiveStr::Ascii("unpacke"), &["unpacked"]),
            (
                dictgen::InsensitiveStr::Ascii("disclipine"),
                &["discipline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissinos"),
                &["permissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("cimetric"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("langyage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("leary"), &["leery"]),
            (
                dictgen::InsensitiveStr::Ascii("orgnization"),
                &["organization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separatin"),
                &["separating", "separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopefullu"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("odly"), &["oddly"]),
            (
                dictgen::InsensitiveStr::Ascii("unparseable"),
                &["unparsable"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangnig"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("loctions"),
                &["locations", "lotions", "lections", "locutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diferrences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wormarounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calncellations"),
                &["cancellations"],
            ),
            (dictgen::InsensitiveStr::Ascii("behaivior"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("recallection"),
                &["recollection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defendents"),
                &["defendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chesckerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("itertate"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("hyperoble"), &["hyperbole"]),
            (dictgen::InsensitiveStr::Ascii("tetxture"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("unknwowns"), &["unknowns"]),
            (dictgen::InsensitiveStr::Ascii("retrievin"), &["retrieving"]),
            (
                dictgen::InsensitiveStr::Ascii("perpertrated"),
                &["perpetrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlcean"), &["unclean"]),
            (dictgen::InsensitiveStr::Ascii("uesrnames"), &["usernames"]),
            (dictgen::InsensitiveStr::Ascii("manfiest"), &["manifest"]),
            (
                dictgen::InsensitiveStr::Ascii("cheesebruger"),
                &["cheeseburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncontrollabley"),
                &["uncontrollably"],
            ),
            (dictgen::InsensitiveStr::Ascii("athism"), &["atheism"]),
            (dictgen::InsensitiveStr::Ascii("acitivate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("trainwreak"),
                &["trainwreck"],
            ),
            (dictgen::InsensitiveStr::Ascii("apending"), &["appending"]),
            (
                dictgen::InsensitiveStr::Ascii("experimdntations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vetored"),
                &["vectored", "vetoed"],
            ),
            (dictgen::InsensitiveStr::Ascii("incresed"), &["increased"]),
            (
                dictgen::InsensitiveStr::Ascii("deserailised"),
                &["deserialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("tradions"), &["traditions"]),
            (
                dictgen::InsensitiveStr::Ascii("nutrutious"),
                &["nutritious"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribue"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("compuslion"),
                &["compulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("coridal"), &["cordial"]),
            (
                dictgen::InsensitiveStr::Ascii("cheezeburger"),
                &["cheeseburger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restucturing"),
                &["restructuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchmarmking"),
                &["benchmarking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proficienty"),
                &["proficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitap"), &["bitmap"]),
            (
                dictgen::InsensitiveStr::Ascii("alievating"),
                &["alienating", "alleviating"],
            ),
            (dictgen::InsensitiveStr::Ascii("destros"), &["destroys"]),
            (dictgen::InsensitiveStr::Ascii("peepels"), &["peoples"]),
            (dictgen::InsensitiveStr::Ascii("ankshios"), &["anxious"]),
            (
                dictgen::InsensitiveStr::Ascii("frogetting"),
                &["forgetting"],
            ),
            (dictgen::InsensitiveStr::Ascii("initilise"), &["initialise"]),
            (
                dictgen::InsensitiveStr::Ascii("definetely"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("stylsheet"), &["stylesheet"]),
            (dictgen::InsensitiveStr::Ascii("amognst"), &["amongst"]),
            (dictgen::InsensitiveStr::Ascii("threhsold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("explicatif"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("statuline"), &["statusline"]),
            (
                dictgen::InsensitiveStr::Ascii("proprogated"),
                &["propagated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprepresents"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("carful"), &["careful"]),
            (dictgen::InsensitiveStr::Ascii("skecthes"), &["sketches"]),
            (dictgen::InsensitiveStr::Ascii("permitis"), &["permits"]),
            (
                dictgen::InsensitiveStr::Ascii("electromechnical"),
                &["electromechanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrobutions"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("searcheable"),
                &["searchable"],
            ),
            (dictgen::InsensitiveStr::Ascii("expandr"), &["expander"]),
            (dictgen::InsensitiveStr::Ascii("immeditate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("infininte"), &["infinite"]),
            (dictgen::InsensitiveStr::Ascii("proproably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("varibility"),
                &["variability"],
            ),
            (dictgen::InsensitiveStr::Ascii("laoding"), &["loading"]),
            (dictgen::InsensitiveStr::Ascii("cecksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("trnaslates"),
                &["translates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bannanas"),
                &["bananas", "bandanas"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differance"),
                &["difference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classificaions"),
                &["classifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("formulados"), &["formulas"]),
            (dictgen::InsensitiveStr::Ascii("menetioned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("signitories"),
                &["signatories"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhautive"), &["exhaustive"]),
            (
                dictgen::InsensitiveStr::Ascii("cannels"),
                &["channels", "cancels", "canals"],
            ),
            (dictgen::InsensitiveStr::Ascii("satisfyied"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuprized"),
                &["unsurprised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distincing"),
                &["distancing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shareholdes"),
                &["shareholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwritted"),
                &["overwrite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legilsation"),
                &["legislation"],
            ),
            (dictgen::InsensitiveStr::Ascii("resycn"), &["resync"]),
            (dictgen::InsensitiveStr::Ascii("muffings"), &["muffins"]),
            (
                dictgen::InsensitiveStr::Ascii("inicialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("senintels"), &["sentinels"]),
            (
                dictgen::InsensitiveStr::Ascii("ideologiers"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("pedistals"), &["pedestals"]),
            (dictgen::InsensitiveStr::Ascii("requiers"), &["requires"]),
            (dictgen::InsensitiveStr::Ascii("airloom"), &["heirloom"]),
            (dictgen::InsensitiveStr::Ascii("recrated"), &["recreated"]),
            (
                dictgen::InsensitiveStr::Ascii("androidextra"),
                &["androidextras"],
            ),
            (dictgen::InsensitiveStr::Ascii("jagid"), &["jagged"]),
            (
                dictgen::InsensitiveStr::Ascii("dimansional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esitmating"),
                &["estimating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envioronments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("useg"), &["user", "usage"]),
            (
                dictgen::InsensitiveStr::Ascii("assoiciations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("possibbe"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("impplement"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("hypertorphy"),
                &["hypertrophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spitirually"),
                &["spiritually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convieniently"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directpries"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("girlfreinds"),
                &["girlfriends"],
            ),
            (dictgen::InsensitiveStr::Ascii("tigthens"), &["tightens"]),
            (dictgen::InsensitiveStr::Ascii("difract"), &["diffract"]),
            (
                dictgen::InsensitiveStr::Ascii("measuremenet"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fials"),
                &["fails", "dials", "vials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschologically"),
                &["psychologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealiarators"),
                &["ameliorators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizated"),
                &["organized"],
            ),
            (dictgen::InsensitiveStr::Ascii("unseless"), &["useless"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarcies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propogates"),
                &["propagates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectivty"),
                &["objectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("thi"), &["the", "this"]),
            (
                dictgen::InsensitiveStr::Ascii("meditationg"),
                &["meditating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplyoments"),
                &["deployments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authrorised"),
                &["authorised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsiblty"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converstons"),
                &["conversations", "conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagerations"),
                &["exaggerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspons"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("datrabase"), &["database"]),
            (
                dictgen::InsensitiveStr::Ascii("unavaiability"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playthrouhgs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaions"), &["locations"]),
            (
                dictgen::InsensitiveStr::Ascii("assassinatin"),
                &["assassination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expierenced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definintions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternetly"),
                &["alternately"],
            ),
            (dictgen::InsensitiveStr::Ascii("stnding"), &["standing"]),
            (
                dictgen::InsensitiveStr::Ascii("pregressed"),
                &["progressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("irraditate"), &["irradiate"]),
            (dictgen::InsensitiveStr::Ascii("solgering"), &["soldiering"]),
            (dictgen::InsensitiveStr::Ascii("tro"), &["to"]),
            (dictgen::InsensitiveStr::Ascii("fropm"), &["from"]),
            (
                dictgen::InsensitiveStr::Ascii("paleolitic"),
                &["paleolithic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expcept"),
                &["except", "expect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rollerocaster"),
                &["rollercoaster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impeckable"),
                &["impeccable"],
            ),
            (dictgen::InsensitiveStr::Ascii("enclosue"), &["enclosure"]),
            (dictgen::InsensitiveStr::Ascii("deleletes"), &["deletes"]),
            (
                dictgen::InsensitiveStr::Ascii("incremeneting"),
                &["incrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowwind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("percentabge"),
                &["percentage"],
            ),
            (dictgen::InsensitiveStr::Ascii("cultueral"), &["cultural"]),
            (dictgen::InsensitiveStr::Ascii("insteances"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitor"),
                &["arbiter", "arbitrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identitties"),
                &["identities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diminuitive"),
                &["diminutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("literaly"), &["literally"]),
            (
                dictgen::InsensitiveStr::Ascii("dynamicly"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("alpahabet"), &["alphabet"]),
            (dictgen::InsensitiveStr::Ascii("docuents"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("reviewl"), &["review"]),
            (
                dictgen::InsensitiveStr::Ascii("suceessful"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggresiveness"),
                &["aggressiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonizers"),
                &["synchronizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandnig"), &["sanding"]),
            (dictgen::InsensitiveStr::Ascii("modefy"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("displys"), &["displays"]),
            (
                dictgen::InsensitiveStr::Ascii("excectional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("quew"), &["queue"]),
            (
                dictgen::InsensitiveStr::Ascii("propblematic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("capablle"), &["capable"]),
            (
                dictgen::InsensitiveStr::Ascii("alternaively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfarily"), &["unfairly"]),
            (
                dictgen::InsensitiveStr::Ascii("activiating"),
                &["activating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predecessores"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exucutables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("delviered"), &["delivered"]),
            (
                dictgen::InsensitiveStr::Ascii("antisipated"),
                &["anticipated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alernately"),
                &["alternately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acconplishments"),
                &["accomplishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapearing"),
                &["disappearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categoical"),
                &["categorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("mortarts"), &["mortars"]),
            (
                dictgen::InsensitiveStr::Ascii("invididually"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrosepct"),
                &["retrospect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("happenes"),
                &["happens", "happened", "happiness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quadrapeds"),
                &["quadrupeds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramadics"),
                &["paramedics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconnaissence"),
                &["reconnaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("tnanks"), &["thanks"]),
            (
                dictgen::InsensitiveStr::Ascii("signifitant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abriviating"),
                &["abbreviating"],
            ),
            (dictgen::InsensitiveStr::Ascii("columsn"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("stationerad"),
                &["stationed"],
            ),
            (dictgen::InsensitiveStr::Ascii("frooze"), &["froze"]),
            (
                dictgen::InsensitiveStr::Ascii("reguarlizer"),
                &["regularizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("nodess"), &["nodes"]),
            (
                dictgen::InsensitiveStr::Ascii("autenticates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lackbuster"),
                &["lackluster"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammongst"), &["amongst"]),
            (dictgen::InsensitiveStr::Ascii("doiuble"), &["double"]),
            (
                dictgen::InsensitiveStr::Ascii("unregnized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("accourding"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("calulus"),
                &["calculus", "callus"],
            ),
            (dictgen::InsensitiveStr::Ascii("exicutes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("collions"), &["collisions"]),
            (
                dictgen::InsensitiveStr::Ascii("custumisations"),
                &["customisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranalation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deisgners"), &["designers"]),
            (dictgen::InsensitiveStr::Ascii("desrired"), &["desired"]),
            (
                dictgen::InsensitiveStr::Ascii("mataphorical"),
                &["metaphorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("reutnr"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("continuousy"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("everybodies"),
                &["everybody"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("fatig"), &["fatigue"]),
            (
                dictgen::InsensitiveStr::Ascii("nocontinuable"),
                &["noncontinuable"],
            ),
            (dictgen::InsensitiveStr::Ascii("undelying"), &["underlying"]),
            (
                dictgen::InsensitiveStr::Ascii("knwoledgable"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nothingess"),
                &["nothingness"],
            ),
            (dictgen::InsensitiveStr::Ascii("ddogers"), &["dodgers"]),
            (dictgen::InsensitiveStr::Ascii("forsekan"), &["forsaken"]),
            (
                dictgen::InsensitiveStr::Ascii("multyplayer"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avaiiability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ukrianians"),
                &["ukrainians"],
            ),
            (dictgen::InsensitiveStr::Ascii("andf"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("eisntances"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("ecomonical"),
                &["economical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underrater"),
                &["undertaker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommmendations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("appropach"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("opimises"), &["optimises"]),
            (dictgen::InsensitiveStr::Ascii("subtitels"), &["subtitles"]),
            (dictgen::InsensitiveStr::Ascii("varibels"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("reprensenting"),
                &["representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sythesizes"),
                &["synthesizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleraptors"),
                &["accelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("iit"), &["it"]),
            (dictgen::InsensitiveStr::Ascii("putpose"), &["purpose"]),
            (dictgen::InsensitiveStr::Ascii("languate"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("genearates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("compatitbilities"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("balcked"), &["blacked"]),
            (dictgen::InsensitiveStr::Ascii("prooduce"), &["produce"]),
            (
                dictgen::InsensitiveStr::Ascii("venezuelean"),
                &["venezuela"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creationsism"),
                &["creationism"],
            ),
            (dictgen::InsensitiveStr::Ascii("truging"), &["trudging"]),
            (dictgen::InsensitiveStr::Ascii("charcter"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("leagelize"), &["legalize"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabillity"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("whenenever"), &["whenever"]),
            (dictgen::InsensitiveStr::Ascii("secrety"), &["secretly"]),
            (
                dictgen::InsensitiveStr::Ascii("milliscond"),
                &["millisecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revewing"),
                &["reviewing", "renewing", "reveling"],
            ),
            (dictgen::InsensitiveStr::Ascii("failiure"), &["failure"]),
            (
                dictgen::InsensitiveStr::Ascii("esponds"),
                &["responds", "desponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("theathre"), &["theatre"]),
            (dictgen::InsensitiveStr::Ascii("ceratine"), &["creatine"]),
            (dictgen::InsensitiveStr::Ascii("siffixes"), &["suffixes"]),
            (
                dictgen::InsensitiveStr::Ascii("sponatneous"),
                &["spontaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extansion"),
                &["extension", "expansion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attactment"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("ppoulator"), &["populator"]),
            (
                dictgen::InsensitiveStr::Ascii("substantually"),
                &["substantially"],
            ),
            (dictgen::InsensitiveStr::Ascii("piolting"), &["piloting"]),
            (dictgen::InsensitiveStr::Ascii("expoldes"), &["explodes"]),
            (dictgen::InsensitiveStr::Ascii("hexogans"), &["hexagons"]),
            (dictgen::InsensitiveStr::Ascii("finilized"), &["finalized"]),
            (dictgen::InsensitiveStr::Ascii("obusing"), &["abusing"]),
            (
                dictgen::InsensitiveStr::Ascii("diassociate"),
                &["disassociate", "dissociate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("membranaphone"),
                &["membranophone"],
            ),
            (dictgen::InsensitiveStr::Ascii("invincil"), &["invincible"]),
            (
                dictgen::InsensitiveStr::Ascii("attribbuted"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("appaluse"), &["applause"]),
            (dictgen::InsensitiveStr::Ascii("durign"), &["during"]),
            (
                dictgen::InsensitiveStr::Ascii("initliazes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("friently"), &["friendly"]),
            (dictgen::InsensitiveStr::Ascii("ciclic"), &["cyclic"]),
            (dictgen::InsensitiveStr::Ascii("authering"), &["authoring"]),
            (
                dictgen::InsensitiveStr::Ascii("triagulated"),
                &["triangulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quesitonnaire"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("algoithms"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("okat"), &["okay"]),
            (dictgen::InsensitiveStr::Ascii("facsimilie"), &["facsimile"]),
            (
                dictgen::InsensitiveStr::Ascii("valudation"),
                &["validation", "valuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undoctrinated"),
                &["indoctrinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("possitive"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("teacer"), &["teacher"]),
            (
                dictgen::InsensitiveStr::Ascii("appliation"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convenctionally"),
                &["conventionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diviation"),
                &["divination", "deviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ower"),
                &["lower", "owner", "over"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disabiltities"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cercumstantial"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detecttives"),
                &["detectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workets"),
                &["worklets", "workers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thrresholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmittin"),
                &["transmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fingerpint"),
                &["fingerprint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intoxicaton"),
                &["intoxication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigsting"),
                &["investigating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciprine"),
                &["discipline"],
            ),
            (dictgen::InsensitiveStr::Ascii("kennedey"), &["kennedy"]),
            (
                dictgen::InsensitiveStr::Ascii("conservitives"),
                &["conservatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistro"), &["maestro"]),
            (
                dictgen::InsensitiveStr::Ascii("inititators"),
                &["initiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guaranteeds"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("carloina"), &["carolina"]),
            (
                dictgen::InsensitiveStr::Ascii("requrement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("underlyin"), &["underlying"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallizers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("occcur"), &["occur"]),
            (dictgen::InsensitiveStr::Ascii("overwhem"), &["overwhelm"]),
            (dictgen::InsensitiveStr::Ascii("sertings"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("staition"), &["station"]),
            (dictgen::InsensitiveStr::Ascii("goin"), &["going"]),
            (
                dictgen::InsensitiveStr::Ascii("presentaional"),
                &["presentational"],
            ),
            (dictgen::InsensitiveStr::Ascii("mitakes"), &["mistakes"]),
            (
                dictgen::InsensitiveStr::Ascii("whisperered"),
                &["whispered"],
            ),
            (dictgen::InsensitiveStr::Ascii("recipet"), &["receipt"]),
            (dictgen::InsensitiveStr::Ascii("publiser"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("rearraned"), &["rearranged"]),
            (dictgen::InsensitiveStr::Ascii("conntact"), &["contact"]),
            (
                dictgen::InsensitiveStr::Ascii("contempoary"),
                &["contemporary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("turncating"),
                &["truncating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comproimise"),
                &["compromise"],
            ),
            (dictgen::InsensitiveStr::Ascii("effeck"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("trogladitism"),
                &["troglodytism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optiionally"),
                &["optionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraives"), &["arrives"]),
            (dictgen::InsensitiveStr::Ascii("foxus"), &["focus", "foxes"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoonding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("exaclty"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("defininitely"),
                &["definitively", "definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("pernament"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("elemntal"), &["elemental"]),
            (
                dictgen::InsensitiveStr::Ascii("alighnments"),
                &["alignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoslver"), &["resolver"]),
            (dictgen::InsensitiveStr::Ascii("traceing"), &["tracing"]),
            (
                dictgen::InsensitiveStr::Ascii("mormalized"),
                &["normalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("traverces"), &["traverses"]),
            (dictgen::InsensitiveStr::Ascii("forunate"), &["fortunate"]),
            (
                dictgen::InsensitiveStr::Ascii("converver"),
                &["conserver", "converter"],
            ),
            (dictgen::InsensitiveStr::Ascii("apppendix"), &["appendix"]),
            (dictgen::InsensitiveStr::Ascii("mkae"), &["make"]),
            (
                dictgen::InsensitiveStr::Ascii("encrytping"),
                &["encrypting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independentally"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("revcieved"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("refrences"), &["references"]),
            (
                dictgen::InsensitiveStr::Ascii("compilication"),
                &["complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("formualte"), &["formulate"]),
            (
                dictgen::InsensitiveStr::Ascii("sefault"),
                &["segfault", "default"],
            ),
            (dictgen::InsensitiveStr::Ascii("disliks"), &["dislikes"]),
            (dictgen::InsensitiveStr::Ascii("checcker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("ameelarative"),
                &["ameliorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("definines"), &["defines"]),
            (
                dictgen::InsensitiveStr::Ascii("isntructed"),
                &["instructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeiencing"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("nagivator"), &["navigator"]),
            (dictgen::InsensitiveStr::Ascii("clcosed"), &["closed"]),
            (
                dictgen::InsensitiveStr::Ascii("inherriting"),
                &["inheriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apocalytic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("preimum"), &["premium"]),
            (dictgen::InsensitiveStr::Ascii("brigh"), &["bright"]),
            (
                dictgen::InsensitiveStr::Ascii("destruktor"),
                &["destructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumulare"),
                &["accumulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkeout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("downovting"),
                &["downvoting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cilivization"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintelligient"),
                &["unintelligent"],
            ),
            (dictgen::InsensitiveStr::Ascii("answeres"), &["answers"]),
            (
                dictgen::InsensitiveStr::Ascii("coordindate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniating"), &["initiating"]),
            (dictgen::InsensitiveStr::Ascii("permanet"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("convincted"), &["convince"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenterade"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("hijkacked"), &["hijacked"]),
            (
                dictgen::InsensitiveStr::Ascii("differiator"),
                &["differentiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaptater"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("aqquires"), &["acquires"]),
            (
                dictgen::InsensitiveStr::Ascii("negoatiates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedestiran"),
                &["pedestrian"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaninful"), &["meaningful"]),
            (
                dictgen::InsensitiveStr::Ascii("superceedes"),
                &["supersedes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boierplate"),
                &["boilerplate"],
            ),
            (dictgen::InsensitiveStr::Ascii("remaped"), &["remapped"]),
            (dictgen::InsensitiveStr::Ascii("meading"), &["meaning"]),
            (dictgen::InsensitiveStr::Ascii("archetipe"), &["archetype"]),
            (
                dictgen::InsensitiveStr::Ascii("configurees"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstrcution"),
                &["reconstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("bodyweigt"), &["bodyweight"]),
            (dictgen::InsensitiveStr::Ascii("assocciate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("stichted"), &["stitched"]),
            (
                dictgen::InsensitiveStr::Ascii("nessesarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("amendmant"), &["amendment"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhopaths"),
                &["psychopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("cathredal"), &["cathedral"]),
            (dictgen::InsensitiveStr::Ascii("didactical"), &["didactic"]),
            (dictgen::InsensitiveStr::Ascii("medicinske"), &["medicine"]),
            (dictgen::InsensitiveStr::Ascii("applyig"), &["applying"]),
            (dictgen::InsensitiveStr::Ascii("hyrdogen"), &["hydrogen"]),
            (
                dictgen::InsensitiveStr::Ascii("negogtiates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filsystems"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterpice"),
                &["masterpiece"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependets"), &["dependents"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeting"),
                &["cheating", "cheering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technnology"),
                &["technology"],
            ),
            (dictgen::InsensitiveStr::Ascii("powerppc"), &["powerpc"]),
            (
                dictgen::InsensitiveStr::Ascii("digets"),
                &["digest", "digests"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranditions"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thorttling"),
                &["throttling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multivriate"),
                &["multivariate"],
            ),
            (dictgen::InsensitiveStr::Ascii("muticast"), &["multicast"]),
            (dictgen::InsensitiveStr::Ascii("ivoking"), &["invoking"]),
            (
                dictgen::InsensitiveStr::Ascii("innaccurately"),
                &["inaccurately"],
            ),
            (dictgen::InsensitiveStr::Ascii("brokened"), &["broken"]),
            (dictgen::InsensitiveStr::Ascii("smooting"), &["smoothing"]),
            (
                dictgen::InsensitiveStr::Ascii("activizion"),
                &["activision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifiied"),
                &["classified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranparently"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("indecate"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("paricular"), &["particular"]),
            (dictgen::InsensitiveStr::Ascii("inventry"), &["inventory"]),
            (
                dictgen::InsensitiveStr::Ascii("chidlbirth"),
                &["childbirth"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirective"), &["directive"]),
            (
                dictgen::InsensitiveStr::Ascii("apparrently"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assistante"),
                &["assistance"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuites"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("parntered"), &["partnered"]),
            (
                dictgen::InsensitiveStr::Ascii("exctractions"),
                &["extractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("theier"), &["their"]),
            (dictgen::InsensitiveStr::Ascii("inbraced"), &["embraced"]),
            (dictgen::InsensitiveStr::Ascii("datatyep"), &["datatype"]),
            (dictgen::InsensitiveStr::Ascii("darkets"), &["darkest"]),
            (dictgen::InsensitiveStr::Ascii("supplided"), &["supplied"]),
            (dictgen::InsensitiveStr::Ascii("collets"), &["collects"]),
            (dictgen::InsensitiveStr::Ascii("proccessed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("resonsive"), &["responsive"]),
            (dictgen::InsensitiveStr::Ascii("sttrict"), &["strict"]),
            (
                dictgen::InsensitiveStr::Ascii("interspursed"),
                &["interspersed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termanology"),
                &["terminology"],
            ),
            (dictgen::InsensitiveStr::Ascii("assymtote"), &["asymptote"]),
            (dictgen::InsensitiveStr::Ascii("dypshoria"), &["dysphoria"]),
            (dictgen::InsensitiveStr::Ascii("kepp"), &["keep", "kelp"]),
            (dictgen::InsensitiveStr::Ascii("nurturin"), &["nurturing"]),
            (dictgen::InsensitiveStr::Ascii("llike"), &["like"]),
            (dictgen::InsensitiveStr::Ascii("indcator"), &["indicator"]),
            (dictgen::InsensitiveStr::Ascii("crutchs"), &["crutches"]),
            (dictgen::InsensitiveStr::Ascii("eariler"), &["earlier"]),
            (dictgen::InsensitiveStr::Ascii("improvee"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("montanous"),
                &["mountainous"],
            ),
            (dictgen::InsensitiveStr::Ascii("concertas"), &["concerts"]),
            (
                dictgen::InsensitiveStr::Ascii("derective"),
                &["directive", "detective", "defective"],
            ),
            (dictgen::InsensitiveStr::Ascii("neolitic"), &["neolithic"]),
            (dictgen::InsensitiveStr::Ascii("diminishs"), &["diminishes"]),
            (
                dictgen::InsensitiveStr::Ascii("substituion"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ominscient"),
                &["omniscient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coallisions"),
                &["collisions"],
            ),
            (dictgen::InsensitiveStr::Ascii("worklords"), &["workloads"]),
            (dictgen::InsensitiveStr::Ascii("succeful"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("devidable"), &["divisible"]),
            (dictgen::InsensitiveStr::Ascii("expaded"), &["expanded"]),
            (dictgen::InsensitiveStr::Ascii("combinine"), &["combine"]),
            (dictgen::InsensitiveStr::Ascii("intity"), &["entity"]),
            (
                dictgen::InsensitiveStr::Ascii("construter"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("positoned"), &["positioned"]),
            (
                dictgen::InsensitiveStr::Ascii("neds"),
                &["needs", "ends", "nods"],
            ),
            (dictgen::InsensitiveStr::Ascii("pretador"), &["predator"]),
            (dictgen::InsensitiveStr::Ascii("actove"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("constractors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckecks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("demaned"),
                &["demand", "demeaned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhausively"),
                &["exhaustively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recgonition"),
                &["recognition"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiners"), &["pointers"]),
            (dictgen::InsensitiveStr::Ascii("thumnail"), &["thumbnail"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasformations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tatgerting"), &["targeting"]),
            (
                dictgen::InsensitiveStr::Ascii("paramerizing"),
                &["parameterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commet"),
                &["comet", "comment"],
            ),
            (dictgen::InsensitiveStr::Ascii("crowshays"), &["crochets"]),
            (
                dictgen::InsensitiveStr::Ascii("hyopthetically"),
                &["hypothetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amerliorate"),
                &["ameliorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparreling"),
                &["appareling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chancnel"),
                &["channel", "cancel"],
            ),
            (dictgen::InsensitiveStr::Ascii("upholstry"), &["upholstery"]),
            (
                dictgen::InsensitiveStr::Ascii("depracates"),
                &["deprecates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containe"),
                &["contain", "contained", "container", "contains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resuce"),
                &["reduce", "rescue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lengedaries"),
                &["legendaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepculating"),
                &["speculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatiors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("argmuments"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("experimtenal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("liscences"), &["licences"]),
            (dictgen::InsensitiveStr::Ascii("presreved"), &["preserved"]),
            (
                dictgen::InsensitiveStr::Ascii("pronted"),
                &["printed", "pronated", "pronged", "proned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negitiates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("dashboar"), &["dashboard"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretaton"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succees"),
                &["succeeds", "success"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automotize"),
                &["automotive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developping"),
                &["developing"],
            ),
            (dictgen::InsensitiveStr::Ascii("namnes"), &["names"]),
            (
                dictgen::InsensitiveStr::Ascii("accomdate"),
                &["accommodate"],
            ),
            (dictgen::InsensitiveStr::Ascii("execune"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("lerning"),
                &["learning", "leaning"],
            ),
            (dictgen::InsensitiveStr::Ascii("leyouts"), &["layouts"]),
            (dictgen::InsensitiveStr::Ascii("vegansim"), &["veganism"]),
            (dictgen::InsensitiveStr::Ascii("brefore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("banhsee"), &["banshee"]),
            (
                dictgen::InsensitiveStr::Ascii("formattring"),
                &["formatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorithies"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumlating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appointmets"),
                &["appointments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimizaing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttaching"), &["attaching"]),
            (dictgen::InsensitiveStr::Ascii("ciphe"), &["cipher"]),
            (
                dictgen::InsensitiveStr::Ascii("behvaiours"),
                &["behaviours"],
            ),
            (dictgen::InsensitiveStr::Ascii("cornmitted"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("phospher"), &["phosphor"]),
            (
                dictgen::InsensitiveStr::Ascii("resitances"),
                &["resistances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefrences"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("unompress"), &["uncompress"]),
            (dictgen::InsensitiveStr::Ascii("revewers"), &["reviewers"]),
            (
                dictgen::InsensitiveStr::Ascii("streatching"),
                &["stretching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmnets"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fot"),
                &["for", "fit", "dot", "rot", "cot", "got", "tot", "fog"],
            ),
            (dictgen::InsensitiveStr::Ascii("coordine"), &["coordinate"]),
            (dictgen::InsensitiveStr::Ascii("trasfers"), &["transfers"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoziations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verificaion"),
                &["verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preditable"),
                &["predictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("referenes"), &["references"]),
            (dictgen::InsensitiveStr::Ascii("shoukd"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("inturrupting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catergorizations"),
                &["categorizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliantro"), &["cilantro"]),
            (dictgen::InsensitiveStr::Ascii("gorup"), &["group"]),
            (dictgen::InsensitiveStr::Ascii("develoed"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("priortisations"),
                &["prioritisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimpplementation"),
                &["reimplementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unadvertantly"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionaryy"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bechmarking"),
                &["benchmarking"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressourced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("arbritarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("examle"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("facourites"),
                &["favourites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civalization"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("machins"), &["machines"]),
            (dictgen::InsensitiveStr::Ascii("proppely"), &["properly"]),
            (dictgen::InsensitiveStr::Ascii("oxzillary"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("commpare"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("indroductions"),
                &["introductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("defintley"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("ccustom"), &["custom"]),
            (
                dictgen::InsensitiveStr::Ascii("transformered"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburguers"),
                &["cheeseburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifikant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficillifed"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("neigbors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("olypmics"), &["olympics"]),
            (
                dictgen::InsensitiveStr::Ascii("condenmation"),
                &["condemnation"],
            ),
            (dictgen::InsensitiveStr::Ascii("remval"), &["removal"]),
            (
                dictgen::InsensitiveStr::Ascii("homecomeing"),
                &["homecoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esolution"),
                &["resolution", "solution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentatly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diverisons"),
                &["diversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoropeat"),
                &["autorepeat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintutive"),
                &["unintuitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrubed"),
                &["unsubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexperssions"),
                &["subexpressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissble"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("sonething"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("torchligth"),
                &["torchlight"],
            ),
            (dictgen::InsensitiveStr::Ascii("uffered"), &["buffered"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenades"),
                &["concatenates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genereration"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commandeur"),
                &["commanders"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectoned"), &["sectioned"]),
            (
                dictgen::InsensitiveStr::Ascii("economisiti"),
                &["economist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowledeable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("documen"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("reprehensable"),
                &["reprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("wintesses"), &["witnesses"]),
            (dictgen::InsensitiveStr::Ascii("aske"), &["ask"]),
            (dictgen::InsensitiveStr::Ascii("corrollary"), &["corollary"]),
            (dictgen::InsensitiveStr::Ascii("ocasioned"), &["occasioned"]),
            (
                dictgen::InsensitiveStr::Ascii("alterntives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("serivce"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("empass"), &["impasse"]),
            (dictgen::InsensitiveStr::Ascii("symptums"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("turorial"), &["tutorial"]),
            (dictgen::InsensitiveStr::Ascii("exposere"), &["exposure"]),
            (
                dictgen::InsensitiveStr::Ascii("periwinkie"),
                &["periwinkle"],
            ),
            (dictgen::InsensitiveStr::Ascii("separtely"), &["separately"]),
            (dictgen::InsensitiveStr::Ascii("calfes"), &["calves"]),
            (
                dictgen::InsensitiveStr::Ascii("launhing"),
                &["launching", "laughing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conyak"), &["cognac"]),
            (
                dictgen::InsensitiveStr::Ascii("playgorund"),
                &["playground"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annonciators"),
                &["annunciators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scateboarding"),
                &["skateboarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("conxtext"), &["context"]),
            (dictgen::InsensitiveStr::Ascii("emualtors"), &["emulators"]),
            (
                dictgen::InsensitiveStr::Ascii("intenionally"),
                &["intentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("deamenor"), &["demeanor"]),
            (dictgen::InsensitiveStr::Ascii("premines"), &["premise"]),
            (dictgen::InsensitiveStr::Ascii("awailable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("competidor"),
                &["competitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("choesive"), &["cohesive"]),
            (dictgen::InsensitiveStr::Ascii("paitience"), &["patience"]),
            (dictgen::InsensitiveStr::Ascii("warrany"), &["warranty"]),
            (dictgen::InsensitiveStr::Ascii("drinkes"), &["drinkers"]),
            (dictgen::InsensitiveStr::Ascii("threadened"), &["threaded"]),
            (dictgen::InsensitiveStr::Ascii("reitres"), &["retires"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchrounously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defualtdict"),
                &["defaultdict"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envireonment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlikley"), &["unlikely"]),
            (dictgen::InsensitiveStr::Ascii("zorbist"), &["zobrist"]),
            (
                dictgen::InsensitiveStr::Ascii("chank"),
                &[
                    "clank", "crank", "chalk", "chant", "chink", "chunk", "shank", "hank", "chonk",
                    "thank",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("completly"), &["completely"]),
            (dictgen::InsensitiveStr::Ascii("contatc"), &["contact"]),
            (dictgen::InsensitiveStr::Ascii("dialig"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("neuronas"), &["neurons"]),
            (
                dictgen::InsensitiveStr::Ascii("funcioning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thrading"),
                &["threading", "trading"],
            ),
            (dictgen::InsensitiveStr::Ascii("temporyry"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("repeadet"), &["repeated"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigeous"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("expectred"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("sastifies"), &["satisfies"]),
            (dictgen::InsensitiveStr::Ascii("renig"), &["renege"]),
            (
                dictgen::InsensitiveStr::Ascii("buildt"),
                &["built", "build"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessiate"),
                &["necessitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("asser"), &["assert"]),
            (dictgen::InsensitiveStr::Ascii("obay"), &["obey"]),
            (
                dictgen::InsensitiveStr::Ascii("optionnally"),
                &["optionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("adusting"), &["adjusting"]),
            (dictgen::InsensitiveStr::Ascii("defintly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("asisting"), &["assisting"]),
            (dictgen::InsensitiveStr::Ascii("addjusting"), &["adjusting"]),
            (dictgen::InsensitiveStr::Ascii("spotifiy"), &["spotify"]),
            (
                dictgen::InsensitiveStr::Ascii("assistence"),
                &["assistance"],
            ),
            (dictgen::InsensitiveStr::Ascii("pospone"), &["postpone"]),
            (
                dictgen::InsensitiveStr::Ascii("precedures"),
                &["procedures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strawbrerry"),
                &["strawberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("uppstreams"), &["upstreams"]),
            (
                dictgen::InsensitiveStr::Ascii("thught"),
                &["thought", "taught"],
            ),
            (dictgen::InsensitiveStr::Ascii("whike"), &["while"]),
            (
                dictgen::InsensitiveStr::Ascii("exacuting"),
                &["executing", "exacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorectly"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ojection"), &["objection"]),
            (dictgen::InsensitiveStr::Ascii("clossion"), &["collision"]),
            (dictgen::InsensitiveStr::Ascii("depdendent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulaton"),
                &["manipulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tommorrow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("copmleting"),
                &["completing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbtirary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilzing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotication"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("defautled"), &["defaulted"]),
            (dictgen::InsensitiveStr::Ascii("adapated"), &["adapted"]),
            (
                dictgen::InsensitiveStr::Ascii("inteacting"),
                &["interacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("synidcate"), &["syndicate"]),
            (
                dictgen::InsensitiveStr::Ascii("initializaes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exectuables"),
                &["executable", "executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("remanins"), &["remains"]),
            (dictgen::InsensitiveStr::Ascii("thriteen"), &["thirteen"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaisation"),
                &["generalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("specail"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("occaisionally"),
                &["occasionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificato"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technnician"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intraverted"),
                &["introverted"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaibility"), &["viability"]),
            (dictgen::InsensitiveStr::Ascii("getlael"), &["getlabel"]),
            (
                dictgen::InsensitiveStr::Ascii("provisoner"),
                &["provisioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspends"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementatin"),
                &["implementations", "implementation", "implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("donwgrading"),
                &["downgrading"],
            ),
            (dictgen::InsensitiveStr::Ascii("histogam"), &["histogram"]),
            (
                dictgen::InsensitiveStr::Ascii("fringeworthy"),
                &["cringeworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iteractively"),
                &["interactively", "iteratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("stricly"), &["strictly"]),
            (dictgen::InsensitiveStr::Ascii("expetancy"), &["expectancy"]),
            (dictgen::InsensitiveStr::Ascii("idesa"), &["ideas", "ides"]),
            (dictgen::InsensitiveStr::Ascii("loacle"), &["locale"]),
            (dictgen::InsensitiveStr::Ascii("obayed"), &["obeyed"]),
            (dictgen::InsensitiveStr::Ascii("satric"), &["satiric"]),
            (
                dictgen::InsensitiveStr::Ascii("multuplicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("prposed"), &["proposed"]),
            (dictgen::InsensitiveStr::Ascii("styrofom"), &["styrofoam"]),
            (
                dictgen::InsensitiveStr::Ascii("impresison"),
                &["impressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blackbarry"),
                &["blackberry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deploymnet"),
                &["deployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("firmawre"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("balnk"), &["blank"]),
            (
                dictgen::InsensitiveStr::Ascii("communciated"),
                &["communicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("animaitions"),
                &["animations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inherents"),
                &["inherits", "inherent"],
            ),
            (dictgen::InsensitiveStr::Ascii("unopenend"), &["unopened"]),
            (
                dictgen::InsensitiveStr::Ascii("disciplen"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("islamisist"), &["islamist"]),
            (dictgen::InsensitiveStr::Ascii("susncreen"), &["sunscreen"]),
            (dictgen::InsensitiveStr::Ascii("gunniess"), &["guinness"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscritptions"),
                &["unsubscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("organzing"), &["organizing"]),
            (dictgen::InsensitiveStr::Ascii("displayied"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("corasir"), &["corsair"]),
            (
                dictgen::InsensitiveStr::Ascii("appropropreate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("considerd"), &["considered"]),
            (dictgen::InsensitiveStr::Ascii("disalbe"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("imeadiatly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participantes"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orhtogonal"),
                &["orthogonal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("warpper"),
                &["warper", "wrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("interrest"), &["interest"]),
            (dictgen::InsensitiveStr::Ascii("coutnown"), &["countdown"]),
            (
                dictgen::InsensitiveStr::Ascii("upsrteamed"),
                &["upstreamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifgurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unattented"),
                &["unattended"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaliting"), &["evaluating"]),
            (
                dictgen::InsensitiveStr::Ascii("eliminatin"),
                &["elimination", "eliminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undermindes"),
                &["undermines"],
            ),
            (dictgen::InsensitiveStr::Ascii("skwalked"), &["squawked"]),
            (
                dictgen::InsensitiveStr::Ascii("generalife"),
                &["generalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rewirtable"),
                &["rewritable"],
            ),
            (dictgen::InsensitiveStr::Ascii("connecor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("mechancially"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semicondictor"),
                &["semiconductor"],
            ),
            (dictgen::InsensitiveStr::Ascii("foreginer"), &["foreigner"]),
            (
                dictgen::InsensitiveStr::Ascii("implementeations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completaste"),
                &["completes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuent"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservaton"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoiurcing"),
                &["resourcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("encorded"), &["encoded"]),
            (
                dictgen::InsensitiveStr::Ascii("presearvation"),
                &["preservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("athiestic"), &["atheistic"]),
            (dictgen::InsensitiveStr::Ascii("asssumes"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("unassinging"),
                &["unassigning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconencts"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenth"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("dezert"), &["dessert"]),
            (
                dictgen::InsensitiveStr::Ascii("partirioning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilzed"), &["utilized"]),
            (dictgen::InsensitiveStr::Ascii("storeble"), &["storable"]),
            (dictgen::InsensitiveStr::Ascii("specialtys"), &["specials"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeards"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherrits"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("apshalt"), &["asphalt"]),
            (
                dictgen::InsensitiveStr::Ascii("generiously"),
                &["generously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miminalist"),
                &["minimalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("accupts"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("mainfested"),
                &["manifested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmitting"),
                &["committing"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrypter"), &["encryptor"]),
            (dictgen::InsensitiveStr::Ascii("redy"), &["ready", "red"]),
            (
                dictgen::InsensitiveStr::Ascii("functionaity"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docmenting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrarys"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("brosing"), &["browsing"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosincracies"),
                &["idiosyncrasies"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostories"), &["histories"]),
            (dictgen::InsensitiveStr::Ascii("intervas"), &["interval"]),
            (
                dictgen::InsensitiveStr::Ascii("permantently"),
                &["permanently"],
            ),
            (dictgen::InsensitiveStr::Ascii("princible"), &["principle"]),
            (dictgen::InsensitiveStr::Ascii("egal"), &["equal"]),
            (dictgen::InsensitiveStr::Ascii("spoiles"), &["spoilers"]),
            (
                dictgen::InsensitiveStr::Ascii("perssiously"),
                &["preciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrumentaion"),
                &["instrumentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unalbe"), &["unable"]),
            (
                dictgen::InsensitiveStr::Ascii("experimtening"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approoximated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crowdsigna"),
                &["crowdsignal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supresing"),
                &["suppressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranported"),
                &["transported"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesless"), &["useless"]),
            (
                dictgen::InsensitiveStr::Ascii("logorithmic"),
                &["logarithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normallisation"),
                &["normalisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terains"),
                &["terrains", "trains"],
            ),
            (dictgen::InsensitiveStr::Ascii("instruted"), &["instructed"]),
            (dictgen::InsensitiveStr::Ascii("cilcks"), &["clicks"]),
            (dictgen::InsensitiveStr::Ascii("tunnles"), &["tunnels"]),
            (
                dictgen::InsensitiveStr::Ascii("roatory"),
                &["rotatory", "oratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("utlizing"), &["utilizing"]),
            (dictgen::InsensitiveStr::Ascii("breserker"), &["berserker"]),
            (dictgen::InsensitiveStr::Ascii("soudns"), &["sounds"]),
            (dictgen::InsensitiveStr::Ascii("avoinding"), &["avoiding"]),
            (
                dictgen::InsensitiveStr::Ascii("intelligable"),
                &["intelligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("dafualts"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("muscicians"), &["musicians"]),
            (
                dictgen::InsensitiveStr::Ascii("transgengered"),
                &["transgendered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimened"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simily"),
                &["simile", "smiley", "simply", "similarly"],
            ),
            (dictgen::InsensitiveStr::Ascii("tansmit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("behavrioural"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("squirte"), &["squirtle"]),
            (dictgen::InsensitiveStr::Ascii("concsious"), &["conscious"]),
            (
                dictgen::InsensitiveStr::Ascii("syntehtically"),
                &["synthetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("oakereous"), &["ocherous"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailible"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommdends"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("aslias"), &["alias"]),
            (
                dictgen::InsensitiveStr::Ascii("quanitization"),
                &["quantization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copilations"),
                &["compilations", "copulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelarating"),
                &["accelerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("untintuitive"),
                &["unintuitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquanited"),
                &["acquainted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preprocesssing"),
                &["preprocessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsmitting"),
                &["transmitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("conifigure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("addessses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("concatination"),
                &["contamination", "concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentciate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("correctin"), &["correction"]),
            (
                dictgen::InsensitiveStr::Ascii("commericals"),
                &["commercials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metorpolitan"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homegeneous"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("socity"), &["society"]),
            (
                dictgen::InsensitiveStr::Ascii("mutilthread"),
                &["multithread"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playtrhoughs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghtend"), &["lengthened"]),
            (
                dictgen::InsensitiveStr::Ascii("averged"),
                &["averaged", "avenged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custoisation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incemented"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacalbes"),
                &["replaceables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrealsitic"),
                &["unrealistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("rococco"), &["rococo"]),
            (dictgen::InsensitiveStr::Ascii("ambiguious"), &["ambiguous"]),
            (dictgen::InsensitiveStr::Ascii("cofrimed"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("biuld"), &["build"]),
            (dictgen::InsensitiveStr::Ascii("seriou"), &["serious"]),
            (dictgen::InsensitiveStr::Ascii("discuassed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("diferent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("derectors"),
                &["detectors", "directors", "erectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("coductor"), &["conductor"]),
            (dictgen::InsensitiveStr::Ascii("dissaster"), &["disaster"]),
            (dictgen::InsensitiveStr::Ascii("reboudns"), &["rebounds"]),
            (dictgen::InsensitiveStr::Ascii("hapends"), &["happens"]),
            (
                dictgen::InsensitiveStr::Ascii("asendancey"),
                &["ascendancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinced"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("essentual"), &["essential"]),
            (dictgen::InsensitiveStr::Ascii("lanched"), &["launched"]),
            (
                dictgen::InsensitiveStr::Ascii("evalutaion"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("feminie"), &["feminine"]),
            (dictgen::InsensitiveStr::Ascii("moutned"), &["mounted"]),
            (dictgen::InsensitiveStr::Ascii("inquring"), &["inquiring"]),
            (dictgen::InsensitiveStr::Ascii("wierdly"), &["weirdly"]),
            (dictgen::InsensitiveStr::Ascii("priesthod"), &["priesthood"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumeably"),
                &["presumably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negosiation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("satruday"), &["saturday"]),
            (dictgen::InsensitiveStr::Ascii("listiner"), &["listener"]),
            (
                dictgen::InsensitiveStr::Ascii("calssified"),
                &["classified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extragavant"),
                &["extravagant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocorrct"),
                &["autocorrect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documeting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimitated"),
                &["intimidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("texured"), &["textured"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinastes"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxiumum"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("amiguously"),
                &["ambiguously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reitterated"),
                &["reiterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("enouf"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("commeted"),
                &["commented", "competed"],
            ),
            (dictgen::InsensitiveStr::Ascii("vrifier"), &["verifier"]),
            (dictgen::InsensitiveStr::Ascii("sproon"), &["spoon"]),
            (dictgen::InsensitiveStr::Ascii("diagnoes"), &["diagnose"]),
            (dictgen::InsensitiveStr::Ascii("retuend"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("erorrs"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("konversation"),
                &["conversation"],
            ),
            (dictgen::InsensitiveStr::Ascii("plently"), &["plenty"]),
            (dictgen::InsensitiveStr::Ascii("marcros"), &["macros"]),
            (dictgen::InsensitiveStr::Ascii("bevcause"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("aussian"),
                &["gaussian", "russian", "austrian"],
            ),
            (dictgen::InsensitiveStr::Ascii("architcts"), &["architects"]),
            (dictgen::InsensitiveStr::Ascii("reatiler"), &["retailer"]),
            (
                dictgen::InsensitiveStr::Ascii("alchoholism"),
                &["alcoholism"],
            ),
            (dictgen::InsensitiveStr::Ascii("requiest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("astronimical"),
                &["astronomical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandenavian"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomizing"),
                &["randomizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wieghted"),
                &["weighted", "weighed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedependents"),
                &["dependents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juggetnaut"),
                &["juggernaut"],
            ),
            (dictgen::InsensitiveStr::Ascii("melodieuse"), &["melodies"]),
            (
                dictgen::InsensitiveStr::Ascii("mension"),
                &["mansion", "mention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeliefable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("paackage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("splitner"), &["splinter"]),
            (dictgen::InsensitiveStr::Ascii("succssion"), &["succession"]),
            (dictgen::InsensitiveStr::Ascii("volatages"), &["voltages"]),
            (dictgen::InsensitiveStr::Ascii("gropuing"), &["grouping"]),
            (
                dictgen::InsensitiveStr::Ascii("individuall"),
                &["individual", "individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("versionm"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("reguirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excepionally"),
                &["exceptionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonished"),
                &["demolished"],
            ),
            (dictgen::InsensitiveStr::Ascii("crypo"), &["crypto"]),
            (dictgen::InsensitiveStr::Ascii("argumen"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("publishs"), &["publishes"]),
            (dictgen::InsensitiveStr::Ascii("bouqet"), &["bouquet"]),
            (
                dictgen::InsensitiveStr::Ascii("selektable"),
                &["selectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propietary"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethibng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("probles"),
                &["probes", "problems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innoculating"),
                &["inoculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgning"), &["designing"]),
            (
                dictgen::InsensitiveStr::Ascii("accelarated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omnipotant"),
                &["omnipotent"],
            ),
            (dictgen::InsensitiveStr::Ascii("currntly"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("disastroso"),
                &["disastrous"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifx"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("concurently"),
                &["concurrently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chekcboxes"),
                &["checkboxes"],
            ),
            (dictgen::InsensitiveStr::Ascii("uprgade"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("positionnes"),
                &["positions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superflouse"),
                &["superfluous"],
            ),
            (dictgen::InsensitiveStr::Ascii("convenant"), &["covenant"]),
            (dictgen::InsensitiveStr::Ascii("talbian"), &["taliban"]),
            (
                dictgen::InsensitiveStr::Ascii("compreses"),
                &["compress", "compresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaion"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("contiained"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("terrformed"),
                &["terraformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaliates"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("recrusevly"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("pthred"), &["pthread"]),
            (dictgen::InsensitiveStr::Ascii("asnwerer"), &["answerer"]),
            (
                dictgen::InsensitiveStr::Ascii("overwieght"),
                &["overweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contants"),
                &["constants", "contents"],
            ),
            (dictgen::InsensitiveStr::Ascii("snapshop"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("srttings"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("nastershem"),
                &["nasturtium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temprarily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presistance"),
                &["persistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coomanding"),
                &["commanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("timstamp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("spinock"), &["spinlock"]),
            (
                dictgen::InsensitiveStr::Ascii("hadshaking"),
                &["handshaking", "headshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("gatherins"), &["gatherings"]),
            (dictgen::InsensitiveStr::Ascii("samuari"), &["samurai"]),
            (
                dictgen::InsensitiveStr::Ascii("reuired"),
                &["required", "rewired"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("futurestic"),
                &["futuristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculcation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuqire"),
                &["required", "require"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufficates"),
                &["suffocates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worskapces"),
                &["workspaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confurmation"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalistes"),
                &["capitalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ethniticies"),
                &["ethnicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressolved"), &["resolved"]),
            (
                dictgen::InsensitiveStr::Ascii("compromizing"),
                &["compromising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meassaged"),
                &["messaged", "massaged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("keybindining"),
                &["keybinding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phychologically"),
                &["psychologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reademe"), &["readme"]),
            (
                dictgen::InsensitiveStr::Ascii("territorrial"),
                &["territorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("memmorys"), &["memories"]),
            (
                dictgen::InsensitiveStr::Ascii("sodo"),
                &["dodo", "sod", "soda", "sods", "solo", "sudo"],
            ),
            (dictgen::InsensitiveStr::Ascii("authrority"), &["authority"]),
            (dictgen::InsensitiveStr::Ascii("survery"), &["survey"]),
            (dictgen::InsensitiveStr::Ascii("funtion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("sarcasim"), &["sarcasm"]),
            (
                dictgen::InsensitiveStr::Ascii("epxression"),
                &["expression"],
            ),
            (dictgen::InsensitiveStr::Ascii("protoypes"), &["prototypes"]),
            (dictgen::InsensitiveStr::Ascii("finihed"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("decmial"), &["decimal"]),
            (dictgen::InsensitiveStr::Ascii("interit"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("intneral"), &["internal"]),
            (
                dictgen::InsensitiveStr::Ascii("probabily"),
                &["probability", "probably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cahing"),
                &["caching", "aching"],
            ),
            (dictgen::InsensitiveStr::Ascii("valuling"), &["valuing"]),
            (dictgen::InsensitiveStr::Ascii("tylenool"), &["tylenol"]),
            (
                dictgen::InsensitiveStr::Ascii("classsifiers"),
                &["classifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("actvating"), &["activating"]),
            (dictgen::InsensitiveStr::Ascii("joytick"), &["joystick"]),
            (dictgen::InsensitiveStr::Ascii("reflexs"), &["reflexes"]),
            (dictgen::InsensitiveStr::Ascii("currenet"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("greenlad"), &["greenland"]),
            (dictgen::InsensitiveStr::Ascii("conection"), &["connection"]),
            (
                dictgen::InsensitiveStr::Ascii("nondeteministic"),
                &["nondeterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("domonstration"),
                &["demonstration"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpshiba"), &["toshiba"]),
            (dictgen::InsensitiveStr::Ascii("cricical"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("connot"), &["cannot"]),
            (
                dictgen::InsensitiveStr::Ascii("distinshish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("konw"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("replacemtn"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incrieasing"),
                &["increasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necesseraly"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmnet"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("channges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("specififcations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularizin"),
                &["regularizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adpatation"),
                &["adaptation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exorbidant"),
                &["exorbitant"],
            ),
            (dictgen::InsensitiveStr::Ascii("portected"), &["protected"]),
            (dictgen::InsensitiveStr::Ascii("repearedly"), &["repealed"]),
            (dictgen::InsensitiveStr::Ascii("brocolee"), &["broccoli"]),
            (
                dictgen::InsensitiveStr::Ascii("dependdency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecesarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialzing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpatible"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("linkely"), &["likely"]),
            (dictgen::InsensitiveStr::Ascii("defenate"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("frustrum"), &["frustum"]),
            (
                dictgen::InsensitiveStr::Ascii("illistration"),
                &["illustration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifigures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compensantion"),
                &["compensation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambassodor"),
                &["ambassador"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentates"),
                &["concentrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("translateing"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("oerflow"), &["overflow"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsstrapped"),
                &["bootstrapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("allegries"), &["allergies"]),
            (dictgen::InsensitiveStr::Ascii("fotgot"), &["forgot"]),
            (
                dictgen::InsensitiveStr::Ascii("pasturised"),
                &["pasteurised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsisteny"),
                &["inconsistency", "inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("thtreads"), &["threads"]),
            (
                dictgen::InsensitiveStr::Ascii("dissassembling"),
                &["disassembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("togueter"), &["together"]),
            (
                dictgen::InsensitiveStr::Ascii("rememberance"),
                &["remembrance"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreration"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("confirmedd"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("intitally"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("intamite"), &["intimate"]),
            (
                dictgen::InsensitiveStr::Ascii("marshamllows"),
                &["marshmallows"],
            ),
            (dictgen::InsensitiveStr::Ascii("generted"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("themplating"),
                &["templating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccopes"), &["scopes"]),
            (dictgen::InsensitiveStr::Ascii("virutes"), &["virtues"]),
            (
                dictgen::InsensitiveStr::Ascii("classsifying"),
                &["classifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("worskpace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("assumnes"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("golakeeper"),
                &["goalkeeper"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlimitied"), &["unlimited"]),
            (dictgen::InsensitiveStr::Ascii("wathcmen"), &["watchmen"]),
            (dictgen::InsensitiveStr::Ascii("inintials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("redistirbution"),
                &["redistribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammendment"), &["amendment"]),
            (
                dictgen::InsensitiveStr::Ascii("bodybuiling"),
                &["bodybuilding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siplifiers"),
                &["simplifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfering"),
                &["transferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcoser"),
                &["transcoder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conchances"),
                &["consciences"],
            ),
            (dictgen::InsensitiveStr::Ascii("adiacent"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("handshakng"),
                &["handshaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rimanisser"),
                &["reminiscer"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhancd"), &["enhanced"]),
            (
                dictgen::InsensitiveStr::Ascii("frinedzoned"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("rehersing"), &["rehearsing"]),
            (
                dictgen::InsensitiveStr::Ascii("alterates"),
                &["alters", "alternates"],
            ),
            (dictgen::InsensitiveStr::Ascii("birgade"), &["brigade"]),
            (
                dictgen::InsensitiveStr::Ascii("attachemnt"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometinh"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("cppp"), &["cpp"]),
            (
                dictgen::InsensitiveStr::Ascii("volunatrily"),
                &["voluntarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("repertoir"), &["repertoire"]),
            (
                dictgen::InsensitiveStr::Ascii("enviromental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("wokrs"), &["works", "woks"]),
            (
                dictgen::InsensitiveStr::Ascii("utillization"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitiation"),
                &["limitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("generare"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("transaltor"),
                &["translator"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowiong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("troling"), &["trolling"]),
            (dictgen::InsensitiveStr::Ascii("chizzled"), &["chiseled"]),
            (dictgen::InsensitiveStr::Ascii("humurous"), &["humorous"]),
            (dictgen::InsensitiveStr::Ascii("releating"), &["relating"]),
            (dictgen::InsensitiveStr::Ascii("solwed"), &["solved"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnreable"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("penaltis"), &["penalties"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatiability"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherting"), &["inheriting"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalistics"),
                &["nationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchngs"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("sgement"), &["segment"]),
            (
                dictgen::InsensitiveStr::Ascii("tolerences"),
                &["tolerances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dublicating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("isse"), &["issue"]),
            (
                dictgen::InsensitiveStr::Ascii("secrataries"),
                &["secretaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("epmty"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstraits"),
                &["demonstrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecors"), &["selectors"]),
            (
                dictgen::InsensitiveStr::Ascii("accelarates"),
                &["accelerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roleplaing"),
                &["roleplaying"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyizing"), &["analyzing"]),
            (dictgen::InsensitiveStr::Ascii("avriables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("apoligized"),
                &["apologized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("universale"), &["universe"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitralily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confermation"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspendingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experementations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("servcing"),
                &["servicing", "serving"],
            ),
            (dictgen::InsensitiveStr::Ascii("rentors"), &["renters"]),
            (dictgen::InsensitiveStr::Ascii("calrkson"), &["clarkson"]),
            (
                dictgen::InsensitiveStr::Ascii("alvorithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("enywhere"), &["anywhere"]),
            (
                dictgen::InsensitiveStr::Ascii("assymtotic"),
                &["asymptotic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructivo"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accompagnies"),
                &["accompanies"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriopted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("wraped"),
                &["wrapped", "warped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surley"),
                &["surly", "surely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priotities"),
                &["priorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("isntall"), &["install"]),
            (dictgen::InsensitiveStr::Ascii("comptible"), &["compatible"]),
            (dictgen::InsensitiveStr::Ascii("scrren"), &["screen"]),
            (dictgen::InsensitiveStr::Ascii("oeprands"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("sterams"), &["streams"]),
            (dictgen::InsensitiveStr::Ascii("revovles"), &["revolves"]),
            (dictgen::InsensitiveStr::Ascii("wrapepr"), &["wrapper"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("esitmated"), &["estimated"]),
            (
                dictgen::InsensitiveStr::Ascii("molestating"),
                &["molestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("anyhing"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("moust"), &["mouse", "most"]),
            (dictgen::InsensitiveStr::Ascii("selecetors"), &["selectors"]),
            (
                dictgen::InsensitiveStr::Ascii("aassignments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("paritition"), &["partition"]),
            (dictgen::InsensitiveStr::Ascii("unmaanaged"), &["unmanaged"]),
            (dictgen::InsensitiveStr::Ascii("flatterd"), &["flattered"]),
            (dictgen::InsensitiveStr::Ascii("counds"), &["counts"]),
            (
                dictgen::InsensitiveStr::Ascii("reparametrized"),
                &["reparameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("retirment"), &["retirement"]),
            (
                dictgen::InsensitiveStr::Ascii("organsiing"),
                &["organising"],
            ),
            (dictgen::InsensitiveStr::Ascii("becase"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("propegators"),
                &["propagators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegoziations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("andtoids"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("telelevision"),
                &["television"],
            ),
            (dictgen::InsensitiveStr::Ascii("fillament"), &["filament"]),
            (dictgen::InsensitiveStr::Ascii("sanctiond"), &["sanctioned"]),
            (dictgen::InsensitiveStr::Ascii("chracter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("rocess"),
                &["process", "recess"],
            ),
            (dictgen::InsensitiveStr::Ascii("previouly"), &["previously"]),
            (dictgen::InsensitiveStr::Ascii("averageadi"), &["averaged"]),
            (
                dictgen::InsensitiveStr::Ascii("subseuqent"),
                &["subsequent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teleproting"),
                &["teleporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scritpures"),
                &["scriptures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punshiments"),
                &["punishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("highschol"), &["highschool"]),
            (dictgen::InsensitiveStr::Ascii("citerion"), &["criterion"]),
            (
                dictgen::InsensitiveStr::Ascii("diffreences"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("wsee"), &["see"]),
            (dictgen::InsensitiveStr::Ascii("cumpus"), &["compass"]),
            (
                dictgen::InsensitiveStr::Ascii("masqueradin"),
                &["masquerading"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethijng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("persecusion"),
                &["persecution"],
            ),
            (dictgen::InsensitiveStr::Ascii("teenages"), &["teenagers"]),
            (
                dictgen::InsensitiveStr::Ascii("defintition"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlaisations"),
                &["generalisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cherck"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("recconects"),
                &["reconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unuxpected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combie"),
                &["combine", "zombie"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknowin"), &["unknowing"]),
            (
                dictgen::InsensitiveStr::Ascii("prefence"),
                &["preference", "presence", "pretence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underliying"),
                &["underlying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srting"),
                &["string", "sorting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dataype"), &["datatype"]),
            (
                dictgen::InsensitiveStr::Ascii("duplictaed"),
                &["duplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("igzort"), &["exhort"]),
            (
                dictgen::InsensitiveStr::Ascii("idenifiable"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("andoirds"), &["androids"]),
            (dictgen::InsensitiveStr::Ascii("guarnetee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhing"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomalises"), &["normalises"]),
            (
                dictgen::InsensitiveStr::Ascii("intercepticons"),
                &["interceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persuasivo"),
                &["persuasion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggressivly"),
                &["aggressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interveening"),
                &["intervening"],
            ),
            (dictgen::InsensitiveStr::Ascii("gracefuly"), &["gracefully"]),
            (
                dictgen::InsensitiveStr::Ascii("subtracion"),
                &["subtraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communitcations"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciphersutes"),
                &["ciphersuites"],
            ),
            (dictgen::InsensitiveStr::Ascii("docsund"), &["dachshund"]),
            (dictgen::InsensitiveStr::Ascii("credists"), &["credits"]),
            (dictgen::InsensitiveStr::Ascii("purposs"), &["purpose"]),
            (
                dictgen::InsensitiveStr::Ascii("uncondisionnaly"),
                &["unconditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("dedly"), &["deadly"]),
            (
                dictgen::InsensitiveStr::Ascii("consquently"),
                &["consequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbelieveble"),
                &["unbelievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aditionnally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heterozygyous"),
                &["heterozygous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charimastic"),
                &["charismatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilties"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("avalaibles"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("nightley"), &["nightly"]),
            (
                dictgen::InsensitiveStr::Ascii("executiong"),
                &["execution", "executing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conractions"),
                &["contractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("quieries"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("thermodymanics"),
                &["thermodynamics"],
            ),
            (dictgen::InsensitiveStr::Ascii("extrudin"), &["extruding"]),
            (dictgen::InsensitiveStr::Ascii("loacles"), &["locales"]),
            (dictgen::InsensitiveStr::Ascii("steathly"), &["stealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("newlsetter"),
                &["newsletter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counry"),
                &["country", "county"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commneting"),
                &["commenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("plecing"), &["placing"]),
            (dictgen::InsensitiveStr::Ascii("compusing"), &["computing"]),
            (dictgen::InsensitiveStr::Ascii("reprhase"), &["rephrase"]),
            (dictgen::InsensitiveStr::Ascii("liberalim"), &["liberalism"]),
            (dictgen::InsensitiveStr::Ascii("aboslved"), &["absolved"]),
            (dictgen::InsensitiveStr::Ascii("unqoute"), &["unquote"]),
            (dictgen::InsensitiveStr::Ascii("muiltiple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("transulcent"),
                &["translucent"],
            ),
            (dictgen::InsensitiveStr::Ascii("exicute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("deamons"), &["daemons"]),
            (dictgen::InsensitiveStr::Ascii("nautres"), &["natures"]),
            (dictgen::InsensitiveStr::Ascii("trinkes"), &["trinkets"]),
            (dictgen::InsensitiveStr::Ascii("arcticle"), &["article"]),
            (dictgen::InsensitiveStr::Ascii("confesso"), &["confession"]),
            (dictgen::InsensitiveStr::Ascii("keybard"), &["keyboard"]),
            (
                dictgen::InsensitiveStr::Ascii("uniliterally"),
                &["unilaterally"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurited"), &["recruited"]),
            (dictgen::InsensitiveStr::Ascii("implantase"), &["implants"]),
            (
                dictgen::InsensitiveStr::Ascii("repbulican"),
                &["republican"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aportionable"),
                &["apportionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("segemnts"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("zookeenei"), &["zucchinis"]),
            (
                dictgen::InsensitiveStr::Ascii("mathematicans"),
                &["mathematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("menally"), &["mentally"]),
            (
                dictgen::InsensitiveStr::Ascii("cyclindrical"),
                &["cylindrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comminucation"),
                &["communications", "communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transisioning"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particualar"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("vinicity"), &["vicinity"]),
            (dictgen::InsensitiveStr::Ascii("ratatui"), &["ratatouille"]),
            (
                dictgen::InsensitiveStr::Ascii("mathimatically"),
                &["mathematically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetifications"),
                &["certifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapointing"),
                &["disappointing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphapeicall"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("mobiltiy"), &["mobility"]),
            (
                dictgen::InsensitiveStr::Ascii("cessationalism"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiavers"), &["waivers"]),
            (
                dictgen::InsensitiveStr::Ascii("trals"),
                &["trails", "trials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certific"),
                &["certificate", "certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveyrone"), &["everyone"]),
            (
                dictgen::InsensitiveStr::Ascii("euqivalents"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extepecting"),
                &["expecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribtutions"),
                &["attributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserailisation"),
                &["deserialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminstering"),
                &["administering"],
            ),
            (dictgen::InsensitiveStr::Ascii("soliders"), &["soldiers"]),
            (dictgen::InsensitiveStr::Ascii("sirynges"), &["syringes"]),
            (
                dictgen::InsensitiveStr::Ascii("patren"),
                &["patron", "pattern"],
            ),
            (dictgen::InsensitiveStr::Ascii("versons"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("delimter"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("swaping"), &["swapping"]),
            (dictgen::InsensitiveStr::Ascii("sctript"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("thermisor"), &["thermistor"]),
            (
                dictgen::InsensitiveStr::Ascii("conatcting"),
                &["contacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("depreacate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("instancse"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("fascisim"), &["fascism"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterprent"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcuer"), &["producer"]),
            (
                dictgen::InsensitiveStr::Ascii("intstructer"),
                &["instructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("decalring"), &["declaring"]),
            (
                dictgen::InsensitiveStr::Ascii("progarmers"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("secoundary"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("throghout"), &["throughout"]),
            (dictgen::InsensitiveStr::Ascii("loacale"), &["locale"]),
            (
                dictgen::InsensitiveStr::Ascii("powersehll"),
                &["powershell"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefiififed"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("assered"), &["asserted"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("isarelis"), &["israelis"]),
            (
                dictgen::InsensitiveStr::Ascii("dimentionnals"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("calid"), &["valid", "canid"]),
            (
                dictgen::InsensitiveStr::Ascii("inevitablly"),
                &["inevitably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immidiately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequnece"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("seletable"), &["selectable"]),
            (
                dictgen::InsensitiveStr::Ascii("amibguities"),
                &["ambiguities"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcuator"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("freindships"),
                &["friendships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expresss"),
                &["expresses", "express"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticlokwise"),
                &["anticlockwise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("morphlogical"),
                &["morphological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autenticator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialzation"),
                &["deserialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("pepare"), &["prepare"]),
            (dictgen::InsensitiveStr::Ascii("guardias"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("previuos"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("prejuduces"),
                &["prejudices"],
            ),
            (dictgen::InsensitiveStr::Ascii("origianal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("certaity"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("vluables"), &["valuables"]),
            (
                dictgen::InsensitiveStr::Ascii("nagivating"),
                &["navigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("assmeble"), &["assemble"]),
            (dictgen::InsensitiveStr::Ascii("forermly"), &["formerly"]),
            (dictgen::InsensitiveStr::Ascii("swalloed"), &["swallowed"]),
            (
                dictgen::InsensitiveStr::Ascii("discrimine"),
                &["discriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("argiculture"),
                &["agriculture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoitating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archetecture"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("preventi"), &["prevent"]),
            (
                dictgen::InsensitiveStr::Ascii("agggregators"),
                &["aggregators"],
            ),
            (dictgen::InsensitiveStr::Ascii("utimately"), &["ultimately"]),
            (dictgen::InsensitiveStr::Ascii("chlroine"), &["chlorine"]),
            (
                dictgen::InsensitiveStr::Ascii("fronkenstein"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supsicious"),
                &["suspicious"],
            ),
            (dictgen::InsensitiveStr::Ascii("elecrton"), &["electron"]),
            (
                dictgen::InsensitiveStr::Ascii("singls"),
                &["singles", "single"],
            ),
            (dictgen::InsensitiveStr::Ascii("seuences"), &["sequences"]),
            (
                dictgen::InsensitiveStr::Ascii("speading"),
                &["spreading", "speeding", "spending", "speaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("defien"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("improrter"), &["importer"]),
            (
                dictgen::InsensitiveStr::Ascii("cyllindrically"),
                &["cylindrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("critquing"), &["critiquing"]),
            (dictgen::InsensitiveStr::Ascii("deocders"), &["decoders"]),
            (
                dictgen::InsensitiveStr::Ascii("identiication"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("expesnce"), &["expense"]),
            (
                dictgen::InsensitiveStr::Ascii("consentrations"),
                &["concentrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematicly"),
                &["mathematical"],
            ),
            (dictgen::InsensitiveStr::Ascii("mediciad"), &["medicaid"]),
            (
                dictgen::InsensitiveStr::Ascii("exhausiveness"),
                &["exhaustiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("creatue"), &["creature"]),
            (
                dictgen::InsensitiveStr::Ascii("histarically"),
                &["historically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australianos"),
                &["australians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nclude"),
                &["include", "conclude"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comandline"),
                &["commandline"],
            ),
            (dictgen::InsensitiveStr::Ascii("decorder"), &["decoder"]),
            (
                dictgen::InsensitiveStr::Ascii("corperation"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininster"),
                &["administer"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooving"), &["proving"]),
            (dictgen::InsensitiveStr::Ascii("democrates"), &["democrats"]),
            (dictgen::InsensitiveStr::Ascii("wiyhout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("peridoically"),
                &["periodically"],
            ),
            (dictgen::InsensitiveStr::Ascii("definilty"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("maanged"), &["managed"]),
            (dictgen::InsensitiveStr::Ascii("recurssed"), &["recursed"]),
            (dictgen::InsensitiveStr::Ascii("algebric"), &["algebraic"]),
            (
                dictgen::InsensitiveStr::Ascii("reolvers"),
                &["resolvers", "revolvers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntrains"),
                &["constrains", "constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soterd"),
                &["stored", "sorted"],
            ),
            (dictgen::InsensitiveStr::Ascii("misspellt"), &["misspelt"]),
            (dictgen::InsensitiveStr::Ascii("singnaled"), &["signaled"]),
            (
                dictgen::InsensitiveStr::Ascii("approximion"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deffensively"),
                &["defensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("gigantisch"), &["gigantic"]),
            (dictgen::InsensitiveStr::Ascii("adust"), &["adjust"]),
            (
                dictgen::InsensitiveStr::Ascii("replicaiton"),
                &["replication"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimized"), &["minimized"]),
            (
                dictgen::InsensitiveStr::Ascii("grapefrukt"),
                &["grapefruit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorespondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("attitides"), &["attitudes"]),
            (dictgen::InsensitiveStr::Ascii("masicer"), &["massacre"]),
            (
                dictgen::InsensitiveStr::Ascii("unilatreally"),
                &["unilaterally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrptors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execeptions"),
                &["exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propagatin"),
                &["propagating", "propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propriatry"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwose"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("apprecaites"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("exisiting"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("oiginating"),
                &["originating"],
            ),
            (dictgen::InsensitiveStr::Ascii("directort"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("nither"), &["neither"]),
            (
                dictgen::InsensitiveStr::Ascii("explicete"),
                &["explicit", "explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrrupted"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transterred"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituton"),
                &["prostitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissasembly"),
                &["disassembly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encounterd"),
                &["encountered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximetions"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallukination"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("ectastic"), &["ecstatic"]),
            (
                dictgen::InsensitiveStr::Ascii("preferental"),
                &["preferential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pairokeality"),
                &["parochiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corinthains"),
                &["corinthians"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuzzes"), &["confuses"]),
            (
                dictgen::InsensitiveStr::Ascii("monotiring"),
                &["monitoring"],
            ),
            (dictgen::InsensitiveStr::Ascii("chcekers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("absolulte"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("parituclar"),
                &["particular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representd"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chocolatos"),
                &["chocolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilses"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("rewirin"), &["rewiring"]),
            (dictgen::InsensitiveStr::Ascii("correctt"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("directoreis"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simle"),
                &["simple", "smile", "simile"],
            ),
            (dictgen::InsensitiveStr::Ascii("aburptly"), &["abruptly"]),
            (
                dictgen::InsensitiveStr::Ascii("recepies"),
                &["recipes", "recopies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubsrcibe"),
                &["unsubscribe"],
            ),
            (dictgen::InsensitiveStr::Ascii("situration"), &["situation"]),
            (
                dictgen::InsensitiveStr::Ascii("revokation"),
                &["revocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("modualar"), &["modular"]),
            (dictgen::InsensitiveStr::Ascii("checkcer"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("repsonders"),
                &["responders"],
            ),
            (dictgen::InsensitiveStr::Ascii("blackshit"), &["blacksmith"]),
            (dictgen::InsensitiveStr::Ascii("diappears"), &["disappears"]),
            (dictgen::InsensitiveStr::Ascii("ges"), &["goes", "guess"]),
            (dictgen::InsensitiveStr::Ascii("gnerals"), &["generals"]),
            (
                dictgen::InsensitiveStr::Ascii("promenently"),
                &["prominently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advantagous"),
                &["advantageous"],
            ),
            (dictgen::InsensitiveStr::Ascii("expecs"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("cruze"), &["cruise"]),
            (
                dictgen::InsensitiveStr::Ascii("selectiose"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("partice"), &["particle"]),
            (dictgen::InsensitiveStr::Ascii("gerneraly"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("opearion"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("emial"), &["email"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaegues"),
                &["colleagues"],
            ),
            (dictgen::InsensitiveStr::Ascii("soureced"), &["sourced"]),
            (dictgen::InsensitiveStr::Ascii("epxanded"), &["expanded"]),
            (
                dictgen::InsensitiveStr::Ascii("accomondate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achieveable"),
                &["achievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorythms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dymamically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("intersted"), &["interested"]),
            (
                dictgen::InsensitiveStr::Ascii("transcendance"),
                &["transcendence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cyphersuites"),
                &["ciphersuites"],
            ),
            (dictgen::InsensitiveStr::Ascii("violoation"), &["violation"]),
            (dictgen::InsensitiveStr::Ascii("stawberry"), &["strawberry"]),
            (dictgen::InsensitiveStr::Ascii("underware"), &["underwear"]),
            (dictgen::InsensitiveStr::Ascii("exitted"), &["exited"]),
            (
                dictgen::InsensitiveStr::Ascii("insersects"),
                &["intersects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffrences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzealis"),
                &["overzealous"],
            ),
            (dictgen::InsensitiveStr::Ascii("crhistmas"), &["christmas"]),
            (dictgen::InsensitiveStr::Ascii("optimissm"), &["optimism"]),
            (dictgen::InsensitiveStr::Ascii("animaties"), &["animates"]),
            (dictgen::InsensitiveStr::Ascii("vaccume"), &["vacuum"]),
            (dictgen::InsensitiveStr::Ascii("exclusivs"), &["exclusives"]),
            (dictgen::InsensitiveStr::Ascii("rewitable"), &["rewritable"]),
            (
                dictgen::InsensitiveStr::Ascii("participatie"),
                &["participated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philippins"),
                &["philippines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invertation"),
                &["inversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("gnerator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("hopelessy"), &["hopelessly"]),
            (dictgen::InsensitiveStr::Ascii("intstant"), &["instant"]),
            (
                dictgen::InsensitiveStr::Ascii("perputrators"),
                &["perpetrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("litecon"), &["litecoin"]),
            (dictgen::InsensitiveStr::Ascii("siblins"), &["siblings"]),
            (dictgen::InsensitiveStr::Ascii("barnching"), &["branching"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotions"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("affraid"), &["afraid"]),
            (
                dictgen::InsensitiveStr::Ascii("braching"),
                &["branching", "breaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successfuly"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subdirectoires"),
                &["subdirectories"],
            ),
            (dictgen::InsensitiveStr::Ascii("obhectify"), &["objectify"]),
            (
                dictgen::InsensitiveStr::Ascii("aggegrates"),
                &["aggregates"],
            ),
            (dictgen::InsensitiveStr::Ascii("qestions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("surpeme"), &["supreme"]),
            (dictgen::InsensitiveStr::Ascii("exhostive"), &["exhaustive"]),
            (
                dictgen::InsensitiveStr::Ascii("partcipant"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibiility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centisencond"),
                &["centisecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desintaitons"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bottomost"), &["bottommost"]),
            (
                dictgen::InsensitiveStr::Ascii("pilons"),
                &["pillions", "pylons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciphersuit"),
                &["ciphersuite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neckbeardese"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifiations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("extermer"), &["extremer"]),
            (dictgen::InsensitiveStr::Ascii("conveyered"), &["conveyed"]),
            (dictgen::InsensitiveStr::Ascii("conepts"), &["concepts"]),
            (
                dictgen::InsensitiveStr::Ascii("pervailing"),
                &["prevailing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difficultly"),
                &["difficulty"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarentee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("hitorical"), &["historical"]),
            (
                dictgen::InsensitiveStr::Ascii("intitiative"),
                &["initiative"],
            ),
            (dictgen::InsensitiveStr::Ascii("innitial"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("outperfoeming"),
                &["outperforming"],
            ),
            (dictgen::InsensitiveStr::Ascii("oherwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("chcking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("posesses"), &["possesses"]),
            (
                dictgen::InsensitiveStr::Ascii("mathametician"),
                &["mathematician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("girlfriens"),
                &["girlfriends"],
            ),
            (dictgen::InsensitiveStr::Ascii("nefarios"), &["nefarious"]),
            (
                dictgen::InsensitiveStr::Ascii("defails"),
                &["details", "derails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represention"),
                &["representing", "representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indigeneous"),
                &["indigenous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afflection"),
                &["affliction"],
            ),
            (dictgen::InsensitiveStr::Ascii("serverite"), &["severity"]),
            (
                dictgen::InsensitiveStr::Ascii("undoed"),
                &["undo", "undone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endiannness"),
                &["endianness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discliamer"),
                &["disclaimer"],
            ),
            (dictgen::InsensitiveStr::Ascii("compillers"), &["compilers"]),
            (dictgen::InsensitiveStr::Ascii("chescksum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("repraesentation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerabile"),
                &["considerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaliated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("mutal"), &["mutual"]),
            (
                dictgen::InsensitiveStr::Ascii("bangaldesh"),
                &["bangladesh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("settubgs"),
                &["settings", "setups"],
            ),
            (dictgen::InsensitiveStr::Ascii("eacg"), &["each"]),
            (
                dictgen::InsensitiveStr::Ascii("standartizing"),
                &["standardizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrusted"), &["entrusted"]),
            (dictgen::InsensitiveStr::Ascii("packeges"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("porgressing"),
                &["progressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("monarkeys"), &["monarchies"]),
            (
                dictgen::InsensitiveStr::Ascii("interstellor"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("updatrs"),
                &["updaters", "updates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microntroller"),
                &["microcontroller"],
            ),
            (dictgen::InsensitiveStr::Ascii("settor"), &["setter"]),
            (dictgen::InsensitiveStr::Ascii("coure"), &["course"]),
            (
                dictgen::InsensitiveStr::Ascii("promter"),
                &["promoter", "prompter"],
            ),
            (dictgen::InsensitiveStr::Ascii("detatched"), &["detached"]),
            (dictgen::InsensitiveStr::Ascii("currenly"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("endevours"), &["endeavours"]),
            (
                dictgen::InsensitiveStr::Ascii("inevitabley"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("depresse"), &["depressive"]),
            (dictgen::InsensitiveStr::Ascii("tiemrs"), &["timers"]),
            (dictgen::InsensitiveStr::Ascii("reccursive"), &["recursive"]),
            (dictgen::InsensitiveStr::Ascii("trainign"), &["training"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenationd"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("referemce"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("constinually"),
                &["continually"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprects"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("expirienced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformase"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clickabled"),
                &["clickable", "clickables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auotmobile"),
                &["automobile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconisder"),
                &["reconsider"],
            ),
            (dictgen::InsensitiveStr::Ascii("discoved"), &["discovered"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherizes"),
                &["characterizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbots"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("onliene"), &["online"]),
            (dictgen::InsensitiveStr::Ascii("enconded"), &["encoded"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsitered"),
                &["administered"],
            ),
            (dictgen::InsensitiveStr::Ascii("aspet"), &["aspect"]),
            (
                dictgen::InsensitiveStr::Ascii("involunary"),
                &["involuntary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thikning"),
                &["thinking", "thickening"],
            ),
            (dictgen::InsensitiveStr::Ascii("primtive"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmectrical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoconpleting"),
                &["autocompleting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undeniabley"),
                &["undeniably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silhouetes"),
                &["silhouettes"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqcuiring"), &["acquiring"]),
            (
                dictgen::InsensitiveStr::Ascii("circumvrent"),
                &["circumvent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notarizaton"),
                &["notarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoned"), &["reasoned"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("exepct"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("clitorious"), &["clitoris"]),
            (
                dictgen::InsensitiveStr::Ascii("ambassidor"),
                &["ambassador"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consficated"),
                &["confiscated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiasists"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("mothods"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("unannomously"),
                &["unanimously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aaproximating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalseced"), &["coalesced"]),
            (
                dictgen::InsensitiveStr::Ascii("probalibity"),
                &["probability"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoudle"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("exitstence"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("intrface"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("incerasing"),
                &["increasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignanle"),
                &["assignable"],
            ),
            (dictgen::InsensitiveStr::Ascii("descriibes"), &["describes"]),
            (dictgen::InsensitiveStr::Ascii("acendency"), &["ascendency"]),
            (dictgen::InsensitiveStr::Ascii("psycology"), &["psychology"]),
            (dictgen::InsensitiveStr::Ascii("motiveated"), &["motivated"]),
            (dictgen::InsensitiveStr::Ascii("layrinxes"), &["larynxes"]),
            (dictgen::InsensitiveStr::Ascii("candiacy"), &["candidacy"]),
            (dictgen::InsensitiveStr::Ascii("mroning"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("verfies"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("developmemt"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serverlesss"),
                &["serverless"],
            ),
            (dictgen::InsensitiveStr::Ascii("infomed"), &["informed"]),
            (dictgen::InsensitiveStr::Ascii("eaxct"), &["exact"]),
            (dictgen::InsensitiveStr::Ascii("rivalrly"), &["rivalry"]),
            (dictgen::InsensitiveStr::Ascii("amatersu"), &["amateurs"]),
            (dictgen::InsensitiveStr::Ascii("thoght"), &["thought"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterizing"),
                &["parameterizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arquitect"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("freecallrelpy"),
                &["freecallreply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("owerflowed"),
                &["overflowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkbench"), &["workbench"]),
            (
                dictgen::InsensitiveStr::Ascii("presuring"),
                &["pressuring", "presuming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficillication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferrably"),
                &["preferably"],
            ),
            (dictgen::InsensitiveStr::Ascii("parnters"), &["partners"]),
            (
                dictgen::InsensitiveStr::Ascii("estuwarries"),
                &["estuaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("forotten"), &["forgotten"]),
            (dictgen::InsensitiveStr::Ascii("variabes"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("defults"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaed"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("surrograte"), &["surrogate"]),
            (dictgen::InsensitiveStr::Ascii("automical"), &["automatic"]),
            (dictgen::InsensitiveStr::Ascii("treasurey"), &["treasury"]),
            (dictgen::InsensitiveStr::Ascii("pthon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("sorceror"), &["sorcerer"]),
            (dictgen::InsensitiveStr::Ascii("vallaying"), &["valeting"]),
            (dictgen::InsensitiveStr::Ascii("discussd"), &["discussed"]),
            (
                dictgen::InsensitiveStr::Ascii("intorductory"),
                &["introductory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpolar"),
                &["interpolator"],
            ),
            (dictgen::InsensitiveStr::Ascii("protion"), &["portion"]),
            (dictgen::InsensitiveStr::Ascii("conatains"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("appearaing"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("apapters"), &["adapters"]),
            (
                dictgen::InsensitiveStr::Ascii("scaffolwing"),
                &["scaffolding"],
            ),
            (dictgen::InsensitiveStr::Ascii("libratie"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("tangentailly"),
                &["tangentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscribir"),
                &["subscriber"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentciated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respone"),
                &["response", "respond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exploraton"),
                &["exploration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunetely"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("incresases"), &["increases"]),
            (
                dictgen::InsensitiveStr::Ascii("counterfit"),
                &["counterfeit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmandline"),
                &["commandline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstanses"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucraps"),
                &["bureaucrats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vitualization"),
                &["visualization", "virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("asissting"), &["assisting"]),
            (dictgen::InsensitiveStr::Ascii("assmebled"), &["assembled"]),
            (dictgen::InsensitiveStr::Ascii("narcotis"), &["narcotics"]),
            (
                dictgen::InsensitiveStr::Ascii("simluation"),
                &["simulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persisitence"),
                &["persistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represenation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("essnetial"), &["essential"]),
            (dictgen::InsensitiveStr::Ascii("realisim"), &["realism"]),
            (dictgen::InsensitiveStr::Ascii("annouing"), &["annoying"]),
            (dictgen::InsensitiveStr::Ascii("ducument"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("iburpofen"), &["ibuprofen"]),
            (dictgen::InsensitiveStr::Ascii("couted"), &["counted"]),
            (dictgen::InsensitiveStr::Ascii("suppurt"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("cultrual"), &["cultural"]),
            (dictgen::InsensitiveStr::Ascii("sytsem"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("forskaen"), &["forsaken"]),
            (dictgen::InsensitiveStr::Ascii("arrray"), &["array"]),
            (
                dictgen::InsensitiveStr::Ascii("humanaties"),
                &["humanities"],
            ),
            (dictgen::InsensitiveStr::Ascii("mataching"), &["matching"]),
            (dictgen::InsensitiveStr::Ascii("broges"), &["brogues"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpectly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("arvg"), &["argv"]),
            (
                dictgen::InsensitiveStr::Ascii("competator"),
                &["competitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("presumaby"), &["presumably"]),
            (dictgen::InsensitiveStr::Ascii("solger"), &["soldier"]),
            (
                dictgen::InsensitiveStr::Ascii("realtively"),
                &["relatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("outsoure"), &["outsourced"]),
            (dictgen::InsensitiveStr::Ascii("inport"), &["import"]),
            (dictgen::InsensitiveStr::Ascii("abstrakt"), &["abstract"]),
            (dictgen::InsensitiveStr::Ascii("squirrtle"), &["squirrel"]),
            (dictgen::InsensitiveStr::Ascii("comodore"), &["commodore"]),
            (
                dictgen::InsensitiveStr::Ascii("comfertably"),
                &["comfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alvorithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("recrods"), &["records"]),
            (
                dictgen::InsensitiveStr::Ascii("excepotins"),
                &["exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintainabillity"),
                &["maintainability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerimentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("speeling"), &["spelling"]),
            (
                dictgen::InsensitiveStr::Ascii("brazilianese"),
                &["brazilians"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("masteires"), &["masteries"]),
            (dictgen::InsensitiveStr::Ascii("engieer"), &["engineer"]),
            (dictgen::InsensitiveStr::Ascii("timning"), &["timing"]),
            (dictgen::InsensitiveStr::Ascii("poritions"), &["portions"]),
            (
                dictgen::InsensitiveStr::Ascii("ommishience"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("silhouwet"), &["silhouette"]),
            (dictgen::InsensitiveStr::Ascii("whithin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("sesnors"), &["sensors"]),
            (
                dictgen::InsensitiveStr::Ascii("constrct"),
                &["construct", "constrict"],
            ),
            (dictgen::InsensitiveStr::Ascii("enitrety"), &["entirety"]),
            (
                dictgen::InsensitiveStr::Ascii("wery"),
                &["very", "wary", "weary"],
            ),
            (dictgen::InsensitiveStr::Ascii("stard"), &["start"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseudicals"),
                &["pharmaceuticals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visting"),
                &["visiting", "listing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underestime"),
                &["underestimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheed"),
                &["spreadsheet"],
            ),
            (dictgen::InsensitiveStr::Ascii("geometyr"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("legalizate"), &["legalize"]),
            (
                dictgen::InsensitiveStr::Ascii("dependenies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("storys"), &["storeys"]),
            (dictgen::InsensitiveStr::Ascii("moast"), &["most", "moat"]),
            (
                dictgen::InsensitiveStr::Ascii("automibile"),
                &["automobile"],
            ),
            (dictgen::InsensitiveStr::Ascii("strentghs"), &["strengths"]),
            (dictgen::InsensitiveStr::Ascii("exersices"), &["exercises"]),
            (dictgen::InsensitiveStr::Ascii("monsterous"), &["monsters"]),
            (dictgen::InsensitiveStr::Ascii("posotion"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("jursidiction"),
                &["jurisdiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deseralising"),
                &["deserialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("lollipoopy"), &["lollipop"]),
            (dictgen::InsensitiveStr::Ascii("banlance"), &["balance"]),
            (dictgen::InsensitiveStr::Ascii("connecion"), &["connection"]),
            (dictgen::InsensitiveStr::Ascii("backedns"), &["backends"]),
            (
                dictgen::InsensitiveStr::Ascii("provied"),
                &["provide", "provided", "proved"],
            ),
            (dictgen::InsensitiveStr::Ascii("hisory"), &["history"]),
            (dictgen::InsensitiveStr::Ascii("debbugers"), &["debuggers"]),
            (
                dictgen::InsensitiveStr::Ascii("canonalizing"),
                &["canonicalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("namaed"), &["named"]),
            (
                dictgen::InsensitiveStr::Ascii("dupications"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctess"),
                &["correctness"],
            ),
            (dictgen::InsensitiveStr::Ascii("acerages"), &["acreages"]),
            (
                dictgen::InsensitiveStr::Ascii("acknwoledge"),
                &["acknowledges", "acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dpeloyment"),
                &["deployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parliement"),
                &["parliament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("louisviile"),
                &["louisville"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aboslutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onces"),
                &["ounces", "once", "ones"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceotors"), &["acceptors"]),
            (dictgen::InsensitiveStr::Ascii("assembe"), &["assemble"]),
            (
                dictgen::InsensitiveStr::Ascii("koordinate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("featur"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatablitiy"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("accceptors"), &["acceptors"]),
            (
                dictgen::InsensitiveStr::Ascii("improvments"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("kayboard"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("sentaor"), &["senator"]),
            (dictgen::InsensitiveStr::Ascii("interestes"), &["interests"]),
            (dictgen::InsensitiveStr::Ascii("sprayade"), &["sprayed"]),
            (dictgen::InsensitiveStr::Ascii("cuurent"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("profesionals"),
                &["professionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precedurally"),
                &["procedurally"],
            ),
            (dictgen::InsensitiveStr::Ascii("becusae"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("impulsivley"),
                &["impulsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regardkess"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotitions"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("adusts"), &["adjusts"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmiating"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalaescing"),
                &["coalescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activistion"),
                &["activision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabalizers"),
                &["stabilizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assication"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffuclt"), &["difficult"]),
            (dictgen::InsensitiveStr::Ascii("loigns"), &["logins"]),
            (dictgen::InsensitiveStr::Ascii("tempraary"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessfull"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scroes"),
                &["scores", "scrods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimtener"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnanteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychopatch"),
                &["psychopath"],
            ),
            (dictgen::InsensitiveStr::Ascii("informella"), &["informal"]),
            (dictgen::InsensitiveStr::Ascii("proctect"), &["protect"]),
            (
                dictgen::InsensitiveStr::Ascii("stauration"),
                &["saturation"],
            ),
            (dictgen::InsensitiveStr::Ascii("virigns"), &["virgins"]),
            (
                dictgen::InsensitiveStr::Ascii("persisitently"),
                &["persistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associcated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("relgious"), &["religious"]),
            (dictgen::InsensitiveStr::Ascii("parameterr"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("follewinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("wrk"), &["work"]),
            (dictgen::InsensitiveStr::Ascii("defendrs"), &["defends"]),
            (
                dictgen::InsensitiveStr::Ascii("unstruction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsequant"),
                &["subsequent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefenreces"),
                &["dereferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indispensibleness"),
                &["indispensableness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cursos"),
                &["cursors", "cursor"],
            ),
            (dictgen::InsensitiveStr::Ascii("serimony"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("stuggling"), &["struggling"]),
            (
                dictgen::InsensitiveStr::Ascii("kickstater"),
                &["kickstarter"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulats"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("airborn"), &["airborne"]),
            (
                dictgen::InsensitiveStr::Ascii("installling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("spheriod"), &["spheroid"]),
            (dictgen::InsensitiveStr::Ascii("opague"), &["opaque"]),
            (dictgen::InsensitiveStr::Ascii("languanges"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("complie"),
                &["compile", "complice", "complied"],
            ),
            (dictgen::InsensitiveStr::Ascii("loacation"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("betwen"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("parmasen"), &["parmesan"]),
            (dictgen::InsensitiveStr::Ascii("victimes"), &["victims"]),
            (dictgen::InsensitiveStr::Ascii("cilyndre"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("undordered"), &["unordered"]),
            (dictgen::InsensitiveStr::Ascii("paackaged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("desciribes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("intrecacies"),
                &["intricacies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsudized"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("perversley"), &["perverse"]),
            (dictgen::InsensitiveStr::Ascii("vershin"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("competense"),
                &["competence"],
            ),
            (dictgen::InsensitiveStr::Ascii("donload"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("coorespondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulataneous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("crapenter"), &["carpenter"]),
            (dictgen::InsensitiveStr::Ascii("villan"), &["villain"]),
            (
                dictgen::InsensitiveStr::Ascii("distingused"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("predjuice"), &["prejudice"]),
            (
                dictgen::InsensitiveStr::Ascii("unappeeling"),
                &["unappealing"],
            ),
            (dictgen::InsensitiveStr::Ascii("gauarentee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("equivalalent"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("somenthing"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("intertwyned"),
                &["intertwined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitutins"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("alchoholic"), &["alcoholic"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrevated"),
                &["aggravated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sendinging"), &["sending"]),
            (dictgen::InsensitiveStr::Ascii("boudler"), &["boulder"]),
            (
                dictgen::InsensitiveStr::Ascii("participante"),
                &["participate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existensial"),
                &["existential"],
            ),
            (dictgen::InsensitiveStr::Ascii("improvin"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("headmap"), &["heatmap"]),
            (
                dictgen::InsensitiveStr::Ascii("appoinment"),
                &["appointment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("labouriously"),
                &["laboriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("chassms"), &["chasms"]),
            (
                dictgen::InsensitiveStr::Ascii("initalizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("featrue"), &["feature"]),
            (dictgen::InsensitiveStr::Ascii("conaint"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("detectetd"), &["detected"]),
            (dictgen::InsensitiveStr::Ascii("miraculos"), &["miraculous"]),
            (dictgen::InsensitiveStr::Ascii("chaecking"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("idenitifiers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("identyty"), &["identity"]),
            (dictgen::InsensitiveStr::Ascii("mimick"), &["mimic"]),
            (
                dictgen::InsensitiveStr::Ascii("unregstered"),
                &["unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mecernaries"),
                &["mercenaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("counteries"), &["counters"]),
            (
                dictgen::InsensitiveStr::Ascii("overriting"),
                &["overwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resolvement"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neibhbours"),
                &["neighbours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schoole"),
                &["schooled", "schools"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfitying"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("ddressing"),
                &["addressing", "dressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("exensions"), &["extensions"]),
            (dictgen::InsensitiveStr::Ascii("exercisin"), &["exercising"]),
            (dictgen::InsensitiveStr::Ascii("iamges"), &["images"]),
            (dictgen::InsensitiveStr::Ascii("domainate"), &["dominate"]),
            (
                dictgen::InsensitiveStr::Ascii("recontsruction"),
                &["reconstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("entitie"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("smewhat"), &["somewhat"]),
            (
                dictgen::InsensitiveStr::Ascii("disconneting"),
                &["disconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferrences"),
                &["preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterized"),
                &["parameterized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomosphere"),
                &["atmosphere"],
            ),
            (dictgen::InsensitiveStr::Ascii("visualzed"), &["visualized"]),
            (dictgen::InsensitiveStr::Ascii("rebulit"), &["rebuilt"]),
            (dictgen::InsensitiveStr::Ascii("opitimize"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("sawtayes"), &["sautés"]),
            (
                dictgen::InsensitiveStr::Ascii("cringewothy"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("defferred"), &["deferred"]),
            (dictgen::InsensitiveStr::Ascii("resintall"), &["reinstall"]),
            (
                dictgen::InsensitiveStr::Ascii("favorise"),
                &["favor", "favorite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overriddable"),
                &["overridable"],
            ),
            (dictgen::InsensitiveStr::Ascii("collasion"), &["collision"]),
            (
                dictgen::InsensitiveStr::Ascii("probabalistic"),
                &["probabilistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("tresure"), &["treasure"]),
            (
                dictgen::InsensitiveStr::Ascii("persistante"),
                &["persistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("siguretes"), &["cigarettes"]),
            (dictgen::InsensitiveStr::Ascii("waranty"), &["warranty"]),
            (
                dictgen::InsensitiveStr::Ascii("unitializes"),
                &["uninitializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqiuem"), &["requiem"]),
            (dictgen::InsensitiveStr::Ascii("retireved"), &["retrieved"]),
            (dictgen::InsensitiveStr::Ascii("negotite"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("subrouines"),
                &["subroutines"],
            ),
            (dictgen::InsensitiveStr::Ascii("comilers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("ananlytically"),
                &["analytically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paria"),
                &["pariah", "parka"],
            ),
            (dictgen::InsensitiveStr::Ascii("greif"), &["grief"]),
            (
                dictgen::InsensitiveStr::Ascii("minamilist"),
                &["minimalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("minuscle"), &["minuscule"]),
            (dictgen::InsensitiveStr::Ascii("tahnked"), &["thanked"]),
            (dictgen::InsensitiveStr::Ascii("flucutate"), &["fluctuate"]),
            (dictgen::InsensitiveStr::Ascii("cylcled"), &["cycled"]),
            (dictgen::InsensitiveStr::Ascii("evidende"), &["evidenced"]),
            (dictgen::InsensitiveStr::Ascii("receeded"), &["receded"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtenticates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incommplete"),
                &["incomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("addig"), &["adding"]),
            (dictgen::InsensitiveStr::Ascii("exceucting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("upploaded"), &["uploaded"]),
            (
                dictgen::InsensitiveStr::Ascii("translteration"),
                &["transliteration"],
            ),
            (dictgen::InsensitiveStr::Ascii("santizing"), &["sanitizing"]),
            (dictgen::InsensitiveStr::Ascii("recpie"), &["recipe"]),
            (dictgen::InsensitiveStr::Ascii("intecept"), &["intercept"]),
            (
                dictgen::InsensitiveStr::Ascii("psycholigically"),
                &["psychologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currespondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("remved"), &["removed"]),
            (dictgen::InsensitiveStr::Ascii("possiably"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("atmoizer"), &["atomizer"]),
            (dictgen::InsensitiveStr::Ascii("ammended"), &["amended"]),
            (
                dictgen::InsensitiveStr::Ascii("earthqauke"),
                &["earthquake"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inveitable"),
                &["inevitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sponsorise"), &["sponsors"]),
            (dictgen::InsensitiveStr::Ascii("reviewtrue"), &["reviewer"]),
            (dictgen::InsensitiveStr::Ascii("obsidain"), &["obsidian"]),
            (
                dictgen::InsensitiveStr::Ascii("sefaults"),
                &["segfaults", "defaults"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pshycological"),
                &["psychological"],
            ),
            (dictgen::InsensitiveStr::Ascii("pathanme"), &["pathname"]),
            (dictgen::InsensitiveStr::Ascii("strang"), &["strange"]),
            (dictgen::InsensitiveStr::Ascii("optimazer"), &["optimizer"]),
            (
                dictgen::InsensitiveStr::Ascii("loocking"),
                &["locking", "looking"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirth"), &["with", "worth"]),
            (
                dictgen::InsensitiveStr::Ascii("assersions"),
                &["assertions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disdvantage"),
                &["disadvantage"],
            ),
            (dictgen::InsensitiveStr::Ascii("hertzs"), &["hertz"]),
            (
                dictgen::InsensitiveStr::Ascii("nagivators"),
                &["navigators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insufficiant"),
                &["insufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("namnespaces"),
                &["namespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("afwully"), &["awfully"]),
            (dictgen::InsensitiveStr::Ascii("enconder"), &["encoder"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvenieces"),
                &["inconveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcue"), &["produce"]),
            (dictgen::InsensitiveStr::Ascii("operatting"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("bitterseet"),
                &["bittersweet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repentable"),
                &["repeatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retardathon"),
                &["retardation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivallents"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invesitgations"),
                &["investigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("engieneer"), &["engineer"]),
            (
                dictgen::InsensitiveStr::Ascii("producables"),
                &["producible"],
            ),
            (dictgen::InsensitiveStr::Ascii("defitenly"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("hamburgaren"),
                &["hamburger"],
            ),
            (dictgen::InsensitiveStr::Ascii("actusl"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("repatwar"), &["repertoire"]),
            (dictgen::InsensitiveStr::Ascii("peninsulla"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("dissassociate"),
                &["disassociate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omplementaion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("parmater"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("patriotisk"), &["patriots"]),
            (dictgen::InsensitiveStr::Ascii("comprator"), &["comparator"]),
            (
                dictgen::InsensitiveStr::Ascii("intrducing"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dcoumentation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialialisation"),
                &["serialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("spanw"), &["spawn"]),
            (dictgen::InsensitiveStr::Ascii("intot"), &["into"]),
            (
                dictgen::InsensitiveStr::Ascii("prevailaing"),
                &["prevailing"],
            ),
            (dictgen::InsensitiveStr::Ascii("armorment"), &["armament"]),
            (dictgen::InsensitiveStr::Ascii("pulish"), &["publish"]),
            (dictgen::InsensitiveStr::Ascii("beter"), &["better"]),
            (dictgen::InsensitiveStr::Ascii("depsoit"), &["deposit"]),
            (
                dictgen::InsensitiveStr::Ascii("dissambiguate"),
                &["disambiguate"],
            ),
            (dictgen::InsensitiveStr::Ascii("romanin"), &["romanian"]),
            (dictgen::InsensitiveStr::Ascii("sereouslly"), &["seriously"]),
            (dictgen::InsensitiveStr::Ascii("helpying"), &["helping"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetentelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolising"),
                &["capitalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("impoartant"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("enviornmentalist"),
                &["environmentalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("faund"), &["found", "fund"]),
            (dictgen::InsensitiveStr::Ascii("cooldwons"), &["cooldowns"]),
            (dictgen::InsensitiveStr::Ascii("matcbh"), &["match"]),
            (dictgen::InsensitiveStr::Ascii("acheeves"), &["achieves"]),
            (
                dictgen::InsensitiveStr::Ascii("ubiquitious"),
                &["ubiquitous"],
            ),
            (dictgen::InsensitiveStr::Ascii("commtiter"), &["committer"]),
            (dictgen::InsensitiveStr::Ascii("exended"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("canoicalize"),
                &["canonicalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operaitons"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serailisation"),
                &["serialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalties"),
                &["nationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appologized"),
                &["apologized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constributes"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrcted"), &["abstracted"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumenetd"),
                &["instrumented"],
            ),
            (dictgen::InsensitiveStr::Ascii("reelvant"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("firey"), &["fiery", "firry"]),
            (dictgen::InsensitiveStr::Ascii("synethic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("astroanut"), &["astronaut"]),
            (dictgen::InsensitiveStr::Ascii("arcive"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("depicitng"), &["depicting"]),
            (dictgen::InsensitiveStr::Ascii("delivere"), &["deliveries"]),
            (dictgen::InsensitiveStr::Ascii("clitorios"), &["clitoris"]),
            (
                dictgen::InsensitiveStr::Ascii("adjectiveus"),
                &["adjectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreasonbly"),
                &["unreasonably"],
            ),
            (dictgen::InsensitiveStr::Ascii("telphony"), &["telephony"]),
            (
                dictgen::InsensitiveStr::Ascii("cusom"),
                &["custom", "cusum"],
            ),
            (dictgen::InsensitiveStr::Ascii("refenrence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("firsrt"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("resepects"), &["respects"]),
            (dictgen::InsensitiveStr::Ascii("embargos"), &["embargoes"]),
            (dictgen::InsensitiveStr::Ascii("comarde"), &["comrade"]),
            (
                dictgen::InsensitiveStr::Ascii("configurates"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitiing"),
                &["delimiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpetect"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutiplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cerinomially"),
                &["ceremonially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desinty"),
                &["density", "destiny"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctroying"),
                &["destroying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proyection"),
                &["projection", "protection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumsision"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanich"), &["launch"]),
            (dictgen::InsensitiveStr::Ascii("permament"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("trggered"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("assoiciated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objecitves"),
                &["objectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("stablized"), &["stabilized"]),
            (dictgen::InsensitiveStr::Ascii("intalled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("plaforms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("sequeze"),
                &["squeeze", "sequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strategisk"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recogizing"),
                &["recognizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reudctions"),
                &["reductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceled"), &["concede"]),
            (dictgen::InsensitiveStr::Ascii("restrait"), &["restraint"]),
            (
                dictgen::InsensitiveStr::Ascii("replacalbe"),
                &["replaceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archimedian"),
                &["archimedean"],
            ),
            (dictgen::InsensitiveStr::Ascii("hardhare"), &["hardware"]),
            (dictgen::InsensitiveStr::Ascii("compount"), &["compound"]),
            (dictgen::InsensitiveStr::Ascii("overhual"), &["overhaul"]),
            (dictgen::InsensitiveStr::Ascii("alhabets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("abstraktion"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("standarad"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("refferences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("lateny"), &["latency"]),
            (
                dictgen::InsensitiveStr::Ascii("arcitechture"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangager"), &["manager"]),
            (
                dictgen::InsensitiveStr::Ascii("expermentelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("augmnetes"), &["augments"]),
            (
                dictgen::InsensitiveStr::Ascii("oscilliscopes"),
                &["oscilloscopes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verticlealign"),
                &["verticalalign"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impprovements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dffered"),
                &["differed", "offered"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimiser"), &["optimiser"]),
            (
                dictgen::InsensitiveStr::Ascii("executations"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("incremint"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("supressions"),
                &["suppressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("loweraced"), &["lowercased"]),
            (dictgen::InsensitiveStr::Ascii("standrats"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("enhabced"), &["enhanced"]),
            (dictgen::InsensitiveStr::Ascii("manifestes"), &["manifests"]),
            (dictgen::InsensitiveStr::Ascii("matchs"), &["matches"]),
            (
                dictgen::InsensitiveStr::Ascii("travellodge"),
                &["travelled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surreptiously"),
                &["surreptitiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("territorries"),
                &["territories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ralationship"),
                &["relationship"],
            ),
            (dictgen::InsensitiveStr::Ascii("alligned"), &["aligned"]),
            (dictgen::InsensitiveStr::Ascii("curerent"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("narrowd"),
                &["narrowed", "narrows"],
            ),
            (dictgen::InsensitiveStr::Ascii("riminder"), &["reminder"]),
            (dictgen::InsensitiveStr::Ascii("horishima"), &["hiroshima"]),
            (dictgen::InsensitiveStr::Ascii("emnity"), &["enmity"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessasarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumisation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registraion"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intentially"),
                &["intentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("workpsace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("settting"), &["setting"]),
            (
                dictgen::InsensitiveStr::Ascii("controllerd"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantasitcally"),
                &["fantastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("nutritent"), &["nutrient"]),
            (
                dictgen::InsensitiveStr::Ascii("fortuantely"),
                &["fortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("preferrin"), &["preferring"]),
            (dictgen::InsensitiveStr::Ascii("chekcing"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("seriosuly"), &["seriously"]),
            (
                dictgen::InsensitiveStr::Ascii("earthqukes"),
                &["earthquakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("insalls"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("leagalise"), &["legalise"]),
            (
                dictgen::InsensitiveStr::Ascii("transformarea"),
                &["transformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postitioning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficialleis"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeatedlt"), &["repealed"]),
            (dictgen::InsensitiveStr::Ascii("thasnks"), &["thanks"]),
            (
                dictgen::InsensitiveStr::Ascii("indpendently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("zellots"), &["zealots"]),
            (
                dictgen::InsensitiveStr::Ascii("lieutanent"),
                &["lieutenant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undergorund"),
                &["underground"],
            ),
            (dictgen::InsensitiveStr::Ascii("custums"), &["customs"]),
            (dictgen::InsensitiveStr::Ascii("becamae"), &["became"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiififed"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("zukeeni"), &["zucchini"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsitant"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtsitutatble"),
                &["substitutable"],
            ),
            (dictgen::InsensitiveStr::Ascii("governered"), &["governed"]),
            (dictgen::InsensitiveStr::Ascii("delviery"), &["delivery"]),
            (
                dictgen::InsensitiveStr::Ascii("compleation"),
                &["compilation", "completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("srcipting"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("corelate"), &["correlate"]),
            (dictgen::InsensitiveStr::Ascii("meddo"), &["meadow"]),
            (
                dictgen::InsensitiveStr::Ascii("equialently"),
                &["equivalently"],
            ),
            (dictgen::InsensitiveStr::Ascii("circuitos"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("porletariat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("sutisfies"), &["satisfies"]),
            (dictgen::InsensitiveStr::Ascii("bounus"), &["bonus"]),
            (dictgen::InsensitiveStr::Ascii("becuaes"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("praameter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("refferer"), &["referrer"]),
            (dictgen::InsensitiveStr::Ascii("elemant"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("finihsed"), &["finished"]),
            (
                dictgen::InsensitiveStr::Ascii("revoluutionary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("miminizes"), &["minimizes"]),
            (dictgen::InsensitiveStr::Ascii("varierty"), &["variety"]),
            (
                dictgen::InsensitiveStr::Ascii("implented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("embeed"), &["embed"]),
            (
                dictgen::InsensitiveStr::Ascii("acheevable"),
                &["achievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unaccessable"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intdicating"),
                &["indicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("trianers"), &["trainers"]),
            (dictgen::InsensitiveStr::Ascii("isalha"), &["isalpha"]),
            (dictgen::InsensitiveStr::Ascii("transferrs"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("clening"), &["cleaning"]),
            (dictgen::InsensitiveStr::Ascii("whth"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("exoitcs"), &["exotics"]),
            (dictgen::InsensitiveStr::Ascii("tarrifs"), &["tariffs"]),
            (dictgen::InsensitiveStr::Ascii("repesents"), &["represents"]),
            (
                dictgen::InsensitiveStr::Ascii("redifintions"),
                &["redefinitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("christain"), &["christian"]),
            (dictgen::InsensitiveStr::Ascii("addrssed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("protmoto"), &["promote"]),
            (
                dictgen::InsensitiveStr::Ascii("trangulating"),
                &["triangulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("oficially"), &["officially"]),
            (dictgen::InsensitiveStr::Ascii("resurce"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("incrememnting"),
                &["incrementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrriden"),
                &["overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("speices"), &["species"]),
            (
                dictgen::InsensitiveStr::Ascii("cloberring"),
                &["clobbering"],
            ),
            (dictgen::InsensitiveStr::Ascii("birhtdays"), &["birthdays"]),
            (
                dictgen::InsensitiveStr::Ascii("defenisvely"),
                &["defensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("simpifies"), &["simplifies"]),
            (dictgen::InsensitiveStr::Ascii("simetries"), &["symmetries"]),
            (dictgen::InsensitiveStr::Ascii("conjugiate"), &["conjugate"]),
            (dictgen::InsensitiveStr::Ascii("muliples"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("inteligence"),
                &["intelligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irridation"),
                &["irritation"],
            ),
            (dictgen::InsensitiveStr::Ascii("critieron"), &["criterion"]),
            (
                dictgen::InsensitiveStr::Ascii("gangsterest"),
                &["gangsters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docummented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interveres"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypocritus"),
                &["hypocrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilizacion"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("bage"), &["bag"]),
            (dictgen::InsensitiveStr::Ascii("charactet"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("particpiate"),
                &["participated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functioally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("runnig"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("conciders"), &["considers"]),
            (
                dictgen::InsensitiveStr::Ascii("resperatory"),
                &["respiratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separetely"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("textrues"), &["textures"]),
            (dictgen::InsensitiveStr::Ascii("exceprts"), &["excerpts"]),
            (dictgen::InsensitiveStr::Ascii("playist"), &["playlist"]),
            (dictgen::InsensitiveStr::Ascii("keyoutch"), &["keytouch"]),
            (dictgen::InsensitiveStr::Ascii("shifed"), &["shifted"]),
            (
                dictgen::InsensitiveStr::Ascii("filessytem"),
                &["filesystem"],
            ),
            (dictgen::InsensitiveStr::Ascii("herustics"), &["heuristics"]),
            (
                dictgen::InsensitiveStr::Ascii("expresssive"),
                &["expressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionair"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participans"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accustumed"),
                &["accustomed"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkk"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("deocde"), &["decode"]),
            (
                dictgen::InsensitiveStr::Ascii("unimpresed"),
                &["unimpressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straightorward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("stirr"), &["stir"]),
            (
                dictgen::InsensitiveStr::Ascii("handicape"),
                &["handicapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("technlogy"), &["technology"]),
            (
                dictgen::InsensitiveStr::Ascii("irrelevent"),
                &["irrelevant"],
            ),
            (dictgen::InsensitiveStr::Ascii("colose"), &["close"]),
            (dictgen::InsensitiveStr::Ascii("arbiatray"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("combonation"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("archioves"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("snowbolling"),
                &["snowballing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensibilty"),
                &["extensibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("compents"), &["components"]),
            (
                dictgen::InsensitiveStr::Ascii("incarciration"),
                &["incarceration"],
            ),
            (dictgen::InsensitiveStr::Ascii("elctrical"), &["electrical"]),
            (
                dictgen::InsensitiveStr::Ascii("bodydbuilder"),
                &["bodybuilder"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuseday"), &["tuesday"]),
            (dictgen::InsensitiveStr::Ascii("upgarted"), &["upgraded"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoor"),
                &["neighbor"],
            ),
            (dictgen::InsensitiveStr::Ascii("rubarb"), &["rhubarb"]),
            (dictgen::InsensitiveStr::Ascii("alirghty"), &["alrighty"]),
            (
                dictgen::InsensitiveStr::Ascii("composicion"),
                &["composition"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizled"), &["chiseled"]),
            (dictgen::InsensitiveStr::Ascii("univeral"), &["universal"]),
            (
                dictgen::InsensitiveStr::Ascii("investigatin"),
                &["investigation", "investigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("liqiuds"), &["liquids"]),
            (dictgen::InsensitiveStr::Ascii("redunant"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("poltically"),
                &["politically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repersentations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("trankated"), &["truncated"]),
            (
                dictgen::InsensitiveStr::Ascii("kidnapning"),
                &["kidnapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrubyte"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("regural"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("neglicence"),
                &["negligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diatancing"),
                &["distancing"],
            ),
            (dictgen::InsensitiveStr::Ascii("renaned"), &["renamed"]),
            (
                dictgen::InsensitiveStr::Ascii("corispondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwats"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("pladed"),
                &["plaided", "pleaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wiuld"),
                &["would", "wild", "wield"],
            ),
            (dictgen::InsensitiveStr::Ascii("witk"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("briuses"), &["bruises"]),
            (
                dictgen::InsensitiveStr::Ascii("cellabrations"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdendet"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("commentare"), &["commenter"]),
            (dictgen::InsensitiveStr::Ascii("speicfied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("cleares"), &["clears"]),
            (dictgen::InsensitiveStr::Ascii("skepticim"), &["skepticism"]),
            (
                dictgen::InsensitiveStr::Ascii("maniuplation"),
                &["manipulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subdirectiories"),
                &["subdirectories"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheve"), &["achieve"]),
            (
                dictgen::InsensitiveStr::Ascii("hypathetical"),
                &["hypothetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccuricies"),
                &["inaccuracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("mushroooms"), &["mushroom"]),
            (dictgen::InsensitiveStr::Ascii("missle"), &["missile"]),
            (dictgen::InsensitiveStr::Ascii("nuisanse"), &["nuisance"]),
            (dictgen::InsensitiveStr::Ascii("fighers"), &["fighters"]),
            (dictgen::InsensitiveStr::Ascii("assigher"), &["assigner"]),
            (dictgen::InsensitiveStr::Ascii("attacs"), &["attacks"]),
            (
                dictgen::InsensitiveStr::Ascii("natinalist"),
                &["nationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reupladers"),
                &["reuploaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insrugency"),
                &["insurgency"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadamics"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("ellegantly"), &["elegantly"]),
            (
                dictgen::InsensitiveStr::Ascii("erroneusly"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterise"),
                &["parameterise"],
            ),
            (dictgen::InsensitiveStr::Ascii("generoator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("thermistat"),
                &["thermostat"],
            ),
            (dictgen::InsensitiveStr::Ascii("genious"), &["genius"]),
            (
                dictgen::InsensitiveStr::Ascii("cercumstances"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proections"),
                &["protections", "projections"],
            ),
            (dictgen::InsensitiveStr::Ascii("asborbing"), &["absorbing"]),
            (dictgen::InsensitiveStr::Ascii("desicive"), &["decisive"]),
            (dictgen::InsensitiveStr::Ascii("subshystem"), &["subsystem"]),
            (dictgen::InsensitiveStr::Ascii("repedetly"), &["repeatedly"]),
            (
                dictgen::InsensitiveStr::Ascii("itnerpretation"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hollistically"),
                &["holistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimints"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("collaps"), &["collapse"]),
            (
                dictgen::InsensitiveStr::Ascii("appoximately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("marixsts"), &["marxists"]),
            (dictgen::InsensitiveStr::Ascii("loacating"), &["locating"]),
            (dictgen::InsensitiveStr::Ascii("bakcward"), &["backward"]),
            (
                dictgen::InsensitiveStr::Ascii("unsharable"),
                &["unshareable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfiabilty"),
                &["satisfiability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structrual"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogenious"),
                &["homogeneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promiscous"),
                &["promiscuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalsecense"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructeds"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlement"),
                &["acknowledgment"],
            ),
            (dictgen::InsensitiveStr::Ascii("partiel"), &["particle"]),
            (
                dictgen::InsensitiveStr::Ascii("rutes"),
                &["brutes", "routes", "rules"],
            ),
            (dictgen::InsensitiveStr::Ascii("spacegoat"), &["scapegoat"]),
            (dictgen::InsensitiveStr::Ascii("agrguments"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("insuffciency"),
                &["insufficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("configed"), &["configured"]),
            (
                dictgen::InsensitiveStr::Ascii("irregularties"),
                &["irregularities"],
            ),
            (dictgen::InsensitiveStr::Ascii("toplogy"), &["topology"]),
            (dictgen::InsensitiveStr::Ascii("aritrary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("algoorithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reliabillity"),
                &["reliability"],
            ),
            (dictgen::InsensitiveStr::Ascii("contoller"), &["controller"]),
            (
                dictgen::InsensitiveStr::Ascii("scuccessfull"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("covince"), &["convince"]),
            (dictgen::InsensitiveStr::Ascii("assiting"), &["assisting"]),
            (dictgen::InsensitiveStr::Ascii("pertrub"), &["perturb"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferencced"),
                &["dereferenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opeational"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neumonicly"),
                &["mnemonically"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenitty"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("bankruptsy"),
                &["bankruptcy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possobilities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("resvolved"), &["resolved"]),
            (dictgen::InsensitiveStr::Ascii("realiably"), &["reliably"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("encouter"), &["encounter"]),
            (
                dictgen::InsensitiveStr::Ascii("attrocities"),
                &["atrocities"],
            ),
            (dictgen::InsensitiveStr::Ascii("varieity"), &["variety"]),
            (
                dictgen::InsensitiveStr::Ascii("resurrecion"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("postively"), &["positively"]),
            (
                dictgen::InsensitiveStr::Ascii("documeneters"),
                &["documenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("langauage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("fermentacion"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("currentfps"), &["currents"]),
            (dictgen::InsensitiveStr::Ascii("conncet"), &["connect"]),
            (dictgen::InsensitiveStr::Ascii("cointains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("desturbing"),
                &["disturbing"],
            ),
            (dictgen::InsensitiveStr::Ascii("chancged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("kindapping"),
                &["kidnapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuited"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("conditons"), &["conditions"]),
            (
                dictgen::InsensitiveStr::Ascii("downgarded"),
                &["downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("cachin"), &["caching"]),
            (dictgen::InsensitiveStr::Ascii("travestry"), &["travesty"]),
            (
                dictgen::InsensitiveStr::Ascii("vacciantion"),
                &["vaccination"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonymouse"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("reults"), &["results"]),
            (dictgen::InsensitiveStr::Ascii("rondayvou"), &["rendezvous"]),
            (
                dictgen::InsensitiveStr::Ascii("attribuiting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("agancy"), &["agency"]),
            (dictgen::InsensitiveStr::Ascii("lookng"), &["looking"]),
            (dictgen::InsensitiveStr::Ascii("phasis"), &["phases"]),
            (
                dictgen::InsensitiveStr::Ascii("innitializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("welathier"), &["wealthier"]),
            (dictgen::InsensitiveStr::Ascii("defectos"), &["defects"]),
            (
                dictgen::InsensitiveStr::Ascii("considdered"),
                &["considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customizble"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissallows"), &["disallows"]),
            (
                dictgen::InsensitiveStr::Ascii("accommdated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggrgated"), &["aggregated"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetamies"),
                &["amphetamines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occurrrences"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("spesific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("istance"),
                &["instance", "distance"],
            ),
            (dictgen::InsensitiveStr::Ascii("ligthly"), &["lightly"]),
            (
                dictgen::InsensitiveStr::Ascii("descrepancies"),
                &["discrepancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("debina"), &["debian"]),
            (dictgen::InsensitiveStr::Ascii("deisgned"), &["designed"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarhical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superceeds"),
                &["supersedes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucome"), &["succumb"]),
            (
                dictgen::InsensitiveStr::Ascii("colloqueal"),
                &["colloquial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soldgering"),
                &["soldiering"],
            ),
            (dictgen::InsensitiveStr::Ascii("gaurdian"), &["guardian"]),
            (dictgen::InsensitiveStr::Ascii("calssic"), &["classic"]),
            (dictgen::InsensitiveStr::Ascii("stpo"), &["stop"]),
            (dictgen::InsensitiveStr::Ascii("configrue"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("espisodically"),
                &["episodically"],
            ),
            (dictgen::InsensitiveStr::Ascii("brackers"), &["brackets"]),
            (
                dictgen::InsensitiveStr::Ascii("appproaches"),
                &["approaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("chatow"), &["château"]),
            (dictgen::InsensitiveStr::Ascii("biulders"), &["builders"]),
            (
                dictgen::InsensitiveStr::Ascii("sementically"),
                &["semantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compataility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deperecations"),
                &["deprecations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relativiser"),
                &["relatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimatly"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibillities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("zylophones"),
                &["xylophones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resuled"),
                &["resulted", "resumed"],
            ),
            (dictgen::InsensitiveStr::Ascii("euqals"), &["equals"]),
            (
                dictgen::InsensitiveStr::Ascii("repeatitively"),
                &["repetitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("autograpgh"), &["autograph"]),
            (
                dictgen::InsensitiveStr::Ascii("implausable"),
                &["implausible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmens"),
                &["segment", "segments"],
            ),
            (dictgen::InsensitiveStr::Ascii("arpeture"), &["aperture"]),
            (dictgen::InsensitiveStr::Ascii("cilcking"), &["clicking"]),
            (
                dictgen::InsensitiveStr::Ascii("rearranment"),
                &["rearrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("seh"), &["she"]),
            (dictgen::InsensitiveStr::Ascii("suposed"), &["supposed"]),
            (dictgen::InsensitiveStr::Ascii("parseing"), &["parsing"]),
            (dictgen::InsensitiveStr::Ascii("crashees"), &["crashes"]),
            (
                dictgen::InsensitiveStr::Ascii("excetutables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("highlane"), &["highlander"]),
            (
                dictgen::InsensitiveStr::Ascii("origninates"),
                &["originates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paravirtualied"),
                &["paravirtualised", "paravirtualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumcition"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("designes"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("cretes"),
                &["creates", "crates"],
            ),
            (dictgen::InsensitiveStr::Ascii("datases"), &["datasets"]),
            (dictgen::InsensitiveStr::Ascii("appproach"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("copmiled"), &["compiled"]),
            (
                dictgen::InsensitiveStr::Ascii("manupilated"),
                &["manipulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("serounds"), &["surrounds"]),
            (
                dictgen::InsensitiveStr::Ascii("alisas"),
                &["alias", "aliases"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrreshold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("enourages"), &["encourages"]),
            (dictgen::InsensitiveStr::Ascii("inititator"), &["initiator"]),
            (
                dictgen::InsensitiveStr::Ascii("diffirentiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sesceding"), &["seceding"]),
            (
                dictgen::InsensitiveStr::Ascii("unconsistency"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ultimatelly"),
                &["ultimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("immobilzed"), &["immobile"]),
            (dictgen::InsensitiveStr::Ascii("uneform"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("importantely"),
                &["importantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosemtic"), &["cosmetic"]),
            (dictgen::InsensitiveStr::Ascii("verctors"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("boudned"), &["bounded"]),
            (dictgen::InsensitiveStr::Ascii("strictist"), &["strictest"]),
            (
                dictgen::InsensitiveStr::Ascii("imputs"),
                &["inputs", "imputes"],
            ),
            (dictgen::InsensitiveStr::Ascii("achevable"), &["achievable"]),
            (
                dictgen::InsensitiveStr::Ascii("respiritory"),
                &["respiratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundrary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("begining"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("opional"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("preveiws"), &["previews"]),
            (
                dictgen::InsensitiveStr::Ascii("dysfunktional"),
                &["dysfunctional"],
            ),
            (dictgen::InsensitiveStr::Ascii("convigs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("defintaly"), &["defiantly"]),
            (dictgen::InsensitiveStr::Ascii("anrachist"), &["anarchist"]),
            (
                dictgen::InsensitiveStr::Ascii("resoltuion"),
                &["resolutions", "resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("keybindnig"),
                &["keybinding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criculating"),
                &["circulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mussils"),
                &["muscles", "mussels"],
            ),
            (dictgen::InsensitiveStr::Ascii("compaires"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("updraged"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("envidenced"), &["evidenced"]),
            (
                dictgen::InsensitiveStr::Ascii("triscaidecaphobia"),
                &["triskaidekaphobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archictecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("encyption"), &["encryption"]),
            (
                dictgen::InsensitiveStr::Ascii("adminitrator"),
                &["administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partitiioning"),
                &["partitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("krytponite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustation"),
                &["frustration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sixtin"),
                &["sistine", "sixteen"],
            ),
            (dictgen::InsensitiveStr::Ascii("ibjects"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("sessoion"), &["session"]),
            (
                dictgen::InsensitiveStr::Ascii("tragectory"),
                &["trajectory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combanatorial"),
                &["combinatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememntation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estiomators"),
                &["estimators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabililty"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphabetcially"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ggod"), &["good"]),
            (dictgen::InsensitiveStr::Ascii("modukles"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("skept"), &["skipped"]),
            (dictgen::InsensitiveStr::Ascii("exicuting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("masakist"), &["masochist"]),
            (dictgen::InsensitiveStr::Ascii("guaruanty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("restirer"), &["restorer"]),
            (
                dictgen::InsensitiveStr::Ascii("firefighers"),
                &["firefighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("limtier"), &["limiter"]),
            (dictgen::InsensitiveStr::Ascii("copver"), &["cover"]),
            (
                dictgen::InsensitiveStr::Ascii("autotoriums"),
                &["auditoriums"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crrespondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("marginaali"), &["marginal"]),
            (
                dictgen::InsensitiveStr::Ascii("inturpratation"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espicially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("equelly"), &["equally"]),
            (
                dictgen::InsensitiveStr::Ascii("commnetator"),
                &["commentator"],
            ),
            (dictgen::InsensitiveStr::Ascii("asnwers"), &["answers"]),
            (dictgen::InsensitiveStr::Ascii("communsit"), &["communist"]),
            (
                dictgen::InsensitiveStr::Ascii("sponteanous"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrker"), &["worker"]),
            (
                dictgen::InsensitiveStr::Ascii("computato"),
                &["computation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthongonality"),
                &["orthogonality"],
            ),
            (dictgen::InsensitiveStr::Ascii("desitny"), &["destiny"]),
            (
                dictgen::InsensitiveStr::Ascii("calculatsion"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdend"), &["depend"]),
            (dictgen::InsensitiveStr::Ascii("aslo"), &["also"]),
            (dictgen::InsensitiveStr::Ascii("blackjak"), &["blackjack"]),
            (dictgen::InsensitiveStr::Ascii("snet"), &["sent"]),
            (dictgen::InsensitiveStr::Ascii("miges"), &["midges"]),
            (dictgen::InsensitiveStr::Ascii("uesers"), &["users"]),
            (
                dictgen::InsensitiveStr::Ascii("negogtiation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificatd"),
                &["certificated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempeartures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabines"), &["cabinets"]),
            (
                dictgen::InsensitiveStr::Ascii("goosepumps"),
                &["goosebumps"],
            ),
            (dictgen::InsensitiveStr::Ascii("sprinke"), &["sprinkle"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncalization"),
                &["canonicalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("convinent"), &["convenient"]),
            (
                dictgen::InsensitiveStr::Ascii("thights"),
                &["tights", "thighs", "fights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processessing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("compiller"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("forewards"),
                &["forwards", "forewords", "forewarns"],
            ),
            (dictgen::InsensitiveStr::Ascii("mmaped"), &["mapped"]),
            (dictgen::InsensitiveStr::Ascii("eevery"), &["every"]),
            (
                dictgen::InsensitiveStr::Ascii("condifuration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anouncement"),
                &["announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incooperate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("idicated"), &["indicated"]),
            (dictgen::InsensitiveStr::Ascii("umberlla"), &["umbrella"]),
            (
                dictgen::InsensitiveStr::Ascii("perihperal"),
                &["peripheral"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheatta"), &["cheetah"]),
            (
                dictgen::InsensitiveStr::Ascii("caterpilar"),
                &["caterpillar"],
            ),
            (dictgen::InsensitiveStr::Ascii("effectly"), &["effectively"]),
            (
                dictgen::InsensitiveStr::Ascii("inidvidual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimality"), &["optimality"]),
            (dictgen::InsensitiveStr::Ascii("hygience"), &["hygiene"]),
            (
                dictgen::InsensitiveStr::Ascii("inportance"),
                &["importance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refletions"),
                &["reflections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproduciable"),
                &["reproduceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("frind"), &["friend"]),
            (
                dictgen::InsensitiveStr::Ascii("bracin"),
                &["bracing", "brain"],
            ),
            (dictgen::InsensitiveStr::Ascii("revereces"), &["references"]),
            (
                dictgen::InsensitiveStr::Ascii("difinition"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("finialised"), &["finalised"]),
            (
                dictgen::InsensitiveStr::Ascii("fumctioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("whitelsit"), &["whitelist"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreements"),
                &["agreements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkkpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("bottlneck"), &["bottleneck"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiamte"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperialst"),
                &["imperialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("donwgrades"),
                &["downgrades"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntrained"),
                &["constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsessful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("opinon"), &["opinion"]),
            (dictgen::InsensitiveStr::Ascii("chercksums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("staticlly"), &["statically"]),
            (dictgen::InsensitiveStr::Ascii("purposedly"), &["purposely"]),
            (dictgen::InsensitiveStr::Ascii("lern"), &["learn", "lean"]),
            (dictgen::InsensitiveStr::Ascii("releeased"), &["released"]),
            (
                dictgen::InsensitiveStr::Ascii("poped"),
                &["popped", "pooped"],
            ),
            (dictgen::InsensitiveStr::Ascii("develps"), &["develops"]),
            (dictgen::InsensitiveStr::Ascii("samckdown"), &["smackdown"]),
            (
                dictgen::InsensitiveStr::Ascii("tesellator"),
                &["tessellator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furtheremore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("keywoards"),
                &["keywords", "keyboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sistemically"),
                &["systemically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interventie"),
                &["intervene"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lietuenant"),
                &["lieutenant"],
            ),
            (dictgen::InsensitiveStr::Ascii("creted"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("stiple"), &["stipple"]),
            (dictgen::InsensitiveStr::Ascii("laoder"), &["loader"]),
            (
                dictgen::InsensitiveStr::Ascii("stadnardised"),
                &["standardised"],
            ),
            (dictgen::InsensitiveStr::Ascii("traslates"), &["translates"]),
            (
                dictgen::InsensitiveStr::Ascii("patchs"),
                &["patches", "paths"],
            ),
            (dictgen::InsensitiveStr::Ascii("insanley"), &["insanely"]),
            (dictgen::InsensitiveStr::Ascii("eugencis"), &["eugenics"]),
            (dictgen::InsensitiveStr::Ascii("resistas"), &["resists"]),
            (
                dictgen::InsensitiveStr::Ascii("supermarkerts"),
                &["supermarkets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viusalization"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("assistans"), &["assistants"]),
            (
                dictgen::InsensitiveStr::Ascii("descirption"),
                &["descriptions", "description"],
            ),
            (dictgen::InsensitiveStr::Ascii("longuest"), &["longest"]),
            (dictgen::InsensitiveStr::Ascii("architetc"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeuthoods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppposing"), &["supposing"]),
            (dictgen::InsensitiveStr::Ascii("accodred"), &["accorded"]),
            (dictgen::InsensitiveStr::Ascii("jugnling"), &["jungling"]),
            (
                dictgen::InsensitiveStr::Ascii("inceasing"),
                &["increasing", "unceasing", "incensing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questoining"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skillshoots"),
                &["skillshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("occouring"), &["occurring"]),
            (
                dictgen::InsensitiveStr::Ascii("unfreindly"),
                &["unfriendly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hosptialized"),
                &["hospitalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootsrap"), &["bootstrap"]),
            (dictgen::InsensitiveStr::Ascii("merecats"), &["meerkats"]),
            (dictgen::InsensitiveStr::Ascii("catgeory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("prospectos"), &["prospects"]),
            (dictgen::InsensitiveStr::Ascii("organzie"), &["organize"]),
            (dictgen::InsensitiveStr::Ascii("eignvalue"), &["eigenvalue"]),
            (
                dictgen::InsensitiveStr::Ascii("midnlessly"),
                &["mindlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entensions"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generocity"),
                &["generosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("monthe"), &["month"]),
            (dictgen::InsensitiveStr::Ascii("assings"), &["assigns"]),
            (dictgen::InsensitiveStr::Ascii("omitable"), &["omittable"]),
            (
                dictgen::InsensitiveStr::Ascii("combanators"),
                &["combinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpuld"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("amonut"), &["amount"]),
            (dictgen::InsensitiveStr::Ascii("gloabl"), &["global"]),
            (dictgen::InsensitiveStr::Ascii("cominings"), &["comings"]),
            (
                dictgen::InsensitiveStr::Ascii("contributs"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalase"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("functionable"),
                &["functional", "functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("versiona"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("gradiation"),
                &["graduation"],
            ),
            (dictgen::InsensitiveStr::Ascii("worrried"), &["worried"]),
            (
                dictgen::InsensitiveStr::Ascii("oscillaters"),
                &["oscillators", "oscillates"],
            ),
            (dictgen::InsensitiveStr::Ascii("jhondoe"), &["johndoe"]),
            (dictgen::InsensitiveStr::Ascii("rednerers"), &["renderers"]),
            (dictgen::InsensitiveStr::Ascii("chasiss"), &["chassis"]),
            (
                dictgen::InsensitiveStr::Ascii("initiailizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acheivements"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("procedue"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("boucing"), &["bouncing"]),
            (dictgen::InsensitiveStr::Ascii("cassim"), &["chasm"]),
            (dictgen::InsensitiveStr::Ascii("addictis"), &["addictions"]),
            (
                dictgen::InsensitiveStr::Ascii("sirectives"),
                &["directives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicatble"),
                &["predictable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associationg"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparancey"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("conncts"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("tablespon"), &["tablespoon"]),
            (
                dictgen::InsensitiveStr::Ascii("evoluation"),
                &["evaluations", "evaluation", "evolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("filty"), &["filthy"]),
            (dictgen::InsensitiveStr::Ascii("updae"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("exeution"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("gunslanger"),
                &["gunslinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("casuing"), &["causing"]),
            (
                dictgen::InsensitiveStr::Ascii("doumentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("corector"), &["corrector"]),
            (
                dictgen::InsensitiveStr::Ascii("ove"),
                &["oven", "over", "move"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condifures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requerimento"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proficienct"),
                &["proficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaneosly"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("naybours"), &["neighbours"]),
            (
                dictgen::InsensitiveStr::Ascii("portestants"),
                &["protestants"],
            ),
            (dictgen::InsensitiveStr::Ascii("passs"), &["pass"]),
            (
                dictgen::InsensitiveStr::Ascii("psycholigical"),
                &["psychological"],
            ),
            (dictgen::InsensitiveStr::Ascii("comsummes"), &["consumes"]),
            (
                dictgen::InsensitiveStr::Ascii("disollusioned"),
                &["disillusioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotionations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("popuplate"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("detaulting"),
                &["defaulting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spearately"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serializaable"),
                &["serializable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppresson"),
                &["suppression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentaties"),
                &["documentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("registrer"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("reorginised"),
                &["reorganised"],
            ),
            (dictgen::InsensitiveStr::Ascii("atatching"), &["attaching"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sublasses"), &["subclasses"]),
            (dictgen::InsensitiveStr::Ascii("changably"), &["changeably"]),
            (
                dictgen::InsensitiveStr::Ascii("temmporarily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autohorized"),
                &["authorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recepy"),
                &["recipe", "recopy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimizaers"),
                &["optimizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("craetes"),
                &["crates", "creates"],
            ),
            (dictgen::InsensitiveStr::Ascii("prejiduce"), &["prejudice"]),
            (dictgen::InsensitiveStr::Ascii("skeep"), &["skip"]),
            (
                dictgen::InsensitiveStr::Ascii("actiavtion"),
                &["activation"],
            ),
            (dictgen::InsensitiveStr::Ascii("launguages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("analogicaly"),
                &["analogically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rountripping"),
                &["roundtripping"],
            ),
            (dictgen::InsensitiveStr::Ascii("qualifers"), &["qualifiers"]),
            (dictgen::InsensitiveStr::Ascii("viusalize"), &["visualize"]),
            (
                dictgen::InsensitiveStr::Ascii("automagicaly"),
                &["automagically"],
            ),
            (dictgen::InsensitiveStr::Ascii("obviosuly"), &["obviously"]),
            (dictgen::InsensitiveStr::Ascii("innocous"), &["innocuous"]),
            (
                dictgen::InsensitiveStr::Ascii("orginization"),
                &["organization"],
            ),
            (dictgen::InsensitiveStr::Ascii("unicors"), &["unicorns"]),
            (dictgen::InsensitiveStr::Ascii("aoto"), &["auto"]),
            (dictgen::InsensitiveStr::Ascii("lisense"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("proprogation"),
                &["propagation"],
            ),
            (dictgen::InsensitiveStr::Ascii("obections"), &["objections"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiifed"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condescencion"),
                &["condescension"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sightstine"),
                &["sightstone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beleivably"),
                &["believably"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefiers"), &["specifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("descritpiton"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("clearted"), &["cleared"]),
            (
                dictgen::InsensitiveStr::Ascii("celebrato"),
                &["celebration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofee"), &["coffee"]),
            (dictgen::InsensitiveStr::Ascii("follwoes"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("ot"),
                &["to", "of", "or", "not", "it"],
            ),
            (dictgen::InsensitiveStr::Ascii("mozillia"), &["mozilla"]),
            (
                dictgen::InsensitiveStr::Ascii("configutration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhooods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrcitions"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streigtish"),
                &["straightish"],
            ),
            (dictgen::InsensitiveStr::Ascii("reomve"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("deivant"), &["deviant"]),
            (dictgen::InsensitiveStr::Ascii("loookup"), &["lookup"]),
            (dictgen::InsensitiveStr::Ascii("surgests"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("konckback"), &["knockback"]),
            (dictgen::InsensitiveStr::Ascii("debudgged"), &["debugged"]),
            (dictgen::InsensitiveStr::Ascii("exanmple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("compialtions"),
                &["compilations", "complications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscciative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lackrimosity"),
                &["lachrymosity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sorced"),
                &["sourced", "forced", "sorted"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhange"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("revolvr"), &["revolver"]),
            (
                dictgen::InsensitiveStr::Ascii("contempalte"),
                &["contemplate"],
            ),
            (dictgen::InsensitiveStr::Ascii("hauty"), &["haughty"]),
            (
                dictgen::InsensitiveStr::Ascii("interrupions"),
                &["interruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("soucre"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("noticiable"),
                &["noticeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("snippert"), &["snippet"]),
            (dictgen::InsensitiveStr::Ascii("worklfows"), &["workflows"]),
            (dictgen::InsensitiveStr::Ascii("generas"), &["generals"]),
            (dictgen::InsensitiveStr::Ascii("rotataion"), &["rotation"]),
            (dictgen::InsensitiveStr::Ascii("thumnails"), &["thumbnails"]),
            (dictgen::InsensitiveStr::Ascii("convicto"), &["conviction"]),
            (dictgen::InsensitiveStr::Ascii("attriburte"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("knuckels"), &["knuckles"]),
            (
                dictgen::InsensitiveStr::Ascii("consisntent"),
                &["consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrrange"), &["arrange"]),
            (
                dictgen::InsensitiveStr::Ascii("freqquencies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloding"), &["closing"]),
            (dictgen::InsensitiveStr::Ascii("awnsering"), &["answering"]),
            (
                dictgen::InsensitiveStr::Ascii("decalarators"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelisense"),
                &["intellisense"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecutes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("featching"), &["fetching"]),
            (dictgen::InsensitiveStr::Ascii("cofirm"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilsed"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylcing"), &["cycling"]),
            (dictgen::InsensitiveStr::Ascii("enounter"), &["encounter"]),
            (
                dictgen::InsensitiveStr::Ascii("governmnet"),
                &["government"],
            ),
            (dictgen::InsensitiveStr::Ascii("baloon"), &["balloon"]),
            (dictgen::InsensitiveStr::Ascii("standerds"), &["standards"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("documetns"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("bluethooth"), &["bluetooth"]),
            (
                dictgen::InsensitiveStr::Ascii("corectable"),
                &["correctable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("couroutines"),
                &["coroutines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prolbematic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicing"),
                &["explaining", "explicating", "explicitly", "explicit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deconstrcutor"),
                &["deconstructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clonning"),
                &["cloning", "clowning", "conning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cammanding"),
                &["commanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsever"), &["observer"]),
            (dictgen::InsensitiveStr::Ascii("triptick"), &["triptych"]),
            (
                dictgen::InsensitiveStr::Ascii("intregrations"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpired"), &["inspired"]),
            (dictgen::InsensitiveStr::Ascii("aproach"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("desroying"), &["destroying"]),
            (
                dictgen::InsensitiveStr::Ascii("assassiante"),
                &["assassinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrruption"),
                &["interruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpaction"),
                &["compaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("routiens"), &["routines"]),
            (
                dictgen::InsensitiveStr::Ascii("catholocisim"),
                &["catholicism"],
            ),
            (dictgen::InsensitiveStr::Ascii("remnans"), &["remnants"]),
            (
                dictgen::InsensitiveStr::Ascii("initilised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("boutqiue"), &["boutique"]),
            (dictgen::InsensitiveStr::Ascii("fractalpus"), &["fractals"]),
            (dictgen::InsensitiveStr::Ascii("collest"), &["collects"]),
            (
                dictgen::InsensitiveStr::Ascii("calcalator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("falshes"), &["flashes"]),
            (dictgen::InsensitiveStr::Ascii("apoints"), &["appoints"]),
            (dictgen::InsensitiveStr::Ascii("outsie"), &["outside"]),
            (dictgen::InsensitiveStr::Ascii("upacks"), &["unpacks"]),
            (dictgen::InsensitiveStr::Ascii("precisionn"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("customsies"),
                &["customises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidaton"),
                &["intimidation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoucning"),
                &["announcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulty"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("privilegeds"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplicy"), &["simplicity"]),
            (dictgen::InsensitiveStr::Ascii("channge"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("spacin"), &["spacing"]),
            (
                dictgen::InsensitiveStr::Ascii("costumisers"),
                &["customisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("canadains"), &["canadians"]),
            (
                dictgen::InsensitiveStr::Ascii("amened"),
                &["amended", "amend"],
            ),
            (dictgen::InsensitiveStr::Ascii("narcissts"), &["narcissist"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdendet"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidential"),
                &["accidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranfered"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfersom"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("notabley"), &["notably"]),
            (dictgen::InsensitiveStr::Ascii("thhe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("ststionary"),
                &["stationary"],
            ),
            (dictgen::InsensitiveStr::Ascii("doplhin"), &["dolphin"]),
            (dictgen::InsensitiveStr::Ascii("yopu"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbouthoods"),
                &["neighbourhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibity"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparts"),
                &["transports"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registre"),
                &["register", "registry"],
            ),
            (dictgen::InsensitiveStr::Ascii("cjanges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("characteors"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninistalling"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("breal"), &["break"]),
            (
                dictgen::InsensitiveStr::Ascii("symptumatically"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availbvility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("inrement"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("procecures"),
                &["procedures"],
            ),
            (dictgen::InsensitiveStr::Ascii("movei"), &["movie"]),
            (
                dictgen::InsensitiveStr::Ascii("restraing"),
                &["restraining"],
            ),
            (dictgen::InsensitiveStr::Ascii("perposals"), &["proposals"]),
            (
                dictgen::InsensitiveStr::Ascii("unexperience"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("exterme"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("commmited"), &["committed"]),
            (
                dictgen::InsensitiveStr::Ascii("underscroe"),
                &["underscore"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrivied"), &["arrived"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitucion"),
                &["prostitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("happpenings"),
                &["happenings"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesssing"), &["accessing"]),
            (
                dictgen::InsensitiveStr::Ascii("tyoe"),
                &["type", "toe", "toey"],
            ),
            (dictgen::InsensitiveStr::Ascii("seinor"), &["senior"]),
            (dictgen::InsensitiveStr::Ascii("enfoce"), &["enforce"]),
            (dictgen::InsensitiveStr::Ascii("coomments"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("necassery"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("exptectation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catagorized"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbak"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("seleceted"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("inbed"), &["imbed"]),
            (dictgen::InsensitiveStr::Ascii("handshkes"), &["handshakes"]),
            (
                dictgen::InsensitiveStr::Ascii("subsequenct"),
                &["subsequent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whould"),
                &["should", "would"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permessioned"),
                &["permissioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoyence"), &["annoyance"]),
            (
                dictgen::InsensitiveStr::Ascii("deliminated"),
                &["delimited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcriptus"),
                &["transcripts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reluctanctly"),
                &["reluctantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("misisons"), &["missions"]),
            (
                dictgen::InsensitiveStr::Ascii("canonalizes"),
                &["canonicalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exposicion"),
                &["exposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expolitation"),
                &["exploitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deskops"), &["desktops"]),
            (
                dictgen::InsensitiveStr::Ascii("overtapping"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatibale"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoective"),
                &["respective"],
            ),
            (dictgen::InsensitiveStr::Ascii("responsd"), &["responds"]),
            (dictgen::InsensitiveStr::Ascii("bounndary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("beleiv"),
                &["believe", "belief"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialialising"),
                &["serialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ulitmately"),
                &["ultimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("puting"), &["putting"]),
            (dictgen::InsensitiveStr::Ascii("brielfy"), &["briefly"]),
            (
                dictgen::InsensitiveStr::Ascii("algoruthmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illigitament"),
                &["illegitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("socail"), &["social"]),
            (dictgen::InsensitiveStr::Ascii("iranains"), &["iranians"]),
            (dictgen::InsensitiveStr::Ascii("declarees"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("objectivily"),
                &["objectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("compltely"), &["completely"]),
            (
                dictgen::InsensitiveStr::Ascii("defrentiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychologyst"),
                &["psychologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dinominator"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("proides"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("alogirthmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosntant"), &["constant"]),
            (
                dictgen::InsensitiveStr::Ascii("annnotations"),
                &["annotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decsriptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("organismer"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("prepering"), &["preparing"]),
            (dictgen::InsensitiveStr::Ascii("grouops"), &["groups"]),
            (dictgen::InsensitiveStr::Ascii("conived"), &["connived"]),
            (dictgen::InsensitiveStr::Ascii("seperaor"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("dircets"), &["directs"]),
            (dictgen::InsensitiveStr::Ascii("apperars"), &["appears"]),
            (dictgen::InsensitiveStr::Ascii("contrsted"), &["contrasted"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaised"),
                &["generalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incerceration"),
                &["incarceration"],
            ),
            (dictgen::InsensitiveStr::Ascii("docoment"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("unrestriced"),
                &["unrestricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlcude"), &["exclude"]),
            (dictgen::InsensitiveStr::Ascii("samle"), &["sample", "same"]),
            (
                dictgen::InsensitiveStr::Ascii("investemnt"),
                &["investments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contingancy"),
                &["contingency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upgradablilty"),
                &["upgradability"],
            ),
            (dictgen::InsensitiveStr::Ascii("thinn"), &["thin"]),
            (dictgen::InsensitiveStr::Ascii("intereste"), &["interested"]),
            (dictgen::InsensitiveStr::Ascii("pertrified"), &["petrified"]),
            (
                dictgen::InsensitiveStr::Ascii("costumization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconicle"), &["reconcile"]),
            (dictgen::InsensitiveStr::Ascii("lowerace"), &["lowercase"]),
            (dictgen::InsensitiveStr::Ascii("applogies"), &["apologies"]),
            (dictgen::InsensitiveStr::Ascii("trubador"), &["troubadour"]),
            (dictgen::InsensitiveStr::Ascii("narative"), &["narrative"]),
            (dictgen::InsensitiveStr::Ascii("functio"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("heruistics"),
                &["heuristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interviened"),
                &["interviewed"],
            ),
            (dictgen::InsensitiveStr::Ascii("perpare"), &["prepare"]),
            (dictgen::InsensitiveStr::Ascii("sprakling"), &["sparkling"]),
            (dictgen::InsensitiveStr::Ascii("cofused"), &["confused"]),
            (dictgen::InsensitiveStr::Ascii("closser"), &["closer"]),
            (
                dictgen::InsensitiveStr::Ascii("mysterieuse"),
                &["mysteries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excepection"),
                &["exception"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insesitive"),
                &["insensitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("illigal"), &["illegal"]),
            (dictgen::InsensitiveStr::Ascii("exractors"), &["extractors"]),
            (dictgen::InsensitiveStr::Ascii("flavous"), &["flavours"]),
            (dictgen::InsensitiveStr::Ascii("rovides"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("rigthfully"),
                &["rightfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemement"),
                &["implement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recgonised"),
                &["recognised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoziate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applciable"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("preserv"), &["preserve"]),
            (dictgen::InsensitiveStr::Ascii("bycicled"), &["bicycled"]),
            (dictgen::InsensitiveStr::Ascii("dpendents"), &["dependents"]),
            (
                dictgen::InsensitiveStr::Ascii("elegibility"),
                &["eligibility", "illegibility", "legibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("abundand"), &["abundant"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistantly"),
                &["inconsistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("hueristic"), &["heuristic"]),
            (
                dictgen::InsensitiveStr::Ascii("overrinding"),
                &["overriding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generetors"),
                &["generators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unterlying"),
                &["underlying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efficientcy"),
                &["efficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occationally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconcs"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("specififed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("benifiting"),
                &["benefiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("withouth"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("burocratic"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("detaills"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("jaguards"), &["jaguars"]),
            (dictgen::InsensitiveStr::Ascii("gorumet"), &["gourmet"]),
            (
                dictgen::InsensitiveStr::Ascii("modprobbing"),
                &["modprobing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spead"),
                &["spread", "speed", "spend", "speak", "spear"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intructional"),
                &["instructional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otrhographic"),
                &["orthographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charismatisch"),
                &["charismatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("thougth"), &["thought"]),
            (
                dictgen::InsensitiveStr::Ascii("pronounciation"),
                &["pronunciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregisterd"),
                &["unregistered"],
            ),
            (dictgen::InsensitiveStr::Ascii("sufixing"), &["suffixing"]),
            (dictgen::InsensitiveStr::Ascii("techneeks"), &["techniques"]),
            (dictgen::InsensitiveStr::Ascii("stictly"), &["strictly"]),
            (dictgen::InsensitiveStr::Ascii("assering"), &["asserting"]),
            (dictgen::InsensitiveStr::Ascii("supposedy"), &["supposedly"]),
            (
                dictgen::InsensitiveStr::Ascii("duplicatas"),
                &["duplicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discracefully"),
                &["disgracefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("collet"), &["collect"]),
            (
                dictgen::InsensitiveStr::Ascii("distribitons"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ggogles"),
                &["goggles", "googles"],
            ),
            (dictgen::InsensitiveStr::Ascii("procceses"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("incompartibility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrittive"),
                &["restrictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contriubting"),
                &["contributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimention"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquairums"), &["aquariums"]),
            (dictgen::InsensitiveStr::Ascii("prejudis"), &["prejudices"]),
            (dictgen::InsensitiveStr::Ascii("parge"), &["large"]),
            (
                dictgen::InsensitiveStr::Ascii("accelarators"),
                &["accelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("implents"), &["implements"]),
            (dictgen::InsensitiveStr::Ascii("princesas"), &["princess"]),
            (dictgen::InsensitiveStr::Ascii("thhinks"), &["thinks"]),
            (dictgen::InsensitiveStr::Ascii("peforms"), &["performs"]),
            (
                dictgen::InsensitiveStr::Ascii("narcissisitc"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("failsaves"), &["failsafes"]),
            (
                dictgen::InsensitiveStr::Ascii("custumisation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalzing"),
                &["capitalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("torphies"), &["trophies"]),
            (dictgen::InsensitiveStr::Ascii("creasin"), &["creasing"]),
            (dictgen::InsensitiveStr::Ascii("publishher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("expration"), &["expiration"]),
            (
                dictgen::InsensitiveStr::Ascii("initaliser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("stadards"), &["standards"]),
            (
                dictgen::InsensitiveStr::Ascii("aesthitics"),
                &["aesthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("capteurs"), &["captures"]),
            (dictgen::InsensitiveStr::Ascii("accsepting"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("termonology"),
                &["terminology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properites"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("crossin"), &["crossing"]),
            (
                dictgen::InsensitiveStr::Ascii("diferrentiates"),
                &["differentiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("diretory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("switchting"), &["switching"]),
            (dictgen::InsensitiveStr::Ascii("vehementy"), &["vehemently"]),
            (
                dictgen::InsensitiveStr::Ascii("consistensy"),
                &["consistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourgh"),
                &["neighbour"],
            ),
            (dictgen::InsensitiveStr::Ascii("charis"), &["chairs"]),
            (dictgen::InsensitiveStr::Ascii("keyowrd"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("projec"), &["project"]),
            (
                dictgen::InsensitiveStr::Ascii("distriutor"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfeccted"), &["perfected"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbateing"),
                &["masturbating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custoisers"),
                &["customisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("acurate"), &["accurate"]),
            (dictgen::InsensitiveStr::Ascii("herculeees"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("htose"), &["these", "those"]),
            (
                dictgen::InsensitiveStr::Ascii("instersectionality"),
                &["intersectionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("precacheed"), &["precached"]),
            (dictgen::InsensitiveStr::Ascii("hestitate"), &["hesitate"]),
            (
                dictgen::InsensitiveStr::Ascii("poindcloud"),
                &["pointcloud"],
            ),
            (dictgen::InsensitiveStr::Ascii("folllowong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("residude"), &["residue"]),
            (
                dictgen::InsensitiveStr::Ascii("unacknowleged"),
                &["unacknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("imediate"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("cababilities"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ontact"),
                &["contact", "intact"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepre"), &["sphere"]),
            (dictgen::InsensitiveStr::Ascii("subsizide"), &["subsidize"]),
            (
                dictgen::InsensitiveStr::Ascii("categoization"),
                &["categorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("benchamark"), &["benchmark"]),
            (dictgen::InsensitiveStr::Ascii("asteriods"), &["asteroids"]),
            (
                dictgen::InsensitiveStr::Ascii("scrools"),
                &["scrolls", "schools"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fermantation"),
                &["fermentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizatons"),
                &["organizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("polyar"), &["polar"]),
            (
                dictgen::InsensitiveStr::Ascii("undeinable"),
                &["undeniable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensinality"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circimcised"),
                &["circumcised"],
            ),
            (dictgen::InsensitiveStr::Ascii("qaulities"), &["qualities"]),
            (dictgen::InsensitiveStr::Ascii("qeustion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("refereneced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("wel"), &["well"]),
            (
                dictgen::InsensitiveStr::Ascii("blockchian"),
                &["blockchain"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interatively"),
                &["interactively", "iteratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customisaton"),
                &["customisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("prsets"), &["presets"]),
            (dictgen::InsensitiveStr::Ascii("baynoet"), &["bayonet"]),
            (dictgen::InsensitiveStr::Ascii("napom"), &["napalm"]),
            (dictgen::InsensitiveStr::Ascii("encrpting"), &["encrypting"]),
            (
                dictgen::InsensitiveStr::Ascii("respresenting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("casses"), &["cases"]),
            (dictgen::InsensitiveStr::Ascii("listenner"), &["listener"]),
            (dictgen::InsensitiveStr::Ascii("arre"), &["are"]),
            (
                dictgen::InsensitiveStr::Ascii("ministerios"),
                &["ministers"],
            ),
            (dictgen::InsensitiveStr::Ascii("oreder"), &["order"]),
            (dictgen::InsensitiveStr::Ascii("restroring"), &["restoring"]),
            (dictgen::InsensitiveStr::Ascii("volcanoe"), &["volcano"]),
            (dictgen::InsensitiveStr::Ascii("blockeras"), &["blockers"]),
            (
                dictgen::InsensitiveStr::Ascii("predictibly"),
                &["predictably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecutres"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("contanti"), &["contacting"]),
            (dictgen::InsensitiveStr::Ascii("poings"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("crusor"), &["cursor"]),
            (dictgen::InsensitiveStr::Ascii("mutisite"), &["multisite"]),
            (dictgen::InsensitiveStr::Ascii("asterisc"), &["asterisk"]),
            (
                dictgen::InsensitiveStr::Ascii("oficianado"),
                &["aficionado"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deauthenication"),
                &["deauthentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("cappably"), &["capably"]),
            (dictgen::InsensitiveStr::Ascii("speperates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("substite"), &["substitute"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangtement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("vieweres"), &["viewers"]),
            (dictgen::InsensitiveStr::Ascii("availabel"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("tryign"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratioon"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cominators"),
                &["combinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("retoric"), &["rhetoric"]),
            (
                dictgen::InsensitiveStr::Ascii("genralizations"),
                &["generalizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metabolitic"),
                &["metabolic"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsett"), &["offset"]),
            (dictgen::InsensitiveStr::Ascii("filterss"), &["filters"]),
            (
                dictgen::InsensitiveStr::Ascii("disresepcting"),
                &["disrespecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("ligthers"), &["lighters"]),
            (dictgen::InsensitiveStr::Ascii("anscestor"), &["ancestor"]),
            (dictgen::InsensitiveStr::Ascii("calgarry"), &["calgary"]),
            (
                dictgen::InsensitiveStr::Ascii("histroicaly"),
                &["historically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vectices"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("downlaods"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("embalance"), &["imbalance"]),
            (
                dictgen::InsensitiveStr::Ascii("thermomether"),
                &["thermometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediated"),
                &["immediate", "immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("inseting"), &["inserting"]),
            (dictgen::InsensitiveStr::Ascii("comresses"), &["compresses"]),
            (
                dictgen::InsensitiveStr::Ascii("omishience"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("whethor"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("stabalizer"),
                &["stabilizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("consensis"), &["consensus"]),
            (
                dictgen::InsensitiveStr::Ascii("pessimisstic"),
                &["pessimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("seacrh"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("collaboratoin"),
                &["collaboration"],
            ),
            (dictgen::InsensitiveStr::Ascii("forlders"), &["folders"]),
            (dictgen::InsensitiveStr::Ascii("cmments"), &["comments"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortuate"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurationns"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("genitaila"), &["genitalia"]),
            (dictgen::InsensitiveStr::Ascii("infiniment"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("unmodfitied"),
                &["unmodified"],
            ),
            (dictgen::InsensitiveStr::Ascii("moteured"), &["motored"]),
            (
                dictgen::InsensitiveStr::Ascii("friednship"),
                &["friendships"],
            ),
            (dictgen::InsensitiveStr::Ascii("dysphoira"), &["dysphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("committments"),
                &["commitments"],
            ),
            (dictgen::InsensitiveStr::Ascii("padds"), &["pads"]),
            (
                dictgen::InsensitiveStr::Ascii("projetctors"),
                &["projectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuestioned"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("rasberry"), &["raspberry"]),
            (
                dictgen::InsensitiveStr::Ascii("replacemenet"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructting"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("traines"), &["trainers"]),
            (dictgen::InsensitiveStr::Ascii("prototpye"), &["prototype"]),
            (
                dictgen::InsensitiveStr::Ascii("comparisms"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepentents"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("attch"), &["attach"]),
            (
                dictgen::InsensitiveStr::Ascii("determanism"),
                &["determinism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interporations"),
                &["interpolations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithums"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("anynmore"), &["anymore"]),
            (
                dictgen::InsensitiveStr::Ascii("dcokerd"),
                &["dockerd", "docked", "docker"],
            ),
            (dictgen::InsensitiveStr::Ascii("numbets"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("expresson"), &["expression"]),
            (
                dictgen::InsensitiveStr::Ascii("milesecond"),
                &["millisecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenendent"),
                &["dependent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("biologicaly"),
                &["biologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subpackges"),
                &["subpackages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimantally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrospekt"),
                &["retrospect"],
            ),
            (dictgen::InsensitiveStr::Ascii("interrface"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("astranauts"),
                &["astronauts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("residentual"),
                &["residential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wheather"),
                &["weather", "whether"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custonising"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrolbars"), &["scrollbars"]),
            (dictgen::InsensitiveStr::Ascii("materil"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("substansen"),
                &["substances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("univeriality"),
                &["universality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dominacion"),
                &["domination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enironment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("noified"), &["notified"]),
            (dictgen::InsensitiveStr::Ascii("vaylkrie"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("appliences"),
                &["appliances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pessimistisch"),
                &["pessimistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transforation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comprehesive"),
                &["comprehensive"],
            ),
            (dictgen::InsensitiveStr::Ascii("emplyees"), &["employees"]),
            (dictgen::InsensitiveStr::Ascii("remvoed"), &["removed"]),
            (
                dictgen::InsensitiveStr::Ascii("expceptional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timestanp"),
                &["timestamp", "timespan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discritization"),
                &["discretization"],
            ),
            (dictgen::InsensitiveStr::Ascii("enablig"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("cosumption"),
                &["consumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailzed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("operaion"), &["operation"]),
            (
                dictgen::InsensitiveStr::Ascii("discusssing"),
                &["discussing"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaied"), &["chained"]),
            (dictgen::InsensitiveStr::Ascii("specfifies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("assserting"), &["asserting"]),
            (dictgen::InsensitiveStr::Ascii("prefected"), &["perfected"]),
            (
                dictgen::InsensitiveStr::Ascii("summar"),
                &["summary", "summer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cipersuite"),
                &["ciphersuite"],
            ),
            (dictgen::InsensitiveStr::Ascii("perisisted"), &["persisted"]),
            (
                dictgen::InsensitiveStr::Ascii("condictioned"),
                &["conditioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underpowererd"),
                &["underpowered"],
            ),
            (dictgen::InsensitiveStr::Ascii("withouht"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregatore"),
                &["aggregate", "aggregator"],
            ),
            (dictgen::InsensitiveStr::Ascii("broacast"), &["broadcast"]),
            (
                dictgen::InsensitiveStr::Ascii("garenteeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("diamons"), &["diamonds"]),
            (
                dictgen::InsensitiveStr::Ascii("interpolatin"),
                &["interpolating", "interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("multline"), &["multiline"]),
            (dictgen::InsensitiveStr::Ascii("behavios"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("deconstructer"),
                &["deconstructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alltogether"),
                &["altogether"],
            ),
            (dictgen::InsensitiveStr::Ascii("knuckel"), &["knuckle"]),
            (dictgen::InsensitiveStr::Ascii("realling"), &["really"]),
            (
                dictgen::InsensitiveStr::Ascii("internatinalism"),
                &["internationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porportion"),
                &["proportion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermomoter"),
                &["thermometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomeotry"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("paackages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("incrememeted"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("supspects"), &["suspects"]),
            (dictgen::InsensitiveStr::Ascii("greator"), &["greater"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvience"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("tructure"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("origial"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("operato"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("recruites"), &["recruits"]),
            (dictgen::InsensitiveStr::Ascii("materaials"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("pregressive"),
                &["progressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("repostory"), &["repository"]),
            (
                dictgen::InsensitiveStr::Ascii("replacaiblity"),
                &["replaceability", "replicability"],
            ),
            (dictgen::InsensitiveStr::Ascii("towrad"), &["toward"]),
            (
                dictgen::InsensitiveStr::Ascii("operatating"),
                &["operating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasturizes"),
                &["pasteurizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifacto"), &["artifact"]),
            (
                dictgen::InsensitiveStr::Ascii("nayborhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("hrlping"), &["helping"]),
            (
                dictgen::InsensitiveStr::Ascii("conneection"),
                &["connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("colorts"), &["colors"]),
            (
                dictgen::InsensitiveStr::Ascii("sensacionalist"),
                &["sensationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrduced"), &["introduced"]),
            (dictgen::InsensitiveStr::Ascii("victorios"), &["victorious"]),
            (
                dictgen::InsensitiveStr::Ascii("developmetn"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clealy"),
                &["cleanly", "clearly"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaize"), &["localize"]),
            (dictgen::InsensitiveStr::Ascii("endpooints"), &["endpoints"]),
            (dictgen::InsensitiveStr::Ascii("millenia"), &["millennia"]),
            (dictgen::InsensitiveStr::Ascii("misile"), &["missile"]),
            (dictgen::InsensitiveStr::Ascii("ende"), &["end"]),
            (
                dictgen::InsensitiveStr::Ascii("reposiroties"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indipendence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawacks"), &["drawbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetrador"),
                &["perpetrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspicison"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdding"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("ameelarator"),
                &["ameliorator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconeects"),
                &["disconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contraddice"),
                &["contradicted"],
            ),
            (dictgen::InsensitiveStr::Ascii("paitient"), &["patient"]),
            (dictgen::InsensitiveStr::Ascii("itnerface"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("dependdents"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("bonnano"), &["bonanno"]),
            (
                dictgen::InsensitiveStr::Ascii("automatonic"),
                &["automation"],
            ),
            (dictgen::InsensitiveStr::Ascii("simliar"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("direktly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("devcies"), &["devices"]),
            (
                dictgen::InsensitiveStr::Ascii("akumulator"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloude"), &["cloud"]),
            (dictgen::InsensitiveStr::Ascii("livels"), &["levels"]),
            (
                dictgen::InsensitiveStr::Ascii("rounted"),
                &["routed", "rounded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rekommendation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("managemnet"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcly"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("correspend"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitmized"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("conciderable"),
                &["considerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetrifying"),
                &["certifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("seeging"), &["sieging"]),
            (dictgen::InsensitiveStr::Ascii("depency"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("structrues"),
                &["structures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknownledges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sotring"),
                &["storing", "sorting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addministrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("marlbes"), &["marbles"]),
            (dictgen::InsensitiveStr::Ascii("oueue"), &["queue"]),
            (
                dictgen::InsensitiveStr::Ascii("conveninence"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigurete"), &["cigarette"]),
            (dictgen::InsensitiveStr::Ascii("atested"), &["attested"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("silders"), &["sliders"]),
            (
                dictgen::InsensitiveStr::Ascii("documemtation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resurection"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("messeage"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("tempratures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("incorrent"), &["incorrect"]),
            (dictgen::InsensitiveStr::Ascii("stimulas"), &["stimulants"]),
            (
                dictgen::InsensitiveStr::Ascii("persceptive"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incrediblely"),
                &["incredibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithims"),
                &["algorithm", "algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporyrily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("uggesting"), &["suggesting"]),
            (dictgen::InsensitiveStr::Ascii("smebody"), &["somebody"]),
            (
                dictgen::InsensitiveStr::Ascii("tablepsaces"),
                &["tablespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("corretly"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("rektifications"),
                &["certifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("islamits"), &["islamist"]),
            (dictgen::InsensitiveStr::Ascii("swasitka"), &["swastika"]),
            (
                dictgen::InsensitiveStr::Ascii("comaparable"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("correspod"), &["correspond"]),
            (
                dictgen::InsensitiveStr::Ascii("subsystyems"),
                &["subsystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviniencys"),
                &["conveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("braodcasted"),
                &["broadcasted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throughiut"),
                &["throughput"],
            ),
            (dictgen::InsensitiveStr::Ascii("valeud"), &["valued"]),
            (
                dictgen::InsensitiveStr::Ascii("competetor"),
                &["competitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleves"), &["believes"]),
            (dictgen::InsensitiveStr::Ascii("credetial"), &["credential"]),
            (dictgen::InsensitiveStr::Ascii("duplicatd"), &["duplicated"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatves"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearranent"),
                &["rearrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrainted"),
                &["constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjuncting"),
                &["conjunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignoers"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("maginitude"), &["magnitude"]),
            (dictgen::InsensitiveStr::Ascii("finallly"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("megerd"), &["merged"]),
            (dictgen::InsensitiveStr::Ascii("complety"), &["completely"]),
            (dictgen::InsensitiveStr::Ascii("imapcted"), &["impacted"]),
            (
                dictgen::InsensitiveStr::Ascii("meranda"),
                &["veranda", "miranda"],
            ),
            (dictgen::InsensitiveStr::Ascii("scipted"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("quadroople"), &["quadruple"]),
            (dictgen::InsensitiveStr::Ascii("selecgting"), &["selecting"]),
            (dictgen::InsensitiveStr::Ascii("rady"), &["ready"]),
            (dictgen::InsensitiveStr::Ascii("carbien"), &["carbine"]),
            (dictgen::InsensitiveStr::Ascii("eliminte"), &["eliminate"]),
            (dictgen::InsensitiveStr::Ascii("alwauys"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("mulithreads"),
                &["multithreads"],
            ),
            (dictgen::InsensitiveStr::Ascii("playwrite"), &["playwright"]),
            (dictgen::InsensitiveStr::Ascii("renoylds"), &["reynolds"]),
            (dictgen::InsensitiveStr::Ascii("presnters"), &["presenters"]),
            (
                dictgen::InsensitiveStr::Ascii("proffesing"),
                &["professing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arciving"), &["archiving"]),
            (dictgen::InsensitiveStr::Ascii("bizarely"), &["bizarrely"]),
            (
                dictgen::InsensitiveStr::Ascii("deprivaton"),
                &["deprivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("lefitst"), &["leftist"]),
            (dictgen::InsensitiveStr::Ascii("thair"), &["their", "there"]),
            (dictgen::InsensitiveStr::Ascii("capitola"), &["capital"]),
            (dictgen::InsensitiveStr::Ascii("enhacning"), &["enhancing"]),
            (
                dictgen::InsensitiveStr::Ascii("sociopathis"),
                &["sociopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("spectraply"), &["spectral"]),
            (dictgen::InsensitiveStr::Ascii("presumin"), &["presuming"]),
            (
                dictgen::InsensitiveStr::Ascii("connotaion"),
                &["connotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transolation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("iserting"), &["inserting"]),
            (
                dictgen::InsensitiveStr::Ascii("corparations"),
                &["corporations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tasbar"), &["taskbar"]),
            (dictgen::InsensitiveStr::Ascii("protocls"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("commiteed"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("privte"), &["private"]),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("midpoins"),
                &["midpoint", "midpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("boths"), &["both"]),
            (dictgen::InsensitiveStr::Ascii("srikeout"), &["strikeout"]),
            (dictgen::InsensitiveStr::Ascii("useage"), &["usage"]),
            (
                dictgen::InsensitiveStr::Ascii("pease"),
                &["lease", "peace", "piece", "please"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("produciton"),
                &["production"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrcts"), &["abstracts"]),
            (dictgen::InsensitiveStr::Ascii("aoplogies"), &["apologies"]),
            (dictgen::InsensitiveStr::Ascii("opimisers"), &["optimisers"]),
            (
                dictgen::InsensitiveStr::Ascii("projectilt"),
                &["projectile"],
            ),
            (dictgen::InsensitiveStr::Ascii("delink"), &["unlink"]),
            (dictgen::InsensitiveStr::Ascii("doble"), &["double"]),
            (
                dictgen::InsensitiveStr::Ascii("challanging"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginating"),
                &["originating"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraside"), &["paradise"]),
            (dictgen::InsensitiveStr::Ascii("templetes"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("unterleaved"),
                &["interleaved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referenecs"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoun"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("enfored"), &["enforced"]),
            (dictgen::InsensitiveStr::Ascii("spoacing"), &["spacing"]),
            (dictgen::InsensitiveStr::Ascii("traverlers"), &["traverse"]),
            (dictgen::InsensitiveStr::Ascii("avatares"), &["avatars"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstuctors"),
                &["reconstructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colleage"),
                &["colleagues", "colleague"],
            ),
            (dictgen::InsensitiveStr::Ascii("watkings"), &["watkins"]),
            (dictgen::InsensitiveStr::Ascii("multpiles"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("determind"), &["determined"]),
            (
                dictgen::InsensitiveStr::Ascii("challeging"),
                &["challenging"],
            ),
            (dictgen::InsensitiveStr::Ascii("alreayd"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("chosse"),
                &["choose", "chose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchagne"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("condascending"),
                &["condescending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbrearable"),
                &["unbreakable"],
            ),
            (dictgen::InsensitiveStr::Ascii("yopur"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("elimnating"),
                &["eliminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbibtarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualizacion"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("formalhaut"), &["fomalhaut"]),
            (dictgen::InsensitiveStr::Ascii("africains"), &["africans"]),
            (dictgen::InsensitiveStr::Ascii("indicato"), &["indication"]),
            (dictgen::InsensitiveStr::Ascii("scracth"), &["scratch"]),
            (
                dictgen::InsensitiveStr::Ascii("confidentally"),
                &["confidentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoft"), &["shift", "short"]),
            (dictgen::InsensitiveStr::Ascii("rgaph"), &["graph"]),
            (
                dictgen::InsensitiveStr::Ascii("encluded"),
                &["included", "excluded"],
            ),
            (dictgen::InsensitiveStr::Ascii("requuest"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("estiamated"), &["estimated"]),
            (
                dictgen::InsensitiveStr::Ascii("identificator"),
                &["identifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("residencial"),
                &["residential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphsyics"),
                &["metaphysics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criminalul"),
                &["criminally"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathcups"), &["matchups"]),
            (dictgen::InsensitiveStr::Ascii("exportint"), &["exporting"]),
            (dictgen::InsensitiveStr::Ascii("cancle"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("parises"), &["praises"]),
            (
                dictgen::InsensitiveStr::Ascii("secification"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefying"), &["specifying"]),
            (dictgen::InsensitiveStr::Ascii("deprecte"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("topoogy"), &["topology"]),
            (
                dictgen::InsensitiveStr::Ascii("apporoximated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("diecable"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("sharloton"), &["charlatan"]),
            (dictgen::InsensitiveStr::Ascii("suppossed"), &["supposed"]),
            (
                dictgen::InsensitiveStr::Ascii("survivebility"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abundancies"),
                &["abundances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("langnguages"),
                &["languages"],
            ),
            (dictgen::InsensitiveStr::Ascii("loaclly"), &["locally"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeorhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("horizontes"), &["horizons"]),
            (
                dictgen::InsensitiveStr::Ascii("configurationn"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocmpleting"),
                &["autocompleting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passangers"),
                &["passengers"],
            ),
            (dictgen::InsensitiveStr::Ascii("algoritms"), &["algorithms"]),
            (
                dictgen::InsensitiveStr::Ascii("exprssions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocne"), &["once"]),
            (dictgen::InsensitiveStr::Ascii("doign"), &["doing"]),
            (dictgen::InsensitiveStr::Ascii("intrduce"), &["introduce"]),
            (dictgen::InsensitiveStr::Ascii("objetive"), &["objective"]),
            (dictgen::InsensitiveStr::Ascii("israelits"), &["israelis"]),
            (
                dictgen::InsensitiveStr::Ascii("tansformer"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("naivity"), &["naivety"]),
            (
                dictgen::InsensitiveStr::Ascii("overclokcing"),
                &["overclocking"],
            ),
            (dictgen::InsensitiveStr::Ascii("spanisch"), &["spanish"]),
            (
                dictgen::InsensitiveStr::Ascii("construstor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproduciability"),
                &["reproducibility", "reproduceability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathamatics"),
                &["mathematics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacifically"),
                &["specifically", "pacifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("similarlly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("beuraucratic"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("defqault"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("complilation"),
                &["complication", "compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("istantiates"),
                &["instantiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interperated"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeling"), &["repelling"]),
            (
                dictgen::InsensitiveStr::Ascii("capabiilities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("socripts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("dissonante"),
                &["dissonance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sintakt"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("languafe"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("custonary"), &["customary"]),
            (
                dictgen::InsensitiveStr::Ascii("fromatters"),
                &["formatters"],
            ),
            (dictgen::InsensitiveStr::Ascii("requestesd"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("girafe"), &["giraffe"]),
            (
                dictgen::InsensitiveStr::Ascii("magnificant"),
                &["magnificent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decorativos"),
                &["decorations"],
            ),
            (dictgen::InsensitiveStr::Ascii("orphanes"), &["orphans"]),
            (
                dictgen::InsensitiveStr::Ascii("trnaslations"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlasion"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metrololitan"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptio"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aletrnated"),
                &["alternated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexiclinic"),
                &["hexaclinic"],
            ),
            (dictgen::InsensitiveStr::Ascii("databsaes"), &["databases"]),
            (
                dictgen::InsensitiveStr::Ascii("partipant"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("joruney"), &["journey"]),
            (dictgen::InsensitiveStr::Ascii("gauranttee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("intraspection"),
                &["introspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstlaled"),
                &["uninstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sessons"),
                &["sessions", "seasons", "lessons"],
            ),
            (dictgen::InsensitiveStr::Ascii("igores"), &["ignores"]),
            (
                dictgen::InsensitiveStr::Ascii("subesquently"),
                &["subsequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("synomyms"), &["synonyms"]),
            (dictgen::InsensitiveStr::Ascii("registrate"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("requestd"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("joanthan"), &["jonathan"]),
            (
                dictgen::InsensitiveStr::Ascii("neuralogical"),
                &["neurological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refereshing"),
                &["refreshing"],
            ),
            (dictgen::InsensitiveStr::Ascii("rednered"), &["rendered"]),
            (dictgen::InsensitiveStr::Ascii("modfifier"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginisers"),
                &["recognisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluculator"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaquesting"),
                &["requesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("initia"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("inevitabile"),
                &["inevitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectopned"), &["sectioned"]),
            (dictgen::InsensitiveStr::Ascii("extrators"), &["extractors"]),
            (
                dictgen::InsensitiveStr::Ascii("additionall"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepresent"), &["represent"]),
            (
                dictgen::InsensitiveStr::Ascii("constructings"),
                &["constructing", "constructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("assimile"), &["assimilate"]),
            (
                dictgen::InsensitiveStr::Ascii("docuenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrumented"),
                &["instrumented"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupliates"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("transormations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("packkage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("accessiblilty"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("resonses"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("transalating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("readigs"), &["readings"]),
            (
                dictgen::InsensitiveStr::Ascii("expited"),
                &["expedited", "expired"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixeing"), &["fixing"]),
            (dictgen::InsensitiveStr::Ascii("atructure"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("katastrophic"),
                &["catastrophic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guaruanted"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serializatin"),
                &["serialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cantained"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("ungreatfull"),
                &["ungrateful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adaquetely"),
                &["adequately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directorguy"),
                &["directory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construtor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asymmectric"),
                &["asymmetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identidied"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("asended"), &["ascended"]),
            (dictgen::InsensitiveStr::Ascii("terminial"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("comandeered"),
                &["commandeered"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkced"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("utlrasound"),
                &["ultrasound"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conseguence"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destributors"),
                &["distributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("lizense"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("repoduces"), &["reproduces"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnmitted"),
                &["transmitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("asure"), &["assure"]),
            (
                dictgen::InsensitiveStr::Ascii("bestiallity"),
                &["bestiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parmaeters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("composute"), &["composite"]),
            (dictgen::InsensitiveStr::Ascii("submitt"), &["submit"]),
            (dictgen::InsensitiveStr::Ascii("agrred"), &["agreed"]),
            (
                dictgen::InsensitiveStr::Ascii("associsted"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("englist"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("opposties"), &["opposites"]),
            (
                dictgen::InsensitiveStr::Ascii("intefacing"),
                &["interfacing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenanet"), &["tenant"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluaters"),
                &["evaluators", "evaluates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticisng"),
                &["criticising"],
            ),
            (dictgen::InsensitiveStr::Ascii("avarege"), &["average"]),
            (dictgen::InsensitiveStr::Ascii("ryamond"), &["raymond"]),
            (
                dictgen::InsensitiveStr::Ascii("indentifies"),
                &["identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicatia"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("posistioned"),
                &["positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerabe"),
                &["considerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("incongito"), &["incognito"]),
            (dictgen::InsensitiveStr::Ascii("geerally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("eddges"), &["edges"]),
            (dictgen::InsensitiveStr::Ascii("bareclona"), &["barcelona"]),
            (dictgen::InsensitiveStr::Ascii("specialz"), &["specialize"]),
            (dictgen::InsensitiveStr::Ascii("districct"), &["district"]),
            (dictgen::InsensitiveStr::Ascii("prfiles"), &["profiles"]),
            (dictgen::InsensitiveStr::Ascii("regestry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("nimphal"), &["nymphal"]),
            (
                dictgen::InsensitiveStr::Ascii("interrutps"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralellise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("expesting"), &["expecting"]),
            (dictgen::InsensitiveStr::Ascii("preculdes"), &["precludes"]),
            (
                dictgen::InsensitiveStr::Ascii("turnk"),
                &["trunk", "turnkey", "turn"],
            ),
            (dictgen::InsensitiveStr::Ascii("acticity"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("namnespace"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("mthod"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("categor"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("stabliize"), &["stabilize"]),
            (dictgen::InsensitiveStr::Ascii("iterleave"), &["interleave"]),
            (
                dictgen::InsensitiveStr::Ascii("idetifiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programemers"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("brnach"), &["branch"]),
            (
                dictgen::InsensitiveStr::Ascii("resemblence"),
                &["resemblance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsolocited"),
                &["unsolicited"],
            ),
            (dictgen::InsensitiveStr::Ascii("paremeter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisaton"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferring"),
                &["preferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaiant"), &["variant"]),
            (dictgen::InsensitiveStr::Ascii("triangels"), &["triangles"]),
            (dictgen::InsensitiveStr::Ascii("cirdular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("executioneers"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("uest"), &["quest"]),
            (dictgen::InsensitiveStr::Ascii("thinkabel"), &["thinkable"]),
            (dictgen::InsensitiveStr::Ascii("reitterate"), &["reiterate"]),
            (dictgen::InsensitiveStr::Ascii("estimatied"), &["estimated"]),
            (
                dictgen::InsensitiveStr::Ascii("dicovering"),
                &["discovering"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsettin"), &["unsetting"]),
            (dictgen::InsensitiveStr::Ascii("invisble"), &["invisible"]),
            (
                dictgen::InsensitiveStr::Ascii("hurricaines"),
                &["hurricanes"],
            ),
            (dictgen::InsensitiveStr::Ascii("ararbic"), &["arabic"]),
            (
                dictgen::InsensitiveStr::Ascii("ddressed"),
                &["addressed", "dressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("debouced"), &["debounced"]),
            (
                dictgen::InsensitiveStr::Ascii("asympyotic"),
                &["asymptotic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continutously"),
                &["continuously"],
            ),
            (dictgen::InsensitiveStr::Ascii("manualyy"), &["manually"]),
            (dictgen::InsensitiveStr::Ascii("addressd"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("inspiraste"), &["inspires"]),
            (
                dictgen::InsensitiveStr::Ascii("separartely"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("decleared"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("indroduces"),
                &["introduces"],
            ),
            (dictgen::InsensitiveStr::Ascii("steramer"), &["streamer"]),
            (
                dictgen::InsensitiveStr::Ascii("cherching"),
                &["checking", "churching"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitamines"), &["vitamins"]),
            (dictgen::InsensitiveStr::Ascii("starding"), &["starting"]),
            (dictgen::InsensitiveStr::Ascii("discourge"), &["discourage"]),
            (dictgen::InsensitiveStr::Ascii("coered"), &["coerced"]),
            (
                dictgen::InsensitiveStr::Ascii("simmetricly"),
                &["symmetricly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("riskay"),
                &["risky", "risqué"],
            ),
            (dictgen::InsensitiveStr::Ascii("acsend"), &["ascend"]),
            (dictgen::InsensitiveStr::Ascii("chalenger"), &["challenger"]),
            (dictgen::InsensitiveStr::Ascii("contenet"), &["content"]),
            (
                dictgen::InsensitiveStr::Ascii("techinician"),
                &["technician"],
            ),
            (dictgen::InsensitiveStr::Ascii("exitsts"), &["exists"]),
            (dictgen::InsensitiveStr::Ascii("secceeds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("comnplexes"), &["complexes"]),
            (dictgen::InsensitiveStr::Ascii("laodouts"), &["loadouts"]),
            (
                dictgen::InsensitiveStr::Ascii("overcompansate"),
                &["overcompensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplication"),
                &["duplication", "replication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deteremined"),
                &["determined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenaticates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cips"), &["chips"]),
            (dictgen::InsensitiveStr::Ascii("delted"), &["deleted"]),
            (
                dictgen::InsensitiveStr::Ascii("respwaning"),
                &["respawning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkpoits"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availavility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrepret"), &["interpret"]),
            (dictgen::InsensitiveStr::Ascii("tuplets"), &["tuples"]),
            (dictgen::InsensitiveStr::Ascii("micorcode"), &["microcode"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptionaly"),
                &["exceptionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("exucuted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("comlicating"),
                &["complicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indistingusihable"),
                &["indistinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("finailse"), &["finalise"]),
            (
                dictgen::InsensitiveStr::Ascii("opitionally"),
                &["optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzelos"),
                &["overzealous"],
            ),
            (dictgen::InsensitiveStr::Ascii("frendlies"), &["friendlies"]),
            (
                dictgen::InsensitiveStr::Ascii("colaborative"),
                &["collaborative"],
            ),
            (dictgen::InsensitiveStr::Ascii("securly"), &["securely"]),
            (dictgen::InsensitiveStr::Ascii("inturprett"), &["interpret"]),
            (dictgen::InsensitiveStr::Ascii("redirecct"), &["redirect"]),
            (
                dictgen::InsensitiveStr::Ascii("checkpiont"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("leanr"),
                &["lean", "learn", "leaner"],
            ),
            (dictgen::InsensitiveStr::Ascii("polgon"), &["polygon"]),
            (
                dictgen::InsensitiveStr::Ascii("manafactured"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabed"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("overwize"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarates"),
                &["ameliorates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intraversion"),
                &["introversion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispalyable"),
                &["displayable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realtionship"),
                &["relationship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatablities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("logaritmically"),
                &["logarithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devleoping"),
                &["developing"],
            ),
            (dictgen::InsensitiveStr::Ascii("protocals"), &["protocols"]),
            (
                dictgen::InsensitiveStr::Ascii("synchroize"),
                &["synchronize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pssibility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambassators"),
                &["ambassadors"],
            ),
            (dictgen::InsensitiveStr::Ascii("copyting"), &["copying"]),
            (dictgen::InsensitiveStr::Ascii("tantrumers"), &["tantrums"]),
            (dictgen::InsensitiveStr::Ascii("literatute"), &["literate"]),
            (
                dictgen::InsensitiveStr::Ascii("translatation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("grobally"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("dervies"), &["derives"]),
            (dictgen::InsensitiveStr::Ascii("metatabe"), &["metatable"]),
            (
                dictgen::InsensitiveStr::Ascii("presistable"),
                &["persistable"],
            ),
            (dictgen::InsensitiveStr::Ascii("objct"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("beight"), &["height"]),
            (
                dictgen::InsensitiveStr::Ascii("autometically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("gorw"), &["gore", "grow"]),
            (
                dictgen::InsensitiveStr::Ascii("opportuniste"),
                &["opportunities"],
            ),
            (dictgen::InsensitiveStr::Ascii("wass"), &["was"]),
            (dictgen::InsensitiveStr::Ascii("accurancy"), &["accuracy"]),
            (
                dictgen::InsensitiveStr::Ascii("stardardize"),
                &["standardize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rendazyvooed"),
                &["rendezvoused"],
            ),
            (dictgen::InsensitiveStr::Ascii("prohpets"), &["prophets"]),
            (dictgen::InsensitiveStr::Ascii("freezin"), &["freezing"]),
            (dictgen::InsensitiveStr::Ascii("diamater"), &["diameter"]),
            (
                dictgen::InsensitiveStr::Ascii("unreoverable"),
                &["unrecoverable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanatizing"),
                &["sanitizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shamen"),
                &["shaman", "shamans"],
            ),
            (dictgen::InsensitiveStr::Ascii("hedgehodge"), &["hedgehog"]),
            (
                dictgen::InsensitiveStr::Ascii("extravagina"),
                &["extravagant"],
            ),
            (dictgen::InsensitiveStr::Ascii("udpdate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("ofisionados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("geniuss"), &["geniuses"]),
            (dictgen::InsensitiveStr::Ascii("fushas"), &["fuchsias"]),
            (
                dictgen::InsensitiveStr::Ascii("ressociation"),
                &["reassociation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underfowing"),
                &["underflowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("zepplein"), &["zeppelin"]),
            (dictgen::InsensitiveStr::Ascii("pedophille"), &["pedophile"]),
            (dictgen::InsensitiveStr::Ascii("wathers"), &["watchers"]),
            (
                dictgen::InsensitiveStr::Ascii("excetpional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("formulars"), &["formulas"]),
            (
                dictgen::InsensitiveStr::Ascii("repersented"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("complicati"), &["complicit"]),
            (dictgen::InsensitiveStr::Ascii("flasghip"), &["flagship"]),
            (dictgen::InsensitiveStr::Ascii("shouldes"), &["shoulders"]),
            (dictgen::InsensitiveStr::Ascii("cimpiler"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("dyrection"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("commans"),
                &["commands", "commas", "commons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interlectually"),
                &["intellectually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relazation"),
                &["relaxation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exectuable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peroperties"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("thinest"), &["thinnest"]),
            (
                dictgen::InsensitiveStr::Ascii("vitualizing"),
                &["virtualizing", "visualizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condfiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("currected"), &["corrected"]),
            (dictgen::InsensitiveStr::Ascii("asorbed"), &["absorbed"]),
            (dictgen::InsensitiveStr::Ascii("asssignees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("exponentialy"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attachmet"),
                &["attachments"],
            ),
            (dictgen::InsensitiveStr::Ascii("manangers"), &["managers"]),
            (dictgen::InsensitiveStr::Ascii("otherviese"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("dictionaly"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("finnished"), &["finished"]),
            (
                dictgen::InsensitiveStr::Ascii("transylvannia"),
                &["transylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carthographer"),
                &["cartographer"],
            ),
            (dictgen::InsensitiveStr::Ascii("expiried"), &["expired"]),
            (
                dictgen::InsensitiveStr::Ascii("respresentation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawng"), &["drawing"]),
            (dictgen::InsensitiveStr::Ascii("qeuries"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenstions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differntiates"),
                &["differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomacically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vesion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("analysisis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("circumcission"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("espressino"), &["espresso"]),
            (dictgen::InsensitiveStr::Ascii("spearated"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("bayblon"), &["babylon"]),
            (dictgen::InsensitiveStr::Ascii("elments"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("encryp"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("poihnter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("investigativos"),
                &["investigations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disussions"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attementing"),
                &["attempting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inflateble"),
                &["inflatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("candidatxs"),
                &["candidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbeut"), &["neighbour"]),
            (
                dictgen::InsensitiveStr::Ascii("registreing"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associetion"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogynisitc"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("attendent"), &["attendant"]),
            (dictgen::InsensitiveStr::Ascii("supoorts"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("pinoeer"), &["pioneer"]),
            (
                dictgen::InsensitiveStr::Ascii("associatates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notififations"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("versionns"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("fisisist"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("bookwarm"), &["bookworm"]),
            (
                dictgen::InsensitiveStr::Ascii("situtaions"),
                &["situations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aciton"), &["action"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotioable"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("hydratin"), &["hydration"]),
            (
                dictgen::InsensitiveStr::Ascii("playthrogh"),
                &["playthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumned"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("experiations"),
                &["expirations"],
            ),
            (dictgen::InsensitiveStr::Ascii("queuable"), &["queueable"]),
            (dictgen::InsensitiveStr::Ascii("boggus"), &["bogus"]),
            (dictgen::InsensitiveStr::Ascii("uninamous"), &["unanimous"]),
            (
                dictgen::InsensitiveStr::Ascii("abtraction"),
                &["abstraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ankshusness"),
                &["anxiousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("justifing"), &["justifying"]),
            (dictgen::InsensitiveStr::Ascii("irritaties"), &["irritate"]),
            (dictgen::InsensitiveStr::Ascii("revivin"), &["reviving"]),
            (
                dictgen::InsensitiveStr::Ascii("adventuers"),
                &["adventures", "adventurers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predocessors"),
                &["predecessors"],
            ),
            (dictgen::InsensitiveStr::Ascii("expernal"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("registratino"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnsupported"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installtions"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pessiary"), &["pessary"]),
            (
                dictgen::InsensitiveStr::Ascii("confedence"),
                &["conference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stanardising"),
                &["standardising"],
            ),
            (dictgen::InsensitiveStr::Ascii("optmized"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("transpacency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autheniticates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incroporate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("curising"), &["cruising"]),
            (
                dictgen::InsensitiveStr::Ascii("counterfiets"),
                &["counterfeit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sourroundings"),
                &["surroundings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrespecing"),
                &["disrespecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppinionated"),
                &["opinionated"],
            ),
            (dictgen::InsensitiveStr::Ascii("appreciat"), &["appreciate"]),
            (
                dictgen::InsensitiveStr::Ascii("supermare"),
                &["supermarket"],
            ),
            (dictgen::InsensitiveStr::Ascii("negothiate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("cll"), &["call"]),
            (dictgen::InsensitiveStr::Ascii("denpended"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("santizes"), &["sanitizes"]),
            (
                dictgen::InsensitiveStr::Ascii("pupulations"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stormfromt"),
                &["stormfront"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntehtics"),
                &["synthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequnence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("compalint"), &["compliant"]),
            (
                dictgen::InsensitiveStr::Ascii("redistrubition"),
                &["redistribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dergeistered"),
                &["deregistered"],
            ),
            (dictgen::InsensitiveStr::Ascii("entchanted"), &["enchanted"]),
            (dictgen::InsensitiveStr::Ascii("mouvement"), &["movement"]),
            (dictgen::InsensitiveStr::Ascii("hense"), &["hence"]),
            (dictgen::InsensitiveStr::Ascii("syncting"), &["syncing"]),
            (
                dictgen::InsensitiveStr::Ascii("correcponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructies"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionned"),
                &["conditioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("immmediate"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminacion"),
                &["contamination"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminalis"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("defeaulting"),
                &["defaulting"],
            ),
            (dictgen::InsensitiveStr::Ascii("linix"), &["linux"]),
            (dictgen::InsensitiveStr::Ascii("reult"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("cratin"),
                &["crating", "gratin"],
            ),
            (dictgen::InsensitiveStr::Ascii("logsitics"), &["logistics"]),
            (dictgen::InsensitiveStr::Ascii("disover"), &["discover"]),
            (dictgen::InsensitiveStr::Ascii("hathcing"), &["hatching"]),
            (dictgen::InsensitiveStr::Ascii("wahts"), &["whats"]),
            (
                dictgen::InsensitiveStr::Ascii("regstration"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrimintor"),
                &["discriminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("execututed"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("resotrer"), &["restorer"]),
            (
                dictgen::InsensitiveStr::Ascii("restructing"),
                &["restricting", "restructuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartilege"), &["cartilage"]),
            (dictgen::InsensitiveStr::Ascii("teched"), &["taught"]),
            (dictgen::InsensitiveStr::Ascii("resevered"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("hijkack"), &["hijack"]),
            (dictgen::InsensitiveStr::Ascii("mclarean"), &["mclaren"]),
            (dictgen::InsensitiveStr::Ascii("spefiies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("depitcs"), &["depicts"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherisations"),
                &["characterisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("abnormaly"), &["abnormally"]),
            (dictgen::InsensitiveStr::Ascii("tyope"), &["type"]),
            (dictgen::InsensitiveStr::Ascii("activley"), &["actively"]),
            (dictgen::InsensitiveStr::Ascii("advertier"), &["advertiser"]),
            (
                dictgen::InsensitiveStr::Ascii("regidstered"),
                &["registered"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncknown"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("birghtens"), &["brightens"]),
            (
                dictgen::InsensitiveStr::Ascii("calles"),
                &["calls", "called", "caller", "callers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorizatoin"),
                &["authorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("hosptials"), &["hospitals"]),
            (
                dictgen::InsensitiveStr::Ascii("perpendicualr"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangmenets"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximim"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("alpah"), &["alpha"]),
            (
                dictgen::InsensitiveStr::Ascii("ivalidated"),
                &["invalidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenanigains"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("fricton"), &["friction"]),
            (dictgen::InsensitiveStr::Ascii("sinistre"), &["sinister"]),
            (
                dictgen::InsensitiveStr::Ascii("handshaling"),
                &["handshaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tournements"),
                &["tournaments"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimze"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("allocatated"),
                &["allocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("equeally"), &["equally"]),
            (
                dictgen::InsensitiveStr::Ascii("playfull"),
                &["playful", "playfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("weridly"), &["weirdly"]),
            (dictgen::InsensitiveStr::Ascii("faiiled"), &["failed"]),
            (dictgen::InsensitiveStr::Ascii("accoutned"), &["accounted"]),
            (dictgen::InsensitiveStr::Ascii("meassured"), &["measured"]),
            (
                dictgen::InsensitiveStr::Ascii("asbolutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchornously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparentcy"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("micrsft"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("desctiptor"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negitiation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valiations"),
                &["validations", "valuations"],
            ),
            (dictgen::InsensitiveStr::Ascii("snapsnot"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("minimizin"), &["minimizing"]),
            (
                dictgen::InsensitiveStr::Ascii("regerenerate"),
                &["regenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("susbsytem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("inititives"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informaional"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gymanstics"),
                &["gymnastics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aletred"),
                &["alerted", "altered"],
            ),
            (dictgen::InsensitiveStr::Ascii("vizually"), &["visually"]),
            (
                dictgen::InsensitiveStr::Ascii("recution"),
                &["reduction", "recursion", "precaution"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnkngs"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("wreckign"), &["wrecking"]),
            (
                dictgen::InsensitiveStr::Ascii("concatincate"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("creatning"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("confidental"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("yoou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("spefifies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("cedential"), &["credential"]),
            (dictgen::InsensitiveStr::Ascii("wnated"), &["wanted"]),
            (dictgen::InsensitiveStr::Ascii("galatic"), &["galactic"]),
            (dictgen::InsensitiveStr::Ascii("saveed"), &["saved"]),
            (
                dictgen::InsensitiveStr::Ascii("envelopped"),
                &["envelope", "enveloped"],
            ),
            (dictgen::InsensitiveStr::Ascii("chasin"), &["chasing"]),
            (dictgen::InsensitiveStr::Ascii("outselves"), &["ourselves"]),
            (
                dictgen::InsensitiveStr::Ascii("infintiely"),
                &["infinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuillity"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("longiudes"), &["longitudes"]),
            (dictgen::InsensitiveStr::Ascii("expediate"), &["expedite"]),
            (dictgen::InsensitiveStr::Ascii("usefl"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("initializaer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("interefere"), &["interfere"]),
            (dictgen::InsensitiveStr::Ascii("messagues"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("differece"), &["difference"]),
            (dictgen::InsensitiveStr::Ascii("exapnded"), &["expanded"]),
            (dictgen::InsensitiveStr::Ascii("vaccinae"), &["vaccinated"]),
            (dictgen::InsensitiveStr::Ascii("infex"), &["index"]),
            (dictgen::InsensitiveStr::Ascii("boook"), &["book"]),
            (dictgen::InsensitiveStr::Ascii("catpturing"), &["capturing"]),
            (dictgen::InsensitiveStr::Ascii("sigalled"), &["signalled"]),
            (dictgen::InsensitiveStr::Ascii("sofware"), &["software"]),
            (
                dictgen::InsensitiveStr::Ascii("eventially"),
                &["eventually"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefifer"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("phisics"), &["physics"]),
            (
                dictgen::InsensitiveStr::Ascii("libitarianisn"),
                &["libertarianism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacified"),
                &["specified", "pacified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skipp"),
                &["skip", "skipped"],
            ),
            (dictgen::InsensitiveStr::Ascii("marix"), &["matrix"]),
            (
                dictgen::InsensitiveStr::Ascii("classificaiton"),
                &["classification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplfiers"),
                &["simplifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("crahsing"), &["crashing"]),
            (dictgen::InsensitiveStr::Ascii("whilrwind"), &["whirlwind"]),
            (dictgen::InsensitiveStr::Ascii("structers"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("invesitgative"),
                &["investigative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatibily"),
                &["incompatibility", "incompatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autochtonous"),
                &["autochthonous"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulgariska"), &["bulgaria"]),
            (
                dictgen::InsensitiveStr::Ascii("dimentional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwealmed"),
                &["overwhelmed"],
            ),
            (dictgen::InsensitiveStr::Ascii("empahty"), &["empathy"]),
            (dictgen::InsensitiveStr::Ascii("coalleced"), &["coalesced"]),
            (
                dictgen::InsensitiveStr::Ascii("translucient"),
                &["translucent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facinations"),
                &["fascinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocassions"), &["occasions"]),
            (dictgen::InsensitiveStr::Ascii("mosquitero"), &["mosquito"]),
            (
                dictgen::InsensitiveStr::Ascii("clens"),
                &["clans", "cleans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininstration"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("purposelly"), &["purposely"]),
            (dictgen::InsensitiveStr::Ascii("noopner"), &["noopener"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccruate"),
                &["inaccurate"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisones"), &["poisons"]),
            (
                dictgen::InsensitiveStr::Ascii("critisicms"),
                &["criticisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alterate"),
                &["alternate", "alter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribued"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("mmatching"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("synchroizing"),
                &["synchronizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("disicples"), &["disciples"]),
            (dictgen::InsensitiveStr::Ascii("analyzier"), &["analyzer"]),
            (
                dictgen::InsensitiveStr::Ascii("suggesiton"),
                &["suggestion"],
            ),
            (dictgen::InsensitiveStr::Ascii("ariflow"), &["airflow"]),
            (
                dictgen::InsensitiveStr::Ascii("reigon"),
                &["region", "reign"],
            ),
            (dictgen::InsensitiveStr::Ascii("maintane"), &["maintain"]),
            (
                dictgen::InsensitiveStr::Ascii("brigthening"),
                &["brightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("insistas"), &["insists"]),
            (
                dictgen::InsensitiveStr::Ascii("placeholds"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("volocity"), &["velocity"]),
            (
                dictgen::InsensitiveStr::Ascii("alighns"),
                &["aligns", "alights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configarable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("lightbulp"), &["lightbulb"]),
            (dictgen::InsensitiveStr::Ascii("chechker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("discreprencies"),
                &["discrepancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("mystql"), &["mysql"]),
            (
                dictgen::InsensitiveStr::Ascii("catergorization"),
                &["categorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheive"), &["achieve"]),
            (dictgen::InsensitiveStr::Ascii("excepions"), &["exceptions"]),
            (dictgen::InsensitiveStr::Ascii("forgiener"), &["foreigner"]),
            (dictgen::InsensitiveStr::Ascii("imroving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("listenned"), &["listened"]),
            (
                dictgen::InsensitiveStr::Ascii("interrogacion"),
                &["interrogation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumalted"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("socialim"), &["socialism"]),
            (
                dictgen::InsensitiveStr::Ascii("caulfilower"),
                &["cauliflower"],
            ),
            (dictgen::InsensitiveStr::Ascii("aceptably"), &["acceptably"]),
            (
                dictgen::InsensitiveStr::Ascii("interlaing"),
                &["interlacing"],
            ),
            (dictgen::InsensitiveStr::Ascii("metdata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("fundirse"), &["fundies"]),
            (
                dictgen::InsensitiveStr::Ascii("shtopping"),
                &["stopping", "shopping"],
            ),
            (dictgen::InsensitiveStr::Ascii("deathcat"), &["deathmatch"]),
            (dictgen::InsensitiveStr::Ascii("repreesnt"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("iteractes"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("assitss"), &["assists"]),
            (dictgen::InsensitiveStr::Ascii("yelding"), &["yielding"]),
            (dictgen::InsensitiveStr::Ascii("euphorica"), &["euphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplies"),
                &["multiplies"],
            ),
            (dictgen::InsensitiveStr::Ascii("recod"), &["record"]),
            (dictgen::InsensitiveStr::Ascii("preverses"), &["preserves"]),
            (
                dictgen::InsensitiveStr::Ascii("curresponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("totat"), &["total"]),
            (
                dictgen::InsensitiveStr::Ascii("redistirbuting"),
                &["redistributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preservativo"),
                &["preservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapoliated"),
                &["extrapolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crendentials"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pitchforkers"),
                &["pitchforks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generelize"),
                &["generalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("destoys"), &["destroys"]),
            (
                dictgen::InsensitiveStr::Ascii("onveniently"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formtatting"),
                &["formatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparason"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("contious"), &["continuous"]),
            (dictgen::InsensitiveStr::Ascii("hyjackers"), &["hijackers"]),
            (
                dictgen::InsensitiveStr::Ascii("approprations"),
                &["appropriations"],
            ),
            (dictgen::InsensitiveStr::Ascii("baasic"), &["basic"]),
            (
                dictgen::InsensitiveStr::Ascii("anihilating"),
                &["annihilating"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpound"), &["compound"]),
            (dictgen::InsensitiveStr::Ascii("levetated"), &["levitated"]),
            (
                dictgen::InsensitiveStr::Ascii("scandianvian"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermaster"),
                &["supermarkets"],
            ),
            (dictgen::InsensitiveStr::Ascii("veresions"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("dockson"), &["dachshund"]),
            (
                dictgen::InsensitiveStr::Ascii("privilages"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("glorfied"), &["glorified"]),
            (
                dictgen::InsensitiveStr::Ascii("permissable"),
                &["permissible"],
            ),
            (dictgen::InsensitiveStr::Ascii("batchleur"), &["bachelor"]),
            (dictgen::InsensitiveStr::Ascii("floatint"), &["floating"]),
            (
                dictgen::InsensitiveStr::Ascii("statemanets"),
                &["statements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fullfiling"),
                &["fulfilling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oeprations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bastardes"), &["bastards"]),
            (
                dictgen::InsensitiveStr::Ascii("camoflagued"),
                &["camouflaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("conole"), &["console"]),
            (
                dictgen::InsensitiveStr::Ascii("desrciptions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("garants"), &["guarantors"]),
            (dictgen::InsensitiveStr::Ascii("downovted"), &["downvoted"]),
            (dictgen::InsensitiveStr::Ascii("zoocheinei"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("disscting"), &["dissecting"]),
            (
                dictgen::InsensitiveStr::Ascii("finded"),
                &["fined", "found", "funded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resursive"),
                &["recursive", "resourceful"],
            ),
            (dictgen::InsensitiveStr::Ascii("berskerer"), &["berserker"]),
            (
                dictgen::InsensitiveStr::Ascii("appeciates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traveres"),
                &["traverse", "traverses"],
            ),
            (dictgen::InsensitiveStr::Ascii("ivoke"), &["invoke"]),
            (
                dictgen::InsensitiveStr::Ascii("democarcies"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silient"),
                &["silent", "salient"],
            ),
            (dictgen::InsensitiveStr::Ascii("awmings"), &["awnings"]),
            (dictgen::InsensitiveStr::Ascii("obssesion"), &["obsession"]),
            (dictgen::InsensitiveStr::Ascii("initiials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("usuall"),
                &["usually", "usual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distiguish"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grativational"),
                &["gravitational"],
            ),
            (dictgen::InsensitiveStr::Ascii("receiever"), &["receiver"]),
            (dictgen::InsensitiveStr::Ascii("discovr"), &["discover"]),
            (
                dictgen::InsensitiveStr::Ascii("invonvenient"),
                &["inconvenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("mastter"), &["master"]),
            (
                dictgen::InsensitiveStr::Ascii("sensititive"),
                &["sensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heder"),
                &["header", "herder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redendancy"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrtibs"), &["contribs"]),
            (
                dictgen::InsensitiveStr::Ascii("transperently"),
                &["transparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmnetation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("solitare"),
                &["solitaire", "solitary"],
            ),
            (dictgen::InsensitiveStr::Ascii("quesetion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("invlovement"),
                &["involvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deverlopments"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisation"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgrating"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("elsiof"), &["elseif"]),
            (
                dictgen::InsensitiveStr::Ascii("disignated"),
                &["designated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootoloader"),
                &["bootloader"],
            ),
            (dictgen::InsensitiveStr::Ascii("displying"), &["displaying"]),
            (dictgen::InsensitiveStr::Ascii("separtates"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("absuolatist"),
                &["absolutist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrequantcies"),
                &["infrequencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paragpraphs"),
                &["paragraphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocrochips"),
                &["microchips"],
            ),
            (dictgen::InsensitiveStr::Ascii("tweeking"), &["tweaking"]),
            (
                dictgen::InsensitiveStr::Ascii("lengthly"),
                &["lengthy", "lengthwise", "lengthily"],
            ),
            (dictgen::InsensitiveStr::Ascii("bivwack"), &["bivouac"]),
            (dictgen::InsensitiveStr::Ascii("avaoid"), &["avoid"]),
            (
                dictgen::InsensitiveStr::Ascii("misgoynistic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimansion"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("familiare"),
                &["familiarize"],
            ),
            (dictgen::InsensitiveStr::Ascii("mnemnonic"), &["mnemonic"]),
            (dictgen::InsensitiveStr::Ascii("huricain"), &["hurricane"]),
            (
                dictgen::InsensitiveStr::Ascii("expensable"),
                &["expendable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coniderate"),
                &["considerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sunscreeen"), &["sunscreen"]),
            (
                dictgen::InsensitiveStr::Ascii("trianglular"),
                &["triangular"],
            ),
            (dictgen::InsensitiveStr::Ascii("aliagn"), &["align"]),
            (dictgen::InsensitiveStr::Ascii("joinning"), &["joining"]),
            (dictgen::InsensitiveStr::Ascii("excesive"), &["excessive"]),
            (dictgen::InsensitiveStr::Ascii("servised"), &["serviced"]),
            (
                dictgen::InsensitiveStr::Ascii("autimatically"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specualtion"),
                &["speculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rewatchign"),
                &["rewatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnaslation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tessealte"), &["tessellate"]),
            (dictgen::InsensitiveStr::Ascii("esacpe"), &["escape"]),
            (
                dictgen::InsensitiveStr::Ascii("documentry"),
                &["documentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prejudgudices"),
                &["prejudices"],
            ),
            (dictgen::InsensitiveStr::Ascii("officialy"), &["officially"]),
            (dictgen::InsensitiveStr::Ascii("nofifying"), &["notifying"]),
            (
                dictgen::InsensitiveStr::Ascii("regconized"),
                &["recognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("doublde"), &["double"]),
            (dictgen::InsensitiveStr::Ascii("aplies"), &["applies"]),
            (
                dictgen::InsensitiveStr::Ascii("commected"),
                &["connected", "commented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insentitive"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approiately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("asthetic"), &["aesthetic"]),
            (dictgen::InsensitiveStr::Ascii("throium"), &["thorium"]),
            (dictgen::InsensitiveStr::Ascii("spefixifes"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("seventeeen"), &["seventeen"]),
            (
                dictgen::InsensitiveStr::Ascii("athletecism"),
                &["athleticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("tecnician"), &["technician"]),
            (dictgen::InsensitiveStr::Ascii("valykrie"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("corresondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alloctions"),
                &["allocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("incresing"), &["increasing"]),
            (dictgen::InsensitiveStr::Ascii("padam"), &["param"]),
            (dictgen::InsensitiveStr::Ascii("vaildates"), &["validates"]),
            (
                dictgen::InsensitiveStr::Ascii("strikethough"),
                &["strikethrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desactivation"),
                &["deactivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofig"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("substancially"),
                &["substantially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregistes"),
                &["unregisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contracictions"),
                &["contradictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("insers"), &["inserts"]),
            (dictgen::InsensitiveStr::Ascii("fof"), &["for"]),
            (dictgen::InsensitiveStr::Ascii("slayign"), &["slaying"]),
            (dictgen::InsensitiveStr::Ascii("litle"), &["little"]),
            (
                dictgen::InsensitiveStr::Ascii("interveiws"),
                &["interviews"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiaze"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("converving"),
                &["conserving", "conversing", "converting", "converging"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnety"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("regardin"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("reestructuring"),
                &["restructuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonimously"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("encolsure"), &["enclosure"]),
            (dictgen::InsensitiveStr::Ascii("caridge"), &["carriage"]),
            (
                dictgen::InsensitiveStr::Ascii("stipping"),
                &["stippling", "stopping"],
            ),
            (dictgen::InsensitiveStr::Ascii("mantanine"), &["maintain"]),
            (
                dictgen::InsensitiveStr::Ascii("withhelding"),
                &["withholding"],
            ),
            (dictgen::InsensitiveStr::Ascii("amunition"), &["ammunition"]),
            (dictgen::InsensitiveStr::Ascii("excaping"), &["escaping"]),
            (dictgen::InsensitiveStr::Ascii("vertival"), &["vertical"]),
            (dictgen::InsensitiveStr::Ascii("orgainzer"), &["organizer"]),
            (
                dictgen::InsensitiveStr::Ascii("hyptotheses"),
                &["hypotheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("quckly"), &["quickly"]),
            (dictgen::InsensitiveStr::Ascii("ernoeous"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("canonoical"), &["canonical"]),
            (
                dictgen::InsensitiveStr::Ascii("typcasting"),
                &["typecasting"],
            ),
            (dictgen::InsensitiveStr::Ascii("mebrain"), &["membrane"]),
            (
                dictgen::InsensitiveStr::Ascii("notificiation"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corproration"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shemed"),
                &["schemed", "shamed"],
            ),
            (dictgen::InsensitiveStr::Ascii("genralize"), &["generalize"]),
            (
                dictgen::InsensitiveStr::Ascii("temprorarly"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terrotories"),
                &["territories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centipeddle"),
                &["centipede"],
            ),
            (dictgen::InsensitiveStr::Ascii("ligthing"), &["lighting"]),
            (dictgen::InsensitiveStr::Ascii("opitmism"), &["optimism"]),
            (dictgen::InsensitiveStr::Ascii("valuie"), &["value"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanthing"),
                &["enchanting"],
            ),
            (dictgen::InsensitiveStr::Ascii("versinos"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("clipoards"), &["clipboards"]),
            (
                dictgen::InsensitiveStr::Ascii("physcedelics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepetor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("plausbile"), &["plausible"]),
            (
                dictgen::InsensitiveStr::Ascii("accommadate"),
                &["accommodate"],
            ),
            (dictgen::InsensitiveStr::Ascii("chosing"), &["choosing"]),
            (dictgen::InsensitiveStr::Ascii("sensisble"), &["sensible"]),
            (
                dictgen::InsensitiveStr::Ascii("switzerand"),
                &["switzerland"],
            ),
            (dictgen::InsensitiveStr::Ascii("corespond"), &["correspond"]),
            (dictgen::InsensitiveStr::Ascii("convirted"), &["converted"]),
            (
                dictgen::InsensitiveStr::Ascii("trianglulates"),
                &["triangulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("custcenes"), &["cutscenes"]),
            (
                dictgen::InsensitiveStr::Ascii("appriproately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workstaions"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("devided"), &["divided"]),
            (
                dictgen::InsensitiveStr::Ascii("refelction"),
                &["reflection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sustainabillity"),
                &["sustainability"],
            ),
            (dictgen::InsensitiveStr::Ascii("pojrector"), &["projector"]),
            (dictgen::InsensitiveStr::Ascii("meathod"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("cahrging"), &["charging"]),
            (
                dictgen::InsensitiveStr::Ascii("abstructs"),
                &["abstracts", "obstructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("becaise"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillifed"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workrounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sovereinty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("vehilce"), &["vehicle"]),
            (
                dictgen::InsensitiveStr::Ascii("assisst"),
                &["assists", "assist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termining"),
                &["terminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convers"),
                &[
                    "converts", "converse", "convert", "covers", "conveys", "confers",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("favourie"), &["favourites"]),
            (
                dictgen::InsensitiveStr::Ascii("professsion"),
                &["professions"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrcypt"), &["encrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("communciations"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantiation"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captialising"),
                &["capitalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bloodporne"),
                &["bloodborne"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("challener"), &["challenger"]),
            (dictgen::InsensitiveStr::Ascii("startig"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("transmitirte"),
                &["transmitter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficillation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decelarations"),
                &["decelerations", "declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("peridic"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("unknwoingly"),
                &["unknowingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subconcsious"),
                &["subconscious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessufll"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comptibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("obejects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("comfrotable"),
                &["comfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("radicas"), &["radicals"]),
            (dictgen::InsensitiveStr::Ascii("acqure"), &["acquire"]),
            (
                dictgen::InsensitiveStr::Ascii("histarical"),
                &["historical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perspectief"),
                &["perspective"],
            ),
            (dictgen::InsensitiveStr::Ascii("spesifies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("steram"), &["stream"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ortolaryngologists"),
                &["otolaryngologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulid"), &["build"]),
            (dictgen::InsensitiveStr::Ascii("guarntees"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("appliacation"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzealusly"),
                &["overzealously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succussive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("dipsose"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("chekclist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("unexepecedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ressurrected"),
                &["resurrected"],
            ),
            (dictgen::InsensitiveStr::Ascii("phycisist"), &["physicist"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenionsalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rescrition"),
                &["restriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("enitities"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("conervative"),
                &["conservative"],
            ),
            (dictgen::InsensitiveStr::Ascii("contails"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("speficififed"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inproperly"),
                &["improperly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improsined"),
                &["imprisoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryfied"), &["verified"]),
            (dictgen::InsensitiveStr::Ascii("ukrainin"), &["ukrainian"]),
            (dictgen::InsensitiveStr::Ascii("outfittet"), &["outfitted"]),
            (
                dictgen::InsensitiveStr::Ascii("indefintly"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinaterics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anniversiary"),
                &["anniversary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fotgetting"),
                &["forgetting"],
            ),
            (dictgen::InsensitiveStr::Ascii("touristsy"), &["touristy"]),
            (dictgen::InsensitiveStr::Ascii("preorded"), &["preordered"]),
            (dictgen::InsensitiveStr::Ascii("tuncated"), &["truncated"]),
            (
                dictgen::InsensitiveStr::Ascii("simultaneos"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capasitors"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidentail"),
                &["presidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("secceed"), &["succeed"]),
            (dictgen::InsensitiveStr::Ascii("impres"), &["impress"]),
            (dictgen::InsensitiveStr::Ascii("skiping"), &["skipping"]),
            (dictgen::InsensitiveStr::Ascii("highlihght"), &["highlight"]),
            (dictgen::InsensitiveStr::Ascii("soloutions"), &["solutions"]),
            (
                dictgen::InsensitiveStr::Ascii("neglibable"),
                &["negligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenicating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mayalsia"), &["malaysia"]),
            (dictgen::InsensitiveStr::Ascii("geenrally"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("unspportable"),
                &["unsupportable"],
            ),
            (dictgen::InsensitiveStr::Ascii("trhreading"), &["threading"]),
            (dictgen::InsensitiveStr::Ascii("upliad"), &["upload"]),
            (
                dictgen::InsensitiveStr::Ascii("selecters"),
                &["selectors", "selects"],
            ),
            (dictgen::InsensitiveStr::Ascii("parmisian"), &["parmesan"]),
            (
                dictgen::InsensitiveStr::Ascii("predicition"),
                &["prediction", "predication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declearators"),
                &["declarators", "decelerators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperiece"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("presumendo"), &["presumed"]),
            (
                dictgen::InsensitiveStr::Ascii("effortlessley"),
                &["effortlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtats"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("reconsturctor"),
                &["reconstructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaysis"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("novmeber"), &["november"]),
            (
                dictgen::InsensitiveStr::Ascii("preinitalizes"),
                &["preinitializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovserve"), &["observe"]),
            (
                dictgen::InsensitiveStr::Ascii("disconecction"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholership"),
                &["scholarship"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumu"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("coporate"), &["corporate"]),
            (dictgen::InsensitiveStr::Ascii("scorpoin"), &["scorpion"]),
            (
                dictgen::InsensitiveStr::Ascii("comfigured"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manigestos"),
                &["manifestos"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventageous"),
                &["advantageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continunes"),
                &["continents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representationer"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("follwoers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("authrorise"), &["authorise"]),
            (dictgen::InsensitiveStr::Ascii("technicial"), &["technical"]),
            (
                dictgen::InsensitiveStr::Ascii("creationsist"),
                &["creationist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sportsmenship"),
                &["sportsmanship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("observacion"),
                &["observation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processs"),
                &["processes", "process"],
            ),
            (dictgen::InsensitiveStr::Ascii("spaw"), &["spawn"]),
            (dictgen::InsensitiveStr::Ascii("origninate"), &["originate"]),
            (dictgen::InsensitiveStr::Ascii("yeilded"), &["yielded"]),
            (
                dictgen::InsensitiveStr::Ascii("sructurally"),
                &["structurally"],
            ),
            (dictgen::InsensitiveStr::Ascii("contatiner"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("storin"), &["storing"]),
            (dictgen::InsensitiveStr::Ascii("knwoingly"), &["knowingly"]),
            (
                dictgen::InsensitiveStr::Ascii("separetedly"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("secifiers"), &["specifiers"]),
            (dictgen::InsensitiveStr::Ascii("occaisons"), &["occasions"]),
            (
                dictgen::InsensitiveStr::Ascii("educatin"),
                &["educating", "education"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negiotated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parititiones"),
                &["partitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("aond"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("instnance"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("unemplyoment"),
                &["unemployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("seleccted"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("abbreivations"),
                &["abbreviations"],
            ),
            (dictgen::InsensitiveStr::Ascii("saerch"), &["search"]),
            (dictgen::InsensitiveStr::Ascii("spefiiers"), &["specifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("habbit"),
                &["habit", "rabbit", "hobbit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contracitions"),
                &["contradictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexperesions"),
                &["subexpressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asumptions"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rachets"),
                &["ratchets", "sachets", "cachets", "rackets"],
            ),
            (dictgen::InsensitiveStr::Ascii("cteation"), &["creation"]),
            (dictgen::InsensitiveStr::Ascii("carmels"), &["caramels"]),
            (dictgen::InsensitiveStr::Ascii("annount"), &["announce"]),
            (
                dictgen::InsensitiveStr::Ascii("parallisation"),
                &["parallelisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("accute"), &["acute"]),
            (dictgen::InsensitiveStr::Ascii("manifst"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("purpouses"), &["purposes"]),
            (
                dictgen::InsensitiveStr::Ascii("animaitons"),
                &["animations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aftterward"), &["afterward"]),
            (dictgen::InsensitiveStr::Ascii("croozer"), &["cruiser"]),
            (dictgen::InsensitiveStr::Ascii("correnti"), &["correcting"]),
            (dictgen::InsensitiveStr::Ascii("encocdings"), &["encodings"]),
            (
                dictgen::InsensitiveStr::Ascii("expestedly"),
                &["expectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("overiew"), &["overview"]),
            (
                dictgen::InsensitiveStr::Ascii("eloquantly"),
                &["eloquently"],
            ),
            (dictgen::InsensitiveStr::Ascii("restesting"), &["retesting"]),
            (dictgen::InsensitiveStr::Ascii("muesums"), &["museums"]),
            (dictgen::InsensitiveStr::Ascii("ugraders"), &["upgraders"]),
            (
                dictgen::InsensitiveStr::Ascii("placegolders"),
                &["placeholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarquism"), &["anarchism"]),
            (
                dictgen::InsensitiveStr::Ascii("delevopment"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("parana"), &["piranha"]),
            (dictgen::InsensitiveStr::Ascii("fomrated"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("surpporting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("exclue"), &["exclude"]),
            (
                dictgen::InsensitiveStr::Ascii("sunglinger"),
                &["gunslinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copiled"),
                &["compiled", "copied"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilynders"), &["cylinders"]),
            (
                dictgen::InsensitiveStr::Ascii("composablity"),
                &["composability"],
            ),
            (dictgen::InsensitiveStr::Ascii("repaird"), &["repaired"]),
            (dictgen::InsensitiveStr::Ascii("aggaravate"), &["aggravate"]),
            (dictgen::InsensitiveStr::Ascii("ignorte"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("intagrates"),
                &["integrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("outfied"), &["outfield"]),
            (dictgen::InsensitiveStr::Ascii("ariable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("constands"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("annullin"), &["annulling"]),
            (dictgen::InsensitiveStr::Ascii("upporterd"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("iterrations"),
                &["iterations"],
            ),
            (dictgen::InsensitiveStr::Ascii("commiting"), &["committing"]),
            (dictgen::InsensitiveStr::Ascii("requesst"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("efectively"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persuassion"),
                &["persuasion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartors"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrinutions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("passabe"), &["passable"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstrcuted"),
                &["reconstructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sruct"), &["struct"]),
            (
                dictgen::InsensitiveStr::Ascii("explecitily"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrpited"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("contrbutes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmeneter"),
                &["implementer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restriktion"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracterizations"),
                &["characterizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychiatirst"),
                &["psychiatrists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoing"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("heps"), &["helps", "heaps"]),
            (dictgen::InsensitiveStr::Ascii("activit"), &["activist"]),
            (dictgen::InsensitiveStr::Ascii("recievers"), &["receivers"]),
            (
                dictgen::InsensitiveStr::Ascii("vulerabuility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermientelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surprizingly"),
                &["surprisingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsting"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("folliwong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("repulicans"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("tme"), &["time"]),
            (dictgen::InsensitiveStr::Ascii("rasied"), &["raised"]),
            (
                dictgen::InsensitiveStr::Ascii("comapatibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("conumed"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("exceotion"),
                &["exemption", "exception"],
            ),
            (dictgen::InsensitiveStr::Ascii("slicable"), &["sliceable"]),
            (
                dictgen::InsensitiveStr::Ascii("manufaturing"),
                &["manufacturing"],
            ),
            (dictgen::InsensitiveStr::Ascii("objetc"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("scrach"), &["scratch"]),
            (dictgen::InsensitiveStr::Ascii("decypted"), &["decrypted"]),
            (dictgen::InsensitiveStr::Ascii("vitrually"), &["virtually"]),
            (dictgen::InsensitiveStr::Ascii("thurdsay"), &["thursday"]),
            (dictgen::InsensitiveStr::Ascii("losened"), &["loosened"]),
            (
                dictgen::InsensitiveStr::Ascii("interporation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("extracotr"), &["extractor"]),
            (
                dictgen::InsensitiveStr::Ascii("workstaion"),
                &["workstation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communiste"),
                &["communities"],
            ),
            (dictgen::InsensitiveStr::Ascii("relised"), &["realised"]),
            (
                dictgen::InsensitiveStr::Ascii("structeral"),
                &["structural"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizzle"), &["chisel"]),
            (dictgen::InsensitiveStr::Ascii("organizaed"), &["organize"]),
            (dictgen::InsensitiveStr::Ascii("primarly"), &["primarily"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotications"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("disappers"), &["disappears"]),
            (dictgen::InsensitiveStr::Ascii("bejond"), &["beyond"]),
            (
                dictgen::InsensitiveStr::Ascii("incremmenting"),
                &["incrementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otates"),
                &["notates", "rotates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fileshystem"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohobited"),
                &["prohibited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confugring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("dfine"), &["define"]),
            (
                dictgen::InsensitiveStr::Ascii("definitevly"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("mamal"), &["mammal"]),
            (dictgen::InsensitiveStr::Ascii("huristics"), &["heuristics"]),
            (dictgen::InsensitiveStr::Ascii("grranted"), &["granted"]),
            (
                dictgen::InsensitiveStr::Ascii("degenaration"),
                &["degeneration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representacion"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramaterizations"),
                &["parameterizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possesives"),
                &["possessive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preservering"),
                &["preserving", "persevering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hamburgurs"),
                &["hamburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallize"),
                &["parallelize"],
            ),
            (dictgen::InsensitiveStr::Ascii("theistc"), &["theistic"]),
            (dictgen::InsensitiveStr::Ascii("sofomore"), &["sophomore"]),
            (
                dictgen::InsensitiveStr::Ascii("transmiting"),
                &["transmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("follwowiwng"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occrrences"),
                &["occurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mediocrisy"),
                &["mediocrity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotaitions"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("screeb"), &["screen"]),
            (
                dictgen::InsensitiveStr::Ascii("sheming"),
                &["scheming", "shaming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatabilty"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilizatons"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intereested"),
                &["interested"],
            ),
            (dictgen::InsensitiveStr::Ascii("predujice"), &["prejudice"]),
            (dictgen::InsensitiveStr::Ascii("scenarious"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("feburary"), &["february"]),
            (dictgen::InsensitiveStr::Ascii("wepon"), &["weapon"]),
            (
                dictgen::InsensitiveStr::Ascii("declaraions"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutneract"),
                &["counteract"],
            ),
            (dictgen::InsensitiveStr::Ascii("sned"), &["send"]),
            (
                dictgen::InsensitiveStr::Ascii("comeptitive"),
                &["competitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurise"),
                &["pasteurise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutipliers"),
                &["multipliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canacelling"),
                &["cancelling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchrounous"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstanging"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("historgam"), &["histogram"]),
            (dictgen::InsensitiveStr::Ascii("disort"), &["distort"]),
            (dictgen::InsensitiveStr::Ascii("greusome"), &["gruesome"]),
            (
                dictgen::InsensitiveStr::Ascii("manupulated"),
                &["manipulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleandes"), &["cleanses"]),
            (dictgen::InsensitiveStr::Ascii("querys"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailabel"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("socratease"), &["socrates"]),
            (
                dictgen::InsensitiveStr::Ascii("genralized"),
                &["generalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("impules"), &["impulse"]),
            (dictgen::InsensitiveStr::Ascii("propety"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("wroked"), &["worked"]),
            (
                dictgen::InsensitiveStr::Ascii("deconstructers"),
                &["deconstructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("expolsion"), &["explosion"]),
            (dictgen::InsensitiveStr::Ascii("dissmis"), &["dismiss"]),
            (dictgen::InsensitiveStr::Ascii("noticibly"), &["noticeably"]),
            (dictgen::InsensitiveStr::Ascii("sometinhg"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("preocupation"),
                &["preoccupation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formmatters"),
                &["formatters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstartes"),
                &["demonstrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("sarters"), &["starters"]),
            (dictgen::InsensitiveStr::Ascii("paperworks"), &["paperwork"]),
            (dictgen::InsensitiveStr::Ascii("disgustosi"), &["disgusts"]),
            (
                dictgen::InsensitiveStr::Ascii("encyclapedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("detectsion"), &["detection"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegoziation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("corected"), &["corrected"]),
            (dictgen::InsensitiveStr::Ascii("exitance"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("lackrimose"),
                &["lachrymose"],
            ),
            (dictgen::InsensitiveStr::Ascii("epheremal"), &["ephemeral"]),
            (
                dictgen::InsensitiveStr::Ascii("modyfication"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstlalation"),
                &["uninstallation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improtations"),
                &["importations"],
            ),
            (dictgen::InsensitiveStr::Ascii("manetains"), &["maintains"]),
            (dictgen::InsensitiveStr::Ascii("judegments"), &["judgments"]),
            (dictgen::InsensitiveStr::Ascii("eihths"), &["eighths"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibiity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remainining"),
                &["remaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("renference"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("regnerative"),
                &["regenerative"],
            ),
            (dictgen::InsensitiveStr::Ascii("stipled"), &["stippled"]),
            (
                dictgen::InsensitiveStr::Ascii("manipluate"),
                &["manipulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("effecient"), &["efficient"]),
            (
                dictgen::InsensitiveStr::Ascii("coinsident"),
                &["coincident"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palistinian"),
                &["palestinian"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticial"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("upack"), &["unpack"]),
            (dictgen::InsensitiveStr::Ascii("theading"), &["threading"]),
            (
                dictgen::InsensitiveStr::Ascii("questionsign"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deslected"),
                &["deselected", "deflected"],
            ),
            (dictgen::InsensitiveStr::Ascii("fastners"), &["fasteners"]),
            (dictgen::InsensitiveStr::Ascii("unprone"), &["unproven"]),
            (dictgen::InsensitiveStr::Ascii("nummeral"), &["numeral"]),
            (dictgen::InsensitiveStr::Ascii("unlkely"), &["unlikely"]),
            (dictgen::InsensitiveStr::Ascii("fragmenets"), &["fragments"]),
            (dictgen::InsensitiveStr::Ascii("precends"), &["precedence"]),
            (dictgen::InsensitiveStr::Ascii("resetable"), &["resettable"]),
            (dictgen::InsensitiveStr::Ascii("subjet"), &["subject"]),
            (dictgen::InsensitiveStr::Ascii("strutural"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("convincersi"),
                &["convinces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritsing"),
                &["prioritising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitailised"),
                &["reinitialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rade"),
                &["read", "raid", "rate"],
            ),
            (dictgen::InsensitiveStr::Ascii("clipbarod"), &["clipboard"]),
            (dictgen::InsensitiveStr::Ascii("januray"), &["january"]),
            (
                dictgen::InsensitiveStr::Ascii("neutrallity"),
                &["neutrality"],
            ),
            (dictgen::InsensitiveStr::Ascii("stucture"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("inbracer"), &["embracer"]),
            (
                dictgen::InsensitiveStr::Ascii("chanches"),
                &["chances", "changes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chrisitans"),
                &["christians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviorement"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("migated"),
                &["migrated", "mitigated"],
            ),
            (dictgen::InsensitiveStr::Ascii("shudown"), &["shutdown"]),
            (dictgen::InsensitiveStr::Ascii("seperetes"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("errupted"), &["erupted"]),
            (dictgen::InsensitiveStr::Ascii("woraround"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("pundent"), &["pundit"]),
            (
                dictgen::InsensitiveStr::Ascii("unselcting"),
                &["unselecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equavilent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troossos"),
                &["trousseaus", "trousseaux"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowleded"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("showvinism"),
                &["chauvinism"],
            ),
            (dictgen::InsensitiveStr::Ascii("roughtly"), &["roughly"]),
            (
                dictgen::InsensitiveStr::Ascii("metamorphysis"),
                &["metamorphosis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylcling"), &["cycling"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaeters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("accort"), &["accord"]),
            (
                dictgen::InsensitiveStr::Ascii("asynnchronously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranlator"), &["translator"]),
            (
                dictgen::InsensitiveStr::Ascii("intrumental"),
                &["instrumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("actaul"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("pyschopath"),
                &["psychopath"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neumonectomies"),
                &["pneumonectomies"],
            ),
            (dictgen::InsensitiveStr::Ascii("retruned"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("similiarties"),
                &["similarities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifying"),
                &["sacrificing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polinators"),
                &["pollinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("socrateks"), &["socrates"]),
            (dictgen::InsensitiveStr::Ascii("descktop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretaciones"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppresssion"),
                &["suppression"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirrir"), &["mirror"]),
            (dictgen::InsensitiveStr::Ascii("whcich"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("susceder"), &["seceder"]),
            (
                dictgen::InsensitiveStr::Ascii("prejudgudicing"),
                &["prejudicing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assistnace"),
                &["assistance"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbously"), &["verbosely"]),
            (
                dictgen::InsensitiveStr::Ascii("photograhpy"),
                &["photography"],
            ),
            (dictgen::InsensitiveStr::Ascii("selektor"), &["selector"]),
            (
                dictgen::InsensitiveStr::Ascii("predetermiend"),
                &["predetermined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downwoters"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("poralized"), &["polarized"]),
            (
                dictgen::InsensitiveStr::Ascii("requiesting"),
                &["requesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrictons"),
                &["restricts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectinos"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistencency"),
                &["consistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatione"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("acceptes"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("correesponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blueburries"),
                &["blueberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("depricate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("refects"), &["reflects"]),
            (
                dictgen::InsensitiveStr::Ascii("constillation"),
                &["constellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("psydonym"), &["pseudonym"]),
            (
                dictgen::InsensitiveStr::Ascii("sarimonial"),
                &["ceremonial"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmitter"), &["emitter"]),
            (
                dictgen::InsensitiveStr::Ascii("microtramsactions"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("flavoures"), &["flavours"]),
            (dictgen::InsensitiveStr::Ascii("aethist"), &["atheist"]),
            (dictgen::InsensitiveStr::Ascii("prusuit"), &["pursuit"]),
            (dictgen::InsensitiveStr::Ascii("gaurd"), &["guard", "gourd"]),
            (dictgen::InsensitiveStr::Ascii("summoenrs"), &["summoners"]),
            (dictgen::InsensitiveStr::Ascii("viserally"), &["viscerally"]),
            (dictgen::InsensitiveStr::Ascii("finicial"), &["financial"]),
            (dictgen::InsensitiveStr::Ascii("encrouch"), &["encroach"]),
            (
                dictgen::InsensitiveStr::Ascii("statisticaly"),
                &["statistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("aneeling"), &["annealing"]),
            (dictgen::InsensitiveStr::Ascii("timestramp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("librraries"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("vaction"), &["vacation"]),
            (
                dictgen::InsensitiveStr::Ascii("dictioanries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiner"), &["pointer"]),
            (dictgen::InsensitiveStr::Ascii("invarient"), &["invariant"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronizing"),
                &["synchronizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("immitators"), &["imitators"]),
            (dictgen::InsensitiveStr::Ascii("trasnfers"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("hindisght"), &["hindsight"]),
            (
                dictgen::InsensitiveStr::Ascii("intangable"),
                &["intangible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiening"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("warninsgs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("applicatoin"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("prohram"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("syntehsises"),
                &["synthesises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminater"),
                &[
                    "discriminator",
                    "discriminated",
                    "discriminates",
                    "discriminate",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aknowledged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininized"), &["minimized"]),
            (dictgen::InsensitiveStr::Ascii("maneuveur"), &["maneuver"]),
            (
                dictgen::InsensitiveStr::Ascii("expetimenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("savy"), &["savvy"]),
            (dictgen::InsensitiveStr::Ascii("througt"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("provdes"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("normlizing"),
                &["normalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("envleope"), &["envelope"]),
            (
                dictgen::InsensitiveStr::Ascii("cointainers"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illigitemate"),
                &["illegitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("predecent"), &["precedent"]),
            (
                dictgen::InsensitiveStr::Ascii("inticement"),
                &["enticement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("orhpans"), &["orphans"]),
            (dictgen::InsensitiveStr::Ascii("apppears"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("appoximated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("collaros"), &["collars"]),
            (dictgen::InsensitiveStr::Ascii("realted"), &["related"]),
            (dictgen::InsensitiveStr::Ascii("wrok"), &["work"]),
            (
                dictgen::InsensitiveStr::Ascii("movied"),
                &["moved", "movie"],
            ),
            (dictgen::InsensitiveStr::Ascii("udpating"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("initiatiated"),
                &["initiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("armenain"), &["armenian"]),
            (dictgen::InsensitiveStr::Ascii("hatchign"), &["hatching"]),
            (
                dictgen::InsensitiveStr::Ascii("desciptively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpetrar"),
                &["perpetrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferable"),
                &["preferable"],
            ),
            (dictgen::InsensitiveStr::Ascii("mitagated"), &["mitigated"]),
            (dictgen::InsensitiveStr::Ascii("noramalize"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("mesoneens"), &["mezzanines"]),
            (
                dictgen::InsensitiveStr::Ascii("dirctional"),
                &["directional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instroments"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessleates"),
                &["tessellates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empiracally"),
                &["empirically"],
            ),
            (dictgen::InsensitiveStr::Ascii("resverves"), &["reserves"]),
            (
                dictgen::InsensitiveStr::Ascii("spesifying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvinience"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noteworthly"),
                &["noteworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("emable"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("mutialted"), &["mutilated"]),
            (
                dictgen::InsensitiveStr::Ascii("unpleasent"),
                &["unpleasant"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequene"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("thughts"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("interections"),
                &["interactions", "intersections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gangsterous"),
                &["gangsters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ilustrated"),
                &["illustrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("geenrator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("sepetately"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnticated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subroutinese"),
                &["subroutines"],
            ),
            (dictgen::InsensitiveStr::Ascii("breaked"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("organiclly"),
                &["organically"],
            ),
            (dictgen::InsensitiveStr::Ascii("dribbels"), &["dribbles"]),
            (
                dictgen::InsensitiveStr::Ascii("autheticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("runn"), &["run"]),
            (dictgen::InsensitiveStr::Ascii("disccusses"), &["discusses"]),
            (dictgen::InsensitiveStr::Ascii("dissappear"), &["disappear"]),
            (
                dictgen::InsensitiveStr::Ascii("inconveinence"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("degrassie"), &["degrasse"]),
            (
                dictgen::InsensitiveStr::Ascii("supermarkert"),
                &["supermarkets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compuslive"),
                &["compulsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("gradualy"), &["gradually"]),
            (
                dictgen::InsensitiveStr::Ascii("cenvention"),
                &["convention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uppstreamer"),
                &["upstreamer"],
            ),
            (dictgen::InsensitiveStr::Ascii("coruption"), &["corruption"]),
            (dictgen::InsensitiveStr::Ascii("tillt"), &["tilt"]),
            (
                dictgen::InsensitiveStr::Ascii("matchamking"),
                &["matchmaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("normalyly"), &["normally"]),
            (
                dictgen::InsensitiveStr::Ascii("constructieve"),
                &["constructive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("medidating"),
                &["meditating"],
            ),
            (dictgen::InsensitiveStr::Ascii("kuberntes"), &["kubernetes"]),
            (
                dictgen::InsensitiveStr::Ascii("avaiblability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("preesnt"), &["present"]),
            (dictgen::InsensitiveStr::Ascii("spefiifeis"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("atendance"), &["attendance"]),
            (
                dictgen::InsensitiveStr::Ascii("canonal"),
                &["canonical", "cantonal", "canola"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspiciuos"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("algortithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("snythesise"),
                &["synthesise"],
            ),
            (dictgen::InsensitiveStr::Ascii("rentention"), &["retention"]),
            (dictgen::InsensitiveStr::Ascii("firt"), &["first", "flirt"]),
            (
                dictgen::InsensitiveStr::Ascii("atomticaly"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlsation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("priotize"), &["prioritize"]),
            (
                dictgen::InsensitiveStr::Ascii("particaulrly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjcence"), &["adjacence"]),
            (dictgen::InsensitiveStr::Ascii("cmpiler"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("abudct"), &["abduct"]),
            (dictgen::InsensitiveStr::Ascii("dolphinese"), &["dolphins"]),
            (
                dictgen::InsensitiveStr::Ascii("extractino"),
                &["extracting", "extraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumberable"),
                &["enumerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complacant"),
                &["complacent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enoucntered"),
                &["encountered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegothiable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficified"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randazyvoued"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recuritment"),
                &["recruitment"],
            ),
            (dictgen::InsensitiveStr::Ascii("troubador"), &["troubadour"]),
            (dictgen::InsensitiveStr::Ascii("trianlge"), &["triangle"]),
            (dictgen::InsensitiveStr::Ascii("saturdey"), &["saturday"]),
            (
                dictgen::InsensitiveStr::Ascii("ojbections"),
                &["objections"],
            ),
            (dictgen::InsensitiveStr::Ascii("intitiated"), &["initiated"]),
            (
                dictgen::InsensitiveStr::Ascii("requirering"),
                &["requiring"],
            ),
            (dictgen::InsensitiveStr::Ascii("corispond"), &["correspond"]),
            (dictgen::InsensitiveStr::Ascii("monnth"), &["month"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirrectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("absorbtion"),
                &["absorption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfior"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatabiity"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("travles"), &["travels"]),
            (
                dictgen::InsensitiveStr::Ascii("unknowinlgy"),
                &["unknowingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("accidens"), &["accidents"]),
            (
                dictgen::InsensitiveStr::Ascii("custonizer"),
                &["customizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibitng"),
                &["prohibiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahnnels"), &["channels"]),
            (dictgen::InsensitiveStr::Ascii("eleents"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("follwowing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("pleaee"), &["please"]),
            (dictgen::InsensitiveStr::Ascii("biginners"), &["beginners"]),
            (
                dictgen::InsensitiveStr::Ascii("structeres"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepulcre"), &["sepulchre"]),
            (
                dictgen::InsensitiveStr::Ascii("improbement"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrspondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conatainers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("depenency"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("ideologije"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("sexuallly"), &["sexually"]),
            (
                dictgen::InsensitiveStr::Ascii("defintiion"),
                &["definitions", "definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdded"), &["embedded"]),
            (
                dictgen::InsensitiveStr::Ascii("spontaneosly"),
                &["spontaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilaging"),
                &["privileging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconencting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("emtied"), &["emptied"]),
            (
                dictgen::InsensitiveStr::Ascii("dispalying"),
                &["displaying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confrontacional"),
                &["confrontational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ridiculouly"),
                &["ridiculously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accuartely"),
                &["accurately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misdemenors"),
                &["misdemeanors"],
            ),
            (dictgen::InsensitiveStr::Ascii("conlict"), &["conflict"]),
            (
                dictgen::InsensitiveStr::Ascii("preferenze"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtitutes"),
                &["substitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("absovles"), &["absolves"]),
            (
                dictgen::InsensitiveStr::Ascii("sistemizing"),
                &["systemizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierchies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressie"),
                &["progressives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convoluded"),
                &["convoluted"],
            ),
            (dictgen::InsensitiveStr::Ascii("coincidnce"), &["coincide"]),
            (
                dictgen::InsensitiveStr::Ascii("cetificate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastinationg"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("deferance"), &["deference"]),
            (dictgen::InsensitiveStr::Ascii("duplactes"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("efficieny"),
                &["efficiency", "efficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("analzyers"), &["analyzers"]),
            (dictgen::InsensitiveStr::Ascii("brimestone"), &["brimstone"]),
            (
                dictgen::InsensitiveStr::Ascii("neccessarly"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("escpae"), &["escape"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapoitned"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("feasabile"), &["feasible"]),
            (
                dictgen::InsensitiveStr::Ascii("authecate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demogrpahics"),
                &["demographics"],
            ),
            (dictgen::InsensitiveStr::Ascii("contestase"), &["contests"]),
            (
                dictgen::InsensitiveStr::Ascii("conneiction"),
                &["connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("worthelss"), &["worthless"]),
            (dictgen::InsensitiveStr::Ascii("undersacn"), &["underscan"]),
            (dictgen::InsensitiveStr::Ascii("phsics"), &["physics"]),
            (
                dictgen::InsensitiveStr::Ascii("purpendicular"),
                &["perpendicular"],
            ),
            (dictgen::InsensitiveStr::Ascii("espacally"), &["especially"]),
            (
                dictgen::InsensitiveStr::Ascii("uncrypted"),
                &["unencrypted"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherient"), &["inherent"]),
            (dictgen::InsensitiveStr::Ascii("parmisan"), &["parmesan"]),
            (dictgen::InsensitiveStr::Ascii("direcives"), &["directives"]),
            (
                dictgen::InsensitiveStr::Ascii("naviagtion"),
                &["navigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inheirted"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("activtions"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gentlemn"), &["gentlemen"]),
            (dictgen::InsensitiveStr::Ascii("runntime"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("regresison"),
                &["regression"],
            ),
            (dictgen::InsensitiveStr::Ascii("femminist"), &["feminist"]),
            (dictgen::InsensitiveStr::Ascii("languahes"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("disapplined"),
                &["disciplined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deveopment"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persicution"),
                &["persecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placeheld"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqared"), &["squared"]),
            (dictgen::InsensitiveStr::Ascii("filenaes"), &["filenames"]),
            (dictgen::InsensitiveStr::Ascii("intreegued"), &["intrigued"]),
            (dictgen::InsensitiveStr::Ascii("determned"), &["determined"]),
            (dictgen::InsensitiveStr::Ascii("reqiuring"), &["requiring"]),
            (
                dictgen::InsensitiveStr::Ascii("scatchs"),
                &["catches", "scratches", "sketches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eveyrwhere"),
                &["everywhere"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apreciating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("workin"), &["working"]),
            (
                dictgen::InsensitiveStr::Ascii("desitination"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgustos"), &["disgusts"]),
            (
                dictgen::InsensitiveStr::Ascii("unfilpping"),
                &["unflipping"],
            ),
            (dictgen::InsensitiveStr::Ascii("attaced"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulationg"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("hadnling"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("conenience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mangaement"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("triptishs"), &["triptychs"]),
            (dictgen::InsensitiveStr::Ascii("platoes"), &["plateaus"]),
            (dictgen::InsensitiveStr::Ascii("opne"), &["open"]),
            (
                dictgen::InsensitiveStr::Ascii("infinetely"),
                &["infinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equievalent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paychedelics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("explaines"), &["explains"]),
            (dictgen::InsensitiveStr::Ascii("eith"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("windwo"), &["window"]),
            (dictgen::InsensitiveStr::Ascii("theread"), &["thread"]),
            (
                dictgen::InsensitiveStr::Ascii("certifacted"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("relocaite"), &["relocate"]),
            (
                dictgen::InsensitiveStr::Ascii("unnesesary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissbled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("inprove"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("serailization"),
                &["serialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("bufefr"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("configuere"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("conspiracys"),
                &["conspiracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("coddlin"), &["coddling"]),
            (dictgen::InsensitiveStr::Ascii("vailidated"), &["validated"]),
            (dictgen::InsensitiveStr::Ascii("imform"), &["inform"]),
            (dictgen::InsensitiveStr::Ascii("inclreased"), &["increased"]),
            (dictgen::InsensitiveStr::Ascii("warnins"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("evebts"), &["events"]),
            (dictgen::InsensitiveStr::Ascii("trinagles"), &["triangles"]),
            (dictgen::InsensitiveStr::Ascii("probelm"), &["problem"]),
            (
                dictgen::InsensitiveStr::Ascii("correccion"),
                &["correction"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytemd"), &["systemd"]),
            (dictgen::InsensitiveStr::Ascii("labirinth"), &["labyrinth"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasparent"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("magolia"), &["magnolia"]),
            (dictgen::InsensitiveStr::Ascii("provdier"), &["provider"]),
            (dictgen::InsensitiveStr::Ascii("tablulate"), &["tabulate"]),
            (
                dictgen::InsensitiveStr::Ascii("addtionally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisments"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("receipient"), &["recipient"]),
            (dictgen::InsensitiveStr::Ascii("adin"), &["admin"]),
            (
                dictgen::InsensitiveStr::Ascii("comprasion"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficeid"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("disporve"), &["disprove"]),
            (
                dictgen::InsensitiveStr::Ascii("autheticated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("counpound"), &["compound"]),
            (
                dictgen::InsensitiveStr::Ascii("impliments"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("momements"), &["moments"]),
            (
                dictgen::InsensitiveStr::Ascii("responsbilty"),
                &["responsibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("alocated"), &["allocated"]),
            (
                dictgen::InsensitiveStr::Ascii("verfication"),
                &["verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriximates"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assumotion"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimphean"), &["nymphean"]),
            (dictgen::InsensitiveStr::Ascii("scuccesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("definate"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("fragmentization"),
                &["fragmentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiinter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("accedental"),
                &["accidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permanetly"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tailgatting"),
                &["tailgating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catterpilar"),
                &["caterpillar"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitmiser"), &["optimiser"]),
            (dictgen::InsensitiveStr::Ascii("folllowed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("languge"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("cencrete"), &["concrete"]),
            (dictgen::InsensitiveStr::Ascii("relaesed"), &["released"]),
            (
                dictgen::InsensitiveStr::Ascii("integartes"),
                &["integrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disucussions"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventages"),
                &["advantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrductions"),
                &["introductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("digist"), &["digits"]),
            (dictgen::InsensitiveStr::Ascii("laugnage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("proccedings"),
                &["proceedings"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabniets"), &["cabinets"]),
            (dictgen::InsensitiveStr::Ascii("politicin"), &["politician"]),
            (dictgen::InsensitiveStr::Ascii("endevors"), &["endeavors"]),
            (dictgen::InsensitiveStr::Ascii("nominatie"), &["nominate"]),
            (
                dictgen::InsensitiveStr::Ascii("appoximates"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metrix"),
                &["matrix", "metric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independedly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("caraboo"), &["caribou"]),
            (
                dictgen::InsensitiveStr::Ascii("derecting"),
                &["directing", "detecting", "defecting", "erecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assomption"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("boradway"), &["broadway"]),
            (
                dictgen::InsensitiveStr::Ascii("milimiters"),
                &["millimeters"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupliagte"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("ngnix"), &["nginx"]),
            (dictgen::InsensitiveStr::Ascii("sigatures"), &["signatures"]),
            (dictgen::InsensitiveStr::Ascii("siginals"), &["signals"]),
            (
                dictgen::InsensitiveStr::Ascii("transcrito"),
                &["transcript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temparily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("imputin"), &["imputing"]),
            (
                dictgen::InsensitiveStr::Ascii("inctroduced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconditoners"),
                &["preconditioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("harased"), &["harassed"]),
            (dictgen::InsensitiveStr::Ascii("cacncel"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("duailty"), &["duality"]),
            (dictgen::InsensitiveStr::Ascii("benfiting"), &["benefiting"]),
            (
                dictgen::InsensitiveStr::Ascii("absoloutely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcribin"),
                &["transcribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("processsed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("pracitcal"), &["practical"]),
            (dictgen::InsensitiveStr::Ascii("optiones"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("specififiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("measurment"),
                &["measurement"],
            ),
            (dictgen::InsensitiveStr::Ascii("chimmenies"), &["chimneys"]),
            (
                dictgen::InsensitiveStr::Ascii("teloportation"),
                &["teleportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrally"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possability"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gloassaries"),
                &["glossaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optoinally"),
                &["optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("orriginal"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("retirbution"),
                &["retribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("usrernames"), &["usernames"]),
            (dictgen::InsensitiveStr::Ascii("latitide"), &["latitude"]),
            (
                dictgen::InsensitiveStr::Ascii("completelyl"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititailiaze"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahched"), &["cached"]),
            (dictgen::InsensitiveStr::Ascii("approches"), &["approaches"]),
            (
                dictgen::InsensitiveStr::Ascii("desciptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounad"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("optimimum"), &["optimum"]),
            (
                dictgen::InsensitiveStr::Ascii("anonymyzed"),
                &["anonymized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthuisast"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("coldplg"), &["coldplug"]),
            (
                dictgen::InsensitiveStr::Ascii("alterantive"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallelipiped"),
                &["parallelepiped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("booltloader"),
                &["bootloader"],
            ),
            (dictgen::InsensitiveStr::Ascii("asisst"), &["assist"]),
            (
                dictgen::InsensitiveStr::Ascii("transevier"),
                &["transceiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcitonal"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("privlege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("chilren"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("obscur"), &["obscure"]),
            (
                dictgen::InsensitiveStr::Ascii("meaninless"),
                &["meaningless"],
            ),
            (dictgen::InsensitiveStr::Ascii("noticible"), &["noticeable"]),
            (
                dictgen::InsensitiveStr::Ascii("anitsocial"),
                &["antisocial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tournmanets"),
                &["tournaments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denonimator"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("seseed"), &["secede"]),
            (dictgen::InsensitiveStr::Ascii("dgetttext"), &["dgettext"]),
            (dictgen::InsensitiveStr::Ascii("anounce"), &["announce"]),
            (dictgen::InsensitiveStr::Ascii("stadard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("farily"), &["fairly"]),
            (dictgen::InsensitiveStr::Ascii("proocecure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("aks"), &["ask"]),
            (dictgen::InsensitiveStr::Ascii("mappeed"), &["mapped"]),
            (dictgen::InsensitiveStr::Ascii("magazins"), &["magazines"]),
            (dictgen::InsensitiveStr::Ascii("exepect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("propoerty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("entquiry"), &["enquiry"]),
            (
                dictgen::InsensitiveStr::Ascii("beginninng"),
                &["beginnings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neccecarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordiantion"),
                &["coordination"],
            ),
            (dictgen::InsensitiveStr::Ascii("covnertor"), &["convertor"]),
            (dictgen::InsensitiveStr::Ascii("moustace"), &["moustache"]),
            (
                dictgen::InsensitiveStr::Ascii("articifially"),
                &["artificially"],
            ),
            (dictgen::InsensitiveStr::Ascii("attmpt"), &["attempt"]),
            (
                dictgen::InsensitiveStr::Ascii("dishinored"),
                &["dishonored"],
            ),
            (dictgen::InsensitiveStr::Ascii("kimerical"), &["chimerical"]),
            (
                dictgen::InsensitiveStr::Ascii("accelarate"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mediam"),
                &["median", "medial", "mediad", "medium"],
            ),
            (dictgen::InsensitiveStr::Ascii("loaclhost"), &["localhost"]),
            (
                dictgen::InsensitiveStr::Ascii("collectieve"),
                &["collective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumualtor"),
                &["accumulator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experssions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoever"), &["however"]),
            (dictgen::InsensitiveStr::Ascii("pairokeal"), &["parochial"]),
            (
                dictgen::InsensitiveStr::Ascii("competitoin"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleizing"),
                &["parallelizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("isuses"), &["issues"]),
            (dictgen::InsensitiveStr::Ascii("annotaor"), &["annotator"]),
            (
                dictgen::InsensitiveStr::Ascii("berforming"),
                &["performing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruptted"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convine"),
                &["convince", "combine", "confine", "convene"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cappabilities"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrariely"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolato"),
                &["consolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighboard"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ominpotent"),
                &["omnipotent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ommisiences"),
                &["omnisciences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afforementioned"),
                &["aforementioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("recide"), &["reside"]),
            (dictgen::InsensitiveStr::Ascii("afetr"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("mamalian"), &["mammalian"]),
            (
                dictgen::InsensitiveStr::Ascii("aound"),
                &["around", "sound"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootstapper"),
                &["bootstrapper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mormalises"),
                &["normalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogrithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ources"),
                &["sources", "ounces"],
            ),
            (dictgen::InsensitiveStr::Ascii("assiciate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("toliet"), &["toilet"]),
            (dictgen::InsensitiveStr::Ascii("assmume"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("locationes"), &["locations"]),
            (
                dictgen::InsensitiveStr::Ascii("unemployent"),
                &["unemployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitimistically"),
                &["optimistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinites"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactuable"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("shange"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("boolen"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("optimizate"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("transmissons"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemntation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compensante"),
                &["compensate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccripted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("specificiations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("brosable"), &["browsable"]),
            (dictgen::InsensitiveStr::Ascii("requestet"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("beethoveen"), &["beethoven"]),
            (dictgen::InsensitiveStr::Ascii("usting"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("approxomate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("coukd"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("listning"), &["listening"]),
            (
                dictgen::InsensitiveStr::Ascii("prelimenary"),
                &["preliminary"],
            ),
            (dictgen::InsensitiveStr::Ascii("workdpace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("selcetors"), &["selectors"]),
            (
                dictgen::InsensitiveStr::Ascii("definitoin"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("deficiet"), &["deficient"]),
            (
                dictgen::InsensitiveStr::Ascii("conveinient"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgareement"),
                &["disagreements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnderability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goverments"),
                &["governments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("managable"),
                &["manageable", "manageably"],
            ),
            (dictgen::InsensitiveStr::Ascii("protols"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("fuly"), &["fully"]),
            (
                dictgen::InsensitiveStr::Ascii("coniderable"),
                &["considerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("poltics"), &["politics"]),
            (
                dictgen::InsensitiveStr::Ascii("obsturction"),
                &["obstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("degreeees"), &["degrees"]),
            (
                dictgen::InsensitiveStr::Ascii("organzation"),
                &["organization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemetnations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterpirse"),
                &["enterprise"],
            ),
            (dictgen::InsensitiveStr::Ascii("deferrin"), &["deferring"]),
            (dictgen::InsensitiveStr::Ascii("outweighes"), &["outweighs"]),
            (
                dictgen::InsensitiveStr::Ascii("superseeds"),
                &["supersedes"],
            ),
            (dictgen::InsensitiveStr::Ascii("qith"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("webiste"), &["website"]),
            (
                dictgen::InsensitiveStr::Ascii("responisbilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("blasphemey"), &["blasphemy"]),
            (
                dictgen::InsensitiveStr::Ascii("springfeild"),
                &["springfield"],
            ),
            (dictgen::InsensitiveStr::Ascii("implicatii"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("eslewhere"), &["elsewhere"]),
            (dictgen::InsensitiveStr::Ascii("mulitcast"), &["multicast"]),
            (dictgen::InsensitiveStr::Ascii("preceision"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("characeters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininstrations"),
                &["administrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influenta"),
                &["influential"],
            ),
            (dictgen::InsensitiveStr::Ascii("comvert"), &["convert"]),
            (
                dictgen::InsensitiveStr::Ascii("warpprs"),
                &["warps", "wraps"],
            ),
            (dictgen::InsensitiveStr::Ascii("juvenlie"), &["juvenile"]),
            (dictgen::InsensitiveStr::Ascii("adjasant"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("speciallized"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("acrage"), &["acreage"]),
            (dictgen::InsensitiveStr::Ascii("expcets"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("alternativelly"),
                &["alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecjture"),
                &["conjecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("insipres"), &["inspires"]),
            (dictgen::InsensitiveStr::Ascii("legt"), &["left", "legit"]),
            (
                dictgen::InsensitiveStr::Ascii("unaped"),
                &["unmapped", "unpaid"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparistions"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("nmber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("nedia"), &["media"]),
            (dictgen::InsensitiveStr::Ascii("defanitly"), &["defiantly"]),
            (dictgen::InsensitiveStr::Ascii("boudaries"), &["boundaries"]),
            (
                dictgen::InsensitiveStr::Ascii("appraching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoprorate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("genertors"), &["generators"]),
            (dictgen::InsensitiveStr::Ascii("longitutde"), &["longitude"]),
            (dictgen::InsensitiveStr::Ascii("hyperbolie"), &["hyperbole"]),
            (dictgen::InsensitiveStr::Ascii("illsuions"), &["illusions"]),
            (dictgen::InsensitiveStr::Ascii("wintson"), &["winston"]),
            (
                dictgen::InsensitiveStr::Ascii("accelorators"),
                &["accelerators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carbohyrdates"),
                &["carbohydrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("prosepct"), &["prospect"]),
            (
                dictgen::InsensitiveStr::Ascii("possbility"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("collpased"), &["collapsed"]),
            (dictgen::InsensitiveStr::Ascii("loactes"), &["locates"]),
            (dictgen::InsensitiveStr::Ascii("effetct"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("isntallations"),
                &["installations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aplication"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technologia"),
                &["technological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exctracting"),
                &["extracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("oveflowed"), &["overflowed"]),
            (dictgen::InsensitiveStr::Ascii("simpy"), &["simply"]),
            (
                dictgen::InsensitiveStr::Ascii("juxtification"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiaters"),
                &["negotiators", "negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desallocates"),
                &["deallocates"],
            ),
            (dictgen::InsensitiveStr::Ascii("automony"), &["autonomy"]),
            (dictgen::InsensitiveStr::Ascii("detroyed"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("alientating"),
                &["alienating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogeneized"),
                &["homogenized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bord"),
                &["board", "bored", "border"],
            ),
            (dictgen::InsensitiveStr::Ascii("alphabt"), &["alphabet"]),
            (dictgen::InsensitiveStr::Ascii("provences"), &["provinces"]),
            (dictgen::InsensitiveStr::Ascii("barbarin"), &["barbarian"]),
            (dictgen::InsensitiveStr::Ascii("unanomous"), &["unanimous"]),
            (dictgen::InsensitiveStr::Ascii("attrubute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("preprocessers"),
                &["preprocessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ourselfe"),
                &["ourselves", "ourself"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicipline"), &["discipline"]),
            (
                dictgen::InsensitiveStr::Ascii("intanciate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("remakin"), &["remaking"]),
            (
                dictgen::InsensitiveStr::Ascii("defibately"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espisodical"),
                &["episodical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ceritfication"),
                &["certifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("gratefull"), &["grateful"]),
            (dictgen::InsensitiveStr::Ascii("liscensed"), &["licensed"]),
            (dictgen::InsensitiveStr::Ascii("mexicanese"), &["mexicans"]),
            (dictgen::InsensitiveStr::Ascii("strusture"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("midfied"), &["midfield"]),
            (
                dictgen::InsensitiveStr::Ascii("experamentations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("descide"), &["decide"]),
            (dictgen::InsensitiveStr::Ascii("copuright"), &["copyright"]),
            (
                dictgen::InsensitiveStr::Ascii("connonation"),
                &["connotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scubscribed"),
                &["subscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporarity"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premutations"),
                &["permutations"],
            ),
            (dictgen::InsensitiveStr::Ascii("oommits"), &["commits"]),
            (dictgen::InsensitiveStr::Ascii("caroseles"), &["carousels"]),
            (
                dictgen::InsensitiveStr::Ascii("sufocation"),
                &["suffocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuploding"),
                &["reuploading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desserialise"),
                &["deserialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("promatory"), &["promontory"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigiose"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("steapness"), &["steepness"]),
            (
                dictgen::InsensitiveStr::Ascii("checlerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creen"),
                &["screen", "green", "creep", "careen"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppenings"), &["openings"]),
            (dictgen::InsensitiveStr::Ascii("meanigns"), &["meanings"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotioates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorhitm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("camboida"), &["cambodia"]),
            (dictgen::InsensitiveStr::Ascii("grooup"), &["group"]),
            (
                dictgen::InsensitiveStr::Ascii("parliamentery"),
                &["parliamentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foundatin"),
                &["foundations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parametrse"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependeing"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("matinay"), &["matinee"]),
            (dictgen::InsensitiveStr::Ascii("prepaired"), &["prepared"]),
            (
                dictgen::InsensitiveStr::Ascii("hypohthezied"),
                &["hypothesized"],
            ),
            (dictgen::InsensitiveStr::Ascii("impoving"), &["improving"]),
            (
                dictgen::InsensitiveStr::Ascii("palces"),
                &["places", "pales"],
            ),
            (dictgen::InsensitiveStr::Ascii("civilains"), &["civilians"]),
            (
                dictgen::InsensitiveStr::Ascii("dissassociated"),
                &["disassociated"],
            ),
            (dictgen::InsensitiveStr::Ascii("wordpresss"), &["wordpress"]),
            (
                dictgen::InsensitiveStr::Ascii("deriatives"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("hwile"), &["while"]),
            (dictgen::InsensitiveStr::Ascii("countrys"), &["countries"]),
            (
                dictgen::InsensitiveStr::Ascii("retreivers"),
                &["retrievers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hosptial"), &["hospital"]),
            (
                dictgen::InsensitiveStr::Ascii("hirachies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ligthhouse"),
                &["lighthouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authendicating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("accurable"), &["accurate"]),
            (
                dictgen::InsensitiveStr::Ascii("canceltation"),
                &["cancellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("compoesd"), &["composed"]),
            (
                dictgen::InsensitiveStr::Ascii("identifiying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("licencin"), &["licencing"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisitior"),
                &["inquisitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exculsives"),
                &["exclusives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disproportionnate"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheeseus"), &["cheeses"]),
            (
                dictgen::InsensitiveStr::Ascii("outstandin"),
                &["outstanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("kighboshed"), &["kiboshed"]),
            (
                dictgen::InsensitiveStr::Ascii("enforsement"),
                &["endorsement", "enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("enclosng"), &["enclosing"]),
            (dictgen::InsensitiveStr::Ascii("redefiend"), &["redefined"]),
            (dictgen::InsensitiveStr::Ascii("descendat"), &["descendant"]),
            (dictgen::InsensitiveStr::Ascii("commondly"), &["commonly"]),
            (
                dictgen::InsensitiveStr::Ascii("regconising"),
                &["recognising"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpet"), &["accept"]),
            (dictgen::InsensitiveStr::Ascii("sirecting"), &["directing"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensiones"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("barrakcs"), &["barracks"]),
            (dictgen::InsensitiveStr::Ascii("systtem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("generilises"),
                &["generalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("intilize"), &["initialize"]),
            (dictgen::InsensitiveStr::Ascii("mollested"), &["molested"]),
            (
                dictgen::InsensitiveStr::Ascii("simulatation"),
                &["simulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signle"),
                &["single", "signal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilesations"),
                &["civilisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("grandient"), &["gradient"]),
            (dictgen::InsensitiveStr::Ascii("insurgeny"), &["insurgency"]),
            (dictgen::InsensitiveStr::Ascii("prufe"), &["proof"]),
            (
                dictgen::InsensitiveStr::Ascii("technicials"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemtation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciying"), &["specifying"]),
            (dictgen::InsensitiveStr::Ascii("srinkd"), &["shrunk"]),
            (dictgen::InsensitiveStr::Ascii("straings"), &["strains"]),
            (dictgen::InsensitiveStr::Ascii("stanalone"), &["standalone"]),
            (dictgen::InsensitiveStr::Ascii("varios"), &["various"]),
            (dictgen::InsensitiveStr::Ascii("segmenst"), &["segments"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretacion"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("asny"), &["any"]),
            (
                dictgen::InsensitiveStr::Ascii("incroporation"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dermined"), &["determined"]),
            (dictgen::InsensitiveStr::Ascii("falgship"), &["flagship"]),
            (dictgen::InsensitiveStr::Ascii("chanched"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("convines"),
                &["combines", "convinces", "confines", "convenes"],
            ),
            (dictgen::InsensitiveStr::Ascii("emables"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("idependently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("sercived"), &["serviced"]),
            (dictgen::InsensitiveStr::Ascii("autmate"), &["automate"]),
            (dictgen::InsensitiveStr::Ascii("entierty"), &["entirety"]),
            (
                dictgen::InsensitiveStr::Ascii("serialiasation"),
                &["serialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("miliary"), &["military"]),
            (dictgen::InsensitiveStr::Ascii("sadistc"), &["sadistic"]),
            (
                dictgen::InsensitiveStr::Ascii("activaste"),
                &["activates", "activate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutipule"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("polarlises"), &["polarises"]),
            (
                dictgen::InsensitiveStr::Ascii("aggaravating"),
                &["aggravating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabitility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("frankiln"), &["franklin"]),
            (dictgen::InsensitiveStr::Ascii("mantains"), &["maintains"]),
            (
                dictgen::InsensitiveStr::Ascii("desctiptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apllicatins"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("velantine"), &["valentine"]),
            (dictgen::InsensitiveStr::Ascii("additivive"), &["additive"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiecs"),
                &["specifics", "specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("purgable"), &["purgeable"]),
            (dictgen::InsensitiveStr::Ascii("desrire"), &["desire"]),
            (dictgen::InsensitiveStr::Ascii("srcollbar"), &["scrollbar"]),
            (
                dictgen::InsensitiveStr::Ascii("expermientations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignroed"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("micoscopies"),
                &["microscopies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gothenberg"),
                &["gothenburg"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archeaological"),
                &["archeological"],
            ),
            (dictgen::InsensitiveStr::Ascii("intevene"), &["intervene"]),
            (
                dictgen::InsensitiveStr::Ascii("describtions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antibitoic"),
                &["antibiotic"],
            ),
            (dictgen::InsensitiveStr::Ascii("anectodes"), &["anecdotes"]),
            (
                dictgen::InsensitiveStr::Ascii("terminte"),
                &["terminate", "termite"],
            ),
            (dictgen::InsensitiveStr::Ascii("cognizent"), &["cognizant"]),
            (dictgen::InsensitiveStr::Ascii("commplain"), &["complain"]),
            (
                dictgen::InsensitiveStr::Ascii("intendation"),
                &["indentation", "intention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitplicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("aesy"), &["easy"]),
            (
                dictgen::InsensitiveStr::Ascii("condictional"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshakre"), &["handshake"]),
            (
                dictgen::InsensitiveStr::Ascii("classsification"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("actibity"), &["activity"]),
            (
                dictgen::InsensitiveStr::Ascii("confussion"),
                &["confession", "confusion", "concussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("defficult"), &["difficult"]),
            (
                dictgen::InsensitiveStr::Ascii("annyoances"),
                &["annoyances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropritately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconviniances"),
                &["inconveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("morever"), &["moreover"]),
            (
                dictgen::InsensitiveStr::Ascii("susbsystems"),
                &["subsystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("phyton"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("miximum"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("incorret"), &["incorrect"]),
            (
                dictgen::InsensitiveStr::Ascii("unacceptible"),
                &["unacceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("practcies"), &["practices"]),
            (dictgen::InsensitiveStr::Ascii("cellulaire"), &["cellular"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfitifers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschiatric"),
                &["psychiatric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benefitd"),
                &["benefited", "benefits"],
            ),
            (dictgen::InsensitiveStr::Ascii("adderess"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("reminiscint"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("conact"), &["contact"]),
            (
                dictgen::InsensitiveStr::Ascii("abour"),
                &["about", "labour"],
            ),
            (dictgen::InsensitiveStr::Ascii("decocded"), &["decoded"]),
            (dictgen::InsensitiveStr::Ascii("candidata"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("unepected"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrviations"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instlalations"),
                &["installations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presonalities"),
                &["personalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("requesing"), &["requesting"]),
            (dictgen::InsensitiveStr::Ascii("amortizin"), &["amortizing"]),
            (dictgen::InsensitiveStr::Ascii("tresuary"), &["treasury"]),
            (dictgen::InsensitiveStr::Ascii("cluter"), &["cluster"]),
            (dictgen::InsensitiveStr::Ascii("pwoer"), &["power"]),
            (dictgen::InsensitiveStr::Ascii("sxmbol"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("admissability"),
                &["admissibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multidimentional"),
                &["multidimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("afforable"), &["affordable"]),
            (
                dictgen::InsensitiveStr::Ascii("specialiazing"),
                &["specializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("universirty"),
                &["university"],
            ),
            (dictgen::InsensitiveStr::Ascii("unitilty"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailiability"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("culterally"),
                &["culturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("mortum"), &["mortem"]),
            (
                dictgen::InsensitiveStr::Ascii("unexcpectedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("documnent"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("decoratored"),
                &["decorated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ege"), &["edge"]),
            (
                dictgen::InsensitiveStr::Ascii("discconetions"),
                &["disconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("roperties"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("descibing"), &["describing"]),
            (dictgen::InsensitiveStr::Ascii("cheksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("suppor"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("ignorng"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("valtage"), &["voltage"]),
            (dictgen::InsensitiveStr::Ascii("modellin"), &["modelling"]),
            (
                dictgen::InsensitiveStr::Ascii("terain"),
                &["terrain", "train"],
            ),
            (dictgen::InsensitiveStr::Ascii("argment"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("repitition"),
                &["repetition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaatl"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("scemas"), &["schemas"]),
            (
                dictgen::InsensitiveStr::Ascii("conceptial"),
                &["conceptual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trascation"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("tootonic"), &["teutonic"]),
            (
                dictgen::InsensitiveStr::Ascii("liberalizm"),
                &["liberalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("jamacia"), &["jamaica"]),
            (
                dictgen::InsensitiveStr::Ascii("earthquaks"),
                &["earthquakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passporters"),
                &["passports"],
            ),
            (dictgen::InsensitiveStr::Ascii("beaviours"), &["behaviours"]),
            (dictgen::InsensitiveStr::Ascii("meaasured"), &["measured"]),
            (dictgen::InsensitiveStr::Ascii("exatcta"), &["exacta"]),
            (
                dictgen::InsensitiveStr::Ascii("disributions"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("storyteling"),
                &["storytelling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompatibilities"),
                &["incompatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pesticidas"),
                &["pesticides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimazers"),
                &["optimizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefernce"),
                &["dereference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenemonal"),
                &["phenomenal"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulatr"), &["regulator"]),
            (
                dictgen::InsensitiveStr::Ascii("transformatted"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocasional"), &["occasional"]),
            (dictgen::InsensitiveStr::Ascii("sourse"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("constituants"),
                &["constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepetaror"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciae"),
                &["appreciates", "appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optisimation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("trubled"), &["troubled"]),
            (dictgen::InsensitiveStr::Ascii("destoyers"), &["destroyers"]),
            (
                dictgen::InsensitiveStr::Ascii("compresing"),
                &["compressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contenstants"),
                &["contestants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentraing"),
                &["concentrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kuberenetes"),
                &["kubernetes"],
            ),
            (dictgen::InsensitiveStr::Ascii("exucuting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("escusing"), &["excusing"]),
            (dictgen::InsensitiveStr::Ascii("histgram"), &["histogram"]),
            (dictgen::InsensitiveStr::Ascii("wokred"), &["worked"]),
            (
                dictgen::InsensitiveStr::Ascii("perssiuosly"),
                &["preciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("givien"), &["given"]),
            (dictgen::InsensitiveStr::Ascii("intalls"), &["installs"]),
            (
                dictgen::InsensitiveStr::Ascii("alegiances"),
                &["allegiances"],
            ),
            (dictgen::InsensitiveStr::Ascii("componeent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitialse"),
                &["uninitialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lighthours"),
                &["lighthouse"],
            ),
            (dictgen::InsensitiveStr::Ascii("approuved"), &["approved"]),
            (
                dictgen::InsensitiveStr::Ascii("inherrently"),
                &["inherently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitratiojn"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("aircarft"), &["aircraft"]),
            (
                dictgen::InsensitiveStr::Ascii("concanetes"),
                &["concatenates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responisble"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eplicit"),
                &["explicit", "elicit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jugdements"),
                &["judgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frankinstein"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capapcitors"),
                &["capacitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporper"), &["improper"]),
            (dictgen::InsensitiveStr::Ascii("calclulate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("migrainers"), &["migraines"]),
            (
                dictgen::InsensitiveStr::Ascii("clera"),
                &["clear", "sclera"],
            ),
            (dictgen::InsensitiveStr::Ascii("manfestos"), &["manifestos"]),
            (
                dictgen::InsensitiveStr::Ascii("expextancy"),
                &["expectancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterers"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constuctors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialisaes"),
                &["deserialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacehlors"), &["bachelors"]),
            (
                dictgen::InsensitiveStr::Ascii("hellucination"),
                &["hallucination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommneds"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("fininsh"), &["finnish"]),
            (dictgen::InsensitiveStr::Ascii("preapare"), &["prepare"]),
            (dictgen::InsensitiveStr::Ascii("documens"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("indicaste"), &["indicates"]),
            (dictgen::InsensitiveStr::Ascii("doamins"), &["domains"]),
            (dictgen::InsensitiveStr::Ascii("invloves"), &["involves"]),
            (dictgen::InsensitiveStr::Ascii("paralely"), &["parallelly"]),
            (dictgen::InsensitiveStr::Ascii("dialectes"), &["dialects"]),
            (
                dictgen::InsensitiveStr::Ascii("unconsicous"),
                &["unconscious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquaintaince"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("reslved"), &["resolved"]),
            (
                dictgen::InsensitiveStr::Ascii("acordingly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mesurements"),
                &["measurements"],
            ),
            (dictgen::InsensitiveStr::Ascii("detectoare"), &["detector"]),
            (
                dictgen::InsensitiveStr::Ascii("chechboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregulared"),
                &["unregulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomeanses"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("featuer"),
                &["feature", "feather"],
            ),
            (dictgen::InsensitiveStr::Ascii("lisetns"), &["listens"]),
            (dictgen::InsensitiveStr::Ascii("storate"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("psychiatic"),
                &["psychiatric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuarble"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("tansform"), &["transform"]),
            (
                dictgen::InsensitiveStr::Ascii("procedding"),
                &["proceeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("contractos"), &["contracts"]),
            (
                dictgen::InsensitiveStr::Ascii("surpressed"),
                &["suppressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("minnimums"), &["minimums"]),
            (dictgen::InsensitiveStr::Ascii("awya"), &["away"]),
            (
                dictgen::InsensitiveStr::Ascii("teatotlers"),
                &["teetotalers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounnd"), &["bound"]),
            (
                dictgen::InsensitiveStr::Ascii("achitecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("earthquacks"),
                &["earthquakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("chalenge"), &["challenge"]),
            (
                dictgen::InsensitiveStr::Ascii("grilfriends"),
                &["girlfriends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactheristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhoood"),
                &["neighborhoods", "neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("overiden"), &["overridden"]),
            (
                dictgen::InsensitiveStr::Ascii("possesions"),
                &["possessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("succusses"), &["successes"]),
            (
                dictgen::InsensitiveStr::Ascii("intresting"),
                &["interesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("revelead"), &["revealed"]),
            (
                dictgen::InsensitiveStr::Ascii("thankfull"),
                &["thankful", "thankfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("victorin"), &["victorian"]),
            (
                dictgen::InsensitiveStr::Ascii("construcion"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cardiovascualar"),
                &["cardiovascular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrants"),
                &["rearrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("idendify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("sepatate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("autonymous"),
                &["autonomous"],
            ),
            (dictgen::InsensitiveStr::Ascii("buils"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("tramautic"), &["traumatic"]),
            (dictgen::InsensitiveStr::Ascii("untrached"), &["untracked"]),
            (dictgen::InsensitiveStr::Ascii("pscyhed"), &["psyched"]),
            (dictgen::InsensitiveStr::Ascii("adrain"), &["adrian"]),
            (dictgen::InsensitiveStr::Ascii("radiaton"), &["radiation"]),
            (dictgen::InsensitiveStr::Ascii("versionn"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("expaneded"), &["expanded"]),
            (dictgen::InsensitiveStr::Ascii("trubbled"), &["troubled"]),
            (dictgen::InsensitiveStr::Ascii("cahanged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("undoubtley"),
                &["undoubtedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delivative"),
                &["derivative"],
            ),
            (dictgen::InsensitiveStr::Ascii("invice"), &["invoice"]),
            (
                dictgen::InsensitiveStr::Ascii("overwrides"),
                &["overwrites", "overrides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appoitnment"),
                &["appointments", "appointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheking"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("contructor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestacion"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("amercians"), &["americans"]),
            (dictgen::InsensitiveStr::Ascii("immenseley"), &["immensely"]),
            (dictgen::InsensitiveStr::Ascii("vaue"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("enrty"), &["entry"]),
            (
                dictgen::InsensitiveStr::Ascii("automcatically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("durantions"), &["durations"]),
            (dictgen::InsensitiveStr::Ascii("tweek"), &["tweak"]),
            (
                dictgen::InsensitiveStr::Ascii("supersedin"),
                &["superseding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steroetypical"),
                &["stereotypical"],
            ),
            (dictgen::InsensitiveStr::Ascii("explane"), &["explain"]),
            (
                dictgen::InsensitiveStr::Ascii("citizinship"),
                &["citizenship"],
            ),
            (dictgen::InsensitiveStr::Ascii("christiaan"), &["christian"]),
            (dictgen::InsensitiveStr::Ascii("langyages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("pubilshers"),
                &["publishers"],
            ),
            (dictgen::InsensitiveStr::Ascii("iiterator"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("playlise"), &["playlist"]),
            (dictgen::InsensitiveStr::Ascii("classfies"), &["classifies"]),
            (
                dictgen::InsensitiveStr::Ascii("collpasing"),
                &["collapsing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recostruct"),
                &["reconstruct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticly"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteritics"),
                &["characteristic", "characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adddressing"),
                &["addressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obstrucion"),
                &["obstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curruption"),
                &["corruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorparate"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periodicaly"),
                &["periodically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("resercher"), &["researcher"]),
            (dictgen::InsensitiveStr::Ascii("frustrats"), &["frustrates"]),
            (dictgen::InsensitiveStr::Ascii("choatic"), &["chaotic"]),
            (dictgen::InsensitiveStr::Ascii("ampliflies"), &["amplifies"]),
            (dictgen::InsensitiveStr::Ascii("folwoiwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("santroum"), &["santorum"]),
            (
                dictgen::InsensitiveStr::Ascii("branche"),
                &["branch", "branched", "branches"],
            ),
            (dictgen::InsensitiveStr::Ascii("hazzle"), &["hassle"]),
            (
                dictgen::InsensitiveStr::Ascii("memeasurement"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admnistrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("passsed"), &["passed"]),
            (
                dictgen::InsensitiveStr::Ascii("dealershp"),
                &["dealerships"],
            ),
            (dictgen::InsensitiveStr::Ascii("appendt"), &["append"]),
            (dictgen::InsensitiveStr::Ascii("alchmey"), &["alchemy"]),
            (dictgen::InsensitiveStr::Ascii("benefica"), &["beneficial"]),
            (dictgen::InsensitiveStr::Ascii("diagnol"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("degeree"), &["degree"]),
            (
                dictgen::InsensitiveStr::Ascii("combiniatorics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaoiding"),
                &["avoiding", "voiding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interprerter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonous"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempraraly"),
                &["temporarily", "temporally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rememembered"),
                &["remembered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsturctions"),
                &["reconstructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberation"),
                &["aberration"],
            ),
            (dictgen::InsensitiveStr::Ascii("condusive"), &["conducive"]),
            (dictgen::InsensitiveStr::Ascii("tounreys"), &["tourneys"]),
            (
                dictgen::InsensitiveStr::Ascii("palestinier"),
                &["palestine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assemalating"),
                &["assimilating"],
            ),
            (dictgen::InsensitiveStr::Ascii("whethre"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("grapgic"), &["graphic"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulising"),
                &["virtualising"],
            ),
            (dictgen::InsensitiveStr::Ascii("coontent"), &["content"]),
            (
                dictgen::InsensitiveStr::Ascii("instinctivelly"),
                &["instinctively"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeuctes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("pathwya"), &["pathway"]),
            (
                dictgen::InsensitiveStr::Ascii("misellaneous"),
                &["miscellaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autmoations"),
                &["automations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sytemwindow"),
                &["systemwindow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protectron"),
                &["protection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denpendents"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasients"), &["transients"]),
            (
                dictgen::InsensitiveStr::Ascii("commonspace"),
                &["commonplace"],
            ),
            (dictgen::InsensitiveStr::Ascii("youi"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("esclusions"),
                &["exclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("directoory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("santorm"), &["santorum"]),
            (dictgen::InsensitiveStr::Ascii("writng"), &["writing"]),
            (
                dictgen::InsensitiveStr::Ascii("dinteractively"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penssylvania"),
                &["pennsylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("satandards"), &["standards"]),
            (
                dictgen::InsensitiveStr::Ascii("chechkerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetimental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanetations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("kibutzim"), &["kibbutzim"]),
            (dictgen::InsensitiveStr::Ascii("seround"), &["surround"]),
            (dictgen::InsensitiveStr::Ascii("eduation"), &["education"]),
            (
                dictgen::InsensitiveStr::Ascii("postcondtion"),
                &["postcondition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponetial"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissmised"), &["dismissed"]),
            (
                dictgen::InsensitiveStr::Ascii("whather"),
                &["whether", "weather"],
            ),
            (dictgen::InsensitiveStr::Ascii("approuves"), &["approves"]),
            (dictgen::InsensitiveStr::Ascii("peedmont"), &["piedmont"]),
            (
                dictgen::InsensitiveStr::Ascii("wokrbenches"),
                &["workbenches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positionial"),
                &["positional"],
            ),
            (dictgen::InsensitiveStr::Ascii("turntabe"), &["turntable"]),
            (dictgen::InsensitiveStr::Ascii("failues"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("trarget"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("temrination"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wikileakers"),
                &["wikileaks"],
            ),
            (dictgen::InsensitiveStr::Ascii("prototyed"), &["prototyped"]),
            (
                dictgen::InsensitiveStr::Ascii("inititializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("lieks"), &["likes"]),
            (dictgen::InsensitiveStr::Ascii("havea"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("appriecating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("shema"), &["schema"]),
            (dictgen::InsensitiveStr::Ascii("jossle"), &["jostle"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitalized"),
                &["reinitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradisional"),
                &["traditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particant"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("unstalbe"), &["unstable"]),
            (
                dictgen::InsensitiveStr::Ascii("revaluated"),
                &["reevaluated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cicrle"), &["circle"]),
            (
                dictgen::InsensitiveStr::Ascii("sunconscious"),
                &["subconscious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transtitioned"),
                &["transitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("contestans"), &["contests"]),
            (dictgen::InsensitiveStr::Ascii("visivble"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("kwnown"), &["known"]),
            (
                dictgen::InsensitiveStr::Ascii("deseralised"),
                &["deserialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("sistemic"), &["systemic"]),
            (dictgen::InsensitiveStr::Ascii("owership"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("buiilders"), &["builders"]),
            (
                dictgen::InsensitiveStr::Ascii("syptomatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("amrenia"), &["armenia"]),
            (dictgen::InsensitiveStr::Ascii("pacjages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("buearucrats"),
                &["bureaucrats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructores"),
                &["instructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenating"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilzations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributin"),
                &["contribution", "contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("deifnitly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("gettter"), &["getter"]),
            (
                dictgen::InsensitiveStr::Ascii("complicatred"),
                &["complicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("allegedy"), &["allegedly"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliazing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarcerato"),
                &["incarceration"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploymer"), &["polymer"]),
            (
                dictgen::InsensitiveStr::Ascii("suspicously"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpahbetically"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("autorizes"), &["authorizes"]),
            (
                dictgen::InsensitiveStr::Ascii("apologiseing"),
                &["apologising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconeccted"),
                &["reconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("flacour"), &["flavour"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("explaing"), &["explaining"]),
            (dictgen::InsensitiveStr::Ascii("noet"), &["note"]),
            (dictgen::InsensitiveStr::Ascii("tighly"), &["tightly"]),
            (dictgen::InsensitiveStr::Ascii("configue"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("mutilplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("socilaism"), &["socialism"]),
            (dictgen::InsensitiveStr::Ascii("damagin"), &["damaging"]),
            (dictgen::InsensitiveStr::Ascii("salughter"), &["slaughter"]),
            (
                dictgen::InsensitiveStr::Ascii("exceoting"),
                &["excepting", "exempting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neckbearders"),
                &["neckbeards"],
            ),
            (dictgen::InsensitiveStr::Ascii("iether"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("centralizin"),
                &["centralizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sociopathes"),
                &["sociopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("tsnuami"), &["tsunami"]),
            (
                dictgen::InsensitiveStr::Ascii("predicitng"),
                &["predicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("opnssl"), &["openssl"]),
            (dictgen::InsensitiveStr::Ascii("quesitons"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("slewthing"), &["sleuthing"]),
            (dictgen::InsensitiveStr::Ascii("coomented"), &["commented"]),
            (dictgen::InsensitiveStr::Ascii("moleculair"), &["molecular"]),
            (
                dictgen::InsensitiveStr::Ascii("meethodology"),
                &["methodology"],
            ),
            (dictgen::InsensitiveStr::Ascii("trageting"), &["targeting"]),
            (
                dictgen::InsensitiveStr::Ascii("authroised"),
                &["authorised"],
            ),
            (dictgen::InsensitiveStr::Ascii("unselcts"), &["unselects"]),
            (
                dictgen::InsensitiveStr::Ascii("saddly"),
                &["saddle", "sadly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persectives"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assistantes"),
                &["assistants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresponsivness"),
                &["unresponsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sweatshits"),
                &["sweatshirt"],
            ),
            (dictgen::InsensitiveStr::Ascii("armstorng"), &["armstrong"]),
            (
                dictgen::InsensitiveStr::Ascii("integradions"),
                &["integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disallusioned"),
                &["disillusioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("anayltics"), &["analytics"]),
            (
                dictgen::InsensitiveStr::Ascii("motivationals"),
                &["motivations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerciseing"),
                &["exercising"],
            ),
            (dictgen::InsensitiveStr::Ascii("keybaords"), &["keyboards"]),
            (
                dictgen::InsensitiveStr::Ascii("addittions"),
                &["addictions", "additions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serailized"),
                &["serialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("tupels"), &["tuples"]),
            (dictgen::InsensitiveStr::Ascii("retrives"), &["retrieves"]),
            (
                dictgen::InsensitiveStr::Ascii("statndalone"),
                &["standalone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opeaaration"),
                &["operation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstantings"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defelction"),
                &["deflection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decompreser"),
                &["decompressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exctraction"),
                &["extraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buinsesses"),
                &["businesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconviniant"),
                &["inconvenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("librarry"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("annoators"), &["annotators"]),
            (
                dictgen::InsensitiveStr::Ascii("disgrunteled"),
                &["disgruntled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiemented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("wokrload"), &["workload"]),
            (
                dictgen::InsensitiveStr::Ascii("ampitheater"),
                &["amphitheater"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniects"), &["injects"]),
            (
                dictgen::InsensitiveStr::Ascii("analysies"),
                &["analyses", "analysis"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyisis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarachically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enfocement"),
                &["enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("taiwanee"), &["taiwanese"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratons"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("patroits"), &["patriots"]),
            (
                dictgen::InsensitiveStr::Ascii("genereations"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcotory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("arbutrarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("excerises"), &["exercises"]),
            (
                dictgen::InsensitiveStr::Ascii("maniputor"),
                &["manipulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("electroinc"), &["electron"]),
            (dictgen::InsensitiveStr::Ascii("licencie"), &["licence"]),
            (dictgen::InsensitiveStr::Ascii("anylzing"), &["analyzing"]),
            (
                dictgen::InsensitiveStr::Ascii("opproximating"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminiation"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tretrahedralization"),
                &["tetrahedralization"],
            ),
            (dictgen::InsensitiveStr::Ascii("eachs"), &["each"]),
            (
                dictgen::InsensitiveStr::Ascii("convineances"),
                &["conveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("classses"), &["classes"]),
            (
                dictgen::InsensitiveStr::Ascii("adjacancies"),
                &["adjacencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suceessors"),
                &["successors"],
            ),
            (dictgen::InsensitiveStr::Ascii("tecniques"), &["techniques"]),
            (dictgen::InsensitiveStr::Ascii("busnesses"), &["businesses"]),
            (dictgen::InsensitiveStr::Ascii("arrranges"), &["arranges"]),
            (dictgen::InsensitiveStr::Ascii("ecstascy"), &["ecstasy"]),
            (dictgen::InsensitiveStr::Ascii("incluging"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("bariier"), &["barrier"]),
            (dictgen::InsensitiveStr::Ascii("etxt"), &["text"]),
            (
                dictgen::InsensitiveStr::Ascii("greivences"),
                &["grievances"],
            ),
            (dictgen::InsensitiveStr::Ascii("automaitc"), &["automatic"]),
            (
                dictgen::InsensitiveStr::Ascii("diferrentiating"),
                &["differentiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("briagde"), &["brigade"]),
            (
                dictgen::InsensitiveStr::Ascii("sistemists"),
                &["systemists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimnts"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("specimines"), &["specimen"]),
            (
                dictgen::InsensitiveStr::Ascii("tempraaily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaluate"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("symptumaticlly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interchangably"),
                &["interchangeably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sience"),
                &["science", "silence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connnections"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("exacute"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("mischevous"),
                &["mischievous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sophisicated"),
                &["sophisticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistmatched"),
                &["mismatched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfromed"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resurses"),
                &["recurses", "resources"],
            ),
            (dictgen::InsensitiveStr::Ascii("trnasform"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("commitin"), &["committing"]),
            (
                dictgen::InsensitiveStr::Ascii("excecised"),
                &["excised", "exercised"],
            ),
            (dictgen::InsensitiveStr::Ascii("protraits"), &["portraits"]),
            (dictgen::InsensitiveStr::Ascii("insalled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("decryp"), &["decrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("termiators"),
                &["terminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("coputer"), &["computer"]),
            (
                dictgen::InsensitiveStr::Ascii("finialising"),
                &["finalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrispondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomalized"), &["normalized"]),
            (dictgen::InsensitiveStr::Ascii("examinerad"), &["examined"]),
            (
                dictgen::InsensitiveStr::Ascii("cyphersuite"),
                &["ciphersuite"],
            ),
            (dictgen::InsensitiveStr::Ascii("parmas"), &["params"]),
            (dictgen::InsensitiveStr::Ascii("allegedley"), &["allegedly"]),
            (dictgen::InsensitiveStr::Ascii("sacrifie"), &["sacrifice"]),
            (dictgen::InsensitiveStr::Ascii("occures"), &["occurs"]),
            (dictgen::InsensitiveStr::Ascii("socekts"), &["sockets"]),
            (dictgen::InsensitiveStr::Ascii("safefy"), &["safety"]),
            (
                dictgen::InsensitiveStr::Ascii("discovereability"),
                &["discoverability"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissmissed"), &["dismissed"]),
            (
                dictgen::InsensitiveStr::Ascii("perpertated"),
                &["perpetuated"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconts"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("expanson"), &["expansion"]),
            (
                dictgen::InsensitiveStr::Ascii("preficiensy"),
                &["proficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elimentary"),
                &["elementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("asignees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("fortunetely"),
                &["fortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("smaler"), &["smaller"]),
            (dictgen::InsensitiveStr::Ascii("souble"), &["double"]),
            (
                dictgen::InsensitiveStr::Ascii("interveing"),
                &["intervening"],
            ),
            (dictgen::InsensitiveStr::Ascii("linkfy"), &["linkify"]),
            (dictgen::InsensitiveStr::Ascii("serverles"), &["serverless"]),
            (
                dictgen::InsensitiveStr::Ascii("messagers"),
                &["messages", "messengers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentracion"),
                &["concentration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistendt"),
                &["consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("processore"), &["processor"]),
            (dictgen::InsensitiveStr::Ascii("aquaruim"), &["aquarium"]),
            (
                dictgen::InsensitiveStr::Ascii("conincidence"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congergation"),
                &["congregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculting"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compablity"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("versatle"), &["versatile"]),
            (dictgen::InsensitiveStr::Ascii("resoluton"), &["resolution"]),
            (
                dictgen::InsensitiveStr::Ascii("suppresssor"),
                &["suppressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("conduict"), &["conduit"]),
            (
                dictgen::InsensitiveStr::Ascii("ressurrection"),
                &["resurrection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contingeny"),
                &["contingency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassemblerr"),
                &["disassembler"],
            ),
            (dictgen::InsensitiveStr::Ascii("exteral"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlatable"),
                &["translatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("libraty"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("solgered"), &["soldiered"]),
            (
                dictgen::InsensitiveStr::Ascii("itersection"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("happes"), &["happens"]),
            (dictgen::InsensitiveStr::Ascii("lcation"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("necassry"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("apppending"), &["appending"]),
            (
                dictgen::InsensitiveStr::Ascii("resaurants"),
                &["restaurants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complextion"),
                &["complexion"],
            ),
            (dictgen::InsensitiveStr::Ascii("anectodal"), &["anecdotal"]),
            (
                dictgen::InsensitiveStr::Ascii("destrcuted"),
                &["destructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("farenheit"), &["fahrenheit"]),
            (dictgen::InsensitiveStr::Ascii("trhread"), &["thread"]),
            (dictgen::InsensitiveStr::Ascii("switchted"), &["switched"]),
            (dictgen::InsensitiveStr::Ascii("dectector"), &["detector"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstandng"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentaly"),
                &["fundamentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("connct"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("intercating"),
                &["interacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("astrisks"), &["asterisks"]),
            (dictgen::InsensitiveStr::Ascii("implct"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("notitied"), &["notified"]),
            (dictgen::InsensitiveStr::Ascii("secnes"), &["scenes"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretarse"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scemes"),
                &["schemes", "scenes"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchenged"), &["unchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("embarassig"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dichotomoy"), &["dichotomy"]),
            (
                dictgen::InsensitiveStr::Ascii("diappearance"),
                &["disappearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("tunnells"), &["tunnels"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenatation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wupport"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("streightly"),
                &["straightly"],
            ),
            (dictgen::InsensitiveStr::Ascii("alreay"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("disasbled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("clarificaiton"),
                &["clarification"],
            ),
            (dictgen::InsensitiveStr::Ascii("librabarie"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("recever"),
                &["receiver", "recover"],
            ),
            (dictgen::InsensitiveStr::Ascii("easliy"), &["easily"]),
            (
                dictgen::InsensitiveStr::Ascii("freqquently"),
                &["frequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspecifieced"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("superme"), &["supreme"]),
            (
                dictgen::InsensitiveStr::Ascii("traisivity"),
                &["transitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("categoies"), &["categories"]),
            (
                dictgen::InsensitiveStr::Ascii("rinosaruses"),
                &["rhinoceroses"],
            ),
            (dictgen::InsensitiveStr::Ascii("wantd"), &["wanted"]),
            (dictgen::InsensitiveStr::Ascii("uplaoder"), &["uploader"]),
            (dictgen::InsensitiveStr::Ascii("qeued"), &["queued"]),
            (dictgen::InsensitiveStr::Ascii("smoe"), &["some"]),
            (
                dictgen::InsensitiveStr::Ascii("bootrapers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maching"),
                &["machine", "marching", "matching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubutes"),
                &["distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("publicher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("follww"), &["follow"]),
            (
                dictgen::InsensitiveStr::Ascii("disrispecting"),
                &["disrespecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("broswer"), &["browser"]),
            (
                dictgen::InsensitiveStr::Ascii("autoritharian"),
                &["authoritarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propage"),
                &["propagate", "propane"],
            ),
            (dictgen::InsensitiveStr::Ascii("dsplays"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("manuver"), &["maneuver"]),
            (dictgen::InsensitiveStr::Ascii("instnaces"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("direrctory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("qualit"), &["quality"]),
            (dictgen::InsensitiveStr::Ascii("purchace"), &["purchase"]),
            (dictgen::InsensitiveStr::Ascii("acale"), &["scale"]),
            (
                dictgen::InsensitiveStr::Ascii("unsibscribed"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("misspeld"), &["misspelled"]),
            (
                dictgen::InsensitiveStr::Ascii("acompanied"),
                &["accompanied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcstion"),
                &["function", "functions"],
            ),
            (dictgen::InsensitiveStr::Ascii("repalced"), &["replaced"]),
            (
                dictgen::InsensitiveStr::Ascii("substancial"),
                &["substantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("minize"), &["minimize"]),
            (
                dictgen::InsensitiveStr::Ascii("firends"),
                &["friends", "fiends"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininizes"), &["minimizes"]),
            (dictgen::InsensitiveStr::Ascii("guitares"), &["guitars"]),
            (
                dictgen::InsensitiveStr::Ascii("uncontrallably"),
                &["uncontrollably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requierements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("momoment"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("expiditions"),
                &["expeditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("lobbyistes"), &["lobbyists"]),
            (
                dictgen::InsensitiveStr::Ascii("minniapolis"),
                &["minneapolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("materiales"), &["materials"]),
            (dictgen::InsensitiveStr::Ascii("versitale"), &["versatile"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnurabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envirnment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resposibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("varriances"), &["variances"]),
            (
                dictgen::InsensitiveStr::Ascii("narciscist"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sestatusbar"),
                &["setstatusbar"],
            ),
            (dictgen::InsensitiveStr::Ascii("lileral"), &["literal"]),
            (
                dictgen::InsensitiveStr::Ascii("signifiying"),
                &["signifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("awkwardsness"),
                &["awkwardness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighrouring"),
                &["neighbouring"],
            ),
            (dictgen::InsensitiveStr::Ascii("groupped"), &["grouped"]),
            (
                dictgen::InsensitiveStr::Ascii("unconciousness"),
                &["unconsciousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ehen"),
                &["when", "hen", "even", "then"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocurrred"), &["occurred"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("coleagues"), &["colleagues"]),
            (dictgen::InsensitiveStr::Ascii("seqeunce"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("severly"), &["severely"]),
            (
                dictgen::InsensitiveStr::Ascii("confurm"),
                &["confirm", "conform"],
            ),
            (dictgen::InsensitiveStr::Ascii("alchmeist"), &["alchemist"]),
            (
                dictgen::InsensitiveStr::Ascii("constriants"),
                &["constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanles"), &["handles"]),
            (dictgen::InsensitiveStr::Ascii("intiative"), &["initiative"]),
            (dictgen::InsensitiveStr::Ascii("procuct"), &["product"]),
            (dictgen::InsensitiveStr::Ascii("satisies"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("envolutionary"),
                &["evolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("silluete"), &["silhouette"]),
            (dictgen::InsensitiveStr::Ascii("cotacting"), &["contacting"]),
            (dictgen::InsensitiveStr::Ascii("librray"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("developmentwise"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hospitallity"),
                &["hospitality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indiviually"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecifiers"),
                &["specifiers", "pacifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("ditance"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("misogenistic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefinitley"),
                &["indefinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("shpere"), &["sphere"]),
            (dictgen::InsensitiveStr::Ascii("buddishts"), &["buddhists"]),
            (dictgen::InsensitiveStr::Ascii("rapidally"), &["rapidly"]),
            (dictgen::InsensitiveStr::Ascii("conveyd"), &["conveyed"]),
            (dictgen::InsensitiveStr::Ascii("accapter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("minimilist"),
                &["minimalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inheretance"),
                &["inheritance"],
            ),
            (dictgen::InsensitiveStr::Ascii("confucing"), &["confusing"]),
            (dictgen::InsensitiveStr::Ascii("tolarable"), &["tolerable"]),
            (dictgen::InsensitiveStr::Ascii("magasine"), &["magazine"]),
            (
                dictgen::InsensitiveStr::Ascii("irreversiable"),
                &["irreversible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camoufalge"),
                &["camouflage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installating"),
                &["installation", "installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("somhowe"), &["somehow"]),
            (
                dictgen::InsensitiveStr::Ascii("cirruculum"),
                &["curriculum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passionnate"),
                &["passionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("gererate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhelemd"),
                &["overwhelmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priort"),
                &["prior", "priory", "priors", "priori"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammused"), &["amused"]),
            (dictgen::InsensitiveStr::Ascii("legth"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("authenitcating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("toloerance"), &["tolerance"]),
            (
                dictgen::InsensitiveStr::Ascii("accompained"),
                &["accompanied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialiazes"),
                &["serializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suplements"),
                &["supplements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consideratie"),
                &["considerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("kyeboshes"), &["kiboshes"]),
            (
                dictgen::InsensitiveStr::Ascii("happing"),
                &["happening", "happen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diagronalization"),
                &["diagonalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirumstance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusomized"), &["customized"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiazed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("balaster"), &["baluster"]),
            (
                dictgen::InsensitiveStr::Ascii("hystarically"),
                &["hysterically"],
            ),
            (dictgen::InsensitiveStr::Ascii("arborts"), &["aborts"]),
            (dictgen::InsensitiveStr::Ascii("userame"), &["username"]),
            (
                dictgen::InsensitiveStr::Ascii("underlaying"),
                &["underlying"],
            ),
            (dictgen::InsensitiveStr::Ascii("subobjecs"), &["subobjects"]),
            (dictgen::InsensitiveStr::Ascii("barbedos"), &["barbados"]),
            (
                dictgen::InsensitiveStr::Ascii("catastrohpic"),
                &["catastrophic"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggestted"), &["suggested"]),
            (
                dictgen::InsensitiveStr::Ascii("aggressivo"),
                &["aggression", "aggressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderd"), &["rendered"]),
            (dictgen::InsensitiveStr::Ascii("annoator"), &["annotator"]),
            (
                dictgen::InsensitiveStr::Ascii("aproximated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcot"), &["forgot"]),
            (dictgen::InsensitiveStr::Ascii("chahging"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("seccesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("acccession"), &["accession"]),
            (
                dictgen::InsensitiveStr::Ascii("undertstanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("lesbianus"), &["lesbians"]),
            (dictgen::InsensitiveStr::Ascii("weigths"), &["weights"]),
            (
                dictgen::InsensitiveStr::Ascii("experimtened"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("desginers"), &["designers"]),
            (dictgen::InsensitiveStr::Ascii("hellifre"), &["hellfire"]),
            (dictgen::InsensitiveStr::Ascii("dinosaurus"), &["dinosaurs"]),
            (dictgen::InsensitiveStr::Ascii("somaila"), &["somalia"]),
            (
                dictgen::InsensitiveStr::Ascii("recipts"),
                &["receipts", "recipes"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanege"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("unresponive"),
                &["unresponsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acomodation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrialzed"),
                &["industrialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("balacers"), &["balancers"]),
            (dictgen::InsensitiveStr::Ascii("obviouly"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("nasturshums"),
                &["nasturtiums"],
            ),
            (dictgen::InsensitiveStr::Ascii("emiitted"), &["emitted"]),
            (dictgen::InsensitiveStr::Ascii("generetor"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("leggacy"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpectendly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("tsream"), &["stream"]),
            (
                dictgen::InsensitiveStr::Ascii("customisatoins"),
                &["customisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressolvers"), &["resolvers"]),
            (dictgen::InsensitiveStr::Ascii("soluable"), &["soluble"]),
            (
                dictgen::InsensitiveStr::Ascii("sometrings"),
                &["somethings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesystyems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("normalied"), &["normalized"]),
            (
                dictgen::InsensitiveStr::Ascii("invincinble"),
                &["invincible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reassocations"),
                &["reassociations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sneeks"), &["sneaks"]),
            (dictgen::InsensitiveStr::Ascii("deliverd"), &["delivered"]),
            (dictgen::InsensitiveStr::Ascii("lmits"), &["limits"]),
            (
                dictgen::InsensitiveStr::Ascii("flexibillity"),
                &["flexibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("summmarise"), &["summarise"]),
            (dictgen::InsensitiveStr::Ascii("manetain"), &["maintain"]),
            (dictgen::InsensitiveStr::Ascii("definion"), &["definition"]),
            (dictgen::InsensitiveStr::Ascii("mathced"), &["matched"]),
            (dictgen::InsensitiveStr::Ascii("concentic"), &["concentric"]),
            (
                dictgen::InsensitiveStr::Ascii("untrasformed"),
                &["untransformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overalpping"),
                &["overlapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("starup"), &["startup"]),
            (dictgen::InsensitiveStr::Ascii("modigied"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("problims"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("denomimator"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaiable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacuter"),
                &["manufacture"],
            ),
            (dictgen::InsensitiveStr::Ascii("utiilties"), &["utilities"]),
            (
                dictgen::InsensitiveStr::Ascii("preapration"),
                &["preparation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vakues"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("bootrappers"),
                &["bootstrappers"],
            ),
            (dictgen::InsensitiveStr::Ascii("fimware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("exteremly"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("snaping"), &["snapping"]),
            (dictgen::InsensitiveStr::Ascii("tpyos"), &["typos"]),
            (dictgen::InsensitiveStr::Ascii("momoments"), &["moments"]),
            (
                dictgen::InsensitiveStr::Ascii("challneger"),
                &["challenger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defianetly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprotionally"),
                &["proportionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sevicing"), &["servicing"]),
            (dictgen::InsensitiveStr::Ascii("probbably"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("coducted"), &["conducted"]),
            (dictgen::InsensitiveStr::Ascii("compunet"), &["compute"]),
            (dictgen::InsensitiveStr::Ascii("uprgades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("captue"), &["capture"]),
            (
                dictgen::InsensitiveStr::Ascii("ambidexterouseness"),
                &["ambidextrousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neconstitutional"),
                &["unconstitutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliks"), &["clicks"]),
            (dictgen::InsensitiveStr::Ascii("abborting"), &["aborting"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticaiton"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ampatheater"),
                &["amphitheater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marshmellow"),
                &["marshmallow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuitor"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmisions"),
                &["transmission", "transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("asertions"), &["assertions"]),
            (
                dictgen::InsensitiveStr::Ascii("invidivual"),
                &["individual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lingusitics"),
                &["linguistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generatr"),
                &["generator", "generate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pattens"),
                &["patterns", "patents"],
            ),
            (dictgen::InsensitiveStr::Ascii("excitin"), &["exciting"]),
            (
                dictgen::InsensitiveStr::Ascii("fabricatin"),
                &["fabricating", "fabrication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrinsecally"),
                &["intrinsically"],
            ),
            (dictgen::InsensitiveStr::Ascii("senarreo"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("repitively"),
                &["repetitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orientacion"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dilpoma"), &["diploma"]),
            (
                dictgen::InsensitiveStr::Ascii("itnerpretaton"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bullest"), &["bullets"]),
            (dictgen::InsensitiveStr::Ascii("coordinte"), &["coordinate"]),
            (
                dictgen::InsensitiveStr::Ascii("descriping"),
                &["describing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eavesdroppin"),
                &["eavesdropping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meenings"),
                &["meanings", "meetings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibilties"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("coyrights"), &["copyrights"]),
            (
                dictgen::InsensitiveStr::Ascii("atomatic"),
                &["automatic", "atomic", "aromatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("receets"), &["receipts"]),
            (
                dictgen::InsensitiveStr::Ascii("civalezations"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("rapsadies"), &["rhapsodies"]),
            (
                dictgen::InsensitiveStr::Ascii("relesase"),
                &["release", "releases"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasive"), &["passive"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("coverters"), &["converters"]),
            (
                dictgen::InsensitiveStr::Ascii("permanenty"),
                &["permanently"],
            ),
            (dictgen::InsensitiveStr::Ascii("gonewidl"), &["gonewild"]),
            (dictgen::InsensitiveStr::Ascii("owernship"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("raelism"), &["realism"]),
            (dictgen::InsensitiveStr::Ascii("secretos"), &["secrets"]),
            (dictgen::InsensitiveStr::Ascii("israelitas"), &["israelis"]),
            (dictgen::InsensitiveStr::Ascii("basiclly"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("enviornmental"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programattic"),
                &["programmatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outnunbered"),
                &["outnumbered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unequalities"),
                &["inequalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectaculalry"),
                &["spectacularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("stangely"), &["strangely"]),
            (dictgen::InsensitiveStr::Ascii("perfoems"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("prorotype"), &["prototype"]),
            (dictgen::InsensitiveStr::Ascii("fillowing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("resitors"), &["resistors"]),
            (
                dictgen::InsensitiveStr::Ascii("standrdization"),
                &["standardization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistakinly"),
                &["mistakenly"],
            ),
            (dictgen::InsensitiveStr::Ascii("alliviate"), &["alleviate"]),
            (dictgen::InsensitiveStr::Ascii("forwaded"), &["forwarded"]),
            (
                dictgen::InsensitiveStr::Ascii("mircoscope"),
                &["microscope"],
            ),
            (dictgen::InsensitiveStr::Ascii("invisilbe"), &["invisible"]),
            (dictgen::InsensitiveStr::Ascii("kweyword"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("behaviuor"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("autoneogotiation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("whe"), &["when", "we"]),
            (dictgen::InsensitiveStr::Ascii("bollocs"), &["bollocks"]),
            (dictgen::InsensitiveStr::Ascii("seesion"), &["session"]),
            (dictgen::InsensitiveStr::Ascii("metadat"), &["metadata"]),
            (
                dictgen::InsensitiveStr::Ascii("experimal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoultes"), &["absolutes"]),
            (dictgen::InsensitiveStr::Ascii("concieved"), &["conceived"]),
            (
                dictgen::InsensitiveStr::Ascii("custumizer"),
                &["customizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplissed"),
                &["accomplished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefencing"),
                &["dereferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libretarian"),
                &["libertarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("synoynm"), &["synonym"]),
            (dictgen::InsensitiveStr::Ascii("kitches"), &["kitchens"]),
            (dictgen::InsensitiveStr::Ascii("percevied"), &["perceived"]),
            (dictgen::InsensitiveStr::Ascii("clampling"), &["clamping"]),
            (dictgen::InsensitiveStr::Ascii("interrups"), &["interrupts"]),
            (
                dictgen::InsensitiveStr::Ascii("favuourites"),
                &["favourites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perparations"),
                &["preparations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initmately"),
                &["intimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("alligns"), &["aligns"]),
            (dictgen::InsensitiveStr::Ascii("palyback"), &["playback"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappointments"),
                &["disappointments"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlytics"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("mak"), &["make", "mask"]),
            (dictgen::InsensitiveStr::Ascii("duoble"), &["double"]),
            (dictgen::InsensitiveStr::Ascii("picoseond"), &["picosecond"]),
            (
                dictgen::InsensitiveStr::Ascii("incoporating"),
                &["incorporating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("othogonally"),
                &["orthogonally"],
            ),
            (dictgen::InsensitiveStr::Ascii("ditial"), &["digital"]),
            (dictgen::InsensitiveStr::Ascii("decoratie"), &["decorative"]),
            (
                dictgen::InsensitiveStr::Ascii("fellowshop"),
                &["fellowship"],
            ),
            (dictgen::InsensitiveStr::Ascii("extensis"), &["extensions"]),
            (dictgen::InsensitiveStr::Ascii("cleanned"), &["cleaned"]),
            (dictgen::InsensitiveStr::Ascii("tyranies"), &["tyrannies"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminarti"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitcally"),
                &["identically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivisuality"),
                &["individuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaialbe"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("folwoers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("diffrent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("coordindates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("americs"), &["americas"]),
            (dictgen::InsensitiveStr::Ascii("gilty"), &["guilty"]),
            (dictgen::InsensitiveStr::Ascii("overrids"), &["overrides"]),
            (dictgen::InsensitiveStr::Ascii("confursed"), &["confused"]),
            (
                dictgen::InsensitiveStr::Ascii("pornograpgy"),
                &["pornography"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrue"), &["thru"]),
            (dictgen::InsensitiveStr::Ascii("accepors"), &["acceptors"]),
            (dictgen::InsensitiveStr::Ascii("avantages"), &["advantages"]),
            (dictgen::InsensitiveStr::Ascii("priview"), &["preview"]),
            (
                dictgen::InsensitiveStr::Ascii("sucseptible"),
                &["susceptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assitsance"),
                &["assistance"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuplods"), &["reuploads"]),
            (
                dictgen::InsensitiveStr::Ascii("californnia"),
                &["californian"],
            ),
            (dictgen::InsensitiveStr::Ascii("channets"), &["channels"]),
            (dictgen::InsensitiveStr::Ascii("opoen"), &["open"]),
            (
                dictgen::InsensitiveStr::Ascii("currespondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unhilight"),
                &["unhighlight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("smarthpone"),
                &["smartphone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrinsisch"),
                &["intrinsic"],
            ),
            (dictgen::InsensitiveStr::Ascii("comress"), &["compress"]),
            (
                dictgen::InsensitiveStr::Ascii("ehancements"),
                &["enhancements"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwhere"), &["anywhere"]),
            (dictgen::InsensitiveStr::Ascii("enumearate"), &["enumerate"]),
            (
                dictgen::InsensitiveStr::Ascii("deficience"),
                &["deficiencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniitals"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("familly"), &["family"]),
            (
                dictgen::InsensitiveStr::Ascii("architectuur"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("wahtever"), &["whatever"]),
            (dictgen::InsensitiveStr::Ascii("connetc"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("deamonising"),
                &["daemonising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rondazyvoo"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("circit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("foriner"), &["foreigner"]),
            (dictgen::InsensitiveStr::Ascii("tournes"), &["tourneys"]),
            (
                dictgen::InsensitiveStr::Ascii("opimisations"),
                &["optimisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journolist"),
                &["journalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("smetime"), &["sometime"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseface"),
                &["cheesecake"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surpression"),
                &["suppression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antogonistic"),
                &["antagonistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrecongized"),
                &["unrecognized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preaparations"),
                &["preparations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vizuals"), &["visuals"]),
            (dictgen::InsensitiveStr::Ascii("anwsering"), &["answering"]),
            (dictgen::InsensitiveStr::Ascii("intputs"), &["inputs"]),
            (dictgen::InsensitiveStr::Ascii("prooses"), &["proposes"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarchichal"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("brewerys"), &["brewers"]),
            (dictgen::InsensitiveStr::Ascii("cimptoms"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("finisse"), &["finishes"]),
            (
                dictgen::InsensitiveStr::Ascii("assiocations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undermiming"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplacate"),
                &["deduplicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("montains"), &["mountains"]),
            (dictgen::InsensitiveStr::Ascii("implemnts"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("themselfes"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarunted"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("hempishere"),
                &["hemisphere"],
            ),
            (dictgen::InsensitiveStr::Ascii("someties"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("modfieis"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("distribiton"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rellocates"),
                &["reallocates", "relocates"],
            ),
            (dictgen::InsensitiveStr::Ascii("trinitiy"), &["trinity"]),
            (dictgen::InsensitiveStr::Ascii("langerie"), &["lingerie"]),
            (
                dictgen::InsensitiveStr::Ascii("coincidente"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapoints"),
                &["disappoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondeing"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbaly"), &["verbally"]),
            (dictgen::InsensitiveStr::Ascii("phantoom"), &["phantom"]),
            (
                dictgen::InsensitiveStr::Ascii("permuataions"),
                &["permutations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symapthetic"),
                &["sympathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perturbe"),
                &["perturb", "perturbed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("freqiencies"),
                &["frequencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memember"),
                &["member", "remember"],
            ),
            (dictgen::InsensitiveStr::Ascii("ome"), &["some"]),
            (
                dictgen::InsensitiveStr::Ascii("unfinsihed"),
                &["unfinished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturucts"), &["structs"]),
            (dictgen::InsensitiveStr::Ascii("channle"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("rythemic"), &["rhythmic"]),
            (dictgen::InsensitiveStr::Ascii("sacalars"), &["scalars"]),
            (
                dictgen::InsensitiveStr::Ascii("ddresses"),
                &["addresses", "dresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("apporves"), &["approves"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowleging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizating"),
                &["organizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplfies"), &["simplifies"]),
            (
                dictgen::InsensitiveStr::Ascii("akkumulators"),
                &["accumulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("metrig"), &["metric"]),
            (
                dictgen::InsensitiveStr::Ascii("challengur"),
                &["challenger"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfors"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("redeisgn"), &["redesign"]),
            (
                dictgen::InsensitiveStr::Ascii("theraputic"),
                &["therapeutic"],
            ),
            (dictgen::InsensitiveStr::Ascii("fxed"), &["fixed"]),
            (dictgen::InsensitiveStr::Ascii("ontacting"), &["contacting"]),
            (
                dictgen::InsensitiveStr::Ascii("erronously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceutables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkclists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpectely"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpertations"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invisibillity"),
                &["invisibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragrph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("sequnced"), &["sequenced"]),
            (dictgen::InsensitiveStr::Ascii("mdoeled"), &["modeled"]),
            (
                dictgen::InsensitiveStr::Ascii("alllocations"),
                &["allocations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramalizing"),
                &["normalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenfify"), &["identify"]),
            (
                dictgen::InsensitiveStr::Ascii("birmingharam"),
                &["birmingham"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coururier"),
                &["courier", "couturier"],
            ),
            (dictgen::InsensitiveStr::Ascii("inluded"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("strictier"), &["stricter"]),
            (dictgen::InsensitiveStr::Ascii("famoulsy"), &["famously"]),
            (dictgen::InsensitiveStr::Ascii("evrybody"), &["everybody"]),
            (dictgen::InsensitiveStr::Ascii("multible"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("gwava"), &["guava"]),
            (
                dictgen::InsensitiveStr::Ascii("natioanlly"),
                &["nationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traingulates"),
                &["triangulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrinseci"), &["intrinsic"]),
            (dictgen::InsensitiveStr::Ascii("exceded"), &["exceeded"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortuantely"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("recieved"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("benges"), &["binges"]),
            (
                dictgen::InsensitiveStr::Ascii("expropriatin"),
                &["expropriating", "expropriation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udnerstand"),
                &["understand"],
            ),
            (dictgen::InsensitiveStr::Ascii("anbd"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("reuquest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("reconected"),
                &["reconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("librabry"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("treatmens"), &["treatments"]),
            (dictgen::InsensitiveStr::Ascii("mitake"), &["mistake"]),
            (
                dictgen::InsensitiveStr::Ascii("clammer"),
                &["clamber", "clamor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreamming"),
                &["upstreaming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geomteries"),
                &["geometries"],
            ),
            (dictgen::InsensitiveStr::Ascii("ibuprofein"), &["ibuprofen"]),
            (
                dictgen::InsensitiveStr::Ascii("compairisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurrin"), &["recurring"]),
            (dictgen::InsensitiveStr::Ascii("scrolable"), &["scrollable"]),
            (dictgen::InsensitiveStr::Ascii("perferrm"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("correcxt"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("aposltes"), &["apostles"]),
            (
                dictgen::InsensitiveStr::Ascii("pocession"),
                &["procession", "possession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disspaointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unenforcable"),
                &["unenforceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theimselves"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("somestimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("easilly"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("adaquit"), &["adequate"]),
            (dictgen::InsensitiveStr::Ascii("validte"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("systematisin"),
                &["systematising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inherert"),
                &["inherit", "inherent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufatured"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("borwsers"), &["browsers"]),
            (dictgen::InsensitiveStr::Ascii("zeors"), &["zeros"]),
            (dictgen::InsensitiveStr::Ascii("simiulated"), &["simulated"]),
            (dictgen::InsensitiveStr::Ascii("othewize"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("separetion"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpresd"), &["compressed"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstates"),
                &["demonstrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("assebly"), &["assembly"]),
            (
                dictgen::InsensitiveStr::Ascii("algotrithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjustabe"), &["adjustable"]),
            (dictgen::InsensitiveStr::Ascii("englishs"), &["english"]),
            (
                dictgen::InsensitiveStr::Ascii("redundency"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfified"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplation"),
                &["deduplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("torandoes"), &["tornadoes"]),
            (
                dictgen::InsensitiveStr::Ascii("assiocation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinador"),
                &["coordinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algotrithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtractor"),
                &["abstractor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omnipitent"),
                &["omnipotent"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynamincs"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("nocture"), &["nocturne"]),
            (
                dictgen::InsensitiveStr::Ascii("segrigated"),
                &["segregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraterrestials"),
                &["extraterrestrials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minipulation"),
                &["manipulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissmising"),
                &["dismissing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performences"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantiator"),
                &["differentiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenominal"),
                &["phenomenal"],
            ),
            (dictgen::InsensitiveStr::Ascii("implantes"), &["implants"]),
            (
                dictgen::InsensitiveStr::Ascii("excecting"),
                &["executing", "expecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deinitalize"),
                &["deinitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternativly"),
                &["alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abandonne"),
                &["abandonment", "abandon"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheslea"), &["chelsea"]),
            (
                dictgen::InsensitiveStr::Ascii("nonsenscial"),
                &["nonsensical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionallities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("scneario"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("vitualising"),
                &["virtualising", "visualising"],
            ),
            (dictgen::InsensitiveStr::Ascii("licesed"), &["licensed"]),
            (dictgen::InsensitiveStr::Ascii("execuable"), &["executable"]),
            (
                dictgen::InsensitiveStr::Ascii("staisfying"),
                &["satisfying"],
            ),
            (dictgen::InsensitiveStr::Ascii("mornig"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("possesive"), &["possessive"]),
            (
                dictgen::InsensitiveStr::Ascii("concatentaion"),
                &["concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essentialls"),
                &["essentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categogically"),
                &["categorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exposese"), &["exposes"]),
            (dictgen::InsensitiveStr::Ascii("develper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("esponiage"), &["espionage"]),
            (
                dictgen::InsensitiveStr::Ascii("derstruion"),
                &["destruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbance"),
                &["disturbance"],
            ),
            (dictgen::InsensitiveStr::Ascii("insallers"), &["installers"]),
            (dictgen::InsensitiveStr::Ascii("peaseful"), &["peaceful"]),
            (dictgen::InsensitiveStr::Ascii("regsister"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("editoro"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("napomms"), &["napalms"]),
            (
                dictgen::InsensitiveStr::Ascii("skillhosts"),
                &["skillshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harrasments"),
                &["harassments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnsoticism"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("decidates"), &["dedicates"]),
            (
                dictgen::InsensitiveStr::Ascii("backgrounts"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuting"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualtions"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstracor"), &["abstractor"]),
            (dictgen::InsensitiveStr::Ascii("recpetors"), &["receptors"]),
            (
                dictgen::InsensitiveStr::Ascii("reprehenisble"),
                &["reprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedemic"), &["academic"]),
            (dictgen::InsensitiveStr::Ascii("flacoring"), &["flavoring"]),
            (dictgen::InsensitiveStr::Ascii("whisperd"), &["whispered"]),
            (dictgen::InsensitiveStr::Ascii("eveywhere"), &["everywhere"]),
            (dictgen::InsensitiveStr::Ascii("aserts"), &["asserts"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilitie"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constistently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("voluntears"),
                &["volunteers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conpressed"),
                &["compressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("centriod"), &["centroid"]),
            (dictgen::InsensitiveStr::Ascii("concesso"), &["concession"]),
            (dictgen::InsensitiveStr::Ascii("paralel"), &["parallel"]),
            (
                dictgen::InsensitiveStr::Ascii("missmatched"),
                &["mismatched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequistite"),
                &["prerequisite"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwais"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("appreceating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplyng"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("dploying"), &["deploying"]),
            (dictgen::InsensitiveStr::Ascii("surbuban"), &["suburban"]),
            (
                dictgen::InsensitiveStr::Ascii("encompasess"),
                &["encompasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirgation"), &["migration"]),
            (
                dictgen::InsensitiveStr::Ascii("assications"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappearnce"),
                &["disappearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportionnal"),
                &["proportional"],
            ),
            (dictgen::InsensitiveStr::Ascii("naturels"), &["natures"]),
            (
                dictgen::InsensitiveStr::Ascii("parrallelizes"),
                &["parallelizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncluded"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("supplimental"),
                &["supplemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particlars"),
                &["particulars"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyeballers"), &["eyeballs"]),
            (
                dictgen::InsensitiveStr::Ascii("convensions"),
                &["conventions", "conversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("plalform"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("suggetss"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("communiyt"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("uninstlal"), &["uninstall"]),
            (dictgen::InsensitiveStr::Ascii("edbiles"), &["edibles"]),
            (dictgen::InsensitiveStr::Ascii("listeral"), &["literal"]),
            (dictgen::InsensitiveStr::Ascii("arrrays"), &["arrays"]),
            (
                dictgen::InsensitiveStr::Ascii("excewptional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegoziations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dostribution"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchanidse"),
                &["merchandise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sizor"),
                &["scissor", "sizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alings"),
                &["alines", "slings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentationalist"),
                &["sensationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("featchuring"),
                &["featuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleetion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filght"),
                &["flight", "fight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternater"),
                &["alternator"],
            ),
            (dictgen::InsensitiveStr::Ascii("decieves"), &["deceives"]),
            (dictgen::InsensitiveStr::Ascii("ilegally"), &["illegally"]),
            (dictgen::InsensitiveStr::Ascii("aeorspace"), &["aerospace"]),
            (dictgen::InsensitiveStr::Ascii("stoyr"), &["story"]),
            (dictgen::InsensitiveStr::Ascii("signatuire"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("interable"),
                &["iterable", "interactable", "integrable", "intolerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyerarchically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrte"), &["write", "wrote"]),
            (
                dictgen::InsensitiveStr::Ascii("satisfiction"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("inddex"), &["index"]),
            (
                dictgen::InsensitiveStr::Ascii("unconscous"),
                &["unconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("myselves"), &["myself"]),
            (dictgen::InsensitiveStr::Ascii("accesing"), &["accessing"]),
            (dictgen::InsensitiveStr::Ascii("confues"), &["confuses"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimmetric"),
                &["dissymmetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unprofessionel"),
                &["unprofessional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsual"),
                &["unusual", "usual"],
            ),
            (dictgen::InsensitiveStr::Ascii("empathie"), &["empathize"]),
            (
                dictgen::InsensitiveStr::Ascii("conchusly"),
                &["consciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowledegable"),
                &["knowledgeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("judgamental"),
                &["judgemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilaitons"),
                &["compilations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represantiation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("surrouded"), &["surrounded"]),
            (
                dictgen::InsensitiveStr::Ascii("collison"),
                &["collision", "collusion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ttempted"),
                &["attempted", "tempted"],
            ),
            (dictgen::InsensitiveStr::Ascii("teamfigt"), &["teamfight"]),
            (
                dictgen::InsensitiveStr::Ascii("notifycations"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("prophetes"), &["prophets"]),
            (
                dictgen::InsensitiveStr::Ascii("mantainers"),
                &["maintainers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temparally"),
                &["temporally", "temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoded"), &["avoided"]),
            (
                dictgen::InsensitiveStr::Ascii("partisipate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("peacful"), &["peaceful"]),
            (dictgen::InsensitiveStr::Ascii("beghaviors"), &["behaviors"]),
            (dictgen::InsensitiveStr::Ascii("vertify"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("notibly"), &["notably"]),
            (dictgen::InsensitiveStr::Ascii("chatou"), &["château"]),
            (
                dictgen::InsensitiveStr::Ascii("distinciton"),
                &["distinction"],
            ),
            (dictgen::InsensitiveStr::Ascii("grooups"), &["groups"]),
            (
                dictgen::InsensitiveStr::Ascii("misunterstood"),
                &["misunderstood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peformance"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colleauges"),
                &["colleagues"],
            ),
            (dictgen::InsensitiveStr::Ascii("reatime"), &["realtime"]),
            (
                dictgen::InsensitiveStr::Ascii("documentataions"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("imortant"), &["important"]),
            (dictgen::InsensitiveStr::Ascii("melineumms"), &["millennia"]),
            (dictgen::InsensitiveStr::Ascii("abitrate"), &["arbitrate"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotioations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertext"), &["vertex"]),
            (dictgen::InsensitiveStr::Ascii("operte"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("ermington"), &["remington"]),
            (
                dictgen::InsensitiveStr::Ascii("insependent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impotant"),
                &["important", "impotent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocorrences"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("vehicules"), &["vehicles"]),
            (
                dictgen::InsensitiveStr::Ascii("muscils"),
                &["muscles", "mussels"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggetion"), &["suggestion"]),
            (dictgen::InsensitiveStr::Ascii("jamaicia"), &["jamaica"]),
            (
                dictgen::InsensitiveStr::Ascii("wednesdsay"),
                &["wednesdays"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartesion"), &["cartesian"]),
            (dictgen::InsensitiveStr::Ascii("archivd"), &["archived"]),
            (
                dictgen::InsensitiveStr::Ascii("missionare"),
                &["missionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deveploment"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("demamgled"), &["demangled"]),
            (dictgen::InsensitiveStr::Ascii("matchend"), &["matched"]),
            (dictgen::InsensitiveStr::Ascii("intruduce"), &["introduce"]),
            (
                dictgen::InsensitiveStr::Ascii("syntehsizes"),
                &["synthesizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("yonger"), &["younger"]),
            (
                dictgen::InsensitiveStr::Ascii("spontaneouly"),
                &["spontaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequenc"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("elemenet"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("discusion"), &["discussion"]),
            (dictgen::InsensitiveStr::Ascii("finalizin"), &["finalizing"]),
            (
                dictgen::InsensitiveStr::Ascii("migation"),
                &["migration", "mitigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("navgiators"),
                &["navigators"],
            ),
            (dictgen::InsensitiveStr::Ascii("penisnula"), &["peninsula"]),
            (dictgen::InsensitiveStr::Ascii("passerbys"), &["passersby"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutoinary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("uption"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("enevlope"), &["envelope"]),
            (dictgen::InsensitiveStr::Ascii("unduely"), &["unduly"]),
            (
                dictgen::InsensitiveStr::Ascii("syntehsizing"),
                &["synthesizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("availalble"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("datecreatedd"),
                &["datecreated"],
            ),
            (dictgen::InsensitiveStr::Ascii("triology"), &["trilogy"]),
            (dictgen::InsensitiveStr::Ascii("interfers"), &["interferes"]),
            (
                dictgen::InsensitiveStr::Ascii("organziing"),
                &["organizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylzed"), &["analyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("analysators"),
                &["analysers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sitautions"),
                &["situations"],
            ),
            (dictgen::InsensitiveStr::Ascii("blos"), &["blobs"]),
            (dictgen::InsensitiveStr::Ascii("nework"), &["network"]),
            (
                dictgen::InsensitiveStr::Ascii("predicitability"),
                &["predictability"],
            ),
            (dictgen::InsensitiveStr::Ascii("zookeeni"), &["zucchini"]),
            (dictgen::InsensitiveStr::Ascii("subseuent"), &["subsequent"]),
            (
                dictgen::InsensitiveStr::Ascii("volunteeers"),
                &["volunteers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntehsising"),
                &["synthesising"],
            ),
            (dictgen::InsensitiveStr::Ascii("govement"), &["government"]),
            (
                dictgen::InsensitiveStr::Ascii("transformatons"),
                &["transformations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transactiona"),
                &["transactional", "transactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeation"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("surroundes"), &["surrounds"]),
            (dictgen::InsensitiveStr::Ascii("tehsors"), &["tensors"]),
            (dictgen::InsensitiveStr::Ascii("laucnhers"), &["launchers"]),
            (
                dictgen::InsensitiveStr::Ascii("complicted"),
                &["complicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calulations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unproduktive"),
                &["unproductive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactued"),
                &["manufactured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rollarcoaster"),
                &["rollercoaster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performes"),
                &["performed", "performs"],
            ),
            (dictgen::InsensitiveStr::Ascii("aninators"), &["animators"]),
            (
                dictgen::InsensitiveStr::Ascii("unregisted"),
                &["unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpletion"),
                &["completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("advsie"), &["advise"]),
            (
                dictgen::InsensitiveStr::Ascii("recutions"),
                &["reductions", "recursions", "precautions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autherising"),
                &["authorising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowiedge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransactioned"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("comlexity"), &["complexity"]),
            (
                dictgen::InsensitiveStr::Ascii("reconneting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("preiodic"), &["periodic"]),
            (dictgen::InsensitiveStr::Ascii("anonyms"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("termporarly"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegitmate"),
                &["illegitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("palidans"), &["paladins"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithmnic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpont"),
                &["counterpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("propblems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("presistence"),
                &["persistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("krypronite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repositorries"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("squirtel"), &["squirtle"]),
            (
                dictgen::InsensitiveStr::Ascii("appoximations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understnading"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("relasped"), &["relapsed"]),
            (
                dictgen::InsensitiveStr::Ascii("definitiion"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteisation"),
                &["characterisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dircetly"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatve"),
                &["alternate", "alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psuedolayer"),
                &["pseudolayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("alocators"), &["allocators"]),
            (dictgen::InsensitiveStr::Ascii("monarcy"), &["monarchy"]),
            (dictgen::InsensitiveStr::Ascii("boith"), &["both"]),
            (dictgen::InsensitiveStr::Ascii("optioin"), &["option"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplented"),
                &["reimplemented", "reimplanted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espected"),
                &["expected", "respected"],
            ),
            (dictgen::InsensitiveStr::Ascii("compliled"), &["compiled"]),
            (
                dictgen::InsensitiveStr::Ascii("contractar"),
                &["contractor"],
            ),
            (dictgen::InsensitiveStr::Ascii("columbina"), &["columbia"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetning"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adaptove"),
                &["adaptive", "adoptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protistant"),
                &["protestant"],
            ),
            (dictgen::InsensitiveStr::Ascii("onot"), &["note", "not"]),
            (
                dictgen::InsensitiveStr::Ascii("cancele"),
                &["cancel", "cancels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretter"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("trampline"), &["trampoline"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronisations"),
                &["synchronisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("unquotte"), &["unquote"]),
            (
                dictgen::InsensitiveStr::Ascii("incorrently"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimaonds"), &["diamonds"]),
            (
                dictgen::InsensitiveStr::Ascii("expanation"),
                &["explanation", "expansion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afect"),
                &["affect", "effect"],
            ),
            (dictgen::InsensitiveStr::Ascii("moer"), &["more"]),
            (dictgen::InsensitiveStr::Ascii("cheif"), &["chief"]),
            (dictgen::InsensitiveStr::Ascii("aer"), &["are"]),
            (dictgen::InsensitiveStr::Ascii("unfiform"), &["uniform"]),
            (dictgen::InsensitiveStr::Ascii("behavio"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("lanuage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("comparres"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("recusion"),
                &["recursion", "reclusion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impleemntation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posicional"),
                &["positional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maitaining"),
                &["maintaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisifaction"),
                &["satisfaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacturerers"),
                &["manufacturers"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedons"), &["freedoms"]),
            (dictgen::InsensitiveStr::Ascii("psoition"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("congradulation"),
                &["congratulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capters"),
                &["captors", "chapters"],
            ),
            (dictgen::InsensitiveStr::Ascii("possobile"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("promted"),
                &["prompted", "promoted"],
            ),
            (dictgen::InsensitiveStr::Ascii("staticts"), &["statist"]),
            (
                dictgen::InsensitiveStr::Ascii("pertrubation"),
                &["perturbation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsidizied"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pecentage"), &["percentage"]),
            (dictgen::InsensitiveStr::Ascii("zimbabwae"), &["zimbabwe"]),
            (
                dictgen::InsensitiveStr::Ascii("penisylvania"),
                &["pennsylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("empy"), &["empty"]),
            (dictgen::InsensitiveStr::Ascii("archivrs"), &["archivers"]),
            (
                dictgen::InsensitiveStr::Ascii("orgiginally"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimal"), &["optimal"]),
            (
                dictgen::InsensitiveStr::Ascii("systemetic"),
                &["systematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("manuelly"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoors"),
                &["neighbors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hightlighted"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyllinder"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("stuggled"), &["struggled"]),
            (dictgen::InsensitiveStr::Ascii("mememory"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("ansd"), &["and", "ansa"]),
            (dictgen::InsensitiveStr::Ascii("synopis"), &["synopsis"]),
            (dictgen::InsensitiveStr::Ascii("akward"), &["awkward"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasembled"),
                &["disassembled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occupaiton"),
                &["occupation"],
            ),
            (dictgen::InsensitiveStr::Ascii("introvet"), &["introvert"]),
            (dictgen::InsensitiveStr::Ascii("reminsces"), &["reminisces"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("conscent"), &["consent"]),
            (dictgen::InsensitiveStr::Ascii("sugned"), &["signed"]),
            (
                dictgen::InsensitiveStr::Ascii("spreasheet"),
                &["spreadsheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repalcement"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manegement"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("elsehwere"), &["elsewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("simplisitically"),
                &["simplistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("falgs"), &["flags"]),
            (
                dictgen::InsensitiveStr::Ascii("succefull"),
                &["successful", "successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clibpoards"),
                &["clipboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genreation"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chocolateers"),
                &["chocolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("utitlities"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("encrpyts"), &["encrypts"]),
            (dictgen::InsensitiveStr::Ascii("permssion"), &["permission"]),
            (
                dictgen::InsensitiveStr::Ascii("nastertium"),
                &["nasturtium"],
            ),
            (dictgen::InsensitiveStr::Ascii("maininly"), &["mainly"]),
            (
                dictgen::InsensitiveStr::Ascii("tecnically"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("covnerts"), &["converts"]),
            (
                dictgen::InsensitiveStr::Ascii("incomptible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irresponsbile"),
                &["irresponsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("microphen"), &["microphone"]),
            (
                dictgen::InsensitiveStr::Ascii("intorduced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posibilty"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("qualifid"), &["qualified"]),
            (dictgen::InsensitiveStr::Ascii("tremelos"), &["tremolos"]),
            (dictgen::InsensitiveStr::Ascii("alocate"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("partirions"),
                &["partitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iresistible"),
                &["irresistible"],
            ),
            (dictgen::InsensitiveStr::Ascii("everyteim"), &["everytime"]),
            (
                dictgen::InsensitiveStr::Ascii("necssarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electorale"),
                &["electorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialice"),
                &["specialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anniversay"),
                &["anniversary"],
            ),
            (dictgen::InsensitiveStr::Ascii("supuser"), &["superuser"]),
            (
                dictgen::InsensitiveStr::Ascii("charactetistics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("livelehood"),
                &["livelihood"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcumented"), &["documented"]),
            (dictgen::InsensitiveStr::Ascii("inersia"), &["inertia"]),
            (
                dictgen::InsensitiveStr::Ascii("attchments"),
                &["attachments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatenaed"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulatorios"),
                &["regulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("movemnet"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("alcoholsim"),
                &["alcoholism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequecences"),
                &["sequences"],
            ),
            (dictgen::InsensitiveStr::Ascii("wholely"), &["wholly"]),
            (
                dictgen::InsensitiveStr::Ascii("uncersored"),
                &["uncensored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomatically"),
                &["automatically", "atomically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagerized"),
                &["plagiarized"],
            ),
            (dictgen::InsensitiveStr::Ascii("phlematic"), &["phlegmatic"]),
            (dictgen::InsensitiveStr::Ascii("coloardo"), &["colorado"]),
            (dictgen::InsensitiveStr::Ascii("knwon"), &["known"]),
            (dictgen::InsensitiveStr::Ascii("rutime"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("otpimistically"),
                &["optimistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("systematisch"),
                &["systematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionare"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("turqoise"), &["turquoise"]),
            (dictgen::InsensitiveStr::Ascii("remotelly"), &["remotely"]),
            (dictgen::InsensitiveStr::Ascii("wrteched"), &["wretched"]),
            (
                dictgen::InsensitiveStr::Ascii("downloding"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peferential"),
                &["preferential"],
            ),
            (dictgen::InsensitiveStr::Ascii("reussing"), &["reusing"]),
            (dictgen::InsensitiveStr::Ascii("dimensiom"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("dissarray"), &["disarray"]),
            (
                dictgen::InsensitiveStr::Ascii("summmarising"),
                &["summarising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rengenerate"),
                &["regenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lozonya"), &["lasagna"]),
            (dictgen::InsensitiveStr::Ascii("inifnity"), &["infinity"]),
            (dictgen::InsensitiveStr::Ascii("attrbuted"), &["attributed"]),
            (dictgen::InsensitiveStr::Ascii("undisered"), &["undesired"]),
            (
                dictgen::InsensitiveStr::Ascii("exclussive"),
                &["exclusives"],
            ),
            (dictgen::InsensitiveStr::Ascii("descize"), &["disguise"]),
            (
                dictgen::InsensitiveStr::Ascii("chowse"),
                &["choose", "chose"],
            ),
            (dictgen::InsensitiveStr::Ascii("detils"), &["details"]),
            (
                dictgen::InsensitiveStr::Ascii("confrontato"),
                &["confrontation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inagurated"),
                &["inaugurated"],
            ),
            (dictgen::InsensitiveStr::Ascii("aheared"), &["adhered"]),
            (
                dictgen::InsensitiveStr::Ascii("identitets"),
                &["identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("insurnace"), &["insurance"]),
            (
                dictgen::InsensitiveStr::Ascii("omnipetent"),
                &["omnipotent"],
            ),
            (dictgen::InsensitiveStr::Ascii("extesions"), &["extensions"]),
            (dictgen::InsensitiveStr::Ascii("dimensins"), &["dimensions"]),
            (dictgen::InsensitiveStr::Ascii("palyed"), &["played"]),
            (
                dictgen::InsensitiveStr::Ascii("vistor"),
                &["visitor", "visor", "victor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefifation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("listners"), &["listeners"]),
            (
                dictgen::InsensitiveStr::Ascii("exictement"),
                &["excitement"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheduing"), &["scheduling"]),
            (
                dictgen::InsensitiveStr::Ascii("programmin"),
                &["programming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("packt"),
                &["packet", "packed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliently"), &["silently"]),
            (
                dictgen::InsensitiveStr::Ascii("negotioating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nimphomaniacs"),
                &["nymphomaniacs"],
            ),
            (dictgen::InsensitiveStr::Ascii("pactched"), &["patched"]),
            (dictgen::InsensitiveStr::Ascii("processses"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("disapparing"),
                &["disappearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwwinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("embold"), &["embolden"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("memoery"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("wokrflows"), &["workflows"]),
            (
                dictgen::InsensitiveStr::Ascii("maitainance"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fiture"),
                &["feature", "future", "figure", "fixture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retcieves"),
                &["retrieves", "receives"],
            ),
            (dictgen::InsensitiveStr::Ascii("hadnled"), &["handled"]),
            (
                dictgen::InsensitiveStr::Ascii("diminushing"),
                &["diminishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substituto"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discretiation"),
                &["discretization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cluthcing"), &["clutching"]),
            (dictgen::InsensitiveStr::Ascii("suface"), &["surface"]),
            (dictgen::InsensitiveStr::Ascii("hopeing"), &["hoping"]),
            (dictgen::InsensitiveStr::Ascii("funcitons"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("skteches"), &["sketches"]),
            (dictgen::InsensitiveStr::Ascii("linaerly"), &["linearly"]),
            (dictgen::InsensitiveStr::Ascii("mininises"), &["minimises"]),
            (
                dictgen::InsensitiveStr::Ascii("functionnaly"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("defatult"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("srinked"), &["shrunk"]),
            (dictgen::InsensitiveStr::Ascii("chckpoint"), &["checkpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("corospondants"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encruption"),
                &["encryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coneninece"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("customied"), &["customized"]),
            (
                dictgen::InsensitiveStr::Ascii("cusomizers"),
                &["customizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("glyserin"), &["glycerin"]),
            (dictgen::InsensitiveStr::Ascii("uppon"), &["upon"]),
            (
                dictgen::InsensitiveStr::Ascii("correlaties"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("bahavior"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("prosectuion"),
                &["prosecution"],
            ),
            (dictgen::InsensitiveStr::Ascii("landingers"), &["landings"]),
            (dictgen::InsensitiveStr::Ascii("fwe"), &["few"]),
            (dictgen::InsensitiveStr::Ascii("ciriticals"), &["criticals"]),
            (dictgen::InsensitiveStr::Ascii("analyzin"), &["analyzing"]),
            (dictgen::InsensitiveStr::Ascii("inititate"), &["initiate"]),
            (dictgen::InsensitiveStr::Ascii("upated"), &["updated"]),
            (
                dictgen::InsensitiveStr::Ascii("approimate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("violatin"),
                &["violation", "violating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermarkeds"),
                &["supermarkets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intestinos"),
                &["intestines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchrounous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoadmap"), &["roadmap"]),
            (
                dictgen::InsensitiveStr::Ascii("addministratively"),
                &["administratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("maradeur"), &["marauder"]),
            (
                dictgen::InsensitiveStr::Ascii("convulutes"),
                &["convolutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exorting"),
                &["exporting", "extorting", "exerting", "exhorting"],
            ),
            (dictgen::InsensitiveStr::Ascii("symol"), &["symbol"]),
            (dictgen::InsensitiveStr::Ascii("exlicite"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("accodion"), &["accordion"]),
            (dictgen::InsensitiveStr::Ascii("invetory"), &["inventory"]),
            (
                dictgen::InsensitiveStr::Ascii("exresses"),
                &["expresses", "excesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("appeneded"), &["appended"]),
            (
                dictgen::InsensitiveStr::Ascii("termigator"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechnically"),
                &["mechanically", "technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coveniently"),
                &["conveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("asign"), &["assign"]),
            (
                dictgen::InsensitiveStr::Ascii("permitions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversubcribes"),
                &["oversubscribes"],
            ),
            (dictgen::InsensitiveStr::Ascii("executre"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("intialsing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installlations"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("milwakuee"), &["milwaukee"]),
            (
                dictgen::InsensitiveStr::Ascii("equivolents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("consice"), &["concise"]),
            (
                dictgen::InsensitiveStr::Ascii("erronenously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siturations"),
                &["situations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pilrgim"), &["pilgrim"]),
            (
                dictgen::InsensitiveStr::Ascii("configruations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sawtay"), &["sauté"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomrpessed"),
                &["uncompressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminastrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("poluted"), &["polluted"]),
            (
                dictgen::InsensitiveStr::Ascii("intemediaries"),
                &["intermediaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordinair"),
                &["extraordinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconencting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("redeemd"), &["redeemed"]),
            (dictgen::InsensitiveStr::Ascii("althoguh"), &["although"]),
            (dictgen::InsensitiveStr::Ascii("happenning"), &["happening"]),
            (
                dictgen::InsensitiveStr::Ascii("antisipates"),
                &["anticipates"],
            ),
            (dictgen::InsensitiveStr::Ascii("instaance"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("coalessense"),
                &["coalescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ser"), &["set"]),
            (
                dictgen::InsensitiveStr::Ascii("stimilated"),
                &["stimulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hihg"), &["high"]),
            (dictgen::InsensitiveStr::Ascii("definiet"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("discconeections"),
                &["disconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organziation"),
                &["organization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintianing"),
                &["maintaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualtors"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practioners"),
                &["practitioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("picniced"), &["picnicked"]),
            (dictgen::InsensitiveStr::Ascii("popoup"), &["popup"]),
            (
                dictgen::InsensitiveStr::Ascii("fianlizing"),
                &["finalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("samles"),
                &["samples", "seamless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninfrom"),
                &["uninform", "uniform"],
            ),
            (dictgen::InsensitiveStr::Ascii("sweatshit"), &["sweatshirt"]),
            (
                dictgen::InsensitiveStr::Ascii("combininators"),
                &["combinators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reigstering"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behvaioural"),
                &["behavioural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autheniticated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("eugneics"), &["eugenics"]),
            (dictgen::InsensitiveStr::Ascii("clared"), &["cleared"]),
            (
                dictgen::InsensitiveStr::Ascii("contrieving"),
                &["contriving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnateeeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("favorties"), &["favorites"]),
            (dictgen::InsensitiveStr::Ascii("collable"), &["callable"]),
            (
                dictgen::InsensitiveStr::Ascii("previoulsy"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocaition"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("currecny"), &["currency"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficiallied"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("langueges"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("altogehter"),
                &["altogether"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regoin"),
                &["regain", "region"],
            ),
            (dictgen::InsensitiveStr::Ascii("capabiity"), &["capability"]),
            (dictgen::InsensitiveStr::Ascii("botto"), &["bottom"]),
            (
                dictgen::InsensitiveStr::Ascii("inhertiance"),
                &["inheritance"],
            ),
            (dictgen::InsensitiveStr::Ascii("singoolar"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("lexigraphically"),
                &["lexicographically"],
            ),
            (dictgen::InsensitiveStr::Ascii("upporting"), &["supporting"]),
            (dictgen::InsensitiveStr::Ascii("somene"), &["someone"]),
            (dictgen::InsensitiveStr::Ascii("messanger"), &["messenger"]),
            (dictgen::InsensitiveStr::Ascii("judasim"), &["judaism"]),
            (
                dictgen::InsensitiveStr::Ascii("possabilites"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professiinal"),
                &["professionalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("retris"), &["retries"]),
            (dictgen::InsensitiveStr::Ascii("functin"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("prooduces"), &["produces"]),
            (
                dictgen::InsensitiveStr::Ascii("terminater"),
                &["terminator", "terminated", "terminates", "terminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internationnalism"),
                &["internationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("excells"), &["excels"]),
            (dictgen::InsensitiveStr::Ascii("adddress"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("renouned"),
                &["renounced", "renowned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmentations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcations"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("pretains"), &["pertains"]),
            (
                dictgen::InsensitiveStr::Ascii("curtesy"),
                &["courtesy", "curtsy"],
            ),
            (dictgen::InsensitiveStr::Ascii("convrt"), &["convert"]),
            (dictgen::InsensitiveStr::Ascii("lattitude"), &["latitude"]),
            (
                dictgen::InsensitiveStr::Ascii("refelections"),
                &["reflections"],
            ),
            (dictgen::InsensitiveStr::Ascii("hillarious"), &["hilarious"]),
            (
                dictgen::InsensitiveStr::Ascii("outoing"),
                &["outdoing", "outgoing", "outing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifeid"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("aniother"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("repetively"),
                &["repetitively", "receptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributins"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("detalis"), &["details"]),
            (
                dictgen::InsensitiveStr::Ascii("deactiviate"),
                &["deactivate"],
            ),
            (dictgen::InsensitiveStr::Ascii("uppercas"), &["uppercase"]),
            (dictgen::InsensitiveStr::Ascii("documnet"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("followes"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("rigeur"),
                &["rigueur", "rigour"],
            ),
            (dictgen::InsensitiveStr::Ascii("prference"), &["preference"]),
            (dictgen::InsensitiveStr::Ascii("dobled"), &["doubled"]),
            (
                dictgen::InsensitiveStr::Ascii("repeatitive"),
                &["repetitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("nurished"), &["nourished"]),
            (dictgen::InsensitiveStr::Ascii("goddammn"), &["goddamn"]),
            (dictgen::InsensitiveStr::Ascii("strippped"), &["stripped"]),
            (
                dictgen::InsensitiveStr::Ascii("victemized"),
                &["victimized"],
            ),
            (dictgen::InsensitiveStr::Ascii("intenseley"), &["intensely"]),
            (
                dictgen::InsensitiveStr::Ascii("convential"),
                &["conventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ternimator"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paychiatrist"),
                &["psychiatrist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilplier"),
                &["multiplier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southernerns"),
                &["southerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quaterions"),
                &["quaternions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alleigance"),
                &["allegiance"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucural"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberbility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alghoritmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosntruct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("includded"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("reknowned"), &["renowned"]),
            (dictgen::InsensitiveStr::Ascii("lifestel"), &["lifesteal"]),
            (dictgen::InsensitiveStr::Ascii("propterty"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("depandants"),
                &["dependants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temprararily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("accquired"), &["acquired"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosynchratic"),
                &["idiosyncratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emperically"),
                &["empirically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighouring"),
                &["neighbouring"],
            ),
            (dictgen::InsensitiveStr::Ascii("plugun"), &["plugin"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuye"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immeditately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("engineerd"), &["engineered"]),
            (
                dictgen::InsensitiveStr::Ascii("santitization"),
                &["sanitization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalsence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undernearth"),
                &["underneath"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pemissions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decsendants"),
                &["descendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmeneting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilcate"),
                &["complicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("agreemnet"), &["agreement"]),
            (
                dictgen::InsensitiveStr::Ascii("annoucnements"),
                &["announcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberterian"),
                &["libertarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrofy"),
                &["catastrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitcal"), &["optical"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactureers"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("tirangle"), &["triangle"]),
            (
                dictgen::InsensitiveStr::Ascii("configiuration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("accessintg"), &["accessing"]),
            (dictgen::InsensitiveStr::Ascii("exetent"), &["extent"]),
            (
                dictgen::InsensitiveStr::Ascii("cohabitating"),
                &["cohabiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitles"),
                &["subtitles", "substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spataializer"),
                &["spatializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("unselecgt"), &["unselect"]),
            (
                dictgen::InsensitiveStr::Ascii("attachements"),
                &["attachments"],
            ),
            (dictgen::InsensitiveStr::Ascii("denisty"), &["density"]),
            (
                dictgen::InsensitiveStr::Ascii("decommisioned"),
                &["decommissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlflowed"),
                &["overflowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decstiption"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("iwht"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoonds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("prdpagate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("screnshot"), &["screenshot"]),
            (dictgen::InsensitiveStr::Ascii("syste"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("metiphors"), &["metaphors"]),
            (
                dictgen::InsensitiveStr::Ascii("preipherals"),
                &["peripherals"],
            ),
            (dictgen::InsensitiveStr::Ascii("normalyze"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("anlyses"), &["analyses"]),
            (dictgen::InsensitiveStr::Ascii("disbility"), &["disability"]),
            (dictgen::InsensitiveStr::Ascii("noninital"), &["noninitial"]),
            (
                dictgen::InsensitiveStr::Ascii("copytights"),
                &["copyrights"],
            ),
            (dictgen::InsensitiveStr::Ascii("seeem"), &["seem"]),
            (
                dictgen::InsensitiveStr::Ascii("pupulating"),
                &["populating"],
            ),
            (dictgen::InsensitiveStr::Ascii("necissary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("acquring"), &["acquiring"]),
            (dictgen::InsensitiveStr::Ascii("broblem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("appoaches"), &["approaches"]),
            (dictgen::InsensitiveStr::Ascii("eliptical"), &["elliptical"]),
            (
                dictgen::InsensitiveStr::Ascii("lamba"),
                &["lambda", "lambs", "lamb"],
            ),
            (dictgen::InsensitiveStr::Ascii("selectig"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("instanciating"),
                &["instantiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepparately"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("exteneded"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("terminiators"),
                &["terminators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatitbly"),
                &["compatibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("keeplive"), &["keepalive"]),
            (dictgen::InsensitiveStr::Ascii("commercent"), &["commence"]),
            (dictgen::InsensitiveStr::Ascii("apporove"), &["approve"]),
            (dictgen::InsensitiveStr::Ascii("singulator"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("assymetriy"), &["asymmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("disporportionately"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlyzed"), &["analyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("deendencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinsim"),
                &["determinism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chane"),
                &["change", "chain"],
            ),
            (dictgen::InsensitiveStr::Ascii("sreturns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("manslaugther"),
                &["manslaughter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ailenating"),
                &["alienating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutliated"), &["mutilated"]),
            (dictgen::InsensitiveStr::Ascii("tiemr"), &["timer"]),
            (
                dictgen::InsensitiveStr::Ascii("cript"),
                &["crypt", "script"],
            ),
            (dictgen::InsensitiveStr::Ascii("mannar"), &["manner"]),
            (dictgen::InsensitiveStr::Ascii("complaind"), &["complained"]),
            (
                dictgen::InsensitiveStr::Ascii("trinagulations"),
                &["triangulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamnetalist"),
                &["fundamentalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ccertificate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interchangeablity"),
                &["interchangeably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horrendeous"),
                &["horrendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("sysem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("colsest"),
                &["closest", "coldest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervieni"),
                &["interviewing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accoring"),
                &["according", "occurring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indestrictible"),
                &["indestructible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteristicly"),
                &["characteristically"],
            ),
            (dictgen::InsensitiveStr::Ascii("simmetry"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("experiemental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangteables"),
                &["rearrangeables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltrare"),
                &["infiltrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("abillities"), &["abilities"]),
            (
                dictgen::InsensitiveStr::Ascii("unbreakble"),
                &["unbreakable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ugrades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("generig"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("repalying"), &["replaying"]),
            (dictgen::InsensitiveStr::Ascii("braevry"), &["bravery"]),
            (
                dictgen::InsensitiveStr::Ascii("emphisized"),
                &["emphasized"],
            ),
            (dictgen::InsensitiveStr::Ascii("quantifiy"), &["quantify"]),
            (dictgen::InsensitiveStr::Ascii("skipable"), &["skippable"]),
            (dictgen::InsensitiveStr::Ascii("vart"), &["cart", "wart"]),
            (dictgen::InsensitiveStr::Ascii("baiscly"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("intangeble"),
                &["intangible"],
            ),
            (dictgen::InsensitiveStr::Ascii("batlimore"), &["baltimore"]),
            (
                dictgen::InsensitiveStr::Ascii("presntation"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sycling"), &["cycling"]),
            (
                dictgen::InsensitiveStr::Ascii("irresistably"),
                &["irresistibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("overidden"), &["overridden"]),
            (
                dictgen::InsensitiveStr::Ascii("confuigred"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("graphcially"),
                &["graphically"],
            ),
            (dictgen::InsensitiveStr::Ascii("spartanops"), &["spartans"]),
            (dictgen::InsensitiveStr::Ascii("deposites"), &["deposits"]),
            (dictgen::InsensitiveStr::Ascii("disconnet"), &["disconnect"]),
            (dictgen::InsensitiveStr::Ascii("reaon"), &["reason"]),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitured"),
                &["substituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemnents"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhoood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eqaulities"),
                &["equalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremmental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funktionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interepretation"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminoligy"),
                &["terminology"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffor"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("conterpart"),
                &["counterpart"],
            ),
            (dictgen::InsensitiveStr::Ascii("massace"), &["massacre"]),
            (dictgen::InsensitiveStr::Ascii("dpeending"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("objetion"), &["objection"]),
            (
                dictgen::InsensitiveStr::Ascii("defauled"),
                &["default", "defaulted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitaitons"),
                &["limitation", "limitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percieving"),
                &["perceiving"],
            ),
            (dictgen::InsensitiveStr::Ascii("banditas"), &["bandits"]),
            (
                dictgen::InsensitiveStr::Ascii("unanomously"),
                &["unanimously"],
            ),
            (dictgen::InsensitiveStr::Ascii("hispancis"), &["hispanics"]),
            (dictgen::InsensitiveStr::Ascii("savelt"), &["svelte"]),
            (
                dictgen::InsensitiveStr::Ascii("copilation"),
                &["compilation", "copulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcstions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("mushrooom"), &["mushroom"]),
            (
                dictgen::InsensitiveStr::Ascii("effectivily"),
                &["effectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("profling"), &["profiling"]),
            (
                dictgen::InsensitiveStr::Ascii("accompagnied"),
                &["accompanied"],
            ),
            (dictgen::InsensitiveStr::Ascii("rturned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("obyekt"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("exising"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("convertie"),
                &["convertible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emulater"),
                &["emulator", "emulated", "emulates", "emulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatio"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calssifiers"),
                &["classifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cald"), &["called"]),
            (dictgen::InsensitiveStr::Ascii("beizer"), &["bezier"]),
            (dictgen::InsensitiveStr::Ascii("wokrbooks"), &["workbooks"]),
            (dictgen::InsensitiveStr::Ascii("functtions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("communisit"), &["communist"]),
            (dictgen::InsensitiveStr::Ascii("pivott"), &["pivot"]),
            (
                dictgen::InsensitiveStr::Ascii("seached"),
                &["searched", "reached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesssive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("shedule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("probilities"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("formallity"), &["formally"]),
            (dictgen::InsensitiveStr::Ascii("derivates"), &["derivate"]),
            (
                dictgen::InsensitiveStr::Ascii("preipheral"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convinience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpoart"),
                &["counterpart"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwar"), &["forward"]),
            (dictgen::InsensitiveStr::Ascii("llok"), &["look"]),
            (dictgen::InsensitiveStr::Ascii("jornaled"), &["journaled"]),
            (dictgen::InsensitiveStr::Ascii("sqaurely"), &["squarely"]),
            (
                dictgen::InsensitiveStr::Ascii("registrain"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partiticipants"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissiable"),
                &["permissible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destnations"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("muext"), &["mutex"]),
            (dictgen::InsensitiveStr::Ascii("amdgput"), &["amdgpu"]),
            (
                dictgen::InsensitiveStr::Ascii("projecters"),
                &["projectors", "projects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheeters"),
                &["spreadsheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("sorpanos"), &["sopranos"]),
            (dictgen::InsensitiveStr::Ascii("pedestral"), &["pedestal"]),
            (
                dictgen::InsensitiveStr::Ascii("integrato"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("adrress"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("pretified"), &["prettified"]),
            (
                dictgen::InsensitiveStr::Ascii("extraordianry"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("pytho"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("deifnetly"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("agrigultural"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temproarly"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proabilities"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deficiencey"),
                &["deficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sharpeneing"),
                &["sharpening"],
            ),
            (dictgen::InsensitiveStr::Ascii("minnestoa"), &["minnesota"]),
            (dictgen::InsensitiveStr::Ascii("gathet"), &["gather"]),
            (dictgen::InsensitiveStr::Ascii("protocools"), &["protocols"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("qiestion"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("delais"), &["delays"]),
            (dictgen::InsensitiveStr::Ascii("wroks"), &["works"]),
            (
                dictgen::InsensitiveStr::Ascii("horizonatally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representaciones"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializiations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("associateed"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("fasade"), &["facade"]),
            (
                dictgen::InsensitiveStr::Ascii("indepandance"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brightnesss"),
                &["brightness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincedentally"),
                &["coincidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("follwoeinwg"),
                &["following"],
            ),
            (dictgen::InsensitiveStr::Ascii("traverced"), &["traversed"]),
            (dictgen::InsensitiveStr::Ascii("omiting"), &["omitting"]),
            (
                dictgen::InsensitiveStr::Ascii("metalurgical"),
                &["metallurgical"],
            ),
            (dictgen::InsensitiveStr::Ascii("convergens"), &["converse"]),
            (
                dictgen::InsensitiveStr::Ascii("kryptonyte"),
                &["kryptonite"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouurs"), &["ours"]),
            (dictgen::InsensitiveStr::Ascii("entquire"), &["enquire"]),
            (
                dictgen::InsensitiveStr::Ascii("effictiveness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coreespondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacialisations"),
                &["specialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comverters"),
                &["converters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delusionnal"),
                &["delusional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decied"),
                &["decide", "decided"],
            ),
            (dictgen::InsensitiveStr::Ascii("happenns"), &["happens"]),
            (dictgen::InsensitiveStr::Ascii("autimatic"), &["automatic"]),
            (
                dictgen::InsensitiveStr::Ascii("ficticious"),
                &["fictitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissalows"), &["disallows"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnderable"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volunteerd"),
                &["volunteered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurized"),
                &["pasteurized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loclaization"),
                &["localization"],
            ),
            (dictgen::InsensitiveStr::Ascii("eytmology"), &["etymology"]),
            (dictgen::InsensitiveStr::Ascii("preferece"), &["preference"]),
            (
                dictgen::InsensitiveStr::Ascii("vaaluations"),
                &["valuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confrimation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("episdoes"), &["episodes"]),
            (dictgen::InsensitiveStr::Ascii("sanatizes"), &["sanitizes"]),
            (dictgen::InsensitiveStr::Ascii("flages"), &["flags"]),
            (
                dictgen::InsensitiveStr::Ascii("expressons"),
                &["expressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approiate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("raplaces"), &["replaces"]),
            (
                dictgen::InsensitiveStr::Ascii("functioing"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("toolar"), &["toolbar"]),
            (dictgen::InsensitiveStr::Ascii("asymmeric"), &["asymmetric"]),
            (dictgen::InsensitiveStr::Ascii("nigthclub"), &["nightclub"]),
            (
                dictgen::InsensitiveStr::Ascii("somethinks"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("opartor"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("corollla"), &["corolla"]),
            (
                dictgen::InsensitiveStr::Ascii("reporductive"),
                &["reproductive"],
            ),
            (dictgen::InsensitiveStr::Ascii("sructural"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("copstruction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unqaulified"),
                &["unqualified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("variale"),
                &["variable", "variate"],
            ),
            (dictgen::InsensitiveStr::Ascii("relatvie"), &["relative"]),
            (dictgen::InsensitiveStr::Ascii("palletted"), &["paletted"]),
            (dictgen::InsensitiveStr::Ascii("dokument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("psuedoclasses"),
                &["pseudoclasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarquist"), &["anarchist"]),
            (dictgen::InsensitiveStr::Ascii("bakcrefs"), &["backrefs"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("permitas"), &["permits"]),
            (
                dictgen::InsensitiveStr::Ascii("exhautivity"),
                &["exhaustivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("functiosn"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("cobination"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confirguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("axisymmectric"),
                &["axisymmetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scorebored"),
                &["scoreboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("selekts"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("relatinoship"),
                &["relationships"],
            ),
            (dictgen::InsensitiveStr::Ascii("headle"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("ptherad"), &["pthread"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusivy"),
                &["exclusivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sycronizing"),
                &["synchronizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invectigation"),
                &["investigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("empahsis"), &["emphasis"]),
            (
                dictgen::InsensitiveStr::Ascii("decscription"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessively"),
                &["successively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excrumentially"),
                &["excruciatingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("desctructs"), &["destructs"]),
            (
                dictgen::InsensitiveStr::Ascii("perposing"),
                &["proposing", "purposing"],
            ),
            (dictgen::InsensitiveStr::Ascii("heroe"), &["hero"]),
            (dictgen::InsensitiveStr::Ascii("substract"), &["subtract"]),
            (dictgen::InsensitiveStr::Ascii("koolot"), &["culotte"]),
            (
                dictgen::InsensitiveStr::Ascii("repsresents"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("tork"), &["torque"]),
            (
                dictgen::InsensitiveStr::Ascii("preorderded"),
                &["preordered"],
            ),
            (dictgen::InsensitiveStr::Ascii("authrorize"), &["authorize"]),
            (
                dictgen::InsensitiveStr::Ascii("disgnostic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sweetheat"), &["sweetheart"]),
            (dictgen::InsensitiveStr::Ascii("breacket"), &["bracket"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaites"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("intial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("ahere"), &["adhere", "here"]),
            (dictgen::InsensitiveStr::Ascii("exceutors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("activly"), &["actively"]),
            (
                dictgen::InsensitiveStr::Ascii("canocically"),
                &["canonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inmediatily"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexualls"),
                &["homosexuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsedent"), &["subsequent"]),
            (dictgen::InsensitiveStr::Ascii("secions"), &["sections"]),
            (
                dictgen::InsensitiveStr::Ascii("dimmensions"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("receips"), &["receipts"]),
            (dictgen::InsensitiveStr::Ascii("temporory"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("whitepsace"),
                &["whitespace"],
            ),
            (dictgen::InsensitiveStr::Ascii("underlows"), &["underflows"]),
            (dictgen::InsensitiveStr::Ascii("clarifiies"), &["clarifies"]),
            (dictgen::InsensitiveStr::Ascii("scaned"), &["scanned"]),
            (dictgen::InsensitiveStr::Ascii("temportary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("actuak"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("invaderas"), &["invaders"]),
            (
                dictgen::InsensitiveStr::Ascii("parmameters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("toleranz"), &["tolerance"]),
            (
                dictgen::InsensitiveStr::Ascii("superviors"),
                &["supervisors"],
            ),
            (dictgen::InsensitiveStr::Ascii("indoensia"), &["indonesia"]),
            (
                dictgen::InsensitiveStr::Ascii("declataions"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bramches"), &["branches"]),
            (
                dictgen::InsensitiveStr::Ascii("prolematic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiprocesser"),
                &["multiprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissmantling"),
                &["dismantling"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximimums"), &["maximums"]),
            (dictgen::InsensitiveStr::Ascii("insaney"), &["insanely"]),
            (dictgen::InsensitiveStr::Ascii("sepcially"), &["specially"]),
            (dictgen::InsensitiveStr::Ascii("rnger"), &["ranger"]),
            (dictgen::InsensitiveStr::Ascii("ignoed"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("arrnages"), &["arranges"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotionate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dianostic"), &["diagnostic"]),
            (dictgen::InsensitiveStr::Ascii("recenet"), &["recent"]),
            (dictgen::InsensitiveStr::Ascii("braceletes"), &["bracelets"]),
            (
                dictgen::InsensitiveStr::Ascii("consultato"),
                &["consultation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highjackers"),
                &["hijackers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("franchizes"),
                &["franchises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoritically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comporisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorithim"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("everyonehas"),
                &["everyones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disscussion"),
                &["discussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("filllers"), &["fillers"]),
            (
                dictgen::InsensitiveStr::Ascii("envrironments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("whrilwind"), &["whirlwind"]),
            (dictgen::InsensitiveStr::Ascii("neighrour"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("dicline"), &["decline"]),
            (
                dictgen::InsensitiveStr::Ascii("componants"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canabelyzed"),
                &["cannibalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainging"), &["training"]),
            (dictgen::InsensitiveStr::Ascii("remvove"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("obstancles"), &["obstacles"]),
            (dictgen::InsensitiveStr::Ascii("demog"), &["demo"]),
            (
                dictgen::InsensitiveStr::Ascii("neccessarrily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("feezer"), &["freezer"]),
            (dictgen::InsensitiveStr::Ascii("colapses"), &["collapses"]),
            (dictgen::InsensitiveStr::Ascii("buidlings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("intrinscis"),
                &["intrinsics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succedfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("clasical"), &["classical"]),
            (dictgen::InsensitiveStr::Ascii("animaiton"), &["animation"]),
            (dictgen::InsensitiveStr::Ascii("camcel"), &["camel"]),
            (
                dictgen::InsensitiveStr::Ascii("conbined"),
                &["combined", "confined"],
            ),
            (dictgen::InsensitiveStr::Ascii("somthing"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("gropu"), &["group"]),
            (dictgen::InsensitiveStr::Ascii("themporary"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("reporsitories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosophia"),
                &["philosophical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalisitic"),
                &["journalistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("kuridsh"), &["kurdish"]),
            (
                dictgen::InsensitiveStr::Ascii("implementate"),
                &["implement"],
            ),
            (dictgen::InsensitiveStr::Ascii("indiates"), &["indicates"]),
            (dictgen::InsensitiveStr::Ascii("realisitc"), &["realistic"]),
            (
                dictgen::InsensitiveStr::Ascii("scholerships"),
                &["scholarships"],
            ),
            (dictgen::InsensitiveStr::Ascii("terrabytes"), &["terabytes"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubted"),
                &["distributed", "disrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandchilder"),
                &["grandchildren"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyoes"), &["types", "toes"]),
            (dictgen::InsensitiveStr::Ascii("limtit"), &["limit"]),
            (
                dictgen::InsensitiveStr::Ascii("baased"),
                &["based", "abased"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promotionnal"),
                &["promotional"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetting"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("checkums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("implimenting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("decondings"), &["decodings"]),
            (dictgen::InsensitiveStr::Ascii("oppponent"), &["opponent"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhoofs"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overeaching"),
                &["overarching"],
            ),
            (dictgen::InsensitiveStr::Ascii("putput"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("exerciese"),
                &["exercises", "exercise"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppposes"), &["supposes"]),
            (dictgen::InsensitiveStr::Ascii("guidline"), &["guideline"]),
            (
                dictgen::InsensitiveStr::Ascii("explcitely"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("statisitc"), &["statistic"]),
            (
                dictgen::InsensitiveStr::Ascii("descroptor"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrseponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepcifying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitiley"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("actualmy"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("alphbetic"), &["alphabetic"]),
            (
                dictgen::InsensitiveStr::Ascii("molestato"),
                &["molestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alls"), &["all", "falls"]),
            (dictgen::InsensitiveStr::Ascii("dabshboard"), &["dashboard"]),
            (dictgen::InsensitiveStr::Ascii("secong"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("recquirement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesion"), &["tension"]),
            (
                dictgen::InsensitiveStr::Ascii("highligter"),
                &["highlighter"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghth"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("experimeners"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("clearifies"), &["clarifies"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumarated"),
                &["enumerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poisitioned"),
                &["positioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("contronl"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("technition"),
                &["technician"],
            ),
            (dictgen::InsensitiveStr::Ascii("brobable"), &["probable"]),
            (
                dictgen::InsensitiveStr::Ascii("cotributing"),
                &["contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("deleteting"), &["deleting"]),
            (
                dictgen::InsensitiveStr::Ascii("viusualisation"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("equtions"), &["equations"]),
            (
                dictgen::InsensitiveStr::Ascii("situationals"),
                &["situations", "situational"],
            ),
            (dictgen::InsensitiveStr::Ascii("annote"), &["annotate"]),
            (
                dictgen::InsensitiveStr::Ascii("reprodiction"),
                &["reproduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("schould"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("appoximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("diasables"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("optomistically"),
                &["optimistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("forenics"), &["forensic"]),
            (
                dictgen::InsensitiveStr::Ascii("uninterrutable"),
                &["uninterruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interwines"),
                &["intertwines"],
            ),
            (dictgen::InsensitiveStr::Ascii("activisiom"), &["activism"]),
            (
                dictgen::InsensitiveStr::Ascii("exahusting"),
                &["exhausting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satistically"),
                &["statistically", "sadistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wew"), &["we"]),
            (dictgen::InsensitiveStr::Ascii("emultors"), &["emulators"]),
            (
                dictgen::InsensitiveStr::Ascii("progressisme"),
                &["progresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("excector"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("shecker"),
                &["checker", "shucker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explizitly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("boddy"), &["body"]),
            (dictgen::InsensitiveStr::Ascii("confued"), &["confused"]),
            (dictgen::InsensitiveStr::Ascii("detction"), &["detection"]),
            (
                dictgen::InsensitiveStr::Ascii("apoligists"),
                &["apologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("execeed"), &["exceed"]),
            (
                dictgen::InsensitiveStr::Ascii("negoators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repsonsibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("volitile"), &["volatile"]),
            (dictgen::InsensitiveStr::Ascii("manuever"), &["maneuver"]),
            (dictgen::InsensitiveStr::Ascii("oridinals"), &["ordinals"]),
            (dictgen::InsensitiveStr::Ascii("arogent"), &["arrogant"]),
            (dictgen::InsensitiveStr::Ascii("annotors"), &["annotators"]),
            (
                dictgen::InsensitiveStr::Ascii("internationaly"),
                &["internationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("rivlary"), &["rivalry"]),
            (dictgen::InsensitiveStr::Ascii("witdth"), &["width"]),
            (
                dictgen::InsensitiveStr::Ascii("neighobring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("squring"),
                &["squaring", "squiring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatenaion"),
                &["concatenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("moreso"), &["more"]),
            (dictgen::InsensitiveStr::Ascii("tabbaray"), &["taboret"]),
            (dictgen::InsensitiveStr::Ascii("upadted"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("manadatory"), &["mandatory"]),
            (dictgen::InsensitiveStr::Ascii("failur"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("hauntig"), &["haunting"]),
            (dictgen::InsensitiveStr::Ascii("ineqality"), &["inequality"]),
            (
                dictgen::InsensitiveStr::Ascii("disciplinin"),
                &["disciplining"],
            ),
            (dictgen::InsensitiveStr::Ascii("referecen"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("abbreivating"),
                &["abbreviating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appromixation"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outreagous"),
                &["outrageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpsiration"),
                &["inspiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swithcboard"),
                &["switchboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valiable"),
                &["variable", "valuable", "available"],
            ),
            (dictgen::InsensitiveStr::Ascii("pulgins"), &["plugins"]),
            (
                dictgen::InsensitiveStr::Ascii("collecions"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perephirals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responnsibilty"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("supposin"), &["supposing"]),
            (dictgen::InsensitiveStr::Ascii("throguh"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("susbtrate"), &["substrate"]),
            (
                dictgen::InsensitiveStr::Ascii("seplicural"),
                &["sepulchral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unforgiveable"),
                &["unforgivable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aritficial"),
                &["artificial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeliebable"),
                &["unbelievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreadhseets"),
                &["spreadsheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("alchemey"), &["alchemy"]),
            (
                dictgen::InsensitiveStr::Ascii("accociations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("paniics"), &["panics"]),
            (dictgen::InsensitiveStr::Ascii("worbook"), &["workbook"]),
            (dictgen::InsensitiveStr::Ascii("regualar"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("idenitities"),
                &["identities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialzers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tablesppon"),
                &["tablespoon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renforcements"),
                &["reinforcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuel"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("partnerd"), &["partnered"]),
            (
                dictgen::InsensitiveStr::Ascii("dericteryes"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("excective"), &["executive"]),
            (dictgen::InsensitiveStr::Ascii("preceds"), &["precedes"]),
            (
                dictgen::InsensitiveStr::Ascii("deambiguage"),
                &["disambiguate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presetation"),
                &["presentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("africas"), &["africans"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreegated"),
                &["aggregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("untuitive"),
                &["unintuitive", "intuitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("thay"), &["they", "that"]),
            (dictgen::InsensitiveStr::Ascii("inifnite"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("benckmarked"),
                &["benchmarked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependensies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anpitheaters"),
                &["amphitheaters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affliciton"),
                &["affliction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfortabil"),
                &["comfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denomonation"),
                &["denomination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncalizing"),
                &["canonicalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wallpapr"), &["wallpapers"]),
            (dictgen::InsensitiveStr::Ascii("expemts"), &["exempts"]),
            (dictgen::InsensitiveStr::Ascii("kknow"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("terimnators"),
                &["terminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("entit"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("termianls"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("impressoin"),
                &["impressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organisationens"),
                &["organisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("senconds"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("obssessed"), &["obsessed"]),
            (
                dictgen::InsensitiveStr::Ascii("compressable"),
                &["compressible"],
            ),
            (dictgen::InsensitiveStr::Ascii("hassel"), &["hassle"]),
            (
                dictgen::InsensitiveStr::Ascii("definetelly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("firdt"), &["first", "third"]),
            (dictgen::InsensitiveStr::Ascii("vertixes"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("deveolps"), &["develops"]),
            (dictgen::InsensitiveStr::Ascii("therefo"), &["thereof"]),
            (dictgen::InsensitiveStr::Ascii("sluggify"), &["slugify"]),
            (dictgen::InsensitiveStr::Ascii("traingles"), &["triangles"]),
            (dictgen::InsensitiveStr::Ascii("anaolgues"), &["analogues"]),
            (
                dictgen::InsensitiveStr::Ascii("intoxicacion"),
                &["intoxication"],
            ),
            (dictgen::InsensitiveStr::Ascii("frontents"), &["frontends"]),
            (dictgen::InsensitiveStr::Ascii("somethis"), &["somethings"]),
            (
                dictgen::InsensitiveStr::Ascii("execurable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("submoules"), &["submodules"]),
            (
                dictgen::InsensitiveStr::Ascii("commpressd"),
                &["compressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("gorvement"), &["government"]),
            (
                dictgen::InsensitiveStr::Ascii("codder"),
                &["coder", "coddler"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoerter"), &["supporter"]),
            (
                dictgen::InsensitiveStr::Ascii("bewteeness"),
                &["betweenness"],
            ),
            (dictgen::InsensitiveStr::Ascii("advatange"), &["advantage"]),
            (
                dictgen::InsensitiveStr::Ascii("figurativelly"),
                &["figuratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("stressers"), &["stresses"]),
            (
                dictgen::InsensitiveStr::Ascii("opproximately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critiziced"),
                &["criticized"],
            ),
            (dictgen::InsensitiveStr::Ascii("sprtie"), &["sprite"]),
            (dictgen::InsensitiveStr::Ascii("postcript"), &["postscript"]),
            (dictgen::InsensitiveStr::Ascii("appreared"), &["appeared"]),
            (dictgen::InsensitiveStr::Ascii("errocode"), &["errorcode"]),
            (
                dictgen::InsensitiveStr::Ascii("frustrasion"),
                &["frustrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("limtis"), &["limits"]),
            (
                dictgen::InsensitiveStr::Ascii("prepetually"),
                &["perpetually"],
            ),
            (dictgen::InsensitiveStr::Ascii("dailogue"), &["dialogue"]),
            (dictgen::InsensitiveStr::Ascii("specififes"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("solating"), &["isolating"]),
            (
                dictgen::InsensitiveStr::Ascii("prealocated"),
                &["preallocated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nanosencond"),
                &["nanosecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("titels"), &["titles"]),
            (
                dictgen::InsensitiveStr::Ascii("prothsetic"),
                &["prosthetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("targtes"), &["targets"]),
            (dictgen::InsensitiveStr::Ascii("claered"), &["cleared"]),
            (dictgen::InsensitiveStr::Ascii("trackes"), &["trackers"]),
            (dictgen::InsensitiveStr::Ascii("opposin"), &["opposing"]),
            (dictgen::InsensitiveStr::Ascii("strctures"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("succeeeding"),
                &["succeeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("supproted"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("destinguish"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undertsand"),
                &["understands", "understand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrensponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("lollipoop"), &["lollipop"]),
            (
                dictgen::InsensitiveStr::Ascii("relected"),
                &["reelected", "reflected"],
            ),
            (dictgen::InsensitiveStr::Ascii("accounds"), &["accounts"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetuaded"),
                &["perpetuated"],
            ),
            (dictgen::InsensitiveStr::Ascii("celebirty"), &["celebrity"]),
            (
                dictgen::InsensitiveStr::Ascii("identtifier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("addign"), &["adding"]),
            (
                dictgen::InsensitiveStr::Ascii("undependend"),
                &["independent", "nondependent"],
            ),
            (dictgen::InsensitiveStr::Ascii("whirpools"), &["whirlpools"]),
            (
                dictgen::InsensitiveStr::Ascii("funciotned"),
                &["functioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidare"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impoversihed"),
                &["impoverished"],
            ),
            (dictgen::InsensitiveStr::Ascii("inculding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("destributing"),
                &["distributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exrensions"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comensurate"),
                &["commensurate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("andle"),
                &["handle", "candle", "angle", "ankle", "dandle"],
            ),
            (dictgen::InsensitiveStr::Ascii("succceeds"), &["succeeds"]),
            (
                dictgen::InsensitiveStr::Ascii("documemted"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unplayeble"),
                &["unplayable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tracsoding"),
                &["transcoding"],
            ),
            (dictgen::InsensitiveStr::Ascii("supeblock"), &["superblock"]),
            (dictgen::InsensitiveStr::Ascii("statefull"), &["stateful"]),
            (
                dictgen::InsensitiveStr::Ascii("instantenious"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("secreenshots"),
                &["screenshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crrespondingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("agregated"), &["aggregated"]),
            (dictgen::InsensitiveStr::Ascii("protool"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("annyo"), &["annoy"]),
            (
                dictgen::InsensitiveStr::Ascii("progating"),
                &["propagating"],
            ),
            (dictgen::InsensitiveStr::Ascii("giltches"), &["glitches"]),
            (dictgen::InsensitiveStr::Ascii("villian"), &["villain"]),
            (dictgen::InsensitiveStr::Ascii("enything"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("unspecifed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependnecy"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inbalanced"),
                &["imbalanced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inintialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valentinas"),
                &["valentines"],
            ),
            (dictgen::InsensitiveStr::Ascii("expesnces"), &["expenses"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilzation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("salvery"), &["slavery"]),
            (dictgen::InsensitiveStr::Ascii("autorized"), &["authorized"]),
            (dictgen::InsensitiveStr::Ascii("furing"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("sav"), &["save"]),
            (dictgen::InsensitiveStr::Ascii("cannabile"), &["cannibal"]),
            (
                dictgen::InsensitiveStr::Ascii("coorespond"),
                &["correspond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decorater"),
                &["decorator", "decorated", "decorates", "decorate"],
            ),
            (dictgen::InsensitiveStr::Ascii("yelow"), &["yellow"]),
            (
                dictgen::InsensitiveStr::Ascii("perfecxion"),
                &["perfection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifyable"),
                &["modifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckecksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("compoennts"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("gord"), &["gourd"]),
            (dictgen::InsensitiveStr::Ascii("multilier"), &["multiplier"]),
            (dictgen::InsensitiveStr::Ascii("copyeight"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("fallabcks"), &["fallbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("intaces"),
                &["instance", "instances", "intakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("specififcs"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("projectyle"),
                &["projectile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nofifications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("stackk"), &["stack"]),
            (
                dictgen::InsensitiveStr::Ascii("indepenedently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phsyicians"),
                &["physicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcdory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("stads"),
                &["stats", "stands"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabeld"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("decrases"),
                &["decreases", "deceases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stimualting"),
                &["stimulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("brightern"), &["brighten"]),
            (dictgen::InsensitiveStr::Ascii("buddihsts"), &["buddhists"]),
            (
                dictgen::InsensitiveStr::Ascii("sistemized"),
                &["systemized"],
            ),
            (dictgen::InsensitiveStr::Ascii("entorpy"), &["entropy"]),
            (
                dictgen::InsensitiveStr::Ascii("supoorting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("featchss"), &["fetches"]),
            (dictgen::InsensitiveStr::Ascii("bachelour"), &["bachelor"]),
            (dictgen::InsensitiveStr::Ascii("writewr"), &["writer"]),
            (
                dictgen::InsensitiveStr::Ascii("commissionned"),
                &["commissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaserchers"),
                &["researchers"],
            ),
            (dictgen::InsensitiveStr::Ascii("rythms"), &["rhythms"]),
            (dictgen::InsensitiveStr::Ascii("corruptin"), &["corruption"]),
            (dictgen::InsensitiveStr::Ascii("opearnd"), &["operand"]),
            (dictgen::InsensitiveStr::Ascii("reamined"), &["remained"]),
            (
                dictgen::InsensitiveStr::Ascii("wakeus"),
                &["wakeups", "walrus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noticeablely"),
                &["noticeably"],
            ),
            (dictgen::InsensitiveStr::Ascii("brasillian"), &["brazilian"]),
            (
                dictgen::InsensitiveStr::Ascii("assertations"),
                &["assertions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultinously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pilgrimmages"),
                &["pilgrimages"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloactor"), &["allocator"]),
            (
                dictgen::InsensitiveStr::Ascii("investmet"),
                &["investments"],
            ),
            (dictgen::InsensitiveStr::Ascii("luckly"), &["luckily"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("exspecting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("sprinklered"),
                &["sprinkled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccurred"),
                &["recurred", "reoccurred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequtive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("proverai"), &["proverbial"]),
            (dictgen::InsensitiveStr::Ascii("therefoer"), &["therefor"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrgations"),
                &["aggregations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyrllic"), &["cyrillic"]),
            (
                dictgen::InsensitiveStr::Ascii("successfullt"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrenging"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("additionalyy"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("exitsting"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("sidelen"), &["sideline"]),
            (dictgen::InsensitiveStr::Ascii("reagrding"), &["regarding"]),
            (dictgen::InsensitiveStr::Ascii("throurh"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("aternies"), &["attorneys"]),
            (dictgen::InsensitiveStr::Ascii("dicarded"), &["discarded"]),
            (dictgen::InsensitiveStr::Ascii("atttackers"), &["attackers"]),
            (
                dictgen::InsensitiveStr::Ascii("pentsylvania"),
                &["pennsylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executbles"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticatin"),
                &["authenticating", "authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("acordion"), &["accordion"]),
            (dictgen::InsensitiveStr::Ascii("excapes"), &["escapes"]),
            (dictgen::InsensitiveStr::Ascii("ymbols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("doubel"), &["double"]),
            (dictgen::InsensitiveStr::Ascii("emphaise"), &["emphasise"]),
            (
                dictgen::InsensitiveStr::Ascii("torublesome"),
                &["troublesome"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrobes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("pleace"),
                &["please", "place"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertaryans"),
                &["libertarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaiser"), &["localiser"]),
            (dictgen::InsensitiveStr::Ascii("capsulets"), &["capsules"]),
            (
                dictgen::InsensitiveStr::Ascii("intallations"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("overright"), &["overwrite"]),
            (
                dictgen::InsensitiveStr::Ascii("optimistaions"),
                &["optimisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemtned"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitiation"),
                &["initiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabilaty"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramlizes"),
                &["normalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("commecen"), &["commence"]),
            (
                dictgen::InsensitiveStr::Ascii("affinitied"),
                &["affinities"],
            ),
            (dictgen::InsensitiveStr::Ascii("strengthes"), &["strengths"]),
            (
                dictgen::InsensitiveStr::Ascii("parameterically"),
                &["parametrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("unitility"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("delares"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("montion"),
                &["mention", "motion"],
            ),
            (dictgen::InsensitiveStr::Ascii("auotmaton"), &["automaton"]),
            (dictgen::InsensitiveStr::Ascii("truned"), &["turned"]),
            (
                dictgen::InsensitiveStr::Ascii("formualating"),
                &["formulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograpy"),
                &["photography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitplayer"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aaggregation"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downagrading"),
                &["downgrading"],
            ),
            (dictgen::InsensitiveStr::Ascii("examplars"), &["exemplars"]),
            (
                dictgen::InsensitiveStr::Ascii("shatterling"),
                &["shattering"],
            ),
            (dictgen::InsensitiveStr::Ascii("furances"), &["furnaces"]),
            (
                dictgen::InsensitiveStr::Ascii("erroneosly"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benckmarking"),
                &["benchmarking"],
            ),
            (dictgen::InsensitiveStr::Ascii("adderesses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("neccesities"),
                &["necessities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curcumstances"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otpimizers"),
                &["optimizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persocuted"),
                &["persecuted"],
            ),
            (dictgen::InsensitiveStr::Ascii("conqured"), &["conquered"]),
            (dictgen::InsensitiveStr::Ascii("marginis"), &["margins"]),
            (dictgen::InsensitiveStr::Ascii("mideval"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("sests"), &["sets"]),
            (dictgen::InsensitiveStr::Ascii("creatse"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("modificatins"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("colected"), &["collected"]),
            (
                dictgen::InsensitiveStr::Ascii("transesxuals"),
                &["transsexuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("austiran"), &["austrian"]),
            (
                dictgen::InsensitiveStr::Ascii("conseqeunces"),
                &["consequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("malignent"), &["malignant"]),
            (
                dictgen::InsensitiveStr::Ascii("aboud"),
                &["about", "abound"],
            ),
            (dictgen::InsensitiveStr::Ascii("systyem"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("isimilar"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("streatches"), &["stretches"]),
            (dictgen::InsensitiveStr::Ascii("trubling"), &["troubling"]),
            (
                dictgen::InsensitiveStr::Ascii("incidential"),
                &["incidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("krypotnite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("audiobookmrs"),
                &["audiobooks"],
            ),
            (dictgen::InsensitiveStr::Ascii("temproary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("seperates"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("consicence"),
                &["conscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representationen"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithmus"),
                &["algorithms", "algorithm"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprciate"), &["appreciate"]),
            (
                dictgen::InsensitiveStr::Ascii("nclusively"),
                &["inclusively", "conclusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothisis"),
                &["hypothesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("referrence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("neighbeors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("crystalens"), &["crystals"]),
            (
                dictgen::InsensitiveStr::Ascii("exorbatant"),
                &["exorbitant"],
            ),
            (dictgen::InsensitiveStr::Ascii("executuon"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("conetenting"),
                &["contenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qudrangles"),
                &["quadrangles"],
            ),
            (dictgen::InsensitiveStr::Ascii("ugrade"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("processiong"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("epiloge"), &["epilogue"]),
            (dictgen::InsensitiveStr::Ascii("exectuted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("suppliementing"),
                &["supplementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transacrtions"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apocaplyse"),
                &["apocalypse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excplictly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("datasset"), &["dataset"]),
            (dictgen::InsensitiveStr::Ascii("absoluthe"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("unicornis"), &["unicorns"]),
            (
                dictgen::InsensitiveStr::Ascii("vitualised"),
                &["virtualised", "visualised"],
            ),
            (dictgen::InsensitiveStr::Ascii("asembling"), &["assembling"]),
            (
                dictgen::InsensitiveStr::Ascii("cofirmation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempate"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("hwoever"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("familiarizate"),
                &["familiarize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtraction"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("govenment"), &["government"]),
            (dictgen::InsensitiveStr::Ascii("tamplate"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("metacharater"),
                &["metacharacter"],
            ),
            (dictgen::InsensitiveStr::Ascii("midifying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("ironicaly"), &["ironically"]),
            (dictgen::InsensitiveStr::Ascii("excelent"), &["excellent"]),
            (
                dictgen::InsensitiveStr::Ascii("creationsim"),
                &["creationism"],
            ),
            (dictgen::InsensitiveStr::Ascii("serailse"), &["serialise"]),
            (
                dictgen::InsensitiveStr::Ascii("collaborant"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("toghter"), &["together"]),
            (
                dictgen::InsensitiveStr::Ascii("completiom"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concervation"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("forutunately"),
                &["fortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeend"), &["append"]),
            (
                dictgen::InsensitiveStr::Ascii("approximite"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplissing"),
                &["accomplishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuillity"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("wednesdy"), &["wednesdays"]),
            (
                dictgen::InsensitiveStr::Ascii("patitions"),
                &["partitions", "petitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lenghtening"),
                &["lengthening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("levetating"),
                &["levitating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplinerad"),
                &["disciplined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acopalyptic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("hiygeinic"), &["hygienic"]),
            (
                dictgen::InsensitiveStr::Ascii("orietation"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nigthlife"), &["nightlife"]),
            (
                dictgen::InsensitiveStr::Ascii("generoisty"),
                &["generosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("atention"), &["attention"]),
            (
                dictgen::InsensitiveStr::Ascii("electromagnectic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contolllers"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolisation"),
                &["capitalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("benchamrk"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("persecutin"),
                &["persecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rquests"),
                &["requests", "quests"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restort"),
                &["restore", "restored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provedd"),
                &["proved", "provided"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hightlighters"),
                &["highlighters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circulatiing"),
                &["circulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqueus"), &["aqueous"]),
            (
                dictgen::InsensitiveStr::Ascii("apperciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("blatanlty"), &["blatantly"]),
            (dictgen::InsensitiveStr::Ascii("internetbs"), &["internets"]),
            (dictgen::InsensitiveStr::Ascii("thereian"), &["therein"]),
            (
                dictgen::InsensitiveStr::Ascii("defintitive"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("hwat"), &["what"]),
            (dictgen::InsensitiveStr::Ascii("attemptes"), &["attempts"]),
            (dictgen::InsensitiveStr::Ascii("constanst"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("foirefox"), &["firefox"]),
            (dictgen::InsensitiveStr::Ascii("vlaid"), &["valid"]),
            (dictgen::InsensitiveStr::Ascii("bouyancy"), &["buoyancy"]),
            (dictgen::InsensitiveStr::Ascii("formely"), &["formerly"]),
            (
                dictgen::InsensitiveStr::Ascii("devisions"),
                &["divisions", "decisions", "revisions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punishmnet"),
                &["punishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("ongoign"), &["ongoing"]),
            (
                dictgen::InsensitiveStr::Ascii("cumilative"),
                &["cumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechnic"), &["mechanic"]),
            (dictgen::InsensitiveStr::Ascii("summonr"), &["summoner"]),
            (dictgen::InsensitiveStr::Ascii("haristyle"), &["hairstyle"]),
            (dictgen::InsensitiveStr::Ascii("mahcines"), &["machines"]),
            (dictgen::InsensitiveStr::Ascii("moziila"), &["mozilla"]),
            (dictgen::InsensitiveStr::Ascii("affinties"), &["affinities"]),
            (dictgen::InsensitiveStr::Ascii("arsenaal"), &["arsenal"]),
            (dictgen::InsensitiveStr::Ascii("peding"), &["pending"]),
            (dictgen::InsensitiveStr::Ascii("rattlin"), &["rattling"]),
            (dictgen::InsensitiveStr::Ascii("veryfy"), &["verify"]),
            (
                dictgen::InsensitiveStr::Ascii("distinctily"),
                &["distinctly"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifiying"), &["unifying"]),
            (dictgen::InsensitiveStr::Ascii("emtying"), &["emptying"]),
            (
                dictgen::InsensitiveStr::Ascii("inormation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroled"), &["scrolled"]),
            (
                dictgen::InsensitiveStr::Ascii("therepeutic"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worring"),
                &["working", "worrying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrisponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("meidcare"), &["medicare"]),
            (
                dictgen::InsensitiveStr::Ascii("expodentially"),
                &["exponentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("spiltter"), &["splitter"]),
            (dictgen::InsensitiveStr::Ascii("pachaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("separtations"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mappaings"), &["mappings"]),
            (dictgen::InsensitiveStr::Ascii("guantano"), &["guantanamo"]),
            (dictgen::InsensitiveStr::Ascii("verex"), &["vertex"]),
            (
                dictgen::InsensitiveStr::Ascii("interaktive"),
                &["interactive", "interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creationg"),
                &["creation", "creating"],
            ),
            (dictgen::InsensitiveStr::Ascii("slase"), &["slash"]),
            (
                dictgen::InsensitiveStr::Ascii("conventioal"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapening"), &["happening"]),
            (
                dictgen::InsensitiveStr::Ascii("checckpoint"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sylabuses"),
                &["syllabuses", "syllabi"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinematagrophy"),
                &["cinematography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alltogetrher"),
                &["altogether"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiififers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depandently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("skpetic"), &["skeptic"]),
            (dictgen::InsensitiveStr::Ascii("oxdiation"), &["oxidation"]),
            (
                dictgen::InsensitiveStr::Ascii("downviting"),
                &["downvoting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutionel"),
                &["institutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reslection"),
                &["reselection"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedos"), &["freedoms"]),
            (dictgen::InsensitiveStr::Ascii("protecton"), &["protection"]),
            (
                dictgen::InsensitiveStr::Ascii("assiciated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politicien"),
                &["politician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modificato"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("oylmpic"), &["olympic"]),
            (dictgen::InsensitiveStr::Ascii("deficienct"), &["deficient"]),
            (dictgen::InsensitiveStr::Ascii("relatabe"), &["relatable"]),
            (
                dictgen::InsensitiveStr::Ascii("conservaties"),
                &["conservatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("characteor"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("positition"), &["position"]),
            (dictgen::InsensitiveStr::Unicode("сontains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("prevoiusly"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("groosomely"),
                &["gruesomely"],
            ),
            (dictgen::InsensitiveStr::Ascii("alements"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("numbesr"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("fullfil"), &["fulfil"]),
            (dictgen::InsensitiveStr::Ascii("yieldin"), &["yielding"]),
            (dictgen::InsensitiveStr::Ascii("snwoden"), &["snowden"]),
            (
                dictgen::InsensitiveStr::Ascii("mintored"),
                &["mentored", "monitored"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispacth"), &["dispatch"]),
            (dictgen::InsensitiveStr::Ascii("tommarow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("correespondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obejctives"),
                &["objectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trancevier"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("atrribute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("interasted"),
                &["interacted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeuts"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("followwing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("finising"), &["finishing"]),
            (
                dictgen::InsensitiveStr::Ascii("retramsmissions"),
                &["retransmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qoutations"),
                &["quotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automodorator"),
                &["automoderator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncontrained"),
                &["unconstrained"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkcbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("launguage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("granparent"),
                &["grandparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimpho"), &["nympho"]),
            (dictgen::InsensitiveStr::Ascii("loggging"), &["logging"]),
            (dictgen::InsensitiveStr::Ascii("goldbeg"), &["goldberg"]),
            (dictgen::InsensitiveStr::Ascii("siezable"), &["sizable"]),
            (
                dictgen::InsensitiveStr::Ascii("inconstitutional"),
                &["unconstitutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marketpalces"),
                &["marketplaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("populaiton"),
                &["populations"],
            ),
            (dictgen::InsensitiveStr::Ascii("smoeone"), &["someone"]),
            (dictgen::InsensitiveStr::Ascii("currebt"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("polulating"),
                &["populating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assumtions"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occurrencs"),
                &["occurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectaul"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("denomals"), &["denormals"]),
            (
                dictgen::InsensitiveStr::Ascii("qeuestionnaire"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("captainers"), &["captains"]),
            (dictgen::InsensitiveStr::Ascii("excplicit"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("fisrt"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("critque"), &["critique"]),
            (dictgen::InsensitiveStr::Ascii("theery"), &["theory"]),
            (dictgen::InsensitiveStr::Ascii("underglo"), &["undergo"]),
            (dictgen::InsensitiveStr::Ascii("supposde"), &["supposed"]),
            (dictgen::InsensitiveStr::Ascii("hygienne"), &["hygiene"]),
            (
                dictgen::InsensitiveStr::Ascii("intialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("depoyment"), &["deployment"]),
            (dictgen::InsensitiveStr::Ascii("intendet"), &["intended"]),
            (
                dictgen::InsensitiveStr::Ascii("autherized"),
                &["authorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("buhddist"), &["buddhist"]),
            (dictgen::InsensitiveStr::Ascii("stataion"), &["station"]),
            (dictgen::InsensitiveStr::Ascii("impuning"), &["impugning"]),
            (dictgen::InsensitiveStr::Ascii("activeted"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("porvids"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("wonderfull"), &["wonderful"]),
            (dictgen::InsensitiveStr::Ascii("serachs"), &["searches"]),
            (
                dictgen::InsensitiveStr::Ascii("snowbaling"),
                &["snowballing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heigthened"),
                &["heightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropriatin"),
                &["appropriation", "appropriating"],
            ),
            (dictgen::InsensitiveStr::Ascii("postiives"), &["positives"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatley"),
                &["alternately"],
            ),
            (dictgen::InsensitiveStr::Ascii("merhcant"), &["merchant"]),
            (
                dictgen::InsensitiveStr::Ascii("erverything"),
                &["everything"],
            ),
            (dictgen::InsensitiveStr::Ascii("agravates"), &["aggravates"]),
            (
                dictgen::InsensitiveStr::Ascii("personallity"),
                &["personally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependends"),
                &["depended", "depends", "dependents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skopped"),
                &["shopped", "skipped", "slopped", "stopped"],
            ),
            (dictgen::InsensitiveStr::Ascii("monitring"), &["monitoring"]),
            (
                dictgen::InsensitiveStr::Ascii("temperatrue"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenfied"), &["identified"]),
            (
                dictgen::InsensitiveStr::Ascii("attriburtes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("regultory"), &["regulatory"]),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonises"),
                &["synchronises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunetly"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevidable"),
                &["inevitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("overlaped"), &["overlapped"]),
            (dictgen::InsensitiveStr::Ascii("treak"), &["treat", "tweak"]),
            (dictgen::InsensitiveStr::Ascii("hyerarchy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("disgracful"),
                &["disgraceful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwriten"),
                &["overwritten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributino"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("voyouristic"),
                &["voyeuristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outpust"),
                &["output", "outputs"],
            ),
            (dictgen::InsensitiveStr::Ascii("iternated"), &["iterated"]),
            (dictgen::InsensitiveStr::Ascii("grahpic"), &["graphic"]),
            (
                dictgen::InsensitiveStr::Ascii("inexperieced"),
                &["inexperienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscriping"),
                &["subscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalleence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resturn"),
                &["return", "returns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ubiquotous"),
                &["ubiquitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapoliate"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("efficienct"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("anomisity"), &["animosity"]),
            (
                dictgen::InsensitiveStr::Ascii("cleas"),
                &["cleans", "clears", "cleats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("personalitits"),
                &["personalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differentiaton"),
                &["differentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("extenally"), &["externally"]),
            (dictgen::InsensitiveStr::Ascii("avod"), &["avoid"]),
            (dictgen::InsensitiveStr::Ascii("lisensing"), &["licensing"]),
            (
                dictgen::InsensitiveStr::Ascii("multupliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("calculed"), &["calculated"]),
            (
                dictgen::InsensitiveStr::Ascii("involnerable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("divisability"),
                &["divisibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("voluntarilly"),
                &["voluntarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("intenst"), &["intents"]),
            (
                dictgen::InsensitiveStr::Ascii("technoligically"),
                &["technologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transprently"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("utlimatum"), &["ultimatum"]),
            (
                dictgen::InsensitiveStr::Ascii("embarassingly"),
                &["embarrassingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpested"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seens"),
                &["seems", "seen", "scenes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concertmate"),
                &["concentrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaboritave"),
                &["collaborative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blackbeary"),
                &["blackberry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appraoches"),
                &["approaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumimg"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("everets"), &["everest"]),
            (dictgen::InsensitiveStr::Ascii("cotents"), &["contents"]),
            (
                dictgen::InsensitiveStr::Ascii("bobybuilding"),
                &["bodybuilding"],
            ),
            (dictgen::InsensitiveStr::Ascii("caled"), &["called"]),
            (dictgen::InsensitiveStr::Ascii("apprently"), &["apparently"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowiedging"),
                &["acknowledging"],
            ),
            (dictgen::InsensitiveStr::Ascii("tjhe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("intuituvely"),
                &["intuitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("absurdley"), &["absurdly"]),
            (dictgen::InsensitiveStr::Ascii("neighbbors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("endpint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("orginally"), &["originally"]),
            (dictgen::InsensitiveStr::Ascii("validted"), &["validated"]),
            (dictgen::InsensitiveStr::Ascii("politican"), &["politician"]),
            (dictgen::InsensitiveStr::Ascii("conneting"), &["connecting"]),
            (dictgen::InsensitiveStr::Ascii("reporing"), &["reporting"]),
            (dictgen::InsensitiveStr::Ascii("wilh"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("syntethics"),
                &["synthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("unroken"), &["unbroken"]),
            (
                dictgen::InsensitiveStr::Ascii("credentaisl"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("rertieve"), &["retrieve"]),
            (
                dictgen::InsensitiveStr::Ascii("cockroachs"),
                &["cockroaches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tengentially"),
                &["tangentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interveign"),
                &["intervening"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceeted"), &["conceited"]),
            (dictgen::InsensitiveStr::Ascii("maylasian"), &["malaysian"]),
            (
                dictgen::InsensitiveStr::Ascii("insue"),
                &["ensue", "insure"],
            ),
            (dictgen::InsensitiveStr::Ascii("tufure"), &["future"]),
            (dictgen::InsensitiveStr::Ascii("maped"), &["mapped"]),
            (dictgen::InsensitiveStr::Ascii("cusomised"), &["customised"]),
            (
                dictgen::InsensitiveStr::Ascii("cartriages"),
                &["cartridges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trinagulates"),
                &["triangulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("verticies"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("eforceable"),
                &["enforceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipeplines"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("partiarchal"),
                &["patriarchal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("horzions"), &["horizons"]),
            (dictgen::InsensitiveStr::Ascii("sendign"), &["sending"]),
            (
                dictgen::InsensitiveStr::Ascii("disription"),
                &["disruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreacited"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("permuatate"), &["permutate"]),
            (
                dictgen::InsensitiveStr::Ascii("commpessed"),
                &["compressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparaters"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independtly"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disproportiante"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("inocence"), &["innocence"]),
            (
                dictgen::InsensitiveStr::Ascii("ascnesions"),
                &["ascensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passwors"),
                &["passwords", "password"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rimanisses"),
                &["reminisces"],
            ),
            (dictgen::InsensitiveStr::Ascii("igoned"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("routeguild"),
                &["routeguide"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premptively"),
                &["preemptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("ralative"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("moleststion"),
                &["molestation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derprecated"),
                &["deprecated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understans"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("decosed"), &["decoded"]),
            (
                dictgen::InsensitiveStr::Ascii("alhpabetical"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amendmet"),
                &["amendments", "amendment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intracicies"),
                &["intricacies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpacted"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitation"),
                &["mutilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuschia"), &["fuchsia"]),
            (dictgen::InsensitiveStr::Ascii("analizing"), &["analyzing"]),
            (
                dictgen::InsensitiveStr::Ascii("repositiory"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("coregated"), &["corrugated"]),
            (
                dictgen::InsensitiveStr::Ascii("distane"),
                &["distance", "distant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requirested"),
                &["requested", "required"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutted"),
                &["cut", "cutter", "gutted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpexctedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structired"),
                &["structured"],
            ),
            (dictgen::InsensitiveStr::Ascii("developp"), &["develop"]),
            (
                dictgen::InsensitiveStr::Ascii("ingration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("chocolet"), &["chocolates"]),
            (dictgen::InsensitiveStr::Ascii("tripple"), &["triple"]),
            (
                dictgen::InsensitiveStr::Ascii("reasearchers"),
                &["researchers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsustainble"),
                &["unsustainable"],
            ),
            (dictgen::InsensitiveStr::Ascii("perharps"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("dissmises"), &["dismisses"]),
            (dictgen::InsensitiveStr::Ascii("execte"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("smoething"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("expement"), &["experiment"]),
            (dictgen::InsensitiveStr::Ascii("dealine"), &["deadline"]),
            (
                dictgen::InsensitiveStr::Ascii("successsive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("associacte"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("lenghten"), &["lengthen"]),
            (
                dictgen::InsensitiveStr::Ascii("conclussive"),
                &["conclusive"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacakaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("checkpointusa"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funchtionnal"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conchusally"),
                &["consciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acedamically"),
                &["academically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iversed"),
                &["inverse", "inversed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emberrassment"),
                &["embarrassment"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("argumengs"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("mainstreem"),
                &["mainstream"],
            ),
            (dictgen::InsensitiveStr::Ascii("activetes"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("appliactions"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("interrtup"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("aproove"), &["approve"]),
            (dictgen::InsensitiveStr::Ascii("conectix"), &["connectix"]),
            (
                dictgen::InsensitiveStr::Ascii("advertiement"),
                &["advertisement"],
            ),
            (dictgen::InsensitiveStr::Ascii("embassey"), &["embassy"]),
            (
                dictgen::InsensitiveStr::Ascii("impracticle"),
                &["impractical"],
            ),
            (dictgen::InsensitiveStr::Ascii("synax"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("unick"), &["eunuch", "unix"]),
            (dictgen::InsensitiveStr::Ascii("refrenced"), &["referenced"]),
            (
                dictgen::InsensitiveStr::Ascii("transmittion"),
                &["transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("caleed"), &["called"]),
            (dictgen::InsensitiveStr::Ascii("mashines"), &["machines"]),
            (dictgen::InsensitiveStr::Ascii("universell"), &["universal"]),
            (dictgen::InsensitiveStr::Ascii("itereator"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("skippning"), &["skipping"]),
            (
                dictgen::InsensitiveStr::Ascii("canonally"),
                &["canonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertanity"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeturns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("musels"),
                &["muscles", "mussels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementeation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pareent"), &["parent"]),
            (dictgen::InsensitiveStr::Ascii("srearch"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("seamlessley"),
                &["seamlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convencience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boootloader"),
                &["bootloader"],
            ),
            (dictgen::InsensitiveStr::Ascii("somwhat"), &["somewhat"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonisers"),
                &["synchronisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonination"),
                &["denominations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gererator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("enoucnter"), &["encounter"]),
            (dictgen::InsensitiveStr::Ascii("beeter"), &["better"]),
            (
                dictgen::InsensitiveStr::Ascii("unvelievable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("asumed"), &["assumed"]),
            (dictgen::InsensitiveStr::Ascii("maltesian"), &["maltese"]),
            (dictgen::InsensitiveStr::Ascii("bcause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("scropion"), &["scorpion"]),
            (dictgen::InsensitiveStr::Ascii("boradwalk"), &["boardwalk"]),
            (dictgen::InsensitiveStr::Ascii("caughty"), &["caught"]),
            (dictgen::InsensitiveStr::Ascii("bealtes"), &["beatles"]),
            (dictgen::InsensitiveStr::Ascii("russion"), &["russian"]),
            (dictgen::InsensitiveStr::Ascii("piecewiese"), &["piecewise"]),
            (dictgen::InsensitiveStr::Ascii("asymptopte"), &["asymptote"]),
            (dictgen::InsensitiveStr::Ascii("arbiturary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("reductibles"),
                &["deductibles"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacnning"), &["scanning"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimiliar"),
                &["dissimilar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminitratively"),
                &["administratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("succecful"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("hirarchy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("leftsits"), &["leftists"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprenure"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("kicthens"), &["kitchens"]),
            (
                dictgen::InsensitiveStr::Ascii("abondone"),
                &["abandoned", "abandon"],
            ),
            (dictgen::InsensitiveStr::Ascii("figgured"), &["figured"]),
            (
                dictgen::InsensitiveStr::Ascii("transofrmation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("orginiate"), &["originate"]),
            (
                dictgen::InsensitiveStr::Ascii("descritptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanatization"),
                &["sanitization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferenc"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqencing"), &["sequencing"]),
            (
                dictgen::InsensitiveStr::Ascii("expendible"),
                &["expendable"],
            ),
            (dictgen::InsensitiveStr::Ascii("rerunnin"), &["rerunning"]),
            (dictgen::InsensitiveStr::Ascii("begnals"), &["bengals"]),
            (dictgen::InsensitiveStr::Ascii("nothigng"), &["nothing"]),
            (
                dictgen::InsensitiveStr::Ascii("itialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massachucetts"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("follder"), &["folder"]),
            (dictgen::InsensitiveStr::Ascii("touchs"), &["touches"]),
            (dictgen::InsensitiveStr::Ascii("exchages"), &["exchanges"]),
            (
                dictgen::InsensitiveStr::Ascii("applicatio"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkaround"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("basnhee"), &["banshee"]),
            (dictgen::InsensitiveStr::Ascii("kidnappin"), &["kidnapping"]),
            (dictgen::InsensitiveStr::Ascii("woould"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("configations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("comptued"), &["computed"]),
            (
                dictgen::InsensitiveStr::Ascii("iformation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("csae"), &["case"]),
            (
                dictgen::InsensitiveStr::Ascii("excists"),
                &["exists", "excises"],
            ),
            (dictgen::InsensitiveStr::Ascii("entirelly"), &["entirely"]),
            (dictgen::InsensitiveStr::Ascii("comemnted"), &["commented"]),
            (dictgen::InsensitiveStr::Ascii("compluter"), &["computer"]),
            (
                dictgen::InsensitiveStr::Ascii("directoories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtitectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("burrtios"), &["burritos"]),
            (
                dictgen::InsensitiveStr::Ascii("coprocesor"),
                &["coprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("leiutenant"),
                &["lieutenant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libretarianism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("evenlope"), &["envelope"]),
            (
                dictgen::InsensitiveStr::Ascii("synchornously"),
                &["synchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("dawrves"), &["dwarves"]),
            (
                dictgen::InsensitiveStr::Ascii("portuguesse"),
                &["portuguese"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("happenend"), &["happened"]),
            (
                dictgen::InsensitiveStr::Ascii("whitout"),
                &["without", "whiteout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essencials"),
                &["essentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilplications"),
                &["multiplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("pinancle"), &["pinnacle"]),
            (dictgen::InsensitiveStr::Ascii("macthes"), &["matches"]),
            (dictgen::InsensitiveStr::Ascii("nitches"), &["niches"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtution"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intoxocated"),
                &["intoxicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("regsion"), &["region"]),
            (dictgen::InsensitiveStr::Ascii("panthen"), &["pantheon"]),
            (
                dictgen::InsensitiveStr::Ascii("susbcription"),
                &["subscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("teraforms"), &["terraforms"]),
            (dictgen::InsensitiveStr::Ascii("architext"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("decremenet"), &["decrement"]),
            (
                dictgen::InsensitiveStr::Ascii("asymptoptically"),
                &["asymptotically"],
            ),
            (dictgen::InsensitiveStr::Ascii("caleee"), &["callee"]),
            (
                dictgen::InsensitiveStr::Ascii("definintely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extiting"),
                &["exciting", "exiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("escaltion"), &["escalation"]),
            (dictgen::InsensitiveStr::Ascii("censable"), &["sensible"]),
            (dictgen::InsensitiveStr::Ascii("cehcking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("trancoder"), &["transcoder"]),
            (
                dictgen::InsensitiveStr::Ascii("antisipation"),
                &["anticipation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capbability"),
                &["capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articualte"),
                &["articulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firghtened"),
                &["frightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitons"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("btye"), &["byte"]),
            (dictgen::InsensitiveStr::Ascii("vectros"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("icluding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("vereigated"),
                &["variegated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosquitto"), &["mosquito"]),
            (dictgen::InsensitiveStr::Ascii("exstream"), &["extreme"]),
            (
                dictgen::InsensitiveStr::Ascii("alcoholicos"),
                &["alcoholics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("migating"),
                &["migrating", "mitigating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accceptable"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contatining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoledgements"),
                &["acknowledgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reitterates"),
                &["reiterates"],
            ),
            (dictgen::InsensitiveStr::Ascii("presitge"), &["prestige"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamently"),
                &["fundamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("corwbar"), &["crowbar"]),
            (
                dictgen::InsensitiveStr::Ascii("cerify"),
                &["certify", "verify"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicfic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("avalibale"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("incrieases"), &["increases"]),
            (dictgen::InsensitiveStr::Ascii("agreee"), &["agree"]),
            (dictgen::InsensitiveStr::Ascii("eeeprom"), &["eeprom"]),
            (
                dictgen::InsensitiveStr::Ascii("hiarchical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("hutnsman"), &["huntsman"]),
            (
                dictgen::InsensitiveStr::Ascii("developemnts"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proficinecy"),
                &["proficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependenceny"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchache"), &["uncache"]),
            (
                dictgen::InsensitiveStr::Ascii("democrocies"),
                &["democracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("binanary"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasembles"),
                &["disassembles"],
            ),
            (dictgen::InsensitiveStr::Ascii("gylphs"), &["glyphs"]),
            (
                dictgen::InsensitiveStr::Ascii("constratin"),
                &["constraint", "constrain"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechisms"), &["mechanisms"]),
            (
                dictgen::InsensitiveStr::Ascii("enumurations"),
                &["enumerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antibitoics"),
                &["antibiotics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specalization"),
                &["specialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continiuous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approoximation"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disurption"),
                &["disruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("inadquate"), &["inadequate"]),
            (
                dictgen::InsensitiveStr::Ascii("troubleshotting"),
                &["troubleshooting"],
            ),
            (dictgen::InsensitiveStr::Ascii("indluces"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("confiugures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exixted"),
                &["existed", "exited"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhence"), &["enhance"]),
            (dictgen::InsensitiveStr::Ascii("diamters"), &["diameters"]),
            (dictgen::InsensitiveStr::Ascii("interesect"), &["intersect"]),
            (dictgen::InsensitiveStr::Ascii("consdired"), &["considered"]),
            (dictgen::InsensitiveStr::Ascii("involvin"), &["involving"]),
            (dictgen::InsensitiveStr::Ascii("exchanches"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("wiht"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("sadistisch"), &["sadistic"]),
            (
                dictgen::InsensitiveStr::Ascii("accleration"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacificts"), &["pacifist"]),
            (dictgen::InsensitiveStr::Ascii("efffort"), &["effort"]),
            (
                dictgen::InsensitiveStr::Ascii("symetrically"),
                &["symmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eloquentely"),
                &["eloquently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envaluation"),
                &["evaluation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitin"),
                &["competitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("binaies"), &["binaries"]),
            (
                dictgen::InsensitiveStr::Ascii("interationalists"),
                &["internationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("liberato"), &["liberation"]),
            (dictgen::InsensitiveStr::Ascii("imporant"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("accomadate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countrying"),
                &["countering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailzes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acheviement"),
                &["achievement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispostion"),
                &["disposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbedded"), &["embedded"]),
            (
                dictgen::InsensitiveStr::Ascii("inadeqaute"),
                &["inadequate"],
            ),
            (dictgen::InsensitiveStr::Ascii("deivative"), &["derivative"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallally"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shortcommings"),
                &["shortcomings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visuabises"),
                &["visualises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presbaterien"),
                &["presbyterian"],
            ),
            (dictgen::InsensitiveStr::Ascii("profilin"), &["profiling"]),
            (dictgen::InsensitiveStr::Ascii("psudeo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("absoluvely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("wokrflow"), &["workflow"]),
            (dictgen::InsensitiveStr::Ascii("properies"), &["properties"]),
            (
                dictgen::InsensitiveStr::Ascii("contraceptivo"),
                &["contraception"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorothmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperetely"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("bullyan"), &["bouillon"]),
            (
                dictgen::InsensitiveStr::Ascii("refferes"),
                &["refers", "referees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterms"),
                &["terms", "items", "interns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("broadcasing"),
                &["broadcasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfaction"),
                &["perfection"],
            ),
            (dictgen::InsensitiveStr::Ascii("connecetd"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("inconistency"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repet"),
                &["repeat", "repent", "repel", "reset"],
            ),
            (dictgen::InsensitiveStr::Ascii("universets"), &["universes"]),
            (dictgen::InsensitiveStr::Ascii("cheecked"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("meaurers"), &["measurers"]),
            (dictgen::InsensitiveStr::Ascii("shapsnot"), &["snapshot"]),
            (
                dictgen::InsensitiveStr::Ascii("antibioitics"),
                &["antibiotics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigle"),
                &["single", "sigil"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronoucning"),
                &["pronouncing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporpriately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacinates"), &["vaccinates"]),
            (dictgen::InsensitiveStr::Ascii("smehow"), &["somehow"]),
            (dictgen::InsensitiveStr::Ascii("trys"), &["tries"]),
            (dictgen::InsensitiveStr::Ascii("cofrim"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("standalonw"),
                &["standalone"],
            ),
            (dictgen::InsensitiveStr::Ascii("medai"), &["media"]),
            (dictgen::InsensitiveStr::Ascii("javasript"), &["javascript"]),
            (dictgen::InsensitiveStr::Ascii("reflet"), &["reflect"]),
            (dictgen::InsensitiveStr::Ascii("preselet"), &["preselect"]),
            (dictgen::InsensitiveStr::Ascii("ierland"), &["ireland"]),
            (dictgen::InsensitiveStr::Ascii("downvore"), &["downvoters"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegements"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("includign"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("futur"), &["future"]),
            (
                dictgen::InsensitiveStr::Ascii("transalting"),
                &["translating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomapatibility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bakend"),
                &["backend", "baked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perdominantly"),
                &["predominantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("lyche"), &["lychee"]),
            (
                dictgen::InsensitiveStr::Ascii("concanetion"),
                &["concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helikopter"),
                &["helicopter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenricates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("rasies"), &["raises"]),
            (dictgen::InsensitiveStr::Ascii("aborb"), &["absorb"]),
            (dictgen::InsensitiveStr::Ascii("shrpanel"), &["shrapnel"]),
            (
                dictgen::InsensitiveStr::Ascii("perticipation"),
                &["participation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execulatble"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("archtitect"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("gpysies"), &["gypsies"]),
            (dictgen::InsensitiveStr::Ascii("formware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("crystalus"), &["crystals"]),
            (dictgen::InsensitiveStr::Ascii("tarbals"), &["tarballs"]),
            (
                dictgen::InsensitiveStr::Ascii("oblitirated"),
                &["obliterated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrtips"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("implemation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("queitly"), &["quietly"]),
            (dictgen::InsensitiveStr::Ascii("peferable"), &["preferable"]),
            (
                dictgen::InsensitiveStr::Ascii("transctional"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("assinees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("otation"),
                &["notation", "rotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("invoekr"), &["invoker"]),
            (dictgen::InsensitiveStr::Ascii("thoroughy"), &["thoroughly"]),
            (
                dictgen::InsensitiveStr::Ascii("myitereator"),
                &["myiterator"],
            ),
            (dictgen::InsensitiveStr::Ascii("reponds"), &["responds"]),
            (
                dictgen::InsensitiveStr::Ascii("distrobute"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expressiong"),
                &["expressing", "expression"],
            ),
            (dictgen::InsensitiveStr::Ascii("clasues"), &["clauses"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhooding"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prophocies"),
                &["prophecies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphoricly"),
                &["metaphorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critizized"),
                &["criticized"],
            ),
            (dictgen::InsensitiveStr::Ascii("dopper"), &["doppler"]),
            (dictgen::InsensitiveStr::Ascii("mdoeling"), &["modeling"]),
            (dictgen::InsensitiveStr::Ascii("carthesian"), &["cartesian"]),
            (
                dictgen::InsensitiveStr::Ascii("trianglulating"),
                &["triangulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("versionms"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("dustification"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coommenting"),
                &["commenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("substarte"), &["substrate"]),
            (
                dictgen::InsensitiveStr::Ascii("serebrally"),
                &["cerebrally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compaibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcuate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("deocrations"),
                &["decorations"],
            ),
            (dictgen::InsensitiveStr::Ascii("desuction"), &["seduction"]),
            (dictgen::InsensitiveStr::Ascii("brwe"), &["brew"]),
            (dictgen::InsensitiveStr::Ascii("runninng"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("scrolll"), &["scroll"]),
            (dictgen::InsensitiveStr::Ascii("frogiven"), &["forgiven"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifires"),
                &["qualifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("delimite"), &["delimit"]),
            (dictgen::InsensitiveStr::Ascii("qubic"), &["cubic", "qubit"]),
            (
                dictgen::InsensitiveStr::Ascii("streights"),
                &["straights", "straits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalutaions"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearte"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("parammeter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("profundly"), &["profoundly"]),
            (
                dictgen::InsensitiveStr::Ascii("preparetion"),
                &["preparation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierarhcies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("accptably"), &["acceptably"]),
            (
                dictgen::InsensitiveStr::Ascii("personalitites"),
                &["personalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prinicipals"),
                &["principals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indixes"),
                &["indexes", "indices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretions"),
                &["interpretations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pozitives"), &["positives"]),
            (
                dictgen::InsensitiveStr::Ascii("agriculure"),
                &["agriculture"],
            ),
            (dictgen::InsensitiveStr::Ascii("broadwalk"), &["boardwalk"]),
            (
                dictgen::InsensitiveStr::Ascii("thes"),
                &["this", "these", "the"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lieklihood"),
                &["likelihood"],
            ),
            (dictgen::InsensitiveStr::Ascii("underlflow"), &["underflow"]),
            (dictgen::InsensitiveStr::Ascii("amung"), &["among"]),
            (
                dictgen::InsensitiveStr::Ascii("congigures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("yementite"),
                &["yemenite", "yemeni"],
            ),
            (dictgen::InsensitiveStr::Ascii("trival"), &["trivial"]),
            (
                dictgen::InsensitiveStr::Ascii("ameelarating"),
                &["ameliorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntetize"), &["synthetize"]),
            (
                dictgen::InsensitiveStr::Ascii("typorgapher"),
                &["typographer"],
            ),
            (dictgen::InsensitiveStr::Ascii("fpor"), &["for"]),
            (
                dictgen::InsensitiveStr::Ascii("understadn"),
                &["understands", "understand"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprhaned"), &["orphaned"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropraitely"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("srtings"), &["strings"]),
            (
                dictgen::InsensitiveStr::Ascii("accomondates"),
                &["accommodates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restructurin"),
                &["restructuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirachically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("formmated"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("negoate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("resoves"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("beautyfully"),
                &["beautifully"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcessing"), &["processing"]),
            (dictgen::InsensitiveStr::Ascii("aroun"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("geograpy"), &["geography"]),
            (
                dictgen::InsensitiveStr::Ascii("constituem"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("behviour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("cathces"), &["catches"]),
            (dictgen::InsensitiveStr::Ascii("priavte"), &["private"]),
            (dictgen::InsensitiveStr::Ascii("tru"), &["thru", "true"]),
            (dictgen::InsensitiveStr::Ascii("prozess"), &["process"]),
            (dictgen::InsensitiveStr::Ascii("valididty"), &["validity"]),
            (dictgen::InsensitiveStr::Ascii("secnarios"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("commpresion"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("requring"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("custumer"), &["customer"]),
            (dictgen::InsensitiveStr::Ascii("arrangte"), &["arrange"]),
            (dictgen::InsensitiveStr::Ascii("unknwon"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("uncompresing"),
                &["uncompressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("peraphs"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("detroying"), &["destroying"]),
            (
                dictgen::InsensitiveStr::Ascii("beahvioral"),
                &["behavioral"],
            ),
            (dictgen::InsensitiveStr::Ascii("negeated"), &["negated"]),
            (
                dictgen::InsensitiveStr::Ascii("intsallations"),
                &["installations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versatiliy"),
                &["versatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undertstand"),
                &["understand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illistrations"),
                &["illustrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("uniquey"), &["uniquely"]),
            (dictgen::InsensitiveStr::Ascii("richochet"), &["ricochet"]),
            (dictgen::InsensitiveStr::Ascii("curencies"), &["currencies"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherizations"),
                &["characterizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("conig"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("paratmers"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("burisers"), &["bruisers"]),
            (dictgen::InsensitiveStr::Ascii("inguenity"), &["ingenuity"]),
            (dictgen::InsensitiveStr::Ascii("amoung"), &["among"]),
            (dictgen::InsensitiveStr::Ascii("quntity"), &["quantity"]),
            (
                dictgen::InsensitiveStr::Ascii("surrenderd"),
                &["surrendered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connec"),
                &["connect", "conned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decoracion"),
                &["decoration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegeratian"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constituient"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarentie"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("configre"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("pubilsher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("correspontents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbrakeable"),
                &["unbreakable"],
            ),
            (dictgen::InsensitiveStr::Ascii("varitey"), &["variety"]),
            (
                dictgen::InsensitiveStr::Ascii("propaging"),
                &["propagating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("officailly"),
                &["officially"],
            ),
            (dictgen::InsensitiveStr::Ascii("sellected"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("inlcuded"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("unaccpetable"),
                &["unacceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evironments"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parital"),
                &["parietal", "partial"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprating"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("gunlsinger"),
                &["gunslinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explotions"),
                &["explosions", "explorations", "exploitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tawianese"), &["taiwanese"]),
            (dictgen::InsensitiveStr::Ascii("reserach"), &["research"]),
            (dictgen::InsensitiveStr::Ascii("reoprting"), &["reporting"]),
            (
                dictgen::InsensitiveStr::Ascii("paticipants"),
                &["participants"],
            ),
            (dictgen::InsensitiveStr::Ascii("architype"), &["archetype"]),
            (dictgen::InsensitiveStr::Ascii("suppying"), &["supplying"]),
            (dictgen::InsensitiveStr::Ascii("sylinders"), &["cylinders"]),
            (dictgen::InsensitiveStr::Ascii("planetos"), &["planets"]),
            (
                dictgen::InsensitiveStr::Ascii("coindidental"),
                &["coincidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frankenstain"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("insticnt"), &["instinct"]),
            (dictgen::InsensitiveStr::Ascii("medicae"), &["medicare"]),
            (
                dictgen::InsensitiveStr::Ascii("desparately"),
                &["desperately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmittted"),
                &["transmitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("medatadata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("encompase"), &["encompass"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapalate"),
                &["extrapolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dymanically"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterfact"),
                &["counteract"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iteraters"),
                &["iterators", "iterates"],
            ),
            (dictgen::InsensitiveStr::Ascii("marrtyred"), &["martyred"]),
            (
                dictgen::InsensitiveStr::Ascii("superficiel"),
                &["superficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("uplading"), &["uploading"]),
            (
                dictgen::InsensitiveStr::Ascii("redifintion"),
                &["redefinition"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanpus"), &["cleanups"]),
            (
                dictgen::InsensitiveStr::Ascii("transisioned"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchmkars"),
                &["benchmarks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peroidical"),
                &["periodical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ttempts"),
                &["attempts", "tempts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("additionnaly"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("godlman"), &["goldman"]),
            (
                dictgen::InsensitiveStr::Ascii("autoritative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("milicously"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concorrent"),
                &["concurrent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teatotalers"),
                &["teetotalers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotranser"), &["cotransfer"]),
            (
                dictgen::InsensitiveStr::Ascii("behing"),
                &["behind", "being"],
            ),
            (dictgen::InsensitiveStr::Ascii("inhumain"), &["inhuman"]),
            (dictgen::InsensitiveStr::Ascii("trohpies"), &["trophies"]),
            (
                dictgen::InsensitiveStr::Ascii("persisitent"),
                &["persistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectec"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("deregistred"),
                &["deregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visiters"),
                &["visitors", "visits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhilating"),
                &["annihilating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acelerated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsopported"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denpendence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("sapeenad"), &["subpoenaed"]),
            (
                dictgen::InsensitiveStr::Ascii("installatron"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("beteween"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("appologised"),
                &["apologised"],
            ),
            (dictgen::InsensitiveStr::Ascii("officals"), &["officials"]),
            (dictgen::InsensitiveStr::Ascii("mathing"), &["matching"]),
            (dictgen::InsensitiveStr::Ascii("sumamrize"), &["summarize"]),
            (
                dictgen::InsensitiveStr::Ascii("earthquate"),
                &["earthquake"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilyndres"), &["cylinders"]),
            (
                dictgen::InsensitiveStr::Ascii("pupularity"),
                &["popularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("vedio"), &["video"]),
            (
                dictgen::InsensitiveStr::Ascii("represenatations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("instsance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("accient"), &["accident"]),
            (dictgen::InsensitiveStr::Ascii("shoyld"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseuticaly"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elementray"),
                &["elementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("passivley"), &["passively"]),
            (dictgen::InsensitiveStr::Ascii("cler"), &["clear"]),
            (
                dictgen::InsensitiveStr::Ascii("parnterships"),
                &["partnerships"],
            ),
            (dictgen::InsensitiveStr::Ascii("succseeded"), &["succeeded"]),
            (dictgen::InsensitiveStr::Ascii("omnicient"), &["omniscient"]),
            (
                dictgen::InsensitiveStr::Ascii("intesection"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cacellations"),
                &["cancellations"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheras"), &["whereas"]),
            (dictgen::InsensitiveStr::Ascii("corects"), &["corrects"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotaitions"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("blatently"), &["blatantly"]),
            (
                dictgen::InsensitiveStr::Ascii("presmuably"),
                &["presumably"],
            ),
            (dictgen::InsensitiveStr::Ascii("escpable"), &["escapable"]),
            (dictgen::InsensitiveStr::Ascii("mittigate"), &["mitigate"]),
            (dictgen::InsensitiveStr::Ascii("genereal"), &["general"]),
            (
                dictgen::InsensitiveStr::Ascii("santization"),
                &["sanitization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marijuannas"),
                &["marijuana"],
            ),
            (dictgen::InsensitiveStr::Ascii("esy"), &["easy"]),
            (
                dictgen::InsensitiveStr::Ascii("condidential"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrythed"), &["writhed"]),
            (
                dictgen::InsensitiveStr::Ascii("stangers"),
                &["strangers", "stingers", "staggers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cariage"), &["carriage"]),
            (
                dictgen::InsensitiveStr::Ascii("prominantly"),
                &["prominently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hel"),
                &["help", "hell", "heal"],
            ),
            (dictgen::InsensitiveStr::Ascii("rabbitts"), &["rabbits"]),
            (dictgen::InsensitiveStr::Ascii("decalres"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("interrutpions"),
                &["interruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonynous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("nagivation"),
                &["navigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("prependet"), &["prepended"]),
            (dictgen::InsensitiveStr::Ascii("registe"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("guarenteer"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("hability"), &["ability"]),
            (dictgen::InsensitiveStr::Ascii("discovred"), &["discovered"]),
            (dictgen::InsensitiveStr::Ascii("somenone"), &["someone"]),
            (
                dictgen::InsensitiveStr::Ascii("transcribtion"),
                &["transcription"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrollade"), &["scrolled"]),
            (dictgen::InsensitiveStr::Ascii("ethniticy"), &["ethnicity"]),
            (dictgen::InsensitiveStr::Ascii("apostel"), &["apostle"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconfortably"),
                &["uncomfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculcations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interspserse"),
                &["intersperse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protectrons"),
                &["protections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unqarantined"),
                &["unquarantined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overvrites"),
                &["overwrites"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodiver"), &["provider"]),
            (
                dictgen::InsensitiveStr::Ascii("constucters"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppots"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("ignorrs"), &["ignores"]),
            (
                dictgen::InsensitiveStr::Ascii("meening"),
                &["meaning", "meeting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandanavia"),
                &["scandinavia"],
            ),
            (dictgen::InsensitiveStr::Ascii("houshold"), &["household"]),
            (
                dictgen::InsensitiveStr::Ascii("minimilast"),
                &["minimalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkload"), &["workload"]),
            (dictgen::InsensitiveStr::Ascii("standrat"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("transfromers"),
                &["transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("trickry"), &["trickery"]),
            (
                dictgen::InsensitiveStr::Ascii("hilighting"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradicion"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrbutions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("distroy"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("permulate"), &["permutate"]),
            (
                dictgen::InsensitiveStr::Ascii("delimitors"),
                &["delimiters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sotrmfront"),
                &["stormfront"],
            ),
            (dictgen::InsensitiveStr::Ascii("wathcdogs"), &["watchdogs"]),
            (
                dictgen::InsensitiveStr::Ascii("transperent"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("sucessses"), &["successes"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepeuner"),
                &["entrepreneur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gettings"),
                &["getting", "settings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nueroimaging"),
                &["neuroimaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("isothem"), &["isotherm"]),
            (dictgen::InsensitiveStr::Ascii("statring"), &["starting"]),
            (dictgen::InsensitiveStr::Ascii("programemd"), &["programed"]),
            (dictgen::InsensitiveStr::Ascii("legecies"), &["legacies"]),
            (dictgen::InsensitiveStr::Ascii("possile"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("barabric"), &["barbaric"]),
            (
                dictgen::InsensitiveStr::Ascii("barabrians"),
                &["barbarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarnato"),
                &["incarnation"],
            ),
            (dictgen::InsensitiveStr::Ascii("backslah"), &["backslash"]),
            (
                dictgen::InsensitiveStr::Ascii("exponentiel"),
                &["exponential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("numerbering"),
                &["numbering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politessen"),
                &["politeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transplats"),
                &["transplants"],
            ),
            (dictgen::InsensitiveStr::Ascii("inferioara"), &["inferior"]),
            (dictgen::InsensitiveStr::Ascii("repalaces"), &["replaces"]),
            (dictgen::InsensitiveStr::Ascii("nihilisim"), &["nihilism"]),
            (
                dictgen::InsensitiveStr::Ascii("possbilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fulfillling"),
                &["fulfilling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivduals"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("encoedes"), &["encodes"]),
            (
                dictgen::InsensitiveStr::Ascii("accummulated"),
                &["accumulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("akumulators"),
                &["accumulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("cementary"), &["cemetery"]),
            (dictgen::InsensitiveStr::Ascii("dealte"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("spritn"), &["sprint"]),
            (dictgen::InsensitiveStr::Ascii("infromed"), &["informed"]),
            (dictgen::InsensitiveStr::Ascii("useers"), &["users"]),
            (dictgen::InsensitiveStr::Ascii("movemnts"), &["movements"]),
            (dictgen::InsensitiveStr::Ascii("eatswood"), &["eastwood"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelisation"),
                &["parallelisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstrcut"),
                &["reconstruct"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssignee"), &["assignee"]),
            (
                dictgen::InsensitiveStr::Ascii("expetimenters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialiasing"),
                &["specialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("sampel"), &["sample"]),
            (dictgen::InsensitiveStr::Ascii("inludes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("leightweight"),
                &["lightweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeust"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("intormation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("createable"), &["creatable"]),
            (
                dictgen::InsensitiveStr::Ascii("implausibile"),
                &["implausible"],
            ),
            (dictgen::InsensitiveStr::Ascii("piplelines"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("subscirbing"),
                &["subscribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("creativey"), &["creatively"]),
            (
                dictgen::InsensitiveStr::Ascii("additionaly"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overreaktion"),
                &["overreaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpitrated"),
                &["perpetrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("compiliers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("ambulancier"),
                &["ambulance", "ambulanceman"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("republicians"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanllenge"), &["challenge"]),
            (dictgen::InsensitiveStr::Ascii("disturbd"), &["disturbed"]),
            (
                dictgen::InsensitiveStr::Ascii("contating"),
                &["contacting", "containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documeneter"),
                &["documenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwowong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("differt"),
                &["different", "differ"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaged"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("survivabililty"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comissions"),
                &["commissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirutalisation"),
                &["devirtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("becomeing"), &["becoming"]),
            (
                dictgen::InsensitiveStr::Ascii("impleneted"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequemces"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("mimmicking"), &["mimicking"]),
            (dictgen::InsensitiveStr::Ascii("tjpanishad"), &["upanishad"]),
            (dictgen::InsensitiveStr::Ascii("modiiers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("crahses"), &["crashes"]),
            (
                dictgen::InsensitiveStr::Ascii("algororithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funtionallity"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsisting"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("connetation"),
                &["connotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathemetically"),
                &["mathematically"],
            ),
            (dictgen::InsensitiveStr::Ascii("angualr"), &["angular"]),
            (dictgen::InsensitiveStr::Ascii("procdure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("checlsum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("transactoin"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("indiate"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("leaveing"), &["leaving"]),
            (
                dictgen::InsensitiveStr::Ascii("reverece"),
                &["reference", "reverence"],
            ),
            (dictgen::InsensitiveStr::Ascii("partiotic"), &["patriotic"]),
            (
                dictgen::InsensitiveStr::Ascii("associatative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sopporter"), &["supporter"]),
            (
                dictgen::InsensitiveStr::Ascii("juggernaunt"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompressible"),
                &["incompressible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagerrates"),
                &["exaggerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notce"),
                &["notice", "notch", "note", "nonce"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechs"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("auxileries"),
                &["auxiliaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepretation"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("existings"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("responsile"),
                &["responsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucutring"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validater"),
                &["validator", "validated", "validates", "validate"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuploade"), &["reupload"]),
            (dictgen::InsensitiveStr::Ascii("krudish"), &["kurdish"]),
            (dictgen::InsensitiveStr::Ascii("tructural"), &["structural"]),
            (dictgen::InsensitiveStr::Ascii("coutns"), &["counts"]),
            (dictgen::InsensitiveStr::Ascii("platflorms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("parilament"),
                &["parliament"],
            ),
            (dictgen::InsensitiveStr::Ascii("refelxes"), &["reflexes"]),
            (dictgen::InsensitiveStr::Ascii("togther"), &["together"]),
            (dictgen::InsensitiveStr::Ascii("unknow"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("byond"), &["beyond"]),
            (
                dictgen::InsensitiveStr::Ascii("independantes"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("backround"), &["background"]),
            (
                dictgen::InsensitiveStr::Ascii("infilitrated"),
                &["infiltrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exturde"), &["extrude"]),
            (
                dictgen::InsensitiveStr::Ascii("authobiography"),
                &["autobiography"],
            ),
            (dictgen::InsensitiveStr::Ascii("libraties"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("algoithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("asiprins"), &["aspirins"]),
            (dictgen::InsensitiveStr::Ascii("assums"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("paragrapgh"),
                &["paragraphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elligibility"),
                &["eligibility", "illegibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constituer"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("intances"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("dokcred"),
                &["dockerd", "docked", "docker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertarains"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharamceutical"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("acitivates"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("siloueted"),
                &["silhouetted"],
            ),
            (dictgen::InsensitiveStr::Ascii("dammages"), &["damages"]),
            (
                dictgen::InsensitiveStr::Ascii("statistct"),
                &["statistic", "statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("supppose"), &["suppose"]),
            (dictgen::InsensitiveStr::Ascii("cthulhlu"), &["cthulhu"]),
            (
                dictgen::InsensitiveStr::Ascii("unshfiting"),
                &["unshifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("homewrold"), &["homeworld"]),
            (
                dictgen::InsensitiveStr::Ascii("scandinaivan"),
                &["scandinavia"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixees"), &["fixes"]),
            (
                dictgen::InsensitiveStr::Ascii("hurricances"),
                &["hurricanes"],
            ),
            (dictgen::InsensitiveStr::Ascii("carriadge"), &["carriage"]),
            (dictgen::InsensitiveStr::Ascii("tribuanl"), &["tribunal"]),
            (dictgen::InsensitiveStr::Ascii("hiptsers"), &["hipsters"]),
            (
                dictgen::InsensitiveStr::Ascii("workwround"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("collaspes"), &["collapses"]),
            (dictgen::InsensitiveStr::Ascii("importes"), &["imports"]),
            (
                dictgen::InsensitiveStr::Ascii("tansferring"),
                &["transferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("functioms"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("protol"), &["protocol"]),
            (
                dictgen::InsensitiveStr::Ascii("essensials"),
                &["essentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimialrly"),
                &["dissimilarly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envrironment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjectivy"),
                &["subjectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("peircing"), &["piercing"]),
            (dictgen::InsensitiveStr::Ascii("zimpaps"), &["zipmaps"]),
            (dictgen::InsensitiveStr::Ascii("larinx"), &["larynx"]),
            (
                dictgen::InsensitiveStr::Ascii("coercin"),
                &["coercing", "coercion"],
            ),
            (dictgen::InsensitiveStr::Ascii("imrpove"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("banching"),
                &["branching", "benching"],
            ),
            (dictgen::InsensitiveStr::Ascii("plasticas"), &["plastics"]),
            (dictgen::InsensitiveStr::Ascii("innovatve"), &["innovate"]),
            (dictgen::InsensitiveStr::Ascii("cornerns"), &["corners"]),
            (
                dictgen::InsensitiveStr::Ascii("requesets"),
                &["requests", "requisites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assymmetries"),
                &["asymmetries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifield"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("trggers"), &["triggers"]),
            (dictgen::InsensitiveStr::Ascii("kwuzines"), &["cuisines"]),
            (dictgen::InsensitiveStr::Ascii("entireity"), &["entirety"]),
            (dictgen::InsensitiveStr::Ascii("minsitry"), &["ministry"]),
            (
                dictgen::InsensitiveStr::Ascii("algoritmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("notity"), &["notify"]),
            (
                dictgen::InsensitiveStr::Ascii("beforerehand"),
                &["beforehand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidentaly"),
                &["confidently"],
            ),
            (dictgen::InsensitiveStr::Ascii("acepts"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("iteator"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("eyebals"), &["eyeballs"]),
            (
                dictgen::InsensitiveStr::Ascii("unsustianable"),
                &["unsustainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimazation"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polarizaiton"),
                &["polarization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selctables"),
                &["selectables", "selectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("expertos"), &["experts"]),
            (dictgen::InsensitiveStr::Ascii("comptown"), &["compton"]),
            (
                dictgen::InsensitiveStr::Ascii("unversoned"),
                &["unversioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playgrouds"),
                &["playgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("chinesse"), &["chinese"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactuers"),
                &["manufacturers", "manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("traige"), &["triage"]),
            (
                dictgen::InsensitiveStr::Ascii("indivdualy"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custonisable"),
                &["customisable"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaper"), &["chapter"]),
            (
                dictgen::InsensitiveStr::Ascii("immedially"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimissing"), &["dismissing"]),
            (dictgen::InsensitiveStr::Ascii("thw"), &["the", "thaw"]),
            (dictgen::InsensitiveStr::Ascii("cyrto"), &["crypto"]),
            (
                dictgen::InsensitiveStr::Ascii("iterater"),
                &["iterator", "iterated", "iterates", "iterate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dircting"), &["directing"]),
            (
                dictgen::InsensitiveStr::Ascii("ramdomizers"),
                &["randomizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("atribs"), &["attribs"]),
            (
                dictgen::InsensitiveStr::Ascii("constrainsts"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cramugins"),
                &["curmudgeons"],
            ),
            (dictgen::InsensitiveStr::Ascii("sais"), &["says"]),
            (
                dictgen::InsensitiveStr::Ascii("circularlly"),
                &["circularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exececutable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("releease"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("adjatating"), &["agitating"]),
            (dictgen::InsensitiveStr::Ascii("provieding"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("playgrouns"),
                &["playgrounds", "playground"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("natioanlist"),
                &["nationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknolwedgement"),
                &["acknowledgement"],
            ),
            (dictgen::InsensitiveStr::Ascii("trintiy"), &["trinity"]),
            (
                dictgen::InsensitiveStr::Ascii("inremental"),
                &["incremental"],
            ),
            (dictgen::InsensitiveStr::Ascii("usrenames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("enviornemnts"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("targest"),
                &["target", "targets", "largest"],
            ),
            (dictgen::InsensitiveStr::Ascii("leagl"), &["legal"]),
            (
                dictgen::InsensitiveStr::Ascii("millionar"),
                &["millionaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depencendencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("authroize"), &["authorize"]),
            (dictgen::InsensitiveStr::Ascii("referreres"), &["referrers"]),
            (dictgen::InsensitiveStr::Ascii("arrys"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("rela"), &["real"]),
            (
                dictgen::InsensitiveStr::Ascii("architecuturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("entiyt"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("villans"), &["villains"]),
            (
                dictgen::InsensitiveStr::Ascii("guarrenteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumend"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("scafolding"),
                &["scaffolding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interploated"),
                &["interpolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurrance"),
                &["recurrence"],
            ),
            (dictgen::InsensitiveStr::Ascii("resaurant"), &["restaurant"]),
            (dictgen::InsensitiveStr::Ascii("androide"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("utitlity"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("exagons"), &["hexagons"]),
            (dictgen::InsensitiveStr::Ascii("featrues"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("stechiometric"),
                &["stoichiometric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticiy"),
                &["authenticity"],
            ),
            (dictgen::InsensitiveStr::Ascii("impusles"), &["impulses"]),
            (
                dictgen::InsensitiveStr::Ascii("privitized"),
                &["privatized"],
            ),
            (dictgen::InsensitiveStr::Ascii("basci"), &["basic"]),
            (dictgen::InsensitiveStr::Ascii("villageois"), &["villages"]),
            (
                dictgen::InsensitiveStr::Ascii("navigater"),
                &["navigator", "navigated", "navigates", "navigate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mediaction"),
                &["medications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpendiculaires"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("casualites"),
                &["casualties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reestructure"),
                &["restructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tructurally"),
                &["structurally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intructing"),
                &["instructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbohood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("uniforum"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("rittle"),
                &["rattle", "riddle"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechk"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("intepret"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("unregluated"),
                &["unregulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peirodicals"),
                &["periodicals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processers"),
                &["processors"],
            ),
            (dictgen::InsensitiveStr::Ascii("hafltime"), &["halftime"]),
            (dictgen::InsensitiveStr::Ascii("nutrieous"), &["nutritious"]),
            (
                dictgen::InsensitiveStr::Ascii("beleagured"),
                &["beleaguered"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperating"), &["separating"]),
            (dictgen::InsensitiveStr::Ascii("alogned"), &["aligned"]),
            (
                dictgen::InsensitiveStr::Ascii("peroendicular"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmenetations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriptue"), &["scriptures"]),
            (dictgen::InsensitiveStr::Ascii("telegramm"), &["telegram"]),
            (dictgen::InsensitiveStr::Ascii("cateogry"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("cominer"), &["combiner"]),
            (
                dictgen::InsensitiveStr::Ascii("anurism"),
                &["aneurism", "aneurysm"],
            ),
            (dictgen::InsensitiveStr::Ascii("accourdion"), &["accordion"]),
            (
                dictgen::InsensitiveStr::Ascii("inpection"),
                &["inception", "inspection"],
            ),
            (dictgen::InsensitiveStr::Ascii("trimmng"), &["trimming"]),
            (dictgen::InsensitiveStr::Ascii("sapphie"), &["sapphire"]),
            (dictgen::InsensitiveStr::Ascii("athesitic"), &["atheistic"]),
            (
                dictgen::InsensitiveStr::Ascii("spacializing"),
                &["specializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provileges"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaialable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("doctrins"), &["doctrines"]),
            (dictgen::InsensitiveStr::Ascii("reseach"), &["research"]),
            (dictgen::InsensitiveStr::Ascii("optimim"), &["optimism"]),
            (dictgen::InsensitiveStr::Ascii("acqured"), &["acquired"]),
            (
                dictgen::InsensitiveStr::Ascii("thunderolt"),
                &["thunderbolt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accociates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompletd"),
                &["incomplete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppressiun"),
                &["oppressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polymophing"),
                &["polymorphing"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnaty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("movesped"), &["movespeed"]),
            (
                dictgen::InsensitiveStr::Ascii("stregnthen"),
                &["strengthen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("citisenship"),
                &["citizenship"],
            ),
            (dictgen::InsensitiveStr::Ascii("alterato"), &["alteration"]),
            (
                dictgen::InsensitiveStr::Ascii("unsaged"),
                &["unsaved", "unstaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimses"), &["minimises"]),
            (dictgen::InsensitiveStr::Ascii("stronly"), &["strongly"]),
            (dictgen::InsensitiveStr::Ascii("shedules"), &["schedules"]),
            (dictgen::InsensitiveStr::Ascii("memwars"), &["memoirs"]),
            (dictgen::InsensitiveStr::Ascii("ond"), &["one", "and"]),
            (
                dictgen::InsensitiveStr::Ascii("experimdntation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolunely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("actvated"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("comit"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("retreives"), &["retrieves"]),
            (
                dictgen::InsensitiveStr::Ascii("imbalences"),
                &["imbalances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transievers"),
                &["transceivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("teancity"), &["tenacity"]),
            (dictgen::InsensitiveStr::Ascii("habeus"), &["habeas"]),
            (
                dictgen::InsensitiveStr::Ascii("ontainging"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("resurect"), &["resurrect"]),
            (
                dictgen::InsensitiveStr::Ascii("arranement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("insluts"), &["insults"]),
            (dictgen::InsensitiveStr::Ascii("learing"), &["learning"]),
            (dictgen::InsensitiveStr::Ascii("regsions"), &["regions"]),
            (dictgen::InsensitiveStr::Ascii("rennovate"), &["renovate"]),
            (
                dictgen::InsensitiveStr::Ascii("descrybing"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dockeed"), &["docked"]),
            (dictgen::InsensitiveStr::Ascii("minite"), &["minute"]),
            (
                dictgen::InsensitiveStr::Ascii("hipotetical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("compille"), &["compile"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepreuners"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructred"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("clibpoard"), &["clipboard"]),
            (dictgen::InsensitiveStr::Ascii("aterny"), &["attorney"]),
            (dictgen::InsensitiveStr::Ascii("themslves"), &["themselves"]),
            (
                dictgen::InsensitiveStr::Ascii("aliver"),
                &["alive", "liver", "sliver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstrumental"),
                &["instrumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("fincial"), &["financial"]),
            (dictgen::InsensitiveStr::Ascii("auxilery"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("condtion"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("epectation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definotely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrate"),
                &["aggregate", "aggravate"],
            ),
            (dictgen::InsensitiveStr::Ascii("intergers"), &["integers"]),
            (
                dictgen::InsensitiveStr::Ascii("enterprishe"),
                &["enterprises"],
            ),
            (dictgen::InsensitiveStr::Ascii("recliam"), &["reclaim"]),
            (
                dictgen::InsensitiveStr::Ascii("ramdomisers"),
                &["randomisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrumetal"),
                &["instrumental"],
            ),
            (dictgen::InsensitiveStr::Ascii("withdrawan"), &["withdrawn"]),
            (
                dictgen::InsensitiveStr::Ascii("organizaions"),
                &["organizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docmuneted"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoatttack"),
                &["autoattack"],
            ),
            (dictgen::InsensitiveStr::Ascii("assignged"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("herarchically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("eccessive"), &["excessive"]),
            (dictgen::InsensitiveStr::Ascii("spritre"), &["sprite"]),
            (
                dictgen::InsensitiveStr::Ascii("deselctable"),
                &["deselectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("correts"), &["corrects"]),
            (dictgen::InsensitiveStr::Ascii("refecting"), &["reflecting"]),
            (dictgen::InsensitiveStr::Ascii("satallite"), &["satellite"]),
            (dictgen::InsensitiveStr::Ascii("myselv"), &["myself"]),
            (dictgen::InsensitiveStr::Ascii("comtainer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("preorderd"), &["preordered"]),
            (
                dictgen::InsensitiveStr::Ascii("supllying"),
                &["supplying", "sullying"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaurer"), &["measurer"]),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedoren"),
                &["tetrahedron"],
            ),
            (dictgen::InsensitiveStr::Ascii("excliuding"), &["excluding"]),
            (
                dictgen::InsensitiveStr::Ascii("thermodynanics"),
                &["thermodynamics"],
            ),
            (dictgen::InsensitiveStr::Ascii("negociate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("aggragations"),
                &["aggregations"],
            ),
            (dictgen::InsensitiveStr::Ascii("obected"), &["objected"]),
            (dictgen::InsensitiveStr::Ascii("digitizin"), &["digitizing"]),
            (
                dictgen::InsensitiveStr::Ascii("specialaization"),
                &["specialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliped"), &["slipped"]),
            (dictgen::InsensitiveStr::Ascii("contct"), &["contact"]),
            (dictgen::InsensitiveStr::Ascii("asists"), &["assists"]),
            (
                dictgen::InsensitiveStr::Ascii("dimesnions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubsribe"),
                &["unsubscribe"],
            ),
            (dictgen::InsensitiveStr::Ascii("saveguard"), &["safeguard"]),
            (
                dictgen::InsensitiveStr::Ascii("defensivley"),
                &["defensively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passionatly"),
                &["passionately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpahabetic"),
                &["alphabetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("accuratly"), &["accurately"]),
            (dictgen::InsensitiveStr::Ascii("medicaly"), &["medically"]),
            (dictgen::InsensitiveStr::Ascii("gerogia"), &["georgia"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficiencts"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("statuser"), &["stature"]),
            (
                dictgen::InsensitiveStr::Ascii("coreespondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("imprioned"), &["imprisoned"]),
            (dictgen::InsensitiveStr::Ascii("engoug"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("bulidings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("aplications"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("imploed"), &["implode"]),
            (
                dictgen::InsensitiveStr::Ascii("southapmton"),
                &["southampton"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadvertedly"),
                &["inadvertently"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopefullly"), &["hopefully"]),
            (
                dictgen::InsensitiveStr::Ascii("libertariens"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poossibilities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqaured"), &["squared"]),
            (
                dictgen::InsensitiveStr::Ascii("maybelleine"),
                &["maybelline"],
            ),
            (dictgen::InsensitiveStr::Ascii("universial"), &["universal"]),
            (
                dictgen::InsensitiveStr::Ascii("architecutrally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("exampes"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("icrementally"),
                &["incrementally"],
            ),
            (dictgen::InsensitiveStr::Ascii("functitons"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("apppeared"), &["appeared"]),
            (dictgen::InsensitiveStr::Ascii("cloest"), &["closest"]),
            (
                dictgen::InsensitiveStr::Ascii("phenomenonly"),
                &["phenomenally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acatemically"),
                &["academically"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomrorow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("instructon"),
                &["instruction", "instructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("delimeted"), &["delimited"]),
            (
                dictgen::InsensitiveStr::Ascii("conten"),
                &["contain", "content"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertisments"),
                &["advertisements"],
            ),
            (dictgen::InsensitiveStr::Ascii("temparary"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("oportunities"),
                &["opportunities"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicards"), &["discards"]),
            (dictgen::InsensitiveStr::Ascii("intevenes"), &["intervenes"]),
            (dictgen::InsensitiveStr::Ascii("delt"), &["dealt"]),
            (dictgen::InsensitiveStr::Ascii("statits"), &["statist"]),
            (
                dictgen::InsensitiveStr::Ascii("streemlining"),
                &["streamlining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pictureskly"),
                &["picturesquely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memebrship"),
                &["membership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensevily"),
                &["extensively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilthreads"),
                &["multithreads"],
            ),
            (dictgen::InsensitiveStr::Ascii("appered"), &["appeared"]),
            (
                dictgen::InsensitiveStr::Ascii("ocasionally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependnt"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("assumong"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("algorithmn"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("dertmine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("automic"),
                &["atomic", "automatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("tarced"), &["traced"]),
            (dictgen::InsensitiveStr::Ascii("exectutes"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("detaulted"), &["defaulted"]),
            (dictgen::InsensitiveStr::Ascii("electirc"), &["electric"]),
            (
                dictgen::InsensitiveStr::Ascii("displacment"),
                &["displacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatiled"), &["detailed"]),
            (dictgen::InsensitiveStr::Ascii("nelink"), &["netlink"]),
            (dictgen::InsensitiveStr::Ascii("ceckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("oversue"), &["overuse"]),
            (
                dictgen::InsensitiveStr::Ascii("exiciting"),
                &["existing", "exciting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possiabilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incroporating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyhiscal"), &["physical"]),
            (
                dictgen::InsensitiveStr::Ascii("assignemnt"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("provding"), &["providing"]),
            (dictgen::InsensitiveStr::Ascii("uinque"), &["unique"]),
            (
                dictgen::InsensitiveStr::Ascii("critizing"),
                &["criticizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insensitve"),
                &["insensitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathces"), &["matches"]),
            (
                dictgen::InsensitiveStr::Ascii("exaplanations"),
                &["explanations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialisae"),
                &["deserialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("polgons"), &["polygons"]),
            (dictgen::InsensitiveStr::Ascii("checkelist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("coexsitence"),
                &["coexistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaible"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("processsor"),
                &["processors", "processor"],
            ),
            (dictgen::InsensitiveStr::Ascii("yaers"), &["years"]),
            (dictgen::InsensitiveStr::Ascii("loacally"), &["locally"]),
            (dictgen::InsensitiveStr::Ascii("epsiodes"), &["episodes"]),
            (
                dictgen::InsensitiveStr::Ascii("comparement"),
                &["compartment", "comparing", "comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("acuqire"), &["acquire"]),
            (
                dictgen::InsensitiveStr::Ascii("survelliance"),
                &["surveillance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecuture"),
                &["architecture"],
            ),
            (dictgen::InsensitiveStr::Ascii("deneirs"), &["deniers"]),
            (dictgen::InsensitiveStr::Ascii("hokay"), &["okay"]),
            (dictgen::InsensitiveStr::Ascii("cominers"), &["combiners"]),
            (
                dictgen::InsensitiveStr::Ascii("vritualization"),
                &["virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbos"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("patttern"), &["pattern"]),
            (
                dictgen::InsensitiveStr::Ascii("eletrically"),
                &["electrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("actul"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("uncoditionally"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomeanse"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expcet"),
                &["except", "expect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aliase"),
                &["aliases", "alias"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutated"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("insantiating"),
                &["instantiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repectively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("directry"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("trasncript"),
                &["transcript"],
            ),
            (dictgen::InsensitiveStr::Ascii("toom"), &["tomb"]),
            (dictgen::InsensitiveStr::Ascii("resul"), &["result"]),
            (dictgen::InsensitiveStr::Ascii("powderade"), &["powdered"]),
            (
                dictgen::InsensitiveStr::Ascii("communicatie"),
                &["communicate", "communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnferred"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("enducing"), &["inducing"]),
            (dictgen::InsensitiveStr::Ascii("bouced"), &["bounced"]),
            (
                dictgen::InsensitiveStr::Ascii("ckeckerboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprasial"), &["appraisal"]),
            (dictgen::InsensitiveStr::Ascii("alchoolic"), &["alcoholic"]),
            (
                dictgen::InsensitiveStr::Ascii("producting"),
                &["production", "producing"],
            ),
            (dictgen::InsensitiveStr::Ascii("asymmmetry"), &["asymmetry"]),
            (dictgen::InsensitiveStr::Ascii("conversie"), &["converse"]),
            (dictgen::InsensitiveStr::Ascii("utilses"), &["utilises"]),
            (dictgen::InsensitiveStr::Ascii("obeject"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("allocatote"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("neighburhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuella"),
                &["individual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understading"),
                &["understanding", "understating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("styless"),
                &["styles", "styleless"],
            ),
            (dictgen::InsensitiveStr::Ascii("forsees"), &["foresees"]),
            (
                dictgen::InsensitiveStr::Ascii("feasiblty"),
                &["feasibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("backeast"), &["backseat"]),
            (dictgen::InsensitiveStr::Ascii("boyant"), &["buoyant"]),
            (
                dictgen::InsensitiveStr::Ascii("resused"),
                &["reused", "refused", "resumed"],
            ),
            (dictgen::InsensitiveStr::Ascii("typcially"), &["typically"]),
            (
                dictgen::InsensitiveStr::Ascii("repare"),
                &["reparse", "repair"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teraformed"),
                &["terraformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomized"),
                &["randomized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentictors"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("competant"), &["competent"]),
            (
                dictgen::InsensitiveStr::Ascii("markerplace"),
                &["marketplace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriminator"),
                &["discriminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helicoptor"),
                &["helicopter"],
            ),
            (dictgen::InsensitiveStr::Ascii("consize"), &["concise"]),
            (dictgen::InsensitiveStr::Ascii("possitives"), &["positives"]),
            (
                dictgen::InsensitiveStr::Ascii("intentually"),
                &["intentionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psuedoinverse"),
                &["pseudoinverse"],
            ),
            (dictgen::InsensitiveStr::Ascii("durtaion"), &["duration"]),
            (dictgen::InsensitiveStr::Ascii("dungoen"), &["dungeon"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibel"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("samuria"), &["samurai"]),
            (dictgen::InsensitiveStr::Ascii("initliaze"), &["initialize"]),
            (dictgen::InsensitiveStr::Ascii("asudo"), &["sudo"]),
            (dictgen::InsensitiveStr::Ascii("unoknown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("metephysical"),
                &["metaphysical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcation"),
                &["transaction", "translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complusive"),
                &["compulsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("hihghlight"), &["highlight"]),
            (dictgen::InsensitiveStr::Ascii("cleandups"), &["cleanups"]),
            (dictgen::InsensitiveStr::Ascii("wierd"), &["weird"]),
            (
                dictgen::InsensitiveStr::Ascii("suspecious"),
                &["suspicious"],
            ),
            (dictgen::InsensitiveStr::Ascii("posiion"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("ugpraded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("unnedded"), &["unneeded"]),
            (dictgen::InsensitiveStr::Ascii("anonymus"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("bacics"), &["basics"]),
            (dictgen::InsensitiveStr::Ascii("dpeends"), &["depends"]),
            (
                dictgen::InsensitiveStr::Ascii("oscilllators"),
                &["oscillators"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaures"), &["measures"]),
            (
                dictgen::InsensitiveStr::Ascii("characteristc"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentcators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolidare"),
                &["consolidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlelating"),
                &["correlating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsitutuents"),
                &["substituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfy"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("denyed"), &["denied"]),
            (
                dictgen::InsensitiveStr::Ascii("intervined"),
                &["intervened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmetn"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfying"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("meessages"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("connecttor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyprocite"), &["hypocrite"]),
            (
                dictgen::InsensitiveStr::Ascii("prioretizes"),
                &["prioritizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyling"), &["cycling"]),
            (
                dictgen::InsensitiveStr::Ascii("experimintally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componoents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unneceesarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deyhdration"),
                &["dehydration"],
            ),
            (dictgen::InsensitiveStr::Ascii("procesing"), &["processing"]),
            (dictgen::InsensitiveStr::Ascii("zucheenei"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("brigated"), &["brigade"]),
            (
                dictgen::InsensitiveStr::Ascii("thsnkfully"),
                &["thankfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("cupon"), &["coupon"]),
            (
                dictgen::InsensitiveStr::Ascii("underneeth"),
                &["underneath"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procucer"),
                &["procurer", "producer"],
            ),
            (dictgen::InsensitiveStr::Ascii("biulder"), &["builder"]),
            (
                dictgen::InsensitiveStr::Ascii("imlpemented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("underlayed"), &["underlaid"]),
            (
                dictgen::InsensitiveStr::Ascii("enlightned"),
                &["enlightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unecesary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contitionals"),
                &["conditionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculcator"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuite"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactersistics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("sintilla"), &["scintilla"]),
            (dictgen::InsensitiveStr::Ascii("vartiables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarhcical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotnents"), &["contents"]),
            (
                dictgen::InsensitiveStr::Ascii("utilisaton"),
                &["utilisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declaritive"),
                &["declarative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alginments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completalbe"),
                &["completable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endiens"),
                &["endians", "indians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preveiwes"),
                &["previews", "previewers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doucmenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accaptance"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("coherance"), &["coherence"]),
            (dictgen::InsensitiveStr::Ascii("saclars"), &["scalars"]),
            (
                dictgen::InsensitiveStr::Ascii("recommmendation"),
                &["recommendation"],
            ),
            (dictgen::InsensitiveStr::Ascii("equeals"), &["equals"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficied"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("catchs"), &["catches"]),
            (dictgen::InsensitiveStr::Ascii("prohibitus"), &["prohibits"]),
            (
                dictgen::InsensitiveStr::Ascii("doccumentation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termiating"),
                &["terminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repressivo"),
                &["repression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencticates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("recommd"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("methdods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("prestes"), &["presets"]),
            (
                dictgen::InsensitiveStr::Ascii("egaletarian"),
                &["egalitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("testsers"), &["testers"]),
            (
                dictgen::InsensitiveStr::Ascii("declarayion"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perspectivas"),
                &["perspectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("recude"), &["reduce"]),
            (
                dictgen::InsensitiveStr::Ascii("responsbile"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("receipe"),
                &["recipe", "receipt"],
            ),
            (dictgen::InsensitiveStr::Ascii("crahsed"), &["crashed"]),
            (
                dictgen::InsensitiveStr::Ascii("respnsibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicated"), &["dictated"]),
            (
                dictgen::InsensitiveStr::Ascii("deficienty"),
                &["deficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enternally"),
                &["internally", "externally", "eternally"],
            ),
            (dictgen::InsensitiveStr::Ascii("rightfuly"), &["rightfully"]),
            (dictgen::InsensitiveStr::Ascii("dfferent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("environementally"),
                &["environmentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("delpoyed"), &["deployed"]),
            (dictgen::InsensitiveStr::Ascii("netflify"), &["netlify"]),
            (dictgen::InsensitiveStr::Ascii("shcemes"), &["schemes"]),
            (dictgen::InsensitiveStr::Ascii("anonimous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("interruptes"),
                &["interrupts"],
            ),
            (dictgen::InsensitiveStr::Ascii("anorexiac"), &["anorexic"]),
            (dictgen::InsensitiveStr::Ascii("enither"), &["either"]),
            (dictgen::InsensitiveStr::Ascii("nickmane"), &["nickname"]),
            (
                dictgen::InsensitiveStr::Ascii("exracts"),
                &["extracts", "exacts"],
            ),
            (dictgen::InsensitiveStr::Ascii("transprot"), &["transport"]),
            (dictgen::InsensitiveStr::Ascii("yrea"), &["year"]),
            (
                dictgen::InsensitiveStr::Ascii("operationg"),
                &["operating", "operation"],
            ),
            (dictgen::InsensitiveStr::Ascii("infrotn"), &["infront"]),
            (dictgen::InsensitiveStr::Ascii("selceting"), &["selecting"]),
            (dictgen::InsensitiveStr::Ascii("thoruogh"), &["thorough"]),
            (
                dictgen::InsensitiveStr::Ascii("convedsion"),
                &["conversion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditianal"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("academcis"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("coersced"), &["coerced"]),
            (
                dictgen::InsensitiveStr::Ascii("harrange"),
                &["arrange", "harangue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimilat"),
                &["dissimilar"],
            ),
            (dictgen::InsensitiveStr::Ascii("nominatin"), &["nomination"]),
            (dictgen::InsensitiveStr::Ascii("revelant"), &["relevant"]),
            (
                dictgen::InsensitiveStr::Ascii("transision"),
                &["transition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotition"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repertwares"),
                &["repertoires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcisscist"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("catalgoue"), &["catalogue"]),
            (dictgen::InsensitiveStr::Ascii("dpended"), &["depended"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventurus"),
                &["adventurous"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectet"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("poimt"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("manipluation"),
                &["manipulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("emplys"), &["employs"]),
            (
                dictgen::InsensitiveStr::Ascii("orientato"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("recommmend"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("declatory"),
                &["declaratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("copyied"), &["copied"]),
            (
                dictgen::InsensitiveStr::Ascii("proprotion"),
                &["proportion"],
            ),
            (dictgen::InsensitiveStr::Ascii("piorities"), &["priorities"]),
            (
                dictgen::InsensitiveStr::Ascii("cronstructs"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alloacation"),
                &["allocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("betweeness"),
                &["betweenness"],
            ),
            (dictgen::InsensitiveStr::Ascii("daa"), &["data"]),
            (dictgen::InsensitiveStr::Ascii("absulute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("respektable"),
                &["respectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("foreinger"), &["foreigner"]),
            (
                dictgen::InsensitiveStr::Ascii("compenstate"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depencence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("frontilne"), &["frontline"]),
            (dictgen::InsensitiveStr::Ascii("thannk"), &["thank"]),
            (dictgen::InsensitiveStr::Ascii("affecfted"), &["affected"]),
            (
                dictgen::InsensitiveStr::Ascii("maintanability"),
                &["maintainability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ancapsulates"),
                &["encapsulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unimplimented"),
                &["unimplemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deveolping"),
                &["developing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturct"), &["struct"]),
            (dictgen::InsensitiveStr::Ascii("parrakeets"), &["parakeets"]),
            (
                dictgen::InsensitiveStr::Ascii("conventionnal"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("macthed"), &["matched"]),
            (
                dictgen::InsensitiveStr::Ascii("comparibility"),
                &["comparability", "compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transistion"),
                &["transition"],
            ),
            (dictgen::InsensitiveStr::Ascii("valetta"), &["valletta"]),
            (
                dictgen::InsensitiveStr::Ascii("predicited"),
                &["predicated", "predicted"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatched"), &["detached"]),
            (dictgen::InsensitiveStr::Ascii("subsrcibe"), &["subscribe"]),
            (
                dictgen::InsensitiveStr::Ascii("tupperwears"),
                &["tupperware"],
            ),
            (dictgen::InsensitiveStr::Ascii("usptart"), &["upstart"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentos"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("overfloww"), &["overflow"]),
            (dictgen::InsensitiveStr::Ascii("recidents"), &["residents"]),
            (
                dictgen::InsensitiveStr::Ascii("troublshoot"),
                &["troubleshoot"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostaglic"), &["nostalgic"]),
            (
                dictgen::InsensitiveStr::Ascii("inheritting"),
                &["inheriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chronosome"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainner"), &["trainer"]),
            (dictgen::InsensitiveStr::Ascii("endpooint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("sepertor"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatent"),
                &["incompetent"],
            ),
            (dictgen::InsensitiveStr::Ascii("baord"), &["board"]),
            (dictgen::InsensitiveStr::Ascii("resotre"), &["restore"]),
            (
                dictgen::InsensitiveStr::Ascii("chaeckerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrtupable"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interresing"),
                &["interfering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatino"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanel"), &["channel"]),
            (
                dictgen::InsensitiveStr::Ascii("investigaron"),
                &["investigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubes"),
                &["troubles", "troupes"],
            ),
            (dictgen::InsensitiveStr::Ascii("activ"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("sumarising"),
                &["summarising"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyisers"), &["analysers"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticty"),
                &["authenticity"],
            ),
            (dictgen::InsensitiveStr::Ascii("styl"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("concurents"),
                &["concurrents", "concurrence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("histerical"),
                &["historical", "hysterical"],
            ),
            (dictgen::InsensitiveStr::Ascii("flipers"), &["flippers"]),
            (
                dictgen::InsensitiveStr::Ascii("priorized"),
                &["prioritized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuemos"),
                &["continues"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reords"),
                &["records", "rewords"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpts"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("expetct"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpreters"),
                &["interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assiciation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consturctions"),
                &["constructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirutalised"),
                &["devirtualised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappering"),
                &["disappearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconciderate"),
                &["inconsiderate"],
            ),
            (dictgen::InsensitiveStr::Ascii("eabled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("corospondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoritive"),
                &["authoritative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensable"), &["sensible"]),
            (
                dictgen::InsensitiveStr::Ascii("protectoras"),
                &["protectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonised"),
                &["synchronised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encryptiing"),
                &["encrypting"],
            ),
            (dictgen::InsensitiveStr::Ascii("otheres"), &["others"]),
            (dictgen::InsensitiveStr::Ascii("folowwer"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("servecing"), &["servicing"]),
            (
                dictgen::InsensitiveStr::Ascii("peases"),
                &["leases", "pieces", "pleases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destinatinos"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("acessable"), &["accessible"]),
            (
                dictgen::InsensitiveStr::Ascii("asserttively"),
                &["assertively"],
            ),
            (dictgen::InsensitiveStr::Ascii("propsed"), &["proposed"]),
            (dictgen::InsensitiveStr::Ascii("usserspace"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("credentails"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contricted"),
                &["contracted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("poilcy"), &["policy"]),
            (
                dictgen::InsensitiveStr::Ascii("shortenting"),
                &["shortening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleizes"),
                &["parallelizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("excat"), &["exact"]),
            (dictgen::InsensitiveStr::Ascii("wirldcard"), &["wildcard"]),
            (
                dictgen::InsensitiveStr::Ascii("mechnanically"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behavious"),
                &["behaviour", "behaviours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectualls"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nessecarilt"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("mundance"), &["mundane"]),
            (
                dictgen::InsensitiveStr::Ascii("progarammable"),
                &["programmable"],
            ),
            (dictgen::InsensitiveStr::Ascii("contex"), &["context"]),
            (
                dictgen::InsensitiveStr::Ascii("metropolitin"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassmbler"),
                &["disassembler"],
            ),
            (dictgen::InsensitiveStr::Ascii("librabies"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("solutino"), &["solution"]),
            (dictgen::InsensitiveStr::Ascii("folowind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("speficifed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("expropiate"),
                &["expropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwrited"),
                &["overwritten", "overwrote"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomodative"),
                &["accommodative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyphenatin"),
                &["hyphenating", "hyphenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrary"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("forntline"), &["frontline"]),
            (
                dictgen::InsensitiveStr::Ascii("consruction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compansation"),
                &["compensation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sandwishes"),
                &["sandwiches"],
            ),
            (dictgen::InsensitiveStr::Ascii("vailidates"), &["validates"]),
            (
                dictgen::InsensitiveStr::Ascii("krypyonite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recgonized"),
                &["recognized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lepracans"),
                &["leprechauns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normallization"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("upstem"), &["upstep"]),
            (
                dictgen::InsensitiveStr::Ascii("beuraucracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuing"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("assiciations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("iif"), &["if"]),
            (dictgen::InsensitiveStr::Ascii("instabces"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("vacinations"),
                &["vaccinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("proedures"), &["procedures"]),
            (dictgen::InsensitiveStr::Ascii("cuddleys"), &["cuddles"]),
            (
                dictgen::InsensitiveStr::Ascii("guarenteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immidatly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnfigures"), &["configures"]),
            (dictgen::InsensitiveStr::Ascii("unicde"), &["unicode"]),
            (dictgen::InsensitiveStr::Ascii("uesername"), &["username"]),
            (
                dictgen::InsensitiveStr::Ascii("palistian"),
                &["palestinian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generationens"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frontlinies"),
                &["frontline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adevnturer"),
                &["adventurer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenedecies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("unhealty"), &["unhealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("disabilitily"),
                &["disability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritizin"),
                &["prioritizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("walkaroud"),
                &["workaround", "walkaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconditionally"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconvectional"),
                &["unconventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrubytes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitplied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloesing"), &["closing"]),
            (
                dictgen::InsensitiveStr::Ascii("investiagtive"),
                &["investigative"],
            ),
            (dictgen::InsensitiveStr::Ascii("assuembly"), &["assembly"]),
            (
                dictgen::InsensitiveStr::Ascii("instraments"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alikes"),
                &["alike", "likes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cadidate"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("preficiently"),
                &["proficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converitble"),
                &["convertible"],
            ),
            (dictgen::InsensitiveStr::Ascii("phenonmena"), &["phenomena"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleartes"),
                &["accelerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripion"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionn"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevtiably"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("proposte"), &["propose"]),
            (
                dictgen::InsensitiveStr::Ascii("stanardizes"),
                &["standardizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirters"), &["writers"]),
            (dictgen::InsensitiveStr::Ascii("iteartors"), &["iterators"]),
            (
                dictgen::InsensitiveStr::Ascii("attribuitions"),
                &["attributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rithmicly"),
                &["rhythmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuratiom"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corospondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reducs"),
                &["reduce", "reduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprentace"),
                &["apprentice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crsytallographic"),
                &["crystallographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("mergable"), &["mergeable"]),
            (
                dictgen::InsensitiveStr::Ascii("extermally"),
                &["externally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funvtional"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("chillend"), &["chilled"]),
            (
                dictgen::InsensitiveStr::Ascii("destrution"),
                &["destruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("flakeness"), &["flakiness"]),
            (
                dictgen::InsensitiveStr::Ascii("breasttaking"),
                &["breathtaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabalizing"),
                &["stabilizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traingulations"),
                &["triangulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tablulated"), &["tabulated"]),
            (dictgen::InsensitiveStr::Ascii("canrage"), &["carnage"]),
            (dictgen::InsensitiveStr::Ascii("bluetooh"), &["bluetooth"]),
            (
                dictgen::InsensitiveStr::Ascii("backbround"),
                &["background"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcions"), &["directions"]),
            (
                dictgen::InsensitiveStr::Ascii("resistnaces"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("bussiness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("conversationy"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfirmation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("appoving"), &["approving"]),
            (dictgen::InsensitiveStr::Ascii("opearator"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("musicaly"), &["musically"]),
            (
                dictgen::InsensitiveStr::Ascii("regulationg"),
                &["regulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotaiations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encountre"),
                &["encounter", "encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("algrithms"), &["algorithms"]),
            (
                dictgen::InsensitiveStr::Ascii("malfuncting"),
                &["malfunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("novisheuts"),
                &["novitiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("crystalls"), &["crystals"]),
            (
                dictgen::InsensitiveStr::Ascii("insntruction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("exspected"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("surrount"), &["surround"]),
            (dictgen::InsensitiveStr::Ascii("sickamores"), &["sycamores"]),
            (dictgen::InsensitiveStr::Ascii("matricies"), &["matrices"]),
            (
                dictgen::InsensitiveStr::Ascii("sccessfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elelmentary"),
                &["elementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("mamuth"), &["mammoth"]),
            (
                dictgen::InsensitiveStr::Ascii("opportinity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nothingsness"),
                &["nothingness"],
            ),
            (dictgen::InsensitiveStr::Ascii("nuber"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("procliamed"),
                &["proclaimed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("situatution"),
                &["situation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camoflages"),
                &["camouflages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partitionned"),
                &["partitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossibley"),
                &["impossibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("integrey"), &["integrity"]),
            (dictgen::InsensitiveStr::Ascii("genererate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("barays"), &["berets"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacturor"),
                &["manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("fillung"), &["filling"]),
            (dictgen::InsensitiveStr::Ascii("frowrads"), &["forwards"]),
            (
                dictgen::InsensitiveStr::Ascii("powetlifting"),
                &["powerlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enformation"),
                &["information", "deformation", "reformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismatching"),
                &["mismatching", "dispatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("nerver"), &["never"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introductionary"),
                &["introductory"],
            ),
            (dictgen::InsensitiveStr::Ascii("liquides"), &["liquids"]),
            (dictgen::InsensitiveStr::Ascii("utlized"), &["utilized"]),
            (dictgen::InsensitiveStr::Ascii("qury"), &["query"]),
            (
                dictgen::InsensitiveStr::Ascii("propostion"),
                &["proposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexadeimcal"),
                &["hexadecimal"],
            ),
            (dictgen::InsensitiveStr::Ascii("moveble"), &["movable"]),
            (dictgen::InsensitiveStr::Ascii("soemthin"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("deinitalized"),
                &["deinitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simularity"),
                &["similarity", "singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbars"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrandizin"),
                &["aggrandizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("proccess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("scirptures"),
                &["scriptures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersction"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genererated"),
                &["generated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("appneds"), &["appends"]),
            (dictgen::InsensitiveStr::Ascii("selecton"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("diferentiating"),
                &["differentiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnmitting"),
                &["transmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustraitor"),
                &["illustrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("psudonyms"), &["pseudonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("reconeecting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsitable"), &["unsuitable"]),
            (dictgen::InsensitiveStr::Ascii("projeted"), &["projected"]),
            (dictgen::InsensitiveStr::Ascii("lightbub"), &["lightbulb"]),
            (
                dictgen::InsensitiveStr::Ascii("canonicied"),
                &["canonified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permisisons"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deterioratin"),
                &["deteriorating", "deterioration"],
            ),
            (dictgen::InsensitiveStr::Ascii("bardford"), &["bradford"]),
            (dictgen::InsensitiveStr::Ascii("resonet"), &["resonate"]),
            (
                dictgen::InsensitiveStr::Ascii("miplementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inventroy"), &["inventory"]),
            (dictgen::InsensitiveStr::Ascii("faiing"), &["failing"]),
            (dictgen::InsensitiveStr::Ascii("scroipts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("polymoprhisms"),
                &["polymorphisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("americain"), &["american"]),
            (dictgen::InsensitiveStr::Ascii("specktor"), &["specter"]),
            (dictgen::InsensitiveStr::Ascii("passiv"), &["passive"]),
            (dictgen::InsensitiveStr::Ascii("swcloumns"), &["swcolumns"]),
            (
                dictgen::InsensitiveStr::Ascii("dimesional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informations"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("artificts"), &["artifacts"]),
            (dictgen::InsensitiveStr::Ascii("compair"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalistisch"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraphraseing"),
                &["paraphrasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremnetd"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinig"),
                &["determining"],
            ),
            (dictgen::InsensitiveStr::Ascii("edittor"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("shinking"), &["shrinking"]),
            (
                dictgen::InsensitiveStr::Ascii("disconeted"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("attandees"), &["attendees"]),
            (dictgen::InsensitiveStr::Ascii("marryied"), &["married"]),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheeticus"),
                &["spreadsheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("perparing"), &["preparing"]),
            (dictgen::InsensitiveStr::Ascii("reprsesent"), &["represent"]),
            (
                dictgen::InsensitiveStr::Ascii("contuinity"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legalizeing"),
                &["legalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlyst"), &["analyst"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtorises"),
                &["authorises"],
            ),
            (dictgen::InsensitiveStr::Ascii("convingi"), &["convincing"]),
            (
                dictgen::InsensitiveStr::Ascii("undeterministic"),
                &["indeterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apppreciative"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worbenches"),
                &["workbenches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notoriuosly"),
                &["notoriously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactheristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("tmplate"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("puchase"), &["purchase"]),
            (dictgen::InsensitiveStr::Ascii("coverred"), &["covered"]),
            (dictgen::InsensitiveStr::Ascii("geoorgia"), &["georgia"]),
            (dictgen::InsensitiveStr::Ascii("remembed"), &["remembered"]),
            (
                dictgen::InsensitiveStr::Ascii("deliberitely"),
                &["deliberately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreedsheet"),
                &["spreadsheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("follwoeiong"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proccesors"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constarint"),
                &["constraint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancelleation"),
                &["cancellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expementel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiailised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("appologize"), &["apologize"]),
            (dictgen::InsensitiveStr::Ascii("amounds"), &["amounts"]),
            (dictgen::InsensitiveStr::Ascii("impreved"), &["improved"]),
            (
                dictgen::InsensitiveStr::Ascii("interacsion"),
                &["interaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pariticipation"),
                &["participation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrieveing"),
                &["retrieving"],
            ),
            (dictgen::InsensitiveStr::Ascii("choicers"), &["choices"]),
            (dictgen::InsensitiveStr::Ascii("amnestry"), &["amnesty"]),
            (
                dictgen::InsensitiveStr::Ascii("execeptional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("assinged"), &["assigned"]),
            (dictgen::InsensitiveStr::Ascii("gruop"), &["group"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegoations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("describtion"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("evauluator"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("nuetrally"), &["neutrally"]),
            (
                dictgen::InsensitiveStr::Ascii("histocompatability"),
                &["histocompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititations"),
                &["initiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereators"), &["generators"]),
            (
                dictgen::InsensitiveStr::Ascii("specularite"),
                &["speculative"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxmimums"), &["maximums"]),
            (
                dictgen::InsensitiveStr::Ascii("lates"),
                &["later", "latest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompasitates"),
                &["incapacitates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referendim"),
                &["referendum"],
            ),
            (dictgen::InsensitiveStr::Ascii("oberver"), &["observer"]),
            (
                dictgen::InsensitiveStr::Ascii("notidications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoroithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplactes"),
                &["deduplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("airpsace"), &["airspace"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscibe"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comptablity"),
                &["compatibility", "computability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trajecotry"),
                &["trajectory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equall"),
                &["equal", "equally"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanegd"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("transitin"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("sorcerors"), &["sorcerers"]),
            (
                dictgen::InsensitiveStr::Ascii("eroneously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrcuted"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpic"), &["topic"]),
            (
                dictgen::InsensitiveStr::Ascii("inhereting"),
                &["inheriting"],
            ),
            (dictgen::InsensitiveStr::Ascii("perforemd"), &["performed"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrided"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("slient"), &["silent"]),
            (dictgen::InsensitiveStr::Ascii("rivarly"), &["rivalry"]),
            (
                dictgen::InsensitiveStr::Ascii("approximiting"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("repalaced"), &["replaced"]),
            (dictgen::InsensitiveStr::Ascii("shangahi"), &["shanghai"]),
            (
                dictgen::InsensitiveStr::Ascii("chipertext"),
                &["ciphertext"],
            ),
            (dictgen::InsensitiveStr::Ascii("assingnees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("abrreviate"),
                &["abbreviate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("livilehood"),
                &["livelihood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjudgation"),
                &["subjugation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workbenchch"),
                &["workbench"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unplaybale"),
                &["unplayable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sxmbols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("elemnts"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("abou"), &["about", "abound"]),
            (dictgen::InsensitiveStr::Ascii("armpitts"), &["armpits"]),
            (dictgen::InsensitiveStr::Ascii("preprared"), &["prepared"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaions"),
                &["reallocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("epsresso"), &["espresso"]),
            (dictgen::InsensitiveStr::Ascii("prinicipal"), &["principal"]),
            (dictgen::InsensitiveStr::Ascii("enfoceing"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("reparamterizing"),
                &["reparameterizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("pretifies"), &["prettifies"]),
            (
                dictgen::InsensitiveStr::Ascii("caterpilars"),
                &["caterpillars"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolument"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("ejacluation"),
                &["ejaculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unlass"), &["unless"]),
            (dictgen::InsensitiveStr::Ascii("refiined"), &["refined"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterprett"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("pixes"), &["pixels"]),
            (
                dictgen::InsensitiveStr::Ascii("initiazed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("documement"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficifed"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("socript"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("tranitioned"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depencendency"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("navtives"), &["natives"]),
            (dictgen::InsensitiveStr::Ascii("unsusual"), &["unusual"]),
            (dictgen::InsensitiveStr::Ascii("chaecker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeurghs"),
                &["neighbours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countinually"),
                &["continually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otimisations"),
                &["optimisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncensord"), &["uncensored"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construits"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressivily"),
                &["progressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misintrepret"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdatal"), &["anecdotal"]),
            (
                dictgen::InsensitiveStr::Ascii("necessitive"),
                &["necessities"],
            ),
            (dictgen::InsensitiveStr::Ascii("elememts"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("collaborare"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mesasge"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("anarquistic"),
                &["anarchistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantiate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beanch"),
                &["branch", "bench", "beach"],
            ),
            (dictgen::InsensitiveStr::Ascii("coutry"), &["country"]),
            (dictgen::InsensitiveStr::Ascii("operare"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("qeries"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("pharamcists"),
                &["pharmacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepcifier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("constructure"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("posibbly"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("structred"), &["structured"]),
            (dictgen::InsensitiveStr::Ascii("tournyes"), &["tourneys"]),
            (
                dictgen::InsensitiveStr::Ascii("strwaberry"),
                &["strawberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("offereing"), &["offering"]),
            (
                dictgen::InsensitiveStr::Ascii("destintation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coproccessor"),
                &["coprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crationalism"),
                &["nationalism", "rationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multicultralism"),
                &["multiculturalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("variante"),
                &["variant", "variance", "variate"],
            ),
            (dictgen::InsensitiveStr::Ascii("yieded"), &["yielded"]),
            (dictgen::InsensitiveStr::Ascii("verifers"), &["verifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("electronicos"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutnerfeit"),
                &["counterfeit"],
            ),
            (dictgen::InsensitiveStr::Ascii("oralces"), &["oracles"]),
            (
                dictgen::InsensitiveStr::Ascii("ventillates"),
                &["ventilates"],
            ),
            (dictgen::InsensitiveStr::Ascii("desciptor"), &["descriptor"]),
            (dictgen::InsensitiveStr::Ascii("olther"), &["other"]),
            (dictgen::InsensitiveStr::Ascii("dumplicate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("occation"), &["occasion"]),
            (dictgen::InsensitiveStr::Ascii("unistall"), &["uninstall"]),
            (dictgen::InsensitiveStr::Ascii("egister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("implementaton"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("listenser"), &["listener"]),
            (
                dictgen::InsensitiveStr::Ascii("animetions"),
                &["animations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accommadating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("neesds"), &["needs"]),
            (
                dictgen::InsensitiveStr::Ascii("conspiriator"),
                &["conspirator"],
            ),
            (dictgen::InsensitiveStr::Ascii("preaparing"), &["preparing"]),
            (dictgen::InsensitiveStr::Ascii("calulates"), &["calculates"]),
            (dictgen::InsensitiveStr::Ascii("gorgous"), &["gorgeous"]),
            (
                dictgen::InsensitiveStr::Ascii("transmittd"),
                &["transmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ameelarates"),
                &["ameliorates"],
            ),
            (dictgen::InsensitiveStr::Ascii("bcuket"), &["bucket"]),
            (
                dictgen::InsensitiveStr::Ascii("extrememely"),
                &["extremely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consultunt"),
                &["consultant"],
            ),
            (dictgen::InsensitiveStr::Ascii("errorrs"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("ideologiske"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("camrbidge"), &["cambridge"]),
            (
                dictgen::InsensitiveStr::Ascii("protagnoist"),
                &["protagonist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sattelites"),
                &["satellites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overridiing"),
                &["overriding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controllled"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstlaling"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("febrary"), &["february"]),
            (
                dictgen::InsensitiveStr::Ascii("interfaves"),
                &["interfaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mislading"),
                &["misleading", "misloading"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuidos"), &["studios"]),
            (
                dictgen::InsensitiveStr::Ascii("immeidately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("wouuld"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("anfd"), &["and"]),
            (
                dictgen::InsensitiveStr::Ascii("constently"),
                &["constantly", "consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parameteres"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("prifix"), &["prefix"]),
            (dictgen::InsensitiveStr::Ascii("succeessor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasformed"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrcutinoal"),
                &["instructional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geopraphically"),
                &["geographically"],
            ),
            (dictgen::InsensitiveStr::Ascii("produses"), &["produces"]),
            (dictgen::InsensitiveStr::Ascii("chizzling"), &["chiseling"]),
            (
                dictgen::InsensitiveStr::Ascii("fisionable"),
                &["fissionable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acitvities"),
                &["activities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integratione"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("populr"), &["popular"]),
            (
                dictgen::InsensitiveStr::Ascii("catapillar"),
                &["caterpillar"],
            ),
            (dictgen::InsensitiveStr::Ascii("startde"), &["startled"]),
            (dictgen::InsensitiveStr::Ascii("atcualy"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("cusomisable"),
                &["customisable"],
            ),
            (dictgen::InsensitiveStr::Ascii("neoghbor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("comaprator"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("safty"), &["safety"]),
            (
                dictgen::InsensitiveStr::Ascii("certifiates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellegence"),
                &["intelligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exsited"),
                &["existed", "excited"],
            ),
            (dictgen::InsensitiveStr::Ascii("comitters"), &["committers"]),
            (dictgen::InsensitiveStr::Ascii("disabilty"), &["disability"]),
            (
                dictgen::InsensitiveStr::Ascii("differentiatiors"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencticators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("basicaly"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("atorney"), &["attorney"]),
            (
                dictgen::InsensitiveStr::Ascii("reperesenting"),
                &["representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporraily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminat"), &["terminate"]),
            (
                dictgen::InsensitiveStr::Ascii("relicate"),
                &["replicate", "relocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrobed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("masquarade"),
                &["masquerade"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snowballling"),
                &["snowballing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartizes"),
                &["standardizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumeents"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("tresuring"), &["treasuring"]),
            (dictgen::InsensitiveStr::Ascii("priotity"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("overarcing"),
                &["overarching"],
            ),
            (dictgen::InsensitiveStr::Ascii("vulacn"), &["vulcan"]),
            (dictgen::InsensitiveStr::Ascii("mdoes"), &["does", "modes"]),
            (dictgen::InsensitiveStr::Ascii("manuevers"), &["maneuvers"]),
            (
                dictgen::InsensitiveStr::Ascii("attriubution"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticting"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("referencs"), &["references"]),
            (
                dictgen::InsensitiveStr::Ascii("unproteced"),
                &["unprotected"],
            ),
            (dictgen::InsensitiveStr::Ascii("indoneisa"), &["indonesia"]),
            (
                dictgen::InsensitiveStr::Ascii("unattractice"),
                &["unattractive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effectionate"),
                &["affectionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tansverse"), &["transverse"]),
            (dictgen::InsensitiveStr::Ascii("hhader"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("photograper"),
                &["photographer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siezed"),
                &["seized", "sized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("voluntereed"),
                &["volunteered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teamplating"),
                &["templating"],
            ),
            (dictgen::InsensitiveStr::Ascii("schemd"), &["schemed"]),
            (
                dictgen::InsensitiveStr::Ascii("scetioning"),
                &["sectioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabilisitic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converzation"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsitutable"),
                &["substitutable"],
            ),
            (dictgen::InsensitiveStr::Ascii("organices"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("renewebles"),
                &["renewables"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheilded"), &["shielded"]),
            (dictgen::InsensitiveStr::Ascii("netwplit"), &["netsplit"]),
            (
                dictgen::InsensitiveStr::Ascii("beutifully"),
                &["beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emprically"),
                &["empirically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("animartions"),
                &["animations"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwiong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("absoluetly"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("heroicas"), &["heroics"]),
            (
                dictgen::InsensitiveStr::Ascii("execuse"),
                &["excuse", "execute"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavoiur"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("isnb"), &["isbn"]),
            (
                dictgen::InsensitiveStr::Ascii("apocalypic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("inporters"), &["importers"]),
            (dictgen::InsensitiveStr::Ascii("retreive"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("presenece"), &["presence"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordint"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("primititves"),
                &["primitives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciffically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("emapthize"), &["empathize"]),
            (
                dictgen::InsensitiveStr::Ascii("consequene"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insentivised"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excetutable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("restors"), &["restores"]),
            (dictgen::InsensitiveStr::Ascii("extranous"), &["extraneous"]),
            (dictgen::InsensitiveStr::Ascii("arrestes"), &["arrests"]),
            (
                dictgen::InsensitiveStr::Ascii("enhanements"),
                &["enhancements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrobuted"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categoized"),
                &["categorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critcially"),
                &["critically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilially"),
                &["initially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reight"),
                &["eight", "freight", "right"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkpont"), &["checkpoint"]),
            (dictgen::InsensitiveStr::Ascii("coercable"), &["coercible"]),
            (
                dictgen::InsensitiveStr::Ascii("coalasence"),
                &["coalescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploygamy"), &["polygamy"]),
            (
                dictgen::InsensitiveStr::Ascii("procastrinating"),
                &["procrastinating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unversities"),
                &["universities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overpovered"),
                &["overpowered"],
            ),
            (dictgen::InsensitiveStr::Ascii("decend"), &["descend"]),
            (
                dictgen::InsensitiveStr::Ascii("restaruants"),
                &["restaurants"],
            ),
            (dictgen::InsensitiveStr::Ascii("stopps"), &["stops"]),
            (
                dictgen::InsensitiveStr::Ascii("rendererers"),
                &["renderers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entends"),
                &["intends", "extends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendants"),
                &["dependants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superposer"),
                &["superpowers"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressue"), &["pressure"]),
            (dictgen::InsensitiveStr::Ascii("exlusions"), &["exclusions"]),
            (dictgen::InsensitiveStr::Ascii("volxels"), &["voxels"]),
            (dictgen::InsensitiveStr::Ascii("transtive"), &["transitive"]),
            (
                dictgen::InsensitiveStr::Ascii("integratoin"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsricption"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tatoo"), &["tattoo"]),
            (
                dictgen::InsensitiveStr::Ascii("intefering"),
                &["interfering"],
            ),
            (dictgen::InsensitiveStr::Ascii("rquiring"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("repalces"), &["replaces"]),
            (
                dictgen::InsensitiveStr::Ascii("staticists"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("easely"), &["easily"]),
            (
                dictgen::InsensitiveStr::Ascii("execting"),
                &["executing", "expecting", "ejecting", "exerting"],
            ),
            (dictgen::InsensitiveStr::Ascii("mailicious"), &["malicious"]),
            (dictgen::InsensitiveStr::Ascii("otpimized"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("songled"),
                &["singled", "dongled"],
            ),
            (dictgen::InsensitiveStr::Ascii("avtaars"), &["avatars"]),
            (
                dictgen::InsensitiveStr::Ascii("genarations"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("altready"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("deserailises"),
                &["deserialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatble"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("escluded"), &["excluded"]),
            (dictgen::InsensitiveStr::Ascii("deleleting"), &["deleting"]),
            (dictgen::InsensitiveStr::Ascii("callbakcs"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("descrtiptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterstically"),
                &["characteristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("megers"),
                &["mergers", "merges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tournemants"),
                &["tournaments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconcider"),
                &["reconsider"],
            ),
            (dictgen::InsensitiveStr::Ascii("virtiual"), &["virtual"]),
            (dictgen::InsensitiveStr::Ascii("conribute"), &["contribute"]),
            (
                dictgen::InsensitiveStr::Ascii("proffessionally"),
                &["professionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formualting"),
                &["formulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("afecting"), &["affecting"]),
            (
                dictgen::InsensitiveStr::Ascii("genarators"),
                &["generators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnateing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("disscover"), &["discover"]),
            (
                dictgen::InsensitiveStr::Ascii("assemalates"),
                &["assimilates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencicators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("favoruites"), &["favorites"]),
            (
                dictgen::InsensitiveStr::Ascii("sucesseding"),
                &["succeeding", "seceding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irreversibel"),
                &["irreversible"],
            ),
            (dictgen::InsensitiveStr::Ascii("adminssion"), &["admission"]),
            (dictgen::InsensitiveStr::Ascii("excludind"), &["excluding"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnearability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kilometros"),
                &["kilometres"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circomvented"),
                &["circumvented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplfication"),
                &["simplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strawburries"),
                &["strawberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("acommpany"), &["accompany"]),
            (
                dictgen::InsensitiveStr::Ascii("comtiguous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recgonizable"),
                &["recognizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("runnung"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("seacrhing"), &["searching"]),
            (dictgen::InsensitiveStr::Ascii("vaaluing"), &["valuing"]),
            (
                dictgen::InsensitiveStr::Ascii("associatied"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("monologe"), &["monologue"]),
            (dictgen::InsensitiveStr::Ascii("dettach"), &["detach"]),
            (
                dictgen::InsensitiveStr::Ascii("expermienters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("inseted"), &["inserted"]),
            (
                dictgen::InsensitiveStr::Ascii("dentifiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremeantal"),
                &["incremental"],
            ),
            (dictgen::InsensitiveStr::Ascii("definely"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("baskteball"),
                &["basketball"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraenous"),
                &["extraneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("obgect"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("finelly"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("extered"), &["exerted"]),
            (dictgen::InsensitiveStr::Ascii("microfost"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("reoslution"),
                &["resolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjast"), &["adjust"]),
            (dictgen::InsensitiveStr::Ascii("smeone"), &["someone"]),
            (
                dictgen::InsensitiveStr::Ascii("rebuplican"),
                &["republican"],
            ),
            (dictgen::InsensitiveStr::Ascii("victum"), &["victim"]),
            (dictgen::InsensitiveStr::Ascii("artifial"), &["artificial"]),
            (
                dictgen::InsensitiveStr::Ascii("algorritms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("onwer"), &["owner"]),
            (dictgen::InsensitiveStr::Ascii("someoene"), &["someone"]),
            (dictgen::InsensitiveStr::Ascii("perfurred"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("paylod"), &["payload"]),
            (dictgen::InsensitiveStr::Ascii("tescase"), &["testcase"]),
            (
                dictgen::InsensitiveStr::Ascii("bakgrounds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("practial"), &["practical"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpetcted"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discourging"),
                &["discouraging"],
            ),
            (dictgen::InsensitiveStr::Ascii("moutning"), &["mounting"]),
            (dictgen::InsensitiveStr::Ascii("recoriding"), &["recording"]),
            (dictgen::InsensitiveStr::Ascii("therstat"), &["thermostat"]),
            (dictgen::InsensitiveStr::Ascii("correspnd"), &["correspond"]),
            (dictgen::InsensitiveStr::Ascii("shelfs"), &["shelves"]),
            (
                dictgen::InsensitiveStr::Ascii("confesssion"),
                &["confessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("readibly"), &["readably"]),
            (
                dictgen::InsensitiveStr::Ascii("cononicalized"),
                &["canonicalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("beaon"), &["beacon"]),
            (
                dictgen::InsensitiveStr::Ascii("antaganistic"),
                &["antagonistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compolsive"),
                &["compulsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiibly"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("nethertheless"),
                &["nevertheless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differerential"),
                &["differential"],
            ),
            (dictgen::InsensitiveStr::Ascii("relecant"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("seralize"), &["serialize"]),
            (
                dictgen::InsensitiveStr::Ascii("bordrelines"),
                &["borderlines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jouranlism"),
                &["journalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuzes"), &["confuses"]),
            (
                dictgen::InsensitiveStr::Ascii("devestating"),
                &["devastating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expoential"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("yorskhire"), &["yorkshire"]),
            (dictgen::InsensitiveStr::Ascii("warehosue"), &["warehouse"]),
            (dictgen::InsensitiveStr::Ascii("widthout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("configer"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("uncorrolated"),
                &["uncorrelated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stanardize"),
                &["standardize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurin"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pocessed"),
                &["processed", "possessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("expirin"), &["expiring"]),
            (dictgen::InsensitiveStr::Ascii("lugages"), &["luggage"]),
            (dictgen::InsensitiveStr::Ascii("influece"), &["influence"]),
            (dictgen::InsensitiveStr::Ascii("destoryed"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("conjuction"),
                &["conjunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("stratgies"), &["strategies"]),
            (dictgen::InsensitiveStr::Ascii("uranuim"), &["uranium"]),
            (dictgen::InsensitiveStr::Ascii("kunckle"), &["knuckle"]),
            (dictgen::InsensitiveStr::Ascii("scenarioes"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("agreeeableness"),
                &["agreeableness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disscusing"),
                &["discussing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastruction"),
                &["infrastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("coolots"), &["culottes"]),
            (dictgen::InsensitiveStr::Ascii("fullanme"), &["fullname"]),
            (
                dictgen::InsensitiveStr::Ascii("unreprocducable"),
                &["unreproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramlises"),
                &["normalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("assistent"), &["assistant"]),
            (dictgen::InsensitiveStr::Ascii("accquire"), &["acquire"]),
            (
                dictgen::InsensitiveStr::Ascii("implmeneted"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("benn"), &["been"]),
            (
                dictgen::InsensitiveStr::Ascii("shenanigangs"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviromentalist"),
                &["environmentalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelertors"),
                &["accelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowwong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("precended"), &["preceded"]),
            (dictgen::InsensitiveStr::Ascii("ultimatuum"), &["ultimatum"]),
            (
                dictgen::InsensitiveStr::Ascii("scandinavan"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplyfied"),
                &["simplified"],
            ),
            (dictgen::InsensitiveStr::Ascii("borow"), &["borrow"]),
            (
                dictgen::InsensitiveStr::Ascii("conclutions"),
                &["conclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chganging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("accomondated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("jennigns"), &["jennings"]),
            (dictgen::InsensitiveStr::Ascii("devider"), &["divider"]),
            (dictgen::InsensitiveStr::Ascii("reconeect"), &["reconnect"]),
            (dictgen::InsensitiveStr::Ascii("sinopsis"), &["synopsis"]),
            (
                dictgen::InsensitiveStr::Ascii("allocatbale"),
                &["allocatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("behaviro"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("thta"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("blessins"), &["blessings"]),
            (dictgen::InsensitiveStr::Ascii("overides"), &["overrides"]),
            (dictgen::InsensitiveStr::Ascii("algoritthm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("simlutaneously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("neccessity"), &["necessity"]),
            (
                dictgen::InsensitiveStr::Ascii("elimanation"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maskerading"),
                &["masquerading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checksumed"),
                &["checksummed"],
            ),
            (dictgen::InsensitiveStr::Ascii("agravate"), &["aggravate"]),
            (dictgen::InsensitiveStr::Ascii("gret"), &["great"]),
            (dictgen::InsensitiveStr::Ascii("xour"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("thinig"), &["thing"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumend"),
                &["instrument"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimneting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicitally"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrition"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("coprights"), &["copyrights"]),
            (
                dictgen::InsensitiveStr::Ascii("alhabetical"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overreacton"),
                &["overreaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furethermore"),
                &["furthermore"],
            ),
            (dictgen::InsensitiveStr::Ascii("interese"), &["interested"]),
            (
                dictgen::InsensitiveStr::Ascii("situatin"),
                &["situating", "situation"],
            ),
            (dictgen::InsensitiveStr::Ascii("finanacier"), &["financier"]),
            (dictgen::InsensitiveStr::Ascii("moderatin"), &["moderation"]),
            (dictgen::InsensitiveStr::Ascii("ensureing"), &["ensuring"]),
            (dictgen::InsensitiveStr::Ascii("adheasive"), &["adhesive"]),
            (
                dictgen::InsensitiveStr::Ascii("partititioning"),
                &["partitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apoligetic"),
                &["apologetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("extenstion"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("cammanded"), &["commanded"]),
            (dictgen::InsensitiveStr::Ascii("enusre"), &["ensure"]),
            (
                dictgen::InsensitiveStr::Ascii("autoeselect"),
                &["autoselect"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirrorring"), &["mirroring"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuillities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provocitive"),
                &["provocative"],
            ),
            (dictgen::InsensitiveStr::Ascii("hollywod"), &["hollywood"]),
            (dictgen::InsensitiveStr::Ascii("creastor"), &["creator"]),
            (
                dictgen::InsensitiveStr::Ascii("regeneratin"),
                &["regeneration", "regenerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("androidos"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("intellectul"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("plarise"), &["polarise"]),
            (
                dictgen::InsensitiveStr::Ascii("distriutors"),
                &["distributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("fianlly"), &["finally"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainted"),
                &["maintained"],
            ),
            (dictgen::InsensitiveStr::Ascii("mantain"), &["maintain"]),
            (dictgen::InsensitiveStr::Ascii("capablly"), &["capably"]),
            (dictgen::InsensitiveStr::Ascii("socialisim"), &["socialism"]),
            (dictgen::InsensitiveStr::Ascii("modfider"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("dislays"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("indxed"), &["indexed"]),
            (dictgen::InsensitiveStr::Ascii("cehck"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("abusrdly"), &["absurdly"]),
            (dictgen::InsensitiveStr::Ascii("offests"), &["offsets"]),
            (dictgen::InsensitiveStr::Ascii("boutnies"), &["bounties"]),
            (
                dictgen::InsensitiveStr::Ascii("extroardinarily"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("claaes"), &["classes"]),
            (dictgen::InsensitiveStr::Ascii("dimenson"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalista"),
                &["fundamentals"],
            ),
            (dictgen::InsensitiveStr::Ascii("deezil"), &["diesel"]),
            (dictgen::InsensitiveStr::Ascii("cilcker"), &["clicker"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithmmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protuguese"),
                &["portuguese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improbabile"),
                &["improbable"],
            ),
            (dictgen::InsensitiveStr::Ascii("languaege"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("interepreting"),
                &["interpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("langulages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomfortablity"),
                &["uncomfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("asume"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("windsoar"), &["windsor"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrviates"),
                &["abbreviates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embded"),
                &["embed", "embedded"],
            ),
            (dictgen::InsensitiveStr::Ascii("acsended"), &["ascended"]),
            (
                dictgen::InsensitiveStr::Ascii("outpreform"),
                &["outperform"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumeent"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("assymmetrically"),
                &["asymmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjusment"), &["adjustment"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplicities"),
                &["multiplicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacturin"),
                &["manufacturing"],
            ),
            (dictgen::InsensitiveStr::Ascii("abutts"), &["abuts"]),
            (dictgen::InsensitiveStr::Ascii("cannabiles"), &["cannibals"]),
            (
                dictgen::InsensitiveStr::Ascii("lightwight"),
                &["lightweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beauracracy"),
                &["bureaucracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentatially"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("forrset"), &["forrest"]),
            (
                dictgen::InsensitiveStr::Ascii("cutomisations"),
                &["customisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutnerproductive"),
                &["counterproductive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradicory"),
                &["contradictory"],
            ),
            (dictgen::InsensitiveStr::Ascii("templers"), &["temples"]),
            (dictgen::InsensitiveStr::Ascii("sapeenas"), &["subpoenas"]),
            (dictgen::InsensitiveStr::Ascii("ebale"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("dispay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("clienta"), &["client"]),
            (dictgen::InsensitiveStr::Ascii("chckers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("competative"),
                &["competitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinoidal"), &["sinusoidal"]),
            (
                dictgen::InsensitiveStr::Ascii("cption"),
                &["caption", "option"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multithreding"),
                &["multithreading"],
            ),
            (dictgen::InsensitiveStr::Ascii("maneveur"), &["maneuver"]),
            (dictgen::InsensitiveStr::Ascii("icrements"), &["increments"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlating"),
                &["translating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribuited"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intolorance"),
                &["intolerance"],
            ),
            (dictgen::InsensitiveStr::Ascii("intereact"), &["interact"]),
            (dictgen::InsensitiveStr::Ascii("convaxiity"), &["convexity"]),
            (dictgen::InsensitiveStr::Ascii("inisials"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("deaults"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("troubed"), &["troubled"]),
            (dictgen::InsensitiveStr::Ascii("wrokloads"), &["workloads"]),
            (dictgen::InsensitiveStr::Ascii("indcating"), &["indicating"]),
            (
                dictgen::InsensitiveStr::Ascii("succeesion"),
                &["succession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condtitional"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stting"),
                &["string", "setting", "sitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpcheck"), &["cppcheck"]),
            (dictgen::InsensitiveStr::Ascii("orfer"), &["offer", "order"]),
            (
                dictgen::InsensitiveStr::Ascii("gundamentalists"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heterosexal"),
                &["heterosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timstamping"),
                &["timestamping"],
            ),
            (dictgen::InsensitiveStr::Ascii("anectode"), &["anecdote"]),
            (
                dictgen::InsensitiveStr::Ascii("assocciated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consrtuctors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalsation"),
                &["capitalisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muliplicities"),
                &["multiplicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("clsoed"), &["closed"]),
            (
                dictgen::InsensitiveStr::Ascii("recconeting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrolld"), &["scrolled"]),
            (dictgen::InsensitiveStr::Ascii("fasening"), &["fastening"]),
            (dictgen::InsensitiveStr::Ascii("projecto"), &["projection"]),
            (
                dictgen::InsensitiveStr::Ascii("picknicking"),
                &["picnicking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referenenced"),
                &["referenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aqcuisition"),
                &["acquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("recrod"), &["record"]),
            (dictgen::InsensitiveStr::Ascii("calymore"), &["claymore"]),
            (dictgen::InsensitiveStr::Ascii("synonymos"), &["synonymous"]),
            (dictgen::InsensitiveStr::Ascii("malicius"), &["malicious"]),
            (dictgen::InsensitiveStr::Ascii("incerement"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("extenisble"),
                &["extensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publich"),
                &["publish", "public"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideologisk"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("annonate"), &["annotate"]),
            (dictgen::InsensitiveStr::Ascii("tooptip"), &["tooltip"]),
            (
                dictgen::InsensitiveStr::Ascii("prdpagated"),
                &["propagated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skillshosts"),
                &["skillshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("leaglize"), &["legalize"]),
            (dictgen::InsensitiveStr::Ascii("pedastal"), &["pedestal"]),
            (dictgen::InsensitiveStr::Ascii("sastisfy"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("checpoint"), &["checkpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("compataibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenteative"), &["tentative"]),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tham"), &["than"]),
            (
                dictgen::InsensitiveStr::Ascii("subtarcting"),
                &["subtracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("descussed"), &["discussed"]),
            (
                dictgen::InsensitiveStr::Ascii("indicateds"),
                &["indicated", "indicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techticians"),
                &["technicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("datset"), &["dataset"]),
            (dictgen::InsensitiveStr::Ascii("cacheing"), &["caching"]),
            (dictgen::InsensitiveStr::Ascii("eis"), &["is"]),
            (
                dictgen::InsensitiveStr::Ascii("unimpresssed"),
                &["unimpressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pagentry"),
                &["pageantry", "plangently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("academcially"),
                &["academically"],
            ),
            (dictgen::InsensitiveStr::Ascii("smpt"), &["smtp", "smtpe"]),
            (
                dictgen::InsensitiveStr::Ascii("christianty"),
                &["christianity"],
            ),
            (dictgen::InsensitiveStr::Ascii("rescaned"), &["rescanned"]),
            (dictgen::InsensitiveStr::Ascii("pendulm"), &["pendulum"]),
            (dictgen::InsensitiveStr::Ascii("intitials"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("infom"), &["inform"]),
            (
                dictgen::InsensitiveStr::Ascii("interprester"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("develoeprs"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("situationly"),
                &["situational", "situationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differenes"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoncement"),
                &["announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decentralisin"),
                &["decentralising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalismus"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difractive"),
                &["diffractive"],
            ),
            (dictgen::InsensitiveStr::Ascii("recipets"), &["recipes"]),
            (
                dictgen::InsensitiveStr::Ascii("repositaries"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("legalizare"), &["legalize"]),
            (
                dictgen::InsensitiveStr::Ascii("requiements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instutution"),
                &["institution"],
            ),
            (dictgen::InsensitiveStr::Ascii("ervery"), &["every"]),
            (
                dictgen::InsensitiveStr::Ascii("rebouncing"),
                &["rebounding"],
            ),
            (dictgen::InsensitiveStr::Ascii("diven"), &["given"]),
            (dictgen::InsensitiveStr::Ascii("moveing"), &["moving"]),
            (
                dictgen::InsensitiveStr::Ascii("registeresd"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entrepreneuers"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarmatically"),
                &["programmatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasportable"),
                &["transportable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibitity"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("recogise"), &["recognise"]),
            (dictgen::InsensitiveStr::Ascii("bigtoed"), &["bigoted"]),
            (
                dictgen::InsensitiveStr::Ascii("transcirpt"),
                &["transcripts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particpate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("pagnation"), &["pagination"]),
            (
                dictgen::InsensitiveStr::Ascii("playgropunds"),
                &["playgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckecklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("subsequenlty"),
                &["subsequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("consoel"), &["console"]),
            (dictgen::InsensitiveStr::Ascii("molelcule"), &["molecule"]),
            (
                dictgen::InsensitiveStr::Ascii("prosicuted"),
                &["prosecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("takin"),
                &["taking", "taken", "akin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexectedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsituting"),
                &["substituting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecifications"),
                &["specifications", "pacifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylzes"), &["analyzes"]),
            (dictgen::InsensitiveStr::Ascii("hostogram"), &["histogram"]),
            (dictgen::InsensitiveStr::Ascii("iproves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("napomes"), &["napalms"]),
            (dictgen::InsensitiveStr::Ascii("wrestlewar"), &["wrestler"]),
            (dictgen::InsensitiveStr::Ascii("massacrare"), &["massacre"]),
            (
                dictgen::InsensitiveStr::Ascii("armageddomon"),
                &["armageddon"],
            ),
            (dictgen::InsensitiveStr::Ascii("projectd"), &["projected"]),
            (dictgen::InsensitiveStr::Ascii("accurrate"), &["accurate"]),
            (dictgen::InsensitiveStr::Ascii("gobally"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("gammeode"), &["gamemode"]),
            (
                dictgen::InsensitiveStr::Ascii("sme"),
                &["some", "same", "seme", "sms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investement"),
                &["investment"],
            ),
            (dictgen::InsensitiveStr::Ascii("unneeeded"), &["unneeded"]),
            (dictgen::InsensitiveStr::Ascii("containr"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("cincinnnati"),
                &["cincinnati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempreature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("kitites"), &["kitties"]),
            (
                dictgen::InsensitiveStr::Ascii("buisnessmen"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("valsues"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("sponsord"), &["sponsored"]),
            (dictgen::InsensitiveStr::Ascii("buillder"), &["builder"]),
            (dictgen::InsensitiveStr::Ascii("receievers"), &["receivers"]),
            (
                dictgen::InsensitiveStr::Ascii("hyperldger"),
                &["hyperledger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serveillance"),
                &["surveillance"],
            ),
            (dictgen::InsensitiveStr::Ascii("irleand"), &["ireland"]),
            (dictgen::InsensitiveStr::Ascii("asssume"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("bouded"), &["bounded"]),
            (
                dictgen::InsensitiveStr::Ascii("thetraedral"),
                &["tetrahedral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ehther"),
                &["ether", "either"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificatons"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("definit"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("expetcting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("complination"),
                &["complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("presssured"), &["pressured"]),
            (dictgen::InsensitiveStr::Ascii("orelans"), &["orleans"]),
            (
                dictgen::InsensitiveStr::Ascii("diskretize"),
                &["discretize"],
            ),
            (dictgen::InsensitiveStr::Ascii("busrting"), &["bursting"]),
            (
                dictgen::InsensitiveStr::Ascii("nurishment"),
                &["nourishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("collums"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("multliples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("believin"), &["believing"]),
            (dictgen::InsensitiveStr::Ascii("dereived"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("sotware"), &["software"]),
            (dictgen::InsensitiveStr::Ascii("stabilzes"), &["stabilizes"]),
            (dictgen::InsensitiveStr::Ascii("accordign"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("excections"),
                &["exceptions", "executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("achivable"), &["achievable"]),
            (dictgen::InsensitiveStr::Ascii("incomng"), &["incoming"]),
            (dictgen::InsensitiveStr::Ascii("tsunamai"), &["tsunami"]),
            (dictgen::InsensitiveStr::Ascii("secnd"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("pedopholia"),
                &["pedophilia"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesult"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("termiates"),
                &["terminates", "termites"],
            ),
            (dictgen::InsensitiveStr::Ascii("principl"), &["principals"]),
            (dictgen::InsensitiveStr::Ascii("failtures"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("denomenation"),
                &["denomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciefics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("negogtiators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("esponders"), &["responders"]),
            (
                dictgen::InsensitiveStr::Ascii("psichologist"),
                &["psychologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("oveflows"), &["overflows"]),
            (
                dictgen::InsensitiveStr::Ascii("experinementations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undoubtebly"),
                &["undoubtedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("studing"), &["studying"]),
            (
                dictgen::InsensitiveStr::Ascii("witnissing"),
                &["witnessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouces"), &["bounces"]),
            (dictgen::InsensitiveStr::Ascii("compilare"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("dumplication"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("defendes"), &["defenders"]),
            (
                dictgen::InsensitiveStr::Ascii("imtimidation"),
                &["intimidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppplies"), &["supplies"]),
            (dictgen::InsensitiveStr::Ascii("triggeresd"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("unreproducble"),
                &["unreproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorported"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("utlity"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("spefcifiy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("plarising"), &["polarising"]),
            (dictgen::InsensitiveStr::Ascii("contry"), &["country"]),
            (dictgen::InsensitiveStr::Ascii("isalmic"), &["islamic"]),
            (dictgen::InsensitiveStr::Ascii("nestolgia"), &["nostalgia"]),
            (dictgen::InsensitiveStr::Ascii("lieuenant"), &["lieutenant"]),
            (
                dictgen::InsensitiveStr::Ascii("girlfirend"),
                &["girlfriend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moisterizing"),
                &["moisturizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("follws"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("ppropriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactorers"),
                &["manufacturers"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlcuding"), &["excluding"]),
            (dictgen::InsensitiveStr::Ascii("lacker"), &["lacquer"]),
            (
                dictgen::InsensitiveStr::Ascii("independente"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbborhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paticipates"),
                &["participates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punctutation"),
                &["punctuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("porper"), &["proper"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretaiton"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catagories"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("pixell"), &["pixel"]),
            (dictgen::InsensitiveStr::Ascii("sublasse"), &["subclasse"]),
            (dictgen::InsensitiveStr::Ascii("commnets"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("integal"), &["integral"]),
            (
                dictgen::InsensitiveStr::Ascii("destruciton"),
                &["destruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("zoocheeni"), &["zucchini"]),
            (
                dictgen::InsensitiveStr::Ascii("questionnair"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistentcies"),
                &["inconsistencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("degenrate"), &["degenerate"]),
            (
                dictgen::InsensitiveStr::Ascii("priviledges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoucnement"),
                &["announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trinagular"),
                &["triangular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heatlhcare"),
                &["healthcare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("popuplarity"),
                &["popularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("facour"), &["favour"]),
            (dictgen::InsensitiveStr::Ascii("thirites"), &["thirties"]),
            (dictgen::InsensitiveStr::Ascii("enginin"), &["engine"]),
            (
                dictgen::InsensitiveStr::Ascii("civalisations"),
                &["civilisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconeccted"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixel"), &["pixel"]),
            (dictgen::InsensitiveStr::Ascii("immerdiate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("creulty"), &["cruelty"]),
            (dictgen::InsensitiveStr::Ascii("reanmes"), &["renames"]),
            (
                dictgen::InsensitiveStr::Ascii("camofluage"),
                &["camouflage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previoously"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possabilties"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shatows"),
                &["châteaus", "châteaux", "shadows"],
            ),
            (dictgen::InsensitiveStr::Ascii("distase"), &["distaste"]),
            (dictgen::InsensitiveStr::Ascii("vizualize"), &["visualize"]),
            (
                dictgen::InsensitiveStr::Ascii("transormers"),
                &["transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("templat"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinants"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("unuseable"), &["unusable"]),
            (dictgen::InsensitiveStr::Ascii("seriban"), &["serbian"]),
            (dictgen::InsensitiveStr::Ascii("althoug"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("oppertunist"),
                &["opportunist"],
            ),
            (dictgen::InsensitiveStr::Ascii("painiting"), &["painting"]),
            (dictgen::InsensitiveStr::Ascii("copright"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("checckers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("bombardeado"),
                &["bombarded"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolkein"), &["tolkien"]),
            (
                dictgen::InsensitiveStr::Ascii("attriubutions"),
                &["attributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politicion"),
                &["politician"],
            ),
            (dictgen::InsensitiveStr::Ascii("apparant"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("recipees"), &["recipes"]),
            (
                dictgen::InsensitiveStr::Ascii("differenecs"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reall"),
                &["real", "really", "recall"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortuante"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("licesnsing"), &["licensing"]),
            (dictgen::InsensitiveStr::Ascii("protestans"), &["protests"]),
            (dictgen::InsensitiveStr::Ascii("iritable"), &["irritable"]),
            (
                dictgen::InsensitiveStr::Ascii("generaging"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("recyclin"), &["recycling"]),
            (
                dictgen::InsensitiveStr::Ascii("prononciation"),
                &["pronunciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("blockse"), &["blocks"]),
            (
                dictgen::InsensitiveStr::Ascii("signguature"),
                &["signature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("billionaries"),
                &["billionaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainifest"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("milennia"), &["millennia"]),
            (
                dictgen::InsensitiveStr::Ascii("isntallation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mcalren"), &["mclaren"]),
            (dictgen::InsensitiveStr::Ascii("adheared"), &["adhered"]),
            (dictgen::InsensitiveStr::Ascii("appluase"), &["applause"]),
            (dictgen::InsensitiveStr::Ascii("datashhet"), &["datasheet"]),
            (
                dictgen::InsensitiveStr::Ascii("continouty"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrariliy"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("objction"), &["objection"]),
            (dictgen::InsensitiveStr::Ascii("opnegroup"), &["opengroup"]),
            (
                dictgen::InsensitiveStr::Ascii("orieations"),
                &["orientations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("christpher"),
                &["christopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exoressuin"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("editer"),
                &["editor", "edited", "edit"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchane"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("trivias"), &["trivia"]),
            (dictgen::InsensitiveStr::Ascii("frequecy"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("eariest"),
                &["earliest", "easiest", "eeriest"],
            ),
            (dictgen::InsensitiveStr::Ascii("hosited"), &["hoisted"]),
            (
                dictgen::InsensitiveStr::Ascii("peridically"),
                &["periodically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeusted"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("assicieting"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortitue"), &["fortitude"]),
            (dictgen::InsensitiveStr::Ascii("grpahed"), &["graphed"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitifed"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("multuples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("publisers"), &["publishers"]),
            (
                dictgen::InsensitiveStr::Ascii("captialize"),
                &["capitalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rimanissent"),
                &["reminiscent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multlications"),
                &["multiplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("pimxap"), &["pixmap"]),
            (dictgen::InsensitiveStr::Ascii("witthin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("inaktively"),
                &["inactively"],
            ),
            (dictgen::InsensitiveStr::Ascii("colllect"), &["collect"]),
            (
                dictgen::InsensitiveStr::Ascii("singificand"),
                &["significand", "significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("glyhs"), &["glyphs"]),
            (dictgen::InsensitiveStr::Ascii("scafoldes"), &["scaffolds"]),
            (
                dictgen::InsensitiveStr::Ascii("considereable"),
                &["considerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaoid"),
                &["avoid", "valid", "void", "vapid"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transtioned"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("yse"),
                &["yes", "use", "nyse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qeustioned"),
                &["questioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autcoimpletion"),
                &["autocompletion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derogitory"),
                &["derogatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associtated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("porftolio"), &["portfolio"]),
            (dictgen::InsensitiveStr::Ascii("selctable"), &["selectable"]),
            (dictgen::InsensitiveStr::Ascii("multixsite"), &["multisite"]),
            (
                dictgen::InsensitiveStr::Ascii("prgressions"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ahtleticism"),
                &["athleticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("startupbus"), &["startups"]),
            (dictgen::InsensitiveStr::Ascii("unfilp"), &["unflip"]),
            (dictgen::InsensitiveStr::Ascii("tempraay"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbethood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decopresses"),
                &["decompresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaition"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("messasges"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("spltter"), &["splitter"]),
            (
                dictgen::InsensitiveStr::Ascii("accoundingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("notifyes"), &["notifies"]),
            (dictgen::InsensitiveStr::Ascii("postitive"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("inclusing"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("polinomials"),
                &["polynomials"],
            ),
            (dictgen::InsensitiveStr::Ascii("faseeshus"), &["facetious"]),
            (dictgen::InsensitiveStr::Ascii("tutorals"), &["tutorials"]),
            (
                dictgen::InsensitiveStr::Ascii("disresepct"),
                &["disrespect"],
            ),
            (dictgen::InsensitiveStr::Ascii("effulence"), &["effluence"]),
            (dictgen::InsensitiveStr::Ascii("exexutors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("preferentes"),
                &["preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotions"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssembler"), &["assembler"]),
            (dictgen::InsensitiveStr::Ascii("publisger"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("athists"), &["atheists"]),
            (dictgen::InsensitiveStr::Ascii("foreard"), &["forward"]),
            (dictgen::InsensitiveStr::Ascii("marz"), &["march", "mars"]),
            (
                dictgen::InsensitiveStr::Ascii("dffers"),
                &["differs", "defers", "offers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downloaed"),
                &["downloaded", "download"],
            ),
            (dictgen::InsensitiveStr::Ascii("aply"), &["apply"]),
            (
                dictgen::InsensitiveStr::Ascii("neighourhood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("callpadding"),
                &["cellpadding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organisating"),
                &["organising"],
            ),
            (dictgen::InsensitiveStr::Ascii("saggital"), &["sagittal"]),
            (
                dictgen::InsensitiveStr::Ascii("implmenetd"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("encryting"), &["encrypting"]),
            (
                dictgen::InsensitiveStr::Ascii("soustraction"),
                &["subtraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("negatice"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("exsisted"), &["existed"]),
            (dictgen::InsensitiveStr::Ascii("querstions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("responser"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("norwegain"), &["norwegian"]),
            (dictgen::InsensitiveStr::Ascii("scratchs"), &["scratches"]),
            (dictgen::InsensitiveStr::Ascii("sufixes"), &["suffixes"]),
            (
                dictgen::InsensitiveStr::Ascii("initalizers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinply"), &["simply"]),
            (dictgen::InsensitiveStr::Ascii("ideintify"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("excecutors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("restaurnad"),
                &["restaurant"],
            ),
            (dictgen::InsensitiveStr::Ascii("defnite"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("prettyier"), &["prettier"]),
            (
                dictgen::InsensitiveStr::Ascii("bevorehand"),
                &["beforehand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrbutive"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("thredhold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("inclinacion"),
                &["inclination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relms"),
                &["elms", "helms", "realms", "reams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southernest"),
                &["southerners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furtheremor"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degenererat"),
                &["degenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("workbennch"), &["workbench"]),
            (
                dictgen::InsensitiveStr::Ascii("appriceated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeed"), &["exceed"]),
            (dictgen::InsensitiveStr::Ascii("supperts"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("humiliatin"),
                &["humiliation"],
            ),
            (dictgen::InsensitiveStr::Ascii("attachd"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("academicas"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("loacalizes"), &["localizes"]),
            (
                dictgen::InsensitiveStr::Ascii("exprations"),
                &["expirations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cassowarry"), &["cassowary"]),
            (dictgen::InsensitiveStr::Ascii("adventrue"), &["adventure"]),
            (dictgen::InsensitiveStr::Ascii("pssible"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("uggests"), &["suggests"]),
            (
                dictgen::InsensitiveStr::Ascii("techinicians"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coallecing"),
                &["coalescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintensional"),
                &["unintentional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immatureity"),
                &["immaturity"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabl"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("comparigon"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("dosseays"), &["dossiers"]),
            (dictgen::InsensitiveStr::Ascii("similir"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("registeres"), &["registers"]),
            (dictgen::InsensitiveStr::Ascii("coloful"), &["colorful"]),
            (
                dictgen::InsensitiveStr::Ascii("segretated"),
                &["segregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techinologies"),
                &["technologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermomenter"),
                &["thermometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("househols"), &["households"]),
            (
                dictgen::InsensitiveStr::Ascii("theoeretically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantiative"),
                &["quantitative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("additoinaly"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("panles"), &["panels"]),
            (dictgen::InsensitiveStr::Ascii("pasuing"), &["pausing"]),
            (dictgen::InsensitiveStr::Ascii("callong"), &["calling"]),
            (dictgen::InsensitiveStr::Ascii("workaroud"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("disgard"), &["discard"]),
            (dictgen::InsensitiveStr::Ascii("esily"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("frivilous"), &["frivolous"]),
            (
                dictgen::InsensitiveStr::Ascii("commentars"),
                &["commenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphatetic"),
                &["empathetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlpoit"), &["exploit"]),
            (dictgen::InsensitiveStr::Ascii("iteraror"), &["iterator"]),
            (
                dictgen::InsensitiveStr::Ascii("interoprability"),
                &["interoperability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standars"),
                &["standards", "standard"],
            ),
            (dictgen::InsensitiveStr::Ascii("susseeds"), &["secedes"]),
            (dictgen::InsensitiveStr::Ascii("paradimes"), &["paradise"]),
            (dictgen::InsensitiveStr::Ascii("scavanges"), &["scavenges"]),
            (dictgen::InsensitiveStr::Ascii("appendign"), &["appending"]),
            (
                dictgen::InsensitiveStr::Ascii("algorihtms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swtichable"),
                &["switchable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palceholder"),
                &["placeholder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspectible"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("finings"), &["findings"]),
            (
                dictgen::InsensitiveStr::Ascii("islamistisk"),
                &["islamists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aligh"),
                &["align", "alight"],
            ),
            (dictgen::InsensitiveStr::Ascii("caloires"), &["calories"]),
            (
                dictgen::InsensitiveStr::Ascii("staurating"),
                &["saturating"],
            ),
            (dictgen::InsensitiveStr::Ascii("postifx"), &["postfix"]),
            (
                dictgen::InsensitiveStr::Ascii("automotion"),
                &["automation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperianced"),
                &["inexperienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconeting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneption"), &["conception"]),
            (dictgen::InsensitiveStr::Ascii("intquiries"), &["inquiries"]),
            (
                dictgen::InsensitiveStr::Ascii("tessleators"),
                &["tessellators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissipointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaiminate"),
                &["contaminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacemnet"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimmic"), &["mimic"]),
            (
                dictgen::InsensitiveStr::Ascii("privelages"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("mchanical"), &["mechanical"]),
            (dictgen::InsensitiveStr::Ascii("computd"), &["computed"]),
            (dictgen::InsensitiveStr::Ascii("regax"), &["regex"]),
            (dictgen::InsensitiveStr::Ascii("uesful"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("intolerate"),
                &["intolerant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exexutions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuslted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("distint"),
                &["distinct", "distant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrioduced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trogloditic"),
                &["troglodytic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playthorughs"),
                &["playthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apperently"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disputandem"),
                &["disputandum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commnunicating"),
                &["communicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auxialliary"),
                &["auxiliary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elimimates"),
                &["eliminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidentily"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("assersed"), &["asserted"]),
            (dictgen::InsensitiveStr::Ascii("fanaticas"), &["fanatics"]),
            (
                dictgen::InsensitiveStr::Ascii("exploting"),
                &["exploiting", "exploring"],
            ),
            (dictgen::InsensitiveStr::Ascii("connecing"), &["connecting"]),
            (dictgen::InsensitiveStr::Ascii("bergamont"), &["bergamot"]),
            (
                dictgen::InsensitiveStr::Ascii("finshes"),
                &["finishes", "finches"],
            ),
            (dictgen::InsensitiveStr::Ascii("passivedns"), &["passives"]),
            (
                dictgen::InsensitiveStr::Ascii("rollercoatser"),
                &["rollercoaster"],
            ),
            (dictgen::InsensitiveStr::Ascii("thatt"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("pitchs"), &["pitches"]),
            (
                dictgen::InsensitiveStr::Ascii("iteractions"),
                &["interactions", "iterations"],
            ),
            (dictgen::InsensitiveStr::Ascii("elsof"), &["elseif"]),
            (dictgen::InsensitiveStr::Ascii("uniquenes"), &["uniqueness"]),
            (
                dictgen::InsensitiveStr::Ascii("delcarations"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("concened"), &["concede"]),
            (dictgen::InsensitiveStr::Ascii("forsting"), &["frosting"]),
            (dictgen::InsensitiveStr::Ascii("traned"), &["trained"]),
            (dictgen::InsensitiveStr::Ascii("tral"), &["trail", "trial"]),
            (
                dictgen::InsensitiveStr::Ascii("attandance"),
                &["attendance"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiints"), &["points"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecessarly"),
                &["unnecessarily", "unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrariy"),
                &["arbitrarily", "arbitrary"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapenned"), &["happened"]),
            (dictgen::InsensitiveStr::Ascii("israelli"), &["israeli"]),
            (
                dictgen::InsensitiveStr::Ascii("interect"),
                &["interacted", "interact", "intersect"],
            ),
            (dictgen::InsensitiveStr::Ascii("nofitied"), &["notified"]),
            (dictgen::InsensitiveStr::Ascii("subgregion"), &["subregion"]),
            (dictgen::InsensitiveStr::Ascii("thnig"), &["thing"]),
            (dictgen::InsensitiveStr::Ascii("boudn"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("juge"), &["judge"]),
            (dictgen::InsensitiveStr::Ascii("neightobr"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("romanna"), &["romanian"]),
            (
                dictgen::InsensitiveStr::Ascii("disasterous"),
                &["disastrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoiciating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerations"), &["creations"]),
            (dictgen::InsensitiveStr::Ascii("allways"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("statamenet"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("dabue"), &["debut"]),
            (dictgen::InsensitiveStr::Ascii("inerta"), &["inertia"]),
            (dictgen::InsensitiveStr::Ascii("fianlized"), &["finalized"]),
            (dictgen::InsensitiveStr::Ascii("scheduel"), &["schedule"]),
            (dictgen::InsensitiveStr::Ascii("coalesed"), &["coalesced"]),
            (dictgen::InsensitiveStr::Ascii("synctax"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("rquired"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("juxtopose"), &["juxtapose"]),
            (
                dictgen::InsensitiveStr::Ascii("effctively"),
                &["effectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("targteing"), &["targeting"]),
            (
                dictgen::InsensitiveStr::Ascii("unspefixifed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotication"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("emprove"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("genearator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiasic"),
                &["enthusiastic"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomatter"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("companines"),
                &["companions"],
            ),
            (dictgen::InsensitiveStr::Ascii("hava"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("sucesful"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("scrrens"), &["screens"]),
            (dictgen::InsensitiveStr::Ascii("exent"), &["extent"]),
            (
                dictgen::InsensitiveStr::Ascii("stetching"),
                &["stretching", "sketching", "stitching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integrading"),
                &["integrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("reson"), &["reason"]),
            (dictgen::InsensitiveStr::Ascii("indiciated"), &["indicated"]),
            (dictgen::InsensitiveStr::Ascii("cocktials"), &["cocktails"]),
            (
                dictgen::InsensitiveStr::Ascii("authethenticators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("colinear"), &["collinear"]),
            (
                dictgen::InsensitiveStr::Ascii("differencial"),
                &["differential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotitaions"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automomous"),
                &["autonomous"],
            ),
            (dictgen::InsensitiveStr::Ascii("movemet"), &["movement"]),
            (dictgen::InsensitiveStr::Ascii("osterages"), &["ostriches"]),
            (dictgen::InsensitiveStr::Ascii("westbroook"), &["westbrook"]),
            (dictgen::InsensitiveStr::Ascii("bcak"), &["back"]),
            (
                dictgen::InsensitiveStr::Ascii("puplication"),
                &["publication", "duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indirectely"),
                &["indirectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hten"),
                &["then", "hen", "the"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoroporating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("inerrupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("recepion"), &["reception"]),
            (
                dictgen::InsensitiveStr::Ascii("kilometeres"),
                &["kilometers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termperatues"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successfullness"),
                &["successfulness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innitializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diskretion"),
                &["discretion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaenous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoulute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("structer"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("indes"), &["index"]),
            (dictgen::InsensitiveStr::Ascii("prision"), &["prison"]),
            (
                dictgen::InsensitiveStr::Ascii("spacializes"),
                &["specializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("opject"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("enconde"), &["encode"]),
            (
                dictgen::InsensitiveStr::Ascii("phenomenol"),
                &["phenomenal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperately"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deomcracies"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predecesores"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("altantic"), &["atlantic"]),
            (dictgen::InsensitiveStr::Ascii("mileu"), &["milieu"]),
            (dictgen::InsensitiveStr::Ascii("bimontly"), &["bimonthly"]),
            (dictgen::InsensitiveStr::Ascii("hypocricy"), &["hypocrisy"]),
            (dictgen::InsensitiveStr::Ascii("secondadry"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("falsley"), &["falsely"]),
            (
                dictgen::InsensitiveStr::Ascii("prefessionalism"),
                &["professionalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sythesizing"),
                &["synthesizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("eventuall"), &["eventually"]),
            (dictgen::InsensitiveStr::Ascii("allcoator"), &["allocator"]),
            (
                dictgen::InsensitiveStr::Ascii("chaging"),
                &["changing", "charging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possesseurs"),
                &["possesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("procide"), &["provide"]),
            (
                dictgen::InsensitiveStr::Ascii("computacion"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("illgal"), &["illegal"]),
            (
                dictgen::InsensitiveStr::Ascii("authos"),
                &["authors", "autos"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactere"),
                &["characterize", "character"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proffessions"),
                &["professions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsatifactory"),
                &["unsatisfactory"],
            ),
            (dictgen::InsensitiveStr::Ascii("patiens"), &["patients"]),
            (
                dictgen::InsensitiveStr::Ascii("experminental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("covection"), &["convection"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilitis"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("videogamemes"),
                &["videogames"],
            ),
            (dictgen::InsensitiveStr::Ascii("tecnology"), &["technology"]),
            (
                dictgen::InsensitiveStr::Ascii("producable"),
                &["producible"],
            ),
            (dictgen::InsensitiveStr::Ascii("sergent"), &["sergeant"]),
            (
                dictgen::InsensitiveStr::Ascii("apdated"),
                &["updated", "adapted"],
            ),
            (dictgen::InsensitiveStr::Ascii("jersyes"), &["jerseys"]),
            (dictgen::InsensitiveStr::Ascii("mesasage"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("ventilato"),
                &["ventilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("febuary"), &["february"]),
            (dictgen::InsensitiveStr::Ascii("childrenis"), &["childrens"]),
            (dictgen::InsensitiveStr::Ascii("wythout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("failiures"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("promtional"),
                &["promotional"],
            ),
            (dictgen::InsensitiveStr::Ascii("tarces"), &["traces"]),
            (dictgen::InsensitiveStr::Ascii("coutndown"), &["countdown"]),
            (dictgen::InsensitiveStr::Ascii("uretrha"), &["urethra"]),
            (dictgen::InsensitiveStr::Ascii("simlify"), &["simplify"]),
            (
                dictgen::InsensitiveStr::Ascii("innovaiton"),
                &["innovations"],
            ),
            (dictgen::InsensitiveStr::Ascii("abselutly"), &["absolutely"]),
            (dictgen::InsensitiveStr::Ascii("modell"), &["model"]),
            (
                dictgen::InsensitiveStr::Ascii("electroylte"),
                &["electrolytes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitions"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("privilge"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("arguemnt"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("nastersiem"),
                &["nasturtium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrinisics"),
                &["intrinsics"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceduling"), &["scheduling"]),
            (dictgen::InsensitiveStr::Ascii("launhers"), &["launchers"]),
            (dictgen::InsensitiveStr::Ascii("kinnect"), &["kinect"]),
            (
                dictgen::InsensitiveStr::Ascii("screenshort"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciritics"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("lengthg"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("transproting"),
                &["transporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("divertions"),
                &["diversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("flordia"), &["florida"]),
            (
                dictgen::InsensitiveStr::Ascii("degenarate"),
                &["degenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("definding"), &["defining"]),
            (
                dictgen::InsensitiveStr::Ascii("orientatied"),
                &["orientated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millimitre"),
                &["millimetre"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiesses"),
                &["acquiesces"],
            ),
            (dictgen::InsensitiveStr::Ascii("socalism"), &["socialism"]),
            (
                dictgen::InsensitiveStr::Ascii("authencates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("documened"), &["documented"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangtes"),
                &["rearranges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjuntions"),
                &["conjunctions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperiencable"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("deplyoing"), &["deploying"]),
            (dictgen::InsensitiveStr::Ascii("frowarded"), &["forwarded"]),
            (
                dictgen::InsensitiveStr::Ascii("libertarainism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("notitying"), &["notifying"]),
            (dictgen::InsensitiveStr::Ascii("constast"), &["constants"]),
            (
                dictgen::InsensitiveStr::Ascii("transplain"),
                &["transplant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similiarities"),
                &["similarities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possiilities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("preciesly"), &["precisely"]),
            (
                dictgen::InsensitiveStr::Ascii("deperecated"),
                &["deprecated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ccertificates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("censequence"),
                &["consequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheeved"), &["achieved"]),
            (
                dictgen::InsensitiveStr::Ascii("cataclismical"),
                &["cataclysmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("varaitions"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("membrans"), &["membranes"]),
            (dictgen::InsensitiveStr::Ascii("monitoing"), &["monitoring"]),
            (dictgen::InsensitiveStr::Ascii("sercurity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("introdus"), &["introduces"]),
            (dictgen::InsensitiveStr::Ascii("santurom"), &["santorum"]),
            (
                dictgen::InsensitiveStr::Ascii("extacting"),
                &["extracting", "exacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportianal"),
                &["proportional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispraportionately"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("itearting"), &["iterating"]),
            (dictgen::InsensitiveStr::Ascii("bioplar"), &["bipolar"]),
            (dictgen::InsensitiveStr::Ascii("thatn"), &["that", "than"]),
            (dictgen::InsensitiveStr::Ascii("censibly"), &["sensibly"]),
            (
                dictgen::InsensitiveStr::Ascii("overhooking"),
                &["overlooking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorizing"),
                &["prioritizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("prviacy"), &["privacy"]),
            (dictgen::InsensitiveStr::Ascii("applaudes"), &["applause"]),
            (dictgen::InsensitiveStr::Ascii("undet"), &["under"]),
            (dictgen::InsensitiveStr::Ascii("peformant"), &["performant"]),
            (dictgen::InsensitiveStr::Ascii("brocolli"), &["broccoli"]),
            (
                dictgen::InsensitiveStr::Ascii("presidencial"),
                &["presidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("crockadile"), &["crocodile"]),
            (
                dictgen::InsensitiveStr::Ascii("neumonically"),
                &["mnemonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagerrate"),
                &["exaggerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("assingee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("brednan"), &["brendan"]),
            (
                dictgen::InsensitiveStr::Ascii("depecreated"),
                &["deprecated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qaurterback"),
                &["quarterback"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnovert"), &["convert"]),
            (
                dictgen::InsensitiveStr::Ascii("inadvertantly"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregetor"),
                &["aggregator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitinely"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("harth"), &["hearth"]),
            (dictgen::InsensitiveStr::Ascii("assocate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("messurement"),
                &["measurement"],
            ),
            (dictgen::InsensitiveStr::Ascii("pleaase"), &["please"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitratrion"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("postitions"), &["positions"]),
            (
                dictgen::InsensitiveStr::Ascii("adventureres"),
                &["adventures", "adventurers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacward"), &["backward"]),
            (
                dictgen::InsensitiveStr::Ascii("initiializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coloering"),
                &["coloring", "colouring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrepetitious"),
                &["surreptitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("travering"), &["traversing"]),
            (
                dictgen::InsensitiveStr::Ascii("contributuions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("prohpet"), &["prophet"]),
            (dictgen::InsensitiveStr::Ascii("bootom"), &["bottom"]),
            (dictgen::InsensitiveStr::Ascii("repubics"), &["republics"]),
            (
                dictgen::InsensitiveStr::Ascii("milktoast"),
                &["milquetoast"],
            ),
            (dictgen::InsensitiveStr::Ascii("proctected"), &["protected"]),
            (dictgen::InsensitiveStr::Ascii("ckeckbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("recends"), &["rescinds"]),
            (dictgen::InsensitiveStr::Ascii("messagess"), &["messages"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomamce"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ctification"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antarcitca"),
                &["antarctica"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivaletn"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selt"),
                &["set", "self", "sold"],
            ),
            (dictgen::InsensitiveStr::Ascii("jennins"), &["jennings"]),
            (dictgen::InsensitiveStr::Ascii("annnounce"), &["announce"]),
            (dictgen::InsensitiveStr::Ascii("buillders"), &["builders"]),
            (
                dictgen::InsensitiveStr::Ascii("preson"),
                &["person", "prison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chekcpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("inejction"), &["injection"]),
            (
                dictgen::InsensitiveStr::Ascii("commanists"),
                &["communists"],
            ),
            (dictgen::InsensitiveStr::Ascii("addrress"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiififer"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("cermonies"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("vocabularly"),
                &["vocabulary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminstrate"),
                &["administrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("decendant"), &["descendant"]),
            (
                dictgen::InsensitiveStr::Ascii("catholisism"),
                &["catholicism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collasping"),
                &["collapsing"],
            ),
            (dictgen::InsensitiveStr::Ascii("oringally"), &["originally"]),
            (
                dictgen::InsensitiveStr::Ascii("algotirhmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemetation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciltoris"), &["clitoris"]),
            (
                dictgen::InsensitiveStr::Ascii("diferring"),
                &["differing", "deferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfacion"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("enyway"), &["anyway"]),
            (dictgen::InsensitiveStr::Ascii("matzch"), &["match"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deregisted"),
                &["deregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresponsibe"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("aftre"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("apeendices"),
                &["appendices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consttructer"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professins"),
                &["professions"],
            ),
            (dictgen::InsensitiveStr::Ascii("flaot"), &["float"]),
            (dictgen::InsensitiveStr::Ascii("sacalar"), &["scalar"]),
            (dictgen::InsensitiveStr::Ascii("uffer"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("rapsberry"), &["raspberry"]),
            (
                dictgen::InsensitiveStr::Ascii("initilisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("feturing"), &["featuring"]),
            (
                dictgen::InsensitiveStr::Ascii("rsource"),
                &["resource", "source"],
            ),
            (dictgen::InsensitiveStr::Ascii("unusued"), &["unused"]),
            (
                dictgen::InsensitiveStr::Ascii("interpereters"),
                &["interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orchestarted"),
                &["orchestrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommedation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpctedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directtions"),
                &["directions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyzies"),
                &["analysis", "analyzes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytles"), &["styles"]),
            (dictgen::InsensitiveStr::Ascii("pateints"), &["patients"]),
            (dictgen::InsensitiveStr::Ascii("extarnal"), &["external"]),
            (dictgen::InsensitiveStr::Ascii("awkwardy"), &["awkwardly"]),
            (
                dictgen::InsensitiveStr::Ascii("cipherntext"),
                &["ciphertext"],
            ),
            (dictgen::InsensitiveStr::Ascii("spearates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("remander"), &["remainder"]),
            (
                dictgen::InsensitiveStr::Ascii("ccordinate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("referece"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("congfigs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("depcrecation"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("itatic"), &["italic"]),
            (dictgen::InsensitiveStr::Ascii("exclosed"), &["enclosed"]),
            (
                dictgen::InsensitiveStr::Ascii("lnowledgable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("constrauct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("rehtoric"), &["rhetoric"]),
            (
                dictgen::InsensitiveStr::Ascii("pittsbrugh"),
                &["pittsburgh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyclopdia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("indeces"), &["indices"]),
            (dictgen::InsensitiveStr::Ascii("festivas"), &["festivals"]),
            (dictgen::InsensitiveStr::Ascii("switiched"), &["switched"]),
            (
                dictgen::InsensitiveStr::Ascii("depencency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hospitalty"),
                &["hospitality"],
            ),
            (dictgen::InsensitiveStr::Ascii("folllower"), &["follower"]),
            (
                dictgen::InsensitiveStr::Ascii("apostorophe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summerizes"),
                &["summarizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intenationalists"),
                &["internationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("threadd"), &["threaded"]),
            (dictgen::InsensitiveStr::Ascii("serviec"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("condictionals"),
                &["conditionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteligently"),
                &["intelligently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curcumference"),
                &["circumference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cricumference"),
                &["circumference"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendirer"), &["renderer"]),
            (
                dictgen::InsensitiveStr::Ascii("controles"),
                &["controls", "controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computs"),
                &["computes", "computus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increamented"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constranit"),
                &["constraint"],
            ),
            (dictgen::InsensitiveStr::Ascii("assigneee"), &["assignee"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulatie"),
                &["manipulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("instlals"), &["instals"]),
            (
                dictgen::InsensitiveStr::Ascii("storeis"),
                &["storeys", "stores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transocder"),
                &["transcoder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwridden"),
                &["overridden", "overwritten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrecpectful"),
                &["disrespectful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instruection"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviroments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("exoticos"), &["exotics"]),
            (
                dictgen::InsensitiveStr::Ascii("undetwater"),
                &["underwater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelereate"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extint"),
                &["extinct", "extant"],
            ),
            (dictgen::InsensitiveStr::Ascii("fianite"), &["finite"]),
            (dictgen::InsensitiveStr::Ascii("reccorded"), &["recorded"]),
            (
                dictgen::InsensitiveStr::Ascii("concatentated"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("boardway"), &["broadway"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilites"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcel"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("periodioc"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("behaviuoral"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abourt"),
                &["abort", "about"],
            ),
            (dictgen::InsensitiveStr::Ascii("disalb"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("contrase"), &["contrast"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("talekd"), &["talked"]),
            (dictgen::InsensitiveStr::Ascii("equallity"), &["equality"]),
            (
                dictgen::InsensitiveStr::Ascii("ingnorar"),
                &["ignore", "ignorant"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloeses"), &["closes"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpendiculaire"),
                &["perpendicular"],
            ),
            (dictgen::InsensitiveStr::Ascii("cracing"), &["gracing"]),
            (dictgen::InsensitiveStr::Ascii("langnguage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("computationnal"),
                &["computational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilging"),
                &["privileging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("screebshots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("measuerer"), &["measurer"]),
            (
                dictgen::InsensitiveStr::Ascii("supposeable"),
                &["supposable"],
            ),
            (dictgen::InsensitiveStr::Ascii("currnt"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("cyrstallize"),
                &["crystallize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representiative"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("foudning"), &["founding"]),
            (dictgen::InsensitiveStr::Ascii("runnint"), &["running"]),
            (
                dictgen::InsensitiveStr::Ascii("charactars"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("foramtter"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("carret"), &["caret"]),
            (dictgen::InsensitiveStr::Ascii("metion"), &["mention"]),
            (dictgen::InsensitiveStr::Ascii("especiall"), &["especially"]),
            (
                dictgen::InsensitiveStr::Ascii("checkalaises"),
                &["checkaliases"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteracte"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("aprticle"), &["particle"]),
            (
                dictgen::InsensitiveStr::Ascii("submissons"),
                &["submissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("armeninas"), &["armenians"]),
            (dictgen::InsensitiveStr::Ascii("meditatin"), &["meditation"]),
            (dictgen::InsensitiveStr::Ascii("reasponse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("labyrnith"), &["labyrinth"]),
            (dictgen::InsensitiveStr::Ascii("methhods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("duplaicate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("midfiled"), &["midfield"]),
            (
                dictgen::InsensitiveStr::Ascii("prosperious"),
                &["prosperous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immedietly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkpoins"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoitate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciallist"),
                &["specialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conserning"),
                &["concerning"],
            ),
            (dictgen::InsensitiveStr::Ascii("visiuble"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("statististics"),
                &["statistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabelized"),
                &["stabilized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalet"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crationalist"),
                &["nationalist", "rationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fotography"),
                &["photography"],
            ),
            (dictgen::InsensitiveStr::Ascii("santitizes"), &["sanitizes"]),
            (dictgen::InsensitiveStr::Ascii("councills"), &["councils"]),
            (dictgen::InsensitiveStr::Ascii("oger"), &["ogre"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnaslating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("wreks"), &["wrecks"]),
            (dictgen::InsensitiveStr::Ascii("attivates"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("constrated"),
                &["contrasted", "constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connnecting"),
                &["connecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persective"),
                &["respective", "perspective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neadlessly"),
                &["needlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pscyhologist"),
                &["psychologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("empasses"), &["impasses"]),
            (
                dictgen::InsensitiveStr::Ascii("wavelenghth"),
                &["wavelength"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disssemblies"),
                &["disassemblies"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuilting"), &["resulting"]),
            (dictgen::InsensitiveStr::Ascii("thoerems"), &["theorems"]),
            (dictgen::InsensitiveStr::Ascii("rocord"), &["record"]),
            (
                dictgen::InsensitiveStr::Ascii("andlers"),
                &["handlers", "antlers"],
            ),
            (dictgen::InsensitiveStr::Ascii("thic"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("resrouce"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("specializare"),
                &["specialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("packeging"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("upadaters"), &["updaters"]),
            (
                dictgen::InsensitiveStr::Ascii("caculations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("enconding"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("politing"), &["piloting"]),
            (
                dictgen::InsensitiveStr::Ascii("strategicaly"),
                &["strategically"],
            ),
            (dictgen::InsensitiveStr::Ascii("minnesotta"), &["minnesota"]),
            (dictgen::InsensitiveStr::Ascii("acccessor"), &["accessor"]),
            (dictgen::InsensitiveStr::Ascii("unbannend"), &["unbanned"]),
            (dictgen::InsensitiveStr::Ascii("diggit"), &["digit"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponing"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstinince"),
                &["abstinence"],
            ),
            (dictgen::InsensitiveStr::Ascii("feautured"), &["featured"]),
            (dictgen::InsensitiveStr::Ascii("gatable"), &["gateable"]),
            (
                dictgen::InsensitiveStr::Ascii("cathegorise"),
                &["categorise"],
            ),
            (dictgen::InsensitiveStr::Ascii("reenfoce"), &["reenforce"]),
            (dictgen::InsensitiveStr::Ascii("symapthy"), &["sympathy"]),
            (
                dictgen::InsensitiveStr::Ascii("esponsible"),
                &["responsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("amplifly"), &["amplify"]),
            (
                dictgen::InsensitiveStr::Ascii("stutterring"),
                &["stuttering"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrat"), &["abstract"]),
            (
                dictgen::InsensitiveStr::Ascii("adborb"),
                &["adsorb", "absorb"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifaction"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lithuaninan"),
                &["lithuania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selfishess"),
                &["selfishness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orcestrating"),
                &["orchestrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("birghter"), &["brighter"]),
            (dictgen::InsensitiveStr::Ascii("succedded"), &["succeeded"]),
            (
                dictgen::InsensitiveStr::Ascii("pringing"),
                &["printing", "springing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unencyrpted"),
                &["unencrypted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimialrity"),
                &["dissimilarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("parsialy"), &["partially"]),
            (dictgen::InsensitiveStr::Ascii("extedner"), &["extender"]),
            (
                dictgen::InsensitiveStr::Ascii("sponserships"),
                &["sponsorships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capatilists"),
                &["capitalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("pach"), &["patch", "path"]),
            (
                dictgen::InsensitiveStr::Ascii("universite"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugguests"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("thesitic"), &["theistic"]),
            (
                dictgen::InsensitiveStr::Ascii("accure"),
                &["accrue", "occur", "acquire"],
            ),
            (dictgen::InsensitiveStr::Ascii("enitre"), &["entire"]),
            (dictgen::InsensitiveStr::Ascii("clickd"), &["clicked"]),
            (dictgen::InsensitiveStr::Ascii("inifinty"), &["infinity"]),
            (
                dictgen::InsensitiveStr::Ascii("misspellling"),
                &["misspelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("partiots"), &["patriots"]),
            (
                dictgen::InsensitiveStr::Ascii("diliminator"),
                &["delimiter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silhouettin"),
                &["silhouetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accuraccies"),
                &["accuracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("endlessley"), &["endlessly"]),
            (dictgen::InsensitiveStr::Ascii("comerce"), &["commerce"]),
            (
                dictgen::InsensitiveStr::Ascii("parametros"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("demaond"), &["demand"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplaication"),
                &["deduplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("containder"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("agregation"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("settig"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("lmit"), &["limit"]),
            (
                dictgen::InsensitiveStr::Ascii("recommednation"),
                &["recommendation"],
            ),
            (dictgen::InsensitiveStr::Ascii("develpos"), &["develops"]),
            (
                dictgen::InsensitiveStr::Ascii("checherboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("methaphor"), &["metaphor"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinetly"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tich"),
                &["thick", "tick", "titch", "stitch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconecct"),
                &["disconnect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dscriptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geogrpahically"),
                &["geographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsconciously"),
                &["subconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algolithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmeentation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mensutral"), &["menstrual"]),
            (dictgen::InsensitiveStr::Ascii("februray"), &["february"]),
            (dictgen::InsensitiveStr::Ascii("cubilce"), &["cubicle"]),
            (dictgen::InsensitiveStr::Ascii("skecthy"), &["sketchy"]),
            (dictgen::InsensitiveStr::Ascii("initialials"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("remannt"), &["remnant"]),
            (
                dictgen::InsensitiveStr::Ascii("vicitmized"),
                &["victimized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accoutants"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("includied"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("cacluate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("electriv"), &["electric"]),
            (
                dictgen::InsensitiveStr::Ascii("afaterwards"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rescritions"),
                &["restrictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instatiated"),
                &["instantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inventer"),
                &["inventor", "invented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dowgrading"),
                &["downgrading"],
            ),
            (dictgen::InsensitiveStr::Ascii("timtouts"), &["timeouts"]),
            (dictgen::InsensitiveStr::Ascii("signallin"), &["signalling"]),
            (dictgen::InsensitiveStr::Ascii("clame"), &["claim"]),
            (
                dictgen::InsensitiveStr::Ascii("algorrithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancl"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("illegas"), &["illegals"]),
            (
                dictgen::InsensitiveStr::Ascii("artificialy"),
                &["artificially"],
            ),
            (dictgen::InsensitiveStr::Ascii("convovle"), &["convolve"]),
            (
                dictgen::InsensitiveStr::Ascii("montors"),
                &["monitors", "motors", "mentors"],
            ),
            (dictgen::InsensitiveStr::Ascii("funture"), &["future"]),
            (
                dictgen::InsensitiveStr::Ascii("reperesented"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("serguns"), &["surgeons"]),
            (
                dictgen::InsensitiveStr::Ascii("refelcting"),
                &["reflecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("atually"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("epilespy"), &["epilepsy"]),
            (
                dictgen::InsensitiveStr::Ascii("underming"),
                &["undermining"],
            ),
            (dictgen::InsensitiveStr::Ascii("becoem"), &["become"]),
            (
                dictgen::InsensitiveStr::Ascii("imprements"),
                &["implements", "improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("eanabled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("menions"),
                &["mentions", "minions"],
            ),
            (dictgen::InsensitiveStr::Ascii("dbe"), &["be"]),
            (dictgen::InsensitiveStr::Ascii("extremisim"), &["extremism"]),
            (dictgen::InsensitiveStr::Ascii("bevore"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("inseperable"),
                &["inseparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("serinomys"), &["ceremonies"]),
            (dictgen::InsensitiveStr::Ascii("supsended"), &["suspended"]),
            (
                dictgen::InsensitiveStr::Ascii("earily"),
                &["easily", "eerily", "early"],
            ),
            (dictgen::InsensitiveStr::Ascii("indutrial"), &["industrial"]),
            (dictgen::InsensitiveStr::Ascii("confugres"), &["configures"]),
            (
                dictgen::InsensitiveStr::Ascii("mositurizer"),
                &["moisturizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegde"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("carrage"), &["carriage"]),
            (dictgen::InsensitiveStr::Ascii("brane"), &["brain"]),
            (
                dictgen::InsensitiveStr::Ascii("catergorizes"),
                &["categorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlyzer"), &["analyzer"]),
            (dictgen::InsensitiveStr::Ascii("poers"), &["powers"]),
            (dictgen::InsensitiveStr::Ascii("dissmiss"), &["dismiss"]),
            (dictgen::InsensitiveStr::Ascii("urfacing"), &["surfacing"]),
            (dictgen::InsensitiveStr::Ascii("barcleona"), &["barcelona"]),
            (
                dictgen::InsensitiveStr::Ascii("aleinating"),
                &["alienating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementas"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfying"),
                &["identifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mainifesting"),
                &["manifesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complianed"),
                &["compliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possissive"),
                &["possessive"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("resotres"), &["restores"]),
            (dictgen::InsensitiveStr::Ascii("cylynders"), &["cylinders"]),
            (dictgen::InsensitiveStr::Ascii("proceesed"), &["processed"]),
            (
                dictgen::InsensitiveStr::Ascii("persepective"),
                &["perspective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ampehtamines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("poduced"), &["produced"]),
            (dictgen::InsensitiveStr::Ascii("calim"), &["claim"]),
            (
                dictgen::InsensitiveStr::Ascii("consrvative"),
                &["conservative"],
            ),
            (dictgen::InsensitiveStr::Ascii("calucalte"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("retreiving"),
                &["retrieving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itentionaly"),
                &["intentionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("displayes"),
                &["displays", "displayed"],
            ),
            (dictgen::InsensitiveStr::Ascii("latitudine"), &["latitude"]),
            (dictgen::InsensitiveStr::Ascii("pharamcy"), &["pharmacy"]),
            (
                dictgen::InsensitiveStr::Ascii("conotroled"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architechtural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomances"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moderm"),
                &["modem", "modern"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictatorhip"),
                &["dictatorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("phising"), &["phishing"]),
            (
                dictgen::InsensitiveStr::Ascii("differntiable"),
                &["differentiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("instace"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("catagorised"),
                &["categorised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vailidators"),
                &["validators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obliverated"),
                &["obliterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("scedule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("transportar"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("playofs"), &["playoffs"]),
            (dictgen::InsensitiveStr::Ascii("roiginal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("baragin"), &["bargain"]),
            (dictgen::InsensitiveStr::Ascii("bundels"), &["bundles"]),
            (dictgen::InsensitiveStr::Ascii("wondow"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("bytetreams"),
                &["bytestreams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promisse"),
                &["promise", "promises", "promised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shoulde"),
                &["should", "shoulder"],
            ),
            (dictgen::InsensitiveStr::Ascii("querried"), &["queried"]),
            (
                dictgen::InsensitiveStr::Ascii("erroniously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("epects"),
                &["expects", "elects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tournement"),
                &["tournament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simillarity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("minsiters"), &["ministers"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvienence"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntructions"),
                &["constructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implicetly"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaiance"), &["variance"]),
            (dictgen::InsensitiveStr::Ascii("sopunding"), &["sounding"]),
            (
                dictgen::InsensitiveStr::Ascii("revoultionary"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalculatin"),
                &["recalculating", "recalculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gerenal"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("acitvates"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("correcteing"),
                &["correcting"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimmics"), &["mimics"]),
            (dictgen::InsensitiveStr::Ascii("cordinade"), &["coordinate"]),
            (
                dictgen::InsensitiveStr::Ascii("defintitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionner"),
                &["conditioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecodtal"), &["anecdotal"]),
            (
                dictgen::InsensitiveStr::Ascii("conntecting"),
                &["connecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloesed"), &["closed"]),
            (dictgen::InsensitiveStr::Ascii("excetution"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("formulatin"),
                &["formulating", "formulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("angirly"), &["angrily"]),
            (dictgen::InsensitiveStr::Ascii("memmick"), &["mimic"]),
            (
                dictgen::InsensitiveStr::Ascii("clairvoiants"),
                &["clairvoyants"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmaping"), &["unmapping"]),
            (dictgen::InsensitiveStr::Ascii("pandroa"), &["pandora"]),
            (dictgen::InsensitiveStr::Ascii("uncludes"), &["includes"]),
            (dictgen::InsensitiveStr::Ascii("canonival"), &["canonical"]),
            (
                dictgen::InsensitiveStr::Ascii("generlizing"),
                &["generalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguires"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("mehcanical"),
                &["mechanical"],
            ),
            (dictgen::InsensitiveStr::Ascii("thigsn"), &["things"]),
            (dictgen::InsensitiveStr::Ascii("functios"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("awefully"), &["awfully"]),
            (
                dictgen::InsensitiveStr::Ascii("transctions"),
                &["transitions", "transactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("regardes"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("nexessary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("mateiral"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("disributes"),
                &["distributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooridnates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("coments"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("anxiosly"), &["anxiously"]),
            (dictgen::InsensitiveStr::Ascii("worskapce"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("teories"),
                &["theories", "tories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disobeidence"),
                &["disobedience"],
            ),
            (dictgen::InsensitiveStr::Ascii("anormally"), &["abnormally"]),
            (dictgen::InsensitiveStr::Ascii("requsted"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("diamaters"), &["diameters"]),
            (
                dictgen::InsensitiveStr::Ascii("standartised"),
                &["standardised"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysens"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("differrently"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("milage"), &["mileage"]),
            (dictgen::InsensitiveStr::Ascii("ignorred"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("implent"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("sensasionalism"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("resaons"), &["reasons"]),
            (dictgen::InsensitiveStr::Ascii("sufferred"), &["suffered"]),
            (dictgen::InsensitiveStr::Ascii("radiu"), &["radius"]),
            (dictgen::InsensitiveStr::Ascii("susseedes"), &["secedes"]),
            (dictgen::InsensitiveStr::Ascii("postivie"), &["positive"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburges"),
                &["cheeseburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procurment"),
                &["procurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complasance"),
                &["complaisance", "compliance"],
            ),
            (dictgen::InsensitiveStr::Ascii("tralier"), &["trailer"]),
            (dictgen::InsensitiveStr::Ascii("consuder"), &["consider"]),
            (
                dictgen::InsensitiveStr::Ascii("temporairly"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("liasion"), &["liaison"]),
            (
                dictgen::InsensitiveStr::Ascii("algortithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminsicently"),
                &["reminiscently"],
            ),
            (dictgen::InsensitiveStr::Ascii("supres"), &["suppress"]),
            (
                dictgen::InsensitiveStr::Ascii("disapporved"),
                &["disapproved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directoryes"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("recordarse"), &["recorder"]),
            (dictgen::InsensitiveStr::Ascii("opeartion"), &["operation"]),
            (
                dictgen::InsensitiveStr::Ascii("traditionel"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("incluiding"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("pacement"), &["placement"]),
            (dictgen::InsensitiveStr::Ascii("hollywoood"), &["hollywood"]),
            (dictgen::InsensitiveStr::Ascii("evaulated"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvinently"),
                &["inconveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("applyable"), &["applicable"]),
            (dictgen::InsensitiveStr::Ascii("andd"), &["and"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuprising"),
                &["unsurprising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archiecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("buddisht"), &["buddhist"]),
            (
                dictgen::InsensitiveStr::Ascii("automonous"),
                &["autonomous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collabore"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("onfigure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("acchievable"),
                &["achievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("benefical"), &["beneficial"]),
            (
                dictgen::InsensitiveStr::Ascii("noticy"),
                &["notify", "notice"],
            ),
            (dictgen::InsensitiveStr::Ascii("natinally"), &["nationally"]),
            (
                dictgen::InsensitiveStr::Ascii("intrepreter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiased"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replaceble"),
                &["replaceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("everyoens"), &["everyones"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbooring"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("featring"), &["featuring"]),
            (dictgen::InsensitiveStr::Ascii("glicthed"), &["glitched"]),
            (dictgen::InsensitiveStr::Ascii("abstrac"), &["abstract"]),
            (dictgen::InsensitiveStr::Ascii("implosed"), &["implode"]),
            (
                dictgen::InsensitiveStr::Ascii("innitializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independnence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aphabetical"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("avances"), &["advances"]),
            (dictgen::InsensitiveStr::Ascii("cleened"), &["cleaned"]),
            (dictgen::InsensitiveStr::Ascii("strories"), &["stories"]),
            (dictgen::InsensitiveStr::Ascii("sucess"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("crresponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstinense"),
                &["abstinence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immeadiately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photogrpahs"),
                &["photographs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elctronics"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strawberies"),
                &["strawberries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miminizing"),
                &["minimizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recheability"),
                &["reachability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcississtic"),
                &["narcissistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("excating"), &["exacting"]),
            (dictgen::InsensitiveStr::Ascii("worldivew"), &["worldview"]),
            (dictgen::InsensitiveStr::Ascii("gyphs"), &["glyphs", "gyps"]),
            (
                dictgen::InsensitiveStr::Ascii("automatycally"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pahtfinder"),
                &["pathfinder"],
            ),
            (dictgen::InsensitiveStr::Ascii("realiste"), &["realise"]),
            (
                dictgen::InsensitiveStr::Ascii("conviniances"),
                &["conveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("includive"), &["inclusive"]),
            (dictgen::InsensitiveStr::Ascii("rememebr"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("goolge"), &["google"]),
            (
                dictgen::InsensitiveStr::Ascii("countrywides"),
                &["countryside"],
            ),
            (dictgen::InsensitiveStr::Ascii("realoded"), &["reloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("highes"),
                &["highest", "highs"],
            ),
            (dictgen::InsensitiveStr::Ascii("combiers"), &["combiners"]),
            (dictgen::InsensitiveStr::Ascii("demaind"), &["demand"]),
            (dictgen::InsensitiveStr::Ascii("flahses"), &["flashes"]),
            (dictgen::InsensitiveStr::Ascii("imagenary"), &["imaginary"]),
            (dictgen::InsensitiveStr::Ascii("listenting"), &["listening"]),
            (
                dictgen::InsensitiveStr::Ascii("superpowereds"),
                &["superpowers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizonatal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondend"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coditioned"),
                &["conditioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoses"), &["those"]),
            (dictgen::InsensitiveStr::Ascii("disporved"), &["disproved"]),
            (dictgen::InsensitiveStr::Ascii("furmulae"), &["formulae"]),
            (
                dictgen::InsensitiveStr::Ascii("threah"),
                &["thread", "threat"],
            ),
            (dictgen::InsensitiveStr::Ascii("dobling"), &["doubling"]),
            (dictgen::InsensitiveStr::Ascii("abdomine"), &["abdomen"]),
            (dictgen::InsensitiveStr::Ascii("exchausts"), &["exhausts"]),
            (
                dictgen::InsensitiveStr::Ascii("monolythic"),
                &["monolithic"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryfiers"), &["verifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("tranitions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("niave"), &["naive"]),
            (
                dictgen::InsensitiveStr::Ascii("radioaktive"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("arquiver"), &["archiver"]),
            (
                dictgen::InsensitiveStr::Ascii("dissambled"),
                &["disassembled"],
            ),
            (dictgen::InsensitiveStr::Ascii("unicornios"), &["unicorns"]),
            (
                dictgen::InsensitiveStr::Ascii("matheticians"),
                &["mathematicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composiiton"),
                &["compositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prophacies"),
                &["prophecies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dffering"),
                &["differing", "offering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertiees"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("spaecing"), &["spacing"]),
            (
                dictgen::InsensitiveStr::Ascii("widepsread"),
                &["widespread"],
            ),
            (dictgen::InsensitiveStr::Ascii("reasssign"), &["reassign"]),
            (
                dictgen::InsensitiveStr::Ascii("informatinal"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archieved"),
                &["achieved", "archived"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derprecates"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("opjects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("documenetation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrappning"), &["wrapping"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatibity"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refereneces"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("somehwere"), &["somewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("configulate"),
                &["configurate", "configure"],
            ),
            (dictgen::InsensitiveStr::Ascii("varify"), &["verify"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbare"),
                &["masturbate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("twiiter"),
                &["twitter", "twister"],
            ),
            (dictgen::InsensitiveStr::Ascii("illnesess"), &["illnesses"]),
            (
                dictgen::InsensitiveStr::Ascii("instralling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforrces"), &["enforces"]),
            (dictgen::InsensitiveStr::Ascii("pakaged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("oscurity"), &["obscurity"]),
            (dictgen::InsensitiveStr::Ascii("incuded"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimisation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeated"), &["competed"]),
            (dictgen::InsensitiveStr::Ascii("secifying"), &["specifying"]),
            (
                dictgen::InsensitiveStr::Ascii("thet"),
                &["that", "the", "they", "theft"],
            ),
            (dictgen::InsensitiveStr::Ascii("converson"), &["conversion"]),
            (dictgen::InsensitiveStr::Ascii("usseful"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("cherckerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanaged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("dicretionary"),
                &["discretionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibilites"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inializations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antidepressents"),
                &["antidepressants"],
            ),
            (dictgen::InsensitiveStr::Ascii("porjects"), &["projects"]),
            (dictgen::InsensitiveStr::Ascii("bcuckets"), &["buckets"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalistisk"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rudimentatry"),
                &["rudimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atferthought"),
                &["afterthought"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceped"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("sepearations"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downloades"),
                &["downloads", "downloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("neraly"), &["neurally"]),
            (dictgen::InsensitiveStr::Ascii("couroutine"), &["coroutine"]),
            (dictgen::InsensitiveStr::Ascii("staus"), &["status"]),
            (
                dictgen::InsensitiveStr::Ascii("juggarnaut"),
                &["juggernaut"],
            ),
            (dictgen::InsensitiveStr::Ascii("everbody"), &["everybody"]),
            (dictgen::InsensitiveStr::Ascii("exchnaged"), &["exchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("coersion"),
                &["coercion", "conversion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condigured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("conchus"), &["conscious"]),
            (dictgen::InsensitiveStr::Ascii("sidelinked"), &["sideline"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmeceutical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchanters"),
                &["merchants"],
            ),
            (dictgen::InsensitiveStr::Ascii("botifies"), &["notifies"]),
            (
                dictgen::InsensitiveStr::Ascii("condomnation"),
                &["condemnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperetion"),
                &["separation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certficiated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("numvber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("scecnario"), &["scenario"]),
            (dictgen::InsensitiveStr::Ascii("askes"), &["asks"]),
            (dictgen::InsensitiveStr::Ascii("entiteld"), &["entitled"]),
            (dictgen::InsensitiveStr::Ascii("riminicer"), &["reminiscer"]),
            (
                dictgen::InsensitiveStr::Ascii("cerimonies"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinining"),
                &["determining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangable"),
                &["rearrangeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successivo"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("outplayd"), &["outplayed"]),
            (
                dictgen::InsensitiveStr::Ascii("historietas"),
                &["histories"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulitply"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("avoding"), &["avoiding"]),
            (dictgen::InsensitiveStr::Ascii("coverd"), &["covered"]),
            (
                dictgen::InsensitiveStr::Ascii("reprentations"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hurricanefps"),
                &["hurricanes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contraining"),
                &["constraining", "containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("shanger"), &["changer"]),
            (
                dictgen::InsensitiveStr::Ascii("hyptothesis"),
                &["hypothesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("morpholgy"), &["morphology"]),
            (
                dictgen::InsensitiveStr::Ascii("gaurenteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("daughterbord"),
                &["daughterboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("leegth"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("restructued"),
                &["restructured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trianglulated"),
                &["triangulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaptaion"), &["adaptation"]),
            (
                dictgen::InsensitiveStr::Ascii("coincidential"),
                &["coincidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alighing"),
                &["aligning", "alighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementaciones"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("entites"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("oberflowing"),
                &["overflowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyposeses"), &["hypotheses"]),
            (
                dictgen::InsensitiveStr::Ascii("guaruntees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("languqges"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("archiecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interwined"),
                &["intertwined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responnsibilties"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consectutively"),
                &["consecutively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overhwelming"),
                &["overwhelming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveniene"),
                &["convenience", "conveniency", "convene"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circulaters"),
                &["circulators", "circulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("litterals"), &["literals"]),
            (dictgen::InsensitiveStr::Ascii("anxios"), &["anxious"]),
            (dictgen::InsensitiveStr::Ascii("torue"), &["torus", "tore"]),
            (
                dictgen::InsensitiveStr::Ascii("introvents"),
                &["introverts"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdder"), &["embedder"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstruciton"),
                &["reconstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("warrantly"), &["warranty"]),
            (
                dictgen::InsensitiveStr::Ascii("openscourced"),
                &["opensourced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwrold"),
                &["underworld"],
            ),
            (dictgen::InsensitiveStr::Ascii("intersecct"), &["intersect"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpretation"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("resarch"), &["research"]),
            (
                dictgen::InsensitiveStr::Ascii("trnaslated"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematicals"),
                &["mathematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("initilial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("systemes"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("permuation"),
                &["permutation"],
            ),
            (dictgen::InsensitiveStr::Ascii("idetifies"), &["identifies"]),
            (
                dictgen::InsensitiveStr::Ascii("gerenations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caterogized"),
                &["categorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intergrates"),
                &["integrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("editon"), &["edition"]),
            (dictgen::InsensitiveStr::Ascii("transitor"), &["transistor"]),
            (
                dictgen::InsensitiveStr::Ascii("disscussions"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siplification"),
                &["simplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditoinal"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresloved"),
                &["unresolved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoluitons"),
                &["revolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("offerred"), &["offered"]),
            (
                dictgen::InsensitiveStr::Ascii("fanservive"),
                &["fanservice"],
            ),
            (dictgen::InsensitiveStr::Ascii("asnwerers"), &["answerers"]),
            (dictgen::InsensitiveStr::Ascii("enocder"), &["encoder"]),
            (dictgen::InsensitiveStr::Ascii("meaured"), &["measured"]),
            (dictgen::InsensitiveStr::Ascii("garentor"), &["guarantor"]),
            (dictgen::InsensitiveStr::Ascii("peroidic"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("intercollegate"),
                &["intercollegiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executible"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfmromed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("mismatchd"), &["mismatched"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilitis"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("documets"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("fimilies"), &["families"]),
            (dictgen::InsensitiveStr::Ascii("disnables"), &["disables"]),
            (dictgen::InsensitiveStr::Ascii("keywors"), &["keywords"]),
            (
                dictgen::InsensitiveStr::Ascii("syntehsise"),
                &["synthesise"],
            ),
            (dictgen::InsensitiveStr::Ascii("concluso"), &["conclusion"]),
            (dictgen::InsensitiveStr::Ascii("anonymos"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("throgh"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("poost"), &["post"]),
            (
                dictgen::InsensitiveStr::Ascii("necromacer"),
                &["necromancer"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovverrides"), &["overrides"]),
            (
                dictgen::InsensitiveStr::Ascii("costumizer"),
                &["customizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equire"),
                &["require", "enquire", "equine", "esquire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("numericable"),
                &["numerical"],
            ),
            (dictgen::InsensitiveStr::Ascii("nto"), &["not"]),
            (
                dictgen::InsensitiveStr::Ascii("cancelllations"),
                &["cancellations"],
            ),
            (dictgen::InsensitiveStr::Ascii("falvours"), &["flavours"]),
            (
                dictgen::InsensitiveStr::Ascii("cartrdiges"),
                &["cartridges"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemtpts"), &["attempts"]),
            (dictgen::InsensitiveStr::Ascii("repetitve"), &["repetitive"]),
            (
                dictgen::InsensitiveStr::Ascii("storelines"),
                &["storylines"],
            ),
            (dictgen::InsensitiveStr::Ascii("selectrion"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("garant"),
                &["guarantor", "guardant"],
            ),
            (dictgen::InsensitiveStr::Ascii("instrinic"), &["intrinsic"]),
            (dictgen::InsensitiveStr::Ascii("operatons"), &["operations"]),
            (dictgen::InsensitiveStr::Ascii("celles"), &["cells"]),
            (dictgen::InsensitiveStr::Ascii("backwares"), &["backwards"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtorizes"),
                &["authorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cycolps"), &["cyclops"]),
            (
                dictgen::InsensitiveStr::Ascii("appliacble"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferal"), &["referral"]),
            (dictgen::InsensitiveStr::Ascii("zoommed"), &["zoomed"]),
            (dictgen::InsensitiveStr::Ascii("opimally"), &["optimally"]),
            (dictgen::InsensitiveStr::Ascii("crosspot"), &["crosspost"]),
            (
                dictgen::InsensitiveStr::Ascii("conditiong"),
                &["conditioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethihng"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("beglian"), &["belgian"]),
            (dictgen::InsensitiveStr::Ascii("enebling"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("instaciates"),
                &["instantiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("acrue"), &["accrue"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnforming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontaneos"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("esctasy"), &["ecstasy"]),
            (dictgen::InsensitiveStr::Ascii("gorillia"), &["gorilla"]),
            (
                dictgen::InsensitiveStr::Ascii("poenis"),
                &["penis", "peonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefiified"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ubiqituous"),
                &["ubiquitous"],
            ),
            (dictgen::InsensitiveStr::Ascii("procative"), &["proactive"]),
            (
                dictgen::InsensitiveStr::Ascii("foresseable"),
                &["foreseeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undergated"),
                &["underrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigal"),
                &["signal", "sigil"],
            ),
            (dictgen::InsensitiveStr::Ascii("delet"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("reciepts"), &["receipts"]),
            (dictgen::InsensitiveStr::Ascii("alumnium"), &["aluminium"]),
            (
                dictgen::InsensitiveStr::Ascii("singulary"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("sponsorees"), &["sponsors"]),
            (dictgen::InsensitiveStr::Ascii("acsends"), &["ascends"]),
            (
                dictgen::InsensitiveStr::Ascii("initiatie"),
                &["initiatives", "initiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrequancy"),
                &["infrequency"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifyers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("writtings"), &["writings"]),
            (dictgen::InsensitiveStr::Ascii("morroco"), &["morocco"]),
            (dictgen::InsensitiveStr::Ascii("chnange"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("drowts"), &["droughts"]),
            (dictgen::InsensitiveStr::Ascii("likelly"), &["likely"]),
            (
                dictgen::InsensitiveStr::Ascii("multplies"),
                &["multiples", "multiplies"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomater"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("miniscue"), &["miniscule"]),
            (
                dictgen::InsensitiveStr::Ascii("fahrenhiet"),
                &["fahrenheit"],
            ),
            (dictgen::InsensitiveStr::Ascii("draggin"), &["dragging"]),
            (
                dictgen::InsensitiveStr::Ascii("depreacated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("episdoe"), &["episode"]),
            (
                dictgen::InsensitiveStr::Ascii("preferecne"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addministering"),
                &["administering"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpeted"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("soovineers"), &["souvenirs"]),
            (
                dictgen::InsensitiveStr::Ascii("wellingotn"),
                &["wellington"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vizualised"),
                &["visualised"],
            ),
            (dictgen::InsensitiveStr::Ascii("researce"), &["researcher"]),
            (
                dictgen::InsensitiveStr::Ascii("indisputible"),
                &["indisputable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fatalaties"),
                &["fatalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramaterise"),
                &["parameterise"],
            ),
            (dictgen::InsensitiveStr::Ascii("delgators"), &["delegators"]),
            (
                dictgen::InsensitiveStr::Ascii("arrancement"),
                &["arrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assumotions"),
                &["assumptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("taxanomic"), &["taxonomic"]),
            (dictgen::InsensitiveStr::Ascii("remainging"), &["remaining"]),
            (dictgen::InsensitiveStr::Ascii("vastecomy"), &["vasectomy"]),
            (
                dictgen::InsensitiveStr::Ascii("eextracting"),
                &["extracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("preapre"), &["prepare"]),
            (
                dictgen::InsensitiveStr::Ascii("experementing"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressivo"),
                &["progression"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuaize"), &["visualize"]),
            (
                dictgen::InsensitiveStr::Ascii("authentified"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("intialled"), &["initialled"]),
            (
                dictgen::InsensitiveStr::Ascii("destinattion"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("optinoal"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("shileded"), &["shielded"]),
            (dictgen::InsensitiveStr::Ascii("suseeder"), &["seceder"]),
            (
                dictgen::InsensitiveStr::Ascii("homogenius"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("squeeky"), &["squeaky"]),
            (dictgen::InsensitiveStr::Ascii("supporteur"), &["supporter"]),
            (dictgen::InsensitiveStr::Ascii("langueages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("insiprational"),
                &["inspirational"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceptipn"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("techincally"),
                &["technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("officiallly"),
                &["officially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommeneded"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhanceds"), &["enhances"]),
            (dictgen::InsensitiveStr::Ascii("dominanace"), &["dominance"]),
            (dictgen::InsensitiveStr::Ascii("attribyte"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("deleteable"), &["deletable"]),
            (dictgen::InsensitiveStr::Ascii("askreddt"), &["askreddit"]),
            (
                dictgen::InsensitiveStr::Ascii("enchantmens"),
                &["enchantments"],
            ),
            (dictgen::InsensitiveStr::Ascii("outter"), &["outer"]),
            (dictgen::InsensitiveStr::Ascii("impolde"), &["implode"]),
            (dictgen::InsensitiveStr::Ascii("opbject"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("inheritcs"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("perpaid"), &["prepaid"]),
            (dictgen::InsensitiveStr::Ascii("guarnatey"), &["guaranty"]),
            (
                dictgen::InsensitiveStr::Ascii("prescedence"),
                &["precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("visiblly"), &["visibly"]),
            (dictgen::InsensitiveStr::Ascii("dedfined"), &["defined"]),
            (
                dictgen::InsensitiveStr::Ascii("entertianment"),
                &["entertainment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisically"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versatilty"),
                &["versatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aulty"),
                &["aunty", "faulty"],
            ),
            (dictgen::InsensitiveStr::Ascii("previsou"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("inifinitely"),
                &["infinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrrespondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("propertys"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("remotley"), &["remotely"]),
            (dictgen::InsensitiveStr::Ascii("declarin"), &["declaring"]),
            (dictgen::InsensitiveStr::Ascii("groosome"), &["gruesome"]),
            (dictgen::InsensitiveStr::Ascii("chizles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("derpivation"),
                &["deprivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("redme"), &["readme"]),
            (dictgen::InsensitiveStr::Ascii("asumtotic"), &["asymptotic"]),
            (dictgen::InsensitiveStr::Ascii("swithced"), &["switched"]),
            (dictgen::InsensitiveStr::Ascii("stomache"), &["stomach"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffictive"),
                &["ineffective"],
            ),
            (dictgen::InsensitiveStr::Ascii("catched"), &["caught"]),
            (
                dictgen::InsensitiveStr::Ascii("trafficing"),
                &["trafficking"],
            ),
            (dictgen::InsensitiveStr::Ascii("bookamrks"), &["bookmarks"]),
            (dictgen::InsensitiveStr::Ascii("resarted"), &["restarted"]),
            (
                dictgen::InsensitiveStr::Ascii("produccion"),
                &["production"],
            ),
            (dictgen::InsensitiveStr::Ascii("strore"), &["store"]),
            (dictgen::InsensitiveStr::Ascii("willfull"), &["willful"]),
            (dictgen::InsensitiveStr::Ascii("limt"), &["limit", "lint"]),
            (
                dictgen::InsensitiveStr::Ascii("analyitcally"),
                &["analytically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statistices"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("oponents"), &["opponents"]),
            (dictgen::InsensitiveStr::Ascii("abandonig"), &["abandoning"]),
            (dictgen::InsensitiveStr::Ascii("neral"), &["neural"]),
            (dictgen::InsensitiveStr::Ascii("changued"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("iranianos"), &["iranians"]),
            (
                dictgen::InsensitiveStr::Ascii("idendified"),
                &["identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pretencious"),
                &["pretentious"],
            ),
            (dictgen::InsensitiveStr::Ascii("appering"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("skepticles"), &["skeptics"]),
            (
                dictgen::InsensitiveStr::Ascii("etender"),
                &["extender", "attender"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superstitios"),
                &["superstitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("erronerous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("validationt"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("respectuos"), &["respects"]),
            (dictgen::InsensitiveStr::Ascii("conntacted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("cuddels"), &["cuddles"]),
            (dictgen::InsensitiveStr::Ascii("reipients"), &["recipients"]),
            (dictgen::InsensitiveStr::Ascii("annnounces"), &["announces"]),
            (
                dictgen::InsensitiveStr::Ascii("simptumatic"),
                &["symptomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("resulsets"), &["resultsets"]),
            (dictgen::InsensitiveStr::Ascii("whtihin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("horizontaly"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterless"),
                &["parameterless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchmkaring"),
                &["benchmarking"],
            ),
            (dictgen::InsensitiveStr::Ascii("disatrous"), &["disastrous"]),
            (dictgen::InsensitiveStr::Ascii("treis"), &["tries"]),
            (
                dictgen::InsensitiveStr::Ascii("girlfrined"),
                &["girlfriends"],
            ),
            (dictgen::InsensitiveStr::Ascii("emulatin"), &["emulation"]),
            (dictgen::InsensitiveStr::Ascii("aicraft"), &["aircraft"]),
            (dictgen::InsensitiveStr::Ascii("emailling"), &["emailing"]),
            (
                dictgen::InsensitiveStr::Ascii("executionor"),
                &["executioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlaized"),
                &["generalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("efffect"), &["effect"]),
            (dictgen::InsensitiveStr::Ascii("delgator"), &["delegator"]),
            (dictgen::InsensitiveStr::Ascii("optmizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("standrdizes"),
                &["standardizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissymmectric"),
                &["dissymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilinders"), &["cylinders"]),
            (dictgen::InsensitiveStr::Ascii("assicate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("vitirol"), &["vitriol"]),
            (
                dictgen::InsensitiveStr::Ascii("advertis"),
                &["adverts", "advertise"],
            ),
            (dictgen::InsensitiveStr::Ascii("grouips"), &["groups"]),
            (dictgen::InsensitiveStr::Ascii("evaluaton"), &["evaluation"]),
            (
                dictgen::InsensitiveStr::Ascii("imaginitve"),
                &["imaginative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calbirations"),
                &["calibrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipl"),
                &["multiple", "multiply"],
            ),
            (dictgen::InsensitiveStr::Ascii("diggested"), &["digested"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronised"),
                &["synchronised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraphrashed"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decrmenting"),
                &["decrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdoels"), &["models"]),
            (
                dictgen::InsensitiveStr::Ascii("understandbly"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaguaries"), &["vagaries"]),
            (dictgen::InsensitiveStr::Ascii("cteate"), &["create"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginzer"),
                &["recognizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docuentation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("condiut"), &["conduit"]),
            (
                dictgen::InsensitiveStr::Ascii("expermiental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("resembe"), &["resemble"]),
            (
                dictgen::InsensitiveStr::Ascii("indoctrinatin"),
                &["indoctrination"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyrmaid"), &["pyramid"]),
            (
                dictgen::InsensitiveStr::Ascii("dynamiclly"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sociopatch"),
                &["sociopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("trinagle"), &["triangle"]),
            (dictgen::InsensitiveStr::Ascii("ipmroves"), &["improves"]),
            (
                dictgen::InsensitiveStr::Ascii("proprietory"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounsary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("trogladites"),
                &["troglodytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("desroyer"), &["destroyer"]),
            (
                dictgen::InsensitiveStr::Ascii("manfiesting"),
                &["manifesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootraper"),
                &["bootstrapper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nightmears"),
                &["nightmares"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discussons"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniuplators"),
                &["manipulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequense"),
                &["consequences"],
            ),
            (dictgen::InsensitiveStr::Ascii("positionof"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("projectiel"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represnetated"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("negoitate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("bannets"), &["bayonets"]),
            (
                dictgen::InsensitiveStr::Ascii("imbalancers"),
                &["imbalances"],
            ),
            (dictgen::InsensitiveStr::Ascii("oringinal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("excempts"), &["exempts"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainence"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("utlised"), &["utilised"]),
            (dictgen::InsensitiveStr::Ascii("graduatin"), &["graduation"]),
            (dictgen::InsensitiveStr::Ascii("sorcerry"), &["sorcery"]),
            (dictgen::InsensitiveStr::Ascii("freesytle"), &["freestyle"]),
            (
                dictgen::InsensitiveStr::Ascii("reducters"),
                &["reducers", "redacters"],
            ),
            (dictgen::InsensitiveStr::Ascii("languaes"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("pipelies"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("sensetivity"),
                &["sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("escpecially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("fantasiose"), &["fantasies"]),
            (
                dictgen::InsensitiveStr::Ascii("dominationg"),
                &["dominating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transisions"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difraction"),
                &["diffraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complicatie"),
                &["complicit"],
            ),
            (dictgen::InsensitiveStr::Ascii("sriarcha"), &["sriracha"]),
            (dictgen::InsensitiveStr::Ascii("differen"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("buoancy"), &["buoyancy"]),
            (dictgen::InsensitiveStr::Ascii("archor"), &["anchor"]),
            (dictgen::InsensitiveStr::Ascii("hearbeat"), &["heartbeat"]),
            (dictgen::InsensitiveStr::Ascii("chractors"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segragated"),
                &["segregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bankrupty"), &["bankruptcy"]),
            (dictgen::InsensitiveStr::Ascii("stoped"), &["stopped"]),
            (
                dictgen::InsensitiveStr::Ascii("uninfromes"),
                &["uninforms", "uniforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("activing"), &["activating"]),
            (
                dictgen::InsensitiveStr::Ascii("preferrable"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispayport"),
                &["displayport"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviguring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resturcture"),
                &["restructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("workpace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("meanins"), &["meanings"]),
            (
                dictgen::InsensitiveStr::Ascii("interploates"),
                &["interpolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrleations"),
                &["correlations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiallizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneted"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("deterimined"),
                &["determined"],
            ),
            (dictgen::InsensitiveStr::Ascii("remian"), &["remain"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburguer"),
                &["cheeseburger"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreation"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("filesnames"), &["filenames"]),
            (
                dictgen::InsensitiveStr::Ascii("solfer"),
                &["solver", "solder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortuntaly"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comporessing"),
                &["compressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("westbrok"), &["westbrook"]),
            (dictgen::InsensitiveStr::Ascii("conacting"), &["contacting"]),
            (dictgen::InsensitiveStr::Ascii("pathign"), &["pathing"]),
            (
                dictgen::InsensitiveStr::Ascii("exicted"),
                &["excited", "existed", "exited", "evicted", "exacted"],
            ),
            (dictgen::InsensitiveStr::Ascii("unwrittern"), &["unwritten"]),
            (
                dictgen::InsensitiveStr::Ascii("apppropriately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("helds"), &["holds", "held"]),
            (dictgen::InsensitiveStr::Ascii("libarary"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("untrasform"),
                &["untransform"],
            ),
            (dictgen::InsensitiveStr::Ascii("singlar"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("hanlde"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("compasso"), &["compassion"]),
            (dictgen::InsensitiveStr::Ascii("attivated"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("minnors"),
                &["minors", "minnows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otating"),
                &["notating", "rotating"],
            ),
            (dictgen::InsensitiveStr::Ascii("initmate"), &["intimate"]),
            (
                dictgen::InsensitiveStr::Ascii("destrcutors"),
                &["destructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comunication"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breacks"),
                &["breaks", "brakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("expensie"), &["expense"]),
            (dictgen::InsensitiveStr::Ascii("wathcing"), &["watching"]),
            (
                dictgen::InsensitiveStr::Ascii("ative"),
                &["active", "native"],
            ),
            (dictgen::InsensitiveStr::Ascii("deternine"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("awakenend"), &["awakened"]),
            (dictgen::InsensitiveStr::Ascii("wraappers"), &["wrappers"]),
            (dictgen::InsensitiveStr::Ascii("reverible"), &["reversible"]),
            (dictgen::InsensitiveStr::Ascii("defenderes"), &["defenders"]),
            (dictgen::InsensitiveStr::Ascii("arizonia"), &["arizona"]),
            (
                dictgen::InsensitiveStr::Ascii("confirgures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narling"),
                &["gnarling", "nailing", "narking", "snarling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characstyle"),
                &["charstyle"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistros"), &["maestros"]),
            (dictgen::InsensitiveStr::Ascii("guaging"), &["gaging"]),
            (dictgen::InsensitiveStr::Ascii("wolrdwide"), &["worldwide"]),
            (dictgen::InsensitiveStr::Ascii("layou"), &["layout"]),
            (
                dictgen::InsensitiveStr::Ascii("protogonist"),
                &["protagonist"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapper"), &["disappear"]),
            (dictgen::InsensitiveStr::Ascii("hybernate"), &["hibernate"]),
            (
                dictgen::InsensitiveStr::Ascii("silouhetting"),
                &["silhouetting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cornel"), &["colonel"]),
            (dictgen::InsensitiveStr::Ascii("neckbead"), &["neckbeard"]),
            (dictgen::InsensitiveStr::Ascii("evolveos"), &["evolves"]),
            (
                dictgen::InsensitiveStr::Ascii("immaturety"),
                &["immaturity"],
            ),
            (dictgen::InsensitiveStr::Ascii("galations"), &["galatians"]),
            (dictgen::InsensitiveStr::Ascii("chromum"), &["chromium"]),
            (dictgen::InsensitiveStr::Ascii("recconecct"), &["reconnect"]),
            (dictgen::InsensitiveStr::Ascii("casues"), &["causes"]),
            (dictgen::InsensitiveStr::Ascii("quation"), &["equation"]),
            (dictgen::InsensitiveStr::Ascii("robocoop"), &["robocop"]),
            (dictgen::InsensitiveStr::Ascii("driection"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("popoulus"), &["populous"]),
            (
                dictgen::InsensitiveStr::Ascii("unknowngly"),
                &["unknowingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("preqeuls"), &["prequels"]),
            (dictgen::InsensitiveStr::Ascii("prodivded"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("represnted"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registeration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("recoveres"), &["recovers"]),
            (
                dictgen::InsensitiveStr::Ascii("conjonctive"),
                &["conjunctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossibile"),
                &["impossible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialiazed"),
                &["serialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaxts"),
                &["contexts", "contacts"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahmonds"), &["almonds"]),
            (dictgen::InsensitiveStr::Ascii("defived"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("intentended"), &["intended"]),
            (
                dictgen::InsensitiveStr::Ascii("constructurs"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("guestures"), &["gestures"]),
            (
                dictgen::InsensitiveStr::Ascii("complience"),
                &["compliance"],
            ),
            (dictgen::InsensitiveStr::Ascii("encodng"), &["encoding"]),
            (
                dictgen::InsensitiveStr::Ascii("appricates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriptype"), &["scripttype"]),
            (dictgen::InsensitiveStr::Ascii("presernt"), &["present"]),
            (dictgen::InsensitiveStr::Ascii("annonced"), &["announced"]),
            (dictgen::InsensitiveStr::Ascii("listner"), &["listener"]),
            (
                dictgen::InsensitiveStr::Ascii("bluebrints"),
                &["blueprints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidunce"),
                &["presidency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropiately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("disctinct"), &["distinct"]),
            (dictgen::InsensitiveStr::Ascii("docrines"), &["doctrines"]),
            (dictgen::InsensitiveStr::Ascii("derfiend"), &["defined"]),
            (
                dictgen::InsensitiveStr::Ascii("parallizes"),
                &["parallelizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("puplations"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publsihing"),
                &["publishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradiciton"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferncencers"),
                &["dereferencers"],
            ),
            (dictgen::InsensitiveStr::Ascii("discrards"), &["discards"]),
            (
                dictgen::InsensitiveStr::Ascii("solication"),
                &["solicitation", "solidification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecions"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("estuwarry"), &["estuary"]),
            (
                dictgen::InsensitiveStr::Ascii("conservatie"),
                &["conservatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diminsions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delearship"),
                &["dealership"],
            ),
            (dictgen::InsensitiveStr::Ascii("buoyancu"), &["buoyancy"]),
            (dictgen::InsensitiveStr::Ascii("borader"), &["broader"]),
            (
                dictgen::InsensitiveStr::Ascii("confituration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("ourselvs"), &["ourselves"]),
            (dictgen::InsensitiveStr::Ascii("elminated"), &["eliminated"]),
            (
                dictgen::InsensitiveStr::Ascii("defensivly"),
                &["defensively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mediterainnean"),
                &["mediterranean"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propageted"),
                &["propagated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanle"), &["handle"]),
            (
                dictgen::InsensitiveStr::Ascii("incorprated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("inmigrant"), &["immigrant"]),
            (dictgen::InsensitiveStr::Ascii("contendor"), &["contender"]),
            (
                dictgen::InsensitiveStr::Ascii("perfectably"),
                &["perfectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("becxause"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("soverignty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("srriacha"), &["sriracha"]),
            (dictgen::InsensitiveStr::Ascii("signigns"), &["signings"]),
            (
                dictgen::InsensitiveStr::Ascii("transhpobic"),
                &["transphobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jounralism"),
                &["journalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("resettin"), &["resetting"]),
            (dictgen::InsensitiveStr::Ascii("containin"), &["containing"]),
            (
                dictgen::InsensitiveStr::Ascii("industrialied"),
                &["industrialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contigents"),
                &["continents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patitioning"),
                &["partitioning", "petitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasparently"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("humands"), &["humans"]),
            (dictgen::InsensitiveStr::Ascii("greenore"), &["greener"]),
            (dictgen::InsensitiveStr::Ascii("accpeting"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("compareison"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sructuring"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("licsensing"), &["licensing"]),
            (
                dictgen::InsensitiveStr::Ascii("deklaration"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstnce"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("wispered"), &["whispered"]),
            (dictgen::InsensitiveStr::Ascii("puchased"), &["purchased"]),
            (dictgen::InsensitiveStr::Ascii("scrol"), &["scroll"]),
            (dictgen::InsensitiveStr::Ascii("campaing"), &["campaign"]),
            (dictgen::InsensitiveStr::Ascii("artcile"), &["article"]),
            (
                dictgen::InsensitiveStr::Ascii("recognision"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggeratted"),
                &["exaggerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("installus"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("arquival"), &["archival"]),
            (
                dictgen::InsensitiveStr::Ascii("participaton"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technicans"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predection"),
                &["prediction"],
            ),
            (dictgen::InsensitiveStr::Ascii("liquour"), &["liquor"]),
            (
                dictgen::InsensitiveStr::Ascii("comparativly"),
                &["comparatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("brighen"), &["brighten"]),
            (
                dictgen::InsensitiveStr::Ascii("interprating"),
                &["interpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("keynseian"), &["keynesian"]),
            (dictgen::InsensitiveStr::Ascii("gnerated"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("alhpabets"), &["alphabets"]),
            (dictgen::InsensitiveStr::Ascii("refrerence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("collogue"), &["colleague"]),
            (
                dictgen::InsensitiveStr::Ascii("jursidictions"),
                &["jurisdictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("evluating"), &["evaluating"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusiveity"),
                &["exclusivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("gaurding"), &["guarding"]),
            (dictgen::InsensitiveStr::Ascii("evalauted"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("sylesheet"), &["stylesheet"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupicate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("symphaty"), &["sympathy"]),
            (dictgen::InsensitiveStr::Ascii("mangesium"), &["magnesium"]),
            (
                dictgen::InsensitiveStr::Ascii("photographes"),
                &["photographs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virutalize"),
                &["virtualize"],
            ),
            (dictgen::InsensitiveStr::Ascii("colllapses"), &["collapses"]),
            (
                dictgen::InsensitiveStr::Ascii("recipercal"),
                &["reciprocal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("storytellling"),
                &["storytelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("convaless"), &["convalesce"]),
            (dictgen::InsensitiveStr::Ascii("rigth"), &["right"]),
            (dictgen::InsensitiveStr::Ascii("ba"), &["by", "be"]),
            (dictgen::InsensitiveStr::Ascii("buferred"), &["buffered"]),
            (dictgen::InsensitiveStr::Ascii("intall"), &["install"]),
            (dictgen::InsensitiveStr::Ascii("insualted"), &["insulated"]),
            (
                dictgen::InsensitiveStr::Ascii("preaprations"),
                &["preparations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momentairly"),
                &["momentarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiull"), &["will"]),
            (dictgen::InsensitiveStr::Ascii("undefinied"), &["undefined"]),
            (dictgen::InsensitiveStr::Ascii("argumemnts"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmisations"),
                &["optimisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanets"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("identfied"), &["identified"]),
            (dictgen::InsensitiveStr::Ascii("makin"), &["making", "main"]),
            (dictgen::InsensitiveStr::Ascii("virualenv"), &["virtualenv"]),
            (
                dictgen::InsensitiveStr::Ascii("represantational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("ubunut"), &["ubuntu"]),
            (dictgen::InsensitiveStr::Ascii("reqquests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("aymptotic"), &["asymptotic"]),
            (
                dictgen::InsensitiveStr::Ascii("compareble"),
                &["comparable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expecetations"),
                &["expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schizophreniiic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("persue"), &["pursue"]),
            (
                dictgen::InsensitiveStr::Ascii("torunaments"),
                &["tournaments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cotributions"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulataeous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortiy"), &["fortify"]),
            (dictgen::InsensitiveStr::Ascii("properrt"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("enforer"), &["enforcer"]),
            (
                dictgen::InsensitiveStr::Ascii("bottlencks"),
                &["bottlenecks"],
            ),
            (dictgen::InsensitiveStr::Ascii("coffe"), &["coffee"]),
            (
                dictgen::InsensitiveStr::Ascii("hipothetical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("applyes"), &["applies"]),
            (dictgen::InsensitiveStr::Ascii("highjacker"), &["hijacker"]),
            (
                dictgen::InsensitiveStr::Ascii("chanllenging"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invaidates"),
                &["invalidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directiories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depracating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("callibri"), &["calibri"]),
            (
                dictgen::InsensitiveStr::Ascii("prirorities"),
                &["priorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiems"), &["times", "items"]),
            (dictgen::InsensitiveStr::Ascii("desided"), &["decided"]),
            (dictgen::InsensitiveStr::Ascii("aheres"), &["adheres"]),
            (dictgen::InsensitiveStr::Ascii("somtime"), &["sometime"]),
            (
                dictgen::InsensitiveStr::Ascii("complettly"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conquerers"),
                &["conquerors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirucumstance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("owened"), &["owned"]),
            (dictgen::InsensitiveStr::Ascii("brigthnes"), &["brightness"]),
            (dictgen::InsensitiveStr::Ascii("inheritate"), &["inherit"]),
            (
                dictgen::InsensitiveStr::Ascii("discriptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momomento"),
                &["memento", "moment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controllors"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limted"),
                &["limited", "linted"],
            ),
            (dictgen::InsensitiveStr::Ascii("ansambles"), &["ensembles"]),
            (dictgen::InsensitiveStr::Ascii("dscribe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("cecked"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("interfacae"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("dublicated"),
                &["duplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electrinics"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alocholism"),
                &["alcoholism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mosue"),
                &["mosque", "mouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remarkablely"),
                &["remarkably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetrificates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiblilty"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destinaiton"),
                &["destinations", "destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chancges"),
                &["changes", "chances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indits"),
                &["indicts", "indites", "indies"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefifes"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("hygene"), &["hygiene"]),
            (dictgen::InsensitiveStr::Ascii("challengs"), &["challenges"]),
            (
                dictgen::InsensitiveStr::Ascii("retunr"),
                &["return", "retune"],
            ),
            (dictgen::InsensitiveStr::Ascii("helpfuly"), &["helpfully"]),
            (
                dictgen::InsensitiveStr::Ascii("expericing"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("teorem"), &["theorem"]),
            (dictgen::InsensitiveStr::Ascii("complient"), &["compliant"]),
            (dictgen::InsensitiveStr::Ascii("channnels"), &["channels"]),
            (dictgen::InsensitiveStr::Ascii("keywork"), &["keyword"]),
            (
                dictgen::InsensitiveStr::Ascii("accelereating"),
                &["accelerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("provders"), &["providers"]),
            (
                dictgen::InsensitiveStr::Ascii("instuction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ventilatin"),
                &["ventilating", "ventilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techincian"),
                &["technician"],
            ),
            (dictgen::InsensitiveStr::Ascii("antyhing"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("kiyackers"), &["kayakers"]),
            (
                dictgen::InsensitiveStr::Ascii("querstionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whatepsaces"),
                &["whitespaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acompanying"),
                &["accompanying"],
            ),
            (dictgen::InsensitiveStr::Ascii("signaure"), &["signature"]),
            (dictgen::InsensitiveStr::Ascii("desplay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("excellant"), &["excellent"]),
            (dictgen::InsensitiveStr::Ascii("corridoor"), &["corridor"]),
            (
                dictgen::InsensitiveStr::Ascii("undreground"),
                &["underground"],
            ),
            (dictgen::InsensitiveStr::Ascii("conlcude"), &["conclude"]),
            (
                dictgen::InsensitiveStr::Ascii("eucledian"),
                &["euclidean", "euclidian"],
            ),
            (dictgen::InsensitiveStr::Ascii("gived"), &["given", "gave"]),
            (dictgen::InsensitiveStr::Ascii("reaed"), &["read"]),
            (
                dictgen::InsensitiveStr::Ascii("unhighlighed"),
                &["unhighlighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("childbrith"),
                &["childbirth"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheparded"), &["shepherded"]),
            (dictgen::InsensitiveStr::Ascii("overlayed"), &["overlaid"]),
            (dictgen::InsensitiveStr::Ascii("hoooks"), &["hooks"]),
            (dictgen::InsensitiveStr::Ascii("pubilshed"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("aggegrate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("desne"), &["dense"]),
            (
                dictgen::InsensitiveStr::Ascii("rendazyvoo"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("represen"), &["represent"]),
            (
                dictgen::InsensitiveStr::Ascii("inconveinent"),
                &["inconvenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("piggypacked"),
                &["piggybacked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accompined"),
                &["accompanied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensibilites"),
                &["sensibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("persuded"), &["persuaded"]),
            (dictgen::InsensitiveStr::Ascii("detaults"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("deficiensies"),
                &["deficiencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("babysittng"),
                &["babysitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheems"), &["schemes"]),
            (
                dictgen::InsensitiveStr::Ascii("initilising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("draged"), &["dragged"]),
            (
                dictgen::InsensitiveStr::Ascii("mitigatin"),
                &["mitigating", "mitigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("glimpes"), &["glimpse"]),
            (dictgen::InsensitiveStr::Ascii("metagem"), &["metagame"]),
            (dictgen::InsensitiveStr::Ascii("poistive"), &["positive"]),
            (
                dictgen::InsensitiveStr::Ascii("overrided"),
                &["overrode", "overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("grestest"), &["greatest"]),
            (
                dictgen::InsensitiveStr::Ascii("visibillity"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("lunguages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("tranmitted"),
                &["transmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expropiating"),
                &["expropriating"],
            ),
            (dictgen::InsensitiveStr::Ascii("convinve"), &["convince"]),
            (
                dictgen::InsensitiveStr::Ascii("impore"),
                &["implore", "impure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimititers"),
                &["delimiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("aninating"), &["animating"]),
            (
                dictgen::InsensitiveStr::Ascii("succefully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contollers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceucte"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("copmletion"),
                &["completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("freeezer"), &["freezer"]),
            (dictgen::InsensitiveStr::Ascii("assumue"), &["assume"]),
            (
                dictgen::InsensitiveStr::Ascii("inconditional"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("straem"), &["stream"]),
            (
                dictgen::InsensitiveStr::Ascii("completition"),
                &["completion", "competition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilates"),
                &["compiles", "complicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("algortim"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("tentant"), &["tenant"]),
            (dictgen::InsensitiveStr::Ascii("cummand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("culiminate"), &["culminate"]),
            (
                dictgen::InsensitiveStr::Ascii("legendarisk"),
                &["legendaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifacation"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("fesitve"), &["festive"]),
            (dictgen::InsensitiveStr::Ascii("quantitity"), &["quantity"]),
            (dictgen::InsensitiveStr::Ascii("afterw"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("assumbed"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("patholigical"),
                &["pathological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomenom"),
                &["phenomenon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fermentaiton"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("neckbeared"), &["neckbeard"]),
            (dictgen::InsensitiveStr::Ascii("pengwens"), &["penguins"]),
            (dictgen::InsensitiveStr::Ascii("nutrally"), &["neutrally"]),
            (
                dictgen::InsensitiveStr::Ascii("unseful"),
                &["useless", "useful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranzitions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminls"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("determinated"),
                &["determined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("machanisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facitilies"),
                &["facilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigare"),
                &["investigate"],
            ),
            (dictgen::InsensitiveStr::Ascii("monkies"), &["monkeys"]),
            (dictgen::InsensitiveStr::Ascii("blohted"), &["bloated"]),
            (dictgen::InsensitiveStr::Ascii("socring"), &["scoring"]),
            (dictgen::InsensitiveStr::Ascii("receieving"), &["receiving"]),
            (dictgen::InsensitiveStr::Ascii("shortct"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("preseved"), &["preserved"]),
            (dictgen::InsensitiveStr::Ascii("bruiseres"), &["bruisers"]),
            (
                dictgen::InsensitiveStr::Ascii("repostions"),
                &["repositions"],
            ),
            (dictgen::InsensitiveStr::Ascii("workingest"), &["workings"]),
            (
                dictgen::InsensitiveStr::Ascii("teather"),
                &[
                    "tether", "teacher", "theater", "feather", "weather", "leather",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associaitions"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lieutenent"),
                &["lieutenant"],
            ),
            (dictgen::InsensitiveStr::Ascii("acordions"), &["accordions"]),
            (dictgen::InsensitiveStr::Ascii("extremisme"), &["extremes"]),
            (dictgen::InsensitiveStr::Ascii("reschdule"), &["reschedule"]),
            (dictgen::InsensitiveStr::Ascii("simlar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("rabinnical"),
                &["rabbinical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segementation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfromer"),
                &["transformers", "transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("collumns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("derriving"), &["deriving"]),
            (dictgen::InsensitiveStr::Ascii("ptichers"), &["pitchers"]),
            (dictgen::InsensitiveStr::Ascii("pojectors"), &["projectors"]),
            (
                dictgen::InsensitiveStr::Ascii("caluclation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("consideras"), &["considers"]),
            (
                dictgen::InsensitiveStr::Ascii("etablishing"),
                &["establishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arpanoid"), &["paranoid"]),
            (dictgen::InsensitiveStr::Ascii("aplicable"), &["applicable"]),
            (
                dictgen::InsensitiveStr::Ascii("connecstatus"),
                &["connectstatus"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprimizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("mercenarios"),
                &["mercenaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adequetely"),
                &["adequately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconeections"),
                &["disconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("featcher"),
                &["feather", "feature", "fetcher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappared"),
                &["disappeared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senisivity"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppert"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("accomdated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("passt"), &["passed", "past"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibitons"),
                &["prohibits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpritations"),
                &["interpretations"],
            ),
            (dictgen::InsensitiveStr::Ascii("broswed"), &["browsed"]),
            (dictgen::InsensitiveStr::Ascii("asscii"), &["ascii"]),
            (
                dictgen::InsensitiveStr::Ascii("continentais"),
                &["continents"],
            ),
            (dictgen::InsensitiveStr::Ascii("resistent"), &["resistant"]),
            (dictgen::InsensitiveStr::Ascii("edditor"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("uncommmon"), &["uncommon"]),
            (
                dictgen::InsensitiveStr::Ascii("neightobrhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginition"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicatoins"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordiing"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("unmarsalling"),
                &["unmarshalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("manigesto"), &["manifesto"]),
            (dictgen::InsensitiveStr::Ascii("assumme"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("beardude"), &["bearded"]),
            (
                dictgen::InsensitiveStr::Ascii("blackbeery"),
                &["blackberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("aersopace"), &["aerospace"]),
            (
                dictgen::InsensitiveStr::Ascii("desctiptively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticdotes"),
                &["anecdotes", "antidotes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possiibilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deallcoate"),
                &["deallocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("expanshion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("chocloate"), &["chocolate"]),
            (dictgen::InsensitiveStr::Ascii("nofify"), &["notify"]),
            (dictgen::InsensitiveStr::Ascii("insteaed"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("complys"), &["complies"]),
            (dictgen::InsensitiveStr::Ascii("trought"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("affilate"), &["affiliate"]),
            (dictgen::InsensitiveStr::Ascii("coprorate"), &["corporate"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoticated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhist"), &["exist"]),
            (
                dictgen::InsensitiveStr::Ascii("ankshiuosly"),
                &["anxiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electoratul"),
                &["electoral"],
            ),
            (dictgen::InsensitiveStr::Ascii("pactching"), &["patching"]),
            (dictgen::InsensitiveStr::Ascii("convig"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("calcaulting"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("alisasing"), &["aliasing"]),
            (
                dictgen::InsensitiveStr::Ascii("ambiguuously"),
                &["ambiguously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operationgs"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exsist"),
                &["exists", "exist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypotetically"),
                &["hypothetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("responsess"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("termianting"),
                &["terminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prgression"),
                &["progression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoaiation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copyirghted"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleann"), &["clean"]),
            (dictgen::InsensitiveStr::Ascii("insidiuos"), &["insidious"]),
            (
                dictgen::InsensitiveStr::Ascii("drasticlly"),
                &["drastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupplications"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independante"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("shawhsank"), &["shawshank"]),
            (dictgen::InsensitiveStr::Ascii("allos"), &["allows"]),
            (dictgen::InsensitiveStr::Ascii("functins"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("inablility"), &["inability"]),
            (dictgen::InsensitiveStr::Ascii("consunmed"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("overcompensatin"),
                &["overcompensating", "overcompensation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reigment"), &["regiment"]),
            (dictgen::InsensitiveStr::Ascii("depdenncy"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("subscripcitons"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anecdotelly"),
                &["anecdotally"],
            ),
            (dictgen::InsensitiveStr::Ascii("articaft"), &["artifact"]),
            (dictgen::InsensitiveStr::Ascii("rectanges"), &["rectangles"]),
            (
                dictgen::InsensitiveStr::Ascii("discrtization"),
                &["discretization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fascinacion"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("assummed"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("archetectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("spectrail"), &["spectral"]),
            (
                dictgen::InsensitiveStr::Ascii("wries"),
                &["writes", "wires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communcated"),
                &["communicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("unabnned"), &["unbanned"]),
            (dictgen::InsensitiveStr::Ascii("upsream"), &["upstream"]),
            (
                dictgen::InsensitiveStr::Ascii("apostrafies"),
                &["apostrophes"],
            ),
            (dictgen::InsensitiveStr::Ascii("desicion"), &["decision"]),
            (
                dictgen::InsensitiveStr::Ascii("connectoin"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispecable"),
                &["despicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionsl"),
                &["professionalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("aethistic"), &["atheistic"]),
            (dictgen::InsensitiveStr::Ascii("gigbayte"), &["gigabyte"]),
            (
                dictgen::InsensitiveStr::Ascii("arrity"),
                &["arity", "parity", "rarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("contuine"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("prosocution"),
                &["prosecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilibriam"),
                &["equilibrium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetner"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooperacion"),
                &["cooperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("iritated"), &["irritated"]),
            (
                dictgen::InsensitiveStr::Ascii("incremenetally"),
                &["incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posessions"),
                &["possessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkpint"), &["checkpoint"]),
            (dictgen::InsensitiveStr::Ascii("suggeste"), &["suggestive"]),
            (
                dictgen::InsensitiveStr::Ascii("shrinked"),
                &["shrunk", "shrank"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incapacitatin"),
                &["incapacitating", "incapacitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("predfined"), &["predefined"]),
            (dictgen::InsensitiveStr::Ascii("grativate"), &["gravitate"]),
            (
                dictgen::InsensitiveStr::Ascii("congresional"),
                &["congressional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priveliged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("entropay"), &["entropy"]),
            (
                dictgen::InsensitiveStr::Ascii("phenomonenon"),
                &["phenomenon"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasmit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptiively"),
                &["descriptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("booleam"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("soddered"), &["soldered"]),
            (dictgen::InsensitiveStr::Ascii("peirced"), &["pierced"]),
            (dictgen::InsensitiveStr::Ascii("belevably"), &["believably"]),
            (dictgen::InsensitiveStr::Ascii("chromose"), &["chromosome"]),
            (
                dictgen::InsensitiveStr::Ascii("intruments"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extradiction"),
                &["extradition"],
            ),
            (dictgen::InsensitiveStr::Ascii("similary"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuded"),
                &["unused", "unsuited"],
            ),
            (dictgen::InsensitiveStr::Ascii("fluttersy"), &["fluttershy"]),
            (dictgen::InsensitiveStr::Ascii("mutilples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("allcators"), &["allocators"]),
            (dictgen::InsensitiveStr::Ascii("nighbor"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("aplied"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("dyanamics"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("descirptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("conensus"), &["consensus"]),
            (dictgen::InsensitiveStr::Ascii("hhttp"), &["http"]),
            (dictgen::InsensitiveStr::Ascii("targesting"), &["targeting"]),
            (
                dictgen::InsensitiveStr::Ascii("roboustness"),
                &["robustness"],
            ),
            (dictgen::InsensitiveStr::Ascii("puplarity"), &["popularity"]),
            (dictgen::InsensitiveStr::Ascii("defautt"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("fronat"),
                &["format", "front"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitian"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("propmted"), &["prompted"]),
            (dictgen::InsensitiveStr::Ascii("commnet"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("intrinic"), &["intrinsic"]),
            (dictgen::InsensitiveStr::Ascii("ammending"), &["amending"]),
            (
                dictgen::InsensitiveStr::Ascii("refinemenet"),
                &["refinement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconcilled"),
                &["reconcile"],
            ),
            (dictgen::InsensitiveStr::Ascii("poposals"), &["proposals"]),
            (
                dictgen::InsensitiveStr::Ascii("encounted"),
                &["encountered", "encounter"],
            ),
            (dictgen::InsensitiveStr::Ascii("titlin"), &["titling"]),
            (
                dictgen::InsensitiveStr::Ascii("calculcating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mercinaries"),
                &["mercenaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("debugggee"), &["debuggee"]),
            (
                dictgen::InsensitiveStr::Ascii("regeneartive"),
                &["regenerative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stalagtite"),
                &["stalactite", "stalagmite"],
            ),
            (dictgen::InsensitiveStr::Ascii("itearte"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("coommanded"), &["commanded"]),
            (dictgen::InsensitiveStr::Ascii("averload"), &["overload"]),
            (dictgen::InsensitiveStr::Ascii("reloccate"), &["relocate"]),
            (
                dictgen::InsensitiveStr::Ascii("unreproducable"),
                &["unreproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsequentely"),
                &["subsequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenifier"), &["identifier"]),
            (
                dictgen::InsensitiveStr::Ascii("transcosers"),
                &["transcoders"],
            ),
            (dictgen::InsensitiveStr::Ascii("neccesary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("handleer"), &["handler"]),
            (
                dictgen::InsensitiveStr::Ascii("practitionners"),
                &["practitioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("stucts"), &["structs"]),
            (dictgen::InsensitiveStr::Ascii("qeried"), &["queried"]),
            (dictgen::InsensitiveStr::Ascii("exproted"), &["exported"]),
            (
                dictgen::InsensitiveStr::Ascii("diferentiates"),
                &["differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insctuction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("lookign"), &["looking"]),
            (dictgen::InsensitiveStr::Ascii("crutchers"), &["crutches"]),
            (dictgen::InsensitiveStr::Ascii("espically"), &["especially"]),
            (dictgen::InsensitiveStr::Ascii("inerests"), &["interests"]),
            (dictgen::InsensitiveStr::Ascii("disanced"), &["distanced"]),
            (dictgen::InsensitiveStr::Ascii("spagetti"), &["spaghetti"]),
            (
                dictgen::InsensitiveStr::Ascii("innovatin"),
                &["innovations"],
            ),
            (dictgen::InsensitiveStr::Ascii("simlifies"), &["simplifies"]),
            (
                dictgen::InsensitiveStr::Ascii("ubiqutious"),
                &["ubiquitous"],
            ),
            (dictgen::InsensitiveStr::Ascii("denila"), &["denial"]),
            (dictgen::InsensitiveStr::Ascii("valie"), &["value", "valid"]),
            (
                dictgen::InsensitiveStr::Ascii("patriarchia"),
                &["patriarchal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assembleing"),
                &["assembling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temmporary"),
                &["temporary", "temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("divsions"), &["divisions"]),
            (dictgen::InsensitiveStr::Ascii("perisist"), &["persist"]),
            (
                dictgen::InsensitiveStr::Ascii("substitite"),
                &["substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("abov"), &["above"]),
            (
                dictgen::InsensitiveStr::Ascii("tempororily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedomers"), &["freedoms"]),
            (dictgen::InsensitiveStr::Ascii("exectuing"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("specialitzed"),
                &["specialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nanoparicle"),
                &["nanoparticle"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutatble"), &["mutable"]),
            (dictgen::InsensitiveStr::Ascii("craotia"), &["croatia"]),
            (dictgen::InsensitiveStr::Ascii("cahgned"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("electricy"),
                &["electricity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concenting"),
                &["connecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calclulated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("midifies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("negotionates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("auotmated"), &["automated"]),
            (dictgen::InsensitiveStr::Ascii("reslver"), &["resolver"]),
            (dictgen::InsensitiveStr::Ascii("algin"), &["align"]),
            (
                dictgen::InsensitiveStr::Ascii("dismantaled"),
                &["dismantled"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajacent"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("confidentuality"),
                &["confidential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucracie"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerbures"), &["cerberus"]),
            (dictgen::InsensitiveStr::Ascii("judisuary"), &["judiciary"]),
            (dictgen::InsensitiveStr::Ascii("fugure"), &["figure"]),
            (
                dictgen::InsensitiveStr::Ascii("ctificates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaiton"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accodrance"),
                &["accordance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallcuinations"),
                &["hallucination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caclulators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existentiella"),
                &["existential"],
            ),
            (dictgen::InsensitiveStr::Ascii("embetting"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("passionale"),
                &["passionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("executig"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("magainzes"), &["magazines"]),
            (
                dictgen::InsensitiveStr::Ascii("behaviorial"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubute"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countination"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("temlated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("datbase"), &["database"]),
            (
                dictgen::InsensitiveStr::Ascii("progapating"),
                &["propagating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependendently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("aliasas"), &["aliases"]),
            (dictgen::InsensitiveStr::Ascii("closley"), &["closely"]),
            (dictgen::InsensitiveStr::Ascii("temr"), &["term"]),
            (dictgen::InsensitiveStr::Ascii("optimzing"), &["optimizing"]),
            (dictgen::InsensitiveStr::Ascii("primsry"), &["primary"]),
            (dictgen::InsensitiveStr::Ascii("dectects"), &["detects"]),
            (dictgen::InsensitiveStr::Ascii("horzontal"), &["horizontal"]),
            (
                dictgen::InsensitiveStr::Ascii("temperment"),
                &["temperament"],
            ),
            (dictgen::InsensitiveStr::Ascii("froget"), &["forget"]),
            (
                dictgen::InsensitiveStr::Ascii("executubles"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("approche"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("cathloics"), &["catholics"]),
            (dictgen::InsensitiveStr::Ascii("candiates"), &["candidates"]),
            (
                dictgen::InsensitiveStr::Ascii("sarinomial"),
                &["ceremonial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintelligant"),
                &["unintelligent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandinavion"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("peice"), &["piece"]),
            (
                dictgen::InsensitiveStr::Ascii("programmetically"),
                &["programmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("finializes"), &["finalizes"]),
            (
                dictgen::InsensitiveStr::Ascii("traveersing"),
                &["traversing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutiplier"), &["multiplier"]),
            (
                dictgen::InsensitiveStr::Ascii("friendsies"),
                &["friendlies"],
            ),
            (dictgen::InsensitiveStr::Ascii("beteshda"), &["bethesda"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluatating"),
                &["evaluating"],
            ),
            (dictgen::InsensitiveStr::Ascii("seletes"), &["deletes"]),
            (dictgen::InsensitiveStr::Ascii("returnst"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("scatching"),
                &["catching", "scratching", "sketching"],
            ),
            (dictgen::InsensitiveStr::Ascii("compoment"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("laoded"), &["loaded"]),
            (dictgen::InsensitiveStr::Ascii("quailfy"), &["qualify"]),
            (dictgen::InsensitiveStr::Ascii("filenmaes"), &["filenames"]),
            (dictgen::InsensitiveStr::Ascii("capactior"), &["capacitor"]),
            (
                dictgen::InsensitiveStr::Ascii("diffusin"),
                &["diffusing", "diffusion"],
            ),
            (dictgen::InsensitiveStr::Ascii("setable"), &["settable"]),
            (
                dictgen::InsensitiveStr::Ascii("bulletbroof"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qiestionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emmigrated"),
                &["emigrated", "immigrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qaulifications"),
                &["qualification"],
            ),
            (dictgen::InsensitiveStr::Ascii("aduiobook"), &["audiobook"]),
            (dictgen::InsensitiveStr::Ascii("expesses"), &["expresses"]),
            (
                dictgen::InsensitiveStr::Ascii("automitive"),
                &["automotive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ooutput"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("dollor"), &["dollar"]),
            (
                dictgen::InsensitiveStr::Ascii("photogragh"),
                &["photograph"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowiedgements"),
                &["acknowledgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstnading"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("girlfirends"),
                &["girlfriends"],
            ),
            (dictgen::InsensitiveStr::Ascii("avanced"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("maintianer"),
                &["maintainer"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispensare"), &["dispenser"]),
            (
                dictgen::InsensitiveStr::Ascii("tempearture"),
                &["temperatures", "temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("survivers"), &["survivors"]),
            (
                dictgen::InsensitiveStr::Ascii("copytighted"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciaing"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("livng"), &["living"]),
            (
                dictgen::InsensitiveStr::Ascii("psychaitrist"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("advenced"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("bottelneck"),
                &["bottleneck"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompasitated"),
                &["incapacitated"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjuscent"), &["adjacent"]),
            (dictgen::InsensitiveStr::Ascii("ehtically"), &["ethically"]),
            (
                dictgen::InsensitiveStr::Ascii("immitations"),
                &["imitations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symmmetrically"),
                &["symmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exersiced"), &["exercised"]),
            (dictgen::InsensitiveStr::Ascii("bleutooth"), &["bluetooth"]),
            (dictgen::InsensitiveStr::Ascii("tthank"), &["thank"]),
            (
                dictgen::InsensitiveStr::Ascii("intercepcion"),
                &["interception"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obstruccion"),
                &["obstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idebntifiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformare"),
                &["transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("intgrated"), &["integrated"]),
            (dictgen::InsensitiveStr::Ascii("capitulo"), &["capitol"]),
            (dictgen::InsensitiveStr::Ascii("ulters"), &["alters"]),
            (
                dictgen::InsensitiveStr::Ascii("porgramming"),
                &["programming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wonce"),
                &["nonce", "once", "ponce", "wince"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgarade"), &["upgrade"]),
            (dictgen::InsensitiveStr::Ascii("bankrupcy"), &["bankruptcy"]),
            (
                dictgen::InsensitiveStr::Ascii("pararmeterization"),
                &["parameterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimenty"),
                &["complimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbelivebly"),
                &["unbelievably"],
            ),
            (dictgen::InsensitiveStr::Ascii("servin"), &["serving"]),
            (
                dictgen::InsensitiveStr::Ascii("civilasation"),
                &["civilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("peple"), &["people"]),
            (dictgen::InsensitiveStr::Ascii("folowws"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("constuctor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlyzers"), &["analyzers"]),
            (
                dictgen::InsensitiveStr::Ascii("chechpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("promtping"), &["prompting"]),
            (
                dictgen::InsensitiveStr::Ascii("effecticeness"),
                &["effectiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("cannister"), &["canister"]),
            (
                dictgen::InsensitiveStr::Ascii("referebces"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereally"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("discustingly"),
                &["disgustingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitation"), &["situation"]),
            (
                dictgen::InsensitiveStr::Ascii("maybellinne"),
                &["maybelline"],
            ),
            (dictgen::InsensitiveStr::Ascii("desperat"), &["desperate"]),
            (dictgen::InsensitiveStr::Ascii("somethins"), &["somethings"]),
            (dictgen::InsensitiveStr::Ascii("gaget"), &["gadget"]),
            (dictgen::InsensitiveStr::Ascii("allo"), &["allow"]),
            (dictgen::InsensitiveStr::Ascii("poylgon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("contracs"), &["contracts"]),
            (dictgen::InsensitiveStr::Ascii("specialits"), &["specials"]),
            (dictgen::InsensitiveStr::Ascii("connnected"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("genearics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("compeltelyt"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communicaion"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empathatic"),
                &["empathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condigurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("inferrin"), &["inferring"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburgler"),
                &["cheeseburger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apperances"),
                &["appearances"],
            ),
            (dictgen::InsensitiveStr::Ascii("evenlopes"), &["envelopes"]),
            (dictgen::InsensitiveStr::Ascii("continutes"), &["continues"]),
            (
                dictgen::InsensitiveStr::Ascii("resintalls"),
                &["reinstalls"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lastest"),
                &["latest", "last"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discountined"),
                &["discontinued"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecognized"), &["recognized"]),
            (dictgen::InsensitiveStr::Ascii("attemping"), &["attempting"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinate"),
                &["indefinite"],
            ),
            (dictgen::InsensitiveStr::Ascii("directionh"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("independentiste"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("linera"), &["linear"]),
            (
                dictgen::InsensitiveStr::Ascii("declerative"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("multicat"), &["multicast"]),
            (
                dictgen::InsensitiveStr::Ascii("initalling"),
                &["initialling"],
            ),
            (dictgen::InsensitiveStr::Ascii("conifigs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("prrottypes"),
                &["prototypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("deciml"), &["decimal"]),
            (
                dictgen::InsensitiveStr::Ascii("unempolyed"),
                &["unemployed"],
            ),
            (dictgen::InsensitiveStr::Ascii("defeines"), &["defines"]),
            (
                dictgen::InsensitiveStr::Ascii("lonlieness"),
                &["loneliness"],
            ),
            (dictgen::InsensitiveStr::Ascii("atteneded"), &["attended"]),
            (dictgen::InsensitiveStr::Ascii("disscuss"), &["discuss"]),
            (
                dictgen::InsensitiveStr::Ascii("capabiltities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("reserded"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("ahlpa"), &["alpha"]),
            (
                dictgen::InsensitiveStr::Ascii("apologistas"),
                &["apologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomands"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("inferioirty"),
                &["inferiority"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dyregulation"),
                &["deregulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepetuate"),
                &["perpetuate"],
            ),
            (dictgen::InsensitiveStr::Ascii("perpetuae"), &["perpetuate"]),
            (
                dictgen::InsensitiveStr::Ascii("hirarchically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miracilously"),
                &["miraculously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reassocation"),
                &["reassociation"],
            ),
            (dictgen::InsensitiveStr::Ascii("miltiline"), &["multiline"]),
            (
                dictgen::InsensitiveStr::Ascii("universtiy"),
                &["university"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumarable"),
                &["enumerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("inprooved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("ananlysts"), &["analysts"]),
            (dictgen::InsensitiveStr::Ascii("assoiates"), &["associates"]),
            (dictgen::InsensitiveStr::Ascii("confugre"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("pointint"), &["pointing"]),
            (
                dictgen::InsensitiveStr::Ascii("compareable"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("althougth"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("transforations"),
                &["transformations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsitutes"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perposterous"),
                &["preposterous"],
            ),
            (dictgen::InsensitiveStr::Ascii("laughabley"), &["laughably"]),
            (dictgen::InsensitiveStr::Ascii("phtreads"), &["pthreads"]),
            (dictgen::InsensitiveStr::Ascii("pasrsed"), &["parsed"]),
            (
                dictgen::InsensitiveStr::Ascii("programmmer"),
                &["programme"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inturruptability"),
                &["interruptibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("satisifes"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("potentionally"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("endnoden"), &["endnode"]),
            (dictgen::InsensitiveStr::Ascii("refered"), &["referred"]),
            (dictgen::InsensitiveStr::Ascii("singolar"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("philosipher"),
                &["philosopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competive"),
                &["competitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("policitally"),
                &["politically"],
            ),
            (dictgen::InsensitiveStr::Ascii("inh"), &["in"]),
            (
                dictgen::InsensitiveStr::Ascii("categoried"),
                &["categorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscribres"),
                &["subscribers", "subscribes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origininate"),
                &["originate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neaded"),
                &["needed", "kneaded", "headed"],
            ),
            (dictgen::InsensitiveStr::Ascii("furance"), &["furnace"]),
            (dictgen::InsensitiveStr::Ascii("scinece"), &["science"]),
            (dictgen::InsensitiveStr::Ascii("quizes"), &["quizzes"]),
            (dictgen::InsensitiveStr::Ascii("defulted"), &["defaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("handshacked"),
                &["handshaked"],
            ),
            (dictgen::InsensitiveStr::Ascii("heistate"), &["hesitate"]),
            (
                dictgen::InsensitiveStr::Ascii("transpanted"),
                &["transplanted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rienforcements"),
                &["reinforcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("revrsing"), &["reversing"]),
            (
                dictgen::InsensitiveStr::Ascii("improsoned"),
                &["imprisoned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precission"),
                &["precision", "percussion", "precession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("bathrom"), &["bathroom"]),
            (
                dictgen::InsensitiveStr::Ascii("regesses"),
                &["regresses", "recesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("intenal"), &["internal"]),
            (
                dictgen::InsensitiveStr::Ascii("definantley"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("intput"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("iserted"), &["inserted"]),
            (dictgen::InsensitiveStr::Ascii("expempt"), &["exempt"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpetcedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expereinced"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circualtion"),
                &["circulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ceckerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("dalse"), &["false"]),
            (dictgen::InsensitiveStr::Ascii("disariable"), &["desirable"]),
            (dictgen::InsensitiveStr::Ascii("secnds"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("measued"), &["measured"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaitions"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disbilities"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annnotating"),
                &["annotating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anayltical"),
                &["analytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("doughter"), &["daughter"]),
            (
                dictgen::InsensitiveStr::Ascii("mutlipler"),
                &["multiplier", "multiple"],
            ),
            (dictgen::InsensitiveStr::Ascii("beign"), &["being", "begin"]),
            (dictgen::InsensitiveStr::Ascii("exudin"), &["exuding"]),
            (
                dictgen::InsensitiveStr::Ascii("notirications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("direectly"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("suficating"),
                &["suffocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttention"), &["attention"]),
            (dictgen::InsensitiveStr::Ascii("cleaness"), &["cleanness"]),
            (dictgen::InsensitiveStr::Ascii("modyfies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("accommadation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recrusions"),
                &["recursions"],
            ),
            (dictgen::InsensitiveStr::Ascii("logwritter"), &["logwriter"]),
            (dictgen::InsensitiveStr::Ascii("multily"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("functiionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benevloent"),
                &["benevolent"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplifys"), &["simplifies"]),
            (dictgen::InsensitiveStr::Ascii("tiemd"), &["timed"]),
            (dictgen::InsensitiveStr::Ascii("tunelling"), &["tunnelling"]),
            (dictgen::InsensitiveStr::Ascii("jerusaleum"), &["jerusalem"]),
            (dictgen::InsensitiveStr::Ascii("evalueates"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("reuplading"),
                &["reuploading"],
            ),
            (dictgen::InsensitiveStr::Ascii("featuing"), &["featuring"]),
            (
                dictgen::InsensitiveStr::Ascii("unspefcifieid"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("conncted"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("assginee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("temaplates"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("malicously"),
                &["maliciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("splig"),
                &["split", "splign"],
            ),
            (dictgen::InsensitiveStr::Ascii("countinued"), &["continued"]),
            (dictgen::InsensitiveStr::Ascii("olypmic"), &["olympic"]),
            (dictgen::InsensitiveStr::Ascii("lcoalized"), &["localized"]),
            (dictgen::InsensitiveStr::Ascii("rescueing"), &["rescuing"]),
            (dictgen::InsensitiveStr::Ascii("ednif"), &["endif"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronizer"),
                &["synchronizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreacting"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("modellled"), &["modelled"]),
            (dictgen::InsensitiveStr::Ascii("follong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("coindice"), &["coincide"]),
            (dictgen::InsensitiveStr::Ascii("infinit"), &["infinite"]),
            (dictgen::InsensitiveStr::Ascii("cesspoool"), &["cesspool"]),
            (
                dictgen::InsensitiveStr::Ascii("tesed"),
                &["used", "teased", "tested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitionned"),
                &["transitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("equls"), &["equals"]),
            (
                dictgen::InsensitiveStr::Ascii("akkumulator"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("notebok"), &["notebook"]),
            (dictgen::InsensitiveStr::Ascii("dependeny"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("magnifecent"),
                &["magnificent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurtion"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("processse"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("realationships"),
                &["relationships"],
            ),
            (dictgen::InsensitiveStr::Ascii("detctable"), &["detectable"]),
            (dictgen::InsensitiveStr::Ascii("seqeunced"), &["sequenced"]),
            (
                dictgen::InsensitiveStr::Ascii("unresgisterd"),
                &["unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definnition"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geographia"),
                &["geographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atribution"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("occurrring"), &["occurring"]),
            (dictgen::InsensitiveStr::Ascii("intack"), &["intact"]),
            (
                dictgen::InsensitiveStr::Ascii("casulaties"),
                &["casualties"],
            ),
            (dictgen::InsensitiveStr::Ascii("proponet"), &["proponents"]),
            (
                dictgen::InsensitiveStr::Ascii("ancedotally"),
                &["anecdotally"],
            ),
            (dictgen::InsensitiveStr::Ascii("unuseal"), &["unusual"]),
            (dictgen::InsensitiveStr::Ascii("procesors"), &["processors"]),
            (dictgen::InsensitiveStr::Ascii("dpeendent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("prependicular"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectuel"),
                &["intellectual"],
            ),
            (dictgen::InsensitiveStr::Ascii("projcets"), &["projects"]),
            (dictgen::InsensitiveStr::Ascii("heighted"), &["heightened"]),
            (
                dictgen::InsensitiveStr::Ascii("claravoyants"),
                &["clairvoyants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveince"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("washingtion"),
                &["washington"],
            ),
            (dictgen::InsensitiveStr::Ascii("sreen"), &["screen"]),
            (dictgen::InsensitiveStr::Ascii("intsaller"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("ramdomize"), &["randomize"]),
            (
                dictgen::InsensitiveStr::Ascii("infomration"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("trubadour"), &["troubadour"]),
            (dictgen::InsensitiveStr::Ascii("trialers"), &["trailers"]),
            (dictgen::InsensitiveStr::Ascii("pege"), &["page"]),
            (
                dictgen::InsensitiveStr::Ascii("physcologically"),
                &["psychologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("supposeded"), &["supposed"]),
            (
                dictgen::InsensitiveStr::Ascii("reccommended"),
                &["recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionairy"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("cathegory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("quittin"), &["quitting"]),
            (dictgen::InsensitiveStr::Ascii("qeueued"), &["queued"]),
            (dictgen::InsensitiveStr::Ascii("potetial"), &["potential"]),
            (dictgen::InsensitiveStr::Ascii("speparator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("checkes"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("expetional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condifured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("resluts"), &["results"]),
            (dictgen::InsensitiveStr::Ascii("holliday"), &["holiday"]),
            (
                dictgen::InsensitiveStr::Ascii("offisianado"),
                &["aficionado"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomeed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("unbouned"), &["unbounded"]),
            (
                dictgen::InsensitiveStr::Ascii("terminaters"),
                &["terminators", "terminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("twpo"), &["two"]),
            (dictgen::InsensitiveStr::Ascii("portriats"), &["portraits"]),
            (
                dictgen::InsensitiveStr::Ascii("coeffitients"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuqired"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("parmter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("anothner"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("hradware"), &["hardware"]),
            (
                dictgen::InsensitiveStr::Ascii("genearations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepednent"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoyying"), &["annoying"]),
            (dictgen::InsensitiveStr::Ascii("confuzzing"), &["confusing"]),
            (dictgen::InsensitiveStr::Ascii("mappping"), &["mapping"]),
            (dictgen::InsensitiveStr::Ascii("wharever"), &["wherever"]),
            (dictgen::InsensitiveStr::Ascii("prompty"), &["promptly"]),
            (
                dictgen::InsensitiveStr::Ascii("intermediete"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("alchimest"), &["alchemist"]),
            (dictgen::InsensitiveStr::Ascii("napommed"), &["napalmed"]),
            (dictgen::InsensitiveStr::Ascii("typpical"), &["typical"]),
            (
                dictgen::InsensitiveStr::Ascii("oppend"),
                &["opened", "append"],
            ),
            (dictgen::InsensitiveStr::Ascii("deviiation"), &["deviation"]),
            (
                dictgen::InsensitiveStr::Ascii("sistematics"),
                &["systematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("exntend"), &["extend"]),
            (dictgen::InsensitiveStr::Ascii("simualtor"), &["simulator"]),
            (dictgen::InsensitiveStr::Ascii("continuu"), &["continuous"]),
            (dictgen::InsensitiveStr::Ascii("metaphore"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("spefiier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("democraphics"),
                &["demographics"],
            ),
            (dictgen::InsensitiveStr::Ascii("bisnes"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("mesures"), &["measures"]),
            (dictgen::InsensitiveStr::Ascii("maximuns"), &["maximums"]),
            (
                dictgen::InsensitiveStr::Ascii("asisstants"),
                &["assistants"],
            ),
            (dictgen::InsensitiveStr::Ascii("optionial"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("currentlly"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("portrayls"), &["portrays"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotatiations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruicble"), &["crucible"]),
            (dictgen::InsensitiveStr::Ascii("empted"), &["emptied"]),
            (dictgen::InsensitiveStr::Ascii("depostis"), &["deposits"]),
            (
                dictgen::InsensitiveStr::Ascii("saskatchawan"),
                &["saskatchewan"],
            ),
            (dictgen::InsensitiveStr::Ascii("hostory"), &["history"]),
            (
                dictgen::InsensitiveStr::Ascii("progresion"),
                &["progression"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceptionn"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("interationalist"),
                &["internationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("truncatin"),
                &["truncating", "truncation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instromenting"),
                &["instrumenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unerstands"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("socailism"), &["socialism"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoitated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("foced"), &["forced"]),
            (
                dictgen::InsensitiveStr::Ascii("refromation"),
                &["reformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchaneged"), &["unchanged"]),
            (dictgen::InsensitiveStr::Ascii("multiplle"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("comletly"), &["completely"]),
            (dictgen::InsensitiveStr::Ascii("onlie"), &["online", "only"]),
            (dictgen::InsensitiveStr::Ascii("expericne"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("unvalidate"),
                &["invalidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepeates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("interept"), &["intercept"]),
            (
                dictgen::InsensitiveStr::Ascii("magnifine"),
                &["magnificent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("userrname"), &["username"]),
            (dictgen::InsensitiveStr::Ascii("conected"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("simmilarly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("corrresponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pursuin"),
                &["pursuing", "pursuit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aaproximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartdrige"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoweldging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreivate"),
                &["abbreviate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzelusly"),
                &["overzealously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cariacture"),
                &["caricature"],
            ),
            (dictgen::InsensitiveStr::Ascii("updgrades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("squareds"), &["squares"]),
            (
                dictgen::InsensitiveStr::Ascii("sctioning"),
                &["sectioning", "suctioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantasmically"),
                &["fantastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exagon"), &["hexagon"]),
            (
                dictgen::InsensitiveStr::Ascii("reather"),
                &["feather", "rather"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dterminated"),
                &["determined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifcant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("clipboads"), &["clipboards"]),
            (
                dictgen::InsensitiveStr::Ascii("resturcturation"),
                &["restructuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatatble"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ccertification"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("definityl"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("regualtors"),
                &["regulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbur"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("brobability"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscribar"),
                &["subscriber"],
            ),
            (dictgen::InsensitiveStr::Ascii("studois"), &["studios"]),
            (dictgen::InsensitiveStr::Ascii("maximiums"), &["maximums"]),
            (dictgen::InsensitiveStr::Ascii("occuppying"), &["occupying"]),
            (
                dictgen::InsensitiveStr::Ascii("settelements"),
                &["settlements"],
            ),
            (dictgen::InsensitiveStr::Ascii("manigest"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("diplayed"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("operared"), &["operated"]),
            (dictgen::InsensitiveStr::Ascii("attribut"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("cababge"), &["cabbage"]),
            (
                dictgen::InsensitiveStr::Ascii("substituding"),
                &["substituting"],
            ),
            (dictgen::InsensitiveStr::Ascii("carraige"), &["carriage"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("minoins"), &["minions"]),
            (
                dictgen::InsensitiveStr::Ascii("expeimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spects"),
                &["aspects", "expects", "specs"],
            ),
            (dictgen::InsensitiveStr::Ascii("infraread"), &["infrared"]),
            (
                dictgen::InsensitiveStr::Ascii("prerelasing"),
                &["prereleasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tribunaal"), &["tribunal"]),
            (
                dictgen::InsensitiveStr::Ascii("toubleshooted"),
                &["troubleshooted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referneces"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polynomals"),
                &["polynomials"],
            ),
            (dictgen::InsensitiveStr::Ascii("expessive"), &["expressive"]),
            (dictgen::InsensitiveStr::Ascii("beeing"), &["being", "been"]),
            (dictgen::InsensitiveStr::Ascii("startd"), &["started"]),
            (dictgen::InsensitiveStr::Ascii("alloqed"), &["allowed"]),
            (dictgen::InsensitiveStr::Ascii("pinnaple"), &["pineapple"]),
            (
                dictgen::InsensitiveStr::Ascii("manifacturer"),
                &["manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("sotrs"), &["sorts"]),
            (dictgen::InsensitiveStr::Ascii("paralzyed"), &["paralyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("frustraties"),
                &["frustrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaintly"), &["defiantly"]),
            (dictgen::InsensitiveStr::Ascii("capitalisn"), &["capitals"]),
            (
                dictgen::InsensitiveStr::Ascii("correponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conductuctances"),
                &["conductances"],
            ),
            (dictgen::InsensitiveStr::Ascii("doomsdaily"), &["doomsday"]),
            (
                dictgen::InsensitiveStr::Ascii("cosntructed"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertians"), &["pertains"]),
            (dictgen::InsensitiveStr::Ascii("uniquelly"), &["uniquely"]),
            (dictgen::InsensitiveStr::Ascii("locaing"), &["locating"]),
            (dictgen::InsensitiveStr::Ascii("agruable"), &["arguable"]),
            (dictgen::InsensitiveStr::Ascii("aloows"), &["allows"]),
            (dictgen::InsensitiveStr::Ascii("obhect"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("conserned"), &["concerned"]),
            (dictgen::InsensitiveStr::Ascii("alraedy"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("initailizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("angshios"), &["anxious"]),
            (
                dictgen::InsensitiveStr::Ascii("resposnive"),
                &["responsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deply"),
                &["deploy", "deeply"],
            ),
            (dictgen::InsensitiveStr::Ascii("insatnt"), &["instant"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessecarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assessement"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipuling"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("libralies"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("authorizin"),
                &["authorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pascheurized"),
                &["pasteurized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disinguish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomon"), &["common"]),
            (dictgen::InsensitiveStr::Ascii("arbatrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("functiton"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("exerimenting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberration"),
                &["aberration"],
            ),
            (dictgen::InsensitiveStr::Ascii("infomer"), &["informer"]),
            (dictgen::InsensitiveStr::Ascii("limitiers"), &["limiters"]),
            (
                dictgen::InsensitiveStr::Ascii("presecution"),
                &["persecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reparamterise"),
                &["reparameterise"],
            ),
            (dictgen::InsensitiveStr::Ascii("candidat"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("formulaical"),
                &["formulaic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prominance"),
                &["prominence"],
            ),
            (dictgen::InsensitiveStr::Ascii("aftherward"), &["afterward"]),
            (
                dictgen::InsensitiveStr::Ascii("ipmrovements"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypnoss"), &["hypnosis"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacure"),
                &["manufacture"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoher"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("decompresser"),
                &["decompressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("tihnks"), &["thinks"]),
            (
                dictgen::InsensitiveStr::Ascii("coeffient"),
                &["coefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifiyng"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploarized"), &["polarized"]),
            (dictgen::InsensitiveStr::Ascii("morisette"), &["morissette"]),
            (
                dictgen::InsensitiveStr::Ascii("reinforcemens"),
                &["reinforcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("almostly"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("thermodynmaics"),
                &["thermodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boilerplatte"),
                &["boilerplate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconditionable"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("diret"), &["direct"]),
            (dictgen::InsensitiveStr::Ascii("diffent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("pijeon"), &["pigeon"]),
            (
                dictgen::InsensitiveStr::Ascii("teamfighters"),
                &["teamfights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premeptively"),
                &["preemptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("teritory"), &["territory"]),
            (dictgen::InsensitiveStr::Ascii("titainum"), &["titanium"]),
            (dictgen::InsensitiveStr::Ascii("ancesters"), &["ancestors"]),
            (
                dictgen::InsensitiveStr::Ascii("disssembles"),
                &["disassembles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnreabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missspelling"),
                &["misspelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("defininite"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("confingured"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipules"),
                &["manipulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("descibed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmicist"),
                &["pharmacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sculpters"),
                &["sculptors", "sculptures"],
            ),
            (dictgen::InsensitiveStr::Ascii("compleetes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("feministers"),
                &["feminists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preinitalizing"),
                &["preinitializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("filtype"), &["filetype"]),
            (dictgen::InsensitiveStr::Ascii("tirgger"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("cloen"), &["clone"]),
            (
                dictgen::InsensitiveStr::Ascii("cylcically"),
                &["cyclically"],
            ),
            (dictgen::InsensitiveStr::Ascii("symantics"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("degenerages"),
                &["degenerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("equelity"), &["equality"]),
            (
                dictgen::InsensitiveStr::Ascii("teleportato"),
                &["teleportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestior"),
                &["manifestation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciaite"),
                &["appreciative"],
            ),
            (dictgen::InsensitiveStr::Ascii("genereic"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("identites"), &["identities"]),
            (dictgen::InsensitiveStr::Ascii("operatoin"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("audince"), &["audience"]),
            (
                dictgen::InsensitiveStr::Ascii("downgradin"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immedeielty"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("vechicle"), &["vehicle"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensionailties"),
                &["dimensionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("prposers"), &["proposers"]),
            (
                dictgen::InsensitiveStr::Ascii("healthercare"),
                &["healthcare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humilitied"),
                &["humiliated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("startegically"),
                &["strategically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncialized"),
                &["canonicalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentional"),
                &["potential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprietery"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("invalaid"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("chosers"), &["choosers"]),
            (dictgen::InsensitiveStr::Ascii("reputpose"), &["repurpose"]),
            (dictgen::InsensitiveStr::Ascii("caucasion"), &["caucasian"]),
            (
                dictgen::InsensitiveStr::Ascii("fianlising"),
                &["finalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticizin"),
                &["criticizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("temlate"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("fetured"), &["featured"]),
            (dictgen::InsensitiveStr::Ascii("impedence"), &["impedance"]),
            (dictgen::InsensitiveStr::Ascii("toggel"), &["toggle"]),
            (dictgen::InsensitiveStr::Ascii("sctipts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("playthrougth"),
                &["playthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assiociation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefiified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("intriduced"),
                &["introduced"],
            ),
            (dictgen::InsensitiveStr::Ascii("readby"), &["read"]),
            (
                dictgen::InsensitiveStr::Ascii("proteection"),
                &["protection"],
            ),
            (dictgen::InsensitiveStr::Ascii("bluetoot"), &["bluetooth"]),
            (
                dictgen::InsensitiveStr::Ascii("occurrrence"),
                &["occurrence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorelation"),
                &["correlation"],
            ),
            (dictgen::InsensitiveStr::Ascii("singltons"), &["singletons"]),
            (
                dictgen::InsensitiveStr::Ascii("introduceras"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("systhemerror"),
                &["systemerror"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configrues"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("interanls"), &["internals"]),
            (dictgen::InsensitiveStr::Ascii("cluprit"), &["culprit"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisin"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satifiable"),
                &["satisfiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranmits"), &["transmits"]),
            (
                dictgen::InsensitiveStr::Ascii("songling"),
                &["singling", "dongling"],
            ),
            (dictgen::InsensitiveStr::Ascii("collumn"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("mentaly"), &["mentally"]),
            (
                dictgen::InsensitiveStr::Ascii("conviniences"),
                &["conveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accepetable"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequenzing"),
                &["sequencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associdative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("emense"), &["immense"]),
            (
                dictgen::InsensitiveStr::Ascii("playthtough"),
                &["playthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenisonal"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilze"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterise"),
                &["parameterise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iummediately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("behviours"), &["behaviours"]),
            (dictgen::InsensitiveStr::Ascii("etroneous"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("negotites"), &["negotiates"]),
            (
                dictgen::InsensitiveStr::Ascii("ivalidating"),
                &["invalidating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cantelope"), &["cantaloupe"]),
            (
                dictgen::InsensitiveStr::Ascii("floding"),
                &["folding", "flooding"],
            ),
            (dictgen::InsensitiveStr::Ascii("biploar"), &["bipolar"]),
            (dictgen::InsensitiveStr::Ascii("emmisarry"), &["emissary"]),
            (
                dictgen::InsensitiveStr::Ascii("throuhgput"),
                &["throughput"],
            ),
            (dictgen::InsensitiveStr::Ascii("settigns"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("imprefections"),
                &["imperfections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semgentations"),
                &["segmentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("deafults"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("epxansion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("orhtodox"), &["orthodox"]),
            (dictgen::InsensitiveStr::Ascii("masterise"), &["masteries"]),
            (
                dictgen::InsensitiveStr::Ascii("confusses"),
                &["confuses", "confesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhorbitent"),
                &["exorbitant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("divable"),
                &["divisible", "disable", "diveable"],
            ),
            (dictgen::InsensitiveStr::Ascii("outisde"), &["outside"]),
            (dictgen::InsensitiveStr::Ascii("bootstras"), &["bootstraps"]),
            (dictgen::InsensitiveStr::Ascii("scrathces"), &["scratches"]),
            (
                dictgen::InsensitiveStr::Ascii("antiboitic"),
                &["antibiotics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conferedate"),
                &["confederate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncode"),
                &["unicode", "encode"],
            ),
            (dictgen::InsensitiveStr::Ascii("follin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("accets"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("prejection"),
                &["projection"],
            ),
            (dictgen::InsensitiveStr::Ascii("longitue"), &["longitude"]),
            (dictgen::InsensitiveStr::Ascii("checing"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("progarammers"),
                &["programmers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constans"),
                &["constants", "constance", "constant"],
            ),
            (dictgen::InsensitiveStr::Ascii("valuees"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("procotol"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("moudle"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("acomplishments"),
                &["accomplishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("strecthed"), &["stretched"]),
            (dictgen::InsensitiveStr::Ascii("gnerating"), &["generating"]),
            (dictgen::InsensitiveStr::Ascii("littl"), &["little"]),
            (
                dictgen::InsensitiveStr::Ascii("parititioning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("monopolis"), &["monopolies"]),
            (
                dictgen::InsensitiveStr::Ascii("unisntalling"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("redece"), &["reduce"]),
            (dictgen::InsensitiveStr::Ascii("vrsion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("paologies"), &["apologies"]),
            (dictgen::InsensitiveStr::Ascii("lithuim"), &["lithium"]),
            (
                dictgen::InsensitiveStr::Ascii("retramsmit"),
                &["retransmit"],
            ),
            (dictgen::InsensitiveStr::Ascii("overridin"), &["overriding"]),
            (dictgen::InsensitiveStr::Ascii("gerenates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("disrepectfully"),
                &["disrespectfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("securty"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("improters"), &["importers"]),
            (
                dictgen::InsensitiveStr::Ascii("guarnteing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accodingly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefifally"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normalyzation"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("epitamy"), &["epitome"]),
            (
                dictgen::InsensitiveStr::Ascii("custonizers"),
                &["customizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("occulusion"), &["occlusion"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdendent"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("euristic"), &["heuristic"]),
            (dictgen::InsensitiveStr::Ascii("paragaph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("poitd"), &["pointed"]),
            (
                dictgen::InsensitiveStr::Ascii("capitilises"),
                &["capitalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encripting"),
                &["encrypting"],
            ),
            (dictgen::InsensitiveStr::Ascii("renteris"), &["renters"]),
            (dictgen::InsensitiveStr::Ascii("optin"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("amound"), &["amount"]),
            (
                dictgen::InsensitiveStr::Ascii("repreciated"),
                &["depreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pottom"), &["bottom"]),
            (dictgen::InsensitiveStr::Ascii("weill"), &["will"]),
            (dictgen::InsensitiveStr::Ascii("includng"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("duplaication"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recusions"),
                &["recursions", "reclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("alterasion"),
                &["alteration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cahes"),
                &["caches", "aches"],
            ),
            (dictgen::InsensitiveStr::Ascii("crucialy"), &["crucially"]),
            (
                dictgen::InsensitiveStr::Ascii("cleanlyness"),
                &["cleanliness"],
            ),
            (dictgen::InsensitiveStr::Ascii("mentioed"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("eaturn"),
                &["return", "saturn", "eaten"],
            ),
            (dictgen::InsensitiveStr::Ascii("opemed"), &["opened"]),
            (dictgen::InsensitiveStr::Ascii("intriguied"), &["intrigue"]),
            (dictgen::InsensitiveStr::Ascii("suntask"), &["subtask"]),
            (
                dictgen::InsensitiveStr::Ascii("avalailable"),
                &["available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analysze"),
                &["analyse", "analyze"],
            ),
            (dictgen::InsensitiveStr::Ascii("grammer"), &["grammar"]),
            (
                dictgen::InsensitiveStr::Ascii("encriptions"),
                &["encryptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depening"),
                &["depending", "deepening"],
            ),
            (dictgen::InsensitiveStr::Ascii("slghtly"), &["slightly"]),
            (dictgen::InsensitiveStr::Ascii("rendayvou"), &["rendezvous"]),
            (
                dictgen::InsensitiveStr::Ascii("unintialsied"),
                &["uninitialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("selectie"), &["selective"]),
            (dictgen::InsensitiveStr::Ascii("eyasr"), &["years", "eyas"]),
            (dictgen::InsensitiveStr::Ascii("caucaisan"), &["caucasian"]),
            (dictgen::InsensitiveStr::Ascii("colums"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("wildget"), &["widget"]),
            (
                dictgen::InsensitiveStr::Ascii("imperailist"),
                &["imperialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanismus"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understoud"),
                &["understood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("begginings"),
                &["beginnings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finialisation"),
                &["finalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cthulluh"), &["cthulhu"]),
            (
                dictgen::InsensitiveStr::Ascii("documentiation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporaches"),
                &["approaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("agregate"), &["aggregate"]),
            (
                dictgen::InsensitiveStr::Ascii("apsotrophe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preste"),
                &["preset", "pressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("labeld"), &["labeled"]),
            (
                dictgen::InsensitiveStr::Ascii("backrgound"),
                &["backgrounds", "background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miraculousy"),
                &["miraculously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fascilitating"),
                &["facilitating"],
            ),
            (dictgen::InsensitiveStr::Ascii("collonies"), &["colonies"]),
            (
                dictgen::InsensitiveStr::Ascii("confuring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calliflowers"),
                &["cauliflowers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algolithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exersised"), &["exercised"]),
            (dictgen::InsensitiveStr::Ascii("exinct"), &["extinct"]),
            (
                dictgen::InsensitiveStr::Ascii("novisheates"),
                &["novitiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("budegets"), &["budgets"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillifes"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trugged"),
                &["trudged", "tugged"],
            ),
            (dictgen::InsensitiveStr::Ascii("requiere"), &["require"]),
            (dictgen::InsensitiveStr::Ascii("connetcor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("soop"),
                &["soup", "scoop", "snoop", "soap"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perserveres"),
                &["perseveres"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleives"), &["believes"]),
            (dictgen::InsensitiveStr::Ascii("gorws"), &["grows"]),
            (dictgen::InsensitiveStr::Ascii("modyfiers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("stopp"), &["stop"]),
            (dictgen::InsensitiveStr::Ascii("heping"), &["helping"]),
            (dictgen::InsensitiveStr::Ascii("toturials"), &["tutorials"]),
            (dictgen::InsensitiveStr::Ascii("gisers"), &["geysers"]),
            (dictgen::InsensitiveStr::Ascii("fomulates"), &["formulates"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnsotic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("hailfax"), &["halifax"]),
            (dictgen::InsensitiveStr::Ascii("missspell"), &["misspell"]),
            (
                dictgen::InsensitiveStr::Ascii("legalizaiton"),
                &["legalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thunberbolt"),
                &["thunderbolt"],
            ),
            (dictgen::InsensitiveStr::Ascii("withih"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("partiton"), &["partition"]),
            (
                dictgen::InsensitiveStr::Ascii("conducters"),
                &["conductors"],
            ),
            (dictgen::InsensitiveStr::Ascii("depedents"), &["dependents"]),
            (dictgen::InsensitiveStr::Ascii("mimimally"), &["minimally"]),
            (dictgen::InsensitiveStr::Ascii("couls"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("guarnate"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("arrivie"), &["arrive"]),
            (dictgen::InsensitiveStr::Ascii("menssage"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("initialie"),
                &["initialize", "initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desperatley"),
                &["desperately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apliance"),
                &["appliance", "alliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immedately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogonistic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environemtal"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demphasized"),
                &["deemphasized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deemphesized"),
                &["deemphasized"],
            ),
            (dictgen::InsensitiveStr::Ascii("stoarge"), &["storage"]),
            (dictgen::InsensitiveStr::Ascii("decerns"), &["discerns"]),
            (dictgen::InsensitiveStr::Ascii("peroperty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("tghe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("payements"),
                &["payments", "pavements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idetically"),
                &["identically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vairance"), &["variance"]),
            (dictgen::InsensitiveStr::Ascii("trunctates"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("lengten"), &["lengthen"]),
            (
                dictgen::InsensitiveStr::Ascii("derefernces"),
                &["dereferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdinight"), &["midnight"]),
            (
                dictgen::InsensitiveStr::Ascii("straightin"),
                &["straighten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefinitrelly"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyzators"),
                &["analyzers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excepional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("insiddes"), &["insides"]),
            (dictgen::InsensitiveStr::Ascii("dectorator"), &["decorator"]),
            (
                dictgen::InsensitiveStr::Ascii("oversubscirbed"),
                &["oversubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrad"), &["thread"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccuraccies"),
                &["inaccuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antisocail"),
                &["antisocial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contstructing"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("pararaph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("happpens"), &["happens"]),
            (
                dictgen::InsensitiveStr::Ascii("initiializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("surport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("portfoilo"), &["portfolio"]),
            (
                dictgen::InsensitiveStr::Ascii("ukowns"),
                &["unknowns", "unknown"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentances"), &["sentences"]),
            (dictgen::InsensitiveStr::Ascii("regimet"), &["regiment"]),
            (dictgen::InsensitiveStr::Ascii("sahre"), &["share"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depoy"),
                &["deploy", "depot", "decoy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amazin"),
                &["amazing", "amazon"],
            ),
            (dictgen::InsensitiveStr::Ascii("apach"), &["apache"]),
            (
                dictgen::InsensitiveStr::Ascii("differntiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("litaral"), &["literal"]),
            (
                dictgen::InsensitiveStr::Ascii("exclamativo"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("multiecast"), &["multicast"]),
            (dictgen::InsensitiveStr::Ascii("locahost"), &["localhost"]),
            (dictgen::InsensitiveStr::Ascii("updateed"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("burrtio"), &["burrito"]),
            (dictgen::InsensitiveStr::Ascii("reviwer"), &["reviewer"]),
            (
                dictgen::InsensitiveStr::Ascii("prdpagator"),
                &["propagator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessuflly"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("margaritte"), &["margaret"]),
            (
                dictgen::InsensitiveStr::Ascii("existencial"),
                &["existential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("therough"),
                &["through", "thorough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automatron"),
                &["automation"],
            ),
            (dictgen::InsensitiveStr::Ascii("teeangers"), &["teenagers"]),
            (dictgen::InsensitiveStr::Ascii("aribrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("miniumn"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("contollling"),
                &["controlling"],
            ),
            (dictgen::InsensitiveStr::Ascii("algebrra"), &["algebra"]),
            (
                dictgen::InsensitiveStr::Ascii("sotred"),
                &["sorted", "stored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cracefully"),
                &["gracefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometine"), &["sometime"]),
            (
                dictgen::InsensitiveStr::Ascii("missinterpreted"),
                &["misinterpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("periperal"), &["peripheral"]),
            (
                dictgen::InsensitiveStr::Ascii("bottlenects"),
                &["bottlenecks"],
            ),
            (dictgen::InsensitiveStr::Ascii("crontrols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("subsciber"), &["subscriber"]),
            (dictgen::InsensitiveStr::Ascii("desireable"), &["desirable"]),
            (
                dictgen::InsensitiveStr::Ascii("maked"),
                &["marked", "made", "makes"],
            ),
            (dictgen::InsensitiveStr::Ascii("opnion"), &["opinion"]),
            (dictgen::InsensitiveStr::Ascii("worte"), &["wrote"]),
            (dictgen::InsensitiveStr::Ascii("conmmutes"), &["commutes"]),
            (dictgen::InsensitiveStr::Ascii("telemitry"), &["telemetry"]),
            (
                dictgen::InsensitiveStr::Ascii("alphapeticaly"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimizer"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("lieve"), &["leave"]),
            (
                dictgen::InsensitiveStr::Ascii("percantile"),
                &["percentile"],
            ),
            (dictgen::InsensitiveStr::Ascii("franework"), &["framework"]),
            (
                dictgen::InsensitiveStr::Ascii("corrispondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("themplated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("blocekd"), &["blocked"]),
            (dictgen::InsensitiveStr::Ascii("pacaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("particulalrly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("braceletts"), &["bracelets"]),
            (dictgen::InsensitiveStr::Ascii("supress"), &["suppress"]),
            (
                dictgen::InsensitiveStr::Ascii("achevements"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("metrisc"), &["metrics"]),
            (
                dictgen::InsensitiveStr::Ascii("stabalizes"),
                &["stabilizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invulernable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reclaimation"),
                &["reclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbck"), &["callback"]),
            (
                dictgen::InsensitiveStr::Ascii("denined"),
                &["denied", "defined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degredation"),
                &["degradation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availailability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmport"), &["export"]),
            (dictgen::InsensitiveStr::Ascii("bigegr"), &["bigger"]),
            (dictgen::InsensitiveStr::Ascii("beliefing"), &["believing"]),
            (
                dictgen::InsensitiveStr::Ascii("continaers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("requet"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("sympethizers"),
                &["sympathizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("mediorce"), &["mediocre"]),
            (dictgen::InsensitiveStr::Ascii("funcsions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("exclued"),
                &["excluded", "exclude"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncoding"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("rummagin"), &["rummaging"]),
            (
                dictgen::InsensitiveStr::Ascii("abanonding"),
                &["abandoning"],
            ),
            (dictgen::InsensitiveStr::Ascii("promtions"), &["promotions"]),
            (
                dictgen::InsensitiveStr::Ascii("restructre"),
                &["restructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("arthroses"), &["arthrosis"]),
            (dictgen::InsensitiveStr::Ascii("spedific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("oportions"),
                &["options", "apportions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolegetics"),
                &["apologetics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossibillity"),
                &["impossibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tansaction"),
                &["transaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("felxibility"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accseptance"),
                &["acceptance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unperturbated"),
                &["unperturbed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incentiveise"),
                &["incentives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoultion"),
                &["revolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("outgoign"), &["outgoing"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapoles"),
                &["extrapolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("lodable"), &["loadable"]),
            (dictgen::InsensitiveStr::Ascii("cutscens"), &["cutscenes"]),
            (
                dictgen::InsensitiveStr::Ascii("desserializing"),
                &["deserializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conserns"), &["concerns"]),
            (
                dictgen::InsensitiveStr::Ascii("constrcuts"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("reacorded"), &["recorded"]),
            (
                dictgen::InsensitiveStr::Ascii("exeriencing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cigarete"),
                &["cigarets", "cigarette"],
            ),
            (dictgen::InsensitiveStr::Ascii("maange"), &["manage"]),
            (dictgen::InsensitiveStr::Ascii("usespace"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("prometheius"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("nautiuls"), &["nautilus"]),
            (dictgen::InsensitiveStr::Ascii("accodrs"), &["accords"]),
            (dictgen::InsensitiveStr::Ascii("overules"), &["overuse"]),
            (dictgen::InsensitiveStr::Ascii("delyaing"), &["delaying"]),
            (
                dictgen::InsensitiveStr::Ascii("archeaologists"),
                &["archeologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("featch"), &["fetch", "each"]),
            (dictgen::InsensitiveStr::Ascii("violetas"), &["violates"]),
            (
                dictgen::InsensitiveStr::Ascii("thogh"),
                &["though", "thigh"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrarly"),
                &["arbitrary", "arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("suddnely"), &["suddenly"]),
            (dictgen::InsensitiveStr::Ascii("atleats"), &["atleast"]),
            (dictgen::InsensitiveStr::Ascii("denosie"), &["denoise"]),
            (
                dictgen::InsensitiveStr::Ascii("notificication"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refeinement"),
                &["refinement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("additivley"),
                &["additive", "additively"],
            ),
            (dictgen::InsensitiveStr::Ascii("rmeote"), &["remote"]),
            (dictgen::InsensitiveStr::Ascii("chargin"), &["charging"]),
            (
                dictgen::InsensitiveStr::Ascii("corsses"),
                &["crosses", "courses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infaltable"),
                &["inflatable", "infallible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infectuous"),
                &["infectious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adventuros"),
                &["adventurous"],
            ),
            (dictgen::InsensitiveStr::Ascii("inser"), &["insert"]),
            (
                dictgen::InsensitiveStr::Ascii("momenterily"),
                &["momentarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exapct"),
                &["except", "exact"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wilderniss"),
                &["wilderness"],
            ),
            (dictgen::InsensitiveStr::Ascii("raisedd"), &["raised"]),
            (dictgen::InsensitiveStr::Ascii("carying"), &["carrying"]),
            (
                dictgen::InsensitiveStr::Ascii("presonality"),
                &["personality"],
            ),
            (dictgen::InsensitiveStr::Ascii("clos"), &["close"]),
            (
                dictgen::InsensitiveStr::Ascii("circumsizion"),
                &["circumcision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innacurate"),
                &["inaccurate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("therapeudic"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabitly"),
                &["compatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingquished"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("litarals"), &["literals"]),
            (dictgen::InsensitiveStr::Ascii("premesis"), &["premises"]),
            (
                dictgen::InsensitiveStr::Ascii("numbes"),
                &["numbers", "numbs", "numbed"],
            ),
            (dictgen::InsensitiveStr::Ascii("prvileged"), &["privileged"]),
            (dictgen::InsensitiveStr::Ascii("preapring"), &["preparing"]),
            (dictgen::InsensitiveStr::Ascii("deadlok"), &["deadlock"]),
            (
                dictgen::InsensitiveStr::Ascii("conservato"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("configurre"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("harrasses"),
                &["harassed", "harasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("posibly"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("seccessfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("benedit"), &["benedict"]),
            (dictgen::InsensitiveStr::Ascii("lawernce"), &["lawrence"]),
            (
                dictgen::InsensitiveStr::Ascii("respones"),
                &["response", "responses"],
            ),
            (dictgen::InsensitiveStr::Ascii("determinne"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("outadted"), &["outdated"]),
            (
                dictgen::InsensitiveStr::Ascii("patriotisim"),
                &["patriotism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciefications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conontation"),
                &["connotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inhabitans"),
                &["inhabitants"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterfaces"), &["interfaces"]),
            (dictgen::InsensitiveStr::Ascii("ddevices"), &["devices"]),
            (dictgen::InsensitiveStr::Ascii("relient"), &["reliant"]),
            (dictgen::InsensitiveStr::Ascii("conrrect"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("alreaydy"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("otimized"), &["optimized"]),
            (dictgen::InsensitiveStr::Ascii("institues"), &["institutes"]),
            (dictgen::InsensitiveStr::Ascii("extermity"), &["extremity"]),
            (dictgen::InsensitiveStr::Ascii("registser"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("runetime"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("preceeded"),
                &["preceded", "proceeded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provier"),
                &["provider", "prover"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertaincy"),
                &["uncertainty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laveling"),
                &["leveling", "labeling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unser"),
                &["under", "unset", "unsure", "user"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepreted"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presedence"),
                &["precedence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configutes"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinece"),
                &["experience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requierments"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalized"),
                &["orthogonalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("diversiy"), &["diversify"]),
            (
                dictgen::InsensitiveStr::Ascii("retailation"),
                &["retaliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implentors"),
                &["implementers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irreversebly"),
                &["irreversible"],
            ),
            (dictgen::InsensitiveStr::Ascii("frindly"), &["friendly"]),
            (dictgen::InsensitiveStr::Ascii("mohammedan"), &["muslim"]),
            (dictgen::InsensitiveStr::Ascii("magincian"), &["magician"]),
            (dictgen::InsensitiveStr::Ascii("forver"), &["forever"]),
            (
                dictgen::InsensitiveStr::Ascii("calcluated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traceablity"),
                &["traceability"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsidary"), &["subsidiary"]),
            (
                dictgen::InsensitiveStr::Ascii("comvolutions"),
                &["convolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilegied"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprted"), &["exported"]),
            (
                dictgen::InsensitiveStr::Ascii("squirrelius"),
                &["squirrels"],
            ),
            (dictgen::InsensitiveStr::Ascii("goruped"), &["grouped"]),
            (
                dictgen::InsensitiveStr::Ascii("snythesised"),
                &["synthesised"],
            ),
            (dictgen::InsensitiveStr::Ascii("abbility"), &["ability"]),
            (dictgen::InsensitiveStr::Ascii("tecnic"), &["technique"]),
            (dictgen::InsensitiveStr::Ascii("atttacker"), &["attacker"]),
            (dictgen::InsensitiveStr::Ascii("simiilar"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("conseed"), &["concede"]),
            (dictgen::InsensitiveStr::Ascii("guttaral"), &["guttural"]),
            (
                dictgen::InsensitiveStr::Ascii("arithmetric"),
                &["arithmetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eveluation"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuilld"), &["rebuild"]),
            (
                dictgen::InsensitiveStr::Ascii("proactivley"),
                &["proactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("posible"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("authenciate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("tupples"), &["tuples"]),
            (dictgen::InsensitiveStr::Ascii("entitees"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("interlaved"),
                &["interlaced", "interleaved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeficent"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("polygone"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("binrary"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("palestinans"),
                &["palestinians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acommodations"),
                &["accommodations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitmizers"),
                &["optimizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundimentally"),
                &["fundamentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optionallly"),
                &["optionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("authroing"), &["authoring"]),
            (dictgen::InsensitiveStr::Ascii("techneek"), &["technique"]),
            (dictgen::InsensitiveStr::Ascii("seviced"), &["serviced"]),
            (dictgen::InsensitiveStr::Ascii("checer"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("littlefiger"),
                &["littlefinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("referign"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("aditional"), &["additional"]),
            (dictgen::InsensitiveStr::Ascii("mysapce"), &["myspace"]),
            (dictgen::InsensitiveStr::Ascii("privilige"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("indictate"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("deactivete"),
                &["deactivate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overnurfed"),
                &["overturned"],
            ),
            (dictgen::InsensitiveStr::Ascii("characts"), &["characters"]),
            (dictgen::InsensitiveStr::Ascii("detectts"), &["detects"]),
            (
                dictgen::InsensitiveStr::Ascii("isntruction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("deceasin"), &["deceasing"]),
            (
                dictgen::InsensitiveStr::Ascii("omniscienct"),
                &["omniscient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priotising"),
                &["prioritising"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqments"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("geraff"), &["giraffe"]),
            (dictgen::InsensitiveStr::Ascii("stdouot"), &["stdout"]),
            (dictgen::InsensitiveStr::Ascii("enduced"), &["induced"]),
            (dictgen::InsensitiveStr::Ascii("alais"), &["alias"]),
            (
                dictgen::InsensitiveStr::Ascii("demosntrations"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intentinal"),
                &["intentional"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeended"), &["appended"]),
            (
                dictgen::InsensitiveStr::Ascii("configurble"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("committi"), &["committee"]),
            (
                dictgen::InsensitiveStr::Ascii("orignality"),
                &["originality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itnerprets"),
                &["interprets"],
            ),
            (dictgen::InsensitiveStr::Ascii("staidums"), &["stadiums"]),
            (dictgen::InsensitiveStr::Ascii("schemin"), &["scheming"]),
            (
                dictgen::InsensitiveStr::Ascii("enumration"),
                &["enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executabel"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganerics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("eroniously"),
                &["erroneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulaters"),
                &["regulators", "regulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thianks"),
                &["thanks", "thinks"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroollbar"), &["scrollbar"]),
            (
                dictgen::InsensitiveStr::Ascii("apporpiately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("interfave"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("perisistently"),
                &["persistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reurns"),
                &["returns", "reruns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rhythimcally"),
                &["rhythmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("tequlia"), &["tequila"]),
            (dictgen::InsensitiveStr::Ascii("butcherd"), &["butchered"]),
            (dictgen::InsensitiveStr::Ascii("distractes"), &["distracts"]),
            (dictgen::InsensitiveStr::Ascii("anaylsts"), &["analysts"]),
            (dictgen::InsensitiveStr::Ascii("wappers"), &["wrappers"]),
            (
                dictgen::InsensitiveStr::Ascii("mannarisms"),
                &["mannerisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("secuity"), &["security"]),
            (
                dictgen::InsensitiveStr::Ascii("asociating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("stupidr"), &["stupider"]),
            (
                dictgen::InsensitiveStr::Ascii("sreams"),
                &["streams", "screams"],
            ),
            (dictgen::InsensitiveStr::Ascii("foriners"), &["foreigners"]),
            (
                dictgen::InsensitiveStr::Ascii("ecnetricity"),
                &["eccentricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("intristic"), &["intrinsic"]),
            (
                dictgen::InsensitiveStr::Ascii("refelecting"),
                &["reflecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistrubutions"),
                &["redistributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromosmes"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsould"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("conntains"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("rapelled"), &["rappelled"]),
            (
                dictgen::InsensitiveStr::Ascii("bounardies"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precondtionner"),
                &["preconditioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("absorve"), &["absorb"]),
            (dictgen::InsensitiveStr::Ascii("defeined"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("braoden"), &["broaden"]),
            (dictgen::InsensitiveStr::Ascii("noifiers"), &["notifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("economisesti"),
                &["economists"],
            ),
            (dictgen::InsensitiveStr::Ascii("scorates"), &["socrates"]),
            (
                dictgen::InsensitiveStr::Ascii("technologicially"),
                &["technologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaluted"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("encluding"),
                &["including", "excluding"],
            ),
            (dictgen::InsensitiveStr::Ascii("analsyt"), &["analyst"]),
            (
                dictgen::InsensitiveStr::Ascii("librarie"),
                &["libraries", "library"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangents"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("counsle"), &["counsel"]),
            (
                dictgen::InsensitiveStr::Ascii("paramenets"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("gagits"), &["gadgets"]),
            (dictgen::InsensitiveStr::Ascii("murderd"), &["murdered"]),
            (dictgen::InsensitiveStr::Ascii("doiubled"), &["doubled"]),
            (dictgen::InsensitiveStr::Ascii("behavioutr"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("progams"), &["programs"]),
            (
                dictgen::InsensitiveStr::Ascii("experimntation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transolates"),
                &["translates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschedelic"),
                &["psychedelic"],
            ),
            (dictgen::InsensitiveStr::Ascii("acordians"), &["accordions"]),
            (dictgen::InsensitiveStr::Ascii("wokrbench"), &["workbench"]),
            (dictgen::InsensitiveStr::Ascii("toogle"), &["toggle"]),
            (dictgen::InsensitiveStr::Ascii("sucidial"), &["suicidal"]),
            (dictgen::InsensitiveStr::Ascii("poicies"), &["policies"]),
            (dictgen::InsensitiveStr::Ascii("imbracing"), &["embracing"]),
            (
                dictgen::InsensitiveStr::Ascii("advertisor"),
                &["advertiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("sabotagin"), &["sabotaging"]),
            (dictgen::InsensitiveStr::Ascii("nieve"), &["naive"]),
            (dictgen::InsensitiveStr::Ascii("aggrivate"), &["aggravate"]),
            (dictgen::InsensitiveStr::Ascii("aout"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("certaintly"), &["certainly"]),
            (
                dictgen::InsensitiveStr::Ascii("techinically"),
                &["technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinematograpy"),
                &["cinematography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomamces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("experimnt"), &["experiment"]),
            (
                dictgen::InsensitiveStr::Ascii("desturcted"),
                &["destructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poentially"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sketckes"),
                &["skates", "sketches"],
            ),
            (dictgen::InsensitiveStr::Ascii("comapt"), &["compat"]),
            (
                dictgen::InsensitiveStr::Ascii("coprrights"),
                &["copyrights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endagnering"),
                &["endangering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistnace"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceedes"), &["exceeds"]),
            (dictgen::InsensitiveStr::Ascii("cofusion"), &["confusion"]),
            (
                dictgen::InsensitiveStr::Ascii("sepeartions"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ineriority"),
                &["interiority", "inferiority"],
            ),
            (dictgen::InsensitiveStr::Ascii("indianas"), &["indians"]),
            (
                dictgen::InsensitiveStr::Ascii("administraters"),
                &["administrators", "administrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discourraged"),
                &["discouraged"],
            ),
            (dictgen::InsensitiveStr::Ascii("schow"), &["show", "chow"]),
            (dictgen::InsensitiveStr::Ascii("ccustoms"), &["customs"]),
            (dictgen::InsensitiveStr::Ascii("teminate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("lengh"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("denomerator"),
                &["enumerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("innovatief"), &["innovate"]),
            (
                dictgen::InsensitiveStr::Ascii("executabable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socialogical"),
                &["sociological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commig"),
                &["commit", "coming"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfoemer"), &["performer"]),
            (dictgen::InsensitiveStr::Ascii("makeing"), &["making"]),
            (dictgen::InsensitiveStr::Ascii("dessigned"), &["designed"]),
            (
                dictgen::InsensitiveStr::Ascii("conecurrency"),
                &["concurrency"],
            ),
            (dictgen::InsensitiveStr::Ascii("chenge"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("beakpoints"),
                &["breakpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("passtime"), &["pastime"]),
            (
                dictgen::InsensitiveStr::Ascii("pariticular"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("proose"), &["propose"]),
            (
                dictgen::InsensitiveStr::Ascii("fertiziler"),
                &["fertilizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaildations"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guaranteing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accepecters"),
                &["accepters"],
            ),
            (dictgen::InsensitiveStr::Ascii("bheavior"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("discogrpahy"),
                &["discography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preposterious"),
                &["preposterous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satistactory"),
                &["satisfactory"],
            ),
            (dictgen::InsensitiveStr::Ascii("currennt"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("stryofoam"), &["styrofoam"]),
            (
                dictgen::InsensitiveStr::Ascii("chaged"),
                &["changed", "charged"],
            ),
            (dictgen::InsensitiveStr::Ascii("tagliate"), &["tailgate"]),
            (dictgen::InsensitiveStr::Ascii("matchter"), &["matcher"]),
            (
                dictgen::InsensitiveStr::Ascii("difficulity"),
                &["difficulty"],
            ),
            (dictgen::InsensitiveStr::Ascii("grainery"), &["granary"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecessarity"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wereabouts"),
                &["whereabouts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postiioned"),
                &["positioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("serchable"), &["searchable"]),
            (dictgen::InsensitiveStr::Ascii("snowfalke"), &["snowflake"]),
            (
                dictgen::InsensitiveStr::Ascii("compatably"),
                &["compatibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("udpate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("creacionism"),
                &["creationism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corelations"),
                &["correlations"],
            ),
            (dictgen::InsensitiveStr::Ascii("burritoes"), &["burritos"]),
            (dictgen::InsensitiveStr::Ascii("immortalls"), &["immortals"]),
            (
                dictgen::InsensitiveStr::Ascii("interrputs"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subisdizing"),
                &["subsidizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("captian"), &["captain"]),
            (dictgen::InsensitiveStr::Ascii("emperical"), &["empirical"]),
            (
                dictgen::InsensitiveStr::Ascii("konsultation"),
                &["consultation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reundantly"),
                &["redundantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facillities"),
                &["facilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuitors"),
                &["distributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("strangreal"), &["strangle"]),
            (dictgen::InsensitiveStr::Ascii("testca"), &["testcase"]),
            (
                dictgen::InsensitiveStr::Ascii("materialsim"),
                &["materialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("bletooth"), &["bluetooth"]),
            (dictgen::InsensitiveStr::Ascii("asssit"), &["assist"]),
            (
                dictgen::InsensitiveStr::Ascii("construces"),
                &["constructs", "construes"],
            ),
            (dictgen::InsensitiveStr::Ascii("posisition"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("instructios"),
                &["instructors", "instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("florene"), &["florence"]),
            (
                dictgen::InsensitiveStr::Ascii("santitizers"),
                &["sanitizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("premuim"), &["premium"]),
            (
                dictgen::InsensitiveStr::Ascii("suspicioulsy"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggurating"),
                &["exaggerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpsondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguarded"), &["regarded"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiasers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diminisheds"),
                &["diminishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initliazed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabel"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("accessments"),
                &["assessments"],
            ),
            (dictgen::InsensitiveStr::Ascii("bracnhing"), &["branching"]),
            (dictgen::InsensitiveStr::Ascii("reliefed"), &["relieved"]),
            (
                dictgen::InsensitiveStr::Ascii("conecntrated"),
                &["concentrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscirbers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("protiens"), &["proteins"]),
            (dictgen::InsensitiveStr::Ascii("coaless"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonise"),
                &["synchronise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accesability"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstraccion"),
                &["abstraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("driectional"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("caost"), &["coast"]),
            (
                dictgen::InsensitiveStr::Ascii("oversemplification"),
                &["oversimplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternar"),
                &["alternator", "alternate"],
            ),
            (dictgen::InsensitiveStr::Ascii("analisers"), &["analysers"]),
            (dictgen::InsensitiveStr::Ascii("indepent"), &["independent"]),
            (
                dictgen::InsensitiveStr::Ascii("stationairy"),
                &["stationary"],
            ),
            (dictgen::InsensitiveStr::Ascii("descriing"), &["describing"]),
            (dictgen::InsensitiveStr::Ascii("enablin"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("egistrations"),
                &["registrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detrmination"),
                &["determination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompataility"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherweise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("lyin"), &["lying", "loin"]),
            (
                dictgen::InsensitiveStr::Ascii("prosecuter"),
                &["prosecutor"],
            ),
            (dictgen::InsensitiveStr::Ascii("upliftng"), &["uplifting"]),
            (
                dictgen::InsensitiveStr::Ascii("contraceptie"),
                &["contraceptives"],
            ),
            (dictgen::InsensitiveStr::Ascii("claerer"), &["clearer"]),
            (dictgen::InsensitiveStr::Ascii("vefiry"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("inproper"), &["improper"]),
            (dictgen::InsensitiveStr::Ascii("dectivate"), &["deactivate"]),
            (dictgen::InsensitiveStr::Ascii("framwork"), &["framework"]),
            (
                dictgen::InsensitiveStr::Ascii("succcessive"),
                &["successive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fulttershy"),
                &["fluttershy"],
            ),
            (dictgen::InsensitiveStr::Ascii("compny"), &["company"]),
            (
                dictgen::InsensitiveStr::Ascii("grandchilren"),
                &["grandchildren"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirfting"), &["drifting"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtenticating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("rythim"), &["rhythm"]),
            (
                dictgen::InsensitiveStr::Ascii("nnovisheate"),
                &["novitiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffenece"), &["difference"]),
            (
                dictgen::InsensitiveStr::Ascii("alernative"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspiciuosly"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opulate"),
                &["populate", "opiate", "opulent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legendarios"),
                &["legendaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcorder"),
                &["transcoder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considderation"),
                &["consideration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteteristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaative"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("multitaking"),
                &["multitasking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaiating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("excede"), &["exceed"]),
            (
                dictgen::InsensitiveStr::Ascii("functionalty"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("rotuers"), &["routers"]),
            (dictgen::InsensitiveStr::Ascii("assisstive"), &["assistive"]),
            (
                dictgen::InsensitiveStr::Ascii("irrelavent"),
                &["irrelevant"],
            ),
            (dictgen::InsensitiveStr::Ascii("paiting"), &["painting"]),
            (dictgen::InsensitiveStr::Ascii("unbouunded"), &["unbounded"]),
            (
                dictgen::InsensitiveStr::Ascii("hankerchif"),
                &["handkerchief"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplier"), &["simpler"]),
            (dictgen::InsensitiveStr::Ascii("symobls"), &["symbols"]),
            (
                dictgen::InsensitiveStr::Ascii("speciallize"),
                &["specialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concurency"),
                &["concurrency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkelists"),
                &["checklists"],
            ),
            (dictgen::InsensitiveStr::Ascii("solemly"), &["solemnly"]),
            (dictgen::InsensitiveStr::Ascii("hander"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("progarms"), &["programs"]),
            (dictgen::InsensitiveStr::Ascii("relevants"), &["relevant"]),
            (
                dictgen::InsensitiveStr::Ascii("environtments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("formua"), &["formula"]),
            (
                dictgen::InsensitiveStr::Ascii("executbales"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lieutennant"),
                &["lieutenant"],
            ),
            (dictgen::InsensitiveStr::Ascii("barnch"), &["branch"]),
            (dictgen::InsensitiveStr::Ascii("deside"), &["decide"]),
            (dictgen::InsensitiveStr::Ascii("enetered"), &["entered"]),
            (
                dictgen::InsensitiveStr::Ascii("equivlalent"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("actine"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronized"),
                &["synchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("derivatio"), &["derivation"]),
            (
                dictgen::InsensitiveStr::Ascii("abriviations"),
                &["abbreviations"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperatd"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("accountabilty"),
                &["accountability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absouletely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("codders"),
                &["coders", "coddlers"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainners"), &["trainers"]),
            (dictgen::InsensitiveStr::Ascii("threatenes"), &["threatens"]),
            (
                dictgen::InsensitiveStr::Ascii("preferablly"),
                &["preferably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priveleged"),
                &["privileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideologisen"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("touh"), &["touch"]),
            (
                dictgen::InsensitiveStr::Ascii("describs"),
                &["describe", "describes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistuently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilplication"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derector"),
                &["detector", "director", "erector"],
            ),
            (dictgen::InsensitiveStr::Ascii("clairfy"), &["clarify"]),
            (
                dictgen::InsensitiveStr::Ascii("supercalifragilisticexpialidoceous"),
                &["supercalifragilisticexpialidocious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transiever"),
                &["transceiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoledgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generilised"),
                &["generalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusor"), &["cursor"]),
            (
                dictgen::InsensitiveStr::Ascii("ankshiuosness"),
                &["anxiousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("athiets"), &["atheist"]),
            (dictgen::InsensitiveStr::Ascii("claerly"), &["clearly"]),
            (
                dictgen::InsensitiveStr::Ascii("polimorphism"),
                &["polymorphism"],
            ),
            (dictgen::InsensitiveStr::Ascii("confeses"), &["confess"]),
            (
                dictgen::InsensitiveStr::Ascii("feasbility"),
                &["feasibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("exector"), &["executor"]),
            (dictgen::InsensitiveStr::Ascii("poulate"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("desensitived"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chould"),
                &["should", "could"],
            ),
            (dictgen::InsensitiveStr::Ascii("exctract"), &["extract"]),
            (
                dictgen::InsensitiveStr::Ascii("runtines"),
                &["runtimes", "routines"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellision"), &["elision"]),
            (dictgen::InsensitiveStr::Ascii("venegance"), &["vengeance"]),
            (
                dictgen::InsensitiveStr::Ascii("incrmented"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("relfect"), &["reflect"]),
            (
                dictgen::InsensitiveStr::Ascii("associeation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("outputed"), &["outputted"]),
            (dictgen::InsensitiveStr::Ascii("gagit"), &["gadget"]),
            (dictgen::InsensitiveStr::Ascii("thurough"), &["thorough"]),
            (
                dictgen::InsensitiveStr::Ascii("reminscent"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("requsets"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("beddin"), &["bedding"]),
            (
                dictgen::InsensitiveStr::Ascii("descripitively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionalbe"),
                &["questionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("publcize"), &["publicize"]),
            (
                dictgen::InsensitiveStr::Ascii("prelimitary"),
                &["preliminary"],
            ),
            (dictgen::InsensitiveStr::Ascii("needlesly"), &["needlessly"]),
            (dictgen::InsensitiveStr::Ascii("critcals"), &["criticals"]),
            (
                dictgen::InsensitiveStr::Ascii("bodybuliding"),
                &["bodybuilding"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssigning"), &["assigning"]),
            (
                dictgen::InsensitiveStr::Ascii("comsumable"),
                &["consumable"],
            ),
            (dictgen::InsensitiveStr::Ascii("prestigue"), &["prestige"]),
            (dictgen::InsensitiveStr::Ascii("osffsets"), &["offsets"]),
            (
                dictgen::InsensitiveStr::Ascii("equest"),
                &["request", "quest"],
            ),
            (dictgen::InsensitiveStr::Ascii("accedent"), &["accident"]),
            (
                dictgen::InsensitiveStr::Ascii("consilidated"),
                &["consolidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("delegaties"), &["delegate"]),
            (
                dictgen::InsensitiveStr::Ascii("directiona"),
                &["directional", "directions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underratted"),
                &["underrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tendence"), &["tendencies"]),
            (dictgen::InsensitiveStr::Ascii("straignt"), &["straighten"]),
            (
                dictgen::InsensitiveStr::Ascii("littlefiner"),
                &["littlefinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruzed"), &["cruised"]),
            (dictgen::InsensitiveStr::Ascii("descisions"), &["decisions"]),
            (
                dictgen::InsensitiveStr::Ascii("negotionate"),
                &["negotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("arlingotn"), &["arlington"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonizes"),
                &["synchronizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deteoriated"),
                &["deteriorated"],
            ),
            (dictgen::InsensitiveStr::Ascii("wednseday"), &["wednesday"]),
            (
                dictgen::InsensitiveStr::Ascii("comaprable"),
                &["comparable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiencin"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expewrimenters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operrations"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenetations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hybird"), &["hybrid"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualted"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("influcence"), &["influence"]),
            (dictgen::InsensitiveStr::Ascii("husbans"), &["husbands"]),
            (dictgen::InsensitiveStr::Ascii("mamento"), &["memento"]),
            (
                dictgen::InsensitiveStr::Ascii("graudating"),
                &["graduating"],
            ),
            (dictgen::InsensitiveStr::Ascii("hirearchy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("highligthers"),
                &["highlighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("usigned"), &["unsigned"]),
            (
                dictgen::InsensitiveStr::Ascii("diagtnostic"),
                &["diagnostic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communiation"),
                &["communication"],
            ),
            (dictgen::InsensitiveStr::Ascii("prfer"), &["prefer"]),
            (dictgen::InsensitiveStr::Ascii("cirlcle"), &["circle"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitalise"),
                &["uninitialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("etherenet"), &["ethernet"]),
            (dictgen::InsensitiveStr::Ascii("spammade"), &["spammed"]),
            (
                dictgen::InsensitiveStr::Ascii("contension"),
                &["contention"],
            ),
            (dictgen::InsensitiveStr::Ascii("debouce"), &["debounce"]),
            (dictgen::InsensitiveStr::Ascii("eygptians"), &["egyptians"]),
            (
                dictgen::InsensitiveStr::Ascii("butterlfies"),
                &["butterflies"],
            ),
            (dictgen::InsensitiveStr::Ascii("pendning"), &["pending"]),
            (dictgen::InsensitiveStr::Ascii("whethe"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("dirbble"), &["dribble"]),
            (
                dictgen::InsensitiveStr::Ascii("initalised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("cautio"), &["caution"]),
            (dictgen::InsensitiveStr::Ascii("medeval"), &["medieval"]),
            (
                dictgen::InsensitiveStr::Ascii("accdientally"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("maitains"), &["maintains"]),
            (dictgen::InsensitiveStr::Ascii("rescuecd"), &["rescued"]),
            (
                dictgen::InsensitiveStr::Ascii("subdivisiond"),
                &["subdivisioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("definety"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("assesments"),
                &["assessments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wrokbenches"),
                &["workbenches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrbed"),
                &["unsubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transalated"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integraions"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcualate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("reprociblbe"),
                &["reproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finantially"),
                &["financially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presuambly"),
                &["presumably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelising"),
                &["parallelising"],
            ),
            (dictgen::InsensitiveStr::Ascii("poltic"), &["politic"]),
            (dictgen::InsensitiveStr::Ascii("aboce"), &["above", "abode"]),
            (dictgen::InsensitiveStr::Ascii("concer"), &["concern"]),
            (dictgen::InsensitiveStr::Ascii("misteeks"), &["mystiques"]),
            (
                dictgen::InsensitiveStr::Ascii("incurruptible"),
                &["incorruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenionality"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subequences"),
                &["subsequences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphasisze"),
                &["emphasize", "emphasise"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerebrawl"), &["cerebral"]),
            (
                dictgen::InsensitiveStr::Ascii("instrcutinos"),
                &["instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simpifiers"),
                &["simplifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hipsanics"), &["hispanics"]),
            (dictgen::InsensitiveStr::Ascii("scatterd"), &["scattered"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasuticaly"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("thte"), &["the", "that"]),
            (dictgen::InsensitiveStr::Ascii("critials"), &["criticals"]),
            (dictgen::InsensitiveStr::Ascii("coppying"), &["copying"]),
            (
                dictgen::InsensitiveStr::Ascii("surprizing"),
                &["surprising"],
            ),
            (dictgen::InsensitiveStr::Ascii("seciton"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("converet"), &["convert"]),
            (
                dictgen::InsensitiveStr::Ascii("geneartion"),
                &["generation"],
            ),
            (dictgen::InsensitiveStr::Ascii("statusses"), &["statuses"]),
            (
                dictgen::InsensitiveStr::Ascii("kighboshing"),
                &["kiboshing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bultin"), &["builtin"]),
            (
                dictgen::InsensitiveStr::Ascii("authoitative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquisiting"),
                &["inquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutices"), &["mutexes"]),
            (dictgen::InsensitiveStr::Ascii("scoll"), &["scroll"]),
            (
                dictgen::InsensitiveStr::Ascii("progamming"),
                &["programming"],
            ),
            (dictgen::InsensitiveStr::Ascii("cielings"), &["ceilings"]),
            (
                dictgen::InsensitiveStr::Ascii("undermineres"),
                &["undermines"],
            ),
            (dictgen::InsensitiveStr::Ascii("identcal"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("obsolet"), &["obsolete"]),
            (
                dictgen::InsensitiveStr::Ascii("accostumed"),
                &["accustomed"],
            ),
            (dictgen::InsensitiveStr::Ascii("bablyon"), &["babylon"]),
            (dictgen::InsensitiveStr::Ascii("supressed"), &["suppressed"]),
            (dictgen::InsensitiveStr::Ascii("disucusses"), &["discusses"]),
            (dictgen::InsensitiveStr::Ascii("retutned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("docmuent"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("dominato"), &["domination"]),
            (dictgen::InsensitiveStr::Ascii("cahanging"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("unitilized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scrennshot"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("umarked"), &["unmarked"]),
            (dictgen::InsensitiveStr::Ascii("cheduling"), &["scheduling"]),
            (
                dictgen::InsensitiveStr::Ascii("applicatable"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("busness"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("sohw"), &["show"]),
            (
                dictgen::InsensitiveStr::Ascii("alloud"),
                &["aloud", "allowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sematical"), &["semantical"]),
            (dictgen::InsensitiveStr::Ascii("mexcians"), &["mexicans"]),
            (dictgen::InsensitiveStr::Ascii("refilles"), &["refills"]),
            (dictgen::InsensitiveStr::Ascii("imbrases"), &["embraces"]),
            (dictgen::InsensitiveStr::Ascii("clearify"), &["clarify"]),
            (
                dictgen::InsensitiveStr::Ascii("succeesors"),
                &["successors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaliable"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("comliance"), &["compliance"]),
            (dictgen::InsensitiveStr::Ascii("docmented"), &["documented"]),
            (dictgen::InsensitiveStr::Ascii("beuatiful"), &["beautiful"]),
            (
                dictgen::InsensitiveStr::Ascii("unnacceptable"),
                &["unacceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprieciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounses"), &["bonuses"]),
            (dictgen::InsensitiveStr::Ascii("rightt"), &["right"]),
            (dictgen::InsensitiveStr::Ascii("flexiable"), &["flexible"]),
            (
                dictgen::InsensitiveStr::Ascii("anitdepressant"),
                &["antidepressants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surveillence"),
                &["surveillance"],
            ),
            (dictgen::InsensitiveStr::Ascii("implict"), &["implicit"]),
            (
                dictgen::InsensitiveStr::Ascii("coordiinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilato"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("leeguns"), &["legions"]),
            (dictgen::InsensitiveStr::Ascii("sylog"), &["syslog"]),
            (dictgen::InsensitiveStr::Ascii("evaliator"), &["evaluator"]),
            (
                dictgen::InsensitiveStr::Ascii("susbcribes"),
                &["subscribes"],
            ),
            (dictgen::InsensitiveStr::Ascii("dymanic"), &["dynamic"]),
            (
                dictgen::InsensitiveStr::Ascii("settelement"),
                &["settlement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivanets"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quesitoned"),
                &["questioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("managemenet"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("haveing"), &["having"]),
            (
                dictgen::InsensitiveStr::Ascii("subsequenty"),
                &["subsequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiializers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitilities"),
                &["utilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("virutual"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("indiciating"),
                &["indicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconditoner"),
                &["preconditioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properlty"),
                &["properly", "property"],
            ),
            (dictgen::InsensitiveStr::Ascii("caspule"), &["capsule"]),
            (dictgen::InsensitiveStr::Ascii("wacther"), &["watcher"]),
            (
                dictgen::InsensitiveStr::Ascii("mcroscopes"),
                &["microscopes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordiantes"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutomization"),
                &["customization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executuable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("invailid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("arhitecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("gererated"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("obejection"), &["objection"]),
            (dictgen::InsensitiveStr::Ascii("sysbols"), &["symbols"]),
            (
                dictgen::InsensitiveStr::Ascii("diliberately"),
                &["deliberately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintellegent"),
                &["unintelligent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coallition"),
                &["coalition", "collation"],
            ),
            (dictgen::InsensitiveStr::Ascii("personnell"), &["personnel"]),
            (
                dictgen::InsensitiveStr::Ascii("stormfornt"),
                &["stormfront"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("debehlper"), &["debhelper"]),
            (
                dictgen::InsensitiveStr::Ascii("laotion"),
                &["laotian", "lotion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poistioned"),
                &["positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lotation"),
                &["rotation", "flotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("emties"), &["empties"]),
            (
                dictgen::InsensitiveStr::Ascii("collaberate"),
                &["collaborate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaats"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrationallly"),
                &["irrationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("incerase"), &["increase"]),
            (dictgen::InsensitiveStr::Ascii("asessment"), &["assessment"]),
            (
                dictgen::InsensitiveStr::Ascii("uncompetetive"),
                &["uncompetitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trotskists"),
                &["trotskyists"],
            ),
            (dictgen::InsensitiveStr::Ascii("nashvillle"), &["nashville"]),
            (dictgen::InsensitiveStr::Ascii("wiat"), &["wait"]),
            (
                dictgen::InsensitiveStr::Ascii("participas"),
                &["participants"],
            ),
            (dictgen::InsensitiveStr::Ascii("begginers"), &["beginners"]),
            (
                dictgen::InsensitiveStr::Ascii("namemespaces"),
                &["namespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("crabine"), &["carbine"]),
            (dictgen::InsensitiveStr::Ascii("curremtly"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("diffenrent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("monolopy"), &["monopoly"]),
            (
                dictgen::InsensitiveStr::Ascii("indoctrication"),
                &["indoctrination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpersonel"),
                &["interpersonal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreigister"),
                &["unregister"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infromation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dominateurs"),
                &["dominates"],
            ),
            (dictgen::InsensitiveStr::Ascii("freshkly"), &["freshly"]),
            (
                dictgen::InsensitiveStr::Ascii("valiation"),
                &["validation", "valuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adovcating"),
                &["advocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundimentalists"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("brighens"), &["brightens"]),
            (dictgen::InsensitiveStr::Ascii("tommorow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("imigrant"),
                &["emigrant", "immigrant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definatelly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactered"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("falt"), &["fault", "flat"]),
            (dictgen::InsensitiveStr::Ascii("sampels"), &["samples"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasembling"),
                &["disassembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("traversier"), &["traverse"]),
            (
                dictgen::InsensitiveStr::Ascii("calcutators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("contolls"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("existense"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("chocolatte"),
                &["chocolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("bevause"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("datections"),
                &["detections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supercedes"),
                &["supersedes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahannel"), &["channel"]),
            (
                dictgen::InsensitiveStr::Ascii("supproting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("hydraluic"), &["hydraulic"]),
            (
                dictgen::InsensitiveStr::Ascii("expetimenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("enteties"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("clairvoiant"),
                &["clairvoyant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythetically"),
                &["synthetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("profirle"), &["profile"]),
            (
                dictgen::InsensitiveStr::Ascii("disgustinly"),
                &["disgustingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classfiers"),
                &["classifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("knwoing"), &["knowing"]),
            (dictgen::InsensitiveStr::Ascii("accesnt"), &["accent"]),
            (
                dictgen::InsensitiveStr::Ascii("contant"),
                &["constant", "content"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assisstance"),
                &["assistance"],
            ),
            (dictgen::InsensitiveStr::Ascii("setences"), &["sentences"]),
            (dictgen::InsensitiveStr::Ascii("stealthray"), &["stealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("independnent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelator"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("refereees"), &["referees"]),
            (
                dictgen::InsensitiveStr::Ascii("transaitions"),
                &["translations", "transitions", "transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infulential"),
                &["influential"],
            ),
            (dictgen::InsensitiveStr::Ascii("automatiek"), &["automate"]),
            (dictgen::InsensitiveStr::Ascii("fritions"), &["frictions"]),
            (dictgen::InsensitiveStr::Ascii("correectly"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("decendentant"),
                &["descendant"],
            ),
            (dictgen::InsensitiveStr::Ascii("otehr"), &["other"]),
            (dictgen::InsensitiveStr::Ascii("pleothra"), &["plethora"]),
            (
                dictgen::InsensitiveStr::Ascii("highligthing"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialilsers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentare"),
                &["documentaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sympathien"),
                &["sympathize"],
            ),
            (dictgen::InsensitiveStr::Ascii("cantact"), &["contact"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtorized"),
                &["authorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("absorbant"), &["absorbent"]),
            (dictgen::InsensitiveStr::Ascii("symnols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("aggenst"), &["against"]),
            (dictgen::InsensitiveStr::Ascii("eanbling"), &["enabling"]),
            (dictgen::InsensitiveStr::Ascii("arugments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("dependenct"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("spredsheet"),
                &["spreadsheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalisti"),
                &["fundamentals"],
            ),
            (dictgen::InsensitiveStr::Ascii("dorced"), &["forced"]),
            (dictgen::InsensitiveStr::Ascii("layour"), &["layout"]),
            (
                dictgen::InsensitiveStr::Ascii("cooparetive"),
                &["cooperative"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabledi"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("compense"), &["compensate"]),
            (dictgen::InsensitiveStr::Ascii("reviwers"), &["reviewers"]),
            (dictgen::InsensitiveStr::Ascii("checkoint"), &["checkpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpriting"),
                &["misinterpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paricipants"),
                &["participants"],
            ),
            (dictgen::InsensitiveStr::Ascii("exapmle"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("sparators"), &["separators"]),
            (dictgen::InsensitiveStr::Ascii("faeture"), &["feature"]),
            (dictgen::InsensitiveStr::Ascii("scavange"), &["scavenge"]),
            (dictgen::InsensitiveStr::Ascii("disgustosa"), &["disgusts"]),
            (dictgen::InsensitiveStr::Ascii("totol"), &["total"]),
            (dictgen::InsensitiveStr::Ascii("versalite"), &["versatile"]),
            (dictgen::InsensitiveStr::Ascii("bloking"), &["blocking"]),
            (
                dictgen::InsensitiveStr::Ascii("instruciton"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eliminetaion"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreteate"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("desision"), &["decision"]),
            (
                dictgen::InsensitiveStr::Ascii("compulsivo"),
                &["compulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcifier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("propagaton"),
                &["propagation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutlisite"), &["multisite"]),
            (dictgen::InsensitiveStr::Ascii("damanging"), &["damaging"]),
            (dictgen::InsensitiveStr::Ascii("diallows"), &["disallows"]),
            (dictgen::InsensitiveStr::Ascii("adfter"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("dungoens"), &["dungeons"]),
            (
                dictgen::InsensitiveStr::Ascii("specificially"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganerally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("agressive"), &["aggressive"]),
            (
                dictgen::InsensitiveStr::Ascii("reenfocement"),
                &["reinforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrao"), &["scrap"]),
            (dictgen::InsensitiveStr::Ascii("shelveys"), &["shelves"]),
            (
                dictgen::InsensitiveStr::Ascii("comute"),
                &["commute", "compute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("translationg"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("gurantee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("censably"), &["sensibly"]),
            (
                dictgen::InsensitiveStr::Ascii("stylessheets"),
                &["stylesheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuarla"),
                &["continual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compromessi"),
                &["compromises"],
            ),
            (dictgen::InsensitiveStr::Ascii("unselctes"), &["unselects"]),
            (dictgen::InsensitiveStr::Ascii("suprises"), &["surprises"]),
            (
                dictgen::InsensitiveStr::Ascii("customizatoin"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("goddammt"), &["goddammit"]),
            (
                dictgen::InsensitiveStr::Ascii("agnosticisim"),
                &["agnosticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administravie"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeod"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("micoscopes"),
                &["microscopes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rittles"),
                &["rattles", "riddles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginising"),
                &["recognising"],
            ),
            (dictgen::InsensitiveStr::Ascii("wraapp"), &["wrap"]),
            (dictgen::InsensitiveStr::Ascii("visulized"), &["visualized"]),
            (dictgen::InsensitiveStr::Ascii("sanetizes"), &["sanitizes"]),
            (dictgen::InsensitiveStr::Ascii("stirng"), &["string"]),
            (dictgen::InsensitiveStr::Ascii("asssigns"), &["assigns"]),
            (dictgen::InsensitiveStr::Ascii("derivaive"), &["derivative"]),
            (
                dictgen::InsensitiveStr::Ascii("parralellised"),
                &["parallelised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concurence"),
                &["concurrence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornogrpahy"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitraryly"),
                &["arbitrary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nutritiuos"),
                &["nutritious"],
            ),
            (dictgen::InsensitiveStr::Ascii("jeffies"), &["jiffies"]),
            (dictgen::InsensitiveStr::Ascii("appaered"), &["appeared"]),
            (dictgen::InsensitiveStr::Ascii("poeple"), &["people"]),
            (
                dictgen::InsensitiveStr::Ascii("astronomicly"),
                &["astronomical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantion"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spiritualiy"),
                &["spiritually"],
            ),
            (dictgen::InsensitiveStr::Ascii("airator"), &["aerator"]),
            (
                dictgen::InsensitiveStr::Ascii("initiases"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intantiated"),
                &["instantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("availbale"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("excetption"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregatted"),
                &["aggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("dencodings"), &["encodings"]),
            (
                dictgen::InsensitiveStr::Ascii("slooth"),
                &["sleuth", "sloth", "sooth"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dehyrdated"),
                &["dehydrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vareigated"),
                &["variegated"],
            ),
            (dictgen::InsensitiveStr::Ascii("anntena"), &["antenna"]),
            (
                dictgen::InsensitiveStr::Ascii("academicese"),
                &["academics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tanslation"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkepoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("algee"), &["algae"]),
            (dictgen::InsensitiveStr::Ascii("arriviing"), &["arriving"]),
            (
                dictgen::InsensitiveStr::Ascii("marchmallows"),
                &["marshmallows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolodate"),
                &["consolidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("combatans"), &["combatants"]),
            (
                dictgen::InsensitiveStr::Ascii("prioirties"),
                &["priorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("brithday"), &["birthday"]),
            (dictgen::InsensitiveStr::Ascii("spefixy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("rebion"), &["region"]),
            (dictgen::InsensitiveStr::Ascii("systemc"), &["systemic"]),
            (dictgen::InsensitiveStr::Ascii("woudl"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("benevelont"),
                &["benevolent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retunred"),
                &["returned", "retuned"],
            ),
            (dictgen::InsensitiveStr::Ascii("toggeld"), &["toggled"]),
            (
                dictgen::InsensitiveStr::Ascii("consructing"),
                &["constructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preposterus"),
                &["preposterous"],
            ),
            (dictgen::InsensitiveStr::Ascii("rgeards"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("misspelld"), &["misspelled"]),
            (
                dictgen::InsensitiveStr::Ascii("amerliorators"),
                &["ameliorators"],
            ),
            (dictgen::InsensitiveStr::Ascii("veriasion"), &["variation"]),
            (dictgen::InsensitiveStr::Ascii("jurny"), &["journey"]),
            (dictgen::InsensitiveStr::Ascii("ignorring"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("reslution"), &["resolution"]),
            (dictgen::InsensitiveStr::Ascii("attemp"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("molelcules"), &["molecules"]),
            (dictgen::InsensitiveStr::Ascii("resourcees"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("decrmenetd"),
                &["decremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discontiuned"),
                &["discontinued"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstruments"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iplementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("entpoints"), &["endpoints"]),
            (dictgen::InsensitiveStr::Ascii("qiests"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("peloponnes"),
                &["peloponnese", "peloponnesus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuared"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmarised"),
                &["summarised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrectness"),
                &["correctness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deamonizes"),
                &["daemonizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmappend"), &["unmapped"]),
            (
                dictgen::InsensitiveStr::Ascii("validions"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declarion"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("oackage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("deffining"), &["defining"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilitie"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferecnes"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("onef"), &["one"]),
            (dictgen::InsensitiveStr::Ascii("termplates"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("memmoryes"), &["memories"]),
            (
                dictgen::InsensitiveStr::Ascii("transforme"),
                &["transfer", "transformed", "transformer", "transform"],
            ),
            (dictgen::InsensitiveStr::Ascii("reveiwed"), &["reviewed"]),
            (dictgen::InsensitiveStr::Ascii("disconet"), &["disconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("photograpers"),
                &["photographers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cansiter"), &["canister"]),
            (
                dictgen::InsensitiveStr::Ascii("vegitarian"),
                &["vegetarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colleection"),
                &["collection"],
            ),
            (dictgen::InsensitiveStr::Ascii("sketchysex"), &["sketches"]),
            (dictgen::InsensitiveStr::Ascii("comapare"), &["compare"]),
            (dictgen::InsensitiveStr::Ascii("sidweays"), &["sideways"]),
            (dictgen::InsensitiveStr::Ascii("probablay"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("hipocritical"),
                &["hypocritical"],
            ),
            (dictgen::InsensitiveStr::Ascii("speaced"), &["spaced"]),
            (dictgen::InsensitiveStr::Ascii("supplyed"), &["supplied"]),
            (
                dictgen::InsensitiveStr::Ascii("mathematitian"),
                &["mathematician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassians"),
                &["assassin", "assassins"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenicator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("mnay"), &["many"]),
            (dictgen::InsensitiveStr::Ascii("enaled"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("foundary"), &["foundry"]),
            (dictgen::InsensitiveStr::Ascii("contfibs"), &["contribs"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseberger"),
                &["cheeseburger"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcaing"), &["spacing"]),
            (
                dictgen::InsensitiveStr::Ascii("recontruct"),
                &["reconstruct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterfest"),
                &["counterfeit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cominations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deinitalizing"),
                &["deinitializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troublehshoot"),
                &["troubleshoot"],
            ),
            (dictgen::InsensitiveStr::Ascii("scedulers"), &["schedulers"]),
            (dictgen::InsensitiveStr::Ascii("lincolin"), &["lincoln"]),
            (dictgen::InsensitiveStr::Ascii("lication"), &["location"]),
            (
                dictgen::InsensitiveStr::Ascii("histroically"),
                &["historically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relentness"),
                &["relentless"],
            ),
            (dictgen::InsensitiveStr::Ascii("gruops"), &["groups"]),
            (
                dictgen::InsensitiveStr::Ascii("executution"),
                &["execution"],
            ),
            (dictgen::InsensitiveStr::Ascii("classifer"), &["classifier"]),
            (dictgen::InsensitiveStr::Ascii("endponit"), &["endpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("supportare"),
                &["supporters"],
            ),
            (dictgen::InsensitiveStr::Ascii("succedds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("exersice"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("paackaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("fabircates"),
                &["fabricates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirimentations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("jumpimng"), &["jumping"]),
            (
                dictgen::InsensitiveStr::Ascii("especcially"),
                &["especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornograpghy"),
                &["pornography"],
            ),
            (dictgen::InsensitiveStr::Ascii("missplaced"), &["misplaced"]),
            (
                dictgen::InsensitiveStr::Ascii("autimagically"),
                &["automagically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guesss"),
                &["guess", "guesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("westminser"),
                &["westminster"],
            ),
            (dictgen::InsensitiveStr::Ascii("agressor"), &["aggressor"]),
            (
                dictgen::InsensitiveStr::Ascii("assosciations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ariports"), &["airports"]),
            (
                dictgen::InsensitiveStr::Ascii("opprtunity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deacriptors"),
                &["decryptors", "descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("disible"), &["divisible"]),
            (dictgen::InsensitiveStr::Ascii("fursther"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("idenifiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execeptionally"),
                &["exceptionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsable"),
                &["responsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("latset"), &["latest"]),
            (dictgen::InsensitiveStr::Ascii("delfation"), &["deflation"]),
            (dictgen::InsensitiveStr::Ascii("exponant"), &["exponent"]),
            (dictgen::InsensitiveStr::Ascii("breifly"), &["briefly"]),
            (dictgen::InsensitiveStr::Ascii("aovoids"), &["avoids"]),
            (
                dictgen::InsensitiveStr::Ascii("intiliser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspindent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mobilitiy"), &["mobility"]),
            (
                dictgen::InsensitiveStr::Ascii("checheckpoits"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("harrang"), &["harangue"]),
            (
                dictgen::InsensitiveStr::Ascii("insuinating"),
                &["insinuating"],
            ),
            (dictgen::InsensitiveStr::Ascii("nagivated"), &["navigated"]),
            (
                dictgen::InsensitiveStr::Ascii("reminscence"),
                &["reminiscence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("boomark"), &["bookmark"]),
            (
                dictgen::InsensitiveStr::Ascii("othopaedic"),
                &["orthopaedic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selectivos"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("bunbled"), &["bundled"]),
            (dictgen::InsensitiveStr::Ascii("infenro"), &["inferno"]),
            (dictgen::InsensitiveStr::Ascii("donn"), &["done", "don"]),
            (dictgen::InsensitiveStr::Ascii("accceptor"), &["acceptor"]),
            (
                dictgen::InsensitiveStr::Ascii("frational"),
                &["fractional", "rational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excetptional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evangelikal"),
                &["evangelical"],
            ),
            (dictgen::InsensitiveStr::Ascii("ansestors"), &["ancestors"]),
            (dictgen::InsensitiveStr::Ascii("avtive"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("reassocitions"),
                &["reassociations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vailidity"), &["validity"]),
            (dictgen::InsensitiveStr::Ascii("aggrevate"), &["aggravate"]),
            (dictgen::InsensitiveStr::Ascii("wheighted"), &["weighted"]),
            (dictgen::InsensitiveStr::Ascii("thurday"), &["thursday"]),
            (
                dictgen::InsensitiveStr::Ascii("ressurecting"),
                &["resurrecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("senfile"), &["sendfile"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapointment"),
                &["disappointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeturning"), &["returning"]),
            (
                dictgen::InsensitiveStr::Ascii("interrutped"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaniments"),
                &["containment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unstalled"),
                &["installed", "uninstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimtiers"),
                &["delimiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatails"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("lightergas"), &["lighters"]),
            (dictgen::InsensitiveStr::Ascii("deferrred"), &["deferred"]),
            (
                dictgen::InsensitiveStr::Ascii("propultion"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("nativley"), &["natively"]),
            (dictgen::InsensitiveStr::Ascii("aggegated"), &["aggregated"]),
            (dictgen::InsensitiveStr::Ascii("loacal"), &["local"]),
            (dictgen::InsensitiveStr::Ascii("mainainer"), &["maintainer"]),
            (dictgen::InsensitiveStr::Ascii("proceessed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("electronik"), &["election"]),
            (
                dictgen::InsensitiveStr::Ascii("databasse"),
                &["database", "databases"],
            ),
            (dictgen::InsensitiveStr::Ascii("lincences"), &["licences"]),
            (
                dictgen::InsensitiveStr::Ascii("conceivabley"),
                &["conceivably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncially"),
                &["canonically"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicet"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("disresepctful"),
                &["disrespectful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomdations"),
                &["accommodations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisions"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("probbable"), &["probable"]),
            (
                dictgen::InsensitiveStr::Ascii("perservering"),
                &["persevering"],
            ),
            (dictgen::InsensitiveStr::Ascii("shrtcuts"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("dregees"), &["degrees"]),
            (
                dictgen::InsensitiveStr::Ascii("horitzontal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparaent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("reture"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("fialed"),
                &["failed", "dialed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructief"),
                &["constructive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initliazers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pascheurised"),
                &["pasteurised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fontends"),
                &["frontends", "contends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("healt"),
                &["health", "heart", "halt", "heal", "heat"],
            ),
            (dictgen::InsensitiveStr::Ascii("fatiuge"), &["fatigue"]),
            (dictgen::InsensitiveStr::Ascii("briliant"), &["brilliant"]),
            (dictgen::InsensitiveStr::Ascii("belguim"), &["belgium"]),
            (
                dictgen::InsensitiveStr::Ascii("experementel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisssioners"),
                &["commissioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("sadisitc"), &["sadistic"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherize"),
                &["characterize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcutated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pubilc"), &["public"]),
            (
                dictgen::InsensitiveStr::Ascii("consulation"),
                &["consolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanitisin"), &["sanitising"]),
            (dictgen::InsensitiveStr::Ascii("coorect"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("anmespace"), &["namespace"]),
            (
                dictgen::InsensitiveStr::Ascii("unregisting"),
                &["unregistering"],
            ),
            (dictgen::InsensitiveStr::Ascii("schmatic"), &["schematic"]),
            (dictgen::InsensitiveStr::Ascii("usrername"), &["username"]),
            (dictgen::InsensitiveStr::Ascii("mediveal"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("immitates"), &["imitates"]),
            (
                dictgen::InsensitiveStr::Ascii("orgainsation"),
                &["organisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurizing"),
                &["pasteurizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customisatoin"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qeuestioned"),
                &["questioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descirbing"),
                &["describing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierchically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funktional"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("facours"), &["favours"]),
            (dictgen::InsensitiveStr::Ascii("probem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("puplisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("unviersally"),
                &["universally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expectaion"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selctor"),
                &["selector", "sector"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmont"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impelementing"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("trianed"), &["trained"]),
            (
                dictgen::InsensitiveStr::Ascii("definded"),
                &["defined", "defunded"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribye"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("intolerace"),
                &["intolerance"],
            ),
            (dictgen::InsensitiveStr::Ascii("plattform"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("gitatributes"),
                &["gitattributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("happan"), &["happen"]),
            (dictgen::InsensitiveStr::Ascii("suppling"), &["supplying"]),
            (dictgen::InsensitiveStr::Ascii("pentalty"), &["penalty"]),
            (
                dictgen::InsensitiveStr::Ascii("strirngification"),
                &["stringification"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizle"), &["chisel"]),
            (dictgen::InsensitiveStr::Ascii("chesckbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("nuaghty"), &["naughty"]),
            (
                dictgen::InsensitiveStr::Ascii("adverticement"),
                &["advertisement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("culmulate"),
                &["cumulate", "culminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwsers"), &["answers"]),
            (
                dictgen::InsensitiveStr::Ascii("throthling"),
                &["throttling"],
            ),
            (dictgen::InsensitiveStr::Ascii("carryintg"), &["carrying"]),
            (
                dictgen::InsensitiveStr::Ascii("provocatief"),
                &["provocative"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkstum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("protcol"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("deploied"), &["deployed"]),
            (
                dictgen::InsensitiveStr::Ascii("critisizms"),
                &["criticisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrantes"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("decreasin"), &["decreasing"]),
            (dictgen::InsensitiveStr::Ascii("doker"), &["docker"]),
            (
                dictgen::InsensitiveStr::Ascii("horozontal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("texsts"), &["texts"]),
            (
                dictgen::InsensitiveStr::Ascii("massachusets"),
                &["massachusetts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installeert"),
                &["installer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosophiae"),
                &["philosophies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interdependecies"),
                &["interdependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subracting"),
                &["subtracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punissable"),
                &["punishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("aligin"), &["align"]),
            (
                dictgen::InsensitiveStr::Ascii("acessors"),
                &["accessors", "accessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("manfests"), &["manifests"]),
            (dictgen::InsensitiveStr::Ascii("calucate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("stakeboard"),
                &["skateboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("loccked"), &["locked"]),
            (dictgen::InsensitiveStr::Ascii("ket"), &["kept", "key"]),
            (dictgen::InsensitiveStr::Ascii("reallign"), &["realign"]),
            (
                dictgen::InsensitiveStr::Ascii("migitation"),
                &["mitigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticapating"),
                &["anticipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("husbandos"), &["husbands"]),
            (dictgen::InsensitiveStr::Ascii("staration"), &["starvation"]),
            (
                dictgen::InsensitiveStr::Ascii("regulirization"),
                &["regularization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequenses"), &["sequences"]),
            (
                dictgen::InsensitiveStr::Ascii("tranciever"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("recrete"), &["recreate"]),
            (dictgen::InsensitiveStr::Ascii("othrodox"), &["orthodox"]),
            (
                dictgen::InsensitiveStr::Ascii("splite"),
                &["split", "splits", "splice"],
            ),
            (dictgen::InsensitiveStr::Ascii("stards"), &["starts"]),
            (
                dictgen::InsensitiveStr::Ascii("deregisteres"),
                &["deregisters"],
            ),
            (dictgen::InsensitiveStr::Ascii("inidcates"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("dimmensioned"),
                &["dimensioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possiblilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undestandable"),
                &["understandable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommmtted"),
                &["uncommitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("qaulity"), &["quality"]),
            (
                dictgen::InsensitiveStr::Ascii("phylosopher"),
                &["philosopher"],
            ),
            (dictgen::InsensitiveStr::Ascii("crictical"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("spefcifications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("frointer"), &["frontier"]),
            (
                dictgen::InsensitiveStr::Ascii("insentively"),
                &["insensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recondifure"),
                &["reconfigure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistences"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("indciator"), &["indicator"]),
            (dictgen::InsensitiveStr::Ascii("gettin"), &["getting"]),
            (dictgen::InsensitiveStr::Ascii("censible"), &["sensible"]),
            (dictgen::InsensitiveStr::Ascii("authroity"), &["authority"]),
            (dictgen::InsensitiveStr::Ascii("ording"), &["ordering"]),
            (
                dictgen::InsensitiveStr::Ascii("miscommunicaiton"),
                &["miscommunication"],
            ),
            (dictgen::InsensitiveStr::Ascii("whoose"), &["whose"]),
            (dictgen::InsensitiveStr::Ascii("normlize"), &["normalize"]),
            (
                dictgen::InsensitiveStr::Ascii("absolately"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("tnanking"), &["thanking"]),
            (dictgen::InsensitiveStr::Ascii("starngely"), &["strangely"]),
            (
                dictgen::InsensitiveStr::Ascii("quarternion"),
                &["quaternion"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyos"), &["typos"]),
            (dictgen::InsensitiveStr::Ascii("earpluggs"), &["earplugs"]),
            (dictgen::InsensitiveStr::Ascii("attractin"), &["attraction"]),
            (
                dictgen::InsensitiveStr::Ascii("paculierly"),
                &["peculiarly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreeableness"),
                &["agreeableness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorythemically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeueuing"), &["queuing"]),
            (
                dictgen::InsensitiveStr::Ascii("annoations"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("referals"), &["referrals"]),
            (dictgen::InsensitiveStr::Ascii("siez"), &["size", "seize"]),
            (dictgen::InsensitiveStr::Ascii("stuggle"), &["struggle"]),
            (dictgen::InsensitiveStr::Ascii("reconises"), &["recognises"]),
            (dictgen::InsensitiveStr::Ascii("anticpate"), &["anticipate"]),
            (dictgen::InsensitiveStr::Ascii("sequeence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("demandas"), &["demands"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinately"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogenes"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("leaglise"), &["legalise"]),
            (
                dictgen::InsensitiveStr::Ascii("serinomial"),
                &["ceremonial"],
            ),
            (dictgen::InsensitiveStr::Ascii("deoends"), &["depends"]),
            (dictgen::InsensitiveStr::Ascii("exctracted"), &["extracted"]),
            (dictgen::InsensitiveStr::Ascii("reconnct"), &["reconnect"]),
            (dictgen::InsensitiveStr::Ascii("dloating"), &["floating"]),
            (dictgen::InsensitiveStr::Ascii("addeed"), &["added"]),
            (
                dictgen::InsensitiveStr::Ascii("autoconplete"),
                &["autocomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("plausable"), &["plausible"]),
            (dictgen::InsensitiveStr::Ascii("bocome"), &["become"]),
            (
                dictgen::InsensitiveStr::Ascii("conciously"),
                &["consciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoziations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("memmics"), &["mimics"]),
            (dictgen::InsensitiveStr::Ascii("openining"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("workaorund"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("noctrune"), &["nocturne"]),
            (
                dictgen::InsensitiveStr::Ascii("francaises"),
                &["franchises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoitable"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("lannguages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("asserttive"), &["assertive"]),
            (
                dictgen::InsensitiveStr::Ascii("techincians"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ther"),
                &["there", "their", "the", "other"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distuingish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("miserabley"), &["miserably"]),
            (dictgen::InsensitiveStr::Ascii("knonw"), &["known"]),
            (dictgen::InsensitiveStr::Ascii("otherwsie"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("blacony"), &["balcony"]),
            (
                dictgen::InsensitiveStr::Ascii("whatepsace"),
                &["whitespace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproduciblility"),
                &["reproducibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculuating"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("evailable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("expirimentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constantins"),
                &["constants"],
            ),
            (dictgen::InsensitiveStr::Ascii("digtal"), &["digital"]),
            (
                dictgen::InsensitiveStr::Ascii("intruction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocaes"), &["allocates"]),
            (
                dictgen::InsensitiveStr::Ascii("desperatly"),
                &["desperately"],
            ),
            (dictgen::InsensitiveStr::Ascii("sctipting"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("musuclar"), &["muscular"]),
            (dictgen::InsensitiveStr::Ascii("itearates"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("cosists"), &["consists"]),
            (dictgen::InsensitiveStr::Ascii("cylnder"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("exlainers"), &["explainers"]),
            (
                dictgen::InsensitiveStr::Ascii("capitilists"),
                &["capitalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strangeshit"),
                &["strangest"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahcing"), &["caching"]),
            (
                dictgen::InsensitiveStr::Ascii("implimentation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funxtionality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("staitc"), &["static"]),
            (dictgen::InsensitiveStr::Ascii("lifetimers"), &["lifetimes"]),
            (dictgen::InsensitiveStr::Ascii("audibel"), &["audible"]),
            (dictgen::InsensitiveStr::Ascii("extermist"), &["extremist"]),
            (
                dictgen::InsensitiveStr::Ascii("abandomnent"),
                &["abandonment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aqcuaintance"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("listbbox"), &["listbox"]),
            (
                dictgen::InsensitiveStr::Ascii("excellement"),
                &["excellent"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheronbyl"), &["chernobyl"]),
            (dictgen::InsensitiveStr::Ascii("mentallity"), &["mentally"]),
            (dictgen::InsensitiveStr::Ascii("advocats"), &["advocates"]),
            (dictgen::InsensitiveStr::Ascii("uodate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("hystericlly"),
                &["hysterically"],
            ),
            (dictgen::InsensitiveStr::Ascii("thorats"), &["throats"]),
            (
                dictgen::InsensitiveStr::Ascii("proplematic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disaterous"),
                &["disastrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pplications"),
                &["applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervento"),
                &["intervention"],
            ),
            (dictgen::InsensitiveStr::Ascii("carnagie"), &["carnegie"]),
            (
                dictgen::InsensitiveStr::Ascii("acknwoledges"),
                &["acknowledges"],
            ),
            (dictgen::InsensitiveStr::Ascii("challange"), &["challenge"]),
            (
                dictgen::InsensitiveStr::Ascii("dissymmectry"),
                &["dissymmetry"],
            ),
            (dictgen::InsensitiveStr::Ascii("serive"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("sepecific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("cononically"),
                &["canonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exluded"),
                &["excluded", "exuded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformus"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceppter"), &["accepter"]),
            (dictgen::InsensitiveStr::Ascii("classrom"), &["classroom"]),
            (
                dictgen::InsensitiveStr::Ascii("documantary"),
                &["documentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("weappon"), &["weapon"]),
            (
                dictgen::InsensitiveStr::Ascii("mutiplicative"),
                &["multiplicative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terimnation"),
                &["termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaions"), &["animations"]),
            (
                dictgen::InsensitiveStr::Ascii("refriderator"),
                &["refrigerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deactiving"),
                &["deactivating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thorugh"),
                &["through", "thorough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilpliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfilps"), &["unflips"]),
            (dictgen::InsensitiveStr::Ascii("conosuer"), &["connoisseur"]),
            (
                dictgen::InsensitiveStr::Ascii("currespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indiginous"),
                &["indigenous"],
            ),
            (dictgen::InsensitiveStr::Ascii("allergey"), &["allergy"]),
            (
                dictgen::InsensitiveStr::Ascii("agriculteral"),
                &["agricultural"],
            ),
            (dictgen::InsensitiveStr::Ascii("graudally"), &["gradually"]),
            (
                dictgen::InsensitiveStr::Ascii("therapeutisch"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trotzkists"),
                &["trotskyists"],
            ),
            (dictgen::InsensitiveStr::Ascii("aronud"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("trivival"), &["trivial"]),
            (dictgen::InsensitiveStr::Ascii("booe"), &["buoy"]),
            (dictgen::InsensitiveStr::Ascii("coercian"), &["coercion"]),
            (
                dictgen::InsensitiveStr::Ascii("catholicsim"),
                &["catholicism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consrtucted"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("maunals"), &["manuals"]),
            (dictgen::InsensitiveStr::Ascii("leageue"), &["league"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretatin"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standrdising"),
                &["standardising"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotrolls"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("sanbbox"), &["sandbox"]),
            (dictgen::InsensitiveStr::Ascii("evokin"), &["evoking"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mataphorically"),
                &["metaphorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilidging"),
                &["privileging"],
            ),
            (dictgen::InsensitiveStr::Ascii("advencing"), &["advancing"]),
            (dictgen::InsensitiveStr::Ascii("headng"), &["heading"]),
            (
                dictgen::InsensitiveStr::Ascii("frustations"),
                &["frustrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofetn"), &["often"]),
            (dictgen::InsensitiveStr::Ascii("softwade"), &["software"]),
            (dictgen::InsensitiveStr::Ascii("accointed"), &["accounted"]),
            (dictgen::InsensitiveStr::Ascii("baasics"), &["basics"]),
            (dictgen::InsensitiveStr::Ascii("abvious"), &["obvious"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtorities"),
                &["authorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("succint"), &["succinct"]),
            (
                dictgen::InsensitiveStr::Ascii("ressemblence"),
                &["resemblance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beanchmarked"),
                &["benchmarked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invoved"),
                &["involved", "invoked"],
            ),
            (dictgen::InsensitiveStr::Ascii("delarator"), &["declarator"]),
            (dictgen::InsensitiveStr::Ascii("cursade"), &["crusade"]),
            (dictgen::InsensitiveStr::Ascii("itselfs"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("flippade"), &["flipped"]),
            (
                dictgen::InsensitiveStr::Ascii("discograhy"),
                &["discography"],
            ),
            (dictgen::InsensitiveStr::Ascii("intially"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("resoourced"), &["resourced"]),
            (dictgen::InsensitiveStr::Ascii("fature"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("complimation"),
                &["complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("precison"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("shouod"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("bracelona"), &["barcelona"]),
            (dictgen::InsensitiveStr::Ascii("simialr"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("sentaors"), &["senators"]),
            (
                dictgen::InsensitiveStr::Ascii("preminence"),
                &["preeminence"],
            ),
            (dictgen::InsensitiveStr::Ascii("pilgirm"), &["pilgrim"]),
            (
                dictgen::InsensitiveStr::Ascii("illustrant"),
                &["illustrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("brighest"), &["brightest"]),
            (dictgen::InsensitiveStr::Ascii("vibratoare"), &["vibrator"]),
            (dictgen::InsensitiveStr::Ascii("appologies"), &["apologies"]),
            (
                dictgen::InsensitiveStr::Ascii("comisioned"),
                &["commissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelearte"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ned"),
                &["need", "end", "nod"],
            ),
            (dictgen::InsensitiveStr::Ascii("targetted"), &["targeted"]),
            (
                dictgen::InsensitiveStr::Ascii("enterpreuners"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("claraty"), &["clarity"]),
            (
                dictgen::InsensitiveStr::Ascii("overlflowing"),
                &["overflowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sise"), &["size", "sisal"]),
            (
                dictgen::InsensitiveStr::Ascii("industiral"),
                &["industrial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanizims"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administatively"),
                &["administratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("touristas"), &["tourists"]),
            (dictgen::InsensitiveStr::Ascii("actuaally"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("summones"), &["summoners"]),
            (dictgen::InsensitiveStr::Ascii("qauntity"), &["quantity"]),
            (dictgen::InsensitiveStr::Ascii("renforce"), &["reinforce"]),
            (
                dictgen::InsensitiveStr::Ascii("standarise"),
                &["standardise"],
            ),
            (dictgen::InsensitiveStr::Ascii("searchin"), &["searching"]),
            (
                dictgen::InsensitiveStr::Ascii("regulacion"),
                &["regulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("filterig"), &["filtering"]),
            (dictgen::InsensitiveStr::Ascii("symmmetric"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("islamistas"), &["islamists"]),
            (
                dictgen::InsensitiveStr::Ascii("officianados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("thaat"), &["that"]),
            (
                dictgen::InsensitiveStr::Ascii("transparnt"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("derrives"), &["derives"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalsim"),
                &["capitalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("snaphsots"), &["snapshots"]),
            (dictgen::InsensitiveStr::Ascii("ceratinly"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("disclosin"), &["disclosing"]),
            (
                dictgen::InsensitiveStr::Ascii("physioligy"),
                &["physiology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregats"),
                &["aggregate", "aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vailidation"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("troble"), &["trouble"]),
            (
                dictgen::InsensitiveStr::Ascii("afterthougt"),
                &["afterthought"],
            ),
            (dictgen::InsensitiveStr::Ascii("auxilliary"), &["auxiliary"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplentations"),
                &["reimplementations", "reimplantations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pn"), &["on"]),
            (dictgen::InsensitiveStr::Ascii("persain"), &["persian"]),
            (dictgen::InsensitiveStr::Ascii("sldiers"), &["sliders"]),
            (dictgen::InsensitiveStr::Ascii("simley"), &["smiley"]),
            (
                dictgen::InsensitiveStr::Ascii("humilitaion"),
                &["humiliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradtion"),
                &["tradition", "traction"],
            ),
            (dictgen::InsensitiveStr::Ascii("commencin"), &["commencing"]),
            (
                dictgen::InsensitiveStr::Ascii("gratuitious"),
                &["gratuitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustrare"),
                &["illustrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("botstrapped"),
                &["bootstrapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("specfiier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("biult"), &["built", "build"]),
            (
                dictgen::InsensitiveStr::Ascii("principaly"),
                &["principality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scences"),
                &["scenes", "sciences", "senses", "census"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrictifs"),
                &["restricts"],
            ),
            (dictgen::InsensitiveStr::Ascii("startgin"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("compiations"),
                &["compilations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complaince"),
                &["compliance", "complaints"],
            ),
            (dictgen::InsensitiveStr::Ascii("dtaa"), &["data"]),
            (
                dictgen::InsensitiveStr::Ascii("vehical"),
                &["vehicle", "vesical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppresion"),
                &["suppression"],
            ),
            (dictgen::InsensitiveStr::Ascii("assgined"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("doocumented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contigouous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visulizing"),
                &["visualizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("voluntered"),
                &["volunteered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hiarchically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspongding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooming"),
                &["coming", "cooling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symptumaticaly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("syllabe"), &["syllable"]),
            (dictgen::InsensitiveStr::Ascii("endopint"), &["endpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchrnously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("reivews"), &["reviews"]),
            (
                dictgen::InsensitiveStr::Ascii("conparator"),
                &["comparator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sheme"),
                &["scheme", "shame"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulitiple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("perpared"), &["prepared"]),
            (
                dictgen::InsensitiveStr::Ascii("susbtantive"),
                &["substantive"],
            ),
            (dictgen::InsensitiveStr::Ascii("universtal"), &["universal"]),
            (
                dictgen::InsensitiveStr::Ascii("embarrasing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reegions"), &["regions"]),
            (dictgen::InsensitiveStr::Ascii("respoted"), &["reposted"]),
            (dictgen::InsensitiveStr::Ascii("grainte"), &["granite"]),
            (
                dictgen::InsensitiveStr::Ascii("attruibutes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oint"),
                &["point", "pint", "joint", "lint"],
            ),
            (dictgen::InsensitiveStr::Ascii("downoload"), &["download"]),
            (dictgen::InsensitiveStr::Ascii("emebedded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("folwoer"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("represnt"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("unechecked"), &["unchecked"]),
            (dictgen::InsensitiveStr::Ascii("paralelle"), &["parallel"]),
            (
                dictgen::InsensitiveStr::Ascii("protoganists"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("lienups"), &["lineups"]),
            (
                dictgen::InsensitiveStr::Ascii("incramental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permuatated"),
                &["permutated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hierachrical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmaticaly"),
                &["programmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("skup"), &["skip"]),
            (dictgen::InsensitiveStr::Ascii("copywrite"), &["copyright"]),
            (
                dictgen::InsensitiveStr::Ascii("decendentants"),
                &["descendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heirarchically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("precdence"), &["precedence"]),
            (dictgen::InsensitiveStr::Ascii("fonud"), &["found"]),
            (
                dictgen::InsensitiveStr::Ascii("cooperativo"),
                &["cooperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimph"), &["nymph"]),
            (dictgen::InsensitiveStr::Ascii("btyes"), &["bytes"]),
            (dictgen::InsensitiveStr::Ascii("chaings"), &["chains"]),
            (
                dictgen::InsensitiveStr::Ascii("intermediare"),
                &["intermediate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecific"),
                &["specific", "pacific"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lowd"),
                &["load", "low", "loud"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operatin"),
                &["operating", "operation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forbidded"),
                &["forbidden", "forbade"],
            ),
            (dictgen::InsensitiveStr::Ascii("movememts"), &["movements"]),
            (dictgen::InsensitiveStr::Ascii("anaylsises"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscirbe"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imporvised"),
                &["improvised"],
            ),
            (dictgen::InsensitiveStr::Ascii("platofmrs"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("cose"), &["case"]),
            (dictgen::InsensitiveStr::Ascii("ammoungst"), &["amongst"]),
            (
                dictgen::InsensitiveStr::Ascii("posiblities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("purposses"), &["purposes"]),
            (dictgen::InsensitiveStr::Ascii("travle"), &["travel"]),
            (
                dictgen::InsensitiveStr::Ascii("porgrammeers"),
                &["programmers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconecction"),
                &["reconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compications"),
                &["complications"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanldle"), &["handle"]),
            (
                dictgen::InsensitiveStr::Ascii("equavalent"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("cappacity"), &["capacity"]),
            (dictgen::InsensitiveStr::Ascii("seahakws"), &["seahawks"]),
            (dictgen::InsensitiveStr::Ascii("compleated"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("tettrahedral"),
                &["tetrahedral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kubenernetes"),
                &["kubernetes"],
            ),
            (dictgen::InsensitiveStr::Ascii("couter"), &["counter"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsitration"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("realtives"), &["relatives"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalits"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("plastis"), &["plastics"]),
            (dictgen::InsensitiveStr::Ascii("northen"), &["northern"]),
            (dictgen::InsensitiveStr::Ascii("arhtrosis"), &["arthrosis"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprise"),
                &["enterprise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncontrolable"),
                &["uncontrollable"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacine"), &["vaccine"]),
            (dictgen::InsensitiveStr::Ascii("sparklie"), &["sparkle"]),
            (dictgen::InsensitiveStr::Ascii("shoulld"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridinated"),
                &["coordinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orchistrated"),
                &["orchestrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("michgian"), &["michigan"]),
            (
                dictgen::InsensitiveStr::Ascii("granularty"),
                &["granularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stromfront"),
                &["stormfront"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negligiable"),
                &["negligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("makse"), &["makes", "masks"]),
            (dictgen::InsensitiveStr::Ascii("bowkay"), &["bouquet"]),
            (
                dictgen::InsensitiveStr::Ascii("represantation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinfornced"),
                &["reinforced"],
            ),
            (dictgen::InsensitiveStr::Ascii("thingey"), &["thingy"]),
            (
                dictgen::InsensitiveStr::Ascii("camouflagued"),
                &["camouflage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disabilites"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inappropirate"),
                &["inappropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singooled"),
                &["singled", "signaled"],
            ),
            (dictgen::InsensitiveStr::Ascii("decrpt"), &["decrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("dangeroys"),
                &["dangerously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiations"),
                &["initiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("thickess"), &["thickness"]),
            (dictgen::InsensitiveStr::Ascii("eachohter"), &["eachother"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubcribes"),
                &["unsubscribes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mastubrate"),
                &["masturbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aproching"),
                &["approaching"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchagned"), &["exchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("infomraiton"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorcers"),
                &["endorsers", "enforcers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsepcified"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalidatiopn"),
                &["invalidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rquested"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("consired"),
                &["conspired", "considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgustigly"),
                &["disgustingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyis"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("primtives"), &["primitives"]),
            (dictgen::InsensitiveStr::Ascii("gingam"), &["gingham"]),
            (dictgen::InsensitiveStr::Ascii("calcurate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("releting"),
                &["deleting", "relating"],
            ),
            (dictgen::InsensitiveStr::Ascii("earthboud"), &["earthbound"]),
            (
                dictgen::InsensitiveStr::Ascii("counsiling"),
                &["counseling"],
            ),
            (dictgen::InsensitiveStr::Ascii("symetric"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("attacment"), &["attachment"]),
            (dictgen::InsensitiveStr::Ascii("dissrupts"), &["disrupts"]),
            (
                dictgen::InsensitiveStr::Ascii("sugguesting"),
                &["suggesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("policiy"), &["policy"]),
            (
                dictgen::InsensitiveStr::Ascii("calbiration"),
                &["calibration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggration"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("beehtoven"), &["beethoven"]),
            (dictgen::InsensitiveStr::Ascii("annyoance"), &["annoyance"]),
            (
                dictgen::InsensitiveStr::Ascii("remenisser"),
                &["reminiscer"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("cyrstallise"),
                &["crystallise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achiving"),
                &["achieving", "archiving"],
            ),
            (dictgen::InsensitiveStr::Ascii("analizers"), &["analyzers"]),
            (
                dictgen::InsensitiveStr::Ascii("inconveniet"),
                &["inconvenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independendet"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("genrated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("vegitarien"),
                &["vegetarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrobing"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("astrixs"), &["asterisks"]),
            (dictgen::InsensitiveStr::Ascii("pandoria"), &["pandora"]),
            (
                dictgen::InsensitiveStr::Ascii("caculaters"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("pached"), &["patched"]),
            (
                dictgen::InsensitiveStr::Ascii("abriviates"),
                &["abbreviates"],
            ),
            (dictgen::InsensitiveStr::Ascii("auhtoring"), &["authoring"]),
            (dictgen::InsensitiveStr::Ascii("entring"), &["entering"]),
            (dictgen::InsensitiveStr::Ascii("advandces"), &["advances"]),
            (
                dictgen::InsensitiveStr::Ascii("establsihment"),
                &["establishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termniator"),
                &["terminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("specialy"), &["specially"]),
            (
                dictgen::InsensitiveStr::Ascii("transfoming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governement"),
                &["government"],
            ),
            (dictgen::InsensitiveStr::Ascii("transprts"), &["transports"]),
            (
                dictgen::InsensitiveStr::Ascii("hypenating"),
                &["hyphenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("themselces"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("micoscope"), &["microscope"]),
            (dictgen::InsensitiveStr::Ascii("bofy"), &["body"]),
            (
                dictgen::InsensitiveStr::Ascii("acknolwedged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("glorios"), &["glorious"]),
            (dictgen::InsensitiveStr::Ascii("incpetion"), &["inception"]),
            (dictgen::InsensitiveStr::Ascii("cotainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("statastics"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("opinyonal"), &["opinional"]),
            (dictgen::InsensitiveStr::Ascii("slq"), &["sql"]),
            (dictgen::InsensitiveStr::Ascii("urface"), &["surface"]),
            (
                dictgen::InsensitiveStr::Ascii("achevement"),
                &["achievement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mississppi"),
                &["mississippi"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experminentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthentication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependens"),
                &["depends", "dependents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reparametrization"),
                &["reparameterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accideny"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("evry"), &["every"]),
            (dictgen::InsensitiveStr::Ascii("ecounter"), &["encounter"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhitms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standerdized"),
                &["standardized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrespondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("insertes"), &["inserts"]),
            (dictgen::InsensitiveStr::Ascii("stimulli"), &["stimuli"]),
            (dictgen::InsensitiveStr::Ascii("coalsece"), &["coalesce"]),
            (dictgen::InsensitiveStr::Ascii("scokets"), &["sockets"]),
            (dictgen::InsensitiveStr::Ascii("nonsens"), &["nonsense"]),
            (dictgen::InsensitiveStr::Ascii("inpupt"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("insecue"), &["insecure"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactueres"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("councilos"), &["councils"]),
            (dictgen::InsensitiveStr::Ascii("fabricas"), &["fabrics"]),
            (dictgen::InsensitiveStr::Ascii("wihout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("togles"), &["toggles"]),
            (
                dictgen::InsensitiveStr::Ascii("ncluding"),
                &["including", "concluding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerablle"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generaotrs"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("poduction"), &["production"]),
            (
                dictgen::InsensitiveStr::Ascii("intereferes"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialsations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("combainer"), &["combiner"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepreneurers"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provileged"),
                &["privileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprevation"),
                &["deprivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wellfare"), &["welfare"]),
            (
                dictgen::InsensitiveStr::Ascii("recommands"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticials"), &["criticals"]),
            (dictgen::InsensitiveStr::Ascii("cenarios"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("startet"), &["started"]),
            (dictgen::InsensitiveStr::Ascii("ceasers"), &["caesars"]),
            (dictgen::InsensitiveStr::Ascii("precurser"), &["precursor"]),
            (
                dictgen::InsensitiveStr::Ascii("dysfunktion"),
                &["dysfunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annihilatron"),
                &["annihilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("consle"), &["console"]),
            (
                dictgen::InsensitiveStr::Ascii("approrpiation"),
                &["appropriation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvoteers"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakcet"), &["packet"]),
            (
                dictgen::InsensitiveStr::Ascii("overcapping"),
                &["overlapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("tood"), &["todo"]),
            (dictgen::InsensitiveStr::Ascii("estiamates"), &["estimates"]),
            (dictgen::InsensitiveStr::Ascii("sarinomys"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("terminaron"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corproations"),
                &["corporations"],
            ),
            (dictgen::InsensitiveStr::Ascii("serice"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("dfines"), &["defines"]),
            (dictgen::InsensitiveStr::Ascii("uise"), &["use"]),
            (
                dictgen::InsensitiveStr::Ascii("unsusbcribed"),
                &["unsubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nanosecods"),
                &["nanoseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parenthesed"),
                &["parenthesized"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuplaod"), &["reupload"]),
            (dictgen::InsensitiveStr::Ascii("comapring"), &["comparing"]),
            (
                dictgen::InsensitiveStr::Ascii("concaneted"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benefitial"),
                &["beneficial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliasations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("autasave"), &["autosave"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoziable"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmached"), &["unmatched"]),
            (dictgen::InsensitiveStr::Ascii("practives"), &["practise"]),
            (dictgen::InsensitiveStr::Ascii("exsitence"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("modefying"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("skeptecism"),
                &["skepticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimentory"),
                &["complimentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("generize"), &["generalize"]),
            (
                dictgen::InsensitiveStr::Ascii("indestructibe"),
                &["indestructible"],
            ),
            (dictgen::InsensitiveStr::Ascii("fiancial"), &["financial"]),
            (dictgen::InsensitiveStr::Ascii("moteurs"), &["motors"]),
            (
                dictgen::InsensitiveStr::Ascii("exhuastion"),
                &["exhaustion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gastly"),
                &["ghastly", "vastly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ortogonalization"),
                &["orthogonalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sendetary"), &["sedentary"]),
            (
                dictgen::InsensitiveStr::Ascii("imperically"),
                &["empirically"],
            ),
            (dictgen::InsensitiveStr::Ascii("daulity"), &["duality"]),
            (
                dictgen::InsensitiveStr::Ascii("coalasece"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retramsmitted"),
                &["retransmitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpontent"), &["content"]),
            (dictgen::InsensitiveStr::Ascii("databses"), &["databases"]),
            (dictgen::InsensitiveStr::Ascii("intiale"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("indivudals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimpliments"),
                &["reimplements"],
            ),
            (dictgen::InsensitiveStr::Ascii("spetsified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("conlcuded"), &["concluded"]),
            (dictgen::InsensitiveStr::Ascii("spezial"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("othersise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("namemespace"),
                &["namespace"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaiilable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("sratch"),
                &["scratch", "snatch", "starch"],
            ),
            (dictgen::InsensitiveStr::Ascii("funchtionn"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("heathern"), &["heathen"]),
            (
                dictgen::InsensitiveStr::Ascii("dstribution"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountent"),
                &["accountant"],
            ),
            (dictgen::InsensitiveStr::Ascii("overivew"), &["overview"]),
            (
                dictgen::InsensitiveStr::Ascii("rethorical"),
                &["rhetorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("leasurely"), &["leisurely"]),
            (dictgen::InsensitiveStr::Ascii("restrcted"), &["restricted"]),
            (dictgen::InsensitiveStr::Ascii("polotical"), &["political"]),
            (
                dictgen::InsensitiveStr::Ascii("breating"),
                &["beating", "breathing"],
            ),
            (dictgen::InsensitiveStr::Ascii("withing"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("transciprt"),
                &["transcripts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentemental"),
                &["sentimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reactiony"),
                &["reactionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("toggleing"), &["toggling"]),
            (dictgen::InsensitiveStr::Ascii("ellegant"), &["elegant"]),
            (
                dictgen::InsensitiveStr::Ascii("inflitrate"),
                &["infiltrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("stamnia"), &["stamina"]),
            (
                dictgen::InsensitiveStr::Ascii("alcohalism"),
                &["alcoholism"],
            ),
            (dictgen::InsensitiveStr::Ascii("mandrian"), &["mandarin"]),
            (dictgen::InsensitiveStr::Ascii("priorise"), &["prioritise"]),
            (dictgen::InsensitiveStr::Ascii("capela"), &["capella"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibiities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unimplemeneted"),
                &["unimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("resistece"), &["resistance"]),
            (dictgen::InsensitiveStr::Ascii("vallgrind"), &["valgrind"]),
            (
                dictgen::InsensitiveStr::Ascii("breackthrough"),
                &["breakthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithemically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailzer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("monotired"), &["monitored"]),
            (
                dictgen::InsensitiveStr::Ascii("individuallity"),
                &["individuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beliefe"),
                &["believe", "belief"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missunderstandings"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("psitoins"), &["positions"]),
            (
                dictgen::InsensitiveStr::Ascii("volunteed"),
                &["volunteered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementantion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gauntelt"), &["gauntlet"]),
            (dictgen::InsensitiveStr::Ascii("resons"), &["reasons"]),
            (
                dictgen::InsensitiveStr::Ascii("pascheurize"),
                &["pasteurize"],
            ),
            (dictgen::InsensitiveStr::Ascii("imaptient"), &["impatient"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitaed"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponts"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("locgical"), &["logical"]),
            (dictgen::InsensitiveStr::Ascii("complais"), &["complains"]),
            (
                dictgen::InsensitiveStr::Ascii("approxamate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("portras"), &["portrays"]),
            (
                dictgen::InsensitiveStr::Ascii("techincality"),
                &["technicality"],
            ),
            (dictgen::InsensitiveStr::Ascii("limk"), &["link"]),
            (
                dictgen::InsensitiveStr::Ascii("behvaioral"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stanardises"),
                &["standardises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executings"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakistain"), &["pakistani"]),
            (dictgen::InsensitiveStr::Ascii("inherts"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("bisct"), &["bisect"]),
            (dictgen::InsensitiveStr::Ascii("stubbron"), &["stubborn"]),
            (dictgen::InsensitiveStr::Ascii("depenedent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeck"),
                &["check", "cheek"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomplie"),
                &["recompile", "recomply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervining"),
                &["intervening"],
            ),
            (dictgen::InsensitiveStr::Ascii("igonoring"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthoughs"),
                &["breakthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similarites"),
                &["similarities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wavelenght"),
                &["wavelength"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unappropriately"),
                &["inappropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conntectors"),
                &["connectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaiations"), &["variations"]),
            (
                dictgen::InsensitiveStr::Ascii("canniballism"),
                &["cannibalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consisteny"),
                &["consistency", "consistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invulnerabe"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depedently"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prokrastination"),
                &["procrastination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplemention"),
                &["reimplementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyou"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("mimimizing"),
                &["minimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bresh"), &["brush", "fresh"]),
            (
                dictgen::InsensitiveStr::Ascii("xenophopic"),
                &["xenophobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetainly"), &["certainly"]),
            (
                dictgen::InsensitiveStr::Ascii("bastractor"),
                &["abstractor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insentivises"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experamentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferentiation"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overhwelmingly"),
                &["overwhelmingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sybol"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("peizometer"),
                &["piezometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("shutted"), &["shut"]),
            (dictgen::InsensitiveStr::Ascii("catholicus"), &["catholics"]),
            (
                dictgen::InsensitiveStr::Ascii("nclusions"),
                &["inclusions", "conclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempraal"), &["temporal"]),
            (
                dictgen::InsensitiveStr::Ascii("abstrakted"),
                &["abstracted"],
            ),
            (dictgen::InsensitiveStr::Ascii("buufers"), &["buffers"]),
            (dictgen::InsensitiveStr::Ascii("annoyying"), &["annoying"]),
            (
                dictgen::InsensitiveStr::Ascii("larges"),
                &["largest", "large"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrebuted"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("zionistm"), &["zionism"]),
            (
                dictgen::InsensitiveStr::Ascii("resignement"),
                &["resignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("vituals"), &["visuals"]),
            (
                dictgen::InsensitiveStr::Ascii("confifuration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("castlers"), &["castles"]),
            (dictgen::InsensitiveStr::Ascii("esclusion"), &["exclusion"]),
            (
                dictgen::InsensitiveStr::Ascii("recognizin"),
                &["recognizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("descedent"), &["descendent"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsitrations"),
                &["administrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("requirs"), &["requires"]),
            (dictgen::InsensitiveStr::Ascii("eleent"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("emenets"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("exprienced"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("memor"), &["memory"]),
            (
                dictgen::InsensitiveStr::Ascii("ciruclating"),
                &["circulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("supposely"), &["supposedly"]),
            (
                dictgen::InsensitiveStr::Ascii("confurmations"),
                &["confirmations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbcripts"),
                &["subscripts"],
            ),
            (dictgen::InsensitiveStr::Ascii("signol"), &["signal"]),
            (dictgen::InsensitiveStr::Ascii("comptues"), &["computes"]),
            (dictgen::InsensitiveStr::Ascii("sematic"), &["semantic"]),
            (
                dictgen::InsensitiveStr::Ascii("thinked"),
                &["thunked", "thought"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favilitates"),
                &["facilitates"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoized"), &["customized"]),
            (dictgen::InsensitiveStr::Ascii("perisan"), &["persian"]),
            (dictgen::InsensitiveStr::Ascii("noifying"), &["notifying"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlations"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bufferd"), &["buffered"]),
            (dictgen::InsensitiveStr::Ascii("vitenam"), &["vietnam"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatoion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hootsba"),
                &["hutzpa", "chutzpah"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonsencial"),
                &["nonsensical"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliboard"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressurrect"), &["resurrect"]),
            (dictgen::InsensitiveStr::Ascii("forenisc"), &["forensic"]),
            (dictgen::InsensitiveStr::Ascii("animaor"), &["animator"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("recruse"), &["recurse"]),
            (dictgen::InsensitiveStr::Ascii("conecct"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("besteality"),
                &["bestiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dirrectives"),
                &["directives"],
            ),
            (dictgen::InsensitiveStr::Ascii("apllied"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("sofmores"), &["sophomores"]),
            (dictgen::InsensitiveStr::Ascii("analyise"), &["analyse"]),
            (dictgen::InsensitiveStr::Ascii("otherwaise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("artument"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("sideboad"), &["sideboard"]),
            (
                dictgen::InsensitiveStr::Ascii("construciton"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continutity"),
                &["continuity"],
            ),
            (dictgen::InsensitiveStr::Ascii("empasis"), &["emphasis"]),
            (
                dictgen::InsensitiveStr::Ascii("mischevious"),
                &["mischievous"],
            ),
            (dictgen::InsensitiveStr::Ascii("latenciy"), &["latency"]),
            (dictgen::InsensitiveStr::Ascii("regrest"), &["regrets"]),
            (
                dictgen::InsensitiveStr::Ascii("interational"),
                &["international"],
            ),
            (dictgen::InsensitiveStr::Ascii("roomate"), &["roommate"]),
            (dictgen::InsensitiveStr::Ascii("meausure"), &["measure"]),
            (dictgen::InsensitiveStr::Ascii("netwrok"), &["network"]),
            (
                dictgen::InsensitiveStr::Ascii("convaluted"),
                &["convoluted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clasification"),
                &["clarification", "classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("eles"), &["eels", "else"]),
            (dictgen::InsensitiveStr::Ascii("suystemic"), &["systemic"]),
            (
                dictgen::InsensitiveStr::Ascii("guaruantes"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architexts"),
                &["architects"],
            ),
            (dictgen::InsensitiveStr::Ascii("apostrope"), &["apostrophe"]),
            (
                dictgen::InsensitiveStr::Ascii("consumersim"),
                &["consumerism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approprated"),
                &["appropriated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particpation"),
                &["participation"],
            ),
            (dictgen::InsensitiveStr::Ascii("patinetly"), &["patiently"]),
            (dictgen::InsensitiveStr::Ascii("reminscer"), &["reminiscer"]),
            (
                dictgen::InsensitiveStr::Ascii("classe"),
                &["class", "classes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ristriction"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimatley"),
                &["legitimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitiary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficeints"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("evoluate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("spartaniis"), &["spartans"]),
            (
                dictgen::InsensitiveStr::Ascii("incoproration"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reworkd"), &["reworked"]),
            (
                dictgen::InsensitiveStr::Ascii("distirbuting"),
                &["distributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prgramming"),
                &["programming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informativenss"),
                &["informativeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("aloowed"), &["allowed"]),
            (
                dictgen::InsensitiveStr::Ascii("peridiocally"),
                &["periodically"],
            ),
            (dictgen::InsensitiveStr::Ascii("leaset"), &["least"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatiblities"),
                &["incompatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("shetler"), &["shelter"]),
            (dictgen::InsensitiveStr::Ascii("graphci"), &["graphic"]),
            (dictgen::InsensitiveStr::Ascii("eczecutive"), &["executive"]),
            (dictgen::InsensitiveStr::Ascii("sorcerey"), &["sorcery"]),
            (
                dictgen::InsensitiveStr::Ascii("consoldiate"),
                &["consolidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("starded"), &["started"]),
            (dictgen::InsensitiveStr::Ascii("botostrap"), &["bootstrap"]),
            (
                dictgen::InsensitiveStr::Ascii("operationes"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("olymipcs"), &["olympics"]),
            (
                dictgen::InsensitiveStr::Ascii("correspionds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefenrece"),
                &["dereference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stronkhold"),
                &["stronghold"],
            ),
            (dictgen::InsensitiveStr::Ascii("thirstay"), &["thirsty"]),
            (
                dictgen::InsensitiveStr::Ascii("temporily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrese"), &["decrease"]),
            (
                dictgen::InsensitiveStr::Ascii("downvotess"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simltaneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commercail"),
                &["commercials"],
            ),
            (dictgen::InsensitiveStr::Ascii("meybe"), &["maybe"]),
            (
                dictgen::InsensitiveStr::Ascii("occasinally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("orientarla"), &["oriental"]),
            (
                dictgen::InsensitiveStr::Ascii("coreespondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persuasing"),
                &["persuasion"],
            ),
            (dictgen::InsensitiveStr::Ascii("thresold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("thiestic"), &["theistic"]),
            (dictgen::InsensitiveStr::Ascii("rembembers"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("glyped"), &["glyphed"]),
            (
                dictgen::InsensitiveStr::Ascii("southamption"),
                &["southampton"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contet"),
                &["contest", "content", "context"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitiablity"),
                &["profitability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisfying"),
                &["satisfying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intergrated"),
                &["integrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progessive"),
                &["progressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bodywieght"),
                &["bodyweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("situtations"),
                &["situations"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearios"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("stereotipe"),
                &["stereotype"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kombinations"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("observerd"), &["observed"]),
            (
                dictgen::InsensitiveStr::Ascii("accumualte"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vietnameese"),
                &["vietnamese"],
            ),
            (dictgen::InsensitiveStr::Ascii("idexes"), &["indices"]),
            (
                dictgen::InsensitiveStr::Ascii("congregacion"),
                &["congregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("influense"), &["influences"]),
            (
                dictgen::InsensitiveStr::Ascii("disticntly"),
                &["distinctly"],
            ),
            (dictgen::InsensitiveStr::Ascii("porgress"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("edittors"), &["editors"]),
            (
                dictgen::InsensitiveStr::Ascii("promethues"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmplify"), &["exemplify"]),
            (dictgen::InsensitiveStr::Ascii("vrsions"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhological"),
                &["psychological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individule"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylider"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("disabeling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("donwvotes"), &["downvotes"]),
            (dictgen::InsensitiveStr::Ascii("runnign"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("lokales"), &["locales"]),
            (
                dictgen::InsensitiveStr::Ascii("informaton"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrgues"), &["argues"]),
            (dictgen::InsensitiveStr::Ascii("peform"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("corperate"), &["corporate"]),
            (dictgen::InsensitiveStr::Ascii("origintea"), &["originate"]),
            (dictgen::InsensitiveStr::Ascii("barbarain"), &["barbarian"]),
            (dictgen::InsensitiveStr::Ascii("ivoker"), &["invoker"]),
            (dictgen::InsensitiveStr::Ascii("andorids"), &["androids"]),
            (dictgen::InsensitiveStr::Ascii("creatiion"), &["creation"]),
            (
                dictgen::InsensitiveStr::Ascii("superviosr"),
                &["supervisors"],
            ),
            (dictgen::InsensitiveStr::Ascii("masteers"), &["masters"]),
            (
                dictgen::InsensitiveStr::Ascii("astronuats"),
                &["astronauts"],
            ),
            (dictgen::InsensitiveStr::Ascii("vairables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("referennce"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("denstructor"),
                &["destructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massachusites"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("inro"), &["into"]),
            (
                dictgen::InsensitiveStr::Ascii("adborbs"),
                &["adsorbs", "absorbs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("birectional"),
                &["bidirectional"],
            ),
            (dictgen::InsensitiveStr::Ascii("accouning"), &["accounting"]),
            (
                dictgen::InsensitiveStr::Ascii("kidknappers"),
                &["kidnappers"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifed"), &["verified"]),
            (
                dictgen::InsensitiveStr::Ascii("intergrating"),
                &["integrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rectangual"),
                &["rectangular", "rectangle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authrorizing"),
                &["authorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("artificail"),
                &["artificial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disobediant"),
                &["disobedient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltraitor"),
                &["infiltrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transistior"),
                &["transistor"],
            ),
            (dictgen::InsensitiveStr::Ascii("despies"), &["despise"]),
            (
                dictgen::InsensitiveStr::Ascii("sigles"),
                &["singles", "sigils"],
            ),
            (dictgen::InsensitiveStr::Ascii("fith"), &["fifth", "filth"]),
            (
                dictgen::InsensitiveStr::Ascii("surrundering"),
                &["surrendering"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimensino"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("tagnet"), &["tangent"]),
            (dictgen::InsensitiveStr::Ascii("kinldy"), &["kindly"]),
            (dictgen::InsensitiveStr::Ascii("ezdrop"), &["eavesdrop"]),
            (dictgen::InsensitiveStr::Ascii("improbes"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("comestic"), &["cosmetic"]),
            (
                dictgen::InsensitiveStr::Ascii("exponantialy"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticipes"),
                &["anticipates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anphibians"),
                &["amphibians"],
            ),
            (dictgen::InsensitiveStr::Ascii("workarund"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("wre"), &["were"]),
            (dictgen::InsensitiveStr::Ascii("schedul"), &["schedule"]),
            (dictgen::InsensitiveStr::Ascii("actvates"), &["activates"]),
            (
                dictgen::InsensitiveStr::Ascii("convseration"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutiplies"), &["multiplies"]),
            (dictgen::InsensitiveStr::Ascii("intsall"), &["install"]),
            (dictgen::InsensitiveStr::Ascii("ptrss"), &["press"]),
            (
                dictgen::InsensitiveStr::Ascii("alise"),
                &["alias", "allies"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempraral"), &["temporal"]),
            (
                dictgen::InsensitiveStr::Ascii("stroed"),
                &["stored", "stroked", "strode"],
            ),
            (dictgen::InsensitiveStr::Ascii("suburburban"), &["suburban"]),
            (dictgen::InsensitiveStr::Ascii("aniwhere"), &["anywhere"]),
            (dictgen::InsensitiveStr::Ascii("modefies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("implemened"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senitmental"),
                &["sentimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("opposuite"), &["opposite"]),
            (
                dictgen::InsensitiveStr::Ascii("reqirement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("intevened"), &["intervened"]),
            (
                dictgen::InsensitiveStr::Ascii("excluses"),
                &["excludes", "excuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tremendeous"),
                &["tremendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("pollenate"), &["pollinate"]),
            (dictgen::InsensitiveStr::Ascii("specifric"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("curcial"), &["crucial"]),
            (dictgen::InsensitiveStr::Ascii("stoopidly"), &["stupidly"]),
            (
                dictgen::InsensitiveStr::Ascii("progatonists"),
                &["protagonists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutithread"),
                &["multithread"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dipsosables"),
                &["disposables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expensions"),
                &["expansions", "extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pomotion"), &["promotion"]),
            (
                dictgen::InsensitiveStr::Ascii("approximitely"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("radiers"), &["raiders"]),
            (dictgen::InsensitiveStr::Ascii("instsall"), &["install"]),
            (
                dictgen::InsensitiveStr::Ascii("reposiotory"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("presuaded"), &["persuaded"]),
            (dictgen::InsensitiveStr::Ascii("comented"), &["commented"]),
            (
                dictgen::InsensitiveStr::Ascii("unreprodicible"),
                &["unreproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("malined"), &["maligned"]),
            (dictgen::InsensitiveStr::Ascii("expeected"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("susbcript"), &["subscript"]),
            (dictgen::InsensitiveStr::Ascii("probaly"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("distrobutes"),
                &["distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("inverded"), &["inverted"]),
            (dictgen::InsensitiveStr::Ascii("levls"), &["levels"]),
            (dictgen::InsensitiveStr::Ascii("havind"), &["having"]),
            (
                dictgen::InsensitiveStr::Ascii("optimiztion"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porgresses"),
                &["progresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchronsouly"),
                &["synchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("polymino"), &["polyomino"]),
            (dictgen::InsensitiveStr::Ascii("dayly"), &["daily"]),
            (dictgen::InsensitiveStr::Ascii("adviced"), &["advised"]),
            (
                dictgen::InsensitiveStr::Ascii("sanitazion"),
                &["sanitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("doxgen"), &["doxygen"]),
            (dictgen::InsensitiveStr::Ascii("contols"), &["controls"]),
            (
                dictgen::InsensitiveStr::Ascii("contigously"),
                &["contiguously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geneartions"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dividens"), &["dividends"]),
            (dictgen::InsensitiveStr::Ascii("imporoving"), &["improving"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstraiton"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tokneization"),
                &["tokenization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheepherds"), &["shepherds"]),
            (dictgen::InsensitiveStr::Ascii("prvacy"), &["privacy"]),
            (
                dictgen::InsensitiveStr::Ascii("simplfying"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("nigthly"), &["nightly"]),
            (
                dictgen::InsensitiveStr::Ascii("iplementations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("staklers"), &["stalkers"]),
            (
                dictgen::InsensitiveStr::Ascii("intgrating"),
                &["integrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("excution"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("complementt"),
                &["complement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinating"),
                &["combining"],
            ),
            (dictgen::InsensitiveStr::Ascii("musicains"), &["musicians"]),
            (dictgen::InsensitiveStr::Ascii("reivison"), &["revision"]),
            (dictgen::InsensitiveStr::Ascii("suspicioun"), &["suspicion"]),
            (dictgen::InsensitiveStr::Ascii("indicies"), &["indices"]),
            (
                dictgen::InsensitiveStr::Ascii("ointers"),
                &["pointers", "linters"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncomprss"), &["uncompress"]),
            (dictgen::InsensitiveStr::Ascii("atribut"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("assisiate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("intercation"),
                &["integration", "interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuurently"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmization"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provisiong"),
                &["provisioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("conersion"), &["conversion"]),
            (dictgen::InsensitiveStr::Ascii("polyomial"), &["polynomial"]),
            (
                dictgen::InsensitiveStr::Ascii("refridgeration"),
                &["refrigeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("utililties"), &["utilities"]),
            (
                dictgen::InsensitiveStr::Ascii("enviornmentally"),
                &["environmentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imporvment"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialisaed"),
                &["deserialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitamis"), &["vitamins"]),
            (
                dictgen::InsensitiveStr::Ascii("drasticaly"),
                &["drastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scheduliers"),
                &["schedulers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executioneer"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadamys"), &["academies"]),
            (dictgen::InsensitiveStr::Ascii("demanling"), &["demangling"]),
            (dictgen::InsensitiveStr::Ascii("coflict"), &["conflict"]),
            (dictgen::InsensitiveStr::Ascii("predecate"), &["predicate"]),
            (dictgen::InsensitiveStr::Ascii("tonigh"), &["tonight"]),
            (dictgen::InsensitiveStr::Ascii("paraser"), &["parser"]),
            (dictgen::InsensitiveStr::Ascii("acidents"), &["accidents"]),
            (
                dictgen::InsensitiveStr::Ascii("compandium"),
                &["compendium"],
            ),
            (dictgen::InsensitiveStr::Ascii("notablly"), &["notably"]),
            (
                dictgen::InsensitiveStr::Ascii("skillfullness"),
                &["skillfulness", "skilfulness"],
            ),
            (dictgen::InsensitiveStr::Ascii("conenient"), &["convenient"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamnce"),
                &["performances", "performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("culturels"), &["cultures"]),
            (
                dictgen::InsensitiveStr::Ascii("depecation"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranasctional"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("rediected"), &["redirected"]),
            (dictgen::InsensitiveStr::Ascii("vientam"), &["vietnam"]),
            (dictgen::InsensitiveStr::Ascii("uninteded"), &["unintended"]),
            (dictgen::InsensitiveStr::Ascii("persisent"), &["persistent"]),
            (dictgen::InsensitiveStr::Ascii("janury"), &["january"]),
            (
                dictgen::InsensitiveStr::Ascii("brach"),
                &["branch", "beach", "brash", "bract", "brachia", "brace"],
            ),
            (dictgen::InsensitiveStr::Ascii("inined"), &["inlined"]),
            (
                dictgen::InsensitiveStr::Ascii("vertexs"),
                &["vertices", "vertex"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisical"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronunication"),
                &["pronunciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistatches"),
                &["mismatches"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitmizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("miscelanious"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("signuture"), &["signature"]),
            (dictgen::InsensitiveStr::Ascii("licesnse"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("andling"),
                &["handling", "dandling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuit"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partitionning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("serialsie"), &["serialise"]),
            (dictgen::InsensitiveStr::Ascii("softare"), &["software"]),
            (
                dictgen::InsensitiveStr::Ascii("irresponisble"),
                &["irresponsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregistert"),
                &["unregistered"],
            ),
            (dictgen::InsensitiveStr::Ascii("sampele"), &["sample"]),
            (dictgen::InsensitiveStr::Ascii("ignorrable"), &["ignorable"]),
            (
                dictgen::InsensitiveStr::Ascii("pojections"),
                &["projections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exapansions"),
                &["expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implikation"),
                &["implication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transgenered"),
                &["transgendered"],
            ),
            (dictgen::InsensitiveStr::Ascii("deliveres"), &["delivers"]),
            (dictgen::InsensitiveStr::Ascii("allopones"), &["allophones"]),
            (
                dictgen::InsensitiveStr::Ascii("psuedorandom"),
                &["pseudorandom"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplemenet"),
                &["reimplement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sude"),
                &["dude", "side", "sudo", "sued", "suede", "sure"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememeber"), &["remember"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowladging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoindence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termiantors"),
                &["terminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("alpahbet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("benetifing"),
                &["benefiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monomorpize"),
                &["monomorphize"],
            ),
            (dictgen::InsensitiveStr::Ascii("froces"), &["forces"]),
            (dictgen::InsensitiveStr::Ascii("sistemics"), &["systemics"]),
            (
                dictgen::InsensitiveStr::Ascii("assemblying"),
                &["assembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("detecttion"), &["detection"]),
            (dictgen::InsensitiveStr::Ascii("deprectes"), &["deprecates"]),
            (
                dictgen::InsensitiveStr::Ascii("domimation"),
                &["domination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worshippig"),
                &["worshipping"],
            ),
            (dictgen::InsensitiveStr::Ascii("preperints"), &["preprints"]),
            (
                dictgen::InsensitiveStr::Ascii("chairtable"),
                &["charitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congratulatin"),
                &["congratulating", "congratulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("selct"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("versining"), &["versioning"]),
            (dictgen::InsensitiveStr::Ascii("earilest"), &["earliest"]),
            (dictgen::InsensitiveStr::Ascii("batitsa"), &["batista"]),
            (
                dictgen::InsensitiveStr::Ascii("constnatly"),
                &["constantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sytlesheet"),
                &["stylesheet"],
            ),
            (dictgen::InsensitiveStr::Ascii("puplation"), &["population"]),
            (dictgen::InsensitiveStr::Ascii("swedisch"), &["swedish"]),
            (dictgen::InsensitiveStr::Ascii("welathy"), &["wealthy"]),
            (dictgen::InsensitiveStr::Ascii("initaily"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("stradegies"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sertification"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propeoperties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnaction"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("asterik"), &["asterisk"]),
            (dictgen::InsensitiveStr::Ascii("inquirie"), &["inquire"]),
            (dictgen::InsensitiveStr::Ascii("cumpuses"), &["compasses"]),
            (
                dictgen::InsensitiveStr::Ascii("progressino"),
                &["progressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("texuture"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("ressolving"), &["resolving"]),
            (dictgen::InsensitiveStr::Ascii("otains"), &["obtains"]),
            (dictgen::InsensitiveStr::Ascii("stromed"), &["stormed"]),
            (
                dictgen::InsensitiveStr::Ascii("deliberant"),
                &["deliberate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucrazy"),
                &["bureaucracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opertunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("barrles"), &["barrels"]),
            (dictgen::InsensitiveStr::Ascii("tesitfy"), &["testify"]),
            (
                dictgen::InsensitiveStr::Ascii("stript"),
                &["stripped", "script"],
            ),
            (dictgen::InsensitiveStr::Ascii("multplied"), &["multiplied"]),
            (dictgen::InsensitiveStr::Ascii("enventual"), &["eventual"]),
            (
                dictgen::InsensitiveStr::Ascii("interpreteer"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropolitain"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempororary"),
                &["temporary"],
            ),
            (dictgen::InsensitiveStr::Ascii("covnerted"), &["converted"]),
            (dictgen::InsensitiveStr::Ascii("bouce"), &["bounce"]),
            (
                dictgen::InsensitiveStr::Ascii("cristalline"),
                &["crystalline"],
            ),
            (dictgen::InsensitiveStr::Ascii("unbouinded"), &["unbounded"]),
            (dictgen::InsensitiveStr::Ascii("erturns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("whatseover"),
                &["whatsoever"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assemalated"),
                &["assimilated"],
            ),
            (dictgen::InsensitiveStr::Ascii("programem"), &["programme"]),
            (dictgen::InsensitiveStr::Ascii("errpr"), &["error"]),
            (
                dictgen::InsensitiveStr::Ascii("philosohpers"),
                &["philosopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondes"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainned"), &["trained"]),
            (
                dictgen::InsensitiveStr::Ascii("configuraton"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procedings"),
                &["proceedings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expiriencing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucceedde"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standarises"),
                &["standardises"],
            ),
            (dictgen::InsensitiveStr::Ascii("miinimised"), &["minimised"]),
            (dictgen::InsensitiveStr::Ascii("cardinales"), &["cardinals"]),
            (
                dictgen::InsensitiveStr::Ascii("andles"),
                &["handles", "candles", "angles", "ankles", "dandles"],
            ),
            (dictgen::InsensitiveStr::Ascii("languqge"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("hiearchically"),
                &["hierarchically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reosource"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("reluctently"),
                &["reluctantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreciateing"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gategories"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("laungages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("complately"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilezation"),
                &["civilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistrubute"),
                &["redistribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticie"), &["criticise"]),
            (dictgen::InsensitiveStr::Ascii("apllies"), &["applies"]),
            (
                dictgen::InsensitiveStr::Ascii("acccurately"),
                &["accurately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reacable"),
                &["reachable", "readable"],
            ),
            (dictgen::InsensitiveStr::Ascii("documenets"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("unjustifyed"),
                &["unjustified"],
            ),
            (dictgen::InsensitiveStr::Ascii("varriable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("pacakged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("pacages"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("hearder"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("playaround"),
                &["playground"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlysis"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("diagnoals"), &["diagonals"]),
            (
                dictgen::InsensitiveStr::Ascii("seperateing"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cassms"), &["chasms"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvince"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accreditted"),
                &["accredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrimanatory"),
                &["discriminatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("aactually"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("langugues"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("islmaic"), &["islamic"]),
            (
                dictgen::InsensitiveStr::Ascii("existentiel"),
                &["existential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ligitamate"),
                &["legitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperaors"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("instructioin"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certiticates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("desogns"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("symmertical"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annhiliates"),
                &["annihilates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootsrappers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("translaters"),
                &["translators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particuallry"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("membran"), &["membrane"]),
            (
                dictgen::InsensitiveStr::Ascii("trealing"),
                &["trailing", "trialing"],
            ),
            (dictgen::InsensitiveStr::Ascii("parralells"), &["parallels"]),
            (dictgen::InsensitiveStr::Ascii("santizers"), &["sanitizers"]),
            (
                dictgen::InsensitiveStr::Ascii("guarenteering"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fotografy"),
                &["photography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpoloated"),
                &["interpolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explititly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("lavendar"), &["lavender"]),
            (dictgen::InsensitiveStr::Ascii("prinicple"), &["principle"]),
            (dictgen::InsensitiveStr::Ascii("cofirmed"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("spefifeis"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("cosistently"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preexising"),
                &["preexisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("beacsue"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbeta"),
                &["masturbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartles"), &["cartels"]),
            (
                dictgen::InsensitiveStr::Ascii("determnines"),
                &["determines"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuple"), &["couple"]),
            (dictgen::InsensitiveStr::Ascii("alimunium"), &["aluminium"]),
            (dictgen::InsensitiveStr::Ascii("compleat"), &["complete"]),
            (
                dictgen::InsensitiveStr::Ascii("interepolating"),
                &["interpolating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shandeleers"),
                &["chandeliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchantbility"),
                &["merchantability"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulicast"), &["multicast"]),
            (
                dictgen::InsensitiveStr::Ascii("creativily"),
                &["creatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empressively"),
                &["impressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesd"), &["used"]),
            (
                dictgen::InsensitiveStr::Ascii("apostropohes"),
                &["apostrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("triyng"), &["trying"]),
            (dictgen::InsensitiveStr::Ascii("ubutu"), &["ubuntu"]),
            (
                dictgen::InsensitiveStr::Ascii("preceeding"),
                &["preceding", "proceeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("drfiting"), &["drifting"]),
            (dictgen::InsensitiveStr::Ascii("analsyed"), &["analysed"]),
            (dictgen::InsensitiveStr::Ascii("docrtine"), &["doctrine"]),
            (dictgen::InsensitiveStr::Ascii("obyect"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("mistatchd"), &["mismatched"]),
            (
                dictgen::InsensitiveStr::Ascii("credintial"),
                &["credential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualizaion"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("valies"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("infalability"),
                &["infallibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializaed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("recvied"), &["received"]),
            (
                dictgen::InsensitiveStr::Ascii("civelezations"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimums"), &["minimums"]),
            (
                dictgen::InsensitiveStr::Ascii("experiation"),
                &["expiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miracurously"),
                &["miraculous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponsds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouinding"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("whataver"), &["whatever"]),
            (dictgen::InsensitiveStr::Ascii("columms"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("mashine"), &["machine"]),
            (
                dictgen::InsensitiveStr::Ascii("misconfiged"),
                &["misconfigured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exptectations"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanses"), &["chances"]),
            (dictgen::InsensitiveStr::Ascii("blokc"), &["block", "bloke"]),
            (dictgen::InsensitiveStr::Ascii("targetin"), &["targeting"]),
            (dictgen::InsensitiveStr::Ascii("prceeding"), &["preceding"]),
            (dictgen::InsensitiveStr::Ascii("androis"), &["androids"]),
            (dictgen::InsensitiveStr::Ascii("anyting"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegothiations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("coutning"), &["counting"]),
            (
                dictgen::InsensitiveStr::Ascii("autoconpleted"),
                &["autocompleted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sarcastisch"),
                &["sarcastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enginineers"),
                &["engineers"],
            ),
            (dictgen::InsensitiveStr::Ascii("memebered"), &["remembered"]),
            (dictgen::InsensitiveStr::Ascii("phsyicist"), &["physicist"]),
            (
                dictgen::InsensitiveStr::Ascii("exercized"),
                &["exercised", "exorcized"],
            ),
            (dictgen::InsensitiveStr::Ascii("insigt"), &["insight"]),
            (
                dictgen::InsensitiveStr::Ascii("excurciating"),
                &["excruciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enginineering"),
                &["engineering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interleiive"),
                &["interleave"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeings"), &["openings"]),
            (
                dictgen::InsensitiveStr::Ascii("installlation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missconfiguration"),
                &["misconfiguration"],
            ),
            (dictgen::InsensitiveStr::Ascii("hisotry"), &["history"]),
            (dictgen::InsensitiveStr::Ascii("prviously"), &["previously"]),
            (dictgen::InsensitiveStr::Ascii("aftrer"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("primiary"), &["primary"]),
            (dictgen::InsensitiveStr::Ascii("inerest"), &["interest"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailiseing"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheduled"), &["scheduled"]),
            (dictgen::InsensitiveStr::Ascii("serached"), &["searched"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotor"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculaton"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("retart"), &["restart"]),
            (
                dictgen::InsensitiveStr::Ascii("alphapetically"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discreditted"),
                &["discredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clearstorys"),
                &["clerestories"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("affort"),
                &["afford", "effort"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unauthoried"),
                &["unauthorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caricuture"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonymusly"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnformed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("leaneant"), &["lenient"]),
            (dictgen::InsensitiveStr::Ascii("charactr"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("unexpeted"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("speficififes"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doucmented"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("publsiher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("referecning"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dperecated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("extermely"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("misakes"), &["mistakes"]),
            (dictgen::InsensitiveStr::Ascii("archvial"), &["archival"]),
            (
                dictgen::InsensitiveStr::Ascii("executabnle"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finction"),
                &["function", "fiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveniance"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("lispticks"), &["lipsticks"]),
            (dictgen::InsensitiveStr::Ascii("anologous"), &["analogous"]),
            (dictgen::InsensitiveStr::Ascii("delemeter"), &["delimiter"]),
            (
                dictgen::InsensitiveStr::Ascii("uneccesarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereigstering"),
                &["deregistering"],
            ),
            (dictgen::InsensitiveStr::Ascii("endlessy"), &["endlessly"]),
            (dictgen::InsensitiveStr::Ascii("roviders"), &["providers"]),
            (
                dictgen::InsensitiveStr::Ascii("modificaitons"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philadelhpia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mehtodical"),
                &["methodical"],
            ),
            (dictgen::InsensitiveStr::Ascii("woithout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("dissabling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("cylclical"), &["cyclical"]),
            (dictgen::InsensitiveStr::Ascii("attribuite"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("timestanps"),
                &["timestamps", "timespans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribuites"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("nuerons"), &["neurons"]),
            (dictgen::InsensitiveStr::Ascii("avialable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("insatnce"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("cristallising"),
                &["crystallising"],
            ),
            (dictgen::InsensitiveStr::Ascii("thme"), &["them"]),
            (
                dictgen::InsensitiveStr::Ascii("algorthynically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("confgures"), &["configures"]),
            (
                dictgen::InsensitiveStr::Ascii("intercepto"),
                &["interception"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("sysyem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("mechanicallly"),
                &["mechanically"],
            ),
            (dictgen::InsensitiveStr::Ascii("abanonds"), &["abandons"]),
            (
                dictgen::InsensitiveStr::Ascii("metiphorically"),
                &["metaphorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("prolems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("vegetatians"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("achieval"), &["achievable"]),
            (dictgen::InsensitiveStr::Ascii("wrappped"), &["wrapped"]),
            (
                dictgen::InsensitiveStr::Ascii("campatibility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommondations"),
                &["recommendations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atmosoheric"),
                &["atmospheric"],
            ),
            (dictgen::InsensitiveStr::Ascii("coording"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("salughtered"),
                &["slaughtered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spected"),
                &["expected", "spectated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfiors"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("welll"), &["well"]),
            (
                dictgen::InsensitiveStr::Ascii("compramise"),
                &["compromise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptivie"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiminter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimening"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("addrerss"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("appologising"),
                &["apologising"],
            ),
            (dictgen::InsensitiveStr::Ascii("withotu"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("represso"), &["repression"]),
            (
                dictgen::InsensitiveStr::Ascii("coerscivity"),
                &["coercivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematisch"),
                &["mathematics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandiser"),
                &["aggrandiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonisation"),
                &["synchronisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unuseed"), &["unused"]),
            (dictgen::InsensitiveStr::Ascii("pakaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("trjaectory"),
                &["trajectory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisitics"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("exipry"), &["expiry"]),
            (
                dictgen::InsensitiveStr::Ascii("commnuative"),
                &["commutative"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehir"), &["their"]),
            (dictgen::InsensitiveStr::Ascii("tracin"), &["tracing"]),
            (dictgen::InsensitiveStr::Ascii("arryas"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("provodes"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("trannsexual"),
                &["transsexual"],
            ),
            (dictgen::InsensitiveStr::Ascii("absloute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("pennensulas"),
                &["peninsulas"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertial"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("specyifing"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructiong"),
                &["constructing", "construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("vinigar"), &["vinegar"]),
            (dictgen::InsensitiveStr::Ascii("relesaing"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("aquit"), &["acquit"]),
            (dictgen::InsensitiveStr::Ascii("tansports"), &["transports"]),
            (dictgen::InsensitiveStr::Ascii("underfows"), &["underflows"]),
            (dictgen::InsensitiveStr::Ascii("pretinent"), &["pertinent"]),
            (dictgen::InsensitiveStr::Ascii("chargeur"), &["charger"]),
            (dictgen::InsensitiveStr::Ascii("redirectd"), &["redirected"]),
            (
                dictgen::InsensitiveStr::Ascii("unprmopted"),
                &["unprompted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfromance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatgram"), &["datagram"]),
            (dictgen::InsensitiveStr::Ascii("torlence"), &["tolerance"]),
            (dictgen::InsensitiveStr::Ascii("limitet"), &["limited"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoticates"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inremediate"),
                &["intermediate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blancing"),
                &["balancing", "glancing", "lancing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adernaline"),
                &["adrenaline"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimals"), &["minimalist"]),
            (
                dictgen::InsensitiveStr::Ascii("unreasonablely"),
                &["unreasonably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesssors"),
                &["successors"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("confurses"), &["confuses"]),
            (dictgen::InsensitiveStr::Ascii("eyebrowns"), &["eyebrows"]),
            (dictgen::InsensitiveStr::Ascii("boraches"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("commenty"), &["commentary"]),
            (
                dictgen::InsensitiveStr::Ascii("projectles"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segregatie"),
                &["segregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginnging"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("documentarios"),
                &["documentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("cariier"), &["carrier"]),
            (
                dictgen::InsensitiveStr::Ascii("valdations"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invervention"),
                &["intervention"],
            ),
            (dictgen::InsensitiveStr::Ascii("publiclly"), &["publicly"]),
            (
                dictgen::InsensitiveStr::Ascii("desillusioned"),
                &["disillusioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructur"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depentence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("anopther"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("varisty"), &["varsity"]),
            (
                dictgen::InsensitiveStr::Ascii("caluculates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("undercore"), &["underscore"]),
            (
                dictgen::InsensitiveStr::Ascii("initiailized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("thnak"), &["thank"]),
            (dictgen::InsensitiveStr::Ascii("movememnt"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("translyvanian"),
                &["transylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenoard"), &["leonard"]),
            (
                dictgen::InsensitiveStr::Ascii("proletariot"),
                &["proletariat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silouhetted"),
                &["silhouetted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("casualries"),
                &["casualties"],
            ),
            (dictgen::InsensitiveStr::Ascii("desigins"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("afterawards"),
                &["afterwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fondamentally"),
                &["fundamentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustratin"),
                &["frustration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excpetionally"),
                &["exceptionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliw"), &["follow"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentatal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serverlsss"),
                &["serverless"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnining"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("peridoic"), &["periodic"]),
            (dictgen::InsensitiveStr::Ascii("exclues"), &["excludes"]),
            (
                dictgen::InsensitiveStr::Ascii("inprisonment"),
                &["imprisonment"],
            ),
            (dictgen::InsensitiveStr::Ascii("teplmated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("deketing"), &["deleting"]),
            (dictgen::InsensitiveStr::Ascii("platofms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("travered"), &["traversed"]),
            (dictgen::InsensitiveStr::Ascii("pariwise"), &["pairwise"]),
            (
                dictgen::InsensitiveStr::Ascii("neigbhours"),
                &["neighbours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatin"),
                &["information", "informing"],
            ),
            (dictgen::InsensitiveStr::Ascii("apilogue"), &["epilogue"]),
            (
                dictgen::InsensitiveStr::Ascii("indexs"),
                &["indexes", "indices"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipilines"), &["pipelines"]),
            (dictgen::InsensitiveStr::Ascii("masscare"), &["massacre"]),
            (
                dictgen::InsensitiveStr::Ascii("paychologists"),
                &["psychologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("naiborly"), &["neighborly"]),
            (
                dictgen::InsensitiveStr::Ascii("notofocation"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("montioring"),
                &["monitoring"],
            ),
            (dictgen::InsensitiveStr::Ascii("conneced"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("yogoslavia"),
                &["yugoslavia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unble"),
                &["unable", "uncle"],
            ),
            (dictgen::InsensitiveStr::Ascii("evevn"), &["even"]),
            (dictgen::InsensitiveStr::Ascii("separes"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("madantory"), &["mandatory"]),
            (
                dictgen::InsensitiveStr::Ascii("australiams"),
                &["australians"],
            ),
            (dictgen::InsensitiveStr::Ascii("wawrick"), &["warwick"]),
            (
                dictgen::InsensitiveStr::Ascii("mentionning"),
                &["mentioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("paralllel"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("americats"), &["americas"]),
            (
                dictgen::InsensitiveStr::Ascii("consensuarlo"),
                &["consensual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orchestraded"),
                &["orchestrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaeged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("informfational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenanets"), &["tenants"]),
            (
                dictgen::InsensitiveStr::Ascii("instelling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("nothihg"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("errects"), &["erects"]),
            (dictgen::InsensitiveStr::Ascii("departer"), &["departure"]),
            (dictgen::InsensitiveStr::Ascii("analizer"), &["analyzer"]),
            (dictgen::InsensitiveStr::Ascii("stivks"), &["sticks"]),
            (
                dictgen::InsensitiveStr::Ascii("braziliians"),
                &["brazilians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assupmtion"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitionable"),
                &["transitional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecent"),
                &["percent", "recent", "decent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("afernoon"), &["afternoon"]),
            (dictgen::InsensitiveStr::Ascii("keeo"), &["keep"]),
            (dictgen::InsensitiveStr::Ascii("sycle"), &["cycle"]),
            (dictgen::InsensitiveStr::Ascii("anuwhere"), &["anywhere"]),
            (dictgen::InsensitiveStr::Ascii("servicin"), &["servicing"]),
            (
                dictgen::InsensitiveStr::Ascii("vinigarette"),
                &["vinaigrette"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulgaia"), &["bulgaria"]),
            (dictgen::InsensitiveStr::Ascii("minumum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("againsst"), &["against"]),
            (dictgen::InsensitiveStr::Ascii("exagerate"), &["exaggerate"]),
            (dictgen::InsensitiveStr::Ascii("gateing"), &["gating"]),
            (
                dictgen::InsensitiveStr::Ascii("characteristerics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuztomizing"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orienate"),
                &["orientate", "orient", "ornate"],
            ),
            (dictgen::InsensitiveStr::Ascii("probllems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("nojifications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesst"),
                &["interests", "interest"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimached"), &["mismatched"]),
            (dictgen::InsensitiveStr::Ascii("hpe"), &["hope"]),
            (
                dictgen::InsensitiveStr::Ascii("accessoirs"),
                &["accessories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environement"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentaions"),
                &["concatenations"],
            ),
            (dictgen::InsensitiveStr::Ascii("stragety"), &["strategy"]),
            (
                dictgen::InsensitiveStr::Ascii("perseverin"),
                &["persevering"],
            ),
            (dictgen::InsensitiveStr::Ascii("seelect"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("unnnecessarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertisment"),
                &["advertisement"],
            ),
            (dictgen::InsensitiveStr::Ascii("constatnt"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("triggerrs"), &["triggers"]),
            (dictgen::InsensitiveStr::Ascii("wrties"), &["writes"]),
            (dictgen::InsensitiveStr::Ascii("mcgergor"), &["mcgregor"]),
            (
                dictgen::InsensitiveStr::Ascii("corrosponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("orginals"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("algortismically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approoximate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demograhpics"),
                &["demographics"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyrrent"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("tongiht"), &["tonight"]),
            (
                dictgen::InsensitiveStr::Ascii("sturucturing"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compinsate"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspind"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirctive"), &["directive"]),
            (
                dictgen::InsensitiveStr::Ascii("substucture"),
                &["substructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("shrtcut"), &["shortcut"]),
            (
                dictgen::InsensitiveStr::Ascii("overcompansates"),
                &["overcompensates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technnicians"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stablization"),
                &["stabilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("namming"), &["naming"]),
            (
                dictgen::InsensitiveStr::Ascii("persectuion"),
                &["persecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dectecte"),
                &["detect", "detected", "detects"],
            ),
            (dictgen::InsensitiveStr::Ascii("tbey"), &["they"]),
            (dictgen::InsensitiveStr::Ascii("quesrions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreagated"),
                &["aggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("redemtion"), &["redemption"]),
            (
                dictgen::InsensitiveStr::Ascii("irrellevant"),
                &["irrelevant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterperts"),
                &["misinterprets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skillfull"),
                &["skillful", "skillfully", "skilful", "skilfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("relocaion"), &["relocation"]),
            (
                dictgen::InsensitiveStr::Ascii("indetifiable"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("humouros"), &["humorous"]),
            (
                dictgen::InsensitiveStr::Ascii("persistance"),
                &["persistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("defualts"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("communisim"), &["communism"]),
            (
                dictgen::InsensitiveStr::Ascii("agnostisicm"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("cemetarey"), &["cemetery"]),
            (dictgen::InsensitiveStr::Ascii("tekst"), &["text"]),
            (
                dictgen::InsensitiveStr::Ascii("inpsection"),
                &["inspection"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggrees"), &["agrees"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenaes"),
                &["concatenates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilaize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrtibution"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("dalta"), &["delta"]),
            (
                dictgen::InsensitiveStr::Ascii("relacatable"),
                &["relocatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("neibhbors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("produkts"), &["products"]),
            (dictgen::InsensitiveStr::Ascii("cheecker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("connetions"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opjections"),
                &["objections"],
            ),
            (dictgen::InsensitiveStr::Ascii("guardains"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("toubles"), &["troubles"]),
            (dictgen::InsensitiveStr::Ascii("depiciton"), &["depiction"]),
            (
                dictgen::InsensitiveStr::Ascii("comprimise"),
                &["compromise"],
            ),
            (dictgen::InsensitiveStr::Ascii("receptos"), &["receptors"]),
            (dictgen::InsensitiveStr::Ascii("historial"), &["historical"]),
            (dictgen::InsensitiveStr::Ascii("nescesarry"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("interprete"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhhod"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("itelf"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("asnychronoue"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("insering"), &["inserting"]),
            (dictgen::InsensitiveStr::Ascii("assmelber"), &["assembler"]),
            (
                dictgen::InsensitiveStr::Ascii("scrool"),
                &["scroll", "school"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferrencing"),
                &["dereferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("stickness"), &["stickiness"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentatin"),
                &["experimentation", "experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dirctories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("constantsm"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("analoges"), &["analogues"]),
            (
                dictgen::InsensitiveStr::Ascii("explicityly"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supresion"),
                &["suppression"],
            ),
            (dictgen::InsensitiveStr::Ascii("gettitem"), &["getitem"]),
            (dictgen::InsensitiveStr::Ascii("professer"), &["professor"]),
            (
                dictgen::InsensitiveStr::Ascii("implementaed"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etended"),
                &["extended", "attended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmical"),
                &["programmatical"],
            ),
            (dictgen::InsensitiveStr::Ascii("esitmator"), &["estimator"]),
            (dictgen::InsensitiveStr::Ascii("importace"), &["importance"]),
            (dictgen::InsensitiveStr::Ascii("launck"), &["launch"]),
            (dictgen::InsensitiveStr::Ascii("benagls"), &["bengals"]),
            (
                dictgen::InsensitiveStr::Ascii("charcteristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("focibly"), &["forcibly"]),
            (dictgen::InsensitiveStr::Ascii("markerts"), &["markers"]),
            (dictgen::InsensitiveStr::Ascii("writiing"), &["writing"]),
            (dictgen::InsensitiveStr::Ascii("dribblin"), &["dribbling"]),
            (
                dictgen::InsensitiveStr::Ascii("functionaility"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thunderbowl"),
                &["thunderbolt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissimliar"),
                &["dissimilar"],
            ),
            (dictgen::InsensitiveStr::Ascii("unpluging"), &["unplugging"]),
            (dictgen::InsensitiveStr::Ascii("formmat"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("linerizations"),
                &["linearizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caronavirus"),
                &["coronavirus"],
            ),
            (dictgen::InsensitiveStr::Ascii("regconize"), &["recognize"]),
            (dictgen::InsensitiveStr::Ascii("unborned"), &["unborn"]),
            (
                dictgen::InsensitiveStr::Ascii("indiviudally"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("widnows"),
                &["windows", "widows"],
            ),
            (dictgen::InsensitiveStr::Ascii("misstaken"), &["mistaken"]),
            (dictgen::InsensitiveStr::Ascii("nagivate"), &["navigate"]),
            (dictgen::InsensitiveStr::Ascii("wriet"), &["write"]),
            (
                dictgen::InsensitiveStr::Ascii("manifestaion"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("coratia"), &["croatia"]),
            (dictgen::InsensitiveStr::Ascii("fauilure"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("exlpoited"), &["exploited"]),
            (dictgen::InsensitiveStr::Ascii("analyziing"), &["analyzing"]),
            (
                dictgen::InsensitiveStr::Ascii("prposes"),
                &["proposes", "purposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goalkeaper"),
                &["goalkeeper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifally"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("netacpe"), &["netscape"]),
            (dictgen::InsensitiveStr::Ascii("emojii"), &["emoji"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondint"),
                &["corresponding", "correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptionnally"),
                &["exceptionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("resseted"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("simly"),
                &["simply", "simile", "smiley"],
            ),
            (dictgen::InsensitiveStr::Ascii("progam"), &["program"]),
            (dictgen::InsensitiveStr::Ascii("presnts"), &["presents"]),
            (dictgen::InsensitiveStr::Ascii("sendinng"), &["sending"]),
            (dictgen::InsensitiveStr::Ascii("scheudler"), &["scheduler"]),
            (dictgen::InsensitiveStr::Ascii("lighitng"), &["lighting"]),
            (dictgen::InsensitiveStr::Ascii("volenteer"), &["volunteer"]),
            (
                dictgen::InsensitiveStr::Ascii("incorperated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfrom"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("arrancing"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("franticlly"),
                &["frantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapearance"),
                &["disappearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatinated"),
                &["contaminated", "concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archetectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("shortned"), &["shortened"]),
            (
                dictgen::InsensitiveStr::Ascii("comceptually"),
                &["conceptually"],
            ),
            (dictgen::InsensitiveStr::Ascii("fluctuand"), &["fluctuant"]),
            (
                dictgen::InsensitiveStr::Ascii("strenghten"),
                &["strengthen"],
            ),
            (dictgen::InsensitiveStr::Ascii("poliy"), &["policy"]),
            (
                dictgen::InsensitiveStr::Ascii("teleportating"),
                &["teleportation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bilt"), &["built"]),
            (dictgen::InsensitiveStr::Ascii("specialops"), &["specials"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliazer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifeteime"), &["lifetime"]),
            (dictgen::InsensitiveStr::Ascii("brodaway"), &["broadway"]),
            (dictgen::InsensitiveStr::Ascii("espreso"), &["espresso"]),
            (dictgen::InsensitiveStr::Ascii("normailze"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("whyt"), &["what", "why"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessfuly"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("excahnges"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("paramtere"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("foramt"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("decomposited"),
                &["decomposed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sandwhiches"),
                &["sandwiches"],
            ),
            (dictgen::InsensitiveStr::Ascii("schaeduler"), &["scheduler"]),
            (dictgen::InsensitiveStr::Ascii("formen"), &["foremen"]),
            (
                dictgen::InsensitiveStr::Ascii("installlers"),
                &["installers"],
            ),
            (dictgen::InsensitiveStr::Ascii("typped"), &["typed"]),
            (
                dictgen::InsensitiveStr::Ascii("exprsesion"),
                &["expression"],
            ),
            (dictgen::InsensitiveStr::Ascii("finnaly"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("sryinge"), &["syringe"]),
            (
                dictgen::InsensitiveStr::Ascii("reformattin"),
                &["reformatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatenadion"),
                &["concatenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("avacados"), &["avocados"]),
            (
                dictgen::InsensitiveStr::Ascii("accomodoate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manupilation"),
                &["manipulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restaurnat"),
                &["restaurant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enity"),
                &["entity", "enmity"],
            ),
            (dictgen::InsensitiveStr::Ascii("pointure"), &["pointer"]),
            (dictgen::InsensitiveStr::Ascii("slightl"), &["slightly"]),
            (
                dictgen::InsensitiveStr::Ascii("benevelant"),
                &["benevolent"],
            ),
            (dictgen::InsensitiveStr::Ascii("exexuting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("jorunals"), &["journals"]),
            (dictgen::InsensitiveStr::Ascii("produly"), &["proudly"]),
            (
                dictgen::InsensitiveStr::Ascii("recipt"),
                &["receipt", "recipe"],
            ),
            (dictgen::InsensitiveStr::Ascii("mek"), &["me"]),
            (
                dictgen::InsensitiveStr::Ascii("abiguously"),
                &["ambiguously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shakespeer"),
                &["shakespeare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programatically"),
                &["programmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("siganled"), &["signaled"]),
            (dictgen::InsensitiveStr::Ascii("biznesses"), &["businesses"]),
            (dictgen::InsensitiveStr::Ascii("ewhwer"), &["where"]),
            (
                dictgen::InsensitiveStr::Ascii("compenstates"),
                &["compensates"],
            ),
            (dictgen::InsensitiveStr::Ascii("manauls"), &["manuals"]),
            (
                dictgen::InsensitiveStr::Ascii("refertences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("quith"), &["quit", "with"]),
            (
                dictgen::InsensitiveStr::Ascii("resues"),
                &["reuses", "resumes", "rescues"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheltic"), &["athletic"]),
            (dictgen::InsensitiveStr::Ascii("adbandon"), &["abandon"]),
            (dictgen::InsensitiveStr::Ascii("remaping"), &["remapping"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitonally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionl"),
                &["professionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placehoulders"),
                &["placeholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("shanges"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("installate"), &["install"]),
            (dictgen::InsensitiveStr::Ascii("chechkout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("contenintal"),
                &["continental"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacum"), &["vacuum"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffcient"),
                &["inefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("cimmand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("limosine"), &["limousine"]),
            (
                dictgen::InsensitiveStr::Ascii("rectengularity"),
                &["rectangularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplication"),
                &["simplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progapators"),
                &["propagators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompability"),
                &["incompatible", "incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahnnel"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("summmer"), &["summer"]),
            (dictgen::InsensitiveStr::Ascii("follwower"), &["follower"]),
            (
                dictgen::InsensitiveStr::Ascii("curreny"),
                &["currency", "current", "currently"],
            ),
            (dictgen::InsensitiveStr::Ascii("igored"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("follfowin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("contianed"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("adequatly"), &["adequately"]),
            (
                dictgen::InsensitiveStr::Ascii("coloers"),
                &["colors", "colours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configureation"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancres"), &["cancers"]),
            (
                dictgen::InsensitiveStr::Ascii("compansate"),
                &["compensate"],
            ),
            (dictgen::InsensitiveStr::Ascii("selcetor"), &["selector"]),
            (
                dictgen::InsensitiveStr::Ascii("accumlation"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inclince"), &["incline"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("broge"), &["brogue"]),
            (dictgen::InsensitiveStr::Ascii("relpy"), &["reply"]),
            (dictgen::InsensitiveStr::Ascii("suposing"), &["supposing"]),
            (dictgen::InsensitiveStr::Ascii("diableld"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("unforgiveble"),
                &["unforgivable"],
            ),
            (dictgen::InsensitiveStr::Ascii("invlove"), &["involve"]),
            (dictgen::InsensitiveStr::Ascii("uplod"), &["upload"]),
            (dictgen::InsensitiveStr::Ascii("chngelog"), &["changelog"]),
            (dictgen::InsensitiveStr::Ascii("spirtual"), &["spiritual"]),
            (dictgen::InsensitiveStr::Ascii("ambigious"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("overwheliming"),
                &["overwhelming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regconizers"),
                &["recognizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("directd"), &["directed"]),
            (
                dictgen::InsensitiveStr::Ascii("trasporter"),
                &["transporter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interationalism"),
                &["internationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("nott"), &["not"]),
            (
                dictgen::InsensitiveStr::Ascii("entrapeneurs"),
                &["entrepreneur", "entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("runnners"), &["runners"]),
            (
                dictgen::InsensitiveStr::Ascii("historicas"),
                &["historians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumized"),
                &["customized"],
            ),
            (dictgen::InsensitiveStr::Ascii("immerisve"), &["immersive"]),
            (dictgen::InsensitiveStr::Ascii("assignting"), &["assigning"]),
            (dictgen::InsensitiveStr::Ascii("ignorrg"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("retieval"), &["retrieval"]),
            (
                dictgen::InsensitiveStr::Ascii("accoordingly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extarctors"),
                &["extractors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoroprate"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meterologist"),
                &["meteorologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("xdpf"), &["xpdf"]),
            (dictgen::InsensitiveStr::Ascii("namespsce"), &["namespace"]),
            (
                dictgen::InsensitiveStr::Ascii("sentationalism"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("destector"), &["detector"]),
            (
                dictgen::InsensitiveStr::Ascii("generatioon"),
                &["generation"],
            ),
            (dictgen::InsensitiveStr::Ascii("critices"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("sockt"), &["socket"]),
            (dictgen::InsensitiveStr::Ascii("anoynimize"), &["anonymize"]),
            (
                dictgen::InsensitiveStr::Ascii("earthquackes"),
                &["earthquakes"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaises"), &["localises"]),
            (dictgen::InsensitiveStr::Ascii("annoation"), &["annotation"]),
            (dictgen::InsensitiveStr::Ascii("unsanfe"), &["unsafe"]),
            (
                dictgen::InsensitiveStr::Ascii("proclamied"),
                &["proclaimed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funxtioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("increates"), &["increases"]),
            (
                dictgen::InsensitiveStr::Ascii("normallizing"),
                &["normalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consideribly"),
                &["considerably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorhythms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ariculations"),
                &["articulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("subfromat"), &["subformat"]),
            (
                dictgen::InsensitiveStr::Ascii("extensivly"),
                &["extensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("odering"), &["ordering"]),
            (dictgen::InsensitiveStr::Ascii("alreasy"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("renderred"), &["rendered"]),
            (
                dictgen::InsensitiveStr::Ascii("dependend"),
                &["dependent", "depended", "depend"],
            ),
            (dictgen::InsensitiveStr::Ascii("relesaes"), &["releases"]),
            (dictgen::InsensitiveStr::Ascii("munchis"), &["munchies"]),
            (
                dictgen::InsensitiveStr::Ascii("addministration"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("hitboxers"), &["hitboxes"]),
            (
                dictgen::InsensitiveStr::Ascii("appllication"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performens"),
                &["performers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prevalecen"),
                &["prevalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("headhpone"), &["headphone"]),
            (
                dictgen::InsensitiveStr::Ascii("specfications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representetives"),
                &["representatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ventilacion"),
                &["ventilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("faling"),
                &["failing", "falling"],
            ),
            (dictgen::InsensitiveStr::Ascii("checck"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenetaions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preservered"),
                &["preserved"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparre"), &["compare"]),
            (dictgen::InsensitiveStr::Ascii("encryts"), &["encrypts"]),
            (dictgen::InsensitiveStr::Ascii("memeber"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("appoitnments"),
                &["appointments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impementaions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("animatiions"),
                &["animations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermaket"),
                &["supermarket"],
            ),
            (dictgen::InsensitiveStr::Ascii("enocding"), &["encoding"]),
            (
                dictgen::InsensitiveStr::Ascii("individualy"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("pooints"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("romotely"), &["remotely"]),
            (
                dictgen::InsensitiveStr::Ascii("matemathical"),
                &["mathematical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambadexterously"),
                &["ambidextrously"],
            ),
            (dictgen::InsensitiveStr::Ascii("configire"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("legitamately"),
                &["legitimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("cjanged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("ternimation"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accesssiiblity"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaneg"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitaes"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checksuming"),
                &["checksumming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apropiate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("porcessed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("sepcifies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("satisifies"), &["satisfies"]),
            (dictgen::InsensitiveStr::Ascii("exapanded"), &["expanded"]),
            (
                dictgen::InsensitiveStr::Ascii("apprieciation"),
                &["appreciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undeniablly"),
                &["undeniably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregnencies"),
                &["pregnancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("animetion"), &["animation"]),
            (dictgen::InsensitiveStr::Ascii("mulit"), &["multi"]),
            (
                dictgen::InsensitiveStr::Ascii("transfert"),
                &["transfer", "transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("metohd"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("sysmbols"), &["symbols"]),
            (
                dictgen::InsensitiveStr::Ascii("attriburted"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scientfically"),
                &["scientifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounaary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("mircoscopic"),
                &["microscopic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanhedrim"), &["sanhedrin"]),
            (
                dictgen::InsensitiveStr::Ascii("idendification"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("valenica"), &["valencia"]),
            (dictgen::InsensitiveStr::Ascii("perminant"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("pacakging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("authroritatively"),
                &["authoritatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritzes"),
                &["prioritizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentaires"),
                &["documentaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitacion"),
                &["limitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinstic"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("holdiay"), &["holiday"]),
            (
                dictgen::InsensitiveStr::Ascii("inspirationnal"),
                &["inspirational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agriculturual"),
                &["agricultural"],
            ),
            (dictgen::InsensitiveStr::Ascii("impproving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("sypported"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("compialtion"),
                &["compilation", "complication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommonder"),
                &["recommender"],
            ),
            (dictgen::InsensitiveStr::Ascii("srot"), &["sort"]),
            (
                dictgen::InsensitiveStr::Ascii("interferens"),
                &["interferes", "interference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impropable"),
                &["improbable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ordred"), &["ordered"]),
            (dictgen::InsensitiveStr::Ascii("axtive"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("forld"), &["fold"]),
            (
                dictgen::InsensitiveStr::Ascii("experimintel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("windoes"), &["windows"]),
            (
                dictgen::InsensitiveStr::Ascii("terrorsits"),
                &["terrorists"],
            ),
            (dictgen::InsensitiveStr::Ascii("positivly"), &["positively"]),
            (
                dictgen::InsensitiveStr::Ascii("exlcamation"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("withint"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("sountrack"), &["soundtrack"]),
            (
                dictgen::InsensitiveStr::Ascii("sharkening"),
                &["sharpening"],
            ),
            (dictgen::InsensitiveStr::Ascii("counciles"), &["councils"]),
            (dictgen::InsensitiveStr::Ascii("toritllas"), &["tortillas"]),
            (dictgen::InsensitiveStr::Ascii("swicthes"), &["switches"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilplicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("ligthroom"), &["lightroom"]),
            (dictgen::InsensitiveStr::Ascii("wrolds"), &["worlds"]),
            (
                dictgen::InsensitiveStr::Ascii("communciate"),
                &["communicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomaptibelities"),
                &["incompatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("pretected"), &["protected"]),
            (
                dictgen::InsensitiveStr::Ascii("perisistence"),
                &["persistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("epecifica"), &["especifica"]),
            (dictgen::InsensitiveStr::Ascii("normallize"), &["normalize"]),
            (
                dictgen::InsensitiveStr::Ascii("constrolling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currectly"),
                &["correctly", "currently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tollerances"),
                &["tolerances"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsettin"), &["offsetting"]),
            (dictgen::InsensitiveStr::Ascii("disgarees"), &["disagrees"]),
            (dictgen::InsensitiveStr::Ascii("acatemics"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("comfusion"), &["confusion"]),
            (dictgen::InsensitiveStr::Ascii("forumlate"), &["formulate"]),
            (
                dictgen::InsensitiveStr::Ascii("drammatically"),
                &["grammatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifing"), &["specifying"]),
            (dictgen::InsensitiveStr::Ascii("filfills"), &["fulfills"]),
            (
                dictgen::InsensitiveStr::Ascii("chancge"),
                &["change", "chance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correnspondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("unanswerd"), &["unanswered"]),
            (dictgen::InsensitiveStr::Ascii("acceot"), &["accept"]),
            (dictgen::InsensitiveStr::Ascii("enumerble"), &["enumerable"]),
            (dictgen::InsensitiveStr::Ascii("criricals"), &["criticals"]),
            (dictgen::InsensitiveStr::Ascii("laywer"), &["lawyer"]),
            (dictgen::InsensitiveStr::Ascii("projetion"), &["projection"]),
            (
                dictgen::InsensitiveStr::Ascii("detemining"),
                &["determining"],
            ),
            (dictgen::InsensitiveStr::Ascii("localtions"), &["locations"]),
            (dictgen::InsensitiveStr::Ascii("constrcut"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("chisiled"), &["chiseled"]),
            (
                dictgen::InsensitiveStr::Ascii("restraunts"),
                &["restaurants", "restraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeiments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frientzoned"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanppening"), &["happening"]),
            (
                dictgen::InsensitiveStr::Ascii("resuing"),
                &["reusing", "resuming", "rescuing"],
            ),
            (dictgen::InsensitiveStr::Ascii("hiarchy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("integeration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("persisten"), &["persistent"]),
            (
                dictgen::InsensitiveStr::Ascii("encaspulate"),
                &["encapsulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shatterring"),
                &["shattering"],
            ),
            (dictgen::InsensitiveStr::Ascii("deciples"), &["disciples"]),
            (
                dictgen::InsensitiveStr::Ascii("crendential"),
                &["credential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oriantation"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("beggining"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("flass"),
                &["class", "glass", "flask", "flash"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internationnaly"),
                &["internationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("germanisch"), &["germanic"]),
            (dictgen::InsensitiveStr::Ascii("servoces"), &["services"]),
            (
                dictgen::InsensitiveStr::Ascii("adventurin"),
                &["adventuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprached"), &["approached"]),
            (dictgen::InsensitiveStr::Ascii("cosnumer"), &["consumer"]),
            (
                dictgen::InsensitiveStr::Ascii("viewtransfromation"),
                &["viewtransformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorythmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prestigous"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("columnt"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("strawburry"),
                &["strawberry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproduccion"),
                &["reproduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buildd"),
                &["build", "builds"],
            ),
            (dictgen::InsensitiveStr::Ascii("secquence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("pepperin"), &["pepperoni"]),
            (
                dictgen::InsensitiveStr::Ascii("restrucuring"),
                &["restructuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterrpise"),
                &["enterprises"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterrators"), &["iterators"]),
            (dictgen::InsensitiveStr::Ascii("availabled"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("decrytion"), &["decryption"]),
            (dictgen::InsensitiveStr::Ascii("subtitltes"), &["subtitle"]),
            (dictgen::InsensitiveStr::Ascii("probabale"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("hyrbids"), &["hybrids"]),
            (dictgen::InsensitiveStr::Ascii("noboday"), &["nobody"]),
            (
                dictgen::InsensitiveStr::Ascii("initialies"),
                &["initialise", "initializes", "initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concusssion"),
                &["concussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("stormend"), &["stormed"]),
            (dictgen::InsensitiveStr::Ascii("linaer"), &["linear"]),
            (
                dictgen::InsensitiveStr::Ascii("simultaneuous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotition"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unchangeble"),
                &["unchangeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exprimental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loaclization"),
                &["localization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conext"),
                &["context", "connect", "connects"],
            ),
            (dictgen::InsensitiveStr::Ascii("faied"), &["failed", "fade"]),
            (dictgen::InsensitiveStr::Ascii("unknwn"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("bassic"), &["basic"]),
            (
                dictgen::InsensitiveStr::Ascii("multuplied"),
                &["multiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("orchesta"), &["orchestra"]),
            (
                dictgen::InsensitiveStr::Ascii("foreshadowning"),
                &["foreshadowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("esimates"), &["estimates"]),
            (dictgen::InsensitiveStr::Ascii("alllowed"), &["allowed"]),
            (
                dictgen::InsensitiveStr::Ascii("guantamano"),
                &["guantanamo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wokrarounds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoize"), &["customize"]),
            (
                dictgen::InsensitiveStr::Ascii("wildebeast"),
                &["wildebeest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positionning"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("portrayels"), &["portrays"]),
            (dictgen::InsensitiveStr::Ascii("artitrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("disire"), &["desire"]),
            (
                dictgen::InsensitiveStr::Ascii("intantiation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoziable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensionailty"),
                &["dimensionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("jusify"), &["justify"]),
            (
                dictgen::InsensitiveStr::Ascii("expoitation"),
                &["exploitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscritptions"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("arive"), &["arrive"]),
            (
                dictgen::InsensitiveStr::Ascii("beuraucrats"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("durind"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("turle"), &["turtle"]),
            (
                dictgen::InsensitiveStr::Ascii("teoretically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ephememeris"),
                &["ephemeris"],
            ),
            (dictgen::InsensitiveStr::Ascii("explaine"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("evaluatate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("sohpomore"), &["sophomore"]),
            (dictgen::InsensitiveStr::Ascii("thatnkful"), &["thankful"]),
            (dictgen::InsensitiveStr::Ascii("plaback"), &["playback"]),
            (
                dictgen::InsensitiveStr::Ascii("progresss"),
                &["progresses", "progress"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administation"),
                &["administration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenricating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vikingos"), &["vikings"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalite"),
                &["nationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectopn"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("culprint"), &["culprit"]),
            (
                dictgen::InsensitiveStr::Ascii("produktion"),
                &["production"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pratically"),
                &["practically"],
            ),
            (dictgen::InsensitiveStr::Ascii("sharpenss"), &["sharpness"]),
            (
                dictgen::InsensitiveStr::Ascii("conencting"),
                &["connecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commnication"),
                &["communication"],
            ),
            (dictgen::InsensitiveStr::Ascii("ramdomise"), &["randomise"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterperting"),
                &["misinterpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriminated"),
                &["discriminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("embeedders"), &["embedders"]),
            (
                dictgen::InsensitiveStr::Ascii("dcoumented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseudically"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("conflcits"), &["conflicts"]),
            (dictgen::InsensitiveStr::Ascii("elaspe"), &["elapse"]),
            (dictgen::InsensitiveStr::Ascii("coalleed"), &["coalesced"]),
            (dictgen::InsensitiveStr::Ascii("slanguages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("momomentum"), &["momentum"]),
            (dictgen::InsensitiveStr::Ascii("cilinder"), &["cylinder"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessecarrily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("penison"), &["pension"]),
            (dictgen::InsensitiveStr::Ascii("asnyc"), &["async"]),
            (
                dictgen::InsensitiveStr::Ascii("speshally"),
                &["specially", "especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tailgateing"),
                &["tailgating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintensionally"),
                &["unintentionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sihlouette"),
                &["silhouette"],
            ),
            (dictgen::InsensitiveStr::Ascii("equivlant"), &["equivalent"]),
            (
                dictgen::InsensitiveStr::Ascii("abstructly"),
                &["abstractly"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuillds"), &["rebuilds"]),
            (dictgen::InsensitiveStr::Ascii("verry"), &["very"]),
            (
                dictgen::InsensitiveStr::Ascii("chandalier"),
                &["chandelier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocorrekt"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("jewlery"), &["jewelry"]),
            (dictgen::InsensitiveStr::Ascii("trabajdo"), &["trabajo"]),
            (dictgen::InsensitiveStr::Ascii("suggessts"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("equaly"), &["equally"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondnt"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("checlists"), &["checklists"]),
            (
                dictgen::InsensitiveStr::Ascii("decstiptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vicotrious"),
                &["victorious"],
            ),
            (dictgen::InsensitiveStr::Ascii("staisfy"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("midifier"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("proclaimation"),
                &["proclamation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deviiations"),
                &["deviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filmamkers"),
                &["filmmakers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmittions"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoically"),
                &["canonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comprehnsive"),
                &["comprehensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interepreters"),
                &["interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crach"),
                &["crash", "crack", "coach", "crutch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playthorugh"),
                &["playthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ditribution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("rythym"), &["rhythm"]),
            (
                dictgen::InsensitiveStr::Ascii("aganda"),
                &["agenda", "uganda"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unpleaseant"),
                &["unpleasant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("euqivalent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquisitons"),
                &["acquisitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("locallly"), &["locally"]),
            (dictgen::InsensitiveStr::Ascii("somehitng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("requrements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyrption"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("integraal"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("specailly"), &["specially"]),
            (
                dictgen::InsensitiveStr::Ascii("franchines"),
                &["franchises"],
            ),
            (dictgen::InsensitiveStr::Ascii("elelements"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("greviances"),
                &["grievances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conceptionally"),
                &["conceptually"],
            ),
            (dictgen::InsensitiveStr::Ascii("srtructure"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("excecutives"),
                &["executives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acompanies"),
                &["accompanies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("erronuosly"),
                &["erroneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("insde"), &["inside"]),
            (dictgen::InsensitiveStr::Ascii("chisiling"), &["chiseling"]),
            (dictgen::InsensitiveStr::Ascii("selekting"), &["selecting"]),
            (dictgen::InsensitiveStr::Ascii("calucates"), &["calculates"]),
            (
                dictgen::InsensitiveStr::Ascii("infrequancies"),
                &["infrequencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retirevers"),
                &["retrievers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("platfromers"),
                &["platformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcors"), &["correctors"]),
            (
                dictgen::InsensitiveStr::Ascii("internationalizin"),
                &["internationalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("paroject"), &["project"]),
            (dictgen::InsensitiveStr::Ascii("muniches"), &["munchies"]),
            (
                dictgen::InsensitiveStr::Ascii("hypocondriac"),
                &["hypochondriac"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significanty"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("myselve"), &["myself"]),
            (dictgen::InsensitiveStr::Ascii("criticat"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpecetdly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cliente"),
                &["client", "clientele"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wether"),
                &["weather", "whether"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgruntld"),
                &["disgruntled"],
            ),
            (dictgen::InsensitiveStr::Ascii("newances"), &["nuances"]),
            (dictgen::InsensitiveStr::Ascii("rectange"), &["rectangle"]),
            (dictgen::InsensitiveStr::Ascii("cyprtic"), &["cryptic"]),
            (dictgen::InsensitiveStr::Ascii("aache"), &["ache", "cache"]),
            (
                dictgen::InsensitiveStr::Ascii("represendations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("balitmore"), &["baltimore"]),
            (
                dictgen::InsensitiveStr::Ascii("temperametal"),
                &["temperamental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scintiallation"),
                &["scintillation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporraies"),
                &["temporaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("arguemnets"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("surounding"),
                &["surrounding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsitencies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pervention"),
                &["prevention"],
            ),
            (dictgen::InsensitiveStr::Ascii("voyours"), &["voyeurs"]),
            (dictgen::InsensitiveStr::Ascii("certanity"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("cuased"), &["caused"]),
            (
                dictgen::InsensitiveStr::Ascii("invulnurable"),
                &["invulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbrirtary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("stlying"), &["styling"]),
            (dictgen::InsensitiveStr::Ascii("srtifact"), &["artifact"]),
            (
                dictgen::InsensitiveStr::Ascii("tranvers"),
                &["traverse", "transverse"],
            ),
            (dictgen::InsensitiveStr::Ascii("undeline"), &["underline"]),
            (
                dictgen::InsensitiveStr::Ascii("catergorizing"),
                &["categorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("gallleries"), &["galleries"]),
            (dictgen::InsensitiveStr::Ascii("wrappping"), &["wrapping"]),
            (dictgen::InsensitiveStr::Ascii("knietic"), &["kinetic"]),
            (
                dictgen::InsensitiveStr::Ascii("sloughtering"),
                &["slaughtering"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticis"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("speeed"), &["speed"]),
            (dictgen::InsensitiveStr::Ascii("prequles"), &["prequels"]),
            (
                dictgen::InsensitiveStr::Ascii("distirbutive"),
                &["distributive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weightlfiting"),
                &["weightlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indiffernce"),
                &["indifference"],
            ),
            (dictgen::InsensitiveStr::Ascii("desepears"), &["disappears"]),
            (dictgen::InsensitiveStr::Ascii("clouflare"), &["cloudflare"]),
            (
                dictgen::InsensitiveStr::Ascii("millenialism"),
                &["millennialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("arlready"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("precedure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("begative"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("exressed"), &["expressed"]),
            (
                dictgen::InsensitiveStr::Ascii("konfirmation"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("enver"), &["never"]),
            (dictgen::InsensitiveStr::Ascii("ciotees"), &["coyotes"]),
            (dictgen::InsensitiveStr::Ascii("abtracter"), &["abstracter"]),
            (
                dictgen::InsensitiveStr::Ascii("redefintion"),
                &["redefinition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportion"),
                &["incorporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("animatiors"),
                &["animators", "animations"],
            ),
            (dictgen::InsensitiveStr::Ascii("invalied"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("ellipsical"),
                &["elliptical"],
            ),
            (dictgen::InsensitiveStr::Ascii("unkonwns"), &["unknowns"]),
            (dictgen::InsensitiveStr::Ascii("uplifitng"), &["uplifting"]),
            (dictgen::InsensitiveStr::Ascii("numberrs"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("indiquate"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("overpad"), &["overpaid"]),
            (
                dictgen::InsensitiveStr::Ascii("manifestating"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mormalize"), &["normalize"]),
            (
                dictgen::InsensitiveStr::Ascii("automicly"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("beling"), &["belong"]),
            (dictgen::InsensitiveStr::Ascii("clippin"), &["clipping"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistences"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolide"),
                &["consolidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("arameters"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("supporing"), &["supporting"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapolerat"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimenstion"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("expexted"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("bakcup"), &["backup"]),
            (dictgen::InsensitiveStr::Ascii("parmaeter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("dependand"), &["dependant"]),
            (
                dictgen::InsensitiveStr::Ascii("agricultual"),
                &["agricultural"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffes"), &["buffers"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinde"),
                &["combined", "combine"],
            ),
            (dictgen::InsensitiveStr::Ascii("regardles"), &["regardless"]),
            (
                dictgen::InsensitiveStr::Ascii("recognzied"),
                &["recognised", "recognized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrerenceing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcents"),
                &["announces", "announcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimneters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("secondaty"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("suppors"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalizer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("headshoot"), &["headshot"]),
            (dictgen::InsensitiveStr::Ascii("obtainig"), &["obtaining"]),
            (
                dictgen::InsensitiveStr::Ascii("entite"),
                &["entire", "entity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paranomral"),
                &["paranormal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptiive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlightented"),
                &["enlightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("comonents"), &["components"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotioated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("regstry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("trnaslate"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("necessariily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("preferr"), &["prefer"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstanes"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("replacd"), &["replaced"]),
            (dictgen::InsensitiveStr::Ascii("iniative"), &["initiative"]),
            (dictgen::InsensitiveStr::Ascii("abiguity"), &["ambiguity"]),
            (
                dictgen::InsensitiveStr::Ascii("uncetainty"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("audiance"), &["audience"]),
            (
                dictgen::InsensitiveStr::Ascii("angosticism"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheistc"), &["atheistic"]),
            (
                dictgen::InsensitiveStr::Ascii("unregsitered"),
                &["unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulaotrs"),
                &["regulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("radiobuttion"),
                &["radiobutton"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masoginistic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegociation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("habbits"),
                &["habits", "rabbits", "hobbits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cahed"),
                &["cached", "ached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemtentation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("desitions"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("configuree"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("harassement"),
                &["harassment"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalasing"), &["coalescing"]),
            (
                dictgen::InsensitiveStr::Ascii("controle"),
                &["control", "controlled", "controller", "controls"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpolite"), &["impolite"]),
            (dictgen::InsensitiveStr::Ascii("complict"), &["complicit"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificeing"),
                &["sacrificing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepretable"),
                &["interpretable"],
            ),
            (dictgen::InsensitiveStr::Ascii("spainish"), &["spanish"]),
            (
                dictgen::InsensitiveStr::Ascii("indocrtination"),
                &["indoctrination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosessed"),
                &["processed", "professed", "possessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("priciple"), &["principle"]),
            (dictgen::InsensitiveStr::Ascii("provdide"), &["provide"]),
            (
                dictgen::InsensitiveStr::Ascii("beuatifully"),
                &["beautifully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interripting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorsment"),
                &["endorsement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("notifers"), &["notifiers"]),
            (dictgen::InsensitiveStr::Ascii("expexct"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("resolvemos"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubstantiaed"),
                &["unsubstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("methons"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("levitatin"),
                &["levitating", "levitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritse"),
                &["priorities", "prioritise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exchaustive"),
                &["exhaustive"],
            ),
            (dictgen::InsensitiveStr::Ascii("rlse"), &["else"]),
            (dictgen::InsensitiveStr::Ascii("torando"), &["tornado"]),
            (
                dictgen::InsensitiveStr::Ascii("compostion"),
                &["composition"],
            ),
            (dictgen::InsensitiveStr::Ascii("misslies"), &["missiles"]),
            (
                dictgen::InsensitiveStr::Ascii("definitieve"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("supsects"), &["suspects"]),
            (dictgen::InsensitiveStr::Ascii("macpakge"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("ingore"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("overrie"),
                &["override", "ovary"],
            ),
            (dictgen::InsensitiveStr::Ascii("imcoming"), &["incoming"]),
            (dictgen::InsensitiveStr::Ascii("poninted"), &["pointed"]),
            (dictgen::InsensitiveStr::Ascii("ananlyser"), &["analyser"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaites"),
                &["reallocates"],
            ),
            (dictgen::InsensitiveStr::Ascii("sufferd"), &["suffered"]),
            (dictgen::InsensitiveStr::Ascii("mecanic"), &["mechanic"]),
            (dictgen::InsensitiveStr::Ascii("changsets"), &["changesets"]),
            (
                dictgen::InsensitiveStr::Ascii("atrack"),
                &["attract", "attack"],
            ),
            (dictgen::InsensitiveStr::Ascii("boostrap"), &["bootstrap"]),
            (
                dictgen::InsensitiveStr::Ascii("susceding"),
                &["seceding", "succeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definetley"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgrouds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("blcokers"), &["blockers"]),
            (
                dictgen::InsensitiveStr::Ascii("classef"),
                &["classdef", "classed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepdendantly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("burbon"), &["bourbon"]),
            (
                dictgen::InsensitiveStr::Ascii("inspiraton"),
                &["inspiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("reccurent"), &["recurrent"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelly"),
                &["parallelly"],
            ),
            (dictgen::InsensitiveStr::Ascii("spezify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("symbls"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("coult"), &["could"]),
            (dictgen::InsensitiveStr::Ascii("anaylst"), &["analyst"]),
            (
                dictgen::InsensitiveStr::Ascii("temportaries"),
                &["temporaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("invaderats"), &["invaders"]),
            (dictgen::InsensitiveStr::Ascii("explining"), &["explaining"]),
            (dictgen::InsensitiveStr::Ascii("wrnagler"), &["wrangler"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenatal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meskeetos"),
                &["mosquitoes", "mosquitos"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiancing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minusculy"),
                &["minusculely", "minuscule"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categorizin"),
                &["categorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("stablize"), &["stabilize"]),
            (dictgen::InsensitiveStr::Ascii("schaeduled"), &["scheduled"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitalization"),
                &["reinitialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementeed"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reveivers"),
                &["receivers", "reviewers", "revivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("contrl"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("mesurment"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("windos"),
                &["windows", "window"],
            ),
            (dictgen::InsensitiveStr::Ascii("filpping"), &["flipping"]),
            (dictgen::InsensitiveStr::Ascii("mapp"), &["map"]),
            (dictgen::InsensitiveStr::Ascii("mentiong"), &["mentioning"]),
            (dictgen::InsensitiveStr::Ascii("delpoy"), &["deploy"]),
            (dictgen::InsensitiveStr::Ascii("residin"), &["residing"]),
            (dictgen::InsensitiveStr::Ascii("unusuable"), &["unusable"]),
            (dictgen::InsensitiveStr::Ascii("interuppt"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("termonologies"),
                &["terminologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("thryoid"), &["thyroid"]),
            (
                dictgen::InsensitiveStr::Ascii("distinktion"),
                &["distinction"],
            ),
            (dictgen::InsensitiveStr::Ascii("bigining"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("momen"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplicites"),
                &["multiplicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("bunble"), &["bundle"]),
            (dictgen::InsensitiveStr::Ascii("abrutply"), &["abruptly"]),
            (dictgen::InsensitiveStr::Ascii("proccesor"), &["processor"]),
            (
                dictgen::InsensitiveStr::Ascii("compromosing"),
                &["compromising"],
            ),
            (dictgen::InsensitiveStr::Ascii("upstread"), &["upstream"]),
            (
                dictgen::InsensitiveStr::Ascii("follfowinwg"),
                &["following"],
            ),
            (dictgen::InsensitiveStr::Ascii("minerales"), &["minerals"]),
            (dictgen::InsensitiveStr::Ascii("anually"), &["annually"]),
            (dictgen::InsensitiveStr::Ascii("satistied"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("demosntrating"),
                &["demonstrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignortig"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("muscels"),
                &["mussels", "muscles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaibility"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eeverything"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapolite"),
                &["extrapolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responces"),
                &["response", "responses"],
            ),
            (dictgen::InsensitiveStr::Ascii("proflier"), &["profiler"]),
            (dictgen::InsensitiveStr::Ascii("standart"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("alaready"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("heirarchical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beautifuly"),
                &["beautifully"],
            ),
            (dictgen::InsensitiveStr::Ascii("receievd"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("dishonet"), &["dishonesty"]),
            (
                dictgen::InsensitiveStr::Ascii("stickyness"),
                &["stickiness"],
            ),
            (dictgen::InsensitiveStr::Ascii("rasiing"), &["raising"]),
            (dictgen::InsensitiveStr::Ascii("spefiying"), &["specifying"]),
            (
                dictgen::InsensitiveStr::Ascii("neturality"),
                &["neutrality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronizes"),
                &["synchronizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsidezed"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaalues"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("deimiter"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("opemess"), &["openness"]),
            (dictgen::InsensitiveStr::Ascii("mountan"), &["mountain"]),
            (dictgen::InsensitiveStr::Ascii("memership"), &["membership"]),
            (
                dictgen::InsensitiveStr::Ascii("positionining"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requris"),
                &["require", "requires"],
            ),
            (dictgen::InsensitiveStr::Ascii("enagement"), &["engagement"]),
            (dictgen::InsensitiveStr::Ascii("mofified"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("autheniticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("reflectd"), &["reflected"]),
            (dictgen::InsensitiveStr::Ascii("britian"), &["britain"]),
            (
                dictgen::InsensitiveStr::Ascii("noifications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operatonal"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("pilgram"), &["pilgrim"]),
            (
                dictgen::InsensitiveStr::Ascii("acumalates"),
                &["accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("orign"), &["origin"]),
            (dictgen::InsensitiveStr::Ascii("penduluum"), &["pendulum"]),
            (
                dictgen::InsensitiveStr::Ascii("territoriella"),
                &["territorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("fasterner"), &["fastener"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterprate"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylnders"), &["cylinders"]),
            (dictgen::InsensitiveStr::Ascii("idealy"), &["ideally"]),
            (
                dictgen::InsensitiveStr::Ascii("integerated"),
                &["integrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recompuuted"),
                &["recomputed"],
            ),
            (dictgen::InsensitiveStr::Ascii("joing"), &["joining"]),
            (dictgen::InsensitiveStr::Ascii("capaciors"), &["capacitors"]),
            (
                dictgen::InsensitiveStr::Ascii("paraphanalia"),
                &["paraphernalia"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuctions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("reprihensible"),
                &["reprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transormed"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distintion"),
                &["distinction"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparare"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("inesrted"), &["inserted"]),
            (dictgen::InsensitiveStr::Ascii("vesioned"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("expest"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmanouverable"),
                &["unmaneuverable", "unmanoeuvrable"],
            ),
            (dictgen::InsensitiveStr::Ascii("avalaible"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("funstions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("symmetetric"),
                &["symmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("veamant"), &["vehement"]),
            (
                dictgen::InsensitiveStr::Ascii("convencionally"),
                &["conventionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("recipents"), &["recipients"]),
            (dictgen::InsensitiveStr::Ascii("proedure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("apapted"), &["adapted"]),
            (dictgen::InsensitiveStr::Ascii("paraneter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("indimidating"),
                &["intimidating"],
            ),
            (dictgen::InsensitiveStr::Ascii("zukeenei"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("speficers"), &["specifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("differantiates"),
                &["differentiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("organiszed"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("pyramind"), &["pyramid"]),
            (dictgen::InsensitiveStr::Ascii("scetch"), &["sketch"]),
            (dictgen::InsensitiveStr::Ascii("physicits"), &["physicist"]),
            (
                dictgen::InsensitiveStr::Ascii("pakistanais"),
                &["pakistani"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlizations"),
                &["generalizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegoitation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("filtypes"), &["filetypes"]),
            (
                dictgen::InsensitiveStr::Ascii("photoshopd"),
                &["photoshopped"],
            ),
            (dictgen::InsensitiveStr::Ascii("framin"), &["framing"]),
            (
                dictgen::InsensitiveStr::Ascii("persuassive"),
                &["persuasive"],
            ),
            (dictgen::InsensitiveStr::Ascii("scolled"), &["scrolled"]),
            (
                dictgen::InsensitiveStr::Ascii("reaaranges"),
                &["rearranges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infalliable"),
                &["infallible"],
            ),
            (dictgen::InsensitiveStr::Ascii("teachnigs"), &["teachings"]),
            (dictgen::InsensitiveStr::Ascii("direcors"), &["directors"]),
            (
                dictgen::InsensitiveStr::Ascii("countinueq"),
                &["continueq", "continue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passionatelly"),
                &["passionately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parokeality"),
                &["parochiality"],
            ),
            (dictgen::InsensitiveStr::Ascii("snwoballs"), &["snowballs"]),
            (dictgen::InsensitiveStr::Ascii("exipres"), &["expires"]),
            (dictgen::InsensitiveStr::Ascii("posiitive"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("realiy"), &["really"]),
            (
                dictgen::InsensitiveStr::Ascii("substituations"),
                &["substitutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australiaan"),
                &["australians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarantee"),
                &["guarantee", "quarantine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidentlay"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("corospond"), &["correspond"]),
            (
                dictgen::InsensitiveStr::Ascii("contextura"),
                &["contextual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentated"),
                &["concentrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("melborne"), &["melbourne"]),
            (dictgen::InsensitiveStr::Ascii("theoeretic"), &["theoretic"]),
            (
                dictgen::InsensitiveStr::Ascii("decryptation"),
                &["decryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premanently"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("locting"),
                &["locating", "locking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("planeswalkr"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("swithes"), &["switches"]),
            (dictgen::InsensitiveStr::Ascii("incremenet"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("midevil"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("siply"), &["simply"]),
            (dictgen::InsensitiveStr::Ascii("prioroty"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("decrytpion"),
                &["decryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiations"),
                &["initializations", "initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpolaed"),
                &["interpolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geometricians"),
                &["geometers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interviewd"),
                &["interviewed"],
            ),
            (dictgen::InsensitiveStr::Ascii("proove"), &["prove"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarcical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilications"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("runable"), &["runnable"]),
            (
                dictgen::InsensitiveStr::Ascii("usre"),
                &["sure", "user", "use"],
            ),
            (dictgen::InsensitiveStr::Ascii("startign"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("symphatizers"),
                &["sympathizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indireclty"),
                &["indirectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("weighys"), &["weights"]),
            (
                dictgen::InsensitiveStr::Ascii("otimizations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("earthquack"),
                &["earthquake"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referentes"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handwritting"),
                &["handwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapprouve"),
                &["disapprove"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthimically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheved"), &["achieved"]),
            (
                dictgen::InsensitiveStr::Ascii("athiest"),
                &["atheist", "athirst"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cantacting"),
                &["contacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precomupte"),
                &["precompute"],
            ),
            (dictgen::InsensitiveStr::Ascii("presisted"), &["persisted"]),
            (
                dictgen::InsensitiveStr::Ascii("replacas"),
                &["replicas", "replaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("protable"), &["portable"]),
            (dictgen::InsensitiveStr::Ascii("stoages"), &["storages"]),
            (dictgen::InsensitiveStr::Ascii("shiiping"), &["shipping"]),
            (
                dictgen::InsensitiveStr::Ascii("undergroud"),
                &["underground"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheving"), &["achieving"]),
            (
                dictgen::InsensitiveStr::Ascii("inexplicaby"),
                &["inexplicably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitilaize"),
                &["reinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("allthough"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("regconiser"),
                &["recogniser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgradded"),
                &["downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("soltuions"), &["solutions"]),
            (dictgen::InsensitiveStr::Ascii("succss"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("attrinute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("hirarcical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetcedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccur"),
                &["recur", "reoccur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchindise"),
                &["merchandise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randomzied"),
                &["randomized"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolune"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("responsiblities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("thottle"), &["throttle"]),
            (
                dictgen::InsensitiveStr::Ascii("phramaceutical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissasociation"),
                &["disassociation"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkd"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("relevat"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("beginnner"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("suddently"), &["suddenly"]),
            (dictgen::InsensitiveStr::Ascii("supportted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("amernia"), &["armenia"]),
            (dictgen::InsensitiveStr::Ascii("authoise"), &["authorise"]),
            (
                dictgen::InsensitiveStr::Ascii("streamlinin"),
                &["streamlining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definatedly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("librarries"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("ocupiers"), &["occupiers"]),
            (dictgen::InsensitiveStr::Ascii("apperance"), &["appearance"]),
            (
                dictgen::InsensitiveStr::Ascii("devirtualiation"),
                &["devirtualisation", "devirtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("assistendo"), &["assisted"]),
            (dictgen::InsensitiveStr::Ascii("wthout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("reprodicible"),
                &["reproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("spwan"), &["spawn"]),
            (dictgen::InsensitiveStr::Ascii("conversare"), &["converse"]),
            (
                dictgen::InsensitiveStr::Ascii("transpantly"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("levae"), &["leave", "levee"]),
            (dictgen::InsensitiveStr::Ascii("somone"), &["someone"]),
            (dictgen::InsensitiveStr::Ascii("newmatic"), &["pneumatic"]),
            (
                dictgen::InsensitiveStr::Ascii("constitutiei"),
                &["constitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periperals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificaiton"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("maccros"), &["macros"]),
            (dictgen::InsensitiveStr::Ascii("produt"), &["product"]),
            (dictgen::InsensitiveStr::Ascii("rnuning"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("teraform"), &["terraform"]),
            (
                dictgen::InsensitiveStr::Ascii("hospitolized"),
                &["hospitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhing"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contens"),
                &["contents", "contains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequenzers"),
                &["sequencers"],
            ),
            (dictgen::InsensitiveStr::Ascii("delimiiter"), &["delimiter"]),
            (
                dictgen::InsensitiveStr::Ascii("inconistently"),
                &["inconsistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("molestarle"), &["molester"]),
            (
                dictgen::InsensitiveStr::Ascii("instersection"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensonalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temprary"),
                &["temporary", "temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("attenion"), &["attention"]),
            (dictgen::InsensitiveStr::Ascii("teorems"), &["theorems"]),
            (
                dictgen::InsensitiveStr::Ascii("unregester"),
                &["unregister"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preventivno"),
                &["prevention"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcalate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("previosly"), &["previously"]),
            (
                dictgen::InsensitiveStr::Ascii("narcissisim"),
                &["narcissism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulneability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabint"), &["cabinet"]),
            (
                dictgen::InsensitiveStr::Ascii("positivitey"),
                &["positivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpetced"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumptuos"),
                &["presumptuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicats"),
                &["indicators", "indicates", "indicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("entilted"), &["entitled"]),
            (
                dictgen::InsensitiveStr::Ascii("considerarle"),
                &["considerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activiti"),
                &["activity", "activities"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqares"), &["squares"]),
            (dictgen::InsensitiveStr::Ascii("atalogs"), &["catalogs"]),
            (dictgen::InsensitiveStr::Ascii("parameger"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("universitites"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("seacrhed"), &["searched"]),
            (
                dictgen::InsensitiveStr::Ascii("coppyrights"),
                &["copyrights"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdots"), &["anecdotes"]),
            (dictgen::InsensitiveStr::Ascii("prmopting"), &["prompting"]),
            (dictgen::InsensitiveStr::Ascii("inut"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("criticos"), &["critics"]),
            (
                dictgen::InsensitiveStr::Ascii("mehodically"),
                &["methodically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutnerplay"),
                &["counterplay"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucesfuly"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhhancement"),
                &["enhancement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrije"),
                &["industries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sopporters"),
                &["supporters"],
            ),
            (dictgen::InsensitiveStr::Ascii("surrrounds"), &["surrounds"]),
            (
                dictgen::InsensitiveStr::Ascii("practiclly"),
                &["practically"],
            ),
            (dictgen::InsensitiveStr::Ascii("returming"), &["returning"]),
            (
                dictgen::InsensitiveStr::Ascii("documentaitons"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("acchieves"), &["achieves"]),
            (dictgen::InsensitiveStr::Ascii("litlle"), &["little"]),
            (
                dictgen::InsensitiveStr::Ascii("explantion"),
                &["explanation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conveted"), &["converted"]),
            (
                dictgen::InsensitiveStr::Ascii("subsysytems"),
                &["subsystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("sprinkeld"), &["sprinkled"]),
            (dictgen::InsensitiveStr::Ascii("approachs"), &["approaches"]),
            (
                dictgen::InsensitiveStr::Ascii("appropaite"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("malaysain"), &["malaysian"]),
            (dictgen::InsensitiveStr::Ascii("objtain"), &["obtain"]),
            (
                dictgen::InsensitiveStr::Ascii("porportinal"),
                &["proportional"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyeballls"), &["eyeballs"]),
            (dictgen::InsensitiveStr::Ascii("exeucted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("nanosenconds"),
                &["nanoseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("frightend"), &["frightened"]),
            (
                dictgen::InsensitiveStr::Ascii("intepretators"),
                &["interpreters"],
            ),
            (dictgen::InsensitiveStr::Ascii("toubling"), &["troubling"]),
            (dictgen::InsensitiveStr::Ascii("fictionnal"), &["fictional"]),
            (dictgen::InsensitiveStr::Ascii("dotrmund"), &["dortmund"]),
            (
                dictgen::InsensitiveStr::Ascii("callenging"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrigation"),
                &["interrogation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timestampes"),
                &["timestamps"],
            ),
            (dictgen::InsensitiveStr::Ascii("egine"), &["engine"]),
            (dictgen::InsensitiveStr::Ascii("truncte"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("coommands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("porven"), &["proven"]),
            (
                dictgen::InsensitiveStr::Ascii("pressenters"),
                &["presenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represtatation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("denmmark"), &["denmark"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaion"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reviewes"), &["reviewers"]),
            (dictgen::InsensitiveStr::Ascii("debuger"), &["debugger"]),
            (dictgen::InsensitiveStr::Ascii("refering"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("fianls"), &["finals"]),
            (
                dictgen::InsensitiveStr::Ascii("ploynomial"),
                &["polynomial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimers"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimesnion"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("penultimante"),
                &["penultimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimaneter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corisponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preject"),
                &["prefect", "project"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workaronuds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("morining"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("maesured"), &["measured"]),
            (
                dictgen::InsensitiveStr::Ascii("temperture"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememebrs"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("erturning"), &["returning"]),
            (dictgen::InsensitiveStr::Ascii("otpimal"), &["optimal"]),
            (
                dictgen::InsensitiveStr::Ascii("skatebaord"),
                &["skateboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unanymously"),
                &["unanimously"],
            ),
            (dictgen::InsensitiveStr::Ascii("treatement"), &["treatment"]),
            (
                dictgen::InsensitiveStr::Ascii("determinanti"),
                &["determination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brughtness"),
                &["brightness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invaulable"),
                &["invaluable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestare"),
                &["protesters"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiails"), &["initials"]),
            (dictgen::InsensitiveStr::Ascii("allievate"), &["alleviate"]),
            (dictgen::InsensitiveStr::Ascii("ismas"), &["isthmus"]),
            (dictgen::InsensitiveStr::Ascii("discusson"), &["discussion"]),
            (dictgen::InsensitiveStr::Ascii("reveiws"), &["reviews"]),
            (
                dictgen::InsensitiveStr::Ascii("extraversion"),
                &["extroversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("wikpedia"), &["wikipedia"]),
            (dictgen::InsensitiveStr::Ascii("requriring"), &["requiring"]),
            (
                dictgen::InsensitiveStr::Ascii("environemtns"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaetes"), &["creates"]),
            (dictgen::InsensitiveStr::Ascii("socartes"), &["socrates"]),
            (dictgen::InsensitiveStr::Ascii("horisons"), &["horizons"]),
            (
                dictgen::InsensitiveStr::Ascii("pestizides"),
                &["pesticides"],
            ),
            (dictgen::InsensitiveStr::Ascii("changees"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("anceints"), &["ancients"]),
            (dictgen::InsensitiveStr::Ascii("monopoloy"), &["monopoly"]),
            (
                dictgen::InsensitiveStr::Ascii("containters"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("varibables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("ndoes"), &["nodes"]),
            (dictgen::InsensitiveStr::Ascii("spaeces"), &["spaces"]),
            (
                dictgen::InsensitiveStr::Ascii("evaliating"),
                &["evaluating"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuaised"), &["visualised"]),
            (dictgen::InsensitiveStr::Ascii("tepmlate"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("substanial"),
                &["substantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("consunmers"), &["consumers"]),
            (dictgen::InsensitiveStr::Ascii("exelent"), &["excellent"]),
            (dictgen::InsensitiveStr::Ascii("toogles"), &["toggles"]),
            (dictgen::InsensitiveStr::Ascii("jericoh"), &["jericho"]),
            (dictgen::InsensitiveStr::Ascii("gouping"), &["grouping"]),
            (
                dictgen::InsensitiveStr::Ascii("apparantely"),
                &["apparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("porves"), &["proves"]),
            (dictgen::InsensitiveStr::Ascii("augumented"), &["augmented"]),
            (dictgen::InsensitiveStr::Ascii("speficieir"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("downgradding"),
                &["downgrading"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinosuars"), &["dinosaurs"]),
            (
                dictgen::InsensitiveStr::Ascii("merchandice"),
                &["merchandise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immeditaly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("startting"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("itialiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vertabrays"),
                &["vertebraes"],
            ),
            (dictgen::InsensitiveStr::Ascii("flacoured"), &["flavoured"]),
            (dictgen::InsensitiveStr::Ascii("dicussed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("clincis"), &["clinics"]),
            (dictgen::InsensitiveStr::Ascii("encypts"), &["encrypts"]),
            (
                dictgen::InsensitiveStr::Ascii("manifiesting"),
                &["manifesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknwns"), &["unknowns"]),
            (dictgen::InsensitiveStr::Ascii("featred"), &["featured"]),
            (dictgen::InsensitiveStr::Ascii("similiar"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("subrutine"), &["subroutine"]),
            (
                dictgen::InsensitiveStr::Ascii("supspecting"),
                &["suspecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orietations"),
                &["orientations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devestated"),
                &["devastated"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicti"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("nimphlike"), &["nymphlike"]),
            (dictgen::InsensitiveStr::Ascii("wokrbook"), &["workbook"]),
            (dictgen::InsensitiveStr::Ascii("cheeta"), &["cheetah"]),
            (dictgen::InsensitiveStr::Ascii("euqation"), &["equation"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretares"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("keywoard"),
                &["keyword", "keyboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabule"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preprocesor"),
                &["preprocessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("derth"), &["dearth"]),
            (
                dictgen::InsensitiveStr::Ascii("compadible"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("desigerns"), &["designers"]),
            (
                dictgen::InsensitiveStr::Ascii("preffering"),
                &["preferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("untrasposed"),
                &["untransposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("lisitng"), &["listing"]),
            (dictgen::InsensitiveStr::Ascii("backtics"), &["backticks"]),
            (dictgen::InsensitiveStr::Ascii("slicings"), &["slices"]),
            (
                dictgen::InsensitiveStr::Ascii("largets"),
                &["largest", "targets"],
            ),
            (dictgen::InsensitiveStr::Ascii("redundat"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("inicialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("portguese"), &["portuguese"]),
            (
                dictgen::InsensitiveStr::Ascii("ideosyncracy"),
                &["idiosyncrasy"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdule"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("overridee"), &["override"]),
            (dictgen::InsensitiveStr::Ascii("wief"), &["wife"]),
            (dictgen::InsensitiveStr::Ascii("arrey"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("allocaites"), &["allocates"]),
            (dictgen::InsensitiveStr::Ascii("chekcsum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("claculator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("miinimize"), &["minimize"]),
            (
                dictgen::InsensitiveStr::Ascii("anesthetia"),
                &["anesthesia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consituted"),
                &["constituted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tangentialy"),
                &["tangentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("contists"), &["consists"]),
            (
                dictgen::InsensitiveStr::Ascii("rocesses"),
                &["processes", "recesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clasas"),
                &["class", "classes"],
            ),
            (dictgen::InsensitiveStr::Ascii("informarla"), &["informal"]),
            (dictgen::InsensitiveStr::Ascii("compairing"), &["comparing"]),
            (
                dictgen::InsensitiveStr::Ascii("mannerisims"),
                &["mannerisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("queires"), &["queries"]),
            (dictgen::InsensitiveStr::Ascii("provdided"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("neigboring"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("incuding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("desensitied"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curruptable"),
                &["corruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("tahnkyou"), &["thankyou"]),
            (
                dictgen::InsensitiveStr::Ascii("receptionsist"),
                &["receptionist"],
            ),
            (dictgen::InsensitiveStr::Ascii("folow"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("exapmles"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("esponded"),
                &["responded", "desponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contructions"),
                &["contractions", "constructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclamantions"),
                &["exclamations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("currancies"),
                &["currencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("prorities"), &["priorities"]),
            (
                dictgen::InsensitiveStr::Ascii("applicaitons"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("keynode"), &["keynote"]),
            (
                dictgen::InsensitiveStr::Ascii("flutteryshy"),
                &["fluttershy"],
            ),
            (dictgen::InsensitiveStr::Ascii("colse"), &["close"]),
            (
                dictgen::InsensitiveStr::Ascii("interacsions"),
                &["interactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigleton"), &["singleton"]),
            (
                dictgen::InsensitiveStr::Ascii("utilitising"),
                &["utilising"],
            ),
            (dictgen::InsensitiveStr::Ascii("wanrings"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("memebr"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("billionairs"),
                &["billionaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percintile"),
                &["percentile"],
            ),
            (dictgen::InsensitiveStr::Ascii("togeter"), &["together"]),
            (
                dictgen::InsensitiveStr::Ascii("polymorhed"),
                &["polymorphed"],
            ),
            (dictgen::InsensitiveStr::Ascii("exturdes"), &["extrudes"]),
            (dictgen::InsensitiveStr::Ascii("bieng"), &["being"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celebratings"),
                &["celebrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redifining"),
                &["redefining"],
            ),
            (dictgen::InsensitiveStr::Ascii("probabile"), &["probable"]),
            (
                dictgen::InsensitiveStr::Ascii("unpertubated"),
                &["unperturbed"],
            ),
            (dictgen::InsensitiveStr::Ascii("conmanded"), &["commanded"]),
            (dictgen::InsensitiveStr::Ascii("ariplane"), &["airplane"]),
            (dictgen::InsensitiveStr::Ascii("packgage"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("aretmis"), &["artemis"]),
            (
                dictgen::InsensitiveStr::Ascii("guantanamero"),
                &["guantanamo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envokes"),
                &["evokes", "invokes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bottomborde"),
                &["bottomborder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coursefork"),
                &["coursework"],
            ),
            (dictgen::InsensitiveStr::Ascii("buisness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instroduced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internetors"),
                &["internets"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorned"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("cliffhangar"),
                &["cliffhanger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environements"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sovereignity"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("assualts"), &["assaults"]),
            (dictgen::InsensitiveStr::Ascii("cahined"), &["chained"]),
            (dictgen::InsensitiveStr::Ascii("specfiies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("negoziating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecetd"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("tessleate"), &["tessellate"]),
            (dictgen::InsensitiveStr::Ascii("yuor"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("organsies"), &["organises"]),
            (dictgen::InsensitiveStr::Ascii("monthes"), &["months"]),
            (
                dictgen::InsensitiveStr::Ascii("realatively"),
                &["relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progresssing"),
                &["progressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependcy"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("curruent"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("satelite"), &["satellite"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptorts"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("siganl"), &["signal"]),
            (
                dictgen::InsensitiveStr::Ascii("enthisiast"),
                &["enthusiast"],
            ),
            (dictgen::InsensitiveStr::Ascii("leaneantly"), &["leniently"]),
            (
                dictgen::InsensitiveStr::Ascii("experamentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwoord"),
                &["underworld"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simetrical"),
                &["symmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("cointrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("registeries"),
                &["registers", "registries"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpeend"), &["depend"]),
            (
                dictgen::InsensitiveStr::Ascii("handwirting"),
                &["handwriting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dificult"), &["difficult"]),
            (dictgen::InsensitiveStr::Ascii("durantion"), &["duration"]),
            (dictgen::InsensitiveStr::Ascii("rotaiton"), &["rotation"]),
            (
                dictgen::InsensitiveStr::Ascii("algoritmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("destinationhash"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opeabcration"),
                &["operation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deley"), &["delay"]),
            (
                dictgen::InsensitiveStr::Ascii("confedential"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("hindiusm"), &["hinduism"]),
            (dictgen::InsensitiveStr::Ascii("gltiched"), &["glitched"]),
            (dictgen::InsensitiveStr::Ascii("adjancency"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("limitiing"), &["limiting"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptionss"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("existnace"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("continoue"), &["continue"]),
            (dictgen::InsensitiveStr::Ascii("forula"), &["formula"]),
            (dictgen::InsensitiveStr::Ascii("morroccan"), &["moroccan"]),
            (
                dictgen::InsensitiveStr::Ascii("appriximations"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorightm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("respecatble"),
                &["respectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aymore"), &["anymore"]),
            (
                dictgen::InsensitiveStr::Ascii("outperfomeing"),
                &["outperforming"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneators"), &["generators"]),
            (
                dictgen::InsensitiveStr::Ascii("progammable"),
                &["programmable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cylidrical"),
                &["cylindrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("suble"), &["subtle"]),
            (dictgen::InsensitiveStr::Ascii("asist"), &["assist"]),
            (dictgen::InsensitiveStr::Ascii("assiging"), &["assigning"]),
            (dictgen::InsensitiveStr::Ascii("movign"), &["moving"]),
            (dictgen::InsensitiveStr::Ascii("summeries"), &["summaries"]),
            (dictgen::InsensitiveStr::Ascii("currentry"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("spectifying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("loaclizes"), &["localizes"]),
            (
                dictgen::InsensitiveStr::Ascii("fragmeneting"),
                &["fragmenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelligenly"),
                &["intelligently"],
            ),
            (dictgen::InsensitiveStr::Ascii("xenobalde"), &["xenoblade"]),
            (
                dictgen::InsensitiveStr::Ascii("contamint"),
                &["contaminant"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminent"), &["remnant"]),
            (dictgen::InsensitiveStr::Ascii("precurosr"), &["precursor"]),
            (dictgen::InsensitiveStr::Ascii("spicified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("irrelivant"),
                &["irrelevant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustrato"),
                &["frustration"],
            ),
            (
                dictgen::InsensitiveStr::Unicode("сontainer"),
                &["container"],
            ),
            (dictgen::InsensitiveStr::Ascii("framming"), &["framing"]),
            (dictgen::InsensitiveStr::Ascii("scaning"), &["scanning"]),
            (dictgen::InsensitiveStr::Ascii("impluse"), &["impulse"]),
            (
                dictgen::InsensitiveStr::Ascii("considereis"),
                &["considers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retricted"),
                &["restricted", "retracted"],
            ),
            (dictgen::InsensitiveStr::Ascii("partick"), &["patrick"]),
            (
                dictgen::InsensitiveStr::Ascii("represantatives"),
                &["representatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("destoyed"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("negothiation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("genreic"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("otherwie"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("opearnds"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("aptitudine"), &["aptitude"]),
            (
                dictgen::InsensitiveStr::Ascii("convenciences"),
                &["conveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("zookeini"), &["zucchini"]),
            (dictgen::InsensitiveStr::Ascii("dupliacate"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("deamonises"),
                &["daemonises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addaptation"),
                &["adaptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("committter"), &["committer"]),
            (dictgen::InsensitiveStr::Ascii("hsell"), &["shell"]),
            (
                dictgen::InsensitiveStr::Ascii("progrom"),
                &["pogrom", "program"],
            ),
            (dictgen::InsensitiveStr::Ascii("deleteed"), &["deleted"]),
            (dictgen::InsensitiveStr::Ascii("gruopd"), &["grouped"]),
            (dictgen::InsensitiveStr::Ascii("humanoind"), &["humanoid"]),
            (
                dictgen::InsensitiveStr::Ascii("brotherhod"),
                &["brotherhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caperbility"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("obessions"), &["obsessions"]),
            (
                dictgen::InsensitiveStr::Ascii("tipled"),
                &["tripled", "tipped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiate"),
                &["negotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degenrated"),
                &["degenerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchanage"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("nieghbourhood"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("euorpean"), &["european"]),
            (dictgen::InsensitiveStr::Ascii("rutine"), &["routine"]),
            (
                dictgen::InsensitiveStr::Ascii("sensationail"),
                &["sensational"],
            ),
            (dictgen::InsensitiveStr::Ascii("marekting"), &["marketing"]),
            (dictgen::InsensitiveStr::Ascii("terrorsim"), &["terrorism"]),
            (dictgen::InsensitiveStr::Ascii("arguemtnt"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("syncronous"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catholizism"),
                &["catholicism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agains"),
                &["against", "again"],
            ),
            (dictgen::InsensitiveStr::Ascii("forwwards"), &["forwards"]),
            (dictgen::InsensitiveStr::Ascii("szies"), &["sizes"]),
            (
                dictgen::InsensitiveStr::Ascii("delarative"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacholers"), &["bachelors"]),
            (
                dictgen::InsensitiveStr::Ascii("uncertainities"),
                &["uncertainties"],
            ),
            (dictgen::InsensitiveStr::Ascii("pupulates"), &["populates"]),
            (dictgen::InsensitiveStr::Ascii("ovservers"), &["observers"]),
            (
                dictgen::InsensitiveStr::Ascii("certaintity"),
                &["certainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("thresshold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("subexpession"),
                &["subexpression"],
            ),
            (dictgen::InsensitiveStr::Ascii("crypric"), &["cryptic"]),
            (
                dictgen::InsensitiveStr::Ascii("disguntingly"),
                &["disgustingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("punsiher"), &["punisher"]),
            (dictgen::InsensitiveStr::Ascii("othwhise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("escalting"), &["escalating"]),
            (dictgen::InsensitiveStr::Ascii("oscilate"), &["oscillate"]),
            (dictgen::InsensitiveStr::Ascii("obeserving"), &["observing"]),
            (
                dictgen::InsensitiveStr::Ascii("opporutnity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("manully"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("catchip"),
                &["catchup", "ketchup"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imediately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreegations"),
                &["aggregations"],
            ),
            (dictgen::InsensitiveStr::Ascii("avearging"), &["averaging"]),
            (
                dictgen::InsensitiveStr::Ascii("definititely"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massectomy"),
                &["mastectomy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ajacencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("createive"), &["creative"]),
            (
                dictgen::InsensitiveStr::Ascii("uncommpressd"),
                &["uncompressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("deketed"), &["deleted"]),
            (
                dictgen::InsensitiveStr::Ascii("accointant"),
                &["accountant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masuclinity"),
                &["masculinity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaguration"),
                &["inauguration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecte"),
                &["architecture", "architect"],
            ),
            (dictgen::InsensitiveStr::Ascii("proviced"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("ouptuts"), &["outputs"]),
            (
                dictgen::InsensitiveStr::Ascii("circumsicions"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("hareware"), &["hardware"]),
            (dictgen::InsensitiveStr::Ascii("follwinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("aononymous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("guarrenteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disagreemtn"),
                &["disagreements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shtops"),
                &["stops", "shops"],
            ),
            (dictgen::InsensitiveStr::Ascii("workspce"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("resued"),
                &["reused", "rescued", "resumed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranducers"),
                &["transducers"],
            ),
            (dictgen::InsensitiveStr::Ascii("swappin"), &["swapping"]),
            (
                dictgen::InsensitiveStr::Ascii("correctionsn"),
                &["corrections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("divices"),
                &["devices", "divides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintuive"),
                &["unintuitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaganative"),
                &["imaginative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mannorisms"),
                &["mannerisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transohobic"),
                &["transphobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("featurs"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("authroization"),
                &["authorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapproves"),
                &["disapproves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurrantly"),
                &["recurrently"],
            ),
            (dictgen::InsensitiveStr::Ascii("vulakn"), &["vulkan"]),
            (
                dictgen::InsensitiveStr::Ascii("directlor"),
                &["directory", "director"],
            ),
            (dictgen::InsensitiveStr::Ascii("climatiser"), &["climates"]),
            (dictgen::InsensitiveStr::Ascii("turncates"), &["truncates"]),
            (
                dictgen::InsensitiveStr::Ascii("terrninates"),
                &["terminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("archatype"), &["archetype"]),
            (dictgen::InsensitiveStr::Ascii("turncated"), &["truncated"]),
            (
                dictgen::InsensitiveStr::Ascii("federativo"),
                &["federation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastinaton"),
                &["procrastination"],
            ),
            (dictgen::InsensitiveStr::Ascii("exepcts"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("instructers"),
                &["instructors", "instructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confessionis"),
                &["confessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("parastics"), &["parasitics"]),
            (
                dictgen::InsensitiveStr::Ascii("disassemlby"),
                &["disassembly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dosseay"), &["dossier"]),
            (dictgen::InsensitiveStr::Ascii("profies"), &["profiles"]),
            (dictgen::InsensitiveStr::Ascii("childrends"), &["childrens"]),
            (dictgen::InsensitiveStr::Ascii("reassing"), &["reassign"]),
            (dictgen::InsensitiveStr::Ascii("titel"), &["title"]),
            (dictgen::InsensitiveStr::Ascii("checkksum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("intitializers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agrgressive"),
                &["aggressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("archtects"), &["architects"]),
            (dictgen::InsensitiveStr::Ascii("alimoney"), &["alimony"]),
            (dictgen::InsensitiveStr::Ascii("sqauwk"), &["squawk"]),
            (dictgen::InsensitiveStr::Ascii("assmebler"), &["assembler"]),
            (dictgen::InsensitiveStr::Ascii("oribtal"), &["orbital"]),
            (
                dictgen::InsensitiveStr::Ascii("finess"),
                &["finesse", "fitness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infromative"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apropiately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rotatd"),
                &["rotated", "rotate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmenet"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimmension"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("listenning"), &["listening"]),
            (dictgen::InsensitiveStr::Ascii("ballancing"), &["balancing"]),
            (
                dictgen::InsensitiveStr::Ascii("normailzes"),
                &["normalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("naem"), &["name"]),
            (
                dictgen::InsensitiveStr::Ascii("asymtomatic"),
                &["asymptomatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precendeces"),
                &["precedences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesctions"),
                &["intersections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialzed"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefection"), &["perfection"]),
            (
                dictgen::InsensitiveStr::Ascii("accomodativeness"),
                &["accommodativeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("brethen"), &["brethren"]),
            (dictgen::InsensitiveStr::Ascii("actuakly"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("deflatin"), &["deflation"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wih"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("duaghters"), &["daughters"]),
            (
                dictgen::InsensitiveStr::Ascii("negoations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negelcting"),
                &["neglecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("archivs"), &["archives"]),
            (dictgen::InsensitiveStr::Ascii("ammong"), &["among"]),
            (dictgen::InsensitiveStr::Ascii("architecut"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("evaulators"),
                &["evaluators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrruptable"),
                &["interruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("frowards"), &["forwards"]),
            (dictgen::InsensitiveStr::Ascii("withinn"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("presedintia"),
                &["presidential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomprehendible"),
                &["incomprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelerato"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("harrases"), &["harasses"]),
            (dictgen::InsensitiveStr::Ascii("neccesity"), &["necessity"]),
            (dictgen::InsensitiveStr::Ascii("whipser"), &["whisper"]),
            (
                dictgen::InsensitiveStr::Ascii("cringeworthey"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutuxes"), &["mutexes"]),
            (dictgen::InsensitiveStr::Ascii("temaplate"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("achor"), &["anchor"]),
            (
                dictgen::InsensitiveStr::Ascii("distibuting"),
                &["distributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neightbourhood"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("dieases"), &["diseases"]),
            (dictgen::InsensitiveStr::Ascii("horray"), &["hooray"]),
            (
                dictgen::InsensitiveStr::Ascii("impresonating"),
                &["impersonating"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucuture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("scrollablbe"),
                &["scrollable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisements"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparantie"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjatated"), &["agitated"]),
            (dictgen::InsensitiveStr::Ascii("gamplay"), &["gameplay"]),
            (dictgen::InsensitiveStr::Ascii("sexualy"), &["sexually"]),
            (
                dictgen::InsensitiveStr::Ascii("autorizing"),
                &["authorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cainster"), &["canister"]),
            (
                dictgen::InsensitiveStr::Ascii("contuguous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silksreened"),
                &["silkscreened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("animatin"),
                &["animating", "animation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugested"), &["suggested"]),
            (dictgen::InsensitiveStr::Ascii("sentinal"), &["sentinel"]),
            (dictgen::InsensitiveStr::Ascii("dispath"), &["dispatch"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallic"),
                &["specific"],
            ),
            (dictgen::InsensitiveStr::Ascii("choos"), &["choose"]),
            (dictgen::InsensitiveStr::Ascii("diphtong"), &["diphthong"]),
            (
                dictgen::InsensitiveStr::Ascii("compressiong"),
                &["compressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conotrolled"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proletaryat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("vigourous"), &["vigorous"]),
            (
                dictgen::InsensitiveStr::Ascii("juggernaugt"),
                &["juggernaut"],
            ),
            (dictgen::InsensitiveStr::Ascii("advisin"), &["advising"]),
            (
                dictgen::InsensitiveStr::Ascii("developmant"),
                &["developmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("islamsit"), &["islamist"]),
            (dictgen::InsensitiveStr::Ascii("alledgely"), &["allegedly"]),
            (
                dictgen::InsensitiveStr::Ascii("acknolwedging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meskeetoes"),
                &["mosquitoes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collborative"),
                &["collaborative"],
            ),
            (dictgen::InsensitiveStr::Ascii("whitchever"), &["whichever"]),
            (
                dictgen::InsensitiveStr::Ascii("transparnecy"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("filiament"), &["filament"]),
            (dictgen::InsensitiveStr::Ascii("itineray"), &["itinerary"]),
            (dictgen::InsensitiveStr::Ascii("cofeee"), &["coffee"]),
            (dictgen::InsensitiveStr::Ascii("meatadata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("parial"), &["partial"]),
            (
                dictgen::InsensitiveStr::Ascii("incremently"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heirachies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurits"), &["recruits"]),
            (
                dictgen::InsensitiveStr::Ascii("unrecogonized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("percison"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("fightings"), &["fighting"]),
            (dictgen::InsensitiveStr::Ascii("sutdown"), &["shutdown"]),
            (
                dictgen::InsensitiveStr::Ascii("adminitered"),
                &["administered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conflifted"),
                &["conflicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contineous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fulfulling"),
                &["fulfilling"],
            ),
            (dictgen::InsensitiveStr::Ascii("macroes"), &["macros"]),
            (dictgen::InsensitiveStr::Ascii("inclde"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("projecticle"),
                &["projectile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("musel"),
                &["muscle", "mussel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consummable"),
                &["consumable"],
            ),
            (dictgen::InsensitiveStr::Ascii("straines"), &["strains"]),
            (
                dictgen::InsensitiveStr::Ascii("hurrycanes"),
                &["hurricanes"],
            ),
            (dictgen::InsensitiveStr::Ascii("gallaries"), &["galleries"]),
            (dictgen::InsensitiveStr::Ascii("vomitting"), &["vomiting"]),
            (dictgen::InsensitiveStr::Ascii("accsepts"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("connceted"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("elimates"), &["eliminates"]),
            (
                dictgen::InsensitiveStr::Ascii("insepction"),
                &["inspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affirmitive"),
                &["affirmative"],
            ),
            (dictgen::InsensitiveStr::Ascii("amrpits"), &["armpits"]),
            (
                dictgen::InsensitiveStr::Ascii("sigantures"),
                &["signatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("codgen"), &["codegen"]),
            (dictgen::InsensitiveStr::Ascii("repertwar"), &["repertoire"]),
            (dictgen::InsensitiveStr::Ascii("afircans"), &["africans"]),
            (dictgen::InsensitiveStr::Ascii("acadimies"), &["academies"]),
            (
                dictgen::InsensitiveStr::Ascii("repetetions"),
                &["repetitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assiociated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semented"),
                &["cemented", "segmented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceutions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("excluder"), &["excluded"]),
            (
                dictgen::InsensitiveStr::Ascii("filessytems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("giong"), &["going"]),
            (
                dictgen::InsensitiveStr::Ascii("implementtions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deseptively"),
                &["deceptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmisary"), &["emissary"]),
            (
                dictgen::InsensitiveStr::Ascii("demograhpic"),
                &["demographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluiflower"),
                &["cauliflower"],
            ),
            (dictgen::InsensitiveStr::Ascii("strory"), &["story"]),
            (
                dictgen::InsensitiveStr::Ascii("visuabizes"),
                &["visualizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquistior"),
                &["inquisitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicover"), &["discover"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoziators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("deployes"), &["deploys"]),
            (dictgen::InsensitiveStr::Ascii("decipting"), &["depicting"]),
            (
                dictgen::InsensitiveStr::Ascii("copyrigthed"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("warninngs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("intsallers"),
                &["installers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pretaining"),
                &["pertaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("jorneys"), &["journeys"]),
            (
                dictgen::InsensitiveStr::Ascii("implemataion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("scannning"), &["scanning"]),
            (dictgen::InsensitiveStr::Ascii("temporay"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("cunting"),
                &["counting", "hunting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefices"),
                &["prefixes", "prefaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("escapin"), &["escaping"]),
            (dictgen::InsensitiveStr::Ascii("tablle"), &["table"]),
            (
                dictgen::InsensitiveStr::Ascii("perspecive"),
                &["perspective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exploitating"),
                &["exploitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terfforming"),
                &["terraforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiantions"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("standardd"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("sateful"), &["stateful"]),
            (dictgen::InsensitiveStr::Ascii("foult"), &["fault"]),
            (dictgen::InsensitiveStr::Ascii("clssrooms"), &["classrooms"]),
            (
                dictgen::InsensitiveStr::Ascii("commodites"),
                &["commodities"],
            ),
            (dictgen::InsensitiveStr::Ascii("uppler"), &["upper"]),
            (dictgen::InsensitiveStr::Ascii("diposes"), &["disposes"]),
            (dictgen::InsensitiveStr::Ascii("fedback"), &["feedback"]),
            (dictgen::InsensitiveStr::Ascii("yaer"), &["year"]),
            (
                dictgen::InsensitiveStr::Ascii("initialializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("meessaged"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("despicible"),
                &["despicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalistische"),
                &["nationalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconeecting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("doygen"), &["doxygen"]),
            (dictgen::InsensitiveStr::Ascii("reponders"), &["responders"]),
            (dictgen::InsensitiveStr::Ascii("instea"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("cateogrically"),
                &["categorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("hombrew"), &["homebrew"]),
            (
                dictgen::InsensitiveStr::Ascii("deintialized"),
                &["deinitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normalls"),
                &["normals", "normally"],
            ),
            (dictgen::InsensitiveStr::Ascii("occurr"), &["occur"]),
            (dictgen::InsensitiveStr::Ascii("ownwer"), &["owner"]),
            (
                dictgen::InsensitiveStr::Ascii("prepatory"),
                &["preparatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("separat"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("movemets"), &["movements"]),
            (
                dictgen::InsensitiveStr::Ascii("enviorenment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeared"), &["compared"]),
            (dictgen::InsensitiveStr::Ascii("featchure"), &["feature"]),
            (dictgen::InsensitiveStr::Ascii("generlaly"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilties"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("mediavel"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("seires"), &["series"]),
            (
                dictgen::InsensitiveStr::Ascii("undersanding"),
                &["understanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abandenment"),
                &["abandonment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticisze"), &["criticise"]),
            (dictgen::InsensitiveStr::Ascii("justfiy"), &["justify"]),
            (dictgen::InsensitiveStr::Ascii("sekect"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("autasaves"), &["autosaves"]),
            (dictgen::InsensitiveStr::Ascii("salaires"), &["salaries"]),
            (dictgen::InsensitiveStr::Ascii("estiomator"), &["estimator"]),
            (dictgen::InsensitiveStr::Ascii("diconnect"), &["disconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("dermatologyst"),
                &["dermatologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastrutture"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consdieration"),
                &["considerations", "consideration"],
            ),
            (dictgen::InsensitiveStr::Ascii("relases"), &["releases"]),
            (dictgen::InsensitiveStr::Ascii("leesurely"), &["leisurely"]),
            (dictgen::InsensitiveStr::Ascii("momemtn"), &["moment"]),
            (dictgen::InsensitiveStr::Ascii("detecttors"), &["detectors"]),
            (dictgen::InsensitiveStr::Ascii("stiched"), &["stitched"]),
            (
                dictgen::InsensitiveStr::Ascii("sufficiantly"),
                &["sufficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dpeendence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("delimitted"), &["delimited"]),
            (dictgen::InsensitiveStr::Ascii("chiuhahua"), &["chihuahua"]),
            (
                dictgen::InsensitiveStr::Ascii("misspelings"),
                &["misspellings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conciderations"),
                &["considerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraprashing"),
                &["paraphrasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("quantiy"), &["quantity"]),
            (dictgen::InsensitiveStr::Ascii("supoorted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("shoudl"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("legende"),
                &["legendre", "legend", "legends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifcations"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("yuforic"), &["euphoric"]),
            (
                dictgen::InsensitiveStr::Ascii("regstering"),
                &["registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("becomses"), &["becomes"]),
            (dictgen::InsensitiveStr::Ascii("follwoe"), &["follow"]),
            (
                dictgen::InsensitiveStr::Ascii("obilterated"),
                &["obliterated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfterbly"),
                &["comfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("zimmaps"), &["zipmaps"]),
            (dictgen::InsensitiveStr::Ascii("whihch"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("ridiculos"), &["ridiculous"]),
            (dictgen::InsensitiveStr::Ascii("errenous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("comrpession"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrgesiter"),
                &["unregister"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supersript"),
                &["superscript"],
            ),
            (dictgen::InsensitiveStr::Ascii("ferbuary"), &["february"]),
            (dictgen::InsensitiveStr::Ascii("hispanicos"), &["hispanics"]),
            (dictgen::InsensitiveStr::Ascii("vampries"), &["vampires"]),
            (dictgen::InsensitiveStr::Ascii("imortes"), &["imports"]),
            (dictgen::InsensitiveStr::Ascii("mentiond"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("expetects"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("prediced"), &["predicted"]),
            (
                dictgen::InsensitiveStr::Ascii("repositotries"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnent"), &["comment"]),
            (
                dictgen::InsensitiveStr::Ascii("naibhorhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("reacheable"), &["reachable"]),
            (
                dictgen::InsensitiveStr::Ascii("invovement"),
                &["involvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("apporve"), &["approve"]),
            (dictgen::InsensitiveStr::Ascii("defition"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("functionailties"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modfiiable"),
                &["modifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyclopeedia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponatenous"),
                &["spontaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryify"), &["verify"]),
            (
                dictgen::InsensitiveStr::Ascii("exahustion"),
                &["exhaustion"],
            ),
            (dictgen::InsensitiveStr::Ascii("desigined"), &["designed"]),
            (dictgen::InsensitiveStr::Ascii("syteme"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("condescenscion"),
                &["condescension"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insuniating"),
                &["insinuating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveniente"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("densitities"),
                &["densities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technoloiges"),
                &["technologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmplar"), &["exemplar"]),
            (dictgen::InsensitiveStr::Ascii("organismes"), &["organisms"]),
            (
                dictgen::InsensitiveStr::Ascii("identificaton"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("wotkspace"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("resilent"),
                &["resilient", "resident"],
            ),
            (dictgen::InsensitiveStr::Ascii("infinate"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("insperational"),
                &["inspirational"],
            ),
            (dictgen::InsensitiveStr::Ascii("forsenic"), &["forensic"]),
            (dictgen::InsensitiveStr::Ascii("releasin"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("blythly"), &["blithely"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetraron"),
                &["perpetrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissiosn"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminiscant"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("judaisim"), &["judaism"]),
            (dictgen::InsensitiveStr::Ascii("texutre"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("eveluates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("leagality"), &["legality"]),
            (dictgen::InsensitiveStr::Ascii("laboratoy"), &["laboratory"]),
            (
                dictgen::InsensitiveStr::Ascii("gymnistics"),
                &["gymnastics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratificaiton"),
                &["gratification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictarorship"),
                &["dictatorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("reslvers"), &["resolvers"]),
            (
                dictgen::InsensitiveStr::Ascii("cleancacne"),
                &["cleancache"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bluebarries"),
                &["blueberries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromosomers"),
                &["chromosomes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartations"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("downlowds"), &["downloads"]),
            (
                dictgen::InsensitiveStr::Ascii("generaterated"),
                &["generated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negogtiable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("subcribes"), &["subscribes"]),
            (
                dictgen::InsensitiveStr::Ascii("accepectors"),
                &["acceptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("praameters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("incremnet"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("dpendences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("ukrianian"), &["ukrainian"]),
            (
                dictgen::InsensitiveStr::Ascii("acquaintence"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("utillize"), &["utilize"]),
            (
                dictgen::InsensitiveStr::Ascii("photoshipped"),
                &["photoshopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilitise"), &["utilise"]),
            (dictgen::InsensitiveStr::Ascii("coincidin"), &["coinciding"]),
            (
                dictgen::InsensitiveStr::Ascii("partaining"),
                &["pertaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrcted"),
                &["constructed", "constricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("organismus"), &["organisms"]),
            (
                dictgen::InsensitiveStr::Ascii("scroed"),
                &["scored", "scrolled", "scrod"],
            ),
            (dictgen::InsensitiveStr::Ascii("varius"), &["various"]),
            (dictgen::InsensitiveStr::Ascii("pertubing"), &["perturbing"]),
            (
                dictgen::InsensitiveStr::Ascii("coincidince"),
                &["coincidence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ststioned"), &["stationed"]),
            (
                dictgen::InsensitiveStr::Ascii("frequenties"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("exntends"), &["extends"]),
            (dictgen::InsensitiveStr::Ascii("whilest"), &["whilst"]),
            (dictgen::InsensitiveStr::Ascii("nothe"), &["north", "note"]),
            (
                dictgen::InsensitiveStr::Ascii("silhoueted"),
                &["silhouetted"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragmet"), &["fragment"]),
            (dictgen::InsensitiveStr::Ascii("sitck"), &["stick"]),
            (dictgen::InsensitiveStr::Ascii("dne"), &["done"]),
            (dictgen::InsensitiveStr::Ascii("premuims"), &["premiums"]),
            (
                dictgen::InsensitiveStr::Ascii("prepration"),
                &["preparation"],
            ),
            (dictgen::InsensitiveStr::Ascii("decidated"), &["dedicated"]),
            (
                dictgen::InsensitiveStr::Ascii("coresponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("accros"), &["across"]),
            (dictgen::InsensitiveStr::Ascii("recconets"), &["reconnects"]),
            (dictgen::InsensitiveStr::Ascii("disbled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("guared"),
                &["guard", "guarded"],
            ),
            (dictgen::InsensitiveStr::Ascii("naseuous"), &["nauseous"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondance"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convienient"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ditributions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("attactor"), &["attractor"]),
            (
                dictgen::InsensitiveStr::Ascii("sequencially"),
                &["sequentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispalys"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("anarchsit"), &["anarchist"]),
            (
                dictgen::InsensitiveStr::Ascii("contiuously"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succicently"),
                &["sufficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("fractalius"), &["fractals"]),
            (
                dictgen::InsensitiveStr::Ascii("specififics"),
                &["specifics"],
            ),
            (dictgen::InsensitiveStr::Ascii("sublass"), &["subclass"]),
            (dictgen::InsensitiveStr::Ascii("lov"), &["love"]),
            (
                dictgen::InsensitiveStr::Ascii("approprate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepentence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irresistable"),
                &["irresistible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decresdendo"),
                &["decrescendo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicaions"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("nubmers"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("paranoa"), &["paranoia"]),
            (dictgen::InsensitiveStr::Ascii("nagative"), &["negative"]),
            (
                dictgen::InsensitiveStr::Ascii("inadequite"),
                &["inadequate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificarions"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("extranal"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("disocgraphy"),
                &["discography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uneeded"),
                &["unneeded", "unheeded", "needed"],
            ),
            (dictgen::InsensitiveStr::Ascii("remve"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("clases"),
                &["cases", "clashes", "classes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquantiance"),
                &["acquaintances", "acquaintance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorgithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeendage"), &["appendage"]),
            (dictgen::InsensitiveStr::Ascii("mimimize"), &["minimize"]),
            (dictgen::InsensitiveStr::Ascii("acive"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("arbirarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("cambrigde"), &["cambridge"]),
            (dictgen::InsensitiveStr::Ascii("joysick"), &["joystick"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstaning"),
                &["misunderstanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impresssion"),
                &["impressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transalators"),
                &["translators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorhithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritpro"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("delimted"), &["delimited"]),
            (
                dictgen::InsensitiveStr::Ascii("exeuctable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("msssge"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("evelation"), &["elevation"]),
            (
                dictgen::InsensitiveStr::Ascii("ovrridable"),
                &["overridable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosisted"), &["consisted"]),
            (dictgen::InsensitiveStr::Ascii("extemes"), &["extremes"]),
            (
                dictgen::InsensitiveStr::Ascii("straightmen"),
                &["straighten"],
            ),
            (dictgen::InsensitiveStr::Ascii("aweful"), &["awful"]),
            (dictgen::InsensitiveStr::Ascii("interferd"), &["interfered"]),
            (dictgen::InsensitiveStr::Ascii("travereses"), &["traverses"]),
            (dictgen::InsensitiveStr::Ascii("spliter"), &["splitter"]),
            (dictgen::InsensitiveStr::Ascii("sysmbol"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("regualtions"),
                &["regulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influenceing"),
                &["influencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legendaies"),
                &["legendaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apocalpytic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("aristolte"), &["aristotle"]),
            (
                dictgen::InsensitiveStr::Ascii("availailable"),
                &["available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabality"),
                &["capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encouttered"),
                &["encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("weightened"), &["weighted"]),
            (
                dictgen::InsensitiveStr::Ascii("irreversable"),
                &["irreversible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commandent"),
                &["commandment"],
            ),
            (dictgen::InsensitiveStr::Ascii("srry"), &["sorry"]),
            (dictgen::InsensitiveStr::Ascii("receiviing"), &["receiving"]),
            (dictgen::InsensitiveStr::Ascii("nescessary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("arquitecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("dyamics"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("dignosed"), &["diagnosed"]),
            (
                dictgen::InsensitiveStr::Ascii("communitcate"),
                &["communicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("hinderence"), &["hindrance"]),
            (
                dictgen::InsensitiveStr::Ascii("expermentally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("edtion"), &["edition"]),
            (
                dictgen::InsensitiveStr::Ascii("filder"),
                &["folder", "filter", "finder", "fielder", "filer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminatoire"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("eventally"), &["eventually"]),
            (
                dictgen::InsensitiveStr::Ascii("corrdinator"),
                &["coordinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibel"),
                &["responsibly", "responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonymousy"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanizer"), &["sanitizer"]),
            (
                dictgen::InsensitiveStr::Ascii("greating"),
                &["greeting", "grating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intace"),
                &["instance", "intact", "intake"],
            ),
            (dictgen::InsensitiveStr::Ascii("foregroud"), &["foreground"]),
            (dictgen::InsensitiveStr::Ascii("keybords"), &["keyboards"]),
            (dictgen::InsensitiveStr::Ascii("recended"), &["rescinded"]),
            (dictgen::InsensitiveStr::Ascii("commeneted"), &["commented"]),
            (dictgen::InsensitiveStr::Ascii("styiling"), &["styling"]),
            (
                dictgen::InsensitiveStr::Ascii("terorrists"),
                &["terrorists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unistallers"),
                &["uninstallers"],
            ),
            (dictgen::InsensitiveStr::Ascii("dscretion"), &["discretion"]),
            (
                dictgen::InsensitiveStr::Ascii("pesitcides"),
                &["pesticides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constist"),
                &["consist", "consists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distincte"),
                &["distinctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irracional"),
                &["irrational"],
            ),
            (dictgen::InsensitiveStr::Ascii("languange"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("constaining"),
                &["constraining", "containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfiy"), &["verify"]),
            (
                dictgen::InsensitiveStr::Ascii("clearifications"),
                &["clarifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fluttersky"),
                &["fluttershy"],
            ),
            (dictgen::InsensitiveStr::Ascii("unabale"), &["unable"]),
            (
                dictgen::InsensitiveStr::Ascii("devirutalize"),
                &["devirtualize"],
            ),
            (dictgen::InsensitiveStr::Ascii("addrssing"), &["addressing"]),
            (dictgen::InsensitiveStr::Ascii("examinates"), &["examines"]),
            (
                dictgen::InsensitiveStr::Ascii("decrase"),
                &["decrease", "decease"],
            ),
            (dictgen::InsensitiveStr::Ascii("turnaroud"), &["turnaround"]),
            (
                dictgen::InsensitiveStr::Ascii("neumonectomy"),
                &["pneumonectomy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mintors"),
                &["mentors", "monitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("labbel"), &["label"]),
            (
                dictgen::InsensitiveStr::Ascii("nonsensicle"),
                &["nonsensical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breackthroughs"),
                &["breakthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distraccion"),
                &["distraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("curageous"), &["courageous"]),
            (
                dictgen::InsensitiveStr::Ascii("statestical"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addittionally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specicification"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcemet"),
                &["announcements", "announcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("backwark"), &["backward"]),
            (
                dictgen::InsensitiveStr::Ascii("exploiding"),
                &["exploiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conpleting"),
                &["completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatabile"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschology"),
                &["psychology"],
            ),
            (dictgen::InsensitiveStr::Ascii("encoutner"), &["encounter"]),
            (
                dictgen::InsensitiveStr::Ascii("singluarity"),
                &["singularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visisbilities"),
                &["visibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrection"),
                &["correction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefence"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("taxonoy"), &["taxonomy"]),
            (dictgen::InsensitiveStr::Ascii("perfomrs"), &["performs"]),
            (
                dictgen::InsensitiveStr::Ascii("eletromagnitic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatatble"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disclipines"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("numbber"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("prdpagation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("answhareing"),
                &["answering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("smouthness"),
                &["smoothness"],
            ),
            (dictgen::InsensitiveStr::Ascii("inheriths"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("assingled"), &["assigned"]),
            (dictgen::InsensitiveStr::Ascii("markaup"), &["markup"]),
            (dictgen::InsensitiveStr::Ascii("kiyacks"), &["kayaks"]),
            (dictgen::InsensitiveStr::Ascii("utillity"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("decremeanted"),
                &["decremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("voume"), &["volume"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficienct"),
                &["coefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mininalist"),
                &["minimalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spontaneoulsy"),
                &["spontaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("accecpted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoemamce"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreeval"), &["retrieval"]),
            (
                dictgen::InsensitiveStr::Ascii("approopriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideologias"),
                &["ideologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifiance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("stte"), &["state"]),
            (
                dictgen::InsensitiveStr::Ascii("demosntrably"),
                &["demonstrably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighboordhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplishs"),
                &["accomplishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construint"),
                &["constraint"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehse"), &["these", "tense"]),
            (dictgen::InsensitiveStr::Ascii("properaly"), &["properly"]),
            (dictgen::InsensitiveStr::Ascii("intregral"), &["integral"]),
            (
                dictgen::InsensitiveStr::Ascii("enginee"),
                &["engine", "engineer"],
            ),
            (dictgen::InsensitiveStr::Ascii("praries"), &["prairies"]),
            (dictgen::InsensitiveStr::Ascii("vigrins"), &["virgins"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("whistels"), &["whistles"]),
            (
                dictgen::InsensitiveStr::Ascii("doesing"),
                &["doing", "does", "dosing", "dozing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathcmaking"),
                &["matchmaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkpoin"), &["checkpoint"]),
            (dictgen::InsensitiveStr::Ascii("algorhtm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("constistency"),
                &["consistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("befurehand"),
                &["beforehand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scarifices"),
                &["sacrifices"],
            ),
            (dictgen::InsensitiveStr::Ascii("ideologis"), &["ideologies"]),
            (dictgen::InsensitiveStr::Ascii("elephantes"), &["elephants"]),
            (
                dictgen::InsensitiveStr::Ascii("serialializing"),
                &["serializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("optonal"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("meomory"), &["memory"]),
            (
                dictgen::InsensitiveStr::Ascii("presipitator"),
                &["precipitator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incredibily"),
                &["incredibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("physicaly"), &["physically"]),
            (dictgen::InsensitiveStr::Ascii("addied"), &["added"]),
            (
                dictgen::InsensitiveStr::Ascii("gadged"),
                &["gadget", "gauged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aaproximately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("compoennt"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("rumatic"), &["rheumatic"]),
            (
                dictgen::InsensitiveStr::Ascii("synronous"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hussel"),
                &["hustle", "mussel"],
            ),
            (dictgen::InsensitiveStr::Ascii("miminums"), &["minimums"]),
            (
                dictgen::InsensitiveStr::Ascii("concsience"),
                &["conscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdends"), &["depends"]),
            (
                dictgen::InsensitiveStr::Ascii("prioritice"),
                &["prioritize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reocurring"),
                &["reoccurring", "recurring"],
            ),
            (dictgen::InsensitiveStr::Ascii("tailred"), &["tailored"]),
            (dictgen::InsensitiveStr::Ascii("coommitted"), &["committed"]),
            (
                dictgen::InsensitiveStr::Ascii("eliminaton"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realtion"),
                &["relation", "reaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("pleaes"), &["please"]),
            (
                dictgen::InsensitiveStr::Ascii("supermarkts"),
                &["supermarkets"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarentyd"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("privcy"), &["privacy"]),
            (
                dictgen::InsensitiveStr::Ascii("pplication"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("dedidate"), &["dedicate"]),
            (dictgen::InsensitiveStr::Ascii("anyawy"), &["anyway"]),
            (dictgen::InsensitiveStr::Ascii("attemtped"), &["attempted"]),
            (dictgen::InsensitiveStr::Ascii("selecing"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("constrolled"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallucenation"),
                &["hallucination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mosntrosity"),
                &["monstrosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("garantuee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("characterisicly"),
                &["characteristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conintuity"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trickyness"),
                &["trickiness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenendence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotainers"), &["containers"]),
            (dictgen::InsensitiveStr::Ascii("fascits"), &["fascist"]),
            (dictgen::InsensitiveStr::Ascii("parmetian"), &["parmesan"]),
            (dictgen::InsensitiveStr::Ascii("dammage"), &["damage"]),
            (
                dictgen::InsensitiveStr::Ascii("necessairy"),
                &["necessarily", "necessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modyfications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("wryth"), &["writhe"]),
            (dictgen::InsensitiveStr::Ascii("optomizer"), &["optimizer"]),
            (
                dictgen::InsensitiveStr::Ascii("notificarion"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identitied"),
                &["identities", "identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiensial"),
                &["experiential"],
            ),
            (dictgen::InsensitiveStr::Ascii("avcoids"), &["avoids"]),
            (dictgen::InsensitiveStr::Ascii("atrillery"), &["artillery"]),
            (
                dictgen::InsensitiveStr::Ascii("accordngly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("angshiuosly"),
                &["anxiously"],
            ),
            (dictgen::InsensitiveStr::Ascii("curcular"), &["circular"]),
            (dictgen::InsensitiveStr::Ascii("empyt"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("reintialize"),
                &["reinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("easiler"), &["easier"]),
            (dictgen::InsensitiveStr::Ascii("fursthest"), &["furthest"]),
            (
                dictgen::InsensitiveStr::Ascii("assiocative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wainting"),
                &["waiting", "wanting"],
            ),
            (dictgen::InsensitiveStr::Ascii("jaugars"), &["jaguars"]),
            (
                dictgen::InsensitiveStr::Ascii("multifuction"),
                &["multifunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("gnerics"), &["generics"]),
            (dictgen::InsensitiveStr::Ascii("conseeds"), &["concedes"]),
            (
                dictgen::InsensitiveStr::Ascii("assiciative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repatition"),
                &["repetition", "repartition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("baged"),
                &["badge", "bagged"],
            ),
            (dictgen::InsensitiveStr::Ascii("algoruthm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("developement"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adreanline"),
                &["adrenaline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asyncrounesouly"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practicianer"),
                &["practitioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcctly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("specififer"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("compartements"),
                &["compartments"],
            ),
            (dictgen::InsensitiveStr::Ascii("bumpting"), &["bumping"]),
            (
                dictgen::InsensitiveStr::Ascii("succcession"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("peaple"), &["people"]),
            (dictgen::InsensitiveStr::Ascii("attentino"), &["attention"]),
            (dictgen::InsensitiveStr::Ascii("singeleton"), &["singleton"]),
            (
                dictgen::InsensitiveStr::Ascii("modifactions"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flatterende"),
                &["flattered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addrerssing"),
                &["addressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstarting"),
                &["demonstrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generaring"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgradring"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("themseves"), &["themselves"]),
            (
                dictgen::InsensitiveStr::Ascii("undupplicated"),
                &["unduplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("trigggers"), &["triggers"]),
            (
                dictgen::InsensitiveStr::Ascii("canabelyse"),
                &["cannibalise"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlcludion"), &["inclusion"]),
            (
                dictgen::InsensitiveStr::Ascii("benechmarked"),
                &["benchmarked"],
            ),
            (dictgen::InsensitiveStr::Ascii("fwirte"), &["fwrite"]),
            (
                dictgen::InsensitiveStr::Ascii("confedarate"),
                &["confederate"],
            ),
            (dictgen::InsensitiveStr::Ascii("anixety"), &["anxiety"]),
            (
                dictgen::InsensitiveStr::Ascii("backgrunds"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exeuctioner"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("earler"), &["earlier"]),
            (dictgen::InsensitiveStr::Ascii("defineable"), &["definable"]),
            (
                dictgen::InsensitiveStr::Ascii("precides"),
                &["precedes", "presides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neruological"),
                &["neurological"],
            ),
            (dictgen::InsensitiveStr::Ascii("grometry"), &["geometry"]),
            (
                dictgen::InsensitiveStr::Ascii("appropropreately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catagorize"),
                &["categorize"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveyrbody"), &["everybody"]),
            (
                dictgen::InsensitiveStr::Ascii("appreicating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrigerater"),
                &[
                    "refrigerator",
                    "refrigerated",
                    "refrigerates",
                    "refrigerate",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consoldiated"),
                &["consolidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endocers"),
                &["endorsers", "enforcers", "encoders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depand"),
                &["depend", "demand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behavorial"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repsitories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coresponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("droput"), &["dropout"]),
            (dictgen::InsensitiveStr::Ascii("immesnely"), &["immensely"]),
            (
                dictgen::InsensitiveStr::Ascii("resignacion"),
                &["resignation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referennces"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moisutrizer"),
                &["moisturizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("tunned"), &["tuned"]),
            (
                dictgen::InsensitiveStr::Ascii("experimaneting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simuations"),
                &["simulations", "situations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliction"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participait"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("sovler"), &["solver"]),
            (dictgen::InsensitiveStr::Ascii("montaran"), &["montana"]),
            (dictgen::InsensitiveStr::Ascii("iterare"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("procesesed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("neighbores"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("experamenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploade"), &["upload"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomenaces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("hammmer"), &["hammer"]),
            (
                dictgen::InsensitiveStr::Ascii("consrtucts"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("blaclist"), &["blacklist"]),
            (
                dictgen::InsensitiveStr::Ascii("uninfroming"),
                &["uninforming", "uniforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatbility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplyifing"),
                &["simplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expositivo"),
                &["exposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("africaness"), &["africans"]),
            (
                dictgen::InsensitiveStr::Ascii("ovveridden"),
                &["overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("uqests"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("correspindingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positionne"),
                &["positioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectopmed"), &["sectioned"]),
            (dictgen::InsensitiveStr::Ascii("latitue"), &["latitude"]),
            (
                dictgen::InsensitiveStr::Ascii("intepreting"),
                &["interpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalulate"),
                &["recalculate"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("achievemint"),
                &["achievement"],
            ),
            (dictgen::InsensitiveStr::Ascii("instruts"), &["instructs"]),
            (dictgen::InsensitiveStr::Ascii("distroys"), &["destroys"]),
            (dictgen::InsensitiveStr::Ascii("desig"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("formath"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("budgest"), &["budgets"]),
            (dictgen::InsensitiveStr::Ascii("secound"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("actvity"), &["activity"]),
            (
                dictgen::InsensitiveStr::Ascii("confgiures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconcary"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("comitted"), &["committed"]),
            (
                dictgen::InsensitiveStr::Ascii("deseralizes"),
                &["deserializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psichologically"),
                &["psychologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispalyed"), &["displayed"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pentuim"), &["pentium"]),
            (dictgen::InsensitiveStr::Ascii("tomorrrow"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("microsfoft"), &["microsoft"]),
            (dictgen::InsensitiveStr::Ascii("scoket"), &["socket"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinatess"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kowledgeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("protrays"), &["portrays"]),
            (dictgen::InsensitiveStr::Ascii("generel"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("imidiate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("sufocated"), &["suffocated"]),
            (dictgen::InsensitiveStr::Ascii("meantings"), &["meanings"]),
            (dictgen::InsensitiveStr::Ascii("wansts"), &["wants"]),
            (dictgen::InsensitiveStr::Ascii("ignormable"), &["ignorable"]),
            (
                dictgen::InsensitiveStr::Ascii("necessaerily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfullfilled"),
                &["unfulfilled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("madturbating"),
                &["masturbating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetices"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("bastractions"),
                &["abstractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahear"), &["adhere", "hear"]),
            (
                dictgen::InsensitiveStr::Ascii("puplishers"),
                &["publishers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminescent"),
                &["reminiscent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vizualisation"),
                &["visualisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accoutnant"),
                &["accountant"],
            ),
            (dictgen::InsensitiveStr::Ascii("levellign"), &["levelling"]),
            (dictgen::InsensitiveStr::Ascii("punisse"), &["punishes"]),
            (dictgen::InsensitiveStr::Ascii("baout"), &["about", "bout"]),
            (
                dictgen::InsensitiveStr::Ascii("declartation"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirlce"), &["circle"]),
            (dictgen::InsensitiveStr::Ascii("cantacts"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("reconeccted"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("markes"),
                &["marks", "marked", "markers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charasmatic"),
                &["charismatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("supposrted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("evovled"), &["evolved"]),
            (dictgen::InsensitiveStr::Ascii("crucibe"), &["crucible"]),
            (
                dictgen::InsensitiveStr::Ascii("relaasing"),
                &["releasing", "relapsing"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrusday"), &["thursday"]),
            (
                dictgen::InsensitiveStr::Ascii("relagation"),
                &["relaxation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implausbile"),
                &["implausible"],
            ),
            (dictgen::InsensitiveStr::Ascii("impacte"), &["impact"]),
            (dictgen::InsensitiveStr::Ascii("receves"), &["receives"]),
            (
                dictgen::InsensitiveStr::Ascii("warehosues"),
                &["warehouses"],
            ),
            (dictgen::InsensitiveStr::Ascii("prioritie"), &["prioritize"]),
            (dictgen::InsensitiveStr::Ascii("tyep"), &["type", "yep"]),
            (
                dictgen::InsensitiveStr::Ascii("transgended"),
                &["transgender"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalzes"),
                &["capitalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firendzoned"),
                &["friendzoned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trangle"),
                &["triangle", "tangle", "strangle", "wrangle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegociator"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diminsioning"),
                &["dimensioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosoper"),
                &["philosopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accordlingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheves"), &["achieves"]),
            (dictgen::InsensitiveStr::Ascii("grpahics"), &["graphics"]),
            (dictgen::InsensitiveStr::Ascii("electroman"), &["electron"]),
            (dictgen::InsensitiveStr::Ascii("otimise"), &["optimise"]),
            (
                dictgen::InsensitiveStr::Ascii("denominatior"),
                &["denominator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destrctured"),
                &["destructured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counstructors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("snapsnots"), &["snapshots"]),
            (
                dictgen::InsensitiveStr::Ascii("caracterisations"),
                &["characterisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("effiency"), &["efficiency"]),
            (
                dictgen::InsensitiveStr::Ascii("distaces"),
                &["distances", "distastes"],
            ),
            (dictgen::InsensitiveStr::Ascii("presering"), &["preserving"]),
            (
                dictgen::InsensitiveStr::Ascii("governemnt"),
                &["government"],
            ),
            (dictgen::InsensitiveStr::Ascii("discussses"), &["discusses"]),
            (dictgen::InsensitiveStr::Ascii("extensios"), &["extensions"]),
            (dictgen::InsensitiveStr::Ascii("certaincy"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("polulation"),
                &["population"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncognized"),
                &["unrecognized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutionss"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incusive"),
                &["inclusive", "incisive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prinf"),
                &["print", "printf", "sprintf"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterprted"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlusive"), &["inclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangents"),
                &["rearrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigae"),
                &["investigate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("powrrlifting"),
                &["powerlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contempory"),
                &["contemporary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("whioch"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("renegatiotiated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("linkink"), &["linking"]),
            (dictgen::InsensitiveStr::Ascii("subracts"), &["subtracts"]),
            (
                dictgen::InsensitiveStr::Ascii("implimented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("nulk"), &["null"]),
            (
                dictgen::InsensitiveStr::Ascii("discrimate"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("fliename"), &["filename"]),
            (dictgen::InsensitiveStr::Ascii("monopoply"), &["monopoly"]),
            (dictgen::InsensitiveStr::Ascii("sibilings"), &["siblings"]),
            (dictgen::InsensitiveStr::Ascii("vishus"), &["vicious"]),
            (dictgen::InsensitiveStr::Ascii("invald"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("connonically"),
                &["canonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specualtive"),
                &["speculative"],
            ),
            (dictgen::InsensitiveStr::Ascii("mofify"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("unselcted"), &["unselected"]),
            (dictgen::InsensitiveStr::Ascii("inegrated"), &["integrated"]),
            (dictgen::InsensitiveStr::Ascii("reserverad"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("instucts"), &["instructs"]),
            (dictgen::InsensitiveStr::Ascii("restriced"), &["restricted"]),
            (dictgen::InsensitiveStr::Ascii("irradated"), &["irradiated"]),
            (dictgen::InsensitiveStr::Ascii("statiskt"), &["statist"]),
            (
                dictgen::InsensitiveStr::Ascii("direcitonal"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("hemingwary"), &["hemingway"]),
            (
                dictgen::InsensitiveStr::Ascii("staggerring"),
                &["staggering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sugestions"),
                &["suggestions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxtracted"), &["extracted"]),
            (dictgen::InsensitiveStr::Ascii("ciculars"), &["circulars"]),
            (
                dictgen::InsensitiveStr::Ascii("restricton"),
                &["restriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("sympaty"), &["sympathy"]),
            (
                dictgen::InsensitiveStr::Ascii("egotystical"),
                &["egotistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetable"),
                &["repeatable", "reputable"],
            ),
            (dictgen::InsensitiveStr::Ascii("lcoality"), &["locality"]),
            (dictgen::InsensitiveStr::Ascii("tagret"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("eletronics"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfarring"),
                &["transferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificacion"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("timout"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("poitlessly"),
                &["pointlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("completeed"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("figgure"), &["figure"]),
            (dictgen::InsensitiveStr::Ascii("circiuts"), &["circuits"]),
            (dictgen::InsensitiveStr::Ascii("othewrise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("deserializaed"),
                &["deserialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("angluar"), &["angular"]),
            (dictgen::InsensitiveStr::Ascii("expatriot"), &["expatriate"]),
            (dictgen::InsensitiveStr::Ascii("cach"), &["catch", "cache"]),
            (
                dictgen::InsensitiveStr::Ascii("contradicing"),
                &["contradicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissmantle"), &["dismantle"]),
            (
                dictgen::InsensitiveStr::Ascii("localte"),
                &["locale", "locate"],
            ),
            (dictgen::InsensitiveStr::Ascii("paladinos"), &["paladins"]),
            (
                dictgen::InsensitiveStr::Ascii("pertitently"),
                &["pertinently", "persistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexualty"),
                &["homosexuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambadextrouseness"),
                &["ambidextrousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gradification"),
                &["gratification"],
            ),
            (dictgen::InsensitiveStr::Ascii("starined"), &["strained"]),
            (dictgen::InsensitiveStr::Ascii("defauling"), &["defaulting"]),
            (
                dictgen::InsensitiveStr::Ascii("dispensarie"),
                &["dispenser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonexistance"),
                &["nonexistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particularlly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trancendent"),
                &["transcendent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitivos"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agresssively"),
                &["aggressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenatication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharamceuticals"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpats"), &["adapts"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiification"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaningles"), &["meanings"]),
            (
                dictgen::InsensitiveStr::Ascii("enumuration"),
                &["enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowldege"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("cashieer"), &["cashier"]),
            (
                dictgen::InsensitiveStr::Ascii("specefically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("iternators"), &["iterators"]),
            (
                dictgen::InsensitiveStr::Ascii("highighter"),
                &["highlighter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colonialisim"),
                &["colonialism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("importntly"),
                &["importantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handshakin"),
                &["handshaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referensed"),
                &["referenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interuption"),
                &["interruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spiritualy"),
                &["spiritually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increaes"),
                &["increases", "increase"],
            ),
            (dictgen::InsensitiveStr::Ascii("mccarthey"), &["mccarthy"]),
            (
                dictgen::InsensitiveStr::Ascii("helicoptors"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfiying"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("langiage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("discconected"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("exttra"), &["extra"]),
            (
                dictgen::InsensitiveStr::Ascii("brotherhoood"),
                &["brotherhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepedently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infomrational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("physqiue"), &["physique"]),
            (dictgen::InsensitiveStr::Ascii("lagacies"), &["legacies"]),
            (
                dictgen::InsensitiveStr::Ascii("pessimistc"),
                &["pessimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("terriblly"), &["terribly"]),
            (
                dictgen::InsensitiveStr::Ascii("conmanding"),
                &["commanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumarator"),
                &["enumerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrphic"),
                &["catastrophic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandards"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("includeed"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("connectivety"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfide"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("foraigner"), &["foreigner"]),
            (
                dictgen::InsensitiveStr::Ascii("insenitive"),
                &["insensitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("earplus"), &["earplugs"]),
            (dictgen::InsensitiveStr::Ascii("lolal"), &["total"]),
            (
                dictgen::InsensitiveStr::Ascii("sacling"),
                &["scaling", "sapling", "sacking", "sailing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicetly"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opjectively"),
                &["objectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("thumbanil"), &["thumbnail"]),
            (dictgen::InsensitiveStr::Ascii("assuing"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("desctroyer"), &["destroyer"]),
            (dictgen::InsensitiveStr::Ascii("queenland"), &["queensland"]),
            (dictgen::InsensitiveStr::Ascii("geneartes"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("funcionality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("neeeds"), &["needs"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunatelly"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructoin"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("swansoon"), &["swanson"]),
            (dictgen::InsensitiveStr::Ascii("systmes"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("cosoled"), &["consoled"]),
            (
                dictgen::InsensitiveStr::Ascii("involunatry"),
                &["involuntary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmacias"),
                &["pharmacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("musil"),
                &["muscle", "mussel"],
            ),
            (dictgen::InsensitiveStr::Ascii("binnary"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("sourcd"),
                &["sourced", "source"],
            ),
            (dictgen::InsensitiveStr::Ascii("keepint"), &["keeping"]),
            (
                dictgen::InsensitiveStr::Ascii("precedessors"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervewing"),
                &["intervening"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficiied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("cencretely"),
                &["concretely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responesiveness"),
                &["responsiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("sizin"), &["sizing"]),
            (
                dictgen::InsensitiveStr::Ascii("confrentational"),
                &["confrontational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbacl"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("rosettta"), &["rosetta"]),
            (
                dictgen::InsensitiveStr::Ascii("improtance"),
                &["importance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insturments"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tuncations"),
                &["truncations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breatsfeeding"),
                &["breastfeeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("clipbaord"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("modificatioon"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarranteing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapdive"), &["adaptive"]),
            (dictgen::InsensitiveStr::Ascii("visuale"), &["visual"]),
            (
                dictgen::InsensitiveStr::Ascii("stablilizing"),
                &["stabilizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("consistens"), &["consists"]),
            (
                dictgen::InsensitiveStr::Ascii("cehckpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resvervation"),
                &["reservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declearations"),
                &["declarations", "decelerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetify"), &["certify"]),
            (dictgen::InsensitiveStr::Ascii("dificulty"), &["difficulty"]),
            (
                dictgen::InsensitiveStr::Ascii("combiantors"),
                &["combinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("follod"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("interrogatin"),
                &["interrogation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglegting"),
                &["neglecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visuable"),
                &["visual", "visible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandchilden"),
                &["grandchildren"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificaction"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("purpuses"), &["purposes"]),
            (dictgen::InsensitiveStr::Ascii("rugters"), &["rutgers"]),
            (dictgen::InsensitiveStr::Ascii("builldings"), &["buildings"]),
            (dictgen::InsensitiveStr::Ascii("shaddowed"), &["shadowed"]),
            (dictgen::InsensitiveStr::Ascii("curch"), &["church"]),
            (dictgen::InsensitiveStr::Ascii("convax"), &["convex"]),
            (
                dictgen::InsensitiveStr::Ascii("incoporates"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetlics"), &["celtics"]),
            (
                dictgen::InsensitiveStr::Ascii("apprarently"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralelised"),
                &["parallelised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retardating"),
                &["retardation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallellizing"),
                &["parallelizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependeded"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("alterady"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("wekends"), &["weekends"]),
            (dictgen::InsensitiveStr::Ascii("collationg"), &["collation"]),
            (dictgen::InsensitiveStr::Ascii("essentals"), &["essentials"]),
            (
                dictgen::InsensitiveStr::Ascii("tesselators"),
                &["tessellators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dusgustingly"),
                &["disgustingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singoled"),
                &["signaled", "singled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("televisivo"),
                &["television"],
            ),
            (dictgen::InsensitiveStr::Ascii("discove"), &["discover"]),
            (dictgen::InsensitiveStr::Ascii("igorned"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("phongraph"), &["phonograph"]),
            (
                dictgen::InsensitiveStr::Ascii("referenceing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("latecny"), &["latency"]),
            (
                dictgen::InsensitiveStr::Ascii("undergating"),
                &["undertaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undetecable"),
                &["undetectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicatins"),
                &["applicants", "applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqare"), &["square"]),
            (dictgen::InsensitiveStr::Ascii("anotrher"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("opporunity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagiarizm"),
                &["plagiarism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triger"),
                &["trigger", "tiger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remaind"),
                &["remained", "remind"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("drunkeness"),
                &["drunkenness"],
            ),
            (dictgen::InsensitiveStr::Ascii("compatble"), &["compatible"]),
            (dictgen::InsensitiveStr::Ascii("synonim"), &["synonym"]),
            (dictgen::InsensitiveStr::Ascii("vaildated"), &["validated"]),
            (
                dictgen::InsensitiveStr::Ascii("traditonal"),
                &["traditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attenuatin"),
                &["attenuating", "attenuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictrionaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("inscets"), &["insects"]),
            (dictgen::InsensitiveStr::Ascii("extreemes"), &["extremes"]),
            (dictgen::InsensitiveStr::Ascii("thos"), &["those", "this"]),
            (dictgen::InsensitiveStr::Ascii("housr"), &["hours", "house"]),
            (dictgen::InsensitiveStr::Ascii("preapared"), &["prepared"]),
            (dictgen::InsensitiveStr::Ascii("travling"), &["traveling"]),
            (dictgen::InsensitiveStr::Ascii("ddoging"), &["dodging"]),
            (
                dictgen::InsensitiveStr::Ascii("tendencije"),
                &["tendencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragarph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("advanaces"), &["advances"]),
            (
                dictgen::InsensitiveStr::Ascii("furthemore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interferred"),
                &["interfere", "interfered"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiith"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("overlfow"), &["overflow"]),
            (dictgen::InsensitiveStr::Ascii("juadism"), &["judaism"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregater"),
                &["aggregator", "aggregated", "aggregates", "aggregate"],
            ),
            (dictgen::InsensitiveStr::Ascii("emiting"), &["emitting"]),
            (dictgen::InsensitiveStr::Ascii("nees"), &["needs"]),
            (
                dictgen::InsensitiveStr::Ascii("shamelessely"),
                &["shamelessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectiing"),
                &["collecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("storege"), &["storage"]),
            (dictgen::InsensitiveStr::Ascii("databas"), &["database"]),
            (dictgen::InsensitiveStr::Ascii("exatract"), &["extract"]),
            (dictgen::InsensitiveStr::Ascii("leack"), &["leak"]),
            (dictgen::InsensitiveStr::Ascii("naame"), &["name"]),
            (dictgen::InsensitiveStr::Ascii("sovereing"), &["sovereign"]),
            (
                dictgen::InsensitiveStr::Ascii("hurse"),
                &["hearse", "nurse"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimzed"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("civilazations"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("treate"), &["treat"]),
            (
                dictgen::InsensitiveStr::Ascii("certifiated"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("abadonded"), &["abandoned"]),
            (
                dictgen::InsensitiveStr::Ascii("initiazation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("definad"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("destrying"), &["destroying"]),
            (dictgen::InsensitiveStr::Ascii("biddings"), &["bidding"]),
            (dictgen::InsensitiveStr::Ascii("curiser"), &["cruiser"]),
            (
                dictgen::InsensitiveStr::Ascii("simetriclly"),
                &["symmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanims"),
                &["mechanism", "mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charnge"),
                &["change", "charge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeiences"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("templtes"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("nkwo"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("curisve"), &["cursive"]),
            (dictgen::InsensitiveStr::Ascii("preadtor"), &["predator"]),
            (dictgen::InsensitiveStr::Ascii("rectangel"), &["rectangle"]),
            (dictgen::InsensitiveStr::Ascii("oepned"), &["opened"]),
            (dictgen::InsensitiveStr::Ascii("sourcees"), &["sources"]),
            (
                dictgen::InsensitiveStr::Ascii("predominanty"),
                &["predominantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("stroyline"), &["storyline"]),
            (dictgen::InsensitiveStr::Ascii("colector"), &["collector"]),
            (
                dictgen::InsensitiveStr::Ascii("renegothiate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubances"),
                &["disturbance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geomterically"),
                &["geometrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("allinace"), &["alliance"]),
            (
                dictgen::InsensitiveStr::Ascii("forcasters"),
                &["forecasters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("navigaton"),
                &["navigation", "navigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregatet"),
                &["aggregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chandaleers"),
                &["chandeliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implicacion"),
                &["implication"],
            ),
            (dictgen::InsensitiveStr::Ascii("quwesant"), &["croissant"]),
            (
                dictgen::InsensitiveStr::Ascii("procalimed"),
                &["proclaimed"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaltion"), &["deflation"]),
            (dictgen::InsensitiveStr::Ascii("harras"), &["harass"]),
            (dictgen::InsensitiveStr::Ascii("formattted"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("calclate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("retaliers"), &["retailers"]),
            (
                dictgen::InsensitiveStr::Ascii("imporvements"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("juming"), &["jumping"]),
            (dictgen::InsensitiveStr::Ascii("receieved"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("mormonisim"), &["mormonism"]),
            (
                dictgen::InsensitiveStr::Ascii("incremnetally"),
                &["incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arhitecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registrered"),
                &["registered"],
            ),
            (dictgen::InsensitiveStr::Ascii("liasons"), &["liaisons"]),
            (dictgen::InsensitiveStr::Ascii("blured"), &["blurred"]),
            (dictgen::InsensitiveStr::Ascii("disaprity"), &["disparity"]),
            (dictgen::InsensitiveStr::Ascii("secenario"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("prixies"),
                &["proxies", "pixies", "privies"],
            ),
            (dictgen::InsensitiveStr::Ascii("secne"), &["scene"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisitir"),
                &["inquisitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrealitic"),
                &["unrealistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("argreement"), &["agreement"]),
            (dictgen::InsensitiveStr::Ascii("flaried"), &["flaired"]),
            (
                dictgen::InsensitiveStr::Ascii("initailising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrugh"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("utlimate"), &["ultimate"]),
            (
                dictgen::InsensitiveStr::Ascii("incidencies"),
                &["incidence"],
            ),
            (dictgen::InsensitiveStr::Ascii("singlers"), &["singles"]),
            (dictgen::InsensitiveStr::Ascii("appicant"), &["applicant"]),
            (dictgen::InsensitiveStr::Ascii("routinly"), &["routinely"]),
            (dictgen::InsensitiveStr::Ascii("wrokers"), &["workers"]),
            (dictgen::InsensitiveStr::Ascii("hastable"), &["hashtable"]),
            (
                dictgen::InsensitiveStr::Ascii("superchager"),
                &["supercharger"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepeator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("ethnaol"), &["ethanol"]),
            (dictgen::InsensitiveStr::Ascii("srtting"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("jagwar"), &["jaguar"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimialr"),
                &["dissimilar"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgustose"), &["disgusts"]),
            (dictgen::InsensitiveStr::Ascii("custoner"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("alternitavely"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("alisased"), &["aliased"]),
            (
                dictgen::InsensitiveStr::Ascii("cheetsheets"),
                &["cheatsheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vetinarian"),
                &["veterinarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("normaize"), &["normalize"]),
            (
                dictgen::InsensitiveStr::Ascii("shouders"),
                &["shoulders", "shudders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconasence"),
                &["reconnaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("migates"),
                &["migrates", "mitigates"],
            ),
            (dictgen::InsensitiveStr::Ascii("persewed"), &["pursued"]),
            (
                dictgen::InsensitiveStr::Ascii("concatentate"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveything"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("thoerem"), &["theorem"]),
            (
                dictgen::InsensitiveStr::Ascii("cirucmference"),
                &["circumference"],
            ),
            (dictgen::InsensitiveStr::Ascii("declres"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("recuriting"),
                &["recruiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorhytmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("compreso"), &["compressor"]),
            (
                dictgen::InsensitiveStr::Ascii("displacments"),
                &["displacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("amneisa"), &["amnesia"]),
            (
                dictgen::InsensitiveStr::Ascii("intellegently"),
                &["intelligently"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperable"), &["separable"]),
            (dictgen::InsensitiveStr::Ascii("comptue"), &["compute"]),
            (dictgen::InsensitiveStr::Ascii("tocuhpad"), &["touchpad"]),
            (dictgen::InsensitiveStr::Ascii("mechnanic"), &["mechanic"]),
            (dictgen::InsensitiveStr::Ascii("berween"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("compabitle"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("submisson"), &["submission"]),
            (dictgen::InsensitiveStr::Ascii("goemetric"), &["geometric"]),
            (dictgen::InsensitiveStr::Ascii("secific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("atmoically"),
                &["atomically"],
            ),
            (dictgen::InsensitiveStr::Ascii("easilest"), &["easiest"]),
            (
                dictgen::InsensitiveStr::Ascii("experimdnting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("insatlls"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("cmulative"), &["cumulative"]),
            (
                dictgen::InsensitiveStr::Ascii("nescessarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minue"),
                &["menu", "minus", "minute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("krpytonite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critisicsm"),
                &["criticisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpritation"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("moduleso"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("exchanes"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("operants"), &["operands"]),
            (
                dictgen::InsensitiveStr::Ascii("mantanined"),
                &["maintained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortuneatly"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependended"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("acoostic"), &["acoustic"]),
            (
                dictgen::InsensitiveStr::Ascii("invesitgating"),
                &["investigating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interchangeablely"),
                &["interchangeably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construccion"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("securtiy"), &["security"]),
            (
                dictgen::InsensitiveStr::Ascii("differantiators"),
                &["differentiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("contribte"), &["contribute"]),
            (
                dictgen::InsensitiveStr::Ascii("applicaitions"),
                &["applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corinthinans"),
                &["corinthians"],
            ),
            (dictgen::InsensitiveStr::Ascii("doubths"), &["doubts"]),
            (
                dictgen::InsensitiveStr::Ascii("insuficiency"),
                &["insufficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("removce"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("distrobution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("clonable"), &["cloneable"]),
            (
                dictgen::InsensitiveStr::Ascii("unreprocible"),
                &["unreproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("withous"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("reconncting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("norhteast"), &["northeast"]),
            (
                dictgen::InsensitiveStr::Ascii("demoliting"),
                &["demolition"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordeon"), &["accordion"]),
            (dictgen::InsensitiveStr::Ascii("combinder"), &["combiner"]),
            (dictgen::InsensitiveStr::Ascii("nedle"), &["needle"]),
            (
                dictgen::InsensitiveStr::Ascii("discontinuted"),
                &["discontinued"],
            ),
            (dictgen::InsensitiveStr::Ascii("distinced"), &["distanced"]),
            (dictgen::InsensitiveStr::Ascii("heree"), &["here"]),
            (dictgen::InsensitiveStr::Ascii("interafce"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("diectors"), &["directors"]),
            (
                dictgen::InsensitiveStr::Ascii("westminter"),
                &["westminster"],
            ),
            (dictgen::InsensitiveStr::Ascii("fortunae"), &["fortune"]),
            (
                dictgen::InsensitiveStr::Ascii("contreversy"),
                &["controversy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceeptably"),
                &["acceptably"],
            ),
            (dictgen::InsensitiveStr::Ascii("depection"), &["deception"]),
            (
                dictgen::InsensitiveStr::Ascii("specialsts"),
                &["specialists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correllates"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("forigven"), &["forgiven"]),
            (
                dictgen::InsensitiveStr::Ascii("autocorect"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("awating"), &["awaiting"]),
            (dictgen::InsensitiveStr::Ascii("cocnerns"), &["concerns"]),
            (
                dictgen::InsensitiveStr::Ascii("interrutions"),
                &["interruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("opperand"), &["operand"]),
            (
                dictgen::InsensitiveStr::Ascii("narcississt"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("inidcate"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("detuschland"),
                &["deutschland"],
            ),
            (dictgen::InsensitiveStr::Ascii("compltion"), &["completion"]),
            (dictgen::InsensitiveStr::Ascii("colleages"), &["colleagues"]),
            (dictgen::InsensitiveStr::Ascii("exampels"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("udater"),
                &["updater", "dater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nominatons"),
                &["nominations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("californiaman"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("togeterness"),
                &["togetherness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliactes"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("australa"), &["australian"]),
            (dictgen::InsensitiveStr::Ascii("maintaner"), &["maintainer"]),
            (dictgen::InsensitiveStr::Ascii("somethiong"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("gettig"), &["getting"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimizing"),
                &["optimizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememtations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("repsoted"), &["reposted"]),
            (dictgen::InsensitiveStr::Ascii("radaince"), &["radiance"]),
            (dictgen::InsensitiveStr::Ascii("desltop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("innecesarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multithreded"),
                &["multithreaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containined"),
                &["contained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demostrates"),
                &["demonstrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("terfforms"), &["terraforms"]),
            (dictgen::InsensitiveStr::Ascii("binay"), &["binary"]),
            (dictgen::InsensitiveStr::Ascii("polygnon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("teturns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("assymtotes"),
                &["asymptotes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsitutute"),
                &["substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("symetriy"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("inobtrusive"),
                &["unobtrusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disruptness"),
                &["disruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotying"),
                &["stereotyping"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorritm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("contaniners"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envirnoment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalsced"), &["coalesced"]),
            (dictgen::InsensitiveStr::Ascii("sufocate"), &["suffocate"]),
            (
                dictgen::InsensitiveStr::Ascii("referense"),
                &["references", "reference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regestration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("corerct"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("coponents"), &["components"]),
            (dictgen::InsensitiveStr::Ascii("maclolm"), &["malcolm"]),
            (dictgen::InsensitiveStr::Ascii("solutons"), &["solutions"]),
            (dictgen::InsensitiveStr::Ascii("pebblets"), &["pebbles"]),
            (dictgen::InsensitiveStr::Ascii("hotshoot"), &["hotshot"]),
            (
                dictgen::InsensitiveStr::Ascii("acccounting"),
                &["accounting"],
            ),
            (dictgen::InsensitiveStr::Ascii("captacity"), &["capacity"]),
            (dictgen::InsensitiveStr::Ascii("traceer"), &["tracer"]),
            (dictgen::InsensitiveStr::Ascii("illuminai"), &["illuminati"]),
            (
                dictgen::InsensitiveStr::Ascii("correspdoning"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appretiation"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("contins"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("attetntion"), &["attention"]),
            (
                dictgen::InsensitiveStr::Ascii("disconeccts"),
                &["disconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentacion"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificaion"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluculators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consituting"),
                &["constituting"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanuches"), &["launches"]),
            (
                dictgen::InsensitiveStr::Ascii("systematicaly"),
                &["systematically"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticts"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("intoto"), &["into"]),
            (
                dictgen::InsensitiveStr::Ascii("probabilies"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inidividual"),
                &["individual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastrofies"),
                &["catastrophes"],
            ),
            (dictgen::InsensitiveStr::Ascii("exploreres"), &["explorers"]),
            (
                dictgen::InsensitiveStr::Ascii("opeaarations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("spageti"), &["spaghetti"]),
            (
                dictgen::InsensitiveStr::Ascii("diffreents"),
                &["different", "difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("whietlist"), &["whitelist"]),
            (
                dictgen::InsensitiveStr::Ascii("unidentifiedly"),
                &["unidentified"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarchim"), &["anarchism"]),
            (
                dictgen::InsensitiveStr::Ascii("introspectivo"),
                &["introspection"],
            ),
            (dictgen::InsensitiveStr::Ascii("marshalls"), &["marshals"]),
            (
                dictgen::InsensitiveStr::Ascii("architetcural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camouglage"),
                &["camouflage"],
            ),
            (dictgen::InsensitiveStr::Ascii("boudnary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("conbinations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indistinguisable"),
                &["indistinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("euqal"), &["equal"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaited"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("rrrors"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("subscritption"),
                &["subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorhytmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predifined"),
                &["predefined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionalisim"),
                &["professionalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bounbdaries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("inerface"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("hases"), &["hashes"]),
            (
                dictgen::InsensitiveStr::Ascii("helathcare"),
                &["healthcare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interviewes"),
                &["interviews"],
            ),
            (dictgen::InsensitiveStr::Ascii("kerenls"), &["kernels"]),
            (dictgen::InsensitiveStr::Ascii("cllouding"), &["clouding"]),
            (dictgen::InsensitiveStr::Ascii("epitomie"), &["epitome"]),
            (
                dictgen::InsensitiveStr::Ascii("paralelization"),
                &["parallelization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipualtive"),
                &["manipulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplciation"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentiations"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomadations"),
                &["accommodations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indigenius"),
                &["indigenous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diskrimination"),
                &["discrimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disaggretating"),
                &["disaggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexpersions"),
                &["subexpressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("employeer"), &["employer"]),
            (dictgen::InsensitiveStr::Ascii("acoording"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("politicains"),
                &["politicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("modellinng"), &["modelling"]),
            (
                dictgen::InsensitiveStr::Ascii("intimitaly"),
                &["intimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreeably"), &["agreeably"]),
            (dictgen::InsensitiveStr::Ascii("corectly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("cange"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("metalurgy"), &["metallurgy"]),
            (
                dictgen::InsensitiveStr::Ascii("prevantative"),
                &["preventative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achievment"),
                &["achievement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribtion"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("saplin"), &["sapling"]),
            (dictgen::InsensitiveStr::Ascii("sleection"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("evrythign"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("colered"), &["colored"]),
            (dictgen::InsensitiveStr::Ascii("familliar"), &["familiar"]),
            (dictgen::InsensitiveStr::Ascii("congnition"), &["cognition"]),
            (dictgen::InsensitiveStr::Ascii("intilise"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("travese"), &["traverse"]),
            (
                dictgen::InsensitiveStr::Ascii("apprpriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsnked"), &["thanked"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreagating"),
                &["aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("writen"), &["written"]),
            (
                dictgen::InsensitiveStr::Ascii("uncommpresion"),
                &["uncompression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncontrollabe"),
                &["uncontrollable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provde"),
                &["provide", "proved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregressions"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibilityt"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("monopols"), &["monopolies"]),
            (dictgen::InsensitiveStr::Ascii("canoncal"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("phycisian"), &["physician"]),
            (
                dictgen::InsensitiveStr::Ascii("expressief"),
                &["expressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialsies"),
                &["serialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ajcencies"),
                &["adjacencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretted"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("queeue"), &["queue"]),
            (
                dictgen::InsensitiveStr::Ascii("sesitiveness"),
                &["sensitiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cocktailers"),
                &["cocktails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailizers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("refereshes"), &["refreshes"]),
            (dictgen::InsensitiveStr::Ascii("consited"), &["consisted"]),
            (
                dictgen::InsensitiveStr::Ascii("homegenous"),
                &["homogenous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backslashs"),
                &["backslashes"],
            ),
            (dictgen::InsensitiveStr::Ascii("summurs"), &["summers"]),
            (
                dictgen::InsensitiveStr::Ascii("pertinately"),
                &["pertinently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motivaiton"),
                &["motivations", "motivation"],
            ),
            (dictgen::InsensitiveStr::Ascii("imapge"), &["image"]),
            (dictgen::InsensitiveStr::Ascii("delvery"), &["delivery"]),
            (
                dictgen::InsensitiveStr::Ascii("skipt"),
                &["skip", "skype", "skipped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialialiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contravenin"),
                &["contravening"],
            ),
            (dictgen::InsensitiveStr::Ascii("soureces"), &["sources"]),
            (
                dictgen::InsensitiveStr::Ascii("tempraarily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concaternate"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifecyles"), &["lifecycles"]),
            (
                dictgen::InsensitiveStr::Ascii("underestemated"),
                &["underestimated"],
            ),
            (dictgen::InsensitiveStr::Ascii("supspect"), &["suspect"]),
            (
                dictgen::InsensitiveStr::Ascii("consilidate"),
                &["consolidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("poducts"), &["products"]),
            (
                dictgen::InsensitiveStr::Ascii("alteranting"),
                &["alternating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistrubution"),
                &["redistribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("attement"), &["attempt"]),
            (
                dictgen::InsensitiveStr::Ascii("reintepret"),
                &["reinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("chello"), &["cello"]),
            (
                dictgen::InsensitiveStr::Ascii("nofification"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("implmement"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("settingss"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("relocatin"),
                &["relocating", "relocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reportoire"),
                &["repertoire"],
            ),
            (dictgen::InsensitiveStr::Ascii("fethed"), &["fetched"]),
            (
                dictgen::InsensitiveStr::Ascii("algoithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inatruction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("wondeful"), &["wonderful"]),
            (dictgen::InsensitiveStr::Ascii("scupt"), &["sculpt"]),
            (dictgen::InsensitiveStr::Ascii("stae"), &["state"]),
            (
                dictgen::InsensitiveStr::Ascii("inevidably"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("blcoked"), &["blocked"]),
            (
                dictgen::InsensitiveStr::Ascii("depreacted"),
                &["deprecated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reusin"),
                &["reusing", "resin"],
            ),
            (dictgen::InsensitiveStr::Ascii("miinimise"), &["minimise"]),
            (
                dictgen::InsensitiveStr::Ascii("dissatisfed"),
                &["dissatisfied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifeis"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("frametyp"), &["frametype"]),
            (
                dictgen::InsensitiveStr::Ascii("incopatible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("backgrace"), &["backtrace"]),
            (
                dictgen::InsensitiveStr::Ascii("compeltion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insepection"),
                &["inspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intillectuals"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devored"),
                &["devoted", "devoured"],
            ),
            (dictgen::InsensitiveStr::Ascii("folfer"), &["folder"]),
            (dictgen::InsensitiveStr::Ascii("requeums"), &["requiems"]),
            (dictgen::InsensitiveStr::Ascii("plugings"), &["plugins"]),
            (dictgen::InsensitiveStr::Ascii("misandy"), &["misandry"]),
            (dictgen::InsensitiveStr::Ascii("relatibe"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("inverleaved"),
                &["interleaved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interoptability"),
                &["interoperability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsibstantiated"),
                &["unsubstantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplmentation"),
                &["reimplementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vernecular"),
                &["vernacular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapprouves"),
                &["disapproves"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorgs"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("otpimizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("modidfication"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("saerching"), &["searching"]),
            (dictgen::InsensitiveStr::Ascii("sciript"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("exculdes"), &["excludes"]),
            (dictgen::InsensitiveStr::Ascii("stirker"), &["striker"]),
            (
                dictgen::InsensitiveStr::Ascii("assiocating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deocrative"),
                &["decorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("difracted"), &["diffracted"]),
            (
                dictgen::InsensitiveStr::Ascii("performamces"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restarants"),
                &["restaurants"],
            ),
            (dictgen::InsensitiveStr::Ascii("maskeraid"), &["masquerade"]),
            (
                dictgen::InsensitiveStr::Ascii("necessarlily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("rejplace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("unxepected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dscription"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similarlity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("decsion"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("yau"), &["you", "yaw"]),
            (dictgen::InsensitiveStr::Ascii("willfuly"), &["willfully"]),
            (
                dictgen::InsensitiveStr::Ascii("christiantiy"),
                &["christianity"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitimans"), &["vitamins"]),
            (
                dictgen::InsensitiveStr::Ascii("enhencment"),
                &["enhancement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinative"),
                &["distinctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersting"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementastions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("umambiguous"),
                &["unambiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("claens"), &["cleans"]),
            (dictgen::InsensitiveStr::Ascii("capturin"), &["capturing"]),
            (
                dictgen::InsensitiveStr::Ascii("timestmaps"),
                &["timestamps"],
            ),
            (dictgen::InsensitiveStr::Ascii("bcucket"), &["bucket"]),
            (
                dictgen::InsensitiveStr::Ascii("exculsivity"),
                &["exclusivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regenration"),
                &["regeneration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sourcse"),
                &["sources", "source"],
            ),
            (dictgen::InsensitiveStr::Ascii("striekrs"), &["strikers"]),
            (
                dictgen::InsensitiveStr::Ascii("acticities"),
                &["activities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributuons"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychologysts"),
                &["psychologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("behabiors"), &["behaviors"]),
            (dictgen::InsensitiveStr::Ascii("surpised"), &["surprised"]),
            (dictgen::InsensitiveStr::Ascii("mirometer"), &["micrometer"]),
            (dictgen::InsensitiveStr::Ascii("cabable"), &["capable"]),
            (dictgen::InsensitiveStr::Ascii("nostaligc"), &["nostalgic"]),
            (dictgen::InsensitiveStr::Ascii("disapoint"), &["disappoint"]),
            (
                dictgen::InsensitiveStr::Ascii("juxtifications"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("channes"),
                &["channels", "chances", "changes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vigilanties"),
                &["vigilante", "vigilantes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devolopers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspodingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("nowe"), &["now"]),
            (
                dictgen::InsensitiveStr::Ascii("egistration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("virtal"), &["virtual"]),
            (dictgen::InsensitiveStr::Ascii("iteratible"), &["iterable"]),
            (
                dictgen::InsensitiveStr::Ascii("negoticators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peirodical"),
                &["periodical"],
            ),
            (dictgen::InsensitiveStr::Ascii("naibhorly"), &["neighborly"]),
            (dictgen::InsensitiveStr::Ascii("rememebers"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("anayslis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("mehtodically"),
                &["methodically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmitsion"),
                &["transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinines"), &["combines"]),
            (dictgen::InsensitiveStr::Ascii("sacale"), &["scale"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisaitons"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("exerise"), &["exercise"]),
            (
                dictgen::InsensitiveStr::Ascii("organizarte"),
                &["organizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirction"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("deflecticon"),
                &["deflection"],
            ),
            (dictgen::InsensitiveStr::Ascii("macthing"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("behvioural"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("coinsides"), &["coincides"]),
            (
                dictgen::InsensitiveStr::Ascii("precondidition"),
                &["precondition"],
            ),
            (dictgen::InsensitiveStr::Ascii("dishoner"), &["dishonored"]),
            (dictgen::InsensitiveStr::Ascii("perfomr"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("quiests"), &["quests"]),
            (dictgen::InsensitiveStr::Ascii("embbeding"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("repubicans"),
                &["republicans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porportional"),
                &["proportional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idealisitc"),
                &["idealistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resourcs"),
                &["resources", "resource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarrasingly"),
                &["embarrassingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("exportas"), &["exports"]),
            (
                dictgen::InsensitiveStr::Ascii("jouranlistic"),
                &["journalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allignement"),
                &["alignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("naybhorhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("adavances"), &["advances"]),
            (
                dictgen::InsensitiveStr::Ascii("defenitively"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ligitamacy"),
                &["legitimacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggestie"), &["suggestive"]),
            (dictgen::InsensitiveStr::Ascii("severley"), &["severely"]),
            (dictgen::InsensitiveStr::Ascii("dout"), &["doubt"]),
            (dictgen::InsensitiveStr::Ascii("mechinism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("preperations"),
                &["preparations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authroizes"),
                &["authorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("shfited"), &["shifted"]),
            (
                dictgen::InsensitiveStr::Ascii("nanseconds"),
                &["nanoseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("expicit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("specicifications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetners"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("coputes"), &["computes"]),
            (
                dictgen::InsensitiveStr::Ascii("liaisin"),
                &["liaising", "liaison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approprpiately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertarain"),
                &["libertarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesrname"), &["username"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliacations"),
                &["duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("ampliflied"), &["amplified"]),
            (
                dictgen::InsensitiveStr::Ascii("compatbility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaptee"), &["adapted"]),
            (dictgen::InsensitiveStr::Ascii("covincing"), &["convincing"]),
            (dictgen::InsensitiveStr::Ascii("reaveled"), &["revealed"]),
            (dictgen::InsensitiveStr::Ascii("begginer"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("scond"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("ketmaine"), &["ketamine"]),
            (dictgen::InsensitiveStr::Ascii("lasr"), &["laser", "last"]),
            (dictgen::InsensitiveStr::Ascii("tranient"), &["transient"]),
            (dictgen::InsensitiveStr::Ascii("jeopary"), &["jeopardy"]),
            (dictgen::InsensitiveStr::Ascii("confursing"), &["confusing"]),
            (
                dictgen::InsensitiveStr::Ascii("disposation"),
                &["disposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strangness"),
                &["strangeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("dlalog"), &["dialog"]),
            (dictgen::InsensitiveStr::Ascii("semanitcs"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("improovement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatil"), &["detail"]),
            (
                dictgen::InsensitiveStr::Ascii("confurms"),
                &["confirms", "conforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desturtors"),
                &["destructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discontinus"),
                &["discontinue", "discontinuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("graphvis"), &["graphviz"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimistic"),
                &["optimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("afair"), &["affair"]),
            (
                dictgen::InsensitiveStr::Ascii("tabulaters"),
                &["tabulators", "tabulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggergating"),
                &["aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mitgate"), &["mitigate"]),
            (
                dictgen::InsensitiveStr::Ascii("insantiations"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discimenation"),
                &["dissemination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("preriodic"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("fondamentalist"),
                &["fundamentalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("surprize"), &["surprise"]),
            (dictgen::InsensitiveStr::Ascii("morphein"), &["morphine"]),
            (dictgen::InsensitiveStr::Ascii("icrement"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("remmove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("raeding"), &["reading"]),
            (
                dictgen::InsensitiveStr::Ascii("colelction"),
                &["collection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perputuates"),
                &["perpetuates"],
            ),
            (dictgen::InsensitiveStr::Ascii("posatives"), &["positives"]),
            (dictgen::InsensitiveStr::Ascii("autmata"), &["automata"]),
            (
                dictgen::InsensitiveStr::Ascii("commenteries"),
                &["commenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parenthasis"),
                &["parenthesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuely"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeusts"), &["requests"]),
            (
                dictgen::InsensitiveStr::Ascii("maniuplated"),
                &["manipulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conductiong"),
                &["conducting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuely"),
                &["continually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiifer"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("benj"), &["binge"]),
            (
                dictgen::InsensitiveStr::Ascii("abominaton"),
                &["abomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("jepordize"), &["jeopardize"]),
            (dictgen::InsensitiveStr::Ascii("purpus"), &["purpose"]),
            (dictgen::InsensitiveStr::Ascii("physision"), &["physician"]),
            (dictgen::InsensitiveStr::Ascii("stiplers"), &["stipplers"]),
            (
                dictgen::InsensitiveStr::Ascii("decomperssor"),
                &["decompressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitraily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("revrese"), &["reverse"]),
            (
                dictgen::InsensitiveStr::Ascii("partion"),
                &["partition", "portion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configruated"),
                &["configured", "configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nueral"),
                &["neural", "numeral"],
            ),
            (dictgen::InsensitiveStr::Ascii("senstive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("scepture"), &["sceptre"]),
            (dictgen::InsensitiveStr::Ascii("enveloppen"), &["envelope"]),
            (dictgen::InsensitiveStr::Ascii("obvilion"), &["oblivion"]),
            (dictgen::InsensitiveStr::Ascii("naploeon"), &["napoleon"]),
            (
                dictgen::InsensitiveStr::Ascii("replases"),
                &["replaces", "relapses", "rephases"],
            ),
            (dictgen::InsensitiveStr::Ascii("bypas"), &["bypass"]),
            (dictgen::InsensitiveStr::Ascii("faulsure"), &["failure"]),
            (
                dictgen::InsensitiveStr::Ascii("flutterhsy"),
                &["fluttershy"],
            ),
            (dictgen::InsensitiveStr::Ascii("procoess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("sitautional"),
                &["situational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amplifiying"),
                &["amplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("masacra"), &["mascara"]),
            (
                dictgen::InsensitiveStr::Ascii("appropritely"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("packade"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("innosence"), &["innocence"]),
            (dictgen::InsensitiveStr::Ascii("multilples"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("qualificatons"),
                &["qualifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceduled"), &["scheduled"]),
            (
                dictgen::InsensitiveStr::Ascii("reproducabely"),
                &["reproducibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("clsuters"), &["clusters"]),
            (dictgen::InsensitiveStr::Ascii("detectected"), &["detected"]),
            (dictgen::InsensitiveStr::Ascii("ubutunu"), &["ubuntu"]),
            (
                dictgen::InsensitiveStr::Ascii("exprts"),
                &["exports", "experts", "expats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aestethics"),
                &["aesthetics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingrediens"),
                &["ingredients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnported"),
                &["transported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occuppational"),
                &["occupational"],
            ),
            (dictgen::InsensitiveStr::Ascii("rolepaly"), &["roleplay"]),
            (
                dictgen::InsensitiveStr::Ascii("policitian"),
                &["politician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlightenting"),
                &["enlightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motorcylces"),
                &["motorcycles"],
            ),
            (dictgen::InsensitiveStr::Ascii("resizeing"), &["resizing"]),
            (dictgen::InsensitiveStr::Ascii("colourd"), &["coloured"]),
            (
                dictgen::InsensitiveStr::Ascii("unsolicted"),
                &["unsolicited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychedelicness"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("bottem"), &["bottom"]),
            (dictgen::InsensitiveStr::Ascii("flext"), &["flex"]),
            (dictgen::InsensitiveStr::Ascii("ingnored"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("kwoledgeable"),
                &["knowledgeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phsyiological"),
                &["physiological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spleling"),
                &["spelling", "spieling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interals"),
                &["internals", "intervals", "integrals"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylic"), &["cyclic"]),
            (
                dictgen::InsensitiveStr::Ascii("comperable"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("navgiated"), &["navigated"]),
            (
                dictgen::InsensitiveStr::Ascii("spefixiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoziates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppresions"),
                &["suppression", "suppressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("linerisations"),
                &["linearisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("commiter"), &["committer"]),
            (dictgen::InsensitiveStr::Ascii("buton"), &["button"]),
            (
                dictgen::InsensitiveStr::Ascii("generalizacion"),
                &["generalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloow"), &["allow"]),
            (dictgen::InsensitiveStr::Ascii("rescources"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("incorporting"),
                &["incorporating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encycolpedia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("landspaces"),
                &["landscapes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optiona"),
                &["optional", "options"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caustion"),
                &["caution", "causation"],
            ),
            (dictgen::InsensitiveStr::Ascii("diamonnd"), &["diamond"]),
            (
                dictgen::InsensitiveStr::Ascii("significanly"),
                &["significantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prodominantly"),
                &["predominantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("olmypics"), &["olympics"]),
            (
                dictgen::InsensitiveStr::Ascii("porportionally"),
                &["proportionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contraints"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifilers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("asronist"), &["arsonist"]),
            (
                dictgen::InsensitiveStr::Ascii("commpressed"),
                &["compressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regnerated"),
                &["regenerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfurs"), &["prefers"]),
            (
                dictgen::InsensitiveStr::Ascii("mulltiplier"),
                &["multiplier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesects"),
                &["intersects"],
            ),
            (dictgen::InsensitiveStr::Ascii("watiers"), &["waiters"]),
            (dictgen::InsensitiveStr::Ascii("plasement"), &["placement"]),
            (
                dictgen::InsensitiveStr::Ascii("asendence"),
                &["ascendence", "ascendance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collataral"),
                &["collateral"],
            ),
            (dictgen::InsensitiveStr::Ascii("soemwhere"), &["somewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("dismatches"),
                &["mismatches", "dispatches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatoin"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("amoutn"), &["amount"]),
            (dictgen::InsensitiveStr::Ascii("monstrum"), &["monster"]),
            (
                dictgen::InsensitiveStr::Ascii("fould"),
                &["foul", "fouled", "fold", "found", "could", "would"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnetly"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("delievery"), &["delivery"]),
            (
                dictgen::InsensitiveStr::Ascii("elementery"),
                &["elementary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parethesis"),
                &["parenthesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authobiographic"),
                &["autobiographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sriraca"), &["sriracha"]),
            (
                dictgen::InsensitiveStr::Ascii("dimentsions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instsallation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wothout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("protaganists"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("violoated"), &["violated"]),
            (dictgen::InsensitiveStr::Ascii("starteld"), &["startled"]),
            (
                dictgen::InsensitiveStr::Ascii("astronomia"),
                &["astronomical"],
            ),
            (dictgen::InsensitiveStr::Ascii("verchew"), &["virtue"]),
            (
                dictgen::InsensitiveStr::Ascii("adderssing"),
                &["addressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understadable"),
                &["understandable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publisehrs"),
                &["publishers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reloades"),
                &["reloads", "reloaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docummenting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("funktion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("istalled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("replacite"), &["replicate"]),
            (dictgen::InsensitiveStr::Ascii("placments"), &["placements"]),
            (dictgen::InsensitiveStr::Ascii("parmesaen"), &["parmesan"]),
            (dictgen::InsensitiveStr::Ascii("crossair"), &["crosshair"]),
            (dictgen::InsensitiveStr::Ascii("foound"), &["found"]),
            (dictgen::InsensitiveStr::Ascii("acorss"), &["across"]),
            (dictgen::InsensitiveStr::Ascii("whants"), &["wants"]),
            (dictgen::InsensitiveStr::Ascii("lauching"), &["launching"]),
            (
                dictgen::InsensitiveStr::Ascii("ligitimate"),
                &["legitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscribbing"),
                &["subscribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cna"), &["can"]),
            (
                dictgen::InsensitiveStr::Ascii("saskatechwan"),
                &["saskatchewan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleating"),
                &["completing"],
            ),
            (dictgen::InsensitiveStr::Ascii("lituhania"), &["lithuania"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibiltiy"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatablility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("donains"), &["domains"]),
            (dictgen::InsensitiveStr::Ascii("wolrdview"), &["worldview"]),
            (dictgen::InsensitiveStr::Ascii("agreemnt"), &["agreement"]),
            (
                dictgen::InsensitiveStr::Ascii("coallecence"),
                &["coalescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooven"), &["proven"]),
            (dictgen::InsensitiveStr::Ascii("recctangle"), &["rectangle"]),
            (
                dictgen::InsensitiveStr::Ascii("mesmerizin"),
                &["mesmerizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("thougths"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("transcrits"),
                &["transcripts"],
            ),
            (dictgen::InsensitiveStr::Ascii("shorlty"), &["shortly"]),
            (
                dictgen::InsensitiveStr::Ascii("dictionarys"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implenmentation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exposion"), &["explosion"]),
            (dictgen::InsensitiveStr::Ascii("countes"), &["counters"]),
            (dictgen::InsensitiveStr::Ascii("assuem"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("palyable"), &["playable"]),
            (dictgen::InsensitiveStr::Ascii("roundtip"), &["roundtrip"]),
            (
                dictgen::InsensitiveStr::Ascii("transations"),
                &["transitions", "transactions", "translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("shedulers"), &["schedulers"]),
            (
                dictgen::InsensitiveStr::Ascii("invove"),
                &["involve", "invoke"],
            ),
            (dictgen::InsensitiveStr::Ascii("tasliman"), &["talisman"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfity"),
                &["identify", "identity"],
            ),
            (dictgen::InsensitiveStr::Ascii("middlware"), &["middleware"]),
            (
                dictgen::InsensitiveStr::Ascii("cacluations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("uernames"), &["usernames"]),
            (dictgen::InsensitiveStr::Ascii("appluad"), &["applaud"]),
            (
                dictgen::InsensitiveStr::Ascii("informtion"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critisiscm"),
                &["criticisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("standinds"), &["standings"]),
            (
                dictgen::InsensitiveStr::Ascii("suggestionss"),
                &["suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contraintes"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("earhtbound"),
                &["earthbound"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnity"), &["community"]),
            (
                dictgen::InsensitiveStr::Ascii("philadeplhia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worl"),
                &["world", "work", "whorl"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgasmos"), &["orgasms"]),
            (
                dictgen::InsensitiveStr::Ascii("countrycide"),
                &["countryside"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivually"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("victemize"), &["victimize"]),
            (dictgen::InsensitiveStr::Ascii("ephipany"), &["epiphany"]),
            (dictgen::InsensitiveStr::Ascii("assamble"), &["assemble"]),
            (dictgen::InsensitiveStr::Ascii("branchs"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("dessigner"), &["designer"]),
            (
                dictgen::InsensitiveStr::Ascii("misaligments"),
                &["misalignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("slaverly"), &["slavery"]),
            (
                dictgen::InsensitiveStr::Ascii("chaces"),
                &["chances", "caches", "chases"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrokbench"), &["workbench"]),
            (dictgen::InsensitiveStr::Ascii("tolerants"), &["tolerates"]),
            (
                dictgen::InsensitiveStr::Ascii("possibity"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("exciption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("oepns"), &["opens"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiatin"),
                &["negotiations", "negotiating", "negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectuors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("assihning"), &["assigning"]),
            (
                dictgen::InsensitiveStr::Ascii("pontuation"),
                &["punctuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("meterials"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("oversimplificating"),
                &["oversimplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronounciations"),
                &["pronunciations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("medicatons"),
                &["medications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportaing"),
                &["incorporating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captalized"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moderninations"),
                &["modernizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbergh"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("glpyh"), &["glyph"]),
            (dictgen::InsensitiveStr::Ascii("anyoneis"), &["anyones"]),
            (
                dictgen::InsensitiveStr::Ascii("regconizer"),
                &["recognizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenaticate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrebution"),
                &["retribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("alege"), &["allege"]),
            (
                dictgen::InsensitiveStr::Ascii("complession"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monorchrome"),
                &["monochrome"],
            ),
            (dictgen::InsensitiveStr::Ascii("lincense"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("ddelete"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("conflit"), &["conflict"]),
            (
                dictgen::InsensitiveStr::Ascii("experimdntally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("subruban"), &["suburban"]),
            (dictgen::InsensitiveStr::Ascii("sypmathy"), &["sympathy"]),
            (dictgen::InsensitiveStr::Ascii("uplaodes"), &["uploads"]),
            (dictgen::InsensitiveStr::Ascii("balencing"), &["balancing"]),
            (dictgen::InsensitiveStr::Ascii("dleaying"), &["delaying"]),
            (dictgen::InsensitiveStr::Ascii("modulie"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("emcompassing"),
                &["encompassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("unqiue"), &["unique"]),
            (
                dictgen::InsensitiveStr::Ascii("millienaire"),
                &["millionaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqash"), &["squash"]),
            (dictgen::InsensitiveStr::Ascii("textfrme"), &["textframe"]),
            (
                dictgen::InsensitiveStr::Ascii("neckbeardus"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tesselatin"),
                &["tessellating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psuedocode"),
                &["pseudocode"],
            ),
            (dictgen::InsensitiveStr::Ascii("misspels"), &["misspells"]),
            (
                dictgen::InsensitiveStr::Ascii("ecountering"),
                &["encountering"],
            ),
            (dictgen::InsensitiveStr::Ascii("itnervals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("simpliifcation"),
                &["simplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("thatnk"), &["thank"]),
            (
                dictgen::InsensitiveStr::Ascii("precedense"),
                &["precedence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missionaires"),
                &["missionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("quesants"), &["croissants"]),
            (
                dictgen::InsensitiveStr::Ascii("jurisidction"),
                &["jurisdiction"],
            ),
            (dictgen::InsensitiveStr::Ascii("bateries"), &["batteries"]),
            (
                dictgen::InsensitiveStr::Ascii("uspported"),
                &["supported", "unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utlization"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cigerattes"),
                &["cigarettes"],
            ),
            (dictgen::InsensitiveStr::Ascii("usgin"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("handlig"), &["handling"]),
            (dictgen::InsensitiveStr::Ascii("suportted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("comsumer"), &["consumer"]),
            (dictgen::InsensitiveStr::Ascii("sudent"), &["student"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleartor"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("vociemail"), &["voicemail"]),
            (
                dictgen::InsensitiveStr::Ascii("entriy"),
                &["entry", "entries"],
            ),
            (dictgen::InsensitiveStr::Ascii("doctines"), &["doctrines"]),
            (
                dictgen::InsensitiveStr::Ascii("algotihmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("includees"), &["includes"]),
            (dictgen::InsensitiveStr::Ascii("windo"), &["window"]),
            (dictgen::InsensitiveStr::Ascii("appearning"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("wirtten"), &["written"]),
            (
                dictgen::InsensitiveStr::Ascii("phenomonon"),
                &["phenomenon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manfiested"),
                &["manifested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctiptions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("monitores"), &["monitors"]),
            (dictgen::InsensitiveStr::Ascii("safefly"), &["safely"]),
            (dictgen::InsensitiveStr::Ascii("ajacence"), &["adjacence"]),
            (
                dictgen::InsensitiveStr::Ascii("agein"),
                &["again", "ageing"],
            ),
            (dictgen::InsensitiveStr::Ascii("peusdo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("consecuitively"),
                &["consecutively"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensure"), &["censure"]),
            (dictgen::InsensitiveStr::Ascii("insistes"), &["insists"]),
            (
                dictgen::InsensitiveStr::Ascii("imbalanaced"),
                &["imbalanced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lifetsyles"),
                &["lifestyles"],
            ),
            (dictgen::InsensitiveStr::Ascii("supprts"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("dangerousely"),
                &["dangerously"],
            ),
            (dictgen::InsensitiveStr::Ascii("learnign"), &["learning"]),
            (
                dictgen::InsensitiveStr::Ascii("affixiation"),
                &["affiliation"],
            ),
            (dictgen::InsensitiveStr::Ascii("phsyics"), &["physics"]),
            (
                dictgen::InsensitiveStr::Ascii("ofisianado"),
                &["aficionado"],
            ),
            (dictgen::InsensitiveStr::Ascii("chihuahau"), &["chihuahua"]),
            (dictgen::InsensitiveStr::Ascii("unicornus"), &["unicorns"]),
            (
                dictgen::InsensitiveStr::Ascii("defitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commantator"),
                &["commentator"],
            ),
            (dictgen::InsensitiveStr::Ascii("variablwes"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("archtiectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("compilie"), &["compile"]),
            (dictgen::InsensitiveStr::Ascii("consunmer"), &["consumer"]),
            (dictgen::InsensitiveStr::Ascii("registy"), &["registry"]),
            (
                dictgen::InsensitiveStr::Ascii("strawberrry"),
                &["strawberry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtorised"),
                &["authorised"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciepient"), &["recipient"]),
            (dictgen::InsensitiveStr::Ascii("insmonia"), &["insomnia"]),
            (dictgen::InsensitiveStr::Ascii("probelms"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("dirrectional"),
                &["directional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequece"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comversions"),
                &["conversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherith"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("ghraphic"), &["graphic"]),
            (
                dictgen::InsensitiveStr::Ascii("probabilit"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precomupted"),
                &["precomputed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorythems"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtiecturally"),
                &["architecturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pestecides"),
                &["pesticides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusuively"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("obiect"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("diffrentiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("complane"), &["complacent"]),
            (
                dictgen::InsensitiveStr::Ascii("defininitively"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("macason"), &["moccasin"]),
            (
                dictgen::InsensitiveStr::Ascii("ligitation"),
                &["litigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustraits"),
                &["frustrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifforms"), &["uniforms"]),
            (dictgen::InsensitiveStr::Ascii("potrayed"), &["portrayed"]),
            (
                dictgen::InsensitiveStr::Ascii("importamtly"),
                &["importantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("flatened"), &["flattened"]),
            (dictgen::InsensitiveStr::Ascii("recorre"), &["recorder"]),
            (
                dictgen::InsensitiveStr::Ascii("agggregating"),
                &["aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("downlod"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("prepetuates"),
                &["perpetuates"],
            ),
            (dictgen::InsensitiveStr::Ascii("burgunday"), &["burgundy"]),
            (
                dictgen::InsensitiveStr::Ascii("electroltyes"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("achilleos"), &["achilles"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsciously"),
                &["unconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("divisons"), &["divisions"]),
            (
                dictgen::InsensitiveStr::Ascii("unssupported"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misrepresentaion"),
                &["misrepresentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("autio"), &["audio"]),
            (dictgen::InsensitiveStr::Ascii("benchark"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("comeptitors"),
                &["competitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requirments"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("releae"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("sligthly"), &["slightly"]),
            (
                dictgen::InsensitiveStr::Ascii("inconcistency"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("secifications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("oktober"), &["october"]),
            (
                dictgen::InsensitiveStr::Ascii("incrementas"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retet"),
                &["reset", "retest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ublication"),
                &["publication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrrespondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnfigured"), &["configured"]),
            (dictgen::InsensitiveStr::Ascii("crucibel"), &["crucible"]),
            (
                dictgen::InsensitiveStr::Ascii("syncronously"),
                &["synchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsaging"), &["unstaging"]),
            (
                dictgen::InsensitiveStr::Ascii("mangaers"),
                &["managers", "mangers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifable"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("negotator"), &["negotiator"]),
            (dictgen::InsensitiveStr::Ascii("seledtors"), &["selectors"]),
            (
                dictgen::InsensitiveStr::Ascii("istant"),
                &["instant", "distant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mississipis"),
                &["mississippi"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expceptions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tkaing"), &["taking"]),
            (dictgen::InsensitiveStr::Ascii("cleses"), &["closes"]),
            (dictgen::InsensitiveStr::Ascii("gernades"), &["grenades"]),
            (dictgen::InsensitiveStr::Ascii("domans"), &["domains"]),
            (dictgen::InsensitiveStr::Ascii("carachter"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("reposonder"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("htey"), &["they"]),
            (dictgen::InsensitiveStr::Ascii("hydorgen"), &["hydrogen"]),
            (dictgen::InsensitiveStr::Ascii("bufferring"), &["buffering"]),
            (
                dictgen::InsensitiveStr::Ascii("lonber"),
                &["longer", "loner"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuppiers"), &["occupiers"]),
            (
                dictgen::InsensitiveStr::Ascii("deseralization"),
                &["deserialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("custonize"), &["customize"]),
            (dictgen::InsensitiveStr::Ascii("pritimive"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("convervable"),
                &["conservable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presnetation"),
                &["presentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encountred"),
                &["encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceets"), &["conceits"]),
            (dictgen::InsensitiveStr::Ascii("priorty"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("environmnent"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("loactions"), &["locations"]),
            (
                dictgen::InsensitiveStr::Ascii("fozen"),
                &["frozen", "dozen"],
            ),
            (dictgen::InsensitiveStr::Ascii("implemen"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("alturism"), &["altruism"]),
            (
                dictgen::InsensitiveStr::Ascii("granuality"),
                &["granularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("theif"), &["thief"]),
            (
                dictgen::InsensitiveStr::Ascii("substantiable"),
                &["substantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experminents"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("hispter"), &["hipster"]),
            (dictgen::InsensitiveStr::Ascii("convento"), &["convention"]),
            (dictgen::InsensitiveStr::Ascii("quickier"), &["quicker"]),
            (dictgen::InsensitiveStr::Ascii("reeasers"), &["releasers"]),
            (dictgen::InsensitiveStr::Ascii("versins"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("partcipations"),
                &["participations"],
            ),
            (dictgen::InsensitiveStr::Ascii("disbaled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("momemtarily"),
                &["momentarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("summerise"), &["summarise"]),
            (dictgen::InsensitiveStr::Ascii("fianlise"), &["finalise"]),
            (
                dictgen::InsensitiveStr::Ascii("sympathiek"),
                &["sympathize"],
            ),
            (dictgen::InsensitiveStr::Ascii("genreally"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("copntroller"),
                &["controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("cooger"), &["cougar"]),
            (dictgen::InsensitiveStr::Ascii("curretn"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("examinated"), &["examined"]),
            (
                dictgen::InsensitiveStr::Ascii("measureable"),
                &["measurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misubderstanding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigaling"), &["signaling"]),
            (dictgen::InsensitiveStr::Ascii("seralizes"), &["serializes"]),
            (dictgen::InsensitiveStr::Ascii("expersing"), &["expressing"]),
            (dictgen::InsensitiveStr::Ascii("personarse"), &["personas"]),
            (
                dictgen::InsensitiveStr::Ascii("presss"),
                &["press", "presses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discret"),
                &["discrete", "discreet"],
            ),
            (dictgen::InsensitiveStr::Ascii("attachted"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("lighing"), &["lighting"]),
            (dictgen::InsensitiveStr::Ascii("calulater"), &["calculator"]),
            (dictgen::InsensitiveStr::Ascii("succed"), &["succeed"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmizations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributivo"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("proifle"), &["profile"]),
            (
                dictgen::InsensitiveStr::Ascii("acustommed"),
                &["accustomed"],
            ),
            (dictgen::InsensitiveStr::Ascii("commetns"), &["comments"]),
            (
                dictgen::InsensitiveStr::Ascii("peripherials"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assihnment"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("devy"), &["divvy"]),
            (dictgen::InsensitiveStr::Ascii("viusals"), &["visuals"]),
            (
                dictgen::InsensitiveStr::Ascii("speficying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percetange"),
                &["percentages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absoluthely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("zukeenee"), &["zucchini"]),
            (dictgen::InsensitiveStr::Ascii("subcripts"), &["subscripts"]),
            (
                dictgen::InsensitiveStr::Ascii("contenated"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sommewhat"), &["somewhat"]),
            (dictgen::InsensitiveStr::Ascii("comnmand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("contraceptivos"),
                &["contraceptives"],
            ),
            (dictgen::InsensitiveStr::Ascii("masiff"), &["massif"]),
            (dictgen::InsensitiveStr::Ascii("succecive"), &["successive"]),
            (dictgen::InsensitiveStr::Ascii("neded"), &["needed"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfitifying"),
                &["identifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comminists"),
                &["communists"],
            ),
            (dictgen::InsensitiveStr::Ascii("appoligise"), &["apologise"]),
            (
                dictgen::InsensitiveStr::Ascii("competiton"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("snyopsis"), &["synopsis"]),
            (dictgen::InsensitiveStr::Ascii("ocmplex"), &["complex"]),
            (dictgen::InsensitiveStr::Ascii("docuements"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("reseting"), &["resetting"]),
            (
                dictgen::InsensitiveStr::Ascii("correting"),
                &["correcting", "correlating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expressios"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetsts"), &["tests"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructus"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustraded"),
                &["frustrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistance"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminlly"), &["terminally"]),
            (
                dictgen::InsensitiveStr::Ascii("commemters"),
                &["commenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleands"), &["cleans"]),
            (
                dictgen::InsensitiveStr::Ascii("occsionally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sourth"), &["south"]),
            (dictgen::InsensitiveStr::Ascii("eqivalent"), &["equivalent"]),
            (dictgen::InsensitiveStr::Ascii("champange"), &["champagne"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterize"),
                &["parameterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffussive"), &["diffusive"]),
            (
                dictgen::InsensitiveStr::Ascii("atttempting"),
                &["attempting"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaningul"), &["meaningful"]),
            (dictgen::InsensitiveStr::Ascii("culller"), &["culler"]),
            (
                dictgen::InsensitiveStr::Ascii("journalistes"),
                &["journalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("leat"),
                &["lead", "leak", "least", "leaf"],
            ),
            (dictgen::InsensitiveStr::Ascii("elephantis"), &["elephants"]),
            (dictgen::InsensitiveStr::Ascii("rquires"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("comercially"),
                &["commercially"],
            ),
            (dictgen::InsensitiveStr::Ascii("fallabck"), &["fallback"]),
            (dictgen::InsensitiveStr::Ascii("humants"), &["humanist"]),
            (dictgen::InsensitiveStr::Ascii("succcess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("handeles"), &["handles"]),
            (dictgen::InsensitiveStr::Ascii("comleted"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("checlpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tansformers"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementating"),
                &["implementation", "implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aforementiond"),
                &["aforementioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misions"),
                &["missions", "minions", "visions"],
            ),
            (dictgen::InsensitiveStr::Ascii("overlcock"), &["overclock"]),
            (dictgen::InsensitiveStr::Ascii("otuput"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("socailize"), &["socialize"]),
            (
                dictgen::InsensitiveStr::Ascii("interuppting"),
                &["interrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("chocolae"), &["chocolates"]),
            (dictgen::InsensitiveStr::Ascii("alrogithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("negitiator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("typpe"), &["type"]),
            (
                dictgen::InsensitiveStr::Ascii("prohabition"),
                &["prohibition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perservered"),
                &["persevered"],
            ),
            (dictgen::InsensitiveStr::Ascii("desgins"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("eczecutives"),
                &["executives"],
            ),
            (dictgen::InsensitiveStr::Ascii("untility"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("pretection"),
                &["protection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangent"),
                &["rearrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referrenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("synoym"), &["synonym"]),
            (dictgen::InsensitiveStr::Ascii("atifacts"), &["artifacts"]),
            (dictgen::InsensitiveStr::Ascii("lanagugs"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("menber"), &["member"]),
            (dictgen::InsensitiveStr::Ascii("equeation"), &["equation"]),
            (
                dictgen::InsensitiveStr::Ascii("correspndents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onmishiences"),
                &["omnisciences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufficienntly"),
                &["sufficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("rquire"), &["require"]),
            (dictgen::InsensitiveStr::Ascii("optionel"), &["optional"]),
            (dictgen::InsensitiveStr::Ascii("pachooly"), &["patchouli"]),
            (dictgen::InsensitiveStr::Ascii("marevlous"), &["marvelous"]),
            (dictgen::InsensitiveStr::Ascii("shnaghai"), &["shanghai"]),
            (
                dictgen::InsensitiveStr::Ascii("insuficient"),
                &["insufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("maskk"), &["mask"]),
            (dictgen::InsensitiveStr::Ascii("cicruits"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("initilializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocurrances"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("brighness"), &["brightness"]),
            (dictgen::InsensitiveStr::Ascii("recyclying"), &["recycling"]),
            (
                dictgen::InsensitiveStr::Ascii("autmobiles"),
                &["automobiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("reccord"), &["record"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnformer"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("arthrite"), &["arthritis"]),
            (
                dictgen::InsensitiveStr::Ascii("swtichboards"),
                &["switchboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curriences"),
                &["currencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("terfform"), &["terraform"]),
            (
                dictgen::InsensitiveStr::Ascii("sceens"),
                &["scenes", "screens", "schemes"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatiner"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("determinin"),
                &["determining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogyinsts"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instutionalized"),
                &["institutionalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenitication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("curiocity"), &["curiosity"]),
            (
                dictgen::InsensitiveStr::Ascii("sudio"),
                &["audio", "sudo", "studio"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenominon"),
                &["phenomenon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avdisoriyes"),
                &["advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("featued"), &["featured"]),
            (dictgen::InsensitiveStr::Ascii("returnn"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("colapsing"), &["collapsing"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuions"),
                &["contributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("deturance"), &["deterrence"]),
            (dictgen::InsensitiveStr::Ascii("publsih"), &["publish"]),
            (
                dictgen::InsensitiveStr::Ascii("supermarktes"),
                &["supermarkets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penninsular"),
                &["peninsular"],
            ),
            (dictgen::InsensitiveStr::Ascii("egenrally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("criticus"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("tsunmai"), &["tsunami"]),
            (dictgen::InsensitiveStr::Ascii("raed"), &["read"]),
            (dictgen::InsensitiveStr::Ascii("awesomeley"), &["awesomely"]),
            (
                dictgen::InsensitiveStr::Ascii("obervation"),
                &["observation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatated"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("raotation"), &["rotation"]),
            (
                dictgen::InsensitiveStr::Ascii("retquiresgpos"),
                &["requiresgpos"],
            ),
            (dictgen::InsensitiveStr::Ascii("docoder"), &["decoder"]),
            (dictgen::InsensitiveStr::Ascii("acivities"), &["activities"]),
            (dictgen::InsensitiveStr::Ascii("througg"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("singulair"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("murdereres"), &["murderers"]),
            (
                dictgen::InsensitiveStr::Ascii("contamints"),
                &["contaminants"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaitor"), &["animator"]),
            (dictgen::InsensitiveStr::Ascii("desturtor"), &["destructor"]),
            (
                dictgen::InsensitiveStr::Ascii("troubeshoot"),
                &["troubleshoot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brimingham"),
                &["birmingham"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjsuted"), &["adjusted"]),
            (dictgen::InsensitiveStr::Ascii("qouoted"), &["quoted"]),
            (dictgen::InsensitiveStr::Ascii("chuncs"), &["chunks"]),
            (
                dictgen::InsensitiveStr::Ascii("portuguease"),
                &["portuguese"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynastry"), &["dynasty"]),
            (
                dictgen::InsensitiveStr::Ascii("perferectly"),
                &["perfectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("communiy"), &["community"]),
            (
                dictgen::InsensitiveStr::Ascii("jurney"),
                &["journey", "gurney"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegitamite"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overpolulation"),
                &["overpopulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nestin"), &["nesting"]),
            (
                dictgen::InsensitiveStr::Ascii("unfamilair"),
                &["unfamiliar"],
            ),
            (dictgen::InsensitiveStr::Ascii("venelope"), &["envelope"]),
            (dictgen::InsensitiveStr::Ascii("workfore"), &["workforce"]),
            (
                dictgen::InsensitiveStr::Ascii("spefifations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interespersed"),
                &["interspersed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcters"),
                &["correctors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charistics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbakc"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("cigurete"), &["cigarette"]),
            (dictgen::InsensitiveStr::Ascii("incudes"), &["includes"]),
            (dictgen::InsensitiveStr::Ascii("dyas"), &["dryas"]),
            (
                dictgen::InsensitiveStr::Ascii("addresssing"),
                &["addressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraging"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("auhtoritatively"),
                &["authoritatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonises"),
                &["synchronises"],
            ),
            (dictgen::InsensitiveStr::Ascii("detectted"), &["detected"]),
            (dictgen::InsensitiveStr::Ascii("vectorr"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("spefiify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("degrated"), &["degrade"]),
            (
                dictgen::InsensitiveStr::Ascii("prontes"),
                &["pronates", "prones"],
            ),
            (dictgen::InsensitiveStr::Ascii("assasin"), &["assassin"]),
            (dictgen::InsensitiveStr::Ascii("sutffing"), &["stuffing"]),
            (dictgen::InsensitiveStr::Ascii("fashoned"), &["fashioned"]),
            (
                dictgen::InsensitiveStr::Ascii("replentishs"),
                &["replenishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("susinkt"), &["succinct"]),
            (dictgen::InsensitiveStr::Ascii("artic"), &["arctic"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsulted"),
                &["encapsulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("contstants"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("accout"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("grahical"), &["graphical"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomrant"),
                &["performant"],
            ),
            (dictgen::InsensitiveStr::Ascii("koreanos"), &["koreans"]),
            (
                dictgen::InsensitiveStr::Ascii("compaytibility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("cihphers"), &["ciphers"]),
            (dictgen::InsensitiveStr::Ascii("nummerical"), &["numerical"]),
            (
                dictgen::InsensitiveStr::Ascii("discusting"),
                &["discussing", "disgusting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexepctedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcuates"), &["calculates"]),
            (dictgen::InsensitiveStr::Ascii("allias"), &["alias"]),
            (
                dictgen::InsensitiveStr::Ascii("reacquirin"),
                &["reacquiring"],
            ),
            (dictgen::InsensitiveStr::Ascii("versoning"), &["versioning"]),
            (dictgen::InsensitiveStr::Ascii("copile"), &["compile"]),
            (
                dictgen::InsensitiveStr::Ascii("corordinate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soundtrakc"),
                &["soundtracks"],
            ),
            (dictgen::InsensitiveStr::Ascii("persewes"), &["pursues"]),
            (
                dictgen::InsensitiveStr::Ascii("reopositories"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("aproval"), &["approval"]),
            (dictgen::InsensitiveStr::Ascii("penguinese"), &["penguins"]),
            (
                dictgen::InsensitiveStr::Ascii("pratictioner"),
                &["practitioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("speechs"), &["speeches"]),
            (dictgen::InsensitiveStr::Ascii("imbrace"), &["embrace"]),
            (dictgen::InsensitiveStr::Ascii("releaving"), &["relieving"]),
            (dictgen::InsensitiveStr::Ascii("furfill"), &["fulfill"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilzes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("yoour"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("lcoal"), &["local", "coal"]),
            (dictgen::InsensitiveStr::Ascii("frpm"), &["from"]),
            (
                dictgen::InsensitiveStr::Ascii("transofrms"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplay"),
                &["delay", "deploy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaocratic"),
                &["bureaucratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justfy"),
                &["justify", "justly"],
            ),
            (dictgen::InsensitiveStr::Ascii("hiden"), &["hidden"]),
            (dictgen::InsensitiveStr::Ascii("reptition"), &["repetition"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratoins"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhane"),
                &["enhance", "ethane"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrationnal"),
                &["irrational"],
            ),
            (dictgen::InsensitiveStr::Ascii("alising"), &["aliasing"]),
            (dictgen::InsensitiveStr::Ascii("roboticus"), &["robotics"]),
            (
                dictgen::InsensitiveStr::Ascii("affordible"),
                &["affordable"],
            ),
            (dictgen::InsensitiveStr::Ascii("debth"), &["depth"]),
            (
                dictgen::InsensitiveStr::Ascii("caclualted"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecction"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("perhpas"), &["perhaps"]),
            (
                dictgen::InsensitiveStr::Ascii("incompasitate"),
                &["incapacitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("inducin"), &["inducing"]),
            (dictgen::InsensitiveStr::Ascii("layed"), &["laid"]),
            (dictgen::InsensitiveStr::Ascii("latley"), &["lately"]),
            (dictgen::InsensitiveStr::Ascii("theard"), &["thread"]),
            (dictgen::InsensitiveStr::Ascii("sythensis"), &["synthesis"]),
            (
                dictgen::InsensitiveStr::Ascii("execuding"),
                &["executing", "excluding"],
            ),
            (dictgen::InsensitiveStr::Ascii("shure"), &["sure"]),
            (
                dictgen::InsensitiveStr::Ascii("guarranty"),
                &["guaranty", "guarantee"],
            ),
            (dictgen::InsensitiveStr::Ascii("connecties"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("ovewrites"), &["overwrites"]),
            (dictgen::InsensitiveStr::Ascii("prasied"), &["praised"]),
            (dictgen::InsensitiveStr::Ascii("perseed"), &["precede"]),
            (dictgen::InsensitiveStr::Ascii("dosument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("observarable"),
                &["observable"],
            ),
            (dictgen::InsensitiveStr::Ascii("presonas"), &["personas"]),
            (
                dictgen::InsensitiveStr::Ascii("engagaments"),
                &["engagements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containign"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiblely"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("evaludate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("managin"), &["managing"]),
            (dictgen::InsensitiveStr::Ascii("correllate"), &["correlate"]),
            (
                dictgen::InsensitiveStr::Ascii("referenncing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberivate"),
                &["abbreviate"],
            ),
            (dictgen::InsensitiveStr::Ascii("referenz"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("transconders"),
                &["transcoders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collecter"),
                &["collector", "collected"],
            ),
            (dictgen::InsensitiveStr::Ascii("subbmitted"), &["submitted"]),
            (
                dictgen::InsensitiveStr::Ascii("derected"),
                &["directed", "detected", "defected", "erected"],
            ),
            (dictgen::InsensitiveStr::Ascii("cataloge"), &["catalogue"]),
            (dictgen::InsensitiveStr::Ascii("finsihed"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("repsonded"), &["responded"]),
            (
                dictgen::InsensitiveStr::Ascii("mistakedly"),
                &["mistakenly"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaquitly"), &["adequately"]),
            (dictgen::InsensitiveStr::Ascii("arrage"), &["arrange"]),
            (
                dictgen::InsensitiveStr::Ascii("pojrecting"),
                &["projecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("isralei"), &["israeli"]),
            (dictgen::InsensitiveStr::Ascii("acedemy"), &["academy"]),
            (
                dictgen::InsensitiveStr::Ascii("similairity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("podmeos"), &["podemos"]),
            (dictgen::InsensitiveStr::Ascii("repuslive"), &["repulsive"]),
            (
                dictgen::InsensitiveStr::Ascii("definstely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missunderstands"),
                &["misunderstands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oreintation"),
                &["orientation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("morrsion"), &["morrison"]),
            (
                dictgen::InsensitiveStr::Ascii("regularisin"),
                &["regularising"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifiecd"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("industriel"),
                &["industrial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinquishing"),
                &["distinguishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nofitications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("nanoseond"), &["nanosecond"]),
            (
                dictgen::InsensitiveStr::Ascii("tansformations"),
                &["transformations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathamatical"),
                &["mathematical"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleandup"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("spripted"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("detroys"), &["destroys"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriagely"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inncrements"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructie"),
                &["instructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("thetering"), &["tethering"]),
            (
                dictgen::InsensitiveStr::Ascii("resteraunt"),
                &["restaurant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responed"),
                &["respond", "responded"],
            ),
            (dictgen::InsensitiveStr::Ascii("bactch"), &["batch"]),
            (dictgen::InsensitiveStr::Ascii("successses"), &["successes"]),
            (
                dictgen::InsensitiveStr::Ascii("subscribbers"),
                &["subscribers"],
            ),
            (dictgen::InsensitiveStr::Ascii("contess"), &["contests"]),
            (
                dictgen::InsensitiveStr::Ascii("dispensory"),
                &["dispensary"],
            ),
            (dictgen::InsensitiveStr::Ascii("missletow"), &["mistletoe"]),
            (
                dictgen::InsensitiveStr::Ascii("quckstarter"),
                &["quickstarter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingrediant"),
                &["ingredient"],
            ),
            (dictgen::InsensitiveStr::Ascii("moniter"), &["monitor"]),
            (dictgen::InsensitiveStr::Ascii("follwow"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("expession"), &["expression"]),
            (
                dictgen::InsensitiveStr::Ascii("prdpagations"),
                &["propagations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mozzaralla"),
                &["mozzarella"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunet"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("personmal"), &["personal"]),
            (dictgen::InsensitiveStr::Ascii("instaslled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("bufers"), &["buffers"]),
            (
                dictgen::InsensitiveStr::Ascii("acheivement"),
                &["achievement"],
            ),
            (dictgen::InsensitiveStr::Ascii("caryy"), &["carry"]),
            (dictgen::InsensitiveStr::Ascii("enusure"), &["ensure"]),
            (dictgen::InsensitiveStr::Ascii("hask"), &["hash"]),
            (dictgen::InsensitiveStr::Ascii("documnents"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("affectes"), &["affects"]),
            (dictgen::InsensitiveStr::Ascii("unmbered"), &["numbered"]),
            (dictgen::InsensitiveStr::Ascii("msesage"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("princessses"),
                &["princesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scracthing"),
                &["scratching"],
            ),
            (dictgen::InsensitiveStr::Ascii("asigment"), &["assignment"]),
            (dictgen::InsensitiveStr::Ascii("revoluion"), &["revolution"]),
            (
                dictgen::InsensitiveStr::Ascii("alnterating"),
                &["alternating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unimiplemented"),
                &["unimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("recieve"), &["receive"]),
            (dictgen::InsensitiveStr::Ascii("wekend"), &["weekend"]),
            (
                dictgen::InsensitiveStr::Ascii("instantialed"),
                &["instantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticae"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongnized"),
                &["recognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("guideance"), &["guidance"]),
            (
                dictgen::InsensitiveStr::Ascii("confirme"),
                &["confirmed", "confirms"],
            ),
            (dictgen::InsensitiveStr::Ascii("secon"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstarted"),
                &["demonstrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("lemosines"), &["limousines"]),
            (dictgen::InsensitiveStr::Ascii("gard"), &["guard"]),
            (dictgen::InsensitiveStr::Ascii("catloag"), &["catalog"]),
            (dictgen::InsensitiveStr::Ascii("problemes"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("flexibiity"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fialing"),
                &["failing", "dialing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstenince"),
                &["abstinence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composistions"),
                &["compositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obeservations"),
                &["observations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administraive"),
                &["administrative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribative"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("bescribed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("dependenices"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scroipttype"),
                &["scripttype"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounites"), &["bounties"]),
            (
                dictgen::InsensitiveStr::Ascii("speicfiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothesus"),
                &["hypotheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throwed"),
                &["threw", "thrown"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparater"),
                &["comparator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("istantiating"),
                &["instantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqquired"), &["acquired"]),
            (
                dictgen::InsensitiveStr::Ascii("interfereing"),
                &["interfering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monitoreada"),
                &["monitored"],
            ),
            (dictgen::InsensitiveStr::Ascii("trooso"), &["trousseau"]),
            (
                dictgen::InsensitiveStr::Ascii("rightfullly"),
                &["rightfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("instande"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("comitte"),
                &["committed", "committee"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condtitionals"),
                &["conditionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffrerence"),
                &["difference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mischevusly"),
                &["mischievously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infallable"),
                &["infallible"],
            ),
            (dictgen::InsensitiveStr::Ascii("admisions"), &["admissions"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutoins"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excecutions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ataching"), &["attaching"]),
            (
                dictgen::InsensitiveStr::Ascii("temproment"),
                &["temperament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transicional"),
                &["transitional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wildebeasts"),
                &["wildebeests"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circunventions"),
                &["circumventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregsiter"),
                &["unregister"],
            ),
            (dictgen::InsensitiveStr::Ascii("backard"), &["backward"]),
            (dictgen::InsensitiveStr::Ascii("acccessd"), &["accessed"]),
            (
                dictgen::InsensitiveStr::Ascii("definitelly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succesfuly"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installatons"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sudeo"), &["sudo"]),
            (
                dictgen::InsensitiveStr::Ascii("wellingston"),
                &["wellington"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeuested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("knowlegdge"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("stetched"),
                &["stretched", "sketched", "stitched"],
            ),
            (dictgen::InsensitiveStr::Ascii("adivsor"), &["advisor"]),
            (
                dictgen::InsensitiveStr::Ascii("coneptions"),
                &["conceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthnically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoildays"), &["holidays"]),
            (dictgen::InsensitiveStr::Ascii("perserved"), &["preserved"]),
            (
                dictgen::InsensitiveStr::Ascii("amealearative"),
                &["ameliorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("numver"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("extimator"), &["estimator"]),
            (
                dictgen::InsensitiveStr::Ascii("consistuent"),
                &["constituent", "consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("swaped"), &["swapped"]),
            (dictgen::InsensitiveStr::Ascii("poointer"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("trasmitter"),
                &["transmitter"],
            ),
            (dictgen::InsensitiveStr::Ascii("deliever"), &["deliver"]),
            (
                dictgen::InsensitiveStr::Ascii("abritrarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interveiwing"),
                &["interviewing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provices"),
                &["provides", "provinces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gererators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccessed"), &["accessed"]),
            (dictgen::InsensitiveStr::Ascii("machettie"), &["machete"]),
            (
                dictgen::InsensitiveStr::Ascii("californains"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucutres"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("senisble"), &["sensible"]),
            (
                dictgen::InsensitiveStr::Ascii("interm"),
                &["interim", "intern"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refererenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("publuc"), &["public"]),
            (dictgen::InsensitiveStr::Ascii("execuetd"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("apoligised"),
                &["apologised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibil"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privatizied"),
                &["privatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypathetically"),
                &["hypothetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplaction"),
                &["deduplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bimillenia"),
                &["bimillennia"],
            ),
            (dictgen::InsensitiveStr::Ascii("memer"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("additoinal"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("deplyo"), &["deploy"]),
            (dictgen::InsensitiveStr::Ascii("iterrupts"), &["interrupts"]),
            (dictgen::InsensitiveStr::Ascii("asssumed"), &["assumed"]),
            (dictgen::InsensitiveStr::Ascii("hardned"), &["hardened"]),
            (dictgen::InsensitiveStr::Ascii("testser"), &["tester"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("difffers"), &["differs"]),
            (dictgen::InsensitiveStr::Ascii("vaulable"), &["valuable"]),
            (
                dictgen::InsensitiveStr::Ascii("mose"),
                &["more", "mouse", "mode"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heared"),
                &["heard", "header"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abnormalty"),
                &["abnormally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectit"),
                &["connecticut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantnciate"),
                &["instantiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simuation"),
                &["simulation", "situation"],
            ),
            (dictgen::InsensitiveStr::Ascii("premire"), &["premier"]),
            (
                dictgen::InsensitiveStr::Ascii("ternimates"),
                &["terminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("closd"), &["closed"]),
            (
                dictgen::InsensitiveStr::Ascii("necessesary"),
                &["necessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appearences"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marshmallons"),
                &["marshmallows"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuallly"), &["visually"]),
            (dictgen::InsensitiveStr::Ascii("teampseak"), &["teamspeak"]),
            (dictgen::InsensitiveStr::Ascii("cusomise"), &["customise"]),
            (
                dictgen::InsensitiveStr::Ascii("shakesperean"),
                &["shakespeare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumiser"),
                &["customiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("ninties"), &["nineties"]),
            (dictgen::InsensitiveStr::Ascii("prgrams"), &["programs"]),
            (dictgen::InsensitiveStr::Ascii("unchangd"), &["unchanged"]),
            (dictgen::InsensitiveStr::Ascii("shufle"), &["shuffle"]),
            (
                dictgen::InsensitiveStr::Ascii("randazyvou"),
                &["rendezvous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbriviates"),
                &["abbreviates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcsription"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statsitical"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilitie"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothises"),
                &["hypotheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("garphs"), &["graphs"]),
            (
                dictgen::InsensitiveStr::Ascii("practicarlo"),
                &["practical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingriedents"),
                &["ingredient"],
            ),
            (dictgen::InsensitiveStr::Ascii("intimite"), &["intimate"]),
            (dictgen::InsensitiveStr::Ascii("addiional"), &["additional"]),
            (dictgen::InsensitiveStr::Ascii("apsergers"), &["aspergers"]),
            (dictgen::InsensitiveStr::Ascii("repeler"), &["repeller"]),
            (dictgen::InsensitiveStr::Ascii("dircetion"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("backgroung"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifiant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("liares"), &["liars"]),
            (dictgen::InsensitiveStr::Ascii("versiones"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("fallbacls"), &["fallbacks"]),
            (dictgen::InsensitiveStr::Ascii("resursion"), &["recursion"]),
            (
                dictgen::InsensitiveStr::Ascii("instabiliy"),
                &["instability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maneouvring"),
                &["manoeuvring"],
            ),
            (dictgen::InsensitiveStr::Ascii("synoymous"), &["synonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("ethnicites"),
                &["ethnicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("miluwakee"), &["milwaukee"]),
            (
                dictgen::InsensitiveStr::Ascii("institutionnal"),
                &["institutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypotehtical"),
                &["hypothetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("vioaltion"), &["violation"]),
            (
                dictgen::InsensitiveStr::Ascii("superfelous"),
                &["superfluous"],
            ),
            (dictgen::InsensitiveStr::Ascii("arcahic"), &["archaic"]),
            (
                dictgen::InsensitiveStr::Ascii("propreties"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("foricbly"), &["forcibly"]),
            (
                dictgen::InsensitiveStr::Ascii("selectiosn"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("teplmates"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("conetivities"),
                &["connectivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millsenconds"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("siz"), &["six", "size"]),
            (
                dictgen::InsensitiveStr::Ascii("adjastment"),
                &["adjustment"],
            ),
            (dictgen::InsensitiveStr::Ascii("deskop"), &["desktop"]),
            (
                dictgen::InsensitiveStr::Ascii("generalizaing"),
                &["generalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acutaly"),
                &["actually", "acutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espescially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("datee"), &["date"]),
            (dictgen::InsensitiveStr::Ascii("confimr"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("packkaging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("uncompreses"),
                &["uncompresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("extrinsinc"), &["extrinsic"]),
            (dictgen::InsensitiveStr::Ascii("exchaning"), &["exchanging"]),
            (dictgen::InsensitiveStr::Ascii("thown"), &["thrown", "town"]),
            (dictgen::InsensitiveStr::Ascii("direccion"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("approximatin"),
                &["approximating", "approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabiliti"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraniod"), &["paranoid"]),
            (dictgen::InsensitiveStr::Ascii("esimator"), &["estimator"]),
            (dictgen::InsensitiveStr::Ascii("protien"), &["protein"]),
            (dictgen::InsensitiveStr::Ascii("specicied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("trignometry"),
                &["trigonometry"],
            ),
            (dictgen::InsensitiveStr::Ascii("languates"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("librarys"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("realignement"),
                &["realignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("thearding"), &["threading"]),
            (
                dictgen::InsensitiveStr::Ascii("addrersssing"),
                &["addressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ansamble"), &["ensemble"]),
            (dictgen::InsensitiveStr::Ascii("typlically"), &["typically"]),
            (dictgen::InsensitiveStr::Ascii("derivin"), &["deriving"]),
            (dictgen::InsensitiveStr::Ascii("timeuots"), &["timeouts"]),
            (dictgen::InsensitiveStr::Ascii("vauled"), &["valued"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoticate"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("casees"), &["cases"]),
            (
                dictgen::InsensitiveStr::Ascii("parlaimentary"),
                &["parliamentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("cetrain"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("relevabt"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("diarhea"), &["diarrhea"]),
            (dictgen::InsensitiveStr::Ascii("androidtvs"), &["androids"]),
            (dictgen::InsensitiveStr::Ascii("definied"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("contigous"), &["contiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("ressourcing"),
                &["resourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("numerial"),
                &["numeral", "numerical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deativated"),
                &["deactivated"],
            ),
            (dictgen::InsensitiveStr::Ascii("archove"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("callsr"), &["calls"]),
            (dictgen::InsensitiveStr::Ascii("ivokers"), &["invokers"]),
            (dictgen::InsensitiveStr::Ascii("expasions"), &["expansions"]),
            (
                dictgen::InsensitiveStr::Ascii("accompagnying"),
                &["accompanying"],
            ),
            (dictgen::InsensitiveStr::Ascii("macchine"), &["machine"]),
            (
                dictgen::InsensitiveStr::Ascii("extacts"),
                &["extracts", "exacts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hydropobic"),
                &["hydrophobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoitator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("doucehbag"), &["douchebag"]),
            (dictgen::InsensitiveStr::Ascii("festivales"), &["festivals"]),
            (dictgen::InsensitiveStr::Ascii("polltry"), &["poultry"]),
            (dictgen::InsensitiveStr::Ascii("somehting"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("obsolted"), &["obsoleted"]),
            (dictgen::InsensitiveStr::Ascii("hinduisum"), &["hinduism"]),
            (
                dictgen::InsensitiveStr::Ascii("refrigerador"),
                &["refrigerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hiearchies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meaningfull"),
                &["meaningful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destructros"),
                &["destructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("centepide"), &["centipede"]),
            (
                dictgen::InsensitiveStr::Ascii("transation"),
                &["transition", "transaction", "translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arguablly"), &["arguably"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotionation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramterer"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("brighening"),
                &["brightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("sistems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("requiremnt"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("thatnks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("standad"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("overklocked"),
                &["overclocked"],
            ),
            (dictgen::InsensitiveStr::Ascii("reverals"), &["reversal"]),
            (
                dictgen::InsensitiveStr::Ascii("frequncies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("infinitiy"), &["infinity"]),
            (
                dictgen::InsensitiveStr::Ascii("monstrasity"),
                &["monstrosity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decraesed"),
                &["decreased", "deceased"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprotions"),
                &["proportions"],
            ),
            (dictgen::InsensitiveStr::Ascii("milennium"), &["millennium"]),
            (dictgen::InsensitiveStr::Ascii("aufter"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("opbjects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnports"),
                &["transports"],
            ),
            (dictgen::InsensitiveStr::Ascii("expecing"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("accoutning"),
                &["accounting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenificator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilplies"),
                &["multiplies"],
            ),
            (dictgen::InsensitiveStr::Ascii("releeases"), &["releases"]),
            (dictgen::InsensitiveStr::Ascii("unifform"), &["uniform"]),
            (
                dictgen::InsensitiveStr::Ascii("spacifications"),
                &["specifications", "pacifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reguardless"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminato"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excrusiating"),
                &["excruciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("explorarea"), &["explorer"]),
            (
                dictgen::InsensitiveStr::Ascii("suffocatin"),
                &["suffocating", "suffocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpise"), &["surprise"]),
            (dictgen::InsensitiveStr::Ascii("abount"), &["about"]),
            (
                dictgen::InsensitiveStr::Ascii("craeting"),
                &["creating", "crating"],
            ),
            (dictgen::InsensitiveStr::Ascii("perist"), &["persist"]),
            (
                dictgen::InsensitiveStr::Ascii("environnemental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("transorm"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("updatess"), &["updates"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactureros"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("separting"), &["separating"]),
            (
                dictgen::InsensitiveStr::Ascii("socilaists"),
                &["socialists"],
            ),
            (dictgen::InsensitiveStr::Ascii("neice"), &["niece", "nice"]),
            (dictgen::InsensitiveStr::Ascii("temporrary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("conceetes"), &["conceits"]),
            (
                dictgen::InsensitiveStr::Ascii("invulnerabile"),
                &["invulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("misouri"), &["missouri"]),
            (dictgen::InsensitiveStr::Ascii("availbable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("interace"),
                &["interacted", "interface", "interlace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomralized"),
                &["normalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("thiniking"), &["thinking"]),
            (dictgen::InsensitiveStr::Ascii("retireus"), &["retires"]),
            (dictgen::InsensitiveStr::Ascii("limitited"), &["limited"]),
            (dictgen::InsensitiveStr::Ascii("riminders"), &["reminders"]),
            (
                dictgen::InsensitiveStr::Ascii("refenenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("someons"), &["someones"]),
            (dictgen::InsensitiveStr::Ascii("mjor"), &["major"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotionable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordinarely"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpute"), &["compute"]),
            (
                dictgen::InsensitiveStr::Ascii("partcipation"),
                &["participation"],
            ),
            (dictgen::InsensitiveStr::Ascii("backards"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("announed"), &["announced"]),
            (dictgen::InsensitiveStr::Ascii("prototyps"), &["prototypes"]),
            (
                dictgen::InsensitiveStr::Ascii("simultanously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("criminaly"), &["criminally"]),
            (
                dictgen::InsensitiveStr::Ascii("meassurably"),
                &["measurably"],
            ),
            (dictgen::InsensitiveStr::Ascii("ilkely"), &["likely"]),
            (
                dictgen::InsensitiveStr::Ascii("connectibity"),
                &["connectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmasists"),
                &["pharmacists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simluations"),
                &["simulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("detecing"), &["detecting"]),
            (dictgen::InsensitiveStr::Ascii("tabeles"), &["tables"]),
            (
                dictgen::InsensitiveStr::Ascii("comversation"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coreesponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreaciating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udnerstood"),
                &["understood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strring"),
                &["string", "stirring", "starring"],
            ),
            (dictgen::InsensitiveStr::Ascii("porotocols"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("hydrolic"), &["hydraulic"]),
            (
                dictgen::InsensitiveStr::Ascii("multisamplimg"),
                &["multisampling"],
            ),
            (dictgen::InsensitiveStr::Ascii("arugements"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("bounadaries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bandwdiths"),
                &["bandwidths"],
            ),
            (dictgen::InsensitiveStr::Ascii("differnet"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("calculaters"),
                &["calculators", "calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcultions"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exicuteable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("temprory"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("consntant"), &["constant"]),
            (
                dictgen::InsensitiveStr::Ascii("championchip"),
                &["championship"],
            ),
            (dictgen::InsensitiveStr::Ascii("resiliant"), &["resilient"]),
            (dictgen::InsensitiveStr::Ascii("beleaved"), &["believed"]),
            (dictgen::InsensitiveStr::Ascii("avaoidable"), &["avoidable"]),
            (
                dictgen::InsensitiveStr::Ascii("undistinghable"),
                &["indistinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("reanamed"), &["renamed"]),
            (
                dictgen::InsensitiveStr::Ascii("contradicton"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomiser"),
                &["randomiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoreticaly"),
                &["theoretically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ballsitic"), &["ballistic"]),
            (
                dictgen::InsensitiveStr::Ascii("propertise"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefifeid"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitutionnal"),
                &["constitutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucccessully"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hiarchies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfromation"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("endopints"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("optomisation"),
                &["optimisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermimentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("higlights"), &["highlights"]),
            (
                dictgen::InsensitiveStr::Ascii("fotografical"),
                &["photographical"],
            ),
            (dictgen::InsensitiveStr::Ascii("utiliizing"), &["utilizing"]),
            (dictgen::InsensitiveStr::Ascii("improoved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("chechsums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("theared"), &["threaded"]),
            (
                dictgen::InsensitiveStr::Ascii("sastisfying"),
                &["satisfying"],
            ),
            (dictgen::InsensitiveStr::Ascii("heriarchy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("marxisim"), &["marxism"]),
            (
                dictgen::InsensitiveStr::Ascii("applcation"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("machinary"), &["machinery"]),
            (dictgen::InsensitiveStr::Ascii("asissts"), &["assists"]),
            (dictgen::InsensitiveStr::Ascii("transated"), &["translated"]),
            (dictgen::InsensitiveStr::Ascii("manadrin"), &["mandarin"]),
            (
                dictgen::InsensitiveStr::Ascii("parsesd"),
                &["parsed", "parses"],
            ),
            (dictgen::InsensitiveStr::Ascii("bassically"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("normaizes"), &["normalizes"]),
            (
                dictgen::InsensitiveStr::Ascii("acumulator"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("souveniers"), &["souvenirs"]),
            (dictgen::InsensitiveStr::Ascii("addapting"), &["adapting"]),
            (dictgen::InsensitiveStr::Ascii("releaase"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("uncatalogued"),
                &["uncataloged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnateeed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrputable"),
                &["interruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsolote"), &["obsolete"]),
            (
                dictgen::InsensitiveStr::Ascii("inicialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("servier"), &["server"]),
            (dictgen::InsensitiveStr::Ascii("calncel"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("kernal"), &["kernel"]),
            (
                dictgen::InsensitiveStr::Ascii("synonimous"),
                &["synonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("regiters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("confogured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("brazeer"), &["brassiere"]),
            (dictgen::InsensitiveStr::Ascii("circels"), &["circles"]),
            (
                dictgen::InsensitiveStr::Ascii("peseudocode"),
                &["pseudocode"],
            ),
            (dictgen::InsensitiveStr::Ascii("irregulier"), &["irregular"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forgeiners"),
                &["foreigners"],
            ),
            (dictgen::InsensitiveStr::Ascii("resulving"), &["resolving"]),
            (
                dictgen::InsensitiveStr::Ascii("applictaion"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("dorceful"), &["forceful"]),
            (
                dictgen::InsensitiveStr::Ascii("concencration"),
                &["concentration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contatenating"),
                &["concatenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitutuents"),
                &["constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("finilises"), &["finalises"]),
            (dictgen::InsensitiveStr::Ascii("explicits"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("thankyooou"), &["thankyou"]),
            (dictgen::InsensitiveStr::Ascii("beacuse"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboorhod"),
                &["neighbourhood", "neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capapbility"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("haorder"), &["hoarder"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonizer"),
                &["synchronizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("idendity"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("microcontrolers"),
                &["microcontrollers"],
            ),
            (dictgen::InsensitiveStr::Ascii("oblisks"), &["obelisks"]),
            (
                dictgen::InsensitiveStr::Ascii("filders"),
                &["folders", "filters", "finders", "fielders", "filers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitll"), &["still"]),
            (dictgen::InsensitiveStr::Ascii("atomobile"), &["automobile"]),
            (dictgen::InsensitiveStr::Ascii("proffessed"), &["professed"]),
            (
                dictgen::InsensitiveStr::Ascii("trogloditism"),
                &["troglodytism"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccurate"), &["accurate"]),
            (
                dictgen::InsensitiveStr::Ascii("vulbearbuility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contnets"),
                &["contents", "continents"],
            ),
            (dictgen::InsensitiveStr::Ascii("addmission"), &["admission"]),
            (
                dictgen::InsensitiveStr::Ascii("neurosciene"),
                &["neuroscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("backgound"), &["background"]),
            (dictgen::InsensitiveStr::Ascii("hjave"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("returens"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("schoalrs"), &["scholars"]),
            (dictgen::InsensitiveStr::Ascii("crusiers"), &["cruisers"]),
            (dictgen::InsensitiveStr::Ascii("ocorrence"), &["occurrence"]),
            (
                dictgen::InsensitiveStr::Ascii("pronunciating"),
                &["pronunciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("scholary"), &["scholarly"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitutin"),
                &["prostitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnagelog"), &["changelog"]),
            (
                dictgen::InsensitiveStr::Ascii("interepolates"),
                &["interpolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subporgram"),
                &["subprogram"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdoelled"), &["modelled"]),
            (
                dictgen::InsensitiveStr::Ascii("insonsistency"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instersction"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("bigtory"), &["bigotry"]),
            (dictgen::InsensitiveStr::Ascii("revovler"), &["revolver"]),
            (
                dictgen::InsensitiveStr::Ascii("corrispondingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("suport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("inutituve"), &["intuitive"]),
            (
                dictgen::InsensitiveStr::Ascii("implementos"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheifs"), &["chiefs"]),
            (dictgen::InsensitiveStr::Ascii("axix"), &["axis"]),
            (
                dictgen::InsensitiveStr::Ascii("uncouncious"),
                &["unconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulitline"), &["multiline"]),
            (
                dictgen::InsensitiveStr::Ascii("alterantely"),
                &["alternately"],
            ),
            (dictgen::InsensitiveStr::Ascii("suspedn"), &["suspend"]),
            (dictgen::InsensitiveStr::Ascii("alha"), &["alpha"]),
            (dictgen::InsensitiveStr::Ascii("sleector"), &["selector"]),
            (dictgen::InsensitiveStr::Ascii("sycronus"), &["synchronous"]),
            (dictgen::InsensitiveStr::Ascii("distroyer"), &["destroyer"]),
            (
                dictgen::InsensitiveStr::Ascii("intreeging"),
                &["intriguing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialitse"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("buiders"), &["builders"]),
            (
                dictgen::InsensitiveStr::Ascii("pomegrenate"),
                &["pomegranate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprssion"), &["expression"]),
            (
                dictgen::InsensitiveStr::Ascii("complciated"),
                &["complicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("feezing"), &["freezing"]),
            (
                dictgen::InsensitiveStr::Ascii("annonators"),
                &["annotators"],
            ),
            (dictgen::InsensitiveStr::Ascii("directyory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("conqure"), &["conquer"]),
            (
                dictgen::InsensitiveStr::Ascii("possitively"),
                &["positively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuntaminated"),
                &["contaminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("allegely"), &["allegedly"]),
            (dictgen::InsensitiveStr::Ascii("callabck"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("provicial"), &["provincial"]),
            (
                dictgen::InsensitiveStr::Ascii("reperesentational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylcle"), &["cycle"]),
            (dictgen::InsensitiveStr::Ascii("exteriour"), &["exterior"]),
            (
                dictgen::InsensitiveStr::Ascii("collaspible"),
                &["collapsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("ogrilla"), &["gorilla"]),
            (dictgen::InsensitiveStr::Ascii("connecto"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("pathalogical"),
                &["pathological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particlarly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("revisionis"), &["revisions"]),
            (dictgen::InsensitiveStr::Ascii("listesning"), &["listening"]),
            (dictgen::InsensitiveStr::Ascii("firmwear"), &["firmware"]),
            (
                dictgen::InsensitiveStr::Ascii("cerifying"),
                &["certifying", "verifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pitchworks"),
                &["pitchforks"],
            ),
            (dictgen::InsensitiveStr::Ascii("consits"), &["consists"]),
            (dictgen::InsensitiveStr::Ascii("indiviaul"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("detur"), &["detour"]),
            (dictgen::InsensitiveStr::Ascii("butiful"), &["beautiful"]),
            (
                dictgen::InsensitiveStr::Ascii("rectengles"),
                &["rectangles"],
            ),
            (dictgen::InsensitiveStr::Ascii("sparated"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("laungage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("changue"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("tranverser"), &["traverser"]),
            (
                dictgen::InsensitiveStr::Ascii("cotent"),
                &["content", "cogent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coommanding"),
                &["commanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completeable"),
                &["completable"],
            ),
            (dictgen::InsensitiveStr::Ascii("databaases"), &["databases"]),
            (dictgen::InsensitiveStr::Ascii("rovide"), &["provide"]),
            (
                dictgen::InsensitiveStr::Ascii("benefitical"),
                &["beneficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("macthup"), &["matchup"]),
            (dictgen::InsensitiveStr::Ascii("depdenent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("requiress"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("unconsciouly"),
                &["unconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quadroopled"),
                &["quadrupled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accessibiliity"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("decording"), &["decoding"]),
            (dictgen::InsensitiveStr::Ascii("artifiacts"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("applicaten"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("enities"), &["entities"]),
            (dictgen::InsensitiveStr::Ascii("agin"), &["again"]),
            (dictgen::InsensitiveStr::Ascii("thnik"), &["think"]),
            (dictgen::InsensitiveStr::Ascii("piplines"), &["pipelines"]),
            (dictgen::InsensitiveStr::Ascii("shelterd"), &["sheltered"]),
            (
                dictgen::InsensitiveStr::Ascii("specificaitons"),
                &["specification", "specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("moounts"), &["mounts"]),
            (
                dictgen::InsensitiveStr::Ascii("parallise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("almsot"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("resulotion"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("luckiy"),
                &["luckily", "lucky"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apocalyspe"),
                &["apocalypse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negosiated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("escate"),
                &["escalate", "escape"],
            ),
            (dictgen::InsensitiveStr::Ascii("foriegner"), &["foreigner"]),
            (dictgen::InsensitiveStr::Ascii("maschine"), &["machine"]),
            (dictgen::InsensitiveStr::Ascii("targed"), &["target"]),
            (dictgen::InsensitiveStr::Ascii("captrure"), &["capture"]),
            (dictgen::InsensitiveStr::Ascii("fthe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("descripors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinclty"),
                &["distinctly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorhitmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("monogmay"), &["monogamy"]),
            (
                dictgen::InsensitiveStr::Ascii("bable"),
                &["babel", "table", "bible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accordintly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("methapor"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("tyes"), &["types", "ties"]),
            (dictgen::InsensitiveStr::Ascii("accronym"), &["acronym"]),
            (
                dictgen::InsensitiveStr::Ascii("foding"),
                &["boding", "coding", "fading", "folding", "fording"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocilators"),
                &["oscillators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifigantly"),
                &["significantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volunteraly"),
                &["voluntarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("saught"), &["sought"]),
            (dictgen::InsensitiveStr::Ascii("cleeshey"), &["cliché"]),
            (
                dictgen::InsensitiveStr::Ascii("depenencis"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trianglulations"),
                &["triangulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoters"), &["supporters"]),
            (
                dictgen::InsensitiveStr::Ascii("permisions"),
                &["permission", "permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verifi"),
                &["verify", "verified"],
            ),
            (dictgen::InsensitiveStr::Ascii("inspries"), &["inspires"]),
            (
                dictgen::InsensitiveStr::Ascii("bitswaping"),
                &["bitswapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("opon"), &["upon"]),
            (
                dictgen::InsensitiveStr::Ascii("ideoligically"),
                &["ideologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasteries"), &["pastries"]),
            (
                dictgen::InsensitiveStr::Ascii("configuraiton"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aqcuaintances"),
                &["acquaintances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempurture"),
                &["temperature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectons"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambadextrously"),
                &["ambidextrously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submisison"),
                &["submissions", "submission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diconnections"),
                &["disconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("delpoying"), &["deploying"]),
            (dictgen::InsensitiveStr::Ascii("legitmate"), &["legitimate"]),
            (
                dictgen::InsensitiveStr::Ascii("satrically"),
                &["satirically"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekcer"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("lybia"), &["libya"]),
            (dictgen::InsensitiveStr::Ascii("ansesters"), &["ancestors"]),
            (dictgen::InsensitiveStr::Ascii("descirbes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("transformis"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometihng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("curentfilter"),
                &["currentfilter"],
            ),
            (dictgen::InsensitiveStr::Ascii("fluroide"), &["fluoride"]),
            (dictgen::InsensitiveStr::Ascii("oberved"), &["observed"]),
            (dictgen::InsensitiveStr::Ascii("repects"), &["respects"]),
            (dictgen::InsensitiveStr::Ascii("coexhist"), &["coexist"]),
            (dictgen::InsensitiveStr::Ascii("poperties"), &["properties"]),
            (
                dictgen::InsensitiveStr::Ascii("tolelerance"),
                &["tolerance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstructions"),
                &["instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecissary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("nutreints"), &["nutrients"]),
            (dictgen::InsensitiveStr::Ascii("narcissit"), &["narcissist"]),
            (
                dictgen::InsensitiveStr::Ascii("communitcation"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimzations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("emiited"), &["emitted"]),
            (
                dictgen::InsensitiveStr::Ascii("abrreviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triagulators"),
                &["triangulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("returing"), &["returning"]),
            (dictgen::InsensitiveStr::Ascii("attandee"), &["attendee"]),
            (
                dictgen::InsensitiveStr::Ascii("visisbility"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("execusion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("instincitvely"),
                &["instinctively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartized"),
                &["standardized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequantial"),
                &["sequential"],
            ),
            (dictgen::InsensitiveStr::Ascii("funrels"), &["funerals"]),
            (dictgen::InsensitiveStr::Ascii("liberaties"), &["liberate"]),
            (dictgen::InsensitiveStr::Ascii("chcker"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("propigate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("desroy"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("seahawkers"), &["seahawks"]),
            (
                dictgen::InsensitiveStr::Ascii("disscusion"),
                &["discussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("compexity"), &["complexity"]),
            (dictgen::InsensitiveStr::Ascii("rasbperry"), &["raspberry"]),
            (dictgen::InsensitiveStr::Ascii("menas"), &["means"]),
            (dictgen::InsensitiveStr::Ascii("valied"), &["valid"]),
            (
                dictgen::InsensitiveStr::Ascii("irreversibelt"),
                &["irreversible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traverese"),
                &["traverse", "traverses"],
            ),
            (dictgen::InsensitiveStr::Ascii("instanses"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("idicators"), &["indicators"]),
            (dictgen::InsensitiveStr::Ascii("allwances"), &["allowances"]),
            (
                dictgen::InsensitiveStr::Ascii("prequesite"),
                &["prerequisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coversation"),
                &["conversation"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumse"), &["assumes"]),
            (dictgen::InsensitiveStr::Ascii("conteins"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("successsor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("jugglenaut"),
                &["juggernaut"],
            ),
            (dictgen::InsensitiveStr::Ascii("arguents"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("downaloads"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("battlaion"), &["battalion"]),
            (dictgen::InsensitiveStr::Ascii("defnitely"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubuting"),
                &["distributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregisterable"),
                &["unregistrable"],
            ),
            (dictgen::InsensitiveStr::Ascii("medacine"), &["medicine"]),
            (
                dictgen::InsensitiveStr::Ascii("intellectualis"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("measureably"),
                &["measurably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglegance"),
                &["negligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substittue"),
                &["substitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("dettaching"), &["detaching"]),
            (dictgen::InsensitiveStr::Ascii("oberves"), &["observes"]),
            (
                dictgen::InsensitiveStr::Ascii("coincidencal"),
                &["coincidental"],
            ),
            (dictgen::InsensitiveStr::Ascii("procesess"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("guaranees"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("capitilising"),
                &["capitalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interceptin"),
                &["interception"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apostrophie"),
                &["apostrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("sligth"), &["slight"]),
            (
                dictgen::InsensitiveStr::Ascii("expexnasion"),
                &["expansion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comunities"),
                &["communities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("produse"),
                &["produces", "produce"],
            ),
            (dictgen::InsensitiveStr::Ascii("recustion"), &["recursion"]),
            (dictgen::InsensitiveStr::Ascii("euphoira"), &["euphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("guarenties"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destablized"),
                &["destabilized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresonded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("gargoil"), &["gargoyle"]),
            (dictgen::InsensitiveStr::Ascii("preffer"), &["prefer"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapolant"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganbia"), &["gambia"]),
            (
                dictgen::InsensitiveStr::Ascii("conttributions"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresping"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntetic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("ballancer"), &["balancer"]),
            (
                dictgen::InsensitiveStr::Ascii("occourrences"),
                &["occurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consturctors"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeding"), &["succeeding"]),
            (dictgen::InsensitiveStr::Ascii("exicte"), &["excite"]),
            (
                dictgen::InsensitiveStr::Ascii("convetions"),
                &["conventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("clishays"), &["clichés"]),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarating"),
                &["ameliorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("plathome"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("latiude"), &["latitude"]),
            (
                dictgen::InsensitiveStr::Ascii("determineing"),
                &["determining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressivisme"),
                &["progressives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("drunkness"),
                &["drunkenness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mainfesting"),
                &["manifesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("databae"), &["database"]),
            (dictgen::InsensitiveStr::Ascii("regisry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("agumented"), &["augmented"]),
            (dictgen::InsensitiveStr::Ascii("helluvva"), &["helluva"]),
            (
                dictgen::InsensitiveStr::Ascii("ncessarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("unplesant"), &["unpleasant"]),
            (
                dictgen::InsensitiveStr::Ascii("sprecialized"),
                &["specialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("czechoslavakia"),
                &["czechoslovakia"],
            ),
            (dictgen::InsensitiveStr::Ascii("shapsnots"), &["snapshots"]),
            (dictgen::InsensitiveStr::Ascii("warcrat"), &["warcraft"]),
            (
                dictgen::InsensitiveStr::Ascii("ressurects"),
                &["resurrects"],
            ),
            (dictgen::InsensitiveStr::Ascii("accodr"), &["accord"]),
            (
                dictgen::InsensitiveStr::Ascii("revisisions"),
                &["revisions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangaged"), &["managed"]),
            (
                dictgen::InsensitiveStr::Ascii("troubleshoting"),
                &["troubleshooting"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupliace"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("sintact"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("diplaying"), &["displaying"]),
            (
                dictgen::InsensitiveStr::Ascii("modificacion"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferncence"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("benchs"), &["benches"]),
            (
                dictgen::InsensitiveStr::Ascii("phemonenon"),
                &["phenomenon"],
            ),
            (dictgen::InsensitiveStr::Ascii("encocder"), &["encoder"]),
            (
                dictgen::InsensitiveStr::Ascii("unintentinal"),
                &["unintentional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generatrs"),
                &["generators", "generates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perposes"),
                &["purposes", "proposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificiation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstator"),
                &["demonstrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("aloowing"), &["allowing"]),
            (dictgen::InsensitiveStr::Ascii("charaters"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("applicalbe"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("recyles"), &["recycles"]),
            (dictgen::InsensitiveStr::Ascii("notifyed"), &["notified"]),
            (dictgen::InsensitiveStr::Ascii("celllular"), &["cellular"]),
            (
                dictgen::InsensitiveStr::Ascii("determinstically"),
                &["deterministically"],
            ),
            (dictgen::InsensitiveStr::Ascii("superisor"), &["superiors"]),
            (dictgen::InsensitiveStr::Ascii("ritable"), &["writable"]),
            (dictgen::InsensitiveStr::Ascii("woulndt"), &["wouldnt"]),
            (
                dictgen::InsensitiveStr::Ascii("adminitrators"),
                &["administrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accortance"),
                &["accordance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autodection"),
                &["autodetection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincidentaly"),
                &["coincidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("pompted"), &["prompted"]),
            (
                dictgen::InsensitiveStr::Ascii("propector"),
                &["projector", "prospector", "protector"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contentino"),
                &["contention"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadamic"), &["academic"]),
            (
                dictgen::InsensitiveStr::Ascii("austrailian"),
                &["australian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infoemational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("instralled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("zimbabe"), &["zimbabwe"]),
            (
                dictgen::InsensitiveStr::Ascii("administrater"),
                &[
                    "administrator",
                    "administrated",
                    "administrates",
                    "administrate",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("sessison"), &["session"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotion"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("obiedence"), &["obedience"]),
            (dictgen::InsensitiveStr::Ascii("shorly"), &["shortly"]),
            (dictgen::InsensitiveStr::Ascii("santize"), &["sanitize"]),
            (dictgen::InsensitiveStr::Ascii("paradse"), &["parades"]),
            (dictgen::InsensitiveStr::Ascii("saterday"), &["saturday"]),
            (dictgen::InsensitiveStr::Ascii("fidality"), &["fidelity"]),
            (dictgen::InsensitiveStr::Ascii("gracefull"), &["graceful"]),
            (
                dictgen::InsensitiveStr::Ascii("accomodated"),
                &["accommodated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moutnpoints"),
                &["mountpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideologiset"),
                &["ideologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resuctions"),
                &["reductions"],
            ),
            (dictgen::InsensitiveStr::Ascii("achevied"), &["achieved"]),
            (dictgen::InsensitiveStr::Ascii("iterat"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("adhear"), &["adhere"]),
            (dictgen::InsensitiveStr::Ascii("requieum"), &["requiem"]),
            (
                dictgen::InsensitiveStr::Ascii("revrievers"),
                &["retrievers"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaecklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("critisiced"),
                &["criticized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meshe"),
                &["mesh", "meshed", "meshes"],
            ),
            (dictgen::InsensitiveStr::Ascii("adequet"), &["adequate"]),
            (dictgen::InsensitiveStr::Ascii("mysefl"), &["myself"]),
            (dictgen::InsensitiveStr::Ascii("pivotting"), &["pivoting"]),
            (dictgen::InsensitiveStr::Ascii("listapck"), &["listpack"]),
            (dictgen::InsensitiveStr::Ascii("depserate"), &["desperate"]),
            (
                dictgen::InsensitiveStr::Ascii("maintiainers"),
                &["maintainers"],
            ),
            (dictgen::InsensitiveStr::Ascii("zink"), &["zinc"]),
            (dictgen::InsensitiveStr::Ascii("arragnes"), &["arranges"]),
            (
                dictgen::InsensitiveStr::Ascii("paricipation"),
                &["participation"],
            ),
            (dictgen::InsensitiveStr::Ascii("manifiests"), &["manifests"]),
            (
                dictgen::InsensitiveStr::Ascii("insistenze"),
                &["insistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("fivety"), &["fifty"]),
            (dictgen::InsensitiveStr::Ascii("diaabled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("incovience"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("yymmetrical"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disctinction"),
                &["distinction"],
            ),
            (dictgen::InsensitiveStr::Ascii("curatin"), &["curtain"]),
            (
                dictgen::InsensitiveStr::Ascii("demosntrate"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desginations"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moutnpoint"),
                &["mountpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("alowed"), &["allowed"]),
            (
                dictgen::InsensitiveStr::Ascii("disconencted"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("zimmap"), &["zipmap"]),
            (
                dictgen::InsensitiveStr::Ascii("transmisson"),
                &["transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("toglles"), &["toggles"]),
            (
                dictgen::InsensitiveStr::Ascii("pigen"),
                &["pigeon", "pigpen"],
            ),
            (dictgen::InsensitiveStr::Ascii("balona"), &["bologna"]),
            (dictgen::InsensitiveStr::Ascii("soltuion"), &["solution"]),
            (dictgen::InsensitiveStr::Ascii("trasposed"), &["transposed"]),
            (
                dictgen::InsensitiveStr::Ascii("corresspondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apportunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("promps"), &["prompts"]),
            (dictgen::InsensitiveStr::Ascii("creationis"), &["creations"]),
            (dictgen::InsensitiveStr::Ascii("compils"), &["compiles"]),
            (
                dictgen::InsensitiveStr::Ascii("exceuctable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("proocess"), &["process"]),
            (dictgen::InsensitiveStr::Ascii("putposed"), &["purposed"]),
            (dictgen::InsensitiveStr::Ascii("ommiting"), &["omitting"]),
            (
                dictgen::InsensitiveStr::Ascii("prioritses"),
                &["prioritises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccessbile"),
                &["inaccessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("beforr"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("callls"), &["calls"]),
            (dictgen::InsensitiveStr::Ascii("folliwind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("suspicios"), &["suspicious"]),
            (
                dictgen::InsensitiveStr::Ascii("persectued"),
                &["persecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprirate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("goole"), &["google"]),
            (
                dictgen::InsensitiveStr::Ascii("interseciton"),
                &["intersections"],
            ),
            (dictgen::InsensitiveStr::Ascii("modyfied"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("arraays"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("ordianry"), &["ordinary"]),
            (dictgen::InsensitiveStr::Ascii("postive"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("theves"), &["thieves"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpexted"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("siezure"), &["seizure"]),
            (dictgen::InsensitiveStr::Ascii("eploits"), &["exploits"]),
            (dictgen::InsensitiveStr::Ascii("silibus"), &["syllabus"]),
            (
                dictgen::InsensitiveStr::Ascii("cleand"),
                &["cleaned", "cleans", "clean"],
            ),
            (dictgen::InsensitiveStr::Ascii("determnine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("parallised"),
                &["parallelised"],
            ),
            (dictgen::InsensitiveStr::Ascii("vai"), &["via", "vie"]),
            (
                dictgen::InsensitiveStr::Ascii("grasshoper"),
                &["grasshopper"],
            ),
            (dictgen::InsensitiveStr::Ascii("modells"), &["models"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geogrpahical"),
                &["geographical"],
            ),
            (dictgen::InsensitiveStr::Ascii("propeht"), &["prophet"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilysed"),
                &["cannibalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contorlers"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homophoboes"),
                &["homophobe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shortcomin"),
                &["shortcoming"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbraces"), &["embraces"]),
            (
                dictgen::InsensitiveStr::Ascii("methodolgy"),
                &["methodology"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapmshire"), &["hampshire"]),
            (dictgen::InsensitiveStr::Ascii("adaquate"), &["adequate"]),
            (dictgen::InsensitiveStr::Ascii("nighther"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("bathrooom"), &["bathroom"]),
            (
                dictgen::InsensitiveStr::Ascii("specializied"),
                &["specialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("refernnce"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("visuaise"), &["visualise"]),
            (
                dictgen::InsensitiveStr::Ascii("foregorund"),
                &["foreground"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lannasters"),
                &["lannisters"],
            ),
            (dictgen::InsensitiveStr::Ascii("objet"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("gometries"), &["geometries"]),
            (
                dictgen::InsensitiveStr::Ascii("cylyndrically"),
                &["cylindrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemps"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("showboarding"),
                &["snowboarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presistently"),
                &["persistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("achored"), &["anchored"]),
            (dictgen::InsensitiveStr::Ascii("panasoic"), &["panasonic"]),
            (dictgen::InsensitiveStr::Ascii("classicus"), &["classics"]),
            (dictgen::InsensitiveStr::Ascii("executble"), &["executable"]),
            (
                dictgen::InsensitiveStr::Ascii("deustchland"),
                &["deutschland"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbhors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("doagonals"), &["diagonals"]),
            (dictgen::InsensitiveStr::Ascii("cmment"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("furnature"), &["furniture"]),
            (
                dictgen::InsensitiveStr::Ascii("multipying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("retailes"), &["retailers"]),
            (dictgen::InsensitiveStr::Ascii("monatge"), &["montage"]),
            (
                dictgen::InsensitiveStr::Ascii("delimititation"),
                &["delimitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosiphers"),
                &["philosophers"],
            ),
            (dictgen::InsensitiveStr::Ascii("elese"), &["else"]),
            (
                dictgen::InsensitiveStr::Ascii("telepotation"),
                &["teleportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassembler"),
                &["disassembler"],
            ),
            (dictgen::InsensitiveStr::Ascii("wawes"), &["waves"]),
            (
                dictgen::InsensitiveStr::Ascii("indenpendence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("scientic"), &["scientific"]),
            (dictgen::InsensitiveStr::Ascii("agregates"), &["aggregates"]),
            (dictgen::InsensitiveStr::Ascii("cirilic"), &["cyrillic"]),
            (dictgen::InsensitiveStr::Ascii("hadles"), &["handles"]),
            (dictgen::InsensitiveStr::Ascii("attribtute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("tehy"), &["they"]),
            (dictgen::InsensitiveStr::Ascii("raplaced"), &["replaced"]),
            (
                dictgen::InsensitiveStr::Ascii("incatation"),
                &["incantation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fromat"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("mchanic"), &["mechanic"]),
            (
                dictgen::InsensitiveStr::Ascii("chronicels"),
                &["chronicles"],
            ),
            (dictgen::InsensitiveStr::Ascii("deocdes"), &["decodes"]),
            (
                dictgen::InsensitiveStr::Ascii("marices"),
                &["matrices", "marines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susequently"),
                &["subsequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyperhtreaded"),
                &["hyperthreaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recrusively"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resevations"),
                &["reservations"],
            ),
            (dictgen::InsensitiveStr::Ascii("distarct"), &["distract"]),
            (
                dictgen::InsensitiveStr::Ascii("reproducbility"),
                &["reproducibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requirying"),
                &["requerying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchnously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("confiugure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("tihng"), &["thing", "ting"]),
            (
                dictgen::InsensitiveStr::Ascii("assinments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("buipd"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("practisioner"),
                &["practitioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaued"), &["valued"]),
            (dictgen::InsensitiveStr::Ascii("olympis"), &["olympics"]),
            (
                dictgen::InsensitiveStr::Ascii("coppyrighted"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggegates"), &["aggregates"]),
            (
                dictgen::InsensitiveStr::Ascii("correctably"),
                &["correctable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duoblequote"),
                &["doublequote"],
            ),
            (dictgen::InsensitiveStr::Ascii("invirant"), &["invariant"]),
            (
                dictgen::InsensitiveStr::Ascii("contsruction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("santity"),
                &["sanctity", "sanity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequenices"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("expempted"), &["exempted"]),
            (
                dictgen::InsensitiveStr::Ascii("rountripped"),
                &["roundtripped"],
            ),
            (dictgen::InsensitiveStr::Ascii("properety"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminista"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimtenation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprciating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mneonics"), &["mnemonics"]),
            (
                dictgen::InsensitiveStr::Ascii("journalistisch"),
                &["journalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquisito"),
                &["inquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrieces"), &["retrieves"]),
            (
                dictgen::InsensitiveStr::Ascii("contitional"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("plasticos"), &["plastics"]),
            (dictgen::InsensitiveStr::Ascii("limtiers"), &["limiters"]),
            (dictgen::InsensitiveStr::Ascii("warniing"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("loweracing"),
                &["lowercasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsolate"), &["obsolete"]),
            (dictgen::InsensitiveStr::Ascii("pubilsh"), &["publish"]),
            (dictgen::InsensitiveStr::Ascii("scandas"), &["scandals"]),
            (
                dictgen::InsensitiveStr::Ascii("releection"),
                &["reelection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refenencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("soceity"), &["society"]),
            (dictgen::InsensitiveStr::Ascii("habsbourg"), &["habsburg"]),
            (dictgen::InsensitiveStr::Ascii("dementa"), &["dementia"]),
            (dictgen::InsensitiveStr::Ascii("skirmiches"), &["skirmish"]),
            (dictgen::InsensitiveStr::Ascii("secrectly"), &["secretly"]),
            (
                dictgen::InsensitiveStr::Ascii("remebering"),
                &["remembering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncontitutional"),
                &["unconstitutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specfication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraiving"), &["arriving"]),
            (dictgen::InsensitiveStr::Ascii("seond"), &["second", "send"]),
            (dictgen::InsensitiveStr::Ascii("menssaging"), &["messaging"]),
            (dictgen::InsensitiveStr::Ascii("rechaged"), &["recharged"]),
            (dictgen::InsensitiveStr::Ascii("numbder"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("intereseted"),
                &["interested"],
            ),
            (dictgen::InsensitiveStr::Ascii("oulined"), &["outlined"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliases"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("brunete"), &["brunette"]),
            (
                dictgen::InsensitiveStr::Ascii("symmetricaly"),
                &["symmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("worbooks"), &["workbooks"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffciency"),
                &["inefficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("setgit"), &["setgid"]),
            (dictgen::InsensitiveStr::Ascii("reundant"), &["redundant"]),
            (dictgen::InsensitiveStr::Ascii("virtalenv"), &["virtualenv"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitonality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("encompas"), &["encompass"]),
            (
                dictgen::InsensitiveStr::Ascii("disctionaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("weeves"), &["weaves"]),
            (dictgen::InsensitiveStr::Ascii("packacged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("knoing"), &["knowing"]),
            (
                dictgen::InsensitiveStr::Ascii("embarrasments"),
                &["embarrassment"],
            ),
            (dictgen::InsensitiveStr::Ascii("pozitive"), &["positive"]),
            (
                dictgen::InsensitiveStr::Ascii("confirmacion"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthagonal"),
                &["orthogonal"],
            ),
            (dictgen::InsensitiveStr::Ascii("striing"), &["string"]),
            (
                dictgen::InsensitiveStr::Ascii("sheped"),
                &["shaped", "shepherd"],
            ),
            (dictgen::InsensitiveStr::Ascii("regualte"), &["regulate"]),
            (dictgen::InsensitiveStr::Ascii("loclaized"), &["localized"]),
            (dictgen::InsensitiveStr::Ascii("constext"), &["context"]),
            (dictgen::InsensitiveStr::Ascii("spirously"), &["spuriously"]),
            (dictgen::InsensitiveStr::Ascii("symetry"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("grandizer"),
                &["aggrandizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mcroscopic"),
                &["microscopic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allowe"),
                &["allowed", "allow", "allows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmenets"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presured"),
                &["pressured", "presumed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrantees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahacters"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("administor"),
                &["administer", "administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presures"),
                &["pressures", "presumes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("imparital"), &["impartial"]),
            (
                dictgen::InsensitiveStr::Ascii("proponentes"),
                &["proponents"],
            ),
            (dictgen::InsensitiveStr::Ascii("excetutive"), &["executive"]),
            (dictgen::InsensitiveStr::Ascii("seahaws"), &["seahawks"]),
            (dictgen::InsensitiveStr::Ascii("erraneous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("stablizing"),
                &["stabilizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cymptumatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("upstrem"), &["upstream"]),
            (
                dictgen::InsensitiveStr::Ascii("authoratitive"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detemination"),
                &["determination"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebulding"), &["rebuilding"]),
            (
                dictgen::InsensitiveStr::Ascii("sympethize"),
                &["sympathize"],
            ),
            (dictgen::InsensitiveStr::Ascii("modul"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("supposdely"),
                &["supposedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("fetaure"), &["feature"]),
            (dictgen::InsensitiveStr::Ascii("happpened"), &["happened"]),
            (dictgen::InsensitiveStr::Ascii("einfahc"), &["einfach"]),
            (dictgen::InsensitiveStr::Ascii("devolvendo"), &["devolved"]),
            (
                dictgen::InsensitiveStr::Ascii("recommenations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("stteting"), &["setting"]),
            (
                dictgen::InsensitiveStr::Ascii("parasitisme"),
                &["parasites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitraton"),
                &["arbitration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispossable"),
                &["disposable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autherization"),
                &["authorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("crayones"), &["crayons"]),
            (dictgen::InsensitiveStr::Ascii("additiona"), &["additional"]),
            (dictgen::InsensitiveStr::Ascii("valudated"), &["validated"]),
            (
                dictgen::InsensitiveStr::Ascii("modifations"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("provicde"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("contries"), &["countries"]),
            (
                dictgen::InsensitiveStr::Ascii("attemtping"),
                &["attempting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separions"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lightheated"),
                &["lighthearted"],
            ),
            (dictgen::InsensitiveStr::Ascii("knowldege"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("manfuacturers"),
                &["manufactures"],
            ),
            (dictgen::InsensitiveStr::Ascii("nect"), &["next"]),
            (
                dictgen::InsensitiveStr::Ascii("desriptively"),
                &["descriptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("inovker"), &["invoker"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfitier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("oriant"), &["orient"]),
            (
                dictgen::InsensitiveStr::Ascii("moistruizer"),
                &["moisturizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotained"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("beginnins"), &["beginnings"]),
            (
                dictgen::InsensitiveStr::Ascii("dperecation"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("droppped"), &["dropped"]),
            (dictgen::InsensitiveStr::Ascii("ailmony"), &["alimony"]),
            (
                dictgen::InsensitiveStr::Ascii("nieghboring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frankenstined"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("spartants"), &["spartans"]),
            (dictgen::InsensitiveStr::Ascii("replacingx"), &["replacing"]),
            (dictgen::InsensitiveStr::Ascii("lisence"), &["licence"]),
            (dictgen::InsensitiveStr::Ascii("reporeted"), &["reported"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomeers"),
                &["performers"],
            ),
            (dictgen::InsensitiveStr::Ascii("libstc"), &["libstdc"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangnig"),
                &["rearranging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisioner"),
                &["commissioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("louisvillle"),
                &["louisville"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prgrammers"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("reounded"), &["rounded"]),
            (dictgen::InsensitiveStr::Ascii("cofriming"), &["confirming"]),
            (
                dictgen::InsensitiveStr::Ascii("feminimity"),
                &["femininity"],
            ),
            (dictgen::InsensitiveStr::Ascii("tommrow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("acquaintainces"),
                &["acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("darw"), &["draw"]),
            (
                dictgen::InsensitiveStr::Ascii("positionned"),
                &["positioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoeer"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("alloco"), &["alloc"]),
            (dictgen::InsensitiveStr::Ascii("amouns"), &["amounts"]),
            (dictgen::InsensitiveStr::Ascii("supeena"), &["subpoena"]),
            (
                dictgen::InsensitiveStr::Ascii("anmes"),
                &["names", "animes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimativo"),
                &["estimation"],
            ),
            (dictgen::InsensitiveStr::Ascii("excludles"), &["excludes"]),
            (
                dictgen::InsensitiveStr::Ascii("asthetically"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("indroduce"), &["introduce"]),
            (
                dictgen::InsensitiveStr::Ascii("cemateries"),
                &["cemeteries"],
            ),
            (dictgen::InsensitiveStr::Ascii("associcate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("becomme"), &["become"]),
            (
                dictgen::InsensitiveStr::Ascii("infoermation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purpusefully"),
                &["purposefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execubales"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("contein"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("regeresses"), &["regresses"]),
            (dictgen::InsensitiveStr::Ascii("offficials"), &["officials"]),
            (
                dictgen::InsensitiveStr::Ascii("gerated"),
                &["generated", "gyrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoruim"), &["thorium"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("catpures"), &["captures"]),
            (dictgen::InsensitiveStr::Ascii("chernyobl"), &["chernobyl"]),
            (
                dictgen::InsensitiveStr::Ascii("optisimations"),
                &["optimisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("porcesses"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("handlng"), &["handling"]),
            (dictgen::InsensitiveStr::Ascii("effct"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("correspended"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("bithced"), &["bitched"]),
            (dictgen::InsensitiveStr::Ascii("addrss"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("eferenced"), &["referenced"]),
            (dictgen::InsensitiveStr::Ascii("instnt"), &["instant"]),
            (dictgen::InsensitiveStr::Ascii("migrianes"), &["migraines"]),
            (dictgen::InsensitiveStr::Ascii("furhtest"), &["furthest"]),
            (dictgen::InsensitiveStr::Ascii("insterting"), &["inserting"]),
            (dictgen::InsensitiveStr::Ascii("statitics"), &["statistics"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differente"),
                &["difference", "different"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliegh"), &["sleigh"]),
            (
                dictgen::InsensitiveStr::Ascii("requiremenet"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("govermet"), &["government"]),
            (
                dictgen::InsensitiveStr::Ascii("criticall"),
                &["critical", "critically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orgainizing"),
                &["organizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("idel"), &["idle"]),
            (
                dictgen::InsensitiveStr::Ascii("addtitional"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleeshays"), &["clichés"]),
            (
                dictgen::InsensitiveStr::Ascii("contributeurs"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnosoles"), &["consoles"]),
            (
                dictgen::InsensitiveStr::Ascii("indikation"),
                &["indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adtodetect"),
                &["autodetect"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepet"), &["accept"]),
            (dictgen::InsensitiveStr::Ascii("varioush"), &["various"]),
            (
                dictgen::InsensitiveStr::Ascii("volonteering"),
                &["volunteering"],
            ),
            (dictgen::InsensitiveStr::Ascii("twoards"), &["towards"]),
            (
                dictgen::InsensitiveStr::Ascii("activiting"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("deveopers"), &["developers"]),
            (
                dictgen::InsensitiveStr::Ascii("dircetives"),
                &["directives"],
            ),
            (dictgen::InsensitiveStr::Ascii("servoced"), &["serviced"]),
            (
                dictgen::InsensitiveStr::Ascii("undertands"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illumniati"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparied"), &["compared"]),
            (dictgen::InsensitiveStr::Ascii("onwed"), &["owned"]),
            (dictgen::InsensitiveStr::Ascii("averagine"), &["averaging"]),
            (
                dictgen::InsensitiveStr::Ascii("foundland"),
                &["newfoundland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalizin"),
                &["nationalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abtractness"),
                &["abstractness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatinations"),
                &["concatenations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superstitiuos"),
                &["superstitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("eviciton"), &["eviction"]),
            (dictgen::InsensitiveStr::Ascii("enque"), &["enqueue"]),
            (
                dictgen::InsensitiveStr::Ascii("privilidged"),
                &["privileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contravercy"),
                &["controversy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexpresssions"),
                &["subexpressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("framewoek"), &["framework"]),
            (dictgen::InsensitiveStr::Ascii("whilelist"), &["whitelist"]),
            (
                dictgen::InsensitiveStr::Ascii("perfromers"),
                &["performers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conparable"),
                &["comparable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscommunciation"),
                &["miscommunication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("torrentbig"),
                &["torrenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("eventualy"), &["eventually"]),
            (dictgen::InsensitiveStr::Ascii("tacticus"), &["tactics"]),
            (dictgen::InsensitiveStr::Ascii("silbings"), &["siblings"]),
            (dictgen::InsensitiveStr::Ascii("sometiles"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubtion"),
                &["distribution", "disruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("persuadin"), &["persuading"]),
            (
                dictgen::InsensitiveStr::Ascii("exercicing"),
                &["exercising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approstraphe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfortbly"),
                &["comfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpress"), &["compress"]),
            (dictgen::InsensitiveStr::Ascii("oaut"), &["oauth"]),
            (
                dictgen::InsensitiveStr::Ascii("deveolpers"),
                &["developers"],
            ),
            (dictgen::InsensitiveStr::Ascii("twelth"), &["twelfth"]),
            (
                dictgen::InsensitiveStr::Ascii("entrie"),
                &["entry", "entries"],
            ),
            (dictgen::InsensitiveStr::Ascii("aviable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("sentinet"), &["sentient"]),
            (
                dictgen::InsensitiveStr::Ascii("fortunatelly"),
                &["fortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("executino"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("depcited"), &["depicted"]),
            (dictgen::InsensitiveStr::Ascii("affintiy"), &["affinity"]),
            (dictgen::InsensitiveStr::Ascii("intgral"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("enoguh"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("superstision"),
                &["superstition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumerisim"),
                &["consumerism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribuor"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("cople"), &["couple"]),
            (
                dictgen::InsensitiveStr::Ascii("possitioning"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapater"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("rewite"), &["rewrite"]),
            (dictgen::InsensitiveStr::Ascii("connfig"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("requir"), &["require"]),
            (
                dictgen::InsensitiveStr::Ascii("clsasified"),
                &["classified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("radioactieve"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("righ"), &["right"]),
            (dictgen::InsensitiveStr::Ascii("ballisitc"), &["ballistic"]),
            (dictgen::InsensitiveStr::Ascii("organizin"), &["organizing"]),
            (dictgen::InsensitiveStr::Ascii("wmpty"), &["empty"]),
            (dictgen::InsensitiveStr::Ascii("caridnals"), &["cardinals"]),
            (dictgen::InsensitiveStr::Ascii("recrooted"), &["recruited"]),
            (dictgen::InsensitiveStr::Ascii("guassian"), &["gaussian"]),
            (dictgen::InsensitiveStr::Ascii("inself"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("beginnnings"),
                &["beginnings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correalation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attened"),
                &["attended", "attend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wonderfullly"),
                &["wonderfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tiemstamps"),
                &["timestamps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribtions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("repalacing"), &["replacing"]),
            (
                dictgen::InsensitiveStr::Ascii("preppent"),
                &["prepend", "preprent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustrasted"),
                &["frustrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstraion"),
                &["demonstration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seletion"),
                &["selection", "deletion"],
            ),
            (dictgen::InsensitiveStr::Ascii("capcacity"), &["capacity"]),
            (
                dictgen::InsensitiveStr::Ascii("formattind"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcuated"), &["calculated"]),
            (
                dictgen::InsensitiveStr::Ascii("differens"),
                &["differences", "difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("pachtes"), &["patches"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlfowing"),
                &["overflowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsenses"), &["offenses"]),
            (dictgen::InsensitiveStr::Ascii("pendantic"), &["pedantic"]),
            (
                dictgen::InsensitiveStr::Ascii("generacional"),
                &["generational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("braziliaan"),
                &["brazilians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relatib"),
                &["relative", "relatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("seplling"), &["spelling"]),
            (dictgen::InsensitiveStr::Ascii("succeesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("ziper"), &["zipper"]),
            (dictgen::InsensitiveStr::Ascii("republi"), &["republic"]),
            (dictgen::InsensitiveStr::Ascii("traved"), &["traversed"]),
            (
                dictgen::InsensitiveStr::Ascii("providince"),
                &["providence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diaram"),
                &["diagram", "diorama"],
            ),
            (dictgen::InsensitiveStr::Ascii("harases"), &["harasses"]),
            (
                dictgen::InsensitiveStr::Ascii("spontaenous"),
                &["spontaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optmisation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("actally"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("retaired"), &["retired"]),
            (dictgen::InsensitiveStr::Ascii("cresh"), &["crèche"]),
            (
                dictgen::InsensitiveStr::Ascii("vertifiable"),
                &["verifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wholeheartedy"),
                &["wholeheartedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extracions"),
                &["extractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefcification"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inidcation"),
                &["indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commutive"),
                &["commutative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conffiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transferrin"),
                &["transferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurisvely"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photoshoppade"),
                &["photoshopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximeted"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhanement"),
                &["enhancement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("establishmet"),
                &["establishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("doxygene"), &["doxygen"]),
            (dictgen::InsensitiveStr::Ascii("dealocate"), &["deallocate"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetns"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("newance"), &["nuance"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulizing"),
                &["virtualizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("profirled"), &["profiled"]),
            (dictgen::InsensitiveStr::Ascii("rabit"), &["rabbit"]),
            (
                dictgen::InsensitiveStr::Ascii("concurancy"),
                &["concurrency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homelesness"),
                &["homelessness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geograhpical"),
                &["geographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unprepaired"),
                &["unprepared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiifications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("checl"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("veritgo"), &["vertigo"]),
            (
                dictgen::InsensitiveStr::Ascii("comaprison"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentating"),
                &["concatenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celibrations"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheighed"), &["weighed"]),
            (dictgen::InsensitiveStr::Ascii("eecutable"), &["executable"]),
            (dictgen::InsensitiveStr::Ascii("infinty"), &["infinity"]),
            (
                dictgen::InsensitiveStr::Ascii("instansiated"),
                &["instantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbutions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("distrcts"), &["districts"]),
            (
                dictgen::InsensitiveStr::Ascii("faciliated"),
                &["facilitated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manigesting"),
                &["manifesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cononical"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("bugix"), &["bugfix"]),
            (dictgen::InsensitiveStr::Ascii("multipls"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("fomulations"),
                &["formulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expodential"),
                &["exponential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcions"),
                &["constructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporpriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("irradates"), &["irradiates"]),
            (dictgen::InsensitiveStr::Ascii("licesned"), &["licensed"]),
            (dictgen::InsensitiveStr::Ascii("acrade"), &["arcade"]),
            (dictgen::InsensitiveStr::Ascii("lsiting"), &["listing"]),
            (dictgen::InsensitiveStr::Ascii("inocuous"), &["innocuous"]),
            (
                dictgen::InsensitiveStr::Ascii("appopriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("qualifieed"), &["qualified"]),
            (
                dictgen::InsensitiveStr::Ascii("converst"),
                &["converts", "convert"],
            ),
            (dictgen::InsensitiveStr::Ascii("ineherited"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("bakups"), &["backups"]),
            (
                dictgen::InsensitiveStr::Ascii("signles"),
                &["singles", "signals"],
            ),
            (dictgen::InsensitiveStr::Ascii("flamable"), &["flammable"]),
            (
                dictgen::InsensitiveStr::Ascii("chrsitianity"),
                &["christianity"],
            ),
            (dictgen::InsensitiveStr::Ascii("checlsums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("formattied"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("beause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("madsion"), &["madison"]),
            (dictgen::InsensitiveStr::Ascii("uprgaded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("symbles"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("abstrator"), &["abstractor"]),
            (
                dictgen::InsensitiveStr::Ascii("differenatiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("hsa"), &["has"]),
            (dictgen::InsensitiveStr::Ascii("actualyl"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("destributions"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetetive"),
                &["repetitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functitonally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restaraunteurs"),
                &["restaurateurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downsteramed"),
                &["downstreamed"],
            ),
            (dictgen::InsensitiveStr::Ascii("devince"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("assistents"),
                &["assistants"],
            ),
            (dictgen::InsensitiveStr::Ascii("minuts"), &["minutes"]),
            (dictgen::InsensitiveStr::Ascii("qeusts"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("advirtisements"),
                &["advertisements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expropiated"),
                &["expropriated"],
            ),
            (dictgen::InsensitiveStr::Ascii("propmt"), &["prompt"]),
            (dictgen::InsensitiveStr::Ascii("sugguest"), &["suggest"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("missigno"), &["mission"]),
            (
                dictgen::InsensitiveStr::Ascii("interseccting"),
                &["intersecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("contiain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("bumb"),
                &["bump", "bomb", "bum"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaners"), &["containers"]),
            (dictgen::InsensitiveStr::Ascii("kindess"), &["kindness"]),
            (
                dictgen::InsensitiveStr::Ascii("architecutures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("acutions"), &["auctions"]),
            (
                dictgen::InsensitiveStr::Ascii("millescond"),
                &["millisecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("linheight"), &["lineheight"]),
            (dictgen::InsensitiveStr::Ascii("ofsetted"), &["offsetted"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibiton"),
                &["prohibition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbtrarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucur"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("chked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("aggravatin"),
                &["aggravating", "aggravation"],
            ),
            (dictgen::InsensitiveStr::Ascii("thumbnial"), &["thumbnail"]),
            (dictgen::InsensitiveStr::Ascii("resrict"), &["restrict"]),
            (dictgen::InsensitiveStr::Ascii("expesncive"), &["expensive"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessasary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuppported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("chlids"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("cofrims"), &["confirms"]),
            (dictgen::InsensitiveStr::Ascii("firebals"), &["fireballs"]),
            (
                dictgen::InsensitiveStr::Ascii("normaliztion"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechanismn"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("substituation"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("brekaout"), &["breakout"]),
            (dictgen::InsensitiveStr::Ascii("hsitories"), &["histories"]),
            (dictgen::InsensitiveStr::Ascii("adquire"), &["acquire"]),
            (dictgen::InsensitiveStr::Ascii("commadns"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylsed"), &["analysed"]),
            (
                dictgen::InsensitiveStr::Ascii("manifestano"),
                &["manifesto"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professiomal"),
                &["professionalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("ansestor"), &["ancestor"]),
            (dictgen::InsensitiveStr::Ascii("constanly"), &["constantly"]),
            (dictgen::InsensitiveStr::Ascii("abondons"), &["abandons"]),
            (dictgen::InsensitiveStr::Ascii("impelents"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("expeirimental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("controled"), &["controlled"]),
            (
                dictgen::InsensitiveStr::Ascii("harranging"),
                &["arranging", "haranguing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhacements"),
                &["enhancements"],
            ),
            (dictgen::InsensitiveStr::Ascii("braoder"), &["broader"]),
            (dictgen::InsensitiveStr::Ascii("ralate"), &["relate"]),
            (
                dictgen::InsensitiveStr::Ascii("advertister"),
                &["advertiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devasteted"),
                &["devastated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perticipates"),
                &["participates"],
            ),
            (dictgen::InsensitiveStr::Ascii("wuld"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("proximty"), &["proximity"]),
            (dictgen::InsensitiveStr::Ascii("nutraly"), &["neutrally"]),
            (
                dictgen::InsensitiveStr::Ascii("philisophical"),
                &["philosophical"],
            ),
            (dictgen::InsensitiveStr::Ascii("guareentee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("expeecting"), &["expecting"]),
            (dictgen::InsensitiveStr::Ascii("countain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("execuatables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleesheys"), &["clichés"]),
            (dictgen::InsensitiveStr::Ascii("agrentina"), &["argentina"]),
            (
                dictgen::InsensitiveStr::Ascii("correcponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("restriant"), &["restraint"]),
            (
                dictgen::InsensitiveStr::Ascii("walkthoughs"),
                &["walkthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comptability"),
                &["compatibility", "computability"],
            ),
            (dictgen::InsensitiveStr::Ascii("elminates"), &["eliminates"]),
            (dictgen::InsensitiveStr::Ascii("camer"), &["camera"]),
            (dictgen::InsensitiveStr::Ascii("decending"), &["descending"]),
            (dictgen::InsensitiveStr::Ascii("seniores"), &["seniors"]),
            (dictgen::InsensitiveStr::Ascii("depercate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("prferred"), &["preferred"]),
            (
                dictgen::InsensitiveStr::Ascii("modfiication"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispacher"), &["dispatcher"]),
            (
                dictgen::InsensitiveStr::Ascii("survivabiliy"),
                &["survivability"],
            ),
            (dictgen::InsensitiveStr::Ascii("cokies"), &["cookies"]),
            (dictgen::InsensitiveStr::Ascii("gaurantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("mutlisampling"),
                &["multisampling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidentially"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("briter"), &["brighter"]),
            (dictgen::InsensitiveStr::Ascii("idenified"), &["identified"]),
            (
                dictgen::InsensitiveStr::Ascii("imcompatibility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missinterprets"),
                &["misinterprets"],
            ),
            (dictgen::InsensitiveStr::Ascii("poicy"), &["policy"]),
            (
                dictgen::InsensitiveStr::Ascii("mathemagics"),
                &["mathematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentiers"), &["sentries"]),
            (
                dictgen::InsensitiveStr::Ascii("unifromity"),
                &["uniformity"],
            ),
            (dictgen::InsensitiveStr::Ascii("machanic"), &["mechanic"]),
            (dictgen::InsensitiveStr::Ascii("vaccineras"), &["vaccines"]),
            (
                dictgen::InsensitiveStr::Ascii("incrementarla"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effecitveness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archictectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("sthe"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("ellected"), &["elected"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnererability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existst"),
                &["exists", "exist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prdpagates"),
                &["propagates"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyzied"), &["analyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("cautionsly"),
                &["cautiously"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeset"), &["reset"]),
            (
                dictgen::InsensitiveStr::Ascii("humantiarian"),
                &["humanitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confunction"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distortian"),
                &["distortion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completipn"),
                &["completion"],
            ),
            (dictgen::InsensitiveStr::Ascii("comeputer"), &["computer"]),
            (
                dictgen::InsensitiveStr::Ascii("complating"),
                &["completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rebuliding"),
                &["rebuilding"],
            ),
            (dictgen::InsensitiveStr::Ascii("finf"), &["find"]),
            (dictgen::InsensitiveStr::Ascii("cernomy"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("organizarme"),
                &["organizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rationalisin"),
                &["rationalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("veriier"), &["verifier"]),
            (dictgen::InsensitiveStr::Ascii("personell"), &["personnel"]),
            (
                dictgen::InsensitiveStr::Ascii("hierarachical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improbabil"),
                &["improbable"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuable"), &["reusable"]),
            (dictgen::InsensitiveStr::Ascii("filesytem"), &["filesystem"]),
            (dictgen::InsensitiveStr::Ascii("eneter"), &["enter"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatio"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportionallly"),
                &["proportionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeuests"), &["requests"]),
            (
                dictgen::InsensitiveStr::Ascii("termplating"),
                &["templating"],
            ),
            (dictgen::InsensitiveStr::Ascii("follfowers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("cmoputers"), &["computers"]),
            (dictgen::InsensitiveStr::Ascii("egistry"), &["registry"]),
            (
                dictgen::InsensitiveStr::Ascii("scientifcly"),
                &["scientifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("raplacements"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teleprotation"),
                &["teleportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallusinations"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteraed"), &["iterated"]),
            (
                dictgen::InsensitiveStr::Ascii("methamatician"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("reachs"), &["reaches"]),
            (
                dictgen::InsensitiveStr::Ascii("conditially"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ertor"),
                &["error", "terror"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vegatable"), &["vegetable"]),
            (
                dictgen::InsensitiveStr::Ascii("ceckerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaase"),
                &["release", "relapse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vizualizes"),
                &["visualizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guareented"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finializing"),
                &["finalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloews"), &["allows"]),
            (dictgen::InsensitiveStr::Ascii("procedre"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("allwing"), &["allowing"]),
            (dictgen::InsensitiveStr::Ascii("expresion"), &["expression"]),
            (dictgen::InsensitiveStr::Ascii("quantite"), &["quantities"]),
            (dictgen::InsensitiveStr::Ascii("acnedote"), &["anecdote"]),
            (dictgen::InsensitiveStr::Ascii("navigatore"), &["navigate"]),
            (
                dictgen::InsensitiveStr::Ascii("multipling"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extractins"),
                &["extractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplissement"),
                &["accomplishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("peepel"), &["people"]),
            (
                dictgen::InsensitiveStr::Ascii("simplies"),
                &["implies", "simplifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiantiate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("omited"), &["omitted"]),
            (dictgen::InsensitiveStr::Ascii("axuiliary"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("expolited"), &["exploited"]),
            (dictgen::InsensitiveStr::Ascii("wiscled"), &["whistled"]),
            (dictgen::InsensitiveStr::Ascii("geogrpahy"), &["geography"]),
            (
                dictgen::InsensitiveStr::Ascii("permissons"),
                &["permissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("syphyllis"), &["syphilis"]),
            (dictgen::InsensitiveStr::Ascii("duplate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("recevie"), &["receive"]),
            (
                dictgen::InsensitiveStr::Ascii("disgnostics"),
                &["diagnostics"],
            ),
            (dictgen::InsensitiveStr::Ascii("useble"), &["usable"]),
            (
                dictgen::InsensitiveStr::Ascii("diplomatisch"),
                &["diplomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("steeleries"), &["steelers"]),
            (dictgen::InsensitiveStr::Ascii("relys"), &["relies"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("procedual"), &["procedural"]),
            (dictgen::InsensitiveStr::Ascii("numbert"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("subcribe"), &["subscribe"]),
            (
                dictgen::InsensitiveStr::Ascii("lingvistic"),
                &["linguistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clonez"),
                &["clones", "cloner"],
            ),
            (dictgen::InsensitiveStr::Ascii("identcial"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("logoic"), &["logic"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schoalrship"),
                &["scholarships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionaire"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("refernce"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("moodule"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("confim"), &["confirm"]),
            (dictgen::InsensitiveStr::Ascii("gerenally"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("regardlesss"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contiainers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkoxes"), &["checkboxes"]),
            (dictgen::InsensitiveStr::Ascii("temperary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("emitable"), &["emittable"]),
            (dictgen::InsensitiveStr::Ascii("backgroun"), &["background"]),
            (dictgen::InsensitiveStr::Ascii("depenancy"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("covnerting"),
                &["converting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivisual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("deletete"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("caricate"), &["caricature"]),
            (
                dictgen::InsensitiveStr::Ascii("uninstaling"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("enrypting"), &["encrypting"]),
            (
                dictgen::InsensitiveStr::Ascii("calcultator"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workstaiton"),
                &["workstation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsusable"), &["unusable"]),
            (dictgen::InsensitiveStr::Ascii("freit"), &["freight"]),
            (dictgen::InsensitiveStr::Ascii("broadley"), &["broadly"]),
            (dictgen::InsensitiveStr::Ascii("hvea"), &["have", "heave"]),
            (
                dictgen::InsensitiveStr::Ascii("cooporation"),
                &["cooperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("talkign"), &["talking"]),
            (
                dictgen::InsensitiveStr::Ascii("descripiton"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("purposin"), &["purposing"]),
            (
                dictgen::InsensitiveStr::Ascii("ackumulating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comapletion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porpose"),
                &["propose", "purpose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("storange"),
                &["storage", "strange"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compremises"),
                &["compromises"],
            ),
            (dictgen::InsensitiveStr::Ascii("drummin"), &["drumming"]),
            (
                dictgen::InsensitiveStr::Ascii("amealiarate"),
                &["ameliorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abitration"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("chilvary"), &["chivalry"]),
            (dictgen::InsensitiveStr::Ascii("engene"), &["engine"]),
            (dictgen::InsensitiveStr::Ascii("respectes"), &["respects"]),
            (dictgen::InsensitiveStr::Ascii("accepance"), &["acceptance"]),
            (dictgen::InsensitiveStr::Ascii("lifestlye"), &["lifestyle"]),
            (
                dictgen::InsensitiveStr::Ascii("intruducing"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trandgender"),
                &["transgender"],
            ),
            (dictgen::InsensitiveStr::Ascii("comiler"), &["compiler"]),
            (
                dictgen::InsensitiveStr::Ascii("comparitively"),
                &["comparatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbaitm"), &["verbatim"]),
            (dictgen::InsensitiveStr::Ascii("casion"), &["caisson"]),
            (
                dictgen::InsensitiveStr::Ascii("prosicutor"),
                &["prosecutor"],
            ),
            (dictgen::InsensitiveStr::Ascii("covriates"), &["covariates"]),
            (dictgen::InsensitiveStr::Ascii("qests"), &["quests"]),
            (dictgen::InsensitiveStr::Ascii("sourece"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("yourselv"),
                &["yourself", "yourselves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemeting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inrements"), &["increments"]),
            (dictgen::InsensitiveStr::Ascii("electivite"), &["elective"]),
            (dictgen::InsensitiveStr::Ascii("cariiers"), &["carriers"]),
            (dictgen::InsensitiveStr::Ascii("chocolats"), &["chocolates"]),
            (dictgen::InsensitiveStr::Ascii("octobear"), &["october"]),
            (dictgen::InsensitiveStr::Ascii("virtuels"), &["virtues"]),
            (
                dictgen::InsensitiveStr::Ascii("allpication"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("longitute"), &["longitude"]),
            (
                dictgen::InsensitiveStr::Ascii("proponenets"),
                &["proponents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innappropriate"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("desktiop"), &["desktop"]),
            (dictgen::InsensitiveStr::Ascii("neether"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("oligrachy"), &["oligarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("deceptionist"),
                &["receptionist"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearitos"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("mocrophone"),
                &["microphone"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceald"), &["concealed"]),
            (dictgen::InsensitiveStr::Ascii("currious"), &["curious"]),
            (dictgen::InsensitiveStr::Ascii("fluorish"), &["flourish"]),
            (dictgen::InsensitiveStr::Ascii("certains"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("perrror"), &["perror"]),
            (dictgen::InsensitiveStr::Ascii("indictes"), &["indicates"]),
            (dictgen::InsensitiveStr::Ascii("provdies"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("bastardos"), &["bastards"]),
            (dictgen::InsensitiveStr::Ascii("psueudo"), &["pseudo"]),
            (dictgen::InsensitiveStr::Ascii("cocument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("notifycation"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presearing"),
                &["preserving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insitutions"),
                &["institutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("interfces"), &["interfaces"]),
            (dictgen::InsensitiveStr::Ascii("modifyes"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("desgination"),
                &["designation"],
            ),
            (dictgen::InsensitiveStr::Ascii("precending"), &["preceding"]),
            (
                dictgen::InsensitiveStr::Ascii("quadrilaterlas"),
                &["quadrilaterals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permuataion"),
                &["permutation"],
            ),
            (dictgen::InsensitiveStr::Ascii("prevous"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("diggestion"), &["digestion"]),
            (
                dictgen::InsensitiveStr::Ascii("configarations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("passwrod"), &["password"]),
            (dictgen::InsensitiveStr::Ascii("relaion"), &["relation"]),
            (dictgen::InsensitiveStr::Ascii("bufufer"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("referes"),
                &["refers", "referees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justifcations"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhouds"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmently"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comiplation"),
                &["compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concrets"),
                &["concerts", "concrete"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteracts"), &["interacts"]),
            (dictgen::InsensitiveStr::Ascii("picthfork"), &["pitchfork"]),
            (dictgen::InsensitiveStr::Ascii("suseeds"), &["secedes"]),
            (
                dictgen::InsensitiveStr::Ascii("unspported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("busniess"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("transfmrom"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("thankfuly"), &["thankfully"]),
            (
                dictgen::InsensitiveStr::Ascii("deambiguates"),
                &["disambiguates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equels"),
                &["equals", "sequels"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacciners"), &["vaccines"]),
            (
                dictgen::InsensitiveStr::Ascii("dernormalization"),
                &["denormalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exadecimal"),
                &["hexadecimal"],
            ),
            (dictgen::InsensitiveStr::Ascii("keboard"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("congifs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("exercice"), &["exercise"]),
            (
                dictgen::InsensitiveStr::Ascii("conncetion"),
                &["connections", "connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whiher"),
                &["whether", "whiter", "whither", "whiner", "whinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("menaing"),
                &["meaning", "menacing", "mending"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapat"), &["adapt"]),
            (
                dictgen::InsensitiveStr::Ascii("uploades"),
                &["uploads", "uploaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("akkumulation"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("automato"), &["automation"]),
            (
                dictgen::InsensitiveStr::Ascii("assoiating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisaction"),
                &["satisfaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanifying"),
                &["quantifying", "qualifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocurred"), &["occurred"]),
            (
                dictgen::InsensitiveStr::Ascii("insentivite"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compononent"),
                &["component"],
            ),
            (dictgen::InsensitiveStr::Ascii("clokwise"), &["clockwise"]),
            (
                dictgen::InsensitiveStr::Ascii("conigurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justifyable"),
                &["justifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("practicle"), &["practical"]),
            (
                dictgen::InsensitiveStr::Ascii("complimente"),
                &["complement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitivo"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insititutional"),
                &["institutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarammatically"),
                &["programmatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onmishints"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("exurpts"), &["excerpts"]),
            (dictgen::InsensitiveStr::Ascii("toogling"), &["toggling"]),
            (dictgen::InsensitiveStr::Ascii("foppy"), &["floppy"]),
            (dictgen::InsensitiveStr::Ascii("conents"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("sufface"), &["surface"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithimically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("direcctory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("flahsing"), &["flashing"]),
            (dictgen::InsensitiveStr::Ascii("messsenger"), &["messenger"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acivations"),
                &["activations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particapated"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("varning"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("oragnised"), &["organised"]),
            (
                dictgen::InsensitiveStr::Ascii("regitering"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("velociries"),
                &["velocities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commanndline"),
                &["commandline"],
            ),
            (dictgen::InsensitiveStr::Ascii("dentistas"), &["dentists"]),
            (dictgen::InsensitiveStr::Ascii("reserched"), &["researched"]),
            (
                dictgen::InsensitiveStr::Ascii("specificly"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("squered"), &["squared"]),
            (dictgen::InsensitiveStr::Ascii("unbounad"), &["unbound"]),
            (dictgen::InsensitiveStr::Ascii("zeored"), &["zeroed"]),
            (dictgen::InsensitiveStr::Ascii("devaulted"), &["defaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("trasmitting"),
                &["transmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatelly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("exaple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("suffication"),
                &["suffocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("navigting"), &["navigating"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalisin"),
                &["capitalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encouttering"),
                &["encountering"],
            ),
            (dictgen::InsensitiveStr::Ascii("enebles"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatibilies"),
                &["incompatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helicotper"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("comfigure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("necessiates"),
                &["necessities"],
            ),
            (dictgen::InsensitiveStr::Ascii("accetable"), &["acceptable"]),
            (dictgen::InsensitiveStr::Ascii("radify"), &["ratify"]),
            (
                dictgen::InsensitiveStr::Ascii("marganilize"),
                &["marginalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rhinosarus"),
                &["rhinoceros"],
            ),
            (dictgen::InsensitiveStr::Ascii("anomolous"), &["anomalous"]),
            (
                dictgen::InsensitiveStr::Ascii("collabarate"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("gude"), &["guide", "good"]),
            (dictgen::InsensitiveStr::Ascii("cellebrate"), &["celebrate"]),
            (dictgen::InsensitiveStr::Ascii("orgininals"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasudicaly"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definltely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acctivities"),
                &["activities"],
            ),
            (dictgen::InsensitiveStr::Ascii("poind"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("speficilleis"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demolision"),
                &["demolition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succcessfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignalble"),
                &["assignable"],
            ),
            (dictgen::InsensitiveStr::Ascii("smallar"), &["smaller"]),
            (dictgen::InsensitiveStr::Ascii("carrears"), &["careers"]),
            (
                dictgen::InsensitiveStr::Ascii("fase"),
                &["faze", "phase", "false"],
            ),
            (dictgen::InsensitiveStr::Ascii("respponses"), &["responses"]),
            (dictgen::InsensitiveStr::Ascii("conrols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("cannobt"), &["cannot"]),
            (dictgen::InsensitiveStr::Ascii("receieve"), &["receive"]),
            (
                dictgen::InsensitiveStr::Ascii("correctons"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("napoming"), &["napalming"]),
            (
                dictgen::InsensitiveStr::Ascii("desstructor"),
                &["destructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("starightforward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("reomvable"), &["removable"]),
            (dictgen::InsensitiveStr::Ascii("coversion"), &["conversion"]),
            (dictgen::InsensitiveStr::Ascii("tryes"), &["tries"]),
            (
                dictgen::InsensitiveStr::Ascii("translatied"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ojbectively"),
                &["objectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracterisation"),
                &["characterisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalutaing"),
                &["evaluating"],
            ),
            (dictgen::InsensitiveStr::Ascii("provieder"), &["provider"]),
            (
                dictgen::InsensitiveStr::Ascii("deplicating"),
                &["duplicating", "replicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neckbeardius"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifyed"),
                &["simplified"],
            ),
            (dictgen::InsensitiveStr::Ascii("mackeral"), &["mackerel"]),
            (
                dictgen::InsensitiveStr::Ascii("cylindical"),
                &["cylindrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuie"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornograghy"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirences"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("intregal"), &["integral"]),
            (
                dictgen::InsensitiveStr::Ascii("reevaludated"),
                &["reevaluated"],
            ),
            (dictgen::InsensitiveStr::Ascii("postioned"), &["positioned"]),
            (
                dictgen::InsensitiveStr::Ascii("tramautized"),
                &["traumatized"],
            ),
            (dictgen::InsensitiveStr::Ascii("disablity"), &["disability"]),
            (dictgen::InsensitiveStr::Ascii("seqeuence"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("incorparation"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("utopain"), &["utopian"]),
            (dictgen::InsensitiveStr::Ascii("thridly"), &["thirdly"]),
            (
                dictgen::InsensitiveStr::Ascii("incorrectled"),
                &["incorrectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovverides"), &["overrides"]),
            (
                dictgen::InsensitiveStr::Ascii("narrativos"),
                &["narratives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmodfided"),
                &["unmodified"],
            ),
            (dictgen::InsensitiveStr::Ascii("malasiya"), &["malaysia"]),
            (dictgen::InsensitiveStr::Ascii("confg"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("operativne"), &["operative"]),
            (dictgen::InsensitiveStr::Ascii("exatct"), &["exact"]),
            (
                dictgen::InsensitiveStr::Ascii("candadates"),
                &["candidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("otiginal"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("tennants"),
                &["tenants", "pennants"],
            ),
            (dictgen::InsensitiveStr::Ascii("symbolisch"), &["symbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstraters"),
                &["demonstrators", "demonstrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("averge"),
                &["average", "avenge"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitre"), &["arbiter"]),
            (
                dictgen::InsensitiveStr::Ascii("determinint"),
                &["determining", "determinant", "determist"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("adaptates"), &["adapts"]),
            (
                dictgen::InsensitiveStr::Ascii("acomodates"),
                &["accommodates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responcibly"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzation"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converstion"),
                &["conversion", "conversation"],
            ),
            (dictgen::InsensitiveStr::Ascii("diasable"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("cylclically"),
                &["cyclically"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncheked"), &["unchecked"]),
            (
                dictgen::InsensitiveStr::Ascii("offisionado"),
                &["aficionado"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regresives"),
                &["regressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("differetnt"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunaly"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("championsship"),
                &["championships"],
            ),
            (dictgen::InsensitiveStr::Ascii("betwean"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("fanatism"), &["fanaticism"]),
            (
                dictgen::InsensitiveStr::Ascii("ressemblance"),
                &["resemblance"],
            ),
            (dictgen::InsensitiveStr::Ascii("lteral"), &["literal"]),
            (dictgen::InsensitiveStr::Ascii("hatesink"), &["heatsink"]),
            (dictgen::InsensitiveStr::Ascii("dchp"), &["dhcp"]),
            (
                dictgen::InsensitiveStr::Ascii("volounteered"),
                &["volunteered"],
            ),
            (dictgen::InsensitiveStr::Ascii("devleoped"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("finctioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("roigin"), &["origin"]),
            (
                dictgen::InsensitiveStr::Ascii("proects"),
                &["protects", "projects"],
            ),
            (dictgen::InsensitiveStr::Ascii("inputing"), &["inputting"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatibilties"),
                &["incompatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congegration"),
                &["congregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pedophils"), &["pedophiles"]),
            (dictgen::InsensitiveStr::Ascii("negotiote"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("unbelievibly"),
                &["unbelievably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fondamentalists"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foreignors"),
                &["foreigners"],
            ),
            (dictgen::InsensitiveStr::Ascii("undefien"), &["undefine"]),
            (
                dictgen::InsensitiveStr::Ascii("backaloriette"),
                &["baccalaureate"],
            ),
            (dictgen::InsensitiveStr::Ascii("turltes"), &["turtles"]),
            (dictgen::InsensitiveStr::Ascii("traitorus"), &["traitors"]),
            (dictgen::InsensitiveStr::Ascii("damamging"), &["damaging"]),
            (dictgen::InsensitiveStr::Ascii("drwaing"), &["drawing"]),
            (dictgen::InsensitiveStr::Ascii("progate"), &["propagate"]),
            (
                dictgen::InsensitiveStr::Ascii("hypenation"),
                &["hyphenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pieceweise"), &["piecewise"]),
            (
                dictgen::InsensitiveStr::Ascii("discriminted"),
                &["discriminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressues"), &["pressures"]),
            (
                dictgen::InsensitiveStr::Ascii("implementated"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("minitaure"), &["miniature"]),
            (
                dictgen::InsensitiveStr::Ascii("mechnical"),
                &["mechanical", "technical"],
            ),
            (dictgen::InsensitiveStr::Ascii("asuterity"), &["austerity"]),
            (
                dictgen::InsensitiveStr::Ascii("moddifications"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigatie"),
                &["investigative"],
            ),
            (dictgen::InsensitiveStr::Ascii("malayisan"), &["malaysian"]),
            (dictgen::InsensitiveStr::Ascii("outpu"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("faciliating"),
                &["facilitating"],
            ),
            (dictgen::InsensitiveStr::Ascii("featchs"), &["fetches"]),
            (
                dictgen::InsensitiveStr::Ascii("descriotively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interates"),
                &["iterates", "interacts", "integrated", "integrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("studoi"), &["studio"]),
            (dictgen::InsensitiveStr::Ascii("overide"), &["override"]),
            (
                dictgen::InsensitiveStr::Ascii("rereated"),
                &["recreated", "rerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harranges"),
                &["arranges", "harangues"],
            ),
            (dictgen::InsensitiveStr::Ascii("telemtry"), &["telemetry"]),
            (dictgen::InsensitiveStr::Ascii("bluebery"), &["blueberry"]),
            (
                dictgen::InsensitiveStr::Ascii("accummulation"),
                &["accumulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algororithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("hesiman"), &["heisman"]),
            (dictgen::InsensitiveStr::Ascii("opressive"), &["oppressive"]),
            (dictgen::InsensitiveStr::Ascii("agruments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("reding"), &["reading"]),
            (
                dictgen::InsensitiveStr::Ascii("ertors"),
                &["errors", "terrors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoisation"),
                &["authorisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arctifacts"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("undersands"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachterising"),
                &["characterising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coummunities"),
                &["communities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acccountant"),
                &["accountant"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocotpus"), &["octopus"]),
            (dictgen::InsensitiveStr::Ascii("provicer"), &["provider"]),
            (
                dictgen::InsensitiveStr::Ascii("traling"),
                &[
                    "trailing",
                    "trialing",
                    "tracing",
                    "trading",
                    "traveling",
                    "trawling",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("contentas"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("nuturing"), &["nurturing"]),
            (
                dictgen::InsensitiveStr::Ascii("architecuts"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strugglign"),
                &["struggling"],
            ),
            (dictgen::InsensitiveStr::Ascii("simluator"), &["simulator"]),
            (
                dictgen::InsensitiveStr::Ascii("transfomers"),
                &["transforms", "transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("languauges"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("closeer"), &["closer"]),
            (dictgen::InsensitiveStr::Ascii("denpending"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("decieded"), &["decided"]),
            (
                dictgen::InsensitiveStr::Ascii("quartenion"),
                &["quaternion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparasion"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accompliss"),
                &["accomplishes", "accomplish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siplifications"),
                &["simplifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribtes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimentions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtructing"),
                &["subtracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("poinnter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("caluse"),
                &["callous", "callus", "clause"],
            ),
            (dictgen::InsensitiveStr::Ascii("landmakrs"), &["landmarks"]),
            (dictgen::InsensitiveStr::Ascii("byast"), &["biased"]),
            (
                dictgen::InsensitiveStr::Ascii("konfiguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("bascic"), &["basic"]),
            (
                dictgen::InsensitiveStr::Ascii("assocaites"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotypian"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pubilshing"),
                &["publishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("kineitc"), &["kinetic"]),
            (dictgen::InsensitiveStr::Ascii("speciffics"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("whsipered"), &["whispered"]),
            (
                dictgen::InsensitiveStr::Ascii("progressivey"),
                &["progressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("strting"), &["string"]),
            (
                dictgen::InsensitiveStr::Ascii("inializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dertermine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("concleanment"),
                &["concealment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconstitutionnal"),
                &["unconstitutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comectivity"),
                &["connectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signoatures"),
                &["signatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("chercker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("annihalated"),
                &["annihilated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("happennings"),
                &["happenings"],
            ),
            (dictgen::InsensitiveStr::Ascii("permut"), &["permute"]),
            (
                dictgen::InsensitiveStr::Ascii("optionsl"),
                &["optional", "options"],
            ),
            (dictgen::InsensitiveStr::Ascii("callenged"), &["challenged"]),
            (dictgen::InsensitiveStr::Ascii("hav"), &["have", "half"]),
            (
                dictgen::InsensitiveStr::Ascii("adjcencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguiring"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("accuraccy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("covereed"), &["covered"]),
            (
                dictgen::InsensitiveStr::Ascii("paramnetrized"),
                &["parameterized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hsitorians"),
                &["historians"],
            ),
            (dictgen::InsensitiveStr::Ascii("enevlopes"), &["envelopes"]),
            (dictgen::InsensitiveStr::Ascii("concibes"), &["concise"]),
            (dictgen::InsensitiveStr::Ascii("arreay"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("impement"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("intrenets"), &["internets"]),
            (
                dictgen::InsensitiveStr::Ascii("sorcing"),
                &["sourcing", "forcing", "sorting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incidentaly"),
                &["incidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesname"),
                &["filename", "filenames"],
            ),
            (dictgen::InsensitiveStr::Ascii("excercises"), &["exercises"]),
            (dictgen::InsensitiveStr::Ascii("evluators"), &["evaluators"]),
            (dictgen::InsensitiveStr::Ascii("windoow"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoitating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("frackign"), &["fracking"]),
            (dictgen::InsensitiveStr::Ascii("cambpell"), &["campbell"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnformers"),
                &["transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("devisible"), &["divisible"]),
            (
                dictgen::InsensitiveStr::Ascii("lexographically"),
                &["lexicographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conffigurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraival"), &["arrival"]),
            (
                dictgen::InsensitiveStr::Ascii("histerically"),
                &["historically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advanteges"),
                &["advantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucuturing"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtatted"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaiton"),
                &["reallocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("describiste"),
                &["describes"],
            ),
            (dictgen::InsensitiveStr::Ascii("brose"), &["browse", "rose"]),
            (dictgen::InsensitiveStr::Ascii("transocde"), &["transcode"]),
            (dictgen::InsensitiveStr::Ascii("failuare"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("bacuse"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("multipicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("symobl"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("eferencing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deacriptor"),
                &["decryptor", "descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("imortable"), &["importable"]),
            (dictgen::InsensitiveStr::Ascii("fucntions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("negoitated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synonumous"),
                &["synonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialte"),
                &["initialise", "initiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("gridles"), &["griddles"]),
            (dictgen::InsensitiveStr::Ascii("litquids"), &["liquids"]),
            (dictgen::InsensitiveStr::Ascii("addessing"), &["addressing"]),
            (dictgen::InsensitiveStr::Ascii("progrma"), &["program"]),
            (dictgen::InsensitiveStr::Ascii("assulated"), &["assaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("instantaintg"),
                &["instantiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralellized"),
                &["parallelized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponantation"),
                &["exponentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("respnse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("luckely"), &["luckily"]),
            (dictgen::InsensitiveStr::Ascii("proccessor"), &["processor"]),
            (dictgen::InsensitiveStr::Ascii("annotaion"), &["annotation"]),
            (
                dictgen::InsensitiveStr::Ascii("bounudaries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unautherized"),
                &["unauthorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerrebral"), &["cerebral"]),
            (
                dictgen::InsensitiveStr::Ascii("accorndingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("morphie"), &["morphine"]),
            (
                dictgen::InsensitiveStr::Ascii("apprearance"),
                &["appearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplogize"), &["apologize"]),
            (dictgen::InsensitiveStr::Ascii("operattes"), &["operates"]),
            (dictgen::InsensitiveStr::Ascii("aditions"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("centeres"), &["centers"]),
            (dictgen::InsensitiveStr::Ascii("segfualts"), &["segfaults"]),
            (dictgen::InsensitiveStr::Ascii("caraboos"), &["caribous"]),
            (dictgen::InsensitiveStr::Ascii("follwoiong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("modidfied"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("knolwedgable"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singel"),
                &["single", "signal"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconsilde"), &["reconcile"]),
            (
                dictgen::InsensitiveStr::Ascii("efficvely"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nescesaries"),
                &["necessaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("additoin"), &["addition"]),
            (
                dictgen::InsensitiveStr::Ascii("privitazed"),
                &["privatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commets"),
                &["comets", "comments"],
            ),
            (dictgen::InsensitiveStr::Ascii("cananbis"), &["cannabis"]),
            (
                dictgen::InsensitiveStr::Ascii("propenents"),
                &["proponents"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgradin"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("eptiome"), &["epitome"]),
            (
                dictgen::InsensitiveStr::Ascii("programmend"),
                &["programmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incentivies"),
                &["incentives"],
            ),
            (dictgen::InsensitiveStr::Ascii("frop"), &["drop"]),
            (dictgen::InsensitiveStr::Ascii("algorythm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("undersandably"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("replusive"), &["repulsive"]),
            (
                dictgen::InsensitiveStr::Ascii("randazyvoos"),
                &["rendezvous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("frowrad"), &["forward"]),
            (dictgen::InsensitiveStr::Ascii("hrough"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("newcaste"), &["newcastle"]),
            (
                dictgen::InsensitiveStr::Ascii("propaganada"),
                &["propaganda"],
            ),
            (dictgen::InsensitiveStr::Ascii("distnaces"), &["distances"]),
            (
                dictgen::InsensitiveStr::Ascii("canabelysing"),
                &["cannibalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpyes"), &["types"]),
            (dictgen::InsensitiveStr::Ascii("libries"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("sepearately"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarrases"),
                &["embarrassed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nclusion"),
                &["inclusion", "conclusion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexuella"),
                &["homosexual"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundong"), &["bounding"]),
            (
                dictgen::InsensitiveStr::Ascii("environmentatlists"),
                &["environmentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissamble"),
                &["disassemble"],
            ),
            (dictgen::InsensitiveStr::Ascii("reparied"), &["repaired"]),
            (dictgen::InsensitiveStr::Ascii("lizensing"), &["licensing"]),
            (dictgen::InsensitiveStr::Ascii("identtical"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("questiond"), &["questioned"]),
            (dictgen::InsensitiveStr::Ascii("catogory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("senitnels"), &["sentinels"]),
            (dictgen::InsensitiveStr::Ascii("folloiwng"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("documneted"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("missign"), &["missing"]),
            (dictgen::InsensitiveStr::Ascii("abstrcter"), &["abstracter"]),
            (
                dictgen::InsensitiveStr::Ascii("reprents"),
                &["repents", "represents", "reprints"],
            ),
            (dictgen::InsensitiveStr::Ascii("yeasterday"), &["yesterday"]),
            (dictgen::InsensitiveStr::Ascii("breif"), &["brief"]),
            (
                dictgen::InsensitiveStr::Ascii("implementates"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("goemetry"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("actualy"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("whistel"), &["whistle"]),
            (
                dictgen::InsensitiveStr::Ascii("outnumberd"),
                &["outnumbered"],
            ),
            (dictgen::InsensitiveStr::Ascii("insatller"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("apppenders"), &["appenders"]),
            (dictgen::InsensitiveStr::Ascii("heightend"), &["heightened"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanging"),
                &["enchanting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransacations"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("switchtes"), &["switches"]),
            (dictgen::InsensitiveStr::Ascii("docrtines"), &["doctrines"]),
            (
                dictgen::InsensitiveStr::Ascii("charachers"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchagnges"), &["exchanges"]),
            (
                dictgen::InsensitiveStr::Ascii("ambassabor"),
                &["ambassador"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprectices"),
                &["apprentices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eficiencies"),
                &["efficiencies", "deficiencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heiroglyphics"),
                &["hieroglyphics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemishperes"),
                &["hemisphere", "hemispheres"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microcontroler"),
                &["microcontroller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projectives"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suitablity"),
                &["suitability"],
            ),
            (dictgen::InsensitiveStr::Ascii("delegage"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("unclude"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("carimonially"),
                &["ceremonially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitoning"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triagulates"),
                &["triangulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("disnabled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("certifaction"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("downgarde"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("provisonal"),
                &["provisional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hystericaly"),
                &["hysterically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscarriege"),
                &["miscarriage"],
            ),
            (dictgen::InsensitiveStr::Ascii("blizzad"), &["blizzard"]),
            (
                dictgen::InsensitiveStr::Ascii("masturpiece"),
                &["masterpiece"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partcularity"),
                &["particularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("moruning"), &["mourning"]),
            (dictgen::InsensitiveStr::Ascii("exahust"), &["exhaust"]),
            (dictgen::InsensitiveStr::Ascii("booleanss"), &["booleans"]),
            (dictgen::InsensitiveStr::Ascii("acitivity"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("faile"), &["failed", "fail"]),
            (
                dictgen::InsensitiveStr::Ascii("accessiiblity"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gringeworthy"),
                &["cringeworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regargless"),
                &["regardless"],
            ),
            (dictgen::InsensitiveStr::Ascii("wesites"), &["websites"]),
            (
                dictgen::InsensitiveStr::Ascii("prescripton"),
                &["prescription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mayballine"),
                &["maybelline"],
            ),
            (dictgen::InsensitiveStr::Ascii("quotted"), &["quoted"]),
            (dictgen::InsensitiveStr::Ascii("everythig"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("aktive"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("prestine"), &["pristine"]),
            (
                dictgen::InsensitiveStr::Ascii("vizualisations"),
                &["visualisation", "visualisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("furnctions"), &["functions"]),
            (dictgen::InsensitiveStr::Ascii("amacingly"), &["amazingly"]),
            (dictgen::InsensitiveStr::Ascii("formet"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("accurrately"),
                &["accurately"],
            ),
            (dictgen::InsensitiveStr::Ascii("screnns"), &["screens"]),
            (dictgen::InsensitiveStr::Ascii("cordiante"), &["coordinate"]),
            (dictgen::InsensitiveStr::Ascii("recided"), &["resided"]),
            (dictgen::InsensitiveStr::Ascii("exampleo"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("experiemnted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("statick"), &["static"]),
            (
                dictgen::InsensitiveStr::Ascii("commants"),
                &["commands", "comments"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilzes"), &["utilizes"]),
            (dictgen::InsensitiveStr::Ascii("souce"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("electrolites"),
                &["electrolytes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("educted"),
                &["educated", "deducted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpriting"),
                &["interpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivolent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invisibily"),
                &["invisibility", "invisibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticall"),
                &["automatically", "automatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aother"),
                &["another", "other", "mother"],
            ),
            (dictgen::InsensitiveStr::Ascii("deryption"), &["decryption"]),
            (dictgen::InsensitiveStr::Ascii("perceded"), &["preceded"]),
            (dictgen::InsensitiveStr::Ascii("horizones"), &["horizons"]),
            (dictgen::InsensitiveStr::Ascii("pipleines"), &["pipelines"]),
            (
                dictgen::InsensitiveStr::Ascii("speceficly"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelorated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninteruppted"),
                &["uninterrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("unqouted"), &["unquoted"]),
            (dictgen::InsensitiveStr::Ascii("exectures"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("hitories"), &["histories"]),
            (dictgen::InsensitiveStr::Ascii("frisday"), &["friday"]),
            (dictgen::InsensitiveStr::Ascii("supoort"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("informate"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntruction"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("varable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("ugprading"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("essentail"), &["essential"]),
            (dictgen::InsensitiveStr::Ascii("whislte"), &["whistle"]),
            (
                dictgen::InsensitiveStr::Ascii("uninmportant"),
                &["unimportant"],
            ),
            (dictgen::InsensitiveStr::Ascii("publisheed"), &["published"]),
            (
                dictgen::InsensitiveStr::Ascii("atttributed"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenitcation"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dorder"),
                &["order", "disorder"],
            ),
            (dictgen::InsensitiveStr::Ascii("listented"), &["listened"]),
            (
                dictgen::InsensitiveStr::Ascii("initializating"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifepsan"), &["lifespan"]),
            (dictgen::InsensitiveStr::Ascii("staticall"), &["statically"]),
            (dictgen::InsensitiveStr::Ascii("accdient"), &["accident"]),
            (dictgen::InsensitiveStr::Ascii("celebritiy"), &["celebrity"]),
            (
                dictgen::InsensitiveStr::Ascii("provisionned"),
                &["provisioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summerised"),
                &["summarised"],
            ),
            (dictgen::InsensitiveStr::Ascii("agreggate"), &["aggregate"]),
            (
                dictgen::InsensitiveStr::Ascii("reasonabley"),
                &["reasonably"],
            ),
            (dictgen::InsensitiveStr::Ascii("upstreem"), &["upstream"]),
            (dictgen::InsensitiveStr::Ascii("impace"), &["impact"]),
            (dictgen::InsensitiveStr::Ascii("privision"), &["provision"]),
            (
                dictgen::InsensitiveStr::Ascii("confirmaton"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defind"),
                &["defined", "defund"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heler"),
                &["helper", "healer"],
            ),
            (dictgen::InsensitiveStr::Ascii("amethsyts"), &["amethysts"]),
            (dictgen::InsensitiveStr::Ascii("penninsula"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterization"),
                &["parameterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retaliaton"),
                &["retaliation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exucute"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("expeirment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremints"),
                &["increments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prinicples"),
                &["principles"],
            ),
            (dictgen::InsensitiveStr::Ascii("getnerate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("operaterrg"), &["operating"]),
            (dictgen::InsensitiveStr::Ascii("nuatilus"), &["nautilus"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetnelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubitive"),
                &["distributive", "disruptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("provieded"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("sylinder"), &["cylinder"]),
            (dictgen::InsensitiveStr::Ascii("surpious"), &["spurious"]),
            (
                dictgen::InsensitiveStr::Ascii("surroundngs"),
                &["surrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autenticating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoaiative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specilazations"),
                &["specializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maeaningless"),
                &["meaningless"],
            ),
            (dictgen::InsensitiveStr::Ascii("everyhing"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("storeing"), &["storing"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessesary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repostories"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("priodical"), &["periodical"]),
            (
                dictgen::InsensitiveStr::Ascii("commissionees"),
                &["commissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("thtreaded"), &["threaded"]),
            (dictgen::InsensitiveStr::Ascii("serrebral"), &["cerebral"]),
            (dictgen::InsensitiveStr::Ascii("exatc"), &["exact"]),
            (dictgen::InsensitiveStr::Ascii("bunji"), &["bungee"]),
            (dictgen::InsensitiveStr::Ascii("montly"), &["monthly"]),
            (dictgen::InsensitiveStr::Ascii("accomplis"), &["accomplish"]),
            (
                dictgen::InsensitiveStr::Ascii("singnificance"),
                &["significance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attributse"),
                &["attributes", "attribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canabelyzing"),
                &["cannibalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("patiant"), &["patient"]),
            (
                dictgen::InsensitiveStr::Ascii("paticipate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("sypport"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("assymthotic"),
                &["asymptotic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ineffizient"),
                &["inefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("hadnle"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("sumodule"), &["submodule"]),
            (dictgen::InsensitiveStr::Ascii("strucgture"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("persauded"), &["persuaded"]),
            (
                dictgen::InsensitiveStr::Ascii("integratron"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytrofoam"), &["styrofoam"]),
            (
                dictgen::InsensitiveStr::Ascii("unilaterallly"),
                &["unilaterally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partioning"),
                &["partitioning", "portioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("wraning"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("endoints"), &["endpoints"]),
            (dictgen::InsensitiveStr::Ascii("reiceved"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("confety"), &["confetti"]),
            (dictgen::InsensitiveStr::Ascii("clenaup"), &["cleanup"]),
            (
                dictgen::InsensitiveStr::Ascii("reconizing"),
                &["recognizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abborts"),
                &["aborts", "abbots"],
            ),
            (dictgen::InsensitiveStr::Ascii("rarelly"), &["rarely"]),
            (
                dictgen::InsensitiveStr::Ascii("genralised"),
                &["generalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foor"),
                &["foot", "for", "door"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origninally"),
                &["originally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialyzes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpeters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("consitute"), &["constitute"]),
            (
                dictgen::InsensitiveStr::Ascii("undoubetdly"),
                &["undoubtedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicksih"), &["dickish"]),
            (dictgen::InsensitiveStr::Ascii("acheives"), &["achieves"]),
            (dictgen::InsensitiveStr::Ascii("chisled"), &["chiseled"]),
            (
                dictgen::InsensitiveStr::Ascii("comfortablely"),
                &["comfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("enitity"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("rto"), &["to"]),
            (
                dictgen::InsensitiveStr::Ascii("exaplanation"),
                &["explanation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosectued"),
                &["prosecuted"],
            ),
            (dictgen::InsensitiveStr::Ascii("backerds"), &["backers"]),
            (dictgen::InsensitiveStr::Ascii("guarneted"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("passess"),
                &["passes", "possess", "passers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmarization"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("humoros"), &["humorous"]),
            (dictgen::InsensitiveStr::Ascii("acccepted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchronious"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessleation"),
                &["tessellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transciver"),
                &["transceiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinforcin"),
                &["reinforcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondig"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuab"), &["visual"]),
            (dictgen::InsensitiveStr::Ascii("funchtion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("sucession"), &["succession"]),
            (dictgen::InsensitiveStr::Ascii("conrrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabity"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("austing"), &["austin"]),
            (dictgen::InsensitiveStr::Ascii("clinicals"), &["clinical"]),
            (dictgen::InsensitiveStr::Ascii("favourits"), &["favourites"]),
            (dictgen::InsensitiveStr::Ascii("genarates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("correxponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outragoues"),
                &["outrageous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriarcial"),
                &["patriarchal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plaing"),
                &["playing", "placing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptionts"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("deocding"), &["decoding"]),
            (
                dictgen::InsensitiveStr::Ascii("compulsury"),
                &["compulsory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workbanches"),
                &["workbenches"],
            ),
            (dictgen::InsensitiveStr::Ascii("condigs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("includee"), &["include"]),
            (dictgen::InsensitiveStr::Ascii("ploygons"), &["polygons"]),
            (dictgen::InsensitiveStr::Ascii("sacrafice"), &["sacrifice"]),
            (dictgen::InsensitiveStr::Ascii("fowll"), &["follow", "foul"]),
            (dictgen::InsensitiveStr::Ascii("colletor"), &["collector"]),
            (
                dictgen::InsensitiveStr::Ascii("engagemnet"),
                &["engagements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilthreaded"),
                &["multithreaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplucated"),
                &["duplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protectiei"),
                &["protective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginzes"),
                &["recognizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakcend"), &["backend"]),
            (
                dictgen::InsensitiveStr::Ascii("deccelerate"),
                &["decelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquisitivo"),
                &["inquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("shuttlin"), &["shuttling"]),
            (
                dictgen::InsensitiveStr::Ascii("consectutive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("publishor"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("excewption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("specufies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("messaing"), &["messaging"]),
            (dictgen::InsensitiveStr::Ascii("exchances"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("contestes"), &["contests"]),
            (
                dictgen::InsensitiveStr::Ascii("disporportionate"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("genertion"), &["generation"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaocrats"),
                &["bureaucrats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counsolers"),
                &["counselors"],
            ),
            (dictgen::InsensitiveStr::Ascii("effiecnet"), &["efficient"]),
            (
                dictgen::InsensitiveStr::Ascii("vietnamnese"),
                &["vietnamese"],
            ),
            (dictgen::InsensitiveStr::Ascii("wihtin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactoring"),
                &["manufacturing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversimplifiction"),
                &["oversimplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("intsalled"), &["installed"]),
            (
                dictgen::InsensitiveStr::Ascii("suppliment"),
                &["supplement"],
            ),
            (dictgen::InsensitiveStr::Ascii("humer"), &["humor"]),
            (dictgen::InsensitiveStr::Ascii("delelete"), &["delete"]),
            (
                dictgen::InsensitiveStr::Ascii("persussion"),
                &["persuasion"],
            ),
            (dictgen::InsensitiveStr::Ascii("pactch"), &["patch"]),
            (
                dictgen::InsensitiveStr::Ascii("dispensiary"),
                &["dispensary"],
            ),
            (dictgen::InsensitiveStr::Ascii("supsends"), &["suspends"]),
            (
                dictgen::InsensitiveStr::Ascii("ocassional"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorotize"),
                &["prioritize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soundrtack"),
                &["soundtracks"],
            ),
            (dictgen::InsensitiveStr::Ascii("futre"), &["future"]),
            (
                dictgen::InsensitiveStr::Ascii("apostrephe"),
                &["apostrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("fialover"), &["failover"]),
            (dictgen::InsensitiveStr::Ascii("previuous"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("prometheous"),
                &["prometheus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingush"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("faciliates"),
                &["facilitates"],
            ),
            (dictgen::InsensitiveStr::Ascii("worskspace"), &["workspace"]),
            (
                dictgen::InsensitiveStr::Ascii("subsquently"),
                &["subsequently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aclhemists"),
                &["alchemists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refernecing"),
                &["referencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collatin"),
                &["collating", "collation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheats"),
                &["spreadsheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("promixity"), &["proximity"]),
            (
                dictgen::InsensitiveStr::Ascii("fonctionnalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("hlep"), &["help"]),
            (dictgen::InsensitiveStr::Ascii("cahrs"), &["chars"]),
            (
                dictgen::InsensitiveStr::Ascii("manipluating"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearable"), &["separable"]),
            (
                dictgen::InsensitiveStr::Ascii("trgistration"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectons"),
                &["connections", "connectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prominantely"),
                &["prominently"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressses"), &["presses"]),
            (dictgen::InsensitiveStr::Ascii("expemtion"), &["exemption"]),
            (
                dictgen::InsensitiveStr::Ascii("interractive"),
                &["interactive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("choises"),
                &["choices", "chooses"],
            ),
            (dictgen::InsensitiveStr::Ascii("minituare"), &["miniature"]),
            (dictgen::InsensitiveStr::Ascii("obejctive"), &["objective"]),
            (dictgen::InsensitiveStr::Ascii("abosulte"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("bloching"), &["blocking"]),
            (
                dictgen::InsensitiveStr::Ascii("transpeorted"),
                &["transported"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhbition"), &["exhibition"]),
            (
                dictgen::InsensitiveStr::Ascii("silhouwetes"),
                &["silhouettes"],
            ),
            (dictgen::InsensitiveStr::Ascii("tipical"), &["typical"]),
            (dictgen::InsensitiveStr::Ascii("princpal"), &["principal"]),
            (dictgen::InsensitiveStr::Ascii("equeations"), &["equations"]),
            (
                dictgen::InsensitiveStr::Ascii("instructioins"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("esttings"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("allegry"), &["allergy"]),
            (dictgen::InsensitiveStr::Ascii("sumitted"), &["submitted"]),
            (dictgen::InsensitiveStr::Ascii("publlisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("awknowledged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identially"),
                &["identically"],
            ),
            (dictgen::InsensitiveStr::Ascii("pursude"), &["pursued"]),
            (dictgen::InsensitiveStr::Ascii("assfalt"), &["asphalt"]),
            (dictgen::InsensitiveStr::Ascii("wrapppers"), &["wrappers"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workbneches"),
                &["workbenches"],
            ),
            (dictgen::InsensitiveStr::Ascii("uptim"), &["uptime"]),
            (dictgen::InsensitiveStr::Ascii("philosopy"), &["philosophy"]),
            (dictgen::InsensitiveStr::Ascii("outfeild"), &["outfield"]),
            (dictgen::InsensitiveStr::Ascii("argumentated"), &["argued"]),
            (dictgen::InsensitiveStr::Ascii("fubding"), &["funding"]),
            (
                dictgen::InsensitiveStr::Ascii("mastrubating"),
                &["masturbating"],
            ),
            (dictgen::InsensitiveStr::Ascii("shippin"), &["shipping"]),
            (
                dictgen::InsensitiveStr::Ascii("underultilization"),
                &["underutilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeditonary"),
                &["expeditionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potientially"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("insall"), &["install"]),
            (dictgen::InsensitiveStr::Ascii("hannging"), &["hanging"]),
            (dictgen::InsensitiveStr::Ascii("ocur"), &["occur"]),
            (dictgen::InsensitiveStr::Ascii("moteur"), &["motor"]),
            (dictgen::InsensitiveStr::Ascii("suposes"), &["supposes"]),
            (
                dictgen::InsensitiveStr::Ascii("whitespcae"),
                &["whitespace"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytnax"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("unpredicable"),
                &["unpredictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("valentein"), &["valentine"]),
            (
                dictgen::InsensitiveStr::Ascii("observaton"),
                &["observation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponantial"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqeunces"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("doucment"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("contintent"),
                &["contingent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexistant"),
                &["inexistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("commuity"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("annonated"), &["annotated"]),
            (dictgen::InsensitiveStr::Ascii("expested"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("folled"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("impove"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("reencarnation"),
                &["reincarnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represended"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indestructuble"),
                &["indestructible"],
            ),
            (dictgen::InsensitiveStr::Ascii("dilemas"), &["dilemmas"]),
            (
                dictgen::InsensitiveStr::Ascii("reimbursment"),
                &["reimbursement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atuhenticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("subitles"), &["subtitles"]),
            (
                dictgen::InsensitiveStr::Ascii("validaiton"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dafualted"), &["defaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("informat"),
                &["inform", "informant"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefial"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("optimziations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depndancies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("appying"), &["applying"]),
            (dictgen::InsensitiveStr::Ascii("regaridng"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("capabitilies"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypoteses"), &["hypotheses"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliacted"),
                &["duplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("nothin"), &["nothing"]),
            (
                dictgen::InsensitiveStr::Ascii("availiability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("firday"), &["friday"]),
            (dictgen::InsensitiveStr::Ascii("elitistm"), &["elitism"]),
            (
                dictgen::InsensitiveStr::Ascii("progaramming"),
                &["programming"],
            ),
            (dictgen::InsensitiveStr::Ascii("faslely"), &["falsely"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensionnal"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("colunms"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("operatng"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("functinoalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propigating"),
                &["propagating"],
            ),
            (dictgen::InsensitiveStr::Ascii("doagonal"), &["diagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("legitamacy"),
                &["legitimacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assicating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("acroynm"), &["acronym"]),
            (
                dictgen::InsensitiveStr::Ascii("experiementally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandanivia"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("gauntlents"), &["gauntlet"]),
            (
                dictgen::InsensitiveStr::Ascii("objectivitiy"),
                &["objectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("colorizoer"), &["colorizer"]),
            (dictgen::InsensitiveStr::Ascii("flasing"), &["flashing"]),
            (
                dictgen::InsensitiveStr::Ascii("embarissing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("succceeded"), &["succeeded"]),
            (
                dictgen::InsensitiveStr::Ascii("equivalentsly"),
                &["equivalently", "equivalency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individaul"),
                &["individual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thsnkfulness"),
                &["thankfulness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenetaion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("andler"),
                &["handler", "antler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coresponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("adaper"), &["adapter"]),
            (
                dictgen::InsensitiveStr::Ascii("appriopriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliaser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decription"),
                &["description", "decryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discusison"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compositionwise"),
                &["compositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detination"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtitects"),
                &["architects"],
            ),
            (dictgen::InsensitiveStr::Ascii("genatilia"), &["genitalia"]),
            (dictgen::InsensitiveStr::Ascii("replcae"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("retreeving"),
                &["retrieving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disscusions"),
                &["discussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("brusting"), &["bursting"]),
            (dictgen::InsensitiveStr::Ascii("stremas"), &["streams"]),
            (dictgen::InsensitiveStr::Ascii("ecstasty"), &["ecstasy"]),
            (dictgen::InsensitiveStr::Ascii("lineupes"), &["lineups"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawalt"),
                &["withdrawal"],
            ),
            (dictgen::InsensitiveStr::Ascii("doocuments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("corportaion"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disimilarity"),
                &["dissimilarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("isplay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("oringal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("browseable"), &["browsable"]),
            (
                dictgen::InsensitiveStr::Ascii("interrupion"),
                &["interruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresponting"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphetmaine"),
                &["amphetamines", "amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralellism"),
                &["parallelism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialiazed"),
                &["specialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncically"),
                &["canonically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insparation"),
                &["inspiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("equipement"), &["equipment"]),
            (dictgen::InsensitiveStr::Ascii("elapsin"), &["elapsing"]),
            (
                dictgen::InsensitiveStr::Ascii("cusomizing"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranmissions"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("myselfe"), &["myself"]),
            (
                dictgen::InsensitiveStr::Ascii("succesions"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("scorin"), &["scoring"]),
            (
                dictgen::InsensitiveStr::Ascii("dulicate"),
                &["duplicate", "delicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("afernoons"), &["afternoons"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilites"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("applogize"), &["apologize"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendy"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("releeasing"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("cusoms"), &["customs"]),
            (
                dictgen::InsensitiveStr::Ascii("communiction"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidenty"),
                &["confidently"],
            ),
            (dictgen::InsensitiveStr::Ascii("takslet"), &["tasklet"]),
            (dictgen::InsensitiveStr::Ascii("evalueator"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("artsit"), &["artist"]),
            (
                dictgen::InsensitiveStr::Ascii("otherwordly"),
                &["otherworldly"],
            ),
            (dictgen::InsensitiveStr::Ascii("speical"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiamted"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepencence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properteis"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrtes"), &["writes"]),
            (
                dictgen::InsensitiveStr::Ascii("infreqency"),
                &["infrequency"],
            ),
            (dictgen::InsensitiveStr::Ascii("accusato"), &["accusation"]),
            (
                dictgen::InsensitiveStr::Ascii("technitians"),
                &["technicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constititional"),
                &["constitutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("tornaodes"), &["tornadoes"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsitantly"),
                &["inconsistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delusionally"),
                &["delusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprecated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("revisons"), &["revisions"]),
            (
                dictgen::InsensitiveStr::Ascii("creationsits"),
                &["creationists"],
            ),
            (dictgen::InsensitiveStr::Ascii("whatch"), &["watch"]),
            (
                dictgen::InsensitiveStr::Ascii("parliamenty"),
                &["parliamentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknwoledging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defferential"),
                &["differential", "deferential"],
            ),
            (dictgen::InsensitiveStr::Ascii("thimng"), &["thing"]),
            (dictgen::InsensitiveStr::Ascii("endcoded"), &["encoded"]),
            (
                dictgen::InsensitiveStr::Ascii("esitmators"),
                &["estimators"],
            ),
            (dictgen::InsensitiveStr::Ascii("lighnting"), &["lightning"]),
            (dictgen::InsensitiveStr::Ascii("indiated"), &["indicated"]),
            (
                dictgen::InsensitiveStr::Ascii("mutithreads"),
                &["multithreads"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understandabley"),
                &["understandably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circulationg"),
                &["circulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rhethorically"),
                &["rhetorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("involtue"), &["involute"]),
            (dictgen::InsensitiveStr::Ascii("negoated"), &["negotiated"]),
            (
                dictgen::InsensitiveStr::Ascii("diconnected"),
                &["disconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("molestaion"),
                &["molestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("autmated"), &["automated"]),
            (
                dictgen::InsensitiveStr::Ascii("enivitable"),
                &["inevitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarased"),
                &["embarrassed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradically"),
                &["contradictory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irritatin"),
                &["irritation", "irritating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arranents"),
                &["arrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transvorming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscirbed"),
                &["unsubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialialise"),
                &["serialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionnally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceide"), &["conceived"]),
            (dictgen::InsensitiveStr::Ascii("meni"), &["menu"]),
            (dictgen::InsensitiveStr::Ascii("commodoere"), &["commodore"]),
            (dictgen::InsensitiveStr::Ascii("hypocrit"), &["hypocrite"]),
            (
                dictgen::InsensitiveStr::Ascii("comparetors"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("documument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("replacation"),
                &["replication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alhpabetically"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("debub"), &["debug"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promethous"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttacking"), &["attacking"]),
            (dictgen::InsensitiveStr::Ascii("priro"), &["prior"]),
            (
                dictgen::InsensitiveStr::Ascii("maniuplates"),
                &["manipulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("repacing"), &["replacing"]),
            (
                dictgen::InsensitiveStr::Ascii("ackowledge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arithmethic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("connetctor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("dimentionalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialiase"),
                &["specialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auotmotive"),
                &["automotive"],
            ),
            (dictgen::InsensitiveStr::Ascii("recenly"), &["recently"]),
            (
                dictgen::InsensitiveStr::Ascii("decommissionned"),
                &["decommissioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("allcating"), &["allocating"]),
            (
                dictgen::InsensitiveStr::Ascii("initailiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondances"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("mmnemonic"), &["mnemonic"]),
            (
                dictgen::InsensitiveStr::Ascii("swtichboard"),
                &["switchboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("onwward"), &["onward"]),
            (dictgen::InsensitiveStr::Ascii("stabel"), &["stable"]),
            (
                dictgen::InsensitiveStr::Ascii("relatin"),
                &["relation", "relating"],
            ),
            (dictgen::InsensitiveStr::Ascii("endevers"), &["endeavors"]),
            (dictgen::InsensitiveStr::Ascii("polarlize"), &["polarize"]),
            (dictgen::InsensitiveStr::Ascii("ypour"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("congratulatons"),
                &["congratulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bulletproff"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renderes"),
                &["renders", "renderers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrating"),
                &["aggregating", "aggravating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contriubte"),
                &["contributes", "contribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoherant"),
                &["incoherent"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimazed"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("customsiation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("krankenstein"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("accademics"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("widgests"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("randome"), &["random"]),
            (
                dictgen::InsensitiveStr::Ascii("ideologice"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("minised"), &["minimised"]),
            (dictgen::InsensitiveStr::Ascii("depdennce"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("condntional"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("intialse"), &["initialise"]),
            (
                dictgen::InsensitiveStr::Ascii("termonological"),
                &["terminological"],
            ),
            (dictgen::InsensitiveStr::Ascii("triptickes"), &["triptychs"]),
            (dictgen::InsensitiveStr::Ascii("pluginss"), &["plugins"]),
            (
                dictgen::InsensitiveStr::Ascii("atmopshere"),
                &["atmosphere"],
            ),
            (dictgen::InsensitiveStr::Ascii("damamges"), &["damages"]),
            (dictgen::InsensitiveStr::Ascii("acedemia"), &["academia"]),
            (dictgen::InsensitiveStr::Ascii("custonise"), &["customise"]),
            (dictgen::InsensitiveStr::Ascii("blackend"), &["blacked"]),
            (dictgen::InsensitiveStr::Ascii("merhcants"), &["merchants"]),
            (dictgen::InsensitiveStr::Ascii("concedered"), &["conceded"]),
            (dictgen::InsensitiveStr::Ascii("contestais"), &["contests"]),
            (dictgen::InsensitiveStr::Ascii("matris"), &["matrix"]),
            (dictgen::InsensitiveStr::Ascii("aligmnets"), &["alignments"]),
            (
                dictgen::InsensitiveStr::Ascii("correktness"),
                &["correctness"],
            ),
            (dictgen::InsensitiveStr::Ascii("responsec"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("oigin"), &["origin"]),
            (
                dictgen::InsensitiveStr::Ascii("monestaries"),
                &["monasteries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representives"),
                &["representatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("picknicing"),
                &["picnicking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismanting"),
                &["dismantling"],
            ),
            (dictgen::InsensitiveStr::Ascii("valueable"), &["valuable"]),
            (dictgen::InsensitiveStr::Ascii("capbably"), &["capably"]),
            (dictgen::InsensitiveStr::Ascii("invertion"), &["inversion"]),
            (
                dictgen::InsensitiveStr::Ascii("hearily"),
                &["heartily", "heavily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agrumentative"),
                &["argumentative"],
            ),
            (dictgen::InsensitiveStr::Ascii("minumun"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("negoatiation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrading"),
                &["overriding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chechklists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overiddden"),
                &["overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("libell"), &["libel"]),
            (
                dictgen::InsensitiveStr::Ascii("partonizing"),
                &["patronizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("targers"),
                &["targets", "taggers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprocuce"),
                &["reproduce", "reprocure"],
            ),
            (dictgen::InsensitiveStr::Ascii("fastr"), &["faster"]),
            (dictgen::InsensitiveStr::Ascii("podemso"), &["podemos"]),
            (
                dictgen::InsensitiveStr::Ascii("oversimplificaiton"),
                &["oversimplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefenrence"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranposes"), &["transposes"]),
            (dictgen::InsensitiveStr::Ascii("freedomest"), &["freedoms"]),
            (
                dictgen::InsensitiveStr::Ascii("extractin"),
                &["extraction", "extracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peported"),
                &["reported", "purported"],
            ),
            (dictgen::InsensitiveStr::Ascii("damamge"), &["damage"]),
            (dictgen::InsensitiveStr::Ascii("amgles"), &["angles"]),
            (dictgen::InsensitiveStr::Ascii("containa"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("rurrent"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("proviving"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphorics"),
                &["metaphors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intoxinated"),
                &["intoxicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optmization"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinemetography"),
                &["cinematography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constroller"),
                &["controller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unoficcial"),
                &["unofficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("chechers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamant"),
                &["performant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annotataions"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraeter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("improting"), &["importing"]),
            (dictgen::InsensitiveStr::Ascii("detct"), &["detect"]),
            (dictgen::InsensitiveStr::Ascii("empitness"), &["emptiness"]),
            (
                dictgen::InsensitiveStr::Ascii("endorce"),
                &["endorse", "enforce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influentual"),
                &["influential"],
            ),
            (dictgen::InsensitiveStr::Ascii("highights"), &["highlights"]),
            (dictgen::InsensitiveStr::Ascii("ellipisis"), &["ellipsis"]),
            (dictgen::InsensitiveStr::Ascii("secction"), &["section"]),
            (
                dictgen::InsensitiveStr::Ascii("decalarative"),
                &["declarative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoziated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechanizim"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("noramalizes"),
                &["normalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miticating"),
                &["mitigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("jaring"), &["jarring"]),
            (dictgen::InsensitiveStr::Ascii("srollbar"), &["scrollbar"]),
            (
                dictgen::InsensitiveStr::Ascii("distinguised"),
                &["distinguished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visuall"),
                &["visual", "visually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conververs"),
                &["conservers", "converters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassemble"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stocahstic"),
                &["stochastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordinaly"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fathest"), &["farthest"]),
            (
                dictgen::InsensitiveStr::Ascii("untrustworty"),
                &["untrustworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("susceders"), &["seceders"]),
            (dictgen::InsensitiveStr::Ascii("presens"), &["presents"]),
            (dictgen::InsensitiveStr::Ascii("bouds"), &["bounds"]),
            (
                dictgen::InsensitiveStr::Ascii("speadsheet"),
                &["spreadsheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approporiate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masrhmallow"),
                &["marshmallow"],
            ),
            (dictgen::InsensitiveStr::Ascii("opereated"), &["operated"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetamene"),
                &["amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respectuflly"),
                &["respectfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("mircoatx"), &["microatx"]),
            (
                dictgen::InsensitiveStr::Ascii("unpredecented"),
                &["unprecedented"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsteam"), &["upstream"]),
            (dictgen::InsensitiveStr::Ascii("empries"), &["empires"]),
            (dictgen::InsensitiveStr::Ascii("prblem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("aplogies"), &["apologies"]),
            (dictgen::InsensitiveStr::Ascii("surronded"), &["surrounded"]),
            (dictgen::InsensitiveStr::Ascii("weappons"), &["weapons"]),
            (
                dictgen::InsensitiveStr::Ascii("estabilished"),
                &["established"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsiby"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjugagion"),
                &["conjugation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendundant"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("extractivo"),
                &["extraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("booleens"), &["booleans"]),
            (
                dictgen::InsensitiveStr::Ascii("simptomaticly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("negitiate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("minimumm"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("initiatiate"), &["initiate"]),
            (dictgen::InsensitiveStr::Ascii("exculde"), &["exclude"]),
            (
                dictgen::InsensitiveStr::Ascii("seachable"),
                &["searchable", "reachable", "teachable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supllemental"),
                &["supplemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrctions"),
                &["corrections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cateogrize"),
                &["categorized", "categorize"],
            ),
            (dictgen::InsensitiveStr::Ascii("trucate"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("retreival"), &["retrieval"]),
            (
                dictgen::InsensitiveStr::Ascii("incoropate"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possabilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specficications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combintaors"),
                &["combinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("werstle"), &["wrestle"]),
            (dictgen::InsensitiveStr::Ascii("collegate"), &["collegiate"]),
            (dictgen::InsensitiveStr::Ascii("wizzard"), &["wizard"]),
            (dictgen::InsensitiveStr::Ascii("chcked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("compensare"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistribution"),
                &["redistribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straightner"),
                &["straighten"],
            ),
            (dictgen::InsensitiveStr::Ascii("segements"), &["segments"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetenters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechaninc"), &["mechanic"]),
            (
                dictgen::InsensitiveStr::Ascii("silouhettes"),
                &["silhouettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("picthforks"),
                &["pitchforks"],
            ),
            (dictgen::InsensitiveStr::Ascii("coment"), &["comment"]),
            (dictgen::InsensitiveStr::Ascii("seperad"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("quaratine"), &["quarantine"]),
            (dictgen::InsensitiveStr::Ascii("jobb"), &["job"]),
            (
                dictgen::InsensitiveStr::Ascii("authentciation"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("prostite"), &["prostitute"]),
            (
                dictgen::InsensitiveStr::Ascii("functionailities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssigned"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("demolishon"),
                &["demolition"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrete"), &["discrete"]),
            (dictgen::InsensitiveStr::Ascii("widgits"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("arbitraly"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanetal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counteratk"),
                &["counteract"],
            ),
            (dictgen::InsensitiveStr::Ascii("wawe"), &["wave"]),
            (dictgen::InsensitiveStr::Ascii("mhytical"), &["mythical"]),
            (dictgen::InsensitiveStr::Ascii("crutially"), &["crucially"]),
            (dictgen::InsensitiveStr::Ascii("lables"), &["labels"]),
            (dictgen::InsensitiveStr::Ascii("fortat"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotionated"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("boudry"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("markerplaces"),
                &["marketplaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantitites"),
                &["quantities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proceesors"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("champiosnhip"),
                &["championships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminiscient"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("usccess"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("assoziations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tonuges"), &["tongues"]),
            (dictgen::InsensitiveStr::Ascii("corolloa"), &["corolla"]),
            (
                dictgen::InsensitiveStr::Ascii("refector"),
                &["refactor", "reflector"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difficultes"),
                &["difficulties"],
            ),
            (dictgen::InsensitiveStr::Ascii("replacong"), &["replacing"]),
            (
                dictgen::InsensitiveStr::Ascii("assasinates"),
                &["assassinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afficionados"),
                &["aficionados"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformarem"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminsced"), &["reminisced"]),
            (
                dictgen::InsensitiveStr::Ascii("neighburhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterests"), &["interests"]),
            (
                dictgen::InsensitiveStr::Ascii("depndencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheeckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("convenion"), &["convention"]),
            (
                dictgen::InsensitiveStr::Ascii("cymptumaticlly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("cellxs"), &["cells"]),
            (dictgen::InsensitiveStr::Ascii("gardai"), &["gardaí"]),
            (dictgen::InsensitiveStr::Ascii("assesed"), &["assessed"]),
            (dictgen::InsensitiveStr::Ascii("ccatch"), &["catch"]),
            (
                dictgen::InsensitiveStr::Ascii("parallellized"),
                &["parallelized"],
            ),
            (dictgen::InsensitiveStr::Ascii("seting"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("elemetal"), &["elemental"]),
            (dictgen::InsensitiveStr::Ascii("wimmens"), &["women"]),
            (dictgen::InsensitiveStr::Ascii("rekursive"), &["recursive"]),
            (
                dictgen::InsensitiveStr::Ascii("autorisation"),
                &["authorisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("juggernat"), &["juggernaut"]),
            (dictgen::InsensitiveStr::Ascii("infalte"), &["inflate"]),
            (dictgen::InsensitiveStr::Ascii("beginnig"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("derefenrecable"),
                &["dereferenceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("temrinals"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("condidtioning"),
                &["conditioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonized"),
                &["synchronized"],
            ),
            (dictgen::InsensitiveStr::Ascii("beavioral"), &["behavioral"]),
            (dictgen::InsensitiveStr::Ascii("variavble"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("prejucided"),
                &["prejudiced"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecions"), &["selections"]),
            (dictgen::InsensitiveStr::Ascii("eqaulity"), &["equality"]),
            (
                dictgen::InsensitiveStr::Ascii("backets"),
                &["baskets", "brackets", "buckets", "backers"],
            ),
            (dictgen::InsensitiveStr::Ascii("balsphemy"), &["blasphemy"]),
            (dictgen::InsensitiveStr::Ascii("mexicanas"), &["mexicans"]),
            (dictgen::InsensitiveStr::Ascii("notoroius"), &["notorious"]),
            (
                dictgen::InsensitiveStr::Ascii("vershining"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("presearver"), &["preserver"]),
            (dictgen::InsensitiveStr::Ascii("dureing"), &["during"]),
            (
                dictgen::InsensitiveStr::Ascii("approximatively"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("valriables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("picknics"), &["picnics"]),
            (
                dictgen::InsensitiveStr::Ascii("regulatorias"),
                &["regulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("disbales"), &["disables"]),
            (dictgen::InsensitiveStr::Ascii("setpember"), &["september"]),
            (
                dictgen::InsensitiveStr::Ascii("inteleaved"),
                &["interleaved"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("bitshifth"), &["bitshift"]),
            (dictgen::InsensitiveStr::Ascii("promtped"), &["prompted"]),
            (dictgen::InsensitiveStr::Ascii("tetsed"), &["tested"]),
            (dictgen::InsensitiveStr::Ascii("resurces"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("guresome"), &["gruesome"]),
            (dictgen::InsensitiveStr::Ascii("trucation"), &["truncation"]),
            (
                dictgen::InsensitiveStr::Ascii("doccumented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparetively"),
                &["comparatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("accupied"), &["occupied"]),
            (
                dictgen::InsensitiveStr::Ascii("furstrates"),
                &["frustrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("singpaore"), &["singapore"]),
            (
                dictgen::InsensitiveStr::Ascii("altetnatively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("probleems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplentation"),
                &["reimplementation", "reimplantation"],
            ),
            (dictgen::InsensitiveStr::Ascii("allcoate"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("perpatrators"),
                &["perpetrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execultables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissble"),
                &["permissible"],
            ),
            (dictgen::InsensitiveStr::Ascii("epxect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("ressource"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("mounths"),
                &["months", "mouths"],
            ),
            (dictgen::InsensitiveStr::Ascii("dilligence"), &["diligence"]),
            (
                dictgen::InsensitiveStr::Ascii("bulders"),
                &["builders", "boulders"],
            ),
            (dictgen::InsensitiveStr::Ascii("freqiency"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("guareded"), &["guarded"]),
            (dictgen::InsensitiveStr::Ascii("linez"), &["lines"]),
            (dictgen::InsensitiveStr::Ascii("execued"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("receve"), &["receive"]),
            (
                dictgen::InsensitiveStr::Ascii("trasversals"),
                &["traversals", "transversals"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewieving"), &["reviewing"]),
            (
                dictgen::InsensitiveStr::Ascii("requriement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coodinates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("comitter"), &["committer"]),
            (
                dictgen::InsensitiveStr::Ascii("amerliorates"),
                &["ameliorates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porjectiles"),
                &["projectiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("religon"), &["religion"]),
            (
                dictgen::InsensitiveStr::Ascii("traingulated"),
                &["triangulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("montaj"), &["montage"]),
            (dictgen::InsensitiveStr::Ascii("sertainty"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("enduce"), &["induce"]),
            (
                dictgen::InsensitiveStr::Ascii("intermittant"),
                &["intermittent"],
            ),
            (dictgen::InsensitiveStr::Ascii("spesifics"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("ftrunacate"), &["ftruncate"]),
            (
                dictgen::InsensitiveStr::Ascii("transconder"),
                &["transcoder"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcoker"), &["docker"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaiators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("arhive"), &["archive"]),
            (
                dictgen::InsensitiveStr::Ascii("boaut"),
                &["bout", "boat", "about"],
            ),
            (dictgen::InsensitiveStr::Ascii("errorneus"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("sesceder"), &["seceder"]),
            (
                dictgen::InsensitiveStr::Ascii("interesecting"),
                &["intersecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenaters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("naibhors"),
                &["neighbor", "neighbors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ganerative"),
                &["generative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("glorifierad"),
                &["glorified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplises"),
                &["accomplishes", "accomplices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalistiska"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("describibg"),
                &["describing"],
            ),
            (dictgen::InsensitiveStr::Ascii("manuells"), &["manuals"]),
            (
                dictgen::InsensitiveStr::Ascii("sorrounding"),
                &["surrounding"],
            ),
            (dictgen::InsensitiveStr::Ascii("hurriance"), &["hurricane"]),
            (
                dictgen::InsensitiveStr::Ascii("expirementation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("proflie"), &["profile"]),
            (dictgen::InsensitiveStr::Ascii("toahiba"), &["toshiba"]),
            (dictgen::InsensitiveStr::Ascii("soverign"), &["sovereign"]),
            (dictgen::InsensitiveStr::Ascii("emtpies"), &["empties"]),
            (
                dictgen::InsensitiveStr::Ascii("presedents"),
                &["presidents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omnishiences"),
                &["omnisciences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complecated"),
                &["complicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("usedfule"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("terimnates"),
                &["terminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("bidning"), &["binding"]),
            (dictgen::InsensitiveStr::Ascii("deverlops"), &["develops"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbboring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("secondy"),
                &["secondly", "secondary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminaci"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloccks"), &["blocks"]),
            (dictgen::InsensitiveStr::Ascii("shaprly"), &["sharply"]),
            (
                dictgen::InsensitiveStr::Ascii("bestaility"),
                &["bestiality"],
            ),
            (dictgen::InsensitiveStr::Ascii("otpional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregrating"),
                &["aggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defininition"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amplifliers"),
                &["amplifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("variavles"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("resuls"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("completelys"),
                &["completes"],
            ),
            (dictgen::InsensitiveStr::Ascii("medeival"), &["medieval"]),
            (dictgen::InsensitiveStr::Ascii("provoded"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("monotirs"), &["monitors"]),
            (dictgen::InsensitiveStr::Ascii("accompony"), &["accompany"]),
            (dictgen::InsensitiveStr::Ascii("tthe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("inpust"),
                &["input", "inputs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrtibution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("relesing"), &["releasing"]),
            (dictgen::InsensitiveStr::Ascii("convenit"), &["convenient"]),
            (
                dictgen::InsensitiveStr::Ascii("domensions"),
                &["dimensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythesizing"),
                &["synthesizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("incidate"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("onws"), &["owns"]),
            (
                dictgen::InsensitiveStr::Ascii("builer"),
                &["builder", "boiler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriecated"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coneectivity"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbossity"), &["verbosity"]),
            (
                dictgen::InsensitiveStr::Ascii("strust"),
                &["strut", "trust"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoslving"), &["resolving"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhhod"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asessments"),
                &["assessments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hiting"),
                &[
                    "hitting", "hiding", "hinting", "hiring", "hating", "biting", "hiking",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randazyvooed"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paricipates"),
                &["participates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iresistably"),
                &["irresistibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("huld"), &["held", "hold"]),
            (dictgen::InsensitiveStr::Ascii("obation"), &["ovation"]),
            (dictgen::InsensitiveStr::Ascii("frotn"), &["front"]),
            (
                dictgen::InsensitiveStr::Ascii("cooperateive"),
                &["cooperative"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcely"), &["forcibly"]),
            (
                dictgen::InsensitiveStr::Ascii("incoherance"),
                &["incoherence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ephermeral"), &["ephemeral"]),
            (
                dictgen::InsensitiveStr::Ascii("secuencial"),
                &["sequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("campagining"),
                &["campaigning"],
            ),
            (dictgen::InsensitiveStr::Ascii("expnese"), &["expense"]),
            (
                dictgen::InsensitiveStr::Ascii("accepecting"),
                &["accepting"],
            ),
            (dictgen::InsensitiveStr::Ascii("clarskon"), &["clarkson"]),
            (dictgen::InsensitiveStr::Ascii("maschines"), &["machines"]),
            (dictgen::InsensitiveStr::Ascii("defiently"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("boycot"), &["boycott"]),
            (
                dictgen::InsensitiveStr::Ascii("specalised"),
                &["specialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("serealize"), &["serialize"]),
            (dictgen::InsensitiveStr::Ascii("terimnals"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("comparitve"),
                &["comparative"],
            ),
            (dictgen::InsensitiveStr::Ascii("vawdville"), &["vaudeville"]),
            (dictgen::InsensitiveStr::Ascii("deubug"), &["debug"]),
            (dictgen::InsensitiveStr::Ascii("gameply"), &["gameplay"]),
            (dictgen::InsensitiveStr::Ascii("machintosh"), &["macintosh"]),
            (dictgen::InsensitiveStr::Ascii("wholey"), &["wholly"]),
            (dictgen::InsensitiveStr::Ascii("icefrong"), &["icefrog"]),
            (dictgen::InsensitiveStr::Ascii("everwhere"), &["everywhere"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcuations"),
                &["recalculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("jailbrek"), &["jailbreak"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapering"),
                &["disappearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("parshally"), &["partially"]),
            (dictgen::InsensitiveStr::Ascii("authro"), &["author"]),
            (
                dictgen::InsensitiveStr::Ascii("sarinomies"),
                &["ceremonies"],
            ),
            (dictgen::InsensitiveStr::Ascii("loenard"), &["leonard"]),
            (dictgen::InsensitiveStr::Ascii("actie"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegoptionsotiations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("elamentry"), &["elementary"]),
            (dictgen::InsensitiveStr::Ascii("almaost"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("remenissed"),
                &["reminisced"],
            ),
            (dictgen::InsensitiveStr::Ascii("overpaied"), &["overpaid"]),
            (
                dictgen::InsensitiveStr::Ascii("misrepresentativ"),
                &["misrepresentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwer"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("beneits"), &["benefits"]),
            (
                dictgen::InsensitiveStr::Ascii("electricial"),
                &["electrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accordignly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("vermillion"), &["vermilion"]),
            (dictgen::InsensitiveStr::Ascii("schedualer"), &["scheduler"]),
            (dictgen::InsensitiveStr::Ascii("thist"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("formtated"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("desactivations"),
                &["deactivations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefcifieis"),
                &["specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superfluos"),
                &["superfluous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializates"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("aapply"), &["apply"]),
            (
                dictgen::InsensitiveStr::Ascii("suppporting"),
                &["supporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repersenting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("manhatten"), &["manhattan"]),
            (dictgen::InsensitiveStr::Ascii("sufixed"), &["suffixed"]),
            (dictgen::InsensitiveStr::Ascii("complte"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("riminding"), &["reminding"]),
            (
                dictgen::InsensitiveStr::Ascii("espect"),
                &["expect", "aspect", "respect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repesented"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorporatin"),
                &["incorporating", "incorporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omnisiences"),
                &["omnisciences"],
            ),
            (dictgen::InsensitiveStr::Ascii("somewho"), &["somehow"]),
            (dictgen::InsensitiveStr::Ascii("combain"), &["combine"]),
            (
                dictgen::InsensitiveStr::Ascii("compromisse"),
                &["compromises"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensably"), &["sensibly"]),
            (dictgen::InsensitiveStr::Ascii("fomulated"), &["formulated"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapperd"),
                &["disappeared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarassign"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("persets"), &["presets"]),
            (dictgen::InsensitiveStr::Ascii("belys"), &["belies"]),
            (dictgen::InsensitiveStr::Ascii("maka"), &["make"]),
            (
                dictgen::InsensitiveStr::Ascii("tupes"),
                &["tuples", "types"],
            ),
            (dictgen::InsensitiveStr::Ascii("traiges"), &["triages"]),
            (
                dictgen::InsensitiveStr::Ascii("formattgin"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("diresired"), &["desired"]),
            (dictgen::InsensitiveStr::Ascii("colomba"), &["colombia"]),
            (dictgen::InsensitiveStr::Ascii("acecess"), &["access"]),
            (
                dictgen::InsensitiveStr::Ascii("terriories"),
                &["territories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multuplication"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positon"),
                &["position", "positron"],
            ),
            (dictgen::InsensitiveStr::Ascii("pentagoon"), &["pentagon"]),
            (dictgen::InsensitiveStr::Ascii("leasy"), &["least"]),
            (
                dictgen::InsensitiveStr::Ascii("calcalating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spesifities"),
                &["specifics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriopriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterising"),
                &["parameterising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partiticipant"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarnaton"),
                &["incarnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indaequate"),
                &["inadequate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supremasist"),
                &["supremacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritive"),
                &["descriptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("coexsited"), &["coexisted"]),
            (
                dictgen::InsensitiveStr::Ascii("opeartions"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalistiske"),
                &["nationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("literallly"), &["literally"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficillied"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("festivalens"),
                &["festivals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prominenty"),
                &["prominently"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahgner"), &["changer"]),
            (dictgen::InsensitiveStr::Ascii("correcly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("abtractor"), &["abstractor"]),
            (dictgen::InsensitiveStr::Ascii("outisders"), &["outsiders"]),
            (
                dictgen::InsensitiveStr::Ascii("succeessfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibile"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessfukk"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("geenric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("leaave"), &["leave"]),
            (
                dictgen::InsensitiveStr::Ascii("deselctables"),
                &["deselectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcaulted"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dellocation"),
                &["deallocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("littlefinder"),
                &["littlefinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posions"),
                &["poisons", "positions", "psions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutionnels"),
                &["institutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("bve"), &["be"]),
            (dictgen::InsensitiveStr::Ascii("timzeones"), &["timezones"]),
            (
                dictgen::InsensitiveStr::Ascii("likleihood"),
                &["likelihood"],
            ),
            (dictgen::InsensitiveStr::Ascii("pecentile"), &["percentile"]),
            (
                dictgen::InsensitiveStr::Ascii("innoculates"),
                &["inoculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appoaching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimintation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discourrage"),
                &["discourage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensinos"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chernoybl"), &["chernobyl"]),
            (
                dictgen::InsensitiveStr::Ascii("generalnie"),
                &["generalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thicknessess"),
                &["thicknesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("arthrits"), &["arthritis"]),
            (
                dictgen::InsensitiveStr::Ascii("suggesstion"),
                &["suggestion"],
            ),
            (dictgen::InsensitiveStr::Ascii("diskrete"), &["discrete"]),
            (dictgen::InsensitiveStr::Ascii("parsering"), &["parsing"]),
            (
                dictgen::InsensitiveStr::Ascii("punsihable"),
                &["punishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesss"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("strcture"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("scheudle"), &["schedule"]),
            (dictgen::InsensitiveStr::Ascii("warninng"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("gogin"),
                &["going", "gauguin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsition"),
                &["transition"],
            ),
            (dictgen::InsensitiveStr::Ascii("buttins"), &["buttons"]),
            (dictgen::InsensitiveStr::Ascii("reomove"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("venezulea"), &["venezuela"]),
            (
                dictgen::InsensitiveStr::Ascii("imprementing"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dolhpin"), &["dolphin"]),
            (
                dictgen::InsensitiveStr::Ascii("multipicities"),
                &["multiplicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("punsihes"), &["punishes"]),
            (
                dictgen::InsensitiveStr::Ascii("reampping"),
                &["remapping", "revamping"],
            ),
            (dictgen::InsensitiveStr::Ascii("supportied"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("datattypes"), &["datatypes"]),
            (dictgen::InsensitiveStr::Ascii("kelogger"), &["keylogger"]),
            (
                dictgen::InsensitiveStr::Ascii("intercahnged"),
                &["interchanged"],
            ),
            (dictgen::InsensitiveStr::Ascii("algining"), &["alining"]),
            (
                dictgen::InsensitiveStr::Ascii("extrems"),
                &["extremes", "extrema"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranition"), &["transition"]),
            (
                dictgen::InsensitiveStr::Ascii("transitiong"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("octive"),
                &["active", "octave"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitley"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("reccords"), &["records"]),
            (dictgen::InsensitiveStr::Ascii("stess"), &["stress"]),
            (dictgen::InsensitiveStr::Ascii("zepplien"), &["zeppelin"]),
            (
                dictgen::InsensitiveStr::Ascii("patnership"),
                &["partnership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("domonstrating"),
                &["demonstrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fowrards"), &["forwards"]),
            (dictgen::InsensitiveStr::Ascii("valdiated"), &["validated"]),
            (
                dictgen::InsensitiveStr::Ascii("wath"),
                &["watch", "wrath", "what"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintinence"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("intreeged"), &["intrigued"]),
            (dictgen::InsensitiveStr::Ascii("govermnet"), &["government"]),
            (dictgen::InsensitiveStr::Ascii("fianlises"), &["finalises"]),
            (
                dictgen::InsensitiveStr::Ascii("intercative"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("suasages"), &["sausages"]),
            (
                dictgen::InsensitiveStr::Ascii("configureed"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrepreted"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressiong"),
                &["progressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("baisc"), &["basic"]),
            (dictgen::InsensitiveStr::Ascii("verion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("feautred"), &["featured"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondning"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("platfromer"),
                &["platformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("remplacing"), &["replacing"]),
            (
                dictgen::InsensitiveStr::Ascii("enthousiastically"),
                &["enthusiastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randomizin"),
                &["randomizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("titanuim"), &["titanium"]),
            (dictgen::InsensitiveStr::Ascii("intrisic"), &["intrinsic"]),
            (dictgen::InsensitiveStr::Ascii("initilize"), &["initialize"]),
            (dictgen::InsensitiveStr::Ascii("threshods"), &["thresholds"]),
            (
                dictgen::InsensitiveStr::Ascii("supposted"),
                &["supposed", "supported"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouunded"), &["bounded"]),
            (dictgen::InsensitiveStr::Ascii("buids"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("aquire"), &["acquire"]),
            (
                dictgen::InsensitiveStr::Ascii("restoreable"),
                &["restorable"],
            ),
            (dictgen::InsensitiveStr::Ascii("repsawn"), &["respawn"]),
            (dictgen::InsensitiveStr::Ascii("stduent"), &["student"]),
            (
                dictgen::InsensitiveStr::Ascii("allianse"),
                &["alliances", "alliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualizaton"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("destkops"), &["desktops"]),
            (
                dictgen::InsensitiveStr::Ascii("indentatoin"),
                &["indentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("austira"), &["austria"]),
            (
                dictgen::InsensitiveStr::Ascii("restuctured"),
                &["restructured"],
            ),
            (dictgen::InsensitiveStr::Ascii("multilne"), &["multiline"]),
            (
                dictgen::InsensitiveStr::Ascii("concures"),
                &["concurs", "conquers"],
            ),
            (dictgen::InsensitiveStr::Ascii("spectum"), &["spectrum"]),
            (dictgen::InsensitiveStr::Ascii("wehrever"), &["wherever"]),
            (dictgen::InsensitiveStr::Ascii("subtitiles"), &["subtitles"]),
            (
                dictgen::InsensitiveStr::Ascii("nercomancer"),
                &["necromancer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuploades"),
                &["reuploads", "reuploaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnfigure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("photograher"),
                &["photographer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assicative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("porfilers"), &["profilers"]),
            (dictgen::InsensitiveStr::Ascii("mecernary"), &["mercenary"]),
            (
                dictgen::InsensitiveStr::Ascii("transtitioning"),
                &["transitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("generrally"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("assiociations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sasauges"), &["sausages"]),
            (dictgen::InsensitiveStr::Ascii("occaision"), &["occasion"]),
            (
                dictgen::InsensitiveStr::Ascii("capitolizes"),
                &["capitalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("langeuagesection"),
                &["languagesection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commemmorating"),
                &["commemorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vandalisim"), &["vandalism"]),
            (dictgen::InsensitiveStr::Ascii("execising"), &["exercising"]),
            (
                dictgen::InsensitiveStr::Ascii("disconenct"),
                &["disconnect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercatively"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referncial"),
                &["referential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prferables"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peizometers"),
                &["piezometers"],
            ),
            (dictgen::InsensitiveStr::Ascii("garentors"), &["guarantors"]),
            (dictgen::InsensitiveStr::Ascii("disscusses"), &["discusses"]),
            (
                dictgen::InsensitiveStr::Ascii("persent"),
                &["present", "percent", "persistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unwarrented"),
                &["unwarranted"],
            ),
            (dictgen::InsensitiveStr::Ascii("usefult"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("statemens"), &["statements"]),
            (
                dictgen::InsensitiveStr::Ascii("dependdences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("anywya"), &["anyway"]),
            (dictgen::InsensitiveStr::Ascii("resuorce"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("unxpected"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("commisssion"),
                &["commissions", "commission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colloection"),
                &["collection"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfur"), &["prefer"]),
            (dictgen::InsensitiveStr::Ascii("etherael"), &["ethereal"]),
            (
                dictgen::InsensitiveStr::Ascii("annotaiotns"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mairadb"), &["mariadb"]),
            (dictgen::InsensitiveStr::Ascii("synatax"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("aprove"), &["approve"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliazes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("typin"), &["typing"]),
            (
                dictgen::InsensitiveStr::Ascii("rearranements"),
                &["rearrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarchits"), &["anarchists"]),
            (dictgen::InsensitiveStr::Ascii("callbacsk"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("sleected"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("notacible"), &["noticeable"]),
            (dictgen::InsensitiveStr::Ascii("messaih"), &["messiah"]),
            (dictgen::InsensitiveStr::Ascii("quantitty"), &["quantity"]),
            (
                dictgen::InsensitiveStr::Ascii("councelors"),
                &["counselors"],
            ),
            (dictgen::InsensitiveStr::Ascii("cluste"), &["cluster"]),
            (
                dictgen::InsensitiveStr::Ascii("interrtuped"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("rapayment"), &["repayment"]),
            (dictgen::InsensitiveStr::Ascii("colect"), &["collect"]),
            (
                dictgen::InsensitiveStr::Ascii("documentstion"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenaticated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternitive"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("imagin"), &["imagine"]),
            (dictgen::InsensitiveStr::Ascii("pretiest"), &["prettiest"]),
            (dictgen::InsensitiveStr::Ascii("begginner"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("interfer"), &["interfere"]),
            (
                dictgen::InsensitiveStr::Ascii("paticipating"),
                &["participating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerature"), &["creature"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithimic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependeces"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("indsutry"), &["industry"]),
            (dictgen::InsensitiveStr::Ascii("prepaird"), &["prepaid"]),
            (dictgen::InsensitiveStr::Ascii("agrs"), &["args"]),
            (dictgen::InsensitiveStr::Ascii("menaings"), &["meanings"]),
            (
                dictgen::InsensitiveStr::Ascii("importend"),
                &["imported", "important"],
            ),
            (dictgen::InsensitiveStr::Ascii("embbedded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("acchiever"), &["achiever"]),
            (dictgen::InsensitiveStr::Ascii("euphoriac"), &["euphoric"]),
            (
                dictgen::InsensitiveStr::Ascii("relatdness"),
                &["relatedness"],
            ),
            (dictgen::InsensitiveStr::Ascii("mairlyn"), &["marilyn"]),
            (dictgen::InsensitiveStr::Ascii("dosposing"), &["disposing"]),
            (
                dictgen::InsensitiveStr::Ascii("assertinos"),
                &["assertions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pxoxy"), &["proxy"]),
            (dictgen::InsensitiveStr::Ascii("parmeters"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("insterested"),
                &["interested"],
            ),
            (dictgen::InsensitiveStr::Ascii("sporuts"), &["sprouts"]),
            (dictgen::InsensitiveStr::Ascii("bitocin"), &["bitcoin"]),
            (
                dictgen::InsensitiveStr::Ascii("intenet"),
                &["internet", "intent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandianvia"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costumizations"),
                &["customization", "customizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytax"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("construktor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undoubedtly"),
                &["undoubtedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeditively"),
                &["competitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("expresse"), &["expressive"]),
            (
                dictgen::InsensitiveStr::Ascii("checerboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("usefutl"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("mailboxe"), &["mailbox"]),
            (dictgen::InsensitiveStr::Ascii("barsita"), &["barista"]),
            (
                dictgen::InsensitiveStr::Ascii("organisera"),
                &["organisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sreenshots"),
                &["screenshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wnidow"),
                &["window", "widow"],
            ),
            (dictgen::InsensitiveStr::Ascii("concenrs"), &["concerns"]),
            (dictgen::InsensitiveStr::Ascii("dyin"), &["dying"]),
            (dictgen::InsensitiveStr::Ascii("procceed"), &["proceed"]),
            (
                dictgen::InsensitiveStr::Ascii("delimiation"),
                &["delimitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("frameworkk"), &["framework"]),
            (
                dictgen::InsensitiveStr::Ascii("modificatioons"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("warrios"), &["warriors"]),
            (dictgen::InsensitiveStr::Ascii("abizmal"), &["abysmal"]),
            (
                dictgen::InsensitiveStr::Ascii("encompases"),
                &["encompasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inintialize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyzie"), &["analyze"]),
            (dictgen::InsensitiveStr::Ascii("assighee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("aneroxia"), &["anorexia"]),
            (dictgen::InsensitiveStr::Ascii("vitroil"), &["vitriol"]),
            (dictgen::InsensitiveStr::Ascii("expanisve"), &["expansive"]),
            (dictgen::InsensitiveStr::Ascii("collapsin"), &["collapsing"]),
            (
                dictgen::InsensitiveStr::Ascii("recconeted"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punihsment"),
                &["punishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregisteres"),
                &["unregisters", "unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expliotation"),
                &["exploitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juridisction"),
                &["jurisdictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perputrator"),
                &["perpetrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persucution"),
                &["persecution"],
            ),
            (dictgen::InsensitiveStr::Ascii("captacitor"), &["capacitor"]),
            (dictgen::InsensitiveStr::Ascii("hamitlon"), &["hamilton"]),
            (
                dictgen::InsensitiveStr::Ascii("extremeophile"),
                &["extremophile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoluttionary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("baceause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("szene"), &["scene"]),
            (
                dictgen::InsensitiveStr::Ascii("implemanted"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analoguously"),
                &["analogously"],
            ),
            (dictgen::InsensitiveStr::Ascii("fornated"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("amlpifier"), &["amplifier"]),
            (dictgen::InsensitiveStr::Ascii("opeatror"), &["operator"]),
            (
                dictgen::InsensitiveStr::Ascii("organiztion"),
                &["organization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inidications"),
                &["indications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complaines"),
                &["complaints"],
            ),
            (dictgen::InsensitiveStr::Ascii("behvaiour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("coninuity"), &["continuity"]),
            (
                dictgen::InsensitiveStr::Ascii("interractial"),
                &["interracial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electrolyes"),
                &["electrolytes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inventings"),
                &["inventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nacionalists"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definateley"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("exersise"), &["exercise"]),
            (
                dictgen::InsensitiveStr::Ascii("referecnes"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrappered"), &["wrapped"]),
            (
                dictgen::InsensitiveStr::Ascii("resistane"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("sautay"), &["sauté"]),
            (dictgen::InsensitiveStr::Ascii("suceeded"), &["succeeded"]),
            (dictgen::InsensitiveStr::Ascii("enrypted"), &["encrypted"]),
            (dictgen::InsensitiveStr::Ascii("accptance"), &["acceptance"]),
            (
                dictgen::InsensitiveStr::Ascii("renegothiator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliboards"), &["clipboards"]),
            (dictgen::InsensitiveStr::Ascii("formar"), &["former"]),
            (
                dictgen::InsensitiveStr::Ascii("repsonding"),
                &["responding"],
            ),
            (dictgen::InsensitiveStr::Ascii("privoded"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("appriciates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("unnistall"), &["uninstall"]),
            (
                dictgen::InsensitiveStr::Ascii("interactie"),
                &["interactive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrativo"),
                &["demonstration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalistisch"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developmet"),
                &["developments", "development"],
            ),
            (dictgen::InsensitiveStr::Ascii("pursuaded"), &["persuaded"]),
            (dictgen::InsensitiveStr::Ascii("brokeness"), &["brokenness"]),
            (dictgen::InsensitiveStr::Ascii("practhett"), &["pratchett"]),
            (dictgen::InsensitiveStr::Ascii("vunerable"), &["vulnerable"]),
            (
                dictgen::InsensitiveStr::Ascii("cerinomies"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filles"),
                &["files", "fills", "filled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplacated"),
                &["deduplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sizors"),
                &["scissors", "sizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentements"),
                &["sentiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trianges"),
                &["triangles", "triages"],
            ),
            (dictgen::InsensitiveStr::Ascii("anythign"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("intilising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfforms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncicalized"),
                &["canonicalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cellers"), &["cellars"]),
            (dictgen::InsensitiveStr::Ascii("speified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("pathalogy"), &["pathology"]),
            (
                dictgen::InsensitiveStr::Ascii("mothebroard"),
                &["motherboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchanise"),
                &["merchandise"],
            ),
            (dictgen::InsensitiveStr::Ascii("handlin"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("calulating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("historicaly"),
                &["historically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninitiliazed"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaberator"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapats"), &["adapts"]),
            (
                dictgen::InsensitiveStr::Ascii("projeciton"),
                &["projection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntitutive"),
                &["constitutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ascneded"), &["ascended"]),
            (dictgen::InsensitiveStr::Ascii("ussuall"), &["usual"]),
            (dictgen::InsensitiveStr::Ascii("boound"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("bunldes"), &["bundles"]),
            (
                dictgen::InsensitiveStr::Ascii("acitvation"),
                &["activation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regradless"),
                &["regardless"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceedes"), &["proceeds"]),
            (
                dictgen::InsensitiveStr::Ascii("capaciance"),
                &["capacitance"],
            ),
            (dictgen::InsensitiveStr::Ascii("curev"), &["curve"]),
            (dictgen::InsensitiveStr::Ascii("aquitting"), &["acquitting"]),
            (dictgen::InsensitiveStr::Ascii("opitmal"), &["optimal"]),
            (
                dictgen::InsensitiveStr::Ascii("deferentiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expectional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("reloction"), &["relocation"]),
            (
                dictgen::InsensitiveStr::Ascii("delegatin"),
                &["delegating", "delegation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigals"),
                &["signals", "sigils"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitibality"),
                &["profitability"],
            ),
            (dictgen::InsensitiveStr::Ascii("squarin"), &["squaring"]),
            (dictgen::InsensitiveStr::Ascii("acadmy"), &["academy"]),
            (dictgen::InsensitiveStr::Ascii("inportant"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("execusions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phsyiology"),
                &["physiology"],
            ),
            (dictgen::InsensitiveStr::Ascii("displayfps"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("entitiy"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("mechamism"), &["mechanism"]),
            (dictgen::InsensitiveStr::Ascii("mebers"), &["members"]),
            (dictgen::InsensitiveStr::Ascii("commmands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("governer"), &["governor"]),
            (
                dictgen::InsensitiveStr::Ascii("executivas"),
                &["executives"],
            ),
            (dictgen::InsensitiveStr::Ascii("embeddder"), &["embedder"]),
            (dictgen::InsensitiveStr::Ascii("shoild"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("discreprancy"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("franzise"), &["franchise"]),
            (dictgen::InsensitiveStr::Ascii("instuctor"), &["instructor"]),
            (dictgen::InsensitiveStr::Ascii("decmials"), &["decimals"]),
            (dictgen::InsensitiveStr::Ascii("clumn"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("muiltiples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("wiyth"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("assoiation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cerificates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("propagana"), &["propaganda"]),
            (
                dictgen::InsensitiveStr::Ascii("remenissently"),
                &["reminiscently"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheterh"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("miliseconds"),
                &["milliseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disturping"),
                &["disrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("romoved"), &["removed"]),
            (dictgen::InsensitiveStr::Ascii("secratery"), &["secretary"]),
            (
                dictgen::InsensitiveStr::Ascii("cumulatative"),
                &["cumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sargeant"), &["sergeant"]),
            (dictgen::InsensitiveStr::Ascii("updte"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("duplaicates"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("pamflet"), &["pamphlet"]),
            (
                dictgen::InsensitiveStr::Ascii("victemizing"),
                &["victimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("opern"), &["open"]),
            (dictgen::InsensitiveStr::Ascii("versined"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("monumnet"), &["monument"]),
            (dictgen::InsensitiveStr::Ascii("mastters"), &["masters"]),
            (dictgen::InsensitiveStr::Ascii("nativ"), &["native"]),
            (
                dictgen::InsensitiveStr::Ascii("occcurrences"),
                &["occurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepdendetly"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramendously"),
                &["tremendously"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytsemic"), &["systemic"]),
            (
                dictgen::InsensitiveStr::Ascii("survavibility"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcational"),
                &["transactional", "translational", "transnational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalizaton"),
                &["capitalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosumable"), &["consumable"]),
            (
                dictgen::InsensitiveStr::Ascii("resorusion"),
                &["resolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("lene"), &["lens"]),
            (dictgen::InsensitiveStr::Ascii("defautl"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("confrontaion"),
                &["confrontation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychopants"),
                &["psychopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("euristics"), &["heuristics"]),
            (
                dictgen::InsensitiveStr::Ascii("impeccabile"),
                &["impeccable"],
            ),
            (dictgen::InsensitiveStr::Ascii("classfy"), &["classify"]),
            (dictgen::InsensitiveStr::Ascii("deferreal"), &["deferral"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminami"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noticications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("asignor"), &["assignor"]),
            (dictgen::InsensitiveStr::Ascii("visiblle"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("integrat"),
                &["integral", "integrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cathegorised"),
                &["categorised"],
            ),
            (dictgen::InsensitiveStr::Ascii("moteefs"), &["motifs"]),
            (
                dictgen::InsensitiveStr::Ascii("advertisiments"),
                &["advertisements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vritualisation"),
                &["virtualisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resizeble"),
                &["resizable", "resizeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investogator"),
                &["investigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invaldiate"),
                &["invalidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("permium"), &["premium"]),
            (dictgen::InsensitiveStr::Ascii("guarateed"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("analitic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("pecifically"),
                &["specifically", "pacifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palid"),
                &["plaid", "pallid"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoising"),
                &["authorising"],
            ),
            (dictgen::InsensitiveStr::Ascii("wihich"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("istantiate"),
                &["instantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("oututs"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("handker"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("expostes"), &["exposes"]),
            (
                dictgen::InsensitiveStr::Ascii("progession"),
                &["progression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represetning"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqiure"), &["require"]),
            (dictgen::InsensitiveStr::Ascii("plesae"), &["please"]),
            (dictgen::InsensitiveStr::Ascii("reoslves"), &["resolves"]),
            (dictgen::InsensitiveStr::Ascii("restorin"), &["restoring"]),
            (dictgen::InsensitiveStr::Ascii("cnannel"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("itenary"), &["itinerary"]),
            (dictgen::InsensitiveStr::Ascii("chalenges"), &["challenges"]),
            (dictgen::InsensitiveStr::Ascii("comfirm"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("administraors"),
                &["administrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrriddden"),
                &["overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("leasat"), &["least"]),
            (dictgen::InsensitiveStr::Ascii("monstorus"), &["monstrous"]),
            (dictgen::InsensitiveStr::Ascii("siwtching"), &["switching"]),
            (
                dictgen::InsensitiveStr::Ascii("follfowiong"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpectedtly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("exced"), &["exceed"]),
            (dictgen::InsensitiveStr::Ascii("genearted"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("checkkout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("croped"), &["coped"]),
            (dictgen::InsensitiveStr::Ascii("reacording"), &["recording"]),
            (
                dictgen::InsensitiveStr::Ascii("ameelarators"),
                &["ameliorators"],
            ),
            (dictgen::InsensitiveStr::Ascii("committs"), &["commits"]),
            (dictgen::InsensitiveStr::Ascii("eqaution"), &["equation"]),
            (dictgen::InsensitiveStr::Ascii("purcahses"), &["purchases"]),
            (dictgen::InsensitiveStr::Ascii("dperecate"), &["deprecate"]),
            (dictgen::InsensitiveStr::Ascii("insterted"), &["inserted"]),
            (dictgen::InsensitiveStr::Ascii("examing"), &["examining"]),
            (dictgen::InsensitiveStr::Ascii("movememnts"), &["movements"]),
            (dictgen::InsensitiveStr::Ascii("recommad"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("oppononent"), &["opponent"]),
            (
                dictgen::InsensitiveStr::Ascii("convertation"),
                &["conversation", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("committis"), &["committees"]),
            (
                dictgen::InsensitiveStr::Ascii("declerations"),
                &["declarations", "decelerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("banches"),
                &["branches", "benches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consituents"),
                &["constituents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgrates"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatior"), &["operator"]),
            (
                dictgen::InsensitiveStr::Ascii("configiuring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoitations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutialtion"),
                &["mutilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sooit"),
                &["suet", "suit", "soot"],
            ),
            (dictgen::InsensitiveStr::Ascii("origanaly"), &["originally"]),
            (
                dictgen::InsensitiveStr::Ascii("exmplifying"),
                &["exemplifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diskretized"),
                &["discretized"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecilpse"), &["eclipse"]),
            (
                dictgen::InsensitiveStr::Ascii("sophistocated"),
                &["sophisticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("nediums"), &["mediums"]),
            (
                dictgen::InsensitiveStr::Ascii("inpossible"),
                &["impossible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquantainces"),
                &["acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("randayvou"), &["rendezvous"]),
            (dictgen::InsensitiveStr::Ascii("messure"), &["measure"]),
            (dictgen::InsensitiveStr::Ascii("duplacate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("referenct"), &["referenced"]),
            (
                dictgen::InsensitiveStr::Ascii("interpreteert"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denpendency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proposterous"),
                &["preposterous"],
            ),
            (dictgen::InsensitiveStr::Ascii("extremaly"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("buliders"), &["builders"]),
            (dictgen::InsensitiveStr::Ascii("sculputre"), &["sculpture"]),
            (dictgen::InsensitiveStr::Ascii("tempray"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("tyoed"),
                &["typed", "toyed", "toed"],
            ),
            (dictgen::InsensitiveStr::Ascii("genreal"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("anchord"), &["anchored"]),
            (dictgen::InsensitiveStr::Ascii("compilant"), &["compliant"]),
            (dictgen::InsensitiveStr::Ascii("germaniac"), &["germanic"]),
            (dictgen::InsensitiveStr::Ascii("booundary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("atatches"), &["attaches"]),
            (
                dictgen::InsensitiveStr::Ascii("alloccating"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agricuture"),
                &["agriculture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpetectedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcessist"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("ublishing"), &["publishing"]),
            (dictgen::InsensitiveStr::Ascii("synonymns"), &["synonyms"]),
            (dictgen::InsensitiveStr::Ascii("ceasser"), &["caesar"]),
            (dictgen::InsensitiveStr::Ascii("samaphore"), &["semaphore"]),
            (dictgen::InsensitiveStr::Ascii("hierarchie"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("corrcect"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("disrepsected"),
                &["disrespected"],
            ),
            (dictgen::InsensitiveStr::Ascii("seacrhers"), &["searchers"]),
            (
                dictgen::InsensitiveStr::Ascii("definitevely"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicatin"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completetly"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creadential"),
                &["credential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilavents"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instealling"),
                &["installing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assciating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("extrimely"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("exerciess"), &["exercises"]),
            (
                dictgen::InsensitiveStr::Ascii("emberrassing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbrasing"), &["embracing"]),
            (dictgen::InsensitiveStr::Ascii("cylclic"), &["cyclic"]),
            (dictgen::InsensitiveStr::Ascii("metrigs"), &["metrics"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregeting"),
                &["aggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepolates"),
                &["interpolates"],
            ),
            (dictgen::InsensitiveStr::Ascii("summenor"), &["summoner"]),
            (
                dictgen::InsensitiveStr::Ascii("hexidecmials"),
                &["hexadecimals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kubernests"),
                &["kubernetes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effortlessely"),
                &["effortlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnearabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seqeuntials"),
                &["sequentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparnecies"),
                &["transparencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("picures"),
                &["pictures", "epicures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("torrentting"),
                &["torrenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contreversial"),
                &["controversial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sprotsmanship"),
                &["sportsmanship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fial"),
                &["fail", "final", "dial", "vial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particuliar"),
                &["particular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactherized"),
                &["characterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("contrrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("identifeis"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapporves"),
                &["disapproves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poping"),
                &["popping", "pooping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("negiotate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticy"),
                &["authenticity"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyramide"), &["pyramid"]),
            (dictgen::InsensitiveStr::Ascii("aspecs"), &["aspects"]),
            (
                dictgen::InsensitiveStr::Ascii("paramterised"),
                &["parameterised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("participati"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("downlods"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("customzie"), &["customize"]),
            (dictgen::InsensitiveStr::Ascii("redesiging"), &["redesign"]),
            (dictgen::InsensitiveStr::Ascii("bouys"), &["buoys"]),
            (
                dictgen::InsensitiveStr::Ascii("architechturally"),
                &["architecturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neverthelss"),
                &["nevertheless"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressious"), &["precious"]),
            (
                dictgen::InsensitiveStr::Ascii("decentralizin"),
                &["decentralizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("comign"), &["coming"]),
            (
                dictgen::InsensitiveStr::Ascii("absolutiste"),
                &["absolutes", "absolutist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercations"),
                &["interactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("seriouslly"), &["seriously"]),
            (
                dictgen::InsensitiveStr::Ascii("otpimisation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplciate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("lcoales"), &["locales"]),
            (dictgen::InsensitiveStr::Ascii("valenca"), &["valencia"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalismo"),
                &["fundamentals"],
            ),
            (dictgen::InsensitiveStr::Ascii("usless"), &["useless"]),
            (dictgen::InsensitiveStr::Ascii("firmwqre"), &["firmware"]),
            (
                dictgen::InsensitiveStr::Ascii("hopeleslly"),
                &["hopelessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propbability"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convining"),
                &["combining", "convincing", "confining", "convening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catapiller"),
                &["caterpillar"],
            ),
            (dictgen::InsensitiveStr::Ascii("angostics"), &["agnostics"]),
            (dictgen::InsensitiveStr::Ascii("exhasuted"), &["exhausted"]),
            (dictgen::InsensitiveStr::Ascii("acient"), &["ancient"]),
            (
                dictgen::InsensitiveStr::Ascii("commparisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("symbolisim"), &["symbolism"]),
            (dictgen::InsensitiveStr::Ascii("disiring"), &["desiring"]),
            (
                dictgen::InsensitiveStr::Ascii("excetptions"),
                &["exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithimes"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knownledgeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("insturct"), &["instruct"]),
            (
                dictgen::InsensitiveStr::Ascii("iconclastic"),
                &["iconoclastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubiting"),
                &["distributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreators"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propogated"),
                &["propagated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaquial"),
                &["colloquial"],
            ),
            (dictgen::InsensitiveStr::Ascii("loosly"), &["loosely"]),
            (
                dictgen::InsensitiveStr::Ascii("convesration"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatoral"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchnge"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("abjects"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("enetity"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("happeing"), &["happening"]),
            (
                dictgen::InsensitiveStr::Ascii("traslations"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propgating"),
                &["propagating"],
            ),
            (dictgen::InsensitiveStr::Ascii("touritsy"), &["touristy"]),
            (
                dictgen::InsensitiveStr::Ascii("suppresssions"),
                &["suppressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccuarate"),
                &["inaccurate"],
            ),
            (dictgen::InsensitiveStr::Ascii("appearrs"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("threatments"),
                &["treatments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedophilea"),
                &["pedophilia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lement"),
                &["element", "clement", "lament"],
            ),
            (dictgen::InsensitiveStr::Ascii("secnario"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("requestested"),
                &["requested"],
            ),
            (dictgen::InsensitiveStr::Ascii("canceld"), &["canceled"]),
            (
                dictgen::InsensitiveStr::Ascii("scipt"),
                &["script", "skipped"],
            ),
            (dictgen::InsensitiveStr::Ascii("typles"), &["tuples"]),
            (dictgen::InsensitiveStr::Ascii("manuell"), &["manual"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptation"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosqueto"), &["mosquitoes"]),
            (
                dictgen::InsensitiveStr::Ascii("interchage"),
                &["interchange"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revelutionary"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investiaging"),
                &["investigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("familiies"), &["families"]),
            (dictgen::InsensitiveStr::Ascii("beleivers"), &["believers"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsitrative"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("bectors"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("rotaions"), &["rotations"]),
            (
                dictgen::InsensitiveStr::Ascii("cateogrizes"),
                &["categorizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deficienies"),
                &["deficiencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("virtualy"), &["virtually"]),
            (dictgen::InsensitiveStr::Ascii("norwegin"), &["norwegian"]),
            (
                dictgen::InsensitiveStr::Ascii("instututed"),
                &["instituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("tonihgt"), &["tonight"]),
            (
                dictgen::InsensitiveStr::Ascii("messagetqueue"),
                &["messagequeue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catholisim"),
                &["catholicism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controvesy"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("similari"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("unorderd"), &["unordered"]),
            (
                dictgen::InsensitiveStr::Ascii("subtitution"),
                &["substitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("governened"), &["governed"]),
            (dictgen::InsensitiveStr::Ascii("wisconsion"), &["wisconsin"]),
            (dictgen::InsensitiveStr::Ascii("statistc"), &["statistic"]),
            (dictgen::InsensitiveStr::Ascii("widges"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("descrived"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("exands"), &["expands"]),
            (
                dictgen::InsensitiveStr::Ascii("approximetion"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarkist"), &["anarchist"]),
            (
                dictgen::InsensitiveStr::Ascii("mutatations"),
                &["mutations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypocracy"), &["hypocrisy"]),
            (dictgen::InsensitiveStr::Ascii("compitent"), &["competent"]),
            (
                dictgen::InsensitiveStr::Ascii("approximition"),
                &["approximation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interwiever"),
                &["interviewer"],
            ),
            (dictgen::InsensitiveStr::Ascii("campanies"), &["companies"]),
            (dictgen::InsensitiveStr::Ascii("bounudary"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("soldeirs"), &["soldiers"]),
            (
                dictgen::InsensitiveStr::Ascii("visibibilty"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("disired"), &["desired"]),
            (dictgen::InsensitiveStr::Ascii("ublisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("progressionwise"),
                &["progressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("frsibee"), &["frisbee"]),
            (dictgen::InsensitiveStr::Ascii("guilia"), &["giulia"]),
            (dictgen::InsensitiveStr::Ascii("rehersals"), &["rehearsals"]),
            (dictgen::InsensitiveStr::Ascii("ampilfies"), &["amplifies"]),
            (dictgen::InsensitiveStr::Ascii("adavancing"), &["advancing"]),
            (dictgen::InsensitiveStr::Ascii("faktors"), &["factors"]),
            (dictgen::InsensitiveStr::Ascii("realative"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("millionarie"),
                &["millionaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conncection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bodyweigth"),
                &["bodyweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("throners"), &["thrones"]),
            (dictgen::InsensitiveStr::Ascii("coerceion"), &["coercion"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("posative"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("besiegin"), &["besieging"]),
            (dictgen::InsensitiveStr::Ascii("charmisa"), &["charisma"]),
            (
                dictgen::InsensitiveStr::Ascii("represenging"),
                &["representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antipenultimate"),
                &["antepenultimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shaprening"),
                &["sharpening"],
            ),
            (dictgen::InsensitiveStr::Ascii("librraies"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("programas"), &["programs"]),
            (
                dictgen::InsensitiveStr::Ascii("degenerage"),
                &["degenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("requesr"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("reniged"), &["reneged"]),
            (
                dictgen::InsensitiveStr::Ascii("revewed"),
                &["reviewed", "renewed", "reveled"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjsence"), &["adjacence"]),
            (
                dictgen::InsensitiveStr::Ascii("amrageddon"),
                &["armageddon"],
            ),
            (dictgen::InsensitiveStr::Ascii("carcuses"), &["carcasses"]),
            (dictgen::InsensitiveStr::Ascii("trunacted"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("picther"), &["pitcher"]),
            (dictgen::InsensitiveStr::Ascii("existant"), &["existent"]),
            (
                dictgen::InsensitiveStr::Ascii("negosiating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("launchign"), &["launching"]),
            (dictgen::InsensitiveStr::Ascii("expaning"), &["expanding"]),
            (
                dictgen::InsensitiveStr::Ascii("frankenstine"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("romotly"), &["remotely"]),
            (
                dictgen::InsensitiveStr::Ascii("sroll"),
                &["scroll", "stroll"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhaces"), &["enhances"]),
            (dictgen::InsensitiveStr::Ascii("begginng"), &["begging"]),
            (
                dictgen::InsensitiveStr::Ascii("transocdes"),
                &["transcodes"],
            ),
            (dictgen::InsensitiveStr::Ascii("jamacain"), &["jamaican"]),
            (dictgen::InsensitiveStr::Ascii("enableded"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("follwed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("strarting"), &["starting"]),
            (
                dictgen::InsensitiveStr::Ascii("psudoterminal"),
                &["pseudoterminal"],
            ),
            (dictgen::InsensitiveStr::Ascii("sokobon"), &["sokoban"]),
            (
                dictgen::InsensitiveStr::Ascii("nostolgicly"),
                &["nostalgically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specilized"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("scoripon"), &["scorpion"]),
            (dictgen::InsensitiveStr::Ascii("corrdior"), &["corridor"]),
            (dictgen::InsensitiveStr::Ascii("modfiy"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("tattooes"), &["tattoos"]),
            (dictgen::InsensitiveStr::Ascii("calclates"), &["calculates"]),
            (
                dictgen::InsensitiveStr::Ascii("multipilers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("loock"), &["lock", "look"]),
            (
                dictgen::InsensitiveStr::Ascii("condiction"),
                &["condition", "conduction", "conviction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparcencies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynmics"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaitions"),
                &["reallocations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpenticular"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersertions"),
                &["insertions", "intersections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedection"),
                &["detection", "deduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("authoer"), &["author"]),
            (dictgen::InsensitiveStr::Ascii("argumnts"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("aksed"), &["asked"]),
            (
                dictgen::InsensitiveStr::Ascii("ineqaulities"),
                &["inequalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ubelieveble"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("healthit"), &["healthiest"]),
            (
                dictgen::InsensitiveStr::Ascii("ceremonije"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randayvous"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforrcing"), &["enforcing"]),
            (dictgen::InsensitiveStr::Ascii("egdes"), &["edges"]),
            (dictgen::InsensitiveStr::Ascii("asexaul"), &["asexual"]),
            (
                dictgen::InsensitiveStr::Ascii("announement"),
                &["announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideosincracies"),
                &["idiosyncrasies"],
            ),
            (dictgen::InsensitiveStr::Ascii("vechile"), &["vehicle"]),
            (dictgen::InsensitiveStr::Ascii("structurd"), &["structured"]),
            (
                dictgen::InsensitiveStr::Ascii("multiprocesor"),
                &["multiprocessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timesamped"),
                &["timestamped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governemnts"),
                &["governments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proproties"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("supprted"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("intepolation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramete"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistence"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("foler"), &["folder"]),
            (
                dictgen::InsensitiveStr::Ascii("transcocders"),
                &["transcoders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requrirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantasyzing"),
                &["fantasizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("invokable"), &["invocable"]),
            (dictgen::InsensitiveStr::Ascii("adress"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("whicht"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("regconizes"),
                &["recognizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("ture"), &["true", "pure"]),
            (dictgen::InsensitiveStr::Ascii("cutomary"), &["customary"]),
            (dictgen::InsensitiveStr::Ascii("suppposed"), &["supposed"]),
            (
                dictgen::InsensitiveStr::Ascii("declerated"),
                &["decelerated", "declared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplment"),
                &["reimplement"],
            ),
            (dictgen::InsensitiveStr::Ascii("conigures"), &["configures"]),
            (dictgen::InsensitiveStr::Ascii("expries"), &["expires"]),
            (dictgen::InsensitiveStr::Ascii("cleanning"), &["cleaning"]),
            (
                dictgen::InsensitiveStr::Ascii("minimalisity"),
                &["minimalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("loadning"), &["loading"]),
            (
                dictgen::InsensitiveStr::Ascii("modiifications"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempearures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canonalized"),
                &["canonicalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminiates"),
                &["terminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicriminating"),
                &["discriminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guareenteing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technnologies"),
                &["technologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("romaina"), &["romania"]),
            (dictgen::InsensitiveStr::Ascii("villians"), &["villains"]),
            (dictgen::InsensitiveStr::Ascii("fiields"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("adjustmens"),
                &["adjustments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialier"),
                &["initializer", "initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuctional"), &["functional"]),
            (
                dictgen::InsensitiveStr::Ascii("noramalising"),
                &["normalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumbenus"), &["submenus"]),
            (dictgen::InsensitiveStr::Ascii("resgined"), &["resigned"]),
            (
                dictgen::InsensitiveStr::Ascii("illuminiati"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("abymsally"), &["abysmally"]),
            (dictgen::InsensitiveStr::Ascii("tentacels"), &["tentacles"]),
            (
                dictgen::InsensitiveStr::Ascii("declerate"),
                &["decelerate", "declare"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceeeded"), &["proceeded"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplmenet"),
                &["reimplement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mothreboard"),
                &["motherboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eignvalues"),
                &["eigenvalues"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpercted"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinics"), &["cynics"]),
            (
                dictgen::InsensitiveStr::Ascii("roats"),
                &["roasts", "roast"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nessessarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architetures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("wesbites"), &["websites"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaucrash"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("awkardly"), &["awkwardly"]),
            (
                dictgen::InsensitiveStr::Ascii("communitcated"),
                &["communicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pbulisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("switcht"),
                &["switched", "switch"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneect"), &["connect"]),
            (dictgen::InsensitiveStr::Ascii("oppens"), &["opens"]),
            (dictgen::InsensitiveStr::Ascii("elemts"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("judgementle"),
                &["judgemental"],
            ),
            (dictgen::InsensitiveStr::Ascii("adapaters"), &["adapters"]),
            (dictgen::InsensitiveStr::Ascii("thr"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("especiallly"),
                &["especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encapsualtes"),
                &["encapsulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaaluable"), &["valuable"]),
            (
                dictgen::InsensitiveStr::Ascii("phsycologically"),
                &["psychologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("speecheasy"), &["speeches"]),
            (dictgen::InsensitiveStr::Ascii("analysied"), &["analysed"]),
            (
                dictgen::InsensitiveStr::Ascii("consideres"),
                &["considered", "considers"],
            ),
            (dictgen::InsensitiveStr::Ascii("welcom"), &["welcome"]),
            (
                dictgen::InsensitiveStr::Ascii("positevely"),
                &["positively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sugggesting"),
                &["suggesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statistisch"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("besided"), &["besides"]),
            (dictgen::InsensitiveStr::Ascii("encodin"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("paradies"), &["paradise"]),
            (dictgen::InsensitiveStr::Ascii("abialble"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("intrduces"), &["introduces"]),
            (dictgen::InsensitiveStr::Ascii("dploys"), &["deploys"]),
            (dictgen::InsensitiveStr::Ascii("robberys"), &["robbers"]),
            (
                dictgen::InsensitiveStr::Ascii("incompetant"),
                &["incompetent"],
            ),
            (dictgen::InsensitiveStr::Ascii("witholds"), &["withholds"]),
            (dictgen::InsensitiveStr::Ascii("insepctor"), &["inspector"]),
            (dictgen::InsensitiveStr::Ascii("modidfier"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("ponits"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("resolverse"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlsates"),
                &["translates"],
            ),
            (dictgen::InsensitiveStr::Ascii("coloskwise"), &["clockwise"]),
            (dictgen::InsensitiveStr::Ascii("retsarts"), &["restarts"]),
            (
                dictgen::InsensitiveStr::Ascii("assimilatie"),
                &["assimilate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conntacting"),
                &["contacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentation"),
                &["concatenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hamburges"), &["hamburgers"]),
            (
                dictgen::InsensitiveStr::Ascii("straightforwad"),
                &["straightforward"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placemets"),
                &["placements", "placemats"],
            ),
            (dictgen::InsensitiveStr::Ascii("causions"), &["cautions"]),
            (dictgen::InsensitiveStr::Ascii("matix"), &["matrix"]),
            (
                dictgen::InsensitiveStr::Ascii("compatative"),
                &["comparative"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanaticus"), &["fanatics"]),
            (
                dictgen::InsensitiveStr::Ascii("governemtn"),
                &["government"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physisians"),
                &["physicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitiations"),
                &["delimitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("relesae"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("inpu"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("conponent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("patriachry"),
                &["patriarchy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parntering"),
                &["partnering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependecys"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("addreess"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("custoizer"), &["customizer"]),
            (dictgen::InsensitiveStr::Ascii("inclused"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("bulder"),
                &["builder", "boulder"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatytpe"), &["datatype"]),
            (dictgen::InsensitiveStr::Ascii("harnesss"), &["harness"]),
            (dictgen::InsensitiveStr::Ascii("dregrees"), &["degrees"]),
            (
                dictgen::InsensitiveStr::Ascii("encounteres"),
                &["encounters"],
            ),
            (dictgen::InsensitiveStr::Ascii("puhsups"), &["pushups"]),
            (
                dictgen::InsensitiveStr::Ascii("theoeretical"),
                &["theoretical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogynisitic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("quesition"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuvre"), &["curve", "cover"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lingeire"), &["lingerie"]),
            (dictgen::InsensitiveStr::Ascii("turtorial"), &["tutorial"]),
            (dictgen::InsensitiveStr::Ascii("kenerls"), &["kernels"]),
            (
                dictgen::InsensitiveStr::Ascii("compiators"),
                &["compilators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crirically"),
                &["critically"],
            ),
            (dictgen::InsensitiveStr::Ascii("nationalis"), &["nationals"]),
            (dictgen::InsensitiveStr::Ascii("ractangle"), &["rectangle"]),
            (
                dictgen::InsensitiveStr::Ascii("sympethetic"),
                &["sympathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninterestin"),
                &["uninteresting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciliazed"),
                &["specialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("animtion"), &["animation"]),
            (
                dictgen::InsensitiveStr::Ascii("disciplince"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("quatized"), &["quantized"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmates"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementastion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("postumus"), &["posthumous"]),
            (dictgen::InsensitiveStr::Ascii("acition"), &["action"]),
            (
                dictgen::InsensitiveStr::Ascii("arbirtrarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangeld"), &["mangled"]),
            (dictgen::InsensitiveStr::Ascii("numbre"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("shephered"), &["shepherd"]),
            (dictgen::InsensitiveStr::Ascii("headest"), &["headset"]),
            (
                dictgen::InsensitiveStr::Ascii("inherithing"),
                &["inheriting"],
            ),
            (dictgen::InsensitiveStr::Ascii("archetyps"), &["archetypes"]),
            (dictgen::InsensitiveStr::Ascii("exccute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("intgrals"), &["integrals"]),
            (
                dictgen::InsensitiveStr::Ascii("healtzh"),
                &["health", "healthz"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obtainible"),
                &["obtainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transylvnia"),
                &["transylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("covention"), &["convention"]),
            (dictgen::InsensitiveStr::Ascii("replacate"), &["replicate"]),
            (dictgen::InsensitiveStr::Ascii("comiles"), &["compiles"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberghs"),
                &["neighbors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminazi"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("anedoctes"), &["anecdotes"]),
            (dictgen::InsensitiveStr::Ascii("intervenue"), &["intervene"]),
            (dictgen::InsensitiveStr::Ascii("pairokeel"), &["parochial"]),
            (
                dictgen::InsensitiveStr::Ascii("irriversible"),
                &["irreversible"],
            ),
            (dictgen::InsensitiveStr::Ascii("decocders"), &["decoders"]),
            (
                dictgen::InsensitiveStr::Ascii("architecutre"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copurights"),
                &["copyrights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterprented"),
                &["misinterpret"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcurating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstracions"),
                &["abstractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("lisetn"), &["listen"]),
            (
                dictgen::InsensitiveStr::Ascii("precendence"),
                &["precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("ises"), &["uses"]),
            (
                dictgen::InsensitiveStr::Ascii("tiangle"),
                &["triangle", "tingle", "tangle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambidexterousness"),
                &["ambidextrousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("coprses"), &["corpses"]),
            (dictgen::InsensitiveStr::Ascii("behavor"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("juxt"), &["just"]),
            (dictgen::InsensitiveStr::Ascii("sipport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("viligant"), &["vigilant"]),
            (dictgen::InsensitiveStr::Ascii("directon"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("oarticles"), &["particles"]),
            (dictgen::InsensitiveStr::Ascii("traveles"), &["travels"]),
            (
                dictgen::InsensitiveStr::Ascii("collobaration"),
                &["collaboration"],
            ),
            (dictgen::InsensitiveStr::Ascii("comverts"), &["converts"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamances"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trogloditical"),
                &["troglodytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("douible"), &["double"]),
            (
                dictgen::InsensitiveStr::Ascii("conveinience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectivy"),
                &["objectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambassabors"),
                &["ambassadors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destoryers"),
                &["destroyers"],
            ),
            (dictgen::InsensitiveStr::Ascii("ourced"), &["sourced"]),
            (dictgen::InsensitiveStr::Ascii("follfower"), &["follower"]),
            (
                dictgen::InsensitiveStr::Ascii("instanseation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supporitng"),
                &["supporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exercizes"),
                &["exercise", "exercises", "exorcizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("maturty"), &["maturity"]),
            (dictgen::InsensitiveStr::Ascii("procteted"), &["protected"]),
            (
                dictgen::InsensitiveStr::Ascii("virtualenf"),
                &["virtualenv"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficillic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("seup"), &["setup"]),
            (dictgen::InsensitiveStr::Ascii("scupted"), &["sculpted"]),
            (
                dictgen::InsensitiveStr::Ascii("confidantly"),
                &["confidently"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingeunity"), &["ingenuity"]),
            (
                dictgen::InsensitiveStr::Ascii("invincibile"),
                &["invincible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartising"),
                &["standardising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cultivative"),
                &["cultivate"],
            ),
            (dictgen::InsensitiveStr::Ascii("discussin"), &["discussion"]),
            (
                dictgen::InsensitiveStr::Ascii("elipse"),
                &["ellipse", "eclipse"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenactiy"), &["tenacity"]),
            (dictgen::InsensitiveStr::Ascii("spedify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("notitication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("osbscure"), &["obscure"]),
            (dictgen::InsensitiveStr::Ascii("obhecting"), &["objecting"]),
            (
                dictgen::InsensitiveStr::Ascii("coeffitient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressence"), &["presence"]),
            (
                dictgen::InsensitiveStr::Ascii("nacionalistic"),
                &["nationalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costructor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("deffines"), &["defines"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnositc"),
                &["diagnostic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processores"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contirbute"),
                &["contribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("romaanin"), &["romanian"]),
            (dictgen::InsensitiveStr::Ascii("disalow"), &["disallow"]),
            (
                dictgen::InsensitiveStr::Ascii("passthruogh"),
                &["passthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("logorithm"), &["logarithm"]),
            (
                dictgen::InsensitiveStr::Ascii("disconeccting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("nullifiy"), &["nullify"]),
            (dictgen::InsensitiveStr::Ascii("cariies"), &["carries"]),
            (dictgen::InsensitiveStr::Ascii("svae"), &["save", "suave"]),
            (dictgen::InsensitiveStr::Ascii("execept"), &["except"]),
            (
                dictgen::InsensitiveStr::Ascii("protostant"),
                &["protestant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nagatively"),
                &["negatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derterming"),
                &["determining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficilleid"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("momnet"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("metropilitan"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afghanisthan"),
                &["afghanistan"],
            ),
            (dictgen::InsensitiveStr::Ascii("phisical"), &["physical"]),
            (dictgen::InsensitiveStr::Ascii("irectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("initailisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimima"), &["minima"]),
            (
                dictgen::InsensitiveStr::Ascii("castatrophe"),
                &["catastrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porgrammer"),
                &["programmer"],
            ),
            (dictgen::InsensitiveStr::Ascii("memorys"), &["memories"]),
            (
                dictgen::InsensitiveStr::Ascii("platformie"),
                &["platformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("grenaders"), &["grenades"]),
            (dictgen::InsensitiveStr::Ascii("reeased"), &["released"]),
            (dictgen::InsensitiveStr::Ascii("derectory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("opyions"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("moodel"), &["model"]),
            (dictgen::InsensitiveStr::Ascii("grahpics"), &["graphics"]),
            (dictgen::InsensitiveStr::Ascii("sphagetti"), &["spaghetti"]),
            (dictgen::InsensitiveStr::Ascii("obect"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("evertything"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unasnwered"),
                &["unanswered"],
            ),
            (dictgen::InsensitiveStr::Ascii("myabe"), &["maybe"]),
            (
                dictgen::InsensitiveStr::Ascii("transactoinal"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("currupts"), &["corrupts"]),
            (dictgen::InsensitiveStr::Ascii("vallidity"), &["validity"]),
            (
                dictgen::InsensitiveStr::Ascii("inspektion"),
                &["inspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dealershits"),
                &["dealerships"],
            ),
            (dictgen::InsensitiveStr::Ascii("pary"), &["party", "parry"]),
            (dictgen::InsensitiveStr::Ascii("pattented"), &["patented"]),
            (
                dictgen::InsensitiveStr::Ascii("disclsoure"),
                &["disclosure"],
            ),
            (dictgen::InsensitiveStr::Ascii("racistas"), &["racists"]),
            (dictgen::InsensitiveStr::Ascii("dract"), &["draft"]),
            (dictgen::InsensitiveStr::Ascii("retur"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("unregstering"),
                &["unregistering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulater"),
                &["regulator", "regulated", "regulates", "regulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("eargely"), &["eagerly"]),
            (dictgen::InsensitiveStr::Ascii("userpace"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("spacialise"),
                &["specialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("srollable"), &["scrollable"]),
            (
                dictgen::InsensitiveStr::Ascii("unmodfieid"),
                &["unmodified"],
            ),
            (dictgen::InsensitiveStr::Ascii("knowledage"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("guareentes"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstations"),
                &["demonstrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tne"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboards"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearet"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("mechancis"), &["mechanics"]),
            (
                dictgen::InsensitiveStr::Ascii("speshelly"),
                &["specially", "especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coloeration"),
                &["coloration", "colouration"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoustic"), &["acoustic"]),
            (dictgen::InsensitiveStr::Ascii("techical"), &["technical"]),
            (dictgen::InsensitiveStr::Ascii("lookinng"), &["looking"]),
            (
                dictgen::InsensitiveStr::Ascii("othographic"),
                &["orthographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiallising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("astersik"), &["asterisk"]),
            (
                dictgen::InsensitiveStr::Ascii("snowbaording"),
                &["snowboarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physiciens"),
                &["physicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dircetories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("foregound"), &["foreground"]),
            (dictgen::InsensitiveStr::Ascii("commnunity"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("surounded"), &["surrounded"]),
            (
                dictgen::InsensitiveStr::Ascii("annoymously"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("legasy"), &["legacy"]),
            (dictgen::InsensitiveStr::Ascii("mudering"), &["murdering"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequisits"),
                &["prerequisites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troughput"),
                &["throughput", "throughout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nin"),
                &["inn", "min", "bin", "nine"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqence"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("autherised"),
                &["authorised"],
            ),
            (dictgen::InsensitiveStr::Ascii("frecklers"), &["freckles"]),
            (dictgen::InsensitiveStr::Ascii("anlyst"), &["analyst"]),
            (
                dictgen::InsensitiveStr::Ascii("insturcted"),
                &["instructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("checled"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("underlflowing"),
                &["underflowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulfunctions"),
                &["malfunctions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roiginating"),
                &["originating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protagoinst"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoyning"), &["annoying"]),
            (dictgen::InsensitiveStr::Ascii("overrrides"), &["overrides"]),
            (
                dictgen::InsensitiveStr::Ascii("enountered"),
                &["encountered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breakthrouhgs"),
                &["breakthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfomer"),
                &["transformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upperacing"),
                &["uppercasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tabelspoon"),
                &["tablespoon"],
            ),
            (dictgen::InsensitiveStr::Ascii("folows"), &["follows"]),
            (dictgen::InsensitiveStr::Ascii("chelsae"), &["chelsea"]),
            (dictgen::InsensitiveStr::Ascii("quiries"), &["queries"]),
            (
                dictgen::InsensitiveStr::Ascii("obects"),
                &["object", "objects"],
            ),
            (dictgen::InsensitiveStr::Ascii("entires"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("arbritary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("depandancy"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangmeent"),
                &["rearrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantate"),
                &["instantiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conductuctance"),
                &["conductance"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramemter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("gerenated"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("yosemeti"), &["yosemite"]),
            (dictgen::InsensitiveStr::Ascii("contatct"), &["contact"]),
            (dictgen::InsensitiveStr::Ascii("requirin"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("inctance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("immsersive"), &["immersive"]),
            (
                dictgen::InsensitiveStr::Ascii("cinematogrpahy"),
                &["cinematography"],
            ),
            (dictgen::InsensitiveStr::Ascii("mandotory"), &["mandatory"]),
            (dictgen::InsensitiveStr::Ascii("throu"), &["thru"]),
            (
                dictgen::InsensitiveStr::Ascii("sirectionality"),
                &["directionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("syatem"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("achievents"),
                &["achieves", "achievements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralelize"),
                &["parallelize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevatible"),
                &["inevitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wakin"),
                &["waking", "waken", "wain", "akin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exercize"),
                &["exercise", "exorcize"],
            ),
            (dictgen::InsensitiveStr::Ascii("explaind"), &["explained"]),
            (dictgen::InsensitiveStr::Ascii("upion"), &["upon"]),
            (
                dictgen::InsensitiveStr::Ascii("condiitonally"),
                &["conditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectable"), &["executable"]),
            (
                dictgen::InsensitiveStr::Ascii("pervented"),
                &["prevented", "perverted"],
            ),
            (dictgen::InsensitiveStr::Ascii("condemnd"), &["condemned"]),
            (
                dictgen::InsensitiveStr::Ascii("registraions"),
                &["registrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("confert"), &["convert"]),
            (
                dictgen::InsensitiveStr::Ascii("algotithmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definicion"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mellinneums"),
                &["millennia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revelution"),
                &["revelation", "revolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("quickyl"), &["quickly"]),
            (dictgen::InsensitiveStr::Ascii("functinal"), &["functional"]),
            (dictgen::InsensitiveStr::Ascii("fileters"), &["filters"]),
            (
                dictgen::InsensitiveStr::Ascii("allegaince"),
                &["allegiance"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploting"), &["plotting"]),
            (dictgen::InsensitiveStr::Ascii("cosoles"), &["consoles"]),
            (dictgen::InsensitiveStr::Ascii("provisiosn"), &["provision"]),
            (
                dictgen::InsensitiveStr::Ascii("exexutable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aktivation"),
                &["activation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenionsality"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterizarion"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corellates"),
                &["correlates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncondtionally"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presistantly"),
                &["persistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perswasion"),
                &["persuasion"],
            ),
            (dictgen::InsensitiveStr::Ascii("norhtern"), &["northern"]),
            (dictgen::InsensitiveStr::Ascii("apriciate"), &["appreciate"]),
            (dictgen::InsensitiveStr::Ascii("isalmist"), &["islamist"]),
            (
                dictgen::InsensitiveStr::Ascii("chormosomes"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("msut"), &["must"]),
            (
                dictgen::InsensitiveStr::Ascii("deambiguages"),
                &["disambiguates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("graduacion"),
                &["graduation"],
            ),
            (dictgen::InsensitiveStr::Ascii("grpahite"), &["graphite"]),
            (dictgen::InsensitiveStr::Ascii("concider"), &["consider"]),
            (dictgen::InsensitiveStr::Ascii("susseed"), &["secede"]),
            (dictgen::InsensitiveStr::Ascii("posssible"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("srvice"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("understad"),
                &["understands", "understand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninititalized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfoemnaces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagonaal"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("initail"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("cehckerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relfective"),
                &["reflective"],
            ),
            (dictgen::InsensitiveStr::Ascii("therfore"), &["therefore"]),
            (
                dictgen::InsensitiveStr::Ascii("accessibilitiy"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("debhlper"), &["debhelper"]),
            (dictgen::InsensitiveStr::Ascii("optimaal"), &["optimal"]),
            (dictgen::InsensitiveStr::Ascii("densitity"), &["density"]),
            (
                dictgen::InsensitiveStr::Ascii("configuratien"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("enqeue"), &["enqueue"]),
            (dictgen::InsensitiveStr::Ascii("influeces"), &["influences"]),
            (
                dictgen::InsensitiveStr::Ascii("suspicians"),
                &["suspicions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenifying"),
                &["identifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupication"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coneection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transportng"),
                &["transporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("similair"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("detectting"), &["detecting"]),
            (dictgen::InsensitiveStr::Ascii("conceous"), &["conscious"]),
            (dictgen::InsensitiveStr::Ascii("prepered"), &["prepared"]),
            (dictgen::InsensitiveStr::Ascii("vacinated"), &["vaccinated"]),
            (
                dictgen::InsensitiveStr::Ascii("uninistallers"),
                &["uninstallers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhoor"),
                &["neighbor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deinstantating"),
                &["deinstantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomoorow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("decompresing"),
                &["decompressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("eperience"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("terratories"),
                &["territories"],
            ),
            (dictgen::InsensitiveStr::Ascii("directorty"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("boutiqe"), &["boutique"]),
            (dictgen::InsensitiveStr::Ascii("anlyze"), &["analyze"]),
            (
                dictgen::InsensitiveStr::Ascii("exectations"),
                &["expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aletrnative"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("langualges"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("staionairy"),
                &["stationary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("writebufer"),
                &["writebuffer"],
            ),
            (dictgen::InsensitiveStr::Ascii("arraows"), &["arrows"]),
            (
                dictgen::InsensitiveStr::Ascii("desriptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rendazyvoued"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alllocators"),
                &["allocators"],
            ),
            (dictgen::InsensitiveStr::Ascii("freited"), &["freighted"]),
            (dictgen::InsensitiveStr::Ascii("lgeacy"), &["legacy"]),
            (dictgen::InsensitiveStr::Ascii("exisited"), &["existed"]),
            (
                dictgen::InsensitiveStr::Ascii("decentraliced"),
                &["decentralized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimmetricly"),
                &["symmetricly"],
            ),
            (dictgen::InsensitiveStr::Ascii("veriosns"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("checllists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identitites"),
                &["identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("muliple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("constructifs"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrainted"),
                &["constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optomizing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfices"), &["offices"]),
            (
                dictgen::InsensitiveStr::Ascii("specifich"),
                &["specify", "specific"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfec"), &["perfect"]),
            (
                dictgen::InsensitiveStr::Ascii("peizometric"),
                &["piezometric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptionel"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("continer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("striings"), &["strings"]),
            (
                dictgen::InsensitiveStr::Ascii("tunning"),
                &["tuning", "running"],
            ),
            (dictgen::InsensitiveStr::Ascii("outputsr"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("symettric"), &["symmetric"]),
            (
                dictgen::InsensitiveStr::Ascii("modifaction"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrbuting"),
                &["contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("montypic"), &["monotypic"]),
            (dictgen::InsensitiveStr::Ascii("elemenst"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("arramged"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("orkids"), &["orchids"]),
            (dictgen::InsensitiveStr::Ascii("coorinate"), &["coordinate"]),
            (
                dictgen::InsensitiveStr::Ascii("disicplined"),
                &["disciplined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associeted"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loadbalncer"),
                &["loadbalancer"],
            ),
            (dictgen::InsensitiveStr::Ascii("controler"), &["controller"]),
            (
                dictgen::InsensitiveStr::Ascii("supernatual"),
                &["supernatural"],
            ),
            (dictgen::InsensitiveStr::Ascii("trigerred"), &["triggered"]),
            (dictgen::InsensitiveStr::Ascii("chaarges"), &["charges"]),
            (dictgen::InsensitiveStr::Ascii("distrcit"), &["district"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasociates"),
                &["disassociates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conidering"),
                &["considering"],
            ),
            (dictgen::InsensitiveStr::Ascii("clanup"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("ocupier"), &["occupier"]),
            (
                dictgen::InsensitiveStr::Ascii("preformated"),
                &["preformatted"],
            ),
            (dictgen::InsensitiveStr::Ascii("registr"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("coulumns"),
                &["columns", "coulombs"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheeve"), &["achieve"]),
            (dictgen::InsensitiveStr::Ascii("hanppen"), &["happen"]),
            (dictgen::InsensitiveStr::Ascii("resopnse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("propably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("communcate"),
                &["communicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demensional"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoicalizing"),
                &["canonicalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substituer"),
                &["substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("acivity"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("erroenous"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("defintian"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("sempahores"),
                &["semaphores"],
            ),
            (dictgen::InsensitiveStr::Ascii("februrary"), &["february"]),
            (
                dictgen::InsensitiveStr::Ascii("cristallizing"),
                &["crystallizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victorioso"),
                &["victorious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envolves"),
                &["involves", "evolves"],
            ),
            (dictgen::InsensitiveStr::Ascii("numebering"), &["numbering"]),
            (
                dictgen::InsensitiveStr::Ascii("necesities"),
                &["necessities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dificulties"),
                &["difficulties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pydandic"),
                &["pydantic", "pedantic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioble"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("impprove"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("redubant"), &["redundant"]),
            (dictgen::InsensitiveStr::Ascii("gien"), &["given"]),
            (dictgen::InsensitiveStr::Ascii("bascially"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("inivisible"), &["invisible"]),
            (dictgen::InsensitiveStr::Ascii("setp"), &["step"]),
            (
                dictgen::InsensitiveStr::Ascii("requiremenht"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("whnever"), &["whenever"]),
            (
                dictgen::InsensitiveStr::Ascii("quanifications"),
                &["qualifications", "quantifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotrols"), &["controls"]),
            (
                dictgen::InsensitiveStr::Ascii("frankenstiens"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("changged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("parocheelity"),
                &["parochiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlelation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camoflaging"),
                &["camouflaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("siwtched"), &["switched"]),
            (dictgen::InsensitiveStr::Ascii("calfs"), &["calves"]),
            (dictgen::InsensitiveStr::Ascii("librayr"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("undersetimate"),
                &["underestimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abtractions"),
                &["abstractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addressible"),
                &["addressable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addministered"),
                &["administered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("claread"), &["cleared"]),
            (
                dictgen::InsensitiveStr::Ascii("heirarchies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("shpuld"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("arbitiray"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("excatly"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("beginining"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplyed"),
                &["multiplied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversatives"),
                &["conservatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferens"),
                &["preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phychedelics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("skipps"), &["skips"]),
            (
                dictgen::InsensitiveStr::Ascii("interescts"),
                &["intersects"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteraor"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("sucessive"), &["successive"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnostioc"),
                &["diagnostic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocationg"),
                &["allocating", "allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("flattend"), &["flattened"]),
            (dictgen::InsensitiveStr::Ascii("edditing"), &["editing"]),
            (
                dictgen::InsensitiveStr::Ascii("microscopice"),
                &["microscope"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplyments"),
                &["deployments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differentiantion"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congifurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categoriezed"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("auidobook"), &["audiobook"]),
            (dictgen::InsensitiveStr::Ascii("ractise"), &["practise"]),
            (
                dictgen::InsensitiveStr::Ascii("idividually"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("conist"), &["consist"]),
            (
                dictgen::InsensitiveStr::Ascii("instaciated"),
                &["instantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("necessar"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("suppprt"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("expermientally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chekcerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("euorpeans"), &["europeans"]),
            (
                dictgen::InsensitiveStr::Ascii("galvinized"),
                &["galvanized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correciton"),
                &["correction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playtgroughs"),
                &["playthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constenting"),
                &["consenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atheistisch"),
                &["atheistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("declring"), &["declaring"]),
            (
                dictgen::InsensitiveStr::Ascii("disrespection"),
                &["disrespecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratutious"),
                &["gratuitous"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalulated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("lugage"), &["luggage"]),
            (dictgen::InsensitiveStr::Ascii("kazakstan"), &["kazakhstan"]),
            (
                dictgen::InsensitiveStr::Ascii("assissinated"),
                &["assassinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terretories"),
                &["territories"],
            ),
            (dictgen::InsensitiveStr::Ascii("geeoteen"), &["guillotine"]),
            (dictgen::InsensitiveStr::Ascii("iteated"), &["iterated"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasudicals"),
                &["pharmaceuticals"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreement"), &["agreement"]),
            (dictgen::InsensitiveStr::Ascii("rmaining"), &["remaining"]),
            (dictgen::InsensitiveStr::Ascii("stuborn"), &["stubborn"]),
            (dictgen::InsensitiveStr::Ascii("segmenets"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("strinc"), &["string"]),
            (dictgen::InsensitiveStr::Ascii("anayltic"), &["analytic"]),
            (
                dictgen::InsensitiveStr::Ascii("reciptocal"),
                &["reciprocal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iluminates"),
                &["illuminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("zeebra"), &["zebra"]),
            (dictgen::InsensitiveStr::Ascii("mezmorize"), &["mesmerize"]),
            (
                dictgen::InsensitiveStr::Ascii("specifcally"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supersticious"),
                &["superstitious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imporantly"),
                &["importantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("amnsety"), &["amnesty"]),
            (
                dictgen::InsensitiveStr::Ascii("propolsion"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("bengazhi"), &["benghazi"]),
            (dictgen::InsensitiveStr::Ascii("utillized"), &["utilized"]),
            (
                dictgen::InsensitiveStr::Ascii("fundimental"),
                &["fundamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("licnese"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("accomodatingly"),
                &["accommodatingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("querstioned"),
                &["questioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("christinas"),
                &["christians"],
            ),
            (dictgen::InsensitiveStr::Ascii("chganges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("intitiatives"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commectivity"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("chlesea"), &["chelsea"]),
            (
                dictgen::InsensitiveStr::Ascii("admininstering"),
                &["administering"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethimng"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("vegeteble"), &["vegetable"]),
            (
                dictgen::InsensitiveStr::Ascii("commnetary"),
                &["commentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("naibhorhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redirectrions"),
                &["redirections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invitiation"),
                &["invitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inactice"), &["inactive"]),
            (dictgen::InsensitiveStr::Ascii("sepeators"), &["separators"]),
            (dictgen::InsensitiveStr::Ascii("adroable"), &["adorable"]),
            (dictgen::InsensitiveStr::Ascii("cours"), &["course"]),
            (
                dictgen::InsensitiveStr::Ascii("initialsed"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternetevely"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicuss"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("dissapper"), &["disappear"]),
            (dictgen::InsensitiveStr::Ascii("observ"), &["observe"]),
            (
                dictgen::InsensitiveStr::Ascii("probablistically"),
                &["probabilistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecttion"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("pseudonyn"), &["pseudonym"]),
            (
                dictgen::InsensitiveStr::Ascii("appriciated"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ckeckboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trama"),
                &["trauma", "tram", "trams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devistating"),
                &["devastating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expections"),
                &["exceptions", "expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwright"),
                &["overwrite", "overweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitons"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("durectories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perdictive"),
                &["predictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fullfilling"),
                &["fulfilling"],
            ),
            (dictgen::InsensitiveStr::Ascii("premit"), &["permit"]),
            (
                dictgen::InsensitiveStr::Ascii("terimnating"),
                &["terminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bycicling"), &["bicycling"]),
            (dictgen::InsensitiveStr::Ascii("belive"), &["believe"]),
            (dictgen::InsensitiveStr::Ascii("classied"), &["classified"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeing"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilavalent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repant"),
                &["repaint", "repent"],
            ),
            (dictgen::InsensitiveStr::Ascii("picnicing"), &["picnicking"]),
            (dictgen::InsensitiveStr::Ascii("meassuring"), &["measuring"]),
            (dictgen::InsensitiveStr::Ascii("adovcacy"), &["advocacy"]),
            (
                dictgen::InsensitiveStr::Ascii("criticisied"),
                &["criticise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overreacing"),
                &["overreacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("prvious"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("securuity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("addressse"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("suscribed"), &["subscribed"]),
            (dictgen::InsensitiveStr::Ascii("shiped"), &["shipped"]),
            (dictgen::InsensitiveStr::Ascii("trasmits"), &["transmits"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellize"),
                &["parallelize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illumaniti"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficilliable"),
                &["specifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("flagg"), &["flag"]),
            (dictgen::InsensitiveStr::Ascii("remakrs"), &["remarks"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprenuer"),
                &["entrepreneur"],
            ),
            (dictgen::InsensitiveStr::Ascii("technnique"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("expermiments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simliarity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("outweights"), &["outweighs"]),
            (
                dictgen::InsensitiveStr::Ascii("electricty"),
                &["electricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("urkaine"), &["ukraine"]),
            (dictgen::InsensitiveStr::Ascii("downvots"), &["downvotes"]),
            (
                dictgen::InsensitiveStr::Ascii("indpendent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valudators"),
                &["validators"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinated"), &["combined"]),
            (dictgen::InsensitiveStr::Ascii("qeuestion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("certifcation"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("aproch"), &["approach"]),
            (
                dictgen::InsensitiveStr::Ascii("converage"),
                &["converge", "coverage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civalasation"),
                &["civilisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pengwins"), &["penguins"]),
            (
                dictgen::InsensitiveStr::Ascii("considerarte"),
                &["considerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("covarages"), &["coverages"]),
            (
                dictgen::InsensitiveStr::Ascii("pattren"),
                &["pattern", "patron"],
            ),
            (dictgen::InsensitiveStr::Ascii("hashses"), &["hashes"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulised"),
                &["virtualised"],
            ),
            (dictgen::InsensitiveStr::Ascii("currentl"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalit"),
                &["nationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecularity"),
                &["peculiarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("comradets"), &["comrades"]),
            (
                dictgen::InsensitiveStr::Ascii("ommishients"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abortificant"),
                &["abortifacient"],
            ),
            (dictgen::InsensitiveStr::Ascii("loopup"), &["lookup"]),
            (dictgen::InsensitiveStr::Ascii("sequece"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("screenwrighter"),
                &["screenwriter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternaties"),
                &["alternates", "alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("compelete"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("ifle"), &["file"]),
            (
                dictgen::InsensitiveStr::Ascii("deductables"),
                &["deductibles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flexbility"),
                &["flexibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenitify"), &["identify"]),
            (
                dictgen::InsensitiveStr::Ascii("psychyatrist"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("variablies"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("readanle"), &["readable"]),
            (
                dictgen::InsensitiveStr::Ascii("identifing"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("consumiste"), &["consumes"]),
            (dictgen::InsensitiveStr::Ascii("acctively"), &["actively"]),
            (
                dictgen::InsensitiveStr::Ascii("dosclosure"),
                &["disclosure"],
            ),
            (dictgen::InsensitiveStr::Ascii("latests"), &["latest"]),
            (
                dictgen::InsensitiveStr::Ascii("embbed"),
                &["embed", "ebbed"],
            ),
            (dictgen::InsensitiveStr::Ascii("pararllell"), &["parallel"]),
            (
                dictgen::InsensitiveStr::Ascii("idenification"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horisontally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unzaturated"),
                &["unsaturated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complacient"),
                &["complacent"],
            ),
            (dictgen::InsensitiveStr::Ascii("esnures"), &["ensures"]),
            (dictgen::InsensitiveStr::Ascii("novermber"), &["november"]),
            (
                dictgen::InsensitiveStr::Ascii("cmobination"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recompence"),
                &["recompense"],
            ),
            (dictgen::InsensitiveStr::Ascii("natuarally"), &["naturally"]),
            (dictgen::InsensitiveStr::Ascii("scientfic"), &["scientific"]),
            (
                dictgen::InsensitiveStr::Ascii("unnecissarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("memove"), &["memmove"]),
            (dictgen::InsensitiveStr::Ascii("dessigning"), &["designing"]),
            (dictgen::InsensitiveStr::Ascii("dishonesy"), &["dishonesty"]),
            (dictgen::InsensitiveStr::Ascii("operaands"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("elsewehere"), &["elsewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("subsitutution"),
                &["substitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("undersog"), &["undergo"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimilary"),
                &["dissimilarly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parametetrized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("seected"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("chaned"),
                &["changed", "chained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adapd"),
                &["adapt", "adapted", "adopt", "adopted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technolgies"),
                &["technologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copilating"),
                &["compiling", "copulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qickly"),
                &["quickly", "sickly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoristhm"),
                &["algorithm", "algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("lmited"), &["limited"]),
            (
                dictgen::InsensitiveStr::Ascii("natinalism"),
                &["nationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgiginal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("cigarett"), &["cigarette"]),
            (dictgen::InsensitiveStr::Ascii("fimiliar"), &["familiar"]),
            (
                dictgen::InsensitiveStr::Ascii("openscource"),
                &["opensource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contirbuting"),
                &["contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("squishly"), &["squishy"]),
            (dictgen::InsensitiveStr::Ascii("spiritial"), &["spiritual"]),
            (
                dictgen::InsensitiveStr::Ascii("experimmenting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reputacion"),
                &["reputation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deivces"), &["devices"]),
            (dictgen::InsensitiveStr::Ascii("enegery"), &["energy"]),
            (
                dictgen::InsensitiveStr::Ascii("conflites"),
                &["conflicts", "conflates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepatators"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("arithetic"), &["arithmetic"]),
            (
                dictgen::InsensitiveStr::Ascii("worke"),
                &["work", "worked", "works"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturtural"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("evironment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooof"), &["proof"]),
            (dictgen::InsensitiveStr::Ascii("prologe"), &["prologue"]),
            (dictgen::InsensitiveStr::Ascii("brancket"), &["bracket"]),
            (
                dictgen::InsensitiveStr::Ascii("supsequent"),
                &["subsequent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriottism"),
                &["patriotism"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependat"), &["dependant"]),
            (
                dictgen::InsensitiveStr::Ascii("versitlity"),
                &["versatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluclates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internatinal"),
                &["international"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unecessarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regressivo"),
                &["regression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisison"),
                &["commissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("babysite"), &["babysitter"]),
            (dictgen::InsensitiveStr::Ascii("stoers"), &["stores"]),
            (
                dictgen::InsensitiveStr::Ascii("replacment"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("youn"),
                &["your", "you", "young"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorizes"),
                &["prioritizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolutey"),
                &["absolutely", "absolute"],
            ),
            (dictgen::InsensitiveStr::Ascii("outllook"), &["outlook"]),
            (dictgen::InsensitiveStr::Ascii("follfowing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("apporoximations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibilitities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncontrollablly"),
                &["uncontrollably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electorlytes"),
                &["electrolytes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmd"),
                &["programmed", "programme"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrevirsible"),
                &["irreversible"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectos"), &["connectors"]),
            (
                dictgen::InsensitiveStr::Ascii("nescesarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlfowed"),
                &["overflowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparitions"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ploynomials"),
                &["polynomials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimunitive"),
                &["diminutive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticaton"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("neightbors"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("identifed"), &["identified"]),
            (dictgen::InsensitiveStr::Ascii("expetc"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("rythims"), &["rhythms"]),
            (
                dictgen::InsensitiveStr::Ascii("aestheticlly"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwrittin"),
                &["overwriting", "overwritten"],
            ),
            (dictgen::InsensitiveStr::Ascii("troup"), &["troupe"]),
            (
                dictgen::InsensitiveStr::Ascii("northwesten"),
                &["northwestern"],
            ),
            (dictgen::InsensitiveStr::Ascii("disguisted"), &["disguise"]),
            (
                dictgen::InsensitiveStr::Ascii("encrytption"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("empoloyees"), &["employees"]),
            (
                dictgen::InsensitiveStr::Ascii("loging"),
                &["logging", "lodging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maketplace"),
                &["marketplace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administrativo"),
                &["administration", "administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("ionde"), &["inode"]),
            (
                dictgen::InsensitiveStr::Ascii("decompresson"),
                &["decompression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asnychronouely"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("disentry"), &["dysentery"]),
            (
                dictgen::InsensitiveStr::Ascii("frankensteen"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preinitalized"),
                &["preinitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("boelean"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("deliminted"), &["delimited"]),
            (dictgen::InsensitiveStr::Ascii("respectons"), &["respects"]),
            (
                dictgen::InsensitiveStr::Ascii("mattern"),
                &["matter", "pattern"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconnnaissance"),
                &["reconnaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("releant"),
                &["relevant", "relent"],
            ),
            (dictgen::InsensitiveStr::Ascii("heaer"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("projecttion"),
                &["projection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correograph"),
                &["choreograph"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vietnamees"),
                &["vietnamese"],
            ),
            (dictgen::InsensitiveStr::Ascii("monstre"), &["monster"]),
            (dictgen::InsensitiveStr::Ascii("aprtment"), &["apartment"]),
            (dictgen::InsensitiveStr::Ascii("shapshots"), &["snapshots"]),
            (
                dictgen::InsensitiveStr::Ascii("orginazational"),
                &["organizational"],
            ),
            (dictgen::InsensitiveStr::Ascii("elimnate"), &["eliminate"]),
            (dictgen::InsensitiveStr::Ascii("dsiabled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("noptifying"), &["notifying"]),
            (dictgen::InsensitiveStr::Ascii("archve"), &["archive"]),
            (
                dictgen::InsensitiveStr::Ascii("analying"),
                &["analysing", "analyzing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distinctie"),
                &["distinctive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("receipients"),
                &["recipients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminastrator"),
                &["administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abriviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerbailities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("guidlines"), &["guidelines"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcuates"),
                &["recalculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("dublicat"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("visuaization"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("slewth"), &["sleuth"]),
            (dictgen::InsensitiveStr::Ascii("pervertes"), &["perverse"]),
            (
                dictgen::InsensitiveStr::Ascii("propagater"),
                &["propagator", "propagated", "propagates", "propagate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoicalization"),
                &["canonicalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permisssions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practie"),
                &["practise", "practice"],
            ),
            (dictgen::InsensitiveStr::Ascii("retstore"), &["restore"]),
            (
                dictgen::InsensitiveStr::Ascii("prellocate"),
                &["preallocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("requst"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("inclding"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("contianer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("commma"), &["comma"]),
            (
                dictgen::InsensitiveStr::Ascii("assoication"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadiquate"),
                &["inadequate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectular"),
                &["spectacular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzier"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("privleges"), &["privileges"]),
            (dictgen::InsensitiveStr::Ascii("comformer"), &["conformer"]),
            (dictgen::InsensitiveStr::Ascii("safepooint"), &["safepoint"]),
            (dictgen::InsensitiveStr::Ascii("varegated"), &["variegated"]),
            (dictgen::InsensitiveStr::Ascii("accrding"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("downoloads"), &["downloads"]),
            (
                dictgen::InsensitiveStr::Ascii("assymetrical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponencial"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoyingy"), &["annoyingly"]),
            (dictgen::InsensitiveStr::Ascii("todat"), &["today"]),
            (
                dictgen::InsensitiveStr::Ascii("recommmends"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancesetor"), &["ancestor"]),
            (dictgen::InsensitiveStr::Ascii("camparing"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("excape"), &["escape"]),
            (dictgen::InsensitiveStr::Ascii("pipiline"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("attibuted"), &["attributed"]),
            (
                dictgen::InsensitiveStr::Ascii("documentarse"),
                &["documentaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soures"),
                &["sources", "sores", "sours", "soured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respiratoy"),
                &["respiratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expewrimentally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("clarrity"), &["clarity"]),
            (dictgen::InsensitiveStr::Ascii("tusnami"), &["tsunami"]),
            (
                dictgen::InsensitiveStr::Ascii("initailise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrresponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrpts"), &["encrypts"]),
            (
                dictgen::InsensitiveStr::Ascii("forcifully"),
                &["forcefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discontiguos"),
                &["discontiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensacional"),
                &["sensational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envrionment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("souvenier"), &["souvenir"]),
            (
                dictgen::InsensitiveStr::Ascii("correpsondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminiters"),
                &["administers"],
            ),
            (dictgen::InsensitiveStr::Ascii("registiry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("regarts"), &["regards"]),
            (
                dictgen::InsensitiveStr::Ascii("cirucmflex"),
                &["circumflex"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanciates"),
                &["instantiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("entrophy"), &["entropy"]),
            (dictgen::InsensitiveStr::Ascii("dircet"), &["direct"]),
            (dictgen::InsensitiveStr::Ascii("alogn"), &["along", "align"]),
            (
                dictgen::InsensitiveStr::Ascii("conecnt"),
                &["connect", "content"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disabilitiles"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explination"),
                &["explanation"],
            ),
            (dictgen::InsensitiveStr::Ascii("positivie"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("opponant"), &["opponent"]),
            (
                dictgen::InsensitiveStr::Ascii("probablity"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purtaining"),
                &["pertaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("inveting"), &["inverting"]),
            (
                dictgen::InsensitiveStr::Ascii("exclicitly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("destoyer"), &["destroyer"]),
            (dictgen::InsensitiveStr::Ascii("magled"), &["mangled"]),
            (dictgen::InsensitiveStr::Ascii("oredes"), &["orders"]),
            (
                dictgen::InsensitiveStr::Ascii("anrachists"),
                &["anarchists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffentiation"),
                &["differentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("traveld"), &["traveled"]),
            (dictgen::InsensitiveStr::Ascii("generatd"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("exercieses"), &["exercises"]),
            (dictgen::InsensitiveStr::Ascii("virttual"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("termniators"),
                &["terminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("leibnitz"), &["leibniz"]),
            (dictgen::InsensitiveStr::Ascii("desigern"), &["designer"]),
            (dictgen::InsensitiveStr::Ascii("cannott"), &["cannot"]),
            (dictgen::InsensitiveStr::Ascii("demandes"), &["demands"]),
            (dictgen::InsensitiveStr::Ascii("impcats"), &["impacts"]),
            (
                dictgen::InsensitiveStr::Ascii("entertainig"),
                &["entertaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("speshal"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("memmicked"), &["mimicked"]),
            (dictgen::InsensitiveStr::Ascii("cursror"), &["cursor"]),
            (dictgen::InsensitiveStr::Ascii("idetity"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("endoce"),
                &["endorse", "enforce", "encode"],
            ),
            (dictgen::InsensitiveStr::Ascii("haev"), &["have", "heave"]),
            (dictgen::InsensitiveStr::Ascii("follwoong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentatations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("trhilling"), &["thrilling"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpectidly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("nigthlies"), &["nightlies"]),
            (dictgen::InsensitiveStr::Ascii("requisits"), &["requisites"]),
            (dictgen::InsensitiveStr::Ascii("belevable"), &["believable"]),
            (
                dictgen::InsensitiveStr::Ascii("contanst"),
                &["constants", "contents"],
            ),
            (dictgen::InsensitiveStr::Ascii("exucution"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("otpionally"),
                &["optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpoarts"),
                &["counterparts"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulr"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("dataets"), &["datasets"]),
            (dictgen::InsensitiveStr::Ascii("srite"), &["sprite"]),
            (
                dictgen::InsensitiveStr::Ascii("dericterys"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqeuence"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("priveledges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustrsted"),
                &["frustrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("mariabd"), &["mariadb"]),
            (dictgen::InsensitiveStr::Ascii("wraapping"), &["wrapping"]),
            (dictgen::InsensitiveStr::Ascii("behoviour"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("authencicates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cardridges"),
                &["cartridges"],
            ),
            (dictgen::InsensitiveStr::Ascii("nationales"), &["nationals"]),
            (dictgen::InsensitiveStr::Ascii("interge"), &["integer"]),
            (dictgen::InsensitiveStr::Ascii("neightbour"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("corrspond"), &["correspond"]),
            (dictgen::InsensitiveStr::Ascii("reprorts"), &["reports"]),
            (
                dictgen::InsensitiveStr::Ascii("enforses"),
                &["endorses", "enforces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fustrating"),
                &["frustrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dementieva"), &["dementia"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaes"),
                &["reallocates"],
            ),
            (dictgen::InsensitiveStr::Ascii("resverv"), &["reserve"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomftorable"),
                &["uncomfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destinataions"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnmitter"),
                &["transmitter"],
            ),
            (dictgen::InsensitiveStr::Ascii("disnable"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsultes"),
                &["encapsulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("seived"), &["sieved"]),
            (
                dictgen::InsensitiveStr::Ascii("crsytallization"),
                &["crystallization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitonally"),
                &["optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entrepeneurs"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circunstances"),
                &["circumstances"],
            ),
            (dictgen::InsensitiveStr::Ascii("swimmin"), &["swimming"]),
            (dictgen::InsensitiveStr::Ascii("decsribed"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("fixwd"), &["fixed"]),
            (dictgen::InsensitiveStr::Ascii("steramed"), &["streamed"]),
            (dictgen::InsensitiveStr::Ascii("paranioa"), &["paranoia"]),
            (
                dictgen::InsensitiveStr::Ascii("deterministisch"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catastropic"),
                &["catastrophic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangment"),
                &["arrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("veritiable"),
                &["verifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("comunity"), &["community"]),
            (
                dictgen::InsensitiveStr::Ascii("notigications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missconfigured"),
                &["misconfigured"],
            ),
            (dictgen::InsensitiveStr::Ascii("reloactes"), &["relocates"]),
            (
                dictgen::InsensitiveStr::Ascii("interolates"),
                &["interpolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attentuations"),
                &["attenuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obversations"),
                &["observations"],
            ),
            (dictgen::InsensitiveStr::Ascii("perposal"), &["proposal"]),
            (
                dictgen::InsensitiveStr::Ascii("capitilization"),
                &["capitalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("prfiled"), &["profiled"]),
            (dictgen::InsensitiveStr::Ascii("follwwer"), &["follower"]),
            (dictgen::InsensitiveStr::Ascii("hypenate"), &["hyphenate"]),
            (
                dictgen::InsensitiveStr::Ascii("politiciens"),
                &["politicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterpolations"),
                &["interpolations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architcture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profesionally"),
                &["professionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("catpured"), &["captured"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangt"),
                &["rearrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("leagalize"), &["legalize"]),
            (dictgen::InsensitiveStr::Ascii("benidect"), &["benedict"]),
            (dictgen::InsensitiveStr::Ascii("invaild"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("initailization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("litature"), &["literature"]),
            (
                dictgen::InsensitiveStr::Ascii("maintenamce"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resistable"),
                &["resistible"],
            ),
            (dictgen::InsensitiveStr::Ascii("luanchers"), &["launchers"]),
            (dictgen::InsensitiveStr::Ascii("cherckbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("accidentilly"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infastructure"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaialbale"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproducin"),
                &["reproducing"],
            ),
            (dictgen::InsensitiveStr::Ascii("widgects"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("vlue"), &["value", "blue"]),
            (
                dictgen::InsensitiveStr::Ascii("reporduction"),
                &["reproduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceucted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("expore"),
                &["explore", "expire", "export", "exposure", "expose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjancently"),
                &["adjacently"],
            ),
            (dictgen::InsensitiveStr::Ascii("manditory"), &["mandatory"]),
            (dictgen::InsensitiveStr::Ascii("duplcated"), &["duplicated"]),
            (
                dictgen::InsensitiveStr::Ascii("augmneters"),
                &["augmenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithmnically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specififically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("supirsed"), &["surprised"]),
            (dictgen::InsensitiveStr::Ascii("grafield"), &["garfield"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("jod"), &["job"]),
            (dictgen::InsensitiveStr::Ascii("ciotee"), &["coyote"]),
            (
                dictgen::InsensitiveStr::Ascii("promiscuos"),
                &["promiscuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("longic"), &["logic"]),
            (dictgen::InsensitiveStr::Ascii("caued"), &["caused"]),
            (dictgen::InsensitiveStr::Ascii("footall"), &["football"]),
            (dictgen::InsensitiveStr::Ascii("chanlder"), &["chandler"]),
            (
                dictgen::InsensitiveStr::Ascii("objektives"),
                &["objectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("blievably"), &["believably"]),
            (
                dictgen::InsensitiveStr::Ascii("onventions"),
                &["conventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significence"),
                &["significance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoultions"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crytpocurrency"),
                &["cryptocurrency"],
            ),
            (dictgen::InsensitiveStr::Ascii("exaustion"), &["exhaustion"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhelimg"),
                &["overwhelm"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramether"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("disadvanteges"),
                &["disadvantages"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherint"), &["inherit"]),
            (
                dictgen::InsensitiveStr::Ascii("wakthrough"),
                &["walkthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humerous"),
                &["humorous", "humerus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabulility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("embaress"), &["embarrass"]),
            (dictgen::InsensitiveStr::Ascii("webhools"), &["webhooks"]),
            (dictgen::InsensitiveStr::Ascii("serenitary"), &["serenity"]),
            (
                dictgen::InsensitiveStr::Ascii("apolegetic"),
                &["apologetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoultion"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exchanaging"),
                &["exchanging"],
            ),
            (dictgen::InsensitiveStr::Ascii("othogonal"), &["orthogonal"]),
            (
                dictgen::InsensitiveStr::Ascii("keyboars"),
                &["keyboards", "keyboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traid"),
                &["triad", "trade", "raid"],
            ),
            (dictgen::InsensitiveStr::Ascii("contantly"), &["constantly"]),
            (dictgen::InsensitiveStr::Ascii("puplicly"), &["publicly"]),
            (
                dictgen::InsensitiveStr::Ascii("equivelents"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homeothapy"),
                &["homeopathy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classsifier"),
                &["classifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdifying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("hannbial"), &["hannibal"]),
            (
                dictgen::InsensitiveStr::Ascii("impplementating"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarantenni"),
                &["quarantine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associateion"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumurated"),
                &["enumerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("deinitely"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("cancled"), &["canceled"]),
            (
                dictgen::InsensitiveStr::Ascii("desactivates"),
                &["deactivates"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupliated"), &["duplicated"]),
            (dictgen::InsensitiveStr::Ascii("buildder"), &["builder"]),
            (dictgen::InsensitiveStr::Ascii("canocical"), &["canonical"]),
            (
                dictgen::InsensitiveStr::Ascii("desciminant"),
                &["discriminant"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsuable"), &["unusable"]),
            (dictgen::InsensitiveStr::Ascii("panc"), &["panic"]),
            (dictgen::InsensitiveStr::Ascii("anecdotle"), &["anecdote"]),
            (dictgen::InsensitiveStr::Ascii("derrive"), &["derive"]),
            (
                dictgen::InsensitiveStr::Ascii("responde"),
                &["respond", "response", "responds", "responded", "responder"],
            ),
            (dictgen::InsensitiveStr::Ascii("charactear"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("judgemant"),
                &["judgemental"],
            ),
            (dictgen::InsensitiveStr::Ascii("transprt"), &["transport"]),
            (
                dictgen::InsensitiveStr::Ascii("savin"),
                &["saving", "satin", "sarin"],
            ),
            (dictgen::InsensitiveStr::Ascii("gogether"), &["together"]),
            (dictgen::InsensitiveStr::Ascii("infallibe"), &["infallible"]),
            (
                dictgen::InsensitiveStr::Ascii("infrasctructure"),
                &["infrastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("seeges"), &["sieges"]),
            (dictgen::InsensitiveStr::Ascii("numhers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("gardian"),
                &["guardian", "gradian"],
            ),
            (dictgen::InsensitiveStr::Ascii("itereating"), &["iterating"]),
            (dictgen::InsensitiveStr::Ascii("herselv"), &["herself"]),
            (
                dictgen::InsensitiveStr::Ascii("signficance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("timeput"), &["timeout"]),
            (dictgen::InsensitiveStr::Ascii("typping"), &["typing"]),
            (
                dictgen::InsensitiveStr::Ascii("neightbours"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("visibilty"), &["visibility"]),
            (dictgen::InsensitiveStr::Ascii("maxiums"), &["maximums"]),
            (
                dictgen::InsensitiveStr::Ascii("varifications"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("decesion"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("packagin"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("arror"), &["error", "arrow"]),
            (dictgen::InsensitiveStr::Ascii("experiene"), &["experience"]),
            (dictgen::InsensitiveStr::Ascii("cancells"), &["cancels"]),
            (
                dictgen::InsensitiveStr::Ascii("disbursment"),
                &["disbursement"],
            ),
            (dictgen::InsensitiveStr::Ascii("revelence"), &["relevance"]),
            (dictgen::InsensitiveStr::Ascii("faboulus"), &["fabulous"]),
            (dictgen::InsensitiveStr::Ascii("ancinet"), &["ancient"]),
            (
                dictgen::InsensitiveStr::Ascii("occuptional"),
                &["occupational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tremendious"),
                &["tremendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("canacel"), &["cancel"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeordhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accdiental"),
                &["accidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alghoritms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmarisation"),
                &["summarisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertarien"),
                &["libertarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("connextion"), &["connexion"]),
            (dictgen::InsensitiveStr::Ascii("undfines"), &["undefines"]),
            (dictgen::InsensitiveStr::Ascii("neglectn"), &["neglecting"]),
            (dictgen::InsensitiveStr::Ascii("accsept"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("repspectively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("potical"), &["optical"]),
            (
                dictgen::InsensitiveStr::Ascii("amerliorator"),
                &["ameliorator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("settlemens"),
                &["settlements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immortalisy"),
                &["immortals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troglidites"),
                &["troglodytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("cherckout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("consistenly"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depened"),
                &["depend", "depended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categoizations"),
                &["categorizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coommittees"),
                &["committees"],
            ),
            (dictgen::InsensitiveStr::Ascii("updatwe"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("misconseptions"),
                &["misconceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("proiders"), &["providers"]),
            (
                dictgen::InsensitiveStr::Ascii("genralisation"),
                &["generalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("resemblin"), &["resembling"]),
            (
                dictgen::InsensitiveStr::Ascii("excludet"),
                &["exclude", "excluded"],
            ),
            (dictgen::InsensitiveStr::Ascii("eczecuted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("modifation"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("larvays"), &["larvae"]),
            (
                dictgen::InsensitiveStr::Ascii("photoshooped"),
                &["photoshopped"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgner"), &["designer"]),
            (
                dictgen::InsensitiveStr::Ascii("disovering"),
                &["discovering"],
            ),
            (dictgen::InsensitiveStr::Ascii("blcoker"), &["blocker"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrevations"),
                &["abbreviations"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomimal"), &["nominal"]),
            (
                dictgen::InsensitiveStr::Ascii("surporting"),
                &["supporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apperciates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("threashold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("macack"), &["macaque"]),
            (dictgen::InsensitiveStr::Ascii("surported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("mozila"), &["mozilla"]),
            (dictgen::InsensitiveStr::Ascii("oberving"), &["observing"]),
            (dictgen::InsensitiveStr::Ascii("sucessiv"), &["successive"]),
            (dictgen::InsensitiveStr::Ascii("memomry"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("croshet"), &["crochet"]),
            (
                dictgen::InsensitiveStr::Ascii("continuativo"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("spllitting"), &["splitting"]),
            (dictgen::InsensitiveStr::Ascii("pupulated"), &["populated"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenaions"),
                &["concatenations"],
            ),
            (dictgen::InsensitiveStr::Ascii("renameing"), &["renaming"]),
            (
                dictgen::InsensitiveStr::Ascii("resulution"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthymically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinig"), &["combining"]),
            (dictgen::InsensitiveStr::Ascii("cleck"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("dominantes"),
                &["dominants", "dominates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triffering"),
                &["triggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("condomes"), &["condoms"]),
            (
                dictgen::InsensitiveStr::Ascii("borwser"),
                &["browsers", "browser"],
            ),
            (dictgen::InsensitiveStr::Ascii("traigers"), &["triagers"]),
            (
                dictgen::InsensitiveStr::Ascii("shold"),
                &["should", "hold", "sold"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraylsis"), &["paralysis"]),
            (dictgen::InsensitiveStr::Ascii("increaed"), &["increased"]),
            (dictgen::InsensitiveStr::Ascii("genearte"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("geometrician"),
                &["geometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("availeble"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("resently"), &["recently"]),
            (dictgen::InsensitiveStr::Ascii("recretion"), &["recreation"]),
            (
                dictgen::InsensitiveStr::Ascii("spefation"),
                &["separation", "specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("reenfoces"), &["reenforces"]),
            (dictgen::InsensitiveStr::Ascii("relaease"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("mischevus"),
                &["mischievous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allpications"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("jonts"), &["joints"]),
            (
                dictgen::InsensitiveStr::Ascii("enivornment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exacutables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preformances"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereivatives"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("clearified"), &["clarified"]),
            (
                dictgen::InsensitiveStr::Ascii("webassemby"),
                &["webassembly"],
            ),
            (dictgen::InsensitiveStr::Ascii("breifing"), &["briefing"]),
            (dictgen::InsensitiveStr::Ascii("endoint"), &["endpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("comamnding"),
                &["commanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("movees"),
                &["moves", "movies"],
            ),
            (dictgen::InsensitiveStr::Ascii("objctive"), &["objective"]),
            (dictgen::InsensitiveStr::Ascii("outputtin"), &["outputting"]),
            (dictgen::InsensitiveStr::Ascii("mutilple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("finge"),
                &["finger", "fringe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indiffernece"),
                &["indifference"],
            ),
            (dictgen::InsensitiveStr::Ascii("conveters"), &["converters"]),
            (
                dictgen::InsensitiveStr::Ascii("replentished"),
                &["replenished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multilplicities"),
                &["multiplicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attenation"),
                &["attenuation", "attention"],
            ),
            (dictgen::InsensitiveStr::Ascii("ssee"), &["see"]),
            (
                dictgen::InsensitiveStr::Ascii("perventions"),
                &["preventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("columes"),
                &["columns", "volumes"],
            ),
            (dictgen::InsensitiveStr::Ascii("filpped"), &["flipped"]),
            (dictgen::InsensitiveStr::Ascii("cautios"), &["cautious"]),
            (
                dictgen::InsensitiveStr::Ascii("indistructible"),
                &["indestructible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alterately"),
                &["alternately"],
            ),
            (dictgen::InsensitiveStr::Ascii("decortator"), &["decorator"]),
            (
                dictgen::InsensitiveStr::Ascii("accumilation"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conirm"), &["confirm"]),
            (dictgen::InsensitiveStr::Ascii("scheduls"), &["schedules"]),
            (dictgen::InsensitiveStr::Ascii("instrcut"), &["instruct"]),
            (dictgen::InsensitiveStr::Ascii("apolgise"), &["apologise"]),
            (
                dictgen::InsensitiveStr::Ascii("apprieciative"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recognies"),
                &["recognizes", "recognises"],
            ),
            (dictgen::InsensitiveStr::Ascii("scetches"), &["sketches"]),
            (dictgen::InsensitiveStr::Ascii("sepereate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("diabling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("starnge"), &["strange"]),
            (dictgen::InsensitiveStr::Ascii("conetents"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("physicks"), &["physics"]),
            (
                dictgen::InsensitiveStr::Ascii("proprieter"),
                &["proprietor"],
            ),
            (dictgen::InsensitiveStr::Ascii("syptom"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("wndows"), &["windows"]),
            (
                dictgen::InsensitiveStr::Ascii("disgustiny"),
                &["disgustingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanaiton"),
                &["explanations", "explanation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplaction"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlapsing"),
                &["overlapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("uplader"), &["uploader"]),
            (
                dictgen::InsensitiveStr::Ascii("cummunicate"),
                &["communicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("varilue"),
                &["value", "variable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("scafolds"), &["scaffolds"]),
            (dictgen::InsensitiveStr::Ascii("propetry"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("wpuld"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("clutchign"), &["clutching"]),
            (
                dictgen::InsensitiveStr::Ascii("recommenend"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("savees"), &["saves"]),
            (
                dictgen::InsensitiveStr::Ascii("comlaining"),
                &["complaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arquitecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alforithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivlently"),
                &["equivalently"],
            ),
            (dictgen::InsensitiveStr::Ascii("mashetty"), &["machete"]),
            (dictgen::InsensitiveStr::Ascii("livel"), &["level"]),
            (dictgen::InsensitiveStr::Ascii("mathes"), &["matches"]),
            (dictgen::InsensitiveStr::Ascii("kwuzine"), &["cuisine"]),
            (
                dictgen::InsensitiveStr::Ascii("retransferd"),
                &["retransferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconeects"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("analise"), &["analyse"]),
            (
                dictgen::InsensitiveStr::Ascii("quesetioned"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("relevnt"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("initialze"), &["initialize"]),
            (dictgen::InsensitiveStr::Ascii("checged"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("extenssion"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("pharse"),
                &["phrase", "parse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proficincy"),
                &["proficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifceycle"), &["lifecycle"]),
            (
                dictgen::InsensitiveStr::Ascii("partical"),
                &["particular", "partial", "particle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coallessence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civelezation"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("demaonds"), &["demands"]),
            (dictgen::InsensitiveStr::Ascii("strnad"), &["strand"]),
            (
                dictgen::InsensitiveStr::Ascii("argumentitive"),
                &["argumentative"],
            ),
            (dictgen::InsensitiveStr::Ascii("disablin"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("imporance"), &["importance"]),
            (
                dictgen::InsensitiveStr::Ascii("parallelpipeds"),
                &["parallelepipeds"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaridan"), &["guardian"]),
            (
                dictgen::InsensitiveStr::Ascii("speparating"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("wranings"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("extactor"), &["extractor"]),
            (dictgen::InsensitiveStr::Ascii("confimred"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("wrehouse"), &["warehouse"]),
            (dictgen::InsensitiveStr::Ascii("considred"), &["considered"]),
            (
                dictgen::InsensitiveStr::Ascii("conviciton"),
                &["conviction"],
            ),
            (dictgen::InsensitiveStr::Ascii("lisetned"), &["listened"]),
            (dictgen::InsensitiveStr::Ascii("regstered"), &["registered"]),
            (
                dictgen::InsensitiveStr::Ascii("bartendars"),
                &["bartenders"],
            ),
            (dictgen::InsensitiveStr::Ascii("origianals"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("distinghishes"),
                &["distinguishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapilly"), &["happily"]),
            (
                dictgen::InsensitiveStr::Ascii("deslect"),
                &["deselect", "deflect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expcetation"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("authered"), &["authored"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undertsanding"),
                &["understanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissambler"),
                &["disassembler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("darwn"),
                &["drawn", "darwin"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbiden"), &["forbidden"]),
            (dictgen::InsensitiveStr::Ascii("changues"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("cropses"), &["corpses"]),
            (dictgen::InsensitiveStr::Ascii("ymbolic"), &["symbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("abonimation"),
                &["abomination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shatnering"),
                &["shattering"],
            ),
            (dictgen::InsensitiveStr::Ascii("slighly"), &["slightly"]),
            (dictgen::InsensitiveStr::Ascii("spotifty"), &["spotify"]),
            (
                dictgen::InsensitiveStr::Ascii("pinter"),
                &["pointer", "printer"],
            ),
            (dictgen::InsensitiveStr::Ascii("imfamus"), &["infamous"]),
            (dictgen::InsensitiveStr::Ascii("thred"), &["thread"]),
            (dictgen::InsensitiveStr::Ascii("prefacin"), &["prefacing"]),
            (
                dictgen::InsensitiveStr::Ascii("engagemnts"),
                &["engagements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presmission"),
                &["permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("rumtimes"), &["runtimes"]),
            (
                dictgen::InsensitiveStr::Ascii("cancellato"),
                &["cancellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("canot"), &["cannot"]),
            (dictgen::InsensitiveStr::Ascii("blokchain"), &["blockchain"]),
            (
                dictgen::InsensitiveStr::Ascii("encapuslates"),
                &["encapsulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("plantium"), &["platinum"]),
            (
                dictgen::InsensitiveStr::Ascii("protistants"),
                &["protestants"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispersa"), &["dispensary"]),
            (dictgen::InsensitiveStr::Ascii("anciliary"), &["ancillary"]),
            (
                dictgen::InsensitiveStr::Ascii("authentiction"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("prgrammer"), &["programmer"]),
            (dictgen::InsensitiveStr::Ascii("lentgh"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("recurion"), &["recursion"]),
            (dictgen::InsensitiveStr::Ascii("soucres"), &["sources"]),
            (dictgen::InsensitiveStr::Ascii("givne"), &["given"]),
            (
                dictgen::InsensitiveStr::Ascii("grammaticaal"),
                &["grammatical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalidtes"),
                &["invalidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimitaded"),
                &["intimidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedupliated"),
                &["deduplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bechmarked"),
                &["benchmarked"],
            ),
            (dictgen::InsensitiveStr::Ascii("chineese"), &["chinese"]),
            (dictgen::InsensitiveStr::Ascii("proce"), &["price"]),
            (dictgen::InsensitiveStr::Ascii("decisisons"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("viees"), &["views"]),
            (
                dictgen::InsensitiveStr::Ascii("phycisians"),
                &["physicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comunicating"),
                &["communicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tangensial"),
                &["tangential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concenrtation"),
                &["concentrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pastery"), &["pastry"]),
            (dictgen::InsensitiveStr::Ascii("parmeter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("coexistance"),
                &["coexistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranalating"),
                &["translating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nastersiums"),
                &["nasturtiums"],
            ),
            (dictgen::InsensitiveStr::Ascii("guages"), &["gages"]),
            (dictgen::InsensitiveStr::Ascii("excedes"), &["exceeds"]),
            (
                dictgen::InsensitiveStr::Ascii("runing"),
                &["running", "ruining"],
            ),
            (dictgen::InsensitiveStr::Ascii("progresed"), &["progressed"]),
            (
                dictgen::InsensitiveStr::Ascii("consequente"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasngender"),
                &["transgender"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceept"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("matric"),
                &["metric", "matrix"],
            ),
            (dictgen::InsensitiveStr::Ascii("mystrow"), &["maestro"]),
            (dictgen::InsensitiveStr::Ascii("generares"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("nagatives"), &["negatives"]),
            (
                dictgen::InsensitiveStr::Ascii("recgonizing"),
                &["recognizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contricting"),
                &["contracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("releifes"), &["relieves"]),
            (
                dictgen::InsensitiveStr::Ascii("custimized"),
                &["customized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pateintly"), &["patiently"]),
            (
                dictgen::InsensitiveStr::Ascii("dumplicates"),
                &["duplicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apocalytpic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("iwithout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("ceratin"),
                &["certain", "keratin"],
            ),
            (dictgen::InsensitiveStr::Ascii("implicatia"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("overwite"), &["overwrite"]),
            (dictgen::InsensitiveStr::Ascii("longuer"), &["longer"]),
            (dictgen::InsensitiveStr::Ascii("reqiured"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("hellucinations"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispatchs"), &["dispatches"]),
            (dictgen::InsensitiveStr::Ascii("lanuchers"), &["launchers"]),
            (dictgen::InsensitiveStr::Ascii("diaciritc"), &["diacritic"]),
            (dictgen::InsensitiveStr::Ascii("terimnal"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("insatnces"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("scritps"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("paket"), &["packet"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfication"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("serverity"), &["severity"]),
            (
                dictgen::InsensitiveStr::Ascii("precessing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wated"), &["wanted"]),
            (dictgen::InsensitiveStr::Ascii("cipertext"), &["ciphertext"]),
            (dictgen::InsensitiveStr::Ascii("farcking"), &["fracking"]),
            (dictgen::InsensitiveStr::Ascii("susseeder"), &["seceder"]),
            (
                dictgen::InsensitiveStr::Ascii("victimizin"),
                &["victimizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diect"),
                &["direct", "dissect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("righteouness"),
                &["righteousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correllations"),
                &["correlations"],
            ),
            (dictgen::InsensitiveStr::Ascii("idetical"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("exept"),
                &["except", "exempt"],
            ),
            (dictgen::InsensitiveStr::Ascii("staidum"), &["stadium"]),
            (
                dictgen::InsensitiveStr::Ascii("extropolate"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyclits"), &["cyclist"]),
            (dictgen::InsensitiveStr::Ascii("requeueing"), &["requeuing"]),
            (
                dictgen::InsensitiveStr::Ascii("possiblity"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("liveing"), &["living"]),
            (
                dictgen::InsensitiveStr::Ascii("redusing"),
                &["reducing", "refusing", "reusing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etsablishment"),
                &["establishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("elepant"), &["elephant"]),
            (
                dictgen::InsensitiveStr::Ascii("nutricional"),
                &["nutritional"],
            ),
            (dictgen::InsensitiveStr::Ascii("quuery"), &["query"]),
            (
                dictgen::InsensitiveStr::Ascii("mataphysical"),
                &["metaphysical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semicondutor"),
                &["semiconductor"],
            ),
            (dictgen::InsensitiveStr::Ascii("stament"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("pacjaged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("negotitaors"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retricting"),
                &["restricting", "retracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antibiodic"),
                &["antibiotic"],
            ),
            (dictgen::InsensitiveStr::Ascii("calue"), &["value"]),
            (
                dictgen::InsensitiveStr::Ascii("distnguished"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("admn"), &["admin"]),
            (
                dictgen::InsensitiveStr::Ascii("mede"),
                &["made", "mere", "mead", "lede", "mete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportaes"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanugages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("defaull"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("stero"), &["stereo"]),
            (dictgen::InsensitiveStr::Ascii("falvored"), &["flavored"]),
            (
                dictgen::InsensitiveStr::Ascii("fieds"),
                &["fields", "fiends"],
            ),
            (dictgen::InsensitiveStr::Ascii("improofing"), &["improving"]),
            (
                dictgen::InsensitiveStr::Ascii("bumbed"),
                &["bombed", "bumped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cringeworty"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("coleague"), &["colleague"]),
            (dictgen::InsensitiveStr::Ascii("omision"), &["omission"]),
            (dictgen::InsensitiveStr::Ascii("refures"), &["refuses"]),
            (
                dictgen::InsensitiveStr::Ascii("unistallation"),
                &["uninstallation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfoulding"), &["unfolding"]),
            (
                dictgen::InsensitiveStr::Ascii("expcetions"),
                &["exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gunstinger"),
                &["gunslinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("tatus"), &["status"]),
            (
                dictgen::InsensitiveStr::Ascii("signure"),
                &["signature", "signer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cockaroches"),
                &["cockroaches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inheritage"),
                &["heritage", "inheritance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encumberance"),
                &["encumbrance"],
            ),
            (dictgen::InsensitiveStr::Ascii("secoundly"), &["secondly"]),
            (dictgen::InsensitiveStr::Ascii("bruisend"), &["bruised"]),
            (dictgen::InsensitiveStr::Ascii("securiyt"), &["security"]),
            (dictgen::InsensitiveStr::Unicode("évaluate"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("simplisitc"),
                &["simplistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoindent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assymptotically"),
                &["asymptotically"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendereded"), &["rendered"]),
            (
                dictgen::InsensitiveStr::Ascii("detrimentul"),
                &["detrimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chatacters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocaited"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("finallt"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("taraget"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("cuvres"),
                &["curves", "covers"],
            ),
            (dictgen::InsensitiveStr::Ascii("filting"), &["filtering"]),
            (
                dictgen::InsensitiveStr::Ascii("configurting"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallifed"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("readyness"), &["readiness"]),
            (
                dictgen::InsensitiveStr::Ascii("clcoksource"),
                &["clocksource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workbennches"),
                &["workbenches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheezecake"),
                &["cheesecake"],
            ),
            (dictgen::InsensitiveStr::Ascii("fromates"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("programmate"),
                &["programme"],
            ),
            (dictgen::InsensitiveStr::Ascii("iproving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("onsalught"), &["onslaught"]),
            (
                dictgen::InsensitiveStr::Ascii("sillybuses"),
                &["syllabuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propegations"),
                &["propagations"],
            ),
            (dictgen::InsensitiveStr::Ascii("excitied"), &["excited"]),
            (dictgen::InsensitiveStr::Ascii("returnexd"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("brakeout"), &["breakout"]),
            (
                dictgen::InsensitiveStr::Ascii("rocordings"),
                &["recordings"],
            ),
            (dictgen::InsensitiveStr::Ascii("quieried"), &["queried"]),
            (dictgen::InsensitiveStr::Ascii("noteriety"), &["notoriety"]),
            (
                dictgen::InsensitiveStr::Ascii("desscriptor"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("imbraces"), &["embraces"]),
            (
                dictgen::InsensitiveStr::Ascii("singualrity"),
                &["singularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("develompents"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinematogrophy"),
                &["cinematography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypotheiss"),
                &["hypotheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("counteres"), &["counters"]),
            (dictgen::InsensitiveStr::Ascii("wraapped"), &["wrapped"]),
            (dictgen::InsensitiveStr::Ascii("unmaped"), &["unmapped"]),
            (dictgen::InsensitiveStr::Ascii("eploit"), &["exploit"]),
            (dictgen::InsensitiveStr::Ascii("symbsol"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("quantifiying"),
                &["quantifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("barrer"),
                &["barred", "barrel", "barren", "barrier"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceduler"), &["scheduler"]),
            (
                dictgen::InsensitiveStr::Ascii("linguisics"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("conveniet"), &["convenient"]),
            (
                dictgen::InsensitiveStr::Ascii("naiborhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coventionally"),
                &["conventionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("retunrs"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("sysematic"), &["systematic"]),
            (
                dictgen::InsensitiveStr::Ascii("generatting"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("seprate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("beneifical"),
                &["beneficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("thererin"), &["therein"]),
            (
                dictgen::InsensitiveStr::Ascii("classifiction"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("hweaton"), &["wheaton"]),
            (dictgen::InsensitiveStr::Ascii("someonelse"), &["someones"]),
            (dictgen::InsensitiveStr::Ascii("ignorged"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("cleeshay"), &["cliché"]),
            (
                dictgen::InsensitiveStr::Ascii("replacates"),
                &["replicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("wroking"), &["working"]),
            (dictgen::InsensitiveStr::Ascii("itens"), &["items"]),
            (
                dictgen::InsensitiveStr::Ascii("appretiate"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("artilces"), &["articles"]),
            (
                dictgen::InsensitiveStr::Ascii("patryarchy"),
                &["patriarchy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sord"),
                &["sword", "sore", "sored", "sawed", "soared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applycable"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("teusdays"), &["tuesdays"]),
            (dictgen::InsensitiveStr::Ascii("amensty"), &["amnesty"]),
            (dictgen::InsensitiveStr::Ascii("chosses"), &["chooses"]),
            (dictgen::InsensitiveStr::Ascii("chinense"), &["chinese"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsing"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("suround"), &["surround"]),
            (dictgen::InsensitiveStr::Ascii("selv"), &["self"]),
            (
                dictgen::InsensitiveStr::Ascii("framentation"),
                &["fragmentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("clasifies"), &["classifies"]),
            (dictgen::InsensitiveStr::Ascii("belond"), &["belong"]),
            (
                dictgen::InsensitiveStr::Ascii("interleive"),
                &["interleave"],
            ),
            (dictgen::InsensitiveStr::Ascii("verison"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("normalyzed"),
                &["normalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemententation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gutiarist"), &["guitarist"]),
            (dictgen::InsensitiveStr::Ascii("enmpty"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("acceptabil"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sotrylines"),
                &["storylines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sudmobules"),
                &["submodules"],
            ),
            (dictgen::InsensitiveStr::Ascii("involed"), &["involved"]),
            (
                dictgen::InsensitiveStr::Ascii("mateial"),
                &["material", "metal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dischare"),
                &["discharged", "discharge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consectuve"),
                &["consecutive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaganitive"),
                &["imaginative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confusin"),
                &["confusing", "confusion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefernece"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("htting"), &["hitting"]),
            (dictgen::InsensitiveStr::Ascii("desings"), &["designs"]),
            (dictgen::InsensitiveStr::Ascii("thraed"), &["thread"]),
            (
                dictgen::InsensitiveStr::Ascii("oblitorated"),
                &["obliterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("adrerss"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("trucating"), &["truncating"]),
            (
                dictgen::InsensitiveStr::Ascii("determiend"),
                &["determined"],
            ),
            (dictgen::InsensitiveStr::Ascii("avalance"), &["avalanche"]),
            (dictgen::InsensitiveStr::Ascii("insidde"), &["inside"]),
            (dictgen::InsensitiveStr::Ascii("operattors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("nessessary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("crytpographic"),
                &["cryptographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisin"), &["poison"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbouing"),
                &["neighbouring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supression"),
                &["suppression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomised"),
                &["randomised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservitism"),
                &["conservatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("acticated"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("battlearts"),
                &["battlestar"],
            ),
            (dictgen::InsensitiveStr::Ascii("eastwoord"), &["eastwood"]),
            (
                dictgen::InsensitiveStr::Ascii("convineient"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifides"),
                &["classified"],
            ),
            (dictgen::InsensitiveStr::Ascii("autograh"), &["autograph"]),
            (
                dictgen::InsensitiveStr::Ascii("unregiters"),
                &["unregisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhencement"),
                &["enhancement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefrencable"),
                &["dereferenceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmonts"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categoizing"),
                &["categorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misundersood"),
                &["misunderstood"],
            ),
            (dictgen::InsensitiveStr::Ascii("thuis"), &["this", "thus"]),
            (dictgen::InsensitiveStr::Ascii("puplating"), &["populating"]),
            (
                dictgen::InsensitiveStr::Ascii("atttribution"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obsolutists"),
                &["absolutists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orignating"),
                &["originating"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellingotn"), &["ellington"]),
            (dictgen::InsensitiveStr::Ascii("activastes"), &["activates"]),
            (dictgen::InsensitiveStr::Ascii("relevane"), &["relevant"]),
            (
                dictgen::InsensitiveStr::Ascii("obsloescence"),
                &["obsolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("dscriptor"), &["descriptor"]),
            (
                dictgen::InsensitiveStr::Ascii("reuire"),
                &["require", "rewire"],
            ),
            (dictgen::InsensitiveStr::Ascii("anytning"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("souch"),
                &["pouch", "sough", "such", "touch"],
            ),
            (dictgen::InsensitiveStr::Ascii("housand"), &["thousand"]),
            (
                dictgen::InsensitiveStr::Ascii("equirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dosin"),
                &["dosing", "rosin"],
            ),
            (dictgen::InsensitiveStr::Ascii("triators"), &["traitors"]),
            (
                dictgen::InsensitiveStr::Ascii("brakets"),
                &["brackets", "brakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitutie"),
                &["constitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preinitalize"),
                &["preinitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encordes"),
                &["encodes", "encores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conventient"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrputible"),
                &["interruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("supoosing"), &["supposing"]),
            (dictgen::InsensitiveStr::Ascii("proprty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("intuitivno"), &["intuition"]),
            (
                dictgen::InsensitiveStr::Ascii("corispondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("adresee"), &["addressee"]),
            (
                dictgen::InsensitiveStr::Ascii("internettes"),
                &["internets"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgainze"), &["organize"]),
            (
                dictgen::InsensitiveStr::Ascii("pronouciation"),
                &["pronunciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheneans"), &["athenians"]),
            (
                dictgen::InsensitiveStr::Ascii("substantivly"),
                &["substantially", "substantively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comissioned"),
                &["commissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rererences"),
                &["references", "reverences"],
            ),
            (dictgen::InsensitiveStr::Ascii("figher"), &["fighter"]),
            (dictgen::InsensitiveStr::Ascii("custons"), &["customs"]),
            (
                dictgen::InsensitiveStr::Ascii("publichers"),
                &["publishers"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavivor"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("triscaidekaphobia"),
                &["triskaidekaphobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintelligable"),
                &["unintelligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumalators"),
                &["accumulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compositong"),
                &["compositing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fiels"),
                &["fields", "feels", "files", "phials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intensites"),
                &["intensities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demenstration"),
                &["demonstration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persentile"),
                &["percentile"],
            ),
            (dictgen::InsensitiveStr::Ascii("specfify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("neturon"), &["neutron"]),
            (
                dictgen::InsensitiveStr::Ascii("politiness"),
                &["politeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("assoiate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("surviellance"),
                &["surveillance"],
            ),
            (dictgen::InsensitiveStr::Ascii("evangers"), &["avengers"]),
            (
                dictgen::InsensitiveStr::Ascii("singal"),
                &["signal", "single"],
            ),
            (dictgen::InsensitiveStr::Ascii("aperatures"), &["apertures"]),
            (
                dictgen::InsensitiveStr::Ascii("contaiminating"),
                &["contaminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("definiate"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("tecnicians"),
                &["technicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("variabel"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("wardobe"), &["wardrobe"]),
            (
                dictgen::InsensitiveStr::Ascii("misygonist"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("craeted"),
                &["created", "crated"],
            ),
            (dictgen::InsensitiveStr::Ascii("minut"), &["minute"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelization"),
                &["parallelization"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemted"), &["attempted"]),
            (dictgen::InsensitiveStr::Ascii("baed"), &["based"]),
            (dictgen::InsensitiveStr::Ascii("unusally"), &["unusually"]),
            (dictgen::InsensitiveStr::Ascii("realitime"), &["realtime"]),
            (dictgen::InsensitiveStr::Ascii("mormonts"), &["mormons"]),
            (dictgen::InsensitiveStr::Ascii("dicount"), &["discount"]),
            (
                dictgen::InsensitiveStr::Ascii("notitications"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogorithmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("babysiting"),
                &["babysitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techinican"),
                &["technician"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrants"), &["arrangements"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficifeid"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registeing"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurize"),
                &["pasteurize"],
            ),
            (dictgen::InsensitiveStr::Ascii("overclcok"), &["overclock"]),
            (
                dictgen::InsensitiveStr::Ascii("comparetor"),
                &["comparator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("portabilty"),
                &["portability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorporare"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exitsted"), &["existed"]),
            (
                dictgen::InsensitiveStr::Ascii("caracterized"),
                &["characterized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propects"),
                &["projects", "prospects", "protects"],
            ),
            (dictgen::InsensitiveStr::Ascii("disalbing"), &["disabling"]),
            (
                dictgen::InsensitiveStr::Ascii("fundemantalists"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colonizacion"),
                &["colonization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsriptions"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiatiator"),
                &["initiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("arkasnas"), &["arkansas"]),
            (dictgen::InsensitiveStr::Ascii("legalizin"), &["legalizing"]),
            (
                dictgen::InsensitiveStr::Ascii("inapproprate"),
                &["inappropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roundaboot"),
                &["roundabout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuratiens"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("deveop"), &["develop"]),
            (
                dictgen::InsensitiveStr::Ascii("minneaplis"),
                &["minneapolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regsitered"),
                &["registered"],
            ),
            (dictgen::InsensitiveStr::Ascii("evition"), &["eviction"]),
            (dictgen::InsensitiveStr::Ascii("elseware"), &["elsewhere"]),
            (dictgen::InsensitiveStr::Ascii("footnoes"), &["footnotes"]),
            (
                dictgen::InsensitiveStr::Ascii("contraitns"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repricussions"),
                &["repercussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alchoolism"),
                &["alcoholism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffuculties"),
                &["difficulties"],
            ),
            (dictgen::InsensitiveStr::Ascii("inteferes"), &["interferes"]),
            (
                dictgen::InsensitiveStr::Ascii("initialysed"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brakethrough"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeator"), &["operator"]),
            (
                dictgen::InsensitiveStr::Ascii("indentification"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("metldown"), &["meltdown"]),
            (
                dictgen::InsensitiveStr::Ascii("comporessed"),
                &["compressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atronomical"),
                &["agronomical", "astronomical"],
            ),
            (dictgen::InsensitiveStr::Ascii("abosulute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("diretive"), &["directive"]),
            (dictgen::InsensitiveStr::Ascii("alsso"), &["also"]),
            (dictgen::InsensitiveStr::Ascii("resrved"), &["reserved"]),
            (
                dictgen::InsensitiveStr::Ascii("immeditely"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("brillant"), &["brilliant"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confirmaiton"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("acii"), &["ascii"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotors"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceeling"), &["ceiling"]),
            (
                dictgen::InsensitiveStr::Ascii("judgementsl"),
                &["judgemental"],
            ),
            (dictgen::InsensitiveStr::Ascii("infromers"), &["informers"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapointments"),
                &["disappointments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compagnons"),
                &["companions"],
            ),
            (dictgen::InsensitiveStr::Ascii("inidices"), &["indices"]),
            (
                dictgen::InsensitiveStr::Ascii("hipotheses"),
                &["hypotheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adresable"),
                &["addressable"],
            ),
            (dictgen::InsensitiveStr::Ascii("experss"), &["express"]),
            (dictgen::InsensitiveStr::Ascii("alph"), &["alpha"]),
            (dictgen::InsensitiveStr::Ascii("looknig"), &["looking"]),
            (
                dictgen::InsensitiveStr::Ascii("authenenticating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("commneted"), &["commented"]),
            (
                dictgen::InsensitiveStr::Ascii("aynchronously"),
                &["synchronously", "asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("edficient"), &["deficient"]),
            (
                dictgen::InsensitiveStr::Ascii("collocweally"),
                &["colloquially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminatie"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("owerflows"), &["overflows"]),
            (
                dictgen::InsensitiveStr::Ascii("satuaration"),
                &["saturation"],
            ),
            (dictgen::InsensitiveStr::Ascii("simimlarly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatibilites"),
                &["incompatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("shfiter"), &["shifter"]),
            (dictgen::InsensitiveStr::Ascii("repellin"), &["repelling"]),
            (
                dictgen::InsensitiveStr::Ascii("participe"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("needlees"), &["needles"]),
            (
                dictgen::InsensitiveStr::Ascii("acusations"),
                &["accusations", "causations"],
            ),
            (dictgen::InsensitiveStr::Ascii("previousl"), &["previously"]),
            (
                dictgen::InsensitiveStr::Ascii("commentery"),
                &["commentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("iten"), &["item"]),
            (
                dictgen::InsensitiveStr::Ascii("experimtenally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("americanss"),
                &["americas", "americans"],
            ),
            (dictgen::InsensitiveStr::Ascii("douchely"), &["douchey"]),
            (dictgen::InsensitiveStr::Ascii("tooptips"), &["tooltips"]),
            (dictgen::InsensitiveStr::Ascii("decidate"), &["dedicate"]),
            (
                dictgen::InsensitiveStr::Ascii("intercahnge"),
                &["interchange"],
            ),
            (dictgen::InsensitiveStr::Ascii("populer"), &["popular"]),
            (
                dictgen::InsensitiveStr::Ascii("initializtion"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissconect"),
                &["disconnect"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylser"), &["analyser"]),
            (dictgen::InsensitiveStr::Ascii("execuate"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatition"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifciation"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hetrogenous"),
                &["heterogenous", "heterogeneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opportunitites"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interchangabley"),
                &["interchangeably"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentece"), &["sentence"]),
            (dictgen::InsensitiveStr::Ascii("contur"), &["contour"]),
            (dictgen::InsensitiveStr::Ascii("inverions"), &["inversions"]),
            (
                dictgen::InsensitiveStr::Ascii("facilitiate"),
                &["facilitate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unpredictablity"),
                &["unpredictability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expolsives"),
                &["explosives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exisint"),
                &["existing", "existent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconventionnal"),
                &["unconventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrationel"),
                &["irrational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelelctuals"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("baiscally"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("decodeing"), &["decoding"]),
            (
                dictgen::InsensitiveStr::Ascii("percentange"),
                &["percentage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stablilized"),
                &["stabilized"],
            ),
            (dictgen::InsensitiveStr::Ascii("mehtod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("croporation"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aleniating"),
                &["alienating"],
            ),
            (dictgen::InsensitiveStr::Ascii("stateman"), &["statesman"]),
            (
                dictgen::InsensitiveStr::Ascii("tempereatures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangs"), &["arranges"]),
            (
                dictgen::InsensitiveStr::Ascii("parkays"),
                &["parkways", "parlays", "parquets"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuipder"), &["stupider"]),
            (dictgen::InsensitiveStr::Ascii("bacic"), &["basic"]),
            (dictgen::InsensitiveStr::Ascii("obmitted"), &["omitted"]),
            (dictgen::InsensitiveStr::Ascii("serch"), &["search"]),
            (dictgen::InsensitiveStr::Ascii("conveyer"), &["conveyor"]),
            (
                dictgen::InsensitiveStr::Ascii("assocication"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experesion"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureuacracy"),
                &["bureaucracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchamarking"),
                &["benchmarking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevetible"),
                &["inevitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substanser"),
                &["substances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ttests"),
                &["tests", "attests"],
            ),
            (dictgen::InsensitiveStr::Ascii("digitial"), &["digital"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchronly"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("thhink"), &["think"]),
            (
                dictgen::InsensitiveStr::Ascii("pluging"),
                &["plugging", "plugin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civelizations"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertiing"),
                &["advertising"],
            ),
            (dictgen::InsensitiveStr::Ascii("reowrked"), &["reworked"]),
            (
                dictgen::InsensitiveStr::Ascii("independantly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("weakenend"), &["weakened"]),
            (
                dictgen::InsensitiveStr::Ascii("blances"),
                &["balances", "glances", "lances"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakcets"), &["packets"]),
            (
                dictgen::InsensitiveStr::Ascii("interpeting"),
                &["interpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chanche"),
                &["chance", "change"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxiting"), &["exciting"]),
            (dictgen::InsensitiveStr::Ascii("igrnore"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("tyoing"),
                &["typing", "toying", "tying", "toeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("twon"), &["town"]),
            (dictgen::InsensitiveStr::Ascii("hostiliy"), &["hostility"]),
            (
                dictgen::InsensitiveStr::Ascii("stroing"),
                &["storing", "string", "strong", "stroking"],
            ),
            (dictgen::InsensitiveStr::Ascii("goilath"), &["goliath"]),
            (dictgen::InsensitiveStr::Ascii("inclinde"), &["incline"]),
            (
                dictgen::InsensitiveStr::Ascii("synchonizing"),
                &["synchronizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exepectedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufatures"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expept"),
                &["expect", "except"],
            ),
            (dictgen::InsensitiveStr::Ascii("wigdet"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("seearch"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("contraccion"),
                &["contraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("secirity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("permited"), &["permitted"]),
            (dictgen::InsensitiveStr::Ascii("saferi"), &["safari"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplebgs"),
                &["multiples"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imlpementation"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("woudld"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("crosshar"), &["crosshair"]),
            (dictgen::InsensitiveStr::Ascii("csses"), &["cases"]),
            (
                dictgen::InsensitiveStr::Ascii("pascheurises"),
                &["pasteurises"],
            ),
            (dictgen::InsensitiveStr::Ascii("existance"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("embarressed"),
                &["embarrassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("inrerface"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("practies"), &["practise"]),
            (dictgen::InsensitiveStr::Ascii("redfined"), &["refined"]),
            (
                dictgen::InsensitiveStr::Ascii("fllowing"),
                &["following", "flowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimizies"), &["optimizes"]),
            (dictgen::InsensitiveStr::Ascii("desabling"), &["disabling"]),
            (
                dictgen::InsensitiveStr::Ascii("congraturations"),
                &["congratulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("authoised"), &["authorised"]),
            (dictgen::InsensitiveStr::Ascii("istead"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("apearances"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subesquent"),
                &["subsequent"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertubates"), &["perturbs"]),
            (
                dictgen::InsensitiveStr::Ascii("charactarized"),
                &["characterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("phython"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("refinmenet"),
                &["refinement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underfowed"),
                &["underflowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("soluation"), &["solution"]),
            (dictgen::InsensitiveStr::Ascii("paritals"), &["partials"]),
            (
                dictgen::InsensitiveStr::Ascii("tracsoders"),
                &["transcoders"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyworkds"), &["keywords"]),
            (dictgen::InsensitiveStr::Ascii("clossions"), &["collisions"]),
            (
                dictgen::InsensitiveStr::Ascii("recommentations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tendacy"), &["tendency"]),
            (
                dictgen::InsensitiveStr::Ascii("unstalls"),
                &["installs", "uninstalls"],
            ),
            (dictgen::InsensitiveStr::Ascii("memcahe"), &["memcache"]),
            (dictgen::InsensitiveStr::Ascii("imroves"), &["improves"]),
            (
                dictgen::InsensitiveStr::Ascii("transferrring"),
                &["transferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrecgonized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("dashbord"), &["dashboard"]),
            (dictgen::InsensitiveStr::Ascii("knowldge"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("rrror"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("termnator"), &["terminator"]),
            (dictgen::InsensitiveStr::Ascii("evluative"), &["evaluative"]),
            (
                dictgen::InsensitiveStr::Ascii("deliberetly"),
                &["deliberately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gauranteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degres"),
                &["degrees", "digress"],
            ),
            (dictgen::InsensitiveStr::Ascii("firmaware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("alwyas"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("chnages"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("componeents"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("poited"), &["pointed"]),
            (
                dictgen::InsensitiveStr::Ascii("breakpiont"),
                &["breakpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("sezuires"), &["seizures"]),
            (
                dictgen::InsensitiveStr::Ascii("calculaution"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realiability"),
                &["reliability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiallizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neveloping"),
                &["enveloping"],
            ),
            (dictgen::InsensitiveStr::Ascii("colonge"), &["cologne"]),
            (
                dictgen::InsensitiveStr::Ascii("requestying"),
                &["requesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("chagned"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("encoeded"), &["encoded"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotionator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("exibit"), &["exhibit"]),
            (dictgen::InsensitiveStr::Ascii("aproves"), &["approves"]),
            (
                dictgen::InsensitiveStr::Ascii("mutlinational"),
                &["multinational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciliazation"),
                &["specialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differntial"),
                &["differential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annonations"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("eletrical"), &["electrical"]),
            (dictgen::InsensitiveStr::Ascii("regitries"), &["registries"]),
            (dictgen::InsensitiveStr::Ascii("notse"), &["notes", "note"]),
            (
                dictgen::InsensitiveStr::Ascii("enountering"),
                &["encountering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playhtroughs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("insurasnce"), &["insurance"]),
            (
                dictgen::InsensitiveStr::Ascii("transparany"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ventillating"),
                &["ventilating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stripted"),
                &["scripted", "stripped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("launhed"),
                &["launched", "laughed"],
            ),
            (dictgen::InsensitiveStr::Ascii("hseldon"), &["sheldon"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretted"),
                &["misinterpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instumentation"),
                &["instrumentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanisims"),
                &["mechanism"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepeartes"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("stylessheet"),
                &["stylesheet", "stylesheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("richochets"), &["ricochets"]),
            (
                dictgen::InsensitiveStr::Ascii("ideolagies"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("noticalbe"), &["noticeable"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterept"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("losslesly"), &["losslessly"]),
            (dictgen::InsensitiveStr::Ascii("downlad"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("postincremend"),
                &["postincrement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirctor"), &["director"]),
            (
                dictgen::InsensitiveStr::Ascii("congfiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tchnically"),
                &["technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preliferation"),
                &["proliferation"],
            ),
            (dictgen::InsensitiveStr::Ascii("uderstood"), &["understood"]),
            (dictgen::InsensitiveStr::Ascii("midfeild"), &["midfield"]),
            (
                dictgen::InsensitiveStr::Ascii("thereotically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceptibly"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referenence"),
                &["reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("headong"), &["heading"]),
            (
                dictgen::InsensitiveStr::Ascii("neckbeardies"),
                &["neckbeards"],
            ),
            (dictgen::InsensitiveStr::Ascii("containd"), &["contained"]),
            (dictgen::InsensitiveStr::Ascii("annnotate"), &["annotate"]),
            (
                dictgen::InsensitiveStr::Ascii("livestrems"),
                &["livestream"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenselss"),
                &["defenseless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("querstioning"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccuraties"),
                &["inaccuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializated"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("trotski"), &["trotsky"]),
            (dictgen::InsensitiveStr::Ascii("diplays"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("apporoval"), &["approval"]),
            (
                dictgen::InsensitiveStr::Ascii("intolerante"),
                &["intolerance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrogativo"),
                &["interrogation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dumbbellls"), &["dumbbells"]),
            (dictgen::InsensitiveStr::Ascii("implemt"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("artbitrary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("deactivatiion"),
                &["deactivation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innacuracies"),
                &["inaccuracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("curso"), &["cursor"]),
            (
                dictgen::InsensitiveStr::Ascii("briteners"),
                &["brighteners"],
            ),
            (dictgen::InsensitiveStr::Ascii("woithin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("catacyslm"), &["cataclysm"]),
            (dictgen::InsensitiveStr::Ascii("obeserved"), &["observed"]),
            (dictgen::InsensitiveStr::Ascii("mentioend"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("inverals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("warpping"),
                &["warping", "wrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaialble"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("medicallly"), &["medically"]),
            (
                dictgen::InsensitiveStr::Ascii("essentails"),
                &["essentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("pahntom"), &["phantom"]),
            (
                dictgen::InsensitiveStr::Ascii("compenstation"),
                &["compensation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleateness"),
                &["completeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggergation"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arcitechtural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partizipation"),
                &["participation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalises"),
                &["orthogonalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("ziped"), &["zipped"]),
            (dictgen::InsensitiveStr::Ascii("sandstom"), &["sandstorm"]),
            (
                dictgen::InsensitiveStr::Ascii("canoicalized"),
                &["canonicalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomeant"),
                &["performant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sandwitching"),
                &["sandwiching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambidexterous"),
                &["ambidextrous"],
            ),
            (dictgen::InsensitiveStr::Ascii("topogy"), &["topology"]),
            (
                dictgen::InsensitiveStr::Ascii("discrepany"),
                &["discrepancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaoning"),
                &["reasoning", "rezoning"],
            ),
            (dictgen::InsensitiveStr::Ascii("butterey"), &["buttery"]),
            (dictgen::InsensitiveStr::Ascii("spoliers"), &["spoilers"]),
            (
                dictgen::InsensitiveStr::Ascii("misellaneously"),
                &["miscellaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("nahsville"), &["nashville"]),
            (dictgen::InsensitiveStr::Ascii("accepably"), &["acceptably"]),
            (
                dictgen::InsensitiveStr::Ascii("deveolpments"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("wednessay"), &["wednesdays"]),
            (dictgen::InsensitiveStr::Ascii("agreably"), &["agreeably"]),
            (dictgen::InsensitiveStr::Ascii("catgory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("timzeone"), &["timezone"]),
            (
                dictgen::InsensitiveStr::Ascii("centerfuge"),
                &["centrifuge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolulionary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("unqoted"), &["unquoted"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infintrator"),
                &["infiltrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissitc"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("simlation"), &["simulation"]),
            (
                dictgen::InsensitiveStr::Ascii("conectivity"),
                &["connectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("averloaded"),
                &["overloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilse"), &["utilise"]),
            (dictgen::InsensitiveStr::Ascii("amking"), &["making"]),
            (
                dictgen::InsensitiveStr::Ascii("konfrontation"),
                &["confrontation"],
            ),
            (dictgen::InsensitiveStr::Ascii("decocdes"), &["decodes"]),
            (
                dictgen::InsensitiveStr::Ascii("authethenticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("reutrns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("literatre"), &["literate"]),
            (
                dictgen::InsensitiveStr::Ascii("resolutoin"),
                &["resolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chunaks"), &["chunks"]),
            (dictgen::InsensitiveStr::Ascii("sucseed"), &["succeed"]),
            (
                dictgen::InsensitiveStr::Ascii("exclamanation"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("susseeding"), &["seceding"]),
            (
                dictgen::InsensitiveStr::Ascii("afforadble"),
                &["affordable"],
            ),
            (dictgen::InsensitiveStr::Ascii("religeons"), &["religions"]),
            (dictgen::InsensitiveStr::Ascii("puplate"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("girlfreind"),
                &["girlfriend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converstations"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phoneticly"),
                &["phonetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotatiation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwyays"), &["anyways"]),
            (
                dictgen::InsensitiveStr::Ascii("infrequentcies"),
                &["infrequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("organizms"), &["organism"]),
            (
                dictgen::InsensitiveStr::Ascii("perpedicularly"),
                &["perpendicularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contanimate"),
                &["contaminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("assemple"), &["assemble"]),
            (dictgen::InsensitiveStr::Ascii("currnent"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamnt"),
                &["performant"],
            ),
            (dictgen::InsensitiveStr::Ascii("resitance"), &["resistance"]),
            (dictgen::InsensitiveStr::Ascii("prescene"), &["presence"]),
            (dictgen::InsensitiveStr::Ascii("chizelles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("intruduced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreprener"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contatcting"),
                &["contacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("diretcx"), &["directx"]),
            (dictgen::InsensitiveStr::Ascii("duplicted"), &["duplicated"]),
            (
                dictgen::InsensitiveStr::Ascii("proffesion"),
                &["profession"],
            ),
            (dictgen::InsensitiveStr::Ascii("requrired"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("postion"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("perimenter"), &["perimeter"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnmissions"),
                &["transmissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reallocaition"),
                &["reallocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongises"),
                &["recognises"],
            ),
            (dictgen::InsensitiveStr::Ascii("execessive"), &["excessive"]),
            (
                dictgen::InsensitiveStr::Ascii("incorpore"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("attributei"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("renderning"), &["rendering"]),
            (dictgen::InsensitiveStr::Ascii("usring"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("aattributes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("displaed"),
                &["displaced", "displayed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilated"),
                &["compiled", "complicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("remmber"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("effexts"), &["effects"]),
            (
                dictgen::InsensitiveStr::Ascii("rebuidling"),
                &["rebuilding"],
            ),
            (dictgen::InsensitiveStr::Ascii("appareal"), &["apparel"]),
            (
                dictgen::InsensitiveStr::Ascii("defintions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("activting"), &["activating"]),
            (dictgen::InsensitiveStr::Ascii("disbale"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("incrmenetd"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refromatting"),
                &["refomatting", "reformatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("resultion"), &["resolution"]),
            (
                dictgen::InsensitiveStr::Ascii("dimesnional"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("doed"), &["does"]),
            (
                dictgen::InsensitiveStr::Ascii("declations"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprectats"),
                &["deprecates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalulating"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cusomisers"),
                &["customisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("transtorm"), &["transform"]),
            (
                dictgen::InsensitiveStr::Ascii("optmizations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnoticable"),
                &["unnoticeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciphr"), &["cipher"]),
            (dictgen::InsensitiveStr::Ascii("failng"), &["failing"]),
            (dictgen::InsensitiveStr::Ascii("zoocheini"), &["zucchini"]),
            (
                dictgen::InsensitiveStr::Ascii("intendations"),
                &["indentations", "intentions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciepents"),
                &["recipients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuiliti"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorisin"),
                &["authorising"],
            ),
            (dictgen::InsensitiveStr::Ascii("stabality"), &["stability"]),
            (
                dictgen::InsensitiveStr::Ascii("reactionnary"),
                &["reactionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("trackling"), &["tracking"]),
            (dictgen::InsensitiveStr::Ascii("sunggle"), &["snuggle"]),
            (
                dictgen::InsensitiveStr::Ascii("reaasigned"),
                &["reassigned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victomized"),
                &["victimized"],
            ),
            (dictgen::InsensitiveStr::Ascii("typicallly"), &["typically"]),
            (dictgen::InsensitiveStr::Ascii("registery"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("imcomming"), &["incoming"]),
            (dictgen::InsensitiveStr::Ascii("pacakaged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("denpendently"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segemnting"),
                &["segmenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metablosim"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifyied"), &["verified"]),
            (dictgen::InsensitiveStr::Ascii("usue"), &["use"]),
            (dictgen::InsensitiveStr::Ascii("speficiies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("clened"), &["cleaned"]),
            (dictgen::InsensitiveStr::Ascii("tooks"), &["takes", "took"]),
            (dictgen::InsensitiveStr::Ascii("mysef"), &["myself"]),
            (
                dictgen::InsensitiveStr::Ascii("apprecaiting"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgorunds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaliate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("succceed"), &["succeed"]),
            (dictgen::InsensitiveStr::Ascii("tranfers"), &["transfers"]),
            (
                dictgen::InsensitiveStr::Ascii("renegothiations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convigurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serimonies"),
                &["ceremonies"],
            ),
            (dictgen::InsensitiveStr::Ascii("hagasses"), &["haggises"]),
            (
                dictgen::InsensitiveStr::Ascii("acidentally"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("extistence"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("zimbabew"), &["zimbabwe"]),
            (dictgen::InsensitiveStr::Ascii("onchage"), &["onchange"]),
            (dictgen::InsensitiveStr::Ascii("cretates"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("discconection"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mirometers"),
                &["micrometers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auotmation"),
                &["automation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expewrimenter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("populatin"),
                &["populations", "population", "populating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questonnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepndences"),
                &["independences"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrun"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("exteremes"), &["extremes"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuitor"),
                &["contributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partitionns"),
                &["partitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("dereives"), &["derives"]),
            (dictgen::InsensitiveStr::Ascii("immuniy"), &["immunity"]),
            (
                dictgen::InsensitiveStr::Ascii("knowlegeabel"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhistance"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("appartments"),
                &["apartments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefernece"),
                &["preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horziontal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("balacing"), &["balancing"]),
            (
                dictgen::InsensitiveStr::Ascii("spacifiers"),
                &["specifiers", "pacifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignemtns"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tralled"),
                &[
                    "trailed",
                    "travelled",
                    "trawled",
                    "trialled",
                    "trilled",
                    "trolled",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clase"),
                &["clause", "class"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architecturial"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intializing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caricaturize"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protagonistas"),
                &["protagonists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipiers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bargin"), &["bargain"]),
            (
                dictgen::InsensitiveStr::Ascii("develepors"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accourdingly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitency"),
                &["consistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("contoler"), &["controller"]),
            (dictgen::InsensitiveStr::Ascii("stetes"), &["states"]),
            (
                dictgen::InsensitiveStr::Ascii("possestions"),
                &["possessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("existane"), &["existence"]),
            (dictgen::InsensitiveStr::Ascii("retutning"), &["returning"]),
            (
                dictgen::InsensitiveStr::Ascii("diffrential"),
                &["differential"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqaurium"), &["aquarium"]),
            (
                dictgen::InsensitiveStr::Ascii("cotribution"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("asymmectry"), &["asymmetry"]),
            (dictgen::InsensitiveStr::Ascii("doubes"), &["doubles"]),
            (dictgen::InsensitiveStr::Ascii("offcially"), &["officially"]),
            (dictgen::InsensitiveStr::Ascii("shoud"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("cathegorising"),
                &["categorising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unorignial"),
                &["unoriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootrapped"),
                &["bootstrapped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrition"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("politicus"), &["politics"]),
            (
                dictgen::InsensitiveStr::Ascii("protastant"),
                &["protestant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defishantly"),
                &["deficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("unquotted"), &["unquoted"]),
            (dictgen::InsensitiveStr::Ascii("firwmare"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("suppoted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("reagrded"), &["regarded"]),
            (dictgen::InsensitiveStr::Ascii("mananager"), &["manager"]),
            (dictgen::InsensitiveStr::Ascii("exptected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabillities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standarised"),
                &["standardised"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimisin"), &["optimising"]),
            (
                dictgen::InsensitiveStr::Ascii("dozin"),
                &["dozing", "dozen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveninetly"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialzie"),
                &["deserialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituer"),
                &["prostitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("publisged"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("comopnent"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("separete"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("turnapound"),
                &["turnaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistecy"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescirption"),
                &["prescriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribtues"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optomization"),
                &["optimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("ulimited"), &["unlimited"]),
            (dictgen::InsensitiveStr::Ascii("follwong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("traled"),
                &[
                    "traced", "traded", "trailed", "traveled", "trawled", "trialed",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("iteract"), &["interact"]),
            (
                dictgen::InsensitiveStr::Ascii("professin"),
                &["professions"],
            ),
            (dictgen::InsensitiveStr::Ascii("memick"), &["mimic"]),
            (
                dictgen::InsensitiveStr::Ascii("legitamicy"),
                &["legitimacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparetive"),
                &["comparative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restatting"),
                &["restarting", "restating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsedized"),
                &["subsidized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architctures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("papaer"), &["paper"]),
            (
                dictgen::InsensitiveStr::Ascii("reursively"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("orangers"), &["oranges"]),
            (
                dictgen::InsensitiveStr::Ascii("hyperlobic"),
                &["hyperbolic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denomitators"),
                &["denominators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carrers"),
                &["careers", "carriers"],
            ),
            (dictgen::InsensitiveStr::Ascii("secondray"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("nationaly"), &["nationally"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitvity"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("termplate"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("chanceller"),
                &["chancellor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protectons"),
                &["protectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aymptotical"),
                &["asymptotical"],
            ),
            (dictgen::InsensitiveStr::Ascii("affter"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("autogroping"),
                &["autogrouping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timin"),
                &["timing", "timid"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrinsics"),
                &["intrinsics"],
            ),
            (dictgen::InsensitiveStr::Ascii("builings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("outpreforms"),
                &["outperform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circluating"),
                &["circulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notced"),
                &["noticed", "notched", "noted"],
            ),
            (dictgen::InsensitiveStr::Ascii("assisance"), &["assistance"]),
            (
                dictgen::InsensitiveStr::Ascii("disccussing"),
                &["discussing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cannabil"), &["cannibal"]),
            (dictgen::InsensitiveStr::Ascii("destuctor"), &["destructor"]),
            (dictgen::InsensitiveStr::Ascii("deivices"), &["devices"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("delateable"), &["deletable"]),
            (dictgen::InsensitiveStr::Ascii("chuncking"), &["chunking"]),
            (dictgen::InsensitiveStr::Ascii("situtaion"), &["situation"]),
            (dictgen::InsensitiveStr::Ascii("cartdridge"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("capaticitors"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgruond"),
                &["backgrounds", "background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithimcally"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functonally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemoraging"),
                &["hemorrhaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("virgings"), &["virgins"]),
            (dictgen::InsensitiveStr::Ascii("larington"), &["arlington"]),
            (
                dictgen::InsensitiveStr::Ascii("chipers"),
                &["ciphers", "chippers", "chimers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eligeblity"),
                &["eligibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("refcounf"), &["refcount"]),
            (
                dictgen::InsensitiveStr::Ascii("watermeleon"),
                &["watermelon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenanagins"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbution"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terretorial"),
                &["territorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("claculates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspisiously"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficilleid"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomenona"),
                &["phenomena"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collcetion"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassemblin"),
                &["disassembling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitmistically"),
                &["optimistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("threding"), &["threading"]),
            (dictgen::InsensitiveStr::Ascii("redudancy"), &["redundancy"]),
            (
                dictgen::InsensitiveStr::Ascii("conttributes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bookkeping"),
                &["bookkeeping"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteating"), &["iterating"]),
            (dictgen::InsensitiveStr::Ascii("asumption"), &["assumption"]),
            (
                dictgen::InsensitiveStr::Ascii("convienence"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behaviorly"),
                &["behavioral"],
            ),
            (dictgen::InsensitiveStr::Ascii("macchines"), &["machines"]),
            (
                dictgen::InsensitiveStr::Ascii("shakepseare"),
                &["shakespeare"],
            ),
            (dictgen::InsensitiveStr::Ascii("vershins"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("sombody"), &["somebody"]),
            (dictgen::InsensitiveStr::Ascii("coersce"), &["coerce"]),
            (dictgen::InsensitiveStr::Ascii("perstige"), &["prestige"]),
            (dictgen::InsensitiveStr::Ascii("indicatore"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("originalis"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("assoziated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("plcaebo"), &["placebo"]),
            (
                dictgen::InsensitiveStr::Ascii("charactaristics"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fulnerabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("typle"), &["tuple"]),
            (
                dictgen::InsensitiveStr::Ascii("thermisors"),
                &["thermistors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converions"),
                &["conversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("atain"), &["attain"]),
            (dictgen::InsensitiveStr::Ascii("assgining"), &["assigning"]),
            (
                dictgen::InsensitiveStr::Ascii("evertyhign"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contempate"),
                &["contemplate"],
            ),
            (dictgen::InsensitiveStr::Ascii("varmit"), &["varmint"]),
            (dictgen::InsensitiveStr::Ascii("tyrranies"), &["tyrannies"]),
            (dictgen::InsensitiveStr::Ascii("buillding"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("temprorily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("withold"), &["withhold"]),
            (
                dictgen::InsensitiveStr::Ascii("failes"),
                &["fails", "failed", "files"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcisissts"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimer"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impementling"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trustworty"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thereotical"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakcgae"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("exersicing"),
                &["exercising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaulating"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheerleeders"),
                &["cheerleaders"],
            ),
            (dictgen::InsensitiveStr::Ascii("manifect"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("uninished"), &["unfinished"]),
            (dictgen::InsensitiveStr::Ascii("ffrom"), &["from"]),
            (dictgen::InsensitiveStr::Ascii("jurneying"), &["journeying"]),
            (dictgen::InsensitiveStr::Ascii("themplates"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilishes"),
                &["establishes", "stabilises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multliplies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanization"),
                &["quantization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstre"),
                &["demonstrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("elgance"), &["elegance"]),
            (
                dictgen::InsensitiveStr::Ascii("succesfully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacable"),
                &["replicable", "replaceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sortcut"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("stardards"), &["standards"]),
            (
                dictgen::InsensitiveStr::Ascii("canonivalization"),
                &["canonicalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("succes"), &["success"]),
            (
                dictgen::InsensitiveStr::Ascii("categeories"),
                &["categories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apologitic"),
                &["apologetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("exntended"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("conveluted"),
                &["convoluted"],
            ),
            (dictgen::InsensitiveStr::Ascii("worler"), &["worker"]),
            (dictgen::InsensitiveStr::Ascii("reseving"), &["reserving"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstratably"),
                &["demonstrably"],
            ),
            (dictgen::InsensitiveStr::Ascii("sention"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("valiating"), &["validating"]),
            (dictgen::InsensitiveStr::Ascii("coalscing"), &["coalescing"]),
            (
                dictgen::InsensitiveStr::Ascii("surroundng"),
                &["surrounding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technicion"),
                &["technician"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuables"), &["visuals"]),
            (
                dictgen::InsensitiveStr::Ascii("locted"),
                &["located", "locked"],
            ),
            (dictgen::InsensitiveStr::Ascii("peom"), &["poem"]),
            (dictgen::InsensitiveStr::Ascii("preimeter"), &["perimeter"]),
            (
                dictgen::InsensitiveStr::Ascii("seflishness"),
                &["selfishness"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebiuld"), &["rebuild"]),
            (dictgen::InsensitiveStr::Ascii("accessble"), &["accessible"]),
            (dictgen::InsensitiveStr::Ascii("otimal"), &["optimal"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatior"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undercling"),
                &["underlying"],
            ),
            (dictgen::InsensitiveStr::Ascii("intregrate"), &["integrate"]),
            (
                dictgen::InsensitiveStr::Ascii("cunted"),
                &["counted", "hunted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jusification"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contimplating"),
                &["contemplating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shoul"),
                &["should", "shawl", "shoal"],
            ),
            (dictgen::InsensitiveStr::Ascii("undestand"), &["understand"]),
            (
                dictgen::InsensitiveStr::Ascii("simulteneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constituie"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("whenver"), &["whenever"]),
            (
                dictgen::InsensitiveStr::Ascii("salughtering"),
                &["slaughtering"],
            ),
            (dictgen::InsensitiveStr::Ascii("repporting"), &["reporting"]),
            (dictgen::InsensitiveStr::Ascii("indulgue"), &["indulge"]),
            (dictgen::InsensitiveStr::Ascii("organisate"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("opitmised"), &["optimised"]),
            (dictgen::InsensitiveStr::Ascii("supossed"), &["supposed"]),
            (dictgen::InsensitiveStr::Ascii("accuarte"), &["accurate"]),
            (dictgen::InsensitiveStr::Ascii("eddits"), &["edits"]),
            (
                dictgen::InsensitiveStr::Ascii("auccessive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("resepected"), &["respected"]),
            (
                dictgen::InsensitiveStr::Ascii("compelxity"),
                &["complexity"],
            ),
            (dictgen::InsensitiveStr::Ascii("visbile"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapered"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("cencus"), &["census"]),
            (dictgen::InsensitiveStr::Ascii("relace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("derections"),
                &["directions", "detections", "erections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projecticles"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proscratination"),
                &["procrastination"],
            ),
            (dictgen::InsensitiveStr::Ascii("carisma"), &["charisma"]),
            (
                dictgen::InsensitiveStr::Ascii("functinoality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("releafes"), &["relieves"]),
            (dictgen::InsensitiveStr::Ascii("timestap"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("captins"), &["captains"]),
            (
                dictgen::InsensitiveStr::Ascii("indepence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("embbedder"), &["embedder"]),
            (dictgen::InsensitiveStr::Ascii("upgradei"), &["upgrade"]),
            (dictgen::InsensitiveStr::Ascii("upadtes"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("demcorats"), &["democrats"]),
            (
                dictgen::InsensitiveStr::Ascii("flexibity"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiifying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("isalmists"), &["islamists"]),
            (dictgen::InsensitiveStr::Ascii("creepes"), &["creepers"]),
            (
                dictgen::InsensitiveStr::Ascii("vulbearbuilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manupulations"),
                &["manipulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("predicits"), &["predicts"]),
            (dictgen::InsensitiveStr::Ascii("wrked"), &["worked"]),
            (dictgen::InsensitiveStr::Ascii("hemlets"), &["helmets"]),
            (
                dictgen::InsensitiveStr::Ascii("acomodate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alighments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspiciois"),
                &["suspicions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subseqence"),
                &["subsequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoure"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("facilitatile"),
                &["facilitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("triggeres"), &["triggers"]),
            (dictgen::InsensitiveStr::Ascii("superioara"), &["superior"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("recevied"), &["received"]),
            (
                dictgen::InsensitiveStr::Ascii("programable"),
                &["programmable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hamburglers"),
                &["hamburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crytocurrencies"),
                &["cryptocurrencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captialist"),
                &["capitalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electricien"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interesing"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("navgiating"),
                &["navigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("unalllowed"), &["unallowed"]),
            (
                dictgen::InsensitiveStr::Ascii("overlcocking"),
                &["overclocking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downrading"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternaives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("aroaras"), &["auroras"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnositcs"),
                &["diagnostic", "diagnostics"],
            ),
            (dictgen::InsensitiveStr::Ascii("positivs"), &["positives"]),
            (dictgen::InsensitiveStr::Ascii("retireve"), &["retrieve"]),
            (
                dictgen::InsensitiveStr::Ascii("standartisator"),
                &["standardiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("coleeg"), &["colleague"]),
            (dictgen::InsensitiveStr::Ascii("plantext"), &["plaintext"]),
            (dictgen::InsensitiveStr::Ascii("paramerts"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("detaisl"), &["details"]),
            (
                dictgen::InsensitiveStr::Ascii("orientantion"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ublishers"), &["publishers"]),
            (
                dictgen::InsensitiveStr::Ascii("adavantages"),
                &["advantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coneccting"),
                &["connecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("noteable"), &["notable"]),
            (dictgen::InsensitiveStr::Ascii("moudules"), &["modules"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriage"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosphetic"),
                &["prosthetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializd"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremenetal"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrpution"),
                &["corruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscrpiption"),
                &["subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moisterizer"),
                &["moisturizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentciator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skilfull"),
                &["skilful", "skillful"],
            ),
            (dictgen::InsensitiveStr::Ascii("syirans"), &["syrians"]),
            (
                dictgen::InsensitiveStr::Ascii("delarations"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interviwer"),
                &["interviewer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbae"),
                &["masturbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("stong"), &["strong"]),
            (
                dictgen::InsensitiveStr::Ascii("accomponies"),
                &["accompanies"],
            ),
            (dictgen::InsensitiveStr::Ascii("swtichs"), &["switches"]),
            (
                dictgen::InsensitiveStr::Ascii("multilplied"),
                &["multiplied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transormation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("treshholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motivatin"),
                &["motivations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaniously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivelency"),
                &["equivalency"],
            ),
            (dictgen::InsensitiveStr::Ascii("diable"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("approxamatly"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("respnding"), &["responding"]),
            (dictgen::InsensitiveStr::Ascii("accpted"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("cosutmes"), &["costumes"]),
            (dictgen::InsensitiveStr::Ascii("seriaized"), &["serialized"]),
            (
                dictgen::InsensitiveStr::Ascii("undersatnd"),
                &["understands", "understand"],
            ),
            (dictgen::InsensitiveStr::Ascii("ccoutant"), &["accountant"]),
            (
                dictgen::InsensitiveStr::Ascii("expads"),
                &["expands", "expats"],
            ),
            (dictgen::InsensitiveStr::Ascii("naughtly"), &["naughty"]),
            (
                dictgen::InsensitiveStr::Ascii("enumarates"),
                &["enumerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("liekd"), &["liked"]),
            (
                dictgen::InsensitiveStr::Ascii("assocaition"),
                &["associations", "association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prexied"),
                &["prefixed", "proxied"],
            ),
            (dictgen::InsensitiveStr::Ascii("discusssed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("incldued"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("continously"),
                &["continuously"],
            ),
            (dictgen::InsensitiveStr::Ascii("reasonble"), &["reasonable"]),
            (dictgen::InsensitiveStr::Ascii("illnesss"), &["illnesses"]),
            (dictgen::InsensitiveStr::Ascii("sommebody"), &["somebody"]),
            (dictgen::InsensitiveStr::Ascii("exceptino"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("consistuted"),
                &["constituted", "consisted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronise"),
                &["synchronise"],
            ),
            (dictgen::InsensitiveStr::Ascii("maraconi"), &["macaroni"]),
            (
                dictgen::InsensitiveStr::Ascii("epxectations"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("inforced"), &["enforced"]),
            (
                dictgen::InsensitiveStr::Ascii("dentically"),
                &["identically"],
            ),
            (dictgen::InsensitiveStr::Ascii("palete"), &["palette"]),
            (
                dictgen::InsensitiveStr::Ascii("trasformers"),
                &["transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("tequilia"), &["tequila"]),
            (
                dictgen::InsensitiveStr::Ascii("reacent"),
                &["recent", "reagent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolgies"),
                &["apologies", "apologise"],
            ),
            (dictgen::InsensitiveStr::Ascii("capalize"), &["capitalize"]),
            (dictgen::InsensitiveStr::Ascii("filesystm"), &["filesystem"]),
            (dictgen::InsensitiveStr::Ascii("analyizers"), &["analyzers"]),
            (
                dictgen::InsensitiveStr::Ascii("correpsonds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("alhough"), &["although"]),
            (dictgen::InsensitiveStr::Ascii("brockend"), &["broken"]),
            (dictgen::InsensitiveStr::Ascii("indure"), &["endure"]),
            (dictgen::InsensitiveStr::Ascii("blugin"), &["plugin"]),
            (
                dictgen::InsensitiveStr::Ascii("techically"),
                &["tactically", "technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survaillance"),
                &["surveillance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanitification"),
                &["quantification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intereaction"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitional"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("listenend"), &["listened"]),
            (dictgen::InsensitiveStr::Ascii("receveing"), &["receiving"]),
            (dictgen::InsensitiveStr::Ascii("handel"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("defensd"), &["defends"]),
            (dictgen::InsensitiveStr::Ascii("reentant"), &["reentrant"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminanted"),
                &["contaminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condolenses"),
                &["condolences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confederatie"),
                &["confederate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fucntionally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initliasations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("testeing"), &["testing"]),
            (
                dictgen::InsensitiveStr::Ascii("reconeted"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzelis"),
                &["overzealous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microoseconds"),
                &["microseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("plottin"), &["plotting"]),
            (
                dictgen::InsensitiveStr::Ascii("difinitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilist"),
                &["capitalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexplicably"),
                &["inexplicably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacent"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("detailling"), &["detailing"]),
            (
                dictgen::InsensitiveStr::Ascii("teatotaler"),
                &["teetotaler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desserialization"),
                &["deserialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("reloccates"), &["relocates"]),
            (
                dictgen::InsensitiveStr::Ascii("notificatino"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("ephermal"), &["ephemeral"]),
            (dictgen::InsensitiveStr::Ascii("evidance"), &["evidence"]),
            (
                dictgen::InsensitiveStr::Ascii("underpining"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("purposfully"),
                &["purposefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("execut"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("priestes"), &["priests"]),
            (dictgen::InsensitiveStr::Ascii("pyschosis"), &["psychosis"]),
            (dictgen::InsensitiveStr::Ascii("stealthely"), &["stealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("comining"),
                &["combining", "coming"],
            ),
            (dictgen::InsensitiveStr::Ascii("justfies"), &["justifies"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitplications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleised"),
                &["parallelised"],
            ),
            (dictgen::InsensitiveStr::Ascii("imort"), &["import"]),
            (dictgen::InsensitiveStr::Ascii("graphial"), &["graphical"]),
            (
                dictgen::InsensitiveStr::Ascii("libertarinaism"),
                &["libertarianism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jailborken"),
                &["jailbroken"],
            ),
            (dictgen::InsensitiveStr::Ascii("emensely"), &["immensely"]),
            (
                dictgen::InsensitiveStr::Ascii("hemorages"),
                &["hemorrhages"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsytem"), &["subsystem"]),
            (dictgen::InsensitiveStr::Ascii("liverpol"), &["liverpool"]),
            (dictgen::InsensitiveStr::Ascii("aovidable"), &["avoidable"]),
            (dictgen::InsensitiveStr::Ascii("dsyphoria"), &["dysphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("invaribaly"),
                &["invariably"],
            ),
            (dictgen::InsensitiveStr::Ascii("aare"), &["are"]),
            (
                dictgen::InsensitiveStr::Ascii("comilations"),
                &["compilations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusivty"),
                &["exclusivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("segemnted"), &["segmented"]),
            (
                dictgen::InsensitiveStr::Ascii("sacntioned"),
                &["sanctioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diferrently"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collabrative"),
                &["collaborative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolist"),
                &["capitalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denines"),
                &["defines", "denies"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardina"), &["cardigan"]),
            (
                dictgen::InsensitiveStr::Ascii("illimunati"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("doube"), &["double"]),
            (dictgen::InsensitiveStr::Ascii("impunes"), &["impugns"]),
            (
                dictgen::InsensitiveStr::Ascii("strenghtend"),
                &["strengthen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplacted"),
                &["deduplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggestsed"), &["suggested"]),
            (dictgen::InsensitiveStr::Ascii("consmer"), &["consumer"]),
            (
                dictgen::InsensitiveStr::Ascii("defins"),
                &["defines", "define"],
            ),
            (dictgen::InsensitiveStr::Ascii("reflectie"), &["reflective"]),
            (dictgen::InsensitiveStr::Ascii("tolernce"), &["tolerance"]),
            (dictgen::InsensitiveStr::Ascii("bitamp"), &["bitmap"]),
            (
                dictgen::InsensitiveStr::Ascii("construced"),
                &["constructed", "construed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribuem"),
                &["contribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("staulk"), &["stalk"]),
            (dictgen::InsensitiveStr::Ascii("attemt"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("melatonion"), &["melatonin"]),
            (dictgen::InsensitiveStr::Ascii("requeried"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("dominats"), &["dominates"]),
            (
                dictgen::InsensitiveStr::Ascii("reenforcement"),
                &["reinforcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assiciates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poulations"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferences"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("achillees"), &["achilles"]),
            (
                dictgen::InsensitiveStr::Ascii("appropraite"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lesibans"), &["lesbians"]),
            (dictgen::InsensitiveStr::Ascii("dicard"), &["discard"]),
            (
                dictgen::InsensitiveStr::Ascii("experinecing"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("fantaic"), &["fanatic"]),
            (dictgen::InsensitiveStr::Ascii("colsely"), &["closely"]),
            (
                dictgen::InsensitiveStr::Ascii("charcaters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("godliek"), &["godlike"]),
            (
                dictgen::InsensitiveStr::Ascii("insubstantiated"),
                &["unsubstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("formelly"), &["formerly"]),
            (
                dictgen::InsensitiveStr::Ascii("probabilites"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("petetion"), &["petition"]),
            (
                dictgen::InsensitiveStr::Ascii("develpment"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("dectected"), &["detected"]),
            (dictgen::InsensitiveStr::Ascii("rapell"), &["rappel"]),
            (dictgen::InsensitiveStr::Ascii("squishey"), &["squishy"]),
            (
                dictgen::InsensitiveStr::Ascii("fascinatie"),
                &["fascinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("minitature"), &["miniature"]),
            (dictgen::InsensitiveStr::Ascii("pythong"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscribbed"),
                &["unsubscribe"],
            ),
            (dictgen::InsensitiveStr::Ascii("characher"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("tageting"), &["targeting"]),
            (
                dictgen::InsensitiveStr::Ascii("videogaems"),
                &["videogames"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adverticements"),
                &["advertisements"],
            ),
            (dictgen::InsensitiveStr::Ascii("arhtritis"), &["arthritis"]),
            (
                dictgen::InsensitiveStr::Ascii("dpends"),
                &["depends", "spends"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovverride"), &["override"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilzers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enligthenment"),
                &["enlightenment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exort"),
                &["export", "exhort", "extort", "exert"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumated"),
                &["consummated"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherrited"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("algorhtms"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("valkiyre"), &["valkyrie"]),
            (
                dictgen::InsensitiveStr::Ascii("cateogrization"),
                &["categorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitution"), &["situation"]),
            (dictgen::InsensitiveStr::Ascii("pyarmid"), &["pyramid"]),
            (dictgen::InsensitiveStr::Ascii("enocde"), &["encode"]),
            (
                dictgen::InsensitiveStr::Ascii("suppliments"),
                &["supplements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarenteedes"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caclucator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("unkknown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("convenvient"),
                &["convenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("offisde"), &["offside"]),
            (
                dictgen::InsensitiveStr::Ascii("occurences"),
                &["occurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularlise"),
                &["regularise"],
            ),
            (dictgen::InsensitiveStr::Ascii("increasin"), &["increasing"]),
            (dictgen::InsensitiveStr::Ascii("machanics"), &["mechanics"]),
            (dictgen::InsensitiveStr::Ascii("instensity"), &["intensity"]),
            (
                dictgen::InsensitiveStr::Ascii("terimnated"),
                &["terminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resitsance"),
                &["resistances"],
            ),
            (dictgen::InsensitiveStr::Ascii("eneable"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("associaed"), &["associated"]),
            (dictgen::InsensitiveStr::Ascii("blaspehmy"), &["blasphemy"]),
            (dictgen::InsensitiveStr::Ascii("abiss"), &["abyss"]),
            (
                dictgen::InsensitiveStr::Ascii("creativelly"),
                &["creatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elementrary"),
                &["elementary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("committment"),
                &["commitment"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewriten"), &["rewritten"]),
            (dictgen::InsensitiveStr::Ascii("exentended"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("sythesized"),
                &["synthesized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transisitioning"),
                &["transitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("falures"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("virgines"), &["virgins"]),
            (
                dictgen::InsensitiveStr::Ascii("multitudine"),
                &["multitude"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neccessarry"),
                &["necessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("scostman"), &["scotsman"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticor"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconeect"),
                &["disconnect"],
            ),
            (dictgen::InsensitiveStr::Ascii("specied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("langulage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilsing"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifeid"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("theroem"), &["theorem"]),
            (dictgen::InsensitiveStr::Ascii("claenups"), &["cleanups"]),
            (
                dictgen::InsensitiveStr::Ascii("precenences"),
                &["preferences", "precedences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrution"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreviaton"),
                &["abbreviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("equeality"), &["equality"]),
            (dictgen::InsensitiveStr::Ascii("febbruary"), &["february"]),
            (
                dictgen::InsensitiveStr::Ascii("inititiave"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("intervies"), &["interviews"]),
            (dictgen::InsensitiveStr::Ascii("sepeartor"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("hierachical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("afe"), &["safe"]),
            (dictgen::InsensitiveStr::Ascii("reidual"), &["residual"]),
            (
                dictgen::InsensitiveStr::Ascii("revoluionary"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accordians"),
                &["accordions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilizies"),
                &["stabilize"],
            ),
            (dictgen::InsensitiveStr::Ascii("contibute"), &["contribute"]),
            (
                dictgen::InsensitiveStr::Ascii("disccussion"),
                &["discussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimizin"), &["optimizing"]),
            (
                dictgen::InsensitiveStr::Ascii("showfer"),
                &["chauffeur", "shower"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconenctions"),
                &["reconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilitises"), &["utilises"]),
            (
                dictgen::InsensitiveStr::Ascii("catgorizes"),
                &["categorizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undersatnding"),
                &["understanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anevironments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("canceres"), &["cancers"]),
            (
                dictgen::InsensitiveStr::Ascii("disensitized"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("middelware"),
                &["middleware"],
            ),
            (dictgen::InsensitiveStr::Ascii("chcksums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("somethiung"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("analisis"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("organisims"), &["organism"]),
            (dictgen::InsensitiveStr::Ascii("seriel"), &["serial"]),
            (dictgen::InsensitiveStr::Ascii("acccepts"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("gereric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("applys"), &["applies"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstaces"),
                &["circumstances"],
            ),
            (dictgen::InsensitiveStr::Ascii("prototyes"), &["prototypes"]),
            (
                dictgen::InsensitiveStr::Ascii("cricitising"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburgare"),
                &["cheeseburger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composit"),
                &["compost", "composite"],
            ),
            (dictgen::InsensitiveStr::Ascii("supploed"), &["supplied"]),
            (
                dictgen::InsensitiveStr::Ascii("princeples"),
                &["principles"],
            ),
            (dictgen::InsensitiveStr::Ascii("touchdwon"), &["touchdown"]),
            (dictgen::InsensitiveStr::Ascii("secedin"), &["seceding"]),
            (
                dictgen::InsensitiveStr::Ascii("chaced"),
                &["cached", "chased"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoratitative"),
                &["authoritative"],
            ),
            (dictgen::InsensitiveStr::Ascii("assition"), &["assertion"]),
            (dictgen::InsensitiveStr::Ascii("smoetime"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("sawnsea"), &["swansea"]),
            (
                dictgen::InsensitiveStr::Ascii("dielectircs"),
                &["dielectrics"],
            ),
            (dictgen::InsensitiveStr::Ascii("tutoral"), &["tutorial"]),
            (dictgen::InsensitiveStr::Ascii("parially"), &["partially"]),
            (dictgen::InsensitiveStr::Ascii("releives"), &["relieves"]),
            (dictgen::InsensitiveStr::Ascii("sepearate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("packkages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("ourselv"),
                &["ourself", "ourselves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possitional"),
                &["positional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fransiscan"),
                &["franciscan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connetations"),
                &["connotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persecutie"),
                &["persecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentiels"),
                &["potentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("wathcdog"), &["watchdog"]),
            (dictgen::InsensitiveStr::Ascii("focued"), &["focused"]),
            (
                dictgen::InsensitiveStr::Ascii("correcteness"),
                &["correctness"],
            ),
            (dictgen::InsensitiveStr::Ascii("uniqu"), &["unique"]),
            (dictgen::InsensitiveStr::Ascii("americam"), &["american"]),
            (dictgen::InsensitiveStr::Ascii("roatated"), &["rotated"]),
            (dictgen::InsensitiveStr::Ascii("incrases"), &["increases"]),
            (dictgen::InsensitiveStr::Ascii("explosin"), &["explosions"]),
            (
                dictgen::InsensitiveStr::Ascii("fingertaps"),
                &["fingertips"],
            ),
            (dictgen::InsensitiveStr::Ascii("tourisim"), &["tourism"]),
            (
                dictgen::InsensitiveStr::Ascii("ublications"),
                &["publications"],
            ),
            (dictgen::InsensitiveStr::Ascii("confinin"), &["confining"]),
            (dictgen::InsensitiveStr::Ascii("rewieved"), &["reviewed"]),
            (dictgen::InsensitiveStr::Ascii("biologiskt"), &["biologist"]),
            (dictgen::InsensitiveStr::Ascii("sebasitan"), &["sebastian"]),
            (
                dictgen::InsensitiveStr::Ascii("oriantated"),
                &["orientated"],
            ),
            (dictgen::InsensitiveStr::Ascii("disappars"), &["disappears"]),
            (
                dictgen::InsensitiveStr::Ascii("uderstands"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("cornels"), &["colonels"]),
            (
                dictgen::InsensitiveStr::Ascii("continuating"),
                &["continuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalsim"),
                &["journalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuiremnt"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("varriation"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("montored"),
                &["monitored", "mentored"],
            ),
            (dictgen::InsensitiveStr::Ascii("homewoner"), &["homeowner"]),
            (dictgen::InsensitiveStr::Ascii("caclculate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("serializng"),
                &["serializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("analsys"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("diffenrently"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("protypted"), &["prototyped"]),
            (
                dictgen::InsensitiveStr::Ascii("reprensents"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniected"), &["injected"]),
            (
                dictgen::InsensitiveStr::Ascii("transporteurs"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmoutned"), &["unmounted"]),
            (
                dictgen::InsensitiveStr::Ascii("catholicsm"),
                &["catholicism"],
            ),
            (dictgen::InsensitiveStr::Ascii("deadlit"), &["deadlift"]),
            (
                dictgen::InsensitiveStr::Ascii("discharded"),
                &["discharged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apllicatin"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celebrationis"),
                &["celebrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("friednzone"),
                &["friendzoned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregisteing"),
                &["registering", "unregistering"],
            ),
            (dictgen::InsensitiveStr::Ascii("cesspol"), &["cesspool"]),
            (dictgen::InsensitiveStr::Ascii("ahtiests"), &["atheists"]),
            (
                dictgen::InsensitiveStr::Ascii("rerurn"),
                &["return", "rerun"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ankshisness"),
                &["anxiousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrcutions"),
                &["instruction", "instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompaitible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstrumented"),
                &["instrumented"],
            ),
            (dictgen::InsensitiveStr::Ascii("singnal"), &["signal"]),
            (
                dictgen::InsensitiveStr::Ascii("chacing"),
                &["caching", "chasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("meachanism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("circomference"),
                &["circumference"],
            ),
            (dictgen::InsensitiveStr::Ascii("buillds"), &["builds"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnphobic"),
                &["transphobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("apartheied"), &["apartheid"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrbe"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprogations"),
                &["propagations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immidiatelly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("riminicent"),
                &["reminiscent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generalizate"),
                &["generalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rounters"),
                &["routers", "rounders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declearatively"),
                &["declaratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percentige"),
                &["percentile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proagation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informacion"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("erticulations"),
                &["articulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("penisular"), &["peninsular"]),
            (dictgen::InsensitiveStr::Ascii("pauload"), &["payload"]),
            (dictgen::InsensitiveStr::Ascii("appoved"), &["approved"]),
            (
                dictgen::InsensitiveStr::Ascii("progressus"),
                &["progresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("commputer"), &["computer"]),
            (dictgen::InsensitiveStr::Ascii("orbtial"), &["orbital"]),
            (
                dictgen::InsensitiveStr::Ascii("consoldate"),
                &["consolidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementatios"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accceptably"),
                &["acceptably"],
            ),
            (dictgen::InsensitiveStr::Ascii("countainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("adjasently"),
                &["adjacently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformating"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("braches"),
                &["branches", "breaches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculatoare"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perjorative"),
                &["pejorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("reapeats"), &["repeats"]),
            (dictgen::InsensitiveStr::Ascii("gamerga"), &["gamertag"]),
            (
                dictgen::InsensitiveStr::Ascii("expetation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotionating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("incement"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("homeoapthy"),
                &["homeopathy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countinuously"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exacarbated"),
                &["exacerbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuggged"), &["debugged"]),
            (dictgen::InsensitiveStr::Ascii("operarors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("outupts"), &["outputs"]),
            (
                dictgen::InsensitiveStr::Ascii("executionner"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceepting"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenaing"),
                &["concatenating"],
            ),
            (dictgen::InsensitiveStr::Ascii("depennding"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("afficianado"),
                &["aficionado"],
            ),
            (dictgen::InsensitiveStr::Ascii("assicated"), &["associated"]),
            (dictgen::InsensitiveStr::Ascii("agreable"), &["agreeable"]),
            (dictgen::InsensitiveStr::Ascii("facilisi"), &["facilities"]),
            (
                dictgen::InsensitiveStr::Ascii("experimet"),
                &["experiments", "experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconvertable"),
                &["unconvertible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computating"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("assassinas"), &["assassins"]),
            (
                dictgen::InsensitiveStr::Ascii("corelating"),
                &["correlating"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalsce"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("intentation"),
                &["indentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ful"), &["full"]),
            (dictgen::InsensitiveStr::Ascii("wiated"), &["waited"]),
            (
                dictgen::InsensitiveStr::Ascii("spezialisation"),
                &["specialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("usin"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("instrunctions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("microwae"), &["microwaves"]),
            (dictgen::InsensitiveStr::Ascii("detaled"), &["detailed"]),
            (
                dictgen::InsensitiveStr::Ascii("combintaorics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complainig"),
                &["complaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecion"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("strengtheing"),
                &["strengthening"],
            ),
            (dictgen::InsensitiveStr::Ascii("extensons"), &["extensions"]),
            (dictgen::InsensitiveStr::Ascii("ellided"), &["elided"]),
            (dictgen::InsensitiveStr::Ascii("redundunt"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("desabilities"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simuating"),
                &["simulating", "situating"],
            ),
            (dictgen::InsensitiveStr::Ascii("streches"), &["stretches"]),
            (
                dictgen::InsensitiveStr::Ascii("commtiting"),
                &["committing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigen"), &["sign"]),
            (
                dictgen::InsensitiveStr::Ascii("procreatin"),
                &["procreation"],
            ),
            (dictgen::InsensitiveStr::Ascii("rocorder"), &["recorder"]),
            (
                dictgen::InsensitiveStr::Ascii("composte"),
                &["compose", "composed", "composite", "compost", "composted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comnpresion"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streigts"),
                &["straights", "straits"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplifiy"), &["simplify"]),
            (dictgen::InsensitiveStr::Ascii("intrpret"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("correspinded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underestamate"),
                &["underestimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wolwide"), &["worldwide"]),
            (
                dictgen::InsensitiveStr::Ascii("defineatly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ource"),
                &["source", "ounce"],
            ),
            (dictgen::InsensitiveStr::Ascii("listenter"), &["listener"]),
            (
                dictgen::InsensitiveStr::Ascii("progatators"),
                &["propagators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("angshiosness"),
                &["anxiousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inefficienty"),
                &["inefficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environent"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separatisme"),
                &["separates"],
            ),
            (dictgen::InsensitiveStr::Ascii("oru"), &["our"]),
            (
                dictgen::InsensitiveStr::Ascii("comflicted"),
                &["conflicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrodinary"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("followeing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("invissible"), &["invisible"]),
            (
                dictgen::InsensitiveStr::Ascii("domonstrates"),
                &["demonstrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("couldnot"), &["couldnt"]),
            (
                dictgen::InsensitiveStr::Ascii("candinates"),
                &["candidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discertation"),
                &["dissertation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adustments"),
                &["adjustments"],
            ),
            (dictgen::InsensitiveStr::Ascii("holdes"), &["holds"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenatin"),
                &["concatenating", "concatenation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurisons"),
                &["recursions"],
            ),
            (dictgen::InsensitiveStr::Ascii("increaded"), &["increased"]),
            (dictgen::InsensitiveStr::Ascii("chizzell"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("professionalsim"),
                &["professionalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("arcived"), &["archived"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuito"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sirectness"),
                &["directness"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntakt"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("contribtions"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redundantcy"),
                &["redundancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sexaulized"),
                &["sexualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("docmunets"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("rewrited"),
                &["rewrote", "rewritten"],
            ),
            (dictgen::InsensitiveStr::Ascii("sreens"), &["screens"]),
            (
                dictgen::InsensitiveStr::Ascii("colectable"),
                &["collectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastrcuture"),
                &["infrastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangde"), &["arranged"]),
            (
                dictgen::InsensitiveStr::Ascii("informatiom"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcripton"),
                &["transcription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusivety"),
                &["exclusivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniitial"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("yourselve"),
                &["yourselves", "yourself"],
            ),
            (dictgen::InsensitiveStr::Ascii("gernerics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("congratualte"),
                &["congratulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("consuments"), &["consumes"]),
            (dictgen::InsensitiveStr::Ascii("eliptic"), &["elliptic"]),
            (dictgen::InsensitiveStr::Ascii("fualts"), &["faults"]),
            (
                dictgen::InsensitiveStr::Ascii("audioboost"),
                &["audiobooks"],
            ),
            (dictgen::InsensitiveStr::Ascii("mounpoint"), &["mountpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("expeimentation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congresssmen"),
                &["congressmen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("xenophoblic"),
                &["xenophobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("randomess"), &["randomness"]),
            (
                dictgen::InsensitiveStr::Ascii("trotzkyism"),
                &["trotskyism"],
            ),
            (dictgen::InsensitiveStr::Ascii("missng"), &["missing"]),
            (
                dictgen::InsensitiveStr::Ascii("intrepretation"),
                &["interpretation"],
            ),
            (dictgen::InsensitiveStr::Ascii("buddah"), &["buddha"]),
            (dictgen::InsensitiveStr::Ascii("sleping"), &["sleeping"]),
            (
                dictgen::InsensitiveStr::Ascii("contingient"),
                &["contingent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspdonents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politicaly"),
                &["politically"],
            ),
            (dictgen::InsensitiveStr::Ascii("activiated"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("emegrency"), &["emergency"]),
            (
                dictgen::InsensitiveStr::Ascii("solidairty"),
                &["solidarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deambiguate"),
                &["disambiguate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qestionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exeggeration"),
                &["exaggeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corisponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("turksih"), &["turkish"]),
            (
                dictgen::InsensitiveStr::Ascii("battlefiled"),
                &["battlefield"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeate"), &["repeat"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("gziped"), &["gzipped"]),
            (dictgen::InsensitiveStr::Ascii("fom"), &["form", "from"]),
            (dictgen::InsensitiveStr::Ascii("traveral"), &["traversal"]),
            (
                dictgen::InsensitiveStr::Ascii("accessiibility"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannonicalize"),
                &["canonicalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similtaneously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charsimatic"),
                &["charismatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("rertieved"), &["retrieved"]),
            (dictgen::InsensitiveStr::Ascii("persiting"), &["persisting"]),
            (dictgen::InsensitiveStr::Ascii("thexture"), &["texture"]),
            (
                dictgen::InsensitiveStr::Ascii("seperators"),
                &["separators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intenationalist"),
                &["internationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuzed"), &["confused"]),
            (dictgen::InsensitiveStr::Ascii("changeed"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("neightboring"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("advsior"), &["advisor"]),
            (dictgen::InsensitiveStr::Ascii("partcle"), &["particle"]),
            (dictgen::InsensitiveStr::Ascii("signul"), &["signal"]),
            (dictgen::InsensitiveStr::Ascii("drummless"), &["drumless"]),
            (
                dictgen::InsensitiveStr::Ascii("notifacations"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("dobule"), &["double"]),
            (dictgen::InsensitiveStr::Ascii("terarform"), &["terraform"]),
            (
                dictgen::InsensitiveStr::Ascii("unterstanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("pluign"), &["plugin"]),
            (
                dictgen::InsensitiveStr::Ascii("conplimenting"),
                &["complimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratiuitous"),
                &["gratuitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contiuning"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("underun"), &["underrun"]),
            (
                dictgen::InsensitiveStr::Ascii("cristallises"),
                &["crystallises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstlals"),
                &["uninstalls"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("istantiations"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("originalet"),
                &["originated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradtionally"),
                &["traditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("entreis"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("unashable"), &["unhashable"]),
            (
                dictgen::InsensitiveStr::Ascii("dedepulicate"),
                &["deduplicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("securre"), &["secure"]),
            (
                dictgen::InsensitiveStr::Ascii("denomonations"),
                &["denominations"],
            ),
            (dictgen::InsensitiveStr::Ascii("difficut"), &["difficulty"]),
            (
                dictgen::InsensitiveStr::Ascii("enthically"),
                &["ethnically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coonstantly"),
                &["constantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("volatiliy"), &["volatility"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegoitation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missleading"),
                &["misleading"],
            ),
            (dictgen::InsensitiveStr::Ascii("adverised"), &["advertised"]),
            (
                dictgen::InsensitiveStr::Ascii("propogating"),
                &["propagating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requseting"),
                &["requesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statictics"),
                &["statistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armagideon"),
                &["armageddon"],
            ),
            (dictgen::InsensitiveStr::Ascii("deapth"), &["depth"]),
            (dictgen::InsensitiveStr::Ascii("firslty"), &["firstly"]),
            (dictgen::InsensitiveStr::Ascii("invirants"), &["invariants"]),
            (dictgen::InsensitiveStr::Ascii("arrrives"), &["arrives"]),
            (
                dictgen::InsensitiveStr::Ascii("underastimated"),
                &["underestimated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consciousely"),
                &["consciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("inferioir"), &["inferior"]),
            (dictgen::InsensitiveStr::Ascii("inverterad"), &["inverted"]),
            (dictgen::InsensitiveStr::Ascii("slighlty"), &["slightly"]),
            (
                dictgen::InsensitiveStr::Ascii("acumalated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hegdehog"), &["hedgehog"]),
            (dictgen::InsensitiveStr::Ascii("presnting"), &["presenting"]),
            (dictgen::InsensitiveStr::Ascii("economicus"), &["economics"]),
            (dictgen::InsensitiveStr::Ascii("abtractly"), &["abstractly"]),
            (
                dictgen::InsensitiveStr::Ascii("negotating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstucting"),
                &["reconstructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("spining"), &["spinning"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("extacy"), &["ecstasy"]),
            (dictgen::InsensitiveStr::Ascii("archetpye"), &["archetype"]),
            (dictgen::InsensitiveStr::Ascii("ilustrate"), &["illustrate"]),
            (
                dictgen::InsensitiveStr::Ascii("verifcation"),
                &["verification"],
            ),
            (dictgen::InsensitiveStr::Ascii("clincial"), &["clinical"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstartions"),
                &["demonstrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("largley"), &["largely"]),
            (dictgen::InsensitiveStr::Ascii("infinife"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubstantianted"),
                &["unsubstantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harrasment"),
                &["harassment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countour"),
                &["contour", "counter"],
            ),
            (dictgen::InsensitiveStr::Ascii("gettters"), &["getters"]),
            (dictgen::InsensitiveStr::Ascii("hopyfully"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("regster"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("manaul"), &["manual"]),
            (dictgen::InsensitiveStr::Ascii("charaset"), &["charset"]),
            (dictgen::InsensitiveStr::Ascii("writebasck"), &["writeback"]),
            (
                dictgen::InsensitiveStr::Ascii("institutionals"),
                &["institutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioator"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("shematic"), &["schematic"]),
            (dictgen::InsensitiveStr::Ascii("asynnc"), &["async"]),
            (
                dictgen::InsensitiveStr::Ascii("problemita"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undercores"),
                &["underscores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordinairly"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("associaes"), &["associates"]),
            (
                dictgen::InsensitiveStr::Ascii("peroidically"),
                &["periodically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritpitons"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomonen"),
                &["phenomenon"],
            ),
            (dictgen::InsensitiveStr::Ascii("cellabrate"), &["celebrate"]),
            (
                dictgen::InsensitiveStr::Ascii("playthrouhg"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("groubpy"), &["groupby"]),
            (
                dictgen::InsensitiveStr::Ascii("formualation"),
                &["formulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("northren"), &["northern"]),
            (dictgen::InsensitiveStr::Ascii("frezes"), &["freezes"]),
            (dictgen::InsensitiveStr::Ascii("corrcts"), &["corrects"]),
            (
                dictgen::InsensitiveStr::Ascii("unrealiable"),
                &["unreliable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qeuestionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quaruntine"),
                &["quarantine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knoweldgable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifyed"), &["verified"]),
            (
                dictgen::InsensitiveStr::Ascii("dignostics"),
                &["diagnostics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilisation"),
                &["capitalisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramtrical"),
                &["parametrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("depenend"), &["depend"]),
            (
                dictgen::InsensitiveStr::Ascii("administro"),
                &["administer"],
            ),
            (dictgen::InsensitiveStr::Ascii("replce"), &["replace"]),
            (dictgen::InsensitiveStr::Ascii("idae"), &["idea"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsrapping"),
                &["bootstrapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resssurecting"),
                &["resurrecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parathesis"),
                &["parenthesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("hovever"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("architure"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implicilty"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectuting"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("abbreivates"),
                &["abbreviates"],
            ),
            (dictgen::InsensitiveStr::Ascii("tigther"), &["tighter"]),
            (
                dictgen::InsensitiveStr::Ascii("neighorhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclsuives"),
                &["exclusives"],
            ),
            (dictgen::InsensitiveStr::Ascii("frequencny"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("dublicates"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaienr"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptivos"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("numeber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("despiste"), &["despise"]),
            (
                dictgen::InsensitiveStr::Ascii("authenenticated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("histroians"),
                &["historians"],
            ),
            (dictgen::InsensitiveStr::Ascii("absail"), &["abseil"]),
            (dictgen::InsensitiveStr::Ascii("philosphy"), &["philosophy"]),
            (dictgen::InsensitiveStr::Ascii("plausibe"), &["plausible"]),
            (dictgen::InsensitiveStr::Ascii("disecting"), &["dissecting"]),
            (
                dictgen::InsensitiveStr::Ascii("apppearances"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stoopid"),
                &["stupid", "stooped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fielesystems"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactuerer"),
                &["manufacturer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhoud"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("explorin"), &["exploring"]),
            (dictgen::InsensitiveStr::Ascii("balnce"), &["balance"]),
            (dictgen::InsensitiveStr::Ascii("burjun"), &["burgeon"]),
            (
                dictgen::InsensitiveStr::Ascii("sustitition"),
                &["substitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallision"), &["collision"]),
            (
                dictgen::InsensitiveStr::Ascii("expreses"),
                &["express", "expresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgarting"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifikation"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cristallized"),
                &["crystallized"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoluth"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("eisntance"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("sarcasticaly"),
                &["sarcastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("identtity"), &["identity"]),
            (dictgen::InsensitiveStr::Ascii("siwtcher"), &["switcher"]),
            (
                dictgen::InsensitiveStr::Ascii("detecttions"),
                &["detections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotition"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("respwaned"), &["respawned"]),
            (dictgen::InsensitiveStr::Ascii("exmaple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("envent"),
                &["event", "invent"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutscences"), &["cutscene"]),
            (dictgen::InsensitiveStr::Ascii("jurnales"), &["journals"]),
            (
                dictgen::InsensitiveStr::Ascii("rondazyvooed"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persmission"),
                &["permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("jenkns"), &["jenkins"]),
            (dictgen::InsensitiveStr::Ascii("commpn"), &["common"]),
            (dictgen::InsensitiveStr::Ascii("soursce"), &["source"]),
            (dictgen::InsensitiveStr::Ascii("docstinrg"), &["docstring"]),
            (dictgen::InsensitiveStr::Ascii("shoppin"), &["shopping"]),
            (dictgen::InsensitiveStr::Ascii("carrotus"), &["carrots"]),
            (
                dictgen::InsensitiveStr::Ascii("nuisanses"),
                &["nuisance", "nuisances"],
            ),
            (dictgen::InsensitiveStr::Ascii("afrer"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("linerly"), &["linearly"]),
            (
                dictgen::InsensitiveStr::Ascii("invarinats"),
                &["invariants"],
            ),
            (dictgen::InsensitiveStr::Ascii("fracton"), &["fraction"]),
            (
                dictgen::InsensitiveStr::Ascii("liberalisim"),
                &["liberalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("buttong"), &["button"]),
            (dictgen::InsensitiveStr::Ascii("archvie"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("offen"), &["often"]),
            (dictgen::InsensitiveStr::Ascii("convense"), &["convince"]),
            (dictgen::InsensitiveStr::Ascii("efort"), &["effort"]),
            (
                dictgen::InsensitiveStr::Ascii("indicentally"),
                &["incidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwse"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("unviersity"),
                &["university"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchrons"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("borded"),
                &["boarded", "border"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitition"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("natinalists"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applyication"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requeseting"),
                &["requesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definiative"),
                &["definitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("enulation"), &["emulation"]),
            (
                dictgen::InsensitiveStr::Ascii("recusrively"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorhitmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("disovery"), &["discovery"]),
            (dictgen::InsensitiveStr::Ascii("molesterd"), &["molested"]),
            (
                dictgen::InsensitiveStr::Ascii("calcutation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asethetics"),
                &["aesthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("contines"), &["continents"]),
            (
                dictgen::InsensitiveStr::Ascii("millionaries"),
                &["millionaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfcation"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("scenariou"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("organziers"),
                &["organizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("transer"), &["transfer"]),
            (
                dictgen::InsensitiveStr::Ascii("geenrating"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dijskstra"), &["dijkstra"]),
            (dictgen::InsensitiveStr::Ascii("hairsytle"), &["hairstyle"]),
            (dictgen::InsensitiveStr::Ascii("frops"), &["drops"]),
            (dictgen::InsensitiveStr::Ascii("chihauhua"), &["chihuahua"]),
            (dictgen::InsensitiveStr::Ascii("iinclude"), &["include"]),
            (dictgen::InsensitiveStr::Ascii("spurios"), &["spurious"]),
            (dictgen::InsensitiveStr::Ascii("exceeeds"), &["exceeds"]),
            (dictgen::InsensitiveStr::Ascii("cecker"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("delgate"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("auxilirary"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("gneric"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("depenendencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devolopment"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("athron"), &["athlon"]),
            (dictgen::InsensitiveStr::Ascii("reaplying"), &["replaying"]),
            (dictgen::InsensitiveStr::Ascii("poitning"), &["pointing"]),
            (dictgen::InsensitiveStr::Ascii("ambulence"), &["ambulance"]),
            (dictgen::InsensitiveStr::Ascii("notoreous"), &["notorious"]),
            (
                dictgen::InsensitiveStr::Ascii("insititutes"),
                &["institutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassine"),
                &["assassinate", "assassin"],
            ),
            (dictgen::InsensitiveStr::Ascii("toxicictiy"), &["toxicity"]),
            (dictgen::InsensitiveStr::Ascii("fatured"), &["featured"]),
            (
                dictgen::InsensitiveStr::Ascii("renweables"),
                &["renewables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralelizing"),
                &["parallelizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("acctual"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("conversin"),
                &["conversions", "conversing", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("preamde"), &["premade"]),
            (
                dictgen::InsensitiveStr::Ascii("unrecocnized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("whant"), &["want"]),
            (
                dictgen::InsensitiveStr::Ascii("separater"),
                &["separator", "separated", "separates", "separate"],
            ),
            (dictgen::InsensitiveStr::Ascii("mian"), &["main", "moan"]),
            (dictgen::InsensitiveStr::Ascii("structur"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("plausability"),
                &["plausibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diconnection"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reorganision"),
                &["reorganisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("auxiluary"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("porgrams"), &["programs"]),
            (dictgen::InsensitiveStr::Ascii("eejus"), &["aegis"]),
            (dictgen::InsensitiveStr::Ascii("vresioned"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("easyer"), &["easier"]),
            (dictgen::InsensitiveStr::Ascii("unregiste"), &["unregister"]),
            (
                dictgen::InsensitiveStr::Ascii("alernatives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("segemented"), &["segmented"]),
            (
                dictgen::InsensitiveStr::Ascii("integrering"),
                &["interfering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimentionnality"),
                &["dimensionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isntalations"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("psots"), &["posts"]),
            (
                dictgen::InsensitiveStr::Ascii("carnige"),
                &["carnage", "carnegie"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algothitmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("improrted"), &["imported"]),
            (dictgen::InsensitiveStr::Ascii("seprates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("postives"), &["positives"]),
            (dictgen::InsensitiveStr::Ascii("interract"), &["interact"]),
            (
                dictgen::InsensitiveStr::Ascii("negotionators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("accruate"), &["accurate"]),
            (
                dictgen::InsensitiveStr::Ascii("admininistered"),
                &["administered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demographs"),
                &["demographics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opproximate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("booundaries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("alllow"), &["allow"]),
            (dictgen::InsensitiveStr::Ascii("mirrioring"), &["mirroring"]),
            (
                dictgen::InsensitiveStr::Ascii("architecturse"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislpay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("petrolem"), &["petroleum"]),
            (
                dictgen::InsensitiveStr::Ascii("cleare"),
                &["cleared", "clear"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exploitatie"),
                &["exploitative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratuituos"),
                &["gratuitous"],
            ),
            (dictgen::InsensitiveStr::Ascii("hertiage"), &["heritage"]),
            (
                dictgen::InsensitiveStr::Ascii("menaingful"),
                &["meaningful"],
            ),
            (dictgen::InsensitiveStr::Ascii("smetimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("beauitful"), &["beautiful"]),
            (dictgen::InsensitiveStr::Ascii("emplore"), &["employer"]),
            (
                dictgen::InsensitiveStr::Ascii("emabrassing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conrrupted"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("chechlists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("develepment"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gracin"),
                &["gracing", "gain", "gratin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concecutive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("insultos"), &["insults"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasmitter"),
                &["transmitter"],
            ),
            (dictgen::InsensitiveStr::Ascii("experssed"), &["expressed"]),
            (
                dictgen::InsensitiveStr::Ascii("animationos"),
                &["animations"],
            ),
            (dictgen::InsensitiveStr::Ascii("diabled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("transalate"), &["translate"]),
            (dictgen::InsensitiveStr::Ascii("cralwer"), &["crawler"]),
            (dictgen::InsensitiveStr::Ascii("exaclinic"), &["hexaclinic"]),
            (dictgen::InsensitiveStr::Ascii("rorated"), &["rotated"]),
            (
                dictgen::InsensitiveStr::Ascii("ansynchronous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recursse"),
                &["recurse", "recurses"],
            ),
            (dictgen::InsensitiveStr::Ascii("procss"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsitancies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hihglights"),
                &["highlights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatins"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ontrolled"), &["controlled"]),
            (dictgen::InsensitiveStr::Ascii("weherever"), &["wherever"]),
            (dictgen::InsensitiveStr::Ascii("messures"), &["measures"]),
            (
                dictgen::InsensitiveStr::Ascii("motherbords"),
                &["motherboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firghtening"),
                &["frightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraploate"),
                &["extrapolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranciver"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("ardvarks"), &["aardvarks"]),
            (dictgen::InsensitiveStr::Ascii("sceleton"), &["skeleton"]),
            (dictgen::InsensitiveStr::Ascii("liberary"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("foloowers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("orginal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("slaches"), &["slashes"]),
            (
                dictgen::InsensitiveStr::Ascii("capalized"),
                &["capitalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("repell"), &["repel"]),
            (dictgen::InsensitiveStr::Ascii("diamter"), &["diameter"]),
            (
                dictgen::InsensitiveStr::Ascii("contemporay"),
                &["contemporary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcitionalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("artciles"), &["articles"]),
            (
                dictgen::InsensitiveStr::Ascii("implementattion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriatchy"),
                &["patriarchy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crystaline"),
                &["crystalline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inoperant"),
                &["inoperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradionally"),
                &["traditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("aliged"), &["aligned"]),
            (
                dictgen::InsensitiveStr::Ascii("imexperience"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("exanded"), &["expanded"]),
            (
                dictgen::InsensitiveStr::Ascii("maintence"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortuately"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonyomous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditianal"),
                &["unconditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calucators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("danceing"), &["dancing"]),
            (dictgen::InsensitiveStr::Ascii("undergrand"), &["undergrad"]),
            (
                dictgen::InsensitiveStr::Ascii("respecitely"),
                &["respectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alltogther"),
                &["altogether"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempatied"),
                &["templatized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloisonay"), &["cloisonné"]),
            (dictgen::InsensitiveStr::Ascii("varialbles"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("symmeterized"),
                &["symmetrized"],
            ),
            (dictgen::InsensitiveStr::Ascii("unpoened"), &["unopened"]),
            (dictgen::InsensitiveStr::Ascii("takign"), &["taking"]),
            (dictgen::InsensitiveStr::Ascii("follwwiwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("describng"), &["describing"]),
            (dictgen::InsensitiveStr::Ascii("prevously"), &["previously"]),
            (dictgen::InsensitiveStr::Ascii("recal"), &["recall"]),
            (dictgen::InsensitiveStr::Ascii("waas"), &["was"]),
            (
                dictgen::InsensitiveStr::Ascii("authenaticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("serialzed"), &["serialized"]),
            (
                dictgen::InsensitiveStr::Ascii("interrumping"),
                &["interrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("elimiante"), &["eliminate"]),
            (dictgen::InsensitiveStr::Ascii("definiton"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("hsitorically"),
                &["historically"],
            ),
            (dictgen::InsensitiveStr::Ascii("promtply"), &["promptly"]),
            (dictgen::InsensitiveStr::Ascii("woukd"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("nedless"),
                &["needless", "needles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("principales"),
                &["principles"],
            ),
            (dictgen::InsensitiveStr::Ascii("soalris"), &["solaris"]),
            (dictgen::InsensitiveStr::Ascii("neees"), &["knees", "needs"]),
            (dictgen::InsensitiveStr::Ascii("diaganol"), &["diagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("differentiaters"),
                &["differentiators", "differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disintereted"),
                &["disinterested"],
            ),
            (dictgen::InsensitiveStr::Ascii("badnits"), &["bandits"]),
            (
                dictgen::InsensitiveStr::Ascii("deprectates"),
                &["deprecates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificsting"),
                &["certificating"],
            ),
            (dictgen::InsensitiveStr::Ascii("authroring"), &["authoring"]),
            (dictgen::InsensitiveStr::Ascii("crusive"), &["cursive"]),
            (
                dictgen::InsensitiveStr::Ascii("disconeecting"),
                &["disconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("navagation"),
                &["navigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("supppress"), &["suppress"]),
            (
                dictgen::InsensitiveStr::Ascii("acclerating"),
                &["accelerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("publusher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("fonctionality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("styiled"), &["styled"]),
            (dictgen::InsensitiveStr::Ascii("algorihm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("tittling"), &["titling"]),
            (
                dictgen::InsensitiveStr::Ascii("creaed"),
                &["created", "creased", "creaked", "creamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocorret"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("oreding"), &["ordering"]),
            (dictgen::InsensitiveStr::Ascii("simualted"), &["simulated"]),
            (dictgen::InsensitiveStr::Ascii("becom"), &["become"]),
            (dictgen::InsensitiveStr::Ascii("specifcy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("forensisch"), &["forensic"]),
            (dictgen::InsensitiveStr::Ascii("porfiles"), &["profiles"]),
            (dictgen::InsensitiveStr::Ascii("benchmkar"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("continuousle"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailsation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origonated"),
                &["originated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insurasnces"),
                &["insurances"],
            ),
            (dictgen::InsensitiveStr::Ascii("nohing"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("pard"), &["part"]),
            (dictgen::InsensitiveStr::Ascii("normallly"), &["normally"]),
            (dictgen::InsensitiveStr::Ascii("publihser"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("pyschedelics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("ninimum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("publisehd"), &["published"]),
            (
                dictgen::InsensitiveStr::Ascii("transvormed"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nastersiems"),
                &["nasturtiums"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcion"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("conclusivley"),
                &["conclusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvent"),
                &["improvement", "improvident"],
            ),
            (dictgen::InsensitiveStr::Ascii("corret"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("dimissed"), &["dismissed"]),
            (
                dictgen::InsensitiveStr::Ascii("comformances"),
                &["conformances"],
            ),
            (dictgen::InsensitiveStr::Ascii("curage"), &["courage"]),
            (
                dictgen::InsensitiveStr::Ascii("considdering"),
                &["considering"],
            ),
            (dictgen::InsensitiveStr::Ascii("wesbite"), &["website"]),
            (
                dictgen::InsensitiveStr::Ascii("snapshoting"),
                &["snapshotting"],
            ),
            (dictgen::InsensitiveStr::Ascii("whilw"), &["while"]),
            (
                dictgen::InsensitiveStr::Ascii("acumulators"),
                &["accumulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("paradym"), &["paradigm"]),
            (
                dictgen::InsensitiveStr::Ascii("experementelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("failiing"), &["failing"]),
            (
                dictgen::InsensitiveStr::Ascii("institutuion"),
                &["institution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disingenious"),
                &["disingenuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("propmtly"), &["promptly"]),
            (
                dictgen::InsensitiveStr::Ascii("fundumentalists"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("voilently"), &["violently"]),
            (
                dictgen::InsensitiveStr::Ascii("instrcution"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retquireseek"),
                &["requireseek"],
            ),
            (dictgen::InsensitiveStr::Ascii("gategory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("ouputted"), &["outputted"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferenccing"),
                &["dereferencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("estabilish"), &["establish"]),
            (
                dictgen::InsensitiveStr::Ascii("bastracted"),
                &["abstracted"],
            ),
            (dictgen::InsensitiveStr::Ascii("caffiene"), &["caffeine"]),
            (
                dictgen::InsensitiveStr::Ascii("overlanded"),
                &["overloaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilizatoin"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanitizisation"),
                &["sanitization"],
            ),
            (dictgen::InsensitiveStr::Ascii("uptions"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("cusomising"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhaned"), &["enhanced"]),
            (dictgen::InsensitiveStr::Ascii("therefour"), &["therefor"]),
            (
                dictgen::InsensitiveStr::Ascii("delections"),
                &["detections", "deletions", "selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("teching"), &["teaching"]),
            (
                dictgen::InsensitiveStr::Ascii("parallellism"),
                &["parallelism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerimenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("josn"), &["json"]),
            (dictgen::InsensitiveStr::Ascii("speperatd"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("reccommends"),
                &["recommends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutue"),
                &["instituted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticicing"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repsective"),
                &["respective"],
            ),
            (dictgen::InsensitiveStr::Ascii("vegitable"), &["vegetable"]),
            (dictgen::InsensitiveStr::Ascii("balue"), &["blue", "value"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplayer"),
                &["multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("elction"), &["election"]),
            (
                dictgen::InsensitiveStr::Ascii("fertilizier"),
                &["fertilizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disussing"),
                &["discussing", "disusing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterises"),
                &["parameterises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peripheria"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formatteded"),
                &["formatted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neurosicence"),
                &["neuroscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermakert"),
                &["supermarkets"],
            ),
            (dictgen::InsensitiveStr::Ascii("conditoin"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("initilises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanenters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispaying"), &["displaying"]),
            (dictgen::InsensitiveStr::Ascii("becauce"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("bookkeeing"),
                &["bookkeeping"],
            ),
            (dictgen::InsensitiveStr::Ascii("ediiting"), &["editing"]),
            (
                dictgen::InsensitiveStr::Ascii("prostituate"),
                &["prostitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifieers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("profier"), &["profiler"]),
            (dictgen::InsensitiveStr::Ascii("domension"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("mexicaness"), &["mexicans"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnurable"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtituting"),
                &["substituting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleisation"),
                &["parallelisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collitions"),
                &["collisions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extremites"),
                &["extremities"],
            ),
            (dictgen::InsensitiveStr::Ascii("caluclate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("agos"),
                &["ago", "ages", "egos", "lagos"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predjuiced"),
                &["prejudiced"],
            ),
            (dictgen::InsensitiveStr::Ascii("netowrks"), &["networks"]),
            (dictgen::InsensitiveStr::Ascii("aswage"), &["assuage"]),
            (dictgen::InsensitiveStr::Ascii("masturbe"), &["masturbate"]),
            (
                dictgen::InsensitiveStr::Ascii("anticiapte"),
                &["anticipate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applictions"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("complexs"), &["complexes"]),
            (
                dictgen::InsensitiveStr::Ascii("pragmatisch"),
                &["pragmatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remanes"),
                &["remains", "renames", "remakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucceeed"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("desinging"), &["designing"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsstraps"),
                &["bootstraps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorites"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassebmler"),
                &["disassembler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egenralized"),
                &["generalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("politicain"),
                &["politician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtractions"),
                &["abstractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocroseconds"),
                &["microseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectins"), &["sections"]),
            (
                dictgen::InsensitiveStr::Ascii("critisized"),
                &["criticized"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbirtrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("apppendage"), &["appendage"]),
            (dictgen::InsensitiveStr::Ascii("grapgics"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacutres"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kingergarten"),
                &["kindergarten"],
            ),
            (dictgen::InsensitiveStr::Ascii("simpicity"), &["simplicity"]),
            (dictgen::InsensitiveStr::Ascii("toggele"), &["toggle"]),
            (
                dictgen::InsensitiveStr::Ascii("impersinating"),
                &["impersonating"],
            ),
            (dictgen::InsensitiveStr::Ascii("virtula"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("appications"),
                &["applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excetutives"),
                &["executives"],
            ),
            (dictgen::InsensitiveStr::Ascii("reprot"), &["report"]),
            (dictgen::InsensitiveStr::Ascii("relaese"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("installe"),
                &["installer", "installed", "install"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocassionaly"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("brewerks"), &["brewers"]),
            (
                dictgen::InsensitiveStr::Ascii("initalisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multithreds"),
                &["multithreads"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laysered"),
                &["lasered", "layered"],
            ),
            (dictgen::InsensitiveStr::Ascii("couply"), &["couple"]),
            (dictgen::InsensitiveStr::Ascii("glicthes"), &["glitches"]),
            (
                dictgen::InsensitiveStr::Ascii("traspositions"),
                &["transpositions"],
            ),
            (dictgen::InsensitiveStr::Ascii("compiing"), &["compiling"]),
            (
                dictgen::InsensitiveStr::Ascii("infrantryman"),
                &["infantryman"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("torubleshoot"),
                &["troubleshoot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconfortable"),
                &["uncomfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disseration"),
                &["dissertation"],
            ),
            (dictgen::InsensitiveStr::Ascii("helpped"), &["helped"]),
            (
                dictgen::InsensitiveStr::Ascii("atriculate"),
                &["articulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("paletts"), &["palettes"]),
            (
                dictgen::InsensitiveStr::Ascii("acclerated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shtoped"),
                &["stopped", "shopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comppliance"),
                &["compliance"],
            ),
            (dictgen::InsensitiveStr::Ascii("merucry"), &["mercury"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamative"),
                &["performative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninitialsed"),
                &["uninitialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornograhpy"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serices"),
                &["services", "series"],
            ),
            (dictgen::InsensitiveStr::Ascii("newslines"), &["newlines"]),
            (dictgen::InsensitiveStr::Ascii("heigt"), &["height"]),
            (
                dictgen::InsensitiveStr::Ascii("earthquakers"),
                &["earthquakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructons"),
                &["instructors", "instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecitons"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attritubes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriecates"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("muteces"), &["mutexes"]),
            (dictgen::InsensitiveStr::Ascii("particale"), &["particle"]),
            (
                dictgen::InsensitiveStr::Ascii("inifintely"),
                &["infinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("oter"), &["other", "otter"]),
            (
                dictgen::InsensitiveStr::Ascii("actiavtions"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakcaged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("milliseonds"),
                &["milliseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramemeter"),
                &["parameter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firest"),
                &["fires", "first"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approimates"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthuisasts"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("shuould"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("enchantents"),
                &["enchantments"],
            ),
            (dictgen::InsensitiveStr::Ascii("sortlst"), &["sortlist"]),
            (
                dictgen::InsensitiveStr::Ascii("explosivas"),
                &["explosives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attosenconds"),
                &["attoseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enterprice"),
                &["enterprise"],
            ),
            (dictgen::InsensitiveStr::Ascii("neuterd"), &["neutered"]),
            (
                dictgen::InsensitiveStr::Ascii("advantadges"),
                &["advantages"],
            ),
            (dictgen::InsensitiveStr::Ascii("comemnts"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("pomising"), &["promising"]),
            (
                dictgen::InsensitiveStr::Ascii("destructers"),
                &["destructors", "destructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("easiliy"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("subtlties"), &["subtitles"]),
            (
                dictgen::InsensitiveStr::Ascii("upsteramed"),
                &["upstreamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciprocoal"),
                &["reciprocal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschological"),
                &["psychological"],
            ),
            (dictgen::InsensitiveStr::Ascii("toekns"), &["tokens"]),
            (
                dictgen::InsensitiveStr::Ascii("interessting"),
                &["interesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("santourm"), &["santorum"]),
            (dictgen::InsensitiveStr::Ascii("magizine"), &["magazine"]),
            (dictgen::InsensitiveStr::Ascii("sematnics"), &["semantics"]),
            (dictgen::InsensitiveStr::Ascii("physcal"), &["physical"]),
            (dictgen::InsensitiveStr::Ascii("auotmata"), &["automata"]),
            (dictgen::InsensitiveStr::Ascii("napcakes"), &["pancakes"]),
            (
                dictgen::InsensitiveStr::Ascii("collecitons"),
                &["collections"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarrente"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("monserrat"), &["montserrat"]),
            (
                dictgen::InsensitiveStr::Ascii("complicato"),
                &["complication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certicicate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("catche"), &["catch"]),
            (
                dictgen::InsensitiveStr::Ascii("highligjts"),
                &["highlights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opeabcrations"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocaitions"),
                &["allocations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synomynous"),
                &["synonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("shortcats"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("mocrochip"), &["microchip"]),
            (dictgen::InsensitiveStr::Ascii("respnders"), &["responders"]),
            (dictgen::InsensitiveStr::Ascii("encryptio"), &["encryption"]),
            (dictgen::InsensitiveStr::Ascii("beucase"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("underfaker"),
                &["undertaker"],
            ),
            (dictgen::InsensitiveStr::Ascii("diector"), &["director"]),
            (dictgen::InsensitiveStr::Ascii("derivces"), &["derives"]),
            (
                dictgen::InsensitiveStr::Ascii("accessoire"),
                &["accessories", "accessory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentataion"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprentince"),
                &["apprentice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condicted"),
                &["conducted", "convicted"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisition"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("initialsations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("commamd"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("permutaion"),
                &["permutation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommendes"),
                &["recommends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("broacasted"),
                &["broadcast", "broadcasted"],
            ),
            (dictgen::InsensitiveStr::Ascii("delelte"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("croucing"), &["crouching"]),
            (
                dictgen::InsensitiveStr::Ascii("bootstapping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("urnaium"), &["uranium"]),
            (
                dictgen::InsensitiveStr::Ascii("respecitvely"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewritte"), &["rewrite"]),
            (dictgen::InsensitiveStr::Ascii("chateao"), &["château"]),
            (dictgen::InsensitiveStr::Ascii("bicthes"), &["bitches"]),
            (
                dictgen::InsensitiveStr::Ascii("idiologically"),
                &["ideologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("swichers"), &["switchers"]),
            (dictgen::InsensitiveStr::Ascii("quereis"), &["queries"]),
            (dictgen::InsensitiveStr::Ascii("mathcup"), &["matchup"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstating"),
                &["demonstrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("speacing"), &["spacing"]),
            (
                dictgen::InsensitiveStr::Ascii("calcutlations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("perhasp"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("divertion"), &["diversion"]),
            (dictgen::InsensitiveStr::Ascii("invarse"), &["inverse"]),
            (dictgen::InsensitiveStr::Ascii("gltiches"), &["glitches"]),
            (dictgen::InsensitiveStr::Ascii("applcable"), &["applicable"]),
            (dictgen::InsensitiveStr::Ascii("peridioc"), &["periodic"]),
            (
                dictgen::InsensitiveStr::Ascii("allocater"),
                &["allocator", "allocated", "allocates", "allocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpendence"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("dedeuplicate"),
                &["deduplicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("systhemmemory"),
                &["systemmemory"],
            ),
            (dictgen::InsensitiveStr::Ascii("egytpian"), &["egyptian"]),
            (dictgen::InsensitiveStr::Ascii("manule"), &["manual"]),
            (dictgen::InsensitiveStr::Ascii("displaing"), &["displaying"]),
            (dictgen::InsensitiveStr::Ascii("shouw"), &["show"]),
            (dictgen::InsensitiveStr::Ascii("napomed"), &["napalmed"]),
            (dictgen::InsensitiveStr::Ascii("exeutable"), &["executable"]),
            (
                dictgen::InsensitiveStr::Ascii("portneitaly"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posativity"),
                &["positivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("limtation"), &["limitation"]),
            (
                dictgen::InsensitiveStr::Ascii("reconsiled"),
                &["reconsider"],
            ),
            (dictgen::InsensitiveStr::Ascii("previsous"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("begininng"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("fomulae"), &["formulae"]),
            (dictgen::InsensitiveStr::Ascii("adbucts"), &["abducts"]),
            (dictgen::InsensitiveStr::Ascii("perorders"), &["preorders"]),
            (dictgen::InsensitiveStr::Ascii("repedable"), &["repeatable"]),
            (
                dictgen::InsensitiveStr::Ascii("volonteered"),
                &["volunteered"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimissive"), &["dismissive"]),
            (dictgen::InsensitiveStr::Ascii("repsots"), &["reposts"]),
            (dictgen::InsensitiveStr::Ascii("malicuous"), &["malicious"]),
            (
                dictgen::InsensitiveStr::Ascii("overruning"),
                &["overrunning"],
            ),
            (dictgen::InsensitiveStr::Ascii("emabssy"), &["embassy"]),
            (dictgen::InsensitiveStr::Ascii("polinator"), &["pollinator"]),
            (dictgen::InsensitiveStr::Ascii("relocaites"), &["relocates"]),
            (
                dictgen::InsensitiveStr::Ascii("propertyst"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decoratin"),
                &["decorations"],
            ),
            (dictgen::InsensitiveStr::Ascii("denile"), &["denial"]),
            (
                dictgen::InsensitiveStr::Ascii("supersition"),
                &["supervision"],
            ),
            (dictgen::InsensitiveStr::Ascii("raotats"), &["rotates"]),
            (
                dictgen::InsensitiveStr::Ascii("denpendencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steroetype"),
                &["stereotype"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgroound"),
                &["background"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prospertity"),
                &["prosperity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inveitably"),
                &["inevitably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversimplifacation"),
                &["oversimplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("intervew"), &["interview"]),
            (dictgen::InsensitiveStr::Ascii("conffigs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("shifitng"), &["shifting"]),
            (dictgen::InsensitiveStr::Ascii("auhtority"), &["authority"]),
            (
                dictgen::InsensitiveStr::Ascii("abstaction"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("repells"), &["repels"]),
            (dictgen::InsensitiveStr::Ascii("repsected"), &["respected"]),
            (dictgen::InsensitiveStr::Ascii("corecable"), &["coercible"]),
            (
                dictgen::InsensitiveStr::Ascii("derpatologist"),
                &["dermatologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("structues"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("guarnateed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("agorithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("pebbels"), &["pebbles"]),
            (dictgen::InsensitiveStr::Ascii("milwuakee"), &["milwaukee"]),
            (
                dictgen::InsensitiveStr::Ascii("somethinig"),
                &["somethings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paschurization"),
                &["pasteurization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soure"),
                &["source", "sure", "sore", "sour", "soured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disaggretated"),
                &["disaggregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancelas"), &["cancels"]),
            (
                dictgen::InsensitiveStr::Ascii("massachusents"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoustics"), &["acoustics"]),
            (dictgen::InsensitiveStr::Ascii("dispenced"), &["dispensed"]),
            (
                dictgen::InsensitiveStr::Ascii("wors"),
                &["wars", "was", "works", "worse", "worst"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statmenets"),
                &["statements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underminging"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nessecerily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pewder"),
                &["lewder", "pewter", "powder"],
            ),
            (dictgen::InsensitiveStr::Ascii("peppermit"), &["peppermint"]),
            (
                dictgen::InsensitiveStr::Ascii("neccessarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("generlly"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("pretious"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("reputposed"),
                &["repurposed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propected"),
                &["projected", "prospected", "protected"],
            ),
            (dictgen::InsensitiveStr::Ascii("localy"), &["locally"]),
            (
                dictgen::InsensitiveStr::Ascii("desctructor"),
                &["destructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copyrighed"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentated"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmen"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disporportionatly"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("equpping"), &["equipping"]),
            (dictgen::InsensitiveStr::Ascii("incunabla"), &["incunabula"]),
            (dictgen::InsensitiveStr::Ascii("argessive"), &["aggressive"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("codepoitn"), &["codepoint"]),
            (
                dictgen::InsensitiveStr::Ascii("manifacturing"),
                &["manufacturing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heades"),
                &["headers", "heads"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceckpoint"), &["checkpoint"]),
            (dictgen::InsensitiveStr::Ascii("analysisto"), &["analysts"]),
            (
                dictgen::InsensitiveStr::Ascii("precictions"),
                &["predictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practicaclly"),
                &["practically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrequantcy"),
                &["infrequency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interaktively"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subconchus"),
                &["subconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("intger"), &["integer"]),
            (dictgen::InsensitiveStr::Ascii("cutomized"), &["customized"]),
            (dictgen::InsensitiveStr::Ascii("sentenal"), &["sentinel"]),
            (
                dictgen::InsensitiveStr::Ascii("interacions"),
                &["interactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comnpleting"),
                &["completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egenralize"),
                &["generalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("digitisin"), &["digitising"]),
            (
                dictgen::InsensitiveStr::Ascii("simpification"),
                &["simplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparements"),
                &["compartments", "comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaul"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("oultiner"), &["outliner"]),
            (
                dictgen::InsensitiveStr::Ascii("consttructors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpsectors"),
                &["inspectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ganerations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acccelerate"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramameter"),
                &["parameter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combainatorics"),
                &["combinatorics"],
            ),
            (dictgen::InsensitiveStr::Ascii("obediant"), &["obedient"]),
            (dictgen::InsensitiveStr::Ascii("earlist"), &["earliest"]),
            (dictgen::InsensitiveStr::Ascii("meaninng"), &["meaning"]),
            (
                dictgen::InsensitiveStr::Ascii("practibable"),
                &["practicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("siftware"), &["software"]),
            (dictgen::InsensitiveStr::Ascii("mamage"), &["manage"]),
            (dictgen::InsensitiveStr::Ascii("alevt"), &["alert"]),
            (dictgen::InsensitiveStr::Ascii("bemusemnt"), &["bemusement"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerarbility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabiliies"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appaise"),
                &["appease", "appraise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicabel"),
                &["applicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indiviudal"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjcent"), &["adjacent"]),
            (dictgen::InsensitiveStr::Ascii("sementics"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("loockup"),
                &["lockup", "lookup"],
            ),
            (dictgen::InsensitiveStr::Ascii("pigin"), &["pigeon"]),
            (dictgen::InsensitiveStr::Ascii("sucesors"), &["successors"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibilites"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsbiility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesulted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("renassaince"),
                &["renaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identtifying"),
                &["identifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weightlifing"),
                &["weightlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eneumerations"),
                &["enumerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajacentcy"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("assited"), &["assisted"]),
            (
                dictgen::InsensitiveStr::Ascii("superiorest"),
                &["superiors"],
            ),
            (dictgen::InsensitiveStr::Ascii("canoical"), &["canonical"]),
            (
                dictgen::InsensitiveStr::Ascii("descprition"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromesomes"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouyant"), &["buoyant"]),
            (dictgen::InsensitiveStr::Ascii("vertabim"), &["verbatim"]),
            (dictgen::InsensitiveStr::Ascii("doosmday"), &["doomsday"]),
            (
                dictgen::InsensitiveStr::Ascii("implemntations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subconscius"),
                &["subconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("castels"), &["castles"]),
            (dictgen::InsensitiveStr::Ascii("categoy"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("prehaps"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("subscrie"), &["subscriber"]),
            (
                dictgen::InsensitiveStr::Ascii("sepulchraly"),
                &["sepulchrally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redicilous"),
                &["ridiculous"],
            ),
            (dictgen::InsensitiveStr::Ascii("notticed"), &["noticed"]),
            (dictgen::InsensitiveStr::Ascii("prodecure"), &["procedure"]),
            (dictgen::InsensitiveStr::Ascii("clense"), &["cleanse"]),
            (
                dictgen::InsensitiveStr::Ascii("congrassman"),
                &["congressman"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trandional"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("asembler"), &["assembler"]),
            (dictgen::InsensitiveStr::Ascii("typicially"), &["typically"]),
            (dictgen::InsensitiveStr::Ascii("poriferal"), &["peripheral"]),
            (dictgen::InsensitiveStr::Ascii("thses"), &["these"]),
            (dictgen::InsensitiveStr::Ascii("skippps"), &["skips"]),
            (
                dictgen::InsensitiveStr::Ascii("reuirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archietectures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mikroseconds"),
                &["microseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ridiculouness"),
                &["ridiculousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulatneously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrmation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supressible"),
                &["suppressible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscociative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmenetal"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforece"), &["enforce"]),
            (dictgen::InsensitiveStr::Ascii("systemn"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("calculte"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("anlaytics"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("symmerty"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoziating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("repaced"), &["replaced"]),
            (dictgen::InsensitiveStr::Ascii("pregant"), &["pregnant"]),
            (dictgen::InsensitiveStr::Ascii("stremaed"), &["streamed"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangtements"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("exapands"), &["expands"]),
            (dictgen::InsensitiveStr::Ascii("inteacted"), &["interacted"]),
            (
                dictgen::InsensitiveStr::Ascii("interrution"),
                &["interruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("reneagde"), &["renegade"]),
            (dictgen::InsensitiveStr::Ascii("scirpt"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("dirve"), &["drive"]),
            (dictgen::InsensitiveStr::Ascii("staf"), &["staff"]),
            (dictgen::InsensitiveStr::Ascii("mortailty"), &["mortality"]),
            (dictgen::InsensitiveStr::Ascii("gerneric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("bandwitdh"), &["bandwidth"]),
            (dictgen::InsensitiveStr::Ascii("negotinate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("opimized"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("polulations"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predomenantly"),
                &["predominantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("prfilers"), &["profilers"]),
            (dictgen::InsensitiveStr::Ascii("officaly"), &["officially"]),
            (
                dictgen::InsensitiveStr::Ascii("esteblishing"),
                &["establishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("poineer"), &["pioneer"]),
            (
                dictgen::InsensitiveStr::Ascii("underestimateing"),
                &["underestimating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obtaien"),
                &["obtain", "obtained"],
            ),
            (dictgen::InsensitiveStr::Ascii("wresseled"), &["wrestled"]),
            (dictgen::InsensitiveStr::Ascii("blantant"), &["blatant"]),
            (dictgen::InsensitiveStr::Ascii("resticted"), &["restricted"]),
            (dictgen::InsensitiveStr::Ascii("heloer"), &["helper"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibitivo"),
                &["prohibition"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabe"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("litarally"), &["literally"]),
            (dictgen::InsensitiveStr::Ascii("higlight"), &["highlight"]),
            (dictgen::InsensitiveStr::Ascii("boun"), &["bound"]),
            (dictgen::InsensitiveStr::Ascii("hadle"), &["handle"]),
            (dictgen::InsensitiveStr::Ascii("unshfit"), &["unshift"]),
            (dictgen::InsensitiveStr::Ascii("intgrate"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("commna"), &["comma"]),
            (dictgen::InsensitiveStr::Ascii("minium"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("conntroller"),
                &["controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("dprecated"), &["deprecated"]),
            (dictgen::InsensitiveStr::Ascii("reagrd"), &["regard"]),
            (dictgen::InsensitiveStr::Ascii("fiw"), &["fix"]),
            (
                dictgen::InsensitiveStr::Ascii("plesing"),
                &["pleasing", "blessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insances"),
                &["instances", "insurances"],
            ),
            (dictgen::InsensitiveStr::Ascii("enttries"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("latst"), &["latest"]),
            (dictgen::InsensitiveStr::Ascii("collapes"), &["collapse"]),
            (
                dictgen::InsensitiveStr::Ascii("initialializations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountt"),
                &["accountant", "account"],
            ),
            (dictgen::InsensitiveStr::Ascii("afterr"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("writed"),
                &["wrote", "written", "write", "writer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentensing"),
                &["sentencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjency"),
                &["adjacency", "agency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misdameanors"),
                &["misdemeanors"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutuallly"), &["mutually"]),
            (
                dictgen::InsensitiveStr::Ascii("strutcures"),
                &["structures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforcment"),
                &["enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("staic"), &["static"]),
            (dictgen::InsensitiveStr::Ascii("manaise"), &["mayonnaise"]),
            (
                dictgen::InsensitiveStr::Ascii("imitatin"),
                &["imitating", "imitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("leprachan"), &["leprechaun"]),
            (dictgen::InsensitiveStr::Ascii("organizate"), &["organize"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("otu"), &["out"]),
            (
                dictgen::InsensitiveStr::Ascii("xinitiazlize"),
                &["xinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("zerp"), &["zero"]),
            (
                dictgen::InsensitiveStr::Ascii("unfourtunately"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("encourae"), &["encourage"]),
            (
                dictgen::InsensitiveStr::Ascii("chromasome"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("africanos"), &["africans"]),
            (dictgen::InsensitiveStr::Ascii("precize"), &["precise"]),
            (dictgen::InsensitiveStr::Ascii("seseeding"), &["seceding"]),
            (dictgen::InsensitiveStr::Ascii("opertaors"), &["operators"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefenite"),
                &["indefinite"],
            ),
            (dictgen::InsensitiveStr::Ascii("buitin"), &["builtin"]),
            (dictgen::InsensitiveStr::Ascii("alocator"), &["allocator"]),
            (dictgen::InsensitiveStr::Ascii("gymnistic"), &["gymnastic"]),
            (dictgen::InsensitiveStr::Ascii("asigned"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("promenantly"),
                &["prominently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unforetunate"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funtionalty"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("implace"), &["inplace"]),
            (dictgen::InsensitiveStr::Ascii("howver"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("reactiving"),
                &["reactivating"],
            ),
            (dictgen::InsensitiveStr::Ascii("combiator"), &["combinator"]),
            (dictgen::InsensitiveStr::Ascii("lovley"), &["lovely"]),
            (dictgen::InsensitiveStr::Ascii("upgradding"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("secunds"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("realitive"), &["relative"]),
            (dictgen::InsensitiveStr::Ascii("bardcode"), &["barcode"]),
            (
                dictgen::InsensitiveStr::Ascii("unambigiously"),
                &["unambiguously"],
            ),
            (dictgen::InsensitiveStr::Ascii("protray"), &["portray"]),
            (
                dictgen::InsensitiveStr::Ascii("exchagning"),
                &["exchanging"],
            ),
            (dictgen::InsensitiveStr::Ascii("cordoroy"), &["corduroy"]),
            (
                dictgen::InsensitiveStr::Ascii("archytypes"),
                &["archetypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrollin"), &["scrolling"]),
            (
                dictgen::InsensitiveStr::Ascii("multithred"),
                &["multithread"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufocating"),
                &["suffocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recogising"),
                &["recognising"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("convulutions"),
                &["convolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pusehd"), &["pushed"]),
            (
                dictgen::InsensitiveStr::Ascii("engineeering"),
                &["engineering"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarentes"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("configurtes"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactivo"),
                &["interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("ultraound"), &["ultrasound"]),
            (
                dictgen::InsensitiveStr::Ascii("permessions"),
                &["permissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("hepler"), &["helper"]),
            (dictgen::InsensitiveStr::Ascii("unnowns"), &["unknowns"]),
            (dictgen::InsensitiveStr::Ascii("fomatted"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("apprecitae"),
                &["appreciates", "appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliverys"),
                &["delivers", "deliveries"],
            ),
            (dictgen::InsensitiveStr::Ascii("cilck"), &["click"]),
            (
                dictgen::InsensitiveStr::Ascii("alforithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriotisch"),
                &["patriotic"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderr"), &["render"]),
            (
                dictgen::InsensitiveStr::Ascii("encahnting"),
                &["enchanting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legionair"),
                &["legionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("anme"), &["name", "anime"]),
            (
                dictgen::InsensitiveStr::Ascii("assymetric"),
                &["asymmetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefference"),
                &["preference"],
            ),
            (dictgen::InsensitiveStr::Ascii("dessigns"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("modication"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("agreemnts"), &["agreements"]),
            (
                dictgen::InsensitiveStr::Ascii("ethicallity"),
                &["ethically"],
            ),
            (dictgen::InsensitiveStr::Ascii("finace"), &["finance"]),
            (
                dictgen::InsensitiveStr::Ascii("threatended"),
                &["threatened"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosture"), &["moisture"]),
            (
                dictgen::InsensitiveStr::Ascii("congugated"),
                &["conjugated"],
            ),
            (dictgen::InsensitiveStr::Ascii("wroldly"), &["worldly"]),
            (dictgen::InsensitiveStr::Ascii("hosthot"), &["hotshot"]),
            (
                dictgen::InsensitiveStr::Ascii("conversley"),
                &["conversely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsiquent"),
                &["subsequent"],
            ),
            (dictgen::InsensitiveStr::Ascii("diaplay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("utlising"), &["utilising"]),
            (dictgen::InsensitiveStr::Ascii("subtabels"), &["subtables"]),
            (dictgen::InsensitiveStr::Ascii("scrpits"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("pleasd"), &["pleased"]),
            (dictgen::InsensitiveStr::Ascii("chekcbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("senesitive"), &["sensitive"]),
            (
                dictgen::InsensitiveStr::Ascii("seceed"),
                &["succeed", "secede"],
            ),
            (dictgen::InsensitiveStr::Ascii("ojbect"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("reproted"), &["reported"]),
            (
                dictgen::InsensitiveStr::Ascii("enmvironments"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troubeshooter"),
                &["troubleshooter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooperly"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("thoeretical"),
                &["theoretical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parellelised"),
                &["parallelised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noticied"),
                &["noticed", "notified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siloweted"),
                &["silhouetted"],
            ),
            (dictgen::InsensitiveStr::Ascii("multuply"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("whicg"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("anothr"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("indivual"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("extracto"), &["extraction"]),
            (
                dictgen::InsensitiveStr::Ascii("controversials"),
                &["controversial"],
            ),
            (dictgen::InsensitiveStr::Ascii("interators"), &["iterators"]),
            (dictgen::InsensitiveStr::Ascii("varriance"), &["variance"]),
            (
                dictgen::InsensitiveStr::Ascii("rearranments"),
                &["rearrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwriding"),
                &["overwriting", "overriding"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancelles"), &["cancels"]),
            (dictgen::InsensitiveStr::Ascii("quesetions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("resepectively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("alechmist"), &["alchemist"]),
            (
                dictgen::InsensitiveStr::Ascii("comisions"),
                &["commissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitions"),
                &["conditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("optoins"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("criticms"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("dezinens"), &["denizens"]),
            (
                dictgen::InsensitiveStr::Ascii("conformace"),
                &["conformance"],
            ),
            (dictgen::InsensitiveStr::Ascii("rference"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("tailord"), &["tailored"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnportation"),
                &["transportation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutliple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("onlye"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("signifigant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("invisivle"), &["invisible"]),
            (
                dictgen::InsensitiveStr::Ascii("execustions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cusomisation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionnairs"),
                &["questionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("revolvin"), &["revolving"]),
            (
                dictgen::InsensitiveStr::Ascii("notigication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("tabblow"), &["tableau"]),
            (dictgen::InsensitiveStr::Ascii("reqiest"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("placin"), &["placing"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpreting"),
                &["interpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("casuation"), &["causation"]),
            (
                dictgen::InsensitiveStr::Ascii("combinaison"),
                &["combinations", "combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyising"), &["analysing"]),
            (dictgen::InsensitiveStr::Ascii("pipeine"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("landscspe"), &["landscapes"]),
            (dictgen::InsensitiveStr::Ascii("essensial"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("bastardisin"),
                &["bastardising"],
            ),
            (dictgen::InsensitiveStr::Ascii("nativly"), &["natively"]),
            (dictgen::InsensitiveStr::Ascii("orider"), &["order"]),
            (dictgen::InsensitiveStr::Ascii("sapphrie"), &["sapphire"]),
            (dictgen::InsensitiveStr::Ascii("blatent"), &["blatant"]),
            (dictgen::InsensitiveStr::Ascii("calcutate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("ostrasiced"),
                &["ostracized"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumbitted"), &["submitted"]),
            (
                dictgen::InsensitiveStr::Ascii("unadvertedly"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevititably"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("zookeper"), &["zookeeper"]),
            (
                dictgen::InsensitiveStr::Ascii("alternateve"),
                &["alternative"],
            ),
            (dictgen::InsensitiveStr::Ascii("forfited"), &["forfeited"]),
            (
                dictgen::InsensitiveStr::Ascii("echancement"),
                &["enhancement"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirucal"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("comparrable"),
                &["comparable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucruring"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorprating"),
                &["incorporating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pupulation"),
                &["population"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("voyouristically"),
                &["voyeuristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("territoral"),
                &["territorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrightty"), &["alrighty"]),
            (dictgen::InsensitiveStr::Ascii("perfomed"), &["performed"]),
            (
                dictgen::InsensitiveStr::Ascii("objectivst"),
                &["objectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentating"),
                &["experimentation", "experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("worser"), &["worse"]),
            (dictgen::InsensitiveStr::Ascii("exerciesed"), &["exercised"]),
            (
                dictgen::InsensitiveStr::Ascii("airpost"),
                &["airports", "airport"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neeedless"),
                &["needless", "needles"],
            ),
            (dictgen::InsensitiveStr::Ascii("preshusly"), &["preciously"]),
            (dictgen::InsensitiveStr::Ascii("toughtful"), &["thoughtful"]),
            (dictgen::InsensitiveStr::Ascii("expetion"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("lsat"),
                &["last", "slat", "sat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coulumn"),
                &["column", "coulomb"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperiecned"),
                &["inexperience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotionation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("quetsions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("contibutors"),
                &["contributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transprots"),
                &["transports"],
            ),
            (dictgen::InsensitiveStr::Ascii("languege"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("disatvantaged"),
                &["disadvantaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("invalidte"), &["invalidate"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalsim"),
                &["nationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulatories"),
                &["regulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indianaoplis"),
                &["indianapolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collapsable"),
                &["collapsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstruction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("booda"), &["buddha"]),
            (
                dictgen::InsensitiveStr::Ascii("tempatised"),
                &["templatised"],
            ),
            (dictgen::InsensitiveStr::Ascii("biulds"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("templaet"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("imbraser"), &["embracer"]),
            (dictgen::InsensitiveStr::Ascii("artsits"), &["artists"]),
            (
                dictgen::InsensitiveStr::Ascii("comercials"),
                &["commercials"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysator"), &["analyser"]),
            (dictgen::InsensitiveStr::Ascii("appendend"), &["appended"]),
            (dictgen::InsensitiveStr::Ascii("truthfuly"), &["truthfully"]),
            (dictgen::InsensitiveStr::Ascii("cartiesian"), &["cartesian"]),
            (
                dictgen::InsensitiveStr::Ascii("expementations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renassiance"),
                &["renaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("memwar"), &["memoir"]),
            (
                dictgen::InsensitiveStr::Ascii("depolyment"),
                &["deployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equlity"),
                &["equality", "equity"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrct"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("sepererated"),
                &["separated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obectively"),
                &["objectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("begginig"),
                &["begging", "beginning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contolller"),
                &["controller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protaganist"),
                &["protagonist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comrpomise"),
                &["compromise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplened"),
                &["disciplined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shepperded"),
                &["shepherded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponsorshop"),
                &["sponsorship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathalogic"),
                &["pathologic"],
            ),
            (dictgen::InsensitiveStr::Ascii("trun"), &["turn"]),
            (dictgen::InsensitiveStr::Ascii("jurnies"), &["journeys"]),
            (
                dictgen::InsensitiveStr::Ascii("partcipating"),
                &["participating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concanete"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lamda"), &["lambda"]),
            (dictgen::InsensitiveStr::Ascii("wiew"), &["view"]),
            (dictgen::InsensitiveStr::Ascii("occurence"), &["occurrence"]),
            (dictgen::InsensitiveStr::Ascii("removeable"), &["removable"]),
            (dictgen::InsensitiveStr::Ascii("symblic"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("innitially"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("tempromental"),
                &["temperamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("bizzare"), &["bizarre"]),
            (dictgen::InsensitiveStr::Ascii("maitained"), &["maintained"]),
            (dictgen::InsensitiveStr::Ascii("zemporary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("junlging"), &["jungling"]),
            (dictgen::InsensitiveStr::Ascii("adovcates"), &["advocates"]),
            (
                dictgen::InsensitiveStr::Ascii("credencials"),
                &["credentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmenet"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("fallign"), &["falling"]),
            (dictgen::InsensitiveStr::Ascii("tangetial"), &["tangential"]),
            (dictgen::InsensitiveStr::Ascii("chivarly"), &["chivalry"]),
            (dictgen::InsensitiveStr::Ascii("eptier"), &["emptier"]),
            (dictgen::InsensitiveStr::Ascii("parctise"), &["practise"]),
            (dictgen::InsensitiveStr::Ascii("implet"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("pyhton"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatablie"),
                &["incompatible", "incompatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electronnegativity"),
                &["electronegativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("kackie"), &["khaki"]),
            (dictgen::InsensitiveStr::Ascii("pretier"), &["prettier"]),
            (dictgen::InsensitiveStr::Ascii("spellig"), &["spelling"]),
            (dictgen::InsensitiveStr::Ascii("dyshporia"), &["dysphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("ducumentation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stattistical"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increminting"),
                &["incrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("contense"), &["contenders"]),
            (dictgen::InsensitiveStr::Ascii("emplyed"), &["employed"]),
            (
                dictgen::InsensitiveStr::Ascii("rememberable"),
                &["memorable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premonasterians"),
                &["premonstratensians"],
            ),
            (dictgen::InsensitiveStr::Ascii("chanse"), &["chance"]),
            (dictgen::InsensitiveStr::Ascii("scarry"), &["scary"]),
            (dictgen::InsensitiveStr::Ascii("varians"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("servced"),
                &["serviced", "served"],
            ),
            (dictgen::InsensitiveStr::Ascii("plattforms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("requies"),
                &["requires", "requites", "requiems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imprevement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("deficites"), &["deficits"]),
            (
                dictgen::InsensitiveStr::Ascii("pregnanices"),
                &["pregnancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("intialy"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("aquaints"), &["acquaints"]),
            (dictgen::InsensitiveStr::Ascii("metrigal"), &["metrical"]),
            (dictgen::InsensitiveStr::Ascii("fromed"), &["formed"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallifers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuf"), &["stuff"]),
            (
                dictgen::InsensitiveStr::Ascii("anedoctally"),
                &["anecdotally"],
            ),
            (dictgen::InsensitiveStr::Ascii("variatino"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("corrsponing"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confirmatino"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regconises"),
                &["recognises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotions"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignesting"), &["ingesting"]),
            (dictgen::InsensitiveStr::Ascii("checkeing"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("cerators"), &["creators"]),
            (dictgen::InsensitiveStr::Ascii("namspace"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("deisplays"), &["displays"]),
            (
                dictgen::InsensitiveStr::Ascii("exaggareted"),
                &["exaggerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhooding"),
                &["neighbouring"],
            ),
            (dictgen::InsensitiveStr::Ascii("themeing"), &["theming"]),
            (
                dictgen::InsensitiveStr::Ascii("sociologial"),
                &["sociological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptins"),
                &["exceptions", "excepting"],
            ),
            (dictgen::InsensitiveStr::Ascii("languag"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("corresopndences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensitivty"),
                &["sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpertator"),
                &["perpetrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuables"), &["reusables"]),
            (dictgen::InsensitiveStr::Ascii("prisitne"), &["pristine"]),
            (
                dictgen::InsensitiveStr::Ascii("containinng"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proseletyzing"),
                &["proselytizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("keboards"), &["keyboards"]),
            (
                dictgen::InsensitiveStr::Ascii("subsecrion"),
                &["subsection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comisioner"),
                &["commissioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("kingdomers"), &["kingdoms"]),
            (
                dictgen::InsensitiveStr::Ascii("beliefes"),
                &["beliefs", "believes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presentaion"),
                &["presentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("doman"), &["domain"]),
            (
                dictgen::InsensitiveStr::Ascii("photografic"),
                &["photographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoroporation"),
                &["incorporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcripto"),
                &["transcription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriorate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("meccob"), &["macabre"]),
            (dictgen::InsensitiveStr::Ascii("nickame"), &["nickname"]),
            (dictgen::InsensitiveStr::Ascii("reroutin"), &["rerouting"]),
            (dictgen::InsensitiveStr::Ascii("contiue"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("disrepectful"),
                &["disrespectful"],
            ),
            (dictgen::InsensitiveStr::Ascii("steraming"), &["streaming"]),
            (
                dictgen::InsensitiveStr::Ascii("wrokaround"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experemental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("exibiting"), &["exhibiting"]),
            (
                dictgen::InsensitiveStr::Ascii("mutater"),
                &["mutator", "mutated", "mutates", "mutate"],
            ),
            (dictgen::InsensitiveStr::Ascii("benedicat"), &["benedict"]),
            (
                dictgen::InsensitiveStr::Ascii("miminising"),
                &["minimising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrevating"),
                &["aggravating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfification"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("aritsts"), &["artists"]),
            (
                dictgen::InsensitiveStr::Ascii("subcommnad"),
                &["subcommand"],
            ),
            (dictgen::InsensitiveStr::Ascii("meged"), &["merged"]),
            (dictgen::InsensitiveStr::Ascii("sumaries"), &["summaries"]),
            (
                dictgen::InsensitiveStr::Ascii("conclution"),
                &["conclusion"],
            ),
            (dictgen::InsensitiveStr::Ascii("doomdsay"), &["doomsday"]),
            (
                dictgen::InsensitiveStr::Ascii("umbelievable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifer"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("succient"), &["succinct"]),
            (
                dictgen::InsensitiveStr::Ascii("conversatino"),
                &["conservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasteurisin"),
                &["pasteurising"],
            ),
            (dictgen::InsensitiveStr::Ascii("asendance"), &["ascendance"]),
            (
                dictgen::InsensitiveStr::Ascii("agonsticism"),
                &["agnosticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("elenment"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("spontainously"),
                &["spontaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versinoing"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplcate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("inovation"), &["innovation"]),
            (dictgen::InsensitiveStr::Ascii("totats"), &["totals"]),
            (dictgen::InsensitiveStr::Ascii("retieved"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("adknowledges"),
                &["acknowledges"],
            ),
            (dictgen::InsensitiveStr::Ascii("moible"), &["mobile"]),
            (
                dictgen::InsensitiveStr::Ascii("simualtions"),
                &["simulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaules"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("collape"), &["collapse"]),
            (dictgen::InsensitiveStr::Ascii("yhat"), &["that"]),
            (
                dictgen::InsensitiveStr::Ascii("childres"),
                &["childrens", "children"],
            ),
            (dictgen::InsensitiveStr::Ascii("charactrs"), &["characters"]),
            (dictgen::InsensitiveStr::Ascii("inersting"), &["inserting"]),
            (
                dictgen::InsensitiveStr::Ascii("stragetically"),
                &["strategically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finness"),
                &["finesse", "fineness", "fitness"],
            ),
            (dictgen::InsensitiveStr::Ascii("syle"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("compososite"),
                &["composite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dereferencces"),
                &["dereferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounading"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("stompling"), &["stomping"]),
            (
                dictgen::InsensitiveStr::Ascii("jugdemental"),
                &["judgemental"],
            ),
            (dictgen::InsensitiveStr::Ascii("cretaed"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("attmpts"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("constactly"),
                &["constantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("antivuris"), &["antivirus"]),
            (dictgen::InsensitiveStr::Ascii("reyonlds"), &["reynolds"]),
            (dictgen::InsensitiveStr::Ascii("regulares"), &["regulars"]),
            (
                dictgen::InsensitiveStr::Ascii("conferming"),
                &["confirming"],
            ),
            (dictgen::InsensitiveStr::Ascii("guage"), &["gage"]),
            (dictgen::InsensitiveStr::Ascii("taveling"), &["traveling"]),
            (dictgen::InsensitiveStr::Ascii("prescense"), &["presence"]),
            (dictgen::InsensitiveStr::Ascii("feministas"), &["feminists"]),
            (dictgen::InsensitiveStr::Ascii("srceen"), &["screen"]),
            (
                dictgen::InsensitiveStr::Ascii("optimisitc"),
                &["optimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("annnoy"), &["annoy"]),
            (dictgen::InsensitiveStr::Ascii("physicion"), &["physician"]),
            (
                dictgen::InsensitiveStr::Ascii("contriubtes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitratily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradices"),
                &["contradicts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retransmition"),
                &["retransmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heartborne"),
                &["heartbroken"],
            ),
            (dictgen::InsensitiveStr::Ascii("registert"), &["registered"]),
            (
                dictgen::InsensitiveStr::Ascii("guantonamo"),
                &["guantanamo"],
            ),
            (dictgen::InsensitiveStr::Ascii("returms"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficified"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("optmizing"), &["optimizing"]),
            (dictgen::InsensitiveStr::Ascii("momment"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("exacberate"),
                &["exacerbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("montrel"), &["montreal"]),
            (
                dictgen::InsensitiveStr::Ascii("transmisssions"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbrio"), &["embryo"]),
            (
                dictgen::InsensitiveStr::Ascii("openen"),
                &["opened", "opening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similaraties"),
                &["similarities"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcaulte"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("straightfoward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("inhibate"), &["inhibit"]),
            (
                dictgen::InsensitiveStr::Ascii("equires"),
                &["requires", "enquires", "equines", "esquires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converve"),
                &["conserve", "converse", "converge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manfiestos"),
                &["manifestos"],
            ),
            (dictgen::InsensitiveStr::Ascii("difficutl"), &["difficult"]),
            (dictgen::InsensitiveStr::Ascii("sescede"), &["secede"]),
            (
                dictgen::InsensitiveStr::Ascii("transfomrations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("slowley"), &["slowly"]),
            (dictgen::InsensitiveStr::Ascii("disocver"), &["discover"]),
            (
                dictgen::InsensitiveStr::Ascii("specifactions"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocuntries"), &["countries"]),
            (
                dictgen::InsensitiveStr::Ascii("neigbhoring"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insurgencey"),
                &["insurgency"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagsing"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("uncreachable"),
                &["unreachable"],
            ),
            (dictgen::InsensitiveStr::Ascii("straigt"), &["straight"]),
            (dictgen::InsensitiveStr::Ascii("decocding"), &["decoding"]),
            (dictgen::InsensitiveStr::Ascii("suprsied"), &["surprised"]),
            (dictgen::InsensitiveStr::Ascii("rathcet"), &["ratchet"]),
            (
                dictgen::InsensitiveStr::Ascii("expermimenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("shelvers"), &["shelves"]),
            (dictgen::InsensitiveStr::Ascii("propvider"), &["provider"]),
            (dictgen::InsensitiveStr::Ascii("aexs"), &["axes"]),
            (
                dictgen::InsensitiveStr::Ascii("entretainment"),
                &["entertainment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficiifed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autherities"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rescourcing"),
                &["resourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recogition"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inheritence"),
                &["inheritance"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancedotal"), &["anecdotal"]),
            (
                dictgen::InsensitiveStr::Ascii("precedene"),
                &["precedence", "precedent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recquirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misintrepreting"),
                &["misinterpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtuing"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("varn"), &["warn"]),
            (dictgen::InsensitiveStr::Ascii("sintaks"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("muliplier"), &["multiplier"]),
            (
                dictgen::InsensitiveStr::Ascii("descritpion"),
                &["descriptions", "description"],
            ),
            (dictgen::InsensitiveStr::Ascii("selcted"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("croporations"),
                &["corporations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchromized"),
                &["synchronized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pensioen"), &["pension"]),
            (
                dictgen::InsensitiveStr::Ascii("commonweath"),
                &["commonwealth"],
            ),
            (dictgen::InsensitiveStr::Ascii("paremsan"), &["parmesan"]),
            (
                dictgen::InsensitiveStr::Ascii("simeltaneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indoensian"),
                &["indonesian"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfiable"), &["modifiable"]),
            (dictgen::InsensitiveStr::Ascii("diaresis"), &["diaeresis"]),
            (dictgen::InsensitiveStr::Ascii("instade"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("delimma"), &["dilemma"]),
            (dictgen::InsensitiveStr::Ascii("arrrival"), &["arrival"]),
            (
                dictgen::InsensitiveStr::Ascii("minoosha"),
                &["minutia", "minutiae"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perliminary"),
                &["preliminary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photografical"),
                &["photographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acccordance"),
                &["accordance"],
            ),
            (dictgen::InsensitiveStr::Ascii("varned"), &["warned"]),
            (
                dictgen::InsensitiveStr::Ascii("imcremented"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("symmectric"), &["symmetric"]),
            (dictgen::InsensitiveStr::Ascii("shifing"), &["shifting"]),
            (
                dictgen::InsensitiveStr::Ascii("intenationalism"),
                &["internationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("assocites"), &["associates"]),
            (dictgen::InsensitiveStr::Ascii("sytems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("explicilty"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyboad"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("argumends"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("sepaate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("makwfile"), &["makefile"]),
            (dictgen::InsensitiveStr::Ascii("previout"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("cick"), &["click"]),
            (dictgen::InsensitiveStr::Ascii("trotilla"), &["tortilla"]),
            (dictgen::InsensitiveStr::Ascii("capbaly"), &["capably"]),
            (
                dictgen::InsensitiveStr::Ascii("undersandable"),
                &["understandable"],
            ),
            (dictgen::InsensitiveStr::Ascii("possily"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("euivalents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("alinging"), &["aligning"]),
            (dictgen::InsensitiveStr::Ascii("subsidie"), &["subsidized"]),
            (
                dictgen::InsensitiveStr::Ascii("alternativos"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("universse"), &["universes"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibliy"),
                &["responsibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("columnms"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("scientficaly"),
                &["scientifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dowloaders"),
                &["downloaders"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighobrs"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("transitiond"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acccountants"),
                &["accountants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviorements"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyord"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("stregth"), &["strength"]),
            (
                dictgen::InsensitiveStr::Ascii("residentail"),
                &["residential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarasaing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfers"), &["prefers"]),
            (dictgen::InsensitiveStr::Ascii("neeeded"), &["needed"]),
            (
                dictgen::InsensitiveStr::Ascii("addresst"),
                &["address", "addressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannibalizm"),
                &["cannibalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("depolys"), &["deploys"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriete"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comnplexity"),
                &["complexity"],
            ),
            (dictgen::InsensitiveStr::Ascii("comedias"), &["comedians"]),
            (dictgen::InsensitiveStr::Ascii("horison"), &["horizon"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliacting"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("adiditon"), &["addition"]),
            (dictgen::InsensitiveStr::Ascii("quetions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("stupidiy"), &["stupidity"]),
            (
                dictgen::InsensitiveStr::Ascii("fucntionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articualtes"),
                &["articulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifical"),
                &["sacrificial"],
            ),
            (dictgen::InsensitiveStr::Ascii("imbrased"), &["embraced"]),
            (dictgen::InsensitiveStr::Ascii("enouh"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("owuld"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("formery"), &["formerly"]),
            (
                dictgen::InsensitiveStr::Ascii("reuplaoding"),
                &["reuploading"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuction"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("subtarcted"),
                &["subtracted"],
            ),
            (dictgen::InsensitiveStr::Ascii("initalise"), &["initialise"]),
            (
                dictgen::InsensitiveStr::Ascii("dispencing"),
                &["dispensing"],
            ),
            (dictgen::InsensitiveStr::Ascii("archviing"), &["archiving"]),
            (dictgen::InsensitiveStr::Ascii("pijun"), &["pigeon"]),
            (
                dictgen::InsensitiveStr::Ascii("ortographic"),
                &["orographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoruthms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("achiveing"), &["achieving"]),
            (
                dictgen::InsensitiveStr::Ascii("ejaculaton"),
                &["ejaculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troosos"),
                &["trousseaus", "trousseaux"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triscaidekafobia"),
                &["triskaidekaphobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("improrts"), &["imports"]),
            (dictgen::InsensitiveStr::Ascii("cycic"), &["cyclic"]),
            (
                dictgen::InsensitiveStr::Ascii("rhinosaruses"),
                &["rhinoceroses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progagates"),
                &["propagates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unedicated"),
                &["uneducated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggarated"),
                &["exaggerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soliditary"),
                &["solidarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horrendeus"),
                &["horrendous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefeid"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgnaised"), &["organised"]),
            (dictgen::InsensitiveStr::Ascii("neuton"), &["neutron"]),
            (dictgen::InsensitiveStr::Ascii("corspes"), &["corpses"]),
            (dictgen::InsensitiveStr::Ascii("monitos"), &["monitors"]),
            (
                dictgen::InsensitiveStr::Ascii("discreddit"),
                &["discredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miracoulus"),
                &["miraculous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lewtenants"),
                &["lieutenants"],
            ),
            (dictgen::InsensitiveStr::Ascii("expecetd"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("disassmebles"),
                &["disassembles"],
            ),
            (dictgen::InsensitiveStr::Ascii("exntending"), &["extending"]),
            (dictgen::InsensitiveStr::Ascii("bytetream"), &["bytestream"]),
            (
                dictgen::InsensitiveStr::Ascii("puplishing"),
                &["publishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difficaulties"),
                &["difficulties"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfifiers"), &["verifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("inititialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captitalized"),
                &["capitalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("stranget"), &["strangest"]),
            (
                dictgen::InsensitiveStr::Ascii("sertificate"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passionatley"),
                &["passionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("formtter"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("superstitous"),
                &["superstitious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("couse"),
                &["course", "cause"],
            ),
            (dictgen::InsensitiveStr::Ascii("tememetry"), &["telemetry"]),
            (dictgen::InsensitiveStr::Ascii("migare"), &["migrate"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphysicans"),
                &["metaphysics"],
            ),
            (dictgen::InsensitiveStr::Ascii("resposive"), &["responsive"]),
            (
                dictgen::InsensitiveStr::Ascii("hypocrytical"),
                &["hypocritical"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifyes"), &["verifies"]),
            (dictgen::InsensitiveStr::Ascii("alloction"), &["allocation"]),
            (dictgen::InsensitiveStr::Ascii("caliased"), &["aliased"]),
            (
                dictgen::InsensitiveStr::Ascii("impossiblity"),
                &["impossibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("kniw"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("multilied"), &["multiplied"]),
            (dictgen::InsensitiveStr::Ascii("atlantc"), &["atlantic"]),
            (dictgen::InsensitiveStr::Ascii("authrors"), &["authors"]),
            (dictgen::InsensitiveStr::Ascii("severl"), &["several"]),
            (dictgen::InsensitiveStr::Ascii("packege"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("temporries"),
                &["temporaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("responce"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("puplated"), &["populated"]),
            (dictgen::InsensitiveStr::Ascii("speciffic"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("expireitme"),
                &["expiretime"],
            ),
            (dictgen::InsensitiveStr::Ascii("necesity"), &["necessity"]),
            (dictgen::InsensitiveStr::Ascii("exameples"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("dionsaurs"), &["dinosaurs"]),
            (dictgen::InsensitiveStr::Ascii("cerainly"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("sensetive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("sorcercy"), &["sorcery"]),
            (
                dictgen::InsensitiveStr::Ascii("vigourosly"),
                &["vigorously"],
            ),
            (dictgen::InsensitiveStr::Ascii("limititers"), &["limiters"]),
            (dictgen::InsensitiveStr::Ascii("decorded"), &["decoded"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certficiate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("deocder"), &["decoder"]),
            (dictgen::InsensitiveStr::Ascii("imporatnt"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unneccesary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleaer"),
                &["clear", "clearer", "cleaner", "cleaver"],
            ),
            (dictgen::InsensitiveStr::Ascii("imapcting"), &["impacting"]),
            (
                dictgen::InsensitiveStr::Ascii("comfigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequtively"),
                &["consecutively"],
            ),
            (dictgen::InsensitiveStr::Ascii("torchligt"), &["torchlight"]),
            (dictgen::InsensitiveStr::Ascii("prfixing"), &["prefixing"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrailly"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceeses"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("regretts"), &["regrets"]),
            (
                dictgen::InsensitiveStr::Ascii("constituional"),
                &["constitutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somethigns"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("kakfa"), &["kafka"]),
            (
                dictgen::InsensitiveStr::Ascii("confgiuring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreation"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conpromising"),
                &["compromising"],
            ),
            (dictgen::InsensitiveStr::Ascii("hienous"), &["heinous"]),
            (dictgen::InsensitiveStr::Ascii("shtop"), &["stop", "shop"]),
            (dictgen::InsensitiveStr::Ascii("claen"), &["clean"]),
            (
                dictgen::InsensitiveStr::Ascii("purpetrators"),
                &["perpetrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("mozzarela"), &["mozzarella"]),
            (dictgen::InsensitiveStr::Ascii("selektors"), &["selectors"]),
            (dictgen::InsensitiveStr::Ascii("dereiving"), &["deriving"]),
            (
                dictgen::InsensitiveStr::Ascii("converstation"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realisticaly"),
                &["realistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vigilantie"), &["vigilante"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegothiations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordintae"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiative"),
                &["negotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("bilzzcon"), &["blizzcon"]),
            (
                dictgen::InsensitiveStr::Ascii("instrunction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("achievin"), &["achieving"]),
            (
                dictgen::InsensitiveStr::Ascii("consitututes"),
                &["constitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("slect"),
                &["select", "slept", "elect", "sect"],
            ),
            (dictgen::InsensitiveStr::Ascii("collaris"), &["collars"]),
            (
                dictgen::InsensitiveStr::Ascii("playthroguh"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("rotateable"), &["rotatable"]),
            (dictgen::InsensitiveStr::Ascii("shaprie"), &["sharpie"]),
            (
                dictgen::InsensitiveStr::Ascii("restartarting"),
                &["restarting"],
            ),
            (dictgen::InsensitiveStr::Ascii("albumns"), &["albums"]),
            (
                dictgen::InsensitiveStr::Ascii("corrrecting"),
                &["correcting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empirialist"),
                &["imperialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninitiaized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticaters"),
                &["authenticators", "authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenately"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generilized"),
                &["generalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmons"), &["commons"]),
            (
                dictgen::InsensitiveStr::Ascii("transperancies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("genreate"), &["generate"]),
            (dictgen::InsensitiveStr::Ascii("christams"), &["christmas"]),
            (dictgen::InsensitiveStr::Ascii("andriods"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacterers"),
                &["manufacturers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poducibility"),
                &["producibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("coommand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("iteations"), &["iterations"]),
            (dictgen::InsensitiveStr::Ascii("scrambe"), &["scramble"]),
            (dictgen::InsensitiveStr::Ascii("vacumeed"), &["vacuumed"]),
            (
                dictgen::InsensitiveStr::Ascii("confirgured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitualenv"), &["virtualenv"]),
            (dictgen::InsensitiveStr::Ascii("rturning"), &["returning"]),
            (dictgen::InsensitiveStr::Ascii("moduls"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("inudstry"), &["industry"]),
            (
                dictgen::InsensitiveStr::Ascii("compeleting"),
                &["completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabillities"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrormatting"),
                &["reformatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unambigously"),
                &["unambiguously"],
            ),
            (dictgen::InsensitiveStr::Ascii("withour"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("indavertently"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extremistes"),
                &["extremists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degoratory"),
                &["derogatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuzing"), &["confusing"]),
            (
                dictgen::InsensitiveStr::Ascii("defensese"),
                &["defenseless"],
            ),
            (dictgen::InsensitiveStr::Ascii("mehcanics"), &["mechanics"]),
            (dictgen::InsensitiveStr::Ascii("validiate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("progessing"),
                &["progressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ritoers"), &["rioters"]),
            (dictgen::InsensitiveStr::Ascii("filmmakes"), &["filmmakers"]),
            (dictgen::InsensitiveStr::Ascii("currus"), &["cirrus"]),
            (dictgen::InsensitiveStr::Ascii("langugue"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("personnalities"),
                &["personalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benechmarking"),
                &["benchmarking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insatllers"),
                &["installers"],
            ),
            (dictgen::InsensitiveStr::Ascii("journys"), &["journeys"]),
            (dictgen::InsensitiveStr::Ascii("monolight"), &["moonlight"]),
            (
                dictgen::InsensitiveStr::Ascii("cuatiously"),
                &["cautiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facter"),
                &["factor", "faster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encosing"),
                &["enclosing", "encoding"],
            ),
            (dictgen::InsensitiveStr::Ascii("ultrason"), &["ultrasound"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiater"),
                &["negotiator", "negotiated", "negotiates", "negotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiallly"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("organizacional"),
                &["organizational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singeled"),
                &["singled", "signaled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrespecful"),
                &["disrespectful"],
            ),
            (dictgen::InsensitiveStr::Ascii("upploading"), &["uploading"]),
            (
                dictgen::InsensitiveStr::Ascii("declerator"),
                &["declarator", "decelerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tormenters"),
                &["tormentors"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyebrowes"), &["eyebrows"]),
            (dictgen::InsensitiveStr::Ascii("bidnings"), &["bindings"]),
            (
                dictgen::InsensitiveStr::Ascii("becnhmarks"),
                &["benchmarks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liteautrue"),
                &["literature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developements"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serivceable"),
                &["serviceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("nationaal"), &["national"]),
            (dictgen::InsensitiveStr::Ascii("domiante"), &["dominate"]),
            (
                dictgen::InsensitiveStr::Ascii("classifiications"),
                &["classifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequsites"),
                &["prerequisites"],
            ),
            (dictgen::InsensitiveStr::Ascii("illinios"), &["illinois"]),
            (dictgen::InsensitiveStr::Ascii("flused"), &["flushed"]),
            (
                dictgen::InsensitiveStr::Ascii("deterrnines"),
                &["determines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associaties"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("substans"), &["substances"]),
            (
                dictgen::InsensitiveStr::Ascii("resolustion"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convervation"),
                &["conservation", "conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incramentally"),
                &["incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acchievement"),
                &["achievement"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdifies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("sirvaylence"),
                &["surveillance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protototype"),
                &["prototype"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizobtal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practicioners"),
                &["practitioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("solated"), &["isolated"]),
            (dictgen::InsensitiveStr::Ascii("extrates"), &["extracts"]),
            (dictgen::InsensitiveStr::Ascii("documemt"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("defautls"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("accumelates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("linerization"),
                &["linearization"],
            ),
            (dictgen::InsensitiveStr::Ascii("supprt"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("aas"), &["as", "ass"]),
            (
                dictgen::InsensitiveStr::Ascii("partitioing"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("tihnk"), &["think"]),
            (dictgen::InsensitiveStr::Ascii("exacutes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("thrustworthy"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precendance"),
                &["precedence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hullucinations"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reaonable"), &["reasonable"]),
            (dictgen::InsensitiveStr::Ascii("famly"), &["family"]),
            (dictgen::InsensitiveStr::Ascii("liitle"), &["little"]),
            (dictgen::InsensitiveStr::Ascii("dokcing"), &["docking"]),
            (dictgen::InsensitiveStr::Ascii("avaiible"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("invloved"), &["involved"]),
            (
                dictgen::InsensitiveStr::Ascii("possibillity"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("adew"), &["adieu"]),
            (dictgen::InsensitiveStr::Ascii("fieldlst"), &["fieldlist"]),
            (
                dictgen::InsensitiveStr::Ascii("documenated"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priveleges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laboratroy"),
                &["laboratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("raphsody"), &["rhapsody"]),
            (dictgen::InsensitiveStr::Ascii("respnsive"), &["responsive"]),
            (
                dictgen::InsensitiveStr::Ascii("equivlants"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breathtakng"),
                &["breathtaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antialialised"),
                &["antialiased"],
            ),
            (dictgen::InsensitiveStr::Ascii("grapichs"), &["graphics"]),
            (dictgen::InsensitiveStr::Ascii("hanshakes"), &["handshakes"]),
            (
                dictgen::InsensitiveStr::Ascii("contorlling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitiute"),
                &["substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("parmenas"), &["parmesan"]),
            (
                dictgen::InsensitiveStr::Ascii("dynmaically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("slience"), &["silence"]),
            (
                dictgen::InsensitiveStr::Ascii("proprocessed"),
                &["preprocessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("synphony"), &["symphony"]),
            (dictgen::InsensitiveStr::Ascii("emprical"), &["empirical"]),
            (dictgen::InsensitiveStr::Ascii("fracturare"), &["fracture"]),
            (dictgen::InsensitiveStr::Ascii("resivwar"), &["reservoir"]),
            (dictgen::InsensitiveStr::Ascii("unordererd"), &["unordered"]),
            (
                dictgen::InsensitiveStr::Ascii("propreitary"),
                &["proprietary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evesdropping"),
                &["eavesdropping"],
            ),
            (dictgen::InsensitiveStr::Ascii("insepects"), &["inspects"]),
            (dictgen::InsensitiveStr::Ascii("toppins"), &["toppings"]),
            (dictgen::InsensitiveStr::Ascii("vioalting"), &["violating"]),
            (dictgen::InsensitiveStr::Ascii("animatons"), &["animations"]),
            (dictgen::InsensitiveStr::Ascii("bogous"), &["bogus"]),
            (dictgen::InsensitiveStr::Ascii("cirumvent"), &["circumvent"]),
            (dictgen::InsensitiveStr::Ascii("sonebody"), &["somebody"]),
            (dictgen::InsensitiveStr::Ascii("aruments"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("unsychronise"),
                &["unsynchronise"],
            ),
            (dictgen::InsensitiveStr::Ascii("litttle"), &["little"]),
            (
                dictgen::InsensitiveStr::Ascii("precaulcate"),
                &["precalculate"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyboads"), &["keyboards"]),
            (dictgen::InsensitiveStr::Ascii("statuys"), &["status"]),
            (
                dictgen::InsensitiveStr::Ascii("queryinterace"),
                &["queryinterface"],
            ),
            (dictgen::InsensitiveStr::Ascii("takoff"), &["takeoff"]),
            (
                dictgen::InsensitiveStr::Ascii("contianing"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humanitatian"),
                &["humanitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissambles"),
                &["disassembles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anormaly"),
                &["abnormally", "anomaly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plasements"),
                &["placements"],
            ),
            (dictgen::InsensitiveStr::Ascii("fibonaacci"), &["fibonacci"]),
            (
                dictgen::InsensitiveStr::Ascii("depoys"),
                &["deploys", "depots", "decoys"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minimzation"),
                &["minimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphreal"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jewelery"),
                &["jewelry", "jewellery"],
            ),
            (dictgen::InsensitiveStr::Ascii("acivated"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("routin"),
                &["routine", "routing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cylndrical"),
                &["cylindrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("imminet"), &["imminent"]),
            (dictgen::InsensitiveStr::Ascii("conects"), &["connects"]),
            (
                dictgen::InsensitiveStr::Ascii("functiong"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genuis"),
                &["genius", "genus"],
            ),
            (dictgen::InsensitiveStr::Ascii("presnt"), &["present"]),
            (dictgen::InsensitiveStr::Ascii("bidrman"), &["birdman"]),
            (dictgen::InsensitiveStr::Ascii("editiors"), &["editors"]),
            (
                dictgen::InsensitiveStr::Ascii("reruirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attritubed"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("publishre"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("shennaigans"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("epxects"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("mergin"),
                &["merging", "margin"],
            ),
            (dictgen::InsensitiveStr::Ascii("fyou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("enpoind"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("defenite"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("disconecctions"),
                &["disconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("freeeze"), &["freeze"]),
            (
                dictgen::InsensitiveStr::Ascii("assginments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("loking"), &["looking"]),
            (
                dictgen::InsensitiveStr::Ascii("adreneline"),
                &["adrenaline"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpuut"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("disucussed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("algorithum"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("greatfully"),
                &["gratefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("somwhow"), &["somehow"]),
            (
                dictgen::InsensitiveStr::Ascii("aesthatically"),
                &["aesthetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mmanifest"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("signitory"), &["signatory"]),
            (dictgen::InsensitiveStr::Ascii("originial"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("veiwing"), &["viewing"]),
            (dictgen::InsensitiveStr::Ascii("repoen"), &["reopen"]),
            (
                dictgen::InsensitiveStr::Ascii("desagreement"),
                &["disagreement"],
            ),
            (dictgen::InsensitiveStr::Ascii("paralelly"), &["parallelly"]),
            (
                dictgen::InsensitiveStr::Ascii("siginificance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("platorms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("tyupos"), &["typos"]),
            (
                dictgen::InsensitiveStr::Ascii("particulaly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corospondant"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterprit"),
                &["misinterpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntesizes"),
                &["synthesizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomerty"), &["geometry"]),
            (
                dictgen::InsensitiveStr::Ascii("conveinences"),
                &["conveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahining"), &["chaining"]),
            (dictgen::InsensitiveStr::Ascii("mutablity"), &["mutability"]),
            (
                dictgen::InsensitiveStr::Ascii("verision"),
                &["version", "revision", "derision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deamiguate"),
                &["disambiguate"],
            ),
            (dictgen::InsensitiveStr::Ascii("stanard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("corpsers"), &["corpses"]),
            (
                dictgen::InsensitiveStr::Ascii("travercing"),
                &["traversing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caligraphy"),
                &["calligraphy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("namepsaces"),
                &["namespaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conflatin"),
                &["conflating", "conflation"],
            ),
            (dictgen::InsensitiveStr::Ascii("charasets"), &["charsets"]),
            (dictgen::InsensitiveStr::Ascii("contiuing"), &["continuing"]),
            (dictgen::InsensitiveStr::Ascii("tehnician"), &["technician"]),
            (dictgen::InsensitiveStr::Ascii("sucsede"), &["succeed"]),
            (dictgen::InsensitiveStr::Ascii("checkks"), &["checks"]),
            (dictgen::InsensitiveStr::Ascii("coneccted"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("tremenduous"),
                &["tremendous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenfies"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("displayig"), &["displaying"]),
            (
                dictgen::InsensitiveStr::Ascii("vegitables"),
                &["vegetables"],
            ),
            (dictgen::InsensitiveStr::Ascii("noifies"), &["notifies"]),
            (
                dictgen::InsensitiveStr::Ascii("elemantary"),
                &["elementary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimensonality"),
                &["dimensionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("underated"), &["underrated"]),
            (dictgen::InsensitiveStr::Ascii("estabish"), &["establish"]),
            (dictgen::InsensitiveStr::Ascii("retrievie"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("utilisin"), &["utilising"]),
            (dictgen::InsensitiveStr::Ascii("modifers"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoledged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("clenup"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("physisans"), &["physicians"]),
            (dictgen::InsensitiveStr::Ascii("objekt"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("corrseponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("noteboks"), &["notebooks"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrevation"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefeirs"), &["specifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("growteskly"),
                &["grotesquely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combanatorics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vertice"),
                &["vertex", "vertices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definietly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionabilities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("cardianls"), &["cardinals"]),
            (
                dictgen::InsensitiveStr::Ascii("alhapeticaly"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bussinesses"),
                &["businesses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genenerating"),
                &["generating"],
            ),
            (dictgen::InsensitiveStr::Ascii("coolot"), &["culotte"]),
            (
                dictgen::InsensitiveStr::Ascii("ontained"),
                &["contained", "obtained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efficeintly"),
                &["efficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photogrpaher"),
                &["photographer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongnises"),
                &["recognises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefininte"),
                &["indefinite"],
            ),
            (dictgen::InsensitiveStr::Ascii("addictes"), &["addicts"]),
            (dictgen::InsensitiveStr::Ascii("ofo"), &["of"]),
            (dictgen::InsensitiveStr::Ascii("excell"), &["excel"]),
            (dictgen::InsensitiveStr::Ascii("scroling"), &["scrolling"]),
            (
                dictgen::InsensitiveStr::Ascii("throttleing"),
                &["throttling"],
            ),
            (dictgen::InsensitiveStr::Ascii("coputed"), &["computed"]),
            (dictgen::InsensitiveStr::Ascii("anroid"), &["android"]),
            (
                dictgen::InsensitiveStr::Ascii("suspiciouns"),
                &["suspicion"],
            ),
            (dictgen::InsensitiveStr::Ascii("verndor"), &["vendor"]),
            (
                dictgen::InsensitiveStr::Ascii("resourceype"),
                &["resourcetype"],
            ),
            (dictgen::InsensitiveStr::Ascii("cursaders"), &["crusaders"]),
            (
                dictgen::InsensitiveStr::Ascii("narciscism"),
                &["narcissism"],
            ),
            (dictgen::InsensitiveStr::Ascii("flouride"), &["fluoride"]),
            (dictgen::InsensitiveStr::Ascii("driective"), &["directive"]),
            (dictgen::InsensitiveStr::Ascii("expresing"), &["expressing"]),
            (dictgen::InsensitiveStr::Ascii("searher"), &["searcher"]),
            (
                dictgen::InsensitiveStr::Ascii("subelemets"),
                &["subelements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalculaion"),
                &["recalculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanservie"), &["fanservice"]),
            (
                dictgen::InsensitiveStr::Ascii("transprting"),
                &["transporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classsifications"),
                &["classifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("distacne"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusuastic"),
                &["enthusiastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrcyptions"),
                &["encryptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenition"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbhorhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ridiculoussness"),
                &["ridiculousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("syantax"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("reallly"), &["really"]),
            (dictgen::InsensitiveStr::Ascii("concers"), &["concerns"]),
            (
                dictgen::InsensitiveStr::Ascii("passthrought"),
                &["passthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("produdes"), &["produces"]),
            (
                dictgen::InsensitiveStr::Ascii("satistified"),
                &["satisfied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ridiculousy"),
                &["ridiculously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("previewd"), &["previewed"]),
            (dictgen::InsensitiveStr::Ascii("deubbers"), &["debuggers"]),
            (dictgen::InsensitiveStr::Ascii("erronous"), &["erroneous"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhtmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferabley"),
                &["preferably"],
            ),
            (dictgen::InsensitiveStr::Ascii("lingerd"), &["lingered"]),
            (
                dictgen::InsensitiveStr::Ascii("registations"),
                &["registrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("truncted"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("betweeb"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("teleportaion"),
                &["teleportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motorcyles"),
                &["motorcycles"],
            ),
            (dictgen::InsensitiveStr::Ascii("optomize"), &["optimize"]),
            (
                dictgen::InsensitiveStr::Ascii("instutition"),
                &["institution"],
            ),
            (dictgen::InsensitiveStr::Ascii("brussells"), &["brussels"]),
            (
                dictgen::InsensitiveStr::Ascii("accodomates"),
                &["accommodates"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynamc"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("pecents"), &["percents"]),
            (dictgen::InsensitiveStr::Ascii("genreated"), &["generated"]),
            (
                dictgen::InsensitiveStr::Ascii("dictionairy"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("elimiate"), &["eliminate"]),
            (dictgen::InsensitiveStr::Ascii("contorl"), &["control"]),
            (dictgen::InsensitiveStr::Ascii("anmesty"), &["amnesty"]),
            (
                dictgen::InsensitiveStr::Ascii("condtionally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurred"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberbuilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegothiation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrybes"), &["describes"]),
            (dictgen::InsensitiveStr::Ascii("exhuast"), &["exhaust"]),
            (dictgen::InsensitiveStr::Ascii("exeeding"), &["exceeding"]),
            (dictgen::InsensitiveStr::Ascii("sticthes"), &["stitches"]),
            (dictgen::InsensitiveStr::Ascii("ediitor"), &["editor"]),
            (
                dictgen::InsensitiveStr::Ascii("reccurences"),
                &["recurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abdominable"),
                &["abdominal", "abominable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dislcaimer"),
                &["disclaimer"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntacks"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("spaning"), &["spanning"]),
            (
                dictgen::InsensitiveStr::Ascii("torrentking"),
                &["torrenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("propagete"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("thereom"), &["theorem"]),
            (dictgen::InsensitiveStr::Ascii("sepertors"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("comamndline"),
                &["commandline"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnge"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("keyords"), &["keywords"]),
            (dictgen::InsensitiveStr::Ascii("convaxly"), &["convexly"]),
            (
                dictgen::InsensitiveStr::Ascii("misogyinst"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("changce"), &["chance"]),
            (dictgen::InsensitiveStr::Ascii("publiches"), &["publishes"]),
            (
                dictgen::InsensitiveStr::Ascii("interlectuals"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confguring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("altenatives"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brianwashing"),
                &["brainwashing"],
            ),
            (dictgen::InsensitiveStr::Ascii("secretery"), &["secretary"]),
            (
                dictgen::InsensitiveStr::Ascii("notifcation"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manuall"),
                &["manually", "manual"],
            ),
            (dictgen::InsensitiveStr::Ascii("bizness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("virutalised"),
                &["virtualised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconnetion"),
                &["reconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("figurin"),
                &["figuring", "figurine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplimented"),
                &["reimplemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priotitize"),
                &["prioritize"],
            ),
            (dictgen::InsensitiveStr::Ascii("trurned"), &["turned"]),
            (
                dictgen::InsensitiveStr::Ascii("poain"),
                &["point", "pain", "plain"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiblly"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenatally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polyphonyic"),
                &["polyphonic"],
            ),
            (dictgen::InsensitiveStr::Ascii("essentias"), &["essentials"]),
            (dictgen::InsensitiveStr::Ascii("reuplad"), &["reupload"]),
            (
                dictgen::InsensitiveStr::Ascii("astronaunts"),
                &["astronaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concating"),
                &["concatenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparoching"),
                &["approaching"],
            ),
            (dictgen::InsensitiveStr::Ascii("natrual"), &["natural"]),
            (dictgen::InsensitiveStr::Ascii("scatchpad"), &["scratchpad"]),
            (
                dictgen::InsensitiveStr::Ascii("transaltion"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reserchers"),
                &["researchers"],
            ),
            (dictgen::InsensitiveStr::Ascii("corruptted"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("manipualtion"),
                &["manipulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignor"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("instantating"),
                &["instantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowiwng"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("passione"), &["passionate"]),
            (
                dictgen::InsensitiveStr::Ascii("carnomially"),
                &["ceremonially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hadlers"),
                &["handler", "handlers"],
            ),
            (dictgen::InsensitiveStr::Ascii("abcissa"), &["abscissa"]),
            (dictgen::InsensitiveStr::Ascii("literatire"), &["literate"]),
            (
                dictgen::InsensitiveStr::Ascii("resuse"),
                &["reuse", "refuse", "resume"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltrait"),
                &["infiltrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("decisiones"), &["decisions"]),
            (
                dictgen::InsensitiveStr::Ascii("algorthins"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("behvavior"), &["behavior"]),
            (dictgen::InsensitiveStr::Ascii("reapired"), &["repaired"]),
            (
                dictgen::InsensitiveStr::Ascii("noticiably"),
                &["noticeably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corinthinas"),
                &["corinthians"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucrure"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("messasging"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("prosessor"),
                &["processor", "professor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sustitutions"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolate"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("initiatiater"),
                &["initiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreagator"),
                &["aggregator"],
            ),
            (dictgen::InsensitiveStr::Ascii("amgle"), &["angle"]),
            (dictgen::InsensitiveStr::Ascii("auxillary"), &["auxiliary"]),
            (
                dictgen::InsensitiveStr::Ascii("dditions"),
                &["additions", "editions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fermentato"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vicseral"), &["visceral"]),
            (
                dictgen::InsensitiveStr::Ascii("delibrately"),
                &["deliberately"],
            ),
            (dictgen::InsensitiveStr::Ascii("goess"), &["goes"]),
            (dictgen::InsensitiveStr::Ascii("amaizing"), &["amazing"]),
            (dictgen::InsensitiveStr::Ascii("annuled"), &["annulled"]),
            (dictgen::InsensitiveStr::Ascii("onyl"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("nomals"),
                &["normals", "nomads"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecplicitly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("progammer"), &["programmer"]),
            (dictgen::InsensitiveStr::Ascii("migth"), &["might"]),
            (dictgen::InsensitiveStr::Ascii("sasuages"), &["sausages"]),
            (dictgen::InsensitiveStr::Ascii("assisants"), &["assistants"]),
            (dictgen::InsensitiveStr::Ascii("celitcs"), &["celtics"]),
            (dictgen::InsensitiveStr::Ascii("queus"), &["queues"]),
            (
                dictgen::InsensitiveStr::Ascii("monopolets"),
                &["monopolies"],
            ),
            (dictgen::InsensitiveStr::Ascii("orcehstra"), &["orchestra"]),
            (
                dictgen::InsensitiveStr::Ascii("consiquently"),
                &["consequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifyies"), &["verifies"]),
            (dictgen::InsensitiveStr::Ascii("hastables"), &["hashtables"]),
            (dictgen::InsensitiveStr::Ascii("raotating"), &["rotating"]),
            (
                dictgen::InsensitiveStr::Ascii("testiclular"),
                &["testicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polarlising"),
                &["polarising"],
            ),
            (dictgen::InsensitiveStr::Ascii("domecrats"), &["democrats"]),
            (dictgen::InsensitiveStr::Ascii("opetion"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("arral"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("oveerun"), &["overrun"]),
            (dictgen::InsensitiveStr::Ascii("oprand"), &["operand"]),
            (dictgen::InsensitiveStr::Ascii("estiamator"), &["estimator"]),
            (dictgen::InsensitiveStr::Ascii("intrisics"), &["intrinsics"]),
            (
                dictgen::InsensitiveStr::Ascii("oprimizing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasturize"), &["pasteurize"]),
            (dictgen::InsensitiveStr::Ascii("emmision"), &["emission"]),
            (
                dictgen::InsensitiveStr::Ascii("distorsion"),
                &["distortion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercetps"),
                &["intercepts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meditarrenean"),
                &["mediterranean"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelereations"),
                &["accelerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otpimization"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfmroms"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presidante"),
                &["presidents"],
            ),
            (dictgen::InsensitiveStr::Ascii("turcoise"), &["turquoise"]),
            (
                dictgen::InsensitiveStr::Ascii("initalisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("stutering"), &["stuttering"]),
            (dictgen::InsensitiveStr::Ascii("secitons"), &["sections"]),
            (dictgen::InsensitiveStr::Ascii("apeture"), &["aperture"]),
            (
                dictgen::InsensitiveStr::Ascii("timestemps"),
                &["timestamps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("returne"),
                &["return", "returned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regardsless"),
                &["regardless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throtted"),
                &["throttled", "trotted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mdifielder"),
                &["midfielder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefenreceable"),
                &["dereferenceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("nedds"), &["needs"]),
            (dictgen::InsensitiveStr::Ascii("fetishiste"), &["fetishes"]),
            (
                dictgen::InsensitiveStr::Ascii("preversing"),
                &["preserving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multlicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("mediocer"), &["mediocre"]),
            (dictgen::InsensitiveStr::Ascii("disguished"), &["disguised"]),
            (dictgen::InsensitiveStr::Ascii("wice"), &["vice"]),
            (dictgen::InsensitiveStr::Ascii("excersize"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("lesbain"), &["lesbian"]),
            (dictgen::InsensitiveStr::Ascii("erorr"), &["error"]),
            (
                dictgen::InsensitiveStr::Ascii("exexutables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparsions"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserailize"),
                &["deserialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspiciosly"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challender"),
                &["challenger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolagized"),
                &["apologized"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedon"), &["freedom"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentan"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliaction"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionarties"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("plauers"), &["players"]),
            (dictgen::InsensitiveStr::Ascii("colletion"), &["collection"]),
            (dictgen::InsensitiveStr::Ascii("euphroic"), &["euphoric"]),
            (dictgen::InsensitiveStr::Ascii("encryt"), &["encrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("pattrens"),
                &["patterns", "patrons"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopmepages"), &["homepages"]),
            (
                dictgen::InsensitiveStr::Ascii("bettery"),
                &["better", "battery"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwis"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("menual"), &["manual"]),
            (dictgen::InsensitiveStr::Ascii("semgent"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("disater"), &["disaster"]),
            (
                dictgen::InsensitiveStr::Ascii("immedaitely"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("soveriegn"), &["sovereign"]),
            (dictgen::InsensitiveStr::Ascii("bacically"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("pepole"), &["people"]),
            (
                dictgen::InsensitiveStr::Ascii("substanitally"),
                &["substantially"],
            ),
            (dictgen::InsensitiveStr::Ascii("reachers"), &["readers"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortuntely"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgarding"),
                &["discarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematicion"),
                &["mathematician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supsension"),
                &["suspension"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("principlaity"),
                &["principality"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteself"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("priorisations"),
                &["prioritisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("supplated"), &["supplanted"]),
            (dictgen::InsensitiveStr::Ascii("relativ"), &["relative"]),
            (dictgen::InsensitiveStr::Ascii("neighbar"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("brutaly"), &["brutally"]),
            (dictgen::InsensitiveStr::Ascii("jersualem"), &["jerusalem"]),
            (
                dictgen::InsensitiveStr::Ascii("transitionnal"),
                &["transitional"],
            ),
            (dictgen::InsensitiveStr::Ascii("raelly"), &["really"]),
            (dictgen::InsensitiveStr::Ascii("disapline"), &["discipline"]),
            (
                dictgen::InsensitiveStr::Ascii("insterrupts"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnearbilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("invlid"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("analyises"), &["analyses"]),
            (dictgen::InsensitiveStr::Ascii("remoives"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("figguring"), &["figuring"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappoints"),
                &["disappoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("oricle"), &["oracle"]),
            (dictgen::InsensitiveStr::Ascii("inpsired"), &["inspired"]),
            (dictgen::InsensitiveStr::Ascii("upgarte"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoties"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("fialures"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("inisialized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selctors"),
                &["selectors", "sectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("examininig"), &["examining"]),
            (
                dictgen::InsensitiveStr::Ascii("acoordingly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performnaces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("spash"), &["splash"]),
            (dictgen::InsensitiveStr::Ascii("clitors"), &["clitoris"]),
            (dictgen::InsensitiveStr::Ascii("systm"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("constaints"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presuasion"),
                &["persuasion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existental"),
                &["existential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromisomes"),
                &["chromosomes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relativitiy"),
                &["relativity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradictary"),
                &["contradictory"],
            ),
            (dictgen::InsensitiveStr::Ascii("elseswhere"), &["elsewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("virtualzation"),
                &["virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("somethimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("reconize"), &["recognize"]),
            (dictgen::InsensitiveStr::Ascii("analysit"), &["analyst"]),
            (
                dictgen::InsensitiveStr::Ascii("capabitily"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("acconding"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("attribbutes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrise"), &["arise"]),
            (
                dictgen::InsensitiveStr::Ascii("drescriptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invisibiliy"),
                &["invisibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depedencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emiss"),
                &["remiss", "amiss", "amass"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficiy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamnetally"),
                &["fundamentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sofisticated"),
                &["sophisticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interepolate"),
                &["interpolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descroptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosiacted"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroipted"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("vartex"), &["vertex"]),
            (
                dictgen::InsensitiveStr::Ascii("syclic"),
                &["cyclic", "psychic"],
            ),
            (dictgen::InsensitiveStr::Ascii("comsumes"), &["consumes"]),
            (
                dictgen::InsensitiveStr::Ascii("overridding"),
                &["overriding"],
            ),
            (dictgen::InsensitiveStr::Ascii("profirler"), &["profiler"]),
            (
                dictgen::InsensitiveStr::Ascii("jpng"),
                &["png", "jpg", "jpeg"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppporter"), &["supporter"]),
            (
                dictgen::InsensitiveStr::Ascii("addiitonal"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("archved"), &["archived"]),
            (
                dictgen::InsensitiveStr::Ascii("redifinition"),
                &["redefinition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperetions"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("enterpris"), &["enterprise"]),
            (dictgen::InsensitiveStr::Ascii("alwast"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("unavaialble"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("creatre"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("recommned"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("speficification"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminscient"),
                &["reminiscent"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifespawn"), &["lifespan"]),
            (dictgen::InsensitiveStr::Ascii("howeverm"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("reccured"),
                &["recurred", "reoccurred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meterology"),
                &["meteorology"],
            ),
            (dictgen::InsensitiveStr::Ascii("disagred"), &["disagreed"]),
            (dictgen::InsensitiveStr::Ascii("earlest"), &["earliest"]),
            (dictgen::InsensitiveStr::Ascii("genarics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("structrually"),
                &["structurally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficitied"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("companis"), &["companions"]),
            (
                dictgen::InsensitiveStr::Ascii("particularely"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitriray"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("luminence"), &["luminance"]),
            (
                dictgen::InsensitiveStr::Ascii("indistinguishible"),
                &["indistinguishable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ahould"),
                &["should", "ahold"],
            ),
            (dictgen::InsensitiveStr::Ascii("servcie"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("capabitilites"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assicietions"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumarizes"), &["summarizes"]),
            (dictgen::InsensitiveStr::Ascii("fruitin"), &["fruition"]),
            (dictgen::InsensitiveStr::Ascii("interlly"), &["internally"]),
            (
                dictgen::InsensitiveStr::Ascii("repsonsible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emminent"),
                &["eminent", "imminent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associcative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("paitiently"), &["patiently"]),
            (dictgen::InsensitiveStr::Ascii("descreased"), &["decreased"]),
            (
                dictgen::InsensitiveStr::Ascii("precident"),
                &["precedent", "president"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ocassioned"),
                &["occasioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automatcially"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimants"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("idaes"), &["ideas"]),
            (dictgen::InsensitiveStr::Ascii("mantatory"), &["mandatory"]),
            (
                dictgen::InsensitiveStr::Ascii("improviserad"),
                &["improvised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemmorhage"),
                &["hemorrhage"],
            ),
            (dictgen::InsensitiveStr::Ascii("entitlied"), &["entitled"]),
            (dictgen::InsensitiveStr::Ascii("inpacting"), &["impacting"]),
            (
                dictgen::InsensitiveStr::Ascii("intoloerance"),
                &["intolerance"],
            ),
            (dictgen::InsensitiveStr::Ascii("norhtwest"), &["northwest"]),
            (dictgen::InsensitiveStr::Ascii("ttrying"), &["trying"]),
            (dictgen::InsensitiveStr::Ascii("sponser"), &["sponsor"]),
            (dictgen::InsensitiveStr::Ascii("befores"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("assertted"), &["asserted"]),
            (
                dictgen::InsensitiveStr::Ascii("thermometre"),
                &["thermometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("scalale"), &["scalable"]),
            (dictgen::InsensitiveStr::Ascii("scecnarios"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("algorothmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("fourten"), &["fourteen"]),
            (
                dictgen::InsensitiveStr::Ascii("compelxities"),
                &["complexities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterract"),
                &["counterpart"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apostrophies"),
                &["apostrophe", "apostrophes"],
            ),
            (dictgen::InsensitiveStr::Ascii("messag"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("diablers"), &["disablers"]),
            (dictgen::InsensitiveStr::Ascii("exempels"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("whitsles"), &["whistles"]),
            (
                dictgen::InsensitiveStr::Ascii("mamagement"),
                &["management"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconeccted"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("recrutied"), &["recruited"]),
            (dictgen::InsensitiveStr::Ascii("ownded"), &["owned"]),
            (
                dictgen::InsensitiveStr::Ascii("anways"),
                &["anyway", "anyways"],
            ),
            (dictgen::InsensitiveStr::Ascii("occassions"), &["occasions"]),
            (dictgen::InsensitiveStr::Ascii("experim"), &["experiment"]),
            (
                dictgen::InsensitiveStr::Ascii("referrences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("shirnk"), &["shrink"]),
            (
                dictgen::InsensitiveStr::Ascii("equivalants"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailsiations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrubution"),
                &["retribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceoted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("fingerpoints"),
                &["fingerprints"],
            ),
            (dictgen::InsensitiveStr::Ascii("nunber"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunalte"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("skagerak"), &["skagerrak"]),
            (
                dictgen::InsensitiveStr::Ascii("preaprooved"),
                &["preapproved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disingeneous"),
                &["disingenuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("approvin"), &["approving"]),
            (
                dictgen::InsensitiveStr::Ascii("preseverance"),
                &["perseverance"],
            ),
            (dictgen::InsensitiveStr::Ascii("bleuberry"), &["blueberry"]),
            (dictgen::InsensitiveStr::Ascii("conqueor"), &["conquer"]),
            (dictgen::InsensitiveStr::Ascii("packaage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("reopsitories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attackerasu"),
                &["attackers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrtiption"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camparisons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrsting"),
                &["contrasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unswithced"),
                &["unswitched"],
            ),
            (dictgen::InsensitiveStr::Ascii("reander"), &["render"]),
            (
                dictgen::InsensitiveStr::Ascii("transscription"),
                &["transcription"],
            ),
            (dictgen::InsensitiveStr::Ascii("furthher"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("excecuts"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("compontents"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("attepts"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("correnspond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("margerat"), &["margaret"]),
            (dictgen::InsensitiveStr::Ascii("dispoase"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("shaodw"), &["shadow"]),
            (
                dictgen::InsensitiveStr::Ascii("tranformers"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaplaination"),
                &["explanation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comformed"),
                &["conformed", "comforted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoitation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stanardise"),
                &["standardise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conspericies"),
                &["conspiracies"],
            ),
            (dictgen::InsensitiveStr::Ascii("fabricus"), &["fabrics"]),
            (dictgen::InsensitiveStr::Ascii("octopuns"), &["octopus"]),
            (dictgen::InsensitiveStr::Ascii("accociate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("begis"), &["begins"]),
            (dictgen::InsensitiveStr::Ascii("tabualted"), &["tabulated"]),
            (dictgen::InsensitiveStr::Ascii("relaly"), &["really"]),
            (dictgen::InsensitiveStr::Ascii("gloabally"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("valnecia"), &["valencia"]),
            (
                dictgen::InsensitiveStr::Ascii("philiphines"),
                &["philippines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthognally"),
                &["orthogonally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supllementary"),
                &["supplementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroipt"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("beacuase"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotions"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cerification"),
                &["verification", "certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("unverifed"), &["unverified"]),
            (dictgen::InsensitiveStr::Ascii("paramerter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("cretator"), &["creator"]),
            (dictgen::InsensitiveStr::Ascii("utillizing"), &["utilizing"]),
            (dictgen::InsensitiveStr::Ascii("thursdsy"), &["thursdays"]),
            (dictgen::InsensitiveStr::Ascii("extremeley"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("explosie"), &["explosive"]),
            (dictgen::InsensitiveStr::Ascii("optomise"), &["optimise"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolerabe"), &["tolerable"]),
            (
                dictgen::InsensitiveStr::Ascii("deseralize"),
                &["deserialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinatro"),
                &["coordinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveince"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("langugaes"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("validaterelase"),
                &["validaterelease"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transacions"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("saveguarded"),
                &["safeguarded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investiage"),
                &["investigate"],
            ),
            (dictgen::InsensitiveStr::Ascii("concent"), &["consent"]),
            (
                dictgen::InsensitiveStr::Ascii("signletons"),
                &["singletons"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprates"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("condiitons"),
                &["conditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("elegently"), &["elegantly"]),
            (
                dictgen::InsensitiveStr::Ascii("curresponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("scondly"), &["secondly"]),
            (
                dictgen::InsensitiveStr::Ascii("expermimented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("polisse"), &["polishes"]),
            (dictgen::InsensitiveStr::Ascii("optimizd"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiifation"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundig"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("mayority"), &["majority"]),
            (
                dictgen::InsensitiveStr::Ascii("constistent"),
                &["consistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vitualisation"),
                &["virtualisation", "visualisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nanoseonds"),
                &["nanoseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("baiscs"), &["basics"]),
            (
                dictgen::InsensitiveStr::Ascii("allocattion"),
                &["allocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counteredit"),
                &["counterfeit"],
            ),
            (dictgen::InsensitiveStr::Ascii("desltops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("asymetic"), &["asymmetric"]),
            (dictgen::InsensitiveStr::Ascii("framwd"), &["framed"]),
            (
                dictgen::InsensitiveStr::Ascii("proceded"),
                &["proceeded", "preceded"],
            ),
            (dictgen::InsensitiveStr::Ascii("fulll"), &["full"]),
            (dictgen::InsensitiveStr::Ascii("terninal"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("revrses"), &["reverses"]),
            (dictgen::InsensitiveStr::Ascii("vaieties"), &["varieties"]),
            (dictgen::InsensitiveStr::Ascii("ignroable"), &["ignorable"]),
            (
                dictgen::InsensitiveStr::Ascii("mimimising"),
                &["minimising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compuler"),
                &["compiler", "computer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compansating"),
                &["compensating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("litthe"),
                &["little", "lithe"],
            ),
            (dictgen::InsensitiveStr::Ascii("regsitry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("infarred"), &["infrared"]),
            (dictgen::InsensitiveStr::Ascii("arrised"), &["arose"]),
            (dictgen::InsensitiveStr::Ascii("tungs"), &["tongues"]),
            (dictgen::InsensitiveStr::Ascii("attemtp"), &["attempt"]),
            (dictgen::InsensitiveStr::Ascii("burriots"), &["burritos"]),
            (dictgen::InsensitiveStr::Ascii("nomalise"), &["normalise"]),
            (dictgen::InsensitiveStr::Ascii("quention"), &["quentin"]),
            (dictgen::InsensitiveStr::Ascii("upate"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("imorted"), &["imported"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegociations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("immoblie"), &["immobile"]),
            (dictgen::InsensitiveStr::Ascii("awfullly"), &["awfully"]),
            (dictgen::InsensitiveStr::Ascii("referesh"), &["refresh"]),
            (
                dictgen::InsensitiveStr::Ascii("rehabilitaton"),
                &["rehabilitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccuring"),
                &["recurring", "reoccurring"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmisions"), &["emissions"]),
            (dictgen::InsensitiveStr::Ascii("bluestooth"), &["bluetooth"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmerically"),
                &["asymmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleve"), &["believe"]),
            (
                dictgen::InsensitiveStr::Ascii("expeptedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstnces"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aautomatically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("threeshold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("loctes"), &["locates"]),
            (
                dictgen::InsensitiveStr::Ascii("metabilism"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("atatched"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("propbabilities"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boungaries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("prejected"), &["projected"]),
            (dictgen::InsensitiveStr::Ascii("stucked"), &["stuck"]),
            (
                dictgen::InsensitiveStr::Ascii("envoking"),
                &["invoking", "evoking"],
            ),
            (dictgen::InsensitiveStr::Ascii("disalbed"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("genarating"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administartion"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("anway"), &["anyway"]),
            (dictgen::InsensitiveStr::Ascii("upadate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiamtion"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sconds"), &["seconds"]),
            (
                dictgen::InsensitiveStr::Ascii("repositries"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massechusetts"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("wealtheir"), &["wealthier"]),
            (
                dictgen::InsensitiveStr::Ascii("expexctedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polititians"),
                &["politicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developmets"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("regalars"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("atend"), &["attend"]),
            (
                dictgen::InsensitiveStr::Ascii("surprisinly"),
                &["surprisingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("northeat"), &["northeast"]),
            (
                dictgen::InsensitiveStr::Ascii("formulatoin"),
                &["formulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relitavely"),
                &["relatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalese"), &["coalesce"]),
            (dictgen::InsensitiveStr::Ascii("uplaod"), &["upload"]),
            (dictgen::InsensitiveStr::Ascii("sanwiches"), &["sandwiches"]),
            (dictgen::InsensitiveStr::Ascii("confirure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("backkgrounds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("virbate"), &["vibrate"]),
            (
                dictgen::InsensitiveStr::Ascii("synchronizaton"),
                &["synchronization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stratagies"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specializaiton"),
                &["specialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrilegeous"),
                &["sacrilegious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temproily"),
                &["temporarily", "temporally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recrational"),
                &["recreational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perspektive"),
                &["perspective"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememer"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("cominator"), &["combinator"]),
            (dictgen::InsensitiveStr::Ascii("begiinners"), &["beginners"]),
            (
                dictgen::InsensitiveStr::Ascii("correlelations"),
                &["correlations"],
            ),
            (dictgen::InsensitiveStr::Ascii("absymal"), &["abysmal"]),
            (
                dictgen::InsensitiveStr::Ascii("reponsible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissasembler"),
                &["disassembler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumating"),
                &["enumerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("remots"), &["remotes"]),
            (
                dictgen::InsensitiveStr::Ascii("disproprotionately"),
                &["disproportionately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autenticators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appologizing"),
                &["apologizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checlpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrsholds"), &["thresholds"]),
            (dictgen::InsensitiveStr::Ascii("fcound"), &["found"]),
            (
                dictgen::InsensitiveStr::Ascii("condiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("problmes"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("deminsion"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("trunaction"),
                &["truncation"],
            ),
            (dictgen::InsensitiveStr::Ascii("methjods"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("transiver"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("mkaes"), &["makes"]),
            (dictgen::InsensitiveStr::Ascii("catgorize"), &["categorize"]),
            (dictgen::InsensitiveStr::Ascii("secondaray"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("reoganizations"),
                &["reorganizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumelated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ameythsts"), &["amethysts"]),
            (dictgen::InsensitiveStr::Ascii("prsent"), &["present"]),
            (dictgen::InsensitiveStr::Ascii("percision"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("regist"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("apprpriately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("balencer"), &["balancer"]),
            (dictgen::InsensitiveStr::Ascii("additinal"), &["additional"]),
            (
                dictgen::InsensitiveStr::Ascii("tryannical"),
                &["tyrannical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutliplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("postulat"), &["postulate"]),
            (
                dictgen::InsensitiveStr::Ascii("conpsiracy"),
                &["conspiracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("fouind"), &["found"]),
            (dictgen::InsensitiveStr::Ascii("mediocry"), &["mediocrity"]),
            (dictgen::InsensitiveStr::Ascii("stuctural"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("exends"),
                &["extends", "expends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordiantor"),
                &["coordinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("equiped"), &["equipped"]),
            (
                dictgen::InsensitiveStr::Ascii("recreacional"),
                &["recreational"],
            ),
            (dictgen::InsensitiveStr::Ascii("probablies"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("syntaxt"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("failue"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("warrandyte"), &["warranty"]),
            (
                dictgen::InsensitiveStr::Ascii("serialialized"),
                &["serialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("singnaling"), &["signaling"]),
            (
                dictgen::InsensitiveStr::Ascii("provoactive"),
                &["provocative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("govermenet"),
                &["government"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecologia"), &["ecological"]),
            (
                dictgen::InsensitiveStr::Ascii("partiulars"),
                &["particulars"],
            ),
            (dictgen::InsensitiveStr::Ascii("suplying"), &["supplying"]),
            (
                dictgen::InsensitiveStr::Ascii("negoziators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("committtee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("guas"), &["gauss"]),
            (
                dictgen::InsensitiveStr::Ascii("establising"),
                &["establishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporrarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("cmmenting"), &["commenting"]),
            (dictgen::InsensitiveStr::Ascii("arpatheid"), &["apartheid"]),
            (dictgen::InsensitiveStr::Ascii("loacalize"), &["localize"]),
            (dictgen::InsensitiveStr::Ascii("contrieves"), &["contrives"]),
            (
                dictgen::InsensitiveStr::Ascii("compesition"),
                &["composition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guestimate"),
                &["guesstimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seplicuraly"),
                &["sepulchrally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resloutions"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegosiating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vechiles"), &["vehicles"]),
            (dictgen::InsensitiveStr::Ascii("comediants"), &["comedians"]),
            (
                dictgen::InsensitiveStr::Ascii("sociapathic"),
                &["sociopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("chceklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("specifieced"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("dominent"), &["dominant"]),
            (
                dictgen::InsensitiveStr::Ascii("propection"),
                &["projection", "prospection", "protection"],
            ),
            (dictgen::InsensitiveStr::Ascii("terrabyte"), &["terabyte"]),
            (
                dictgen::InsensitiveStr::Ascii("illumunati"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apresheation"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("asembly"), &["assembly"]),
            (
                dictgen::InsensitiveStr::Ascii("unbeliveable"),
                &["unbelievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficillier"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("overreide"), &["override"]),
            (
                dictgen::InsensitiveStr::Ascii("inconprehensible"),
                &["incomprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivilantly"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liberterianism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjecent"), &["adjacent"]),
            (dictgen::InsensitiveStr::Ascii("suspesion"), &["suspension"]),
            (
                dictgen::InsensitiveStr::Ascii("diects"),
                &["directs", "dissects"],
            ),
            (dictgen::InsensitiveStr::Ascii("concole"), &["console"]),
            (dictgen::InsensitiveStr::Ascii("ambien"), &["ambient"]),
            (
                dictgen::InsensitiveStr::Ascii("customizabe"),
                &["customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beneficary"),
                &["beneficiary"],
            ),
            (dictgen::InsensitiveStr::Ascii("reigonal"), &["regional"]),
            (
                dictgen::InsensitiveStr::Ascii("retun"),
                &["return", "rerun", "retune"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementationer"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investogators"),
                &["investigators"],
            ),
            (dictgen::InsensitiveStr::Ascii("herslef"), &["herself"]),
            (dictgen::InsensitiveStr::Ascii("evetually"), &["eventually"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalismos"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vunerability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("muiltipart"), &["multipart"]),
            (
                dictgen::InsensitiveStr::Ascii("multipllication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("settng"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("messagses"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("horribley"), &["horribly"]),
            (
                dictgen::InsensitiveStr::Ascii("algorthmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressumes"),
                &["presumes", "pressures"],
            ),
            (dictgen::InsensitiveStr::Ascii("consquent"), &["consequent"]),
            (
                dictgen::InsensitiveStr::Ascii("senging"),
                &["sending", "singing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configueration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("boungind"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("allone"), &["alone"]),
            (
                dictgen::InsensitiveStr::Ascii("psydonymous"),
                &["pseudonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parellelizes"),
                &["parallelizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("organaize"), &["organize"]),
            (dictgen::InsensitiveStr::Ascii("activted"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("differentate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("nutral"), &["neutral"]),
            (
                dictgen::InsensitiveStr::Ascii("coutnerparts"),
                &["counterparts"],
            ),
            (dictgen::InsensitiveStr::Ascii("moduels"), &["modules"]),
            (
                dictgen::InsensitiveStr::Ascii("unsettleing"),
                &["unsettling"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreed"), &["agreed"]),
            (
                dictgen::InsensitiveStr::Ascii("clutser"),
                &["cluster", "clutter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delaerships"),
                &["dealerships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supressors"),
                &["suppressors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintaning"),
                &["maintaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepdencente"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comporison"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("crashess"), &["crashes"]),
            (dictgen::InsensitiveStr::Ascii("oveflow"), &["overflow"]),
            (dictgen::InsensitiveStr::Ascii("confitm"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("conections"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("consisent"), &["consistent"]),
            (
                dictgen::InsensitiveStr::Ascii("communties"),
                &["communities"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciontrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("progresison"),
                &["progressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("exersize"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("connnect"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("discovring"),
                &["discovering"],
            ),
            (dictgen::InsensitiveStr::Ascii("implicatie"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("occuppants"), &["occupants"]),
            (dictgen::InsensitiveStr::Ascii("sittin"), &["sitting"]),
            (
                dictgen::InsensitiveStr::Ascii("speculatin"),
                &["speculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranlsate"), &["translate"]),
            (dictgen::InsensitiveStr::Ascii("diangose"), &["diagnose"]),
            (dictgen::InsensitiveStr::Ascii("conservice"), &["conserve"]),
            (dictgen::InsensitiveStr::Ascii("grpah"), &["graph"]),
            (
                dictgen::InsensitiveStr::Ascii("americants"),
                &["americas", "americans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endocement"),
                &["endorsement", "enforcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configursable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("greate"),
                &["greater", "create", "grate", "great"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparrison"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("ballance"), &["balance"]),
            (
                dictgen::InsensitiveStr::Ascii("chromasomes"),
                &["chromosomes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implenmentations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hipoteses"), &["hypotheses"]),
            (
                dictgen::InsensitiveStr::Ascii("everytihng"),
                &["everything"],
            ),
            (dictgen::InsensitiveStr::Ascii("browswe"), &["browse"]),
            (dictgen::InsensitiveStr::Ascii("seeverity"), &["severity"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegding"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contstructor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructrs"),
                &["constructors", "constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("jeresys"), &["jerseys"]),
            (dictgen::InsensitiveStr::Ascii("dsytopian"), &["dystopian"]),
            (dictgen::InsensitiveStr::Ascii("tabluar"), &["tabular"]),
            (
                dictgen::InsensitiveStr::Ascii("schedulling"),
                &["scheduling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initaition"),
                &["initiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semaphones"),
                &["semaphores"],
            ),
            (dictgen::InsensitiveStr::Ascii("blueberrry"), &["blueberry"]),
            (dictgen::InsensitiveStr::Ascii("vampirs"), &["vampires"]),
            (
                dictgen::InsensitiveStr::Ascii("capbaility"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("atmoshere"), &["atmosphere"]),
            (dictgen::InsensitiveStr::Ascii("exchng"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("maneuveurs"), &["maneuver"]),
            (dictgen::InsensitiveStr::Ascii("childisch"), &["childish"]),
            (dictgen::InsensitiveStr::Ascii("esimation"), &["estimation"]),
            (dictgen::InsensitiveStr::Ascii("primaires"), &["primaries"]),
            (
                dictgen::InsensitiveStr::Ascii("apprecaition"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("huanting"), &["haunting"]),
            (dictgen::InsensitiveStr::Ascii("prolbem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("blegium"), &["belgium"]),
            (dictgen::InsensitiveStr::Ascii("soemone"), &["someone"]),
            (
                dictgen::InsensitiveStr::Ascii("assimilatin"),
                &["assimilating", "assimilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continential"),
                &["continental"],
            ),
            (dictgen::InsensitiveStr::Ascii("chocolste"), &["chocolates"]),
            (dictgen::InsensitiveStr::Ascii("diosese"), &["diocese"]),
            (
                dictgen::InsensitiveStr::Ascii("repetitivo"),
                &["repetition"],
            ),
            (dictgen::InsensitiveStr::Ascii("triniy"), &["trinity"]),
            (
                dictgen::InsensitiveStr::Ascii("critizisms"),
                &["criticisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impementations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misalignement"),
                &["misalignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("walkalbe"), &["walkable"]),
            (dictgen::InsensitiveStr::Ascii("griaffe"), &["giraffe"]),
            (dictgen::InsensitiveStr::Ascii("uplads"), &["uploads"]),
            (dictgen::InsensitiveStr::Ascii("elephantus"), &["elephants"]),
            (
                dictgen::InsensitiveStr::Ascii("flametrhower"),
                &["flamethrower"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geometrc"),
                &["geometric", "geometry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitailize"),
                &["reinitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consolidad"),
                &["consolidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiating"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("pojected"), &["projected"]),
            (dictgen::InsensitiveStr::Ascii("disapled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("bindng"), &["binding"]),
            (dictgen::InsensitiveStr::Ascii("comapred"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("internelized"),
                &["internalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("recommeds"), &["recommends"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimentelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resoluition"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignemnts"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("firther"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("bufffered"), &["buffered"]),
            (dictgen::InsensitiveStr::Ascii("travesed"), &["traversed"]),
            (
                dictgen::InsensitiveStr::Ascii("ministerens"),
                &["ministers"],
            ),
            (dictgen::InsensitiveStr::Ascii("similates"), &["simulates"]),
            (dictgen::InsensitiveStr::Ascii("simplist"), &["simplest"]),
            (
                dictgen::InsensitiveStr::Ascii("necessarly"),
                &["necessary", "necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonation"),
                &["denomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnibg"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("minimumn"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("authencication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("requsites"), &["requisites"]),
            (dictgen::InsensitiveStr::Ascii("valeu"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("figthing"), &["fighting"]),
            (dictgen::InsensitiveStr::Ascii("cotact"), &["contact"]),
            (
                dictgen::InsensitiveStr::Ascii("parititioned"),
                &["partitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consummatin"),
                &["consummating", "consummation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cntain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("serialsing"),
                &["serialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("checs"), &["checks"]),
            (dictgen::InsensitiveStr::Ascii("muste"), &["must"]),
            (dictgen::InsensitiveStr::Ascii("hink"), &["think"]),
            (dictgen::InsensitiveStr::Ascii("valdated"), &["validated"]),
            (dictgen::InsensitiveStr::Ascii("alllocator"), &["allocator"]),
            (
                dictgen::InsensitiveStr::Ascii("antiobitics"),
                &["antibiotic", "antibiotics"],
            ),
            (dictgen::InsensitiveStr::Ascii("argentinia"), &["argentina"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("specficy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("coalescsing"),
                &["coalescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challengue"),
                &["challenger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endoces"),
                &["endorses", "enforces", "encodes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatoinal"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("pathneame"), &["pathname"]),
            (dictgen::InsensitiveStr::Ascii("delaloc"), &["delalloc"]),
            (dictgen::InsensitiveStr::Ascii("iternal"), &["internal"]),
            (
                dictgen::InsensitiveStr::Ascii("transofrmations"),
                &["transformations"],
            ),
            (dictgen::InsensitiveStr::Ascii("embarass"), &["embarrass"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaizing"),
                &["generalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("libraie"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("messgaer"), &["messenger"]),
            (
                dictgen::InsensitiveStr::Ascii("recomments"),
                &["recommends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplaer"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("exapansion"), &["expansion"]),
            (
                dictgen::InsensitiveStr::Ascii("desserialized"),
                &["deserialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refactrive"),
                &["refractive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succceeding"),
                &["succeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisan"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("abpout"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("unerlying"), &["underlying"]),
            (dictgen::InsensitiveStr::Ascii("spagheti"), &["spaghetti"]),
            (dictgen::InsensitiveStr::Ascii("alchemsit"), &["alchemist"]),
            (
                dictgen::InsensitiveStr::Ascii("convertirte"),
                &["converter"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcer"), &["producer"]),
            (
                dictgen::InsensitiveStr::Ascii("resourcsed"),
                &["resourced", "resource"],
            ),
            (dictgen::InsensitiveStr::Ascii("affulent"), &["affluent"]),
            (dictgen::InsensitiveStr::Ascii("vaidating"), &["validating"]),
            (
                dictgen::InsensitiveStr::Ascii("misdemeaner"),
                &["misdemeanor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soemthings"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyopcrite"), &["hypocrite"]),
            (dictgen::InsensitiveStr::Ascii("charector"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("tempraarly"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("uisng"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("disussed"),
                &["discussed", "disused"],
            ),
            (dictgen::InsensitiveStr::Ascii("stap"), &["step", "stop"]),
            (dictgen::InsensitiveStr::Ascii("bounnded"), &["bounded"]),
            (dictgen::InsensitiveStr::Ascii("mopre"), &["more"]),
            (dictgen::InsensitiveStr::Ascii("macinery"), &["machinery"]),
            (dictgen::InsensitiveStr::Ascii("discssion"), &["discussion"]),
            (dictgen::InsensitiveStr::Ascii("thyorid"), &["thyroid"]),
            (
                dictgen::InsensitiveStr::Ascii("emphazised"),
                &["emphasised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaidations"),
                &["validations"],
            ),
            (dictgen::InsensitiveStr::Ascii("asigns"), &["assigns"]),
            (dictgen::InsensitiveStr::Ascii("regenade"), &["renegade"]),
            (
                dictgen::InsensitiveStr::Ascii("appliction"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylced"), &["cycled"]),
            (
                dictgen::InsensitiveStr::Ascii("asbtractors"),
                &["abstractors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchrnization"),
                &["synchronization"],
            ),
            (dictgen::InsensitiveStr::Ascii("improbabe"), &["improbable"]),
            (
                dictgen::InsensitiveStr::Ascii("commond"),
                &["command", "common", "commend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procdurally"),
                &["procedurally"],
            ),
            (dictgen::InsensitiveStr::Ascii("gauntlent"), &["gauntlet"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappeard"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("heroices"), &["heroics"]),
            (
                dictgen::InsensitiveStr::Ascii("experinementation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compinsation"),
                &["compensation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiation"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("acessible"), &["accessible"]),
            (dictgen::InsensitiveStr::Ascii("belssings"), &["blessings"]),
            (
                dictgen::InsensitiveStr::Ascii("reteroactively"),
                &["retroactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprietry"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmplars"), &["exemplars"]),
            (
                dictgen::InsensitiveStr::Ascii("proprogators"),
                &["propagators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coverges"),
                &["coverages", "converges"],
            ),
            (dictgen::InsensitiveStr::Ascii("ulterioara"), &["ulterior"]),
            (dictgen::InsensitiveStr::Ascii("bastract"), &["abstract"]),
            (
                dictgen::InsensitiveStr::Ascii("procceding"),
                &["proceeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliws"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("handicaped"),
                &["handicapped"],
            ),
            (dictgen::InsensitiveStr::Ascii("froma"), &["from"]),
            (
                dictgen::InsensitiveStr::Ascii("cetrifications"),
                &["certifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dcumenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undocomented"),
                &["undocumented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copyrigths"),
                &["copyrights"],
            ),
            (dictgen::InsensitiveStr::Ascii("repecting"), &["respecting"]),
            (
                dictgen::InsensitiveStr::Ascii("satisfactority"),
                &["satisfactorily"],
            ),
            (dictgen::InsensitiveStr::Ascii("langeuage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("suggestings"),
                &["suggesting", "suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmimates"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overprised"),
                &["overpriced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delievering"),
                &["delivering"],
            ),
            (dictgen::InsensitiveStr::Ascii("micrphone"), &["microphone"]),
            (
                dictgen::InsensitiveStr::Ascii("dependcies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("implicati"), &["implicit"]),
            (
                dictgen::InsensitiveStr::Ascii("accommadations"),
                &["accommodations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blackhakws"),
                &["blackhawks"],
            ),
            (dictgen::InsensitiveStr::Ascii("arund"), &["around"]),
            (
                dictgen::InsensitiveStr::Ascii("substitutue"),
                &["substitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacume"), &["vacuum"]),
            (dictgen::InsensitiveStr::Ascii("gorgoeus"), &["gorgeous"]),
            (dictgen::InsensitiveStr::Ascii("reffered"), &["referred"]),
            (dictgen::InsensitiveStr::Ascii("imporoting"), &["importing"]),
            (
                dictgen::InsensitiveStr::Ascii("checkts"),
                &["checked", "checks"],
            ),
            (dictgen::InsensitiveStr::Ascii("futureus"), &["futures"]),
            (
                dictgen::InsensitiveStr::Ascii("confidantally"),
                &["confidentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("powderd"), &["powdered"]),
            (dictgen::InsensitiveStr::Ascii("wathdog"), &["watchdog"]),
            (dictgen::InsensitiveStr::Ascii("demamd"), &["demand"]),
            (
                dictgen::InsensitiveStr::Ascii("indext"),
                &["index", "indent"],
            ),
            (dictgen::InsensitiveStr::Ascii("kunckles"), &["knuckles"]),
            (
                dictgen::InsensitiveStr::Ascii("initalialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commishioned"),
                &["commissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responisbility"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juggernath"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differntiating"),
                &["differentiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("pronoune"), &["pronounced"]),
            (
                dictgen::InsensitiveStr::Ascii("terminacion"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substituing"),
                &["substituting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pascheurisation"),
                &["pasteurisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("probblezs"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("spectaculary"),
                &["spectacularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalessed"), &["coalesced"]),
            (dictgen::InsensitiveStr::Ascii("socekt"), &["socket"]),
            (
                dictgen::InsensitiveStr::Ascii("repeatations"),
                &["repetitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprehinsible"),
                &["reprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("delibery"), &["delivery"]),
            (dictgen::InsensitiveStr::Ascii("geeral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("tirbunal"), &["tribunal"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacters"),
                &["manufactures", "manufacturers"],
            ),
            (dictgen::InsensitiveStr::Ascii("calncelled"), &["cancelled"]),
            (dictgen::InsensitiveStr::Ascii("convets"), &["converts"]),
            (
                dictgen::InsensitiveStr::Ascii("overhooked"),
                &["overlooked"],
            ),
            (dictgen::InsensitiveStr::Ascii("onatrio"), &["ontario"]),
            (dictgen::InsensitiveStr::Ascii("seconadry"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("operaror"), &["operator"]),
            (
                dictgen::InsensitiveStr::Ascii("inevitablely"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("personnal"), &["personal"]),
            (dictgen::InsensitiveStr::Ascii("dedidated"), &["dedicated"]),
            (
                dictgen::InsensitiveStr::Ascii("inconveniantly"),
                &["inconveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("precisou"), &["precious"]),
            (
                dictgen::InsensitiveStr::Ascii("paychologically"),
                &["psychologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("consitant"), &["consistent"]),
            (dictgen::InsensitiveStr::Ascii("prvode"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("aleinates"), &["alienates"]),
            (
                dictgen::InsensitiveStr::Ascii("vulerabuilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intergerated"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("clenaser"), &["cleanser"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalzed"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requriment"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("holded"), &["held"]),
            (
                dictgen::InsensitiveStr::Ascii("presumebly"),
                &["presumably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommments"),
                &["uncomments"],
            ),
            (dictgen::InsensitiveStr::Ascii("orthewr"), &["other"]),
            (
                dictgen::InsensitiveStr::Ascii("transaltions"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebellios"), &["rebellious"]),
            (
                dictgen::InsensitiveStr::Ascii("distengish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryed"), &["tried"]),
            (dictgen::InsensitiveStr::Ascii("noght"), &["night"]),
            (
                dictgen::InsensitiveStr::Ascii("speciefication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approimating"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homeopothy"),
                &["homeopathy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconecting"),
                &["reconnecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surgestion"),
                &["suggestion"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfires"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("dentifying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("larg"), &["large"]),
            (dictgen::InsensitiveStr::Ascii("golbal"), &["global"]),
            (dictgen::InsensitiveStr::Ascii("fidn"), &["find"]),
            (
                dictgen::InsensitiveStr::Ascii("deines"),
                &["defined", "defines", "denies"],
            ),
            (dictgen::InsensitiveStr::Ascii("dogmatisch"), &["dogmatic"]),
            (dictgen::InsensitiveStr::Ascii("surrround"), &["surround"]),
            (dictgen::InsensitiveStr::Ascii("auxilary"), &["auxiliary"]),
            (
                dictgen::InsensitiveStr::Ascii("internationaal"),
                &["international"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedophilac"),
                &["pedophilia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memebership"),
                &["membership"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuigre"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("innocentes"), &["innocents"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilites"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inclanation"),
                &["inclination"],
            ),
            (dictgen::InsensitiveStr::Ascii("beecause"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillations"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuncating"), &["truncating"]),
            (dictgen::InsensitiveStr::Ascii("exliled"), &["exiled"]),
            (
                dictgen::InsensitiveStr::Ascii("docuemented"),
                &["documented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addressibility"),
                &["addressability"],
            ),
            (dictgen::InsensitiveStr::Ascii("tusday"), &["tuesday"]),
            (dictgen::InsensitiveStr::Ascii("currrently"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("grassrooots"),
                &["grassroots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fomratting"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("schoodle"), &["schooled"]),
            (dictgen::InsensitiveStr::Ascii("useer"), &["user"]),
            (dictgen::InsensitiveStr::Ascii("alphabeat"), &["alphabet"]),
            (dictgen::InsensitiveStr::Ascii("aovidably"), &["avoidably"]),
            (dictgen::InsensitiveStr::Ascii("desroyers"), &["destroyers"]),
            (dictgen::InsensitiveStr::Ascii("plaftorm"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("segmenations"),
                &["segmentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contorling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vetinarians"),
                &["veterinarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socialistisk"),
                &["socialists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmaximice"),
                &["unmaximize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paret"),
                &["parent", "parrot"],
            ),
            (dictgen::InsensitiveStr::Ascii("morphinate"), &["morphine"]),
            (dictgen::InsensitiveStr::Ascii("internels"), &["internals"]),
            (dictgen::InsensitiveStr::Ascii("replyign"), &["replying"]),
            (dictgen::InsensitiveStr::Ascii("mercanery"), &["mercenary"]),
            (dictgen::InsensitiveStr::Ascii("impresive"), &["impressive"]),
            (
                dictgen::InsensitiveStr::Ascii("defenitions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrruption"),
                &["corruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("reprorted"), &["reported"]),
            (
                dictgen::InsensitiveStr::Ascii("intervier"),
                &["interviewer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreataon"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("libray"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("lineseaches"),
                &["linesearches"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoprted"), &["reported"]),
            (dictgen::InsensitiveStr::Ascii("renditioon"), &["rendition"]),
            (dictgen::InsensitiveStr::Ascii("secontly"), &["secondly"]),
            (
                dictgen::InsensitiveStr::Ascii("instromentation"),
                &["instrumentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sasketchewan"),
                &["saskatchewan"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoif"), &["avoid"]),
            (dictgen::InsensitiveStr::Ascii("sooaside"), &["suicide"]),
            (
                dictgen::InsensitiveStr::Ascii("tihngs"),
                &["things", "tings"],
            ),
            (dictgen::InsensitiveStr::Ascii("nown"), &["known", "noun"]),
            (dictgen::InsensitiveStr::Ascii("dateteim"), &["datetime"]),
            (dictgen::InsensitiveStr::Ascii("abusrdity"), &["absurdity"]),
            (
                dictgen::InsensitiveStr::Ascii("strucutred"),
                &["structured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envinronments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbiten"), &["forbidden"]),
            (
                dictgen::InsensitiveStr::Ascii("confgiurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vesseles"), &["vessels"]),
            (
                dictgen::InsensitiveStr::Ascii("expriences"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("resopnses"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("cincinatti"),
                &["cincinnati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parametized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("satistfy"), &["satisfy"]),
            (
                dictgen::InsensitiveStr::Ascii("regularlizing"),
                &["regularizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("clausens"), &["clauses"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetnally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlefeild"),
                &["battlefield"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationilism"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("consisant"), &["consistent"]),
            (dictgen::InsensitiveStr::Ascii("ensalved"), &["enslaved"]),
            (
                dictgen::InsensitiveStr::Ascii("envinroment"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("frquently"), &["frequently"]),
            (dictgen::InsensitiveStr::Ascii("deafeted"), &["defeated"]),
            (
                dictgen::InsensitiveStr::Ascii("syntaxical"),
                &["syntactical"],
            ),
            (dictgen::InsensitiveStr::Ascii("archver"), &["archiver"]),
            (dictgen::InsensitiveStr::Ascii("schmatics"), &["schematics"]),
            (dictgen::InsensitiveStr::Ascii("statisics"), &["statistics"]),
            (dictgen::InsensitiveStr::Ascii("fomulate"), &["formulate"]),
            (
                dictgen::InsensitiveStr::Ascii("assignemtn"),
                &["assignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concievable"),
                &["conceivable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repositionning"),
                &["repositioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconnets"), &["reconnects"]),
            (dictgen::InsensitiveStr::Ascii("necesary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("mezmoriser"),
                &["mesmeriser"],
            ),
            (dictgen::InsensitiveStr::Ascii("nestalgic"), &["nostalgic"]),
            (
                dictgen::InsensitiveStr::Ascii("helicpoter"),
                &["helicopter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("belived"),
                &["believed", "beloved"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conetxt"),
                &["context", "connect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgroudn"),
                &["backgrounds", "background"],
            ),
            (dictgen::InsensitiveStr::Ascii("spectified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("javascipt"), &["javascript"]),
            (
                dictgen::InsensitiveStr::Ascii("generlisations"),
                &["generalisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dyanmically"),
                &["dynamically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commuincate"),
                &["communicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanwiched"), &["sandwiched"]),
            (dictgen::InsensitiveStr::Ascii("dleivery"), &["delivery"]),
            (
                dictgen::InsensitiveStr::Ascii("propersity"),
                &["propensity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("athletisism"),
                &["athleticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propectless"),
                &["prospectless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electrobytes"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("widgeds"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("cylindres"), &["cylinders"]),
            (dictgen::InsensitiveStr::Ascii("barcelets"), &["bracelets"]),
            (
                dictgen::InsensitiveStr::Ascii("weightligting"),
                &["weightlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmisakably"),
                &["unmistakably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternitivs"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("charactes"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("competenze"),
                &["competence"],
            ),
            (dictgen::InsensitiveStr::Ascii("belogs"), &["belongs"]),
            (dictgen::InsensitiveStr::Ascii("pacakges"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("pocesses"),
                &["processes", "possesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("traitour"), &["traitor"]),
            (dictgen::InsensitiveStr::Ascii("coducts"), &["conducts"]),
            (dictgen::InsensitiveStr::Ascii("tempets"), &["tempest"]),
            (
                dictgen::InsensitiveStr::Ascii("comparigons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrumenetation"),
                &["instrumentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechine"), &["machine"]),
            (
                dictgen::InsensitiveStr::Ascii("reaktivate"),
                &["reactivate"],
            ),
            (dictgen::InsensitiveStr::Ascii("rephrasse"), &["rephrase"]),
            (
                dictgen::InsensitiveStr::Ascii("disconnnect"),
                &["disconnect"],
            ),
            (dictgen::InsensitiveStr::Ascii("earlies"), &["earliest"]),
            (
                dictgen::InsensitiveStr::Ascii("interpoloation"),
                &["interpolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioations"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoweldgements"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("recylcing"), &["recycling"]),
            (
                dictgen::InsensitiveStr::Ascii("analyticaly"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlained"), &["explained"]),
            (
                dictgen::InsensitiveStr::Ascii("encorporating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacumes"), &["vacuums"]),
            (
                dictgen::InsensitiveStr::Ascii("coorespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquisiton"),
                &["acquisition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tournamnets"),
                &["tournaments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isolatin"),
                &["isolating", "isolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritsations"),
                &["prioritisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("valiators"), &["validators"]),
            (
                dictgen::InsensitiveStr::Ascii("positivitiy"),
                &["positivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trusthworthiness"),
                &["trustworthiness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordonation"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abord"),
                &["abort", "aboard", "board"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaiton"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("overlordes"), &["overlords"]),
            (
                dictgen::InsensitiveStr::Ascii("pyhsically"),
                &["physically"],
            ),
            (dictgen::InsensitiveStr::Ascii("snappin"), &["snapping"]),
            (dictgen::InsensitiveStr::Ascii("orignated"), &["originated"]),
            (
                dictgen::InsensitiveStr::Ascii("transational"),
                &[
                    "transitional",
                    "transactional",
                    "translational",
                    "transnational",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("utlimately"),
                &["ultimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("ulterioare"), &["ulterior"]),
            (dictgen::InsensitiveStr::Ascii("resgistry"), &["registry"]),
            (
                dictgen::InsensitiveStr::Ascii("admininstered"),
                &["administered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precendece"),
                &["precedence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegitariens"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mainaining"),
                &["maintaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("attracties"), &["attracts"]),
            (dictgen::InsensitiveStr::Ascii("menthods"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("preorderers"),
                &["preorders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accuray"),
                &["accuracy", "actuary"],
            ),
            (dictgen::InsensitiveStr::Ascii("asborbed"), &["absorbed"]),
            (
                dictgen::InsensitiveStr::Ascii("pluin"),
                &["plugin", "plain"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacialisation"),
                &["specialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reedemed"), &["redeemed"]),
            (
                dictgen::InsensitiveStr::Ascii("consttruction"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("featchures"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("mountpionts"),
                &["mountpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("colmns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("speration"), &["separation"]),
            (dictgen::InsensitiveStr::Ascii("badmitten"), &["badminton"]),
            (dictgen::InsensitiveStr::Ascii("specficied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("multy"), &["multi"]),
            (dictgen::InsensitiveStr::Ascii("customn"), &["custom"]),
            (dictgen::InsensitiveStr::Ascii("secotr"), &["sector"]),
            (
                dictgen::InsensitiveStr::Ascii("publushing"),
                &["publishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comunications"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiescin"),
                &["acquiescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inherintly"),
                &["inherently"],
            ),
            (dictgen::InsensitiveStr::Ascii("trigerr"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("scoprion"), &["scorpion"]),
            (dictgen::InsensitiveStr::Ascii("extesion"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("vascetomy"), &["vasectomy"]),
            (
                dictgen::InsensitiveStr::Ascii("perdicting"),
                &["predicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("execured"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("corrseponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("poligon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("seacrher"), &["searcher"]),
            (
                dictgen::InsensitiveStr::Ascii("homosexuales"),
                &["homosexuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelaration"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("alernates"), &["alternates"]),
            (dictgen::InsensitiveStr::Ascii("afternon"), &["afternoon"]),
            (
                dictgen::InsensitiveStr::Ascii("unsoliticed"),
                &["unsolicited"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveyr"), &["every"]),
            (dictgen::InsensitiveStr::Ascii("arraned"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("oscilated"), &["oscillated"]),
            (
                dictgen::InsensitiveStr::Ascii("placeholdrs"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firts"),
                &["flirts", "first"],
            ),
            (dictgen::InsensitiveStr::Ascii("knockously"), &["noxiously"]),
            (dictgen::InsensitiveStr::Ascii("detatil"), &["detail"]),
            (
                dictgen::InsensitiveStr::Ascii("obtaine"),
                &["obtain", "obtained", "obtains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aplycation"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("visist"), &["visit"]),
            (
                dictgen::InsensitiveStr::Ascii("refelective"),
                &["reflective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timstamped"),
                &["timestamped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporarilu"),
                &["temporary"],
            ),
            (dictgen::InsensitiveStr::Ascii("appedn"), &["append"]),
            (dictgen::InsensitiveStr::Ascii("shadasloo"), &["shadaloo"]),
            (
                dictgen::InsensitiveStr::Ascii("appropirate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforrcement"),
                &["enforcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advocading"),
                &["advocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("reynlods"), &["reynolds"]),
            (
                dictgen::InsensitiveStr::Ascii("assigmnent"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("percusson"), &["percussion"]),
            (dictgen::InsensitiveStr::Ascii("depeneded"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("delimeter"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("wser"), &["user"]),
            (
                dictgen::InsensitiveStr::Ascii("frustracion"),
                &["frustration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interchangeble"),
                &["interchangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("hesistate"), &["hesitate"]),
            (dictgen::InsensitiveStr::Ascii("socketted"), &["socketed"]),
            (dictgen::InsensitiveStr::Ascii("synomym"), &["synonym"]),
            (
                dictgen::InsensitiveStr::Ascii("discritized"),
                &["discretized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corressponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequancies"),
                &["frequencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythesises"),
                &["synthesises"],
            ),
            (dictgen::InsensitiveStr::Ascii("balnaced"), &["balanced"]),
            (
                dictgen::InsensitiveStr::Ascii("forunately"),
                &["fortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prominetly"),
                &["prominently"],
            ),
            (dictgen::InsensitiveStr::Ascii("frequencey"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("equilvalently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypocrties"),
                &["hypocrites"],
            ),
            (dictgen::InsensitiveStr::Ascii("aspectos"), &["aspects"]),
            (dictgen::InsensitiveStr::Ascii("expetcts"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("verfy"), &["verify"]),
            (
                dictgen::InsensitiveStr::Ascii("polititian"),
                &["politician"],
            ),
            (dictgen::InsensitiveStr::Ascii("airrplane"), &["airplane"]),
            (
                dictgen::InsensitiveStr::Ascii("temproraly"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("decondes"), &["decodes"]),
            (
                dictgen::InsensitiveStr::Ascii("cervial"),
                &["cervical", "servile", "serval"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rectanguar"),
                &["rectangular"],
            ),
            (dictgen::InsensitiveStr::Ascii("zomebies"), &["zombies"]),
            (dictgen::InsensitiveStr::Ascii("crypitc"), &["cryptic"]),
            (dictgen::InsensitiveStr::Ascii("trickey"), &["trickery"]),
            (dictgen::InsensitiveStr::Ascii("adavanced"), &["advanced"]),
            (dictgen::InsensitiveStr::Ascii("acustom"), &["accustom"]),
            (dictgen::InsensitiveStr::Ascii("realods"), &["reloads"]),
            (dictgen::InsensitiveStr::Ascii("sublassed"), &["subclassed"]),
            (
                dictgen::InsensitiveStr::Ascii("enterpises"),
                &["enterprises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportinal"),
                &["proportional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetectedly"),
                &["expectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("obisdian"), &["obsidian"]),
            (dictgen::InsensitiveStr::Ascii("ammend"), &["amend"]),
            (dictgen::InsensitiveStr::Ascii("updatd"), &["updated"]),
            (
                dictgen::InsensitiveStr::Ascii("reliabilty"),
                &["reliability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survibability"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivlance"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathcer"), &["matcher"]),
            (
                dictgen::InsensitiveStr::Ascii("soveriegnty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("underog"), &["undergo"]),
            (
                dictgen::InsensitiveStr::Ascii("lieutanant"),
                &["lieutenant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condescening"),
                &["condescending"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialgos"), &["dialogs"]),
            (
                dictgen::InsensitiveStr::Ascii("tast"),
                &["taste", "task", "test"],
            ),
            (dictgen::InsensitiveStr::Ascii("hipotesis"), &["hypothesis"]),
            (
                dictgen::InsensitiveStr::Ascii("necromaner"),
                &["necromancer"],
            ),
            (dictgen::InsensitiveStr::Ascii("frineds"), &["friends"]),
            (
                dictgen::InsensitiveStr::Ascii("opionated"),
                &["opinionated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfeares"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custoization"),
                &["customization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activitites"),
                &["activities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delocate"),
                &["deallocate", "delicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("gullibe"), &["gullible"]),
            (dictgen::InsensitiveStr::Ascii("symbollic"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("plantiff"), &["plaintiff"]),
            (dictgen::InsensitiveStr::Ascii("lubricat"), &["lubricant"]),
            (dictgen::InsensitiveStr::Ascii("adhearing"), &["adhering"]),
            (
                dictgen::InsensitiveStr::Ascii("corrrespondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architeture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gaurantees"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostitutke"),
                &["prostitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("packacging"), &["packaging"]),
            (
                dictgen::InsensitiveStr::Ascii("quadroopling"),
                &["quadrupling"],
            ),
            (dictgen::InsensitiveStr::Ascii("sipported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("mangage"), &["manage"]),
            (
                dictgen::InsensitiveStr::Ascii("effeceively"),
                &["effectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("suh"), &["such"]),
            (dictgen::InsensitiveStr::Ascii("announcess"), &["announces"]),
            (dictgen::InsensitiveStr::Ascii("coverate"), &["coverage"]),
            (dictgen::InsensitiveStr::Ascii("mandase"), &["mandates"]),
            (dictgen::InsensitiveStr::Ascii("valkirye"), &["valkyrie"]),
            (dictgen::InsensitiveStr::Ascii("unnused"), &["unused"]),
            (
                dictgen::InsensitiveStr::Ascii("blueberris"),
                &["blueberries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inflamable"),
                &["inflatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("befire"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepreneures"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("rwite"), &["write"]),
            (dictgen::InsensitiveStr::Ascii("cigeratte"), &["cigarette"]),
            (dictgen::InsensitiveStr::Ascii("refferred"), &["referred"]),
            (
                dictgen::InsensitiveStr::Ascii("initailised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("secondardy"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("capitilised"),
                &["capitalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("custum"), &["custom"]),
            (
                dictgen::InsensitiveStr::Ascii("aspriation"),
                &["aspiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approprietly"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticisin"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cehckpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("mateirals"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("precendences"),
                &["precedences"],
            ),
            (dictgen::InsensitiveStr::Ascii("nulify"), &["nullify"]),
            (dictgen::InsensitiveStr::Ascii("vicotries"), &["victories"]),
            (
                dictgen::InsensitiveStr::Ascii("corrolated"),
                &["correlated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symmectrically"),
                &["symmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("dogfather"), &["godfather"]),
            (
                dictgen::InsensitiveStr::Ascii("executionar"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("employeurs"), &["employer"]),
            (dictgen::InsensitiveStr::Ascii("sectionis"), &["sections"]),
            (
                dictgen::InsensitiveStr::Ascii("exponentielly"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("monentary"), &["momentary"]),
            (
                dictgen::InsensitiveStr::Ascii("predecated"),
                &["predicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reasonbly"), &["reasonably"]),
            (
                dictgen::InsensitiveStr::Ascii("desperato"),
                &["desperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("undefiend"), &["undefined"]),
            (dictgen::InsensitiveStr::Ascii("japanease"), &["japanese"]),
            (dictgen::InsensitiveStr::Ascii("sinically"), &["cynically"]),
            (
                dictgen::InsensitiveStr::Ascii("hallusination"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("thereads"), &["threads"]),
            (
                dictgen::InsensitiveStr::Ascii("accleometers"),
                &["accelerometers"],
            ),
            (dictgen::InsensitiveStr::Ascii("wolud"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("enforcmement"),
                &["enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("librarse"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("califonria"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brianwashed"),
                &["brainwashed"],
            ),
            (dictgen::InsensitiveStr::Ascii("possbily"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("prepetuated"),
                &["perpetuated"],
            ),
            (dictgen::InsensitiveStr::Ascii("suttlety"), &["subtlety"]),
            (dictgen::InsensitiveStr::Ascii("obleekly"), &["obliquely"]),
            (
                dictgen::InsensitiveStr::Ascii("infestating"),
                &["infestation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclcoked"),
                &["overclocked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replicaions"),
                &["replications"],
            ),
            (dictgen::InsensitiveStr::Ascii("quibic"), &["cubic"]),
            (
                dictgen::InsensitiveStr::Ascii("prograaming"),
                &["programming"],
            ),
            (dictgen::InsensitiveStr::Ascii("buddhisim"), &["buddhism"]),
            (
                dictgen::InsensitiveStr::Ascii("inialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legenadries"),
                &["legendaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("autum"), &["autumn"]),
            (
                dictgen::InsensitiveStr::Ascii("constraings"),
                &["constraints", "constrains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correnspondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("nagivates"), &["navigates"]),
            (dictgen::InsensitiveStr::Ascii("earch"), &["search", "each"]),
            (dictgen::InsensitiveStr::Ascii("stealhty"), &["stealthy"]),
            (dictgen::InsensitiveStr::Ascii("referrring"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("mirrorn"), &["mirror"]),
            (dictgen::InsensitiveStr::Ascii("queueuing"), &["queueing"]),
            (dictgen::InsensitiveStr::Ascii("databsase"), &["database"]),
            (
                dictgen::InsensitiveStr::Ascii("afther"),
                &["after", "father"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrinics"), &["intrinsics"]),
            (
                dictgen::InsensitiveStr::Ascii("sensibilisiert"),
                &["sensibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("invovling"), &["involving"]),
            (dictgen::InsensitiveStr::Ascii("mommentum"), &["momentum"]),
            (
                dictgen::InsensitiveStr::Ascii("diasspointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspitions"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("conveter"), &["converter"]),
            (dictgen::InsensitiveStr::Ascii("serialze"), &["serialize"]),
            (
                dictgen::InsensitiveStr::Ascii("partiarchy"),
                &["patriarchy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("matchine"),
                &["machine", "matching"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteratoin"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("oprational"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirorred"), &["mirrored"]),
            (
                dictgen::InsensitiveStr::Ascii("charaterized"),
                &["characterized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miror"),
                &["mirror", "minor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknownledging"),
                &["acknowledging"],
            ),
            (dictgen::InsensitiveStr::Ascii("releative"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("deccelerating"),
                &["decelerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("seporate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("cjoice"), &["choice"]),
            (dictgen::InsensitiveStr::Ascii("expexting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("amont"),
                &["among", "amount"],
            ),
            (dictgen::InsensitiveStr::Ascii("justifiy"), &["justify"]),
            (
                dictgen::InsensitiveStr::Ascii("maximazing"),
                &["maximizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inheritied"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidiate"),
                &["invalidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ilterate"), &["literate"]),
            (
                dictgen::InsensitiveStr::Ascii("torunament"),
                &["tournament"],
            ),
            (dictgen::InsensitiveStr::Ascii("seseeder"), &["seceder"]),
            (dictgen::InsensitiveStr::Ascii("regstries"), &["registries"]),
            (
                dictgen::InsensitiveStr::Ascii("compilicates"),
                &["complicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("becouse"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("railraod"), &["railroad"]),
            (
                dictgen::InsensitiveStr::Ascii("skatebored"),
                &["skateboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("uggestion"), &["suggestion"]),
            (
                dictgen::InsensitiveStr::Ascii("propesterous"),
                &["preposterous"],
            ),
            (dictgen::InsensitiveStr::Ascii("cataclysim"), &["cataclysm"]),
            (dictgen::InsensitiveStr::Ascii("ediit"), &["edit"]),
            (dictgen::InsensitiveStr::Ascii("mkssing"), &["missing"]),
            (dictgen::InsensitiveStr::Ascii("relavant"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("friendlis"), &["friendlies"]),
            (dictgen::InsensitiveStr::Ascii("referrrers"), &["referrers"]),
            (
                dictgen::InsensitiveStr::Ascii("coopeartes"),
                &["cooperates"],
            ),
            (dictgen::InsensitiveStr::Ascii("supsense"), &["suspense"]),
            (dictgen::InsensitiveStr::Ascii("labriynth"), &["labyrinth"]),
            (dictgen::InsensitiveStr::Ascii("semestre"), &["semester"]),
            (
                dictgen::InsensitiveStr::Ascii("overlaoded"),
                &["overloaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parttioning"),
                &["partitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffentiates"),
                &["differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquision"),
                &["acquisition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltrarte"),
                &["infiltrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("collegaue"), &["colleague"]),
            (dictgen::InsensitiveStr::Ascii("dialgoue"), &["dialogue"]),
            (dictgen::InsensitiveStr::Ascii("mulltiple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("constructior"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("turrest"), &["turrets"]),
            (dictgen::InsensitiveStr::Ascii("maximnum"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("nulable"), &["nullable"]),
            (dictgen::InsensitiveStr::Ascii("alchool"), &["alcohol"]),
            (dictgen::InsensitiveStr::Ascii("moprhine"), &["morphine"]),
            (dictgen::InsensitiveStr::Ascii("livestrem"), &["livestream"]),
            (dictgen::InsensitiveStr::Ascii("mixxing"), &["mixing"]),
            (dictgen::InsensitiveStr::Ascii("explined"), &["explained"]),
            (
                dictgen::InsensitiveStr::Ascii("absed"),
                &["abased", "based"],
            ),
            (dictgen::InsensitiveStr::Ascii("blockin"), &["blocking"]),
            (dictgen::InsensitiveStr::Ascii("acccused"), &["accused"]),
            (dictgen::InsensitiveStr::Ascii("rescource"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("correcs"), &["corrects"]),
            (
                dictgen::InsensitiveStr::Ascii("goosebumbps"),
                &["goosebumps"],
            ),
            (dictgen::InsensitiveStr::Ascii("wezzals"), &["weasels"]),
            (
                dictgen::InsensitiveStr::Ascii("assigntments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("linakage"), &["linkage"]),
            (dictgen::InsensitiveStr::Ascii("maintaied"), &["maintained"]),
            (
                dictgen::InsensitiveStr::Ascii("distinghished"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("oif"), &["of"]),
            (
                dictgen::InsensitiveStr::Ascii("seriolization"),
                &["serialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixifers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("standatd"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("studnts"), &["students"]),
            (dictgen::InsensitiveStr::Ascii("altought"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("sentinents"),
                &["sentiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrption"),
                &["discretion"],
            ),
            (dictgen::InsensitiveStr::Ascii("fridey"), &["friday"]),
            (
                dictgen::InsensitiveStr::Ascii("punishible"),
                &["punishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("grahpite"), &["graphite"]),
            (dictgen::InsensitiveStr::Ascii("eveolution"), &["evolution"]),
            (dictgen::InsensitiveStr::Ascii("iteate"), &["iterate"]),
            (
                dictgen::InsensitiveStr::Ascii("inquizitor"),
                &["inquisitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resotrative"),
                &["restorative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agressivley"),
                &["aggressive", "aggressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("desturbed"), &["disturbed"]),
            (
                dictgen::InsensitiveStr::Ascii("downvotest"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovelap"), &["overlap"]),
            (
                dictgen::InsensitiveStr::Ascii("acommpanying"),
                &["accompanying"],
            ),
            (dictgen::InsensitiveStr::Ascii("atcive"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("usera"), &["users"]),
            (dictgen::InsensitiveStr::Ascii("muscualr"), &["muscular"]),
            (
                dictgen::InsensitiveStr::Ascii("predisposte"),
                &["predisposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("stoppoing"), &["stopping"]),
            (dictgen::InsensitiveStr::Ascii("seelction"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("electrodan"), &["electron"]),
            (dictgen::InsensitiveStr::Ascii("encondings"), &["encodings"]),
            (dictgen::InsensitiveStr::Ascii("nativelyx"), &["natively"]),
            (dictgen::InsensitiveStr::Ascii("forumulas"), &["formulas"]),
            (
                dictgen::InsensitiveStr::Ascii("capitarist"),
                &["capitalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precondtioners"),
                &["preconditioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("syllabels"), &["syllables"]),
            (
                dictgen::InsensitiveStr::Ascii("persitentely"),
                &["persistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("satements"), &["statements"]),
            (
                dictgen::InsensitiveStr::Ascii("calculuated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("immortales"), &["immortals"]),
            (dictgen::InsensitiveStr::Ascii("endiif"), &["endif"]),
            (
                dictgen::InsensitiveStr::Ascii("unrealesed"),
                &["unreleased"],
            ),
            (dictgen::InsensitiveStr::Ascii("brendamn"), &["brendan"]),
            (dictgen::InsensitiveStr::Ascii("analyitics"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("alog"), &["along"]),
            (
                dictgen::InsensitiveStr::Ascii("intervension"),
                &["intervention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaproppriate"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("fronend"), &["frontend"]),
            (
                dictgen::InsensitiveStr::Ascii("perhiperals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizonlly"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("closeed"), &["closed"]),
            (dictgen::InsensitiveStr::Ascii("enthusiam"), &["enthusiasm"]),
            (
                dictgen::InsensitiveStr::Ascii("casaulties"),
                &["casualties"],
            ),
            (dictgen::InsensitiveStr::Ascii("souvinears"), &["souvenirs"]),
            (
                dictgen::InsensitiveStr::Ascii("anarkistic"),
                &["anarchistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australien"),
                &["australian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communites"),
                &["communities"],
            ),
            (dictgen::InsensitiveStr::Ascii("instaances"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("msot"), &["most"]),
            (dictgen::InsensitiveStr::Ascii("smaurai"), &["samurai"]),
            (
                dictgen::InsensitiveStr::Ascii("repreciates"),
                &["depreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreands"), &["operands"]),
            (
                dictgen::InsensitiveStr::Ascii("vailidations"),
                &["validations"],
            ),
            (dictgen::InsensitiveStr::Ascii("remianing"), &["remaining"]),
            (
                dictgen::InsensitiveStr::Ascii("unathorised"),
                &["unauthorised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trustwothy"),
                &["trustworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("internetis"), &["internets"]),
            (dictgen::InsensitiveStr::Ascii("forlder"), &["folder"]),
            (dictgen::InsensitiveStr::Ascii("furture"), &["future"]),
            (dictgen::InsensitiveStr::Ascii("italianos"), &["italians"]),
            (dictgen::InsensitiveStr::Ascii("nottices"), &["notices"]),
            (
                dictgen::InsensitiveStr::Ascii("spectification"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("commision"), &["commission"]),
            (
                dictgen::InsensitiveStr::Ascii("authencating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("jornaling"), &["journaling"]),
            (dictgen::InsensitiveStr::Ascii("enuf"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("mimimisation"),
                &["minimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("refusla"), &["refusal"]),
            (dictgen::InsensitiveStr::Ascii("abiguous"), &["ambiguous"]),
            (dictgen::InsensitiveStr::Ascii("scitilla"), &["scintilla"]),
            (
                dictgen::InsensitiveStr::Ascii("tablulations"),
                &["tabulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("multipul"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("tablulation"),
                &["tabulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("orignials"), &["originals"]),
            (dictgen::InsensitiveStr::Ascii("thosse"), &["those"]),
            (dictgen::InsensitiveStr::Ascii("regrds"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("filies"), &["files"]),
            (
                dictgen::InsensitiveStr::Ascii("highscholl"),
                &["highschool"],
            ),
            (dictgen::InsensitiveStr::Ascii("bourgeios"), &["bourgeois"]),
            (dictgen::InsensitiveStr::Ascii("varibles"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("executabes"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranistions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("wouold"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("exploition"),
                &["exploiting", "explosion", "exploitation", "exploit"],
            ),
            (dictgen::InsensitiveStr::Ascii("frequenly"), &["frequently"]),
            (dictgen::InsensitiveStr::Ascii("referenece"), &["reference"]),
            (
                dictgen::InsensitiveStr::Ascii("proffession"),
                &["profession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pscyhiatric"),
                &["psychiatric"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicussion"), &["discussion"]),
            (dictgen::InsensitiveStr::Ascii("sourthern"), &["southern"]),
            (
                dictgen::InsensitiveStr::Ascii("funktionally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("scheduale"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("evenhtually"),
                &["eventually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowedgements"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("genreates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("reploying"),
                &["redeploying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("westernese"),
                &["westerners"],
            ),
            (dictgen::InsensitiveStr::Ascii("peapels"), &["peoples"]),
            (
                dictgen::InsensitiveStr::Ascii("coneections"),
                &["connections"],
            ),
            (dictgen::InsensitiveStr::Ascii("algins"), &["aligns"]),
            (
                dictgen::InsensitiveStr::Ascii("termniates"),
                &["terminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhausto"), &["exhaustion"]),
            (dictgen::InsensitiveStr::Ascii("geerating"), &["generating"]),
            (
                dictgen::InsensitiveStr::Ascii("expectataions"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("boootstrap"), &["bootstrap"]),
            (dictgen::InsensitiveStr::Ascii("ftbs"), &["ftbfs"]),
            (dictgen::InsensitiveStr::Ascii("mandats"), &["mandates"]),
            (
                dictgen::InsensitiveStr::Ascii("franchieses"),
                &["franchises"],
            ),
            (dictgen::InsensitiveStr::Ascii("reviewd"), &["reviewed"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistancy"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containuations"),
                &["continuations"],
            ),
            (dictgen::InsensitiveStr::Ascii("respnses"), &["responses"]),
            (dictgen::InsensitiveStr::Ascii("oppposite"), &["opposite"]),
            (dictgen::InsensitiveStr::Ascii("commputed"), &["computed"]),
            (dictgen::InsensitiveStr::Ascii("parrent"), &["parent"]),
            (dictgen::InsensitiveStr::Ascii("propertu"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("deamiguation"),
                &["disambiguation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exilerates"),
                &["exhilarates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metabolsim"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("afinity"), &["affinity"]),
            (dictgen::InsensitiveStr::Ascii("insersts"), &["inserts"]),
            (
                dictgen::InsensitiveStr::Ascii("importamce"),
                &["importance"],
            ),
            (dictgen::InsensitiveStr::Ascii("arciver"), &["archiver"]),
            (
                dictgen::InsensitiveStr::Ascii("unorigonal"),
                &["unoriginal"],
            ),
            (dictgen::InsensitiveStr::Ascii("airplans"), &["airplanes"]),
            (dictgen::InsensitiveStr::Ascii("partititon"), &["partition"]),
            (
                dictgen::InsensitiveStr::Ascii("sistematising"),
                &["systematising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizzontal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specficication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("coucil"), &["council"]),
            (
                dictgen::InsensitiveStr::Ascii("depercates"),
                &["deprecates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimpplementating"),
                &["reimplementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalistes"),
                &["nationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggestin"), &["suggestion"]),
            (dictgen::InsensitiveStr::Ascii("direcoty"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("dribbeled"), &["dribbled"]),
            (
                dictgen::InsensitiveStr::Ascii("dfferently"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("eletric"), &["electric"]),
            (
                dictgen::InsensitiveStr::Ascii("conventiences"),
                &["conveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("thrird"), &["third"]),
            (
                dictgen::InsensitiveStr::Ascii("insturmental"),
                &["instrumental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("losin"),
                &["losing", "login", "loin", "rosin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpresed"),
                &["compressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estabishing"),
                &["establishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("explodeds"), &["explodes"]),
            (dictgen::InsensitiveStr::Ascii("fighing"), &["fighting"]),
            (dictgen::InsensitiveStr::Ascii("delegrate"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("headseat"), &["headset"]),
            (
                dictgen::InsensitiveStr::Ascii("adminstrations"),
                &["administrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theraupetic"),
                &["therapeutic"],
            ),
            (dictgen::InsensitiveStr::Ascii("definat"), &["defiant"]),
            (dictgen::InsensitiveStr::Ascii("featchured"), &["featured"]),
            (
                dictgen::InsensitiveStr::Ascii("trainling"),
                &["trailing", "training"],
            ),
            (dictgen::InsensitiveStr::Ascii("avataras"), &["avatars"]),
            (
                dictgen::InsensitiveStr::Ascii("threadsave"),
                &["threadsafe"],
            ),
            (dictgen::InsensitiveStr::Ascii("tosbiba"), &["toshiba"]),
            (
                dictgen::InsensitiveStr::Ascii("needlessley"),
                &["needlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocurs"), &["occurs"]),
            (
                dictgen::InsensitiveStr::Ascii("compramises"),
                &["compromises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("powerlisting"),
                &["powerlifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("sunscren"), &["sunscreen"]),
            (
                dictgen::InsensitiveStr::Ascii("sympithizers"),
                &["sympathizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singificant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulatin"),
                &["simulation", "simulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("lipstics"), &["lipsticks"]),
            (
                dictgen::InsensitiveStr::Ascii("alternitives"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compensationg"),
                &["compensating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comissioning"),
                &["commissioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("viloently"), &["violently"]),
            (dictgen::InsensitiveStr::Ascii("privillege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("allergisch"), &["allergic"]),
            (
                dictgen::InsensitiveStr::Ascii("journalistische"),
                &["journalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsturcted"),
                &["reconstructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("polarlizes"), &["polarizes"]),
            (dictgen::InsensitiveStr::Ascii("smooter"), &["smoother"]),
            (dictgen::InsensitiveStr::Ascii("campiagns"), &["campaigns"]),
            (dictgen::InsensitiveStr::Ascii("bracese"), &["braces"]),
            (
                dictgen::InsensitiveStr::Ascii("adges"),
                &["edges", "badges", "adages"],
            ),
            (dictgen::InsensitiveStr::Ascii("tutorual"), &["tutorial"]),
            (dictgen::InsensitiveStr::Ascii("coundition"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("conversationa"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("updaes"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("picutres"), &["pictures"]),
            (
                dictgen::InsensitiveStr::Ascii("approxamately"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("miltiant"), &["militant"]),
            (dictgen::InsensitiveStr::Ascii("passworss"), &["passwords"]),
            (
                dictgen::InsensitiveStr::Ascii("docuementing"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("dowhill"), &["downhill"]),
            (dictgen::InsensitiveStr::Ascii("surveyer"), &["surveyor"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegoitations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizil"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("alternitiveness"),
                &["alternativeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("actiavtor"), &["activator"]),
            (dictgen::InsensitiveStr::Ascii("swepth"), &["swept"]),
            (dictgen::InsensitiveStr::Ascii("clairify"), &["clarify"]),
            (
                dictgen::InsensitiveStr::Ascii("poportional"),
                &["proportional"],
            ),
            (dictgen::InsensitiveStr::Ascii("caputures"), &["captures"]),
            (
                dictgen::InsensitiveStr::Ascii("experencing"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("acept"), &["accept"]),
            (
                dictgen::InsensitiveStr::Ascii("funchtionning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distiction"),
                &["distinction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motherbord"),
                &["motherboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shuits"),
                &["shuts", "suits", "shits"],
            ),
            (dictgen::InsensitiveStr::Ascii("publsh"), &["publish"]),
            (dictgen::InsensitiveStr::Ascii("specfiy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("neigbour"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("workround"), &["workaround"]),
            (dictgen::InsensitiveStr::Ascii("micoscopy"), &["microscopy"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenional"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifer"), &["verifier"]),
            (dictgen::InsensitiveStr::Ascii("depden"), &["depend"]),
            (
                dictgen::InsensitiveStr::Ascii("minimalisitc"),
                &["minimalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetex"), &["vertex"]),
            (
                dictgen::InsensitiveStr::Ascii("winnters"),
                &["winters", "winners"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitis"), &["limits"]),
            (
                dictgen::InsensitiveStr::Ascii("compednium"),
                &["compendium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necormancer"),
                &["necromancer"],
            ),
            (dictgen::InsensitiveStr::Ascii("pressiuos"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("itsef"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("respwan"), &["respawn"]),
            (
                dictgen::InsensitiveStr::Ascii("programattically"),
                &["programmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("interveal"), &["interval"]),
            (
                dictgen::InsensitiveStr::Ascii("degenarated"),
                &["degenerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronuncuation"),
                &["pronunciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aboluste"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("deccrement"), &["decrement"]),
            (dictgen::InsensitiveStr::Ascii("actualty"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("hyginic"), &["hygienic"]),
            (
                dictgen::InsensitiveStr::Ascii("challegner"),
                &["challenger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelegently"),
                &["intelligently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battelship"),
                &["battleship"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporing"), &["importing"]),
            (dictgen::InsensitiveStr::Ascii("expetcing"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("transferrable"),
                &["transferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotitation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("intstead"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("monoploy"), &["monopoly"]),
            (
                dictgen::InsensitiveStr::Ascii("payement"),
                &["payment", "pavement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internationnally"),
                &["internationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuls"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepated"), &["prepared"]),
            (dictgen::InsensitiveStr::Ascii("epxlicit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("configurses"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considert"),
                &["considerate", "considered", "consider"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrepency"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("executres"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("ceremone"), &["ceremonies"]),
            (dictgen::InsensitiveStr::Ascii("sicinctly"), &["succinctly"]),
            (dictgen::InsensitiveStr::Ascii("messege"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("diferrentiate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomramces"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administrativne"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlyser"), &["analyser"]),
            (dictgen::InsensitiveStr::Ascii("nevere"), &["never"]),
            (
                dictgen::InsensitiveStr::Ascii("californiia"),
                &["californian"],
            ),
            (dictgen::InsensitiveStr::Ascii("geenrics"), &["generics"]),
            (dictgen::InsensitiveStr::Ascii("actully"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("depecated"), &["deprecated"]),
            (
                dictgen::InsensitiveStr::Ascii("appreaciting"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprectice"),
                &["apprentice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reparamterising"),
                &["reparameterising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differenly"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilizin"), &["utilizing"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifyers"),
                &["qualifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("dominno"), &["dominion"]),
            (dictgen::InsensitiveStr::Ascii("vaelues"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("exhilerated"),
                &["exhilarated"],
            ),
            (dictgen::InsensitiveStr::Ascii("spedified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("permutions"),
                &["permutations"],
            ),
            (dictgen::InsensitiveStr::Ascii("enligthen"), &["enlighten"]),
            (
                dictgen::InsensitiveStr::Ascii("stardardizing"),
                &["standardizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("breshing"), &["brushing"]),
            (
                dictgen::InsensitiveStr::Ascii("experimntations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hepls"), &["helps"]),
            (dictgen::InsensitiveStr::Ascii("smller"), &["smaller"]),
            (dictgen::InsensitiveStr::Ascii("remvoved"), &["removed"]),
            (dictgen::InsensitiveStr::Ascii("editt"), &["edit"]),
            (dictgen::InsensitiveStr::Ascii("canoncical"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("piecin"), &["piecing"]),
            (
                dictgen::InsensitiveStr::Ascii("condiguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagease"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("tatoos"), &["tattoos"]),
            (dictgen::InsensitiveStr::Ascii("famousy"), &["famously"]),
            (dictgen::InsensitiveStr::Ascii("forgettin"), &["forgetting"]),
            (dictgen::InsensitiveStr::Ascii("weerd"), &["weird"]),
            (
                dictgen::InsensitiveStr::Ascii("authencicating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("nesseccary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("infreqeuncy"),
                &["infrequency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimntel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("appetitie"), &["appetite"]),
            (dictgen::InsensitiveStr::Ascii("epxecting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("calcucated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("humaniod"), &["humanoid"]),
            (dictgen::InsensitiveStr::Ascii("phlemy"), &["phlegmy"]),
            (dictgen::InsensitiveStr::Ascii("wiscle"), &["whistle"]),
            (
                dictgen::InsensitiveStr::Ascii("destinarions"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuasing"), &["causing"]),
            (dictgen::InsensitiveStr::Ascii("physial"), &["physical"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivly"),
                &["sensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveninece"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("opimism"), &["optimism"]),
            (dictgen::InsensitiveStr::Ascii("rimaniss"), &["reminisce"]),
            (
                dictgen::InsensitiveStr::Ascii("concantenating"),
                &["concatenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponteanously"),
                &["spontaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specailizes"),
                &["specializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrapency"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("improtant"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondense"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("viseral"), &["visceral"]),
            (
                dictgen::InsensitiveStr::Ascii("automobilies"),
                &["automobile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisin"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("beyone"), &["beyond"]),
            (dictgen::InsensitiveStr::Ascii("refersher"), &["refresher"]),
            (dictgen::InsensitiveStr::Ascii("expeses"), &["expenses"]),
            (
                dictgen::InsensitiveStr::Ascii("availality"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintelligble"),
                &["unintelligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("constented"), &["consented"]),
            (dictgen::InsensitiveStr::Ascii("sharpning"), &["sharpening"]),
            (
                dictgen::InsensitiveStr::Ascii("intitialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("executign"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("undescores"),
                &["underscores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applikation"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoerxia"), &["anorexia"]),
            (
                dictgen::InsensitiveStr::Ascii("singeletons"),
                &["singletons"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpsiring"), &["inspiring"]),
            (dictgen::InsensitiveStr::Ascii("temproral"), &["temporal"]),
            (
                dictgen::InsensitiveStr::Ascii("involentary"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("procudure"), &["procedure"]),
            (
                dictgen::InsensitiveStr::Ascii("categoically"),
                &["categorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southbrige"),
                &["southbridge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomrnaces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("crrespond"), &["correspond"]),
            (
                dictgen::InsensitiveStr::Ascii("jurisdicitons"),
                &["jurisdictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("assymetry"), &["asymmetry"]),
            (dictgen::InsensitiveStr::Ascii("allegedely"), &["allegedly"]),
            (
                dictgen::InsensitiveStr::Ascii("fluxtuations"),
                &["fluctuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotioations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sustainabilty"),
                &["sustainability"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuplaods"), &["reuploads"]),
            (
                dictgen::InsensitiveStr::Ascii("downsterams"),
                &["downstreams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refenrencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("pointzer"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("hearthsone"),
                &["hearthstone"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensitrive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("ukriane"), &["ukraine"]),
            (
                dictgen::InsensitiveStr::Ascii("alpbabetical"),
                &["alphabetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("vissible"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("swiches"),
                &["switches", "swishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sparating"), &["separating"]),
            (dictgen::InsensitiveStr::Ascii("meerkrat"), &["meerkat"]),
            (
                dictgen::InsensitiveStr::Ascii("pozitively"),
                &["positively"],
            ),
            (dictgen::InsensitiveStr::Ascii("boundays"), &["boundaries"]),
            (
                dictgen::InsensitiveStr::Ascii("successed"),
                &["succeeded", "success", "successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("preveiw"), &["preview"]),
            (
                dictgen::InsensitiveStr::Ascii("experamenters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("zionsit"), &["zionist"]),
            (
                dictgen::InsensitiveStr::Ascii("embaressing"),
                &["embarrassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooridated"),
                &["coordinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convorsations"),
                &["conversations"],
            ),
            (dictgen::InsensitiveStr::Ascii("beforere"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhelminly"),
                &["overwhelmingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("genrating"), &["generating"]),
            (dictgen::InsensitiveStr::Ascii("latitudie"), &["latitude"]),
            (
                dictgen::InsensitiveStr::Ascii("architypes"),
                &["archetypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("validtors"), &["validators"]),
            (
                dictgen::InsensitiveStr::Ascii("consttructor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("floride"), &["fluoride"]),
            (dictgen::InsensitiveStr::Ascii("encorder"), &["encoder"]),
            (dictgen::InsensitiveStr::Ascii("neesd"), &["needs"]),
            (
                dictgen::InsensitiveStr::Ascii("carismatic"),
                &["charismatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articluates"),
                &["articulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininal"), &["minimal"]),
            (dictgen::InsensitiveStr::Ascii("proceeeds"), &["proceeds"]),
            (dictgen::InsensitiveStr::Ascii("subsitute"), &["substitute"]),
            (
                dictgen::InsensitiveStr::Ascii("avaiability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("preced"), &["precede"]),
            (
                dictgen::InsensitiveStr::Ascii("correctionals"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("atrifact"), &["artifact"]),
            (
                dictgen::InsensitiveStr::Ascii("comandeering"),
                &["commandeering"],
            ),
            (dictgen::InsensitiveStr::Ascii("monocrome"), &["monochrome"]),
            (dictgen::InsensitiveStr::Ascii("wherease"), &["whereas"]),
            (dictgen::InsensitiveStr::Ascii("portrayl"), &["portrayal"]),
            (
                dictgen::InsensitiveStr::Ascii("explenations"),
                &["explanations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesssfuly"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("qwesants"), &["croissants"]),
            (
                dictgen::InsensitiveStr::Ascii("cimmanding"),
                &["commanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("rythmic"), &["rhythmic"]),
            (
                dictgen::InsensitiveStr::Ascii("aninteresting"),
                &["uninteresting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succussors"),
                &["successors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definaetly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertieces"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("eletricity"),
                &["electricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("reodering"), &["reordering"]),
            (
                dictgen::InsensitiveStr::Ascii("seccessful"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fascianted"),
                &["fascinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouside"), &["outside"]),
            (
                dictgen::InsensitiveStr::Ascii("calucating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phychiatrists"),
                &["psychiatrists"],
            ),
            (dictgen::InsensitiveStr::Ascii("dunegon"), &["dungeon"]),
            (
                dictgen::InsensitiveStr::Ascii("inexplicabil"),
                &["inexplicably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authendicated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("syndonic"), &["syntonic"]),
            (dictgen::InsensitiveStr::Ascii("clearence"), &["clearance"]),
            (
                dictgen::InsensitiveStr::Ascii("franticaly"),
                &["frantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clearouput"),
                &["clearoutput"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insturment"),
                &["instrument"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terrestial"),
                &["terrestrial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuplodad"),
                &["reupload", "reuploaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("highjacks"), &["hijacks"]),
            (
                dictgen::InsensitiveStr::Ascii("comapators"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipilated"),
                &["manipulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compliacted"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluataions"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferers"), &["referrers"]),
            (
                dictgen::InsensitiveStr::Ascii("compenduim"),
                &["compendium"],
            ),
            (dictgen::InsensitiveStr::Ascii("consting"), &["consisting"]),
            (
                dictgen::InsensitiveStr::Ascii("conotations"),
                &["connotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traverls"),
                &["travels", "traversals"],
            ),
            (dictgen::InsensitiveStr::Ascii("possilby"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("poisitionning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutliation"),
                &["mutilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("overrdies"), &["overrides"]),
            (dictgen::InsensitiveStr::Ascii("intructed"), &["instructed"]),
            (dictgen::InsensitiveStr::Ascii("cyles"), &["cycles"]),
            (
                dictgen::InsensitiveStr::Ascii("vaiant"),
                &["variant", "valiant"],
            ),
            (dictgen::InsensitiveStr::Ascii("somwwhat"), &["somewhat"]),
            (
                dictgen::InsensitiveStr::Ascii("dependenceis"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("enetering"), &["entering"]),
            (
                dictgen::InsensitiveStr::Ascii("installaiton"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("misisng"), &["missing"]),
            (
                dictgen::InsensitiveStr::Ascii("specificaiton"),
                &["specifications", "specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("expemplars"), &["exemplars"]),
            (
                dictgen::InsensitiveStr::Ascii("factuallity"),
                &["factually"],
            ),
            (dictgen::InsensitiveStr::Ascii("relvantly"), &["relevantly"]),
            (dictgen::InsensitiveStr::Ascii("utlise"), &["utilise"]),
            (
                dictgen::InsensitiveStr::Ascii("initialysing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrsepondence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("levellin"), &["levelling"]),
            (
                dictgen::InsensitiveStr::Ascii("anotations"),
                &["annotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encorporates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announements"),
                &["announcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("browing"), &["browsing"]),
            (dictgen::InsensitiveStr::Ascii("chisle"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("transgendred"),
                &["transgendered"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuetherst"), &["furthest"]),
            (
                dictgen::InsensitiveStr::Ascii("decrasing"),
                &["decreasing", "deceasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("rmeoves"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("canges"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("vigilence"), &["vigilance"]),
            (
                dictgen::InsensitiveStr::Ascii("priorisation"),
                &["prioritisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("challene"), &["challenge"]),
            (dictgen::InsensitiveStr::Ascii("multiplpy"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("counterpar"),
                &["counterpart"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabyl"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("convension"),
                &["convention", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("insetead"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("prosocuted"),
                &["prosecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maliciusly"),
                &["maliciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("messiach"), &["messiah"]),
            (dictgen::InsensitiveStr::Ascii("reasoable"), &["reasonable"]),
            (
                dictgen::InsensitiveStr::Ascii("origiginally"),
                &["originally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posisitions"),
                &["positions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimises"), &["minimises"]),
            (
                dictgen::InsensitiveStr::Ascii("neurologial"),
                &["neurological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplicaiton"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("insepect"), &["inspect"]),
            (
                dictgen::InsensitiveStr::Ascii("quiestioning"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disingenuos"),
                &["disingenuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derogotory"),
                &["derogatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sumarizing"),
                &["summarizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("surprisin"), &["surprising"]),
            (
                dictgen::InsensitiveStr::Ascii("resolutionary"),
                &["revolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevtiable"),
                &["inevitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alhabetically"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("diference"), &["difference"]),
            (dictgen::InsensitiveStr::Ascii("precendent"), &["precedent"]),
            (dictgen::InsensitiveStr::Ascii("integrarlo"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("evovling"), &["evolving"]),
            (dictgen::InsensitiveStr::Ascii("suppurts"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("saven"), &["save"]),
            (
                dictgen::InsensitiveStr::Ascii("multidimensinal"),
                &["multidimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customsied"),
                &["customised"],
            ),
            (dictgen::InsensitiveStr::Ascii("looseley"), &["loosely"]),
            (dictgen::InsensitiveStr::Ascii("demonisin"), &["demonising"]),
            (
                dictgen::InsensitiveStr::Ascii("appreteated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("outrageos"), &["outrageous"]),
            (dictgen::InsensitiveStr::Ascii("valiates"), &["validates"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessfullly"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferentiate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegatiotiations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosessing"),
                &["processing", "professing", "possessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("addiiton"), &["addition"]),
            (dictgen::InsensitiveStr::Ascii("atributed"), &["attributed"]),
            (
                dictgen::InsensitiveStr::Ascii("catagorie"),
                &["category", "categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("algohm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("polarizin"), &["polarizing"]),
            (dictgen::InsensitiveStr::Ascii("analysiers"), &["analysers"]),
            (
                dictgen::InsensitiveStr::Ascii("imporbable"),
                &["improbable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asutralian"),
                &["australian"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomteric"), &["geometric"]),
            (
                dictgen::InsensitiveStr::Ascii("intitiations"),
                &["initiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("desingers"), &["designers"]),
            (
                dictgen::InsensitiveStr::Ascii("charactoristics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("softwre"), &["software"]),
            (
                dictgen::InsensitiveStr::Ascii("regulatin"),
                &["regulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("wereas"), &["whereas"]),
            (dictgen::InsensitiveStr::Ascii("defult"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("reevaulate"),
                &["reevaluate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplicitly"),
                &["simplicity", "implicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipyling"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("arlignton"), &["arlington"]),
            (dictgen::InsensitiveStr::Ascii("manageing"), &["managing"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensionailities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notmalized"),
                &["normalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("tto"), &["to"]),
            (
                dictgen::InsensitiveStr::Ascii("overwelming"),
                &["overwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("staurated"), &["saturated"]),
            (
                dictgen::InsensitiveStr::Ascii("decalaration"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproducting"),
                &["reproduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autocmplete"),
                &["autocomplete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("messges"),
                &["messages", "messes"],
            ),
            (dictgen::InsensitiveStr::Ascii("edoema"), &["edema"]),
            (
                dictgen::InsensitiveStr::Ascii("deviceremoveable"),
                &["deviceremovable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexperctedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressers"),
                &["progresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("tarbal"), &["tarball"]),
            (dictgen::InsensitiveStr::Ascii("sceenes"), &["scenes"]),
            (dictgen::InsensitiveStr::Ascii("boudning"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("wordd"), &["world", "word"]),
            (
                dictgen::InsensitiveStr::Ascii("presonalized"),
                &["personalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bookmakr"), &["bookmark"]),
            (
                dictgen::InsensitiveStr::Ascii("figurativly"),
                &["figuratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("amiguous"), &["ambiguous"]),
            (dictgen::InsensitiveStr::Ascii("certaionly"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("ntegrity"), &["integrity"]),
            (dictgen::InsensitiveStr::Ascii("availavble"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("installater"),
                &["installer"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinnical"), &["cynical"]),
            (dictgen::InsensitiveStr::Ascii("examnples"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("faiure"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("upadter"), &["updater"]),
            (dictgen::InsensitiveStr::Ascii("oppossed"), &["opposed"]),
            (dictgen::InsensitiveStr::Ascii("amibguous"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("collding"),
                &["colliding", "colluding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nazionalists"),
                &["nationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifiy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("dimnension"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("subarmines"),
                &["submarines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadvertant"),
                &["inadvertent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteegration"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("tecnics"), &["techniques"]),
            (dictgen::InsensitiveStr::Ascii("prefrence"), &["preference"]),
            (
                dictgen::InsensitiveStr::Ascii("undertunes"),
                &["undertones"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuypos"), &["typos"]),
            (
                dictgen::InsensitiveStr::Ascii("substractive"),
                &["subtractive"],
            ),
            (dictgen::InsensitiveStr::Ascii("negtive"), &["negative"]),
            (
                dictgen::InsensitiveStr::Ascii("equaivalents"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintented"),
                &["unintended", "unindented"],
            ),
            (dictgen::InsensitiveStr::Ascii("unceasin"), &["unceasing"]),
            (
                dictgen::InsensitiveStr::Ascii("sodu"),
                &["sod", "soda", "sods", "sudo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorization"),
                &["prioritization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("floading"),
                &["floating", "flooding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulaters"),
                &["simulators", "simulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("composet"), &["compost"]),
            (dictgen::InsensitiveStr::Ascii("settlin"), &["settling"]),
            (dictgen::InsensitiveStr::Ascii("exhasut"), &["exhaust"]),
            (
                dictgen::InsensitiveStr::Ascii("realiztions"),
                &["realizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determintes"),
                &["determines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evelope"),
                &["envelope", "envelop"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intecepted"),
                &["intercepted"],
            ),
            (dictgen::InsensitiveStr::Ascii("ascendend"), &["ascended"]),
            (dictgen::InsensitiveStr::Ascii("liekely"), &["likely"]),
            (dictgen::InsensitiveStr::Ascii("odasee"), &["odyssey"]),
            (dictgen::InsensitiveStr::Ascii("imbracer"), &["embracer"]),
            (
                dictgen::InsensitiveStr::Ascii("confuguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("adderted"), &["asserted"]),
            (
                dictgen::InsensitiveStr::Ascii("hotizontally"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmit"), &["commit"]),
            (
                dictgen::InsensitiveStr::Ascii("exploitaion"),
                &["exploitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("verticaly"), &["vertically"]),
            (dictgen::InsensitiveStr::Ascii("itterated"), &["iterated"]),
            (
                dictgen::InsensitiveStr::Ascii("mathematicks"),
                &["mathematics"],
            ),
            (dictgen::InsensitiveStr::Ascii("selifes"), &["selfies"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnarable"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terrirories"),
                &["territories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartises"),
                &["standardises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philisophy"),
                &["philosophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("materiasl"),
                &["materials", "material"],
            ),
            (dictgen::InsensitiveStr::Ascii("tthanks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("parlament"), &["parliament"]),
            (dictgen::InsensitiveStr::Ascii("suspened"), &["suspend"]),
            (dictgen::InsensitiveStr::Ascii("compilier"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("glithces"), &["glitches"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterises"),
                &["parameterises"],
            ),
            (dictgen::InsensitiveStr::Ascii("paniic"), &["panic"]),
            (dictgen::InsensitiveStr::Ascii("calller"), &["caller"]),
            (
                dictgen::InsensitiveStr::Ascii("indefinitiley"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documetations"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architcturally"),
                &["architecturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swithcover"),
                &["switchover"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succcessul"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterbating"),
                &["masturbating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colourpsaces"),
                &["colourspaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdencente"),
                &["dependence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downsteamer"),
                &["downstreamer"],
            ),
            (dictgen::InsensitiveStr::Ascii("viewe"), &["view", "viewed"]),
            (dictgen::InsensitiveStr::Ascii("licesnsed"), &["licensed"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibation"),
                &["prohibition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pring"),
                &["print", "bring", "ping", "spring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("haders"),
                &["headers", "shaders", "haters"],
            ),
            (dictgen::InsensitiveStr::Ascii("muliplies"), &["multiplies"]),
            (
                dictgen::InsensitiveStr::Ascii("intersperesed"),
                &["interspersed"],
            ),
            (dictgen::InsensitiveStr::Ascii("guardiands"), &["guardians"]),
            (dictgen::InsensitiveStr::Ascii("popolated"), &["populated"]),
            (
                dictgen::InsensitiveStr::Ascii("recreationnal"),
                &["recreational"],
            ),
            (dictgen::InsensitiveStr::Ascii("sonehow"), &["somehow"]),
            (dictgen::InsensitiveStr::Ascii("remoiving"), &["removing"]),
            (dictgen::InsensitiveStr::Ascii("accoount"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("laucnhing"), &["launching"]),
            (
                dictgen::InsensitiveStr::Ascii("breackpoint"),
                &["breakpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("downloadas"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("delpoys"), &["deploys"]),
            (
                dictgen::InsensitiveStr::Ascii("carser"),
                &["parser", "coarser", "cursor"],
            ),
            (dictgen::InsensitiveStr::Ascii("wondring"), &["wondering"]),
            (
                dictgen::InsensitiveStr::Ascii("massachusettians"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("quadrantic"), &["quadratic"]),
            (
                dictgen::InsensitiveStr::Ascii("consctructs"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("shleter"), &["shelter"]),
            (dictgen::InsensitiveStr::Ascii("cutsceens"), &["cutscenes"]),
            (dictgen::InsensitiveStr::Ascii("releif"), &["relief"]),
            (
                dictgen::InsensitiveStr::Ascii("capicitors"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitually"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("deselcted"), &["deselected"]),
            (dictgen::InsensitiveStr::Ascii("comined"), &["combined"]),
            (dictgen::InsensitiveStr::Ascii("falsks"), &["flasks"]),
            (
                dictgen::InsensitiveStr::Ascii("experimations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("indusrty"), &["industry"]),
            (dictgen::InsensitiveStr::Ascii("beautifys"), &["beautifies"]),
            (
                dictgen::InsensitiveStr::Ascii("indespensible"),
                &["indispensable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boostraping"),
                &["bootstrapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direrctories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undoubtadly"),
                &["undoubtedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilating"),
                &["compiling", "complicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaegs"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("redifined"), &["redefined"]),
            (
                dictgen::InsensitiveStr::Ascii("sinificant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("detemine"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("targtets"), &["targets"]),
            (
                dictgen::InsensitiveStr::Ascii("depenencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertigro"), &["vertigo"]),
            (
                dictgen::InsensitiveStr::Ascii("individals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactureed"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutatiohn"), &["mutation"]),
            (dictgen::InsensitiveStr::Ascii("recreatief"), &["recreate"]),
            (
                dictgen::InsensitiveStr::Ascii("headhpones"),
                &["headphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("vangurad"), &["vanguard"]),
            (
                dictgen::InsensitiveStr::Ascii("trianglulate"),
                &["triangulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ejaculatie"), &["ejaculate"]),
            (
                dictgen::InsensitiveStr::Ascii("continuato"),
                &["continuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatilibility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atrtributes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confingurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("situtation"), &["situation"]),
            (
                dictgen::InsensitiveStr::Ascii("transacional"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("transpot"), &["transport"]),
            (dictgen::InsensitiveStr::Ascii("constucts"), &["constructs"]),
            (
                dictgen::InsensitiveStr::Ascii("approriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afganhistan"),
                &["afghanistan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foundaiton"),
                &["foundations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consentrate"),
                &["concentrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynmic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("verbage"), &["verbiage"]),
            (dictgen::InsensitiveStr::Ascii("miminaly"), &["minimally"]),
            (
                dictgen::InsensitiveStr::Ascii("tourch"),
                &["torch", "touch"],
            ),
            (dictgen::InsensitiveStr::Ascii("itms"), &["items"]),
            (dictgen::InsensitiveStr::Ascii("fillement"), &["filament"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpierenced"),
                &["inexperienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritpor"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asssertions"),
                &["assertions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbreakbale"),
                &["unbreakable"],
            ),
            (dictgen::InsensitiveStr::Ascii("magnitiude"), &["magnitude"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillied"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("acessor"), &["accessor"]),
            (
                dictgen::InsensitiveStr::Ascii("foreginers"),
                &["foreigners"],
            ),
            (dictgen::InsensitiveStr::Ascii("abundence"), &["abundance"]),
            (
                dictgen::InsensitiveStr::Ascii("particuarly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("circlular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("ine"),
                &[
                    "one", "in", "dine", "fine", "line", "mine", "nine", "pine", "sine", "tine",
                    "vine", "wine",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("alloccate"), &["allocate"]),
            (dictgen::InsensitiveStr::Ascii("correcteed"), &["corrected"]),
            (dictgen::InsensitiveStr::Ascii("excecpt"), &["except"]),
            (
                dictgen::InsensitiveStr::Ascii("mutilplayer"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dermetologist"),
                &["dermatologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("packes"),
                &["packages", "packed", "packs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfiers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hestiate"), &["hesitate"]),
            (
                dictgen::InsensitiveStr::Ascii("accomplissments"),
                &["accomplishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefeid"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("lexographical"),
                &["lexicographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompitent"),
                &["incompetent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disagremeent"),
                &["disagreements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doulbelift"),
                &["doublelift"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplifes"), &["simplifies"]),
            (dictgen::InsensitiveStr::Ascii("chizzelled"), &["chiselled"]),
            (dictgen::InsensitiveStr::Ascii("notidies"), &["notifies"]),
            (dictgen::InsensitiveStr::Ascii("follews"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("coinsidental"),
                &["coincidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegosiation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("colleauge"), &["colleague"]),
            (dictgen::InsensitiveStr::Ascii("obtrain"), &["obtain"]),
            (dictgen::InsensitiveStr::Ascii("componoent"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("respesct"), &["respects"]),
            (dictgen::InsensitiveStr::Ascii("compleates"), &["completes"]),
            (dictgen::InsensitiveStr::Ascii("faulures"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("cavlary"), &["cavalry"]),
            (
                dictgen::InsensitiveStr::Ascii("curricullum"),
                &["curriculum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chechkerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complicarte"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authrorisation"),
                &["authorisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagerating"),
                &["exaggerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoleted"), &["obsoleted"]),
            (
                dictgen::InsensitiveStr::Ascii("represenational"),
                &["representational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nuerological"),
                &["neurological"],
            ),
            (dictgen::InsensitiveStr::Ascii("visble"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("idendifier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("copeing"), &["copying"]),
            (
                dictgen::InsensitiveStr::Ascii("antisosial"),
                &["antisocial"],
            ),
            (dictgen::InsensitiveStr::Ascii("balanse"), &["balances"]),
            (dictgen::InsensitiveStr::Ascii("juadaism"), &["judaism"]),
            (dictgen::InsensitiveStr::Ascii("substitte"), &["substitute"]),
            (dictgen::InsensitiveStr::Ascii("refrehsed"), &["refreshed"]),
            (dictgen::InsensitiveStr::Ascii("politey"), &["politely"]),
            (
                dictgen::InsensitiveStr::Ascii("transtition"),
                &["transition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manangement"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("streamd"), &["streamed"]),
            (
                dictgen::InsensitiveStr::Ascii("cotrolling"),
                &["controlling"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuld"), &["result"]),
            (dictgen::InsensitiveStr::Ascii("correect"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("croshairs"), &["crosshairs"]),
            (
                dictgen::InsensitiveStr::Ascii("regsitering"),
                &["registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("excliuded"), &["excluded"]),
            (dictgen::InsensitiveStr::Ascii("compain"), &["complain"]),
            (dictgen::InsensitiveStr::Ascii("traslate"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("floaring"),
                &["floating", "flooring"],
            ),
            (dictgen::InsensitiveStr::Ascii("receved"), &["received"]),
            (
                dictgen::InsensitiveStr::Ascii("repbulicans"),
                &["republicans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continuarlo"),
                &["continual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indadvertently"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incedentally"),
                &["incidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("thourghly"), &["thoroughly"]),
            (dictgen::InsensitiveStr::Ascii("conetented"), &["contented"]),
            (dictgen::InsensitiveStr::Ascii("closses"), &["closes"]),
            (
                dictgen::InsensitiveStr::Ascii("interractively"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diasabilities"),
                &["disabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("happnes"),
                &["happens", "happiness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otpimisers"),
                &["optimisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("descibe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("abodminal"), &["abdominal"]),
            (dictgen::InsensitiveStr::Ascii("chiziles"), &["chisels"]),
            (
                dictgen::InsensitiveStr::Ascii("presearving"),
                &["preserving"],
            ),
            (dictgen::InsensitiveStr::Ascii("progagate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("coumn"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("liscencing"), &["licencing"]),
            (dictgen::InsensitiveStr::Ascii("mulitpath"), &["multipath"]),
            (
                dictgen::InsensitiveStr::Ascii("differeny"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initate"),
                &["initiate", "imitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("surfae"), &["surface"]),
            (dictgen::InsensitiveStr::Ascii("scrpiting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("perferction"),
                &["perfection"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimising"), &["optimising"]),
            (dictgen::InsensitiveStr::Ascii("inlcude"), &["include"]),
            (dictgen::InsensitiveStr::Ascii("possibile"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("photograhic"),
                &["photographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("agani"), &["again"]),
            (dictgen::InsensitiveStr::Ascii("assemblare"), &["assemble"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstansing"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciritc"), &["critic"]),
            (dictgen::InsensitiveStr::Ascii("eletron"), &["electron"]),
            (dictgen::InsensitiveStr::Ascii("belye"), &["belie"]),
            (dictgen::InsensitiveStr::Ascii("refershed"), &["refreshed"]),
            (dictgen::InsensitiveStr::Ascii("exacty"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("garrsion"), &["garrison"]),
            (dictgen::InsensitiveStr::Ascii("eceptions"), &["exceptions"]),
            (
                dictgen::InsensitiveStr::Ascii("beanching"),
                &["branching", "benching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("phyiscs"), &["physics"]),
            (
                dictgen::InsensitiveStr::Ascii("authethicator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbiddent"), &["forbidden"]),
            (dictgen::InsensitiveStr::Ascii("cotnainer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("cahnged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("cerifies"),
                &["certifies", "verifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("abondoned"), &["abandoned"]),
            (dictgen::InsensitiveStr::Ascii("aritfact"), &["artifact"]),
            (
                dictgen::InsensitiveStr::Ascii("corporatoin"),
                &["corporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("paritions"), &["partitions"]),
            (dictgen::InsensitiveStr::Ascii("arives"), &["arrives"]),
            (
                dictgen::InsensitiveStr::Ascii("lightbulps"),
                &["lightbulbs"],
            ),
            (dictgen::InsensitiveStr::Ascii("instaead"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("machanical"),
                &["mechanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutionlized"),
                &["institutionalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknwons"), &["unknowns"]),
            (dictgen::InsensitiveStr::Ascii("abnove"), &["above"]),
            (dictgen::InsensitiveStr::Ascii("malasyia"), &["malaysia"]),
            (
                dictgen::InsensitiveStr::Ascii("controllin"),
                &["controlling"],
            ),
            (dictgen::InsensitiveStr::Ascii("morriosn"), &["morrison"]),
            (
                dictgen::InsensitiveStr::Ascii("enqeueuing"),
                &["enqueueing"],
            ),
            (dictgen::InsensitiveStr::Ascii("acroynms"), &["acronyms"]),
            (
                dictgen::InsensitiveStr::Ascii("gauaranteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarranteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("relatie"), &["relative"]),
            (dictgen::InsensitiveStr::Ascii("corrupterd"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("simulaiton"),
                &["simulations", "simulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintiallized"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("commtted"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("liablity"), &["liability"]),
            (dictgen::InsensitiveStr::Ascii("sumbodule"), &["submodule"]),
            (dictgen::InsensitiveStr::Ascii("conqeur"), &["conquer"]),
            (
                dictgen::InsensitiveStr::Ascii("uneccessary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("graviometric"),
                &["gravimetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapole"),
                &["extrapolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physciatric"),
                &["psychiatric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfigurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("reaname"), &["rename"]),
            (dictgen::InsensitiveStr::Ascii("betrween"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("propecting"),
                &["projecting", "prospecting", "protecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("snpashots"), &["snapshots"]),
            (
                dictgen::InsensitiveStr::Ascii("contaienrs"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("explaning"), &["explaining"]),
            (dictgen::InsensitiveStr::Ascii("subsribes"), &["subscribes"]),
            (dictgen::InsensitiveStr::Ascii("seqeuences"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("freqeuncy"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("checkeers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("spefiy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("bisexaul"), &["bisexual"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuors"),
                &["distributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("fedreated"), &["federated"]),
            (dictgen::InsensitiveStr::Ascii("luminaces"), &["luminances"]),
            (dictgen::InsensitiveStr::Ascii("reomved"), &["removed"]),
            (dictgen::InsensitiveStr::Ascii("mastieres"), &["masteries"]),
            (
                dictgen::InsensitiveStr::Ascii("fullsecreen"),
                &["fullscreen"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaiiable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("successs"),
                &["successes", "success"],
            ),
            (dictgen::InsensitiveStr::Ascii("prevews"), &["previews"]),
            (dictgen::InsensitiveStr::Ascii("publihsed"), &["published"]),
            (
                dictgen::InsensitiveStr::Ascii("invication"),
                &["invocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("consdider"), &["consider"]),
            (
                dictgen::InsensitiveStr::Ascii("symmectrical"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hospitilized"),
                &["hospitalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("traveerses"), &["traverses"]),
            (dictgen::InsensitiveStr::Ascii("concernig"), &["concerning"]),
            (
                dictgen::InsensitiveStr::Ascii("equlibrium"),
                &["equilibrium"],
            ),
            (dictgen::InsensitiveStr::Ascii("priorority"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("arquitectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("throuhout"), &["throughout"]),
            (
                dictgen::InsensitiveStr::Ascii("perpitrator"),
                &["perpetrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fragmenation"),
                &["fragmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regualators"),
                &["regulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mosquiters"),
                &["mosquitoes"],
            ),
            (dictgen::InsensitiveStr::Ascii("canvase"), &["canvas"]),
            (
                dictgen::InsensitiveStr::Ascii("earier"),
                &["earlier", "easier", "eerier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contamporary"),
                &["contemporary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpetation"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhilerate"),
                &["exhilarate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comporator"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("defauld"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("tortillita"), &["tortilla"]),
            (dictgen::InsensitiveStr::Ascii("etension"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiting"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronizing"),
                &["synchronizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("paculier"), &["peculiar"]),
            (dictgen::InsensitiveStr::Ascii("vvalues"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("transparceny"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacraficed"),
                &["sacrificed"],
            ),
            (dictgen::InsensitiveStr::Ascii("overflw"), &["overflow"]),
            (
                dictgen::InsensitiveStr::Ascii("cotrollers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("yto"), &["to"]),
            (dictgen::InsensitiveStr::Ascii("visul"), &["visual"]),
            (
                dictgen::InsensitiveStr::Ascii("indide"),
                &["inside", "indeed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neads"),
                &["heads", "kneads", "needs"],
            ),
            (dictgen::InsensitiveStr::Ascii("exampeles"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("downgraddes"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("devlopers"), &["developers"]),
            (
                dictgen::InsensitiveStr::Ascii("progressionists"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autotorium"),
                &["auditorium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bioligical"),
                &["biological"],
            ),
            (dictgen::InsensitiveStr::Ascii("assembed"), &["assembled"]),
            (dictgen::InsensitiveStr::Ascii("cosmeticas"), &["cosmetics"]),
            (dictgen::InsensitiveStr::Ascii("unitl"), &["until"]),
            (dictgen::InsensitiveStr::Ascii("remoove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("porgrammed"),
                &["programmed"],
            ),
            (dictgen::InsensitiveStr::Ascii("secur"), &["secure"]),
            (dictgen::InsensitiveStr::Ascii("exapnsive"), &["expansive"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterises"),
                &["characterises"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloowinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("eclpise"), &["eclipse"]),
            (dictgen::InsensitiveStr::Ascii("failurs"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("hanlers"), &["handlers"]),
            (
                dictgen::InsensitiveStr::Ascii("parttioned"),
                &["partitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("vakue"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("negotates"), &["negotiates"]),
            (
                dictgen::InsensitiveStr::Ascii("horicontal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaating"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("prescritpion"),
                &["prescriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postitional"),
                &["positional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apppendices"),
                &["appendices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polymorpic"),
                &["polymorphic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classificiation"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("tangnet"), &["tangent"]),
            (dictgen::InsensitiveStr::Ascii("havr"), &["have"]),
            (
                dictgen::InsensitiveStr::Ascii("investirat"),
                &["investigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conssitent"),
                &["consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("obtaions"), &["obtains"]),
            (dictgen::InsensitiveStr::Ascii("largerly"), &["largely"]),
            (dictgen::InsensitiveStr::Ascii("acceotor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("argumeng"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("unistalling"),
                &["uninstalling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("followd"),
                &["followed", "follows", "follow"],
            ),
            (dictgen::InsensitiveStr::Ascii("similairy"), &["similarity"]),
            (dictgen::InsensitiveStr::Ascii("apyoons"), &["oppugns"]),
            (dictgen::InsensitiveStr::Ascii("enfore"), &["enforce"]),
            (
                dictgen::InsensitiveStr::Ascii("unsatisifiable"),
                &["unsatisfiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("physisict"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("commector"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("conditiner"),
                &["conditioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("netwroking"),
                &["networking"],
            ),
            (dictgen::InsensitiveStr::Ascii("hexigon"), &["hexagon"]),
            (
                dictgen::InsensitiveStr::Ascii("enumerater"),
                &["enumerator", "enumerated", "enumerates", "enumerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("westerse"), &["westerners"]),
            (dictgen::InsensitiveStr::Ascii("inteval"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("passwor"), &["password"]),
            (
                dictgen::InsensitiveStr::Ascii("experimdnts"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("diminsh"), &["diminish"]),
            (dictgen::InsensitiveStr::Ascii("cintaner"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("appeding"), &["appending"]),
            (dictgen::InsensitiveStr::Ascii("notaly"), &["notably"]),
            (dictgen::InsensitiveStr::Ascii("hthe"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("clatified"), &["clarified"]),
            (dictgen::InsensitiveStr::Ascii("changess"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("datatpye"), &["datatype"]),
            (
                dictgen::InsensitiveStr::Ascii("sepcifications"),
                &["specification", "specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provintial"),
                &["provincial"],
            ),
            (dictgen::InsensitiveStr::Ascii("commandis"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("accessable"),
                &["accessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emergancies"),
                &["emergencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossable"),
                &["impossible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survallience"),
                &["surveillance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ramdomization"),
                &["randomization"],
            ),
            (dictgen::InsensitiveStr::Ascii("polarlised"), &["polarised"]),
            (dictgen::InsensitiveStr::Ascii("disharge"), &["discharge"]),
            (dictgen::InsensitiveStr::Ascii("febraury"), &["february"]),
            (dictgen::InsensitiveStr::Ascii("preffixed"), &["prefixed"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubutors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquantiances"),
                &["acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("worldveiw"), &["worldview"]),
            (
                dictgen::InsensitiveStr::Ascii("catstrophe"),
                &["catastrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formes"),
                &["forms", "formed"],
            ),
            (dictgen::InsensitiveStr::Ascii("adatpive"), &["adaptive"]),
            (dictgen::InsensitiveStr::Ascii("paritally"), &["partially"]),
            (
                dictgen::InsensitiveStr::Ascii("functionnalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpetuationg"),
                &["perpetuating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theshholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropropiate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("variabls"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("formost"), &["foremost"]),
            (dictgen::InsensitiveStr::Ascii("deciption"), &["depiction"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregrates"),
                &["aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimietrs"),
                &["delimiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruthces"), &["crutches"]),
            (dictgen::InsensitiveStr::Ascii("delevops"), &["develops"]),
            (dictgen::InsensitiveStr::Ascii("directrly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("crusies"), &["cruises"]),
            (
                dictgen::InsensitiveStr::Ascii("algolithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fragmeneted"),
                &["fragmented"],
            ),
            (dictgen::InsensitiveStr::Ascii("liberatin"), &["liberation"]),
            (dictgen::InsensitiveStr::Ascii("flaoting"), &["floating"]),
            (dictgen::InsensitiveStr::Ascii("edeycats"), &["etiquettes"]),
            (
                dictgen::InsensitiveStr::Ascii("performancepcs"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proclomation"),
                &["proclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("adivsers"), &["advisers"]),
            (dictgen::InsensitiveStr::Ascii("annoates"), &["annotates"]),
            (
                dictgen::InsensitiveStr::Ascii("exceuctables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("licensin"), &["licensing"]),
            (dictgen::InsensitiveStr::Ascii("pouint"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonation"),
                &["automation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifyng"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("esitmate"), &["estimate"]),
            (
                dictgen::InsensitiveStr::Ascii("speciyfing"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("veriosn"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("historisch"), &["historic"]),
            (
                dictgen::InsensitiveStr::Ascii("progerssion"),
                &["progressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("countinue"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("extited"),
                &["excited", "exited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiatorial"),
                &["combinatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redefinitons"),
                &["redefinitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("desktopbsd"), &["desktops"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsulatin"),
                &["encapsulating", "encapsulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surgestions"),
                &["suggestions"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfying"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("asymetries"),
                &["asymmetries"],
            ),
            (dictgen::InsensitiveStr::Ascii("surrouns"), &["surrounds"]),
            (dictgen::InsensitiveStr::Ascii("intgrates"), &["integrates"]),
            (
                dictgen::InsensitiveStr::Ascii("translyvania"),
                &["transylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("standlone"), &["standalone"]),
            (dictgen::InsensitiveStr::Ascii("adjument"), &["adjustment"]),
            (
                dictgen::InsensitiveStr::Ascii("capitolism"),
                &["capitalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("languaqes"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("apserger"), &["asperger"]),
            (dictgen::InsensitiveStr::Ascii("cretate"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("cloesd"), &["closed"]),
            (dictgen::InsensitiveStr::Ascii("appoach"), &["approach"]),
            (
                dictgen::InsensitiveStr::Ascii("transpacencies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("officail"), &["official"]),
            (dictgen::InsensitiveStr::Ascii("inproove"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("copmetitors"),
                &["competitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("disagress"), &["disagrees"]),
            (dictgen::InsensitiveStr::Ascii("repsponse"), &["response"]),
            (
                dictgen::InsensitiveStr::Ascii("disciplanary"),
                &["disciplinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beeforehand"),
                &["beforehand"],
            ),
            (dictgen::InsensitiveStr::Ascii("negotitor"), &["negotiator"]),
            (dictgen::InsensitiveStr::Ascii("fertily"), &["fertility"]),
            (
                dictgen::InsensitiveStr::Ascii("comprossor"),
                &["compressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("caeful"), &["careful"]),
            (dictgen::InsensitiveStr::Ascii("beleiver"), &["believer"]),
            (
                dictgen::InsensitiveStr::Ascii("cusstomizable"),
                &["customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socilaized"),
                &["socialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("srouce"), &["source"]),
            (dictgen::InsensitiveStr::Ascii("eanable"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("concentate"),
                &["concentrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silohuette"),
                &["silhouette"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mainifested"),
                &["manifested"],
            ),
            (dictgen::InsensitiveStr::Ascii("destrcut"), &["destruct"]),
            (dictgen::InsensitiveStr::Ascii("proposse"), &["proposes"]),
            (dictgen::InsensitiveStr::Ascii("ultering"), &["altering"]),
            (
                dictgen::InsensitiveStr::Ascii("caclulating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projektion"),
                &["projection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempertaure"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("targteed"), &["targeted"]),
            (
                dictgen::InsensitiveStr::Ascii("probabililty"),
                &["probability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clasifying"),
                &["classifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("assersing"), &["asserting"]),
            (dictgen::InsensitiveStr::Ascii("nevers"), &["never"]),
            (
                dictgen::InsensitiveStr::Ascii("unbeleivable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("yesteday"), &["yesterday"]),
            (dictgen::InsensitiveStr::Ascii("objetcs"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("docuemtn"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("usesfull"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("occured"), &["occurred"]),
            (dictgen::InsensitiveStr::Ascii("viwer"), &["viewer"]),
            (dictgen::InsensitiveStr::Ascii("unsernames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("intitialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("creasoat"), &["creosote"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnetations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infiltrade"),
                &["infiltrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuratio"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libreoficekit"),
                &["libreofficekit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deterministincally"),
                &["deterministically"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhaspody"), &["rhapsody"]),
            (
                dictgen::InsensitiveStr::Ascii("experimments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rendere"),
                &["renderer", "render"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mezmorises"),
                &["mesmerises"],
            ),
            (dictgen::InsensitiveStr::Ascii("proffeses"), &["professes"]),
            (
                dictgen::InsensitiveStr::Ascii("specifyinhg"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executiner"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("imgage"), &["image"]),
            (dictgen::InsensitiveStr::Ascii("stamentent"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("itsel"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("nawseous"), &["nauseous"]),
            (dictgen::InsensitiveStr::Ascii("quiestion"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("reoports"), &["reports"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropraitely"),
                &["inappropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("invovled"), &["involved"]),
            (
                dictgen::InsensitiveStr::Ascii("environmets"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encryptiong"),
                &["encryption", "encrypting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexperesion"),
                &["subexpression"],
            ),
            (dictgen::InsensitiveStr::Ascii("opponenets"), &["opponent"]),
            (dictgen::InsensitiveStr::Ascii("messenges"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("inavlis"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("attribuet"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("inidcators"),
                &["indicators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifiestos"),
                &["manifestos"],
            ),
            (dictgen::InsensitiveStr::Ascii("stubbon"), &["stubborn"]),
            (dictgen::InsensitiveStr::Ascii("androides"), &["androids"]),
            (dictgen::InsensitiveStr::Ascii("recrusive"), &["recursive"]),
            (
                dictgen::InsensitiveStr::Ascii("revlalidation"),
                &["revalidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hittin"), &["hitting"]),
            (dictgen::InsensitiveStr::Ascii("ingegral"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("panethon"), &["pantheon"]),
            (
                dictgen::InsensitiveStr::Ascii("envionrment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("leegun"), &["legion"]),
            (dictgen::InsensitiveStr::Ascii("addrsses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("liuetenant"),
                &["lieutenant"],
            ),
            (dictgen::InsensitiveStr::Ascii("kimeric"), &["chimeric"]),
            (
                dictgen::InsensitiveStr::Ascii("witespaces"),
                &["whitespaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decendents"),
                &["descendents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interupting"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumualtors"),
                &["accumulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryifing"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoemanse"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("porfiled"), &["profiled"]),
            (
                dictgen::InsensitiveStr::Ascii("wraparround"),
                &["wraparound"],
            ),
            (dictgen::InsensitiveStr::Ascii("sercives"), &["services"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnurability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacakge"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("accommidating"),
                &["accommodating"],
            ),
            (dictgen::InsensitiveStr::Ascii("usefuly"), &["usefully"]),
            (dictgen::InsensitiveStr::Ascii("sistemist"), &["systemist"]),
            (dictgen::InsensitiveStr::Ascii("avioding"), &["avoiding"]),
            (
                dictgen::InsensitiveStr::Ascii("compulstion"),
                &["compulsion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asethetically"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalcualtion"),
                &["recalculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplyer"),
                &["multiple", "multiplier", "multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("catalcysm"), &["cataclysm"]),
            (dictgen::InsensitiveStr::Ascii("errons"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("enterprize"),
                &["enterprise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilized"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("languanages"),
                &["languages"],
            ),
            (dictgen::InsensitiveStr::Ascii("carnvial"), &["carnival"]),
            (dictgen::InsensitiveStr::Ascii("vaildator"), &["validator"]),
            (dictgen::InsensitiveStr::Ascii("execue"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("deathmath"), &["deathmatch"]),
            (dictgen::InsensitiveStr::Ascii("fomating"), &["formatting"]),
            (dictgen::InsensitiveStr::Ascii("sekects"), &["selects"]),
            (dictgen::InsensitiveStr::Ascii("notciable"), &["noticeable"]),
            (
                dictgen::InsensitiveStr::Ascii("trangulated"),
                &["triangulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installatin"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("inabled"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("sprech"), &["speech"]),
            (
                dictgen::InsensitiveStr::Ascii("hortizontally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefinetely"),
                &["indefinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("formates"), &["formats"]),
            (dictgen::InsensitiveStr::Ascii("drwawings"), &["drawings"]),
            (dictgen::InsensitiveStr::Ascii("adivser"), &["adviser"]),
            (
                dictgen::InsensitiveStr::Ascii("experiminting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provocatie"),
                &["provocative"],
            ),
            (dictgen::InsensitiveStr::Ascii("curiculum"), &["curriculum"]),
            (
                dictgen::InsensitiveStr::Ascii("envionments"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconventional"),
                &["unconventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dividors"),
                &["dividers", "divisors"],
            ),
            (dictgen::InsensitiveStr::Ascii("attributre"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("upadte"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("congifurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosumer"), &["consumer"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficific"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("subshystems"),
                &["subsystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("fotget"), &["forget"]),
            (dictgen::InsensitiveStr::Ascii("adrersser"), &["addresser"]),
            (
                dictgen::InsensitiveStr::Ascii("ownenership"),
                &["ownership"],
            ),
            (dictgen::InsensitiveStr::Ascii("columnes"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("subsitutions"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("technial"), &["technical"]),
            (dictgen::InsensitiveStr::Ascii("intriuge"), &["intrigue"]),
            (dictgen::InsensitiveStr::Ascii("comarator"), &["comparator"]),
            (
                dictgen::InsensitiveStr::Ascii("documetning"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("exauhsted"), &["exhausted"]),
            (dictgen::InsensitiveStr::Ascii("cylcical"), &["cyclical"]),
            (dictgen::InsensitiveStr::Ascii("distnace"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("upgade"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("ambadextrous"),
                &["ambidextrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configutring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("protestat"), &["protestant"]),
            (dictgen::InsensitiveStr::Ascii("impproves"), &["improves"]),
            (
                dictgen::InsensitiveStr::Ascii("availalbility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsisently"),
                &["inconsistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloekss"), &["blokes"]),
            (dictgen::InsensitiveStr::Ascii("covereages"), &["coverages"]),
            (
                dictgen::InsensitiveStr::Ascii("chages"),
                &["changes", "charges"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeate"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("ocilate"), &["oscillate"]),
            (dictgen::InsensitiveStr::Ascii("breakdows"), &["breakdowns"]),
            (dictgen::InsensitiveStr::Ascii("claymer"), &["claymore"]),
            (
                dictgen::InsensitiveStr::Ascii("dispooables"),
                &["disposables"],
            ),
            (dictgen::InsensitiveStr::Ascii("keybord"), &["keyboard"]),
            (
                dictgen::InsensitiveStr::Ascii("konservative"),
                &["conservative"],
            ),
            (dictgen::InsensitiveStr::Ascii("confured"), &["configured"]),
            (
                dictgen::InsensitiveStr::Ascii("tesellation"),
                &["tessellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellipsises"), &["ellipsis"]),
            (dictgen::InsensitiveStr::Ascii("portugese"), &["portuguese"]),
            (
                dictgen::InsensitiveStr::Ascii("impotrt"),
                &["import", "imported", "importer"],
            ),
            (dictgen::InsensitiveStr::Ascii("finishd"), &["finished"]),
            (
                dictgen::InsensitiveStr::Ascii("envioment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacutre"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rademption"),
                &["redemption"],
            ),
            (dictgen::InsensitiveStr::Ascii("rivised"), &["revised"]),
            (
                dictgen::InsensitiveStr::Ascii("workaroungs"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("discribes"), &["describes"]),
            (dictgen::InsensitiveStr::Ascii("apdat"), &["adapt"]),
            (dictgen::InsensitiveStr::Ascii("snigles"), &["singles"]),
            (
                dictgen::InsensitiveStr::Ascii("determenism"),
                &["determinism"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfrom"), &["perform"]),
            (
                dictgen::InsensitiveStr::Ascii("parsef"),
                &["parsec", "parsed", "parser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilism"),
                &["capitalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("relaoded"), &["reloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("awknowledging"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantaneoulsy"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electrolyts"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiscling"), &["whistling"]),
            (dictgen::InsensitiveStr::Ascii("juptier"), &["jupiter"]),
            (dictgen::InsensitiveStr::Ascii("guarantie"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("positons"),
                &["positions", "positrons"],
            ),
            (dictgen::InsensitiveStr::Ascii("roduct"), &["product"]),
            (
                dictgen::InsensitiveStr::Ascii("repersents"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innoculated"),
                &["inoculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behavoural"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("detectt"), &["detect"]),
            (dictgen::InsensitiveStr::Ascii("accseptors"), &["acceptors"]),
            (
                dictgen::InsensitiveStr::Ascii("dupliation"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocrocontroller"),
                &["microcontroller"],
            ),
            (dictgen::InsensitiveStr::Ascii("cimpilers"), &["compilers"]),
            (dictgen::InsensitiveStr::Ascii("outherwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("accepector"), &["acceptor"]),
            (
                dictgen::InsensitiveStr::Ascii("shenanigins"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconeccts"),
                &["reconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerablity"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simler"),
                &["simpler", "similar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explotion"),
                &["explosion", "exploration", "exploitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opreational"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("extractred"), &["extracted"]),
            (dictgen::InsensitiveStr::Ascii("origingal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("entery"), &["entry"]),
            (dictgen::InsensitiveStr::Ascii("webstie"), &["website"]),
            (dictgen::InsensitiveStr::Ascii("resoured"), &["resourced"]),
            (dictgen::InsensitiveStr::Ascii("circuvent"), &["circumvent"]),
            (dictgen::InsensitiveStr::Ascii("partital"), &["partial"]),
            (
                dictgen::InsensitiveStr::Ascii("scandaniva"),
                &["scandinavia"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminis"), &["reminisce"]),
            (
                dictgen::InsensitiveStr::Ascii("nautral"),
                &["natural", "neutral"],
            ),
            (dictgen::InsensitiveStr::Ascii("pojnt"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("specifices"),
                &["specifics", "specifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discovere"),
                &["discoveries"],
            ),
            (dictgen::InsensitiveStr::Ascii("couseling"), &["counseling"]),
            (dictgen::InsensitiveStr::Ascii("ehternet"), &["ethernet"]),
            (dictgen::InsensitiveStr::Ascii("lifecicle"), &["lifecycle"]),
            (
                dictgen::InsensitiveStr::Ascii("initliased"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independets"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloce"), &["alloc"]),
            (dictgen::InsensitiveStr::Ascii("midle"), &["middle"]),
            (
                dictgen::InsensitiveStr::Ascii("deinitions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("avarage"), &["average"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiazer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapplining"),
                &["disciplining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponention"),
                &["exponentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiantences"),
                &["acquaintances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calucations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("termostat"), &["thermostat"]),
            (dictgen::InsensitiveStr::Ascii("didi"), &["did"]),
            (dictgen::InsensitiveStr::Ascii("pebbleos"), &["pebbles"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapponted"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("cannoical"), &["canonical"]),
            (dictgen::InsensitiveStr::Ascii("extremelly"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("hyunday"), &["hyundai"]),
            (
                dictgen::InsensitiveStr::Ascii("inevitabily"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheived"), &["achieved"]),
            (
                dictgen::InsensitiveStr::Ascii("recommdended"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("engrames"), &["engrams"]),
            (
                dictgen::InsensitiveStr::Ascii("terminatino"),
                &["termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("rigt"), &["right"]),
            (dictgen::InsensitiveStr::Ascii("elagantly"), &["elegantly"]),
            (dictgen::InsensitiveStr::Ascii("chganged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("divisiones"), &["divisions"]),
            (dictgen::InsensitiveStr::Ascii("sommetimes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("cengter"), &["center"]),
            (
                dictgen::InsensitiveStr::Ascii("componenete"),
                &["component", "components"],
            ),
            (dictgen::InsensitiveStr::Ascii("nly"), &["only"]),
            (dictgen::InsensitiveStr::Ascii("nieghbour"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("perseedes"), &["precedes"]),
            (
                dictgen::InsensitiveStr::Ascii("disscovering"),
                &["discovering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rennovating"),
                &["renovating"],
            ),
            (dictgen::InsensitiveStr::Ascii("engieers"), &["engineers"]),
            (
                dictgen::InsensitiveStr::Ascii("equivlantly"),
                &["equivalently"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnance"), &["chance"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caclulations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("adition"), &["addition"]),
            (
                dictgen::InsensitiveStr::Ascii("unapretiative"),
                &["unappreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthenticate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fascinatinf"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("patckets"), &["packets"]),
            (dictgen::InsensitiveStr::Ascii("bugus"), &["bogus"]),
            (
                dictgen::InsensitiveStr::Ascii("distincitons"),
                &["distinctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("whetther"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("embedd"), &["embed"]),
            (
                dictgen::InsensitiveStr::Ascii("conceeting"),
                &["conceiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("exturd"), &["extrude"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrality"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("eplison"), &["epsilon"]),
            (dictgen::InsensitiveStr::Ascii("perfro"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("spliiter"), &["splitter"]),
            (
                dictgen::InsensitiveStr::Ascii("powerliftng"),
                &["powerlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circuncised"),
                &["circumcised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeateadly"),
                &["repeatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("debounding"),
                &["debouncing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruted"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanners"), &["cleaners"]),
            (dictgen::InsensitiveStr::Ascii("excutable"), &["executable"]),
            (dictgen::InsensitiveStr::Ascii("propeties"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("originaly"), &["originally"]),
            (dictgen::InsensitiveStr::Ascii("friendsi"), &["friendlies"]),
            (dictgen::InsensitiveStr::Ascii("provoding"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("taks"),
                &["task", "tasks", "takes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplisment"),
                &["accomplishment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonnegarive"),
                &["nonnegative"],
            ),
            (dictgen::InsensitiveStr::Ascii("lanuched"), &["launched"]),
            (
                dictgen::InsensitiveStr::Ascii("appologizes"),
                &["apologizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fiding"),
                &["finding", "fidling"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitmise"), &["optimise"]),
            (
                dictgen::InsensitiveStr::Ascii("constrats"),
                &["contrasts", "constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenfy"), &["identify"]),
            (
                dictgen::InsensitiveStr::Ascii("morphlogies"),
                &["morphologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("unbouind"), &["unbound"]),
            (
                dictgen::InsensitiveStr::Ascii("evalueators"),
                &["evaluators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("involement"),
                &["involvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("instaling"), &["installing"]),
            (
                dictgen::InsensitiveStr::Ascii("bodybuidling"),
                &["bodybuilding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nayborhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("eelments"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("restictions"),
                &["restrictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inerrupted"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quesionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connnection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miinimising"),
                &["minimising"],
            ),
            (dictgen::InsensitiveStr::Ascii("refenced"), &["referenced"]),
            (dictgen::InsensitiveStr::Ascii("qoutient"), &["quotient"]),
            (
                dictgen::InsensitiveStr::Ascii("differnencing"),
                &["differencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tanslate"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("devastaing"),
                &["devastating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formualates"),
                &["formulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("attatches"), &["attaches"]),
            (
                dictgen::InsensitiveStr::Ascii("abreviations"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loades"),
                &["loads", "loaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreperneur"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("farwards"), &["forwards"]),
            (dictgen::InsensitiveStr::Ascii("nonsesne"), &["nonsense"]),
            (dictgen::InsensitiveStr::Ascii("curerntly"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("southampon"),
                &["southampton"],
            ),
            (dictgen::InsensitiveStr::Ascii("strated"), &["started"]),
            (dictgen::InsensitiveStr::Ascii("birghtly"), &["brightly"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthorugh"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatpes"), &["datatypes"]),
            (
                dictgen::InsensitiveStr::Ascii("interploatin"),
                &["interpolating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accointing"),
                &["accounting"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitwiedh"), &["bitwidth"]),
            (
                dictgen::InsensitiveStr::Ascii("enlightining"),
                &["enlightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("fuctiones"), &["functioned"]),
            (
                dictgen::InsensitiveStr::Ascii("selecrtions"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("aribter"), &["arbiter"]),
            (dictgen::InsensitiveStr::Ascii("lasanga"), &["lasagna"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistendly"),
                &["inconsistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whiltelisted"),
                &["whitelisted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limter"),
                &["limiter", "linter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contriubtion"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitions"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determing"),
                &["determining", "determine"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbsoe"), &["verbose"]),
            (dictgen::InsensitiveStr::Ascii("stimulat"), &["stimulants"]),
            (dictgen::InsensitiveStr::Ascii("boleen"), &["boolean"]),
            (
                dictgen::InsensitiveStr::Ascii("desallocate"),
                &["deallocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("vrilog"), &["verilog"]),
            (dictgen::InsensitiveStr::Ascii("optinos"), &["options"]),
            (
                dictgen::InsensitiveStr::Ascii("accidently"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("currupt"), &["corrupt"]),
            (dictgen::InsensitiveStr::Ascii("proclamed"), &["proclaimed"]),
            (dictgen::InsensitiveStr::Ascii("numbfer"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("reacord"), &["record"]),
            (dictgen::InsensitiveStr::Ascii("enitrely"), &["entirely"]),
            (dictgen::InsensitiveStr::Ascii("raplacing"), &["replacing"]),
            (
                dictgen::InsensitiveStr::Ascii("indroduction"),
                &["introduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectualisme"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttach"), &["attach"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticde"),
                &["authenticode"],
            ),
            (dictgen::InsensitiveStr::Ascii("logcal"), &["logical"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnporter"),
                &["transporter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvienenced"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survivabiltiy"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skillshoot"),
                &["skillshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regstrations"),
                &["registrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("warrantles"),
                &["warranties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("relevence"), &["relevance"]),
            (
                dictgen::InsensitiveStr::Ascii("creatd"),
                &["created", "create"],
            ),
            (dictgen::InsensitiveStr::Ascii("mismached"), &["mismatched"]),
            (dictgen::InsensitiveStr::Ascii("chisil"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("tipling"),
                &["tripling", "tipping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invaldates"),
                &["invalidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitailise"),
                &["reinitialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trushworthy"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efficently"),
                &["efficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("protocl"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("expeiment"), &["experiment"]),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossibry"),
                &["impossibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibilty"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("assimtote"), &["asymptote"]),
            (dictgen::InsensitiveStr::Ascii("nitch"), &["niche"]),
            (dictgen::InsensitiveStr::Ascii("weeked"), &["weekend"]),
            (
                dictgen::InsensitiveStr::Ascii("straming"),
                &["streaming", "steaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferees"), &["referees"]),
            (dictgen::InsensitiveStr::Ascii("aspriins"), &["aspirins"]),
            (
                dictgen::InsensitiveStr::Ascii("punctiation"),
                &["punctuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfaccion"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("sintak"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("precussions"),
                &["percussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("rrroring"), &["erroring"]),
            (
                dictgen::InsensitiveStr::Ascii("aplhabetically"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevitible"),
                &["inevitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("tanslator"), &["translator"]),
            (dictgen::InsensitiveStr::Ascii("csutomer"), &["customer"]),
            (dictgen::InsensitiveStr::Ascii("alignmenet"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("submittitng"),
                &["submitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("aperutre"), &["aperture"]),
            (dictgen::InsensitiveStr::Ascii("convinved"), &["convinced"]),
            (
                dictgen::InsensitiveStr::Ascii("workpsaces"),
                &["workspaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyizer"), &["analyzer"]),
            (
                dictgen::InsensitiveStr::Ascii("endandering"),
                &["endangering"],
            ),
            (dictgen::InsensitiveStr::Ascii("ununsed"), &["unused"]),
            (dictgen::InsensitiveStr::Ascii("sciprt"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("subscritpiton"),
                &["subscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterstion"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("lossleslly"),
                &["losslessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inqusitior"),
                &["inquisitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandenavia"),
                &["scandinavia"],
            ),
            (dictgen::InsensitiveStr::Ascii("builld"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("explortation"),
                &["exploitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartiuon"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("browswers"), &["browsers"]),
            (
                dictgen::InsensitiveStr::Ascii("distace"),
                &["distance", "distaste"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiem"), &["time", "item"]),
            (dictgen::InsensitiveStr::Ascii("checkox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatiable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cogegen"), &["codegen"]),
            (dictgen::InsensitiveStr::Ascii("finisching"), &["finishing"]),
            (
                dictgen::InsensitiveStr::Ascii("persepctives"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accummulating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerions"),
                &["considerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("stik"), &["stick"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadmic"), &["academic"]),
            (dictgen::InsensitiveStr::Ascii("genisues"), &["geniuses"]),
            (dictgen::InsensitiveStr::Ascii("problaem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("reslves"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("implememntations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjatate"), &["agitate"]),
            (dictgen::InsensitiveStr::Ascii("issuin"), &["issuing"]),
            (
                dictgen::InsensitiveStr::Ascii("differntiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commonded"),
                &["commanded", "commended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrtupion"),
                &["interruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representatief"),
                &["representative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("principielt"),
                &["principle"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctivo"),
                &["correction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrensponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("disaple"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("negosiates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("automatice"), &["automate"]),
            (dictgen::InsensitiveStr::Ascii("fortess"), &["fortress"]),
            (dictgen::InsensitiveStr::Ascii("repsonder"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("barnched"), &["branched"]),
            (dictgen::InsensitiveStr::Ascii("seperatos"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("carridge"),
                &["carriage", "cartridge"],
            ),
            (dictgen::InsensitiveStr::Ascii("descided"), &["decided"]),
            (dictgen::InsensitiveStr::Ascii("disances"), &["distances"]),
            (dictgen::InsensitiveStr::Ascii("getttime"), &["gettime"]),
            (dictgen::InsensitiveStr::Ascii("cussess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("upladers"), &["uploaders"]),
            (dictgen::InsensitiveStr::Ascii("remoce"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("applicatons"),
                &["applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinato"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmets"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("continuted"), &["continued"]),
            (
                dictgen::InsensitiveStr::Ascii("addresable"),
                &["addressable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vibraters"),
                &["vibrators", "vibrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ilness"), &["illness"]),
            (dictgen::InsensitiveStr::Ascii("succeesor"), &["successor"]),
            (dictgen::InsensitiveStr::Ascii("rocording"), &["recording"]),
            (
                dictgen::InsensitiveStr::Ascii("asychronize"),
                &["asynchronize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ovewritable"),
                &["overwritable"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrorker"), &["worker"]),
            (
                dictgen::InsensitiveStr::Ascii("crigneworthy"),
                &["cringeworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("sythesize"), &["synthesize"]),
            (
                dictgen::InsensitiveStr::Ascii("spacialised"),
                &["specialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expaned"),
                &["expand", "expanded", "explained"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirgate"), &["migrate"]),
            (
                dictgen::InsensitiveStr::Ascii("kronicling"),
                &["chronicling"],
            ),
            (dictgen::InsensitiveStr::Ascii("navagates"), &["navigates"]),
            (dictgen::InsensitiveStr::Ascii("maually"), &["manually"]),
            (dictgen::InsensitiveStr::Ascii("cumlative"), &["cumulative"]),
            (dictgen::InsensitiveStr::Ascii("addresess"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("corect"), &["correct"]),
            (
                dictgen::InsensitiveStr::Ascii("iluminations"),
                &["illuminations"],
            ),
            (dictgen::InsensitiveStr::Ascii("whishlist"), &["wishlist"]),
            (
                dictgen::InsensitiveStr::Ascii("highligthed"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertion"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("religeon"), &["religion"]),
            (dictgen::InsensitiveStr::Ascii("evalues"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("apratheid"), &["apartheid"]),
            (
                dictgen::InsensitiveStr::Ascii("disception"),
                &["discretion"],
            ),
            (dictgen::InsensitiveStr::Ascii("siplified"), &["simplified"]),
            (dictgen::InsensitiveStr::Ascii("champage"), &["champagne"]),
            (dictgen::InsensitiveStr::Ascii("synthedic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("cleaups"), &["cleanups"]),
            (dictgen::InsensitiveStr::Ascii("dribbeling"), &["dribbling"]),
            (dictgen::InsensitiveStr::Ascii("choising"), &["choosing"]),
            (dictgen::InsensitiveStr::Ascii("separetes"), &["separates"]),
            (
                dictgen::InsensitiveStr::Ascii("unitilize"),
                &["uninitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resturaunts"),
                &["restaurants"],
            ),
            (dictgen::InsensitiveStr::Ascii("passisve"), &["passives"]),
            (
                dictgen::InsensitiveStr::Ascii("trememdously"),
                &["tremendously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interveiwer"),
                &["interviewer"],
            ),
            (dictgen::InsensitiveStr::Ascii("earpeice"), &["earpiece"]),
            (dictgen::InsensitiveStr::Ascii("partically"), &["partially"]),
            (dictgen::InsensitiveStr::Ascii("expexts"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("klenex"), &["kleenex"]),
            (dictgen::InsensitiveStr::Ascii("mirrorrs"), &["mirrors"]),
            (
                dictgen::InsensitiveStr::Ascii("hexidecimal"),
                &["hexadecimal"],
            ),
            (dictgen::InsensitiveStr::Ascii("ridiculus"), &["ridiculous"]),
            (dictgen::InsensitiveStr::Ascii("contoll"), &["control"]),
            (dictgen::InsensitiveStr::Ascii("haave"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("peson"), &["person"]),
            (dictgen::InsensitiveStr::Ascii("encrcypted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("corosponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderatanding"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resolveable"),
                &["resolvable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenedecy"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("stagnat"), &["stagnant"]),
            (dictgen::InsensitiveStr::Ascii("exension"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("freqient"), &["frequent"]),
            (
                dictgen::InsensitiveStr::Ascii("intestinces"),
                &["intestines"],
            ),
            (dictgen::InsensitiveStr::Ascii("meesaged"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivitiy"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("flechter"), &["fletcher"]),
            (dictgen::InsensitiveStr::Ascii("buik"), &["bulk"]),
            (dictgen::InsensitiveStr::Ascii("leackage"), &["leakage"]),
            (
                dictgen::InsensitiveStr::Ascii("repblicans"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("miliitas"), &["militias"]),
            (dictgen::InsensitiveStr::Ascii("indiacte"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("althogh"), &["although"]),
            (dictgen::InsensitiveStr::Ascii("simmilar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdendance"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("expliots"), &["exploits"]),
            (dictgen::InsensitiveStr::Ascii("hokpins"), &["hopkins"]),
            (
                dictgen::InsensitiveStr::Ascii("continueing"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserialsies"),
                &["deserialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("deciedd"), &["decided"]),
            (dictgen::InsensitiveStr::Ascii("devined"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("follwoed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("bakward"), &["backward"]),
            (dictgen::InsensitiveStr::Ascii("apeals"), &["appeals"]),
            (dictgen::InsensitiveStr::Ascii("devloped"), &["developed"]),
            (dictgen::InsensitiveStr::Ascii("insersect"), &["intersect"]),
            (dictgen::InsensitiveStr::Ascii("asburdly"), &["absurdly"]),
            (
                dictgen::InsensitiveStr::Ascii("appointmnets"),
                &["appointments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalscece"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agorithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("victems"), &["victims"]),
            (
                dictgen::InsensitiveStr::Ascii("groupe"),
                &["group", "grouped"],
            ),
            (dictgen::InsensitiveStr::Ascii("addditions"), &["additions"]),
            (
                dictgen::InsensitiveStr::Ascii("intialzations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("projct"), &["project"]),
            (dictgen::InsensitiveStr::Ascii("execion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("overbraking"),
                &["overbearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("fatc"), &["fact"]),
            (dictgen::InsensitiveStr::Ascii("illusiones"), &["illusions"]),
            (
                dictgen::InsensitiveStr::Ascii("accomidates"),
                &["accommodates"],
            ),
            (dictgen::InsensitiveStr::Ascii("garphics"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequite"),
                &["prerequisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circulacion"),
                &["circulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("worklow"), &["workflow"]),
            (dictgen::InsensitiveStr::Ascii("fogets"), &["forgets"]),
            (
                dictgen::InsensitiveStr::Ascii("veriations"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("tramsmits"), &["transmits"]),
            (dictgen::InsensitiveStr::Ascii("cashies"), &["cashiers"]),
            (
                dictgen::InsensitiveStr::Ascii("indepenendence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("amtheyst"), &["amethyst"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomeative"),
                &["performative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpatrated"),
                &["perpetrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("generotor"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("statisticly"),
                &["statistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogynystic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheeseburgie"),
                &["cheeseburger"],
            ),
            (dictgen::InsensitiveStr::Ascii("incomping"), &["incoming"]),
            (
                dictgen::InsensitiveStr::Ascii("disadvatange"),
                &["disadvantage"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiggers"), &["triggers"]),
            (dictgen::InsensitiveStr::Ascii("corruptued"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("enviorenmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhethoric"), &["rhetoric"]),
            (dictgen::InsensitiveStr::Ascii("everytone"), &["everyone"]),
            (dictgen::InsensitiveStr::Ascii("aparrent"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("jurnyed"), &["journeyed"]),
            (dictgen::InsensitiveStr::Ascii("dackery"), &["daiquiri"]),
            (dictgen::InsensitiveStr::Ascii("approvement"), &["approval"]),
            (dictgen::InsensitiveStr::Ascii("brige"), &["bridge"]),
            (dictgen::InsensitiveStr::Ascii("contryie"), &["countryie"]),
            (dictgen::InsensitiveStr::Ascii("skecth"), &["sketch"]),
            (
                dictgen::InsensitiveStr::Ascii("calssification"),
                &["classification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultainously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capatilize"),
                &["capitalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspodence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("puritannical"),
                &["puritanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditonally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protagonistes"),
                &["protagonists"],
            ),
            (dictgen::InsensitiveStr::Ascii("perphas"), &["perhaps"]),
            (
                dictgen::InsensitiveStr::Ascii("syntehsized"),
                &["synthesized"],
            ),
            (dictgen::InsensitiveStr::Ascii("exturded"), &["extruded"]),
            (dictgen::InsensitiveStr::Ascii("ailenate"), &["alienate"]),
            (dictgen::InsensitiveStr::Ascii("condtiion"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("betwenn"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("continuare"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("sensetionalism"),
                &["sensationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("insigth"), &["insight"]),
            (dictgen::InsensitiveStr::Ascii("exploted"), &["explode"]),
            (
                dictgen::InsensitiveStr::Ascii("comminicate"),
                &["communicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placeholdr"),
                &["placeholder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investagated"),
                &["investigated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fontisizing"),
                &["fontifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("resouces"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("finnishing"), &["finishing"]),
            (
                dictgen::InsensitiveStr::Ascii("motherbaords"),
                &["motherboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquisiter"),
                &["inquisitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sculpter"),
                &["sculpture", "sculptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("concurent"), &["concurrent"]),
            (dictgen::InsensitiveStr::Ascii("calculcate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("stip"), &["stop", "strip"]),
            (dictgen::InsensitiveStr::Ascii("recorvery"), &["recovery"]),
            (dictgen::InsensitiveStr::Ascii("collares"), &["collars"]),
            (
                dictgen::InsensitiveStr::Ascii("mauals"),
                &["manuals", "mauls"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insecurites"),
                &["insecurities"],
            ),
            (dictgen::InsensitiveStr::Ascii("chescking"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("extenstivly"),
                &["extensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinmaic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("admision"), &["admission"]),
            (dictgen::InsensitiveStr::Ascii("sprinkel"), &["sprinkle"]),
            (dictgen::InsensitiveStr::Ascii("sythetics"), &["synthetics"]),
            (
                dictgen::InsensitiveStr::Ascii("understnadably"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("mey"), &["may"]),
            (dictgen::InsensitiveStr::Ascii("dectection"), &["detection"]),
            (dictgen::InsensitiveStr::Ascii("intruct"), &["instruct"]),
            (
                dictgen::InsensitiveStr::Ascii("ameelarate"),
                &["ameliorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("managmenet"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("visheator"), &["vitiator"]),
            (
                dictgen::InsensitiveStr::Ascii("demosntrated"),
                &["demonstrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mininizing"),
                &["minimizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponencially"),
                &["exponentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("polocies"), &["policies"]),
            (
                dictgen::InsensitiveStr::Ascii("groudwater"),
                &["groundwater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scripte"),
                &["script", "scripted"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempalte"), &["template"]),
            (dictgen::InsensitiveStr::Ascii("booard"), &["board"]),
            (dictgen::InsensitiveStr::Ascii("optioanl"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitialzes"),
                &["uninitializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavies"), &["behaves"]),
            (dictgen::InsensitiveStr::Ascii("rocords"), &["records"]),
            (dictgen::InsensitiveStr::Ascii("exmaples"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("hunderd"), &["hundred"]),
            (dictgen::InsensitiveStr::Ascii("regularily"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("reanaming"), &["renaming"]),
            (dictgen::InsensitiveStr::Ascii("whitleist"), &["whitelist"]),
            (
                dictgen::InsensitiveStr::Ascii("carniverous"),
                &["carnivorous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customsiations"),
                &["customisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mormones"), &["mormons"]),
            (
                dictgen::InsensitiveStr::Ascii("infilitration"),
                &["infiltration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combindators"),
                &["combinators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countercoat"),
                &["counteract"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("epilouge"),
                &["epilogue", "epilog"],
            ),
            (dictgen::InsensitiveStr::Ascii("advacne"), &["advance"]),
            (
                dictgen::InsensitiveStr::Ascii("boradcasts"),
                &["broadcasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presisting"),
                &["persisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpress"), &["suppress"]),
            (dictgen::InsensitiveStr::Ascii("screnn"), &["screen"]),
            (dictgen::InsensitiveStr::Ascii("alsready"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("avator"), &["avatar"]),
            (dictgen::InsensitiveStr::Ascii("reprorting"), &["reporting"]),
            (dictgen::InsensitiveStr::Ascii("sandlas"), &["sandals"]),
            (
                dictgen::InsensitiveStr::Ascii("combanator"),
                &["combinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependancies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affliation"),
                &["affliction", "affiliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snapshoted"),
                &["snapshotted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maesurement"),
                &["measurement"],
            ),
            (dictgen::InsensitiveStr::Ascii("ommitting"), &["omitting"]),
            (
                dictgen::InsensitiveStr::Ascii("negotioators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repostes"),
                &["reposts", "ripostes"],
            ),
            (dictgen::InsensitiveStr::Ascii("micorwave"), &["microwave"]),
            (dictgen::InsensitiveStr::Ascii("defautted"), &["defaulted"]),
            (
                dictgen::InsensitiveStr::Ascii("parrallelize"),
                &["parallelize"],
            ),
            (dictgen::InsensitiveStr::Ascii("impropre"), &["improper"]),
            (dictgen::InsensitiveStr::Ascii("screeshot"), &["screenshot"]),
            (
                dictgen::InsensitiveStr::Ascii("equilibrum"),
                &["equilibrium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continueous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("universella"),
                &["universally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sade"), &["sad"]),
            (
                dictgen::InsensitiveStr::Ascii("distintcly"),
                &["distinctly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consturctor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doamine"),
                &["dopamine", "domain"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlodge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("togehter"), &["together"]),
            (dictgen::InsensitiveStr::Ascii("countres"), &["counters"]),
            (dictgen::InsensitiveStr::Ascii("sqashed"), &["squashed"]),
            (dictgen::InsensitiveStr::Ascii("accpetor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("romainan"), &["romanian"]),
            (dictgen::InsensitiveStr::Ascii("triancle"), &["triangle"]),
            (dictgen::InsensitiveStr::Ascii("grubing"), &["grubbing"]),
            (dictgen::InsensitiveStr::Ascii("hundres"), &["hundreds"]),
            (
                dictgen::InsensitiveStr::Ascii("orchastrates"),
                &["orchestrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disctintions"),
                &["distinctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("fassinate"), &["fascinate"]),
            (
                dictgen::InsensitiveStr::Ascii("immedeatly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("gigabtye"), &["gigabyte"]),
            (
                dictgen::InsensitiveStr::Ascii("gibabyte"),
                &["gibibyte", "gigabyte"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceepts"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("consier"), &["consider"]),
            (
                dictgen::InsensitiveStr::Ascii("coinsidentally"),
                &["coincidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("chec"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("poitivity"), &["positivity"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstsnding"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prodictions"),
                &["productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("regluar"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("transffered"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("summarizen"), &["summarize"]),
            (dictgen::InsensitiveStr::Ascii("ciper"), &["cipher"]),
            (dictgen::InsensitiveStr::Ascii("staduim"), &["stadium"]),
            (
                dictgen::InsensitiveStr::Ascii("pecular"),
                &["peculiar", "specular", "secular"],
            ),
            (dictgen::InsensitiveStr::Ascii("defendas"), &["defends"]),
            (
                dictgen::InsensitiveStr::Ascii("cimetricly"),
                &["symmetricly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcations"),
                &["transactions", "translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconscioulsy"),
                &["unconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("traking"), &["tracking"]),
            (dictgen::InsensitiveStr::Ascii("newmonia"), &["pneumonia"]),
            (dictgen::InsensitiveStr::Ascii("downstar"), &["downstairs"]),
            (dictgen::InsensitiveStr::Ascii("thottled"), &["throttled"]),
            (dictgen::InsensitiveStr::Ascii("modfify"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("rewriite"), &["rewrite"]),
            (
                dictgen::InsensitiveStr::Ascii("journyes"),
                &["journeyed", "journeys"],
            ),
            (dictgen::InsensitiveStr::Ascii("mounc"), &["mount"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsequencial"),
                &["inconsequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializationa"),
                &["initialization", "initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("disscted"), &["dissected"]),
            (dictgen::InsensitiveStr::Ascii("staitions"), &["stations"]),
            (dictgen::InsensitiveStr::Ascii("dreasm"), &["dreams"]),
            (dictgen::InsensitiveStr::Ascii("butterry"), &["buttery"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessessarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("perferct"), &["perfect"]),
            (dictgen::InsensitiveStr::Ascii("perdict"), &["predict"]),
            (
                dictgen::InsensitiveStr::Ascii("garanteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecesserily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agressivnes"),
                &["aggressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("firsr"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("enumable"), &["enumerable"]),
            (
                dictgen::InsensitiveStr::Ascii("prededence"),
                &["precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("scecified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("orhtogonally"),
                &["orthogonally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nostolgically"),
                &["nostalgically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ein"), &["in"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseutical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rounter"),
                &["router", "rounder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funchtional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantations"),
                &["instantiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("rtemove"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("pinteret"), &["pinterest"]),
            (
                dictgen::InsensitiveStr::Ascii("successeurs"),
                &["successes"],
            ),
            (dictgen::InsensitiveStr::Ascii("samue"), &["same", "samuel"]),
            (dictgen::InsensitiveStr::Ascii("intitially"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("handwriten"),
                &["handwritten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrtupible"),
                &["interruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("disposse"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("fimrware"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("appent"), &["append"]),
            (dictgen::InsensitiveStr::Ascii("timestan"), &["timespan"]),
            (
                dictgen::InsensitiveStr::Ascii("reprecussion"),
                &["repercussions", "repercussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("operad"), &["operand"]),
            (dictgen::InsensitiveStr::Ascii("vailidate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("regisering"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthinic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agregations"),
                &["aggregations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinement"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("queation"), &["equation"]),
            (
                dictgen::InsensitiveStr::Ascii("quesitonnaires"),
                &["questionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("lablel"), &["label"]),
            (dictgen::InsensitiveStr::Ascii("monololy"), &["monopoly"]),
            (dictgen::InsensitiveStr::Ascii("potiential"), &["potential"]),
            (dictgen::InsensitiveStr::Ascii("qaulify"), &["qualify"]),
            (dictgen::InsensitiveStr::Ascii("redundany"), &["redundancy"]),
            (
                dictgen::InsensitiveStr::Ascii("deseralise"),
                &["deserialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumary"), &["summary"]),
            (
                dictgen::InsensitiveStr::Ascii("projctions"),
                &["projections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcalators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("togater"), &["together"]),
            (
                dictgen::InsensitiveStr::Ascii("exprectation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistend"),
                &["consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("choses"), &["chooses"]),
            (
                dictgen::InsensitiveStr::Ascii("accumalator"),
                &["accumulator"],
            ),
            (dictgen::InsensitiveStr::Ascii("enourage"), &["encourage"]),
            (
                dictgen::InsensitiveStr::Ascii("disturbence"),
                &["disturbance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissits"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("defautlt"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("denomitator"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("jetsream"), &["jetstream"]),
            (dictgen::InsensitiveStr::Ascii("elitisim"), &["elitism"]),
            (
                dictgen::InsensitiveStr::Ascii("aotomaticall"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepdendant"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oeprational"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugget"), &["suggest"]),
            (dictgen::InsensitiveStr::Ascii("intorvert"), &["introvert"]),
            (dictgen::InsensitiveStr::Ascii("seacrched"), &["searched"]),
            (
                dictgen::InsensitiveStr::Ascii("unspecializated"),
                &["unspecialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transeiver"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("prevoius"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("ossociate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("projcet"), &["project"]),
            (dictgen::InsensitiveStr::Ascii("ballooon"), &["balloon"]),
            (dictgen::InsensitiveStr::Ascii("attmepted"), &["attempted"]),
            (
                dictgen::InsensitiveStr::Ascii("dictoinary"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("referrrer"), &["referrer"]),
            (dictgen::InsensitiveStr::Ascii("acctivity"), &["activity"]),
            (dictgen::InsensitiveStr::Ascii("loadad"), &["loaded"]),
            (dictgen::InsensitiveStr::Ascii("oragnered"), &["orangered"]),
            (dictgen::InsensitiveStr::Ascii("interfeer"), &["interfere"]),
            (dictgen::InsensitiveStr::Ascii("begiinner"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("aprrovement"), &["approval"]),
            (dictgen::InsensitiveStr::Ascii("certai"), &["certain"]),
            (
                dictgen::InsensitiveStr::Ascii("fullfillment"),
                &["fulfillment"],
            ),
            (dictgen::InsensitiveStr::Ascii("requrest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("possibiliy"),
                &["possibility", "possibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sovling"), &["solving"]),
            (
                dictgen::InsensitiveStr::Ascii("paermission"),
                &["permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("awoid"), &["avoid"]),
            (dictgen::InsensitiveStr::Ascii("faiway"), &["fairway"]),
            (
                dictgen::InsensitiveStr::Ascii("acknodledgments"),
                &["acknowledgments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("magnitudine"),
                &["magnitude"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("romanticaly"),
                &["romantically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reptiel"), &["reptile"]),
            (
                dictgen::InsensitiveStr::Ascii("equel"),
                &["equal", "sequel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motherboads"),
                &["motherboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("tihnking"), &["thinking"]),
            (
                dictgen::InsensitiveStr::Ascii("suspention"),
                &["suspension"],
            ),
            (dictgen::InsensitiveStr::Ascii("refeshing"), &["refreshing"]),
            (
                dictgen::InsensitiveStr::Ascii("defrentiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("generetes"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("disputerad"), &["disputed"]),
            (
                dictgen::InsensitiveStr::Ascii("restiction"),
                &["restriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressemle"), &["resemble"]),
            (dictgen::InsensitiveStr::Ascii("inventoty"), &["inventory"]),
            (
                dictgen::InsensitiveStr::Ascii("adminstrative"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("preprend"), &["prepend"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelize"),
                &["parallelize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustratior"),
                &["illustrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinstatiate"),
                &["reinstantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispaches"), &["dispatches"]),
            (dictgen::InsensitiveStr::Ascii("unduee"), &["undue"]),
            (dictgen::InsensitiveStr::Ascii("multile"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("accepatnce"),
                &["acceptance"],
            ),
            (dictgen::InsensitiveStr::Ascii("scollable"), &["scrollable"]),
            (dictgen::InsensitiveStr::Ascii("thtreading"), &["threading"]),
            (
                dictgen::InsensitiveStr::Ascii("corresoponing"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moew"),
                &["meow", "more", "mow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctoin"),
                &["correction"],
            ),
            (dictgen::InsensitiveStr::Ascii("biogtry"), &["bigotry"]),
            (dictgen::InsensitiveStr::Ascii("separarte"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("limitating"),
                &["limitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconveived"),
                &["preconceived"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbracing"), &["embracing"]),
            (dictgen::InsensitiveStr::Ascii("vegtable"), &["vegetable"]),
            (dictgen::InsensitiveStr::Ascii("commense"), &["commenters"]),
            (dictgen::InsensitiveStr::Ascii("hieght"), &["height"]),
            (dictgen::InsensitiveStr::Ascii("opetions"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("amendmend"), &["amendment"]),
            (dictgen::InsensitiveStr::Ascii("componets"), &["components"]),
            (dictgen::InsensitiveStr::Ascii("selenoid"), &["solenoid"]),
            (
                dictgen::InsensitiveStr::Ascii("explicitely"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ehich"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("interefaces"),
                &["interfaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("murderered"), &["murdered"]),
            (dictgen::InsensitiveStr::Ascii("lackers"), &["lacquers"]),
            (dictgen::InsensitiveStr::Ascii("strenous"), &["strenuous"]),
            (
                dictgen::InsensitiveStr::Ascii("autoattaks"),
                &["autoattack"],
            ),
            (dictgen::InsensitiveStr::Ascii("reaquest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("suppressio"),
                &["suppressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticateion"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("passin"), &["passing"]),
            (
                dictgen::InsensitiveStr::Ascii("objectivication"),
                &["objectification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terimnator"),
                &["terminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("noifier"), &["notifier"]),
            (dictgen::InsensitiveStr::Ascii("disrete"), &["discrete"]),
            (dictgen::InsensitiveStr::Ascii("mwcos"), &["macos"]),
            (
                dictgen::InsensitiveStr::Ascii("offficially"),
                &["officially"],
            ),
            (dictgen::InsensitiveStr::Ascii("attmepts"), &["attempts"]),
            (dictgen::InsensitiveStr::Ascii("silimiar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("inaugures"),
                &["inaugurates"],
            ),
            (dictgen::InsensitiveStr::Ascii("avdisory"), &["advisory"]),
            (dictgen::InsensitiveStr::Ascii("reayd"), &["ready", "read"]),
            (dictgen::InsensitiveStr::Ascii("expersed"), &["expressed"]),
            (
                dictgen::InsensitiveStr::Ascii("mysogenistic"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("detctive"), &["detective"]),
            (dictgen::InsensitiveStr::Ascii("spefiifeid"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("kronicles"), &["chronicles"]),
            (dictgen::InsensitiveStr::Ascii("tookkits"), &["toolkits"]),
            (
                dictgen::InsensitiveStr::Ascii("microtransacrions"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("standtard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("expetcs"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("overthining"),
                &["overthinking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conceitual"),
                &["conceptual"],
            ),
            (dictgen::InsensitiveStr::Ascii("colescing"), &["coalescing"]),
            (dictgen::InsensitiveStr::Ascii("origninal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("meging"), &["merging"]),
            (
                dictgen::InsensitiveStr::Ascii("recepients"),
                &["recipients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undesitable"),
                &["undesirable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singulariy"),
                &["singularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equvivalent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilizaiton"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahinets"), &["cabinets"]),
            (dictgen::InsensitiveStr::Ascii("ue"), &["use", "due"]),
            (dictgen::InsensitiveStr::Ascii("aletrnate"), &["alternate"]),
            (dictgen::InsensitiveStr::Ascii("tansition"), &["transition"]),
            (dictgen::InsensitiveStr::Ascii("analoguous"), &["analogous"]),
            (dictgen::InsensitiveStr::Ascii("mimatches"), &["mismatches"]),
            (
                dictgen::InsensitiveStr::Ascii("deail"),
                &["deal", "detail", "derail"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheighs"), &["weighs"]),
            (dictgen::InsensitiveStr::Ascii("ags"), &["ages", "tags"]),
            (dictgen::InsensitiveStr::Ascii("opimizer"), &["optimizer"]),
            (
                dictgen::InsensitiveStr::Ascii("govenrment"),
                &["government"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrsepondingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("distroyed"), &["destroyed"]),
            (dictgen::InsensitiveStr::Ascii("cirticise"), &["criticise"]),
            (dictgen::InsensitiveStr::Ascii("libraier"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("sieze"),
                &["seize", "size", "siege"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proection"),
                &["protection", "projection"],
            ),
            (dictgen::InsensitiveStr::Ascii("cient"), &["client"]),
            (
                dictgen::InsensitiveStr::Ascii("polysaccharid"),
                &["polysaccharide"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predecesors"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassociation"),
                &["disassociation", "dissociation"],
            ),
            (dictgen::InsensitiveStr::Ascii("poverful"), &["powerful"]),
            (dictgen::InsensitiveStr::Ascii("donwloads"), &["downloads"]),
            (
                dictgen::InsensitiveStr::Ascii("civalizations"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confidencialy"),
                &["confidentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreicate"),
                &["appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorespondends"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incosistency"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("netween"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("prugatory"), &["purgatory"]),
            (
                dictgen::InsensitiveStr::Ascii("subtaks"),
                &["subtask", "subtasks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copilers"),
                &["compilers", "copiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antagonisic"),
                &["antagonistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("recovr"), &["recover"]),
            (dictgen::InsensitiveStr::Ascii("chercking"), &["checking"]),
            (
                dictgen::InsensitiveStr::Ascii("negociable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("envriron"), &["environ"]),
            (
                dictgen::InsensitiveStr::Ascii("uncontrolablly"),
                &["uncontrollably"],
            ),
            (dictgen::InsensitiveStr::Ascii("calenday"), &["calendar"]),
            (dictgen::InsensitiveStr::Ascii("foloww"), &["follow"]),
            (
                dictgen::InsensitiveStr::Ascii("rehighlightes"),
                &["rehighlights", "rehighlighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalisitc"),
                &["journalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cardiovasculaire"),
                &["cardiovascular"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterrator"), &["iterator"]),
            (
                dictgen::InsensitiveStr::Ascii("detremental"),
                &["detrimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bactracked"),
                &["backtracked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converstions"),
                &["conversions", "conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intemediary"),
                &["intermediary"],
            ),
            (dictgen::InsensitiveStr::Ascii("renyolds"), &["reynolds"]),
            (dictgen::InsensitiveStr::Ascii("iound"), &["round", "wound"]),
            (
                dictgen::InsensitiveStr::Ascii("degenerite"),
                &["degenerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atructurally"),
                &["structurally"],
            ),
            (dictgen::InsensitiveStr::Ascii("attivate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("perpretated"),
                &["perpetrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("speek"), &["speak"]),
            (dictgen::InsensitiveStr::Ascii("multipled"), &["multiplied"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunatey"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("enymore"), &["anymore"]),
            (dictgen::InsensitiveStr::Ascii("numeraotr"), &["numerator"]),
            (
                dictgen::InsensitiveStr::Ascii("proposito"),
                &["proposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpbabetic"),
                &["alphabetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwiwng"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("moistorizing"),
                &["moisturizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributiors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("intherited"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("proceder"), &["procedure"]),
            (
                dictgen::InsensitiveStr::Ascii("descriotions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apperarance"),
                &["appearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("spece"), &["space", "spice"]),
            (dictgen::InsensitiveStr::Ascii("completeds"), &["completes"]),
            (dictgen::InsensitiveStr::Ascii("wildgets"), &["widgets"]),
            (
                dictgen::InsensitiveStr::Ascii("desttructor"),
                &["destructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("cnanging"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("debudgger"), &["debugger"]),
            (
                dictgen::InsensitiveStr::Ascii("amargeddon"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentrato"),
                &["concentration"],
            ),
            (dictgen::InsensitiveStr::Ascii("genral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("qerying"), &["querying"]),
            (
                dictgen::InsensitiveStr::Ascii("singelarity"),
                &["singularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("dowload"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("directrives"),
                &["directives"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrappng"), &["wrapping"]),
            (dictgen::InsensitiveStr::Ascii("silybus"), &["syllabus"]),
            (dictgen::InsensitiveStr::Ascii("shepards"), &["shepherds"]),
            (dictgen::InsensitiveStr::Ascii("seantors"), &["senators"]),
            (dictgen::InsensitiveStr::Ascii("repid"), &["rapid"]),
            (
                dictgen::InsensitiveStr::Ascii("exectutables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hetergeneous"),
                &["heterogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("protoype"), &["prototype"]),
            (dictgen::InsensitiveStr::Ascii("bugg"), &["bug"]),
            (
                dictgen::InsensitiveStr::Ascii("confifured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("dentence"), &["sentence"]),
            (
                dictgen::InsensitiveStr::Ascii("pestacides"),
                &["pesticides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reppository"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifictaion"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplied"),
                &["simplified", "implied"],
            ),
            (dictgen::InsensitiveStr::Ascii("desomnd"), &["desmond"]),
            (dictgen::InsensitiveStr::Ascii("classicos"), &["classics"]),
            (dictgen::InsensitiveStr::Ascii("exsample"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("ojbecting"), &["objecting"]),
            (
                dictgen::InsensitiveStr::Ascii("constituion"),
                &["constitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronunciacion"),
                &["pronunciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrpting"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("kenendy"), &["kennedy"]),
            (
                dictgen::InsensitiveStr::Ascii("intentionnal"),
                &["intentional"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfsets"), &["offsets"]),
            (dictgen::InsensitiveStr::Ascii("cotrolled"), &["controlled"]),
            (dictgen::InsensitiveStr::Ascii("streamys"), &["streams"]),
            (dictgen::InsensitiveStr::Ascii("anytrhing"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("deccremented"),
                &["decremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("apreciate"), &["appreciate"]),
            (dictgen::InsensitiveStr::Ascii("mounring"), &["mourning"]),
            (
                dictgen::InsensitiveStr::Ascii("mocrotransactions"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommaded"),
                &["recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certficiations"),
                &["certifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("mairjuana"), &["marijuana"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherises"),
                &["characterises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiatiing"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulantaneously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("villify"), &["vilify"]),
            (dictgen::InsensitiveStr::Ascii("opeartios"), &["operations"]),
            (dictgen::InsensitiveStr::Ascii("rotater"), &["rotator"]),
            (
                dictgen::InsensitiveStr::Ascii("recursily"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluatiing"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dosen"),
                &["dozen", "dose", "doesn"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedamies"), &["academies"]),
            (
                dictgen::InsensitiveStr::Ascii("anevironment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("discribe"), &["describe"]),
            (
                dictgen::InsensitiveStr::Ascii("enventually"),
                &["eventually"],
            ),
            (dictgen::InsensitiveStr::Ascii("diciding"), &["deciding"]),
            (
                dictgen::InsensitiveStr::Ascii("unfirendly"),
                &["unfriendly"],
            ),
            (dictgen::InsensitiveStr::Ascii("awesomley"), &["awesomely"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidting"),
                &["invalidating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoticated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremenent"),
                &["increment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("namespaeed"),
                &["namespaced"],
            ),
            (dictgen::InsensitiveStr::Ascii("checlout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("extennsion"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("asncronous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("ilegle"), &["illegal"]),
            (
                dictgen::InsensitiveStr::Ascii("catastrphe"),
                &["catastrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("flate"), &["flat"]),
            (dictgen::InsensitiveStr::Ascii("apreteate"), &["appreciate"]),
            (dictgen::InsensitiveStr::Ascii("zookeenee"), &["zucchini"]),
            (
                dictgen::InsensitiveStr::Ascii("transconding"),
                &["transcoding"],
            ),
            (dictgen::InsensitiveStr::Ascii("resereved"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("corsari"), &["corsair"]),
            (dictgen::InsensitiveStr::Ascii("defaut"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("independnce"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsoloted"), &["obsoleted"]),
            (dictgen::InsensitiveStr::Ascii("obsloeted"), &["obsoleted"]),
            (
                dictgen::InsensitiveStr::Ascii("technologicaly"),
                &["technologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("revisiones"), &["revisions"]),
            (
                dictgen::InsensitiveStr::Ascii("rendazyvou"),
                &["rendezvous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masternind"),
                &["mastermind"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strenghtens"),
                &["strengthen"],
            ),
            (dictgen::InsensitiveStr::Ascii("directiory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("appropached"),
                &["approached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("torhclight"),
                &["torchlight"],
            ),
            (dictgen::InsensitiveStr::Ascii("dekstops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("recourced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("condradicting"),
                &["contradicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutioners"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tabualtors"),
                &["tabulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("chearset"), &["charset"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaitally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("rerwite"), &["rewrite"]),
            (
                dictgen::InsensitiveStr::Ascii("coversations"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentailly"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("puplications"),
                &["publications", "duplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("contenst"), &["contents"]),
            (
                dictgen::InsensitiveStr::Ascii("prefferring"),
                &["preferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("feets"), &["feet", "feats"]),
            (
                dictgen::InsensitiveStr::Ascii("brance"),
                &["branch", "brace", "branches"],
            ),
            (dictgen::InsensitiveStr::Ascii("esclusive"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("originalty"),
                &["originality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepetrator"),
                &["perpetrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("stattered"), &["scattered"]),
            (
                dictgen::InsensitiveStr::Ascii("sligh"),
                &["slight", "sleigh"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapens"), &["happens"]),
            (dictgen::InsensitiveStr::Ascii("circluars"), &["circulars"]),
            (
                dictgen::InsensitiveStr::Ascii("gatin"),
                &["gating", "latin", "satin", "gain", "gratin"],
            ),
            (dictgen::InsensitiveStr::Ascii("forign"), &["foreign"]),
            (dictgen::InsensitiveStr::Ascii("comunism"), &["communism"]),
            (dictgen::InsensitiveStr::Ascii("consiting"), &["consisting"]),
            (dictgen::InsensitiveStr::Ascii("vindow"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("procuced"),
                &["procured", "produced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selcetable"),
                &["selectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstalable"),
                &["uninstallable"],
            ),
            (dictgen::InsensitiveStr::Ascii("documeted"), &["documented"]),
            (
                dictgen::InsensitiveStr::Ascii("experimmented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviroiment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succussion"),
                &["succession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compareisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilis"), &["utilise"]),
            (
                dictgen::InsensitiveStr::Ascii("dependening"),
                &["depending"],
            ),
            (dictgen::InsensitiveStr::Ascii("cringely"), &["cringey"]),
            (dictgen::InsensitiveStr::Ascii("dictionay"), &["dictionary"]),
            (
                dictgen::InsensitiveStr::Ascii("difining"),
                &["defining", "divining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comopnents"),
                &["components"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderig"), &["rendering"]),
            (dictgen::InsensitiveStr::Ascii("brutallity"), &["brutally"]),
            (dictgen::InsensitiveStr::Ascii("standerd"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotatied"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uniquiness"),
                &["uniqueness"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterrates"), &["iterates"]),
            (dictgen::InsensitiveStr::Ascii("occuppant"), &["occupant"]),
            (dictgen::InsensitiveStr::Ascii("simiply"), &["simply"]),
            (dictgen::InsensitiveStr::Ascii("userapace"), &["userspace"]),
            (dictgen::InsensitiveStr::Ascii("oustide"), &["outside"]),
            (
                dictgen::InsensitiveStr::Ascii("bunding"),
                &["binding", "bounding", "bundling"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrnaging"), &["arranging"]),
            (dictgen::InsensitiveStr::Ascii("opimizers"), &["optimizers"]),
            (dictgen::InsensitiveStr::Ascii("attnetion"), &["attention"]),
            (
                dictgen::InsensitiveStr::Ascii("strcutures"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("clapses"), &["collapses"]),
            (dictgen::InsensitiveStr::Ascii("labaelled"), &["labelled"]),
            (dictgen::InsensitiveStr::Ascii("cuasation"), &["causation"]),
            (
                dictgen::InsensitiveStr::Ascii("specializin"),
                &["specializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contenplating"),
                &["contemplating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosist"),
                &["consist", "cosiest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalascence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconected"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("activasted"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("subolders"), &["subfolders"]),
            (dictgen::InsensitiveStr::Ascii("characteer"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("blueberies"),
                &["blueberries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleises"),
                &["parallelises"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciefic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("anyoens"), &["anyones"]),
            (dictgen::InsensitiveStr::Ascii("eletrode"), &["electrode"]),
            (
                dictgen::InsensitiveStr::Ascii("requie"),
                &["require", "requite", "requiem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abovmentioned"),
                &["abovementioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("austeriy"), &["austerity"]),
            (
                dictgen::InsensitiveStr::Ascii("phylosophical"),
                &["philosophical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devie"),
                &["device", "divvy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manslaugter"),
                &["manslaughter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoucenment"),
                &["announcements", "announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pigens"),
                &["pigeons", "pigpens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fumctional"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreates"), &["operates"]),
            (dictgen::InsensitiveStr::Ascii("althought"), &["although"]),
            (dictgen::InsensitiveStr::Ascii("throuth"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("directgories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("queisce"), &["quiesce"]),
            (dictgen::InsensitiveStr::Ascii("confirmas"), &["confirms"]),
            (
                dictgen::InsensitiveStr::Ascii("armageddeon"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concantenates"),
                &["concatenates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranformer"),
                &["transformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cofiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquisation"),
                &["acquisition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miselaneous"),
                &["miscellaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("brockened"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("genererates"),
                &["generates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechandise"),
                &["merchandise"],
            ),
            (dictgen::InsensitiveStr::Ascii("methhod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("disapointment"),
                &["disappointment"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqquire"), &["acquire"]),
            (dictgen::InsensitiveStr::Ascii("hed"), &["head"]),
            (
                dictgen::InsensitiveStr::Ascii("acheviable"),
                &["achievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("reduntant"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("insufficent"),
                &["insufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionaly"),
                &["functionally", "functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("ddons"), &["addons"]),
            (dictgen::InsensitiveStr::Ascii("notority"), &["notoriety"]),
            (dictgen::InsensitiveStr::Ascii("butthoe"), &["butthole"]),
            (dictgen::InsensitiveStr::Ascii("executs"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("avriable"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("counterfited"),
                &["counterfeit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standarizing"),
                &["standardizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explinations"),
                &["explanations"],
            ),
            (dictgen::InsensitiveStr::Ascii("desctroy"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("eleiminate"), &["eliminate"]),
            (
                dictgen::InsensitiveStr::Ascii("uncontrallable"),
                &["uncontrollable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cateogries"),
                &["categories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accourdance"),
                &["accordance"],
            ),
            (dictgen::InsensitiveStr::Ascii("htat"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("meessage"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("repulics"), &["republics"]),
            (
                dictgen::InsensitiveStr::Ascii("warpprer"),
                &["warper", "wrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechanicus"), &["mechanics"]),
            (dictgen::InsensitiveStr::Ascii("forgivance"), &["forgiven"]),
            (dictgen::InsensitiveStr::Ascii("concensus"), &["consensus"]),
            (
                dictgen::InsensitiveStr::Ascii("adaptatiton"),
                &["adaptation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimmetriclly"),
                &["symmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("singlton"), &["singleton"]),
            (
                dictgen::InsensitiveStr::Ascii("experementally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("billionairres"),
                &["billionaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntaks"), &["syntax"]),
            (dictgen::InsensitiveStr::Ascii("larinxes"), &["larynxes"]),
            (dictgen::InsensitiveStr::Ascii("nitching"), &["niching"]),
            (
                dictgen::InsensitiveStr::Ascii("synchronuously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ctificated"),
                &["certificated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrbuting"),
                &["distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("orthognal"), &["orthogonal"]),
            (dictgen::InsensitiveStr::Ascii("sodder"), &["solder"]),
            (
                dictgen::InsensitiveStr::Ascii("differentitation"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understnad"),
                &["understands", "understand"],
            ),
            (dictgen::InsensitiveStr::Ascii("globelly"), &["globally"]),
            (
                dictgen::InsensitiveStr::Ascii("inconfort"),
                &["discomfort", "uncomfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("becausw"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("kronicled"), &["chronicled"]),
            (dictgen::InsensitiveStr::Ascii("dublicade"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("maintenaince"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("sptintf"), &["sprintf"]),
            (
                dictgen::InsensitiveStr::Ascii("definetlly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("fullfils"), &["fulfils"]),
            (
                dictgen::InsensitiveStr::Ascii("cerated"),
                &["created", "serrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirarchical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cilindrical"),
                &["cylindrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upsteramer"),
                &["upstreamer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vacinating"),
                &["vaccinating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitition"),
                &["constitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antiobitic"),
                &["antibiotic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juggernault"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coloublind"),
                &["colorblind"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumelate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speicalist"),
                &["specialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semconductor"),
                &["semiconductor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redesinged"),
                &["redesigned"],
            ),
            (dictgen::InsensitiveStr::Ascii("sserver"), &["server"]),
            (
                dictgen::InsensitiveStr::Ascii("decice"),
                &["device", "decide", "decile"],
            ),
            (dictgen::InsensitiveStr::Ascii("downlaodes"), &["downloads"]),
            (
                dictgen::InsensitiveStr::Ascii("roundabounts"),
                &["roundabout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenetation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pervents"),
                &["prevents", "perverts"],
            ),
            (dictgen::InsensitiveStr::Ascii("motoroloa"), &["motorola"]),
            (dictgen::InsensitiveStr::Ascii("zellot"), &["zealot"]),
            (
                dictgen::InsensitiveStr::Ascii("influenting"),
                &["influencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appointmnet"),
                &["appointment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consecutivly"),
                &["consecutively"],
            ),
            (dictgen::InsensitiveStr::Ascii("fufilled"), &["fulfilled"]),
            (
                dictgen::InsensitiveStr::Ascii("conert"),
                &["convert", "concert"],
            ),
            (dictgen::InsensitiveStr::Ascii("applide"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("durabiliy"), &["durability"]),
            (dictgen::InsensitiveStr::Ascii("stagess"), &["stages"]),
            (
                dictgen::InsensitiveStr::Ascii("coordonated"),
                &["coordinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("naopleon"), &["napoleon"]),
            (dictgen::InsensitiveStr::Ascii("ilegal"), &["illegal"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetamin"),
                &["amphetamine"],
            ),
            (dictgen::InsensitiveStr::Ascii("existng"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("balacer"), &["balancer"]),
            (dictgen::InsensitiveStr::Ascii("parralell"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("sopports"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("hetreosexual"),
                &["heterosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documention"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiantion"),
                &["combinations", "combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("confins"), &["confines"]),
            (dictgen::InsensitiveStr::Ascii("jitterr"), &["jitter"]),
            (dictgen::InsensitiveStr::Ascii("ading"), &["adding"]),
            (
                dictgen::InsensitiveStr::Ascii("paschurising"),
                &["pasteurising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achiver"),
                &["achiever", "archiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacemet"),
                &["replacements", "replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomalize"), &["normalize"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccurrate"),
                &["inaccurate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lackeres"), &["lacquers"]),
            (dictgen::InsensitiveStr::Ascii("trunlde"), &["trundle"]),
            (
                dictgen::InsensitiveStr::Ascii("anitialising"),
                &["antialiasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temperarure"),
                &["temperature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cerimonious"),
                &["ceremonious"],
            ),
            (dictgen::InsensitiveStr::Ascii("dfinition"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("reccursively"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylse"), &["analyse"]),
            (
                dictgen::InsensitiveStr::Ascii("sucseeding"),
                &["succeeding"],
            ),
            (dictgen::InsensitiveStr::Ascii("statemet"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("expireds"), &["expires"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuccessul"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lousiville"),
                &["louisville"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronizations"),
                &["synchronizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mercaneries"),
                &["mercenaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("hefer"), &["heifer"]),
            (
                dictgen::InsensitiveStr::Ascii("claculating"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("compsite"), &["composite"]),
            (dictgen::InsensitiveStr::Ascii("udnerdog"), &["underdog"]),
            (dictgen::InsensitiveStr::Ascii("mirgaine"), &["migraine"]),
            (dictgen::InsensitiveStr::Ascii("deaemon"), &["daemon"]),
            (
                dictgen::InsensitiveStr::Ascii("infromational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaialbale"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("assotiative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnetes"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitnational"),
                &["multinational"],
            ),
            (dictgen::InsensitiveStr::Ascii("violance"), &["violence"]),
            (dictgen::InsensitiveStr::Ascii("wihch"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("coinjunction"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermimentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negothiating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cermonial"), &["ceremonial"]),
            (
                dictgen::InsensitiveStr::Ascii("intepolations"),
                &["interpolations"],
            ),
            (dictgen::InsensitiveStr::Ascii("insipred"), &["inspired"]),
            (
                dictgen::InsensitiveStr::Ascii("utilitizing"),
                &["utilizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomranse"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowlageable"),
                &["knowledgeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philisophies"),
                &["philosophies"],
            ),
            (dictgen::InsensitiveStr::Ascii("overand"), &["operand"]),
            (dictgen::InsensitiveStr::Ascii("statictic"), &["statistic"]),
            (dictgen::InsensitiveStr::Ascii("excetra"), &["etcetera"]),
            (dictgen::InsensitiveStr::Ascii("recurrant"), &["recurrent"]),
            (dictgen::InsensitiveStr::Ascii("behavriors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("ogranisation"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chidlrens"), &["childrens"]),
            (
                dictgen::InsensitiveStr::Ascii("presreving"),
                &["preserving"],
            ),
            (dictgen::InsensitiveStr::Ascii("corelates"), &["correlates"]),
            (dictgen::InsensitiveStr::Ascii("hamliton"), &["hamilton"]),
            (
                dictgen::InsensitiveStr::Ascii("functinoed"),
                &["functioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivilent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominato"),
                &["denomination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempurature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("ehr"), &["her"]),
            (dictgen::InsensitiveStr::Ascii("reviwed"), &["reviewed"]),
            (
                dictgen::InsensitiveStr::Ascii("curcumcision"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("marter"), &["martyr"]),
            (dictgen::InsensitiveStr::Ascii("packgs"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("geerator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("tremelo"), &["tremolo"]),
            (dictgen::InsensitiveStr::Ascii("failitate"), &["facilitate"]),
            (
                dictgen::InsensitiveStr::Ascii("imaginitave"),
                &["imaginative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controverisal"),
                &["controversial"],
            ),
            (dictgen::InsensitiveStr::Ascii("disbeleif"), &["disbelief"]),
            (
                dictgen::InsensitiveStr::Ascii("schizophrania"),
                &["schizophrenia"],
            ),
            (dictgen::InsensitiveStr::Ascii("etsts"), &["tests"]),
            (dictgen::InsensitiveStr::Ascii("applikay"), &["appliqué"]),
            (
                dictgen::InsensitiveStr::Ascii("pedohpiles"),
                &["pedophiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternarives"),
                &["alternatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nieghborhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("expectany"), &["expectancy"]),
            (
                dictgen::InsensitiveStr::Ascii("attributres"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("updrages"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("dictacted"), &["dictated"]),
            (
                dictgen::InsensitiveStr::Ascii("genearators"),
                &["generators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexedacimals"),
                &["hexadecimals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progresively"),
                &["progressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scarificed"),
                &["sacrificed"],
            ),
            (dictgen::InsensitiveStr::Ascii("curren"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("murderes"), &["murders"]),
            (
                dictgen::InsensitiveStr::Ascii("enviromnentally"),
                &["environmentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("asgolute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("eelment"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("overthinkig"),
                &["overthinking"],
            ),
            (dictgen::InsensitiveStr::Ascii("abosrb"), &["absorb"]),
            (dictgen::InsensitiveStr::Ascii("durationm"), &["duration"]),
            (
                dictgen::InsensitiveStr::Ascii("agricultral"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacifier"),
                &["specifier", "pacifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaild"), &["valid"]),
            (dictgen::InsensitiveStr::Ascii("direciton"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("continuoous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("observabil"),
                &["observable"],
            ),
            (dictgen::InsensitiveStr::Ascii("unitled"), &["untitled"]),
            (
                dictgen::InsensitiveStr::Ascii("coordintaes"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perisistent"),
                &["persistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sizemologist"),
                &["seismologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("instralls"), &["installs"]),
            (
                dictgen::InsensitiveStr::Ascii("occasionly"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exercices"),
                &["exercise", "exercises"],
            ),
            (dictgen::InsensitiveStr::Ascii("polyedral"), &["polyhedral"]),
            (dictgen::InsensitiveStr::Ascii("shitstrom"), &["shitstorm"]),
            (
                dictgen::InsensitiveStr::Ascii("precentages"),
                &["percentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("manfested"), &["manifested"]),
            (
                dictgen::InsensitiveStr::Ascii("experamentelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfured"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("electrial"), &["electrical"]),
            (
                dictgen::InsensitiveStr::Ascii("meditatiing"),
                &["meditating"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesing"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("allocats"), &["allocates"]),
            (dictgen::InsensitiveStr::Ascii("shetlers"), &["shelters"]),
            (dictgen::InsensitiveStr::Ascii("coninuous"), &["continuous"]),
            (dictgen::InsensitiveStr::Ascii("abandonds"), &["abandons"]),
            (dictgen::InsensitiveStr::Ascii("pointetr"), &["pointer"]),
            (dictgen::InsensitiveStr::Ascii("qucikest"), &["quickest"]),
            (
                dictgen::InsensitiveStr::Ascii("paremeters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("republicanas"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("fixeme"), &["fixme"]),
            (
                dictgen::InsensitiveStr::Ascii("simultaneoulys"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprearances"),
                &["appearances"],
            ),
            (dictgen::InsensitiveStr::Ascii("vcontains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("overpoweing"),
                &["overpowering"],
            ),
            (dictgen::InsensitiveStr::Ascii("decsribe"), &["describe"]),
            (
                dictgen::InsensitiveStr::Ascii("homophibia"),
                &["homophobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("develpoer"), &["developer"]),
            (
                dictgen::InsensitiveStr::Ascii("accoridngly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignested"), &["ingested"]),
            (dictgen::InsensitiveStr::Ascii("correcton"), &["correction"]),
            (dictgen::InsensitiveStr::Ascii("ddivision"), &["division"]),
            (
                dictgen::InsensitiveStr::Ascii("customziable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("confuzzed"), &["confused"]),
            (dictgen::InsensitiveStr::Ascii("updatete"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("machienry"), &["machinery"]),
            (dictgen::InsensitiveStr::Ascii("forwads"), &["forwards"]),
            (dictgen::InsensitiveStr::Ascii("biologia"), &["biological"]),
            (dictgen::InsensitiveStr::Ascii("commmitter"), &["committer"]),
            (
                dictgen::InsensitiveStr::Ascii("individuati"),
                &["individuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremetrically"),
                &["parametrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradictin"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philisopher"),
                &["philosopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracteristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("descized"), &["disguised"]),
            (dictgen::InsensitiveStr::Ascii("cthluhu"), &["cthulhu"]),
            (dictgen::InsensitiveStr::Ascii("creeated"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("impactes"), &["impacts"]),
            (
                dictgen::InsensitiveStr::Ascii("asyncronous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("birgading"), &["brigading"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosincratic"),
                &["idiosyncratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("closue"), &["closure"]),
            (dictgen::InsensitiveStr::Ascii("argumement"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("comected"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("hundread"), &["hundred"]),
            (dictgen::InsensitiveStr::Ascii("recogizer"), &["recognizer"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregatin"),
                &["aggregating", "aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("plaftorms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("notifictation"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individuels"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("actualpy"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpretations"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrospection"),
                &["introspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("feministisk"),
                &["feminists"],
            ),
            (dictgen::InsensitiveStr::Ascii("clossely"), &["closely"]),
            (
                dictgen::InsensitiveStr::Ascii("pepperocini"),
                &["pepperoni"],
            ),
            (dictgen::InsensitiveStr::Ascii("notities"), &["notifies"]),
            (
                dictgen::InsensitiveStr::Ascii("cigarrettes"),
                &["cigarettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecountered"),
                &["encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("volcando"), &["volcano"]),
            (
                dictgen::InsensitiveStr::Ascii("intstructed"),
                &["instructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clearification"),
                &["clarification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concusions"),
                &["concussion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confugurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accessibiliy"),
                &["accessibility", "accessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superceded"),
                &["superseded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortune"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archictectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("ssudo"), &["sudo"]),
            (
                dictgen::InsensitiveStr::Ascii("sutable"),
                &["suitable", "stable"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreieved"), &["retrieved"]),
            (dictgen::InsensitiveStr::Ascii("chaecksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("legitemacy"),
                &["legitimacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlinse"), &["inlines"]),
            (dictgen::InsensitiveStr::Ascii("tamplates"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("configuaration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("ryour"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("specialisin"),
                &["specialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("suseeding"), &["seceding"]),
            (dictgen::InsensitiveStr::Ascii("comande"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("virtialized"),
                &["virtualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainling"), &["mailing"]),
            (dictgen::InsensitiveStr::Ascii("beacause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("remainst"), &["remains"]),
            (
                dictgen::InsensitiveStr::Ascii("conglaturation"),
                &["congratulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("fasicsts"), &["fascists"]),
            (
                dictgen::InsensitiveStr::Ascii("proselytizin"),
                &["proselytizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("implemenrt"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("evaulation"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chiop"), &["chip", "chop"]),
            (dictgen::InsensitiveStr::Ascii("closeely"), &["closely"]),
            (dictgen::InsensitiveStr::Ascii("tempelated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("languales"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("ligitamassy"),
                &["legitimacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schizophernia"),
                &["schizophrenia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("addrerssed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("penitum"), &["pentium"]),
            (
                dictgen::InsensitiveStr::Ascii("plaestinian"),
                &["palestinian"],
            ),
            (dictgen::InsensitiveStr::Ascii("ophan"), &["orphan"]),
            (dictgen::InsensitiveStr::Ascii("dinasours"), &["dinosaurs"]),
            (
                dictgen::InsensitiveStr::Ascii("associateive"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invlunerable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosophics"),
                &["philosophies"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabal"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("checkponts"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("astrisk"), &["asterisk"]),
            (dictgen::InsensitiveStr::Ascii("prjecting"), &["projecting"]),
            (dictgen::InsensitiveStr::Ascii("improvd"), &["improved"]),
            (
                dictgen::InsensitiveStr::Ascii("modles"),
                &["models", "modules"],
            ),
            (dictgen::InsensitiveStr::Ascii("storeable"), &["storable"]),
            (dictgen::InsensitiveStr::Ascii("rebuilded"), &["rebuilt"]),
            (dictgen::InsensitiveStr::Ascii("singluar"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("chanching"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("explonation"),
                &["exploration"],
            ),
            (dictgen::InsensitiveStr::Ascii("trunction"), &["truncation"]),
            (
                dictgen::InsensitiveStr::Ascii("distingish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("uplaoding"), &["uploading"]),
            (dictgen::InsensitiveStr::Ascii("prallel"), &["parallel"]),
            (dictgen::InsensitiveStr::Ascii("examplar"), &["exemplar"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasmitted"),
                &["transmitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsreams"), &["upstreams"]),
            (dictgen::InsensitiveStr::Ascii("instanbul"), &["istanbul"]),
            (dictgen::InsensitiveStr::Ascii("mutatin"), &["mutation"]),
            (dictgen::InsensitiveStr::Ascii("separtate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("simualtors"),
                &["simulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("forrmatter"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("phsyically"),
                &["physically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonwithstanding"),
                &["notwithstanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("promissed"), &["promised"]),
            (
                dictgen::InsensitiveStr::Ascii("wrokarounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detetion"),
                &["detection", "deletion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speace"),
                &["peace", "space"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeury"), &["query"]),
            (dictgen::InsensitiveStr::Ascii("oftenly"), &["often"]),
            (
                dictgen::InsensitiveStr::Ascii("workarunds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("restrored"), &["restored"]),
            (
                dictgen::InsensitiveStr::Ascii("unspeficialleid"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcitng"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("infite"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrievated"),
                &["abbreviated"],
            ),
            (dictgen::InsensitiveStr::Ascii("heade"), &["head", "header"]),
            (dictgen::InsensitiveStr::Ascii("referants"), &["referents"]),
            (dictgen::InsensitiveStr::Ascii("anarchsim"), &["anarchism"]),
            (dictgen::InsensitiveStr::Ascii("vaccuum"), &["vacuum"]),
            (
                dictgen::InsensitiveStr::Ascii("branc"),
                &["branch", "brand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclucivity"),
                &["exclusivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpectad"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("falseley"), &["falsely"]),
            (
                dictgen::InsensitiveStr::Ascii("volleyboll"),
                &["volleyball"],
            ),
            (dictgen::InsensitiveStr::Ascii("daybue"), &["debut"]),
            (
                dictgen::InsensitiveStr::Ascii("comppatible"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("navigatie"), &["navigate"]),
            (
                dictgen::InsensitiveStr::Ascii("augumenting"),
                &["augmenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("swicthed"), &["switched"]),
            (
                dictgen::InsensitiveStr::Ascii("expansivos"),
                &["expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particiapted"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("invocies"), &["invoices"]),
            (dictgen::InsensitiveStr::Ascii("calculdate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("somme"), &["some"]),
            (
                dictgen::InsensitiveStr::Ascii("fumctionally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiatiats"),
                &["initiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficitic"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("logisticly"),
                &["logistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("hsitorian"), &["historian"]),
            (
                dictgen::InsensitiveStr::Ascii("kenel"),
                &["kernel", "kennel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agricultre"),
                &["agriculture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heavyweght"),
                &["heavyweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commentar"),
                &["commentator"],
            ),
            (dictgen::InsensitiveStr::Ascii("betcause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("suffuxes"), &["suffixes"]),
            (
                dictgen::InsensitiveStr::Ascii("syllabills"),
                &["syllabus", "syllabification"],
            ),
            (dictgen::InsensitiveStr::Ascii("vigilanty"), &["vigilante"]),
            (dictgen::InsensitiveStr::Ascii("ueries"), &["queries"]),
            (dictgen::InsensitiveStr::Ascii("bigfooot"), &["bigfoot"]),
            (
                dictgen::InsensitiveStr::Ascii("malfonction"),
                &["malfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("petroluem"), &["petroleum"]),
            (
                dictgen::InsensitiveStr::Ascii("differenciating"),
                &["differentiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repatwars"),
                &["repertoires"],
            ),
            (dictgen::InsensitiveStr::Ascii("swatiska"), &["swastika"]),
            (
                dictgen::InsensitiveStr::Ascii("sociopati"),
                &["sociopathic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perhiperal"),
                &["peripheral"],
            ),
            (dictgen::InsensitiveStr::Ascii("brewrey"), &["brewery"]),
            (dictgen::InsensitiveStr::Ascii("actualey"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("profund"), &["profound"]),
            (
                dictgen::InsensitiveStr::Ascii("substititions"),
                &["substitutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discogrophy"),
                &["discography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("desingned"), &["designed"]),
            (
                dictgen::InsensitiveStr::Ascii("appeareance"),
                &["appearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eductions"),
                &["reductions", "deductions", "educations", "seductions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fashoinable"),
                &["fashionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("soource"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("overzealosly"),
                &["overzealously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lavel"),
                &["level", "laravel", "label"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlargments"),
                &["enlargements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desscription"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("follfowind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("asyncronus"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catagorically"),
                &["categorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("timouts"), &["timeouts"]),
            (dictgen::InsensitiveStr::Ascii("chidlren"), &["children"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("increated"), &["increased"]),
            (
                dictgen::InsensitiveStr::Ascii("intializtion"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("dyanmic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("mandarion"), &["mandarin"]),
            (dictgen::InsensitiveStr::Ascii("underbog"), &["undergo"]),
            (dictgen::InsensitiveStr::Ascii("estonija"), &["estonia"]),
            (dictgen::InsensitiveStr::Ascii("averagin"), &["averaging"]),
            (
                dictgen::InsensitiveStr::Ascii("recontruction"),
                &["reconstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcetory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("identifiees"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginizations"),
                &["organizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reacursive"), &["recursive"]),
            (dictgen::InsensitiveStr::Ascii("kayout"), &["layout"]),
            (
                dictgen::InsensitiveStr::Ascii("intoxination"),
                &["intoxication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countrs"),
                &["counters", "contours", "countries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transgeder"),
                &["transgender"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implentations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tarcebacks"),
                &["tracebacks"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreegate"), &["aggregate"]),
            (
                dictgen::InsensitiveStr::Ascii("babysittter"),
                &["babysitter"],
            ),
            (dictgen::InsensitiveStr::Ascii("varriables"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("wherether"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("verchews"), &["virtues"]),
            (
                dictgen::InsensitiveStr::Ascii("delopyment"),
                &["deployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assestment"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misspeeling"),
                &["misspelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("foto"), &["photo"]),
            (dictgen::InsensitiveStr::Ascii("ohters"), &["others"]),
            (
                dictgen::InsensitiveStr::Ascii("deprecting"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("shitlasses"), &["shitless"]),
            (
                dictgen::InsensitiveStr::Ascii("prejections"),
                &["projections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captabilities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("spoitfy"), &["spotify"]),
            (
                dictgen::InsensitiveStr::Ascii("adaptating"),
                &["adaptation", "adapting"],
            ),
            (dictgen::InsensitiveStr::Ascii("spawed"), &["spawned"]),
            (dictgen::InsensitiveStr::Ascii("isladn"), &["island"]),
            (
                dictgen::InsensitiveStr::Ascii("commenes"),
                &["comments", "commends", "commences", "communes"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfomes"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("kimera"), &["chimera"]),
            (
                dictgen::InsensitiveStr::Ascii("participacion"),
                &["participation"],
            ),
            (dictgen::InsensitiveStr::Ascii("patronos"), &["patrons"]),
            (dictgen::InsensitiveStr::Ascii("droped"), &["dropped"]),
            (
                dictgen::InsensitiveStr::Ascii("meassage"),
                &["message", "massage"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodused"), &["produced"]),
            (dictgen::InsensitiveStr::Ascii("cilpboard"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropraite"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("acronysm"), &["acronyms"]),
            (dictgen::InsensitiveStr::Ascii("reuplod"), &["reupload"]),
            (dictgen::InsensitiveStr::Ascii("exectes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("succefuly"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("chizelle"), &["chisel"]),
            (dictgen::InsensitiveStr::Ascii("diasble"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("bypased"), &["bypassed"]),
            (
                dictgen::InsensitiveStr::Ascii("defned"),
                &["defend", "defined"],
            ),
            (dictgen::InsensitiveStr::Ascii("labeledby"), &["labelledby"]),
            (dictgen::InsensitiveStr::Ascii("furthe"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("nigthmares"),
                &["nightmares"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proficeint"),
                &["proficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooresponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshk"), &["handshake"]),
            (
                dictgen::InsensitiveStr::Ascii("prototipes"),
                &["prototypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrleation"),
                &["correlation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integreres"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finishe"),
                &["finished", "finish"],
            ),
            (dictgen::InsensitiveStr::Ascii("anomoly"), &["anomaly"]),
            (
                dictgen::InsensitiveStr::Ascii("cluters"),
                &["clusters", "clutters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customsiing"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("hesistates"), &["hesitates"]),
            (dictgen::InsensitiveStr::Ascii("numbrs"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhemlingly"),
                &["overwhelmingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("someimes"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("implemataions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("itselt"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("repulic"), &["republic"]),
            (
                dictgen::InsensitiveStr::Ascii("misrepresentating"),
                &["misrepresenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppenent"), &["opponent"]),
            (dictgen::InsensitiveStr::Ascii("doller"), &["dollar"]),
            (
                dictgen::InsensitiveStr::Ascii("consilation"),
                &["consolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dislpaying"),
                &["displaying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exlcusions"),
                &["exclusions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelertor"),
                &["accelerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accountat"),
                &["accountants", "accountant"],
            ),
            (dictgen::InsensitiveStr::Ascii("furure"), &["future"]),
            (dictgen::InsensitiveStr::Ascii("succeeeded"), &["succeeded"]),
            (dictgen::InsensitiveStr::Ascii("abandonned"), &["abandoned"]),
            (
                dictgen::InsensitiveStr::Ascii("atrosities"),
                &["atrocities"],
            ),
            (dictgen::InsensitiveStr::Ascii("depenence"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactureres"),
                &["manufacturers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializion"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sponsered"), &["sponsored"]),
            (dictgen::InsensitiveStr::Ascii("inplicit"), &["implicit"]),
            (
                dictgen::InsensitiveStr::Ascii("philippenes"),
                &["philippines"],
            ),
            (dictgen::InsensitiveStr::Ascii("contraint"), &["constraint"]),
            (dictgen::InsensitiveStr::Ascii("pesduo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("officianado"),
                &["aficionado"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanitary"),
                &["explanatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("archines"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("confiugrations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemnet"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("aphabetically"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sycronous"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autodected"),
                &["autodetected"],
            ),
            (dictgen::InsensitiveStr::Ascii("pssed"), &["passed"]),
            (
                dictgen::InsensitiveStr::Ascii("problomatic"),
                &["problematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("exploint"), &["exploit"]),
            (dictgen::InsensitiveStr::Ascii("experice"), &["experience"]),
            (dictgen::InsensitiveStr::Ascii("floruish"), &["flourish"]),
            (dictgen::InsensitiveStr::Ascii("preventin"), &["prevention"]),
            (dictgen::InsensitiveStr::Ascii("redcue"), &["reduce"]),
            (
                dictgen::InsensitiveStr::Ascii("quiestionnaires"),
                &["questionnaires"],
            ),
            (dictgen::InsensitiveStr::Ascii("ongoin"), &["ongoing"]),
            (
                dictgen::InsensitiveStr::Ascii("inteligable"),
                &["intelligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("garriosn"), &["garrison"]),
            (
                dictgen::InsensitiveStr::Ascii("cryptographc"),
                &["cryptographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("irradate"), &["irradiate"]),
            (
                dictgen::InsensitiveStr::Ascii("friendboned"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("integere"), &["integer"]),
            (
                dictgen::InsensitiveStr::Ascii("condfigurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequnce"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("tranluscent"),
                &["translucent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availebilities"),
                &["availabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertions"), &["operations"]),
            (dictgen::InsensitiveStr::Ascii("struture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("supposedely"),
                &["supposedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("regresson"), &["regression"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnferring"),
                &["transferring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliverees"),
                &["deliveries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disasemble"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspicioso"),
                &["suspicions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertires"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentiomenter"),
                &["potentiometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validatins"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosiphies"),
                &["philosophies"],
            ),
            (dictgen::InsensitiveStr::Ascii("conuntry"), &["country"]),
            (
                dictgen::InsensitiveStr::Ascii("encylopedia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciefying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parm"),
                &["param", "pram", "parma"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expores"),
                &["explores", "expires", "exports", "exposures", "exposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stoed"),
                &["stored", "stood"],
            ),
            (dictgen::InsensitiveStr::Ascii("caffine"), &["caffeine"]),
            (dictgen::InsensitiveStr::Ascii("picthed"), &["pitched"]),
            (dictgen::InsensitiveStr::Ascii("timne"), &["time"]),
            (
                dictgen::InsensitiveStr::Ascii("conflcting"),
                &["conflating"],
            ),
            (dictgen::InsensitiveStr::Ascii("relocaes"), &["relocates"]),
            (
                dictgen::InsensitiveStr::Ascii("acccessible"),
                &["accessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("removs"), &["removes"]),
            (
                dictgen::InsensitiveStr::Ascii("analysised"),
                &["analyses", "analysed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumizing"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iterpolates"),
                &["interpolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refelectivity"),
                &["reflectivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpresonal"),
                &["interpersonal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppotunities"),
                &["opportunities"],
            ),
            (dictgen::InsensitiveStr::Ascii("assfalted"), &["asphalted"]),
            (
                dictgen::InsensitiveStr::Ascii("attributess"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classsified"),
                &["classified"],
            ),
            (dictgen::InsensitiveStr::Ascii("balaced"), &["balanced"]),
            (dictgen::InsensitiveStr::Ascii("othres"), &["others"]),
            (dictgen::InsensitiveStr::Ascii("nofities"), &["notifies"]),
            (
                dictgen::InsensitiveStr::Ascii("inicializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("actucally"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("disaples"), &["disables"]),
            (dictgen::InsensitiveStr::Ascii("meaure"), &["measure"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuccees"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothenuse"),
                &["hypotenuse"],
            ),
            (dictgen::InsensitiveStr::Ascii("yot"), &["yacht"]),
            (dictgen::InsensitiveStr::Ascii("questiosn"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("shletered"), &["sheltered"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpirienced"),
                &["inexperienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagd"), &["messaged"]),
            (dictgen::InsensitiveStr::Ascii("colombus"), &["columbus"]),
            (dictgen::InsensitiveStr::Ascii("jamaicains"), &["jamaican"]),
            (dictgen::InsensitiveStr::Ascii("conector"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("tournamnet"),
                &["tournament"],
            ),
            (dictgen::InsensitiveStr::Ascii("masia"), &["messiah"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatatible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("drob"), &["drop"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetned"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounday"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("preconcieve"),
                &["preconceived"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infromatoinal"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("pronoucned"),
                &["pronounced"],
            ),
            (dictgen::InsensitiveStr::Ascii("beileve"), &["believe"]),
            (
                dictgen::InsensitiveStr::Ascii("psychedilic"),
                &["psychedelic"],
            ),
            (dictgen::InsensitiveStr::Ascii("penaltys"), &["penalties"]),
            (
                dictgen::InsensitiveStr::Ascii("comfortabley"),
                &["comfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubutions"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inevitablity"),
                &["inevitably"],
            ),
            (dictgen::InsensitiveStr::Ascii("contexual"), &["contextual"]),
            (dictgen::InsensitiveStr::Ascii("techincal"), &["technical"]),
            (dictgen::InsensitiveStr::Ascii("regularlas"), &["regulars"]),
            (
                dictgen::InsensitiveStr::Ascii("authencticating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramboline"),
                &["trampoline"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrvieved"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("priviledged"),
                &["privileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placehoder"),
                &["placeholder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compitible"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degenration"),
                &["degeneration"],
            ),
            (dictgen::InsensitiveStr::Ascii("rationalle"), &["rationale"]),
            (dictgen::InsensitiveStr::Ascii("trucates"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("systms"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitutos"),
                &["prostitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invcocation"),
                &["invocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nauesous"), &["nauseous"]),
            (dictgen::InsensitiveStr::Ascii("generlize"), &["generalize"]),
            (dictgen::InsensitiveStr::Ascii("gti"), &["git"]),
            (dictgen::InsensitiveStr::Ascii("bicast"), &["bitcast"]),
            (dictgen::InsensitiveStr::Ascii("paticles"), &["particles"]),
            (
                dictgen::InsensitiveStr::Ascii("replacacing"),
                &["replacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stetch"),
                &["stretch", "sketch", "stitch", "stench"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("osicllations"),
                &["oscillations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beahvioural"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("vailidator"), &["validator"]),
            (
                dictgen::InsensitiveStr::Ascii("citicenship"),
                &["citizenship"],
            ),
            (dictgen::InsensitiveStr::Ascii("upstiars"), &["upstairs"]),
            (dictgen::InsensitiveStr::Ascii("bulding"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("arbiturarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poen"),
                &["open", "pen", "poem", "phone"],
            ),
            (dictgen::InsensitiveStr::Ascii("balckjack"), &["blackjack"]),
            (dictgen::InsensitiveStr::Ascii("sloted"), &["slotted"]),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedran"),
                &["tetrahedron"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaative"), &["creative"]),
            (dictgen::InsensitiveStr::Ascii("seuqence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("eamcs"), &["emacs"]),
            (dictgen::InsensitiveStr::Ascii("definitv"), &["definitive"]),
            (dictgen::InsensitiveStr::Ascii("globaly"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("serialzize"), &["serialize"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriaton"),
                &["appropriation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememnts"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("deattach"), &["detach"]),
            (
                dictgen::InsensitiveStr::Ascii("obligerated"),
                &["obliterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumens"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("availabelity"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("mislabled"), &["mislabeled"]),
            (
                dictgen::InsensitiveStr::Ascii("refromists"),
                &["reformists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worhsiping"),
                &["worshiping"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitaution"), &["situation"]),
            (dictgen::InsensitiveStr::Ascii("deaclock"), &["deadlock"]),
            (
                dictgen::InsensitiveStr::Ascii("missunderstand"),
                &["misunderstand"],
            ),
            (dictgen::InsensitiveStr::Ascii("basides"), &["besides"]),
            (
                dictgen::InsensitiveStr::Ascii("unbelievabley"),
                &["unbelievably"],
            ),
            (dictgen::InsensitiveStr::Ascii("scartches"), &["scratches"]),
            (dictgen::InsensitiveStr::Ascii("persistes"), &["persists"]),
            (
                dictgen::InsensitiveStr::Ascii("reolves"),
                &["resolves", "revolves"],
            ),
            (dictgen::InsensitiveStr::Ascii("procces"), &["process"]),
            (dictgen::InsensitiveStr::Ascii("expetcted"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalizes"),
                &["orthogonalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("polotics"), &["politics"]),
            (
                dictgen::InsensitiveStr::Ascii("catastrophies"),
                &["catastrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("lithunaia"), &["lithuania"]),
            (dictgen::InsensitiveStr::Ascii("savely"), &["safely"]),
            (dictgen::InsensitiveStr::Ascii("methode"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("constributed"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explainging"),
                &["explaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("outut"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("derefences"),
                &["dereferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("diretly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("vegtables"), &["vegetables"]),
            (dictgen::InsensitiveStr::Ascii("harcoded"), &["hardcoded"]),
            (
                dictgen::InsensitiveStr::Ascii("somethinmg"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("buttonn"), &["button"]),
            (
                dictgen::InsensitiveStr::Ascii("unnescessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumking"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("fractalus"), &["fractals"]),
            (
                dictgen::InsensitiveStr::Ascii("monontonicity"),
                &["monotonicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("rearrangte"), &["rearrange"]),
            (
                dictgen::InsensitiveStr::Ascii("inacessible"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approrpriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccomendation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destanation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("murdererous"),
                &["murderers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colletions"),
                &["collections"],
            ),
            (dictgen::InsensitiveStr::Ascii("unque"), &["unique"]),
            (dictgen::InsensitiveStr::Ascii("parth"), &["path"]),
            (
                dictgen::InsensitiveStr::Ascii("rqeuest"),
                &["request", "quest"],
            ),
            (dictgen::InsensitiveStr::Ascii("increas"), &["increase"]),
            (dictgen::InsensitiveStr::Ascii("thirtsy"), &["thirsty"]),
            (
                dictgen::InsensitiveStr::Ascii("discualified"),
                &["disqualified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissistc"),
                &["narcissistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impecabbly"),
                &["impeccably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readibility"),
                &["readability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vizualizations"),
                &["visualizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aded"), &["added"]),
            (
                dictgen::InsensitiveStr::Ascii("activiites"),
                &["activist", "activities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorreectly"),
                &["incorrectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defail"),
                &["detail", "derail"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surroundigs"),
                &["surrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("tanturms"), &["tantrums"]),
            (
                dictgen::InsensitiveStr::Ascii("strugglebus"),
                &["struggles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affinitized"),
                &["affinities", "affinity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenteries"),
                &["documentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("acsii"), &["ascii"]),
            (dictgen::InsensitiveStr::Ascii("clen"), &["clan", "clean"]),
            (
                dictgen::InsensitiveStr::Ascii("consdierations"),
                &["considerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appllicable"),
                &["applicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranasctions"),
                &["transactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("avergaed"), &["averaged"]),
            (dictgen::InsensitiveStr::Ascii("inate"), &["innate"]),
            (dictgen::InsensitiveStr::Ascii("shourtcur"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("desables"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("complusions"),
                &["compulsion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comepndium"),
                &["compendium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditonal"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("paradim"), &["paradigm"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluationg"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicitably"),
                &["predictably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("departmnet"),
                &["departments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delegater"),
                &["delegator", "delegated", "delegates", "delegate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contolling"),
                &["controlling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cellebrations"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("spilnter"), &["splinter"]),
            (dictgen::InsensitiveStr::Ascii("macasons"), &["moccasins"]),
            (dictgen::InsensitiveStr::Ascii("againsg"), &["against"]),
            (dictgen::InsensitiveStr::Ascii("anoerxic"), &["anorexic"]),
            (dictgen::InsensitiveStr::Ascii("otherweis"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("haircuit"), &["haircut"]),
            (dictgen::InsensitiveStr::Ascii("omlette"), &["omelette"]),
            (dictgen::InsensitiveStr::Ascii("bitmp"), &["bitmap"]),
            (
                dictgen::InsensitiveStr::Ascii("custumises"),
                &["customises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preventitive"),
                &["preventative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconstituional"),
                &["unconstitutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumenst"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("fnish"), &["finish"]),
            (
                dictgen::InsensitiveStr::Ascii("exercies"),
                &["exercise", "exercises"],
            ),
            (dictgen::InsensitiveStr::Ascii("bargins"), &["bargains"]),
            (dictgen::InsensitiveStr::Ascii("basied"), &["based"]),
            (
                dictgen::InsensitiveStr::Ascii("unconvential"),
                &["unconventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("backsta"), &["backseat"]),
            (dictgen::InsensitiveStr::Ascii("comradre"), &["comrade"]),
            (
                dictgen::InsensitiveStr::Ascii("shortcomming"),
                &["shortcoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("urkainians"),
                &["ukrainians"],
            ),
            (dictgen::InsensitiveStr::Ascii("grabge"), &["garbage"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("revelaed"), &["revealed"]),
            (
                dictgen::InsensitiveStr::Ascii("unreprucible"),
                &["unreproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuratoin"),
                &["configurations", "configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrances"),
                &["instances", "insurances"],
            ),
            (dictgen::InsensitiveStr::Ascii("sempphore"), &["semaphore"]),
            (
                dictgen::InsensitiveStr::Ascii("transgemder"),
                &["transgender"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatte"), &["operate"]),
            (dictgen::InsensitiveStr::Ascii("plarizing"), &["polarizing"]),
            (dictgen::InsensitiveStr::Ascii("uncarefull"), &["uncareful"]),
            (dictgen::InsensitiveStr::Ascii("architec"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("uncler"),
                &["uncle", "unclear", "uncles"],
            ),
            (dictgen::InsensitiveStr::Ascii("realible"), &["reliable"]),
            (
                dictgen::InsensitiveStr::Ascii("charactersitics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("vectore"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("compelted"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("maneagable"),
                &["manageable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transylvanai"),
                &["transylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("collecion"), &["collection"]),
            (
                dictgen::InsensitiveStr::Ascii("implemations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("comedlib"), &["comedilib"]),
            (dictgen::InsensitiveStr::Ascii("candicate"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("thankyu"), &["thankyou"]),
            (dictgen::InsensitiveStr::Ascii("consumin"), &["consuming"]),
            (dictgen::InsensitiveStr::Ascii("continaed"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplied"),
                &["multiplied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mocrophones"),
                &["microphones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanitites"),
                &["quantities"],
            ),
            (dictgen::InsensitiveStr::Ascii("amkes"), &["makes"]),
            (
                dictgen::InsensitiveStr::Ascii("intepreter"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("arogant"), &["arrogant"]),
            (
                dictgen::InsensitiveStr::Ascii("historyans"),
                &["historians"],
            ),
            (dictgen::InsensitiveStr::Ascii("distibute"), &["distribute"]),
            (
                dictgen::InsensitiveStr::Ascii("prograstination"),
                &["procrastination"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckecker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("competitio"),
                &["competitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronus"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("circual"), &["circular"]),
            (dictgen::InsensitiveStr::Ascii("downoad"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("unhilights"),
                &["unhighlights"],
            ),
            (dictgen::InsensitiveStr::Ascii("aresnal"), &["arsenal"]),
            (dictgen::InsensitiveStr::Ascii("soluton"), &["solution"]),
            (dictgen::InsensitiveStr::Ascii("protocoll"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("snowbals"), &["snowballs"]),
            (dictgen::InsensitiveStr::Ascii("natioanls"), &["nationals"]),
            (
                dictgen::InsensitiveStr::Ascii("announcent"),
                &["announcement", "announce"],
            ),
            (dictgen::InsensitiveStr::Ascii("gravitatie"), &["gravitate"]),
            (
                dictgen::InsensitiveStr::Ascii("acheevement"),
                &["achievement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograpical"),
                &["photographical"],
            ),
            (dictgen::InsensitiveStr::Ascii("auotmates"), &["automates"]),
            (dictgen::InsensitiveStr::Ascii("acually"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("intitiator"), &["initiator"]),
            (dictgen::InsensitiveStr::Ascii("autmating"), &["automating"]),
            (
                dictgen::InsensitiveStr::Ascii("dumbfoundeads"),
                &["dumbfounded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramlised"),
                &["normalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("objets"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("purgest"), &["purges"]),
            (
                dictgen::InsensitiveStr::Ascii("gravitaitonal"),
                &["gravitational"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccript"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("dolphines"), &["dolphins"]),
            (
                dictgen::InsensitiveStr::Ascii("dessertation"),
                &["dissertation"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmended"), &["segmented"]),
            (dictgen::InsensitiveStr::Ascii("sequelce"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("twords"), &["towards"]),
            (
                dictgen::InsensitiveStr::Ascii("presidancy"),
                &["presidency"],
            ),
            (dictgen::InsensitiveStr::Ascii("electorn"), &["electron"]),
            (dictgen::InsensitiveStr::Ascii("consensul"), &["consensual"]),
            (
                dictgen::InsensitiveStr::Ascii("compressio"),
                &["compressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspontent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reniassance"),
                &["renaissance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoziator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilaton"),
                &["compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretated"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupilcates"),
                &["duplicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execptions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("alhpabet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("homogeneos"),
                &["homogeneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roducing"),
                &["reducing", "producing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relativily"),
                &["relativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceppters"), &["accepters"]),
            (
                dictgen::InsensitiveStr::Ascii("anihilated"),
                &["annihilated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analystics"),
                &["analysts", "analytics"],
            ),
            (dictgen::InsensitiveStr::Ascii("equialent"), &["equivalent"]),
            (
                dictgen::InsensitiveStr::Ascii("transistioned"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exlcusivity"),
                &["exclusivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomanon"),
                &["phenomenon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introdused"),
                &["introduces", "introduced"],
            ),
            (dictgen::InsensitiveStr::Ascii("exptects"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsiter"),
                &["administer"],
            ),
            (dictgen::InsensitiveStr::Ascii("varibales"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("resilency"),
                &["resiliency", "residency"],
            ),
            (dictgen::InsensitiveStr::Ascii("preselets"), &["preselects"]),
            (dictgen::InsensitiveStr::Ascii("resoources"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("tosday"),
                &["today", "tuesday"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generaton"),
                &["generation", "generator"],
            ),
            (dictgen::InsensitiveStr::Ascii("devfined"), &["defined"]),
            (
                dictgen::InsensitiveStr::Ascii("sturctural"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expored"),
                &["explored", "exported", "expired", "exposed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suttle"),
                &["subtle", "shuttle"],
            ),
            (dictgen::InsensitiveStr::Ascii("characther"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("sacles"),
                &["scales", "sales", "sables"],
            ),
            (dictgen::InsensitiveStr::Ascii("possibel"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("reduntancy"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("unnable"), &["unable"]),
            (dictgen::InsensitiveStr::Ascii("shapshot"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("competend"), &["competed"]),
            (dictgen::InsensitiveStr::Ascii("tomorrwo"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("writtten"), &["written"]),
            (dictgen::InsensitiveStr::Ascii("orkid"), &["orchid"]),
            (dictgen::InsensitiveStr::Ascii("exectues"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("anount"), &["amount"]),
            (
                dictgen::InsensitiveStr::Ascii("guarantted"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catergorical"),
                &["categorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("boungary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("outperfoming"),
                &["outperforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializong"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disclipinary"),
                &["disciplinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("quantites"), &["quantities"]),
            (dictgen::InsensitiveStr::Ascii("enforrced"), &["enforced"]),
            (
                dictgen::InsensitiveStr::Ascii("kintergarten"),
                &["kindergarten"],
            ),
            (dictgen::InsensitiveStr::Ascii("smewhere"), &["somewhere"]),
            (dictgen::InsensitiveStr::Ascii("continuem"), &["continuum"]),
            (
                dictgen::InsensitiveStr::Ascii("indefineable"),
                &["undefinable"],
            ),
            (dictgen::InsensitiveStr::Ascii("virutally"), &["virtually"]),
            (
                dictgen::InsensitiveStr::Ascii("motherbaord"),
                &["motherboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confifure"),
                &["configure", "confiture"],
            ),
            (dictgen::InsensitiveStr::Ascii("turbins"), &["turbines"]),
            (
                dictgen::InsensitiveStr::Ascii("permuating"),
                &["permutating", "permuting"],
            ),
            (dictgen::InsensitiveStr::Ascii("varadic"), &["variadic"]),
            (
                dictgen::InsensitiveStr::Ascii("unconcistency"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("esgers"), &["edgers"]),
            (
                dictgen::InsensitiveStr::Ascii("represntation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("zoinist"), &["zionist"]),
            (dictgen::InsensitiveStr::Ascii("aviods"), &["avoids"]),
            (dictgen::InsensitiveStr::Ascii("missles"), &["missiles"]),
            (dictgen::InsensitiveStr::Ascii("inilne"), &["inline"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasociate"),
                &["disassociate"],
            ),
            (dictgen::InsensitiveStr::Ascii("createdd"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("compoents"), &["components"]),
            (dictgen::InsensitiveStr::Ascii("usualy"), &["usually"]),
            (dictgen::InsensitiveStr::Ascii("anlyzes"), &["analyzes"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvenietly"),
                &["inconveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("netwon"), &["newton"]),
            (
                dictgen::InsensitiveStr::Ascii("ropository"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differense"),
                &["differences", "difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("acciental"), &["accidental"]),
            (
                dictgen::InsensitiveStr::Ascii("exciptions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tabluates"), &["tabulates"]),
            (dictgen::InsensitiveStr::Ascii("donwvoted"), &["downvoted"]),
            (
                dictgen::InsensitiveStr::Ascii("completeion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("receptionst"),
                &["receptionist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advisoriy"),
                &["advisory", "advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("ascneder"), &["ascender"]),
            (
                dictgen::InsensitiveStr::Ascii("excecutables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissrupting"),
                &["disrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinsitic"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("negoable"), &["negotiable"]),
            (dictgen::InsensitiveStr::Ascii("militiades"), &["militias"]),
            (dictgen::InsensitiveStr::Ascii("sproutes"), &["sprouts"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangde"),
                &["rearranged"],
            ),
            (dictgen::InsensitiveStr::Ascii("separtator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("reguarlly"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("otimiser"), &["optimiser"]),
            (dictgen::InsensitiveStr::Ascii("proivde"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("spoaced"), &["spaced"]),
            (
                dictgen::InsensitiveStr::Ascii("percievable"),
                &["perceivable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lighning"),
                &["lightning", "lighting"],
            ),
            (dictgen::InsensitiveStr::Ascii("benifit"), &["benefit"]),
            (dictgen::InsensitiveStr::Ascii("macroses"), &["macros"]),
            (dictgen::InsensitiveStr::Ascii("testsed"), &["tested"]),
            (
                dictgen::InsensitiveStr::Ascii("discontinuos"),
                &["discontinuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("decifits"), &["deficits"]),
            (dictgen::InsensitiveStr::Ascii("gourmelt"), &["gourmet"]),
            (dictgen::InsensitiveStr::Ascii("promoteurs"), &["promotes"]),
            (
                dictgen::InsensitiveStr::Ascii("phychologist"),
                &["psychologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marganilized"),
                &["marginalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriopting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("accessibile"),
                &["accessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gongratulations"),
                &["congratulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("alchohol"), &["alcohol"]),
            (
                dictgen::InsensitiveStr::Ascii("essentailly"),
                &["essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrington"), &["arlington"]),
            (dictgen::InsensitiveStr::Ascii("createing"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("alltogeher"),
                &["altogether"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcoks"), &["docks"]),
            (
                dictgen::InsensitiveStr::Ascii("phycisists"),
                &["physicists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blurr"),
                &["blur", "blurred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtualiation"),
                &["virtualisation", "virtualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("deveoping"), &["developing"]),
            (
                dictgen::InsensitiveStr::Ascii("furstration"),
                &["frustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("languare"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("apauled"), &["appalled"]),
            (
                dictgen::InsensitiveStr::Ascii("dicitonaries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parenthesies"),
                &["parenthesis", "parentheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("primives"), &["primitives"]),
            (dictgen::InsensitiveStr::Ascii("upagrading"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("agancies"), &["agencies"]),
            (dictgen::InsensitiveStr::Ascii("absuolate"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("ceckers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("nreferenced"),
                &["unreferenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esponsibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("refences"), &["references"]),
            (
                dictgen::InsensitiveStr::Ascii("medicaiton"),
                &["medication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoinding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continenets"),
                &["continents"],
            ),
            (dictgen::InsensitiveStr::Ascii("saame"), &["same"]),
            (
                dictgen::InsensitiveStr::Ascii("indocrtinated"),
                &["indoctrinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("scenearios"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("formuls"), &["formulas"]),
            (dictgen::InsensitiveStr::Ascii("corrctly"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("photoshoppad"),
                &["photoshopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptionm"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsnks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("explictly"), &["explicitly"]),
            (dictgen::InsensitiveStr::Ascii("assync"), &["async"]),
            (
                dictgen::InsensitiveStr::Ascii("complatness"),
                &["completeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstading"),
                &["misunderstanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensiable"),
                &["extensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotioable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harranger"),
                &["arranger", "haranguer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recevers"),
                &["receivers", "recovers"],
            ),
            (dictgen::InsensitiveStr::Ascii("withouyt"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("dyspohria"), &["dysphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("exerimenters"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stakeboarding"),
                &["skateboarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoprts"), &["reports"]),
            (
                dictgen::InsensitiveStr::Ascii("depndently"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("followins"),
                &["following", "followings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propechies"),
                &["prophecies"],
            ),
            (dictgen::InsensitiveStr::Ascii("precsion"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("lightneing"),
                &["lightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassocation"),
                &["disassociation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duetschland"),
                &["deutschland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrutions"),
                &["instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scatchss"),
                &["catches", "scratches", "sketches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconcistencies"),
                &["inconsistencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("priortize"), &["prioritize"]),
            (
                dictgen::InsensitiveStr::Ascii("architechural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemts"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("interstaller"),
                &["interstellar"],
            ),
            (dictgen::InsensitiveStr::Ascii("ojbective"), &["objective"]),
            (
                dictgen::InsensitiveStr::Ascii("conteining"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("userames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("restauring"),
                &["restarting"],
            ),
            (dictgen::InsensitiveStr::Ascii("itereation"), &["iteration"]),
            (dictgen::InsensitiveStr::Ascii("reflecters"), &["reflects"]),
            (dictgen::InsensitiveStr::Ascii("melieux"), &["milieux"]),
            (dictgen::InsensitiveStr::Ascii("ciruits"), &["circuits"]),
            (dictgen::InsensitiveStr::Ascii("epectedly"), &["expectedly"]),
            (dictgen::InsensitiveStr::Ascii("declarase"), &["declares"]),
            (dictgen::InsensitiveStr::Ascii("cusstomers"), &["customers"]),
            (
                dictgen::InsensitiveStr::Ascii("bidrectional"),
                &["bidirectional"],
            ),
            (dictgen::InsensitiveStr::Ascii("felds"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("fortunantly"),
                &["fortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsusbtantiated"),
                &["unsubstantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadimia"), &["academia"]),
            (dictgen::InsensitiveStr::Ascii("warings"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("screem"),
                &["scream", "screen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aditionaly"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deccelerates"),
                &["decelerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("pregress"), &["progress"]),
            (
                dictgen::InsensitiveStr::Ascii("insersected"),
                &["intersected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecouraging"),
                &["encouraging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterazation"),
                &["characterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("destrose"), &["destroyers"]),
            (dictgen::InsensitiveStr::Ascii("weridness"), &["weirdness"]),
            (
                dictgen::InsensitiveStr::Ascii("referancing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("joepardy"), &["jeopardy"]),
            (
                dictgen::InsensitiveStr::Ascii("hightly"),
                &["highly", "nightly"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliamtes"), &["climates"]),
            (
                dictgen::InsensitiveStr::Ascii("instinctivley"),
                &["instinctively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shardholders"),
                &["shareholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sceme"),
                &["scheme", "scene"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificateion"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("mecanical"), &["mechanical"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthrouh"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("penguiness"), &["penguins"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensinoal"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotion"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongised"),
                &["recognised"],
            ),
            (dictgen::InsensitiveStr::Ascii("disapears"), &["disappears"]),
            (dictgen::InsensitiveStr::Ascii("diasabled"), &["disabled"]),
            (
                dictgen::InsensitiveStr::Ascii("trespessing"),
                &["trespassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("commuable"), &["commutable"]),
            (dictgen::InsensitiveStr::Ascii("contemt"), &["contempt"]),
            (
                dictgen::InsensitiveStr::Ascii("uncessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("toubled"), &["troubled"]),
            (dictgen::InsensitiveStr::Ascii("asistants"), &["assistants"]),
            (dictgen::InsensitiveStr::Ascii("unaproved"), &["unapproved"]),
            (dictgen::InsensitiveStr::Ascii("direcot"), &["direct"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiased"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("socialy"), &["socially"]),
            (dictgen::InsensitiveStr::Ascii("hree"), &["here", "three"]),
            (dictgen::InsensitiveStr::Ascii("execuation"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("corvering"), &["covering"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnactions"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("penatentury"),
                &["penitentiary"],
            ),
            (dictgen::InsensitiveStr::Ascii("depnded"), &["depended"]),
            (dictgen::InsensitiveStr::Ascii("broard"), &["broad"]),
            (dictgen::InsensitiveStr::Ascii("critisism"), &["criticism"]),
            (
                dictgen::InsensitiveStr::Ascii("dissymmectrical"),
                &["dissymmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("instancees"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("reincatnation"),
                &["reincarnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conveniente"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwhleming"),
                &["underwhelming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabilitiy"),
                &["capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convencion"),
                &["convention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintinance"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orgainzation"),
                &["organizations", "organization"],
            ),
            (dictgen::InsensitiveStr::Ascii("copmleted"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("wezzal"), &["weasel"]),
            (
                dictgen::InsensitiveStr::Ascii("ttempt"),
                &["attempt", "tempt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superfluu"),
                &["superfluous"],
            ),
            (dictgen::InsensitiveStr::Ascii("nusaince"), &["nuisance"]),
            (dictgen::InsensitiveStr::Ascii("ontainors"), &["containers"]),
            (dictgen::InsensitiveStr::Ascii("thrusdays"), &["thursdays"]),
            (dictgen::InsensitiveStr::Ascii("distinces"), &["distances"]),
            (dictgen::InsensitiveStr::Ascii("indicat"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("constitucional"),
                &["constitutional"],
            ),
            (dictgen::InsensitiveStr::Ascii("spanwed"), &["spawned"]),
            (dictgen::InsensitiveStr::Ascii("alrorythm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("reposiroty"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("likley"), &["likely"]),
            (dictgen::InsensitiveStr::Ascii("avancing"), &["advancing"]),
            (dictgen::InsensitiveStr::Ascii("memwoir"), &["memoir"]),
            (
                dictgen::InsensitiveStr::Ascii("documentatation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefectly"), &["perfectly"]),
            (
                dictgen::InsensitiveStr::Ascii("thubmnails"),
                &["thumbnails"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcluating"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("omaje"), &["homage"]),
            (dictgen::InsensitiveStr::Ascii("depedency"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("proabaly"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("carbohydrats"),
                &["carbohydrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("represnet"), &["represent"]),
            (
                dictgen::InsensitiveStr::Ascii("namin"),
                &["naming", "admin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarentied"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proposicion"),
                &["proposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordiary"),
                &["extraordinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionaily"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatability"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("nstallers"), &["installers"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicaes"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("abriviate"), &["abbreviate"]),
            (dictgen::InsensitiveStr::Ascii("covereage"), &["coverage"]),
            (dictgen::InsensitiveStr::Ascii("attemtps"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("amibguously"),
                &["ambiguously"],
            ),
            (dictgen::InsensitiveStr::Ascii("extremiste"), &["extremes"]),
            (dictgen::InsensitiveStr::Ascii("contextos"), &["contexts"]),
            (
                dictgen::InsensitiveStr::Ascii("restraunt"),
                &["restraint", "restaurant"],
            ),
            (dictgen::InsensitiveStr::Ascii("playgroud"), &["playground"]),
            (dictgen::InsensitiveStr::Ascii("prmitives"), &["primitives"]),
            (
                dictgen::InsensitiveStr::Ascii("abbriviating"),
                &["abbreviating"],
            ),
            (dictgen::InsensitiveStr::Ascii("matcing"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("articluate"),
                &["articulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instansiation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convined"),
                &["combined", "convinced", "confined", "convened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requrirements"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsturcts"),
                &["reconstructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congfigures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uniquesness"),
                &["uniqueness"],
            ),
            (dictgen::InsensitiveStr::Ascii("cashire"), &["cashier"]),
            (dictgen::InsensitiveStr::Ascii("descusses"), &["discusses"]),
            (dictgen::InsensitiveStr::Ascii("theroems"), &["theorems"]),
            (dictgen::InsensitiveStr::Ascii("downsteam"), &["downstream"]),
            (dictgen::InsensitiveStr::Ascii("sketpics"), &["skeptics"]),
            (
                dictgen::InsensitiveStr::Ascii("beahviours"),
                &["behaviours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighrours"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("collater"), &["collator"]),
            (dictgen::InsensitiveStr::Ascii("controlles"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("repition"), &["repetition"]),
            (dictgen::InsensitiveStr::Ascii("chargehr"), &["charger"]),
            (
                dictgen::InsensitiveStr::Ascii("ckecklists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exectuions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("doubth"), &["doubt"]),
            (
                dictgen::InsensitiveStr::Ascii("merchandies"),
                &["merchandise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occurrance"),
                &["occurrence"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorithmi"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("limting"),
                &["limiting", "linting"],
            ),
            (dictgen::InsensitiveStr::Ascii("loadead"), &["loaded"]),
            (dictgen::InsensitiveStr::Ascii("achilleus"), &["achilles"]),
            (dictgen::InsensitiveStr::Ascii("disgonal"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("alloctor"), &["allocator"]),
            (dictgen::InsensitiveStr::Ascii("journaal"), &["journal"]),
            (
                dictgen::InsensitiveStr::Ascii("presonally"),
                &["personally"],
            ),
            (dictgen::InsensitiveStr::Ascii("perimetre"), &["perimeter"]),
            (
                dictgen::InsensitiveStr::Ascii("discepline"),
                &["discipline"],
            ),
            (dictgen::InsensitiveStr::Ascii("tlaking"), &["talking"]),
            (dictgen::InsensitiveStr::Ascii("runtume"), &["runtime"]),
            (dictgen::InsensitiveStr::Ascii("fingersi"), &["fingertips"]),
            (
                dictgen::InsensitiveStr::Ascii("uestionnaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("athleticisim"),
                &["athleticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("booolean"), &["boolean"]),
            (
                dictgen::InsensitiveStr::Ascii("configuarable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("epxectation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrailized"),
                &["industrialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("analagous"), &["analogous"]),
            (
                dictgen::InsensitiveStr::Ascii("handelbars"),
                &["handlebars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inportantly"),
                &["importantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprixamate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supervison"),
                &["supervision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheesecave"),
                &["cheesecake"],
            ),
            (dictgen::InsensitiveStr::Ascii("skeptis"), &["skeptics"]),
            (
                dictgen::InsensitiveStr::Ascii("maintanence"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("pitty"), &["pity"]),
            (
                dictgen::InsensitiveStr::Ascii("powerhours"),
                &["powerhouse"],
            ),
            (dictgen::InsensitiveStr::Ascii("westrbook"), &["westbrook"]),
            (dictgen::InsensitiveStr::Ascii("evenually"), &["eventually"]),
            (
                dictgen::InsensitiveStr::Ascii("singol"),
                &["signal", "single"],
            ),
            (dictgen::InsensitiveStr::Ascii("operration"), &["operation"]),
            (
                dictgen::InsensitiveStr::Ascii("backslahes"),
                &["backslashes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somes"),
                &["some", "sums", "semes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructers"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paravirutalized"),
                &["paravirtualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("packgaging"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("prodcut"), &["product"]),
            (dictgen::InsensitiveStr::Ascii("tagggen"), &["taggen"]),
            (
                dictgen::InsensitiveStr::Ascii("represants"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afternarket"),
                &["aftermarket"],
            ),
            (dictgen::InsensitiveStr::Ascii("soverein"), &["sovereign"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitraryily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("poducing"), &["producing"]),
            (
                dictgen::InsensitiveStr::Ascii("workaorunds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specailization"),
                &["specialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("achoring"), &["anchoring"]),
            (
                dictgen::InsensitiveStr::Ascii("executuions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standrdise"),
                &["standardise"],
            ),
            (dictgen::InsensitiveStr::Ascii("tjat"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("speciallty"), &["specially"]),
            (dictgen::InsensitiveStr::Ascii("tawk"), &["talk"]),
            (
                dictgen::InsensitiveStr::Ascii("quanitifying"),
                &["quantifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("foreksin"), &["foreskin"]),
            (dictgen::InsensitiveStr::Ascii("viatmin"), &["vitamin"]),
            (
                dictgen::InsensitiveStr::Ascii("optimsitic"),
                &["optimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("coodinat"), &["coordinate"]),
            (
                dictgen::InsensitiveStr::Ascii("organisaion"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("formaters"), &["formatters"]),
            (dictgen::InsensitiveStr::Ascii("unseting"), &["unsetting"]),
            (
                dictgen::InsensitiveStr::Ascii("equivilently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disctionary"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("insepsion"), &["inception"]),
            (dictgen::InsensitiveStr::Ascii("newletter"), &["newsletter"]),
            (dictgen::InsensitiveStr::Ascii("inquier"), &["inquire"]),
            (
                dictgen::InsensitiveStr::Ascii("prefitlered"),
                &["prefiltered"],
            ),
            (dictgen::InsensitiveStr::Ascii("hinudism"), &["hinduism"]),
            (
                dictgen::InsensitiveStr::Ascii("applycations"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstracly"), &["abstractly"]),
            (dictgen::InsensitiveStr::Ascii("callbcks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficiant"),
                &["coefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transactoins"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrytpion"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("detauls"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("riminiced"), &["reminisced"]),
            (dictgen::InsensitiveStr::Ascii("susbsystem"), &["subsystem"]),
            (dictgen::InsensitiveStr::Ascii("intsance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("seachkey"), &["searchkey"]),
            (
                dictgen::InsensitiveStr::Ascii("paraphenalia"),
                &["paraphernalia"],
            ),
            (dictgen::InsensitiveStr::Ascii("multply"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("procedres"), &["procedures"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotications"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instroment"),
                &["instrument"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conicidence"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatorical"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("promose"), &["promotes"]),
            (dictgen::InsensitiveStr::Ascii("acccounts"), &["accounts"]),
            (dictgen::InsensitiveStr::Ascii("copmiles"), &["compiles"]),
            (dictgen::InsensitiveStr::Ascii("animonee"), &["anemone"]),
            (
                dictgen::InsensitiveStr::Ascii("interlacin"),
                &["interlacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcendant"),
                &["transcendent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallellise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("ede"), &["edge"]),
            (dictgen::InsensitiveStr::Ascii("reource"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("cariers"),
                &["carriers", "careers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pedophilla"),
                &["pedophilia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interepreted"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("aclhemist"), &["alchemist"]),
            (
                dictgen::InsensitiveStr::Ascii("permenently"),
                &["permanently"],
            ),
            (dictgen::InsensitiveStr::Ascii("severeid"), &["severed"]),
            (
                dictgen::InsensitiveStr::Ascii("hamburgles"),
                &["hamburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redction"),
                &["reduction", "redaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("auxiliar"), &["auxiliary"]),
            (
                dictgen::InsensitiveStr::Ascii("oder"),
                &["order", "odor", "older", "coder", "odder", "over", "doer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undenaible"),
                &["undeniable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("therapeutuc"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotication"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectuion"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("captialism"),
                &["capitalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("releaseds"),
                &["released", "releases"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnatees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("sawteing"), &["sautéing"]),
            (
                dictgen::InsensitiveStr::Ascii("compatiblity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninstlalations"),
                &["uninstallations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirience"),
                &["experience"],
            ),
            (dictgen::InsensitiveStr::Ascii("cretae"), &["create"]),
            (
                dictgen::InsensitiveStr::Ascii("quantit"),
                &["quantity", "quantic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitemate"),
                &["legitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("anythimng"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("disaapointed"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("entquiries"), &["enquiries"]),
            (dictgen::InsensitiveStr::Ascii("covere"), &["cover"]),
            (dictgen::InsensitiveStr::Ascii("adajcent"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("microwavers"),
                &["microwaves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directionl"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("aleviates"), &["alleviates"]),
            (
                dictgen::InsensitiveStr::Ascii("convertiable"),
                &["convertible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convulution"),
                &["convolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppoerting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("shotuout"), &["shoutout"]),
            (
                dictgen::InsensitiveStr::Ascii("crystalisation"),
                &["crystallisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequitive"),
                &["consecutive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interal"),
                &["internal", "interval", "integral"],
            ),
            (dictgen::InsensitiveStr::Ascii("snappng"), &["snapping"]),
            (
                dictgen::InsensitiveStr::Ascii("associdating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknolwedgements"),
                &["acknowledgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unefined"),
                &["undefined", "unrefined"],
            ),
            (dictgen::InsensitiveStr::Ascii("enigne"), &["engine"]),
            (
                dictgen::InsensitiveStr::Ascii("conetions"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactored"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("doctorial"), &["doctoral"]),
            (dictgen::InsensitiveStr::Ascii("acctive"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("splatooon"), &["splatoon"]),
            (dictgen::InsensitiveStr::Ascii("preeceding"), &["preceding"]),
            (dictgen::InsensitiveStr::Ascii("berkelely"), &["berkeley"]),
            (
                dictgen::InsensitiveStr::Ascii("discolsure"),
                &["disclosure"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhacing"), &["enhancing"]),
            (
                dictgen::InsensitiveStr::Ascii("coordiante"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimand"),
                &["demand", "diamond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descirpition"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommeding"),
                &["recommending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dumbfouned"),
                &["dumbfounded"],
            ),
            (dictgen::InsensitiveStr::Ascii("sertain"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("coule"), &["could"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiing"),
                &["initializing", "initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnces"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggregrate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("toxen"), &["toxin"]),
            (
                dictgen::InsensitiveStr::Ascii("absolutelys"),
                &["absolutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcocding"),
                &["transcoding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opimizations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("algotihms"), &["algorithms"]),
            (
                dictgen::InsensitiveStr::Ascii("sereous"),
                &["serious", "serous"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplictes"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("cuztomized"),
                &["customized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsupprted"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("attching"), &["attaching"]),
            (
                dictgen::InsensitiveStr::Ascii("paraphrazing"),
                &["paraphrasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoluve"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("inferiour"), &["inferior"]),
            (dictgen::InsensitiveStr::Ascii("mappaing"), &["mapping"]),
            (dictgen::InsensitiveStr::Ascii("desroyed"), &["destroyed"]),
            (dictgen::InsensitiveStr::Ascii("estimages"), &["estimates"]),
            (dictgen::InsensitiveStr::Ascii("lisetener"), &["listener"]),
            (dictgen::InsensitiveStr::Ascii("essencial"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("introspektion"),
                &["introspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alernating"),
                &["alternating"],
            ),
            (dictgen::InsensitiveStr::Ascii("exteion"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("cleansiness"),
                &["cleanliness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permmission"),
                &["permission"],
            ),
            (dictgen::InsensitiveStr::Ascii("isrealis"), &["israelis"]),
            (dictgen::InsensitiveStr::Ascii("fogot"), &["forgot"]),
            (
                dictgen::InsensitiveStr::Ascii("capabilties"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("apparoch"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("algoirthm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("libraaries"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("compilgers"), &["compilers"]),
            (
                dictgen::InsensitiveStr::Ascii("constent"),
                &["consent", "constant", "content", "consistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("fulsh"), &["flush"]),
            (
                dictgen::InsensitiveStr::Ascii("unrecoverd"),
                &["unrecovered"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoerists"), &["theorists"]),
            (dictgen::InsensitiveStr::Ascii("drawning"), &["drawing"]),
            (dictgen::InsensitiveStr::Ascii("contcat"), &["contact"]),
            (
                dictgen::InsensitiveStr::Ascii("equavilently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substituions"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("detalied"), &["detailed"]),
            (dictgen::InsensitiveStr::Ascii("distrobxx"), &["distrobox"]),
            (dictgen::InsensitiveStr::Ascii("offlaoded"), &["offloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("ambulences"),
                &["ambulances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcirucit"),
                &["subcircuit"],
            ),
            (dictgen::InsensitiveStr::Ascii("supplt"), &["supply"]),
            (
                dictgen::InsensitiveStr::Ascii("othopedics"),
                &["orthopedics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configutred"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucures"), &["structures"]),
            (dictgen::InsensitiveStr::Ascii("somentimes"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("reconetions"),
                &["reconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("temporry"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("coalaced"), &["coalesced"]),
            (dictgen::InsensitiveStr::Ascii("mutuxs"), &["mutexes"]),
            (dictgen::InsensitiveStr::Ascii("statemenet"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("benchmar"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("optomisations"),
                &["optimisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cathlic"), &["catholic"]),
            (dictgen::InsensitiveStr::Ascii("paitent"), &["patient"]),
            (
                dictgen::InsensitiveStr::Ascii("emphacizing"),
                &["emphasizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitted"), &["limited"]),
            (
                dictgen::InsensitiveStr::Ascii("postiional"),
                &["positional"],
            ),
            (dictgen::InsensitiveStr::Ascii("chivlary"), &["chivalry"]),
            (
                dictgen::InsensitiveStr::Ascii("plcaeholders"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentimentos"),
                &["sentiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlightnment"),
                &["enlightenment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplactions"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendetly"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captializing"),
                &["capitalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("visisbly"), &["visibly"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmisers"),
                &["optimisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliclated"),
                &["duplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("humon"), &["human"]),
            (
                dictgen::InsensitiveStr::Ascii("asbtracting"),
                &["abstracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convertations"),
                &["conversations", "conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanctionne"),
                &["sanctioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encylcopedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminises"), &["reminisces"]),
            (
                dictgen::InsensitiveStr::Ascii("louisivlle"),
                &["louisville"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throughoput"),
                &["throughput"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meachanisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterpeice"),
                &["masterpiece"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resursing"),
                &["recursing", "resourcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reamins"), &["remains"]),
            (dictgen::InsensitiveStr::Ascii("eneble"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("gernerators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulrs"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("exerciced"), &["exercised"]),
            (
                dictgen::InsensitiveStr::Ascii("deamonization"),
                &["daemonization"],
            ),
            (dictgen::InsensitiveStr::Ascii("bsaed"), &["based"]),
            (dictgen::InsensitiveStr::Ascii("levaridge"), &["leverage"]),
            (dictgen::InsensitiveStr::Ascii("noticying"), &["notifying"]),
            (dictgen::InsensitiveStr::Ascii("enoufh"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("comittee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("imilar"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("implemant"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("interrim"), &["interim"]),
            (
                dictgen::InsensitiveStr::Ascii("enironments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("submiited"), &["submitted"]),
            (
                dictgen::InsensitiveStr::Ascii("unimplemeted"),
                &["unimplemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealiarative"),
                &["ameliorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("spartas"), &["spartans"]),
            (
                dictgen::InsensitiveStr::Ascii("initilaliser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialzies"),
                &["serializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiialisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contriversial"),
                &["controversial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thunderboat"),
                &["thunderbolt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trustowrthy"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inspectoras"),
                &["inspectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("poin"), &["point"]),
            (dictgen::InsensitiveStr::Ascii("expessing"), &["expressing"]),
            (dictgen::InsensitiveStr::Ascii("enery"), &["energy"]),
            (dictgen::InsensitiveStr::Ascii("worlers"), &["workers"]),
            (dictgen::InsensitiveStr::Ascii("invokved"), &["invoked"]),
            (dictgen::InsensitiveStr::Ascii("protocall"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("ofcoruse"), &["ofcourse"]),
            (dictgen::InsensitiveStr::Ascii("istelf"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("marketting"), &["marketing"]),
            (dictgen::InsensitiveStr::Ascii("blong"), &["belong"]),
            (
                dictgen::InsensitiveStr::Ascii("corresondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accommidate"),
                &["accommodate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cobine"), &["combine"]),
            (dictgen::InsensitiveStr::Ascii("chandlure"), &["chandler"]),
            (dictgen::InsensitiveStr::Ascii("asssign"), &["assign"]),
            (
                dictgen::InsensitiveStr::Ascii("intanjible"),
                &["intangible"],
            ),
            (dictgen::InsensitiveStr::Ascii("dectecting"), &["detecting"]),
            (dictgen::InsensitiveStr::Ascii("screenhot"), &["screenshot"]),
            (dictgen::InsensitiveStr::Ascii("fastner"), &["fastener"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstucted"),
                &["reconstructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("remannts"), &["remnants"]),
            (
                dictgen::InsensitiveStr::Ascii("proteen"),
                &["preteen", "protean", "protein"],
            ),
            (dictgen::InsensitiveStr::Ascii("cllose"), &["close"]),
            (
                dictgen::InsensitiveStr::Ascii("submisisons"),
                &["submissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bankrupcty"),
                &["bankruptcy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inofficial"),
                &["unofficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("orientied"), &["oriented"]),
            (dictgen::InsensitiveStr::Ascii("exspensed"), &["expensed"]),
            (dictgen::InsensitiveStr::Ascii("relasing"), &["releasing"]),
            (
                dictgen::InsensitiveStr::Ascii("randayvooed"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotible"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribtes"),
                &["attributes", "attribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inmediatelly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conterparts"),
                &["counterparts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depedencys"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipresistion"),
                &["multiprecision"],
            ),
            (dictgen::InsensitiveStr::Ascii("yatch"), &["yacht"]),
            (
                dictgen::InsensitiveStr::Ascii("reprent"),
                &["repent", "represent", "reprint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complcated"),
                &["complicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associatons"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("leagacy"), &["legacy"]),
            (dictgen::InsensitiveStr::Ascii("disribute"), &["distribute"]),
            (
                dictgen::InsensitiveStr::Ascii("calssically"),
                &["classically"],
            ),
            (dictgen::InsensitiveStr::Ascii("larvas"), &["larvae"]),
            (dictgen::InsensitiveStr::Ascii("tiawanese"), &["taiwanese"]),
            (dictgen::InsensitiveStr::Ascii("highlighs"), &["highlights"]),
            (
                dictgen::InsensitiveStr::Ascii("accured"),
                &["accrued", "occurred", "acquired"],
            ),
            (dictgen::InsensitiveStr::Ascii("assignent"), &["assignment"]),
            (dictgen::InsensitiveStr::Ascii("palcebo"), &["placebo"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellizes"),
                &["parallelizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwright"),
                &["underweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sythetically"),
                &["synthetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("functinos"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("papanicalou"),
                &["papanicolaou"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preaorocessing"),
                &["preprocessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("utitity"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("infiltraron"),
                &["infiltrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinlge"), &["single"]),
            (dictgen::InsensitiveStr::Ascii("chossing"), &["choosing"]),
            (
                dictgen::InsensitiveStr::Ascii("revolations"),
                &["revolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffeur"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("excorciating"),
                &["excruciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("mesured"), &["measured"]),
            (
                dictgen::InsensitiveStr::Ascii("contriception"),
                &["contraception"],
            ),
            (dictgen::InsensitiveStr::Ascii("infinet"), &["infinite"]),
            (dictgen::InsensitiveStr::Ascii("occrred"), &["occurred"]),
            (dictgen::InsensitiveStr::Ascii("parastic"), &["parasitic"]),
            (dictgen::InsensitiveStr::Ascii("toglle"), &["toggle"]),
            (dictgen::InsensitiveStr::Ascii("adjsut"), &["adjust"]),
            (dictgen::InsensitiveStr::Ascii("examplary"), &["exemplary"]),
            (
                dictgen::InsensitiveStr::Ascii("instantanious"),
                &["instantaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("explainig"), &["explaining"]),
            (dictgen::InsensitiveStr::Ascii("prexixed"), &["prefixed"]),
            (dictgen::InsensitiveStr::Ascii("contian"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("finitel"), &["finite"]),
            (
                dictgen::InsensitiveStr::Ascii("capitolized"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promethius"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("fallbck"), &["fallback"]),
            (dictgen::InsensitiveStr::Ascii("clientes"), &["clients"]),
            (
                dictgen::InsensitiveStr::Ascii("superseeded"),
                &["superseded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astromonical"),
                &["astronomical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularlize"),
                &["regularize"],
            ),
            (dictgen::InsensitiveStr::Ascii("exntension"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("bisexuella"), &["bisexual"]),
            (dictgen::InsensitiveStr::Ascii("interfacs"), &["interfaces"]),
            (
                dictgen::InsensitiveStr::Ascii("regnerates"),
                &["regenerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("homewords"), &["homeworlds"]),
            (
                dictgen::InsensitiveStr::Ascii("misdomeanor"),
                &["misdemeanor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shcedulers"),
                &["schedulers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pscyhiatrists"),
                &["psychiatrists"],
            ),
            (dictgen::InsensitiveStr::Ascii("aotomated"), &["automated"]),
            (
                dictgen::InsensitiveStr::Ascii("calcutating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sesitivity"),
                &["sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imlementing"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconsciosly"),
                &["unconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("falback"), &["fallback"]),
            (dictgen::InsensitiveStr::Ascii("fatser"), &["faster"]),
            (dictgen::InsensitiveStr::Ascii("surfcae"), &["surface"]),
            (dictgen::InsensitiveStr::Ascii("beleifing"), &["believing"]),
            (
                dictgen::InsensitiveStr::Ascii("meassages"),
                &["messages", "massages"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignmet"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("gankign"), &["ganking"]),
            (
                dictgen::InsensitiveStr::Ascii("nutricious"),
                &["nutritious"],
            ),
            (dictgen::InsensitiveStr::Ascii("removeed"), &["removed"]),
            (
                dictgen::InsensitiveStr::Ascii("docomenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticiped"),
                &["anticipated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalisties"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuigrable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("starwberry"),
                &["strawberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("frst"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("roleply"), &["roleplay"]),
            (dictgen::InsensitiveStr::Ascii("availablle"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("multuplicities"),
                &["multiplicities"],
            ),
            (dictgen::InsensitiveStr::Ascii("cataclyms"), &["cataclysm"]),
            (dictgen::InsensitiveStr::Ascii("buffrer"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("apppends"), &["appends"]),
            (dictgen::InsensitiveStr::Ascii("addapters"), &["adapters"]),
            (
                dictgen::InsensitiveStr::Ascii("dissaperd"),
                &["disappeared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaised"),
                &["realised", "raised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compability"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apppriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comutes"),
                &["commutes", "computes"],
            ),
            (dictgen::InsensitiveStr::Ascii("trhreaded"), &["threaded"]),
            (dictgen::InsensitiveStr::Ascii("portrais"), &["portraits"]),
            (
                dictgen::InsensitiveStr::Ascii("formts"),
                &["formats", "forms"],
            ),
            (dictgen::InsensitiveStr::Ascii("sincerelly"), &["sincerely"]),
            (dictgen::InsensitiveStr::Ascii("keystoke"), &["keystroke"]),
            (dictgen::InsensitiveStr::Ascii("psyhic"), &["psychic"]),
            (
                dictgen::InsensitiveStr::Ascii("platfarmer"),
                &["platformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communcates"),
                &["communicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("disonnect"), &["disconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("propriatory"),
                &["proprietary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stanardization"),
                &["standardization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumilated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("genearal"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("integreate"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("asent"), &["ascent"]),
            (dictgen::InsensitiveStr::Ascii("balanceado"), &["balanced"]),
            (
                dictgen::InsensitiveStr::Ascii("slection"),
                &["selection", "section", "election"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disticntion"),
                &["distinctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("comission"), &["commission"]),
            (dictgen::InsensitiveStr::Ascii("repetoire"), &["repertoire"]),
            (dictgen::InsensitiveStr::Ascii("metadataa"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("floatation"), &["flotation"]),
            (dictgen::InsensitiveStr::Ascii("hexogan"), &["hexagon"]),
            (dictgen::InsensitiveStr::Ascii("nammes"), &["names"]),
            (dictgen::InsensitiveStr::Ascii("franchie"), &["franchises"]),
            (dictgen::InsensitiveStr::Ascii("poitned"), &["pointed"]),
            (dictgen::InsensitiveStr::Ascii("simulaties"), &["simulate"]),
            (dictgen::InsensitiveStr::Ascii("requesite"), &["requisite"]),
            (dictgen::InsensitiveStr::Ascii("satifsy"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("termnial"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("skillslots"),
                &["skillshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("airsfot"), &["airsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("rectengular"),
                &["rectangular"],
            ),
            (dictgen::InsensitiveStr::Ascii("reflets"), &["reflects"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaite"),
                &["reallocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traveresing"),
                &["traversing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("translusent"),
                &["translucent"],
            ),
            (dictgen::InsensitiveStr::Ascii("szenarios"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("wendesdays"),
                &["wednesdays"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendereing"), &["rendering"]),
            (dictgen::InsensitiveStr::Ascii("exhisted"), &["existed"]),
            (dictgen::InsensitiveStr::Ascii("cigarrets"), &["cigarets"]),
            (dictgen::InsensitiveStr::Ascii("advocades"), &["advocates"]),
            (dictgen::InsensitiveStr::Ascii("thoroidal"), &["toroidal"]),
            (
                dictgen::InsensitiveStr::Ascii("guarnanted"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoorted"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("japanse"), &["japanese"]),
            (
                dictgen::InsensitiveStr::Ascii("jounralist"),
                &["journalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fabrikation"),
                &["fabrication"],
            ),
            (dictgen::InsensitiveStr::Ascii("syptoms"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("enntries"), &["entries"]),
            (
                dictgen::InsensitiveStr::Ascii("configuuration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomralize"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("woh"), &["who"]),
            (
                dictgen::InsensitiveStr::Ascii("negligeable"),
                &["negligible"],
            ),
            (dictgen::InsensitiveStr::Ascii("mornng"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("mechnanics"), &["mechanics"]),
            (dictgen::InsensitiveStr::Ascii("buro"), &["bureau", "burro"]),
            (
                dictgen::InsensitiveStr::Ascii("discriminatin"),
                &["discrimination", "discriminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("distrubed"), &["disturbed"]),
            (dictgen::InsensitiveStr::Ascii("turretts"), &["turrets"]),
            (dictgen::InsensitiveStr::Ascii("cruasder"), &["crusader"]),
            (dictgen::InsensitiveStr::Ascii("brackes"), &["brackets"]),
            (dictgen::InsensitiveStr::Ascii("relfects"), &["reflects"]),
            (
                dictgen::InsensitiveStr::Ascii("configuartion"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exersizing"),
                &["exercising"],
            ),
            (dictgen::InsensitiveStr::Ascii("transorms"), &["transforms"]),
            (
                dictgen::InsensitiveStr::Ascii("acelerations"),
                &["accelerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsilding"),
                &["reconciling"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncehcked"), &["unchecked"]),
            (dictgen::InsensitiveStr::Ascii("deparment"), &["department"]),
            (dictgen::InsensitiveStr::Ascii("addesses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("customzier"),
                &["customizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("yeald"), &["yield"]),
            (dictgen::InsensitiveStr::Ascii("inpouts"), &["inputs"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfified"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("hairstlye"), &["hairstyle"]),
            (dictgen::InsensitiveStr::Ascii("yeaterday"), &["yesterday"]),
            (
                dictgen::InsensitiveStr::Ascii("defititions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatlly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("claritiy"), &["clarity"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconveniances"),
                &["inconveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("yeilding"), &["yielding"]),
            (dictgen::InsensitiveStr::Ascii("wiythin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("secuencing"),
                &["sequencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("journy"), &["journey"]),
            (dictgen::InsensitiveStr::Ascii("labled"), &["labeled"]),
            (dictgen::InsensitiveStr::Ascii("asendency"), &["ascendency"]),
            (dictgen::InsensitiveStr::Ascii("extemely"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("blugaria"), &["bulgaria"]),
            (dictgen::InsensitiveStr::Ascii("forgeting"), &["forgetting"]),
            (dictgen::InsensitiveStr::Ascii("inclues"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("communmication"),
                &["communication"],
            ),
            (dictgen::InsensitiveStr::Ascii("violentce"), &["violence"]),
            (
                dictgen::InsensitiveStr::Ascii("reintantiating"),
                &["reinstantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("reasearch"), &["research"]),
            (dictgen::InsensitiveStr::Ascii("cloesest"), &["closest"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridnator"),
                &["coordinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("polymore"), &["polymer"]),
            (dictgen::InsensitiveStr::Ascii("beleive"), &["believe"]),
            (
                dictgen::InsensitiveStr::Ascii("recreativo"),
                &["recreation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dataum"), &["datum"]),
            (dictgen::InsensitiveStr::Ascii("oritinal"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("contenxt"),
                &["context", "content"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devritualisation"),
                &["devirtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("insteance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("genrally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("exportfs"), &["exports"]),
            (
                dictgen::InsensitiveStr::Ascii("researchs"),
                &["researchers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("verisoning"),
                &["versioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scipts"),
                &["scripts", "skips"],
            ),
            (dictgen::InsensitiveStr::Ascii("certainlyt"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("conersions"),
                &["conversions"],
            ),
            (dictgen::InsensitiveStr::Ascii("obvisously"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("indepently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionials"),
                &["professionalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("amzing"), &["amazing"]),
            (
                dictgen::InsensitiveStr::Ascii("investegated"),
                &["investigated"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbibtrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("suprize"), &["surprise"]),
            (
                dictgen::InsensitiveStr::Ascii("catergorically"),
                &["categorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fonctionnaly"),
                &["functionally", "functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("otpion"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("alpabetic"), &["alphabetic"]),
            (dictgen::InsensitiveStr::Ascii("succeedes"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("debugget"), &["debugger"]),
            (dictgen::InsensitiveStr::Ascii("fnaatic"), &["fanatic"]),
            (
                dictgen::InsensitiveStr::Ascii("patten"),
                &["pattern", "patent"],
            ),
            (dictgen::InsensitiveStr::Ascii("sevirity"), &["severity"]),
            (
                dictgen::InsensitiveStr::Ascii("incrementall"),
                &["incremental", "incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintiallised"),
                &["uninitialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desscriptively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visiblitity"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("anitrez"), &["antirez"]),
            (dictgen::InsensitiveStr::Ascii("colmuned"), &["columned"]),
            (
                dictgen::InsensitiveStr::Ascii("oversubscirbe"),
                &["oversubscribe"],
            ),
            (dictgen::InsensitiveStr::Ascii("parastie"), &["parasite"]),
            (dictgen::InsensitiveStr::Ascii("rusult"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("erasablocks"),
                &["eraseblocks"],
            ),
            (dictgen::InsensitiveStr::Ascii("whoch"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("loacalized"), &["localized"]),
            (dictgen::InsensitiveStr::Ascii("bulgaira"), &["bulgaria"]),
            (
                dictgen::InsensitiveStr::Ascii("tempelating"),
                &["templating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimdntelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loger"),
                &["logger", "lodger", "longer"],
            ),
            (dictgen::InsensitiveStr::Ascii("cloesr"), &["closer"]),
            (
                dictgen::InsensitiveStr::Ascii("apppreciate"),
                &["appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desscriptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplising"),
                &["accomplishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("totaly"), &["totally"]),
            (
                dictgen::InsensitiveStr::Ascii("experimantelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("designet"), &["designated"]),
            (
                dictgen::InsensitiveStr::Ascii("introducation"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("scuccess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("broozes"), &["bruises"]),
            (
                dictgen::InsensitiveStr::Ascii("recods"),
                &["records", "recodes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactionns"),
                &["interactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("boquet"), &["bouquet"]),
            (dictgen::InsensitiveStr::Ascii("torpedeo"), &["torpedo"]),
            (
                dictgen::InsensitiveStr::Ascii("provsional"),
                &["provisional"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirrected"), &["directed"]),
            (dictgen::InsensitiveStr::Ascii("extensino"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("apear"), &["appear"]),
            (
                dictgen::InsensitiveStr::Ascii("cholestrol"),
                &["cholesterol"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlcoking"),
                &["overlooking"],
            ),
            (dictgen::InsensitiveStr::Ascii("hashnap"), &["hashmap"]),
            (dictgen::InsensitiveStr::Ascii("isreali"), &["israeli"]),
            (dictgen::InsensitiveStr::Ascii("shrelock"), &["sherlock"]),
            (dictgen::InsensitiveStr::Ascii("addmin"), &["admin"]),
            (dictgen::InsensitiveStr::Ascii("arivals"), &["arrivals"]),
            (
                dictgen::InsensitiveStr::Ascii("ramdomises"),
                &["randomises"],
            ),
            (dictgen::InsensitiveStr::Ascii("acchieved"), &["achieved"]),
            (
                dictgen::InsensitiveStr::Ascii("derefernced"),
                &["dereferenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unaccpeted"),
                &["unaccepted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countours"),
                &["contours", "counters"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffert"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("descroptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphatizes"),
                &["emphasizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheter"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("sepeartely"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("esential"), &["essential"]),
            (dictgen::InsensitiveStr::Ascii("spehrical"), &["spherical"]),
            (
                dictgen::InsensitiveStr::Ascii("overzelous"),
                &["overzealous"],
            ),
            (dictgen::InsensitiveStr::Ascii("thefore"), &["therefore"]),
            (
                dictgen::InsensitiveStr::Ascii("imbalanaces"),
                &["imbalances"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchanche"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("wreslter"), &["wrestler"]),
            (dictgen::InsensitiveStr::Ascii("seez"), &["seize"]),
            (dictgen::InsensitiveStr::Ascii("preadtors"), &["predators"]),
            (dictgen::InsensitiveStr::Ascii("creaded"), &["created"]),
            (
                dictgen::InsensitiveStr::Ascii("executabil"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstuctor"),
                &["reconstructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("sesssion"), &["session"]),
            (dictgen::InsensitiveStr::Ascii("tye"), &["type", "tie"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalizating"),
                &["capitalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("romoves"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("ocilated"), &["oscillated"]),
            (
                dictgen::InsensitiveStr::Ascii("concatentaed"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perspectie"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guanatanmo"),
                &["guantanamo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrrounding"),
                &["surrounding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evluations"),
                &["evaluations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucratisch"),
                &["bureaucratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inputsream"),
                &["inputstream"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fuctioning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrroundings"),
                &["surroundings"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximims"), &["maximums"]),
            (dictgen::InsensitiveStr::Ascii("desipite"), &["despite"]),
            (dictgen::InsensitiveStr::Ascii("personaly"), &["personally"]),
            (dictgen::InsensitiveStr::Ascii("itterator"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("chooose"), &["choose"]),
            (dictgen::InsensitiveStr::Ascii("extactors"), &["extractors"]),
            (dictgen::InsensitiveStr::Ascii("predicit"), &["predict"]),
            (dictgen::InsensitiveStr::Ascii("annonces"), &["announces"]),
            (dictgen::InsensitiveStr::Ascii("archictect"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("reposotory"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constransi"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unhilighted"),
                &["unhighlighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constiutents"),
                &["constituents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cootdinated"),
                &["coordinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("rondayvoo"), &["rendezvous"]),
            (dictgen::InsensitiveStr::Ascii("neagtive"), &["negative"]),
            (
                dictgen::InsensitiveStr::Ascii("encoutters"),
                &["encounters"],
            ),
            (dictgen::InsensitiveStr::Ascii("asisted"), &["assisted"]),
            (dictgen::InsensitiveStr::Ascii("recommens"), &["recommends"]),
            (dictgen::InsensitiveStr::Ascii("mechnics"), &["mechanics"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpectadely"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforse"),
                &["endorse", "enforce"],
            ),
            (dictgen::InsensitiveStr::Ascii("implcitly"), &["implicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("commulative"),
                &["cumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychaitrists"),
                &["psychiatrists"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependnce"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("linguisitic"),
                &["linguistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstructional"),
                &["instructional"],
            ),
            (dictgen::InsensitiveStr::Ascii("dividendes"), &["dividends"]),
            (dictgen::InsensitiveStr::Ascii("deplying"), &["deploying"]),
            (dictgen::InsensitiveStr::Ascii("lewtenant"), &["lieutenant"]),
            (
                dictgen::InsensitiveStr::Ascii("unspecificed"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarassed"),
                &["embarrassed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirementally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("calendays"), &["calendars"]),
            (dictgen::InsensitiveStr::Ascii("disign"), &["design"]),
            (
                dictgen::InsensitiveStr::Ascii("corrupcion"),
                &["corruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnante"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("sinnicaly"), &["cynically"]),
            (
                dictgen::InsensitiveStr::Ascii("rapresents"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capialized"),
                &["capitalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuition"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("phenemona"), &["phenomena"]),
            (dictgen::InsensitiveStr::Ascii("speparated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("calculationg"),
                &["calculating", "calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combainators"),
                &["combinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("queuin"), &["queuing"]),
            (
                dictgen::InsensitiveStr::Ascii("efficencty"),
                &["efficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("engilsh"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("obselete"), &["obsolete"]),
            (dictgen::InsensitiveStr::Ascii("inpsect"), &["inspect"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsstrapping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("complexers"), &["complexes"]),
            (
                dictgen::InsensitiveStr::Ascii("correspindences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clapse"),
                &["collapse", "clasp", "claps", "lapse"],
            ),
            (dictgen::InsensitiveStr::Ascii("congigs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("medhod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("correspdoned"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("wishlisht"), &["wishlist"]),
            (dictgen::InsensitiveStr::Ascii("panalty"), &["penalty"]),
            (
                dictgen::InsensitiveStr::Ascii("currenctly"),
                &["currency", "currently"],
            ),
            (dictgen::InsensitiveStr::Ascii("failt"), &["fail", "failed"]),
            (dictgen::InsensitiveStr::Ascii("seseedes"), &["secedes"]),
            (
                dictgen::InsensitiveStr::Ascii("conceivablely"),
                &["conceivably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decalratiosn"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("varialbes"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("runnin"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("realates"), &["relates"]),
            (
                dictgen::InsensitiveStr::Ascii("mirros"),
                &["mirrors", "mirror"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("daplicating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("nashvile"), &["nashville"]),
            (dictgen::InsensitiveStr::Ascii("lincensing"), &["licensing"]),
            (dictgen::InsensitiveStr::Ascii("sumbarine"), &["submarine"]),
            (
                dictgen::InsensitiveStr::Ascii("impelmenting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("behaiviour"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("ttested"),
                &["tested", "attested"],
            ),
            (dictgen::InsensitiveStr::Ascii("setings"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("sheeps"), &["sheep"]),
            (dictgen::InsensitiveStr::Ascii("shoucl"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("argumend"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("conceeding"), &["conceding"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillifer"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("interpert"), &["interpret"]),
            (dictgen::InsensitiveStr::Ascii("distractin"), &["district"]),
            (
                dictgen::InsensitiveStr::Ascii("cominatorics"),
                &["combinatorics"],
            ),
            (dictgen::InsensitiveStr::Ascii("aaccess"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("specifiede"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("liev"), &["live"]),
            (
                dictgen::InsensitiveStr::Ascii("techonology"),
                &["technology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosiates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("yiur"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("positonal"), &["positional"]),
            (
                dictgen::InsensitiveStr::Ascii("chirstianity"),
                &["christianity"],
            ),
            (dictgen::InsensitiveStr::Ascii("throghput"), &["throughput"]),
            (
                dictgen::InsensitiveStr::Ascii("registartions"),
                &["registrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horrendious"),
                &["horrendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("actialy"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("expeienced"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("happenned"), &["happened"]),
            (
                dictgen::InsensitiveStr::Ascii("refurbushed"),
                &["refurbished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("africaner"),
                &["african", "afrikaner", "africander", "afrikander"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minisation"),
                &["minimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("partchett"), &["pratchett"]),
            (dictgen::InsensitiveStr::Ascii("benghazhi"), &["benghazi"]),
            (dictgen::InsensitiveStr::Ascii("inmutable"), &["immutable"]),
            (
                dictgen::InsensitiveStr::Ascii("carinomies"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fromatting"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("genereally"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("paterns"), &["patterns"]),
            (dictgen::InsensitiveStr::Ascii("maneveurs"), &["maneuvers"]),
            (dictgen::InsensitiveStr::Ascii("hadnler"), &["handler"]),
            (
                dictgen::InsensitiveStr::Ascii("prefentation"),
                &["presentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulteaneously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("gurantess"), &["guarantees"]),
            (dictgen::InsensitiveStr::Ascii("advanatage"), &["advantage"]),
            (
                dictgen::InsensitiveStr::Ascii("irrelivent"),
                &["irrelevant"],
            ),
            (dictgen::InsensitiveStr::Ascii("bunlde"), &["bundle"]),
            (
                dictgen::InsensitiveStr::Ascii("agants"),
                &["agents", "against"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuztomization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("ampilfier"), &["amplifier"]),
            (dictgen::InsensitiveStr::Ascii("alhpa"), &["alpha"]),
            (
                dictgen::InsensitiveStr::Ascii("celebratons"),
                &["celebrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("fornt"), &["front"]),
            (dictgen::InsensitiveStr::Ascii("obeidence"), &["obedience"]),
            (
                dictgen::InsensitiveStr::Ascii("synchrous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("excutor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("congressen"),
                &["congressmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyops"), &["typos"]),
            (
                dictgen::InsensitiveStr::Ascii("bourgeousie"),
                &["bourgeois"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("popualtion"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substansially"),
                &["substantially"],
            ),
            (dictgen::InsensitiveStr::Ascii("transcipt"), &["transcript"]),
            (dictgen::InsensitiveStr::Ascii("bechmark"), &["benchmark"]),
            (dictgen::InsensitiveStr::Ascii("withen"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("teamplate"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("indoctrinatie"),
                &["indoctrinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desconstructed"),
                &["deconstructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countainers"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repacements"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("excetuting"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("grandises"),
                &["aggrandises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalulators"),
                &["evaluators"],
            ),
            (dictgen::InsensitiveStr::Ascii("fane"), &["fan", "feign"]),
            (dictgen::InsensitiveStr::Ascii("symtpom"), &["symptom"]),
            (
                dictgen::InsensitiveStr::Ascii("dcoumenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficififers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiaton"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corisponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("haversting"),
                &["harvesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retramsmission"),
                &["retransmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circunvent"),
                &["circumvent"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabeles"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("akkumulate"),
                &["accumulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exibition"), &["exhibition"]),
            (
                dictgen::InsensitiveStr::Ascii("homepoathy"),
                &["homeopathy"],
            ),
            (dictgen::InsensitiveStr::Ascii("caculater"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("ttempting"),
                &["attempting", "tempting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("akkumulated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exersizes"), &["exercises"]),
            (
                dictgen::InsensitiveStr::Ascii("assasinate"),
                &["assassinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("datastucture"),
                &["datastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("relapsin"), &["relapsing"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaitve"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfonction"),
                &["dysfunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutiei"),
                &["institute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decalrations"),
                &["declarations"],
            ),
            (dictgen::InsensitiveStr::Ascii("watchemn"), &["watchmen"]),
            (
                dictgen::InsensitiveStr::Ascii("consequenes"),
                &["consequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteteristics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("colmuns"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastructre"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filmmakare"),
                &["filmmakers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disclipined"),
                &["disciplined"],
            ),
            (dictgen::InsensitiveStr::Ascii("thak"), &["thank"]),
            (
                dictgen::InsensitiveStr::Ascii("retribuito"),
                &["retribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostitude"),
                &["prostitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capaciters"),
                &["capacitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("occupance"), &["occupancy"]),
            (
                dictgen::InsensitiveStr::Ascii("termintions"),
                &["terminations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexplicablly"),
                &["inexplicably"],
            ),
            (dictgen::InsensitiveStr::Ascii("juni"), &["june"]),
            (
                dictgen::InsensitiveStr::Ascii("regualting"),
                &["regulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lsit"),
                &["list", "slit", "sit"],
            ),
            (dictgen::InsensitiveStr::Ascii("resolveres"), &["resolves"]),
            (dictgen::InsensitiveStr::Ascii("buliding"), &["building"]),
            (
                dictgen::InsensitiveStr::Ascii("regeression"),
                &["regression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administratief"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("htem"), &["them"]),
            (
                dictgen::InsensitiveStr::Ascii("mysteriousy"),
                &["mysteriously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philospher"),
                &["philosopher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmant"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("atmoic"), &["atomic"]),
            (dictgen::InsensitiveStr::Ascii("plethoria"), &["plethora"]),
            (dictgen::InsensitiveStr::Ascii("variing"), &["varying"]),
            (dictgen::InsensitiveStr::Ascii("developped"), &["developed"]),
            (dictgen::InsensitiveStr::Ascii("permiere"), &["premiere"]),
            (dictgen::InsensitiveStr::Ascii("redeemeed"), &["redeemed"]),
            (dictgen::InsensitiveStr::Ascii("leftits"), &["leftist"]),
            (dictgen::InsensitiveStr::Ascii("syndacite"), &["syndicate"]),
            (dictgen::InsensitiveStr::Ascii("sterilze"), &["sterile"]),
            (
                dictgen::InsensitiveStr::Ascii("recolleciton"),
                &["recollection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginizes"),
                &["recognizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emcompassed"),
                &["encompassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("furethest"), &["furthest"]),
            (
                dictgen::InsensitiveStr::Ascii("decalartion"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactins"),
                &["interacts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefferences"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("snse"), &["sense"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotitaion"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("namd"), &["named", "name"]),
            (dictgen::InsensitiveStr::Ascii("insultas"), &["insults"]),
            (dictgen::InsensitiveStr::Ascii("ambigous"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("innovacion"),
                &["innovation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceution"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("stringnet"), &["stringent"]),
            (
                dictgen::InsensitiveStr::Ascii("satellittes"),
                &["satellites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succeesfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("backword"), &["backward"]),
            (
                dictgen::InsensitiveStr::Ascii("playgerize"),
                &["plagiarize"],
            ),
            (dictgen::InsensitiveStr::Ascii("chuned"), &["chunked"]),
            (
                dictgen::InsensitiveStr::Ascii("offisianados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("extrime"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("chnace"), &["chance"]),
            (dictgen::InsensitiveStr::Ascii("freqquent"), &["frequent"]),
            (
                dictgen::InsensitiveStr::Ascii("practicianers"),
                &["practitioners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvents"),
                &["improves", "improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("undertood"), &["understood"]),
            (
                dictgen::InsensitiveStr::Ascii("precent"),
                &["percent", "prescient", "present"],
            ),
            (dictgen::InsensitiveStr::Ascii("peformed"), &["performed"]),
            (
                dictgen::InsensitiveStr::Ascii("conveniant"),
                &["convenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranients"), &["transients"]),
            (
                dictgen::InsensitiveStr::Ascii("investigaton"),
                &["investigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("howerver"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("calcalates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("bazerk"), &["berserk"]),
            (dictgen::InsensitiveStr::Ascii("assighing"), &["assigning"]),
            (dictgen::InsensitiveStr::Ascii("maestries"), &["masteries"]),
            (dictgen::InsensitiveStr::Ascii("vectorss"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("activw"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("unemployemnt"),
                &["unemployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funciotnally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeeded"), &["exceeded"]),
            (
                dictgen::InsensitiveStr::Ascii("forgeround"),
                &["foreground"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifiies"),
                &["classifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delection"),
                &["detection", "deletion", "selection"],
            ),
            (dictgen::InsensitiveStr::Ascii("thornes"), &["thrones"]),
            (dictgen::InsensitiveStr::Ascii("consdier"), &["consider"]),
            (
                dictgen::InsensitiveStr::Ascii("ascconciated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("falso"), &["false"]),
            (dictgen::InsensitiveStr::Ascii("consunmes"), &["consumes"]),
            (dictgen::InsensitiveStr::Ascii("excho"), &["echo"]),
            (dictgen::InsensitiveStr::Ascii("relises"), &["realises"]),
            (
                dictgen::InsensitiveStr::Ascii("statisitcal"),
                &["statistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("benchmars"), &["benchmarks"]),
            (dictgen::InsensitiveStr::Ascii("aceptors"), &["acceptors"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditinaly"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("medicineras"),
                &["medicines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibicion"),
                &["prohibition"],
            ),
            (dictgen::InsensitiveStr::Ascii("antry"), &["entry"]),
            (dictgen::InsensitiveStr::Ascii("existin"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("decrpted"), &["decrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("cahracters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("specfy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("circuventions"),
                &["circumventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncommpressed"),
                &["uncompressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("apersands"), &["ampersands"]),
            (
                dictgen::InsensitiveStr::Ascii("calulation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassembly"),
                &["disassembly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consicousness"),
                &["consciousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("whule"), &["while", "whole"]),
            (dictgen::InsensitiveStr::Ascii("recident"), &["resident"]),
            (dictgen::InsensitiveStr::Ascii("marginable"), &["marginal"]),
            (
                dictgen::InsensitiveStr::Ascii("experimmentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("suprise"), &["surprise"]),
            (
                dictgen::InsensitiveStr::Ascii("perscribed"),
                &["prescribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumbing"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("interrugum"),
                &["interregnum"],
            ),
            (dictgen::InsensitiveStr::Ascii("stragetic"), &["strategic"]),
            (dictgen::InsensitiveStr::Ascii("suiete"), &["suite"]),
            (
                dictgen::InsensitiveStr::Ascii("philosophycally"),
                &["philosophically"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostaliga"), &["nostalgia"]),
            (dictgen::InsensitiveStr::Ascii("wtih"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("confunses"), &["confuses"]),
            (
                dictgen::InsensitiveStr::Ascii("recognicing"),
                &["recognizing", "recognising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referene"),
                &["referee", "reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("filnames"), &["filenames"]),
            (dictgen::InsensitiveStr::Ascii("exapend"), &["expand"]),
            (
                dictgen::InsensitiveStr::Ascii("priorietor"),
                &["proprietor"],
            ),
            (dictgen::InsensitiveStr::Ascii("prewview"), &["preview"]),
            (dictgen::InsensitiveStr::Ascii("analizes"), &["analyzes"]),
            (
                dictgen::InsensitiveStr::Ascii("conters"),
                &["conteurs", "counters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrigueing"),
                &["intriguing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ritlain"), &["ritalin"]),
            (dictgen::InsensitiveStr::Ascii("incluing"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("htttp"), &["http"]),
            (dictgen::InsensitiveStr::Ascii("heusitic"), &["heuristic"]),
            (
                dictgen::InsensitiveStr::Ascii("ultilization"),
                &["utilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("illegaly"), &["illegally"]),
            (dictgen::InsensitiveStr::Ascii("termiator"), &["terminator"]),
            (dictgen::InsensitiveStr::Ascii("handlling"), &["handling"]),
            (dictgen::InsensitiveStr::Ascii("startsup"), &["startups"]),
            (
                dictgen::InsensitiveStr::Ascii("missinterpret"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("bundel"), &["bundle"]),
            (dictgen::InsensitiveStr::Ascii("elsewhwere"), &["elsewhere"]),
            (dictgen::InsensitiveStr::Ascii("contacing"), &["contacting"]),
            (
                dictgen::InsensitiveStr::Ascii("interporated"),
                &["interpolated", "interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("occoured"), &["occurred"]),
            (dictgen::InsensitiveStr::Ascii("malcom"), &["malcolm"]),
            (dictgen::InsensitiveStr::Ascii("pretify"), &["prettify"]),
            (dictgen::InsensitiveStr::Ascii("negetive"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("originary"), &["originally"]),
            (
                dictgen::InsensitiveStr::Ascii("correcness"),
                &["correctness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stablilizes"),
                &["stabilizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("primitave"), &["primitive"]),
            (dictgen::InsensitiveStr::Ascii("filse"), &["files"]),
            (dictgen::InsensitiveStr::Ascii("organisier"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("flexebility"),
                &["flexibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoluiton"),
                &["revolutions", "revolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentification"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minoritets"),
                &["minorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("poalogies"), &["apologies"]),
            (
                dictgen::InsensitiveStr::Ascii("negociations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quirkyness"),
                &["quirkiness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnetiate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mesurement"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fileststem"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreacations"),
                &["deprecations"],
            ),
            (dictgen::InsensitiveStr::Ascii("elemenent"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsters"),
                &["administers"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostaglia"), &["nostalgia"]),
            (
                dictgen::InsensitiveStr::Ascii("destorying"),
                &["destroying"],
            ),
            (dictgen::InsensitiveStr::Ascii("sheepel"), &["sheeple"]),
            (
                dictgen::InsensitiveStr::Ascii("skitsofrinic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("squeakly"), &["squeaky"]),
            (dictgen::InsensitiveStr::Ascii("specfied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("therapetic"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statestics"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("academicus"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("appneders"), &["appenders"]),
            (
                dictgen::InsensitiveStr::Ascii("matieralism"),
                &["materialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("messured"), &["measured"]),
            (dictgen::InsensitiveStr::Ascii("compliace"), &["compliance"]),
            (
                dictgen::InsensitiveStr::Ascii("databaes"),
                &["database", "databases"],
            ),
            (dictgen::InsensitiveStr::Ascii("diamonnds"), &["diamonds"]),
            (dictgen::InsensitiveStr::Ascii("develpers"), &["developers"]),
            (dictgen::InsensitiveStr::Ascii("gorups"), &["groups"]),
            (dictgen::InsensitiveStr::Ascii("runnng"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("tansient"), &["transient"]),
            (
                dictgen::InsensitiveStr::Ascii("interstelar"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbituarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("errected"), &["erected"]),
            (
                dictgen::InsensitiveStr::Ascii("evaliations"),
                &["evaluations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifiters"),
                &["identifier", "identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("pateint"), &["patient"]),
            (dictgen::InsensitiveStr::Ascii("cyrptic"), &["cryptic"]),
            (dictgen::InsensitiveStr::Ascii("sovles"), &["solves"]),
            (dictgen::InsensitiveStr::Ascii("delgated"), &["delegated"]),
            (
                dictgen::InsensitiveStr::Ascii("rondayvooed"),
                &["rendezvoused"],
            ),
            (dictgen::InsensitiveStr::Ascii("actzal"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("precedeed"), &["preceded"]),
            (dictgen::InsensitiveStr::Ascii("tecture"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("aready"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("isntances"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("regorded"), &["recorded"]),
            (
                dictgen::InsensitiveStr::Ascii("demograpic"),
                &["demographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnticating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("irratic"), &["erratic"]),
            (dictgen::InsensitiveStr::Ascii("outsid"), &["outside"]),
            (dictgen::InsensitiveStr::Ascii("negotited"), &["negotiated"]),
            (
                dictgen::InsensitiveStr::Ascii("confugurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theromdynamics"),
                &["thermodynamics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eigth"),
                &["eighth", "eight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("saskatchewn"),
                &["saskatchewan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woekspaces"),
                &["workspaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpeptedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ccannot"), &["cannot"]),
            (dictgen::InsensitiveStr::Ascii("fanaticals"), &["fanatics"]),
            (dictgen::InsensitiveStr::Ascii("follewind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("occasionnal"),
                &["occasional"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucurs"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("custonization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("primeter"), &["perimeter"]),
            (
                dictgen::InsensitiveStr::Ascii("schziophrenic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("insteads"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("abstratly"), &["abstractly"]),
            (dictgen::InsensitiveStr::Ascii("stregnths"), &["strengths"]),
            (dictgen::InsensitiveStr::Ascii("gerneral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("nurisher"), &["nourisher"]),
            (dictgen::InsensitiveStr::Ascii("turntabel"), &["turntable"]),
            (
                dictgen::InsensitiveStr::Ascii("depoyments"),
                &["deployments"],
            ),
            (dictgen::InsensitiveStr::Ascii("provisios"), &["provisions"]),
            (
                dictgen::InsensitiveStr::Ascii("critisizing"),
                &["criticizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepreters"),
                &["interpreters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiablities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fashionalbe"),
                &["fashionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("clloudes"), &["clouds"]),
            (
                dictgen::InsensitiveStr::Ascii("beliebable"),
                &["believable"],
            ),
            (dictgen::InsensitiveStr::Ascii("persited"), &["persisted"]),
            (dictgen::InsensitiveStr::Ascii("addtitions"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("sturcture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("itializations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestons"),
                &["manifests"],
            ),
            (dictgen::InsensitiveStr::Ascii("satistfied"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("contengent"),
                &["contingent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sistematic"),
                &["systematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("asbtracts"), &["abstracts"]),
            (
                dictgen::InsensitiveStr::Ascii("elecrtomagnetic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qeustionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declaraton"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pocess"),
                &["process", "possess"],
            ),
            (dictgen::InsensitiveStr::Ascii("umarshal"), &["unmarshal"]),
            (
                dictgen::InsensitiveStr::Ascii("dulicating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("caould"), &["could"]),
            (
                dictgen::InsensitiveStr::Ascii("imperialsim"),
                &["imperialism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintiained"),
                &["maintained"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheisim"), &["atheism"]),
            (dictgen::InsensitiveStr::Ascii("celsuis"), &["celsius"]),
            (dictgen::InsensitiveStr::Ascii("resistend"), &["resisted"]),
            (
                dictgen::InsensitiveStr::Ascii("repreciating"),
                &["depreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("challened"), &["challenged"]),
            (dictgen::InsensitiveStr::Ascii("subjectifs"), &["subjects"]),
            (
                dictgen::InsensitiveStr::Ascii("precidences"),
                &["precedences"],
            ),
            (dictgen::InsensitiveStr::Ascii("slin"), &["sling"]),
            (
                dictgen::InsensitiveStr::Ascii("figuartively"),
                &["figuratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validattion"),
                &["validation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialigs"), &["dialogs"]),
            (dictgen::InsensitiveStr::Ascii("sistemize"), &["systemize"]),
            (
                dictgen::InsensitiveStr::Ascii("propigated"),
                &["propagated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruped"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deteministic"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cornithians"),
                &["corinthians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formalizng"),
                &["formalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bottonm"), &["bottom"]),
            (dictgen::InsensitiveStr::Ascii("giagbyte"), &["gigabyte"]),
            (dictgen::InsensitiveStr::Ascii("ptherads"), &["pthreads"]),
            (
                dictgen::InsensitiveStr::Ascii("confidantal"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("phospor"), &["phosphor"]),
            (dictgen::InsensitiveStr::Ascii("interaval"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("headachs"), &["headaches"]),
            (dictgen::InsensitiveStr::Ascii("optomal"), &["optimal"]),
            (dictgen::InsensitiveStr::Ascii("islandes"), &["islanders"]),
            (dictgen::InsensitiveStr::Ascii("herculeas"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("regularlos"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("wranlger"), &["wrangler"]),
            (dictgen::InsensitiveStr::Ascii("inbrased"), &["embraced"]),
            (
                dictgen::InsensitiveStr::Ascii("stretegically"),
                &["strategically"],
            ),
            (dictgen::InsensitiveStr::Ascii("compnent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("oaker"),
                &[
                    "baker", "faker", "laker", "maker", "oaken", "oakier", "ocher", "taker",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("cotains"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("aviaiton"), &["aviation"]),
            (dictgen::InsensitiveStr::Ascii("decresses"), &["decreases"]),
            (
                dictgen::InsensitiveStr::Ascii("warpps"),
                &["warps", "wraps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downloas"),
                &["downloads", "download"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procede"),
                &["proceed", "precede"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplefying"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlyze"), &["analyze"]),
            (
                dictgen::InsensitiveStr::Ascii("targert"),
                &["targets", "target"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("birthdayers"),
                &["birthdays"],
            ),
            (dictgen::InsensitiveStr::Ascii("decison"), &["decision"]),
            (
                dictgen::InsensitiveStr::Ascii("difined"),
                &["defined", "divined"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupliacte"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("cofigure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("selecrtor"), &["selector"]),
            (dictgen::InsensitiveStr::Ascii("acceping"), &["accepting"]),
            (
                dictgen::InsensitiveStr::Ascii("descritpions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciifer"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalizing"),
                &["orthogonalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrcts"),
                &["constructs", "constricts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoaiated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("uesr"), &["user"]),
            (
                dictgen::InsensitiveStr::Ascii("impotr"),
                &["import", "importer"],
            ),
            (dictgen::InsensitiveStr::Ascii("segaults"), &["segfaults"]),
            (dictgen::InsensitiveStr::Ascii("pursuades"), &["persuades"]),
            (
                dictgen::InsensitiveStr::Ascii("conttribute"),
                &["contribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparizon"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensivelly"),
                &["extensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("consruct"), &["construct"]),
            (
                dictgen::InsensitiveStr::Ascii("guarrentes"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("delevoper"), &["developer"]),
            (
                dictgen::InsensitiveStr::Ascii("identfying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("failrue"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("sepereates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("shouls"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("compilling"), &["compiling"]),
            (dictgen::InsensitiveStr::Ascii("slipperly"), &["slippery"]),
            (dictgen::InsensitiveStr::Ascii("athetlics"), &["athletics"]),
            (dictgen::InsensitiveStr::Ascii("altenate"), &["alternate"]),
            (dictgen::InsensitiveStr::Ascii("blindy"), &["blindly"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitiveties"),
                &["sensitivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exressions"),
                &["expressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlaise"),
                &["generalise"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffirent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("stderrr"), &["stderr"]),
            (dictgen::InsensitiveStr::Ascii("satisying"), &["satisfying"]),
            (dictgen::InsensitiveStr::Ascii("cahches"), &["caches"]),
            (
                dictgen::InsensitiveStr::Ascii("chcekerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("timeing"), &["timing"]),
            (
                dictgen::InsensitiveStr::Ascii("exponental"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("simlpe"), &["simple"]),
            (dictgen::InsensitiveStr::Ascii("moutns"), &["mounts"]),
            (
                dictgen::InsensitiveStr::Ascii("exite"),
                &["exit", "excite", "exits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condtitions"),
                &["conditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("acend"), &["ascend"]),
            (dictgen::InsensitiveStr::Ascii("communiss"), &["communists"]),
            (dictgen::InsensitiveStr::Ascii("teamspek"), &["teamspeak"]),
            (
                dictgen::InsensitiveStr::Ascii("transgenderd"),
                &["transgendered"],
            ),
            (dictgen::InsensitiveStr::Ascii("dortmud"), &["dortmund"]),
            (dictgen::InsensitiveStr::Ascii("cacnel"), &["cancel"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpectetly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apologisms"),
                &["apologists", "apologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissobediance"),
                &["disobedience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconditonal"),
                &["unconditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoucenments"),
                &["announcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invulenrable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrivable"),
                &["retrievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dsiable"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("endolithes"), &["endoliths"]),
            (
                dictgen::InsensitiveStr::Ascii("descallocate"),
                &["deallocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regretably"),
                &["regrettably"],
            ),
            (dictgen::InsensitiveStr::Ascii("destrutor"), &["destructor"]),
            (
                dictgen::InsensitiveStr::Ascii("compolsury"),
                &["compulsory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tage"),
                &["stage", "take", "tag", "tagged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succcessors"),
                &["successors"],
            ),
            (dictgen::InsensitiveStr::Ascii("cereates"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("photograpic"),
                &["photographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invinsible"),
                &["invincible"],
            ),
            (dictgen::InsensitiveStr::Ascii("asissted"), &["assisted"]),
            (
                dictgen::InsensitiveStr::Ascii("priotised"),
                &["prioritised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bringtofont"),
                &["bringtofront"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefrancable"),
                &["dereferenceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("behviors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("embbeddings"),
                &["embeddings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apperantly"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatinating"),
                &["concatenating"],
            ),
            (dictgen::InsensitiveStr::Ascii("backfied"), &["backfield"]),
            (
                dictgen::InsensitiveStr::Ascii("perspecitve"),
                &["perspective"],
            ),
            (dictgen::InsensitiveStr::Ascii("aligins"), &["aligns"]),
            (
                dictgen::InsensitiveStr::Ascii("definitiions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("minum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("tempset"), &["tempest"]),
            (dictgen::InsensitiveStr::Ascii("ignors"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("girzzly"), &["grizzly"]),
            (
                dictgen::InsensitiveStr::Ascii("polypoylgons"),
                &["polypolygons"],
            ),
            (dictgen::InsensitiveStr::Ascii("addds"), &["adds"]),
            (dictgen::InsensitiveStr::Ascii("dipthong"), &["diphthong"]),
            (
                dictgen::InsensitiveStr::Ascii("tablulating"),
                &["tabulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experamenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaasure"), &["measure"]),
            (dictgen::InsensitiveStr::Ascii("exta"), &["extra"]),
            (
                dictgen::InsensitiveStr::Ascii("cassaroles"),
                &["casseroles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconeting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchaned"), &["unchanged"]),
            (dictgen::InsensitiveStr::Ascii("criterias"), &["criteria"]),
            (
                dictgen::InsensitiveStr::Ascii("distinguising"),
                &["distinguishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomadates"),
                &["accommodates"],
            ),
            (dictgen::InsensitiveStr::Ascii("obective"), &["objective"]),
            (dictgen::InsensitiveStr::Ascii("marnies"), &["marines"]),
            (dictgen::InsensitiveStr::Ascii("usal"), &["usual"]),
            (
                dictgen::InsensitiveStr::Ascii("insignificatly"),
                &["insignificantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("artical"), &["article"]),
            (dictgen::InsensitiveStr::Ascii("seperater"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("pathced"), &["patched"]),
            (
                dictgen::InsensitiveStr::Ascii("pagenation"),
                &["pagination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("streigtness"),
                &["straightness"],
            ),
            (dictgen::InsensitiveStr::Ascii("standarts"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("elemetn"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("overreactin"),
                &["overreaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthuseasticly"),
                &["enthusiastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsformer"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("deltion"), &["deletion"]),
            (dictgen::InsensitiveStr::Ascii("wokring"), &["working"]),
            (
                dictgen::InsensitiveStr::Ascii("consistutes"),
                &["constitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("suiteable"), &["suitable"]),
            (
                dictgen::InsensitiveStr::Ascii("specificaly"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("emac"), &["emacs"]),
            (dictgen::InsensitiveStr::Ascii("redundate"), &["redundant"]),
            (dictgen::InsensitiveStr::Ascii("nknown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("unencrpted"),
                &["unencrypted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordenated"),
                &["coordinated"],
            ),
            (dictgen::InsensitiveStr::Ascii("heastink"), &["heatsink"]),
            (
                dictgen::InsensitiveStr::Ascii("supresssion"),
                &["suppression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeltely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnostisism"),
                &["agnosticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algothitmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("occaisions"), &["occasions"]),
            (
                dictgen::InsensitiveStr::Ascii("suriously"),
                &[
                    "seriously",
                    "spuriously",
                    "curiously",
                    "furiously",
                    "usuriously",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("lengts"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("dialoge"),
                &["dialogue", "dialog"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopsitals"), &["hospitals"]),
            (
                dictgen::InsensitiveStr::Ascii("horizonals"),
                &["horizontals"],
            ),
            (dictgen::InsensitiveStr::Ascii("triggger"), &["trigger"]),
            (
                dictgen::InsensitiveStr::Ascii("choosed"),
                &["chose", "chosen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referting"),
                &["reverting", "referring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefice"),
                &["prefix", "preface"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impelmented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autosaveperodical"),
                &["autosaveperiodical"],
            ),
            (dictgen::InsensitiveStr::Ascii("intimide"), &["intimidate"]),
            (
                dictgen::InsensitiveStr::Ascii("algorritmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("opended"), &["opened"]),
            (dictgen::InsensitiveStr::Ascii("exhausive"), &["exhaustive"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerbuilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconviniently"),
                &["inconveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstending"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoriding"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("coaslescing"),
                &["coalescing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficities"),
                &["specificities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expad"),
                &["expand", "expat"],
            ),
            (dictgen::InsensitiveStr::Ascii("vedctor"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("picthers"), &["pitchers"]),
            (
                dictgen::InsensitiveStr::Ascii("bodybuilidng"),
                &["bodybuilding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congigured"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copurighted"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satistifying"),
                &["satisfying"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuslt"), &["result"]),
            (dictgen::InsensitiveStr::Ascii("endpounts"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("highting"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sectiont"),
                &["sectioned", "section"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminaters"),
                &["discriminators", "discriminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertexes"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("compilger"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("physcis"), &["physics"]),
            (
                dictgen::InsensitiveStr::Ascii("importent"),
                &["important", "impotent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cotributes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satifsying"),
                &["satisfying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gradiant"),
                &["gradient", "radiant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delcarator"),
                &["declarator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilizaitons"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("demolito"), &["demolition"]),
            (dictgen::InsensitiveStr::Ascii("checcksum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("combiing"),
                &["combining", "combing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnateeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciatly"), &["specialty"]),
            (
                dictgen::InsensitiveStr::Ascii("exceoptions"),
                &["exemptions", "exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("wimmen"), &["women"]),
            (dictgen::InsensitiveStr::Ascii("nintento"), &["nintendo"]),
            (dictgen::InsensitiveStr::Ascii("varoius"), &["various"]),
            (dictgen::InsensitiveStr::Ascii("exericse"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("anywys"), &["anyways"]),
            (
                dictgen::InsensitiveStr::Ascii("speicfying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("developres"), &["develops"]),
            (
                dictgen::InsensitiveStr::Ascii("particalarly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilitiy"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("aviailable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("performous"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("stifness"), &["stiffness"]),
            (dictgen::InsensitiveStr::Ascii("meges"), &["merges"]),
            (dictgen::InsensitiveStr::Ascii("ussing"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("bounsaries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("joinded"), &["joined"]),
            (dictgen::InsensitiveStr::Ascii("intrisnic"), &["intrinsic"]),
            (dictgen::InsensitiveStr::Ascii("aggrator"), &["aggregator"]),
            (
                dictgen::InsensitiveStr::Ascii("deplorabile"),
                &["deplorable"],
            ),
            (dictgen::InsensitiveStr::Ascii("presists"), &["persists"]),
            (
                dictgen::InsensitiveStr::Ascii("transacitons"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excisin"),
                &["excising", "excision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfactoin"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("counding"), &["counting"]),
            (dictgen::InsensitiveStr::Ascii("knarling"), &["gnarling"]),
            (dictgen::InsensitiveStr::Ascii("hurldes"), &["hurdles"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondece"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embodyment"),
                &["embodiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stubborness"),
                &["stubbornness"],
            ),
            (dictgen::InsensitiveStr::Ascii("tatgert"), &["target"]),
            (dictgen::InsensitiveStr::Ascii("seconfs"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("neighborgh"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("restting"),
                &["resetting", "resting"],
            ),
            (dictgen::InsensitiveStr::Ascii("roung"), &["round"]),
            (dictgen::InsensitiveStr::Ascii("propertus"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("mastereis"), &["masteries"]),
            (
                dictgen::InsensitiveStr::Ascii("terratorial"),
                &["territorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentially"),
                &["experimental", "experientially"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprated"), &["operated"]),
            (
                dictgen::InsensitiveStr::Ascii("downwoting"),
                &["downvoting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cousing"),
                &["cousin", "causing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("belows"),
                &["below", "bellows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declaratiom"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("listend"), &["listened"]),
            (
                dictgen::InsensitiveStr::Ascii("masculinty"),
                &["masculinity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrgators"),
                &["aggregators"],
            ),
            (dictgen::InsensitiveStr::Ascii("autonomos"), &["autonomous"]),
            (
                dictgen::InsensitiveStr::Ascii("crached"),
                &["crashed", "cached", "cracked", "coached"],
            ),
            (dictgen::InsensitiveStr::Ascii("integeral"), &["integral"]),
            (dictgen::InsensitiveStr::Ascii("sautays"), &["sautés"]),
            (dictgen::InsensitiveStr::Ascii("consrtuct"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("biger"), &["bigger"]),
            (dictgen::InsensitiveStr::Ascii("nwe"), &["new"]),
            (dictgen::InsensitiveStr::Ascii("heirachy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("millon"), &["million"]),
            (
                dictgen::InsensitiveStr::Ascii("recombiniation"),
                &["recombination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respnsibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("fromm"), &["from"]),
            (dictgen::InsensitiveStr::Ascii("messave"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("vews"), &["views", "vows"]),
            (dictgen::InsensitiveStr::Ascii("concerened"), &["concerned"]),
            (dictgen::InsensitiveStr::Ascii("sheduling"), &["scheduling"]),
            (dictgen::InsensitiveStr::Ascii("gloably"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("readeable"), &["readable"]),
            (
                dictgen::InsensitiveStr::Ascii("symettrical"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appricated"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proyections"),
                &["projections", "protections"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurions"), &["recursions"]),
            (dictgen::InsensitiveStr::Ascii("palcement"), &["placement"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponging"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("perperty"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("unimportent"),
                &["unimportant"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcion"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("pasteing"), &["pasting"]),
            (
                dictgen::InsensitiveStr::Ascii("selctions"),
                &["selections", "sections"],
            ),
            (dictgen::InsensitiveStr::Ascii("vlaues"), &["values"]),
            (
                dictgen::InsensitiveStr::Ascii("reicarnation"),
                &["reincarnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconviniences"),
                &["inconveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("reakpoint"), &["breakpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretating"),
                &["interpretation", "interpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("fieids"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("finallization"),
                &["finalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("platofrm"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("adviseable"), &["advisable"]),
            (dictgen::InsensitiveStr::Ascii("sponosred"), &["sponsored"]),
            (dictgen::InsensitiveStr::Ascii("trak"), &["track"]),
            (
                dictgen::InsensitiveStr::Ascii("fasens"),
                &["fasels", "fastens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutomisation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mercenarias"),
                &["mercenaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("divinition"),
                &["divination", "definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structrued"),
                &["structured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aletring"),
                &["alerting", "altering"],
            ),
            (dictgen::InsensitiveStr::Ascii("resembes"), &["resembles"]),
            (dictgen::InsensitiveStr::Ascii("anotated"), &["annotated"]),
            (dictgen::InsensitiveStr::Ascii("tabke"), &["table"]),
            (dictgen::InsensitiveStr::Ascii("englissh"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("relady"), &["ready"]),
            (dictgen::InsensitiveStr::Ascii("genialia"), &["genitalia"]),
            (dictgen::InsensitiveStr::Ascii("depented"), &["depended"]),
            (
                dictgen::InsensitiveStr::Ascii("serviving"),
                &["servicing", "surviving"],
            ),
            (dictgen::InsensitiveStr::Ascii("deposint"), &["deposing"]),
            (dictgen::InsensitiveStr::Ascii("positivies"), &["positives"]),
            (dictgen::InsensitiveStr::Ascii("doublded"), &["doubled"]),
            (
                dictgen::InsensitiveStr::Ascii("remarcably"),
                &["remarkably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgartes"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("hashi"), &["hash"]),
            (
                dictgen::InsensitiveStr::Ascii("ostraziced"),
                &["ostracized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sterotypes"),
                &["stereotypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrrespond"),
                &["correspond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progations"),
                &["propagations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stucturing"),
                &["structuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosisting"), &["consisting"]),
            (dictgen::InsensitiveStr::Ascii("asemblers"), &["assemblers"]),
            (
                dictgen::InsensitiveStr::Ascii("parallellize"),
                &["parallelize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepented"), &["prepended"]),
            (
                dictgen::InsensitiveStr::Ascii("condigures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("obtaion"), &["obtain"]),
            (dictgen::InsensitiveStr::Ascii("temporafy"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("investions"),
                &["inventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("reallocae"), &["reallocate"]),
            (
                dictgen::InsensitiveStr::Ascii("definently"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnigs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("dissmisive"),
                &["dismissive"],
            ),
            (dictgen::InsensitiveStr::Ascii("absrobing"), &["absorbing"]),
            (dictgen::InsensitiveStr::Ascii("utililty"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("retricts"),
                &["restricts", "retracts"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("confidenciality"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("iwll"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("paragarphs"),
                &["paragraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("separtes"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("relvancy"), &["relevancy"]),
            (dictgen::InsensitiveStr::Ascii("attitide"), &["attitude"]),
            (dictgen::InsensitiveStr::Ascii("yeasr"), &["years"]),
            (
                dictgen::InsensitiveStr::Ascii("apporoximates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeived"), &["received"]),
            (
                dictgen::InsensitiveStr::Ascii("relationshits"),
                &["relationships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boostrapping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("devlop"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("primarliy"), &["primarily"]),
            (
                dictgen::InsensitiveStr::Ascii("distrbutions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("fiorst"), &["first"]),
            (
                dictgen::InsensitiveStr::Ascii("compassione"),
                &["compassionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comphrehensive"),
                &["comprehensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenitive"),
                &["definitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coventionnel"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("fianl"), &["final"]),
            (dictgen::InsensitiveStr::Ascii("cuples"), &["couples"]),
            (dictgen::InsensitiveStr::Ascii("perforams"), &["performs"]),
            (
                dictgen::InsensitiveStr::Ascii("thaught"),
                &["taught", "thought"],
            ),
            (dictgen::InsensitiveStr::Ascii("prostitue"), &["prostitute"]),
            (dictgen::InsensitiveStr::Ascii("netiher"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("exteme"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("effieient"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("begines"), &["begins"]),
            (dictgen::InsensitiveStr::Ascii("evoluates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("ommitted"), &["omitted"]),
            (dictgen::InsensitiveStr::Ascii("dashs"), &["dashes"]),
            (dictgen::InsensitiveStr::Ascii("locaise"), &["localise"]),
            (
                dictgen::InsensitiveStr::Ascii("trigerring"),
                &["triggering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schmas"),
                &["schemas", "schwas"],
            ),
            (dictgen::InsensitiveStr::Ascii("evnet"), &["event"]),
            (
                dictgen::InsensitiveStr::Ascii("compillation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("roylaties"), &["royalties"]),
            (
                dictgen::InsensitiveStr::Ascii("influencian"),
                &["influencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bandiwdth"), &["bandwidth"]),
            (dictgen::InsensitiveStr::Ascii("traverce"), &["traverse"]),
            (
                dictgen::InsensitiveStr::Ascii("sypmtomatic"),
                &["symptomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("suttles"), &["shuttles"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotion"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transalator"),
                &["translator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialyze"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrresponds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("crossgne"), &["crossgen"]),
            (dictgen::InsensitiveStr::Ascii("insteadd"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("possibilies"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("purchse"), &["purchase"]),
            (
                dictgen::InsensitiveStr::Ascii("associateions"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confrontating"),
                &["confrontation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publcation"),
                &["publication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliverate"),
                &["deliberate"],
            ),
            (dictgen::InsensitiveStr::Ascii("distange"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("deivce"), &["device"]),
            (dictgen::InsensitiveStr::Ascii("marxisit"), &["marxist"]),
            (dictgen::InsensitiveStr::Ascii("villegas"), &["villages"]),
            (
                dictgen::InsensitiveStr::Ascii("wednessday"),
                &["wednesdays", "wednesday"],
            ),
            (dictgen::InsensitiveStr::Ascii("varagated"), &["variegated"]),
            (
                dictgen::InsensitiveStr::Ascii("unprecented"),
                &["unprecedented", "unpresented"],
            ),
            (dictgen::InsensitiveStr::Ascii("symols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("kittiens"), &["kitties"]),
            (dictgen::InsensitiveStr::Ascii("propperty"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("compeitions"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computarized"),
                &["computerized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenthicating"),
                &["authenticating"],
            ),
            (dictgen::InsensitiveStr::Ascii("relocae"), &["relocate"]),
            (
                dictgen::InsensitiveStr::Ascii("comparasons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquiesses"), &["acquiesces"]),
            (
                dictgen::InsensitiveStr::Ascii("insecuritites"),
                &["insecurities"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimized"), &["optimized"]),
            (dictgen::InsensitiveStr::Ascii("channcel"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("objtained"), &["obtained"]),
            (dictgen::InsensitiveStr::Ascii("hoeks"), &["hoax"]),
            (
                dictgen::InsensitiveStr::Ascii("intatiated"),
                &["instantiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exarcebated"),
                &["exacerbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrieveds"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("tought"),
                &["thought", "taught", "tough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialisere"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electon"),
                &["election", "electron"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("apreteated"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presciusly"),
                &["preciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcifics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("indended"),
                &["intended", "indented"],
            ),
            (dictgen::InsensitiveStr::Ascii("desciding"), &["deciding"]),
            (
                dictgen::InsensitiveStr::Ascii("initializtions"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thansparent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimenstionalities"),
                &["dimensionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("asethetic"), &["aesthetic"]),
            (
                dictgen::InsensitiveStr::Ascii("recontructed"),
                &["reconstructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbad"), &["forbade"]),
            (dictgen::InsensitiveStr::Ascii("warnin"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("nurish"), &["nourish"]),
            (
                dictgen::InsensitiveStr::Ascii("anonymsly"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrestircted"),
                &["unrestricted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statemenets"),
                &["statements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismatch"),
                &["mismatch", "dispatch"],
            ),
            (dictgen::InsensitiveStr::Ascii("battalin"), &["battalion"]),
            (dictgen::InsensitiveStr::Ascii("antennna"), &["antenna"]),
            (dictgen::InsensitiveStr::Ascii("goldifsh"), &["goldfish"]),
            (dictgen::InsensitiveStr::Ascii("labaelling"), &["labelling"]),
            (dictgen::InsensitiveStr::Ascii("boofay"), &["buffet"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondeingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("unverfied"), &["unverified"]),
            (dictgen::InsensitiveStr::Ascii("keybroads"), &["keyboards"]),
            (
                dictgen::InsensitiveStr::Ascii("bootlaoder"),
                &["bootloader"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restricitng"),
                &["restricting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affekt"),
                &["affect", "effect"],
            ),
            (dictgen::InsensitiveStr::Ascii("disimilar"), &["dissimilar"]),
            (dictgen::InsensitiveStr::Ascii("emmitted"), &["emitted"]),
            (
                dictgen::InsensitiveStr::Ascii("convienant"),
                &["convenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("howewer"), &["however"]),
            (dictgen::InsensitiveStr::Ascii("submoule"), &["submodule"]),
            (dictgen::InsensitiveStr::Ascii("itegrals"), &["integrals"]),
            (
                dictgen::InsensitiveStr::Ascii("consorcium"),
                &["consortium"],
            ),
            (dictgen::InsensitiveStr::Ascii("biseuxal"), &["bisexual"]),
            (dictgen::InsensitiveStr::Ascii("thaks"), &["thanks"]),
            (
                dictgen::InsensitiveStr::Ascii("implamenting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("faoming"), &["foaming"]),
            (
                dictgen::InsensitiveStr::Ascii("farenheight"),
                &["fahrenheit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offisionados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("deferal"), &["deferral"]),
            (
                dictgen::InsensitiveStr::Ascii("calrification"),
                &["clarification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinassance"),
                &["renaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("globa"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("coeffcients"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("explizit"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("expain"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("awekened"), &["weakened"]),
            (dictgen::InsensitiveStr::Ascii("tempalted"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("propertery"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("moderninationz"),
                &["modernizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolete"),
                &["absolute", "obsolete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complicatied"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactiristic"),
                &["characteristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootstraper"),
                &["bootstrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("lok"), &["look"]),
            (
                dictgen::InsensitiveStr::Ascii("endoresments"),
                &["endorsement"],
            ),
            (dictgen::InsensitiveStr::Ascii("inquries"), &["inquiries"]),
            (
                dictgen::InsensitiveStr::Ascii("circustantial"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("expresso"), &["espresso"]),
            (
                dictgen::InsensitiveStr::Ascii("progamatic"),
                &["programmatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("divereged"), &["diverged"]),
            (dictgen::InsensitiveStr::Ascii("aassumed"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("insturctor"),
                &["instructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("stirrin"), &["stirring"]),
            (dictgen::InsensitiveStr::Ascii("prabable"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("kyeboshing"), &["kiboshing"]),
            (
                dictgen::InsensitiveStr::Ascii("acquaintince"),
                &["acquaintance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impressons"),
                &["impressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("manyal"), &["manual"]),
            (dictgen::InsensitiveStr::Ascii("caclulate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisssist"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("anithing"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("eleminates"),
                &["eliminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("visbly"), &["visibly"]),
            (
                dictgen::InsensitiveStr::Ascii("thermostast"),
                &["thermostat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactivly"),
                &["interactively"],
            ),
            (dictgen::InsensitiveStr::Ascii("vrigins"), &["virgins"]),
            (dictgen::InsensitiveStr::Ascii("maraudeur"), &["marauder"]),
            (
                dictgen::InsensitiveStr::Ascii("prosthethic"),
                &["prosthetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("ttached"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnmission"),
                &["transmissions", "transmission"],
            ),
            (dictgen::InsensitiveStr::Ascii("accompt"), &["account"]),
            (
                dictgen::InsensitiveStr::Ascii("conseutive"),
                &["consecutive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphatising"),
                &["emphasising"],
            ),
            (dictgen::InsensitiveStr::Ascii("nutirents"), &["nutrients"]),
            (dictgen::InsensitiveStr::Ascii("monickers"), &["monikers"]),
            (dictgen::InsensitiveStr::Ascii("suboutine"), &["subroutine"]),
            (dictgen::InsensitiveStr::Ascii("rceate"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("peiece"), &["piece"]),
            (dictgen::InsensitiveStr::Ascii("glimspe"), &["glimpse"]),
            (dictgen::InsensitiveStr::Ascii("chater"), &["chapter"]),
            (
                dictgen::InsensitiveStr::Ascii("tanslators"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("vewers"), &["viewers"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpetected"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("caputuring"), &["capturing"]),
            (
                dictgen::InsensitiveStr::Ascii("witholding"),
                &["withholding"],
            ),
            (dictgen::InsensitiveStr::Ascii("releate"), &["relate"]),
            (dictgen::InsensitiveStr::Ascii("gometry"), &["geometry"]),
            (
                dictgen::InsensitiveStr::Ascii("definiation"),
                &["definition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contridictory"),
                &["contradictory"],
            ),
            (dictgen::InsensitiveStr::Ascii("gaphics"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("correespondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanis"),
                &["mechanics", "mechanism"],
            ),
            (dictgen::InsensitiveStr::Ascii("eralier"), &["earlier"]),
            (
                dictgen::InsensitiveStr::Ascii("contatenation"),
                &["concatenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("desogn"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("relead"), &["reload"]),
            (
                dictgen::InsensitiveStr::Ascii("exporing"),
                &["exploring", "expiring", "exporting", "exposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambassator"),
                &["ambassador"],
            ),
            (dictgen::InsensitiveStr::Ascii("vallayed"), &["valeted"]),
            (
                dictgen::InsensitiveStr::Ascii("guarnantes"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intterupts"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influensed"),
                &["influences"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarante"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("symbolizin"),
                &["symbolizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("municiple"), &["municipal"]),
            (dictgen::InsensitiveStr::Ascii("masos"), &["macos"]),
            (dictgen::InsensitiveStr::Ascii("waitres"), &["waiters"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitilized"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("immedaite"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("frementation"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("osbidian"), &["obsidian"]),
            (
                dictgen::InsensitiveStr::Ascii("themeselves"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("remtoe"), &["remote"]),
            (
                dictgen::InsensitiveStr::Ascii("toubleshooting"),
                &["troubleshooting"],
            ),
            (dictgen::InsensitiveStr::Ascii("gabagge"), &["garbage"]),
            (dictgen::InsensitiveStr::Ascii("appertures"), &["apertures"]),
            (
                dictgen::InsensitiveStr::Ascii("assocaiting"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("visiable"), &["visible"]),
            (dictgen::InsensitiveStr::Ascii("sydnrome"), &["syndrome"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailizer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramlisation"),
                &["normalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("buisiness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("skiped"),
                &["skipped", "skyped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informatonal"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("employeur"), &["employer"]),
            (dictgen::InsensitiveStr::Ascii("enebaling"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("conjucntion"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sugggestions"),
                &["suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelarator"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("pillaris"), &["pillars"]),
            (
                dictgen::InsensitiveStr::Ascii("routie"),
                &["routine", "route"],
            ),
            (dictgen::InsensitiveStr::Ascii("alunimum"), &["aluminum"]),
            (dictgen::InsensitiveStr::Ascii("enbracer"), &["embracer"]),
            (dictgen::InsensitiveStr::Ascii("initiaise"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("blittting"), &["blitting"]),
            (dictgen::InsensitiveStr::Ascii("counsils"), &["counsels"]),
            (dictgen::InsensitiveStr::Ascii("qwesant"), &["croissant"]),
            (
                dictgen::InsensitiveStr::Ascii("defiantley"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("fand"), &["fang", "find"]),
            (
                dictgen::InsensitiveStr::Ascii("prediccion"),
                &["prediction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confrentation"),
                &["confrontation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefations"),
                &["separations", "specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projetctions"),
                &["projections"],
            ),
            (dictgen::InsensitiveStr::Ascii("regio"), &["region"]),
            (
                dictgen::InsensitiveStr::Ascii("equivelantly"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supperssor"),
                &["suppressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispirsed"), &["dispersed"]),
            (
                dictgen::InsensitiveStr::Ascii("regularing"),
                &["regulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conciousness"),
                &["consciousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("conept"), &["concept"]),
            (dictgen::InsensitiveStr::Ascii("persumed"), &["presumed"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivy"),
                &["sensitivity", "sensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mischieveous"),
                &["mischievous"],
            ),
            (dictgen::InsensitiveStr::Ascii("modyfy"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("industires"),
                &["industries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compteting"),
                &["completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecctivities"),
                &["connectivities"],
            ),
            (dictgen::InsensitiveStr::Ascii("immenint"), &["imminent"]),
            (
                dictgen::InsensitiveStr::Ascii("conversaiton"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransactiosn"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentialy"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("backgrund"), &["background"]),
            (dictgen::InsensitiveStr::Ascii("embeding"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("eferences"),
                &["references", "preferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissecter"),
                &["dissector", "dissenter", "dissected", "dissect"],
            ),
            (dictgen::InsensitiveStr::Ascii("turnes"), &["turns"]),
            (
                dictgen::InsensitiveStr::Ascii("enchancement"),
                &["enhancement", "enchantment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architetcs"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotypeing"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strem"),
                &["stream", "stem", "strum", "strep"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protagonsit"),
                &["protagonists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extnesions"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downloadbale"),
                &["downloadable"],
            ),
            (dictgen::InsensitiveStr::Ascii("deugging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("escpaes"), &["escapes"]),
            (
                dictgen::InsensitiveStr::Ascii("persecutted"),
                &["persecuted"],
            ),
            (dictgen::InsensitiveStr::Ascii("stauts"), &["status"]),
            (dictgen::InsensitiveStr::Ascii("bootle"), &["bottle"]),
            (dictgen::InsensitiveStr::Ascii("intitiates"), &["initiates"]),
            (
                dictgen::InsensitiveStr::Ascii("distintively"),
                &["distinctively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resstrictions"),
                &["restrictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedomes"), &["freedoms"]),
            (
                dictgen::InsensitiveStr::Ascii("convinving"),
                &["convincing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wisper"), &["whisper"]),
            (
                dictgen::InsensitiveStr::Ascii("identiy"),
                &["identify", "identity"],
            ),
            (dictgen::InsensitiveStr::Ascii("altenates"), &["alternates"]),
            (
                dictgen::InsensitiveStr::Ascii("overhwelmed"),
                &["overwhelmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derefenrecing"),
                &["dereferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimnet"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanciate"),
                &["instantiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stimulatin"),
                &["stimulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnecs"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("mongoles"), &["mongols"]),
            (dictgen::InsensitiveStr::Ascii("partols"), &["patrols"]),
            (
                dictgen::InsensitiveStr::Ascii("intepolated"),
                &["interpolated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconection"),
                &["disconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("origianl"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("poperee"), &["potpourri"]),
            (
                dictgen::InsensitiveStr::Ascii("epecting"),
                &["expecting", "electing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurato"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asymmmetric"),
                &["asymmetric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aquainting"),
                &["acquainting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionaliy"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("settin"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("graniet"), &["granite"]),
            (dictgen::InsensitiveStr::Ascii("striaght"), &["straight"]),
            (dictgen::InsensitiveStr::Ascii("confucion"), &["confusion"]),
            (dictgen::InsensitiveStr::Ascii("warrned"), &["warned"]),
            (dictgen::InsensitiveStr::Ascii("accptors"), &["acceptors"]),
            (
                dictgen::InsensitiveStr::Ascii("kryptinite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convervatives"),
                &["conservatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expanations"),
                &["explanations", "expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approrpiations"),
                &["appropriations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dilaog"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("detet"),
                &["detect", "delete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roduces"),
                &["reduces", "produces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exacuted"),
                &["executed", "exacted"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrtibute"), &["attribute"]),
            (dictgen::InsensitiveStr::Ascii("dissbling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("chceked"), &["checked"]),
            (
                dictgen::InsensitiveStr::Ascii("perspecitves"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jsutification"),
                &["justifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ansalisation"),
                &["nasalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aspergerer"), &["asperger"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtutes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititially"),
                &["initially"],
            ),
            (dictgen::InsensitiveStr::Ascii("harware"), &["hardware"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplated"),
                &["deduplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peasing"),
                &["leasing", "piecing", "pleasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("geeneric"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("mainifestos"),
                &["manifestos"],
            ),
            (dictgen::InsensitiveStr::Ascii("capusles"), &["capsules"]),
            (dictgen::InsensitiveStr::Ascii("inherated"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("simplificacion"),
                &["simplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("submittted"), &["submitted"]),
            (dictgen::InsensitiveStr::Ascii("algorhytm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("chacheline"), &["cacheline"]),
            (
                dictgen::InsensitiveStr::Ascii("discconecting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("crowm"), &["crown"]),
            (dictgen::InsensitiveStr::Ascii("someoens"), &["someones"]),
            (dictgen::InsensitiveStr::Ascii("funxtions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("exeption"),
                &["exception", "exemption"],
            ),
            (dictgen::InsensitiveStr::Ascii("sautaying"), &["sautéing"]),
            (dictgen::InsensitiveStr::Ascii("asending"), &["ascending"]),
            (
                dictgen::InsensitiveStr::Ascii("disproportionatley"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("memics"), &["mimics"]),
            (dictgen::InsensitiveStr::Ascii("desturb"), &["disturb"]),
            (dictgen::InsensitiveStr::Ascii("repeatae"), &["repeatable"]),
            (dictgen::InsensitiveStr::Ascii("vehicels"), &["vehicles"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegoatiation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallucinatin"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vectoize"), &["vectorize"]),
            (
                dictgen::InsensitiveStr::Ascii("commericial"),
                &["commercial"],
            ),
            (dictgen::InsensitiveStr::Ascii("somentime"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("lentills"), &["lentils"]),
            (
                dictgen::InsensitiveStr::Ascii("addittion"),
                &["addition", "addiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concervatives"),
                &["conservatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prilileged"),
                &["privileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entusiastic"),
                &["enthusiastic"],
            ),
            (dictgen::InsensitiveStr::Ascii("realeasing"), &["releasing"]),
            (
                dictgen::InsensitiveStr::Ascii("throtte"),
                &["throttle", "trot"],
            ),
            (dictgen::InsensitiveStr::Ascii("thomspon"), &["thompson"]),
            (
                dictgen::InsensitiveStr::Ascii("classificato"),
                &["classification"],
            ),
            (dictgen::InsensitiveStr::Ascii("gignatic"), &["gigantic"]),
            (dictgen::InsensitiveStr::Ascii("sattelite"), &["satellite"]),
            (dictgen::InsensitiveStr::Ascii("anoying"), &["annoying"]),
            (
                dictgen::InsensitiveStr::Ascii("substitudes"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outsoursing"),
                &["outsourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciporical"),
                &["reciprocal"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignroing"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("archatypes"),
                &["archetypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertikal"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("insterrupted"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execurables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("stirkers"), &["strikers"]),
            (
                dictgen::InsensitiveStr::Ascii("objctively"),
                &["objectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegimacy"),
                &["illegitimacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("follfow"), &["follow"]),
            (
                dictgen::InsensitiveStr::Ascii("possiibility"),
                &["possibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("craption"), &["caption"]),
            (
                dictgen::InsensitiveStr::Ascii("particpates"),
                &["participates"],
            ),
            (dictgen::InsensitiveStr::Ascii("pinuckle"), &["pinochle"]),
            (dictgen::InsensitiveStr::Ascii("keyosk"), &["kiosk"]),
            (
                dictgen::InsensitiveStr::Ascii("stadnardized"),
                &["standardized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cann"), &["can"]),
            (dictgen::InsensitiveStr::Ascii("glaicer"), &["glacier"]),
            (dictgen::InsensitiveStr::Ascii("frankin"), &["franklin"]),
            (dictgen::InsensitiveStr::Ascii("hagases"), &["haggises"]),
            (
                dictgen::InsensitiveStr::Ascii("anihilation"),
                &["annihilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("heartbeart"), &["heartbeat"]),
            (dictgen::InsensitiveStr::Ascii("expexcting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("aforememtioned"),
                &["aforementioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alignements"),
                &["alignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdotel"), &["anecdotal"]),
            (dictgen::InsensitiveStr::Ascii("derivated"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("folowwiong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("debbuggers"), &["debuggers"]),
            (
                dictgen::InsensitiveStr::Ascii("desigin"),
                &["design", "designing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitude"),
                &["constitute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("piggypack"), &["piggyback"]),
            (dictgen::InsensitiveStr::Ascii("struggels"), &["struggles"]),
            (
                dictgen::InsensitiveStr::Ascii("dandidates"),
                &["candidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("apepars"), &["appears"]),
            (dictgen::InsensitiveStr::Ascii("unkowns"), &["unknowns"]),
            (
                dictgen::InsensitiveStr::Ascii("reouted"),
                &["routed", "rerouted"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoint"), &["anoint"]),
            (dictgen::InsensitiveStr::Ascii("follwind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("morbildy"), &["morbidly"]),
            (
                dictgen::InsensitiveStr::Ascii("blackahwks"),
                &["blackhawks"],
            ),
            (dictgen::InsensitiveStr::Ascii("lubicrant"), &["lubricant"]),
            (dictgen::InsensitiveStr::Ascii("witespace"), &["whitespace"]),
            (
                dictgen::InsensitiveStr::Ascii("directsions"),
                &["directions"],
            ),
            (dictgen::InsensitiveStr::Ascii("surprizes"), &["surprises"]),
            (
                dictgen::InsensitiveStr::Ascii("realationship"),
                &["relationship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifiation"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("alll"), &["all", "ally"]),
            (dictgen::InsensitiveStr::Ascii("marskman"), &["marksman"]),
            (
                dictgen::InsensitiveStr::Ascii("distinguishible"),
                &["distinguishable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitabil"),
                &["profitable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experianced"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("apender"), &["appender"]),
            (
                dictgen::InsensitiveStr::Ascii("windhsield"),
                &["windshield"],
            ),
            (dictgen::InsensitiveStr::Ascii("grillig"), &["grilling"]),
            (dictgen::InsensitiveStr::Ascii("fezent"), &["pheasant"]),
            (
                dictgen::InsensitiveStr::Ascii("criminalty"),
                &["criminally"],
            ),
            (dictgen::InsensitiveStr::Ascii("supspicion"), &["suspicion"]),
            (dictgen::InsensitiveStr::Ascii("actieve"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("hierachrically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seonds"),
                &["seconds", "sends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triskadecaphobia"),
                &["triskaidekaphobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyiser"), &["analyser"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstancial"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggestinos"),
                &["suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternetively"),
                &["alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynsaty"), &["dynasty"]),
            (
                dictgen::InsensitiveStr::Ascii("contentants"),
                &["contestants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deployd"),
                &["deployed", "deploy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deutshcland"),
                &["deutschland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substistutions"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("positiond"), &["positioned"]),
            (dictgen::InsensitiveStr::Ascii("retrned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("cartlage"), &["cartilage"]),
            (dictgen::InsensitiveStr::Ascii("optiional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("retquiresgsub"),
                &["requiresgsub"],
            ),
            (dictgen::InsensitiveStr::Ascii("docuent"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("indictrinated"),
                &["indoctrinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambassedors"),
                &["ambassadors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geneartors"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("epilgoue"), &["epilogue"]),
            (
                dictgen::InsensitiveStr::Ascii("asssociative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncalized"),
                &["canonicalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccuraces"),
                &["inaccuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manupulative"),
                &["manipulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromozome"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("insititute"), &["institute"]),
            (dictgen::InsensitiveStr::Ascii("returrns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("econmic"), &["economic"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticaors"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civalesation"),
                &["civilisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasformers"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglacting"),
                &["neglecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("sedereal"), &["sidereal"]),
            (
                dictgen::InsensitiveStr::Ascii("rectiinear"),
                &["rectilinear"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unleess"),
                &["unleash", "unless"],
            ),
            (dictgen::InsensitiveStr::Ascii("chatoic"), &["chaotic"]),
            (
                dictgen::InsensitiveStr::Ascii("converesly"),
                &["conversely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wheighting"),
                &["weighting", "weighing"],
            ),
            (dictgen::InsensitiveStr::Ascii("superiour"), &["superior"]),
            (
                dictgen::InsensitiveStr::Ascii("physcopathic"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("assigntee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("execuete"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("witholded"), &["withheld"]),
            (
                dictgen::InsensitiveStr::Ascii("formidabil"),
                &["formidable"],
            ),
            (dictgen::InsensitiveStr::Ascii("puragtory"), &["purgatory"]),
            (
                dictgen::InsensitiveStr::Ascii("commentarea"),
                &["commenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchrnous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("exractor"), &["extractor"]),
            (
                dictgen::InsensitiveStr::Ascii("bitterswet"),
                &["bittersweet"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragmant"), &["fragment"]),
            (
                dictgen::InsensitiveStr::Ascii("regconition"),
                &["recognition"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforeces"), &["enforces"]),
            (dictgen::InsensitiveStr::Ascii("teamfigth"), &["teamfight"]),
            (dictgen::InsensitiveStr::Ascii("yelds"), &["yields"]),
            (
                dictgen::InsensitiveStr::Ascii("bounradies"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substantie"),
                &["substantive"],
            ),
            (dictgen::InsensitiveStr::Ascii("realsie"), &["realise"]),
            (
                dictgen::InsensitiveStr::Ascii("accuring"),
                &["accruing", "occurring"],
            ),
            (dictgen::InsensitiveStr::Ascii("excliude"), &["exclude"]),
            (
                dictgen::InsensitiveStr::Ascii("resperitory"),
                &["respiratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afects"),
                &["affects", "effects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrationella"),
                &["irrational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modiles"),
                &["mobiles", "modules"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthuiasts"),
                &["enthusiast"],
            ),
            (dictgen::InsensitiveStr::Ascii("satistify"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("mechanichs"), &["mechanics"]),
            (
                dictgen::InsensitiveStr::Ascii("functinalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("attracs"), &["attracts"]),
            (
                dictgen::InsensitiveStr::Ascii("inresponsible"),
                &["irresponsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("funciotns"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("combinatorc"),
                &["combinatoric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaildation"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("relfexes"), &["reflexes"]),
            (dictgen::InsensitiveStr::Ascii("fuctioned"), &["functioned"]),
            (
                dictgen::InsensitiveStr::Ascii("reqiurements"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("clustred"), &["clustered"]),
            (dictgen::InsensitiveStr::Ascii("numerater"), &["numerator"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenetion"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priporitzes"),
                &["prioritizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("restoreing"), &["restoring"]),
            (dictgen::InsensitiveStr::Ascii("downrade"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("certifice"),
                &["certificate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associatation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successing"),
                &["succession", "successive", "succeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bottlebeck"),
                &["bottleneck"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("therapudic"),
                &["therapeutic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphirals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlatin"),
                &["correlating", "correlation"],
            ),
            (dictgen::InsensitiveStr::Ascii("microdode"), &["microcode"]),
            (
                dictgen::InsensitiveStr::Ascii("spectarors"),
                &["spectators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capatible"),
                &["compatible", "capable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qualtitative"),
                &["quantitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("revelas"), &["reveals"]),
            (dictgen::InsensitiveStr::Ascii("metifor"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("insepct"), &["inspect"]),
            (dictgen::InsensitiveStr::Ascii("coalaesce"), &["coalesce"]),
            (dictgen::InsensitiveStr::Ascii("rememver"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("likly"), &["likely"]),
            (dictgen::InsensitiveStr::Ascii("cavaets"), &["caveats"]),
            (
                dictgen::InsensitiveStr::Ascii("annhilation"),
                &["annihilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lcoalizing"),
                &["localizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dolhpins"), &["dolphins"]),
            (dictgen::InsensitiveStr::Ascii("cancling"), &["canceling"]),
            (dictgen::InsensitiveStr::Ascii("comnpleted"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("clampled"), &["clamped"]),
            (
                dictgen::InsensitiveStr::Ascii("stuructures"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("decremeant"), &["decrement"]),
            (dictgen::InsensitiveStr::Ascii("unbounnded"), &["unbounded"]),
            (
                dictgen::InsensitiveStr::Ascii("specilisation"),
                &["specialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insersecting"),
                &["intersecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrbuted"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("settlemnts"),
                &["settlements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alingments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relashionship"),
                &["relationship"],
            ),
            (dictgen::InsensitiveStr::Ascii("likek"), &["like"]),
            (
                dictgen::InsensitiveStr::Ascii("cooperatve"),
                &["cooperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpratations"),
                &["interpretations"],
            ),
            (dictgen::InsensitiveStr::Ascii("exra"), &["extra"]),
            (dictgen::InsensitiveStr::Ascii("upsreamer"), &["upstreamer"]),
            (dictgen::InsensitiveStr::Ascii("gerenic"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("technoligies"),
                &["technologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("oximoron"), &["oxymoron"]),
            (dictgen::InsensitiveStr::Ascii("bottomn"), &["bottom"]),
            (dictgen::InsensitiveStr::Ascii("respoded"), &["responded"]),
            (dictgen::InsensitiveStr::Ascii("perford"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("deleteion"), &["deletion"]),
            (dictgen::InsensitiveStr::Ascii("veryfies"), &["verifies"]),
            (dictgen::InsensitiveStr::Ascii("beliveing"), &["believing"]),
            (
                dictgen::InsensitiveStr::Ascii("expercted"),
                &["expected", "excerpted"],
            ),
            (dictgen::InsensitiveStr::Ascii("higest"), &["highest"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("developps"), &["develops"]),
            (
                dictgen::InsensitiveStr::Ascii("impelentations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ninteenth"), &["nineteenth"]),
            (
                dictgen::InsensitiveStr::Ascii("destinatins"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ontroller"), &["controller"]),
            (dictgen::InsensitiveStr::Ascii("opearors"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("bifgoot"), &["bigfoot"]),
            (
                dictgen::InsensitiveStr::Ascii("propabilistic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elipsis"),
                &["ellipsis", "eclipses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapprouval"),
                &["disapproval"],
            ),
            (dictgen::InsensitiveStr::Ascii("acedamia"), &["academia"]),
            (dictgen::InsensitiveStr::Ascii("numberic"), &["numeric"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotatiation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("backwwards"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("validting"), &["validating"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccurasies"),
                &["inaccuracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalessing"),
                &["coalescing"],
            ),
            (dictgen::InsensitiveStr::Ascii("physicst"), &["physicist"]),
            (
                dictgen::InsensitiveStr::Ascii("partiticular"),
                &["particular"],
            ),
            (dictgen::InsensitiveStr::Ascii("everyoneis"), &["everyones"]),
            (dictgen::InsensitiveStr::Ascii("groupd"), &["grouped"]),
            (dictgen::InsensitiveStr::Ascii("rpocess"), &["process"]),
            (dictgen::InsensitiveStr::Ascii("follwo"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("warnibgs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("lexicographycally"),
                &["lexicographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcocdes"),
                &["transcodes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressenting"),
                &["presenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("folse"), &["false"]),
            (
                dictgen::InsensitiveStr::Ascii("interfcaes"),
                &["interfaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positionaning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nitpciking"),
                &["nitpicking"],
            ),
            (dictgen::InsensitiveStr::Ascii("foudn"), &["found"]),
            (dictgen::InsensitiveStr::Ascii("assing"), &["assign"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenatel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonesensical"),
                &["nonsensical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ostrecized"),
                &["ostracized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyes"),
                &["analyse", "analyses", "analyzes", "analyze"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distriuted"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notoreously"),
                &["notoriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("arhiving"), &["archiving"]),
            (dictgen::InsensitiveStr::Ascii("debbugging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("fing"), &["find"]),
            (
                dictgen::InsensitiveStr::Ascii("termendously"),
                &["tremendously"],
            ),
            (dictgen::InsensitiveStr::Ascii("missings"), &["missing"]),
            (dictgen::InsensitiveStr::Ascii("disgracin"), &["disgracing"]),
            (
                dictgen::InsensitiveStr::Ascii("abbreveation"),
                &["abbreviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explossion"),
                &["explosions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mnenonic"), &["mnemonic"]),
            (
                dictgen::InsensitiveStr::Ascii("benchamrking"),
                &["benchmarking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("justificativo"),
                &["justification"],
            ),
            (dictgen::InsensitiveStr::Ascii("heighteen"), &["eighteen"]),
            (dictgen::InsensitiveStr::Ascii("decordings"), &["decodings"]),
            (
                dictgen::InsensitiveStr::Ascii("numeraters"),
                &["numerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangitude"), &["magnitude"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitriarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("worthwile"), &["worthwhile"]),
            (dictgen::InsensitiveStr::Ascii("proceeed"), &["proceed"]),
            (
                dictgen::InsensitiveStr::Ascii("initiailizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("nervanic"), &["nirvanic"]),
            (dictgen::InsensitiveStr::Ascii("renwal"), &["renewal"]),
            (
                dictgen::InsensitiveStr::Ascii("wallpapaers"),
                &["wallpapers"],
            ),
            (dictgen::InsensitiveStr::Ascii("terrritory"), &["territory"]),
            (
                dictgen::InsensitiveStr::Ascii("electormagnetic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpolayed"),
                &["interpolated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tidibt"), &["tidbit"]),
            (dictgen::InsensitiveStr::Ascii("auxillery"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("chuch"), &["church"]),
            (dictgen::InsensitiveStr::Ascii("resullt"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("blacksmitch"),
                &["blacksmith"],
            ),
            (dictgen::InsensitiveStr::Ascii("intendend"), &["intended"]),
            (dictgen::InsensitiveStr::Ascii("docmunet"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("bureauracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("ogerish"), &["ogreish"]),
            (dictgen::InsensitiveStr::Ascii("weekened"), &["weekend"]),
            (dictgen::InsensitiveStr::Ascii("editots"), &["editors"]),
            (dictgen::InsensitiveStr::Ascii("grabbin"), &["grabbing"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatabilty"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amelearating"),
                &["ameliorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("annotaing"), &["annotating"]),
            (
                dictgen::InsensitiveStr::Ascii("unifes"),
                &["unifies", "unites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trianwreck"),
                &["trainwreck"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversacion"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apppreciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleaered"), &["cleared"]),
            (
                dictgen::InsensitiveStr::Ascii("problematisch"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descrpition"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chcksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("curature"), &["curvature"]),
            (dictgen::InsensitiveStr::Ascii("generte"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("performanse"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("piars"),
                &["pairs", "piers", "pliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("structres"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("competitivo"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylsing"), &["analysing"]),
            (dictgen::InsensitiveStr::Ascii("resuilt"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("inevetably"),
                &["inevitably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aprreciate"),
                &["appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspong"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("dillema"), &["dilemma"]),
            (
                dictgen::InsensitiveStr::Ascii("misinfomed"),
                &["misinformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressolve"), &["resolve"]),
            (dictgen::InsensitiveStr::Ascii("hunagrian"), &["hungarian"]),
            (dictgen::InsensitiveStr::Ascii("acadamia"), &["academia"]),
            (
                dictgen::InsensitiveStr::Ascii("transcordes"),
                &["transcodes"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqiring"), &["requiring"]),
            (dictgen::InsensitiveStr::Ascii("polulates"), &["populates"]),
            (
                dictgen::InsensitiveStr::Ascii("massagebox"),
                &["messagebox"],
            ),
            (dictgen::InsensitiveStr::Ascii("assitsing"), &["assisting"]),
            (dictgen::InsensitiveStr::Ascii("itnerest"), &["interest"]),
            (
                dictgen::InsensitiveStr::Ascii("renderadble"),
                &["renderable"],
            ),
            (dictgen::InsensitiveStr::Ascii("raputre"), &["rapture"]),
            (
                dictgen::InsensitiveStr::Ascii("potententially"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("compones"), &["compose"]),
            (
                dictgen::InsensitiveStr::Ascii("negoitation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("erronoeusly"),
                &["erroneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignoreing"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("prejucides"),
                &["prejudices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilate"),
                &["compile", "complicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("valuess"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("cobining"), &["combining"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibililty"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("electonic"), &["electronic"]),
            (dictgen::InsensitiveStr::Ascii("monochrom"), &["monochrome"]),
            (
                dictgen::InsensitiveStr::Ascii("reachablity"),
                &["reachability"],
            ),
            (dictgen::InsensitiveStr::Ascii("calidate"), &["validate"]),
            (dictgen::InsensitiveStr::Ascii("prooperty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("highligt"), &["highlight"]),
            (dictgen::InsensitiveStr::Ascii("generallly"), &["generally"]),
            (
                dictgen::InsensitiveStr::Ascii("poposed"),
                &["proposed", "opposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("roahces"), &["roaches"]),
            (dictgen::InsensitiveStr::Ascii("optmimzie"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("deductibe"), &["deductible"]),
            (dictgen::InsensitiveStr::Ascii("seuqnece"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("filesysems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("ibuprofine"), &["ibuprofen"]),
            (
                dictgen::InsensitiveStr::Ascii("percentanges"),
                &["percentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("resiual"), &["residual"]),
            (dictgen::InsensitiveStr::Ascii("zealtos"), &["zealots"]),
            (dictgen::InsensitiveStr::Ascii("belongning"), &["belonging"]),
            (
                dictgen::InsensitiveStr::Ascii("aknowledges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expetiments"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("variablens"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("portarying"),
                &["portraying"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrfaced"), &["interfaced"]),
            (
                dictgen::InsensitiveStr::Ascii("asymetrically"),
                &["asymmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fixe"),
                &["fixed", "fixes", "fix", "fixme", "fixer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normailzed"),
                &["normalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("adminster"), &["administer"]),
            (
                dictgen::InsensitiveStr::Ascii("convoulted"),
                &["convoluted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aftermaket"),
                &["aftermarket"],
            ),
            (dictgen::InsensitiveStr::Ascii("domiance"), &["dominance"]),
            (
                dictgen::InsensitiveStr::Ascii("distribustion"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percaution"),
                &["precaution"],
            ),
            (dictgen::InsensitiveStr::Ascii("familes"), &["families"]),
            (
                dictgen::InsensitiveStr::Ascii("correcttness"),
                &["correctness"],
            ),
            (dictgen::InsensitiveStr::Ascii("enormass"), &["enormous"]),
            (
                dictgen::InsensitiveStr::Ascii("altetnative"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oveloading"),
                &["overloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissagreement"),
                &["disagreement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extansions"),
                &["extensions", "expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("punsihments"),
                &["punishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("stoppping"), &["stopping"]),
            (dictgen::InsensitiveStr::Ascii("simiarly"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("reconcille"), &["reconcile"]),
            (dictgen::InsensitiveStr::Ascii("sammon"), &["salmon"]),
            (
                dictgen::InsensitiveStr::Ascii("productivo"),
                &["production"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generattion"),
                &["generation"],
            ),
            (dictgen::InsensitiveStr::Ascii("anf"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("bceuase"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("clarifiy"), &["clarify"]),
            (
                dictgen::InsensitiveStr::Ascii("mormalizes"),
                &["normalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("poritioned"), &["portioned"]),
            (
                dictgen::InsensitiveStr::Ascii("simulteniously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtualisaiton"),
                &["virtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahinet"), &["cabinet"]),
            (dictgen::InsensitiveStr::Ascii("balasters"), &["balusters"]),
            (dictgen::InsensitiveStr::Ascii("bahaving"), &["behaving"]),
            (dictgen::InsensitiveStr::Ascii("libraly"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("approate"), &["appropriate"]),
            (dictgen::InsensitiveStr::Ascii("sentions"), &["sections"]),
            (dictgen::InsensitiveStr::Ascii("validitiy"), &["validity"]),
            (dictgen::InsensitiveStr::Ascii("uncommment"), &["uncomment"]),
            (
                dictgen::InsensitiveStr::Ascii("reaserching"),
                &["researching"],
            ),
            (dictgen::InsensitiveStr::Ascii("porshon"), &["portion"]),
            (dictgen::InsensitiveStr::Ascii("hamsphire"), &["hampshire"]),
            (dictgen::InsensitiveStr::Ascii("devlopper"), &["developer"]),
            (
                dictgen::InsensitiveStr::Ascii("imlemented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmar"),
                &["programmer"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreator"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("refocussed"), &["refocused"]),
            (
                dictgen::InsensitiveStr::Ascii("althorithms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("pithces"), &["pitches"]),
            (dictgen::InsensitiveStr::Ascii("referres"), &["refers"]),
            (
                dictgen::InsensitiveStr::Ascii("perfektion"),
                &["perfection"],
            ),
            (dictgen::InsensitiveStr::Ascii("sampl"), &["sample"]),
            (dictgen::InsensitiveStr::Ascii("comlpex"), &["complex"]),
            (dictgen::InsensitiveStr::Ascii("leeg"), &["league"]),
            (dictgen::InsensitiveStr::Ascii("spefeir"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("expexcts"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("construsts"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("deug"), &["debug"]),
            (dictgen::InsensitiveStr::Ascii("daugter"), &["daughter"]),
            (dictgen::InsensitiveStr::Ascii("indxes"), &["indexes"]),
            (dictgen::InsensitiveStr::Ascii("dragones"), &["dragons"]),
            (
                dictgen::InsensitiveStr::Ascii("depenending"),
                &["depending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calbration"),
                &["calibration"],
            ),
            (dictgen::InsensitiveStr::Ascii("miagic"), &["magic"]),
            (dictgen::InsensitiveStr::Ascii("rewirtten"), &["rewritten"]),
            (
                dictgen::InsensitiveStr::Ascii("homelessess"),
                &["homelessness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficillifers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("strians"), &["strains"]),
            (
                dictgen::InsensitiveStr::Ascii("deceleratin"),
                &["decelerating", "deceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resposnibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovelapped"), &["overlapped"]),
            (
                dictgen::InsensitiveStr::Ascii("experinementer"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depresssion"),
                &["depression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbalanaced"),
                &["unbalanced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effeciveness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicitors"),
                &["predictors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalant"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignament"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("pricipal"), &["principal"]),
            (dictgen::InsensitiveStr::Ascii("segwaying"), &["segueing"]),
            (dictgen::InsensitiveStr::Ascii("sanitizin"), &["sanitizing"]),
            (
                dictgen::InsensitiveStr::Ascii("counsilling"),
                &["counselling"],
            ),
            (dictgen::InsensitiveStr::Ascii("hideen"), &["hidden"]),
            (dictgen::InsensitiveStr::Ascii("confimed"), &["confirmed"]),
            (dictgen::InsensitiveStr::Ascii("evalulates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("platofmss"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("spearators"),
                &["separators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particpants"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enhanchements"),
                &["enhancements"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehnical"), &["technical"]),
            (
                dictgen::InsensitiveStr::Ascii("opertaional"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislikse"), &["dislikes"]),
            (
                dictgen::InsensitiveStr::Ascii("infiltratred"),
                &["infiltrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connetcion"),
                &["connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("comnpress"), &["compress"]),
            (dictgen::InsensitiveStr::Ascii("streamming"), &["streaming"]),
            (
                dictgen::InsensitiveStr::Ascii("standalown"),
                &["standalone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vietmanese"),
                &["vietnamese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertisiers"),
                &["advertisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("comliancy"), &["compliancy"]),
            (dictgen::InsensitiveStr::Ascii("tht"), &["the", "that"]),
            (
                dictgen::InsensitiveStr::Ascii("amealearates"),
                &["ameliorates"],
            ),
            (dictgen::InsensitiveStr::Ascii("retrivers"), &["retrievers"]),
            (dictgen::InsensitiveStr::Ascii("connetor"), &["connector"]),
            (dictgen::InsensitiveStr::Ascii("receivs"), &["receives"]),
            (dictgen::InsensitiveStr::Ascii("difusion"), &["diffusion"]),
            (
                dictgen::InsensitiveStr::Ascii("multitaksing"),
                &["multitasking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amelearators"),
                &["ameliorators"],
            ),
            (dictgen::InsensitiveStr::Ascii("browines"), &["brownies"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbborhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("sticthing"), &["stitching"]),
            (
                dictgen::InsensitiveStr::Ascii("arangement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("presssure"), &["pressure"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalistisk"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mambers"),
                &["members", "ambers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorditate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indentifier"),
                &["identifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprodicibility"),
                &["reproducibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("refcound"), &["refcount"]),
            (dictgen::InsensitiveStr::Ascii("alread"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("requestests"),
                &["requests", "requested"],
            ),
            (dictgen::InsensitiveStr::Ascii("derectly"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("discconenctions"),
                &["disconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("synthtic"), &["synthetic"]),
            (
                dictgen::InsensitiveStr::Ascii("formallized"),
                &["formalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacthes"), &["patches"]),
            (
                dictgen::InsensitiveStr::Ascii("desserialize"),
                &["deserialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcing"), &["producing"]),
            (
                dictgen::InsensitiveStr::Ascii("abstraktions"),
                &["abstractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phillipine"),
                &["philippine"],
            ),
            (dictgen::InsensitiveStr::Ascii("flacored"), &["flavored"]),
            (dictgen::InsensitiveStr::Ascii("accroding"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("vizualized"),
                &["visualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsbilities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralelise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("cassarole"), &["casserole"]),
            (dictgen::InsensitiveStr::Ascii("rigurous"), &["rigorous"]),
            (dictgen::InsensitiveStr::Ascii("slases"), &["slashes"]),
            (dictgen::InsensitiveStr::Ascii("intenals"), &["internals"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhhood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ellipitcals"),
                &["elliptical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessesity"),
                &["necessity"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarging"), &["guarding"]),
            (dictgen::InsensitiveStr::Ascii("apearance"), &["appearance"]),
            (dictgen::InsensitiveStr::Ascii("atttached"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("impelented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("ppeline"), &["pipeline"]),
            (
                dictgen::InsensitiveStr::Ascii("inclenation"),
                &["inclination"],
            ),
            (dictgen::InsensitiveStr::Ascii("sandwhich"), &["sandwich"]),
            (dictgen::InsensitiveStr::Ascii("chcekbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("intellectals"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjectivelly"),
                &["subjectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("definetly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("sectons"), &["sections"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiats"),
                &["enthusiast"],
            ),
            (dictgen::InsensitiveStr::Ascii("lithaunia"), &["lithuania"]),
            (dictgen::InsensitiveStr::Ascii("commitee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("coomit"), &["commit"]),
            (dictgen::InsensitiveStr::Ascii("succeeeds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("scrutinty"), &["scrutiny"]),
            (dictgen::InsensitiveStr::Ascii("sourecing"), &["sourcing"]),
            (
                dictgen::InsensitiveStr::Ascii("insensistive"),
                &["insensitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("aseuxal"), &["asexual"]),
            (
                dictgen::InsensitiveStr::Ascii("tindergarten"),
                &["kindergarten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorgithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metrolopis"),
                &["metropolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("basicliy"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("warnongs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnderabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("levaithan"), &["leviathan"]),
            (dictgen::InsensitiveStr::Ascii("calcultor"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("corrsepond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("scketch"), &["sketch"]),
            (
                dictgen::InsensitiveStr::Ascii("poer"),
                &["power", "poor", "pour"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("otimistically"),
                &["optimistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("midfeilder"),
                &["midfielder"],
            ),
            (dictgen::InsensitiveStr::Ascii("obesrver"), &["observer"]),
            (dictgen::InsensitiveStr::Ascii("nutriens"), &["nutrients"]),
            (
                dictgen::InsensitiveStr::Ascii("apporoximate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocilater"), &["oscillator"]),
            (
                dictgen::InsensitiveStr::Ascii("opearatios"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("orpanage"), &["orphanage"]),
            (dictgen::InsensitiveStr::Ascii("beautiy"), &["beauty"]),
            (dictgen::InsensitiveStr::Ascii("developors"), &["develops"]),
            (dictgen::InsensitiveStr::Ascii("exchaged"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("embeedded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("identicals"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("culmulates"),
                &["cumulates", "culminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("codeen"), &["codeine"]),
            (
                dictgen::InsensitiveStr::Ascii("intrisically"),
                &["intrinsically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anachrists"),
                &["anarchists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rectanfles"),
                &["rectangles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenaticators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("kyeboshed"), &["kiboshed"]),
            (dictgen::InsensitiveStr::Ascii("maxmum"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("laboraty"), &["laboratory"]),
            (
                dictgen::InsensitiveStr::Ascii("replacability"),
                &["replaceability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foreseeble"),
                &["foreseeable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cailbration"),
                &["calibration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reperesentations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinsoaur"), &["dinosaur"]),
            (dictgen::InsensitiveStr::Ascii("therby"), &["thereby"]),
            (dictgen::InsensitiveStr::Ascii("dispode"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("networksm"), &["networks"]),
            (dictgen::InsensitiveStr::Ascii("deviiate"), &["deviate"]),
            (
                dictgen::InsensitiveStr::Ascii("introverse"),
                &["introverts"],
            ),
            (dictgen::InsensitiveStr::Ascii("ampilfied"), &["amplified"]),
            (
                dictgen::InsensitiveStr::Ascii("sumbitting"),
                &["submitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consturcting"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("requrid"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("termperatures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("libraires"), &["libraries"]),
            (
                dictgen::InsensitiveStr::Ascii("imolicitly"),
                &["implicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenciators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("emphaises"), &["emphasises"]),
            (dictgen::InsensitiveStr::Ascii("desabled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("absrobs"), &["absorbs"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonnaire"),
                &["opinionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializier"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("splitted"), &["split"]),
            (dictgen::InsensitiveStr::Ascii("manifestas"), &["manifests"]),
            (dictgen::InsensitiveStr::Ascii("interript"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("nversion"), &["inversion"]),
            (dictgen::InsensitiveStr::Ascii("meassure"), &["measure"]),
            (
                dictgen::InsensitiveStr::Ascii("archetipes"),
                &["archetypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotatiors"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devides"),
                &["divides", "devices"],
            ),
            (dictgen::InsensitiveStr::Ascii("varibale"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("degeneratin"),
                &["degenerating", "degeneration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticallly"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroolbar"), &["scrollbar"]),
            (
                dictgen::InsensitiveStr::Ascii("homewoners"),
                &["homeowners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilability"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellegent"),
                &["intelligent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisaion"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parliamentry"),
                &["parliamentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("hierachy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("ajusted"), &["adjusted"]),
            (dictgen::InsensitiveStr::Ascii("byppases"), &["bypasses"]),
            (
                dictgen::InsensitiveStr::Ascii("monolitihic"),
                &["monolithic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inderictly"),
                &["indirectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("striaghtens"),
                &["straightens"],
            ),
            (dictgen::InsensitiveStr::Ascii("junglig"), &["jungling"]),
            (
                dictgen::InsensitiveStr::Ascii("chocloates"),
                &["chocolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autoctonous"),
                &["autochthonous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosecuters"),
                &["prosecutors"],
            ),
            (dictgen::InsensitiveStr::Ascii("skipe"), &["skip", "skype"]),
            (dictgen::InsensitiveStr::Ascii("cointainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("renegociated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("argreements"),
                &["agreements"],
            ),
            (dictgen::InsensitiveStr::Ascii("corruptd"), &["corrupted"]),
            (
                dictgen::InsensitiveStr::Ascii("tenatively"),
                &["tentatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransactional"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("thios"), &["this"]),
            (
                dictgen::InsensitiveStr::Ascii("systimatically"),
                &["systematically"],
            ),
            (dictgen::InsensitiveStr::Ascii("occations"), &["occasions"]),
            (
                dictgen::InsensitiveStr::Ascii("experices"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("semicolumn"), &["semicolon"]),
            (dictgen::InsensitiveStr::Ascii("bulletted"), &["bulleted"]),
            (dictgen::InsensitiveStr::Ascii("variablves"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("rythimic"), &["rhythmic"]),
            (
                dictgen::InsensitiveStr::Ascii("fonctionalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("actvie"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("belnd"), &["blend"]),
            (dictgen::InsensitiveStr::Ascii("brooz"), &["bruise"]),
            (
                dictgen::InsensitiveStr::Ascii("fration"),
                &["fraction", "ration"],
            ),
            (dictgen::InsensitiveStr::Ascii("renegated"), &["renegade"]),
            (dictgen::InsensitiveStr::Ascii("pachtches"), &["patches"]),
            (
                dictgen::InsensitiveStr::Ascii("viusalisation"),
                &["visualisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("councellor"),
                &["councillor", "counsellor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilizes"),
                &["capitalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("malfucntion"),
                &["malfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("symptons"), &["symptoms"]),
            (
                dictgen::InsensitiveStr::Ascii("comination"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pseudopoentials"),
                &["pseudopotentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("fertil"), &["fertile"]),
            (
                dictgen::InsensitiveStr::Ascii("autocoplete"),
                &["autocomplete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplificating"),
                &["simplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seletions"),
                &["selections", "deletions"],
            ),
            (dictgen::InsensitiveStr::Ascii("abanden"), &["abandon"]),
            (dictgen::InsensitiveStr::Ascii("everday"), &["everyday"]),
            (dictgen::InsensitiveStr::Ascii("periodos"), &["periods"]),
            (
                dictgen::InsensitiveStr::Ascii("introversi"),
                &["introverts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curerncies"),
                &["currencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("extry"), &["entry"]),
            (
                dictgen::InsensitiveStr::Ascii("agriculteur"),
                &["agriculture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissallowing"),
                &["disallowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifiles"),
                &["identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaining"), &["meaning"]),
            (
                dictgen::InsensitiveStr::Ascii("interrtuping"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundametal"),
                &["fundamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("tightely"), &["tightly"]),
            (dictgen::InsensitiveStr::Ascii("percepted"), &["perceived"]),
            (dictgen::InsensitiveStr::Ascii("attibutes"), &["attributes"]),
            (dictgen::InsensitiveStr::Ascii("anylsing"), &["analysing"]),
            (dictgen::InsensitiveStr::Ascii("patheitc"), &["pathetic"]),
            (dictgen::InsensitiveStr::Ascii("meancing"), &["menacing"]),
            (
                dictgen::InsensitiveStr::Ascii("relatonship"),
                &["relationship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separarting"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constituates"),
                &["constitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchronuous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("actiavte"), &["activate"]),
            (dictgen::InsensitiveStr::Ascii("represant"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("intterupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("exspect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("conroller"), &["controller"]),
            (dictgen::InsensitiveStr::Ascii("patchhing"), &["patching"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelize"),
                &["parallelize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particulary"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sergun"), &["surgeon"]),
            (dictgen::InsensitiveStr::Ascii("preffixing"), &["prefixing"]),
            (dictgen::InsensitiveStr::Ascii("cutscenses"), &["cutscene"]),
            (
                dictgen::InsensitiveStr::Ascii("outsorucing"),
                &["outsourcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("seclector"), &["selector"]),
            (
                dictgen::InsensitiveStr::Ascii("modifcation"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchrnous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("withount"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaitor"),
                &["renegotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("bogos"), &["bogus"]),
            (
                dictgen::InsensitiveStr::Ascii("confrontacion"),
                &["confrontation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("slectors"),
                &["selectors", "sectors", "electors"],
            ),
            (dictgen::InsensitiveStr::Ascii("strutture"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimmetry"),
                &["dissymmetry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrruptable"),
                &["corruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphatheater"),
                &["amphitheater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stetches"),
                &["stretches", "sketches", "stitches", "stenches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassinato"),
                &["assassination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialiaze"),
                &["specialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("brocher"), &["brochure"]),
            (dictgen::InsensitiveStr::Ascii("djikstra"), &["dijkstra"]),
            (dictgen::InsensitiveStr::Ascii("tages"), &["stages", "tags"]),
            (dictgen::InsensitiveStr::Ascii("genrators"), &["generators"]),
            (dictgen::InsensitiveStr::Ascii("insstance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("happans"), &["happens"]),
            (
                dictgen::InsensitiveStr::Ascii("comminicated"),
                &["communicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precuation"),
                &["precautions", "precaution"],
            ),
            (dictgen::InsensitiveStr::Ascii("entriess"), &["entries"]),
            (
                dictgen::InsensitiveStr::Ascii("hexidecmial"),
                &["hexadecimal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contitions"),
                &["conditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspongingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("revisting"), &["revisiting"]),
            (
                dictgen::InsensitiveStr::Ascii("reconsidder"),
                &["reconsider"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooperatin"),
                &["cooperating", "cooperation"],
            ),
            (dictgen::InsensitiveStr::Ascii("intuative"), &["intuitive"]),
            (dictgen::InsensitiveStr::Ascii("dsplayed"), &["displayed"]),
            (
                dictgen::InsensitiveStr::Ascii("versitality"),
                &["versatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unportected"),
                &["unprotected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingiushing"),
                &["distinguishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalness"),
                &["initialness"],
            ),
            (dictgen::InsensitiveStr::Ascii("siginify"), &["signify"]),
            (dictgen::InsensitiveStr::Ascii("exeecute"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("helikopters"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanetised"), &["sanitised"]),
            (dictgen::InsensitiveStr::Ascii("inhibates"), &["inhibits"]),
            (
                dictgen::InsensitiveStr::Ascii("devoloping"),
                &["developing"],
            ),
            (dictgen::InsensitiveStr::Ascii("machiness"), &["machines"]),
            (
                dictgen::InsensitiveStr::Ascii("origionally"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("triggerred"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("funtioning"),
                &["functioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("confugred"), &["configured"]),
            (dictgen::InsensitiveStr::Ascii("reprtoire"), &["repertoire"]),
            (dictgen::InsensitiveStr::Ascii("dispayed"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("suspensie"), &["suspense"]),
            (
                dictgen::InsensitiveStr::Ascii("inconcistent"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("culmulative"),
                &["cumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasrers"),
                &["parsers", "passers"],
            ),
            (dictgen::InsensitiveStr::Ascii("chech"), &["check", "czech"]),
            (dictgen::InsensitiveStr::Ascii("continum"), &["continuum"]),
            (dictgen::InsensitiveStr::Ascii("scrtip"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("cosole"), &["console"]),
            (dictgen::InsensitiveStr::Ascii("updadate"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("eagrely"), &["eagerly"]),
            (
                dictgen::InsensitiveStr::Ascii("accessisble"),
                &["accessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nesesarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("tahnking"), &["thanking"]),
            (
                dictgen::InsensitiveStr::Ascii("embarressment"),
                &["embarrassment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precided"),
                &["preceded", "presided"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeliavable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("pshycotic"), &["psychotic"]),
            (dictgen::InsensitiveStr::Ascii("mimmick"), &["mimic"]),
            (dictgen::InsensitiveStr::Ascii("preordes"), &["preorders"]),
            (dictgen::InsensitiveStr::Ascii("ojbected"), &["objected"]),
            (
                dictgen::InsensitiveStr::Ascii("cconfiguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("earleist"), &["earliest"]),
            (dictgen::InsensitiveStr::Ascii("keyevente"), &["keyevent"]),
            (dictgen::InsensitiveStr::Ascii("isloation"), &["isolation"]),
            (dictgen::InsensitiveStr::Ascii("datyatype"), &["datatype"]),
            (
                dictgen::InsensitiveStr::Ascii("assimliated"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tommorw"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("unreacheable"),
                &["unreachable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coppies"),
                &["copies", "coppices"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupicates"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("propriatery"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("incerased"), &["increased"]),
            (
                dictgen::InsensitiveStr::Ascii("expierencing"),
                &["experiencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("satarts"), &["starts"]),
            (dictgen::InsensitiveStr::Ascii("polymorh"), &["polymorph"]),
            (
                dictgen::InsensitiveStr::Ascii("depcitions"),
                &["depictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonmyous"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("appoached"), &["approached"]),
            (dictgen::InsensitiveStr::Ascii("creaetor"), &["creator"]),
            (
                dictgen::InsensitiveStr::Ascii("disencouraged"),
                &["discouraged"],
            ),
            (dictgen::InsensitiveStr::Ascii("parrameter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("accesibility"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpeption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("scrpts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("bratenders"),
                &["bartenders"],
            ),
            (dictgen::InsensitiveStr::Ascii("seprated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("neightobring"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("baray"), &["beret"]),
            (dictgen::InsensitiveStr::Ascii("passabel"), &["passable"]),
            (
                dictgen::InsensitiveStr::Ascii("technicien"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermettent"),
                &["intermittent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proejctive"),
                &["projective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interstela"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothecis"),
                &["hypothesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("caucasain"), &["caucasian"]),
            (dictgen::InsensitiveStr::Ascii("battailon"), &["battalion"]),
            (dictgen::InsensitiveStr::Ascii("preprepared"), &["prepared"]),
            (dictgen::InsensitiveStr::Ascii("dereive"), &["derive"]),
            (
                dictgen::InsensitiveStr::Ascii("enumertaion"),
                &["enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seralizing"),
                &["serializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tilda"), &["tilde"]),
            (dictgen::InsensitiveStr::Ascii("requset"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("obsucrity"), &["obscurity"]),
            (
                dictgen::InsensitiveStr::Ascii("schematrio"),
                &["schematron"],
            ),
            (dictgen::InsensitiveStr::Ascii("abrupty"), &["abruptly"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistentcy"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("thiks"), &["thinks"]),
            (
                dictgen::InsensitiveStr::Ascii("competetions"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulipliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("uupdate"), &["update"]),
            (
                dictgen::InsensitiveStr::Ascii("permissin"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interuppts"),
                &["interrupts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivudually"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prodctions"),
                &["productions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("productin"),
                &["productions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicapility"),
                &["applicability"],
            ),
            (dictgen::InsensitiveStr::Ascii("skalar"), &["scalar"]),
            (dictgen::InsensitiveStr::Ascii("renmant"), &["remnant"]),
            (
                dictgen::InsensitiveStr::Ascii("propegation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impressin"),
                &["impressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("sequencess"), &["sequences"]),
            (
                dictgen::InsensitiveStr::Ascii("omisiences"),
                &["omnisciences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declratively"),
                &["declaratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapperarance"),
                &["disappearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virualises"),
                &["virtualises", "visualises"],
            ),
            (dictgen::InsensitiveStr::Ascii("spendour"), &["splendour"]),
            (dictgen::InsensitiveStr::Ascii("sedning"), &["sending"]),
            (dictgen::InsensitiveStr::Ascii("dectect"), &["detect"]),
            (dictgen::InsensitiveStr::Ascii("checling"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("plcaed"), &["placed"]),
            (dictgen::InsensitiveStr::Ascii("bypasing"), &["bypassing"]),
            (
                dictgen::InsensitiveStr::Ascii("infrigning"),
                &["infringing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inproductive"),
                &["unproductive"],
            ),
            (dictgen::InsensitiveStr::Ascii("critized"), &["criticized"]),
            (dictgen::InsensitiveStr::Ascii("wayword"), &["wayward"]),
            (dictgen::InsensitiveStr::Ascii("relatd"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("terminiations"),
                &["terminations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganerator"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("relateds"), &["relates"]),
            (dictgen::InsensitiveStr::Ascii("prarie"), &["prairie"]),
            (
                dictgen::InsensitiveStr::Ascii("mainenance"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overloades"),
                &["overloads", "overloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("thjese"), &["these"]),
            (
                dictgen::InsensitiveStr::Ascii("trianglulation"),
                &["triangulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libguistics"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("confise"), &["confines"]),
            (dictgen::InsensitiveStr::Ascii("edibels"), &["edibles"]),
            (dictgen::InsensitiveStr::Ascii("lmiting"), &["limiting"]),
            (dictgen::InsensitiveStr::Ascii("adjacant"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("capitolise"),
                &["capitalise"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcltly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("sovle"), &["solve"]),
            (
                dictgen::InsensitiveStr::Ascii("negociated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("formualas"), &["formulas"]),
            (
                dictgen::InsensitiveStr::Ascii("oprimizers"),
                &["optimizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulalry"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("deceivin"), &["deceiving"]),
            (
                dictgen::InsensitiveStr::Ascii("vewiership"),
                &["viewership"],
            ),
            (dictgen::InsensitiveStr::Ascii("maping"), &["mapping"]),
            (dictgen::InsensitiveStr::Ascii("prassing"), &["passing"]),
            (dictgen::InsensitiveStr::Ascii("recurssive"), &["recursive"]),
            (
                dictgen::InsensitiveStr::Ascii("hamburgesa"),
                &["hamburgers"],
            ),
            (dictgen::InsensitiveStr::Ascii("deconde"), &["decode"]),
            (dictgen::InsensitiveStr::Ascii("overal"), &["overall"]),
            (dictgen::InsensitiveStr::Ascii("aboue"), &["about", "above"]),
            (dictgen::InsensitiveStr::Ascii("desribes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("satisfactorally"),
                &["satisfactory", "satisfactorily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallellising"),
                &["parallelising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confugrations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleetely"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("registrers"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("desctiptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("escalted"), &["escalated"]),
            (
                dictgen::InsensitiveStr::Ascii("configuraion"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("functionsn"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("friendzoneado"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("upported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("soloution"), &["solution"]),
            (dictgen::InsensitiveStr::Ascii("attiribute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("thermostaat"),
                &["thermostat"],
            ),
            (dictgen::InsensitiveStr::Ascii("categogy"), &["category"]),
            (
                dictgen::InsensitiveStr::Ascii("curvelinear"),
                &["curvilinear"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skatebaording"),
                &["skateboarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriubtion"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialgo"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("concersion"),
                &["concession", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("seledted"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("togather"), &["together"]),
            (dictgen::InsensitiveStr::Ascii("totorial"), &["tutorial"]),
            (
                dictgen::InsensitiveStr::Ascii("uable"),
                &["unable", "usable", "table", "able"],
            ),
            (dictgen::InsensitiveStr::Ascii("coulmns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("occaison"), &["occasion"]),
            (dictgen::InsensitiveStr::Ascii("accsepter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("exchanching"),
                &["exchanging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espeically"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("heigth"), &["height"]),
            (
                dictgen::InsensitiveStr::Ascii("merceneries"),
                &["mercenaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conexts"),
                &["contexts", "connects"],
            ),
            (dictgen::InsensitiveStr::Ascii("timoeut"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("connecticuit"),
                &["connecticut"],
            ),
            (dictgen::InsensitiveStr::Ascii("modfifiers"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("overlcocked"),
                &["overclocked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replaceemnts"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentaiton"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("schedue"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("documment"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("materails"), &["materials"]),
            (dictgen::InsensitiveStr::Ascii("labed"), &["labeled"]),
            (dictgen::InsensitiveStr::Ascii("grobal"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("quantititive"),
                &["quantitative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circulaton"),
                &["circulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("interstae"), &["interstate"]),
            (
                dictgen::InsensitiveStr::Ascii("decompostion"),
                &["decomposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimenison"), &["dimension"]),
            (dictgen::InsensitiveStr::Ascii("stylin"), &["styling"]),
            (dictgen::InsensitiveStr::Ascii("thann"), &["than", "thank"]),
            (
                dictgen::InsensitiveStr::Ascii("enchance"),
                &["enhance", "enchant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conitnuing"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polylon"),
                &["polygon", "pylon"],
            ),
            (dictgen::InsensitiveStr::Ascii("macors"), &["macros"]),
            (dictgen::InsensitiveStr::Ascii("lodaer"), &["loader"]),
            (dictgen::InsensitiveStr::Ascii("civillian"), &["civilian"]),
            (
                dictgen::InsensitiveStr::Ascii("trangulation"),
                &["strangulation", "triangulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("molestered"), &["molested"]),
            (dictgen::InsensitiveStr::Ascii("feauturing"), &["featuring"]),
            (dictgen::InsensitiveStr::Ascii("templats"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("golbins"), &["goblins"]),
            (
                dictgen::InsensitiveStr::Ascii("accoutnants"),
                &["accountants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unorigional"),
                &["unoriginal"],
            ),
            (dictgen::InsensitiveStr::Ascii("refereces"), &["references"]),
            (
                dictgen::InsensitiveStr::Ascii("specficially"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crockodiles"),
                &["crocodiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("fysical"), &["physical"]),
            (dictgen::InsensitiveStr::Ascii("matrie"), &["matrix"]),
            (dictgen::InsensitiveStr::Ascii("numberous"), &["numerous"]),
            (
                dictgen::InsensitiveStr::Ascii("comporesses"),
                &["compresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essesitaly"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relisation"),
                &["realisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("doens"), &["does", "doesn"]),
            (
                dictgen::InsensitiveStr::Ascii("manoeuverability"),
                &["maneuverability"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeuing"), &["queuing"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitialze"),
                &["uninitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probly"),
                &["probably", "provably"],
            ),
            (dictgen::InsensitiveStr::Ascii("fullfills"), &["fulfills"]),
            (dictgen::InsensitiveStr::Ascii("sentenses"), &["sentences"]),
            (dictgen::InsensitiveStr::Ascii("teamates"), &["teammates"]),
            (dictgen::InsensitiveStr::Ascii("minimisin"), &["minimising"]),
            (
                dictgen::InsensitiveStr::Ascii("speficeirs"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sagital"), &["sagittal"]),
            (dictgen::InsensitiveStr::Ascii("disatnce"), &["distance"]),
            (dictgen::InsensitiveStr::Ascii("nubered"), &["numbered"]),
            (dictgen::InsensitiveStr::Ascii("secrion"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("fileter"), &["filter"]),
            (
                dictgen::InsensitiveStr::Ascii("paralleization"),
                &["parallelization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announcemnet"),
                &["announcements", "announcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("expecected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciaes"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbreakabie"),
                &["unbreakable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuss"),
                &["confuse", "confess", "confuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifiled"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("otimised"), &["optimised"]),
            (
                dictgen::InsensitiveStr::Ascii("unabailability"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsisntency"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersecction"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compensative"),
                &["compensate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("didsapointed"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("suuports"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("ciphes"), &["ciphers"]),
            (
                dictgen::InsensitiveStr::Ascii("peroidicals"),
                &["periodicals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palestininas"),
                &["palestinians"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciever"), &["receiver"]),
            (
                dictgen::InsensitiveStr::Ascii("kollaboration"),
                &["collaboration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ligh"),
                &["light", "lie", "lye"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommanded"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("excepetion"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterised"),
                &["parameterised"],
            ),
            (dictgen::InsensitiveStr::Ascii("szenes"), &["scenes"]),
            (dictgen::InsensitiveStr::Ascii("gormay"), &["gourmet"]),
            (dictgen::InsensitiveStr::Ascii("beyong"), &["beyond"]),
            (
                dictgen::InsensitiveStr::Ascii("acomodating"),
                &["accommodating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disproportionaltely"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("certanty"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("platforma"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("alphanuermic"),
                &["alphanumeric"],
            ),
            (dictgen::InsensitiveStr::Ascii("handledy"), &["handedly"]),
            (
                dictgen::InsensitiveStr::Ascii("privaticed"),
                &["privatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retpresenting"),
                &["representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mone"),
                &["mono", "money", "none"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addittional"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("fracional"), &["fractional"]),
            (
                dictgen::InsensitiveStr::Ascii("indianapols"),
                &["indianapolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("loadig"), &["loading"]),
            (dictgen::InsensitiveStr::Ascii("direclty"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("pairocheel"), &["parochial"]),
            (dictgen::InsensitiveStr::Ascii("iy"), &["it"]),
            (
                dictgen::InsensitiveStr::Ascii("unneceesary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rerpesentation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deinitalization"),
                &["deinitialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("authoizes"), &["authorizes"]),
            (
                dictgen::InsensitiveStr::Ascii("writter"),
                &["writer", "written"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabinent"), &["cabinet"]),
            (
                dictgen::InsensitiveStr::Ascii("isntrumental"),
                &["instrumental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("eaisly"), &["easily"]),
            (
                dictgen::InsensitiveStr::Ascii("temrinating"),
                &["terminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("midpiont"), &["midpoint"]),
            (dictgen::InsensitiveStr::Ascii("beastiary"), &["bestiary"]),
            (
                dictgen::InsensitiveStr::Ascii("unarchvied"),
                &["unarchived"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unassinged"),
                &["unassigned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highligters"),
                &["highlighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("repors"), &["reports"]),
            (
                dictgen::InsensitiveStr::Ascii("hullucination"),
                &["hallucination"],
            ),
            (dictgen::InsensitiveStr::Ascii("colummns"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("executon"),
                &["execution", "executor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterisation"),
                &["parameterisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjustements"),
                &["adjustments"],
            ),
            (dictgen::InsensitiveStr::Ascii("menmonic"), &["mnemonic"]),
            (dictgen::InsensitiveStr::Ascii("relatiate"), &["retaliate"]),
            (dictgen::InsensitiveStr::Ascii("grantie"), &["granite"]),
            (dictgen::InsensitiveStr::Ascii("functuon"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("dieing"),
                &["dying", "dyeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorproate"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("vould"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("completess"),
                &["completeness", "completes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eligiblity"),
                &["eligibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superflous"),
                &["superfluous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tighetning"),
                &["tightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extermists"),
                &["extremists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recard"),
                &["record", "regard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisitic"),
                &["characteristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recocation"),
                &["revocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poposes"),
                &["proposes", "opposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independant"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spere"),
                &["sphere", "spare", "spire", "spore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermarkers"),
                &["supermarkets"],
            ),
            (dictgen::InsensitiveStr::Ascii("witin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("childs"), &["children"]),
            (
                dictgen::InsensitiveStr::Ascii("subisdized"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cuztomize"), &["customize"]),
            (dictgen::InsensitiveStr::Ascii("refrence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("upgerading"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("doubldes"), &["doubles"]),
            (dictgen::InsensitiveStr::Ascii("wil"), &["will", "well"]),
            (dictgen::InsensitiveStr::Ascii("portaits"), &["portraits"]),
            (dictgen::InsensitiveStr::Ascii("cacahe"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("tallets"), &["tallest"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhhods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("inteface"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("inherets"), &["inherits"]),
            (
                dictgen::InsensitiveStr::Ascii("followe"),
                &["follow", "followed"],
            ),
            (dictgen::InsensitiveStr::Ascii("seelected"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("worshoping"),
                &["worshiping"],
            ),
            (dictgen::InsensitiveStr::Ascii("anad"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("dentifier"), &["identifier"]),
            (dictgen::InsensitiveStr::Ascii("elemen"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangting"),
                &["rearranging"],
            ),
            (dictgen::InsensitiveStr::Ascii("itegrates"), &["integrates"]),
            (dictgen::InsensitiveStr::Ascii("inferioare"), &["inferior"]),
            (dictgen::InsensitiveStr::Ascii("extenion"), &["extension"]),
            (dictgen::InsensitiveStr::Ascii("changuing"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("disillutioned"),
                &["disillusioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("aposlte"), &["apostle"]),
            (dictgen::InsensitiveStr::Ascii("langurages"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("fialogs"), &["dialogs"]),
            (
                dictgen::InsensitiveStr::Ascii("insatlling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoule"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("xeonphobic"),
                &["xenophobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknonws"), &["unknowns"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomitted"),
                &["uncommitted"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaetions"), &["creations"]),
            (
                dictgen::InsensitiveStr::Ascii("intergations"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("exitt"), &["exit"]),
            (dictgen::InsensitiveStr::Ascii("idealogy"), &["ideology"]),
            (dictgen::InsensitiveStr::Ascii("aktivity"), &["activity"]),
            (
                dictgen::InsensitiveStr::Ascii("aftter"),
                &["after", "fatter"],
            ),
            (dictgen::InsensitiveStr::Ascii("reoported"), &["reported"]),
            (
                dictgen::InsensitiveStr::Ascii("settelment"),
                &["settlement"],
            ),
            (dictgen::InsensitiveStr::Ascii("fininshed"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("logiacl"), &["logical"]),
            (dictgen::InsensitiveStr::Ascii("tonguers"), &["tongues"]),
            (dictgen::InsensitiveStr::Ascii("committin"), &["committing"]),
            (dictgen::InsensitiveStr::Ascii("saterdays"), &["saturdays"]),
            (dictgen::InsensitiveStr::Ascii("mayalsian"), &["malaysian"]),
            (
                dictgen::InsensitiveStr::Ascii("deprectating"),
                &["deprecating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizonally"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("yatching"), &["yachting"]),
            (
                dictgen::InsensitiveStr::Ascii("atuhenticating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alhapetical"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoticators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supspicious"),
                &["suspicious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conclussion"),
                &["conclusions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chizelling"),
                &["chiselling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limtiation"),
                &["limitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhouds"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practitioneer"),
                &["practitioners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscrive"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mutilplies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goegraphical"),
                &["geographical"],
            ),
            (dictgen::InsensitiveStr::Ascii("wuthout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("quanities"),
                &["quantities", "qualities"],
            ),
            (dictgen::InsensitiveStr::Ascii("taskelt"), &["tasklet"]),
            (dictgen::InsensitiveStr::Ascii("cntroller"), &["controller"]),
            (dictgen::InsensitiveStr::Ascii("preiod"), &["period"]),
            (dictgen::InsensitiveStr::Ascii("priorize"), &["prioritize"]),
            (
                dictgen::InsensitiveStr::Ascii("initisliase"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsforming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ophtalmology"),
                &["ophthalmology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetn"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeming"), &["opening"]),
            (dictgen::InsensitiveStr::Ascii("replics"), &["replicas"]),
            (dictgen::InsensitiveStr::Ascii("plagerise"), &["plagiarise"]),
            (dictgen::InsensitiveStr::Ascii("unplayabe"), &["unplayable"]),
            (
                dictgen::InsensitiveStr::Ascii("experiemntal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("preivous"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("modifiy"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("eorrors"), &["errors"]),
            (dictgen::InsensitiveStr::Ascii("appologise"), &["apologise"]),
            (dictgen::InsensitiveStr::Ascii("reacahble"), &["reachable"]),
            (dictgen::InsensitiveStr::Ascii("intested"), &["interested"]),
            (
                dictgen::InsensitiveStr::Ascii("whiltelisting"),
                &["whitelisting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmunicator"),
                &["communicator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agggregations"),
                &["aggregations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associatin"),
                &["associations", "associating", "association"],
            ),
            (dictgen::InsensitiveStr::Ascii("orpans"), &["orphans"]),
            (dictgen::InsensitiveStr::Ascii("youo"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("cologen"), &["cologne"]),
            (dictgen::InsensitiveStr::Ascii("familiy"), &["family"]),
            (
                dictgen::InsensitiveStr::Ascii("determinato"),
                &["determination"],
            ),
            (dictgen::InsensitiveStr::Ascii("everyting"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("checkebox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("captued"), &["captured"]),
            (
                dictgen::InsensitiveStr::Ascii("congresman"),
                &["congressman"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactetrs"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhhod"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("polulate"), &["populate"]),
            (
                dictgen::InsensitiveStr::Ascii("proprition"),
                &["proportion"],
            ),
            (dictgen::InsensitiveStr::Ascii("transiets"), &["transients"]),
            (
                dictgen::InsensitiveStr::Ascii("corrisponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southernese"),
                &["southerners"],
            ),
            (dictgen::InsensitiveStr::Ascii("seflies"), &["selfies"]),
            (
                dictgen::InsensitiveStr::Ascii("sepecially"),
                &["especially", "specially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmations"),
                &["summations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mergeing"), &["merging"]),
            (dictgen::InsensitiveStr::Ascii("genric"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("homogeniously"),
                &["homogeneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrogato"),
                &["interrogation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinforcemnets"),
                &["reinforcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lexigraphical"),
                &["lexicographical"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifable"), &["modifiable"]),
            (dictgen::InsensitiveStr::Ascii("systemisch"), &["systemic"]),
            (dictgen::InsensitiveStr::Ascii("watn"), &["want"]),
            (dictgen::InsensitiveStr::Ascii("tetss"), &["tests"]),
            (dictgen::InsensitiveStr::Ascii("campares"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("contentss"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("browswing"), &["browsing"]),
            (dictgen::InsensitiveStr::Ascii("prooduced"), &["produced"]),
            (
                dictgen::InsensitiveStr::Ascii("expectatoins"),
                &["expectations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discoteque"),
                &["discotheque"],
            ),
            (dictgen::InsensitiveStr::Ascii("deregiter"), &["deregister"]),
            (
                dictgen::InsensitiveStr::Ascii("sexualixed"),
                &["sexualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accodomate"),
                &["accommodate"],
            ),
            (dictgen::InsensitiveStr::Ascii("pevention"), &["prevention"]),
            (dictgen::InsensitiveStr::Ascii("shorcuts"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("hypervior"), &["hypervisor"]),
            (dictgen::InsensitiveStr::Ascii("ambiguuous"), &["ambiguous"]),
            (dictgen::InsensitiveStr::Ascii("deomcrats"), &["democrats"]),
            (dictgen::InsensitiveStr::Ascii("exhausion"), &["exhaustion"]),
            (dictgen::InsensitiveStr::Ascii("mexicanus"), &["mexicans"]),
            (dictgen::InsensitiveStr::Ascii("releafs"), &["relieves"]),
            (
                dictgen::InsensitiveStr::Ascii("resgistration"),
                &["registration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intructions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("imense"), &["immense"]),
            (
                dictgen::InsensitiveStr::Ascii("intialsation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignroe"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("belever"), &["believer"]),
            (
                dictgen::InsensitiveStr::Ascii("bureacuracy"),
                &["bureaucracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("balconny"), &["balcony"]),
            (
                dictgen::InsensitiveStr::Ascii("unimpemented"),
                &["unimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependning"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("fahernheit"),
                &["fahrenheit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliasation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("metahpor"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("returmed"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("critizised"),
                &["criticised"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheteher"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("inapropriately"),
                &["inappropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("defintive"), &["definitive"]),
            (
                dictgen::InsensitiveStr::Ascii("diggest"),
                &["digest", "biggest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sceintifically"),
                &["scientifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("fizeek"), &["physique"]),
            (dictgen::InsensitiveStr::Ascii("cylical"), &["cyclical"]),
            (dictgen::InsensitiveStr::Ascii("sclar"), &["scalar"]),
            (dictgen::InsensitiveStr::Ascii("ignornd"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("tomarrow"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("horizontall"),
                &["horizontal", "horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extactions"),
                &["extractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecution"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("occrrance"), &["occurrence"]),
            (dictgen::InsensitiveStr::Ascii("knwos"), &["knows"]),
            (
                dictgen::InsensitiveStr::Ascii("prevolence"),
                &["prevalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("aiplanes"), &["airplanes"]),
            (dictgen::InsensitiveStr::Ascii("staisfies"), &["satisfies"]),
            (dictgen::InsensitiveStr::Ascii("assumtion"), &["assumption"]),
            (
                dictgen::InsensitiveStr::Ascii("undefinitely"),
                &["indefinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("confiured"), &["configured"]),
            (dictgen::InsensitiveStr::Ascii("encourges"), &["encourages"]),
            (
                dictgen::InsensitiveStr::Ascii("deficientcy"),
                &["deficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifiing"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("toggeles"), &["toggles"]),
            (dictgen::InsensitiveStr::Ascii("goind"), &["going"]),
            (dictgen::InsensitiveStr::Ascii("boyfried"), &["boyfriend"]),
            (dictgen::InsensitiveStr::Ascii("proejcts"), &["projects"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiants"),
                &["enthusiast"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provsioned"),
                &["provisioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifictions"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("getoe"), &["ghetto"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangemenet"),
                &["arrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiers"),
                &["initializers", "initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("renegae"), &["renegade"]),
            (
                dictgen::InsensitiveStr::Ascii("collaboratin"),
                &["collaboration"],
            ),
            (dictgen::InsensitiveStr::Ascii("chasim"), &["chasm"]),
            (
                dictgen::InsensitiveStr::Ascii("concioussness"),
                &["consciousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termperatue"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("milion"), &["million"]),
            (dictgen::InsensitiveStr::Ascii("propt"), &["prompt"]),
            (
                dictgen::InsensitiveStr::Ascii("bombardement"),
                &["bombardment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplorabil"),
                &["deplorable"],
            ),
            (dictgen::InsensitiveStr::Ascii("malplace"), &["misplace"]),
            (
                dictgen::InsensitiveStr::Ascii("polarlizing"),
                &["polarizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mindlessely"),
                &["mindlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("onn"), &["on"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriato"),
                &["appropriation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aperures"), &["apertures"]),
            (
                dictgen::InsensitiveStr::Ascii("modificaton"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("survivavility"),
                &["survivability"],
            ),
            (dictgen::InsensitiveStr::Ascii("achievs"), &["achieves"]),
            (
                dictgen::InsensitiveStr::Ascii("philippinoes"),
                &["philippines"],
            ),
            (dictgen::InsensitiveStr::Ascii("repulisve"), &["repulsive"]),
            (dictgen::InsensitiveStr::Ascii("liqudity"), &["liquidity"]),
            (
                dictgen::InsensitiveStr::Ascii("perentages"),
                &["percentages", "parentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("bandwith"), &["bandwidth"]),
            (
                dictgen::InsensitiveStr::Ascii("programmier"),
                &["programmer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stabilzied"),
                &["stabilized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opinyonator"),
                &["opinionator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("getfastproperyvalue"),
                &["getfastpropertyvalue"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceptul"), &["conceptual"]),
            (
                dictgen::InsensitiveStr::Ascii("understadns"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatset"), &["dataset"]),
            (dictgen::InsensitiveStr::Ascii("gerat"), &["great"]),
            (
                dictgen::InsensitiveStr::Ascii("liousville"),
                &["louisville"],
            ),
            (dictgen::InsensitiveStr::Ascii("speeped"), &["sped"]),
            (dictgen::InsensitiveStr::Ascii("arbtrary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("socred"),
                &["scored", "sacred"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakistnai"), &["pakistani"]),
            (dictgen::InsensitiveStr::Ascii("deketes"), &["deletes"]),
            (dictgen::InsensitiveStr::Ascii("selecctors"), &["selectors"]),
            (dictgen::InsensitiveStr::Ascii("strager"), &["stranger"]),
            (
                dictgen::InsensitiveStr::Ascii("internatinalist"),
                &["internationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("aleinated"), &["alienated"]),
            (
                dictgen::InsensitiveStr::Ascii("trustworthyness"),
                &["trustworthiness"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepterad"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("biscuts"), &["biscuits"]),
            (
                dictgen::InsensitiveStr::Ascii("congradulating"),
                &["congratulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("addos"), &["addons"]),
            (dictgen::InsensitiveStr::Ascii("strean"), &["stream"]),
            (dictgen::InsensitiveStr::Ascii("expec"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("noramlizer"),
                &["normalizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabilitiies"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertie"),
                &["property", "properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("valudes"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("owuldve"), &["wouldve"]),
            (
                dictgen::InsensitiveStr::Ascii("pedophilles"),
                &["pedophile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effordlessly"),
                &["effortlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifcied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("negothiated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("journing"), &["journeying"]),
            (
                dictgen::InsensitiveStr::Ascii("transportaion"),
                &["transportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cummulative"),
                &["cumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promotheus"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("illinoians"), &["illinois"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilyses"),
                &["cannibalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjonction"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiams"),
                &["enthusiasm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresores"),
                &["compressor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pormetheus"),
                &["prometheus"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("litllefinger"),
                &["littlefinger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowldegable"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fingerpies"),
                &["fingertips"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwwiong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("toriodal"), &["toroidal"]),
            (dictgen::InsensitiveStr::Ascii("termporary"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("depolyments"),
                &["deployments"],
            ),
            (dictgen::InsensitiveStr::Ascii("indciates"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("conditoinally"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optomistic"),
                &["optimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("committe"), &["committee"]),
            (
                dictgen::InsensitiveStr::Ascii("contravercial"),
                &["controversial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detremining"),
                &["determining"],
            ),
            (dictgen::InsensitiveStr::Ascii("pabel"), &["panel"]),
            (
                dictgen::InsensitiveStr::Ascii("rapresenting"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("asbestoes"), &["asbestos"]),
            (dictgen::InsensitiveStr::Ascii("somethigng"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("coeffiecients"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("annouce"), &["announce"]),
            (dictgen::InsensitiveStr::Ascii("approuving"), &["approving"]),
            (
                dictgen::InsensitiveStr::Ascii("wieghtlifting"),
                &["weightlifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("cautius"), &["cautious"]),
            (dictgen::InsensitiveStr::Ascii("nuetron"), &["neutron"]),
            (
                dictgen::InsensitiveStr::Ascii("circomvent"),
                &["circumvent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhibicion"),
                &["exhibition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementtion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramters"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("heuristcs"), &["heuristics"]),
            (dictgen::InsensitiveStr::Ascii("quizs"), &["quizzes"]),
            (
                dictgen::InsensitiveStr::Ascii("indciators"),
                &["indicators"],
            ),
            (dictgen::InsensitiveStr::Ascii("freame"), &["frame"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("belevers"), &["believers"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomative"),
                &["performative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instansiates"),
                &["instantiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotnain"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("createad"), &["created"]),
            (
                dictgen::InsensitiveStr::Ascii("preccedents"),
                &["precedents"],
            ),
            (dictgen::InsensitiveStr::Ascii("contiues"), &["continues"]),
            (
                dictgen::InsensitiveStr::Ascii("devine"),
                &["define", "divine"],
            ),
            (dictgen::InsensitiveStr::Ascii("aborption"), &["absorption"]),
            (
                dictgen::InsensitiveStr::Ascii("commen"),
                &["commend", "comment", "common"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprimizer"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("dynamicpsf"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("conumes"), &["consumes"]),
            (
                dictgen::InsensitiveStr::Ascii("impovements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparititively"),
                &["comparatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("populair"), &["popular"]),
            (dictgen::InsensitiveStr::Ascii("dirctions"), &["directions"]),
            (
                dictgen::InsensitiveStr::Ascii("accessiable"),
                &["accessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("previious"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("altanta"), &["atlanta"]),
            (dictgen::InsensitiveStr::Ascii("skooled"), &["schooled"]),
            (dictgen::InsensitiveStr::Ascii("insatntly"), &["instantly"]),
            (dictgen::InsensitiveStr::Ascii("cerberal"), &["cerebral"]),
            (
                dictgen::InsensitiveStr::Ascii("supersticion"),
                &["superstition"],
            ),
            (dictgen::InsensitiveStr::Ascii("acuracies"), &["accuracies"]),
            (dictgen::InsensitiveStr::Ascii("ojbection"), &["objection"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowldged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("historycal"),
                &["historical"],
            ),
            (dictgen::InsensitiveStr::Ascii("viralence"), &["virulence"]),
            (
                dictgen::InsensitiveStr::Ascii("distributin"),
                &["distribution", "distributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordinag"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("sinle"), &["single"]),
            (dictgen::InsensitiveStr::Ascii("regalar"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("parameteric"),
                &["parametric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvieniences"),
                &["inconveniences"],
            ),
            (dictgen::InsensitiveStr::Ascii("motiviated"), &["motivated"]),
            (dictgen::InsensitiveStr::Ascii("brwoser"), &["browser"]),
            (dictgen::InsensitiveStr::Ascii("amiguity"), &["ambiguity"]),
            (dictgen::InsensitiveStr::Ascii("apsect"), &["aspect"]),
            (dictgen::InsensitiveStr::Ascii("cathloic"), &["catholic"]),
            (
                dictgen::InsensitiveStr::Ascii("excutables"),
                &["executables"],
            ),
            (dictgen::InsensitiveStr::Ascii("stealthly"), &["stealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("withrdawing"),
                &["withdrawing"],
            ),
            (dictgen::InsensitiveStr::Ascii("arranes"), &["arranges"]),
            (dictgen::InsensitiveStr::Ascii("everythin"), &["everything"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnearbility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highjacking"),
                &["hijacking"],
            ),
            (dictgen::InsensitiveStr::Ascii("explotied"), &["exploited"]),
            (dictgen::InsensitiveStr::Ascii("everyons"), &["everyones"]),
            (dictgen::InsensitiveStr::Ascii("patronis"), &["patrons"]),
            (dictgen::InsensitiveStr::Ascii("cpation"), &["caption"]),
            (
                dictgen::InsensitiveStr::Ascii("revoltuionary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("santiy"), &["sanity"]),
            (dictgen::InsensitiveStr::Ascii("executung"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("parctical"), &["practical"]),
            (dictgen::InsensitiveStr::Ascii("retriece"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("identfier"), &["identifier"]),
            (
                dictgen::InsensitiveStr::Ascii("camoflagues"),
                &["camouflages"],
            ),
            (dictgen::InsensitiveStr::Ascii("accaptors"), &["acceptors"]),
            (
                dictgen::InsensitiveStr::Ascii("declars"),
                &["declares", "declared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrcutions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("versioms"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("tuesdey"), &["tuesday"]),
            (
                dictgen::InsensitiveStr::Ascii("reuploders"),
                &["reuploaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiminated"),
                &["intimidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("raterizing"),
                &["rasterizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("productie"), &["productive"]),
            (dictgen::InsensitiveStr::Ascii("extracor"), &["extractor"]),
            (dictgen::InsensitiveStr::Ascii("deffine"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("mouvements"), &["movements"]),
            (
                dictgen::InsensitiveStr::Ascii("priviliging"),
                &["privileging"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchenge"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("appearent"), &["apparent"]),
            (
                dictgen::InsensitiveStr::Ascii("correpondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterperted"),
                &["misinterpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifyers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scientiss"),
                &["scientist", "scientists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonpriviledged"),
                &["nonprivileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intercection"),
                &["interception"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condencing"),
                &["condemning"],
            ),
            (dictgen::InsensitiveStr::Ascii("parterns"), &["partners"]),
            (
                dictgen::InsensitiveStr::Ascii("connetcing"),
                &["connecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conslutants"),
                &["consultant"],
            ),
            (dictgen::InsensitiveStr::Ascii("trully"), &["truly"]),
            (
                dictgen::InsensitiveStr::Ascii("avdisoriy"),
                &["advisory", "advisories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermiented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("morbidley"), &["morbidly"]),
            (
                dictgen::InsensitiveStr::Ascii("resoultions"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abrreviations"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seleted"),
                &["selected", "deleted"],
            ),
            (dictgen::InsensitiveStr::Ascii("vokume"), &["volume"]),
            (
                dictgen::InsensitiveStr::Ascii("candidiates"),
                &["candidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("superseed"), &["supersede"]),
            (dictgen::InsensitiveStr::Ascii("unweildly"), &["unwieldy"]),
            (dictgen::InsensitiveStr::Ascii("veryified"), &["verified"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficientss"),
                &["coefficients"],
            ),
            (dictgen::InsensitiveStr::Ascii("remvs"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("occupided"), &["occupied"]),
            (
                dictgen::InsensitiveStr::Ascii("depandence"),
                &["dependence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disturbace"),
                &["disturbance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overreidden"),
                &["overridden"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutionale"),
                &["institutionalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepatator"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("parokeelly"),
                &["parochially"],
            ),
            (dictgen::InsensitiveStr::Ascii("ssame"), &["same"]),
            (dictgen::InsensitiveStr::Ascii("ringotne"), &["ringtone"]),
            (
                dictgen::InsensitiveStr::Ascii("interprer"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misteryous"),
                &["mysterious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entartaining"),
                &["entertaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chainged"),
                &["changed", "chained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("linguistis"),
                &["linguistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replicaites"),
                &["replicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("agianst"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("displacemnts"),
                &["displacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenenticates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partiitons"),
                &["partitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplmenting"),
                &["reimplementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervieuw"),
                &["interviews"],
            ),
            (dictgen::InsensitiveStr::Ascii("sircle"), &["circle"]),
            (
                dictgen::InsensitiveStr::Ascii("parentheesis"),
                &["parenthesis"],
            ),
            (dictgen::InsensitiveStr::Ascii("auotuning"), &["autotuning"]),
            (
                dictgen::InsensitiveStr::Ascii("adolencent"),
                &["adolescent"],
            ),
            (dictgen::InsensitiveStr::Ascii("qustion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstrucion"),
                &["reconstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conffigured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("experesses"), &["expresses"]),
            (
                dictgen::InsensitiveStr::Ascii("homecuming"),
                &["homecoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repote"),
                &["report", "remote", "repute"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcty"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("blackcawks"),
                &["blackhawks"],
            ),
            (dictgen::InsensitiveStr::Ascii("fileding"), &["fielding"]),
            (
                dictgen::InsensitiveStr::Ascii("restricitve"),
                &["restrictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("dictacte"), &["dictate"]),
            (dictgen::InsensitiveStr::Ascii("neessesary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("insticnts"), &["instincts"]),
            (dictgen::InsensitiveStr::Ascii("inifity"), &["infinity"]),
            (
                dictgen::InsensitiveStr::Ascii("excusively"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("runnnig"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("freeed"), &["freed"]),
            (dictgen::InsensitiveStr::Ascii("presing"), &["pressing"]),
            (dictgen::InsensitiveStr::Ascii("chcek"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("preseves"), &["preserves"]),
            (
                dictgen::InsensitiveStr::Ascii("anonimusly"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amalagmation"),
                &["amalgamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("degenracy"), &["degeneracy"]),
            (dictgen::InsensitiveStr::Ascii("uniquness"), &["uniqueness"]),
            (dictgen::InsensitiveStr::Ascii("resverve"), &["reserve"]),
            (
                dictgen::InsensitiveStr::Ascii("incompatibile"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fucntional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brosed"),
                &["browsed", "rosed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millioniares"),
                &["millionaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proably"),
                &["probably", "provably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperfectionists"),
                &["imperfections"],
            ),
            (dictgen::InsensitiveStr::Ascii("constists"), &["consists"]),
            (
                dictgen::InsensitiveStr::Ascii("classifcation"),
                &["classification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evalutating"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconenctions"),
                &["disconnections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sirectionals"),
                &["directionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("powre"), &["power"]),
            (dictgen::InsensitiveStr::Ascii("enchanct"), &["enchant"]),
            (
                dictgen::InsensitiveStr::Ascii("disconctinued"),
                &["discontinued"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versatily"),
                &["versatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("receptical"),
                &["receptacle"],
            ),
            (dictgen::InsensitiveStr::Ascii("elememt"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("certiticate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchnage"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("chormosome"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("combiens"), &["combines"]),
            (dictgen::InsensitiveStr::Ascii("constrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("expoentntially"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physcially"),
                &["physically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mexicain"), &["mexican"]),
            (dictgen::InsensitiveStr::Ascii("desirin"), &["desiring"]),
            (dictgen::InsensitiveStr::Ascii("fromtier"), &["frontier"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetuaters"),
                &["perpetuates"],
            ),
            (dictgen::InsensitiveStr::Ascii("normalis"), &["normals"]),
            (
                dictgen::InsensitiveStr::Ascii("elimanates"),
                &["eliminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sastifying"),
                &["satisfying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicaiton"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrtized"),
                &["discretized"],
            ),
            (dictgen::InsensitiveStr::Ascii("perseeds"), &["precedes"]),
            (dictgen::InsensitiveStr::Ascii("encrpyt"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("reuslting"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("interuptable"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrepicies"),
                &["discrepancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("shicane"), &["chicane"]),
            (dictgen::InsensitiveStr::Ascii("integirty"), &["integrity"]),
            (
                dictgen::InsensitiveStr::Ascii("inefficency"),
                &["inefficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("toubleshoots"),
                &["troubleshoots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("changgeling"),
                &["changeling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsequnce"),
                &["subsequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("capibly"), &["capably"]),
            (dictgen::InsensitiveStr::Ascii("hipocrisy"), &["hypocrisy"]),
            (dictgen::InsensitiveStr::Ascii("virginus"), &["virgins"]),
            (dictgen::InsensitiveStr::Ascii("simarly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("introduktion"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("openned"), &["opened"]),
            (
                dictgen::InsensitiveStr::Ascii("excemptions"),
                &["exemptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exectuives"),
                &["executives"],
            ),
            (dictgen::InsensitiveStr::Ascii("chubks"), &["chunks"]),
            (
                dictgen::InsensitiveStr::Ascii("protraying"),
                &["portraying"],
            ),
            (dictgen::InsensitiveStr::Ascii("hestiant"), &["hesitant"]),
            (dictgen::InsensitiveStr::Ascii("cartensian"), &["cartesian"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegment"),
                &["acknowledgment"],
            ),
            (dictgen::InsensitiveStr::Ascii("swithcing"), &["switching"]),
            (dictgen::InsensitiveStr::Ascii("mariens"), &["marines"]),
            (dictgen::InsensitiveStr::Ascii("likeley"), &["likely"]),
            (
                dictgen::InsensitiveStr::Ascii("foramtters"),
                &["formatters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkeerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alteracion"),
                &["alteration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gererating"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nclusive"),
                &["inclusive", "conclusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomblishable"),
                &["accomplishable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silkscreend"),
                &["silkscreened"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouding"), &["bounding"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("freits"), &["freights"]),
            (
                dictgen::InsensitiveStr::Ascii("checkcpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perseverence"),
                &["perseverance"],
            ),
            (dictgen::InsensitiveStr::Ascii("indentical"), &["identical"]),
            (
                dictgen::InsensitiveStr::Ascii("continuacion"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("appricate"), &["appreciate"]),
            (
                dictgen::InsensitiveStr::Ascii("generater"),
                &["generator", "generated", "generates", "generate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requeriment"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributio"),
                &["distributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compliation"),
                &["compilation", "complication"],
            ),
            (dictgen::InsensitiveStr::Ascii("utlize"), &["utilize"]),
            (
                dictgen::InsensitiveStr::Ascii("compies"),
                &["compiles", "copies", "complies"],
            ),
            (dictgen::InsensitiveStr::Ascii("downagrade"), &["downgrade"]),
            (dictgen::InsensitiveStr::Ascii("objeccts"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("untrianed"), &["untrained"]),
            (dictgen::InsensitiveStr::Ascii("reamin"), &["remain"]),
            (dictgen::InsensitiveStr::Ascii("levander"), &["lavender"]),
            (dictgen::InsensitiveStr::Ascii("destroyr"), &["destroyers"]),
            (dictgen::InsensitiveStr::Ascii("misquito"), &["mosquito"]),
            (dictgen::InsensitiveStr::Ascii("chameloen"), &["chameleon"]),
            (
                dictgen::InsensitiveStr::Ascii("escavation"),
                &["escalation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alusion"),
                &["allusion", "illusion"],
            ),
            (dictgen::InsensitiveStr::Ascii("institude"), &["instituted"]),
            (
                dictgen::InsensitiveStr::Ascii("recommnded"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("criteron"), &["criterion"]),
            (dictgen::InsensitiveStr::Ascii("describ"), &["describe"]),
            (
                dictgen::InsensitiveStr::Ascii("muscil"),
                &["muscle", "mussel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactertistic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("godsped"), &["godspeed"]),
            (
                dictgen::InsensitiveStr::Ascii("focues"),
                &["focuses", "focus"],
            ),
            (dictgen::InsensitiveStr::Ascii("neested"), &["nested"]),
            (
                dictgen::InsensitiveStr::Ascii("separtately"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("recyling"), &["recycling"]),
            (dictgen::InsensitiveStr::Ascii("servering"), &["serving"]),
            (dictgen::InsensitiveStr::Ascii("threds"), &["threads"]),
            (
                dictgen::InsensitiveStr::Ascii("simplisitic"),
                &["simplistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("halluzinations"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("authroed"), &["authored"]),
            (
                dictgen::InsensitiveStr::Ascii("chrashes"),
                &["crashes", "thrashes", "trashes"],
            ),
            (dictgen::InsensitiveStr::Ascii("respectos"), &["respects"]),
            (
                dictgen::InsensitiveStr::Ascii("avertising"),
                &["advertising"],
            ),
            (dictgen::InsensitiveStr::Ascii("analzye"), &["analyze"]),
            (dictgen::InsensitiveStr::Ascii("declered"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("descritptions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("relavance"), &["relevance"]),
            (dictgen::InsensitiveStr::Ascii("messagsed"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("unconsisntency"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("insitutes"), &["institutes"]),
            (
                dictgen::InsensitiveStr::Ascii("speicfically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vesions"), &["versions"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnporting"),
                &["transporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blcokchains"),
                &["blockchains"],
            ),
            (dictgen::InsensitiveStr::Ascii("opinons"), &["opinions"]),
            (
                dictgen::InsensitiveStr::Ascii("refencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomation"), &["automation"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithems"),
                &["algorithm", "algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mear"),
                &["wear", "mere", "mare"],
            ),
            (dictgen::InsensitiveStr::Ascii("loded"), &["loaded"]),
            (dictgen::InsensitiveStr::Ascii("aqcuire"), &["acquire"]),
            (
                dictgen::InsensitiveStr::Ascii("devolopments"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sheperedly"),
                &["shepherdly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficially"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extracs"),
                &["extracts", "extras"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosqutio"), &["mosquito"]),
            (
                dictgen::InsensitiveStr::Ascii("reporisaty"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hieghtened"),
                &["heightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandized"),
                &["aggrandized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("linguisitcs"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("maintaned"), &["maintained"]),
            (
                dictgen::InsensitiveStr::Ascii("favourates"),
                &["favourites"],
            ),
            (dictgen::InsensitiveStr::Ascii("framgent"), &["fragment"]),
            (
                dictgen::InsensitiveStr::Ascii("workd"),
                &["worked", "works", "word", "world"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differencies"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exchancing"),
                &["exchanging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cymptumatic"),
                &["symptomatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proberty"),
                &["property", "properly"],
            ),
            (dictgen::InsensitiveStr::Ascii("devilers"), &["delivers"]),
            (dictgen::InsensitiveStr::Ascii("balnks"), &["blanks"]),
            (dictgen::InsensitiveStr::Ascii("cursot"), &["cursor"]),
            (dictgen::InsensitiveStr::Ascii("partialy"), &["partially"]),
            (
                dictgen::InsensitiveStr::Ascii("squirin"),
                &["squiring", "squirm"],
            ),
            (dictgen::InsensitiveStr::Ascii("kake"), &["cake", "take"]),
            (
                dictgen::InsensitiveStr::Ascii("instruccions"),
                &["instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certificats"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("modues"), &["modules"]),
            (
                dictgen::InsensitiveStr::Ascii("associdation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("apostrofe"), &["apostrophe"]),
            (
                dictgen::InsensitiveStr::Ascii("attriubuted"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("frquency"), &["frequency"]),
            (dictgen::InsensitiveStr::Ascii("olny"), &["only"]),
            (dictgen::InsensitiveStr::Ascii("agian"), &["again"]),
            (
                dictgen::InsensitiveStr::Ascii("adapatations"),
                &["adaptations"],
            ),
            (dictgen::InsensitiveStr::Ascii("onlyl"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("animatior"),
                &["animator", "animation"],
            ),
            (dictgen::InsensitiveStr::Ascii("captials"), &["capitals"]),
            (
                dictgen::InsensitiveStr::Ascii("scoreborad"),
                &["scoreboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqain"), &["again"]),
            (dictgen::InsensitiveStr::Ascii("simlarity"), &["similarity"]),
            (dictgen::InsensitiveStr::Ascii("liness"), &["lines"]),
            (
                dictgen::InsensitiveStr::Ascii("heirarchic"),
                &["hierarchic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iteraction"),
                &["interaction", "iteration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifification"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibillities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphabeticall"),
                &["alphabetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranformative"),
                &["transformative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compicated"),
                &["complicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overclocled"),
                &["overclocked"],
            ),
            (dictgen::InsensitiveStr::Ascii("imbrasing"), &["embracing"]),
            (dictgen::InsensitiveStr::Ascii("sqaure"), &["square"]),
            (dictgen::InsensitiveStr::Ascii("threschold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("doublellift"),
                &["doublelift"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reveiving"),
                &["receiving", "reviewing", "reviving"],
            ),
            (dictgen::InsensitiveStr::Ascii("simlarly"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("fauilures"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("electivre"), &["elective"]),
            (dictgen::InsensitiveStr::Ascii("traffice"), &["traffic"]),
            (
                dictgen::InsensitiveStr::Ascii("adnimistrator"),
                &["administrators", "administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnacquired"),
                &["unacquired"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoiciates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connvertion"),
                &["conversion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multitaskng"),
                &["multitasking"],
            ),
            (dictgen::InsensitiveStr::Ascii("interfce"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("complier"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("noticin"), &["noticing"]),
            (
                dictgen::InsensitiveStr::Ascii("othogonality"),
                &["orthogonality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hospitaly"),
                &["hospitality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathemathics"),
                &["mathematics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglicable"),
                &["negligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opportunisticly"),
                &["opportunistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grandisement"),
                &["aggrandisement"],
            ),
            (dictgen::InsensitiveStr::Ascii("paoition"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("gravitiation"),
                &["gravitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("opponet"), &["opponent"]),
            (dictgen::InsensitiveStr::Ascii("choser"), &["chooser"]),
            (dictgen::InsensitiveStr::Ascii("overwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("expolde"), &["explode"]),
            (
                dictgen::InsensitiveStr::Ascii("relationsship"),
                &["relationships"],
            ),
            (dictgen::InsensitiveStr::Ascii("neigther"), &["neither"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigios"),
                &["prestigious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruping"),
                &["interrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsts"), &["exists"]),
            (
                dictgen::InsensitiveStr::Ascii("wokraround"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowlegdements"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceses"), &["accesses"]),
            (
                dictgen::InsensitiveStr::Ascii("manefestation"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("terninals"), &["terminals"]),
            (
                dictgen::InsensitiveStr::Ascii("executabables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siutational"),
                &["situational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocaiation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("piont"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpicative"),
                &["multiplicative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmorged"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conribution"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unprobable"),
                &["improbable", "unprovable"],
            ),
            (dictgen::InsensitiveStr::Ascii("conain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterizes"),
                &["parameterizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("semicercle"),
                &["semicircle"],
            ),
            (dictgen::InsensitiveStr::Ascii("understnd"), &["understand"]),
            (dictgen::InsensitiveStr::Ascii("enforing"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("suggessting"),
                &["suggesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inturpratasion"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playhtrough"),
                &["playthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administations"),
                &["administrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orignially"),
                &["originally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referted"),
                &["reverted", "referred", "refereed"],
            ),
            (dictgen::InsensitiveStr::Ascii("kewword"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("wachdog"), &["watchdog"]),
            (dictgen::InsensitiveStr::Ascii("swnasea"), &["swansea"]),
            (
                dictgen::InsensitiveStr::Ascii("propperties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequisit"),
                &["prerequisite"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorent"), &["ignorant"]),
            (dictgen::InsensitiveStr::Ascii("texually"), &["textually"]),
            (
                dictgen::InsensitiveStr::Ascii("summurised"),
                &["summarised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiziing"),
                &["optimizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projetction"),
                &["projection"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("scriputres"),
                &["scriptures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assumpetion"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("isloate"), &["isolate"]),
            (dictgen::InsensitiveStr::Ascii("conecepts"), &["concepts"]),
            (
                dictgen::InsensitiveStr::Ascii("generlizes"),
                &["generalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initators"),
                &["initiators", "imitators"],
            ),
            (dictgen::InsensitiveStr::Ascii("projectio"), &["projection"]),
            (
                dictgen::InsensitiveStr::Ascii("initialilsation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interferance"),
                &["interference"],
            ),
            (dictgen::InsensitiveStr::Ascii("extsing"), &["existing"]),
            (
                dictgen::InsensitiveStr::Ascii("xenophopia"),
                &["xenophobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocateing"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertatians"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cunters"),
                &["counters", "hunters"],
            ),
            (dictgen::InsensitiveStr::Ascii("diabetees"), &["diabetes"]),
            (dictgen::InsensitiveStr::Ascii("downgarade"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("embarrased"),
                &["embarrassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("wilcard"), &["wildcard"]),
            (
                dictgen::InsensitiveStr::Ascii("lighthearthed"),
                &["lighthearted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pararagraph"),
                &["paragraph"],
            ),
            (dictgen::InsensitiveStr::Ascii("devicec"), &["device"]),
            (dictgen::InsensitiveStr::Ascii("uiu"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("prfixed"), &["prefixed"]),
            (dictgen::InsensitiveStr::Ascii("magisine"), &["magazine"]),
            (dictgen::InsensitiveStr::Ascii("severide"), &["severed"]),
            (dictgen::InsensitiveStr::Ascii("excahcnge"), &["exchange"]),
            (dictgen::InsensitiveStr::Ascii("cotnains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("reveives"),
                &["receives", "revives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporiately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("siteu"), &["site"]),
            (dictgen::InsensitiveStr::Ascii("elasped"), &["elapsed"]),
            (dictgen::InsensitiveStr::Ascii("sectopns"), &["sections"]),
            (dictgen::InsensitiveStr::Ascii("succsessor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("coorperations"),
                &["corporations", "cooperations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadvatanges"),
                &["disadvantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucuring"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exagguration"),
                &["exaggeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nvironment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispence"), &["dispense"]),
            (
                dictgen::InsensitiveStr::Ascii("overzelisly"),
                &["overzealously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boardcasts"),
                &["broadcasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pheriparials"),
                &["peripherals"],
            ),
            (dictgen::InsensitiveStr::Ascii("comitting"), &["committing"]),
            (dictgen::InsensitiveStr::Ascii("hystera"), &["hysteria"]),
            (dictgen::InsensitiveStr::Ascii("humidiy"), &["humidity"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferrence"),
                &["dereference"],
            ),
            (dictgen::InsensitiveStr::Ascii("threholds"), &["thresholds"]),
            (dictgen::InsensitiveStr::Ascii("assersion"), &["assertion"]),
            (dictgen::InsensitiveStr::Ascii("aproving"), &["approving"]),
            (
                dictgen::InsensitiveStr::Ascii("configuerd"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresssion"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sasktachewan"),
                &["saskatchewan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreciatin"),
                &["depreciating", "depreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("condesned"), &["condensed"]),
            (
                dictgen::InsensitiveStr::Ascii("periwinlke"),
                &["periwinkle"],
            ),
            (dictgen::InsensitiveStr::Ascii("warrents"), &["warrants"]),
            (
                dictgen::InsensitiveStr::Ascii("destinitation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interally"),
                &["internally", "integrally"],
            ),
            (dictgen::InsensitiveStr::Ascii("booleams"), &["booleans"]),
            (dictgen::InsensitiveStr::Ascii("querstion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("vietnamiese"),
                &["vietnamese"],
            ),
            (dictgen::InsensitiveStr::Ascii("clacium"), &["calcium"]),
            (dictgen::InsensitiveStr::Ascii("copyrught"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("teesting"), &["testing"]),
            (
                dictgen::InsensitiveStr::Ascii("cannibaljim"),
                &["cannibalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rehabilitacion"),
                &["rehabilitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("porbable"), &["probable"]),
            (dictgen::InsensitiveStr::Ascii("scenerio"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("techinques"),
                &["techniques"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribured"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shdowing"),
                &["shadowing", "showing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assumptious"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninistaller"),
                &["uninstaller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoweldgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoters"),
                &["reporters", "repeaters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conntection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("originiating"),
                &["originating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susequence"),
                &["subsequence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publiching"),
                &["publishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predominatly"),
                &["predominantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preconfiged"),
                &["preconfigured"],
            ),
            (dictgen::InsensitiveStr::Ascii("disricts"), &["districts"]),
            (
                dictgen::InsensitiveStr::Ascii("inintelligent"),
                &["unintelligent"],
            ),
            (dictgen::InsensitiveStr::Ascii("redesgin"), &["redesign"]),
            (dictgen::InsensitiveStr::Ascii("messsages"), &["messages"]),
            (dictgen::InsensitiveStr::Ascii("oberflows"), &["overflows"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialisers"),
                &["initialisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hich"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("placholders"),
                &["placeholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("faield"), &["failed"]),
            (dictgen::InsensitiveStr::Ascii("generla"), &["general"]),
            (
                dictgen::InsensitiveStr::Ascii("sirvayling"),
                &["surveiling"],
            ),
            (dictgen::InsensitiveStr::Ascii("emittter"), &["emitter"]),
            (
                dictgen::InsensitiveStr::Ascii("hypocrticial"),
                &["hypocritical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscritpitons"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noticabely"),
                &["noticeably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conceously"),
                &["consciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculuates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheched"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("detcetion"), &["detection"]),
            (dictgen::InsensitiveStr::Ascii("tencaity"), &["tenacity"]),
            (dictgen::InsensitiveStr::Ascii("wotk"), &["work"]),
            (dictgen::InsensitiveStr::Ascii("pojrects"), &["projects"]),
            (
                dictgen::InsensitiveStr::Ascii("asdignment"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmaine"), &["examine"]),
            (
                dictgen::InsensitiveStr::Ascii("georeferncing"),
                &["georeferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecctivity"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("generatie"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("harasments"),
                &["harassments"],
            ),
            (dictgen::InsensitiveStr::Ascii("installler"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("quetion"), &["question"]),
            (
                dictgen::InsensitiveStr::Ascii("devirtualied"),
                &["devirtualised", "devirtualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabilitic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enure"),
                &["ensure", "endure"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploaeding"), &["uploading"]),
            (dictgen::InsensitiveStr::Ascii("toruisty"), &["touristy"]),
            (dictgen::InsensitiveStr::Ascii("diarreha"), &["diarrhea"]),
            (dictgen::InsensitiveStr::Ascii("thie"), &["the", "this"]),
            (dictgen::InsensitiveStr::Ascii("fragement"), &["fragment"]),
            (
                dictgen::InsensitiveStr::Ascii("accerelated"),
                &["accelerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muncipalities"),
                &["municipalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("partiula"), &["particular"]),
            (
                dictgen::InsensitiveStr::Ascii("diaganostics"),
                &["diagnostics"],
            ),
            (dictgen::InsensitiveStr::Ascii("contiains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("commenwealth"),
                &["commonwealth"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressumably"),
                &["presumably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicaton"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depndantly"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("instale"), &["install"]),
            (
                dictgen::InsensitiveStr::Ascii("consciencious"),
                &["conscientious"],
            ),
            (dictgen::InsensitiveStr::Ascii("calucator"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("beleivable"),
                &["believable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propery"),
                &["property", "properly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("screenchots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("hendler"), &["handler"]),
            (
                dictgen::InsensitiveStr::Ascii("necasserily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("tescases"), &["testcases"]),
            (dictgen::InsensitiveStr::Ascii("egpytian"), &["egyptian"]),
            (
                dictgen::InsensitiveStr::Ascii("conceptally"),
                &["conceptually"],
            ),
            (dictgen::InsensitiveStr::Ascii("needeed"), &["needed"]),
            (dictgen::InsensitiveStr::Ascii("consiered"), &["considered"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuie"),
                &["contribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("hachish"), &["hackish"]),
            (
                dictgen::InsensitiveStr::Ascii("flexiblity"),
                &["flexibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("capicitor"), &["capacitor"]),
            (
                dictgen::InsensitiveStr::Ascii("supportting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("uplad"), &["upload"]),
            (dictgen::InsensitiveStr::Ascii("dropse"), &["drops"]),
            (dictgen::InsensitiveStr::Ascii("orded"), &["ordered"]),
            (
                dictgen::InsensitiveStr::Ascii("reprentative"),
                &["representative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envirionment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("impliciete"), &["implicit"]),
            (
                dictgen::InsensitiveStr::Ascii("intricicies"),
                &["intricacies"],
            ),
            (dictgen::InsensitiveStr::Ascii("vavriables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("forwareding"),
                &["forwarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containmemt"),
                &["containment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosiation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difines"),
                &["defines", "divines"],
            ),
            (dictgen::InsensitiveStr::Ascii("barvery"), &["bravery"]),
            (dictgen::InsensitiveStr::Ascii("visinble"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("inicialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceene"), &["scene"]),
            (dictgen::InsensitiveStr::Ascii("parallells"), &["parallels"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nostalgisch"),
                &["nostalgic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insturctors"),
                &["instructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasword"), &["password"]),
            (
                dictgen::InsensitiveStr::Ascii("decapsulting"),
                &["decapsulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloacte"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("termniated"),
                &["terminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("acornym"), &["acronym"]),
            (dictgen::InsensitiveStr::Ascii("structed"), &["structured"]),
            (dictgen::InsensitiveStr::Ascii("paintile"), &["painttile"]),
            (
                dictgen::InsensitiveStr::Ascii("locall"),
                &["local", "locale"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenicate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacholer"), &["bachelor"]),
            (dictgen::InsensitiveStr::Ascii("damge"), &["damage"]),
            (dictgen::InsensitiveStr::Ascii("licenceing"), &["licencing"]),
            (
                dictgen::InsensitiveStr::Ascii("consciouly"),
                &["consciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("windshied"), &["windshield"]),
            (dictgen::InsensitiveStr::Ascii("groshury"), &["grocery"]),
            (
                dictgen::InsensitiveStr::Ascii("contaminatin"),
                &["contaminating", "contamination"],
            ),
            (dictgen::InsensitiveStr::Ascii("famlies"), &["families"]),
            (
                dictgen::InsensitiveStr::Ascii("alighs"),
                &["aligns", "alights"],
            ),
            (dictgen::InsensitiveStr::Ascii("southerton"), &["southern"]),
            (
                dictgen::InsensitiveStr::Ascii("incoremental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("graphicaly"),
                &["graphically"],
            ),
            (dictgen::InsensitiveStr::Ascii("interesct"), &["intersect"]),
            (dictgen::InsensitiveStr::Ascii("packgement"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("congig"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("realsitic"), &["realistic"]),
            (
                dictgen::InsensitiveStr::Ascii("conmnection"),
                &["connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteratior"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("handboook"), &["handbook"]),
            (dictgen::InsensitiveStr::Ascii("patriotes"), &["patriots"]),
            (
                dictgen::InsensitiveStr::Ascii("intersecrion"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("malplaced"), &["misplaced"]),
            (
                dictgen::InsensitiveStr::Ascii("initiatlize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarrantie"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("speperating"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perodically"),
                &["periodically"],
            ),
            (dictgen::InsensitiveStr::Ascii("organismed"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("attractice"),
                &["attractive"],
            ),
            (dictgen::InsensitiveStr::Ascii("fifferent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("extists"), &["exists"]),
            (
                dictgen::InsensitiveStr::Ascii("escartment"),
                &["escarpment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("midfileders"),
                &["midfielders"],
            ),
            (dictgen::InsensitiveStr::Ascii("sekected"), &["selected"]),
            (dictgen::InsensitiveStr::Ascii("correclty"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("activatin"),
                &["activation", "activating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concantenated"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("consummer"), &["consumer"]),
            (dictgen::InsensitiveStr::Ascii("criteak"), &["critique"]),
            (
                dictgen::InsensitiveStr::Ascii("transfersal"),
                &["transversal"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefered"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("snippent"), &["snippet"]),
            (
                dictgen::InsensitiveStr::Ascii("subclassess"),
                &["subclasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("deepo"), &["depot"]),
            (
                dictgen::InsensitiveStr::Ascii("corellated"),
                &["correlated"],
            ),
            (dictgen::InsensitiveStr::Ascii("soomewhat"), &["somewhat"]),
            (dictgen::InsensitiveStr::Ascii("portgual"), &["portugal"]),
            (
                dictgen::InsensitiveStr::Ascii("instramental"),
                &["instrumental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disributed"),
                &["distributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("udloaded"), &["unloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("translatoin"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("warlking"), &["walking"]),
            (
                dictgen::InsensitiveStr::Ascii("compresssor"),
                &["compressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("referant"), &["referent"]),
            (
                dictgen::InsensitiveStr::Ascii("innitializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("malaysa"), &["malaysia"]),
            (dictgen::InsensitiveStr::Ascii("appliyed"), &["applied"]),
            (
                dictgen::InsensitiveStr::Ascii("interated"),
                &["iterated", "interacted", "integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("useed"), &["used"]),
            (dictgen::InsensitiveStr::Ascii("mathcers"), &["matchers"]),
            (dictgen::InsensitiveStr::Ascii("addd"), &["add", "added"]),
            (
                dictgen::InsensitiveStr::Ascii("workds"),
                &["works", "words", "worlds"],
            ),
            (dictgen::InsensitiveStr::Ascii("falcones"), &["falcons"]),
            (dictgen::InsensitiveStr::Ascii("broked"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("inconvertable"),
                &["inconvertible"],
            ),
            (dictgen::InsensitiveStr::Ascii("noticiing"), &["noticing"]),
            (dictgen::InsensitiveStr::Ascii("meesaging"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("regestries"),
                &["registries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resuces"),
                &["rescues", "reduces"],
            ),
            (dictgen::InsensitiveStr::Ascii("sculpteur"), &["sculpture"]),
            (
                dictgen::InsensitiveStr::Ascii("restouration"),
                &["restoration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formualations"),
                &["formulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepend"),
                &["depend", "suspend"],
            ),
            (dictgen::InsensitiveStr::Ascii("healither"), &["healthier"]),
            (dictgen::InsensitiveStr::Ascii("ligth"), &["light"]),
            (dictgen::InsensitiveStr::Ascii("sysamdin"), &["sysadmin"]),
            (dictgen::InsensitiveStr::Ascii("sucesor"), &["successor"]),
            (dictgen::InsensitiveStr::Ascii("definaty"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("argumentates"), &["argues"]),
            (dictgen::InsensitiveStr::Ascii("aparthied"), &["apartheid"]),
            (
                dictgen::InsensitiveStr::Ascii("expressino"),
                &["expression"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfould"), &["unfold"]),
            (dictgen::InsensitiveStr::Ascii("addresesd"), &["addressed"]),
            (
                dictgen::InsensitiveStr::Ascii("inintializing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("engineed"), &["engineered"]),
            (
                dictgen::InsensitiveStr::Ascii("tramploine"),
                &["trampoline"],
            ),
            (dictgen::InsensitiveStr::Ascii("frosaken"), &["forsaken"]),
            (
                dictgen::InsensitiveStr::Ascii("corinthans"),
                &["corinthians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrtibutes"),
                &["distributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularlized"),
                &["regularized"],
            ),
            (dictgen::InsensitiveStr::Ascii("bangquits"), &["banquets"]),
            (
                dictgen::InsensitiveStr::Ascii("intenationally"),
                &["internationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("longst"), &["longest"]),
            (
                dictgen::InsensitiveStr::Ascii("indistiguishable"),
                &["indistinguishable"],
            ),
            (dictgen::InsensitiveStr::Ascii("packacges"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("poerful"), &["powerful"]),
            (
                dictgen::InsensitiveStr::Ascii("phyiscally"),
                &["physically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proflied"),
                &["profiled", "profiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avaloability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("superceed"), &["supersede"]),
            (
                dictgen::InsensitiveStr::Ascii("evalutions"),
                &["evaluations", "evolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("staions"), &["stations"]),
            (dictgen::InsensitiveStr::Ascii("catgories"), &["categories"]),
            (dictgen::InsensitiveStr::Ascii("hispanis"), &["hispanics"]),
            (dictgen::InsensitiveStr::Ascii("sanatizer"), &["sanitizer"]),
            (
                dictgen::InsensitiveStr::Ascii("accidnetly"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("gagues"), &["gages"]),
            (dictgen::InsensitiveStr::Ascii("flushs"), &["flushes"]),
            (dictgen::InsensitiveStr::Ascii("montiored"), &["monitored"]),
            (
                dictgen::InsensitiveStr::Ascii("familierize"),
                &["familiarize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreadshets"),
                &["spreadsheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbait"),
                &["masturbate"],
            ),
            (dictgen::InsensitiveStr::Ascii("undertoe"), &["undertones"]),
            (dictgen::InsensitiveStr::Ascii("brokend"), &["broken"]),
            (dictgen::InsensitiveStr::Ascii("mabye"), &["maybe"]),
            (dictgen::InsensitiveStr::Ascii("modtified"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("librararies"),
                &["libraries"],
            ),
            (dictgen::InsensitiveStr::Ascii("nonthing"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("sceletal"), &["skeletal"]),
            (
                dictgen::InsensitiveStr::Ascii("temporali"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congigurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("feisable"), &["feasible"]),
            (dictgen::InsensitiveStr::Ascii("versioed"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("heda"), &["head"]),
            (
                dictgen::InsensitiveStr::Ascii("avaibility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neglagence"),
                &["negligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializaiton"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlightes"),
                &["highlights", "highlighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discracefulness"),
                &["disgracefulness"],
            ),
            (dictgen::InsensitiveStr::Ascii("assnage"), &["assange"]),
            (
                dictgen::InsensitiveStr::Ascii("destructivo"),
                &["destruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreaciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("laugnages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("improvemens"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unviersality"),
                &["universality"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensentive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("follwowers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("pressurin"), &["pressuring"]),
            (
                dictgen::InsensitiveStr::Ascii("interpersonnal"),
                &["interpersonal"],
            ),
            (dictgen::InsensitiveStr::Ascii("servivor"), &["survivor"]),
            (dictgen::InsensitiveStr::Ascii("decleares"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("visulisation"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpatains"), &["captains"]),
            (
                dictgen::InsensitiveStr::Ascii("unneccessary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deutcshland"),
                &["deutschland"],
            ),
            (dictgen::InsensitiveStr::Ascii("coachig"), &["coaching"]),
            (dictgen::InsensitiveStr::Ascii("comanding"), &["commanding"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluatates"),
                &["evaluates"],
            ),
            (dictgen::InsensitiveStr::Ascii("descriibe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("reteiving"), &["retrieving"]),
            (dictgen::InsensitiveStr::Ascii("destiney"), &["destiny"]),
            (dictgen::InsensitiveStr::Ascii("widthn"), &["width"]),
            (
                dictgen::InsensitiveStr::Ascii("restriciton"),
                &["restrictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("storign"), &["storing"]),
            (dictgen::InsensitiveStr::Ascii("ehtereal"), &["ethereal"]),
            (dictgen::InsensitiveStr::Ascii("vlaue"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("sheelpe"), &["sheeple"]),
            (dictgen::InsensitiveStr::Ascii("tablles"), &["tables"]),
            (
                dictgen::InsensitiveStr::Ascii("overpopulaton"),
                &["overpopulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("muscluar"), &["muscular"]),
            (dictgen::InsensitiveStr::Ascii("bisiness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("differrences"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("produceras"), &["produces"]),
            (
                dictgen::InsensitiveStr::Ascii("califronian"),
                &["californian"],
            ),
            (dictgen::InsensitiveStr::Ascii("devcent"), &["decent"]),
            (dictgen::InsensitiveStr::Ascii("specifc"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("tkae"), &["take"]),
            (dictgen::InsensitiveStr::Ascii("coherant"), &["coherent"]),
            (dictgen::InsensitiveStr::Ascii("juvenille"), &["juvenile"]),
            (dictgen::InsensitiveStr::Ascii("contanct"), &["contact"]),
            (dictgen::InsensitiveStr::Ascii("clsoes"), &["closes"]),
            (
                dictgen::InsensitiveStr::Ascii("unidimensionnal"),
                &["unidimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("chceksum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("selekt"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("idepotency"),
                &["idempotency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summerisation"),
                &["summarisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("relocaton"), &["relocation"]),
            (
                dictgen::InsensitiveStr::Ascii("separtions"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deploymenets"),
                &["deployments"],
            ),
            (dictgen::InsensitiveStr::Ascii("extensiv"), &["extensive"]),
            (dictgen::InsensitiveStr::Ascii("obsleted"), &["obsoleted"]),
            (dictgen::InsensitiveStr::Ascii("signatur"), &["signature"]),
            (dictgen::InsensitiveStr::Ascii("stil"), &["still"]),
            (
                dictgen::InsensitiveStr::Ascii("predickted"),
                &["predicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hesitatin"),
                &["hesitation", "hesitating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotypers"),
                &["stereotypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("radis"), &["radix"]),
            (
                dictgen::InsensitiveStr::Ascii("alignmenets"),
                &["alignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("theads"), &["threads"]),
            (dictgen::InsensitiveStr::Ascii("mistic"), &["mystic"]),
            (dictgen::InsensitiveStr::Ascii("fialure"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("paradime"), &["paradigm"]),
            (
                dictgen::InsensitiveStr::Ascii("homeophaty"),
                &["homeopathy"],
            ),
            (dictgen::InsensitiveStr::Ascii("divded"), &["divided"]),
            (dictgen::InsensitiveStr::Ascii("buildins"), &["buildings"]),
            (dictgen::InsensitiveStr::Ascii("funcation"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("utill"), &["until", "utils"]),
            (
                dictgen::InsensitiveStr::Ascii("reposiotry"),
                &["repository"],
            ),
            (dictgen::InsensitiveStr::Ascii("sporanos"), &["sopranos"]),
            (dictgen::InsensitiveStr::Ascii("ignormig"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("generared"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("fesitvals"), &["festivals"]),
            (
                dictgen::InsensitiveStr::Ascii("anecdotices"),
                &["anecdotes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromesome"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("assmumed"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("origniality"),
                &["originality"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixeis"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("ackumulated"),
                &["accumulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypertrohy"),
                &["hypertrophy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipliciative"),
                &["multiplicative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplicaton"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("essentiatl"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("cofrimations"),
                &["confirmations"],
            ),
            (dictgen::InsensitiveStr::Ascii("packae"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("datasheetes"),
                &["datasheets"],
            ),
            (dictgen::InsensitiveStr::Ascii("fradulent"), &["fraudulent"]),
            (
                dictgen::InsensitiveStr::Ascii("exitations"),
                &["excitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("advences"), &["advances"]),
            (
                dictgen::InsensitiveStr::Ascii("conataining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estiamation"),
                &["estimation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarative"),
                &["ameliorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoinded"), &["avoided"]),
            (
                dictgen::InsensitiveStr::Ascii("commun"),
                &["common", "commune"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humillating"),
                &["humiliating"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("kryptolite"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactifs"),
                &["interacts"],
            ),
            (dictgen::InsensitiveStr::Ascii("palesine"), &["palestine"]),
            (dictgen::InsensitiveStr::Ascii("specifers"), &["specifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("begiinnings"),
                &["beginnings"],
            ),
            (dictgen::InsensitiveStr::Ascii("sreenshot"), &["screenshot"]),
            (
                dictgen::InsensitiveStr::Ascii("imporovement"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriebuted"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("withtin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("chechpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("reprecate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelises"),
                &["parallelises"],
            ),
            (dictgen::InsensitiveStr::Ascii("managemnt"), &["management"]),
            (
                dictgen::InsensitiveStr::Ascii("sturcturing"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoke"),
                &["invoke", "income"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insensative"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victoriuous"),
                &["victorious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumibles"),
                &["consumables"],
            ),
            (dictgen::InsensitiveStr::Ascii("faktor"), &["factor"]),
            (
                dictgen::InsensitiveStr::Ascii("comporessor"),
                &["compressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("remianed"), &["remained"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprenours"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("istalls"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("fossiles"), &["fossils"]),
            (dictgen::InsensitiveStr::Ascii("addtion"), &["addition"]),
            (
                dictgen::InsensitiveStr::Ascii("provocativley"),
                &["provocative"],
            ),
            (dictgen::InsensitiveStr::Ascii("poitners"), &["pointers"]),
            (
                dictgen::InsensitiveStr::Ascii("strengsten"),
                &["strengthen"],
            ),
            (dictgen::InsensitiveStr::Ascii("feeks"), &["feels"]),
            (
                dictgen::InsensitiveStr::Ascii("maximixing"),
                &["maximizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoptionsotiator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phlematically"),
                &["phlegmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("souvinneer"), &["souvenir"]),
            (
                dictgen::InsensitiveStr::Ascii("developmnet"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ncluded"),
                &["included", "concluded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolize"),
                &["capitalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downstrean"),
                &["downstream"],
            ),
            (dictgen::InsensitiveStr::Ascii("viwes"), &["views"]),
            (dictgen::InsensitiveStr::Ascii("hostpot"), &["hotspot"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliasing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("stright"), &["straight"]),
            (dictgen::InsensitiveStr::Ascii("exchance"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("traditilnal"),
                &["traditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspomdences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscoitaion"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("proovread"), &["proofread"]),
            (dictgen::InsensitiveStr::Ascii("convovled"), &["convolved"]),
            (
                dictgen::InsensitiveStr::Ascii("necessarilyn"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oviously"),
                &["obviously", "odiously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victoriuos"),
                &["victorious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precentiles"),
                &["percentiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("plarform"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("supporteurs"),
                &["supporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("markeplace"),
                &["marketplace"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancerus"), &["cancers"]),
            (dictgen::InsensitiveStr::Ascii("highligjt"), &["highlight"]),
            (
                dictgen::InsensitiveStr::Ascii("inattractive"),
                &["unattractive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unempolyment"),
                &["unemployment"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessecery"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("backwrads"), &["backwards"]),
            (
                dictgen::InsensitiveStr::Ascii("acusing"),
                &["causing", "accusing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacromento"),
                &["sacramento"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publushers"),
                &["publishers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacend"), &["backend"]),
            (dictgen::InsensitiveStr::Ascii("enbable"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("covnerter"), &["converter"]),
            (
                dictgen::InsensitiveStr::Ascii("eveerything"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperiened"),
                &["inexperienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("blockss"), &["blocks"]),
            (dictgen::InsensitiveStr::Ascii("reaons"), &["reasons"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacturered"),
                &["manufactured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheeds"),
                &["spreadsheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremneting"),
                &["incrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("condenmed"), &["condemned"]),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responisve"),
                &["responsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("precedin"), &["preceding"]),
            (dictgen::InsensitiveStr::Ascii("checkcout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibillity"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispprove"),
                &["disprove", "disapprove"],
            ),
            (dictgen::InsensitiveStr::Ascii("priorties"), &["priorities"]),
            (dictgen::InsensitiveStr::Ascii("negitive"), &["negative"]),
            (
                dictgen::InsensitiveStr::Ascii("ridicilously"),
                &["ridiculously"],
            ),
            (dictgen::InsensitiveStr::Ascii("tou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("ignorting"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("perpedicular"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoperate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("pedophila"), &["pedophilia"]),
            (dictgen::InsensitiveStr::Ascii("chechkbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("copleted"), &["completed"]),
            (dictgen::InsensitiveStr::Ascii("funxtion"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("conversatin"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remore"),
                &["remote", "remove"],
            ),
            (dictgen::InsensitiveStr::Ascii("porcelan"), &["porcelain"]),
            (dictgen::InsensitiveStr::Ascii("remignton"), &["remington"]),
            (dictgen::InsensitiveStr::Ascii("ulter"), &["alter"]),
            (
                dictgen::InsensitiveStr::Ascii("denegerate"),
                &["degenerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atheistical"),
                &["atheistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("taget"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("massachuestts"),
                &["massachusetts"],
            ),
            (dictgen::InsensitiveStr::Ascii("sive"), &["save", "sieve"]),
            (dictgen::InsensitiveStr::Ascii("forbit"), &["forbid"]),
            (dictgen::InsensitiveStr::Ascii("anaylsers"), &["analysers"]),
            (dictgen::InsensitiveStr::Ascii("apprearing"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("whle"), &["while"]),
            (dictgen::InsensitiveStr::Ascii("protitute"), &["prostitute"]),
            (dictgen::InsensitiveStr::Ascii("dowgraded"), &["downgraded"]),
            (dictgen::InsensitiveStr::Ascii("parmesian"), &["parmesan"]),
            (
                dictgen::InsensitiveStr::Ascii("cocentrated"),
                &["concentrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("decipted"), &["depicted"]),
            (dictgen::InsensitiveStr::Ascii("replcace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("commerical"),
                &["commercial"],
            ),
            (dictgen::InsensitiveStr::Ascii("snippetts"), &["snippets"]),
            (
                dictgen::InsensitiveStr::Ascii("philadlephia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imlementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("publcise"), &["publicise"]),
            (dictgen::InsensitiveStr::Ascii("arrangting"), &["arranging"]),
            (dictgen::InsensitiveStr::Ascii("poistives"), &["positives"]),
            (dictgen::InsensitiveStr::Ascii("lisences"), &["licences"]),
            (
                dictgen::InsensitiveStr::Ascii("spectaculair"),
                &["spectacular"],
            ),
            (dictgen::InsensitiveStr::Ascii("counsole"), &["counsel"]),
            (dictgen::InsensitiveStr::Ascii("unatteded"), &["unattended"]),
            (
                dictgen::InsensitiveStr::Ascii("functionallity"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("profle"), &["profile"]),
            (
                dictgen::InsensitiveStr::Ascii("segregacion"),
                &["segregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconsciouslly"),
                &["unconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("klicks"), &["clicks"]),
            (
                dictgen::InsensitiveStr::Ascii("capitolizing"),
                &["capitalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capapbilities"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("withot"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("incure"), &["incur"]),
            (dictgen::InsensitiveStr::Ascii("easist"), &["easiest"]),
            (
                dictgen::InsensitiveStr::Ascii("difficullty"),
                &["difficulty"],
            ),
            (dictgen::InsensitiveStr::Ascii("scafolder"), &["scaffolder"]),
            (
                dictgen::InsensitiveStr::Ascii("explicited"),
                &["explicit", "explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("decorde"), &["decode"]),
            (
                dictgen::InsensitiveStr::Ascii("transprencies"),
                &["transparencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executionier"),
                &["executioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crsytallize"),
                &["crystallize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmodifable"),
                &["unmodifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("descovery"), &["discovery"]),
            (dictgen::InsensitiveStr::Ascii("insead"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("undeniabely"),
                &["undeniably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orcestrator"),
                &["orchestrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefferable"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correnspondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applcations"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("contol"), &["control"]),
            (dictgen::InsensitiveStr::Ascii("proceesses"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("massacer"), &["massacre"]),
            (dictgen::InsensitiveStr::Ascii("remoote"), &["remote"]),
            (
                dictgen::InsensitiveStr::Ascii("anphatheater"),
                &["amphitheater"],
            ),
            (dictgen::InsensitiveStr::Ascii("seezures"), &["seizures"]),
            (dictgen::InsensitiveStr::Ascii("extisted"), &["existed"]),
            (dictgen::InsensitiveStr::Ascii("causees"), &["causes"]),
            (dictgen::InsensitiveStr::Ascii("modiflers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("sabatoshes"), &["sabotages"]),
            (dictgen::InsensitiveStr::Ascii("avalialbe"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("interploation"),
                &["interpolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modiration"),
                &["moderation"],
            ),
            (dictgen::InsensitiveStr::Ascii("symoblic"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("feilding"), &["fielding"]),
            (
                dictgen::InsensitiveStr::Ascii("deverloping"),
                &["developing"],
            ),
            (dictgen::InsensitiveStr::Ascii("otput"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("horrifing"), &["horrifying"]),
            (dictgen::InsensitiveStr::Ascii("possesse"), &["possessive"]),
            (dictgen::InsensitiveStr::Ascii("sevaral"), &["several"]),
            (
                dictgen::InsensitiveStr::Ascii("benificial"),
                &["beneficial"],
            ),
            (dictgen::InsensitiveStr::Ascii("religius"), &["religious"]),
            (dictgen::InsensitiveStr::Ascii("conntected"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("inerpolation"),
                &["interpolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generalizating"),
                &["generalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reveived"),
                &["received", "reviewed", "revived"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriptus"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("consrtuctor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintanance"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alterantes"),
                &["alternates"],
            ),
            (dictgen::InsensitiveStr::Ascii("survivours"), &["survivor"]),
            (
                dictgen::InsensitiveStr::Ascii("appropreation"),
                &["appropriation"],
            ),
            (dictgen::InsensitiveStr::Ascii("barbaris"), &["barbarians"]),
            (
                dictgen::InsensitiveStr::Ascii("initailize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communcation"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transactino"),
                &["transaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doucmentation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbolutelly"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undesriable"),
                &["undesirable"],
            ),
            (dictgen::InsensitiveStr::Ascii("deamand"), &["demand"]),
            (
                dictgen::InsensitiveStr::Ascii("reolver"),
                &["resolver", "revolver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expliclity"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransacton"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovrwrt"), &["overwrite"]),
            (dictgen::InsensitiveStr::Ascii("differet"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("transacion"),
                &["transaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectioners"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("donwgraded"),
                &["downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("incoginto"), &["incognito"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidades"),
                &["invalidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esspecially"),
                &["especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recreacion"),
                &["recreation"],
            ),
            (dictgen::InsensitiveStr::Ascii("asignors"), &["assignors"]),
            (dictgen::InsensitiveStr::Ascii("pionts"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("grieifng"), &["griefing"]),
            (
                dictgen::InsensitiveStr::Ascii("egalitara"),
                &["egalitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gereration"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconstently"),
                &["inconsistently", "inconstantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissalow"), &["disallow"]),
            (
                dictgen::InsensitiveStr::Ascii("downvoteres"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disproportiate"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("thinkig"), &["thinking"]),
            (
                dictgen::InsensitiveStr::Ascii("horiztonally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aramgeddon"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concured"),
                &["concurred", "conquered"],
            ),
            (dictgen::InsensitiveStr::Ascii("foriegn"), &["foreign"]),
            (dictgen::InsensitiveStr::Ascii("perade"), &["parade"]),
            (
                dictgen::InsensitiveStr::Ascii("tecnologies"),
                &["technologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("apend"), &["append"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactorer"),
                &["manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("emproved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("fetchs"), &["fetches"]),
            (
                dictgen::InsensitiveStr::Ascii("initilialisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generetion"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("increaseing"),
                &["increasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("herucles"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("auxliliary"), &["auxiliary"]),
            (
                dictgen::InsensitiveStr::Ascii("roundtriped"),
                &["roundtripped"],
            ),
            (dictgen::InsensitiveStr::Ascii("instnat"), &["instant"]),
            (dictgen::InsensitiveStr::Ascii("anscestors"), &["ancestors"]),
            (
                dictgen::InsensitiveStr::Ascii("transacrtion"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("worflows"), &["workflows"]),
            (
                dictgen::InsensitiveStr::Ascii("incemental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("originalt"),
                &["originality"],
            ),
            (dictgen::InsensitiveStr::Ascii("versiob"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("accumalates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contiunation"),
                &["continuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scroing"),
                &["scoring", "scrolling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technologicly"),
                &["technological"],
            ),
            (dictgen::InsensitiveStr::Ascii("mucisians"), &["musicians"]),
            (dictgen::InsensitiveStr::Ascii("ussage"), &["usage"]),
            (
                dictgen::InsensitiveStr::Ascii("declrative"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("theifs"), &["thieves"]),
            (
                dictgen::InsensitiveStr::Ascii("contencion"),
                &["contention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projectlies"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactertistically"),
                &["characteristically"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedome"), &["freedom"]),
            (
                dictgen::InsensitiveStr::Ascii("deteting"),
                &["detecting", "deleting"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppresso"), &["oppression"]),
            (dictgen::InsensitiveStr::Ascii("featchd"), &["fetched"]),
            (
                dictgen::InsensitiveStr::Ascii("indonesion"),
                &["indonesian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appericating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symmectrized"),
                &["symmetrized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pheriperials"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("businessa"),
                &["businessman"],
            ),
            (dictgen::InsensitiveStr::Ascii("accusin"), &["accusing"]),
            (dictgen::InsensitiveStr::Ascii("datastrem"), &["datastream"]),
            (dictgen::InsensitiveStr::Ascii("rouinely"), &["routinely"]),
            (dictgen::InsensitiveStr::Ascii("sqeuaky"), &["squeaky"]),
            (dictgen::InsensitiveStr::Ascii("enetities"), &["entities"]),
            (
                dictgen::InsensitiveStr::Ascii("predicte"),
                &["predictive", "predicted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbituary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("plansewalker"),
                &["planeswalker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertarias"),
                &["libertarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("skippd"), &["skipped"]),
            (
                dictgen::InsensitiveStr::Ascii("qualitaitve"),
                &["qualitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumentas"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("hyperparamters"),
                &["hyperparameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("vikigns"), &["vikings"]),
            (
                dictgen::InsensitiveStr::Ascii("uncertianty"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepperate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("coordonator"),
                &["coordinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntehsize"),
                &["synthesize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalisim"),
                &["nationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("unistalls"), &["uninstalls"]),
            (dictgen::InsensitiveStr::Ascii("framents"), &["fragments"]),
            (
                dictgen::InsensitiveStr::Ascii("highligther"),
                &["highlighter"],
            ),
            (dictgen::InsensitiveStr::Ascii("motononic"), &["monotonic"]),
            (dictgen::InsensitiveStr::Ascii("apdative"), &["adaptive"]),
            (
                dictgen::InsensitiveStr::Ascii("counselers"),
                &["counselors"],
            ),
            (dictgen::InsensitiveStr::Ascii("mofifies"), &["modifies"]),
            (dictgen::InsensitiveStr::Ascii("reauires"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("colaborations"),
                &["collaborations"],
            ),
            (dictgen::InsensitiveStr::Ascii("parisitic"), &["parasitic"]),
            (dictgen::InsensitiveStr::Ascii("signfiy"), &["signify"]),
            (dictgen::InsensitiveStr::Ascii("cartriges"), &["cartridges"]),
            (dictgen::InsensitiveStr::Ascii("brwosers"), &["browsers"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleation"),
                &["acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrrespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celar"),
                &["cellar", "clear"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhoofs"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyjacks"), &["hijacks"]),
            (dictgen::InsensitiveStr::Ascii("maruader"), &["marauder"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronising"),
                &["synchronising"],
            ),
            (dictgen::InsensitiveStr::Ascii("capaciaty"), &["capacity"]),
            (dictgen::InsensitiveStr::Ascii("assumbe"), &["assume"]),
            (dictgen::InsensitiveStr::Ascii("buttton"), &["button"]),
            (dictgen::InsensitiveStr::Ascii("authoity"), &["authority"]),
            (dictgen::InsensitiveStr::Ascii("manglade"), &["mangled"]),
            (dictgen::InsensitiveStr::Ascii("mke"), &["make"]),
            (
                dictgen::InsensitiveStr::Ascii("permuate"),
                &["permutate", "permute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decomposion"),
                &["decomposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("texures"), &["textures"]),
            (
                dictgen::InsensitiveStr::Ascii("enumarators"),
                &["enumerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("exepected"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("owful"), &["awful"]),
            (dictgen::InsensitiveStr::Ascii("aincents"), &["ancients"]),
            (
                dictgen::InsensitiveStr::Ascii("exponentional"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("detecttor"), &["detector"]),
            (dictgen::InsensitiveStr::Ascii("peristing"), &["persisting"]),
            (
                dictgen::InsensitiveStr::Ascii("crocodilule"),
                &["crocodile"],
            ),
            (dictgen::InsensitiveStr::Ascii("foratting"), &["formatting"]),
            (
                dictgen::InsensitiveStr::Ascii("expeiriment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antrophology"),
                &["anthropology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporarely"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternateively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("atributd"), &["attributed"]),
            (
                dictgen::InsensitiveStr::Ascii("advertisin"),
                &["advertising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readdrssed"),
                &["readdressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sourcedrectory"),
                &["sourcedirectory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skrawberries"),
                &["strawberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("measurments"),
                &["measurements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conselation"),
                &["consolation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaital"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ineficient"),
                &["inefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differantion"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coyrighted"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconeected"),
                &["reconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("lightnign"), &["lightning"]),
            (dictgen::InsensitiveStr::Ascii("neighter"), &["neither"]),
            (dictgen::InsensitiveStr::Ascii("exerience"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("progresing"),
                &["progressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efficeincy"),
                &["efficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("interestd"), &["interested"]),
            (dictgen::InsensitiveStr::Ascii("asociate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("targts"), &["targets"]),
            (dictgen::InsensitiveStr::Ascii("peroperly"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("misoganist"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("rapresent"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("mulitiples"), &["multiples"]),
            (dictgen::InsensitiveStr::Ascii("shuttin"), &["shutting"]),
            (dictgen::InsensitiveStr::Ascii("vigliante"), &["vigilante"]),
            (dictgen::InsensitiveStr::Ascii("sheilding"), &["shielding"]),
            (
                dictgen::InsensitiveStr::Ascii("correlatie"),
                &["correlated"],
            ),
            (dictgen::InsensitiveStr::Ascii("candadate"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("procoessed"), &["processed"]),
            (dictgen::InsensitiveStr::Ascii("ritalian"), &["ritalin"]),
            (
                dictgen::InsensitiveStr::Ascii("conviluted"),
                &["convoluted"],
            ),
            (dictgen::InsensitiveStr::Ascii("micropone"), &["microphone"]),
            (
                dictgen::InsensitiveStr::Ascii("recconections"),
                &["reconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("exatca"), &["exacta"]),
            (dictgen::InsensitiveStr::Ascii("lobbysits"), &["lobbyists"]),
            (dictgen::InsensitiveStr::Ascii("orginx"), &["originx"]),
            (dictgen::InsensitiveStr::Ascii("recreting"), &["recreating"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfitifer"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("functons"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("questonnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parralelising"),
                &["parallelising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desinterested"),
                &["disinterested"],
            ),
            (dictgen::InsensitiveStr::Ascii("recipie"), &["recipe"]),
            (
                dictgen::InsensitiveStr::Ascii("complicatted"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealearated"),
                &["ameliorated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invesigate"),
                &["investigate"],
            ),
            (dictgen::InsensitiveStr::Ascii("newines"), &["newlines"]),
            (
                dictgen::InsensitiveStr::Ascii("requeset"),
                &["request", "requisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangtement"),
                &["rearrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("floting"), &["floating"]),
            (
                dictgen::InsensitiveStr::Ascii("parralellize"),
                &["parallelize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("santitizing"),
                &["sanitizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekcs"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("subproccese"),
                &["subprocess"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrictie"),
                &["restrictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suficcient"),
                &["sufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commentries"),
                &["commenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abreviates"),
                &["abbreviates"],
            ),
            (dictgen::InsensitiveStr::Ascii("esges"), &["edges"]),
            (dictgen::InsensitiveStr::Ascii("wresselers"), &["wrestlers"]),
            (dictgen::InsensitiveStr::Ascii("singlular"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("deperecation"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restaurent"),
                &["restaurant"],
            ),
            (dictgen::InsensitiveStr::Ascii("storeda"), &["stored"]),
            (
                dictgen::InsensitiveStr::Ascii("chaecklists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotpye"),
                &["stereotypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("atittude"), &["attitude"]),
            (dictgen::InsensitiveStr::Ascii("callling"), &["calling"]),
            (dictgen::InsensitiveStr::Ascii("dicotomy"), &["dichotomy"]),
            (
                dictgen::InsensitiveStr::Ascii("indentifying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependey"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("experiense"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoiurce"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("associte"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("interrdchange"),
                &["interchange"],
            ),
            (dictgen::InsensitiveStr::Ascii("atferward"), &["afterward"]),
            (dictgen::InsensitiveStr::Ascii("strored"), &["stored"]),
            (dictgen::InsensitiveStr::Ascii("surrorate"), &["surrogate"]),
            (
                dictgen::InsensitiveStr::Ascii("repondents"),
                &["respondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persciusly"),
                &["preciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("elicided"), &["elicited"]),
            (
                dictgen::InsensitiveStr::Ascii("initalisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefererred"),
                &["preferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("capaacity"), &["capacity"]),
            (
                dictgen::InsensitiveStr::Ascii("distniguish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("sirect"), &["direct"]),
            (
                dictgen::InsensitiveStr::Ascii("concatonate"),
                &["concatenate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referennced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("assemnly"), &["assembly"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagramas"), &["diagrams"]),
            (dictgen::InsensitiveStr::Ascii("duble"), &["double"]),
            (
                dictgen::InsensitiveStr::Ascii("anniversarry"),
                &["anniversary"],
            ),
            (dictgen::InsensitiveStr::Ascii("memmicking"), &["mimicking"]),
            (dictgen::InsensitiveStr::Ascii("fornatter"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("sawtes"), &["sautés"]),
            (dictgen::InsensitiveStr::Ascii("lenguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("documetn"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("subconsiously"),
                &["subconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encoutnered"),
                &["encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("reinstals"), &["reinstalls"]),
            (dictgen::InsensitiveStr::Ascii("austair"), &["austere"]),
            (dictgen::InsensitiveStr::Ascii("witn"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("contributer"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuotirals"), &["tutorials"]),
            (dictgen::InsensitiveStr::Ascii("curevd"), &["curved"]),
            (
                dictgen::InsensitiveStr::Ascii("perpretators"),
                &["perpetrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archetpyes"),
                &["archetypes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvotear"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstartion"),
                &["demonstration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preperation"),
                &["preparation"],
            ),
            (dictgen::InsensitiveStr::Ascii("follolwing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("resporatory"),
                &["respiratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("speficieis"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("compede"), &["competed"]),
            (
                dictgen::InsensitiveStr::Ascii("nationailty"),
                &["nationality"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouputarea"), &["outputarea"]),
            (dictgen::InsensitiveStr::Ascii("readahaed"), &["readahead"]),
            (
                dictgen::InsensitiveStr::Ascii("notifiations"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unselectabe"),
                &["unselectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("wher"), &["where", "were"]),
            (dictgen::InsensitiveStr::Ascii("ununsual"), &["unusual"]),
            (
                dictgen::InsensitiveStr::Ascii("backgroudns"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calucaltions"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensivitive"),
                &["sensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expecatation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regenratet"),
                &["regenerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cteating"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("misudnerstanding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("filiaments"), &["filaments"]),
            (dictgen::InsensitiveStr::Ascii("calbacks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("convinations"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("expertss"), &["experts"]),
            (
                dictgen::InsensitiveStr::Ascii("secifically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithmes"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutermeasure"),
                &["countermeasure"],
            ),
            (dictgen::InsensitiveStr::Ascii("taylored"), &["tailored"]),
            (dictgen::InsensitiveStr::Ascii("ascripts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("apartmet"),
                &["apartments", "apartment"],
            ),
            (dictgen::InsensitiveStr::Ascii("havve"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("reneweal"), &["renewal"]),
            (
                dictgen::InsensitiveStr::Ascii("availalable"),
                &["available"],
            ),
            (dictgen::InsensitiveStr::Ascii("insidios"), &["insidious"]),
            (
                dictgen::InsensitiveStr::Ascii("comsumption"),
                &["consumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rondayvoued"),
                &["rendezvoused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defeference"),
                &["dereference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartions"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onservation"),
                &["conservation", "observation"],
            ),
            (dictgen::InsensitiveStr::Ascii("acendance"), &["ascendance"]),
            (
                dictgen::InsensitiveStr::Ascii("anecodtally"),
                &["anecdotally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncomfortablely"),
                &["uncomfortably"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlysts"), &["analysts"]),
            (dictgen::InsensitiveStr::Ascii("oligatory"), &["obligatory"]),
            (
                dictgen::InsensitiveStr::Ascii("australiens"),
                &["australians"],
            ),
            (dictgen::InsensitiveStr::Ascii("mysterise"), &["mysteries"]),
            (
                dictgen::InsensitiveStr::Ascii("overcrouded"),
                &["overcrowded"],
            ),
            (dictgen::InsensitiveStr::Ascii("addapt"), &["adapt"]),
            (
                dictgen::InsensitiveStr::Ascii("demostrate"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demenstrations"),
                &["demonstrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mofdified"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("humber"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("selcetions"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("srunken"), &["shrunken"]),
            (dictgen::InsensitiveStr::Ascii("initisl"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("degenarates"),
                &["degenerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreammer"),
                &["upstreamer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normalisin"),
                &["normalising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requestors"),
                &["requesters"],
            ),
            (dictgen::InsensitiveStr::Ascii("descibes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerarble"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("persisit"), &["persist"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("traslator"), &["translator"]),
            (
                dictgen::InsensitiveStr::Ascii("canoncialization"),
                &["canonicalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incredibile"),
                &["incredible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mouthpeice"),
                &["mouthpiece"],
            ),
            (dictgen::InsensitiveStr::Ascii("trucnate"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("senquence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("commtis"), &["commits"]),
            (
                dictgen::InsensitiveStr::Ascii("persentage"),
                &["percentage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("onveience"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("ministeres"), &["ministers"]),
            (dictgen::InsensitiveStr::Ascii("follwwind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("photograpgh"),
                &["photographs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precalulated"),
                &["precalculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertubate"), &["perturb"]),
            (
                dictgen::InsensitiveStr::Ascii("irradiatin"),
                &["irradiating", "irradiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraemeter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("alcohalic"), &["alcoholic"]),
            (dictgen::InsensitiveStr::Ascii("conainer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("zombe"), &["zombie"]),
            (dictgen::InsensitiveStr::Ascii("previosu"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("entraces"), &["entrances"]),
            (
                dictgen::InsensitiveStr::Ascii("playthourgh"),
                &["playthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakcups"), &["backups"]),
            (dictgen::InsensitiveStr::Ascii("desctroyed"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("newmaticly"),
                &["pneumatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("postuminus"),
                &["posthumous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maangement"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("validaded"), &["validated"]),
            (
                dictgen::InsensitiveStr::Ascii("negitiated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missionaire"),
                &["missionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatibl"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosiative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exercizer"),
                &["exerciser", "exorcist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimener"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collisins"),
                &["collisions", "collusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainigs"), &["training"]),
            (dictgen::InsensitiveStr::Ascii("mnochrome"), &["monochrome"]),
            (dictgen::InsensitiveStr::Ascii("itterable"), &["iterable"]),
            (dictgen::InsensitiveStr::Ascii("attidute"), &["attitude"]),
            (
                dictgen::InsensitiveStr::Ascii("ressurrecting"),
                &["resurrecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("unisgned"), &["unsigned"]),
            (dictgen::InsensitiveStr::Ascii("proxys"), &["proxies"]),
            (dictgen::InsensitiveStr::Ascii("lisiting"), &["listing"]),
            (
                dictgen::InsensitiveStr::Ascii("definitiely"),
                &["definitively", "definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("downraded"), &["downgraded"]),
            (
                dictgen::InsensitiveStr::Ascii("netowrking"),
                &["networking"],
            ),
            (dictgen::InsensitiveStr::Ascii("selceted"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("wednsedays"),
                &["wednesdays"],
            ),
            (dictgen::InsensitiveStr::Ascii("vengence"), &["vengeance"]),
            (
                dictgen::InsensitiveStr::Ascii("scarched"),
                &["searched", "scorched", "scratched"],
            ),
            (dictgen::InsensitiveStr::Ascii("hedgehorg"), &["hedgehog"]),
            (dictgen::InsensitiveStr::Ascii("xyou"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscibes"),
                &["unsubscribes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendancy"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimetrical"),
                &["symmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppurted"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("quartically"),
                &["quadratically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normalizin"),
                &["normalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomance"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referecences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("gilotine"), &["guillotine"]),
            (
                dictgen::InsensitiveStr::Ascii("anualization"),
                &["annualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("queing"), &["queuing"]),
            (
                dictgen::InsensitiveStr::Ascii("spefiable"),
                &["specifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenitcated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticisims"),
                &["criticisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememember"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("preppended"), &["prepended"]),
            (
                dictgen::InsensitiveStr::Ascii("thoughs"),
                &["thoughts", "though"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acommodation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stockpilled"),
                &["stockpile"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algothitms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("critisising"),
                &["criticising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refridgerator"),
                &["refrigerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adivsoriyes"),
                &["advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("countinues"), &["continues"]),
            (dictgen::InsensitiveStr::Ascii("molestarme"), &["molester"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitializaed"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imperialstic"),
                &["imperialist"],
            ),
            (dictgen::InsensitiveStr::Ascii("siringe"), &["syringe"]),
            (dictgen::InsensitiveStr::Ascii("publushes"), &["publishes"]),
            (dictgen::InsensitiveStr::Ascii("downcale"), &["downscale"]),
            (
                dictgen::InsensitiveStr::Ascii("consisiting"),
                &["consisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("inerited"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("dispathes"), &["dispatches"]),
            (
                dictgen::InsensitiveStr::Ascii("wrokstations"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("whlch"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("attrirbution"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("diease"), &["disease"]),
            (
                dictgen::InsensitiveStr::Ascii("backened"),
                &["backend", "blackened"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfoem"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("reerence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("incrediby"), &["incredibly"]),
            (dictgen::InsensitiveStr::Ascii("terrforms"), &["terraforms"]),
            (
                dictgen::InsensitiveStr::Ascii("polotically"),
                &["politically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconstrutions"),
                &["reconstructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classfying"),
                &["classifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("attaached"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("occasioanl"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reestructured"),
                &["restructured"],
            ),
            (dictgen::InsensitiveStr::Ascii("burzwah"), &["bourgeois"]),
            (dictgen::InsensitiveStr::Ascii("abandond"), &["abandoned"]),
            (dictgen::InsensitiveStr::Ascii("ourselfs"), &["ourselves"]),
            (dictgen::InsensitiveStr::Ascii("backshlash"), &["backslash"]),
            (dictgen::InsensitiveStr::Ascii("subjest"), &["subjects"]),
            (
                dictgen::InsensitiveStr::Ascii("sizemologogically"),
                &["seismologically"],
            ),
            (dictgen::InsensitiveStr::Ascii("anormal"), &["abnormal"]),
            (dictgen::InsensitiveStr::Ascii("copyed"), &["copied"]),
            (dictgen::InsensitiveStr::Ascii("altruisim"), &["altruism"]),
            (dictgen::InsensitiveStr::Ascii("inferface"), &["interface"]),
            (
                dictgen::InsensitiveStr::Ascii("affilitation"),
                &["affiliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclaimation"),
                &["exclamation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreammed"),
                &["upstreamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geneartive"),
                &["generative"],
            ),
            (dictgen::InsensitiveStr::Ascii("assmebles"), &["assembles"]),
            (
                dictgen::InsensitiveStr::Ascii("prpose"),
                &["propose", "purpose"],
            ),
            (dictgen::InsensitiveStr::Ascii("hashs"), &["hashes"]),
            (dictgen::InsensitiveStr::Ascii("wonderig"), &["wondering"]),
            (dictgen::InsensitiveStr::Ascii("vecors"), &["vectors"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiting"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuztomizers"),
                &["customizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negativitiy"),
                &["negativity"],
            ),
            (dictgen::InsensitiveStr::Ascii("blievable"), &["believable"]),
            (dictgen::InsensitiveStr::Ascii("purose"), &["purpose"]),
            (
                dictgen::InsensitiveStr::Ascii("devicecoordiinates"),
                &["devicecoordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sufficiant"),
                &["sufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intered"),
                &["interred", "interned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppurting"),
                &["supporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egenralised"),
                &["generalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("termially"), &["terminally"]),
            (
                dictgen::InsensitiveStr::Ascii("cuestionable"),
                &["questionable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninitalizable"),
                &["uninitializable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remaned"),
                &["remained", "renamed"],
            ),
            (dictgen::InsensitiveStr::Ascii("absrobed"), &["absorbed"]),
            (
                dictgen::InsensitiveStr::Ascii("precide"),
                &["precede", "precise", "preside"],
            ),
            (dictgen::InsensitiveStr::Ascii("destryoed"), &["destroyed"]),
            (
                dictgen::InsensitiveStr::Ascii("slefishness"),
                &["selfishness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrutping"),
                &["interrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuploded"), &["reuploaded"]),
            (
                dictgen::InsensitiveStr::Ascii("opereational"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representitives"),
                &["representatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("pash"), &["hash"]),
            (
                dictgen::InsensitiveStr::Ascii("reigons"),
                &["regions", "reigns"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucesssor"), &["successor"]),
            (dictgen::InsensitiveStr::Ascii("presideny"), &["presidency"]),
            (dictgen::InsensitiveStr::Ascii("unbuntu"), &["ubuntu"]),
            (
                dictgen::InsensitiveStr::Ascii("interperting"),
                &["interpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("assualt"), &["assault"]),
            (
                dictgen::InsensitiveStr::Ascii("intermitten"),
                &["intermittent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neumonic"),
                &["mnemonic", "pneumonic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bracketting"),
                &["bracketing"],
            ),
            (dictgen::InsensitiveStr::Ascii("scripst"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("configurted"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mateials"),
                &["materials", "metals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiementing"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("carrefully"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("circunstantial"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprogator"),
                &["propagator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporpiate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preformant"),
                &["performant"],
            ),
            (dictgen::InsensitiveStr::Ascii("procliam"), &["proclaim"]),
            (dictgen::InsensitiveStr::Ascii("algorithmm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("tigthly"), &["tightly"]),
            (dictgen::InsensitiveStr::Ascii("astarisk"), &["asterisk"]),
            (dictgen::InsensitiveStr::Ascii("minising"), &["minimising"]),
            (dictgen::InsensitiveStr::Ascii("jounals"), &["journals"]),
            (
                dictgen::InsensitiveStr::Ascii("retquirement"),
                &["requirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("brouter"), &["router"]),
            (
                dictgen::InsensitiveStr::Ascii("cahdidates"),
                &["candidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versoioning"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("depndent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("emision"), &["emission"]),
            (dictgen::InsensitiveStr::Ascii("curatures"), &["curvatures"]),
            (
                dictgen::InsensitiveStr::Ascii("diposing"),
                &["disposing", "deposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprecatedf"),
                &["deprecated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentisity"),
                &["authenticity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transofrmed"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customimzation"),
                &["customization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptior"),
                &["descriptor"],
            ),
            (dictgen::InsensitiveStr::Ascii("garflied"), &["garfield"]),
            (dictgen::InsensitiveStr::Ascii("paor"), &["pair"]),
            (
                dictgen::InsensitiveStr::Ascii("resursed"),
                &["recursed", "resourced"],
            ),
            (dictgen::InsensitiveStr::Ascii("teached"), &["taught"]),
            (dictgen::InsensitiveStr::Ascii("detabase"), &["database"]),
            (dictgen::InsensitiveStr::Ascii("deafualt"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("dialaog"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("disapprouved"),
                &["disapproved"],
            ),
            (dictgen::InsensitiveStr::Ascii("dulicates"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenats"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassebler"),
                &["disassembler"],
            ),
            (dictgen::InsensitiveStr::Ascii("trewthful"), &["truthful"]),
            (dictgen::InsensitiveStr::Ascii("behinde"), &["behind"]),
            (dictgen::InsensitiveStr::Ascii("cacluater"), &["calculator"]),
            (
                dictgen::InsensitiveStr::Ascii("instrcutional"),
                &["instructional"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimim"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("functionnality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("birdamn"), &["birdman"]),
            (dictgen::InsensitiveStr::Ascii("tyepd"), &["typed"]),
            (dictgen::InsensitiveStr::Ascii("provede"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("wesbtrook"), &["westbrook"]),
            (
                dictgen::InsensitiveStr::Ascii("reay"),
                &["ready", "really", "ray"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anulls"),
                &["annulled", "annulls"],
            ),
            (dictgen::InsensitiveStr::Ascii("organiske"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("serimonial"),
                &["ceremonial"],
            ),
            (dictgen::InsensitiveStr::Ascii("succintly"), &["succinctly"]),
            (
                dictgen::InsensitiveStr::Ascii("insetion"),
                &["insertion", "insection"],
            ),
            (dictgen::InsensitiveStr::Ascii("toruney"), &["tourney"]),
            (
                dictgen::InsensitiveStr::Ascii("doubleleft"),
                &["doublelift"],
            ),
            (dictgen::InsensitiveStr::Ascii("constured"), &["construed"]),
            (
                dictgen::InsensitiveStr::Ascii("insufficeint"),
                &["insufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("implememt"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("canonivalizing"),
                &["canonicalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("accumlate"), &["accumulate"]),
            (
                dictgen::InsensitiveStr::Ascii("conatining"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dessicate"), &["desiccate"]),
            (dictgen::InsensitiveStr::Ascii("seezure"), &["seizure"]),
            (dictgen::InsensitiveStr::Ascii("permier"), &["premier"]),
            (
                dictgen::InsensitiveStr::Ascii("augument"),
                &["argument", "augment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribuion"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("alterante"), &["alternate"]),
            (
                dictgen::InsensitiveStr::Ascii("ehancement"),
                &["enhancement"],
            ),
            (dictgen::InsensitiveStr::Ascii("chagnes"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("pytnon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("bracnhes"), &["branches"]),
            (dictgen::InsensitiveStr::Ascii("annd"), &["and"]),
            (
                dictgen::InsensitiveStr::Ascii("referenences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("yeilds"), &["yields"]),
            (dictgen::InsensitiveStr::Ascii("alledgedly"), &["allegedly"]),
            (
                dictgen::InsensitiveStr::Ascii("deconstract"),
                &["deconstruct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidante"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encoses"),
                &["encloses", "encodes"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceckbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("offfence"), &["offence"]),
            (
                dictgen::InsensitiveStr::Ascii("compuations"),
                &["computations"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodive"), &["provide"]),
            (
                dictgen::InsensitiveStr::Ascii("atternuation"),
                &["attenuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sumarization"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("amatuers"), &["amateurs"]),
            (
                dictgen::InsensitiveStr::Ascii("definetion"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeturned"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("collecties"), &["collects"]),
            (dictgen::InsensitiveStr::Ascii("alogrithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("generationnal"),
                &["generational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restucture"),
                &["restructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("temrinal"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("consiousness"),
                &["consciousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpsonded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempereature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("depeneding"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("lins"),
                &["lines", "links", "lions", "loins", "limns"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rotatio"),
                &["rotation", "ratio"],
            ),
            (dictgen::InsensitiveStr::Ascii("decsribes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("conditionel"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconnectus"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("communtiy"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("croppin"), &["cropping"]),
            (dictgen::InsensitiveStr::Ascii("highligh"), &["highlight"]),
            (
                dictgen::InsensitiveStr::Ascii("multuplies"),
                &["multiplies"],
            ),
            (dictgen::InsensitiveStr::Ascii("splittr"), &["splitter"]),
            (dictgen::InsensitiveStr::Ascii("anaylses"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffiently"),
                &["inefficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderrer"), &["renderer"]),
            (dictgen::InsensitiveStr::Ascii("bakc"), &["back"]),
            (dictgen::InsensitiveStr::Ascii("dipthongs"), &["diphthongs"]),
            (dictgen::InsensitiveStr::Ascii("sparate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("contriceptives"),
                &["contraceptives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compositons"),
                &["compositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agnostiscm"),
                &["agnosticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representes"),
                &["represents", "represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowldegement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("involvment"),
                &["involvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illumimati"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuering"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("passiones"), &["passions"]),
            (
                dictgen::InsensitiveStr::Ascii("replicaitions"),
                &["replications"],
            ),
            (dictgen::InsensitiveStr::Ascii("singnals"), &["signals"]),
            (
                dictgen::InsensitiveStr::Ascii("interpratation"),
                &["interpretation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swtichbacks"),
                &["switchbacks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itentionally"),
                &["intentionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgamise"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("sovereinity"),
                &["sovereignty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrction"),
                &["construction", "constriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingorance"), &["ignorance"]),
            (dictgen::InsensitiveStr::Ascii("analsyis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("electriciy"),
                &["electricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("eastwod"), &["eastwood"]),
            (
                dictgen::InsensitiveStr::Ascii("naybhorhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("sufocates"), &["suffocates"]),
            (dictgen::InsensitiveStr::Ascii("codesc"), &["codecs"]),
            (dictgen::InsensitiveStr::Ascii("indenx"), &["index"]),
            (
                dictgen::InsensitiveStr::Ascii("registrating"),
                &["registration", "registering"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhabces"), &["enhances"]),
            (
                dictgen::InsensitiveStr::Ascii("filesytems"),
                &["filesystem", "filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("veriables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("mercineries"),
                &["mercenaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("worshippping"),
                &["worshipping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersparsed"),
                &["interspersed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peacefullly"),
                &["peacefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intstructor"),
                &["instructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("culturs"), &["cultures"]),
            (
                dictgen::InsensitiveStr::Ascii("chckerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("resovlers"), &["resolvers"]),
            (dictgen::InsensitiveStr::Ascii("outernmost"), &["outermost"]),
            (
                dictgen::InsensitiveStr::Ascii("monestary"),
                &["monastery", "monetary"],
            ),
            (dictgen::InsensitiveStr::Ascii("edxpected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("indigenuous"),
                &["indigenous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectief"),
                &["collective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supressing"),
                &["suppressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("toglling"), &["toggling"]),
            (
                dictgen::InsensitiveStr::Ascii("successsfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("whem"), &["when"]),
            (dictgen::InsensitiveStr::Ascii("exapnd"), &["expand"]),
            (dictgen::InsensitiveStr::Ascii("stakler"), &["stalker"]),
            (
                dictgen::InsensitiveStr::Ascii("crimanally"),
                &["criminally"],
            ),
            (dictgen::InsensitiveStr::Ascii("extraxt"), &["extract"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitalizes"),
                &["reinitializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firball"),
                &["fireball", "furball"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesystme"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfarmers"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivelently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parocheally"),
                &["parochially"],
            ),
            (dictgen::InsensitiveStr::Ascii("sarcasam"), &["sarcasm"]),
            (dictgen::InsensitiveStr::Ascii("refferrer"), &["referrer"]),
            (
                dictgen::InsensitiveStr::Ascii("parenthisis"),
                &["parenthesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthuseastically"),
                &["enthusiastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceptioins"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("compariing"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("realease"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("preverse"),
                &["perverse", "preserve"],
            ),
            (dictgen::InsensitiveStr::Ascii("jaimacan"), &["jamaican"]),
            (dictgen::InsensitiveStr::Ascii("insteaf"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("ralated"), &["related"]),
            (dictgen::InsensitiveStr::Ascii("natievly"), &["natively"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotor"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("zealotes"), &["zealots"]),
            (
                dictgen::InsensitiveStr::Ascii("manifacture"),
                &["manufacture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propsective"),
                &["prospective"],
            ),
            (dictgen::InsensitiveStr::Ascii("snpashot"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("dissmisal"), &["dismissal"]),
            (
                dictgen::InsensitiveStr::Ascii("develoments"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilcates"),
                &["complicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("sybolic"), &["symbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("tehnologies"),
                &["technologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representstion"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrigirator"),
                &["refrigerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("refference"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("charicter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("supremacits"),
                &["supremacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragph"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("implicited"),
                &["implicitly", "implicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighrors"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("obervations"),
                &["observations"],
            ),
            (dictgen::InsensitiveStr::Ascii("assigh"), &["assign"]),
            (
                dictgen::InsensitiveStr::Ascii("reproducably"),
                &["reproducibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distiungished"),
                &["distinguished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accourdions"),
                &["accordions"],
            ),
            (dictgen::InsensitiveStr::Ascii("buil"), &["build", "built"]),
            (dictgen::InsensitiveStr::Ascii("hositlity"), &["hostility"]),
            (
                dictgen::InsensitiveStr::Ascii("expliciltly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("chispet"), &["chipset"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgeraded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("miroring"), &["mirroring"]),
            (dictgen::InsensitiveStr::Ascii("enalbed"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("goemetries"),
                &["geometries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrievating"),
                &["abbreviating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpency"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrspondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("misstakes"), &["mistakes"]),
            (dictgen::InsensitiveStr::Ascii("regulats"), &["regulators"]),
            (
                dictgen::InsensitiveStr::Ascii("calanders"),
                &["calendars", "colanders"],
            ),
            (dictgen::InsensitiveStr::Ascii("massasge"), &["massages"]),
            (
                dictgen::InsensitiveStr::Ascii("comutativity"),
                &["commutativity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exessively"),
                &["excessively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simptomatic"),
                &["symptomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("comtains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("hopelessely"),
                &["hopelessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("resetore"), &["restore"]),
            (dictgen::InsensitiveStr::Ascii("oriente"), &["oriented"]),
            (dictgen::InsensitiveStr::Ascii("summay"), &["summary"]),
            (
                dictgen::InsensitiveStr::Ascii("magnificient"),
                &["magnificent"],
            ),
            (dictgen::InsensitiveStr::Ascii("alows"), &["allows"]),
            (dictgen::InsensitiveStr::Ascii("lifetyle"), &["lifestyle"]),
            (
                dictgen::InsensitiveStr::Ascii("inititalisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("productino"),
                &["productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploted"), &["plotted"]),
            (dictgen::InsensitiveStr::Ascii("dependign"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("siezures"), &["seizures"]),
            (dictgen::InsensitiveStr::Ascii("capbable"), &["capable"]),
            (
                dictgen::InsensitiveStr::Ascii("abreviating"),
                &["abbreviating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("smarpthone"),
                &["smartphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("metri"), &["metric"]),
            (dictgen::InsensitiveStr::Ascii("scannign"), &["scanning"]),
            (
                dictgen::InsensitiveStr::Ascii("confiugured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("dezibel"), &["decibel"]),
            (dictgen::InsensitiveStr::Ascii("placemens"), &["placements"]),
            (dictgen::InsensitiveStr::Ascii("graphcis"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("extremitys"),
                &["extremities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesseful"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepulchure"), &["sepulchre"]),
            (
                dictgen::InsensitiveStr::Ascii("coomanders"),
                &["commanders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmasudically"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocmputed"), &["computed"]),
            (dictgen::InsensitiveStr::Ascii("rewriet"), &["rewrite"]),
            (
                dictgen::InsensitiveStr::Ascii("tranmitting"),
                &["transmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("populairty"),
                &["popularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("multitute"), &["multitude"]),
            (dictgen::InsensitiveStr::Ascii("extenson"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("termniation"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbsytems"),
                &["subsystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("unasigned"), &["unassigned"]),
            (dictgen::InsensitiveStr::Ascii("excipt"), &["except"]),
            (dictgen::InsensitiveStr::Ascii("mentionnes"), &["mentions"]),
            (
                dictgen::InsensitiveStr::Ascii("descovering"),
                &["discovering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overcharing"),
                &["overarching"],
            ),
            (dictgen::InsensitiveStr::Ascii("smoothign"), &["smoothing"]),
            (dictgen::InsensitiveStr::Ascii("thst"), &["that"]),
            (
                dictgen::InsensitiveStr::Ascii("perferrmences"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silience"),
                &["silence", "salience"],
            ),
            (dictgen::InsensitiveStr::Ascii("robocorp"), &["robocop"]),
            (
                dictgen::InsensitiveStr::Ascii("sylesheets"),
                &["stylesheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confimration"),
                &["confirmation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overshaddow"),
                &["overshadowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacifying"),
                &["specifying", "pacifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("ccurred"), &["occurred"]),
            (
                dictgen::InsensitiveStr::Ascii("custonizing"),
                &["customizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("polgyamy"), &["polygamy"]),
            (dictgen::InsensitiveStr::Ascii("barabrian"), &["barbarian"]),
            (
                dictgen::InsensitiveStr::Ascii("abstrakting"),
                &["abstracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permsistent"),
                &["persistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("kindgom"), &["kingdom"]),
            (dictgen::InsensitiveStr::Ascii("magor"), &["major"]),
            (
                dictgen::InsensitiveStr::Ascii("appoximate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("auotmate"), &["automate"]),
            (
                dictgen::InsensitiveStr::Ascii("evidentally"),
                &["evidently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstactions"),
                &["abstractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommdation"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neuroscienze"),
                &["neuroscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fluctuative"),
                &["fluctuate"],
            ),
            (dictgen::InsensitiveStr::Ascii("asycn"), &["async"]),
            (dictgen::InsensitiveStr::Ascii("destoying"), &["destroying"]),
            (dictgen::InsensitiveStr::Ascii("negoates"), &["negotiates"]),
            (dictgen::InsensitiveStr::Ascii("gereating"), &["generating"]),
            (
                dictgen::InsensitiveStr::Ascii("sacramenno"),
                &["sacramento"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementaions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("commenciez"), &["commence"]),
            (
                dictgen::InsensitiveStr::Ascii("unedcuated"),
                &["uneducated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qestionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comrpesses"),
                &["compresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amperstands"),
                &["ampersands"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaiances"), &["variances"]),
            (dictgen::InsensitiveStr::Ascii("menue"), &["menu"]),
            (dictgen::InsensitiveStr::Ascii("seconcly"), &["secondly"]),
            (
                dictgen::InsensitiveStr::Ascii("respection"),
                &["respecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enormassly"),
                &["enormously"],
            ),
            (dictgen::InsensitiveStr::Ascii("notod"), &["noted"]),
            (dictgen::InsensitiveStr::Ascii("opponenet"), &["opponent"]),
            (dictgen::InsensitiveStr::Ascii("ubtton"), &["button"]),
            (
                dictgen::InsensitiveStr::Ascii("cherckpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("difuses"),
                &["diffuses", "defused", "defuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibilies"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("damons"),
                &["daemons", "demons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uniniteresting"),
                &["uninteresting"],
            ),
            (dictgen::InsensitiveStr::Ascii("fillay"), &["fillet"]),
            (dictgen::InsensitiveStr::Ascii("folx"), &["folks"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendantly"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resintalled"),
                &["reinstalled"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfform"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("travellin"), &["travelling"]),
            (dictgen::InsensitiveStr::Ascii("organie"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("elektrolytes"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniitially"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("solataire"), &["solitaire"]),
            (
                dictgen::InsensitiveStr::Ascii("profesions"),
                &["profession", "professions", "profusions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("propperly"), &["properly"]),
            (dictgen::InsensitiveStr::Ascii("neighbeor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("pariticipate"),
                &["participate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expanshions"),
                &["expansions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pramater"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("mentined"), &["mentioned"]),
            (dictgen::InsensitiveStr::Ascii("hapy"), &["happy"]),
            (dictgen::InsensitiveStr::Ascii("resonse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("specfiied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("checler"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("transito"), &["transition"]),
            (
                dictgen::InsensitiveStr::Ascii("exlanations"),
                &["explanations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insensitiv"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessellater"),
                &["tessellator", "tessellated", "tessellates", "tessellate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpecteed"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sould"),
                &["could", "should", "sold"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabkes"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("ovverridden"),
                &["overridden"],
            ),
            (dictgen::InsensitiveStr::Ascii("constasnt"), &["constants"]),
            (
                dictgen::InsensitiveStr::Ascii("indvidually"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("momory"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("medioker"), &["mediocre"]),
            (
                dictgen::InsensitiveStr::Ascii("interupted"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("annonymous"), &["anonymous"]),
            (
                dictgen::InsensitiveStr::Ascii("ambadexterous"),
                &["ambidextrous"],
            ),
            (dictgen::InsensitiveStr::Ascii("incude"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("avaialbility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrpted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("incrmentally"),
                &["incrementally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identiable"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsequnt"), &["subsequent"]),
            (dictgen::InsensitiveStr::Ascii("absuridty"), &["absurdity"]),
            (dictgen::InsensitiveStr::Ascii("resarting"), &["restarting"]),
            (
                dictgen::InsensitiveStr::Ascii("finanically"),
                &["financially"],
            ),
            (dictgen::InsensitiveStr::Ascii("fouund"), &["found"]),
            (dictgen::InsensitiveStr::Ascii("linrary"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("certifiations"),
                &["certifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("oragnes"), &["oranges"]),
            (dictgen::InsensitiveStr::Ascii("suop"), &["soup"]),
            (
                dictgen::InsensitiveStr::Ascii("parralelization"),
                &["parallelization"],
            ),
            (dictgen::InsensitiveStr::Ascii("thngs"), &["things"]),
            (
                dictgen::InsensitiveStr::Ascii("initilaizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjectivily"),
                &["subjectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("avgerage"), &["average"]),
            (dictgen::InsensitiveStr::Ascii("opimizes"), &["optimizes"]),
            (dictgen::InsensitiveStr::Ascii("qustioned"), &["questioned"]),
            (
                dictgen::InsensitiveStr::Ascii("insiginifcant"),
                &["insignificant"],
            ),
            (dictgen::InsensitiveStr::Ascii("faulsures"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("undersatndable"),
                &["understandable"],
            ),
            (dictgen::InsensitiveStr::Ascii("peferred"), &["preferred"]),
            (
                dictgen::InsensitiveStr::Ascii("downvotees"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("chiuldren"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("downrgade"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("cipertexts"),
                &["ciphertexts"],
            ),
            (dictgen::InsensitiveStr::Ascii("necause"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("verfifications"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("futuristc"), &["futuristic"]),
            (dictgen::InsensitiveStr::Ascii("bloker"), &["blocker"]),
            (
                dictgen::InsensitiveStr::Ascii("underpowed"),
                &["underpowered"],
            ),
            (dictgen::InsensitiveStr::Ascii("worchester"), &["worcester"]),
            (
                dictgen::InsensitiveStr::Ascii("senstivities"),
                &["sensitivities"],
            ),
            (dictgen::InsensitiveStr::Ascii("highter"), &["higher"]),
            (
                dictgen::InsensitiveStr::Ascii("dobulelift"),
                &["doublelift"],
            ),
            (dictgen::InsensitiveStr::Ascii("extrat"), &["extract"]),
            (
                dictgen::InsensitiveStr::Ascii("satifactory"),
                &["satisfactory"],
            ),
            (dictgen::InsensitiveStr::Ascii("gymnist"), &["gymnast"]),
            (
                dictgen::InsensitiveStr::Ascii("commandmends"),
                &["commandments"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometihing"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailizers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derect"),
                &["direct", "detect", "defect", "erect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instersectioned"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("rapair"), &["repair"]),
            (
                dictgen::InsensitiveStr::Ascii("advertistments"),
                &["advertisements"],
            ),
            (dictgen::InsensitiveStr::Ascii("queations"), &["equations"]),
            (
                dictgen::InsensitiveStr::Ascii("stroytelling"),
                &["storytelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("baclony"), &["balcony"]),
            (dictgen::InsensitiveStr::Ascii("geeration"), &["generation"]),
            (
                dictgen::InsensitiveStr::Ascii("exaustively"),
                &["exhaustively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowiedges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoperating"),
                &["incorporating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correlelate"),
                &["correlate"],
            ),
            (dictgen::InsensitiveStr::Ascii("visably"), &["visibly"]),
            (dictgen::InsensitiveStr::Ascii("cihpher"), &["cipher"]),
            (
                dictgen::InsensitiveStr::Ascii("torchlgiht"),
                &["torchlight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomalisation"),
                &["normalisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allegiancies"),
                &["allegiance", "allegiances"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxmimum"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("fluctiations"),
                &["fluctuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("garphically"),
                &["graphically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ellipticle"),
                &["elliptical"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounary"), &["boundary"]),
            (
                dictgen::InsensitiveStr::Ascii("corperations"),
                &["corporations"],
            ),
            (dictgen::InsensitiveStr::Ascii("improvie"), &["improvised"]),
            (
                dictgen::InsensitiveStr::Ascii("shakeapeare"),
                &["shakespeare"],
            ),
            (dictgen::InsensitiveStr::Ascii("otherwize"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("jefferty"), &["jeffery"]),
            (
                dictgen::InsensitiveStr::Ascii("linearily"),
                &["linearly", "linearity"],
            ),
            (dictgen::InsensitiveStr::Ascii("drifitng"), &["drifting"]),
            (dictgen::InsensitiveStr::Ascii("inrinsics"), &["intrinsics"]),
            (dictgen::InsensitiveStr::Ascii("avaible"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("critiically"),
                &["critically"],
            ),
            (dictgen::InsensitiveStr::Ascii("detatch"), &["detach"]),
            (dictgen::InsensitiveStr::Ascii("emphsis"), &["emphasis"]),
            (dictgen::InsensitiveStr::Ascii("behavoueir"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusing"),
                &["exclusion", "excluding", "excusing"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuaized"), &["visualized"]),
            (dictgen::InsensitiveStr::Ascii("mandess"), &["madness"]),
            (dictgen::InsensitiveStr::Ascii("personhod"), &["personhood"]),
            (dictgen::InsensitiveStr::Ascii("returs"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("fumction"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("govrement"), &["government"]),
            (
                dictgen::InsensitiveStr::Ascii("libertarianisim"),
                &["libertarianism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downaloaded"),
                &["downloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniect"), &["inject"]),
            (
                dictgen::InsensitiveStr::Ascii("heavyweigth"),
                &["heavyweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beuraucratically"),
                &["bureaucratically"],
            ),
            (dictgen::InsensitiveStr::Ascii("conetors"), &["connectors"]),
            (dictgen::InsensitiveStr::Ascii("cathc"), &["catch"]),
            (
                dictgen::InsensitiveStr::Ascii("specficying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("enquirin"), &["enquiring"]),
            (
                dictgen::InsensitiveStr::Ascii("incorprate"),
                &["incorporate"],
            ),
            (dictgen::InsensitiveStr::Ascii("specrtal"), &["spectral"]),
            (
                dictgen::InsensitiveStr::Ascii("disapearing"),
                &["disappearing"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteraions"), &["iterations"]),
            (dictgen::InsensitiveStr::Ascii("pioneeer"), &["pioneer"]),
            (dictgen::InsensitiveStr::Ascii("protrayal"), &["portrayal"]),
            (dictgen::InsensitiveStr::Ascii("buinseses"), &["businesses"]),
            (
                dictgen::InsensitiveStr::Ascii("partecipate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoised"), &["customised"]),
            (dictgen::InsensitiveStr::Ascii("lewow"), &["luau"]),
            (
                dictgen::InsensitiveStr::Ascii("exntensions"),
                &["extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("complates"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("incremintally"),
                &["incrementally"],
            ),
            (dictgen::InsensitiveStr::Ascii("apsaragus"), &["asparagus"]),
            (
                dictgen::InsensitiveStr::Ascii("serializeable"),
                &["serializable"],
            ),
            (dictgen::InsensitiveStr::Ascii("devoloper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("usees"), &["uses"]),
            (
                dictgen::InsensitiveStr::Ascii("nonsencical"),
                &["nonsensical"],
            ),
            (dictgen::InsensitiveStr::Ascii("semster"), &["semester"]),
            (dictgen::InsensitiveStr::Ascii("praisse"), &["praises"]),
            (dictgen::InsensitiveStr::Ascii("algebriac"), &["algebraic"]),
            (dictgen::InsensitiveStr::Ascii("snadler"), &["sandler"]),
            (dictgen::InsensitiveStr::Ascii("expressio"), &["expression"]),
            (
                dictgen::InsensitiveStr::Ascii("grandised"),
                &["aggrandised"],
            ),
            (dictgen::InsensitiveStr::Ascii("boslter"), &["bolster"]),
            (dictgen::InsensitiveStr::Ascii("silouwet"), &["silhouette"]),
            (
                dictgen::InsensitiveStr::Ascii("cunter"),
                &["counter", "hunter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirence"),
                &["experience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essentiall"),
                &["essential", "essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("exprect"), &["expect"]),
            (dictgen::InsensitiveStr::Ascii("strucure"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("skeletos"), &["skeletons"]),
            (dictgen::InsensitiveStr::Ascii("tupless"), &["tuples"]),
            (
                dictgen::InsensitiveStr::Ascii("allwoances"),
                &["allowances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rebuil"),
                &["rebuild", "rebuilt"],
            ),
            (dictgen::InsensitiveStr::Ascii("mannualy"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("unreasonabley"),
                &["unreasonably"],
            ),
            (dictgen::InsensitiveStr::Ascii("doubthed"), &["doubted"]),
            (
                dictgen::InsensitiveStr::Ascii("donesticated"),
                &["domesticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proceesing"),
                &["proceeding", "processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("impossibe"), &["impossible"]),
            (dictgen::InsensitiveStr::Ascii("nothink"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("furser"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("parralellizing"),
                &["parallelizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("muilticast"), &["multicast"]),
            (
                dictgen::InsensitiveStr::Ascii("misintrepreted"),
                &["misinterpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("ukrainain"), &["ukrainian"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitiute"),
                &["prostitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("raspoberry"), &["raspberry"]),
            (
                dictgen::InsensitiveStr::Ascii("atructured"),
                &["structured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backbrounds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("anomolee"), &["anomaly"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastrucure"),
                &["infrastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("deficieny"), &["deficiency"]),
            (dictgen::InsensitiveStr::Ascii("loders"), &["loaders"]),
            (dictgen::InsensitiveStr::Ascii("emittin"), &["emitting"]),
            (dictgen::InsensitiveStr::Ascii("separor"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("carinomy"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("enterily"), &["entirely"]),
            (
                dictgen::InsensitiveStr::Ascii("domin"),
                &["doming", "domain", "domino"],
            ),
            (dictgen::InsensitiveStr::Ascii("critial"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("mutabiltiy"),
                &["mutability"],
            ),
            (dictgen::InsensitiveStr::Ascii("obtrains"), &["obtains"]),
            (
                dictgen::InsensitiveStr::Ascii("statisitically"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autherizes"),
                &["authorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("religous"), &["religious"]),
            (
                dictgen::InsensitiveStr::Ascii("conqeuring"),
                &["conquering"],
            ),
            (dictgen::InsensitiveStr::Ascii("coumter"), &["counter"]),
            (dictgen::InsensitiveStr::Ascii("expepting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("trailling"),
                &["trailing", "trialling", "trilling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("littlefigner"),
                &["littlefinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecclectic"), &["eclectic"]),
            (dictgen::InsensitiveStr::Ascii("instrut"), &["instruct"]),
            (dictgen::InsensitiveStr::Ascii("varieable"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("recominant"),
                &["recombinant"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmenet"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("nestalgia"), &["nostalgia"]),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonising"),
                &["synchronising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberviations"),
                &["abbreviations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplying"),
                &["multiplying"],
            ),
            (dictgen::InsensitiveStr::Ascii("infalting"), &["inflating"]),
            (
                dictgen::InsensitiveStr::Ascii("suppresors"),
                &["suppressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignortg"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("recoid"), &["recoil"]),
            (dictgen::InsensitiveStr::Ascii("beginer"), &["beginner"]),
            (dictgen::InsensitiveStr::Ascii("infinte"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("proliferatin"),
                &["proliferating", "proliferation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incarserated"),
                &["incarcerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("producibles"),
                &["producible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compillations"),
                &["compilations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complictaed"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatnated"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minimalstic"),
                &["minimalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("fugured"), &["figured"]),
            (
                dictgen::InsensitiveStr::Ascii("permaturely"),
                &["prematurely"],
            ),
            (dictgen::InsensitiveStr::Ascii("envorce"), &["enforce"]),
            (dictgen::InsensitiveStr::Ascii("execututes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("aggretator"),
                &["aggregator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satelittes"),
                &["satellites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("waterproff"),
                &["waterproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terrurists"),
                &["terrorists"],
            ),
            (dictgen::InsensitiveStr::Ascii("editting"), &["editing"]),
            (dictgen::InsensitiveStr::Ascii("isnpiron"), &["inspiron"]),
            (dictgen::InsensitiveStr::Ascii("lincencing"), &["licencing"]),
            (dictgen::InsensitiveStr::Ascii("breshes"), &["brushes"]),
            (
                dictgen::InsensitiveStr::Ascii("compresions"),
                &["compressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experses"),
                &["expenses", "expresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremmented"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("dogamtic"), &["dogmatic"]),
            (dictgen::InsensitiveStr::Ascii("rickoshay"), &["ricochet"]),
            (dictgen::InsensitiveStr::Ascii("kenrel"), &["kernel"]),
            (
                dictgen::InsensitiveStr::Ascii("interelated"),
                &["interrelated"],
            ),
            (dictgen::InsensitiveStr::Ascii("coloringh"), &["coloring"]),
            (
                dictgen::InsensitiveStr::Ascii("pennisulas"),
                &["peninsulas"],
            ),
            (dictgen::InsensitiveStr::Ascii("backedn"), &["backend"]),
            (
                dictgen::InsensitiveStr::Ascii("xenohpobic"),
                &["xenophobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("strictiest"), &["strictest"]),
            (dictgen::InsensitiveStr::Ascii("sastisfied"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("vishusly"), &["viciously"]),
            (
                dictgen::InsensitiveStr::Ascii("intialser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("qualifieds"), &["qualifies"]),
            (dictgen::InsensitiveStr::Ascii("byted"), &["byte", "bytes"]),
            (
                dictgen::InsensitiveStr::Ascii("condescendion"),
                &["condescension"],
            ),
            (dictgen::InsensitiveStr::Ascii("compaints"), &["complaints"]),
            (
                dictgen::InsensitiveStr::Ascii("downgarding"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intelegence"),
                &["intelligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theirselves"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("referrs"), &["refers"]),
            (dictgen::InsensitiveStr::Ascii("triagle"), &["triangle"]),
            (
                dictgen::InsensitiveStr::Ascii("technlogies"),
                &["technologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialag"), &["dialog"]),
            (dictgen::InsensitiveStr::Ascii("obtaiens"), &["obtains"]),
            (dictgen::InsensitiveStr::Ascii("trnasfer"), &["transfer"]),
            (
                dictgen::InsensitiveStr::Ascii("addresss"),
                &["address", "addresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("croozed"), &["cruised"]),
            (dictgen::InsensitiveStr::Ascii("escalatie"), &["escalate"]),
            (
                dictgen::InsensitiveStr::Ascii("tempatized"),
                &["templatized"],
            ),
            (dictgen::InsensitiveStr::Ascii("ideitifed"), &["identified"]),
            (dictgen::InsensitiveStr::Ascii("fomed"), &["formed"]),
            (dictgen::InsensitiveStr::Ascii("sequenes"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("homineim"), &["hominem"]),
            (
                dictgen::InsensitiveStr::Ascii("statisician"),
                &["statistician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiailise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassmbled"),
                &["disassembled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investige"),
                &["investigate"],
            ),
            (dictgen::InsensitiveStr::Ascii("otionally"), &["optionally"]),
            (dictgen::InsensitiveStr::Ascii("packkaged"), &["packaged"]),
            (
                dictgen::InsensitiveStr::Ascii("acitivating"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("symptum"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("experessed"), &["expressed"]),
            (dictgen::InsensitiveStr::Ascii("cahching"), &["caching"]),
            (dictgen::InsensitiveStr::Ascii("servoce"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("counsoling"),
                &["counseling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("karbohydrates"),
                &["carbohydrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("exucutors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("braodway"), &["broadway"]),
            (
                dictgen::InsensitiveStr::Ascii("correespond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("polynomal"), &["polynomial"]),
            (dictgen::InsensitiveStr::Ascii("diagional"), &["diagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("constacting"),
                &["contacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propriatary"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("deamonize"), &["daemonize"]),
            (dictgen::InsensitiveStr::Ascii("absloutes"), &["absolutes"]),
            (
                dictgen::InsensitiveStr::Ascii("representatieve"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("dordered"), &["ordered"]),
            (dictgen::InsensitiveStr::Ascii("decicated"), &["dedicated"]),
            (
                dictgen::InsensitiveStr::Ascii("enouncters"),
                &["encounters"],
            ),
            (dictgen::InsensitiveStr::Ascii("exampt"), &["exempt"]),
            (dictgen::InsensitiveStr::Ascii("sigular"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("orpahns"), &["orphans"]),
            (
                dictgen::InsensitiveStr::Ascii("linguistisch"),
                &["linguistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gracefullly"),
                &["gracefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persumption"),
                &["presumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremened"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("bck"), &["back"]),
            (
                dictgen::InsensitiveStr::Ascii("structores"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("arborted"), &["aborted"]),
            (dictgen::InsensitiveStr::Ascii("coyright"), &["copyright"]),
            (
                dictgen::InsensitiveStr::Ascii("diectories"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("radioactief"),
                &["radioactive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configruing"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilaized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("aninate"), &["animate"]),
            (dictgen::InsensitiveStr::Ascii("possilbe"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("acssumed"), &["assumed"]),
            (dictgen::InsensitiveStr::Ascii("cliens"), &["clients"]),
            (dictgen::InsensitiveStr::Ascii("likelood"), &["likelihood"]),
            (dictgen::InsensitiveStr::Ascii("privaye"), &["private"]),
            (dictgen::InsensitiveStr::Ascii("fariar"), &["farrier"]),
            (dictgen::InsensitiveStr::Ascii("somwone"), &["someone"]),
            (dictgen::InsensitiveStr::Ascii("pointeres"), &["pointers"]),
            (
                dictgen::InsensitiveStr::Ascii("calculuator"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occrrances"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeariton"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("sourc"), &["source"]),
            (
                dictgen::InsensitiveStr::Ascii("scrooled"),
                &["scrolled", "schooled"],
            ),
            (dictgen::InsensitiveStr::Ascii("ectsasy"), &["ecstasy"]),
            (
                dictgen::InsensitiveStr::Ascii("momentaraly"),
                &["momentarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacrin"), &["saccharin"]),
            (dictgen::InsensitiveStr::Ascii("evalulate"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("equippin"), &["equipping"]),
            (
                dictgen::InsensitiveStr::Ascii("contstructed"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("organes"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthroughts"),
                &["breakthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horizantal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accpetable"),
                &["acceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abandondment"),
                &["abandonment"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconets"), &["reconnects"]),
            (dictgen::InsensitiveStr::Ascii("sufferered"), &["suffered"]),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalize"),
                &["orthogonalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("swithching"), &["switching"]),
            (
                dictgen::InsensitiveStr::Ascii("concidered"),
                &["considered"],
            ),
            (dictgen::InsensitiveStr::Ascii("triathalon"), &["triathlon"]),
            (dictgen::InsensitiveStr::Ascii("raplidly"), &["rapidly"]),
            (
                dictgen::InsensitiveStr::Ascii("assotiates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscarridge"),
                &["miscarriage"],
            ),
            (dictgen::InsensitiveStr::Ascii("firsly"), &["firstly"]),
            (
                dictgen::InsensitiveStr::Ascii("triskadekaphobia"),
                &["triskaidekaphobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("moure"), &["more"]),
            (dictgen::InsensitiveStr::Ascii("psitoin"), &["position"]),
            (dictgen::InsensitiveStr::Ascii("aswering"), &["answering"]),
            (
                dictgen::InsensitiveStr::Ascii("respiratiory"),
                &["respiratory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experesions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("suddnly"), &["suddenly"]),
            (dictgen::InsensitiveStr::Ascii("odered"), &["ordered"]),
            (dictgen::InsensitiveStr::Ascii("checbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourbood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("developmently"),
                &["developmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximium"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("crytstal"), &["crystal"]),
            (dictgen::InsensitiveStr::Ascii("chatacter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("phyisology"),
                &["physiology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaginating"),
                &["imagination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepocessor"),
                &["preprocessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("veiwings"), &["viewings"]),
            (dictgen::InsensitiveStr::Ascii("israeliens"), &["israelis"]),
            (dictgen::InsensitiveStr::Ascii("menstraul"), &["menstrual"]),
            (
                dictgen::InsensitiveStr::Ascii("chechkpoint"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchrnoized"),
                &["synchronized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenet"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependenta"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("continour"), &["continuous"]),
            (dictgen::InsensitiveStr::Ascii("anicent"), &["ancient"]),
            (dictgen::InsensitiveStr::Ascii("upercase"), &["uppercase"]),
            (
                dictgen::InsensitiveStr::Ascii("embarasssing"),
                &["embarrassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depentents"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("spririous"), &["spurious"]),
            (dictgen::InsensitiveStr::Ascii("inverion"), &["inversion"]),
            (dictgen::InsensitiveStr::Ascii("initlal"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("fomaing"), &["foaming"]),
            (
                dictgen::InsensitiveStr::Ascii("electrican"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnetiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specfiication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("ourselfes"), &["ourselves"]),
            (
                dictgen::InsensitiveStr::Ascii("garuanteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("victimas"), &["victims"]),
            (
                dictgen::InsensitiveStr::Ascii("equaivalently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompitable"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostituto"),
                &["prostitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleraptor"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("munute"), &["minute"]),
            (dictgen::InsensitiveStr::Ascii("quanitity"), &["quantity"]),
            (dictgen::InsensitiveStr::Ascii("comparios"), &["comparison"]),
            (dictgen::InsensitiveStr::Ascii("ulitmate"), &["ultimate"]),
            (
                dictgen::InsensitiveStr::Ascii("desqualified"),
                &["disqualified"],
            ),
            (dictgen::InsensitiveStr::Ascii("occasins"), &["occasions"]),
            (dictgen::InsensitiveStr::Ascii("userspase"), &["userspace"]),
            (
                dictgen::InsensitiveStr::Ascii("functiioned"),
                &["functioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("impelment"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("allignments"),
                &["alignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysie"), &["analyse"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginsed"),
                &["recognised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramalised"),
                &["normalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumulate"), &["simulate"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondigly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("diasabling"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("noramlize"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("debguging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("decosers"), &["decoders"]),
            (
                dictgen::InsensitiveStr::Ascii("inaccessable"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componsites"),
                &["composites"],
            ),
            (dictgen::InsensitiveStr::Ascii("asbolute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("repetions"),
                &["repetitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("observabl"), &["observable"]),
            (
                dictgen::InsensitiveStr::Ascii("dissimmetrical"),
                &["dissymmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("mothod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("superintendant"),
                &["superintendent"],
            ),
            (dictgen::InsensitiveStr::Ascii("panickin"), &["panicking"]),
            (dictgen::InsensitiveStr::Ascii("algotirhm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("contnet"),
                &["content", "continent"],
            ),
            (dictgen::InsensitiveStr::Ascii("androider"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("apearing"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("estimlator"), &["estimator"]),
            (
                dictgen::InsensitiveStr::Ascii("unstallation"),
                &["installation", "uninstallation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persentation"),
                &["presentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parallism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("culutral"), &["cultural"]),
            (
                dictgen::InsensitiveStr::Ascii("canditates"),
                &["candidates"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgradded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("comapares"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("encrytp"), &["encrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("graudation"),
                &["graduation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motehrboard"),
                &["motherboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasposition"),
                &["transposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("disolve"), &["dissolve"]),
            (dictgen::InsensitiveStr::Ascii("recore"), &["record"]),
            (
                dictgen::InsensitiveStr::Ascii("contretely"),
                &["concretely"],
            ),
            (dictgen::InsensitiveStr::Ascii("charcaol"), &["charcoal"]),
            (dictgen::InsensitiveStr::Ascii("reportign"), &["reporting"]),
            (dictgen::InsensitiveStr::Ascii("puchases"), &["purchases"]),
            (
                dictgen::InsensitiveStr::Ascii("recreateation"),
                &["recreation"],
            ),
            (dictgen::InsensitiveStr::Ascii("absouldly"), &["absolutely"]),
            (dictgen::InsensitiveStr::Ascii("snaphots"), &["snapshots"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanced"),
                &["enhanced", "enchanted"],
            ),
            (dictgen::InsensitiveStr::Ascii("solarmutx"), &["solarmutex"]),
            (dictgen::InsensitiveStr::Ascii("occsions"), &["occasions"]),
            (dictgen::InsensitiveStr::Ascii("sucseeds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("ocupied"), &["occupied"]),
            (dictgen::InsensitiveStr::Ascii("physisian"), &["physician"]),
            (dictgen::InsensitiveStr::Ascii("benjes"), &["binges"]),
            (dictgen::InsensitiveStr::Ascii("materail"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("fraterntiy"),
                &["fraternity"],
            ),
            (dictgen::InsensitiveStr::Ascii("clipboad"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("navagating"),
                &["navigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("nullalbe"), &["nullable"]),
            (dictgen::InsensitiveStr::Ascii("thursdsay"), &["thursdays"]),
            (dictgen::InsensitiveStr::Ascii("capcaity"), &["capacity"]),
            (
                dictgen::InsensitiveStr::Ascii("cockraoches"),
                &["cockroaches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crankenstein"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("conteiner"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("conciderably"),
                &["considerably"],
            ),
            (dictgen::InsensitiveStr::Ascii("terribley"), &["terribly"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisidor"),
                &["inquisitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("dokced"), &["docked"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelogram"),
                &["parallelogram"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functinoing"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relativisme"),
                &["relatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("overpiad"), &["overpaid"]),
            (dictgen::InsensitiveStr::Ascii("conrrupts"), &["corrupts"]),
            (
                dictgen::InsensitiveStr::Ascii("legitimite"),
                &["legitimate"],
            ),
            (dictgen::InsensitiveStr::Ascii("gnerators"), &["generators"]),
            (
                dictgen::InsensitiveStr::Ascii("predeterminded"),
                &["predetermined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enexpected"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("logrithms"), &["logarithms"]),
            (
                dictgen::InsensitiveStr::Ascii("encaspulation"),
                &["encapsulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimacion"),
                &["estimation"],
            ),
            (dictgen::InsensitiveStr::Ascii("charactar"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("idendities"),
                &["identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("replcaof"), &["replicaof"]),
            (
                dictgen::InsensitiveStr::Ascii("tranformation"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deferals"), &["deferrals"]),
            (
                dictgen::InsensitiveStr::Ascii("hopfull"),
                &["hopeful", "hopefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("egyptains"), &["egyptians"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomrers"),
                &["performers"],
            ),
            (dictgen::InsensitiveStr::Ascii("appliable"), &["applicable"]),
            (
                dictgen::InsensitiveStr::Ascii("formidabel"),
                &["formidable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescribtions"),
                &["prescriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("heatskin"), &["heatsink"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstandimg"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisioned"),
                &["commissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vallay"),
                &["valet", "valley"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("correctely"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("valdity"), &["validity"]),
            (
                dictgen::InsensitiveStr::Ascii("negogtiating"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignornace"), &["ignorance"]),
            (dictgen::InsensitiveStr::Ascii("callenger"), &["challenger"]),
            (
                dictgen::InsensitiveStr::Ascii("comemoretion"),
                &["commemoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("rnage"), &["rage", "range"]),
            (dictgen::InsensitiveStr::Ascii("exemples"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("dispair"), &["despair"]),
            (dictgen::InsensitiveStr::Ascii("proccesses"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("intencional"),
                &["intentional"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimachd"), &["mismatched"]),
            (
                dictgen::InsensitiveStr::Ascii("mulfunctioned"),
                &["malfunctioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("mcreggor"), &["mcgregor"]),
            (dictgen::InsensitiveStr::Ascii("flem"), &["phlegm"]),
            (dictgen::InsensitiveStr::Ascii("assihned"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("intangibil"),
                &["intangible"],
            ),
            (dictgen::InsensitiveStr::Ascii("asburdity"), &["absurdity"]),
            (
                dictgen::InsensitiveStr::Ascii("paralleize"),
                &["parallelize"],
            ),
            (dictgen::InsensitiveStr::Ascii("opauqe"), &["opaque"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitvely"),
                &["sensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuztomizes"),
                &["customizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("rumtime"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("openning"), &["opening"]),
            (dictgen::InsensitiveStr::Ascii("intertial"), &["inertial"]),
            (
                dictgen::InsensitiveStr::Ascii("accelorate"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infallibale"),
                &["infallible"],
            ),
            (dictgen::InsensitiveStr::Ascii("assiatnce"), &["assistance"]),
            (dictgen::InsensitiveStr::Ascii("empressed"), &["impressed"]),
            (
                dictgen::InsensitiveStr::Ascii("formallised"),
                &["formalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensability"),
                &["extensibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initalialisation"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characyers"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jusridiction"),
                &["jurisdiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprodicing"),
                &["reproducing"],
            ),
            (dictgen::InsensitiveStr::Ascii("incremment"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("technicalty"),
                &["technicality"],
            ),
            (dictgen::InsensitiveStr::Ascii("anglular"), &["angular"]),
            (dictgen::InsensitiveStr::Ascii("formend"), &["formed"]),
            (dictgen::InsensitiveStr::Ascii("rivlaries"), &["rivalries"]),
            (dictgen::InsensitiveStr::Ascii("tolearnce"), &["tolerance"]),
            (dictgen::InsensitiveStr::Ascii("developpe"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("wirhin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("purchsed"), &["purchased"]),
            (dictgen::InsensitiveStr::Ascii("spefific"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("whistlin"), &["whistling"]),
            (dictgen::InsensitiveStr::Ascii("margart"), &["margaret"]),
            (
                dictgen::InsensitiveStr::Ascii("cotributed"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unjustifed"),
                &["unjustified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotioation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronounched"),
                &["pronounced"],
            ),
            (dictgen::InsensitiveStr::Ascii("direrctors"), &["directors"]),
            (
                dictgen::InsensitiveStr::Ascii("indivuduality"),
                &["individuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreiciate"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("britener"), &["brightener"]),
            (
                dictgen::InsensitiveStr::Ascii("meditereanean"),
                &["mediterranean"],
            ),
            (dictgen::InsensitiveStr::Ascii("entety"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("suces"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("quadranle"), &["quadrangle"]),
            (dictgen::InsensitiveStr::Ascii("overrride"), &["override"]),
            (
                dictgen::InsensitiveStr::Ascii("quantitties"),
                &["quantities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalance"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("samled"), &["sampled"]),
            (
                dictgen::InsensitiveStr::Ascii("leves"),
                &["levees", "levels"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirrirs"), &["mirrors"]),
            (dictgen::InsensitiveStr::Ascii("qeuest"), &["quest"]),
            (
                dictgen::InsensitiveStr::Ascii("capapcitance"),
                &["capacitance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recursivley"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("mamaging"), &["managing"]),
            (
                dictgen::InsensitiveStr::Ascii("valyes"),
                &["values", "valses", "valves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rembered"),
                &["remembered", "renumbered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremments"),
                &["increments"],
            ),
            (dictgen::InsensitiveStr::Ascii("verfifying"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("partitionss"),
                &["partitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpression"),
                &["compression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transolations"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("wronf"), &["wrong"]),
            (dictgen::InsensitiveStr::Ascii("follwowed"), &["followed"]),
            (dictgen::InsensitiveStr::Ascii("rasterer"), &["rasterizer"]),
            (
                dictgen::InsensitiveStr::Ascii("quailfiers"),
                &["qualifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("percenage"), &["percentage"]),
            (
                dictgen::InsensitiveStr::Ascii("satisifying"),
                &["satisfying"],
            ),
            (dictgen::InsensitiveStr::Ascii("admisssion"), &["admission"]),
            (
                dictgen::InsensitiveStr::Ascii("ubelievebly"),
                &["unbelievably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiator"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fluroescent"),
                &["fluorescent"],
            ),
            (dictgen::InsensitiveStr::Ascii("inisial"), &["initial"]),
            (dictgen::InsensitiveStr::Ascii("direcitve"), &["directive"]),
            (
                dictgen::InsensitiveStr::Ascii("supplament"),
                &["supplement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginizers"),
                &["recognizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooved"), &["proved"]),
            (
                dictgen::InsensitiveStr::Ascii("manupilating"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepeting"), &["accepting"]),
            (dictgen::InsensitiveStr::Ascii("indicence"), &["incidence"]),
            (dictgen::InsensitiveStr::Ascii("hallaluja"), &["hallelujah"]),
            (dictgen::InsensitiveStr::Ascii("ulrasonic"), &["ultrasonic"]),
            (
                dictgen::InsensitiveStr::Ascii("encrcyption"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("shematics"), &["schematics"]),
            (
                dictgen::InsensitiveStr::Ascii("equialence"),
                &["equivalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aysnchronously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepresented"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funciotnal"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("ppolygons"), &["polygons"]),
            (
                dictgen::InsensitiveStr::Ascii("adventureous"),
                &["adventures", "adventurous"],
            ),
            (dictgen::InsensitiveStr::Ascii("unaquired"), &["unacquired"]),
            (
                dictgen::InsensitiveStr::Ascii("lightweigth"),
                &["lightweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistenties"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpeded"),
                &["unexpected", "unexpended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percantage"),
                &["percentage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rememebred"),
                &["remembered"],
            ),
            (dictgen::InsensitiveStr::Ascii("nottice"), &["notice"]),
            (
                dictgen::InsensitiveStr::Ascii("relinguishing"),
                &["relinquishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requiremensts"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contestents"),
                &["contestants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orgnaisation"),
                &["organisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("imperiaal"), &["imperial"]),
            (dictgen::InsensitiveStr::Ascii("courtesey"), &["courtesy"]),
            (
                dictgen::InsensitiveStr::Ascii("premutation"),
                &["permutation"],
            ),
            (dictgen::InsensitiveStr::Ascii("caharacter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("stratigies"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efinitions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("melanotin"), &["melatonin"]),
            (
                dictgen::InsensitiveStr::Ascii("hypvervisor"),
                &["hypervisor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnaslators"),
                &["translators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinces"),
                &["experiences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarranties"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthogonalizaion"),
                &["orthogonalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("extesnion"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("discrepance"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("imrpoving"), &["improving"]),
            (
                dictgen::InsensitiveStr::Ascii("apparatuss"),
                &["apparatuses", "apparatus"],
            ),
            (dictgen::InsensitiveStr::Ascii("negogtiate"), &["negotiate"]),
            (dictgen::InsensitiveStr::Ascii("clasify"), &["classify"]),
            (dictgen::InsensitiveStr::Ascii("searhing"), &["searching"]),
            (
                dictgen::InsensitiveStr::Ascii("qualitifications"),
                &["qualifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrresponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoduction"),
                &["reproduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bulletprof"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("physciatrist"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("masteriers"), &["masteries"]),
            (dictgen::InsensitiveStr::Ascii("aleinate"), &["alienate"]),
            (dictgen::InsensitiveStr::Ascii("relaise"), &["realise"]),
            (
                dictgen::InsensitiveStr::Ascii("mechinical"),
                &["mechanical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contriubted"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priveledged"),
                &["privileged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realting"),
                &["relating", "reacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("substask"), &["subtask"]),
            (
                dictgen::InsensitiveStr::Ascii("bloodborbe"),
                &["bloodborne"],
            ),
            (dictgen::InsensitiveStr::Ascii("laoders"), &["loaders"]),
            (
                dictgen::InsensitiveStr::Ascii("transcondings"),
                &["transcodings"],
            ),
            (dictgen::InsensitiveStr::Ascii("captably"), &["capably"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitialses"),
                &["uninitialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("flatteur"), &["flatter"]),
            (dictgen::InsensitiveStr::Ascii("syntehsis"), &["synthesis"]),
            (dictgen::InsensitiveStr::Ascii("antivrius"), &["antivirus"]),
            (dictgen::InsensitiveStr::Ascii("rebulids"), &["rebuilds"]),
            (
                dictgen::InsensitiveStr::Ascii("extrordinary"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("canistre"), &["canister"]),
            (
                dictgen::InsensitiveStr::Ascii("intereting"),
                &["interesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("pythonl"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhagan"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("projecs"), &["projects"]),
            (
                dictgen::InsensitiveStr::Ascii("neast"),
                &["nearest", "beast"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsunscribed"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("driagram"), &["diagram"]),
            (dictgen::InsensitiveStr::Ascii("causailty"), &["causality"]),
            (dictgen::InsensitiveStr::Ascii("cotrolles"), &["controls"]),
            (
                dictgen::InsensitiveStr::Ascii("pairocheality"),
                &["parochiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considerabley"),
                &["considerably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparition"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("ambigunous"), &["ambiguous"]),
            (dictgen::InsensitiveStr::Ascii("emminently"), &["eminently"]),
            (dictgen::InsensitiveStr::Ascii("addtions"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("genereric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("descrobe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("percente"), &["percentile"]),
            (
                dictgen::InsensitiveStr::Ascii("victorinus"),
                &["victorious"],
            ),
            (dictgen::InsensitiveStr::Ascii("desitned"), &["destined"]),
            (dictgen::InsensitiveStr::Ascii("switcing"), &["switching"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlsating"),
                &["translating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexspected"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("cenitpede"), &["centipede"]),
            (dictgen::InsensitiveStr::Ascii("etensible"), &["extensible"]),
            (
                dictgen::InsensitiveStr::Ascii("inefficieny"),
                &["inefficiency", "inefficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relentlessy"),
                &["relentlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persisiting"),
                &["persisting"],
            ),
            (dictgen::InsensitiveStr::Ascii("neutraal"), &["neutral"]),
            (dictgen::InsensitiveStr::Ascii("utiliies"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("opinyons"), &["opinions"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondind"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composiblity"),
                &["composability"],
            ),
            (dictgen::InsensitiveStr::Ascii("uggest"), &["suggest"]),
            (
                dictgen::InsensitiveStr::Ascii("understsnd"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("sawtayed"), &["sautéd"]),
            (dictgen::InsensitiveStr::Ascii("merory"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("usege"), &["usage"]),
            (dictgen::InsensitiveStr::Ascii("describin"), &["describing"]),
            (
                dictgen::InsensitiveStr::Ascii("defference"),
                &["difference", "deference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("montioning"),
                &["mentioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("esctatic"), &["ecstatic"]),
            (dictgen::InsensitiveStr::Ascii("exemple"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("daspora"), &["diaspora"]),
            (dictgen::InsensitiveStr::Ascii("acadamies"), &["academies"]),
            (dictgen::InsensitiveStr::Ascii("bttr"), &["better"]),
            (dictgen::InsensitiveStr::Ascii("reoslvers"), &["resolvers"]),
            (
                dictgen::InsensitiveStr::Ascii("dictaionaries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ckeckpoint"),
                &["checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("movemnt"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("communicted"),
                &["communicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philisophers"),
                &["philosophers"],
            ),
            (dictgen::InsensitiveStr::Ascii("addpress"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("opperator"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("securiy"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("modifires"), &["modifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("sentimentals"),
                &["sentiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappearring"),
                &["disappearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regenrating"),
                &["regenerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prodcution"),
                &["production"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresopnd"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("decryped"), &["decrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("ventillated"),
                &["ventilated"],
            ),
            (dictgen::InsensitiveStr::Ascii("printting"), &["printing"]),
            (dictgen::InsensitiveStr::Ascii("fomrats"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("behavoueirs"),
                &["behaviours"],
            ),
            (dictgen::InsensitiveStr::Ascii("capiton"), &["caption"]),
            (dictgen::InsensitiveStr::Ascii("accorting"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("overextimating"),
                &["overestimating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depencencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cathering"),
                &["catering", "gathering"],
            ),
            (dictgen::InsensitiveStr::Ascii("condtions"), &["conditions"]),
            (
                dictgen::InsensitiveStr::Ascii("inable"),
                &["enable", "unable"],
            ),
            (dictgen::InsensitiveStr::Ascii("reapeat"), &["repeat"]),
            (dictgen::InsensitiveStr::Ascii("chatolics"), &["catholics"]),
            (
                dictgen::InsensitiveStr::Ascii("highighted"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("amendend"), &["amended"]),
            (dictgen::InsensitiveStr::Ascii("aboiut"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("stndard"), &["standard"]),
            (dictgen::InsensitiveStr::Ascii("fazin"), &["fazing"]),
            (
                dictgen::InsensitiveStr::Ascii("enchancments"),
                &["enchantments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instatiates"),
                &["instantiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("winchestor"),
                &["winchester"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementiaon"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("conexant"), &["connexant"]),
            (
                dictgen::InsensitiveStr::Ascii("sypmathetic"),
                &["sympathetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("booelan"), &["boolean"]),
            (
                dictgen::InsensitiveStr::Ascii("reserachers"),
                &["researchers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evertyhing"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassmebled"),
                &["disassembled"],
            ),
            (dictgen::InsensitiveStr::Ascii("transconde"), &["transcode"]),
            (dictgen::InsensitiveStr::Ascii("hotname"), &["hostname"]),
            (dictgen::InsensitiveStr::Ascii("notifer"), &["notifier"]),
            (dictgen::InsensitiveStr::Ascii("interet"), &["interest"]),
            (dictgen::InsensitiveStr::Ascii("synthsize"), &["synthesize"]),
            (dictgen::InsensitiveStr::Ascii("swicthing"), &["switching"]),
            (
                dictgen::InsensitiveStr::Ascii("metephorical"),
                &["metaphorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("changset"), &["changeset"]),
            (dictgen::InsensitiveStr::Ascii("organsim"), &["organism"]),
            (dictgen::InsensitiveStr::Ascii("critcials"), &["criticals"]),
            (
                dictgen::InsensitiveStr::Ascii("uptreaming"),
                &["upstreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("assignes"), &["assigns"]),
            (dictgen::InsensitiveStr::Ascii("herculees"), &["hercules"]),
            (
                dictgen::InsensitiveStr::Ascii("concentraze"),
                &["concentrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("relegato"), &["relegation"]),
            (
                dictgen::InsensitiveStr::Ascii("eveyrthing"),
                &["everything"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulament"), &["regulate"]),
            (dictgen::InsensitiveStr::Ascii("romotes"), &["remotes"]),
            (
                dictgen::InsensitiveStr::Ascii("proffesions"),
                &["professions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udnerstands"),
                &["understands"],
            ),
            (dictgen::InsensitiveStr::Ascii("randomally"), &["randomly"]),
            (dictgen::InsensitiveStr::Ascii("obesssion"), &["obsession"]),
            (dictgen::InsensitiveStr::Ascii("comfigs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("metaprogamming"),
                &["metaprogramming"],
            ),
            (dictgen::InsensitiveStr::Ascii("doucments"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("cliping"), &["clipping"]),
            (dictgen::InsensitiveStr::Ascii("throug"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphisical"),
                &["metaphysical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowldgment"),
                &["acknowledgment"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahche"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("certfy"), &["certify"]),
            (
                dictgen::InsensitiveStr::Ascii("mathematitians"),
                &["mathematicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingishes"),
                &["distinguishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practicly"),
                &["practically"],
            ),
            (dictgen::InsensitiveStr::Ascii("progarmer"), &["programmer"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfeared"),
                &["interfered"],
            ),
            (dictgen::InsensitiveStr::Ascii("aracnids"), &["arachnids"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepraneurs"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("niot"), &["not"]),
            (
                dictgen::InsensitiveStr::Ascii("determinisitically"),
                &["deterministically"],
            ),
            (dictgen::InsensitiveStr::Ascii("acerage"), &["acreage"]),
            (dictgen::InsensitiveStr::Ascii("tortila"), &["tortilla"]),
            (
                dictgen::InsensitiveStr::Ascii("declearation"),
                &["declaration", "deceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backkground"),
                &["background"],
            ),
            (dictgen::InsensitiveStr::Ascii("pratical"), &["practical"]),
            (
                dictgen::InsensitiveStr::Ascii("acelerates"),
                &["accelerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represente"),
                &["represents", "represented", "represent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infoemation"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprhan"), &["orphan"]),
            (dictgen::InsensitiveStr::Ascii("passionais"), &["passions"]),
            (
                dictgen::InsensitiveStr::Ascii("affortable"),
                &["affordable"],
            ),
            (dictgen::InsensitiveStr::Ascii("amercia"), &["america"]),
            (
                dictgen::InsensitiveStr::Ascii("pilgramidge"),
                &["pilgrimage"],
            ),
            (dictgen::InsensitiveStr::Ascii("dischage"), &["discharge"]),
            (
                dictgen::InsensitiveStr::Ascii("unrealistisch"),
                &["unrealistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connstrained"),
                &["constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permuations"),
                &["permutations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chractor"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("stimulatie"),
                &["stimulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supplamented"),
                &["supplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("chrenobyl"), &["chernobyl"]),
            (dictgen::InsensitiveStr::Ascii("dribbel"), &["dribble"]),
            (dictgen::InsensitiveStr::Ascii("matych"), &["match"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiating"),
                &["approximating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unoptimiced"),
                &["unoptimised", "unoptimized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contstructs"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("psycothic"), &["psychotic"]),
            (
                dictgen::InsensitiveStr::Ascii("descentants"),
                &["descendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arcitecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("apmplify"), &["amplify"]),
            (
                dictgen::InsensitiveStr::Ascii("idendifies"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphabeticallly"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("summmation"), &["summation"]),
            (
                dictgen::InsensitiveStr::Ascii("compilcations"),
                &["complication", "complications"],
            ),
            (dictgen::InsensitiveStr::Ascii("recogises"), &["recognises"]),
            (dictgen::InsensitiveStr::Ascii("westernes"), &["westerners"]),
            (
                dictgen::InsensitiveStr::Ascii("definitaly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeucte"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("archievements"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("subscribt"), &["subscript"]),
            (dictgen::InsensitiveStr::Ascii("peopels"), &["peoples"]),
            (dictgen::InsensitiveStr::Ascii("dependece"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("inddividuals"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("conainter"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("importante"), &["important"]),
            (dictgen::InsensitiveStr::Ascii("explian"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("tittled"), &["titled"]),
            (
                dictgen::InsensitiveStr::Ascii("distrtibutions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("stanards"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("cuurrent"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("regualtion"),
                &["regulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particaularly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("actially"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("underwheling"),
                &["underwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("fornater"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("signficant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fatest"),
                &["fastest", "fattest"],
            ),
            (dictgen::InsensitiveStr::Ascii("dowloader"), &["downloader"]),
            (dictgen::InsensitiveStr::Ascii("chassim"), &["chasm"]),
            (
                dictgen::InsensitiveStr::Ascii("finalizatiom"),
                &["finalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciefied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("timstamps"), &["timestamps"]),
            (
                dictgen::InsensitiveStr::Ascii("approoximates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("protcols"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("whitre"), &["white"]),
            (
                dictgen::InsensitiveStr::Ascii("distinguises"),
                &["distinguishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("crowshay"), &["crochet"]),
            (dictgen::InsensitiveStr::Ascii("trotzkist"), &["trotskyist"]),
            (dictgen::InsensitiveStr::Ascii("subltety"), &["subtlety"]),
            (dictgen::InsensitiveStr::Ascii("erquest"), &["request"]),
            (dictgen::InsensitiveStr::Ascii("whathever"), &["whatever"]),
            (
                dictgen::InsensitiveStr::Ascii("judgemetal"),
                &["judgemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferably"),
                &["preferably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabiliites"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("perforam"), &["perform"]),
            (
                dictgen::InsensitiveStr::Ascii("democracize"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirimented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gratificacion"),
                &["gratification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracteristically"),
                &["characteristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrubutor"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpecting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("unfurtunate"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprecate"), &["appreciate"]),
            (dictgen::InsensitiveStr::Ascii("particel"), &["particle"]),
            (
                dictgen::InsensitiveStr::Ascii("limtiations"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("widged"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("driects"), &["directs"]),
            (dictgen::InsensitiveStr::Ascii("compearing"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("typpos"), &["typos"]),
            (dictgen::InsensitiveStr::Ascii("cleanes"), &["cleanse"]),
            (dictgen::InsensitiveStr::Ascii("acended"), &["ascended"]),
            (dictgen::InsensitiveStr::Ascii("acces"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("artuments"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("outpusts"), &["outputs"]),
            (
                dictgen::InsensitiveStr::Ascii("intiailizeing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vitually"),
                &["virtually", "visually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selectoins"),
                &["selections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scandinvia"),
                &["scandinavia"],
            ),
            (dictgen::InsensitiveStr::Ascii("pezier"), &["bezier"]),
            (
                dictgen::InsensitiveStr::Ascii("anonymouly"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("evice"), &["evict"]),
            (
                dictgen::InsensitiveStr::Ascii("placehodlers"),
                &["placeholders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cusomisations"),
                &["customisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocatve"), &["octave"]),
            (
                dictgen::InsensitiveStr::Ascii("vegaterians"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alcoholc"),
                &["alcoholics", "alcoholic"],
            ),
            (dictgen::InsensitiveStr::Ascii("animatie"), &["animate"]),
            (
                dictgen::InsensitiveStr::Ascii("dictoinaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("resverved"), &["reserved"]),
            (
                dictgen::InsensitiveStr::Ascii("functionallly"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemeneted"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("genenric"), &["generic"]),
            (dictgen::InsensitiveStr::Ascii("closeing"), &["closing"]),
            (dictgen::InsensitiveStr::Ascii("mophology"), &["morphology"]),
            (dictgen::InsensitiveStr::Ascii("rouine"), &["routine"]),
            (dictgen::InsensitiveStr::Ascii("elimiated"), &["eliminated"]),
            (dictgen::InsensitiveStr::Ascii("carreer"), &["career"]),
            (dictgen::InsensitiveStr::Ascii("connectred"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("beaurocratically"),
                &["bureaucratically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pepperment"),
                &["peppermint"],
            ),
            (dictgen::InsensitiveStr::Ascii("prominant"), &["prominent"]),
            (
                dictgen::InsensitiveStr::Ascii("clearstory"),
                &["clerestory"],
            ),
            (dictgen::InsensitiveStr::Ascii("charcol"), &["charcoal"]),
            (dictgen::InsensitiveStr::Ascii("enteries"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("leyer"), &["layer"]),
            (dictgen::InsensitiveStr::Ascii("aotomatic"), &["automatic"]),
            (
                dictgen::InsensitiveStr::Ascii("dictatorshop"),
                &["dictatorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragphs"), &["paragraphs"]),
            (dictgen::InsensitiveStr::Ascii("excetute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("bycicles"), &["bicycles"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeseborger"),
                &["cheeseburger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probbailities"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("dashbords"), &["dashboards"]),
            (
                dictgen::InsensitiveStr::Ascii("enivronment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("cildren"), &["children"]),
            (dictgen::InsensitiveStr::Ascii("meaningess"), &["meanings"]),
            (
                dictgen::InsensitiveStr::Ascii("deregulaiton"),
                &["deregulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consructions"),
                &["constructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("removees"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("reacll"), &["recall"]),
            (dictgen::InsensitiveStr::Ascii("detectas"), &["detects"]),
            (dictgen::InsensitiveStr::Ascii("egpytians"), &["egyptians"]),
            (
                dictgen::InsensitiveStr::Ascii("althorithmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("balcksmith"),
                &["blacksmith"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralellise"),
                &["parallelise"],
            ),
            (dictgen::InsensitiveStr::Ascii("peferring"), &["preferring"]),
            (dictgen::InsensitiveStr::Ascii("cantains"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("symphoney"), &["symphony"]),
            (
                dictgen::InsensitiveStr::Ascii("servces"),
                &["services", "serves"],
            ),
            (dictgen::InsensitiveStr::Ascii("fantasty"), &["fantasy"]),
            (
                dictgen::InsensitiveStr::Ascii("addtitionally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoretial"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("followin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("acctivate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("repore"),
                &["rapport", "report", "repose"],
            ),
            (dictgen::InsensitiveStr::Ascii("britenes"), &["brightenes"]),
            (dictgen::InsensitiveStr::Ascii("electorns"), &["electrons"]),
            (dictgen::InsensitiveStr::Ascii("preocess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("countiguous"),
                &["contiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("reslut"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpanced"),
                &["unexpanded"],
            ),
            (dictgen::InsensitiveStr::Ascii("behngazi"), &["benghazi"]),
            (dictgen::InsensitiveStr::Ascii("blessure"), &["blessing"]),
            (
                dictgen::InsensitiveStr::Ascii("foundatoin"),
                &["foundations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmnets"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoluet"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("predictivo"),
                &["prediction"],
            ),
            (dictgen::InsensitiveStr::Ascii("comapany"), &["company"]),
            (dictgen::InsensitiveStr::Ascii("pubish"), &["publish"]),
            (
                dictgen::InsensitiveStr::Ascii("connectivty"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("pengwin"), &["penguin"]),
            (
                dictgen::InsensitiveStr::Ascii("comptuation"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("americaps"), &["americas"]),
            (
                dictgen::InsensitiveStr::Ascii("omishints"),
                &["omniscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("explcitly"), &["explicitly"]),
            (dictgen::InsensitiveStr::Ascii("advanes"), &["advances"]),
            (dictgen::InsensitiveStr::Ascii("versoined"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("viatmins"), &["vitamins"]),
            (
                dictgen::InsensitiveStr::Ascii("beastiality"),
                &["bestiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramaterless"),
                &["parameterless"],
            ),
            (dictgen::InsensitiveStr::Ascii("opiod"), &["opioid"]),
            (dictgen::InsensitiveStr::Ascii("versuons"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("leesure"), &["leisure"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplicty"),
                &["multiplicity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hystarical"),
                &["hysterical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exucutions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foded"),
                &["boded", "coded", "faded", "folded", "forded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undermimes"),
                &["undermines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nformation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vlues"),
                &["values", "blues"],
            ),
            (dictgen::InsensitiveStr::Ascii("nubmer"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("mutux"), &["mutex"]),
            (dictgen::InsensitiveStr::Ascii("maximuim"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonators"),
                &["opinionators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarenteerd"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("givven"), &["given"]),
            (dictgen::InsensitiveStr::Ascii("eegerly"), &["eagerly"]),
            (
                dictgen::InsensitiveStr::Ascii("postponinig"),
                &["postponing"],
            ),
            (dictgen::InsensitiveStr::Ascii("causion"), &["caution"]),
            (dictgen::InsensitiveStr::Ascii("uplaoded"), &["uploaded"]),
            (dictgen::InsensitiveStr::Ascii("resinged"), &["resigned"]),
            (
                dictgen::InsensitiveStr::Ascii("attivations"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cigaretes"), &["cigarettes"]),
            (
                dictgen::InsensitiveStr::Ascii("condenming"),
                &["condemning"],
            ),
            (dictgen::InsensitiveStr::Ascii("recogized"), &["recognized"]),
            (
                dictgen::InsensitiveStr::Ascii("identidication"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("vluing"), &["valuing"]),
            (
                dictgen::InsensitiveStr::Ascii("illustratin"),
                &["illustrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("surakata"), &["surakarta"]),
            (
                dictgen::InsensitiveStr::Ascii("tendancies"),
                &["tendencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppresing"),
                &["suppressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mdified"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("iinterval"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("enqueud"), &["enqueued"]),
            (dictgen::InsensitiveStr::Ascii("verfifier"), &["verifier"]),
            (dictgen::InsensitiveStr::Ascii("guarteed"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("transphonic"),
                &["transphobic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gererations"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("contetxt"), &["context"]),
            (dictgen::InsensitiveStr::Ascii("truw"), &["true"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourood"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decalrator"),
                &["declarator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alterated"),
                &["altered", "alternated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promiscious"),
                &["promiscuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("calibrte"), &["calibrate"]),
            (
                dictgen::InsensitiveStr::Ascii("disproportinate"),
                &["disproportionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmenetaions"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acitivation"),
                &["activation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("domisticated"),
                &["domesticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("golbally"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("quwessant"), &["croissant"]),
            (dictgen::InsensitiveStr::Ascii("clloser"), &["closer"]),
            (dictgen::InsensitiveStr::Ascii("disabl"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("generaes"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("millisenconds"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouptuted"), &["outputted"]),
            (dictgen::InsensitiveStr::Ascii("furctose"), &["fructose"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstruccion"),
                &["reconstruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("supooses"), &["supposes"]),
            (dictgen::InsensitiveStr::Ascii("remainign"), &["remaining"]),
            (dictgen::InsensitiveStr::Ascii("realated"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("anitaliasing"),
                &["antialiasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("nodulated"), &["modulated"]),
            (
                dictgen::InsensitiveStr::Ascii("diversed"),
                &["diverse", "diverged"],
            ),
            (dictgen::InsensitiveStr::Ascii("wlaking"), &["walking"]),
            (
                dictgen::InsensitiveStr::Ascii("calcluation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicslosures"),
                &["disclosures"],
            ),
            (dictgen::InsensitiveStr::Ascii("seuence"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("valiator"), &["validator"]),
            (dictgen::InsensitiveStr::Ascii("instntly"), &["instantly"]),
            (dictgen::InsensitiveStr::Ascii("expexcted"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("teanant"), &["tenant"]),
            (
                dictgen::InsensitiveStr::Ascii("intergals"),
                &["integrals", "intervals", "internals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialsers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadvandage"),
                &["disadvantaged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misalligned"),
                &["misaligned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctructive"),
                &["destructive"],
            ),
            (dictgen::InsensitiveStr::Ascii("separion"), &["separation"]),
            (dictgen::InsensitiveStr::Ascii("impreving"), &["improving"]),
            (
                dictgen::InsensitiveStr::Ascii("constituante"),
                &["constituents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowledgabe"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("enble"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("hyperboly"), &["hyperbole"]),
            (
                dictgen::InsensitiveStr::Ascii("emphazising"),
                &["emphasising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissasemblies"),
                &["disassemblies"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrollnig"), &["scrolling"]),
            (
                dictgen::InsensitiveStr::Ascii("manifactured"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorge"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("permieter"), &["perimeter"]),
            (dictgen::InsensitiveStr::Ascii("seperataed"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("unchallengable"),
                &["unchallengeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("delivey"), &["delivery"]),
            (
                dictgen::InsensitiveStr::Ascii("antractica"),
                &["antarctica"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacciante"), &["vaccinate"]),
            (
                dictgen::InsensitiveStr::Ascii("propectable"),
                &["projectable", "protectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discretiztion"),
                &["discretization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insallations"),
                &["installations"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifomly"), &["uniformly"]),
            (
                dictgen::InsensitiveStr::Ascii("coner"),
                &["corner", "coroner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("audioboooks"),
                &["audiobook"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphetemines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtractss"), &["subtracts"]),
            (
                dictgen::InsensitiveStr::Ascii("tramsformers"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resiliantly"),
                &["resiliently"],
            ),
            (dictgen::InsensitiveStr::Ascii("valentis"), &["valentines"]),
            (dictgen::InsensitiveStr::Ascii("initaled"), &["initialed"]),
            (dictgen::InsensitiveStr::Ascii("molcule"), &["molecule"]),
            (dictgen::InsensitiveStr::Ascii("reapirs"), &["repairs"]),
            (dictgen::InsensitiveStr::Ascii("casulaty"), &["casualty"]),
            (dictgen::InsensitiveStr::Ascii("coyping"), &["copying"]),
            (dictgen::InsensitiveStr::Ascii("polarlise"), &["polarise"]),
            (dictgen::InsensitiveStr::Ascii("accrdion"), &["accordion"]),
            (
                dictgen::InsensitiveStr::Ascii("oridinarily"),
                &["ordinarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contirbution"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alcoholicas"),
                &["alcoholics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecctors"),
                &["connectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplinare"),
                &["discipline"],
            ),
            (dictgen::InsensitiveStr::Ascii("executale"), &["executable"]),
            (
                dictgen::InsensitiveStr::Ascii("dissopointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chancs"),
                &["chances", "chunks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compariable"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("driectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("predictible"),
                &["predictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("patrcik"), &["patrick"]),
            (dictgen::InsensitiveStr::Ascii("prepareg"), &["prepare"]),
            (
                dictgen::InsensitiveStr::Ascii("derefenced"),
                &["dereferenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecuting"), &["executing"]),
            (dictgen::InsensitiveStr::Ascii("innocens"), &["innocents"]),
            (
                dictgen::InsensitiveStr::Ascii("unspefixied"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributie"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aktivating"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepectral"), &["spectral"]),
            (
                dictgen::InsensitiveStr::Ascii("inreractive"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyprt"), &["crypt"]),
            (
                dictgen::InsensitiveStr::Ascii("competitve"),
                &["competitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorthin"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("contstraining"),
                &["constraining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiugres"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("desplays"), &["displays"]),
            (dictgen::InsensitiveStr::Ascii("admiting"), &["admitting"]),
            (
                dictgen::InsensitiveStr::Ascii("expiriments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychaitric"),
                &["psychiatric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("historycaly"),
                &["historically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialisiation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("retriver"), &["retriever"]),
            (dictgen::InsensitiveStr::Ascii("enfoces"), &["enforces"]),
            (dictgen::InsensitiveStr::Ascii("insteaad"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("protext"), &["protect"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiazation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("populos"), &["populous"]),
            (
                dictgen::InsensitiveStr::Ascii("uninspiried"),
                &["uninspired"],
            ),
            (dictgen::InsensitiveStr::Ascii("inspirerad"), &["inspired"]),
            (
                dictgen::InsensitiveStr::Ascii("intriduction"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("addres"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("visisble"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronises"),
                &["synchronises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apolagizing"),
                &["apologizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mecahnisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incombatibility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understadnably"),
                &["understandably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatatible"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjancent"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentacion"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuring"), &["occurring"]),
            (
                dictgen::InsensitiveStr::Ascii("motherboars"),
                &["motherboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lannsiters"),
                &["lannisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unilatreal"),
                &["unilateral"],
            ),
            (dictgen::InsensitiveStr::Ascii("aloowance"), &["allowance"]),
            (
                dictgen::InsensitiveStr::Ascii("experssing"),
                &["expressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("suporter"), &["supporter"]),
            (dictgen::InsensitiveStr::Ascii("virtuell"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("approoximating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("stattues"), &["statutes"]),
            (dictgen::InsensitiveStr::Ascii("intialize"), &["initialize"]),
            (dictgen::InsensitiveStr::Ascii("acticely"), &["actively"]),
            (
                dictgen::InsensitiveStr::Ascii("conflite"),
                &["conflict", "conflate"],
            ),
            (dictgen::InsensitiveStr::Ascii("resover"), &["resolver"]),
            (dictgen::InsensitiveStr::Ascii("ncessary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("fime"), &["fixme", "time"]),
            (dictgen::InsensitiveStr::Ascii("climats"), &["climates"]),
            (dictgen::InsensitiveStr::Ascii("neihgbor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("comutative"),
                &["commutative", "computative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advantageos"),
                &["advantageous"],
            ),
            (dictgen::InsensitiveStr::Ascii("vey"), &["very"]),
            (
                dictgen::InsensitiveStr::Ascii("supspicions"),
                &["suspicions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphisics"),
                &["metaphysics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreceation"),
                &["depreciation", "deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reveald"), &["revealed"]),
            (dictgen::InsensitiveStr::Ascii("hopefuly"), &["hopefully"]),
            (
                dictgen::InsensitiveStr::Ascii("subscryber"),
                &["subscriber"],
            ),
            (dictgen::InsensitiveStr::Ascii("unandled"), &["unhandled"]),
            (dictgen::InsensitiveStr::Ascii("voilates"), &["violates"]),
            (
                dictgen::InsensitiveStr::Ascii("concsiousness"),
                &["consciousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aspiratons"),
                &["aspirations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reputposes"),
                &["repurposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devleopers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminadi"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opitmizing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("folliwers"), &["followers"]),
            (
                dictgen::InsensitiveStr::Ascii("wireframws"),
                &["wireframes"],
            ),
            (dictgen::InsensitiveStr::Ascii("tunelled"), &["tunnelled"]),
            (dictgen::InsensitiveStr::Ascii("eleemnts"), &["elements"]),
            (dictgen::InsensitiveStr::Ascii("checkesum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("quesioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("recivers"), &["receivers"]),
            (
                dictgen::InsensitiveStr::Ascii("compilicate"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronizations"),
                &["synchronizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obligitary"),
                &["obligatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethentication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("benetifs"), &["benefits"]),
            (
                dictgen::InsensitiveStr::Ascii("expiriment"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("gentials"), &["genitals"]),
            (
                dictgen::InsensitiveStr::Ascii("siginificantly"),
                &["significantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attendents"),
                &["attendants"],
            ),
            (dictgen::InsensitiveStr::Ascii("brocoli"), &["broccoli"]),
            (
                dictgen::InsensitiveStr::Ascii("authrorizes"),
                &["authorizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("updadating"), &["updating"]),
            (dictgen::InsensitiveStr::Ascii("privledge"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("addiation"), &["additional"]),
            (
                dictgen::InsensitiveStr::Ascii("sexualzied"),
                &["sexualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supossedly"),
                &["supposedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("statemen"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("terriory"), &["territory"]),
            (dictgen::InsensitiveStr::Ascii("contat"), &["contact"]),
            (dictgen::InsensitiveStr::Ascii("wha"), &["what"]),
            (dictgen::InsensitiveStr::Ascii("lantren"), &["lantern"]),
            (
                dictgen::InsensitiveStr::Ascii("compatable"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contentios"),
                &["contentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dehydraded"),
                &["dehydrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("characte"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("misogonyst"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("maintian"), &["maintain"]),
            (
                dictgen::InsensitiveStr::Ascii("automaitcally"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relinguish"),
                &["relinquish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fragmenetd"),
                &["fragmented"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdens"), &["depends"]),
            (
                dictgen::InsensitiveStr::Ascii("changability"),
                &["changeability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimazations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("egals"), &["equals"]),
            (
                dictgen::InsensitiveStr::Ascii("contiunity"),
                &["continuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceotably"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnererabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("alyways"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("intialzed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("formallise"), &["formalise"]),
            (
                dictgen::InsensitiveStr::Ascii("competitiion"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("segemnt"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("migrantes"), &["migrants"]),
            (
                dictgen::InsensitiveStr::Ascii("exectutives"),
                &["executives"],
            ),
            (dictgen::InsensitiveStr::Ascii("atach"), &["attach"]),
            (
                dictgen::InsensitiveStr::Ascii("cimptumaticlly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("millenian"), &["millennia"]),
            (
                dictgen::InsensitiveStr::Ascii("afgahnistan"),
                &["afghanistan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associateing"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("androgeny"), &["androgyny"]),
            (dictgen::InsensitiveStr::Ascii("exceuted"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("perdators"), &["predators"]),
            (
                dictgen::InsensitiveStr::Ascii("fonctional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagerising"),
                &["plagiarising"],
            ),
            (dictgen::InsensitiveStr::Ascii("downgradde"), &["downgrade"]),
            (dictgen::InsensitiveStr::Ascii("priased"), &["praised"]),
            (
                dictgen::InsensitiveStr::Ascii("preperties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ussefulness"),
                &["usefulness"],
            ),
            (dictgen::InsensitiveStr::Ascii("optons"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("acnedotes"), &["anecdotes"]),
            (dictgen::InsensitiveStr::Ascii("collegues"), &["colleagues"]),
            (dictgen::InsensitiveStr::Ascii("futal"), &["futile"]),
            (
                dictgen::InsensitiveStr::Ascii("reevaulated"),
                &["reevaluate"],
            ),
            (dictgen::InsensitiveStr::Ascii("continut"), &["continuity"]),
            (
                dictgen::InsensitiveStr::Ascii("converagence"),
                &["convergence"],
            ),
            (dictgen::InsensitiveStr::Ascii("caridnal"), &["cardinal"]),
            (dictgen::InsensitiveStr::Ascii("withtout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("inutitive"), &["intuitive"]),
            (
                dictgen::InsensitiveStr::Ascii("authentciates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyborad"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("monolite"), &["monolithic"]),
            (dictgen::InsensitiveStr::Ascii("exchaust"), &["exhaust"]),
            (dictgen::InsensitiveStr::Ascii("lesbianest"), &["lesbians"]),
            (dictgen::InsensitiveStr::Ascii("negotion"), &["negotiation"]),
            (
                dictgen::InsensitiveStr::Ascii("contributos"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("nitched"), &["niched"]),
            (dictgen::InsensitiveStr::Ascii("noralize"), &["normalize"]),
            (dictgen::InsensitiveStr::Ascii("emphazied"), &["emphasized"]),
            (dictgen::InsensitiveStr::Ascii("maccro"), &["macro"]),
            (dictgen::InsensitiveStr::Ascii("sinature"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("intimidant"),
                &["intimidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("srinking"), &["shrinking"]),
            (
                dictgen::InsensitiveStr::Ascii("sustainabiltiy"),
                &["sustainability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranmitters"),
                &["transmitters"],
            ),
            (dictgen::InsensitiveStr::Ascii("negligble"), &["negligible"]),
            (
                dictgen::InsensitiveStr::Ascii("advisoriyes"),
                &["advisories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repsonsibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("matterss"), &["mattress"]),
            (
                dictgen::InsensitiveStr::Ascii("seperations"),
                &["separations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformated"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("elatic"), &["elastic"]),
            (
                dictgen::InsensitiveStr::Ascii("reffers"),
                &["refers", "reefers"],
            ),
            (dictgen::InsensitiveStr::Ascii("beauquet"), &["bouquet"]),
            (
                dictgen::InsensitiveStr::Ascii("reciproce"),
                &["reciprocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("canged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("focusf"), &["focus"]),
            (
                dictgen::InsensitiveStr::Ascii("parias"),
                &["pariahs", "parkas"],
            ),
            (dictgen::InsensitiveStr::Ascii("genuinelly"), &["genuinely"]),
            (dictgen::InsensitiveStr::Ascii("jounal"), &["journal"]),
            (
                dictgen::InsensitiveStr::Ascii("persepectives"),
                &["perspectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdenences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("mercurcy"), &["mercury"]),
            (
                dictgen::InsensitiveStr::Ascii("centenals"),
                &["centennials", "sentinels"],
            ),
            (dictgen::InsensitiveStr::Ascii("kidknaps"), &["kidnaps"]),
            (dictgen::InsensitiveStr::Ascii("falke"), &["false"]),
            (dictgen::InsensitiveStr::Ascii("thoguht"), &["thought"]),
            (
                dictgen::InsensitiveStr::Ascii("deparmental"),
                &["departmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantitize"),
                &["quantities"],
            ),
            (dictgen::InsensitiveStr::Ascii("finnisch"), &["finnish"]),
            (
                dictgen::InsensitiveStr::Ascii("represnetation"),
                &["representations", "representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropoliten"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccomending"),
                &["recommending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origianaly"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutilply"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("categort"), &["category"]),
            (
                dictgen::InsensitiveStr::Ascii("simualtion"),
                &["simulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("kibutzes"), &["kibbutzim"]),
            (
                dictgen::InsensitiveStr::Ascii("expodition"),
                &["exposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinementelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envents"),
                &["events", "invents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appication"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("slaugterhouses"),
                &["slaughterhouses"],
            ),
            (dictgen::InsensitiveStr::Ascii("vendettta"), &["vendetta"]),
            (
                dictgen::InsensitiveStr::Ascii("pictureskness"),
                &["picturesqueness"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutomer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("derivtives"),
                &["derivatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfoemnace"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appripriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empahsizes"),
                &["emphasizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configrable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("percieve"), &["perceive"]),
            (dictgen::InsensitiveStr::Ascii("refferrals"), &["referrals"]),
            (dictgen::InsensitiveStr::Ascii("amongsts"), &["amongst"]),
            (dictgen::InsensitiveStr::Ascii("ukrainisn"), &["ukrainians"]),
            (
                dictgen::InsensitiveStr::Ascii("expermienter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("assiocate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("superioris"), &["superiors"]),
            (dictgen::InsensitiveStr::Ascii("logggers"), &["loggers"]),
            (dictgen::InsensitiveStr::Ascii("continuim"), &["continuum"]),
            (dictgen::InsensitiveStr::Ascii("custoemrs"), &["customers"]),
            (
                dictgen::InsensitiveStr::Ascii("unapprieciated"),
                &["unappreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camouflague"),
                &["camouflage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extroardinary"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("corsor"), &["cursor"]),
            (
                dictgen::InsensitiveStr::Ascii("disassociatin"),
                &["disassociating", "disassociation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporyries"),
                &["temporaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acciddently"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatcts"), &["contacts"]),
            (dictgen::InsensitiveStr::Ascii("stroring"), &["storing"]),
            (dictgen::InsensitiveStr::Ascii("confustion"), &["confusion"]),
            (
                dictgen::InsensitiveStr::Ascii("dpeloyments"),
                &["deployments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("territores"),
                &["territories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preventitve"),
                &["preventative"],
            ),
            (dictgen::InsensitiveStr::Ascii("summurize"), &["summarize"]),
            (dictgen::InsensitiveStr::Ascii("unziped"), &["unzipped"]),
            (dictgen::InsensitiveStr::Ascii("pedohpile"), &["pedophile"]),
            (dictgen::InsensitiveStr::Ascii("cplipboard"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidaiton"),
                &["invalidation"],
            ),
            (dictgen::InsensitiveStr::Ascii("argentia"), &["argentina"]),
            (
                dictgen::InsensitiveStr::Ascii("optimiation"),
                &["optimization", "optimisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconvenience"),
                &["inconvenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disastruous"),
                &["disastrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideall"),
                &["ideally", "ideal"],
            ),
            (dictgen::InsensitiveStr::Ascii("emoty"), &["empty"]),
            (dictgen::InsensitiveStr::Ascii("xepected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("adminitration"),
                &["administration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyrpess"), &["cypress"]),
            (dictgen::InsensitiveStr::Ascii("formattin"), &["formatting"]),
            (
                dictgen::InsensitiveStr::Ascii("negothiable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("calllback"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("compilied"), &["compiled"]),
            (dictgen::InsensitiveStr::Ascii("refferee"), &["referee"]),
            (dictgen::InsensitiveStr::Ascii("scandaloso"), &["scandals"]),
            (
                dictgen::InsensitiveStr::Ascii("correspindong"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("perserve"), &["preserve"]),
            (
                dictgen::InsensitiveStr::Ascii("priroritize"),
                &["prioritize"],
            ),
            (dictgen::InsensitiveStr::Ascii("asteriod"), &["asteroid"]),
            (
                dictgen::InsensitiveStr::Ascii("requestes"),
                &["requests", "requested"],
            ),
            (dictgen::InsensitiveStr::Ascii("answe"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("aovoiding"), &["avoiding"]),
            (dictgen::InsensitiveStr::Ascii("simillar"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("amrenians"), &["armenians"]),
            (dictgen::InsensitiveStr::Ascii("adhears"), &["adheres"]),
            (dictgen::InsensitiveStr::Ascii("reuplader"), &["reuploader"]),
            (dictgen::InsensitiveStr::Ascii("suppresed"), &["suppressed"]),
            (dictgen::InsensitiveStr::Ascii("nothign"), &["nothing"]),
            (
                dictgen::InsensitiveStr::Ascii("negitiations"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("diggits"), &["digits"]),
            (
                dictgen::InsensitiveStr::Ascii("acopalypse"),
                &["apocalypse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mecahnical"),
                &["mechanical"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectng"), &["connecting"]),
            (dictgen::InsensitiveStr::Ascii("differernt"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("resgistries"),
                &["registries"],
            ),
            (dictgen::InsensitiveStr::Ascii("adminiter"), &["administer"]),
            (dictgen::InsensitiveStr::Ascii("oaram"), &["param"]),
            (
                dictgen::InsensitiveStr::Ascii("philippinos"),
                &["philippines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnescessarily"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alocholics"),
                &["alcoholics"],
            ),
            (dictgen::InsensitiveStr::Ascii("exectuive"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailbility"),
                &["unavailability"],
            ),
            (dictgen::InsensitiveStr::Ascii("transfrom"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("exhibites"), &["exhibits"]),
            (dictgen::InsensitiveStr::Ascii("theaded"), &["threaded"]),
            (
                dictgen::InsensitiveStr::Ascii("atributions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("presenst"), &["presets"]),
            (dictgen::InsensitiveStr::Ascii("terorrism"), &["terrorism"]),
            (dictgen::InsensitiveStr::Ascii("limosines"), &["limousines"]),
            (dictgen::InsensitiveStr::Ascii("updadated"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("connectd"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("catistrophic"),
                &["catastrophic"],
            ),
            (dictgen::InsensitiveStr::Ascii("prooblem"), &["problem"]),
            (
                dictgen::InsensitiveStr::Ascii("contollled"),
                &["controlled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistmatching"),
                &["mismatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("spagehtti"), &["spaghetti"]),
            (dictgen::InsensitiveStr::Ascii("otherewise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("visulizations"),
                &["visualizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emplies"),
                &["implies", "empties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allcommnads"),
                &["allcommands"],
            ),
            (dictgen::InsensitiveStr::Ascii("realating"), &["relating"]),
            (
                dictgen::InsensitiveStr::Ascii("redistirbutes"),
                &["redistributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empressive"),
                &["impressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifcation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appeciating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("structral"), &["structural"]),
            (
                dictgen::InsensitiveStr::Ascii("synthesizin"),
                &["synthesizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaings"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("feminint"), &["femininity"]),
            (dictgen::InsensitiveStr::Ascii("fedality"), &["fidelity"]),
            (dictgen::InsensitiveStr::Ascii("attracters"), &["attracts"]),
            (dictgen::InsensitiveStr::Ascii("despawing"), &["despawning"]),
            (
                dictgen::InsensitiveStr::Ascii("ruidmentary"),
                &["rudimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoatiations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gruping"), &["grouping"]),
            (
                dictgen::InsensitiveStr::Ascii("chronichles"),
                &["chronicles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chouse"),
                &["choose", "chose", "choux"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificaions"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("magnicifent"),
                &["magnificent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propegated"),
                &["propagated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempalrs"), &["templars"]),
            (dictgen::InsensitiveStr::Ascii("nostirls"), &["nostrils"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installataion"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bordeom"), &["boredom"]),
            (dictgen::InsensitiveStr::Ascii("presumely"), &["presumably"]),
            (dictgen::InsensitiveStr::Ascii("discuessed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("retrievies"), &["retrieves"]),
            (dictgen::InsensitiveStr::Ascii("condiion"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("caefully"), &["carefully"]),
            (dictgen::InsensitiveStr::Ascii("triggern"), &["triggering"]),
            (dictgen::InsensitiveStr::Ascii("lke"), &["like"]),
            (dictgen::InsensitiveStr::Ascii("explin"), &["explain"]),
            (dictgen::InsensitiveStr::Ascii("utillizes"), &["utilizes"]),
            (
                dictgen::InsensitiveStr::Ascii("sucessfual"),
                &["successful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collistions"),
                &["collisions"],
            ),
            (dictgen::InsensitiveStr::Ascii("suposses"), &["supposes"]),
            (
                dictgen::InsensitiveStr::Ascii("interviewr"),
                &["interviewer"],
            ),
            (dictgen::InsensitiveStr::Ascii("wege"), &["wedge"]),
            (dictgen::InsensitiveStr::Ascii("weeekend"), &["weekend"]),
            (
                dictgen::InsensitiveStr::Ascii("conspirace"),
                &["conspiracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("cookoo"), &["cuckoo"]),
            (
                dictgen::InsensitiveStr::Ascii("bootrapper"),
                &["bootstrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("decprecate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("communisits"),
                &["communists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decomposits"),
                &["decomposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differientiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoinded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("reveresed"), &["reversed"]),
            (dictgen::InsensitiveStr::Ascii("aphabet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("mannersims"),
                &["mannerisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stired"),
                &["stored", "stirred"],
            ),
            (dictgen::InsensitiveStr::Ascii("basiclay"), &["basically"]),
            (dictgen::InsensitiveStr::Ascii("relective"), &["reflective"]),
            (dictgen::InsensitiveStr::Ascii("disale"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("configutations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("durin"),
                &["during", "durian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replasement"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("annaying"), &["annoying"]),
            (dictgen::InsensitiveStr::Ascii("intefere"), &["interfere"]),
            (dictgen::InsensitiveStr::Ascii("languares"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("procurin"), &["procuring"]),
            (
                dictgen::InsensitiveStr::Ascii("difused"),
                &["diffused", "defused"],
            ),
            (dictgen::InsensitiveStr::Ascii("internel"), &["internal"]),
            (dictgen::InsensitiveStr::Ascii("tqrget"), &["target"]),
            (
                dictgen::InsensitiveStr::Ascii("descriminates"),
                &["discriminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equalibrium"),
                &["equilibrium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assesement"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("groubdbreaking"),
                &["groundbreaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("reisntall"), &["reinstall"]),
            (
                dictgen::InsensitiveStr::Ascii("uncompletness"),
                &["incompleteness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selecetable"),
                &["selectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("endpoings"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("counterporductive"),
                &["counterproductive"],
            ),
            (dictgen::InsensitiveStr::Ascii("popuation"), &["population"]),
            (
                dictgen::InsensitiveStr::Ascii("associdations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sovereignety"),
                &["sovereignty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("friendhsip"),
                &["friendship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acumulated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("expandes"), &["expands"]),
            (dictgen::InsensitiveStr::Ascii("packaages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("istantly"),
                &["instantly", "distantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("ollipop"), &["lollipop"]),
            (dictgen::InsensitiveStr::Ascii("acceppted"), &["accepted"]),
            (
                dictgen::InsensitiveStr::Ascii("proectors"),
                &["protectors", "projectors", "proctors"],
            ),
            (dictgen::InsensitiveStr::Ascii("reliefes"), &["relieves"]),
            (dictgen::InsensitiveStr::Ascii("embbedding"), &["embedding"]),
            (dictgen::InsensitiveStr::Ascii("intiates"), &["initiates"]),
            (dictgen::InsensitiveStr::Ascii("refroms"), &["reforms"]),
            (
                dictgen::InsensitiveStr::Ascii("collectinos"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comeptition"),
                &["competition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dianostics"),
                &["diagnostics"],
            ),
            (dictgen::InsensitiveStr::Ascii("protlet"), &["portlet"]),
            (dictgen::InsensitiveStr::Ascii("beiginning"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("opptional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("polymomials"),
                &["polynomials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraneterized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosuming"), &["consuming"]),
            (dictgen::InsensitiveStr::Ascii("inex"), &["index"]),
            (dictgen::InsensitiveStr::Ascii("comparign"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("marixst"), &["marxist"]),
            (
                dictgen::InsensitiveStr::Ascii("argiculturally"),
                &["agriculturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("soem"), &["some"]),
            (dictgen::InsensitiveStr::Ascii("disabels"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("successsion"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("cataclism"), &["cataclysm"]),
            (
                dictgen::InsensitiveStr::Ascii("executioness"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("douibled"), &["doubled"]),
            (dictgen::InsensitiveStr::Ascii("arhival"), &["archival"]),
            (dictgen::InsensitiveStr::Ascii("aritst"), &["artist"]),
            (
                dictgen::InsensitiveStr::Ascii("experiemnts"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("feauture"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("accelorator"),
                &["accelerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidenty"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approperiately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("illegsl"), &["illegals"]),
            (dictgen::InsensitiveStr::Ascii("quantitiy"), &["quantity"]),
            (
                dictgen::InsensitiveStr::Ascii("dimands"),
                &["demands", "diamonds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimistisch"),
                &["optimistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("presentors"), &["presents"]),
            (
                dictgen::InsensitiveStr::Ascii("legendaires"),
                &["legendaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpletes"), &["completes"]),
            (dictgen::InsensitiveStr::Ascii("veify"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("initaited"), &["initiated"]),
            (
                dictgen::InsensitiveStr::Ascii("challengin"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescriptionists"),
                &["prescriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("supercede"), &["supersede"]),
            (
                dictgen::InsensitiveStr::Ascii("assigments"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eneumretaion"),
                &["enumeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("fullset"), &["fullest"]),
            (dictgen::InsensitiveStr::Ascii("vally"), &["valley"]),
            (
                dictgen::InsensitiveStr::Ascii("interraccial"),
                &["interracial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traffickin"),
                &["trafficking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructts"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informtaion"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regestered"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenitcates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("artilley"), &["artillery"]),
            (
                dictgen::InsensitiveStr::Ascii("quartenions"),
                &["quaternions"],
            ),
            (dictgen::InsensitiveStr::Ascii("indirct"), &["indirect"]),
            (dictgen::InsensitiveStr::Ascii("remians"), &["remains"]),
            (
                dictgen::InsensitiveStr::Ascii("concatanation"),
                &["concatenation"],
            ),
            (dictgen::InsensitiveStr::Ascii("deprectat"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("destrcutor"),
                &["destructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woul"),
                &["would", "wool", "soul", "foul"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regeressions"),
                &["regressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("kinghts"), &["knights"]),
            (dictgen::InsensitiveStr::Ascii("finanlize"), &["finalize"]),
            (
                dictgen::InsensitiveStr::Ascii("handshakring"),
                &["handshaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("temlating"), &["templating"]),
            (
                dictgen::InsensitiveStr::Ascii("regarged"),
                &["regarded", "recharged"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrrected"), &["corrected"]),
            (
                dictgen::InsensitiveStr::Ascii("alhpabeticaly"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("condfigure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("conlcusions"),
                &["conclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("calucated"), &["calculated"]),
            (
                dictgen::InsensitiveStr::Ascii("maitainers"),
                &["maintainers"],
            ),
            (dictgen::InsensitiveStr::Ascii("arragned"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("mounment"), &["monument"]),
            (dictgen::InsensitiveStr::Ascii("uptreamed"), &["upstreamed"]),
            (
                dictgen::InsensitiveStr::Ascii("substiture"),
                &["substitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("examplees"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("manoeuvrin"),
                &["manoeuvring"],
            ),
            (dictgen::InsensitiveStr::Ascii("mantainer"), &["maintainer"]),
            (dictgen::InsensitiveStr::Ascii("incldue"), &["include"]),
            (dictgen::InsensitiveStr::Ascii("elimnated"), &["eliminated"]),
            (dictgen::InsensitiveStr::Ascii("quaterly"), &["quarterly"]),
            (
                dictgen::InsensitiveStr::Ascii("antagonistc"),
                &["antagonistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("enocdings"), &["encodings"]),
            (dictgen::InsensitiveStr::Ascii("losen"), &["loosen"]),
            (
                dictgen::InsensitiveStr::Ascii("vitual"),
                &["virtual", "visual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilializing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuction"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfecction"),
                &["perfection"],
            ),
            (dictgen::InsensitiveStr::Ascii("institue"), &["institute"]),
            (
                dictgen::InsensitiveStr::Ascii("visualizataion"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("baceuse"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("evesdropper"),
                &["eavesdropper"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanp"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("organizare"), &["organizer"]),
            (
                dictgen::InsensitiveStr::Ascii("canabelysed"),
                &["cannibalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tyhe"),
                &["they", "the", "type"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("victumized"),
                &["victimized"],
            ),
            (dictgen::InsensitiveStr::Ascii("easiyl"), &["easily"]),
            (
                dictgen::InsensitiveStr::Ascii("fondamentals"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoatations"),
                &["annotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("franchies"), &["franchise"]),
            (dictgen::InsensitiveStr::Ascii("tmeselves"), &["themselves"]),
            (
                dictgen::InsensitiveStr::Ascii("friendlines"),
                &["friendlies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjustmants"),
                &["adjustments"],
            ),
            (dictgen::InsensitiveStr::Ascii("comipling"), &["compiling"]),
            (
                dictgen::InsensitiveStr::Ascii("chckpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranzformer"),
                &["transformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("pigoens"), &["pigeons"]),
            (
                dictgen::InsensitiveStr::Ascii("anonmyously"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comprimises"),
                &["compromises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ammendments"),
                &["amendments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calclating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cryptograhic"),
                &["cryptographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("futurustic"),
                &["futuristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("retify"), &["rectify"]),
            (dictgen::InsensitiveStr::Ascii("bayonent"), &["bayonet"]),
            (dictgen::InsensitiveStr::Ascii("temepst"), &["tempest"]),
            (dictgen::InsensitiveStr::Ascii("flexbile"), &["flexible"]),
            (dictgen::InsensitiveStr::Ascii("loacales"), &["locales"]),
            (dictgen::InsensitiveStr::Ascii("sctripted"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("attritube"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("correxponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arthrimetic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("furhest"), &["furthest"]),
            (
                dictgen::InsensitiveStr::Ascii("contrained"),
                &["contained", "constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recontructors"),
                &["reconstructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanteniously"),
                &["instantaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assumtpion"),
                &["assumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrinsicaly"),
                &["intrinsically"],
            ),
            (dictgen::InsensitiveStr::Ascii("catious"), &["cautious"]),
            (dictgen::InsensitiveStr::Ascii("assserts"), &["asserts"]),
            (dictgen::InsensitiveStr::Ascii("greyscalce"), &["greyscale"]),
            (
                dictgen::InsensitiveStr::Ascii("resturctures"),
                &["restructures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("granulatiry"),
                &["granularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("poduces"), &["produces"]),
            (dictgen::InsensitiveStr::Ascii("eanables"), &["enables"]),
            (
                dictgen::InsensitiveStr::Ascii("depdennces"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlcusion"), &["exclusion"]),
            (
                dictgen::InsensitiveStr::Ascii("refererence"),
                &["reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("granjure"), &["grandeur"]),
            (
                dictgen::InsensitiveStr::Ascii("homecomming"),
                &["homecoming"],
            ),
            (dictgen::InsensitiveStr::Ascii("anmore"), &["anymore"]),
            (dictgen::InsensitiveStr::Ascii("prescibed"), &["prescribed"]),
            (dictgen::InsensitiveStr::Ascii("exceptons"), &["exceptions"]),
            (dictgen::InsensitiveStr::Ascii("beligum"), &["belgium"]),
            (
                dictgen::InsensitiveStr::Ascii("visualzations"),
                &["visualizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referencial"),
                &["referential"],
            ),
            (dictgen::InsensitiveStr::Ascii("goodlcuk"), &["goodluck"]),
            (dictgen::InsensitiveStr::Ascii("widder"), &["wider"]),
            (dictgen::InsensitiveStr::Ascii("retures"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("postproccessing"),
                &["postprocessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("adddresses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("executtions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("widgte"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("oridinate"), &["ordinate"]),
            (dictgen::InsensitiveStr::Ascii("guyser"), &["geyser"]),
            (
                dictgen::InsensitiveStr::Ascii("shecking"),
                &["checking", "shucking"],
            ),
            (dictgen::InsensitiveStr::Ascii("deconder"), &["decoder"]),
            (
                dictgen::InsensitiveStr::Ascii("agracultural"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silouhette"),
                &["silhouette"],
            ),
            (dictgen::InsensitiveStr::Ascii("supporre"), &["supporters"]),
            (dictgen::InsensitiveStr::Ascii("espisodic"), &["episodic"]),
            (dictgen::InsensitiveStr::Ascii("corrision"), &["corrosion"]),
            (
                dictgen::InsensitiveStr::Ascii("delievries"),
                &["deliveries"],
            ),
            (dictgen::InsensitiveStr::Ascii("advizable"), &["advisable"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnotics"),
                &["diagnostic", "diagnostics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corressponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoitators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behavoiral"),
                &["behavioral"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofthen"), &["often"]),
            (dictgen::InsensitiveStr::Ascii("iterrated"), &["iterated"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiation"),
                &["initialization", "initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("garuntee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("arithmentic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("protectiv"), &["protective"]),
            (dictgen::InsensitiveStr::Ascii("mariage"), &["marriage"]),
            (dictgen::InsensitiveStr::Ascii("paitients"), &["patients"]),
            (dictgen::InsensitiveStr::Ascii("librery"), &["library"]),
            (dictgen::InsensitiveStr::Ascii("sincerley"), &["sincerely"]),
            (
                dictgen::InsensitiveStr::Ascii("organistions"),
                &["organisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimential"),
                &["experimental", "experiential"],
            ),
            (dictgen::InsensitiveStr::Ascii("liasing"), &["liaising"]),
            (
                dictgen::InsensitiveStr::Ascii("enthicities"),
                &["ethnicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seachers"),
                &["searchers", "teachers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyts"),
                &["analyst", "analysts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contoling"),
                &["controlling", "consoling", "condoling"],
            ),
            (dictgen::InsensitiveStr::Ascii("unviersal"), &["universal"]),
            (dictgen::InsensitiveStr::Ascii("canconical"), &["canonical"]),
            (
                dictgen::InsensitiveStr::Ascii("referencially"),
                &["referentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occurrencies"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("liminted"), &["limited"]),
            (dictgen::InsensitiveStr::Ascii("statisfy"), &["satisfy"]),
            (
                dictgen::InsensitiveStr::Ascii("pitchfolks"),
                &["pitchforks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retoractively"),
                &["retroactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmisive"),
                &["transmissive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scource"),
                &["source", "scouse"],
            ),
            (dictgen::InsensitiveStr::Ascii("ontrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("generattors"),
                &["generators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comemmorates"),
                &["commemorates"],
            ),
            (dictgen::InsensitiveStr::Ascii("defalt"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("opf"), &["of"]),
            (
                dictgen::InsensitiveStr::Ascii("expewriment"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subexpersion"),
                &["subexpression"],
            ),
            (dictgen::InsensitiveStr::Ascii("luanched"), &["launched"]),
            (
                dictgen::InsensitiveStr::Ascii("decremeneted"),
                &["decremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("attibures"), &["attributes"]),
            (dictgen::InsensitiveStr::Ascii("pithcers"), &["pitchers"]),
            (dictgen::InsensitiveStr::Ascii("combier"), &["combiner"]),
            (dictgen::InsensitiveStr::Ascii("depoyed"), &["deployed"]),
            (dictgen::InsensitiveStr::Ascii("bithday"), &["birthday"]),
            (
                dictgen::InsensitiveStr::Ascii("descripitors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmitor"),
                &["transistor"],
            ),
            (dictgen::InsensitiveStr::Ascii("dessign"), &["design"]),
            (dictgen::InsensitiveStr::Ascii("itseltf"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("calclulator"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neccisarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("elgantly"), &["elegantly"]),
            (dictgen::InsensitiveStr::Ascii("defaoults"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("sastisfaction"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininised"), &["minimised"]),
            (
                dictgen::InsensitiveStr::Ascii("accordiingly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referncences"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresolveable"),
                &["unresolvable"],
            ),
            (dictgen::InsensitiveStr::Ascii("lerans"), &["learns"]),
            (dictgen::InsensitiveStr::Ascii("addmit"), &["admit"]),
            (
                dictgen::InsensitiveStr::Ascii("interefacing"),
                &["interfacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proyected"),
                &["projected", "protected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraphraze"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anialates"),
                &["annihilates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corruptiuon"),
                &["corruption"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhytmic"), &["rhythmic"]),
            (dictgen::InsensitiveStr::Ascii("itervals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("motherborads"),
                &["motherboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microships"),
                &["microchips"],
            ),
            (dictgen::InsensitiveStr::Ascii("quatation"), &["quotation"]),
            (
                dictgen::InsensitiveStr::Ascii("recognization"),
                &["recognition"],
            ),
            (dictgen::InsensitiveStr::Ascii("signater"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("equaivalence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("severite"), &["severity"]),
            (dictgen::InsensitiveStr::Ascii("corrrects"), &["corrects"]),
            (dictgen::InsensitiveStr::Ascii("bradfrod"), &["bradford"]),
            (
                dictgen::InsensitiveStr::Ascii("inidcations"),
                &["indications"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbitten"), &["forbidden"]),
            (
                dictgen::InsensitiveStr::Ascii("misconcpetion"),
                &["misconceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("trnasmits"), &["transmits"]),
            (dictgen::InsensitiveStr::Ascii("penaly"), &["penalty"]),
            (dictgen::InsensitiveStr::Ascii("remmember"), &["remember"]),
            (dictgen::InsensitiveStr::Ascii("synthethic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("expandas"), &["expands"]),
            (
                dictgen::InsensitiveStr::Ascii("restricive"),
                &["restrictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("anactive"), &["inactive"]),
            (
                dictgen::InsensitiveStr::Ascii("culminatin"),
                &["culminating", "culmination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiantorial"),
                &["combinatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("covenience"),
                &["convenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("toally"), &["totally"]),
            (dictgen::InsensitiveStr::Ascii("stragely"), &["strangely"]),
            (
                dictgen::InsensitiveStr::Ascii("metalness"),
                &["metallicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("preppented"), &["prepended"]),
            (dictgen::InsensitiveStr::Ascii("carreers"), &["careers"]),
            (dictgen::InsensitiveStr::Ascii("indcation"), &["indication"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanentelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("ankshis"), &["anxious"]),
            (dictgen::InsensitiveStr::Ascii("embedde"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("statememt"), &["statement"]),
            (dictgen::InsensitiveStr::Ascii("asymetry"), &["asymmetry"]),
            (dictgen::InsensitiveStr::Ascii("tcahces"), &["caches"]),
            (
                dictgen::InsensitiveStr::Ascii("excellance"),
                &["excellence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavialable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spcifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwowind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("accepected"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("niear"), &["near"]),
            (dictgen::InsensitiveStr::Ascii("conlusion"), &["conclusion"]),
            (dictgen::InsensitiveStr::Ascii("sicne"), &["since"]),
            (
                dictgen::InsensitiveStr::Ascii("prepetrated"),
                &["perpetrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electromegnetic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delaraction"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigurette"), &["cigarette"]),
            (
                dictgen::InsensitiveStr::Ascii("clearstories"),
                &["clerestories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimenal"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attributet"),
                &["attribute", "attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("setitng"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("coutned"), &["counted"]),
            (
                dictgen::InsensitiveStr::Ascii("foregroung"),
                &["foreground"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundementals"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astiimation"),
                &["estimation"],
            ),
            (dictgen::InsensitiveStr::Ascii("napoloen"), &["napoleon"]),
            (
                dictgen::InsensitiveStr::Ascii("replacament"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equaliteral"),
                &["equilateral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirtuiting"),
                &["circuiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("expaining"), &["explaining"]),
            (dictgen::InsensitiveStr::Ascii("collidies"), &["collides"]),
            (dictgen::InsensitiveStr::Ascii("includuing"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("presumpton"),
                &["presumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("ruslted"), &["rustled"]),
            (
                dictgen::InsensitiveStr::Ascii("accumilates"),
                &["accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("daty"), &["data", "date"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregrator"),
                &["aggregator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inetrrupts"),
                &["interrupts"],
            ),
            (dictgen::InsensitiveStr::Ascii("immensly"), &["immensely"]),
            (dictgen::InsensitiveStr::Ascii("corecctly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("suposable"), &["supposable"]),
            (
                dictgen::InsensitiveStr::Ascii("oyurselves"),
                &["yourselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("spript"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("seaching"),
                &["searching", "reaching", "teaching"],
            ),
            (dictgen::InsensitiveStr::Ascii("milyew"), &["milieu"]),
            (
                dictgen::InsensitiveStr::Ascii("directin"),
                &["directions", "directing", "direction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allowd"),
                &["allowed", "allow", "allows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifiction"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prevencion"),
                &["prevention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metrapolis"),
                &["metropolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("ugrading"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("biappicative"),
                &["biapplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("excact"), &["exact"]),
            (
                dictgen::InsensitiveStr::Ascii("ditactorship"),
                &["dictatorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("serios"), &["serious"]),
            (
                dictgen::InsensitiveStr::Ascii("artificiell"),
                &["artificial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initliasing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sumarisation"),
                &["summarisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entchanting"),
                &["enchanting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ungeneralizeable"),
                &["ungeneralizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("raoming"), &["roaming"]),
            (
                dictgen::InsensitiveStr::Ascii("scientests"),
                &["scientists"],
            ),
            (dictgen::InsensitiveStr::Ascii("sillicon"), &["silicon"]),
            (dictgen::InsensitiveStr::Ascii("alignement"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("interruptt"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("calucaltion"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomrnace"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("statisfied"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("foramts"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("insensity"),
                &["insensitivity", "intensity"],
            ),
            (dictgen::InsensitiveStr::Ascii("ailenates"), &["alienates"]),
            (dictgen::InsensitiveStr::Ascii("exceptoin"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("traveersal"), &["traversal"]),
            (dictgen::InsensitiveStr::Ascii("ukrainias"), &["ukrainians"]),
            (dictgen::InsensitiveStr::Ascii("enfocing"), &["enforcing"]),
            (dictgen::InsensitiveStr::Ascii("numbere"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("sructs"), &["structs"]),
            (dictgen::InsensitiveStr::Ascii("unregiter"), &["unregister"]),
            (
                dictgen::InsensitiveStr::Ascii("factization"),
                &["factorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstandigs"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descvription"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsetes"), &["offsets"]),
            (dictgen::InsensitiveStr::Ascii("regisster"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("fragementation"),
                &["fragmentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimage"), &["minigame"]),
            (dictgen::InsensitiveStr::Ascii("optain"), &["obtain"]),
            (
                dictgen::InsensitiveStr::Ascii("autoformating"),
                &["autoformatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("meachnism"), &["mechanism"]),
            (dictgen::InsensitiveStr::Ascii("wilfull"), &["wilful"]),
            (dictgen::InsensitiveStr::Ascii("aneroxic"), &["anorexic"]),
            (
                dictgen::InsensitiveStr::Ascii("envoke"),
                &["evoke", "invoke"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("liturature"),
                &["literature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprixamated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ariving"), &["arriving"]),
            (
                dictgen::InsensitiveStr::Ascii("contingecy"),
                &["contingency"],
            ),
            (dictgen::InsensitiveStr::Ascii("defauly"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("mutches"), &["matches"]),
            (dictgen::InsensitiveStr::Ascii("vreity"), &["variety"]),
            (dictgen::InsensitiveStr::Ascii("provdied"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("maniputed"),
                &["manipulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("totalitara"),
                &["totalitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitiation"),
                &["delimitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("oeprator"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("queueud"), &["queued"]),
            (
                dictgen::InsensitiveStr::Ascii("transsexal"),
                &["transsexual"],
            ),
            (dictgen::InsensitiveStr::Ascii("sergeat"), &["sergeant"]),
            (
                dictgen::InsensitiveStr::Ascii("supsending"),
                &["suspending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acendancey"),
                &["ascendancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchamarked"),
                &["benchmarked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specfiying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondants"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfamilar"), &["unfamiliar"]),
            (dictgen::InsensitiveStr::Ascii("recurssing"), &["recursing"]),
            (
                dictgen::InsensitiveStr::Ascii("excepcional"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("premiare"), &["premier"]),
            (
                dictgen::InsensitiveStr::Ascii("aassignment"),
                &["assignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conlcusion"),
                &["conclusion"],
            ),
            (dictgen::InsensitiveStr::Ascii("shashes"), &["slashes"]),
            (dictgen::InsensitiveStr::Ascii("buidler"), &["builder"]),
            (
                dictgen::InsensitiveStr::Ascii("engingeering"),
                &["engineering"],
            ),
            (dictgen::InsensitiveStr::Ascii("persewer"), &["pursuer"]),
            (
                dictgen::InsensitiveStr::Ascii("instancec"),
                &["instance", "instances", "instanced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("receipes"),
                &["recipes", "receipts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assimption"),
                &["assumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosparity"),
                &["prosperity"],
            ),
            (dictgen::InsensitiveStr::Ascii("ongewild"), &["gonewild"]),
            (
                dictgen::InsensitiveStr::Ascii("guantanano"),
                &["guantanamo"],
            ),
            (dictgen::InsensitiveStr::Ascii("virbation"), &["vibration"]),
            (dictgen::InsensitiveStr::Ascii("prind"), &["print"]),
            (
                dictgen::InsensitiveStr::Ascii("restrcited"),
                &["restricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("workig"), &["working"]),
            (dictgen::InsensitiveStr::Ascii("mustated"), &["mutated"]),
            (dictgen::InsensitiveStr::Ascii("taxomonmy"), &["taxonomy"]),
            (
                dictgen::InsensitiveStr::Ascii("constituit"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("threeof"), &["thereof"]),
            (dictgen::InsensitiveStr::Ascii("prarent"), &["parent"]),
            (
                dictgen::InsensitiveStr::Ascii("lik"),
                &["like", "lick", "link"],
            ),
            (dictgen::InsensitiveStr::Ascii("assesment"), &["assessment"]),
            (
                dictgen::InsensitiveStr::Ascii("multilplayer"),
                &["multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("starngers"), &["strangers"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfifying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffring"), &["buffering"]),
            (
                dictgen::InsensitiveStr::Ascii("extensiosn"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validadors"),
                &["validators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weakess"),
                &["weakness", "weakest"],
            ),
            (dictgen::InsensitiveStr::Ascii("donig"), &["doing"]),
            (dictgen::InsensitiveStr::Ascii("worlwide"), &["worldwide"]),
            (
                dictgen::InsensitiveStr::Ascii("irreverant"),
                &["irrelevant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffirence"),
                &["difference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressivelly"),
                &["progressively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrairy"),
                &["arbitrarily", "arbitrary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholorship"),
                &["scholarship"],
            ),
            (dictgen::InsensitiveStr::Ascii("disagreeed"), &["disagreed"]),
            (dictgen::InsensitiveStr::Ascii("orherwise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionalble"),
                &["questionable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lambas"),
                &["lambdas", "lambs", "lambast", "lambaste"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corporativos"),
                &["corporations"],
            ),
            (dictgen::InsensitiveStr::Ascii("documnted"), &["documented"]),
            (dictgen::InsensitiveStr::Ascii("remplaces"), &["replaces"]),
            (
                dictgen::InsensitiveStr::Ascii("supplaments"),
                &["supplements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colonianism"),
                &["colonialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("crutchetts"), &["crutches"]),
            (dictgen::InsensitiveStr::Ascii("accounded"), &["accounted"]),
            (dictgen::InsensitiveStr::Ascii("cjange"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("posistioning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normlaized"),
                &["normalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inforrmational"),
                &["informational"],
            ),
            (dictgen::InsensitiveStr::Ascii("duplates"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("perrogative"),
                &["prerogative"],
            ),
            (dictgen::InsensitiveStr::Ascii("pancaeks"), &["pancakes"]),
            (dictgen::InsensitiveStr::Ascii("duobling"), &["doubling"]),
            (
                dictgen::InsensitiveStr::Ascii("developpers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscarrage"),
                &["miscarriage"],
            ),
            (dictgen::InsensitiveStr::Ascii("currenntly"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("participatd"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("geurrilla"), &["guerrilla"]),
            (
                dictgen::InsensitiveStr::Ascii("consultata"),
                &["consultant"],
            ),
            (dictgen::InsensitiveStr::Ascii("fireing"), &["firing"]),
            (dictgen::InsensitiveStr::Ascii("uner"), &["under"]),
            (dictgen::InsensitiveStr::Ascii("tokne"), &["token"]),
            (
                dictgen::InsensitiveStr::Ascii("recontructor"),
                &["reconstructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("critican"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("commpliant"), &["compliant"]),
            (
                dictgen::InsensitiveStr::Ascii("priorisin"),
                &["prioritising"],
            ),
            (dictgen::InsensitiveStr::Ascii("docoments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboorhoud"),
                &["neighbourhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfered"), &["offered"]),
            (dictgen::InsensitiveStr::Ascii("swithches"), &["switches"]),
            (dictgen::InsensitiveStr::Ascii("valudates"), &["validates"]),
            (dictgen::InsensitiveStr::Ascii("nautulis"), &["nautilus"]),
            (
                dictgen::InsensitiveStr::Ascii("intrinically"),
                &["intrinsically"],
            ),
            (dictgen::InsensitiveStr::Ascii("contorled"), &["controlled"]),
            (
                dictgen::InsensitiveStr::Ascii("intiialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjacenet"), &["adjacent"]),
            (dictgen::InsensitiveStr::Ascii("atcions"), &["actions"]),
            (dictgen::InsensitiveStr::Ascii("sunsday"), &["sunday"]),
            (dictgen::InsensitiveStr::Ascii("dynamis"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("polical"), &["political"]),
            (
                dictgen::InsensitiveStr::Ascii("photogrpahed"),
                &["photographed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("warpped"),
                &["warped", "wrapped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archietecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitifers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssessed"), &["assessed"]),
            (dictgen::InsensitiveStr::Ascii("licens"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("variatoins"),
                &["variations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allcoateng"),
                &["allocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangin"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("juggernuat"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrident"),
                &["overridden"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experamentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoed"), &["supposed"]),
            (dictgen::InsensitiveStr::Ascii("retrvieves"), &["retrieves"]),
            (
                dictgen::InsensitiveStr::Ascii("equivilance"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("handeling"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("guaruantees"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivalancy"),
                &["equivalency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconisistent"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comnpletely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gaurantteeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("marvelos"), &["marvelous"]),
            (dictgen::InsensitiveStr::Ascii("funtions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("indefine"),
                &["indefinite", "undefine"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqular"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("checheckpoit"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resturcturing"),
                &["restructuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("atlesat"), &["atleast"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphorial"),
                &["metaphorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("excetions"), &["exceptions"]),
            (dictgen::InsensitiveStr::Ascii("feasbile"), &["feasible"]),
            (
                dictgen::InsensitiveStr::Ascii("annotaions"),
                &["annotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcteions"),
                &["corrections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sincronus"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeast"), &["repeats"]),
            (
                dictgen::InsensitiveStr::Ascii("sprties"),
                &["sprites", "sorties"],
            ),
            (dictgen::InsensitiveStr::Ascii("appaear"), &["appear"]),
            (dictgen::InsensitiveStr::Ascii("propoagate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("revrsed"), &["reversed"]),
            (dictgen::InsensitiveStr::Ascii("invisisble"), &["invisible"]),
            (dictgen::InsensitiveStr::Ascii("attantion"), &["attention"]),
            (
                dictgen::InsensitiveStr::Ascii("countermeausure"),
                &["countermeasure"],
            ),
            (dictgen::InsensitiveStr::Ascii("scenarion"), &["scenario"]),
            (
                dictgen::InsensitiveStr::Ascii("conservitive"),
                &["conservative"],
            ),
            (dictgen::InsensitiveStr::Ascii("silienced"), &["silenced"]),
            (
                dictgen::InsensitiveStr::Ascii("uninstallng"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("publiched"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("sublcass"), &["subclass"]),
            (dictgen::InsensitiveStr::Ascii("habilities"), &["abilities"]),
            (
                dictgen::InsensitiveStr::Ascii("unenployment"),
                &["unemployment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completent"),
                &["complement"],
            ),
            (dictgen::InsensitiveStr::Ascii("wressels"), &["wrestles"]),
            (dictgen::InsensitiveStr::Ascii("languace"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("renewabels"),
                &["renewables"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoprt"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("omiitted"), &["omitted"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhod"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metodologies"),
                &["methodologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordng"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("unfilpped"), &["unflipped"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalistista"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("releationship"),
                &["relationship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horitontal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("rocessors"), &["processors"]),
            (
                dictgen::InsensitiveStr::Ascii("scrrenshots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("obvisouly"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("simplyfication"),
                &["simplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggragator"),
                &["aggregator"],
            ),
            (dictgen::InsensitiveStr::Ascii("containrs"), &["containers"]),
            (dictgen::InsensitiveStr::Ascii("testiclies"), &["testicle"]),
            (dictgen::InsensitiveStr::Ascii("hight"), &["height", "high"]),
            (
                dictgen::InsensitiveStr::Ascii("senaphores"),
                &["semaphores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermedite"),
                &["intermediate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informartion"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumferance"),
                &["circumference"],
            ),
            (dictgen::InsensitiveStr::Ascii("wou"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("superham"), &["superhuman"]),
            (
                dictgen::InsensitiveStr::Ascii("accelearion"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("excecutive"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("maintainig"),
                &["maintaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("variabla"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("oppinions"), &["opinions"]),
            (dictgen::InsensitiveStr::Ascii("latnern"), &["lantern"]),
            (dictgen::InsensitiveStr::Ascii("statastic"), &["statistic"]),
            (dictgen::InsensitiveStr::Ascii("jurneyed"), &["journeyed"]),
            (
                dictgen::InsensitiveStr::Ascii("downloadbig"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aunthenticated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrantys"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elamentary"),
                &["elementary"],
            ),
            (dictgen::InsensitiveStr::Ascii("inititated"), &["initiated"]),
            (
                dictgen::InsensitiveStr::Ascii("ridicoulusness"),
                &["ridiculousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("incurr"), &["incur"]),
            (
                dictgen::InsensitiveStr::Ascii("exted"),
                &["extend", "exited", "texted"],
            ),
            (dictgen::InsensitiveStr::Ascii("versinoed"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("shadhow"), &["shadow"]),
            (
                dictgen::InsensitiveStr::Ascii("separatron"),
                &["separation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deparments"),
                &["departments"],
            ),
            (dictgen::InsensitiveStr::Ascii("defeintly"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessully"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mysoganistic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuportable"),
                &["unsupportable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomidation"),
                &["accommodation"],
            ),
            (dictgen::InsensitiveStr::Ascii("upports"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("cleasning"), &["cleansing"]),
            (
                dictgen::InsensitiveStr::Ascii("attriubtions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("delcare"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("immitate"), &["imitate"]),
            (dictgen::InsensitiveStr::Ascii("effecs"), &["effects"]),
            (dictgen::InsensitiveStr::Ascii("succesion"), &["succession"]),
            (dictgen::InsensitiveStr::Ascii("preimer"), &["premier"]),
            (dictgen::InsensitiveStr::Ascii("graduaste"), &["graduates"]),
            (dictgen::InsensitiveStr::Ascii("ipmrove"), &["improve"]),
            (
                dictgen::InsensitiveStr::Ascii("undelering"),
                &["underlying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acent"),
                &["ascent", "accent"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialecs"), &["dialects"]),
            (dictgen::InsensitiveStr::Ascii("arquivers"), &["archivers"]),
            (dictgen::InsensitiveStr::Ascii("indentin"), &["indent"]),
            (dictgen::InsensitiveStr::Ascii("crossifre"), &["crossfire"]),
            (
                dictgen::InsensitiveStr::Ascii("credentils"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("distrupts"), &["distrust"]),
            (
                dictgen::InsensitiveStr::Ascii("historgrams"),
                &["histograms"],
            ),
            (dictgen::InsensitiveStr::Ascii("starnger"), &["stranger"]),
            (
                dictgen::InsensitiveStr::Ascii("altnerately"),
                &["alternately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuatbles"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summerises"),
                &["summarises"],
            ),
            (dictgen::InsensitiveStr::Ascii("derieved"), &["derived"]),
            (dictgen::InsensitiveStr::Ascii("eleases"), &["releases"]),
            (
                dictgen::InsensitiveStr::Ascii("interceptons"),
                &["interceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pittaburgh"),
                &["pittsburgh"],
            ),
            (dictgen::InsensitiveStr::Ascii("instanecs"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("assinee"), &["assignee"]),
            (dictgen::InsensitiveStr::Ascii("exampe"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("supervisar"),
                &["supervisor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understsand"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininistrators"),
                &["administrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaors"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ressurection"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("velidate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("cumbursome"),
                &["cumbersome"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotaited"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bundele"), &["bundle"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("kookoos"), &["cuckoos"]),
            (dictgen::InsensitiveStr::Ascii("skurge"), &["scourge"]),
            (dictgen::InsensitiveStr::Ascii("abruplty"), &["abruptly"]),
            (
                dictgen::InsensitiveStr::Ascii("manfesting"),
                &["manifesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("genuin"), &["genuine"]),
            (dictgen::InsensitiveStr::Ascii("benhgazi"), &["benghazi"]),
            (dictgen::InsensitiveStr::Ascii("bigorty"), &["bigotry"]),
            (
                dictgen::InsensitiveStr::Ascii("extimation"),
                &["estimation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cojunction"),
                &["conjunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("ediing"), &["editing"]),
            (
                dictgen::InsensitiveStr::Ascii("micropones"),
                &["microphones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instituions"),
                &["institutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpracticality"),
                &["impracticality"],
            ),
            (dictgen::InsensitiveStr::Ascii("comitish"), &["commitish"]),
            (dictgen::InsensitiveStr::Ascii("angshusly"), &["anxiously"]),
            (dictgen::InsensitiveStr::Ascii("establis"), &["establish"]),
            (dictgen::InsensitiveStr::Ascii("impartirla"), &["impartial"]),
            (dictgen::InsensitiveStr::Ascii("resoution"), &["resolution"]),
            (
                dictgen::InsensitiveStr::Ascii("appericate"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("obectives"), &["objectives"]),
            (dictgen::InsensitiveStr::Ascii("paladinlst"), &["paladins"]),
            (dictgen::InsensitiveStr::Ascii("repalcing"), &["replacing"]),
            (
                dictgen::InsensitiveStr::Ascii("quiestionnaire"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("preserrved"), &["preserved"]),
            (
                dictgen::InsensitiveStr::Ascii("internilized"),
                &["internalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifiication"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moderatedly"),
                &["moderately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stiffneing"),
                &["stiffening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congradulate"),
                &["congratulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("underates"), &["underrates"]),
            (
                dictgen::InsensitiveStr::Ascii("constituinte"),
                &["constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyposesis"), &["hypothesis"]),
            (
                dictgen::InsensitiveStr::Ascii("virtaulized"),
                &["virtualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administrar"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("accurring"), &["occurring"]),
            (dictgen::InsensitiveStr::Ascii("amrpit"), &["armpit"]),
            (dictgen::InsensitiveStr::Ascii("currennts"), &["currents"]),
            (dictgen::InsensitiveStr::Ascii("vengenace"), &["vengeance"]),
            (dictgen::InsensitiveStr::Ascii("shortern"), &["shorten"]),
            (
                dictgen::InsensitiveStr::Ascii("triagulate"),
                &["triangulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philosopers"),
                &["philosophers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refere"),
                &["refer", "referee"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instructivo"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyrstals"), &["crystals"]),
            (dictgen::InsensitiveStr::Ascii("visibale"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("logaritmic"),
                &["logarithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorparating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacheed"), &["cached"]),
            (
                dictgen::InsensitiveStr::Ascii("severirities"),
                &["severities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadaquete"),
                &["inadequate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unathenticated"),
                &["unauthenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeetable"),
                &["unbeatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforecing"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("bottons"),
                &["bottoms", "buttons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoroithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("repressin"), &["repression"]),
            (dictgen::InsensitiveStr::Ascii("scarifice"), &["sacrifice"]),
            (dictgen::InsensitiveStr::Ascii("proprety"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("incroporates"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("realistc"), &["realistic"]),
            (
                dictgen::InsensitiveStr::Ascii("convesation"),
                &["conversation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoitations"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("writhin"), &["writhing"]),
            (
                dictgen::InsensitiveStr::Ascii("presearvations"),
                &["preservations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathwes"), &["mathews"]),
            (
                dictgen::InsensitiveStr::Ascii("continants"),
                &["continents"],
            ),
            (dictgen::InsensitiveStr::Ascii("conent"), &["content"]),
            (
                dictgen::InsensitiveStr::Ascii("distribuer"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuarations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analitical"),
                &["analytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("respectifs"), &["respects"]),
            (
                dictgen::InsensitiveStr::Ascii("acelerators"),
                &["accelerators"],
            ),
            (dictgen::InsensitiveStr::Ascii("reliabe"), &["reliable"]),
            (
                dictgen::InsensitiveStr::Ascii("psydonymity"),
                &["pseudonymity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assertation"),
                &["assertion"],
            ),
            (dictgen::InsensitiveStr::Ascii("laguange"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("excellenet"),
                &["excellence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermisson"),
                &["intermission"],
            ),
            (dictgen::InsensitiveStr::Ascii("componemt"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensinalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delevopments"),
                &["developments"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingeniuty"), &["ingenuity"]),
            (dictgen::InsensitiveStr::Ascii("valentins"), &["valentines"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaization"),
                &["generalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uresponsive"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("deine"), &["define"]),
            (
                dictgen::InsensitiveStr::Ascii("virually"),
                &["virtually", "visually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconenction"),
                &["disconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performacne"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("snycing"), &["syncing"]),
            (dictgen::InsensitiveStr::Ascii("falshed"), &["flashed"]),
            (dictgen::InsensitiveStr::Ascii("edeycat"), &["etiquette"]),
            (dictgen::InsensitiveStr::Ascii("orginials"), &["originals"]),
            (dictgen::InsensitiveStr::Ascii("enbale"), &["enable"]),
            (dictgen::InsensitiveStr::Ascii("tarffic"), &["traffic"]),
            (dictgen::InsensitiveStr::Ascii("reaarange"), &["rearrange"]),
            (
                dictgen::InsensitiveStr::Ascii("highpander"),
                &["highlander"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pivacy"),
                &["privacy", "piracy"],
            ),
            (dictgen::InsensitiveStr::Ascii("adressses"), &["addresses"]),
            (dictgen::InsensitiveStr::Ascii("fractiom"), &["fraction"]),
            (
                dictgen::InsensitiveStr::Ascii("alternaters"),
                &["alternators"],
            ),
            (dictgen::InsensitiveStr::Ascii("recurison"), &["recursion"]),
            (
                dictgen::InsensitiveStr::Ascii("comarators"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("heavely"), &["heavily"]),
            (dictgen::InsensitiveStr::Ascii("singels"), &["singles"]),
            (dictgen::InsensitiveStr::Ascii("acuires"), &["acquires"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalitys"),
                &["nationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("melinneums"), &["millennia"]),
            (
                dictgen::InsensitiveStr::Ascii("ackward"),
                &["awkward", "backward"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operatoes"),
                &["operates", "operators"],
            ),
            (dictgen::InsensitiveStr::Ascii("maddness"), &["madness"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnactional"),
                &["transactional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degeneraging"),
                &["degenerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("misake"), &["mistake"]),
            (dictgen::InsensitiveStr::Ascii("checksms"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("dfinitions"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("geenral"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("accepptor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("swtichers"), &["switchers"]),
            (dictgen::InsensitiveStr::Ascii("loign"), &["login"]),
            (
                dictgen::InsensitiveStr::Ascii("recconecctions"),
                &["reconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("euphorisch"), &["euphoric"]),
            (
                dictgen::InsensitiveStr::Ascii("countertrap"),
                &["counterpart"],
            ),
            (dictgen::InsensitiveStr::Ascii("disablle"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("attractifs"), &["attracts"]),
            (dictgen::InsensitiveStr::Ascii("reates"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("weith"),
                &["weight", "weigh", "with", "width"],
            ),
            (dictgen::InsensitiveStr::Ascii("partision"), &["partition"]),
            (
                dictgen::InsensitiveStr::Ascii("syntesising"),
                &["synthesising"],
            ),
            (dictgen::InsensitiveStr::Ascii("disirably"), &["desirably"]),
            (
                dictgen::InsensitiveStr::Ascii("specfifying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occaisional"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conferrencing"),
                &["conferencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("leaglization"),
                &["legalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceeed"), &["exceed"]),
            (dictgen::InsensitiveStr::Ascii("affeccting"), &["affecting"]),
            (dictgen::InsensitiveStr::Ascii("sievin"), &["sieving"]),
            (dictgen::InsensitiveStr::Ascii("libraryes"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("converaged"), &["converged"]),
            (
                dictgen::InsensitiveStr::Ascii("absolurely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depanded"),
                &["depended", "demanded"],
            ),
            (dictgen::InsensitiveStr::Ascii("pankaces"), &["pancakes"]),
            (dictgen::InsensitiveStr::Ascii("balcklist"), &["blacklist"]),
            (
                dictgen::InsensitiveStr::Ascii("exectuioners"),
                &["executioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimzes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("abominacion"),
                &["abomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("agre"), &["agree"]),
            (
                dictgen::InsensitiveStr::Ascii("unconcsiously"),
                &["unconsciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("blasphmey"), &["blasphemy"]),
            (
                dictgen::InsensitiveStr::Ascii("designstion"),
                &["designation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerainty"), &["certainty"]),
            (
                dictgen::InsensitiveStr::Ascii("interuptions"),
                &["interruptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contolers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("genertes"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("pantehon"), &["pantheon"]),
            (dictgen::InsensitiveStr::Ascii("winnins"), &["winnings"]),
            (dictgen::InsensitiveStr::Ascii("coflicted"), &["conflicted"]),
            (
                dictgen::InsensitiveStr::Ascii("unecrypted"),
                &["unencrypted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explixitely"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivolence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("blieved"), &["believed"]),
            (
                dictgen::InsensitiveStr::Ascii("skateboader"),
                &["skateboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legalizaton"),
                &["legalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cacluating"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("quesrion"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("puesdo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendency"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorgig"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("succussor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("assassintation"),
                &["assassination"],
            ),
            (dictgen::InsensitiveStr::Ascii("nasvhille"), &["nashville"]),
            (dictgen::InsensitiveStr::Ascii("mybe"), &["maybe"]),
            (
                dictgen::InsensitiveStr::Ascii("conffigurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("patroling"), &["patrolling"]),
            (
                dictgen::InsensitiveStr::Ascii("declration"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialzed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("appllying"), &["applying"]),
            (
                dictgen::InsensitiveStr::Ascii("stereotpying"),
                &["stereotyping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementatins"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classfication"),
                &["classification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infreqentcy"),
                &["infrequency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("udated"),
                &["updated", "dated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informativo"),
                &["information"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimizer"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("twoard"), &["toward"]),
            (
                dictgen::InsensitiveStr::Ascii("infograpic"),
                &["infographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("jonction"), &["junction"]),
            (dictgen::InsensitiveStr::Ascii("als"), &["also"]),
            (dictgen::InsensitiveStr::Ascii("repubican"), &["republican"]),
            (dictgen::InsensitiveStr::Ascii("laer"), &["later", "layer"]),
            (
                dictgen::InsensitiveStr::Ascii("administren"),
                &["administer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacification"),
                &["specification", "pacification"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconect"), &["reconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("acqusition"),
                &["acquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("craceless"), &["graceless"]),
            (
                dictgen::InsensitiveStr::Ascii("progarammatic"),
                &["programmatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("dominaton"), &["domination"]),
            (
                dictgen::InsensitiveStr::Ascii("concideration"),
                &["consideration"],
            ),
            (dictgen::InsensitiveStr::Ascii("encrytps"), &["encrypts"]),
            (dictgen::InsensitiveStr::Ascii("produde"), &["produce"]),
            (dictgen::InsensitiveStr::Ascii("tumbnail"), &["thumbnail"]),
            (dictgen::InsensitiveStr::Ascii("decmeber"), &["december"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentatally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substiutions"),
                &["substitutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtiel"), &["subtitle"]),
            (
                dictgen::InsensitiveStr::Ascii("equivallently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directorys"),
                &["directors", "directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("toshibe"), &["toshiba"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscriped"),
                &["unsubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sandwhiched"),
                &["sandwiched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrumentos"),
                &["instruments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contanimation"),
                &["contamination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholarhsips"),
                &["scholarships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thourgh"),
                &["thorough", "through"],
            ),
            (dictgen::InsensitiveStr::Ascii("timestampe"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("bitcoints"), &["bitcoins"]),
            (dictgen::InsensitiveStr::Ascii("disktops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("architecs"), &["architects"]),
            (dictgen::InsensitiveStr::Ascii("methapors"), &["metaphors"]),
            (dictgen::InsensitiveStr::Ascii("abondon"), &["abandon"]),
            (dictgen::InsensitiveStr::Ascii("monstruos"), &["monstrous"]),
            (
                dictgen::InsensitiveStr::Ascii("missconfigure"),
                &["misconfigure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangted"),
                &["rearranged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forwording"),
                &["forwarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("deallot"), &["dealloc"]),
            (dictgen::InsensitiveStr::Ascii("mesage"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivety"),
                &["sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("productivty"),
                &["productivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmantation"),
                &["segmentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsurprized"),
                &["unsurprised"],
            ),
            (dictgen::InsensitiveStr::Ascii("emergend"), &["emerged"]),
            (dictgen::InsensitiveStr::Ascii("stroage"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("speficitier"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotioation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipulitive"),
                &["manipulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decompressd"),
                &["decompressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("causalties"),
                &["casualties"],
            ),
            (dictgen::InsensitiveStr::Ascii("vershined"), &["versioned"]),
            (
                dictgen::InsensitiveStr::Ascii("ecoder"),
                &["encoder", "decoder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diplacement"),
                &["displacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ontents"),
                &["contents", "intents"],
            ),
            (dictgen::InsensitiveStr::Ascii("nstall"), &["install"]),
            (
                dictgen::InsensitiveStr::Ascii("trotzkyite"),
                &["trotskyist"],
            ),
            (dictgen::InsensitiveStr::Ascii("yaching"), &["yachting"]),
            (dictgen::InsensitiveStr::Ascii("elemenents"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("ilter"),
                &["filter", "liter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reduse"),
                &["reduce", "reuse", "refuse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitalised"),
                &["reinitialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalizations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntesized"),
                &["synthesized"],
            ),
            (dictgen::InsensitiveStr::Ascii("transimit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("attrirbutes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaule"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("chanisaw"), &["chainsaw"]),
            (dictgen::InsensitiveStr::Ascii("nuissance"), &["nuisance"]),
            (
                dictgen::InsensitiveStr::Ascii("classs"),
                &["class", "classes"],
            ),
            (dictgen::InsensitiveStr::Ascii("calandars"), &["calendars"]),
            (dictgen::InsensitiveStr::Ascii("winthin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("annnotation"),
                &["annotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alghoritm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("comemnt"), &["comment"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutionens"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mezmorised"),
                &["mesmerised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muliplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deconnecting"),
                &["disconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinoids"), &["sinusoids"]),
            (dictgen::InsensitiveStr::Ascii("agregator"), &["aggregator"]),
            (dictgen::InsensitiveStr::Ascii("emipres"), &["empires"]),
            (
                dictgen::InsensitiveStr::Ascii("eigenevectors"),
                &["eigenvectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicfy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("regisries"), &["registries"]),
            (
                dictgen::InsensitiveStr::Ascii("consultating"),
                &["consultation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rebuillding"),
                &["rebuilding"],
            ),
            (dictgen::InsensitiveStr::Ascii("molule"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("loacations"), &["locations"]),
            (
                dictgen::InsensitiveStr::Ascii("arbirtarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxiumn"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("seletor"), &["selector"]),
            (dictgen::InsensitiveStr::Ascii("recongnize"), &["recognize"]),
            (
                dictgen::InsensitiveStr::Ascii("behavioual"),
                &["behavioural"],
            ),
            (dictgen::InsensitiveStr::Ascii("anologon"), &["analogon"]),
            (dictgen::InsensitiveStr::Ascii("asignment"), &["assignment"]),
            (dictgen::InsensitiveStr::Ascii("nessecarry"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("porprty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("eqaulizer"), &["equalizer"]),
            (
                dictgen::InsensitiveStr::Ascii("neurosceince"),
                &["neuroscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthims"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("owervrite"), &["overwrite"]),
            (
                dictgen::InsensitiveStr::Ascii("uncoverted"),
                &["unconverted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprensentation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("regylar"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("skeletaal"), &["skeletal"]),
            (dictgen::InsensitiveStr::Ascii("strema"), &["stream"]),
            (
                dictgen::InsensitiveStr::Ascii("restuarant"),
                &["restaurant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apologizeing"),
                &["apologizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbbox"), &["inbox"]),
            (dictgen::InsensitiveStr::Ascii("challengr"), &["challenger"]),
            (dictgen::InsensitiveStr::Ascii("rerelase"), &["rerelease"]),
            (dictgen::InsensitiveStr::Ascii("orgiin"), &["origin"]),
            (dictgen::InsensitiveStr::Ascii("snaphot"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("opyion"), &["option"]),
            (dictgen::InsensitiveStr::Ascii("acocunted"), &["accounted"]),
            (
                dictgen::InsensitiveStr::Ascii("proclaming"),
                &["proclaiming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimdnted"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("riges"),
                &["rides", "ridges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bilateraly"),
                &["bilaterally"],
            ),
            (dictgen::InsensitiveStr::Ascii("stich"), &["stitch"]),
            (
                dictgen::InsensitiveStr::Ascii("accodomations"),
                &["accommodations"],
            ),
            (dictgen::InsensitiveStr::Ascii("agrc"), &["argc"]),
            (dictgen::InsensitiveStr::Ascii("drection"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("approprite"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgrated"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("itslef"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("pijuns"), &["pigeons"]),
            (
                dictgen::InsensitiveStr::Ascii("cusomization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("tigher"), &["tighter"]),
            (
                dictgen::InsensitiveStr::Ascii("inpterpreter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nestolgically"),
                &["nostalgically"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifu"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("intsances"), &["instances"]),
            (
                dictgen::InsensitiveStr::Ascii("impossiblly"),
                &["impossibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structires"),
                &["structures"],
            ),
            (dictgen::InsensitiveStr::Ascii("implemts"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("implemention"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabilties"),
                &["probabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("useing"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("aktivate"), &["activate"]),
            (dictgen::InsensitiveStr::Ascii("switfly"), &["swiftly"]),
            (
                dictgen::InsensitiveStr::Ascii("unsages"),
                &["usages", "unstages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifcations"),
                &["notifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stablilizer"),
                &["stabilizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confifuring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civizilation"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gholding"), &["holding"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interveiwed"),
                &["interviewed"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaiable"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("continetal"),
                &["continental"],
            ),
            (dictgen::InsensitiveStr::Ascii("degreeee"), &["degree"]),
            (dictgen::InsensitiveStr::Ascii("carefuly"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("performancetest"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mdifielders"),
                &["midfielders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autheticators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predecesser"),
                &["predecessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrulates"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("technic"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("invaildate"),
                &["invalidates", "invalidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concanented"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("pensino"), &["pension"]),
            (
                dictgen::InsensitiveStr::Ascii("omnishience"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerablilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("appies"), &["applies"]),
            (dictgen::InsensitiveStr::Ascii("agnostisch"), &["agnostic"]),
            (dictgen::InsensitiveStr::Ascii("dployed"), &["deployed"]),
            (dictgen::InsensitiveStr::Ascii("abitrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("promis"), &["promise"]),
            (
                dictgen::InsensitiveStr::Ascii("triggerring"),
                &["triggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("compalins"), &["complains"]),
            (dictgen::InsensitiveStr::Ascii("situdio"), &["studio"]),
            (
                dictgen::InsensitiveStr::Ascii("metropolitian"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construted"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplifed"), &["simplified"]),
            (
                dictgen::InsensitiveStr::Ascii("deragotory"),
                &["derogatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integreted"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecct"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("separatelly"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("catory"), &["factory"]),
            (dictgen::InsensitiveStr::Ascii("leaglity"), &["legality"]),
            (dictgen::InsensitiveStr::Ascii("widgetas"), &["widgets"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigiu"),
                &["prestigious"],
            ),
            (dictgen::InsensitiveStr::Ascii("decicing"), &["deciding"]),
            (dictgen::InsensitiveStr::Ascii("intorduce"), &["introduce"]),
            (dictgen::InsensitiveStr::Ascii("optimzer"), &["optimizer"]),
            (
                dictgen::InsensitiveStr::Ascii("dictonnaries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compleatly"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniators"), &["initiators"]),
            (dictgen::InsensitiveStr::Ascii("creaations"), &["creations"]),
            (dictgen::InsensitiveStr::Ascii("accesssors"), &["accessors"]),
            (
                dictgen::InsensitiveStr::Ascii("interrrupted"),
                &["interrupted"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheeots"), &["cheetos"]),
            (
                dictgen::InsensitiveStr::Ascii("tansparent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonisations"),
                &["synchronisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condradiction"),
                &["contradiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("passionite"),
                &["passionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sunglesses"),
                &["sunglasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assigntment"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("monchrome"), &["monochrome"]),
            (
                dictgen::InsensitiveStr::Ascii("demolution"),
                &["demolition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("technnically"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("groouping"), &["grouping"]),
            (dictgen::InsensitiveStr::Ascii("technoloy"), &["technology"]),
            (
                dictgen::InsensitiveStr::Ascii("expecation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satsifactorily"),
                &["satisfactorily"],
            ),
            (dictgen::InsensitiveStr::Ascii("nutton"), &["button"]),
            (dictgen::InsensitiveStr::Ascii("assigntees"), &["assignees"]),
            (dictgen::InsensitiveStr::Ascii("reelease"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("reportedy"), &["reportedly"]),
            (dictgen::InsensitiveStr::Ascii("comput"), &["compute"]),
            (
                dictgen::InsensitiveStr::Ascii("symphatizer"),
                &["sympathizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawack"), &["drawback"]),
            (
                dictgen::InsensitiveStr::Ascii("transpants"),
                &["transplants"],
            ),
            (dictgen::InsensitiveStr::Ascii("contatcs"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("partcipate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("colision"), &["collision"]),
            (dictgen::InsensitiveStr::Ascii("hsi"), &["his"]),
            (dictgen::InsensitiveStr::Ascii("dake"), &["take"]),
            (dictgen::InsensitiveStr::Ascii("dorce"), &["force"]),
            (dictgen::InsensitiveStr::Ascii("vinyets"), &["vignettes"]),
            (dictgen::InsensitiveStr::Ascii("wissled"), &["whistled"]),
            (dictgen::InsensitiveStr::Ascii("tirdent"), &["trident"]),
            (
                dictgen::InsensitiveStr::Ascii("disasembled"),
                &["disassembled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecification"),
                &["specification", "pacification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememnting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triskadekafobia"),
                &["triskaidekaphobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitmuli"), &["stimuli"]),
            (
                dictgen::InsensitiveStr::Ascii("conceptification"),
                &["conceptualisation", "conceptualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("margines"), &["margins"]),
            (
                dictgen::InsensitiveStr::Ascii("unsurprisin"),
                &["unsurprising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conttribution"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconencted"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cherckerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woodworkign"),
                &["woodworking"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyize"), &["analyze"]),
            (dictgen::InsensitiveStr::Ascii("eqully"), &["equally"]),
            (
                dictgen::InsensitiveStr::Ascii("conctructor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitie"),
                &["competitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propgation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conculsions"),
                &["conclusions"],
            ),
            (dictgen::InsensitiveStr::Ascii("soveits"), &["soviets"]),
            (dictgen::InsensitiveStr::Ascii("caads"), &["cards"]),
            (
                dictgen::InsensitiveStr::Ascii("fundemental"),
                &["fundamental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impractial"),
                &["impractical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limiations"),
                &["limitations"],
            ),
            (dictgen::InsensitiveStr::Ascii("intracite"), &["intricate"]),
            (
                dictgen::InsensitiveStr::Ascii("draughtman"),
                &["draughtsman"],
            ),
            (dictgen::InsensitiveStr::Ascii("propgate"), &["propagate"]),
            (
                dictgen::InsensitiveStr::Ascii("unnessecary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("forsakn"), &["forsaken"]),
            (
                dictgen::InsensitiveStr::Ascii("profitabel"),
                &["profitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("preperies"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("statemnet"), &["statement"]),
            (
                dictgen::InsensitiveStr::Ascii("comminication"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assicietive"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstante"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("clishey"), &["cliché"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscirbes"),
                &["unsubscribes"],
            ),
            (dictgen::InsensitiveStr::Ascii("incorrrect"), &["incorrect"]),
            (
                dictgen::InsensitiveStr::Ascii("irradation"),
                &["irradiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expericed"),
                &["experienced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accompagned"),
                &["accompanied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconditionaly"),
                &["unconditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("anytiem"), &["anytime"]),
            (dictgen::InsensitiveStr::Ascii("iput"), &["input"]),
            (
                dictgen::InsensitiveStr::Ascii("adaptacion"),
                &["adaptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquairum"), &["aquarium"]),
            (
                dictgen::InsensitiveStr::Ascii("auxillaries"),
                &["auxiliaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initailsations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("variebles"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("mutecies"), &["mutexes"]),
            (dictgen::InsensitiveStr::Ascii("advicable"), &["advisable"]),
            (dictgen::InsensitiveStr::Ascii("iternship"), &["internship"]),
            (
                dictgen::InsensitiveStr::Ascii("vairations"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuggge"), &["debuggee"]),
            (
                dictgen::InsensitiveStr::Ascii("origininates"),
                &["originates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differientiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("swaers"), &["swears"]),
            (dictgen::InsensitiveStr::Ascii("rewuired"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("authentcate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("indeveres"), &["endeavors"]),
            (dictgen::InsensitiveStr::Ascii("visbily"), &["visibly"]),
            (
                dictgen::InsensitiveStr::Ascii("swearshirt"),
                &["sweatshirt"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifued"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("descened"), &["descend"]),
            (dictgen::InsensitiveStr::Ascii("soudn"), &["sound"]),
            (
                dictgen::InsensitiveStr::Ascii("compatators"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoving"), &["resolving"]),
            (
                dictgen::InsensitiveStr::Ascii("interoperabilty"),
                &["interoperability"],
            ),
            (dictgen::InsensitiveStr::Ascii("accedents"), &["accidents"]),
            (dictgen::InsensitiveStr::Ascii("incensin"), &["incensing"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacterer"),
                &["manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("expct"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("propigates"),
                &["propagates"],
            ),
            (dictgen::InsensitiveStr::Ascii("moroever"), &["moreover"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaible"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foldes"),
                &["folders", "folder", "folds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unprecidented"),
                &["unprecedented"],
            ),
            (dictgen::InsensitiveStr::Ascii("identifes"), &["identifies"]),
            (
                dictgen::InsensitiveStr::Ascii("hypothesies"),
                &["hypothesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discreting"),
                &["discretion"],
            ),
            (dictgen::InsensitiveStr::Ascii("litte"), &["little"]),
            (
                dictgen::InsensitiveStr::Ascii("enthaplies"),
                &["enthalpies"],
            ),
            (dictgen::InsensitiveStr::Ascii("blak"), &["black", "blank"]),
            (dictgen::InsensitiveStr::Ascii("startus"), &["startups"]),
            (
                dictgen::InsensitiveStr::Ascii("shiftd"),
                &["shifted", "shifts"],
            ),
            (dictgen::InsensitiveStr::Ascii("licesnes"), &["licenses"]),
            (
                dictgen::InsensitiveStr::Ascii("chocolatie"),
                &["chocolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defenitely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definettly"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teminating"),
                &["terminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchngd"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("eptied"), &["emptied"]),
            (dictgen::InsensitiveStr::Ascii("comiling"), &["compiling"]),
            (dictgen::InsensitiveStr::Ascii("truimph"), &["triumph"]),
            (dictgen::InsensitiveStr::Ascii("execucted"), &["executed"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoeming"),
                &["performing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distraktion"),
                &["distraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasteurizin"),
                &["pasteurizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tenatious"), &["tenacious"]),
            (
                dictgen::InsensitiveStr::Ascii("listensed"),
                &["listened", "licensed"],
            ),
            (dictgen::InsensitiveStr::Ascii("desnely"), &["densely"]),
            (
                dictgen::InsensitiveStr::Ascii("calculater"),
                &["calculator", "calculated", "calculates", "calculate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auttoatack"),
                &["autoattack"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repositiories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideoligies"),
                &["ideologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("iranina"), &["iranian"]),
            (
                dictgen::InsensitiveStr::Ascii("opereration"),
                &["operation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congratuate"),
                &["congratulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("aritcle"), &["article"]),
            (dictgen::InsensitiveStr::Ascii("obslete"), &["obsolete"]),
            (dictgen::InsensitiveStr::Ascii("emptyy"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("encompasing"),
                &["encompassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innvoation"),
                &["innovation"],
            ),
            (dictgen::InsensitiveStr::Ascii("prfiling"), &["profiling"]),
            (dictgen::InsensitiveStr::Ascii("upacking"), &["unpacking"]),
            (dictgen::InsensitiveStr::Ascii("dicovery"), &["discovery"]),
            (
                dictgen::InsensitiveStr::Ascii("suspene"),
                &["suspend", "suspense"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("choise"),
                &["choice", "choose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informitive"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("whitch"),
                &["which", "switch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chararcters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoicates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cocurrent"), &["concurrent"]),
            (dictgen::InsensitiveStr::Ascii("neigboud"), &["neighbour"]),
            (
                dictgen::InsensitiveStr::Ascii("additionnal"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("recoverys"), &["recovers"]),
            (dictgen::InsensitiveStr::Ascii("crahes"), &["crashes"]),
            (dictgen::InsensitiveStr::Ascii("exsitance"), &["existence"]),
            (
                dictgen::InsensitiveStr::Ascii("repetetion"),
                &["repetition"],
            ),
            (dictgen::InsensitiveStr::Ascii("tweeks"), &["tweaks"]),
            (
                dictgen::InsensitiveStr::Ascii("strigifies"),
                &["stringifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permution"),
                &["permutation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pshycologists"),
                &["psychologists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectino"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contstrains"),
                &["constrains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenanighans"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calidations"),
                &["validations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somethinkg"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("resurced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("representaton"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ismalist"), &["islamist"]),
            (dictgen::InsensitiveStr::Ascii("famliar"), &["familiar"]),
            (
                dictgen::InsensitiveStr::Ascii("preparacion"),
                &["preparation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nitorgen"), &["nitrogen"]),
            (
                dictgen::InsensitiveStr::Ascii("descriotive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("billioniares"),
                &["billionaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enchantmets"),
                &["enchantments"],
            ),
            (dictgen::InsensitiveStr::Ascii("undersand"), &["understand"]),
            (dictgen::InsensitiveStr::Ascii("avilable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("transeivers"),
                &["transceivers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refectors"),
                &["refactors", "reflectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestors"),
                &["manifests"],
            ),
            (dictgen::InsensitiveStr::Ascii("prescrie"), &["prescribed"]),
            (dictgen::InsensitiveStr::Ascii("annouced"), &["announced"]),
            (
                dictgen::InsensitiveStr::Ascii("contrainer"),
                &["container", "constrained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensibity"),
                &["extensibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requriments"),
                &["requirements"],
            ),
            (dictgen::InsensitiveStr::Ascii("rininging"), &["ringing"]),
            (dictgen::InsensitiveStr::Ascii("zars"), &["czars"]),
            (
                dictgen::InsensitiveStr::Ascii("fortunaly"),
                &["fortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boorjwazee"),
                &["bourgeoisie"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialistions"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpeloying"), &["deploying"]),
            (dictgen::InsensitiveStr::Ascii("exibits"), &["exhibits"]),
            (
                dictgen::InsensitiveStr::Ascii("chronocles"),
                &["chronicles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simlutaneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socialini"),
                &["socializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pennsylvaina"),
                &["pennsylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalisn"),
                &["nationals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecction"),
                &["connection"],
            ),
            (dictgen::InsensitiveStr::Ascii("usred"), &["used"]),
            (dictgen::InsensitiveStr::Ascii("devation"), &["deviation"]),
            (dictgen::InsensitiveStr::Ascii("orginial"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("fotgotten"), &["forgotten"]),
            (dictgen::InsensitiveStr::Ascii("rewtched"), &["wretched"]),
            (dictgen::InsensitiveStr::Ascii("exausted"), &["exhausted"]),
            (
                dictgen::InsensitiveStr::Ascii("crsytallise"),
                &["crystallise"],
            ),
            (dictgen::InsensitiveStr::Ascii("hygeinic"), &["hygienic"]),
            (
                dictgen::InsensitiveStr::Ascii("unparented"),
                &["parentless"],
            ),
            (dictgen::InsensitiveStr::Ascii("borken"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("utilizaton"),
                &["utilization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadventaged"),
                &["disadvantaged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lonelyness"),
                &["loneliness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialisaiton"),
                &["initialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depserately"),
                &["desperately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intermperance"),
                &["intemperance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demogrpahic"),
                &["demographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenitity"), &["identity"]),
            (dictgen::InsensitiveStr::Ascii("cavaet"), &["caveat"]),
            (
                dictgen::InsensitiveStr::Ascii("compleness"),
                &["completeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("businesman"),
                &["businessman"],
            ),
            (dictgen::InsensitiveStr::Ascii("ploaded"), &["uploaded"]),
            (
                dictgen::InsensitiveStr::Ascii("contirbutor"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("precding"), &["preceding"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisitio"),
                &["inquisitor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correesponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("unicks"), &["eunuchs"]),
            (dictgen::InsensitiveStr::Ascii("fromatted"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("floopy"), &["floppy"]),
            (dictgen::InsensitiveStr::Ascii("lenguages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("comparations"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificaition"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congragulate"),
                &["congratulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("valulation"), &["valuation"]),
            (
                dictgen::InsensitiveStr::Ascii("recommendors"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("checcksums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("matchign"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("annnouncing"),
                &["announcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defintivly"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("densley"), &["densely"]),
            (dictgen::InsensitiveStr::Ascii("deifned"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("advandced"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("anxiosness"),
                &["anxiousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabitilibies"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sereousally"),
                &["seriously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithmical"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrest"),
                &["interest", "insert"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accoringly"),
                &["accordingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberrations"),
                &["aberrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("conparing"), &["comparing"]),
            (dictgen::InsensitiveStr::Ascii("reponse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("ministerns"), &["ministers"]),
            (
                dictgen::InsensitiveStr::Ascii("exectuableness"),
                &["executableness"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatct"), &["contact"]),
            (
                dictgen::InsensitiveStr::Ascii("plagerises"),
                &["plagiarises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endienness"),
                &["endianness"],
            ),
            (dictgen::InsensitiveStr::Ascii("unehanced"), &["unenhanced"]),
            (dictgen::InsensitiveStr::Ascii("propator"), &["propagator"]),
            (dictgen::InsensitiveStr::Ascii("pasasword"), &["password"]),
            (
                dictgen::InsensitiveStr::Ascii("interiting"),
                &["inheriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chartiable"),
                &["charitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsited"), &["offside"]),
            (dictgen::InsensitiveStr::Ascii("cappable"), &["capable"]),
            (
                dictgen::InsensitiveStr::Ascii("availibilty"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carinomial"),
                &["ceremonial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooedinates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("memmories"), &["memories"]),
            (dictgen::InsensitiveStr::Ascii("divisior"), &["divisor"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessfuly"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("formmatter"), &["formatter"]),
            (
                dictgen::InsensitiveStr::Ascii("considerble"),
                &["considerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrecoginized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("maragret"), &["margaret"]),
            (dictgen::InsensitiveStr::Ascii("litterate"), &["literate"]),
            (
                dictgen::InsensitiveStr::Ascii("blockcahin"),
                &["blockchain"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimizare"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("nucelus"), &["nucleus"]),
            (dictgen::InsensitiveStr::Ascii("dwscended"), &["descended"]),
            (
                dictgen::InsensitiveStr::Ascii("ostrocized"),
                &["ostracized"],
            ),
            (dictgen::InsensitiveStr::Ascii("garbarge"), &["garbage"]),
            (dictgen::InsensitiveStr::Ascii("chanined"), &["chained"]),
            (dictgen::InsensitiveStr::Ascii("woth"), &["worth", "with"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbrohood"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addiditions"),
                &["additions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptoins"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("defalts"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("skepticisim"),
                &["skepticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("repports"), &["reports"]),
            (
                dictgen::InsensitiveStr::Ascii("compatiblilities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("detectionns"),
                &["detections"],
            ),
            (dictgen::InsensitiveStr::Ascii("logrithm"), &["logarithm"]),
            (
                dictgen::InsensitiveStr::Ascii("theromstat"),
                &["thermostat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catasthrope"),
                &["catastrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unaccaptable"),
                &["unacceptable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expementers"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavailble"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("markeys"), &["marquees"]),
            (dictgen::InsensitiveStr::Ascii("extatic"), &["ecstatic"]),
            (dictgen::InsensitiveStr::Ascii("traslalate"), &["translate"]),
            (dictgen::InsensitiveStr::Ascii("incvalid"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("calculuate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("eduction"),
                &["education", "reduction", "deduction", "seduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sceintists"),
                &["scientists"],
            ),
            (dictgen::InsensitiveStr::Ascii("upploader"), &["uploader"]),
            (dictgen::InsensitiveStr::Ascii("maintiains"), &["maintains"]),
            (
                dictgen::InsensitiveStr::Ascii("instaciating"),
                &["instantiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bastardous"), &["bastards"]),
            (
                dictgen::InsensitiveStr::Ascii("authroities"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privaleged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("lecutres"), &["lectures"]),
            (
                dictgen::InsensitiveStr::Ascii("transparcenies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefifier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("absolautely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartells"), &["cartels"]),
            (dictgen::InsensitiveStr::Ascii("decition"), &["decision"]),
            (dictgen::InsensitiveStr::Ascii("expeted"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("determinisctic"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("oberve"), &["observe"]),
            (dictgen::InsensitiveStr::Ascii("funguses"), &["fungi"]),
            (
                dictgen::InsensitiveStr::Ascii("beforehands"),
                &["beforehand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustrants"),
                &["frustrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("adddressed"), &["addressed"]),
            (dictgen::InsensitiveStr::Ascii("comensate"), &["compensate"]),
            (
                dictgen::InsensitiveStr::Ascii("comparitive"),
                &["comparative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southmapton"),
                &["southampton"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowladges"),
                &["acknowledges"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsiting"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("regaring"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstanial"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("heigts"), &["heights"]),
            (
                dictgen::InsensitiveStr::Ascii("prosessors"),
                &["processors", "professors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computition"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("democarcy"), &["democracy"]),
            (dictgen::InsensitiveStr::Ascii("maflormed"), &["malformed"]),
            (
                dictgen::InsensitiveStr::Ascii("scandinacian"),
                &["scandinavian"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbatum"), &["verbatim"]),
            (dictgen::InsensitiveStr::Ascii("exlpoits"), &["exploits"]),
            (
                dictgen::InsensitiveStr::Ascii("eneumretaions"),
                &["enumerations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cconverter"), &["converter"]),
            (
                dictgen::InsensitiveStr::Ascii("graeter"),
                &["greater", "grater"],
            ),
            (dictgen::InsensitiveStr::Ascii("analsyer"), &["analyser"]),
            (
                dictgen::InsensitiveStr::Ascii("speficifeirs"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhoor"),
                &["neighbour"],
            ),
            (dictgen::InsensitiveStr::Ascii("solatary"), &["solitary"]),
            (dictgen::InsensitiveStr::Ascii("commplex"), &["complex"]),
            (
                dictgen::InsensitiveStr::Ascii("desciriptions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("rahpsody"), &["rhapsody"]),
            (dictgen::InsensitiveStr::Ascii("soceities"), &["societies"]),
            (dictgen::InsensitiveStr::Ascii("progess"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("arsneal"), &["arsenal"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fortuntely"),
                &["fortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("kno"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("starighten"),
                &["straighten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inintialising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("reckognize"), &["recognize"]),
            (dictgen::InsensitiveStr::Ascii("workspacce"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("serveice"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("cannotation"),
                &["connotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("neutraly"), &["neutrality"]),
            (dictgen::InsensitiveStr::Ascii("vibratao"), &["vibrato"]),
            (
                dictgen::InsensitiveStr::Ascii("tempature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("algothitm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("passwird"), &["password"]),
            (dictgen::InsensitiveStr::Ascii("naviagtes"), &["navigates"]),
            (
                dictgen::InsensitiveStr::Ascii("paramiters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastenating"),
                &["procrastinating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibiteds"),
                &["prohibits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ntegration"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosntraints"),
                &["constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("altogher"), &["altogether"]),
            (dictgen::InsensitiveStr::Ascii("deatches"), &["detaches"]),
            (
                dictgen::InsensitiveStr::Ascii("troubelsome"),
                &["troublesome"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interracted"),
                &["interacted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipuled"),
                &["manipulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("succssful"), &["successful"]),
            (
                dictgen::InsensitiveStr::Ascii("philoshopically"),
                &["philosophically"],
            ),
            (dictgen::InsensitiveStr::Ascii("valentie"), &["valentines"]),
            (
                dictgen::InsensitiveStr::Ascii("miraculosly"),
                &["miraculously"],
            ),
            (dictgen::InsensitiveStr::Ascii("suceed"), &["succeed"]),
            (dictgen::InsensitiveStr::Ascii("presonal"), &["personal"]),
            (dictgen::InsensitiveStr::Ascii("addtition"), &["addition"]),
            (dictgen::InsensitiveStr::Ascii("descripe"), &["describe"]),
            (
                dictgen::InsensitiveStr::Ascii("reveral"),
                &["reversal", "referral"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoureces"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("tidibts"), &["tidbits"]),
            (
                dictgen::InsensitiveStr::Ascii("exolicitly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("attepting"), &["attempting"]),
            (dictgen::InsensitiveStr::Ascii("coliding"), &["colliding"]),
            (dictgen::InsensitiveStr::Ascii("resorced"), &["resourced"]),
            (dictgen::InsensitiveStr::Ascii("scandalose"), &["scandals"]),
            (
                dictgen::InsensitiveStr::Ascii("avaiilability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("reguster"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("sewdonims"), &["pseudonyms"]),
            (dictgen::InsensitiveStr::Ascii("againgst"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("impleemntations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("handedley"), &["handedly"]),
            (dictgen::InsensitiveStr::Ascii("temporray"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("orienation"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaruante"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("additoinally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplmented"),
                &["reimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("refererd"), &["referred"]),
            (dictgen::InsensitiveStr::Ascii("depositos"), &["deposits"]),
            (dictgen::InsensitiveStr::Ascii("invalide"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("inizially"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("asymetricaly"),
                &["asymmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reosurce"), &["resource"]),
            (
                dictgen::InsensitiveStr::Ascii("positivley"),
                &["positively"],
            ),
            (dictgen::InsensitiveStr::Ascii("fileld"), &["field"]),
            (
                dictgen::InsensitiveStr::Ascii("enthuasists"),
                &["enthusiasts"],
            ),
            (dictgen::InsensitiveStr::Ascii("brilinear"), &["bilinear"]),
            (dictgen::InsensitiveStr::Ascii("intergrals"), &["integrals"]),
            (dictgen::InsensitiveStr::Ascii("shfting"), &["shifting"]),
            (dictgen::InsensitiveStr::Ascii("apprach"), &["approach"]),
            (dictgen::InsensitiveStr::Ascii("alowable"), &["allowable"]),
            (dictgen::InsensitiveStr::Ascii("misssions"), &["missions"]),
            (
                dictgen::InsensitiveStr::Ascii("concentratin"),
                &["concentration", "concentrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bersekr"), &["berserk"]),
            (dictgen::InsensitiveStr::Ascii("mached"), &["matched"]),
            (
                dictgen::InsensitiveStr::Ascii("regularlised"),
                &["regularised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompleete"),
                &["incomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("incomin"), &["incoming"]),
            (dictgen::InsensitiveStr::Ascii("remmapped"), &["remapped"]),
            (dictgen::InsensitiveStr::Ascii("packaeg"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("contrictions"),
                &["contractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astericks"),
                &["asterisk", "asterisks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meassurable"),
                &["measurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instutitions"),
                &["institutions"],
            ),
            (
                dictgen::InsensitiveStr::Unicode("сontainor"),
                &["container"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifciation"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ovservation"),
                &["observation"],
            ),
            (dictgen::InsensitiveStr::Ascii("exampel"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("pertub"), &["perturb"]),
            (
                dictgen::InsensitiveStr::Ascii("slecting"),
                &["selecting", "electing"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertor"), &["operator"]),
            (dictgen::InsensitiveStr::Ascii("dispach"), &["dispatch"]),
            (
                dictgen::InsensitiveStr::Ascii("conventience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregetion"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreprenaur"),
                &["entrepreneur"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissables"), &["disables"]),
            (
                dictgen::InsensitiveStr::Ascii("compustion"),
                &["compulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxeption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("oprands"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("emabrgo"), &["embargo"]),
            (
                dictgen::InsensitiveStr::Ascii("roiginally"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("attriubte"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("correcpond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("pointetrs"), &["pointers"]),
            (
                dictgen::InsensitiveStr::Ascii("advisorys"),
                &["advisors", "advisories"],
            ),
            (dictgen::InsensitiveStr::Ascii("phlema"), &["phlegma"]),
            (
                dictgen::InsensitiveStr::Ascii("idetifiable"),
                &["identifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("bargian"), &["bargain"]),
            (dictgen::InsensitiveStr::Ascii("chgange"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("comparion"), &["comparison"]),
            (dictgen::InsensitiveStr::Ascii("monitary"), &["monetary"]),
            (dictgen::InsensitiveStr::Ascii("vaccuming"), &["vacuuming"]),
            (
                dictgen::InsensitiveStr::Ascii("addicitons"),
                &["addictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("datea"), &["date", "data"]),
            (
                dictgen::InsensitiveStr::Ascii("optionalliy"),
                &["optionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("folwoin"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstalations"),
                &["reinstallations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipulaitng"),
                &["manipulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conneciton"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuplaoded"),
                &["reuploaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("overloard"), &["overload"]),
            (
                dictgen::InsensitiveStr::Ascii("exectation"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("truamatic"), &["traumatic"]),
            (dictgen::InsensitiveStr::Ascii("nagated"), &["negated"]),
            (dictgen::InsensitiveStr::Ascii("dalmation"), &["dalmatian"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboordhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("figurativley"),
                &["figuratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrbiuted"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prohibitve"),
                &["prohibited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declearative"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("detais"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("unusualy"), &["unusually"]),
            (dictgen::InsensitiveStr::Ascii("ethe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("mininterpret"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouned"), &["bounded"]),
            (dictgen::InsensitiveStr::Ascii("prfixes"), &["prefixes"]),
            (
                dictgen::InsensitiveStr::Ascii("tempreatures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegosiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("productiviy"),
                &["productivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("indluge"), &["indulge"]),
            (
                dictgen::InsensitiveStr::Ascii("confortable"),
                &["comfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geomerties"),
                &["geometries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spirutually"),
                &["spiritually"],
            ),
            (dictgen::InsensitiveStr::Ascii("experims"), &["experiments"]),
            (
                dictgen::InsensitiveStr::Ascii("situatuions"),
                &["situations"],
            ),
            (dictgen::InsensitiveStr::Ascii("domian"), &["domain"]),
            (
                dictgen::InsensitiveStr::Ascii("unrecovarable"),
                &["unrecoverable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeimenters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("corale"), &["chorale"]),
            (dictgen::InsensitiveStr::Ascii("pubisher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("performa"), &["performs"]),
            (dictgen::InsensitiveStr::Ascii("formattter"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("hsyteria"), &["hysteria"]),
            (
                dictgen::InsensitiveStr::Ascii("opinyonist"),
                &["opinionist"],
            ),
            (dictgen::InsensitiveStr::Ascii("crurrent"), &["current"]),
            (dictgen::InsensitiveStr::Ascii("yosimete"), &["yosemite"]),
            (dictgen::InsensitiveStr::Ascii("adverst"), &["adverts"]),
            (
                dictgen::InsensitiveStr::Ascii("mercanaries"),
                &["mercenaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoprorated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapenning"), &["happening"]),
            (dictgen::InsensitiveStr::Ascii("reqesting"), &["requesting"]),
            (
                dictgen::InsensitiveStr::Ascii("prodecures"),
                &["procedures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chasnge"),
                &["change", "changes"],
            ),
            (dictgen::InsensitiveStr::Ascii("classrooom"), &["classroom"]),
            (
                dictgen::InsensitiveStr::Ascii("walkthrthrough"),
                &["walkthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("maesures"), &["measures"]),
            (dictgen::InsensitiveStr::Ascii("seaonal"), &["seasonal"]),
            (dictgen::InsensitiveStr::Ascii("endponts"), &["endpoints"]),
            (
                dictgen::InsensitiveStr::Ascii("expirement"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("polishs"), &["polishes"]),
            (
                dictgen::InsensitiveStr::Ascii("convenciently"),
                &["conveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("chceksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("thanksgivng"),
                &["thanksgiving"],
            ),
            (dictgen::InsensitiveStr::Ascii("nulcear"), &["nuclear"]),
            (dictgen::InsensitiveStr::Ascii("elemnetal"), &["elemental"]),
            (
                dictgen::InsensitiveStr::Ascii("autentication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("calssroom"), &["classroom"]),
            (dictgen::InsensitiveStr::Ascii("fule"), &["file", "rule"]),
            (
                dictgen::InsensitiveStr::Ascii("parameterical"),
                &["parametrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("nauitlus"), &["nautilus"]),
            (dictgen::InsensitiveStr::Ascii("blanace"), &["balance"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunaltely"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("persume"), &["presume"]),
            (
                dictgen::InsensitiveStr::Ascii("colaborate"),
                &["collaborate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservatibe"),
                &["conservative"],
            ),
            (dictgen::InsensitiveStr::Ascii("seesions"), &["sessions"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrtily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("assising"), &["assisting"]),
            (dictgen::InsensitiveStr::Ascii("anid"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("truelly"), &["truly"]),
            (dictgen::InsensitiveStr::Ascii("yestrday"), &["yesterday"]),
            (dictgen::InsensitiveStr::Ascii("idetifer"), &["identifier"]),
            (
                dictgen::InsensitiveStr::Ascii("innitialiser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustraties"),
                &["illustrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("throuhg"), &["through"]),
            (dictgen::InsensitiveStr::Ascii("annnotates"), &["annotates"]),
            (
                dictgen::InsensitiveStr::Ascii("replicaiing"),
                &["replicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fruitsations"),
                &["frustrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ontrolling"),
                &["controlling"],
            ),
            (dictgen::InsensitiveStr::Ascii("peference"), &["preference"]),
            (dictgen::InsensitiveStr::Ascii("dispplay"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("definetions"),
                &["definitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalsits"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apreciated"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repostiories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecifying"),
                &["specifying", "pacifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("assciated"), &["associated"]),
            (dictgen::InsensitiveStr::Ascii("intital"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("coomitting"),
                &["committing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considereed"),
                &["considered"],
            ),
            (dictgen::InsensitiveStr::Ascii("receptie"), &["receptive"]),
            (
                dictgen::InsensitiveStr::Ascii("psycholigist"),
                &["psychologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("pennals"), &["panels"]),
            (dictgen::InsensitiveStr::Ascii("hygenic"), &["hygienic"]),
            (dictgen::InsensitiveStr::Ascii("hadned"), &["handed"]),
            (
                dictgen::InsensitiveStr::Ascii("intrumentation"),
                &["instrumentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cassettte"), &["cassette"]),
            (
                dictgen::InsensitiveStr::Ascii("usefullness"),
                &["usefulness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accellerator"),
                &["accelerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trustwhorty"),
                &["trustworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucsession"),
                &["succession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downlading"),
                &["downloading"],
            ),
            (dictgen::InsensitiveStr::Ascii("poignat"), &["poignant"]),
            (dictgen::InsensitiveStr::Ascii("faulyt"), &["faulty"]),
            (
                dictgen::InsensitiveStr::Ascii("annilihation"),
                &["annihilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("backfiled"), &["backfield"]),
            (dictgen::InsensitiveStr::Ascii("tomororw"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("loacls"), &["locals"]),
            (dictgen::InsensitiveStr::Ascii("supposeds"), &["supposed"]),
            (dictgen::InsensitiveStr::Ascii("arcivers"), &["archivers"]),
            (
                dictgen::InsensitiveStr::Ascii("specualting"),
                &["speculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininisters"),
                &["administers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfoprmation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlihgters"),
                &["highlighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("aribtrary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("coeffficient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("icesickle"), &["icicle"]),
            (
                dictgen::InsensitiveStr::Ascii("louieville"),
                &["louisville"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutully"), &["mutually"]),
            (
                dictgen::InsensitiveStr::Ascii("pennsilvania"),
                &["pennsylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("shriley"), &["shirley"]),
            (dictgen::InsensitiveStr::Ascii("dsplaced"), &["displaced"]),
            (dictgen::InsensitiveStr::Ascii("betweent"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("similerly"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("traddition"), &["tradition"]),
            (dictgen::InsensitiveStr::Ascii("vinyet"), &["vignette"]),
            (
                dictgen::InsensitiveStr::Ascii("expewrimentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("secuely"), &["securely"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisong"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("charas"), &["chars"]),
            (dictgen::InsensitiveStr::Ascii("mathch"), &["match"]),
            (dictgen::InsensitiveStr::Ascii("tthere"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("authenenticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpacities"), &["capacities"]),
            (dictgen::InsensitiveStr::Ascii("offpsring"), &["offspring"]),
            (dictgen::InsensitiveStr::Ascii("eiegn"), &["eigen", "reign"]),
            (
                dictgen::InsensitiveStr::Ascii("apparellin"),
                &["apparelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("opjective"), &["objective"]),
            (
                dictgen::InsensitiveStr::Ascii("cootdinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cellebrated"),
                &["celebrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornogrphy"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulatenous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("clese"), &["close"]),
            (dictgen::InsensitiveStr::Ascii("immobilzes"), &["immobile"]),
            (dictgen::InsensitiveStr::Ascii("territoy"), &["territory"]),
            (dictgen::InsensitiveStr::Ascii("deoder"), &["decoder"]),
            (dictgen::InsensitiveStr::Ascii("someghing"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("unterstood"),
                &["understood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metroploitan"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpoloate"),
                &["interpolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrictins"),
                &["restricts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modernizin"),
                &["modernizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plased"),
                &["placed", "pleased", "phased"],
            ),
            (dictgen::InsensitiveStr::Ascii("souvinear"), &["souvenir"]),
            (
                dictgen::InsensitiveStr::Ascii("roduced"),
                &["reduced", "produced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shatow"),
                &["château", "châteaux", "shadow"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supposidly"),
                &["supposedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescibing"),
                &["prescribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cronological"),
                &["chronological"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissapate"), &["dissipate"]),
            (
                dictgen::InsensitiveStr::Ascii("spacifies"),
                &["specifies", "pacifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracterization"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recepit"),
                &["recipe", "receipt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincidentia"),
                &["coincidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individully"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulerabilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compairison"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliacation"),
                &["duplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("fibonnacci"), &["fibonacci"]),
            (dictgen::InsensitiveStr::Ascii("enbedding"), &["embedding"]),
            (dictgen::InsensitiveStr::Ascii("ptd"), &["pdf"]),
            (
                dictgen::InsensitiveStr::Ascii("inhertiing"),
                &["inheriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iverse"),
                &["diverse", "inverse"],
            ),
            (dictgen::InsensitiveStr::Ascii("handlade"), &["handled"]),
            (
                dictgen::InsensitiveStr::Ascii("defintitely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opjectified"),
                &["objectified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rranslations"),
                &["translations"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodivers"), &["providers"]),
            (dictgen::InsensitiveStr::Ascii("dowlinks"), &["downlinks"]),
            (dictgen::InsensitiveStr::Ascii("cahrged"), &["charged"]),
            (dictgen::InsensitiveStr::Ascii("proram"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("deslection"),
                &["deselection", "deflection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("actiavtors"),
                &["activators"],
            ),
            (dictgen::InsensitiveStr::Ascii("unpopulair"), &["unpopular"]),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appontments"),
                &["appointments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcularon"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("vairant"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("permissie"),
                &["permissible"],
            ),
            (dictgen::InsensitiveStr::Ascii("consepts"), &["concepts"]),
            (
                dictgen::InsensitiveStr::Ascii("wavelegnths"),
                &["wavelength"],
            ),
            (dictgen::InsensitiveStr::Ascii("persones"), &["persons"]),
            (dictgen::InsensitiveStr::Ascii("balenced"), &["balanced"]),
            (
                dictgen::InsensitiveStr::Ascii("funtionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("couterparts"),
                &["counterparts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teamifghts"),
                &["teamfights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meagthread"),
                &["megathread"],
            ),
            (dictgen::InsensitiveStr::Ascii("falis"), &["fails"]),
            (dictgen::InsensitiveStr::Ascii("stategies"), &["strategies"]),
            (dictgen::InsensitiveStr::Ascii("inouts"), &["inputs"]),
            (dictgen::InsensitiveStr::Ascii("tyeps"), &["types"]),
            (
                dictgen::InsensitiveStr::Ascii("unseccessful"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independenantly"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determenant"),
                &["determinant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarnanteed"),
                &["guaranteed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitckers"), &["stickers"]),
            (dictgen::InsensitiveStr::Ascii("appoval"), &["approval"]),
            (dictgen::InsensitiveStr::Ascii("eveyone"), &["everyone"]),
            (dictgen::InsensitiveStr::Ascii("presist"), &["persist"]),
            (dictgen::InsensitiveStr::Ascii("lessson"), &["lesson"]),
            (dictgen::InsensitiveStr::Ascii("prooxies"), &["proxies"]),
            (dictgen::InsensitiveStr::Ascii("helerps"), &["helpers"]),
            (
                dictgen::InsensitiveStr::Ascii("friendzonded"),
                &["friendzoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("withoug"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("identitiy"), &["identity"]),
            (
                dictgen::InsensitiveStr::Ascii("aloable"),
                &["allowable", "available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychotisch"),
                &["psychotic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equilibirum"),
                &["equilibrium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbute"),
                &["distribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("upadating"), &["updating"]),
            (dictgen::InsensitiveStr::Ascii("offfice"), &["office"]),
            (
                dictgen::InsensitiveStr::Ascii("transscript"),
                &["transcript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropritae"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fashionalble"),
                &["fashionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("jelusey"), &["jealousy"]),
            (
                dictgen::InsensitiveStr::Ascii("declinining"),
                &["declining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agggregator"),
                &["aggregator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presntations"),
                &["presentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sunfiber"), &["sunfire"]),
            (
                dictgen::InsensitiveStr::Ascii("breat"),
                &["breast", "beat", "bread"],
            ),
            (dictgen::InsensitiveStr::Ascii("vegtarian"), &["vegetarian"]),
            (dictgen::InsensitiveStr::Ascii("seperete"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("relaspe"), &["relapse"]),
            (
                dictgen::InsensitiveStr::Ascii("accepptable"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("peformer"), &["performer"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wrkarounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impilcitly"),
                &["implicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmenetd"), &["segmented"]),
            (dictgen::InsensitiveStr::Ascii("amonst"), &["amongst"]),
            (
                dictgen::InsensitiveStr::Ascii("intergation"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vizualization"),
                &["visualization", "virtualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lible"),
                &["libel", "liable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ministy"), &["ministry"]),
            (
                dictgen::InsensitiveStr::Ascii("schiziphrenic"),
                &["schizophrenic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninintialized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hornor"),
                &["honor", "horror"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwasy"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("encourging"),
                &["encouraging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("involontary"),
                &["involuntary"],
            ),
            (dictgen::InsensitiveStr::Ascii("claibre"), &["calibre"]),
            (dictgen::InsensitiveStr::Ascii("exucutes"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("desentralized"),
                &["decentralized"],
            ),
            (dictgen::InsensitiveStr::Ascii("winnigns"), &["winnings"]),
            (
                dictgen::InsensitiveStr::Ascii("prefectches"),
                &["prefetches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstators"),
                &["demonstrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranlate"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("customziers"),
                &["customizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaes"), &["animates"]),
            (dictgen::InsensitiveStr::Ascii("testimoney"), &["testimony"]),
            (
                dictgen::InsensitiveStr::Ascii("preformers"),
                &["performers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containiner"),
                &["container"],
            ),
            (dictgen::InsensitiveStr::Ascii("austrai"), &["austria"]),
            (
                dictgen::InsensitiveStr::Ascii("explecitely"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("pedantisch"), &["pedantic"]),
            (
                dictgen::InsensitiveStr::Ascii("substituitions"),
                &["substitutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("christainity"),
                &["christianity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straigthens"),
                &["straightens"],
            ),
            (dictgen::InsensitiveStr::Ascii("desponse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("aboslves"), &["absolves"]),
            (
                dictgen::InsensitiveStr::Ascii("corresopnded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enginees"),
                &["engineers", "engines"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("javasrript"),
                &["javascript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scripe"),
                &["scrape", "scribe", "scrip", "script", "stripe"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniate"), &["initiate"]),
            (
                dictgen::InsensitiveStr::Ascii("intervenire"),
                &["intervene"],
            ),
            (dictgen::InsensitiveStr::Ascii("answerd"), &["answered"]),
            (dictgen::InsensitiveStr::Ascii("facttories"), &["factories"]),
            (
                dictgen::InsensitiveStr::Ascii("detecs"),
                &["detects", "deters", "detect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrival"),
                &["retrieval", "retrial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrains"),
                &["constrains", "constraints", "contains"],
            ),
            (dictgen::InsensitiveStr::Ascii("fieled"), &["field"]),
            (dictgen::InsensitiveStr::Ascii("contaxted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("suppoerts"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("hopefulle"), &["hopefully"]),
            (dictgen::InsensitiveStr::Ascii("sissered"), &["scissored"]),
            (dictgen::InsensitiveStr::Ascii("specualte"), &["speculate"]),
            (dictgen::InsensitiveStr::Ascii("wating"), &["waiting"]),
            (dictgen::InsensitiveStr::Ascii("chisell"), &["chisel"]),
            (
                dictgen::InsensitiveStr::Ascii("archaelogical"),
                &["archaeological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cercomstance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("delicous"), &["delicious"]),
            (dictgen::InsensitiveStr::Ascii("ojekts"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalitic"),
                &["nationalistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("homapages"), &["homepages"]),
            (dictgen::InsensitiveStr::Ascii("redonly"), &["readonly"]),
            (dictgen::InsensitiveStr::Ascii("varifier"), &["verifier"]),
            (
                dictgen::InsensitiveStr::Ascii("assiocates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preidction"),
                &["prediction"],
            ),
            (dictgen::InsensitiveStr::Ascii("agrandize"), &["aggrandize"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferncers"),
                &["dereferencers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composibility"),
                &["composability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cachse"),
                &["cache", "caches"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoecracy"), &["theocracy"]),
            (dictgen::InsensitiveStr::Ascii("conceps"), &["concepts"]),
            (
                dictgen::InsensitiveStr::Ascii("jurneys"),
                &["journeys", "gurneys"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convertes"),
                &["converts", "converted", "converters"],
            ),
            (dictgen::InsensitiveStr::Ascii("aktion"), &["action"]),
            (dictgen::InsensitiveStr::Ascii("ommit"), &["omit"]),
            (dictgen::InsensitiveStr::Ascii("continas"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("fufills"), &["fulfills"]),
            (
                dictgen::InsensitiveStr::Ascii("diffentiating"),
                &["differentiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranferred"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hightlighter"),
                &["highlighter"],
            ),
            (dictgen::InsensitiveStr::Ascii("itteration"), &["iteration"]),
            (dictgen::InsensitiveStr::Ascii("fnction"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("specifified"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgaraded"),
                &["downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("hitboxs"), &["hitboxes"]),
            (
                dictgen::InsensitiveStr::Ascii("thumbmails"),
                &["thumbnails"],
            ),
            (dictgen::InsensitiveStr::Ascii("succcessor"), &["successor"]),
            (dictgen::InsensitiveStr::Ascii("cresent"), &["crescent"]),
            (
                dictgen::InsensitiveStr::Ascii("caracterising"),
                &["characterising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projecitle"),
                &["projectiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("totalty"),
                &["totally", "totality"],
            ),
            (dictgen::InsensitiveStr::Ascii("lolipops"), &["lollipops"]),
            (dictgen::InsensitiveStr::Ascii("erro"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("pendinig"), &["pending"]),
            (dictgen::InsensitiveStr::Ascii("finialize"), &["finalize"]),
            (
                dictgen::InsensitiveStr::Ascii("documanting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("meatfile"), &["metafile"]),
            (dictgen::InsensitiveStr::Ascii("legact"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("uderstandable"),
                &["understandable"],
            ),
            (dictgen::InsensitiveStr::Ascii("shuting"), &["shutting"]),
            (
                dictgen::InsensitiveStr::Ascii("muktitasking"),
                &["multitasking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reliablity"),
                &["reliability"],
            ),
            (dictgen::InsensitiveStr::Ascii("fasodds"), &["facades"]),
            (dictgen::InsensitiveStr::Ascii("specufy"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("reposonses"), &["responses"]),
            (
                dictgen::InsensitiveStr::Ascii("idenitfiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millsecond"),
                &["millisecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declarayions"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathematican"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("associted"), &["associated"]),
            (
                dictgen::InsensitiveStr::Ascii("interersted"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interactiv"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("epsionage"), &["espionage"]),
            (dictgen::InsensitiveStr::Ascii("veichles"), &["vehicles"]),
            (
                dictgen::InsensitiveStr::Ascii("certificiates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profesisonal"),
                &["professionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("unassing"), &["unassign"]),
            (dictgen::InsensitiveStr::Ascii("writeble"), &["writable"]),
            (dictgen::InsensitiveStr::Ascii("diagonales"), &["diagonals"]),
            (dictgen::InsensitiveStr::Ascii("equaation"), &["equation"]),
            (
                dictgen::InsensitiveStr::Ascii("kidknapping"),
                &["kidnapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoing"), &["avoiding"]),
            (dictgen::InsensitiveStr::Ascii("pengiuns"), &["penguins"]),
            (dictgen::InsensitiveStr::Ascii("predatobr"), &["predator"]),
            (
                dictgen::InsensitiveStr::Ascii("countinuation"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("emptniess"), &["emptiness"]),
            (dictgen::InsensitiveStr::Ascii("assessmet"), &["assessment"]),
            (
                dictgen::InsensitiveStr::Ascii("reprucible"),
                &["reproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingnoring"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("artcles"), &["articles"]),
            (
                dictgen::InsensitiveStr::Ascii("compresseed"),
                &["compressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheecksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("depdendents"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("spcifiers"), &["specifiers"]),
            (dictgen::InsensitiveStr::Ascii("wotked"), &["worked"]),
            (
                dictgen::InsensitiveStr::Ascii("culiminates"),
                &["culminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisssioned"),
                &["commissioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribure"),
                &["distribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("exercide"), &["exercised"]),
            (dictgen::InsensitiveStr::Ascii("rebult"), &["rebuilt"]),
            (dictgen::InsensitiveStr::Ascii("likey"), &["likely"]),
            (dictgen::InsensitiveStr::Ascii("gerenate"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("addicionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometines"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("spesificaly"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("visulizer"), &["visualizer"]),
            (dictgen::InsensitiveStr::Ascii("neareste"), &["nearest"]),
            (dictgen::InsensitiveStr::Ascii("theocracry"), &["theocracy"]),
            (dictgen::InsensitiveStr::Ascii("prodcuts"), &["products"]),
            (dictgen::InsensitiveStr::Ascii("exceutor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("considerablely"),
                &["considerably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outrageosly"),
                &["outrageously"],
            ),
            (dictgen::InsensitiveStr::Ascii("missplaces"), &["misplaces"]),
            (dictgen::InsensitiveStr::Ascii("typcasts"), &["typecasts"]),
            (
                dictgen::InsensitiveStr::Ascii("seperatedly"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conbine"),
                &["combine", "confine"],
            ),
            (dictgen::InsensitiveStr::Ascii("collonade"), &["colonnade"]),
            (dictgen::InsensitiveStr::Ascii("differemt"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("comparitor"),
                &["comparator"],
            ),
            (dictgen::InsensitiveStr::Ascii("spltting"), &["splitting"]),
            (dictgen::InsensitiveStr::Ascii("liekly"), &["likely"]),
            (dictgen::InsensitiveStr::Ascii("phemonena"), &["phenomena"]),
            (
                dictgen::InsensitiveStr::Ascii("sparce"),
                &["sparse", "space", "spare", "scarce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legistration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("homapage"), &["homepage"]),
            (
                dictgen::InsensitiveStr::Ascii("origniating"),
                &["originating"],
            ),
            (dictgen::InsensitiveStr::Ascii("matacher"), &["matcher"]),
            (
                dictgen::InsensitiveStr::Ascii("ascic"),
                &["ascii", "aspic", "ascetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("extreemly"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("defininte"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("relexation"),
                &["relegation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reulator"), &["regulator"]),
            (dictgen::InsensitiveStr::Ascii("restor"), &["restore"]),
            (dictgen::InsensitiveStr::Ascii("nimphs"), &["nymphs"]),
            (dictgen::InsensitiveStr::Ascii("obstables"), &["obstacles"]),
            (
                dictgen::InsensitiveStr::Ascii("extradtion"),
                &["extraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("addind"), &["adding"]),
            (
                dictgen::InsensitiveStr::Ascii("declaratinos"),
                &["declarations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transaltors"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacup"), &["backup"]),
            (dictgen::InsensitiveStr::Ascii("heigher"), &["higher"]),
            (
                dictgen::InsensitiveStr::Ascii("conpletion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionning"),
                &["conditioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commnandline"),
                &["commandline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("playthrougs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("enflamed"), &["inflamed"]),
            (
                dictgen::InsensitiveStr::Ascii("overwelmed"),
                &["overwhelmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anarchistes"),
                &["anarchists"],
            ),
            (dictgen::InsensitiveStr::Ascii("htacccess"), &["htaccess"]),
            (
                dictgen::InsensitiveStr::Ascii("balckberry"),
                &["blackberry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("logarighmic"),
                &["logarithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("requitin"), &["requiting"]),
            (dictgen::InsensitiveStr::Ascii("focusses"), &["focuses"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowlwdge"),
                &["acknowledge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempories"),
                &["temporaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("fucntion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("morbidy"), &["morbidly"]),
            (
                dictgen::InsensitiveStr::Ascii("chosed"),
                &["chose", "chosen"],
            ),
            (dictgen::InsensitiveStr::Ascii("lossing"), &["losing"]),
            (
                dictgen::InsensitiveStr::Ascii("sucesssfull"),
                &["successful"],
            ),
            (dictgen::InsensitiveStr::Ascii("availbe"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("porvides"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("parrameters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valye"),
                &["valse", "value", "valve"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schematrion"),
                &["schematron"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conatiners"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astronautas"),
                &["astronauts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replicaiting"),
                &["replicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("doapmine"), &["dopamine"]),
            (dictgen::InsensitiveStr::Ascii("confogure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("resevation"),
                &["reservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meassaging"),
                &["messaging", "massaging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomenonal"),
                &["phenomenal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qualificaiton"),
                &["qualifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("amout"), &["amount"]),
            (
                dictgen::InsensitiveStr::Ascii("implementators"),
                &["implementers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credibillity"),
                &["credibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("mapppings"), &["mappings"]),
            (dictgen::InsensitiveStr::Ascii("creadit"), &["credit"]),
            (dictgen::InsensitiveStr::Ascii("militis"), &["militias"]),
            (dictgen::InsensitiveStr::Ascii("defaulhs"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("coditions"), &["conditions"]),
            (dictgen::InsensitiveStr::Ascii("pakced"), &["packed"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretare"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concurrect"),
                &["concurrent"],
            ),
            (dictgen::InsensitiveStr::Ascii("schduled"), &["scheduled"]),
            (
                dictgen::InsensitiveStr::Ascii("unbelizeable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("certianty"), &["certainty"]),
            (dictgen::InsensitiveStr::Ascii("polciy"), &["policy"]),
            (dictgen::InsensitiveStr::Ascii("depentent"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("offlaod"), &["offload"]),
            (dictgen::InsensitiveStr::Ascii("whaterver"), &["whatever"]),
            (
                dictgen::InsensitiveStr::Ascii("disucssing"),
                &["discussing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggregrations"),
                &["aggregations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fontend"),
                &["frontend", "contend"],
            ),
            (dictgen::InsensitiveStr::Ascii("monochome"), &["monochrome"]),
            (dictgen::InsensitiveStr::Ascii("efinition"), &["definition"]),
            (dictgen::InsensitiveStr::Ascii("iterrupt"), &["interrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("relete"),
                &["delete", "relate"],
            ),
            (dictgen::InsensitiveStr::Ascii("handfull"), &["handful"]),
            (dictgen::InsensitiveStr::Ascii("deflecto"), &["deflection"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegociations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofice"), &["office"]),
            (dictgen::InsensitiveStr::Ascii("accetps"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("assynchronous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutscence"), &["cutscene"]),
            (
                dictgen::InsensitiveStr::Ascii("significat"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemnting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strategice"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterss"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("recursses"), &["recurses"]),
            (
                dictgen::InsensitiveStr::Ascii("slaugthering"),
                &["slaughtering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithyms"),
                &["algorithm"],
            ),
            (dictgen::InsensitiveStr::Ascii("throrough"), &["thorough"]),
            (dictgen::InsensitiveStr::Ascii("calcualte"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("neareast"), &["nearest"]),
            (dictgen::InsensitiveStr::Ascii("accoint"), &["account"]),
            (
                dictgen::InsensitiveStr::Ascii("alighned"),
                &["aligned", "alighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeators"), &["creators"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpreter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governmently"),
                &["governmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("reductin"), &["reduction"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthruogh"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("otehrwise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutins"),
                &["revolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("sktechy"), &["sketchy"]),
            (dictgen::InsensitiveStr::Ascii("shystems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("hirachical"),
                &["hierarchical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alforithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("clairity"), &["clarity"]),
            (
                dictgen::InsensitiveStr::Ascii("witable"),
                &["writeable", "writable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singificantly"),
                &["significantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versitilaty"),
                &["versatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathimatical"),
                &["mathematical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recogintion"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introdiction"),
                &["introduction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializae"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("naibhor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("cententenial"),
                &["centennial"],
            ),
            (dictgen::InsensitiveStr::Ascii("tresured"), &["treasured"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellizing"),
                &["parallelizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("happne"), &["happen"]),
            (dictgen::InsensitiveStr::Ascii("doocument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("origniates"),
                &["originates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescirbed"),
                &["prescribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("reomves"), &["removes"]),
            (
                dictgen::InsensitiveStr::Ascii("repsectively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepered"), &["shepherd"]),
            (dictgen::InsensitiveStr::Ascii("spliitting"), &["splitting"]),
            (dictgen::InsensitiveStr::Ascii("flusing"), &["flushing"]),
            (
                dictgen::InsensitiveStr::Ascii("evaliation"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrking"), &["working"]),
            (dictgen::InsensitiveStr::Ascii("dirrective"), &["directive"]),
            (dictgen::InsensitiveStr::Ascii("aparently"), &["apparently"]),
            (dictgen::InsensitiveStr::Ascii("caese"), &["cease"]),
            (
                dictgen::InsensitiveStr::Ascii("accomplishemnts"),
                &["accomplishments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuccesful"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocuntry"), &["country"]),
            (
                dictgen::InsensitiveStr::Ascii("chanks"),
                &[
                    "clanks", "cranks", "chants", "chinks", "chunks", "shanks", "hanks", "chonks",
                    "thanks",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("contstruct"), &["construct"]),
            (
                dictgen::InsensitiveStr::Ascii("annhiliating"),
                &["annihilating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explanetary"),
                &["explanatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatelty"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denomenator"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("stabilze"), &["stabilize"]),
            (dictgen::InsensitiveStr::Ascii("asolute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("sanitzed"), &["sanitized"]),
            (
                dictgen::InsensitiveStr::Ascii("inconveneince"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("coluns"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentais"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retciever"),
                &["receiver", "retriever"],
            ),
            (dictgen::InsensitiveStr::Ascii("entiteis"), &["entities"]),
            (dictgen::InsensitiveStr::Ascii("sheck"), &["check", "shuck"]),
            (
                dictgen::InsensitiveStr::Ascii("conifigurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("thex"), &["the", "then"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("originall"),
                &["original", "originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("quroum"), &["quorum"]),
            (dictgen::InsensitiveStr::Ascii("patethic"), &["pathetic"]),
            (dictgen::InsensitiveStr::Ascii("ocluded"), &["occluded"]),
            (dictgen::InsensitiveStr::Ascii("anyy"), &["any"]),
            (dictgen::InsensitiveStr::Ascii("knwoledge"), &["knowledge"]),
            (dictgen::InsensitiveStr::Ascii("languae"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("appearane"),
                &["appearances", "appearance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transylmania"),
                &["transylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notces"),
                &["notices", "notches", "notes", "nonces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determindes"),
                &["determined", "determines"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdening"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("periphereals"),
                &["peripherals"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitaminas"), &["vitamins"]),
            (dictgen::InsensitiveStr::Ascii("versoion"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("gometric"), &["geometric"]),
            (dictgen::InsensitiveStr::Ascii("crsytal"), &["crystal"]),
            (dictgen::InsensitiveStr::Ascii("expetect"), &["expect"]),
            (
                dictgen::InsensitiveStr::Ascii("symphatized"),
                &["sympathize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppotunity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuillities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("recomends"), &["recommends"]),
            (
                dictgen::InsensitiveStr::Ascii("sentencian"),
                &["sentencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("alcehmist"), &["alchemist"]),
            (dictgen::InsensitiveStr::Ascii("licesnses"), &["licenses"]),
            (
                dictgen::InsensitiveStr::Ascii("muliplications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vurnerability"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similiraties"),
                &["similarities"],
            ),
            (dictgen::InsensitiveStr::Ascii("difenitly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("youy"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitryarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phychiatrist"),
                &["psychiatrist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environemnt"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("sotres"), &["stores"]),
            (
                dictgen::InsensitiveStr::Ascii("distingishing"),
                &["distinguishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("impactos"), &["impacts"]),
            (dictgen::InsensitiveStr::Ascii("speakign"), &["speaking"]),
            (
                dictgen::InsensitiveStr::Ascii("compfortable"),
                &["comfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehnique"), &["technique"]),
            (
                dictgen::InsensitiveStr::Ascii("concentraited"),
                &["concentrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("effec"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("unreasonabily"),
                &["unreasonably"],
            ),
            (dictgen::InsensitiveStr::Ascii("inveryed"), &["inverted"]),
            (dictgen::InsensitiveStr::Ascii("retutn"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("comrpessing"),
                &["compressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("goalkeepr"), &["goalkeeper"]),
            (dictgen::InsensitiveStr::Ascii("oriended"), &["oriented"]),
            (dictgen::InsensitiveStr::Ascii("receptores"), &["receptors"]),
            (dictgen::InsensitiveStr::Ascii("istself"), &["itself"]),
            (dictgen::InsensitiveStr::Ascii("memoriez"), &["memorize"]),
            (
                dictgen::InsensitiveStr::Ascii("predeccesors"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("coneector"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("bootstappers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependentes"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("minmum"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("adventagously"),
                &["advantageously"],
            ),
            (dictgen::InsensitiveStr::Ascii("probabaly"), &["probably"]),
            (dictgen::InsensitiveStr::Ascii("possitions"), &["positions"]),
            (dictgen::InsensitiveStr::Ascii("desroys"), &["destroys"]),
            (
                dictgen::InsensitiveStr::Ascii("integation"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("aliasses"), &["aliases"]),
            (dictgen::InsensitiveStr::Ascii("journying"), &["journeying"]),
            (dictgen::InsensitiveStr::Ascii("robberts"), &["robbers"]),
            (
                dictgen::InsensitiveStr::Ascii("multipectral"),
                &["multispectral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pilon"),
                &["pillion", "pylon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sctions"),
                &["sections", "suctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("marineris"), &["mariners"]),
            (
                dictgen::InsensitiveStr::Ascii("resopnsible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitued"),
                &["substitute", "substituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("tropedo"), &["torpedo"]),
            (dictgen::InsensitiveStr::Ascii("stoppped"), &["stopped"]),
            (dictgen::InsensitiveStr::Ascii("imforming"), &["informing"]),
            (
                dictgen::InsensitiveStr::Ascii("apporpriation"),
                &["appropriation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resultions"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recursivion"),
                &["recursion"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearating"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("fillter"),
                &["filter", "filler"],
            ),
            (dictgen::InsensitiveStr::Ascii("parttions"), &["partitions"]),
            (dictgen::InsensitiveStr::Ascii("whiltelist"), &["whitelist"]),
            (
                dictgen::InsensitiveStr::Ascii("functionlity"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviromnental"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declaraction"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("trye"), &["true", "try"]),
            (dictgen::InsensitiveStr::Ascii("wressel"), &["wrestle"]),
            (dictgen::InsensitiveStr::Ascii("elmenet"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("quesrioning"),
                &["questioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provinvial"),
                &["provincial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassember"),
                &["disassembler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispesnary"),
                &["dispensary"],
            ),
            (dictgen::InsensitiveStr::Ascii("frequet"), &["frequent"]),
            (dictgen::InsensitiveStr::Ascii("nummeric"), &["numeric"]),
            (
                dictgen::InsensitiveStr::Ascii("dissalowed"),
                &["disallowed"],
            ),
            (dictgen::InsensitiveStr::Ascii("conclusie"), &["conclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("acquistions"),
                &["acquisitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chekcout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("optionl"),
                &["optional", "option"],
            ),
            (dictgen::InsensitiveStr::Ascii("comediac"), &["comedic"]),
            (dictgen::InsensitiveStr::Ascii("grapical"), &["graphical"]),
            (
                dictgen::InsensitiveStr::Ascii("counterracts"),
                &["counterparts"],
            ),
            (dictgen::InsensitiveStr::Ascii("soemtime"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("onwing"), &["owning"]),
            (dictgen::InsensitiveStr::Ascii("decribing"), &["describing"]),
            (
                dictgen::InsensitiveStr::Ascii("developemnt"),
                &["developments", "development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtualizaiton"),
                &["virtualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stablilizers"),
                &["stabilizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silhouweting"),
                &["silhouetting"],
            ),
            (dictgen::InsensitiveStr::Ascii("northwet"), &["northwest"]),
            (
                dictgen::InsensitiveStr::Ascii("ougoing"),
                &["outgoing", "ongoing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respondible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procuces"),
                &["procures", "produces"],
            ),
            (dictgen::InsensitiveStr::Ascii("resullted"), &["resulted"]),
            (
                dictgen::InsensitiveStr::Ascii("multilplicative"),
                &["multiplicative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lavelling"),
                &["levelling", "labelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("definitie"), &["definitive"]),
            (
                dictgen::InsensitiveStr::Ascii("negosiator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("asapragus"), &["asparagus"]),
            (
                dictgen::InsensitiveStr::Ascii("instlls"),
                &["installs", "instills"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacarmento"),
                &["sacramento"],
            ),
            (dictgen::InsensitiveStr::Ascii("dyanamic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("licesing"), &["licensing"]),
            (dictgen::InsensitiveStr::Ascii("execusive"), &["exclusive"]),
            (dictgen::InsensitiveStr::Ascii("whipsers"), &["whispers"]),
            (
                dictgen::InsensitiveStr::Ascii("prostetution"),
                &["prostitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registerdns"),
                &["registers"],
            ),
            (dictgen::InsensitiveStr::Ascii("failiver"), &["failover"]),
            (dictgen::InsensitiveStr::Ascii("pathces"), &["patches"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmiates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("tumbnails"), &["thumbnails"]),
            (
                dictgen::InsensitiveStr::Ascii("charateristics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaegd"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("acitvely"), &["actively"]),
            (dictgen::InsensitiveStr::Ascii("ciricuit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("hypen"), &["hyphen"]),
            (dictgen::InsensitiveStr::Ascii("picknicks"), &["picnics"]),
            (
                dictgen::InsensitiveStr::Ascii("uncertaity"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("serounded"), &["surrounded"]),
            (dictgen::InsensitiveStr::Ascii("scorpiomon"), &["scorpion"]),
            (dictgen::InsensitiveStr::Ascii("sumation"), &["summation"]),
            (dictgen::InsensitiveStr::Ascii("tortilllas"), &["tortilla"]),
            (dictgen::InsensitiveStr::Ascii("arised"), &["arose"]),
            (dictgen::InsensitiveStr::Ascii("glicthy"), &["glitchy"]),
            (
                dictgen::InsensitiveStr::Ascii("ajdectives"),
                &["adjectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("deskys"), &["disguise"]),
            (
                dictgen::InsensitiveStr::Ascii("worled"),
                &["world", "worked", "whorled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phillipines"),
                &["philippines"],
            ),
            (dictgen::InsensitiveStr::Ascii("amke"), &["make"]),
            (
                dictgen::InsensitiveStr::Ascii("documetnation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unbannd"), &["unbanned"]),
            (
                dictgen::InsensitiveStr::Ascii("simulatanously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("realod"), &["reload"]),
            (dictgen::InsensitiveStr::Ascii("deepos"), &["depots"]),
            (
                dictgen::InsensitiveStr::Ascii("perists"),
                &["persists", "persist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaboratore"),
                &["collaborate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylically"), &["cyclically"]),
            (dictgen::InsensitiveStr::Ascii("oppressin"), &["oppression"]),
            (dictgen::InsensitiveStr::Ascii("rrecorder"), &["recorder"]),
            (dictgen::InsensitiveStr::Ascii("excecutor"), &["executor"]),
            (dictgen::InsensitiveStr::Ascii("sharipe"), &["sharpie"]),
            (
                dictgen::InsensitiveStr::Ascii("spearating"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delagation"),
                &["delegation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacrasm"), &["sarcasm"]),
            (dictgen::InsensitiveStr::Ascii("evryday"), &["everyday"]),
            (dictgen::InsensitiveStr::Ascii("speciaal"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("ccordinats"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permissibe"),
                &["permissible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsceptible"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("clickare"), &["clicker"]),
            (dictgen::InsensitiveStr::Ascii("slighty"), &["slightly"]),
            (
                dictgen::InsensitiveStr::Ascii("regenarate"),
                &["regenerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dasporas"), &["diasporas"]),
            (dictgen::InsensitiveStr::Ascii("shepereds"), &["shepherds"]),
            (
                dictgen::InsensitiveStr::Ascii("tremondous"),
                &["tremendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("guerrila"), &["guerrilla"]),
            (
                dictgen::InsensitiveStr::Ascii("charceters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("depreacts"), &["deprecates"]),
            (dictgen::InsensitiveStr::Ascii("paramenter"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("mississipi"),
                &["mississippi"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitation"),
                &["renegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("woith"), &["with", "worth"]),
            (dictgen::InsensitiveStr::Ascii("arbritrary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("misuing"), &["misusing"]),
            (dictgen::InsensitiveStr::Ascii("suported"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("negociation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappiont"),
                &["disappoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negothiators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprentise"),
                &["apprentice"],
            ),
            (dictgen::InsensitiveStr::Ascii("valdators"), &["validators"]),
            (
                dictgen::InsensitiveStr::Ascii("relatiation"),
                &["retaliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implimentations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("shfit"), &["shift"]),
            (dictgen::InsensitiveStr::Ascii("sigining"), &["signing"]),
            (
                dictgen::InsensitiveStr::Ascii("questional"),
                &["questionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("loclahost"), &["localhost"]),
            (
                dictgen::InsensitiveStr::Ascii("unfourtunet"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggegrating"),
                &["aggregating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fulfiled"), &["fulfilled"]),
            (dictgen::InsensitiveStr::Ascii("jaques"), &["jacques"]),
            (
                dictgen::InsensitiveStr::Ascii("cheetsheet"),
                &["cheatsheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pariticipants"),
                &["participants"],
            ),
            (dictgen::InsensitiveStr::Ascii("fimxe"), &["fixme"]),
            (
                dictgen::InsensitiveStr::Ascii("colonizators"),
                &["colonizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinded"),
                &["determine", "determined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifclly"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpolaton"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("doin"), &["doing"]),
            (dictgen::InsensitiveStr::Ascii("inteprets"), &["interprets"]),
            (dictgen::InsensitiveStr::Ascii("relaser"), &["releaser"]),
            (
                dictgen::InsensitiveStr::Ascii("additioanlly"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpports"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("personnaly"),
                &["personally"],
            ),
            (dictgen::InsensitiveStr::Ascii("lveo"), &["love"]),
            (dictgen::InsensitiveStr::Ascii("excacta"), &["exacta"]),
            (dictgen::InsensitiveStr::Ascii("noramls"), &["normals"]),
            (dictgen::InsensitiveStr::Ascii("runns"), &["runs"]),
            (
                dictgen::InsensitiveStr::Ascii("hypotehtically"),
                &["hypothetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concanetions"),
                &["concatenations"],
            ),
            (dictgen::InsensitiveStr::Ascii("binairy"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("apprciated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuotiral"), &["tutorial"]),
            (dictgen::InsensitiveStr::Ascii("tabluated"), &["tabulated"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtued"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("offficers"), &["officers"]),
            (
                dictgen::InsensitiveStr::Ascii("buis"),
                &["buoy", "buoys", "buys"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciffying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benchamrked"),
                &["benchmarked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subrouties"),
                &["subroutines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimlators"),
                &["estimators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependenciens"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("nieghbor"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("repling"), &["replying"]),
            (
                dictgen::InsensitiveStr::Ascii("discograpy"),
                &["discography"],
            ),
            (dictgen::InsensitiveStr::Ascii("templas"), &["templars"]),
            (
                dictgen::InsensitiveStr::Ascii("increamental"),
                &["incremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previsously"),
                &["previously"],
            ),
            (dictgen::InsensitiveStr::Ascii("studioes"), &["studios"]),
            (dictgen::InsensitiveStr::Ascii("arguemets"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("aesthethics"),
                &["aesthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("reproting"), &["reporting"]),
            (
                dictgen::InsensitiveStr::Ascii("caronaviruses"),
                &["coronaviruses"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciricles"), &["circles"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluatated"),
                &["evaluated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exspects"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("aborte"),
                &["aborted", "abort", "aborts"],
            ),
            (dictgen::InsensitiveStr::Ascii("exoticas"), &["exotics"]),
            (dictgen::InsensitiveStr::Ascii("noticeing"), &["noticing"]),
            (
                dictgen::InsensitiveStr::Ascii("requess"),
                &["requests", "request"],
            ),
            (dictgen::InsensitiveStr::Ascii("whisle"), &["whistle"]),
            (
                dictgen::InsensitiveStr::Ascii("pessimitic"),
                &["pessimistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reinitilize"),
                &["reinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("garnola"), &["granola"]),
            (dictgen::InsensitiveStr::Ascii("savanha"), &["savanna"]),
            (
                dictgen::InsensitiveStr::Ascii("comeptitively"),
                &["competitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("buildning"), &["building"]),
            (dictgen::InsensitiveStr::Ascii("histigram"), &["histogram"]),
            (
                dictgen::InsensitiveStr::Ascii("preservacion"),
                &["preservation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eveluations"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("trendig"), &["trending"]),
            (dictgen::InsensitiveStr::Ascii("unsgined"), &["unsigned"]),
            (dictgen::InsensitiveStr::Ascii("dirtyed"), &["dirtied"]),
            (dictgen::InsensitiveStr::Ascii("flawess"), &["flawless"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalister"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("bnndler"), &["bundler"]),
            (dictgen::InsensitiveStr::Ascii("spports"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("dispacement"),
                &["displacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("erturn"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("severeal"), &["several"]),
            (
                dictgen::InsensitiveStr::Ascii("amerliorative"),
                &["ameliorative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appretiated"),
                &["appreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeimentel"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("possition"), &["position"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubcriptions"),
                &["unsubscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recconencts"),
                &["reconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("notatin"), &["notation"]),
            (dictgen::InsensitiveStr::Ascii("anlayze"), &["analyze"]),
            (
                dictgen::InsensitiveStr::Ascii("entired"),
                &["entered", "entire"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrollling"), &["scrolling"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetimine"),
                &["amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authrorization"),
                &["authorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dinamiclly"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitmize"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("muscician"), &["musician"]),
            (dictgen::InsensitiveStr::Ascii("accoording"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("presonhood"),
                &["personhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temeratures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcissstic"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supposts"),
                &["supports", "supposes"],
            ),
            (dictgen::InsensitiveStr::Ascii("anchestors"), &["ancestors"]),
            (
                dictgen::InsensitiveStr::Ascii("shattening"),
                &["shattering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrangse"),
                &["rearranges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoceated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lateration"),
                &["alteration"],
            ),
            (dictgen::InsensitiveStr::Ascii("replecate"), &["replicate"]),
            (
                dictgen::InsensitiveStr::Ascii("billionarie"),
                &["billionaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("praticle"), &["particle"]),
            (dictgen::InsensitiveStr::Ascii("dcoking"), &["docking"]),
            (dictgen::InsensitiveStr::Ascii("cluser"), &["cluster"]),
            (
                dictgen::InsensitiveStr::Ascii("organistaion"),
                &["organisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premillenial"),
                &["premillennial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conductuve"),
                &["conductive", "conducive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motorcylce"),
                &["motorcycle"],
            ),
            (dictgen::InsensitiveStr::Ascii("opthers"), &["others"]),
            (
                dictgen::InsensitiveStr::Ascii("paramatric"),
                &["parametric"],
            ),
            (dictgen::InsensitiveStr::Ascii("setteings"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("immobille"), &["immobile"]),
            (dictgen::InsensitiveStr::Ascii("analyseres"), &["analyses"]),
            (dictgen::InsensitiveStr::Ascii("ommited"), &["omitted"]),
            (
                dictgen::InsensitiveStr::Ascii("differental"),
                &["differential"],
            ),
            (dictgen::InsensitiveStr::Ascii("artilce"), &["article"]),
            (dictgen::InsensitiveStr::Ascii("phisicaly"), &["physically"]),
            (
                dictgen::InsensitiveStr::Ascii("assosiacition"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("wekeend"), &["weekend"]),
            (
                dictgen::InsensitiveStr::Ascii("exeprimental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spermatozoan"),
                &["spermatozoon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confersation"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("executd"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("macrow"), &["macro"]),
            (
                dictgen::InsensitiveStr::Ascii("unsintalled"),
                &["uninstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inventiones"),
                &["inventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("langth"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("provincie"), &["province"]),
            (
                dictgen::InsensitiveStr::Ascii("condidtions"),
                &["conditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deslections"),
                &["deselections", "deflections"],
            ),
            (dictgen::InsensitiveStr::Ascii("siduction"), &["seduction"]),
            (dictgen::InsensitiveStr::Ascii("mertics"), &["metrics"]),
            (
                dictgen::InsensitiveStr::Ascii("calculcates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immortalty"),
                &["immortality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normalizationis"),
                &["normalization", "normalizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("escates"),
                &["escalates", "escapes"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepaires"), &["prepares"]),
            (
                dictgen::InsensitiveStr::Ascii("undergrund"),
                &["underground"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcultating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabality"),
                &["probability"],
            ),
            (dictgen::InsensitiveStr::Ascii("evnts"), &["events"]),
            (dictgen::InsensitiveStr::Ascii("forfits"), &["forfeits"]),
            (dictgen::InsensitiveStr::Ascii("apointed"), &["appointed"]),
            (
                dictgen::InsensitiveStr::Ascii("coodrinate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("previiously"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acelerator"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcuses"), &["focuses"]),
            (dictgen::InsensitiveStr::Ascii("requeted"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("verastility"),
                &["versatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("prohbit"), &["prohibit"]),
            (dictgen::InsensitiveStr::Ascii("normales"), &["normals"]),
            (
                dictgen::InsensitiveStr::Ascii("economisit"),
                &["economists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphedamine"),
                &["amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallies"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("technican"), &["technician"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatin"),
                &["alternating", "alternation"],
            ),
            (dictgen::InsensitiveStr::Ascii("scritping"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("prusues"), &["pursues"]),
            (
                dictgen::InsensitiveStr::Ascii("collectins"),
                &["collections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availabiltiy"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("autosavegs"), &["autosaves"]),
            (
                dictgen::InsensitiveStr::Ascii("acclerator"),
                &["accelerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("othopaedics"),
                &["orthopaedics"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpating"), &["adapting"]),
            (
                dictgen::InsensitiveStr::Ascii("unreleated"),
                &["unreleased", "unrelated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infulenced"),
                &["influenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("incidates"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("logers"),
                &["loggers", "lodgers"],
            ),
            (dictgen::InsensitiveStr::Ascii("buddhim"), &["buddhism"]),
            (
                dictgen::InsensitiveStr::Ascii("deinitailze"),
                &["deinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("reduceable"), &["reducible"]),
            (
                dictgen::InsensitiveStr::Ascii("commemerative"),
                &["commemorative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disproportionatly"),
                &["disproportionately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cleanilness"),
                &["cleanliness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represenatation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriubting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sporatic"), &["sporadic"]),
            (dictgen::InsensitiveStr::Ascii("qunetin"), &["quentin"]),
            (dictgen::InsensitiveStr::Ascii("fornatted"), &["formatted"]),
            (
                dictgen::InsensitiveStr::Ascii("offsest"),
                &["offset", "offsets"],
            ),
            (dictgen::InsensitiveStr::Ascii("predictie"), &["predictive"]),
            (dictgen::InsensitiveStr::Ascii("childrents"), &["childrens"]),
            (
                dictgen::InsensitiveStr::Ascii("authentcator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("hispters"), &["hipsters"]),
            (dictgen::InsensitiveStr::Ascii("advnced"), &["advanced"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplyers"),
                &["multipliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenattions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("disputs"), &["disputes"]),
            (dictgen::InsensitiveStr::Ascii("domians"), &["domains"]),
            (dictgen::InsensitiveStr::Ascii("cliped"), &["clipped"]),
            (
                dictgen::InsensitiveStr::Ascii("temperatuur"),
                &["temperature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitiarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsupressed"),
                &["unsuppressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistrubuted"),
                &["redistributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("imapcts"), &["impacts"]),
            (
                dictgen::InsensitiveStr::Ascii("terminilogy"),
                &["terminology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addititionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("relyed"), &["relied"]),
            (
                dictgen::InsensitiveStr::Ascii("kalidescope"),
                &["kaleidoscope"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipelin"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("logitude"), &["longitude"]),
            (
                dictgen::InsensitiveStr::Ascii("colourfull"),
                &["colourful", "colourfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomaptibele"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impplementation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presentaions"),
                &["presentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucesssfull"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("deriffed"), &["derived"]),
            (
                dictgen::InsensitiveStr::Ascii("representatons"),
                &["representations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priviliges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refenrences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("revoling"), &["revolving"]),
            (dictgen::InsensitiveStr::Ascii("utilies"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("observered"), &["observed"]),
            (dictgen::InsensitiveStr::Ascii("ogers"), &["ogres"]),
            (
                dictgen::InsensitiveStr::Ascii("prefentations"),
                &["presentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("froniter"), &["frontier"]),
            (dictgen::InsensitiveStr::Ascii("distrub"), &["disturb"]),
            (dictgen::InsensitiveStr::Ascii("heloers"), &["helpers"]),
            (
                dictgen::InsensitiveStr::Ascii("causin"),
                &["causing", "cousin"],
            ),
            (dictgen::InsensitiveStr::Ascii("bandwidht"), &["bandwidth"]),
            (
                dictgen::InsensitiveStr::Ascii("eptrapolates"),
                &["extrapolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periphereal"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gubnatorial"),
                &["gubernatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potenially"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cosumables"),
                &["consumables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponantially"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interferisse"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimistaion"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("delgates"), &["delegates"]),
            (dictgen::InsensitiveStr::Ascii("impcated"), &["impacted"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalizin"),
                &["capitalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginnning"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeurhooding"),
                &["neighbouring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviormentally"),
                &["environmentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspndent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("outpalyed"), &["outplayed"]),
            (
                dictgen::InsensitiveStr::Ascii("dyamic"),
                &["dynamic", "dyadic"],
            ),
            (dictgen::InsensitiveStr::Ascii("instucted"), &["instructed"]),
            (dictgen::InsensitiveStr::Ascii("pythin"), &["python"]),
            (
                dictgen::InsensitiveStr::Ascii("migations"),
                &["migrations", "mitigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("arispace"), &["airspace"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginiser"),
                &["recogniser"],
            ),
            (dictgen::InsensitiveStr::Ascii("readabout"), &["readable"]),
            (
                dictgen::InsensitiveStr::Ascii("internediate"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("overriddes"), &["overrides"]),
            (dictgen::InsensitiveStr::Ascii("italianess"), &["italians"]),
            (
                dictgen::InsensitiveStr::Ascii("particalar"),
                &["particular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memeberships"),
                &["memberships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connetcors"),
                &["connectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sohpisticated"),
                &["sophisticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("averadge"), &["averaged"]),
            (dictgen::InsensitiveStr::Ascii("uncerain"), &["uncertain"]),
            (
                dictgen::InsensitiveStr::Ascii("enumuratable"),
                &["enumerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponatenously"),
                &["spontaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraylzed"), &["paralyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("cometed"),
                &["commented", "competed"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdedders"), &["embedders"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacteur"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanted"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("numberators"),
                &["numerators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscrible"),
                &["subscriber"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoluute"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("iluminated"),
                &["illuminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("framebufer"),
                &["framebuffer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provisiones"),
                &["provisions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bankrutpcy"),
                &["bankruptcy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("swalled"),
                &["stalled", "swapped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraprhased"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expotential"),
                &["exponential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randmoness"),
                &["randomness"],
            ),
            (dictgen::InsensitiveStr::Ascii("autherise"), &["authorise"]),
            (
                dictgen::InsensitiveStr::Ascii("claculations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("allowence"), &["allowance"]),
            (dictgen::InsensitiveStr::Ascii("cloes"), &["close"]),
            (dictgen::InsensitiveStr::Ascii("efficacity"), &["efficacy"]),
            (
                dictgen::InsensitiveStr::Ascii("conglaturations"),
                &["congratulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("icluded"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("lfiesteal"), &["lifesteal"]),
            (
                dictgen::InsensitiveStr::Ascii("superstituous"),
                &["superstitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("supprting"), &["supporting"]),
            (
                dictgen::InsensitiveStr::Ascii("retransfert"),
                &["retransfer", "retransferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiziers"),
                &["optimizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("univerity"), &["university"]),
            (dictgen::InsensitiveStr::Ascii("sarinomy"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("navigatin"),
                &["navigation", "navigating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("witnesssing"),
                &["witnessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conigurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("invarinat"), &["invariant"]),
            (dictgen::InsensitiveStr::Ascii("somethign"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("mannor"), &["manner"]),
            (dictgen::InsensitiveStr::Ascii("continute"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("incomfortably"),
                &["uncomfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresspondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extimating"),
                &["estimating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occaisonaly"),
                &["occasionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particuraly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("newmonectomies"),
                &["pneumonectomies"],
            ),
            (dictgen::InsensitiveStr::Ascii("goegraphy"), &["geography"]),
            (
                dictgen::InsensitiveStr::Ascii("identies"),
                &["identities", "identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("dorp"), &["drop"]),
            (dictgen::InsensitiveStr::Ascii("minifys"), &["minifies"]),
            (dictgen::InsensitiveStr::Ascii("encoeding"), &["encoding"]),
            (
                dictgen::InsensitiveStr::Ascii("configulation"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentificators"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethicate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compansated"),
                &["compensated"],
            ),
            (dictgen::InsensitiveStr::Ascii("shaodws"), &["shadows"]),
            (dictgen::InsensitiveStr::Ascii("opereation"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("battey"), &["battery"]),
            (
                dictgen::InsensitiveStr::Ascii("laysering"),
                &["lasering", "layering"],
            ),
            (dictgen::InsensitiveStr::Ascii("combains"), &["combines"]),
            (dictgen::InsensitiveStr::Ascii("rapsady"), &["rhapsody"]),
            (
                dictgen::InsensitiveStr::Ascii("misoginyst"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excessivly"),
                &["excessively"],
            ),
            (dictgen::InsensitiveStr::Ascii("regarted"), &["regarded"]),
            (dictgen::InsensitiveStr::Ascii("atends"), &["attends"]),
            (
                dictgen::InsensitiveStr::Ascii("southamtpon"),
                &["southampton"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proable"),
                &["probable", "provable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expiriences"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("bui"), &["buoy", "buy"]),
            (dictgen::InsensitiveStr::Ascii("frmo"), &["from"]),
            (
                dictgen::InsensitiveStr::Ascii("headquatered"),
                &["headquartered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parellism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("airosft"), &["airsoft"]),
            (dictgen::InsensitiveStr::Ascii("erroneos"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("bitwse"), &["bitwise"]),
            (dictgen::InsensitiveStr::Ascii("assumng"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("visualzation"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("demsond"), &["desmond"]),
            (
                dictgen::InsensitiveStr::Ascii("temprement"),
                &["temperament"],
            ),
            (dictgen::InsensitiveStr::Ascii("prvided"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("merficul"), &["merciful"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborhoors"),
                &["neighbors"],
            ),
            (dictgen::InsensitiveStr::Ascii("hamphsire"), &["hampshire"]),
            (dictgen::InsensitiveStr::Ascii("shemas"), &["schemas"]),
            (dictgen::InsensitiveStr::Ascii("utilitizes"), &["utilizes"]),
            (dictgen::InsensitiveStr::Ascii("behaviuors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("hyperbollic"),
                &["hyperbolic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paremeterised"),
                &["parameterised"],
            ),
            (dictgen::InsensitiveStr::Ascii("pushpus"), &["pushups"]),
            (dictgen::InsensitiveStr::Ascii("fnished"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("pyramidas"), &["pyramids"]),
            (dictgen::InsensitiveStr::Ascii("blaim"), &["blame"]),
            (dictgen::InsensitiveStr::Ascii("heusitics"), &["heuristics"]),
            (dictgen::InsensitiveStr::Ascii("propertyn"), &["property"]),
            (
                dictgen::InsensitiveStr::Ascii("copenhaegen"),
                &["copenhagen"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projektile"),
                &["projectile"],
            ),
            (dictgen::InsensitiveStr::Ascii("assebmly"), &["assembly"]),
            (dictgen::InsensitiveStr::Ascii("phillipse"), &["phillies"]),
            (dictgen::InsensitiveStr::Ascii("rewievs"), &["reviews"]),
            (
                dictgen::InsensitiveStr::Ascii("trangulates"),
                &["triangulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("accseptor"), &["acceptor"]),
            (dictgen::InsensitiveStr::Ascii("caprenter"), &["carpenter"]),
            (dictgen::InsensitiveStr::Ascii("archvies"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("automactically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("emedded"), &["embedded"]),
            (
                dictgen::InsensitiveStr::Ascii("geographycally"),
                &["geographically"],
            ),
            (dictgen::InsensitiveStr::Ascii("lettter"), &["letter"]),
            (
                dictgen::InsensitiveStr::Ascii("marhsmallow"),
                &["marshmallow"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacis"), &["basic", "basis"]),
            (
                dictgen::InsensitiveStr::Ascii("architechture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalidade"),
                &["invalidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intantiating"),
                &["instantiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customziation"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("matcheable"), &["matchable"]),
            (
                dictgen::InsensitiveStr::Ascii("rescheudle"),
                &["reschedule"],
            ),
            (dictgen::InsensitiveStr::Ascii("caclium"), &["calcium"]),
            (
                dictgen::InsensitiveStr::Ascii("custumized"),
                &["customized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheapeast"), &["cheapest"]),
            (
                dictgen::InsensitiveStr::Ascii("accessibiilty"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strawberrries"),
                &["strawberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("verson"), &["version"]),
            (dictgen::InsensitiveStr::Ascii("seahwaks"), &["seahawks"]),
            (dictgen::InsensitiveStr::Ascii("exchanages"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("pronomial"), &["pronominal"]),
            (dictgen::InsensitiveStr::Ascii("deafulted"), &["defaulted"]),
            (dictgen::InsensitiveStr::Ascii("deopsited"), &["deposited"]),
            (dictgen::InsensitiveStr::Ascii("becauae"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("alternatr"), &["alternator"]),
            (dictgen::InsensitiveStr::Ascii("insigths"), &["insights"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("brightess"), &["brightness"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitalized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadvantges"),
                &["disadvantages"],
            ),
            (dictgen::InsensitiveStr::Ascii("dosplay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("utlises"), &["utilises"]),
            (
                dictgen::InsensitiveStr::Ascii("experiementations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasitionally"),
                &["transitionally", "traditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("dominiant"), &["dominant"]),
            (
                dictgen::InsensitiveStr::Ascii("humanitarion"),
                &["humanitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brekpoints"),
                &["breakpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funchtionned"),
                &["functioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscribber"),
                &["subscriber"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributiv"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("ather"), &["other"]),
            (dictgen::InsensitiveStr::Ascii("nayborly"), &["neighborly"]),
            (dictgen::InsensitiveStr::Ascii("expirese"), &["expires"]),
            (
                dictgen::InsensitiveStr::Ascii("potencials"),
                &["potentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkepoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminitrations"),
                &["administrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("manouver"), &["maneuver"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropreate"),
                &["inappropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryig"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("interepeted"),
                &["interpreted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cotributors"),
                &["contributors"],
            ),
            (dictgen::InsensitiveStr::Ascii("buidl"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("spreadhseet"),
                &["spreadsheet"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exceuctives"),
                &["executives"],
            ),
            (dictgen::InsensitiveStr::Ascii("succesors"), &["successors"]),
            (dictgen::InsensitiveStr::Ascii("droppong"), &["dropping"]),
            (
                dictgen::InsensitiveStr::Ascii("strategisch"),
                &["strategic"],
            ),
            (dictgen::InsensitiveStr::Ascii("defeaults"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("ediitors"), &["editors"]),
            (dictgen::InsensitiveStr::Ascii("daatsheet"), &["datasheet"]),
            (dictgen::InsensitiveStr::Ascii("porperly"), &["properly"]),
            (
                dictgen::InsensitiveStr::Ascii("distriutions"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unforgetable"),
                &["unforgettable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvinences"),
                &["inconveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thir"),
                &["this", "third", "their"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aforementioend"),
                &["aforementioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("scarmble"), &["scramble"]),
            (dictgen::InsensitiveStr::Ascii("chmabers"), &["chambers"]),
            (
                dictgen::InsensitiveStr::Ascii("preficientsy"),
                &["proficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("guysers"), &["geysers"]),
            (dictgen::InsensitiveStr::Ascii("surrond"), &["surround"]),
            (dictgen::InsensitiveStr::Ascii("standbay"), &["standby"]),
            (dictgen::InsensitiveStr::Ascii("acitions"), &["actions"]),
            (dictgen::InsensitiveStr::Ascii("aremnian"), &["armenian"]),
            (
                dictgen::InsensitiveStr::Ascii("fundemantal"),
                &["fundamental"],
            ),
            (dictgen::InsensitiveStr::Ascii("searhc"), &["search"]),
            (
                dictgen::InsensitiveStr::Ascii("decryptyon"),
                &["decryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("nauseuos"), &["nauseous"]),
            (
                dictgen::InsensitiveStr::Ascii("tintinabulation"),
                &["tintinnabulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("backlght"), &["backlight"]),
            (dictgen::InsensitiveStr::Ascii("employeers"), &["employees"]),
            (dictgen::InsensitiveStr::Ascii("wihite"), &["white"]),
            (
                dictgen::InsensitiveStr::Ascii("additioally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orgins"),
                &["origins", "organs"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancilliary"), &["ancillary"]),
            (
                dictgen::InsensitiveStr::Ascii("stanardized"),
                &["standardized"],
            ),
            (dictgen::InsensitiveStr::Ascii("challenege"), &["challenge"]),
            (
                dictgen::InsensitiveStr::Ascii("mangaes"),
                &["manages", "mangoes", "mangas"],
            ),
            (dictgen::InsensitiveStr::Ascii("idicator"), &["indicator"]),
            (
                dictgen::InsensitiveStr::Ascii("autherizing"),
                &["authorizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("behtesda"), &["bethesda"]),
            (
                dictgen::InsensitiveStr::Ascii("dedicacion"),
                &["dedication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("governmant"),
                &["governmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("analzer"), &["analyzer"]),
            (dictgen::InsensitiveStr::Ascii("alliasing"), &["aliasing"]),
            (
                dictgen::InsensitiveStr::Ascii("desciptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprection"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scheudling"),
                &["scheduling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("furnctional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psuedoterminal"),
                &["pseudoterminal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknoledges"),
                &["acknowledges"],
            ),
            (dictgen::InsensitiveStr::Ascii("nifes"), &["knives"]),
            (
                dictgen::InsensitiveStr::Ascii("deprecatation"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alogns"), &["aligns"]),
            (dictgen::InsensitiveStr::Ascii("annnotated"), &["annotated"]),
            (
                dictgen::InsensitiveStr::Ascii("butterfleye"),
                &["butterfly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decorrellation"),
                &["decorrelation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrustructure"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prematurily"),
                &["prematurely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexedacimal"),
                &["hexadecimal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deutschand"),
                &["deutschland"],
            ),
            (dictgen::InsensitiveStr::Ascii("leve"), &["levee", "level"]),
            (dictgen::InsensitiveStr::Ascii("droping"), &["dropping"]),
            (dictgen::InsensitiveStr::Ascii("nirtogen"), &["nitrogen"]),
            (
                dictgen::InsensitiveStr::Ascii("accumlators"),
                &["accumulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("additonal"), &["additional"]),
            (
                dictgen::InsensitiveStr::Ascii("descripter"),
                &["descriptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ceremoniis"),
                &["ceremonies"],
            ),
            (dictgen::InsensitiveStr::Ascii("willk"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("tansporting"),
                &["transporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexplicablely"),
                &["inexplicably"],
            ),
            (dictgen::InsensitiveStr::Ascii("gagdets"), &["gadgets"]),
            (dictgen::InsensitiveStr::Ascii("constly"), &["costly"]),
            (dictgen::InsensitiveStr::Ascii("teplmate"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("commnetaries"),
                &["commentaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("systyems"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("blok"), &["block", "bloke"]),
            (
                dictgen::InsensitiveStr::Ascii("impementing"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combonations"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyists"), &["analysts"]),
            (
                dictgen::InsensitiveStr::Ascii("unsychronised"),
                &["unsynchronised"],
            ),
            (dictgen::InsensitiveStr::Ascii("wanring"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("animatonic"),
                &["animatronic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armagedden"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetentations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaberatively"),
                &["collaboratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("stategic"), &["strategic"]),
            (
                dictgen::InsensitiveStr::Ascii("euqivalence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("paltform"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("acccessibility"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctruction"),
                &["destruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irriration"),
                &["irritation"],
            ),
            (dictgen::InsensitiveStr::Ascii("eqaul"), &["equal"]),
            (dictgen::InsensitiveStr::Ascii("withces"), &["witches"]),
            (
                dictgen::InsensitiveStr::Ascii("restrictios"),
                &["restricts"],
            ),
            (dictgen::InsensitiveStr::Ascii("attriebute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("temprorarily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("scepotre"), &["sceptre"]),
            (dictgen::InsensitiveStr::Ascii("gnored"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("maximice"),
                &["maximize", "maximise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanizms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encosed"),
                &["enclosed", "encoded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vetors"),
                &["vectors", "vetoes"],
            ),
            (dictgen::InsensitiveStr::Ascii("suplanted"), &["supplanted"]),
            (
                dictgen::InsensitiveStr::Ascii("necromanser"),
                &["necromancer"],
            ),
            (dictgen::InsensitiveStr::Ascii("cringery"), &["cringey"]),
            (dictgen::InsensitiveStr::Ascii("frqeuency"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("threshholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("austrlaian"),
                &["australians", "australian"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrrupt"), &["corrupt"]),
            (
                dictgen::InsensitiveStr::Ascii("transpancy"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encoutering"),
                &["encountering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefarably"),
                &["preferably"],
            ),
            (dictgen::InsensitiveStr::Ascii("moslty"), &["mostly"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscritption"),
                &["unsubscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proveribal"),
                &["proverbial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commandemnts"),
                &["commandment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woking"),
                &["working", "waking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deactiviates"),
                &["deactivates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesssion"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("auhtored"), &["authored"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpolation"),
                &["interpolation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fabircate"), &["fabricate"]),
            (
                dictgen::InsensitiveStr::Ascii("deflaction"),
                &["deflection"],
            ),
            (dictgen::InsensitiveStr::Ascii("wihin"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("timly"), &["timely"]),
            (
                dictgen::InsensitiveStr::Ascii("determistically"),
                &["deterministically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabiliy"),
                &["probability"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancell"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("reguarly"), &["regularly"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdence"),
                &["independence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anecdotallly"),
                &["anecdotally"],
            ),
            (dictgen::InsensitiveStr::Ascii("creamic"), &["ceramic"]),
            (dictgen::InsensitiveStr::Ascii("writted"), &["written"]),
            (
                dictgen::InsensitiveStr::Ascii("intiialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("sutffs"), &["stuffs"]),
            (dictgen::InsensitiveStr::Ascii("guyzer"), &["geyser"]),
            (
                dictgen::InsensitiveStr::Ascii("electrpnics"),
                &["electronics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvized"),
                &["improvised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unappretiated"),
                &["unappreciated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seramonies"),
                &["ceremonies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adversarian"),
                &["adversarial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatiblity"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("brancless"), &["branchless"]),
            (dictgen::InsensitiveStr::Ascii("deviiating"), &["deviating"]),
            (dictgen::InsensitiveStr::Ascii("foucs"), &["focus"]),
            (
                dictgen::InsensitiveStr::Ascii("respecfively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulys"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("capacitaron"),
                &["capacitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("benifite"), &["benefit"]),
            (dictgen::InsensitiveStr::Ascii("modfity"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("contaniner"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("abouve"), &["above"]),
            (
                dictgen::InsensitiveStr::Ascii("retransmitions"),
                &["retransmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("chating"), &["chatting"]),
            (dictgen::InsensitiveStr::Ascii("exteremely"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("exceds"), &["exceeds"]),
            (dictgen::InsensitiveStr::Ascii("assumming"), &["assuming"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitriraly"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissecters"),
                &["dissectors", "dissenters", "dissects"],
            ),
            (dictgen::InsensitiveStr::Ascii("cecks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("fianlisation"),
                &["finalisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocurrence"), &["occurrence"]),
            (dictgen::InsensitiveStr::Ascii("clobal"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("evoluations"),
                &["evaluations", "evolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintancne"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meory"),
                &["maori", "memory"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomulaic"), &["formulaic"]),
            (
                dictgen::InsensitiveStr::Ascii("transction"),
                &["transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("jamiacan"), &["jamaican"]),
            (dictgen::InsensitiveStr::Ascii("enccode"), &["encode"]),
            (dictgen::InsensitiveStr::Ascii("necessarry"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("attacted"), &["attached"]),
            (dictgen::InsensitiveStr::Ascii("installesd"), &["installs"]),
            (dictgen::InsensitiveStr::Ascii("grat"), &["great"]),
            (
                dictgen::InsensitiveStr::Ascii("constandly"),
                &["constantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("familiarlize"),
                &["familiarize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execptional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transplantees"),
                &["transplants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicialise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insenstive"),
                &["insensitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuggin"), &["debugging"]),
            (
                dictgen::InsensitiveStr::Ascii("traslalating"),
                &["translating"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginnin"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("cataclysym"), &["cataclysm"]),
            (dictgen::InsensitiveStr::Ascii("adapating"), &["adapting"]),
            (dictgen::InsensitiveStr::Ascii("disguss"), &["disgusts"]),
            (
                dictgen::InsensitiveStr::Ascii("contribtors"),
                &["contributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("righetousness"),
                &["righteousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("rutinely"), &["routinely"]),
            (dictgen::InsensitiveStr::Ascii("ongly"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("aprroximates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("sabatour"), &["saboteur"]),
            (
                dictgen::InsensitiveStr::Ascii("traids"),
                &["traits", "triads", "trades", "raids"],
            ),
            (dictgen::InsensitiveStr::Ascii("interprate"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("intestions"),
                &["intestines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descerning"),
                &["descending"],
            ),
            (dictgen::InsensitiveStr::Ascii("nescesary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("inbrace"), &["embrace"]),
            (dictgen::InsensitiveStr::Ascii("resurects"), &["resurrects"]),
            (dictgen::InsensitiveStr::Ascii("amphasis"), &["emphasis"]),
            (
                dictgen::InsensitiveStr::Ascii("annnouncers"),
                &["announcers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrobuts"),
                &["distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("mesure"), &["measure"]),
            (
                dictgen::InsensitiveStr::Ascii("nationallity"),
                &["nationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplenary"),
                &["disciplinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modfifications"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statuse"),
                &[
                    "statutes", "status", "statuses", "statues", "statute", "stature",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstraktly"),
                &["abstractly"],
            ),
            (dictgen::InsensitiveStr::Ascii("swown"), &["shown"]),
            (dictgen::InsensitiveStr::Ascii("termiated"), &["terminated"]),
            (
                dictgen::InsensitiveStr::Ascii("orieation"),
                &["orientation"],
            ),
            (dictgen::InsensitiveStr::Ascii("shtiless"), &["shitless"]),
            (
                dictgen::InsensitiveStr::Ascii("discrepency"),
                &["discrepancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularliser"),
                &["regulariser"],
            ),
            (dictgen::InsensitiveStr::Ascii("deapths"), &["depths"]),
            (dictgen::InsensitiveStr::Ascii("mordern"), &["modern"]),
            (dictgen::InsensitiveStr::Ascii("annonates"), &["annotates"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmiation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("chenged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("editores"), &["editors"]),
            (dictgen::InsensitiveStr::Ascii("fowards"), &["forwards"]),
            (dictgen::InsensitiveStr::Ascii("comback"), &["comeback"]),
            (
                dictgen::InsensitiveStr::Ascii("purcahsing"),
                &["purchasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("collektion"),
                &["collection"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplogise"), &["apologise"]),
            (dictgen::InsensitiveStr::Ascii("pireced"), &["pierced"]),
            (
                dictgen::InsensitiveStr::Ascii("initializaton"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("continuuum"), &["continuum"]),
            (dictgen::InsensitiveStr::Ascii("wdith"), &["width"]),
            (dictgen::InsensitiveStr::Ascii("defenitly"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("brillaintly"),
                &["brilliantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcotry"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("golbaly"), &["globally"]),
            (
                dictgen::InsensitiveStr::Ascii("sufficency"),
                &["sufficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convetionally"),
                &["conventionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substituition"),
                &["substitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propagaters"),
                &["propagators", "propagates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assest"),
                &["asset", "assess"],
            ),
            (dictgen::InsensitiveStr::Ascii("reposond"), &["respond"]),
            (dictgen::InsensitiveStr::Ascii("axpressed"), &["expressed"]),
            (dictgen::InsensitiveStr::Ascii("rewritin"), &["rewriting"]),
            (
                dictgen::InsensitiveStr::Ascii("doublin"),
                &["doubling", "dublin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underestemate"),
                &["underestimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraprhase"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paratheses"),
                &["parentheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envioremental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("radiane"), &["radiance"]),
            (
                dictgen::InsensitiveStr::Ascii("highalnder"),
                &["highlander"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standrdised"),
                &["standardised"],
            ),
            (dictgen::InsensitiveStr::Ascii("typcial"), &["typical"]),
            (dictgen::InsensitiveStr::Ascii("somethig"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("teahers"), &["teachers"]),
            (
                dictgen::InsensitiveStr::Ascii("archetypus"),
                &["archetypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("statrt"), &["start"]),
            (dictgen::InsensitiveStr::Ascii("coalleces"), &["coalesces"]),
            (dictgen::InsensitiveStr::Ascii("nightlcub"), &["nightclub"]),
            (dictgen::InsensitiveStr::Ascii("adbucted"), &["abducted"]),
            (dictgen::InsensitiveStr::Ascii("slowy"), &["slowly"]),
            (dictgen::InsensitiveStr::Ascii("thest"), &["test"]),
            (
                dictgen::InsensitiveStr::Ascii("conjugaging"),
                &["conjugating"],
            ),
            (dictgen::InsensitiveStr::Ascii("relly"), &["really"]),
            (
                dictgen::InsensitiveStr::Ascii("retramsmitting"),
                &["retransmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisicians"),
                &["statisticians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imcomprehensible"),
                &["incomprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("hundrets"), &["hundreds"]),
            (
                dictgen::InsensitiveStr::Ascii("compatitbility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("balckhawks"),
                &["blackhawks"],
            ),
            (dictgen::InsensitiveStr::Ascii("eextracted"), &["extracted"]),
            (dictgen::InsensitiveStr::Ascii("sentenals"), &["sentinels"]),
            (dictgen::InsensitiveStr::Ascii("freeezing"), &["freezing"]),
            (dictgen::InsensitiveStr::Ascii("beahviors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaes"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("loations"), &["locations"]),
            (dictgen::InsensitiveStr::Ascii("newanced"), &["nuanced"]),
            (
                dictgen::InsensitiveStr::Ascii("temparament"),
                &["temperament"],
            ),
            (dictgen::InsensitiveStr::Ascii("indrectly"), &["indirectly"]),
            (
                dictgen::InsensitiveStr::Ascii("accepptably"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motherborad"),
                &["motherboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("dinasour"), &["dinosaur"]),
            (
                dictgen::InsensitiveStr::Ascii("raidoactive"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("stutying"), &["studying"]),
            (
                dictgen::InsensitiveStr::Ascii("coyprighted"),
                &["copyrighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("seledtor"), &["selector"]),
            (
                dictgen::InsensitiveStr::Ascii("complexety"),
                &["complexity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitnationals"),
                &["multinational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persumably"),
                &["presumably"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyrhard"), &["tryhard"]),
            (dictgen::InsensitiveStr::Ascii("exlposion"), &["explosion"]),
            (
                dictgen::InsensitiveStr::Ascii("concervatism"),
                &["conservatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("tupple"), &["tuple"]),
            (
                dictgen::InsensitiveStr::Ascii("favilitators"),
                &["facilitators"],
            ),
            (dictgen::InsensitiveStr::Ascii("transates"), &["translates"]),
            (
                dictgen::InsensitiveStr::Ascii("downoloading"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalsitic"),
                &["nationalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demoninators"),
                &["denominators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shped"),
                &["shaped", "shipped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prgressing"),
                &["progressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representions"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("trinkst"), &["trinkets"]),
            (
                dictgen::InsensitiveStr::Ascii("legitimatcy"),
                &["legitimacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("everyonelse"),
                &["everyones"],
            ),
            (dictgen::InsensitiveStr::Ascii("actve"), &["active"]),
            (
                dictgen::InsensitiveStr::Ascii("interchangebly"),
                &["interchangeably"],
            ),
            (dictgen::InsensitiveStr::Ascii("notificated"), &["notified"]),
            (
                dictgen::InsensitiveStr::Ascii("draing"),
                &["drawing", "drain", "draining", "draping"],
            ),
            (dictgen::InsensitiveStr::Ascii("specyfing"), &["specifying"]),
            (
                dictgen::InsensitiveStr::Ascii("condional"),
                &["conditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sentaments"),
                &["sentiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpetion"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("discrimianted"),
                &["discriminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("evolveds"), &["evolves"]),
            (
                dictgen::InsensitiveStr::Ascii("unsurprizingly"),
                &["unsurprisingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloggeur"), &["blogger"]),
            (dictgen::InsensitiveStr::Ascii("simplicty"), &["simplicity"]),
            (dictgen::InsensitiveStr::Ascii("subtances"), &["substances"]),
            (dictgen::InsensitiveStr::Ascii("scorebord"), &["scoreboard"]),
            (
                dictgen::InsensitiveStr::Ascii("internationnalist"),
                &["internationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("ifinity"), &["infinity"]),
            (dictgen::InsensitiveStr::Ascii("arrriving"), &["arriving"]),
            (dictgen::InsensitiveStr::Ascii("algorihtm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("regualrs"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("arisoft"), &["airsoft"]),
            (dictgen::InsensitiveStr::Ascii("hollistic"), &["holistic"]),
            (
                dictgen::InsensitiveStr::Ascii("connvenient"),
                &["convenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("directtory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("simptom"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("overfowed"), &["overflowed"]),
            (dictgen::InsensitiveStr::Ascii("moduless"), &["modules"]),
            (
                dictgen::InsensitiveStr::Ascii("lifceycles"),
                &["lifecycles"],
            ),
            (dictgen::InsensitiveStr::Ascii("messsage"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("optomally"), &["optimally"]),
            (dictgen::InsensitiveStr::Ascii("allwoance"), &["allowance"]),
            (
                dictgen::InsensitiveStr::Ascii("bootleneck"),
                &["bottleneck"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underhwleming"),
                &["underwhelming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenitcator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("crosshiar"), &["crosshair"]),
            (dictgen::InsensitiveStr::Ascii("alhapet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("capatibly"),
                &["compatibly", "capably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convinently"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticle"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contentuous"),
                &["contentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evesdropped"),
                &["eavesdropped"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaitors"), &["animators"]),
            (dictgen::InsensitiveStr::Ascii("homeopaty"), &["homeopathy"]),
            (dictgen::InsensitiveStr::Ascii("enabeled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("lighweight"),
                &["lightweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentas"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistenty"),
                &["consistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("licesning"), &["licensing"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosincracy"),
                &["idiosyncrasy"],
            ),
            (dictgen::InsensitiveStr::Ascii("generatig"), &["generating"]),
            (
                dictgen::InsensitiveStr::Ascii("dstination"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("typ"), &["typo", "type"]),
            (dictgen::InsensitiveStr::Ascii("firmare"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("candidiate"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("cusine"), &["cuisine"]),
            (
                dictgen::InsensitiveStr::Ascii("constracting"),
                &["constructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("parargaph"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("overreides"), &["overrides"]),
            (
                dictgen::InsensitiveStr::Ascii("repalcements"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("enfoced"), &["enforced"]),
            (
                dictgen::InsensitiveStr::Ascii("signifcance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("raedme"), &["readme"]),
            (dictgen::InsensitiveStr::Ascii("masqurade"), &["masquerade"]),
            (
                dictgen::InsensitiveStr::Ascii("facedrwaing"),
                &["facedrawing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotiotation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availanility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orchastrated"),
                &["orchestrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeturn"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("asychronized"),
                &["asynchronized"],
            ),
            (dictgen::InsensitiveStr::Ascii("casheirs"), &["cashiers"]),
            (
                dictgen::InsensitiveStr::Ascii("paritioned"),
                &["partitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprensentational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("experties"), &["expertise"]),
            (dictgen::InsensitiveStr::Ascii("dependncy"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("expeimentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("excluse"),
                &["excludes", "exclude", "excuse", "exclusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procesed"),
                &["proceeds", "processed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exlude"),
                &["exclude", "exude"],
            ),
            (dictgen::InsensitiveStr::Ascii("scematics"), &["schematics"]),
            (
                dictgen::InsensitiveStr::Ascii("expementation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("excects"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("schems"),
                &["schemes", "schema", "schemas"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("refrerences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("encocders"), &["encoders"]),
            (
                dictgen::InsensitiveStr::Ascii("quanlified"),
                &["qualified", "quantified"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoping"), &["shopping"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetualy"),
                &["perpetually"],
            ),
            (dictgen::InsensitiveStr::Ascii("sampeld"), &["sampled"]),
            (
                dictgen::InsensitiveStr::Ascii("unresponcive"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("naturaly"), &["naturally"]),
            (dictgen::InsensitiveStr::Ascii("paragaraph"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("siver"),
                &["diver", "silver", "sliver"],
            ),
            (dictgen::InsensitiveStr::Ascii("cymptums"), &["symptoms"]),
            (
                dictgen::InsensitiveStr::Ascii("ressembled"),
                &["reassembled", "resembled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inded"),
                &["indeed", "index"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaising"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("leaading"), &["leading"]),
            (dictgen::InsensitiveStr::Ascii("upraded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("alighnment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("dissappointed"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("doumentc"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("chnaging"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("abbrevate"), &["abbreviate"]),
            (dictgen::InsensitiveStr::Ascii("beahviour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("ment"), &["meant"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabiliti"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("sourcin"), &["sourcing"]),
            (dictgen::InsensitiveStr::Ascii("mavrick"), &["maverick"]),
            (dictgen::InsensitiveStr::Ascii("cachup"), &["catchup"]),
            (dictgen::InsensitiveStr::Ascii("lanauages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("blcokchain"),
                &["blockchain"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeorhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acompaniment"),
                &["accompaniment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegatiotiations"),
                &["autonegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("displaybles"),
                &["displayables"],
            ),
            (dictgen::InsensitiveStr::Ascii("woill"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("unaivalable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harvestgain"),
                &["harvesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypocritial"),
                &["hypocritical"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesors"), &["accessors"]),
            (
                dictgen::InsensitiveStr::Ascii("inutitively"),
                &["intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valuses"),
                &["values", "valises"],
            ),
            (dictgen::InsensitiveStr::Ascii("builters"), &["builders"]),
            (dictgen::InsensitiveStr::Ascii("monitered"), &["monitored"]),
            (dictgen::InsensitiveStr::Ascii("tenacle"), &["tentacle"]),
            (
                dictgen::InsensitiveStr::Ascii("relavation"),
                &["relaxation"],
            ),
            (dictgen::InsensitiveStr::Ascii("puchasing"), &["purchasing"]),
            (
                dictgen::InsensitiveStr::Ascii("algorthnic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asycronous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulteanously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicateurs"),
                &["indicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraxting"),
                &["extracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("kurbick"), &["kubrick"]),
            (dictgen::InsensitiveStr::Ascii("simulatore"), &["simulate"]),
            (dictgen::InsensitiveStr::Ascii("jewerly"), &["jewelry"]),
            (dictgen::InsensitiveStr::Ascii("tcheckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("commandd"), &["commanded"]),
            (dictgen::InsensitiveStr::Ascii("pacthed"), &["patched"]),
            (
                dictgen::InsensitiveStr::Ascii("permanantly"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpararently"),
                &["transparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("periperhals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficififed"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("attributs"), &["attributes"]),
            (dictgen::InsensitiveStr::Ascii("tutrles"), &["turtles"]),
            (dictgen::InsensitiveStr::Ascii("unpopluar"), &["unpopular"]),
            (
                dictgen::InsensitiveStr::Ascii("coorespondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neurologia"),
                &["neurological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("minises"),
                &["minimises", "minuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readly"),
                &["readily", "ready"],
            ),
            (dictgen::InsensitiveStr::Ascii("degenere"), &["degenerate"]),
            (dictgen::InsensitiveStr::Ascii("shcooled"), &["schooled"]),
            (
                dictgen::InsensitiveStr::Ascii("annontation"),
                &["annotation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temptetion"),
                &["temptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fprmat"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("seleccts"), &["selects"]),
            (dictgen::InsensitiveStr::Ascii("susupend"), &["suspend"]),
            (
                dictgen::InsensitiveStr::Ascii("inititiaves"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("coommits"), &["commits"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplexd"),
                &["multiplexed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("steryotypical"),
                &["stereotypical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envioronment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("destryed"), &["destroyed"]),
            (dictgen::InsensitiveStr::Ascii("uests"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("expite"),
                &["expedite", "expire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requierement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsustainible"),
                &["unsustainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incremintd"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfering"), &["offering"]),
            (dictgen::InsensitiveStr::Ascii("comared"), &["compared"]),
            (dictgen::InsensitiveStr::Ascii("valudate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("appreicative"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalcualtes"),
                &["recalculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("hopkings"), &["hopkins"]),
            (dictgen::InsensitiveStr::Ascii("puposes"), &["purposes"]),
            (dictgen::InsensitiveStr::Ascii("effectice"), &["effective"]),
            (dictgen::InsensitiveStr::Ascii("purpuse"), &["purpose"]),
            (
                dictgen::InsensitiveStr::Ascii("monogymous"),
                &["monogamous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("longititude"),
                &["longitude"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("sccripts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("continuious"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonising"),
                &["synchronising"],
            ),
            (dictgen::InsensitiveStr::Ascii("acomplish"), &["accomplish"]),
            (dictgen::InsensitiveStr::Ascii("dynamcs"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("fullets"), &["fullest"]),
            (
                dictgen::InsensitiveStr::Ascii("regluarize"),
                &["regularize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializad"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activitions"),
                &["activations"],
            ),
            (dictgen::InsensitiveStr::Ascii("arriveis"), &["arrives"]),
            (dictgen::InsensitiveStr::Ascii("verteices"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("cholrine"), &["chlorine"]),
            (
                dictgen::InsensitiveStr::Ascii("contribues"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priviliged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("curreents"), &["currents"]),
            (dictgen::InsensitiveStr::Ascii("gnaking"), &["ganking"]),
            (dictgen::InsensitiveStr::Ascii("coallesce"), &["coalesce"]),
            (dictgen::InsensitiveStr::Ascii("stobes"), &["strobes"]),
            (
                dictgen::InsensitiveStr::Ascii("thannkfully"),
                &["thankfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("sprecially"), &["specially"]),
            (
                dictgen::InsensitiveStr::Ascii("intiializations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elementries"),
                &["elementaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("guardin"), &["guardian"]),
            (
                dictgen::InsensitiveStr::Ascii("comestical"),
                &["cosmetical"],
            ),
            (dictgen::InsensitiveStr::Ascii("enrtries"), &["entries"]),
            (
                dictgen::InsensitiveStr::Ascii("statisitcs"),
                &["statistics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aknowledgements"),
                &["acknowledgements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecxeptional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritptors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconsturct"),
                &["reconstruct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traceiver"),
                &["transceiver"],
            ),
            (dictgen::InsensitiveStr::Ascii("advantags"), &["advantages"]),
            (dictgen::InsensitiveStr::Ascii("widgect"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("stratgy"), &["strategy"]),
            (
                dictgen::InsensitiveStr::Ascii("depressies"),
                &["depressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("infantus"), &["infants"]),
            (dictgen::InsensitiveStr::Ascii("speeaking"), &["speaking"]),
            (dictgen::InsensitiveStr::Ascii("vluations"), &["valuations"]),
            (
                dictgen::InsensitiveStr::Ascii("ddress"),
                &["address", "dress"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resolutin"),
                &["resolutions", "resolution"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranlated"), &["translated"]),
            (dictgen::InsensitiveStr::Ascii("synegry"), &["synergy"]),
            (dictgen::InsensitiveStr::Ascii("cacahes"), &["caches"]),
            (dictgen::InsensitiveStr::Ascii("gutiars"), &["guitars"]),
            (
                dictgen::InsensitiveStr::Ascii("mysogonistic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nutrituous"),
                &["nutritious"],
            ),
            (dictgen::InsensitiveStr::Ascii("hemlholtz"), &["helmholtz"]),
            (dictgen::InsensitiveStr::Ascii("ategories"), &["categories"]),
            (dictgen::InsensitiveStr::Ascii("asssits"), &["assists"]),
            (dictgen::InsensitiveStr::Ascii("esccaped"), &["escaped"]),
            (
                dictgen::InsensitiveStr::Ascii("somethinfg"),
                &["somethings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entusiastically"),
                &["enthusiastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("worrrying"), &["worrying"]),
            (dictgen::InsensitiveStr::Ascii("possobly"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("arugable"), &["arguable"]),
            (
                dictgen::InsensitiveStr::Ascii("reqiurement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palyerbase"),
                &["playerbase"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyponsis"), &["hypnosis"]),
            (
                dictgen::InsensitiveStr::Ascii("notifiation"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("rety"), &["retry"]),
            (dictgen::InsensitiveStr::Ascii("effet"), &["effect"]),
            (dictgen::InsensitiveStr::Ascii("coalesing"), &["coalescing"]),
            (dictgen::InsensitiveStr::Ascii("enytime"), &["anytime"]),
            (dictgen::InsensitiveStr::Ascii("dismantel"), &["dismantle"]),
            (dictgen::InsensitiveStr::Ascii("chirstmas"), &["christmas"]),
            (dictgen::InsensitiveStr::Ascii("trigered"), &["triggered"]),
            (
                dictgen::InsensitiveStr::Ascii("canonalize"),
                &["canonicalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicslosing"),
                &["disclosing"],
            ),
            (dictgen::InsensitiveStr::Ascii("parametes"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("terminiator"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxamated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphetamenes"),
                &["amphetamines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("greedly"),
                &["greedily", "greedy"],
            ),
            (dictgen::InsensitiveStr::Ascii("srirachra"), &["sriracha"]),
            (dictgen::InsensitiveStr::Ascii("plasitcs"), &["plastics"]),
            (dictgen::InsensitiveStr::Ascii("rplace"), &["replace"]),
            (dictgen::InsensitiveStr::Ascii("anyonse"), &["anyones"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomanses"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("halucination"),
                &["hallucination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strucutral"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathimaticians"),
                &["mathematicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("viwport"), &["viewport"]),
            (dictgen::InsensitiveStr::Ascii("argumetn"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("temproay"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("becaus"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("fidnings"), &["findings"]),
            (dictgen::InsensitiveStr::Ascii("sinse"), &["sines", "since"]),
            (
                dictgen::InsensitiveStr::Ascii("retranmission"),
                &["retransmission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projetions"),
                &["projections"],
            ),
            (dictgen::InsensitiveStr::Ascii("expresed"), &["expressed"]),
            (dictgen::InsensitiveStr::Ascii("releatd"), &["related"]),
            (dictgen::InsensitiveStr::Ascii("fruther"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("exposse"), &["exposes"]),
            (
                dictgen::InsensitiveStr::Ascii("propabilities"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sereverless"),
                &["serverless"],
            ),
            (dictgen::InsensitiveStr::Ascii("diemsion"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("adquisitions"),
                &["acquisitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stategized"),
                &["strategized"],
            ),
            (dictgen::InsensitiveStr::Ascii("apihelions"), &["aphelions"]),
            (
                dictgen::InsensitiveStr::Ascii("defintaley"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifigation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millimiter"),
                &["millimeter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partcipants"),
                &["participants"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulf"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("yoesmite"), &["yosemite"]),
            (
                dictgen::InsensitiveStr::Ascii("classifcations"),
                &["classifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("susinct"), &["succinct"]),
            (
                dictgen::InsensitiveStr::Ascii("distinations"),
                &["distinctions", "destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polarisaiton"),
                &["polarisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasitioning"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isdefinitely"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countoring"),
                &["countering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dashbaords"),
                &["dashboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("bettter"), &["better"]),
            (dictgen::InsensitiveStr::Ascii("generlise"), &["generalise"]),
            (
                dictgen::InsensitiveStr::Ascii("iminent"),
                &["eminent", "imminent", "immanent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ideolouges"),
                &["ideologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("runned"),
                &["ran", "run", "ruined"],
            ),
            (dictgen::InsensitiveStr::Ascii("chhange"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("rapresented"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("replaceing"), &["replacing"]),
            (dictgen::InsensitiveStr::Ascii("manaully"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("diganostic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepard"), &["shepherd"]),
            (
                dictgen::InsensitiveStr::Ascii("clodes"),
                &["closed", "clothes"],
            ),
            (dictgen::InsensitiveStr::Ascii("hierarchjy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("supporession"),
                &["suppression"],
            ),
            (dictgen::InsensitiveStr::Ascii("swithing"), &["switching"]),
            (
                dictgen::InsensitiveStr::Ascii("subscribbed"),
                &["subscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idiosynchrasies"),
                &["idiosyncrasies"],
            ),
            (dictgen::InsensitiveStr::Ascii("lingiere"), &["lingerie"]),
            (dictgen::InsensitiveStr::Ascii("dynamicus"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("optimizied"),
                &["optimize", "optimized"],
            ),
            (dictgen::InsensitiveStr::Ascii("puplar"), &["popular"]),
            (
                dictgen::InsensitiveStr::Ascii("environemental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("recending"), &["rescinding"]),
            (
                dictgen::InsensitiveStr::Ascii("naerly"),
                &["nearly", "gnarly"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounadry"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("thurver"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("bootles"), &["bottles"]),
            (dictgen::InsensitiveStr::Ascii("thounsands"), &["thousands"]),
            (dictgen::InsensitiveStr::Ascii("collasped"), &["collapsed"]),
            (dictgen::InsensitiveStr::Ascii("expaneds"), &["expands"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphyiscal"),
                &["metaphysical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standartise"),
                &["standardise"],
            ),
            (dictgen::InsensitiveStr::Ascii("lectureres"), &["lectures"]),
            (dictgen::InsensitiveStr::Ascii("btiches"), &["bitches"]),
            (dictgen::InsensitiveStr::Ascii("reqroduce"), &["reproduce"]),
            (dictgen::InsensitiveStr::Ascii("recogize"), &["recognize"]),
            (
                dictgen::InsensitiveStr::Ascii("diappeares"),
                &["disappears"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingerit"), &["inherit"]),
            (
                dictgen::InsensitiveStr::Ascii("carbohyrdate"),
                &["carbohydrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciallly"), &["specially"]),
            (
                dictgen::InsensitiveStr::Ascii("mdification"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirorrs"), &["mirrors"]),
            (dictgen::InsensitiveStr::Ascii("sortnr"), &["sorter"]),
            (
                dictgen::InsensitiveStr::Ascii("analiticaly"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimzed"), &["minimized"]),
            (dictgen::InsensitiveStr::Ascii("netscpe"), &["netscape"]),
            (
                dictgen::InsensitiveStr::Ascii("tansactions"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partisioned"),
                &["partitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("animeted"), &["animated"]),
            (
                dictgen::InsensitiveStr::Ascii("perferance"),
                &["preference"],
            ),
            (dictgen::InsensitiveStr::Ascii("entrace"), &["entrance"]),
            (
                dictgen::InsensitiveStr::Ascii("distinquish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepatae"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("abutt"), &["abut"]),
            (dictgen::InsensitiveStr::Ascii("ignorts"), &["ignores"]),
            (
                dictgen::InsensitiveStr::Ascii("outher"),
                &["other", "outer"],
            ),
            (dictgen::InsensitiveStr::Ascii("temselves"), &["themselves"]),
            (dictgen::InsensitiveStr::Ascii("symmety"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("obhectifies"),
                &["objectifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("aassign"), &["assign"]),
            (dictgen::InsensitiveStr::Ascii("istened"), &["listened"]),
            (
                dictgen::InsensitiveStr::Ascii("reperesentation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cartiledge"), &["cartilage"]),
            (
                dictgen::InsensitiveStr::Ascii("multipleies"),
                &["multiples"],
            ),
            (dictgen::InsensitiveStr::Ascii("qurter"), &["quarter"]),
            (
                dictgen::InsensitiveStr::Ascii("themperature"),
                &["temperature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boyfirends"),
                &["boyfriends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contets"),
                &["contents", "contexts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benerate"),
                &["generate", "venerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaidate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("virutalized"),
                &["virtualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("rturn"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("obliderated"),
                &["obliterated"],
            ),
            (dictgen::InsensitiveStr::Ascii("distictly"), &["distinctly"]),
            (
                dictgen::InsensitiveStr::Ascii("applicible"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacups"), &["backups"]),
            (
                dictgen::InsensitiveStr::Ascii("indefintiely"),
                &["indefinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("contiguos"), &["contiguous"]),
            (dictgen::InsensitiveStr::Ascii("pirric"), &["pyrrhic"]),
            (
                dictgen::InsensitiveStr::Ascii("enviormental"),
                &["environmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cockroackes"),
                &["cockroaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("cognitivie"), &["cognitive"]),
            (
                dictgen::InsensitiveStr::Ascii("recrooters"),
                &["recruiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleareance"), &["clearance"]),
            (dictgen::InsensitiveStr::Ascii("discosure"), &["discourse"]),
            (
                dictgen::InsensitiveStr::Ascii("progressvie"),
                &["progressives"],
            ),
            (dictgen::InsensitiveStr::Ascii("confirgure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("ingreediants"),
                &["ingredients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("behavriours"),
                &["behaviours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associeations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressult"), &["result"]),
            (dictgen::InsensitiveStr::Ascii("elevatin"), &["elevation"]),
            (dictgen::InsensitiveStr::Ascii("triggerd"), &["triggered"]),
            (dictgen::InsensitiveStr::Ascii("naybhor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("regartless"),
                &["regardless"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahnges"), &["changes"]),
            (
                dictgen::InsensitiveStr::Ascii("configiring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apostrobphies"),
                &["apostrophe"],
            ),
            (dictgen::InsensitiveStr::Ascii("textture"), &["texture"]),
            (dictgen::InsensitiveStr::Ascii("intiials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("excensions"),
                &["extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanldes"), &["handles"]),
            (
                dictgen::InsensitiveStr::Ascii("sasketchawan"),
                &["saskatchewan"],
            ),
            (dictgen::InsensitiveStr::Ascii("prioryty"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("acused"),
                &["accused", "caused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overriddden"),
                &["overridden"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dellocates"),
                &["deallocates"],
            ),
            (dictgen::InsensitiveStr::Ascii("perferr"), &["prefer"]),
            (
                dictgen::InsensitiveStr::Ascii("absoliutely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("spiltting"), &["splitting"]),
            (
                dictgen::InsensitiveStr::Ascii("paschurisation"),
                &["pasteurisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("faireness"), &["fairness"]),
            (dictgen::InsensitiveStr::Ascii("azma"), &["asthma"]),
            (dictgen::InsensitiveStr::Ascii("reconise"), &["recognise"]),
            (dictgen::InsensitiveStr::Ascii("afected"), &["affected"]),
            (dictgen::InsensitiveStr::Ascii("separeted"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("boundin"), &["bounding"]),
            (
                dictgen::InsensitiveStr::Ascii("differenty"),
                &["differently"],
            ),
            (dictgen::InsensitiveStr::Ascii("criteriom"), &["criterion"]),
            (
                dictgen::InsensitiveStr::Ascii("signifigance"),
                &["significance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprenting"),
                &["repenting", "representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("asteriscs"), &["asterisks"]),
            (dictgen::InsensitiveStr::Ascii("oprimized"), &["optimized"]),
            (dictgen::InsensitiveStr::Ascii("befow"), &["below"]),
            (dictgen::InsensitiveStr::Ascii("enougth"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("abrreviating"),
                &["abbreviating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kuberentes"),
                &["kubernetes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poseesions"),
                &["possessions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappearing"),
                &["disappearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workarbounds"),
                &["workarounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quesrionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decremenst"),
                &["decrements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existantial"),
                &["existential"],
            ),
            (dictgen::InsensitiveStr::Ascii("capapcity"), &["capacity"]),
            (dictgen::InsensitiveStr::Ascii("folowwing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("adjacentcy"),
                &["adjacency", "adjacence"],
            ),
            (dictgen::InsensitiveStr::Ascii("hadler"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("vectoized"), &["vectorized"]),
            (dictgen::InsensitiveStr::Ascii("unquotting"), &["unquoting"]),
            (
                dictgen::InsensitiveStr::Ascii("delcaration"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("rqeuested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("trigging"), &["triggering"]),
            (
                dictgen::InsensitiveStr::Ascii("converseley"),
                &["conversely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unassgined"),
                &["unassigned"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulle"), &["bullet"]),
            (
                dictgen::InsensitiveStr::Ascii("incoherancy"),
                &["incoherency"],
            ),
            (dictgen::InsensitiveStr::Ascii("leegs"), &["leagues"]),
            (dictgen::InsensitiveStr::Ascii("releaeing"), &["releasing"]),
            (
                dictgen::InsensitiveStr::Ascii("sketcking"),
                &["skating", "sketching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discoruage"),
                &["discourages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disambiguiation"),
                &["disambiguation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gradiants"), &["gradients"]),
            (dictgen::InsensitiveStr::Ascii("sccessful"), &["successful"]),
            (
                dictgen::InsensitiveStr::Ascii("algoristhms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("implimcit"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("vaccins"), &["vaccines"]),
            (
                dictgen::InsensitiveStr::Ascii("reinitiailize"),
                &["reinitialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstracer"), &["abstracter"]),
            (dictgen::InsensitiveStr::Ascii("hapiness"), &["happiness"]),
            (dictgen::InsensitiveStr::Ascii("hiesman"), &["heisman"]),
            (dictgen::InsensitiveStr::Ascii("sharraid"), &["charade"]),
            (
                dictgen::InsensitiveStr::Ascii("transtormed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepecifed"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("artifcacts"), &["artifacts"]),
            (dictgen::InsensitiveStr::Ascii("changings"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("defaoult"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("forresst"), &["forrest"]),
            (dictgen::InsensitiveStr::Ascii("instroduce"), &["introduce"]),
            (
                dictgen::InsensitiveStr::Ascii("europeaners"),
                &["europeans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recipercally"),
                &["reciprocally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphetimins"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("licate"), &["locate"]),
            (dictgen::InsensitiveStr::Ascii("equipt"), &["equipped"]),
            (dictgen::InsensitiveStr::Ascii("wihtout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("resorce"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("apealing"), &["appealing"]),
            (dictgen::InsensitiveStr::Ascii("suspicous"), &["suspicious"]),
            (
                dictgen::InsensitiveStr::Ascii("suggesttion"),
                &["suggestion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guie"),
                &["guide", "guile", "guise"],
            ),
            (dictgen::InsensitiveStr::Ascii("headders"), &["headers"]),
            (dictgen::InsensitiveStr::Ascii("begginnig"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("iniated"), &["initiated"]),
            (
                dictgen::InsensitiveStr::Ascii("realtions"),
                &["relations", "reactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agrandizing"),
                &["aggrandizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("framworks"), &["frameworks"]),
            (
                dictgen::InsensitiveStr::Ascii("exprimenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("docucument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitites"),
                &["prostitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proffesional"),
                &["professional"],
            ),
            (dictgen::InsensitiveStr::Ascii("trunacte"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("understan"), &["understand"]),
            (
                dictgen::InsensitiveStr::Ascii("resposniveness"),
                &["responsiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("plattoes"), &["plateaus"]),
            (dictgen::InsensitiveStr::Ascii("dinghys"), &["dinghies"]),
            (
                dictgen::InsensitiveStr::Ascii("pretensious"),
                &["pretentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ivalidates"),
                &["invalidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisitc"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siving"),
                &["saving", "sieving"],
            ),
            (dictgen::InsensitiveStr::Ascii("succede"), &["succeed"]),
            (dictgen::InsensitiveStr::Ascii("morphlogy"), &["morphology"]),
            (dictgen::InsensitiveStr::Ascii("ardvark"), &["aardvark"]),
            (dictgen::InsensitiveStr::Ascii("compartor"), &["comparator"]),
            (
                dictgen::InsensitiveStr::Ascii("uability"),
                &["usability", "ability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prodcutions"),
                &["productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tripeld"), &["tripled"]),
            (
                dictgen::InsensitiveStr::Ascii("subscritpion"),
                &["subscriptions", "subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearches"),
                &["searches", "researches", "reaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("tanenhill"), &["tannehill"]),
            (
                dictgen::InsensitiveStr::Ascii("vegetariens"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloowind"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("pertubated"), &["perturbed"]),
            (dictgen::InsensitiveStr::Ascii("regarless"), &["regardless"]),
            (dictgen::InsensitiveStr::Ascii("anedocte"), &["anecdote"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphotically"),
                &["metaphorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("spported"), &["supported"]),
            (dictgen::InsensitiveStr::Ascii("inculdes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("trancations"),
                &["truncations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterizaton"),
                &["characterization"],
            ),
            (dictgen::InsensitiveStr::Ascii("martail"), &["martial"]),
            (
                dictgen::InsensitiveStr::Ascii("moistutizer"),
                &["moisturizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chornicles"),
                &["chronicles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clene"),
                &["clean", "cleaned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apporpriates"),
                &["appropriates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prevalente"),
                &["prevalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternetive"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indiffernt"),
                &["indifferent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spirituallity"),
                &["spiritually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applyications"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("serveices"), &["services"]),
            (dictgen::InsensitiveStr::Ascii("backoor"), &["backdoor"]),
            (
                dictgen::InsensitiveStr::Ascii("actiavting"),
                &["activating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocaiting"),
                &["allocating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sanwiching"),
                &["sandwiching"],
            ),
            (dictgen::InsensitiveStr::Ascii("rigths"), &["rights"]),
            (dictgen::InsensitiveStr::Ascii("expetedly"), &["expectedly"]),
            (
                dictgen::InsensitiveStr::Ascii("indciations"),
                &["indications"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuhmbnail"), &["thumbnail"]),
            (
                dictgen::InsensitiveStr::Ascii("repetive"),
                &["repetitive", "receptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("inmediate"), &["immediate"]),
            (
                dictgen::InsensitiveStr::Ascii("preduction"),
                &["prediction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("potentally"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("variblae"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("websit"), &["website"]),
            (
                dictgen::InsensitiveStr::Ascii("commnities"),
                &["communities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generatin"),
                &["generating", "generations", "generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifized"),
                &["sacrificed"],
            ),
            (dictgen::InsensitiveStr::Ascii("omlettes"), &["omelettes"]),
            (dictgen::InsensitiveStr::Ascii("strippin"), &["stripping"]),
            (dictgen::InsensitiveStr::Ascii("sampleing"), &["sampling"]),
            (
                dictgen::InsensitiveStr::Ascii("evolutionairy"),
                &["evolutionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derivitivs"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("recretes"), &["recreates"]),
            (dictgen::InsensitiveStr::Ascii("updtates"), &["updates"]),
            (
                dictgen::InsensitiveStr::Ascii("uninstallled"),
                &["uninstalled"],
            ),
            (dictgen::InsensitiveStr::Ascii("filname"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("progesssive"),
                &["progressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("corutine"), &["coroutine"]),
            (dictgen::InsensitiveStr::Ascii("confunsing"), &["confusing"]),
            (dictgen::InsensitiveStr::Ascii("tryinng"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("succeessors"),
                &["successors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hopelessley"),
                &["hopelessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressoin"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adoptor"),
                &["adopter", "adaptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("republicanos"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("foloowiong"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("ontainers"), &["containers"]),
            (
                dictgen::InsensitiveStr::Ascii("approxiamtes"),
                &["approximates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modesettting"),
                &["modesetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promiscuious"),
                &["promiscuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypenated"), &["hyphenated"]),
            (
                dictgen::InsensitiveStr::Ascii("photograps"),
                &["photographs"],
            ),
            (dictgen::InsensitiveStr::Ascii("pipline"), &["pipeline"]),
            (
                dictgen::InsensitiveStr::Ascii("identidier"),
                &["identifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictioanry"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("buit"), &["built"]),
            (
                dictgen::InsensitiveStr::Ascii("delegaters"),
                &["delegators", "delegates"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffr"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("checckerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadvantge"),
                &["disadvantage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequites"),
                &["prerequisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convices"),
                &["convinces", "convicts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordinare"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("stricteir"), &["stricter"]),
            (dictgen::InsensitiveStr::Ascii("disection"), &["dissection"]),
            (dictgen::InsensitiveStr::Ascii("malyasian"), &["malaysian"]),
            (dictgen::InsensitiveStr::Ascii("hiena"), &["hyena"]),
            (
                dictgen::InsensitiveStr::Ascii("vegtarians"),
                &["vegetarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("seemless"), &["seamless"]),
            (dictgen::InsensitiveStr::Ascii("lengtext"), &["longtext"]),
            (
                dictgen::InsensitiveStr::Ascii("articuled"),
                &["articulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("disporves"), &["disproves"]),
            (dictgen::InsensitiveStr::Ascii("exeuctor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("censability"),
                &["sensibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigstions"),
                &["investigations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communicae"),
                &["communicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polymomial"),
                &["polynomial"],
            ),
            (dictgen::InsensitiveStr::Ascii("widjet"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("delerious"), &["delirious"]),
            (dictgen::InsensitiveStr::Ascii("sycned"), &["synced"]),
            (
                dictgen::InsensitiveStr::Ascii("rockerfeller"),
                &["rockefeller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeffients"),
                &["coefficients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructued"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("differend"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("serching"), &["searching"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considertaion"),
                &["considerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("classifers"),
                &["classifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("vassalos"), &["vassals"]),
            (dictgen::InsensitiveStr::Ascii("trasnmits"), &["transmits"]),
            (dictgen::InsensitiveStr::Ascii("reletive"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("corrilates"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("singlin"), &["singling"]),
            (
                dictgen::InsensitiveStr::Ascii("recquired"),
                &["required", "reacquired"],
            ),
            (dictgen::InsensitiveStr::Ascii("bouncin"), &["bouncing"]),
            (
                dictgen::InsensitiveStr::Ascii("temptating"),
                &["temptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("manfiests"), &["manifests"]),
            (
                dictgen::InsensitiveStr::Ascii("expectantcy"),
                &["expectancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peerowetes"),
                &["pirouettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinging"),
                &["determining"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsitent"), &["existent"]),
            (
                dictgen::InsensitiveStr::Ascii("disrispect"),
                &["disrespect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derivitaves"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("depracted"), &["deprecated"]),
            (dictgen::InsensitiveStr::Ascii("steriods"), &["steroids"]),
            (
                dictgen::InsensitiveStr::Ascii("istruction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subconsicously"),
                &["subconsciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parititioner"),
                &["partitioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("endpdoint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("undorder"), &["unorder"]),
            (
                dictgen::InsensitiveStr::Ascii("cerifications"),
                &["certifications", "verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("edditable"), &["editable"]),
            (dictgen::InsensitiveStr::Ascii("preprares"), &["prepares"]),
            (
                dictgen::InsensitiveStr::Ascii("efining"),
                &["defining", "refining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configutation"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("naturual"), &["natural"]),
            (dictgen::InsensitiveStr::Ascii("subtetly"), &["subtlety"]),
            (
                dictgen::InsensitiveStr::Ascii("efficively"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoducibly"),
                &["reproducibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("enveloppe"), &["envelope"]),
            (
                dictgen::InsensitiveStr::Ascii("mutiplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("congitive"), &["cognitive"]),
            (
                dictgen::InsensitiveStr::Ascii("lements"),
                &["elements", "laments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introdue"),
                &["introduces", "introduce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socioeconimic"),
                &["socioeconomic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affilations"),
                &["affiliations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investigatons"),
                &["investigations"],
            ),
            (dictgen::InsensitiveStr::Ascii("definltey"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("daugther"), &["daughter"]),
            (dictgen::InsensitiveStr::Ascii("compagny"), &["company"]),
            (
                dictgen::InsensitiveStr::Ascii("durabillity"),
                &["durability"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyrany"), &["tyranny"]),
            (
                dictgen::InsensitiveStr::Ascii("cardiovasculaires"),
                &["cardiovascular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conside"),
                &["coincide", "consider"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definiteyl"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crocodiller"),
                &["crocodile"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepending"), &["depending"]),
            (dictgen::InsensitiveStr::Ascii("rised"), &["raised", "rose"]),
            (dictgen::InsensitiveStr::Ascii("startted"), &["started"]),
            (
                dictgen::InsensitiveStr::Ascii("pasturization"),
                &["pasteurization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commanted"),
                &["commanded", "commented"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceated"), &["created"]),
            (dictgen::InsensitiveStr::Ascii("labaeling"), &["labeling"]),
            (dictgen::InsensitiveStr::Ascii("rutgerus"), &["rutgers"]),
            (
                dictgen::InsensitiveStr::Ascii("priortised"),
                &["prioritised"],
            ),
            (dictgen::InsensitiveStr::Ascii("qeue"), &["queue"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstrationens"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propotion"),
                &["proportion", "promotion"],
            ),
            (dictgen::InsensitiveStr::Ascii("iceforg"), &["icefrog"]),
            (dictgen::InsensitiveStr::Ascii("opetional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("justfication"),
                &["justification"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirroed"), &["mirrored"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoptionsotiable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inequalitiy"),
                &["inequality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomplishents"),
                &["accomplishes", "accomplishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("randomus"), &["randoms"]),
            (
                dictgen::InsensitiveStr::Ascii("structurel"),
                &["structural"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumnt"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("devrice"),
                &["device", "dervice"],
            ),
            (dictgen::InsensitiveStr::Ascii("admisible"), &["admissible"]),
            (
                dictgen::InsensitiveStr::Ascii("agrocultural"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chromozomes"),
                &["chromosomes"],
            ),
            (dictgen::InsensitiveStr::Ascii("niusance"), &["nuisance"]),
            (
                dictgen::InsensitiveStr::Ascii("randumness"),
                &["randomness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initliases"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("eveidence"), &["evidence"]),
            (dictgen::InsensitiveStr::Ascii("buffere"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("uggly"), &["ugly"]),
            (
                dictgen::InsensitiveStr::Ascii("provisoned"),
                &["provisioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("initilials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenationa"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipulare"),
                &["manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginially"),
                &["originally"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpyical"), &["typical"]),
            (
                dictgen::InsensitiveStr::Ascii("deverlopers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applicablility"),
                &["applicability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extrapoliations"),
                &["extrapolations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relpacement"),
                &["replacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insatllation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infinitley"),
                &["infinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dpending"),
                &["depending", "spending"],
            ),
            (dictgen::InsensitiveStr::Ascii("publishrs"), &["publishers"]),
            (dictgen::InsensitiveStr::Ascii("euphroia"), &["euphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("avialibility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integreated"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("arcvhive"), &["archive"]),
            (dictgen::InsensitiveStr::Ascii("cleandars"), &["calendars"]),
            (
                dictgen::InsensitiveStr::Ascii("colunn"),
                &["column", "colon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiplicites"),
                &["multiplicities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("califorinan"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convetional"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("thye"), &["they"]),
            (
                dictgen::InsensitiveStr::Ascii("themsevles"),
                &["themselves"],
            ),
            (dictgen::InsensitiveStr::Ascii("sociopats"), &["sociopaths"]),
            (dictgen::InsensitiveStr::Ascii("segmeneted"), &["segmented"]),
            (
                dictgen::InsensitiveStr::Ascii("calssifies"),
                &["classifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innocenters"),
                &["innocents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respresentational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirrord"), &["mirrored"]),
            (dictgen::InsensitiveStr::Ascii("floursih"), &["flourish"]),
            (
                dictgen::InsensitiveStr::Ascii("eptrapolated"),
                &["extrapolated"],
            ),
            (dictgen::InsensitiveStr::Ascii("dployment"), &["deployment"]),
            (dictgen::InsensitiveStr::Ascii("casim"), &["chasm"]),
            (
                dictgen::InsensitiveStr::Ascii("temporarilty"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compabillities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("argumentate"),
                &["argumentative", "argue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emotionella"),
                &["emotionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("alow"), &["allow"]),
            (dictgen::InsensitiveStr::Ascii("startegy"), &["strategy"]),
            (
                dictgen::InsensitiveStr::Ascii("totalitatian"),
                &["totalitarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("remnandts"), &["remnants"]),
            (dictgen::InsensitiveStr::Ascii("exlainer"), &["explainer"]),
            (dictgen::InsensitiveStr::Ascii("skimrish"), &["skirmish"]),
            (
                dictgen::InsensitiveStr::Ascii("hankerchief"),
                &["handkerchief"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcpondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionalties"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("deom"), &["demo"]),
            (dictgen::InsensitiveStr::Ascii("integarte"), &["integrate"]),
            (dictgen::InsensitiveStr::Ascii("warniongs"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("squirrells"), &["squirrels"]),
            (dictgen::InsensitiveStr::Ascii("deattached"), &["detached"]),
            (
                dictgen::InsensitiveStr::Ascii("casesensetive"),
                &["casesensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expantions"),
                &["expansions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endoctrination"),
                &["indoctrination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instruktion"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepositons"),
                &["prepositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("practisioners"),
                &["practitioners"],
            ),
            (dictgen::InsensitiveStr::Ascii("compexes"), &["complexes"]),
            (dictgen::InsensitiveStr::Ascii("antiquae"), &["antique"]),
            (dictgen::InsensitiveStr::Ascii("usses"), &["uses"]),
            (dictgen::InsensitiveStr::Ascii("enaugh"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaucratics"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("evidencd"), &["evidenced"]),
            (dictgen::InsensitiveStr::Ascii("stingent"), &["stringent"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangment"),
                &["rearrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perliferated"),
                &["proliferated"],
            ),
            (dictgen::InsensitiveStr::Ascii("evauluates"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("circomvents"),
                &["circumvents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intertainment"),
                &["entertainment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stringifed"),
                &["stringified"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrfaces"), &["interfaces"]),
            (
                dictgen::InsensitiveStr::Ascii("occurately"),
                &["accurately"],
            ),
            (dictgen::InsensitiveStr::Ascii("laucnh"), &["launch"]),
            (
                dictgen::InsensitiveStr::Ascii("accusating"),
                &["accusation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decoritive"),
                &["decorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("preambule"), &["preamble"]),
            (dictgen::InsensitiveStr::Ascii("onself"), &["oneself"]),
            (
                dictgen::InsensitiveStr::Ascii("falshbacks"),
                &["flashbacks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influenser"),
                &["influences"],
            ),
            (dictgen::InsensitiveStr::Ascii("alised"), &["aliased"]),
            (
                dictgen::InsensitiveStr::Ascii("competution"),
                &["computation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfears"),
                &["interferes"],
            ),
            (dictgen::InsensitiveStr::Ascii("intquired"), &["inquired"]),
            (
                dictgen::InsensitiveStr::Ascii("milliseond"),
                &["millisecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("caclucate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("revealtions"),
                &["revelations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congugates"),
                &["conjugates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantatitive"),
                &["quantitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("faiways"), &["fairways"]),
            (dictgen::InsensitiveStr::Ascii("assertting"), &["asserting"]),
            (
                dictgen::InsensitiveStr::Ascii("confirmming"),
                &["confirming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("botstrapping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("emergancy"), &["emergency"]),
            (
                dictgen::InsensitiveStr::Ascii("intimidades"),
                &["intimidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bundeling"), &["bundling"]),
            (
                dictgen::InsensitiveStr::Ascii("pagents"),
                &["pageants", "plangents"],
            ),
            (dictgen::InsensitiveStr::Ascii("radaint"), &["radiant"]),
            (
                dictgen::InsensitiveStr::Ascii("esepecially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("sempahore"), &["semaphore"]),
            (dictgen::InsensitiveStr::Ascii("serveicing"), &["servicing"]),
            (
                dictgen::InsensitiveStr::Ascii("origonally"),
                &["originally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propagetes"),
                &["propagates"],
            ),
            (dictgen::InsensitiveStr::Ascii("jusified"), &["justified"]),
            (
                dictgen::InsensitiveStr::Ascii("limts"),
                &["limits", "lints"],
            ),
            (dictgen::InsensitiveStr::Ascii("meatballls"), &["meatballs"]),
            (
                dictgen::InsensitiveStr::Ascii("modifcications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("regiister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("apostrohpe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contiuation"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("leaast"), &["least"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaucratius"),
                &["bureaucrats"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predjuices"),
                &["prejudices"],
            ),
            (dictgen::InsensitiveStr::Ascii("pacifit"), &["pacifist"]),
            (
                dictgen::InsensitiveStr::Ascii("categrories"),
                &["categories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreprenuers"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprecates"),
                &["deprecates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomenen"),
                &["phenomenon"],
            ),
            (dictgen::InsensitiveStr::Ascii("eforts"), &["efforts"]),
            (dictgen::InsensitiveStr::Ascii("consit"), &["consist"]),
            (dictgen::InsensitiveStr::Ascii("channeles"), &["channels"]),
            (
                dictgen::InsensitiveStr::Ascii("specilizations"),
                &["specializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependencey"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeknowst"),
                &["unbeknownst"],
            ),
            (dictgen::InsensitiveStr::Ascii("mileau"), &["milieu"]),
            (dictgen::InsensitiveStr::Ascii("scolling"), &["scrolling"]),
            (dictgen::InsensitiveStr::Ascii("enaable"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("prediciting"),
                &["predicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inituialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("voilating"), &["violating"]),
            (dictgen::InsensitiveStr::Ascii("snipets"), &["snippets"]),
            (dictgen::InsensitiveStr::Ascii("checlers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("designato"),
                &["designation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensistive"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("areea"), &["area"]),
            (dictgen::InsensitiveStr::Ascii("connstrain"), &["constrain"]),
            (dictgen::InsensitiveStr::Ascii("intiators"), &["initiators"]),
            (dictgen::InsensitiveStr::Ascii("altas"), &["atlas"]),
            (
                dictgen::InsensitiveStr::Ascii("infectation"),
                &["infestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("annays"), &["annoys", "any"]),
            (dictgen::InsensitiveStr::Ascii("dollors"), &["dollars"]),
            (
                dictgen::InsensitiveStr::Ascii("creadentials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("sastisfies"), &["satisfies"]),
            (dictgen::InsensitiveStr::Ascii("prodcuers"), &["producers"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditially"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corprocessor"),
                &["coprocessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("tehsor"), &["tensor"]),
            (dictgen::InsensitiveStr::Ascii("moodules"), &["modules"]),
            (
                dictgen::InsensitiveStr::Ascii("generousity"),
                &["generosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("savety"), &["safety"]),
            (
                dictgen::InsensitiveStr::Ascii("execuitables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consituencies"),
                &["constituencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camouflauge"),
                &["camouflage"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecplicit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("appriximating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("purposal"), &["proposal"]),
            (
                dictgen::InsensitiveStr::Ascii("syclical"),
                &["cyclical", "physical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plagiarizin"),
                &["plagiarizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reporteros"), &["reporters"]),
            (
                dictgen::InsensitiveStr::Ascii("dealloacte"),
                &["deallocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecgted"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("cyllindrical"),
                &["cylindrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucsesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("landmarsk"), &["landmarks"]),
            (
                dictgen::InsensitiveStr::Ascii("hemicircles"),
                &["semicircles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecessarely"),
                &["unnecessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("malpractive"),
                &["malpractice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncomplete"),
                &["incomplete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monocrhome"),
                &["monochrome"],
            ),
            (dictgen::InsensitiveStr::Ascii("conacted"), &["contacted"]),
            (
                dictgen::InsensitiveStr::Ascii("behaviroal"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoinds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temperatrues"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("stablizes"), &["stabilizes"]),
            (dictgen::InsensitiveStr::Ascii("wednesdaay"), &["wednesday"]),
            (
                dictgen::InsensitiveStr::Ascii("organiztions"),
                &["organizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encaspulated"),
                &["encapsulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("compilled"), &["compiled"]),
            (
                dictgen::InsensitiveStr::Ascii("calcalation"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controvercial"),
                &["controversial"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotacts"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaition"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dangeroulsy"),
                &["dangerously"],
            ),
            (dictgen::InsensitiveStr::Ascii("electronix"), &["election"]),
            (
                dictgen::InsensitiveStr::Ascii("autocoreect"),
                &["autocorrect"],
            ),
            (dictgen::InsensitiveStr::Ascii("chalange"), &["challenge"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiatically"),
                &["enthusiastically"],
            ),
            (dictgen::InsensitiveStr::Ascii("analagus"), &["analogous"]),
            (dictgen::InsensitiveStr::Ascii("shortcat"), &["shortcut"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteraming"),
                &["downstreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdotice"), &["anecdote"]),
            (dictgen::InsensitiveStr::Ascii("connecte"), &["connected"]),
            (
                dictgen::InsensitiveStr::Ascii("ministery"),
                &["ministry", "minister"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrion"),
                &["arbitration"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefiic"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("representive"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("dionsaur"), &["dinosaur"]),
            (
                dictgen::InsensitiveStr::Ascii("indescriminent"),
                &["indiscriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalizers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("squence"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("intervencion"),
                &["intervention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotications"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorcing"),
                &["endorsing", "enforcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documetned"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("filps"), &["flips"]),
            (
                dictgen::InsensitiveStr::Ascii("reasonnable"),
                &["reasonable"],
            ),
            (dictgen::InsensitiveStr::Ascii("nadly"), &["badly"]),
            (
                dictgen::InsensitiveStr::Ascii("dysfunciton"),
                &["dysfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("sleect"), &["select"]),
            (dictgen::InsensitiveStr::Ascii("quwesants"), &["croissants"]),
            (dictgen::InsensitiveStr::Ascii("distaced"), &["distanced"]),
            (
                dictgen::InsensitiveStr::Ascii("successtion"),
                &["succession", "suggestion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("translateable"),
                &["translatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphyiscs"),
                &["metaphysics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucessfully"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicretion"), &["discretion"]),
            (
                dictgen::InsensitiveStr::Ascii("destinataion"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accpetance"),
                &["acceptance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recun"),
                &["reckon", "recon", "recur"],
            ),
            (dictgen::InsensitiveStr::Ascii("satart"), &["start"]),
            (dictgen::InsensitiveStr::Ascii("recruses"), &["recurses"]),
            (
                dictgen::InsensitiveStr::Ascii("concatonated"),
                &["concatenated"],
            ),
            (dictgen::InsensitiveStr::Ascii("meanign"), &["meaning"]),
            (
                dictgen::InsensitiveStr::Ascii("definatrly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("availabble"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferncences"),
                &["dereferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corprorations"),
                &["corporations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranditional"),
                &["transitional"],
            ),
            (dictgen::InsensitiveStr::Ascii("aksreddit"), &["askreddit"]),
            (dictgen::InsensitiveStr::Ascii("taggs"), &["tags"]),
            (
                dictgen::InsensitiveStr::Ascii("neoroscience"),
                &["neuroscience"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysys"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("afflcition"),
                &["affliction"],
            ),
            (dictgen::InsensitiveStr::Ascii("malicous"), &["malicious"]),
            (
                dictgen::InsensitiveStr::Ascii("interrtupions"),
                &["interruptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investmens"),
                &["investments"],
            ),
            (dictgen::InsensitiveStr::Ascii("sceletons"), &["skeletons"]),
            (
                dictgen::InsensitiveStr::Ascii("fundemantalist"),
                &["fundamentalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("apparenly"), &["apparently"]),
            (dictgen::InsensitiveStr::Ascii("intereset"), &["interest"]),
            (
                dictgen::InsensitiveStr::Ascii("tempelate"),
                &["template", "temperate"],
            ),
            (dictgen::InsensitiveStr::Ascii("brackeds"), &["brackets"]),
            (
                dictgen::InsensitiveStr::Ascii("comnplexities"),
                &["complexities"],
            ),
            (dictgen::InsensitiveStr::Ascii("succssor"), &["successor"]),
            (
                dictgen::InsensitiveStr::Ascii("cooedinating"),
                &["coordinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cmmanded"), &["commanded"]),
            (
                dictgen::InsensitiveStr::Ascii("situatinal"),
                &["situational"],
            ),
            (dictgen::InsensitiveStr::Ascii("scavanger"), &["scavenger"]),
            (dictgen::InsensitiveStr::Ascii("exciation"), &["excitation"]),
            (dictgen::InsensitiveStr::Ascii("sudents"), &["students"]),
            (dictgen::InsensitiveStr::Ascii("fointers"), &["pointers"]),
            (dictgen::InsensitiveStr::Ascii("warrriors"), &["warriors"]),
            (dictgen::InsensitiveStr::Ascii("resloved"), &["resolved"]),
            (
                dictgen::InsensitiveStr::Ascii("kernuls"),
                &["colonels", "kernels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intracacies"),
                &["intricacies"],
            ),
            (dictgen::InsensitiveStr::Ascii("decembre"), &["december"]),
            (dictgen::InsensitiveStr::Ascii("egytpians"), &["egyptians"]),
            (dictgen::InsensitiveStr::Ascii("packagd"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("peridical"), &["periodical"]),
            (dictgen::InsensitiveStr::Ascii("transfors"), &["transforms"]),
            (dictgen::InsensitiveStr::Ascii("folllowing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("internationl"),
                &["international"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aproximates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("raionale"), &["rationale"]),
            (dictgen::InsensitiveStr::Ascii("manues"), &["menus"]),
            (
                dictgen::InsensitiveStr::Ascii("replactes"),
                &["replaces", "replicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configared"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialzied"),
                &["serialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("famoust"), &["famous"]),
            (dictgen::InsensitiveStr::Ascii("wordl"), &["world"]),
            (
                dictgen::InsensitiveStr::Ascii("intelluctuals"),
                &["intellectuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parliamentiary"),
                &["parliamentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("gouped"), &["grouped"]),
            (
                dictgen::InsensitiveStr::Ascii("springfied"),
                &["springfield"],
            ),
            (dictgen::InsensitiveStr::Ascii("accapters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("chaeckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("geometrys"), &["geometries"]),
            (
                dictgen::InsensitiveStr::Ascii("persentages"),
                &["percentages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underraker"),
                &["undertaker"],
            ),
            (dictgen::InsensitiveStr::Ascii("reegion"), &["region"]),
            (
                dictgen::InsensitiveStr::Ascii("environmentt"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("govermental"),
                &["governmental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promiscus"),
                &["promiscuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srolled"),
                &["scrolled", "strolled"],
            ),
            (dictgen::InsensitiveStr::Ascii("closeley"), &["closely"]),
            (dictgen::InsensitiveStr::Ascii("teusday"), &["tuesday"]),
            (dictgen::InsensitiveStr::Ascii("relyable"), &["reliable"]),
            (
                dictgen::InsensitiveStr::Ascii("festure"),
                &["feature", "gesture"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpative"), &["adaptive"]),
            (dictgen::InsensitiveStr::Ascii("sunglases"), &["sunglasses"]),
            (
                dictgen::InsensitiveStr::Ascii("terrorities"),
                &["territories"],
            ),
            (dictgen::InsensitiveStr::Ascii("acclamied"), &["acclaimed"]),
            (dictgen::InsensitiveStr::Ascii("beigin"), &["begin"]),
            (dictgen::InsensitiveStr::Ascii("cthuluh"), &["cthulhu"]),
            (dictgen::InsensitiveStr::Ascii("upgarde"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("insurgance"),
                &["insurgency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copiles"),
                &["compiles", "copies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifaction"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boostrapper"),
                &["bootstrapper"],
            ),
            (dictgen::InsensitiveStr::Ascii("expcetion"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("acronyme"), &["acronym"]),
            (dictgen::InsensitiveStr::Ascii("alomst"), &["almost"]),
            (dictgen::InsensitiveStr::Ascii("nessassary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("encrpyted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("gerneration"),
                &["generation"],
            ),
            (dictgen::InsensitiveStr::Ascii("repsitory"), &["repository"]),
            (
                dictgen::InsensitiveStr::Ascii("confgiuration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearand"), &["operand"]),
            (
                dictgen::InsensitiveStr::Ascii("questoning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("exteder"), &["extender"]),
            (
                dictgen::InsensitiveStr::Ascii("simulteanous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrbes"), &["describes"]),
            (dictgen::InsensitiveStr::Ascii("campusers"), &["campuses"]),
            (dictgen::InsensitiveStr::Ascii("cnange"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("stawberries"),
                &["strawberries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrctness"),
                &["correctness"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmon"), &["common"]),
            (dictgen::InsensitiveStr::Ascii("effecitve"), &["effective"]),
            (
                dictgen::InsensitiveStr::Ascii("archetects"),
                &["architects"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawdow"), &["drawdown"]),
            (dictgen::InsensitiveStr::Ascii("classess"), &["classes"]),
            (dictgen::InsensitiveStr::Ascii("ridiculued"), &["ridicule"]),
            (dictgen::InsensitiveStr::Ascii("selets"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("laotions"),
                &["laotians", "lotions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterfait"),
                &["counterfeit"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepration"), &["separation"]),
            (
                dictgen::InsensitiveStr::Ascii("initialses"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descendats"),
                &["descendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revelaiton"),
                &["revelations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transitiosn"),
                &["transitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspicision"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("apoligist"), &["apologist"]),
            (dictgen::InsensitiveStr::Ascii("dillemas"), &["dilemmas"]),
            (
                dictgen::InsensitiveStr::Ascii("beleife"),
                &["belief", "believe"],
            ),
            (dictgen::InsensitiveStr::Ascii("vidoe"), &["video"]),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarated"),
                &["ameliorated"],
            ),
            (dictgen::InsensitiveStr::Ascii("secert"), &["secret"]),
            (
                dictgen::InsensitiveStr::Ascii("justication"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backpackng"),
                &["backpacking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactarization"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nessecarrily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("requiread"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("annotaes"), &["annotates"]),
            (dictgen::InsensitiveStr::Ascii("filcker"), &["flicker"]),
            (
                dictgen::InsensitiveStr::Ascii("portrating"),
                &["portraying"],
            ),
            (dictgen::InsensitiveStr::Ascii("suprized"), &["surprised"]),
            (dictgen::InsensitiveStr::Ascii("retreve"), &["retrieve"]),
            (dictgen::InsensitiveStr::Ascii("fractales"), &["fractals"]),
            (
                dictgen::InsensitiveStr::Ascii("temperatue"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("warninds"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("terrotorial"),
                &["territorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("sebsatian"), &["sebastian"]),
            (dictgen::InsensitiveStr::Ascii("createin"), &["creatine"]),
            (dictgen::InsensitiveStr::Ascii("minstry"), &["ministry"]),
            (
                dictgen::InsensitiveStr::Ascii("aknowledgment"),
                &["acknowledgment"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysiis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("palestenians"),
                &["palestinians"],
            ),
            (dictgen::InsensitiveStr::Ascii("blackbery"), &["blackberry"]),
            (dictgen::InsensitiveStr::Ascii("desiar"), &["desire"]),
            (dictgen::InsensitiveStr::Ascii("playstye"), &["playstyle"]),
            (
                dictgen::InsensitiveStr::Ascii("istantiation"),
                &["instantiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mellineum"), &["millennium"]),
            (dictgen::InsensitiveStr::Ascii("iternate"), &["iterate"]),
            (
                dictgen::InsensitiveStr::Ascii("decrmented"),
                &["decremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrevable"),
                &["retrievable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptiions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negogtiated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("erliear"), &["earlier"]),
            (
                dictgen::InsensitiveStr::Ascii("manafacturing"),
                &["manufacturing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repleacable"),
                &["replaceable"],
            ),
            (dictgen::InsensitiveStr::Ascii("deprectad"), &["deprecated"]),
            (dictgen::InsensitiveStr::Ascii("cynicisim"), &["cynicism"]),
            (
                dictgen::InsensitiveStr::Ascii("initialyzing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("simlified"), &["simplified"]),
            (dictgen::InsensitiveStr::Ascii("mutilatin"), &["mutilation"]),
            (
                dictgen::InsensitiveStr::Ascii("egenralises"),
                &["generalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expriments"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("primimitive"),
                &["primitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuilty"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approximing"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("flashligt"), &["flashlight"]),
            (
                dictgen::InsensitiveStr::Ascii("advertisting"),
                &["advertising"],
            ),
            (dictgen::InsensitiveStr::Ascii("indentended"), &["indented"]),
            (dictgen::InsensitiveStr::Ascii("fetaures"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("swich"),
                &["switch", "swish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regessing"),
                &["regressing", "recessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accessment"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propriotary"),
                &["proprietary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insenitively"),
                &["insensitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("mormalise"), &["normalise"]),
            (
                dictgen::InsensitiveStr::Ascii("assignmnet"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("symemetric"), &["symmetric"]),
            (
                dictgen::InsensitiveStr::Ascii("inplementation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("certanly"), &["certainly"]),
            (
                dictgen::InsensitiveStr::Ascii("approparite"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dditional"), &["additional"]),
            (dictgen::InsensitiveStr::Ascii("archtype"), &["archetype"]),
            (dictgen::InsensitiveStr::Ascii("memner"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("diferentiated"),
                &["differentiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("directrion"), &["direction"]),
            (
                dictgen::InsensitiveStr::Ascii("simptomatically"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorcer"),
                &["endorser", "enforcer"],
            ),
            (dictgen::InsensitiveStr::Ascii("issure"), &["issue"]),
            (dictgen::InsensitiveStr::Ascii("burjuns"), &["burgeons"]),
            (dictgen::InsensitiveStr::Ascii("feilds"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("responsably"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misproportionate"),
                &["disproportionate"],
            ),
            (dictgen::InsensitiveStr::Ascii("reallize"), &["realize"]),
            (dictgen::InsensitiveStr::Ascii("acceoter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("beofrehand"),
                &["beforehand"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigits"), &["digits"]),
            (
                dictgen::InsensitiveStr::Ascii("responisbly"),
                &["responsibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("importanly"),
                &["importantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("broblematic"),
                &["problematic"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifcies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("improvemen"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normallised"),
                &["normalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimantation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("quicklyu"), &["quickly"]),
            (
                dictgen::InsensitiveStr::Ascii("inevetable"),
                &["inevitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkesums"), &["checksums"]),
            (dictgen::InsensitiveStr::Ascii("visuabise"), &["visualise"]),
            (dictgen::InsensitiveStr::Ascii("exportes"), &["exports"]),
            (
                dictgen::InsensitiveStr::Ascii("diagrammin"),
                &["diagramming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsabilities"),
                &["responsibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("launchs"),
                &["launch", "launches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferrring"),
                &["preferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("thottling"), &["throttling"]),
            (dictgen::InsensitiveStr::Ascii("minnimum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("simliarly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("initiallizer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvienient"),
                &["inconvenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posioning"),
                &["poisoning", "positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parametere"),
                &["parameter", "parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("conetent"), &["content"]),
            (dictgen::InsensitiveStr::Ascii("amatuer"), &["amateur"]),
            (
                dictgen::InsensitiveStr::Ascii("austrailan"),
                &["australian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abnornally"),
                &["abnormally"],
            ),
            (dictgen::InsensitiveStr::Ascii("currenty"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("corelation"),
                &["correlation"],
            ),
            (dictgen::InsensitiveStr::Ascii("condolin"), &["condoling"]),
            (dictgen::InsensitiveStr::Ascii("wrkers"), &["workers"]),
            (dictgen::InsensitiveStr::Ascii("pahses"), &["phases"]),
            (dictgen::InsensitiveStr::Ascii("enhaning"), &["enhancing"]),
            (
                dictgen::InsensitiveStr::Ascii("compalined"),
                &["complained"],
            ),
            (dictgen::InsensitiveStr::Ascii("benged"), &["binged"]),
            (dictgen::InsensitiveStr::Ascii("signaturs"), &["signatures"]),
            (
                dictgen::InsensitiveStr::Ascii("dissapointd"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assesmenet"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mointers"),
                &["pointers", "mounters", "monitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("navagated"), &["navigated"]),
            (
                dictgen::InsensitiveStr::Ascii("desability"),
                &["disability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessariy"),
                &["necessary", "necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influecing"),
                &["influencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condifuring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sluaghtering"),
                &["slaughtering"],
            ),
            (dictgen::InsensitiveStr::Ascii("lucious"), &["luscious"]),
            (
                dictgen::InsensitiveStr::Ascii("meassurement"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("craching"),
                &["crashing", "caching", "cracking", "coaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciphersute"),
                &["ciphersuite"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximu"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("assemby"), &["assembly"]),
            (
                dictgen::InsensitiveStr::Ascii("cootdination"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicsriminated"),
                &["discriminated"],
            ),
            (dictgen::InsensitiveStr::Ascii("allong"), &["along"]),
            (
                dictgen::InsensitiveStr::Ascii("micorcodes"),
                &["microcodes"],
            ),
            (dictgen::InsensitiveStr::Ascii("missons"), &["missions"]),
            (dictgen::InsensitiveStr::Ascii("noiser"), &["noisier"]),
            (
                dictgen::InsensitiveStr::Ascii("infestato"),
                &["infestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("snaphost"), &["snapshot"]),
            (dictgen::InsensitiveStr::Ascii("coalescin"), &["coalescing"]),
            (
                dictgen::InsensitiveStr::Ascii("temproarily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pubilcation"),
                &["publication"],
            ),
            (dictgen::InsensitiveStr::Ascii("prusue"), &["pursue"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucesfull"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("pluse"), &["pulse", "plus"]),
            (dictgen::InsensitiveStr::Ascii("yoru"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("individauly"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("recroots"), &["recruits"]),
            (dictgen::InsensitiveStr::Ascii("ocntext"), &["context"]),
            (dictgen::InsensitiveStr::Ascii("allignment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("characteritiscs"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexperiente"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("expaneding"), &["expanding"]),
            (dictgen::InsensitiveStr::Ascii("quoshants"), &["quotients"]),
            (
                dictgen::InsensitiveStr::Ascii("oustanding"),
                &["outstanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accellerating"),
                &["accelerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impelenting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handkerchif"),
                &["handkerchief"],
            ),
            (dictgen::InsensitiveStr::Ascii("ayway"), &["anyway", "away"]),
            (dictgen::InsensitiveStr::Ascii("equeal"), &["equal"]),
            (
                dictgen::InsensitiveStr::Ascii("declarating"),
                &["declaring"],
            ),
            (dictgen::InsensitiveStr::Ascii("sparsly"), &["sparsely"]),
            (dictgen::InsensitiveStr::Ascii("vairants"), &["variants"]),
            (dictgen::InsensitiveStr::Ascii("balence"), &["balance"]),
            (
                dictgen::InsensitiveStr::Ascii("indiviuals"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("welp"), &["whelp"]),
            (dictgen::InsensitiveStr::Ascii("varian"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("rationnals"),
                &["rationales", "rationals"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifcat"), &["artifact"]),
            (
                dictgen::InsensitiveStr::Ascii("existenta"),
                &["existential"],
            ),
            (dictgen::InsensitiveStr::Ascii("scrtiped"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("streching"), &["stretching"]),
            (
                dictgen::InsensitiveStr::Ascii("premissible"),
                &["permissible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placmenets"),
                &["placements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("menbership"),
                &["membership"],
            ),
            (dictgen::InsensitiveStr::Ascii("initalled"), &["initialled"]),
            (dictgen::InsensitiveStr::Ascii("callaback"), &["callback"]),
            (dictgen::InsensitiveStr::Ascii("gramar"), &["grammar"]),
            (
                dictgen::InsensitiveStr::Ascii("reccurring"),
                &["recurring", "reoccurring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sives"),
                &["saves", "sieves"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cutomizers"),
                &["customizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyer"),
                &["analyser", "analyzer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersectoin"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrayes"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("enclosung"), &["enclosing"]),
            (
                dictgen::InsensitiveStr::Ascii("guarenteeded"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enteratinment"),
                &["entertainment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signficiant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algotrithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("refroming"), &["reforming"]),
            (dictgen::InsensitiveStr::Ascii("lesbianas"), &["lesbians"]),
            (
                dictgen::InsensitiveStr::Ascii("tokenizaing"),
                &["tokenizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("boungin"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("trigerrs"), &["triggers"]),
            (
                dictgen::InsensitiveStr::Ascii("cunterpart"),
                &["counterpart"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agumenting"),
                &["augmenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acomplished"),
                &["accomplished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interractions"),
                &["interactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("adhevise"), &["adhesive"]),
            (
                dictgen::InsensitiveStr::Ascii("collocweall"),
                &["colloquial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endoresment"),
                &["endorsement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manupulating"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("canclled"), &["cancelled"]),
            (
                dictgen::InsensitiveStr::Ascii("decremeantal"),
                &["decremental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horrendeously"),
                &["horrendously"],
            ),
            (dictgen::InsensitiveStr::Ascii("provisons"), &["provisions"]),
            (
                dictgen::InsensitiveStr::Ascii("executablble"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("imperfet"), &["imperfect"]),
            (
                dictgen::InsensitiveStr::Ascii("overwhemled"),
                &["overwhelmed"],
            ),
            (dictgen::InsensitiveStr::Ascii("wathcer"), &["watcher"]),
            (dictgen::InsensitiveStr::Ascii("jorunal"), &["journal"]),
            (
                dictgen::InsensitiveStr::Ascii("unsatifiable"),
                &["unsatisfiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prcoessing"),
                &["processing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consdiering"),
                &["considering"],
            ),
            (dictgen::InsensitiveStr::Ascii("trnsfers"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("cavren"), &["cavern"]),
            (dictgen::InsensitiveStr::Ascii("alteast"), &["atleast"]),
            (dictgen::InsensitiveStr::Ascii("randomns"), &["randoms"]),
            (
                dictgen::InsensitiveStr::Ascii("typoes"),
                &["typos", "types"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genitaliban"),
                &["genitalia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initmidate"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilialization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphatamines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("recoginze"), &["recognize"]),
            (dictgen::InsensitiveStr::Ascii("boostraps"), &["bootstraps"]),
            (
                dictgen::InsensitiveStr::Ascii("indepentend"),
                &["independents", "independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funciotning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distributs"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reasercher"),
                &["researcher"],
            ),
            (dictgen::InsensitiveStr::Ascii("becasue"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("dtoring"), &["storing"]),
            (dictgen::InsensitiveStr::Ascii("syntaxe"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("observerats"),
                &["observers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporaraly"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("loation"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("chek"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("skippeable"), &["skippable"]),
            (dictgen::InsensitiveStr::Ascii("arrgued"), &["argued"]),
            (dictgen::InsensitiveStr::Ascii("embeddding"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("unecssary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("nastly"), &["nasty"]),
            (
                dictgen::InsensitiveStr::Ascii("interneting"),
                &["interesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("represes"), &["represents"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunatetly"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("leniant"), &["lenient"]),
            (
                dictgen::InsensitiveStr::Ascii("underwieght"),
                &["underweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("expiers"), &["expires"]),
            (dictgen::InsensitiveStr::Ascii("infered"), &["inferred"]),
            (dictgen::InsensitiveStr::Ascii("electorial"), &["electoral"]),
            (dictgen::InsensitiveStr::Ascii("releasead"), &["released"]),
            (dictgen::InsensitiveStr::Ascii("expcted"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("exressing"), &["expressing"]),
            (
                dictgen::InsensitiveStr::Ascii("congradulated"),
                &["congratulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("canabel"), &["cannibal"]),
            (dictgen::InsensitiveStr::Ascii("combinders"), &["combiners"]),
            (
                dictgen::InsensitiveStr::Ascii("dispensaires"),
                &["dispensaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depandancies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignonre"), &["ignore"]),
            (
                dictgen::InsensitiveStr::Ascii("misundrestanding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("appneder"), &["appender"]),
            (dictgen::InsensitiveStr::Ascii("follewin"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("sanboxing"), &["sandboxing"]),
            (dictgen::InsensitiveStr::Ascii("checkoout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("esle"), &["else"]),
            (
                dictgen::InsensitiveStr::Ascii("unitialising"),
                &["uninitialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("neceassary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitutas"),
                &["prostitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("primairly"), &["primarily"]),
            (dictgen::InsensitiveStr::Ascii("assymmetry"), &["asymmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("influentes"),
                &["influences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfamilliar"),
                &["unfamiliar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimters"),
                &["delimiter", "delimiters"],
            ),
            (dictgen::InsensitiveStr::Ascii("enrtry"), &["entry"]),
            (
                dictgen::InsensitiveStr::Ascii("neigborhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conprehension"),
                &["comprehension"],
            ),
            (dictgen::InsensitiveStr::Ascii("excesed"), &["exceeded"]),
            (
                dictgen::InsensitiveStr::Ascii("idenitfying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("disusin"), &["disusing"]),
            (
                dictgen::InsensitiveStr::Ascii("algorythims"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropolies"),
                &["metropolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmarizing"),
                &["summarizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("celverly"), &["cleverly"]),
            (dictgen::InsensitiveStr::Ascii("passvies"), &["passives"]),
            (
                dictgen::InsensitiveStr::Ascii("fanficitons"),
                &["fanfiction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misdeamenor"),
                &["misdemeanor"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebulid"), &["rebuild"]),
            (dictgen::InsensitiveStr::Ascii("dequeing"), &["dequeuing"]),
            (
                dictgen::InsensitiveStr::Ascii("assymetries"),
                &["asymmetries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reoccuring"),
                &["reoccurring"],
            ),
            (dictgen::InsensitiveStr::Ascii("embedds"), &["embeds"]),
            (dictgen::InsensitiveStr::Ascii("worstened"), &["worsened"]),
            (
                dictgen::InsensitiveStr::Ascii("rotaton"),
                &["rotation", "rotator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destraction"),
                &["destruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("displyed"), &["displayed"]),
            (dictgen::InsensitiveStr::Ascii("despoited"), &["deposited"]),
            (dictgen::InsensitiveStr::Ascii("trotskist"), &["trotskyist"]),
            (dictgen::InsensitiveStr::Ascii("terminted"), &["terminated"]),
            (
                dictgen::InsensitiveStr::Ascii("higlighters"),
                &["highlighters"],
            ),
            (dictgen::InsensitiveStr::Ascii("thowards"), &["towards"]),
            (
                dictgen::InsensitiveStr::Ascii("parametreless"),
                &["parameterless"],
            ),
            (dictgen::InsensitiveStr::Ascii("butons"), &["buttons"]),
            (dictgen::InsensitiveStr::Ascii("excpects"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("settiings"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("utillise"), &["utilise"]),
            (dictgen::InsensitiveStr::Ascii("pronouced"), &["pronounced"]),
            (dictgen::InsensitiveStr::Ascii("canacels"), &["cancels"]),
            (
                dictgen::InsensitiveStr::Ascii("undertable"),
                &["understandable", "understand"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoutns"), &["accounts"]),
            (
                dictgen::InsensitiveStr::Ascii("additiionally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensogram"), &["sensorgram"]),
            (
                dictgen::InsensitiveStr::Ascii("stampeed"),
                &["stamped", "stampede"],
            ),
            (dictgen::InsensitiveStr::Ascii("auguest"), &["august"]),
            (dictgen::InsensitiveStr::Ascii("wheh"), &["when"]),
            (
                dictgen::InsensitiveStr::Ascii("elememtary"),
                &["elementary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investingate"),
                &["investigate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeimenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("repostas"), &["reposts"]),
            (
                dictgen::InsensitiveStr::Ascii("misundersranding"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("suffient"), &["sufficient"]),
            (dictgen::InsensitiveStr::Ascii("ckecked"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("adquired"), &["acquired"]),
            (dictgen::InsensitiveStr::Ascii("inteview"), &["interview"]),
            (dictgen::InsensitiveStr::Ascii("rewirte"), &["rewrite"]),
            (dictgen::InsensitiveStr::Ascii("safequard"), &["safeguard"]),
            (dictgen::InsensitiveStr::Ascii("usag"), &["usage"]),
            (dictgen::InsensitiveStr::Ascii("covnert"), &["convert"]),
            (
                dictgen::InsensitiveStr::Ascii("hierachies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("tocuhdown"), &["touchdown"]),
            (
                dictgen::InsensitiveStr::Ascii("juggernout"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressin"),
                &["progression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininstrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("asperges"), &["aspergers"]),
            (dictgen::InsensitiveStr::Ascii("arragne"), &["arrange"]),
            (
                dictgen::InsensitiveStr::Ascii("intestents"),
                &["intestines"],
            ),
            (dictgen::InsensitiveStr::Ascii("racthet"), &["ratchet"]),
            (
                dictgen::InsensitiveStr::Ascii("perosnally"),
                &["personally"],
            ),
            (dictgen::InsensitiveStr::Ascii("potemtial"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("contiunally"),
                &["continually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decommpress"),
                &["decompress"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intregration"),
                &["integration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenetation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("farward"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("regulary"),
                &["regularly", "regular"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfecly"), &["perfectly"]),
            (
                dictgen::InsensitiveStr::Ascii("separacion"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("follfows"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("cocumentation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplenting"),
                &["reimplementing", "reimplanting"],
            ),
            (dictgen::InsensitiveStr::Ascii("proerties"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("normaized"), &["normalized"]),
            (
                dictgen::InsensitiveStr::Ascii("hyptothetically"),
                &["hypothetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitialzied"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncondisionnal"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("cretical"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("producitons"),
                &["productions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initliazing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identiers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("validing"), &["validating"]),
            (dictgen::InsensitiveStr::Ascii("tacticas"), &["tactics"]),
            (
                dictgen::InsensitiveStr::Ascii("amealiarator"),
                &["ameliorator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subequence"),
                &["subsequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("wlacott"), &["walcott"]),
            (
                dictgen::InsensitiveStr::Ascii("translatin"),
                &["translations", "translating", "translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("significane"),
                &["significance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncrhonizes"),
                &["synchronizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("feminisim"), &["feminism"]),
            (
                dictgen::InsensitiveStr::Ascii("striaghtforward"),
                &["straightforward"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("develpoers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standrdized"),
                &["standardized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnmitters"),
                &["transmitters"],
            ),
            (dictgen::InsensitiveStr::Ascii("intercecpt"), &["intercept"]),
            (
                dictgen::InsensitiveStr::Ascii("silowetist"),
                &["silhouettist"],
            ),
            (dictgen::InsensitiveStr::Ascii("unispired"), &["uninspired"]),
            (dictgen::InsensitiveStr::Ascii("variabble"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("delcared"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("umbiguity"), &["ambiguity"]),
            (
                dictgen::InsensitiveStr::Ascii("checkkerboards"),
                &["checkerboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("proepr"), &["proper"]),
            (dictgen::InsensitiveStr::Ascii("exepecting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("bigest"),
                &["biggest", "digest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resvervations"),
                &["reservations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorgable"), &["ignorable"]),
            (
                dictgen::InsensitiveStr::Ascii("verificiations"),
                &["verifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("navagator"), &["navigator"]),
            (
                dictgen::InsensitiveStr::Ascii("regconizing"),
                &["recognizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("unqouting"), &["unquoting"]),
            (
                dictgen::InsensitiveStr::Ascii("settubg"),
                &["setting", "setup"],
            ),
            (dictgen::InsensitiveStr::Ascii("garfied"), &["garfield"]),
            (dictgen::InsensitiveStr::Ascii("ifself"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("coordiates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("gimmickers"), &["gimmicks"]),
            (
                dictgen::InsensitiveStr::Ascii("corispondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hallucenations"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("shifer"), &["shifter"]),
            (dictgen::InsensitiveStr::Ascii("templateas"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("marshmallowiest"),
                &["marshmallows"],
            ),
            (dictgen::InsensitiveStr::Ascii("rmeoved"), &["removed"]),
            (dictgen::InsensitiveStr::Ascii("studnets"), &["students"]),
            (dictgen::InsensitiveStr::Ascii("paralles"), &["parallels"]),
            (
                dictgen::InsensitiveStr::Ascii("capabilites"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formattting"),
                &["formatting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recognices"),
                &["recognizes", "recognises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percpetion"),
                &["perceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("unios"), &["unions"]),
            (dictgen::InsensitiveStr::Ascii("potocol"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("reseource"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("versiobed"), &["versioned"]),
            (
                dictgen::InsensitiveStr::Ascii("appraently"),
                &["apparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("buuilds"), &["builds"]),
            (
                dictgen::InsensitiveStr::Ascii("authethenticated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertiary"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("georiga"), &["georgia"]),
            (
                dictgen::InsensitiveStr::Ascii("distunguish"),
                &["distinguish"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppplying"), &["supplying"]),
            (
                dictgen::InsensitiveStr::Ascii("intergating"),
                &["integrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponetially"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inintializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyperblic"), &["hyperbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("indispensibly"),
                &["indispensably"],
            ),
            (dictgen::InsensitiveStr::Ascii("depdents"), &["dependents"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamnetal"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coreesponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("steppign"), &["stepping"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialzies"),
                &["deserializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aquiantance"),
                &["acquaintance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centisenconds"),
                &["centiseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conttributing"),
                &["contributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wheight"),
                &["weight", "height"],
            ),
            (dictgen::InsensitiveStr::Ascii("assciates"), &["associates"]),
            (dictgen::InsensitiveStr::Ascii("datetiems"), &["datetimes"]),
            (dictgen::InsensitiveStr::Ascii("expercts"), &["expects"]),
            (
                dictgen::InsensitiveStr::Ascii("dericteries"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("birthdaymas"),
                &["birthdays"],
            ),
            (dictgen::InsensitiveStr::Ascii("exurpt"), &["excerpt"]),
            (
                dictgen::InsensitiveStr::Ascii("pasre"),
                &["parse", "paste", "pare"],
            ),
            (dictgen::InsensitiveStr::Ascii("propetys"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("imprort"), &["import"]),
            (
                dictgen::InsensitiveStr::Ascii("prefereble"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoitable"),
                &["negotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("remves"), &["removes"]),
            (dictgen::InsensitiveStr::Ascii("nedium"), &["medium"]),
            (
                dictgen::InsensitiveStr::Ascii("pshycopathic"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("referal"), &["referral"]),
            (dictgen::InsensitiveStr::Ascii("titanim"), &["titanium"]),
            (dictgen::InsensitiveStr::Ascii("argeument"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("scandonavian"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("threesholds"),
                &["thresholds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unattanded"),
                &["unattended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeatbale"),
                &["unbeatable"],
            ),
            (dictgen::InsensitiveStr::Ascii("productos"), &["products"]),
            (dictgen::InsensitiveStr::Ascii("simliiar"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("consideraste"),
                &["considerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positoined"),
                &["positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conlusions"),
                &["conclusions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribitors"),
                &["distributors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("topicaizer"),
                &["topicalizer"],
            ),
            (dictgen::InsensitiveStr::Ascii("spsaces"), &["spaces"]),
            (dictgen::InsensitiveStr::Ascii("sintacks"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("revelaitons"),
                &["revelations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicialize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("physially"), &["physically"]),
            (
                dictgen::InsensitiveStr::Ascii("translatoins"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seqeuncing"),
                &["sequencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dtermine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitialise"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("accouting"), &["accounting"]),
            (dictgen::InsensitiveStr::Ascii("pulgin"), &["plugin"]),
            (dictgen::InsensitiveStr::Ascii("graphhics"), &["graphics"]),
            (dictgen::InsensitiveStr::Ascii("interfcae"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("velociry"), &["velocity"]),
            (dictgen::InsensitiveStr::Ascii("indevered"), &["endeavored"]),
            (
                dictgen::InsensitiveStr::Ascii("recconected"),
                &["reconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsollicited"),
                &["unsolicited"],
            ),
            (dictgen::InsensitiveStr::Ascii("inhumaan"), &["inhuman"]),
            (dictgen::InsensitiveStr::Ascii("becaue"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("imnage"), &["image"]),
            (dictgen::InsensitiveStr::Ascii("dickisch"), &["dickish"]),
            (dictgen::InsensitiveStr::Ascii("sceanrios"), &["scenarios"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubite"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucrasy"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("hospitales"), &["hospitals"]),
            (
                dictgen::InsensitiveStr::Ascii("peodphiles"),
                &["pedophiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("poingant"), &["poignant"]),
            (
                dictgen::InsensitiveStr::Ascii("diffrerences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interacteve"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("deriviated"), &["derived"]),
            (
                dictgen::InsensitiveStr::Ascii("legitimatelly"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consonents"),
                &["consonants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concaneting"),
                &["concatenating"],
            ),
            (dictgen::InsensitiveStr::Ascii("invertedd"), &["inverted"]),
            (
                dictgen::InsensitiveStr::Ascii("akumulated"),
                &["accumulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefentions"),
                &["preventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destuctive"),
                &["destructive"],
            ),
            (dictgen::InsensitiveStr::Ascii("schme"), &["scheme"]),
            (dictgen::InsensitiveStr::Ascii("collaped"), &["collapsed"]),
            (
                dictgen::InsensitiveStr::Ascii("liveatream"),
                &["livestream"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("messae"),
                &["message", "messy"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploding"), &["uploading"]),
            (
                dictgen::InsensitiveStr::Ascii("definifiton"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("custumize"), &["customize"]),
            (
                dictgen::InsensitiveStr::Ascii("aproximate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalcence"),
                &["coalescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleasner"), &["cleanser"]),
            (dictgen::InsensitiveStr::Ascii("airrplanes"), &["airplanes"]),
            (dictgen::InsensitiveStr::Ascii("patirot"), &["patriot"]),
            (
                dictgen::InsensitiveStr::Ascii("compolation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("visualyse"), &["visualise"]),
            (dictgen::InsensitiveStr::Ascii("multipier"), &["multiplier"]),
            (
                dictgen::InsensitiveStr::Ascii("enumeratiors"),
                &["enumerators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intolorence"),
                &["intolerance"],
            ),
            (dictgen::InsensitiveStr::Ascii("astarisks"), &["asterisks"]),
            (
                dictgen::InsensitiveStr::Ascii("meatballers"),
                &["meatballs"],
            ),
            (dictgen::InsensitiveStr::Ascii("monumentos"), &["monuments"]),
            (
                dictgen::InsensitiveStr::Ascii("incorportation"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("objump"), &["objdump"]),
            (dictgen::InsensitiveStr::Ascii("encouraes"), &["encourages"]),
            (dictgen::InsensitiveStr::Ascii("unknwown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("harrassments"),
                &["harassments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionnalism"),
                &["professionalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("baraches"), &["branches"]),
            (
                dictgen::InsensitiveStr::Ascii("contongency"),
                &["contingency"],
            ),
            (dictgen::InsensitiveStr::Ascii("operatie"), &["operative"]),
            (dictgen::InsensitiveStr::Ascii("bitfiled"), &["bitfield"]),
            (dictgen::InsensitiveStr::Ascii("tehsion"), &["tension"]),
            (
                dictgen::InsensitiveStr::Ascii("justificacion"),
                &["justification"],
            ),
            (dictgen::InsensitiveStr::Ascii("shortenend"), &["shortened"]),
            (dictgen::InsensitiveStr::Ascii("secerts"), &["secrets"]),
            (dictgen::InsensitiveStr::Ascii("suficates"), &["suffocates"]),
            (dictgen::InsensitiveStr::Ascii("editedt"), &["edited"]),
            (dictgen::InsensitiveStr::Ascii("mofule"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("backgroungs"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correpsond"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("zimbawbe"), &["zimbabwe"]),
            (
                dictgen::InsensitiveStr::Ascii("represantations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("chcklists"), &["checklists"]),
            (
                dictgen::InsensitiveStr::Ascii("inproovments"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expalnations"),
                &["explanations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relevations"),
                &["revelations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugggests"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("unhandeled"), &["unhandled"]),
            (
                dictgen::InsensitiveStr::Ascii("civilizatin"),
                &["civilizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("metagaem"), &["metagame"]),
            (dictgen::InsensitiveStr::Ascii("continuty"), &["continuity"]),
            (
                dictgen::InsensitiveStr::Ascii("bounndaries"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proccessing"),
                &["processing"],
            ),
            (dictgen::InsensitiveStr::Ascii("symbo"), &["symbol"]),
            (
                dictgen::InsensitiveStr::Ascii("simulataneously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedects"),
                &["deducts", "detects"],
            ),
            (dictgen::InsensitiveStr::Ascii("methds"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("acadimics"), &["academics"]),
            (
                dictgen::InsensitiveStr::Ascii("crationalists"),
                &["nationalists", "rationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("strarts"), &["starts"]),
            (
                dictgen::InsensitiveStr::Ascii("ceasar"),
                &["cesar", "caesar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respresentations"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("altenated"), &["alternated"]),
            (
                dictgen::InsensitiveStr::Ascii("undertanded"),
                &["understood"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiomeouts"), &["timeouts"]),
            (
                dictgen::InsensitiveStr::Ascii("craches"),
                &["crashes", "caches", "crutches", "coaches", "crèches"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathed"), &["matched"]),
            (dictgen::InsensitiveStr::Ascii("searhable"), &["searchable"]),
            (
                dictgen::InsensitiveStr::Ascii("availlability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurres"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collelective"),
                &["collective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imaginatiei"),
                &["imaginative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("historicians"),
                &["historians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desensatized"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supposebly"),
                &["supposedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seina"),
                &["seine", "sienna"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disucssions"),
                &["discussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("threatining"),
                &["threatening"],
            ),
            (dictgen::InsensitiveStr::Ascii("exmample"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("withdrawel"),
                &["withdrawal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moel"),
                &["model", "noel", "mole"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticisim"), &["criticism"]),
            (
                dictgen::InsensitiveStr::Ascii("coexsitance"),
                &["coexistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aditionnaly"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undelrying"),
                &["underlying"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgiuse"), &["disguise"]),
            (dictgen::InsensitiveStr::Ascii("hyerlink"), &["hyperlink"]),
            (dictgen::InsensitiveStr::Ascii("convertet"), &["converted"]),
            (dictgen::InsensitiveStr::Ascii("windwows"), &["windows"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhytms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("utiltity"), &["utility"]),
            (
                dictgen::InsensitiveStr::Ascii("intialzes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatgrams"), &["datagrams"]),
            (dictgen::InsensitiveStr::Ascii("immediante"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("secrest"), &["secrets"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstantal"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("pysically"), &["physically"]),
            (dictgen::InsensitiveStr::Ascii("unerstood"), &["understood"]),
            (dictgen::InsensitiveStr::Ascii("hradwares"), &["hardwares"]),
            (
                dictgen::InsensitiveStr::Ascii("cylidrically"),
                &["cylindrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poductions"),
                &["productions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fallthrought"),
                &["fallthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheigh"), &["weigh"]),
            (
                dictgen::InsensitiveStr::Ascii("umcomfortable"),
                &["uncomfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estabilishes"),
                &["establishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("erros"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("consturcts"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("metafata"), &["metadata"]),
            (dictgen::InsensitiveStr::Ascii("chaeck"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("projecor"), &["projector"]),
            (
                dictgen::InsensitiveStr::Ascii("conrrespondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presssuring"),
                &["pressuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrovating"),
                &["aggravating"],
            ),
            (dictgen::InsensitiveStr::Ascii("resemblace"), &["resemble"]),
            (dictgen::InsensitiveStr::Ascii("eliminas"), &["eliminates"]),
            (
                dictgen::InsensitiveStr::Ascii("implemetations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("bowkays"), &["bouquets"]),
            (dictgen::InsensitiveStr::Ascii("devleopper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("stockplie"), &["stockpile"]),
            (dictgen::InsensitiveStr::Ascii("functioal"), &["functional"]),
            (dictgen::InsensitiveStr::Ascii("gost"), &["ghost"]),
            (dictgen::InsensitiveStr::Ascii("cooridate"), &["coordinate"]),
            (
                dictgen::InsensitiveStr::Ascii("infiltrerat"),
                &["infiltrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("measuremenets"),
                &["measurements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("argumentating"),
                &["arguing"],
            ),
            (dictgen::InsensitiveStr::Ascii("entirly"), &["entirely"]),
            (dictgen::InsensitiveStr::Ascii("widht"), &["width"]),
            (
                dictgen::InsensitiveStr::Ascii("opproximations"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("otpoins"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("involtued"), &["involuted"]),
            (dictgen::InsensitiveStr::Ascii("enrties"), &["entries"]),
            (dictgen::InsensitiveStr::Ascii("ignroes"), &["ignores"]),
            (dictgen::InsensitiveStr::Ascii("emipre"), &["empire"]),
            (
                dictgen::InsensitiveStr::Ascii("fanficiton"),
                &["fanfiction"],
            ),
            (dictgen::InsensitiveStr::Ascii("partcular"), &["particular"]),
            (dictgen::InsensitiveStr::Ascii("zeebras"), &["zebras"]),
            (dictgen::InsensitiveStr::Ascii("tiemzones"), &["timezones"]),
            (dictgen::InsensitiveStr::Ascii("comiting"), &["committing"]),
            (
                dictgen::InsensitiveStr::Ascii("neeedles"),
                &["needles", "needless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appretiative"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimptumaticaly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emergencias"),
                &["emergencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("onwership"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("clumsly"), &["clumsily"]),
            (
                dictgen::InsensitiveStr::Ascii("exercizable"),
                &["exercisable"],
            ),
            (dictgen::InsensitiveStr::Ascii("torerable"), &["tolerable"]),
            (dictgen::InsensitiveStr::Ascii("ppublisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("recongizer"),
                &["recognizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrievates"),
                &["abbreviates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaions"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipalyer"),
                &["multiplayer"],
            ),
            (dictgen::InsensitiveStr::Ascii("trainning"), &["training"]),
            (
                dictgen::InsensitiveStr::Ascii("interaktion"),
                &["interaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("maletonin"), &["melatonin"]),
            (dictgen::InsensitiveStr::Ascii("exlcludes"), &["excludes"]),
            (dictgen::InsensitiveStr::Ascii("staright"), &["straight"]),
            (
                dictgen::InsensitiveStr::Ascii("explicitily"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empirial"),
                &["empirical", "imperial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perviews"),
                &["previews", "purviews"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iteresting"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ligthweights"),
                &["lightweights"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acheevements"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("gyph"), &["glyph", "gyp"]),
            (
                dictgen::InsensitiveStr::Ascii("distintive"),
                &["distinctive"],
            ),
            (dictgen::InsensitiveStr::Ascii("functoin"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("penality"), &["penalty"]),
            (
                dictgen::InsensitiveStr::Ascii("cerificate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dummmy"), &["dummy"]),
            (
                dictgen::InsensitiveStr::Ascii("radiactive"),
                &["radioactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("livetimes"), &["lifetimes"]),
            (dictgen::InsensitiveStr::Ascii("inidans"), &["indians"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisisstic"),
                &["narcissistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("scholarley"), &["scholarly"]),
            (dictgen::InsensitiveStr::Ascii("dimention"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("sympathsizer"),
                &["sympathizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("firmwart"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("gettting"), &["getting"]),
            (dictgen::InsensitiveStr::Ascii("widgit"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("operatios"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("convertion"),
                &["conversion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pobularity"),
                &["popularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demosntration"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cilent"),
                &["client", "silent"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderring"), &["rendering"]),
            (dictgen::InsensitiveStr::Ascii("objecs"), &["objects"]),
            (dictgen::InsensitiveStr::Ascii("warniings"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("graphcial"), &["graphical"]),
            (dictgen::InsensitiveStr::Ascii("sopported"), &["supported"]),
            (
                dictgen::InsensitiveStr::Ascii("permanentely"),
                &["permanently"],
            ),
            (dictgen::InsensitiveStr::Ascii("btiched"), &["bitched"]),
            (
                dictgen::InsensitiveStr::Ascii("initailziation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cherch"), &["church"]),
            (
                dictgen::InsensitiveStr::Ascii("expetectly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimitely"),
                &["legitimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffrents"),
                &["different", "difference"],
            ),
            (dictgen::InsensitiveStr::Ascii("paddin"), &["padding"]),
            (dictgen::InsensitiveStr::Ascii("delets"), &["deletes"]),
            (dictgen::InsensitiveStr::Ascii("spefiifier"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("equirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorwwithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("emptyed"), &["emptied"]),
            (
                dictgen::InsensitiveStr::Ascii("navagators"),
                &["navigators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prooblematic"),
                &["problematic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constucter"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("possbly"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("execuions"), &["executions"]),
            (
                dictgen::InsensitiveStr::Ascii("pairocheelity"),
                &["parochiality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diposition"),
                &["disposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("benghazzi"), &["benghazi"]),
            (
                dictgen::InsensitiveStr::Ascii("tansforming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adknowledged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instersctions"),
                &["intersections"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrappend"), &["wrapped"]),
            (
                dictgen::InsensitiveStr::Ascii("addoptive"),
                &["adoptive", "adaptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clearifying"),
                &["clarifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("repoerter"), &["reporter"]),
            (dictgen::InsensitiveStr::Ascii("mkea"), &["make"]),
            (
                dictgen::InsensitiveStr::Ascii("partititioned"),
                &["partitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typoe"),
                &["typo", "type", "types"],
            ),
            (dictgen::InsensitiveStr::Ascii("capitalse"), &["capitalise"]),
            (
                dictgen::InsensitiveStr::Ascii("attribiutions"),
                &["attributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("situbbornness"),
                &["stubbornness"],
            ),
            (dictgen::InsensitiveStr::Ascii("sperator"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("contining"),
                &["containing", "continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequences"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("astronaunt"), &["astronaut"]),
            (
                dictgen::InsensitiveStr::Ascii("confessino"),
                &["confessions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absoolutely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("ebedded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("validtes"), &["validates"]),
            (dictgen::InsensitiveStr::Ascii("buildds"), &["builds"]),
            (
                dictgen::InsensitiveStr::Ascii("confugirable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sketcked"),
                &["skated", "sketched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaiter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconnaisance"),
                &["reconnaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpeloyed"), &["deployed"]),
            (dictgen::InsensitiveStr::Ascii("trigers"), &["triggers"]),
            (
                dictgen::InsensitiveStr::Ascii("automatcally"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ofsset"), &["offset"]),
            (
                dictgen::InsensitiveStr::Ascii("communicaiton"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresponse"),
                &["unresponsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxiately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucessfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("leasures"), &["leisures"]),
            (dictgen::InsensitiveStr::Ascii("ninimal"), &["minimal"]),
            (
                dictgen::InsensitiveStr::Ascii("ceritificates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordiniate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expalnation"),
                &["explanation"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilisa"), &["utilise"]),
            (
                dictgen::InsensitiveStr::Ascii("escating"),
                &["escalating", "escaping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correkting"),
                &["correcting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brobabilities"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relizations"),
                &["realizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cedentials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapend"), &["happened"]),
            (dictgen::InsensitiveStr::Ascii("imlicitly"), &["implicitly"]),
            (dictgen::InsensitiveStr::Ascii("tthat"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("secrelty"), &["secretly"]),
            (dictgen::InsensitiveStr::Ascii("poitively"), &["positively"]),
            (
                dictgen::InsensitiveStr::Ascii("satisactorily"),
                &["satisfactorily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparantly"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiblble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("directoris"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffrentiated"),
                &["differentiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoporation"),
                &["incorporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conenctration"),
                &["concentrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reording"),
                &["recording", "reordering", "rewording"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbarhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inernal"),
                &["internal", "infernal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certicates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("unchached"), &["uncached"]),
            (dictgen::InsensitiveStr::Ascii("treament"), &["treatment"]),
            (dictgen::InsensitiveStr::Ascii("forwardig"), &["forwarding"]),
            (dictgen::InsensitiveStr::Ascii("initiaize"), &["initialize"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomanse"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("committes"), &["committees"]),
            (dictgen::InsensitiveStr::Ascii("estiamted"), &["estimated"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallisers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decoraters"),
                &["decorators", "decorates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvinient"),
                &["inconvenient"],
            ),
            (dictgen::InsensitiveStr::Ascii("requsting"), &["requesting"]),
            (
                dictgen::InsensitiveStr::Ascii("inmmediately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("ivalid"), &["invalid"]),
            (dictgen::InsensitiveStr::Ascii("withs"), &["with", "widths"]),
            (
                dictgen::InsensitiveStr::Ascii("alignemnts"),
                &["alignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("poltry"), &["poultry"]),
            (
                dictgen::InsensitiveStr::Ascii("aggaravates"),
                &["aggravates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initizlized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallesses"), &["coalesces"]),
            (
                dictgen::InsensitiveStr::Ascii("subsitututing"),
                &["substituting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("melineumm"),
                &["millennia", "millennium"],
            ),
            (dictgen::InsensitiveStr::Ascii("divideneds"), &["dividend"]),
            (
                dictgen::InsensitiveStr::Ascii("strucutured"),
                &["structured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supllied"),
                &["supplied", "sullied"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frowrading"),
                &["forwarding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outsanding"),
                &["outstanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irrespecitve"),
                &["irrespective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archiever"),
                &["achiever", "archiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retrocatively"),
                &["retroactively"],
            ),
            (dictgen::InsensitiveStr::Ascii("mongolos"), &["mongols"]),
            (dictgen::InsensitiveStr::Ascii("neeed"), &["need", "needed"]),
            (
                dictgen::InsensitiveStr::Ascii("anticapate"),
                &["anticipate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organiserad"),
                &["organised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispencary"),
                &["dispensary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fauturing"), &["featuring"]),
            (dictgen::InsensitiveStr::Ascii("favortie"), &["favorite"]),
            (
                dictgen::InsensitiveStr::Ascii("genearative"),
                &["generative"],
            ),
            (dictgen::InsensitiveStr::Ascii("aneel"), &["anneal"]),
            (dictgen::InsensitiveStr::Ascii("probems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("philedalphia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorrecly"),
                &["incorrectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrasturcture"),
                &["infrastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirh"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("menetioning"),
                &["mentioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("pepperino"), &["pepperoni"]),
            (
                dictgen::InsensitiveStr::Ascii("programtically"),
                &["programmatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkbook"), &["workbook"]),
            (
                dictgen::InsensitiveStr::Ascii("lavels"),
                &["levels", "labels"],
            ),
            (dictgen::InsensitiveStr::Ascii("inflatoin"), &["inflation"]),
            (
                dictgen::InsensitiveStr::Ascii("construin"),
                &["constrain", "construing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administator"),
                &["administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmatter"),
                &["transmitter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproducability"),
                &["reproducibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transgener"),
                &["transgender"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apologes"),
                &["apologise", "apologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressieve"),
                &["progressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterbat"),
                &["counteract"],
            ),
            (dictgen::InsensitiveStr::Ascii("signnals"), &["signals"]),
            (dictgen::InsensitiveStr::Ascii("inteneded"), &["intended"]),
            (
                dictgen::InsensitiveStr::Ascii("quanititative"),
                &["quantitative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fillters"),
                &["filters", "fillers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bristool"), &["bristol"]),
            (dictgen::InsensitiveStr::Ascii("eorror"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("zoinism"), &["zionism"]),
            (
                dictgen::InsensitiveStr::Ascii("progagator"),
                &["propagator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementato"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencity"),
                &["authenticity"],
            ),
            (dictgen::InsensitiveStr::Ascii("estiamte"), &["estimate"]),
            (
                dictgen::InsensitiveStr::Ascii("duplifaces"),
                &["duplicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predecence"),
                &["precedence"],
            ),
            (dictgen::InsensitiveStr::Ascii("honeymon"), &["honeymoon"]),
            (dictgen::InsensitiveStr::Ascii("enterring"), &["entering"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatvie"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("littiefinger"),
                &["littlefinger"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlclude"), &["exclude"]),
            (dictgen::InsensitiveStr::Ascii("vanillla"), &["vanilla"]),
            (dictgen::InsensitiveStr::Ascii("palcing"), &["placing"]),
            (
                dictgen::InsensitiveStr::Ascii("implemntors"),
                &["implementors"],
            ),
            (dictgen::InsensitiveStr::Ascii("declaraste"), &["declares"]),
            (dictgen::InsensitiveStr::Ascii("roundtripp"), &["roundtrip"]),
            (dictgen::InsensitiveStr::Ascii("parition"), &["partition"]),
            (dictgen::InsensitiveStr::Ascii("wrestlin"), &["wrestling"]),
            (dictgen::InsensitiveStr::Ascii("itterators"), &["iterators"]),
            (dictgen::InsensitiveStr::Ascii("pebblers"), &["pebbles"]),
            (
                dictgen::InsensitiveStr::Ascii("privilidges"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatabiity"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asssociating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessaily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chanpionships"),
                &["championships"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaiable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("crusier"), &["cruiser"]),
            (
                dictgen::InsensitiveStr::Ascii("expereicne"),
                &["experience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaigous"),
                &["contagious"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciries"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("alliased"), &["aliased"]),
            (dictgen::InsensitiveStr::Ascii("lonley"), &["lonely"]),
            (
                dictgen::InsensitiveStr::Ascii("condtiions"),
                &["conditions"],
            ),
            (dictgen::InsensitiveStr::Ascii("hightest"), &["highest"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriatally"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("plarises"), &["polarises"]),
            (dictgen::InsensitiveStr::Ascii("unexpecte"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("conpletely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contendors"),
                &["contenders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camparable"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("enocded"), &["encoded"]),
            (dictgen::InsensitiveStr::Ascii("lightnin"), &["lightning"]),
            (
                dictgen::InsensitiveStr::Ascii("missmanaged"),
                &["mismanaged"],
            ),
            (dictgen::InsensitiveStr::Ascii("finised"), &["finished"]),
            (
                dictgen::InsensitiveStr::Ascii("summorized"),
                &["summarized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitutues"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apparentely"),
                &["apparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("flethcer"), &["fletcher"]),
            (
                dictgen::InsensitiveStr::Ascii("connectiong"),
                &["connecting", "connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilegies"),
                &["privilege"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compraison"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cymptumaticaly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("televsion"), &["television"]),
            (dictgen::InsensitiveStr::Ascii("sepraate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("thise"), &["these", "this"]),
            (dictgen::InsensitiveStr::Ascii("ciritcals"), &["criticals"]),
            (dictgen::InsensitiveStr::Ascii("arbitratry"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("accrdions"), &["accordions"]),
            (
                dictgen::InsensitiveStr::Ascii("nightmarket"),
                &["nightmare"],
            ),
            (dictgen::InsensitiveStr::Ascii("withuout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("undestood"), &["understood"]),
            (dictgen::InsensitiveStr::Ascii("acccess"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("compase"), &["compares"]),
            (dictgen::InsensitiveStr::Ascii("chinees"), &["chinese"]),
            (dictgen::InsensitiveStr::Ascii("soscket"), &["socket"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiant"),
                &["enthusiast"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshks"), &["handshakes"]),
            (dictgen::InsensitiveStr::Ascii("ammounts"), &["amounts"]),
            (
                dictgen::InsensitiveStr::Ascii("celebirties"),
                &["celebrities"],
            ),
            (dictgen::InsensitiveStr::Ascii("reosurces"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("forcasting"),
                &["forecasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colloborate"),
                &["collaborate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challegers"),
                &["challengers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpertators"),
                &["perpetrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("noramalises"),
                &["normalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infaillible"),
                &["infallible"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomorrw"), &["tomorrow"]),
            (
                dictgen::InsensitiveStr::Ascii("containors"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seranomies"),
                &["ceremonies"],
            ),
            (dictgen::InsensitiveStr::Ascii("synbolic"), &["symbolic"]),
            (
                dictgen::InsensitiveStr::Ascii("intermideate"),
                &["intermediate"],
            ),
            (dictgen::InsensitiveStr::Ascii("inercia"), &["inertia"]),
            (dictgen::InsensitiveStr::Ascii("grafical"), &["graphical"]),
            (dictgen::InsensitiveStr::Ascii("dispencers"), &["dispenser"]),
            (
                dictgen::InsensitiveStr::Ascii("realitvely"),
                &["relatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compataible"),
                &["compatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bullitproof"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electorite"),
                &["electorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("handeld"),
                &["handled", "handheld"],
            ),
            (dictgen::InsensitiveStr::Ascii("creaetors"), &["creators"]),
            (dictgen::InsensitiveStr::Ascii("huristic"), &["heuristic"]),
            (dictgen::InsensitiveStr::Ascii("viewpoert"), &["viewport"]),
            (
                dictgen::InsensitiveStr::Ascii("survivabily"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficififer"),
                &["specifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("newsans"), &["nuisance"]),
            (dictgen::InsensitiveStr::Ascii("insertos"), &["inserts"]),
            (
                dictgen::InsensitiveStr::Ascii("degradato"),
                &["degradation"],
            ),
            (dictgen::InsensitiveStr::Ascii("privious"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("champain"), &["champagne"]),
            (dictgen::InsensitiveStr::Ascii("asembled"), &["assembled"]),
            (
                dictgen::InsensitiveStr::Ascii("inidicators"),
                &["indicators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estabishd"),
                &["established"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flexibily"),
                &["flexibility", "flexibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkuot"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("compoent"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("magneisum"), &["magnesium"]),
            (dictgen::InsensitiveStr::Ascii("quessant"), &["croissant"]),
            (
                dictgen::InsensitiveStr::Ascii("coomittees"),
                &["committees"],
            ),
            (dictgen::InsensitiveStr::Ascii("wtinesses"), &["witnesses"]),
            (
                dictgen::InsensitiveStr::Ascii("associetive"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nastershems"),
                &["nasturtiums"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesytsems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("messeging"), &["messaging"]),
            (dictgen::InsensitiveStr::Ascii("olreans"), &["orleans"]),
            (dictgen::InsensitiveStr::Ascii("endpionts"), &["endpoints"]),
            (dictgen::InsensitiveStr::Ascii("strippen"), &["stripped"]),
            (dictgen::InsensitiveStr::Ascii("mineras"), &["minerals"]),
            (
                dictgen::InsensitiveStr::Ascii("aggresions"),
                &["aggression"],
            ),
            (dictgen::InsensitiveStr::Ascii("endoding"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("multple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("neigbhors"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("grammitical"),
                &["grammatical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuratrion"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutiindex"), &["multiindex"]),
            (dictgen::InsensitiveStr::Ascii("ajative"), &["adjective"]),
            (
                dictgen::InsensitiveStr::Ascii("ptichforks"),
                &["pitchforks"],
            ),
            (dictgen::InsensitiveStr::Ascii("clik"), &["click"]),
            (dictgen::InsensitiveStr::Ascii("replayd"), &["replayed"]),
            (
                dictgen::InsensitiveStr::Ascii("distributvie"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("pracitse"), &["practise"]),
            (dictgen::InsensitiveStr::Ascii("houres"), &["hours"]),
            (
                dictgen::InsensitiveStr::Ascii("combindatorics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissppointed"),
                &["disappointed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspondnce"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oberflowed"),
                &["overflowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strategiske"),
                &["strategies"],
            ),
            (dictgen::InsensitiveStr::Ascii("traing"), &["training"]),
            (
                dictgen::InsensitiveStr::Ascii("counsilors"),
                &["counselors"],
            ),
            (dictgen::InsensitiveStr::Ascii("verboce"), &["verbose"]),
            (
                dictgen::InsensitiveStr::Ascii("optimizible"),
                &["optimizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorgithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("deliberite"),
                &["deliberate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaion"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("summmarize"), &["summarize"]),
            (dictgen::InsensitiveStr::Ascii("causious"), &["cautious"]),
            (dictgen::InsensitiveStr::Ascii("nomalizes"), &["normalizes"]),
            (dictgen::InsensitiveStr::Ascii("bithces"), &["bitches"]),
            (dictgen::InsensitiveStr::Ascii("chanels"), &["channels"]),
            (
                dictgen::InsensitiveStr::Ascii("subsciptions"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("wotking"), &["working"]),
            (
                dictgen::InsensitiveStr::Ascii("natinal"),
                &["matinal", "national"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disposeable"),
                &["disposable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fatalitites"),
                &["fatalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("agurement"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("overloaed"), &["overload"]),
            (
                dictgen::InsensitiveStr::Ascii("annhiliate"),
                &["annihilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("stangant"), &["stagnant"]),
            (dictgen::InsensitiveStr::Ascii("laguague"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("unprotexted"),
                &["unprotected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicater"),
                &["indicator", "indicated", "indicates", "indicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("theroies"), &["theories"]),
            (dictgen::InsensitiveStr::Ascii("cranberrry"), &["cranberry"]),
            (
                dictgen::InsensitiveStr::Ascii("perparation"),
                &["preparation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeimenting"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("spawnig"), &["spawning"]),
            (dictgen::InsensitiveStr::Ascii("esprots"), &["esports"]),
            (dictgen::InsensitiveStr::Ascii("temproray"), &["temporary"]),
            (
                dictgen::InsensitiveStr::Ascii("envolve"),
                &["involve", "evolve"],
            ),
            (dictgen::InsensitiveStr::Ascii("deeemed"), &["deemed"]),
            (
                dictgen::InsensitiveStr::Ascii("competiors"),
                &["competitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apocalpyse"),
                &["apocalypse"],
            ),
            (dictgen::InsensitiveStr::Ascii("defererred"), &["deferred"]),
            (
                dictgen::InsensitiveStr::Ascii("especiially"),
                &["especially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segmendation"),
                &["segmentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("directgory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("implimentor"),
                &["implementor", "implementer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceletion"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("acuire"), &["acquire"]),
            (
                dictgen::InsensitiveStr::Ascii("pervalence"),
                &["prevalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("browswed"), &["browsed"]),
            (dictgen::InsensitiveStr::Ascii("listenners"), &["listeners"]),
            (dictgen::InsensitiveStr::Ascii("absoulely"), &["absolutely"]),
            (dictgen::InsensitiveStr::Ascii("vidoegame"), &["videogame"]),
            (
                dictgen::InsensitiveStr::Ascii("everythings"),
                &["everything"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("layser"),
                &["laser", "layer"],
            ),
            (dictgen::InsensitiveStr::Ascii("functined"), &["functioned"]),
            (dictgen::InsensitiveStr::Ascii("comsumers"), &["consumers"]),
            (
                dictgen::InsensitiveStr::Ascii("microtranasctions"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pitckforks"),
                &["pitchforks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pomegranite"),
                &["pomegranate"],
            ),
            (dictgen::InsensitiveStr::Ascii("assigees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("proporpotional"),
                &["proportional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distriubtion"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mysterioulsy"),
                &["mysteriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleandar"), &["calendar"]),
            (
                dictgen::InsensitiveStr::Ascii("configuerations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pennslyvania"),
                &["pennsylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordding"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("suports"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("adaptibe"), &["adaptive"]),
            (dictgen::InsensitiveStr::Ascii("compaign"), &["campaign"]),
            (dictgen::InsensitiveStr::Ascii("itesm"), &["items"]),
            (
                dictgen::InsensitiveStr::Ascii("unregistrs"),
                &["unregisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializiation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("logisitics"), &["logistics"]),
            (
                dictgen::InsensitiveStr::Ascii("coordingates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("introvers"), &["introverts"]),
            (dictgen::InsensitiveStr::Ascii("ediors"), &["editors"]),
            (
                dictgen::InsensitiveStr::Ascii("insuffiently"),
                &["insufficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ulterations"),
                &["alterations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exctractors"),
                &["extractors"],
            ),
            (dictgen::InsensitiveStr::Ascii("derictery"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("changable"), &["changeable"]),
            (dictgen::InsensitiveStr::Ascii("inbewteen"), &["inbetween"]),
            (
                dictgen::InsensitiveStr::Ascii("transparanies"),
                &["transparencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constraucts"),
                &["constructs"],
            ),
            (dictgen::InsensitiveStr::Ascii("thyat"), &["that"]),
            (
                dictgen::InsensitiveStr::Ascii("descallocated"),
                &["deallocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("unnhandled"), &["unhandled"]),
            (dictgen::InsensitiveStr::Ascii("cathderal"), &["cathedral"]),
            (dictgen::InsensitiveStr::Ascii("tnankful"), &["thankful"]),
            (
                dictgen::InsensitiveStr::Ascii("negotions"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functtionally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ststionery"),
                &["stationery"],
            ),
            (dictgen::InsensitiveStr::Ascii("meaasures"), &["measures"]),
            (
                dictgen::InsensitiveStr::Ascii("absoprtion"),
                &["absorption"],
            ),
            (dictgen::InsensitiveStr::Ascii("activste"), &["activate"]),
            (dictgen::InsensitiveStr::Ascii("eczecuting"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("copenahgen"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("levareges"), &["leverages"]),
            (
                dictgen::InsensitiveStr::Ascii("expentiture"),
                &["expenditure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workaraounds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("santitize"), &["sanitize"]),
            (
                dictgen::InsensitiveStr::Ascii("caracterised"),
                &["characterised"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutiple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("straind"), &["strained"]),
            (dictgen::InsensitiveStr::Ascii("docekr"), &["docker"]),
            (
                dictgen::InsensitiveStr::Ascii("intialliser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("resut"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("prestigiosa"),
                &["prestigious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scnadinavia"),
                &["scandinavia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vresioning"),
                &["versioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sweerheart"),
                &["sweetheart"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superivsor"),
                &["supervisors"],
            ),
            (dictgen::InsensitiveStr::Ascii("tragets"), &["targets"]),
            (dictgen::InsensitiveStr::Ascii("childerns"), &["childrens"]),
            (dictgen::InsensitiveStr::Ascii("andoid"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("obejcted"), &["objected"]),
            (dictgen::InsensitiveStr::Ascii("sceonds"), &["seconds"]),
            (dictgen::InsensitiveStr::Ascii("deciced"), &["decided"]),
            (
                dictgen::InsensitiveStr::Ascii("legalizaing"),
                &["legalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("flud"), &["flood"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdenence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("thigny"), &["thingy"]),
            (
                dictgen::InsensitiveStr::Ascii("quanitifies"),
                &["quantifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definintively"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prinicipally"),
                &["principally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approppriate"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("buriser"), &["bruiser"]),
            (dictgen::InsensitiveStr::Ascii("cigarret"), &["cigaret"]),
            (
                dictgen::InsensitiveStr::Ascii("interferience"),
                &["interference"],
            ),
            (dictgen::InsensitiveStr::Ascii("complelte"), &["complete"]),
            (
                dictgen::InsensitiveStr::Ascii("enoucntering"),
                &["encountering"],
            ),
            (dictgen::InsensitiveStr::Ascii("begiins"), &["begins"]),
            (dictgen::InsensitiveStr::Ascii("ludricous"), &["ludicrous"]),
            (dictgen::InsensitiveStr::Ascii("conpare"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("pyschopathic"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("respnder"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("trivally"), &["trivially"]),
            (
                dictgen::InsensitiveStr::Ascii("perpsectives"),
                &["perspectives"],
            ),
            (dictgen::InsensitiveStr::Ascii("macinthosh"), &["macintosh"]),
            (dictgen::InsensitiveStr::Ascii("percieved"), &["perceived"]),
            (dictgen::InsensitiveStr::Ascii("athlets"), &["athletes"]),
            (dictgen::InsensitiveStr::Ascii("nmemonic"), &["mnemonic"]),
            (dictgen::InsensitiveStr::Ascii("seacrching"), &["searching"]),
            (
                dictgen::InsensitiveStr::Ascii("aprroximated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("statuer"), &["stature"]),
            (dictgen::InsensitiveStr::Ascii("squirlte"), &["squirtle"]),
            (dictgen::InsensitiveStr::Ascii("allocaion"), &["allocation"]),
            (dictgen::InsensitiveStr::Ascii("multine"), &["multiline"]),
            (dictgen::InsensitiveStr::Ascii("exhuasted"), &["exhausted"]),
            (
                dictgen::InsensitiveStr::Ascii("temeprature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("clogure"), &["clojure"]),
            (
                dictgen::InsensitiveStr::Ascii("volatiltiy"),
                &["volatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("desiginers"), &["designers"]),
            (
                dictgen::InsensitiveStr::Ascii("confedarcy"),
                &["confederacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("strangel"), &["strangle"]),
            (dictgen::InsensitiveStr::Ascii("rorates"), &["rotates"]),
            (dictgen::InsensitiveStr::Ascii("controid"), &["centroid"]),
            (dictgen::InsensitiveStr::Ascii("angshiosly"), &["anxiously"]),
            (dictgen::InsensitiveStr::Ascii("amunt"), &["amount"]),
            (
                dictgen::InsensitiveStr::Ascii("annnouncement"),
                &["announcement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("removefromat"),
                &["removeformat"],
            ),
            (dictgen::InsensitiveStr::Ascii("tredning"), &["trending"]),
            (
                dictgen::InsensitiveStr::Ascii("attosencond"),
                &["attosecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professsor"),
                &["professors"],
            ),
            (dictgen::InsensitiveStr::Ascii("litteral"), &["literal"]),
            (dictgen::InsensitiveStr::Ascii("combatting"), &["combating"]),
            (
                dictgen::InsensitiveStr::Ascii("reposonders"),
                &["responders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatiable"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wirded"),
                &["wired", "weird"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeeses"), &["successes"]),
            (
                dictgen::InsensitiveStr::Ascii("automoderater"),
                &["automoderator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncomfrotable"),
                &["uncomfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("elementay"), &["elementary"]),
            (dictgen::InsensitiveStr::Ascii("suystems"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("manifiest"), &["manifest"]),
            (dictgen::InsensitiveStr::Ascii("filenae"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("anonotate"), &["annotate"]),
            (
                dictgen::InsensitiveStr::Ascii("frankensten"),
                &["frankenstein"],
            ),
            (dictgen::InsensitiveStr::Ascii("unpluged"), &["unplugged"]),
            (
                dictgen::InsensitiveStr::Ascii("condigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downloadeble"),
                &["downloadable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("considere"),
                &["consider", "considered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwheling"),
                &["overwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimensons"), &["dimensions"]),
            (dictgen::InsensitiveStr::Ascii("familys"), &["families"]),
            (dictgen::InsensitiveStr::Ascii("emuation"), &["emulation"]),
            (dictgen::InsensitiveStr::Ascii("ciruit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("buitlins"), &["builtins"]),
            (dictgen::InsensitiveStr::Ascii("mecahnics"), &["mechanics"]),
            (dictgen::InsensitiveStr::Ascii("metiphor"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("occuppied"), &["occupied"]),
            (
                dictgen::InsensitiveStr::Ascii("estabilishment"),
                &["establishment"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugggest"), &["suggest"]),
            (dictgen::InsensitiveStr::Ascii("stoppin"), &["stopping"]),
            (dictgen::InsensitiveStr::Ascii("similia"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("armamant"), &["armament"]),
            (dictgen::InsensitiveStr::Ascii("reoporting"), &["reporting"]),
            (dictgen::InsensitiveStr::Ascii("unannomous"), &["unanimous"]),
            (dictgen::InsensitiveStr::Ascii("absoluely"), &["absolutely"]),
            (dictgen::InsensitiveStr::Ascii("proejctor"), &["projector"]),
            (dictgen::InsensitiveStr::Ascii("gernerates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("investiagtion"),
                &["investigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmodfide"), &["unmodified"]),
            (
                dictgen::InsensitiveStr::Ascii("progresson"),
                &["progression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validatation"),
                &["validation"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrighy"), &["alrighty"]),
            (
                dictgen::InsensitiveStr::Ascii("possiblities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("stickes"), &["stickers"]),
            (
                dictgen::InsensitiveStr::Ascii("autonamous"),
                &["autonomous"],
            ),
            (dictgen::InsensitiveStr::Ascii("mentione"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliazation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("xenbolade"), &["xenoblade"]),
            (dictgen::InsensitiveStr::Ascii("winsdor"), &["windsor"]),
            (
                dictgen::InsensitiveStr::Ascii("caricaturise"),
                &["caricature"],
            ),
            (dictgen::InsensitiveStr::Ascii("admittin"), &["admitting"]),
            (dictgen::InsensitiveStr::Ascii("cotacted"), &["contacted"]),
            (dictgen::InsensitiveStr::Ascii("platofmr"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastuctures"),
                &["infrastructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("desitable"), &["desirable"]),
            (dictgen::InsensitiveStr::Ascii("obaying"), &["obeying"]),
            (dictgen::InsensitiveStr::Ascii("yuou"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("yorkshrie"), &["yorkshire"]),
            (dictgen::InsensitiveStr::Ascii("merecat"), &["meerkat"]),
            (dictgen::InsensitiveStr::Ascii("aiport"), &["airport"]),
            (
                dictgen::InsensitiveStr::Ascii("installmant"),
                &["installment"],
            ),
            (dictgen::InsensitiveStr::Ascii("reduncant"), &["redundant"]),
            (dictgen::InsensitiveStr::Ascii("vaiants"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("intrinsincally"),
                &["intrinsically"],
            ),
            (dictgen::InsensitiveStr::Ascii("stoping"), &["stopping"]),
            (dictgen::InsensitiveStr::Ascii("diectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginsing"),
                &["recognising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("christains"),
                &["christians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualiztion"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("schduling"), &["scheduling"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilizier"),
                &["stabilize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coallesing"),
                &["coalescing"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomitter"), &["committer"]),
            (
                dictgen::InsensitiveStr::Ascii("rin"),
                &["ring", "rink", "rind", "rain", "rein", "ruin", "grin"],
            ),
            (dictgen::InsensitiveStr::Ascii("actvities"), &["activities"]),
            (dictgen::InsensitiveStr::Ascii("maanging"), &["managing"]),
            (dictgen::InsensitiveStr::Ascii("oout"), &["out"]),
            (dictgen::InsensitiveStr::Ascii("engranving"), &["engraving"]),
            (dictgen::InsensitiveStr::Ascii("commnad"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("exaggarate"),
                &["exaggerate"],
            ),
            (dictgen::InsensitiveStr::Ascii("preventors"), &["prevents"]),
            (
                dictgen::InsensitiveStr::Ascii("clincially"),
                &["clinically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ermergency"), &["emergency"]),
            (dictgen::InsensitiveStr::Ascii("theards"), &["threads"]),
            (dictgen::InsensitiveStr::Ascii("onother"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("primordal"), &["primordial"]),
            (
                dictgen::InsensitiveStr::Ascii("qiestionnaire"),
                &["questionnaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("occsion"), &["occasion"]),
            (dictgen::InsensitiveStr::Ascii("dicrete"), &["discrete"]),
            (dictgen::InsensitiveStr::Ascii("benchmak"), &["benchmark"]),
            (dictgen::InsensitiveStr::Ascii("nedlessly"), &["endlessly"]),
            (dictgen::InsensitiveStr::Ascii("abundace"), &["abundance"]),
            (
                dictgen::InsensitiveStr::Ascii("starnation"),
                &["starvation"],
            ),
            (dictgen::InsensitiveStr::Ascii("decembeard"), &["december"]),
            (dictgen::InsensitiveStr::Ascii("cahnge"), &["change"]),
            (
                dictgen::InsensitiveStr::Ascii("underminning"),
                &["undermining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisically"),
                &["characteristically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schozophrenic"),
                &["schizophrenic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referandum"),
                &["referendum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organziational"),
                &["organizational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crystalised"),
                &["crystallised"],
            ),
            (dictgen::InsensitiveStr::Ascii("conume"), &["consume"]),
            (dictgen::InsensitiveStr::Ascii("inerval"), &["interval"]),
            (
                dictgen::InsensitiveStr::Ascii("perimetres"),
                &["perimeters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invinvible"),
                &["invincible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinates"),
                &["combines", "combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("falired"), &["flaired"]),
            (dictgen::InsensitiveStr::Ascii("spefified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("ommision"),
                &["omission", "commission"],
            ),
            (dictgen::InsensitiveStr::Ascii("bion"), &["bio"]),
            (
                dictgen::InsensitiveStr::Ascii("precondtion"),
                &["precondition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exprimentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimmenter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqeuencer"), &["sequencer"]),
            (
                dictgen::InsensitiveStr::Ascii("addministrator"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("portrayes"), &["portrays"]),
            (dictgen::InsensitiveStr::Ascii("callcacks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("intillectual"),
                &["intellectual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disproportiantely"),
                &["disproportionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("reigsters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("rhethorical"),
                &["rhetorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unistaller"),
                &["uninstaller"],
            ),
            (dictgen::InsensitiveStr::Ascii("resourcin"), &["resourcing"]),
            (dictgen::InsensitiveStr::Ascii("qiestions"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("conditialy"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("drankenstein"),
                &["frankenstein"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminanti"),
                &["illuminati"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanled"), &["handled"]),
            (
                dictgen::InsensitiveStr::Ascii("pervent"),
                &["prevent", "percent", "pervert", "fervent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effectivly"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangts"),
                &["arrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continious"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intersetion"),
                &["intersection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sophisticaed"),
                &["sophisticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("autometic"), &["automatic"]),
            (dictgen::InsensitiveStr::Ascii("elagant"), &["elegant"]),
            (dictgen::InsensitiveStr::Ascii("kwoledge"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("mutiplications"),
                &["multiplications"],
            ),
            (dictgen::InsensitiveStr::Ascii("tourits"), &["tourist"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangin"),
                &["rearranging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parademics"),
                &["paramedics"],
            ),
            (dictgen::InsensitiveStr::Ascii("entirity"), &["entirety"]),
            (dictgen::InsensitiveStr::Ascii("cafay"), &["cafe"]),
            (
                dictgen::InsensitiveStr::Ascii("antiqvated"),
                &["antiquated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progamability"),
                &["programmability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eligibilty"),
                &["eligibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepraring"), &["preparing"]),
            (dictgen::InsensitiveStr::Ascii("ignorgd"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("opactiy"), &["opacity"]),
            (
                dictgen::InsensitiveStr::Ascii("parapharsed"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outperfomring"),
                &["outperforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decelaration"),
                &["deceleration", "declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infridging"),
                &["infringing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assignmets"),
                &["assignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compromisso"),
                &["compromises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaitng"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("senaphore"), &["semaphore"]),
            (
                dictgen::InsensitiveStr::Ascii("homophonia"),
                &["homophobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decraesing"),
                &["decreasing", "deceasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mouthpeace"),
                &["mouthpiece"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cataline"),
                &["catiline", "catalina"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("euclidan"),
                &["euclidean", "euclidian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articluated"),
                &["articulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grammarically"),
                &["grammatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxamates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("cpacity"), &["capacity"]),
            (
                dictgen::InsensitiveStr::Ascii("contraticted"),
                &["contradicted"],
            ),
            (dictgen::InsensitiveStr::Ascii("samwiches"), &["sandwiches"]),
            (dictgen::InsensitiveStr::Ascii("symbolc"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("mispelled"), &["misspelled"]),
            (dictgen::InsensitiveStr::Ascii("dissaper"), &["disappear"]),
            (dictgen::InsensitiveStr::Ascii("organiste"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("pagent"),
                &["pageant", "plangent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automanifactured"),
                &["automanufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("faceis"), &["faces"]),
            (dictgen::InsensitiveStr::Ascii("impliying"), &["implying"]),
            (
                dictgen::InsensitiveStr::Ascii("coordintes"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconveniences"),
                &["inconveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncalcualted"),
                &["uncalculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exress"),
                &["express", "excess"],
            ),
            (dictgen::InsensitiveStr::Ascii("readning"), &["reading"]),
            (
                dictgen::InsensitiveStr::Ascii("discretited"),
                &["discredited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficiallication"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("autofomat"), &["autoformat"]),
            (
                dictgen::InsensitiveStr::Ascii("direcitves"),
                &["directives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credibilty"),
                &["credibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("voluem"), &["volume"]),
            (dictgen::InsensitiveStr::Ascii("montania"), &["montana"]),
            (
                dictgen::InsensitiveStr::Ascii("datahasheets"),
                &["datasheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescients"),
                &["presidents"],
            ),
            (dictgen::InsensitiveStr::Ascii("tetsted"), &["tested"]),
            (dictgen::InsensitiveStr::Ascii("outpupt"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhology"),
                &["psychology"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preferibly"),
                &["preferably"],
            ),
            (dictgen::InsensitiveStr::Ascii("hemingwavy"), &["hemingway"]),
            (
                dictgen::InsensitiveStr::Ascii("combintaor"),
                &["combinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifciers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cimmetric"), &["symmetric"]),
            (
                dictgen::InsensitiveStr::Ascii("ostensiably"),
                &["ostensibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("idenitfy"), &["identify"]),
            (dictgen::InsensitiveStr::Ascii("desribed"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("ugrader"), &["upgrader"]),
            (dictgen::InsensitiveStr::Ascii("detete"), &["delete"]),
            (dictgen::InsensitiveStr::Ascii("melancoly"), &["melancholy"]),
            (
                dictgen::InsensitiveStr::Ascii("descrptively"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitaing"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusomize"), &["customize"]),
            (dictgen::InsensitiveStr::Ascii("birdge"), &["bridge"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaises"),
                &["generalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deperecating"),
                &["deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("speically"), &["specially"]),
            (dictgen::InsensitiveStr::Ascii("recoreded"), &["recorded"]),
            (dictgen::InsensitiveStr::Ascii("recenlty"), &["recently"]),
            (dictgen::InsensitiveStr::Ascii("pogress"), &["progress"]),
            (dictgen::InsensitiveStr::Ascii("someime"), &["sometime"]),
            (dictgen::InsensitiveStr::Ascii("inverval"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("openeing"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidentually"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("portrail"),
                &["portrayal", "portrait"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationable"),
                &["sensational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitfier"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("onw"), &["own", "now"]),
            (
                dictgen::InsensitiveStr::Ascii("serialzation"),
                &["serialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("halloweeen"), &["halloween"]),
            (
                dictgen::InsensitiveStr::Ascii("archtitecturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewiever"), &["reviewer"]),
            (dictgen::InsensitiveStr::Ascii("viusal"), &["visual"]),
            (
                dictgen::InsensitiveStr::Ascii("enlightend"),
                &["enlightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgininal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("snugglie"), &["snuggle"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellises"),
                &["parallelises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particpations"),
                &["participations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performanc"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("antennaes"), &["antennas"]),
            (dictgen::InsensitiveStr::Ascii("signifiy"), &["signify"]),
            (
                dictgen::InsensitiveStr::Ascii("disrespekted"),
                &["disrespected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confuration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fanserivce"),
                &["fanservice"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifromed"), &["uniformed"]),
            (
                dictgen::InsensitiveStr::Ascii("fontain"),
                &["fountain", "contain"],
            ),
            (dictgen::InsensitiveStr::Ascii("forfiting"), &["forfeiting"]),
            (
                dictgen::InsensitiveStr::Ascii("derivaties"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("moduel"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("mangagement"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("numnber"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("voluntairy"),
                &["voluntarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("nervana"), &["nirvana"]),
            (
                dictgen::InsensitiveStr::Ascii("depretiating"),
                &["depreciating", "deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("boyfreind"), &["boyfriend"]),
            (
                dictgen::InsensitiveStr::Ascii("expeimented"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggurate"),
                &["exaggerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condemantion"),
                &["condemnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synonamous"),
                &["synonymous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemants"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("resullting"), &["resulting"]),
            (dictgen::InsensitiveStr::Ascii("compilin"), &["compiling"]),
            (
                dictgen::InsensitiveStr::Ascii("declearator"),
                &["declarator", "decelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("exahusted"), &["exhausted"]),
            (dictgen::InsensitiveStr::Ascii("homeword"), &["homeworld"]),
            (dictgen::InsensitiveStr::Ascii("targetting"), &["targeting"]),
            (dictgen::InsensitiveStr::Ascii("columm"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("aymptote"), &["asymptote"]),
            (
                dictgen::InsensitiveStr::Ascii("analises"),
                &["analysis", "analyses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antagonisitc"),
                &["antagonistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adapations"),
                &["adaptations", "adoptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("decompse"), &["decompose"]),
            (
                dictgen::InsensitiveStr::Ascii("basektball"),
                &["basketball"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpractical"),
                &["impractical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterbation"),
                &["masturbation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oscilliscope"),
                &["oscilloscope"],
            ),
            (dictgen::InsensitiveStr::Ascii("invitaion"), &["invitation"]),
            (dictgen::InsensitiveStr::Ascii("trucnated"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("zealoths"), &["zealots"]),
            (dictgen::InsensitiveStr::Ascii("sylistic"), &["stylistic"]),
            (
                dictgen::InsensitiveStr::Ascii("disfunctionality"),
                &["dysfunctionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("deverloper"), &["developer"]),
            (
                dictgen::InsensitiveStr::Ascii("idependent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("konservatives"),
                &["conservatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("worklaod"), &["workload"]),
            (
                dictgen::InsensitiveStr::Ascii("invesitgation"),
                &["investigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincedince"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmodfified"),
                &["unmodified"],
            ),
            (dictgen::InsensitiveStr::Ascii("algolithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("confiure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("omittin"), &["omitting"]),
            (dictgen::InsensitiveStr::Ascii("othwerwise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("gere"), &["gear", "here"]),
            (dictgen::InsensitiveStr::Ascii("asigning"), &["assigning"]),
            (dictgen::InsensitiveStr::Ascii("overlodes"), &["overloads"]),
            (
                dictgen::InsensitiveStr::Ascii("attachtment"),
                &["attachment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("firectories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("elepahnts"), &["elephants"]),
            (
                dictgen::InsensitiveStr::Ascii("converagent"),
                &["convergent"],
            ),
            (dictgen::InsensitiveStr::Ascii("concieve"), &["conceive"]),
            (dictgen::InsensitiveStr::Ascii("thatnking"), &["thanking"]),
            (
                dictgen::InsensitiveStr::Ascii("copyirghts"),
                &["copyrights"],
            ),
            (dictgen::InsensitiveStr::Ascii("analsye"), &["analyse"]),
            (dictgen::InsensitiveStr::Ascii("caue"), &["cause"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactuter"),
                &["manufacture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supervisin"),
                &["supervision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chronicales"),
                &["chronicles"],
            ),
            (dictgen::InsensitiveStr::Ascii("medicinens"), &["medicines"]),
            (
                dictgen::InsensitiveStr::Ascii("missspelled"),
                &["misspelled"],
            ),
            (dictgen::InsensitiveStr::Ascii("beeen"), &["been"]),
            (
                dictgen::InsensitiveStr::Ascii("transalations"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privisional"),
                &["provisional"],
            ),
            (dictgen::InsensitiveStr::Ascii("vailable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("specefied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("masterbate"),
                &["masturbate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frusteration"),
                &["frustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("synomyns"), &["synonyms"]),
            (dictgen::InsensitiveStr::Ascii("checlbox"), &["checkbox"]),
            (
                dictgen::InsensitiveStr::Ascii("spesification"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("adn"), &["and"]),
            (
                dictgen::InsensitiveStr::Ascii("speciyfying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("composants"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspciously"),
                &["suspiciously"],
            ),
            (dictgen::InsensitiveStr::Ascii("dualtiy"), &["duality"]),
            (dictgen::InsensitiveStr::Ascii("epxected"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("alernate"), &["alternate"]),
            (dictgen::InsensitiveStr::Ascii("coducting"), &["conducting"]),
            (
                dictgen::InsensitiveStr::Ascii("uncensorred"),
                &["uncensored"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("actiation"),
                &["activation", "actuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("porcessing"),
                &["processing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scienctific"),
                &["scientific"],
            ),
            (dictgen::InsensitiveStr::Ascii("expanasion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("genaral"), &["general"]),
            (
                dictgen::InsensitiveStr::Ascii("cringewhorty"),
                &["cringeworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerbated"),
                &["exacerbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bounraries"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("extendsion"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("capcitance"),
                &["capacitance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millionairres"),
                &["millionaire"],
            ),
            (dictgen::InsensitiveStr::Ascii("noddin"), &["nodding"]),
            (dictgen::InsensitiveStr::Ascii("listenn"), &["listen"]),
            (
                dictgen::InsensitiveStr::Ascii("intertained"),
                &["intertwined"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throughtput"),
                &["throughput"],
            ),
            (dictgen::InsensitiveStr::Ascii("thorw"), &["throw"]),
            (dictgen::InsensitiveStr::Ascii("restarant"), &["restaurant"]),
            (
                dictgen::InsensitiveStr::Ascii("intialised"),
                &["initialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("regars"), &["regards"]),
            (
                dictgen::InsensitiveStr::Ascii("identifeirs"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("writebak"), &["writeback"]),
            (
                dictgen::InsensitiveStr::Ascii("mediterraean"),
                &["mediterranean"],
            ),
            (dictgen::InsensitiveStr::Ascii("soberity"), &["sobriety"]),
            (dictgen::InsensitiveStr::Ascii("ultered"), &["altered"]),
            (dictgen::InsensitiveStr::Ascii("downgade"), &["downgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("conncurrent"),
                &["concurrent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assosciated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("expoerted"), &["exported"]),
            (dictgen::InsensitiveStr::Ascii("obsticals"), &["obstacles"]),
            (
                dictgen::InsensitiveStr::Ascii("pkaythroughs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("kentucy"), &["kentucky"]),
            (
                dictgen::InsensitiveStr::Ascii("reccomended"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("staurday"), &["saturday"]),
            (dictgen::InsensitiveStr::Ascii("ingenuitiy"), &["ingenuity"]),
            (dictgen::InsensitiveStr::Ascii("moounting"), &["mounting"]),
            (
                dictgen::InsensitiveStr::Ascii("sciencists"),
                &["scientists"],
            ),
            (dictgen::InsensitiveStr::Ascii("michelline"), &["michelle"]),
            (dictgen::InsensitiveStr::Ascii("avisories"), &["advisories"]),
            (
                dictgen::InsensitiveStr::Ascii("distory"),
                &["destroy", "distort", "history"],
            ),
            (dictgen::InsensitiveStr::Ascii("epliepsy"), &["epilepsy"]),
            (dictgen::InsensitiveStr::Ascii("bitfied"), &["bitfield"]),
            (
                dictgen::InsensitiveStr::Ascii("equavalently"),
                &["equivalently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demodulater"),
                &["demodulator", "demodulated", "demodulates", "demodulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("extremally"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("docmunetation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("predictin"), &["prediction"]),
            (dictgen::InsensitiveStr::Ascii("sucseeded"), &["succeeded"]),
            (
                dictgen::InsensitiveStr::Ascii("configutrable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aesthestic"),
                &["aesthetics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occassioned"),
                &["occasioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pitchforcks"),
                &["pitchforks"],
            ),
            (dictgen::InsensitiveStr::Ascii("chatows"), &["châteaux"]),
            (
                dictgen::InsensitiveStr::Ascii("encouraing"),
                &["encouraging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteactive"),
                &["interactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("palyboy"), &["playboy"]),
            (
                dictgen::InsensitiveStr::Ascii("cipheruites"),
                &["ciphersuites"],
            ),
            (dictgen::InsensitiveStr::Ascii("plent"), &["plenty"]),
            (dictgen::InsensitiveStr::Ascii("columnss"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("cavrens"), &["caverns"]),
            (dictgen::InsensitiveStr::Ascii("disaply"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("generlaize"),
                &["generalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("properities"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("importatly"),
                &["importantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooridinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomodates"),
                &["accommodates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("establised"),
                &["established"],
            ),
            (dictgen::InsensitiveStr::Ascii("configiure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("specififcation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreggator"),
                &["aggregator"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlcluded"), &["included"]),
            (dictgen::InsensitiveStr::Ascii("authoritay"), &["authority"]),
            (dictgen::InsensitiveStr::Ascii("bettern"), &["better"]),
            (dictgen::InsensitiveStr::Ascii("fileds"), &["fields"]),
            (
                dictgen::InsensitiveStr::Ascii("workaroudns"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("previou"), &["previous"]),
            (dictgen::InsensitiveStr::Ascii("antoher"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("supproter"), &["supporter"]),
            (dictgen::InsensitiveStr::Ascii("decrese"), &["decrease"]),
            (
                dictgen::InsensitiveStr::Ascii("documenattion"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("overdirve"), &["overdrive"]),
            (dictgen::InsensitiveStr::Ascii("perfercted"), &["perfected"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimizations"),
                &["optimizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("applyting"), &["applying"]),
            (
                dictgen::InsensitiveStr::Ascii("fahrenheight"),
                &["fahrenheit"],
            ),
            (dictgen::InsensitiveStr::Ascii("finalisin"), &["finalising"]),
            (
                dictgen::InsensitiveStr::Ascii("resstriction"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("frutcose"), &["fructose"]),
            (dictgen::InsensitiveStr::Ascii("anonoymous"), &["anonymous"]),
            (dictgen::InsensitiveStr::Ascii("inlclusive"), &["inclusive"]),
            (dictgen::InsensitiveStr::Ascii("adpapter"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("processeed"), &["processed"]),
            (
                dictgen::InsensitiveStr::Ascii("shakesphere"),
                &["shakespeare"],
            ),
            (dictgen::InsensitiveStr::Ascii("ckeckers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("portrat"), &["portrait"]),
            (dictgen::InsensitiveStr::Ascii("sutff"), &["stuff"]),
            (
                dictgen::InsensitiveStr::Ascii("callibrating"),
                &["calibrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bicyclin"), &["bicycling"]),
            (dictgen::InsensitiveStr::Ascii("mcroscopy"), &["microscopy"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanement"),
                &["enchantment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inizialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("recenetly"), &["recently"]),
            (dictgen::InsensitiveStr::Ascii("seperatedy"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("alienet"), &["alienate"]),
            (
                dictgen::InsensitiveStr::Ascii("sysmbolically"),
                &["symbolically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touples"),
                &["tuples", "couples", "topples", "toupees"],
            ),
            (dictgen::InsensitiveStr::Ascii("sensores"), &["sensors"]),
            (
                dictgen::InsensitiveStr::Ascii("friendzowned"),
                &["friendzoned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestantisk"),
                &["protestants"],
            ),
            (dictgen::InsensitiveStr::Ascii("intership"), &["internship"]),
            (dictgen::InsensitiveStr::Ascii("thoward"), &["toward"]),
            (dictgen::InsensitiveStr::Ascii("ropeat"), &["repeat"]),
            (
                dictgen::InsensitiveStr::Ascii("opreations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("includdes"), &["includes"]),
            (dictgen::InsensitiveStr::Ascii("availabed"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("rsourcing"),
                &["resourcing", "sourcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiolets"), &["toilets"]),
            (
                dictgen::InsensitiveStr::Ascii("charicterization"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parliamnetary"),
                &["parliamentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unneccesarily"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("organsims"), &["organisms"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaiator"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("physicts"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("responeses"), &["responses"]),
            (dictgen::InsensitiveStr::Ascii("chambear"), &["chamber"]),
            (dictgen::InsensitiveStr::Ascii("segnemt"), &["segment"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simaltaneously"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("nigthtime"), &["nighttime"]),
            (
                dictgen::InsensitiveStr::Ascii("anthropolgist"),
                &["anthropologist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathameticians"),
                &["mathematicians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoziating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasactional"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("questoins"), &["questions"]),
            (
                dictgen::InsensitiveStr::Ascii("assue"),
                &["assume", "assure"],
            ),
            (dictgen::InsensitiveStr::Ascii("scripoting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("costumizers"),
                &["customizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shoudered"),
                &["shouldered", "shuddered"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimism"), &["optimism"]),
            (
                dictgen::InsensitiveStr::Ascii("civelization"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("specififc"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("indiviudual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecrted"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("trancending"),
                &["transcending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superstitiosi"),
                &["superstitious"],
            ),
            (dictgen::InsensitiveStr::Ascii("protectie"), &["protective"]),
            (dictgen::InsensitiveStr::Ascii("enhancin"), &["enhancing"]),
            (dictgen::InsensitiveStr::Ascii("absorbes"), &["absorbs"]),
            (
                dictgen::InsensitiveStr::Ascii("unbolievable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepere"), &["prepare"]),
            (
                dictgen::InsensitiveStr::Ascii("apparrelled"),
                &["apparelled"],
            ),
            (dictgen::InsensitiveStr::Ascii("depcition"), &["depiction"]),
            (dictgen::InsensitiveStr::Ascii("redifines"), &["redefines"]),
            (dictgen::InsensitiveStr::Ascii("laguanges"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("claravoyant"),
                &["clairvoyant"],
            ),
            (dictgen::InsensitiveStr::Ascii("laybrinth"), &["labyrinth"]),
            (
                dictgen::InsensitiveStr::Ascii("convenction"),
                &["convention", "convection"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbaitim"), &["verbatim"]),
            (dictgen::InsensitiveStr::Ascii("vetor"), &["vector", "veto"]),
            (
                dictgen::InsensitiveStr::Ascii("converstaions"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("envolution"), &["evolution"]),
            (
                dictgen::InsensitiveStr::Ascii("covnertible"),
                &["convertible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lannesters"),
                &["lannisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expceptionally"),
                &["exceptionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("remplaced"), &["replaced"]),
            (dictgen::InsensitiveStr::Ascii("trailins"), &["trailing"]),
            (dictgen::InsensitiveStr::Ascii("propoerly"), &["properly"]),
            (dictgen::InsensitiveStr::Ascii("origonal"), &["original"]),
            (dictgen::InsensitiveStr::Ascii("spashes"), &["splashes"]),
            (
                dictgen::InsensitiveStr::Ascii("nomalization"),
                &["normalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conncetors"),
                &["connectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("persit"), &["persist"]),
            (dictgen::InsensitiveStr::Ascii("genersl"), &["generals"]),
            (dictgen::InsensitiveStr::Ascii("sectin"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("releive"), &["relieve"]),
            (dictgen::InsensitiveStr::Ascii("genitales"), &["genitals"]),
            (
                dictgen::InsensitiveStr::Ascii("boostrapped"),
                &["bootstrapped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conffiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("texutres"), &["textures"]),
            (
                dictgen::InsensitiveStr::Ascii("effecitvely"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("skandinavian"),
                &["scandinavian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strageties"),
                &["strategies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("domonstrations"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intereprets"),
                &["interprets"],
            ),
            (dictgen::InsensitiveStr::Ascii("groshuries"), &["groceries"]),
            (dictgen::InsensitiveStr::Ascii("alreadh"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("elemnets"), &["elements"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleratored"),
                &["accelerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("animete"), &["animate"]),
            (dictgen::InsensitiveStr::Ascii("arbiture"), &["arbiter"]),
            (
                dictgen::InsensitiveStr::Ascii("redistributeable"),
                &["redistributable"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirosoft"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("champinoship"),
                &["championships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regocnition"),
                &["recognition"],
            ),
            (dictgen::InsensitiveStr::Ascii("doulbe"), &["double"]),
            (
                dictgen::InsensitiveStr::Ascii("importently"),
                &["importantly", "impotently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("topoplogical"),
                &["topological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disnability"),
                &["disability"],
            ),
            (dictgen::InsensitiveStr::Ascii("qulaity"), &["quality"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrandisin"),
                &["aggrandising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarmability"),
                &["programmability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tablespooon"),
                &["tablespoon"],
            ),
            (dictgen::InsensitiveStr::Ascii("unworhty"), &["unworthy"]),
            (dictgen::InsensitiveStr::Ascii("discound"), &["discounted"]),
            (
                dictgen::InsensitiveStr::Ascii("fraternaty"),
                &["fraternity"],
            ),
            (dictgen::InsensitiveStr::Ascii("academys"), &["academies"]),
            (dictgen::InsensitiveStr::Ascii("commnents"), &["comments"]),
            (
                dictgen::InsensitiveStr::Ascii("thether"),
                &["tether", "whether"],
            ),
            (dictgen::InsensitiveStr::Ascii("unampped"), &["unmapped"]),
            (
                dictgen::InsensitiveStr::Ascii("funvtionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parrallelism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("archytype"), &["archetype"]),
            (dictgen::InsensitiveStr::Ascii("mirrired"), &["mirrored"]),
            (
                dictgen::InsensitiveStr::Ascii("secrect"),
                &["secret", "secrecy"],
            ),
            (dictgen::InsensitiveStr::Ascii("heathy"), &["healthy"]),
            (
                dictgen::InsensitiveStr::Ascii("accumulato"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrangd"), &["arranged"]),
            (
                dictgen::InsensitiveStr::Ascii("perforemance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispossal"), &["disposal"]),
            (
                dictgen::InsensitiveStr::Ascii("fortunatley"),
                &["fortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("coprright"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("enigneer"), &["engineer"]),
            (dictgen::InsensitiveStr::Ascii("suspision"), &["suspicion"]),
            (dictgen::InsensitiveStr::Ascii("bruse"), &["bruise"]),
            (dictgen::InsensitiveStr::Ascii("folowwiwng"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("quarantaine"),
                &["quarantine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fraudelent"),
                &["fraudulent"],
            ),
            (dictgen::InsensitiveStr::Ascii("downloa"), &["download"]),
            (dictgen::InsensitiveStr::Ascii("generattor"), &["generator"]),
            (dictgen::InsensitiveStr::Ascii("alreday"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("socre"), &["score"]),
            (dictgen::InsensitiveStr::Ascii("overwirte"), &["overwrite"]),
            (
                dictgen::InsensitiveStr::Ascii("entty"),
                &["entry", "entity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discconets"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("wll"), &["will", "well"]),
            (dictgen::InsensitiveStr::Ascii("articats"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("peridinkle"),
                &["periwinkle"],
            ),
            (dictgen::InsensitiveStr::Ascii("archivr"), &["archiver"]),
            (
                dictgen::InsensitiveStr::Ascii("cumbersomem"),
                &["cumbersome"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antaliasing"),
                &["antialiasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dangeros"), &["dangers"]),
            (
                dictgen::InsensitiveStr::Ascii("expepecting"),
                &["expecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irresponsable"),
                &["irresponsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallees"), &["coalesces"]),
            (
                dictgen::InsensitiveStr::Ascii("reimpplemented"),
                &["reimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("likehood"), &["likelihood"]),
            (dictgen::InsensitiveStr::Ascii("decie"), &["decide"]),
            (dictgen::InsensitiveStr::Ascii("repentent"), &["repentant"]),
            (
                dictgen::InsensitiveStr::Ascii("apostropes"),
                &["apostrophe", "apostrophes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compromiss"),
                &["compromises"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissabled"), &["disabled"]),
            (dictgen::InsensitiveStr::Ascii("traliers"), &["trailers"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubtes"),
                &["distrust", "distributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upercasing"),
                &["uppercasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsio"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("projet"), &["project"]),
            (
                dictgen::InsensitiveStr::Ascii("deregiters"),
                &["deregisters"],
            ),
            (dictgen::InsensitiveStr::Ascii("htpts"), &["https"]),
            (
                dictgen::InsensitiveStr::Ascii("concentraion"),
                &["concentration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annihilaton"),
                &["annihilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("yugoslac"), &["yugoslav"]),
            (dictgen::InsensitiveStr::Ascii("blackslash"), &["backslash"]),
            (dictgen::InsensitiveStr::Ascii("sequcne"), &["sequence"]),
            (dictgen::InsensitiveStr::Ascii("intenses"), &["intents"]),
            (dictgen::InsensitiveStr::Ascii("abstracs"), &["abstracts"]),
            (
                dictgen::InsensitiveStr::Ascii("performaces"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("champoinship"),
                &["championship"],
            ),
            (dictgen::InsensitiveStr::Ascii("precceding"), &["preceding"]),
            (dictgen::InsensitiveStr::Ascii("sxmbolic"), &["symbolic"]),
            (dictgen::InsensitiveStr::Ascii("renetkon"), &["renekton"]),
            (dictgen::InsensitiveStr::Ascii("tsnunami"), &["tsunami"]),
            (
                dictgen::InsensitiveStr::Ascii("paralellised"),
                &["parallelised"],
            ),
            (dictgen::InsensitiveStr::Ascii("updgraded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("destoryer"), &["destroyer"]),
            (
                dictgen::InsensitiveStr::Ascii("deletiong"),
                &["deletion", "deleting"],
            ),
            (dictgen::InsensitiveStr::Ascii("althouth"), &["although"]),
            (
                dictgen::InsensitiveStr::Ascii("declrators"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responcibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertica"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("destrucive"),
                &["destructive"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahc"), &["cache"]),
            (
                dictgen::InsensitiveStr::Ascii("acclerates"),
                &["accelerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("easly"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("partsian"), &["partisan"]),
            (dictgen::InsensitiveStr::Ascii("buittons"), &["buttons"]),
            (dictgen::InsensitiveStr::Ascii("reagrds"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("cotnrols"), &["controls"]),
            (
                dictgen::InsensitiveStr::Ascii("nedded"),
                &["needed", "nodded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipution"),
                &["manipulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("druation"), &["duration"]),
            (dictgen::InsensitiveStr::Ascii("imgaine"), &["imagine"]),
            (dictgen::InsensitiveStr::Ascii("revoming"), &["removing"]),
            (dictgen::InsensitiveStr::Ascii("miminum"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("segragation"),
                &["segregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nawsea"), &["nausea"]),
            (dictgen::InsensitiveStr::Ascii("turrains"), &["terrains"]),
            (dictgen::InsensitiveStr::Ascii("genererics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("paragrahps"),
                &["paragraphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimneter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermarkt"),
                &["supermarket"],
            ),
            (dictgen::InsensitiveStr::Ascii("agreeement"), &["agreement"]),
            (
                dictgen::InsensitiveStr::Ascii("ungreatful"),
                &["ungrateful"],
            ),
            (dictgen::InsensitiveStr::Ascii("invoive"), &["invoice"]),
            (
                dictgen::InsensitiveStr::Ascii("specifictions"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("aorund"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("feeture"), &["feature"]),
            (dictgen::InsensitiveStr::Ascii("downloded"), &["downloaded"]),
            (
                dictgen::InsensitiveStr::Ascii("propterties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsrciption"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curcumstance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("optismied"), &["optimised"]),
            (
                dictgen::InsensitiveStr::Ascii("functionlities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalits"),
                &["capitalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("signinged"), &["signing"]),
            (
                dictgen::InsensitiveStr::Ascii("dissepointed"),
                &["disappointed"],
            ),
            (dictgen::InsensitiveStr::Ascii("unrelying"), &["underlying"]),
            (
                dictgen::InsensitiveStr::Ascii("gatherig"),
                &["gatherings", "gathering"],
            ),
            (dictgen::InsensitiveStr::Ascii("producin"), &["producing"]),
            (
                dictgen::InsensitiveStr::Ascii("commentaar"),
                &["commentary"],
            ),
            (dictgen::InsensitiveStr::Ascii("exepmtion"), &["exemption"]),
            (dictgen::InsensitiveStr::Ascii("crusdaer"), &["crusader"]),
            (
                dictgen::InsensitiveStr::Ascii("influenses"),
                &["influences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desscriptors"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calibratin"),
                &["calibration", "calibrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sirectionalities"),
                &["directionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("triallin"), &["trialling"]),
            (
                dictgen::InsensitiveStr::Ascii("omnsicient"),
                &["omniscient"],
            ),
            (dictgen::InsensitiveStr::Ascii("corfirms"), &["confirms"]),
            (
                dictgen::InsensitiveStr::Ascii("intelelctual"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("prelease"), &["prerelease"]),
            (
                dictgen::InsensitiveStr::Ascii("attriebuting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("excetutors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("expermimentel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcational"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catagorizes"),
                &["categorizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orentation"),
                &["orientation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adevntured"),
                &["adventured"],
            ),
            (dictgen::InsensitiveStr::Ascii("satisied"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("abosultely"),
                &["absolutely"],
            ),
            (dictgen::InsensitiveStr::Ascii("oveloaded"), &["overloaded"]),
            (dictgen::InsensitiveStr::Ascii("workfroce"), &["workforce"]),
            (
                dictgen::InsensitiveStr::Ascii("circomstance"),
                &["circumstance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trajcectory"),
                &["trajectory"],
            ),
            (dictgen::InsensitiveStr::Ascii("simspon"), &["simpson"]),
            (dictgen::InsensitiveStr::Ascii("confue"), &["confuse"]),
            (dictgen::InsensitiveStr::Ascii("snetries"), &["sentries"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenated"),
                &["experimented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegoations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiifiers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("varant"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("applie"),
                &["applied", "apply"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collisiton"),
                &["collusion", "collision"],
            ),
            (dictgen::InsensitiveStr::Ascii("vavriable"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("resposnible"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favoritisme"),
                &["favorites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executeable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("primay"), &["primary"]),
            (dictgen::InsensitiveStr::Ascii("extrenal"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("dfferences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funktionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remarkibly"),
                &["remarkably"],
            ),
            (dictgen::InsensitiveStr::Ascii("featruing"), &["featuring"]),
            (
                dictgen::InsensitiveStr::Ascii("depracated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutaor"), &["evaluator"]),
            (dictgen::InsensitiveStr::Ascii("coice"), &["choice"]),
            (dictgen::InsensitiveStr::Ascii("alloccates"), &["allocates"]),
            (
                dictgen::InsensitiveStr::Ascii("subcouncious"),
                &["subconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("expensve"), &["expense"]),
            (dictgen::InsensitiveStr::Ascii("analogus"), &["analogous"]),
            (dictgen::InsensitiveStr::Ascii("chernobl"), &["chernobyl"]),
            (dictgen::InsensitiveStr::Ascii("useses"), &["uses"]),
            (dictgen::InsensitiveStr::Ascii("whispace"), &["whitespace"]),
            (dictgen::InsensitiveStr::Ascii("hom"), &["home"]),
            (
                dictgen::InsensitiveStr::Ascii("interating"),
                &["iterating", "interacting", "integrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuggg"), &["debug"]),
            (
                dictgen::InsensitiveStr::Ascii("countinuing"),
                &["continuing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistreaded"),
                &["mistreated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("longe"),
                &["longer", "lounge"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trucnation"),
                &["truncation"],
            ),
            (dictgen::InsensitiveStr::Ascii("backen"), &["backend"]),
            (
                dictgen::InsensitiveStr::Ascii("reproducive"),
                &["reproductive"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicat"), &["explicate"]),
            (dictgen::InsensitiveStr::Ascii("easyest"), &["easiest"]),
            (dictgen::InsensitiveStr::Ascii("inpact"), &["impact"]),
            (dictgen::InsensitiveStr::Ascii("organisin"), &["organising"]),
            (
                dictgen::InsensitiveStr::Ascii("biridectionality"),
                &["bidirectionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("embdders"), &["embedders"]),
            (dictgen::InsensitiveStr::Ascii("sidde"), &["side"]),
            (
                dictgen::InsensitiveStr::Ascii("susceded"),
                &["seceded", "succeeded"],
            ),
            (dictgen::InsensitiveStr::Ascii("affitnity"), &["affinity"]),
            (
                dictgen::InsensitiveStr::Ascii("uncoming"),
                &["upcoming", "oncoming"],
            ),
            (dictgen::InsensitiveStr::Ascii("miminise"), &["minimise"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialzize"),
                &["deserialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitimg"),
                &["delimiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("employement"),
                &["employment"],
            ),
            (dictgen::InsensitiveStr::Ascii("certaint"), &["certain"]),
            (dictgen::InsensitiveStr::Ascii("contniue"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("certificstes"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implamentation"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independly"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bimillenium"),
                &["bimillennium"],
            ),
            (dictgen::InsensitiveStr::Ascii("brusseles"), &["brussels"]),
            (dictgen::InsensitiveStr::Ascii("roates"), &["rotates"]),
            (dictgen::InsensitiveStr::Ascii("latiudes"), &["latitudes"]),
            (
                dictgen::InsensitiveStr::Ascii("abstacting"),
                &["abstracting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calidators"),
                &["validators"],
            ),
            (dictgen::InsensitiveStr::Ascii("palstics"), &["plastics"]),
            (dictgen::InsensitiveStr::Ascii("desicate"), &["desiccate"]),
            (
                dictgen::InsensitiveStr::Ascii("coners"),
                &["corners", "coroners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supsecting"),
                &["suspecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reporposed"),
                &["repurposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("semaphone"), &["semaphore"]),
            (dictgen::InsensitiveStr::Ascii("comitment"), &["commitment"]),
            (
                dictgen::InsensitiveStr::Ascii("adverticed"),
                &["advertised"],
            ),
            (dictgen::InsensitiveStr::Ascii("afater"), &["after"]),
            (
                dictgen::InsensitiveStr::Ascii("trampolene"),
                &["trampoline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubcribing"),
                &["unsubscribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounads"), &["bounds"]),
            (
                dictgen::InsensitiveStr::Ascii("businessnes"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("nusance"), &["nuisance"]),
            (
                dictgen::InsensitiveStr::Ascii("horizintally"),
                &["horizontally"],
            ),
            (dictgen::InsensitiveStr::Ascii("invesment"), &["investment"]),
            (dictgen::InsensitiveStr::Ascii("cermony"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("modile"),
                &["mobile", "module"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecttions"),
                &["connections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpoloating"),
                &["interpolating"],
            ),
            (dictgen::InsensitiveStr::Ascii("myu"), &["my"]),
            (
                dictgen::InsensitiveStr::Ascii("certificed"),
                &["certificated"],
            ),
            (dictgen::InsensitiveStr::Ascii("shineing"), &["shining"]),
            (dictgen::InsensitiveStr::Ascii("mimc"), &["mimic"]),
            (dictgen::InsensitiveStr::Ascii("wuestions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("schol"), &["school"]),
            (dictgen::InsensitiveStr::Ascii("grbber"), &["grabber"]),
            (
                dictgen::InsensitiveStr::Ascii("significanlty"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("outlieing"), &["outlying"]),
            (dictgen::InsensitiveStr::Ascii("redering"), &["rendering"]),
            (
                dictgen::InsensitiveStr::Ascii("automoblie"),
                &["automobile"],
            ),
            (dictgen::InsensitiveStr::Ascii("plesant"), &["pleasant"]),
            (
                dictgen::InsensitiveStr::Ascii("standarization"),
                &["standardization"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusstomer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaiators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("filal"), &["final"]),
            (dictgen::InsensitiveStr::Ascii("pelase"), &["please"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("differes"), &["differs"]),
            (
                dictgen::InsensitiveStr::Ascii("homophibic"),
                &["homophobic"],
            ),
            (dictgen::InsensitiveStr::Ascii("effeect"), &["effect"]),
            (
                dictgen::InsensitiveStr::Ascii("mediocrety"),
                &["mediocrity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestato"),
                &["protestant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesysthems"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresopnds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extractinos"),
                &["extractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("personhoood"),
                &["personhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misspellng"),
                &["misspelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("aesthatic"), &["aesthetic"]),
            (dictgen::InsensitiveStr::Ascii("checke"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("sisser"),
                &["scissor", "sissier", "sister"],
            ),
            (dictgen::InsensitiveStr::Ascii("govenor"), &["governor"]),
            (
                dictgen::InsensitiveStr::Ascii("possibililty"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypvervisors"),
                &["hypervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjustification"),
                &["justification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jacksonvillle"),
                &["jacksonville"],
            ),
            (dictgen::InsensitiveStr::Ascii("rescuin"), &["rescuing"]),
            (dictgen::InsensitiveStr::Ascii("bastardus"), &["bastards"]),
            (
                dictgen::InsensitiveStr::Ascii("flucutations"),
                &["fluctuations"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaiation"), &["variation"]),
            (dictgen::InsensitiveStr::Ascii("hegiht"), &["height"]),
            (dictgen::InsensitiveStr::Ascii("occcurs"), &["occurs"]),
            (dictgen::InsensitiveStr::Ascii("adresss"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("simgle"), &["single"]),
            (
                dictgen::InsensitiveStr::Ascii("calcualtes"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predomiantly"),
                &["predominately"],
            ),
            (dictgen::InsensitiveStr::Ascii("analysier"), &["analyser"]),
            (
                dictgen::InsensitiveStr::Ascii("forwwarding"),
                &["forwarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("treutned"), &["returned"]),
            (
                dictgen::InsensitiveStr::Ascii("raisons"),
                &["reasons", "raisins"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expectansy"),
                &["expectancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trucnating"),
                &["truncating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrributed"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siezes"),
                &["sizes", "seizes", "sieges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranporting"),
                &["transporting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquisistor"),
                &["inquisitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("romove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("celebritites"),
                &["celebrities"],
            ),
            (dictgen::InsensitiveStr::Ascii("beraded"), &["bearded"]),
            (
                dictgen::InsensitiveStr::Ascii("trunctation"),
                &["truncation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inserstor"), &["inserter"]),
            (
                dictgen::InsensitiveStr::Ascii("appretiating"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimitation"),
                &["intimidation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comilation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("keynoard"), &["keyboard"]),
            (dictgen::InsensitiveStr::Ascii("includeing"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("conains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("reaserched"),
                &["researched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizaitonal"),
                &["organizational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termiate"),
                &["terminate", "termite"],
            ),
            (dictgen::InsensitiveStr::Ascii("accumule"), &["accumulate"]),
            (
                dictgen::InsensitiveStr::Ascii("invurnerable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chekcerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("importen"),
                &["importance", "important"],
            ),
            (dictgen::InsensitiveStr::Ascii("mebmer"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("homosexuel"),
                &["homosexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintianers"),
                &["maintainers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sympathiers"),
                &["sympathizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimzies"), &["optimizes"]),
            (dictgen::InsensitiveStr::Ascii("cloumns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("chambre"), &["chamber"]),
            (dictgen::InsensitiveStr::Ascii("arcives"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("thowing"),
                &["throwing", "towing", "showing", "thawing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("karismatically"),
                &["charismatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitializing"),
                &["uninitializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("realtable"), &["relatable"]),
            (
                dictgen::InsensitiveStr::Ascii("twiter"),
                &["twitter", "twister"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragmenot"), &["fragment"]),
            (dictgen::InsensitiveStr::Ascii("somalija"), &["somalia"]),
            (dictgen::InsensitiveStr::Ascii("decrmenet"), &["decrement"]),
            (dictgen::InsensitiveStr::Ascii("symmtery"), &["symmetry"]),
            (
                dictgen::InsensitiveStr::Ascii("anonamously"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("opinyon"), &["opinion"]),
            (dictgen::InsensitiveStr::Ascii("eluether"), &["eleuther"]),
            (dictgen::InsensitiveStr::Ascii("tuntime"), &["runtime"]),
            (
                dictgen::InsensitiveStr::Ascii("palceholders"),
                &["placeholders"],
            ),
            (dictgen::InsensitiveStr::Ascii("anchestor"), &["ancestor"]),
            (
                dictgen::InsensitiveStr::Ascii("misteriously"),
                &["mysteriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("refreh"), &["refresh"]),
            (
                dictgen::InsensitiveStr::Ascii("impementor"),
                &["implementer"],
            ),
            (dictgen::InsensitiveStr::Ascii("previousy"), &["previously"]),
            (dictgen::InsensitiveStr::Ascii("ono"), &["one"]),
            (dictgen::InsensitiveStr::Ascii("cacel"), &["cancel"]),
            (dictgen::InsensitiveStr::Ascii("forgotton"), &["forgotten"]),
            (
                dictgen::InsensitiveStr::Ascii("resolutino"),
                &["resolutions", "resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mmaping"),
                &["mmapping", "mapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconsistently"),
                &["inconsistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("sqaudron"), &["squadron"]),
            (
                dictgen::InsensitiveStr::Ascii("succcessully"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimaneters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("mumber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("gourment"), &["gourmet"]),
            (dictgen::InsensitiveStr::Ascii("towrds"), &["towards"]),
            (dictgen::InsensitiveStr::Ascii("beleave"), &["believe"]),
            (dictgen::InsensitiveStr::Ascii("availiable"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("prohibites"), &["prohibits"]),
            (dictgen::InsensitiveStr::Ascii("callng"), &["calling"]),
            (
                dictgen::InsensitiveStr::Ascii("discoveribility"),
                &["discoverability"],
            ),
            (dictgen::InsensitiveStr::Ascii("maybelle"), &["maybelline"]),
            (dictgen::InsensitiveStr::Ascii("thius"), &["this"]),
            (
                dictgen::InsensitiveStr::Ascii("paralelizing"),
                &["parallelizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("kinf"), &["kind"]),
            (dictgen::InsensitiveStr::Ascii("navigato"), &["navigation"]),
            (
                dictgen::InsensitiveStr::Ascii("representerad"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perserverance"),
                &["perseverance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioretize"),
                &["prioritize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recogizers"),
                &["recognizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("oragnise"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("canditate"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("buggs"), &["bugs"]),
            (
                dictgen::InsensitiveStr::Ascii("recursivelly"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seacrhable"),
                &["searchable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commisssioner"),
                &["commissioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuatble"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffues"),
                &["defuse", "diffuse"],
            ),
            (dictgen::InsensitiveStr::Ascii("retarting"), &["restarting"]),
            (dictgen::InsensitiveStr::Ascii("muslces"), &["muscles"]),
            (
                dictgen::InsensitiveStr::Ascii("weathliest"),
                &["wealthiest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controlelrs"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragent"), &["fragment"]),
            (
                dictgen::InsensitiveStr::Ascii("survices"),
                &["services", "survives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singulaties"),
                &["singularities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorihtmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subirectory"),
                &["subdirectory"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetrans"), &["veterans"]),
            (dictgen::InsensitiveStr::Ascii("quered"), &["queried"]),
            (dictgen::InsensitiveStr::Ascii("tuscon"), &["tucson"]),
            (
                dictgen::InsensitiveStr::Ascii("ininterested"),
                &["uninterested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precendes"),
                &["precedences", "precedes"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinin"), &["combining"]),
            (dictgen::InsensitiveStr::Ascii("chcukles"), &["chuckles"]),
            (dictgen::InsensitiveStr::Ascii("retoring"), &["restoring"]),
            (
                dictgen::InsensitiveStr::Ascii("contstructors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiests"),
                &["enthusiasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("localzation"),
                &["localization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priorizations"),
                &["prioritizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consequentely"),
                &["consequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("decenders"), &["descenders"]),
            (dictgen::InsensitiveStr::Ascii("embeedding"), &["embedding"]),
            (dictgen::InsensitiveStr::Ascii("sometims"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("matchmakeing"),
                &["matchmaking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("keybidings"),
                &["keybindings"],
            ),
            (dictgen::InsensitiveStr::Ascii("immadiate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("modfitier"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("fonctionnal"),
                &["functional"],
            ),
            (dictgen::InsensitiveStr::Ascii("repulsie"), &["repulsive"]),
            (
                dictgen::InsensitiveStr::Ascii("disadventages"),
                &["disadvantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expresisons"),
                &["expressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconfiged"),
                &["reconfigured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funxtioning"),
                &["functioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backtacking"),
                &["backpacking"],
            ),
            (dictgen::InsensitiveStr::Ascii("compay"), &["company"]),
            (dictgen::InsensitiveStr::Ascii("reaplugs"), &["earplugs"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmacyst"),
                &["pharmacist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installators"),
                &["installers"],
            ),
            (dictgen::InsensitiveStr::Ascii("enhabcing"), &["enhancing"]),
            (dictgen::InsensitiveStr::Ascii("ancondas"), &["anacondas"]),
            (dictgen::InsensitiveStr::Ascii("timestimp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("adjasency"), &["adjacency"]),
            (
                dictgen::InsensitiveStr::Ascii("microprecessor"),
                &["microprocessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsucated"), &["obfuscated"]),
            (dictgen::InsensitiveStr::Ascii("dekstop"), &["desktop"]),
            (dictgen::InsensitiveStr::Ascii("anythin"), &["anything"]),
            (dictgen::InsensitiveStr::Ascii("calback"), &["callback"]),
            (
                dictgen::InsensitiveStr::Ascii("compatiabilities"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivlence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("itterate"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("ategory"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("registred"), &["registered"]),
            (
                dictgen::InsensitiveStr::Ascii("apoligises"),
                &["apologises"],
            ),
            (dictgen::InsensitiveStr::Ascii("bulagria"), &["bulgaria"]),
            (dictgen::InsensitiveStr::Ascii("carribbean"), &["caribbean"]),
            (dictgen::InsensitiveStr::Ascii("chasiers"), &["cashiers"]),
            (dictgen::InsensitiveStr::Ascii("utiliy"), &["utility"]),
            (dictgen::InsensitiveStr::Ascii("furneture"), &["furniture"]),
            (
                dictgen::InsensitiveStr::Ascii("exceotions"),
                &["exemptions", "exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropolys"),
                &["metropolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("meeitng"), &["meeting"]),
            (dictgen::InsensitiveStr::Ascii("monring"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("prooves"), &["proves"]),
            (dictgen::InsensitiveStr::Ascii("acceots"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("othe"), &["other"]),
            (
                dictgen::InsensitiveStr::Ascii("reallocaion"),
                &["reallocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("desctroys"), &["destroys"]),
            (
                dictgen::InsensitiveStr::Ascii("encountres"),
                &["encounters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("democratis"),
                &["democratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("fasened"), &["fastened"]),
            (
                dictgen::InsensitiveStr::Ascii("apdatation"),
                &["adaptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("transate"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("retierment"),
                &["retirement"],
            ),
            (dictgen::InsensitiveStr::Ascii("scraches"), &["scratches"]),
            (dictgen::InsensitiveStr::Ascii("oridinary"), &["ordinary"]),
            (
                dictgen::InsensitiveStr::Ascii("controlleras"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("defelct"), &["deflect"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphyics"),
                &["metaphysics"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyhsicals"), &["physicals"]),
            (
                dictgen::InsensitiveStr::Ascii("dependencly"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palastinians"),
                &["palestinians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encapsulatzion"),
                &["encapsulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("inocent"), &["innocent"]),
            (dictgen::InsensitiveStr::Ascii("kidknappee"), &["kidnappee"]),
            (dictgen::InsensitiveStr::Ascii("mythraic"), &["mithraic"]),
            (
                dictgen::InsensitiveStr::Ascii("caclculations"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("milawukee"), &["milwaukee"]),
            (dictgen::InsensitiveStr::Ascii("simillarly"), &["similarly"]),
            (
                dictgen::InsensitiveStr::Ascii("addministers"),
                &["administers"],
            ),
            (dictgen::InsensitiveStr::Ascii("celler"), &["cellar"]),
            (dictgen::InsensitiveStr::Ascii("ocupying"), &["occupying"]),
            (
                dictgen::InsensitiveStr::Ascii("circumcison"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("respnded"), &["responded"]),
            (dictgen::InsensitiveStr::Ascii("markey"), &["marquee"]),
            (dictgen::InsensitiveStr::Ascii("perosns"), &["persons"]),
            (
                dictgen::InsensitiveStr::Ascii("corresopndents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("remeber"), &["remember"]),
            (
                dictgen::InsensitiveStr::Ascii("unindended"),
                &["unintended"],
            ),
            (dictgen::InsensitiveStr::Ascii("beig"), &["being"]),
            (
                dictgen::InsensitiveStr::Ascii("activasting"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("backwar"), &["backward"]),
            (
                dictgen::InsensitiveStr::Ascii("modfiications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("tusedays"), &["tuesdays"]),
            (
                dictgen::InsensitiveStr::Ascii("negotation"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionare"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siplifying"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("immeidate"), &["immediate"]),
            (dictgen::InsensitiveStr::Ascii("douments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("spreadsheeet"),
                &["spreadsheets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastrucutre"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interogate"),
                &["interrogate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armagedddon"),
                &["armageddon"],
            ),
            (dictgen::InsensitiveStr::Ascii("versioons"), &["versions"]),
            (dictgen::InsensitiveStr::Ascii("delimitier"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("transfoms"), &["transforms"]),
            (dictgen::InsensitiveStr::Ascii("constrasts"), &["contrasts"]),
            (
                dictgen::InsensitiveStr::Ascii("aynchronous"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("atributs"), &["attributes"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacturerd"),
                &["manufactured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essentialy"),
                &["essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanetises"), &["sanitises"]),
            (dictgen::InsensitiveStr::Ascii("licker"), &["liquor"]),
            (dictgen::InsensitiveStr::Ascii("ilogical"), &["illogical"]),
            (dictgen::InsensitiveStr::Ascii("prtinf"), &["printf"]),
            (dictgen::InsensitiveStr::Ascii("prgress"), &["progress"]),
            (
                dictgen::InsensitiveStr::Ascii("hallicunations"),
                &["hallucinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provate"),
                &["private", "provide"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pluins"),
                &["plugins", "plains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foder"),
                &["coder", "folder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cerimonial"),
                &["ceremonial"],
            ),
            (dictgen::InsensitiveStr::Ascii("againts"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("woodowrking"),
                &["woodworking"],
            ),
            (dictgen::InsensitiveStr::Ascii("structore"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("auxialiary"), &["auxiliary"]),
            (dictgen::InsensitiveStr::Ascii("apperaring"), &["appearing"]),
            (dictgen::InsensitiveStr::Ascii("polgyon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("speciffied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("repressentation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("discuse"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("speciefy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("appen"),
                &["append", "happen", "aspen"],
            ),
            (dictgen::InsensitiveStr::Ascii("importat"), &["important"]),
            (dictgen::InsensitiveStr::Ascii("optmizer"), &["optimizer"]),
            (
                dictgen::InsensitiveStr::Ascii("algororithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preivously"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etenders"),
                &["extenders", "attenders"],
            ),
            (dictgen::InsensitiveStr::Ascii("nkow"), &["know"]),
            (dictgen::InsensitiveStr::Ascii("ascned"), &["ascend"]),
            (dictgen::InsensitiveStr::Ascii("sepation"), &["separation"]),
            (
                dictgen::InsensitiveStr::Ascii("resourcd"),
                &["resourced", "resource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contsructor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("mesaged"), &["messaged"]),
            (dictgen::InsensitiveStr::Ascii("mutablyy"), &["mutably"]),
            (dictgen::InsensitiveStr::Ascii("explcit"), &["explicit"]),
            (
                dictgen::InsensitiveStr::Ascii("informaitonal"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afganistan"),
                &["afghanistan"],
            ),
            (dictgen::InsensitiveStr::Ascii("exstacy"), &["ecstasy"]),
            (dictgen::InsensitiveStr::Ascii("graoh"), &["graph"]),
            (dictgen::InsensitiveStr::Ascii("caesed"), &["ceased"]),
            (
                dictgen::InsensitiveStr::Ascii("ciricularise"),
                &["circularise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexplicabley"),
                &["inexplicably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportionella"),
                &["proportionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarentees"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("querstionable"),
                &["questionable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downrgaded"),
                &["downgraded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("captialised"),
                &["capitalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("procssed"), &["processed"]),
            (
                dictgen::InsensitiveStr::Ascii("expets"),
                &["expects", "experts", "expats"],
            ),
            (dictgen::InsensitiveStr::Ascii("chidren"), &["children"]),
            (
                dictgen::InsensitiveStr::Ascii("insurences"),
                &["insurances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environemt"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sporstmanship"),
                &["sportsmanship"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlysis"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("evrywhere"), &["everywhere"]),
            (dictgen::InsensitiveStr::Ascii("dirtectory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("chiphers"), &["ciphers"]),
            (
                dictgen::InsensitiveStr::Ascii("progammability"),
                &["programmability"],
            ),
            (dictgen::InsensitiveStr::Ascii("dashbaord"), &["dashboard"]),
            (dictgen::InsensitiveStr::Ascii("unwrappede"), &["unwrapped"]),
            (
                dictgen::InsensitiveStr::Ascii("explantions"),
                &["explanations"],
            ),
            (dictgen::InsensitiveStr::Ascii("accordin"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("communcations"),
                &["communications"],
            ),
            (dictgen::InsensitiveStr::Ascii("accound"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("sacarin"), &["saccharin"]),
            (dictgen::InsensitiveStr::Ascii("iteraion"), &["iteration"]),
            (dictgen::InsensitiveStr::Ascii("somwthing"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("hirearcy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("parralellisation"),
                &["parallelisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("limitaiton"),
                &["limitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occational"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasferring"),
                &["transferring"],
            ),
            (dictgen::InsensitiveStr::Ascii("artemios"), &["artemis"]),
            (
                dictgen::InsensitiveStr::Ascii("quantitatve"),
                &["quantitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("creting"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("adjasensies"),
                &["adjacencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generall"),
                &["generally", "general"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trunctating"),
                &["truncating"],
            ),
            (dictgen::InsensitiveStr::Ascii("ortherwise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("dperecating"),
                &["deprecating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underpowerd"),
                &["underpowered"],
            ),
            (dictgen::InsensitiveStr::Ascii("livetime"), &["lifetime"]),
            (dictgen::InsensitiveStr::Ascii("dyspotian"), &["dystopian"]),
            (dictgen::InsensitiveStr::Ascii("somwwhere"), &["somewhere"]),
            (dictgen::InsensitiveStr::Ascii("overhal"), &["overhaul"]),
            (
                dictgen::InsensitiveStr::Ascii("indentified"),
                &["identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstation"),
                &["demonstration"],
            ),
            (dictgen::InsensitiveStr::Ascii("plad"), &["plaid", "plead"]),
            (
                dictgen::InsensitiveStr::Ascii("recommened"),
                &["recommend", "recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assotiation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partnetship"),
                &["partnership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occurrances"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeatd"), &["repeated"]),
            (dictgen::InsensitiveStr::Ascii("pefixes"), &["prefixes"]),
            (
                dictgen::InsensitiveStr::Ascii("intregrated"),
                &["integrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("contactos"), &["contacts"]),
            (
                dictgen::InsensitiveStr::Ascii("subscribin"),
                &["subscribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("doplhins"), &["dolphins"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomnaces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("moinitored"), &["monitored"]),
            (
                dictgen::InsensitiveStr::Ascii("watermalon"),
                &["watermelon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electoraat"),
                &["electorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professores"),
                &["professors"],
            ),
            (dictgen::InsensitiveStr::Ascii("declard"), &["declared"]),
            (dictgen::InsensitiveStr::Ascii("seconday"), &["secondary"]),
            (dictgen::InsensitiveStr::Ascii("uniquley"), &["uniquely"]),
            (dictgen::InsensitiveStr::Ascii("sepereted"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("applicances"),
                &["appliances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coloer"),
                &["color", "colour"],
            ),
            (dictgen::InsensitiveStr::Ascii("newtork"), &["network"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronizers"),
                &["synchronizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("reptuable"), &["reputable"]),
            (
                dictgen::InsensitiveStr::Ascii("experimant"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statitically"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reletnless"),
                &["relentless"],
            ),
            (dictgen::InsensitiveStr::Ascii("colokwise"), &["clockwise"]),
            (
                dictgen::InsensitiveStr::Ascii("vizualizing"),
                &["visualizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influencin"),
                &["influencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("plarized"), &["polarized"]),
            (
                dictgen::InsensitiveStr::Ascii("prealocates"),
                &["preallocates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enounctered"),
                &["encountered"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomtrys"), &["geometries"]),
            (
                dictgen::InsensitiveStr::Ascii("vocabularlies"),
                &["vocabularies"],
            ),
            (dictgen::InsensitiveStr::Ascii("techinique"), &["technique"]),
            (dictgen::InsensitiveStr::Ascii("staturday"), &["saturday"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacturedd"),
                &["manufactured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannotations"),
                &["connotations"],
            ),
            (dictgen::InsensitiveStr::Ascii("debia"), &["debian"]),
            (dictgen::InsensitiveStr::Ascii("initaling"), &["initialing"]),
            (dictgen::InsensitiveStr::Ascii("eleemnt"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("finsished"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("exisitent"), &["existent"]),
            (
                dictgen::InsensitiveStr::Ascii("mimatching"),
                &["mismatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("entyr"), &["entry", "enter"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstansial"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorthns"), &["algorithms"]),
            (dictgen::InsensitiveStr::Ascii("geomery"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("unrary"), &["unary"]),
            (
                dictgen::InsensitiveStr::Ascii("metaphysicals"),
                &["metaphysics"],
            ),
            (dictgen::InsensitiveStr::Ascii("seramony"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("familairize"),
                &["familiarize"],
            ),
            (dictgen::InsensitiveStr::Ascii("acitve"), &["active"]),
            (dictgen::InsensitiveStr::Ascii("partitons"), &["partitions"]),
            (
                dictgen::InsensitiveStr::Ascii("advertisted"),
                &["advertised"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabints"), &["cabinets"]),
            (
                dictgen::InsensitiveStr::Ascii("containes"),
                &["contains", "contained", "container"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nofication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("strng"), &["string"]),
            (dictgen::InsensitiveStr::Ascii("dokcs"), &["docks"]),
            (dictgen::InsensitiveStr::Ascii("sharplay"), &["sharply"]),
            (dictgen::InsensitiveStr::Ascii("nwo"), &["now"]),
            (
                dictgen::InsensitiveStr::Ascii("unitesstates"),
                &["unitedstates"],
            ),
            (dictgen::InsensitiveStr::Ascii("reeaser"), &["releaser"]),
            (dictgen::InsensitiveStr::Ascii("karisma"), &["charisma"]),
            (dictgen::InsensitiveStr::Ascii("converion"), &["conversion"]),
            (
                dictgen::InsensitiveStr::Ascii("altrusitic"),
                &["altruistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exacberated"),
                &["exacerbated"],
            ),
            (dictgen::InsensitiveStr::Ascii("speperater"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("gimmickey"), &["gimmicky"]),
            (dictgen::InsensitiveStr::Ascii("mintute"), &["minute"]),
            (
                dictgen::InsensitiveStr::Ascii("cimptumatic"),
                &["symptomatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("switiches"), &["switches"]),
            (dictgen::InsensitiveStr::Ascii("wornged"), &["wronged"]),
            (
                dictgen::InsensitiveStr::Ascii("depandents"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("tthese"), &["these"]),
            (
                dictgen::InsensitiveStr::Ascii("vizualises"),
                &["visualises"],
            ),
            (dictgen::InsensitiveStr::Ascii("enseble"), &["ensemble"]),
            (dictgen::InsensitiveStr::Ascii("refletion"), &["reflection"]),
            (dictgen::InsensitiveStr::Ascii("srirachia"), &["sriracha"]),
            (dictgen::InsensitiveStr::Ascii("jorurnals"), &["journals"]),
            (
                dictgen::InsensitiveStr::Ascii("reenforcements"),
                &["reinforcements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscirbes"),
                &["subscribes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categorues"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("confures"), &["configures"]),
            (
                dictgen::InsensitiveStr::Ascii("aviability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envrionments"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("companys"), &["companies"]),
            (
                dictgen::InsensitiveStr::Ascii("addiitonally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("relvance"), &["relevance"]),
            (
                dictgen::InsensitiveStr::Ascii("associationthis"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreacates"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("carnomy"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("appars"), &["appears"]),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiated"),
                &["renegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opinyonists"),
                &["opinionists"],
            ),
            (dictgen::InsensitiveStr::Ascii("ballances"), &["balances"]),
            (dictgen::InsensitiveStr::Ascii("alignmnts"), &["alignments"]),
            (dictgen::InsensitiveStr::Ascii("seramonys"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("mutithreaded"),
                &["multithreaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("swaskita"), &["swastika"]),
            (
                dictgen::InsensitiveStr::Ascii("genererally"),
                &["generally"],
            ),
            (dictgen::InsensitiveStr::Ascii("aerosapce"), &["aerospace"]),
            (dictgen::InsensitiveStr::Ascii("generl"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("subtance"), &["substance"]),
            (
                dictgen::InsensitiveStr::Ascii("tranceviers"),
                &["transceivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("thsose"), &["those"]),
            (
                dictgen::InsensitiveStr::Ascii("paarameters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("alloved"), &["allowed"]),
            (
                dictgen::InsensitiveStr::Ascii("flourescence"),
                &["fluorescence", "florescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofuses"), &["confuses"]),
            (dictgen::InsensitiveStr::Ascii("aggrement"), &["agreement"]),
            (dictgen::InsensitiveStr::Ascii("presever"), &["preserve"]),
            (dictgen::InsensitiveStr::Ascii("protocos"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("percentil"), &["percentile"]),
            (
                dictgen::InsensitiveStr::Ascii("mediterranen"),
                &["mediterranean"],
            ),
            (dictgen::InsensitiveStr::Ascii("invalud"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("availabbility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutate"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptonal"),
                &["exceptional"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplly"), &["apply"]),
            (
                dictgen::InsensitiveStr::Ascii("proficiant"),
                &["proficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("drapin"), &["draping"]),
            (dictgen::InsensitiveStr::Ascii("nostlagia"), &["nostalgia"]),
            (
                dictgen::InsensitiveStr::Ascii("regenrated"),
                &["regenerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theroetically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mabyelline"),
                &["maybelline"],
            ),
            (dictgen::InsensitiveStr::Ascii("poisioning"), &["poisoning"]),
            (
                dictgen::InsensitiveStr::Ascii("decentrilized"),
                &["decentralized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arugmentative"),
                &["argumentative"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaaluables"), &["valuables"]),
            (dictgen::InsensitiveStr::Ascii("emphysyma"), &["emphysema"]),
            (dictgen::InsensitiveStr::Ascii("informas"), &["informs"]),
            (
                dictgen::InsensitiveStr::Ascii("horisontal"),
                &["horizontal"],
            ),
            (dictgen::InsensitiveStr::Ascii("acatemia"), &["academia"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciatied"),
                &["appreciative", "appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("openin"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("dicotomies"),
                &["dichotomies"],
            ),
            (dictgen::InsensitiveStr::Ascii("deatlhy"), &["deathly"]),
            (
                dictgen::InsensitiveStr::Ascii("ecoligical"),
                &["ecological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approimated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("somewhow"), &["somehow"]),
            (
                dictgen::InsensitiveStr::Ascii("assemalate"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("critisicm"), &["criticism"]),
            (
                dictgen::InsensitiveStr::Ascii("develpoment"),
                &["developments", "development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommonded"),
                &["recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particaluarly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seledtions"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("assighs"), &["assigns"]),
            (dictgen::InsensitiveStr::Ascii("srunkn"), &["shrunken"]),
            (dictgen::InsensitiveStr::Ascii("requiretd"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("missable"), &["miscible"]),
            (
                dictgen::InsensitiveStr::Ascii("intilizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignoded"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("parralellises"),
                &["parallelises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efectiveness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("govenrments"),
                &["governments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactersistic"),
                &["characteristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conributor"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("horizntal"), &["horizontal"]),
            (
                dictgen::InsensitiveStr::Ascii("lackrimosly"),
                &["lachrymosely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitialised"),
                &["uninitialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frientships"),
                &["friendships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprectaed"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("adnroid"), &["android"]),
            (dictgen::InsensitiveStr::Ascii("naviagtor"), &["navigator"]),
            (
                dictgen::InsensitiveStr::Ascii("indigineous"),
                &["indigenous"],
            ),
            (dictgen::InsensitiveStr::Ascii("locaizer"), &["localizer"]),
            (dictgen::InsensitiveStr::Ascii("obsosete"), &["obsolete"]),
            (dictgen::InsensitiveStr::Ascii("chechks"), &["checks"]),
            (dictgen::InsensitiveStr::Ascii("receviing"), &["receiving"]),
            (
                dictgen::InsensitiveStr::Ascii("approprietely"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neaders"),
                &["headers", "kneaders"],
            ),
            (dictgen::InsensitiveStr::Ascii("adpaters"), &["adapters"]),
            (dictgen::InsensitiveStr::Ascii("arbitralry"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("catastraphic"),
                &["catastrophic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrutpable"),
                &["interruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoery"), &["theory"]),
            (dictgen::InsensitiveStr::Ascii("abailable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("undrestand"),
                &["understand"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misintepreted"),
                &["misinterpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("currnet"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("distingushed"),
                &["distinguished"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaite"), &["animate"]),
            (
                dictgen::InsensitiveStr::Ascii("necessarilly"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("cehckers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("varaition"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("nortifications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainains"), &["maintains"]),
            (dictgen::InsensitiveStr::Ascii("organites"), &["organise"]),
            (
                dictgen::InsensitiveStr::Ascii("oppertunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("caceled"), &["canceled"]),
            (dictgen::InsensitiveStr::Ascii("ariplanes"), &["airplanes"]),
            (
                dictgen::InsensitiveStr::Ascii("misrepresention"),
                &["misrepresenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("banchmarking"),
                &["benchmarking"],
            ),
            (dictgen::InsensitiveStr::Ascii("personatus"), &["personas"]),
            (dictgen::InsensitiveStr::Ascii("sacales"), &["scales"]),
            (
                dictgen::InsensitiveStr::Ascii("dereferances"),
                &["dereferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("relpased"), &["relapsed"]),
            (dictgen::InsensitiveStr::Ascii("therofer"), &["therefore"]),
            (dictgen::InsensitiveStr::Ascii("reename"), &["rename"]),
            (
                dictgen::InsensitiveStr::Ascii("authrorises"),
                &["authorises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ovveriding"),
                &["overriding"],
            ),
            (dictgen::InsensitiveStr::Ascii("expediated"), &["expedited"]),
            (
                dictgen::InsensitiveStr::Ascii("bunded"),
                &["binded", "bounded", "bundled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimateors"),
                &["estimators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegeterian"),
                &["vegetarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancestores"), &["ancestors"]),
            (
                dictgen::InsensitiveStr::Ascii("conjuctions"),
                &["conjunctions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mosty"), &["mostly"]),
            (
                dictgen::InsensitiveStr::Ascii("asscoitaed"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("esimations"),
                &["estimations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalisten"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("reasy"), &["ready", "easy"]),
            (
                dictgen::InsensitiveStr::Ascii("moderatley"),
                &["moderately"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomorr"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("declere"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("rapelling"), &["rappelling"]),
            (dictgen::InsensitiveStr::Ascii("restrics"), &["restricts"]),
            (dictgen::InsensitiveStr::Ascii("compleate"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("fillements"), &["filaments"]),
            (dictgen::InsensitiveStr::Ascii("siganture"), &["signature"]),
            (dictgen::InsensitiveStr::Ascii("staatus"), &["status"]),
            (dictgen::InsensitiveStr::Ascii("termial"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("overvrite"), &["overwrite"]),
            (dictgen::InsensitiveStr::Ascii("tablulates"), &["tabulates"]),
            (dictgen::InsensitiveStr::Ascii("prodided"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("renegogtiates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("modefiers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("inherrit"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("ananlyzers"), &["analyzers"]),
            (dictgen::InsensitiveStr::Ascii("inisially"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("reoder"), &["reorder"]),
            (dictgen::InsensitiveStr::Ascii("yoiu"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("mutltiple"), &["multiple"]),
            (dictgen::InsensitiveStr::Ascii("frenquency"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("colissions"),
                &["collisions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensativity"),
                &["sensitivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("reminiser"), &["reminiscer"]),
            (
                dictgen::InsensitiveStr::Ascii("groupes"),
                &["groups", "grouped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cratashous"),
                &["cretaceous"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpout"), &["input"]),
            (
                dictgen::InsensitiveStr::Ascii("verfifiable"),
                &["verifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("influening"),
                &["influencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ebaled"), &["enabled"]),
            (
                dictgen::InsensitiveStr::Ascii("discracing"),
                &["disgracing"],
            ),
            (dictgen::InsensitiveStr::Ascii("activistes"), &["activists"]),
            (
                dictgen::InsensitiveStr::Ascii("microphonies"),
                &["microphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("suffux"), &["suffix"]),
            (dictgen::InsensitiveStr::Ascii("meansure"), &["measure"]),
            (dictgen::InsensitiveStr::Ascii("apartmets"), &["apartments"]),
            (dictgen::InsensitiveStr::Ascii("collectifs"), &["collects"]),
            (dictgen::InsensitiveStr::Ascii("uanble"), &["unable"]),
            (dictgen::InsensitiveStr::Ascii("enounters"), &["encounters"]),
            (
                dictgen::InsensitiveStr::Ascii("additianally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demoninator"),
                &["denominator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("promting"),
                &["promoting", "prompting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prediciment"),
                &["predicament"],
            ),
            (dictgen::InsensitiveStr::Ascii("requrire"), &["require"]),
            (
                dictgen::InsensitiveStr::Ascii("munnicipality"),
                &["municipality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosporous"),
                &["prosperous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volitality"),
                &["volatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("assement"), &["assessment"]),
            (
                dictgen::InsensitiveStr::Ascii("interactios"),
                &["interacts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asyncronously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferrmance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("frequnt"), &["frequent"]),
            (dictgen::InsensitiveStr::Ascii("mannual"), &["manual"]),
            (dictgen::InsensitiveStr::Ascii("netcape"), &["netscape"]),
            (
                dictgen::InsensitiveStr::Ascii("oveflowing"),
                &["overflowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructcor"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintance"),
                &["maintenance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geomtrically"),
                &["geometrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitifies"),
                &["identifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornagraphy"),
                &["pornography"],
            ),
            (dictgen::InsensitiveStr::Ascii("resapwn"), &["respawn"]),
            (dictgen::InsensitiveStr::Ascii("bouunds"), &["bounds"]),
            (dictgen::InsensitiveStr::Ascii("anymoer"), &["anymore"]),
            (
                dictgen::InsensitiveStr::Ascii("defintiely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experementer"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deuplicated"),
                &["deduplicated", "duplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armageddan"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modfification"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradicional"),
                &["traditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implememetation"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("formallize"), &["formalize"]),
            (
                dictgen::InsensitiveStr::Ascii("intened"),
                &["intended", "interned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achiveable"),
                &["achievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dignonals"), &["diagonals"]),
            (
                dictgen::InsensitiveStr::Ascii("bracelettes"),
                &["bracelets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inhabitat"),
                &["inhabitants"],
            ),
            (dictgen::InsensitiveStr::Ascii("hande"), &["handle", "hand"]),
            (dictgen::InsensitiveStr::Ascii("coallesed"), &["coalesced"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistented"),
                &["inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("furished"), &["furnished"]),
            (
                dictgen::InsensitiveStr::Ascii("adrelanine"),
                &["adrenaline"],
            ),
            (dictgen::InsensitiveStr::Ascii("thiests"), &["theists"]),
            (dictgen::InsensitiveStr::Ascii("anwered"), &["answered"]),
            (
                dictgen::InsensitiveStr::Ascii("guareenteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expontentially"),
                &["exponentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reoutes"),
                &["routes", "reroutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchoronously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("beachead"), &["beachhead"]),
            (dictgen::InsensitiveStr::Ascii("interfac"), &["interface"]),
            (dictgen::InsensitiveStr::Ascii("lenngth"), &["length"]),
            (
                dictgen::InsensitiveStr::Ascii("tranformed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("firstg"), &["first"]),
            (
                dictgen::InsensitiveStr::Ascii("cernomially"),
                &["ceremonially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estabilshment"),
                &["establishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("librabries"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("wyth"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetents"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respons"),
                &["response", "respond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appproached"),
                &["approached"],
            ),
            (dictgen::InsensitiveStr::Ascii("retutns"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("curisoity"), &["curiosity"]),
            (dictgen::InsensitiveStr::Ascii("serched"), &["searched"]),
            (dictgen::InsensitiveStr::Ascii("allredy"), &["already"]),
            (dictgen::InsensitiveStr::Ascii("cymk"), &["cmyk"]),
            (dictgen::InsensitiveStr::Ascii("arument"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("microseonds"),
                &["microseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorithmmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("invesre"), &["inverse"]),
            (
                dictgen::InsensitiveStr::Ascii("inflatin"),
                &["inflating", "inflation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensetional"),
                &["sensational"],
            ),
            (dictgen::InsensitiveStr::Ascii("offical"), &["official"]),
            (dictgen::InsensitiveStr::Ascii("peroid"), &["period"]),
            (
                dictgen::InsensitiveStr::Ascii("straigthening"),
                &["straightening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thetrahedron"),
                &["tetrahedron"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implimention"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tansfomed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("blieves"), &["believes"]),
            (
                dictgen::InsensitiveStr::Ascii("concludendo"),
                &["concluded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("providencie"),
                &["providence"],
            ),
            (dictgen::InsensitiveStr::Ascii("addding"), &["adding"]),
            (
                dictgen::InsensitiveStr::Ascii("exibitions"),
                &["exhibitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomparible"),
                &["incompatible", "incomparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("coerscion"), &["coercion"]),
            (
                dictgen::InsensitiveStr::Ascii("reposititioning"),
                &["repositioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("optained"), &["obtained"]),
            (
                dictgen::InsensitiveStr::Ascii("nessasarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizates"),
                &["organizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("exatctly"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("experimintelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("rusteld"), &["rustled"]),
            (dictgen::InsensitiveStr::Ascii("quiting"), &["quitting"]),
            (
                dictgen::InsensitiveStr::Ascii("infrasctructures"),
                &["infrastructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("apenders"), &["appenders"]),
            (
                dictgen::InsensitiveStr::Ascii("intelligenty"),
                &["intelligently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comperhend"),
                &["comprehend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acculumates"),
                &["accumulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("multipies"), &["multiplies"]),
            (dictgen::InsensitiveStr::Ascii("insurace"), &["insurance"]),
            (dictgen::InsensitiveStr::Ascii("exporintg"), &["exporting"]),
            (
                dictgen::InsensitiveStr::Ascii("consultati"),
                &["consultant"],
            ),
            (dictgen::InsensitiveStr::Ascii("liase"), &["liaise"]),
            (dictgen::InsensitiveStr::Ascii("unvierses"), &["universes"]),
            (dictgen::InsensitiveStr::Ascii("materal"), &["material"]),
            (dictgen::InsensitiveStr::Ascii("palyoffs"), &["playoffs"]),
            (dictgen::InsensitiveStr::Ascii("worng"), &["wrong", "worn"]),
            (
                dictgen::InsensitiveStr::Ascii("questionairre"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intendes"),
                &["intends", "intended"],
            ),
            (dictgen::InsensitiveStr::Ascii("styile"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitvities"),
                &["sensitivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deprectated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tailban"), &["taliban"]),
            (
                dictgen::InsensitiveStr::Ascii("appreaciate"),
                &["appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothesizin"),
                &["hypothesizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trignametry"),
                &["trigonometry"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ratatooee"),
                &["ratatouille"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optinally"),
                &["optimally", "optionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disabiguation"),
                &["disambiguation"],
            ),
            (dictgen::InsensitiveStr::Ascii("apeends"), &["appends"]),
            (dictgen::InsensitiveStr::Ascii("deativate"), &["deactivate"]),
            (
                dictgen::InsensitiveStr::Ascii("combies"),
                &["combines", "zombies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initates"),
                &["imitates", "initiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("defusin"), &["defusing"]),
            (dictgen::InsensitiveStr::Ascii("interfal"), &["interval"]),
            (dictgen::InsensitiveStr::Ascii("critiical"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("ouline"),
                &["outline", "online"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congfigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("courthosue"),
                &["courthouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caculation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("amplifyer"), &["amplifier"]),
            (
                dictgen::InsensitiveStr::Ascii("koordinates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conectivities"),
                &["connectivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marketpalce"),
                &["marketplace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeting"),
                &["repeating", "reputing", "repenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sympathyzers"),
                &["sympathizers"],
            ),
            (dictgen::InsensitiveStr::Ascii("loosley"), &["loosely"]),
            (dictgen::InsensitiveStr::Ascii("ncie"), &["nice"]),
            (
                dictgen::InsensitiveStr::Ascii("transisitioned"),
                &["transitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("withion"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("infiltratie"),
                &["infiltrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specialistes"),
                &["specialists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unmber"),
                &["number", "umber"],
            ),
            (dictgen::InsensitiveStr::Ascii("surports"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("abreviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bureaucracts"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("toal"), &["total", "toad"]),
            (dictgen::InsensitiveStr::Ascii("easili"), &["easily"]),
            (
                dictgen::InsensitiveStr::Ascii("downvotesd"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimazing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("bistream"), &["bitstream"]),
            (
                dictgen::InsensitiveStr::Ascii("participents"),
                &["participants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermenets"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("avoind"), &["avoid"]),
            (dictgen::InsensitiveStr::Ascii("sinnically"), &["cynically"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomramce"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuffes"), &["debuffs"]),
            (
                dictgen::InsensitiveStr::Ascii("amelioratin"),
                &["ameliorating", "amelioration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constcurts"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loaclizing"),
                &["localizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selecte"),
                &["selected", "select"],
            ),
            (dictgen::InsensitiveStr::Ascii("expetected"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("pariticularly"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("hass"), &["hash"]),
            (dictgen::InsensitiveStr::Ascii("expceted"), &["expected"]),
            (
                dictgen::InsensitiveStr::Ascii("intitiating"),
                &["initiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condicional"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("normalyy"), &["normally"]),
            (
                dictgen::InsensitiveStr::Ascii("commtitees"),
                &["committees"],
            ),
            (dictgen::InsensitiveStr::Ascii("motorolja"), &["motorola"]),
            (
                dictgen::InsensitiveStr::Ascii("progressivley"),
                &["progressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("regitered"), &["registered"]),
            (dictgen::InsensitiveStr::Ascii("segmends"), &["segments"]),
            (dictgen::InsensitiveStr::Ascii("repeatedy"), &["repeatedly"]),
            (dictgen::InsensitiveStr::Ascii("genearl"), &["general"]),
            (dictgen::InsensitiveStr::Ascii("monopolly"), &["monopoly"]),
            (dictgen::InsensitiveStr::Ascii("stancels"), &["stances"]),
            (dictgen::InsensitiveStr::Ascii("coopearte"), &["cooperate"]),
            (dictgen::InsensitiveStr::Ascii("authror"), &["author"]),
            (dictgen::InsensitiveStr::Ascii("universels"), &["universes"]),
            (dictgen::InsensitiveStr::Ascii("empricical"), &["empirical"]),
            (
                dictgen::InsensitiveStr::Ascii("philosiphical"),
                &["philosophical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imrpovements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilalizing"),
                &["initializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("taked"), &["taken", "took"]),
            (dictgen::InsensitiveStr::Ascii("removin"), &["removing"]),
            (
                dictgen::InsensitiveStr::Ascii("depretiats"),
                &["depreciates", "deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("subscirbe"), &["subscribe"]),
            (dictgen::InsensitiveStr::Ascii("solum"), &["solemn"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreegator"),
                &["aggregator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complimentarity"),
                &["complimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimeted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("coalace"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("cumbersone"),
                &["cumbersome"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranmist"), &["transmit"]),
            (dictgen::InsensitiveStr::Ascii("cauing"), &["causing"]),
            (dictgen::InsensitiveStr::Ascii("comision"), &["commission"]),
            (dictgen::InsensitiveStr::Ascii("vinrator"), &["vibrator"]),
            (
                dictgen::InsensitiveStr::Ascii("uncomfortabley"),
                &["uncomfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correctoins"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("singulat"), &["singular"]),
            (dictgen::InsensitiveStr::Ascii("hallaluya"), &["hallelujah"]),
            (
                dictgen::InsensitiveStr::Ascii("exaplaining"),
                &["explaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inexpirience"),
                &["inexperience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspinds"),
                &["corresponds"],
            ),
            (dictgen::InsensitiveStr::Ascii("grapics"), &["graphics"]),
            (
                dictgen::InsensitiveStr::Ascii("amelearates"),
                &["ameliorates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recourcing"),
                &["resourcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbelieveable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("tripticks"), &["triptychs"]),
            (
                dictgen::InsensitiveStr::Ascii("statis"),
                &["stasis", "status"],
            ),
            (dictgen::InsensitiveStr::Ascii("successul"), &["successful"]),
            (
                dictgen::InsensitiveStr::Ascii("awesoneness"),
                &["awesomeness"],
            ),
            (dictgen::InsensitiveStr::Ascii("emoiji"), &["emoji"]),
            (
                dictgen::InsensitiveStr::Ascii("masculenity"),
                &["masculinity"],
            ),
            (dictgen::InsensitiveStr::Ascii("mismash"), &["mishmash"]),
            (
                dictgen::InsensitiveStr::Ascii("apllication"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("checcks"), &["checks"]),
            (
                dictgen::InsensitiveStr::Ascii("fiter"),
                &["filter", "fighter", "fitter", "fiver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("empirialism"),
                &["imperialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("logiteh"), &["logitech"]),
            (
                dictgen::InsensitiveStr::Ascii("accumalation"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dadlock"), &["deadlock"]),
            (
                dictgen::InsensitiveStr::Ascii("exporers"),
                &["explorers", "exporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tesing"),
                &["testing", "teasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiatied"),
                &["negotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("assingees"), &["assignees"]),
            (
                dictgen::InsensitiveStr::Ascii("characaters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("origniated"),
                &["originated"],
            ),
            (dictgen::InsensitiveStr::Ascii("disscts"), &["dissects"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospoinding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reachabililty"),
                &["reachability"],
            ),
            (dictgen::InsensitiveStr::Ascii("standartds"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("sattelit"), &["satellite"]),
            (dictgen::InsensitiveStr::Ascii("diggestive"), &["digestive"]),
            (
                dictgen::InsensitiveStr::Ascii("asscciations"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comlicates"),
                &["complicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scarches"),
                &["searches", "scorches", "scratches", "starches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ellementals"),
                &["elementals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("halluciantions"),
                &["hallucinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalueated"), &["evaluated"]),
            (dictgen::InsensitiveStr::Ascii("jounaling"), &["journaling"]),
            (dictgen::InsensitiveStr::Ascii("apparenty"), &["apparently"]),
            (dictgen::InsensitiveStr::Ascii("terminiate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("arramges"), &["arranges"]),
            (dictgen::InsensitiveStr::Ascii("curefuly"), &["carefully"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstaling"),
                &["reinstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("conifgure"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("extraccion"),
                &["extraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suplementing"),
                &["supplementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presention"),
                &["presenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responc"),
                &["response", "respond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanitified"),
                &["quantified"],
            ),
            (dictgen::InsensitiveStr::Ascii("venyet"), &["vignette"]),
            (dictgen::InsensitiveStr::Ascii("requred"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("excplicitly"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("writingm"), &["writing"]),
            (dictgen::InsensitiveStr::Ascii("tiemout"), &["timeout"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcualtions"),
                &["recalculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("uptreamer"), &["upstreamer"]),
            (
                dictgen::InsensitiveStr::Ascii("juggeranut"),
                &["juggernaut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priodically"),
                &["periodically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mulitple"), &["multiple"]),
            (
                dictgen::InsensitiveStr::Ascii("repositotry"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inprovement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("barbarina"), &["barbarian"]),
            (dictgen::InsensitiveStr::Ascii("ananlyses"), &["analyses"]),
            (
                dictgen::InsensitiveStr::Ascii("articualting"),
                &["articulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delimitters"),
                &["delimiters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastiantion"),
                &["procrastination"],
            ),
            (dictgen::InsensitiveStr::Ascii("helth"), &["health"]),
            (
                dictgen::InsensitiveStr::Ascii("midfieldiers"),
                &["midfielders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegoates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("noteboook"), &["notebook"]),
            (dictgen::InsensitiveStr::Ascii("similart"), &["similarity"]),
            (dictgen::InsensitiveStr::Ascii("glitchey"), &["glitchy"]),
            (
                dictgen::InsensitiveStr::Ascii("supermakret"),
                &["supermarkets"],
            ),
            (dictgen::InsensitiveStr::Ascii("documments"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("onbaord"), &["onboard"]),
            (dictgen::InsensitiveStr::Ascii("repetead"), &["repeated"]),
            (
                dictgen::InsensitiveStr::Ascii("dissasemble"),
                &["disassemble"],
            ),
            (dictgen::InsensitiveStr::Ascii("whithe"), &["with", "white"]),
            (dictgen::InsensitiveStr::Ascii("senes"), &["sense"]),
            (
                dictgen::InsensitiveStr::Ascii("aftherwards"),
                &["afterwards"],
            ),
            (dictgen::InsensitiveStr::Ascii("medifor"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("booshelf"), &["bookshelf"]),
            (
                dictgen::InsensitiveStr::Ascii("enouraging"),
                &["encouraging"],
            ),
            (dictgen::InsensitiveStr::Ascii("furmalae"), &["formulae"]),
            (dictgen::InsensitiveStr::Ascii("parmters"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("micrcontrollers"),
                &["microcontrollers"],
            ),
            (dictgen::InsensitiveStr::Ascii("propertes"), &["properties"]),
            (
                dictgen::InsensitiveStr::Ascii("skatebaords"),
                &["skateboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("osciallator"),
                &["oscillator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("travellerhd"),
                &["travelled"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparred"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("pronouning"),
                &["pronouncing"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagonale"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("cihers"), &["ciphers"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciateive"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconcsious"),
                &["unconscious"],
            ),
            (dictgen::InsensitiveStr::Ascii("abothe"), &["above"]),
            (dictgen::InsensitiveStr::Ascii("scubscribe"), &["subscribe"]),
            (dictgen::InsensitiveStr::Ascii("omage"), &["homage"]),
            (dictgen::InsensitiveStr::Ascii("cimptom"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("ryenolds"), &["reynolds"]),
            (dictgen::InsensitiveStr::Ascii("tranfsers"), &["transfers"]),
            (dictgen::InsensitiveStr::Ascii("localizin"), &["localizing"]),
            (dictgen::InsensitiveStr::Ascii("validaty"), &["validity"]),
            (dictgen::InsensitiveStr::Ascii("maangers"), &["managers"]),
            (
                dictgen::InsensitiveStr::Ascii("environmet"),
                &["environments", "environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counsellin"),
                &["counselling"],
            ),
            (dictgen::InsensitiveStr::Ascii("paltinum"), &["platinum"]),
            (
                dictgen::InsensitiveStr::Ascii("cotrasferred"),
                &["cotransferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("ipmroving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("waislines"), &["waistlines"]),
            (
                dictgen::InsensitiveStr::Ascii("cyrstallization"),
                &["crystallization"],
            ),
            (dictgen::InsensitiveStr::Ascii("bedore"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("approriate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inadvertenly"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("longevitity"),
                &["longevity"],
            ),
            (dictgen::InsensitiveStr::Ascii("condamned"), &["condemned"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciaters"),
                &["appreciates"],
            ),
            (dictgen::InsensitiveStr::Ascii("mcuh"), &["much"]),
            (dictgen::InsensitiveStr::Ascii("comflicts"), &["conflicts"]),
            (dictgen::InsensitiveStr::Ascii("recommpile"), &["recompile"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhologically"),
                &["psychologically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ovverriding"),
                &["overriding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circuvents"),
                &["circumvents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rewarching"),
                &["rewatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("machiine"), &["machine"]),
            (dictgen::InsensitiveStr::Ascii("unisntall"), &["uninstall"]),
            (
                dictgen::InsensitiveStr::Ascii("appproaching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcrips"),
                &["transcripts"],
            ),
            (dictgen::InsensitiveStr::Ascii("neabled"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("amkeup"), &["makeup"]),
            (dictgen::InsensitiveStr::Ascii("prioity"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("falsly"), &["falsely"]),
            (
                dictgen::InsensitiveStr::Ascii("combinatators"),
                &["combinators"],
            ),
            (dictgen::InsensitiveStr::Ascii("doubthing"), &["doubting"]),
            (
                dictgen::InsensitiveStr::Ascii("plansewalkers"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("secrects"), &["secrets"]),
            (dictgen::InsensitiveStr::Ascii("acatemies"), &["academies"]),
            (
                dictgen::InsensitiveStr::Ascii("metropolin"),
                &["metropolitan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acnowledge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("decies"), &["decides"]),
            (
                dictgen::InsensitiveStr::Ascii("interopeable"),
                &["interoperable"],
            ),
            (dictgen::InsensitiveStr::Ascii("bisness"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("discconeected"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("veretx"), &["vertex"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaite"),
                &["renegotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("nessasary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("preferencfes"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("ennumerate"), &["enumerate"]),
            (dictgen::InsensitiveStr::Ascii("disasble"), &["disable"]),
            (dictgen::InsensitiveStr::Ascii("adwances"), &["advances"]),
            (dictgen::InsensitiveStr::Ascii("ephemereal"), &["ephemeral"]),
            (dictgen::InsensitiveStr::Ascii("receptoras"), &["receptors"]),
            (
                dictgen::InsensitiveStr::Ascii("accerelation"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("correctyly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("backned"), &["backend"]),
            (dictgen::InsensitiveStr::Ascii("carrages"), &["carriages"]),
            (dictgen::InsensitiveStr::Ascii("pount"), &["point", "pound"]),
            (dictgen::InsensitiveStr::Ascii("agrument"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("commisssions"),
                &["commissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("austrain"), &["austrian"]),
            (
                dictgen::InsensitiveStr::Ascii("contoroller"),
                &["controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("imperavi"), &["imperative"]),
            (dictgen::InsensitiveStr::Ascii("vasslas"), &["vassals"]),
            (
                dictgen::InsensitiveStr::Ascii("modfications"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("versionned"), &["versioned"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisiton"),
                &["inquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceoters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("equailty"), &["equality"]),
            (dictgen::InsensitiveStr::Ascii("execption"), &["exception"]),
            (dictgen::InsensitiveStr::Ascii("diea"), &["idea", "die"]),
            (
                dictgen::InsensitiveStr::Ascii("approproximate"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("asteroides"), &["asteroids"]),
            (dictgen::InsensitiveStr::Ascii("discrace"), &["disgrace"]),
            (dictgen::InsensitiveStr::Ascii("backfil"), &["backfill"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplikation"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pich"),
                &["pitch", "pick", "pinch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirucmstances"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("logisitcal"),
                &["logistical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substanta"),
                &["substantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referecenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("classesss"), &["classes"]),
            (
                dictgen::InsensitiveStr::Ascii("operaters"),
                &["operators", "operates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuarions"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("skrimish"), &["skirmish"]),
            (dictgen::InsensitiveStr::Ascii("errprs"), &["errors"]),
            (dictgen::InsensitiveStr::Ascii("blutooth"), &["bluetooth"]),
            (
                dictgen::InsensitiveStr::Ascii("unrestrictred"),
                &["unrestricted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabiliy"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("tallerable"), &["tolerable"]),
            (
                dictgen::InsensitiveStr::Ascii("distortron"),
                &["distortion"],
            ),
            (dictgen::InsensitiveStr::Ascii("motivet"), &["motivate"]),
            (dictgen::InsensitiveStr::Ascii("rythyms"), &["rhythms"]),
            (
                dictgen::InsensitiveStr::Ascii("deficiancies"),
                &["deficiencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentrarte"),
                &["concentrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("overwritte"), &["overwrite"]),
            (dictgen::InsensitiveStr::Ascii("arrya"), &["array"]),
            (dictgen::InsensitiveStr::Ascii("whsipers"), &["whispers"]),
            (dictgen::InsensitiveStr::Ascii("retured"), &["returned"]),
            (dictgen::InsensitiveStr::Ascii("requsts"), &["requests"]),
            (
                dictgen::InsensitiveStr::Ascii("comissioner"),
                &["commissioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("branced"),
                &["branched", "braced"],
            ),
            (dictgen::InsensitiveStr::Ascii("eamil"), &["email"]),
            (
                dictgen::InsensitiveStr::Ascii("traversible"),
                &["traversable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("componnents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiallize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("nteger"), &["integer"]),
            (
                dictgen::InsensitiveStr::Ascii("larget"),
                &["larger", "largest", "target"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continiuty"),
                &["continuity"],
            ),
            (dictgen::InsensitiveStr::Ascii("exucutor"), &["executor"]),
            (
                dictgen::InsensitiveStr::Ascii("deambigates"),
                &["disambiguates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thime"),
                &["time", "theme", "thyme", "thine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaiation"),
                &["negotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("appon"), &["upon"]),
            (dictgen::InsensitiveStr::Ascii("followind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("geographisch"),
                &["geographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("depedence"), &["dependence"]),
            (
                dictgen::InsensitiveStr::Ascii("pontentially"),
                &["potentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("celle"), &["cell"]),
            (dictgen::InsensitiveStr::Ascii("terific"), &["terrific"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitalized"),
                &["initialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciporcated"),
                &["reciprocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("lincese"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("transpable"),
                &["transposable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cockroahes"),
                &["cockroaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("tansfer"), &["transfer"]),
            (dictgen::InsensitiveStr::Ascii("minifyies"), &["minifies"]),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedronization"),
                &["tetrahedralization"],
            ),
            (dictgen::InsensitiveStr::Ascii("appendent"), &["appended"]),
            (dictgen::InsensitiveStr::Ascii("abscound"), &["abscond"]),
            (dictgen::InsensitiveStr::Ascii("characers"), &["characters"]),
            (dictgen::InsensitiveStr::Ascii("opearands"), &["operands"]),
            (dictgen::InsensitiveStr::Ascii("correpond"), &["correspond"]),
            (
                dictgen::InsensitiveStr::Ascii("geneticaly"),
                &["genetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meassurements"),
                &["measurements"],
            ),
            (dictgen::InsensitiveStr::Ascii("santitized"), &["sanitized"]),
            (
                dictgen::InsensitiveStr::Ascii("desctructing"),
                &["destructing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimatched"), &["mismatched"]),
            (
                dictgen::InsensitiveStr::Ascii("misoginystic"),
                &["misogynistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fiters"),
                &["filters", "fighters", "fitters", "fivers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opimistically"),
                &["optimistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("asthetics"), &["aesthetics"]),
            (dictgen::InsensitiveStr::Ascii("mistakey"), &["mistakenly"]),
            (
                dictgen::InsensitiveStr::Ascii("communicationg"),
                &["communicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("edditors"), &["editors"]),
            (
                dictgen::InsensitiveStr::Ascii("swaetshirt"),
                &["sweatshirt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distirbitive"),
                &["distributive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscellaenous"),
                &["miscellaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrastrinate"),
                &["procrastinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("meganism"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("coordanates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("targering"), &["targeting"]),
            (dictgen::InsensitiveStr::Ascii("masteer"), &["master"]),
            (dictgen::InsensitiveStr::Ascii("ambiant"), &["ambient"]),
            (
                dictgen::InsensitiveStr::Ascii("specifially"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coporation"),
                &["corporation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ditributive"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("airporta"), &["airports"]),
            (
                dictgen::InsensitiveStr::Ascii("embeddeding"),
                &["embedding"],
            ),
            (dictgen::InsensitiveStr::Ascii("unles"), &["unless"]),
            (
                dictgen::InsensitiveStr::Ascii("perferrence"),
                &["preference"],
            ),
            (dictgen::InsensitiveStr::Ascii("instering"), &["inserting"]),
            (dictgen::InsensitiveStr::Ascii("realstic"), &["realistic"]),
            (dictgen::InsensitiveStr::Ascii("recylced"), &["recycled"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotates"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("folwoong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("purpses"),
                &["purposes", "purses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirumventing"),
                &["circumventing"],
            ),
            (dictgen::InsensitiveStr::Ascii("exceeeded"), &["exceeded"]),
            (dictgen::InsensitiveStr::Ascii("arbitriary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsualting"),
                &["encapsulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("squarley"), &["squarely"]),
            (dictgen::InsensitiveStr::Ascii("shotcuts"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("allocte"), &["allocate"]),
            (dictgen::InsensitiveStr::Ascii("advsied"), &["advised"]),
            (dictgen::InsensitiveStr::Ascii("exceirpts"), &["excerpts"]),
            (
                dictgen::InsensitiveStr::Ascii("coexinstence"),
                &["coexistence"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabele"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("invesitgated"),
                &["investigated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegetarions"),
                &["vegetarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitally"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("shiping"), &["shipping"]),
            (dictgen::InsensitiveStr::Ascii("unablet"), &["unable"]),
            (
                dictgen::InsensitiveStr::Ascii("initalisms"),
                &["initialisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disadvantadge"),
                &["disadvantaged", "disadvantage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviniance"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversiones"),
                &["conversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architural"),
                &["architectural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organizaion"),
                &["organization"],
            ),
            (dictgen::InsensitiveStr::Ascii("privleged"), &["privileged"]),
            (dictgen::InsensitiveStr::Ascii("struc"), &["struct"]),
            (dictgen::InsensitiveStr::Ascii("knowladge"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("insititutions"),
                &["institutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procrestinating"),
                &["procrastinating"],
            ),
            (dictgen::InsensitiveStr::Ascii("disposte"), &["dispose"]),
            (dictgen::InsensitiveStr::Ascii("debbugger"), &["debugger"]),
            (
                dictgen::InsensitiveStr::Ascii("introductiory"),
                &["introductory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficilleirs"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpubles"), &["doubles"]),
            (
                dictgen::InsensitiveStr::Ascii("initilialize"),
                &["initialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnsformed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("conntects"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("navgiates"), &["navigates"]),
            (dictgen::InsensitiveStr::Ascii("broadcasti"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("gloobal"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("authoritation"),
                &["authorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("manal"), &["manual"]),
            (
                dictgen::InsensitiveStr::Ascii("capaticitance"),
                &["capacitance"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignest"), &["ingest"]),
            (
                dictgen::InsensitiveStr::Ascii("decensitized"),
                &["desensitized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplcating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtarct"), &["subtract"]),
            (dictgen::InsensitiveStr::Ascii("sometmes"), &["sometimes"]),
            (dictgen::InsensitiveStr::Ascii("plehtora"), &["plethora"]),
            (dictgen::InsensitiveStr::Ascii("memorie"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("sanytize"), &["sanitize"]),
            (dictgen::InsensitiveStr::Ascii("dentistes"), &["dentists"]),
            (dictgen::InsensitiveStr::Ascii("shuld"), &["should"]),
            (dictgen::InsensitiveStr::Ascii("logile"), &["logfile"]),
            (dictgen::InsensitiveStr::Ascii("depecate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("alogorithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("devault"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("communsim"), &["communism"]),
            (dictgen::InsensitiveStr::Ascii("ignorning"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("socialiss"), &["socialists"]),
            (dictgen::InsensitiveStr::Ascii("reconizer"), &["recognizer"]),
            (
                dictgen::InsensitiveStr::Ascii("suscribing"),
                &["subscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possibities"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("quafeured"), &["coiffured"]),
            (
                dictgen::InsensitiveStr::Ascii("enchancing"),
                &["enchanting", "enhancing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terriroties"),
                &["territories"],
            ),
            (dictgen::InsensitiveStr::Ascii("proflies"), &["profiles"]),
            (dictgen::InsensitiveStr::Ascii("combainers"), &["combiners"]),
            (dictgen::InsensitiveStr::Ascii("serialiaze"), &["serialize"]),
            (
                dictgen::InsensitiveStr::Ascii("conpassionate"),
                &["compassionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscribtion"),
                &["subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insuffucient"),
                &["insufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlcuding"), &["including"]),
            (dictgen::InsensitiveStr::Ascii("betweend"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("stregnth"), &["strength"]),
            (
                dictgen::InsensitiveStr::Ascii("retanslate"),
                &["retranslate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australain"),
                &["australians"],
            ),
            (dictgen::InsensitiveStr::Ascii("remarkes"), &["remarks"]),
            (dictgen::InsensitiveStr::Ascii("rotuines"), &["routines"]),
            (dictgen::InsensitiveStr::Ascii("wrtiten"), &["written"]),
            (
                dictgen::InsensitiveStr::Ascii("annaverserys"),
                &["anniversaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplcations"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioroties"),
                &["priorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("savigns"), &["savings"]),
            (
                dictgen::InsensitiveStr::Ascii("tolorances"),
                &["tolerances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commanline"),
                &["commandline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obvisiously"),
                &["obviously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gnerations"),
                &["generations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entended"),
                &["intended", "extended"],
            ),
            (dictgen::InsensitiveStr::Ascii("writting"), &["writing"]),
            (dictgen::InsensitiveStr::Ascii("finsih"), &["finish"]),
            (
                dictgen::InsensitiveStr::Ascii("regeistration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofirms"), &["confirms"]),
            (dictgen::InsensitiveStr::Ascii("tecnique"), &["technique"]),
            (dictgen::InsensitiveStr::Ascii("stresss"), &["stresses"]),
            (dictgen::InsensitiveStr::Ascii("constisted"), &["consisted"]),
            (
                dictgen::InsensitiveStr::Ascii("silscreened"),
                &["silkscreened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eference"),
                &["reference", "preference", "deference"],
            ),
            (dictgen::InsensitiveStr::Ascii("semapthore"), &["semaphore"]),
            (dictgen::InsensitiveStr::Ascii("cpaturing"), &["capturing"]),
            (dictgen::InsensitiveStr::Ascii("culteral"), &["cultural"]),
            (dictgen::InsensitiveStr::Ascii("hitted"), &["hit", "fitted"]),
            (
                dictgen::InsensitiveStr::Ascii("unform"),
                &["uniform", "inform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intemittent"),
                &["intermittent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegotatiations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sttring"), &["string"]),
            (dictgen::InsensitiveStr::Ascii("tetry"), &["retry"]),
            (
                dictgen::InsensitiveStr::Ascii("impresonate"),
                &["impersonate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algebraical"),
                &["algebraic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neckneards"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mercentile"),
                &["mercantile"],
            ),
            (dictgen::InsensitiveStr::Ascii("pincher"), &["pinscher"]),
            (dictgen::InsensitiveStr::Ascii("detachs"), &["detaches"]),
            (dictgen::InsensitiveStr::Ascii("depreated"), &["deprecated"]),
            (
                dictgen::InsensitiveStr::Ascii("autoneotiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("vanugard"), &["vanguard"]),
            (
                dictgen::InsensitiveStr::Ascii("representn"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("noveau"), &["nouveau"]),
            (
                dictgen::InsensitiveStr::Ascii("acquries"),
                &["acquires", "equerries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regularlization"),
                &["regularization"],
            ),
            (dictgen::InsensitiveStr::Ascii("traiger"), &["triager"]),
            (
                dictgen::InsensitiveStr::Ascii("adminstrators"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleaves"), &["believes"]),
            (
                dictgen::InsensitiveStr::Ascii("cootdinate"),
                &["coordinate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accissible"),
                &["accessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bargainning"),
                &["bargaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("lavae"), &["larvae", "lava"]),
            (dictgen::InsensitiveStr::Ascii("corection"), &["correction"]),
            (dictgen::InsensitiveStr::Ascii("execture"), &["execute"]),
            (
                dictgen::InsensitiveStr::Ascii("rembers"),
                &["remembers", "renumbers", "members"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sempaphores"),
                &["semaphores"],
            ),
            (dictgen::InsensitiveStr::Ascii("versino"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("indipendent"),
                &["independent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proeprties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppoorting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("alpahabets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("simplictic"),
                &["simplistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("loacting"), &["locating"]),
            (dictgen::InsensitiveStr::Ascii("replacin"), &["replacing"]),
            (dictgen::InsensitiveStr::Ascii("outpout"), &["output"]),
            (
                dictgen::InsensitiveStr::Ascii("pleass"),
                &["pleases", "bless"],
            ),
            (dictgen::InsensitiveStr::Ascii("tragicaly"), &["tragically"]),
            (
                dictgen::InsensitiveStr::Ascii("suspicsion"),
                &["suspicions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gound"),
                &["bound", "ground"],
            ),
            (dictgen::InsensitiveStr::Ascii("fonctions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("auguments"),
                &["arguments", "augments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substracted"),
                &["subtracted"],
            ),
            (dictgen::InsensitiveStr::Ascii("unrakned"), &["unranked"]),
            (
                dictgen::InsensitiveStr::Ascii("bloodbonre"),
                &["bloodborne"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispensiries"),
                &["dispensaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruisin"), &["cruising"]),
            (dictgen::InsensitiveStr::Ascii("interents"), &["internets"]),
            (
                dictgen::InsensitiveStr::Ascii("imprement"),
                &["implement", "improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pertrubations"),
                &["perturbations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unplacable"),
                &["unplayable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kimerically"),
                &["chimerically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complianse"),
                &["compliance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equatin"),
                &["equating", "equation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unselecgting"),
                &["unselecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indviduals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protecte"),
                &["protective", "protected", "protect"],
            ),
            (dictgen::InsensitiveStr::Ascii("ministr"), &["minister"]),
            (
                dictgen::InsensitiveStr::Ascii("finsish"),
                &["finish", "finnish"],
            ),
            (dictgen::InsensitiveStr::Ascii("millitary"), &["military"]),
            (
                dictgen::InsensitiveStr::Ascii("directorios"),
                &["directors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanicaly"),
                &["mechanically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construcive"),
                &["constructive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gauranttees"),
                &["guarantees"],
            ),
            (dictgen::InsensitiveStr::Ascii("secontary"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("reserations"),
                &["reservations"],
            ),
            (dictgen::InsensitiveStr::Ascii("whaeton"), &["wheaton"]),
            (dictgen::InsensitiveStr::Ascii("witll"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("bombarderad"),
                &["bombarded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggeratting"),
                &["exaggerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articules"),
                &["articulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("screnshots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("icongnito"), &["incognito"]),
            (dictgen::InsensitiveStr::Ascii("puplates"), &["populates"]),
            (
                dictgen::InsensitiveStr::Ascii("manupilations"),
                &["manipulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notcing"),
                &["noticing", "nothing", "notching", "noting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramteres"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("cahracter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("configulations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commerically"),
                &["commercially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simplifers"),
                &["simplifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("butterfy"), &["butterfly"]),
            (dictgen::InsensitiveStr::Ascii("beetwen"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("morni"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("treaing"), &["treating"]),
            (dictgen::InsensitiveStr::Ascii("gaurds"), &["guards"]),
            (dictgen::InsensitiveStr::Ascii("theroist"), &["theorist"]),
            (
                dictgen::InsensitiveStr::Ascii("presumabley"),
                &["presumably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrecpected"),
                &["disrespected"],
            ),
            (dictgen::InsensitiveStr::Ascii("kwnoledge"), &["knowledge"]),
            (
                dictgen::InsensitiveStr::Ascii("processesor"),
                &["processors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proocecures"),
                &["procedures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("koncentrations"),
                &["concentrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ackowledgements"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciver"), &["receiver"]),
            (dictgen::InsensitiveStr::Ascii("thirten"), &["thirteen"]),
            (dictgen::InsensitiveStr::Ascii("forcasted"), &["forecasted"]),
            (dictgen::InsensitiveStr::Ascii("sigit"), &["digit"]),
            (
                dictgen::InsensitiveStr::Ascii("reaquired"),
                &["required", "reacquired"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anarchistm"),
                &["anarchism", "anarchist"],
            ),
            (dictgen::InsensitiveStr::Ascii("britsol"), &["bristol"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obesrvation"),
                &["observation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predescesor"),
                &["predecessors"],
            ),
            (dictgen::InsensitiveStr::Ascii("smll"), &["small", "smell"]),
            (
                dictgen::InsensitiveStr::Ascii("countermeausures"),
                &["countermeasures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derivitives"),
                &["derivatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positionnal"),
                &["positional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preficed"),
                &["prefixed", "prefaced"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtletly"), &["subtlety"]),
            (
                dictgen::InsensitiveStr::Ascii("competatively"),
                &["competitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("tailsman"), &["talisman"]),
            (dictgen::InsensitiveStr::Ascii("quikc"), &["quick"]),
            (dictgen::InsensitiveStr::Ascii("varainces"), &["variances"]),
            (dictgen::InsensitiveStr::Ascii("providin"), &["providing"]),
            (
                dictgen::InsensitiveStr::Ascii("oppurtinity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comformance"),
                &["conformance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("xenophobical"),
                &["xenophobia"],
            ),
            (dictgen::InsensitiveStr::Ascii("fouth"), &["fourth"]),
            (dictgen::InsensitiveStr::Ascii("messgae"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("prejucide"), &["prejudice"]),
            (dictgen::InsensitiveStr::Ascii("encouraed"), &["encouraged"]),
            (dictgen::InsensitiveStr::Ascii("recedin"), &["receding"]),
            (
                dictgen::InsensitiveStr::Ascii("occourrence"),
                &["occurrence"],
            ),
            (dictgen::InsensitiveStr::Ascii("datastet"), &["dataset"]),
            (dictgen::InsensitiveStr::Ascii("otimizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("bureuacrats"),
                &["bureaucrats"],
            ),
            (dictgen::InsensitiveStr::Ascii("iniecting"), &["injecting"]),
            (
                dictgen::InsensitiveStr::Ascii("timestmped"),
                &["timestamped"],
            ),
            (dictgen::InsensitiveStr::Ascii("elelement"), &["element"]),
            (dictgen::InsensitiveStr::Ascii("sirvayles"), &["surveils"]),
            (dictgen::InsensitiveStr::Ascii("negotitae"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("semgenting"),
                &["segmenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categorice"),
                &["categorize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmimation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaidators"), &["validators"]),
            (
                dictgen::InsensitiveStr::Ascii("paychiatrists"),
                &["psychiatrists"],
            ),
            (dictgen::InsensitiveStr::Ascii("nawshes"), &["nauseous"]),
            (
                dictgen::InsensitiveStr::Ascii("theoreticly"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("optiions"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("paser"), &["parser"]),
            (dictgen::InsensitiveStr::Ascii("regulariry"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("unwinindg"), &["unwinding"]),
            (dictgen::InsensitiveStr::Ascii("problam"), &["problem"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucceeds"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("apauling"), &["appalling"]),
            (dictgen::InsensitiveStr::Ascii("jestream"), &["jetstream"]),
            (
                dictgen::InsensitiveStr::Ascii("chartroose"),
                &["chartreuse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eveluators"),
                &["evaluators"],
            ),
            (dictgen::InsensitiveStr::Ascii("desigining"), &["designing"]),
            (dictgen::InsensitiveStr::Ascii("submited"), &["submitted"]),
            (
                dictgen::InsensitiveStr::Ascii("successefully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("relevante"), &["relevant"]),
            (dictgen::InsensitiveStr::Ascii("soloban"), &["sokoban"]),
            (dictgen::InsensitiveStr::Ascii("defaulkt"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("eextract"), &["extract"]),
            (dictgen::InsensitiveStr::Ascii("basicallly"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("carachteristically"),
                &["characteristically"],
            ),
            (dictgen::InsensitiveStr::Ascii("winwow"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("concatonating"),
                &["concatenating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfrontational"),
                &["confrontational"],
            ),
            (dictgen::InsensitiveStr::Ascii("clossing"), &["closing"]),
            (dictgen::InsensitiveStr::Ascii("dcoument"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("opportunites"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kubernetest"),
                &["kubernetes"],
            ),
            (dictgen::InsensitiveStr::Ascii("contiunal"), &["continual"]),
            (dictgen::InsensitiveStr::Ascii("tradgic"), &["tragic"]),
            (dictgen::InsensitiveStr::Ascii("tetster"), &["tester"]),
            (
                dictgen::InsensitiveStr::Ascii("optizimation"),
                &["optimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsterams"), &["upstreams"]),
            (dictgen::InsensitiveStr::Ascii("optimizie"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("fleat"), &["float"]),
            (dictgen::InsensitiveStr::Ascii("speficeir"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("compleets"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("accommadated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("needeing"), &["needing"]),
            (
                dictgen::InsensitiveStr::Ascii("uninstallling"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("panedmic"), &["pandemic"]),
            (dictgen::InsensitiveStr::Ascii("recurive"), &["recursive"]),
            (
                dictgen::InsensitiveStr::Ascii("vacination"),
                &["vaccination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncoverged"),
                &["unconverged"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeition"), &["repetition"]),
            (dictgen::InsensitiveStr::Ascii("sigalling"), &["signalling"]),
            (
                dictgen::InsensitiveStr::Ascii("filesystmes"),
                &["filesystems"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggegators"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporaly"),
                &["temporary", "temporarily", "temporally"],
            ),
            (dictgen::InsensitiveStr::Ascii("verbse"), &["verbose"]),
            (
                dictgen::InsensitiveStr::Ascii("proletariaat"),
                &["proletariat"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribtue"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("psaced"),
                &["paced", "spaced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrifising"),
                &["sacrificing"],
            ),
            (dictgen::InsensitiveStr::Ascii("laballed"), &["labelled"]),
            (
                dictgen::InsensitiveStr::Ascii("environmetal"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("counded"), &["counted"]),
            (dictgen::InsensitiveStr::Ascii("miserabel"), &["miserable"]),
            (
                dictgen::InsensitiveStr::Ascii("mediocraty"),
                &["mediocrity"],
            ),
            (dictgen::InsensitiveStr::Ascii("accouts"), &["accounts"]),
            (dictgen::InsensitiveStr::Ascii("fakse"), &["fake"]),
            (
                dictgen::InsensitiveStr::Ascii("decompostition"),
                &["decomposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improvemnts"),
                &["improvements"],
            ),
            (dictgen::InsensitiveStr::Ascii("youe"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("tentacreul"), &["tentacle"]),
            (
                dictgen::InsensitiveStr::Ascii("reuiring"),
                &["requiring", "rewiring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compremising"),
                &["compromising"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticaly"), &["critically"]),
            (
                dictgen::InsensitiveStr::Ascii("rlease"),
                &["release", "please", "lease"],
            ),
            (dictgen::InsensitiveStr::Ascii("awesomey"), &["awesomely"]),
            (
                dictgen::InsensitiveStr::Ascii("comprehenvise"),
                &["comprehensive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commnunication"),
                &["communication"],
            ),
            (dictgen::InsensitiveStr::Ascii("conditionn"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithsmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("vertrex"), &["vertex"]),
            (dictgen::InsensitiveStr::Ascii("enogh"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("believ"), &["believe"]),
            (
                dictgen::InsensitiveStr::Ascii("speculaties"),
                &["speculative"],
            ),
            (dictgen::InsensitiveStr::Ascii("realies"), &["realise"]),
            (
                dictgen::InsensitiveStr::Ascii("specializaitons"),
                &["specializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unserlying"),
                &["underlying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confescated"),
                &["confiscated"],
            ),
            (dictgen::InsensitiveStr::Ascii("correcion"), &["correction"]),
            (dictgen::InsensitiveStr::Ascii("instaed"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("reposoitories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transolating"),
                &["translating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntehsised"),
                &["synthesised"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignornig"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("intstrumenting"),
                &["instrumenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("targer"),
                &["target", "larger", "tagger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arresst"),
                &["arrests", "arrest"],
            ),
            (dictgen::InsensitiveStr::Ascii("yearm"), &["year"]),
            (
                dictgen::InsensitiveStr::Ascii("specifiction"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("prrcision"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("asssociates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("intentas"), &["intents"]),
            (dictgen::InsensitiveStr::Ascii("anorexica"), &["anorexia"]),
            (
                dictgen::InsensitiveStr::Ascii("overhearting"),
                &["overreacting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreactes"),
                &["deprecates"],
            ),
            (dictgen::InsensitiveStr::Ascii("claus"), &["clause"]),
            (
                dictgen::InsensitiveStr::Ascii("costomizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("espisode"), &["episode"]),
            (
                dictgen::InsensitiveStr::Ascii("commericially"),
                &["commercially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoynaces"),
                &["annoyances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefifeirs"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commissons"),
                &["commissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("impuns"), &["impugns"]),
            (dictgen::InsensitiveStr::Ascii("whethere"), &["whether"]),
            (
                dictgen::InsensitiveStr::Ascii("havin"),
                &["having", "haven"],
            ),
            (dictgen::InsensitiveStr::Ascii("strucured"), &["structured"]),
            (
                dictgen::InsensitiveStr::Ascii("alternavtive"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimetriclly"),
                &["symmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supposting"),
                &["supporting", "supposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upsupported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("roiters"), &["rioters"]),
            (dictgen::InsensitiveStr::Ascii("commmunity"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("scupting"), &["sculpting"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistancies"),
                &["inconsistencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("larwence"), &["lawrence"]),
            (dictgen::InsensitiveStr::Ascii("rythymic"), &["rhythmic"]),
            (dictgen::InsensitiveStr::Ascii("recivable"), &["receivable"]),
            (dictgen::InsensitiveStr::Ascii("codesbase"), &["codebase"]),
            (
                dictgen::InsensitiveStr::Ascii("competitevly"),
                &["competitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("politelly"), &["politely"]),
            (
                dictgen::InsensitiveStr::Ascii("desscriptive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoveniently"),
                &["inconveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cominatorial"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("appparent"), &["apparent"]),
            (
                dictgen::InsensitiveStr::Ascii("comnpleteness"),
                &["completeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accontants"),
                &["accountants"],
            ),
            (dictgen::InsensitiveStr::Ascii("revsion"), &["revision"]),
            (dictgen::InsensitiveStr::Ascii("uper"), &["super", "upper"]),
            (
                dictgen::InsensitiveStr::Ascii("replicae"),
                &["replicated", "replicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumsice"),
                &["circumcised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiziation"),
                &["optimization"],
            ),
            (dictgen::InsensitiveStr::Ascii("thristy"), &["thirsty"]),
            (
                dictgen::InsensitiveStr::Ascii("objectiviser"),
                &["objectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differneces"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("thursters"), &["thrusters"]),
            (
                dictgen::InsensitiveStr::Ascii("capitlaize"),
                &["capitalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neigbhorhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("lunatis"), &["lunatics"]),
            (
                dictgen::InsensitiveStr::Ascii("homogenuous"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("noticable"), &["noticeable"]),
            (
                dictgen::InsensitiveStr::Ascii("bootstrappin"),
                &["bootstrapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seding"),
                &["sending", "seeding", "ceding", "sleding"],
            ),
            (dictgen::InsensitiveStr::Ascii("easyly"), &["easily"]),
            (
                dictgen::InsensitiveStr::Ascii("transaition"),
                &["translation", "transition", "transaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("codeing"), &["coding"]),
            (dictgen::InsensitiveStr::Ascii("mesenger"), &["messenger"]),
            (
                dictgen::InsensitiveStr::Ascii("viusalising"),
                &["visualising"],
            ),
            (dictgen::InsensitiveStr::Ascii("secont"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("schyzophrenia"),
                &["schizophrenia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementd"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionnals"),
                &["conditionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explioting"),
                &["exploiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categroies"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("tangeld"), &["tangled"]),
            (
                dictgen::InsensitiveStr::Ascii("spefixiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconencts"),
                &["reconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("tagg"), &["tag"]),
            (dictgen::InsensitiveStr::Ascii("dirvers"), &["drivers"]),
            (
                dictgen::InsensitiveStr::Ascii("brazilions"),
                &["brazilians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sycronise"),
                &["synchronise"],
            ),
            (dictgen::InsensitiveStr::Ascii("tung"), &["tongue"]),
            (
                dictgen::InsensitiveStr::Ascii("unsupposed"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("relpaced"), &["replaced"]),
            (
                dictgen::InsensitiveStr::Ascii("valnetines"),
                &["valentines"],
            ),
            (dictgen::InsensitiveStr::Ascii("noctunre"), &["nocturne"]),
            (dictgen::InsensitiveStr::Ascii("frogotten"), &["forgotten"]),
            (dictgen::InsensitiveStr::Ascii("resovler"), &["resolver"]),
            (dictgen::InsensitiveStr::Ascii("indciated"), &["indicated"]),
            (dictgen::InsensitiveStr::Ascii("covenental"), &["covenant"]),
            (dictgen::InsensitiveStr::Ascii("campagin"), &["campaign"]),
            (
                dictgen::InsensitiveStr::Ascii("messged"),
                &["messaged", "messed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authoitatively"),
                &["authoritatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("accouned"), &["accounted"]),
            (dictgen::InsensitiveStr::Ascii("covenient"), &["convenient"]),
            (dictgen::InsensitiveStr::Ascii("commad"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("farmasudical"),
                &["pharmaceutical"],
            ),
            (dictgen::InsensitiveStr::Ascii("bridman"), &["birdman"]),
            (
                dictgen::InsensitiveStr::Ascii("trustworhty"),
                &["trustworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwys"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("tesselations"),
                &["tessellations"],
            ),
            (dictgen::InsensitiveStr::Ascii("unmbering"), &["numbering"]),
            (
                dictgen::InsensitiveStr::Ascii("camparability"),
                &["comparability"],
            ),
            (dictgen::InsensitiveStr::Ascii("secounds"), &["seconds"]),
            (
                dictgen::InsensitiveStr::Ascii("submssions"),
                &["submissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hyperlogic"),
                &["hyperbolic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sumplifies"),
                &["simplifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("bookmarkd"), &["bookmarked"]),
            (
                dictgen::InsensitiveStr::Ascii("exisitng"),
                &["existing", "exiting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bodybuildng"),
                &["bodybuilding"],
            ),
            (dictgen::InsensitiveStr::Ascii("playabe"), &["playable"]),
            (
                dictgen::InsensitiveStr::Ascii("definitoins"),
                &["definitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiter"), &["pointer"]),
            (dictgen::InsensitiveStr::Ascii("propmting"), &["prompting"]),
            (
                dictgen::InsensitiveStr::Ascii("contientous"),
                &["contentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opputunity"),
                &["opportunity"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitrty"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("sherifs"), &["sheriffs"]),
            (dictgen::InsensitiveStr::Ascii("suppory"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("prodives"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("dictrionary"),
                &["dictionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stadnardising"),
                &["standardising"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinic"), &["cynic", "sonic"]),
            (dictgen::InsensitiveStr::Ascii("suffenct"), &["sufficient"]),
            (
                dictgen::InsensitiveStr::Ascii("obejctively"),
                &["objectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifgurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiztaions"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriont"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("moint"), &["mount"]),
            (dictgen::InsensitiveStr::Ascii("destrys"), &["destroys"]),
            (
                dictgen::InsensitiveStr::Ascii("indicatiors"),
                &["indicators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statutses"),
                &["statuses", "statutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordiniates"),
                &["coordinates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchroneously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illuminite"),
                &["illuminati"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preveiwers"),
                &["previewers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsufficient"),
                &["insufficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exucutable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("inctances"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("uniformely"), &["uniformly"]),
            (dictgen::InsensitiveStr::Ascii("dafaults"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitors"),
                &["arbiters", "arbitrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("driagrammed"),
                &["diagrammed"],
            ),
            (dictgen::InsensitiveStr::Ascii("parterned"), &["partnered"]),
            (dictgen::InsensitiveStr::Ascii("escased"), &["escaped"]),
            (
                dictgen::InsensitiveStr::Ascii("hanager"),
                &["hanger", "manager"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulit"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("accpets"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("wholeheartadly"),
                &["wholeheartedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("benifits"), &["benefits"]),
            (dictgen::InsensitiveStr::Ascii("materaial"), &["material"]),
            (dictgen::InsensitiveStr::Ascii("mutatable"), &["mutable"]),
            (dictgen::InsensitiveStr::Ascii("dymamic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("eaiser"), &["easier"]),
            (
                dictgen::InsensitiveStr::Ascii("elctrically"),
                &["electrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("aautomatic"), &["automatic"]),
            (dictgen::InsensitiveStr::Ascii("franscico"), &["francisco"]),
            (dictgen::InsensitiveStr::Ascii("seive"), &["sieve"]),
            (dictgen::InsensitiveStr::Ascii("noteably"), &["notably"]),
            (dictgen::InsensitiveStr::Ascii("misison"), &["mission"]),
            (dictgen::InsensitiveStr::Ascii("coalcece"), &["coalescence"]),
            (
                dictgen::InsensitiveStr::Ascii("strengtens"),
                &["strengthens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reqeuesting"),
                &["requesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulberabuility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncessfully"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("loacalization"),
                &["localization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mikrosecond"),
                &["microsecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ddition"),
                &["addition", "edition"],
            ),
            (dictgen::InsensitiveStr::Ascii("trapeziod"), &["trapezoid"]),
            (
                dictgen::InsensitiveStr::Ascii("consitutute"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("pactivity"), &["captivity"]),
            (dictgen::InsensitiveStr::Ascii("horader"), &["hoarder"]),
            (
                dictgen::InsensitiveStr::Ascii("excpetional"),
                &["exceptional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifcations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinfiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modificatons"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("hapenn"), &["happen"]),
            (
                dictgen::InsensitiveStr::Ascii("bastractness"),
                &["abstractness"],
            ),
            (dictgen::InsensitiveStr::Ascii("associat"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("generase"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("manuly"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("oprimization"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("radiocative"),
                &["radioactive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifiable"),
                &["specifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggregious"), &["egregious"]),
            (dictgen::InsensitiveStr::Ascii("lienar"), &["linear"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialisations"),
                &["initialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("balaces"), &["balances"]),
            (dictgen::InsensitiveStr::Ascii("channnel"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("localation"), &["location"]),
            (
                dictgen::InsensitiveStr::Ascii("serialsiation"),
                &["serialisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighboing"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreagations"),
                &["aggregations"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeates"), &["creates"]),
            (
                dictgen::InsensitiveStr::Ascii("repets"),
                &["repeats", "repents", "repels", "resets"],
            ),
            (dictgen::InsensitiveStr::Ascii("directoty"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("differnec"), &["difference"]),
            (
                dictgen::InsensitiveStr::Ascii("approacing"),
                &["approaching"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanns"), &["cleans"]),
            (
                dictgen::InsensitiveStr::Ascii("descripively"),
                &["descriptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("minisucle"), &["miniscule"]),
            (dictgen::InsensitiveStr::Ascii("enouch"), &["enough"]),
            (dictgen::InsensitiveStr::Ascii("callacks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("symetrical"),
                &["symmetrical"],
            ),
            (dictgen::InsensitiveStr::Ascii("aadded"), &["added"]),
            (dictgen::InsensitiveStr::Ascii("coolent"), &["coolant"]),
            (dictgen::InsensitiveStr::Ascii("laguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("chnage"), &["change"]),
            (dictgen::InsensitiveStr::Ascii("cosntrain"), &["constrain"]),
            (dictgen::InsensitiveStr::Ascii("lieutenat"), &["lieutenant"]),
            (dictgen::InsensitiveStr::Ascii("psace"), &["space", "pace"]),
            (dictgen::InsensitiveStr::Ascii("firsth"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("follwwing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("repondent"), &["respondent"]),
            (dictgen::InsensitiveStr::Ascii("fales"), &["false", "fails"]),
            (dictgen::InsensitiveStr::Ascii("millenium"), &["millennium"]),
            (dictgen::InsensitiveStr::Ascii("quntities"), &["quantities"]),
            (
                dictgen::InsensitiveStr::Ascii("dispensaire"),
                &["dispensaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constrctor"),
                &["constructor", "constrictor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("datattype"), &["datatype"]),
            (
                dictgen::InsensitiveStr::Ascii("alogirhtms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("specificed"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("cooncept"), &["concept"]),
            (dictgen::InsensitiveStr::Ascii("critiquin"), &["critiquing"]),
            (dictgen::InsensitiveStr::Ascii("possesing"), &["possessing"]),
            (dictgen::InsensitiveStr::Ascii("visaul"), &["visual"]),
            (dictgen::InsensitiveStr::Ascii("ansers"), &["answers"]),
            (
                dictgen::InsensitiveStr::Ascii("algorhythmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wite"), &["write", "white"]),
            (
                dictgen::InsensitiveStr::Ascii("aequivalent"),
                &["equivalent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deactiveate"),
                &["deactivate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("variantes"),
                &["variants", "variances", "variates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("costitution"),
                &["constitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessleator"),
                &["tessellator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resiliance"),
                &["resilience"],
            ),
            (dictgen::InsensitiveStr::Ascii("circulary"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("omnisience"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propogator"),
                &["propagator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrcutor"),
                &["instructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("proceseses"), &["processes"]),
            (dictgen::InsensitiveStr::Ascii("programe"), &["programme"]),
            (dictgen::InsensitiveStr::Ascii("depnding"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("anecdatally"),
                &["anecdotally"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsense"), &["offense"]),
            (dictgen::InsensitiveStr::Ascii("smartare"), &["smarter"]),
            (dictgen::InsensitiveStr::Ascii("premeir"), &["premier"]),
            (dictgen::InsensitiveStr::Ascii("extereme"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("dsiplay"), &["display"]),
            (dictgen::InsensitiveStr::Ascii("cmoplex"), &["complex"]),
            (
                dictgen::InsensitiveStr::Ascii("seplicurally"),
                &["sepulchrally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confugrable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("starightened"),
                &["straightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("replaints"), &["repaints"]),
            (dictgen::InsensitiveStr::Ascii("outout"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("geenrates"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("symlinc"), &["symlink"]),
            (dictgen::InsensitiveStr::Ascii("aliginment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("subjectivly"),
                &["subjectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("poygon"), &["polygon"]),
            (dictgen::InsensitiveStr::Ascii("shrapenl"), &["shrapnel"]),
            (
                dictgen::InsensitiveStr::Ascii("partisions"),
                &["partitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enivronmentally"),
                &["environmentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("moleculs"), &["molecules"]),
            (dictgen::InsensitiveStr::Ascii("callendar"), &["calendar"]),
            (dictgen::InsensitiveStr::Ascii("warehose"), &["warehouse"]),
            (dictgen::InsensitiveStr::Ascii("inegrate"), &["integrate"]),
            (
                dictgen::InsensitiveStr::Ascii("sanctionning"),
                &["sanctioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparanet"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straighforward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("browswable"), &["browsable"]),
            (dictgen::InsensitiveStr::Ascii("orhcestra"), &["orchestra"]),
            (dictgen::InsensitiveStr::Ascii("awnser"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("upgradde"), &["upgrade"]),
            (dictgen::InsensitiveStr::Ascii("upgaraded"), &["upgraded"]),
            (
                dictgen::InsensitiveStr::Ascii("procees"),
                &["proceeds", "process"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breackpoints"),
                &["breakpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("poiting"), &["pointing"]),
            (dictgen::InsensitiveStr::Ascii("hve"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("uploders"), &["uploaders"]),
            (
                dictgen::InsensitiveStr::Ascii("patition"),
                &["partition", "petition"],
            ),
            (dictgen::InsensitiveStr::Ascii("blockhain"), &["blockchain"]),
            (
                dictgen::InsensitiveStr::Ascii("efficienctly"),
                &["efficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appontment"),
                &["appointment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beging"),
                &["begging", "beginning", "being"],
            ),
            (dictgen::InsensitiveStr::Ascii("pharmacs"), &["pharmacist"]),
            (dictgen::InsensitiveStr::Ascii("confirmeed"), &["confirmed"]),
            (
                dictgen::InsensitiveStr::Ascii("industiralized"),
                &["industrialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("verteces"), &["vertices"]),
            (
                dictgen::InsensitiveStr::Ascii("dereivative"),
                &["derivative"],
            ),
            (dictgen::InsensitiveStr::Ascii("condtition"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("beefore"), &["before"]),
            (
                dictgen::InsensitiveStr::Ascii("brances"),
                &["branches", "braces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conntainers"),
                &["containers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepererate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("regons"), &["regions"]),
            (
                dictgen::InsensitiveStr::Ascii("subsctitution"),
                &["substitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("billbaord"), &["billboard"]),
            (dictgen::InsensitiveStr::Ascii("inidcated"), &["indicated"]),
            (dictgen::InsensitiveStr::Ascii("broswers"), &["browsers"]),
            (dictgen::InsensitiveStr::Ascii("spacebr"), &["spacebar"]),
            (
                dictgen::InsensitiveStr::Ascii("accelorating"),
                &["accelerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimpremented"),
                &["reimplemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynamincal"), &["dynamical"]),
            (
                dictgen::InsensitiveStr::Ascii("connecitng"),
                &["connecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("whish"), &["wish", "whisk"]),
            (dictgen::InsensitiveStr::Ascii("fomats"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialzes"),
                &["deserializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ubiquituous"),
                &["ubiquitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countinously"),
                &["continuously"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytheses"), &["syntheses"]),
            (dictgen::InsensitiveStr::Ascii("isolatie"), &["isolate"]),
            (
                dictgen::InsensitiveStr::Ascii("pithon"),
                &["python", "piton"],
            ),
            (dictgen::InsensitiveStr::Ascii("althete"), &["athlete"]),
            (
                dictgen::InsensitiveStr::Ascii("circuventing"),
                &["circumventing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bahavioral"),
                &["behavioral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goosebumbs"),
                &["goosebumps"],
            ),
            (dictgen::InsensitiveStr::Ascii("absance"), &["absence"]),
            (
                dictgen::InsensitiveStr::Ascii("incorreclty"),
                &["incorrectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dahsboards"),
                &["dashboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("instane"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("deattachment"),
                &["detachment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integreation"),
                &["integration"],
            ),
            (dictgen::InsensitiveStr::Ascii("bycicle"), &["bicycle"]),
            (dictgen::InsensitiveStr::Ascii("horozons"), &["horizons"]),
            (
                dictgen::InsensitiveStr::Ascii("boycottting"),
                &["boycotting"],
            ),
            (dictgen::InsensitiveStr::Ascii("seperadly"), &["separately"]),
            (
                dictgen::InsensitiveStr::Ascii("disambigouate"),
                &["disambiguate"],
            ),
            (dictgen::InsensitiveStr::Ascii("execude"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("scipting"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("chesckers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("blackmsith"),
                &["blacksmith"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimiss"), &["dismiss"]),
            (dictgen::InsensitiveStr::Ascii("pedning"), &["pending"]),
            (dictgen::InsensitiveStr::Ascii("abilityes"), &["abilities"]),
            (dictgen::InsensitiveStr::Ascii("performmed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("traingle"), &["triangle"]),
            (dictgen::InsensitiveStr::Ascii("exlporers"), &["explorers"]),
            (dictgen::InsensitiveStr::Ascii("carmonies"), &["ceremonies"]),
            (dictgen::InsensitiveStr::Ascii("implacted"), &["impacted"]),
            (dictgen::InsensitiveStr::Ascii("debuf"), &["debug"]),
            (dictgen::InsensitiveStr::Ascii("benefied"), &["benefited"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallied"),
                &["specified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bakcers"), &["backers"]),
            (dictgen::InsensitiveStr::Ascii("isralies"), &["israelis"]),
            (dictgen::InsensitiveStr::Ascii("moteuring"), &["motoring"]),
            (
                dictgen::InsensitiveStr::Ascii("subscriptons"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avilability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlighing"),
                &["highlighting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninterupted"),
                &["uninterrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legendaris"),
                &["legendaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("coudlnt"), &["couldnt"]),
            (
                dictgen::InsensitiveStr::Ascii("appropriat"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("nectode"), &["netcode"]),
            (
                dictgen::InsensitiveStr::Ascii("perfoemative"),
                &["performative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enegineering"),
                &["engineering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theretically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teansylvania"),
                &["transylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("regaurding"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("pshycopaths"),
                &["psychopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("unfailry"), &["unfairly"]),
            (dictgen::InsensitiveStr::Ascii("suttled"), &["shuttled"]),
            (
                dictgen::InsensitiveStr::Ascii("monts"),
                &["months", "mounts", "mints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uestionnaires"),
                &["questionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deducitble"),
                &["deductible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microcentro"),
                &["microcenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premultipled"),
                &["premultiplied"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomoetric"), &["geometric"]),
            (
                dictgen::InsensitiveStr::Ascii("disssociating"),
                &["dissociating"],
            ),
            (dictgen::InsensitiveStr::Ascii("centriods"), &["centroids"]),
            (
                dictgen::InsensitiveStr::Ascii("unrealible"),
                &["unreliable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifification"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("timesamp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("comapres"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("rudimentery"),
                &["rudimentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suseptible"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicslosed"), &["disclosed"]),
            (dictgen::InsensitiveStr::Ascii("theature"), &["theatre"]),
            (
                dictgen::InsensitiveStr::Ascii("abstracing"),
                &["abstracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("conpleted"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("potentielly"),
                &["potentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomralizing"),
                &["normalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completley"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("winodw"), &["window"]),
            (dictgen::InsensitiveStr::Ascii("arrranging"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("switzlerand"),
                &["switzerland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consides"),
                &["coincides", "considers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reopository"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restuarants"),
                &["restaurants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfriednly"),
                &["unfriendly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarneteeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutiline"), &["multiline"]),
            (
                dictgen::InsensitiveStr::Ascii("psacing"),
                &["pacing", "spacing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circustance"),
                &["circumstance"],
            ),
            (dictgen::InsensitiveStr::Ascii("aleviate"), &["alleviate"]),
            (
                dictgen::InsensitiveStr::Ascii("mislabaled"),
                &["mislabeled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fabricacion"),
                &["fabrication"],
            ),
            (dictgen::InsensitiveStr::Ascii("threahold"), &["threshold"]),
            (
                dictgen::InsensitiveStr::Ascii("impprovement"),
                &["improvement"],
            ),
            (dictgen::InsensitiveStr::Ascii("sersies"), &["series"]),
            (
                dictgen::InsensitiveStr::Ascii("contriversy"),
                &["controversy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefiififes"),
                &["specifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhymme"), &["rhyme"]),
            (dictgen::InsensitiveStr::Ascii("upccoming"), &["upcoming"]),
            (dictgen::InsensitiveStr::Ascii("firmwre"), &["firmware"]),
            (dictgen::InsensitiveStr::Ascii("follers"), &["followers"]),
            (dictgen::InsensitiveStr::Ascii("checckbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("supersed"), &["superseded"]),
            (dictgen::InsensitiveStr::Ascii("saandbox"), &["sandbox"]),
            (
                dictgen::InsensitiveStr::Ascii("accelertion"),
                &["acceleration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchronious"),
                &["synchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perputuated"),
                &["perpetuated"],
            ),
            (dictgen::InsensitiveStr::Ascii("yor"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("backrgounds"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highligjting"),
                &["highlighting"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetran"), &["veteran"]),
            (dictgen::InsensitiveStr::Ascii("sccesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("peronal"), &["personal"]),
            (dictgen::InsensitiveStr::Ascii("persaude"), &["persuade"]),
            (
                dictgen::InsensitiveStr::Ascii("configurtoins"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("pruposely"), &["purposely"]),
            (dictgen::InsensitiveStr::Ascii("cristals"), &["crystals"]),
            (dictgen::InsensitiveStr::Ascii("offic"), &["office"]),
            (dictgen::InsensitiveStr::Ascii("invitied"), &["invited"]),
            (dictgen::InsensitiveStr::Ascii("dowloads"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("romoving"), &["removing"]),
            (dictgen::InsensitiveStr::Ascii("alomost"), &["almost"]),
            (
                dictgen::InsensitiveStr::Ascii("incremeting"),
                &["incrementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("expessed"), &["expressed"]),
            (dictgen::InsensitiveStr::Ascii("anumber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("secial"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("upddater"), &["updater"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstalation"),
                &["reinstallation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("throuhgout"),
                &["throughout"],
            ),
            (dictgen::InsensitiveStr::Ascii("traveses"), &["traverses"]),
            (dictgen::InsensitiveStr::Ascii("aribitary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("suvives"), &["survives"]),
            (dictgen::InsensitiveStr::Ascii("finshed"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("zimbabwaen"), &["zimbabwe"]),
            (
                dictgen::InsensitiveStr::Ascii("psychistrist"),
                &["psychiatrist"],
            ),
            (dictgen::InsensitiveStr::Ascii("agravated"), &["aggravated"]),
            (dictgen::InsensitiveStr::Ascii("specefies"), &["specifies"]),
            (
                dictgen::InsensitiveStr::Ascii("deffers"),
                &["differs", "defers"],
            ),
            (dictgen::InsensitiveStr::Ascii("fiorget"), &["forget"]),
            (dictgen::InsensitiveStr::Ascii("satisfiy"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("slignt"), &["slight"]),
            (
                dictgen::InsensitiveStr::Ascii("brainwasing"),
                &["brainwashing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripive"),
                &["descriptive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seranomial"),
                &["ceremonial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trottled"),
                &["throttled", "trotted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highlighers"),
                &["highlighters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repreesnts"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("sleecting"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("negilgence"),
                &["negligence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soveriengty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("imge"), &["image"]),
            (dictgen::InsensitiveStr::Ascii("copyirght"), &["copyright"]),
            (dictgen::InsensitiveStr::Ascii("anaylze"), &["analyze"]),
            (dictgen::InsensitiveStr::Ascii("otherwsise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("acedemics"), &["academics"]),
            (
                dictgen::InsensitiveStr::Ascii("extens"),
                &["extends", "externs"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingeger"), &["integer"]),
            (dictgen::InsensitiveStr::Ascii("sytlish"), &["stylish"]),
            (dictgen::InsensitiveStr::Ascii("jerusalm"), &["jerusalem"]),
            (
                dictgen::InsensitiveStr::Ascii("continutous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatentae"),
                &["concatenate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interessted"),
                &["interested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpactedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginazations"),
                &["organizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("guarnantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("inappropriatly"),
                &["inappropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("genomew"), &["genome"]),
            (
                dictgen::InsensitiveStr::Ascii("attachement"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("exposte"), &["expose"]),
            (dictgen::InsensitiveStr::Ascii("intquires"), &["inquires"]),
            (
                dictgen::InsensitiveStr::Ascii("generatore"),
                &["generate", "generator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peretrator"),
                &["perpetrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progroms"),
                &["pogroms", "programs"],
            ),
            (dictgen::InsensitiveStr::Ascii("accsess"), &["access"]),
            (
                dictgen::InsensitiveStr::Ascii("inquisator"),
                &["inquisitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("woekspace"), &["workspace"]),
            (dictgen::InsensitiveStr::Ascii("relized"), &["realized"]),
            (dictgen::InsensitiveStr::Ascii("amtheysts"), &["amethysts"]),
            (
                dictgen::InsensitiveStr::Ascii("setquential"),
                &["sequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mainainers"),
                &["maintainers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifiler"),
                &["identifier"],
            ),
            (dictgen::InsensitiveStr::Ascii("ucrrently"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("thottles"), &["throttles"]),
            (
                dictgen::InsensitiveStr::Ascii("movemements"),
                &["movements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diffences"),
                &["differences", "defences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsituted"),
                &["substituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("presentas"), &["presents"]),
            (
                dictgen::InsensitiveStr::Ascii("comminutes"),
                &["communities"],
            ),
            (dictgen::InsensitiveStr::Ascii("weant"), &["want", "wean"]),
            (
                dictgen::InsensitiveStr::Ascii("mathemetician"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("altitute"), &["altitude"]),
            (dictgen::InsensitiveStr::Ascii("ouputting"), &["outputting"]),
            (
                dictgen::InsensitiveStr::Ascii("lsits"),
                &["lists", "slits", "sits"],
            ),
            (dictgen::InsensitiveStr::Ascii("recruting"), &["recruiting"]),
            (
                dictgen::InsensitiveStr::Ascii("revaildating"),
                &["revalidating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affrimative"),
                &["affirmative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralleism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomanded"), &["commanded"]),
            (dictgen::InsensitiveStr::Ascii("acadmies"), &["academies"]),
            (dictgen::InsensitiveStr::Ascii("apendage"), &["appendage"]),
            (dictgen::InsensitiveStr::Ascii("colomn"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("eclpse"), &["eclipse"]),
            (dictgen::InsensitiveStr::Ascii("bringin"), &["bringing"]),
            (
                dictgen::InsensitiveStr::Ascii("ternimations"),
                &["terminations"],
            ),
            (dictgen::InsensitiveStr::Ascii("skippped"), &["skipped"]),
            (dictgen::InsensitiveStr::Ascii("fighitng"), &["fighting"]),
            (dictgen::InsensitiveStr::Ascii("execunting"), &["executing"]),
            (
                dictgen::InsensitiveStr::Ascii("authenfication"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prototying"),
                &["prototyping"],
            ),
            (dictgen::InsensitiveStr::Ascii("questins"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("dialouges"), &["dialogues"]),
            (dictgen::InsensitiveStr::Ascii("explioted"), &["exploited"]),
            (dictgen::InsensitiveStr::Ascii("perfromer"), &["performer"]),
            (
                dictgen::InsensitiveStr::Ascii("costructer"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photograped"),
                &["photographed"],
            ),
            (dictgen::InsensitiveStr::Ascii("packaeging"), &["packaging"]),
            (dictgen::InsensitiveStr::Ascii("ontainer"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("sciense"), &["sciences"]),
            (dictgen::InsensitiveStr::Ascii("althetic"), &["athletic"]),
            (
                dictgen::InsensitiveStr::Ascii("outraegously"),
                &["outrageously"],
            ),
            (dictgen::InsensitiveStr::Ascii("commplete"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("commere"), &["commerce"]),
            (dictgen::InsensitiveStr::Ascii("fristly"), &["firstly"]),
            (
                dictgen::InsensitiveStr::Ascii("indecise"),
                &["indices", "indecisive"],
            ),
            (dictgen::InsensitiveStr::Ascii("torrenters"), &["torrents"]),
            (dictgen::InsensitiveStr::Ascii("nominet"), &["nominate"]),
            (
                dictgen::InsensitiveStr::Ascii("incrememnts"),
                &["increments"],
            ),
            (dictgen::InsensitiveStr::Ascii("psudonym"), &["pseudonym"]),
            (
                dictgen::InsensitiveStr::Ascii("termiantion"),
                &["termination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correespondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifty"), &["verify"]),
            (dictgen::InsensitiveStr::Ascii("maesure"), &["measure"]),
            (dictgen::InsensitiveStr::Ascii("signes"), &["signs"]),
            (dictgen::InsensitiveStr::Ascii("premits"), &["permits"]),
            (dictgen::InsensitiveStr::Ascii("messaes"), &["messages"]),
            (
                dictgen::InsensitiveStr::Ascii("udating"),
                &["updating", "dating"],
            ),
            (dictgen::InsensitiveStr::Ascii("resovle"), &["resolve"]),
            (
                dictgen::InsensitiveStr::Ascii("manetaining"),
                &["maintaining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unproccessed"),
                &["unprocessed"],
            ),
            (dictgen::InsensitiveStr::Ascii("quadrange"), &["quadrangle"]),
            (dictgen::InsensitiveStr::Ascii("apsotles"), &["apostles"]),
            (dictgen::InsensitiveStr::Ascii("effeects"), &["effects"]),
            (dictgen::InsensitiveStr::Ascii("upsreamed"), &["upstreamed"]),
            (
                dictgen::InsensitiveStr::Ascii("emapthetic"),
                &["empathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compresson"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("sotry"), &["story", "sorry"]),
            (
                dictgen::InsensitiveStr::Ascii("butterfries"),
                &["butterflies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pernamently"),
                &["permanently"],
            ),
            (dictgen::InsensitiveStr::Ascii("tryied"), &["tried"]),
            (dictgen::InsensitiveStr::Ascii("socilaist"), &["socialist"]),
            (dictgen::InsensitiveStr::Ascii("clinicos"), &["clinics"]),
            (dictgen::InsensitiveStr::Ascii("servece"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitues"),
                &["prostitutes"],
            ),
            (dictgen::InsensitiveStr::Ascii("attetion"), &["attention"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsiderant"),
                &["inconsiderate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enveloppes"),
                &["envelope", "envelopes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cusstomization"),
                &["customization"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppened"), &["opened"]),
            (
                dictgen::InsensitiveStr::Ascii("duplciations"),
                &["duplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("managenment"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("multipied"), &["multiplied"]),
            (dictgen::InsensitiveStr::Ascii("woulf"), &["would"]),
            (dictgen::InsensitiveStr::Ascii("unerstand"), &["understand"]),
            (
                dictgen::InsensitiveStr::Ascii("insantiated"),
                &["instantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("endur"), &["endure"]),
            (
                dictgen::InsensitiveStr::Ascii("providance"),
                &["providence"],
            ),
            (dictgen::InsensitiveStr::Ascii("barbarisch"), &["barbaric"]),
            (dictgen::InsensitiveStr::Ascii("agrain"), &["again"]),
            (dictgen::InsensitiveStr::Ascii("recordare"), &["recorder"]),
            (
                dictgen::InsensitiveStr::Ascii("ameeliarators"),
                &["ameliorators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attrbution"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxiamting"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("modiify"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("othwerise"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("unce"), &["once"]),
            (dictgen::InsensitiveStr::Ascii("acconts"), &["accounts"]),
            (
                dictgen::InsensitiveStr::Ascii("neighboorhods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wokrstations"),
                &["workstations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensititivities"),
                &["sensitivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commandmants"),
                &["commandments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbcribed"),
                &["subscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instanciations"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelerateor"),
                &["accelerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("disucss"), &["discuss"]),
            (dictgen::InsensitiveStr::Ascii("glithced"), &["glitched"]),
            (
                dictgen::InsensitiveStr::Ascii("distributivos"),
                &["distributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catterpilars"),
                &["caterpillars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seuxalized"),
                &["sexualized"],
            ),
            (dictgen::InsensitiveStr::Ascii("devfine"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("directos"), &["directors"]),
            (dictgen::InsensitiveStr::Ascii("enterie"), &["entry"]),
            (
                dictgen::InsensitiveStr::Ascii("simulataneity"),
                &["simultaneity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insensitivy"),
                &["insensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configurated"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("conacts"), &["contacts"]),
            (dictgen::InsensitiveStr::Ascii("liecester"), &["leicester"]),
            (
                dictgen::InsensitiveStr::Ascii("adeqautely"),
                &["adequately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relfecting"),
                &["reflecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represntative"),
                &["representative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalidtion"),
                &["invalidation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconfugre"),
                &["reconfigure"],
            ),
            (dictgen::InsensitiveStr::Ascii("origially"), &["originally"]),
            (dictgen::InsensitiveStr::Ascii("prohibts"), &["prohibits"]),
            (dictgen::InsensitiveStr::Ascii("safetly"), &["safely"]),
            (dictgen::InsensitiveStr::Ascii("allocagtor"), &["allocator"]),
            (
                dictgen::InsensitiveStr::Ascii("professorn"),
                &["profession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cappacitance"),
                &["capacitance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infomation"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seriousally"),
                &["seriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("novemeber"), &["november"]),
            (dictgen::InsensitiveStr::Ascii("extremised"), &["extremes"]),
            (
                dictgen::InsensitiveStr::Ascii("comventions"),
                &["conventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("numbrer"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("defensea"), &["defenseman"]),
            (dictgen::InsensitiveStr::Ascii("cosoling"), &["consoling"]),
            (dictgen::InsensitiveStr::Ascii("actvation"), &["activation"]),
            (
                dictgen::InsensitiveStr::Ascii("trasformer"),
                &["transformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insitution"),
                &["institution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crystalises"),
                &["crystallises"],
            ),
            (dictgen::InsensitiveStr::Ascii("susepect"), &["suspect"]),
            (dictgen::InsensitiveStr::Ascii("sexuallity"), &["sexually"]),
            (dictgen::InsensitiveStr::Ascii("separtated"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("advocaded"), &["advocated"]),
            (
                dictgen::InsensitiveStr::Ascii("quwessants"),
                &["croissants"],
            ),
            (dictgen::InsensitiveStr::Ascii("sescrets"), &["secrets"]),
            (dictgen::InsensitiveStr::Ascii("machiens"), &["machines"]),
            (dictgen::InsensitiveStr::Ascii("sotyr"), &["satyr", "story"]),
            (dictgen::InsensitiveStr::Ascii("documemts"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("descrptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurences"),
                &["recurrences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotioted"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortisms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("scematic"), &["schematic"]),
            (
                dictgen::InsensitiveStr::Ascii("substantied"),
                &["substantive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congifuring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("populare"), &["popular"]),
            (dictgen::InsensitiveStr::Ascii("comited"), &["committed"]),
            (
                dictgen::InsensitiveStr::Ascii("collabortation"),
                &["collaboration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sencond"), &["second"]),
            (
                dictgen::InsensitiveStr::Ascii("nodeterministic"),
                &["nondeterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("totatlly"), &["totally"]),
            (
                dictgen::InsensitiveStr::Ascii("asychronous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caculating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunderstamding"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accesiibility"),
                &["accessibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("maake"), &["make"]),
            (
                dictgen::InsensitiveStr::Ascii("resue"),
                &["reuse", "rescue", "resume"],
            ),
            (dictgen::InsensitiveStr::Ascii("metedata"), &["metadata"]),
            (
                dictgen::InsensitiveStr::Ascii("imeediately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undoubtedy"),
                &["undoubtedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrecthed"), &["wretched"]),
            (
                dictgen::InsensitiveStr::Ascii("diagnistic"),
                &["diagnostic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polymorhic"),
                &["polymorphic"],
            ),
            (dictgen::InsensitiveStr::Ascii("metricas"), &["metrics"]),
            (dictgen::InsensitiveStr::Ascii("seperator"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("decrmeneted"),
                &["decremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellectuels"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("qoutation"), &["quotation"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidted"),
                &["invalidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwirting"),
                &["overwriting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predictave"),
                &["predictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("peroper"), &["proper"]),
            (dictgen::InsensitiveStr::Ascii("attands"), &["attends"]),
            (dictgen::InsensitiveStr::Ascii("releifing"), &["relieving"]),
            (
                dictgen::InsensitiveStr::Ascii("cliffbanger"),
                &["cliffhanger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memerization"),
                &["memorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("fied"), &["field", "fiend"]),
            (
                dictgen::InsensitiveStr::Ascii("transolated"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vist"),
                &[
                    "visit", "fist", "gist", "list", "mist", "vast", "vest", "vista",
                ],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installent"),
                &["installment"],
            ),
            (dictgen::InsensitiveStr::Ascii("opjecteing"), &["objecting"]),
            (dictgen::InsensitiveStr::Ascii("privilage"), &["privilege"]),
            (
                dictgen::InsensitiveStr::Ascii("imigrate"),
                &["immigrate", "emigrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("awat"), &["await"]),
            (dictgen::InsensitiveStr::Ascii("conifig"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("lacklusture"),
                &["lacklustre"],
            ),
            (dictgen::InsensitiveStr::Ascii("picknic"), &["picnic"]),
            (dictgen::InsensitiveStr::Ascii("intiating"), &["initiating"]),
            (
                dictgen::InsensitiveStr::Ascii("envryption"),
                &["encryption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberviation"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("foults"), &["faults"]),
            (
                dictgen::InsensitiveStr::Ascii("disrputing"),
                &["disrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquisito"),
                &["acquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("bluetootn"), &["bluetooth"]),
            (dictgen::InsensitiveStr::Ascii("adaptaters"), &["adapters"]),
            (
                dictgen::InsensitiveStr::Ascii("regeneracion"),
                &["regeneration"],
            ),
            (dictgen::InsensitiveStr::Ascii("confif"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("palestininan"),
                &["palestinian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summurising"),
                &["summarising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effecitively"),
                &["effectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("sectopms"), &["sections"]),
            (dictgen::InsensitiveStr::Ascii("speciied"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenion"),
                &["dimension", "dominion"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumulates"), &["simulates"]),
            (
                dictgen::InsensitiveStr::Ascii("symtomatically"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("enebale"), &["enable"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfitied"),
                &["identified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpections"),
                &["inspections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualizaiton"),
                &["visualization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compromisses"),
                &["compromises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stimluating"),
                &["stimulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lightweigt"),
                &["lightweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("optimizier"), &["optimizer"]),
            (dictgen::InsensitiveStr::Ascii("unusabe"), &["unusable"]),
            (dictgen::InsensitiveStr::Ascii("formant"), &["format"]),
            (
                dictgen::InsensitiveStr::Ascii("fashonable"),
                &["fashionable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mounth"),
                &["month", "mouth"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simptumaticly"),
                &["symptomatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("whisltes"), &["whistles"]),
            (dictgen::InsensitiveStr::Ascii("contentes"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("exixst"), &["exist"]),
            (dictgen::InsensitiveStr::Ascii("sciprting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("overlodaded"),
                &["overloaded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriotics"),
                &["patriotism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmunicate"),
                &["communicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("referinng"), &["referring"]),
            (
                dictgen::InsensitiveStr::Ascii("tolearnces"),
                &["tolerances"],
            ),
            (dictgen::InsensitiveStr::Ascii("leavong"), &["leaving"]),
            (dictgen::InsensitiveStr::Ascii("connnector"), &["connector"]),
            (dictgen::InsensitiveStr::Ascii("probbed"), &["probed"]),
            (
                dictgen::InsensitiveStr::Ascii("opitimisers"),
                &["optimisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relocaions"),
                &["relocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reinfoced"), &["reinforced"]),
            (dictgen::InsensitiveStr::Ascii("loev"), &["love"]),
            (dictgen::InsensitiveStr::Ascii("spcify"), &["specify"]),
            (dictgen::InsensitiveStr::Ascii("damamged"), &["damaged"]),
            (
                dictgen::InsensitiveStr::Ascii("downstaris"),
                &["downstairs"],
            ),
            (dictgen::InsensitiveStr::Ascii("touristes"), &["tourists"]),
            (
                dictgen::InsensitiveStr::Ascii("concentraiton"),
                &["concentrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("parameer"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("mainfestos"),
                &["manifestos"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thoese"),
                &["these", "those"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entending"),
                &["intending", "extending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreceating"),
                &["depreciating", "deprecating"],
            ),
            (dictgen::InsensitiveStr::Ascii("inputtin"), &["inputting"]),
            (dictgen::InsensitiveStr::Ascii("releasse"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("enfornces"), &["enforces"]),
            (dictgen::InsensitiveStr::Ascii("expeting"), &["expecting"]),
            (dictgen::InsensitiveStr::Ascii("resove"), &["resolve"]),
            (
                dictgen::InsensitiveStr::Ascii("screenshoots"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("recovry"), &["recovery"]),
            (
                dictgen::InsensitiveStr::Ascii("ideosincracy"),
                &["idiosyncrasy"],
            ),
            (dictgen::InsensitiveStr::Ascii("exerciesd"), &["exercised"]),
            (dictgen::InsensitiveStr::Ascii("unsused"), &["unused"]),
            (
                dictgen::InsensitiveStr::Ascii("parellelizing"),
                &["parallelizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presentatin"),
                &["presentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependnces"),
                &["dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accademically"),
                &["academically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensititivies"),
                &["sensitivities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egonomic"),
                &["ergonomic", "economic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cousines"), &["cousins"]),
            (
                dictgen::InsensitiveStr::Ascii("theim"),
                &["thiem", "them", "theism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customsier"),
                &["customiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("theere"), &["there"]),
            (dictgen::InsensitiveStr::Ascii("simlest"), &["simplest"]),
            (
                dictgen::InsensitiveStr::Ascii("dupplicating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleanner"), &["cleaner"]),
            (
                dictgen::InsensitiveStr::Ascii("gauntanamo"),
                &["guantanamo"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contextful"),
                &["contextual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("savve"),
                &["save", "savvy", "salve"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rectangualr"),
                &["rectangular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomputaion"),
                &["recomputation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("saksatchewan"),
                &["saskatchewan"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempated"), &["templated"]),
            (
                dictgen::InsensitiveStr::Ascii("camparatively"),
                &["comparatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypocrytes"),
                &["hypocrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intenational"),
                &["international"],
            ),
            (dictgen::InsensitiveStr::Ascii("prioroity"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("hysteriia"), &["hysteria"]),
            (
                dictgen::InsensitiveStr::Ascii("efficiancy"),
                &["efficiency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expirementations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("temtation"), &["temptation"]),
            (
                dictgen::InsensitiveStr::Ascii("transformaton"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silouweting"),
                &["silhouetting"],
            ),
            (dictgen::InsensitiveStr::Ascii("convetion"), &["convention"]),
            (
                dictgen::InsensitiveStr::Ascii("destrination"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacemets"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("chlidfree"), &["childfree"]),
            (dictgen::InsensitiveStr::Ascii("critcal"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("sypnosis"), &["synopsis"]),
            (
                dictgen::InsensitiveStr::Ascii("operatations"),
                &["operations"],
            ),
            (dictgen::InsensitiveStr::Ascii("auccess"), &["success"]),
            (dictgen::InsensitiveStr::Ascii("declarnig"), &["declaring"]),
            (dictgen::InsensitiveStr::Ascii("adamently"), &["adamantly"]),
            (dictgen::InsensitiveStr::Ascii("explicatie"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("interfals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("crachs"),
                &["cracks", "crashes"],
            ),
            (dictgen::InsensitiveStr::Ascii("anciallary"), &["ancillary"]),
            (dictgen::InsensitiveStr::Ascii("symoblism"), &["symbolism"]),
            (
                dictgen::InsensitiveStr::Ascii("irrationatly"),
                &["irrationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typopgrahical"),
                &["typographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troube"),
                &["trouble", "troupe"],
            ),
            (dictgen::InsensitiveStr::Ascii("peanockle"), &["pinochle"]),
            (dictgen::InsensitiveStr::Ascii("paralleles"), &["parallels"]),
            (dictgen::InsensitiveStr::Ascii("jospeh"), &["joseph"]),
            (dictgen::InsensitiveStr::Ascii("vigilane"), &["vigilante"]),
            (
                dictgen::InsensitiveStr::Ascii("traslalated"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appreicated"),
                &["appreciated"],
            ),
            (dictgen::InsensitiveStr::Ascii("airfow"), &["airflow"]),
            (
                dictgen::InsensitiveStr::Ascii("masturdate"),
                &["masturbate"],
            ),
            (dictgen::InsensitiveStr::Ascii("acousitc"), &["acoustic"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanntment"),
                &["enchantment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transculent"),
                &["translucent"],
            ),
            (dictgen::InsensitiveStr::Ascii("smoewhere"), &["somewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("disrepancies"),
                &["discrepancies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuncurency"),
                &["concurrency"],
            ),
            (dictgen::InsensitiveStr::Ascii("bginning"), &["beginning"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatons"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("descripor"), &["descriptor"]),
            (dictgen::InsensitiveStr::Ascii("cliuents"), &["clients"]),
            (dictgen::InsensitiveStr::Ascii("enginge"), &["engine"]),
            (dictgen::InsensitiveStr::Ascii("painfullly"), &["painfully"]),
            (dictgen::InsensitiveStr::Ascii("gneration"), &["generation"]),
            (dictgen::InsensitiveStr::Ascii("adjasensy"), &["adjacency"]),
            (
                dictgen::InsensitiveStr::Ascii("mistmatches"),
                &["mismatches"],
            ),
            (dictgen::InsensitiveStr::Ascii("brwosable"), &["browsable"]),
            (
                dictgen::InsensitiveStr::Ascii("descenting"),
                &["descending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eigrnvector"),
                &["eigenvector"],
            ),
            (dictgen::InsensitiveStr::Ascii("multipliy"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("sanaty"), &["sanity"]),
            (dictgen::InsensitiveStr::Ascii("possoble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("indciation"),
                &["indication"],
            ),
            (dictgen::InsensitiveStr::Ascii("rebuid"), &["rebuild"]),
            (dictgen::InsensitiveStr::Ascii("faliure"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("gauarana"), &["guaraná"]),
            (
                dictgen::InsensitiveStr::Ascii("combindation"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("asnwering"), &["answering"]),
            (dictgen::InsensitiveStr::Ascii("upgrates"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("consums"), &["consumes"]),
            (
                dictgen::InsensitiveStr::Ascii("sascatchewan"),
                &["saskatchewan"],
            ),
            (dictgen::InsensitiveStr::Ascii("generatos"), &["generators"]),
            (dictgen::InsensitiveStr::Ascii("remaner"), &["remainder"]),
            (dictgen::InsensitiveStr::Ascii("contagen"), &["contagion"]),
            (
                dictgen::InsensitiveStr::Ascii("accessiblei"),
                &["accessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("confilcts"), &["conflicts"]),
            (
                dictgen::InsensitiveStr::Ascii("sensacionalism"),
                &["sensationalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrarity"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("professionnals"),
                &["professionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrrived"), &["arrived"]),
            (
                dictgen::InsensitiveStr::Ascii("acelerometer"),
                &["accelerometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("gardient"), &["gradient"]),
            (dictgen::InsensitiveStr::Ascii("algotithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("accurracy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("teaspooon"), &["teaspoon"]),
            (dictgen::InsensitiveStr::Ascii("stretchs"), &["stretches"]),
            (dictgen::InsensitiveStr::Ascii("missusing"), &["misusing"]),
            (
                dictgen::InsensitiveStr::Ascii("conspiricy"),
                &["conspiracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensinos"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucesscient"),
                &["sufficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("termiante"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("enabes"), &["enables"]),
            (dictgen::InsensitiveStr::Ascii("noteboooks"), &["notebooks"]),
            (
                dictgen::InsensitiveStr::Ascii("confings"),
                &["configs", "confines"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyars"), &["years", "eyas"]),
            (
                dictgen::InsensitiveStr::Ascii("flamethorwer"),
                &["flamethrower"],
            ),
            (dictgen::InsensitiveStr::Ascii("barlkey"), &["barkley"]),
            (
                dictgen::InsensitiveStr::Ascii("receptionnist"),
                &["receptionist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equelibrium"),
                &["equilibrium"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exerienced"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("joysitck"), &["joystick"]),
            (
                dictgen::InsensitiveStr::Ascii("monumentous"),
                &["monuments"],
            ),
            (dictgen::InsensitiveStr::Ascii("aspestus"), &["asbestos"]),
            (dictgen::InsensitiveStr::Ascii("ankshisly"), &["anxiously"]),
            (dictgen::InsensitiveStr::Ascii("instaces"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("opten"), &["often", "open"]),
            (
                dictgen::InsensitiveStr::Ascii("derects"),
                &["directs", "detects", "defects", "erects"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttempts"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialziation"),
                &["deserialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regitrations"),
                &["registrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("thannks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("mesengers"), &["messengers"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitator"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatable"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("comapnies"), &["companies"]),
            (dictgen::InsensitiveStr::Ascii("linclon"), &["lincoln"]),
            (
                dictgen::InsensitiveStr::Ascii("comemnting"),
                &["commenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("bunbling"), &["bundling"]),
            (dictgen::InsensitiveStr::Ascii("expirence"), &["experience"]),
            (
                dictgen::InsensitiveStr::Ascii("synchroous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("nighlties"), &["nightlies"]),
            (dictgen::InsensitiveStr::Ascii("missused"), &["misused"]),
            (dictgen::InsensitiveStr::Ascii("apocraful"), &["apocryphal"]),
            (
                dictgen::InsensitiveStr::Ascii("circumsised"),
                &["circumcised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctivates"),
                &["deactivates"],
            ),
            (dictgen::InsensitiveStr::Ascii("hedaer"), &["header"]),
            (dictgen::InsensitiveStr::Ascii("jagaurs"), &["jaguars"]),
            (
                dictgen::InsensitiveStr::Ascii("dessicated"),
                &["desiccated"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxinum"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("indentically"),
                &["identically"],
            ),
            (dictgen::InsensitiveStr::Ascii("awsome"), &["awesome"]),
            (
                dictgen::InsensitiveStr::Ascii("reciporcate"),
                &["reciprocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhalted"), &["exalted"]),
            (dictgen::InsensitiveStr::Ascii("warnigns"), &["warnings"]),
            (dictgen::InsensitiveStr::Ascii("superceede"), &["supersede"]),
            (dictgen::InsensitiveStr::Ascii("feminim"), &["feminism"]),
            (
                dictgen::InsensitiveStr::Ascii("renedered"),
                &["rende", "rendered"],
            ),
            (dictgen::InsensitiveStr::Ascii("failre"), &["failure"]),
            (
                dictgen::InsensitiveStr::Ascii("bodybuildig"),
                &["bodybuilding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mediocrily"),
                &["mediocrity"],
            ),
            (dictgen::InsensitiveStr::Ascii("gnorung"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("inlclude"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("expirements"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("depreates"), &["deprecates"]),
            (dictgen::InsensitiveStr::Ascii("warrent"), &["warrant"]),
            (dictgen::InsensitiveStr::Ascii("symbolsim"), &["symbolism"]),
            (dictgen::InsensitiveStr::Ascii("booleamns"), &["booleans"]),
            (dictgen::InsensitiveStr::Ascii("blockise"), &["blockwise"]),
            (dictgen::InsensitiveStr::Ascii("socailly"), &["socially"]),
            (dictgen::InsensitiveStr::Ascii("scramling"), &["scrambling"]),
            (dictgen::InsensitiveStr::Ascii("preceedes"), &["precedes"]),
            (dictgen::InsensitiveStr::Ascii("tehcies"), &["techies"]),
            (
                dictgen::InsensitiveStr::Ascii("censabilities"),
                &["sensibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endoctrinated"),
                &["indoctrinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distruction"),
                &["destruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("pugins"), &["plugins"]),
            (dictgen::InsensitiveStr::Ascii("enforecers"), &["enforcers"]),
            (
                dictgen::InsensitiveStr::Ascii("confussions"),
                &["confessions", "confusions", "concussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenitfiable"),
                &["identifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monopollies"),
                &["monopolies"],
            ),
            (dictgen::InsensitiveStr::Ascii("allcaters"), &["allocators"]),
            (dictgen::InsensitiveStr::Ascii("accordling"), &["according"]),
            (dictgen::InsensitiveStr::Ascii("regading"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("dahboards"), &["dashboards"]),
            (dictgen::InsensitiveStr::Ascii("mornin"), &["morning"]),
            (dictgen::InsensitiveStr::Ascii("psudo"), &["pseudo"]),
            (
                dictgen::InsensitiveStr::Ascii("collocalized"),
                &["colocalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("nigth"), &["night", "nigh"]),
            (dictgen::InsensitiveStr::Ascii("singelar"), &["singular"]),
            (
                dictgen::InsensitiveStr::Ascii("algotithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparte"), &["compare"]),
            (dictgen::InsensitiveStr::Ascii("booees"), &["buoys"]),
            (dictgen::InsensitiveStr::Ascii("encoede"), &["encode"]),
            (
                dictgen::InsensitiveStr::Ascii("progressief"),
                &["progressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concentrace"),
                &["concentrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parocheality"),
                &["parochiality"],
            ),
            (dictgen::InsensitiveStr::Ascii("requries"), &["requires"]),
            (
                dictgen::InsensitiveStr::Ascii("chercked"),
                &["checked", "checkered"],
            ),
            (dictgen::InsensitiveStr::Ascii("overvrides"), &["overrides"]),
            (dictgen::InsensitiveStr::Ascii("reserch"), &["research"]),
            (dictgen::InsensitiveStr::Ascii("tablular"), &["tabular"]),
            (
                dictgen::InsensitiveStr::Ascii("legimitacy"),
                &["legitimacy"],
            ),
            (dictgen::InsensitiveStr::Ascii("comlain"), &["complain"]),
            (dictgen::InsensitiveStr::Ascii("rmoved"), &["removed"]),
            (dictgen::InsensitiveStr::Ascii("feautring"), &["featuring"]),
            (dictgen::InsensitiveStr::Ascii("mypsace"), &["myspace"]),
            (
                dictgen::InsensitiveStr::Ascii("categorie"),
                &["category", "categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("mirored"), &["mirrored"]),
            (dictgen::InsensitiveStr::Ascii("torotise"), &["tortoise"]),
            (
                dictgen::InsensitiveStr::Ascii("ovewriting"),
                &["overwriting"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiby"), &["possibly"]),
            (dictgen::InsensitiveStr::Ascii("prairy"), &["prairie"]),
            (
                dictgen::InsensitiveStr::Ascii("connceting"),
                &["connecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeatetly"),
                &["repeatedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("redefiende"), &["redefined"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprehinsible"),
                &["incomprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rysurrection"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("mistics"), &["mystics"]),
            (dictgen::InsensitiveStr::Ascii("folowwinwg"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("beofre"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("ralating"), &["relating"]),
            (dictgen::InsensitiveStr::Ascii("alwans"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("angshisly"), &["anxiously"]),
            (dictgen::InsensitiveStr::Ascii("cheecking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("negaive"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("frontent"), &["frontend"]),
            (
                dictgen::InsensitiveStr::Ascii("intializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("combinding"), &["combining"]),
            (dictgen::InsensitiveStr::Ascii("hismelf"), &["himself"]),
            (dictgen::InsensitiveStr::Ascii("inizial"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("inxes"),
                &["indexes", "indices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bltting"),
                &["blitting", "blotting"],
            ),
            (dictgen::InsensitiveStr::Ascii("qaulifies"), &["qualifies"]),
            (dictgen::InsensitiveStr::Ascii("sovlers"), &["solvers"]),
            (dictgen::InsensitiveStr::Ascii("repetiton"), &["repetition"]),
            (dictgen::InsensitiveStr::Ascii("precieved"), &["perceived"]),
            (
                dictgen::InsensitiveStr::Ascii("nevetheless"),
                &["nevertheless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assoiciative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabicity"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("nrivana"), &["nirvana"]),
            (dictgen::InsensitiveStr::Ascii("scooterers"), &["scooters"]),
            (
                dictgen::InsensitiveStr::Ascii("wornings"),
                &["warnings", "mornings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relaxiation"),
                &["relaxation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpetraters"),
                &["perpetrators", "perpetrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeled"), &["repelled"]),
            (dictgen::InsensitiveStr::Ascii("navagate"), &["navigate"]),
            (
                dictgen::InsensitiveStr::Ascii("contorllers"),
                &["controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ostenisbly"),
                &["ostensibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("exptecting"), &["expecting"]),
            (
                dictgen::InsensitiveStr::Ascii("vew"),
                &["view", "vow", "vex"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misunterstanding"),
                &["misunderstandings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("objectivley"),
                &["objectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("siwzzle"), &["swizzle"]),
            (
                dictgen::InsensitiveStr::Ascii("followwings"),
                &["followings"],
            ),
            (dictgen::InsensitiveStr::Ascii("idefinite"), &["indefinite"]),
            (
                dictgen::InsensitiveStr::Ascii("auxilaries"),
                &["auxiliaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("buiilds"), &["builds"]),
            (dictgen::InsensitiveStr::Ascii("curcible"), &["crucible"]),
            (
                dictgen::InsensitiveStr::Ascii("detets"),
                &["detects", "deletes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregsters"),
                &["unregisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endorces"),
                &["endorses", "enforces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrruptible"),
                &["corruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("alledged"), &["alleged"]),
            (
                dictgen::InsensitiveStr::Ascii("screenshoot"),
                &["screenshot"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactors"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wearns"),
                &["warns", "wears", "earns", "learns", "yearns"],
            ),
            (dictgen::InsensitiveStr::Ascii("paramers"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("prescence"), &["presence"]),
            (dictgen::InsensitiveStr::Ascii("variabe"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("iornman"), &["ironman"]),
            (dictgen::InsensitiveStr::Ascii("binar"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("confogurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neccessaty"),
                &["necessity", "necessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplaicates"),
                &["deduplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("seprator"), &["separator"]),
            (dictgen::InsensitiveStr::Ascii("ninimally"), &["minimally"]),
            (
                dictgen::InsensitiveStr::Ascii("experimenations"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutilpart"), &["multipart"]),
            (
                dictgen::InsensitiveStr::Ascii("istances"),
                &["instances", "distances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicionaries"),
                &["dictionaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirearcies"),
                &["hierarchies"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaulg"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("aranges"), &["arranges"]),
            (
                dictgen::InsensitiveStr::Ascii("instruktor"),
                &["instructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("accually"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("telocom"), &["telecom"]),
            (
                dictgen::InsensitiveStr::Ascii("descritions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discreprancies"),
                &["discrepancies"],
            ),
            (dictgen::InsensitiveStr::Ascii("levetates"), &["levitates"]),
            (
                dictgen::InsensitiveStr::Ascii("arrancements"),
                &["arrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bloodboorne"),
                &["bloodborne"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrondings"),
                &["surroundings"],
            ),
            (dictgen::InsensitiveStr::Ascii("annotaed"), &["annotated"]),
            (
                dictgen::InsensitiveStr::Ascii("embaressed"),
                &["embarrassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghs"), &["lengths"]),
            (dictgen::InsensitiveStr::Ascii("ignote"), &["ignore"]),
            (dictgen::InsensitiveStr::Ascii("includind"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("desription"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("psychedilics"),
                &["psychedelics"],
            ),
            (dictgen::InsensitiveStr::Ascii("funnnily"), &["funnily"]),
            (dictgen::InsensitiveStr::Ascii("boxs"), &["box", "boxes"]),
            (
                dictgen::InsensitiveStr::Ascii("disttribution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("warks"), &["works"]),
            (
                dictgen::InsensitiveStr::Ascii("incentivate"),
                &["incentive"],
            ),
            (dictgen::InsensitiveStr::Ascii("cehcksum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("negitiable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumurates"),
                &["enumerates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretions"),
                &["misinterpreting"],
            ),
            (dictgen::InsensitiveStr::Ascii("reanme"), &["rename"]),
            (
                dictgen::InsensitiveStr::Ascii("atomspheric"),
                &["atmospheric"],
            ),
            (dictgen::InsensitiveStr::Ascii("increses"), &["increases"]),
            (
                dictgen::InsensitiveStr::Ascii("exchangable"),
                &["exchangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("mumbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("conjugages"),
                &["conjugates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowledgeing"),
                &["acknowledging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkpoit"),
                &["checkpoints", "checkpoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("operaions"), &["operations"]),
            (
                dictgen::InsensitiveStr::Ascii("prezidential"),
                &["presidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("externaly"), &["externally"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilies"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cathegorized"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("riaders"), &["raiders"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitioned"),
                &["functioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diagonstic"),
                &["diagnostic"],
            ),
            (dictgen::InsensitiveStr::Ascii("ackwards"), &["backwards"]),
            (dictgen::InsensitiveStr::Ascii("vestage"), &["vestige"]),
            (dictgen::InsensitiveStr::Ascii("tnank"), &["thank"]),
            (dictgen::InsensitiveStr::Ascii("shoule"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("immediantly"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("internts"), &["interns"]),
            (dictgen::InsensitiveStr::Ascii("suppliad"), &["supplied"]),
            (dictgen::InsensitiveStr::Ascii("orccurs"), &["occurs"]),
            (dictgen::InsensitiveStr::Ascii("bracess"), &["braces"]),
            (
                dictgen::InsensitiveStr::Ascii("caterpillers"),
                &["caterpillars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("juptyer"),
                &["jupyter", "jupiter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasturises"),
                &["pasteurises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determenation"),
                &["determination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improbment"),
                &["improvement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("improtation"),
                &["importation"],
            ),
            (dictgen::InsensitiveStr::Ascii("releas"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("converastion"),
                &["conversations"],
            ),
            (dictgen::InsensitiveStr::Ascii("handcufs"), &["handcuffs"]),
            (dictgen::InsensitiveStr::Ascii("setteling"), &["settling"]),
            (dictgen::InsensitiveStr::Ascii("arhiver"), &["archiver"]),
            (dictgen::InsensitiveStr::Ascii("monrachy"), &["monarchy"]),
            (dictgen::InsensitiveStr::Ascii("exending"), &["extending"]),
            (
                dictgen::InsensitiveStr::Ascii("libertariansism"),
                &["libertarianism"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxept"), &["except"]),
            (
                dictgen::InsensitiveStr::Ascii("adolescance"),
                &["adolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("esponder"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("soltion"), &["solution"]),
            (
                dictgen::InsensitiveStr::Ascii("calcultion"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suggetions"),
                &["suggestions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triskadecafobia"),
                &["triskaidekaphobia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extremests"),
                &["extremists"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbords"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("acccount"), &["account"]),
            (dictgen::InsensitiveStr::Ascii("totorials"), &["tutorials"]),
            (
                dictgen::InsensitiveStr::Ascii("convenctional"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("consept"), &["concept"]),
            (
                dictgen::InsensitiveStr::Ascii("abrreviates"),
                &["abbreviates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legalazing"),
                &["legalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contraversy"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("paralllels"), &["parallels"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthough"),
                &["breakthrough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("validaton"),
                &["validation", "validator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collectivly"),
                &["collectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensivley"),
                &["extensively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gayity"),
                &["gayety", "gaiety"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prooperties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arragements"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("semestres"), &["semesters"]),
            (
                dictgen::InsensitiveStr::Ascii("spectaters"),
                &["spectators"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlysed"), &["analysed"]),
            (dictgen::InsensitiveStr::Ascii("willl"), &["will"]),
            (
                dictgen::InsensitiveStr::Ascii("specififcally"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializatons"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doublelfit"),
                &["doublelift"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explenation"),
                &["explanation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongising"),
                &["recognising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporraries"),
                &["temporaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availaiable"),
                &["available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bastraction"),
                &["abstraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desposit"),
                &["deposit", "deposition"],
            ),
            (dictgen::InsensitiveStr::Ascii("abilties"), &["abilities"]),
            (
                dictgen::InsensitiveStr::Ascii("adminsiters"),
                &["administers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeliavably"),
                &["unbelievably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiblility"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumarize"), &["summarize"]),
            (dictgen::InsensitiveStr::Ascii("systme"), &["system"]),
            (
                dictgen::InsensitiveStr::Ascii("aggregetes"),
                &["aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reguarliser"),
                &["regulariser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspindents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("penicls"), &["pencils"]),
            (
                dictgen::InsensitiveStr::Ascii("predominantely"),
                &["predominantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atrocoties"),
                &["atrocities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wellignton"),
                &["wellington"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sympothize"),
                &["sympathize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quinessential"),
                &["quintessential"],
            ),
            (dictgen::InsensitiveStr::Ascii("genereates"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("reaquested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("laucnher"), &["launcher"]),
            (
                dictgen::InsensitiveStr::Ascii("neighborehood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogynistc"),
                &["misogynistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeptions"), &["exceptions"]),
            (
                dictgen::InsensitiveStr::Ascii("accumlates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conviniently"),
                &["conveniently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nighboring"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("detrmines"), &["determines"]),
            (
                dictgen::InsensitiveStr::Ascii("deprectaion"),
                &["deprecation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reevaluatin"),
                &["reevaluating", "reevaluation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traditionaly"),
                &["traditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscoitaive"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("incriment"), &["increment"]),
            (
                dictgen::InsensitiveStr::Ascii("coordenate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("writet"), &["writes"]),
            (dictgen::InsensitiveStr::Ascii("dichomoty"), &["dichotomy"]),
            (dictgen::InsensitiveStr::Ascii("sepecified"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("programmeur"),
                &["programmer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delightlful"),
                &["delightful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pattened"),
                &["patented", "patterned"],
            ),
            (dictgen::InsensitiveStr::Ascii("consiers"), &["considers"]),
            (dictgen::InsensitiveStr::Ascii("rebiulds"), &["rebuilds"]),
            (dictgen::InsensitiveStr::Ascii("assumas"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("impelments"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("catapul"), &["catapult"]),
            (
                dictgen::InsensitiveStr::Ascii("multplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("ralates"), &["relates"]),
            (dictgen::InsensitiveStr::Ascii("succesed"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("indiviual"), &["individual"]),
            (
                dictgen::InsensitiveStr::Ascii("inititaliser"),
                &["initialiser"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluclators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incredibley"),
                &["incredibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undenyable"),
                &["undeniable"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplikay"), &["appliqué"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolorant"), &["tolerant"]),
            (dictgen::InsensitiveStr::Ascii("ands"), &["and"]),
            (
                dictgen::InsensitiveStr::Ascii("assumtpions"),
                &["assumptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocupies"), &["occupies"]),
            (
                dictgen::InsensitiveStr::Ascii("formualated"),
                &["formulated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterisics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("tagnetial"), &["tangential"]),
            (
                dictgen::InsensitiveStr::Ascii("palesitnian"),
                &["palestinians"],
            ),
            (dictgen::InsensitiveStr::Ascii("spindels"), &["spindles"]),
            (dictgen::InsensitiveStr::Ascii("paragraphy"), &["paragraph"]),
            (
                dictgen::InsensitiveStr::Ascii("trasngendered"),
                &["transgendered"],
            ),
            (dictgen::InsensitiveStr::Ascii("absuer"), &["abuser"]),
            (
                dictgen::InsensitiveStr::Ascii("hypothosis"),
                &["hypothesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chekerboard"),
                &["checkerboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarassment"),
                &["embarrassment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronization"),
                &["synchronization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("staturation"),
                &["saturation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ninima"), &["minima"]),
            (
                dictgen::InsensitiveStr::Ascii("crrespondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elecricity"),
                &["electricity"],
            ),
            (dictgen::InsensitiveStr::Ascii("bianries"), &["binaries"]),
            (dictgen::InsensitiveStr::Ascii("jewl"), &["jew", "jewel"]),
            (
                dictgen::InsensitiveStr::Ascii("compramising"),
                &["compromising"],
            ),
            (dictgen::InsensitiveStr::Ascii("referig"), &["referring"]),
            (dictgen::InsensitiveStr::Ascii("sparately"), &["separately"]),
            (dictgen::InsensitiveStr::Ascii("wnen"), &["when", "wen"]),
            (dictgen::InsensitiveStr::Ascii("tcahce"), &["cache"]),
            (dictgen::InsensitiveStr::Ascii("underrrun"), &["underrun"]),
            (dictgen::InsensitiveStr::Ascii("pakcage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("intpreter"),
                &["interpreter"],
            ),
            (dictgen::InsensitiveStr::Ascii("deveolper"), &["developer"]),
            (
                dictgen::InsensitiveStr::Ascii("voluntairly"),
                &["voluntarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transportarte"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("mismatchs"), &["mismatches"]),
            (
                dictgen::InsensitiveStr::Ascii("regestrations"),
                &["registrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grassrooters"),
                &["grassroots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unreasonal"),
                &["unreasonably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subscibres"),
                &["subscribers", "subscribes"],
            ),
            (dictgen::InsensitiveStr::Ascii("comotion"), &["commotion"]),
            (
                dictgen::InsensitiveStr::Ascii("supriously"),
                &["spuriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("especally"), &["especially"]),
            (dictgen::InsensitiveStr::Ascii("errorr"), &["error"]),
            (
                dictgen::InsensitiveStr::Ascii("lighthorse"),
                &["lighthouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instnsiation"),
                &["instantiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("fogotten"), &["forgotten"]),
            (
                dictgen::InsensitiveStr::Ascii("roductions"),
                &["reductions", "productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("necesssary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("innaccuracy"),
                &["inaccuracy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rqeuesting"),
                &["requesting"],
            ),
            (dictgen::InsensitiveStr::Ascii("detaction"), &["detection"]),
            (
                dictgen::InsensitiveStr::Ascii("frequentily"),
                &["frequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("switich"), &["switch"]),
            (
                dictgen::InsensitiveStr::Ascii("differencess"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("endding"), &["ending"]),
            (
                dictgen::InsensitiveStr::Ascii("architechtures"),
                &["architectures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsequest"),
                &["subsequent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predujices"),
                &["prejudices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decallocate"),
                &["deallocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asymetical"),
                &["asymmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotioting"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monstrosoty"),
                &["monstrosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("posessed"), &["possessed"]),
            (dictgen::InsensitiveStr::Ascii("unknonw"), &["unknown"]),
            (dictgen::InsensitiveStr::Ascii("timestmap"), &["timestamp"]),
            (
                dictgen::InsensitiveStr::Ascii("synchrononous"),
                &["synchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoated"), &["annotated"]),
            (
                dictgen::InsensitiveStr::Ascii("scarmbling"),
                &["scrambling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenricated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assitsants"),
                &["assistants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flourescent"),
                &["fluorescent", "florescent"],
            ),
            (dictgen::InsensitiveStr::Ascii("stratled"), &["startled"]),
            (dictgen::InsensitiveStr::Ascii("younget"), &["youngest"]),
            (dictgen::InsensitiveStr::Ascii("recjected"), &["rejected"]),
            (dictgen::InsensitiveStr::Ascii("sigurets"), &["cigarets"]),
            (dictgen::InsensitiveStr::Ascii("destorys"), &["destroys"]),
            (
                dictgen::InsensitiveStr::Ascii("overpopluation"),
                &["overpopulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("corupts"), &["corrupts"]),
            (
                dictgen::InsensitiveStr::Ascii("explecitly"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("believeble"),
                &["believable"],
            ),
            (dictgen::InsensitiveStr::Ascii("dictats"), &["dictates"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplcation"),
                &["multiplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhhods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenciated"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("encondes"), &["encodes"]),
            (dictgen::InsensitiveStr::Ascii("sawte"), &["sauté"]),
            (
                dictgen::InsensitiveStr::Ascii("sensitivies"),
                &["sensitivities"],
            ),
            (dictgen::InsensitiveStr::Ascii("ansering"), &["answering"]),
            (
                dictgen::InsensitiveStr::Ascii("respresents"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documnting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theorical"),
                &["theoretical"],
            ),
            (dictgen::InsensitiveStr::Ascii("confort"), &["comfort"]),
            (dictgen::InsensitiveStr::Ascii("resoruce"), &["resource"]),
            (dictgen::InsensitiveStr::Ascii("upstaris"), &["upstairs"]),
            (
                dictgen::InsensitiveStr::Ascii("veriasions"),
                &["variations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("belittleing"),
                &["belittling"],
            ),
            (dictgen::InsensitiveStr::Ascii("encourged"), &["encouraged"]),
            (dictgen::InsensitiveStr::Ascii("recommond"), &["recommend"]),
            (dictgen::InsensitiveStr::Ascii("suporior"), &["superior"]),
            (dictgen::InsensitiveStr::Ascii("discuused"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("maintaints"), &["maintains"]),
            (dictgen::InsensitiveStr::Ascii("assasssins"), &["assassins"]),
            (
                dictgen::InsensitiveStr::Ascii("commonding"),
                &["commanding", "commending"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchnged"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("scritp"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("reenfocing"),
                &["reenforcing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installator"),
                &["installer"],
            ),
            (dictgen::InsensitiveStr::Ascii("createed"), &["created"]),
            (
                dictgen::InsensitiveStr::Ascii("indepdendence"),
                &["independence"],
            ),
            (dictgen::InsensitiveStr::Ascii("pageing"), &["paging"]),
            (
                dictgen::InsensitiveStr::Ascii("duplicatin"),
                &["duplicating", "duplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfarms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("subsript"), &["subscript"]),
            (dictgen::InsensitiveStr::Ascii("terimate"), &["terminate"]),
            (dictgen::InsensitiveStr::Ascii("indianna"), &["indiana"]),
            (dictgen::InsensitiveStr::Ascii("naturens"), &["natures"]),
            (dictgen::InsensitiveStr::Ascii("savannh"), &["savannah"]),
            (
                dictgen::InsensitiveStr::Ascii("encorporated"),
                &["incorporated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statamenets"),
                &["statements"],
            ),
            (dictgen::InsensitiveStr::Ascii("massewer"), &["masseur"]),
            (dictgen::InsensitiveStr::Ascii("docment"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("preferentail"),
                &["preferential"],
            ),
            (dictgen::InsensitiveStr::Ascii("asain"), &["asian"]),
            (dictgen::InsensitiveStr::Ascii("interfeare"), &["interfere"]),
            (
                dictgen::InsensitiveStr::Ascii("interfact"),
                &["interact", "interface"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reveiver"),
                &["receiver", "reviewer", "reviver"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conculsive"),
                &["conclusive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rquirement"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrrangements"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("debuggging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("constrols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("viweed"), &["viewed"]),
            (
                dictgen::InsensitiveStr::Ascii("proffesionally"),
                &["professionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("heavywieght"),
                &["heavyweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("builr"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("dpeendently"),
                &["dependently"],
            ),
            (dictgen::InsensitiveStr::Ascii("backound"), &["background"]),
            (
                dictgen::InsensitiveStr::Ascii("serialialises"),
                &["serialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("movment"), &["movement"]),
            (
                dictgen::InsensitiveStr::Ascii("protagnoists"),
                &["protagonists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peodphilia"),
                &["pedophilia"],
            ),
            (dictgen::InsensitiveStr::Ascii("icludes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("ommissions"),
                &["omissions", "immissions", "commissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("enrypts"), &["encrypts"]),
            (dictgen::InsensitiveStr::Ascii("sensitve"), &["sensitive"]),
            (
                dictgen::InsensitiveStr::Ascii("contrubutes"),
                &["contributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socioeconimc"),
                &["socioeconomic"],
            ),
            (dictgen::InsensitiveStr::Ascii("operaand"), &["operand"]),
            (dictgen::InsensitiveStr::Ascii("checkker"), &["checker"]),
            (
                dictgen::InsensitiveStr::Ascii("psaces"),
                &["spaces", "paces"],
            ),
            (dictgen::InsensitiveStr::Ascii("empahsize"), &["emphasize"]),
            (dictgen::InsensitiveStr::Ascii("exexutor"), &["executor"]),
            (dictgen::InsensitiveStr::Ascii("stoer"), &["store"]),
            (
                dictgen::InsensitiveStr::Ascii("stuation"),
                &["situation", "station"],
            ),
            (dictgen::InsensitiveStr::Ascii("amibguity"), &["ambiguity"]),
            (dictgen::InsensitiveStr::Ascii("annoncer"), &["announcer"]),
            (
                dictgen::InsensitiveStr::Ascii("presedency"),
                &["presidency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mousturizing"),
                &["moisturizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("handeler"), &["handler"]),
            (
                dictgen::InsensitiveStr::Ascii("undesrtand"),
                &["understands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scrennshots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("questoin"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("timespanp"), &["timespan"]),
            (dictgen::InsensitiveStr::Ascii("insistis"), &["insists"]),
            (dictgen::InsensitiveStr::Ascii("minuites"), &["minutes"]),
            (dictgen::InsensitiveStr::Ascii("wth"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("duplacte"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("manufactuered"),
                &["manufactured"],
            ),
            (dictgen::InsensitiveStr::Ascii("falvors"), &["flavors"]),
            (
                dictgen::InsensitiveStr::Ascii("mellinneum"),
                &["millennium"],
            ),
            (dictgen::InsensitiveStr::Ascii("polocys"), &["policies"]),
            (dictgen::InsensitiveStr::Ascii("balences"), &["balances"]),
            (dictgen::InsensitiveStr::Ascii("buiness"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("arreays"), &["arrays"]),
            (
                dictgen::InsensitiveStr::Ascii("decresing"),
                &["decreasing", "deceasing"],
            ),
            (dictgen::InsensitiveStr::Ascii("nevelops"), &["envelops"]),
            (
                dictgen::InsensitiveStr::Ascii("conmpression"),
                &["compression"],
            ),
            (dictgen::InsensitiveStr::Ascii("hightlight"), &["highlight"]),
            (
                dictgen::InsensitiveStr::Ascii("custumizes"),
                &["customizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arranments"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifict"), &["artifact"]),
            (dictgen::InsensitiveStr::Ascii("languagee"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("involuntairy"),
                &["involuntary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conenctors"),
                &["connectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distination"),
                &["destination", "distinction"],
            ),
            (dictgen::InsensitiveStr::Ascii("keealive"), &["keepalive"]),
            (dictgen::InsensitiveStr::Ascii("nuetered"), &["neutered"]),
            (dictgen::InsensitiveStr::Ascii("staticly"), &["statically"]),
            (dictgen::InsensitiveStr::Ascii("herustic"), &["heuristic"]),
            (
                dictgen::InsensitiveStr::Ascii("firefigher"),
                &["firefighter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotes"),
                &["renegotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogorithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explainations"),
                &["explanations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctructively"),
                &["destructively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancellling"),
                &["cancelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("possisble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("commissionner"),
                &["commissioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("atomzier"), &["atomizer"]),
            (dictgen::InsensitiveStr::Ascii("formual"), &["formula"]),
            (
                dictgen::InsensitiveStr::Ascii("personalites"),
                &["personalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accummulators"),
                &["accumulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreating"), &["operating"]),
            (
                dictgen::InsensitiveStr::Ascii("brached"),
                &["branched", "breached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompleate"),
                &["incomplete"],
            ),
            (dictgen::InsensitiveStr::Ascii("grephic"), &["graphic"]),
            (dictgen::InsensitiveStr::Ascii("rebuit"), &["rebuilt"]),
            (dictgen::InsensitiveStr::Ascii("vandalsim"), &["vandalism"]),
            (
                dictgen::InsensitiveStr::Ascii("competance"),
                &["competence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapperead"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("implemetn"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("tocken"), &["token"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginzed"),
                &["recognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("tyrannia"), &["tyrannical"]),
            (dictgen::InsensitiveStr::Ascii("atempting"), &["attempting"]),
            (
                dictgen::InsensitiveStr::Ascii("iniitializations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("resovers"), &["resolvers"]),
            (dictgen::InsensitiveStr::Ascii("ecstacys"), &["ecstasy"]),
            (
                dictgen::InsensitiveStr::Ascii("commishioner"),
                &["commissioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sloothed"),
                &["sleuthing", "sleuthed"],
            ),
            (dictgen::InsensitiveStr::Ascii("eror"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("oulining"), &["outlining"]),
            (
                dictgen::InsensitiveStr::Ascii("millinnium"),
                &["millennium"],
            ),
            (dictgen::InsensitiveStr::Ascii("fnuction"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("transistioning"),
                &["transitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("exacly"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("mangment"), &["management"]),
            (
                dictgen::InsensitiveStr::Ascii("prerequesit"),
                &["prerequisite"],
            ),
            (dictgen::InsensitiveStr::Ascii("cacelling"), &["cancelling"]),
            (dictgen::InsensitiveStr::Ascii("calculats"), &["calculates"]),
            (dictgen::InsensitiveStr::Ascii("outuput"), &["output"]),
            (dictgen::InsensitiveStr::Ascii("targte"), &["target"]),
            (dictgen::InsensitiveStr::Ascii("pertinante"), &["pertinent"]),
            (dictgen::InsensitiveStr::Ascii("retunrn"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("alogirthmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programmare"),
                &["programmer"],
            ),
            (dictgen::InsensitiveStr::Ascii("speach"), &["speech"]),
            (dictgen::InsensitiveStr::Ascii("liceses"), &["licenses"]),
            (
                dictgen::InsensitiveStr::Ascii("coverted"),
                &["converted", "covered", "coveted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catostrophic"),
                &["catastrophic"],
            ),
            (dictgen::InsensitiveStr::Ascii("whic"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("occuppancies"),
                &["occupancies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiugrable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("oscure"), &["obscure"]),
            (
                dictgen::InsensitiveStr::Ascii("paramerizes"),
                &["parameterizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfuring"), &["preferring"]),
            (
                dictgen::InsensitiveStr::Ascii("cleanleness"),
                &["cleanliness"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranforms"), &["transforms"]),
            (
                dictgen::InsensitiveStr::Ascii("realtive"),
                &["relative", "reactive"],
            ),
            (dictgen::InsensitiveStr::Ascii("logisitcs"), &["logistics"]),
            (dictgen::InsensitiveStr::Ascii("delaied"), &["delayed"]),
            (
                dictgen::InsensitiveStr::Ascii("initializin"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exection"),
                &["execution", "ejection", "election", "exertion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmenents"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decaffinated"),
                &["decaffeinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thorughput"),
                &["throughput"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitplicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("alwaus"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("backounds"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protaginist"),
                &["protagonist"],
            ),
            (dictgen::InsensitiveStr::Ascii("bever"), &["never"]),
            (dictgen::InsensitiveStr::Ascii("spefcific"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("jeruselam"), &["jerusalem"]),
            (dictgen::InsensitiveStr::Ascii("postiions"), &["positions"]),
            (
                dictgen::InsensitiveStr::Ascii("trasversing"),
                &["traversing", "transversing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intilised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pased"),
                &["passed", "parsed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrievations"),
                &["abbreviation", "abbreviations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ganster"), &["gangster"]),
            (dictgen::InsensitiveStr::Ascii("providewd"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("descreases"), &["decreases"]),
            (dictgen::InsensitiveStr::Ascii("ontrols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("inital"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("algoroithms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("clearnance"), &["clearance"]),
            (dictgen::InsensitiveStr::Ascii("pupup"), &["popup"]),
            (dictgen::InsensitiveStr::Ascii("anachrist"), &["anarchist"]),
            (
                dictgen::InsensitiveStr::Ascii("separartion"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("awnsers"), &["answers"]),
            (dictgen::InsensitiveStr::Ascii("verisoned"), &["versioned"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilizare"),
                &["stabilize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predominently"),
                &["predominantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("opertes"), &["operates"]),
            (
                dictgen::InsensitiveStr::Ascii("roating"),
                &["rotating", "roasting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conenctivity"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("campared"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("informatikon"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmissiones"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrteady"), &["already"]),
            (
                dictgen::InsensitiveStr::Ascii("karismatic"),
                &["charismatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgnanized"), &["organized"]),
            (dictgen::InsensitiveStr::Ascii("animatiion"), &["animation"]),
            (
                dictgen::InsensitiveStr::Ascii("naiborhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astronautes"),
                &["astronauts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breakthruoghs"),
                &["breakthroughs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archtiecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confinguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("rigntone"), &["ringtone"]),
            (
                dictgen::InsensitiveStr::Ascii("comparitavely"),
                &["comparatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("wininpeg"), &["winnipeg"]),
            (dictgen::InsensitiveStr::Ascii("battary"), &["battery"]),
            (dictgen::InsensitiveStr::Ascii("encript"), &["encrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("enthuseastic"),
                &["enthusiastic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protfolios"),
                &["portfolios"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("userful"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("methodss"), &["methods"]),
            (
                dictgen::InsensitiveStr::Ascii("caricaturale"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ambadextrousness"),
                &["ambidextrousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("myraid"), &["myriad"]),
            (dictgen::InsensitiveStr::Ascii("contan"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("repetion"), &["repetition"]),
            (
                dictgen::InsensitiveStr::Ascii("associeating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conclusies"),
                &["conclusive"],
            ),
            (dictgen::InsensitiveStr::Ascii("surporter"), &["supporter"]),
            (dictgen::InsensitiveStr::Ascii("estiamate"), &["estimate"]),
            (
                dictgen::InsensitiveStr::Ascii("canceleable"),
                &["cancelable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deveopments"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrgation"),
                &["aggregation"],
            ),
            (dictgen::InsensitiveStr::Ascii("threded"), &["threaded"]),
            (
                dictgen::InsensitiveStr::Ascii("restartalbe"),
                &["restartable"],
            ),
            (dictgen::InsensitiveStr::Ascii("instiate"), &["instantiate"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalisin"),
                &["nationalising"],
            ),
            (dictgen::InsensitiveStr::Ascii("connetion"), &["connection"]),
            (dictgen::InsensitiveStr::Ascii("timeputs"), &["timeouts"]),
            (
                dictgen::InsensitiveStr::Ascii("decriptive"),
                &["descriptive"],
            ),
            (dictgen::InsensitiveStr::Ascii("openign"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeckerboards"),
                &["checkerboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("datatstructures"),
                &["datastructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("espaced"), &["escaped"]),
            (dictgen::InsensitiveStr::Ascii("loclaity"), &["locality"]),
            (dictgen::InsensitiveStr::Ascii("tbe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("pessimestic"),
                &["pessimistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("besics"), &["basics"]),
            (dictgen::InsensitiveStr::Ascii("tatgerts"), &["targets"]),
            (
                dictgen::InsensitiveStr::Ascii("envelopping"),
                &["enveloping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corruptable"),
                &["corruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registerin"),
                &["registering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparecy"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presice"),
                &["precise", "preside"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimum"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("thansk"), &["thanks"]),
            (
                dictgen::InsensitiveStr::Ascii("affiliction"),
                &["affiliation", "affliction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythesizes"),
                &["synthesizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ampilfiers"),
                &["amplifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auotmobiles"),
                &["automobiles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amphatamine"),
                &["amphetamine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sheninigans"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("formulaes"), &["formulas"]),
            (dictgen::InsensitiveStr::Ascii("varibable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("degrase"), &["degrasse"]),
            (dictgen::InsensitiveStr::Ascii("exlpicit"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("bubbels"), &["bubbles"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respository"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilcation"),
                &["complication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libertarianismo"),
                &["libertarians"],
            ),
            (dictgen::InsensitiveStr::Ascii("torward"), &["toward"]),
            (dictgen::InsensitiveStr::Ascii("multpart"), &["multipart"]),
            (dictgen::InsensitiveStr::Ascii("uypdate"), &["update"]),
            (dictgen::InsensitiveStr::Ascii("swtich"), &["switch"]),
            (
                dictgen::InsensitiveStr::Ascii("infroamtion"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palesitnians"),
                &["palestinians"],
            ),
            (dictgen::InsensitiveStr::Ascii("preshisly"), &["preciously"]),
            (
                dictgen::InsensitiveStr::Ascii("expierience"),
                &["experience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formmatting"),
                &["formatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("matatable"), &["metatable"]),
            (
                dictgen::InsensitiveStr::Ascii("repetedly"),
                &["repeatedly", "reputedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriipt"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("intuitivelly"),
                &["intuitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("nummber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("thiking"), &["thinking"]),
            (
                dictgen::InsensitiveStr::Ascii("substantitve"),
                &["substantive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriciate"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("croppped"), &["cropped"]),
            (dictgen::InsensitiveStr::Ascii("equpment"), &["equipment"]),
            (
                dictgen::InsensitiveStr::Ascii("permuatation"),
                &["permutation"],
            ),
            (dictgen::InsensitiveStr::Ascii("examinate"), &["examine"]),
            (dictgen::InsensitiveStr::Ascii("clarfiy"), &["clarify"]),
            (dictgen::InsensitiveStr::Ascii("hrlped"), &["helped"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetnal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("assmption"), &["assumption"]),
            (
                dictgen::InsensitiveStr::Ascii("intiializers"),
                &["initializers"],
            ),
            (dictgen::InsensitiveStr::Ascii("apperared"), &["appeared"]),
            (
                dictgen::InsensitiveStr::Ascii("assers"),
                &["asserts", "assert"],
            ),
            (dictgen::InsensitiveStr::Ascii("attanding"), &["attending"]),
            (dictgen::InsensitiveStr::Ascii("specialt"), &["specialist"]),
            (
                dictgen::InsensitiveStr::Ascii("foders"),
                &["coders", "folders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stereotypcial"),
                &["stereotypical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internatinally"),
                &["internationally"],
            ),
            (dictgen::InsensitiveStr::Ascii("appatures"), &["apertures"]),
            (dictgen::InsensitiveStr::Ascii("randam"), &["random"]),
            (
                dictgen::InsensitiveStr::Ascii("inbankments"),
                &["embankments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("morrisette"),
                &["morissette"],
            ),
            (dictgen::InsensitiveStr::Ascii("knowned"), &["known"]),
            (dictgen::InsensitiveStr::Ascii("revrieve"), &["retrieve"]),
            (
                dictgen::InsensitiveStr::Ascii("expcetations"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("amacing"), &["amazing"]),
            (dictgen::InsensitiveStr::Ascii("pasesd"), &["passed"]),
            (dictgen::InsensitiveStr::Ascii("notifiy"), &["notify"]),
            (
                dictgen::InsensitiveStr::Ascii("ecspecially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("rigtheous"), &["righteous"]),
            (dictgen::InsensitiveStr::Ascii("closin"), &["closing"]),
            (
                dictgen::InsensitiveStr::Ascii("pidgeons"),
                &["pigeons", "pidgins"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alghorithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enchancements"),
                &["enhancements", "enchantments"],
            ),
            (dictgen::InsensitiveStr::Ascii("wehen"), &["when"]),
            (
                dictgen::InsensitiveStr::Ascii("sumbission"),
                &["submission"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempates"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("recognisin"),
                &["recognising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementator"),
                &["implementer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submitions"),
                &["submissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trespasing"),
                &["trespassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prdpagators"),
                &["propagators"],
            ),
            (dictgen::InsensitiveStr::Ascii("poinits"), &["points"]),
            (
                dictgen::InsensitiveStr::Ascii("archaeolgy"),
                &["archaeology"],
            ),
            (dictgen::InsensitiveStr::Ascii("setteled"), &["settled"]),
            (dictgen::InsensitiveStr::Ascii("extreamily"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("activites"),
                &["activities", "activates"],
            ),
            (dictgen::InsensitiveStr::Ascii("moddel"), &["model"]),
            (dictgen::InsensitiveStr::Ascii("prodce"), &["produce"]),
            (
                dictgen::InsensitiveStr::Ascii("interactevely"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reporsitory"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sceen"),
                &["scene", "seen", "screen", "scheme"],
            ),
            (dictgen::InsensitiveStr::Ascii("arges"), &["args"]),
            (
                dictgen::InsensitiveStr::Ascii("techonologies"),
                &["technologies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adressible"),
                &["addressable"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimphian"), &["nymphean"]),
            (
                dictgen::InsensitiveStr::Ascii("aliginments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momomentarily"),
                &["momentarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partcipated"),
                &["participated"],
            ),
            (dictgen::InsensitiveStr::Ascii("indiactor"), &["indicator"]),
            (dictgen::InsensitiveStr::Ascii("highight"), &["highlight"]),
            (dictgen::InsensitiveStr::Ascii("farses"), &["farces"]),
            (
                dictgen::InsensitiveStr::Ascii("descrimination"),
                &["discrimination"],
            ),
            (dictgen::InsensitiveStr::Ascii("diferrent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("carnege"),
                &["carnage", "carnegie"],
            ),
            (dictgen::InsensitiveStr::Ascii("charactre"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("captiol"), &["capitol"]),
            (dictgen::InsensitiveStr::Ascii("condident"), &["confident"]),
            (dictgen::InsensitiveStr::Ascii("shanghi"), &["shanghai"]),
            (dictgen::InsensitiveStr::Ascii("rorate"), &["rotate"]),
            (dictgen::InsensitiveStr::Ascii("collapeses"), &["collapses"]),
            (dictgen::InsensitiveStr::Ascii("gueswork"), &["guesswork"]),
            (dictgen::InsensitiveStr::Ascii("perfered"), &["preferred"]),
            (dictgen::InsensitiveStr::Ascii("sueful"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("simetricly"),
                &["symmetricly"],
            ),
            (dictgen::InsensitiveStr::Ascii("pading"), &["padding"]),
            (dictgen::InsensitiveStr::Ascii("paremters"), &["parameters"]),
            (dictgen::InsensitiveStr::Ascii("cetrainly"), &["certainly"]),
            (
                dictgen::InsensitiveStr::Ascii("commuications"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downloadble"),
                &["downloadable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggeratin"),
                &["exaggeration", "exaggerating"],
            ),
            (dictgen::InsensitiveStr::Ascii("defauts"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("overcomeing"),
                &["overcoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patrionism"),
                &["patriotism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrosponsive"),
                &["unresponsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscciated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appologises"),
                &["apologises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executbale"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("travelodge"), &["traveled"]),
            (dictgen::InsensitiveStr::Ascii("harrassing"), &["harassing"]),
            (dictgen::InsensitiveStr::Ascii("niverse"), &["inverse"]),
            (dictgen::InsensitiveStr::Ascii("furnction"), &["function"]),
            (
                dictgen::InsensitiveStr::Ascii("addaptations"),
                &["adaptations"],
            ),
            (dictgen::InsensitiveStr::Ascii("exaust"), &["exhaust"]),
            (
                dictgen::InsensitiveStr::Ascii("colorpsace"),
                &["colorspace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destinatation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advirtisement"),
                &["advertisement"],
            ),
            (dictgen::InsensitiveStr::Ascii("testsdata"), &["testdata"]),
            (dictgen::InsensitiveStr::Ascii("persuits"), &["pursuits"]),
            (
                dictgen::InsensitiveStr::Ascii("sinnagogs"),
                &["synagogs", "synagogues"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaidation"), &["validation"]),
            (
                dictgen::InsensitiveStr::Ascii("botstrappers"),
                &["bootstrappers"],
            ),
            (dictgen::InsensitiveStr::Ascii("reformes"), &["reforms"]),
            (
                dictgen::InsensitiveStr::Ascii("observible"),
                &["observable"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuppies"), &["occupies"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbarhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("fanserve"), &["fanservice"]),
            (
                dictgen::InsensitiveStr::Ascii("prostitition"),
                &["prostitution"],
            ),
            (dictgen::InsensitiveStr::Ascii("usig"), &["using"]),
            (
                dictgen::InsensitiveStr::Ascii("negotioator"),
                &["negotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proejcting"),
                &["projecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distatseful"),
                &["distasteful"],
            ),
            (dictgen::InsensitiveStr::Ascii("includings"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("experiances"),
                &["experiences"],
            ),
            (dictgen::InsensitiveStr::Ascii("defendre"), &["defender"]),
            (
                dictgen::InsensitiveStr::Ascii("undoubtely"),
                &["undoubtedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("licesne"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("definetive"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("americanas"), &["americans"]),
            (dictgen::InsensitiveStr::Ascii("thn"), &["then"]),
            (
                dictgen::InsensitiveStr::Ascii("enchanetment"),
                &["enchantment"],
            ),
            (dictgen::InsensitiveStr::Ascii("observeras"), &["observers"]),
            (
                dictgen::InsensitiveStr::Ascii("underating"),
                &["underrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecessarilly"),
                &["unnecessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteartion"), &["iteration"]),
            (
                dictgen::InsensitiveStr::Ascii("developmetns"),
                &["developments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unidrectional"),
                &["unidirectional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maitenance"),
                &["maintenance"],
            ),
            (dictgen::InsensitiveStr::Ascii("inporting"), &["importing"]),
            (
                dictgen::InsensitiveStr::Ascii("reasonablely"),
                &["reasonably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiantance"),
                &["acquaintances", "acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("maybed"), &["maybe"]),
            (
                dictgen::InsensitiveStr::Ascii("temporarly"),
                &["temporary", "temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("desiginer"), &["designer"]),
            (dictgen::InsensitiveStr::Ascii("prioritity"), &["priority"]),
            (dictgen::InsensitiveStr::Ascii("dymamics"), &["dynamics"]),
            (dictgen::InsensitiveStr::Ascii("prebious"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("predefiend"),
                &["predefined"],
            ),
            (dictgen::InsensitiveStr::Ascii("animaed"), &["animated"]),
            (dictgen::InsensitiveStr::Ascii("encypting"), &["encrypting"]),
            (
                dictgen::InsensitiveStr::Ascii("disssembler"),
                &["disassembler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitors"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("friendzonie"),
                &["friendzoned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schizophrentic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("respose"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("utilsed"), &["utilised"]),
            (
                dictgen::InsensitiveStr::Ascii("sheakspeare"),
                &["shakespeare"],
            ),
            (dictgen::InsensitiveStr::Ascii("extendos"), &["extends"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrificied"),
                &["sacrifice"],
            ),
            (dictgen::InsensitiveStr::Ascii("shortand"), &["shorthand"]),
            (dictgen::InsensitiveStr::Ascii("iclude"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("socioecenomic"),
                &["socioeconomic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congifures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrucion"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("trancates"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("pysched"), &["psyched"]),
            (dictgen::InsensitiveStr::Ascii("bouunding"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("avaoided"), &["avoided"]),
            (
                dictgen::InsensitiveStr::Ascii("vurnerabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("megatorn"), &["megatron"]),
            (
                dictgen::InsensitiveStr::Ascii("possiblilites"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("fysisit"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("enebales"), &["enables"]),
            (dictgen::InsensitiveStr::Ascii("diaster"), &["disaster"]),
            (dictgen::InsensitiveStr::Ascii("moount"), &["mount"]),
            (
                dictgen::InsensitiveStr::Ascii("fullscreeen"),
                &["fullscreen"],
            ),
            (dictgen::InsensitiveStr::Ascii("generalyl"), &["generally"]),
            (dictgen::InsensitiveStr::Ascii("anaylsis"), &["analysis"]),
            (
                dictgen::InsensitiveStr::Ascii("resourcse"),
                &["resources", "resource"],
            ),
            (dictgen::InsensitiveStr::Ascii("enegeries"), &["energies"]),
            (
                dictgen::InsensitiveStr::Ascii("significace"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("cointained"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnererable"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("regresse"), &["regressive"]),
            (
                dictgen::InsensitiveStr::Ascii("brakpoints"),
                &["breakpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represenatives"),
                &["representatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiting"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resepctively"),
                &["respectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("facepam"), &["facepalm"]),
            (dictgen::InsensitiveStr::Ascii("tyhat"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("definitio"), &["definition"]),
            (
                dictgen::InsensitiveStr::Ascii("cuve"),
                &["curve", "cube", "cave"],
            ),
            (dictgen::InsensitiveStr::Ascii("plasticus"), &["plastics"]),
            (dictgen::InsensitiveStr::Ascii("asignmend"), &["assignment"]),
            (dictgen::InsensitiveStr::Ascii("hieararchy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("opimise"), &["optimise"]),
            (dictgen::InsensitiveStr::Ascii("admrials"), &["admirals"]),
            (
                dictgen::InsensitiveStr::Ascii("aotomation"),
                &["automation"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeates"), &["competes"]),
            (
                dictgen::InsensitiveStr::Ascii("aqquaintance"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgarting"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("anticiping"),
                &["anticipating"],
            ),
            (dictgen::InsensitiveStr::Ascii("holocasut"), &["holocaust"]),
            (dictgen::InsensitiveStr::Ascii("diagnonals"), &["diagonals"]),
            (dictgen::InsensitiveStr::Ascii("togling"), &["toggling"]),
            (
                dictgen::InsensitiveStr::Ascii("assimliate"),
                &["assimilate"],
            ),
            (dictgen::InsensitiveStr::Ascii("allowinging"), &["allowing"]),
            (dictgen::InsensitiveStr::Ascii("broadcase"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("outsite"), &["outside"]),
            (dictgen::InsensitiveStr::Ascii("inervals"), &["intervals"]),
            (
                dictgen::InsensitiveStr::Ascii("identtifiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconaissance"),
                &["reconnaissance"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhythem"), &["rhythm"]),
            (
                dictgen::InsensitiveStr::Ascii("volativity"),
                &["volatility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insteallers"),
                &["installers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thrustworthiness"),
                &["trustworthiness"],
            ),
            (dictgen::InsensitiveStr::Ascii("siwtch"), &["switch"]),
            (
                dictgen::InsensitiveStr::Ascii("painkilers"),
                &["painkillers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cigarrete"), &["cigarette"]),
            (
                dictgen::InsensitiveStr::Ascii("eficient"),
                &["efficient", "deficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tansformed"),
                &["transformed"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciffy"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("impreative"),
                &["imperative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfirming"),
                &["confirming"],
            ),
            (dictgen::InsensitiveStr::Ascii("adecuate"), &["adequate"]),
            (dictgen::InsensitiveStr::Ascii("errots"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaions"),
                &["renegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clerification"),
                &["clarification"],
            ),
            (dictgen::InsensitiveStr::Ascii("solates"), &["isolates"]),
            (dictgen::InsensitiveStr::Ascii("comparin"), &["comparing"]),
            (
                dictgen::InsensitiveStr::Ascii("esepcially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("malcious"), &["malicious"]),
            (dictgen::InsensitiveStr::Ascii("chescker"), &["checker"]),
            (dictgen::InsensitiveStr::Ascii("obtrained"), &["obtained"]),
            (
                dictgen::InsensitiveStr::Ascii("asssociated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("passpost"), &["passports"]),
            (
                dictgen::InsensitiveStr::Ascii("appreication"),
                &["appreciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("midifelders"),
                &["midfielders"],
            ),
            (dictgen::InsensitiveStr::Ascii("nimutes"), &["minutes"]),
            (
                dictgen::InsensitiveStr::Ascii("addititional"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("detials"), &["details"]),
            (
                dictgen::InsensitiveStr::Ascii("concentation"),
                &["concentration"],
            ),
            (dictgen::InsensitiveStr::Ascii("vrifies"), &["verifies"]),
            (
                dictgen::InsensitiveStr::Ascii("inializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attributrions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("nationsl"), &["nationals"]),
            (dictgen::InsensitiveStr::Ascii("chenges"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("raoches"), &["roaches"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstrants"),
                &["demonstrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissmantled"),
                &["dismantled"],
            ),
            (dictgen::InsensitiveStr::Ascii("comittish"), &["committish"]),
            (dictgen::InsensitiveStr::Ascii("forsekin"), &["foreskin"]),
            (
                dictgen::InsensitiveStr::Ascii("customizatoins"),
                &["customizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantacising"),
                &["fantasising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procelains"),
                &["porcelains"],
            ),
            (dictgen::InsensitiveStr::Ascii("sexaully"), &["sexually"]),
            (dictgen::InsensitiveStr::Ascii("combind"), &["combined"]),
            (
                dictgen::InsensitiveStr::Ascii("incestigator"),
                &["investigator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flters"),
                &["filters", "falters", "flutters", "flatters", "floaters"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitcions"), &["bitcoins"]),
            (
                dictgen::InsensitiveStr::Ascii("perforamance"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anticipare"),
                &["anticipate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprendice"),
                &["apprentice"],
            ),
            (dictgen::InsensitiveStr::Ascii("trasnsmit"), &["transmit"]),
            (dictgen::InsensitiveStr::Ascii("messqage"), &["message"]),
            (
                dictgen::InsensitiveStr::Ascii("apprended"),
                &["appended", "apprehended"],
            ),
            (dictgen::InsensitiveStr::Ascii("booloader"), &["bootloader"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastruture"),
                &["infrastructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("whenn"), &["when"]),
            (
                dictgen::InsensitiveStr::Ascii("cofirmations"),
                &["confirmations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surrrounded"),
                &["surrounded"],
            ),
            (dictgen::InsensitiveStr::Ascii("geraphics"), &["graphics"]),
            (dictgen::InsensitiveStr::Ascii("scritpts"), &["scripts"]),
            (dictgen::InsensitiveStr::Ascii("trackres"), &["trackers"]),
            (dictgen::InsensitiveStr::Ascii("stirngs"), &["strings"]),
            (
                dictgen::InsensitiveStr::Ascii("justificstion"),
                &["justifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("cmoputer"), &["computer"]),
            (
                dictgen::InsensitiveStr::Ascii("initative"),
                &["initiative", "imitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("documant"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("repubic"), &["republic"]),
            (
                dictgen::InsensitiveStr::Ascii("predessor"),
                &["predecessor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearranents"),
                &["rearrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implmenets"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("benhind"), &["behind"]),
            (dictgen::InsensitiveStr::Ascii("timedlta"), &["timedelta"]),
            (
                dictgen::InsensitiveStr::Ascii("unbelivable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("idicate"), &["indicate"]),
            (
                dictgen::InsensitiveStr::Ascii("hyerarchies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soecialize"),
                &["specialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("fashon"), &["fashion"]),
            (dictgen::InsensitiveStr::Ascii("truely"), &["truly"]),
            (dictgen::InsensitiveStr::Ascii("correcttly"), &["correctly"]),
            (dictgen::InsensitiveStr::Ascii("sequemce"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("guarranteeing"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allacritty"),
                &["alacritty", "alacrity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chescklists"),
                &["checklists"],
            ),
            (dictgen::InsensitiveStr::Ascii("atttack"), &["attack"]),
            (
                dictgen::InsensitiveStr::Ascii("embarrissing"),
                &["embarrassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connectting"),
                &["connecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("bboolean"), &["boolean"]),
            (
                dictgen::InsensitiveStr::Ascii("hopsitality"),
                &["hospitality"],
            ),
            (dictgen::InsensitiveStr::Ascii("tippin"), &["tipping"]),
            (dictgen::InsensitiveStr::Ascii("upladaded"), &["uploaded"]),
            (
                dictgen::InsensitiveStr::Ascii("executabnles"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boundaires"),
                &["boundaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microscophic"),
                &["microscopic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remiander"),
                &["remainder", "reminder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pairocheelly"),
                &["parochially"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntehtic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("invokve"), &["invoke"]),
            (dictgen::InsensitiveStr::Ascii("specificer"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("heurisitc"), &["heuristic"]),
            (
                dictgen::InsensitiveStr::Ascii("pararmeterize"),
                &["parameterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnanged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("utlities"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("espressso"), &["espresso"]),
            (
                dictgen::InsensitiveStr::Ascii("instructiosn"),
                &["instruction", "instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaminent"),
                &["contaminant", "containment"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaluete"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("authencticator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("mandetory"), &["mandatory"]),
            (dictgen::InsensitiveStr::Ascii("annyoing"), &["annoying"]),
            (
                dictgen::InsensitiveStr::Ascii("seperetors"),
                &["separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("describbed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("exoskelaton"),
                &["exoskeleton"],
            ),
            (dictgen::InsensitiveStr::Ascii("updatees"), &["updates"]),
            (
                dictgen::InsensitiveStr::Ascii("cration"),
                &["creation", "nation", "ration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dircetions"),
                &["directions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconstent"),
                &["inconsistent", "inconstant"],
            ),
            (dictgen::InsensitiveStr::Ascii("sapcebar"), &["spacebar"]),
            (dictgen::InsensitiveStr::Ascii("silabus"), &["syllabus"]),
            (
                dictgen::InsensitiveStr::Ascii("egenralise"),
                &["generalise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("christimas"),
                &["christians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessairly"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("believs"), &["believes"]),
            (dictgen::InsensitiveStr::Ascii("fulfull"), &["fulfill"]),
            (
                dictgen::InsensitiveStr::Ascii("ajsencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("gernerator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("overlaping"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imlimented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calclations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominar"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("proprogate"), &["propagate"]),
            (dictgen::InsensitiveStr::Ascii("preferes"), &["prefers"]),
            (dictgen::InsensitiveStr::Ascii("accliamed"), &["acclaimed"]),
            (dictgen::InsensitiveStr::Ascii("rekative"), &["relative"]),
            (dictgen::InsensitiveStr::Ascii("porposing"), &["proposing"]),
            (dictgen::InsensitiveStr::Ascii("coponent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("attackerats"),
                &["attackers"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaccinet"), &["vaccinated"]),
            (dictgen::InsensitiveStr::Ascii("resluted"), &["resulted"]),
            (dictgen::InsensitiveStr::Ascii("covriance"), &["covariance"]),
            (dictgen::InsensitiveStr::Ascii("enitty"), &["entity"]),
            (
                dictgen::InsensitiveStr::Ascii("coefficent"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("secertary"), &["secretary"]),
            (
                dictgen::InsensitiveStr::Ascii("adjustmenet"),
                &["adjustment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aplicabile"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("nuculear"), &["nuclear"]),
            (dictgen::InsensitiveStr::Ascii("aborbed"), &["absorbed"]),
            (
                dictgen::InsensitiveStr::Ascii("phenemenon"),
                &["phenomenon"],
            ),
            (dictgen::InsensitiveStr::Ascii("analogout"), &["analogous"]),
            (
                dictgen::InsensitiveStr::Ascii("registerred"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sems"),
                &["seems", "stems", "semis", "sens", "seams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shartening"),
                &["sharpening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suppoorters"),
                &["supporters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horrenduous"),
                &["horrendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("retailate"), &["retaliate"]),
            (
                dictgen::InsensitiveStr::Ascii("ofisianados"),
                &["aficionados"],
            ),
            (dictgen::InsensitiveStr::Ascii("capaciy"), &["capacity"]),
            (dictgen::InsensitiveStr::Ascii("turly"), &["truly"]),
            (dictgen::InsensitiveStr::Ascii("halper"), &["helper"]),
            (
                dictgen::InsensitiveStr::Ascii("sutisfaction"),
                &["satisfaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accurary"),
                &["accuracy", "actuary"],
            ),
            (dictgen::InsensitiveStr::Ascii("anination"), &["animation"]),
            (
                dictgen::InsensitiveStr::Ascii("persan"),
                &["person", "persian"],
            ),
            (dictgen::InsensitiveStr::Ascii("oursleves"), &["ourselves"]),
            (
                dictgen::InsensitiveStr::Ascii("battleshop"),
                &["battleship"],
            ),
            (dictgen::InsensitiveStr::Ascii("schmea"), &["schema"]),
            (
                dictgen::InsensitiveStr::Ascii("asssumption"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("proprties"), &["properties"]),
            (
                dictgen::InsensitiveStr::Ascii("indivisuals"),
                &["individuals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scholorships"),
                &["scholarships"],
            ),
            (dictgen::InsensitiveStr::Ascii("treatmet"), &["treatments"]),
            (dictgen::InsensitiveStr::Ascii("expess"), &["express"]),
            (dictgen::InsensitiveStr::Ascii("boundry"), &["boundary"]),
            (dictgen::InsensitiveStr::Ascii("climers"), &["climbers"]),
            (
                dictgen::InsensitiveStr::Ascii("esclusively"),
                &["exclusively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realistisch"),
                &["realistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("moderetely"),
                &["moderately"],
            ),
            (dictgen::InsensitiveStr::Ascii("chioce"), &["choice"]),
            (dictgen::InsensitiveStr::Ascii("duplcates"), &["duplicates"]),
            (
                dictgen::InsensitiveStr::Ascii("presitigous"),
                &["prestigious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shtopes"),
                &["stops", "shops"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priotizing"),
                &["prioritizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compition"),
                &["competition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("motovational"),
                &["motivational"],
            ),
            (dictgen::InsensitiveStr::Ascii("orthgonal"), &["orthogonal"]),
            (
                dictgen::InsensitiveStr::Ascii("autoamtically"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mezmorise"), &["mesmerise"]),
            (
                dictgen::InsensitiveStr::Ascii("attribiution"),
                &["attribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("debbugs"),
                &["debugs", "bedbugs"],
            ),
            (dictgen::InsensitiveStr::Ascii("armagedon"), &["armageddon"]),
            (
                dictgen::InsensitiveStr::Ascii("sleections"),
                &["selections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviomental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("curernt"), &["current"]),
            (
                dictgen::InsensitiveStr::Ascii("subclasssing"),
                &["subclassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommeded"),
                &["recommended"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("materialsm"),
                &["materialism"],
            ),
            (dictgen::InsensitiveStr::Ascii("reappeares"), &["reappears"]),
            (dictgen::InsensitiveStr::Ascii("subtltey"), &["subtlety"]),
            (
                dictgen::InsensitiveStr::Ascii("alternativs"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtrates"), &["substrates"]),
            (dictgen::InsensitiveStr::Ascii("lingerine"), &["lingerie"]),
            (dictgen::InsensitiveStr::Ascii("cauesd"), &["caused"]),
            (
                dictgen::InsensitiveStr::Ascii("representible"),
                &["representable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compulers"),
                &["compilers", "computers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsucessflly"),
                &["unsuccessfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("addapter"), &["adapter"]),
            (dictgen::InsensitiveStr::Ascii("intesnity"), &["intensity"]),
            (
                dictgen::InsensitiveStr::Ascii("algoorithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ejaculaion"),
                &["ejaculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cantacted"), &["contacted"]),
            (
                dictgen::InsensitiveStr::Ascii("methdod"),
                &["method", "methods"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpeended"), &["depended"]),
            (
                dictgen::InsensitiveStr::Ascii("investiagtions"),
                &["investigations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrputing"),
                &["interrupting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfermences"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ossociation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pasing"),
                &["passing", "posing", "parsing"],
            ),
            (dictgen::InsensitiveStr::Ascii("conffigure"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("ilkley"), &["likely"]),
            (
                dictgen::InsensitiveStr::Ascii("cateogories"),
                &["categories"],
            ),
            (dictgen::InsensitiveStr::Ascii("caried"), &["carried"]),
            (dictgen::InsensitiveStr::Ascii("ubunutu"), &["ubuntu"]),
            (
                dictgen::InsensitiveStr::Ascii("eviserating"),
                &["eviscerating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genererators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplet"), &["simplest"]),
            (dictgen::InsensitiveStr::Ascii("cutomize"), &["customize"]),
            (dictgen::InsensitiveStr::Ascii("devleoper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("navitvely"), &["natively"]),
            (dictgen::InsensitiveStr::Ascii("weeken"), &["weekend"]),
            (
                dictgen::InsensitiveStr::Ascii("documentstions"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiatives"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("aproches"), &["approaches"]),
            (
                dictgen::InsensitiveStr::Ascii("prefents"),
                &["prevents", "presents", "prefects"],
            ),
            (dictgen::InsensitiveStr::Ascii("assoiated"), &["associated"]),
            (dictgen::InsensitiveStr::Ascii("brigthens"), &["brightens"]),
            (
                dictgen::InsensitiveStr::Ascii("cathegorically"),
                &["categorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condmenation"),
                &["condemnation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sopce"), &["scope"]),
            (
                dictgen::InsensitiveStr::Ascii("vaccinatie"),
                &["vaccinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transivers"),
                &["transceivers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aysnchronous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufactores"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementationen"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("conainers"), &["containers"]),
            (dictgen::InsensitiveStr::Ascii("languags"), &["languages"]),
            (dictgen::InsensitiveStr::Ascii("convenin"), &["convening"]),
            (
                dictgen::InsensitiveStr::Ascii("unssuccesful"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buttoms"),
                &["buttons", "bottom", "bottoms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ergnomically"),
                &["ergonomically"],
            ),
            (dictgen::InsensitiveStr::Ascii("crutial"), &["crucial"]),
            (
                dictgen::InsensitiveStr::Ascii("initiativs"),
                &["initiatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconeection"),
                &["reconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randomisin"),
                &["randomising"],
            ),
            (dictgen::InsensitiveStr::Ascii("torpdeo"), &["torpedo"]),
            (
                dictgen::InsensitiveStr::Ascii("unspefiifeid"),
                &["unspecified"],
            ),
            (dictgen::InsensitiveStr::Ascii("itheirs"), &["theirs"]),
            (
                dictgen::InsensitiveStr::Ascii("posssibility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interviewie"),
                &["interviewer"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagoe"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("avriant"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitratrily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("processess"), &["processes"]),
            (
                dictgen::InsensitiveStr::Ascii("preint"),
                &["print", "preprint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consructed"),
                &["constructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("redudant"), &["redundant"]),
            (dictgen::InsensitiveStr::Ascii("binanaries"), &["binaries"]),
            (
                dictgen::InsensitiveStr::Ascii("elctromagnetic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encyptions"),
                &["encryptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("motoroal"), &["motorola"]),
            (dictgen::InsensitiveStr::Ascii("dortumnd"), &["dortmund"]),
            (
                dictgen::InsensitiveStr::Ascii("craces"),
                &["crates", "graces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("natvigation"),
                &["navigation"],
            ),
            (dictgen::InsensitiveStr::Ascii("tangeant"), &["tangent"]),
            (
                dictgen::InsensitiveStr::Ascii("serounding"),
                &["surrounding"],
            ),
            (dictgen::InsensitiveStr::Ascii("finacee"), &["fiancee"]),
            (dictgen::InsensitiveStr::Ascii("telegrpah"), &["telegraph"]),
            (
                dictgen::InsensitiveStr::Ascii("resignating"),
                &["resignation"],
            ),
            (dictgen::InsensitiveStr::Ascii("contetns"), &["contents"]),
            (dictgen::InsensitiveStr::Ascii("varriant"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("bullerproof"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conferenze"),
                &["conference"],
            ),
            (dictgen::InsensitiveStr::Ascii("abiltiy"), &["ability"]),
            (
                dictgen::InsensitiveStr::Ascii("optimiziations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cateogrized"),
                &["categorized"],
            ),
            (dictgen::InsensitiveStr::Ascii("loopp"), &["loop"]),
            (dictgen::InsensitiveStr::Ascii("extream"), &["extreme"]),
            (dictgen::InsensitiveStr::Ascii("asnwer"), &["answer"]),
            (
                dictgen::InsensitiveStr::Ascii("alteratives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("aplikays"), &["appliqués"]),
            (
                dictgen::InsensitiveStr::Ascii("copengahen"),
                &["copenhagen"],
            ),
            (dictgen::InsensitiveStr::Ascii("closeest"), &["closest"]),
            (dictgen::InsensitiveStr::Ascii("brusied"), &["bruised"]),
            (
                dictgen::InsensitiveStr::Ascii("instituttet"),
                &["institute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("democrazies"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scriopttype"),
                &["scripttype"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intolorent"),
                &["intolerant"],
            ),
            (dictgen::InsensitiveStr::Ascii("calroies"), &["calories"]),
            (dictgen::InsensitiveStr::Ascii("cheeckbox"), &["checkbox"]),
            (dictgen::InsensitiveStr::Ascii("aci"), &["acpi"]),
            (dictgen::InsensitiveStr::Ascii("smove"), &["move"]),
            (dictgen::InsensitiveStr::Ascii("unhapppy"), &["unhappy"]),
            (dictgen::InsensitiveStr::Ascii("dumptser"), &["dumpster"]),
            (
                dictgen::InsensitiveStr::Ascii("parametic"),
                &["parametric", "paramedic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introversy"),
                &["introverts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("representaion"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("examied"), &["examined"]),
            (dictgen::InsensitiveStr::Ascii("visulizes"), &["visualizes"]),
            (dictgen::InsensitiveStr::Ascii("diffrerent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("reglster"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("orignal"), &["original"]),
            (
                dictgen::InsensitiveStr::Ascii("cringyworthy"),
                &["cringeworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("awknowledges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootlaoders"),
                &["bootloaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asssignment"),
                &["assignment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comunicate"),
                &["communicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ssoaiating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("buiildings"), &["buildings"]),
            (dictgen::InsensitiveStr::Ascii("absould"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("addapted"), &["adapted"]),
            (
                dictgen::InsensitiveStr::Ascii("compostiion"),
                &["compositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constributing"),
                &["contributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etablishd"),
                &["established"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbunded"),
                &["unbundled", "unbounded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indentities"),
                &["identities"],
            ),
            (dictgen::InsensitiveStr::Ascii("abotu"), &["about"]),
            (dictgen::InsensitiveStr::Ascii("preists"), &["priests"]),
            (
                dictgen::InsensitiveStr::Ascii("experimantations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diferently"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progresso"),
                &["progression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upstreemed"),
                &["upstreamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rapresentation"),
                &["representation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsupresses"),
                &["unsuppresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighborhhood"),
                &["neighborhood"],
            ),
            (dictgen::InsensitiveStr::Ascii("gurantees"), &["guarantees"]),
            (dictgen::InsensitiveStr::Ascii("volatily"), &["volatility"]),
            (
                dictgen::InsensitiveStr::Ascii("appeciation"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("assult"), &["assault"]),
            (dictgen::InsensitiveStr::Ascii("leace"), &["leave"]),
            (
                dictgen::InsensitiveStr::Ascii("indendation"),
                &["indentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creenshots"),
                &["screenshots"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquitted"), &["acquitted"]),
            (dictgen::InsensitiveStr::Ascii("specic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("kighboshes"), &["kiboshes"]),
            (
                dictgen::InsensitiveStr::Ascii("bacome"),
                &["become", "became"],
            ),
            (dictgen::InsensitiveStr::Ascii("ungly"), &["ugly"]),
            (dictgen::InsensitiveStr::Ascii("impilcit"), &["implicit"]),
            (dictgen::InsensitiveStr::Ascii("guarnated"), &["guaranteed"]),
            (dictgen::InsensitiveStr::Ascii("arrauys"), &["arrays"]),
            (
                dictgen::InsensitiveStr::Ascii("anticiapted"),
                &["anticipated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangages"), &["manages"]),
            (
                dictgen::InsensitiveStr::Ascii("cannibalsim"),
                &["cannibalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("buddhit"), &["buddhist"]),
            (dictgen::InsensitiveStr::Ascii("repeates"), &["repeats"]),
            (dictgen::InsensitiveStr::Ascii("recives"), &["receives"]),
            (
                dictgen::InsensitiveStr::Ascii("architecurally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("availalbe"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("mimimization"),
                &["minimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optomizations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frientship"),
                &["friendship"],
            ),
            (dictgen::InsensitiveStr::Ascii("enabing"), &["enabling"]),
            (
                dictgen::InsensitiveStr::Ascii("currect"),
                &["correct", "current"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enviromnent"),
                &["environments", "environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("terriroty"), &["territory"]),
            (
                dictgen::InsensitiveStr::Ascii("immedietely"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assessmants"),
                &["assessments"],
            ),
            (dictgen::InsensitiveStr::Ascii("retored"), &["restored"]),
            (dictgen::InsensitiveStr::Ascii("requier"), &["require"]),
            (dictgen::InsensitiveStr::Ascii("queueus"), &["queues"]),
            (
                dictgen::InsensitiveStr::Ascii("dissociatin"),
                &["dissociating", "dissociation"],
            ),
            (dictgen::InsensitiveStr::Ascii("calidator"), &["validator"]),
            (dictgen::InsensitiveStr::Ascii("insided"), &["inside"]),
            (
                dictgen::InsensitiveStr::Ascii("stimulantes"),
                &["stimulants"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("delimitor"), &["delimiter"]),
            (dictgen::InsensitiveStr::Ascii("ethicallly"), &["ethically"]),
            (
                dictgen::InsensitiveStr::Ascii("interfacts"),
                &["interacts", "interfaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("algoritm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("sentimenal"),
                &["sentimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invertions"),
                &["inventions", "inversions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instiating"),
                &["instantiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emphesizing"),
                &["emphasizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("performent"),
                &["performant"],
            ),
            (dictgen::InsensitiveStr::Ascii("encording"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("assigend"), &["assigned"]),
            (dictgen::InsensitiveStr::Ascii("ressits"), &["resists"]),
            (dictgen::InsensitiveStr::Ascii("balck"), &["black", "balk"]),
            (dictgen::InsensitiveStr::Ascii("bouy"), &["buoy"]),
            (dictgen::InsensitiveStr::Ascii("crasheed"), &["crashed"]),
            (
                dictgen::InsensitiveStr::Ascii("progrmmers"),
                &["programmers"],
            ),
            (dictgen::InsensitiveStr::Ascii("raptros"), &["raptors"]),
            (
                dictgen::InsensitiveStr::Ascii("alorithmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirriculum"),
                &["curriculum"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alleviatin"),
                &["alleviating", "alleviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signifanct"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quaterback"),
                &["quarterback"],
            ),
            (dictgen::InsensitiveStr::Ascii("alcholic"), &["alcoholic"]),
            (dictgen::InsensitiveStr::Ascii("caluses"), &["clauses"]),
            (
                dictgen::InsensitiveStr::Ascii("equivilants"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("sampes"), &["samples"]),
            (dictgen::InsensitiveStr::Ascii("monstly"), &["mostly"]),
            (
                dictgen::InsensitiveStr::Ascii("bootlenecks"),
                &["bottlenecks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perferrance"),
                &["preference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificier"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuprisingly"),
                &["unsurprisingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("apauls"), &["appalls"]),
            (
                dictgen::InsensitiveStr::Ascii("privilegde"),
                &["privileges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schizoprhenia"),
                &["schizophrenia"],
            ),
            (dictgen::InsensitiveStr::Ascii("varing"), &["varying"]),
            (dictgen::InsensitiveStr::Ascii("sipplies"), &["supplies"]),
            (
                dictgen::InsensitiveStr::Ascii("tradditionally"),
                &["traditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functiionally"),
                &["functionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("descritor"), &["descriptor"]),
            (
                dictgen::InsensitiveStr::Ascii("correespondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidentlaly"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("callenge"), &["challenge"]),
            (dictgen::InsensitiveStr::Ascii("abreviate"), &["abbreviate"]),
            (
                dictgen::InsensitiveStr::Ascii("qualificato"),
                &["qualification"],
            ),
            (dictgen::InsensitiveStr::Ascii("occrured"), &["occurred"]),
            (
                dictgen::InsensitiveStr::Ascii("prescritions"),
                &["prescriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinisitc"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("limitting"), &["limiting"]),
            (
                dictgen::InsensitiveStr::Ascii("conviccion"),
                &["conviction"],
            ),
            (dictgen::InsensitiveStr::Ascii("devotin"), &["devotion"]),
            (
                dictgen::InsensitiveStr::Ascii("despirately"),
                &["desperately"],
            ),
            (dictgen::InsensitiveStr::Ascii("intenisty"), &["intensity"]),
            (dictgen::InsensitiveStr::Ascii("classsify"), &["classify"]),
            (dictgen::InsensitiveStr::Ascii("enlgish"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("dollers"), &["dollars"]),
            (dictgen::InsensitiveStr::Ascii("thrshold"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("misssing"), &["missing"]),
            (
                dictgen::InsensitiveStr::Ascii("paricipate"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("fashism"), &["fascism"]),
            (
                dictgen::InsensitiveStr::Ascii("negitiating"),
                &["negotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("favilitated"),
                &["facilitated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("livin"),
                &["living", "liven", "livid"],
            ),
            (dictgen::InsensitiveStr::Ascii("handleing"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("pointes"),
                &["pointers", "points"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environnement"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimiztaion"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scenegraaph"),
                &["scenegraph"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomanal"),
                &["phenomenal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necesserily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operarions"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abtracting"),
                &["abstracting"],
            ),
            (dictgen::InsensitiveStr::Ascii("excludeds"), &["excludes"]),
            (dictgen::InsensitiveStr::Ascii("popolates"), &["populates"]),
            (
                dictgen::InsensitiveStr::Ascii("cenventions"),
                &["conventions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phasepsace"),
                &["phasespace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comforms"),
                &["conforms", "comforts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdenencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("unqiuely"), &["uniquely"]),
            (dictgen::InsensitiveStr::Ascii("tracsode"), &["transcode"]),
            (dictgen::InsensitiveStr::Ascii("itheir"), &["their"]),
            (
                dictgen::InsensitiveStr::Ascii("deutchsland"),
                &["deutschland"],
            ),
            (dictgen::InsensitiveStr::Ascii("chroline"), &["chlorine"]),
            (dictgen::InsensitiveStr::Ascii("natuilus"), &["nautilus"]),
            (dictgen::InsensitiveStr::Ascii("parshaly"), &["partially"]),
            (dictgen::InsensitiveStr::Ascii("extrator"), &["extractor"]),
            (dictgen::InsensitiveStr::Ascii("weppon"), &["weapon"]),
            (
                dictgen::InsensitiveStr::Ascii("indivudual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("commandbox"), &["commando"]),
            (
                dictgen::InsensitiveStr::Ascii("hoemopathy"),
                &["homeopathy"],
            ),
            (dictgen::InsensitiveStr::Ascii("poninting"), &["pointing"]),
            (dictgen::InsensitiveStr::Ascii("idelogy"), &["ideology"]),
            (
                dictgen::InsensitiveStr::Ascii("pasrer"),
                &["parser", "passer"],
            ),
            (dictgen::InsensitiveStr::Ascii("noralized"), &["normalized"]),
            (
                dictgen::InsensitiveStr::Ascii("quanified"),
                &["quantified", "qualified"],
            ),
            (dictgen::InsensitiveStr::Ascii("allign"), &["align"]),
            (
                dictgen::InsensitiveStr::Ascii("integerations"),
                &["integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identificable"),
                &["identifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architexture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respriatory"),
                &["respiratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("tresurer"), &["treasurer"]),
            (
                dictgen::InsensitiveStr::Ascii("genralization"),
                &["generalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permuatations"),
                &["permutations"],
            ),
            (dictgen::InsensitiveStr::Ascii("einstance"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("groosum"), &["gruesome"]),
            (dictgen::InsensitiveStr::Ascii("itterating"), &["iterating"]),
            (
                dictgen::InsensitiveStr::Ascii("statictically"),
                &["statistically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialses"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("dieletric"), &["dielectric"]),
            (dictgen::InsensitiveStr::Ascii("deatail"), &["detail"]),
            (
                dictgen::InsensitiveStr::Ascii("renegin"),
                &["reneging", "rebegin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirearcical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("manifeso"), &["manifesto"]),
            (
                dictgen::InsensitiveStr::Ascii("interferin"),
                &["interfering", "interferon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjsencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("nofified"), &["notified"]),
            (
                dictgen::InsensitiveStr::Ascii("reversable"),
                &["reversible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("billingualism"),
                &["bilingualism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credendials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusomises"), &["customises"]),
            (dictgen::InsensitiveStr::Ascii("terminials"), &["terminals"]),
            (dictgen::InsensitiveStr::Ascii("ingenius"), &["ingenious"]),
            (
                dictgen::InsensitiveStr::Ascii("catgorizing"),
                &["categorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("planeswaker"),
                &["planeswalker"],
            ),
            (dictgen::InsensitiveStr::Ascii("perservere"), &["persevere"]),
            (dictgen::InsensitiveStr::Ascii("continuos"), &["continuous"]),
            (
                dictgen::InsensitiveStr::Ascii("incremeneted"),
                &["incremented"],
            ),
            (dictgen::InsensitiveStr::Ascii("cambirdge"), &["cambridge"]),
            (dictgen::InsensitiveStr::Ascii("flor"), &["floor", "flow"]),
            (dictgen::InsensitiveStr::Ascii("elemet"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("consitutents"),
                &["constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("discoure"), &["discourse"]),
            (
                dictgen::InsensitiveStr::Ascii("vigeur"),
                &["vigueur", "vigour"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsticles"), &["obstacles"]),
            (
                dictgen::InsensitiveStr::Ascii("electromagnetc"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algohmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("camoflague"),
                &["camouflage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaprisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("definmed"), &["defined"]),
            (dictgen::InsensitiveStr::Ascii("teamplates"), &["templates"]),
            (
                dictgen::InsensitiveStr::Ascii("declartiuve"),
                &["declarative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchoniser"),
                &["synchroniser"],
            ),
            (dictgen::InsensitiveStr::Ascii("monetizare"), &["monetize"]),
            (dictgen::InsensitiveStr::Ascii("regenere"), &["regenerate"]),
            (
                dictgen::InsensitiveStr::Ascii("conditoinals"),
                &["conditionals"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnges"), &["changes"]),
            (dictgen::InsensitiveStr::Ascii("entitee"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("thannkful"), &["thankful"]),
            (dictgen::InsensitiveStr::Ascii("qurorum"), &["quorum"]),
            (dictgen::InsensitiveStr::Ascii("langht"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("afteward"), &["afterward"]),
            (dictgen::InsensitiveStr::Ascii("guarented"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpted"),
                &["misinterpret"],
            ),
            (dictgen::InsensitiveStr::Ascii("plugion"), &["plugin"]),
            (dictgen::InsensitiveStr::Ascii("plcaes"), &["places"]),
            (dictgen::InsensitiveStr::Ascii("casaulity"), &["causality"]),
            (
                dictgen::InsensitiveStr::Ascii("postiioning"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicily"), &["explicitly"]),
            (dictgen::InsensitiveStr::Ascii("sequensed"), &["sequenced"]),
            (dictgen::InsensitiveStr::Ascii("provideded"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("scafolded"), &["scaffolded"]),
            (dictgen::InsensitiveStr::Ascii("euhporic"), &["euphoric"]),
            (dictgen::InsensitiveStr::Ascii("tetst"), &["test"]),
            (dictgen::InsensitiveStr::Ascii("huricaine"), &["hurricane"]),
            (
                dictgen::InsensitiveStr::Ascii("modfifiable"),
                &["modifiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("pojrect"), &["project"]),
            (dictgen::InsensitiveStr::Ascii("tangenet"), &["tangent"]),
            (dictgen::InsensitiveStr::Ascii("structue"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("origninality"),
                &["originality"],
            ),
            (dictgen::InsensitiveStr::Ascii("revloves"), &["revolves"]),
            (dictgen::InsensitiveStr::Ascii("ententries"), &["entries"]),
            (
                dictgen::InsensitiveStr::Ascii("indiens"),
                &["indians", "endians"],
            ),
            (dictgen::InsensitiveStr::Ascii("triky"), &["tricky"]),
            (dictgen::InsensitiveStr::Ascii("tagets"), &["targets"]),
            (
                dictgen::InsensitiveStr::Ascii("distrution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("campuss"), &["campuses"]),
            (
                dictgen::InsensitiveStr::Ascii("forthcominng"),
                &["forthcoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transorming"),
                &["transforming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertaintity"),
                &["uncertainty"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanlify"),
                &["qualify", "quantify"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signess"),
                &["signedness", "signs"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterpret"), &["interpret"]),
            (dictgen::InsensitiveStr::Ascii("multilpes"), &["multiples"]),
            (
                dictgen::InsensitiveStr::Ascii("configursation"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititalizing"),
                &["initializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infinitelly"),
                &["infinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearates"), &["operates"]),
            (dictgen::InsensitiveStr::Ascii("vertgio"), &["vertigo"]),
            (dictgen::InsensitiveStr::Ascii("termninal"), &["terminal"]),
            (dictgen::InsensitiveStr::Ascii("auot"), &["auto"]),
            (dictgen::InsensitiveStr::Ascii("varient"), &["variant"]),
            (
                dictgen::InsensitiveStr::Ascii("assocciation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predesposed"),
                &["predisposed"],
            ),
            (dictgen::InsensitiveStr::Ascii("retuns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("attahcment"),
                &["attachment"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifyed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("implimentions"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("grappel"), &["grapple"]),
            (dictgen::InsensitiveStr::Ascii("laiter"), &["later"]),
            (dictgen::InsensitiveStr::Ascii("additiion"), &["addition"]),
            (
                dictgen::InsensitiveStr::Ascii("interations"),
                &["iterations", "interactions", "integrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depreacation"),
                &["deprecation"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiable"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("whever"), &["wherever"]),
            (dictgen::InsensitiveStr::Ascii("taxnomy"), &["taxonomy"]),
            (dictgen::InsensitiveStr::Ascii("situaion"), &["situation"]),
            (
                dictgen::InsensitiveStr::Ascii("schyzophrenic"),
                &["schizophrenic"],
            ),
            (dictgen::InsensitiveStr::Ascii("asertion"), &["assertion"]),
            (
                dictgen::InsensitiveStr::Ascii("reigstered"),
                &["registered"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpyo"), &["typo"]),
            (dictgen::InsensitiveStr::Ascii("becaause"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("contiuous"), &["continuous"]),
            (dictgen::InsensitiveStr::Ascii("repsonse"), &["response"]),
            (dictgen::InsensitiveStr::Ascii("procucts"), &["products"]),
            (dictgen::InsensitiveStr::Ascii("coverter"), &["converter"]),
            (
                dictgen::InsensitiveStr::Ascii("possessivize"),
                &["possessive"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowwers"), &["followers"]),
            (
                dictgen::InsensitiveStr::Ascii("connectiviy"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("genrates"), &["generates"]),
            (
                dictgen::InsensitiveStr::Ascii("correspdonences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("tortise"), &["tortoise"]),
            (dictgen::InsensitiveStr::Ascii("fashons"), &["fashions"]),
            (dictgen::InsensitiveStr::Ascii("leagel"), &["legal"]),
            (
                dictgen::InsensitiveStr::Ascii("customsiable"),
                &["customisable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unabailable"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("compieler"), &["compiler"]),
            (dictgen::InsensitiveStr::Ascii("instensive"), &["intensive"]),
            (dictgen::InsensitiveStr::Ascii("trudgin"), &["trudging"]),
            (dictgen::InsensitiveStr::Ascii("nesessary"), &["necessary"]),
            (dictgen::InsensitiveStr::Ascii("spurrious"), &["spurious"]),
            (
                dictgen::InsensitiveStr::Ascii("contineously"),
                &["continuously"],
            ),
            (dictgen::InsensitiveStr::Ascii("folowwin"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibilty"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("mixted"), &["mixed"]),
            (
                dictgen::InsensitiveStr::Ascii("certificstion"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entartainment"),
                &["entertainment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certifificates"),
                &["certificates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selecctable"),
                &["selectable"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfer"), &["prefer"]),
            (dictgen::InsensitiveStr::Ascii("scrubed"), &["scrubbed"]),
            (
                dictgen::InsensitiveStr::Ascii("appriciating"),
                &["appreciating"],
            ),
            (dictgen::InsensitiveStr::Ascii("marrage"), &["marriage"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotation"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functinoal"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthuseasm"),
                &["enthusiasm"],
            ),
            (dictgen::InsensitiveStr::Ascii("umoutn"), &["umount"]),
            (dictgen::InsensitiveStr::Ascii("pionter"), &["pointer"]),
            (
                dictgen::InsensitiveStr::Ascii("conifguration"),
                &["configurations", "configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("fieid"), &["field"]),
            (
                dictgen::InsensitiveStr::Ascii("hypentated"),
                &["hyphenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beliveable"),
                &["believable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourghs"),
                &["neighbours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dameon"),
                &["daemon", "demon", "damien"],
            ),
            (dictgen::InsensitiveStr::Ascii("signign"), &["signing"]),
            (dictgen::InsensitiveStr::Ascii("athients"), &["atheist"]),
            (dictgen::InsensitiveStr::Ascii("commitd"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("innner"), &["inner"]),
            (
                dictgen::InsensitiveStr::Ascii("canadianese"),
                &["canadians"],
            ),
            (dictgen::InsensitiveStr::Ascii("perormed"), &["performed"]),
            (dictgen::InsensitiveStr::Ascii("divy"), &["divvy"]),
            (
                dictgen::InsensitiveStr::Ascii("excecutable"),
                &["executable"],
            ),
            (dictgen::InsensitiveStr::Ascii("derypted"), &["decrypted"]),
            (dictgen::InsensitiveStr::Ascii("commutin"), &["commuting"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridinate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("pimxaps"), &["pixmaps"]),
            (
                dictgen::InsensitiveStr::Ascii("encorporate"),
                &["incorporate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multiliers"),
                &["multipliers"],
            ),
            (dictgen::InsensitiveStr::Ascii("ncrement"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("mege"), &["merge"]),
            (
                dictgen::InsensitiveStr::Ascii("inititating"),
                &["initiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("analzying"), &["analyzing"]),
            (dictgen::InsensitiveStr::Ascii("rountrip"), &["roundtrip"]),
            (
                dictgen::InsensitiveStr::Ascii("cryptograpic"),
                &["cryptographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("resulst"), &["results"]),
            (dictgen::InsensitiveStr::Ascii("srbg"), &["srgb"]),
            (dictgen::InsensitiveStr::Ascii("strechtes"), &["stretches"]),
            (dictgen::InsensitiveStr::Ascii("decsend"), &["descend"]),
            (dictgen::InsensitiveStr::Ascii("rechable"), &["reachable"]),
            (
                dictgen::InsensitiveStr::Ascii("checpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("bubblews"), &["bubbles"]),
            (dictgen::InsensitiveStr::Ascii("audeince"), &["audience"]),
            (
                dictgen::InsensitiveStr::Ascii("infidility"),
                &["infidelity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotatiate"),
                &["negotiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("legecy"), &["legacy"]),
            (
                dictgen::InsensitiveStr::Ascii("submergered"),
                &["submerged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reord"),
                &["record", "reword"],
            ),
            (dictgen::InsensitiveStr::Ascii("froms"), &["forms"]),
            (dictgen::InsensitiveStr::Ascii("regresos"), &["regress"]),
            (dictgen::InsensitiveStr::Ascii("lincensed"), &["licensed"]),
            (dictgen::InsensitiveStr::Ascii("labling"), &["labeling"]),
            (
                dictgen::InsensitiveStr::Ascii("restricters"),
                &["restricts"],
            ),
            (dictgen::InsensitiveStr::Ascii("mesaureed"), &["measured"]),
            (dictgen::InsensitiveStr::Ascii("porposal"), &["proposal"]),
            (dictgen::InsensitiveStr::Ascii("cannpt"), &["cannot"]),
            (dictgen::InsensitiveStr::Ascii("havew"), &["have"]),
            (dictgen::InsensitiveStr::Ascii("acsension"), &["ascension"]),
            (dictgen::InsensitiveStr::Ascii("chrcking"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("lod"), &["load"]),
            (
                dictgen::InsensitiveStr::Ascii("correspondant"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("contac"), &["contact"]),
            (dictgen::InsensitiveStr::Ascii("divding"), &["dividing"]),
            (dictgen::InsensitiveStr::Ascii("sequenze"), &["sequence"]),
            (
                dictgen::InsensitiveStr::Ascii("franeworks"),
                &["frameworks"],
            ),
            (dictgen::InsensitiveStr::Ascii("unblcok"), &["unblock"]),
            (dictgen::InsensitiveStr::Ascii("modfel"), &["model"]),
            (dictgen::InsensitiveStr::Ascii("dirrects"), &["directs"]),
            (dictgen::InsensitiveStr::Ascii("annoating"), &["annotating"]),
            (
                dictgen::InsensitiveStr::Ascii("resiliancy"),
                &["resiliency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("activacion"),
                &["activation"],
            ),
            (dictgen::InsensitiveStr::Ascii("channl"), &["channel"]),
            (
                dictgen::InsensitiveStr::Ascii("misrepresentaiton"),
                &["misrepresentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsence"), &["offence"]),
            (dictgen::InsensitiveStr::Ascii("realsim"), &["realism"]),
            (dictgen::InsensitiveStr::Ascii("programa"), &["program"]),
            (dictgen::InsensitiveStr::Ascii("opperation"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("frist"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("diviced"), &["divided"]),
            (dictgen::InsensitiveStr::Ascii("multipleye"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("defently"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("implementationn"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mastrubated"),
                &["masturbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tiple"),
                &["triple", "tuple"],
            ),
            (dictgen::InsensitiveStr::Ascii("dubug"), &["debug"]),
            (dictgen::InsensitiveStr::Ascii("delcares"), &["declares"]),
            (
                dictgen::InsensitiveStr::Ascii("coreespondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("amoutns"), &["amounts"]),
            (dictgen::InsensitiveStr::Ascii("comestics"), &["cosmetics"]),
            (dictgen::InsensitiveStr::Ascii("misstake"), &["mistake"]),
            (dictgen::InsensitiveStr::Ascii("furhter"), &["further"]),
            (dictgen::InsensitiveStr::Ascii("relaoding"), &["reloading"]),
            (
                dictgen::InsensitiveStr::Ascii("extensilbe"),
                &["extensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evaluetions"),
                &["evaluations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicussing"), &["discussing"]),
            (dictgen::InsensitiveStr::Ascii("aleniate"), &["alienate"]),
            (dictgen::InsensitiveStr::Ascii("strig"), &["string"]),
            (dictgen::InsensitiveStr::Ascii("exectuor"), &["executor"]),
            (dictgen::InsensitiveStr::Ascii("verbouse"), &["verbose"]),
            (
                dictgen::InsensitiveStr::Ascii("interesection"),
                &["intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("tripples"), &["triples"]),
            (
                dictgen::InsensitiveStr::Ascii("destintations"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("achieveing"), &["achieving"]),
            (
                dictgen::InsensitiveStr::Ascii("visuabisation"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("encodning"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("aligined"), &["aligned"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronization"),
                &["synchronization"],
            ),
            (dictgen::InsensitiveStr::Ascii("geneology"), &["genealogy"]),
            (dictgen::InsensitiveStr::Ascii("slaugther"), &["slaughter"]),
            (dictgen::InsensitiveStr::Ascii("wtihout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("spesifier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("ballistc"), &["ballistic"]),
            (dictgen::InsensitiveStr::Ascii("obtaines"), &["obtains"]),
            (
                dictgen::InsensitiveStr::Ascii("transpalnt"),
                &["transplants"],
            ),
            (dictgen::InsensitiveStr::Ascii("exepects"), &["expects"]),
            (dictgen::InsensitiveStr::Ascii("raidant"), &["radiant"]),
            (
                dictgen::InsensitiveStr::Ascii("helicotpers"),
                &["helicopters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrecovable"),
                &["unrecoverable"],
            ),
            (dictgen::InsensitiveStr::Ascii("compiant"), &["compliant"]),
            (dictgen::InsensitiveStr::Ascii("creaated"), &["created"]),
            (
                dictgen::InsensitiveStr::Ascii("experiemnt"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeeedds"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("remebers"), &["remembers"]),
            (dictgen::InsensitiveStr::Ascii("nievely"), &["naively"]),
            (
                dictgen::InsensitiveStr::Ascii("seperateely"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paraphrashing"),
                &["paraphrasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphoricial"),
                &["metaphorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trotzkyist"),
                &["trotskyist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seach"),
                &["search", "each", "reach", "teach", "beach"],
            ),
            (dictgen::InsensitiveStr::Ascii("ninj"), &["ninja"]),
            (
                dictgen::InsensitiveStr::Ascii("percautions"),
                &["precautions"],
            ),
            (dictgen::InsensitiveStr::Ascii("libertea"), &["liberate"]),
            (dictgen::InsensitiveStr::Ascii("expepted"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("graineries"), &["granaries"]),
            (dictgen::InsensitiveStr::Ascii("pararm"), &["param"]),
            (dictgen::InsensitiveStr::Ascii("caridgan"), &["cardigan"]),
            (dictgen::InsensitiveStr::Ascii("referede"), &["referee"]),
            (
                dictgen::InsensitiveStr::Ascii("traditionnally"),
                &["traditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encompased"),
                &["encompassed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("publicitan"),
                &["publication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strawbrary"),
                &["strawberry"],
            ),
            (dictgen::InsensitiveStr::Ascii("keybooards"), &["keyboards"]),
            (
                dictgen::InsensitiveStr::Ascii("spectecular"),
                &["spectacular"],
            ),
            (dictgen::InsensitiveStr::Ascii("sancutary"), &["sanctuary"]),
            (dictgen::InsensitiveStr::Ascii("parasties"), &["parasites"]),
            (
                dictgen::InsensitiveStr::Ascii("indentatin"),
                &["indentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("treadet"),
                &["treated", "threaded", "treaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("cllosed"), &["closed"]),
            (dictgen::InsensitiveStr::Ascii("hydrualic"), &["hydraulic"]),
            (
                dictgen::InsensitiveStr::Ascii("musicallity"),
                &["musically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitaions"),
                &["negotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("codos"), &["codes"]),
            (dictgen::InsensitiveStr::Ascii("unfairy"), &["unfairly"]),
            (dictgen::InsensitiveStr::Ascii("hwole"), &["whole"]),
            (dictgen::InsensitiveStr::Ascii("sintac"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("syntetized"),
                &["synthetized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centrifugeable"),
                &["centrifugable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsbily"),
                &["responsibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("stressfull"), &["stressful"]),
            (dictgen::InsensitiveStr::Ascii("godesses"), &["goddesses"]),
            (dictgen::InsensitiveStr::Ascii("confgured"), &["configured"]),
            (dictgen::InsensitiveStr::Ascii("exhibitin"), &["exhibition"]),
            (dictgen::InsensitiveStr::Ascii("collony"), &["colony"]),
            (dictgen::InsensitiveStr::Ascii("bilbical"), &["biblical"]),
            (
                dictgen::InsensitiveStr::Ascii("uppstreamed"),
                &["upstreamed"],
            ),
            (dictgen::InsensitiveStr::Ascii("rovider"), &["provider"]),
            (dictgen::InsensitiveStr::Ascii("truble"), &["trouble"]),
            (dictgen::InsensitiveStr::Ascii("irritato"), &["irritation"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditionnal"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("strorage"), &["storage"]),
            (dictgen::InsensitiveStr::Ascii("coommander"), &["commander"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("explict"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("acivating"), &["activating"]),
            (dictgen::InsensitiveStr::Ascii("greates"), &["greats"]),
            (dictgen::InsensitiveStr::Ascii("genereics"), &["generics"]),
            (
                dictgen::InsensitiveStr::Ascii("partiotism"),
                &["patriotism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reprsenting"),
                &["representing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deliberatly"),
                &["deliberately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comlication"),
                &["complication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitiors"),
                &["competitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replased"),
                &["relapsed", "replaced", "rephased"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qualifieres"),
                &["qualifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cotrasfers"),
                &["cotransfers"],
            ),
            (dictgen::InsensitiveStr::Ascii("cofigures"), &["configures"]),
            (dictgen::InsensitiveStr::Ascii("munbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("algorwwithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumarise"), &["summarise"]),
            (
                dictgen::InsensitiveStr::Ascii("missionarry"),
                &["missionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomrater"), &["formatter"]),
            (dictgen::InsensitiveStr::Ascii("libarry"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("inspektors"),
                &["inspectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializate"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("quersts"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("availebility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irreverisble"),
                &["irreversible"],
            ),
            (dictgen::InsensitiveStr::Ascii("challendge"), &["challenge"]),
            (dictgen::InsensitiveStr::Ascii("darcula"), &["dracula"]),
            (dictgen::InsensitiveStr::Ascii("terminarlo"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("emergenices"),
                &["emergencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("weigthlifting"),
                &["weightlifting"],
            ),
            (dictgen::InsensitiveStr::Ascii("swiming"), &["swimming"]),
            (
                dictgen::InsensitiveStr::Ascii("interent"),
                &["internet", "interest", "inherent", "interment"],
            ),
            (dictgen::InsensitiveStr::Ascii("bejiing"), &["beijing"]),
            (dictgen::InsensitiveStr::Ascii("noo"), &["no"]),
            (dictgen::InsensitiveStr::Ascii("dnymaic"), &["dynamic"]),
            (dictgen::InsensitiveStr::Ascii("relataive"), &["relative"]),
            (
                dictgen::InsensitiveStr::Ascii("symapthizers"),
                &["sympathizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assemblie"),
                &["assemble", "assembly"],
            ),
            (dictgen::InsensitiveStr::Ascii("truncats"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("immitator"), &["imitator"]),
            (dictgen::InsensitiveStr::Ascii("virtuose"), &["virtues"]),
            (
                dictgen::InsensitiveStr::Ascii("contast"),
                &["contacts", "contrast", "contest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kilomteres"),
                &["kilometers"],
            ),
            (dictgen::InsensitiveStr::Ascii("muliticast"), &["multicast"]),
            (dictgen::InsensitiveStr::Ascii("captues"), &["captures"]),
            (
                dictgen::InsensitiveStr::Ascii("arranements"),
                &["arrangements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condictions"),
                &["conditions", "convictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remmebered"),
                &["remembered"],
            ),
            (dictgen::InsensitiveStr::Ascii("curros"), &["cursor"]),
            (
                dictgen::InsensitiveStr::Ascii("formaly"),
                &["formally", "formerly"],
            ),
            (dictgen::InsensitiveStr::Ascii("encoser"), &["encoder"]),
            (
                dictgen::InsensitiveStr::Ascii("techinology"),
                &["technology"],
            ),
            (dictgen::InsensitiveStr::Ascii("spoilerd"), &["spoiled"]),
            (
                dictgen::InsensitiveStr::Ascii("generlized"),
                &["generalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("aranged"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("numbres"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("deviece"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("immidiatly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("claculated"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("belligerante"),
                &["belligerent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calliflower"),
                &["cauliflower"],
            ),
            (dictgen::InsensitiveStr::Ascii("colummn"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("ameliorater"),
                &["ameliorator", "ameliorated", "ameliorates", "ameliorate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fotografic"),
                &["photographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("wron"), &["wrong"]),
            (
                dictgen::InsensitiveStr::Ascii("compatatbility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unaffliated"),
                &["unaffiliated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nenviroment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapaited"),
                &["dissipated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syntaces"),
                &["syntax", "syntaxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alchoolics"),
                &["alcoholics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deverlopment"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequencial"),
                &["sequential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descendends"),
                &["descended"],
            ),
            (dictgen::InsensitiveStr::Ascii("accepteed"), &["accepted"]),
            (dictgen::InsensitiveStr::Ascii("pratices"), &["practices"]),
            (dictgen::InsensitiveStr::Ascii("cycloen"), &["cyclone"]),
            (
                dictgen::InsensitiveStr::Ascii("initiater"),
                &["initiator", "initiated", "initiates", "initiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserilize"),
                &["deserialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("damed"),
                &["damned", "damped", "domed", "gamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nationalisic"),
                &["nationalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strcutural"),
                &["structural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valudations"),
                &["validations", "valuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompreessed"),
                &["uncompressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("morgs"), &["morgues"]),
            (
                dictgen::InsensitiveStr::Ascii("primatively"),
                &["primitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("offseted"), &["offsetted"]),
            (dictgen::InsensitiveStr::Ascii("porve"), &["prove"]),
            (dictgen::InsensitiveStr::Ascii("proffessor"), &["professor"]),
            (
                dictgen::InsensitiveStr::Ascii("distribtuions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("suposedly"), &["supposedly"]),
            (
                dictgen::InsensitiveStr::Ascii("imporvments"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilitis"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggreagators"),
                &["aggregators"],
            ),
            (dictgen::InsensitiveStr::Ascii("repeatly"), &["repeatedly"]),
            (
                dictgen::InsensitiveStr::Ascii("asperge"),
                &["asparagus", "asperger"],
            ),
            (dictgen::InsensitiveStr::Ascii("comapared"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("pssibilities"),
                &["possibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intesections"),
                &["intersections"],
            ),
            (dictgen::InsensitiveStr::Ascii("familier"), &["familiar"]),
            (
                dictgen::InsensitiveStr::Ascii("organisated"),
                &["organised"],
            ),
            (dictgen::InsensitiveStr::Ascii("apprent"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("incliuding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("depdenents"),
                &["dependents"],
            ),
            (dictgen::InsensitiveStr::Ascii("exausting"), &["exhausting"]),
            (dictgen::InsensitiveStr::Ascii("embbeded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("mutiliated"), &["mutilated"]),
            (dictgen::InsensitiveStr::Ascii("updateded"), &["updated"]),
            (dictgen::InsensitiveStr::Ascii("casaulty"), &["casualty"]),
            (dictgen::InsensitiveStr::Ascii("flawessly"), &["flawlessly"]),
            (dictgen::InsensitiveStr::Ascii("secctors"), &["sectors"]),
            (
                dictgen::InsensitiveStr::Ascii("occaisionaly"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("difrent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("createa"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("fomrat"), &["format"]),
            (dictgen::InsensitiveStr::Ascii("utilites"), &["utilities"]),
            (
                dictgen::InsensitiveStr::Ascii("deamonizing"),
                &["daemonizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("altitide"), &["altitude"]),
            (dictgen::InsensitiveStr::Ascii("hexgaons"), &["hexagons"]),
            (dictgen::InsensitiveStr::Ascii("tansport"), &["transport"]),
            (
                dictgen::InsensitiveStr::Ascii("executuables"),
                &["executables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contiaining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unforttunate"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrees"),
                &["industries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suscriptions"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("plesure"), &["pleasure"]),
            (
                dictgen::InsensitiveStr::Ascii("screenshat"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("tobot"), &["robot"]),
            (dictgen::InsensitiveStr::Ascii("specifes"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("boides"), &["bodies"]),
            (dictgen::InsensitiveStr::Ascii("piblisher"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("hotstpots"), &["hotspots"]),
            (dictgen::InsensitiveStr::Ascii("hyundia"), &["hyundai"]),
            (dictgen::InsensitiveStr::Ascii("emphazise"), &["emphasise"]),
            (dictgen::InsensitiveStr::Ascii("sinds"), &["since"]),
            (dictgen::InsensitiveStr::Ascii("froozen"), &["frozen"]),
            (
                dictgen::InsensitiveStr::Ascii("unselecgted"),
                &["unselected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempretures"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availeability"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coordianted"),
                &["coordinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forcus"),
                &["focus", "forces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerbaility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("embrodery"), &["embroidery"]),
            (
                dictgen::InsensitiveStr::Ascii("decompessor"),
                &["decompressor"],
            ),
            (dictgen::InsensitiveStr::Ascii("happaned"), &["happened"]),
            (
                dictgen::InsensitiveStr::Ascii("butifully"),
                &["beautifully"],
            ),
            (dictgen::InsensitiveStr::Ascii("kerenl"), &["kernel"]),
            (
                dictgen::InsensitiveStr::Ascii("undertakeing"),
                &["undertaking"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrptor"), &["descriptor"]),
            (dictgen::InsensitiveStr::Ascii("chekboxes"), &["checkboxes"]),
            (
                dictgen::InsensitiveStr::Ascii("manouevring"),
                &["manoeuvring"],
            ),
            (dictgen::InsensitiveStr::Ascii("armorments"), &["armaments"]),
            (dictgen::InsensitiveStr::Ascii("initaitor"), &["initiator"]),
            (dictgen::InsensitiveStr::Ascii("resturns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("tranversing"),
                &["traversing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparities"),
                &["transparencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inacurrate"),
                &["inaccurate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rosponsive"),
                &["responsive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighberhodds"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("updgrade"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("approproximating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("kernerl"), &["kernel"]),
            (dictgen::InsensitiveStr::Ascii("defalut"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("exameple"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("operstions"),
                &["operations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impossibal"),
                &["impossibly"],
            ),
            (dictgen::InsensitiveStr::Ascii("corridoors"), &["corridors"]),
            (dictgen::InsensitiveStr::Ascii("checckout"), &["checkout"]),
            (dictgen::InsensitiveStr::Ascii("solgers"), &["soldiers"]),
            (
                dictgen::InsensitiveStr::Ascii("functtioned"),
                &["functioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copie"),
                &["copied", "copier", "copies", "copy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repants"),
                &["repaints", "repents"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifiyng"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacutred"),
                &["manufacture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asycronously"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversatism"),
                &["conservatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("workfow"), &["workflow"]),
            (
                dictgen::InsensitiveStr::Ascii("milliesconds"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("twosday"), &["tuesday"]),
            (
                dictgen::InsensitiveStr::Ascii("planetwalker"),
                &["planeswalker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depdendancies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("folwoed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("vegetarien"),
                &["vegetarian"],
            ),
            (dictgen::InsensitiveStr::Ascii("shriks"), &["shrinks"]),
            (dictgen::InsensitiveStr::Ascii("preoperty"), &["property"]),
            (dictgen::InsensitiveStr::Ascii("capusle"), &["capsule"]),
            (dictgen::InsensitiveStr::Ascii("defaullt"), &["default"]),
            (dictgen::InsensitiveStr::Ascii("vehicals"), &["vehicles"]),
            (
                dictgen::InsensitiveStr::Ascii("cousnelors"),
                &["counselors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("greatfull"),
                &["grateful", "gratefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indoctornated"),
                &["indoctrinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coincedence"),
                &["coincidence"],
            ),
            (dictgen::InsensitiveStr::Ascii("checksume"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("arraning"), &["arranging"]),
            (
                dictgen::InsensitiveStr::Ascii("componbents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complites"),
                &["completes", "compiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforeced"), &["enforced"]),
            (
                dictgen::InsensitiveStr::Ascii("overrwritten"),
                &["overwritten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manipualte"),
                &["manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepdendency"),
                &["independency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasmitters"),
                &["transmitters"],
            ),
            (dictgen::InsensitiveStr::Ascii("assihn"), &["assign"]),
            (dictgen::InsensitiveStr::Ascii("socalist"), &["socialist"]),
            (dictgen::InsensitiveStr::Ascii("sepetated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("chroniclin"),
                &["chronicling"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraeters"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinatore"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearted"), &["operated"]),
            (dictgen::InsensitiveStr::Ascii("detectie"), &["detectives"]),
            (
                dictgen::InsensitiveStr::Ascii("greated"),
                &["greater", "grated", "graded"],
            ),
            (dictgen::InsensitiveStr::Ascii("recorods"), &["records"]),
            (dictgen::InsensitiveStr::Ascii("carrear"), &["career"]),
            (
                dictgen::InsensitiveStr::Ascii("purpolsion"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrrectly"), &["correctly"]),
            (
                dictgen::InsensitiveStr::Ascii("retribuiton"),
                &["retribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("columnular"), &["columnar"]),
            (dictgen::InsensitiveStr::Ascii("sumarised"), &["summarised"]),
            (dictgen::InsensitiveStr::Ascii("repected"), &["respected"]),
            (dictgen::InsensitiveStr::Ascii("burgundry"), &["burgundy"]),
            (
                dictgen::InsensitiveStr::Ascii("distibutes"),
                &["distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("paradyse"), &["parades"]),
            (dictgen::InsensitiveStr::Ascii("expedito"), &["expedition"]),
            (
                dictgen::InsensitiveStr::Ascii("paticipation"),
                &["participation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretes"),
                &["interprets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("highshcool"),
                &["highschool"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coinsiding"),
                &["coinciding"],
            ),
            (dictgen::InsensitiveStr::Ascii("profilic"), &["prolific"]),
            (
                dictgen::InsensitiveStr::Ascii("abomanation"),
                &["abomination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforsed"),
                &["endorsed", "enforced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transporing"),
                &["transporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffred"), &["buffered"]),
            (dictgen::InsensitiveStr::Ascii("typeo"), &["typo"]),
            (dictgen::InsensitiveStr::Ascii("somehwat"), &["somewhat"]),
            (dictgen::InsensitiveStr::Ascii("initilally"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("chairmain"), &["chairman"]),
            (dictgen::InsensitiveStr::Ascii("implment"), &["implement"]),
            (
                dictgen::InsensitiveStr::Ascii("workststion"),
                &["workstation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminiated"),
                &["terminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construtors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dysfucntional"),
                &["dysfunctional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colorblend"),
                &["colorblind"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplicants"),
                &["duplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("propmter"), &["prompter"]),
            (dictgen::InsensitiveStr::Ascii("padd"), &["pad", "padded"]),
            (
                dictgen::InsensitiveStr::Ascii("distancef"),
                &["distanced", "distances", "distance"],
            ),
            (dictgen::InsensitiveStr::Ascii("pooint"), &["point"]),
            (
                dictgen::InsensitiveStr::Ascii("operationel"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriuting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("actuallin"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("alleivate"), &["alleviate"]),
            (
                dictgen::InsensitiveStr::Ascii("reaquires"),
                &["requires", "reacquires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initaitives"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("unalligned"), &["unaligned"]),
            (
                dictgen::InsensitiveStr::Ascii("prefenting"),
                &["preventing", "presenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("restrictes"), &["restricts"]),
            (
                dictgen::InsensitiveStr::Ascii("mistatched"),
                &["mismatched"],
            ),
            (dictgen::InsensitiveStr::Ascii("separeated"), &["separated"]),
            (
                dictgen::InsensitiveStr::Ascii("camparators"),
                &["comparators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valentians"),
                &["valentines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asscciates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("veriation"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("conferance"),
                &["conference"],
            ),
            (dictgen::InsensitiveStr::Ascii("extractt"), &["extract"]),
            (
                dictgen::InsensitiveStr::Ascii("bloorborne"),
                &["bloodborne"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistirbuted"),
                &["redistributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsessive"),
                &["successive"],
            ),
            (dictgen::InsensitiveStr::Ascii("leanred"), &["learned"]),
            (
                dictgen::InsensitiveStr::Ascii("prejudicin"),
                &["prejudicing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complicaties"),
                &["complicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("handshak"), &["handshake"]),
            (dictgen::InsensitiveStr::Ascii("aneeled"), &["annealed"]),
            (
                dictgen::InsensitiveStr::Ascii("apprecaitive"),
                &["appreciative"],
            ),
            (dictgen::InsensitiveStr::Ascii("infintie"), &["infinite"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotications"),
                &["renegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("vonfig"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("pastural"), &["pastoral"]),
            (dictgen::InsensitiveStr::Ascii("suposse"), &["suppose"]),
            (
                dictgen::InsensitiveStr::Ascii("firefigther"),
                &["firefighters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsurprizing"),
                &["unsurprising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attributring"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("reversees"), &["reverses"]),
            (dictgen::InsensitiveStr::Ascii("opppsite"), &["opposite"]),
            (
                dictgen::InsensitiveStr::Ascii("communuication"),
                &["communication"],
            ),
            (dictgen::InsensitiveStr::Ascii("skurged"), &["scourged"]),
            (
                dictgen::InsensitiveStr::Ascii("neccsessary"),
                &["necessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monolithisch"),
                &["monolithic"],
            ),
            (dictgen::InsensitiveStr::Ascii("propages"), &["propagates"]),
            (dictgen::InsensitiveStr::Ascii("eveybody"), &["everybody"]),
            (dictgen::InsensitiveStr::Ascii("presance"), &["presence"]),
            (dictgen::InsensitiveStr::Ascii("adventus"), &["adventures"]),
            (
                dictgen::InsensitiveStr::Ascii("backlghting"),
                &["backlighting"],
            ),
            (dictgen::InsensitiveStr::Ascii("detetmine"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("unigned"), &["unsigned"]),
            (dictgen::InsensitiveStr::Ascii("poits"), &["points"]),
            (
                dictgen::InsensitiveStr::Ascii("corrspondences"),
                &["correspondences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsmitter"),
                &["transmitter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("institutionella"),
                &["institutional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponteneously"),
                &["spontaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradititions"),
                &["traditions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nestalgically"),
                &["nostalgically"],
            ),
            (dictgen::InsensitiveStr::Ascii("regulas"), &["regulars"]),
            (dictgen::InsensitiveStr::Ascii("callendars"), &["calendars"]),
            (dictgen::InsensitiveStr::Ascii("chekpoint"), &["checkpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("doubellift"),
                &["doublelift"],
            ),
            (dictgen::InsensitiveStr::Ascii("connectted"), &["connected"]),
            (dictgen::InsensitiveStr::Ascii("peope"), &["people"]),
            (
                dictgen::InsensitiveStr::Ascii("dicationary"),
                &["dictionary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immortallity"),
                &["immortality"],
            ),
            (dictgen::InsensitiveStr::Ascii("ditribute"), &["distribute"]),
            (
                dictgen::InsensitiveStr::Ascii("experminentation"),
                &["experimentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("qusts"), &["quests"]),
            (
                dictgen::InsensitiveStr::Ascii("accumalating"),
                &["accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("michina"), &["michigan"]),
            (dictgen::InsensitiveStr::Ascii("oposed"), &["opposed"]),
            (dictgen::InsensitiveStr::Ascii("dirctory"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("serivces"), &["services"]),
            (
                dictgen::InsensitiveStr::Ascii("saskatchwan"),
                &["saskatchewan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapparance"),
                &["disappearance"],
            ),
            (dictgen::InsensitiveStr::Ascii("blosums"), &["blossoms"]),
            (dictgen::InsensitiveStr::Ascii("additons"), &["additions"]),
            (dictgen::InsensitiveStr::Ascii("categry"), &["category"]),
            (dictgen::InsensitiveStr::Ascii("intrerrupt"), &["interrupt"]),
            (dictgen::InsensitiveStr::Ascii("belyw"), &["below"]),
            (
                dictgen::InsensitiveStr::Ascii("rendayvooed"),
                &["rendezvoused"],
            ),
            (dictgen::InsensitiveStr::Ascii("toogled"), &["toggled"]),
            (dictgen::InsensitiveStr::Ascii("deterimine"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("scientis"), &["scientist"]),
            (dictgen::InsensitiveStr::Ascii("otehrwice"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("buzilla"), &["bugzilla"]),
            (dictgen::InsensitiveStr::Ascii("stdanards"), &["standards"]),
            (dictgen::InsensitiveStr::Ascii("esclated"), &["escalated"]),
            (dictgen::InsensitiveStr::Ascii("litterally"), &["literally"]),
            (
                dictgen::InsensitiveStr::Ascii("anarquists"),
                &["anarchists"],
            ),
            (dictgen::InsensitiveStr::Ascii("quottes"), &["quotes"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("intiailly"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("contructing"),
                &["contracting", "constructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("altheticism"),
                &["athleticism"],
            ),
            (dictgen::InsensitiveStr::Ascii("traners"), &["trainers"]),
            (dictgen::InsensitiveStr::Ascii("numebrs"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("sructures"), &["structures"]),
            (
                dictgen::InsensitiveStr::Ascii("catagorization"),
                &["categorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("succedful"), &["successful"]),
            (
                dictgen::InsensitiveStr::Ascii("photogrpahy"),
                &["photography"],
            ),
            (dictgen::InsensitiveStr::Ascii("jurnied"), &["journeyed"]),
            (dictgen::InsensitiveStr::Ascii("passowrds"), &["passwords"]),
            (
                dictgen::InsensitiveStr::Ascii("unknowingy"),
                &["unknowingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("compresas"), &["compress"]),
            (dictgen::InsensitiveStr::Ascii("represens"), &["represents"]),
            (dictgen::InsensitiveStr::Ascii("remving"), &["removing"]),
            (dictgen::InsensitiveStr::Ascii("experess"), &["express"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotioators"),
                &["renegotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("everset"), &["everest"]),
            (dictgen::InsensitiveStr::Ascii("geting"), &["getting"]),
            (
                dictgen::InsensitiveStr::Ascii("orcestrated"),
                &["orchestrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("detectarlo"), &["detector"]),
            (
                dictgen::InsensitiveStr::Ascii("hipothesis"),
                &["hypothesis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aquisitions"),
                &["acquisitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aknowledgments"),
                &["acknowledgments"],
            ),
            (dictgen::InsensitiveStr::Ascii("multiplie"), &["multiply"]),
            (dictgen::InsensitiveStr::Ascii("defendent"), &["defendant"]),
            (dictgen::InsensitiveStr::Ascii("recipent"), &["recipient"]),
            (dictgen::InsensitiveStr::Ascii("answhare"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("wirefame"), &["wireframe"]),
            (dictgen::InsensitiveStr::Ascii("noticably"), &["noticeably"]),
            (dictgen::InsensitiveStr::Ascii("acccepter"), &["accepter"]),
            (
                dictgen::InsensitiveStr::Ascii("canonicalizations"),
                &["canonicalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("atractive"), &["attractive"]),
            (dictgen::InsensitiveStr::Ascii("withoit"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("xode"), &["code", "xcode"]),
            (dictgen::InsensitiveStr::Ascii("dleay"), &["delay"]),
            (dictgen::InsensitiveStr::Ascii("releafed"), &["relieved"]),
            (
                dictgen::InsensitiveStr::Ascii("experiecing"),
                &["experiencing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecxeptions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("evn"), &["even"]),
            (dictgen::InsensitiveStr::Ascii("coloumn"), &["column"]),
            (
                dictgen::InsensitiveStr::Ascii("adolescense"),
                &["adolescence"],
            ),
            (dictgen::InsensitiveStr::Ascii("duraiton"), &["duration"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerbuility"),
                &["vulnerability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expences"),
                &["expense", "expenses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contenents"),
                &["continents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominatons"),
                &["denominations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tanslations"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sythesised"),
                &["synthesised"],
            ),
            (dictgen::InsensitiveStr::Ascii("miminized"), &["minimized"]),
            (
                dictgen::InsensitiveStr::Ascii("incidentially"),
                &["incidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("decimials"), &["decimals"]),
            (
                dictgen::InsensitiveStr::Ascii("accelearted"),
                &["accelerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissapers"), &["disappears"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialization"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("anming"), &["naming"]),
            (dictgen::InsensitiveStr::Ascii("vincinity"), &["vicinity"]),
            (dictgen::InsensitiveStr::Ascii("rotine"), &["routine"]),
            (dictgen::InsensitiveStr::Ascii("clipboar"), &["clipboard"]),
            (
                dictgen::InsensitiveStr::Ascii("consistancy"),
                &["consistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("lodgin"), &["lodging"]),
            (
                dictgen::InsensitiveStr::Ascii("fitures"),
                &["features", "futures", "figures", "fixtures"],
            ),
            (dictgen::InsensitiveStr::Ascii("qantized"), &["quantized"]),
            (dictgen::InsensitiveStr::Ascii("exlpoding"), &["exploding"]),
            (
                dictgen::InsensitiveStr::Ascii("capatibilities"),
                &["capabilities", "compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charactarize"),
                &["characterize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depencences"),
                &["dependences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("researvations"),
                &["reservations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configires"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("refelected"), &["reflected"]),
            (
                dictgen::InsensitiveStr::Ascii("inequailty"),
                &["inequality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exploititive"),
                &["exploitative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("buildpackges"),
                &["buildpackages"],
            ),
            (dictgen::InsensitiveStr::Ascii("ailenated"), &["alienated"]),
            (dictgen::InsensitiveStr::Ascii("sebastiaan"), &["sebastian"]),
            (dictgen::InsensitiveStr::Ascii("mudules"), &["modules"]),
            (dictgen::InsensitiveStr::Ascii("aletr"), &["alert", "alter"]),
            (
                dictgen::InsensitiveStr::Ascii("convencional"),
                &["conventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("packageid"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("muffinus"), &["muffins"]),
            (
                dictgen::InsensitiveStr::Ascii("volums"),
                &["volume", "volumes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inceases"),
                &["increases", "incenses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orchastration"),
                &["orchestration"],
            ),
            (dictgen::InsensitiveStr::Ascii("completi"), &["complexity"]),
            (dictgen::InsensitiveStr::Ascii("defaulrts"), &["defaults"]),
            (dictgen::InsensitiveStr::Ascii("resisten"), &["resistance"]),
            (dictgen::InsensitiveStr::Ascii("usrespace"), &["userspace"]),
            (dictgen::InsensitiveStr::Ascii("typicaly"), &["typically"]),
            (
                dictgen::InsensitiveStr::Ascii("unsearcahble"),
                &["unsearchable"],
            ),
            (dictgen::InsensitiveStr::Ascii("instsalled"), &["installed"]),
            (
                dictgen::InsensitiveStr::Ascii("reconeccting"),
                &["reconnecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("folloling"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("evolvin"), &["evolving"]),
            (dictgen::InsensitiveStr::Ascii("rsut"), &["rust"]),
            (dictgen::InsensitiveStr::Ascii("blisteres"), &["blisters"]),
            (
                dictgen::InsensitiveStr::Ascii("incomptable"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("ontaining"), &["containing"]),
            (
                dictgen::InsensitiveStr::Ascii("transcripcion"),
                &["transcription"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitry"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("shoutdown"), &["shutdown"]),
            (
                dictgen::InsensitiveStr::Ascii("reisntalling"),
                &["reinstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("initialse"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("rimitives"), &["primitives"]),
            (dictgen::InsensitiveStr::Ascii("furst"), &["first"]),
            (dictgen::InsensitiveStr::Ascii("everyon"), &["everyone"]),
            (dictgen::InsensitiveStr::Ascii("clanand"), &["clannad"]),
            (
                dictgen::InsensitiveStr::Ascii("unnistalling"),
                &["uninstalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("ficticous"), &["fictitious"]),
            (dictgen::InsensitiveStr::Ascii("behaviou"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("reproductible"),
                &["reproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("freze"), &["freeze"]),
            (dictgen::InsensitiveStr::Ascii("crossign"), &["crossing"]),
            (dictgen::InsensitiveStr::Ascii("pipleline"), &["pipeline"]),
            (dictgen::InsensitiveStr::Ascii("exaplains"), &["explains"]),
            (
                dictgen::InsensitiveStr::Ascii("personnality"),
                &["personality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identidiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ephememeral"),
                &["ephemeral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compalation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("astrix"), &["asterisk"]),
            (dictgen::InsensitiveStr::Ascii("virginas"), &["virgins"]),
            (dictgen::InsensitiveStr::Ascii("accracy"), &["accuracy"]),
            (
                dictgen::InsensitiveStr::Ascii("initiialises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("resumt"), &["resume"]),
            (
                dictgen::InsensitiveStr::Ascii("blokchains"),
                &["blockchains"],
            ),
            (dictgen::InsensitiveStr::Ascii("selfs"), &["self"]),
            (
                dictgen::InsensitiveStr::Ascii("underlflowed"),
                &["underflowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscatalogued"),
                &["miscataloged"],
            ),
            (dictgen::InsensitiveStr::Ascii("objec"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("accesssories"),
                &["accessories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temproaily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descricing"),
                &["describing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarming"),
                &["programming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitolises"),
                &["capitalises"],
            ),
            (dictgen::InsensitiveStr::Ascii("misssion"), &["mission"]),
            (dictgen::InsensitiveStr::Ascii("refence"), &["reference"]),
            (dictgen::InsensitiveStr::Ascii("alogorithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("registies"), &["registries"]),
            (
                dictgen::InsensitiveStr::Ascii("contoled"),
                &["controlled", "consoled", "condoled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alignaments"),
                &["alignments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boganveelia"),
                &["bougainvillea"],
            ),
            (dictgen::InsensitiveStr::Ascii("subsystyem"), &["subsystem"]),
            (
                dictgen::InsensitiveStr::Ascii("comunicates"),
                &["communicates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adapatation"),
                &["adaptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("cricling"), &["circling"]),
            (dictgen::InsensitiveStr::Ascii("simpelst"), &["simplest"]),
            (dictgen::InsensitiveStr::Ascii("autistc"), &["autistic"]),
            (dictgen::InsensitiveStr::Ascii("movments"), &["movements"]),
            (dictgen::InsensitiveStr::Ascii("naibors"), &["neighbors"]),
            (
                dictgen::InsensitiveStr::Ascii("reboudning"),
                &["rebounding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demosntrates"),
                &["demonstrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deferentiates"),
                &["differentiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionned"),
                &["functioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armistises"),
                &["armistices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("doocumentaries"),
                &["documentaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaxting"),
                &["contacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("swtiches"), &["switches"]),
            (dictgen::InsensitiveStr::Ascii("suttling"), &["shuttling"]),
            (
                dictgen::InsensitiveStr::Ascii("descussions"),
                &["discussions"],
            ),
            (dictgen::InsensitiveStr::Ascii("consummed"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("diecting"),
                &["directing", "dissecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understandibly"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("cammand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("vernicular"),
                &["vernacular"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchausted"), &["exhausted"]),
            (
                dictgen::InsensitiveStr::Ascii("dictionnary"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("snyergy"), &["synergy"]),
            (
                dictgen::InsensitiveStr::Ascii("aggreation"),
                &["aggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consumbales"),
                &["consumables"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bloodbrone"),
                &["bloodborne"],
            ),
            (dictgen::InsensitiveStr::Ascii("knwoling"), &["knowing"]),
            (
                dictgen::InsensitiveStr::Ascii("sustainible"),
                &["sustainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crowsourced"),
                &["crowdsourced"],
            ),
            (dictgen::InsensitiveStr::Ascii("recipies"), &["recipes"]),
            (dictgen::InsensitiveStr::Ascii("nostolgia"), &["nostalgia"]),
            (dictgen::InsensitiveStr::Ascii("arbitryary"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("sortner"), &["sorter"]),
            (
                dictgen::InsensitiveStr::Ascii("ambidexterously"),
                &["ambidextrously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionaly"),
                &["conditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("statemnts"), &["statements"]),
            (
                dictgen::InsensitiveStr::Ascii("allcoation"),
                &["allocation"],
            ),
            (dictgen::InsensitiveStr::Ascii("intialise"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("progamm"), &["program"]),
            (
                dictgen::InsensitiveStr::Ascii("conifigurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("integrades"),
                &["integrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininistrations"),
                &["administrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expectency"),
                &["expectancy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beanchmarking"),
                &["benchmarking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionalilty"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copmletely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vietnamesse"),
                &["vietnamese"],
            ),
            (dictgen::InsensitiveStr::Ascii("insectes"), &["insects"]),
            (
                dictgen::InsensitiveStr::Ascii("protoganist"),
                &["protagonist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("childrenmrs"),
                &["childrens"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serailised"),
                &["serialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diasability"),
                &["disability"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximumn"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("seemd"), &["seemed"]),
            (
                dictgen::InsensitiveStr::Ascii("toothrbush"),
                &["toothbrush"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sirectional"),
                &["directional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agknowledged"),
                &["acknowledged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execused"),
                &["excused", "executed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitivly"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agricultrual"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imprsioned"),
                &["imprisoned"],
            ),
            (dictgen::InsensitiveStr::Ascii("manipute"), &["manipulate"]),
            (dictgen::InsensitiveStr::Ascii("atttaches"), &["attaches"]),
            (
                dictgen::InsensitiveStr::Ascii("delimition"),
                &["delimitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtlely"),
                &["subtly", "subtlety"],
            ),
            (dictgen::InsensitiveStr::Ascii("toghether"), &["together"]),
            (
                dictgen::InsensitiveStr::Ascii("guaruanteeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("plesently"), &["pleasantly"]),
            (
                dictgen::InsensitiveStr::Ascii("highligted"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("pologon"), &["polygon"]),
            (
                dictgen::InsensitiveStr::Ascii("brazillans"),
                &["brazilians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("natioanlists"),
                &["nationalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aumatically"),
                &["automatically", "traumatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("sponsoered"), &["sponsored"]),
            (dictgen::InsensitiveStr::Ascii("plaintest"), &["plaintext"]),
            (dictgen::InsensitiveStr::Ascii("tehniques"), &["techniques"]),
            (
                dictgen::InsensitiveStr::Ascii("referanced"),
                &["referenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportaion"),
                &["incorporation"],
            ),
            (dictgen::InsensitiveStr::Ascii("valiated"), &["validated"]),
            (dictgen::InsensitiveStr::Ascii("modfitied"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("labrynth"), &["labyrinth"]),
            (
                dictgen::InsensitiveStr::Ascii("fabircated"),
                &["fabricated"],
            ),
            (dictgen::InsensitiveStr::Ascii("coppy"), &["copy", "choppy"]),
            (dictgen::InsensitiveStr::Ascii("nocturen"), &["nocturne"]),
            (
                dictgen::InsensitiveStr::Ascii("permamently"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsing"),
                &["unsigned", "using"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefitlers"),
                &["prefilters"],
            ),
            (dictgen::InsensitiveStr::Ascii("calidated"), &["validated"]),
            (
                dictgen::InsensitiveStr::Ascii("mutilthreading"),
                &["multithreading"],
            ),
            (dictgen::InsensitiveStr::Ascii("clenups"), &["cleanups"]),
            (
                dictgen::InsensitiveStr::Ascii("devleoppers"),
                &["developers"],
            ),
            (dictgen::InsensitiveStr::Ascii("licemse"), &["license"]),
            (
                dictgen::InsensitiveStr::Ascii("alpahbetic"),
                &["alphabetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stationnary"),
                &["stationary"],
            ),
            (dictgen::InsensitiveStr::Ascii("compairs"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("compressers"),
                &["compressors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grilfriend"),
                &["girlfriend"],
            ),
            (dictgen::InsensitiveStr::Ascii("englsih"), &["english"]),
            (dictgen::InsensitiveStr::Ascii("wwith"), &["with"]),
            (
                dictgen::InsensitiveStr::Ascii("invoving"),
                &["involving", "invoking"],
            ),
            (dictgen::InsensitiveStr::Ascii("womens"), &["women"]),
            (
                dictgen::InsensitiveStr::Ascii("capicators"),
                &["capacitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("volital"), &["volatile"]),
            (dictgen::InsensitiveStr::Ascii("langugae"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("connecton"),
                &["connection", "connector"],
            ),
            (dictgen::InsensitiveStr::Ascii("bretheren"), &["brethren"]),
            (
                dictgen::InsensitiveStr::Ascii("piankillers"),
                &["painkillers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccomendations"),
                &["recommendations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensationnal"),
                &["sensational"],
            ),
            (dictgen::InsensitiveStr::Ascii("setment"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("biyous"), &["bayous"]),
            (
                dictgen::InsensitiveStr::Ascii("condfiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maks"),
                &["mask", "masks", "makes", "make"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insperation"),
                &["inspiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("streamus"), &["streams"]),
            (dictgen::InsensitiveStr::Ascii("spacifics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("construnctor"),
                &["constructor"],
            ),
            (dictgen::InsensitiveStr::Ascii("trundel"), &["trundle"]),
            (dictgen::InsensitiveStr::Ascii("ensture"), &["ensure"]),
            (
                dictgen::InsensitiveStr::Ascii("desingable"),
                &["designable"],
            ),
            (dictgen::InsensitiveStr::Ascii("propertis"), &["properties"]),
            (dictgen::InsensitiveStr::Ascii("graphie"), &["graphite"]),
            (dictgen::InsensitiveStr::Ascii("chiansaw"), &["chainsaw"]),
            (
                dictgen::InsensitiveStr::Ascii("unfourtunate"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desactivate"),
                &["deactivate"],
            ),
            (dictgen::InsensitiveStr::Ascii("binnaries"), &["binaries"]),
            (
                dictgen::InsensitiveStr::Ascii("modifiations"),
                &["modifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("filnal"), &["final"]),
            (
                dictgen::InsensitiveStr::Ascii("administartors"),
                &["administrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpt"), &["except"]),
            (dictgen::InsensitiveStr::Ascii("timezeno"), &["timezone"]),
            (dictgen::InsensitiveStr::Ascii("thtere"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("assciative"),
                &["associative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repsentation"),
                &["representation"],
            ),
            (dictgen::InsensitiveStr::Ascii("microsofot"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("progresssive"),
                &["progressives"],
            ),
            (dictgen::InsensitiveStr::Ascii("dancin"), &["dancing"]),
            (dictgen::InsensitiveStr::Ascii("evloved"), &["evolved"]),
            (dictgen::InsensitiveStr::Ascii("janaury"), &["january"]),
            (dictgen::InsensitiveStr::Ascii("renketon"), &["renekton"]),
            (
                dictgen::InsensitiveStr::Ascii("discontigous"),
                &["discontiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefectly"), &["perfectly"]),
            (
                dictgen::InsensitiveStr::Ascii("corections"),
                &["corrections"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquired"), &["acquired"]),
            (dictgen::InsensitiveStr::Ascii("dispossing"), &["disposing"]),
            (dictgen::InsensitiveStr::Ascii("satrical"), &["satirical"]),
            (dictgen::InsensitiveStr::Ascii("bufffer"), &["buffer"]),
            (
                dictgen::InsensitiveStr::Ascii("ratatoolee"),
                &["ratatouille"],
            ),
            (dictgen::InsensitiveStr::Ascii("externel"), &["external"]),
            (dictgen::InsensitiveStr::Ascii("columbidae"), &["columbia"]),
            (
                dictgen::InsensitiveStr::Ascii("applicatiion"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relocatated"),
                &["relocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("transalt"), &["translate"]),
            (
                dictgen::InsensitiveStr::Ascii("funciotnality"),
                &["functionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("swiotch"), &["switch"]),
            (
                dictgen::InsensitiveStr::Ascii("congiguring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inpsections"),
                &["inspections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diections"),
                &["directions", "dissections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authroizing"),
                &["authorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resulotions"),
                &["resolutions"],
            ),
            (dictgen::InsensitiveStr::Ascii("scondary"), &["secondary"]),
            (
                dictgen::InsensitiveStr::Ascii("consumating"),
                &["consummating"],
            ),
            (dictgen::InsensitiveStr::Ascii("anaylised"), &["analysed"]),
            (
                dictgen::InsensitiveStr::Ascii("abstration"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpos"), &["typos"]),
            (dictgen::InsensitiveStr::Ascii("descirbed"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("bloked"), &["blocked"]),
            (dictgen::InsensitiveStr::Ascii("adedd"), &["added"]),
            (dictgen::InsensitiveStr::Ascii("oblitque"), &["oblique"]),
            (
                dictgen::InsensitiveStr::Ascii("provdidet"),
                &["provided", "provider", "provident"],
            ),
            (dictgen::InsensitiveStr::Ascii("devloper"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("beaviour"), &["behaviour"]),
            (
                dictgen::InsensitiveStr::Ascii("preprosessor"),
                &["preprocessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("occassion"), &["occasion"]),
            (
                dictgen::InsensitiveStr::Ascii("inconsistensies"),
                &["inconsistencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrrupting"),
                &["interrupting"],
            ),
            (dictgen::InsensitiveStr::Ascii("develope"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("compeat"), &["compete"]),
            (dictgen::InsensitiveStr::Ascii("withon"), &["within"]),
            (dictgen::InsensitiveStr::Ascii("datatyoes"), &["datatypes"]),
            (
                dictgen::InsensitiveStr::Ascii("negogtiations"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticaion"),
                &["authentication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoducible"),
                &["reproducible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uneforceable"),
                &["unenforceable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humillated"),
                &["humiliated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competetive"),
                &["competitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("poeples"), &["peoples"]),
            (dictgen::InsensitiveStr::Ascii("campagins"), &["campaigns"]),
            (dictgen::InsensitiveStr::Ascii("reuinon"), &["reunion"]),
            (dictgen::InsensitiveStr::Ascii("imblance"), &["imbalance"]),
            (
                dictgen::InsensitiveStr::Ascii("deserailise"),
                &["deserialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcualting"),
                &["calculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("awarness"), &["awareness"]),
            (dictgen::InsensitiveStr::Ascii("exapnsion"), &["expansion"]),
            (dictgen::InsensitiveStr::Ascii("eailier"), &["earlier"]),
            (dictgen::InsensitiveStr::Ascii("analsyts"), &["analysts"]),
            (dictgen::InsensitiveStr::Ascii("retreived"), &["retrieved"]),
            (
                dictgen::InsensitiveStr::Ascii("implememting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("ohone"), &["phone"]),
            (dictgen::InsensitiveStr::Ascii("cloeser"), &["closer"]),
            (
                dictgen::InsensitiveStr::Ascii("trasformed"),
                &["transformed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrictivo"),
                &["restriction"],
            ),
            (dictgen::InsensitiveStr::Ascii("capaticity"), &["capacity"]),
            (
                dictgen::InsensitiveStr::Ascii("intantiations"),
                &["instantiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transsexuel"),
                &["transsexual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transportatin"),
                &["transportation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inintializers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifgures"),
                &["configures"],
            ),
            (dictgen::InsensitiveStr::Ascii("huntmsan"), &["huntsman"]),
            (
                dictgen::InsensitiveStr::Ascii("multilplications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversoin"),
                &["conversions", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("tichened"), &["thickened"]),
            (dictgen::InsensitiveStr::Ascii("wjat"), &["what"]),
            (
                dictgen::InsensitiveStr::Ascii("communciation"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaible"),
                &["unavailable"],
            ),
            (dictgen::InsensitiveStr::Ascii("remmebers"), &["remembers"]),
            (
                dictgen::InsensitiveStr::Ascii("inahbitants"),
                &["inhabitants"],
            ),
            (dictgen::InsensitiveStr::Ascii("anytying"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("historycally"),
                &["historically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mathching"), &["matching"]),
            (
                dictgen::InsensitiveStr::Ascii("statistisk"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerinomy"), &["ceremony"]),
            (dictgen::InsensitiveStr::Ascii("assassinos"), &["assassins"]),
            (dictgen::InsensitiveStr::Ascii("amateures"), &["amateurs"]),
            (
                dictgen::InsensitiveStr::Ascii("revolutionnary"),
                &["revolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("thumbbnail"), &["thumbnail"]),
            (dictgen::InsensitiveStr::Ascii("inpurt"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("righht"), &["right"]),
            (
                dictgen::InsensitiveStr::Ascii("outrageuosly"),
                &["outrageously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfiguration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("sesssions"), &["sessions"]),
            (
                dictgen::InsensitiveStr::Ascii("correspontence"),
                &["correspondence"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleannup"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("publshed"), &["published"]),
            (dictgen::InsensitiveStr::Ascii("comatible"), &["compatible"]),
            (
                dictgen::InsensitiveStr::Ascii("arcitectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("puproses"), &["purposes"]),
            (
                dictgen::InsensitiveStr::Ascii("commnd"),
                &["command", "commend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organisates"),
                &["organises"],
            ),
            (dictgen::InsensitiveStr::Ascii("diregard"), &["disregard"]),
            (dictgen::InsensitiveStr::Ascii("triffer"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("enconders"), &["encoders"]),
            (
                dictgen::InsensitiveStr::Ascii("invaraibly"),
                &["invariably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impoverised"),
                &["impoverished"],
            ),
            (dictgen::InsensitiveStr::Ascii("trickyer"), &["trickery"]),
            (
                dictgen::InsensitiveStr::Ascii("expolsions"),
                &["explosions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separtation"),
                &["separation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ressolutions"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millseconds"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("acccepters"), &["accepters"]),
            (dictgen::InsensitiveStr::Ascii("vecotrs"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("custmer"), &["customer"]),
            (
                dictgen::InsensitiveStr::Ascii("siginificant"),
                &["significant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patriarical"),
                &["patriarchal"],
            ),
            (dictgen::InsensitiveStr::Ascii("furzzer"), &["fuzzer"]),
            (
                dictgen::InsensitiveStr::Ascii("courced"),
                &["coerced", "sourced", "coursed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ditinguishes"),
                &["distinguishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("vairation"), &["variation"]),
            (
                dictgen::InsensitiveStr::Ascii("vindicitve"),
                &["vindictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dicationaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuets"), &["executes"]),
            (dictgen::InsensitiveStr::Ascii("conifg"), &["config"]),
            (
                dictgen::InsensitiveStr::Ascii("mointoring"),
                &["monitoring"],
            ),
            (dictgen::InsensitiveStr::Ascii("pursuade"), &["persuade"]),
            (
                dictgen::InsensitiveStr::Ascii("introspeccion"),
                &["introspection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helpe"),
                &["helper", "help", "helped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidad"),
                &["intimidated"],
            ),
            (dictgen::InsensitiveStr::Ascii("liberales"), &["liberals"]),
            (dictgen::InsensitiveStr::Ascii("bouinded"), &["bounded"]),
            (dictgen::InsensitiveStr::Ascii("meetigns"), &["meetings"]),
            (
                dictgen::InsensitiveStr::Ascii("abitrarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("addess"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("decomoposition"),
                &["decomposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preaparation"),
                &["preparation"],
            ),
            (dictgen::InsensitiveStr::Ascii("suprizing"), &["surprising"]),
            (
                dictgen::InsensitiveStr::Ascii("transgrassion"),
                &["transgression"],
            ),
            (dictgen::InsensitiveStr::Ascii("containn"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("speciefier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("setteles"), &["settles"]),
            (dictgen::InsensitiveStr::Ascii("addtional"), &["additional"]),
            (dictgen::InsensitiveStr::Ascii("executie"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("adequatedly"),
                &["adequately"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicals"), &["specials"]),
            (dictgen::InsensitiveStr::Ascii("construc"), &["construct"]),
            (dictgen::InsensitiveStr::Ascii("settins"), &["settings"]),
            (
                dictgen::InsensitiveStr::Ascii("attribuetd"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("unsignd"), &["unsigned"]),
            (
                dictgen::InsensitiveStr::Ascii("presomption"),
                &["presumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("corsshair"), &["crosshair"]),
            (dictgen::InsensitiveStr::Ascii("nams"), &["names"]),
            (
                dictgen::InsensitiveStr::Ascii("embaresses"),
                &["embarrasses"],
            ),
            (dictgen::InsensitiveStr::Ascii("tmis"), &["this"]),
            (
                dictgen::InsensitiveStr::Ascii("capbabilities"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transofrmers"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substitutie"),
                &["substitutes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhooods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("counld"), &["could"]),
            (
                dictgen::InsensitiveStr::Ascii("devirtualisaiton"),
                &["devirtualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wourd"), &["word"]),
            (
                dictgen::InsensitiveStr::Ascii("streightish"),
                &["straightish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reciporicate"),
                &["reciprocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identifeid"),
                &["identified"],
            ),
            (dictgen::InsensitiveStr::Ascii("incuring"), &["incurring"]),
            (
                dictgen::InsensitiveStr::Ascii("loggee"),
                &["logged", "logger"],
            ),
            (dictgen::InsensitiveStr::Ascii("herculeus"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("wednsday"), &["wednesday"]),
            (dictgen::InsensitiveStr::Ascii("storted"), &["stored"]),
            (dictgen::InsensitiveStr::Ascii("chaarcter"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("guarantes"), &["guarantees"]),
            (
                dictgen::InsensitiveStr::Ascii("aling"),
                &["aline", "along", "ailing", "sling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equests"),
                &["requests", "quests"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recoginises"),
                &["recognises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absolultely"),
                &["absolutely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continging"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administative"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumbes"), &["assumes"]),
            (dictgen::InsensitiveStr::Ascii("consome"), &["consume"]),
            (
                dictgen::InsensitiveStr::Ascii("carismatically"),
                &["charismatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descriptvely"),
                &["descriptively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixying"),
                &["specifying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converning"),
                &["converting"],
            ),
            (dictgen::InsensitiveStr::Ascii("througout"), &["throughout"]),
            (dictgen::InsensitiveStr::Ascii("uptream"), &["upstream"]),
            (dictgen::InsensitiveStr::Ascii("neccassry"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("appicalitons"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("ingocnito"), &["incognito"]),
            (dictgen::InsensitiveStr::Ascii("sibtitles"), &["subtitles"]),
            (dictgen::InsensitiveStr::Ascii("specturm"), &["spectrum"]),
            (
                dictgen::InsensitiveStr::Ascii("alhapetically"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descroptively"),
                &["descriptively"],
            ),
            (dictgen::InsensitiveStr::Ascii("conmpress"), &["compress"]),
            (
                dictgen::InsensitiveStr::Ascii("outsourcade"),
                &["outsourced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfertable"),
                &["comfortable"],
            ),
            (dictgen::InsensitiveStr::Ascii("wlll"), &["will", "well"]),
            (
                dictgen::InsensitiveStr::Ascii("reminising"),
                &["reminiscing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associaitons"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninistalled"),
                &["uninstalled"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawbale"), &["drawable"]),
            (
                dictgen::InsensitiveStr::Ascii("reccommending"),
                &["recommending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mechanicle"),
                &["mechanical"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaguley"), &["vaguely"]),
            (
                dictgen::InsensitiveStr::Ascii("regulamentations"),
                &["regulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cappacitors"),
                &["capacitors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasitioned"),
                &["transitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phylosophically"),
                &["philosophically"],
            ),
            (dictgen::InsensitiveStr::Ascii("serveced"), &["serviced"]),
            (dictgen::InsensitiveStr::Ascii("datasts"), &["datasets"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuing"),
                &["contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("healtheir"), &["healthier"]),
            (
                dictgen::InsensitiveStr::Ascii("associatations"),
                &["associations"],
            ),
            (dictgen::InsensitiveStr::Ascii("byties"), &["bytes"]),
            (
                dictgen::InsensitiveStr::Ascii("visualizin"),
                &["visualizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("resposes"), &["responses"]),
            (dictgen::InsensitiveStr::Ascii("parametr"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("respecive"), &["respective"]),
            (dictgen::InsensitiveStr::Ascii("midigate"), &["mitigate"]),
            (
                dictgen::InsensitiveStr::Ascii("controlable"),
                &["controllable"],
            ),
            (dictgen::InsensitiveStr::Ascii("casaroles"), &["casseroles"]),
            (
                dictgen::InsensitiveStr::Ascii("exepmtions"),
                &["exemptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("quickets"), &["quickest"]),
            (dictgen::InsensitiveStr::Ascii("cheetoos"), &["cheetos"]),
            (dictgen::InsensitiveStr::Ascii("iamge"), &["image"]),
            (dictgen::InsensitiveStr::Ascii("arica"), &["africa"]),
            (
                dictgen::InsensitiveStr::Ascii("unconventinal"),
                &["unconventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissalowing"),
                &["disallowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("unopossed"), &["unopposed"]),
            (
                dictgen::InsensitiveStr::Ascii("exorbatent"),
                &["exorbitant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segergation"),
                &["segregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absentse"),
                &["absence", "absentee"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessaties"),
                &["necessities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environtmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("comparemos"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("recoginsers"),
                &["recognisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approxmimated"),
                &["approximated"],
            ),
            (dictgen::InsensitiveStr::Ascii("excpected"), &["expected"]),
            (dictgen::InsensitiveStr::Ascii("contaiun"), &["contain"]),
            (dictgen::InsensitiveStr::Ascii("jamaicain"), &["jamaican"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiaser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("tarmigan"), &["ptarmigan"]),
            (
                dictgen::InsensitiveStr::Ascii("virual"),
                &["virtual", "viral", "visual"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepeated"), &["separated"]),
            (dictgen::InsensitiveStr::Ascii("fllenames"), &["filenames"]),
            (dictgen::InsensitiveStr::Ascii("russina"), &["russian"]),
            (dictgen::InsensitiveStr::Ascii("hadnles"), &["handles"]),
            (
                dictgen::InsensitiveStr::Ascii("convervative"),
                &["conservative"],
            ),
            (dictgen::InsensitiveStr::Ascii("apendages"), &["appendages"]),
            (
                dictgen::InsensitiveStr::Ascii("cumulattive"),
                &["cumulative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defauls"),
                &["defaults", "default"],
            ),
            (dictgen::InsensitiveStr::Ascii("localtion"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("forawrd"), &["forward"]),
            (dictgen::InsensitiveStr::Ascii("imedatly"), &["immediately"]),
            (dictgen::InsensitiveStr::Ascii("reasom"), &["reason"]),
            (dictgen::InsensitiveStr::Ascii("apenines"), &["apennines"]),
            (
                dictgen::InsensitiveStr::Ascii("forewarding"),
                &["forwarding", "forewarning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chasr"),
                &["chaser", "chase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forecfully"),
                &["forcefully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ingredientes"),
                &["ingredients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermometor"),
                &["thermometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introducion"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("zippin"), &["zipping"]),
            (
                dictgen::InsensitiveStr::Ascii("tansported"),
                &["transported"],
            ),
            (dictgen::InsensitiveStr::Ascii("sigature"), &["signature"]),
            (
                dictgen::InsensitiveStr::Ascii("parammeters"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilcated"),
                &["complicate", "complicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("mismach"), &["mismatch"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhhoods"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("emitetd"), &["emitted"]),
            (
                dictgen::InsensitiveStr::Ascii("underestimeted"),
                &["underestimated"],
            ),
            (dictgen::InsensitiveStr::Ascii("speciman"), &["specimen"]),
            (
                dictgen::InsensitiveStr::Ascii("morpholgical"),
                &["morphological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microtransactons"),
                &["microtransactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underminde"),
                &["undermines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paticipant"),
                &["participant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electircal"),
                &["electrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unadvertently"),
                &["inadvertently"],
            ),
            (dictgen::InsensitiveStr::Ascii("unloadins"), &["unloading"]),
            (dictgen::InsensitiveStr::Ascii("accalimed"), &["acclaimed"]),
            (
                dictgen::InsensitiveStr::Ascii("proect"),
                &["protect", "project"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retriction"),
                &["restriction", "retraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("astiimate"), &["estimate"]),
            (dictgen::InsensitiveStr::Ascii("anitque"), &["antique"]),
            (
                dictgen::InsensitiveStr::Ascii("sovereighnty"),
                &["sovereignty"],
            ),
            (dictgen::InsensitiveStr::Ascii("consition"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("useername"), &["username"]),
            (dictgen::InsensitiveStr::Ascii("rocorded"), &["recorded"]),
            (
                dictgen::InsensitiveStr::Ascii("temporarirly"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("horading"), &["hoarding"]),
            (dictgen::InsensitiveStr::Ascii("iteratable"), &["iterable"]),
            (
                dictgen::InsensitiveStr::Ascii("baises"),
                &["biases", "raises", "braises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranposing"),
                &["transposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equavilents"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchnages"), &["exchanges"]),
            (dictgen::InsensitiveStr::Ascii("astersisks"), &["asterisks"]),
            (
                dictgen::InsensitiveStr::Ascii("attrbiution"),
                &["attribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("appicable"), &["applicable"]),
            (
                dictgen::InsensitiveStr::Ascii("advertisters"),
                &["advertisers"],
            ),
            (dictgen::InsensitiveStr::Ascii("rechek"), &["recheck"]),
            (dictgen::InsensitiveStr::Ascii("goe"), &["geo", "go"]),
            (dictgen::InsensitiveStr::Ascii("quesant"), &["croissant"]),
            (
                dictgen::InsensitiveStr::Ascii("vehemontly"),
                &["vehemently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unadvertent"),
                &["inadvertent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contributie"),
                &["contribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("anarchiszm"), &["anarchism"]),
            (dictgen::InsensitiveStr::Ascii("redneckese"), &["rednecks"]),
            (dictgen::InsensitiveStr::Ascii("freindly"), &["friendly"]),
            (dictgen::InsensitiveStr::Ascii("prevelant"), &["prevalent"]),
            (
                dictgen::InsensitiveStr::Ascii("relativley"),
                &["relatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("prohibis"), &["prohibits"]),
            (
                dictgen::InsensitiveStr::Ascii("cloure"),
                &["clojure", "closure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjasantly"),
                &["adjacently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vocabluary"),
                &["vocabulary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspefiied"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfacce"),
                &["interfaces", "interface"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("emhpasizing"),
                &["emphasizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("acronymns"), &["acronyms"]),
            (
                dictgen::InsensitiveStr::Ascii("mainpulate"),
                &["manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaneusly"),
                &["simultaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("abudcts"), &["abducts"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalistiske"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("physicsts"), &["physicists"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mmonitoring"),
                &["monitoring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathamatician"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("commpleted"), &["completed"]),
            (
                dictgen::InsensitiveStr::Ascii("associatie"),
                &["associated", "associate", "associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("feautures"), &["features"]),
            (dictgen::InsensitiveStr::Ascii("progaming"), &["programing"]),
            (dictgen::InsensitiveStr::Ascii("restraind"), &["restrained"]),
            (
                dictgen::InsensitiveStr::Ascii("replacememt"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("instaled"), &["installed"]),
            (dictgen::InsensitiveStr::Ascii("foldeer"), &["folder"]),
            (
                dictgen::InsensitiveStr::Ascii("asteriks"),
                &["asterisk", "asterisks"],
            ),
            (dictgen::InsensitiveStr::Ascii("allcate"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprehencible"),
                &["incomprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeffcient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("projcts"), &["projects"]),
            (dictgen::InsensitiveStr::Ascii("buildders"), &["builders"]),
            (
                dictgen::InsensitiveStr::Ascii("speciallity"),
                &["specially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protestina"),
                &["protestant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desent"),
                &["descent", "decent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decalrators"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("humilitaing"),
                &["humiliating"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoing"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("certifcates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("zuser"), &["user"]),
            (dictgen::InsensitiveStr::Ascii("netowrk"), &["network"]),
            (dictgen::InsensitiveStr::Ascii("insepector"), &["inspector"]),
            (dictgen::InsensitiveStr::Ascii("eveolve"), &["evolve"]),
            (dictgen::InsensitiveStr::Ascii("definelty"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("stompted"), &["stomped"]),
            (
                dictgen::InsensitiveStr::Ascii("signautres"),
                &["signatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assessmant"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcutator"),
                &["calculator"],
            ),
            (dictgen::InsensitiveStr::Ascii("helful"), &["helpful"]),
            (
                dictgen::InsensitiveStr::Ascii("sourbraten"),
                &["sauerbraten"],
            ),
            (dictgen::InsensitiveStr::Ascii("settigs"), &["settings"]),
            (dictgen::InsensitiveStr::Ascii("vaidated"), &["validated"]),
            (dictgen::InsensitiveStr::Ascii("atlantia"), &["atlanta"]),
            (dictgen::InsensitiveStr::Ascii("spekaing"), &["speaking"]),
            (
                dictgen::InsensitiveStr::Ascii("converages"),
                &["coverages", "converges"],
            ),
            (dictgen::InsensitiveStr::Ascii("manouvers"), &["maneuvers"]),
            (dictgen::InsensitiveStr::Ascii("errror"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("socalists"), &["socialists"]),
            (dictgen::InsensitiveStr::Ascii("lauout"), &["layout"]),
            (
                dictgen::InsensitiveStr::Ascii("algotithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypotethically"),
                &["hypothetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("bunlding"), &["bundling"]),
            (dictgen::InsensitiveStr::Ascii("chihuaha"), &["chihuahua"]),
            (dictgen::InsensitiveStr::Ascii("oepnapi"), &["openapi"]),
            (
                dictgen::InsensitiveStr::Ascii("consqeuences"),
                &["consequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("approched"), &["approached"]),
            (dictgen::InsensitiveStr::Ascii("goblings"), &["goblins"]),
            (
                dictgen::InsensitiveStr::Ascii("instantanous"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horinzontal"),
                &["horizontal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibile"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intiializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contradictiong"),
                &["contradicting"],
            ),
            (dictgen::InsensitiveStr::Ascii("napomming"), &["napalming"]),
            (dictgen::InsensitiveStr::Ascii("arranced"), &["arranged"]),
            (
                dictgen::InsensitiveStr::Ascii("confidense"),
                &["confidence"],
            ),
            (dictgen::InsensitiveStr::Ascii("deltayed"), &["delayed"]),
            (dictgen::InsensitiveStr::Ascii("durationns"), &["durations"]),
            (dictgen::InsensitiveStr::Ascii("applaued"), &["applaud"]),
            (
                dictgen::InsensitiveStr::Ascii("transister"),
                &["transistor"],
            ),
            (dictgen::InsensitiveStr::Ascii("segmet"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("affilates"), &["affiliates"]),
            (dictgen::InsensitiveStr::Ascii("defintie"), &["definite"]),
            (
                dictgen::InsensitiveStr::Ascii("esssentially"),
                &["essentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("privides"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("exmapl"), &["example"]),
            (dictgen::InsensitiveStr::Ascii("reverced"), &["reversed"]),
            (dictgen::InsensitiveStr::Ascii("extendded"), &["extended"]),
            (
                dictgen::InsensitiveStr::Ascii("hald"),
                &["held", "half", "hall", "hold"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trogliditical"),
                &["troglodytical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confifurations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("weitght"), &["weight"]),
            (
                dictgen::InsensitiveStr::Ascii("intoduction"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("pamplet"), &["pamphlet"]),
            (dictgen::InsensitiveStr::Ascii("suseeded"), &["seceded"]),
            (
                dictgen::InsensitiveStr::Ascii("declation"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifyied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("anytthing"), &["anything"]),
            (
                dictgen::InsensitiveStr::Ascii("divised"),
                &["divided", "devised"],
            ),
            (dictgen::InsensitiveStr::Ascii("execuctors"), &["executors"]),
            (dictgen::InsensitiveStr::Ascii("analzyed"), &["analyzed"]),
            (
                dictgen::InsensitiveStr::Ascii("exclamacion"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sonewhere"), &["somewhere"]),
            (dictgen::InsensitiveStr::Ascii("tabable"), &["tabbable"]),
            (
                dictgen::InsensitiveStr::Ascii("statisitical"),
                &["statistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("algerba"), &["algebra"]),
            (dictgen::InsensitiveStr::Ascii("reedeming"), &["redeeming"]),
            (dictgen::InsensitiveStr::Ascii("syntethic"), &["synthetic"]),
            (dictgen::InsensitiveStr::Ascii("effiecient"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("precisie"), &["precise"]),
            (
                dictgen::InsensitiveStr::Ascii("filessystem"),
                &["filesystem"],
            ),
            (dictgen::InsensitiveStr::Ascii("attrbutes"), &["attributes"]),
            (dictgen::InsensitiveStr::Ascii("cleaering"), &["clearing"]),
            (
                dictgen::InsensitiveStr::Ascii("mulitplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catogerized"),
                &["categorized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caracteres"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invalidaitons"),
                &["invalidations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documneting"),
                &["documenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("specialis"), &["specials"]),
            (dictgen::InsensitiveStr::Ascii("devicess"), &["devices"]),
            (dictgen::InsensitiveStr::Ascii("cehckout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("corrsponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annotions"),
                &["annotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constraintes"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoter"),
                &["reporter", "repeater", "remoter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiales"),
                &["initials", "initialise", "initializes", "initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissagregation"),
                &["dissaggregation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timestmping"),
                &["timestamping"],
            ),
            (dictgen::InsensitiveStr::Ascii("actaully"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("diferences"),
                &["differences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reppositories"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("enlcave"), &["enclave"]),
            (dictgen::InsensitiveStr::Ascii("semseters"), &["semesters"]),
            (
                dictgen::InsensitiveStr::Ascii("dissastified"),
                &["dissatisfied"],
            ),
            (dictgen::InsensitiveStr::Ascii("teamifght"), &["teamfight"]),
            (dictgen::InsensitiveStr::Ascii("conjugage"), &["conjugate"]),
            (dictgen::InsensitiveStr::Ascii("selcetion"), &["selection"]),
            (
                dictgen::InsensitiveStr::Ascii("destributor"),
                &["distributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("saddness"), &["sadness"]),
            (
                dictgen::InsensitiveStr::Ascii("pase"),
                &["pass", "pace", "parse", "phase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specfically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anxiuosness"),
                &["anxiousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rademtions"),
                &["redemptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defininely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summmarizes"),
                &["summarizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undernieth"),
                &["underneath"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacl"), &["back"]),
            (dictgen::InsensitiveStr::Ascii("poossibly"), &["possibly"]),
            (
                dictgen::InsensitiveStr::Ascii("identiications"),
                &["identifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("entrpoint"), &["entrypoint"]),
            (dictgen::InsensitiveStr::Ascii("outwieghs"), &["outweighs"]),
            (dictgen::InsensitiveStr::Ascii("hexoganal"), &["hexagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("physcologists"),
                &["psychologists"],
            ),
            (dictgen::InsensitiveStr::Ascii("immersve"), &["immerse"]),
            (dictgen::InsensitiveStr::Ascii("symobols"), &["symbols"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sweidsh"), &["swedish"]),
            (
                dictgen::InsensitiveStr::Ascii("socioeconmic"),
                &["socioeconomic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arangements"),
                &["arrangements"],
            ),
            (dictgen::InsensitiveStr::Ascii("moteef"), &["motif"]),
            (
                dictgen::InsensitiveStr::Ascii("cockroachers"),
                &["cockroaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("poitive"), &["positive"]),
            (
                dictgen::InsensitiveStr::Ascii("epect"),
                &["expect", "elect"],
            ),
            (dictgen::InsensitiveStr::Ascii("quotaions"), &["quotations"]),
            (
                dictgen::InsensitiveStr::Ascii("prometheyus"),
                &["prometheus"],
            ),
            (dictgen::InsensitiveStr::Ascii("cature"), &["capture"]),
            (
                dictgen::InsensitiveStr::Ascii("implementaitons"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antialised"),
                &["antialiased"],
            ),
            (dictgen::InsensitiveStr::Ascii("filetering"), &["filtering"]),
            (dictgen::InsensitiveStr::Ascii("eyou"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("taligating"),
                &["tailgating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cathegorize"),
                &["categorize"],
            ),
            (dictgen::InsensitiveStr::Ascii("delibrate"), &["deliberate"]),
            (
                dictgen::InsensitiveStr::Ascii("inresponsive"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("enryption"), &["encryption"]),
            (
                dictgen::InsensitiveStr::Ascii("otpimizing"),
                &["optimizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("gutteral"), &["guttural"]),
            (dictgen::InsensitiveStr::Ascii("commtitee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("versoin"), &["version"]),
            (
                dictgen::InsensitiveStr::Ascii("playthrouh"),
                &["playthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("implictly"), &["implicitly"]),
            (dictgen::InsensitiveStr::Ascii("offically"), &["officially"]),
            (
                dictgen::InsensitiveStr::Ascii("excetpions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("expese"), &["expense"]),
            (
                dictgen::InsensitiveStr::Ascii("infomatrion"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convertirme"),
                &["converter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surious"),
                &["serious", "spurious", "curious", "furious", "usurious"],
            ),
            (dictgen::InsensitiveStr::Ascii("shreak"), &["shriek"]),
            (
                dictgen::InsensitiveStr::Ascii("visulisations"),
                &["visualisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("terminatd"), &["terminated"]),
            (
                dictgen::InsensitiveStr::Ascii("altruisitc"),
                &["altruistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("poducibly"), &["producibly"]),
            (dictgen::InsensitiveStr::Ascii("focument"), &["document"]),
            (dictgen::InsensitiveStr::Ascii("synonomy"), &["synonym"]),
            (
                dictgen::InsensitiveStr::Ascii("tansitioning"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribuem"),
                &["distribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connnnection"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persausive"),
                &["persuasive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conversible"),
                &["convertible"],
            ),
            (dictgen::InsensitiveStr::Ascii("tangeants"), &["tangents"]),
            (
                dictgen::InsensitiveStr::Ascii("mindleslly"),
                &["mindlessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserailization"),
                &["deserialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interaxctions"),
                &["interactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complection"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defferently"),
                &["differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conerts"),
                &["converts", "concerts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resourse"),
                &["resources", "recourse", "resource"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asisstance"),
                &["assistance"],
            ),
            (dictgen::InsensitiveStr::Ascii("accees"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("goign"), &["going"]),
            (
                dictgen::InsensitiveStr::Ascii("electromagentic"),
                &["electromagnetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("harcodes"),
                &["hardcodes", "charcodes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrupable"),
                &["interruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconenct"), &["reconnect"]),
            (dictgen::InsensitiveStr::Ascii("decration"), &["decoration"]),
            (dictgen::InsensitiveStr::Ascii("ngix"), &["nginx"]),
            (
                dictgen::InsensitiveStr::Ascii("scince"),
                &["science", "since"],
            ),
            (dictgen::InsensitiveStr::Ascii("collectie"), &["collective"]),
            (dictgen::InsensitiveStr::Ascii("guareenty"), &["guaranty"]),
            (dictgen::InsensitiveStr::Ascii("docuhebag"), &["douchebag"]),
            (
                dictgen::InsensitiveStr::Ascii("recipeints"),
                &["recipients"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availbility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valiability"),
                &["variability", "availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("concious"), &["conscious"]),
            (dictgen::InsensitiveStr::Ascii("phote"), &["photo"]),
            (
                dictgen::InsensitiveStr::Ascii("correpondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sigificance"),
                &["significance"],
            ),
            (dictgen::InsensitiveStr::Ascii("heighit"), &["height"]),
            (dictgen::InsensitiveStr::Ascii("acquiess"), &["acquiesce"]),
            (dictgen::InsensitiveStr::Ascii("howevere"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("deviatin"),
                &["deviating", "deviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miniscully"),
                &["minusculely"],
            ),
            (dictgen::InsensitiveStr::Ascii("prameters"), &["parameters"]),
            (
                dictgen::InsensitiveStr::Ascii("socialsits"),
                &["socialists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sytemerror"),
                &["systemerror"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("obersvations"),
                &["observations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unappropriate"),
                &["inappropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intagrating"),
                &["integrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dergistration"),
                &["deregistration"],
            ),
            (dictgen::InsensitiveStr::Ascii("constatns"), &["constants"]),
            (dictgen::InsensitiveStr::Ascii("pyramides"), &["pyramids"]),
            (
                dictgen::InsensitiveStr::Ascii("sheat"),
                &["sheath", "sheet", "cheat"],
            ),
            (dictgen::InsensitiveStr::Ascii("permutin"), &["permuting"]),
            (dictgen::InsensitiveStr::Ascii("expolsive"), &["explosive"]),
            (
                dictgen::InsensitiveStr::Ascii("syntaxically"),
                &["syntactically"],
            ),
            (dictgen::InsensitiveStr::Ascii("toursits"), &["tourists"]),
            (dictgen::InsensitiveStr::Ascii("cruical"), &["crucial"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithsmic"),
                &["algorithmic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiaties"),
                &["initiatives", "initiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("secceeded"),
                &["seceded", "succeeded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confugiring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("guilio"), &["giulio"]),
            (
                dictgen::InsensitiveStr::Ascii("uncontrollaby"),
                &["uncontrollably"],
            ),
            (dictgen::InsensitiveStr::Ascii("modefier"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("argumants"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("hotsport"), &["hotspot"]),
            (
                dictgen::InsensitiveStr::Ascii("fuctionalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("sucessul"), &["successful"]),
            (dictgen::InsensitiveStr::Ascii("ownerhsip"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("analyist"), &["analyst"]),
            (dictgen::InsensitiveStr::Ascii("edtor"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("livecycle"), &["lifecycle"]),
            (
                dictgen::InsensitiveStr::Ascii("explosvies"),
                &["explosives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interruptors"),
                &["interrupts", "interrupters"],
            ),
            (dictgen::InsensitiveStr::Ascii("agred"), &["agreed"]),
            (dictgen::InsensitiveStr::Ascii("contrct"), &["contract"]),
            (
                dictgen::InsensitiveStr::Ascii("correspdon"),
                &["correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("streming"), &["streaming"]),
            (dictgen::InsensitiveStr::Ascii("misnadry"), &["misandry"]),
            (
                dictgen::InsensitiveStr::Ascii("engieering"),
                &["engineering"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainifesto"), &["manifesto"]),
            (dictgen::InsensitiveStr::Ascii("elaspes"), &["elapses"]),
            (
                dictgen::InsensitiveStr::Ascii("amministrative"),
                &["administrative"],
            ),
            (dictgen::InsensitiveStr::Ascii("endanged"), &["endangered"]),
            (dictgen::InsensitiveStr::Ascii("encodingt"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("requiured"), &["required"]),
            (dictgen::InsensitiveStr::Ascii("specator"), &["spectator"]),
            (dictgen::InsensitiveStr::Ascii("channles"), &["channels"]),
            (
                dictgen::InsensitiveStr::Ascii("catostraphic"),
                &["catastrophic"],
            ),
            (dictgen::InsensitiveStr::Ascii("unknowun"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("recepcionist"),
                &["receptionist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerbility"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("quesion"), &["question"]),
            (dictgen::InsensitiveStr::Ascii("ontext"), &["context"]),
            (dictgen::InsensitiveStr::Ascii("connnects"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("ciritcal"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("techinical"), &["technical"]),
            (
                dictgen::InsensitiveStr::Ascii("notifition"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prioritsed"),
                &["prioritised"],
            ),
            (dictgen::InsensitiveStr::Ascii("storge"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("moderatorn"),
                &["moderation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stadistics"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("destroyes"), &["destroys"]),
            (dictgen::InsensitiveStr::Ascii("minimsed"), &["minimised"]),
            (dictgen::InsensitiveStr::Ascii("campare"), &["compare"]),
            (dictgen::InsensitiveStr::Ascii("repostus"), &["reposts"]),
            (
                dictgen::InsensitiveStr::Ascii("powerlifing"),
                &["powerlifting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychnronization"),
                &["synchronization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constitued"),
                &["constitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("srcipted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("acclimitization"),
                &["acclimatization"],
            ),
            (dictgen::InsensitiveStr::Ascii("sommetime"), &["sometime"]),
            (
                dictgen::InsensitiveStr::Ascii("porpuse"),
                &["purpose", "porpoise"],
            ),
            (dictgen::InsensitiveStr::Ascii("langugages"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("jalibroken"),
                &["jailbroken"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighbrs"), &["neighbors"]),
            (dictgen::InsensitiveStr::Ascii("loook"), &["look"]),
            (
                dictgen::InsensitiveStr::Ascii("increminted"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volentiering"),
                &["volunteering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("teathering"),
                &["tether", "feathering", "weathering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sequentialy"),
                &["sequentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermimentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aquiessing"),
                &["acquiescing"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoutot"), &["shoutout"]),
            (dictgen::InsensitiveStr::Ascii("iluminate"), &["illuminate"]),
            (dictgen::InsensitiveStr::Ascii("classsical"), &["classical"]),
            (dictgen::InsensitiveStr::Ascii("accureate"), &["accurate"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhoofs"),
                &["neighborhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("sciprted"), &["scripted"]),
            (
                dictgen::InsensitiveStr::Ascii("versionning"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("cantain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("astornauts"),
                &["astronauts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conecnts"),
                &["connects", "contents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modifciations"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranistion"),
                &["transition"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwwed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("businesmen"),
                &["businessmen"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomitted"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("downrades"), &["downgrades"]),
            (dictgen::InsensitiveStr::Ascii("unublish"), &["unpublish"]),
            (dictgen::InsensitiveStr::Ascii("destinty"), &["destiny"]),
            (dictgen::InsensitiveStr::Ascii("arreys"), &["arrays"]),
            (dictgen::InsensitiveStr::Ascii("relpaces"), &["replaces"]),
            (
                dictgen::InsensitiveStr::Ascii("paraprhasing"),
                &["paraphrasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thermomter"),
                &["thermometer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcribing"),
                &["subscribing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insuccessful"),
                &["unsuccessful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipications"),
                &["multiplications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incosistente"),
                &["inconsistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("critially"), &["critically"]),
            (dictgen::InsensitiveStr::Ascii("regualrly"), &["regularly"]),
            (dictgen::InsensitiveStr::Ascii("prool"), &["pool"]),
            (dictgen::InsensitiveStr::Ascii("funtional"), &["functional"]),
            (dictgen::InsensitiveStr::Ascii("surrogage"), &["surrogate"]),
            (
                dictgen::InsensitiveStr::Ascii("itnroduced"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similiarites"),
                &["similarities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fortunetly"),
                &["fortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explossive"),
                &["explosives"],
            ),
            (dictgen::InsensitiveStr::Ascii("secodly"), &["secondly"]),
            (dictgen::InsensitiveStr::Ascii("tatgerted"), &["targeted"]),
            (
                dictgen::InsensitiveStr::Ascii("unsupportd"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occucences"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("seacrhes"), &["searches"]),
            (
                dictgen::InsensitiveStr::Ascii("abomonation"),
                &["abomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("constan"), &["constant"]),
            (dictgen::InsensitiveStr::Ascii("propsects"), &["prospects"]),
            (dictgen::InsensitiveStr::Ascii("headear"), &["header"]),
            (
                dictgen::InsensitiveStr::Ascii("reconfugure"),
                &["reconfigure"],
            ),
            (dictgen::InsensitiveStr::Ascii("makfiles"), &["makefiles"]),
            (
                dictgen::InsensitiveStr::Ascii("discontinous"),
                &["discontinuous"],
            ),
            (dictgen::InsensitiveStr::Ascii("dedecated"), &["dedicated"]),
            (dictgen::InsensitiveStr::Ascii("sidelinien"), &["sideline"]),
            (dictgen::InsensitiveStr::Ascii("combininer"), &["combiner"]),
            (dictgen::InsensitiveStr::Ascii("picnicer"), &["picnicker"]),
            (
                dictgen::InsensitiveStr::Ascii("cateogrizing"),
                &["categorizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centrafuges"),
                &["centrifuges"],
            ),
            (dictgen::InsensitiveStr::Ascii("metrcis"), &["metrics"]),
            (
                dictgen::InsensitiveStr::Ascii("conveniance"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superceeded"),
                &["superseded"],
            ),
            (dictgen::InsensitiveStr::Ascii("basline"), &["baseline"]),
            (
                dictgen::InsensitiveStr::Ascii("expeimental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("tanslated"), &["translated"]),
            (dictgen::InsensitiveStr::Ascii("exlusion"), &["exclusion"]),
            (
                dictgen::InsensitiveStr::Ascii("establishmnet"),
                &["establishments"],
            ),
            (dictgen::InsensitiveStr::Ascii("diminshed"), &["diminished"]),
            (
                dictgen::InsensitiveStr::Ascii("transparrently"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("slelected"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("speficillics"),
                &["specifics"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicifed"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("overhwelm"), &["overwhelm"]),
            (dictgen::InsensitiveStr::Ascii("acordian"), &["accordion"]),
            (
                dictgen::InsensitiveStr::Ascii("pakageimpl"),
                &["packageimpl"],
            ),
            (dictgen::InsensitiveStr::Ascii("pariticle"), &["particle"]),
            (dictgen::InsensitiveStr::Ascii("objctives"), &["objectives"]),
            (
                dictgen::InsensitiveStr::Ascii("expemented"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("inlin"), &["inline"]),
            (dictgen::InsensitiveStr::Ascii("runnning"), &["running"]),
            (dictgen::InsensitiveStr::Ascii("offred"), &["offered"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalt"),
                &["fundamentalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlysers"), &["analysers"]),
            (dictgen::InsensitiveStr::Ascii("annhilate"), &["annihilate"]),
            (dictgen::InsensitiveStr::Ascii("acitvated"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("impementaion"),
                &["implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("methaphors"), &["metaphors"]),
            (dictgen::InsensitiveStr::Ascii("baseslines"), &["baselines"]),
            (
                dictgen::InsensitiveStr::Ascii("supsicions"),
                &["suspicions"],
            ),
            (dictgen::InsensitiveStr::Ascii("passanger"), &["passenger"]),
            (dictgen::InsensitiveStr::Ascii("charactor"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("instread"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("interesant"), &["interest"]),
            (dictgen::InsensitiveStr::Ascii("zookeinees"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("splittng"), &["splitting"]),
            (
                dictgen::InsensitiveStr::Ascii("propegating"),
                &["propagating"],
            ),
            (dictgen::InsensitiveStr::Ascii("innactive"), &["inactive"]),
            (
                dictgen::InsensitiveStr::Ascii("retributivo"),
                &["retribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("attampts"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("restransmits"),
                &["retransmits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggaravated"),
                &["aggravated"],
            ),
            (dictgen::InsensitiveStr::Ascii("refected"), &["reflected"]),
            (dictgen::InsensitiveStr::Ascii("durining"), &["during"]),
            (dictgen::InsensitiveStr::Ascii("downgrate"), &["downgrade"]),
            (dictgen::InsensitiveStr::Ascii("suscripts"), &["subscripts"]),
            (dictgen::InsensitiveStr::Ascii("leyered"), &["layered"]),
            (dictgen::InsensitiveStr::Ascii("ueses"), &["uses"]),
            (
                dictgen::InsensitiveStr::Ascii("documations"),
                &["documentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimnetal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("rech"), &["reach"]),
            (
                dictgen::InsensitiveStr::Ascii("thudnerbolt"),
                &["thunderbolt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrenteing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("scorpin"), &["scorpion"]),
            (dictgen::InsensitiveStr::Ascii("wayoints"), &["waypoints"]),
            (
                dictgen::InsensitiveStr::Ascii("sychronization"),
                &["synchronization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outrageuos"),
                &["outrageous"],
            ),
            (dictgen::InsensitiveStr::Ascii("baptsim"), &["baptism"]),
            (dictgen::InsensitiveStr::Ascii("architets"), &["architects"]),
            (dictgen::InsensitiveStr::Ascii("discareded"), &["discarded"]),
            (dictgen::InsensitiveStr::Ascii("comeing"), &["coming"]),
            (dictgen::InsensitiveStr::Ascii("concelaed"), &["concealed"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridination"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("finilizing"),
                &["finalizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dpeendencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dentity"),
                &["identity", "density", "entity"],
            ),
            (dictgen::InsensitiveStr::Ascii("dcuments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("intentionnally"),
                &["intentionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homeonwers"),
                &["homeowners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrigement"),
                &["infringement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caselessely"),
                &["caselessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepeate"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("underneight"),
                &["underweight"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntesize"), &["synthesize"]),
            (dictgen::InsensitiveStr::Ascii("diration"), &["duration"]),
            (
                dictgen::InsensitiveStr::Ascii("asymmeries"),
                &["asymmetries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reproducablitity"),
                &["reproducibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statistcal"),
                &["statistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("witth"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("excercised"), &["exercised"]),
            (
                dictgen::InsensitiveStr::Ascii("conparatively"),
                &["comparatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismatched"),
                &["mismatched", "dispatched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progressos"),
                &["progresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinatatorial"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("coninient"), &["convenient"]),
            (
                dictgen::InsensitiveStr::Ascii("conduiting"),
                &["conducting"],
            ),
            (dictgen::InsensitiveStr::Ascii("securrely"), &["securely"]),
            (
                dictgen::InsensitiveStr::Ascii("messagse"),
                &["messages", "message"],
            ),
            (dictgen::InsensitiveStr::Ascii("louisiania"), &["louisiana"]),
            (
                dictgen::InsensitiveStr::Ascii("assempling"),
                &["assembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("croozes"), &["cruises"]),
            (
                dictgen::InsensitiveStr::Ascii("calcalated"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("versiooned"), &["versioned"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalizm"),
                &["nationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("tsts"), &["tests"]),
            (
                dictgen::InsensitiveStr::Ascii("sociopatas"),
                &["sociopaths"],
            ),
            (dictgen::InsensitiveStr::Ascii("maneuveres"), &["maneuvers"]),
            (
                dictgen::InsensitiveStr::Ascii("invicinble"),
                &["invincible"],
            ),
            (dictgen::InsensitiveStr::Ascii("measurin"), &["measuring"]),
            (
                dictgen::InsensitiveStr::Ascii("authroises"),
                &["authorises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("soudntrack"),
                &["soundtrack"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cultrually"),
                &["culturally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearangement"),
                &["rearrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalites"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("multismapling"),
                &["multisampling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redirectrion"),
                &["redirection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("athorization"),
                &["authorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instabilt"),
                &["instability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algortihmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("confifs"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("reigster"), &["register"]),
            (dictgen::InsensitiveStr::Ascii("worloads"), &["workloads"]),
            (
                dictgen::InsensitiveStr::Ascii("crrespondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialzation"),
                &["initialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("taregts"), &["targets"]),
            (dictgen::InsensitiveStr::Ascii("unpacket"), &["unpacked"]),
            (dictgen::InsensitiveStr::Ascii("permises"), &["premises"]),
            (
                dictgen::InsensitiveStr::Ascii("subdoamins"),
                &["subdomains"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitialising"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expressview"),
                &["expressive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("containning"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understadning"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("knowningly"), &["knowingly"]),
            (
                dictgen::InsensitiveStr::Ascii("weakensses"),
                &["weaknesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("althorithm"), &["algorithm"]),
            (dictgen::InsensitiveStr::Ascii("suprious"), &["spurious"]),
            (
                dictgen::InsensitiveStr::Ascii("championsip"),
                &["championship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vewing"),
                &["viewing", "vowing", "vexing"],
            ),
            (dictgen::InsensitiveStr::Ascii("seeged"), &["sieged"]),
            (
                dictgen::InsensitiveStr::Ascii("colorpsaces"),
                &["colorspaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphbetically"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prersistent"),
                &["persistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversimplificiation"),
                &["oversimplification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaor"),
                &["renegotiator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionnaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("interprett"), &["interpret"]),
            (
                dictgen::InsensitiveStr::Ascii("outperfroms"),
                &["outperform"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characteds"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projectlie"),
                &["projectiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorence"), &["ignorance"]),
            (
                dictgen::InsensitiveStr::Ascii("announcemnt"),
                &["announcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("storng"), &["strong"]),
            (dictgen::InsensitiveStr::Ascii("renforces"), &["reinforces"]),
            (dictgen::InsensitiveStr::Ascii("passses"), &["passes"]),
            (
                dictgen::InsensitiveStr::Ascii("bootsstrapper"),
                &["bootstrapper"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcctries"),
                &["directories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prjections"),
                &["projections"],
            ),
            (dictgen::InsensitiveStr::Ascii("sppeches"), &["speeches"]),
            (dictgen::InsensitiveStr::Ascii("packges"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("prioritzed"),
                &["prioritized"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterae"), &["iterate"]),
            (dictgen::InsensitiveStr::Ascii("conseated"), &["conceited"]),
            (dictgen::InsensitiveStr::Ascii("chancks"), &["chunks"]),
            (
                dictgen::InsensitiveStr::Ascii("constarints"),
                &["constraint", "constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("quialified"), &["qualified"]),
            (dictgen::InsensitiveStr::Ascii("calcultate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("adviseer"), &["adviser"]),
            (dictgen::InsensitiveStr::Ascii("sovled"), &["solved"]),
            (dictgen::InsensitiveStr::Ascii("executues"), &["executes"]),
            (
                dictgen::InsensitiveStr::Ascii("responsing"),
                &["responding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("remenissense"),
                &["reminiscence"],
            ),
            (dictgen::InsensitiveStr::Ascii("meraj"), &["mirage"]),
            (
                dictgen::InsensitiveStr::Ascii("tansportable"),
                &["transportable"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqrites"), &["rewrites"]),
            (dictgen::InsensitiveStr::Ascii("slidseshow"), &["slideshow"]),
            (dictgen::InsensitiveStr::Ascii("foreced"), &["forced"]),
            (dictgen::InsensitiveStr::Ascii("unluckly"), &["unlucky"]),
            (
                dictgen::InsensitiveStr::Ascii("persoanlly"),
                &["personally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addiitonall"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("emty"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("perferances"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("wrkspaces"), &["workspaces"]),
            (dictgen::InsensitiveStr::Ascii("procuded"), &["produced"]),
            (dictgen::InsensitiveStr::Ascii("someoen"), &["someone"]),
            (
                dictgen::InsensitiveStr::Ascii("conservanti"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("confurse"), &["confuse"]),
            (dictgen::InsensitiveStr::Ascii("dsktops"), &["desktops"]),
            (dictgen::InsensitiveStr::Ascii("profirles"), &["profiles"]),
            (dictgen::InsensitiveStr::Ascii("priortity"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("survivabilty"),
                &["survivability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organistion"),
                &["organisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overshadowd"),
                &["overshadowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capatilism"),
                &["capitalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("dpendency"), &["dependency"]),
            (dictgen::InsensitiveStr::Ascii("freee"), &["free"]),
            (
                dictgen::InsensitiveStr::Ascii("deseralisation"),
                &["deserialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mebran"), &["membrane"]),
            (
                dictgen::InsensitiveStr::Ascii("disproportianately"),
                &["disproportionately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("techncially"),
                &["technically"],
            ),
            (dictgen::InsensitiveStr::Ascii("reister"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("initliazer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foriegners"),
                &["foreigners"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accuratelly"),
                &["accurately"],
            ),
            (dictgen::InsensitiveStr::Ascii("srcipt"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("primarely"), &["primarily"]),
            (
                dictgen::InsensitiveStr::Ascii("acitivities"),
                &["activities"],
            ),
            (dictgen::InsensitiveStr::Ascii("compaire"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("constitutent"),
                &["constituent"],
            ),
            (dictgen::InsensitiveStr::Ascii("appaer"), &["appear"]),
            (
                dictgen::InsensitiveStr::Ascii("comporessers"),
                &["compressors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantitization"),
                &["quantization"],
            ),
            (dictgen::InsensitiveStr::Ascii("incliding"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("corrsponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("belog"), &["belong"]),
            (dictgen::InsensitiveStr::Ascii("asnwered"), &["answered"]),
            (
                dictgen::InsensitiveStr::Ascii("improvemenets"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operasional"),
                &["operational"],
            ),
            (dictgen::InsensitiveStr::Ascii("sidebaord"), &["sideboard"]),
            (dictgen::InsensitiveStr::Ascii("afaraid"), &["afraid"]),
            (dictgen::InsensitiveStr::Ascii("conquerd"), &["conquered"]),
            (dictgen::InsensitiveStr::Ascii("suppot"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("intolerence"),
                &["intolerance"],
            ),
            (dictgen::InsensitiveStr::Ascii("schdules"), &["schedules"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegatiotiation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcions"),
                &["corrections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimisticly"),
                &["optimistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("dystopain"), &["dystopian"]),
            (
                dictgen::InsensitiveStr::Ascii("sysmograph"),
                &["seismograph"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefferred"), &["preferred"]),
            (
                dictgen::InsensitiveStr::Ascii("independetly"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("islamiskt"), &["islamist"]),
            (
                dictgen::InsensitiveStr::Ascii("adventuous"),
                &["adventurous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qualitification"),
                &["qualification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoroporated"),
                &["incorporated"],
            ),
            (dictgen::InsensitiveStr::Ascii("conncets"), &["connects"]),
            (dictgen::InsensitiveStr::Ascii("tpe"), &["type"]),
            (
                dictgen::InsensitiveStr::Ascii("evelopes"),
                &["envelopes", "envelops"],
            ),
            (dictgen::InsensitiveStr::Ascii("remappin"), &["remapping"]),
            (dictgen::InsensitiveStr::Ascii("temparate"), &["temperate"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstod"),
                &["misunderstood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derepecated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("termimal"), &["terminal"]),
            (
                dictgen::InsensitiveStr::Ascii("farwarding"),
                &["forwarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("hilight"), &["highlight"]),
            (dictgen::InsensitiveStr::Ascii("transator"), &["translator"]),
            (
                dictgen::InsensitiveStr::Ascii("cathegories"),
                &["categories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupliacating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("descritpr"), &["descriptor"]),
            (dictgen::InsensitiveStr::Ascii("prescripe"), &["prescribe"]),
            (dictgen::InsensitiveStr::Ascii("contrlled"), &["controlled"]),
            (dictgen::InsensitiveStr::Ascii("mediphor"), &["metaphor"]),
            (dictgen::InsensitiveStr::Ascii("memwoirs"), &["memoirs"]),
            (dictgen::InsensitiveStr::Ascii("sponsores"), &["sponsors"]),
            (
                dictgen::InsensitiveStr::Ascii("legionaires"),
                &["legionnaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppurtunites"),
                &["opportunities"],
            ),
            (dictgen::InsensitiveStr::Ascii("seters"), &["setters"]),
            (dictgen::InsensitiveStr::Ascii("mismatich"), &["mismatch"]),
            (dictgen::InsensitiveStr::Ascii("specifcs"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("screewed"), &["screwed"]),
            (dictgen::InsensitiveStr::Ascii("memort"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("encrcypts"), &["encrypts"]),
            (
                dictgen::InsensitiveStr::Ascii("decsiptors"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaibles"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("demmangled"), &["demangled"]),
            (dictgen::InsensitiveStr::Ascii("devloved"), &["devolved"]),
            (dictgen::InsensitiveStr::Ascii("ineritor"), &["inheritor"]),
            (
                dictgen::InsensitiveStr::Ascii("prepherial"),
                &["peripheral"],
            ),
            (dictgen::InsensitiveStr::Ascii("emabled"), &["enabled"]),
            (dictgen::InsensitiveStr::Ascii("colunns"), &["columns"]),
            (dictgen::InsensitiveStr::Ascii("stupidy"), &["stupidly"]),
            (
                dictgen::InsensitiveStr::Ascii("unprofessionnal"),
                &["unprofessional"],
            ),
            (dictgen::InsensitiveStr::Ascii("joruneys"), &["journeys"]),
            (
                dictgen::InsensitiveStr::Ascii("determinisic"),
                &["deterministic"],
            ),
            (dictgen::InsensitiveStr::Ascii("juxtify"), &["justify"]),
            (dictgen::InsensitiveStr::Ascii("nowdays"), &["nowadays"]),
            (
                dictgen::InsensitiveStr::Ascii("substentially"),
                &["substantially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maniputions"),
                &["manipulations"],
            ),
            (dictgen::InsensitiveStr::Ascii("mthods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("feld"), &["field"]),
            (dictgen::InsensitiveStr::Ascii("uniocde"), &["unicode"]),
            (
                dictgen::InsensitiveStr::Ascii("confiugring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimenions"), &["dimensions"]),
            (dictgen::InsensitiveStr::Ascii("tiwards"), &["towards"]),
            (
                dictgen::InsensitiveStr::Ascii("unproducive"),
                &["unproductive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partialially"),
                &["partially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rever"),
                &["revert", "refer", "fever"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intepretations"),
                &["interpretations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("competitioners"),
                &["competitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operationable"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tradititionally"),
                &["traditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("perfome"), &["perform"]),
            (dictgen::InsensitiveStr::Ascii("calulate"), &["calculate"]),
            (dictgen::InsensitiveStr::Ascii("dscrete"), &["discrete"]),
            (dictgen::InsensitiveStr::Ascii("paylaod"), &["payload"]),
            (
                dictgen::InsensitiveStr::Ascii("exilerated"),
                &["exhilarated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preductive"),
                &["predictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("settter"), &["setter"]),
            (dictgen::InsensitiveStr::Ascii("varibel"), &["variable"]),
            (
                dictgen::InsensitiveStr::Ascii("personilized"),
                &["personalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("alochol"), &["alcohol"]),
            (
                dictgen::InsensitiveStr::Ascii("tradtional"),
                &["traditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("plese"), &["please"]),
            (dictgen::InsensitiveStr::Ascii("sytled"), &["styled"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilyze"),
                &["cannibalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscairrage"),
                &["miscarriage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbreviatin"),
                &["abbreviation", "abbreviating"],
            ),
            (dictgen::InsensitiveStr::Ascii("entereing"), &["entering"]),
            (dictgen::InsensitiveStr::Ascii("coerscible"), &["coercible"]),
            (dictgen::InsensitiveStr::Ascii("arrenged"), &["arranged"]),
            (dictgen::InsensitiveStr::Ascii("cartisian"), &["cartesian"]),
            (dictgen::InsensitiveStr::Ascii("pronunce"), &["pronounce"]),
            (dictgen::InsensitiveStr::Ascii("hybrides"), &["hybrids"]),
            (dictgen::InsensitiveStr::Ascii("ofcoures"), &["ofcourse"]),
            (dictgen::InsensitiveStr::Ascii("helpfull"), &["helpful"]),
            (
                dictgen::InsensitiveStr::Ascii("compleletey"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrices"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("trasnlator"),
                &["translator"],
            ),
            (dictgen::InsensitiveStr::Ascii("inidicator"), &["indicator"]),
            (
                dictgen::InsensitiveStr::Ascii("duplciating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatsets"), &["datasets"]),
            (
                dictgen::InsensitiveStr::Ascii("conrollers"),
                &["controllers"],
            ),
            (dictgen::InsensitiveStr::Ascii("increasd"), &["increased"]),
            (
                dictgen::InsensitiveStr::Ascii("respectfuly"),
                &["respectfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("roduceer"), &["producer"]),
            (dictgen::InsensitiveStr::Ascii("arbitarary"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("asssignments"),
                &["assignments"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodcuing"), &["producing"]),
            (
                dictgen::InsensitiveStr::Ascii("confugirble"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("blashpemy"), &["blasphemy"]),
            (dictgen::InsensitiveStr::Ascii("grandeeos"), &["grandiose"]),
            (
                dictgen::InsensitiveStr::Ascii("psychiatrits"),
                &["psychiatrists"],
            ),
            (dictgen::InsensitiveStr::Ascii("infalted"), &["inflated"]),
            (dictgen::InsensitiveStr::Ascii("fullfill"), &["fulfill"]),
            (
                dictgen::InsensitiveStr::Ascii("favourties"),
                &["favourites"],
            ),
            (dictgen::InsensitiveStr::Ascii("synonom"), &["synonym"]),
            (dictgen::InsensitiveStr::Ascii("caliming"), &["claiming"]),
            (
                dictgen::InsensitiveStr::Ascii("implemencted"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("rouglhy"), &["roughly"]),
            (
                dictgen::InsensitiveStr::Ascii("correktions"),
                &["corrections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overzealus"),
                &["overzealous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimeters"),
                &["experimenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("resason"), &["reason"]),
            (
                dictgen::InsensitiveStr::Ascii("instatutionalized"),
                &["institutionalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("filaname"), &["filename"]),
            (
                dictgen::InsensitiveStr::Ascii("distatesful"),
                &["distasteful"],
            ),
            (dictgen::InsensitiveStr::Ascii("regonize"), &["recognize"]),
            (
                dictgen::InsensitiveStr::Ascii("admissable"),
                &["admissible"],
            ),
            (dictgen::InsensitiveStr::Ascii("realations"), &["relations"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoweldges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("counterpath"),
                &["counterpart"],
            ),
            (dictgen::InsensitiveStr::Ascii("guaruntee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("allreight"), &["alright"]),
            (
                dictgen::InsensitiveStr::Ascii("civilazation"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("vulnerble"), &["vulnerable"]),
            (dictgen::InsensitiveStr::Ascii("humanite"), &["humanities"]),
            (
                dictgen::InsensitiveStr::Ascii("simplifiyng"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("specfic"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("loacted"), &["located"]),
            (dictgen::InsensitiveStr::Ascii("clearling"), &["clearing"]),
            (
                dictgen::InsensitiveStr::Ascii("notificaiton"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exractions"),
                &["extractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivuals"),
                &["individuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("updrage"), &["upgrade"]),
            (
                dictgen::InsensitiveStr::Ascii("countr"),
                &["counter", "contour", "country", "county"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulitiplication"),
                &["multiplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootsram"), &["bootram"]),
            (dictgen::InsensitiveStr::Ascii("attrubtes"), &["attributes"]),
            (dictgen::InsensitiveStr::Ascii("tather"), &["rather"]),
            (dictgen::InsensitiveStr::Ascii("recommnd"), &["recommend"]),
            (
                dictgen::InsensitiveStr::Ascii("composuted"),
                &["composited"],
            ),
            (dictgen::InsensitiveStr::Ascii("shareed"), &["shared"]),
            (
                dictgen::InsensitiveStr::Ascii("unpresedented"),
                &["unprecedented"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuncate"), &["truncate"]),
            (
                dictgen::InsensitiveStr::Ascii("rendesvous"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("certiainly"), &["certainly"]),
            (dictgen::InsensitiveStr::Ascii("undefuned"), &["undefined"]),
            (dictgen::InsensitiveStr::Ascii("unrelatd"), &["unrelated"]),
            (
                dictgen::InsensitiveStr::Ascii("generatios"),
                &["generators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priviledging"),
                &["privileging"],
            ),
            (dictgen::InsensitiveStr::Ascii("aftewards"), &["afterwards"]),
            (dictgen::InsensitiveStr::Ascii("brievety"), &["brevity"]),
            (dictgen::InsensitiveStr::Ascii("slooths"), &["sleuths"]),
            (dictgen::InsensitiveStr::Ascii("resquests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("alocates"), &["allocates"]),
            (dictgen::InsensitiveStr::Ascii("namepsace"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("esiest"), &["easiest"]),
            (
                dictgen::InsensitiveStr::Ascii("resulutions"),
                &["resolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkklists"),
                &["checklists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singnalling"),
                &["signalling"],
            ),
            (dictgen::InsensitiveStr::Ascii("witrh"), &["with"]),
            (dictgen::InsensitiveStr::Ascii("anyay"), &["anyway"]),
            (
                dictgen::InsensitiveStr::Ascii("bannana"),
                &["banana", "bandana"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignormed"), &["ignored"]),
            (
                dictgen::InsensitiveStr::Ascii("pressume"),
                &["presume", "pressure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorthyns"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antiquaited"),
                &["antiquated"],
            ),
            (dictgen::InsensitiveStr::Ascii("shpped"), &["shipped"]),
            (dictgen::InsensitiveStr::Ascii("exuberent"), &["exuberant"]),
            (dictgen::InsensitiveStr::Ascii("conencts"), &["connects"]),
            (
                dictgen::InsensitiveStr::Ascii("challeneged"),
                &["challenged"],
            ),
            (dictgen::InsensitiveStr::Ascii("privide"), &["provide"]),
            (
                dictgen::InsensitiveStr::Ascii("atrributes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("varieables"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("optimizacion"),
                &["optimization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denominaton"),
                &["denomination"],
            ),
            (dictgen::InsensitiveStr::Ascii("distractiv"), &["district"]),
            (
                dictgen::InsensitiveStr::Ascii("contredict"),
                &["contradict"],
            ),
            (dictgen::InsensitiveStr::Ascii("documet"), &["document"]),
            (
                dictgen::InsensitiveStr::Ascii("trascations"),
                &["transactions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propencity"),
                &["propensity"],
            ),
            (dictgen::InsensitiveStr::Ascii("actusally"), &["actually"]),
            (dictgen::InsensitiveStr::Ascii("satisified"), &["satisfied"]),
            (dictgen::InsensitiveStr::Ascii("ellipce"), &["ellipse"]),
            (dictgen::InsensitiveStr::Ascii("opimised"), &["optimised"]),
            (dictgen::InsensitiveStr::Ascii("paramenet"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("expination"),
                &["expiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("folwos"), &["follows"]),
            (dictgen::InsensitiveStr::Ascii("mathc"), &["match"]),
            (dictgen::InsensitiveStr::Ascii("loosing"), &["losing"]),
            (
                dictgen::InsensitiveStr::Ascii("impletations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("expliot"), &["exploit"]),
            (dictgen::InsensitiveStr::Ascii("legngths"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("experimnter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("rediect"), &["redirect"]),
            (dictgen::InsensitiveStr::Ascii("dimonds"), &["diamonds"]),
            (dictgen::InsensitiveStr::Ascii("exchnges"), &["exchanges"]),
            (
                dictgen::InsensitiveStr::Ascii("equivelence"),
                &["equivalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fileshystems"),
                &["filesystems"],
            ),
            (dictgen::InsensitiveStr::Ascii("propoer"), &["proper"]),
            (
                dictgen::InsensitiveStr::Ascii("exlamation"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("besdies"), &["besides"]),
            (dictgen::InsensitiveStr::Ascii("degbugging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("diablical"), &["diabolical"]),
            (dictgen::InsensitiveStr::Ascii("etiher"), &["either"]),
            (
                dictgen::InsensitiveStr::Ascii("maybellene"),
                &["maybelline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pereptually"),
                &["perpetually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsers"),
                &["responders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maybelliene"),
                &["maybelline"],
            ),
            (dictgen::InsensitiveStr::Ascii("visious"), &["vicious"]),
            (
                dictgen::InsensitiveStr::Ascii("sepatately"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncannny"), &["uncanny"]),
            (dictgen::InsensitiveStr::Ascii("spreaded"), &["spread"]),
            (
                dictgen::InsensitiveStr::Ascii("completin"),
                &["completion", "completing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("somethintg"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("solidier"), &["soldier"]),
            (dictgen::InsensitiveStr::Ascii("ensalve"), &["enslave"]),
            (dictgen::InsensitiveStr::Ascii("loacl"), &["local"]),
            (
                dictgen::InsensitiveStr::Ascii("infogrpahic"),
                &["infographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descritption"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("newancing"), &["nuancing"]),
            (
                dictgen::InsensitiveStr::Ascii("canonivalized"),
                &["canonicalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemtations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interectioned"),
                &["interaction", "intersection"],
            ),
            (dictgen::InsensitiveStr::Ascii("apoligies"), &["apologies"]),
            (
                dictgen::InsensitiveStr::Ascii("calibrtion"),
                &["calibration"],
            ),
            (dictgen::InsensitiveStr::Ascii("addresds"), &["address"]),
            (dictgen::InsensitiveStr::Ascii("correcing"), &["correcting"]),
            (
                dictgen::InsensitiveStr::Ascii("escated"),
                &["escalated", "escaped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accodomation"),
                &["accommodation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catostraphically"),
                &["catastrophically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapporve"),
                &["disapprove"],
            ),
            (dictgen::InsensitiveStr::Ascii("staps"), &["steps", "stops"]),
            (dictgen::InsensitiveStr::Ascii("cahe"), &["cache", "ache"]),
            (
                dictgen::InsensitiveStr::Ascii("intelectual"),
                &["intellectual"],
            ),
            (dictgen::InsensitiveStr::Ascii("annualy"), &["annually"]),
            (
                dictgen::InsensitiveStr::Ascii("spefcifically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("objetives"), &["objectives"]),
            (dictgen::InsensitiveStr::Ascii("corosion"), &["corrosion"]),
            (dictgen::InsensitiveStr::Ascii("recahed"), &["reached"]),
            (
                dictgen::InsensitiveStr::Ascii("seperatally"),
                &["separately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envirment"),
                &["environment"],
            ),
            (dictgen::InsensitiveStr::Ascii("experimen"), &["experiment"]),
            (
                dictgen::InsensitiveStr::Ascii("desintations"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("wuthin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("sytemmemory"),
                &["systemmemory"],
            ),
            (dictgen::InsensitiveStr::Ascii("attendat"), &["attendant"]),
            (
                dictgen::InsensitiveStr::Ascii("dimmensioning"),
                &["dimensioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("striaghten"),
                &["straighten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cofiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("nativelly"), &["natively"]),
            (
                dictgen::InsensitiveStr::Ascii("offser"),
                &["offset", "offer"],
            ),
            (dictgen::InsensitiveStr::Ascii("obervers"), &["observers"]),
            (
                dictgen::InsensitiveStr::Ascii("operatation"),
                &["operation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnicator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("acornyms"), &["acronyms"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("satisty"), &["satisfy"]),
            (dictgen::InsensitiveStr::Ascii("vitailty"), &["vitality"]),
            (dictgen::InsensitiveStr::Ascii("ontains"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("recongifure"),
                &["reconfigure"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorothm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("markeplaces"),
                &["marketplaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherid"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("oligarcy"), &["oligarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("inizialised"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceptible"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("asutrian"), &["austrian"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfities"),
                &["identifies", "identities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summurizes"),
                &["summarizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hospitible"),
                &["hospitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("procides"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("befirend"), &["befriend"]),
            (dictgen::InsensitiveStr::Ascii("apyooning"), &["oppugning"]),
            (
                dictgen::InsensitiveStr::Ascii("interferisce"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsituding"),
                &["substituting"],
            ),
            (dictgen::InsensitiveStr::Ascii("retriebe"), &["retrieve"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotaiation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifics"),
                &["specifics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("beauracratically"),
                &["bureaucratically"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaliable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("aestethically"),
                &["aesthetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispathing"),
                &["dispatching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisins"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meachnisms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lew"),
                &["dew", "hew", "lieu", "sew"],
            ),
            (dictgen::InsensitiveStr::Ascii("fictious"), &["fictitious"]),
            (dictgen::InsensitiveStr::Ascii("furuther"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("geenrational"),
                &["generational"],
            ),
            (dictgen::InsensitiveStr::Ascii("juli"), &["july"]),
            (dictgen::InsensitiveStr::Ascii("clonne"), &["clone"]),
            (dictgen::InsensitiveStr::Ascii("printerest"), &["printers"]),
            (dictgen::InsensitiveStr::Ascii("prodicing"), &["producing"]),
            (dictgen::InsensitiveStr::Ascii("ceating"), &["creating"]),
            (dictgen::InsensitiveStr::Ascii("physisan"), &["physician"]),
            (
                dictgen::InsensitiveStr::Ascii("underyling"),
                &["underlying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perlciritc"),
                &["perlcritic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assingnment"),
                &["assignment"],
            ),
            (dictgen::InsensitiveStr::Ascii("examime"), &["examine"]),
            (
                dictgen::InsensitiveStr::Ascii("centerfuges"),
                &["centrifuges"],
            ),
            (dictgen::InsensitiveStr::Ascii("destryoer"), &["destroyer"]),
            (dictgen::InsensitiveStr::Ascii("travesal"), &["traversal"]),
            (dictgen::InsensitiveStr::Ascii("brutailty"), &["brutality"]),
            (dictgen::InsensitiveStr::Ascii("tunges"), &["tongues"]),
            (
                dictgen::InsensitiveStr::Ascii("fequencies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("terriblely"), &["terribly"]),
            (dictgen::InsensitiveStr::Ascii("egregrious"), &["egregious"]),
            (dictgen::InsensitiveStr::Ascii("excemption"), &["exemption"]),
            (
                dictgen::InsensitiveStr::Ascii("persence"),
                &["presence", "persistence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacialize"),
                &["specialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efffectiveness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interfearence"),
                &["interference"],
            ),
            (dictgen::InsensitiveStr::Ascii("usueful"), &["useful"]),
            (
                dictgen::InsensitiveStr::Ascii("unreacahble"),
                &["unreachable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimistacally"),
                &["optimistically"],
            ),
            (dictgen::InsensitiveStr::Ascii("seives"), &["sieves"]),
            (
                dictgen::InsensitiveStr::Ascii("parallalized"),
                &["parallelized"],
            ),
            (dictgen::InsensitiveStr::Ascii("monumet"), &["monument"]),
            (dictgen::InsensitiveStr::Ascii("fashists"), &["fascists"]),
            (dictgen::InsensitiveStr::Ascii("folwoind"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("dosclosing"),
                &["disclosing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standrdizing"),
                &["standardizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("configare"), &["configure"]),
            (
                dictgen::InsensitiveStr::Ascii("presons"),
                &["persons", "prisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("persistens"), &["persists"]),
            (
                dictgen::InsensitiveStr::Ascii("legislativo"),
                &["legislation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatibiliy"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("scketching"), &["sketching"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprehensable"),
                &["incomprehensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initations"),
                &["imitations", "initiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphsyical"),
                &["metaphysical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repores"),
                &["reports", "reposes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anualizing"),
                &["annualizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimentaat"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimets"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("organicaly"),
                &["organically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transplate"),
                &["transplant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("understanting"),
                &["understanding", "understating"],
            ),
            (dictgen::InsensitiveStr::Ascii("preffixes"), &["prefixes"]),
            (dictgen::InsensitiveStr::Ascii("analysy"), &["analysis"]),
            (dictgen::InsensitiveStr::Ascii("sleepp"), &["sleep"]),
            (dictgen::InsensitiveStr::Ascii("apporoving"), &["approving"]),
            (
                dictgen::InsensitiveStr::Ascii("cathegorises"),
                &["categorises"],
            ),
            (dictgen::InsensitiveStr::Ascii("montaban"), &["montana"]),
            (dictgen::InsensitiveStr::Ascii("buiding"), &["building"]),
            (dictgen::InsensitiveStr::Ascii("somthign"), &["something"]),
            (dictgen::InsensitiveStr::Ascii("algoroithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("succeession"),
                &["succession"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("succsession"),
                &["succession"],
            ),
            (dictgen::InsensitiveStr::Ascii("cmmand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("defnietly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("blatimore"), &["baltimore"]),
            (
                dictgen::InsensitiveStr::Ascii("oll"),
                &["all", "ole", "old", "olly", "oil"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delcarators"),
                &["declarators"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyperbel"), &["hyperbole"]),
            (
                dictgen::InsensitiveStr::Ascii("ecouragement"),
                &["encouragement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anpatheater"),
                &["amphitheater"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desctructed"),
                &["destructed"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempaltes"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("unanmious"), &["unanimous"]),
            (dictgen::InsensitiveStr::Ascii("transiton"), &["transition"]),
            (
                dictgen::InsensitiveStr::Ascii("cancelllation"),
                &["cancellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("benjed"), &["binged"]),
            (
                dictgen::InsensitiveStr::Ascii("interepreter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimetricaly"),
                &["symmetrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prequisite"),
                &["prerequisite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("viusalizing"),
                &["visualizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("disucsses"), &["discusses"]),
            (dictgen::InsensitiveStr::Ascii("rezonin"), &["rezoning"]),
            (
                dictgen::InsensitiveStr::Ascii("phialdelphia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conenctivities"),
                &["connectivities"],
            ),
            (dictgen::InsensitiveStr::Ascii("abruptley"), &["abruptly"]),
            (
                dictgen::InsensitiveStr::Ascii("derivatvies"),
                &["derivatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trancriptions"),
                &["transcriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elimiation"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acedemically"),
                &["academically"],
            ),
            (dictgen::InsensitiveStr::Ascii("frontapge"), &["frontpage"]),
            (
                dictgen::InsensitiveStr::Ascii("reccurrence"),
                &["recurrence"],
            ),
            (dictgen::InsensitiveStr::Ascii("heridity"), &["heredity"]),
            (dictgen::InsensitiveStr::Ascii("deisgns"), &["designs"]),
            (
                dictgen::InsensitiveStr::Ascii("segregaded"),
                &["segregated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrutable"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribue"),
                &["distributed", "distribute"],
            ),
            (dictgen::InsensitiveStr::Ascii("etcc"), &["etc"]),
            (
                dictgen::InsensitiveStr::Ascii("servived"),
                &["serviced", "survived"],
            ),
            (dictgen::InsensitiveStr::Ascii("teetotler"), &["teetotaler"]),
            (dictgen::InsensitiveStr::Ascii("popuplates"), &["populates"]),
            (
                dictgen::InsensitiveStr::Ascii("ofter"),
                &["after", "offer", "often"],
            ),
            (dictgen::InsensitiveStr::Ascii("gausian"), &["gaussian"]),
            (dictgen::InsensitiveStr::Ascii("arcitects"), &["architects"]),
            (
                dictgen::InsensitiveStr::Ascii("higlighter"),
                &["highlighter"],
            ),
            (dictgen::InsensitiveStr::Ascii("dealed"), &["dealt"]),
            (
                dictgen::InsensitiveStr::Ascii("additianal"),
                &["additional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotitators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("bianry"), &["binary"]),
            (
                dictgen::InsensitiveStr::Ascii("unconvenient"),
                &["inconvenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connitations"),
                &["connotations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconditianaly"),
                &["unconditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traingulation"),
                &["triangulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convience"),
                &["convince", "convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brotherwood"),
                &["brotherhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("realibility"),
                &["reliability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precautios"),
                &["precautions"],
            ),
            (dictgen::InsensitiveStr::Ascii("childresn"), &["childrens"]),
            (dictgen::InsensitiveStr::Ascii("suppres"), &["suppress"]),
            (dictgen::InsensitiveStr::Ascii("initails"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("civalisation"),
                &["civilisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinination"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("delayis"), &["delays"]),
            (dictgen::InsensitiveStr::Ascii("copmose"), &["compose"]),
            (
                dictgen::InsensitiveStr::Ascii("associtions"),
                &["associations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multipler"),
                &["multiplier", "multiple"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbrievation"),
                &["abbreviation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supervsior"),
                &["supervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interecptor"),
                &["interceptor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condmening"),
                &["condemning"],
            ),
            (dictgen::InsensitiveStr::Ascii("tpyed"), &["typed"]),
            (dictgen::InsensitiveStr::Ascii("controlls"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("embeded"), &["embedded"]),
            (dictgen::InsensitiveStr::Ascii("messags"), &["messages"]),
            (
                dictgen::InsensitiveStr::Ascii("justifaction"),
                &["justification"],
            ),
            (dictgen::InsensitiveStr::Ascii("valud"), &["valid", "value"]),
            (dictgen::InsensitiveStr::Ascii("fundation"), &["foundation"]),
            (
                dictgen::InsensitiveStr::Ascii("soldiarity"),
                &["solidarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calibraton"),
                &["calibration"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferals"), &["referrals"]),
            (
                dictgen::InsensitiveStr::Ascii("monomophization"),
                &["monomorphization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("normalysed"),
                &["normalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("iteraing"), &["iterating"]),
            (dictgen::InsensitiveStr::Ascii("classfied"), &["classified"]),
            (
                dictgen::InsensitiveStr::Ascii("diplimatic"),
                &["diplomatic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optmisations"),
                &["optimisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("forot"), &["forgot"]),
            (
                dictgen::InsensitiveStr::Ascii("particullary"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("paragragh"), &["paragraph"]),
            (dictgen::InsensitiveStr::Ascii("mimatchd"), &["mismatched"]),
            (
                dictgen::InsensitiveStr::Ascii("microtranscations"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangets"), &["magnets"]),
            (dictgen::InsensitiveStr::Ascii("acutally"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("appling"),
                &["applying", "appalling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presumibly"),
                &["presumably"],
            ),
            (dictgen::InsensitiveStr::Ascii("moning"), &["morning"]),
            (
                dictgen::InsensitiveStr::Ascii("roduce"),
                &["reduce", "produce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conributed"),
                &["contributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("throttlin"), &["throttling"]),
            (
                dictgen::InsensitiveStr::Ascii("acocunting"),
                &["accounting"],
            ),
            (dictgen::InsensitiveStr::Ascii("disspiate"), &["dissipate"]),
            (
                dictgen::InsensitiveStr::Ascii("comparisations"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("passwirds"), &["passwords"]),
            (dictgen::InsensitiveStr::Ascii("ablity"), &["ability"]),
            (
                dictgen::InsensitiveStr::Ascii("throughtout"),
                &["throughout"],
            ),
            (dictgen::InsensitiveStr::Ascii("incidentes"), &["incidents"]),
            (dictgen::InsensitiveStr::Ascii("fromats"), &["formats"]),
            (
                dictgen::InsensitiveStr::Ascii("comperator"),
                &["comparator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interferce"),
                &["interferes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sanbdox"), &["sandbox"]),
            (
                dictgen::InsensitiveStr::Ascii("geographich"),
                &["geographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("windoows"), &["windows"]),
            (
                dictgen::InsensitiveStr::Ascii("visuably"),
                &["visually", "visibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mulittasking"),
                &["multitasking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumsied"),
                &["circumcised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplaicated"),
                &["deduplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("hexgaon"), &["hexagon"]),
            (dictgen::InsensitiveStr::Ascii("vluation"), &["valuation"]),
            (
                dictgen::InsensitiveStr::Ascii("accompianed"),
                &["accompanied"],
            ),
            (dictgen::InsensitiveStr::Ascii("apllying"), &["applying"]),
            (dictgen::InsensitiveStr::Ascii("systesms"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("trianglutaion"),
                &["triangulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allcoaters"),
                &["allocators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsability"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("jion"), &["join"]),
            (dictgen::InsensitiveStr::Ascii("wresseling"), &["wrestling"]),
            (
                dictgen::InsensitiveStr::Ascii("appraoched"),
                &["approached"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimnetation"),
                &["experimentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proffessors"),
                &["professors"],
            ),
            (dictgen::InsensitiveStr::Ascii("witdh"), &["width"]),
            (
                dictgen::InsensitiveStr::Ascii("decsriptions"),
                &["descriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delearships"),
                &["dealerships"],
            ),
            (dictgen::InsensitiveStr::Ascii("recconeect"), &["reconnect"]),
            (dictgen::InsensitiveStr::Ascii("delimitied"), &["delimited"]),
            (
                dictgen::InsensitiveStr::Ascii("injectiosn"),
                &["injections"],
            ),
            (dictgen::InsensitiveStr::Ascii("simplfied"), &["simplified"]),
            (dictgen::InsensitiveStr::Ascii("anxeity"), &["anxiety"]),
            (dictgen::InsensitiveStr::Ascii("libertae"), &["liberate"]),
            (dictgen::InsensitiveStr::Ascii("extenal"), &["external"]),
            (
                dictgen::InsensitiveStr::Ascii("unfortunatley"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("misusin"), &["misusing"]),
            (dictgen::InsensitiveStr::Ascii("avcoid"), &["avoid"]),
            (
                dictgen::InsensitiveStr::Ascii("generilize"),
                &["generalize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsidizng"),
                &["subsidizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("anualize"), &["annualize"]),
            (
                dictgen::InsensitiveStr::Ascii("unexspectedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("pysical"), &["physical"]),
            (dictgen::InsensitiveStr::Ascii("stardard"), &["standard"]),
            (
                dictgen::InsensitiveStr::Ascii("masturabting"),
                &["masturbating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("jouranlists"),
                &["journalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("rorations"), &["rotations"]),
            (dictgen::InsensitiveStr::Ascii("dicide"), &["decide"]),
            (
                dictgen::InsensitiveStr::Ascii("documenatation"),
                &["documentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("placemenet"),
                &["placements", "placement"],
            ),
            (dictgen::InsensitiveStr::Ascii("spactator"), &["spectator"]),
            (dictgen::InsensitiveStr::Ascii("guarnted"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("implemantations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthousiasm"),
                &["enthusiasm"],
            ),
            (dictgen::InsensitiveStr::Ascii("extit"), &["exit"]),
            (
                dictgen::InsensitiveStr::Ascii("equivilence"),
                &["equivalence"],
            ),
            (dictgen::InsensitiveStr::Ascii("replicaite"), &["replicate"]),
            (
                dictgen::InsensitiveStr::Ascii("norifications"),
                &["notifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("raparse"), &["reparse"]),
            (dictgen::InsensitiveStr::Ascii("cranbery"), &["cranberry"]),
            (
                dictgen::InsensitiveStr::Ascii("compilance"),
                &["compliance"],
            ),
            (dictgen::InsensitiveStr::Ascii("hampster"), &["hamster"]),
            (
                dictgen::InsensitiveStr::Ascii("activits"),
                &["activities", "activists"],
            ),
            (dictgen::InsensitiveStr::Ascii("qulifier"), &["qualifier"]),
            (dictgen::InsensitiveStr::Ascii("presearve"), &["preserve"]),
            (dictgen::InsensitiveStr::Ascii("retirever"), &["retriever"]),
            (dictgen::InsensitiveStr::Ascii("calculatd"), &["calculated"]),
            (
                dictgen::InsensitiveStr::Ascii("incompataible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("grigorian"), &["gregorian"]),
            (
                dictgen::InsensitiveStr::Ascii("intersepts"),
                &["intercepts", "intersteps"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incomfort"),
                &["discomfort", "uncomfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insensistively"),
                &["insensitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("resetart"), &["restart"]),
            (
                dictgen::InsensitiveStr::Ascii("chipersuite"),
                &["ciphersuite"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecxite"), &["excite"]),
            (
                dictgen::InsensitiveStr::Ascii("motherobard"),
                &["motherboards"],
            ),
            (dictgen::InsensitiveStr::Ascii("lenghts"), &["lengths"]),
            (
                dictgen::InsensitiveStr::Ascii("destnation"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confiuguring"),
                &["configuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecitcut"),
                &["connecticut"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arrangmenet"),
                &["arrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformarse"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degridation"),
                &["degradation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sotfware"), &["software"]),
            (
                dictgen::InsensitiveStr::Ascii("anthromorphization"),
                &["anthropomorphization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definatlly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("annotes"), &["annotates"]),
            (
                dictgen::InsensitiveStr::Ascii("precende"),
                &["precedence", "precede"],
            ),
            (dictgen::InsensitiveStr::Ascii("dormtund"), &["dortmund"]),
            (dictgen::InsensitiveStr::Ascii("ananlyse"), &["analyse"]),
            (
                dictgen::InsensitiveStr::Ascii("exauhsting"),
                &["exhausting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invokation"),
                &["invocation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("execuctions"),
                &["executions"],
            ),
            (dictgen::InsensitiveStr::Ascii("respnds"), &["responds"]),
            (dictgen::InsensitiveStr::Ascii("happliy"), &["happily"]),
            (
                dictgen::InsensitiveStr::Ascii("paramemters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("zionistas"), &["zionists"]),
            (
                dictgen::InsensitiveStr::Ascii("unexpestedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overlappping"),
                &["overlapping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standrdises"),
                &["standardises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interstallar"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("percpetions"),
                &["perceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("cushins"), &["cushions"]),
            (dictgen::InsensitiveStr::Ascii("destrory"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("modiified"), &["modified"]),
            (
                dictgen::InsensitiveStr::Ascii("reconenction"),
                &["reconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppening"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("seaches"),
                &["searches", "reaches", "teaches", "beaches"],
            ),
            (dictgen::InsensitiveStr::Ascii("biyou"), &["bayou"]),
            (dictgen::InsensitiveStr::Ascii("settiing"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("defintite"), &["definite"]),
            (dictgen::InsensitiveStr::Ascii("inflatabe"), &["inflatable"]),
            (dictgen::InsensitiveStr::Ascii("delimiited"), &["delimited"]),
            (dictgen::InsensitiveStr::Ascii("requilme"), &["requiem"]),
            (dictgen::InsensitiveStr::Ascii("increaing"), &["increasing"]),
            (
                dictgen::InsensitiveStr::Ascii("similerity"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("curtian"), &["curtain"]),
            (
                dictgen::InsensitiveStr::Ascii("perpetraded"),
                &["perpetrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("exeqution"), &["execution"]),
            (dictgen::InsensitiveStr::Ascii("assihns"), &["assigns"]),
            (
                dictgen::InsensitiveStr::Ascii("adevnturers"),
                &["adventurers"],
            ),
            (dictgen::InsensitiveStr::Ascii("premeired"), &["premiered"]),
            (dictgen::InsensitiveStr::Ascii("viariable"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("momemt"), &["moment"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusivily"),
                &["exclusivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speads"),
                &["spreads", "speeds", "spends", "speaks", "spears"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrection"),
                &["discretion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbcriptions"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("garilla"), &["guerilla"]),
            (dictgen::InsensitiveStr::Ascii("yeras"), &["years"]),
            (
                dictgen::InsensitiveStr::Ascii("amphetmaines"),
                &["amphetamines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snadwiches"),
                &["sandwiches"],
            ),
            (dictgen::InsensitiveStr::Ascii("goodluk"), &["goodluck"]),
            (dictgen::InsensitiveStr::Ascii("acutually"), &["actually"]),
            (
                dictgen::InsensitiveStr::Ascii("surpirously"),
                &["spuriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("bector"), &["vector"]),
            (
                dictgen::InsensitiveStr::Ascii("comparrators"),
                &["comparators"],
            ),
            (dictgen::InsensitiveStr::Ascii("activily"), &["actively"]),
            (
                dictgen::InsensitiveStr::Ascii("repoting"),
                &["reporting", "reposting", "reputing", "repeating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("committtees"),
                &["committees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unanimuously"),
                &["unanimously"],
            ),
            (dictgen::InsensitiveStr::Ascii("utpoian"), &["utopian"]),
            (
                dictgen::InsensitiveStr::Ascii("prefere"),
                &["prefer", "preferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catgorically"),
                &["categorically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coorination"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conver"),
                &["convert", "cover", "convex", "convey", "confer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("portablity"),
                &["portability"],
            ),
            (dictgen::InsensitiveStr::Ascii("oprators"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("parokeal"), &["parochial"]),
            (dictgen::InsensitiveStr::Ascii("robocoup"), &["robocop"]),
            (dictgen::InsensitiveStr::Ascii("termnials"), &["terminals"]),
            (dictgen::InsensitiveStr::Ascii("diversifiy"), &["diversify"]),
            (dictgen::InsensitiveStr::Ascii("cellls"), &["cells"]),
            (
                dictgen::InsensitiveStr::Ascii("govornment"),
                &["government"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspoondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miletsones"),
                &["milestones"],
            ),
            (dictgen::InsensitiveStr::Ascii("invidual"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("unlimted"), &["unlimited"]),
            (
                dictgen::InsensitiveStr::Ascii("placegolder"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("muh"), &["much"]),
            (
                dictgen::InsensitiveStr::Ascii("condfigurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("approvel"), &["approval"]),
            (
                dictgen::InsensitiveStr::Ascii("propserous"),
                &["prosperous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cackground"),
                &["background"],
            ),
            (dictgen::InsensitiveStr::Ascii("foir"), &["for"]),
            (
                dictgen::InsensitiveStr::Ascii("playgropund"),
                &["playground"],
            ),
            (dictgen::InsensitiveStr::Ascii("epscially"), &["specially"]),
            (dictgen::InsensitiveStr::Ascii("rearely"), &["rarely"]),
            (dictgen::InsensitiveStr::Ascii("trabsform"), &["transform"]),
            (dictgen::InsensitiveStr::Ascii("asceding"), &["ascending"]),
            (dictgen::InsensitiveStr::Ascii("solfes"), &["solves"]),
            (
                dictgen::InsensitiveStr::Ascii("specifcying"),
                &["specifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("opearation"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("tavelled"), &["travelled"]),
            (dictgen::InsensitiveStr::Ascii("immensley"), &["immensely"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeted"),
                &["cheated", "cheered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrators"),
                &["aggregators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indispensible"),
                &["indispensable"],
            ),
            (dictgen::InsensitiveStr::Ascii("assoaiate"), &["associate"]),
            (dictgen::InsensitiveStr::Ascii("downvotted"), &["downvote"]),
            (
                dictgen::InsensitiveStr::Ascii("boostrappers"),
                &["bootstrappers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deambigate"),
                &["disambiguate"],
            ),
            (dictgen::InsensitiveStr::Ascii("momemtum"), &["momentum"]),
            (dictgen::InsensitiveStr::Ascii("parahaps"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("initals"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("whitepaces"),
                &["whitespaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("oragsms"), &["orgasms"]),
            (dictgen::InsensitiveStr::Ascii("emptry"), &["empty"]),
            (dictgen::InsensitiveStr::Ascii("idicating"), &["indicating"]),
            (dictgen::InsensitiveStr::Ascii("unshfits"), &["unshifts"]),
            (
                dictgen::InsensitiveStr::Ascii("comatability"),
                &["compatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("ypur"), &["your"]),
            (dictgen::InsensitiveStr::Ascii("multiplays"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("recrooting"),
                &["recruiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("nawshesly"), &["nauseously"]),
            (dictgen::InsensitiveStr::Ascii("lisetners"), &["listeners"]),
            (
                dictgen::InsensitiveStr::Ascii("procrastinazione"),
                &["procrastination"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("tiggering"), &["triggering"]),
            (dictgen::InsensitiveStr::Ascii("ineritors"), &["inheritors"]),
            (
                dictgen::InsensitiveStr::Ascii("opproximates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("underlfow"), &["underflow"]),
            (dictgen::InsensitiveStr::Ascii("beteeen"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("minneosta"), &["minnesota"]),
            (dictgen::InsensitiveStr::Ascii("reveokes"), &["revokes"]),
            (
                dictgen::InsensitiveStr::Ascii("prediciotn"),
                &["prediction"],
            ),
            (dictgen::InsensitiveStr::Ascii("rezize"), &["resize"]),
            (dictgen::InsensitiveStr::Ascii("reesult"), &["result"]),
            (
                dictgen::InsensitiveStr::Ascii("crucifiction"),
                &["crucifixion"],
            ),
            (dictgen::InsensitiveStr::Ascii("abborted"), &["aborted"]),
            (dictgen::InsensitiveStr::Ascii("shtopp"), &["stop", "shop"]),
            (
                dictgen::InsensitiveStr::Ascii("experinementing"),
                &["experimenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("polarlized"), &["polarized"]),
            (
                dictgen::InsensitiveStr::Ascii("descriminant"),
                &["discriminant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("breastfeedig"),
                &["breastfeeding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegociation"),
                &["subnegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiotions"),
                &["negotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralellisation"),
                &["parallelisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("embarress"),
                &["embarrassed"],
            ),
            (dictgen::InsensitiveStr::Ascii("confliced"), &["conflicted"]),
            (dictgen::InsensitiveStr::Ascii("makfile"), &["makefile"]),
            (dictgen::InsensitiveStr::Ascii("allocatedi"), &["allocated"]),
            (dictgen::InsensitiveStr::Ascii("follwows"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbethoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represendative"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("aqauriums"), &["aquariums"]),
            (dictgen::InsensitiveStr::Ascii("multiplr"), &["multiplier"]),
            (
                dictgen::InsensitiveStr::Ascii("configurtable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignests"), &["ingests"]),
            (
                dictgen::InsensitiveStr::Ascii("precomiled"),
                &["precompiled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partiitoning"),
                &["partitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("varieble"), &["variable"]),
            (dictgen::InsensitiveStr::Ascii("awnsered"), &["answered"]),
            (dictgen::InsensitiveStr::Ascii("loggged"), &["logged"]),
            (dictgen::InsensitiveStr::Ascii("dignitiy"), &["dignity"]),
            (
                dictgen::InsensitiveStr::Ascii("prerogotive"),
                &["prerogative"],
            ),
            (dictgen::InsensitiveStr::Ascii("fashoning"), &["fashioning"]),
            (
                dictgen::InsensitiveStr::Ascii("bencmarking"),
                &["benchmarking"],
            ),
            (dictgen::InsensitiveStr::Ascii("planetas"), &["planets"]),
            (dictgen::InsensitiveStr::Ascii("raduis"), &["radius"]),
            (
                dictgen::InsensitiveStr::Ascii("disitributions"),
                &["distributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("regesters"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("narcississts"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("armageddin"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homosexulaity"),
                &["homosexuality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prescriptons"),
                &["prescriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transkript"),
                &["transcript"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocatoor"), &["allocator"]),
            (dictgen::InsensitiveStr::Ascii("posibble"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("micrpohone"),
                &["microphones"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correxponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavoidble"),
                &["unavoidable"],
            ),
            (dictgen::InsensitiveStr::Ascii("skwalking"), &["squawking"]),
            (
                dictgen::InsensitiveStr::Ascii("shamelessley"),
                &["shamelessly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disgustingy"),
                &["disgustingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("nuclean"), &["unclean"]),
            (dictgen::InsensitiveStr::Ascii("motnage"), &["montage"]),
            (
                dictgen::InsensitiveStr::Ascii("ddictionary"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("attches"), &["attaches"]),
            (
                dictgen::InsensitiveStr::Ascii("funcitionally"),
                &["functionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachterizing"),
                &["characterizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afterhtought"),
                &["afterthought"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encrpytions"),
                &["encryptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepates"), &["prepares"]),
            (dictgen::InsensitiveStr::Ascii("defaulat"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplicacion"),
                &["deduplication"],
            ),
            (dictgen::InsensitiveStr::Ascii("qutie"), &["quite", "quiet"]),
            (dictgen::InsensitiveStr::Ascii("parants"), &["parents"]),
            (dictgen::InsensitiveStr::Ascii("valdating"), &["validating"]),
            (dictgen::InsensitiveStr::Ascii("hypter"), &["hyper"]),
            (
                dictgen::InsensitiveStr::Ascii("deyhdrated"),
                &["dehydrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereral"), &["general"]),
            (
                dictgen::InsensitiveStr::Ascii("experiminted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("pickyunes"), &["picayunes"]),
            (dictgen::InsensitiveStr::Ascii("platfoem"), &["platform"]),
            (dictgen::InsensitiveStr::Ascii("admrial"), &["admiral"]),
            (dictgen::InsensitiveStr::Ascii("fitlering"), &["filtering"]),
            (
                dictgen::InsensitiveStr::Ascii("descandant"),
                &["descendant"],
            ),
            (dictgen::InsensitiveStr::Ascii("lcoale"), &["locale"]),
            (dictgen::InsensitiveStr::Ascii("wrkflow"), &["workflow"]),
            (dictgen::InsensitiveStr::Ascii("tyical"), &["typical"]),
            (dictgen::InsensitiveStr::Ascii("normalos"), &["normals"]),
            (dictgen::InsensitiveStr::Ascii("coallee"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("unsable"),
                &["unusable", "usable", "unstable", "unable"],
            ),
            (dictgen::InsensitiveStr::Ascii("aritmetic"), &["arithmetic"]),
            (dictgen::InsensitiveStr::Ascii("willams"), &["williams"]),
            (
                dictgen::InsensitiveStr::Ascii("distiguishes"),
                &["distinguishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("artcle"), &["article"]),
            (
                dictgen::InsensitiveStr::Ascii("generalizare"),
                &["generalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("bencmarks"), &["benchmarks"]),
            (dictgen::InsensitiveStr::Ascii("leathal"), &["lethal"]),
            (
                dictgen::InsensitiveStr::Ascii("homosexuallity"),
                &["homosexuality"],
            ),
            (dictgen::InsensitiveStr::Ascii("stytle"), &["style"]),
            (
                dictgen::InsensitiveStr::Ascii("vunlerable"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharamcist"),
                &["pharmacist"],
            ),
            (dictgen::InsensitiveStr::Ascii("craweld"), &["crawled"]),
            (
                dictgen::InsensitiveStr::Ascii("accomodate"),
                &["accommodate"],
            ),
            (dictgen::InsensitiveStr::Ascii("cernomial"), &["ceremonial"]),
            (
                dictgen::InsensitiveStr::Ascii("circumsiced"),
                &["circumcised"],
            ),
            (dictgen::InsensitiveStr::Ascii("fileystem"), &["filesystem"]),
            (dictgen::InsensitiveStr::Ascii("branier"), &["brainer"]),
            (dictgen::InsensitiveStr::Ascii("perhabs"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("recordarle"), &["recorder"]),
            (dictgen::InsensitiveStr::Ascii("bulletts"), &["bullets"]),
            (
                dictgen::InsensitiveStr::Ascii("trasforming"),
                &["transforming"],
            ),
            (dictgen::InsensitiveStr::Ascii("boradcast"), &["broadcast"]),
            (
                dictgen::InsensitiveStr::Ascii("reflextion"),
                &["reflection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("similairty"),
                &["similarity"],
            ),
            (dictgen::InsensitiveStr::Ascii("birghtest"), &["brightest"]),
            (dictgen::InsensitiveStr::Ascii("intalling"), &["installing"]),
            (dictgen::InsensitiveStr::Ascii("availablen"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("shoudlnt"), &["shouldnt"]),
            (
                dictgen::InsensitiveStr::Ascii("apprixamately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("endianesse"),
                &["endianness"],
            ),
            (dictgen::InsensitiveStr::Ascii("cycular"), &["circular"]),
            (
                dictgen::InsensitiveStr::Ascii("disqualifyed"),
                &["disqualified"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsevrer"), &["observer"]),
            (
                dictgen::InsensitiveStr::Ascii("disassbly"),
                &["disassembly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vocabularity"),
                &["vocabulary"],
            ),
            (dictgen::InsensitiveStr::Ascii("escalato"), &["escalation"]),
            (
                dictgen::InsensitiveStr::Ascii("consituency"),
                &["constituency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predescesors"),
                &["predecessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patroitism"),
                &["patriotism"],
            ),
            (dictgen::InsensitiveStr::Ascii("obscruity"), &["obscurity"]),
            (dictgen::InsensitiveStr::Ascii("fomratted"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("resquested"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("deligthful"),
                &["delightful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generalizin"),
                &["generalizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("completey"), &["completely"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteaming"),
                &["downstreaming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southtampon"),
                &["southampton"],
            ),
            (dictgen::InsensitiveStr::Ascii("regresso"), &["regression"]),
            (
                dictgen::InsensitiveStr::Ascii("destinctions"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analysising"),
                &["analysing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesults"), &["results"]),
            (dictgen::InsensitiveStr::Ascii("trangular"), &["triangular"]),
            (
                dictgen::InsensitiveStr::Ascii("speicalize"),
                &["specialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("premissions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philospohers"),
                &["philosophers"],
            ),
            (dictgen::InsensitiveStr::Ascii("serailsed"), &["serialised"]),
            (dictgen::InsensitiveStr::Ascii("piecewice"), &["piecewise"]),
            (dictgen::InsensitiveStr::Ascii("initailze"), &["initialize"]),
            (dictgen::InsensitiveStr::Ascii("phillipses"), &["phillies"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegoation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("consles"), &["consoles"]),
            (
                dictgen::InsensitiveStr::Ascii("unsolicated"),
                &["unsolicited"],
            ),
            (dictgen::InsensitiveStr::Ascii("weakneses"), &["weakness"]),
            (
                dictgen::InsensitiveStr::Ascii("appearnaces"),
                &["appearances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequisities"),
                &["prerequisite", "prerequisites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dispencaries"),
                &["dispensaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("helictopers"),
                &["helicopters"],
            ),
            (dictgen::InsensitiveStr::Ascii("maco"), &["macro"]),
            (dictgen::InsensitiveStr::Ascii("meight"), &["might"]),
            (dictgen::InsensitiveStr::Ascii("bitwize"), &["bitwise"]),
            (dictgen::InsensitiveStr::Ascii("locla"), &["local"]),
            (
                dictgen::InsensitiveStr::Ascii("methematical"),
                &["mathematical"],
            ),
            (dictgen::InsensitiveStr::Ascii("pretect"), &["protect"]),
            (
                dictgen::InsensitiveStr::Ascii("frquencies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("abillity"), &["ability"]),
            (dictgen::InsensitiveStr::Ascii("cancellled"), &["cancelled"]),
            (
                dictgen::InsensitiveStr::Ascii("furute"),
                &["fruit", "future"],
            ),
            (dictgen::InsensitiveStr::Ascii("multiplayr"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("recogisers"),
                &["recognisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramterisation"),
                &["parameterisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimina"), &["minima"]),
            (dictgen::InsensitiveStr::Ascii("buuilder"), &["builder"]),
            (
                dictgen::InsensitiveStr::Ascii("environemts"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("generae"), &["generate"]),
            (
                dictgen::InsensitiveStr::Ascii("disastrosi"),
                &["disastrous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misogynstic"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triks"),
                &["tricks", "trikes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revolutin"),
                &["revolutions", "revolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virualizes"),
                &["virtualizes", "visualizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abbriviation"),
                &["abbreviation"],
            ),
            (dictgen::InsensitiveStr::Ascii("islmaists"), &["islamists"]),
            (dictgen::InsensitiveStr::Ascii("tuncation"), &["truncation"]),
            (
                dictgen::InsensitiveStr::Ascii("distribitive"),
                &["distributive"],
            ),
            (dictgen::InsensitiveStr::Ascii("syntac"), &["syntax"]),
            (
                dictgen::InsensitiveStr::Ascii("execututions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleraton"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("pathames"), &["pathnames"]),
            (dictgen::InsensitiveStr::Ascii("tranpose"), &["transpose"]),
            (dictgen::InsensitiveStr::Ascii("stuat"), &["status"]),
            (
                dictgen::InsensitiveStr::Ascii("appriecate"),
                &["appreciates", "appreciate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("readdrssing"),
                &["readdressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imcompetence"),
                &["incompetence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environemntally"),
                &["environmentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("initiase"), &["initialise"]),
            (
                dictgen::InsensitiveStr::Ascii("responsed"),
                &["responded", "responses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cannobalism"),
                &["cannibalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trush"),
                &["trust", "trash", "thrush"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mistakengly"),
                &["mistakenly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilizes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cetrificate"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("knive"), &["knife"]),
            (dictgen::InsensitiveStr::Ascii("cheesees"), &["cheeses"]),
            (
                dictgen::InsensitiveStr::Ascii("survaillence"),
                &["surveillance"],
            ),
            (dictgen::InsensitiveStr::Ascii("coulor"), &["color"]),
            (dictgen::InsensitiveStr::Ascii("proirity"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("disruptron"),
                &["disruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restrucure"),
                &["restructure"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgainse"), &["organise"]),
            (dictgen::InsensitiveStr::Ascii("researve"), &["reserve"]),
            (dictgen::InsensitiveStr::Ascii("copyingn"), &["copying"]),
            (
                dictgen::InsensitiveStr::Ascii("confifurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifried"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("alarams"), &["alarms"]),
            (dictgen::InsensitiveStr::Ascii("generice"), &["generic"]),
            (
                dictgen::InsensitiveStr::Ascii("javascropt"),
                &["javascript"],
            ),
            (dictgen::InsensitiveStr::Ascii("quanitfy"), &["quantify"]),
            (dictgen::InsensitiveStr::Ascii("exceute"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("preveiwed"), &["previewed"]),
            (dictgen::InsensitiveStr::Ascii("redings"), &["readings"]),
            (dictgen::InsensitiveStr::Ascii("happends"), &["happens"]),
            (dictgen::InsensitiveStr::Ascii("exponend"), &["exponent"]),
            (dictgen::InsensitiveStr::Ascii("ungodley"), &["ungodly"]),
            (dictgen::InsensitiveStr::Ascii("symantic"), &["semantic"]),
            (dictgen::InsensitiveStr::Ascii("appartment"), &["apartment"]),
            (
                dictgen::InsensitiveStr::Ascii("supposidely"),
                &["supposedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dimesionality"),
                &["dimensionality"],
            ),
            (dictgen::InsensitiveStr::Ascii("identifiy"), &["identify"]),
            (
                dictgen::InsensitiveStr::Ascii("alcoholis"),
                &["alcoholics", "alcoholic", "alcoholism"],
            ),
            (dictgen::InsensitiveStr::Ascii("similiarly"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("neckbread"), &["neckbeard"]),
            (dictgen::InsensitiveStr::Ascii("usualyl"), &["usually"]),
            (dictgen::InsensitiveStr::Ascii("summuries"), &["summaries"]),
            (dictgen::InsensitiveStr::Ascii("alphbet"), &["alphabet"]),
            (
                dictgen::InsensitiveStr::Ascii("deduplacte"),
                &["deduplicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("summurisation"),
                &["summarisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplins"),
                &["disciplines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functionailty"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commendment"),
                &["commandment"],
            ),
            (dictgen::InsensitiveStr::Ascii("perhaphs"), &["perhaps"]),
            (dictgen::InsensitiveStr::Ascii("sigrynges"), &["syringes"]),
            (dictgen::InsensitiveStr::Ascii("perosnas"), &["personas"]),
            (dictgen::InsensitiveStr::Ascii("pleae"), &["please"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteams"),
                &["downstreams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("debbug"),
                &["debug", "bedbug"],
            ),
            (dictgen::InsensitiveStr::Ascii("suggestes"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("chlid"), &["child"]),
            (dictgen::InsensitiveStr::Ascii("disocunt"), &["discount"]),
            (dictgen::InsensitiveStr::Ascii("instller"), &["installer"]),
            (
                dictgen::InsensitiveStr::Ascii("examin"),
                &["examine", "examining"],
            ),
            (dictgen::InsensitiveStr::Ascii("individal"), &["individual"]),
            (dictgen::InsensitiveStr::Ascii("tarce"), &["trace"]),
            (
                dictgen::InsensitiveStr::Ascii("destrutors"),
                &["destructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tresspasing"),
                &["trespassing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quantative"),
                &["quantitative"],
            ),
            (dictgen::InsensitiveStr::Ascii("chemicaly"), &["chemically"]),
            (dictgen::InsensitiveStr::Ascii("filterd"), &["filtered"]),
            (dictgen::InsensitiveStr::Ascii("carrets"), &["carets"]),
            (
                dictgen::InsensitiveStr::Ascii("necromencer"),
                &["necromancer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custimizable"),
                &["customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interpretates"),
                &["interprets"],
            ),
            (dictgen::InsensitiveStr::Ascii("basiton"), &["bastion"]),
            (
                dictgen::InsensitiveStr::Ascii("sumulations"),
                &["simulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenicaion"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("manaually"), &["manually"]),
            (
                dictgen::InsensitiveStr::Ascii("authenicaors"),
                &["authenticators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fantasticlly"),
                &["fantastically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reeiving"),
                &["receiving", "reviving"],
            ),
            (dictgen::InsensitiveStr::Ascii("resticts"), &["restricts"]),
            (
                dictgen::InsensitiveStr::Ascii("artificually"),
                &["artificially"],
            ),
            (dictgen::InsensitiveStr::Ascii("setevn"), &["setenv"]),
            (
                dictgen::InsensitiveStr::Ascii("addititonal"),
                &["additional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approbiate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("friendlinees"),
                &["friendliness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coodrinates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("zylophone"), &["xylophone"]),
            (
                dictgen::InsensitiveStr::Ascii("cylndrically"),
                &["cylindrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovrlapped"), &["overlapped"]),
            (
                dictgen::InsensitiveStr::Ascii("improsonment"),
                &["imprisonment"],
            ),
            (dictgen::InsensitiveStr::Ascii("satisy"), &["satisfy"]),
            (
                dictgen::InsensitiveStr::Ascii("specializeds"),
                &["specializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("prferable"), &["preferable"]),
            (
                dictgen::InsensitiveStr::Ascii("stabilizin"),
                &["stabilizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("encypted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("opition"),
                &["option", "opinion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("foregronds"),
                &["foregrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forthcomming"),
                &["forthcoming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sustainaiblity"),
                &["sustainability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transproted"),
                &["transported"],
            ),
            (dictgen::InsensitiveStr::Ascii("tratiors"), &["traitors"]),
            (dictgen::InsensitiveStr::Ascii("exchanaged"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("pormises"), &["promises"]),
            (dictgen::InsensitiveStr::Ascii("polygones"), &["polygons"]),
            (dictgen::InsensitiveStr::Ascii("unrptorect"), &["unprotect"]),
            (
                dictgen::InsensitiveStr::Ascii("curteusly"),
                &["courteously"],
            ),
            (dictgen::InsensitiveStr::Ascii("messaage"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("percusive"), &["percussive"]),
            (
                dictgen::InsensitiveStr::Ascii("contirbutions"),
                &["contributions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitalizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("makss"), &["masks", "makes"]),
            (
                dictgen::InsensitiveStr::Ascii("porperties"),
                &["properties"],
            ),
            (dictgen::InsensitiveStr::Ascii("aribters"), &["arbiters"]),
            (
                dictgen::InsensitiveStr::Ascii("chiled"),
                &["child", "chilled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompable"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flexibiliy"),
                &["flexibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepertator"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegosiations"),
                &["subnegotiations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialiasing"),
                &["initialising"],
            ),
            (dictgen::InsensitiveStr::Ascii("custoners"), &["customers"]),
            (dictgen::InsensitiveStr::Ascii("dafualt"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("corrsponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flourishment"),
                &["flourishing"],
            ),
            (dictgen::InsensitiveStr::Ascii("uplaoders"), &["uploaders"]),
            (dictgen::InsensitiveStr::Ascii("deletd"), &["deleted"]),
            (
                dictgen::InsensitiveStr::Ascii("mensioning"),
                &["mentioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("wispers"), &["whispers"]),
            (dictgen::InsensitiveStr::Ascii("choosen"), &["chosen"]),
            (dictgen::InsensitiveStr::Ascii("activte"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("decalrative"),
                &["declarative"],
            ),
            (dictgen::InsensitiveStr::Ascii("sednetary"), &["sedentary"]),
            (dictgen::InsensitiveStr::Ascii("onboad"), &["onboard"]),
            (dictgen::InsensitiveStr::Ascii("recieves"), &["receives"]),
            (
                dictgen::InsensitiveStr::Ascii("excection"),
                &["exception", "execution"],
            ),
            (dictgen::InsensitiveStr::Ascii("truns"), &["turns"]),
            (dictgen::InsensitiveStr::Ascii("anuls"), &["annulls"]),
            (dictgen::InsensitiveStr::Ascii("waht"), &["what"]),
            (dictgen::InsensitiveStr::Ascii("sarcastc"), &["sarcastic"]),
            (
                dictgen::InsensitiveStr::Ascii("enumeratior"),
                &["enumerator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("babysister"),
                &["babysitter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configiurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstracion"),
                &["abstraction"],
            ),
            (dictgen::InsensitiveStr::Ascii("packte"), &["packet"]),
            (
                dictgen::InsensitiveStr::Ascii("susbtantially"),
                &["substantially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repective"),
                &["receptive", "respective"],
            ),
            (dictgen::InsensitiveStr::Ascii("exportede"), &["exported"]),
            (dictgen::InsensitiveStr::Ascii("buddhsim"), &["buddhism"]),
            (
                dictgen::InsensitiveStr::Ascii("charactherization"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discontinutiy"),
                &["discontinuity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inspecticon"),
                &["inspection"],
            ),
            (dictgen::InsensitiveStr::Ascii("annouces"), &["announces"]),
            (dictgen::InsensitiveStr::Ascii("gotees"), &["goatees"]),
            (dictgen::InsensitiveStr::Ascii("wirter"), &["writer"]),
            (
                dictgen::InsensitiveStr::Ascii("enginnering"),
                &["engineering"],
            ),
            (dictgen::InsensitiveStr::Ascii("obivously"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("nomralization"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyhthon"), &["python"]),
            (dictgen::InsensitiveStr::Ascii("vanishs"), &["vanishes"]),
            (dictgen::InsensitiveStr::Ascii("loacality"), &["locality"]),
            (
                dictgen::InsensitiveStr::Ascii("elemenental"),
                &["elemental"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximun"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("predictment"),
                &["predicament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("floresent"),
                &["fluorescent", "florescent"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploder"), &["uploader"]),
            (dictgen::InsensitiveStr::Ascii("excahnge"), &["exchange"]),
            (
                dictgen::InsensitiveStr::Ascii("certifactes"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("mchanisms"), &["mechanisms"]),
            (dictgen::InsensitiveStr::Ascii("tutoriels"), &["tutorials"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridnation"),
                &["coordination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cherckpoint"),
                &["checkpoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sttings"),
                &["strings", "settings", "sittings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("headquartes"),
                &["headquarters"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalate"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("compotition"),
                &["composition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplicity"),
                &["multiplicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("beginn"), &["begin"]),
            (dictgen::InsensitiveStr::Ascii("dirctly"), &["directly"]),
            (
                dictgen::InsensitiveStr::Ascii("physcological"),
                &["psychological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhodd"),
                &["neighbourhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sreampropinfo"),
                &["streampropinfo"],
            ),
            (dictgen::InsensitiveStr::Ascii("devasted"), &["devastated"]),
            (
                dictgen::InsensitiveStr::Ascii("apporoximately"),
                &["approximately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supppresses"),
                &["suppresses"],
            ),
            (dictgen::InsensitiveStr::Ascii("bject"), &["object"]),
            (
                dictgen::InsensitiveStr::Ascii("consructors"),
                &["constructors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("characterists"),
                &["characteristics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calcluators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("mailstrum"), &["maelstrom"]),
            (dictgen::InsensitiveStr::Ascii("cylindre"), &["cylinder"]),
            (
                dictgen::InsensitiveStr::Ascii("registertd"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinaton"),
                &["combination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("functtionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reccomends"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("platfarm"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("containt"),
                &["contain", "content"],
            ),
            (dictgen::InsensitiveStr::Ascii("recating"), &["reacting"]),
            (
                dictgen::InsensitiveStr::Ascii("idenically"),
                &["identically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("directrions"),
                &["directions"],
            ),
            (dictgen::InsensitiveStr::Ascii("adevnture"), &["adventure"]),
            (dictgen::InsensitiveStr::Ascii("intensley"), &["intensely"]),
            (
                dictgen::InsensitiveStr::Ascii("simulantaneous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("viligante"), &["vigilante"]),
            (dictgen::InsensitiveStr::Ascii("rasing"), &["raising"]),
            (
                dictgen::InsensitiveStr::Ascii("plceholder"),
                &["placeholder"],
            ),
            (dictgen::InsensitiveStr::Ascii("ceromony"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("disterbance"),
                &["disturbance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brough"),
                &["brought", "borough", "bough", "trough"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incombatibilities"),
                &["incompatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trrigering"),
                &["triggering"],
            ),
            (dictgen::InsensitiveStr::Ascii("agaisnt"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("advocade"),
                &["advocated", "advocate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invisilibity"),
                &["invisibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anesthisia"),
                &["anesthesia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathologial"),
                &["pathological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteferring"),
                &["interfering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("awesomenss"),
                &["awesomeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsusbcribe"),
                &["unsubscribe"],
            ),
            (dictgen::InsensitiveStr::Ascii("bijcetive"), &["bijective"]),
            (dictgen::InsensitiveStr::Ascii("willimas"), &["williams"]),
            (dictgen::InsensitiveStr::Ascii("fasetr"), &["faster"]),
            (dictgen::InsensitiveStr::Ascii("extrimly"), &["extremely"]),
            (dictgen::InsensitiveStr::Ascii("capitain"), &["captain"]),
            (
                dictgen::InsensitiveStr::Ascii("samilarity"),
                &["similarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("randazyvous"),
                &["rendezvous"],
            ),
            (dictgen::InsensitiveStr::Ascii("spruious"), &["spurious"]),
            (dictgen::InsensitiveStr::Ascii("componbent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("depenedence"),
                &["dependence"],
            ),
            (dictgen::InsensitiveStr::Ascii("commnt"), &["comment"]),
            (
                dictgen::InsensitiveStr::Ascii("procecess"),
                &["process", "processes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conservitave"),
                &["conservatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("existatus"), &["exitstatus"]),
            (dictgen::InsensitiveStr::Ascii("roigins"), &["origins"]),
            (
                dictgen::InsensitiveStr::Ascii("circunference"),
                &["circumference"],
            ),
            (dictgen::InsensitiveStr::Ascii("softend"), &["softened"]),
            (dictgen::InsensitiveStr::Ascii("identite"), &["identities"]),
            (dictgen::InsensitiveStr::Ascii("analgoue"), &["analogue"]),
            (dictgen::InsensitiveStr::Ascii("cretaing"), &["creating"]),
            (dictgen::InsensitiveStr::Ascii("resoect"), &["respect"]),
            (
                dictgen::InsensitiveStr::Ascii("desentizised"),
                &["desensitised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tramsmitters"),
                &["transmitters"],
            ),
            (dictgen::InsensitiveStr::Ascii("flashig"), &["flashing"]),
            (
                dictgen::InsensitiveStr::Ascii("corrent"),
                &["correct", "current"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rollorcoaster"),
                &["rollercoaster"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrctions"),
                &["instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("executons"),
                &["executions", "executors"],
            ),
            (dictgen::InsensitiveStr::Ascii("derfine"), &["define"]),
            (
                dictgen::InsensitiveStr::Ascii("partitoning"),
                &["partitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aborignial"),
                &["aboriginal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ntegrating"),
                &["integrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cnfigurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ajasencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("invalde"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("optimation"),
                &["optimization", "optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("neighobr"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("worarounds"),
                &["workarounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("shorctuts"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("bruglar"), &["burglar"]),
            (
                dictgen::InsensitiveStr::Ascii("compatabily"),
                &["compatibility", "compatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capibility"),
                &["capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forach"),
                &["foreach", "orach"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerless"),
                &["parameterless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crresponding"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typographc"),
                &["typographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regenrative"),
                &["regenerative"],
            ),
            (dictgen::InsensitiveStr::Ascii("behavoir"), &["behavior"]),
            (
                dictgen::InsensitiveStr::Ascii("democractic"),
                &["democratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pressentation"),
                &["presentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coniguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("resloving"), &["resolving"]),
            (
                dictgen::InsensitiveStr::Ascii("immunosupressant"),
                &["immunosuppressant"],
            ),
            (dictgen::InsensitiveStr::Ascii("etst"), &["test"]),
            (dictgen::InsensitiveStr::Ascii("adequitly"), &["adequately"]),
            (dictgen::InsensitiveStr::Ascii("presicely"), &["precisely"]),
            (dictgen::InsensitiveStr::Ascii("mixxed"), &["mixed"]),
            (dictgen::InsensitiveStr::Ascii("intead"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("callbabck"), &["callback"]),
            (
                dictgen::InsensitiveStr::Ascii("algoprithmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("acrlyic"), &["acrylic"]),
            (
                dictgen::InsensitiveStr::Ascii("subdirecty"),
                &["subdirectory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraordinairy"),
                &["extraordinary"],
            ),
            (dictgen::InsensitiveStr::Ascii("comiples"), &["compiles"]),
            (
                dictgen::InsensitiveStr::Ascii("effectivness"),
                &["effectiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("storaged"),
                &["stored", "storages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electricly"),
                &["electrically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("morpholgies"),
                &["morphologies"],
            ),
            (dictgen::InsensitiveStr::Ascii("moniters"), &["monitors"]),
            (dictgen::InsensitiveStr::Ascii("disapait"), &["dissipate"]),
            (dictgen::InsensitiveStr::Ascii("functdion"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("delepted"), &["depleted"]),
            (
                dictgen::InsensitiveStr::Ascii("concnetration"),
                &["concentrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("gald"), &["glad"]),
            (dictgen::InsensitiveStr::Ascii("palying"), &["playing"]),
            (dictgen::InsensitiveStr::Ascii("sucesfull"), &["successful"]),
            (
                dictgen::InsensitiveStr::Ascii("intrduction"),
                &["introduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("paraiste"), &["parasite"]),
            (dictgen::InsensitiveStr::Ascii("uerying"), &["querying"]),
            (dictgen::InsensitiveStr::Ascii("portraing"), &["portraying"]),
            (dictgen::InsensitiveStr::Ascii("derepcate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("intercoarse"),
                &["intercourse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destractions"),
                &["distractions"],
            ),
            (dictgen::InsensitiveStr::Ascii("alorithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("extensiones"),
                &["extensions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terroristas"),
                &["terrorists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrativos"),
                &["demonstrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwhelmingy"),
                &["overwhelmingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("gereation"), &["generation"]),
            (
                dictgen::InsensitiveStr::Ascii("distinquishes"),
                &["distinguishes"],
            ),
            (dictgen::InsensitiveStr::Ascii("regressos"), &["regressors"]),
            (
                dictgen::InsensitiveStr::Ascii("chancillor"),
                &["chancellor"],
            ),
            (dictgen::InsensitiveStr::Ascii("sattellite"), &["satellite"]),
            (
                dictgen::InsensitiveStr::Ascii("unifnished"),
                &["unfinished"],
            ),
            (dictgen::InsensitiveStr::Ascii("envrirons"), &["environs"]),
            (
                dictgen::InsensitiveStr::Ascii("polyonimial"),
                &["polynomial"],
            ),
            (dictgen::InsensitiveStr::Ascii("killins"), &["killings"]),
            (
                dictgen::InsensitiveStr::Ascii("tranactional"),
                &["transactional"],
            ),
            (dictgen::InsensitiveStr::Ascii("karoake"), &["karaoke"]),
            (dictgen::InsensitiveStr::Ascii("pugin"), &["plugin"]),
            (
                dictgen::InsensitiveStr::Ascii("capitolization"),
                &["capitalization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recurssively"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("corredct"), &["correct"]),
            (dictgen::InsensitiveStr::Ascii("comnparing"), &["comparing"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalisim"),
                &["capitalism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gaurantteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorruptable"),
                &["incorruptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("benckmark"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("attriutions"),
                &["attributions"],
            ),
            (dictgen::InsensitiveStr::Ascii("envrion"), &["environ"]),
            (dictgen::InsensitiveStr::Ascii("otained"), &["obtained"]),
            (
                dictgen::InsensitiveStr::Ascii("culiminating"),
                &["culminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cyphertexts"),
                &["ciphertexts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("framewoeks"),
                &["frameworks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propositon"),
                &["proposition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individally"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insufficency"),
                &["insufficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("finialises"), &["finalises"]),
            (
                dictgen::InsensitiveStr::Ascii("criticizms"),
                &["criticisms"],
            ),
            (dictgen::InsensitiveStr::Ascii("odrer"), &["order"]),
            (
                dictgen::InsensitiveStr::Ascii("spontanuously"),
                &["spontaneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("renforced"), &["reinforced"]),
            (
                dictgen::InsensitiveStr::Ascii("perihperals"),
                &["peripherals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinements"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abstructions"),
                &["abstractions", "obstructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("filesytsem"),
                &["filesystem"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sketck"),
                &["skate", "sketch"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experirations"),
                &["expirations"],
            ),
            (dictgen::InsensitiveStr::Ascii("implcit"), &["implicit"]),
            (
                dictgen::InsensitiveStr::Ascii("cerinomial"),
                &["ceremonial"],
            ),
            (dictgen::InsensitiveStr::Ascii("seconde"), &["second"]),
            (dictgen::InsensitiveStr::Ascii("playgroun"), &["playground"]),
            (
                dictgen::InsensitiveStr::Ascii("complusory"),
                &["compulsory"],
            ),
            (dictgen::InsensitiveStr::Ascii("forcin"), &["forcing"]),
            (
                dictgen::InsensitiveStr::Ascii("fraebuffer"),
                &["framebuffer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wnidows"),
                &["windows", "widows"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informaiton"),
                &["information"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregressing"),
                &["progressing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spoffed"),
                &["spoofed", "spaffed", "scoffed"],
            ),
            (dictgen::InsensitiveStr::Ascii("simptum"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("midpont"), &["midpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("lonileness"),
                &["loneliness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabilityies"),
                &["vulnerability", "vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronisly"),
                &["synchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("nurons"), &["neurons"]),
            (dictgen::InsensitiveStr::Ascii("dieties"), &["deities"]),
            (dictgen::InsensitiveStr::Ascii("wimmenly"), &["womanly"]),
            (dictgen::InsensitiveStr::Ascii("regiones"), &["regions"]),
            (
                dictgen::InsensitiveStr::Ascii("exceptin"),
                &[
                    "exceptions",
                    "excepting",
                    "exception",
                    "expecting",
                    "accepting",
                ],
            ),
            (dictgen::InsensitiveStr::Ascii("trasnlate"), &["translate"]),
            (dictgen::InsensitiveStr::Ascii("inherti"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("specifify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("termanator"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cancelations"),
                &["cancellations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diferentiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("expors"), &["exports"]),
            (
                dictgen::InsensitiveStr::Ascii("custonizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("beforhand"), &["beforehand"]),
            (
                dictgen::InsensitiveStr::Ascii("virualising"),
                &["virtualising", "visualising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("connecticon"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authecator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ommishints"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indefinity"),
                &["indefinitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwhlemingly"),
                &["overwhelmingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("sumarises"), &["summarises"]),
            (
                dictgen::InsensitiveStr::Ascii("parternship"),
                &["partnership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philidalphia"),
                &["philadelphia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questioable"),
                &["questionable"],
            ),
            (dictgen::InsensitiveStr::Ascii("klicked"), &["clicked"]),
            (dictgen::InsensitiveStr::Ascii("folwo"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("extensie"), &["extensive"]),
            (
                dictgen::InsensitiveStr::Ascii("exsits"),
                &["exists", "exist"],
            ),
            (dictgen::InsensitiveStr::Ascii("brutallly"), &["brutally"]),
            (
                dictgen::InsensitiveStr::Ascii("insensetive"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conducter"),
                &["conductor", "conducted"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrang"), &["arrange"]),
            (dictgen::InsensitiveStr::Ascii("arguemnts"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("negativly"), &["negatively"]),
            (
                dictgen::InsensitiveStr::Ascii("daugtherboard"),
                &["daughterboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("filll"), &["fill"]),
            (
                dictgen::InsensitiveStr::Ascii("exhorbitently"),
                &["exorbitantly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constucting"),
                &["constructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canonivalize"),
                &["canonicalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("languale"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("conversts"), &["converts"]),
            (dictgen::InsensitiveStr::Ascii("greandes"), &["grenades"]),
            (dictgen::InsensitiveStr::Ascii("besed"), &["based"]),
            (dictgen::InsensitiveStr::Ascii("forasken"), &["forsaken"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstuction"),
                &["reconstruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooridates"),
                &["coordinates"],
            ),
            (dictgen::InsensitiveStr::Ascii("endcoder"), &["encoder"]),
            (dictgen::InsensitiveStr::Ascii("feasture"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("mutlithreads"),
                &["multithreads"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mircoscopes"),
                &["microscopes"],
            ),
            (dictgen::InsensitiveStr::Ascii("ducuments"), &["documents"]),
            (
                dictgen::InsensitiveStr::Ascii("fertelizer"),
                &["fertilizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reformating"),
                &["reformatting"],
            ),
            (dictgen::InsensitiveStr::Ascii("differant"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("facutally"), &["factually"]),
            (dictgen::InsensitiveStr::Ascii("levicitus"), &["leviticus"]),
            (dictgen::InsensitiveStr::Ascii("mauarder"), &["marauder"]),
            (
                dictgen::InsensitiveStr::Ascii("percentagens"),
                &["percentages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specificiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasversal"),
                &["traversal", "transversal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumuluate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missonaries"),
                &["missionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifiyng"), &["modifying"]),
            (
                dictgen::InsensitiveStr::Ascii("specifix"),
                &["specific", "specifics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("traslalations"),
                &["translations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciplinairy"),
                &["disciplinary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocaitons"),
                &["allocations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnessecarry"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sensititively"),
                &["sensitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("imlements"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("supppressing"),
                &["suppressing"],
            ),
            (dictgen::InsensitiveStr::Ascii("shadoloo"), &["shadaloo"]),
            (
                dictgen::InsensitiveStr::Ascii("rearching"),
                &["searching", "researching", "reaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acessibility"),
                &["accessibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enciclopedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("historyan"), &["historian"]),
            (dictgen::InsensitiveStr::Ascii("problams"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("manipule"), &["manipulate"]),
            (
                dictgen::InsensitiveStr::Ascii("deficiencias"),
                &["deficiencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewatchd"), &["rewatched"]),
            (
                dictgen::InsensitiveStr::Ascii("tablulators"),
                &["tabulators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aparrently"),
                &["apparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("germanus"), &["germans"]),
            (dictgen::InsensitiveStr::Ascii("wigets"), &["widgets"]),
            (dictgen::InsensitiveStr::Ascii("explose"), &["explodes"]),
            (dictgen::InsensitiveStr::Ascii("ascnesion"), &["ascension"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensinoalities"),
                &["dimensionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("academicaly"),
                &["academically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extennsions"),
                &["extensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("theough"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("excellenze"),
                &["excellence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaitor"),
                &["negotiator"],
            ),
            (dictgen::InsensitiveStr::Ascii("overclok"), &["overclock"]),
            (
                dictgen::InsensitiveStr::Ascii("sistemizes"),
                &["systemizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("despensaries"),
                &["dispensaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("astronatus"),
                &["astronauts"],
            ),
            (dictgen::InsensitiveStr::Ascii("egyptain"), &["egyptian"]),
            (dictgen::InsensitiveStr::Ascii("teksts"), &["texts"]),
            (
                dictgen::InsensitiveStr::Ascii("enviroment"),
                &["environment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("situatutions"),
                &["situations"],
            ),
            (dictgen::InsensitiveStr::Ascii("immobilze"), &["immobile"]),
            (
                dictgen::InsensitiveStr::Ascii("delarators"),
                &["declarators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("everythign"),
                &["everything"],
            ),
            (dictgen::InsensitiveStr::Ascii("preprended"), &["prepended"]),
            (
                dictgen::InsensitiveStr::Ascii("paht"),
                &["path", "pat", "part"],
            ),
            (dictgen::InsensitiveStr::Ascii("toxicitiy"), &["toxicity"]),
            (dictgen::InsensitiveStr::Ascii("debgug"), &["debug"]),
            (
                dictgen::InsensitiveStr::Ascii("intressting"),
                &["interesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prejudicies"),
                &["prejudice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribting"),
                &["contributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("critreria"), &["criteria"]),
            (dictgen::InsensitiveStr::Ascii("supsected"), &["suspected"]),
            (
                dictgen::InsensitiveStr::Ascii("perserving"),
                &["preserving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kompensation"),
                &["compensation"],
            ),
            (dictgen::InsensitiveStr::Ascii("provded"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("mechanizm"), &["mechanism"]),
            (
                dictgen::InsensitiveStr::Ascii("certfications"),
                &["certifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaxt"),
                &["context", "contact"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispached"), &["dispatched"]),
            (dictgen::InsensitiveStr::Ascii("enforncing"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("borderlads"),
                &["borderlands"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocataves"), &["octaves"]),
            (
                dictgen::InsensitiveStr::Ascii("incrediable"),
                &["incredible"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiblle"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("camouflagin"),
                &["camouflaging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mussil"),
                &["muscle", "mussel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equivelant"),
                &["equivalent"],
            ),
            (dictgen::InsensitiveStr::Ascii("potention"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("archieves"),
                &["achieves", "archives"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlcluded"), &["excluded"]),
            (
                dictgen::InsensitiveStr::Ascii("intellecutals"),
                &["intellectuals"],
            ),
            (dictgen::InsensitiveStr::Ascii("comdeic"), &["comedic"]),
            (dictgen::InsensitiveStr::Ascii("channael"), &["channel"]),
            (dictgen::InsensitiveStr::Ascii("defenetly"), &["definitely"]),
            (dictgen::InsensitiveStr::Ascii("leviathn"), &["leviathan"]),
            (dictgen::InsensitiveStr::Ascii("wavesfront"), &["wavefront"]),
            (dictgen::InsensitiveStr::Ascii("paswords"), &["passwords"]),
            (dictgen::InsensitiveStr::Ascii("depedent"), &["dependent"]),
            (
                dictgen::InsensitiveStr::Ascii("transparenty"),
                &["transparency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transalation"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("univerisites"),
                &["universities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tempororaries"),
                &["temporaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absuolatists"),
                &["absolutists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("socializng"),
                &["socializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thruoghout"),
                &["throughout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generlises"),
                &["generalises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assiociates"),
                &["associates"],
            ),
            (dictgen::InsensitiveStr::Ascii("overwiew"), &["overview"]),
            (dictgen::InsensitiveStr::Ascii("protocolls"), &["protocols"]),
            (dictgen::InsensitiveStr::Ascii("apects"), &["aspects"]),
            (dictgen::InsensitiveStr::Ascii("yourslef"), &["yourself"]),
            (
                dictgen::InsensitiveStr::Ascii("archtecture"),
                &["architecture"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aggrates"),
                &["aggregates", "aggravates"],
            ),
            (dictgen::InsensitiveStr::Ascii("arival"), &["arrival"]),
            (dictgen::InsensitiveStr::Ascii("cummulated"), &["cumulated"]),
            (
                dictgen::InsensitiveStr::Ascii("reguarlise"),
                &["regularise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partiy"),
                &["parity", "party"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentificating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pregnance"),
                &["pregnancies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troublehsooting"),
                &["troubleshooting"],
            ),
            (dictgen::InsensitiveStr::Ascii("annyoed"), &["annoyed"]),
            (
                dictgen::InsensitiveStr::Ascii("supervisores"),
                &["supervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transperantly"),
                &["transparently"],
            ),
            (dictgen::InsensitiveStr::Ascii("panicing"), &["panicking"]),
            (
                dictgen::InsensitiveStr::Ascii("neighrorhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preinitalization"),
                &["preinitialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("opjecting"), &["objecting"]),
            (dictgen::InsensitiveStr::Ascii("annimal"), &["animal"]),
            (
                dictgen::InsensitiveStr::Ascii("seledtable"),
                &["selectable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deamonised"),
                &["daemonised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kilomoters"),
                &["kilometers"],
            ),
            (dictgen::InsensitiveStr::Ascii("prodct"), &["product"]),
            (
                dictgen::InsensitiveStr::Ascii("contrain"),
                &["constrain", "contain"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmissions"), &["emissions"]),
            (dictgen::InsensitiveStr::Ascii("lushisly"), &["lusciously"]),
            (
                dictgen::InsensitiveStr::Ascii("opertunities"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simmetricaly"),
                &["symmetrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("explaned"), &["explained"]),
            (
                dictgen::InsensitiveStr::Ascii("rerferences"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("kenerl"), &["kernel"]),
            (dictgen::InsensitiveStr::Ascii("refereshed"), &["refreshed"]),
            (
                dictgen::InsensitiveStr::Ascii("simplicifaction"),
                &["simplification"],
            ),
            (dictgen::InsensitiveStr::Ascii("suppoert"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("straightie"),
                &["straighten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacating"),
                &["replicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("resrouced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("examplifies"),
                &["exemplifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unvulnerable"),
                &["invulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bidings"),
                &["bidding", "bindings"],
            ),
            (dictgen::InsensitiveStr::Ascii("releafing"), &["relieving"]),
            (dictgen::InsensitiveStr::Ascii("unkwown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("charactersistically"),
                &["characteristically"],
            ),
            (dictgen::InsensitiveStr::Ascii("advancin"), &["advancing"]),
            (
                dictgen::InsensitiveStr::Ascii("renegeration"),
                &["regeneration"],
            ),
            (dictgen::InsensitiveStr::Ascii("shperical"), &["spherical"]),
            (dictgen::InsensitiveStr::Ascii("remebered"), &["remembered"]),
            (
                dictgen::InsensitiveStr::Ascii("archictects"),
                &["architects"],
            ),
            (dictgen::InsensitiveStr::Ascii("totation"), &["rotation"]),
            (
                dictgen::InsensitiveStr::Ascii("mastrubates"),
                &["masturbate"],
            ),
            (dictgen::InsensitiveStr::Ascii("canadan"), &["canadian"]),
            (
                dictgen::InsensitiveStr::Ascii("experimetent"),
                &["experiment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destruktion"),
                &["destruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrbutor"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("vewer"), &["viewer"]),
            (
                dictgen::InsensitiveStr::Ascii("breakthrought"),
                &["breakthrough"],
            ),
            (dictgen::InsensitiveStr::Ascii("dynamlic"), &["dynamic"]),
            (
                dictgen::InsensitiveStr::Ascii("extraordinairily"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("substracts"), &["subtracts"]),
            (dictgen::InsensitiveStr::Ascii("adjuct"), &["adjust"]),
            (dictgen::InsensitiveStr::Ascii("attements"), &["attempts"]),
            (
                dictgen::InsensitiveStr::Ascii("intiialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcrted"), &["directed"]),
            (
                dictgen::InsensitiveStr::Ascii("faseeshusly"),
                &["facetiously"],
            ),
            (dictgen::InsensitiveStr::Ascii("animostiy"), &["animosity"]),
            (dictgen::InsensitiveStr::Ascii("applicance"), &["appliance"]),
            (
                dictgen::InsensitiveStr::Ascii("corinthins"),
                &["corinthians"],
            ),
            (dictgen::InsensitiveStr::Ascii("sctripts"), &["scripts"]),
            (
                dictgen::InsensitiveStr::Ascii("excecptions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("relese"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("appreiate"), &["appreciate"]),
            (dictgen::InsensitiveStr::Ascii("reputin"), &["reputing"]),
            (
                dictgen::InsensitiveStr::Ascii("onmisiences"),
                &["omnisciences"],
            ),
            (dictgen::InsensitiveStr::Ascii("producirse"), &["producers"]),
            (
                dictgen::InsensitiveStr::Ascii("conservativo"),
                &["conservation"],
            ),
            (dictgen::InsensitiveStr::Ascii("prsesense"), &["presence"]),
            (
                dictgen::InsensitiveStr::Ascii("undersatndably"),
                &["understandably"],
            ),
            (dictgen::InsensitiveStr::Ascii("wendsay"), &["wednesday"]),
            (
                dictgen::InsensitiveStr::Ascii("monolitich"),
                &["monolithic"],
            ),
            (dictgen::InsensitiveStr::Ascii("conceptos"), &["concepts"]),
            (dictgen::InsensitiveStr::Ascii("techers"), &["teachers"]),
            (dictgen::InsensitiveStr::Ascii("utilitize"), &["utilize"]),
            (dictgen::InsensitiveStr::Ascii("proces"), &["process"]),
            (dictgen::InsensitiveStr::Ascii("contorls"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("freind"), &["friend"]),
            (dictgen::InsensitiveStr::Ascii("jewls"), &["jews", "jewels"]),
            (dictgen::InsensitiveStr::Ascii("encolsed"), &["enclosed"]),
            (
                dictgen::InsensitiveStr::Ascii("destinction"),
                &["destination"],
            ),
            (dictgen::InsensitiveStr::Ascii("manuales"), &["manuals"]),
            (
                dictgen::InsensitiveStr::Ascii("comformity"),
                &["conformity"],
            ),
            (dictgen::InsensitiveStr::Ascii("bufferent"), &["buffered"]),
            (
                dictgen::InsensitiveStr::Ascii("intolarant"),
                &["intolerant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elemination"),
                &["elimination"],
            ),
            (dictgen::InsensitiveStr::Ascii("emput"), &["empty"]),
            (
                dictgen::InsensitiveStr::Ascii("commandmant"),
                &["commandment"],
            ),
            (dictgen::InsensitiveStr::Ascii("encocded"), &["encoded"]),
            (dictgen::InsensitiveStr::Ascii("millepede"), &["millipede"]),
            (
                dictgen::InsensitiveStr::Ascii("aquries"),
                &["acquires", "equerries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("requierment"),
                &["requirement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espectations"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("difffered"), &["differed"]),
            (
                dictgen::InsensitiveStr::Ascii("trancodings"),
                &["transcodings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("palys"),
                &["plays", "polys", "palsy"],
            ),
            (dictgen::InsensitiveStr::Ascii("zoinists"), &["zionists"]),
            (
                dictgen::InsensitiveStr::Ascii("reaquiring"),
                &["requiring", "reacquiring"],
            ),
            (dictgen::InsensitiveStr::Ascii("criticim"), &["criticisms"]),
            (dictgen::InsensitiveStr::Ascii("gererics"), &["generics"]),
            (dictgen::InsensitiveStr::Ascii("treting"), &["treating"]),
            (
                dictgen::InsensitiveStr::Ascii("dominante"),
                &["dominant", "dominate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admininsters"),
                &["administers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perliferate"),
                &["proliferate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissapoint"),
                &["disappoint"],
            ),
            (dictgen::InsensitiveStr::Ascii("busines"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("triginometry"),
                &["trigonometry"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuesdsay"), &["tuesdays"]),
            (dictgen::InsensitiveStr::Ascii("happpen"), &["happen"]),
            (
                dictgen::InsensitiveStr::Ascii("coincidance"),
                &["coincidence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extimations"),
                &["estimations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incovenience"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("vetod"), &["vetoed"]),
            (
                dictgen::InsensitiveStr::Ascii("proejctors"),
                &["projectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reuplaoder"),
                &["reuploader"],
            ),
            (dictgen::InsensitiveStr::Ascii("upcomming"), &["upcoming"]),
            (dictgen::InsensitiveStr::Ascii("delaing"), &["delaying"]),
            (
                dictgen::InsensitiveStr::Ascii("geneologies"),
                &["genealogies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isntalling"),
                &["installing"],
            ),
            (dictgen::InsensitiveStr::Ascii("wheighing"), &["weighing"]),
            (dictgen::InsensitiveStr::Ascii("inititally"), &["initially"]),
            (dictgen::InsensitiveStr::Ascii("utitled"), &["untitled"]),
            (
                dictgen::InsensitiveStr::Ascii("eligability"),
                &["eligibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introducted"),
                &["introduced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relegetion"),
                &["relegation"],
            ),
            (dictgen::InsensitiveStr::Ascii("comporess"), &["compress"]),
            (dictgen::InsensitiveStr::Ascii("submti"), &["submit"]),
            (dictgen::InsensitiveStr::Ascii("onlly"), &["only"]),
            (
                dictgen::InsensitiveStr::Ascii("capabability"),
                &["capability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("architectes"),
                &["architects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplimentations"),
                &["reimplementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blackade"),
                &["blacked", "blockade"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implementatino"),
                &["implementations", "implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transeviers"),
                &["transceivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bengalas"), &["bengals"]),
            (
                dictgen::InsensitiveStr::Ascii("organizarea"),
                &["organizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modidfications"),
                &["modifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protagainist"),
                &["protagonist"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememor"), &["remember"]),
            (
                dictgen::InsensitiveStr::Ascii("narcissictic"),
                &["narcissistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("aagain"), &["again"]),
            (
                dictgen::InsensitiveStr::Ascii("diffenrences"),
                &["differences"],
            ),
            (dictgen::InsensitiveStr::Ascii("alernated"), &["alternated"]),
            (dictgen::InsensitiveStr::Ascii("veiw"), &["view"]),
            (dictgen::InsensitiveStr::Ascii("witrhin"), &["within"]),
            (
                dictgen::InsensitiveStr::Ascii("seacrchable"),
                &["searchable"],
            ),
            (dictgen::InsensitiveStr::Ascii("monumnets"), &["monuments"]),
            (dictgen::InsensitiveStr::Ascii("enahncing"), &["enhancing"]),
            (
                dictgen::InsensitiveStr::Ascii("temperaturs"),
                &["temperatures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cardiovascualr"),
                &["cardiovascular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjsutment"),
                &["adjustment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pairlament"),
                &["parliament"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("plasing"),
                &["placing", "pleasing", "phasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circustances"),
                &["circumstances"],
            ),
            (dictgen::InsensitiveStr::Ascii("iventory"), &["inventory"]),
            (dictgen::InsensitiveStr::Ascii("cleff"), &["clef"]),
            (
                dictgen::InsensitiveStr::Ascii("confirmatinon"),
                &["confirmation"],
            ),
            (dictgen::InsensitiveStr::Ascii("midify"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("encoutnering"),
                &["encountering"],
            ),
            (dictgen::InsensitiveStr::Ascii("coomittee"), &["committee"]),
            (dictgen::InsensitiveStr::Ascii("clory"), &["glory"]),
            (
                dictgen::InsensitiveStr::Ascii("catagorical"),
                &["categorical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achievemnts"),
                &["achievements"],
            ),
            (dictgen::InsensitiveStr::Ascii("autsitic"), &["autistic"]),
            (dictgen::InsensitiveStr::Ascii("shortcutt"), &["shortcut"]),
            (dictgen::InsensitiveStr::Ascii("pacht"), &["patch"]),
            (
                dictgen::InsensitiveStr::Ascii("standardizin"),
                &["standardizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffult"), &["difficult"]),
            (dictgen::InsensitiveStr::Ascii("vaticina"), &["vatican"]),
            (dictgen::InsensitiveStr::Ascii("visuabize"), &["visualize"]),
            (
                dictgen::InsensitiveStr::Ascii("capablility"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("withoout"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("neigbours"), &["neighbours"]),
            (
                dictgen::InsensitiveStr::Ascii("exorted"),
                &["exported", "extorted", "exerted", "exhorted"],
            ),
            (dictgen::InsensitiveStr::Ascii("phasin"), &["phasing"]),
            (dictgen::InsensitiveStr::Ascii("abudance"), &["abundance"]),
            (
                dictgen::InsensitiveStr::Ascii("algorithmns"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("paranoica"), &["paranoia"]),
            (dictgen::InsensitiveStr::Ascii("creeating"), &["creating"]),
            (dictgen::InsensitiveStr::Ascii("availabele"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("faliures"), &["failures"]),
            (dictgen::InsensitiveStr::Ascii("realisin"), &["realising"]),
            (
                dictgen::InsensitiveStr::Ascii("chesckpoints"),
                &["checkpoints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchonously"),
                &["synchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscomunnication"),
                &["miscommunication"],
            ),
            (dictgen::InsensitiveStr::Ascii("explicatii"), &["explicit"]),
            (dictgen::InsensitiveStr::Ascii("unannimous"), &["unanimous"]),
            (
                dictgen::InsensitiveStr::Ascii("commitable"),
                &["committable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alcoholical"),
                &["alcoholic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wokr"),
                &["work", "woke", "wok"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnkng"), &["warning"]),
            (dictgen::InsensitiveStr::Ascii("tath"), &["that"]),
            (dictgen::InsensitiveStr::Ascii("viw"), &["view"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusigasm"),
                &["enthusiasm"],
            ),
            (dictgen::InsensitiveStr::Ascii("causalty"), &["casualty"]),
            (dictgen::InsensitiveStr::Ascii("dimissal"), &["dismissal"]),
            (dictgen::InsensitiveStr::Ascii("impicitly"), &["implicitly"]),
            (dictgen::InsensitiveStr::Ascii("messsaging"), &["messaging"]),
            (dictgen::InsensitiveStr::Ascii("bootsstrap"), &["bootstrap"]),
            (
                dictgen::InsensitiveStr::Ascii("influencial"),
                &["influential"],
            ),
            (dictgen::InsensitiveStr::Ascii("parites"), &["parties"]),
            (dictgen::InsensitiveStr::Ascii("imrpoves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("candinate"), &["candidate"]),
            (dictgen::InsensitiveStr::Ascii("colleted"), &["collected"]),
            (
                dictgen::InsensitiveStr::Ascii("intricasies"),
                &["intricacies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermimental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eliminting"),
                &["eliminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("leafs"), &["leaves"]),
            (
                dictgen::InsensitiveStr::Ascii("assued"),
                &["assumed", "assured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nipticking"),
                &["nitpicking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasferred"),
                &["transferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cinematograhy"),
                &["cinematography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innecessary"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("indefinet"), &["indefinite"]),
            (dictgen::InsensitiveStr::Ascii("acquirin"), &["acquiring"]),
            (
                dictgen::InsensitiveStr::Ascii("inititation"),
                &["initiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simpifications"),
                &["simplifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("allcator"), &["allocator"]),
            (dictgen::InsensitiveStr::Ascii("collge"), &["college"]),
            (
                dictgen::InsensitiveStr::Ascii("expiermental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("ois"), &["is"]),
            (
                dictgen::InsensitiveStr::Ascii("perfermance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("garunteed"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("reconstrcutions"),
                &["reconstructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("betweeen"), &["between"]),
            (
                dictgen::InsensitiveStr::Ascii("emphazises"),
                &["emphasises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rearrant"),
                &["rearrangement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsuported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("exatled"), &["exalted"]),
            (
                dictgen::InsensitiveStr::Ascii("combinatatorics"),
                &["combinatorics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vawdvillian"),
                &["vaudevillian"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimiced"), &["mimicked"]),
            (
                dictgen::InsensitiveStr::Ascii("occaisonal"),
                &["occasional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("boootstrapping"),
                &["bootstrapping"],
            ),
            (dictgen::InsensitiveStr::Ascii("inactvie"), &["inactive"]),
            (dictgen::InsensitiveStr::Ascii("insertas"), &["inserts"]),
            (
                dictgen::InsensitiveStr::Ascii("righteoussness"),
                &["righteousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devlopment"),
                &["development"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notifiaction"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("winnig"), &["winning"]),
            (
                dictgen::InsensitiveStr::Ascii("mircoservice"),
                &["microservice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecourage"),
                &["encourage", "courage"],
            ),
            (dictgen::InsensitiveStr::Ascii("unhealhty"), &["unhealthy"]),
            (
                dictgen::InsensitiveStr::Ascii("impelmentations"),
                &["implementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("matche"), &["match"]),
            (
                dictgen::InsensitiveStr::Ascii("corespondence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redunduncy"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("hierchy"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("surprisinlgy"),
                &["surprisingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("numners"), &["numbers"]),
            (dictgen::InsensitiveStr::Ascii("speperats"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("fiedling"), &["fielding"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlators"),
                &["translators"],
            ),
            (dictgen::InsensitiveStr::Ascii("androig"), &["android"]),
            (
                dictgen::InsensitiveStr::Ascii("transaltes"),
                &["translates"],
            ),
            (dictgen::InsensitiveStr::Ascii("analising"), &["analysing"]),
            (dictgen::InsensitiveStr::Ascii("heiarchy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("drawm"), &["drawn"]),
            (dictgen::InsensitiveStr::Ascii("increament"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("shittown"), &["shitton"]),
            (dictgen::InsensitiveStr::Ascii("adjasancy"), &["adjacency"]),
            (
                dictgen::InsensitiveStr::Ascii("dimenionsal"),
                &["dimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("filltered"), &["filtered"]),
            (
                dictgen::InsensitiveStr::Ascii("attribtuted"),
                &["attributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appicaliton"),
                &["application"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permanantely"),
                &["permanently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficifcations"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatonates"),
                &["concatenates"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupelicate"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("specialiases"),
                &["specialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("pryamid"), &["pyramid"]),
            (
                dictgen::InsensitiveStr::Ascii("seeverities"),
                &["severities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacturs"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regulaotry"),
                &["regulatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("shandow"), &["shadow"]),
            (dictgen::InsensitiveStr::Ascii("absurditiy"), &["absurdity"]),
            (dictgen::InsensitiveStr::Ascii("uers"), &["users", "user"]),
            (
                dictgen::InsensitiveStr::Ascii("natioanlism"),
                &["nationalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("europen"), &["european"]),
            (
                dictgen::InsensitiveStr::Ascii("portoguese"),
                &["portuguese"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sandiwches"),
                &["sandwiches"],
            ),
            (dictgen::InsensitiveStr::Ascii("extracors"), &["extractors"]),
            (dictgen::InsensitiveStr::Ascii("freinds"), &["friends"]),
            (dictgen::InsensitiveStr::Ascii("parttion"), &["partition"]),
            (dictgen::InsensitiveStr::Ascii("kicthen"), &["kitchen"]),
            (
                dictgen::InsensitiveStr::Ascii("equivelance"),
                &["equivalence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modefication"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("luminose"), &["luminous"]),
            (dictgen::InsensitiveStr::Ascii("wendesday"), &["wednesday"]),
            (dictgen::InsensitiveStr::Ascii("claer"), &["clear"]),
            (dictgen::InsensitiveStr::Ascii("opend"), &["opened"]),
            (
                dictgen::InsensitiveStr::Ascii("battlestsr"),
                &["battlestar"],
            ),
            (dictgen::InsensitiveStr::Ascii("bliting"), &["blitting"]),
            (dictgen::InsensitiveStr::Ascii("destoring"), &["destroying"]),
            (dictgen::InsensitiveStr::Ascii("broady"), &["broadly"]),
            (dictgen::InsensitiveStr::Ascii("decieds"), &["decides"]),
            (dictgen::InsensitiveStr::Ascii("dublicate"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("nostlagic"), &["nostalgic"]),
            (dictgen::InsensitiveStr::Ascii("aqueos"), &["aqueous"]),
            (
                dictgen::InsensitiveStr::Ascii("multilpliers"),
                &["multipliers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shemes"),
                &["schemes", "shames"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimtenelly"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("equivalen"), &["equivalent"]),
            (
                dictgen::InsensitiveStr::Ascii("reponsibility"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("auctionrs"), &["auctions"]),
            (dictgen::InsensitiveStr::Ascii("driectly"), &["directly"]),
            (dictgen::InsensitiveStr::Ascii("vitaliy"), &["vitality"]),
            (dictgen::InsensitiveStr::Ascii("thtat"), &["that"]),
            (
                dictgen::InsensitiveStr::Ascii("representated"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("legitimancy"),
                &["legitimacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reperesents"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("crypted"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotion"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srolls"),
                &["scrolls", "strolls"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accumilating"),
                &["accumulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("computaion"),
                &["computation"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeedde"), &["succeeded"]),
            (dictgen::InsensitiveStr::Ascii("buddishm"), &["buddhism"]),
            (dictgen::InsensitiveStr::Ascii("inpending"), &["impending"]),
            (dictgen::InsensitiveStr::Ascii("academicos"), &["academics"]),
            (dictgen::InsensitiveStr::Ascii("realeases"), &["releases"]),
            (dictgen::InsensitiveStr::Ascii("judical"), &["judicial"]),
            (
                dictgen::InsensitiveStr::Ascii("disporving"),
                &["disproving"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atuhenticator"),
                &["authenticator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coersitive"),
                &["coercitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("oeprators"), &["operators"]),
            (dictgen::InsensitiveStr::Ascii("murdererd"), &["murdered"]),
            (
                dictgen::InsensitiveStr::Ascii("somethines"),
                &["somethings"],
            ),
            (dictgen::InsensitiveStr::Ascii("arived"), &["arrived"]),
            (
                dictgen::InsensitiveStr::Ascii("baasing"),
                &["basing", "abasing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civalasations"),
                &["civilisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delfection"),
                &["deflection"],
            ),
            (dictgen::InsensitiveStr::Ascii("scources"), &["sources"]),
            (dictgen::InsensitiveStr::Ascii("jkd"), &["jdk"]),
            (dictgen::InsensitiveStr::Ascii("followiing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("activeta"), &["activate"]),
            (dictgen::InsensitiveStr::Ascii("wronly"), &["wrongly"]),
            (dictgen::InsensitiveStr::Ascii("speeeds"), &["speeds"]),
            (dictgen::InsensitiveStr::Ascii("uninistall"), &["uninstall"]),
            (dictgen::InsensitiveStr::Ascii("modulu"), &["modulo"]),
            (dictgen::InsensitiveStr::Ascii("capapbly"), &["capably"]),
            (
                dictgen::InsensitiveStr::Ascii("mandatatory"),
                &["mandatory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spacialization"),
                &["specialization"],
            ),
            (dictgen::InsensitiveStr::Ascii("auotmatic"), &["automatic"]),
            (
                dictgen::InsensitiveStr::Ascii("applicaable"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("visuaizes"), &["visualizes"]),
            (dictgen::InsensitiveStr::Ascii("sugests"), &["suggests"]),
            (dictgen::InsensitiveStr::Ascii("kwno"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("accesseries"),
                &["accessories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("affiliaton"),
                &["affiliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polypoygon"),
                &["polypolygon"],
            ),
            (dictgen::InsensitiveStr::Ascii("ervice"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("confiurable"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("downoads"), &["downloads"]),
            (dictgen::InsensitiveStr::Ascii("exculsion"), &["exclusion"]),
            (dictgen::InsensitiveStr::Ascii("vacumme"), &["vacuum"]),
            (dictgen::InsensitiveStr::Ascii("rendevous"), &["rendezvous"]),
            (dictgen::InsensitiveStr::Ascii("analysits"), &["analysts"]),
            (dictgen::InsensitiveStr::Ascii("moonligt"), &["moonlight"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplicative"),
                &["multiplicative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paltformer"),
                &["platformer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chaeckerboard"),
                &["checkerboard"],
            ),
            (dictgen::InsensitiveStr::Ascii("defuct"), &["defunct"]),
            (dictgen::InsensitiveStr::Ascii("variuos"), &["various"]),
            (dictgen::InsensitiveStr::Ascii("manifoild"), &["manifold"]),
            (dictgen::InsensitiveStr::Ascii("maximnums"), &["maximums"]),
            (dictgen::InsensitiveStr::Ascii("incrased"), &["increased"]),
            (
                dictgen::InsensitiveStr::Ascii("disconnectme"),
                &["disconnected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturbaton"),
                &["masturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("draclua"), &["dracula"]),
            (
                dictgen::InsensitiveStr::Ascii("circunferences"),
                &["circumferences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemenations"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatiblilties"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sportmansship"),
                &["sportsmanship"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaluetes"), &["evaluates"]),
            (
                dictgen::InsensitiveStr::Ascii("cotnainers"),
                &["containers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silksceened"),
                &["silkscreened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminare"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("conetion"), &["connection"]),
            (
                dictgen::InsensitiveStr::Ascii("reasonabily"),
                &["reasonably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facisnated"),
                &["fascinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retetting"),
                &["resetting", "retesting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwhleming"),
                &["overwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirucit"), &["circuit"]),
            (
                dictgen::InsensitiveStr::Ascii("representetive"),
                &["representative"],
            ),
            (dictgen::InsensitiveStr::Ascii("garanteed"), &["guaranteed"]),
            (
                dictgen::InsensitiveStr::Ascii("accruately"),
                &["accurately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrignement"),
                &["infringement"],
            ),
            (dictgen::InsensitiveStr::Ascii("comlained"), &["complained"]),
            (dictgen::InsensitiveStr::Ascii("coninue"), &["continue"]),
            (dictgen::InsensitiveStr::Ascii("univrsal"), &["universal"]),
            (dictgen::InsensitiveStr::Ascii("pevent"), &["prevent"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprenaurs"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permuatates"),
                &["permutates"],
            ),
            (dictgen::InsensitiveStr::Ascii("vestigal"), &["vestigial"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisisst"),
                &["narcissist"],
            ),
            (dictgen::InsensitiveStr::Ascii("analsyers"), &["analysers"]),
            (
                dictgen::InsensitiveStr::Ascii("hemipshere"),
                &["hemisphere"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appriciative"),
                &["appreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satasfactory"),
                &["satisfactory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transconded"),
                &["transcoded"],
            ),
            (dictgen::InsensitiveStr::Ascii("lefitsts"), &["leftists"]),
            (dictgen::InsensitiveStr::Ascii("platoforms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("implemnted"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrament"),
                &["instrument"],
            ),
            (dictgen::InsensitiveStr::Ascii("exchanched"), &["exchanged"]),
            (dictgen::InsensitiveStr::Ascii("reactquire"), &["reacquire"]),
            (dictgen::InsensitiveStr::Ascii("collaegue"), &["colleague"]),
            (
                dictgen::InsensitiveStr::Ascii("traingular"),
                &["triangular"],
            ),
            (dictgen::InsensitiveStr::Ascii("cambodican"), &["cambodia"]),
            (dictgen::InsensitiveStr::Ascii("pryamids"), &["pyramids"]),
            (
                dictgen::InsensitiveStr::Ascii("ghostscipt"),
                &["ghostscript"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exhustiveness"),
                &["exhaustiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("occors"), &["occurs"]),
            (
                dictgen::InsensitiveStr::Ascii("estiamators"),
                &["estimators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nojification"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("paticular"), &["particular"]),
            (dictgen::InsensitiveStr::Ascii("sophmores"), &["sophomores"]),
            (dictgen::InsensitiveStr::Ascii("cynaide"), &["cyanide"]),
            (dictgen::InsensitiveStr::Ascii("exposees"), &["exposes"]),
            (dictgen::InsensitiveStr::Ascii("varmits"), &["varmints"]),
            (
                dictgen::InsensitiveStr::Ascii("visiter"),
                &["visitor", "visit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("farmasudics"),
                &["pharmaceutics"],
            ),
            (dictgen::InsensitiveStr::Ascii("arugement"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("definde"),
                &["define", "defined", "defines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recpetionist"),
                &["receptionist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("altriustic"),
                &["altruistic"],
            ),
            (dictgen::InsensitiveStr::Ascii("clasroom"), &["classroom"]),
            (dictgen::InsensitiveStr::Ascii("meail"), &["email"]),
            (
                dictgen::InsensitiveStr::Ascii("neightbourhoods"),
                &["neighbourhoods"],
            ),
            (dictgen::InsensitiveStr::Ascii("pathwyas"), &["pathways"]),
            (
                dictgen::InsensitiveStr::Ascii("inernally"),
                &["internally", "infernally"],
            ),
            (dictgen::InsensitiveStr::Ascii("massagin"), &["massaging"]),
            (dictgen::InsensitiveStr::Ascii("cylcles"), &["cycles"]),
            (
                dictgen::InsensitiveStr::Ascii("cacluators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconfortability"),
                &["discomfort"],
            ),
            (dictgen::InsensitiveStr::Ascii("opimising"), &["optimising"]),
            (dictgen::InsensitiveStr::Ascii("dispatcgh"), &["dispatch"]),
            (dictgen::InsensitiveStr::Ascii("guideded"), &["guided"]),
            (dictgen::InsensitiveStr::Ascii("repsects"), &["respects"]),
            (
                dictgen::InsensitiveStr::Ascii("calibraiton"),
                &["calibration"],
            ),
            (dictgen::InsensitiveStr::Ascii("intaller"), &["installer"]),
            (dictgen::InsensitiveStr::Ascii("agrement"), &["agreement"]),
            (dictgen::InsensitiveStr::Ascii("uplaods"), &["uploads"]),
            (dictgen::InsensitiveStr::Ascii("sysmtes"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("santioned"), &["sanctioned"]),
            (dictgen::InsensitiveStr::Ascii("miminises"), &["minimises"]),
            (dictgen::InsensitiveStr::Ascii("initliase"), &["initialise"]),
            (dictgen::InsensitiveStr::Ascii("handlong"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("cicrulating"),
                &["circulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheching"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("insturcts"), &["instructs"]),
            (dictgen::InsensitiveStr::Ascii("medisinal"), &["medicinal"]),
            (
                dictgen::InsensitiveStr::Ascii("conisderation"),
                &["considerations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vehementely"),
                &["vehemently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fellowhsip"),
                &["fellowship"],
            ),
            (dictgen::InsensitiveStr::Ascii("lightrom"), &["lightroom"]),
            (dictgen::InsensitiveStr::Ascii("beloning"), &["belonging"]),
            (dictgen::InsensitiveStr::Ascii("epliogue"), &["epilogue"]),
            (
                dictgen::InsensitiveStr::Ascii("delclaration"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoghts"), &["thoughts"]),
            (
                dictgen::InsensitiveStr::Ascii("minipulating"),
                &["manipulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fededing"), &["feeding"]),
            (
                dictgen::InsensitiveStr::Ascii("quicky"),
                &["quickie", "quickly"],
            ),
            (dictgen::InsensitiveStr::Ascii("packge"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("securtity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("healthiet"), &["healthiest"]),
            (dictgen::InsensitiveStr::Ascii("tentaive"), &["tentative"]),
            (
                dictgen::InsensitiveStr::Ascii("controversity"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("checllist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("mememto"), &["memento"]),
            (dictgen::InsensitiveStr::Ascii("chariman"), &["chairman"]),
            (
                dictgen::InsensitiveStr::Ascii("denine"),
                &["define", "adenine", "dentine"],
            ),
            (dictgen::InsensitiveStr::Ascii("worfklows"), &["workflows"]),
            (dictgen::InsensitiveStr::Ascii("mappped"), &["mapped"]),
            (
                dictgen::InsensitiveStr::Ascii("architeturally"),
                &["architecturally"],
            ),
            (dictgen::InsensitiveStr::Ascii("addted"), &["added"]),
            (
                dictgen::InsensitiveStr::Ascii("customizaton"),
                &["customization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cimptomaticlly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("procuction"),
                &["production"],
            ),
            (dictgen::InsensitiveStr::Ascii("exterioare"), &["exterior"]),
            (
                dictgen::InsensitiveStr::Ascii("chromisome"),
                &["chromosome"],
            ),
            (dictgen::InsensitiveStr::Ascii("consonate"), &["consonant"]),
            (
                dictgen::InsensitiveStr::Ascii("sensasionalist"),
                &["sensationalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("spwaned"), &["spawned"]),
            (dictgen::InsensitiveStr::Ascii("exhaused"), &["exhausted"]),
            (dictgen::InsensitiveStr::Ascii("wonderus"), &["wonders"]),
            (
                dictgen::InsensitiveStr::Ascii("flattenning"),
                &["flattening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("signularities"),
                &["singularities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explinatory"),
                &["explanatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliences"), &["silences"]),
            (dictgen::InsensitiveStr::Ascii("counterd"), &["countered"]),
            (dictgen::InsensitiveStr::Ascii("syrap"), &["syrup"]),
            (dictgen::InsensitiveStr::Ascii("extermly"), &["extremely"]),
            (
                dictgen::InsensitiveStr::Ascii("lingusitic"),
                &["linguistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("relativno"), &["relation"]),
            (dictgen::InsensitiveStr::Ascii("esteblish"), &["establish"]),
            (
                dictgen::InsensitiveStr::Ascii("crations"),
                &["creations", "nations", "rations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinementers"),
                &["experimenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribiuting"),
                &["attributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("peirods"), &["periods"]),
            (dictgen::InsensitiveStr::Ascii("scracthes"), &["scratches"]),
            (dictgen::InsensitiveStr::Ascii("narcoticos"), &["narcotics"]),
            (
                dictgen::InsensitiveStr::Ascii("entreprenur"),
                &["entrepreneur"],
            ),
            (dictgen::InsensitiveStr::Ascii("mainain"), &["maintain"]),
            (dictgen::InsensitiveStr::Ascii("simialir"), &["similar"]),
            (
                dictgen::InsensitiveStr::Ascii("pojrections"),
                &["projections"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynchornous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regessions"),
                &["regressions", "recessions"],
            ),
            (dictgen::InsensitiveStr::Ascii("ablum"), &["album"]),
            (
                dictgen::InsensitiveStr::Ascii("acculumated"),
                &["accumulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("speechers"), &["speeches"]),
            (dictgen::InsensitiveStr::Ascii("encalve"), &["enclave"]),
            (dictgen::InsensitiveStr::Ascii("modfiers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("pannel"), &["panel"]),
            (dictgen::InsensitiveStr::Ascii("montior"), &["monitor"]),
            (
                dictgen::InsensitiveStr::Ascii("presentaiton"),
                &["presentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforser"),
                &["endorser", "enforcer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("owerpowering"),
                &["overpowering"],
            ),
            (dictgen::InsensitiveStr::Ascii("anecdotles"), &["anecdotes"]),
            (
                dictgen::InsensitiveStr::Ascii("govnerment"),
                &["government"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassamble"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("measurmements"),
                &["measurements"],
            ),
            (dictgen::InsensitiveStr::Ascii("extremaste"), &["extremes"]),
            (
                dictgen::InsensitiveStr::Ascii("muncipality"),
                &["municipality"],
            ),
            (dictgen::InsensitiveStr::Ascii("materils"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("construt"),
                &["constructs", "construct"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expilicitely"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("merchantablity"),
                &["merchantability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redefinintions"),
                &["redefinitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approprately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("vektors"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("platfomr"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("encrypitng"),
                &["encrypting"],
            ),
            (dictgen::InsensitiveStr::Ascii("falure"), &["failure"]),
            (
                dictgen::InsensitiveStr::Ascii("communiaction"),
                &["communications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exmplified"),
                &["exemplified"],
            ),
            (dictgen::InsensitiveStr::Ascii("simualtes"), &["simulates"]),
            (
                dictgen::InsensitiveStr::Ascii("correspnds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotites"),
                &["renegotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("vermen"), &["vermin"]),
            (dictgen::InsensitiveStr::Ascii("assma"), &["asthma"]),
            (dictgen::InsensitiveStr::Ascii("numberical"), &["numerical"]),
            (dictgen::InsensitiveStr::Ascii("celesital"), &["celestial"]),
            (dictgen::InsensitiveStr::Ascii("traitoris"), &["traitors"]),
            (dictgen::InsensitiveStr::Ascii("scalarr"), &["scalar"]),
            (dictgen::InsensitiveStr::Ascii("otpimises"), &["optimises"]),
            (dictgen::InsensitiveStr::Ascii("nothind"), &["nothing"]),
            (dictgen::InsensitiveStr::Ascii("distirted"), &["distorted"]),
            (dictgen::InsensitiveStr::Ascii("extremers"), &["extremes"]),
            (
                dictgen::InsensitiveStr::Ascii("approximete"),
                &["approximate"],
            ),
            (dictgen::InsensitiveStr::Ascii("spred"), &["spread"]),
            (dictgen::InsensitiveStr::Ascii("asembles"), &["assembles"]),
            (dictgen::InsensitiveStr::Ascii("dentified"), &["identified"]),
            (dictgen::InsensitiveStr::Ascii("wrting"), &["writing"]),
            (dictgen::InsensitiveStr::Ascii("paltforms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("callibrated"),
                &["calibrated"],
            ),
            (dictgen::InsensitiveStr::Ascii("cooments"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("chnaged"), &["changed"]),
            (
                dictgen::InsensitiveStr::Ascii("contraversies"),
                &["controversies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serializin"),
                &["serializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("picknicked"), &["picnicked"]),
            (dictgen::InsensitiveStr::Ascii("stealthboy"), &["stealthy"]),
            (dictgen::InsensitiveStr::Ascii("inidicated"), &["indicated"]),
            (dictgen::InsensitiveStr::Ascii("terurn"), &["return"]),
            (dictgen::InsensitiveStr::Ascii("migt"), &["might", "midget"]),
            (
                dictgen::InsensitiveStr::Ascii("tablepsace"),
                &["tablespace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extensable"),
                &["extensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("divsion"), &["division"]),
            (
                dictgen::InsensitiveStr::Ascii("pronnounced"),
                &["pronounced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copilate"),
                &["compile", "copilot", "copulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("activaed"), &["activated"]),
            (dictgen::InsensitiveStr::Ascii("tolerent"), &["tolerant"]),
            (
                dictgen::InsensitiveStr::Ascii("cholosterol"),
                &["cholesterol"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guatamalan"),
                &["guatemalan"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outnumbred"),
                &["outnumbered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistenty"),
                &["inconsistency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("structoring"),
                &["structuring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condamnation"),
                &["condemnation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deffinitely"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immedialty"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("membershup"),
                &["membership"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reaises"),
                &["realises", "raises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropriateely"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("intefaces"), &["interfaces"]),
            (dictgen::InsensitiveStr::Ascii("efficiant"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("recieving"), &["receiving"]),
            (dictgen::InsensitiveStr::Ascii("deadlfit"), &["deadlift"]),
            (
                dictgen::InsensitiveStr::Ascii("implemenetd"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("theef"), &["thief"]),
            (dictgen::InsensitiveStr::Ascii("popullated"), &["populated"]),
            (
                dictgen::InsensitiveStr::Ascii("optimisitic"),
                &["optimistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vrsatility"),
                &["versatility"],
            ),
            (dictgen::InsensitiveStr::Ascii("therad"), &["thread"]),
            (dictgen::InsensitiveStr::Ascii("ists"), &["its", "lists"]),
            (dictgen::InsensitiveStr::Ascii("parocheel"), &["parochial"]),
            (
                dictgen::InsensitiveStr::Ascii("simutaneously"),
                &["simultaneously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("philippinnes"),
                &["philippines"],
            ),
            (dictgen::InsensitiveStr::Ascii("stremear"), &["streamer"]),
            (dictgen::InsensitiveStr::Ascii("withourt"), &["without"]),
            (dictgen::InsensitiveStr::Ascii("packgages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("experimally"),
                &["experimentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("essental"), &["essential"]),
            (
                dictgen::InsensitiveStr::Ascii("developes"),
                &["develops", "developed", "developers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sarter"), &["starter"]),
            (
                dictgen::InsensitiveStr::Ascii("indepentently"),
                &["independently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innaccible"),
                &["inaccessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("onslaugth"), &["onslaught"]),
            (dictgen::InsensitiveStr::Ascii("buguous"), &["bogus"]),
            (
                dictgen::InsensitiveStr::Ascii("disconneted"),
                &["disconnected"],
            ),
            (dictgen::InsensitiveStr::Ascii("scroolling"), &["scrolling"]),
            (
                dictgen::InsensitiveStr::Ascii("whiltelists"),
                &["whitelists"],
            ),
            (dictgen::InsensitiveStr::Ascii("posess"), &["possess"]),
            (
                dictgen::InsensitiveStr::Ascii("retargetted"),
                &["retargeted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coeffiecient"),
                &["coefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("hedgehoog"), &["hedgehog"]),
            (dictgen::InsensitiveStr::Ascii("substace"), &["substance"]),
            (dictgen::InsensitiveStr::Ascii("formmatted"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("coommented"), &["commented"]),
            (
                dictgen::InsensitiveStr::Ascii("indavidual"),
                &["individual"],
            ),
            (dictgen::InsensitiveStr::Ascii("upgradingn"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("correspods"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trinagulate"),
                &["triangulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chalengers"),
                &["challengers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contection"),
                &["contention"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expoentntial"),
                &["exponential"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defniately"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contentous"),
                &["contentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itnerfaces"),
                &["interfaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commadline"),
                &["commandline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introduceres"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carboyhdrates"),
                &["carbohydrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("structrue"), &["structure"]),
            (
                dictgen::InsensitiveStr::Ascii("assiociating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfomational"),
                &["transformational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accustommed"),
                &["accustomed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("customzies"),
                &["customizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intitalization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("facilitarte"),
                &["facilitate"],
            ),
            (dictgen::InsensitiveStr::Ascii("konstant"), &["constant"]),
            (
                dictgen::InsensitiveStr::Ascii("extraordinaryly"),
                &["extraordinarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("hodling"), &["holding"]),
            (
                dictgen::InsensitiveStr::Ascii("oscillatin"),
                &["oscillating", "oscillation"],
            ),
            (dictgen::InsensitiveStr::Ascii("succeded"), &["succeeded"]),
            (
                dictgen::InsensitiveStr::Ascii("ternimators"),
                &["terminators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qaulification"),
                &["qualification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deccelerated"),
                &["decelerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("chck"), &["check"]),
            (dictgen::InsensitiveStr::Ascii("maanges"), &["manages"]),
            (dictgen::InsensitiveStr::Ascii("bloaded"), &["bloated"]),
            (
                dictgen::InsensitiveStr::Ascii("identificato"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bounderies"),
                &["boundaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("monolgoue"), &["monologue"]),
            (
                dictgen::InsensitiveStr::Ascii("narcisissim"),
                &["narcissism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cemetaries"),
                &["cemeteries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pornograhy"),
                &["pornography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexdecimal"),
                &["hexadecimal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrubutor"),
                &["contributor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpbabetically"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("nummerator"), &["numerator"]),
            (dictgen::InsensitiveStr::Ascii("inizials"), &["initials"]),
            (
                dictgen::InsensitiveStr::Ascii("penetrading"),
                &["penetrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interestigly"),
                &["interestingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bellweathers"),
                &["bellwethers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estimaters"),
                &["estimators", "estimates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mirors"),
                &["mirrors", "minors"],
            ),
            (dictgen::InsensitiveStr::Ascii("powerfull"), &["powerful"]),
            (dictgen::InsensitiveStr::Ascii("lowcase"), &["lowercase"]),
            (dictgen::InsensitiveStr::Ascii("bengalos"), &["bengals"]),
            (
                dictgen::InsensitiveStr::Ascii("curruptions"),
                &["corruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemtpted"), &["attempted"]),
            (
                dictgen::InsensitiveStr::Ascii("exeriments"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("egde"), &["edge"]),
            (
                dictgen::InsensitiveStr::Ascii("mutatbility"),
                &["mutability"],
            ),
            (dictgen::InsensitiveStr::Ascii("inpt"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("punishs"), &["punishes"]),
            (
                dictgen::InsensitiveStr::Ascii("interaces"),
                &["interfaces", "interlaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infinitie"),
                &["infinite", "infinity"],
            ),
            (dictgen::InsensitiveStr::Ascii("anwser"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("atfernoon"), &["afternoon"]),
            (
                dictgen::InsensitiveStr::Ascii("electrictian"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transactinos"),
                &["transactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("toosday"), &["tuesday"]),
            (dictgen::InsensitiveStr::Ascii("expressin"), &["expression"]),
            (dictgen::InsensitiveStr::Ascii("supportin"), &["supporting"]),
            (dictgen::InsensitiveStr::Ascii("circkets"), &["crickets"]),
            (
                dictgen::InsensitiveStr::Ascii("consumation"),
                &["consumption", "consummation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instantialte"),
                &["instantiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ossociated"),
                &["associated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controversary"),
                &["controversy"],
            ),
            (dictgen::InsensitiveStr::Ascii("drawble"), &["drawable"]),
            (dictgen::InsensitiveStr::Ascii("bastars"), &["bastards"]),
            (dictgen::InsensitiveStr::Ascii("knigths"), &["knights"]),
            (
                dictgen::InsensitiveStr::Ascii("approxamations"),
                &["approximations"],
            ),
            (dictgen::InsensitiveStr::Ascii("inoquous"), &["innocuous"]),
            (dictgen::InsensitiveStr::Ascii("negatve"), &["negative"]),
            (dictgen::InsensitiveStr::Ascii("disconect"), &["disconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("documntation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("brokem"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("assosication"),
                &["assassination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetes"),
                &["repeats", "reputes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrairly"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geomoetrically"),
                &["geometrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("mata"), &["meta", "mater"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("pumkpins"), &["pumpkins"]),
            (dictgen::InsensitiveStr::Ascii("traning"), &["training"]),
            (dictgen::InsensitiveStr::Ascii("decisiion"), &["decision"]),
            (
                dictgen::InsensitiveStr::Ascii("autocommiting"),
                &["autocommitting"],
            ),
            (dictgen::InsensitiveStr::Ascii("aimation"), &["animation"]),
            (dictgen::InsensitiveStr::Ascii("menstrul"), &["menstrual"]),
            (
                dictgen::InsensitiveStr::Ascii("otification"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("questios"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("sucesses"), &["successes"]),
            (dictgen::InsensitiveStr::Ascii("trotskism"), &["trotskyism"]),
            (dictgen::InsensitiveStr::Ascii("dowt"), &["doubt"]),
            (
                dictgen::InsensitiveStr::Ascii("interperts"),
                &["interprets"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesory"), &["accessory"]),
            (
                dictgen::InsensitiveStr::Ascii("disqaulified"),
                &["disqualified"],
            ),
            (dictgen::InsensitiveStr::Ascii("heirarchy"), &["hierarchy"]),
            (dictgen::InsensitiveStr::Ascii("releiving"), &["relieving"]),
            (
                dictgen::InsensitiveStr::Ascii("arbitrers"),
                &["arbiters", "arbitrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("econimical"),
                &["economical"],
            ),
            (dictgen::InsensitiveStr::Ascii("prgrammed"), &["programmed"]),
            (
                dictgen::InsensitiveStr::Ascii("thease"),
                &["these", "tease"],
            ),
            (dictgen::InsensitiveStr::Ascii("lifcycle"), &["lifecycle"]),
            (dictgen::InsensitiveStr::Ascii("waritable"), &["writable"]),
            (
                dictgen::InsensitiveStr::Ascii("malfunktion"),
                &["malfunction"],
            ),
            (dictgen::InsensitiveStr::Ascii("rithmic"), &["rhythmic"]),
            (
                dictgen::InsensitiveStr::Ascii("delcaratively"),
                &["declaratively"],
            ),
            (dictgen::InsensitiveStr::Ascii("cordump"), &["coredump"]),
            (
                dictgen::InsensitiveStr::Ascii("creaes"),
                &["creates", "creases", "crease"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supplimenting"),
                &["supplementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("altenative"),
                &["alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fertilizar"),
                &["fertilizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaccessiable"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approching"),
                &["approaching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consquence"),
                &["consequence"],
            ),
            (dictgen::InsensitiveStr::Ascii("avaition"), &["aviation"]),
            (dictgen::InsensitiveStr::Ascii("advane"), &["advance"]),
            (
                dictgen::InsensitiveStr::Ascii("probailistic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialziation"),
                &["serialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consturction"),
                &["construction"],
            ),
            (dictgen::InsensitiveStr::Ascii("ergnomic"), &["ergonomic"]),
            (dictgen::InsensitiveStr::Ascii("abanond"), &["abandon"]),
            (
                dictgen::InsensitiveStr::Ascii("midfielers"),
                &["midfielders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elliminates"),
                &["eliminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arcitectural"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("exacutors"), &["executors"]),
            (
                dictgen::InsensitiveStr::Ascii("playthroughts"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("conigs"), &["configs"]),
            (
                dictgen::InsensitiveStr::Ascii("shampionship"),
                &["championship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symphatize"),
                &["sympathize"],
            ),
            (dictgen::InsensitiveStr::Ascii("distorto"), &["distortion"]),
            (
                dictgen::InsensitiveStr::Ascii("apologizin"),
                &["apologizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compulsorary"),
                &["compulsory"],
            ),
            (dictgen::InsensitiveStr::Ascii("popoulator"), &["populator"]),
            (dictgen::InsensitiveStr::Ascii("commpile"), &["compile"]),
            (
                dictgen::InsensitiveStr::Ascii("notwithstandin"),
                &["notwithstanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("verticle"), &["vertical"]),
            (
                dictgen::InsensitiveStr::Ascii("anitquated"),
                &["antiquated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undestanding"),
                &["understanding"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalute"), &["evaluate"]),
            (
                dictgen::InsensitiveStr::Ascii("defautling"),
                &["defaulting"],
            ),
            (dictgen::InsensitiveStr::Ascii("delievers"), &["delivers"]),
            (
                dictgen::InsensitiveStr::Ascii("connotacion"),
                &["connotation"],
            ),
            (dictgen::InsensitiveStr::Ascii("asscciate"), &["associate"]),
            (
                dictgen::InsensitiveStr::Ascii("accodomated"),
                &["accommodated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ressolution"),
                &["resolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compeletely"),
                &["completely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propulsing"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("nervouse"), &["nervous"]),
            (
                dictgen::InsensitiveStr::Ascii("reconsilded"),
                &["reconciled"],
            ),
            (dictgen::InsensitiveStr::Ascii("coloumns"), &["columns"]),
            (
                dictgen::InsensitiveStr::Ascii("precisision"),
                &["precision"],
            ),
            (dictgen::InsensitiveStr::Ascii("tranpsort"), &["transport"]),
            (
                dictgen::InsensitiveStr::Ascii("depenences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("insigned"), &["unsigned"]),
            (dictgen::InsensitiveStr::Ascii("substasks"), &["subtasks"]),
            (
                dictgen::InsensitiveStr::Ascii("akumulative"),
                &["accumulative"],
            ),
            (dictgen::InsensitiveStr::Ascii("inherithed"), &["inherited"]),
            (
                dictgen::InsensitiveStr::Ascii("cnofiguration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referecencing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("changeing"), &["changing"]),
            (
                dictgen::InsensitiveStr::Ascii("unrecogized"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("devic"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("implemente"),
                &["implement", "implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("gental"), &["gentle"]),
            (dictgen::InsensitiveStr::Ascii("levleing"), &["leveling"]),
            (dictgen::InsensitiveStr::Ascii("confiug"), &["config"]),
            (dictgen::InsensitiveStr::Ascii("varns"), &["warns"]),
            (
                dictgen::InsensitiveStr::Ascii("checkpints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("attcahed"), &["attached"]),
            (
                dictgen::InsensitiveStr::Ascii("teminators"),
                &["terminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("higheset"), &["highest"]),
            (dictgen::InsensitiveStr::Ascii("unencrpt"), &["unencrypt"]),
            (
                dictgen::InsensitiveStr::Ascii("cooridnated"),
                &["coordinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnistalled"),
                &["uninstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provice"),
                &["provide", "province"],
            ),
            (dictgen::InsensitiveStr::Ascii("requeseted"), &["requested"]),
            (
                dictgen::InsensitiveStr::Ascii("exhaustivo"),
                &["exhaustion"],
            ),
            (dictgen::InsensitiveStr::Ascii("mchanism"), &["mechanism"]),
            (dictgen::InsensitiveStr::Ascii("literall"), &["literal"]),
            (
                dictgen::InsensitiveStr::Ascii("aforementionned"),
                &["aforementioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comfterble"),
                &["comfortable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celllar"),
                &["cellular", "cellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roduction"),
                &["reduction", "production"],
            ),
            (dictgen::InsensitiveStr::Ascii("dirrection"), &["direction"]),
            (dictgen::InsensitiveStr::Ascii("checg"), &["check"]),
            (
                dictgen::InsensitiveStr::Ascii("celebracion"),
                &["celebration"],
            ),
            (dictgen::InsensitiveStr::Ascii("htacess"), &["htaccess"]),
            (
                dictgen::InsensitiveStr::Ascii("assests"),
                &["assets", "assess"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illegitamate"),
                &["illegitimate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cotributor"),
                &["contributor"],
            ),
            (dictgen::InsensitiveStr::Ascii("emtional"), &["emotional"]),
            (dictgen::InsensitiveStr::Ascii("boardcast"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("passord"), &["password"]),
            (
                dictgen::InsensitiveStr::Ascii("screns"),
                &["screens", "screws"],
            ),
            (dictgen::InsensitiveStr::Ascii("uppload"), &["upload"]),
            (dictgen::InsensitiveStr::Ascii("detriot"), &["detroit"]),
            (dictgen::InsensitiveStr::Ascii("thingie"), &["thingy"]),
            (dictgen::InsensitiveStr::Ascii("worht"), &["worth"]),
            (
                dictgen::InsensitiveStr::Ascii("specically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("montanna"), &["montana"]),
            (dictgen::InsensitiveStr::Ascii("yeild"), &["yield"]),
            (dictgen::InsensitiveStr::Ascii("embding"), &["embedding"]),
            (dictgen::InsensitiveStr::Ascii("perssious"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("qeues"), &["queues"]),
            (
                dictgen::InsensitiveStr::Ascii("hypotherical"),
                &["hypothetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regrigerator"),
                &["refrigerator"],
            ),
            (dictgen::InsensitiveStr::Ascii("sfety"), &["safety"]),
            (dictgen::InsensitiveStr::Ascii("triffers"), &["triggers"]),
            (
                dictgen::InsensitiveStr::Ascii("contracto"),
                &["contraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transmited"),
                &["transmitted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschologists"),
                &["psychologists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metrices"),
                &["metrics", "matrices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassemle"),
                &["disassemble"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidiate"),
                &["intimidate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaiting"),
                &["negotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("combaining"), &["combining"]),
            (dictgen::InsensitiveStr::Ascii("selekted"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrevating"),
                &["abbreviating"],
            ),
            (dictgen::InsensitiveStr::Ascii("veitnam"), &["vietnam"]),
            (
                dictgen::InsensitiveStr::Ascii("cigarretes"),
                &["cigarettes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pathfidner"),
                &["pathfinder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discrimanted"),
                &["discriminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shenanigens"),
                &["shenanigans"],
            ),
            (dictgen::InsensitiveStr::Ascii("subfoilder"), &["subdolder"]),
            (dictgen::InsensitiveStr::Ascii("meesage"), &["message"]),
            (dictgen::InsensitiveStr::Ascii("shortkut"), &["shortcut"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibitied"),
                &["prohibited"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meditacion"),
                &["medications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immedeately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("remenber"), &["remember"]),
            (
                dictgen::InsensitiveStr::Ascii("propuslion"),
                &["propulsion"],
            ),
            (dictgen::InsensitiveStr::Ascii("autorize"), &["authorize"]),
            (
                dictgen::InsensitiveStr::Ascii("serialializes"),
                &["serializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laughablely"),
                &["laughably"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaing"), &["chain"]),
            (dictgen::InsensitiveStr::Ascii("chillade"), &["chilled"]),
            (dictgen::InsensitiveStr::Ascii("differred"), &["differed"]),
            (dictgen::InsensitiveStr::Ascii("rickoshays"), &["ricochets"]),
            (dictgen::InsensitiveStr::Ascii("agaian"), &["again"]),
            (dictgen::InsensitiveStr::Ascii("embrio"), &["embryo"]),
            (dictgen::InsensitiveStr::Ascii("ouur"), &["our"]),
            (
                dictgen::InsensitiveStr::Ascii("culiminated"),
                &["culminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimanter"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pruposefully"),
                &["purposefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqeuest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("tremendoulsy"),
                &["tremendously"],
            ),
            (dictgen::InsensitiveStr::Ascii("deecorator"), &["decorator"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensionals"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("adbuction"), &["abduction"]),
            (
                dictgen::InsensitiveStr::Ascii("nessarily"),
                &["necessarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custamizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("quielty"), &["quietly"]),
            (dictgen::InsensitiveStr::Ascii("apyoon"), &["oppugn"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerarbilities"),
                &["vulnerabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reletes"),
                &["deletes", "relates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammortize"), &["amortize"]),
            (
                dictgen::InsensitiveStr::Ascii("uninitialzied"),
                &["uninitialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("unstabel"), &["unstable"]),
            (
                dictgen::InsensitiveStr::Ascii("bufferes"),
                &["buffered", "buffers"],
            ),
            (dictgen::InsensitiveStr::Ascii("discovey"), &["discovery"]),
            (
                dictgen::InsensitiveStr::Ascii("reinfocing"),
                &["reinforcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("deffault"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("outclasssed"),
                &["outclassed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metamophosis"),
                &["metamorphosis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associatate"),
                &["associate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegoptionsotiation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("yotube"), &["youtube"]),
            (
                dictgen::InsensitiveStr::Ascii("deliberante"),
                &["deliberate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parargaphs"),
                &["paragraphs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discriminatorie"),
                &["discriminate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inicializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countrywise"),
                &["countryside"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coalesence"),
                &["coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descussion"),
                &["discussion"],
            ),
            (dictgen::InsensitiveStr::Ascii("adventues"), &["adventures"]),
            (
                dictgen::InsensitiveStr::Ascii("complitations"),
                &["compilations", "complications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docuentations"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("aligment"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("wighing"), &["weighing"]),
            (
                dictgen::InsensitiveStr::Ascii("inherating"),
                &["inheriting"],
            ),
            (dictgen::InsensitiveStr::Ascii("volatilty"), &["volatility"]),
            (dictgen::InsensitiveStr::Ascii("varaints"), &["variants"]),
            (
                dictgen::InsensitiveStr::Ascii("traiing"),
                &["trailing", "training"],
            ),
            (dictgen::InsensitiveStr::Ascii("wolrd"), &["world"]),
            (dictgen::InsensitiveStr::Ascii("neighbous"), &["neighbours"]),
            (dictgen::InsensitiveStr::Ascii("carring"), &["carrying"]),
            (dictgen::InsensitiveStr::Ascii("internus"), &["interns"]),
            (
                dictgen::InsensitiveStr::Ascii("coursin"),
                &["coursing", "cousin"],
            ),
            (dictgen::InsensitiveStr::Ascii("prototye"), &["prototype"]),
            (
                dictgen::InsensitiveStr::Ascii("perforance"),
                &["performance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("miscellanous"),
                &["miscellaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destionations"),
                &["destinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assocation"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("orer"), &["order"]),
            (
                dictgen::InsensitiveStr::Ascii("participare"),
                &["participate"],
            ),
            (dictgen::InsensitiveStr::Ascii("emplyee"), &["employee"]),
            (dictgen::InsensitiveStr::Ascii("interste"), &["interstate"]),
            (
                dictgen::InsensitiveStr::Ascii("colouising"),
                &["colourising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workes"),
                &["works", "workers"],
            ),
            (dictgen::InsensitiveStr::Ascii("responive"), &["responsive"]),
            (dictgen::InsensitiveStr::Ascii("remvoving"), &["removing"]),
            (dictgen::InsensitiveStr::Ascii("maintans"), &["maintains"]),
            (dictgen::InsensitiveStr::Ascii("supportt"), &["support"]),
            (
                dictgen::InsensitiveStr::Ascii("cannonicalization"),
                &["canonicalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("reperesent"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("trancode"), &["transcode"]),
            (
                dictgen::InsensitiveStr::Ascii("corrensponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expropiates"),
                &["expropriates"],
            ),
            (dictgen::InsensitiveStr::Ascii("selfiers"), &["selfies"]),
            (dictgen::InsensitiveStr::Ascii("breakpint"), &["breakpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("indirektly"),
                &["indirectly"],
            ),
            (dictgen::InsensitiveStr::Ascii("authroise"), &["authorise"]),
            (
                dictgen::InsensitiveStr::Ascii("correlelates"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("negosiate"), &["negotiate"]),
            (
                dictgen::InsensitiveStr::Ascii("repsectful"),
                &["respectful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnformations"),
                &["transformations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumeraters"),
                &["enumerators", "enumerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("moelcules"), &["molecules"]),
            (dictgen::InsensitiveStr::Ascii("serialzes"), &["serializes"]),
            (dictgen::InsensitiveStr::Ascii("toggeled"), &["toggled"]),
            (dictgen::InsensitiveStr::Ascii("celcius"), &["celsius"]),
            (dictgen::InsensitiveStr::Ascii("exponet"), &["exponent"]),
            (dictgen::InsensitiveStr::Ascii("regeressed"), &["regressed"]),
            (dictgen::InsensitiveStr::Ascii("confrence"), &["conference"]),
            (
                dictgen::InsensitiveStr::Ascii("transalted"),
                &["translated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("informacional"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeatatly"),
                &["repeatedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("conatain"), &["contain"]),
            (
                dictgen::InsensitiveStr::Ascii("overcompansating"),
                &["overcompensating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("photoshope"),
                &["photoshopped"],
            ),
            (dictgen::InsensitiveStr::Ascii("candiate"), &["candidate"]),
            (
                dictgen::InsensitiveStr::Ascii("outperfoms"),
                &["outperform"],
            ),
            (dictgen::InsensitiveStr::Ascii("bollean"), &["boolean"]),
            (
                dictgen::InsensitiveStr::Ascii("expectatoin"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mormalised"),
                &["normalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("staion"), &["station"]),
            (dictgen::InsensitiveStr::Ascii("exculsive"), &["exclusive"]),
            (
                dictgen::InsensitiveStr::Ascii("abbrviating"),
                &["abbreviating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparment"),
                &["compartment"],
            ),
            (dictgen::InsensitiveStr::Ascii("bracnh"), &["branch"]),
            (dictgen::InsensitiveStr::Ascii("thsoe"), &["those"]),
            (dictgen::InsensitiveStr::Ascii("isues"), &["issues"]),
            (
                dictgen::InsensitiveStr::Unicode("сontained"),
                &["contained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commandered"),
                &["commanded"],
            ),
            (dictgen::InsensitiveStr::Ascii("debugee"), &["debuggee"]),
            (dictgen::InsensitiveStr::Ascii("interupts"), &["interrupts"]),
            (
                dictgen::InsensitiveStr::Ascii("tranformable"),
                &["transformable"],
            ),
            (dictgen::InsensitiveStr::Ascii("popuplated"), &["populated"]),
            (dictgen::InsensitiveStr::Ascii("unocde"), &["unicode"]),
            (dictgen::InsensitiveStr::Ascii("entitys"), &["entities"]),
            (dictgen::InsensitiveStr::Ascii("uprgading"), &["upgrading"]),
            (dictgen::InsensitiveStr::Ascii("accessess"), &["accesses"]),
            (dictgen::InsensitiveStr::Ascii("provived"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("loacalizing"),
                &["localizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("fullfiled"), &["fulfilled"]),
            (dictgen::InsensitiveStr::Ascii("thhanks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("sofmore"), &["sophomore"]),
            (
                dictgen::InsensitiveStr::Ascii("aprroximate"),
                &["approximate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devastaded"),
                &["devastated"],
            ),
            (dictgen::InsensitiveStr::Ascii("stitchs"), &["stitches"]),
            (
                dictgen::InsensitiveStr::Ascii("dissassembled"),
                &["disassembled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedect"),
                &["deduct", "detect"],
            ),
            (dictgen::InsensitiveStr::Ascii("snalshot"), &["snapshot"]),
            (
                dictgen::InsensitiveStr::Ascii("mobules"),
                &["modules", "mobiles"],
            ),
            (dictgen::InsensitiveStr::Ascii("repalace"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("countinuous"),
                &["continuous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imrpovised"),
                &["improvised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introduing"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dedected"),
                &["detected", "deducted"],
            ),
            (dictgen::InsensitiveStr::Ascii("immobilier"), &["immobile"]),
            (
                dictgen::InsensitiveStr::Ascii("establsihed"),
                &["established"],
            ),
            (dictgen::InsensitiveStr::Ascii("wtiness"), &["witness"]),
            (dictgen::InsensitiveStr::Ascii("datsets"), &["datasets"]),
            (dictgen::InsensitiveStr::Ascii("persisited"), &["persisted"]),
            (dictgen::InsensitiveStr::Ascii("exilerate"), &["exhilarate"]),
            (dictgen::InsensitiveStr::Ascii("broswe"), &["browse"]),
            (
                dictgen::InsensitiveStr::Ascii("corosponds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represeted"),
                &["represented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrepoducible"),
                &["unreproducible"],
            ),
            (dictgen::InsensitiveStr::Ascii("overun"), &["overrun"]),
            (dictgen::InsensitiveStr::Ascii("kidknapper"), &["kidnapper"]),
            (dictgen::InsensitiveStr::Ascii("subtrate"), &["substrate"]),
            (
                dictgen::InsensitiveStr::Ascii("exclamantion"),
                &["exclamation"],
            ),
            (dictgen::InsensitiveStr::Ascii("previwes"), &["previews"]),
            (dictgen::InsensitiveStr::Ascii("intentsity"), &["intensity"]),
            (
                dictgen::InsensitiveStr::Ascii("ilustration"),
                &["illustration"],
            ),
            (dictgen::InsensitiveStr::Ascii("youre"), &["your"]),
            (
                dictgen::InsensitiveStr::Ascii("arrengement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("pyramis"), &["pyramids"]),
            (
                dictgen::InsensitiveStr::Ascii("cannabilyse"),
                &["cannibalise"],
            ),
            (dictgen::InsensitiveStr::Ascii("escuse"), &["excuse"]),
            (
                dictgen::InsensitiveStr::Ascii("combininations"),
                &["combinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("purtain"), &["pertain"]),
            (
                dictgen::InsensitiveStr::Ascii("restransmitting"),
                &["retransmitting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nanosceconds"),
                &["nanoseconds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialsation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("subpsace"), &["subspace"]),
            (
                dictgen::InsensitiveStr::Ascii("unistalled"),
                &["uninstalled"],
            ),
            (dictgen::InsensitiveStr::Ascii("inisides"), &["insides"]),
            (dictgen::InsensitiveStr::Ascii("hedeghog"), &["hedgehog"]),
            (dictgen::InsensitiveStr::Ascii("serioulsy"), &["seriously"]),
            (dictgen::InsensitiveStr::Ascii("ingoring"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("criticarlo"), &["critical"]),
            (dictgen::InsensitiveStr::Ascii("shoudld"), &["should"]),
            (
                dictgen::InsensitiveStr::Ascii("electricain"),
                &["electrician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conergence"),
                &["convergence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnciate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accounding"),
                &["according", "accounting"],
            ),
            (dictgen::InsensitiveStr::Ascii("forceing"), &["forcing"]),
            (
                dictgen::InsensitiveStr::Ascii("resovlable"),
                &["resolvable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electromangetic"),
                &["electromagnetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("failicy"), &["facility"]),
            (
                dictgen::InsensitiveStr::Ascii("heartbraker"),
                &["heartbreak"],
            ),
            (dictgen::InsensitiveStr::Ascii("spporting"), &["supporting"]),
            (dictgen::InsensitiveStr::Ascii("secrity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("sequetial"), &["sequential"]),
            (dictgen::InsensitiveStr::Ascii("namaes"), &["names"]),
            (dictgen::InsensitiveStr::Ascii("encocding"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("dissctors"), &["dissectors"]),
            (dictgen::InsensitiveStr::Ascii("divicing"), &["dividing"]),
            (
                dictgen::InsensitiveStr::Ascii("represetnation"),
                &["representations"],
            ),
            (dictgen::InsensitiveStr::Ascii("advesary"), &["adversary"]),
            (
                dictgen::InsensitiveStr::Ascii("disconncet"),
                &["disconnects"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calculattion"),
                &["calculation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incorportates"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("incliudes"), &["includes"]),
            (dictgen::InsensitiveStr::Ascii("guatamala"), &["guatemala"]),
            (dictgen::InsensitiveStr::Ascii("fourh"), &["fourth"]),
            (dictgen::InsensitiveStr::Ascii("dynamicdns"), &["dynamics"]),
            (
                dictgen::InsensitiveStr::Ascii("weeved"),
                &["weaved", "wove"],
            ),
            (dictgen::InsensitiveStr::Ascii("callabacks"), &["callbacks"]),
            (
                dictgen::InsensitiveStr::Ascii("interpretier"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("statisfieds"),
                &["satisfies"],
            ),
            (dictgen::InsensitiveStr::Ascii("spawining"), &["spawning"]),
            (
                dictgen::InsensitiveStr::Ascii("confurming"),
                &["confirming", "conforming"],
            ),
            (dictgen::InsensitiveStr::Ascii("obiously"), &["obviously"]),
            (dictgen::InsensitiveStr::Ascii("degerate"), &["degenerate"]),
            (dictgen::InsensitiveStr::Ascii("isloated"), &["isolated"]),
            (dictgen::InsensitiveStr::Ascii("earleir"), &["earlier"]),
            (
                dictgen::InsensitiveStr::Ascii("responsble"),
                &["responsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catapillers"),
                &["caterpillars"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contaminents"),
                &["contaminants"],
            ),
            (dictgen::InsensitiveStr::Ascii("ealiest"), &["earliest"]),
            (
                dictgen::InsensitiveStr::Ascii("startparanthesis"),
                &["startparentheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prefection"),
                &["perfection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asympyotically"),
                &["asymptotically"],
            ),
            (dictgen::InsensitiveStr::Ascii("relativiy"), &["relativity"]),
            (
                dictgen::InsensitiveStr::Ascii("reintantiate"),
                &["reinstantiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("ststr"), &["strstr"]),
            (
                dictgen::InsensitiveStr::Ascii("consctructions"),
                &["constructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("virbator"), &["vibrator"]),
            (dictgen::InsensitiveStr::Ascii("iteratons"), &["iterations"]),
            (
                dictgen::InsensitiveStr::Ascii("chizzelling"),
                &["chiselling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wrkbenches"),
                &["workbenches"],
            ),
            (dictgen::InsensitiveStr::Ascii("reuqests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("reserches"), &["researches"]),
            (dictgen::InsensitiveStr::Ascii("politley"), &["politely"]),
            (dictgen::InsensitiveStr::Ascii("cardborad"), &["cardboard"]),
            (dictgen::InsensitiveStr::Ascii("taxanomy"), &["taxonomy"]),
            (dictgen::InsensitiveStr::Ascii("architct"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("variatinos"),
                &["variations"],
            ),
            (dictgen::InsensitiveStr::Ascii("anotehr"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("initilializes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomrative"),
                &["performative"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggrgator"), &["aggregator"]),
            (
                dictgen::InsensitiveStr::Ascii("downtraded"),
                &["downgraded"],
            ),
            (dictgen::InsensitiveStr::Ascii("colud"), &["cloud", "could"]),
            (dictgen::InsensitiveStr::Ascii("yealry"), &["yearly"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptons"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("countere"), &["counter"]),
            (
                dictgen::InsensitiveStr::Ascii("prealocating"),
                &["preallocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("peerowets"), &["pirouettes"]),
            (dictgen::InsensitiveStr::Ascii("spammend"), &["spammed"]),
            (dictgen::InsensitiveStr::Ascii("portary"), &["portray"]),
            (dictgen::InsensitiveStr::Ascii("hting"), &["thing"]),
            (
                dictgen::InsensitiveStr::Ascii("quitely"),
                &["quietly", "quite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("midfileder"),
                &["midfielder"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prerequisets"),
                &["prerequisite"],
            ),
            (dictgen::InsensitiveStr::Ascii("pliars"), &["pliers"]),
            (
                dictgen::InsensitiveStr::Ascii("geenration"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("standadise"),
                &["standardise"],
            ),
            (dictgen::InsensitiveStr::Ascii("carolan"), &["carolina"]),
            (
                dictgen::InsensitiveStr::Ascii("itegrating"),
                &["integrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("meens"), &["means"]),
            (
                dictgen::InsensitiveStr::Ascii("comlicated"),
                &["complicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("contans"), &["contains"]),
            (
                dictgen::InsensitiveStr::Ascii("suposeable"),
                &["supposable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superviso"),
                &["supervision"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exmplifies"),
                &["exemplifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dcompressed"),
                &["decompressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefication"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simulatanous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginisation"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("lighbar"), &["lightbar"]),
            (
                dictgen::InsensitiveStr::Ascii("expressiosn"),
                &["expressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converison"),
                &["conversions", "conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("velcome"), &["welcome"]),
            (
                dictgen::InsensitiveStr::Ascii("stimualtion"),
                &["stimulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("dimesion"), &["dimension"]),
            (
                dictgen::InsensitiveStr::Ascii("accidentlly"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("generlas"), &["generals"]),
            (dictgen::InsensitiveStr::Ascii("persuated"), &["persuade"]),
            (
                dictgen::InsensitiveStr::Ascii("unsuccesfully"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separato"),
                &["separation", "separator"],
            ),
            (dictgen::InsensitiveStr::Ascii("asssassans"), &["assassins"]),
            (
                dictgen::InsensitiveStr::Ascii("insiginficant"),
                &["insignificant"],
            ),
            (dictgen::InsensitiveStr::Ascii("yoman"), &["yeoman"]),
            (dictgen::InsensitiveStr::Ascii("minimse"), &["minimise"]),
            (
                dictgen::InsensitiveStr::Ascii("workaroung"),
                &["workaround"],
            ),
            (dictgen::InsensitiveStr::Ascii("uscaled"), &["unscaled"]),
            (dictgen::InsensitiveStr::Ascii("lesstiff"), &["lesstif"]),
            (dictgen::InsensitiveStr::Ascii("failer"), &["failure"]),
            (dictgen::InsensitiveStr::Ascii("communits"), &["communist"]),
            (dictgen::InsensitiveStr::Ascii("directoy"), &["directory"]),
            (dictgen::InsensitiveStr::Ascii("ambiguitiy"), &["ambiguity"]),
            (
                dictgen::InsensitiveStr::Ascii("transparancies"),
                &["transparencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("courrier"), &["courier"]),
            (dictgen::InsensitiveStr::Ascii("cirucits"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("acctivating"),
                &["activating"],
            ),
            (dictgen::InsensitiveStr::Ascii("bronocs"), &["broncos"]),
            (dictgen::InsensitiveStr::Ascii("torpedos"), &["torpedoes"]),
            (
                dictgen::InsensitiveStr::Ascii("confugired"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recompuutes"),
                &["recomputes"],
            ),
            (dictgen::InsensitiveStr::Ascii("sartup"), &["startup"]),
            (dictgen::InsensitiveStr::Ascii("limtied"), &["limited"]),
            (dictgen::InsensitiveStr::Ascii("neumonics"), &["mnemonics"]),
            (dictgen::InsensitiveStr::Ascii("sommeone"), &["someone"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstratable"),
                &["demonstrable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manouverable"),
                &["maneuverable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sliblings"), &["siblings"]),
            (
                dictgen::InsensitiveStr::Ascii("customicable"),
                &["customisable", "customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complelely"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhcuasted"), &["exhausted"]),
            (
                dictgen::InsensitiveStr::Ascii("accorddions"),
                &["accordions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continaing"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("cirularly"), &["circularly"]),
            (
                dictgen::InsensitiveStr::Ascii("caclucation"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("obgects"), &["objects"]),
            (
                dictgen::InsensitiveStr::Ascii("apdaters"),
                &["updaters", "adapters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribucion"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distrabution"),
                &["distribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforcin"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("atatchments"),
                &["attachments"],
            ),
            (dictgen::InsensitiveStr::Ascii("promtps"), &["prompts"]),
            (dictgen::InsensitiveStr::Ascii("alterior"), &["ulterior"]),
            (
                dictgen::InsensitiveStr::Ascii("criticisize"),
                &["criticise"],
            ),
            (dictgen::InsensitiveStr::Ascii("thnigs"), &["things"]),
            (
                dictgen::InsensitiveStr::Ascii("configutrations"),
                &["configurations"],
            ),
            (dictgen::InsensitiveStr::Ascii("anlge"), &["angle"]),
            (
                dictgen::InsensitiveStr::Ascii("trinagulation"),
                &["triangulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialsing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastractures"),
                &["infrastructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("incriease"), &["increase"]),
            (dictgen::InsensitiveStr::Ascii("repoduced"), &["reproduced"]),
            (
                dictgen::InsensitiveStr::Ascii("juristictions"),
                &["jurisdictions"],
            ),
            (dictgen::InsensitiveStr::Ascii("rerturns"), &["returns"]),
            (
                dictgen::InsensitiveStr::Ascii("percantages"),
                &["percentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("gayshas"), &["geishas"]),
            (
                dictgen::InsensitiveStr::Ascii("dicriminators"),
                &["discriminators"],
            ),
            (dictgen::InsensitiveStr::Ascii("cubcile"), &["cubicle"]),
            (dictgen::InsensitiveStr::Ascii("waktins"), &["watkins"]),
            (
                dictgen::InsensitiveStr::Ascii("oppresssion"),
                &["oppression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intrfacing"),
                &["interfacing"],
            ),
            (dictgen::InsensitiveStr::Ascii("deffined"), &["defined"]),
            (
                dictgen::InsensitiveStr::Ascii("convicing"),
                &["convincing", "convicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("announciators"),
                &["annunciators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatanete"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("resoiurces"), &["resources"]),
            (dictgen::InsensitiveStr::Ascii("upagraded"), &["upgraded"]),
            (dictgen::InsensitiveStr::Ascii("subtitute"), &["substitute"]),
            (dictgen::InsensitiveStr::Ascii("whlie"), &["while"]),
            (dictgen::InsensitiveStr::Ascii("partcles"), &["particles"]),
            (dictgen::InsensitiveStr::Ascii("gagnsters"), &["gangsters"]),
            (dictgen::InsensitiveStr::Ascii("weridest"), &["weirdest"]),
            (dictgen::InsensitiveStr::Ascii("pyrmaids"), &["pyramids"]),
            (
                dictgen::InsensitiveStr::Ascii("instrumeneted"),
                &["instrumented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satsifaction"),
                &["satisfaction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronize"),
                &["synchronize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reletively"),
                &["relatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("kindom"), &["kingdom"]),
            (
                dictgen::InsensitiveStr::Ascii("overriabled"),
                &["overridable"],
            ),
            (dictgen::InsensitiveStr::Ascii("amouts"), &["amounts"]),
            (
                dictgen::InsensitiveStr::Ascii("requencies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("thhese"), &["these"]),
            (
                dictgen::InsensitiveStr::Ascii("videojames"),
                &["videogames"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunete"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("relase"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("silhouwetist"),
                &["silhouettist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thoroughty"),
                &["thoroughly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trnasformation"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referemces"),
                &["references"],
            ),
            (dictgen::InsensitiveStr::Ascii("oulines"), &["outlines"]),
            (
                dictgen::InsensitiveStr::Ascii("ommission"),
                &["omission", "immission", "commission"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authentcates"),
                &["authenticates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sertificates"),
                &["certificates"],
            ),
            (dictgen::InsensitiveStr::Ascii("maraudeurs"), &["marauder"]),
            (
                dictgen::InsensitiveStr::Ascii("abbraviated"),
                &["abbreviated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unconditonally"),
                &["unconditionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoucing"), &["announcing"]),
            (dictgen::InsensitiveStr::Ascii("oreintal"), &["oriental"]),
            (dictgen::InsensitiveStr::Ascii("casheir"), &["cashier"]),
            (dictgen::InsensitiveStr::Ascii("contiune"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("reposiries"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("dioreha"), &["diarrhea"]),
            (
                dictgen::InsensitiveStr::Ascii("ofisionado"),
                &["aficionado"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continguous"),
                &["contiguous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imputted"),
                &["inputted", "imputed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("managament"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("crirical"), &["critical"]),
            (
                dictgen::InsensitiveStr::Ascii("accumelating"),
                &["accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissapear"), &["disappear"]),
            (
                dictgen::InsensitiveStr::Ascii("accumuates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranparent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("woild"),
                &["would", "world", "wild"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalite"), &["evaluate"]),
            (dictgen::InsensitiveStr::Ascii("enemey"), &["enemy"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiastisch"),
                &["enthusiastic"],
            ),
            (dictgen::InsensitiveStr::Ascii("missorui"), &["missouri"]),
            (
                dictgen::InsensitiveStr::Ascii("impressario"),
                &["impresario"],
            ),
            (dictgen::InsensitiveStr::Ascii("ennpoint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("compsers"), &["compress"]),
            (
                dictgen::InsensitiveStr::Ascii("predicitve"),
                &["predictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conjuntion"),
                &["conjunction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowleges"),
                &["acknowledges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protectores"),
                &["protectors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transcluent"),
                &["translucent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacments"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompataibility"),
                &["incompatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("omniverously"),
                &["omnivorously"],
            ),
            (dictgen::InsensitiveStr::Ascii("hyprocisy"), &["hypocrisy"]),
            (dictgen::InsensitiveStr::Ascii("modifdied"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("lauguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("briused"), &["bruised"]),
            (dictgen::InsensitiveStr::Ascii("decodin"), &["decoding"]),
            (dictgen::InsensitiveStr::Ascii("iniital"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("retcievers"),
                &["receivers", "retrievers"],
            ),
            (dictgen::InsensitiveStr::Ascii("picnicks"), &["picnics"]),
            (dictgen::InsensitiveStr::Ascii("accoutn"), &["account"]),
            (
                dictgen::InsensitiveStr::Ascii("transforners"),
                &["transformers"],
            ),
            (dictgen::InsensitiveStr::Ascii("alphbets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("contimplate"),
                &["contemplate"],
            ),
            (dictgen::InsensitiveStr::Ascii("conerted"), &["converted"]),
            (
                dictgen::InsensitiveStr::Ascii("prepetrators"),
                &["perpetrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotiaion"),
                &["negotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experinementel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("halucinate"),
                &["hallucinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("varity"), &["variety"]),
            (
                dictgen::InsensitiveStr::Ascii("pscyhopathic"),
                &["psychopathic"],
            ),
            (dictgen::InsensitiveStr::Ascii("expence"), &["expense"]),
            (
                dictgen::InsensitiveStr::Ascii("confidantals"),
                &["confidentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("presude"), &["presumed"]),
            (
                dictgen::InsensitiveStr::Ascii("supervisior"),
                &["supervisor"],
            ),
            (dictgen::InsensitiveStr::Ascii("assighed"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("indivisually"),
                &["individually"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apaptation"),
                &["adaptation"],
            ),
            (dictgen::InsensitiveStr::Ascii("apparreled"), &["appareled"]),
            (dictgen::InsensitiveStr::Ascii("uknown"), &["unknown"]),
            (
                dictgen::InsensitiveStr::Ascii("calculataed"),
                &["calculated"],
            ),
            (dictgen::InsensitiveStr::Ascii("scholalry"), &["scholarly"]),
            (
                dictgen::InsensitiveStr::Ascii("popose"),
                &["propose", "oppose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intricaces"),
                &["intricacies"],
            ),
            (dictgen::InsensitiveStr::Ascii("tesited"), &["tested"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucessufll"),
                &["unsuccessful"],
            ),
            (dictgen::InsensitiveStr::Ascii("tablulator"), &["tabulator"]),
            (dictgen::InsensitiveStr::Ascii("clipoard"), &["clipboard"]),
            (dictgen::InsensitiveStr::Ascii("bounidng"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("expressens"), &["expresses"]),
            (dictgen::InsensitiveStr::Ascii("rumuors"), &["rumors"]),
            (dictgen::InsensitiveStr::Ascii("reular"), &["regular"]),
            (
                dictgen::InsensitiveStr::Ascii("responcibilities"),
                &["responsibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("spawnign"), &["spawning"]),
            (
                dictgen::InsensitiveStr::Ascii("identificativo"),
                &["identification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redmeption"),
                &["redemption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configutable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prostitutiei"),
                &["prostitute"],
            ),
            (dictgen::InsensitiveStr::Ascii("reapeater"), &["repeater"]),
            (dictgen::InsensitiveStr::Ascii("areospace"), &["aerospace"]),
            (dictgen::InsensitiveStr::Ascii("macther"), &["matcher"]),
            (
                dictgen::InsensitiveStr::Ascii("responiveness"),
                &["responsiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("determaine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("unsucesssfully"),
                &["unsuccessfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backwad"),
                &["backward", "backwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charaterizations"),
                &["characterizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shecks"),
                &["checks", "shucks"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuill"), &["still"]),
            (dictgen::InsensitiveStr::Ascii("twiligt"), &["twilight"]),
            (dictgen::InsensitiveStr::Ascii("addert"), &["assert"]),
            (
                dictgen::InsensitiveStr::Ascii("retrictions"),
                &["restrictions", "retractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predessecor"),
                &["predecessor"],
            ),
            (dictgen::InsensitiveStr::Ascii("tre"), &["tree", "the"]),
            (
                dictgen::InsensitiveStr::Ascii("commtiters"),
                &["committers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("syncronising"),
                &["synchronising"],
            ),
            (dictgen::InsensitiveStr::Ascii("shcedules"), &["schedules"]),
            (dictgen::InsensitiveStr::Ascii("cirquit"), &["circuit"]),
            (dictgen::InsensitiveStr::Ascii("appature"), &["aperture"]),
            (
                dictgen::InsensitiveStr::Ascii("akumulation"),
                &["accumulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("geerates"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("laucher"), &["launcher"]),
            (
                dictgen::InsensitiveStr::Ascii("prodecural"),
                &["procedural"],
            ),
            (dictgen::InsensitiveStr::Ascii("dissimlar"), &["dissimilar"]),
            (dictgen::InsensitiveStr::Ascii("argements"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("distastful"),
                &["distasteful"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expermeneter"),
                &["experimenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("architech"), &["architect"]),
            (
                dictgen::InsensitiveStr::Ascii("donwloaded"),
                &["downloaded"],
            ),
            (dictgen::InsensitiveStr::Ascii("cteations"), &["creations"]),
            (dictgen::InsensitiveStr::Ascii("separtor"), &["separator"]),
            (
                dictgen::InsensitiveStr::Ascii("propertly"),
                &["property", "properly"],
            ),
            (dictgen::InsensitiveStr::Ascii("diptheria"), &["diphtheria"]),
            (
                dictgen::InsensitiveStr::Ascii("universiality"),
                &["universality"],
            ),
            (dictgen::InsensitiveStr::Ascii("imedialy"), &["immediately"]),
            (dictgen::InsensitiveStr::Ascii("cartirdge"), &["cartridge"]),
            (
                dictgen::InsensitiveStr::Ascii("infallibile"),
                &["infallible"],
            ),
            (dictgen::InsensitiveStr::Ascii("langage"), &["language"]),
            (
                dictgen::InsensitiveStr::Ascii("distrubited"),
                &["distributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("madturbation"),
                &["masturbation"],
            ),
            (dictgen::InsensitiveStr::Ascii("jorurnal"), &["journal"]),
            (dictgen::InsensitiveStr::Ascii("straigh"), &["straight"]),
            (dictgen::InsensitiveStr::Ascii("changeble"), &["changeable"]),
            (dictgen::InsensitiveStr::Ascii("seciion"), &["section"]),
            (dictgen::InsensitiveStr::Ascii("suspectes"), &["suspects"]),
            (
                dictgen::InsensitiveStr::Ascii("tuesdsy"),
                &["tuesdays", "tuesday"],
            ),
            (dictgen::InsensitiveStr::Ascii("phoen"), &["phone"]),
            (
                dictgen::InsensitiveStr::Ascii("apostropohe"),
                &["apostrophe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wokrspaces"),
                &["workspaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahving"), &["having"]),
            (
                dictgen::InsensitiveStr::Ascii("verificacion"),
                &["verification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evolutionnary"),
                &["evolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("fascistes"), &["fascists"]),
            (dictgen::InsensitiveStr::Ascii("commited"), &["committed"]),
            (
                dictgen::InsensitiveStr::Ascii("accrdingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("mysitcal"), &["mystical"]),
            (dictgen::InsensitiveStr::Ascii("autoattk"), &["autoattack"]),
            (
                dictgen::InsensitiveStr::Ascii("techtician"),
                &["technician"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cessationalist"),
                &["sensationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manifestato"),
                &["manifesto"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("microcentre"),
                &["microcenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lerned"),
                &["learned", "leaned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mysogynist"),
                &["misogynist"],
            ),
            (dictgen::InsensitiveStr::Ascii("debain"), &["debian"]),
            (
                dictgen::InsensitiveStr::Ascii("provacotive"),
                &["provocative"],
            ),
            (dictgen::InsensitiveStr::Ascii("alingment"), &["alignment"]),
            (
                dictgen::InsensitiveStr::Ascii("publicani"),
                &["publication"],
            ),
            (dictgen::InsensitiveStr::Ascii("bangquit"), &["banquet"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchroous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciminator"),
                &["discriminator"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbitualy"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("porotocol"), &["protocol"]),
            (dictgen::InsensitiveStr::Ascii("publisherr"), &["publisher"]),
            (dictgen::InsensitiveStr::Ascii("kenyesian"), &["keynesian"]),
            (
                dictgen::InsensitiveStr::Ascii("merchantibility"),
                &["merchantability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprieters"),
                &["proprietors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specifiecally"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("definned"), &["defined"]),
            (
                dictgen::InsensitiveStr::Ascii("inprovements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restirction"),
                &["restriction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encounteerd"),
                &["encountered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convencient"),
                &["convenient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availe"),
                &["avail", "available"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metropolitaine"),
                &["metropolitan"],
            ),
            (dictgen::InsensitiveStr::Ascii("intravert"), &["introvert"]),
            (
                dictgen::InsensitiveStr::Ascii("aluminim"),
                &["aluminium", "aluminum"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalited"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("calculationgs"),
                &["calculations"],
            ),
            (dictgen::InsensitiveStr::Ascii("abstrated"), &["abstracted"]),
            (
                dictgen::InsensitiveStr::Ascii("dependendencies"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("projecte"), &["projectile"]),
            (
                dictgen::InsensitiveStr::Ascii("collisons"),
                &["collisions", "collusion", "collusions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compileable"),
                &["compilable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assotiated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("raotate"), &["rotate"]),
            (
                dictgen::InsensitiveStr::Ascii("incompetente"),
                &["incompetence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ancapsulate"),
                &["encapsulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconvinence"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajacency"), &["adjacency"]),
            (dictgen::InsensitiveStr::Ascii("prining"), &["printing"]),
            (
                dictgen::InsensitiveStr::Ascii("manifistation"),
                &["manifestation"],
            ),
            (dictgen::InsensitiveStr::Ascii("whipsered"), &["whispered"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("renderingg"), &["rendering"]),
            (
                dictgen::InsensitiveStr::Ascii("convienently"),
                &["conveniently"],
            ),
            (dictgen::InsensitiveStr::Ascii("eplore"), &["explore"]),
            (
                dictgen::InsensitiveStr::Ascii("mellinnium"),
                &["millennium"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoult"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("thiknesses"),
                &["thicknesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("contaner"), &["container"]),
            (dictgen::InsensitiveStr::Ascii("exectly"), &["exactly"]),
            (
                dictgen::InsensitiveStr::Ascii("trogliditism"),
                &["troglodytism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spearations"),
                &["separations"],
            ),
            (dictgen::InsensitiveStr::Ascii("undefind"), &["undefined"]),
            (
                dictgen::InsensitiveStr::Ascii("preferabely"),
                &["preferable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deduplacates"),
                &["deduplicates"],
            ),
            (dictgen::InsensitiveStr::Ascii("mocrosoft"), &["microsoft"]),
            (dictgen::InsensitiveStr::Ascii("treasurery"), &["treasury"]),
            (
                dictgen::InsensitiveStr::Ascii("constructes"),
                &["constructs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialialisation"),
                &["initialisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("challenes"), &["challenges"]),
            (
                dictgen::InsensitiveStr::Ascii("hierachries"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpabetically"),
                &["alphabetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("depositers"), &["deposits"]),
            (
                dictgen::InsensitiveStr::Ascii("contribted"),
                &["contributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("deviiated"), &["deviated"]),
            (
                dictgen::InsensitiveStr::Ascii("unimporant"),
                &["unimportant"],
            ),
            (dictgen::InsensitiveStr::Ascii("adquires"), &["acquires"]),
            (
                dictgen::InsensitiveStr::Ascii("implementes"),
                &["implements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operatators"),
                &["operators"],
            ),
            (dictgen::InsensitiveStr::Ascii("fecthing"), &["fetching"]),
            (
                dictgen::InsensitiveStr::Ascii("calcurators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("langerray"), &["lingerie"]),
            (dictgen::InsensitiveStr::Ascii("validtion"), &["validation"]),
            (
                dictgen::InsensitiveStr::Ascii("backpropogation"),
                &["backpropagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avalaibility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("selektions"),
                &["selections"],
            ),
            (dictgen::InsensitiveStr::Ascii("fecthed"), &["fetched"]),
            (dictgen::InsensitiveStr::Ascii("summurise"), &["summarise"]),
            (
                dictgen::InsensitiveStr::Ascii("agressions"),
                &["aggressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoropates"),
                &["incorporates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infilitrate"),
                &["infiltrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("materializatons"),
                &["materializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dstinations"),
                &["destinations"],
            ),
            (dictgen::InsensitiveStr::Ascii("minizing"), &["minimizing"]),
            (
                dictgen::InsensitiveStr::Ascii("borderlans"),
                &["borderlands"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrepresentation"),
                &["misrepresentation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoitators"),
                &["negotiators"],
            ),
            (dictgen::InsensitiveStr::Ascii("prolouge"), &["prologue"]),
            (dictgen::InsensitiveStr::Ascii("minsiter"), &["minister"]),
            (
                dictgen::InsensitiveStr::Ascii("obseravtion"),
                &["observation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subtargers"),
                &["subtargets"],
            ),
            (dictgen::InsensitiveStr::Ascii("infor"), &["info", "inform"]),
            (dictgen::InsensitiveStr::Ascii("arquiving"), &["archiving"]),
            (dictgen::InsensitiveStr::Ascii("servie"), &["service"]),
            (dictgen::InsensitiveStr::Ascii("feautre"), &["feature"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscirbing"),
                &["unsubscribing"],
            ),
            (dictgen::InsensitiveStr::Ascii("jugdement"), &["judgement"]),
            (dictgen::InsensitiveStr::Ascii("syarcuse"), &["syracuse"]),
            (dictgen::InsensitiveStr::Ascii("aqquiring"), &["acquiring"]),
            (dictgen::InsensitiveStr::Ascii("abigious"), &["ambiguous"]),
            (
                dictgen::InsensitiveStr::Ascii("charatable"),
                &["charitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignoreable"), &["ignorable"]),
            (dictgen::InsensitiveStr::Ascii("eazier"), &["easier"]),
            (
                dictgen::InsensitiveStr::Ascii("successully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("dialouge"), &["dialogue"]),
            (
                dictgen::InsensitiveStr::Ascii("comuter"),
                &["computer", "commuter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcpondences"),
                &["correspondences"],
            ),
            (dictgen::InsensitiveStr::Ascii("useernames"), &["usernames"]),
            (
                dictgen::InsensitiveStr::Ascii("differenciations"),
                &["differentiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("eamples"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("insttruction"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("releaed"), &["released"]),
            (
                dictgen::InsensitiveStr::Ascii("exceoption"),
                &["exemption", "exception"],
            ),
            (dictgen::InsensitiveStr::Ascii("uwinding"), &["unwinding"]),
            (dictgen::InsensitiveStr::Ascii("labaels"), &["labels"]),
            (dictgen::InsensitiveStr::Ascii("indecies"), &["indices"]),
            (dictgen::InsensitiveStr::Ascii("democract"), &["democrat"]),
            (
                dictgen::InsensitiveStr::Ascii("organsiation"),
                &["organisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appropriatee"),
                &["appropriate"],
            ),
            (dictgen::InsensitiveStr::Ascii("seldomly"), &["seldom"]),
            (
                dictgen::InsensitiveStr::Ascii("falmethrower"),
                &["flamethrower"],
            ),
            (dictgen::InsensitiveStr::Ascii("upddaters"), &["updaters"]),
            (dictgen::InsensitiveStr::Ascii("classis"), &["classics"]),
            (
                dictgen::InsensitiveStr::Ascii("denomintor"),
                &["denominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("endangerd"), &["endangered"]),
            (
                dictgen::InsensitiveStr::Ascii("speficications"),
                &["specifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("taligate"), &["tailgate"]),
            (dictgen::InsensitiveStr::Ascii("luckilly"), &["luckily"]),
            (dictgen::InsensitiveStr::Ascii("esier"), &["easier"]),
            (dictgen::InsensitiveStr::Ascii("propectus"), &["prospectus"]),
            (
                dictgen::InsensitiveStr::Ascii("availibity"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("absoulete"),
                &["absolute", "obsolete"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immediatedly"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("categoricaly"),
                &["categorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("preivews"), &["previews"]),
            (
                dictgen::InsensitiveStr::Ascii("priororities"),
                &["priorities"],
            ),
            (dictgen::InsensitiveStr::Ascii("kryptonie"), &["kryptonite"]),
            (
                dictgen::InsensitiveStr::Ascii("traslation"),
                &["translation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("identtifies"),
                &["identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("wensday"), &["wednesday"]),
            (
                dictgen::InsensitiveStr::Ascii("montor"),
                &["monitor", "motor", "mentor"],
            ),
            (dictgen::InsensitiveStr::Ascii("pefixed"), &["prefixed"]),
            (dictgen::InsensitiveStr::Ascii("spawnve"), &["spawn"]),
            (
                dictgen::InsensitiveStr::Ascii("hambergers"),
                &["hamburgers"],
            ),
            (dictgen::InsensitiveStr::Ascii("airboner"), &["airborne"]),
            (
                dictgen::InsensitiveStr::Ascii("monolopies"),
                &["monopolies"],
            ),
            (dictgen::InsensitiveStr::Ascii("dominaters"), &["dominates"]),
            (
                dictgen::InsensitiveStr::Ascii("unapprieciative"),
                &["unappreciative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("injusticas"),
                &["injustices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perforamnces"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanacing"), &["advancing"]),
            (
                dictgen::InsensitiveStr::Ascii("overridde"),
                &["overridden", "override"],
            ),
            (dictgen::InsensitiveStr::Ascii("imbraced"), &["embraced"]),
            (dictgen::InsensitiveStr::Ascii("spanws"), &["spawns"]),
            (
                dictgen::InsensitiveStr::Ascii("bettr"),
                &["better", "bettor"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabes"), &["disables"]),
            (dictgen::InsensitiveStr::Ascii("yersa"), &["years"]),
            (dictgen::InsensitiveStr::Ascii("derfien"), &["define"]),
            (dictgen::InsensitiveStr::Ascii("foulded"), &["folded"]),
            (
                dictgen::InsensitiveStr::Ascii("featches"),
                &["fetches", "features"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unobstrusive"),
                &["unobtrusive"],
            ),
            (dictgen::InsensitiveStr::Ascii("upsteamer"), &["upstreamer"]),
            (dictgen::InsensitiveStr::Ascii("exteranl"), &["external"]),
            (dictgen::InsensitiveStr::Ascii("reveiw"), &["review"]),
            (
                dictgen::InsensitiveStr::Ascii("alogrithmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("intiger"), &["integer"]),
            (dictgen::InsensitiveStr::Ascii("forcused"), &["focused"]),
            (
                dictgen::InsensitiveStr::Ascii("nnovisheates"),
                &["novitiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumnets"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("distnces"), &["distances"]),
            (dictgen::InsensitiveStr::Ascii("assitsive"), &["assistive"]),
            (
                dictgen::InsensitiveStr::Ascii("sepearation"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("provicers"), &["providers"]),
            (
                dictgen::InsensitiveStr::Ascii("mispelling"),
                &["misspelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("lasagnia"), &["lasagna"]),
            (
                dictgen::InsensitiveStr::Ascii("confgurable"),
                &["configurable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curnilinear"),
                &["curvilinear"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desserialised"),
                &["deserialised"],
            ),
            (dictgen::InsensitiveStr::Ascii("faries"), &["fairies"]),
            (
                dictgen::InsensitiveStr::Ascii("consentrates"),
                &["concentrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("ajusts"), &["adjusts"]),
            (
                dictgen::InsensitiveStr::Ascii("overwriteable"),
                &["overwritable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("libretarians"),
                &["libertarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("optimier"),
                &["optimizer", "optimiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("diiferent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("subscribtions"),
                &["subscriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkclist"), &["checklist"]),
            (dictgen::InsensitiveStr::Ascii("scrpit"), &["script"]),
            (dictgen::InsensitiveStr::Ascii("enlcosed"), &["enclosed"]),
            (dictgen::InsensitiveStr::Ascii("wesite"), &["website"]),
            (dictgen::InsensitiveStr::Ascii("ecomonics"), &["economics"]),
            (dictgen::InsensitiveStr::Ascii("rapore"), &["rapport"]),
            (
                dictgen::InsensitiveStr::Ascii("responsibiliy"),
                &["responsibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curvasious"),
                &["curvacious", "curvaceous"],
            ),
            (dictgen::InsensitiveStr::Ascii("brfore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("numering"), &["numbering"]),
            (dictgen::InsensitiveStr::Ascii("comverter"), &["converter"]),
            (
                dictgen::InsensitiveStr::Ascii("receptionest"),
                &["receptionist"],
            ),
            (dictgen::InsensitiveStr::Ascii("exisits"), &["exists"]),
            (
                dictgen::InsensitiveStr::Ascii("transmutter"),
                &["transmitter"],
            ),
            (dictgen::InsensitiveStr::Ascii("witdhs"), &["widths"]),
            (dictgen::InsensitiveStr::Ascii("opreate"), &["operate"]),
            (
                dictgen::InsensitiveStr::Ascii("evaluataion"),
                &["evaluation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mempry"), &["memory"]),
            (
                dictgen::InsensitiveStr::Ascii("apacolyptic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("brazillian"), &["brazilian"]),
            (dictgen::InsensitiveStr::Ascii("rainbowe"), &["rainbow"]),
            (dictgen::InsensitiveStr::Ascii("depndants"), &["dependants"]),
            (dictgen::InsensitiveStr::Ascii("calues"), &["values"]),
            (dictgen::InsensitiveStr::Ascii("attirbs"), &["attribs"]),
            (dictgen::InsensitiveStr::Ascii("ascneders"), &["ascenders"]),
            (
                dictgen::InsensitiveStr::Ascii("millsencond"),
                &["millisecond"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nestolgicly"),
                &["nostalgically"],
            ),
            (dictgen::InsensitiveStr::Ascii("disnegage"), &["disengage"]),
            (
                dictgen::InsensitiveStr::Ascii("affectionatley"),
                &["affectionate", "affectionately"],
            ),
            (dictgen::InsensitiveStr::Ascii("reapper"), &["reappear"]),
            (
                dictgen::InsensitiveStr::Ascii("instrucions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("searchign"), &["searching"]),
            (dictgen::InsensitiveStr::Ascii("publicy"), &["publicly"]),
            (dictgen::InsensitiveStr::Ascii("deafault"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("unpredictabe"),
                &["unpredictable"],
            ),
            (dictgen::InsensitiveStr::Ascii("relpacing"), &["replacing"]),
            (dictgen::InsensitiveStr::Ascii("starins"), &["strains"]),
            (dictgen::InsensitiveStr::Ascii("parellels"), &["parallels"]),
            (dictgen::InsensitiveStr::Ascii("shareing"), &["sharing"]),
            (dictgen::InsensitiveStr::Ascii("hypevisor"), &["hypervisor"]),
            (
                dictgen::InsensitiveStr::Ascii("dissobedience"),
                &["disobedience"],
            ),
            (dictgen::InsensitiveStr::Ascii("minature"), &["miniature"]),
            (dictgen::InsensitiveStr::Ascii("initalism"), &["initialism"]),
            (dictgen::InsensitiveStr::Ascii("pretendias"), &["pretends"]),
            (
                dictgen::InsensitiveStr::Ascii("aditionnal"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("recuces"), &["reduces"]),
            (dictgen::InsensitiveStr::Ascii("magicain"), &["magician"]),
            (
                dictgen::InsensitiveStr::Ascii("pretentios"),
                &["pretentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsizided"),
                &["subsidized"],
            ),
            (dictgen::InsensitiveStr::Ascii("allwos"), &["allows"]),
            (
                dictgen::InsensitiveStr::Ascii("tranparency"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("emulaction"), &["emulation"]),
            (
                dictgen::InsensitiveStr::Ascii("assistanat"),
                &["assistants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satasfaction"),
                &["satisfaction"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfense"), &["offense"]),
            (
                dictgen::InsensitiveStr::Ascii("encyclopadia"),
                &["encyclopedia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civilasations"),
                &["civilisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbcribers"),
                &["subscribers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documenatries"),
                &["documentaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("presmissions"),
                &["permissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initializaitons"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fortitudine"),
                &["fortitude"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thunderjolt"),
                &["thunderbolt"],
            ),
            (dictgen::InsensitiveStr::Ascii("reudced"), &["reduced"]),
            (
                dictgen::InsensitiveStr::Ascii("imperialfist"),
                &["imperialist"],
            ),
            (dictgen::InsensitiveStr::Ascii("yiou"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("mutliplicity"),
                &["multiplicity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomendations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reigmes"), &["regimes"]),
            (
                dictgen::InsensitiveStr::Ascii("conotroling"),
                &["controlling"],
            ),
            (dictgen::InsensitiveStr::Ascii("emurator"), &["emulator"]),
            (
                dictgen::InsensitiveStr::Ascii("wotkspaces"),
                &["workspaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contridicted"),
                &["contradicted"],
            ),
            (dictgen::InsensitiveStr::Ascii("pologons"), &["polygons"]),
            (dictgen::InsensitiveStr::Ascii("researving"), &["reserving"]),
            (dictgen::InsensitiveStr::Ascii("cahced"), &["cached"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalsed"),
                &["capitalised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phenomemon"),
                &["phenomenon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acction"),
                &["action", "auction"],
            ),
            (dictgen::InsensitiveStr::Ascii("elctron"), &["electron"]),
            (dictgen::InsensitiveStr::Ascii("mormal"), &["normal"]),
            (dictgen::InsensitiveStr::Ascii("yoy"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("exceirpt"), &["excerpt"]),
            (dictgen::InsensitiveStr::Ascii("missmatch"), &["mismatch"]),
            (dictgen::InsensitiveStr::Ascii("crystyal"), &["crystal"]),
            (dictgen::InsensitiveStr::Ascii("catpuring"), &["capturing"]),
            (
                dictgen::InsensitiveStr::Ascii("thoeretically"),
                &["theoretically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("particually"),
                &["particularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("influens"), &["influences"]),
            (dictgen::InsensitiveStr::Ascii("morter"), &["mortar"]),
            (dictgen::InsensitiveStr::Ascii("fempto"), &["femto"]),
            (dictgen::InsensitiveStr::Ascii("freqquency"), &["frequency"]),
            (
                dictgen::InsensitiveStr::Ascii("exerimental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("controll"), &["control"]),
            (dictgen::InsensitiveStr::Ascii("belifes"), &["beliefs"]),
            (dictgen::InsensitiveStr::Ascii("angshis"), &["anxious"]),
            (
                dictgen::InsensitiveStr::Ascii("attitional"),
                &["additional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribuets"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("asisstant"), &["assistant"]),
            (dictgen::InsensitiveStr::Ascii("hollowcost"), &["holocaust"]),
            (
                dictgen::InsensitiveStr::Ascii("authoroties"),
                &["authorities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprciates"),
                &["appreciates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diasbilities"),
                &["disabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("exerpt"), &["excerpt"]),
            (dictgen::InsensitiveStr::Ascii("adddition"), &["addition"]),
            (dictgen::InsensitiveStr::Ascii("bouinds"), &["bounds"]),
            (
                dictgen::InsensitiveStr::Ascii("contestaste"),
                &["contestants"],
            ),
            (dictgen::InsensitiveStr::Ascii("lesiure"), &["leisure"]),
            (
                dictgen::InsensitiveStr::Ascii("postdomiator"),
                &["postdominator"],
            ),
            (dictgen::InsensitiveStr::Ascii("druming"), &["drumming"]),
            (
                dictgen::InsensitiveStr::Ascii("depencendencys"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pocessing"),
                &["processing", "possessing"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepearte"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("genenerates"),
                &["generates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privilaged"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("registed"), &["registered"]),
            (
                dictgen::InsensitiveStr::Ascii("broadcastes"),
                &["broadcasts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surpisingly"),
                &["surprisingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contingincy"),
                &["contingency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumsicion"),
                &["circumcision"],
            ),
            (dictgen::InsensitiveStr::Ascii("roation"), &["rotation"]),
            (dictgen::InsensitiveStr::Ascii("dicides"), &["decides"]),
            (
                dictgen::InsensitiveStr::Ascii("acomodated"),
                &["accommodated"],
            ),
            (dictgen::InsensitiveStr::Ascii("enginner"), &["engineer"]),
            (
                dictgen::InsensitiveStr::Ascii("oncompressed"),
                &["uncompressed"],
            ),
            (dictgen::InsensitiveStr::Ascii("akransas"), &["arkansas"]),
            (dictgen::InsensitiveStr::Ascii("kyebosh"), &["kibosh"]),
            (dictgen::InsensitiveStr::Ascii("respoding"), &["responding"]),
            (dictgen::InsensitiveStr::Ascii("felisatus"), &["felicitous"]),
            (dictgen::InsensitiveStr::Ascii("satisties"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentes"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bunds"),
                &["binds", "bounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("gloab"), &["globe"]),
            (
                dictgen::InsensitiveStr::Ascii("comaptibilities"),
                &["compatibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("optomized"), &["optimized"]),
            (
                dictgen::InsensitiveStr::Ascii("thermomiter"),
                &["thermometer"],
            ),
            (dictgen::InsensitiveStr::Ascii("ouputs"), &["outputs"]),
            (dictgen::InsensitiveStr::Ascii("jounral"), &["journal"]),
            (dictgen::InsensitiveStr::Ascii("piegons"), &["pigeons"]),
            (
                dictgen::InsensitiveStr::Ascii("battelstar"),
                &["battlestar"],
            ),
            (dictgen::InsensitiveStr::Ascii("acident"), &["accident"]),
            (
                dictgen::InsensitiveStr::Ascii("athleticsim"),
                &["athleticism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sacrificies"),
                &["sacrifice"],
            ),
            (dictgen::InsensitiveStr::Ascii("introuduce"), &["introduce"]),
            (
                dictgen::InsensitiveStr::Ascii("autheniticate"),
                &["authenticate"],
            ),
            (dictgen::InsensitiveStr::Ascii("inteacts"), &["interacts"]),
            (
                dictgen::InsensitiveStr::Ascii("interrutpible"),
                &["interruptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consultate"),
                &["consultant"],
            ),
            (dictgen::InsensitiveStr::Ascii("enbracing"), &["embracing"]),
            (
                dictgen::InsensitiveStr::Ascii("flimmakers"),
                &["filmmakers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bootstaps"), &["bootstraps"]),
            (dictgen::InsensitiveStr::Ascii("pertient"), &["pertinent"]),
            (dictgen::InsensitiveStr::Ascii("sheduler"), &["scheduler"]),
            (dictgen::InsensitiveStr::Ascii("compots"), &["compost"]),
            (
                dictgen::InsensitiveStr::Ascii("tetrahedralizaion"),
                &["tetrahedralization"],
            ),
            (dictgen::InsensitiveStr::Ascii("flitering"), &["filtering"]),
            (dictgen::InsensitiveStr::Ascii("eixsting"), &["existing"]),
            (dictgen::InsensitiveStr::Ascii("feitshes"), &["fetishes"]),
            (
                dictgen::InsensitiveStr::Ascii("flase"),
                &["false", "flake", "flame", "flare", "flash", "flask"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concusison"),
                &["concussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asbtracted"),
                &["abstracted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cristallization"),
                &["crystallization"],
            ),
            (dictgen::InsensitiveStr::Ascii("stirrs"), &["stirs"]),
            (
                dictgen::InsensitiveStr::Ascii("accidentely"),
                &["accidentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contrbuted"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reseptionist"),
                &["receptionist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certicated"),
                &["certificated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insenstivie"),
                &["insensitive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("artifactig"),
                &["artifacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("streamare"), &["streamer"]),
            (dictgen::InsensitiveStr::Ascii("diagnonal"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("cigaretts"), &["cigarettes"]),
            (dictgen::InsensitiveStr::Ascii("varances"), &["variances"]),
            (dictgen::InsensitiveStr::Ascii("promiss"), &["promise"]),
            (dictgen::InsensitiveStr::Ascii("solveing"), &["solving"]),
            (dictgen::InsensitiveStr::Ascii("obtaioned"), &["obtained"]),
            (dictgen::InsensitiveStr::Ascii("vaildity"), &["validity"]),
            (
                dictgen::InsensitiveStr::Ascii("relfectivity"),
                &["reflectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("concurret"), &["concurrent"]),
            (
                dictgen::InsensitiveStr::Ascii("bureaucractic"),
                &["bureaucratic"],
            ),
            (dictgen::InsensitiveStr::Ascii("superseede"), &["supersede"]),
            (dictgen::InsensitiveStr::Ascii("tomorow"), &["tomorrow"]),
            (dictgen::InsensitiveStr::Ascii("acceess"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("watchign"), &["watching"]),
            (
                dictgen::InsensitiveStr::Ascii("jalusey"),
                &["jealousy", "jalousie"],
            ),
            (dictgen::InsensitiveStr::Ascii("readius"), &["radius"]),
            (dictgen::InsensitiveStr::Ascii("spefifers"), &["specifiers"]),
            (
                dictgen::InsensitiveStr::Ascii("altogheter"),
                &["altogether"],
            ),
            (dictgen::InsensitiveStr::Ascii("broncoes"), &["broncos"]),
            (dictgen::InsensitiveStr::Ascii("oustpoken"), &["outspoken"]),
            (
                dictgen::InsensitiveStr::Ascii("intitalizer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("unifiy"), &["unify"]),
            (
                dictgen::InsensitiveStr::Ascii("charictors"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("offerd"), &["offered"]),
            (
                dictgen::InsensitiveStr::Ascii("specificfation"),
                &["specification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("leuitenant"),
                &["lieutenant"],
            ),
            (dictgen::InsensitiveStr::Ascii("quaility"), &["quality"]),
            (dictgen::InsensitiveStr::Ascii("deplys"), &["deploys"]),
            (dictgen::InsensitiveStr::Ascii("midifeld"), &["midfield"]),
            (
                dictgen::InsensitiveStr::Ascii("combintaorial"),
                &["combinatorial"],
            ),
            (dictgen::InsensitiveStr::Ascii("nuclues"), &["nucleus"]),
            (dictgen::InsensitiveStr::Ascii("acconted"), &["accounted"]),
            (
                dictgen::InsensitiveStr::Ascii("genralisations"),
                &["generalisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("conflifts"), &["conflicts"]),
            (dictgen::InsensitiveStr::Ascii("eloquenty"), &["eloquently"]),
            (
                dictgen::InsensitiveStr::Ascii("experimanet"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("packhages"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("caclulus"), &["calculus"]),
            (dictgen::InsensitiveStr::Ascii("maxmium"), &["maximum"]),
            (dictgen::InsensitiveStr::Ascii("funciotn"), &["function"]),
            (dictgen::InsensitiveStr::Ascii("trianlges"), &["triangles"]),
            (dictgen::InsensitiveStr::Ascii("insectos"), &["insects"]),
            (
                dictgen::InsensitiveStr::Ascii("confirgurations"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefixeirs"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("hatachi"), &["hitachi"]),
            (
                dictgen::InsensitiveStr::Ascii("supplemant"),
                &["supplemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("allocateng"),
                &["allocating"],
            ),
            (dictgen::InsensitiveStr::Ascii("protoge"), &["protege"]),
            (dictgen::InsensitiveStr::Ascii("opned"), &["opened"]),
            (dictgen::InsensitiveStr::Ascii("passsword"), &["password"]),
            (dictgen::InsensitiveStr::Ascii("trianing"), &["training"]),
            (dictgen::InsensitiveStr::Ascii("carinval"), &["carnival"]),
            (dictgen::InsensitiveStr::Ascii("editted"), &["edited"]),
            (dictgen::InsensitiveStr::Ascii("rilvaries"), &["rivalries"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronize"),
                &["synchronize"],
            ),
            (dictgen::InsensitiveStr::Ascii("bsaically"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("experamental"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("exlated"), &["exalted"]),
            (
                dictgen::InsensitiveStr::Ascii("pretifying"),
                &["prettifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("nofitying"), &["notifying"]),
            (
                dictgen::InsensitiveStr::Ascii("ineffiecent"),
                &["inefficient"],
            ),
            (dictgen::InsensitiveStr::Ascii("perpetue"), &["perpetuate"]),
            (
                dictgen::InsensitiveStr::Ascii("summariztion"),
                &["summarization"],
            ),
            (dictgen::InsensitiveStr::Ascii("acceppts"), &["accepts"]),
            (dictgen::InsensitiveStr::Ascii("artifiact"), &["artifact"]),
            (dictgen::InsensitiveStr::Ascii("optioinal"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("distrbution"),
                &["distribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invensions"),
                &["inventions"],
            ),
            (dictgen::InsensitiveStr::Ascii("permenant"), &["permanent"]),
            (dictgen::InsensitiveStr::Ascii("alotted"), &["allotted"]),
            (dictgen::InsensitiveStr::Ascii("criticas"), &["critics"]),
            (dictgen::InsensitiveStr::Ascii("spefiifeir"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("frustradet"),
                &["frustrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convervatism"),
                &["conservatism"],
            ),
            (dictgen::InsensitiveStr::Ascii("enterity"), &["entirety"]),
            (dictgen::InsensitiveStr::Ascii("croporate"), &["corporate"]),
            (
                dictgen::InsensitiveStr::Ascii("intrepreting"),
                &["interpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backpacing"),
                &["backpacking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("segegrated"),
                &["segregated"],
            ),
            (dictgen::InsensitiveStr::Ascii("bubbless"), &["bubbles"]),
            (dictgen::InsensitiveStr::Ascii("abstacts"), &["abstracts"]),
            (dictgen::InsensitiveStr::Ascii("scoreboad"), &["scoreboard"]),
            (
                dictgen::InsensitiveStr::Ascii("supervisoras"),
                &["supervisors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carachteristics"),
                &["characteristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("affaire"), &["affair"]),
            (dictgen::InsensitiveStr::Ascii("supresed"), &["suppressed"]),
            (dictgen::InsensitiveStr::Ascii("reliabley"), &["reliably"]),
            (
                dictgen::InsensitiveStr::Ascii("exponentiell"),
                &["exponential"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffreent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("catergorize"),
                &["categorize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimantal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("processpr"), &["processor"]),
            (dictgen::InsensitiveStr::Ascii("synonymus"), &["synonyms"]),
            (
                dictgen::InsensitiveStr::Ascii("jupytrer"),
                &["jupyter", "jupiter"],
            ),
            (dictgen::InsensitiveStr::Ascii("diagonaly"), &["diagonally"]),
            (dictgen::InsensitiveStr::Ascii("ofrward"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("neightborhood"),
                &["neighborhood"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commomwealth"),
                &["commonwealth"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhanged"), &["exchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("socioligical"),
                &["sociological"],
            ),
            (dictgen::InsensitiveStr::Ascii("cleaup"), &["cleanup"]),
            (dictgen::InsensitiveStr::Ascii("spectify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("typograhpy"),
                &["typography"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mispronounciation"),
                &["mispronunciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("wehereas"), &["whereas"]),
            (
                dictgen::InsensitiveStr::Ascii("regularlizer"),
                &["regularizer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("multplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("secreteries"),
                &["secretaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blackslashes"),
                &["backslashes"],
            ),
            (dictgen::InsensitiveStr::Ascii("aremnia"), &["armenia"]),
            (dictgen::InsensitiveStr::Ascii("physcial"), &["physical"]),
            (
                dictgen::InsensitiveStr::Ascii("controlers"),
                &["controls", "controllers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("approproximates"),
                &["approximates"],
            ),
            (dictgen::InsensitiveStr::Ascii("gloabal"), &["global"]),
            (
                dictgen::InsensitiveStr::Ascii("depdences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("maximimum"), &["maximum"]),
            (
                dictgen::InsensitiveStr::Ascii("deprecat"),
                &["deprecated", "deprecate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("maintianed"),
                &["maintained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("democracries"),
                &["democracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disributing"),
                &["distributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eratically"),
                &["erratically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opjectification"),
                &["objectification"],
            ),
            (dictgen::InsensitiveStr::Ascii("accesor"), &["accessor"]),
            (dictgen::InsensitiveStr::Ascii("coalesses"), &["coalesces"]),
            (
                dictgen::InsensitiveStr::Ascii("coniderably"),
                &["considerably"],
            ),
            (dictgen::InsensitiveStr::Ascii("glancin"), &["glancing"]),
            (dictgen::InsensitiveStr::Ascii("tecnical"), &["technical"]),
            (dictgen::InsensitiveStr::Ascii("linz"), &["lines"]),
            (dictgen::InsensitiveStr::Ascii("onotolgy"), &["ontology"]),
            (dictgen::InsensitiveStr::Ascii("thev"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("retension"),
                &["retention", "pretension", "recension"],
            ),
            (dictgen::InsensitiveStr::Ascii("regrests"), &["regress"]),
            (
                dictgen::InsensitiveStr::Ascii("coalacece"),
                &["coalesce", "coalescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converations"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constribution"),
                &["contribution"],
            ),
            (dictgen::InsensitiveStr::Ascii("summmary"), &["summary"]),
            (dictgen::InsensitiveStr::Ascii("incluide"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("progressivly"),
                &["progressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("dmangling"), &["demangling"]),
            (dictgen::InsensitiveStr::Ascii("minimmum"), &["minimum"]),
            (
                dictgen::InsensitiveStr::Ascii("imablances"),
                &["imbalances"],
            ),
            (dictgen::InsensitiveStr::Ascii("toplogies"), &["topologies"]),
            (
                dictgen::InsensitiveStr::Ascii("doocumenting"),
                &["documenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repeatidly"),
                &["repeatedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perview"),
                &["preview", "purview"],
            ),
            (dictgen::InsensitiveStr::Ascii("chidlfree"), &["childfree"]),
            (
                dictgen::InsensitiveStr::Ascii("sponsorhsip"),
                &["sponsorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("medicalert"), &["medicare"]),
            (dictgen::InsensitiveStr::Ascii("arguent"), &["argument"]),
            (
                dictgen::InsensitiveStr::Ascii("troublehsoot"),
                &["troubleshoot"],
            ),
            (dictgen::InsensitiveStr::Ascii("respoder"), &["responder"]),
            (dictgen::InsensitiveStr::Ascii("speficies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("wishpered"), &["whispered"]),
            (dictgen::InsensitiveStr::Ascii("facsist"), &["fascist"]),
            (dictgen::InsensitiveStr::Ascii("rebounce"), &["rebound"]),
            (
                dictgen::InsensitiveStr::Ascii("threre"),
                &["there", "three"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasturise"), &["pasteurise"]),
            (
                dictgen::InsensitiveStr::Ascii("expetecting"),
                &["expecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("futher"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("itnerpreter"),
                &["interpreter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("matainers"),
                &["maintainers"],
            ),
            (dictgen::InsensitiveStr::Ascii("wolffram"), &["wolfram"]),
            (
                dictgen::InsensitiveStr::Ascii("absitnence"),
                &["abstinence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millenials"),
                &["millennials"],
            ),
            (dictgen::InsensitiveStr::Ascii("versoned"), &["versioned"]),
            (dictgen::InsensitiveStr::Ascii("mexcian"), &["mexican"]),
            (dictgen::InsensitiveStr::Ascii("szied"), &["sized"]),
            (dictgen::InsensitiveStr::Ascii("antarctia"), &["antarctica"]),
            (dictgen::InsensitiveStr::Ascii("tollerant"), &["tolerant"]),
            (
                dictgen::InsensitiveStr::Ascii("intimidative"),
                &["intimidate"],
            ),
            (dictgen::InsensitiveStr::Ascii("subystems"), &["subsystems"]),
            (
                dictgen::InsensitiveStr::Ascii("mutilplier"),
                &["multiplier"],
            ),
            (dictgen::InsensitiveStr::Ascii("acatemy"), &["academy"]),
            (dictgen::InsensitiveStr::Ascii("elementy"), &["elementary"]),
            (dictgen::InsensitiveStr::Ascii("parfay"), &["parfait"]),
            (
                dictgen::InsensitiveStr::Ascii("scrooling"),
                &["scrolling", "schooling"],
            ),
            (dictgen::InsensitiveStr::Ascii("applly"), &["apply"]),
            (dictgen::InsensitiveStr::Ascii("dscribes"), &["describes"]),
            (
                dictgen::InsensitiveStr::Ascii("performaing"),
                &["performing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addministrations"),
                &["administrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("continous"), &["continuous"]),
            (dictgen::InsensitiveStr::Ascii("issus"), &["issues"]),
            (dictgen::InsensitiveStr::Ascii("directrory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("misspeling"),
                &["misspelling"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheklists"), &["checklists"]),
            (
                dictgen::InsensitiveStr::Ascii("organziations"),
                &["organizations"],
            ),
            (dictgen::InsensitiveStr::Ascii("weerdly"), &["weirdly"]),
            (
                dictgen::InsensitiveStr::Ascii("acciedential"),
                &["accidental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cointaining"),
                &["containing"],
            ),
            (dictgen::InsensitiveStr::Ascii("syndroms"), &["syndromes"]),
            (dictgen::InsensitiveStr::Ascii("alegiance"), &["allegiance"]),
            (dictgen::InsensitiveStr::Ascii("aberation"), &["aberration"]),
            (
                dictgen::InsensitiveStr::Ascii("logner"),
                &["logger", "longer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revoltuion"),
                &["revolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lifesytles"),
                &["lifestyles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresopndent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respectivily"),
                &["respectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canneled"),
                &["channeled", "canceled"],
            ),
            (dictgen::InsensitiveStr::Ascii("domecracy"), &["democracy"]),
            (dictgen::InsensitiveStr::Ascii("arro"), &["arrow"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmimating"),
                &["approximating"],
            ),
            (dictgen::InsensitiveStr::Ascii("creeations"), &["creations"]),
            (
                dictgen::InsensitiveStr::Ascii("parms"),
                &["params", "prams"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transocding"),
                &["transcoding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demoninations"),
                &["denominations"],
            ),
            (dictgen::InsensitiveStr::Ascii("registger"), &["register"]),
            (
                dictgen::InsensitiveStr::Ascii("origiginals"),
                &["originals"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoirding"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("registation"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("goup"), &["group"]),
            (dictgen::InsensitiveStr::Ascii("homeowrk"), &["homework"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensionsal"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("begininngs"),
                &["beginnings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addditional"),
                &["additional"],
            ),
            (dictgen::InsensitiveStr::Ascii("chatedral"), &["cathedral"]),
            (dictgen::InsensitiveStr::Ascii("resumsed"), &["resumed"]),
            (dictgen::InsensitiveStr::Ascii("iteartor"), &["iterator"]),
            (dictgen::InsensitiveStr::Ascii("collectros"), &["collects"]),
            (
                dictgen::InsensitiveStr::Ascii("ubiquoutous"),
                &["ubiquitous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reseracher"),
                &["researchers", "researcher"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotnent"), &["content"]),
            (
                dictgen::InsensitiveStr::Ascii("intreguing"),
                &["intriguing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continiously"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reposities"),
                &["repositories"],
            ),
            (dictgen::InsensitiveStr::Ascii("atheletes"), &["athletes"]),
            (
                dictgen::InsensitiveStr::Ascii("unresolvabvle"),
                &["unresolvable"],
            ),
            (dictgen::InsensitiveStr::Ascii("detaling"), &["detailing"]),
            (
                dictgen::InsensitiveStr::Ascii("propageting"),
                &["propagating"],
            ),
            (dictgen::InsensitiveStr::Ascii("plalforms"), &["platforms"]),
            (
                dictgen::InsensitiveStr::Ascii("armagedeon"),
                &["armageddon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowlegdeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("retorical"), &["rhetorical"]),
            (
                dictgen::InsensitiveStr::Ascii("investagate"),
                &["investigate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("oculd"), &["could"]),
            (
                dictgen::InsensitiveStr::Ascii("exracted"),
                &["extracted", "exacted"],
            ),
            (dictgen::InsensitiveStr::Ascii("featched"), &["fetched"]),
            (dictgen::InsensitiveStr::Ascii("speficifes"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("iechart"), &["piechart"]),
            (
                dictgen::InsensitiveStr::Ascii("metropolitcan"),
                &["metropolitan"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignormg"), &["ignoring"]),
            (dictgen::InsensitiveStr::Ascii("revivel"), &["revival"]),
            (dictgen::InsensitiveStr::Ascii("advertice"), &["advertise"]),
            (dictgen::InsensitiveStr::Ascii("pltforms"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("differring"), &["differing"]),
            (dictgen::InsensitiveStr::Ascii("forumla"), &["formula"]),
            (
                dictgen::InsensitiveStr::Ascii("recommandations"),
                &["recommendations"],
            ),
            (dictgen::InsensitiveStr::Ascii("variblaes"), &["variables"]),
            (
                dictgen::InsensitiveStr::Ascii("correspontingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("revriever"), &["retriever"]),
            (dictgen::InsensitiveStr::Ascii("discpline"), &["discipline"]),
            (
                dictgen::InsensitiveStr::Ascii("inplementing"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("providfers"), &["providers"]),
            (dictgen::InsensitiveStr::Ascii("edgin"), &["edging"]),
            (
                dictgen::InsensitiveStr::Ascii("antidepressent"),
                &["antidepressant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meditaciones"),
                &["medications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("undoubtetly"),
                &["undoubtedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exicting"),
                &["exciting", "existing", "exiting", "evicting", "exacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("luicfer"), &["lucifer"]),
            (
                dictgen::InsensitiveStr::Ascii("missspellings"),
                &["misspellings"],
            ),
            (dictgen::InsensitiveStr::Ascii("activties"), &["activities"]),
            (dictgen::InsensitiveStr::Ascii("unifroms"), &["uniforms"]),
            (
                dictgen::InsensitiveStr::Ascii("accelereates"),
                &["accelerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("utillises"), &["utilises"]),
            (dictgen::InsensitiveStr::Ascii("shepperd"), &["shepherd"]),
            (
                dictgen::InsensitiveStr::Ascii("cheerleeder"),
                &["cheerleader"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resposiblity"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("quafilied"), &["qualified"]),
            (dictgen::InsensitiveStr::Ascii("dirceting"), &["directing"]),
            (
                dictgen::InsensitiveStr::Ascii("incarcirated"),
                &["incarcerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("timzones"), &["timezones"]),
            (
                dictgen::InsensitiveStr::Ascii("integations"),
                &["integrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("analyised"), &["analysed"]),
            (dictgen::InsensitiveStr::Ascii("experment"), &["experiment"]),
            (dictgen::InsensitiveStr::Ascii("sillibus"), &["syllabus"]),
            (
                dictgen::InsensitiveStr::Ascii("transitionning"),
                &["transitioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communters"),
                &["commenters"],
            ),
            (dictgen::InsensitiveStr::Ascii("insluated"), &["insulated"]),
            (
                dictgen::InsensitiveStr::Ascii("exract"),
                &["extract", "exact"],
            ),
            (dictgen::InsensitiveStr::Ascii("availeable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("infleunced"),
                &["influenced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenedences"),
                &["dependences"],
            ),
            (dictgen::InsensitiveStr::Ascii("autmates"), &["automates"]),
            (
                dictgen::InsensitiveStr::Ascii("fundimentals"),
                &["fundamentals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simpifying"),
                &["simplifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("tthis"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("compains"), &["complains"]),
            (dictgen::InsensitiveStr::Ascii("apyooned"), &["oppugned"]),
            (dictgen::InsensitiveStr::Ascii("impuned"), &["impugned"]),
            (dictgen::InsensitiveStr::Ascii("acuiring"), &["acquiring"]),
            (
                dictgen::InsensitiveStr::Ascii("underastimating"),
                &["underestimating"],
            ),
            (dictgen::InsensitiveStr::Ascii("gabage"), &["garbage"]),
            (
                dictgen::InsensitiveStr::Ascii("occourences"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("nubmering"), &["numbering"]),
            (dictgen::InsensitiveStr::Ascii("paricle"), &["particle"]),
            (dictgen::InsensitiveStr::Ascii("comptuing"), &["computing"]),
            (
                dictgen::InsensitiveStr::Ascii("prohibiion"),
                &["prohibition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenricate"),
                &["authenticate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cooedinate"),
                &["coordinate"],
            ),
            (dictgen::InsensitiveStr::Ascii("reconized"), &["recognized"]),
            (
                dictgen::InsensitiveStr::Ascii("dimensionail"),
                &["dimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("registrated"),
                &["registered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sratched"),
                &["scratched", "snatched"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalutaed"), &["evaluated"]),
            (
                dictgen::InsensitiveStr::Ascii("parliamantary"),
                &["parliamentary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disrepsectful"),
                &["disrespectful"],
            ),
            (dictgen::InsensitiveStr::Ascii("eactly"), &["exactly"]),
            (dictgen::InsensitiveStr::Ascii("inentory"), &["inventory"]),
            (
                dictgen::InsensitiveStr::Ascii("multliplier"),
                &["multiplier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncondisional"),
                &["unconditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("servivce"), &["service"]),
            (
                dictgen::InsensitiveStr::Ascii("plagiarios"),
                &["plagiarism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combinaterial"),
                &["combinatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("milliescond"),
                &["millisecond"],
            ),
            (dictgen::InsensitiveStr::Ascii("arbort"), &["abort"]),
            (
                dictgen::InsensitiveStr::Ascii("equiring"),
                &["requiring", "enquiring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grpahically"),
                &["graphically"],
            ),
            (dictgen::InsensitiveStr::Ascii("wondows"), &["windows"]),
            (
                dictgen::InsensitiveStr::Ascii("heirachical"),
                &["hierarchical"],
            ),
            (dictgen::InsensitiveStr::Ascii("invitato"), &["invitation"]),
            (dictgen::InsensitiveStr::Ascii("fillin"), &["filling"]),
            (dictgen::InsensitiveStr::Ascii("crusdae"), &["crusade"]),
            (dictgen::InsensitiveStr::Ascii("abstactor"), &["abstractor"]),
            (
                dictgen::InsensitiveStr::Ascii("functiionality"),
                &["functionality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qhich"),
                &["quiche", "which"],
            ),
            (dictgen::InsensitiveStr::Ascii("repesent"), &["represent"]),
            (
                dictgen::InsensitiveStr::Ascii("disturben"),
                &["disturbance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alloed"),
                &["allowed", "aloud"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("addionally"),
                &["additionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manouverability"),
                &["maneuverability"],
            ),
            (dictgen::InsensitiveStr::Ascii("contatcts"), &["contacts"]),
            (dictgen::InsensitiveStr::Ascii("simpified"), &["simplified"]),
            (dictgen::InsensitiveStr::Ascii("pakge"), &["package"]),
            (dictgen::InsensitiveStr::Ascii("encompus"), &["encompass"]),
            (
                dictgen::InsensitiveStr::Ascii("eleiminates"),
                &["eliminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workaruond"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antiboitics"),
                &["antibiotics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chace"),
                &["cache", "chance", "chase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("covnersion"),
                &["conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("publsher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("resepecting"),
                &["respecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("bookmar"), &["bookmark"]),
            (
                dictgen::InsensitiveStr::Ascii("brainwahsed"),
                &["brainwashed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convenienes"),
                &["conveniences", "conveniencies", "convenes"],
            ),
            (dictgen::InsensitiveStr::Ascii("scourgin"), &["scourging"]),
            (
                dictgen::InsensitiveStr::Ascii("correnspondingly"),
                &["correspondingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bulletprrof"),
                &["bulletproof"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compactible"),
                &["compactable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfifiers"),
                &["identifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("validades"), &["validates"]),
            (
                dictgen::InsensitiveStr::Ascii("necromamcer"),
                &["necromancer"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepharded"), &["shepherd"]),
            (dictgen::InsensitiveStr::Ascii("toches"), &["touches"]),
            (dictgen::InsensitiveStr::Ascii("confrim"), &["confirm"]),
            (
                dictgen::InsensitiveStr::Ascii("sandviches"),
                &["sandwiches"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("countercat"),
                &["counteract"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismanteled"),
                &["dismantled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reouting"),
                &["routing", "rerouting"],
            ),
            (dictgen::InsensitiveStr::Ascii("elease"), &["release"]),
            (dictgen::InsensitiveStr::Ascii("versioing"), &["versioning"]),
            (dictgen::InsensitiveStr::Ascii("cruzing"), &["cruising"]),
            (dictgen::InsensitiveStr::Ascii("ehtier"), &["either"]),
            (dictgen::InsensitiveStr::Ascii("languague"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("imporove"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("survye"), &["survey"]),
            (dictgen::InsensitiveStr::Ascii("anaolgue"), &["analogue"]),
            (dictgen::InsensitiveStr::Ascii("nurtients"), &["nutrients"]),
            (
                dictgen::InsensitiveStr::Ascii("innacurately"),
                &["inaccurately"],
            ),
            (dictgen::InsensitiveStr::Ascii("socripted"), &["scripted"]),
            (dictgen::InsensitiveStr::Ascii("lepards"), &["leopards"]),
            (
                dictgen::InsensitiveStr::Ascii("arbtirarily"),
                &["arbitrarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("rednerer"), &["renderer"]),
            (dictgen::InsensitiveStr::Ascii("allcation"), &["allocation"]),
            (dictgen::InsensitiveStr::Ascii("unexpeced"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("gernal"),
                &["general", "journal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sceintific"),
                &["scientific"],
            ),
            (dictgen::InsensitiveStr::Ascii("obsevrers"), &["observers"]),
            (dictgen::InsensitiveStr::Ascii("excetpion"), &["exception"]),
            (
                dictgen::InsensitiveStr::Ascii("reposible"),
                &["responsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumud"), &["assumed"]),
            (
                dictgen::InsensitiveStr::Ascii("raison"),
                &["reason", "raisin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteference"),
                &["interference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("documentaions"),
                &["documentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("efel"), &["evil"]),
            (
                dictgen::InsensitiveStr::Ascii("overbearring"),
                &["overbearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquianted"),
                &["acquainted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequecies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("conservin"), &["conserving"]),
            (dictgen::InsensitiveStr::Ascii("piecwise"), &["piecewise"]),
            (dictgen::InsensitiveStr::Ascii("margarent"), &["margaret"]),
            (
                dictgen::InsensitiveStr::Ascii("territorist"),
                &["terrorist"],
            ),
            (dictgen::InsensitiveStr::Ascii("preppendet"), &["prepended"]),
            (dictgen::InsensitiveStr::Ascii("deffirent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("solf"), &["solve", "sold"]),
            (
                dictgen::InsensitiveStr::Ascii("transporteur"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("accoridng"), &["according"]),
            (
                dictgen::InsensitiveStr::Ascii("exageration"),
                &["exaggeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgarades"),
                &["downgrades"],
            ),
            (dictgen::InsensitiveStr::Ascii("plehora"), &["plethora"]),
            (dictgen::InsensitiveStr::Ascii("mnemoic"), &["mnemonic"]),
            (dictgen::InsensitiveStr::Ascii("basf"), &["base"]),
            (
                dictgen::InsensitiveStr::Ascii("overwrittes"),
                &["overwrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("irritacion"),
                &["irritation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derivativs"),
                &["derivatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("downloade"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("entchantment"),
                &["enchantment"],
            ),
            (dictgen::InsensitiveStr::Ascii("releses"), &["releases"]),
            (
                dictgen::InsensitiveStr::Ascii("asynschron"),
                &["asynchronous"],
            ),
            (dictgen::InsensitiveStr::Ascii("jurnys"), &["journeys"]),
            (dictgen::InsensitiveStr::Ascii("tarvisci"), &["travisci"]),
            (dictgen::InsensitiveStr::Ascii("impuslive"), &["impulsive"]),
            (
                dictgen::InsensitiveStr::Ascii("automatical"),
                &["automatically", "automatic", "automated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("konfidential"),
                &["confidential"],
            ),
            (dictgen::InsensitiveStr::Ascii("prepand"), &["prepend"]),
            (
                dictgen::InsensitiveStr::Ascii("delimitiaion"),
                &["delimitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("disscuses"), &["discusses"]),
            (
                dictgen::InsensitiveStr::Ascii("apropreate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommendataion"),
                &["recommendation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("susbcribe"), &["subscribe"]),
            (dictgen::InsensitiveStr::Ascii("annotor"), &["annotator"]),
            (dictgen::InsensitiveStr::Ascii("lifestye"), &["lifestyle"]),
            (dictgen::InsensitiveStr::Ascii("messagepad"), &["messaged"]),
            (
                dictgen::InsensitiveStr::Ascii("divisionals"),
                &["divisions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("snythetics"),
                &["synthetics"],
            ),
            (dictgen::InsensitiveStr::Ascii("crewsant"), &["croissant"]),
            (
                dictgen::InsensitiveStr::Ascii("addictin"),
                &["addictions", "addicting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("confedercy"),
                &["confederacy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("routet"),
                &["routed", "route", "router"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inticracies"),
                &["intricacies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commeting"),
                &["commenting", "competing"],
            ),
            (dictgen::InsensitiveStr::Ascii("experes"), &["express"]),
            (dictgen::InsensitiveStr::Ascii("fodler"), &["folder"]),
            (
                dictgen::InsensitiveStr::Ascii("councels"),
                &["councils", "counsels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sipporting"),
                &["supporting"],
            ),
            (dictgen::InsensitiveStr::Ascii("tansforms"), &["transforms"]),
            (dictgen::InsensitiveStr::Ascii("focu"), &["focus"]),
            (
                dictgen::InsensitiveStr::Ascii("cercomstances"),
                &["circumstances"],
            ),
            (dictgen::InsensitiveStr::Ascii("epslion"), &["epsilon"]),
            (
                dictgen::InsensitiveStr::Ascii("choppin"),
                &["chopping", "chopin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tortoitse"),
                &["tortoise", "tortoises"],
            ),
            (dictgen::InsensitiveStr::Ascii("adreses"), &["addresses"]),
            (
                dictgen::InsensitiveStr::Ascii("capitalizaiton"),
                &["capitalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("pharoah"), &["pharaoh"]),
            (
                dictgen::InsensitiveStr::Ascii("measuement"),
                &["measurement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("achitectures"),
                &["architectures"],
            ),
            (dictgen::InsensitiveStr::Ascii("searhers"), &["searchers"]),
            (
                dictgen::InsensitiveStr::Ascii("documeneted"),
                &["documented"],
            ),
            (dictgen::InsensitiveStr::Ascii("decleare"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("stength"), &["strength"]),
            (dictgen::InsensitiveStr::Ascii("drity"), &["dirty"]),
            (
                dictgen::InsensitiveStr::Ascii("affiliato"),
                &["affiliation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pertitent"),
                &["pertinent", "persistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reallocaitons"),
                &["reallocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerative"), &["creative"]),
            (dictgen::InsensitiveStr::Ascii("carribean"), &["caribbean"]),
            (
                dictgen::InsensitiveStr::Ascii("summmarises"),
                &["summarises"],
            ),
            (dictgen::InsensitiveStr::Ascii("uique"), &["unique"]),
            (
                dictgen::InsensitiveStr::Ascii("entretaining"),
                &["entertaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("augest"), &["august"]),
            (dictgen::InsensitiveStr::Ascii("deram"), &["dram", "dream"]),
            (dictgen::InsensitiveStr::Ascii("togglin"), &["toggling"]),
            (dictgen::InsensitiveStr::Ascii("twoo"), &["two", "too"]),
            (dictgen::InsensitiveStr::Ascii("scrunity"), &["scrutiny"]),
            (
                dictgen::InsensitiveStr::Ascii("comprehand"),
                &["comprehend"],
            ),
            (dictgen::InsensitiveStr::Ascii("continius"), &["continuous"]),
            (
                dictgen::InsensitiveStr::Ascii("affilation"),
                &["affiliation"],
            ),
            (dictgen::InsensitiveStr::Ascii("anoys"), &["annoys"]),
            (
                dictgen::InsensitiveStr::Ascii("instantaenous"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("copyritghted"),
                &["copyrighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalistc"),
                &["journalistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expalining"),
                &["explaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("suspendeds"), &["suspense"]),
            (
                dictgen::InsensitiveStr::Ascii("comprehention"),
                &["comprehension"],
            ),
            (dictgen::InsensitiveStr::Ascii("miltisite"), &["multisite"]),
            (dictgen::InsensitiveStr::Ascii("emought"), &["enough"]),
            (
                dictgen::InsensitiveStr::Ascii("hexidecimals"),
                &["hexadecimals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exacutable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunat"),
                &["unfortunate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specailist"),
                &["specialist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicationg"),
                &["indicating", "indication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acculumating"),
                &["accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("declaired"), &["declared"]),
            (
                dictgen::InsensitiveStr::Ascii("diffenrence"),
                &["difference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indicitive"),
                &["indicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaildate"), &["validate"]),
            (
                dictgen::InsensitiveStr::Ascii("ineqaulity"),
                &["inequality"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorihmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("phsyician"), &["physician"]),
            (dictgen::InsensitiveStr::Ascii("seperat"), &["separate"]),
            (
                dictgen::InsensitiveStr::Ascii("transmist"),
                &["transmit", "transmits"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completor"),
                &["completer", "completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implemanting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("evolutionarly"),
                &["evolutionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("argubaly"), &["arguably"]),
            (
                dictgen::InsensitiveStr::Ascii("intagrated"),
                &["integrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abording"),
                &["aborting", "boarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuctures"), &["structures"]),
            (dictgen::InsensitiveStr::Ascii("lisetning"), &["listening"]),
            (
                dictgen::InsensitiveStr::Ascii("prejiduced"),
                &["prejudiced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ternimating"),
                &["terminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insance"),
                &["instance", "insane", "insurance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elimintation"),
                &["elimination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reminsicent"),
                &["reminiscent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconviences"),
                &["inconveniences"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interprated"),
                &["interpreted"],
            ),
            (dictgen::InsensitiveStr::Ascii("offsrping"), &["offspring"]),
            (dictgen::InsensitiveStr::Ascii("cecklists"), &["checklists"]),
            (
                dictgen::InsensitiveStr::Ascii("reconfifure"),
                &["reconfigure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ackumulation"),
                &["accumulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transfender"),
                &["transgender"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expresssion"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referncence"),
                &["reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("logtiech"), &["logitech"]),
            (
                dictgen::InsensitiveStr::Ascii("informatice"),
                &["informative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termniating"),
                &["terminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("knowladgable"),
                &["knowledgable"],
            ),
            (dictgen::InsensitiveStr::Ascii("feeback"), &["feedback"]),
            (
                dictgen::InsensitiveStr::Ascii("semiphores"),
                &["semaphores"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("regeneraton"),
                &["regeneration"],
            ),
            (dictgen::InsensitiveStr::Ascii("regresion"), &["regression"]),
            (
                dictgen::InsensitiveStr::Ascii("penetratiing"),
                &["penetrating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misrepreseted"),
                &["misrepresented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitalsit"),
                &["capitalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleratrion"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("projction"), &["projection"]),
            (dictgen::InsensitiveStr::Ascii("idenities"), &["identities"]),
            (dictgen::InsensitiveStr::Ascii("fysisist"), &["physicist"]),
            (dictgen::InsensitiveStr::Ascii("persuit"), &["pursuit"]),
            (
                dictgen::InsensitiveStr::Ascii("authetnicating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convered"),
                &["converted", "covered", "converged", "conveyed"],
            ),
            (dictgen::InsensitiveStr::Ascii("machiery"), &["machinery"]),
            (dictgen::InsensitiveStr::Ascii("dequed"), &["dequeued"]),
            (
                dictgen::InsensitiveStr::Ascii("devived"),
                &["derived", "devised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aktivities"),
                &["activities"],
            ),
            (dictgen::InsensitiveStr::Ascii("statment"), &["statement"]),
            (
                dictgen::InsensitiveStr::Ascii("backgournd"),
                &["background"],
            ),
            (dictgen::InsensitiveStr::Ascii("stalkes"), &["stalkers"]),
            (
                dictgen::InsensitiveStr::Ascii("asympyotical"),
                &["asymptotical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unflexible"),
                &["inflexible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underneith"),
                &["underneath"],
            ),
            (dictgen::InsensitiveStr::Ascii("truncat"), &["truncate"]),
            (
                dictgen::InsensitiveStr::Ascii("calucalted"),
                &["calculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("versiooning"),
                &["versioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("seeked"), &["sought"]),
            (
                dictgen::InsensitiveStr::Ascii("exclused"),
                &["excluded", "excused"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquiesence"),
                &["acquiescence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authethicators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("careing"), &["caring"]),
            (
                dictgen::InsensitiveStr::Ascii("coherantly"),
                &["coherently"],
            ),
            (dictgen::InsensitiveStr::Ascii("justs"), &["just"]),
            (
                dictgen::InsensitiveStr::Ascii("exnternally"),
                &["externally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("courcework"),
                &["coursework"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimdntel"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mintor"),
                &["mentor", "monitor", "minor"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutomiser"), &["customiser"]),
            (dictgen::InsensitiveStr::Ascii("bronken"), &["broken"]),
            (
                dictgen::InsensitiveStr::Ascii("redeclaation"),
                &["redeclaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("afhganistan"),
                &["afghanistan"],
            ),
            (dictgen::InsensitiveStr::Ascii("indciate"), &["indicate"]),
            (dictgen::InsensitiveStr::Ascii("capapble"), &["capable"]),
            (dictgen::InsensitiveStr::Ascii("sometring"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("invariabil"),
                &["invariably"],
            ),
            (dictgen::InsensitiveStr::Ascii("quater"), &["quarter"]),
            (dictgen::InsensitiveStr::Ascii("nstaller"), &["installer"]),
            (
                dictgen::InsensitiveStr::Ascii("catgorization"),
                &["categorization"],
            ),
            (dictgen::InsensitiveStr::Ascii("critiicals"), &["criticals"]),
            (dictgen::InsensitiveStr::Ascii("bountries"), &["boundaries"]),
            (
                dictgen::InsensitiveStr::Ascii("sacreligious"),
                &["sacrilegious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disresepcted"),
                &["disrespected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wearn"),
                &["warn", "wear", "earn", "learn", "worn", "yearn"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disciminate"),
                &["discriminate"],
            ),
            (dictgen::InsensitiveStr::Ascii("dvided"), &["divided"]),
            (
                dictgen::InsensitiveStr::Ascii("misinterpretts"),
                &["misinterprets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("engagemet"),
                &["engagements"],
            ),
            (dictgen::InsensitiveStr::Ascii("nerally"), &["neurally"]),
            (dictgen::InsensitiveStr::Ascii("recgonize"), &["recognize"]),
            (dictgen::InsensitiveStr::Ascii("accademic"), &["academic"]),
            (dictgen::InsensitiveStr::Ascii("betwene"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("checsum"), &["checksum"]),
            (dictgen::InsensitiveStr::Ascii("decidely"), &["decidedly"]),
            (
                dictgen::InsensitiveStr::Ascii("flutterspy"),
                &["fluttershy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attriubted"),
                &["attributed"],
            ),
            (dictgen::InsensitiveStr::Ascii("apon"), &["upon", "apron"]),
            (dictgen::InsensitiveStr::Ascii("compontent"), &["component"]),
            (dictgen::InsensitiveStr::Ascii("forumula"), &["formula"]),
            (dictgen::InsensitiveStr::Ascii("comparble"), &["comparable"]),
            (
                dictgen::InsensitiveStr::Ascii("opprotunities"),
                &["opportunities"],
            ),
            (dictgen::InsensitiveStr::Ascii("prioriy"), &["priority"]),
            (
                dictgen::InsensitiveStr::Ascii("hypothesees"),
                &["hypotheses"],
            ),
            (dictgen::InsensitiveStr::Ascii("monlayers"), &["monolayers"]),
            (dictgen::InsensitiveStr::Ascii("buld"), &["build"]),
            (dictgen::InsensitiveStr::Ascii("mentinned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("sacksonville"),
                &["jacksonville"],
            ),
            (dictgen::InsensitiveStr::Ascii("efinitely"), &["definitely"]),
            (
                dictgen::InsensitiveStr::Ascii("contriction"),
                &["contraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depenedency"),
                &["dependency"],
            ),
            (dictgen::InsensitiveStr::Ascii("yoh"), &["you"]),
            (dictgen::InsensitiveStr::Ascii("containter"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("realiztion"),
                &["realization"],
            ),
            (dictgen::InsensitiveStr::Ascii("becaues"), &["because"]),
            (dictgen::InsensitiveStr::Ascii("smouth"), &["smooth"]),
            (dictgen::InsensitiveStr::Ascii("correspon"), &["correspond"]),
            (dictgen::InsensitiveStr::Ascii("alreight"), &["alright"]),
            (
                dictgen::InsensitiveStr::Ascii("noteworhty"),
                &["noteworthy"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("blueprients"),
                &["blueprints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investingating"),
                &["investigating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dekete"), &["delete"]),
            (
                dictgen::InsensitiveStr::Ascii("republicon"),
                &["republican"],
            ),
            (dictgen::InsensitiveStr::Ascii("exapnding"), &["expanding"]),
            (dictgen::InsensitiveStr::Ascii("redudnant"), &["redundant"]),
            (
                dictgen::InsensitiveStr::Ascii("reproducion"),
                &["reproduction"],
            ),
            (dictgen::InsensitiveStr::Ascii("retuens"), &["returns"]),
            (dictgen::InsensitiveStr::Ascii("dispays"), &["displays"]),
            (
                dictgen::InsensitiveStr::Ascii("callibrations"),
                &["calibrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqment"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("disscussed"), &["discussed"]),
            (dictgen::InsensitiveStr::Ascii("coallese"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("interstellaire"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successfullly"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("depcrecate"), &["deprecate"]),
            (
                dictgen::InsensitiveStr::Ascii("appplicable"),
                &["applicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reimplementaions"),
                &["reimplementations"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocaite"), &["allocate"]),
            (
                dictgen::InsensitiveStr::Ascii("ocurrences"),
                &["occurrences"],
            ),
            (dictgen::InsensitiveStr::Ascii("crashs"), &["crashes"]),
            (
                dictgen::InsensitiveStr::Ascii("templacing"),
                &["templating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("claculators"),
                &["calculators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfamilier"),
                &["unfamiliar"],
            ),
            (dictgen::InsensitiveStr::Ascii("dateime"), &["datetime"]),
            (dictgen::InsensitiveStr::Ascii("fourteeen"), &["fourteen"]),
            (
                dictgen::InsensitiveStr::Ascii("austrailans"),
                &["australians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalulated"),
                &["recalculated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparabile"),
                &["comparable"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffferent"), &["different"]),
            (dictgen::InsensitiveStr::Ascii("reuests"), &["requests"]),
            (dictgen::InsensitiveStr::Ascii("destry"), &["destroy"]),
            (dictgen::InsensitiveStr::Ascii("sizo"), &["size"]),
            (dictgen::InsensitiveStr::Ascii("uupload"), &["upload"]),
            (
                dictgen::InsensitiveStr::Ascii("initailzers"),
                &["initializers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unitialize"),
                &["uninitialize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("profitible"),
                &["profitable"],
            ),
            (dictgen::InsensitiveStr::Ascii("intesect"), &["intersect"]),
            (
                dictgen::InsensitiveStr::Ascii("decompresssion"),
                &["decompression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("laysers"),
                &["lasers", "layers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unintialized"),
                &["uninitialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nevelope"),
                &["envelop", "envelope"],
            ),
            (dictgen::InsensitiveStr::Ascii("triying"), &["trying"]),
            (
                dictgen::InsensitiveStr::Ascii("signitures"),
                &["signatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("senitive"), &["sensitive"]),
            (
                dictgen::InsensitiveStr::Ascii("idenfitifes"),
                &["identifies"],
            ),
            (dictgen::InsensitiveStr::Ascii("algorihms"), &["algorithms"]),
            (
                dictgen::InsensitiveStr::Ascii("confuigres"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valulations"),
                &["valuations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ineffecitve"),
                &["ineffective"],
            ),
            (dictgen::InsensitiveStr::Ascii("chambres"), &["chambers"]),
            (dictgen::InsensitiveStr::Ascii("defeaulted"), &["defaulted"]),
            (dictgen::InsensitiveStr::Ascii("povided"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("sanetizers"),
                &["sanitizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("millieseconds"),
                &["milliseconds"],
            ),
            (dictgen::InsensitiveStr::Ascii("camapign"), &["campaign"]),
            (dictgen::InsensitiveStr::Ascii("enpdoint"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("ponting"), &["pointing"]),
            (dictgen::InsensitiveStr::Ascii("victorain"), &["victorian"]),
            (
                dictgen::InsensitiveStr::Ascii("sunderlona"),
                &["sunderland"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tokenizaiton"),
                &["tokenization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indivdually"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitmally"), &["optimally"]),
            (dictgen::InsensitiveStr::Ascii("serialzie"), &["serialize"]),
            (dictgen::InsensitiveStr::Ascii("featuires"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("consequetive"),
                &["consecutive"],
            ),
            (dictgen::InsensitiveStr::Ascii("nostrills"), &["nostrils"]),
            (dictgen::InsensitiveStr::Ascii("irelevant"), &["irrelevant"]),
            (
                dictgen::InsensitiveStr::Ascii("otpimizations"),
                &["optimizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("newsettler"),
                &["newsletter"],
            ),
            (dictgen::InsensitiveStr::Ascii("specality"), &["speciality"]),
            (
                dictgen::InsensitiveStr::Ascii("expermental"),
                &["experimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaparisons"),
                &["comparisons"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configuraitons"),
                &["configurations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expementing"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("direcional"),
                &["directional"],
            ),
            (dictgen::InsensitiveStr::Ascii("courtey"), &["courtesy"]),
            (dictgen::InsensitiveStr::Ascii("dispensin"), &["dispensing"]),
            (dictgen::InsensitiveStr::Ascii("encypt"), &["encrypt"]),
            (dictgen::InsensitiveStr::Ascii("compond"), &["compound"]),
            (dictgen::InsensitiveStr::Ascii("functioed"), &["functioned"]),
            (
                dictgen::InsensitiveStr::Ascii("insectiverous"),
                &["insectivorous"],
            ),
            (dictgen::InsensitiveStr::Ascii("indiating"), &["indicating"]),
            (
                dictgen::InsensitiveStr::Ascii("australianas"),
                &["australians"],
            ),
            (dictgen::InsensitiveStr::Ascii("minimzes"), &["minimizes"]),
            (dictgen::InsensitiveStr::Ascii("heratige"), &["heritage"]),
            (
                dictgen::InsensitiveStr::Ascii("investegator"),
                &["investigator"],
            ),
            (dictgen::InsensitiveStr::Ascii("legnthy"), &["lengthy"]),
            (
                dictgen::InsensitiveStr::Ascii("restauranteur"),
                &["restaurateur"],
            ),
            (dictgen::InsensitiveStr::Ascii("turtlehs"), &["turtles"]),
            (
                dictgen::InsensitiveStr::Ascii("detetions"),
                &["deletions", "detections"],
            ),
            (dictgen::InsensitiveStr::Ascii("protectes"), &["protects"]),
            (dictgen::InsensitiveStr::Ascii("schdeule"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("capabilitieis"),
                &["capabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("oeprand"), &["operand"]),
            (dictgen::InsensitiveStr::Ascii("trimmin"), &["trimming"]),
            (
                dictgen::InsensitiveStr::Ascii("recrusivelly"),
                &["recursively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassociates"),
                &["disassociates"],
            ),
            (dictgen::InsensitiveStr::Ascii("selectes"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("guarneteed"),
                &["guaranteed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unanimoulsy"),
                &["unanimous"],
            ),
            (dictgen::InsensitiveStr::Ascii("watned"), &["wanted"]),
            (dictgen::InsensitiveStr::Ascii("thhis"), &["this"]),
            (
                dictgen::InsensitiveStr::Ascii("combinatations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simles"),
                &["smiles", "similes"],
            ),
            (dictgen::InsensitiveStr::Ascii("garentees"), &["guarantees"]),
            (dictgen::InsensitiveStr::Ascii("adjasent"), &["adjacent"]),
            (
                dictgen::InsensitiveStr::Ascii("margenalized"),
                &["marginalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("manaze"), &["mayonnaise"]),
            (dictgen::InsensitiveStr::Ascii("reguested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("dirves"), &["drives"]),
            (
                dictgen::InsensitiveStr::Ascii("compatbile"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuting"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("dynically"),
                &["dynamically"],
            ),
            (dictgen::InsensitiveStr::Ascii("codition"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("tehm"), &["them"]),
            (
                dictgen::InsensitiveStr::Ascii("compulisve"),
                &["compulsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcional"), &["functional"]),
            (
                dictgen::InsensitiveStr::Ascii("orangatangs"),
                &["orangutangs", "orangutans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspuported"),
                &["unsupported"],
            ),
            (dictgen::InsensitiveStr::Ascii("autority"), &["authority"]),
            (
                dictgen::InsensitiveStr::Ascii("masterbates"),
                &["masturbates"],
            ),
            (dictgen::InsensitiveStr::Ascii("disperive"), &["dispersive"]),
            (
                dictgen::InsensitiveStr::Ascii("exercizing"),
                &["exercising", "exorcizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("adbominal"), &["abdominal"]),
            (
                dictgen::InsensitiveStr::Ascii("vindictave"),
                &["vindictive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determinisim"),
                &["determinism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("princesess"),
                &["princesses"],
            ),
            (dictgen::InsensitiveStr::Ascii("strotage"), &["storage"]),
            (dictgen::InsensitiveStr::Ascii("solveable"), &["solvable"]),
            (
                dictgen::InsensitiveStr::Ascii("difficults"),
                &["difficulties", "difficult"],
            ),
            (dictgen::InsensitiveStr::Ascii("affirmates"), &["affirms"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sucsessors"),
                &["successors"],
            ),
            (dictgen::InsensitiveStr::Ascii("expandrs"), &["expanders"]),
            (
                dictgen::InsensitiveStr::Ascii("cicle"),
                &["cycle", "circle", "icicle"],
            ),
            (dictgen::InsensitiveStr::Ascii("inquirey"), &["inquiry"]),
            (dictgen::InsensitiveStr::Ascii("coerscing"), &["coercing"]),
            (dictgen::InsensitiveStr::Ascii("engien"), &["engine"]),
            (
                dictgen::InsensitiveStr::Ascii("initilalizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("estiamtes"), &["estimates"]),
            (
                dictgen::InsensitiveStr::Ascii("dermining"),
                &["determining"],
            ),
            (dictgen::InsensitiveStr::Ascii("asert"), &["assert"]),
            (dictgen::InsensitiveStr::Ascii("apihelion"), &["aphelion"]),
            (dictgen::InsensitiveStr::Ascii("dischages"), &["discharges"]),
            (dictgen::InsensitiveStr::Ascii("protrayed"), &["portrayed"]),
            (dictgen::InsensitiveStr::Ascii("furiosuly"), &["furiously"]),
            (dictgen::InsensitiveStr::Ascii("surpport"), &["support"]),
            (dictgen::InsensitiveStr::Ascii("frobidden"), &["forbidden"]),
            (dictgen::InsensitiveStr::Ascii("vailid"), &["valid"]),
            (dictgen::InsensitiveStr::Ascii("directove"), &["directive"]),
            (
                dictgen::InsensitiveStr::Ascii("explanaitons"),
                &["explanations"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancestoral"), &["ancestral"]),
            (
                dictgen::InsensitiveStr::Ascii("covnerters"),
                &["converters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fermentating"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("insultes"), &["insults"]),
            (
                dictgen::InsensitiveStr::Ascii("relie"),
                &["rely", "relies", "really", "relief"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visualisaton"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("informis"), &["informs"]),
            (
                dictgen::InsensitiveStr::Ascii("unitialises"),
                &["uninitialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomrances"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("sphereos"), &["spheres"]),
            (dictgen::InsensitiveStr::Ascii("compelxes"), &["complexes"]),
            (dictgen::InsensitiveStr::Ascii("tuhmb"), &["thumb"]),
            (dictgen::InsensitiveStr::Ascii("reuturn"), &["return"]),
            (
                dictgen::InsensitiveStr::Ascii("elliminating"),
                &["eliminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cteater"), &["creator"]),
            (
                dictgen::InsensitiveStr::Ascii("execised"),
                &["excised", "exercised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recongition"),
                &["recognition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("srolling"),
                &["scrolling", "strolling"],
            ),
            (dictgen::InsensitiveStr::Ascii("commitees"), &["committees"]),
            (dictgen::InsensitiveStr::Ascii("composiste"), &["composite"]),
            (dictgen::InsensitiveStr::Ascii("cehcked"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("disconnec"), &["disconnect"]),
            (dictgen::InsensitiveStr::Ascii("guarnte"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("knwoledgeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("opreated"), &["operated"]),
            (dictgen::InsensitiveStr::Ascii("customns"), &["customs"]),
            (dictgen::InsensitiveStr::Ascii("repsond"), &["respond"]),
            (
                dictgen::InsensitiveStr::Ascii("corregated"),
                &["correlated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exten"),
                &["extent", "extend"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessiated"),
                &["necessitated"],
            ),
            (dictgen::InsensitiveStr::Ascii("ergnomics"), &["ergonomics"]),
            (dictgen::InsensitiveStr::Ascii("scalled"), &["scaled"]),
            (
                dictgen::InsensitiveStr::Ascii("osbservable"),
                &["observable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specatcular"),
                &["spectacular"],
            ),
            (dictgen::InsensitiveStr::Ascii("interst"), &["interest"]),
            (dictgen::InsensitiveStr::Ascii("capcity"), &["capacity"]),
            (dictgen::InsensitiveStr::Ascii("marinens"), &["marines"]),
            (dictgen::InsensitiveStr::Ascii("repported"), &["reported"]),
            (dictgen::InsensitiveStr::Ascii("continure"), &["continue"]),
            (
                dictgen::InsensitiveStr::Ascii("asssessment"),
                &["assessment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplacting"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("diabets"), &["diabetes"]),
            (
                dictgen::InsensitiveStr::Ascii("provisoning"),
                &["provisioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrualized"),
                &["industrialized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unpriviliged"),
                &["unprivileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("refeeres"), &["referees"]),
            (dictgen::InsensitiveStr::Ascii("intiail"), &["initial"]),
            (
                dictgen::InsensitiveStr::Ascii("cheezeburgers"),
                &["cheeseburgers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficitied"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("symtoms"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("otions"), &["options"]),
            (dictgen::InsensitiveStr::Ascii("downloaad"), &["download"]),
            (
                dictgen::InsensitiveStr::Ascii("interploate"),
                &["interpolate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("denigratin"),
                &["denigrating", "denigration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accummulates"),
                &["accumulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nastershum"),
                &["nasturtium"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacaling"), &["scaling"]),
            (dictgen::InsensitiveStr::Ascii("impreves"), &["improves"]),
            (dictgen::InsensitiveStr::Ascii("repackge"), &["repackage"]),
            (
                dictgen::InsensitiveStr::Ascii("standaradized"),
                &["standardized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("communicaton"),
                &["communication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adverstising"),
                &["advertising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seperatist"),
                &["separatist"],
            ),
            (dictgen::InsensitiveStr::Ascii("amongs"), &["among"]),
            (
                dictgen::InsensitiveStr::Ascii("accessbile"),
                &["accessible"],
            ),
            (dictgen::InsensitiveStr::Ascii("genereator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("collpasible"),
                &["collapsible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inappropriae"),
                &["inappropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacuosly"), &["vacuously"]),
            (dictgen::InsensitiveStr::Ascii("centuty"), &["century"]),
            (
                dictgen::InsensitiveStr::Ascii("engaving"),
                &["engraving", "engaging"],
            ),
            (dictgen::InsensitiveStr::Ascii("lepracan"), &["leprechaun"]),
            (dictgen::InsensitiveStr::Ascii("folloeing"), &["following"]),
            (dictgen::InsensitiveStr::Ascii("paching"), &["patching"]),
            (
                dictgen::InsensitiveStr::Ascii("unconditinal"),
                &["unconditional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("providse"),
                &["provides", "provide"],
            ),
            (dictgen::InsensitiveStr::Ascii("callbaclk"), &["callback"]),
            (
                dictgen::InsensitiveStr::Ascii("visuaisation"),
                &["visualisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloster"), &["bolster"]),
            (dictgen::InsensitiveStr::Ascii("uppwards"), &["upwards"]),
            (dictgen::InsensitiveStr::Ascii("degister"), &["deregister"]),
            (
                dictgen::InsensitiveStr::Ascii("maintenaing"),
                &["maintaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("resered"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("genuienly"), &["genuinely"]),
            (dictgen::InsensitiveStr::Ascii("corousel"), &["carousel"]),
            (
                dictgen::InsensitiveStr::Ascii("transicion"),
                &["transition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availavbility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("multilply"), &["multiply"]),
            (
                dictgen::InsensitiveStr::Ascii("attenutation"),
                &["attenuation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("poisitioning"),
                &["positioning"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notfication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("uppper"), &["upper"]),
            (
                dictgen::InsensitiveStr::Ascii("experimens"),
                &["experiments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitiraly"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("produktions"),
                &["productions"],
            ),
            (dictgen::InsensitiveStr::Ascii("uncommong"), &["uncommon"]),
            (
                dictgen::InsensitiveStr::Ascii("excewptions"),
                &["exceptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("relesased"), &["released"]),
            (dictgen::InsensitiveStr::Ascii("hagas"), &["haggis"]),
            (dictgen::InsensitiveStr::Ascii("defaulkts"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("initailizations"),
                &["initializations"],
            ),
            (dictgen::InsensitiveStr::Ascii("evloving"), &["evolving"]),
            (
                dictgen::InsensitiveStr::Ascii("agggregates"),
                &["aggregates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyitically"),
                &["analytically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ellemental"), &["elemental"]),
            (
                dictgen::InsensitiveStr::Ascii("voilations"),
                &["violations"],
            ),
            (dictgen::InsensitiveStr::Ascii("signleton"), &["singleton"]),
            (
                dictgen::InsensitiveStr::Ascii("internall"),
                &["internal", "internally"],
            ),
            (dictgen::InsensitiveStr::Ascii("reulting"), &["resulting"]),
            (dictgen::InsensitiveStr::Ascii("chechkers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("mimicing"), &["mimicking"]),
            (dictgen::InsensitiveStr::Ascii("ststic"), &["static"]),
            (
                dictgen::InsensitiveStr::Ascii("amelioraters"),
                &["ameliorators", "ameliorates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthogonnal"),
                &["orthogonal"],
            ),
            (dictgen::InsensitiveStr::Ascii("bersekrer"), &["berserker"]),
            (dictgen::InsensitiveStr::Ascii("exraction"), &["extraction"]),
            (dictgen::InsensitiveStr::Ascii("obviusly"), &["obviously"]),
            (dictgen::InsensitiveStr::Ascii("webbooks"), &["webhooks"]),
            (
                dictgen::InsensitiveStr::Ascii("intiallizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("absoute"), &["absolute"]),
            (dictgen::InsensitiveStr::Ascii("fially"), &["finally"]),
            (dictgen::InsensitiveStr::Ascii("euopean"), &["european"]),
            (
                dictgen::InsensitiveStr::Ascii("speficiallier"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mecahinsms"),
                &["mechanisms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("possbile"),
                &["possible", "possibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decremeants"),
                &["decrements"],
            ),
            (dictgen::InsensitiveStr::Ascii("preseving"), &["preserving"]),
            (dictgen::InsensitiveStr::Ascii("perferctly"), &["perfectly"]),
            (
                dictgen::InsensitiveStr::Ascii("determistic"),
                &["deterministic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unapplicable"),
                &["inapplicable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfomranses"),
                &["performances"],
            ),
            (dictgen::InsensitiveStr::Ascii("grpahing"), &["graphing"]),
            (dictgen::InsensitiveStr::Ascii("automattic"), &["automatic"]),
            (
                dictgen::InsensitiveStr::Ascii("paremeterize"),
                &["parameterize"],
            ),
            (dictgen::InsensitiveStr::Ascii("enforeing"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("microtrasnactions"),
                &["microtransactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("endpiont"), &["endpoint"]),
            (dictgen::InsensitiveStr::Ascii("diaginal"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("provsions"), &["provisions"]),
            (dictgen::InsensitiveStr::Ascii("comune"), &["commune"]),
            (
                dictgen::InsensitiveStr::Ascii("furthurmore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contruction"),
                &["construction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acelerometers"),
                &["accelerometers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complited"),
                &["completed", "compiled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inferiorty"),
                &["inferiority"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aborigonal"),
                &["aboriginal"],
            ),
            (dictgen::InsensitiveStr::Ascii("biogted"), &["bigoted"]),
            (dictgen::InsensitiveStr::Ascii("chechsum"), &["checksum"]),
            (
                dictgen::InsensitiveStr::Ascii("uminportant"),
                &["unimportant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amendmenters"),
                &["amendments"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacumed"), &["vacuumed"]),
            (dictgen::InsensitiveStr::Ascii("emiiter"), &["emitter"]),
            (dictgen::InsensitiveStr::Ascii("posseses"), &["possesses"]),
            (
                dictgen::InsensitiveStr::Ascii("instriction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exponentionally"),
                &["exponentially"],
            ),
            (dictgen::InsensitiveStr::Ascii("degraged"), &["degrade"]),
            (dictgen::InsensitiveStr::Ascii("subptimal"), &["suboptimal"]),
            (dictgen::InsensitiveStr::Ascii("switchign"), &["switching"]),
            (dictgen::InsensitiveStr::Ascii("snadstorm"), &["sandstorm"]),
            (dictgen::InsensitiveStr::Ascii("molestarte"), &["molester"]),
            (
                dictgen::InsensitiveStr::Ascii("parenthesees"),
                &["parentheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misquitoes"),
                &["mosquitoes"],
            ),
            (dictgen::InsensitiveStr::Ascii("charakter"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("splels"),
                &["spells", "spiels"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("zeores"),
                &["zeros", "zeroes"],
            ),
            (dictgen::InsensitiveStr::Ascii("agruing"), &["arguing"]),
            (dictgen::InsensitiveStr::Ascii("showerd"), &["showered"]),
            (dictgen::InsensitiveStr::Ascii("travere"), &["traverse"]),
            (dictgen::InsensitiveStr::Ascii("eit"), &["it"]),
            (dictgen::InsensitiveStr::Ascii("embezelled"), &["embezzled"]),
            (
                dictgen::InsensitiveStr::Ascii("pixelx"),
                &["pixels", "pixel"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controlller"),
                &["controller"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misterious"),
                &["mysterious"],
            ),
            (dictgen::InsensitiveStr::Ascii("penguis"), &["penguins"]),
            (dictgen::InsensitiveStr::Ascii("aviliable"), &["available"]),
            (
                dictgen::InsensitiveStr::Ascii("manufacturier"),
                &["manufacturer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("restauranteurs"),
                &["restaurateurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("delievry"), &["delivery"]),
            (dictgen::InsensitiveStr::Ascii("interits"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("paied"), &["paid", "paired"]),
            (
                dictgen::InsensitiveStr::Ascii("definitifely"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("ressizes"), &["resizes"]),
            (dictgen::InsensitiveStr::Ascii("stablilize"), &["stabilize"]),
            (dictgen::InsensitiveStr::Ascii("vietual"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("trnasmissions"),
                &["transmissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pleaseure"), &["pleasure"]),
            (
                dictgen::InsensitiveStr::Ascii("winnter"),
                &["winter", "winner"],
            ),
            (dictgen::InsensitiveStr::Ascii("componts"), &["compost"]),
            (dictgen::InsensitiveStr::Ascii("charcters"), &["characters"]),
            (dictgen::InsensitiveStr::Ascii("regitry"), &["registry"]),
            (dictgen::InsensitiveStr::Ascii("strangets"), &["strangest"]),
            (
                dictgen::InsensitiveStr::Ascii("perpsective"),
                &["perspective"],
            ),
            (dictgen::InsensitiveStr::Ascii("tentalces"), &["tentacles"]),
            (
                dictgen::InsensitiveStr::Ascii("stran"),
                &["strand", "strain"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstntial"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("bloddy"), &["bloody"]),
            (dictgen::InsensitiveStr::Ascii("disance"), &["distance"]),
            (
                dictgen::InsensitiveStr::Ascii("suppporters"),
                &["supporters"],
            ),
            (dictgen::InsensitiveStr::Ascii("romanica"), &["romania"]),
            (dictgen::InsensitiveStr::Ascii("opbjective"), &["objective"]),
            (
                dictgen::InsensitiveStr::Ascii("overstreching"),
                &["overstretching"],
            ),
            (dictgen::InsensitiveStr::Ascii("packhage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("beauracratic"),
                &["bureaucratic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsistenly"),
                &["inconsistency", "inconsistently"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepres"), &["spheres"]),
            (
                dictgen::InsensitiveStr::Ascii("cheeckboxes"),
                &["checkboxes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("degress"),
                &["degrees", "digress"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credidential"),
                &["credential"],
            ),
            (dictgen::InsensitiveStr::Ascii("wedeged"), &["wedged"]),
            (dictgen::InsensitiveStr::Ascii("anywyas"), &["anyways"]),
            (dictgen::InsensitiveStr::Ascii("draview"), &["drawview"]),
            (
                dictgen::InsensitiveStr::Ascii("concatanations"),
                &["concatenations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supplmental"),
                &["supplemental"],
            ),
            (dictgen::InsensitiveStr::Ascii("backslase"), &["backslash"]),
            (
                dictgen::InsensitiveStr::Ascii("assiciating"),
                &["associating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decyphered"),
                &["deciphered"],
            ),
            (dictgen::InsensitiveStr::Ascii("penalities"), &["penalties"]),
            (
                dictgen::InsensitiveStr::Ascii("summerizing"),
                &["summarizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("dalegate"), &["delegate"]),
            (dictgen::InsensitiveStr::Ascii("arrivin"), &["arriving"]),
            (dictgen::InsensitiveStr::Ascii("strrings"), &["strings"]),
            (dictgen::InsensitiveStr::Ascii("inerits"), &["inherits"]),
            (dictgen::InsensitiveStr::Ascii("excersice"), &["exercise"]),
            (dictgen::InsensitiveStr::Ascii("exsamples"), &["examples"]),
            (dictgen::InsensitiveStr::Ascii("amongts"), &["amongst"]),
            (
                dictgen::InsensitiveStr::Ascii("arrrangement"),
                &["arrangement"],
            ),
            (dictgen::InsensitiveStr::Ascii("fhurter"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("regisration"),
                &["registration"],
            ),
            (dictgen::InsensitiveStr::Ascii("importamt"), &["important"]),
            (
                dictgen::InsensitiveStr::Ascii("republcian"),
                &["republican"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calssifications"),
                &["classifications"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanves"), &["advances"]),
            (
                dictgen::InsensitiveStr::Ascii("applictaions"),
                &["applications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nomal"),
                &["normal", "nominal", "nomad", "nodal"],
            ),
            (dictgen::InsensitiveStr::Ascii("reigstry"), &["registry"]),
            (
                dictgen::InsensitiveStr::Ascii("prealocate"),
                &["preallocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("euhporia"), &["euphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("availbaility"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("arametric"), &["parametric"]),
            (dictgen::InsensitiveStr::Ascii("trubinal"), &["tribunal"]),
            (
                dictgen::InsensitiveStr::Ascii("empathethic"),
                &["empathetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innerstellar"),
                &["interstellar"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniitializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("delevop"), &["develop"]),
            (dictgen::InsensitiveStr::Ascii("validador"), &["validator"]),
            (dictgen::InsensitiveStr::Ascii("assissts"), &["assists"]),
            (
                dictgen::InsensitiveStr::Ascii("requeriments"),
                &["requirements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("collaterol"),
                &["collateral"],
            ),
            (dictgen::InsensitiveStr::Ascii("througput"), &["throughput"]),
            (
                dictgen::InsensitiveStr::Ascii("initiaitives"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("ignorin"), &["ignoring"]),
            (
                dictgen::InsensitiveStr::Ascii("crytpocurrencies"),
                &["cryptocurrencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("equested"), &["requested"]),
            (dictgen::InsensitiveStr::Ascii("colombina"), &["colombia"]),
            (
                dictgen::InsensitiveStr::Ascii("distuingishing"),
                &["distinguishing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differientation"),
                &["differentiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immedeiately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("qualifiy"), &["qualify"]),
            (dictgen::InsensitiveStr::Ascii("modidfy"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("traversin"), &["traversing"]),
            (dictgen::InsensitiveStr::Ascii("agrieved"), &["aggrieved"]),
            (dictgen::InsensitiveStr::Ascii("vacinity"), &["vicinity"]),
            (
                dictgen::InsensitiveStr::Ascii("embedidngs"),
                &["embeddings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termiations"),
                &["terminations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corispondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overwhelmigly"),
                &["overwhelmingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("amealearating"),
                &["ameliorating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dravadian"), &["dravidian"]),
            (dictgen::InsensitiveStr::Ascii("abstraced"), &["abstracted"]),
            (dictgen::InsensitiveStr::Ascii("idetified"), &["identified"]),
            (
                dictgen::InsensitiveStr::Ascii("recommonds"),
                &["recommends"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakcages"), &["packages"]),
            (
                dictgen::InsensitiveStr::Ascii("incomptetent"),
                &["incompetent"],
            ),
            (dictgen::InsensitiveStr::Ascii("floruide"), &["fluoride"]),
            (dictgen::InsensitiveStr::Ascii("reguar"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("proided"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("downgradingn"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkkpoints"),
                &["checkpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("incluudes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("authethicated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatabile"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hirearchies"),
                &["hierarchies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genearting"),
                &["generating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intoxicatin"),
                &["intoxication"],
            ),
            (dictgen::InsensitiveStr::Ascii("desriable"), &["desirable"]),
            (dictgen::InsensitiveStr::Ascii("penatly"), &["penalty"]),
            (
                dictgen::InsensitiveStr::Ascii("civilziation"),
                &["civilizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("associatating"),
                &["associating"],
            ),
            (dictgen::InsensitiveStr::Ascii("disposle"), &["dispose"]),
            (
                dictgen::InsensitiveStr::Ascii("enviorenments"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternativley"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("ammount"), &["amount"]),
            (dictgen::InsensitiveStr::Ascii("ovrrides"), &["overrides"]),
            (
                dictgen::InsensitiveStr::Ascii("processeurs"),
                &["processes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expecetation"),
                &["expectation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgroounds"),
                &["backgrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("curremts"), &["currents"]),
            (
                dictgen::InsensitiveStr::Ascii("preisthood"),
                &["priesthood"],
            ),
            (dictgen::InsensitiveStr::Ascii("editter"), &["editor"]),
            (
                dictgen::InsensitiveStr::Ascii("descripotrs"),
                &["descriptors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitlly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("adament"), &["adamant"]),
            (dictgen::InsensitiveStr::Ascii("projets"), &["projects"]),
            (dictgen::InsensitiveStr::Ascii("crystas"), &["crystals"]),
            (dictgen::InsensitiveStr::Ascii("exlaining"), &["explaining"]),
            (
                dictgen::InsensitiveStr::Ascii("benchmkared"),
                &["benchmarked"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("higlighted"),
                &["highlighted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pharmaseudical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("certication"),
                &["certification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("actications"),
                &["activations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disappering"),
                &["disappearing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("individaully"),
                &["individually"],
            ),
            (dictgen::InsensitiveStr::Ascii("incognitio"), &["incognito"]),
            (dictgen::InsensitiveStr::Ascii("warninig"), &["warning"]),
            (
                dictgen::InsensitiveStr::Ascii("reduses"),
                &["reduces", "refuses", "reuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commnds"),
                &["commands", "commends"],
            ),
            (dictgen::InsensitiveStr::Ascii("againnst"), &["against"]),
            (
                dictgen::InsensitiveStr::Ascii("essentuals"),
                &["essentials"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasnparency"),
                &["transparency"],
            ),
            (dictgen::InsensitiveStr::Ascii("decress"), &["decrees"]),
            (dictgen::InsensitiveStr::Ascii("silibuses"), &["syllabuses"]),
            (
                dictgen::InsensitiveStr::Ascii("unforunate"),
                &["unfortunate"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifed"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("cusomiser"), &["customiser"]),
            (
                dictgen::InsensitiveStr::Ascii("disatvantages"),
                &["disadvantages"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("administerd"),
                &["administered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("egotisticle"),
                &["egotistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("flourine"), &["fluorine"]),
            (dictgen::InsensitiveStr::Ascii("ambluance"), &["ambulance"]),
            (dictgen::InsensitiveStr::Ascii("shspe"), &["shape"]),
            (dictgen::InsensitiveStr::Ascii("undefied"), &["undefined"]),
            (dictgen::InsensitiveStr::Ascii("decopose"), &["decompose"]),
            (dictgen::InsensitiveStr::Ascii("improrters"), &["importers"]),
            (dictgen::InsensitiveStr::Ascii("witnesss"), &["witnesses"]),
            (dictgen::InsensitiveStr::Ascii("behavors"), &["behaviors"]),
            (
                dictgen::InsensitiveStr::Ascii("veiwpoints"),
                &["viewpoints"],
            ),
            (dictgen::InsensitiveStr::Ascii("autopsec"), &["autospec"]),
            (
                dictgen::InsensitiveStr::Ascii("preponderence"),
                &["preponderance"],
            ),
            (dictgen::InsensitiveStr::Ascii("nighly"), &["nightly"]),
            (
                dictgen::InsensitiveStr::Ascii("neuorscience"),
                &["neuroscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fermentaion"),
                &["fermentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("worrries"), &["worries"]),
            (
                dictgen::InsensitiveStr::Ascii("aninations"),
                &["animations"],
            ),
            (dictgen::InsensitiveStr::Ascii("strutured"), &["structured"]),
            (
                dictgen::InsensitiveStr::Ascii("endinanness"),
                &["endianness"],
            ),
            (dictgen::InsensitiveStr::Ascii("vectorrs"), &["vectors"]),
            (dictgen::InsensitiveStr::Ascii("jounaled"), &["journaled"]),
            (dictgen::InsensitiveStr::Ascii("privilged"), &["privileged"]),
            (
                dictgen::InsensitiveStr::Ascii("chandaleer"),
                &["chandelier"],
            ),
            (dictgen::InsensitiveStr::Ascii("continuin"), &["continuing"]),
            (dictgen::InsensitiveStr::Ascii("pruchases"), &["purchases"]),
            (
                dictgen::InsensitiveStr::Ascii("heartsthone"),
                &["hearthstone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("configired"),
                &["configured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("austensible"),
                &["ostensible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desination"),
                &["destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acadmically"),
                &["academically"],
            ),
            (dictgen::InsensitiveStr::Ascii("pancaks"), &["pancakes"]),
            (
                dictgen::InsensitiveStr::Ascii("cuztomizable"),
                &["customizable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temporrily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consipracies"),
                &["conspiracies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasncripts"),
                &["transcripts"],
            ),
            (dictgen::InsensitiveStr::Ascii("tihkn"), &["think"]),
            (
                dictgen::InsensitiveStr::Ascii("comversations"),
                &["conversations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defrentiating"),
                &["differentiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("pinapple"), &["pineapple"]),
            (dictgen::InsensitiveStr::Ascii("facorite"), &["favorite"]),
            (
                dictgen::InsensitiveStr::Ascii("billborads"),
                &["billboards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("credidentials"),
                &["credentials"],
            ),
            (dictgen::InsensitiveStr::Ascii("althetes"), &["athletes"]),
            (
                dictgen::InsensitiveStr::Ascii("satistying"),
                &["satisfying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fabircatings"),
                &["fabricating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cristallise"),
                &["crystallise"],
            ),
            (dictgen::InsensitiveStr::Ascii("desogning"), &["designing"]),
            (
                dictgen::InsensitiveStr::Ascii("presse"),
                &["pressed", "press"],
            ),
            (dictgen::InsensitiveStr::Ascii("eletrons"), &["electrons"]),
            (dictgen::InsensitiveStr::Ascii("northeren"), &["northern"]),
            (dictgen::InsensitiveStr::Ascii("inclode"), &["include"]),
            (
                dictgen::InsensitiveStr::Ascii("underwager"),
                &["underwater"],
            ),
            (dictgen::InsensitiveStr::Ascii("instanct"), &["instance"]),
            (dictgen::InsensitiveStr::Ascii("provods"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("aparantly"), &["apparently"]),
            (
                dictgen::InsensitiveStr::Ascii("transfmromation"),
                &["transformation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pagckage"), &["package"]),
            (
                dictgen::InsensitiveStr::Ascii("aritmethic"),
                &["arithmetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("tahnks"), &["thanks"]),
            (dictgen::InsensitiveStr::Ascii("inferecene"), &["inference"]),
            (
                dictgen::InsensitiveStr::Ascii("cosntructing"),
                &["constructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constructore"),
                &["constructor"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completenes"),
                &["completes"],
            ),
            (dictgen::InsensitiveStr::Ascii("previoous"), &["previous"]),
            (
                dictgen::InsensitiveStr::Ascii("peasefully"),
                &["peacefully"],
            ),
            (dictgen::InsensitiveStr::Ascii("advanace"), &["advance"]),
            (
                dictgen::InsensitiveStr::Ascii("medicority"),
                &["mediocrity"],
            ),
            (dictgen::InsensitiveStr::Ascii("tomatoe"), &["tomato"]),
            (
                dictgen::InsensitiveStr::Ascii("appriproate"),
                &["appropriate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abandonding"),
                &["abandoning"],
            ),
            (dictgen::InsensitiveStr::Ascii("inialise"), &["initialise"]),
            (
                dictgen::InsensitiveStr::Ascii("comperatively"),
                &["comparatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replecated"),
                &["replicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("programm"),
                &["program", "programme"],
            ),
            (dictgen::InsensitiveStr::Ascii("misteek"), &["mystique"]),
            (dictgen::InsensitiveStr::Ascii("sarimonys"), &["ceremonies"]),
            (
                dictgen::InsensitiveStr::Ascii("downloaing"),
                &["downloading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overvlocked"),
                &["overclocked"],
            ),
            (dictgen::InsensitiveStr::Ascii("becoz"), &["because"]),
            (
                dictgen::InsensitiveStr::Ascii("complext"),
                &["complex", "complexity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlefornt"),
                &["battlefront"],
            ),
            (dictgen::InsensitiveStr::Ascii("probblem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("follwws"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("segemntations"),
                &["segmentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("priotise"), &["prioritise"]),
            (
                dictgen::InsensitiveStr::Ascii("principly"),
                &["principally"],
            ),
            (dictgen::InsensitiveStr::Ascii("remoive"), &["remove"]),
            (dictgen::InsensitiveStr::Ascii("indianos"), &["indians"]),
            (
                dictgen::InsensitiveStr::Ascii("imperialsm"),
                &["imperialism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correcteion"),
                &["correction"],
            ),
            (dictgen::InsensitiveStr::Ascii("mentionned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("stragedy"),
                &["strategy", "tragedy"],
            ),
            (dictgen::InsensitiveStr::Ascii("antennea"), &["antenna"]),
            (dictgen::InsensitiveStr::Ascii("answharing"), &["answering"]),
            (
                dictgen::InsensitiveStr::Ascii("challengeing"),
                &["challenging"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrumenet"),
                &["instrument"],
            ),
            (dictgen::InsensitiveStr::Ascii("overbaord"), &["overboard"]),
            (
                dictgen::InsensitiveStr::Ascii("decoratieve"),
                &["decorative"],
            ),
            (dictgen::InsensitiveStr::Ascii("attemented"), &["attempted"]),
            (
                dictgen::InsensitiveStr::Ascii("bastractly"),
                &["abstractly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corresonds"),
                &["corresponds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("derection"),
                &["direction", "detection", "erection"],
            ),
            (dictgen::InsensitiveStr::Ascii("acn"), &["can", "acne"]),
            (dictgen::InsensitiveStr::Ascii("implments"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("certificste"),
                &["certificate"],
            ),
            (dictgen::InsensitiveStr::Ascii("descrybed"), &["described"]),
            (
                dictgen::InsensitiveStr::Ascii("sustainabily"),
                &["sustainability"],
            ),
            (dictgen::InsensitiveStr::Ascii("literatue"), &["literate"]),
            (
                dictgen::InsensitiveStr::Ascii("defninition"),
                &["definition"],
            ),
            (dictgen::InsensitiveStr::Ascii("pathame"), &["pathname"]),
            (dictgen::InsensitiveStr::Ascii("consenus"), &["consensus"]),
            (
                dictgen::InsensitiveStr::Ascii("neighberhooding"),
                &["neighboring"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("completness"),
                &["completeness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unserstood"),
                &["understood"],
            ),
            (dictgen::InsensitiveStr::Ascii("shoing"), &["showing"]),
            (dictgen::InsensitiveStr::Ascii("tranmit"), &["transmit"]),
            (dictgen::InsensitiveStr::Ascii("attribted"), &["attributed"]),
            (dictgen::InsensitiveStr::Ascii("reactangle"), &["rectangle"]),
            (dictgen::InsensitiveStr::Ascii("cardinalis"), &["cardinals"]),
            (dictgen::InsensitiveStr::Ascii("balancin"), &["balancing"]),
            (dictgen::InsensitiveStr::Ascii("dailog"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("expectatons"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturggles"), &["struggles"]),
            (
                dictgen::InsensitiveStr::Ascii("opporunities"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("closee"),
                &["closed", "close"],
            ),
            (dictgen::InsensitiveStr::Ascii("freestlye"), &["freestyle"]),
            (
                dictgen::InsensitiveStr::Ascii("corrsepondent"),
                &["correspondent"],
            ),
            (dictgen::InsensitiveStr::Ascii("lightres"), &["lighters"]),
            (
                dictgen::InsensitiveStr::Ascii("calcuration"),
                &["calculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("apdates"), &["updates"]),
            (dictgen::InsensitiveStr::Ascii("endevered"), &["endeavored"]),
            (dictgen::InsensitiveStr::Ascii("negativy"), &["negativity"]),
            (
                dictgen::InsensitiveStr::Ascii("supplemetal"),
                &["supplemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singlularly"),
                &["singularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("dscribed"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("mysteris"), &["mysteries"]),
            (dictgen::InsensitiveStr::Ascii("reconet"), &["reconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("evauluating"),
                &["evaluating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("formitable"),
                &["formidable"],
            ),
            (dictgen::InsensitiveStr::Ascii("compaigns"), &["campaigns"]),
            (
                dictgen::InsensitiveStr::Ascii("administratior"),
                &["administrator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiate"),
                &["renegotiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("primitaves"),
                &["primitives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immideately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("desides"), &["decides"]),
            (dictgen::InsensitiveStr::Ascii("gamergat"), &["gamertag"]),
            (
                dictgen::InsensitiveStr::Ascii("permtuation"),
                &["permutation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomanded"),
                &["recommended"],
            ),
            (dictgen::InsensitiveStr::Ascii("beffer"), &["buffer"]),
            (dictgen::InsensitiveStr::Ascii("missuse"), &["misuse"]),
            (
                dictgen::InsensitiveStr::Ascii("managerment"),
                &["management"],
            ),
            (dictgen::InsensitiveStr::Ascii("nanosecod"), &["nanosecond"]),
            (dictgen::InsensitiveStr::Ascii("mechanincs"), &["mechanic"]),
            (
                dictgen::InsensitiveStr::Ascii("uncontrollablely"),
                &["uncontrollably"],
            ),
            (dictgen::InsensitiveStr::Ascii("cutomizes"), &["customizes"]),
            (
                dictgen::InsensitiveStr::Ascii("intimatley"),
                &["intimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("checkeboxes"),
                &["checkboxes"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimally"), &["optimally"]),
            (dictgen::InsensitiveStr::Ascii("secuenced"), &["sequenced"]),
            (
                dictgen::InsensitiveStr::Ascii("amohetamines"),
                &["amphetamines"],
            ),
            (dictgen::InsensitiveStr::Ascii("decisivie"), &["decisive"]),
            (
                dictgen::InsensitiveStr::Ascii("equired"),
                &["required", "enquired"],
            ),
            (dictgen::InsensitiveStr::Ascii("closig"), &["closing"]),
            (
                dictgen::InsensitiveStr::Ascii("nofitication"),
                &["notification"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleiving"), &["believing"]),
            (
                dictgen::InsensitiveStr::Ascii("subrountine"),
                &["subroutine"],
            ),
            (dictgen::InsensitiveStr::Ascii("supportes"), &["supports"]),
            (dictgen::InsensitiveStr::Ascii("requeum"), &["requiem"]),
            (
                dictgen::InsensitiveStr::Ascii("automicaly"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temprature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("variablesm"), &["variables"]),
            (dictgen::InsensitiveStr::Ascii("decisons"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("issused"), &["issued"]),
            (
                dictgen::InsensitiveStr::Ascii("preceeds"),
                &["precedes", "proceeds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("wedensdays"),
                &["wednesdays"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialsers"),
                &["initialisers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accomodiate"),
                &["accommodate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accommidates"),
                &["accommodates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recontructing"),
                &["reconstructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frequensies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("favoruite"), &["favorite"]),
            (dictgen::InsensitiveStr::Ascii("zealotus"), &["zealots"]),
            (dictgen::InsensitiveStr::Ascii("palettte"), &["palette"]),
            (
                dictgen::InsensitiveStr::Ascii("competitior"),
                &["competitor"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendired"), &["rendered"]),
            (
                dictgen::InsensitiveStr::Ascii("exis"),
                &["exist", "exists", "exit", "exits", "axis", "lexis", "exes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subsitututed"),
                &["substituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("exvent"), &["event"]),
            (dictgen::InsensitiveStr::Ascii("recongize"), &["recognize"]),
            (
                dictgen::InsensitiveStr::Ascii("discreprency"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("seqenced"), &["sequenced"]),
            (dictgen::InsensitiveStr::Ascii("promplty"), &["promptly"]),
            (dictgen::InsensitiveStr::Ascii("streamtrue"), &["streamer"]),
            (dictgen::InsensitiveStr::Ascii("diphtongs"), &["diphthongs"]),
            (
                dictgen::InsensitiveStr::Ascii("librabaries"),
                &["libraries"],
            ),
            (dictgen::InsensitiveStr::Ascii("consern"), &["concern"]),
            (dictgen::InsensitiveStr::Ascii("pairocheal"), &["parochial"]),
            (dictgen::InsensitiveStr::Ascii("threshod"), &["threshold"]),
            (dictgen::InsensitiveStr::Ascii("paramater"), &["parameter"]),
            (
                dictgen::InsensitiveStr::Ascii("videostreamming"),
                &["videostreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("chuncks"), &["chunks"]),
            (
                dictgen::InsensitiveStr::Ascii("handshakres"),
                &["handshakes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compililation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("imput"), &["input"]),
            (dictgen::InsensitiveStr::Ascii("regrding"), &["regarding"]),
            (
                dictgen::InsensitiveStr::Ascii("extracters"),
                &["extractors"],
            ),
            (dictgen::InsensitiveStr::Ascii("methjod"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("duplcation"),
                &["duplication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definitivelly"),
                &["definitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("cruciaal"), &["crucial"]),
            (
                dictgen::InsensitiveStr::Ascii("knoweldgeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("endcodings"), &["encodings"]),
            (dictgen::InsensitiveStr::Ascii("indiciates"), &["indicates"]),
            (
                dictgen::InsensitiveStr::Ascii("discrecion"),
                &["discretion"],
            ),
            (dictgen::InsensitiveStr::Ascii("contronls"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("trainngs"), &["training"]),
            (
                dictgen::InsensitiveStr::Ascii("artificiely"),
                &["artificially"],
            ),
            (dictgen::InsensitiveStr::Ascii("plausibel"), &["plausible"]),
            (
                dictgen::InsensitiveStr::Ascii("transsexaul"),
                &["transsexual"],
            ),
            (dictgen::InsensitiveStr::Ascii("silowet"), &["silhouette"]),
            (dictgen::InsensitiveStr::Ascii("shouldt"), &["shouldnt"]),
            (
                dictgen::InsensitiveStr::Ascii("ultimatley"),
                &["ultimately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("docmenetation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("nmbers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("apocaliptic"),
                &["apocalyptic"],
            ),
            (dictgen::InsensitiveStr::Ascii("modiffers"), &["modifiers"]),
            (dictgen::InsensitiveStr::Ascii("declar"), &["declare"]),
            (dictgen::InsensitiveStr::Ascii("tanget"), &["tangent"]),
            (dictgen::InsensitiveStr::Ascii("doccuments"), &["documents"]),
            (dictgen::InsensitiveStr::Ascii("refelcts"), &["reflects"]),
            (dictgen::InsensitiveStr::Ascii("pensies"), &["penises"]),
            (dictgen::InsensitiveStr::Ascii("executred"), &["executed"]),
            (dictgen::InsensitiveStr::Ascii("closesly"), &["closely"]),
            (
                dictgen::InsensitiveStr::Ascii("settelments"),
                &["settlements"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjancey"), &["adjacency"]),
            (
                dictgen::InsensitiveStr::Ascii("discalimer"),
                &["disclaimer"],
            ),
            (dictgen::InsensitiveStr::Ascii("inheret"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("formuala"), &["formula"]),
            (dictgen::InsensitiveStr::Ascii("isntalled"), &["installed"]),
            (
                dictgen::InsensitiveStr::Ascii("enhencments"),
                &["enhancements"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinnics"), &["cynics"]),
            (
                dictgen::InsensitiveStr::Ascii("materialisimo"),
                &["materialism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extraxtors"),
                &["extractors"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimimun"), &["minimum"]),
            (dictgen::InsensitiveStr::Ascii("loooking"), &["looking"]),
            (
                dictgen::InsensitiveStr::Ascii("terminiating"),
                &["terminating"],
            ),
            (dictgen::InsensitiveStr::Ascii("sortcuts"), &["shortcuts"]),
            (dictgen::InsensitiveStr::Ascii("arond"), &["around"]),
            (dictgen::InsensitiveStr::Ascii("marixsm"), &["marxism"]),
            (dictgen::InsensitiveStr::Ascii("faturing"), &["featuring"]),
            (dictgen::InsensitiveStr::Ascii("octect"), &["octet"]),
            (
                dictgen::InsensitiveStr::Ascii("transocdings"),
                &["transcodings"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separtating"),
                &["separating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("norhtwestern"),
                &["northwestern"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abovemtioned"),
                &["abovementioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effectivelly"),
                &["effectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("termiantes"),
                &["terminates"],
            ),
            (dictgen::InsensitiveStr::Ascii("affleunt"), &["affluent"]),
            (dictgen::InsensitiveStr::Ascii("stoprage"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("reparamterises"),
                &["reparameterises"],
            ),
            (dictgen::InsensitiveStr::Ascii("explods"), &["explodes"]),
            (dictgen::InsensitiveStr::Ascii("prevert"), &["pervert"]),
            (
                dictgen::InsensitiveStr::Ascii("descchedules"),
                &["deschedules"],
            ),
            (dictgen::InsensitiveStr::Ascii("extenision"), &["extension"]),
            (
                dictgen::InsensitiveStr::Ascii("resurrektion"),
                &["resurrection"],
            ),
            (dictgen::InsensitiveStr::Ascii("efficent"), &["efficient"]),
            (dictgen::InsensitiveStr::Ascii("herculeasy"), &["hercules"]),
            (dictgen::InsensitiveStr::Ascii("clasified"), &["classified"]),
            (dictgen::InsensitiveStr::Ascii("euivalent"), &["equivalent"]),
            (
                dictgen::InsensitiveStr::Ascii("intented"),
                &["intended", "indented"],
            ),
            (dictgen::InsensitiveStr::Ascii("extreem"), &["extreme"]),
            (
                dictgen::InsensitiveStr::Ascii("corresponsd"),
                &["corresponds", "correspond"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosidered"), &["considered"]),
            (dictgen::InsensitiveStr::Ascii("registes"), &["registers"]),
            (dictgen::InsensitiveStr::Ascii("medhods"), &["methods"]),
            (dictgen::InsensitiveStr::Ascii("pointss"), &["points"]),
            (dictgen::InsensitiveStr::Ascii("upddating"), &["updating"]),
            (
                dictgen::InsensitiveStr::Ascii("unconventianal"),
                &["unconventional"],
            ),
            (dictgen::InsensitiveStr::Ascii("falshing"), &["flashing"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubscirption"),
                &["unsubscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("transcorde"), &["transcode"]),
            (dictgen::InsensitiveStr::Ascii("stirrig"), &["stirring"]),
            (
                dictgen::InsensitiveStr::Ascii("reprodiced"),
                &["reproduced"],
            ),
            (dictgen::InsensitiveStr::Ascii("bastane"), &["bastante"]),
            (dictgen::InsensitiveStr::Ascii("furthere"), &["further"]),
            (
                dictgen::InsensitiveStr::Ascii("microprocessprs"),
                &["microprocessors"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redifinitions"),
                &["redefinitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("spalsher"), &["splasher"]),
            (dictgen::InsensitiveStr::Ascii("parnoia"), &["paranoia"]),
            (dictgen::InsensitiveStr::Ascii("diagnoal"), &["diagonal"]),
            (dictgen::InsensitiveStr::Ascii("scalin"), &["scaling"]),
            (dictgen::InsensitiveStr::Ascii("compteurs"), &["computers"]),
            (
                dictgen::InsensitiveStr::Ascii("overflowerd"),
                &["overflowed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacemnnt"),
                &["replacement"],
            ),
            (dictgen::InsensitiveStr::Ascii("therw"), &["there"]),
            (
                dictgen::InsensitiveStr::Ascii("ommisience"),
                &["omniscience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparistion"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbourhoors"),
                &["neighbours"],
            ),
            (dictgen::InsensitiveStr::Ascii("captivitiy"), &["captivity"]),
            (
                dictgen::InsensitiveStr::Ascii("approximitions"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfurtunately"),
                &["unfortunately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiailises"),
                &["initialises"],
            ),
            (dictgen::InsensitiveStr::Ascii("mutipart"), &["multipart"]),
            (dictgen::InsensitiveStr::Ascii("sterlie"), &["sterile"]),
            (dictgen::InsensitiveStr::Ascii("poossible"), &["possible"]),
            (dictgen::InsensitiveStr::Ascii("goosebums"), &["goosebumps"]),
            (
                dictgen::InsensitiveStr::Ascii("desactivated"),
                &["deactivated"],
            ),
            (dictgen::InsensitiveStr::Ascii("eqauls"), &["equals"]),
            (dictgen::InsensitiveStr::Ascii("udner"), &["under"]),
            (
                dictgen::InsensitiveStr::Ascii("homeogeneous"),
                &["homogeneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acknowldeged"),
                &["acknowledged"],
            ),
            (dictgen::InsensitiveStr::Ascii("casarole"), &["casserole"]),
            (dictgen::InsensitiveStr::Ascii("notticing"), &["noticing"]),
            (dictgen::InsensitiveStr::Ascii("exsted"), &["existed"]),
            (
                dictgen::InsensitiveStr::Ascii("aggrgating"),
                &["aggregating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provinence"),
                &["providence"],
            ),
            (dictgen::InsensitiveStr::Ascii("databsae"), &["database"]),
            (dictgen::InsensitiveStr::Ascii("listents"), &["listens"]),
            (
                dictgen::InsensitiveStr::Ascii("interferres"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initilization"),
                &["initialization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sysmatically"),
                &["systematically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimint"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomety"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("inlalid"), &["invalid"]),
            (
                dictgen::InsensitiveStr::Ascii("recalcution"),
                &["recalculation"],
            ),
            (dictgen::InsensitiveStr::Ascii("hanling"), &["handling"]),
            (dictgen::InsensitiveStr::Ascii("polyginal"), &["polygonal"]),
            (
                dictgen::InsensitiveStr::Ascii("phenonemon"),
                &["phenomenon"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comatibility"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crystalise"),
                &["crystallise"],
            ),
            (dictgen::InsensitiveStr::Ascii("heplers"), &["helpers"]),
            (dictgen::InsensitiveStr::Ascii("identied"), &["identified"]),
            (
                dictgen::InsensitiveStr::Ascii("visibliity"),
                &["visibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("mocule"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("suppy"), &["supply"]),
            (dictgen::InsensitiveStr::Ascii("cetrify"), &["certify"]),
            (dictgen::InsensitiveStr::Ascii("melboure"), &["melbourne"]),
            (
                dictgen::InsensitiveStr::Ascii("deserailized"),
                &["deserialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cecklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("internation"),
                &["international"],
            ),
            (dictgen::InsensitiveStr::Ascii("acronmy"), &["acronym"]),
            (
                dictgen::InsensitiveStr::Ascii("equivilents"),
                &["equivalents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exporation"),
                &["exploration", "expiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ristricted"),
                &["restricted"],
            ),
            (dictgen::InsensitiveStr::Ascii("hstory"), &["history"]),
            (dictgen::InsensitiveStr::Ascii("cameleon"), &["chameleon"]),
            (
                dictgen::InsensitiveStr::Ascii("interrputed"),
                &["interrupted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rittling"),
                &["rattling", "riddling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deactiveted"),
                &["deactivated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comines"),
                &["combines", "comings"],
            ),
            (dictgen::InsensitiveStr::Ascii("jaguras"), &["jaguars"]),
            (dictgen::InsensitiveStr::Ascii("vebrose"), &["verbose"]),
            (dictgen::InsensitiveStr::Ascii("sufficit"), &["sufficient"]),
            (dictgen::InsensitiveStr::Ascii("itentical"), &["identical"]),
            (dictgen::InsensitiveStr::Ascii("militat"), &["militant"]),
            (
                dictgen::InsensitiveStr::Ascii("paricipant"),
                &["participant"],
            ),
            (dictgen::InsensitiveStr::Ascii("stuty"), &["study"]),
            (dictgen::InsensitiveStr::Ascii("degnerate"), &["degenerate"]),
            (
                dictgen::InsensitiveStr::Ascii("apparences"),
                &["appearances"],
            ),
            (dictgen::InsensitiveStr::Ascii("theoery"), &["theory"]),
            (dictgen::InsensitiveStr::Ascii("revewer"), &["reviewer"]),
            (dictgen::InsensitiveStr::Ascii("inlcudes"), &["includes"]),
            (
                dictgen::InsensitiveStr::Ascii("dispalcement"),
                &["displacement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custumising"),
                &["customising"],
            ),
            (dictgen::InsensitiveStr::Ascii("nd"), &["and"]),
            (
                dictgen::InsensitiveStr::Ascii("termendous"),
                &["tremendous"],
            ),
            (dictgen::InsensitiveStr::Ascii("resuling"), &["resulting"]),
            (
                dictgen::InsensitiveStr::Ascii("circumstantional"),
                &["circumstantial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complerely"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("eaziest"), &["easiest"]),
            (
                dictgen::InsensitiveStr::Ascii("neighbeurhoor"),
                &["neighbour"],
            ),
            (dictgen::InsensitiveStr::Ascii("makr"), &["maker", "mark"]),
            (dictgen::InsensitiveStr::Ascii("perjery"), &["perjury"]),
            (
                dictgen::InsensitiveStr::Ascii("incomatible"),
                &["incompatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitsteam"), &["bitstream"]),
            (
                dictgen::InsensitiveStr::Ascii("datastrcutures"),
                &["datastructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("diginal"), &["diagonal"]),
            (
                dictgen::InsensitiveStr::Ascii("autheritative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commpilation"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("outskirst"), &["outskirts"]),
            (
                dictgen::InsensitiveStr::Ascii("applicato"),
                &["application", "applicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("insters"), &["inserts"]),
            (
                dictgen::InsensitiveStr::Ascii("tracability"),
                &["traceability"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkng"), &["checking"]),
            (dictgen::InsensitiveStr::Ascii("unselecgts"), &["unselects"]),
            (
                dictgen::InsensitiveStr::Ascii("underestamated"),
                &["underestimated"],
            ),
            (dictgen::InsensitiveStr::Ascii("detctions"), &["detections"]),
            (
                dictgen::InsensitiveStr::Ascii("definiately"),
                &["definitely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistribucion"),
                &["redistribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("friendzond"),
                &["friendzoned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cercomstancial"),
                &["circumstantial"],
            ),
            (dictgen::InsensitiveStr::Ascii("diffrence"), &["difference"]),
            (
                dictgen::InsensitiveStr::Ascii("confugures"),
                &["configures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cirsumstances"),
                &["circumstances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overrite"),
                &["overwrite", "override", "overrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("chnged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("ecept"), &["except"]),
            (dictgen::InsensitiveStr::Ascii("segmening"), &["segmenting"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptiion"),
                &["description"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("occasionnally"),
                &["occasionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("compeletes"), &["completes"]),
            (
                dictgen::InsensitiveStr::Ascii("suspencion"),
                &["suspension"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condiitonal"),
                &["conditional"],
            ),
            (dictgen::InsensitiveStr::Ascii("secruity"), &["security"]),
            (dictgen::InsensitiveStr::Ascii("varience"), &["variance"]),
            (dictgen::InsensitiveStr::Ascii("issueing"), &["issuing"]),
            (
                dictgen::InsensitiveStr::Ascii("shapt"),
                &["shape", "shaped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissassociating"),
                &["disassociating"],
            ),
            (dictgen::InsensitiveStr::Ascii("edtior"), &["editor"]),
            (dictgen::InsensitiveStr::Ascii("agreeeing"), &["agreeing"]),
            (dictgen::InsensitiveStr::Ascii("elctric"), &["electric"]),
            (
                dictgen::InsensitiveStr::Ascii("indonesien"),
                &["indonesian"],
            ),
            (dictgen::InsensitiveStr::Ascii("sturtured"), &["structured"]),
            (dictgen::InsensitiveStr::Ascii("refereers"), &["referees"]),
            (
                dictgen::InsensitiveStr::Ascii("attriubutes"),
                &["attributes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conributing"),
                &["contributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commecting"),
                &["connecting", "commenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserializaing"),
                &["deserializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("verifiying"), &["verifying"]),
            (
                dictgen::InsensitiveStr::Ascii("kontamination"),
                &["contamination"],
            ),
            (dictgen::InsensitiveStr::Ascii("themselvs"), &["themselves"]),
            (
                dictgen::InsensitiveStr::Ascii("pretentous"),
                &["pretentious"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sychronmode"),
                &["synchronmode"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crytography"),
                &["cryptography"],
            ),
            (dictgen::InsensitiveStr::Ascii("radeus"), &["radius"]),
            (dictgen::InsensitiveStr::Ascii("simmetric"), &["symmetric"]),
            (
                dictgen::InsensitiveStr::Ascii("complexitiy"),
                &["complexity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oscilations"),
                &["oscillations"],
            ),
            (dictgen::InsensitiveStr::Ascii("fnishes"), &["finishes"]),
            (
                dictgen::InsensitiveStr::Ascii("developpement"),
                &["development"],
            ),
            (dictgen::InsensitiveStr::Ascii("athros"), &["atheros"]),
            (dictgen::InsensitiveStr::Ascii("profie"), &["profile"]),
            (
                dictgen::InsensitiveStr::Ascii("contageous"),
                &["contagious"],
            ),
            (dictgen::InsensitiveStr::Ascii("buffors"), &["buffers"]),
            (dictgen::InsensitiveStr::Ascii("afoter"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("recommnds"), &["recommends"]),
            (dictgen::InsensitiveStr::Ascii("mashined"), &["machined"]),
            (
                dictgen::InsensitiveStr::Ascii("divising"),
                &["dividing", "devising"],
            ),
            (dictgen::InsensitiveStr::Ascii("chocies"), &["choices"]),
            (dictgen::InsensitiveStr::Ascii("deviice"), &["device"]),
            (
                dictgen::InsensitiveStr::Ascii("deploiments"),
                &["deployments"],
            ),
            (dictgen::InsensitiveStr::Ascii("evaluetor"), &["evaluator"]),
            (
                dictgen::InsensitiveStr::Ascii("charateristic"),
                &["characteristic"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytsems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("assoction"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("allocatrd"), &["allocated"]),
            (
                dictgen::InsensitiveStr::Ascii("subsitutatble"),
                &["substitutable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resotration"),
                &["restoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("comare"), &["compare"]),
            (
                dictgen::InsensitiveStr::Ascii("vaccinatin"),
                &["vaccinating", "vaccination"],
            ),
            (dictgen::InsensitiveStr::Ascii("filetimes"), &["lifetimes"]),
            (
                dictgen::InsensitiveStr::Ascii("unpreductive"),
                &["unproductive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patronozing"),
                &["patronizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("utilsing"), &["utilising"]),
            (dictgen::InsensitiveStr::Ascii("allocting"), &["allocating"]),
            (dictgen::InsensitiveStr::Ascii("littelry"), &["literally"]),
            (dictgen::InsensitiveStr::Ascii("sparlking"), &["sparkling"]),
            (dictgen::InsensitiveStr::Ascii("billng"), &["billing"]),
            (
                dictgen::InsensitiveStr::Ascii("transformees"),
                &["transforms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trusthworthy"),
                &["trustworthy"],
            ),
            (dictgen::InsensitiveStr::Ascii("garbadge"), &["garbage"]),
            (
                dictgen::InsensitiveStr::Ascii("substanties"),
                &["substantive"],
            ),
            (dictgen::InsensitiveStr::Ascii("howvere"), &["however"]),
            (
                dictgen::InsensitiveStr::Ascii("assumpation"),
                &["assumption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proprogating"),
                &["propagating"],
            ),
            (dictgen::InsensitiveStr::Ascii("decerned"), &["discerned"]),
            (dictgen::InsensitiveStr::Ascii("virtural"), &["virtual"]),
            (dictgen::InsensitiveStr::Ascii("efffects"), &["effects"]),
            (dictgen::InsensitiveStr::Ascii("speficied"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("victoris"), &["victories"]),
            (dictgen::InsensitiveStr::Ascii("instrace"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("mesmerisin"),
                &["mesmerising"],
            ),
            (dictgen::InsensitiveStr::Ascii("servcies"), &["services"]),
            (
                dictgen::InsensitiveStr::Ascii("calbirates"),
                &["calibrates"],
            ),
            (dictgen::InsensitiveStr::Ascii("muscial"), &["musical"]),
            (
                dictgen::InsensitiveStr::Ascii("conceousally"),
                &["consciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alpanumeric"),
                &["alphanumeric"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underwhemling"),
                &["underwhelming"],
            ),
            (dictgen::InsensitiveStr::Ascii("fialog"), &["dialog"]),
            (
                dictgen::InsensitiveStr::Ascii("opitmising"),
                &["optimising"],
            ),
            (dictgen::InsensitiveStr::Ascii("commmand"), &["command"]),
            (
                dictgen::InsensitiveStr::Ascii("normallizes"),
                &["normalizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("conncetor"), &["connector"]),
            (
                dictgen::InsensitiveStr::Ascii("demensions"),
                &["dimensions"],
            ),
            (dictgen::InsensitiveStr::Ascii("bistreams"), &["bitstreams"]),
            (dictgen::InsensitiveStr::Ascii("cascace"), &["cascade"]),
            (dictgen::InsensitiveStr::Ascii("lengt"), &["length"]),
            (dictgen::InsensitiveStr::Ascii("paintin"), &["painting"]),
            (
                dictgen::InsensitiveStr::Ascii("guranteeing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("proffesses"), &["professes"]),
            (dictgen::InsensitiveStr::Ascii("legel"), &["legal"]),
            (dictgen::InsensitiveStr::Ascii("neutralt"), &["neutrality"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinaton"),
                &["coordination"],
            ),
            (dictgen::InsensitiveStr::Ascii("preblems"), &["problems"]),
            (
                dictgen::InsensitiveStr::Ascii("straigntened"),
                &["straightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmiting"),
                &["committing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tibetian"), &["tibetan"]),
            (
                dictgen::InsensitiveStr::Ascii("theologia"),
                &["theological"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("imporv"),
                &["improve", "improv"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatabillity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("permenantly"),
                &["permanently"],
            ),
            (dictgen::InsensitiveStr::Ascii("decodding"), &["decoding"]),
            (
                dictgen::InsensitiveStr::Ascii("pennyslvania"),
                &["pennsylvania"],
            ),
            (dictgen::InsensitiveStr::Ascii("openening"), &["opening"]),
            (
                dictgen::InsensitiveStr::Ascii("facsination"),
                &["fascination"],
            ),
            (dictgen::InsensitiveStr::Ascii("woulld"), &["would"]),
            (
                dictgen::InsensitiveStr::Ascii("configury"),
                &["configurable"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuorial"), &["tutorial"]),
            (
                dictgen::InsensitiveStr::Ascii("specilisations"),
                &["specialisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dummp"), &["dump", "dummy"]),
            (
                dictgen::InsensitiveStr::Ascii("identication"),
                &["identification"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacuming"), &["vacuuming"]),
            (
                dictgen::InsensitiveStr::Ascii("conditionnaly"),
                &["conditionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("typopgrahic"),
                &["typographic"],
            ),
            (dictgen::InsensitiveStr::Ascii("resloves"), &["resolves"]),
            (
                dictgen::InsensitiveStr::Ascii("discontiued"),
                &["discontinued"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reoute"),
                &["route", "reroute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priortization"),
                &["prioritization"],
            ),
            (dictgen::InsensitiveStr::Ascii("suffcient"), &["sufficient"]),
            (dictgen::InsensitiveStr::Ascii("hienas"), &["hyenas"]),
            (dictgen::InsensitiveStr::Ascii("delivared"), &["delivered"]),
            (
                dictgen::InsensitiveStr::Ascii("follwowinwg"),
                &["following"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("platformar"),
                &["platformer"],
            ),
            (dictgen::InsensitiveStr::Ascii("onslaugt"), &["onslaught"]),
            (dictgen::InsensitiveStr::Ascii("weaknes"), &["weakness"]),
            (
                dictgen::InsensitiveStr::Ascii("coommanders"),
                &["commanders"],
            ),
            (dictgen::InsensitiveStr::Ascii("ibject"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("begninning"), &["beginning"]),
            (dictgen::InsensitiveStr::Ascii("exitts"), &["exits"]),
            (dictgen::InsensitiveStr::Ascii("lsip"), &["lisp"]),
            (
                dictgen::InsensitiveStr::Ascii("struggleing"),
                &["struggling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interlectual"),
                &["intellectual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("proportionalty"),
                &["proportionally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intellecutal"),
                &["intellectual"],
            ),
            (dictgen::InsensitiveStr::Ascii("maybee"), &["maybe"]),
            (dictgen::InsensitiveStr::Ascii("similairly"), &["similarly"]),
            (dictgen::InsensitiveStr::Ascii("acrost"), &["across"]),
            (
                dictgen::InsensitiveStr::Ascii("defintiively"),
                &["definitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarrentees"),
                &["guarantees"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("givin"),
                &["giving", "given"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alternatie"),
                &["alternatives", "alternate", "alternative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interacively"),
                &["interactively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("existenial"),
                &["existential"],
            ),
            (dictgen::InsensitiveStr::Ascii("sylablles"), &["syllables"]),
            (dictgen::InsensitiveStr::Ascii("animaion"), &["animation"]),
            (dictgen::InsensitiveStr::Ascii("isntead"), &["instead"]),
            (dictgen::InsensitiveStr::Ascii("verablly"), &["verbally"]),
            (dictgen::InsensitiveStr::Ascii("commandes"), &["commands"]),
            (
                dictgen::InsensitiveStr::Ascii("trochlight"),
                &["torchlight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("situacional"),
                &["situational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("processibg"),
                &["processing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibilty"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("poential"), &["potential"]),
            (
                dictgen::InsensitiveStr::Ascii("sentimant"),
                &["sentimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphyscial"),
                &["metaphysical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegociated"),
                &["autonegotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recomending"),
                &["recommending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respresented"),
                &["represented"],
            ),
            (dictgen::InsensitiveStr::Ascii("comsume"), &["consume"]),
            (dictgen::InsensitiveStr::Ascii("sibiling"), &["sibling"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialise"),
                &["initialise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generilise"),
                &["generalise"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypothenuses"),
                &["hypotenuses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("importan"),
                &["important", "importance"],
            ),
            (dictgen::InsensitiveStr::Ascii("ephemerial"), &["ephemeral"]),
            (dictgen::InsensitiveStr::Ascii("aovid"), &["avoid"]),
            (
                dictgen::InsensitiveStr::Ascii("coordinacion"),
                &["coordination"],
            ),
            (dictgen::InsensitiveStr::Ascii("theives"), &["thieves"]),
            (
                dictgen::InsensitiveStr::Ascii("enchantmet"),
                &["enchantments"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitimizes"), &["optimizes"]),
            (
                dictgen::InsensitiveStr::Ascii("exclusvies"),
                &["exclusives"],
            ),
            (dictgen::InsensitiveStr::Ascii("stongest"), &["strongest"]),
            (
                dictgen::InsensitiveStr::Ascii("regulatios"),
                &["regulators"],
            ),
            (dictgen::InsensitiveStr::Ascii("sometiems"), &["sometimes"]),
            (
                dictgen::InsensitiveStr::Ascii("combindatorial"),
                &["combinatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("threataning"),
                &["threatening"],
            ),
            (dictgen::InsensitiveStr::Ascii("archetect"), &["architect"]),
            (dictgen::InsensitiveStr::Ascii("arhived"), &["archived"]),
            (
                dictgen::InsensitiveStr::Ascii("simptomaticaly"),
                &["symptomatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attactments"),
                &["attachments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nonexixtent"),
                &["nonexistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("recconenct"), &["reconnect"]),
            (
                dictgen::InsensitiveStr::Ascii("automaatically"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constainted"),
                &["constrained", "contained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encription"),
                &["encryption"],
            ),
            (dictgen::InsensitiveStr::Ascii("molelcular"), &["molecular"]),
            (dictgen::InsensitiveStr::Ascii("montitored"), &["monitored"]),
            (dictgen::InsensitiveStr::Ascii("isareli"), &["israeli"]),
            (
                dictgen::InsensitiveStr::Ascii("organisationers"),
                &["organisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sctional"),
                &["sectional", "suctional"],
            ),
            (dictgen::InsensitiveStr::Ascii("chipstes"), &["chipsets"]),
            (dictgen::InsensitiveStr::Ascii("mopdule"), &["module"]),
            (dictgen::InsensitiveStr::Ascii("expensess"), &["expenses"]),
            (dictgen::InsensitiveStr::Ascii("disaters"), &["disasters"]),
            (
                dictgen::InsensitiveStr::Ascii("incremeted"),
                &["incremented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inteligant"),
                &["intelligent"],
            ),
            (dictgen::InsensitiveStr::Ascii("bilsters"), &["blisters"]),
            (dictgen::InsensitiveStr::Ascii("molestarse"), &["molester"]),
            (dictgen::InsensitiveStr::Ascii("flciker"), &["flicker"]),
            (dictgen::InsensitiveStr::Ascii("namesspace"), &["namespace"]),
            (dictgen::InsensitiveStr::Ascii("aboout"), &["about"]),
            (
                dictgen::InsensitiveStr::Ascii("implmenting"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("demenaor"), &["demeanor"]),
            (dictgen::InsensitiveStr::Ascii("decoser"), &["decoder"]),
            (
                dictgen::InsensitiveStr::Ascii("insturcion"),
                &["instruction"],
            ),
            (dictgen::InsensitiveStr::Ascii("mininise"), &["minimise"]),
            (dictgen::InsensitiveStr::Ascii("faceplam"), &["facepalm"]),
            (
                dictgen::InsensitiveStr::Ascii("optimiced"),
                &["optimised", "optimized"],
            ),
            (dictgen::InsensitiveStr::Ascii("naugthy"), &["naughty"]),
            (
                dictgen::InsensitiveStr::Ascii("hilighted"),
                &["highlighted"],
            ),
            (dictgen::InsensitiveStr::Ascii("droppend"), &["dropped"]),
            (dictgen::InsensitiveStr::Ascii("characetr"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("verifable"), &["verifiable"]),
            (dictgen::InsensitiveStr::Ascii("similiars"), &["similar"]),
            (dictgen::InsensitiveStr::Ascii("htere"), &["there", "here"]),
            (dictgen::InsensitiveStr::Ascii("arcehtype"), &["archetype"]),
            (
                dictgen::InsensitiveStr::Ascii("paramaterising"),
                &["parameterising"],
            ),
            (dictgen::InsensitiveStr::Ascii("antivirs"), &["antivirus"]),
            (dictgen::InsensitiveStr::Ascii("duplicat"), &["duplicate"]),
            (dictgen::InsensitiveStr::Ascii("elimate"), &["eliminate"]),
            (
                dictgen::InsensitiveStr::Ascii("alternavtively"),
                &["alternatively"],
            ),
            (dictgen::InsensitiveStr::Ascii("adjsuting"), &["adjusting"]),
            (
                dictgen::InsensitiveStr::Ascii("workarbound"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repetetively"),
                &["repetitively"],
            ),
            (dictgen::InsensitiveStr::Ascii("abolutely"), &["absolutely"]),
            (
                dictgen::InsensitiveStr::Ascii("ankshiosness"),
                &["anxiousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("persued"), &["pursued"]),
            (
                dictgen::InsensitiveStr::Ascii("breack"),
                &["brake", "break"],
            ),
            (dictgen::InsensitiveStr::Ascii("detatiling"), &["detailing"]),
            (
                dictgen::InsensitiveStr::Ascii("bordreline"),
                &["borderline"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interriptions"),
                &["interruptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("phenomon"), &["phenomenon"]),
            (dictgen::InsensitiveStr::Ascii("flattire"), &["flatter"]),
            (
                dictgen::InsensitiveStr::Ascii("compoenents"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synonomous"),
                &["synonymous"],
            ),
            (dictgen::InsensitiveStr::Ascii("paratmer"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("labmda"), &["lambda"]),
            (
                dictgen::InsensitiveStr::Ascii("universites"),
                &["universities"],
            ),
            (dictgen::InsensitiveStr::Ascii("mercahnt"), &["merchant"]),
            (dictgen::InsensitiveStr::Ascii("erors"), &["errors"]),
            (
                dictgen::InsensitiveStr::Ascii("opinionater"),
                &["opinionator", "opinionated", "opinionates", "opinionate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabilitis"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arithemetic"),
                &["arithmetic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hexadeciml"),
                &["hexadecimal"],
            ),
            (dictgen::InsensitiveStr::Ascii("checkksums"), &["checksums"]),
            (
                dictgen::InsensitiveStr::Ascii("negotatied"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("meditaiton"),
                &["meditation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("funcitional"),
                &["functional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privledges"),
                &["privileges"],
            ),
            (dictgen::InsensitiveStr::Ascii("convenent"), &["convenient"]),
            (dictgen::InsensitiveStr::Ascii("cruzer"), &["cruiser"]),
            (
                dictgen::InsensitiveStr::Ascii("perfomrance"),
                &["performances", "performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("emmited"), &["emitted"]),
            (dictgen::InsensitiveStr::Ascii("thsnkful"), &["thankful"]),
            (dictgen::InsensitiveStr::Ascii("breserk"), &["berserk"]),
            (
                dictgen::InsensitiveStr::Ascii("frustartion"),
                &["frustrations"],
            ),
            (dictgen::InsensitiveStr::Ascii("dealying"), &["delaying"]),
            (dictgen::InsensitiveStr::Ascii("montsh"), &["months"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowiedgement"),
                &["acknowledgement"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("independentisme"),
                &["independents"],
            ),
            (dictgen::InsensitiveStr::Ascii("answere"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("cacausian"), &["caucasian"]),
            (dictgen::InsensitiveStr::Ascii("deesil"), &["diesel"]),
            (dictgen::InsensitiveStr::Ascii("montnana"), &["montana"]),
            (
                dictgen::InsensitiveStr::Ascii("prodiction"),
                &["production"],
            ),
            (dictgen::InsensitiveStr::Ascii("provives"), &["provides"]),
            (dictgen::InsensitiveStr::Ascii("orgasmus"), &["orgasms"]),
            (
                dictgen::InsensitiveStr::Ascii("apperant"),
                &["apparent", "aberrant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indepedent"),
                &["independent"],
            ),
            (dictgen::InsensitiveStr::Ascii("answhars"), &["answers"]),
            (
                dictgen::InsensitiveStr::Ascii("initialially"),
                &["initially"],
            ),
            (dictgen::InsensitiveStr::Ascii("diahrrea"), &["diarrhea"]),
            (dictgen::InsensitiveStr::Ascii("chaninging"), &["changing"]),
            (dictgen::InsensitiveStr::Ascii("easisly"), &["easily"]),
            (dictgen::InsensitiveStr::Ascii("rcall"), &["recall"]),
            (
                dictgen::InsensitiveStr::Ascii("containered"),
                &["contained"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explotation"),
                &["exploitation", "exploration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tabualting"),
                &["tabulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conifiguring"),
                &["configuring"],
            ),
            (dictgen::InsensitiveStr::Ascii("apetures"), &["apertures"]),
            (dictgen::InsensitiveStr::Ascii("trancate"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("sycing"), &["syncing"]),
            (dictgen::InsensitiveStr::Ascii("libraris"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("occcured"), &["occurred"]),
            (dictgen::InsensitiveStr::Ascii("bigallic"), &["bigalloc"]),
            (
                dictgen::InsensitiveStr::Ascii("photoshoppped"),
                &["photoshopped"],
            ),
            (dictgen::InsensitiveStr::Ascii("supoose"), &["suppose"]),
            (dictgen::InsensitiveStr::Ascii("createor"), &["creator"]),
            (
                dictgen::InsensitiveStr::Ascii("deallcoated"),
                &["deallocated"],
            ),
            (dictgen::InsensitiveStr::Ascii("receieves"), &["receives"]),
            (dictgen::InsensitiveStr::Ascii("agreees"), &["agrees"]),
            (dictgen::InsensitiveStr::Ascii("startes"), &["starts"]),
            (
                dictgen::InsensitiveStr::Ascii("indentaion"),
                &["indentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("ahaed"), &["ahead"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteram"),
                &["downstream"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("illustator"),
                &["illustrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("divdes"), &["divides"]),
            (dictgen::InsensitiveStr::Ascii("packaeges"), &["packages"]),
            (dictgen::InsensitiveStr::Ascii("apche"), &["apache"]),
            (
                dictgen::InsensitiveStr::Ascii("autheticates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("nasted"), &["nested"]),
            (dictgen::InsensitiveStr::Ascii("seleced"), &["selected"]),
            (
                dictgen::InsensitiveStr::Ascii("degeneraged"),
                &["degenerated"],
            ),
            (dictgen::InsensitiveStr::Ascii("peformers"), &["performers"]),
            (
                dictgen::InsensitiveStr::Ascii("sepcification"),
                &["specification"],
            ),
            (dictgen::InsensitiveStr::Ascii("ihs"), &["his"]),
            (
                dictgen::InsensitiveStr::Ascii("severitie"),
                &["severity", "severities"],
            ),
            (dictgen::InsensitiveStr::Ascii("meetin"), &["meeting"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmacueticals"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulerabule"),
                &["vulnerable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masturabted"),
                &["masturbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isntruments"),
                &["instruments"],
            ),
            (dictgen::InsensitiveStr::Ascii("assumues"), &["assumes"]),
            (
                dictgen::InsensitiveStr::Ascii("probablilities"),
                &["probabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acccordances"),
                &["accordances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innaccurate"),
                &["inaccurate"],
            ),
            (dictgen::InsensitiveStr::Ascii("reflecte"), &["reflective"]),
            (
                dictgen::InsensitiveStr::Ascii("conributes"),
                &["contributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("dependecy"), &["dependency"]),
            (
                dictgen::InsensitiveStr::Ascii("minneaplois"),
                &["minneapolis"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challeneges"),
                &["challenges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hererosexual"),
                &["heterosexual"],
            ),
            (dictgen::InsensitiveStr::Ascii("dending"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("certificaton"),
                &["certification"],
            ),
            (dictgen::InsensitiveStr::Ascii("insalling"), &["installing"]),
            (
                dictgen::InsensitiveStr::Ascii("expecations"),
                &["expectations"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadamy"), &["academy"]),
            (dictgen::InsensitiveStr::Ascii("oppenly"), &["openly"]),
            (dictgen::InsensitiveStr::Ascii("strucutre"), &["structure"]),
            (dictgen::InsensitiveStr::Ascii("releveant"), &["relevant"]),
            (
                dictgen::InsensitiveStr::Ascii("intersetllar"),
                &["interstellar"],
            ),
            (dictgen::InsensitiveStr::Ascii("dogding"), &["dodging"]),
            (dictgen::InsensitiveStr::Ascii("ammounting"), &["amounting"]),
            (dictgen::InsensitiveStr::Ascii("annoints"), &["anoints"]),
            (dictgen::InsensitiveStr::Ascii("elecrto"), &["electro"]),
            (
                dictgen::InsensitiveStr::Ascii("sectopming"),
                &["sectioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("testiing"), &["testing"]),
            (dictgen::InsensitiveStr::Ascii("anouther"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("filted"),
                &["filtered", "filed"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacntuary"), &["sanctuary"]),
            (dictgen::InsensitiveStr::Ascii("guarrantee"), &["guarantee"]),
            (
                dictgen::InsensitiveStr::Ascii("compeditors"),
                &["competitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("thign"), &["thing"]),
            (dictgen::InsensitiveStr::Ascii("stribe"), &["strive"]),
            (
                dictgen::InsensitiveStr::Ascii("irrationably"),
                &["irrationally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quadrooples"),
                &["quadruples"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serverities"),
                &["severities"],
            ),
            (dictgen::InsensitiveStr::Ascii("preblem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("poweshell"), &["powershell"]),
            (
                dictgen::InsensitiveStr::Ascii("funxtionalities"),
                &["functionalities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symmetria"),
                &["symmetrical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lacalization"),
                &["localization"],
            ),
            (dictgen::InsensitiveStr::Ascii("protugal"), &["portugal"]),
            (
                dictgen::InsensitiveStr::Ascii("expidetion"),
                &["expedition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconetion"),
                &["reconnection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inusrgency"),
                &["insurgency"],
            ),
            (dictgen::InsensitiveStr::Ascii("seledtion"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("aquiess"), &["acquiesce"]),
            (dictgen::InsensitiveStr::Ascii("apperar"), &["appear"]),
            (dictgen::InsensitiveStr::Ascii("maidson"), &["madison"]),
            (dictgen::InsensitiveStr::Ascii("quadddec"), &["quaddec"]),
            (
                dictgen::InsensitiveStr::Ascii("exectutable"),
                &["executable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformas"),
                &["transforms"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreving"), &["retrieving"]),
            (dictgen::InsensitiveStr::Ascii("sidleine"), &["sideline"]),
            (dictgen::InsensitiveStr::Ascii("iputs"), &["inputs"]),
            (dictgen::InsensitiveStr::Ascii("dokcre"), &["docker"]),
            (
                dictgen::InsensitiveStr::Ascii("corrispondants"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anestheisa"),
                &["anesthesia"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intaxication"),
                &["intoxication"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvotters"),
                &["downvoters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("necessites"),
                &["necessities"],
            ),
            (dictgen::InsensitiveStr::Ascii("descriped"), &["described"]),
            (dictgen::InsensitiveStr::Ascii("qualifer"), &["qualifier"]),
            (dictgen::InsensitiveStr::Ascii("drammatic"), &["dramatic"]),
            (
                dictgen::InsensitiveStr::Ascii("intercource"),
                &["intercourse"],
            ),
            (dictgen::InsensitiveStr::Ascii("ehtanol"), &["ethanol"]),
            (dictgen::InsensitiveStr::Ascii("tryin"), &["trying"]),
            (dictgen::InsensitiveStr::Ascii("ocmpared"), &["compared"]),
            (
                dictgen::InsensitiveStr::Ascii("navgiation"),
                &["navigation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prooerties"),
                &["properties"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definetally"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("pblisher"), &["publisher"]),
            (
                dictgen::InsensitiveStr::Ascii("represenatational"),
                &["representational"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrested"), &["interested"]),
            (dictgen::InsensitiveStr::Ascii("tupel"), &["tuple"]),
            (dictgen::InsensitiveStr::Ascii("transtion"), &["transition"]),
            (dictgen::InsensitiveStr::Ascii("selecs"), &["selects"]),
            (
                dictgen::InsensitiveStr::Ascii("makrs"),
                &["makes", "makers", "macros", "marks"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacutrers"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("masterbated"),
                &["masturbated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inisializer"),
                &["initializer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("releagtion"),
                &["relegation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticalyl"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("oppener"), &["opener"]),
            (dictgen::InsensitiveStr::Ascii("cammands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("sequnces"), &["sequences"]),
            (
                dictgen::InsensitiveStr::Ascii("upsteaming"),
                &["upstreaming"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deserailizing"),
                &["deserializing"],
            ),
            (dictgen::InsensitiveStr::Ascii("palystyle"), &["playstyle"]),
            (dictgen::InsensitiveStr::Ascii("trancated"), &["truncated"]),
            (dictgen::InsensitiveStr::Ascii("symbole"), &["symbol"]),
            (dictgen::InsensitiveStr::Ascii("checket"), &["checked"]),
            (dictgen::InsensitiveStr::Ascii("tansfers"), &["transfers"]),
            (
                dictgen::InsensitiveStr::Ascii("faliing"),
                &["failing", "falling"],
            ),
            (dictgen::InsensitiveStr::Ascii("moudule"), &["module"]),
            (
                dictgen::InsensitiveStr::Ascii("implemening"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("morgtages"), &["mortgages"]),
            (dictgen::InsensitiveStr::Ascii("obations"), &["ovations"]),
            (
                dictgen::InsensitiveStr::Ascii("envireonmental"),
                &["environmental"],
            ),
            (dictgen::InsensitiveStr::Ascii("lablels"), &["labels"]),
            (dictgen::InsensitiveStr::Ascii("aesthethic"), &["aesthetic"]),
            (
                dictgen::InsensitiveStr::Ascii("confuigration"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("positiv"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("modifed"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("accessses"), &["accesses"]),
            (dictgen::InsensitiveStr::Ascii("condmen"), &["condemn"]),
            (
                dictgen::InsensitiveStr::Ascii("expoliting"),
                &["exploiting"],
            ),
            (dictgen::InsensitiveStr::Ascii("varts"), &["carts", "warts"]),
            (
                dictgen::InsensitiveStr::Ascii("propators"),
                &["propagators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatitble"),
                &["compatible"],
            ),
            (dictgen::InsensitiveStr::Ascii("reciding"), &["residing"]),
            (
                dictgen::InsensitiveStr::Ascii("dezentralized"),
                &["decentralized"],
            ),
            (dictgen::InsensitiveStr::Ascii("spicify"), &["specify"]),
            (
                dictgen::InsensitiveStr::Ascii("derrivatives"),
                &["derivatives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("devirutalized"),
                &["devirtualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonymousely"),
                &["anonymously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roundabount"),
                &["roundabout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metaphoricaly"),
                &["metaphorically"],
            ),
            (dictgen::InsensitiveStr::Ascii("leanring"), &["learning"]),
            (dictgen::InsensitiveStr::Ascii("desogned"), &["designed"]),
            (dictgen::InsensitiveStr::Ascii("bookkepp"), &["bookkeep"]),
            (dictgen::InsensitiveStr::Ascii("succesd"), &["succeeds"]),
            (dictgen::InsensitiveStr::Ascii("simpilify"), &["simplify"]),
            (
                dictgen::InsensitiveStr::Ascii("instnatiation"),
                &["instantiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("effecticely"),
                &["effectively"],
            ),
            (dictgen::InsensitiveStr::Ascii("fossilis"), &["fossils"]),
            (dictgen::InsensitiveStr::Ascii("contancts"), &["contacts"]),
            (dictgen::InsensitiveStr::Ascii("repoduce"), &["reproduce"]),
            (dictgen::InsensitiveStr::Ascii("rasie"), &["raise"]),
            (
                dictgen::InsensitiveStr::Ascii("recongized"),
                &["recognized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stuructural"),
                &["structural"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggegate"), &["aggregate"]),
            (dictgen::InsensitiveStr::Ascii("nessicary"), &["necessary"]),
            (
                dictgen::InsensitiveStr::Ascii("manipulant"),
                &["manipulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("combiations"),
                &["combinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependancys"),
                &["dependencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("answhares"), &["answers"]),
            (
                dictgen::InsensitiveStr::Ascii("correspomdent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("partiitoned"),
                &["partitioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("duplaicated"),
                &["duplicated"],
            ),
            (dictgen::InsensitiveStr::Ascii("modififies"), &["modifies"]),
            (
                dictgen::InsensitiveStr::Ascii("symantically"),
                &["semantically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsitent"),
                &["inconsistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authetnticates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("detals"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("lineary"), &["linearly"]),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalistiska"),
                &["fundamentalists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oscilating"),
                &["oscillating"],
            ),
            (dictgen::InsensitiveStr::Ascii("fountan"), &["fountain"]),
            (
                dictgen::InsensitiveStr::Ascii("assements"),
                &["assessments"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolutly"), &["absolutely"]),
            (dictgen::InsensitiveStr::Ascii("meixcan"), &["mexican"]),
            (dictgen::InsensitiveStr::Ascii("seezing"), &["seizing"]),
            (
                dictgen::InsensitiveStr::Ascii("sepatation"),
                &["separation"],
            ),
            (dictgen::InsensitiveStr::Ascii("seletors"), &["selectors"]),
            (dictgen::InsensitiveStr::Ascii("ponitd"), &["pointed"]),
            (dictgen::InsensitiveStr::Ascii("thiunk"), &["think"]),
            (
                dictgen::InsensitiveStr::Ascii("elminating"),
                &["eliminating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("augmneting"),
                &["augmenting"],
            ),
            (dictgen::InsensitiveStr::Ascii("possiible"), &["possible"]),
            (
                dictgen::InsensitiveStr::Ascii("altetnatives"),
                &["alternatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("operaton"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("durectory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("transportr"),
                &["transporter"],
            ),
            (dictgen::InsensitiveStr::Ascii("licsense"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("licated"), &["located"]),
            (dictgen::InsensitiveStr::Ascii("posessing"), &["possessing"]),
            (dictgen::InsensitiveStr::Ascii("phyical"), &["physical"]),
            (
                dictgen::InsensitiveStr::Ascii("undergratuate"),
                &["undergraduate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("polymoprhic"),
                &["polymorphic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temproraily"),
                &["temporarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundamentalistiskt"),
                &["fundamentalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("snowballls"), &["snowballs"]),
            (
                dictgen::InsensitiveStr::Ascii("unexepectedly"),
                &["unexpectedly"],
            ),
            (dictgen::InsensitiveStr::Ascii("usere"), &["user"]),
            (
                dictgen::InsensitiveStr::Ascii("opinionaters"),
                &["opinionators", "opinionates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precedural"),
                &["procedural"],
            ),
            (dictgen::InsensitiveStr::Ascii("transolate"), &["translate"]),
            (dictgen::InsensitiveStr::Ascii("depenting"), &["depending"]),
            (
                dictgen::InsensitiveStr::Ascii("unmainted"),
                &["unmaintained"],
            ),
            (dictgen::InsensitiveStr::Ascii("rewatchig"), &["rewatching"]),
            (
                dictgen::InsensitiveStr::Ascii("concervative"),
                &["conservative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alphanumberic"),
                &["alphanumeric"],
            ),
            (dictgen::InsensitiveStr::Ascii("alocation"), &["allocation"]),
            (dictgen::InsensitiveStr::Ascii("alphapets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("necesssarily"),
                &["necessarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("inbrios"), &["embryos"]),
            (dictgen::InsensitiveStr::Ascii("workir"), &["worker"]),
            (dictgen::InsensitiveStr::Ascii("undercunt"), &["undercut"]),
            (
                dictgen::InsensitiveStr::Ascii("customizin"),
                &["customizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ciclically"),
                &["cyclically"],
            ),
            (dictgen::InsensitiveStr::Ascii("loaction"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("generaates"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("cannibas"), &["cannabis"]),
            (
                dictgen::InsensitiveStr::Ascii("infideltiy"),
                &["infidelity"],
            ),
            (dictgen::InsensitiveStr::Ascii("abusin"), &["abusing"]),
            (dictgen::InsensitiveStr::Ascii("semding"), &["sending"]),
            (
                dictgen::InsensitiveStr::Ascii("colorfull"),
                &["colorful", "colorfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("imrove"), &["improve"]),
            (dictgen::InsensitiveStr::Ascii("dribbeld"), &["dribbled"]),
            (
                dictgen::InsensitiveStr::Ascii("reimplments"),
                &["reimplements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("definte"),
                &["definite", "define"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anarchsits"),
                &["anarchists"],
            ),
            (dictgen::InsensitiveStr::Ascii("yhank"), &["thank", "yank"]),
            (dictgen::InsensitiveStr::Ascii("mehcanic"), &["mechanic"]),
            (dictgen::InsensitiveStr::Ascii("fasr"), &["fast"]),
            (dictgen::InsensitiveStr::Ascii("literarry"), &["literary"]),
            (
                dictgen::InsensitiveStr::Ascii("posibilties"),
                &["possibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("alrightly"), &["alrighty"]),
            (dictgen::InsensitiveStr::Ascii("beleving"), &["believing"]),
            (dictgen::InsensitiveStr::Ascii("usnig"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("numbr"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("systesm"), &["systems"]),
            (dictgen::InsensitiveStr::Ascii("alegbra"), &["algebra"]),
            (
                dictgen::InsensitiveStr::Ascii("responese"),
                &["responses", "response"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negoatiators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geomitrically"),
                &["geometrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("ovrriding"), &["overriding"]),
            (
                dictgen::InsensitiveStr::Ascii("percengtage"),
                &["percentages"],
            ),
            (dictgen::InsensitiveStr::Ascii("quantaty"), &["quantity"]),
            (dictgen::InsensitiveStr::Ascii("accoird"), &["accord"]),
            (dictgen::InsensitiveStr::Ascii("stationd"), &["stationed"]),
            (dictgen::InsensitiveStr::Ascii("directrx"), &["directx"]),
            (dictgen::InsensitiveStr::Ascii("membrances"), &["membrane"]),
            (dictgen::InsensitiveStr::Ascii("hierachry"), &["hierarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("pararagraphs"),
                &["paragraphs"],
            ),
            (dictgen::InsensitiveStr::Ascii("communty"), &["community"]),
            (dictgen::InsensitiveStr::Ascii("speficifer"), &["specifier"]),
            (
                dictgen::InsensitiveStr::Ascii("cartirdges"),
                &["cartridges"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("backgrouns"),
                &["backgrounds"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("positioing"),
                &["positioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("preffix"), &["prefix"]),
            (dictgen::InsensitiveStr::Ascii("skipd"), &["skipped"]),
            (
                dictgen::InsensitiveStr::Ascii("excentuating"),
                &["accentuating"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleif"), &["belief"]),
            (dictgen::InsensitiveStr::Ascii("intruiged"), &["intrigued"]),
            (
                dictgen::InsensitiveStr::Ascii("calcultater"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visiblilities"),
                &["visibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("koolots"), &["culottes"]),
            (
                dictgen::InsensitiveStr::Ascii("tranlsator"),
                &["translator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demonstrant"),
                &["demonstrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disassocate"),
                &["disassociate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monopolios"),
                &["monopolies"],
            ),
            (dictgen::InsensitiveStr::Ascii("warnngs"), &["warnings"]),
            (
                dictgen::InsensitiveStr::Ascii("intelegent"),
                &["intelligent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colaboration"),
                &["collaboration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("continusly"),
                &["continuously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suspeciously"),
                &["suspiciously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expementer"),
                &["experimenter"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cumpolsory"),
                &["compulsory"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("valus"),
                &["values", "talus", "value"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spefcifieirs"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("provsioning"),
                &["provisioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("natique"), &["antique"]),
            (
                dictgen::InsensitiveStr::Ascii("trancievers"),
                &["transceivers"],
            ),
            (dictgen::InsensitiveStr::Ascii("prabably"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("iplemented"),
                &["implemented"],
            ),
            (dictgen::InsensitiveStr::Ascii("configurte"), &["configure"]),
            (dictgen::InsensitiveStr::Ascii("neurton"), &["neutron"]),
            (
                dictgen::InsensitiveStr::Ascii("statitical"),
                &["statistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("snyc"), &["sync"]),
            (dictgen::InsensitiveStr::Ascii("unsucess"), &["unsuccess"]),
            (dictgen::InsensitiveStr::Ascii("sybols"), &["symbols"]),
            (dictgen::InsensitiveStr::Ascii("ukranie"), &["ukraine"]),
            (
                dictgen::InsensitiveStr::Ascii("porgrammers"),
                &["programmers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inefficenty"),
                &["inefficiently"],
            ),
            (dictgen::InsensitiveStr::Ascii("archivin"), &["archiving"]),
            (dictgen::InsensitiveStr::Ascii("desinger"), &["designer"]),
            (dictgen::InsensitiveStr::Ascii("keywprd"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("settters"), &["setters"]),
            (
                dictgen::InsensitiveStr::Ascii("inkompetence"),
                &["incompetence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("challanger"),
                &["challenger"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compremised"),
                &["compromised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simliarities"),
                &["similarities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expresscoin"),
                &["expression"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parellelises"),
                &["parallelises"],
            ),
            (dictgen::InsensitiveStr::Ascii("timeour"), &["timeout"]),
            (dictgen::InsensitiveStr::Ascii("arguemtns"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("currnetly"), &["currently"]),
            (
                dictgen::InsensitiveStr::Ascii("antisipate"),
                &["anticipate"],
            ),
            (dictgen::InsensitiveStr::Ascii("capcitor"), &["capacitor"]),
            (
                dictgen::InsensitiveStr::Ascii("benchmakrs"),
                &["benchmarks"],
            ),
            (dictgen::InsensitiveStr::Ascii("edites"), &["edits"]),
            (dictgen::InsensitiveStr::Ascii("specfield"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("disconenction"),
                &["disconnection"],
            ),
            (dictgen::InsensitiveStr::Ascii("szes"), &["sizes"]),
            (
                dictgen::InsensitiveStr::Ascii("termiantor"),
                &["terminator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intervied"),
                &["interviewed"],
            ),
            (dictgen::InsensitiveStr::Ascii("disporue"), &["disparue"]),
            (
                dictgen::InsensitiveStr::Ascii("competiting"),
                &["competition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orchestratin"),
                &["orchestrating", "orchestration"],
            ),
            (dictgen::InsensitiveStr::Ascii("instuct"), &["instruct"]),
            (dictgen::InsensitiveStr::Ascii("ascpects"), &["aspects"]),
            (dictgen::InsensitiveStr::Ascii("plesantly"), &["pleasantly"]),
            (dictgen::InsensitiveStr::Ascii("moodify"), &["modify"]),
            (dictgen::InsensitiveStr::Ascii("espianoge"), &["espionage"]),
            (
                dictgen::InsensitiveStr::Ascii("hypocritcal"),
                &["hypocritical"],
            ),
            (dictgen::InsensitiveStr::Ascii("articels"), &["articles"]),
            (dictgen::InsensitiveStr::Ascii("almigthy"), &["almighty"]),
            (dictgen::InsensitiveStr::Ascii("skooling"), &["schooling"]),
            (dictgen::InsensitiveStr::Ascii("preserverd"), &["preserved"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchroneous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commissionor"),
                &["commissioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constallation"),
                &["constellation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sistematically"),
                &["systematically"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacoming"), &["becoming"]),
            (dictgen::InsensitiveStr::Ascii("urgentness"), &["urgency"]),
            (
                dictgen::InsensitiveStr::Ascii("organisatons"),
                &["organisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvotres"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("selecctor"), &["selector"]),
            (
                dictgen::InsensitiveStr::Ascii("initiailizes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("marjority"), &["majority"]),
            (
                dictgen::InsensitiveStr::Ascii("comparaisons"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("respecte"), &["respective"]),
            (
                dictgen::InsensitiveStr::Ascii("notorize"),
                &["notarize", "motorize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeptions"),
                &["exceptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assynchronously"),
                &["asynchronously"],
            ),
            (dictgen::InsensitiveStr::Ascii("woudlve"), &["wouldve"]),
            (
                dictgen::InsensitiveStr::Ascii("restrictied"),
                &["restrictive"],
            ),
            (dictgen::InsensitiveStr::Ascii("arrguing"), &["arguing"]),
            (
                dictgen::InsensitiveStr::Ascii("appliaction"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaultet"), &["defaulted"]),
            (dictgen::InsensitiveStr::Ascii("nearset"), &["nearest"]),
            (dictgen::InsensitiveStr::Ascii("portayed"), &["portrayed"]),
            (
                dictgen::InsensitiveStr::Ascii("asssociate"),
                &["associated", "associate"],
            ),
            (dictgen::InsensitiveStr::Ascii("facorty"), &["factory"]),
            (dictgen::InsensitiveStr::Ascii("varance"), &["variance"]),
            (dictgen::InsensitiveStr::Ascii("consuemd"), &["consumed"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubcribed"),
                &["unsubscribed"],
            ),
            (dictgen::InsensitiveStr::Ascii("hepl"), &["help"]),
            (
                dictgen::InsensitiveStr::Ascii("stablizers"),
                &["stabilizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pecularities"),
                &["peculiarities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represening"),
                &["representing"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuncates"), &["truncates"]),
            (dictgen::InsensitiveStr::Ascii("pedistal"), &["pedestal"]),
            (
                dictgen::InsensitiveStr::Ascii("enthusiasim"),
                &["enthusiasm"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delgations"),
                &["delegations"],
            ),
            (dictgen::InsensitiveStr::Ascii("compears"), &["compares"]),
            (
                dictgen::InsensitiveStr::Ascii("calcutates"),
                &["calculates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("protectoin"),
                &["protections", "protection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bulnerabilities"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("peacd"), &["peace"]),
            (dictgen::InsensitiveStr::Ascii("seledting"), &["selecting"]),
            (
                dictgen::InsensitiveStr::Ascii("sinnagog"),
                &["synagog", "synagogue"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("auhtenticators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("rhe"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("catchi"), &["catch"]),
            (
                dictgen::InsensitiveStr::Ascii("canocicalizes"),
                &["canonicalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peripherial"),
                &["peripheral"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimentally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("analyers"),
                &["analysers", "analyzers"],
            ),
            (dictgen::InsensitiveStr::Ascii("expesive"), &["expensive"]),
            (
                dictgen::InsensitiveStr::Ascii("precrastination"),
                &["procrastination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cryptocraphic"),
                &["cryptographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stravation"),
                &["starvation"],
            ),
            (dictgen::InsensitiveStr::Ascii("gloassary"), &["glossary"]),
            (
                dictgen::InsensitiveStr::Ascii("retrvievers"),
                &["retrievers"],
            ),
            (dictgen::InsensitiveStr::Ascii("unregster"), &["unregister"]),
            (dictgen::InsensitiveStr::Ascii("separte"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("massters"), &["masters"]),
            (dictgen::InsensitiveStr::Ascii("treasurs"), &["treasures"]),
            (dictgen::InsensitiveStr::Ascii("atachment"), &["attachment"]),
            (
                dictgen::InsensitiveStr::Ascii("paragarapha"),
                &["paragraph"],
            ),
            (dictgen::InsensitiveStr::Ascii("shepperds"), &["shepherds"]),
            (
                dictgen::InsensitiveStr::Ascii("infinintely"),
                &["infinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("pakets"), &["packets"]),
            (dictgen::InsensitiveStr::Ascii("insid"), &["inside"]),
            (dictgen::InsensitiveStr::Ascii("decpetion"), &["deception"]),
            (dictgen::InsensitiveStr::Ascii("coreutil"), &["coreutils"]),
            (
                dictgen::InsensitiveStr::Ascii("criticallity"),
                &["critically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dupicating"),
                &["duplicating"],
            ),
            (dictgen::InsensitiveStr::Ascii("couloumb"), &["coulomb"]),
            (dictgen::InsensitiveStr::Ascii("impproved"), &["improved"]),
            (dictgen::InsensitiveStr::Ascii("dominanted"), &["dominated"]),
            (dictgen::InsensitiveStr::Ascii("lcoalize"), &["localize"]),
            (
                dictgen::InsensitiveStr::Ascii("vulberabilites"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("marginaly"), &["marginally"]),
            (
                dictgen::InsensitiveStr::Ascii("treatements"),
                &["treatments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strams"),
                &["steams", "streams", "trams"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcess"), &["process"]),
            (
                dictgen::InsensitiveStr::Ascii("plagiariam"),
                &["plagiarism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpepectedly"),
                &["unexpectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("susbcriber"),
                &["subscriber"],
            ),
            (dictgen::InsensitiveStr::Ascii("merchantos"), &["merchants"]),
            (
                dictgen::InsensitiveStr::Ascii("cutomizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("solfing"), &["solving"]),
            (
                dictgen::InsensitiveStr::Ascii("deaktivated"),
                &["deactivated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("geographicly"),
                &["geographical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("criticizng"),
                &["criticizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("owerwrites"),
                &["overwrites"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rememoring"),
                &["remembering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mantion"),
                &["mention", "mansion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("charecterization"),
                &["characterization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indenpendently"),
                &["independently"],
            ),
            (dictgen::InsensitiveStr::Ascii("dispuse"), &["disputes"]),
            (dictgen::InsensitiveStr::Ascii("uncomon"), &["uncommon"]),
            (dictgen::InsensitiveStr::Ascii("aanother"), &["another"]),
            (
                dictgen::InsensitiveStr::Ascii("neccessities"),
                &["necessities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trasiently"),
                &["transiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itnerpreting"),
                &["interpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prarameters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("exression"), &["expression"]),
            (
                dictgen::InsensitiveStr::Ascii("bulnerable"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("shorterm"), &["shorter"]),
            (dictgen::InsensitiveStr::Ascii("buffereed"), &["buffered"]),
            (
                dictgen::InsensitiveStr::Ascii("enviorments"),
                &["environments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touple"),
                &["tuple", "couple", "topple", "toupee"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interbally"),
                &["internally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiliazed"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("enocoding"), &["encoding"]),
            (dictgen::InsensitiveStr::Ascii("reguarl"), &["regular"]),
            (dictgen::InsensitiveStr::Ascii("aparment"), &["apartment"]),
            (
                dictgen::InsensitiveStr::Ascii("influencie"),
                &["influences"],
            ),
            (dictgen::InsensitiveStr::Ascii("proplem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("itsself"), &["itself"]),
            (
                dictgen::InsensitiveStr::Ascii("gymnasitcs"),
                &["gymnastics"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eliminaste"),
                &["eliminates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplayer"),
                &["multiplayer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaptibelity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differnty"),
                &["different", "differently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("annoncements"),
                &["announcements"],
            ),
            (dictgen::InsensitiveStr::Ascii("expiered"), &["expired"]),
            (dictgen::InsensitiveStr::Ascii("eprror"), &["error"]),
            (dictgen::InsensitiveStr::Ascii("exectutive"), &["executive"]),
            (
                dictgen::InsensitiveStr::Ascii("elimnation"),
                &["elimination"],
            ),
            (dictgen::InsensitiveStr::Ascii("crsytals"), &["crystals"]),
            (dictgen::InsensitiveStr::Ascii("analysise"), &["analyses"]),
            (dictgen::InsensitiveStr::Ascii("nrtwork"), &["network"]),
            (
                dictgen::InsensitiveStr::Ascii("predefiened"),
                &["predefined"],
            ),
            (dictgen::InsensitiveStr::Ascii("novisheut"), &["novitiate"]),
            (dictgen::InsensitiveStr::Ascii("constanty"), &["constantly"]),
            (dictgen::InsensitiveStr::Ascii("sepetates"), &["separates"]),
            (dictgen::InsensitiveStr::Ascii("fiedled"), &["fielded"]),
            (
                dictgen::InsensitiveStr::Ascii("controllor"),
                &["controller"],
            ),
            (dictgen::InsensitiveStr::Ascii("acheiving"), &["achieving"]),
            (
                dictgen::InsensitiveStr::Ascii("renegoticable"),
                &["renegotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schowing"),
                &["showing", "chowing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questionaire"),
                &["questionnaire"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simutaneous"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compilatons"),
                &["compilations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("operattional"),
                &["operational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precondtioner"),
                &["preconditioner"],
            ),
            (dictgen::InsensitiveStr::Ascii("sxl"), &["xsl"]),
            (dictgen::InsensitiveStr::Ascii("diaspra"), &["diaspora"]),
            (
                dictgen::InsensitiveStr::Ascii("coddes"),
                &["codes", "coddles"],
            ),
            (dictgen::InsensitiveStr::Ascii("debths"), &["depths"]),
            (dictgen::InsensitiveStr::Ascii("esacape"), &["escape"]),
            (dictgen::InsensitiveStr::Ascii("igzorted"), &["exhorted"]),
            (
                dictgen::InsensitiveStr::Ascii("ivolving"),
                &["involving", "evolving"],
            ),
            (dictgen::InsensitiveStr::Ascii("cricles"), &["circles"]),
            (dictgen::InsensitiveStr::Ascii("alghorithm"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("creationst"),
                &["creationist"],
            ),
            (dictgen::InsensitiveStr::Ascii("tke"), &["take"]),
            (
                dictgen::InsensitiveStr::Ascii("unaproving"),
                &["unapproving"],
            ),
            (dictgen::InsensitiveStr::Ascii("coleeges"), &["colleagues"]),
            (dictgen::InsensitiveStr::Ascii("filterring"), &["filtering"]),
            (
                dictgen::InsensitiveStr::Ascii("suscription"),
                &["subscription"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attraktive"),
                &["attractive"],
            ),
            (dictgen::InsensitiveStr::Ascii("cmak"), &["cmake"]),
            (dictgen::InsensitiveStr::Ascii("segement"), &["segment"]),
            (
                dictgen::InsensitiveStr::Ascii("instrucitons"),
                &["instruction", "instructions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hearhtstone"),
                &["hearthstone"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inefficently"),
                &["inefficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commentsry"),
                &["commenters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straightforeward"),
                &["straightforward"],
            ),
            (dictgen::InsensitiveStr::Ascii("calllers"), &["callers"]),
            (
                dictgen::InsensitiveStr::Ascii("unsoclicited"),
                &["unsolicited"],
            ),
            (dictgen::InsensitiveStr::Ascii("depndance"), &["dependance"]),
            (dictgen::InsensitiveStr::Ascii("overlows"), &["overflows"]),
            (
                dictgen::InsensitiveStr::Ascii("belown"),
                &["belong", "below", "blown"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shilouette"),
                &["silhouette"],
            ),
            (dictgen::InsensitiveStr::Ascii("accpetors"), &["acceptors"]),
            (dictgen::InsensitiveStr::Ascii("seemes"), &["seems"]),
            (
                dictgen::InsensitiveStr::Ascii("licencse"),
                &["licence", "license", "licenses"],
            ),
            (dictgen::InsensitiveStr::Ascii("jurnals"), &["journals"]),
            (dictgen::InsensitiveStr::Ascii("aligne"), &["align"]),
            (dictgen::InsensitiveStr::Ascii("scroipting"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("cellebrates"),
                &["celebrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extravogant"),
                &["extravagant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unrecognixed"),
                &["unrecognized"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecape"), &["escape"]),
            (dictgen::InsensitiveStr::Ascii("upgerades"), &["upgrades"]),
            (dictgen::InsensitiveStr::Ascii("genrator"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("opportuinity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silhoueting"),
                &["silhouetting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perpandicular"),
                &["perpendicular"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intuatively"),
                &["intuitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("judgenental"),
                &["judgemental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("homogeoneous"),
                &["homogeneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("enthusiat"), &["enthusiast"]),
            (dictgen::InsensitiveStr::Ascii("optomism"), &["optimism"]),
            (
                dictgen::InsensitiveStr::Ascii("hirin"),
                &["hiring", "mirin"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("applikations"),
                &["applications"],
            ),
            (dictgen::InsensitiveStr::Ascii("garph"), &["graph"]),
            (
                dictgen::InsensitiveStr::Ascii("exected"),
                &["executed", "expected", "ejected", "exerted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suscepitble"),
                &["susceptible"],
            ),
            (dictgen::InsensitiveStr::Ascii("fianlizes"), &["finalizes"]),
            (dictgen::InsensitiveStr::Ascii("hyrdation"), &["hydration"]),
            (dictgen::InsensitiveStr::Ascii("unhelpfull"), &["unhelpful"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotatiators"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frustrantes"),
                &["frustrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orthongonalised"),
                &["orthogonalised"],
            ),
            (dictgen::InsensitiveStr::Ascii("cteated"), &["created"]),
            (
                dictgen::InsensitiveStr::Ascii("accseptable"),
                &["acceptable"],
            ),
            (dictgen::InsensitiveStr::Ascii("fragmanted"), &["fragment"]),
            (dictgen::InsensitiveStr::Ascii("resonable"), &["reasonable"]),
            (dictgen::InsensitiveStr::Ascii("follwos"), &["follows"]),
            (
                dictgen::InsensitiveStr::Ascii("accaptably"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("generatuons"),
                &["generations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reclutant"), &["reluctant"]),
            (
                dictgen::InsensitiveStr::Ascii("efford"),
                &["effort", "afford"],
            ),
            (dictgen::InsensitiveStr::Ascii("attampted"), &["attempted"]),
            (
                dictgen::InsensitiveStr::Ascii("cilincers"),
                &["cylinders", "silencers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarenteede"),
                &["guarantee"],
            ),
            (dictgen::InsensitiveStr::Ascii("debudgging"), &["debugging"]),
            (dictgen::InsensitiveStr::Ascii("headding"), &["heading"]),
            (
                dictgen::InsensitiveStr::Ascii("explosiones"),
                &["explosions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expeirimentelly"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("descripcion"),
                &["description"],
            ),
            (dictgen::InsensitiveStr::Ascii("comonly"), &["commonly"]),
            (dictgen::InsensitiveStr::Ascii("kepoint"), &["keypoint"]),
            (dictgen::InsensitiveStr::Ascii("cimmands"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("vartices"), &["vertices"]),
            (dictgen::InsensitiveStr::Ascii("snythetic"), &["synthetic"]),
            (
                dictgen::InsensitiveStr::Ascii("comapnionship"),
                &["companionship"],
            ),
            (dictgen::InsensitiveStr::Ascii("labenese"), &["lebanese"]),
            (
                dictgen::InsensitiveStr::Ascii("associatated"),
                &["associated"],
            ),
            (dictgen::InsensitiveStr::Ascii("portolio"), &["portfolio"]),
            (
                dictgen::InsensitiveStr::Ascii("whitepsaces"),
                &["whitespaces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("satisfactorly"),
                &["satisfactory"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiondow"), &["window"]),
            (
                dictgen::InsensitiveStr::Ascii("initialisaing"),
                &["initialising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("replacemnts"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("attribbute"), &["attribute"]),
            (
                dictgen::InsensitiveStr::Ascii("metropilis"),
                &["metropolis"],
            ),
            (dictgen::InsensitiveStr::Ascii("abbilities"), &["abilities"]),
            (
                dictgen::InsensitiveStr::Ascii("integreating"),
                &["integrating"],
            ),
            (dictgen::InsensitiveStr::Ascii("varaince"), &["variance"]),
            (
                dictgen::InsensitiveStr::Ascii("reflectivos"),
                &["reflections"],
            ),
            (dictgen::InsensitiveStr::Ascii("semphores"), &["semaphores"]),
            (dictgen::InsensitiveStr::Ascii("naybour"), &["neighbour"]),
            (dictgen::InsensitiveStr::Ascii("equalls"), &["equals"]),
            (
                dictgen::InsensitiveStr::Ascii("caculators"),
                &["calculators"],
            ),
            (dictgen::InsensitiveStr::Ascii("pronounds"), &["pronouns"]),
            (
                dictgen::InsensitiveStr::Ascii("subnegotionations"),
                &["subnegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("whill"), &["while", "will"]),
            (
                dictgen::InsensitiveStr::Ascii("rehabiliation"),
                &["rehabilitation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quanlifies"),
                &["qualifies", "quantifies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introduse"),
                &["introduces", "introduce"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosess"),
                &["process", "profess", "prowess", "possess"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responcive"),
                &["responsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("higeinic"), &["hygienic"]),
            (
                dictgen::InsensitiveStr::Ascii("disaggretate"),
                &["disaggregate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experenced"),
                &["experienced"],
            ),
            (dictgen::InsensitiveStr::Ascii("conseeded"), &["conceded"]),
            (dictgen::InsensitiveStr::Ascii("alwas"), &["always"]),
            (
                dictgen::InsensitiveStr::Ascii("termintes"),
                &["terminates", "termites"],
            ),
            (dictgen::InsensitiveStr::Ascii("imporoted"), &["imported"]),
            (dictgen::InsensitiveStr::Ascii("pecision"), &["precision"]),
            (dictgen::InsensitiveStr::Ascii("checers"), &["checkers"]),
            (dictgen::InsensitiveStr::Ascii("stesp"), &["steps"]),
            (dictgen::InsensitiveStr::Ascii("reveiwer"), &["reviewer"]),
            (dictgen::InsensitiveStr::Ascii("wiget"), &["widget"]),
            (dictgen::InsensitiveStr::Ascii("diagostic"), &["diagnostic"]),
            (
                dictgen::InsensitiveStr::Ascii("authenitcators"),
                &["authenticators"],
            ),
            (dictgen::InsensitiveStr::Ascii("endpount"), &["endpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("coninuation"),
                &["continuation"],
            ),
            (dictgen::InsensitiveStr::Ascii("betweek"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("nmme"), &["name"]),
            (dictgen::InsensitiveStr::Ascii("simptoms"), &["symptoms"]),
            (dictgen::InsensitiveStr::Ascii("asserions"), &["assertions"]),
            (dictgen::InsensitiveStr::Ascii("divison"), &["division"]),
            (
                dictgen::InsensitiveStr::Ascii("containining"),
                &["containing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("etror"),
                &["error", "terror"],
            ),
            (dictgen::InsensitiveStr::Ascii("speicfier"), &["specifier"]),
            (dictgen::InsensitiveStr::Ascii("afternnon"), &["afternoon"]),
            (
                dictgen::InsensitiveStr::Ascii("serializble"),
                &["serializable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiazes"),
                &["initializes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trotskyites"),
                &["trotskyists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cliffhander"),
                &["cliffhanger"],
            ),
            (dictgen::InsensitiveStr::Ascii("postional"), &["positional"]),
            (
                dictgen::InsensitiveStr::Ascii("exerciesing"),
                &["exercising"],
            ),
            (dictgen::InsensitiveStr::Ascii("syrcause"), &["syracuse"]),
            (
                dictgen::InsensitiveStr::Ascii("distorition"),
                &["distortion"],
            ),
            (dictgen::InsensitiveStr::Ascii("atual"), &["actual"]),
            (dictgen::InsensitiveStr::Ascii("aphabets"), &["alphabets"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibily"),
                &["compatibility", "compatibly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("antripanewers"),
                &["entrepreneurs"],
            ),
            (dictgen::InsensitiveStr::Ascii("avriants"), &["variants"]),
            (dictgen::InsensitiveStr::Ascii("cubpoard"), &["cupboard"]),
            (
                dictgen::InsensitiveStr::Ascii("negligente"),
                &["negligence"],
            ),
            (dictgen::InsensitiveStr::Ascii("platflorm"), &["platform"]),
            (
                dictgen::InsensitiveStr::Ascii("undesireble"),
                &["undesirable"],
            ),
            (dictgen::InsensitiveStr::Ascii("eextracts"), &["extracts"]),
            (dictgen::InsensitiveStr::Ascii("exampl"), &["example"]),
            (
                dictgen::InsensitiveStr::Ascii("implementig"),
                &["implementing"],
            ),
            (dictgen::InsensitiveStr::Ascii("opitional"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("performancs"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asociative"),
                &["associative"],
            ),
            (dictgen::InsensitiveStr::Ascii("flemmish"), &["flemish"]),
            (
                dictgen::InsensitiveStr::Ascii("permuatating"),
                &["permutating"],
            ),
            (dictgen::InsensitiveStr::Ascii("treee"), &["tree"]),
            (
                dictgen::InsensitiveStr::Ascii("assmumptions"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straigthened"),
                &["straightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vulernable"),
                &["vulnerable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinnic"), &["cynic"]),
            (
                dictgen::InsensitiveStr::Ascii("approximely"),
                &["approximately"],
            ),
            (dictgen::InsensitiveStr::Ascii("scriptiong"), &["scripting"]),
            (
                dictgen::InsensitiveStr::Ascii("automaticaly"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theoreticall"),
                &["theoretically"],
            ),
            (dictgen::InsensitiveStr::Ascii("aquiring"), &["acquiring"]),
            (
                dictgen::InsensitiveStr::Ascii("aimprovements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("redistributin"),
                &["redistribution", "redistributing"],
            ),
            (dictgen::InsensitiveStr::Ascii("forsee"), &["foresee"]),
            (
                dictgen::InsensitiveStr::Ascii("chrminance"),
                &["chrominance"],
            ),
            (dictgen::InsensitiveStr::Ascii("comprese"), &["compressed"]),
            (
                dictgen::InsensitiveStr::Ascii("descenants"),
                &["descendants"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pronuntiation"),
                &["pronunciation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimntal"),
                &["experimental"],
            ),
            (dictgen::InsensitiveStr::Ascii("faktoring"), &["factoring"]),
            (
                dictgen::InsensitiveStr::Ascii("tansitions"),
                &["transitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("inteded"), &["intended"]),
            (dictgen::InsensitiveStr::Ascii("gausssian"), &["gaussian"]),
            (
                dictgen::InsensitiveStr::Ascii("wrose"),
                &["worse", "wrote", "whose"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("construtced"),
                &["constructed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("grather"),
                &["greater", "gather"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramerises"),
                &["parameterises"],
            ),
            (dictgen::InsensitiveStr::Ascii("esacaped"), &["escaped"]),
            (dictgen::InsensitiveStr::Ascii("optoinal"), &["optional"]),
            (
                dictgen::InsensitiveStr::Ascii("referenzes"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("extermities"),
                &["extremities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("progarammer"),
                &["programmer"],
            ),
            (dictgen::InsensitiveStr::Ascii("forbbiden"), &["forbidden"]),
            (
                dictgen::InsensitiveStr::Ascii("mininterpreting"),
                &["misinterpreting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("custonisation"),
                &["customisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("immmediately"),
                &["immediately"],
            ),
            (dictgen::InsensitiveStr::Ascii("editoras"), &["editors"]),
            (dictgen::InsensitiveStr::Ascii("alloted"), &["allotted"]),
            (dictgen::InsensitiveStr::Ascii("fale"), &["false", "fail"]),
            (
                dictgen::InsensitiveStr::Ascii("abbout"),
                &["about", "abbot"],
            ),
            (dictgen::InsensitiveStr::Ascii("smapled"), &["sampled"]),
            (dictgen::InsensitiveStr::Ascii("sterio"), &["stereo"]),
            (dictgen::InsensitiveStr::Ascii("ountline"), &["outline"]),
            (dictgen::InsensitiveStr::Ascii("advaning"), &["advancing"]),
            (dictgen::InsensitiveStr::Ascii("unsername"), &["username"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegoatiations"),
                &["autonegotiations"],
            ),
            (dictgen::InsensitiveStr::Ascii("docuhes"), &["douches"]),
            (
                dictgen::InsensitiveStr::Ascii("deplacements"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("matrials"), &["materials"]),
            (
                dictgen::InsensitiveStr::Ascii("steriotype"),
                &["stereotype"],
            ),
            (dictgen::InsensitiveStr::Ascii("wissles"), &["whistles"]),
            (dictgen::InsensitiveStr::Ascii("seelct"), &["select"]),
            (
                dictgen::InsensitiveStr::Ascii("metalurgic"),
                &["metallurgic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("battlehsips"),
                &["battleship"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offesnively"),
                &["offensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("messagin"), &["messaging"]),
            (
                dictgen::InsensitiveStr::Ascii("implemememnt"),
                &["implement"],
            ),
            (dictgen::InsensitiveStr::Ascii("iscrated"), &["iscreated"]),
            (dictgen::InsensitiveStr::Ascii("statuets"), &["statutes"]),
            (dictgen::InsensitiveStr::Ascii("beleieve"), &["believe"]),
            (dictgen::InsensitiveStr::Ascii("imporots"), &["imports"]),
            (
                dictgen::InsensitiveStr::Ascii("refractice"),
                &["refractive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specailists"),
                &["specialists"],
            ),
            (dictgen::InsensitiveStr::Ascii("disaable"), &["disable"]),
            (
                dictgen::InsensitiveStr::Ascii("atmoshperic"),
                &["atmospheric"],
            ),
            (dictgen::InsensitiveStr::Ascii("tickness"), &["thickness"]),
            (
                dictgen::InsensitiveStr::Ascii("transalates"),
                &["translates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unemploymed"),
                &["unemployed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contribuent"),
                &["contribute"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorritmically"),
                &["algorithmically"],
            ),
            (dictgen::InsensitiveStr::Ascii("bofore"), &["before"]),
            (dictgen::InsensitiveStr::Ascii("constols"), &["controls"]),
            (dictgen::InsensitiveStr::Ascii("derrived"), &["derived"]),
            (
                dictgen::InsensitiveStr::Ascii("reputposing"),
                &["repurposing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("isntalation"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("sugguested"), &["suggested"]),
            (dictgen::InsensitiveStr::Ascii("typpo"), &["typo"]),
            (dictgen::InsensitiveStr::Ascii("squrriels"), &["squirrels"]),
            (dictgen::InsensitiveStr::Ascii("meeans"), &["means"]),
            (dictgen::InsensitiveStr::Ascii("escludes"), &["excludes"]),
            (dictgen::InsensitiveStr::Ascii("optomises"), &["optimises"]),
            (
                dictgen::InsensitiveStr::Ascii("infogrpahics"),
                &["infographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("separaters"),
                &["separates", "separators"],
            ),
            (dictgen::InsensitiveStr::Ascii("blegian"), &["belgian"]),
            (dictgen::InsensitiveStr::Ascii("bulgrian"), &["bulgarian"]),
            (
                dictgen::InsensitiveStr::Ascii("dployments"),
                &["deployments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lightenning"),
                &["lightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("paymet"), &["payment"]),
            (
                dictgen::InsensitiveStr::Ascii("sies"),
                &["size", "sighs", "sides"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apaptations"),
                &["adaptations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("avaialbaility"),
                &["availability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("senstively"),
                &["sensitively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialisated"),
                &["initialised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interasting"),
                &["interacting"],
            ),
            (dictgen::InsensitiveStr::Ascii("quantuty"), &["quantity"]),
            (
                dictgen::InsensitiveStr::Ascii("decoratrion"),
                &["decoration"],
            ),
            (dictgen::InsensitiveStr::Ascii("gaysha"), &["geisha"]),
            (dictgen::InsensitiveStr::Ascii("servicies"), &["services"]),
            (dictgen::InsensitiveStr::Ascii("aparant"), &["apparent"]),
            (dictgen::InsensitiveStr::Ascii("surprized"), &["surprised"]),
            (
                dictgen::InsensitiveStr::Ascii("simulatneous"),
                &["simultaneous"],
            ),
            (dictgen::InsensitiveStr::Ascii("wihle"), &["while"]),
            (
                dictgen::InsensitiveStr::Ascii("differnetial"),
                &["differential"],
            ),
            (dictgen::InsensitiveStr::Ascii("packeged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("comming"), &["coming"]),
            (
                dictgen::InsensitiveStr::Ascii("compatibilties"),
                &["compatibilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instllations"),
                &["installations", "instillations"],
            ),
            (dictgen::InsensitiveStr::Ascii("symnol"), &["symbol"]),
            (dictgen::InsensitiveStr::Ascii("sctripting"), &["scripting"]),
            (dictgen::InsensitiveStr::Ascii("speeak"), &["speak"]),
            (dictgen::InsensitiveStr::Ascii("decoses"), &["decodes"]),
            (
                dictgen::InsensitiveStr::Ascii("calbirators"),
                &["calibrators"],
            ),
            (dictgen::InsensitiveStr::Ascii("aritfacts"), &["artifacts"]),
            (dictgen::InsensitiveStr::Ascii("clucthing"), &["clutching"]),
            (
                dictgen::InsensitiveStr::Ascii("coneectivities"),
                &["connectivities"],
            ),
            (dictgen::InsensitiveStr::Ascii("acadimic"), &["academic"]),
            (dictgen::InsensitiveStr::Ascii("hardwirted"), &["hardwired"]),
            (dictgen::InsensitiveStr::Ascii("structued"), &["structured"]),
            (dictgen::InsensitiveStr::Ascii("analyized"), &["analyzed"]),
            (dictgen::InsensitiveStr::Ascii("criterea"), &["criteria"]),
            (
                dictgen::InsensitiveStr::Ascii("disrepancy"),
                &["discrepancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("equestions"), &["questions"]),
            (dictgen::InsensitiveStr::Ascii("gypises"), &["gypsies"]),
            (dictgen::InsensitiveStr::Ascii("mortards"), &["mortars"]),
            (
                dictgen::InsensitiveStr::Ascii("censur"),
                &["censor", "censure"],
            ),
            (dictgen::InsensitiveStr::Ascii("gospells"), &["gospels"]),
            (
                dictgen::InsensitiveStr::Ascii("noice"),
                &["noise", "nice", "notice"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investiaged"),
                &["investigated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monumentals"),
                &["monuments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetnations"),
                &["experimentations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innaccessible"),
                &["inaccessible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dissappers"),
                &["disappears"],
            ),
            (dictgen::InsensitiveStr::Ascii("generalis"), &["generals"]),
            (dictgen::InsensitiveStr::Ascii("vluetooth"), &["bluetooth"]),
            (dictgen::InsensitiveStr::Ascii("traveresal"), &["traversal"]),
            (
                dictgen::InsensitiveStr::Ascii("accumualting"),
                &["accumulating"],
            ),
            (dictgen::InsensitiveStr::Ascii("cousre"), &["course"]),
            (dictgen::InsensitiveStr::Ascii("broadcsat"), &["broadcast"]),
            (dictgen::InsensitiveStr::Ascii("buttrey"), &["buttery"]),
            (dictgen::InsensitiveStr::Ascii("exchagnged"), &["exchanged"]),
            (
                dictgen::InsensitiveStr::Ascii("annilihated"),
                &["annihilated"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefiics"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("racisst"), &["racists"]),
            (dictgen::InsensitiveStr::Ascii("voicemal"), &["voicemail"]),
            (
                dictgen::InsensitiveStr::Ascii("condigdialog"),
                &["configdialog"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pennsylvainia"),
                &["pennsylvania"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrolates"),
                &["correlates"],
            ),
            (dictgen::InsensitiveStr::Ascii("propbable"), &["probable"]),
            (
                dictgen::InsensitiveStr::Ascii("preposturous"),
                &["preposterous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genralizes"),
                &["generalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anphitheater"),
                &["amphitheater"],
            ),
            (dictgen::InsensitiveStr::Ascii("instnace"), &["instance"]),
            (
                dictgen::InsensitiveStr::Ascii("propiertary"),
                &["proprietary"],
            ),
            (dictgen::InsensitiveStr::Ascii("gamme"), &["gamma", "game"]),
            (dictgen::InsensitiveStr::Ascii("lazyness"), &["laziness"]),
            (dictgen::InsensitiveStr::Ascii("perfixes"), &["prefixes"]),
            (dictgen::InsensitiveStr::Ascii("orientarlo"), &["oriental"]),
            (
                dictgen::InsensitiveStr::Ascii("declaraion"),
                &["declaration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flashligth"),
                &["flashlight"],
            ),
            (dictgen::InsensitiveStr::Ascii("furiuosly"), &["furiously"]),
            (
                dictgen::InsensitiveStr::Ascii("inbelievable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("phsyical"), &["physical"]),
            (
                dictgen::InsensitiveStr::Ascii("automiclly"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("affliated"), &["affiliated"]),
            (
                dictgen::InsensitiveStr::Ascii("communite"),
                &["communities"],
            ),
            (dictgen::InsensitiveStr::Ascii("chassids"), &["chassis"]),
            (dictgen::InsensitiveStr::Ascii("excludled"), &["excluded"]),
            (
                dictgen::InsensitiveStr::Ascii("corrospondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opproximated"),
                &["approximated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("straightoward"),
                &["straightforward", "straightwards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("congratualted"),
                &["congratulate", "congratulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("customie"), &["customize"]),
            (
                dictgen::InsensitiveStr::Ascii("charachters"),
                &["characters"],
            ),
            (dictgen::InsensitiveStr::Ascii("sacrastic"), &["sarcastic"]),
            (
                dictgen::InsensitiveStr::Ascii("futurisitic"),
                &["futuristic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("propertions"),
                &["proportions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("quarentines"),
                &["quarantines"],
            ),
            (dictgen::InsensitiveStr::Ascii("veryfying"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("condidtion"), &["condition"]),
            (
                dictgen::InsensitiveStr::Ascii("californain"),
                &["californian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perfoemance"),
                &["performance"],
            ),
            (dictgen::InsensitiveStr::Ascii("patchtes"), &["patches"]),
            (
                dictgen::InsensitiveStr::Ascii("resucing"),
                &["reducing", "rescuing"],
            ),
            (dictgen::InsensitiveStr::Ascii("storag"), &["storage"]),
            (
                dictgen::InsensitiveStr::Ascii("loction"),
                &["location", "lotion", "lection", "locution"],
            ),
            (dictgen::InsensitiveStr::Ascii("sepecial"), &["special"]),
            (
                dictgen::InsensitiveStr::Ascii("inflexable"),
                &["inflexible"],
            ),
            (dictgen::InsensitiveStr::Ascii("yieleded"), &["yielded"]),
            (dictgen::InsensitiveStr::Ascii("missplace"), &["misplace"]),
            (dictgen::InsensitiveStr::Ascii("aritcles"), &["articles"]),
            (
                dictgen::InsensitiveStr::Ascii("assymentric"),
                &["asymmetric"],
            ),
            (dictgen::InsensitiveStr::Ascii("sarted"), &["started"]),
            (
                dictgen::InsensitiveStr::Ascii("sertainties"),
                &["certainties"],
            ),
            (dictgen::InsensitiveStr::Ascii("platformt"), &["platforms"]),
            (dictgen::InsensitiveStr::Ascii("anmed"), &["named"]),
            (
                dictgen::InsensitiveStr::Ascii("provacative"),
                &["provocative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("probabilstic"),
                &["probabilistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assimtotes"),
                &["asymptotes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sophisticted"),
                &["sophisticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("paymetn"), &["payment"]),
            (
                dictgen::InsensitiveStr::Ascii("linke"),
                &["linked", "like", "link", "links", "line"],
            ),
            (dictgen::InsensitiveStr::Ascii("educte"), &["educate"]),
            (dictgen::InsensitiveStr::Ascii("divsor"), &["divisor"]),
            (dictgen::InsensitiveStr::Ascii("scenerios"), &["scenarios"]),
            (dictgen::InsensitiveStr::Ascii("curelty"), &["cruelty"]),
            (dictgen::InsensitiveStr::Ascii("abandone"), &["abandon"]),
            (dictgen::InsensitiveStr::Ascii("relvant"), &["relevant"]),
            (
                dictgen::InsensitiveStr::Ascii("comparisns"),
                &["comparisons"],
            ),
            (dictgen::InsensitiveStr::Ascii("engramms"), &["engrams"]),
            (dictgen::InsensitiveStr::Ascii("chassies"), &["chassis"]),
            (dictgen::InsensitiveStr::Ascii("experies"), &["expires"]),
            (dictgen::InsensitiveStr::Ascii("fpt"), &["ftp"]),
            (dictgen::InsensitiveStr::Ascii("locaes"), &["locates"]),
            (
                dictgen::InsensitiveStr::Ascii("stripts"),
                &["scripts", "strips"],
            ),
            (dictgen::InsensitiveStr::Ascii("recgonise"), &["recognise"]),
            (dictgen::InsensitiveStr::Ascii("reasult"), &["result"]),
            (dictgen::InsensitiveStr::Ascii("writeing"), &["writing"]),
            (
                dictgen::InsensitiveStr::Ascii("evangalical"),
                &["evangelical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("southernes"),
                &["southerners"],
            ),
            (dictgen::InsensitiveStr::Ascii("solatare"), &["solitaire"]),
            (dictgen::InsensitiveStr::Ascii("exaples"), &["examples"]),
            (
                dictgen::InsensitiveStr::Ascii("excectedly"),
                &["expectedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instituded"),
                &["instituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("activiate"), &["activate"]),
            (
                dictgen::InsensitiveStr::Ascii("parametised"),
                &["parametrised"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unisntalled"),
                &["uninstalled"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convcition"),
                &["convictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighburing"),
                &["neighboring"],
            ),
            (dictgen::InsensitiveStr::Ascii("comiled"), &["compiled"]),
            (dictgen::InsensitiveStr::Ascii("govoners"), &["governors"]),
            (dictgen::InsensitiveStr::Ascii("unexpect"), &["unexpected"]),
            (
                dictgen::InsensitiveStr::Ascii("alphabeticaly"),
                &["alphabetically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referefences"),
                &["references"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("committting"),
                &["committing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reduncancy"),
                &["redundancy"],
            ),
            (dictgen::InsensitiveStr::Ascii("blythely"), &["blithely"]),
            (dictgen::InsensitiveStr::Ascii("historiens"), &["histories"]),
            (dictgen::InsensitiveStr::Ascii("exeuction"), &["execution"]),
            (
                dictgen::InsensitiveStr::Ascii("replacents"),
                &["replacements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("stragegically"),
                &["strategically"],
            ),
            (dictgen::InsensitiveStr::Ascii("loreplay"), &["roleplay"]),
            (dictgen::InsensitiveStr::Ascii("scandalosa"), &["scandals"]),
            (
                dictgen::InsensitiveStr::Ascii("metabalism"),
                &["metabolism"],
            ),
            (dictgen::InsensitiveStr::Ascii("intructor"), &["instructor"]),
            (
                dictgen::InsensitiveStr::Ascii("expermenetally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("troublehshooting"),
                &["troubleshooting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conttributed"),
                &["contributed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conscioussness"),
                &["consciousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shecked"),
                &["checked", "shucked"],
            ),
            (dictgen::InsensitiveStr::Ascii("assitance"), &["assistance"]),
            (dictgen::InsensitiveStr::Ascii("smething"), &["something"]),
            (
                dictgen::InsensitiveStr::Ascii("acknoledge"),
                &["acknowledge"],
            ),
            (dictgen::InsensitiveStr::Ascii("pormising"), &["promising"]),
            (dictgen::InsensitiveStr::Ascii("nimber"), &["number"]),
            (dictgen::InsensitiveStr::Ascii("broacasts"), &["broadcasts"]),
            (dictgen::InsensitiveStr::Ascii("carmalite"), &["carmelite"]),
            (
                dictgen::InsensitiveStr::Ascii("refferenced"),
                &["referenced"],
            ),
            (dictgen::InsensitiveStr::Ascii("corporatie"), &["corporate"]),
            (
                dictgen::InsensitiveStr::Ascii("deserialsise"),
                &["deserialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("recepie"), &["recipe"]),
            (
                dictgen::InsensitiveStr::Ascii("multuplicative"),
                &["multiplicative"],
            ),
            (dictgen::InsensitiveStr::Ascii("recroot"), &["recruit"]),
            (dictgen::InsensitiveStr::Ascii("latops"), &["laptops"]),
            (dictgen::InsensitiveStr::Ascii("upcomin"), &["upcoming"]),
            (dictgen::InsensitiveStr::Ascii("asssuming"), &["assuming"]),
            (dictgen::InsensitiveStr::Ascii("interited"), &["inherited"]),
            (dictgen::InsensitiveStr::Ascii("reteriver"), &["retriever"]),
            (
                dictgen::InsensitiveStr::Ascii("scatched"),
                &["caught", "scratched", "sketched"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("estination"),
                &["estimation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convovling"),
                &["convolving"],
            ),
            (dictgen::InsensitiveStr::Ascii("tiemouts"), &["timeouts"]),
            (dictgen::InsensitiveStr::Ascii("requiests"), &["requests"]),
            (
                dictgen::InsensitiveStr::Ascii("particalurly"),
                &["particularly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hemmorhagic"),
                &["haemorrhagic", "hemorrhagic"],
            ),
            (dictgen::InsensitiveStr::Ascii("cogniscent"), &["cognisant"]),
            (
                dictgen::InsensitiveStr::Ascii("interaxtions"),
                &["interactions"],
            ),
            (dictgen::InsensitiveStr::Ascii("zookeenees"), &["zucchinis"]),
            (dictgen::InsensitiveStr::Ascii("depositd"), &["deposited"]),
            (
                dictgen::InsensitiveStr::Ascii("healtier"),
                &["healthier", "heartier"],
            ),
            (dictgen::InsensitiveStr::Ascii("blantantly"), &["blatantly"]),
            (
                dictgen::InsensitiveStr::Ascii("sentincing"),
                &["sentencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("kenndey"), &["kennedy"]),
            (
                dictgen::InsensitiveStr::Ascii("modfication"),
                &["modification"],
            ),
            (dictgen::InsensitiveStr::Ascii("ciricuits"), &["circuits"]),
            (dictgen::InsensitiveStr::Ascii("holocuast"), &["holocaust"]),
            (
                dictgen::InsensitiveStr::Ascii("embarasing"),
                &["embarrassing"],
            ),
            (dictgen::InsensitiveStr::Ascii("veragated"), &["variegated"]),
            (dictgen::InsensitiveStr::Ascii("reconizes"), &["recognizes"]),
            (
                dictgen::InsensitiveStr::Ascii("derectives"),
                &["directives", "detectives"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("substuctures"),
                &["substructures"],
            ),
            (dictgen::InsensitiveStr::Ascii("incrment"), &["increment"]),
            (dictgen::InsensitiveStr::Ascii("eneabling"), &["enabling"]),
            (dictgen::InsensitiveStr::Ascii("selektion"), &["selection"]),
            (dictgen::InsensitiveStr::Ascii("rmeove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("sustainble"),
                &["sustainable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("lenghtened"),
                &["lengthened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("retcieve"),
                &["receive", "retrieve"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electronagnetic"),
                &["electromagnetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("provsion"), &["provision"]),
            (
                dictgen::InsensitiveStr::Ascii("constainte"),
                &["constrained", "constraint", "constraints"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyandie"), &["cyanide"]),
            (dictgen::InsensitiveStr::Ascii("receice"), &["receive"]),
            (dictgen::InsensitiveStr::Ascii("vry"), &["very"]),
            (dictgen::InsensitiveStr::Ascii("precence"), &["presence"]),
            (
                dictgen::InsensitiveStr::Ascii("depricated"),
                &["deprecated"],
            ),
            (dictgen::InsensitiveStr::Ascii("slanguage"), &["language"]),
            (dictgen::InsensitiveStr::Ascii("arbritray"), &["arbitrary"]),
            (dictgen::InsensitiveStr::Ascii("booteek"), &["boutique"]),
            (dictgen::InsensitiveStr::Ascii("quierying"), &["querying"]),
            (
                dictgen::InsensitiveStr::Ascii("successfullies"),
                &["successfully"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compatabilitiy"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unfortunaletly"),
                &["unfortunately"],
            ),
            (dictgen::InsensitiveStr::Ascii("alignmment"), &["alignment"]),
            (dictgen::InsensitiveStr::Ascii("audiacity"), &["audacity"]),
            (
                dictgen::InsensitiveStr::Ascii("refert"),
                &["revert", "refer", "refers"],
            ),
            (dictgen::InsensitiveStr::Ascii("calld"), &["called"]),
            (dictgen::InsensitiveStr::Ascii("ceasars"), &["caesars"]),
            (dictgen::InsensitiveStr::Ascii("tehtering"), &["tethering"]),
            (
                dictgen::InsensitiveStr::Ascii("tehnically"),
                &["ethnically", "technically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("crytographically"),
                &["cryptographically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memembers"),
                &["members", "remembers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("archievement"),
                &["achievement"],
            ),
            (dictgen::InsensitiveStr::Ascii("scuplture"), &["sculpture"]),
            (
                dictgen::InsensitiveStr::Ascii("configiration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modprobeing"),
                &["modprobing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scourced"),
                &["sourced", "scoured"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("logorithms"),
                &["logarithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("chahged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("maximuims"), &["maximums"]),
            (
                dictgen::InsensitiveStr::Ascii("otional"),
                &["optional", "notional"],
            ),
            (dictgen::InsensitiveStr::Ascii("radomizer"), &["randomizer"]),
            (
                dictgen::InsensitiveStr::Ascii("dedeuplicated"),
                &["deduplicated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("volounteers"),
                &["volunteers"],
            ),
            (dictgen::InsensitiveStr::Ascii("despenced"), &["dispensed"]),
            (dictgen::InsensitiveStr::Ascii("migrains"), &["migraines"]),
            (
                dictgen::InsensitiveStr::Ascii("encapsulte"),
                &["encapsulate"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrction"), &["correction"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotiotable"),
                &["renegotiable"],
            ),
            (dictgen::InsensitiveStr::Ascii("recordarte"), &["recorder"]),
            (
                dictgen::InsensitiveStr::Ascii("sepatating"),
                &["separating"],
            ),
            (dictgen::InsensitiveStr::Ascii("dubled"), &["doubled"]),
            (dictgen::InsensitiveStr::Ascii("emplyment"), &["employment"]),
            (
                dictgen::InsensitiveStr::Ascii("matrial"),
                &["martial", "material"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranistioning"),
                &["transitioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("availible"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("deciamls"), &["decimals"]),
            (
                dictgen::InsensitiveStr::Ascii("recurssions"),
                &["recursions"],
            ),
            (dictgen::InsensitiveStr::Ascii("accurarcy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("digitially"), &["digitally"]),
            (
                dictgen::InsensitiveStr::Ascii("triagulating"),
                &["triangulating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diassociating"),
                &["disassociating", "dissociating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("thesre"),
                &["these", "there"],
            ),
            (dictgen::InsensitiveStr::Ascii("wiats"), &["waits"]),
            (dictgen::InsensitiveStr::Ascii("faillures"), &["failures"]),
            (
                dictgen::InsensitiveStr::Ascii("tightining"),
                &["tightening"],
            ),
            (dictgen::InsensitiveStr::Ascii("benig"), &["being"]),
            (
                dictgen::InsensitiveStr::Ascii("serevrless"),
                &["serverless"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotioates"),
                &["negotiates"],
            ),
            (dictgen::InsensitiveStr::Ascii("lcoation"), &["location"]),
            (dictgen::InsensitiveStr::Ascii("sebastin"), &["sebastian"]),
            (
                dictgen::InsensitiveStr::Ascii("emmisaries"),
                &["emissaries"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pararmeterized"),
                &["parameterized"],
            ),
            (dictgen::InsensitiveStr::Ascii("cliuster"), &["cluster"]),
            (dictgen::InsensitiveStr::Ascii("regargs"), &["regards"]),
            (dictgen::InsensitiveStr::Ascii("openapig"), &["openapi"]),
            (
                dictgen::InsensitiveStr::Ascii("constitutents"),
                &["constitutes", "constituents"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefiifics"), &["specifics"]),
            (
                dictgen::InsensitiveStr::Ascii("assoziates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admnistrator"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("twilgiht"), &["twilight"]),
            (dictgen::InsensitiveStr::Ascii("alaising"), &["aliasing"]),
            (
                dictgen::InsensitiveStr::Ascii("suseptable"),
                &["susceptible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mozzeralla"),
                &["mozzarella"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transpararent"),
                &["transparent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complecate"),
                &["complicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repititions"),
                &["repetitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumstnaces"),
                &["circumstance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("barberians"),
                &["barbarians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("atomtically"),
                &["automatically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("referens"),
                &["references", "reference"],
            ),
            (dictgen::InsensitiveStr::Ascii("murderus"), &["murders"]),
            (dictgen::InsensitiveStr::Ascii("staurdays"), &["saturdays"]),
            (
                dictgen::InsensitiveStr::Ascii("apppreciation"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("helo"), &["hello"]),
            (
                dictgen::InsensitiveStr::Ascii("birghtened"),
                &["brightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("segument"), &["segment"]),
            (dictgen::InsensitiveStr::Ascii("unoredered"), &["unordered"]),
            (dictgen::InsensitiveStr::Ascii("validty"), &["validity"]),
            (
                dictgen::InsensitiveStr::Ascii("helicopteros"),
                &["helicopters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internatioanl"),
                &["international"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contractin"),
                &["contraction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supscriptions"),
                &["subscriptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constratint"),
                &["constraint"],
            ),
            (dictgen::InsensitiveStr::Ascii("themsef"), &["themself"]),
            (
                dictgen::InsensitiveStr::Ascii("registres"),
                &["registries", "registers"],
            ),
            (dictgen::InsensitiveStr::Ascii("evalouates"), &["evaluates"]),
            (dictgen::InsensitiveStr::Ascii("geerators"), &["generators"]),
            (
                dictgen::InsensitiveStr::Ascii("conresponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("wront"), &["wrong"]),
            (
                dictgen::InsensitiveStr::Ascii("previoutly"),
                &["previously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("patitioned"),
                &["partitioned", "petitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("tust"), &["trust"]),
            (
                dictgen::InsensitiveStr::Ascii("inscpeting"),
                &["inspecting"],
            ),
            (dictgen::InsensitiveStr::Ascii("proplems"), &["problems"]),
            (dictgen::InsensitiveStr::Ascii("histroic"), &["historic"]),
            (dictgen::InsensitiveStr::Ascii("coallece"), &["coalesce"]),
            (
                dictgen::InsensitiveStr::Ascii("infiltrase"),
                &["infiltrate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("equitorial"),
                &["equatorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("australlian"),
                &["australian"],
            ),
            (dictgen::InsensitiveStr::Ascii("databaeses"), &["databases"]),
            (dictgen::InsensitiveStr::Ascii("kernals"), &["kernels"]),
            (
                dictgen::InsensitiveStr::Ascii("singaling"),
                &["signaling", "singling", "singalong"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underbaker"),
                &["undertaker"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("electornics"),
                &["electronics"],
            ),
            (dictgen::InsensitiveStr::Ascii("pasttime"), &["pastime"]),
            (dictgen::InsensitiveStr::Ascii("reprensent"), &["represent"]),
            (dictgen::InsensitiveStr::Ascii("abymsal"), &["abysmal"]),
            (
                dictgen::InsensitiveStr::Ascii("compatiablity"),
                &["compatibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("articulted"),
                &["articulated"],
            ),
            (dictgen::InsensitiveStr::Ascii("derivtive"), &["derivative"]),
            (
                dictgen::InsensitiveStr::Ascii("compilaiton"),
                &["compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("steping"), &["stepping"]),
            (
                dictgen::InsensitiveStr::Ascii("intrepretations"),
                &["interpretations"],
            ),
            (dictgen::InsensitiveStr::Ascii("breathos"), &["breaths"]),
            (
                dictgen::InsensitiveStr::Ascii("streightened"),
                &["straightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("opems"), &["opens"]),
            (dictgen::InsensitiveStr::Ascii("caracters"), &["characters"]),
            (
                dictgen::InsensitiveStr::Ascii("progarammed"),
                &["programmed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencicator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("litigatin"), &["litigation"]),
            (
                dictgen::InsensitiveStr::Ascii("appopriately"),
                &["appropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exclusivelly"),
                &["exclusively"],
            ),
            (dictgen::InsensitiveStr::Ascii("ancesetors"), &["ancestors"]),
            (dictgen::InsensitiveStr::Ascii("esnured"), &["ensured"]),
            (
                dictgen::InsensitiveStr::Ascii("insignifacnt"),
                &["insignificant"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consciouness"),
                &["consciousness"],
            ),
            (dictgen::InsensitiveStr::Ascii("amateurest"), &["amateurs"]),
            (
                dictgen::InsensitiveStr::Ascii("inifnitely"),
                &["infinitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("burssels"), &["brussels"]),
            (dictgen::InsensitiveStr::Ascii("disokay"), &["display"]),
            (
                dictgen::InsensitiveStr::Ascii("understaning"),
                &["understanding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaelue"),
                &["value", "valued"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberivates"),
                &["abbreviates"],
            ),
            (dictgen::InsensitiveStr::Ascii("copmilers"), &["compilers"]),
            (dictgen::InsensitiveStr::Ascii("inventarlo"), &["inventor"]),
            (dictgen::InsensitiveStr::Ascii("nihther"), &["neither"]),
            (
                dictgen::InsensitiveStr::Ascii("calculalating"),
                &["calculating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negligable"),
                &["negligible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("aspergerers"),
                &["aspergers"],
            ),
            (dictgen::InsensitiveStr::Ascii("parokeel"), &["parochial"]),
            (
                dictgen::InsensitiveStr::Ascii("carachterisations"),
                &["characterisations"],
            ),
            (dictgen::InsensitiveStr::Ascii("reaveling"), &["revealing"]),
            (dictgen::InsensitiveStr::Ascii("intregity"), &["integrity"]),
            (
                dictgen::InsensitiveStr::Ascii("scheduluing"),
                &["scheduling"],
            ),
            (dictgen::InsensitiveStr::Ascii("auctioners"), &["auctions"]),
            (dictgen::InsensitiveStr::Ascii("scinerio"), &["scenario"]),
            (dictgen::InsensitiveStr::Ascii("growtesk"), &["grotesque"]),
            (
                dictgen::InsensitiveStr::Ascii("sigificantly"),
                &["significantly"],
            ),
            (dictgen::InsensitiveStr::Ascii("periwinke"), &["periwinkle"]),
            (dictgen::InsensitiveStr::Ascii("reserv"), &["reserve"]),
            (
                dictgen::InsensitiveStr::Ascii("uninistalls"),
                &["uninstalls"],
            ),
            (dictgen::InsensitiveStr::Ascii("diggital"), &["digital"]),
            (
                dictgen::InsensitiveStr::Ascii("sepearetly"),
                &["separately"],
            ),
            (dictgen::InsensitiveStr::Ascii("esoterisch"), &["esoteric"]),
            (
                dictgen::InsensitiveStr::Ascii("countryies"),
                &["countryside"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminstration"),
                &["administration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("industrail"),
                &["industrial"],
            ),
            (dictgen::InsensitiveStr::Ascii("chesckout"), &["checkout"]),
            (
                dictgen::InsensitiveStr::Ascii("execise"),
                &["excise", "exercise"],
            ),
            (dictgen::InsensitiveStr::Ascii("inclusinve"), &["inclusive"]),
            (dictgen::InsensitiveStr::Ascii("occupato"), &["occupation"]),
            (dictgen::InsensitiveStr::Ascii("inacive"), &["inactive"]),
            (dictgen::InsensitiveStr::Ascii("resonably"), &["reasonably"]),
            (
                dictgen::InsensitiveStr::Ascii("consentious"),
                &["conscientious"],
            ),
            (dictgen::InsensitiveStr::Ascii("callled"), &["called"]),
            (
                dictgen::InsensitiveStr::Ascii("depencies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comptition"),
                &["competition"],
            ),
            (dictgen::InsensitiveStr::Ascii("dataet"), &["dataset"]),
            (dictgen::InsensitiveStr::Ascii("accumuate"), &["accumulate"]),
            (dictgen::InsensitiveStr::Ascii("compained"), &["complained"]),
            (dictgen::InsensitiveStr::Ascii("horrendos"), &["horrendous"]),
            (dictgen::InsensitiveStr::Ascii("brownei"), &["brownie"]),
            (dictgen::InsensitiveStr::Ascii("habe"), &["have", "babe"]),
            (dictgen::InsensitiveStr::Ascii("chocolot"), &["chocolate"]),
            (
                dictgen::InsensitiveStr::Ascii("endiannity"),
                &["endianness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incidently"),
                &["incidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("musilms"), &["muslims"]),
            (
                dictgen::InsensitiveStr::Ascii("interaced"),
                &["interacted", "interlaced", "interfaced"],
            ),
            (dictgen::InsensitiveStr::Ascii("specificl"), &["specific"]),
            (
                dictgen::InsensitiveStr::Ascii("normlization"),
                &["normalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("commentes"), &["commenters"]),
            (dictgen::InsensitiveStr::Ascii("entiti"), &["entity"]),
            (dictgen::InsensitiveStr::Ascii("distrcits"), &["districts"]),
            (dictgen::InsensitiveStr::Ascii("bolor"), &["color"]),
            (
                dictgen::InsensitiveStr::Ascii("negotionated"),
                &["negotiated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("theripists"),
                &["therapists"],
            ),
            (dictgen::InsensitiveStr::Ascii("croozing"), &["cruising"]),
            (
                dictgen::InsensitiveStr::Ascii("protohypes"),
                &["prototypes"],
            ),
            (dictgen::InsensitiveStr::Ascii("nomrally"), &["normally"]),
            (dictgen::InsensitiveStr::Ascii("modifler"), &["modifier"]),
            (dictgen::InsensitiveStr::Ascii("asssess"), &["assess"]),
            (dictgen::InsensitiveStr::Ascii("opitimised"), &["optimised"]),
            (dictgen::InsensitiveStr::Ascii("contrst"), &["contrast"]),
            (dictgen::InsensitiveStr::Ascii("andoids"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("egaliterian"),
                &["egalitarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repitions"),
                &["repetitions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preventetive"),
                &["preventative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coutermeasueres"),
                &["countermeasures"],
            ),
            (dictgen::InsensitiveStr::Ascii("massivelly"), &["massively"]),
            (
                dictgen::InsensitiveStr::Ascii("procudures"),
                &["procedures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monochrone"),
                &["monochrome"],
            ),
            (dictgen::InsensitiveStr::Ascii("incrasing"), &["increasing"]),
            (dictgen::InsensitiveStr::Ascii("sequneces"), &["sequences"]),
            (dictgen::InsensitiveStr::Ascii("ennemies"), &["enemies"]),
            (
                dictgen::InsensitiveStr::Ascii("monochromo"),
                &["monochrome"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manafacturer"),
                &["manufacturer"],
            ),
            (dictgen::InsensitiveStr::Ascii("contrac"), &["contract"]),
            (dictgen::InsensitiveStr::Ascii("fronends"), &["frontends"]),
            (
                dictgen::InsensitiveStr::Ascii("contaiminated"),
                &["contaminated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("benevelent"),
                &["benevolent"],
            ),
            (dictgen::InsensitiveStr::Ascii("oiginals"), &["originals"]),
            (
                dictgen::InsensitiveStr::Ascii("regenarated"),
                &["regenerated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experimetenting"),
                &["experimenting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ninty"),
                &["ninety", "minty"],
            ),
            (dictgen::InsensitiveStr::Ascii("requried"), &["required"]),
            (
                dictgen::InsensitiveStr::Ascii("diassembling"),
                &["disassembling"],
            ),
            (dictgen::InsensitiveStr::Ascii("cohesie"), &["cohesive"]),
            (dictgen::InsensitiveStr::Ascii("publicaly"), &["publicly"]),
            (
                dictgen::InsensitiveStr::Ascii("initiliazations"),
                &["initializations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("spectacuarly"),
                &["spectacular"],
            ),
            (dictgen::InsensitiveStr::Ascii("altso"), &["also", "altos"]),
            (dictgen::InsensitiveStr::Ascii("satisfyies"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("neckbeardest"),
                &["neckbeards"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("schziophrenia"),
                &["schizophrenia"],
            ),
            (dictgen::InsensitiveStr::Ascii("annoynace"), &["annoyance"]),
            (dictgen::InsensitiveStr::Ascii("plantes"), &["planets"]),
            (
                dictgen::InsensitiveStr::Ascii("disconections"),
                &["disconnections"],
            ),
            (dictgen::InsensitiveStr::Ascii("anthor"), &["another"]),
            (dictgen::InsensitiveStr::Ascii("vaalued"), &["valued"]),
            (dictgen::InsensitiveStr::Ascii("satellie"), &["satellites"]),
            (
                dictgen::InsensitiveStr::Ascii("subsequntly"),
                &["subsequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("oepration"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("oyurself"), &["yourself"]),
            (
                dictgen::InsensitiveStr::Ascii("algoirthmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabiltiy"),
                &["capability"],
            ),
            (dictgen::InsensitiveStr::Ascii("differrent"), &["different"]),
            (
                dictgen::InsensitiveStr::Ascii("condfigured"),
                &["configured"],
            ),
            (dictgen::InsensitiveStr::Ascii("farwarded"), &["forwarded"]),
            (
                dictgen::InsensitiveStr::Ascii("conecntrate"),
                &["concentrate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wepons"), &["weapons"]),
            (dictgen::InsensitiveStr::Ascii("abnrormal"), &["abnormal"]),
            (dictgen::InsensitiveStr::Ascii("inherat"), &["inherit"]),
            (dictgen::InsensitiveStr::Ascii("parant"), &["parent"]),
            (
                dictgen::InsensitiveStr::Ascii("reoresents"),
                &["represents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("centrilized"),
                &["centralized"],
            ),
            (dictgen::InsensitiveStr::Ascii("disabing"), &["disabling"]),
            (dictgen::InsensitiveStr::Ascii("benglas"), &["bengals"]),
            (
                dictgen::InsensitiveStr::Ascii("substante"),
                &["substantive"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entreprenour"),
                &["entrepreneur"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("offensivley"),
                &["offensively"],
            ),
            (dictgen::InsensitiveStr::Ascii("caraff"), &["carafe"]),
            (
                dictgen::InsensitiveStr::Ascii("priniciples"),
                &["principles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internatinalists"),
                &["internationalists"],
            ),
            (dictgen::InsensitiveStr::Ascii("battaries"), &["batteries"]),
            (dictgen::InsensitiveStr::Ascii("stabilty"), &["stability"]),
            (dictgen::InsensitiveStr::Ascii("analyziers"), &["analyzers"]),
            (dictgen::InsensitiveStr::Ascii("precidely"), &["precisely"]),
            (dictgen::InsensitiveStr::Ascii("anser"), &["answer"]),
            (dictgen::InsensitiveStr::Ascii("specical"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("resturant"), &["restaurant"]),
            (dictgen::InsensitiveStr::Ascii("vloned"), &["cloned"]),
            (
                dictgen::InsensitiveStr::Ascii("cinimatography"),
                &["cinematography"],
            ),
            (dictgen::InsensitiveStr::Ascii("closees"), &["closes"]),
            (
                dictgen::InsensitiveStr::Ascii("instnatiated"),
                &["instantiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("contribue"), &["contribute"]),
            (dictgen::InsensitiveStr::Ascii("appied"), &["applied"]),
            (dictgen::InsensitiveStr::Ascii("packgaged"), &["packaged"]),
            (dictgen::InsensitiveStr::Ascii("workbanch"), &["workbench"]),
            (
                dictgen::InsensitiveStr::Ascii("introducting"),
                &["introduction", "introducing"],
            ),
            (dictgen::InsensitiveStr::Ascii("hnalder"), &["handler"]),
            (dictgen::InsensitiveStr::Ascii("otherwiese"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("straightend"),
                &["straightened"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaiation"),
                &["aviation", "variation"],
            ),
            (dictgen::InsensitiveStr::Ascii("bitmast"), &["bitmask"]),
            (dictgen::InsensitiveStr::Ascii("statments"), &["statements"]),
            (
                dictgen::InsensitiveStr::Ascii("trogladitic"),
                &["troglodytic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unambigous"),
                &["unambiguous"],
            ),
            (dictgen::InsensitiveStr::Ascii("liraries"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("impleter"), &["implementer"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapolare"),
                &["extrapolate"],
            ),
            (dictgen::InsensitiveStr::Ascii("prcision"), &["precision"]),
            (
                dictgen::InsensitiveStr::Ascii("qualifiactions"),
                &["qualification"],
            ),
            (dictgen::InsensitiveStr::Ascii("sentionel"), &["sentinel"]),
            (
                dictgen::InsensitiveStr::Ascii("deplications"),
                &["duplications", "replications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exaggaration"),
                &["exaggeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("apartmen"), &["apartment"]),
            (
                dictgen::InsensitiveStr::Ascii("worls"),
                &["world", "works", "worlds", "whorls"],
            ),
            (dictgen::InsensitiveStr::Ascii("perferect"), &["perfect"]),
            (dictgen::InsensitiveStr::Ascii("eraticly"), &["erratically"]),
            (dictgen::InsensitiveStr::Ascii("monopace"), &["monospace"]),
            (dictgen::InsensitiveStr::Ascii("globablly"), &["globally"]),
            (dictgen::InsensitiveStr::Ascii("thid"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("capitalis"), &["capitals"]),
            (
                dictgen::InsensitiveStr::Ascii("catholicisim"),
                &["catholicism"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("journalisim"),
                &["journalism"],
            ),
            (dictgen::InsensitiveStr::Ascii("valudator"), &["validator"]),
            (
                dictgen::InsensitiveStr::Ascii("sacrefices"),
                &["sacrifices"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluclator"),
                &["calculator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virtualied"),
                &["virtualised", "virtualized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resposible"),
                &["responsible"],
            ),
            (dictgen::InsensitiveStr::Ascii("tagued"), &["tagged"]),
            (
                dictgen::InsensitiveStr::Ascii("avialability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("uprading"), &["upgrading"]),
            (
                dictgen::InsensitiveStr::Ascii("collisin"),
                &["collisions", "collision", "collusion"],
            ),
            (dictgen::InsensitiveStr::Ascii("plcaement"), &["placement"]),
            (
                dictgen::InsensitiveStr::Ascii("repertwars"),
                &["repertoires"],
            ),
            (dictgen::InsensitiveStr::Ascii("gthe"), &["the"]),
            (
                dictgen::InsensitiveStr::Ascii("apologistes"),
                &["apologists", "apologises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repoistories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frecuencies"),
                &["frequencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("uupper"), &["upper"]),
            (dictgen::InsensitiveStr::Ascii("seacrch"), &["search"]),
            (dictgen::InsensitiveStr::Ascii("tortillera"), &["tortilla"]),
            (
                dictgen::InsensitiveStr::Ascii("commentaren"),
                &["commenter"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefies"), &["specifies"]),
            (dictgen::InsensitiveStr::Ascii("airators"), &["aerators"]),
            (
                dictgen::InsensitiveStr::Ascii("impossibilty"),
                &["impossibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compariston"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("spezified"), &["specified"]),
            (dictgen::InsensitiveStr::Ascii("judegment"), &["judgment"]),
            (dictgen::InsensitiveStr::Ascii("releast"), &["release"]),
            (
                dictgen::InsensitiveStr::Ascii("accumilate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorihmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depressieve"),
                &["depressive"],
            ),
            (dictgen::InsensitiveStr::Ascii("conquerer"), &["conqueror"]),
            (dictgen::InsensitiveStr::Ascii("bumpter"), &["bumper"]),
            (dictgen::InsensitiveStr::Ascii("alowing"), &["allowing"]),
            (
                dictgen::InsensitiveStr::Ascii("congratualtions"),
                &["congratulations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unavaialable"),
                &["unavailable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contibuting"),
                &["contributing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("driections"),
                &["directions"],
            ),
            (dictgen::InsensitiveStr::Ascii("algortism"), &["algorithm"]),
            (
                dictgen::InsensitiveStr::Ascii("shenanegans"),
                &["shenanigans"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superopeator"),
                &["superoperator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commened"),
                &["commented", "commended", "commend", "commenced", "communed"],
            ),
            (dictgen::InsensitiveStr::Ascii("likebale"), &["likeable"]),
            (dictgen::InsensitiveStr::Ascii("availabale"), &["available"]),
            (dictgen::InsensitiveStr::Ascii("reove"), &["remove"]),
            (
                dictgen::InsensitiveStr::Ascii("disasssembler"),
                &["disassembler"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speciffiers"),
                &["specifiers"],
            ),
            (dictgen::InsensitiveStr::Ascii("bounnding"), &["bounding"]),
            (dictgen::InsensitiveStr::Ascii("traileras"), &["trailers"]),
            (dictgen::InsensitiveStr::Ascii("versitle"), &["versatile"]),
            (
                dictgen::InsensitiveStr::Ascii("encoutners"),
                &["encounters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("alogirthms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimidatie"),
                &["intimidated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distingiush"),
                &["distinguish"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accelerare"),
                &["accelerate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("primitivies"),
                &["primitives"],
            ),
            (dictgen::InsensitiveStr::Ascii("seaparate"), &["separate"]),
            (dictgen::InsensitiveStr::Ascii("pararmeter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("compoonent"), &["component"]),
            (
                dictgen::InsensitiveStr::Ascii("tessellaters"),
                &["tessellators", "tessellates"],
            ),
            (dictgen::InsensitiveStr::Ascii("packats"), &["packets"]),
            (
                dictgen::InsensitiveStr::Ascii("connectivitiy"),
                &["connectivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("driected"), &["directed"]),
            (
                dictgen::InsensitiveStr::Ascii("knowladgeable"),
                &["knowledgeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("procided"), &["provided"]),
            (
                dictgen::InsensitiveStr::Ascii("negatiotiates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vaoids"),
                &["avoids", "voids"],
            ),
            (dictgen::InsensitiveStr::Ascii("memeory"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("chassm"), &["chasm"]),
            (dictgen::InsensitiveStr::Ascii("exectue"), &["execute"]),
            (dictgen::InsensitiveStr::Ascii("fre"), &["free"]),
            (dictgen::InsensitiveStr::Ascii("defualt"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("correpsondent"),
                &["correspondent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("idenfitying"),
                &["identifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("thiw"), &["this"]),
            (dictgen::InsensitiveStr::Ascii("shceduler"), &["scheduler"]),
            (dictgen::InsensitiveStr::Ascii("minitures"), &["miniatures"]),
            (dictgen::InsensitiveStr::Ascii("intriguin"), &["intriguing"]),
            (dictgen::InsensitiveStr::Ascii("preovided"), &["provided"]),
            (dictgen::InsensitiveStr::Ascii("perxoide"), &["peroxide"]),
            (dictgen::InsensitiveStr::Ascii("foward"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("submittions"),
                &["submissions"],
            ),
            (dictgen::InsensitiveStr::Ascii("regeress"), &["regress"]),
            (dictgen::InsensitiveStr::Ascii("usefule"), &["useful"]),
            (dictgen::InsensitiveStr::Ascii("eveolves"), &["evolves"]),
            (dictgen::InsensitiveStr::Ascii("resore"), &["restore"]),
            (dictgen::InsensitiveStr::Ascii("chunck"), &["chunk"]),
            (
                dictgen::InsensitiveStr::Ascii("capablilities"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assasination"),
                &["assassination"],
            ),
            (dictgen::InsensitiveStr::Ascii("andorid"), &["android"]),
            (
                dictgen::InsensitiveStr::Ascii("appliacable"),
                &["applicable"],
            ),
            (dictgen::InsensitiveStr::Ascii("esnure"), &["ensure"]),
            (dictgen::InsensitiveStr::Ascii("briusers"), &["bruisers"]),
            (
                dictgen::InsensitiveStr::Ascii("apperciation"),
                &["appreciation"],
            ),
            (dictgen::InsensitiveStr::Ascii("miinimized"), &["minimized"]),
            (
                dictgen::InsensitiveStr::Ascii("algororithm"),
                &["algorithm"],
            ),
            (dictgen::InsensitiveStr::Ascii("opimizing"), &["optimizing"]),
            (
                dictgen::InsensitiveStr::Ascii("stanger"),
                &["stranger", "stinger", "stagger"],
            ),
            (dictgen::InsensitiveStr::Ascii("parsin"), &["parsing"]),
            (
                dictgen::InsensitiveStr::Ascii("acceleratoin"),
                &["acceleration"],
            ),
            (dictgen::InsensitiveStr::Ascii("bridg"), &["bridge"]),
            (
                dictgen::InsensitiveStr::Ascii("inconviniency"),
                &["inconvenience"],
            ),
            (dictgen::InsensitiveStr::Ascii("proproable"), &["probable"]),
            (
                dictgen::InsensitiveStr::Ascii("custumised"),
                &["customised"],
            ),
            (dictgen::InsensitiveStr::Ascii("enclousre"), &["enclosure"]),
            (
                dictgen::InsensitiveStr::Ascii("confugiration"),
                &["configuration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rocessing"),
                &["processing", "recessing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("workstaitons"),
                &["workstations"],
            ),
            (dictgen::InsensitiveStr::Ascii("theyr"), &["their", "there"]),
            (dictgen::InsensitiveStr::Ascii("rlated"), &["related"]),
            (
                dictgen::InsensitiveStr::Ascii("asscociation"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("missclassify"),
                &["misclassify"],
            ),
            (dictgen::InsensitiveStr::Ascii("destryer"), &["destroyer"]),
            (
                dictgen::InsensitiveStr::Ascii("sycronising"),
                &["synchronising"],
            ),
            (dictgen::InsensitiveStr::Ascii("pysics"), &["physics"]),
            (dictgen::InsensitiveStr::Ascii("capcitors"), &["capacitors"]),
            (
                dictgen::InsensitiveStr::Ascii("discconenct"),
                &["disconnect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("eliminato"),
                &["elimination"],
            ),
            (dictgen::InsensitiveStr::Ascii("drakest"), &["darkest"]),
            (dictgen::InsensitiveStr::Ascii("percursor"), &["precursor"]),
            (dictgen::InsensitiveStr::Ascii("shrinds"), &["shrinks"]),
            (dictgen::InsensitiveStr::Ascii("acurately"), &["accurately"]),
            (
                dictgen::InsensitiveStr::Ascii("subtile"),
                &["subfile", "subtitle", "subtle"],
            ),
            (dictgen::InsensitiveStr::Ascii("mechnisms"), &["mechanisms"]),
            (dictgen::InsensitiveStr::Ascii("swarmin"), &["swarming"]),
            (dictgen::InsensitiveStr::Ascii("aquires"), &["acquires"]),
            (dictgen::InsensitiveStr::Ascii("optains"), &["obtains"]),
            (
                dictgen::InsensitiveStr::Ascii("memerships"),
                &["memberships"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("monochorome"),
                &["monochrome"],
            ),
            (dictgen::InsensitiveStr::Ascii("addional"), &["additional"]),
            (
                dictgen::InsensitiveStr::Ascii("asssumptions"),
                &["assumptions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("anonymoys"),
                &["anonymously"],
            ),
            (dictgen::InsensitiveStr::Ascii("stawk"), &["stalk"]),
            (
                dictgen::InsensitiveStr::Ascii("componemts"),
                &["components"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("memorizacion"),
                &["memorization"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disconects"),
                &["disconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("inutition"), &["intuition"]),
            (dictgen::InsensitiveStr::Ascii("consttruct"), &["construct"]),
            (
                dictgen::InsensitiveStr::Ascii("crystalising"),
                &["crystallising"],
            ),
            (dictgen::InsensitiveStr::Ascii("incliuded"), &["included"]),
            (
                dictgen::InsensitiveStr::Ascii("entaglements"),
                &["entanglements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colleteral"),
                &["collateral"],
            ),
            (dictgen::InsensitiveStr::Ascii("progessor"), &["progressor"]),
            (dictgen::InsensitiveStr::Ascii("cirtcuit"), &["circuit"]),
            (
                dictgen::InsensitiveStr::Ascii("sansitizers"),
                &["sanitizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accredidation"),
                &["accreditation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negociators"),
                &["negotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("triangulatin"),
                &["triangulating", "triangulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mergged"), &["merged"]),
            (dictgen::InsensitiveStr::Ascii("potisive"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("includin"), &["including"]),
            (
                dictgen::InsensitiveStr::Ascii("sublcasses"),
                &["subclasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transformes"),
                &["transformers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inquizition"),
                &["inquisition"],
            ),
            (dictgen::InsensitiveStr::Ascii("senatores"), &["senators"]),
            (dictgen::InsensitiveStr::Ascii("calulated"), &["calculated"]),
            (dictgen::InsensitiveStr::Ascii("proviedes"), &["provides"]),
            (
                dictgen::InsensitiveStr::Ascii("signifficant"),
                &["significant"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifv"), &["specific"]),
            (dictgen::InsensitiveStr::Ascii("commads"), &["commands"]),
            (dictgen::InsensitiveStr::Ascii("conntainer"), &["container"]),
            (
                dictgen::InsensitiveStr::Ascii("commerciales"),
                &["commercials"],
            ),
            (dictgen::InsensitiveStr::Ascii("sctipt"), &["script"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotition"),
                &["renegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("americ"),
                &["america", "amebic"],
            ),
            (dictgen::InsensitiveStr::Ascii("markdowne"), &["markdown"]),
            (dictgen::InsensitiveStr::Ascii("neigbor"), &["neighbor"]),
            (
                dictgen::InsensitiveStr::Ascii("typesrript"),
                &["typescript"],
            ),
            (dictgen::InsensitiveStr::Ascii("liscense"), &["license"]),
            (dictgen::InsensitiveStr::Ascii("springst"), &["sprints"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotaitors"),
                &["renegotiators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("horziontally"),
                &["horizontally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comaparison"),
                &["comparison"],
            ),
            (dictgen::InsensitiveStr::Ascii("commputes"), &["computes"]),
            (
                dictgen::InsensitiveStr::Ascii("entrepreners"),
                &["entrepreneurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condamning"),
                &["condemning"],
            ),
            (dictgen::InsensitiveStr::Ascii("reqrite"), &["rewrite"]),
            (dictgen::InsensitiveStr::Ascii("remining"), &["remaining"]),
            (
                dictgen::InsensitiveStr::Ascii("unprofessinal"),
                &["unprofessional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simialrity"),
                &["similarity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discoraged"),
                &["discouraged"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capabilitires"),
                &["capabilities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admis"),
                &["admins", "admits", "admix"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cordinator"),
                &["coordinator"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algoprithmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscripe"),
                &["unsubscribe"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downgading"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("clearner"),
                &["cleaner", "clearer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subcommannd"),
                &["subcommand"],
            ),
            (dictgen::InsensitiveStr::Ascii("stichting"), &["stitching"]),
            (
                dictgen::InsensitiveStr::Ascii("screenchot"),
                &["screenshot"],
            ),
            (dictgen::InsensitiveStr::Ascii("respest"), &["respects"]),
            (
                dictgen::InsensitiveStr::Ascii("inintialiser"),
                &["initialiser"],
            ),
            (dictgen::InsensitiveStr::Ascii("guniness"), &["guinness"]),
            (
                dictgen::InsensitiveStr::Ascii("infrormational"),
                &["informational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resonsiveness"),
                &["responsiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("sustems"), &["systems"]),
            (
                dictgen::InsensitiveStr::Ascii("prefernces"),
                &["preferences"],
            ),
            (dictgen::InsensitiveStr::Ascii("bivouaced"), &["bivouacked"]),
            (
                dictgen::InsensitiveStr::Ascii("apologistic"),
                &["apologetic", "apologist"],
            ),
            (dictgen::InsensitiveStr::Ascii("resorvoir"), &["reservoir"]),
            (dictgen::InsensitiveStr::Ascii("adter"), &["after"]),
            (dictgen::InsensitiveStr::Ascii("retyr"), &["retry"]),
            (dictgen::InsensitiveStr::Ascii("exersized"), &["exercised"]),
            (
                dictgen::InsensitiveStr::Ascii("brazileans"),
                &["brazilians"],
            ),
            (dictgen::InsensitiveStr::Ascii("xdescribe"), &["describe"]),
            (dictgen::InsensitiveStr::Ascii("sommewhere"), &["somewhere"]),
            (
                dictgen::InsensitiveStr::Ascii("coutnering"),
                &["countering"],
            ),
            (dictgen::InsensitiveStr::Ascii("eyesahdow"), &["eyeshadow"]),
            (
                dictgen::InsensitiveStr::Ascii("creationsit"),
                &["creationist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capitilize"),
                &["capitalize"],
            ),
            (dictgen::InsensitiveStr::Ascii("prarams"), &["params"]),
            (dictgen::InsensitiveStr::Ascii("annoucne"), &["announce"]),
            (dictgen::InsensitiveStr::Ascii("buuild"), &["build"]),
            (
                dictgen::InsensitiveStr::Ascii("exectutions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("hypokrites"),
                &["hypocrites"],
            ),
            (dictgen::InsensitiveStr::Ascii("partisain"), &["partisan"]),
            (dictgen::InsensitiveStr::Ascii("identties"), &["identities"]),
            (
                dictgen::InsensitiveStr::Ascii("recendable"),
                &["rescindable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disabiltitiy"),
                &["disability"],
            ),
            (dictgen::InsensitiveStr::Ascii("humoural"), &["humoral"]),
            (
                dictgen::InsensitiveStr::Ascii("substituded"),
                &["substituted"],
            ),
            (dictgen::InsensitiveStr::Ascii("ocupy"), &["occupy"]),
            (dictgen::InsensitiveStr::Ascii("cilanto"), &["cilantro"]),
            (dictgen::InsensitiveStr::Ascii("dircetive"), &["directive"]),
            (
                dictgen::InsensitiveStr::Ascii("algorythmic"),
                &["algorithmic"],
            ),
            (dictgen::InsensitiveStr::Ascii("concicely"), &["concisely"]),
            (dictgen::InsensitiveStr::Ascii("openess"), &["openness"]),
            (
                dictgen::InsensitiveStr::Ascii("authenticted"),
                &["authenticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("tuhmbs"), &["thumbs"]),
            (dictgen::InsensitiveStr::Ascii("refelcted"), &["reflected"]),
            (dictgen::InsensitiveStr::Ascii("turncate"), &["truncate"]),
            (dictgen::InsensitiveStr::Ascii("heistant"), &["hesitant"]),
            (
                dictgen::InsensitiveStr::Ascii("combinato"),
                &["combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("submdoule"), &["submodule"]),
            (
                dictgen::InsensitiveStr::Ascii("intitialized"),
                &["initialized"],
            ),
            (dictgen::InsensitiveStr::Ascii("engineeers"), &["engineers"]),
            (
                dictgen::InsensitiveStr::Ascii("interefers"),
                &["interferes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unbeleifable"),
                &["unbelievable"],
            ),
            (dictgen::InsensitiveStr::Ascii("sitrring"), &["stirring"]),
            (
                dictgen::InsensitiveStr::Ascii("preccedence"),
                &["precedence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("qeuestioning"),
                &["questioning"],
            ),
            (dictgen::InsensitiveStr::Ascii("cabbagge"), &["cabbage"]),
            (
                dictgen::InsensitiveStr::Ascii("predetermind"),
                &["predetermined"],
            ),
            (dictgen::InsensitiveStr::Ascii("proffess"), &["profess"]),
            (
                dictgen::InsensitiveStr::Ascii("incomaptible"),
                &["incompatible"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("creaters"),
                &["craters", "creators", "creates"],
            ),
            (dictgen::InsensitiveStr::Ascii("divicer"), &["divider"]),
            (
                dictgen::InsensitiveStr::Ascii("palythroughs"),
                &["playthroughs"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcluate"), &["calculate"]),
            (
                dictgen::InsensitiveStr::Ascii("aprroximation"),
                &["approximation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mermory"), &["memory"]),
            (dictgen::InsensitiveStr::Ascii("createions"), &["creations"]),
            (
                dictgen::InsensitiveStr::Ascii("heartbrakes"),
                &["heartbreak"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("circumvernt"),
                &["circumvent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notificatin"),
                &["notification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenaticating"),
                &["authenticating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("corrsepondents"),
                &["correspondents"],
            ),
            (dictgen::InsensitiveStr::Ascii("attennuate"), &["attenuate"]),
            (dictgen::InsensitiveStr::Ascii("addrersser"), &["addresser"]),
            (dictgen::InsensitiveStr::Ascii("biuldings"), &["buildings"]),
            (
                dictgen::InsensitiveStr::Ascii("indifferance"),
                &["indifference"],
            ),
            (dictgen::InsensitiveStr::Ascii("utillities"), &["utilities"]),
            (dictgen::InsensitiveStr::Ascii("keesh"), &["quiche"]),
            (
                dictgen::InsensitiveStr::Ascii("mistatching"),
                &["mismatching"],
            ),
            (dictgen::InsensitiveStr::Ascii("specisl"), &["specials"]),
            (
                dictgen::InsensitiveStr::Ascii("policitians"),
                &["politicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("bacisally"), &["basically"]),
            (
                dictgen::InsensitiveStr::Ascii("impleting"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sponsorhip"),
                &["sponsorship"],
            ),
            (dictgen::InsensitiveStr::Ascii("uploded"), &["uploaded"]),
            (dictgen::InsensitiveStr::Ascii("benetifed"), &["benefited"]),
            (
                dictgen::InsensitiveStr::Ascii("parallising"),
                &["parallelising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("declartative"),
                &["declarative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("determimnes"),
                &["determines"],
            ),
            (dictgen::InsensitiveStr::Ascii("inprooving"), &["improving"]),
            (dictgen::InsensitiveStr::Ascii("contais"), &["contains"]),
            (dictgen::InsensitiveStr::Ascii("penalites"), &["penalties"]),
            (
                dictgen::InsensitiveStr::Ascii("iterpolated"),
                &["interpolated"],
            ),
            (dictgen::InsensitiveStr::Ascii("abosrbing"), &["absorbing"]),
            (
                dictgen::InsensitiveStr::Ascii("inexpeirenced"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("bisuness"), &["business"]),
            (dictgen::InsensitiveStr::Ascii("schedual"), &["schedule"]),
            (
                dictgen::InsensitiveStr::Ascii("alhpabetic"),
                &["alphabetic"],
            ),
            (dictgen::InsensitiveStr::Ascii("absolure"), &["absolute"]),
            (
                dictgen::InsensitiveStr::Ascii("genererator"),
                &["generator"],
            ),
            (dictgen::InsensitiveStr::Ascii("methdo"), &["method"]),
            (
                dictgen::InsensitiveStr::Ascii("smarthpones"),
                &["smartphones"],
            ),
            (dictgen::InsensitiveStr::Ascii("preccedes"), &["precedes"]),
            (dictgen::InsensitiveStr::Ascii("registsers"), &["registers"]),
            (
                dictgen::InsensitiveStr::Ascii("parenthesys"),
                &["parentheses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comutability"),
                &["commutability", "computability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("savves"),
                &["salves", "saves"],
            ),
            (dictgen::InsensitiveStr::Ascii("gererates"), &["generates"]),
            (dictgen::InsensitiveStr::Ascii("reuslts"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("arrangaeble"),
                &["arrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("modidfies"), &["modifies"]),
            (dictgen::InsensitiveStr::Ascii("bnack"), &["back"]),
            (
                dictgen::InsensitiveStr::Ascii("volentiers"),
                &["volunteers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subjektive"),
                &["subjective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("movebackwrd"),
                &["movebackward"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("outsourcad"),
                &["outsourced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instutitionalized"),
                &["institutionalized"],
            ),
            (dictgen::InsensitiveStr::Ascii("elswhere"), &["elsewhere"]),
            (dictgen::InsensitiveStr::Ascii("troleld"), &["trolled"]),
            (dictgen::InsensitiveStr::Ascii("fragmenet"), &["fragment"]),
            (
                dictgen::InsensitiveStr::Ascii("chainging"),
                &["changing", "chaining"],
            ),
            (dictgen::InsensitiveStr::Ascii("specifieed"), &["specified"]),
            (
                dictgen::InsensitiveStr::Ascii("aletrnates"),
                &["alternates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("essentally"),
                &["essentially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supportd"),
                &["supported", "supports"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adpatations"),
                &["adaptations"],
            ),
            (dictgen::InsensitiveStr::Ascii("prety"), &["pretty", "prey"]),
            (
                dictgen::InsensitiveStr::Ascii("partititions"),
                &["partitions"],
            ),
            (dictgen::InsensitiveStr::Ascii("setted"), &["set"]),
            (dictgen::InsensitiveStr::Ascii("paksitani"), &["pakistani"]),
            (dictgen::InsensitiveStr::Ascii("nullalble"), &["nullable"]),
            (
                dictgen::InsensitiveStr::Ascii("elipses"),
                &["ellipses", "eclipses", "ellipsis"],
            ),
            (dictgen::InsensitiveStr::Ascii("conotrol"), &["control"]),
            (
                dictgen::InsensitiveStr::Ascii("origininated"),
                &["originated"],
            ),
            (dictgen::InsensitiveStr::Ascii("gallary"), &["gallery"]),
            (dictgen::InsensitiveStr::Ascii("volonteer"), &["volunteer"]),
            (dictgen::InsensitiveStr::Ascii("pleanty"), &["plenty"]),
            (
                dictgen::InsensitiveStr::Ascii("argessively"),
                &["aggressively"],
            ),
            (dictgen::InsensitiveStr::Ascii("trnasmit"), &["transmit"]),
            (
                dictgen::InsensitiveStr::Ascii("shystemmemory"),
                &["systemmemory"],
            ),
            (dictgen::InsensitiveStr::Ascii("pertoleum"), &["petroleum"]),
            (
                dictgen::InsensitiveStr::Ascii("troglidite"),
                &["troglodyte"],
            ),
            (dictgen::InsensitiveStr::Ascii("dubly"), &["doubly"]),
            (
                dictgen::InsensitiveStr::Ascii("replasements"),
                &["replacements"],
            ),
            (dictgen::InsensitiveStr::Ascii("objecct"), &["object"]),
            (dictgen::InsensitiveStr::Ascii("trrigger"), &["trigger"]),
            (dictgen::InsensitiveStr::Ascii("ampilfy"), &["amplify"]),
            (dictgen::InsensitiveStr::Ascii("consulant"), &["consultant"]),
            (dictgen::InsensitiveStr::Ascii("thorwing"), &["throwing"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteramer"),
                &["downstreamer"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("visiblities"),
                &["visibilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("subtak"), &["subtask"]),
            (dictgen::InsensitiveStr::Ascii("stablizer"), &["stabilizer"]),
            (dictgen::InsensitiveStr::Ascii("romaing"), &["roaming"]),
            (dictgen::InsensitiveStr::Ascii("cofigured"), &["configured"]),
            (
                dictgen::InsensitiveStr::Ascii("combiantor"),
                &["combinator"],
            ),
            (dictgen::InsensitiveStr::Ascii("spoace"), &["space"]),
            (dictgen::InsensitiveStr::Ascii("porvide"), &["provide"]),
            (dictgen::InsensitiveStr::Ascii("loockdown"), &["lockdown"]),
            (dictgen::InsensitiveStr::Ascii("foregin"), &["foreign"]),
            (
                dictgen::InsensitiveStr::Ascii("architecutral"),
                &["architectural"],
            ),
            (dictgen::InsensitiveStr::Ascii("argumant"), &["argument"]),
            (dictgen::InsensitiveStr::Ascii("poweful"), &["powerful"]),
            (
                dictgen::InsensitiveStr::Ascii("assasinations"),
                &["assassinations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("guarentyinging"),
                &["guaranteeing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("exlicitely"),
                &["explicitly"],
            ),
            (dictgen::InsensitiveStr::Ascii("conusmer"), &["consumer"]),
            (
                dictgen::InsensitiveStr::Ascii("enceclopedia"),
                &["encyclopedia"],
            ),
            (dictgen::InsensitiveStr::Ascii("cotten"), &["cotton"]),
            (
                dictgen::InsensitiveStr::Ascii("proportianally"),
                &["proportionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("mimicks"), &["mimics"]),
            (dictgen::InsensitiveStr::Ascii("sigrynge"), &["syringe"]),
            (
                dictgen::InsensitiveStr::Ascii("espacially"),
                &["especially"],
            ),
            (dictgen::InsensitiveStr::Ascii("terriffic"), &["terrific"]),
            (dictgen::InsensitiveStr::Ascii("qeuried"), &["queried"]),
            (dictgen::InsensitiveStr::Ascii("falg"), &["flag"]),
            (
                dictgen::InsensitiveStr::Ascii("unbreakabke"),
                &["unbreakable"],
            ),
            (dictgen::InsensitiveStr::Ascii("obhection"), &["objection"]),
            (dictgen::InsensitiveStr::Ascii("meber"), &["member"]),
            (
                dictgen::InsensitiveStr::Ascii("algorwwithms"),
                &["algorithms"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cicles"),
                &["cycles", "circles", "icicles"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppourtunity"),
                &["opportunity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("questoined"),
                &["questioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("frition"), &["friction"]),
            (
                dictgen::InsensitiveStr::Ascii("separartors"),
                &["separators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("complation"),
                &["completion", "compilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("interger"), &["integer"]),
            (dictgen::InsensitiveStr::Ascii("presentes"), &["presents"]),
            (
                dictgen::InsensitiveStr::Ascii("implementatoins"),
                &["implementations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("serialzing"),
                &["serializing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sepcifically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("touchscren"),
                &["touchscreen"],
            ),
            (dictgen::InsensitiveStr::Ascii("currecy"), &["currency"]),
            (
                dictgen::InsensitiveStr::Ascii("schisophrenic"),
                &["schizophrenic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("respecively"),
                &["respectively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adaped"),
                &["adapted", "adapt", "adopted", "adopt"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("controception"),
                &["contraception"],
            ),
            (dictgen::InsensitiveStr::Ascii("psuedo"), &["pseudo"]),
            (dictgen::InsensitiveStr::Ascii("encosure"), &["enclosure"]),
            (
                dictgen::InsensitiveStr::Ascii("opportunties"),
                &["opportunities"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suceessfully"),
                &["successfully"],
            ),
            (dictgen::InsensitiveStr::Ascii("cancellin"), &["cancelling"]),
            (dictgen::InsensitiveStr::Ascii("conditon"), &["condition"]),
            (dictgen::InsensitiveStr::Ascii("unmodifed"), &["unmodified"]),
            (
                dictgen::InsensitiveStr::Ascii("liklelihood"),
                &["likelihood"],
            ),
            (dictgen::InsensitiveStr::Ascii("liek"), &["like"]),
            (dictgen::InsensitiveStr::Ascii("fluctuatie"), &["fluctuate"]),
            (dictgen::InsensitiveStr::Ascii("insted"), &["instead"]),
            (
                dictgen::InsensitiveStr::Ascii("partecipants"),
                &["participants"],
            ),
            (dictgen::InsensitiveStr::Ascii("hefers"), &["heifers"]),
            (
                dictgen::InsensitiveStr::Ascii("pressuming"),
                &["pressuring", "presuming"],
            ),
            (dictgen::InsensitiveStr::Ascii("glassses"), &["glasses"]),
            (dictgen::InsensitiveStr::Ascii("colmn"), &["column"]),
            (dictgen::InsensitiveStr::Ascii("unoin"), &["union"]),
            (
                dictgen::InsensitiveStr::Ascii("improvision"),
                &["improvisation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("epected"),
                &["expected", "elected"],
            ),
            (dictgen::InsensitiveStr::Ascii("complets"), &["completes"]),
            (dictgen::InsensitiveStr::Ascii("encrytped"), &["encrypted"]),
            (
                dictgen::InsensitiveStr::Ascii("prolateriat"),
                &["proletariat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cusomizable"),
                &["customizable"],
            ),
            (dictgen::InsensitiveStr::Ascii("cusomers"), &["customers"]),
            (dictgen::InsensitiveStr::Ascii("thm"), &["them"]),
            (dictgen::InsensitiveStr::Ascii("caf"), &["calf"]),
            (dictgen::InsensitiveStr::Ascii("geometrie"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("resquest"), &["request"]),
            (
                dictgen::InsensitiveStr::Ascii("unsubcribe"),
                &["unsubscribe"],
            ),
            (dictgen::InsensitiveStr::Ascii("accontant"), &["accountant"]),
            (
                dictgen::InsensitiveStr::Ascii("recommenting"),
                &["recommending"],
            ),
            (dictgen::InsensitiveStr::Ascii("refferral"), &["referral"]),
            (dictgen::InsensitiveStr::Ascii("veiwpoint"), &["viewpoint"]),
            (
                dictgen::InsensitiveStr::Ascii("transparren"),
                &["transparent"],
            ),
            (dictgen::InsensitiveStr::Ascii("corupt"), &["corrupt"]),
            (dictgen::InsensitiveStr::Ascii("lising"), &["listing"]),
            (
                dictgen::InsensitiveStr::Ascii("seplicurlly"),
                &["sepulchrally"],
            ),
            (dictgen::InsensitiveStr::Ascii("agrements"), &["agreements"]),
            (
                dictgen::InsensitiveStr::Ascii("documnetation"),
                &["documentation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pagin"), &["paging"]),
            (dictgen::InsensitiveStr::Ascii("cerate"), &["create"]),
            (dictgen::InsensitiveStr::Ascii("accopunt"), &["account"]),
            (
                dictgen::InsensitiveStr::Ascii("depdencency"),
                &["dependency"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nerative"),
                &["narrative", "negative"],
            ),
            (dictgen::InsensitiveStr::Ascii("patriarcy"), &["patriarchy"]),
            (dictgen::InsensitiveStr::Ascii("mutatios"), &["mutations"]),
            (dictgen::InsensitiveStr::Ascii("airbore"), &["airborne"]),
            (
                dictgen::InsensitiveStr::Ascii("pregresses"),
                &["progresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recommendeds"),
                &["recommends"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("misognyist"),
                &["misogynist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncertanty"),
                &["uncertainty"],
            ),
            (dictgen::InsensitiveStr::Ascii("repalce"), &["replace"]),
            (
                dictgen::InsensitiveStr::Ascii("proletariant"),
                &["proletariat"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("councilers"),
                &["councils", "councilors"],
            ),
            (dictgen::InsensitiveStr::Ascii("percious"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("instnances"), &["instances"]),
            (dictgen::InsensitiveStr::Ascii("streigtly"), &["straightly"]),
            (dictgen::InsensitiveStr::Ascii("explitly"), &["explicitly"]),
            (
                dictgen::InsensitiveStr::Ascii("analystical"),
                &["analytical"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempopary"), &["temporary"]),
            (dictgen::InsensitiveStr::Ascii("direcive"), &["directive"]),
            (dictgen::InsensitiveStr::Ascii("sysadmn"), &["sysadmin"]),
            (
                dictgen::InsensitiveStr::Ascii("incorprates"),
                &["incorporates"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoeries"), &["theories"]),
            (dictgen::InsensitiveStr::Ascii("corcodile"), &["crocodile"]),
            (dictgen::InsensitiveStr::Ascii("otherwice"), &["otherwise"]),
            (dictgen::InsensitiveStr::Ascii("freamework"), &["framework"]),
            (
                dictgen::InsensitiveStr::Ascii("hyptothetical"),
                &["hypothetical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intialzer"),
                &["initializer"],
            ),
            (dictgen::InsensitiveStr::Ascii("frecuents"), &["frequents"]),
            (dictgen::InsensitiveStr::Ascii("rademtion"), &["redemption"]),
            (
                dictgen::InsensitiveStr::Ascii("springfeld"),
                &["springfield"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enumaration"),
                &["enumeration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narcassist"),
                &["narcissist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("goemetrically"),
                &["geometrically"],
            ),
            (dictgen::InsensitiveStr::Ascii("sinicaly"), &["cynically"]),
            (
                dictgen::InsensitiveStr::Ascii("confrontativo"),
                &["confrontation"],
            ),
            (dictgen::InsensitiveStr::Ascii("explodie"), &["explode"]),
            (dictgen::InsensitiveStr::Ascii("stuties"), &["studies"]),
            (dictgen::InsensitiveStr::Ascii("analystes"), &["analysts"]),
            (
                dictgen::InsensitiveStr::Ascii("configurtation"),
                &["configuration"],
            ),
            (dictgen::InsensitiveStr::Ascii("opeate"), &["operate"]),
            (
                dictgen::InsensitiveStr::Ascii("matrics"),
                &["matrix", "matrices", "metrics"],
            ),
            (dictgen::InsensitiveStr::Ascii("btton"), &["button"]),
            (dictgen::InsensitiveStr::Ascii("sipports"), &["supports"]),
            (
                dictgen::InsensitiveStr::Ascii("spefifications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("condionals"),
                &["conditionals"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("genearation"),
                &["generation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("vegaterian"),
                &["vegetarian"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preficient"),
                &["proficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sie"),
                &["size", "sigh", "side"],
            ),
            (dictgen::InsensitiveStr::Ascii("sourcee"), &["source"]),
            (dictgen::InsensitiveStr::Ascii("meethod"), &["method"]),
            (dictgen::InsensitiveStr::Ascii("issu"), &["issue"]),
            (
                dictgen::InsensitiveStr::Ascii("mutlivolume"),
                &["multivolume"],
            ),
            (dictgen::InsensitiveStr::Ascii("lnow"), &["know"]),
            (
                dictgen::InsensitiveStr::Ascii("hirarcically"),
                &["hierarchically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dependetly"),
                &["dependently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inititialises"),
                &["initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("opimisation"),
                &["optimisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("spefixics"), &["specifics"]),
            (dictgen::InsensitiveStr::Ascii("forcibley"), &["forcibly"]),
            (dictgen::InsensitiveStr::Ascii("molcules"), &["molecules"]),
            (
                dictgen::InsensitiveStr::Ascii("embarasses"),
                &["embarrasses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unefficient"),
                &["inefficient"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathmaticians"),
                &["mathematicians"],
            ),
            (dictgen::InsensitiveStr::Ascii("vitmain"), &["vitamin"]),
            (
                dictgen::InsensitiveStr::Ascii("ethose"),
                &["those", "ethos"],
            ),
            (dictgen::InsensitiveStr::Ascii("overvise"), &["otherwise"]),
            (
                dictgen::InsensitiveStr::Ascii("northereastern"),
                &["northeastern"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paramenters"),
                &["parameters"],
            ),
            (dictgen::InsensitiveStr::Ascii("progreses"), &["progresses"]),
            (
                dictgen::InsensitiveStr::Ascii("aeactivate"),
                &["deactivate", "activate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("orginasation"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("roomates"), &["roommates"]),
            (dictgen::InsensitiveStr::Ascii("severat"), &["several"]),
            (
                dictgen::InsensitiveStr::Ascii("charecters"),
                &["characters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attempeting"),
                &["attempting"],
            ),
            (dictgen::InsensitiveStr::Ascii("filp"), &["flip"]),
            (dictgen::InsensitiveStr::Ascii("folwoinwg"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("skatebrand"),
                &["skateboard"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negotaties"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ventelation"),
                &["ventilation"],
            ),
            (dictgen::InsensitiveStr::Ascii("oponent"), &["opponent"]),
            (dictgen::InsensitiveStr::Ascii("modiying"), &["modifying"]),
            (dictgen::InsensitiveStr::Ascii("cgywin"), &["cygwin"]),
            (
                dictgen::InsensitiveStr::Ascii("atuhenticates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomtry"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("whch"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("segway"), &["segue"]),
            (dictgen::InsensitiveStr::Ascii("oprhans"), &["orphans"]),
            (
                dictgen::InsensitiveStr::Ascii("sanatizers"),
                &["sanitizers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("discresion"),
                &["discretion"],
            ),
            (dictgen::InsensitiveStr::Ascii("perscius"), &["precious"]),
            (dictgen::InsensitiveStr::Ascii("natioanl"), &["national"]),
            (dictgen::InsensitiveStr::Ascii("obsure"), &["obscure"]),
            (dictgen::InsensitiveStr::Ascii("udno"), &["undo", "uno"]),
            (dictgen::InsensitiveStr::Ascii("cosnider"), &["consider"]),
            (dictgen::InsensitiveStr::Ascii("ignered"), &["ignored"]),
            (dictgen::InsensitiveStr::Ascii("mensioned"), &["mentioned"]),
            (
                dictgen::InsensitiveStr::Ascii("equivlaent"),
                &["equivalents"],
            ),
            (dictgen::InsensitiveStr::Ascii("lineseach"), &["linesearch"]),
            (dictgen::InsensitiveStr::Ascii("traversie"), &["traverse"]),
            (
                dictgen::InsensitiveStr::Ascii("proporcion"),
                &["proportion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("momentarely"),
                &["momentarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("coallates"), &["collates"]),
            (dictgen::InsensitiveStr::Ascii("reveiwers"), &["reviewers"]),
            (
                dictgen::InsensitiveStr::Ascii("progressioin"),
                &["progressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defererence"),
                &["dereference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("innocentius"),
                &["innocents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incompatablility"),
                &["incompatibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("namme"), &["name"]),
            (
                dictgen::InsensitiveStr::Ascii("inconstencies"),
                &["inconsistencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("particlar"), &["particular"]),
            (
                dictgen::InsensitiveStr::Ascii("iterstions"),
                &["iterations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initialyse"),
                &["initialise"],
            ),
            (dictgen::InsensitiveStr::Ascii("atitude"), &["attitude"]),
            (
                dictgen::InsensitiveStr::Ascii("negotiotates"),
                &["negotiates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appraochable"),
                &["approachable"],
            ),
            (dictgen::InsensitiveStr::Ascii("disgustes"), &["disgusts"]),
            (
                dictgen::InsensitiveStr::Ascii("rotatios"),
                &["rotations", "ratios"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transposin"),
                &["transposing"],
            ),
            (dictgen::InsensitiveStr::Ascii("prouldy"), &["proudly"]),
            (dictgen::InsensitiveStr::Ascii("inproving"), &["improving"]),
            (
                dictgen::InsensitiveStr::Ascii("unbeareble"),
                &["unbearable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infograhpic"),
                &["infographic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deffered"),
                &["differed", "deferred"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("abberations"),
                &["aberrations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distustingly"),
                &["disgustingly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresgistered"),
                &["unregistered"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("indifferente"),
                &["indifference"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("transparrenly"),
                &["transparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("virualizing"),
                &["virtualizing", "visualizing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("delevopers"),
                &["developers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("overcompansation"),
                &["overcompensation"],
            ),
            (dictgen::InsensitiveStr::Ascii("embeddders"), &["embedders"]),
            (
                dictgen::InsensitiveStr::Ascii("correponsing"),
                &["corresponding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("silouetist"),
                &["silhouettist"],
            ),
            (dictgen::InsensitiveStr::Ascii("benechmark"), &["benchmark"]),
            (
                dictgen::InsensitiveStr::Ascii("overridded"),
                &["overrode", "overridden"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("resposiveness"),
                &["responsiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("iniatives"),
                &["initiatives"],
            ),
            (dictgen::InsensitiveStr::Ascii("orgainsed"), &["organised"]),
            (
                dictgen::InsensitiveStr::Ascii("unessecary"),
                &["unnecessary"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instrctional"),
                &["instructional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comparisino"),
                &["comparison"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("downvotting"),
                &["downvoting"],
            ),
            (dictgen::InsensitiveStr::Ascii("doubely"), &["doubly"]),
            (dictgen::InsensitiveStr::Ascii("connet"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("imidiately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unexpepected"),
                &["unexpected"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("impelmentation"),
                &["implementations", "implementation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pupups"), &["popups"]),
            (
                dictgen::InsensitiveStr::Ascii("speficialleir"),
                &["specifier"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("efined"),
                &["defined", "refined"],
            ),
            (dictgen::InsensitiveStr::Ascii("mananger"), &["manager"]),
            (dictgen::InsensitiveStr::Ascii("accquires"), &["acquires"]),
            (dictgen::InsensitiveStr::Ascii("finishied"), &["finished"]),
            (dictgen::InsensitiveStr::Ascii("scarmbled"), &["scrambled"]),
            (dictgen::InsensitiveStr::Ascii("anlaysed"), &["analysed"]),
            (dictgen::InsensitiveStr::Ascii("burtally"), &["brutally"]),
            (
                dictgen::InsensitiveStr::Ascii("appreciationg"),
                &["appreciating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commendments"),
                &["commandments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dismanteld"),
                &["dismantled"],
            ),
            (dictgen::InsensitiveStr::Ascii("sotryline"), &["storyline"]),
            (dictgen::InsensitiveStr::Ascii("nationella"), &["national"]),
            (dictgen::InsensitiveStr::Ascii("arbitraty"), &["arbitrary"]),
            (
                dictgen::InsensitiveStr::Ascii("multidimensionnal"),
                &["multidimensional"],
            ),
            (dictgen::InsensitiveStr::Ascii("dupplicate"), &["duplicate"]),
            (
                dictgen::InsensitiveStr::Ascii("unverfified"),
                &["unverified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calclulations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsubscritpion"),
                &["unsubscription"],
            ),
            (dictgen::InsensitiveStr::Ascii("framwework"), &["framework"]),
            (dictgen::InsensitiveStr::Ascii("appply"), &["apply"]),
            (dictgen::InsensitiveStr::Ascii("accepets"), &["accepts"]),
            (
                dictgen::InsensitiveStr::Ascii("inocorrectly"),
                &["incorrectly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("superstiton"),
                &["superstition"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyosks"), &["kiosks"]),
            (dictgen::InsensitiveStr::Ascii("charator"), &["character"]),
            (
                dictgen::InsensitiveStr::Ascii("distorsional"),
                &["distortional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("elctricity"),
                &["electricity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("civalazation"),
                &["civilization"],
            ),
            (dictgen::InsensitiveStr::Ascii("herlo"), &["hello"]),
            (dictgen::InsensitiveStr::Ascii("cnter"), &["center"]),
            (dictgen::InsensitiveStr::Ascii("currence"), &["currencies"]),
            (dictgen::InsensitiveStr::Ascii("sempaphore"), &["semaphore"]),
            (
                dictgen::InsensitiveStr::Ascii("assocition"),
                &["association"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oversubscibed"),
                &["oversubscribed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("curroption"),
                &["corruption"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bobard"),
                &["board", "bombard"],
            ),
            (dictgen::InsensitiveStr::Ascii("apapting"), &["adapting"]),
            (dictgen::InsensitiveStr::Ascii("porvider"), &["provider"]),
            (dictgen::InsensitiveStr::Ascii("deamonise"), &["daemonise"]),
            (
                dictgen::InsensitiveStr::Ascii("expendetures"),
                &["expenditures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatanate"),
                &["concatenate"],
            ),
            (dictgen::InsensitiveStr::Ascii("linnaena"), &["linnaean"]),
            (
                dictgen::InsensitiveStr::Ascii("genreators"),
                &["generators"],
            ),
            (dictgen::InsensitiveStr::Ascii("resourses"), &["resources"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasutical"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consitution"),
                &["constitution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bivouacing"),
                &["bivouacking"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("demolitian"),
                &["demolition"],
            ),
            (dictgen::InsensitiveStr::Ascii("ananlyzes"), &["analyzes"]),
            (dictgen::InsensitiveStr::Ascii("stratergy"), &["strategy"]),
            (
                dictgen::InsensitiveStr::Ascii("injustaces"),
                &["injustices"],
            ),
            (dictgen::InsensitiveStr::Ascii("trotzky"), &["trotsky"]),
            (dictgen::InsensitiveStr::Ascii("spwans"), &["spawns"]),
            (dictgen::InsensitiveStr::Ascii("enfocers"), &["enforcers"]),
            (dictgen::InsensitiveStr::Ascii("revisted"), &["revisited"]),
            (dictgen::InsensitiveStr::Ascii("predicitor"), &["predictor"]),
            (
                dictgen::InsensitiveStr::Ascii("spefically"),
                &["specifically"],
            ),
            (dictgen::InsensitiveStr::Ascii("pluigns"), &["plugins"]),
            (
                dictgen::InsensitiveStr::Ascii("assumuption"),
                &["assumption"],
            ),
            (dictgen::InsensitiveStr::Ascii("whehther"), &["whether"]),
            (dictgen::InsensitiveStr::Ascii("minggw"), &["mingw"]),
            (dictgen::InsensitiveStr::Ascii("postions"), &["positions"]),
            (dictgen::InsensitiveStr::Ascii("collaspe"), &["collapse"]),
            (
                dictgen::InsensitiveStr::Ascii("attrributes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("empyty"), &["empty"]),
            (dictgen::InsensitiveStr::Ascii("prupose"), &["purpose"]),
            (dictgen::InsensitiveStr::Ascii("agreeeably"), &["agreeably"]),
            (
                dictgen::InsensitiveStr::Ascii("reculrively"),
                &["recursively"],
            ),
            (dictgen::InsensitiveStr::Ascii("trhough"), &["through"]),
            (
                dictgen::InsensitiveStr::Ascii("encupsulates"),
                &["encapsulates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explnations"),
                &["explanations"],
            ),
            (dictgen::InsensitiveStr::Ascii("numner"), &["number"]),
            (
                dictgen::InsensitiveStr::Ascii("snythesising"),
                &["synthesising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspomdence"),
                &["correspondence"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("consistentcy"),
                &["consistently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compositon"),
                &["composition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("implenments"),
                &["implements"],
            ),
            (dictgen::InsensitiveStr::Ascii("thearpy"), &["therapy"]),
            (dictgen::InsensitiveStr::Ascii("sproels"), &["sproles"]),
            (dictgen::InsensitiveStr::Ascii("crocodille"), &["crocodile"]),
            (
                dictgen::InsensitiveStr::Ascii("inintializes"),
                &["initializes"],
            ),
            (dictgen::InsensitiveStr::Ascii("devialet"), &["deviate"]),
            (dictgen::InsensitiveStr::Ascii("operatio"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("asnd"), &["and"]),
            (dictgen::InsensitiveStr::Ascii("sewdonim"), &["pseudonym"]),
            (dictgen::InsensitiveStr::Ascii("relgion"), &["religion"]),
            (dictgen::InsensitiveStr::Ascii("weridos"), &["weirdos"]),
            (
                dictgen::InsensitiveStr::Ascii("disaapoint"),
                &["disappoint"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intimitely"),
                &["intimately"],
            ),
            (dictgen::InsensitiveStr::Ascii("porcelian"), &["porcelain"]),
            (
                dictgen::InsensitiveStr::Ascii("downsteamed"),
                &["downstreamed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authorotative"),
                &["authoritative"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inmediately"),
                &["immediately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cheesebuger"),
                &["cheeseburger"],
            ),
            (dictgen::InsensitiveStr::Ascii("verything"), &["everything"]),
            (dictgen::InsensitiveStr::Ascii("excceds"), &["exceeds"]),
            (dictgen::InsensitiveStr::Ascii("ostridges"), &["ostriches"]),
            (dictgen::InsensitiveStr::Ascii("fitlers"), &["filters"]),
            (
                dictgen::InsensitiveStr::Ascii("coordingly"),
                &["accordingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("baordwalk"), &["boardwalk"]),
            (dictgen::InsensitiveStr::Ascii("elemnt"), &["element"]),
            (
                dictgen::InsensitiveStr::Ascii("luminousity"),
                &["luminosity"],
            ),
            (dictgen::InsensitiveStr::Ascii("forard"), &["forward"]),
            (
                dictgen::InsensitiveStr::Ascii("identifyable"),
                &["identifiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apprecaite"),
                &["appreciate"],
            ),
            (dictgen::InsensitiveStr::Ascii("wich"), &["which"]),
            (dictgen::InsensitiveStr::Ascii("dependedn"), &["dependent"]),
            (dictgen::InsensitiveStr::Ascii("implenets"), &["implements"]),
            (
                dictgen::InsensitiveStr::Ascii("audovisual"),
                &["audiovisual"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simpliy"),
                &["simplify", "simply"],
            ),
            (dictgen::InsensitiveStr::Ascii("analtyics"), &["analytics"]),
            (dictgen::InsensitiveStr::Ascii("scrolbar"), &["scrollbar"]),
            (
                dictgen::InsensitiveStr::Ascii("reinstallled"),
                &["reinstalled"],
            ),
            (dictgen::InsensitiveStr::Ascii("corrolate"), &["correlate"]),
            (dictgen::InsensitiveStr::Ascii("gautnlet"), &["gauntlet"]),
            (dictgen::InsensitiveStr::Ascii("threadad"), &["threaded"]),
            (
                dictgen::InsensitiveStr::Ascii("misunderstandig"),
                &["misunderstandings"],
            ),
            (dictgen::InsensitiveStr::Ascii("memcahed"), &["memcached"]),
            (dictgen::InsensitiveStr::Ascii("numvers"), &["numbers"]),
            (
                dictgen::InsensitiveStr::Ascii("trnsfered"),
                &["transferred"],
            ),
            (dictgen::InsensitiveStr::Ascii("genarated"), &["generated"]),
            (dictgen::InsensitiveStr::Ascii("emplyer"), &["employer"]),
            (dictgen::InsensitiveStr::Ascii("loaader"), &["loader"]),
            (dictgen::InsensitiveStr::Ascii("versitile"), &["versatile"]),
            (
                dictgen::InsensitiveStr::Ascii("pharmasutically"),
                &["pharmaceutical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roundabaout"),
                &["roundabout"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("diretories"),
                &["directories"],
            ),
            (dictgen::InsensitiveStr::Ascii("famliliar"), &["familiar"]),
            (dictgen::InsensitiveStr::Ascii("gemoetry"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("leggacies"), &["legacies"]),
            (
                dictgen::InsensitiveStr::Ascii("fictionaries"),
                &["dictionaries"],
            ),
            (dictgen::InsensitiveStr::Ascii("sticthed"), &["stitched"]),
            (dictgen::InsensitiveStr::Ascii("fusha"), &["fuchsia"]),
            (dictgen::InsensitiveStr::Ascii("sistem"), &["system"]),
            (dictgen::InsensitiveStr::Ascii("lcarify"), &["clarify"]),
            (dictgen::InsensitiveStr::Ascii("remeniss"), &["reminisce"]),
            (
                dictgen::InsensitiveStr::Ascii("nationalits"),
                &["nationalist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contstraints"),
                &["constraints"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kryptoninte"),
                &["kryptonite"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("residental"),
                &["residential"],
            ),
            (dictgen::InsensitiveStr::Ascii("launhes"), &["launches"]),
            (dictgen::InsensitiveStr::Ascii("crticals"), &["criticals"]),
            (
                dictgen::InsensitiveStr::Ascii("execuitions"),
                &["executions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recalcualting"),
                &["recalculating"],
            ),
            (dictgen::InsensitiveStr::Ascii("srtifacts"), &["artifacts"]),
            (
                dictgen::InsensitiveStr::Ascii("crossfiters"),
                &["crossfire"],
            ),
            (dictgen::InsensitiveStr::Ascii("ficks"), &["flicks", "fix"]),
            (dictgen::InsensitiveStr::Ascii("buddhits"), &["buddhist"]),
            (
                dictgen::InsensitiveStr::Ascii("reportadly"),
                &["reportedly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("synchronisin"),
                &["synchronising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tranlation"),
                &["translation"],
            ),
            (dictgen::InsensitiveStr::Ascii("reapeted"), &["repeated"]),
            (
                dictgen::InsensitiveStr::Ascii("eneumeration"),
                &["enumeration"],
            ),
            (dictgen::InsensitiveStr::Ascii("milesone"), &["milestone"]),
            (dictgen::InsensitiveStr::Ascii("trik"), &["trick", "trike"]),
            (dictgen::InsensitiveStr::Ascii("occucence"), &["occurrence"]),
            (dictgen::InsensitiveStr::Ascii("deprecatd"), &["deprecated"]),
            (dictgen::InsensitiveStr::Ascii("harnomics"), &["harmonics"]),
            (dictgen::InsensitiveStr::Ascii("kiyack"), &["kayak"]),
            (dictgen::InsensitiveStr::Ascii("behaivors"), &["behaviors"]),
            (dictgen::InsensitiveStr::Ascii("symptoom"), &["symptom"]),
            (dictgen::InsensitiveStr::Ascii("typs"), &["typos"]),
            (
                dictgen::InsensitiveStr::Ascii("implementaiton"),
                &["implementation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("subnegotionation"),
                &["subnegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("boolan"), &["boolean"]),
            (dictgen::InsensitiveStr::Ascii("operattion"), &["operation"]),
            (dictgen::InsensitiveStr::Ascii("intiallly"), &["initially"]),
            (
                dictgen::InsensitiveStr::Ascii("needleslly"),
                &["needlessly"],
            ),
            (dictgen::InsensitiveStr::Ascii("prefixe"), &["prefix"]),
            (
                dictgen::InsensitiveStr::Ascii("systematizin"),
                &["systematizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("adheisve"), &["adhesive"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaible"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("desperatelly"),
                &["desperately"],
            ),
            (dictgen::InsensitiveStr::Ascii("acsess"), &["access"]),
            (dictgen::InsensitiveStr::Ascii("argumernts"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("exipration"),
                &["expiration"],
            ),
            (dictgen::InsensitiveStr::Ascii("btween"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("peplica"), &["replica"]),
            (
                dictgen::InsensitiveStr::Ascii("associaton"),
                &["association"],
            ),
            (dictgen::InsensitiveStr::Ascii("progrss"), &["progress"]),
            (
                dictgen::InsensitiveStr::Ascii("descriptots"),
                &["descriptors"],
            ),
            (dictgen::InsensitiveStr::Ascii("freedums"), &["freedoms"]),
            (
                dictgen::InsensitiveStr::Ascii("disastorus"),
                &["disastrous"],
            ),
            (dictgen::InsensitiveStr::Ascii("ommand"), &["command"]),
            (dictgen::InsensitiveStr::Ascii("visibl"), &["visible"]),
            (
                dictgen::InsensitiveStr::Ascii("applicaition"),
                &["application"],
            ),
            (dictgen::InsensitiveStr::Ascii("summaried"), &["summarized"]),
            (
                dictgen::InsensitiveStr::Ascii("caracterises"),
                &["characterises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("sandwitched"),
                &["sandwiched"],
            ),
            (dictgen::InsensitiveStr::Ascii("archves"), &["archives"]),
            (
                dictgen::InsensitiveStr::Ascii("systemetically"),
                &["systematically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("capactiors"),
                &["capacitors"],
            ),
            (dictgen::InsensitiveStr::Ascii("momenet"), &["moment"]),
            (dictgen::InsensitiveStr::Ascii("sybling"), &["sibling"]),
            (dictgen::InsensitiveStr::Ascii("scandlas"), &["scandals"]),
            (
                dictgen::InsensitiveStr::Ascii("initialialize"),
                &["initialize"],
            ),
            (dictgen::InsensitiveStr::Ascii("clarety"), &["clarity"]),
            (
                dictgen::InsensitiveStr::Ascii("worksapces"),
                &["workspaces"],
            ),
            (dictgen::InsensitiveStr::Ascii("delting"), &["deleting"]),
            (
                dictgen::InsensitiveStr::Ascii("colelctors"),
                &["collectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("whithout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("parapharse"),
                &["paraphrase"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("carcas"),
                &["carcass", "caracas"],
            ),
            (dictgen::InsensitiveStr::Ascii("matser"), &["master"]),
            (dictgen::InsensitiveStr::Ascii("follfowong"), &["following"]),
            (
                dictgen::InsensitiveStr::Ascii("consdiered"),
                &["considered"],
            ),
            (dictgen::InsensitiveStr::Ascii("javacript"), &["javascript"]),
            (dictgen::InsensitiveStr::Ascii("listes"), &["listed"]),
            (
                dictgen::InsensitiveStr::Ascii("requiresd"),
                &["requires", "required"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislaimer"), &["disclaimer"]),
            (dictgen::InsensitiveStr::Ascii("potatos"), &["potatoes"]),
            (
                dictgen::InsensitiveStr::Ascii("publc"),
                &["public", "pubic"],
            ),
            (dictgen::InsensitiveStr::Ascii("shal"), &["shall"]),
            (
                dictgen::InsensitiveStr::Ascii("abstracness"),
                &["abstractness"],
            ),
            (dictgen::InsensitiveStr::Ascii("spaece"), &["space"]),
            (dictgen::InsensitiveStr::Ascii("primitve"), &["primitive"]),
            (
                dictgen::InsensitiveStr::Ascii("agriculutral"),
                &["agricultural"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unresponisve"),
                &["unresponsive"],
            ),
            (dictgen::InsensitiveStr::Ascii("intrige"), &["intrigue"]),
            (dictgen::InsensitiveStr::Ascii("reamining"), &["remaining"]),
            (dictgen::InsensitiveStr::Ascii("materiaal"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("expeditivo"),
                &["expedition"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enthusiatic"),
                &["enthusiastic"],
            ),
            (dictgen::InsensitiveStr::Ascii("supoosed"), &["supposed"]),
            (
                dictgen::InsensitiveStr::Ascii("influances"),
                &["influences"],
            ),
            (dictgen::InsensitiveStr::Ascii("hypenates"), &["hyphenates"]),
            (
                dictgen::InsensitiveStr::Ascii("canocicalization"),
                &["canonicalization"],
            ),
            (dictgen::InsensitiveStr::Ascii("dicisions"), &["decisions"]),
            (dictgen::InsensitiveStr::Ascii("combusion"), &["combustion"]),
            (
                dictgen::InsensitiveStr::Ascii("crytpography"),
                &["cryptography"],
            ),
            (dictgen::InsensitiveStr::Ascii("napoms"), &["napalms"]),
            (dictgen::InsensitiveStr::Ascii("fomated"), &["formatted"]),
            (dictgen::InsensitiveStr::Ascii("passinig"), &["passing"]),
            (dictgen::InsensitiveStr::Ascii("luetooth"), &["bluetooth"]),
            (dictgen::InsensitiveStr::Ascii("srpouts"), &["sprouts"]),
            (dictgen::InsensitiveStr::Ascii("medicins"), &["medicines"]),
            (
                dictgen::InsensitiveStr::Ascii("introducinc"),
                &["introducing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disshearteningly"),
                &["dishearteningly"],
            ),
            (dictgen::InsensitiveStr::Ascii("couning"), &["counting"]),
            (dictgen::InsensitiveStr::Ascii("disbelife"), &["disbelief"]),
            (dictgen::InsensitiveStr::Ascii("cresits"), &["credits"]),
            (dictgen::InsensitiveStr::Ascii("captial"), &["capital"]),
            (
                dictgen::InsensitiveStr::Ascii("cicling"),
                &["cycling", "circling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authencator"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("corgnito"), &["cognito"]),
            (dictgen::InsensitiveStr::Ascii("collistion"), &["collision"]),
            (
                dictgen::InsensitiveStr::Ascii("hystericallly"),
                &["hysterically"],
            ),
            (dictgen::InsensitiveStr::Ascii("animetors"), &["animators"]),
            (
                dictgen::InsensitiveStr::Ascii("homogeniusly"),
                &["homogeneously"],
            ),
            (dictgen::InsensitiveStr::Ascii("fowarding"), &["forwarding"]),
            (
                dictgen::InsensitiveStr::Ascii("microscrope"),
                &["microscope"],
            ),
            (dictgen::InsensitiveStr::Ascii("headerrs"), &["headers"]),
            (dictgen::InsensitiveStr::Ascii("abundunt"), &["abundant"]),
            (
                dictgen::InsensitiveStr::Ascii("botloaders"),
                &["bootloaders"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncerainties"),
                &["uncertainties"],
            ),
            (dictgen::InsensitiveStr::Ascii("notthing"), &["nothing"]),
            (
                dictgen::InsensitiveStr::Ascii("longitudonal"),
                &["longitudinal"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("catagorizations"),
                &["categorizations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("automaticalyy"),
                &["automatically"],
            ),
            (dictgen::InsensitiveStr::Ascii("staduims"), &["stadiums"]),
            (dictgen::InsensitiveStr::Ascii("decieding"), &["deciding"]),
            (
                dictgen::InsensitiveStr::Ascii("inspectres"),
                &["inspectors"],
            ),
            (dictgen::InsensitiveStr::Ascii("anuglar"), &["angular"]),
            (dictgen::InsensitiveStr::Ascii("microsof"), &["microsoft"]),
            (
                dictgen::InsensitiveStr::Ascii("translater"),
                &["translator"],
            ),
            (dictgen::InsensitiveStr::Ascii("ownershop"), &["ownership"]),
            (dictgen::InsensitiveStr::Ascii("neighror"), &["neighbor"]),
            (dictgen::InsensitiveStr::Ascii("peanuckle"), &["pinochle"]),
            (dictgen::InsensitiveStr::Ascii("uou"), &["you"]),
            (
                dictgen::InsensitiveStr::Ascii("inexperiance"),
                &["inexperience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coninciding"),
                &["coinciding"],
            ),
            (dictgen::InsensitiveStr::Ascii("rememberd"), &["remembered"]),
            (
                dictgen::InsensitiveStr::Ascii("suphisticated"),
                &["sophisticated"],
            ),
            (dictgen::InsensitiveStr::Ascii("symetri"), &["symmetry"]),
            (dictgen::InsensitiveStr::Ascii("lirary"), &["library"]),
            (
                dictgen::InsensitiveStr::Ascii("powerhosue"),
                &["powerhouse"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cellpading"),
                &["cellpadding"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("angshiuosness"),
                &["anxiousness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("audiobookas"),
                &["audiobooks"],
            ),
            (dictgen::InsensitiveStr::Ascii("graidents"), &["gradients"]),
            (
                dictgen::InsensitiveStr::Ascii("exclemation"),
                &["exclamation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("muiltiplies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("timestaping"),
                &["timestamping"],
            ),
            (dictgen::InsensitiveStr::Ascii("parliment"), &["parliament"]),
            (
                dictgen::InsensitiveStr::Ascii("regularlisation"),
                &["regularisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("iterested"), &["interested"]),
            (
                dictgen::InsensitiveStr::Ascii("anticapates"),
                &["anticipates"],
            ),
            (dictgen::InsensitiveStr::Ascii("depecates"), &["deprecates"]),
            (dictgen::InsensitiveStr::Ascii("ziping"), &["zipping"]),
            (
                dictgen::InsensitiveStr::Ascii("partirioned"),
                &["partitioned"],
            ),
            (dictgen::InsensitiveStr::Ascii("dislayed"), &["displayed"]),
            (
                dictgen::InsensitiveStr::Ascii("momentarilly"),
                &["momentarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("decutables"),
                &["deductibles"],
            ),
            (dictgen::InsensitiveStr::Ascii("devicest"), &["devices"]),
            (
                dictgen::InsensitiveStr::Ascii("countercut"),
                &["counteract"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("intruduces"),
                &["introduces"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("figurativeley"),
                &["figuratively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbouts"),
                &["neighbours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("kindergaten"),
                &["kindergarten"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("perference"),
                &["preference"],
            ),
            (dictgen::InsensitiveStr::Ascii("applizes"), &["applies"]),
            (
                dictgen::InsensitiveStr::Ascii("convienience"),
                &["convenience"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("celebrites"),
                &["celebrities"],
            ),
            (dictgen::InsensitiveStr::Ascii("humanit"), &["humanist"]),
            (
                dictgen::InsensitiveStr::Ascii("skillsshot"),
                &["skillshots"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prematuraly"),
                &["prematurely"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("revelutions"),
                &["revolutions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shtopped"),
                &["stopped", "shopped"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repearable"),
                &["repeatable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prepostions"),
                &["prepositions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("precentage"),
                &["percentage"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("correspndingly"),
                &["correspondingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("uage"), &["usage"]),
            (
                dictgen::InsensitiveStr::Ascii("senstitivity"),
                &["sensitivity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("trangles"),
                &["triangles", "tangles", "strangles", "wrangles"],
            ),
            (dictgen::InsensitiveStr::Ascii("sovren"), &["sovereign"]),
            (dictgen::InsensitiveStr::Ascii("tooic"), &["topic"]),
            (dictgen::InsensitiveStr::Ascii("egistered"), &["registered"]),
            (dictgen::InsensitiveStr::Ascii("cretaes"), &["creates"]),
            (dictgen::InsensitiveStr::Ascii("crooz"), &["cruise"]),
            (dictgen::InsensitiveStr::Ascii("itmes"), &["items", "times"]),
            (dictgen::InsensitiveStr::Ascii("anlayse"), &["analyse"]),
            (dictgen::InsensitiveStr::Ascii("vertabray"), &["vertebrae"]),
            (
                dictgen::InsensitiveStr::Ascii("demonstras"),
                &["demonstrates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("democraticaly"),
                &["democratically"],
            ),
            (dictgen::InsensitiveStr::Ascii("occuracy"), &["accuracy"]),
            (dictgen::InsensitiveStr::Ascii("hanlding"), &["handling"]),
            (
                dictgen::InsensitiveStr::Ascii("incomprehenisble"),
                &["incomprehensible"],
            ),
            (dictgen::InsensitiveStr::Ascii("succssive"), &["successive"]),
            (
                dictgen::InsensitiveStr::Ascii("rearrangteable"),
                &["rearrangeable"],
            ),
            (dictgen::InsensitiveStr::Ascii("tlak"), &["talk"]),
            (dictgen::InsensitiveStr::Ascii("expemted"), &["exempted"]),
            (
                dictgen::InsensitiveStr::Ascii("referenecing"),
                &["referencing"],
            ),
            (dictgen::InsensitiveStr::Ascii("inerit"), &["inherit"]),
            (
                dictgen::InsensitiveStr::Ascii("renegotitaed"),
                &["renegotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("datatores"), &["datastores"]),
            (
                dictgen::InsensitiveStr::Ascii("negotaiated"),
                &["negotiated"],
            ),
            (dictgen::InsensitiveStr::Ascii("claenup"), &["cleanup"]),
            (
                dictgen::InsensitiveStr::Ascii("trcking"),
                &["tracking", "tricking", "trucking"],
            ),
            (dictgen::InsensitiveStr::Ascii("studens"), &["students"]),
            (dictgen::InsensitiveStr::Ascii("nominae"), &["nominate"]),
            (dictgen::InsensitiveStr::Ascii("resulve"), &["resolve"]),
            (
                dictgen::InsensitiveStr::Ascii("philipines"),
                &["philippines"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("differate"),
                &["differentiate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unspeficieid"),
                &["unspecified"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underestemating"),
                &["underestimating"],
            ),
            (dictgen::InsensitiveStr::Ascii("frome"), &["from"]),
            (dictgen::InsensitiveStr::Ascii("savere"), &["severe"]),
            (
                dictgen::InsensitiveStr::Ascii("popoulations"),
                &["populations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ancapsulation"),
                &["encapsulation"],
            ),
            (dictgen::InsensitiveStr::Ascii("detailes"), &["details"]),
            (dictgen::InsensitiveStr::Ascii("searhes"), &["searches"]),
            (dictgen::InsensitiveStr::Ascii("graceufl"), &["graceful"]),
            (
                dictgen::InsensitiveStr::Ascii("implmented"),
                &["implemented"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("itereations"),
                &["iterations"],
            ),
            (dictgen::InsensitiveStr::Ascii("labrynths"), &["labyrinths"]),
            (dictgen::InsensitiveStr::Ascii("refrehshes"), &["refreshes"]),
            (
                dictgen::InsensitiveStr::Ascii("alternatiely"),
                &["alternately", "alternatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("contersink"),
                &["countersink"],
            ),
            (dictgen::InsensitiveStr::Ascii("hoep"), &["hope"]),
            (dictgen::InsensitiveStr::Ascii("staisfied"), &["satisfied"]),
            (
                dictgen::InsensitiveStr::Ascii("inexeprienced"),
                &["inexperience"],
            ),
            (dictgen::InsensitiveStr::Ascii("espinoage"), &["espionage"]),
            (dictgen::InsensitiveStr::Ascii("strippping"), &["stripping"]),
            (
                dictgen::InsensitiveStr::Ascii("victemizes"),
                &["victimizes"],
            ),
            (dictgen::InsensitiveStr::Ascii("usingg"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("comlpeter"), &["completer"]),
            (
                dictgen::InsensitiveStr::Ascii("prejuduced"),
                &["prejudiced"],
            ),
            (dictgen::InsensitiveStr::Ascii("coudl"), &["could", "cloud"]),
            (
                dictgen::InsensitiveStr::Ascii("compleately"),
                &["completely"],
            ),
            (dictgen::InsensitiveStr::Ascii("rearrane"), &["rearrange"]),
            (dictgen::InsensitiveStr::Ascii("programes"), &["programs"]),
            (
                dictgen::InsensitiveStr::Ascii("extrapoliating"),
                &["extrapolating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fucntionalities"),
                &["functionalities"],
            ),
            (dictgen::InsensitiveStr::Ascii("grifeing"), &["griefing"]),
            (dictgen::InsensitiveStr::Ascii("triptishes"), &["triptychs"]),
            (dictgen::InsensitiveStr::Ascii("briely"), &["briefly"]),
            (
                dictgen::InsensitiveStr::Ascii("appropirately"),
                &["appropriately"],
            ),
            (dictgen::InsensitiveStr::Ascii("urkainian"), &["ukrainian"]),
            (dictgen::InsensitiveStr::Ascii("implemnt"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("tyies"), &["tries"]),
            (dictgen::InsensitiveStr::Ascii("encorcing"), &["enforcing"]),
            (
                dictgen::InsensitiveStr::Ascii("curresponded"),
                &["corresponded"],
            ),
            (dictgen::InsensitiveStr::Ascii("geomtery"), &["geometry"]),
            (dictgen::InsensitiveStr::Ascii("remainer"), &["remainder"]),
            (dictgen::InsensitiveStr::Ascii("nutrientes"), &["nutrients"]),
            (dictgen::InsensitiveStr::Ascii("inferir"), &["inferior"]),
            (dictgen::InsensitiveStr::Ascii("coul"), &["could"]),
            (
                dictgen::InsensitiveStr::Ascii("eficiency"),
                &["efficiency", "deficiency"],
            ),
            (dictgen::InsensitiveStr::Ascii("tempelates"), &["templates"]),
            (dictgen::InsensitiveStr::Ascii("contence"), &["contents"]),
            (
                dictgen::InsensitiveStr::Ascii("multilplying"),
                &["multiplying"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("preparato"),
                &["preparation"],
            ),
            (dictgen::InsensitiveStr::Ascii("aggreggate"), &["aggregate"]),
            (
                dictgen::InsensitiveStr::Ascii("accepatbly"),
                &["acceptably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("externelly"),
                &["externally"],
            ),
            (dictgen::InsensitiveStr::Ascii("designade"), &["designated"]),
            (
                dictgen::InsensitiveStr::Ascii("authentificated"),
                &["authenticated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("repositorry"),
                &["repository"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rembembering"),
                &["remembering"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("persucuted"),
                &["persecuted"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("brazilianess"),
                &["brazilians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interting"),
                &["interacting", "inserting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("chemcially"),
                &["chemically"],
            ),
            (dictgen::InsensitiveStr::Ascii("comlete"), &["complete"]),
            (dictgen::InsensitiveStr::Ascii("ciricle"), &["circle"]),
            (
                dictgen::InsensitiveStr::Ascii("intermitent"),
                &["intermittent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("admi"),
                &["admin", "admit", "admix"],
            ),
            (dictgen::InsensitiveStr::Ascii("delele"), &["delete"]),
            (
                dictgen::InsensitiveStr::Ascii("reprecussions"),
                &["repercussions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("commmitters"),
                &["committers"],
            ),
            (dictgen::InsensitiveStr::Ascii("sideral"), &["sidereal"]),
            (
                dictgen::InsensitiveStr::Ascii("dipections"),
                &["depictions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("submisions"),
                &["submissions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("ecapsulates"),
                &["encapsulates"],
            ),
            (dictgen::InsensitiveStr::Ascii("beleifed"), &["believed"]),
            (dictgen::InsensitiveStr::Ascii("profilees"), &["profiles"]),
            (
                dictgen::InsensitiveStr::Ascii("temination"),
                &["termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("desingage"), &["disengage"]),
            (dictgen::InsensitiveStr::Ascii("samplin"), &["sampling"]),
            (dictgen::InsensitiveStr::Ascii("serting"), &["setting"]),
            (dictgen::InsensitiveStr::Ascii("applay"), &["apply"]),
            (dictgen::InsensitiveStr::Ascii("casette"), &["cassette"]),
            (
                dictgen::InsensitiveStr::Ascii("idiosynchracies"),
                &["idiosyncrasies"],
            ),
            (dictgen::InsensitiveStr::Ascii("parnter"), &["partner"]),
            (dictgen::InsensitiveStr::Ascii("hanshaked"), &["handshaked"]),
            (
                dictgen::InsensitiveStr::Ascii("jacksonvile"),
                &["jacksonville"],
            ),
            (dictgen::InsensitiveStr::Ascii("verctor"), &["vector"]),
            (dictgen::InsensitiveStr::Ascii("shitfer"), &["shifter"]),
            (
                dictgen::InsensitiveStr::Ascii("declatation"),
                &["declaration"],
            ),
            (dictgen::InsensitiveStr::Ascii("custon"), &["custom"]),
            (dictgen::InsensitiveStr::Ascii("ogliarchy"), &["oligarchy"]),
            (
                dictgen::InsensitiveStr::Ascii("ukraineans"),
                &["ukrainians"],
            ),
            (dictgen::InsensitiveStr::Ascii("adnroids"), &["androids"]),
            (
                dictgen::InsensitiveStr::Ascii("identfiers"),
                &["identifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("upsteraming"),
                &["upstreaming"],
            ),
            (dictgen::InsensitiveStr::Ascii("joystik"), &["joystick"]),
            (dictgen::InsensitiveStr::Ascii("secods"), &["seconds"]),
            (
                dictgen::InsensitiveStr::Ascii("stroe"),
                &["store", "stroke", "strobe", "strove", "strode"],
            ),
            (dictgen::InsensitiveStr::Ascii("defaullts"), &["defaults"]),
            (
                dictgen::InsensitiveStr::Ascii("instlalation"),
                &["installation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terristrial"),
                &["territorial"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("assassinatd"),
                &["assassinated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("namesapced"),
                &["namespaced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("temepratures"),
                &["temperatures"],
            ),
            (dictgen::InsensitiveStr::Ascii("shittoon"), &["shitton"]),
            (
                dictgen::InsensitiveStr::Ascii("coventional"),
                &["conventional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responisveness"),
                &["responsiveness"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("instucting"),
                &["instructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conditionar"),
                &["conditioner"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("incoherrent"),
                &["incoherent"],
            ),
            (dictgen::InsensitiveStr::Ascii("commments"), &["comments"]),
            (dictgen::InsensitiveStr::Ascii("asssembly"), &["assembly"]),
            (dictgen::InsensitiveStr::Ascii("activited"), &["activated"]),
            (
                dictgen::InsensitiveStr::Ascii("converstaional"),
                &["conversational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("fundimentalist"),
                &["fundamentalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("simeple"), &["simple"]),
            (
                dictgen::InsensitiveStr::Ascii("egotastical"),
                &["egotistical"],
            ),
            (dictgen::InsensitiveStr::Ascii("offfers"), &["offers"]),
            (dictgen::InsensitiveStr::Ascii("shorthly"), &["shortly"]),
            (dictgen::InsensitiveStr::Ascii("collogues"), &["colleagues"]),
            (
                dictgen::InsensitiveStr::Ascii("correspoondents"),
                &["correspondents"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algotirhms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("surpising"), &["surprising"]),
            (dictgen::InsensitiveStr::Ascii("bloggare"), &["blogger"]),
            (dictgen::InsensitiveStr::Ascii("homophobie"), &["homophobe"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowldgments"),
                &["acknowledgments"],
            ),
            (dictgen::InsensitiveStr::Ascii("artifical"), &["artificial"]),
            (dictgen::InsensitiveStr::Ascii("veridying"), &["verifying"]),
            (dictgen::InsensitiveStr::Ascii("modfier"), &["modifier"]),
            (
                dictgen::InsensitiveStr::Ascii("acknowledgemnts"),
                &["acknowledgements"],
            ),
            (dictgen::InsensitiveStr::Ascii("deafult"), &["default"]),
            (
                dictgen::InsensitiveStr::Ascii("formatings"),
                &["formations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("surroundins"),
                &["surrounds"],
            ),
            (dictgen::InsensitiveStr::Ascii("carrefull"), &["careful"]),
            (dictgen::InsensitiveStr::Ascii("argumenys"), &["arguments"]),
            (
                dictgen::InsensitiveStr::Ascii("organisiation"),
                &["organisation"],
            ),
            (dictgen::InsensitiveStr::Ascii("assighned"), &["assigned"]),
            (
                dictgen::InsensitiveStr::Ascii("visuallization"),
                &["visualization"],
            ),
            (dictgen::InsensitiveStr::Ascii("exsistent"), &["existent"]),
            (dictgen::InsensitiveStr::Ascii("ameythst"), &["amethyst"]),
            (dictgen::InsensitiveStr::Ascii("seezed"), &["seized"]),
            (
                dictgen::InsensitiveStr::Ascii("correponding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("follw"), &["follow"]),
            (dictgen::InsensitiveStr::Ascii("extortin"), &["extortion"]),
            (
                dictgen::InsensitiveStr::Ascii("generilizes"),
                &["generalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("compination"),
                &["compilation", "combination"],
            ),
            (dictgen::InsensitiveStr::Ascii("septmeber"), &["september"]),
            (dictgen::InsensitiveStr::Ascii("separely"), &["separately"]),
            (dictgen::InsensitiveStr::Ascii("hwihc"), &["which"]),
            (
                dictgen::InsensitiveStr::Ascii("sacled"),
                &["scaled", "sacred", "sacked", "sailed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("friendzies"),
                &["friendlies"],
            ),
            (dictgen::InsensitiveStr::Ascii("funcions"), &["functions"]),
            (
                dictgen::InsensitiveStr::Ascii("differintiate"),
                &["differentiate"],
            ),
            (dictgen::InsensitiveStr::Ascii("chaeckers"), &["checkers"]),
            (
                dictgen::InsensitiveStr::Ascii("approxmimations"),
                &["approximations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("baases"),
                &["bases", "abases"],
            ),
            (dictgen::InsensitiveStr::Ascii("connaect"), &["connect"]),
            (
                dictgen::InsensitiveStr::Ascii("imprevements"),
                &["improvements"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convoultion"),
                &["convolution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("nutritivos"),
                &["nutritious"],
            ),
            (dictgen::InsensitiveStr::Ascii("oveall"), &["overall"]),
            (
                dictgen::InsensitiveStr::Ascii("labatory"),
                &["lavatory", "laboratory"],
            ),
            (dictgen::InsensitiveStr::Ascii("sytematic"), &["systematic"]),
            (dictgen::InsensitiveStr::Ascii("emebedding"), &["embedding"]),
            (
                dictgen::InsensitiveStr::Ascii("parametesr"),
                &["parameters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accurs"),
                &["accurse", "occurs"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reconeccts"),
                &["reconnects"],
            ),
            (dictgen::InsensitiveStr::Ascii("featutures"), &["features"]),
            (
                dictgen::InsensitiveStr::Ascii("copmutations"),
                &["computations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("caluculations"),
                &["calculations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("modiofication"),
                &["modification"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("algorightmically"),
                &["algorithmically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("underraged"),
                &["underrated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simularly"),
                &["similarly", "singularly"],
            ),
            (dictgen::InsensitiveStr::Ascii("aperure"), &["aperture"]),
            (dictgen::InsensitiveStr::Ascii("hospot"), &["hotspot"]),
            (
                dictgen::InsensitiveStr::Ascii("asynchronsly"),
                &["asynchronously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("appparently"),
                &["apparently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("peroidicity"),
                &["periodicity"],
            ),
            (dictgen::InsensitiveStr::Ascii("forigener"), &["foreigner"]),
            (dictgen::InsensitiveStr::Ascii("techniquet"), &["technique"]),
            (dictgen::InsensitiveStr::Ascii("desciribe"), &["describe"]),
            (
                dictgen::InsensitiveStr::Ascii("disapeard"),
                &["disappeared"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("agreeements"),
                &["agreements"],
            ),
            (dictgen::InsensitiveStr::Ascii("alle"), &["all", "alley"]),
            (dictgen::InsensitiveStr::Ascii("guarenntee"), &["guarantee"]),
            (dictgen::InsensitiveStr::Ascii("leyout"), &["layout"]),
            (
                dictgen::InsensitiveStr::Ascii("masturbatin"),
                &["masturbation", "masturbating"],
            ),
            (dictgen::InsensitiveStr::Ascii("termine"), &["determine"]),
            (dictgen::InsensitiveStr::Ascii("behabiour"), &["behaviour"]),
            (dictgen::InsensitiveStr::Ascii("addjusted"), &["adjusted"]),
            (dictgen::InsensitiveStr::Ascii("namepaces"), &["namespaces"]),
            (dictgen::InsensitiveStr::Ascii("unusde"), &["unused"]),
            (dictgen::InsensitiveStr::Ascii("thna"), &["than"]),
            (dictgen::InsensitiveStr::Ascii("virttually"), &["virtually"]),
            (
                dictgen::InsensitiveStr::Ascii("minnor"),
                &["minor", "minnow"],
            ),
            (dictgen::InsensitiveStr::Ascii("armonic"), &["harmonic"]),
            (
                dictgen::InsensitiveStr::Ascii("raquetball"),
                &["racquetball"],
            ),
            (dictgen::InsensitiveStr::Ascii("assaultes"), &["assaults"]),
            (
                dictgen::InsensitiveStr::Ascii("compiation"),
                &["compilation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("tessellatin"),
                &["tessellating", "tessellation"],
            ),
            (dictgen::InsensitiveStr::Ascii("separatos"), &["separators"]),
            (
                dictgen::InsensitiveStr::Ascii("concatenadions"),
                &["concatenations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("rertievers"),
                &["retrievers"],
            ),
            (dictgen::InsensitiveStr::Ascii("seleccting"), &["selecting"]),
            (dictgen::InsensitiveStr::Ascii("postiive"), &["positive"]),
            (dictgen::InsensitiveStr::Ascii("staically"), &["statically"]),
            (
                dictgen::InsensitiveStr::Ascii("presistant"),
                &["persistent"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inherritance"),
                &["inheritance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("frecuently"),
                &["frequently"],
            ),
            (dictgen::InsensitiveStr::Ascii("calcuim"), &["calcium"]),
            (
                dictgen::InsensitiveStr::Ascii("persuation"),
                &["persuasion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("distribues"),
                &["distributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("chateos"), &["châteaux"]),
            (dictgen::InsensitiveStr::Ascii("cannonical"), &["canonical"]),
            (
                dictgen::InsensitiveStr::Ascii("compilicating"),
                &["complicating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforsing"),
                &["endorsing", "enforcing"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaccum"), &["vacuum"]),
            (dictgen::InsensitiveStr::Ascii("soluations"), &["solutions"]),
            (dictgen::InsensitiveStr::Ascii("lifecyle"), &["lifecycle"]),
            (
                dictgen::InsensitiveStr::Ascii("contribuito"),
                &["contribution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("internetese"),
                &["internets"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("releted"),
                &["deleted", "related"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("donwvoters"),
                &["downvoters"],
            ),
            (dictgen::InsensitiveStr::Ascii("decern"), &["discern"]),
            (dictgen::InsensitiveStr::Ascii("obvoiusly"), &["obviously"]),
            (
                dictgen::InsensitiveStr::Ascii("esteblishes"),
                &["establishes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("interrested"),
                &["interested"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("deplicate"),
                &["duplicate", "replicate", "delicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("marshmallowness"),
                &["marshmallows"],
            ),
            (dictgen::InsensitiveStr::Ascii("paide"), &["paid"]),
            (dictgen::InsensitiveStr::Ascii("oppsofite"), &["opposite"]),
            (
                dictgen::InsensitiveStr::Ascii("knowledgible"),
                &["knowledgable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enligthened"),
                &["enlightened"],
            ),
            (dictgen::InsensitiveStr::Ascii("testss"), &["tests"]),
            (
                dictgen::InsensitiveStr::Ascii("electrocytes"),
                &["electrolytes"],
            ),
            (dictgen::InsensitiveStr::Ascii("reosurced"), &["resourced"]),
            (
                dictgen::InsensitiveStr::Ascii("considerion"),
                &["consideration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("relatation"),
                &["relaxation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("metiphorical"),
                &["metaphorical"],
            ),
            (dictgen::InsensitiveStr::Ascii("circits"), &["circuits"]),
            (
                dictgen::InsensitiveStr::Ascii("definataly"),
                &["definitely"],
            ),
            (dictgen::InsensitiveStr::Ascii("porfiling"), &["profiling"]),
            (dictgen::InsensitiveStr::Ascii("summur"), &["summer"]),
            (dictgen::InsensitiveStr::Ascii("postmage"), &["postimage"]),
            (dictgen::InsensitiveStr::Ascii("normlly"), &["normally"]),
            (dictgen::InsensitiveStr::Ascii("coytoes"), &["coyotes"]),
            (
                dictgen::InsensitiveStr::Ascii("infrastracture"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authenticaor"),
                &["authenticator"],
            ),
            (dictgen::InsensitiveStr::Ascii("sycology"), &["psychology"]),
            (dictgen::InsensitiveStr::Ascii("bisines"), &["business"]),
            (
                dictgen::InsensitiveStr::Ascii("initialiss"),
                &["initialise", "initialises"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authroitatively"),
                &["authoritatively"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explitictly"),
                &["explicitly"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("paralelism"),
                &["parallelism"],
            ),
            (dictgen::InsensitiveStr::Ascii("anormalies"), &["anomalies"]),
            (
                dictgen::InsensitiveStr::Ascii("comnpletion"),
                &["completion"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("privelidge"),
                &["privileged"],
            ),
            (dictgen::InsensitiveStr::Ascii("exhaustn"), &["exhaustion"]),
            (
                dictgen::InsensitiveStr::Ascii("inadvertantely"),
                &["inadvertently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("granualtiry"),
                &["granularity"],
            ),
            (dictgen::InsensitiveStr::Ascii("aleays"), &["always"]),
            (dictgen::InsensitiveStr::Ascii("satifies"), &["satisfies"]),
            (
                dictgen::InsensitiveStr::Ascii("standy"),
                &["standby", "sandy", "standee"],
            ),
            (dictgen::InsensitiveStr::Ascii("singelton"), &["singleton"]),
            (
                dictgen::InsensitiveStr::Ascii("sentimetal"),
                &["sentimental"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bussineses"),
                &["businesses", "business"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unnecesasry"),
                &["unnecessary"],
            ),
            (dictgen::InsensitiveStr::Ascii("pecuilar"), &["peculiar"]),
            (dictgen::InsensitiveStr::Ascii("sumberged"), &["submerged"]),
            (
                dictgen::InsensitiveStr::Ascii("normallized"),
                &["normalized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("strechting"),
                &["stretching"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmentals"),
                &["environments"],
            ),
            (dictgen::InsensitiveStr::Ascii("impelent"), &["implement"]),
            (dictgen::InsensitiveStr::Ascii("diagnossed"), &["diagnose"]),
            (dictgen::InsensitiveStr::Ascii("develoer"), &["developer"]),
            (dictgen::InsensitiveStr::Ascii("rabbitos"), &["rabbits"]),
            (
                dictgen::InsensitiveStr::Ascii("correspnding"),
                &["corresponding"],
            ),
            (dictgen::InsensitiveStr::Ascii("teated"), &["treated"]),
            (dictgen::InsensitiveStr::Ascii("evloves"), &["evolves"]),
            (
                dictgen::InsensitiveStr::Ascii("fursermore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("authecates"),
                &["authenticates"],
            ),
            (dictgen::InsensitiveStr::Ascii("follwoeed"), &["followed"]),
            (
                dictgen::InsensitiveStr::Ascii("workaronud"),
                &["workaround"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("singalling"),
                &["signalling"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegociating"),
                &["renegotiating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("posisitioned"),
                &["positioned"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("roundtriping"),
                &["roundtripping"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("speficillically"),
                &["specifically"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insttuction"),
                &["instruction"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("apostraphes"),
                &["apostrophes"],
            ),
            (dictgen::InsensitiveStr::Ascii("termplated"), &["templated"]),
            (dictgen::InsensitiveStr::Ascii("tounrey"), &["tourney"]),
            (dictgen::InsensitiveStr::Ascii("apparance"), &["appearance"]),
            (dictgen::InsensitiveStr::Ascii("colisions"), &["collisions"]),
            (dictgen::InsensitiveStr::Ascii("skwalks"), &["squawks"]),
            (
                dictgen::InsensitiveStr::Ascii("downgarading"),
                &["downgrading"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("projetctive"),
                &["projective"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("defininitive"),
                &["definitive"],
            ),
            (dictgen::InsensitiveStr::Ascii("critiera"), &["criteria"]),
            (
                dictgen::InsensitiveStr::Ascii("superbocks"),
                &["superblocks"],
            ),
            (dictgen::InsensitiveStr::Ascii("cosistent"), &["consistent"]),
            (dictgen::InsensitiveStr::Ascii("infomred"), &["informed"]),
            (
                dictgen::InsensitiveStr::Ascii("triagulation"),
                &["triangulation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("supermarkedet"),
                &["supermarket"],
            ),
            (dictgen::InsensitiveStr::Ascii("amkefile"), &["makefile"]),
            (dictgen::InsensitiveStr::Ascii("sentivite"), &["sensitive"]),
            (dictgen::InsensitiveStr::Ascii("prevew"), &["preview"]),
            (dictgen::InsensitiveStr::Ascii("escpaed"), &["escaped"]),
            (dictgen::InsensitiveStr::Ascii("semantis"), &["semantics"]),
            (
                dictgen::InsensitiveStr::Ascii("previaling"),
                &["prevailing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("princibles"),
                &["principles"],
            ),
            (dictgen::InsensitiveStr::Ascii("ecouraged"), &["encouraged"]),
            (
                dictgen::InsensitiveStr::Ascii("administraor"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("cheecklist"), &["checklist"]),
            (
                dictgen::InsensitiveStr::Ascii("millionairs"),
                &["millionaires"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("destinatin"),
                &["destinations", "destination"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cariactures"),
                &["caricature"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uninteressting"),
                &["uninteresting"],
            ),
            (dictgen::InsensitiveStr::Ascii("identifer"), &["identifier"]),
            (dictgen::InsensitiveStr::Ascii("recevieved"), &["received"]),
            (dictgen::InsensitiveStr::Ascii("libraiesr"), &["libraries"]),
            (dictgen::InsensitiveStr::Ascii("comtained"), &["contained"]),
            (
                dictgen::InsensitiveStr::Ascii("syntethically"),
                &["synthetically"],
            ),
            (dictgen::InsensitiveStr::Ascii("debateable"), &["debatable"]),
            (
                dictgen::InsensitiveStr::Ascii("epxiressions"),
                &["expressions"],
            ),
            (dictgen::InsensitiveStr::Ascii("pupular"), &["popular"]),
            (dictgen::InsensitiveStr::Ascii("elphant"), &["elephant"]),
            (dictgen::InsensitiveStr::Ascii("montonic"), &["monotonic"]),
            (dictgen::InsensitiveStr::Ascii("speack"), &["speak"]),
            (
                dictgen::InsensitiveStr::Ascii("unavailibility"),
                &["unavailability"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("renegotioating"),
                &["renegotiating"],
            ),
            (dictgen::InsensitiveStr::Ascii("wraapper"), &["wrapper"]),
            (
                dictgen::InsensitiveStr::Ascii("locatin"),
                &["location", "locating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("asynnchronous"),
                &["asynchronous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("terminatin"),
                &["terminating", "termination"],
            ),
            (dictgen::InsensitiveStr::Ascii("commiteted"), &["committed"]),
            (dictgen::InsensitiveStr::Ascii("espoinage"), &["espionage"]),
            (
                dictgen::InsensitiveStr::Ascii("inheratance"),
                &["inheritance"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("calrifying"),
                &["clarifying"],
            ),
            (dictgen::InsensitiveStr::Ascii("clssroom"), &["classroom"]),
            (dictgen::InsensitiveStr::Ascii("switchs"), &["switches"]),
            (
                dictgen::InsensitiveStr::Ascii("expermeneted"),
                &["experimented"],
            ),
            (dictgen::InsensitiveStr::Ascii("unitilties"), &["utilities"]),
            (
                dictgen::InsensitiveStr::Ascii("parocheelly"),
                &["parochially"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pyschiatrists"),
                &["psychiatrists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("entertwined"),
                &["entertained"],
            ),
            (dictgen::InsensitiveStr::Ascii("dacquiris"), &["daiquiris"]),
            (dictgen::InsensitiveStr::Ascii("previlege"), &["privilege"]),
            (dictgen::InsensitiveStr::Ascii("sparation"), &["separation"]),
            (
                dictgen::InsensitiveStr::Ascii("speciefiers"),
                &["specifiers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inaproppriately"),
                &["inappropriately"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conrresponded"),
                &["corresponded"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enlargment"),
                &["enlargement"],
            ),
            (dictgen::InsensitiveStr::Ascii("rendtion"), &["rendition"]),
            (
                dictgen::InsensitiveStr::Ascii("autonegotitation"),
                &["autonegotiation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbitrors"),
                &["arbiters", "arbitrators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("experiration"),
                &["expiration"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("successsors"),
                &["successors"],
            ),
            (dictgen::InsensitiveStr::Ascii("plyotropy"), &["pleiotropy"]),
            (
                dictgen::InsensitiveStr::Ascii("additinoally"),
                &["additionally"],
            ),
            (dictgen::InsensitiveStr::Ascii("respberry"), &["raspberry"]),
            (
                dictgen::InsensitiveStr::Ascii("alocations"),
                &["allocations"],
            ),
            (dictgen::InsensitiveStr::Ascii("maxmiums"), &["maximums"]),
            (dictgen::InsensitiveStr::Ascii("statemanet"), &["statement"]),
            (
                dictgen::InsensitiveStr::Ascii("experament"),
                &["experiment"],
            ),
            (dictgen::InsensitiveStr::Ascii("honsetly"), &["honestly"]),
            (dictgen::InsensitiveStr::Ascii("dericed"), &["derived"]),
            (
                dictgen::InsensitiveStr::Ascii("aithentication"),
                &["authentication"],
            ),
            (dictgen::InsensitiveStr::Ascii("acutual"), &["actual"]),
            (
                dictgen::InsensitiveStr::Ascii("recipricate"),
                &["reciprocate"],
            ),
            (dictgen::InsensitiveStr::Ascii("deleteted"), &["deleted"]),
            (dictgen::InsensitiveStr::Ascii("spcial"), &["special"]),
            (dictgen::InsensitiveStr::Ascii("detrmine"), &["determine"]),
            (
                dictgen::InsensitiveStr::Ascii("transisition"),
                &["transition"],
            ),
            (dictgen::InsensitiveStr::Ascii("immutible"), &["immutable"]),
            (
                dictgen::InsensitiveStr::Ascii("connectiona"),
                &["connection"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("responsibily"),
                &["responsibility"],
            ),
            (dictgen::InsensitiveStr::Ascii("brigher"), &["brighter"]),
            (dictgen::InsensitiveStr::Ascii("potrait"), &["portrait"]),
            (dictgen::InsensitiveStr::Ascii("pennensula"), &["peninsula"]),
            (
                dictgen::InsensitiveStr::Ascii("interpoloates"),
                &["interpolates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("depcrecating"),
                &["deprecating"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pscyhedelic"),
                &["psychedelic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("recognied"),
                &["recognized", "recognised"],
            ),
            (dictgen::InsensitiveStr::Ascii("makefles"), &["makefiles"]),
            (
                dictgen::InsensitiveStr::Ascii("assosciates"),
                &["associates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acummulate"),
                &["accumulate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("concatinate"),
                &["concatenate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adminsitrator"),
                &["administrator"],
            ),
            (dictgen::InsensitiveStr::Ascii("youself"), &["yourself"]),
            (dictgen::InsensitiveStr::Ascii("heavilly"), &["heavily"]),
            (dictgen::InsensitiveStr::Ascii("ocotber"), &["october"]),
            (dictgen::InsensitiveStr::Ascii("cleens"), &["cleans"]),
            (
                dictgen::InsensitiveStr::Ascii("algortithms"),
                &["algorithms"],
            ),
            (dictgen::InsensitiveStr::Ascii("safly"), &["safely"]),
            (dictgen::InsensitiveStr::Ascii("poppin"), &["popping"]),
            (
                dictgen::InsensitiveStr::Ascii("bidimentionnal"),
                &["bidimensional"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("arbiitrarily"),
                &["arbitrarily"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cuted"),
                &["cut", "cute", "cuter"],
            ),
            (dictgen::InsensitiveStr::Ascii("presest"), &["presets"]),
            (
                dictgen::InsensitiveStr::Ascii("probagation"),
                &["propagation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acceleraters"),
                &["accelerators", "accelerates"],
            ),
            (dictgen::InsensitiveStr::Ascii("broser"), &["browser"]),
            (
                dictgen::InsensitiveStr::Ascii("dependicies"),
                &["dependencies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("specififications"),
                &["specifications"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disregaring"),
                &["disregarding"],
            ),
            (dictgen::InsensitiveStr::Ascii("cerimony"), &["ceremony"]),
            (
                dictgen::InsensitiveStr::Ascii("reccursions"),
                &["recursions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("pshyciatrist"),
                &["psychiatrist"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("springfiled"),
                &["springfield"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("initiallisations"),
                &["initialisations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("produtcion"),
                &["productions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("birghtness"),
                &["brightness"],
            ),
            (dictgen::InsensitiveStr::Ascii("construcs"), &["constructs"]),
            (dictgen::InsensitiveStr::Ascii("infomr"), &["inform"]),
            (
                dictgen::InsensitiveStr::Ascii("blueberrries"),
                &["blueberries"],
            ),
            (dictgen::InsensitiveStr::Ascii("euphora"), &["euphoria"]),
            (
                dictgen::InsensitiveStr::Ascii("temerature"),
                &["temperature"],
            ),
            (dictgen::InsensitiveStr::Ascii("zionsim"), &["zionism"]),
            (dictgen::InsensitiveStr::Ascii("subarmine"), &["submarine"]),
            (dictgen::InsensitiveStr::Ascii("frequntly"), &["frequently"]),
            (dictgen::InsensitiveStr::Ascii("flavores"), &["flavors"]),
            (
                dictgen::InsensitiveStr::Ascii("matricess"),
                &["matrices", "mattresses"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("mathmatician"),
                &["mathematician"],
            ),
            (dictgen::InsensitiveStr::Ascii("confessin"), &["confession"]),
            (dictgen::InsensitiveStr::Ascii("siplify"), &["simplify"]),
            (
                dictgen::InsensitiveStr::Ascii("demoloshed"),
                &["demolished"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("siliently"),
                &["silently", "saliently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unregsiters"),
                &["unregisters"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjustmant"),
                &["adjustment"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("coppermines"),
                &["coppermine"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("converssion"),
                &["conversion"],
            ),
            (dictgen::InsensitiveStr::Ascii("decompres"), &["decompress"]),
            (dictgen::InsensitiveStr::Ascii("bwtween"), &["between"]),
            (dictgen::InsensitiveStr::Ascii("confgis"), &["configs"]),
            (dictgen::InsensitiveStr::Ascii("tenosr"), &["tensor"]),
            (
                dictgen::InsensitiveStr::Ascii("experimentaitons"),
                &["experimentations"],
            ),
            (dictgen::InsensitiveStr::Ascii("modifid"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("modififed"), &["modified"]),
            (dictgen::InsensitiveStr::Ascii("whre"), &["where"]),
            (
                dictgen::InsensitiveStr::Ascii("invalidas"),
                &["invalidates"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expresssions"),
                &["expressions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("productivy"),
                &["productivity"],
            ),
            (dictgen::InsensitiveStr::Ascii("vacummed"), &["vacuumed"]),
            (
                dictgen::InsensitiveStr::Ascii("psychopats"),
                &["psychopaths"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("priotized"),
                &["prioritized"],
            ),
            (dictgen::InsensitiveStr::Ascii("retreeves"), &["retrieves"]),
            (
                dictgen::InsensitiveStr::Ascii("transfomration"),
                &["transformation"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("advertistment"),
                &["advertisement"],
            ),
            (dictgen::InsensitiveStr::Ascii("worling"), &["working"]),
            (dictgen::InsensitiveStr::Ascii("unrelted"), &["unrelated"]),
            (
                dictgen::InsensitiveStr::Ascii("fundumentalist"),
                &["fundamentalist"],
            ),
            (dictgen::InsensitiveStr::Ascii("virginis"), &["virgins"]),
            (dictgen::InsensitiveStr::Ascii("icreases"), &["increases"]),
            (
                dictgen::InsensitiveStr::Ascii("statustics"),
                &["statistics"],
            ),
            (dictgen::InsensitiveStr::Ascii("caharcter"), &["character"]),
            (dictgen::InsensitiveStr::Ascii("preimere"), &["premiere"]),
            (
                dictgen::InsensitiveStr::Ascii("multiplpies"),
                &["multiplies"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("futhermore"),
                &["furthermore"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inconsitency"),
                &["inconsistency"],
            ),
            (dictgen::InsensitiveStr::Ascii("reenfoced"), &["reenforced"]),
            (dictgen::InsensitiveStr::Ascii("dropodwn"), &["dropdown"]),
            (
                dictgen::InsensitiveStr::Ascii("behavoiurs"),
                &["behaviours"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("encludes"),
                &["includes", "excludes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("constantivs"),
                &["constants"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyboaards"), &["keyboards"]),
            (dictgen::InsensitiveStr::Ascii("assissted"), &["assisted"]),
            (
                dictgen::InsensitiveStr::Ascii("dircts"),
                &["directs", "dicts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acomodations"),
                &["accommodations"],
            ),
            (dictgen::InsensitiveStr::Ascii("assocates"), &["associates"]),
            (dictgen::InsensitiveStr::Ascii("proir"), &["prior"]),
            (
                dictgen::InsensitiveStr::Ascii("auotcorrect"),
                &["autocorrect"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("escartments"),
                &["escarpments"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("scroolable"),
                &["scrollable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("seacher"),
                &["searcher", "teacher"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insufficently"),
                &["insufficiently"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("notoriosly"),
                &["notoriously"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirtual"), &["virtual"]),
            (
                dictgen::InsensitiveStr::Ascii("polimorphic"),
                &["polymorphic"],
            ),
            (dictgen::InsensitiveStr::Ascii("mealflur"), &["millefleur"]),
            (dictgen::InsensitiveStr::Ascii("featuerd"), &["featured"]),
            (dictgen::InsensitiveStr::Ascii("mistrow"), &["maestro"]),
            (dictgen::InsensitiveStr::Ascii("barncher"), &["brancher"]),
            (
                dictgen::InsensitiveStr::Ascii("shakespere"),
                &["shakespeare"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("introvertis"),
                &["introverts"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("canoncalizes"),
                &["canonicalizes"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("comptuations"),
                &["computations"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncompresed"),
                &["uncompressed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("convirtible"),
                &["convertible"],
            ),
            (dictgen::InsensitiveStr::Ascii("closly"), &["closely"]),
            (dictgen::InsensitiveStr::Ascii("defargkey"), &["defragkey"]),
            (dictgen::InsensitiveStr::Ascii("resullts"), &["results"]),
            (
                dictgen::InsensitiveStr::Ascii("calculalates"),
                &["calculates"],
            ),
            (dictgen::InsensitiveStr::Ascii("direcory"), &["directory"]),
            (
                dictgen::InsensitiveStr::Ascii("combinining"),
                &["combining"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("autonegoptionsotiation"),
                &["autonegotiation"],
            ),
            (dictgen::InsensitiveStr::Ascii("pulishing"), &["publishing"]),
            (dictgen::InsensitiveStr::Ascii("cerating"), &["creating"]),
            (
                dictgen::InsensitiveStr::Ascii("understantable"),
                &["understandable"],
            ),
            (dictgen::InsensitiveStr::Ascii("connecors"), &["connectors"]),
            (dictgen::InsensitiveStr::Ascii("snippes"), &["snippets"]),
            (dictgen::InsensitiveStr::Ascii("fron"), &["from", "front"]),
            (
                dictgen::InsensitiveStr::Ascii("transpanting"),
                &["transplanting"],
            ),
            (dictgen::InsensitiveStr::Ascii("duratiob"), &["duration"]),
            (
                dictgen::InsensitiveStr::Ascii("singulairty"),
                &["singularity"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("explicty"),
                &["explicitly", "explicit"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("neighbeordhoods"),
                &["neighborhoods"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("flatenning"),
                &["flattening"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("installaton"),
                &["installation"],
            ),
            (dictgen::InsensitiveStr::Ascii("retores"), &["restores"]),
            (dictgen::InsensitiveStr::Ascii("happilly"), &["happily"]),
            (
                dictgen::InsensitiveStr::Ascii("descripitions"),
                &["descriptions"],
            ),
            (dictgen::InsensitiveStr::Ascii("persew"), &["pursue"]),
            (
                dictgen::InsensitiveStr::Ascii("cheesecase"),
                &["cheesecake"],
            ),
            (dictgen::InsensitiveStr::Ascii("probbaly"), &["probably"]),
            (
                dictgen::InsensitiveStr::Ascii("instanteneous"),
                &["instantaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("colourpsace"),
                &["colourspace"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("shouder"),
                &["shoulder", "shudder"],
            ),
            (dictgen::InsensitiveStr::Ascii("nowadys"), &["nowadays"]),
            (
                dictgen::InsensitiveStr::Ascii("consctructing"),
                &["constructing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("simultaneus"),
                &["simultaneous"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("environmentality"),
                &["environmentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("insctuctions"),
                &["instructions"],
            ),
            (dictgen::InsensitiveStr::Ascii("tangets"), &["tangents"]),
            (dictgen::InsensitiveStr::Ascii("trailes"), &["trailers"]),
            (dictgen::InsensitiveStr::Ascii("allcater"), &["allocator"]),
            (
                dictgen::InsensitiveStr::Ascii("wollowing"),
                &["following", "wallowing"],
            ),
            (dictgen::InsensitiveStr::Ascii("thoguhts"), &["thoughts"]),
            (dictgen::InsensitiveStr::Ascii("timzezone"), &["timezone"]),
            (dictgen::InsensitiveStr::Ascii("promblem"), &["problem"]),
            (dictgen::InsensitiveStr::Ascii("stainlees"), &["stainless"]),
            (
                dictgen::InsensitiveStr::Ascii("explaintory"),
                &["explanatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("emal"), &["email"]),
            (
                dictgen::InsensitiveStr::Ascii("bulletpoof"),
                &["bulletproof"],
            ),
            (dictgen::InsensitiveStr::Ascii("matieral"), &["material"]),
            (
                dictgen::InsensitiveStr::Ascii("standarising"),
                &["standardising"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("symobolical"),
                &["symbolical"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("unsespecting"),
                &["unsuspecting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("newspappers"),
                &["newspapers"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("attribubtes"),
                &["attributes"],
            ),
            (dictgen::InsensitiveStr::Ascii("tronado"), &["tornado"]),
            (dictgen::InsensitiveStr::Ascii("creaeted"), &["created"]),
            (
                dictgen::InsensitiveStr::Ascii("experimntally"),
                &["experimentally"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("manufacterurs"),
                &["manufactures"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("narssicistic"),
                &["narcissistic"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("represends"),
                &["represents"],
            ),
            (dictgen::InsensitiveStr::Ascii("keyworkd"), &["keyword"]),
            (dictgen::InsensitiveStr::Ascii("raseon"), &["reason"]),
            (
                dictgen::InsensitiveStr::Ascii("ukraininans"),
                &["ukrainians"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("uncomfortablly"),
                &["uncomfortably"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("bootleader"),
                &["bootloader"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("cahins"),
                &["chains", "cabins"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("conqouring"),
                &["conquering"],
            ),
            (dictgen::InsensitiveStr::Ascii("wirhout"), &["without"]),
            (
                dictgen::InsensitiveStr::Ascii("acquaintences"),
                &["acquaintances"],
            ),
            (dictgen::InsensitiveStr::Ascii("causacian"), &["caucasian"]),
            (
                dictgen::InsensitiveStr::Ascii("ceremonias"),
                &["ceremonies"],
            ),
            (dictgen::InsensitiveStr::Ascii("deprected"), &["deprecated"]),
            (
                dictgen::InsensitiveStr::Ascii("unsychronize"),
                &["unsynchronize"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expectaton"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("vaule"), &["value"]),
            (dictgen::InsensitiveStr::Ascii("deprssed"), &["depressed"]),
            (dictgen::InsensitiveStr::Ascii("subtruct"), &["subtract"]),
            (
                dictgen::InsensitiveStr::Ascii("abstrctions"),
                &["abstractions"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("parlaiment"),
                &["parliament"],
            ),
            (dictgen::InsensitiveStr::Ascii("treees"), &["trees"]),
            (dictgen::InsensitiveStr::Ascii("reseverd"), &["reserved"]),
            (dictgen::InsensitiveStr::Ascii("optomised"), &["optimised"]),
            (
                dictgen::InsensitiveStr::Ascii("republicants"),
                &["republicans"],
            ),
            (dictgen::InsensitiveStr::Ascii("currebtly"), &["currently"]),
            (dictgen::InsensitiveStr::Ascii("commedic"), &["comedic"]),
            (
                dictgen::InsensitiveStr::Ascii("speficialleid"),
                &["specified"],
            ),
            (dictgen::InsensitiveStr::Ascii("stabilite"), &["stabilize"]),
            (
                dictgen::InsensitiveStr::Ascii("descandent"),
                &["descendent"],
            ),
            (dictgen::InsensitiveStr::Ascii("mobify"), &["modify"]),
            (
                dictgen::InsensitiveStr::Ascii("implemetning"),
                &["implementing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("adjasencies"),
                &["adjacencies"],
            ),
            (dictgen::InsensitiveStr::Ascii("swatsika"), &["swastika"]),
            (dictgen::InsensitiveStr::Ascii("chaneged"), &["changed"]),
            (dictgen::InsensitiveStr::Ascii("lanauges"), &["languages"]),
            (
                dictgen::InsensitiveStr::Ascii("downlaodable"),
                &["downloadable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("oppertunities"),
                &["opportunities"],
            ),
            (dictgen::InsensitiveStr::Ascii("tructured"), &["structured"]),
            (
                dictgen::InsensitiveStr::Ascii("snowfalling"),
                &["snowballing"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("dictionarty"),
                &["dictionary"],
            ),
            (dictgen::InsensitiveStr::Ascii("hungray"), &["hungary"]),
            (
                dictgen::InsensitiveStr::Ascii("suprisingly"),
                &["surprisingly"],
            ),
            (dictgen::InsensitiveStr::Ascii("brighened"), &["brightened"]),
            (
                dictgen::InsensitiveStr::Ascii("vulnerabuilitis"),
                &["vulnerabilities"],
            ),
            (dictgen::InsensitiveStr::Ascii("brigth"), &["bright"]),
            (dictgen::InsensitiveStr::Ascii("kinteic"), &["kinetic"]),
            (dictgen::InsensitiveStr::Ascii("timmestamp"), &["timestamp"]),
            (dictgen::InsensitiveStr::Ascii("interactes"), &["interacts"]),
            (
                dictgen::InsensitiveStr::Ascii("traumetized"),
                &["traumatized"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("negosiable"),
                &["negotiable"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapeared"),
                &["disappeared"],
            ),
            (dictgen::InsensitiveStr::Ascii("arragning"), &["arranging"]),
            (dictgen::InsensitiveStr::Ascii("sokets"), &["sockets"]),
            (dictgen::InsensitiveStr::Ascii("emcas"), &["emacs"]),
            (dictgen::InsensitiveStr::Ascii("certainily"), &["certainly"]),
            (
                dictgen::InsensitiveStr::Ascii("derogatary"),
                &["derogatory"],
            ),
            (dictgen::InsensitiveStr::Ascii("fomatting"), &["formatting"]),
            (dictgen::InsensitiveStr::Ascii("teh"), &["the"]),
            (dictgen::InsensitiveStr::Ascii("thar"), &["than", "that"]),
            (dictgen::InsensitiveStr::Ascii("relased"), &["released"]),
            (
                dictgen::InsensitiveStr::Ascii("contatenated"),
                &["concatenated"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("investagators"),
                &["investigators"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("espectation"),
                &["expectation"],
            ),
            (dictgen::InsensitiveStr::Ascii("mangement"), &["management"]),
            (dictgen::InsensitiveStr::Ascii("contetn"), &["content"]),
            (
                dictgen::InsensitiveStr::Ascii("healthyest"),
                &["healthiest"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("prosicution"),
                &["prosecution"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("invicibility"),
                &["invisibility", "invincibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("reporitories"),
                &["repositories"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("availbability"),
                &["availability"],
            ),
            (dictgen::InsensitiveStr::Ascii("quering"), &["querying"]),
            (dictgen::InsensitiveStr::Ascii("wrold"), &["world"]),
            (dictgen::InsensitiveStr::Ascii("argemnts"), &["arguments"]),
            (dictgen::InsensitiveStr::Ascii("mesuring"), &["measuring"]),
            (
                dictgen::InsensitiveStr::Ascii("lightwright"),
                &["lightweight"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("predicat"),
                &["predict", "predicate"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("forgivenss"),
                &["forgiveness"],
            ),
            (dictgen::InsensitiveStr::Ascii("themplate"), &["template"]),
            (
                dictgen::InsensitiveStr::Ascii("vulernability"),
                &["vulnerability"],
            ),
            (dictgen::InsensitiveStr::Ascii("usinf"), &["using"]),
            (dictgen::InsensitiveStr::Ascii("critieria"), &["criteria"]),
            (
                dictgen::InsensitiveStr::Ascii("precisi"),
                &["precision", "precise"],
            ),
            (dictgen::InsensitiveStr::Ascii("cyrilic"), &["cyrillic"]),
            (dictgen::InsensitiveStr::Ascii("porving"), &["proving"]),
            (dictgen::InsensitiveStr::Ascii("clousre"), &["closure"]),
            (
                dictgen::InsensitiveStr::Ascii("construters"),
                &["constructors"],
            ),
            (dictgen::InsensitiveStr::Ascii("deveoped"), &["developed"]),
            (
                dictgen::InsensitiveStr::Ascii("defeind"),
                &["defined", "defend"],
            ),
            (dictgen::InsensitiveStr::Ascii("asserttion"), &["assertion"]),
            (
                dictgen::InsensitiveStr::Ascii("adjudivate"),
                &["adjudicate"],
            ),
            (dictgen::InsensitiveStr::Ascii("civillians"), &["civilians"]),
            (dictgen::InsensitiveStr::Ascii("errornoeus"), &["erroneous"]),
            (dictgen::InsensitiveStr::Ascii("submiting"), &["submitting"]),
            (dictgen::InsensitiveStr::Ascii("collpsed"), &["collapsed"]),
            (
                dictgen::InsensitiveStr::Ascii("performantes"),
                &["performances"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("infrastucture"),
                &["infrastructure"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("disapplins"),
                &["disciplines"],
            ),
            (dictgen::InsensitiveStr::Ascii("intstructs"), &["instructs"]),
            (
                dictgen::InsensitiveStr::Ascii("limititation"),
                &["limitation"],
            ),
            (dictgen::InsensitiveStr::Ascii("themslef"), &["themself"]),
            (
                dictgen::InsensitiveStr::Ascii("condiscending"),
                &["condescending"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("envrionmentally"),
                &["environmentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("felxible"), &["flexible"]),
            (
                dictgen::InsensitiveStr::Ascii("communikay"),
                &["communiqué"],
            ),
            (dictgen::InsensitiveStr::Ascii("fnd"), &["find"]),
            (dictgen::InsensitiveStr::Ascii("aferwards"), &["afterwards"]),
            (
                dictgen::InsensitiveStr::Ascii("possibiility"),
                &["possibility"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("enforement"),
                &["enforcement"],
            ),
            (dictgen::InsensitiveStr::Ascii("prvide"), &["provide"]),
            (
                dictgen::InsensitiveStr::Ascii("tempraraily"),
                &["temporarily"],
            ),
            (dictgen::InsensitiveStr::Ascii("cencel"), &["cancel"]),
            (
                dictgen::InsensitiveStr::Ascii("guarentieing"),
                &["guaranteeing"],
            ),
            (dictgen::InsensitiveStr::Ascii("primaray"), &["primary"]),
            (dictgen::InsensitiveStr::Ascii("otimize"), &["optimize"]),
            (dictgen::InsensitiveStr::Ascii("parareter"), &["parameter"]),
            (dictgen::InsensitiveStr::Ascii("augmnets"), &["augments"]),
            (dictgen::InsensitiveStr::Ascii("aboslutes"), &["absolutes"]),
            (dictgen::InsensitiveStr::Ascii("machien"), &["machine"]),
            (
                dictgen::InsensitiveStr::Ascii("heuristiscs"),
                &["heuristics"],
            ),
            (dictgen::InsensitiveStr::Ascii("misplacin"), &["misplacing"]),
            (
                dictgen::InsensitiveStr::Ascii("diposed"),
                &["disposed", "deposed"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("phsyicists"),
                &["physicists"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("inspeccion"),
                &["inspection"],
            ),
            (dictgen::InsensitiveStr::Ascii("romanain"), &["romanian"]),
            (dictgen::InsensitiveStr::Ascii("journalis"), &["journals"]),
            (dictgen::InsensitiveStr::Ascii("disprovin"), &["disproving"]),
            (
                dictgen::InsensitiveStr::Ascii("moisturizng"),
                &["moisturizing"],
            ),
            (dictgen::InsensitiveStr::Ascii("descripto"), &["descriptor"]),
            (dictgen::InsensitiveStr::Ascii("nostriles"), &["nostrils"]),
            (
                dictgen::InsensitiveStr::Ascii("presistent"),
                &["persistent"],
            ),
            (dictgen::InsensitiveStr::Ascii("tolen"), &["token"]),
            (
                dictgen::InsensitiveStr::Ascii("usupported"),
                &["unsupported"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("accidantly"),
                &["accidentally"],
            ),
            (dictgen::InsensitiveStr::Ascii("hindusim"), &["hinduism"]),
            (dictgen::InsensitiveStr::Ascii("paitents"), &["patients"]),
            (dictgen::InsensitiveStr::Ascii("boraden"), &["broaden"]),
            (dictgen::InsensitiveStr::Ascii("primative"), &["primitive"]),
            (dictgen::InsensitiveStr::Ascii("crystallis"), &["crystals"]),
            (
                dictgen::InsensitiveStr::Ascii("unexecpted"),
                &["unexpected"],
            ),
            (dictgen::InsensitiveStr::Ascii("compsable"), &["composable"]),
            (
                dictgen::InsensitiveStr::Ascii("overpirced"),
                &["overpriced"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("gravitationnal"),
                &["gravitational"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("suplanting"),
                &["supplanting"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("expension"),
                &["expansion", "extension"],
            ),
            (dictgen::InsensitiveStr::Ascii("cylcists"), &["cyclists"]),
            (
                dictgen::InsensitiveStr::Ascii("programatic"),
                &["programmatic"],
            ),
            (dictgen::InsensitiveStr::Ascii("souvineers"), &["souvenirs"]),
            (dictgen::InsensitiveStr::Ascii("theoeries"), &["theories"]),
            (dictgen::InsensitiveStr::Ascii("genertor"), &["generator"]),
            (
                dictgen::InsensitiveStr::Ascii("experimtens"),
                &["experiments"],
            ),
            (dictgen::InsensitiveStr::Ascii("adres"), &["address"]),
            (
                dictgen::InsensitiveStr::Ascii("incoprorating"),
                &["incorporating"],
            ),
            (dictgen::InsensitiveStr::Ascii("loaclity"), &["locality"]),
            (
                dictgen::InsensitiveStr::Ascii("collaterial"),
                &["collateral"],
            ),
            (dictgen::InsensitiveStr::Ascii("secuirty"), &["security"]),
            (
                dictgen::InsensitiveStr::Ascii("overzelously"),
                &["overzealously"],
            ),
            (
                dictgen::InsensitiveStr::Ascii("acquaintace"),
                &["acquaintance"],
            ),
            (dictgen::InsensitiveStr::Ascii("patterno"), &["patterson"]),
            (dictgen::InsensitiveStr::Ascii("chedulers"), &["schedulers"]),
            (dictgen::InsensitiveStr::Ascii("assitive"), &["assistive"]),
            (
                dictgen::InsensitiveStr::Ascii("hueristics"),
                &["heuristics"],
            ),
        ],
    },
    range: 2..=34,
};
